/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : agent.c
 *
 * Description  : Agile Mirror (AM) agent daemon.
 *
 * The AM agent is one of the 3 components of the Agile Mirror:
 * - AM controller (on SSC/external/???)
 * - AM agent (on FP on N-NIM)
 * - AM data mover (on TxRx on N-NIM)
 *
 * There is only one instance of the AM agent daemon .
 * It is running from NFX startup till NFX shutdown.
 * It receives all the messages, callbacks for all the AM agent instances.
 * It is responsible to dispatch them to the correct instances.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 02-OCT-2001
 * Re-designed  : 25-JAN-2002
 *
 *------------------------------------------------------------------------------
 */

#ifndef SSC

#include "agent.h"
#include "exports/hash-api.h"
//#define SHOULD_USE_COOKIE 1

/*
 * Global variables.
 */

/*
 * My slot and CPU identifiers..
 */
extern uint32 nfxMySlotId;
extern uint32 nfxMyCPUId;

/*
 * EEE address port of the AM controler.
 */
uint32   ctrlAddr;

/*
 * My own application id and EEE address port.
 */
int32    amAppId;
uint32   amAddr;

/*
 * EEE address port of the data mover.
 */
uint32   datamoverAddr;

/*
 * EEE address port of the EFS.
 */
uint32   efsAddr;

/*
 * EEE address port of the EVM (FP).
 */
uint32   evmAddr;

/*
 * EEE address port of the EVM SSC.
 */
uint32   evmSscAddr;

/*
 * DM hash table ID
 */
address_t     dmHashID;

char amNewSnapName[] = {"AM_SS_"};

char eventStr[256];

/*
 * Local functions
 */

/*
 *------------------------------------------------------------------------------
 * Name         : am_free_control_block
 *
 * Description  : free control block
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 28-JAN-2002
 *
 *------------------------------------------------------------------------------
 */

void am_free_control_block(am_agCb_t *agtCb)
{
    AM_DBG_PRINT(("am_free_control_block..."));

    if (agtCb->cbType == AM_SOURCE_CB)
        am_free_block_list(agtCb);
    else
        am_free_queued_requests(agtCb);

    am_eee_ramDealloc(agtCb);

    AM_DBG_PRINT(("...am_free_control_block"));
    return;
}


/*
 *------------------------------------------------------------------------------
 * Name         : am_getHashKey
 *
 * Description  : 
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 08-FEB-2002
 *
 *------------------------------------------------------------------------------
 */
uint32 am_getHashKey(void *key)
{
    dm_hdl_key_t *dmKey  = (dm_hdl_key_t *)key;
    uint32       hashKey;

    AM_DBG_PRINT(("am_getHashKey..."));

    if (dmKey->hdlType != AM_HDL_KEY_DM_HDL)
    {
        AM_DBG_PRINT(("Key allowed only for DM Handle!!"));
        hashKey =0;
    }
    else
        hashKey = ((uint32)dmKey->u.hdl)%AM_MAX_HASH_INDEXS;

    AM_DBG_PRINT(("...am_getHashKey"));

    return hashKey;
}


/*
 *------------------------------------------------------------------------------
 * Name         : am_compareHashKey
 *
 * Description  : 
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 08-FEB-2002
 *
 *------------------------------------------------------------------------------
 */
int32 am_compareHashKey(void *ht_data, void *search_data)
{
    am_agCb_t           *agtCb = (am_agCb_t *)ht_data;
    dm_hdl_key_t        *searchKey = (dm_hdl_key_t *)search_data;
    int32               rc = 1;
    uint64              searchHdl = 0;

    AM_DBG_PRINT(("am_compareHashKey..."));

    if (searchKey->hdlType == AM_HDL_KEY_AM_CB)
        searchHdl = searchKey->u.agt->dmHdl;
    else
        searchHdl = searchKey->u.hdl;

    if (agtCb->dmHdl == searchHdl)
    {
        AM_DBG_PRINT(("Found matching DM key"));
        rc = 0;
    }

    AM_DBG_PRINT(("...am_compareHashKey"));

    return rc;
}


/*
 *------------------------------------------------------------------------------
 * Name         : am_releaseHashObj
 *
 * Description  : 
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 08-FEB-2002
 *
 *------------------------------------------------------------------------------
 */
void am_releaseHashObj(void *data)
{
    AM_DBG_PRINT(("am_releaseHashObj..."));

    /*
     * Nothing to do here
     */

    AM_DBG_PRINT(("...am_releaseHashObj"));

    return;
}

/*
 * Global functions.
 */

/*------------------------------------------------------------------------------
 * Name         : am_initTgt
 *
 * Description  : At the time of start up let us bind to the 
 *				  well known port as server to accept any mirror
 *				  related request.
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 02/12/02
 *
 *------------------------------------------------------------------------------
*/
void am_initTgt()
{
    am_agCb_t          	*newCb = NULL;
    dm_listenReq_t     	dmReq;
    uint64       		appHdl = 0;
    int32        		rc;

	AM_DBG_PRINT(("> am_initTgt send DM_LISTEN_REQ"));

    /*
     * Get free control block. Index 0 is always for daemon
     */
    newCb = (am_agCb_t *) am_eee_ramAlloc(sizeof(am_agCb_t));
    NFX_ASSERT(newCb != NULL);
    bzero (newCb, sizeof(am_agCb_t));
    appHdl = (uint64)newCb;

    newCb->cbType     = AM_SOURCE_CB;
    newCb->ctrlHdl    = 0;

	/*
	 * If we bind for a port on address 0.0.0.0 TPL will 
	 * forward anything rcvd on the registered port to us
	 */
    newCb->lclAddr    = 0;  			// IP IF on which to listen
    newCb->lclPort    = htons(11000);   // server/target port

    bzero (&dmReq, sizeof(dm_listenReq_t));
    dmReq.hdr.msgType = DM_LISTEN_REQ;
    dmReq.hdr.appHdl  = appHdl;         // control block
    dmReq.hdr.dmHdl   = 0;              // 0 is for DM daemon
    dmReq.protocol    = TPL_PROT_TCP;   // TCP protocol always
    dmReq.lclAddr     = newCb->lclAddr; // Ip address
    dmReq.lclPort     = newCb->lclPort; // well known port

    /* 
	 * Send message to the data mover 
	 */
    rc = eee_sendMessage(&dmReq, sizeof(dm_listenReq_t),
                                 datamoverAddr, amAddr);
    NFX_ASSERT(rc == NFX_OK);

	AM_DBG_PRINT(("< ... am_initTgt"));
}

/*
 *------------------------------------------------------------------------------
 * Name         : am_ag_init
 *
 * Description  : Initialization of the AM agent daemon.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 02-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
void am_ag_init(void)
{
    int32   rc;
    
    AM_DBG_PRINT(("am_ag_init..."));
    
    /*
     * My application id.
     */
    amAppId = eee_allocateApplicationId("am");
    NFX_ASSERT(amAppId != EEE_ERR_INVALID_APP_NAME);

    /*
     * My EEE address,
     */
    amAddr = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
                                       nfxMyCPUId,
                                       amAppId,
                                       0, 0, 0);

    /*
     * EEE address of AM controller
     */
    ctrlAddr = PACK_UNICAST_PORT_ADDRESS(0,
                                        NFX_SSC_UNIX,
                                        eee_getApplicationId("am"),
                                        EEE_MGMT, NFX_COS_HIGH, PORT_BEST_FIT);

    /*
     * Datamover EEE address
     */
    datamoverAddr = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
                                              NFX_NFP_TXRX_ANY,
                                              eee_getApplicationId("dm"),
                                              0, 0, 0);

    /*
     * EEE address port of the EFS.
     */
    efsAddr = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
                                        nfxMyCPUId,
                                        eee_getApplicationId("efs"),
                                        0, 0, 0);

    /*
     * EEE address port of the EVM (FP).
     */
    evmAddr = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
                                        nfxMyCPUId,
                                        eee_getApplicationId("evm"),
                                        0, 0, 0);

    /*
     * EEE address port of the EVM SSC.
     */
    evmSscAddr = PACK_UNICAST_PORT_ADDRESS(0,
                                        NFX_SSC_UNIX,
                                        eee_getApplicationId("evm"),
                                        EEE_MGMT, NFX_COS_HIGH, PORT_BEST_FIT);

    /*
     * Register to receive messages.
     */
    rc = eee_registerReceiveFunc(amAppId,
                                 am_ag_rcvMsg,
                                 am_ag_rcvMcastMsg);
    NFX_ASSERT(rc == NFX_OK);

    {
        AM_FCT_PRINT(("  nfxMySlotId = %d", nfxMySlotId));
        AM_FCT_PRINT(("   nfxMyCPUId = %d", nfxMyCPUId));
        AM_FCT_PRINT(("      amAppId = %d", amAppId));
        AM_FCT_PRINT(("       amAddr = 0x%x", amAddr));
        AM_FCT_PRINT(("     ctrlAddr = 0x%x", ctrlAddr));
        AM_FCT_PRINT(("datamoverAddr = 0x%x", datamoverAddr));
        AM_FCT_PRINT(("      efsAddr = 0x%x", efsAddr));
        AM_FCT_PRINT(("      evmAddr = 0x%x", evmAddr));
    }

    dmHashID = hash_create(AM_MAX_HASH_INDEXS, am_getHashKey,
                           am_compareHashKey, am_releaseHashObj);
    if (dmHashID == 0)
    {
        AM_DBG_PRINT(("Unable to create hash table"));
        am_ag_cleanup();
        AM_DBG_PRINT(("Agile Mirror not initialized!!!"));
    }

	/*
	 * Initialize the Tgt to start listening
	 * on the well known port
	 */
	am_initTgt();


    AM_DBG_PRINT(("...am_ag_init"));
    return;
}


/*
 *------------------------------------------------------------------------------
 * Name         : am_ag_cleanup
 *
 * Description  : Clean up the AM agent daemon.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
void am_ag_cleanup(void)
{
    //int32   rc;
    
    AM_DBG_PRINT(("am_ag_cleanup..."));
    
    /* Unregister to receive messages.
     */
    eee_unregisterReceiveFunc(amAppId);

    /*
     * Deallocate my application id.
     */
    //rc = eee_deallocateApplicationId(amAppId); - not yet implemented!!!
    //NFX_ASSERT(rc == NFX_OK);

    if (dmHashID != 0)
        hash_destroy(dmHashID);
    
    AM_DBG_PRINT(("...am_ag_cleanup"));
    return;
}


/*
 *------------------------------------------------------------------------------
 * Name         : am_post_event_abort
 *
 * Description  : Post and abort event to the state machine.
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 12-FEB-2002
 *
 *------------------------------------------------------------------------------
 */
void am_post_event_abort(am_agCb_t *agtCb, char *msg)
{
    esm_event_t    newEvent;
    am_rc_t        rc;

    AM_DBG_PRINT(("am_post_event_abort..."));

    AM_DBG_PRINT(("%s : %s", M_GET_CB_TYPE_STRING(agtCb), msg));

    bzero(&newEvent, sizeof(esm_event_t));
    newEvent.event   = M_AM_MIRROR_ABORT_EVT(agtCb);
    newEvent.p1.u.dw = 0;
    newEvent.p2.u.dw = 0;

	if (newEvent.event == SOURCE_EVENT_MIRROR_ABORT)
		strcpy(eventStr, "SOURCE_EVENT_MIRROR_ABORT");
	else 
		strcpy(eventStr, "TARGET_EVENT_MIRROR_ABORT");

    AM_FCT_PRINT(("Posting %s %d to %d SM", \
                       eventStr, newEvent.event, agtCb->amSmHdl));
    rc = esm_deliverEvent(agtCb->amSmHdl, &newEvent);
    NFX_ASSERT(rc == NFX_OK);

    AM_DBG_PRINT(("...am_post_event_abort"));
    return;
}



/*
 *------------------------------------------------------------------------------
 * Name         : am_ag_rcvMsg()
 *
 * Description  : Process received message.
 *
 * - put it in a contiguous buffer
 * - forward it to the right instance of the AM agent
 * - free it
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 02-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
void am_ag_rcvMsg(eee_descPtr_t  edesc)
{
    char         *msgBuf = NULL;
    char         *evtMsgBuf = NULL;
    uint32       msgLen = 0;
    uint32       bufIsInSharedMemory;
    int32        *msgType = NULL;
    esm_event_t  esmEvent;
    int32        rc;
    uint64       appHdl = 0;
    dm_hdl_key_t insertKey;
    dm_hdl_key_t dataKey;
    dm_hdl_key_t searchKey;
	uint32		 srcPort;

    AM_DBG_PRINT(("am_ag_rcvMsg..."));

    /*
     * Get the message.
     * - calculate its length
     * - if complete message is in only one buffer
     *   - keep it in the system buffer
     *   - free the EEE descriptor
     * - if message is in several buffers
     *   - copy it into a contiguous malloc buffer
     *   - free the received packet (EEE descriptor + buffers).
     */
    {
        int32   i, numBufs;

		srcPort = edesc->hdr.src_port;

        numBufs = EEE_GET_NUM_BUFS(edesc);

        for (i = 0; i < numBufs; i++)
        {
            msgLen += edesc->bd[i+1].len;   // first buffer not used
        } // for

        if (numBufs == 1)
        {
            // AM_DBG_PRINT(("Buffer in SHM-RAM"));
            bufIsInSharedMemory = TRUE;
            msgBuf = edesc->bd[1].buf;      // first buffer not used
            eee_deallocateDesc(edesc);
        }
        else
        {
            //AM_DBG_PRINT(("Buffer in RAM"));
            bufIsInSharedMemory = FALSE;
            msgBuf = (char *)am_eee_ramAlloc(msgLen);
            NFX_ASSERT(msgBuf != NULL);
            rc = eee_copyPacketToLocalBuffer(msgBuf, edesc, msgLen, 0);
            NFX_ASSERT(rc == msgLen);
            eee_freePacket(edesc);
        }
    }

    /*
     * Process the message according to its type.
     */
    bzero(&esmEvent, sizeof(esm_event_t));
    msgType = (int32 *)msgBuf;
    AM_DBG_PRINT(("am_ag_rcvMsg msg type is %d", *msgType));

    switch (*msgType)
    {
        /*
         * Message received from the source AM controller
         * ----------------------------------------------
         */
        case AM_AGENT_LISTEN_REQ:
         // AM_AGENT_LISTEN_REQ

        case AM_START_MIRROR_REQ:
        {
            /*
             * On the source the am controller starts a mirror session
             * based on time trigger or operator initiated request.
             */
            am_startMirrorReq_t   *req   = (am_startMirrorReq_t *)msgBuf;
            am_agCb_t             *newCb = NULL;
            am_startMirrorRsp_t    amRsp;

            AM_FCT_PRINT(("AM_START_MIRROR_REQ Rmt 0x%llx %d Lcl 0x%llx %d",
					req->rmtAddr, ntohs(req->rmtPort), req->lclAddr, 0));

            /*
             * Get free control block.
             */

            newCb    = am_eee_ramAlloc(sizeof(am_agCb_t));
            NFX_ASSERT(newCb != NULL);
            bzero (newCb, sizeof(am_agCb_t));

            newCb->ctrlHdl  = req->hdr.ctrlHdl;
            newCb->cbType   = AM_SOURCE_CB;
            newCb->lclAddr  = req->lclAddr;  // IP IF on which to listen
            newCb->lclPort  = 0;             // Will be assigned by DM
            newCb->rmtAddr  = req->rmtAddr;  // IP IF on which to listen
            newCb->rmtPort  = req->rmtPort;  // Will be assigned by DM
#ifndef INCLUDE_FS_ON_SOURCE
            newCb->nextExt  = TRUE;
#endif

            memcpy(&newCb->amVolId,&req->amVolId,sizeof(evm_volId_t));
            newCb->maxDataRate = req->maxDataRate; // Max data rate
			strcpy(newCb->tgtFsysName, req->tgtFsysName);

            sprintf(newCb->newSnap, "%s%s_%03d", amNewSnapName, 
					req->amName, req->snapSeqSuffix);

            AM_FCT_PRINT(("START MIRROR %s", newCb->newSnap));
            printf("\n\nSTART MIRROR %s\n", newCb->newSnap);

            rc = source_sm_create(&newCb->amSmHdl, newCb);
			if (rc == NFX_OK)
			{
            	esmEvent.event = SOURCE_EVENT_CONN_REQ;
				strcpy(eventStr, "SOURCE_EVENT_CONN_REQ");
            	esmEvent.p1.u.dw = 0;
            	esmEvent.p2.u.dw = 0;
			
            	AM_FCT_PRINT(("Posting %s %d to %d SM", \
                           eventStr, esmEvent.event, newCb->amSmHdl));
            	rc = esm_deliverEvent(newCb->amSmHdl, &esmEvent);
				if (rc != NFX_OK)
				{
					AM_ERR_PRINT(("Could not start mirror operation for %s",
								req->amName));
				}
			}

            /*
             * Send the start mirror response
             */

            amRsp.hdr.msgType   = AM_START_MIRROR_RSP; // Type of message
			amRsp.hdr.ctrlHdl	= req->hdr.ctrlHdl;
			amRsp.hdr.amApplHdl	= (uint64) newCb;
            strcpy(amRsp.amName, req->amName);
            amRsp.result        = rc; 

            rc = eee_sendMessage(&amRsp,
                                 sizeof(am_startMirrorRsp_t),
                                 srcPort,
                                 amAddr);

            NFX_ASSERT(rc == NFX_OK);
            break;
        } // AM_START_MIRROR_REQ

        case AM_AGENT_ABORT_REQ:
        {
            /*
             * AM controller or the operator wants to stop the mirror session
             */
            am_abortAgentReq_t    *req = (am_abortAgentReq_t *)msgBuf;
            am_agCb_t             *agtCb = NULL;
            am_abortAgentRsp_t    amRsp;

            AM_FCT_PRINT(("AM_AGENT_ABORT_REQ"));

            agtCb = (am_agCb_t *)req->hdr.amApplHdl;
            if (agtCb != NULL       &&
                agtCb->amSmHdl != 0 &&
                agtCb->amVolId.whole_thing == req->amVolId.whole_thing)
            {
                agtCb->ctrlHdl = req->hdr.ctrlHdl;  /* In case it has changed */
                esmEvent.event = M_AM_MIRROR_ABORT_EVT(agtCb);
				if (esmEvent.event == SOURCE_EVENT_MIRROR_ABORT)
					strcpy(eventStr, "SOURCE_EVENT_MIRROR_ABORT");
				else 
					strcpy(eventStr, "TARGET_EVENT_MIRROR_ABORT");
                esmEvent.p1.u.dw = 0;
                esmEvent.p2.u.dw = 0;
	
	            AM_FCT_PRINT(("Posting %s %d to %d SM", \
                           eventStr, esmEvent.event, agtCb->amSmHdl));

                rc = esm_deliverEvent(agtCb->amSmHdl, &esmEvent);
                if (rc != NFX_OK)
				{
					AM_ERR_PRINT(("Mirror Abort request for %s failed",
								req->amName));
				}
            }

            /*
             * Send abort response
             */

            amRsp.hdr.msgType   = AM_AGENT_ABORT_RSP; // Type of message
            amRsp.hdr.amApplHdl = req->hdr.amApplHdl;  // CB index
            amRsp.hdr.ctrlHdl   = req->hdr.ctrlHdl;   // return to request
            memcpy(&amRsp.amVolId,&req->amVolId,sizeof(evm_volId_t));
            strcpy(amRsp.amName, req->amName);

			if (rc == NFX_OK)
            	amRsp.result = NFX_OK;
			else
				amRsp.result = NFX_ERR;

            rc = eee_sendMessage(&amRsp,
                                 sizeof(am_abortAgentRsp_t),
                                 srcPort,
                                 amAddr);
            NFX_ASSERT(rc == NFX_OK);
            break;
        } // AM_AGENT_ABORT_REQ

        /*
         * Message interface between AM and DM
         * -----------------------------------
         */
        case DM_LISTEN_RSP:
        {
            dm_listenRsp_t    *rsp = (dm_listenRsp_t *)msgBuf;
            am_listenRsp_t    amRsp;
            am_agCb_t         *agtCb = NULL;

            AM_FCT_PRINT(("DM_LISTEN_RSP"));

            AM_DBG_PRINT(("AM handle : %lld", rsp->hdr.appHdl));

            appHdl = rsp->hdr.appHdl;
            agtCb = (am_agCb_t *)appHdl;
            bzero (&amRsp, sizeof(am_listenRsp_t));
            amRsp.hdr.msgType   = AM_AGENT_LISTEN_RSP; // Type of message

            if (agtCb != NULL)
                amRsp.hdr.ctrlHdl  = agtCb->ctrlHdl;   // controller handle
            else
                amRsp.hdr.ctrlHdl  = 0;               // No controller handle

            if (rsp->result != NFX_OK)
                amRsp.result       = NFX_ERR;   
            else
                amRsp.result       = NFX_OK;   

            rc = eee_sendMessage(&amRsp,
                                 sizeof(am_listenRsp_t),
                                 ctrlAddr,
                                 amAddr);
            NFX_ASSERT(rc == NFX_OK);

            /*
             * Since we are done with listening process. Free the control blk
             */
            if(agtCb)
                am_eee_ramDealloc(agtCb);
			AM_DBG_PRINT(("Done here after DM_LISTEN_RSP"));
            break;
        } // DM_LISTEN_RSP

        case DM_ACCEPT_REQ:
        {
            /*
             * On target this is the daemon part
             */
            dm_acceptReq_t        *req = (dm_acceptReq_t *)msgBuf;
            am_agCb_t             *newCb = NULL;

            AM_FCT_PRINT(("DM_ACCEPT_REQ"));

            /*
             * Get new control block
             */
            newCb = (am_agCb_t *)am_eee_ramAlloc(sizeof(am_agCb_t));
            NFX_ASSERT (newCb != NULL);
            bzero (newCb, sizeof(am_agCb_t));

            newCb->dmHdl   = req->hdr.dmHdl;
            newCb->cbType  = AM_TARGET_CB;
            newCb->lclAddr = req->lclAddr;
            newCb->lclPort = req->lclPort;
            newCb->rmtAddr = req->rmtAddr;
            newCb->rmtPort = req->rmtPort;
#ifndef INCLUDE_FS_ON_SOURCE
            newCb->nextExt  = TRUE;          // Does not matter
#endif
            rc = target_sm_create(&newCb->amSmHdl, newCb);
            NFX_ASSERT(rc == NFX_OK);

            insertKey.hdlType = AM_HDL_KEY_DM_HDL;
            insertKey.u.hdl   = req->hdr.dmHdl;
            rc = hash_insert(dmHashID, &insertKey, newCb);
            NFX_ASSERT(rc == NFX_OK);

            esmEvent.event = TARGET_EVENT_AUTH_CONN_REQ;
			strcpy(eventStr, "TARGET_EVENT_AUTH_CONN_REQ");
            esmEvent.p1.u.dw = 0;
            esmEvent.p2.u.dw = 0;

            AM_FCT_PRINT(("Posting %s %d to %d SM", \
                           eventStr, esmEvent.event, newCb->amSmHdl));
            rc = esm_deliverEvent(newCb->amSmHdl, &esmEvent);
            NFX_ASSERT(rc == NFX_OK);
            break;
        } // DM_ACCEPT_REQ

        case DM_OPEN_RSP:
        {
            /*
             * On source. This is the response from DM when TCP conn.
             * is established with the target
             */
            dm_openRsp_t          *rsp = (dm_openRsp_t *)msgBuf;
            am_agCb_t             *agtCb = NULL;

            AM_FCT_PRINT(("DM_OPEN_RSP"));

            appHdl = rsp->hdr.appHdl;
            if (appHdl != 0)
                agtCb = (am_agCb_t *)appHdl;
            else
            {
                AM_DBG_PRINT(("DM_OPEN_RSP lookup 0x%llx", rsp->hdr.appHdl));
                dataKey.hdlType = AM_HDL_KEY_DM_HDL;
                dataKey.u.hdl   = rsp->hdr.dmHdl;
                agtCb = (am_agCb_t *)hash_lookup(dmHashID, &dataKey, &dataKey);
            }

            if ((agtCb == NULL) || (agtCb->amSmHdl == 0))
            {
                AM_ERR_PRINT(("Unknown handle in DM_OPEN_RSP"));
                break;
            }
            else
            {
                if (rsp->lclAddr != agtCb->lclAddr)
                {
                    AM_ERR_PRINT(("Incorrect lcl@ address in DM_OPEN_RSP"));
                    break;
                }
                agtCb->dmHdl = rsp->hdr.dmHdl; /* In case it changed */
                if (rsp->result == NFX_OK)
                {
                    agtCb->lclPort = rsp->lclPort; /* allocated port */
                    esmEvent.event = SOURCE_EVENT_CONN_REQ_OK;
					strcpy(eventStr, "SOURCE_EVENT_CONN_REQ_OK");
                }
                else
				{
                    esmEvent.event = SOURCE_EVENT_CONN_REQ_NOK;
					strcpy(eventStr, "SOURCE_EVENT_CONN_REQ_NOK");
				}

                esmEvent.p1.u.dw = 0;
                esmEvent.p2.u.dw = 0;

                AM_FCT_PRINT(("Posting %s %d to %d SM", \
                              eventStr, esmEvent.event, agtCb->amSmHdl));
                rc = esm_deliverEvent(agtCb->amSmHdl, &esmEvent);
                NFX_ASSERT(rc == NFX_OK);

            }
            
            break;
        } // DM_OPEN_RSP

        case DM_SEND_RSP:
        {
            /*
             * Indication that message has been sent by TCP
             */
            dm_sendRsp_t       *rsp = (dm_sendRsp_t *)msgBuf;
            am_agCb_t          *agtCb = NULL;

            AM_FCT_PRINT(("DM_SEND_RSP"));

            appHdl = rsp->hdr.appHdl;
            if (appHdl != 0)
                agtCb = (am_agCb_t *)appHdl;
            else
            {
                AM_DBG_PRINT(("DM_SEND_RSP lookup 0x%llx", rsp->hdr.appHdl));
                dataKey.hdlType = AM_HDL_KEY_DM_HDL;
                dataKey.u.hdl   = rsp->hdr.dmHdl;
                agtCb = (am_agCb_t *)hash_lookup(dmHashID, &dataKey, &dataKey);
            }
            if ((agtCb == NULL) || (agtCb->amSmHdl == 0))
            {
                AM_ERR_PRINT(("Unknown handle in DM_SEND_RSP"));
                break;
            }
            else
            {
                agtCb->dmHdl = rsp->hdr.dmHdl; /* In case it changed */
                if (rsp->result == NFX_OK)
				{
                    esmEvent.event = M_AM_SEND_RSP_EVT_OK(agtCb);
					if (esmEvent.event == SOURCE_EVENT_SEND_DATA_OK)
						strcpy(eventStr, "SOURCE_EVENT_SEND_DATA_OK");
					else
						strcpy(eventStr, "TARGET_EVENT_SEND_DATA_OK");
				}
                else
				{
                    esmEvent.event = M_AM_SEND_RSP_EVT_NOK(agtCb);
					if (esmEvent.event == SOURCE_EVENT_SEND_DATA_NOK)
						strcpy(eventStr, "SOURCE_EVENT_SEND_DATA_NOK");
					else
						strcpy(eventStr, "TARGET_EVENT_SEND_DATA_NOK");
				}

                esmEvent.p1.u.dw = 0;
                esmEvent.p2.u.dw = 0;

                AM_FCT_PRINT(("Posting %s %d to %d SM", \
                               eventStr, esmEvent.event, agtCb->amSmHdl));
                rc = esm_deliverEvent(agtCb->amSmHdl, &esmEvent);
                NFX_ASSERT(rc == NFX_OK);

            }
            break;
        } // DM_SEND_RSP

        case DM_RCV_IND:
        {
            /*
             * Indication that message has been received
             */
            dm_rcvInd_t       *rsp = (dm_rcvInd_t *)msgBuf;
            am_agCb_t         *agtCb = NULL;
            am_intfMsg_t      *am_msg_type = NULL;

            AM_FCT_PRINT(("DM_RCV_IND"));

            appHdl = rsp->hdr.appHdl;
            if (appHdl != 0)
                agtCb = (am_agCb_t *)appHdl;
            else
            {
                AM_DBG_PRINT(("DM_RCV_IND lookup 0x%llx", rsp->hdr.appHdl));
                dataKey.hdlType = AM_HDL_KEY_DM_HDL;
                dataKey.u.hdl   = rsp->hdr.dmHdl;
                agtCb = (am_agCb_t *)hash_lookup(dmHashID, &dataKey, &dataKey);
            }

            if ((agtCb == NULL) || (agtCb->amSmHdl == 0))
            {
                AM_ERR_PRINT(("Unknown handle 0x%llx in DM_RCV_IND", rsp->hdr.appHdl));
                break;
            }
            else
            {
                agtCb->dmHdl = rsp->hdr.dmHdl; /* In case it changed */
                evtMsgBuf = am_eee_ramAlloc(msgLen);
                NFX_ASSERT(evtMsgBuf != NULL);
                memcpy(evtMsgBuf, msgBuf, msgLen);
                esmEvent.p1.u.p  = evtMsgBuf;
                esmEvent.p2.u.dw = 0;

                am_msg_type = (am_intfMsg_t *)(msgBuf+sizeof(dm_rcvInd_t));
                switch (*am_msg_type)
                {
                    case AM_GET_VOLID_REQ:
                    {
                        AM_FCT_PRINT(("AM message type : AM_GET_VOLID_REQ"));
                        esmEvent.event = TARGET_EVENT_GET_VOLID_REQ;
						strcpy(eventStr, "TARGET_EVENT_GET_VOLID_REQ");
                        break;
                    }
                    case AM_GET_VOLID_RSP:
                    {
                        AM_FCT_PRINT(("AM message type : AM_GET_VOLID_RSP"));
                        esmEvent.event = SOURCE_EVENT_GET_VOLID_OK;
						strcpy(eventStr, "SOURCE_EVENT_GET_VOLID_OK");
                        break;
                    }
                    case AM_LAST_SNAP_REQ:
                    {
                        AM_FCT_PRINT(("AM message type : AM_LAST_SNAP_REQ"));
                        esmEvent.event = TARGET_EVENT_RCV_SNAP_REQ;
						strcpy(eventStr, "TARGET_EVENT_RCV_SNAP_REQ");
                        break;
                    }
                    case AM_LAST_SNAP_RSP:
                    {
                        AM_FCT_PRINT(("AM message type : AM_LAST_SNAP_RSP"));
                        esmEvent.event = SOURCE_EVENT_RCV_SNAP_INFO;
						strcpy(eventStr, "SOURCE_EVENT_RCV_SNAP_INFO");
                        break;
                    }
                    case AM_SET_SNAP_REQ:
                    {
                        AM_FCT_PRINT(("AM message type : AM_SET_SNAP_REQ"));
                        esmEvent.event = TARGET_EVENT_SNAP_INSTL_REQ;
						strcpy(eventStr, "TARGET_EVENT_SNAP_INSTL_REQ");
                        break;
                    }
                    case AM_SET_SNAP_OK:
                    {
                        AM_FCT_PRINT(("AM message type : AM_SET_SNAP_OK"));
                        esmEvent.event = SOURCE_EVENT_RCV_SNAP_OK;
						strcpy(eventStr, "SOURCE_EVENT_RCV_SNAP_OK");
                        break;
                    }
                    case AM_SET_SNAP_NOK:
                    {
                        AM_FCT_PRINT(("AM message type : AM_SET_SNAP_NOK"));
                        esmEvent.event = SOURCE_EVENT_RCV_SNAP_NOK;
						strcpy(eventStr, "SOURCE_EVENT_RCV_SNAP_NOK");
                        break;
                    }

                    case AM_SEND_DATA_IND:
                    {
                        AM_FCT_PRINT(("AM message type : AM_SEND_DATA_IND"));
                        esmEvent.event = TARGET_EVENT_RCV_DATA_IND;
						strcpy(eventStr, "TARGET_EVENT_RCV_DATA_IND");
                        break;
                    }

                    case AM_SEND_DONE_IND:
                    {
                        AM_FCT_PRINT(("AM message type : AM_SEND_DONE_IND"));
                        esmEvent.event = TARGET_EVENT_RCV_DATA_DONE;
						strcpy(eventStr, "TARGET_EVENT_RCV_DATA_DONE");
                        break;
                    }
                    case AM_RCVD_DONE_OK:
                    {
                        AM_FCT_PRINT(("AM message type : AM_RCVD_DONE_OK"));
                        esmEvent.event = SOURCE_EVENT_RCV_DONE_OK;
						strcpy(eventStr, "SOURCE_EVENT_RCV_DONE_OK");
                        break;
                    }
                    case AM_RCVD_DONE_NOK:
                    {
                        AM_FCT_PRINT(("AM message type : AM_RCVD_DONE_NOK"));
                        esmEvent.event = SOURCE_EVENT_RCV_DONE_NOK;
						strcpy(eventStr, "SOURCE_EVENT_RCV_DONE_NOK");
                        break;
                    }
                    default:
                    {
                        AM_FCT_PRINT(("AM message type : Unknown"));
                        am_eee_ramDealloc(evtMsgBuf);
                        break;
                    }
                }

                AM_FCT_PRINT(("Posting %s %d to %d SM", \
                               eventStr, esmEvent.event, agtCb->amSmHdl));
                rc = esm_deliverEvent(agtCb->amSmHdl, &esmEvent);
                NFX_ASSERT(rc == NFX_OK);
            }
            break;
        } // DM_RCV_IND

        case DM_CLOSE_RSP:
        {
            /*
             * Indication that DM has been closed
             */
            dm_closeRsp_t     *rsp = (dm_closeRsp_t *)msgBuf;
            am_agCb_t         *agtCb = NULL;

            AM_FCT_PRINT(("DM_CLOSE_RSP"));

            appHdl = rsp->hdr.appHdl;
            if (appHdl != 0)
                agtCb = (am_agCb_t *)appHdl;
            else
            {
                AM_DBG_PRINT(("DM_CLOSE_RSP lookup 0x%llx", rsp->hdr.appHdl));
                dataKey.hdlType = AM_HDL_KEY_DM_HDL;
                dataKey.u.hdl   = rsp->hdr.dmHdl;
                agtCb = (am_agCb_t *)hash_lookup(dmHashID, &dataKey, &dataKey);
            }
            if ((agtCb == NULL) || (agtCb->amSmHdl == 0))
            {
                AM_ERR_PRINT(("Unknown handle in DM_CLOSE_RSP"));
                break;
            }
            else
            {
                dataKey.hdlType = AM_HDL_KEY_DM_HDL;
                dataKey.u.hdl   = rsp->hdr.dmHdl;
                searchKey.hdlType = AM_HDL_KEY_AM_CB;
                searchKey.u.agt   = agtCb;
                hash_delete(dmHashID, &dataKey, &searchKey);
                AM_FCT_PRINT(("DM_CLOSE_RSP delete 0x%llx", rsp->hdr.appHdl));
                esmEvent.event = M_AM_SEND_DM_CLOSE_EVT(agtCb);
				if (esmEvent.event == SOURCE_EVENT_KILL_DM_OK)
					strcpy(eventStr, "SOURCE_EVENT_KILL_DM_OK");
				else
					strcpy(eventStr, "TARGET_EVENT_DM_CLOSED");
                esmEvent.p1.u.dw = 0;
                esmEvent.p2.u.dw = 0;

                AM_FCT_PRINT(("Posting %s %d to %d SM", \
                               eventStr, esmEvent.event, agtCb->amSmHdl));
                rc = esm_deliverEvent(agtCb->amSmHdl, &esmEvent);
                //NFX_ASSERT(rc == NFX_OK);
				agtCb->amSmHdl = 0;
            }
            break;
        } // DM_CLOSE_RSP

        /*
         * Message interface between AM and EFS
         * ------------------------------------
         */
        case FS_REQ_AM_LAST_SNAPSHOT:
        {
            /*
             * Response from EFS with last snap shot name
             */
            fs_MsgAmLastSnapshotResp_t    *rsp = NULL;
            am_agCb_t                     *agtCb = NULL;

            AM_FCT_PRINT(("FS_REQ_AM_LAST_SNAPSHOT"));

            rsp   = (fs_MsgAmLastSnapshotResp_t *)msgBuf;
#ifdef SHOULD_USE_COOKIE
            appHdl = rsp->cookie;
#else
            appHdl = rsp->hdr.clientPrivate;
#endif
            agtCb = (am_agCb_t *)appHdl;

            if ((agtCb == NULL) || (agtCb->amSmHdl == 0))
            {
                AM_ERR_PRINT(("Unknown handle in FS_REQ_AM_LAST_SNAPSHOT"));
                break;
            }
            else
            {
                esmEvent.event = TARGET_EVENT_LAST_SNAP_RSP;
				strcpy(eventStr, "TARGET_EVENT_LAST_SNAP_RSP");
                evtMsgBuf = am_eee_ramAlloc(msgLen);
                NFX_ASSERT(evtMsgBuf != NULL);
                memcpy(evtMsgBuf, msgBuf, msgLen);
                esmEvent.p1.u.p  = evtMsgBuf;
                esmEvent.p2.u.dw = 0;

                AM_FCT_PRINT(("Posting %s %d to %d SM", \
                               eventStr, esmEvent.event, agtCb->amSmHdl));
                rc = esm_deliverEvent(agtCb->amSmHdl, &esmEvent);
                NFX_ASSERT(rc == NFX_OK);
            }
            break;
        } // FS_REQ_AM_LAST_SNAPSHOT

        case FS_REQ_AM_DELTA_INSTALL_SETUP:
        {
            fs_MsgAmDeltaInstallSetupResp_t  *rsp = NULL;
            am_agCb_t                        *agtCb = NULL;

            AM_FCT_PRINT(("FS_REQ_AM_DELTA_INSTALL_SETUP"));

            rsp   = (fs_MsgAmDeltaInstallSetupResp_t *)msgBuf;
#ifdef SHOULD_USE_COOKIE
            appHdl = rsp->cookie;
#else
            appHdl = rsp->hdr.clientPrivate;
#endif
            agtCb = (am_agCb_t *)appHdl;

            if ((agtCb == NULL) || (agtCb->amSmHdl == 0))
            {
                AM_ERR_PRINT(("Unknown handle in FS_REQ_AM_DELTA_INSTALL_SETUP"));
                break;
            }
            else
            {
                if (rsp->hdr.fsStatus == FS_ERR_NONE)
				{
                    esmEvent.event = TARGET_EVENT_EFS_INSTL_OK;
					strcpy(eventStr, "TARGET_EVENT_EFS_INSTL_OK");
				}
                else
				{
                    esmEvent.event = TARGET_EVENT_EFS_INSTL_NOK;
					strcpy(eventStr, "TARGET_EVENT_EFS_INSTL_NOK");
				}
                evtMsgBuf = am_eee_ramAlloc(msgLen);
                NFX_ASSERT(evtMsgBuf != NULL);
                memcpy(evtMsgBuf, msgBuf, msgLen);
                esmEvent.p1.u.p  = evtMsgBuf;
                esmEvent.p2.u.dw = 0;

                AM_FCT_PRINT(("Posting %s %d to %d SM", \
                               eventStr, esmEvent.event, agtCb->amSmHdl));
                rc = esm_deliverEvent(agtCb->amSmHdl, &esmEvent);
                NFX_ASSERT(rc == NFX_OK);
            }
            break;
        } // FS_REQ_AM_DELTA_INSTALL_SETUP

#ifdef THIS_INTERFACE_HAS_BEEN_REMOVED
        case FS_REQ_AM_DELTA_INSTALL_EXTENT:
        {
            fs_MsgAmDeltaInstallExtentResp_t    *rsp = NULL;
            am_agCb_t                        *agtCb = NULL;

            AM_FCT_PRINT(("FS_REQ_AM_DELTA_INSTALL_EXTENT"));

            rsp   = (fs_MsgAmDeltaInstallExtentResp_t *)msgBuf;
#ifdef SHOULD_USE_COOKIE
            appHdl = rsp->cookie;
#else
            appHdl = rsp->hdr.clientPrivate;
#endif
            agtCb = (am_agCb_t *)appHdl;
            if ((agtCb == NULL) || (agtCb->amSmHdl == 0))
            {
                AM_ERR_PRINT(("Unknown handle in FS_REQ_AM_DELTA_INSTALL_EXTENT"));
                break;
            }
            else
            {
                agtCb->efsHdl = rsp->extentCookie;
                if (rsp->hdr.fsStatus == FS_ERR_NONE)
				{
                    esmEvent.event = TARGET_EVENT_EFS_UPDATED;
					strcpy(eventStr, "TARGET_EVENT_EFS_UPDATED");
				}
                else
				{
                    esmEvent.event = TARGET_EVENT_EFS_UPDATE_NOK;
					strcpy(eventStr, "TARGET_EVENT_EFS_UPDATE_NOK");
				}
                evtMsgBuf = am_eee_ramAlloc(msgLen);
                NFX_ASSERT(evtMsgBuf != NULL);
                memcpy(evtMsgBuf, msgBuf, msgLen);
                esmEvent.p1.u.p  = evtMsgBuf;
                esmEvent.p2.u.dw = 0;

                AM_FCT_PRINT(("Posting %s %d to %d SM", \
                               eventStr, esmEvent.event, agtCb->amSmHdl));
                rc = esm_deliverEvent(agtCb->amSmHdl, &esmEvent);
                NFX_ASSERT(rc == NFX_OK);
            }
            break;
        } // FS_REQ_AM_DELTA_INSTALL_EXTENT
#endif

        case FS_REQ_AM_DELTA_INSTALL_COMMIT:
        {
            fs_MsgAmDeltaInstallCommitResp_t    *rsp = NULL;
            am_agCb_t                        *agtCb = NULL;

            AM_FCT_PRINT(("FS_REQ_AM_DELTA_INSTALL_COMMIT"));

            rsp   = (fs_MsgAmDeltaInstallCommitResp_t *)msgBuf;
#ifdef SHOULD_USE_COOKIE
            appHdl = rsp->cookie;
#else
            appHdl = rsp->hdr.clientPrivate;
#endif
            agtCb = (am_agCb_t *)appHdl;
            if ((agtCb == NULL) || (agtCb->amSmHdl == 0))
            {
                AM_ERR_PRINT(("Unknown handle in FS_REQ_AM_DELTA_INSTALL_COMMIT"));
                break;
            }
            else
            {
                if (rsp->hdr.fsStatus == FS_ERR_NONE)
				{
                    esmEvent.event = TARGET_EVENT_SNAP_COMMIT_OK;
					strcpy(eventStr, "TARGET_EVENT_SNAP_COMMIT_OK");
				}
                else
				{
                    esmEvent.event = TARGET_EVENT_SNAP_COMMIT_NOK;
					strcpy(eventStr, "TARGET_EVENT_SNAP_COMMIT_NOK");
				}
                evtMsgBuf = am_eee_ramAlloc(msgLen);
                NFX_ASSERT(evtMsgBuf != NULL);
                memcpy(evtMsgBuf, msgBuf, msgLen);
                esmEvent.p1.u.p  = evtMsgBuf;
                esmEvent.p2.u.dw = 0;

                AM_FCT_PRINT(("Posting %s %d to %d SM", \
                               eventStr, esmEvent.event, agtCb->amSmHdl));
                rc = esm_deliverEvent(agtCb->amSmHdl, &esmEvent);
                NFX_ASSERT(rc == NFX_OK);
            }
            break;
        } // FS_REQ_AM_DELTA_INSTALL_COMMIT

        case FS_REQ_AM_DELTA_SETUP:
        {
            fs_MsgAmDeltaSetupResp_t    *rsp = NULL;
            am_agCb_t                   *agtCb = NULL;

            AM_FCT_PRINT(("FS_REQ_AM_DELTA_SETUP RSP"));

            rsp   = (fs_MsgAmDeltaSetupResp_t *)msgBuf;
#ifdef SHOULD_USE_COOKIE
            appHdl = rsp->cookie;
#else
            appHdl = rsp->hdr.clientPrivate;
#endif
            agtCb = (am_agCb_t *)appHdl;
            if ((agtCb == NULL) || (agtCb->amSmHdl == 0))
            {
                AM_ERR_PRINT(("Unknown handle in FS_REQ_AM_DELTA_SETUP RSP"));
                break;
            }
            else
            {
                agtCb->efsHdl = rsp->fsCookie;

				AM_FCT_PRINT(("FS_REQ_AM_DELTA_SETUP RSP status code %d", rsp->hdr.fsStatus));

                if (rsp->hdr.fsStatus == FS_ERR_NONE)
				{
                    esmEvent.event = SOURCE_EVENT_DLTA_CRT_OK;
					strcpy(eventStr, "SOURCE_EVENT_DLTA_CRT_OK");
				}
                else
				{
                    esmEvent.event = SOURCE_EVENT_DLTA_CRT_NOK;
					strcpy(eventStr, "SOURCE_EVENT_DLTA_CRT_NOK");
				}

                evtMsgBuf = am_eee_ramAlloc(msgLen);
                NFX_ASSERT(evtMsgBuf != NULL);
                memcpy(evtMsgBuf, msgBuf, msgLen);
                esmEvent.p1.u.p  = evtMsgBuf;
                esmEvent.p2.u.dw = 0;

                AM_FCT_PRINT(("Posting %s %d to %d SM", \
                               eventStr, esmEvent.event, agtCb->amSmHdl));
                rc = esm_deliverEvent(agtCb->amSmHdl, &esmEvent);
                NFX_ASSERT(rc == NFX_OK);
            }
            break;
        } // FS_REQ_AM_DELTA_SETUP

        case FS_REQ_AM_DELTA_GET_EXTENT:
        {
            fs_MsgAmDeltaGetExtentResp_t   *rsp = NULL;
            am_agCb_t                      *agtCb = NULL;

            AM_FCT_PRINT(("FS_REQ_AM_DELTA_GET_EXTENT"));

            rsp   = (fs_MsgAmDeltaGetExtentResp_t *)msgBuf;
#ifdef SHOULD_USE_COOKIE
            appHdl = rsp->cookie;
#else
            appHdl = rsp->hdr.clientPrivate;
#endif
            agtCb = (am_agCb_t *)appHdl;
            if ((agtCb == NULL) || (agtCb->amSmHdl == 0))
            {
                AM_ERR_PRINT(("Unknown handle in FS_REQ_AM_DELTA_GET_EXTENT"));
                break;
            }
            else
            {
                agtCb->efsHdl = rsp->fsCookie;
                if (rsp->hdr.fsStatus == FS_ERR_NONE)
                {
                    if (rsp->numExtents != 0)
					{
                        esmEvent.event = SOURCE_EVENT_DLTA_EXT_LIST;
						strcpy(eventStr, "SOURCE_EVENT_DLTA_EXT_LIST"); 
					}
                    else
					{
                        esmEvent.event = SOURCE_EVENT_DLTA_EXT_NONE;
						strcpy(eventStr, "SOURCE_EVENT_DLTA_EXT_NONE"); 
					}
                }
                else if (rsp->hdr.fsStatus == FS_ERR_NOENT)
				{
                    esmEvent.event = SOURCE_EVENT_DLTA_EXT_NONE;
					strcpy(eventStr, "SOURCE_EVENT_DLTA_EXT_NONE"); 
				}
                else
				{
                    /* In case of error, just abort the session !!! */
                    esmEvent.event = M_AM_MIRROR_ABORT_EVT(agtCb);
					if (esmEvent.event == SOURCE_EVENT_MIRROR_ABORT)
						strcpy(eventStr, "SOURCE_EVENT_MIRROR_ABORT");
					else
						strcpy(eventStr, "TARGET_EVENT_MIRROR_ABORT");
				}
                evtMsgBuf = am_eee_ramAlloc(msgLen);
                NFX_ASSERT(evtMsgBuf != NULL);
                memcpy(evtMsgBuf, msgBuf, msgLen);
                esmEvent.p1.u.p  = evtMsgBuf;
                esmEvent.p2.u.dw = 0;

                AM_FCT_PRINT(("Posting %s %d to %d SM", \
                               eventStr, esmEvent.event, agtCb->amSmHdl));
                rc = esm_deliverEvent(agtCb->amSmHdl, &esmEvent);
                NFX_ASSERT(rc == NFX_OK);
            }
            break;
        } // FS_REQ_AM_DELTA_GET_EXTENT

        case FS_REQ_AM_DELTA_COMPLETE:
        {
            fs_MsgAmDeltaCompleteResp_t   *rsp = NULL;
            am_agCb_t                     *agtCb = NULL;

            AM_FCT_PRINT(("FS_REQ_AM_DELTA_COMPLETE"));

            rsp   = (fs_MsgAmDeltaCompleteResp_t *)msgBuf;
#ifdef SHOULD_USE_COOKIE
            appHdl = rsp->cookie;
#else
            appHdl = rsp->hdr.clientPrivate;
#endif
            agtCb = (am_agCb_t *)appHdl;
            if ((agtCb == NULL) || (agtCb->amSmHdl == 0))
            {
                AM_ERR_PRINT(("Unknown handle in FS_REQ_AM_DELTA_COMPLETE"));
                break;
            }
            else
            {
                if (rsp->hdr.fsStatus == FS_ERR_NONE)
				{
                    esmEvent.event = SOURCE_EVENT_DLTA_COMPLETE;
					strcpy(eventStr, "SOURCE_EVENT_DLTA_COMPLETE");
				}
                else
				{
                    /* In case of error, just abort the session !!! */
                    /* It does not make much difference on source */
                    esmEvent.event = M_AM_MIRROR_ABORT_EVT(agtCb);
					if (esmEvent.event == SOURCE_EVENT_MIRROR_ABORT)
						strcpy(eventStr, "SOURCE_EVENT_MIRROR_ABORT");
					else
						strcpy(eventStr, "TARGET_EVENT_MIRROR_ABORT");
				}
                evtMsgBuf = am_eee_ramAlloc(msgLen);
                NFX_ASSERT(evtMsgBuf != NULL);
                memcpy(evtMsgBuf, msgBuf, msgLen);
                esmEvent.p1.u.p  = evtMsgBuf;
                esmEvent.p2.u.dw = 0;

                AM_FCT_PRINT(("Posting %s %d to %d SM", \
                               eventStr, esmEvent.event, agtCb->amSmHdl));
                rc = esm_deliverEvent(agtCb->amSmHdl, &esmEvent);
                NFX_ASSERT(rc == NFX_OK);
            }
            break;
        } // FS_REQ_AM_DELTA_COMPLETE

        /*
         * Message interface between AM and EVM (SSC)
         * ------------------------------------------
         */
        case EVM_MSG_STAT:
        {
            /*
             * Response from EVM (SSC with target vol id)
             */
            evm_MsgStat_t		*rsp = NULL;
            am_agCb_t           *agtCb = NULL;

            AM_FCT_PRINT(("EVM_MSG_STAT"));

            rsp   = (evm_MsgStat_t *)msgBuf;
#ifdef SHOULD_USE_COOKIE
            appHdl = rsp->cookie;
#else
            appHdl = rsp->hdr.clientPrivate;
#endif
            agtCb = (am_agCb_t *)appHdl;

            if ((agtCb == NULL) || (agtCb->amSmHdl == 0))
            {
                AM_ERR_PRINT(("Unknown handle in EVM_MSG_STAT"));
                break;
            }
            else
            {
                esmEvent.event = TARGET_EVENT_GET_VOLID_RSP;
				strcpy(eventStr, "TARGET_EVENT_GET_VOLID_RSP");
                evtMsgBuf = am_eee_ramAlloc(msgLen);
                NFX_ASSERT(evtMsgBuf != NULL);
                memcpy(evtMsgBuf, msgBuf, msgLen);
                esmEvent.p1.u.p  = evtMsgBuf;
                esmEvent.p2.u.dw = 0;

                AM_FCT_PRINT(("Posting %s %d to %d SM", \
                           	eventStr, esmEvent.event, agtCb->amSmHdl));
                rc = esm_deliverEvent(agtCb->amSmHdl, &esmEvent);
                NFX_ASSERT(rc == NFX_OK);
            }
            break;
        } // EVM_MSG_STAT

    } // switch

    /*
     * Free the message buffer.
     */
    if (bufIsInSharedMemory == TRUE)
    {
        eee_deallocateBuffer(msgBuf);
    }
    else
    {
        rc = am_eee_ramDealloc(msgBuf);
        NFX_ASSERT(rc == NFX_OK);
    }

    AM_DBG_PRINT(("...am_ag_rcvMsg"));
    return;
}


/*
 *------------------------------------------------------------------------------
 * Name         : am_ag_rcvMcastMsg()
 *
 * Description  : Process received multicast message.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 02-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
void am_ag_rcvMcastMsg(eee_descPtr_t  edesc)

{
    AM_DBG_PRINT(("am_ag_rcvMcastMsg..."));
    
    AM_ERR_PRINT(("No multicast messages are expected by the AM agent."));
    AM_ERR_PRINT(("Message is discarded."));
    eee_freePacket(edesc);
    
    AM_DBG_PRINT(("...am_ag_rcvMcastMsg"));
}


/*
 *------------------------------------------------------------------------------
 * Name         : source_mirrorStatus()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created :
 *
 *------------------------------------------------------------------------------
 */

void source_mirrorStatus(int32 ctrlHdl, int32 status)
{
	am_mirrorStatusInd_t  amStatusInd;

	AM_DBG_PRINT(("Source AM_MIRROR_STATUS idx %d status %s",
					ctrlHdl, (status == AM_MIRROR_COMPLETED) 
									? "COMPLETED" : "ABORTED"));

	printf("Source AM_MIRROR_STATUS idx %d status %s\n",
					ctrlHdl, (status == AM_MIRROR_COMPLETED) 
									? "COMPLETED" : "ABORTED");

	/*
	 * Prepare and send status indication
	 */

	memset(&amStatusInd, 0, sizeof(am_mirrorStatusInd_t));
	amStatusInd.hdr.msgType   = AM_MIRROR_STATUS_IND; // Type of message
	amStatusInd.hdr.ctrlHdl   = ctrlHdl;   // return to request
	amStatusInd.status = status;

	eee_sendMessage(&amStatusInd, sizeof(am_mirrorStatusInd_t),
							ctrlAddr, amAddr);
}

/*
 *------------------------------------------------------------------------------
 * Name         : source_evmIOStat()
 *
 * Description  : Callback from EVM with 'read' result.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 09-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

void source_evmIOStat(evm_IOStat_t * stat)
{
    esm_event_t esmEvent;
    int32       rc;
    am_agCb_t   *agtCb = NULL;
    evm_IOStat_t *evmMsg = NULL;
    uint64       appHdl;

    AM_DBG_PRINT(("source_evmIOStat..."));

    /*
     * Forward the response to the right instance.
     */
    appHdl = stat->clientPrivate;
    agtCb = (am_agCb_t *)appHdl;
    if ((agtCb == NULL) || (agtCb->amSmHdl == 0))
	{
        AM_ERR_PRINT(("Unknown control block in source_evmIOStat"));
	}
    else
    {
        bzero(&esmEvent, sizeof(esm_event_t));
        if (stat->status == EVM_STAT_NO_ERROR)
		{
            esmEvent.event = SOURCE_EVENT_EVM_COD_OK;
			strcpy(eventStr, "SOURCE_EVENT_EVM_COD_OK");
		}
        else
		{
            esmEvent.event = SOURCE_EVENT_EVM_COD_NOK;
			strcpy(eventStr, "SOURCE_EVENT_EVM_COD_NOK");
		}
        evmMsg = am_eee_ramAlloc(sizeof(evm_IOStat_t));
        NFX_ASSERT(evmMsg != NULL);
        memcpy (evmMsg, stat, sizeof(evm_IOStat_t));
        esmEvent.p1.u.p  = evmMsg;
        esmEvent.p2.u.dw = 0;
        AM_FCT_PRINT(("Posting %s %d to %d SM", \
                       eventStr, esmEvent.event, agtCb->amSmHdl));
        rc = esm_deliverEvent(agtCb->amSmHdl, &esmEvent);
        NFX_ASSERT(rc == NFX_OK);
    }
    
    AM_DBG_PRINT(("...source_evmIOStat"));
    return;
}


/*
 *------------------------------------------------------------------------------
 * Name         : target_evmIOStat()
 *
 * Description  : Callback from EVM with 'write' result.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 10-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
void target_evmIOStat(evm_IOStat_t * stat)
{
    esm_event_t  esmEvent;
    int32        rc;
    am_agCb_t    *agtCb = NULL;
    evm_IOStat_t *evmMsg = NULL;
    uint64       appHdl;

    AM_DBG_PRINT(("target_evmIOStat..."));

    /*
     * Forward the response to the right instance
     */
    appHdl = stat->clientPrivate;
    agtCb = (am_agCb_t *)appHdl;
    if ((agtCb == NULL) || (agtCb->amSmHdl == 0))
	{
        AM_ERR_PRINT(("Unknown control block in target_evmIOStat"));
	}
    else
    {
        bzero(&esmEvent, sizeof(esm_event_t));
        if (stat->status == EVM_STAT_NO_ERROR)
		{
            esmEvent.event = TARGET_EVENT_EVM_UPDATED;
			strcpy(eventStr, "TARGET_EVENT_EVM_UPDATED");
		}
        else
		{
            esmEvent.event = TARGET_EVENT_EVM_UPDATE_NOK;
			strcpy(eventStr, "TARGET_EVENT_EVM_UPDATE_NOK");
		}
        evmMsg = am_eee_ramAlloc(sizeof(evm_IOStat_t));
        NFX_ASSERT(evmMsg != NULL);
        memcpy (evmMsg, stat, sizeof(evm_IOStat_t));
        esmEvent.p1.u.p  = evmMsg;
        esmEvent.p2.u.dw = 0;
        AM_FCT_PRINT(("Posting %s %d to %d SM", \
                       eventStr, esmEvent.event, agtCb->amSmHdl));
        rc = esm_deliverEvent(agtCb->amSmHdl, &esmEvent);
        NFX_ASSERT(rc == NFX_OK);
    }

    AM_DBG_PRINT(("...target_evmIOStat"));
    return;
}

/*----------------------------------------------------------------------------*/
#endif  /* SSC */
