/*
 *------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : source.c
 *
 * RCS          :
 *
 * Description  : Action functions for the AM agent on the source.
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 23-JAN-2002
 *
 * History      : Date       Name         Comments
 *                --------------------------------
 *                1/23/02    Nitin        created
 * 
 *------------------------------------------------------------------------------
 */

#ifndef SSC

#include <nfx-incl.h>
#include <exports/timer-api.h>
#include "agent.h"

extern char amNewSnapName[];

/*
 * Prototypes of local functions.
 *
 */

void am_post_event_abort(am_agCb_t *agtCb, char *msg);

/*
 * Local functions.
 */

/*
 *------------------------------------------------------------------------------
 * Name         : am_convert_extents_to_blocks
 *
 * Description  : Converts the extents to blocks and returns number of blocks
 *
 * Created by   : Nitin Jain
 *
 *------------------------------------------------------------------------------
 */
static uint32 am_convert_extents_to_blocks(am_agCb_t *agtCb,
                                           fs_MsgAmDeltaGetExtentResp_t *fsRsp)
{
    int32           ext_cnt = 0;
    uint32          ext_len;
    uint64          ext_base;
    am_blk_list_t   *head, *new, *prev;
    int32           cnt = 0;

    AM_DBG_PRINT(("am_convert_extents_to_blocks..."));

    head = NULL;
    prev = NULL;
    new  = NULL;
    for (ext_cnt = 0; ext_cnt < fsRsp->numExtents; ext_cnt++)
    {
        ext_base = fsRsp->extents[ext_cnt].blknum;
        ext_len  = fsRsp->extents[ext_cnt].len;  // number of blocks
        while (ext_len > 0)
        {
            prev = new;
            new = am_eee_ramAlloc(sizeof(am_blk_list_t));
            NFX_ASSERT(new != NULL);
            /*
             * EVM has a view of 512 Bytes blocks. It is better to transfer
             * Those blocks to the target.
             * The EFS always returns blocks in the size of EFS_BLOCK_SIZE
             * i.e. 8K
             * Convert these blocks to EVM blocks
             */
            new->blkOffSet = ext_base * (EFS_BLOCK_SIZE/EVM_LBLK_SIZE);
            new->blkCnt = EFS_BLOCK_SIZE/EVM_LBLK_SIZE;
            AM_FCT_PRINT(("EFS block number and count : %lld %d", \
                           ext_base, (fsRsp->extents[ext_cnt].len - ext_len)));
            AM_DBG_PRINT(("EVM block number and count : %lld %d", \
                                              new->blkOffSet, new->blkCnt));
            ext_len--;
            ext_base++;
            new->next = NULL;
            if (head == NULL)
                head = new;
            else
                prev->next = new;
            cnt++;
         }
     }
     agtCb->lastBlk = new;
     agtCb->blks    = head;
     agtCb->currBlk = head;

     AM_DBG_PRINT(("%d blocks in %d extents", cnt, fsRsp->numExtents));

     AM_DBG_PRINT(("...am_convert_extents_to_blocks"));
     return cnt;
}


/*
 *------------------------------------------------------------------------------
 * Name         : am_free_block_list
 *
 * Description  : Free the block list
 *
 * Created by   : Nitin Jain
 *
 *------------------------------------------------------------------------------
 */
void am_free_block_list(am_agCb_t *agtCb)
{
    am_blk_list_t  *nxt = NULL, *cur = NULL;

    AM_DBG_PRINT(("am_free_block_list..."));

    cur = agtCb->blks;
    while(cur)
    {
        nxt = cur->next;
        am_eee_ramDealloc(cur);
        cur = nxt;
    }
    agtCb->blks    = NULL;
    agtCb->lastBlk = NULL;
    agtCb->currBlk = NULL;

    return;
}


/*
 * Global (action) functions.
 */

/*
 *------------------------------------------------------------------------------
 * Name         : source_create
 *
 * Description  : Create function for source state machine.
 *
 * Created by   : Nitin Jain
 *
 *------------------------------------------------------------------------------
 */
int32 source_create (esm_handle_t  handle,
                     int32         nextState,
                     esm_event_t   thisEvent,
                     void          *userInfo)
{
    AM_FCT_PRINT(("source_create..."));


    /*
     * Nothing to do here.
     */

    AM_DBG_PRINT(("...source_create"));
    return (nextState);
}


/*
 *------------------------------------------------------------------------------
 * Name         : source_bad_state
 *
 * Description  :
 *
 * Created by   : Nitin Jain
 *
 *------------------------------------------------------------------------------
 */
int32 source_bad_state (esm_handle_t  handle,
                        int32         nextState,
                        esm_event_t   thisEvent,
                        void          *userInfo)
{
    AM_ERR_PRINT(("source_bad_state..."));

    /*
     * Nothing to do here.
     */
    if (thisEvent.p1.u.p)
        am_eee_ramDealloc(thisEvent.p1.u.p);

    AM_ERR_PRINT(("...source_bad_state"));
    return (nextState);
}


/*
 *------------------------------------------------------------------------------
 * Name         : source_connReq
 *
 * Description  : Event received from controller, to initiate connection
 *
 * Created by   : Nitin Jain
 *
 *------------------------------------------------------------------------------
 */
int32 source_connReq (esm_handle_t  handle,
                      int32         nextState,
                      esm_event_t   thisEvent,
                      void          *userInfo)
{
    am_agCb_t       *agtCb = (am_agCb_t *)userInfo;
    dm_openReq_t    dmReq;
    int32           rc = 0;
    uint64          appHdl = (uint64)userInfo;

    AM_DBG_PRINT(("source_connReq..."));

    AM_FCT_PRINT(("source_connReq send DM_OPEN_REQ"));

    dmReq.hdr.msgType = DM_OPEN_REQ;
    dmReq.hdr.appHdl  = appHdl;
    dmReq.hdr.dmHdl   = agtCb->dmHdl;

    dmReq.protocol    = TPL_PROT_TCP;   // Should be TCP
    dmReq.lclAddr     = agtCb->lclAddr;
    dmReq.lclPort     = agtCb->lclPort;
    dmReq.rmtAddr     = agtCb->rmtAddr;
    dmReq.rmtPort     = agtCb->rmtPort;

    rc = eee_sendMessage(&dmReq, sizeof(dm_openReq_t), datamoverAddr, amAddr);
    NFX_ASSERT(rc == NFX_OK);
    if (thisEvent.p1.u.p)
        am_eee_ramDealloc(thisEvent.p1.u.p);

    AM_DBG_PRINT(("...source_connReq"));
    return (nextState);
}

int32 sendReqVolId (uint32 tmr_handle,
					void *userInfo,
					void *par2,
					void *par3,
					void *par4)
{
    am_agCb_t       *agtCb = (am_agCb_t *)userInfo;
    dm_sendReq_t    *dmReq = NULL;
    char            *req = NULL;
    int32           rc = 0, msgLen;
    am_getVolIdReq_t  *amMsg = NULL;
    uint64          appHdl = (uint64)userInfo;

    AM_FCT_PRINT(("source_reqVolId... tgtVolName %s ", agtCb->tgtFsysName));

    msgLen = sizeof(dm_sendReq_t) + sizeof(am_getVolIdReq_t);
    req = (char *)am_eee_ramAlloc(msgLen);
    NFX_ASSERT(req != NULL);

    dmReq = (dm_sendReq_t *)req;
    dmReq->hdr.msgType = DM_SEND_REQ;
    dmReq->hdr.appHdl  = appHdl;
    dmReq->hdr.dmHdl   = agtCb->dmHdl;

    dmReq->dataSize    = 0;
    dmReq->cod         = 0;
    dmReq->offset      = 0;
    dmReq->appHdrSize  = sizeof(am_getVolIdReq_t);

    amMsg = (am_getVolIdReq_t *)(req+sizeof(dm_sendReq_t));
    amMsg->hdr.msgType  = AM_GET_VOLID_REQ;
	strcpy(amMsg->tgtFsysName, agtCb->tgtFsysName);

    rc = eee_sendMessage(req, msgLen, datamoverAddr, amAddr);
    NFX_ASSERT(rc == NFX_OK);
    am_eee_ramDealloc(req);

	return 0;
}

/*
 *------------------------------------------------------------------------------
 * Name         : source_getVol
 *
 * Description  :
 *
 * Created by   : Pankaj Gupta
 *
 *------------------------------------------------------------------------------
 */
int32 source_reqVolId (esm_handle_t  handle,
                      int32         nextState,
                      esm_event_t   thisEvent,
                      void          *userInfo)
{
    am_agCb_t       *agtCb = (am_agCb_t *)userInfo;

    AM_FCT_PRINT(("source_reqVolId... "));

	agtCb->tmrHandle = eee_timerCreate(2000, EEE_TIMER_REPEAT);

	/* 
	 * Send the first request
	 */
	sendReqVolId(agtCb->tmrHandle, userInfo, NULL, NULL, NULL);

	/*
	 * Start retry timer
	 */
	eee_timerStart(agtCb->tmrHandle, sendReqVolId, userInfo, 
						NULL, NULL, NULL);

    if (thisEvent.p1.u.p)
        am_eee_ramDealloc(thisEvent.p1.u.p);

    AM_DBG_PRINT(("...source_reqVolId"));
    return (nextState);
}


/*
 *------------------------------------------------------------------------------
 * Name         : source_getSnap
 *
 * Description  :
 *
 * Created by   : Nitin Jain
 *
 *------------------------------------------------------------------------------
 */
int32 source_getSnap (esm_handle_t  handle,
                      int32         nextState,
                      esm_event_t   thisEvent,
                      void          *userInfo)
{
    am_agCb_t       *agtCb = (am_agCb_t *)userInfo;
	char            *msg   = (char *)thisEvent.p1.u.p;
	am_getVolIdRsp_t *amVolIdRsp = 	
							(am_getVolIdRsp_t *)(msg+sizeof(dm_rcvInd_t));

    dm_sendReq_t    *dmReq = NULL;
    char            *req = NULL;
    int32           rc = 0, msgLen;
    am_msg_hdr_t    *amMsg = NULL;
    uint64          appHdl = (uint64)userInfo;

    AM_DBG_PRINT(("source_getSnap..."));

	/*
	 * Got target vol id. Stop request timer
	 * Update vol id and send request for SNAP
	 */
	eee_timerStop(agtCb->tmrHandle);
	eee_timerDestroy(agtCb->tmrHandle);

	AM_FCT_PRINT(("Source -> got vol id 0x%llx for vol %s",
			amVolIdRsp->tgtVolId.whole_thing, amVolIdRsp->tgtFsysName));
	memcpy(&agtCb->tgtVolId, &amVolIdRsp->tgtVolId, sizeof(evm_volId_t));

	/*
	 * Send get SNAP request
	 */
    msgLen = sizeof(dm_sendReq_t) + sizeof(am_msg_hdr_t);
    req = (char *)am_eee_ramAlloc(msgLen);
    NFX_ASSERT(req != NULL);

    dmReq = (dm_sendReq_t *)req;
    dmReq->hdr.msgType = DM_SEND_REQ;
    dmReq->hdr.appHdl  = appHdl;
    dmReq->hdr.dmHdl   = agtCb->dmHdl;

    dmReq->dataSize    = 0;
    dmReq->cod         = 0;
    dmReq->offset      = 0;
    dmReq->appHdrSize  = sizeof(am_msg_hdr_t);

    amMsg = (am_msg_hdr_t *)(req+sizeof(dm_sendReq_t));
    amMsg->am_msgType  = AM_LAST_SNAP_REQ;
    memcpy (&(amMsg->volId), &(agtCb->tgtVolId), sizeof(evm_volId_t));

    rc = eee_sendMessage(req, msgLen, datamoverAddr, amAddr);
    NFX_ASSERT(rc == NFX_OK);
    am_eee_ramDealloc(req);
    if (thisEvent.p1.u.p)
        am_eee_ramDealloc(thisEvent.p1.u.p);

    AM_DBG_PRINT(("...source_getSnap"));
    return (nextState);
}


/*
 *------------------------------------------------------------------------------
 * Name         : source_crtDlta
 *
 * Description  :
 *
 * Created by   : Nitin Jain
 *
 *------------------------------------------------------------------------------
 */
int32 source_crtDlta (esm_handle_t  handle,
                      int32         nextState,
                      esm_event_t   thisEvent,
                      void          *userInfo)
{
#ifdef INCLUDE_FS_ON_SOURCE
    am_agCb_t                *agtCb = (am_agCb_t *)userInfo;
    am_rc_t                  rc;
    char                     *tgtRsp = (char *)thisEvent.p1.u.p;
    int32                    dmMsgLen = sizeof(dm_sendReq_t);
    am_snap_msg_hdr_t        *amMsg = (am_snap_msg_hdr_t *)(tgtRsp+dmMsgLen);
    fs_MsgAmDeltaSetupReq_t  fsReq;
    int32                    fsReqSize = sizeof(fs_MsgAmDeltaSetupReq_t);
    uint64                   appHdl = (uint64)userInfo;

    AM_DBG_PRINT(("source_crtDlta..."));

    bzero(&fsReq, fsReqSize);
    fsReq.hdr.requestType   = FS_REQ_AM_DELTA_SETUP;
    fsReq.hdr.len           = fsReqSize;
    fsReq.hdr.clientPrivate = appHdl;
    fsReq.cookie            = appHdl;

    if (strlen(amMsg->snapName) > strlen(amNewSnapName))
        AM_FCT_PRINT(("Source -> Last snap on target : %s", amMsg->snapName));
    else
        AM_FCT_PRINT(("Source -> Last snap on target : none"));

	memcpy(agtCb->lastSnap, amMsg->snapName, sizeof(am_snapName_t));
    memcpy(&(fsReq.volId), &(agtCb->amVolId), sizeof(evm_volId_t));
    memcpy(fsReq.oldName, agtCb->lastSnap, sizeof(am_snapName_t));
    memcpy(fsReq.newName, agtCb->newSnap, sizeof(am_snapName_t));

	AM_FCT_PRINT(("Sending FS_REQ_AM_DELTA_SETUP to 0x%x\n", efsAddr));

    rc = eee_sendMessage(&fsReq, fsReqSize, efsAddr, amAddr);
    NFX_ASSERT(rc == NFX_OK);
    if (thisEvent.p1.u.p)
        am_eee_ramDealloc(thisEvent.p1.u.p);

    AM_DBG_PRINT(("...source_crtDlta"));
    return (nextState);
#else
    am_agCb_t                *agtCb = (am_agCb_t *)userInfo;
    am_rc_t                  rc;
    char                     *tgtRsp = (char *)thisEvent.p1.u.p;
    fs_MsgAmDeltaSetupResp_t fsRsp;
    int32                    fsRspSize = sizeof(fs_MsgAmDeltaSetupResp_t);
    uint64                   appHdl = (uint64)userInfo;

    AM_DBG_PRINT(("source_crtDlta..."));

    bzero(&fsRsp, fsRspSize);
    fsRsp.hdr.requestType   = FS_REQ_AM_DELTA_SETUP;
    fsRsp.hdr.len           = fsRspSize;
    fsRsp.hdr.fsStatus      = FS_ERR_NONE;
    fsRsp.hdr.clientPrivate = appHdl;
    fsRsp.cookie            = appHdl;
    fsRsp.fsCookie          = 0;
    fsRsp.snapNum           = 1;

    rc = eee_sendMessage(&fsRsp, fsRspSize, amAddr, amAddr);
    NFX_ASSERT(rc == NFX_OK);
    if (thisEvent.p1.u.p)
        am_eee_ramDealloc(thisEvent.p1.u.p);

    AM_DBG_PRINT(("...source_crtDlta"));
    return (nextState);
#endif
}


/*
 *------------------------------------------------------------------------------
 * Name         : source_sndSnap
 *
 * Description  :
 *
 * Created by   : Nitin Jain
 *
 *------------------------------------------------------------------------------
 */
int32 source_sndSnap (esm_handle_t  handle,
                      int32         nextState,
                      esm_event_t   thisEvent,
                      void          *userInfo)
{
    am_agCb_t                 *agtCb = (am_agCb_t *)userInfo;
    char                      *req = NULL;
    dm_sendReq_t              *dmReq = NULL;
    am_snap_msg_hdr_t         *amMsg = NULL;
    int32                     rc = 0, msgLen;
    int32                     dmReqLen = sizeof(dm_sendReq_t);
    int32                     amMsgLen = sizeof(am_snap_msg_hdr_t);
    fs_MsgAmDeltaSetupResp_t  *fsRsp = NULL;
    uint64                    appHdl = (uint64)userInfo;


    AM_DBG_PRINT(("source_sndSnap..."));

    fsRsp = (fs_MsgAmDeltaSetupResp_t *)thisEvent.p1.u.p;
    msgLen = dmReqLen + amMsgLen;
    req = (char *)am_eee_ramAlloc(msgLen);
    NFX_ASSERT(req != NULL);

    dmReq = (dm_sendReq_t *)req;
    dmReq->hdr.msgType = DM_SEND_REQ;
    dmReq->hdr.appHdl  = appHdl;
    dmReq->hdr.dmHdl   = agtCb->dmHdl;

    dmReq->dataSize    = 0;
    dmReq->cod         = 0;
    dmReq->offset      = 0;
    dmReq->appHdrSize  = amMsgLen;

    amMsg = (am_snap_msg_hdr_t *)(req+dmReqLen);
    amMsg->am_msgType  = AM_SET_SNAP_REQ;
    memcpy (&(amMsg->volId), &(agtCb->tgtVolId), sizeof(evm_volId_t));
    amMsg->snapNum     = fsRsp->snapNum;
    memcpy (amMsg->snapName, agtCb->newSnap, sizeof(am_snapName_t));
    AM_FCT_PRINT(("Source -> New snap (%d) : %s", fsRsp->snapNum,
					agtCb->newSnap));

    rc = eee_sendMessage(req, msgLen, datamoverAddr, amAddr);
    NFX_ASSERT(rc == NFX_OK);
    am_eee_ramDealloc(req);
    if (thisEvent.p1.u.p)
        am_eee_ramDealloc(thisEvent.p1.u.p);

    AM_DBG_PRINT(("...source_sndSnap"));
    return (nextState);
}


/*
 *------------------------------------------------------------------------------
 * Name         : source_getDlta
 *
 * Description  :
 *
 * Created by   : Nitin Jain
 *
 *------------------------------------------------------------------------------
 */
int32 source_getDlta (esm_handle_t  handle,
                      int32         nextState,
                      esm_event_t   thisEvent,
                      void          *userInfo)
{
#ifdef INCLUDE_FS_ON_SOURCE
    am_agCb_t                   *agtCb = (am_agCb_t *)userInfo;
    uint32                      fsReqLen = sizeof(fs_MsgAmDeltaGetExtentReq_t);
    am_rc_t                     rc;
    fs_MsgAmDeltaGetExtentReq_t fsReq;
    uint64                      appHdl = (uint64)userInfo;

    AM_DBG_PRINT(("source_getDlta..."));

    bzero(&fsReq, fsReqLen);
    fsReq.hdr.requestType   = FS_REQ_AM_DELTA_GET_EXTENT;
    fsReq.hdr.len           = fsReqLen;
    fsReq.hdr.clientPrivate = appHdl;
    fsReq.cookie            = appHdl;
    fsReq.fsCookie          = agtCb->efsHdl;
    fsReq.maxExtents        = 1; // Max value FS_Am_MAX_EXTENT_TRANSFER
    memcpy(&(fsReq.volId), &(agtCb->amVolId), sizeof(evm_volId_t));
    AM_DBG_PRINT(("Source -> get delta fsCookie : %lld", agtCb->efsHdl));

	AM_FCT_PRINT(("Sending FS_REQ_AM_DELTA_GET_EXTENT to 0x%x\n", efsAddr));
    rc = eee_sendMessage(&fsReq, fsReqLen, efsAddr, amAddr);
    NFX_ASSERT(rc == NFX_OK);
    if (thisEvent.p1.u.p)
        am_eee_ramDealloc(thisEvent.p1.u.p);

    AM_DBG_PRINT(("...source_getDlta"));
    return (nextState);
#else
    am_agCb_t                   *agtCb = (am_agCb_t *)userInfo;
    am_rc_t                     rc;
    uint64                      appHdl = (uint64)userInfo;
    fs_MsgAmDeltaGetExtentResp_t  fsRsp;

    AM_DBG_PRINT(("source_getDlta..."));

    bzero(&fsRsp, sizeof(fs_MsgAmDeltaGetExtentResp_t));
    fsRsp.hdr.requestType   = FS_REQ_AM_DELTA_GET_EXTENT;
    fsRsp.hdr.len = sizeof(fs_MsgAmDeltaGetExtentResp_t);
    fsRsp.hdr.clientPrivate = appHdl;
    fsRsp.cookie = appHdl;
    fsRsp.fsCookie = 1;
    if (agtCb->nextExt == TRUE)
    {
        fsRsp.hdr.fsStatus = FS_ERR_NONE;
        fsRsp.numExtents = 1;
        fsRsp.extents[0].blknum = 2;
        fsRsp.extents[0].len    = 1;
    }
    else
    {
        fsRsp.hdr.fsStatus = FS_ERR_NOENT;
        fsRsp.numExtents = 0;
        fsRsp.extents[0].blknum = 0;
        fsRsp.extents[0].len    = 0;
    }
    rc = eee_sendMessage(&fsRsp, 
                         sizeof(fs_MsgAmDeltaGetExtentResp_t),
                         amAddr, amAddr);
    NFX_ASSERT(rc == NFX_OK);
    if (thisEvent.p1.u.p)
        am_eee_ramDealloc(thisEvent.p1.u.p);

    AM_DBG_PRINT(("...source_getDlta"));
    return (nextState);
#endif
}


/*
 *------------------------------------------------------------------------------
 * Name         : source_procList
 *
 * Description  :
 *
 * Created by   : Nitin Jain
 *
 *------------------------------------------------------------------------------
 */
int32 source_procList (esm_handle_t  handle,
                       int32         nextState,
                       esm_event_t   thisEvent,
                       void          *userInfo)
{
    am_agCb_t                    *agtCb = (am_agCb_t *)userInfo;
    evm_IOReq_t                  evmReq;
    am_blk_list_t                *c_blk = NULL;
    am_rc_t                      rc = NFX_OK;
    esm_event_t                  newEvent;
    fs_MsgAmDeltaGetExtentResp_t *fsRsp = NULL;
    uint64                       appHdl = (uint64)userInfo;
    uint32                       numBlks = 0;

    AM_DBG_PRINT(("source_procList..."));

    if (thisEvent.p1.u.p != NULL)
    {
        /*
         * This is response to getDlta sequence
         */
        AM_DBG_PRINT(("Source -> process response from FS"));
        fsRsp = (fs_MsgAmDeltaGetExtentResp_t *)thisEvent.p1.u.p;
        numBlks = am_convert_extents_to_blocks(agtCb, fsRsp);
        AM_FCT_PRINT(("Number of blocks converted : %d", numBlks));
    }
    else
    {
        /*
         * This is response to procList -> EVM_COD_??? -> sendCOD sequence
         */
        // Move the current processing block to next
        AM_DBG_PRINT(("Source -> process response from getDlta"));
        agtCb->currBlk = (agtCb->currBlk)->next;
    }

    c_blk = agtCb->currBlk;
    if (c_blk != NULL)
    {
#ifdef INCLUDE_EVM_ON_SOURCE
        /*
         * EVM request
         */
        bzero(&evmReq, sizeof(evm_IOReq_t));
        evmReq.reqType       = EVM_READ_REQ;
        evmReq.dataType      = EVM_USER_DATA_REQ;
        memcpy (&(evmReq.id), &(agtCb->amVolId), sizeof(evm_volId_t));
        evmReq.offset        = c_blk->blkOffSet;
        evmReq.reqSize       = c_blk->blkCnt;
        evmReq.clientPrivate = appHdl;

        rc = evm_IOReq(&evmReq, source_evmIOStat);
        M_POST_ABORT_EVENT(agtCb, "EVM-IO READ Failed", rc);
#else
        evm_IOStat_t evmStat;

        bzero(&evmStat, sizeof(evm_IOStat_t));
        evmStat.clientPrivate = appHdl;
        evmStat.status        = EVM_STAT_NO_ERROR;
        evmStat.rDataHandle.codHandle = 0;
        source_evmIOStat(&evmStat);
#endif
    }
    else
    {
        am_free_block_list(agtCb);
        bzero(&newEvent, sizeof(esm_event_t));
#ifndef INCLUDE_FS_ON_SOURCE
        agtCb->nextExt = FALSE;
#endif
        newEvent.event = SOURCE_EVENT_GET_NXT_LIST;
        newEvent.p1.u.dw = 0;
        newEvent.p2.u.dw = 0;

    	AM_FCT_PRINT(("Posting SOURCE_EVENT_GET_NXT_LIST %d to %d SM", \
                       newEvent.event, handle));
        rc = esm_deliverEvent(handle, &newEvent);
        NFX_ASSERT(rc == NFX_OK);
    }
    if (thisEvent.p1.u.p)
        am_eee_ramDealloc(thisEvent.p1.u.p);

    AM_DBG_PRINT(("...source_procList"));
    return (nextState);
}


/*
 *------------------------------------------------------------------------------
 * Name         : source_sendCOD
 *
 * Description  :
 *
 * Created by   : Nitin Jain
 *
 *------------------------------------------------------------------------------
 */
int32 source_sendCOD (esm_handle_t  handle,
                      int32         nextState,
                      esm_event_t   thisEvent,
                      void          *userInfo)
{
    am_agCb_t           *agtCb = (am_agCb_t *)userInfo;
    char                *req = NULL;
    dm_sendReq_t        *dmReq = NULL;
    am_data_msg_hdr_t   *amMsg = NULL;
    int32               dmReqLen = sizeof(dm_sendReq_t);
    int32               amMsgLen = sizeof(am_data_msg_hdr_t);
    int32               rc = 0, msgLen;
    evm_IOStat_t        *evmRsp = NULL;
    uint64              appHdl = (uint64)userInfo;

    AM_DBG_PRINT(("source_sendCOD..."));

    msgLen = dmReqLen + amMsgLen;
    req = (char *)am_eee_ramAlloc(msgLen);
    NFX_ASSERT(req != NULL);

    evmRsp = (evm_IOStat_t *)thisEvent.p1.u.p;
    dmReq = (dm_sendReq_t *)req;
    dmReq->hdr.msgType = DM_SEND_REQ;
    dmReq->hdr.appHdl  = appHdl;
    dmReq->hdr.dmHdl   = agtCb->dmHdl;

#ifdef INCLUDE_EVM_ON_SOURCE
    dmReq->dataSize    = EFS_BLOCK_SIZE;
    dmReq->cod         = evmRsp->rDataHandle.codHandle;
    dmReq->offset      = 0;
#else
    dmReq->dataSize    = 0;
    dmReq->cod         = 0;
    dmReq->offset      = 0;
#endif
    dmReq->appHdrSize  = amMsgLen;

    amMsg = (am_data_msg_hdr_t *)(req+dmReqLen);
    amMsg->am_msgType  = AM_SEND_DATA_IND;
    memcpy(&(amMsg->volId), &(agtCb->tgtVolId), sizeof(evm_volId_t));
    amMsg->blkOffSet   = agtCb->currBlk->blkOffSet;
    amMsg->blkCnt      = agtCb->currBlk->blkCnt;  // # of blocks

    AM_DBG_PRINT(("Source -> Send COD:  blk: %d cod: %lld datasize: %d",
            amMsg->blkOffSet, (uint64)dmReq->cod, (uint32)dmReq->dataSize));

    rc = eee_sendMessage(req, msgLen, datamoverAddr, amAddr);
    NFX_ASSERT(rc == NFX_OK);
    am_eee_ramDealloc(req);
    if (thisEvent.p1.u.p)
        am_eee_ramDealloc(thisEvent.p1.u.p);

    AM_DBG_PRINT(("...source_sendCOD"));
    return (nextState);
}


/*
 *------------------------------------------------------------------------------
 * Name         : source_sendDone
 *
 * Description  :
 *
 * Created by   : Nitin Jain
 *
 *------------------------------------------------------------------------------
 */
int32 source_sendDone (esm_handle_t  handle,
                       int32         nextState,
                       esm_event_t   thisEvent,
                       void          *userInfo)
{
    am_agCb_t           *agtCb = (am_agCb_t *)userInfo;
    char                *req = NULL;
    dm_sendReq_t        *dmReq = NULL;
    am_msg_hdr_t        *amMsg = NULL;
    int32               amMsgLen = sizeof(am_msg_hdr_t);
    int32               dmReqLen = sizeof(dm_sendReq_t);
    int32               rc = 0, msgLen;
    uint64              appHdl = (uint64)userInfo;

    AM_DBG_PRINT(("source_sendDone..."));

    msgLen = dmReqLen + amMsgLen;
    req = (char *)am_eee_ramAlloc(msgLen);
    NFX_ASSERT(req != NULL);

    dmReq = (dm_sendReq_t *)req;
    dmReq->hdr.msgType = DM_SEND_REQ;
    dmReq->hdr.appHdl  = appHdl;
    dmReq->hdr.dmHdl   = agtCb->dmHdl;

    dmReq->dataSize    = 0;
    dmReq->cod         = 0;
    dmReq->offset      = 0;
    dmReq->appHdrSize  = amMsgLen;

    amMsg = (am_msg_hdr_t *)(req+dmReqLen);
    amMsg->am_msgType  = AM_SEND_DONE_IND;
    memcpy (&(amMsg->volId), &(agtCb->tgtVolId), sizeof(evm_volId_t));

    AM_DBG_PRINT(("Source -> send done to target"));

    rc = eee_sendMessage(req, msgLen, datamoverAddr, amAddr);
    NFX_ASSERT(rc == NFX_OK);
    am_eee_ramDealloc(req);
    if (thisEvent.p1.u.p)
        am_eee_ramDealloc(thisEvent.p1.u.p);

    AM_DBG_PRINT(("...source_sendDone"));
    return (nextState);
}


#if 0
/*
 *------------------------------------------------------------------------------
 * Name         : source_nop
 *
 * Description  :
 *
 * Created by   : Nitin Jain
 *
 *------------------------------------------------------------------------------
 */
int32 source_nop (esm_handle_t  handle,
                  int32         nextState,
                  esm_event_t   thisEvent,
                  void          *userInfo)
{
    AM_DBG_PRINT(("source_nop..."));

    /*
     * Really!!! don't waste time :)
     */
    if (thisEvent.p1.u.p)
        am_eee_ramDealloc(thisEvent.p1.u.p);

    AM_DBG_PRINT(("...source_nop"));
    return (nextState);
}
#endif


/*
 *------------------------------------------------------------------------------
 * Name         : source_closeDlta
 *
 * Description  :
 *
 * Created by   : Nitin Jain
 *
 *------------------------------------------------------------------------------
 */
int32 source_closeDlta (esm_handle_t  handle,
                        int32         nextState,
                        esm_event_t   thisEvent,
                        void          *userInfo)
{
#ifdef INCLUDE_FS_ON_SOURCE
    am_agCb_t                  *agtCb = (am_agCb_t *)userInfo;
    fs_MsgAmDeltaCompleteReq_t fsReq;
    am_rc_t                    rc;
    uint64                     appHdl = (uint64)userInfo;

    AM_DBG_PRINT(("source_closeDlta..."));

    bzero(&fsReq, sizeof(fs_MsgAmDeltaCompleteReq_t));
    fsReq.hdr.requestType   = FS_REQ_AM_DELTA_COMPLETE;
    fsReq.hdr.len           = sizeof(fs_MsgAmDeltaCompleteReq_t);
    fsReq.hdr.clientPrivate = appHdl;
    fsReq.cookie            = appHdl;
    memcpy(&(fsReq.volId), &(agtCb->amVolId), sizeof(evm_volId_t));
	strcpy(fsReq.snapName, agtCb->newSnap);

	AM_FCT_PRINT(("Sending FS_REQ_AM_DELTA_COMPLETE to 0x%x\n", efsAddr));
    rc = eee_sendMessage(&fsReq,
                         sizeof(fs_MsgAmDeltaCompleteReq_t),
                         efsAddr,
                         amAddr);
    NFX_ASSERT(rc == NFX_OK);
    if (thisEvent.p1.u.p)
        am_eee_ramDealloc(thisEvent.p1.u.p);

	//source_mirrorStatus(agtCb->ctrlHdl, AM_MIRROR_COMPLETED);

    AM_DBG_PRINT(("...source_closeDlta"));
    return (nextState);
#else
    am_agCb_t                   *agtCb = (am_agCb_t *)userInfo;
    am_rc_t                     rc;
    fs_MsgAmDeltaCompleteResp_t fsRsp;
    uint64                      appHdl = (uint64)userInfo;

    AM_DBG_PRINT(("source_closeDlta..."));

    bzero(&fsRsp, sizeof(fs_MsgAmDeltaCompleteResp_t));
    fsRsp.hdr.requestType   = FS_REQ_AM_DELTA_COMPLETE;
    fsRsp.hdr.len           = sizeof(fs_MsgAmDeltaCompleteResp_t);
    fsRsp.hdr.clientPrivate = appHdl;
    fsRsp.hdr.fsStatus      = FS_ERR_NONE;
    fsRsp.cookie            = appHdl;

    rc = eee_sendMessage(&fsRsp,
                         sizeof(fs_MsgAmDeltaCompleteResp_t),
                         amAddr,
                         amAddr);
    NFX_ASSERT(rc == NFX_OK);
    if (thisEvent.p1.u.p)
        am_eee_ramDealloc(thisEvent.p1.u.p);

	source_mirrorStatus(agtCb->ctrlHdl, AM_MIRROR_COMPLETED);

    AM_DBG_PRINT(("...source_closeDlta"));
    return (nextState);
#endif

}


/*
 *------------------------------------------------------------------------------
 * Name         : source_kill
 *
 * Description  :
 *
 * Created by   : Nitin Jain
 *
 *------------------------------------------------------------------------------
 */
int32 source_kill (esm_handle_t  handle,
                   int32         nextState,
                   esm_event_t   thisEvent,
                   void          *userInfo)
{
    am_agCb_t       *agtCb = (am_agCb_t *)userInfo;
    am_rc_t   rc;
	int32 curState;

    AM_FCT_PRINT(("source_kill..."));

	curState = esm_getCurrState(handle);

	if (curState == SOURCE_STATE_KILL_DM) 
	{
		source_mirrorStatus(agtCb->ctrlHdl, AM_MIRROR_COMPLETED);
	}
	else
	{
		source_mirrorStatus(agtCb->ctrlHdl, AM_MIRROR_ABORTED);
	}

    am_free_control_block(agtCb);

	printf("Terminating source SM %d\n", handle);

    rc = source_sm_destroy(handle);
    if (rc != NFX_OK)
        AM_ERR_PRINT(("Source : could not destroy sm handle %d", handle));

    if (thisEvent.p1.u.p)
        am_eee_ramDealloc(thisEvent.p1.u.p);

    AM_DBG_PRINT(("...source_kill"));
    return (nextState);
}


/*
 *------------------------------------------------------------------------------
 * Name         : source_killDm
 *
 * Description  :
 *
 * Created by   : Nitin Jain
 *
 *------------------------------------------------------------------------------
 */
int32 source_killDm (esm_handle_t  handle,
                     int32         nextState,
                     esm_event_t   thisEvent,
                     void          *userInfo)
{
    am_agCb_t       *agtCb = (am_agCb_t *)userInfo;
    dm_closeReq_t   dmReq;
    int32           rc = 0;
    uint64          appHdl = (uint64)userInfo;
	int32 			curState;

    AM_FCT_PRINT(("source_killDm..."));

	curState = esm_getCurrState(handle);

    dmReq.hdr.msgType = DM_CLOSE_REQ;
    dmReq.hdr.appHdl  = appHdl;
    dmReq.hdr.dmHdl   = agtCb->dmHdl;

    rc = eee_sendMessage(&dmReq, sizeof(dm_closeReq_t), datamoverAddr, amAddr);
    NFX_ASSERT(rc == NFX_OK);
    if (thisEvent.p1.u.p)
        am_eee_ramDealloc(thisEvent.p1.u.p);

	if (curState != SOURCE_STATE_DONE)
		source_mirrorStatus(agtCb->ctrlHdl, AM_MIRROR_ABORTED);

    AM_DBG_PRINT(("...source_killDm"));
    return (nextState);
}


/*
 *------------------------------------------------------------------------------
 * Name         : source_resetDlta
 *
 * Description  :
 *
 * Created by   : Nitin Jain
 *
 *------------------------------------------------------------------------------
 */
int32 source_resetDlta (esm_handle_t  handle,
                        int32         nextState,
                        esm_event_t   thisEvent,
                        void          *userInfo)
{
    AM_DBG_PRINT(("source_resetDlta..."));

    /*
     * TBD after the API from EFS!!!
     */

    if (thisEvent.p1.u.p)
        am_eee_ramDealloc(thisEvent.p1.u.p);

    AM_DBG_PRINT(("...source_resetDlta"));
    return (nextState);
}


/*
 *------------------------------------------------------------------------------
 * Name         : source_err
 *
 * Description  :
 *
 * Created by   : Nitin Jain
 *
 *------------------------------------------------------------------------------
 */
int32 source_err (esm_handle_t  handle,
                  int32         nextState,
                  esm_event_t   thisEvent,
                  void          *userInfo)
{
    AM_DBG_PRINT(("source_err..."));

    AM_ERR_PRINT(("SM err source: %d (handle) %d (next)", handle, nextState));

    if (thisEvent.p1.u.p)
        am_eee_ramDealloc(thisEvent.p1.u.p);

    AM_DBG_PRINT(("...source_err"));
    return (nextState);
}


/*----------------------------------------------------------------------------*/
#endif  /* SSC */
