/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: Agile Network Protocol on the SSC
 *
 * RCS:  $Id: anp_ssc.c,v 1.11 2002/01/28 23:47:05 raster Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Implements the functions to allow easy-to-use access
 *              of messages traveling around the intra-NFX network.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 7 June 2001
 *
 *-----------------------------------------------------------------
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include "nfx-incl.h"
#include "anpssc-api.h"
#include <exports/eee-api.h>

static const char *AgileName = "Agile";

int regAgileApp (char *appName)
{
	struct servent *serv;
	struct protoent *prot;
	struct sockaddr_in sockAdd;
	struct hostent *host;
	int sock;
/*
	Is this a proper "service"?
 */
	setservent(0);
	serv = getservbyname (appName, "udp");
	endservent();
	if (!serv) {
		perror ("regAgileApp: Invalid application");
		return -1;
	}
/*
	Create the socket and bind it to the requested port.
 */
	setprotoent(0);
	prot = getprotobyname ("udp");
	endprotoent();
	if (!prot) {
		perror ("regAgileApp: getprotobyname failed");
		return -1;
	}

	sock = socket (PF_INET, SOCK_DGRAM, prot->p_proto);
	if (sock < 0) {
		perror("regAgileApp: socket error");
		return -1;
	}

	memset (&sockAdd, 0, sizeof(sockAdd));
	sockAdd.sin_family = AF_INET;
	sockAdd.sin_port = serv->s_port;
/*
	Only accept connections from localhost. It would be a security hole
	to accept from anyone on the network.
 */
	host = gethostbyname("localhost");
	if (!host) {
		perror ("sendAgileMsg: gethostbyname failed");
		return -1;
	}
	memcpy (&sockAdd.sin_addr, host->h_addr, host->h_length);

	if (bind (sock, (struct sockaddr *)&sockAdd, sizeof(sockAdd)) < 0) {
		perror ("regAgileApp: bind error");
		close(sock);
		return -1;
	}

	return sock;
}

void unRegAgileApp(int fd)
{
	close(fd);
}

int recvAgileMsg (int fd,                   // From regAgileApp()
                  void *msgBuff,            // Where to put the message
                  ssize_t msgLen,           // and how much room there is
                  struct timeval *timeout,  // (opt) How long to wait
                  char *appName,			// (opt) From which application
                  uint32_t *cpuId,			// (opt) and where
                  uint32_t *slotId, 		// (opt) it is on in the NFX
                  uint32_t *rPlane)			// (opt) On which plane it came?
{
	struct sockaddr_in from;
	socklen_t fromlen;
	int stat;
	struct servent *serv;
	struct timeval to;
	char *rcvbuf;
	char *cpyfrom;
	int bufsize;
/*
	Wait for a message. Use select() if they specify a timeout.
	Otherwise, we just use recvfrom() and block until a message arrives.
 */
	if (timeout) {
		fd_set fds;
		FD_ZERO (&fds);
		FD_SET (fd, &fds);
		to = *timeout;		// Copy because Linux whacks it
		if (select (fd+1, &fds, NULL, NULL, &to) <= 0) {
			return 0;
		}
	}
/*
	Get the message
 */
	memset (&from, 0, sizeof(from));
	fromlen = sizeof(from);
	bufsize = sizeof (ans_ssc_msghdr_t) + msgLen;
	cpyfrom = rcvbuf = malloc (bufsize);
	stat = recvfrom (fd, rcvbuf, bufsize, 0, &from, &fromlen);
/*
	If this is from the routing daemon, it will have an ans_ssc_msghdr_t
	prepended to it. We need to convert that into the proper app/slot/cpu
	triplet.
 */
	if (ntohs(from.sin_port) >= AGILE_BASE) { /* It's a registered port */
		setservent(0);
		serv = getservbyport (from.sin_port, "udp");
		endservent();

		if (!serv) {
			fprintf (stderr, "Didn't find %d\n", ntohs (from.sin_port));
			perror ("recvAgileMsg: getservent error");
			stat = -1;
		}
		else {
/*
		When things come from the ssc router, they are from off-board and will
		have a header to allow us to figure out where they came from. We'll
		use that to translate into something that's usable by sendAgileMsg().
 */
			if (!strcmp (serv->s_name, "ssc")) {
				from.sin_port = 
				     htons(((ans_ssc_msghdr_t *)rcvbuf)->appid + AGILE_BASE);
/*
			Get the REAL application name
 */
				setservent(0);
				serv = getservbyport (from.sin_port, "udp");
				endservent();

				if (cpuId) {
					*cpuId = ((ans_ssc_msghdr_t *)rcvbuf)->cpu;
				}
				if (slotId) {
					*slotId = ((ans_ssc_msghdr_t *)rcvbuf)->slot;
				}
				if (rPlane) {
					*rPlane = ((ans_ssc_msghdr_t *)rcvbuf)->plane;
				}
/*
			Now we need to move the message payload to where it's expected.
			This evaluation had better always be true otherwise it's a NULL
			message. Not very useful.
 */
				stat -= sizeof(ans_ssc_msghdr_t);
				cpyfrom = &rcvbuf[sizeof(ans_ssc_msghdr_t)];
			}
			else {
				if (cpuId) {
					*cpuId = NFX_SSC_UNIX;
				}
				if (slotId) {
					*slotId = SSC_SLOT;
				}
				if (rPlane) {
					*rPlane = 0;
				}
				if (stat > msgLen) {
					stat = msgLen;
				}
			}
/*
	Return the name of the sender if they want it
 */
			if (appName) {
				strcpy (appName, serv->s_name);
			}
		}
	}
	else {
		if (cpuId) {
			*cpuId = NFX_SSC_UNIX;
		}
		if (slotId) {
			*slotId = SSC_SLOT;
		}
		if (rPlane) {
			*rPlane = 0;
		}
		if (appName) {		/* port on SSC waiting for response */
			sprintf (appName,"%s%d", AgileName, ntohs(from.sin_port));
		}
	}
	if (stat > msgLen) {
		stat = msgLen;
	}
	if (stat > 0) {
		memcpy (msgBuff, cpyfrom, stat);
	}
	free (rcvbuf);

	return stat;
}

int sendAgileMsg (int reg_sock,				// Use this socket please
                  void *msgBuff,            // The message
                  ssize_t msgLen,           // How many bytes it be
                  char *appName,            // To which application
                  uint32_t cpuId,           // and where
                  uint32_t slotId,          // it is on in the NFX
                  uint32_t whichPlane,      // 0=mgmt; 1=data
                  struct timeval *timeout,  // (opt) How long to wait for resp
                  void *respBuff,           // (opt) Where to put resp msg
                  ssize_t respLen)          // (opt) how big it is
{
	struct servent *serv;
	struct sockaddr_in sin;
	struct protoent *prot;
	struct hostent *host;
	int port;
	int sock;
	int stat;
	struct timeval to;
	unsigned char *tmpbuf;
	unsigned char *cpyfrom;
	ans_ssc_msghdr_t *pmsghdr;
	int create_sock = reg_sock < 0;
	char *dest_app;
	struct sockaddr_in resp_from;
	socklen_t resp_fromlen;
	int bufsize;
/*
	Let's first make sure this message will fit into an eLayer descriptor
	and JUMBO frame if it is to go outside UNIX. sendto() will limit the
	messages internal to UNIX.
 */

	tmpbuf = NULL;
	if (slotId == SSC_SLOT && cpuId == NFX_SSC_UNIX) {
		pmsghdr = (ans_ssc_msghdr_t *) msgBuff;
		dest_app = appName;
	} else {
		/* this is a message for the embedded world */
		if (msgLen > 8192) {
			errno = EMSGSIZE;
			return -1;
		}
		tmpbuf = malloc (sizeof(ans_ssc_msghdr_t) + msgLen);
		pmsghdr = (ans_ssc_msghdr_t *) tmpbuf;

		setservent(0);
		serv = getservbyname (appName, "udp");
		endservent();
		if (!serv) {
			printf("Invalid application name: %s\n", appName);
			return -1;
		}

		pmsghdr->appid = ntohs(serv->s_port) - AGILE_BASE;
		pmsghdr->slot  = slotId;
		pmsghdr->cpu   = cpuId;
		pmsghdr->plane = EEE_MGMT;

		memcpy(tmpbuf+sizeof(ans_ssc_msghdr_t), msgBuff, msgLen);
		msgLen += sizeof(ans_ssc_msghdr_t);
		dest_app = "ssc";
	}
/*
	Is this to a registered app or a response?
 */
	if (strncmp (dest_app, AgileName, strlen(AgileName))) {
/*
	If this is for NIM, need to send it to "ssc" daemon
*/
		setservent(0);
		serv = getservbyname (dest_app, "udp");
		endservent();
		if (!serv) {
			perror ("sendAgileMsg: Invalid application");
			return -1;
		}
		port = serv->s_port;
	}
	else { /* Response port */
		port = htons(atoi (&dest_app[strlen(AgileName)]));
	}
	
	if (create_sock) {
/*
	Create the socket and bind it to the requested port.
 */
		setprotoent(0);
		prot = getprotobyname ("udp");
		endprotoent();
		if (!prot) {
			perror ("sendAgileMsg: getprotobyname failed");
			return -1;
		}
/*
	Get the socket and send the message
 */
		sock = socket (PF_INET, SOCK_DGRAM, prot->p_proto);
		if (sock < 0) {
			perror ("sendAgileMsg: socket failed");
			return -1;
		}
	}
	else { // They gave us this socket to use
		sock = reg_sock;
	}
/*
	We send all messages to the localhost router
 */
	memset (&sin, 0, sizeof(sin));
	sin.sin_family = AF_INET;
	sin.sin_port = port;

	host = gethostbyname("localhost");
	if (!host) {
		perror ("sendAgileMsg: gethostbyname failed");
		return -1;
	}
	memcpy (&sin.sin_addr, host->h_addr, host->h_length);

	stat = sendto (sock, pmsghdr, msgLen, 0, &sin, sizeof(sin));
	if (stat != msgLen) {
		perror ("sendAgileMsg: sendto error");
		if (create_sock) {
			close (sock);
		}
		return -1;
	}

	if (tmpbuf) {	// Done with this for now
		free (tmpbuf);
		tmpbuf = NULL;
	}
/*
	Do we wait for a response?
 */
	stat = 0;
	if (respBuff) {
/*
		And do we wait only a specified time ?
 */
		if (timeout) {
			fd_set fds;
			FD_ZERO (&fds);
			FD_SET (sock, &fds);
			to = *timeout;		// Linux whacks it
			if (!select (sock+1, &fds, NULL, NULL, &to)) {
				if (create_sock) {
					close(sock);
				}
				return 0;
			}
		}
/*
		Get the message
 */
		bufsize = sizeof(ans_ssc_msghdr_t) + respLen;
		cpyfrom = tmpbuf = malloc (bufsize);

		resp_fromlen = sizeof(resp_from);
		stat = recvfrom (sock, tmpbuf, bufsize, 0, 
		                 &resp_from, &resp_fromlen);
/*
	If this is from the routing daemon, it will have an ans_ssc_msghdr_t
	prepended to it. We need get rid of it and just return the payload.
 */
		if (ntohs(resp_from.sin_port) >= AGILE_BASE) {	// ssc is registered
			setservent(0);
			serv = getservbyport (resp_from.sin_port, "udp");
			endservent();

			if (!serv) {
				fprintf (stderr, "Didn't find %d\n", 
				         ntohs (resp_from.sin_port));
				perror ("sendAgileMsg: getservent error");
				stat = -1;
			}
			else {
				if (!strcmp (serv->s_name, "ssc")) {
/*
			Now we need to move the message payload to where it's expected.
			This evaluation had better always be true otherwise it's a NULL
			message. Not very useful.
 */
					stat -= sizeof(ans_ssc_msghdr_t);
					cpyfrom = &tmpbuf[sizeof(ans_ssc_msghdr_t)];
				}
			}
		}
/*
	Now to return the response message without overflowing the caller's
	buffer.
 */
		if (stat > respLen) {
			stat = respLen;
		}

		if (stat > 0) {
			memcpy (respBuff, cpyfrom, stat);
		}
		free (tmpbuf);
	}

	if (create_sock) {
		close(sock);
	}

	return stat;
}
