/*
 *=============================================================================
 * File Name   :    bcm5600-vlan.c
 *
 *  RCS: $Id: bcm5600-vlan.c,v 1.1.1.1 2001/12/20 04:53:01 tonyw Exp $		Copied and modified from Broadcom
 *
 *  Description:
 *
 *  Created    :    09-04-2001
 *
 *  Author     :    Tony Wong
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *=============================================================================
 */

#include <nfx-incl.h>

#include "bcm5600-soctypes.h"
#include "bcm5600-cmic.h"
#include "bcm5600-schanmsg.h"
#include "bcm5600-error.h"
#include "bcm5600-port.h"
#include "bcm5600-l2.h"

#define BCM5600_VTABLE_MAX_INDEX        63

static int _bcm_vlan_init[N_SOC_DEVICES];
static vlan_data_t default_vlan[N_SOC_DEVICES];
static vtable_entry_t vtable_entry;

static uint32 bcm_zeroes[20] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

#define VLAN_INIT(unit) \
        if (!_bcm_vlan_init[unit]) {return BCM_E_INIT;}

/*
 * Function: bcm_vlan_init_5600
 *
 * Description:
 *	Build in-memory list of VLANs.
 *
 * Parameters:
 *      unit - StrataSwitch PCI device unit number (driver internal).
 *
 * Returns:
 *	BCM_E_NONE		Success.
 *	BCM_E_INTERNAL		Chip access failure.
 *
 * Notes:
 *      This routine registers VLAN module with PMUX.
 *      Handles packets which come to CPU because of clearn bit set in
 *      in PTABLE -- FOR VLAN DESTRUCTION ONLY. 
 */

int 
bcm_vlan_init_5600(int unit)
{
    int		index_min, index_max, index;

    vtable_entry.vlan_tag = ~0;

    /* Clear VTABLE */
    index_min = 0;
    index_max = BCM5600_VTABLE_MAX_INDEX;

    for (index = 0; index <= index_max; index++) {
	SOC_IF_ERROR_RETURN(
		cmic_write_memory(unit, VTABLE_m, COPYNO_ALL, index,
				  index < index_min ?
				  bcm_zeroes :
				  (void*)&vtable_entry));
    }

    default_vlan[unit].vlan_tag = 1;
    default_vlan[unit].port_bitmap = PBMP_ALL;
    default_vlan[unit].ut_port_bitmap = PBMP_ALL;

    /* insert calls to bcm_vlan_create_5600() here to add entries to VTABLE */
    /* start with index = 0,1,2,...for VLAN IDs in ascending order */

    _bcm_vlan_init[unit] = TRUE;

    return BCM_E_NONE;
}

/*
 * Function: bcm_vlan_create_5600
 *
 * Description:
 *	Add entry to VLAN Table (VTABLE) on StrataSwitch.
 *
 *      NOTE: VLAN table is 64 entries deep and sorted by VLAN ID.
 *            Call this routine with ascending order index and VLAN IDs only.
 *
 * Parameters:
 *      unit - StrataSwitch PCI device unit number (driver internal).
 *      vid - VLAN ID to create.
 *	pbmp - port bit map
 *	ubmp - untagged port bit map
 *
 * Returns:
 *	BCM_E_NONE		Success.
 *	BCM_E_INTERNAL		Chip access failure.
 *	BCM_E_VLAN_EXISTS	VLAN ID already in use.
 *
 * Notes:
 *	None.
 */

int
bcm_vlan_create_5600(int unit, vlan_id_t vid, int index, pbmp_t pbmp, pbmp_t ubmp)
{
    vtable_entry_t	vEntry;
    int rv = BCM_E_NONE;

    VLAN_INIT(unit);

    if (rv == BCM_E_NONE) {
        vEntry.vlan_tag = vid;
        vEntry.port_bitmap = pbmp;
        vEntry.ut_port_bitmap = ubmp;

        if (cmic_write_memory(unit, VTABLE_m, COPYNO_ALL, index, &vEntry) < 0) {
            rv = BCM_E_INTERNAL;
        }
    }

    return rv;
}

/*
 * Function: void bcm5600_vlan_default_get(int unit, vlan_data_t *vdata)
 *
 * Description:
 *	Get the default vlan information.
 *
 * Parameters:
 *	unit - StrataSwitch PCI device unit number (driver internal).
 *	vdata- Target to copy vlan data to.
 *
 * Returns:
 *	BCM_E_NONE		Success.
 *
 */
void
bcm5600_vlan_default_get(int unit, vlan_data_t *vdata)
{
    memcpy(vdata, &default_vlan[unit], sizeof(vlan_data_t));
}

/*
 * Function: void bcm5600_vlan_default_set(int unit, vlan_data_t *vdata)
 *
 * Description:
 *	Get the default vlan information.
 *
 * Parameters:
 *	unit - StrataSwitch PCI device unit number (driver internal).
 *	vdata- Target to copy vlan data to.
 *
 * Returns:
 *	BCM_E_NONE		Success.
 *
 */
int
bcm5600_vlan_default_set(int unit, vlan_data_t *vdata)
{
    VLAN_INIT(unit);
    memcpy(&default_vlan[unit], vdata, sizeof(vlan_data_t));
    return BCM_E_NONE;
}

/*
 * Internal functions to convert between marl_entry_t and bcm_mcast_addr_t
 */

void _bcm_mcast_from_marl(bcm_mcast_addr_t *mcaddr, marl_entry_t *marl_entry)
{
    ARL_MAC_GET(*marl_entry, mcaddr->mac);
    mcaddr->vid = marl_entry->vlan_tag;
    mcaddr->cos_dst = marl_entry->cos_dst;
    mcaddr->pbmp = marl_entry->port_bitmap;
    mcaddr->ubmp = marl_entry->ut_port_bitmap;
}

void _bcm_mcast_to_marl(marl_entry_t *marl_entry, bcm_mcast_addr_t *mcaddr)
{
    memset(marl_entry, 0, sizeof (*marl_entry));
    ARL_MAC_SET(*marl_entry, mcaddr->mac);
    marl_entry->vlan_tag = mcaddr->vid;
    marl_entry->cos_dst = mcaddr->cos_dst;
    marl_entry->port_bitmap = mcaddr->pbmp;
    marl_entry->ut_port_bitmap = mcaddr->ubmp;
}

/*
 * Function: bcm_mcast_addr_init
 * Description:
 *	Initialize a bcm_mcast_addr_t to a specified MAC address and VLAN,
 *	while zeroing all other fields.
 * Parameters:
 *	mcaddr - Pointer to bcm_mcast_addr_t
 * Returns:
 *	Nothing.
 * Notes:
 *	None.
 */

void bcm_mcast_addr_init(bcm_mcast_addr_t *mcaddr,
			 mac_addr_t mac, vlan_id_t vid)
{
    memset(mcaddr, 0, sizeof (*mcaddr));
    memcpy(mcaddr->mac, mac, sizeof (mac_addr_t));
    mcaddr->vid = vid;
}

/*
 * Function:
 *      bcm_mcast_addr_add
 * Purpose:
 *      Add an entry in the multicast table
 * Parameters:
 *      unit - StrataSwitch unit number.
 *	mcaddr - Pointer to fully filled-in bcm_mcast_addr_t
 * Returns:
 *      BCM_E_NONE
 *      BCM_E_INTERNAL if table write failed.
 * Notes:
 */

int 
bcm_mcast_addr_add(int unit, bcm_mcast_addr_t *mcaddr)
{
    marl_entry_t 	mc_entry;
    int			rv;

    _bcm_mcast_to_marl(&mc_entry, mcaddr);

    rv = cmic_insert_memory(unit, MARL_m, COPYNO_ALL, &mc_entry);

    return(rv);
}

