/*
 *=============================================================================
 * File Name   :    bcm5632-link.c
 *
 *  RCS: $Id: bcm5632-link.c,v 1.2 2002/10/09 20:34:21 tonyw Exp $
 *
 *  Description: Copied and modified from Broadcom
 *
 *  This module contains routines for detecting changes in link state and
 *  performing the appropriate actions, such as reprogramming the MAC
 *  speed/duplex to correspond with the negotiated PHY speed/duplex.
 *
 *  Created    :    09-04-2001
 *
 *  Author     :    Tony Wong
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *=============================================================================
 */
#include <nfx-incl.h>

#include "bcm5632-soctypes.h"
#include "bcm5632-cmic.h"
#include "bcm5632-error.h"
#include "bcm5632-port.h"

extern uint32 reg_addr_5632(soc_reg_t reg, ...);

typedef struct ls_cntl_s {
    int			lc_init;	/* TRUE if initialized */
    pbmp_t		lc_pbm_hw;	/* Hardware link scan ports */
    pbmp_t		lc_pbm_sw;	/* Software link scan ports */
    pbmp_t		lc_pbm_last;	/* Last scan value */
} ls_cntl_t;

ls_cntl_t	bcm5632_link_control[N_SOC_DEVICES];

typedef struct _linkscan_save_s {
    int   aneg_saved;       /* if the force mode being saved set to 1 else 0 */
    int   aneg_enb;         /* Auto-Negotiation Enable bit original value */
    int   force_mode_saved; /* force mode setting */ 
    int   force_mode;       /* force mode setting */ 
} linkscan_save_t;

static linkscan_save_t _linkscan_save[N_SOC_DEVICES][NUM_E_PORT_5632];

/*
 * Function: 	
 *	_bcm_linkscan_update_sw
 * Purpose:	
 *	Look at all the PHYs and check for link UP/DOWN.
 * Parameters:	
 *	unit - StrataSwitch unit #.
 * Returns:	
 *	Port bit map, 1 indicates link active, 0 indicates link NOT active.
 */
static	pbmp_t
_bcm_linkscan_update_sw(int unit)
{
    ls_cntl_t		*lc = &bcm5632_link_control[unit];
    int			up;
    int			rv;
    pbmp_t		ls;		/* Return value */
    soc_port_t		port;		/* Port # */
    
    ls = 0;

    PBMP_ITER(lc->lc_pbm_sw, port) {
		if ((rv = bcm5632_port_link_status(unit, port, &up)) != BCM_E_NONE) {
	    	printf("LinkScan: Unit %d Port %d: Failed to recover link status\n",
		       		unit, port + 1);
		} 
		else 
		if (up) {
	    	ls |= PBMP_PORT(port);
		}
    }
    return(ls);
}

/*
 * Function: 	
 *	_bcm_linkscan_update_hw
 * Purpose:	
 *	Get current link status for ports being scanned by hardware.
 * Parameters:	
 *	unit - StrataSwitch unit #.
 * Returns:	
 *	Port bit map, 1 indicates link active, 0 indicates link NOT active.
 */
static pbmp_t
_bcm_linkscan_update_hw(int unit)
{
    ls_cntl_t	*lc = &bcm5632_link_control[unit];
    pbmp_t     	ls = 0;
    uint32     	data;
    int	       	rv, up;
    soc_port_t 	port;

    PBMP_ITER(lc->lc_pbm_hw, port) {
		if ((rv = cmic_read_reg_5632(unit, 
				  reg_addr_5632(P_STATUSr,port), &data)) != BCM_E_NONE) {
	    	printf("LinkScan: Unit %d Port %d: Failed to get link status\n",
		       		unit, port + 1);
		} 
		else 
		if (!GET_P_STATUSr_LinkFail(data)) { /* 0=link up */
	    	ls |= PBMP_PORT(port);
		} 
		else { /* the bit needed to be reset by software */
	    	SET_P_STATUSr_LinkFail(data, 0); 
            cmic_write_reg_5632(unit, reg_addr_5632(P_STATUSr, port), data);
            /* if this bit is set, it might not synchronize with PHY */
	    	if ((rv = bcm5632_port_link_status(unit, port, &up)) != BCM_E_NONE)
			{
	        	printf("LinkScan: Unit %d Port %d: Failed to get link status\n",
		           		unit, port + 1);
	    	} 
			else 
			if (up) {
	        	ls |= PBMP_PORT(port);
	    	}
        }
    }
    return(ls);
}

/*
 * Function:
 *	bcm5632_linkscan_update
 * Purpose:
 *	Performs actual scan and update of all links that have changed.
 * Parameters:
 *	unit - StrataSwitch unit #
 * Returns:
 *	Nothing
 */
void
bcm5632_linkscan_update(int unit)
{
    ls_cntl_t		*lc = &bcm5632_link_control[unit];
    int			link;
    int			rv = 0;
    soc_port_t		port;
    pbmp_t		ls_current, ls_changed, ls_ports;
    bcm_port_info_t	info;

    ls_ports   = lc->lc_pbm_hw | lc->lc_pbm_sw;
    ls_current = _bcm_linkscan_update_sw(unit) | _bcm_linkscan_update_hw(unit);
    ls_changed = (ls_current ^ lc->lc_pbm_last) & ls_ports & PBMP_E_ALL;

    PBMP_ITER(ls_changed, port) {
		if ((rv = bcm5632_port_update(unit, port, &link)) != BCM_E_NONE) {
	    	if (rv == BCM_E_AUTONEG_BUSY) {
				/* This requires another scan */
				ls_current &= ~PBMP_PORT(port);
	    	} else {
				 printf("LinkScan: Unit %d Port %d: bcm5632_port_update failed\n", 
			   		     unit, port + 1);
				 continue;
	    	}
		} 
		if ((rv = bcm5632_port_info_get(unit, port, &info)) != BCM_E_NONE) {
	    	printf("LinkScan: Unit %d Port %d: bcm5632_port_info_get failed\n", 
		           unit, port + 1);
	    	continue;
		}

		/* Display state changes if requested. */

		if (info.linkstatus) {
	    	printf("LinkScan: Unit %d Port %d: Link active: %dMb/s %s\n", unit, 
               port + 1, info.speed, info.duplex ? "Full-duplex":"Half-duplex");
		} else {	
	         printf("LinkScan: Unit %d Port %d: Link inactive\n", unit, port+1);
		}    
    }
    lc->lc_pbm_last = ls_current;
}

/*
 * Function: 	
 *	_save_linkscan
 * Purpose:
 *      Save the current setting of the auto-negotiation and force_mode.
 * Parameters:
 *	unit - StrataSwitch Unit #.
 *	port - StrataSwitch Port #.
 * Returns:
 *	BCM_E_XXX
 */
int
_save_linkscan(int unit, soc_port_t port) 
{
    uint32   data;

    /* set force_mode to 0x0b, and enable A/N, so the linkup status in switch 
       will be updated automatically */

    if (_linkscan_save[unit][port].aneg_saved == 0 ) { /* not saved before */
        SOC_IF_ERROR_RETURN(cmic_read_reg_5632(unit, 
							reg_addr_5632(P_CONF_2r, port), &data));
        _linkscan_save[unit][port].aneg_enb   = GET_P_CONF_2r_AnegEnb(data);
        SET_P_CONF_2r_AnegEnb(data, 1);
        SOC_IF_ERROR_RETURN(cmic_write_reg_5632(unit, 
							reg_addr_5632(P_CONF_2r, port), data));
    
        _linkscan_save[unit][port].aneg_saved = 1; 
    }

    if (_linkscan_save[unit][port].force_mode_saved == 0) { 
        SOC_IF_ERROR_RETURN(cmic_read_reg_5632(unit, 
							reg_addr_5632(P_CONF_3r, port), &data));
        _linkscan_save[unit][port].force_mode = GET_P_CONF_3r_Force_mode(data); 
        /* set the force_mode to 0xb */
        SET_P_CONF_3r_Force_mode(data, 0xb);
        SOC_IF_ERROR_RETURN(cmic_write_reg_5632(unit, 
							reg_addr_5632(P_CONF_3r, port), data));
        _linkscan_save[unit][port].force_mode_saved = 1; 
    }
    return(BCM_E_NONE);
}

/*
 * Function: 	
 *	_restore_linkscan
 * Purpose:
 *      Restore the current setting of the auto-negotiation and force_mode.	
 * Parameters:
 *	unit - StrataSwitch Unit #.
 *	port - StrataSwitch Port #.
 * Returns:
 *	BCM_E_XXX
 */
int
_restore_linkscan(int unit, soc_port_t port) 
{
    uint32   data;
    int      rv1, rv2;

    rv1 = rv2 = BCM_E_NONE;
    
    if (_linkscan_save[unit][port].aneg_saved != 0) { /* saved before */
        rv1 = cmic_read_reg_5632(unit, reg_addr_5632(P_CONF_2r, port), &data);
        if (rv1 == BCM_E_NONE) {
            SET_P_CONF_2r_AnegEnb(data, _linkscan_save[unit][port].aneg_enb);
            rv1 = cmic_write_reg_5632(unit, reg_addr_5632(P_CONF_2r,port),data);
            if (rv1 == BCM_E_NONE) 
                _linkscan_save[unit][port].aneg_saved = 0;
        }
    } 
    if (_linkscan_save[unit][port].force_mode_saved != 0) { /* saved before */
        rv2 = cmic_read_reg_5632(unit, reg_addr_5632(P_CONF_3r, port), &data);
        if (rv2 == BCM_E_NONE) {
            SET_P_CONF_3r_Force_mode(data, _linkscan_save[unit][port].force_mode);
            rv2 = cmic_write_reg_5632(unit, reg_addr_5632(P_CONF_3r,port),data);
            if (rv2 == BCM_E_NONE)
                _linkscan_save[unit][port].force_mode_saved = 0;
        }
    }
    return((rv1 == BCM_E_NONE) ? rv2 : rv1);
}

/*
 * Function: 	
 *	bcm_linkscan_port_set
 * Purpose:
 *	Set the current scanning mode for the specified port.
 * Parameters:
 *	unit - StrataSwitch Unit #.
 *	port - StrataSwitch Port #.
 *	mode - Net scan mode for specified port.
 * Returns:
 *	BCM_E_XXX
 */
int
bcm5632_linkscan_port_set(int unit, soc_port_t port, int mode)
{
    ls_cntl_t	*lc = &bcm5632_link_control[unit];
    int		rv = BCM_E_NONE;

    if (!PBMP_MEMBER(PBMP_E_ALL, port)) {
		return(BCM_E_PARAM);
    }
    if (lc->lc_pbm_hw & PBMP_PORT(port)) { /* Disable H/W link scan */
		lc->lc_pbm_hw &= ~PBMP_PORT(port);
        _restore_linkscan(unit, port);
    } 
	else 
	if (lc->lc_pbm_sw & PBMP_PORT(port)) {
		lc->lc_pbm_sw &= ~PBMP_PORT(port);
    }
    
    switch(mode) {
    case BCM_PORT_LINKSCAN_NONE:	/* Do nothing */
		break;
    case BCM_PORT_LINKSCAN_SW:
		lc->lc_pbm_sw |= PBMP_PORT(port);
		break;
    case BCM_PORT_LINKSCAN_HW:
		lc->lc_pbm_hw |= PBMP_PORT(port);
        _save_linkscan(unit, port);
		break;
    default:
		rv = BCM_E_PARAM;
		break;
    }
    return(rv);
}

