/*
 *=============================================================================
 * File Name   :    bcm5632-vlan.c
 *
 *  RCS: $Id: bcm5632-vlan.c,v 1.2 2002/10/09 20:27:31 tonyw Exp $
 *
 *  Description:    Copied and modified from Broadcom
 *
 *  Created    :    09-04-2001
 *
 *  Author     :    Tony Wong
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *=============================================================================
 */
#include <nfx-incl.h>
#include <exports/eee-api.h>

#include "bcm5632-soctypes.h"
#include "bcm5632-cmic.h"
#include "bcm5632-socregs.h"
#include "bcm5632-error.h"
#include "bcm5632-port.h"
#include "bcm5632-phy.h"

extern uint32 reg_addr_5632(soc_reg_t reg, ...);
extern int _BCM5632_table_management(int unit, soc_mem_t mem, void *key_data, int index, int cmd);


static int _bcm_vlan_init[N_SOC_DEVICES];
static vlan_data_t default_vlan[N_SOC_DEVICES];

#define VLAN_INIT(unit) \
        if (!_bcm_vlan_init[unit]) {return BCM_E_INIT;}

/* Declare static functions */
static int _bcm_vlan_port_add(int unit, vlan_id_t vid, pbmp_t pbmp, pbmp_t ubmp);

/*
 * Function: bcm5632_vlan_init
 *
 * Description:
 *	Build in-memory list of VLANs.
 *
 * Parameters:
 *      unit - StrataSwitch PCI device unit number (driver internal).
 *
 * Returns:
 *	BCM_E_NONE		Success.
 *	BCM_E_INTERNAL		Chip access failure.
 *
 * Notes:
 *      This routine registers VLAN module with PMUX.
 *      Handles packets which come to CPU because of clearn bit set in
 *      in PTABLE -- FOR VLAN DESTRUCTION ONLY. 
 */

int 
bcm5632_vlan_init(int unit)
{
	uint32 index;

    default_vlan[unit].vlan_tag = 1;
    default_vlan[unit].port_bitmap = PBMP_ALL;
    default_vlan[unit].ut_port_bitmap = PBMP_ALL;

/* in BCM5632 the chip init vlan entry to all 0 already */
    _bcm_vlan_init[unit] = TRUE;

	/* Init VLAN port map */
	for (index = 0; index < NUM_GE_PORT_5632; index++) {
		cmic_write_reg_5632(unit, CMIC_REG(P_VLAN_MAP_0r + index), 0xfff);
	}
    return BCM_E_NONE;
}

/*
 * Function: bcm5632_vlan_port_add
 *
 * Description:
 *      Add a port to the specified vlan.
 *
 * Parameters:
 *      unit - StrataSwitch PCI device unit number (driver internal).
 *      vid - VLAN ID to add port to as a member.
 *      pbmp - port bitmap for members of VLAN
 *      ubmp - untagged members of VLAN
 *
 * Returns:
 *      BCM_E_NONE              Success.
 *      BCM_E_INTERNAL          Chip access failure.
 *      BCM_E_VLAN_NOT_FOUND    VLAN ID not in use.
 *
 * Notes:
 *      None.
 */
int
bcm5632_vlan_port_add(int unit, vlan_id_t vid, pbmp_t pbmp, pbmp_t ubmp)
{
    int rv;

    VLAN_INIT(unit);
    rv = _bcm_vlan_port_add(unit, vid, pbmp, ubmp);
    return rv;
}

/*
 * Function: _bcm_vlan_port_add
 *    Guts of bcm5632_vlan_port_add.  See above.
 */
static int
_bcm_vlan_port_add(int unit, vlan_id_t vid, pbmp_t pbmp, pbmp_t ubmp)
{
    uint32 buf[2];
    int index;
    VLAN_m_entry_t entry;

	memset(&entry, 0, sizeof(VLAN_m_entry_t));
	index = vid;
    ubmp &= pbmp;  /* only allow untagged ports on the vlan. */
#if 0
	_BCM5632_table_management(unit, VLAN_m, &buf[0], index, TABLE_READ);  
	buf2struct_VLAN_5632(&entry, &buf[0]);     
#endif
    /*
     * Update the entry for this VLAN.
     */
    entry.PortMap |= pbmp;
    entry.OutputTagUntag |= ubmp;
	entry.PortState = 0xffffff; /* set all ports to forwarding state */
/*
    Struct2Buf(VLAN_m, (unsigned char *)&entry, (unsigned char *)&buf[0], 1);
*/
	struct2buf_VLAN_5632(&entry, &buf[0]);
/*
printf("vlan_port_add: PortMap = %0x, OutputTagUntag = %0x, raw = %0x %0x\n", 
		entry.PortMap, entry.OutputTagUntag, buf[0], buf[1]);
*/
	_BCM5632_table_management(unit, VLAN_m, &buf[0], index, TABLE_WRITE);  

    if (vid == default_vlan[unit].vlan_tag) {
        default_vlan[unit].port_bitmap |= pbmp;
        default_vlan[unit].ut_port_bitmap |= ubmp;
    }

    return BCM_E_NONE;
}

/*
 * Function: bcm5632_vlan_port_get
 *
 * Description:
 *      Retrieves the contents of an existing VLAN.
 *
 * Parameters:
 *      unit - StrataSwitch PCI device unit number (driver internal).
 *      vid - VLAN ID to lookup.
 *      tag_pbmp - (output) member port bitmap (ignored if NULL)
 *      untag_pbmp - (output) untagged port bitmap (ignored if NULL)
 *
 * Returns:
 *      BCM_E_NONE              Success (port bitmaps filled in).
 *      BCM_E_INTERNAL          Chip access failure.
 *      BCM_E_VLAN_NOT_FOUND    No such VLAN defined.
 *
 * Notes:
 *      Must not lock VLAN stuff.
 */

int
bcm5632_vlan_port_get(int unit, vlan_id_t vid, pbmp_t *pbmp, pbmp_t *ubmp)
{
    VLAN_m_entry_t entry;
    int rc;
    _u32 buf[MAX_MEM_WORDS];

    rc = cmic_read_memory_5632(unit, VLAN_m, COPYNO_ALL, vid, buf);
    if (rc != SOC_E_NONE) {
       return rc;
    }

    Buf2Struct(VLAN_m, (unsigned char *)&buf[0], (unsigned char *)&entry, 1);
 
    if (pbmp) {
        *pbmp = entry.PortMap;
    }

    if (ubmp) {
        *ubmp = entry.OutputTagUntag;
    }

    return BCM_E_NONE;
}

/*
 * Function: bcm5632_vlan_ifilter_set
 *
 * Description:
 *	Set input filter mode for all ports in a VLAN.
 *
 * Parameters:
 *	unit - StrataSwitch PCI device unit number (driver internal).
 *	vlan - VLAN number to operate on
 *	mode - BCM_VLAN_IFILTER_ON or BCM_VLAN_IFILTER_OFF.
 *
 * Returns:
 *	BCM_E_NONE		Success.
 *	BCM_E_INTERNAL		Chip access failure.
 *
 * Notes:
 *	All ports should be added to VLAN prior to setting filter mode.
 *	When input filtering is turned on for a vlan, packets received
 *	on any of the VLAN ports that do not match the port's VLAN
 *	classifications are discarded.
 *	See also: bcm_port_ifilter_set
 */

int
bcm5632_vlan_ifilter_set(int unit, vlan_id_t vid, int mode)
{
    pbmp_t		pbmp, ubmp;
    soc_port_t		port;

    VLAN_INIT(unit);
    BCM_IF_ERROR_RETURN(bcm5632_vlan_port_get(unit, vid, &pbmp, &ubmp));
    PBMP_ITER(pbmp, port) {
	BCM_IF_ERROR_RETURN(bcm5632_port_ifilter_set(unit, port, mode));
    }

    return BCM_E_NONE;
}

/*
 * Function: void bcm5632_vlan_default_get(int unit, vlan_data_t *vdata)
 *
 * Description:
 *	Get the default vlan information.
 *
 * Parameters:
 *	unit - StrataSwitch PCI device unit number (driver internal).
 *	vdata- Target to copy vlan data to.
 *
 * Returns:
 *	BCM_E_NONE		Success.
 *
 */
void
bcm5632_vlan_default_get(int unit, vlan_data_t *vdata)
{
    memcpy(vdata, &default_vlan[unit], sizeof(vlan_data_t));
}

/*
 * Function: void bcm5632_vlan_default_set(int unit, vlan_data_t *vdata)
 *
 * Description:
 *	Get the default vlan information.
 *
 * Parameters:
 *	unit - StrataSwitch PCI device unit number (driver internal).
 *	vdata- Target to copy vlan data to.
 *
 * Returns:
 *	BCM_E_NONE		Success.
 *
 */
int
bcm5632_vlan_default_set(int unit, vlan_data_t *vdata)
{
    VLAN_INIT(unit);
    memcpy(&default_vlan[unit], vdata, sizeof(vlan_data_t));
    return BCM_E_NONE;
}

