/*  File Name  :    agilefsetmain.c 
 *
 *  Description:    HTTPD action functions for main FileSet Page.
 *
 *  Created    :    3/19/2002
 *
 *  Author     :    Tiffany Chiang
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <anpssc-api.h>

#include <nfx-defs.h>
#include <nfx-types.h>
#include <evm_com_defs.h>
#include <evm-api.h>
#include <fs-api.h>

#include "agilepost.h"
#include "agileutil.h"

#define INVALID_TAG			0
#define FS_VOLID			1
#define FS_VOLNAME			2
#define FS_COLSPAN			3
#define FS_PATH				4
#define FS_INODE			5
#define FS_BLOCK			6
#define FS_QUOTASTS			7
#define FS_QUOTASTS_N		8
#define FS_QUOTADETAIL		9
#define FS_DELETE			10
#define NUM_OF_TAGS			10

static tagNameTable_t tagTable[NUM_OF_TAGS] = {
	{FS_VOLID,		"args.vol_id"},
	{FS_VOLNAME,	"args.vol_name"},
	{FS_COLSPAN,	"fileset_entry.colspan"},
	{FS_PATH,		"fileset_entry.path"},
	{FS_INODE,		"fileset_entry.inode_used"},
	{FS_BLOCK,		"fileset_entry.block_used"},
	{FS_QUOTASTS,	"fileset_entry.quota_status"},
	{FS_QUOTASTS_N,	"fileset_entry.quota_status.not"},
	{FS_QUOTADETAIL,"quotadetail"},
	{FS_DELETE,		"delete"},
};

#define PARAM_ACT			1
#define PARAM_VOLID			2
#define PARAM_VOLNAME		3
#define PARAM_FSPATH		4
#define PARAM_STATUS		5
#define NUM_OF_PARAMS		5

static paramNameTable_t paramTable[NUM_OF_PARAMS] = {
	{PARAM_ACT,		"act"},
	{PARAM_VOLID,	"id"},
	{PARAM_VOLNAME,	"vn"},
	{PARAM_FSPATH,	"path"},
	{PARAM_STATUS,	"status"},
};


/* HERE ARE VARIABLES FOR FILESET INFORMATION */
#define MAX_NUM_OF_FS		16

typedef struct _fileset_entry_t
{
    uint32		id;			// fileset id
    uint64		rootInode;	// inode number of the fileset
	char8		path[FS_PATH_MAX];
	uint64		inodeUsed;
	uint64		blockUsed;
	int			quotaOn;
} fileset_entry_t;

static char inAct[10];
static uint64 inId;
static char inName[VOL_MAX_NAME+1];
static char inPath[FS_PATH_MAX+1];
static int inStatus;

static int dataCount;
static int dataCurr;
static fileset_entry_t stData;
static fileset_entry_t *fsList;


/***********************************************************
 *  Main routines to handle Add/Delete/Qouta On/Off actions
 **********************************************************/

static int getInputParams(int count, char **argname, char **argval)
{
	int argind, paramType;

	for (argind = 0; argind < count; argind ++)
	{
		paramType = GetParamType(argname[argind], paramTable, NUM_OF_PARAMS);
		if (paramType == INVALID_PARAM)
			continue;

		switch (paramType) 
		{
		case PARAM_ACT:
			strcpy(inAct, argval[argind]);
			break;
		case PARAM_VOLID:
			sscanf(argval[argind], "%lld", &inId);
			break;
		case PARAM_VOLNAME:
			strcpy(inName, argval[argind]);
			break;
		case PARAM_FSPATH:
			strcpy(inPath, argval[argind]);
			convertUrlCharToSpecChar(inPath);
			break;
		case PARAM_STATUS:
			if (!strcmp(argval[argind], "On") || !strcmp(argval[argind], "on"))
				inStatus = 1;
			else
				inStatus = 0;
			break;
		}
	}

	return 0;
}

static int DoFilesetDel()
{
    int							stat;
	fs_MsgRemoveFileSetReq_t	req;
	fs_MsgRemoveFileSetResp_t	rsp;
	struct timeval			timeout = {60, 0};

    /* Fill the request. */
    memset((void *) &req, 0, sizeof(fs_MsgRemoveFileSetReq_t));
	req.msgHdr.requestType = FS_REQ_REMOVE_FILESET;
	req.fsid.whole_thing = inId;
	strcpy(req.path, inPath);

    /* Send message to FS. */
    stat = sendAgileMsg(-1, &req, sizeof(req),
                        "efs", NFX_NFP_FP2, 1, MGMTPLANE, &timeout,
                        &rsp, sizeof(rsp));

    /* Check response from FS. */
	if (stat <= 0) 
		return -1;
	if (rsp.msgHdr.fsStatus != FS_ERR_NONE)
		return -1;

	return 1;
}

static int DoFilesetMod()
{
    int							stat;
	fs_MsgQuotaOnReq_t			onreq;
	fs_MsgQuotaOnResp_t			onrsp;
	fs_MsgQuotaOffReq_t			offreq;
	fs_MsgQuotaOffResp_t		offrsp;
	struct timeval			onTimeout = {600, 0};
	struct timeval			offTimeout = {60, 0};

	// turn quota on
	if (inStatus == 1)
	{
	    /* Fill the request. */
		memset((void *) &onreq, 0, sizeof(fs_MsgQuotaOnReq_t));
		onreq.msgHdr.requestType = FS_REQ_QUOTA_ON;
		onreq.fsid.whole_thing = inId;
		strcpy(onreq.path, inPath);

		/* Send message to FS. */
		stat = sendAgileMsg(-1, &onreq, sizeof(onreq),
			                "efs", NFX_NFP_FP2, 1, MGMTPLANE,
							&onTimeout, &onrsp, sizeof(onrsp));

	    /* Check response from FS. */
		if (stat <= 0) 
			return -1;

		if (onrsp.msgHdr.fsStatus != FS_ERR_NONE)
			return -1;
	}
	// turn quota off
	else
	{
	    /* Fill the request. */
		memset((void *) &offreq, 0, sizeof(fs_MsgQuotaOffReq_t));
		offreq.msgHdr.requestType = FS_REQ_QUOTA_OFF;
		offreq.fsid.whole_thing = inId;
		strcpy(offreq.path, inPath);

		/* Send message to FS. */
		stat = sendAgileMsg(-1, &offreq, sizeof(offreq),
			                "efs", NFX_NFP_FP2, 1, MGMTPLANE,
				            &offTimeout, &offrsp, sizeof(offrsp));

	    /* Check response from FS. */
		if (stat <= 0) 
			return -1;
		if (offrsp.msgHdr.fsStatus != FS_ERR_NONE)
			return -1;
	}
	return 1;
}

static int DoFilesetAdd()
{
    int							stat;
	fs_MsgCreateFileSetReq_t	req;
	fs_MsgCreateFileSetResp_t	rsp;
	struct timeval			timeout = {60, 0};

    /* Fill the request. */
    memset((void *) &req, 0, sizeof(fs_MsgCreateFileSetReq_t));
	req.msgHdr.requestType = FS_REQ_CREATE_FILESET;
	req.fsid.whole_thing = inId;
	strcpy(req.path, inPath);

    /* Send message to FS. */
    stat = sendAgileMsg(-1, &req, sizeof(req),
                        "efs", NFX_NFP_FP2, 1, MGMTPLANE, &timeout,
                        &rsp, sizeof(rsp));

    /* Check response from FS. */
	if (stat < 0) 
		return -1;
	else if (!stat)
		return -1;
    if (rsp.msgHdr.requestType != FS_REQ_CREATE_FILESET)
		return -1;
	if (rsp.msgHdr.fsStatus != FS_ERR_NONE)
		return -1;

	return 1;
}

static int HandleFsAction(char *newfname, int count, char **argname, char **argval)
{
	int res;

	inAct[0] = '\0';
	inId = 0;
	inName[0] = '\0';
	inPath[0] = '\0';
	inStatus = 0;

	getInputParams(count, argname, argval);

	res = 1;
	if (!strcmp(inAct, "del"))
	{
		res = DoFilesetDel();
		if (res <= 0)
			strcpy(newfname, "error_deletefs.html");
	}
	else if (!strcmp(inAct, "mod"))
	{
		res = DoFilesetMod();
		if (res <= 0)
			strcpy(newfname, "error_modifyfs.html");
	}
	else if (!strcmp(inAct, "add"))
	{
		res = DoFilesetAdd();
		if (res <= 0)
			strcpy(newfname, "error_addfs.html");
	}
	return res;
}


/***********************************************************
 * routines to get fileset information thru anpssc
 **********************************************************/

static void GetFsPath(uint64 volid, uint32 fsid, uint64 inodenum, char* buf)
{
    int							stat;
	fs_MsgGetFilesetPathReq_t	req;
	fs_MsgGetFilesetPathResp_t	rsp;
	struct timeval			timeout = {60, 0};
	strcpy(buf, "\0");

    /* Fill the request. */
    memset((void *) &req, 0, sizeof(fs_MsgGetFilesetPathReq_t));
	req.msgHdr.requestType = FS_REQ_GET_FILESET_PATH;
	req.volId.whole_thing = volid;
	req.fsetId  = fsid;
	req.inum    = inodenum;
	req.snapnum = 0;

    /* Send message to FS. */
    stat = sendAgileMsg(-1, &req, sizeof(req),
                        "efs", NFX_NFP_FP2, 1, MGMTPLANE, &timeout,
                        &rsp, sizeof(rsp));

    /* Check response from FS. */
	if (stat <= 0)
		return;
    if (rsp.msgHdr.requestType != FS_REQ_GET_FILESET_PATH)
		return;
	if (rsp.msgHdr.fsStatus != FS_ERR_NONE)
		return;

	strcpy(buf, rsp.path);
	return;
}

static int GetFsQuotaStatus(uint64 volid,
				uint64 inodeid, char* fspath, fileset_entry_t* data)
{
    int							stat;
	int							quotaIsOn = 0;
	fs_MsgGetQuotaReq_t			req;
	fs_MsgGetQuotaResp_t		rsp;
	struct timeval				timeout = {60, 0};

	stData.inodeUsed = 0;
	stData.blockUsed = 0;

    /* Fill the request. */
    memset(&req, 0, sizeof(fs_MsgGetQuotaReq_t));
	req.msgHdr.requestType = FS_REQ_GET_QUOTA;
	req.fsid.whole_thing = volid;
	strcpy(req.path, fspath);
	req.numberEntities = 1;
	req.entities[0].id = 0;
	req.entities[0].flag = FS_QUOTA_AGGREGATE;

    /* Send message to FS. */
    stat = sendAgileMsg(-1, &req, sizeof(req),
                        "efs", NFX_NFP_FP2, 1, MGMTPLANE,
						&timeout, &rsp, sizeof(rsp));

    /* Check response from FS. */
	if ((stat <= 0) || (rsp.numberResponse < 1))
		return quotaIsOn;

	stData.inodeUsed = rsp.responses[0].inodesConsumed;
	stData.blockUsed = rsp.responses[0].blocksConsumed;
	if (rsp.responses[0].flag & FS_QUOTA_ON)
		quotaIsOn = 1;

	return quotaIsOn;
}

static int getTableRowCount(void)
{
    int							stat, i, numEntries;
    char						respbuff[10240];
	fs_MsgListFileSetsReq_t		req;
	fs_MsgListFileSetsResp_t	*rsp;
	struct timeval			timeout = {60, 0};

    /* Fill the request. */
    rsp = (fs_MsgListFileSetsResp_t *)respbuff;
    memset((void *) &req, 0, sizeof(fs_MsgListFileSetsReq_t));
	numEntries = -1;

	req.msgHdr.requestType = FS_REQ_LIST_FILESETS;
	req.fsid.whole_thing = inId;
	req.startIndex = 0;
	req.numberEntries = MAX_NUM_OF_FS;

	while (1)
	{
    	/* Send message to FS. */
    	stat = sendAgileMsg(-1, &req, sizeof(req),
                        "efs", NFX_NFP_FP2, 1, MGMTPLANE, &timeout,
                        respbuff, sizeof(respbuff));

	    /* Check response from FS. */
		if (stat < 0) 
			return numEntries;
		else if (!stat)
			return numEntries;


	    if ((rsp->msgHdr.requestType != FS_REQ_LIST_FILESETS) ||
			(rsp->msgHdr.fsStatus != NFX_OK) ||
			(rsp->numberEntries <= 0))
			return numEntries;

		if (fsList)
		{
			fileset_entry_t *p;
			p = realloc(fsList,
				(numEntries + rsp->numberEntries) * sizeof(fileset_entry_t));
			if (p == NULL)
				return numEntries;
			fsList = p;
		}
		else
		{
			fsList = malloc((rsp->numberEntries) * sizeof(fileset_entry_t));
			if (fsList == NULL)
				return numEntries;
			numEntries = 0;
		}

		for (i = 0; i < rsp->numberEntries; i ++)
		{
			int idx = numEntries + i;

			fsList[idx].id = rsp->filesets[i].id;
			fsList[idx].rootInode = rsp->filesets[i].rootInode;
			GetFsPath(inId, fsList[idx].id, fsList[idx].rootInode, fsList[idx].path);
			fsList[idx].quotaOn = GetFsQuotaStatus(inId,
					rsp->filesets[i].rootInode, fsList[idx].path, &fsList[idx]);
		}
		numEntries += rsp->numberEntries;

		if (rsp->numberEntries < MAX_NUM_OF_FS)
			break;

		req.startIndex = fsList[numEntries-1].id + 1;
	}

	return numEntries;
}

static int getTableRowData(int *index, fileset_entry_t *data)
{
	if (*index < 1 || *index > dataCount)
		return -1;
	if (fsList == NULL)
		return -1;

	memcpy(data, &fsList[(*index)-1], sizeof(fileset_entry_t));
	return 0;
}


/***********************************************************
 *  Main routines to process tags
 **********************************************************/

static int getTagValue(char *tag, char *value)
{
	int i, tagType, priv;
	
	tagType = INVALID_TAG;
	for (i = 0; i < NUM_OF_TAGS; i++) {
		if (!strcmp(tag, tagTable[i].tagName)) {
			tagType = tagTable[i].tagType;
			break;
		}
	}
	if (tagType == INVALID_TAG)
		return -1;

	if (tagType == FS_VOLID)
		sprintf(value, "%lld", inId);
	else if (tagType == FS_VOLNAME)
		strcpy(value, inName);
	else {
		if ((dataCurr <= 0) || (dataCurr > dataCount))
			return -1;

		switch (tagType) 
		{
#if 0
		case FS_VOLID:
			sprintf(value, "%lld", inId);
			break;
		case FS_VOLNAME:
			strcpy(value, inName);
			break;
#endif
		case FS_COLSPAN:
			priv = GetUserPrivilege();
			if (priv >= PRIV_FULL)
				sprintf(value, "%d", 2);
			else if (priv >= PRIV_ADD)
				sprintf(value, "%d", 1);
			break;
		case FS_PATH:
			strcpy(value, stData.path);
			break;
		case FS_INODE:
			sprintf(value, "%lld", stData.inodeUsed);
			break;
		case FS_BLOCK:
			sprintf(value, "%lld", stData.blockUsed);
			break;
		case FS_QUOTASTS:
			if (stData.quotaOn)
				strcpy(value, "On");
			else
				strcpy(value, "Off");
			break;
		case FS_QUOTASTS_N:
			if (stData.quotaOn)
				strcpy(value, "Off");
			else
				strcpy(value, "On");
			break;
		case FS_QUOTADETAIL:
			if (stData.quotaOn)
			{
				char tmpbuf[1024];

				sprintf(value, "<a href=\"/agilerequestprefix/FilesetFQuo?id=%lld&vn=%s&path=%s\">FileSet</a>&nbsp;&nbsp;",
						inId, inName, stData.path);
				sprintf(tmpbuf, "<a href=\"/agilerequestprefix/FilesetUGQuo?type=g&id=%lld&vn=%s&path=%s\">Group</a>&nbsp;&nbsp;",
						inId, inName, stData.path);
				strcat(value, tmpbuf);
				sprintf(tmpbuf, "<a href=\"/agilerequestprefix/FilesetUGQuo?type=u&id=%lld&vn=%s&path=%s\">User</a>&nbsp;&nbsp;",
						inId, inName, stData.path);
				strcat(value, tmpbuf);
				sprintf(tmpbuf, "<a href=\"/agilerequestprefix/FilesetDQuo?id=%lld&vn=%s&path=%s\">Default</a>",
						inId, inName, stData.path);
				strcat(value, tmpbuf);
			}
			else
				strcpy(value, "&nbsp");
			break;
		case FS_DELETE:
			if (strcmp(stData.path, "/"))
				sprintf(value, "<a href=\"/agilerequestprefix/FilesetDel?id=%lld&vn=%s&path=%s\">Delete</a>",
					inId, inName, stData.path);
			else
				strcpy(value, "&nbsp");
			break;
		default:
			return -1;
		}
	}

	if (value && !value[0])
		strcpy(value, "&nbsp");
	return 1;
}


/*
 * actRptTagFunc()
 *	tag   - tag name
 *	value - replacement string for tag
 */
static int actRptTagFunc(tagCallbackParams_t *tp)
{
char *tag = tp->tag;
char *value = tp->value;

	if (!strcmp(tag, "TABLE_BEGIN")) {
		/* TABLE_BEGIN tag, need to see if there is any data */
		dataCount = getTableRowCount();
		if (dataCount > 0) {
			/* need to retrieve first table's data */
			dataCurr = 1;
			if (getTableRowData(&dataCurr, &stData) >= 0)
				return 1;
		}

		dataCurr = 0;
		return -1;
	}
	else if (!strcmp(tag, "TABLE_END")) {
		/* TABLE_END tag, check if end of data */
		if ((dataCount <= 0) || (dataCurr <= 0) || (dataCurr >= dataCount))
			return -1;

		dataCurr++;
		if (getTableRowData(&dataCurr, &stData) < 0)
			return -1;
		else {
			return 1;
		}
	}
	else {
		/* other tags */
		return getTagValue(tag, value);
	}

	/* shouldn't reach here... */
	return -1;
}

int actionFsetMain(void *r, int count, char **argname, char **argval)
{
	char newfname[128];

	/* initialize fileset information variables */
	memset(newfname, 0, sizeof(newfname));

	dataCount = 0;
	dataCurr = 0;
	fsList = NULL;

	/* handle Add/Delete/Quota On/Off actions */
	HandleFsAction(newfname, count, argname, argval);

	prepare_agile_template(r, newfname);

	agile_scan_and_send(r, actRptTagFunc);

	if (fsList != NULL)
		free(fsList);
	return 1;
}
