/*
 *  ipifhelp.c - IP interface related helper functions
 *
 */

#include <config.h>
#include <string.h>
#include <sys/types.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>
#include <netinet/in_var.h>
#include <net/if_types.h>
#include <net/if_dl.h>

#include "tools.h"
#include "mibincl.h"
#include "util_funcs.h"

#include "auto_nlist.h"
#include "kernel.h"

#include <ipmd-lib.h>
#include <snmpmsgapi.h>
#include "ipifhelp.h"


/* IP interface table timeout is longer - about 3 minutes */
#define IPIF_TABLE_CACHE_TIMEOUT	(MIB_STATS_CACHE_TIMEOUT*36)
marker_t ipif_table_cache_marker = NULL;



static struct in_ifaddr savein_ifaddr;
static struct ifnet *ifnetaddr, saveifnet, *saveifnetaddr;
static int saveIndex=0;
static char saveName[16];

void
Interface_Scan_Init (void)
{
	auto_nlist(IFNET_SYMBOL, (char *)&ifnetaddr, sizeof(ifnetaddr));
	saveIndex = 0;
}

#define ia_next ia_list.tqe_next
#define if_next if_list.tqe_next

int
Interface_Scan_Next(short *Index,
			char *Name,
			struct ifnet *Retifnet,
			struct in_ifaddr *Retin_ifaddr)
{
	struct ifnet ifnet;
	struct in_ifaddr *ia, in_ifaddr;
	short has_ipaddr=0;

	if (ifnetaddr) {
		/*
		 *		Get the "ifnet" structure and extract the device name
		 */
		klookup((unsigned long)ifnetaddr, (char *)&ifnet, sizeof ifnet);
		strncpy(saveName, ifnet.if_xname, sizeof(saveName));
		saveName[sizeof(saveName)-1] = '\0';

		/*
		 *  Try to find an address for this interface
		 */

		auto_nlist(IFADDR_SYMBOL, (char *)&ia, sizeof(ia));
		while (ia) {
			klookup((unsigned long)ia , (char *)&in_ifaddr, sizeof(in_ifaddr));
			if (in_ifaddr.ia_ifp == ifnetaddr) {
				has_ipaddr = 1; /* this IF has IP-address */
				break;
			}
			ia = in_ifaddr.ia_next;
		}

/*		ifnet.if_addrlist = (struct ifaddr *)&ia->ia_ifa;   */  /* WRONG DATA TYPE; ONLY A FLAG */

		if (Index)
			*Index = ++saveIndex;

		if (Retifnet)
			*Retifnet = ifnet;

		/* assign the in_ifaddr only if the IF has IP-address */
		if (Retin_ifaddr && has_ipaddr)
			*Retin_ifaddr = in_ifaddr;

		if (Name)
			strcpy(Name, saveName);

		saveifnet = ifnet;
		saveifnetaddr = ifnetaddr;
		savein_ifaddr = in_ifaddr;
		ifnetaddr = ifnet.if_next;

		return(1);	/* DONE */
	}
	return(0);		/* EOF */
}


int Interface_Scan_By_Index(int Index,
				   char *Name,
				   struct ifnet *Retifnet,
				   struct in_ifaddr *Retin_ifaddr)
{
	short i;

		Interface_Scan_Init();
		while (Interface_Scan_Next(&i, Name, Retifnet, Retin_ifaddr)) {
		  if (i == Index) break;
		}
		if (i != Index) return(-1);	 /* Error, doesn't exist */
	return(0);	/* DONE */
}

int Interface_Get_Ether_By_Index(int Index,
					u_char *EtherAddr)
{
	short i;
	struct arpcom {
	  char ac_enaddr[6];
	} arpcom;
	struct sockaddr_dl sadl;
	struct ifaddr ifaddr;
	u_long ifaddraddr;

	memset(&arpcom.ac_enaddr, 0, sizeof(arpcom.ac_enaddr));
	memset(EtherAddr, 0, sizeof(arpcom.ac_enaddr));

	if (saveIndex != Index) {	/* Optimization! */

		Interface_Scan_Init();

		while (Interface_Scan_Next((short *)&i, NULL, NULL, NULL) != 0) {
			if (i == Index) break;
		}
		if (i != Index)
			return(-1);	 /* Error, doesn't exist */
	}

	/*
	 *  the arpcom structure is an extended ifnet structure which
	 *  contains the ethernet address.
	 */
#define if_addrlist if_addrlist.tqh_first
#define ifa_next	ifa_list.tqe_next

	ifaddraddr = (unsigned long)saveifnet.if_addrlist;
	while (ifaddraddr) {
		klookup(ifaddraddr, (char *)&ifaddr, sizeof ifaddr);
		klookup((unsigned long)ifaddr.ifa_addr, (char *)&sadl, sizeof sadl);
		if (sadl.sdl_family == AF_LINK &&
				(saveifnet.if_type == IFT_ETHER ||
				saveifnet.if_type == IFT_ISO88025 ||
				saveifnet.if_type == IFT_FDDI)) {
			memcpy(arpcom.ac_enaddr, sadl.sdl_data + sadl.sdl_nlen,
				   sizeof(arpcom.ac_enaddr));
		 	break;
		}
		ifaddraddr = (unsigned long)ifaddr.ifa_next;
	}

	if (strncmp("lo", saveName, 2) == 0) {
		/*
		 *  Loopback doesn't have a HW addr, so return 00:00:00:00:00:00
		 */
		memset(EtherAddr, 0, sizeof(arpcom.ac_enaddr));
	} else {
		memcpy(EtherAddr,(char *) &arpcom.ac_enaddr, sizeof (arpcom.ac_enaddr));
	}
	return(0);	/* DONE */
}


/*
 * IpInterface_Get_IP - get the slot/port info of specified IP addr
 *    ipaddr - IP address in network order
 *    slot   - return slot number on which this IP resides
 *    physical_port - return physical port no. on which this IP resides
 * return - 0 or positive number: fine
 *          -1: failed
 */
int
IpInterface_Get_IP(uint32_t ipaddr, int *slot, int *physical_port)
{
	short i;
	char ifname[16];
	struct in_ifaddr ifaddr;

	memset(&ifaddr, 0, sizeof(ifaddr));

	Interface_Scan_Init();
	while (Interface_Scan_Next(&i, ifname, NULL, &ifaddr)) {
		/* ifname - interface name                 */
		/* retifnet.if_xname - same interface name */
		/* retin_ifaddr                            */
		/*   ((struct sockaddr_in *) &(ifaddr.ia_addr))->sin_addr.s_addr */
		/*                  IP address             */
		/*   .ia_subnetmask - subnetmask           */
		/*   .ia_netbroadcast - broadcast address  */

		if (ipaddr ==
			((struct sockaddr_in *) &(ifaddr.ia_addr))->sin_addr.s_addr) {
			/* found the specified address */
			if (slot)
				*slot = ifname[1] - '0';
			if (physical_port)
				*physical_port = ifname[2] - '0';
			return 0;
		}
		memset(&ifaddr, 0, sizeof(ifaddr));
	}

	/* can't find specified address */
	return -1;
}

int
IpInterface_Get_Next_IP(uint32_t *ipaddr, int *slot, int *physical_port)
{
	short i;
	uint32_t curr_ip, in_ipaddr, lowest = 0;
	char ifname[16];
	struct in_ifaddr ifaddr;

	memset(&ifaddr, 0, sizeof(ifaddr));
	in_ipaddr = ntohl(*ipaddr);

	Interface_Scan_Init();
	while (Interface_Scan_Next(&i, ifname, NULL, &ifaddr)) {
		curr_ip = ntohl(((struct sockaddr_in *)
							&(ifaddr.ia_addr))->sin_addr.s_addr);
		if (curr_ip > in_ipaddr) {
			/* possible next target */
			if ((lowest == 0) || (curr_ip < lowest)) {
				lowest = curr_ip;
				if (slot)
					*slot = ifname[1] - '0';
				if (physical_port)
					*physical_port = ifname[2] - '0';
			}
		}
		memset(&ifaddr, 0, sizeof(ifaddr));
	}

	if (lowest != 0) {
		/* yup, we have a "next" candidate */
		*ipaddr = htonl(lowest);
		return 0;
	}
	else
		return -1;
}

/*
 * slot - a array of integer for this fucntion to return all
 *        the GbE slot number
 * size - the max number of slot number you can put into this
 *        slot array. In return, size indicate how many were
 *        put into the slot array
 */
int
IpInterface_Get_GbE_Slot(int *slot, int *size, int *need)
{
int i, count, ipmd_needed = 0;
ipm_if_t ipm_ifdata[32]; /* 32 ports should be enough for now */
int sindex = 0;
int ret_val;

	/* talk to IPMD to get all the GbE interfaces information */
	count = sizeof(ipm_ifdata) / sizeof(ipm_if_t);
	ret_val = ipmd_GetIfByName(NULL, ipm_ifdata, sizeof(ipm_ifdata),
											&count, &ipmd_needed);
	if ((ret_val == IPM_GET_NO_BUF) || (ret_val == IPM_CMDERR_APPNOBUF)) {
		/* should do something to increase the buffer size */
		return -1;
	}
	if (ret_val == IPM_GET_ERR) {
		return -1;
	}

	if (slot && size) {
		/* find out what slot is GbE NIM located */
		sindex = 0;
		for (i = 0; i < count; i++) {
			int j, newslot, exists;

			newslot = ipm_ifdata[i].if_slot;
			if (newslot > 1) {  /* slot 1 is SSC slot */
				exists = 0;
				for (j = 0; j < sindex; j++) {
					if (slot[j] == newslot) {
						exists = 1;
						break;
					}
				}
				if (exists == 0) {
					slot[sindex] = newslot;
					sindex++;
					if (sindex >= *size)
						break;		/* no more space available */
				}
			}
		}
	}

	if (need) {
		int j, found;

		if ((sindex > 0) && (sindex < *size))
			*need = sindex;
		else {
			*need = 1;
			for (i = 1; i < count; i++) {
				found = 0;
				for (j = 0; j < i; j++) {
					if (ipm_ifdata[i].if_slot == ipm_ifdata[j].if_slot) {
						found = 1;
						break;
					}
				}
				if (found == 0)
					*need += 1;
			}
		}
	}

	*size = sindex;

	return 0;
}


#if 0
int
Build_Internal_IP_Table()
{
	short i;
	char ifname[16];
	struct ifnet ifdata;
	struct in_ifaddr ifaddr;
	int ret_value = 0;

    if (ipif_table_cache_marker &&
		(!atime_ready(ipif_table_cache_marker, IPIF_TABLE_CACHE_TIMEOUT*1000)))
		return 0;

    if (ipif_table_cache_marker )
		atime_setMarker( ipif_table_cache_marker );
    else
		ipif_table_cache_marker = atime_newMarker();

	/* build the internal IP interface table */
	Interface_Scan_Init();
	while (Interface_Scan_Next(&i, ifname, &ifdata, &ifaddr)) {
	}

    if ( ret_value == -1 ) {
		free( ipif_table_cache_marker );
		ipif_table_cache_marker = NULL;
    }
    return ret_value;
}
#endif
