/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2002, Agile Storage Inc.
 * 
 * RCS:  $Id: cifs-rpc-wkssvc.c,v 1.1.1.1 2002/04/23 17:47:36 laurentb Exp $
 *
 * Name			: cifs-rpc-wkssvc.c
 *
 * Description	: Contains CIFS DCE/RPC workstation services api functions. 
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/28/2002
 *
 *-----------------------------------------------------------------
 */

#ifdef FS_STANDALONE
#include "../sm-cifs/cifs-api.h"
#include "../sm-cifs/cifs-unicode.h"
#include "../sm-cifs/cifs-utils.h"
#include "../sm-cifs/cifs-err.h"
#else
#include <exports/cifs-api.h>
#include <exports/cifs-utils.h>
#include <exports/cifs-unicode.h>
#include <exports/cifs-err.h>
#endif

#include "cifs-rpc-utils.h"
#include "cifs-rpc.x"
#include "cifs-rpc-utils.h"
#include "cifs-rpc-wkssvc.h"


static cifsrpc_function wkssvc_functions[] = {
{ WKSSVC_QUERYINFO,			"queryInfo",			wkssvc_queryInfo },
{ 0,						NULL,					NULL }
};


//***************************************************************************
// WKSSVC encode / decode functions
//***************************************************************************

/*-----------------------------------------------------------------
 * Name			: cifsrpcEDwksInfoLv100
 *
 * Description	: WKSSVC infomation (level 100) encode / decode
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/28/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
cifsrpcEDwksInfoLv100(cifsrpc_ed		*edp,
					wkssvc_infoLv100	*info,
					uchar8				lv)
{
	CIFS_PRINT(CIFS_DBGLV_DBG3, ("%swksInfoLv100", cifs_spaces(lv)));
	lv++;
	if (cifsrpcEDalign (edp)										!= NFX_OK ||
		cifsrpcEDuint32(edp, &info->platformId, "platformId",	lv)	!= NFX_OK ||
		cifsrpcEDuint32(edp, &info->serverNamep, "serverNamep",	lv)	!= NFX_OK ||
		cifsrpcEDuint32(edp, &info->domainNamep, "domainNamep",	lv)	!= NFX_OK ||
		cifsrpcEDuint32(edp, &info->majorVersion,"majorVersion",lv) != NFX_OK ||
		cifsrpcEDuint32(edp, &info->minorVersion,"minorVersion",lv) != NFX_OK ||
		cifsrpcEDuniStruc(edp, &info->serverName, "serverName",	lv)	!= NFX_OK ||
		cifsrpcEDuniStruc(edp, &info->domainName, "domainName",	lv)	!= NFX_OK ||
		cifsrpcEDalign	 (edp)										!= NFX_OK)
		return (NFX_ERR);
	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name			: cifsrpcEDwkssvcQueryInfoCall
 *
 * Description	: WKSSVC query info function call encode / decode
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/28/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
cifsrpcEDwkssvcQueryInfoCall(cifsrpc_ed				*edp,
							wkssvc_queryInfoCall	*p,
							uchar8					lv)
{
	CIFS_PRINT(CIFS_DBGLV_DBG3, ("%swkssvcQueryInfoCall", cifs_spaces(lv)));
	lv++;
	if (cifsrpcEDalign	 (edp)										!= NFX_OK ||
		cifsrpcEDuint32	 (edp, &p->serverNamep,	"serverNamep",	lv)	!= NFX_OK ||
		cifsrpcEDuniStruc(edp, &p->serverName,	"serverName ",	lv)	!= NFX_OK ||
		cifsrpcEDalign	 (edp)										!= NFX_OK ||
		cifsrpcEDushort16(edp, &p->level,		"level      ",	lv)	!= NFX_OK ||
		cifsrpcEDalign	 (edp)										!= NFX_OK)
		return(NFX_ERR);

	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name			: cifsrpcEDwkssvcQueryInfoResp
 *
 * Description	: WKSSVC query info function response encode / decode
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/28/2002
 *
 *-----------------------------------------------------------------
 */
static int32	//NFX_OK or NFX_ERR
cifsrpcEDwkssvcQueryInfoResp(cifsrpc_ed				*edp,
							wkssvc_queryInfoResp	*p,
							uchar8					lv)
{
	CIFS_PRINT(CIFS_DBGLV_DBG3, ("%swkssvcQueryInfoResp", cifs_spaces(lv)));
	lv++;

	if (cifsrpcEDalign	 (edp)										!= NFX_OK ||
		cifsrpcEDushort16(edp, &p->level,		"level      ",	lv)	!= NFX_OK ||
		cifsrpcEDalign	 (edp)										!= NFX_OK ||
		cifsrpcEDuint32	 (edp, &p->ptr,			"ptr        ",	lv)	!= NFX_OK ||
		cifsrpcEDwksInfoLv100 (edp, &p->wksInfoLv100, lv)			!= NFX_OK ||
		cifsrpcEDuint32	 (edp, &p->returnValue,	"returnValue ",	lv) != NFX_OK)
	return (NFX_ERR);
	
	return (NFX_OK);
}


//***************************************************************************
// WKSSVC subcommands
//***************************************************************************

/*-----------------------------------------------------------------
 * Name			: wkssvcQueryInfo 
 *
 * Description	: WKSSVC query info on CIFS server
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/28/2002
 *
 *-----------------------------------------------------------------
 */
static uint32	//CIFSTATUS
wkssvcQueryInfo(cifsrpc_pipe			*rpip,
				wkssvc_queryInfoCall	*callp,
				wkssvc_queryInfoResp	*respp)
{
	wkssvc_infoLv100	*info = &respp->wksInfoLv100;
	cifs_server			*svp = rpip->clp->serverp;

	switch(callp->level)
	{
		case 100:
			info->platformId = 0x000001f4;
			info->serverNamep = RPC_DUMMY_PTR;
			info->domainNamep = RPC_DUMMY_PTR;
			info->majorVersion = CIFS_SERVER_VERSION_MAJOR;
			info->minorVersion = CIFS_SERVER_VERSION_MINOR;
			fillUniStrucWithoutAlloc(&info->serverName, svp->cfg.uniName);
			fillUniStrucWithoutAlloc(&info->domainName, svp->cfg.uniDomainName);
			break;

		default:
			return (CIFSTATUS_INVALID_INFO_CLASS);
	}
	return (CIFSTATUS_OK);
}


/*-----------------------------------------------------------------
 * Name			: wkssvc_queryInfo
 *
 * Description	: WKSSVC query information of CIFS server
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/28/2002
 *
 *-----------------------------------------------------------------
 */
int32			//NFX_OK or NFX_ERR
wkssvc_queryInfo(cifsrpc_pipe *rpip)
{
	wkssvc_queryInfoCall	queryInfoCall;
	wkssvc_queryInfoResp	queryInfoResp;

	cifsrpc_ed				*iDecp = &rpip->inPdu.dec;
	cifsrpc_ed				*oEncp = &rpip->outPdu.enc;

	memset(&queryInfoCall, 0, sizeof(wkssvc_queryInfoCall));
	memset(&queryInfoResp, 0, sizeof(wkssvc_queryInfoResp));

	//Decode WKSSVC queryInfo function call data
	if (cifsrpcEDwkssvcQueryInfoCall(iDecp, &queryInfoCall, 1) != NFX_OK)
		return (NFX_ERR);

	//Prepare WKSSVC queryInfo function response data
	queryInfoResp.level = queryInfoCall.level;
	queryInfoResp.ptr = RPC_DUMMY_PTR;
	queryInfoResp.returnValue = wkssvcQueryInfo(rpip,
												&queryInfoCall,
												&queryInfoResp);
	
	//Encode WKSSVC queryInfo function response data
	if (cifsrpcEDwkssvcQueryInfoResp(oEncp, &queryInfoResp, 1) != NFX_OK)
		return (NFX_ERR);

	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name			: cifsrpc_wkssvc
 *
 * Description	: process workstation services pipe requests
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/28/2002
 *
 *-----------------------------------------------------------------
 */
int32	//NFX_OK or NFX_ERR
cifsrpc_wkssvc(cifsrpc_pipe *rpip)
{
	return (cifsrpc_callFunction(rpip, wkssvc_functions));	
}
