/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * RCS:  $Id: cifs-intmsg.c,v 1.1.1.1 2002/04/23 17:41:16 laurentb Exp $
 *
 * Name			: cifs-intmsg.c
 *
 * Description	: Contains functions for internal message processing.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/08/2002
 *
 *-----------------------------------------------------------------
 */
#include "cifs-api.h"
#include "cifs-intmsg.h"


/*-----------------------------------------------------------------
 * Name			: nfs_msgBuildFSHdr 
 *
 * Description	: Build a file system request message header for message 
 * 				  that will be sent to file system.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/11/2002
 *
 *-----------------------------------------------------------------
 */
void
cifs_msgBuildFSHdr(fs_MsgReqHdr		*fsMsgReqHdrp,
					uint32			reqType,	//Request message type
					fs_AttrType		attrType,	//Requested attribute types.
					boolean			caseInsens,	//case insensitive
					uint64			userInfo)	//Thread ID 
{
	memset(fsMsgReqHdrp, 0 , sizeof(fs_MsgReqHdr));
	fsMsgReqHdrp->requestType	= reqType;
	fsMsgReqHdrp->attrType		= attrType;
	fsMsgReqHdrp->proto			= FS_PROTO_CIFS;
	fsMsgReqHdrp->clientPrivate = userInfo;
	fsMsgReqHdrp->flags			= FS_BACKSLASH_SEPARATOR;
	if (caseInsens)
		fsMsgReqHdrp->flags |= FS_CASE_INSENSITIVE;
	return;
}


/*-----------------------------------------------------------------
 * Name			: nfs_msgBuildFssfHdr 
 *
 * Description	: Build a file system stateful request message header 
 *				  for message that will be sent to file system stateful.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/11/2002
 *
 *-----------------------------------------------------------------
 */
void
cifs_msgBuildFssfHdr(fssf_msgHdr_t	*fssfMsgHdrp,
					fssf_msgType_t	reqType,		//Request message type
					cifs_operation	*opp)			//associated operation
{
	fssfMsgHdrp->msgType		= reqType;
	fssfMsgHdrp->clientPrivate	= opp->threadId;
	fssfMsgHdrp->protocol		= FS_PROTO_CIFS;
	if (opp->clp->clientPlatform == CIFS_CLIENT_WIN9x)
		fssfMsgHdrp->cifs = FSSF_CIFS_98;
	else 
		fssfMsgHdrp->cifs = FSSF_CIFS_NT;
}


/*-----------------------------------------------------------------
 * Name			: cifs_msgXmtCmpToFs
 *
 * Description	: Send a transmit complete message to FS.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 04/09/2002
 *
 *-----------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
cifs_msgXmtCmpToFs(cifs_operation	*opp)
{
	int32					rc;
	uchar8					reqBuf[sizeof(fs_MsgTransmitComplete)];
	fs_MsgTransmitComplete	*msgp = (fs_MsgTransmitComplete *)reqBuf;

	cifs_msgBuildFSHdr(&msgp->hdr, FS_REQ_TX_COMPL, 0, FALSE, opp->threadId);
	msgp->fsPrivate = opp->fsPrivateForCod;

	rc = cifs_msgSend(CIFS_INTMSG_TO_FS,
						reqBuf,
						sizeof(fs_MsgTransmitComplete),
						FALSE);
	return (rc);
} /* cifs_xmtCmpIndToFs() */


/*-----------------------------------------------------------------
 * Name			: cifs_msgSend
 * Description	: Send a CIFS control message to modules, such as
 *				  file system, stateful file system, volumn manager,
 *				  and cifsauth, for processing. 
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/11/2002
 *
 *-----------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
cifs_msgSend(uint32		dest,			//Destination module
			 uchar8		*msgp,			//Message buffer
			 uint32		msgLen,			//Message length
			 boolean	freebuf)		//Free message buffer or not
{
	int32	rc;
	uint32	srcPort, dstPort;

#ifndef FS_STANDALONE
	switch(dest)
	{
		case CIFS_INTMSG_TO_FS:
			srcPort = cifs_globalCfgp->ctrlSrcPort;
			dstPort = cifs_globalCfgp->efsDstPort;
			break;

		case CIFS_INTMSG_TO_FSSF:
			srcPort = cifs_globalCfgp->ctrlSrcPort;
			dstPort = cifs_globalCfgp->fssfDstPort;
			break;

		case CIFS_INTMSG_TO_EVM:
			srcPort = cifs_globalCfgp->mgmtSrcPort;
			dstPort = cifs_globalCfgp->evmDstPort;
			break;

		case CIFS_INTMSG_TO_AUTH:
			srcPort = cifs_globalCfgp->mgmtSrcPort;
			dstPort = cifs_globalCfgp->cifsauthDstPort;
			break;

		default:
			break;
	}

	// Send message 
	rc = eee_sendMessage(msgp, msgLen, dstPort, srcPort);
	
	if (rc != NFX_OK)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_msgSend: eee_sendMessage failed!"));
		return (rc);	//We need to know what went wrong
	}

	// Free local buffer
	if (freebuf)
	{
		if (eee_ramDealloc(msgp) != NFX_OK)
		{
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_msgSend: "
						"Failed to deallocate buffer!"));
			return (NFX_ERR);
		}
	}
	
#else //FS_STANDALONE
	switch(dest)
	{
		case CIFS_INTMSG_TO_FS:
			rc = fs_simSendMsg(FS_SIM_MSG_ADDR_CIFS,
								(int32) pthread_getspecific(fs_simKeyCpuId),
								msgp, msgLen);
			break;

		case CIFS_INTMSG_TO_FSSF:
			rc = fs_simSendMsg(FS_SIM_MSG_ADDR_CIFS,
								FS_SIM_MSG_ADDR_FSSF,
								msgp, msgLen);
			break;
		case CIFS_INTMSG_TO_AUTH:	//in FS standalone, we want to time out
			rc = NFX_OK;			//and use anonymous access, so send has to
			break;					//succeed even if not done.
	}
	
	if (rc != NFX_OK)
    {
        CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_msgSend: fs_simSendMsg failed!"));
        return (NFX_ERR);
	}
#endif	//FS_STANDALONE

	return (NFX_OK);
} /* cifs_msgSend() */


#ifndef FS_STANDALONE

/*-----------------------------------------------------------------
 * Name			: cifs_msgRcv
 *
 * Description	: This function is registered with the EEE layer to 
 *				  receive all response messages.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/17/2002
 *
 *-----------------------------------------------------------------
 */
void 
cifs_msgRcv(eee_desc_t *eDescP)
{
	eee_descBuf_t	*bd;
	esm_event_t		evt;
	esm_handle_t	threadId;
	boolean			qevent = TRUE;	//Call thread_event or not

	CIFS_PRINT(CIFS_DBGLV_DBG3, ("> cifs_msgRcv"));

	if (eDescP == NULL)
		return;

	// TODO: Should we consider multiple buffer case?

	bd = (eDescP->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER) ?
										&eDescP->bd[0] : &eDescP->bd[1];

	
	evt.event = *(int32 *)bd->buf;
	
	do {
		if (evt.event >= FS_MSG_BASE && evt.event < FS_MSG_BASE + 1024)
		{	//File system event
			threadId = (esm_handle_t)
						(((fs_MsgRespHdr *)(bd->buf))->clientPrivate);
			break;
		}
		if (evt.event >= FSSF_MSG_BASE && evt.event < FSSF_MSG_BASE + 40)
		{	//Statefull file system event
			threadId = (esm_handle_t)
						(((fssf_msgHdr_t *)(bd->buf))->clientPrivate);
			break;
		}
		if (evt.event == CIFSAUTH_INFO)
		{	// Received message from cifsauth
			threadId = (esm_handle_t)
						(((cifsauthInfoMsg_t *)(bd->buf))->hdr.clientPrivate);
			break;
		}
		if (evt.event == EVM_MSG_STAT)
		{	// Received "get vol id" response from EVM 
			threadId = (esm_handle_t)
						(((evm_MsgStat_t *)(bd->buf))->hdr.clientPrivate);
			break;
		}

		CIFS_PRINT(CIFS_DBGLV_DBG3, ("cifs_msgRcv: Invalid message type"));
		qevent = FALSE;
	}while (0);

	CIFS_PRINT(CIFS_DBGLV_INFO, ("cifs_msgRcv: event = %d threadId = %d",
									evt.event, threadId));

	evt.p1.u.p = eDescP;
	evt.p2.u.p = bd->buf;
	if ((qevent == FALSE) || (thread_event(threadId, &evt) != NFX_OK))
		eee_freePacket(eDescP);
	
	CIFS_PRINT(CIFS_DBGLV_DBG3, ("< cifs_msgRcv"));
} /* cifs_msgRcv() */


#else	//FS_STANDALONE

#include "testcifs.h"

void
cifs_fsStandAloneMsgRcv(char*	msgBuf,		//pointer to the message
						uint32	srcPort)	//sender
{
	esm_event_t		evt;
	esm_handle_t	threadId;
	boolean			qevent = TRUE;	//Call thread_event or not
	
	evt.event = *(int32 *)msgBuf;
	
	do {
		if (evt.event >= FS_MSG_BASE && evt.event < FS_MSG_BASE + 1024)
		{	//File system event
			threadId = (esm_handle_t)
						(((fs_MsgRespHdr *)(msgBuf))->clientPrivate);
			break;
		}
		if (evt.event >= FSSF_MSG_BASE && evt.event < FSSF_MSG_BASE + 40)
		{	//Statefull file system event
			threadId = (esm_handle_t)
						(((fssf_msgHdr_t *)(msgBuf))->clientPrivate);
			break;
		}
		if (evt.event == CIFSAUTH_INFO)
		{	// Received message from cifsauth
			threadId = (esm_handle_t)
						(((cifsauthInfoMsg_t *)(msgBuf))->hdr.clientPrivate);
			break;
		}
/*		
		if (evt.event == CIFS_MSG_BASE)
		{	//Debug command
			static uint32 opCnt = 0;
			
			CIFS_PRINT(CIFS_DBGLV_DBG3, ("cifs_processMsg: NULL cifs op"));
			switch (opCnt)
			{
				case 0: cifstest_init(); break;
//				case 1: cifstest_createSession(); break;
//				case 2: cifstest_treeConnectNX1(); break;
			}
			opCnt++;
			qevent = FALSE;
			break;
		}
*/
		CIFS_PRINT(CIFS_DBGLV_DBG3, ("cifs_processMsg: Invalid message type"));
		qevent = FALSE;
	}while (0);

	CIFS_PRINT(CIFS_DBGLV_INFO, ("cifs_processMsg: event = %d threadId = %d",
									evt.event, threadId));

	evt.p1.u.p = NULL;
	evt.p2.u.p = msgBuf;
	if (qevent && thread_event(threadId, &evt) != NFX_OK)
		CIFS_PRINT(CIFS_DBGLV_ERR, 
					("cifs_processMsg: failed to send event to thread"));
}
#endif	//FS_STANDALONE
