/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 * 
 * RCS:  $Id: cifs-ipc.c,v 1.2 2002/12/10 18:13:18 laurentb Exp $
 * 
 * Name			: cifs-ipc.c
 *
 * Description	: Contains CIFS ipc related functions. 
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/29/2001
 *
 *-----------------------------------------------------------------
 */

#include "cifs-api.h"
#include "cifs-utils.h"
#include "cifs-unicode.h"
#include "cifs-string.h"
#include "cifs-transport.h"
#include "cifs-encdec.h"
#include "cifs-err.h"
#include "cifs-msg.h"
#include "cifs-ipc.h"
#include "cifs-rap.h"

#define IFDBG(lvl)  if (lvl <= cifs_debugLevel)


#define cifs_setBufOverflowWarning(opp, osmbp)								\
	{																		\
		if (opp->clp->capabilities & CAP_STATUS32)							\
		{																	\
			PUTLE4(osmbp, SMBO_ERCL, CIFSTATUS_BUFFER_OVERFLOW);			\
			PUTLE4(osmbp, SMBO_FLG2, 										\
					(GETLE2(osmbp, SMBO_FLG2) | SMB_FLG2_ERR_STATUS));		\
		} else {															\
			PCAR(osmbp, SMBO_ERCL, CIFS_ERRDOS);							\
			PUTLE2(osmbp, SMBO_ERR, CIFSERR_moredata);						\
		}																	\
	}																		


//Named pipe write mode (writeX)
#define CIFS_PIPE_RAW_MODE		0x4
#define CIFS_PIPE_START_MESSAGE	0x8
#define CIFS_PIPE_STARTMSG_RAW	(CIFS_PIPE_RAW_MODE|CIFS_PIPE_START_MESSAGE)

//Browse announcement 
#define CIFS_BROWSECMD_HOSTANNOUNCEMENT	1

//IPC
#define  CIFS_TRANSACERR_unsupported	50

//Unicode strings
uni16 unistr_lanman[]   = { 'L','A','N','M','A','N', 0 };
extern uni16 unistr_backpipe[6];

/*-----------------------------------------------------------------
 * Name			: cifs_createPipe
 *
 * Description	: create a pipe on an IPC share
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/30/2001
 *
 *-----------------------------------------------------------------
 */
static cifs_pipe *
cifs_createPipe(uni16 *name,
				cifs_client *clp, 
				cifs_session *ssp,
				cifs_shareInUse *siup,
				cifs_operation *opp)
{
	cifs_pipe *pip;
	int	index;
	
    IFDBG(CIFS_DBGLV_DBG2) 
	{
		uniprintf("cifs_createPipe: opening pipe ", name);
		printf("\n");
	}
	
	//Check if pipe is supported
	index = cifsrpc_pipeIsSupported(name);
	if (index == -1)
	{	//Pipe not supported
        cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_access, CIFSTATUS_ACCESS_DENIED);
		return (NULL);
	}

	//Create the pipe
	pip = cifs_addPipe(siup);
	if (pip == NULL) {
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_createPipe: no avail pipe obj"));
		cifs_errResp(opp, CIFS_ERRDOS, CIFSERR_nofids,
						CIFSTATUS_TOO_MANY_OPENED_FILES);
		return (NULL);
	}

	//Initialize pipe
	if (cifsrpc_initPipe(&pip->u.rpc, index, clp, ssp) != NFX_OK)
	{
		cifs_closeShareInUsePipe(siup, pip);
		cifs_errResp(opp, CIFS_ERRDOS, CIFSERR_nomem, 
						CIFSTATUS_INSUFF_SERVER_RESOURCES);
		return (NULL);	
	}
		
	return (pip);
}


/*-----------------------------------------------------------------
 * Name			: cifs_closePipe
 *
 * Description	: close a pipe
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/29/2002
 *
 *-----------------------------------------------------------------
 */
static void cifs_closePipe(cifs_shareInUse *siup, cifs_pipe *pip)
{

    CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_closePipe: fid = 0x%x", pip->fid));
	
	//Close the pipe on RPC side
	cifsrpc_closePipe(&pip->u.rpc);

	//remove the pipe from the open pipes list and free the pipe object
	cifs_closeShareInUsePipe(siup, pip);	
}


/*-----------------------------------------------------------------
 * Name			: cifs_writeOnPipe
 *
 * Description	: write on a named pipe
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 12/05/2001
 *
 *-----------------------------------------------------------------
 */
static int32
cifs_writeOnPipe(	cifs_pipe	*pip,
					uchar8 		*datp,
					uint32		datCnt)
{
	cifs_cmdDesc *reqMsgp = &pip->opp->reqMsg;
	int32	bytesRemaining, bytesWritten;

	
	bytesRemaining = datCnt;
	
	if (cifs_msgIsFragmented(reqMsgp) == FALSE)
	{			//Do it the simple way

		CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_writeOnPipe: %d bytes", datCnt));
		IFDBG(CIFS_DBGLV_DBG3)
		{
			memDump(datp, datCnt);
		    printf("\n");
		}

		while (bytesRemaining > 0)
		{
			bytesWritten = cifsrpc_writeData(&pip->u.rpc, datp, bytesRemaining);

		    CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_writeOnPipe: "
						"%d bytes written", bytesWritten));
	
			if (bytesWritten <= 0)
				return (NFX_ERR);
			bytesRemaining -= bytesWritten;
			datp 		   += bytesWritten;
		}
		
	} else {	//Message is fragmented
		uint32 datOff, fragOff, fragSize;	
		int32 bytesToWrite;
		pkt_desc_t *fragp = reqMsgp->tplPktDescp;

		//Calculate back the data offset
		datOff = (address_t)(datp - reqMsgp->smbp);	
		//Set the current fragment offset to 0
		fragOff = 0;
		//Calculate first fragment size
		fragSize = cifs_msgFragSize(reqMsgp);

		//Data should be in first fragment, but...
		if (datOff > fragSize)
		{
			//Look for the fragment where data start
			while (datOff > fragOff + fragSize)
			{
				fragOff += fragSize;
				fragp = fragp->pNext;
				if (fragp == NULL)
					return (NFX_ERR);
				fragSize = fragp->hdrSize;
			}
			datp = fragp->pHdr + (datOff - fragOff);
		}
		
		//Write data
		bytesToWrite = MIN(bytesRemaining, fragSize - (datOff - fragOff));
		
		for (;;)
		{
			CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_writeOnPipe: "
						"%d bytes (Remain=%u hSize=%u)", 
						bytesToWrite, bytesRemaining, fragp->hdrSize));
			IFDBG(CIFS_DBGLV_DBG3)
			{
				memDump(datp, bytesToWrite);
			    printf("\n");
			}
			while (bytesToWrite > 0)
			{
				
				bytesWritten = cifsrpc_writeData(&pip->u.rpc,
												datp, 
												bytesToWrite);
			    CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_writeOnPipe: "
							"%d bytes written",bytesWritten));
				if (bytesWritten <= 0)
					return (NFX_ERR);
				bytesToWrite	-= bytesWritten;
				bytesRemaining	-= bytesWritten;
				datp			+= bytesWritten;
			}
			if (bytesRemaining <= 0)
				break;
			fragp = fragp->pNext;
			if (fragp == NULL)
				return (NFX_ERR);
			bytesToWrite = MIN(bytesRemaining, fragp->hdrSize);
			datp = fragp->pHdr;
		}
	}
	
	return (datCnt);
}



/*-----------------------------------------------------------------
 * Name			: cifsm_ipcNtCreateNX 
 *
 * Description	: NT Create and X on the IPC share
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/29/2001
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_ipcNtCreateNX(	cifs_client *clp, 
							cifs_session *ssp,
							cifs_shareInUse *siup,
							cifs_operation *opp,
							uchar8 *icmdp,
							uchar8 *ocmdp)
{
	cifs_pipe *pip;
	uchar8 *bp;
	ushort16 *wp;

	bp = SMBCP_BPAR(icmdp);
	//Assume that string is in unicode without checking SMB_FLG2_UNICODE flag
	//to protect from W2K bug
	//TODO Check if this can apply to all clients or only to W2K
	bp += ((address_t)bp) % 2;	//Unicode string are aligned on word
	wp = (ushort16*)bp;

	//Skip backslash
	wp++;
	
	//Create the pipe
	pip = cifs_createPipe(wp, clp, ssp, siup, opp);
	if (pip == NULL)
		return (CIFS_SENDRESPNOW);

	//Build response
	cifs_setCmdWordByteCnt(&opp->respMsg, 34, 0, TRUE);
	cifs_initNXCmdHeader(ocmdp);
	
	PUTLE2(ocmdp, SMBCO_WP2+1, pip->fid);						//file ID
	PUTLE4(ocmdp, SMBCO_WP3+1, SEC_FILE_WAS_OPENED);			//create action
	PUTLE4(ocmdp, SMBCO_WP21+1,SEC_FILE_ATTRIBUTE_NORMAL);		//NT file attr.
	PUTLE2(ocmdp, SMBCO_WP31+1,SEC_FILE_TYPE_MSG_MODE_PIPE);	//file type
	PUTLE2(ocmdp, SMBCO_WP32+1,0x5ff);							//device state
	
	if (SMB_BATCHEDCMDFOLLOWS(icmdp))
		return cifs_processNextCmd(opp, ocmdp);
	else	
		return (CIFS_SENDRESPNOW);
}


/*-----------------------------------------------------------------
 * Name			: cifsm_ipcOpenNX 
 *
 * Description	: Open and X on the IPC share
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/12/2002
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_ipcOpenNX(	cifs_client *clp, 
						cifs_session *ssp,
						cifs_shareInUse *siup,
						cifs_operation *opp,
						uchar8 *icmdp,
						uchar8 *ocmdp)
{
	cifs_pipe *pip;
	uchar8 *bp;
	ushort16 *wp;
	uni16 pipeName[CIFS_S_PIPENAME+6+1];	//+6 in case "\PIPE\" is present

	//Get pipe name
	bp = SMBCP_BPAR(icmdp);
	cifs_getSmbUniStr(pipeName, bp, unisizeof(pipeName), opp->reqMsg.smbp);
	
	wp = pipeName;
	
	//Skip PIPE string if present
	if (unistrncasecmp(wp, unistr_backpipe, 5) == 0)
		wp += sizeof(unistr_backpipe)/sizeof(uni16) - 1; //ending 0
				
	//Skip Backslash character if present
	if (*wp == UNICHR_BACKSLASH)
		wp++;
	
	//Create the pipe
	pip = cifs_createPipe(wp, clp, ssp, siup, opp);
	if (pip == NULL)
		return (CIFS_SENDRESPNOW);

	//Build response
	cifs_setCmdWordByteCnt(&opp->respMsg, 15, 0, TRUE);
	cifs_initNXCmdHeader(ocmdp);
	
	PUTLE2(ocmdp, SMBCO_WP2, pip->fid);							//file ID
	PUTLE2(ocmdp, SMBCO_WP9, SEC_FILE_TYPE_MSG_MODE_PIPE);		//file type
	PUTLE2(ocmdp, SMBCO_WP11, SEC_FILE_WAS_OPENED);			//action taken
	
	if (SMB_BATCHEDCMDFOLLOWS(icmdp))
		return cifs_processNextCmd(opp, ocmdp);
	else	
		return (CIFS_SENDRESPNOW);
}


/*-----------------------------------------------------------------
 * Name			: cifsm_ipcReadNX
 *
 * Description	: read data from a pipe
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/13/2002
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_ipcReadNX(	cifs_shareInUse *siup,
						cifs_operation *opp,
						uchar8 *icmdp,
						uchar8 *ocmdp,
						ushort16 fid)
{
	uchar8 *osmbp = opp->respMsg.smbp;
	cifs_pipe *pip;
	ushort16 maxCnt, dataOff;
	int32 bytesRead;
	boolean moreData;
	uchar8 *bp, *bpar, *datp;

	pip = cifs_findShareInUsePipe(siup, fid);
	if (pip == NULL)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_ipcReadNX: unknown fid 0x%x", fid));
		cifs_errResp(opp, CIFS_ERRDOS, CIFSERR_badfid, CIFSTATUS_FILE_CLOSED);
		return (CIFS_SENDRESPNOW);
	}

	maxCnt = GETLE2(icmdp, SMBCO_READNX_MAX_CNT_LOW); 

	bytesRead = cifsrpc_readData(&pip->u.rpc, &datp, maxCnt, &moreData);
	if (bytesRead < 0)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, 
					("cifsm_ipcReadNX: failed to read (fid 0x%x)", fid));
		cifs_errResp(opp, 0, 0, CIFSTATUS_ACCESS_DENIED);
		return (CIFS_SENDRESPNOW);
	}
    CIFS_PRINT(CIFS_DBGLV_DBG2, 
				("cifs_ipcReadNX: %d bytes from %d requested (fid 0x%x)", 
				 bytesRead, maxCnt, pip->fid));
	IFDBG(CIFS_DBGLV_DBG3)
	{
		memDump(datp, bytesRead);
	    printf("\n");
	}

	//Build response
	
	if (moreData)
		cifs_setBufOverflowWarning(opp, osmbp);

	//Align data on a 32 bits boundary (optimize memcpy)
	//TODO: see if we want to go to 64 bits alignment on Sibyte CPU
	bp = bpar = SMBCP_BPAR(ocmdp);
	bp += CIFS_ALIGN_32 - (((address_t)bp) % CIFS_ALIGN_32);
	dataOff = SMB_DIFF(bp, osmbp);

	cifs_setCmdWordCnt(&opp->respMsg, 12, TRUE);//Only zero the word section
	cifs_initNXCmdHeader(ocmdp);
	PUTLE2(ocmdp, SMBCO_WP5, bytesRead);
	PUTLE2(ocmdp, SMBCO_WP6, dataOff);
	cifs_setCmdByteCnt(&opp->respMsg, ((address_t)(bp-bpar))+bytesRead, FALSE);
	memcpy(bp, datp, bytesRead);
					
	if (SMB_BATCHEDCMDFOLLOWS(icmdp))
		return cifs_processNextCmd(opp, ocmdp);
	else	
		return (CIFS_SENDRESPNOW);
}


/*-----------------------------------------------------------------
 * Name			: cifsm_ipcWriteNX
 *
 * Description	: write data to a pipe
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/13/2002
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_ipcWriteNX(	cifs_shareInUse *siup,
						cifs_operation *opp,
						uchar8 *icmdp,
						uchar8 *ocmdp,
						ushort16 fid)
{
	uchar8 *osmbp = opp->respMsg.smbp;
	cifs_pipe *pip;
	boolean startMsgWriteRawNamedPipe;
	ushort16 writeMode, bytesToWrite;
	int32 bytesWritten = 0;
	uchar8 *datp;

	pip = cifs_findShareInUsePipe(siup, fid);
	if (pip == NULL)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_ipcWriteNX: unknown fid 0x%x", fid));
		cifs_errResp(opp, CIFS_ERRDOS, CIFSERR_badfid, CIFSTATUS_FILE_CLOSED);
		return (CIFS_SENDRESPNOW);
	}

	writeMode	 = GETLE2(icmdp, SMBCO_WP7);
	startMsgWriteRawNamedPipe = ((writeMode & CIFS_PIPE_STARTMSG_RAW) 
									== CIFS_PIPE_STARTMSG_RAW);
	bytesToWrite = GETLE2(icmdp, SMBCO_WP10); 
	datp		 = opp->reqMsg.smbp + GETLE2(icmdp, SMBCO_WP11);

	if (bytesToWrite != 0)
	{
		if (startMsgWriteRawNamedPipe)
		{
			if (bytesToWrite < 2)
			{
				CIFS_PRINT(CIFS_DBGLV_ERR,
							("cifsm_ipcWriteNX: start of msg with bad size"));
				cifs_errResp(opp, 0, 0, CIFSTATUS_ACCESS_DENIED);
				return (CIFS_SENDRESPNOW);
			}
			datp		 += 2;
			bytesToWrite -= 2;
		}
		
		bytesWritten = cifs_writeOnPipe(pip, datp, bytesToWrite);
		if ( bytesWritten < 0 || (bytesWritten == 0 && bytesToWrite != 0))
		{
			CIFS_PRINT(CIFS_DBGLV_ERR, 
						("cifsm_ipcWriteNX: failed to write (fid 0x%x)", fid));
			cifs_errResp(opp, 0, 0, CIFSTATUS_ACCESS_DENIED);
			return (CIFS_SENDRESPNOW);
		}
	    CIFS_PRINT(CIFS_DBGLV_DBG2, 
					("cifs_ipcWriteNX: %d bytes from %d requested (fid 0x%x)", 
					 bytesWritten, bytesToWrite, pip->fid));
		IFDBG(CIFS_DBGLV_DBG3)
		{
			memDump(datp, bytesWritten);
			    printf("\n");
		}
	}

	//Build response
	
	cifs_setCmdWordByteCnt(&opp->respMsg, 6, 0, TRUE);
	cifs_initNXCmdHeader(ocmdp);
	
	if (startMsgWriteRawNamedPipe)
		bytesWritten += 2;
	PUTLE2(ocmdp, SMBCO_WP2, bytesWritten);
					
	if (SMB_BATCHEDCMDFOLLOWS(icmdp))
		return cifs_processNextCmd(opp, ocmdp);
	else	
		return (CIFS_SENDRESPNOW);
}


/*-----------------------------------------------------------------
 * Name			: cifsm_ipcClose 
 *
 * Description	: close a pipe open on the IPC share
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/29/2002
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_ipcClose(cifs_shareInUse *siup,	cifs_operation *opp, ushort16 fid)
{
	cifs_pipe *pip;

	pip = cifs_findShareInUsePipe(siup, fid);
	if (pip == NULL)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_ipcClose: "
					"unknown pipe fid 0x%x", fid));
		cifs_errResp(opp, CIFS_ERRDOS, CIFSERR_badfid, CIFSTATUS_FILE_CLOSED);
		return (CIFS_SENDRESPNOW);
	}

	//Close the pipe
	cifs_closePipe(siup, pip);
	
	//Build the response
	
	// Set word count and byte count to 0
	cifs_setCmdWordByteCnt(&opp->respMsg, 0, 0, TRUE);
	
	return (CIFS_SENDRESPNOW);
}


/*-----------------------------------------------------------------
 * Name			: cifs_replyTransac
 *
 * Description	: answer to a transaction
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 12/05/2001
 *
 *-----------------------------------------------------------------
 */
void cifs_replyTransac(	cifs_operation	*opp,
						uchar8			*parp,
						ushort16		paraToSendCnt, 
						uchar8			*datp,
						ushort16		dataToSendCnt,
						boolean 		moreData)
{
	uchar8 *ocmdp = opp->respMsg.cmdp;
	uchar8 *osmbp = opp->respMsg.smbp;
	uchar8 *bp, *bparp, *bdatp;
	int32 replyMaxSize, paraCnt, dataCnt, paraSentCnt, dataSentCnt;
	int32 paraOff, dataOff;

	//Data in each reponse SMB is expected to be truncated to the maximum 
	//number of 512 byte blocks which will fit (starting at a 32 bit boundary)
	//in the negociated buffer size with the odd bytes remaining (if any) in 
	//the final buffer (CIFS 0.9, paragraph 3.13)
	replyMaxSize = opp->clp->maxBufSize;
	replyMaxSize -= replyMaxSize % 512;
			
	paraCnt = MIN(paraToSendCnt, replyMaxSize);
	dataCnt = MIN(dataToSendCnt, replyMaxSize - paraCnt);

    CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_replyTransac: "
				"pSend=%d, dSend=%d, moreData=%d, "
				"replyMax=%d, pCnt=%d, dCnt=%d", 
				paraToSendCnt, dataToSendCnt, moreData, 
				replyMaxSize, paraCnt, dataCnt));

	//We initialize only the word part of the message, since the byte part
	//is memcopied and is big (so not initializing it is more efficient)
	cifs_setCmdWordCnt(&opp->respMsg, 10, TRUE);

	if (moreData)
		cifs_setBufOverflowWarning(opp, osmbp);

	//Align transaction param and data on a 32 bits boundary (optimize memcpy)
	//TODO: see if we want to go to 64 bits alignment on Sibyte CPU
	bp = bparp = SMBCP_BPAR(ocmdp);
	bparp += CIFS_ALIGN_32 - (((address_t)bparp) % CIFS_ALIGN_32);
	bdatp = bparp + paraCnt;
	bdatp += CIFS_ALIGN_32 - (((address_t)bdatp) % CIFS_ALIGN_32);
	paraOff = SMB_DIFF(bparp, osmbp);
	dataOff = SMB_DIFF(bdatp, osmbp);

	//Now we know the size of the byte section
	cifs_setCmdByteCnt(&opp->respMsg, ((address_t)(bdatp-bp))+dataCnt, FALSE);
	
	PUTLE2(ocmdp, SMBCO_WP0, paraToSendCnt);
	PUTLE2(ocmdp, SMBCO_WP1, dataToSendCnt);
	PUTLE2(ocmdp, SMBCO_WP3, paraCnt);
	PUTLE2(ocmdp, SMBCO_WP4, paraOff);
	PUTLE2(ocmdp, SMBCO_WP6, dataCnt);
	PUTLE2(ocmdp, SMBCO_WP7, dataOff);
	
	if (paraCnt)
		memcpy(bparp, parp, paraCnt);
	if (dataCnt)
		memcpy(bdatp, datp, dataCnt);

	paraSentCnt = paraCnt;
	dataSentCnt = dataCnt;
	
	if (paraSentCnt < paraToSendCnt || dataSentCnt < dataToSendCnt)
	{
		pkt_desc_t *tplPktDescp;
			
		NFX_ASSERT(FALSE);
		//TODO see how to handle multiple outgoing SMBs linked with 1 operation
		
		/*
		//Get another buffer to send
		tplPktDescp = cifs_getClientBuf(opp->clp, &bp);
		
		//Copy smb header
		memcpy(bp, osmbp, SMBO_WCNT);
		
		cifs_sendClientMsg(opp->clp, opp);
	
		while (paraSentCnt < paraToSendCnt || dataSentCnt < dataToSendCnt)
		{
		}
		*/
	}
}
	
	
/*-----------------------------------------------------------------
 * Name			: cifs_unsupportedTransac
 *
 * Description	: answer to an unsupported transaction
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 12/05/2001
 *
 *-----------------------------------------------------------------
 */
int32 cifs_unsupportedTransac(cifs_operation *opp)
{
	uchar8 paramb[4];

	PUTLE2(paramb, 0, CIFS_TRANSACERR_unsupported);
	PUTLE2(paramb, 2, 0);

	cifs_replyTransac(opp, paramb, sizeof(paramb), NULL, 0, FALSE);
	
	return (CIFS_SENDRESPNOW);
}

	
/*-----------------------------------------------------------------
 * Name			: cifs_readTransacReply
 *
 * Description	: read from a named pipe
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 12/05/2001
 *
 *-----------------------------------------------------------------
 */
int32 cifs_readTransacReply(cifs_pipe *pip)
{
	uchar8 *datp;
	int32 bytesRead;
	boolean moreData;

	bytesRead = cifsrpc_readData(	&pip->u.rpc,
									&datp,
									pip->maxDataBytes,
									&moreData);
    CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_readTransacReply: "
				"%d bytes (fid 0x%x)", bytesRead, pip->fid));
	IFDBG(CIFS_DBGLV_DBG3)
	{
		memDump(datp, bytesRead);
	    printf("\n");
	}
	if (bytesRead < 0)
		return (bytesRead);
	
	cifs_replyTransac(pip->opp, NULL, 0, datp, bytesRead, moreData);
	
	return (bytesRead);
}


/*-----------------------------------------------------------------
 * Name			: cifs_processTransac
 *
 * Description	: process a transaction once extracted from the cifs
 * 					message
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 12/04/2001
 *
 *-----------------------------------------------------------------
 */
static int32 
cifs_processTransac(cifs_operation *opp,
					uni16	 *pipeName,
					ushort16 *setupp,
					uchar8	 *parp,
					uchar8	 *datp,
					ushort16 swsCnt,
					ushort16 pbsCnt, 
					ushort16 dbsCnt,
					ushort16 mswrCnt,
					ushort16 mpbrCnt,
					ushort16 mdbrCnt)
{
	cifs_pipe *pip;
	ushort16 pipeFid, subCmd;
//	int32 result;
	
    IFDBG(CIFS_DBGLV_DBG2) 
	{
		uniprintf("cifs_processTransac: cmd on pipe ", pipeName);
		printf("\n");
	}
	
	if (unistrlen(pipeName) == 0)
	{	//pipe should be already open
		if (swsCnt != 2)
		{
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_processTransac: "
						"unsupported setup count"));
			cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_error,
							CIFSTATUS_UNSUCCESSFUL);
			return (CIFS_SENDRESPNOW);
		}
		subCmd	= GETLE2(setupp, 0);
		pipeFid	= GETLE2(setupp, 1);
		pip = cifs_findShareInUsePipe(opp->siup, pipeFid);
		if (pip == NULL)
		{
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifs_processTransac: "
						"unknown pipe fid 0x%x", pipeFid));
			return cifs_unsupportedTransac(opp);
		}
		pip->opp = opp;
		pip->maxDataBytes = mdbrCnt;		
		
		switch(subCmd)
		{
			case CIFS_NAMEDPIPESUBCMD_TRANSAC:
				if ((cifs_writeOnPipe(pip, datp, dbsCnt) >= 0)
				 && (cifs_readTransacReply(pip) >= 0))
					return (CIFS_SENDRESPNOW);
				break;
				
			case CIFS_NAMEDPIPESUBCMD_SETHANDSTATE:
				cifs_replyTransac(opp, NULL, 0, NULL, 0, FALSE);
				return (CIFS_SENDRESPNOW);
				
			case CIFS_NAMEDPIPESUBCMD_WAITHANDSTATE:
				cifs_replyTransac(opp, NULL, 0, NULL, 0, FALSE);
				return (CIFS_SENDRESPNOW);
		}

	}
	else
	{	//pipe is not open yet (Win9x family)
		if (unistrcasecmp(pipeName, unistr_lanman) == 0)
		{	//lanman pipe - RAP operations

			//For now we defragment any fragmented RAP request
			//Should not happen, but...
			if (cifs_msgIsFragmented(&opp->reqMsg))
				cifs_defragMsg( opp->clp,
								opp->reqMsg.tplPktDescp,
								opp->reqMsg.smbp,
								opp->reqMsg.smbLen);
								
			return cifs_msrap(opp, parp, datp, pbsCnt, dbsCnt);
		}
	}
	
	return cifs_unsupportedTransac(opp);
}


/*-----------------------------------------------------------------
 * Name			: cifsm_transac 
 *
 * Description	: transaction (0x25)
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 12/03/2001
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_transac(cifs_client *clp, 
					cifs_session *ssp,
					cifs_shareInUse *siup,
					cifs_operation *opp,
					uchar8 *icmdp,
					uchar8 *ocmdp)
{
	uchar8 *bp, *parp, *datp;
	ushort16 *wp, *setupp;
	ushort16 tpbsCnt, tdbsCnt, mpbrCnt, mdbrCnt, mswrCnt, pbsCnt, pbsOff;
	ushort16 dbsCnt, dbsOff, swsCnt, i;
	boolean discTid, noAnswerReq;
	uni16 pipeName[CIFS_S_PIPENAME+6+1];	//+6 in case "\PIPE\" is present
	int32 result;

	//Decode word fields
	tpbsCnt		= GETLE2(icmdp, SMBCO_WP0);
	tdbsCnt 	= GETLE2(icmdp, SMBCO_WP1);
	mpbrCnt		= GETLE2(icmdp, SMBCO_WP2);
	mdbrCnt		= GETLE2(icmdp, SMBCO_WP3);
	mswrCnt		= GETLE2(icmdp, SMBCO_WP4);
	discTid		= SMB_TESTBIT2(icmdp, SMBCO_WP5, 0);
	noAnswerReq	= SMB_TESTBIT2(icmdp, SMBCO_WP5, 1);
	pbsCnt		= GETLE2(icmdp, SMBCO_WP9);
	pbsOff		= GETLE2(icmdp, SMBCO_WP10);
	dbsCnt		= GETLE2(icmdp, SMBCO_WP11);
	dbsOff		= GETLE2(icmdp, SMBCO_WP12);
	swsCnt		= GETLE2(icmdp, SMBCO_WP13);
		
	if (pbsCnt > tpbsCnt || dbsCnt > tdbsCnt || swsCnt > 2)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_transac: "
					"transaction params error"));
		cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_error, CIFSTATUS_UNSUCCESSFUL);
		return (CIFS_SENDRESPNOW);
	}
	
	//Alloc all the necessary memory
	setupp = NULL;
	parp = datp = NULL;

	//Here we assume all setup words are in the first fragment of the message
	if (swsCnt != 0)
		setupp = (ushort16*)(icmdp + SMBCO_WP14); 

	if (tpbsCnt != 0)
	{				
		parp = opp->reqMsg.smbp + pbsOff;
		/* For now we assume that all commands with params will be defragmented
		 * before being processed. So we set parp to the right offset, assuming
		 * that parameters will there when processed 
		if (cifs_msgOffsetIsInFirstFragment(&opp->reqMsg, pbsOff + pbsCnt))
		{			//param bytes are in this buffer
			parp = opp->reqMsg.smbp + pbsOff;
		} else {	//param bytes are not all in this buffer
			CIFS_RAM_ALLOC(parp, uchar8, tpbsCnt);
			if (parp == NULL)
			{
				CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_transac: "
							"no memory for %d params bytes", tpbsCnt));
				cifs_errResp(opp, CIFS_ERRDOS, CIFSERR_nomem, 
								CIFSTATUS_INSUFF_SERVER_RESOURCES);
				return (CIFS_SENDRESPNOW);
			}
			if (cifs_msgDataCpy(parp, &opp->reqMsg, pbsOff, pbsCnt) == NULL)
			{
				CIFS_RAM_DEALLOC(parp);
				CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_transac: "
							"failed to get param from msg"));
				cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_error, 
							CIFSTATUS_UNSUCCESSFUL);			
				return (CIFS_SENDRESPNOW);
			}
		}
		*/
	}
	
	if (tdbsCnt != 0)
	{
		datp = opp->reqMsg.smbp + dbsOff;			
		/* For now, we assume that commands not supporting fragmented data
		 * will be defragmented before being processed. So we set datp to the
		 * right offset assuming that data will be there when processed.
		 * Commands supporting fragmented data will use datp to know where
		 * to look for the data (use it to recreate dbsOff)*/
	}
	
	//Get pipe name
	bp = SMBCP_BPAR(icmdp);
	cifs_getSmbUniStr(pipeName, bp, unisizeof(pipeName), opp->reqMsg.smbp);
	
	wp = pipeName;
	
	//Skip PIPE string if present
	if (unistrncasecmp(wp, unistr_backpipe, 5) == 0)
		wp += sizeof(unistr_backpipe)/sizeof(uni16) - 1; //ending 0
				
	//Skip Backslash character if present
	if (*wp == UNICHR_BACKSLASH)
		wp++;

	if (pbsCnt == tpbsCnt && dbsCnt == tdbsCnt)
	{	//All info are here, process the transaction
		
		result = cifs_processTransac(opp, wp, setupp, parp, datp, swsCnt,
									pbsCnt,	dbsCnt, mswrCnt, mpbrCnt, mdbrCnt);
			

	} else {
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_transac: "
					"fragmented transaction not supported !"));
		cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_error, CIFSTATUS_UNSUCCESSFUL);
	}
	
	if (discTid)
		cifs_closeSessionShareInUse(ssp, siup);
	
	if (noAnswerReq)
	{
		cifs_closeOperation(opp);
		return (CIFS_DONTSENDRESP);
	}

	return (CIFS_SENDRESPNOW);
}



/*-----------------------------------------------------------------
 * Name			: cifs_encodeTransaction
 *
 * Description	: Encode a Transaction to be sent out
 * 				  Used for browse host announcement
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 12/07/2001
 *
 *-----------------------------------------------------------------
 */
void cifs_encodeTransaction(cifs_cmdDesc *msgp,
							char	 	*pipeName,
							ushort16 	*setupp,
							uchar8	 	*parp,
							uchar8	 	*datp,
							ushort16 	setupCnt,
							ushort16 	parCnt, 
							ushort16 	datCnt)
{
	uchar8 *ocmdp = msgp->cmdp;
	uchar8 *bpar, *bp;
	int32 pipeNameSize, alignData, i;

	pipeNameSize = strlen(pipeName) + 1;
	
	//Set command type
	*(msgp->cmdTypep) = SMB_COM_TRANSACTION;

	//Align transaction data on a 32 bits boundary
	alignData = parCnt % 4;
		
    CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_encodeTransac: "
				"sCnt=%d, pCnt=%d, dCnt=%d, pNS=%d al=%d",
				setupCnt, parCnt, datCnt, pipeNameSize, alignData));
	
	cifs_setCmdWordByteCnt(msgp, 14+setupCnt, 
								 parCnt+pipeNameSize+alignData+datCnt, TRUE);

	//Fill byte section
	bpar = SMBCP_BPAR(ocmdp);
	bp = bpar = cifs_strcpy(bpar, pipeName);
	if (parCnt)
		memcpy(bp, parp, parCnt);
	bp += parCnt + alignData;
	if (datCnt)
		memcpy(bp, datp, datCnt);

	//Fill word section
	PUTLE2(ocmdp, SMBCO_WP0,  parCnt);
	PUTLE2(ocmdp, SMBCO_WP1,  datCnt);
	PUTLE2(ocmdp, SMBCO_WP8,  parCnt);
	PUTLE2(ocmdp, SMBCO_WP9,  SMB_DIFF(bpar, msgp->smbp));
	PUTLE2(ocmdp, SMBCO_WP10, datCnt);
	PUTLE2(ocmdp, SMBCO_WP11, SMB_DIFF(bp, msgp->smbp));
	PCAR(ocmdp, SMBCO_WP12, setupCnt);
	for (i=0; i<setupCnt; i++)
		PUTLE2(ocmdp, SMBCO_WP13 + i*2, setupp[i]);

}



/*-----------------------------------------------------------------
 * Name			: cifs_buildBrowseAnnouncement
 *
 * Description	: build browse host announcement transaction
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 12/07/2001
 *
 *-----------------------------------------------------------------
 */
void cifs_buildBrowseAnnouncementMsg(cifs_server *svp, cifs_cmdDesc *cmddp)
{
	static uchar8 setup[] = { 1, 0, 0, 0, 2, 0 };
	static uchar8 data[80];
	char *bp;

	cifs_initSMBHeader(cmddp);
	
	//Prepare the data section of the transaction
	data[0] = CIFS_BROWSECMD_HOSTANNOUNCEMENT;
	data[1] = 0;
	PUTLE4(data, 2, (12 * 60 * 1000));			//Announcement freq: 12 minutes
	strncpy(data+6, svp->cfg.name, NB_S_NAME+1);//our name
	data[22] = CIFS_SERVER_VERSION_MAJOR;		//Major version number
	data[23] = CIFS_SERVER_VERSION_MINOR;		//Minor version number
	PUTLE4(data, 24, CIFS_SERVER_TYPE);			//Windows NT server
	PUTLE2(data, 30, 0x10F);					//Browser Election version
	PUTLE2(data, 32, 0xAA55);					//Very useful...
	bp = cifs_strcpy(data+34, "Here comes the fantastic NFX !");

	cifs_encodeTransaction(cmddp, "\\MAILSLOT\\BROWSE", (ushort16*)setup, NULL,
							data, 3, 0, SMB_DIFF(bp, data));
	
}


