/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * RCS:  $Id: cifs-msg.c,v 1.2 2002/12/10 18:13:18 laurentb Exp $
 *
 * Name			: cifs-msg.c
 *
 * Description	: Contains CIFS encoding and decoding functions. 
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/10/2001
 *
 *-----------------------------------------------------------------
 */

#include "cifs-api.h"
#include "cifs-utils.h"
#include "cifs-unicode.h"
#include "cifs-string.h"
#include "cifs-transport.h"
#include "cifs-encdec.h"
#include "cifs-err.h"
#include "cifs-msg.h"
#include "cifs-ipc.h"
#include "cifs-dc.h"
#include "cifs-intmsg.h"

#define IFDBG(lvl)  if (lvl <= cifs_debugLevel)


//Unicode strings


/*-----------------------------------------------------------------
 * Name			: cifsm_negociate 
 *
 * Description	: protocol negociation (0x72)
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/10/2001
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_negociate(	cifs_client *clp,
						cifs_session *ssp,
						cifs_shareInUse *siup,
						cifs_operation *opp,
						uchar8 *icmdp,
						uchar8 *ocmdp)
{
	cifs_serverCfg *cfgp = &clp->serverp->cfg;
	uchar8 *bpar, *pstr;
	ushort16 bcnt, idx;
	short16 protIdx;
	boolean encryption;
	uchar8 encryptKLen = 0;
	
	if (clp->sessionKey != 0)
	{	//Negociate protocol message already received !
        cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_error, CIFSTATUS_UNSUCCESSFUL);
		return (CIFS_SENDRESPNOW);
	}
	
	pstr    = bpar = SMBCP_BPAR(icmdp) + 1;
	bcnt    = GETLE2(icmdp, SMBCO_BCNT(icmdp));
	idx 	= 0;
	protIdx = -1;
	
    CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifsm_negociate: bcnt=%d", bcnt));
	
	while (pstr < (bpar + bcnt)) {
	    CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifsm_negociate: "
					"dialect %d: [%s]", idx, pstr));
		if (strcmp(pstr, "NT LM 0.12") == 0)
		{
			protIdx = idx;
			break;
		}
		pstr += strlen(pstr) + 2;
		idx++;
	}
	
    CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifsm_negociate: "
				"idx=%d protIdx=%d", idx, protIdx));
	
	//TODO find out more about client platform
	if (idx == 6) 
		clp->clientPlatform = CIFS_CLIENT_WIN9x;
	if (idx == 8)
	{
		 clp->clientPlatform = CIFS_CLIENT_WINNT;
		//TODO check with W2K client
		//if (GETLE2(opp->ReqMsg.smbp, SMBO_FLG2) & SMB_FLG2_EXTENDSEC)
		//  clp->clientPlatform = CIFS_CLIENT_WIN2K;
	}

	// Build the response
	if (protIdx == -1) 
	{	//No supported protocol
		cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_error, CIFSTATUS_UNSUCCESSFUL);
		cifs_setCmdWordByteCnt(&opp->respMsg, 1, 0, FALSE);		
		PUTLE2(ocmdp, SMBCO_WP0, protIdx);
	} else {
		encryption = cfgp->options & CIFS_SVCFG_ENCRYPTPWD;
		if (encryption)
            encryptKLen = CIFS_S_ENCRYPTKEY;
		bcnt = encryptKLen + strlen(cfgp->domainName) + 1;	//include ending 0
		cifs_setCmdWordByteCnt(&opp->respMsg, 17, bcnt, TRUE);
			
		PUTLE2(ocmdp, SMBCO_WP0, protIdx);
		//What a nice word ! MS couldn't have kept security mode as a short...
		PUTLE2(ocmdp, SMBCO_WP1+1, CIFS_N_MPXREQS);
		PUTLE2(ocmdp, SMBCO_WP2+1, CIFS_N_VCS);
		PUTLE4(ocmdp, SMBCO_WP3+1, CIFS_S_BUFFER);
		PUTLE4(ocmdp, SMBCO_WP5+1, CIFS_S_BUFFER);
		PUTLE4(ocmdp, SMBCO_WP7+1, clp);	//sessionKey = client object ptr
		PUTLE4(ocmdp, SMBCO_WP9+1, cfgp->capabilities);
				
		//TODO Init time !
		
		PCAR(ocmdp, SMBCO_WP16+1, encryptKLen);
		if (encryption)
		{
			PCAR(ocmdp, SMBCO_WP1, CIFS_SEC_USER|CIFS_SEC_ENCRYPTPWD);
			cifs_createEncryptionKey(clp->encryptKey);
			memcpy(SMBCP_BPAR(ocmdp), clp->encryptKey, CIFS_S_ENCRYPTKEY);
		}
		else
			PCAR(ocmdp, SMBCO_WP1, CIFS_SEC_USER);
		strcpy(SMBCP_BPAR(ocmdp) + encryptKLen, cfgp->domainName);

		clp->sessionKey = (uint32)clp;	//store the session key
	}
	return (CIFS_SENDRESPNOW);
}



/*-----------------------------------------------------------------
 * Name			: cifs_authenticateUser
 *
 * Description	: go and ask users credentials to DC
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/12/2002
 *
 *-----------------------------------------------------------------
 */
uint32	//CIFSTATUS
cifs_authenticateUser(	cifs_client *clp,
						cifs_session *ssp,
						cifs_operation *opp)
{
	esm_event_t	evt;
	cifsauthInfoMsg_t	*authInfoP;
	
	//Send the authentication request
	if (cifs_authReqToSsc(opp) != NFX_OK)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_authenticateUser: "
					"Failed to send request"));
		return (CIFSTATUS_UNSUCCESSFUL);
	}
	
	// Add the operation to the pending list of the session
	opp->status = cifs_opstatus_processing;
	DLL_INSERT_TAIL(&ssp->pendingOps, opp, entry);

	//Just wait for the response from DC to come back
#ifndef FS_STANDALONE
	thread_yield(&evt, 10);			//timeout in seconds
#else
	evt.event = THREAD_EVENT_TIMEOUT;	//FS STAND ALONE DEBUG
#endif
	
	// Remove the pending operation from operation pending list
	DLL_REMOVE(&ssp->pendingOps, opp, entry);
		
	switch (evt.event)	{
	
		case CIFSAUTH_INFO:
			authInfoP = (cifsauthInfoMsg_t *)(evt.p2.u.p);
			if (authInfoP->status == stat_ok)
				break;
			eee_freePacket((eee_desc_t*)evt.p1.u.p);
			if (authInfoP->status == failed_logon)
			{	//DC was contacted but authentication failed
		    	CIFS_PRINT(CIFS_DBGLV_INFO, 
					("cifsm_authenticateUser: auth failed (%x)", 
					 authInfoP->ntstatus));
				return (authInfoP->ntstatus);
			} else {
			//We didn't find the DC
			if (authInfoP->status == no_trust_pw)
		    	CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_authenticateUser: "
							"server has not joined the domain yet !"));
			return (CIFSTATUS_NO_LOGON_SERVERS);
		}
			
		case THREAD_EVENT_TIMEOUT:
			//Login as guest for now. Use anonymous token, until guest
			//is understood. TODO
			CIFS_PRINT(CIFS_DBGLV_DBG, ("cifsm_authenticateUser: "
						"timeout, use guest account"));
			ssp->accessTokenp = cifs_secCreateAccessToken(NULL);
			ssp->options |= CIFS_SESS_GUEST;
			return (CIFSTATUS_OK);
		
		default:
			NFX_ASSERT(FALSE);
	}

	IFDBG(CIFS_DBGLV_DBG3)
	{
	    CIFS_PRINT(CIFS_DBGLV_DBG3, ("cifsm_authenticateUser: success !"));
		cifs_showUserCreds(ssp, &(authInfoP->userCred));
	}

	//Save user credentials
	memcpy(&ssp->userCreds, &(authInfoP->userCred), sizeof(cifsauthCred_t));

	//Free received message
	eee_freePacket((eee_desc_t*)evt.p1.u.p);

	//Create access token for the session
	ssp->accessTokenp = cifs_secCreateAccessToken(&ssp->userCreds);

	return (CIFSTATUS_OK);
}


/*-----------------------------------------------------------------
 * Name			: cifsm_ssSetupNX
 *
 * Description	: session setup and X (0x73)
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/19/2001
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_ssSetupNX(	cifs_client *clp,
						cifs_session *ssp,
						cifs_shareInUse *siup,
						cifs_operation *opp,
						uchar8 *icmdp,
						uchar8 *ocmdp)
{
	cifs_serverCfg *cfgp = &clp->serverp->cfg;
	uchar8 *bpar, *bp, *smbp;
	uint32 len;
	uint32 rc = CIFSTATUS_UNSUCCESSFUL;
	boolean firstSession, unicode;

	if (clp->sessionKey == 0)
	{	//Negociate protocol message not yet received !
        cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_error, CIFSTATUS_UNSUCCESSFUL);
		return (CIFS_SENDRESPNOW);
	}
	
	if (DLL_IS_EMPTY(&clp->sessions))
		firstSession = TRUE;
	else
		firstSession = FALSE;

	//Create a session
	ssp = cifs_addSession(clp);
	if (ssp == NULL) {
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_ssSetupNX: no avail session obj"));
		cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_toomanyuids,
						CIFSTATUS_TOO_MANY_SESSIONS);
		return (CIFS_SENDRESPNOW);
	}
	
	if (firstSession)
	{
		clp->maxBufSize		= GETLE2(icmdp, SMBCO_WP2);
		clp->capabilities	= GETLE4(icmdp, SMBCO_WP11);
	}
//	ssp->maxMpxReqs		= GETLE2(icmdp, SMBCO_WP3);
	//TODO see what to do with VC number
	ssp->caseInsenPwdLen= GETLE2(icmdp, SMBCO_WP7);
	ssp->caseSenPwdLen	= GETLE2(icmdp, SMBCO_WP8);

    CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifsm_ssSetupNX: "
				"bufSize=%d, ipwdl=%d, spwd=%d", 
				clp->maxBufSize, ssp->caseInsenPwdLen,ssp->caseSenPwdLen));
	
	if ((ssp->caseInsenPwdLen > CIFS_S_ENCRYPTPWD)
	 || (ssp->caseSenPwdLen   > CIFS_S_ENCRYPTPWD))
	{	rc = CIFSTATUS_WRONG_PASSWORD;	goto ssError; }
	//TODO test with other clients and see what to do with passwords
	//and encrypted/plain text pwds.
	
	bp = SMBCP_BPAR(icmdp);
	memcpy(ssp->caseInsenPwd, bp, ssp->caseInsenPwdLen);
	bp += ssp->caseInsenPwdLen;
	memcpy(ssp->caseSenPwd, bp, ssp->caseSenPwdLen);
	bp += ssp->caseSenPwdLen;
	
	unicode = ((GETLE2(opp->reqMsg.smbp, SMBO_FLG2) & SMB_FLG2_UNICODE) != 0);

    CIFS_PRINT(CIFS_DBGLV_DBG3, ("cifsm_ssSetupNX: user@=%p (%s)", bp, bp));
	bp = cifs_getUniStr(ssp->user, bp, unisizeof(ssp->user), unicode);
	if (bp == NULL)
	{   
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_ssSetupNX: user name too long"));
		goto ssError;
	}
    CIFS_PRINT(CIFS_DBGLV_DBG3, ("cifsm_ssSetupNX: domain@=%p", bp));
	bp = cifs_getUniStr(ssp->domain, bp,  unisizeof(ssp->domain), unicode);		
	if (bp == NULL)
	{	
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_ssSetupNX: user name too long"));
		goto ssError;
	}

    IFDBG(CIFS_DBGLV_DBG2) 
	{
		uniprintf("cifsm_ssSetupNX: user:", ssp->user);
		uniprintf(", domain:", ssp->domain);
		printf("\n\tuserlen=%d, usersize=%d, domainlen=%d, domainsize=%d\n",
				unistrlen(ssp->user), unistrsize(ssp->user),
				unistrlen(ssp->domain), unistrsize(ssp->domain));
	}
	
	//assign operation's session 
	opp->ssp = ssp;
	
	if (*ssp->user == 0)	//anonymous access
	{
		CIFS_PRINT(CIFS_DBGLV_DBG, ("cifsm_ssSetupNX: anonymous access"));
		ssp->accessTokenp = cifs_secCreateAccessToken(NULL);
		ssp->options |= CIFS_SESS_ANONYMOUS;
	} else {
		rc = cifs_authenticateUser(clp, ssp, opp);
		if (rc != CIFSTATUS_OK)
			goto ssError;
	}
	
	//Create a new uid for this session
	ssp->uid = cifs_newUid(clp);

	//Build the response
	
	smbp = opp->respMsg.smbp;
	
	//Set unicode bit if supported by client
	if (clp->capabilities & CAP_UNICODE)
	{
		ushort16 flags2 = GETLE2(smbp, SMBO_FLG2)|SMB_FLG2_UNICODE;
		PUTLE2(smbp, SMBO_FLG2, flags2);
	}

	PUTLE2(smbp, SMBO_UID, ssp->uid);

	//Fill word section
	cifs_setCmdWordCnt(&opp->respMsg, 3, TRUE);
	cifs_initNXCmdHeader(ocmdp);
	PUTLE2(ocmdp, SMBCO_WP2, (ssp->options & CIFS_SESS_GUEST));
	
	//Fill byte section
	bpar = bp = SMBCP_BPAR(ocmdp);
	bp += cifs_putStr(bp, CIFS_OS, smbp, &len, 32);
	bp += cifs_putStr(bp, CIFS_LANMAN_TYPE, smbp, &len, 64);
	bp += cifs_putStr(bp, clp->serverp->cfg.domainName, smbp, &len, 64);
	cifs_setCmdByteCnt(&opp->respMsg, SMB_DIFF(bp, bpar), FALSE);

    CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifsm_ssSetupNX: smbl=%d, GCAR=%x, batch=%d", 
				opp->respMsg.smbLen, GCAR(icmdp, SMBCO_WP0),
				SMB_BATCHEDCMDFOLLOWS(icmdp)));

	if (SMB_BATCHEDCMDFOLLOWS(icmdp))
		return (cifs_processNextCmd(opp, ocmdp));
	else
		return (CIFS_SENDRESPNOW);
	
ssError:
	cifs_errResp(opp, 0, 0, rc);
	cifs_closeClientSession(clp, ssp);
	return (CIFS_SENDRESPNOW);
}


/*-----------------------------------------------------------------
 * Name			: cifsm_logoffNX
 *
 * Description	: User logoff (0x74)
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 12/12/2001
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_logoffNX(	cifs_client *clp,
						cifs_session *ssp,
						cifs_shareInUse *siup,
						cifs_operation *opp,
						uchar8 *icmdp,
						uchar8 *ocmdp)
{

	CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifsm_logoffNX: uid = %d", siup->tid));

	// Close client session
	// TODO - We may need a return value from the function to be
	//        either CIFS_SENDRESPNOW or CIFS_DONTSENDRESP
	cifs_closeClientSession(clp, ssp);

	//Build the response
	cifs_setCmdWordByteCnt(&opp->respMsg, 2, 0, TRUE);
	cifs_initNXCmdHeader(ocmdp);

	if (SMB_BATCHEDCMDFOLLOWS(icmdp))
		return (cifs_processNextCmd(opp, ocmdp));
	else
		return (CIFS_SENDRESPNOW);
} /* cifsm_logoffNX() */



/*-----------------------------------------------------------------
 * Name			: cifsm_treeConNX
 *
 * Description	: Tree Connect and X (0x75)
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/28/2001
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_treeConNX(	cifs_client *clp,  
						cifs_session *ssp,
						cifs_shareInUse *siup,
						cifs_operation *opp,
						uchar8 *icmdp,
						uchar8 *ocmdp)
{
	cifs_share *shp;
	uchar8 *bpar, *bp, *smbp;
	ushort16 *wp;
	uint32 len, rc = CIFSTATUS_BAD_NETWORK_NAME;
	boolean unicode;
	uni16 shareName[CIFS_S_SHARENAME+1];

	//Check if we have to close a previous share in use
	if (GETLE2(icmdp, SMBCO_WP2) & 0x1)
	{
		if (siup == NULL)
			siup = cifs_findClientSessionShareInUse(clp, ssp, 
										GETLE2(opp->reqMsg.smbp, SMBO_TID));
		if (siup != NULL)
			cifs_closeSessionShareInUse(ssp, siup);
	}
		
	//Check password length
	len = GETLE2(icmdp, SMBCO_WP3);
	if (len > CIFS_S_ENCRYPTPWD)
	{
        cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_badpw, CIFSTATUS_WRONG_PASSWORD);
		return (CIFS_SENDRESPNOW);
	}

	//Create a share in use object
	siup = cifs_addShareInUse(ssp);
	if (siup == NULL) {
		CIFS_PRINT(CIFS_DBGLV_ERR,
					("cifsm_treeConNX: no avail shareInUse obj"));
		cifs_errResp(opp, 0, 0,	CIFSTATUS_INSUFFICIENT_RESOURCES);
		return (CIFS_SENDRESPNOW);
	}
	
	bp = SMBCP_BPAR(icmdp) + len;	//Skip password
	
	unicode = ((GETLE2(opp->reqMsg.smbp, SMBO_FLG2) & SMB_FLG2_UNICODE) != 0);
			
	if (unicode)
	{			//resource name is in UNICODE
		bp += ((address_t)bp) % 2;	//Unicode string are aligned on word
		wp = (ushort16*)bp;

		//Skip server name, assume it is ours
		if ((*(wp++) != UNICHR_BACKSLASH)
		 || (*(wp++) != UNICHR_BACKSLASH)
		 || ((wp = unistrchr(wp, UNICHR_BACKSLASH)) == NULL))
		{
			CIFS_PRINT(CIFS_DBGLV_ERR, 
						("cifsm_treeConNX: wrong share name format"));
			goto tcError;
		}
		wp++;
		bp = (uchar8*)wp;
	} else {	//resource name is in ASCII
		//Skip server name, assume it is ours
		if ((*(bp++) != '\\')
		 || (*(bp++) != '\\')
		 || ((bp = strchr(bp, '\\')) == NULL))
		{
			CIFS_PRINT(CIFS_DBGLV_ERR,
						("cifsm_treeConNX: wrong share name format"));
			goto tcError;
		}
		bp++;
	}
	bp = cifs_getUniStr(shareName, bp, unisizeof(shareName), unicode);
	if (bp == NULL)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_treeConNX: share name too long"));
		goto tcError;
	}
	
	//Find the share name in the server's configuration		
	shp = cifs_findServerShareByName(clp->serverp, shareName);
	if (shp == NULL) {
		IFDBG(CIFS_DBGLV_ERR)
		{
			uniprintf("cifsm_treeConNX: share ", shareName);
			printf(" does not exist\n");
		}
		goto tcError;
	}
	
	if (shp->cfg.flags & CIFS_SHCFG_ISADMIN)
	{	//It's a pipe
		siup->flags |= CIFS_SIU_ISIPC;
		siup->access = SEC_GENR_FILE_ALL;
		//Init pipes list
		DLL_INIT(&siup->u.pipe.openPipes);
	}
	else 
	{
		//We check if we can at least Read data on the share
		if (cifs_secCheckAccess(shp->cfg.secDescp, ssp->accessTokenp,
								SEC_FILE_READ_DATA|SEC_MAXIMUM_ALLOWED_ACCESS,
								&siup->access) != CIFSTATUS_OK)
		{
			rc = CIFSTATUS_ACCESS_DENIED;
			goto tcError;
		}
		//Here we have in siup->access the maximum allowed access
		
		//Init disk lists
		DLL_INIT(&siup->u.disk.openFiles);
		DLL_INIT(&siup->u.disk.searches);
	}
	siup->sharep = shp;

    IFDBG(CIFS_DBGLV_DBG) 
	{
		uniprintf("cifsm_treeConNX: connecting to share:", shareName);
		printf(" with access: 0x%08x\n", siup->access);
	}
	
	//Increment share use
	shp->useCnt++;
	
	//Create a new tid for this share in use
	siup->tid = cifs_newTid(clp);
	

	//Build the response

	smbp = opp->respMsg.smbp;
	
	PUTLE2(smbp, SMBO_TID, siup->tid);
	
	//Fill word section
	cifs_setCmdWordCnt(&opp->respMsg, 3, TRUE);
	cifs_initNXCmdHeader(ocmdp);
	PUTLE2(ocmdp, SMBCO_WP2, SMB_SUPPORT_SEARCH_BITS);
	
	//Fill byte section
	bpar = bp = SMBCP_BPAR(ocmdp);
	if (siup->flags & CIFS_SIU_ISIPC)
	{
		bp = cifs_strcpy(bp, "IPC");	//Service name is never in unicode
		bp += cifs_putStr(bp, "", smbp, &len, 16);
	} else {
		bp = cifs_strcpy(bp, "A:");		//Service name is never in unicode
		bp += cifs_putStr(bp, "NTFS", smbp, &len, 16);
	}
	cifs_setCmdByteCnt(&opp->respMsg, SMB_DIFF(bp, bpar), FALSE);
	
	opp->siup = siup;			//assign operation's shareInUse for batched cmds
	if (SMB_BATCHEDCMDFOLLOWS(icmdp))
		return cifs_processNextCmd(opp, ocmdp);
	else	
		return (CIFS_SENDRESPNOW);
	
tcError:
	cifs_errResp(opp, 0, 0,	rc);
	cifs_closeSessionShareInUse(ssp, siup);
	return (CIFS_SENDRESPNOW);
}



/*-----------------------------------------------------------------
 * Name			: cifsm_treeDisc
 *
 * Description	: Tree Disconnect (0x71)
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 12/12/2001
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_treeDisc(	cifs_client *clp,
						cifs_session *ssp,
						cifs_shareInUse *siup,
						cifs_operation *opp,
						uchar8 *icmdp,
						uchar8 *ocmdp)
{
	CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifsm_treeDisc: tid = %d", siup->tid));

	//Decrement share use
	siup->sharep->useCnt--;

	// Close share in use
	// TODO - We may need a return value from the function to be
	//        either CIFS_SENDRESPNOW or CIFS_DONTSENDRESP
	cifs_closeSessionShareInUse(ssp, siup);

	// Build the response

	// Set word count and byte count to 0
	cifs_setCmdWordByteCnt(&opp->respMsg, 0, 0, TRUE);

	return (CIFS_SENDRESPNOW);
} /* cifsm_treeDisc() */



int32 cifsm_treeCon(cifs_client *clp, cifs_session *ssp,
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


/*-----------------------------------------------------------------
 * Name			: cifsm_echo
 *
 * Description	: echo (0x2B) 
 * 				  Ping the server.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/12/2002
 *
 *-----------------------------------------------------------------
 */
int32 cifsm_echo(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	ushort16		echoCnt, byteCnt, sequence;
	uchar8			*echoData, *obytep;
	esm_event_t		evt;

	// Parse echo request
	echoCnt = GETLE2(icmdp, SMBCO_WP0);	// Number of times to echo data back
	byteCnt = GETLE2(icmdp, SMBCO_BCNT(icmdp));
	echoData = SMBCP_BPAR(icmdp);

	CIFS_PRINT(CIFS_DBGLV_DBG3, ("cifsm_echo: "
				"echoCnt = %d, byteCnt = %d", echoCnt, byteCnt));

	if (echoCnt == 0)
		return (CIFS_DONTSENDRESP);

	sequence = 0;

	while(echoCnt)
	{
		//Build the response
		cifs_setCmdWordByteCnt(&opp->respMsg, 1, byteCnt, TRUE);

		sequence++;
		PUTLE2(ocmdp, SMBCO_WP0, sequence);		// Sequence number

		obytep = SMBCP_BPAR(ocmdp);
		memcpy(obytep, echoData, byteCnt);

		//Send the response message
		cifs_sendClientMsg(opp->clp, opp);
		if (--echoCnt)
			opp->flags = CIFS_OPFLAG_MORE_TO_SEND;

		CIFS_PRINT(CIFS_DBGLV_DBG3, ("cifsm_echo: echoCnt = %d", echoCnt));

		//To send another response message
		if (echoCnt)
		{
			// Send out multiple echo responses for one request

			// Get and init xmt buffer
			if (cifs_initXmtBuffer(opp, NULL, NULL) == NULL)
			{
				// No available buffer
				opp->status = cifs_opstatus_sendPending;

				thread_yield(&evt, CIFS_INTMSG_TIMEOUT_WAITBUF);

				cifs_treatKillEvent(&evt);

				if (evt.event == CIFS_THREADEVT_BUF_AVAILABLE)
				{
					cifs_initXmtBuffer(opp,
										(pkt_desc_t *)(evt.p1.u.p),     //pktp
										(uchar8 *)(evt.p2.u.p));        //osmbp
				}
				else
				{
					CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_echo: "
								"Received unexpected event"));
					break;
				}
			}
		}
	}

	return (CIFS_DONTSENDRESP);
}


/*-----------------------------------------------------------------
 * Name			: cifsm_ 
 *
 * Description	: 
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 02/10/2002
 *
 *-----------------------------------------------------------------
 */



int32 cifsm_flush(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_rename(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_queryInfo(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_setInfo(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_lockByte(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_ulockByte(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_createTmp(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_exit(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_seek(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_lockRead(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_wrUnlock(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_setInf2(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_queryInf2(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_lockNX(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_transacS(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_ioctl(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_copy(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_move(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_wrClose(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_transac2S(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_findNClos(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_infoDisk(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_search(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_findClose(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_ntTrans(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_ntTransS(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


int32 cifsm_ntCancel(cifs_client *clp, cifs_session *ssp, 
					cifs_shareInUse *siup, cifs_operation *opp,
					uchar8 *icmdp, uchar8 *ocmdp)
{
	return (CIFS_SENDRESPNOW);
}


