/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 * 
 * RCS:  $Id: cifs-objects.c,v 1.1.1.1 2002/04/23 17:41:16 laurentb Exp $
 *
 * Name			: cifs-objects.c
 *
 * Description	: Contains CIFS objects functions.
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 10/31/2001
 *
 *-----------------------------------------------------------------
 */

#include "cifs-api.h"
#include "cifs-utils.h"
#include "cifs-unicode.h"
#include "cifs-transport.h"



//Unicode strings
uni16 unistr_ipcDol[] = { 'I', 'P', 'C', '$', 0 };
uni16 unistr_adminDol[] = { 'A', 'D', 'M', 'I', 'N', '$', 0 };


//***************************************************************************
//*********************** Finding objects Tools *****************************
//***************************************************************************

/*-----------------------------------------------------------------
 * Name			: cifs_findSearch
 *
 * Description	: look for a search in 'share in use' associated with
 * 				  SMB find operation
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/05/2002
 *-----------------------------------------------------------------
 */
cifs_search *	// NULL if no match
cifs_findSearch(cifs_operation *opp, ushort16 sid)
{
	cifs_search	*searchp;

	DLL_ITERATE(searchp, &opp->siup->u.disk.searches, entry)
	{
		if (searchp->findHandle == sid)
			return (searchp);
	}
	return (NULL);
}


/*-----------------------------------------------------------------
 * Name			: cifs_findShareInUsePipe
 *
 * Description	: find an open pipe in share in use environment based on fid
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 12/05/2001
 *-----------------------------------------------------------------
 */
cifs_pipe *
cifs_findShareInUsePipe(cifs_shareInUse *siup, ushort16 fid)
{
	cifs_pipe *pip;
	
	DLL_ITERATE(pip, &siup->u.pipe.openPipes, entry)
	{
/*	DEBUG
		if (pip->fid == fid)
*/		
			return pip;
	}
	return (NULL);
}


/*-----------------------------------------------------------------
 * Name			: cifs_findClientSessionShareInUse
 *
 * Description	: find share in use environment based on tid
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/14/2001
 *-----------------------------------------------------------------
 */
cifs_shareInUse *
cifs_findClientSessionShareInUse(cifs_client *clp,
								cifs_session *ssp,
								ushort16 tid)
{
	cifs_shareInUse *siup;
	cifs_session *clssp;	

	DLL_ITERATE(siup, &ssp->sharesInUse, entry)
	{
/*	DEBUG
		if (siup->tid == tid)
*/		
if ((siup->flags & CIFS_SIU_ISIPC) == 0)//DEBUG for disk share

			return (siup);
	}

	//Client may use a TID got with another UID if it is the IPC share(at least)
	//Look through all the sessions of this client to find the TID and
	//check whether it is an IPC share.
	//Otherwise for now say we have not found it
	DLL_ITERATE(clssp, &clp->sessions, entry)
	{
		DLL_ITERATE(siup, &clssp->sharesInUse, entry)
		{
//DEBUG			if (siup->tid == tid && (siup->flags & CIFS_SIU_ISIPC))
if (siup->flags & CIFS_SIU_ISIPC)				
				return (siup);
		}
	}
	
	return (NULL);
}


/*-----------------------------------------------------------------
 * Name			: cifs_findServerShareByName
 *
 * Description	: find share from its name
 * 
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/28/2001
 *-----------------------------------------------------------------
 */
cifs_share *
cifs_findServerShareByName(cifs_server *svp, uni16 *shareName)
{
	cifs_share *shp;
	
	DLL_ITERATE(shp, &svp->shares, entry)
	{
		if (unistrcasecmp(shareName, shp->cfg.name) == 0)
			return (shp);
	}
	return (NULL);
}

/*-----------------------------------------------------------------
 * Name			: cifs_findServerShareByThread
 *
 * Description	: find share from its thread id
 * 
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/15/2002
 *-----------------------------------------------------------------
 */
cifs_share *
cifs_findServerShareByThread(cifs_server *svp, esm_handle_t threadId)
{
	cifs_share *shp;
	
	DLL_ITERATE(shp, &svp->shares, entry)
	{
		if (threadId == shp->threadId)
			return (shp);
	}
	return (NULL);
}


/*-----------------------------------------------------------------
 * Name			: cifs_findClientSession
 *
 * Description	: find session environment based on uid
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/14/2001
 *-----------------------------------------------------------------
 */
cifs_session *
cifs_findClientSession(cifs_client *clp, ushort16 uid)
{
	cifs_session *ssp;

	DLL_ITERATE(ssp, &clp->sessions, entry)
	{
/*	DEBUG
		if (ssp->uid == uid)
*/		
			return (ssp);
	}
	return (NULL);
}


/*-----------------------------------------------------------------
 * Name			: cifs_findFileOpenByFid
 *
 * Description	: find file open from SMB FID
 * 
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/26/2002
 *-----------------------------------------------------------------
 */
cifs_fileOpen *
cifs_findFileOpenByFid(cifs_shareInUse *siup, ushort16 fid)
{
	cifs_fileOpen	*fop;

	DLL_ITERATE(fop, &siup->u.disk.openFiles, entry)
	{
/*	DEBUG			
		if (fop->fid == fid)
*/		
			return (fop);
	}
	return (NULL);
}


//***************************************************************************
//******************** Closing/Deleting objects Tools ***********************
//***************************************************************************

/*-----------------------------------------------------------------
 * Name			: cifs_closeSearchBySid
 *
 * Description	: Clean search environment when a search is closed.
 * 				  If sid is -1, clean the search list.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/05/2002
 *-----------------------------------------------------------------
 */
void
cifs_closeSearchBySid(cifs_shareInUse *siup, short16 sid)
{
	cifs_search *searchp, *nsearchp;

	if (sid == -1)
	{
		// Clean up the whole search list
		DLL_ITERATE_NEXT(searchp, nsearchp, &siup->u.disk.searches, entry)
			cifs_freeSearch(searchp);
	} else {
		// Remove one search that matches sid
		DLL_ITERATE(searchp, &siup->u.disk.searches, entry)
		{
			if (searchp->findHandle == sid)
			{
				DLL_REMOVE(&siup->u.disk.searches, searchp, entry);
				cifs_freeSearch(searchp);
			}
		}
	}
}

/*-----------------------------------------------------------------
 * Name			: cifs_closeSearchByPtr
 *
 * Description	: Clean search environment when a search is closed.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/05/2002
 *-----------------------------------------------------------------
 */
void
cifs_closeSearchByPtr(cifs_shareInUse *siup, cifs_search *searchp)
{
	DLL_REMOVE(&siup->u.disk.searches, searchp, entry);
	cifs_freeSearch(searchp);
}

/*-----------------------------------------------------------------
 * Name			: cifs_closeOperation
 *
 * Description	: clean operation environment when operation is completed
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/14/2001
 *-----------------------------------------------------------------
 */
void
cifs_closeOperation(cifs_operation *opp)
{
	pkt_desc_t		*currp, *nextp;
	
	//Free in buffer
	cifs_freeInBuf(opp->clp, &opp->reqMsg.tplPktDescp);

	//Free out buffer(s)
	currp = nextp = opp->respMsg.tplPktDescp;
#ifndef CIFSTEST
	while(currp)
	{
		nextp = currp->pNext;
		//TODO understand and remove this
		//nb_freeOutgoingPkt(opp->clp->transportCBp, currp);
		cifs_freeOutBuf(opp->clp, opp, &currp);
		currp = nextp;
	}
#endif

	//Free operation structure
	cifs_freeOperation(opp);	
}

		
/*-----------------------------------------------------------------
 * Name			: cifs_closeShareInUseFileOpen
 *
 * Description	: clean file open environment
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 03/04/2002
 *-----------------------------------------------------------------
 */
void
cifs_closeShareInUseFileOpen(cifs_shareInUse *siup, cifs_fileOpen *fop)
{
	//free file open file name
	if (fop->pathName != NULL)
		CIFS_RAM_DEALLOC(fop->pathName);
	
	//Get out of the open files list
	DLL_REMOVE(&siup->u.disk.openFiles, fop, entry);
	
	//Free fileOpen structure
	cifs_freeFileOpen(fop);
}
		
		
/*-----------------------------------------------------------------
 * Name			: cifs_closeShareInUsePipe
 *
 * Description	: clean pipe environment
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/30/2001
 *-----------------------------------------------------------------
 */
void
cifs_closeShareInUsePipe(cifs_shareInUse *siup, cifs_pipe *pip)
{

	//Get out of the open pipes list
	DLL_REMOVE(&siup->u.pipe.openPipes, pip, entry);
	
	//Free pipe structure
	cifs_freePipe(pip);
}
		
		
/*-----------------------------------------------------------------
 * Name			: cifs_closeShareInUse
 *
 * Description	: clean share in use environment
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/12/2001
 *-----------------------------------------------------------------
 */
void
cifs_closeSessionShareInUse(cifs_session *ssp, cifs_shareInUse *siup)
{
	//TODO...

}
		
/*-----------------------------------------------------------------
 * Name			: cifs_closeClientSession
 *
 * Description	: clean client session environment when client session ends
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/12/2001
 *-----------------------------------------------------------------
 */
void
cifs_closeClientSession(cifs_client *clp, cifs_session *ssp)
{
	cifs_shareInUse *siup, *nsiup;
	cifs_operation *opp, *nopp;

	//Close all shares in use by the session
	DLL_ITERATE_NEXT(siup, nsiup, &ssp->sharesInUse, entry)
	{
		//Decrement share use
		siup->sharep->useCnt--;
		cifs_closeSessionShareInUse(ssp, siup);
	}

	//TODO Change this !!!!!!!!!!!!!!!
	//Mark all current pending operations as canceled
	DLL_ITERATE_NEXT(opp, nopp, &ssp->pendingOps, entry)
		opp->status = cifs_opstatus_canceled;
	
//TODO FIX THIS !!!	ssp->pendingOpsp = NULL;

	//Free access token
	cifs_secDeleteAccessToken(ssp->accessTokenp);
	
	//Get out of the client's sessions list
	DLL_REMOVE(&clp->sessions, ssp, entry);

	//Free session structure
	cifs_freeSession(ssp);
}

/*-----------------------------------------------------------------
 * Name			: cifs_closeClient
 *
 * Description	: clean client environment when client connection ends
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/12/2001
 *
 *-----------------------------------------------------------------
 */
void
cifs_closeClient(cifs_client *clp)
{
	cifs_session *ssp, *nssp;
	cifs_client *nextClp;

	//Close all client's sessions
	DLL_ITERATE_NEXT(ssp, nssp, &clp->sessions, entry)
		cifs_closeClientSession(clp, ssp);
	
	//Get out of the server's clients list
	DLL_REMOVE(&clp->serverp->clients, clp, entry);

	//Free client structure
	cifs_freeClient(clp);
}


/*-----------------------------------------------------------------
 * Name			: cifs_abortClientConnection
 *
 * Description	: abort the client connection and clean
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/09/2001
 *-----------------------------------------------------------------
 */
void
cifs_abortClientConnection(cifs_client *clp)
{
	cifs_closeSessionLayer(clp);
	cifs_closeClient(clp);
}


/*-----------------------------------------------------------------
 * Name			: cifs_deleteShare
 *
 * Description	: clean share environment when a share is deleted
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/16/2002
 *
 *-----------------------------------------------------------------
 */
void
cifs_deleteShare(cifs_server *svp, cifs_share *shp)
{
	cifs_client *clp;
	cifs_session *ssp;
	cifs_shareInUse *siup;

	//Close all server's clients's session's shareInUses linked with this share
	DLL_ITERATE(clp, &svp->clients, entry)
	{
		DLL_ITERATE(ssp, &clp->sessions, entry)
		{
			DLL_ITERATE(siup, &ssp->sharesInUse, entry)
			{
				if (siup->sharep == shp)
					cifs_closeSessionShareInUse(ssp, siup);
			}
		}
	}

	//Get out of the server's shares list
	DLL_REMOVE(&svp->shares, shp, entry);

	//Free the security descriptor associated with the share, if any    
	if (shp->cfg.secDescp != NULL)
		cifs_secFreeSecDesc(shp->cfg.secDescp, TRUE);

	cifs_freeShare(shp);		//Free share structure

	//Decrement server's share count
	svp->shareCnt--;
}


/*-----------------------------------------------------------------
 * Name			: cifs_deleteServer
 *
 * Description	: clean server environment when a server is deleted
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/16/2002
 *
 *-----------------------------------------------------------------
 */
void
cifs_deleteServer(cifs_server *svp)
{
	cifs_client *clp, *nclp;
	cifs_share  *shp, *nshp;

	//Close all server's clients
	DLL_ITERATE_NEXT(clp, nclp, &svp->clients, entry)
		cifs_abortClientConnection(clp);

	//Delete all server's shares
	DLL_ITERATE_NEXT(shp, nshp, &svp->shares, entry)
		cifs_deleteShare(svp, shp);
	
	//Get out of the servers list
	DLL_REMOVE(&cifs_globalCfgp->servers, svp, entry);

	cifs_freeServer(svp);		//Free server structure
}



//***************************************************************************
//*********************** creating objects Tools ****************************
//***************************************************************************

/*-----------------------------------------------------------------
 * Name			: cifs_addSearch
 *
 * Description	: add a new searh to 'share in use' associated with
 * 				  SMB find operation
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/05/2002
 *-----------------------------------------------------------------
 */
cifs_search *	// NULL if error
cifs_addSearch(cifs_operation *opp)
{
	static ushort16	id = 0;

	cifs_shareInUse	*siup;
	cifs_search		*searchp;

	searchp = cifs_allocSearch();
	if (searchp == NULL)
		return (NULL);

	siup = opp->siup;
	
	//Insert search in the share in use's searches list
	DLL_INSERT_TAIL(&siup->u.disk.searches, searchp, entry);

	// Generate a search handle
	id++;
	searchp->findHandle = (ushort16)((address_t)searchp) + id;
	return (searchp);
}


/*-----------------------------------------------------------------
 * Name			: cifs_addFileOpen
 *
 * Description	: add a file open to a share in use
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 04/03/2002
 *-----------------------------------------------------------------
 */
cifs_fileOpen *	//NULL if error
cifs_addFileOpen(cifs_shareInUse *siup)
{
	cifs_fileOpen *fop;

	fop = cifs_allocFileOpen();
	if (fop == NULL) 
		return (NULL);

	//generate fid
	fop->fid = cifs_newFid(siup);
			
	//Insert fop in the share in use's file open list
	DLL_INSERT_TAIL(&siup->u.disk.openFiles, fop, entry);
	
	return (fop);
}


/*-----------------------------------------------------------------
 * Name			: cifs_addPipe
 *
 * Description	: add a pipe to an ipc share in use
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/30/2001
 *-----------------------------------------------------------------
 */
cifs_pipe *	//NULL if error
cifs_addPipe(cifs_shareInUse *siup)
{
	cifs_pipe *pip;

	//NB: cifs_allocPipe DOES NOT set to 0 the pipe structure because it is
	//too big
	pip = cifs_allocPipe();
	if (pip == NULL) 
		return (NULL);

	//generate fid
	pip->fid = cifs_newFid(siup);
			
	//Insert pipe in the share in use's open pipes list
	DLL_INSERT_TAIL(&siup->u.pipe.openPipes, pip, entry);
	
	return (pip);
}


/*-----------------------------------------------------------------
 * Name			: cifs_addShareInUse
 *
 * Description	: add a share in use to a session
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/28/2001
 *-----------------------------------------------------------------
 */
cifs_shareInUse *	//NULL if error
cifs_addShareInUse(cifs_session *ssp)
{
	cifs_shareInUse *siup;

	siup = cifs_allocShareInUse();
	if (siup == NULL) 
		return (NULL);

	//Insert share in use in the session's share in use list
	DLL_INSERT_TAIL(&ssp->sharesInUse, siup, entry);
	
	return (siup);
}


/*-----------------------------------------------------------------
 * Name			: cifs_addSession
 *
 * Description	: add a session to a client
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/19/2001
 *-----------------------------------------------------------------
 */
cifs_session *	//NULL if error
cifs_addSession(cifs_client *clp)
{
	cifs_session *ssp;
	static ushort16	id = 0;

	ssp = cifs_allocSession();
	if (ssp == NULL) 
		return (NULL);

	clp->tidCnt  = (address_t)ssp;	//get very <> tids from other sessions
	clp->tidCnt += id++;
	
	//Init session lists
	DLL_INIT(&ssp->sharesInUse);
	DLL_INIT(&ssp->pendingOps);
	
	//Insert session in the client's sessions list
	DLL_INSERT_TAIL(&clp->sessions, ssp, entry);
	
	return (ssp);
}


/*-----------------------------------------------------------------
 * Name			: cifs_addClient
 *
 * Description	: add a client to a server
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/12/2001
 *-----------------------------------------------------------------
 */
cifs_client *	//NULL if error
cifs_addClient(cifs_server *svp)
{
	cifs_client *clp;
	static ushort16	id = 0;

	clp = cifs_allocClient();
	if (clp == NULL) 
		return (NULL);
	clp->serverp = svp;
	clp->uidCnt  = (address_t)clp;	//get very <> uids from other clients
	clp->uidCnt += id++;

	//Init client lists
	DLL_INIT(&clp->sessions);
	DLL_INIT(&clp->waitOutBufOps);
	
	//Insert client in the server's clients list
	DLL_INSERT_TAIL(&svp->clients, clp, entry);
	
	return (clp);
}


/*-----------------------------------------------------------------
 * Name			: cifs_addShare
 *
 * Description	: add a share to a CIFS server
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/16/2002
 *-----------------------------------------------------------------
 */
cifs_share*	//NULL if error
cifs_addShare(cifs_server *svp, cifs_shareCfg *shCfgp)
{
	cifs_share *shp;

	shp = cifs_allocShare();
	if (shp == NULL) 
		return (NULL);

	//Insert share in the server's shares list
	DLL_INSERT_TAIL(&svp->shares, shp, entry);
	
	//Increment server's share count
	svp->shareCnt++;
	
	//set share configuration
	memcpy(&shp->cfg, shCfgp, sizeof(cifs_shareCfg));

	return (shp);
}


/*-----------------------------------------------------------------
 * Name			: cifs_addServer
 *
 * Description	: add a CIFS server
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/16/2002
 *-----------------------------------------------------------------
 */
cifs_server*	//NULL if error
cifs_addServer(cifs_serverCfg *svCfgp)
{
	cifs_server *svp;

	svp = cifs_allocServer();
	if (svp == NULL) 
		return (NULL);

	//Init server lists
	DLL_INIT(&svp->shares);
	DLL_INIT(&svp->clients);

	//Insert server in the list of servers
	DLL_INSERT_TAIL(&cifs_globalCfgp->servers, svp, entry);
	
	//set server configuration
	memcpy(&svp->cfg, svCfgp, sizeof(cifs_serverCfg));
	
	return (svp);
}


/*-----------------------------------------------------------------
 * Name			: cifs_createServer
 *
 * Description	: create a CIFS server with its default shares
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 01/16/2002
 *-----------------------------------------------------------------
 */
cifs_server*	//NULL if error
cifs_createServer(cifs_serverCfg *svCfgp)
{
	cifs_server	*svp;
	static cifs_shareCfg shCfg;

	svp = cifs_addServer(svCfgp);
	if (svp == NULL) 
		return (NULL);
	
	//create IPC$ share
	unistrcpy(shCfg.name, unistr_ipcDol);
	strDos2Uni(shCfg.comment, "Remote IPC", unisizeof(shCfg.comment), TRUE);
	shCfg.pathName[0] = '\0';
	shCfg.flags = CIFS_SHCFG_IPC;
	if (cifs_addShare(svp, &shCfg) == NULL)
	{
		cifs_deleteServer(svp);
		return (NULL);
	}

	//create ADMIN$ share
	unistrcpy(shCfg.name, unistr_adminDol);
	strDos2Uni(shCfg.comment, "Remote Admin", unisizeof(shCfg.comment), TRUE);
	shCfg.pathName[0] = '\0';
	shCfg.flags = CIFS_SHCFG_ADMIN;
	if (cifs_addShare(svp, &shCfg) == NULL)
	{
		cifs_deleteServer(svp);
		return (NULL);
	}
	
	return (svp);
}


/*-----------------------------------------------------------------
 * Name			: cifs_initObjects
 *
 * Description	: Initialize CIFS internal objects
 *
 * Created by	: Laurent Brard
 *
 * Date Created	: 11/09/2001
 *-----------------------------------------------------------------
 */
int32 	//NFX_OK or NFX_ERR
cifs_initObjects(void)
{

	CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_initObjects: start"));
//printf("cifs_initObjects: start\n");
					
	CIFS_RAM_ALLOC(cifs_objectsp, cifs_objLists, sizeof(cifs_objLists));
	NFX_ASSERT(cifs_objectsp != NULL);

	memset(cifs_objectsp, 0, sizeof(cifs_objLists));
	
	nfp_objInit(&cifs_objectsp->serverList, "cifs_server",
					sizeof(cifs_server), CIFS_N_SERVERS);

	nfp_objInit(&cifs_objectsp->shareList, "cifs_share", 
					sizeof(cifs_share), CIFS_N_SHARES);
		
	nfp_objInit(&cifs_objectsp->clientList, "cifs_client",
					sizeof(cifs_client), CIFS_N_CLIENTS);
	
	nfp_objInit(&cifs_objectsp->sessionList, "cifs_session",
					sizeof(cifs_session), CIFS_N_SESSIONS);
	
	nfp_objInit(&cifs_objectsp->shareInUseList,"cifs_shareInUse",
					sizeof(cifs_shareInUse), CIFS_N_SHARESINUSE);
	
	nfp_objInit(&cifs_objectsp->operationList, "cifs_operation",
					sizeof(cifs_operation), CIFS_N_OPERATIONS);
	
	nfp_objInit(&cifs_objectsp->pipeList, "cifs_pipe",
					sizeof(cifs_pipe), CIFS_N_PIPES);
	
	nfp_objInit(&cifs_objectsp->fileOpenList, "cifs_fileOpen",
					sizeof(cifs_fileOpen), CIFS_N_FILESOPEN);
	
	nfp_objInit(&cifs_objectsp->searchList, "cifs_search",
					sizeof(cifs_search), CIFS_N_SEARCHES);
	
	CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifs_initObjects: end"));
//printf("cifs_initObjects: end\n");

	return (NFX_OK);
}


