/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * RCS:  $Id: cifs-transac2.c,v 1.2 2002/12/10 18:13:18 laurentb Exp $
 *
 * Name			: cifs-transac2.c
 *
 * Description	: Contains SMB_COM_TRANSACTION2 functions. 
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/30/2002
 *
 *-----------------------------------------------------------------
 */

#include "cifs-api.h"
#include "cifs-utils.h"
#include "cifs-unicode.h"
#include "cifs-transport.h"
#include "cifs-encdec.h"
#include "cifs-string.h"
#include "cifs-err.h"
#include "cifs-msg.h"
#include "cifs-intmsg.h"
#include "cifs-filetools.h"
#include "cifs-filereqs.h"
#include "cifs-transac2.h"
 
#define IFDBG(lvl)  if (lvl <= cifs_debugLevel)

struct cifs_msg_transac2_subcmd
{
	ushort16	subcmd;
	int32		(*procSubcmd)(cifs_operation	*opp,
							  cifs_transac2		*transacp,
							  uchar8			*parp,
							  ushort16			pbsCnt,
							  uchar8			*datp,
							  ushort16			dbsCnt);
} cifs_msg_transac2_subcmd[CIFS_S_TRANS2_SUBCMD] = {
	{CIFS_TRANS2_OPEN2,				cifstrans2_open2},			// 0x00
	{CIFS_TRANS2_FIND_FIRST2,		cifstrans2_findFirst2},		// 0x01
	{CIFS_TRANS2_FIND_NEXT2,		cifstrans2_findNext2},		// 0x02
	{CIFS_TRANS2_QUERY_FS_INFO,		cifstrans2_queryFsInfo},	// 0x03
	{CIFS_TRANS2_SET_FS_INFO,		NULL},						// 0x04
	{CIFS_TRANS2_QUERY_PATH_INFO,	cifstrans2_queryPathInfo},	// 0x05
	{CIFS_TRANS2_SET_PATH_INFO,		cifstrans2_setPathInfo},	// 0x06
	{CIFS_TRANS2_QUERY_FILE_INFO,	cifstrans2_queryFileInfo},	// 0x07
	{CIFS_TRANS2_SET_FILE_INFO,		cifstrans2_setFileInfo},	// 0x08
	{CIFS_TRANS2_FSCTL,				NULL},						// 0x09
	{CIFS_TRANS2_IOCTL2,			NULL},						// 0x0A
	{CIFS_TRANS2_FIND_NOTIFY_FIRST,	NULL},						// 0x0B
	{CIFS_TRANS2_FIND_NOTIFY_NEXT,	NULL},						// 0x0C
	{CIFS_TRANS2_CREATE_DIR,		NULL},						// 0x0D
	{CIFS_TRANS2_SESSION_SETUP,		NULL},						// 0x0E
	{0x0F,							NULL},						// 0x0F
	{CIFS_TRANS2_GET_DFS_REFERRAL,	NULL},						// 0x10
	{CIFS_TRANS2_REPORT_DFS_INCONSISTENCY, NULL}				// 0x11
};

// Size of an entry in response to Trans2_Find_First2 or Trans2_Find_Next2, 
// excluding the length of file name. The size depends on information level.
struct cifstransac2_findrsp_entry_len
{
	short16		infoLevel;
	ushort16	entrySize;
} cifstransac2_findrsp_entry_size[9] = {
	{CIFS_INFOLV_STANDARD,				23},
	{CIFS_INFOLV_QUERY_EA_SIZE,			27},
	{CIFS_INFOLV_QUERY_EAS_FROM_LIST,	27},
	{CIFS_INFOLV_FIND_FILE_DIR,			64},
	{CIFS_INFOLV_FIND_FILE_FULL_DIR,	68},
	{CIFS_INFOLV_FIND_FILE_NAMES,		12},
	{CIFS_INFOLV_FIND_FILE_BOTH_DIR,	94},
	{CIFS_INFOLV_FIND_FILE_UNIX,		108},
	{-1, 0}
};

// Assume average length of a file name is 50 characters.
#define CIFS_TRANS2_FINDRSP_AVERAGE_NAME_LEN	50


//********************* Transaction2 Reply **********************************

/*-----------------------------------------------------------------
 * Name			: cifstrans2_replyInit
 *
 * Description	: Initialize an answer to a transaction2 (0x32) 
 *				  This function should be called after the total
 *				  number of setup words and paremeter bytes to send 
 *				  in response is known.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/30/2002
 *
 *-----------------------------------------------------------------
 */
boolean			// TRUE if no response required
cifstrans2_replyInit(cifs_operation *opp, cifs_transac2 *transacp)
{
	uchar8			*ocmdp = opp->respMsg.cmdp;
	uchar8			*osmbp = opp->respMsg.smbp;
	ushort16		wordCnt, paraPad, dataPad;
	uchar8			*bp;

	if (transacp->flags & CIFS_TRANS2_FLAG_NO_ANSWER)
	{
		cifs_closeOperation(opp);
		return (TRUE);
	}

	//----------------
	//Set word counter
	//----------------
	wordCnt = 10 + transacp->swsCnt;
	cifs_setCmdWordCnt(&opp->respMsg, wordCnt, TRUE);

	//------------------------------------------------------------
	//Align parameter bytes and data bytes on a 32 bits maxBytes
	//TODO: see if we want to go to 64-bit alignment on Sibyte CPU
	//------------------------------------------------------------
	bp = SMBCP_BPAR(ocmdp);				// Start of SMB byte parameters
	paraPad = (address_t)bp % CIFS_ALIGN_32;
	if (paraPad)
		paraPad = CIFS_ALIGN_32 - paraPad;
	transacp->parap = bp + paraPad;

	// Assume that the total number of parameter bytes to send 
	// is always less than the max reply buffer size.

	dataPad = (address_t)(transacp->parap + transacp->rpbsCnt) % CIFS_ALIGN_32;
	if (dataPad)
		dataPad = CIFS_ALIGN_32 - dataPad;
	transacp->datap = transacp->parap + transacp->rpbsCnt + dataPad;

	transacp->swp = (ushort16 *)(ocmdp + SMBCO_WP10);

	return (FALSE);
} /* cifstrans2_replyInit() */


/*-----------------------------------------------------------------
 * Name			: cifstrans2_replyOneMsg
 *
 * Description	: Send one message for Transaction2 response 
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/30/2002
 *
 *-----------------------------------------------------------------
 */
boolean			// TRUE if more data to send
cifstrans2_replyOneMsg(cifs_operation *opp, cifs_transac2 *transacp)
{
	uchar8			*ocmdp = opp->respMsg.cmdp;
	uchar8			*osmbp = opp->respMsg.smbp;
	ushort16		wordCnt, maxReplySize, paraCnt, dataCnt, dataPad, i;
	boolean			moreDataToSend;
	uchar8			*bp, *datap, *oparap, *odatap;
	pkt_desc_t		*opktp;

	//------------------------------------
	//Calculate byte count and set counter
	//------------------------------------

	maxReplySize = transacp->maxReplySize; 

	// Assume that the total number of parameter bytes to send 
	// is always less than the max reply buffer size.
	paraCnt = transacp->rpbsCnt;
	dataCnt = MIN(transacp->rdbsCnt, maxReplySize - paraCnt);

	bp = SMBCP_BPAR(ocmdp);
	oparap = transacp->parap;
	odatap = transacp->datap;

	cifs_setCmdByteCnt(&opp->respMsg, 
					   ((address_t)(odatap-bp))+dataCnt, 
					   FALSE);

	if ((transacp->internalDatap != NULL) && (dataCnt > 0))
	{
		datap = transacp->internalDatap + (transacp->tdbsCnt - transacp->rdbsCnt);
		memcpy(odatap, datap, dataCnt);
	}

	//-------------------
	//Set parameter words
	//-------------------
	PUTLE2(ocmdp, SMBCO_WP0, transacp->tpbsCnt);
	PUTLE2(ocmdp, SMBCO_WP1, transacp->tdbsCnt);
	PUTLE2(ocmdp, SMBCO_WP2, 0);
	PUTLE2(ocmdp, SMBCO_WP3, paraCnt);
	PUTLE2(ocmdp, SMBCO_WP4, SMB_DIFF(oparap, osmbp));
	PUTLE2(ocmdp, SMBCO_WP5, transacp->tpbsCnt - transacp->rpbsCnt);
											// displacement of parameter bytes
	PUTLE2(ocmdp, SMBCO_WP6, dataCnt);
	PUTLE2(ocmdp, SMBCO_WP7, SMB_DIFF(odatap, osmbp));
	PUTLE2(ocmdp, SMBCO_WP8, transacp->tdbsCnt - transacp->rdbsCnt);
											// displacement of data bytes

	// Update opp->transaction
	transacp->rpbsCnt -= paraCnt;
	transacp->rdbsCnt -= dataCnt;

	if (transacp->rdbsCnt == 0)
	{
		moreDataToSend = FALSE;
		transacp->flags |= CIFS_TRANS2_FLAG_LAST_MSG;
	}
	else
	{
		moreDataToSend = TRUE;
		transacp->flags &= ~CIFS_TRANS2_FLAG_LAST_MSG;
		opp->flags = CIFS_OPFLAG_MORE_TO_SEND;
	}

	CIFS_PRINT(CIFS_DBGLV_DBG2, ("cifstrans2_replyOneMsg: "
				"pSend=%d, dSend=%d, moreDat=%d, "
				"replyMax=%d, pCnt=%d, dCnt=%d",
				transacp->tpbsCnt, transacp->tdbsCnt, 
				(moreDataToSend ? 1 : 0),
				maxReplySize, paraCnt, dataCnt));

	cifs_sendClientMsg(opp->clp, opp);

	return (moreDataToSend);
} /* cifstrans2_replyOneMsg() */


/*-----------------------------------------------------------------
 * Name			: cifstrans2_reply
 *
 * Description	: answer to a transaction2 (0x32) 
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/30/2002
 *
 *-----------------------------------------------------------------
 */
void
cifstrans2_reply(cifs_operation *opp, cifs_transac2 *transacp)
{
	pkt_desc_t	opktp;
	boolean		moreDataToSend;
	esm_event_t	evt;

	moreDataToSend = cifstrans2_replyOneMsg(opp, transacp);

	while(moreDataToSend == TRUE)
	{
		//More response messages to send for the smb request
		//Get and init xmt buffer
		if (cifs_initXmtBuffer(opp, NULL, NULL) == NULL)
		{
			//No available buffer
			opp->status = cifs_opstatus_sendPending;
			thread_yield(&evt, CIFS_INTMSG_TIMEOUT_WAITBUF);

			// Treat Kill event
			if (evt.event == CIFS_THREADEVT_KILL)
			{	
				// Deallocate memory
				if (transacp->internalDatap != NULL)
				{
					CIFS_RAM_DEALLOC(transacp->internalDatap);
					transacp->internalDatap = NULL;
				}
				thread_exit();
			}

			if (evt.event == CIFS_THREADEVT_BUF_AVAILABLE)
			{
				 cifs_initXmtBuffer(opp,
								 	(pkt_desc_t *)(evt.p1.u.p),		//pktp
									(uchar8 *)(evt.p2.u.p));		//osmbp
			}
			else
			{
				CIFS_PRINT(CIFS_DBGLV_ERR, ("cifstrans2_reply: "
							"Received unexpected event"));
				continue;	//TODO: Should continue to wait or quit?
			}
		}

		cifstrans2_replyInit(opp, transacp);
		moreDataToSend = cifstrans2_replyOneMsg(opp, transacp);
	}

	return;
} /* cifstrans2_reply() */


//******* Functions for filling file info in response messagge ************

/*-----------------------------------------------------------------
 * Name			: cifst2_fillFileStandard
 *
 * Description	: Fill in SMB INFO_STANDARD structure.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/22/2002
 *
 *-----------------------------------------------------------------
 */
short16			// Number of bytes filled, or -1 if there is error
cifst2_fillFileStandard(uchar8			*msgBuf,
					    fs_Attr			*fsAttrp,
						ushort16		maxBytes)
{
	ushort16	dataCnt = 22;

	if (maxBytes < dataCnt)
		return (-1);

	//Fill CreationDate and CreationTime
	PUTLE4(msgBuf, 0, 0);
	//Fill LastAccessDate and LastAccessTime
	cifs_fs2SmbDateTime(msgBuf + 4, &fsAttrp->atime);
	//Fill LastWriteDate and LastWriteTime
	cifs_fs2SmbDateTime(msgBuf + 8, &fsAttrp->mtime);

	//Fill file size
	PUTLE4(msgBuf, 12, (uint32)fsAttrp->size);
	//Fill filesystem allocation unit size 
	//(TODO - what is the meaning of this?)
	PUTLE4(msgBuf, 16, (uint32)(fsAttrp->blockSize * fsAttrp->blocks));
	//Fill SMB file attributes
	cifs_fillSmbFileAttr(msgBuf + 20, fsAttrp, FALSE);	//2 bytes

	return (dataCnt);
} /* cifst2_fillFileStandard() */


/*-----------------------------------------------------------------
 * Name			: cifst2_fillFindStandard
 *
 * Description	: Fill in SMB INFOSTANDARD structure
 *				  in TRANS2 FIND FIRST2 or FIND NEXT2 response
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/28/2002
 *
 *-----------------------------------------------------------------
 */
short16			// Number of bytes filled, or -1 if there is error
cifst2_fillFindStandard(cifs_operation	*opp,
						uchar8			*msgBuf,
						fs_Attr			*fsAttrp,
						uni16			*fileName,
						uchar8			**namePos,
						ushort16		maxBytes)
{
	short16		dataCnt, nameDataCnt;
	int32		nameLen;

	if (maxBytes < 23 + unistrsize(fileName) + 1)
	{
		//This entry's size is bigger than available buffer size
		//For simplicity, we assume the fileName is in UNICODE.
		//We add "1" because UNICODE is 2-byte aligned.
		return (-1);
	}

	//Fill STANDARD file info
	dataCnt = cifst2_fillFileStandard(msgBuf, fsAttrp, maxBytes);

	msgBuf += dataCnt;		//dataCnt == 22

	//Fill FileName
	nameDataCnt = cifs_putUniStr(msgBuf + 1, fileName, opp->respMsg.smbp, 
								  &nameLen, maxBytes - 23);
	//Fill fileNameLength
	PCAR(msgBuf, 0, nameLen);

	msgBuf++;

	*namePos = msgBuf + (nameDataCnt - nameLen); //(nameDataCnt-nameLen) is pad

	return(23 + nameDataCnt);
} /* cifst2_fillFindStandard() */

/*-----------------------------------------------------------------
 * Name			: cifst2_fillFindFileCommon
 *
 * Description	: Fill in some common SMB fields for different info level
 *				  in TRANS2 FIND FIRST2 or FIND NEXT2 response
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/28/2002
 *
 *-----------------------------------------------------------------
 */
short16			// Number of bytes filled
cifst2_fillFindFileCommon(cifs_transac2		*transacp,
						  uchar8			*msgBuf,
						  fs_Attr			*fsAttrp)
{
	//Fill FileIndex
	PUTLE4(msgBuf, 0, transacp->searchp->cookie);
	//Fill CreationTime
	PUTLE8(msgBuf, 4, 0);
	//Fill LastAccessTime
	cifs_fillLongTime(msgBuf + 12, &fsAttrp->atime);
	//Fill LastWriteTime
	cifs_fillLongTime(msgBuf + 20, &fsAttrp->mtime);
	//Fill ChangeTime
	cifs_fillLongTime(msgBuf + 28, &fsAttrp->ctime);
	//Fill EndOfFile
	PUTLE8(msgBuf, 36, (uint64)(fsAttrp->size));
	//Fill allocationSize
	PUTLE8(msgBuf, 44, (uint64)(fsAttrp->blockSize * fsAttrp->blocks)); //TODO
	//Fill SMB extFileAttributes
	cifs_fillSmbFileAttr(msgBuf + 52, fsAttrp, TRUE);	//4 bytes
	
	return (56);
} /* cifst2_fillFindFileCommon */

/*-----------------------------------------------------------------
 * Name			: cifst2_fillFindFileDir
 *
 * Description	: Fill in SMB FIND_FILE_DIRECTORY_INFO structure
 *				  in TRANS2 FIND FIRST2 or FIND NEXT2 response
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/28/2002
 *
 *-----------------------------------------------------------------
 */
short16			// Number of bytes filled, or -1 if there is error
cifst2_fillFindFileDir(cifs_operation	*opp,
						cifs_transac2	*transacp,
						uchar8			*msgBuf,
						fs_Attr			*fsAttrp,
						uni16			*fileName,
						uchar8			**namePos,
						ushort16		maxBytes)
{
	uchar8		*startPos = msgBuf;
	int32		dataCnt, nameDataCnt, nameLen;

	if (maxBytes < 64 + unistrsize(fileName) + 1)
	{
		//This entry's size is bigger than available buffer size
		return (-1);
	}

	msgBuf += 4;

	//Fill common fields in find first/next response
	dataCnt = cifst2_fillFindFileCommon(transacp, msgBuf, fsAttrp);
	
	msgBuf += dataCnt;		//msgBuf = startPos + 60

	//Fill FileName
	nameDataCnt = cifs_putUniStr(msgBuf + 4, fileName, opp->respMsg.smbp,
								  &nameLen, maxBytes - 64);
	//Fill fileNameLength
	PUTLE4(msgBuf, 0, nameLen);

	msgBuf += 4;			//msgBuf = startPos + 64

	*namePos = msgBuf + (nameDataCnt - nameLen);

	//Fill NextEntryOffset
	PUTLE4(startPos, 0, msgBuf + nameDataCnt - startPos);

	return(64 + nameDataCnt);
} /* cifst2_fillFindFileDir() */

/*-----------------------------------------------------------------
 * Name			: cifst2_fillFindFileNames
 *
 * Description	: Fill in SMB FIND_FILE_NAMES_INFO structure
 *				  in TRANS2 FIND FIRST2 or FIND NEXT2 response
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/28/2002
 *
 *-----------------------------------------------------------------
 */
short16			// Number of bytes filled, or -1 if there is error
cifst2_fillFindFileNames(cifs_operation	*opp,
						cifs_transac2	*transacp,
						uchar8			*msgBuf,
						uni16			*fileName,
						uchar8			**namePos,
						ushort16		maxBytes)
{
	uchar8		*startPos = msgBuf;
	int32		dataCnt, nameDataCnt, nameLen;

	if (maxBytes < 12 + unistrsize(fileName) + 1)
	{
		//This entry's size is bigger than available buffer size
		return (-1);
	}

	msgBuf += 4;

	//Fill FileIndex
	PUTLE4(msgBuf, 0, transacp->searchp->cookie);

	msgBuf += 4;

	//Fill FileName
	nameDataCnt = cifs_putUniStr(msgBuf + 4, fileName, opp->respMsg.smbp,
								  &nameLen, maxBytes - 12);
	//Fill fileNameLength
	PUTLE4(msgBuf, 0, nameLen);

	msgBuf += 4;			//msgBuf = startPos + 12

	*namePos = msgBuf + (nameDataCnt - nameLen);

	//Fill NextEntryOffset
	PUTLE4(startPos, 0, msgBuf + nameDataCnt - startPos);

	return(12 + nameDataCnt);
} /* cifst2_fillFindFileNames() */

/*-----------------------------------------------------------------
 * Name			: cifst2_fillFindFileFullDir
 *
 * Description	: Fill in SMB FIND_FILE_FULL_DIRECTORY_INFO structure
 *				  in TRANS2 FIND FIRST2 or FIND NEXT2 response
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/28/2002
 *
 *-----------------------------------------------------------------
 */
short16			// Number of bytes filled, or -1 if there is error
cifst2_fillFindFileFullDir(cifs_operation	*opp,
							cifs_transac2	*transacp,
							uchar8			*msgBuf,
							fs_Attr			*fsAttrp,
							uni16			*fileName,
							uchar8			**namePos,
							ushort16		maxBytes)
{
	uchar8		*startPos = msgBuf, *nameLenPos;
	int32		dataCnt, nameDataCnt, nameLen;

	if (maxBytes < 68 + unistrsize(fileName) + 1)
	{
		//This entry's size is bigger than available buffer size
		return (-1);
	}

	msgBuf += 4;

	//Fill common fields in find first/next response
	dataCnt = cifst2_fillFindFileCommon(transacp, msgBuf + 4, fsAttrp);
	
	msgBuf += dataCnt;		//msgBuf = startPos + 60

	//Remember fileNameLength position
	nameLenPos = msgBuf;

	//Fill EaSize (Do not support)
	PUTLE4(msgBuf, 4, 0);

	msgBuf += 8;			//msgBuf = startPos + 68

	//Fill FileName
	nameDataCnt = cifs_putUniStr(msgBuf, fileName, opp->respMsg.smbp, 
								  &nameLen, maxBytes - 68);

	//Fill fileNameLength
	PUTLE4(nameLenPos, 0, nameLen);

	*namePos = msgBuf + (nameDataCnt - nameLen);

	//Fill NextEntryOffset
	PUTLE4(startPos, 0, msgBuf + nameDataCnt - startPos);

	return (68 + nameDataCnt);
} /* cifst2_fillFindFileFullDir() */

/*-----------------------------------------------------------------
 * Name			: cifst2_fillFindFileBothDir
 *
 * Description	: Fill in SMB FIND_FILE_BOTH_DIRECTORY_INFO structure
 *				  in TRANS2 FIND FIRST2 or FIND NEXT2 response
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/28/2002
 *
 *-----------------------------------------------------------------
 */
short16			// Number of bytes filled, or -1 if there is error
cifst2_fillFindFileBothDir(cifs_operation	*opp,
						  cifs_transac2		*transacp,
						  uchar8			*msgBuf,
						  fs_Attr			*fsAttrp,
						  uni16				*fileName,
						  uchar8			**namePos,
						  ushort16			maxBytes)
{
	uchar8		*startPos = msgBuf, *nameLenPos;
	int32		dataCnt, nameDataCnt, nameLen;

	if (maxBytes < 94 + unistrsize(fileName) + 1)
	{
		//This entry's size is bigger than available buffer size
		return (-1);
	}

	msgBuf += 4;

	//Fill common fields in find first/next response
	dataCnt = cifst2_fillFindFileCommon(transacp, msgBuf + 4, fsAttrp);
	
	msgBuf += dataCnt;		//msgBuf = startPos + 60

	//Remember fileNameLength position
	nameLenPos = msgBuf;

	//Fill EaSize (Do not support)
	PUTLE4(msgBuf, 4, 0);
	//Fill ShortNameLength (Do not support)
	PCAR(msgBuf, 8, 0);
	PCAR(msgBuf, 9, 0);
	//Fill ShortName (Do not support)
	nameLen = cifs_putShortName(msgBuf + 10, fileName, 12*2);

	msgBuf += 34;			//msgBuf = startPos + 94
	
	//Fill FileName
	nameDataCnt = cifs_putUniStr(msgBuf, fileName, opp->respMsg.smbp, 
								  &nameLen, maxBytes - 94);

	//Fill fileNameLength
	PUTLE4(nameLenPos, 0, nameLen);

	*namePos = msgBuf + (nameDataCnt - nameLen);

	//Fill NextEntryOffset
	PUTLE4(startPos, 0, msgBuf + nameDataCnt - startPos);

	return (94 + nameDataCnt);
} /* cifst2_fillFindFileBothDir() */


/*-----------------------------------------------------------------
 * Name			: cifst2_fillFileQueryFileBasic
 *
 * Description	: Fill in SMB QUERY_FILE_BASIC_INFO structure in 
 *				  Transaction2 Query Path/File Information response.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/21/2002
 *
 *-----------------------------------------------------------------
 */
short16			// Number of bytes filled, or -1 if there is error
cifst2_fillFileQueryFileBasic(cifs_operation	*opp,
							  uchar8			*msgBuf,
							  fs_Attr			*fsAttrp,
							  ushort16			maxBytes)
{
	int32	dataCnt = 36;

	if (maxBytes < dataCnt)
		return (-1);

	//Fill CreationTime
	PUTLE8(msgBuf, 4, 0);
	//Fill LastAccessTime
	cifs_fillLongTime(msgBuf + 8, &fsAttrp->atime);
	//Fill LastWriteTime
	cifs_fillLongTime(msgBuf + 16, &fsAttrp->mtime);
	//Fill ChangeTime
	cifs_fillLongTime(msgBuf + 24, &fsAttrp->ctime);

	//Fill SMB file attributes
	cifs_fillSmbFileAttr(msgBuf + 32, fsAttrp, TRUE);	//4 bytes

	return (dataCnt);
} /* cifst2_fillFileQueryFileBasic() */

/*-----------------------------------------------------------------
 * Name			: cifst2_fillFileQueryFileStandard 
 *
 * Description	: Fill in SMB QUERY_FILE_STANDARD_INFO structure in 
 *				  Transaction2 Query Path/File Information response.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/25/2002
 *
 *-----------------------------------------------------------------
 */
int32			// Number of bytes filled, or -1 if there is error
cifst2_fillFileQueryFileStandard(cifs_operation	*opp,
								 uchar8			*msgBuf,
								 fs_Attr		*fsAttrp,
								 ushort16		maxBytes)
{
	int32	dataCnt = 24;	//Only 22 bytes are filled.
							//Trace shows data bytes count filed is 24.
							//Samba also set 24 here.

	if (maxBytes < dataCnt)
		return (-1);

	//Fill allocationSize
	PUTLE8(msgBuf, 0, (uint64)(fsAttrp->blockSize * fsAttrp->blocks));
	//Fill EndOfFile
	PUTLE8(msgBuf, 8, (uint64)(fsAttrp->size));
	//Fill NumberOfLinks
	PUTLE4(msgBuf, 16, fsAttrp->numLinks);
	//Fill DeletePending
	//TODO:
	PCAR(msgBuf, 20, 0);
	//Fill Directory
	PCAR(msgBuf, 21, (fsAttrp->type == FS_FTYPE_DIR)?1:0);
	
	return (dataCnt);
} /* cifst2_fillFileQueryFileStandard() */

/*-----------------------------------------------------------------
 * Name			: cifst2_fillFileQueryFileAll
 *
 * Description	: Fill in SMB QUERY_FILE_ALL_INFO structure in 
 *				  Transaction2 Query Path/File Information response.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/25/2002
 *
 *-----------------------------------------------------------------
 */
int32			// Number of bytes filled, or -1 if there is error
cifst2_fillFileQueryFileAll(cifs_operation	*opp,
							uchar8			*msgBuf,
							fs_Attr			*fsAttrp,
							/*file attributes from fssf*/
							uni16			*pathName,
							ushort16		maxBytes)
{
	uchar8	*startPos = msgBuf;
	int32	dataCnt, nameLength, nameDataCnt;

	if (maxBytes < 44)
		return (-1);

	//Fill CreationTime
	PUTLE8(msgBuf, 4, 0);
	//Fill LastAccessTime
	cifs_fillLongTime(msgBuf + 8, &fsAttrp->atime);
	//Fill LastWriteTime
	cifs_fillLongTime(msgBuf + 16, &fsAttrp->mtime);
	//Fill ChangeTime
	cifs_fillLongTime(msgBuf + 24, &fsAttrp->ctime);
	//Fill SMB file attributes
	cifs_fillSmbFileAttr(msgBuf + 32, fsAttrp, FALSE);	//2 bytes

	msgBuf += 34;	//Samba set it to be 40. Does it require 64-bit alignment? 

	//Fill allocationSize
	PUTLE8(msgBuf, 0, (uint64)(fsAttrp->blockSize * fsAttrp->blocks));
	//Fill EndOfFile
	PUTLE8(msgBuf, 8, (uint64)(fsAttrp->size));
	//Fill NumberOfLinks
	PUTLE4(msgBuf, 16, fsAttrp->numLinks);
	//Fill DeletePending
	//TODO:
	PCAR(msgBuf, 20, 0);
	//Fill Directory
	PCAR(msgBuf, 21, (fsAttrp->type == FS_FTYPE_DIR));
	
	msgBuf += 22;	//Samba set it to be 24. Why?

	//Fill IndexNumber (Here it is inode)
	PUTLE8(msgBuf, 0, fsAttrp->inode);
	//Fill EASize (No EA info)
	PUTLE4(msgBuf, 8, 0);
	//Fill AccessFlags
	cifs_getUserAccessToFile(opp, (msgBuf + 12), fsAttrp);	//4 bytes
	//Fill IndexNumber1
	PUTLE8(msgBuf, 16, 0);
	//Fill CurrentByteOffset
	//TODO:
	PUTLE8(msgBuf, 24, 0);
	//Fill crrent open Mode
	//TODO
	PUTLE4(msgBuf, 32, 0);
	//Fill AlignmentRequirement
	//TODO
	PUTLE4(msgBuf, 36, 0);
	
	msgBuf += 40;

	//Fill FileNameLength and FileName
	nameDataCnt = cifs_putUniStr(msgBuf + 4,
								pathName,
								opp->respMsg.smbp,
								&nameLength,
								maxBytes - 44);
	if (nameDataCnt < 0)
		return (-1);

	PUTLE4(msgBuf, 0, nameLength);
	msgBuf += 4 + nameDataCnt;

	dataCnt = SMB_DIFF(msgBuf, startPos);

	return (dataCnt);
} /* cifst2_fillFileQueryFileAll() */


//******* Functions for extracting file info from response messagge ********

/*-----------------------------------------------------------------
 * Name			: cifst2_getFileStandard
 *
 * Description	: Get SMB INFO_STANDARD from Transaction2 Set Path/File
 *				  Information request, and set fileds in fs_Attr.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 04/06/2002
 *
 *-----------------------------------------------------------------
 */
int32			//CIFSTATUS
cifst2_getFileStandard(cifs_operation	*opp,
						fs_Attr			*fsAttrp,
						uchar8			*msgBuf) // File info in SMB message
{
	//TODO

	return (CIFSTATUS_OK);
} /* cifst2_getFileStandard() */

/*-----------------------------------------------------------------
 * Name			: cifst2_getFileBasic
 *
 * Description	: Get SMB FILE_BASIC_INFO from Transaction2 Set Path/File
 *				  Information request, and set fileds in fs_Attr.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 04/06/2002
 *
 *-----------------------------------------------------------------
 */
int32			//CIFSTATUS
cifst2_getFileBasic(cifs_operation	*opp,
					fs_Attr			*fsAttrp,
					uchar8			*msgBuf) // File info in SMB message
{
	//TODO

	return (CIFSTATUS_OK);
} /* cifst2_getFileBasic() */

/*-----------------------------------------------------------------
 * Name			: cifst2_getFileDisposition
 *
 * Description	: Get SMB FILE_DISPOSITION_INFO from Transaction2 Set Path/File
 *				  Information request, and set fileds in fs_Attr.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 04/06/2002
 *
 *-----------------------------------------------------------------
 */
int32			//CIFSTATUS
cifst2_getFileDisposition(cifs_operation	*opp,
						  fs_Attr			*fsAttrp,
						  uchar8			*msgBuf) // File info in SMB message
{
	//TODO

	return (CIFSTATUS_OK);
} /* cifst2_getFileDisposition() */

/*-----------------------------------------------------------------
 * Name			: cifst2_getFileAllocation 
 *
 * Description	: Get SMB FILE_ALLOCATION_INFO from Transaction2 Set Path/File
 *				  Information request, and set fileds in fs_Attr.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 04/06/2002
 *
 *-----------------------------------------------------------------
 */
int32			//CIFSTATUS
cifst2_getFileAllocation(cifs_operation	*opp,
						fs_Attr			*fsAttrp,
						uchar8			*msgBuf) // File info in SMB message
{
	//TODO

	return (CIFSTATUS_OK);
} /* cifst2_getFileAllocation() */

/*-----------------------------------------------------------------
 * Name			: cifst2_getFileEndOfFile
 *
 * Description	: Get SMB FILE_END_OF_FILE_INFO from Transaction2 Set Path/File
 *				  Information request, and set fileds in fs_Attr.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 04/06/2002
 *
 *-----------------------------------------------------------------
 */
int32			//CIFSTATUS
cifst2_getFileEndOfFile(cifs_operation	*opp,
						fs_Attr			*fsAttrp,
						uchar8			*msgBuf) // File info in SMB message
{
	//TODO

	return (CIFSTATUS_OK);
} /* cifst2_getFileEndOfFile() */



//******************** Transaction2 Open2 **********************************

/*-----------------------------------------------------------------
 * Name			: cifstrans2_open2
 *
 * Description	: TRANS2 OPEN2
 * 				  Create file with extended attrributtes.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/31/2002
 *
 *-----------------------------------------------------------------
 */
int32			// CIFS_SENDRESPNOW or CIFS_DONTSENDRESP
cifstrans2_open2(cifs_operation	*opp,
				cifs_transac2	*transacp,
				uchar8			*parp,
				ushort16		pbsCnt,
				uchar8			*datp,
				ushort16		dbsCnt)
{
	//TODO

	return (CIFS_SENDRESPNOW);
} /* cifstrans2_open2() */


/*-----------------------------------------------------------------
 * Name			: cifstrans2_replyOpen2
 *
 * Description	: Reply to TRANS2 OPEN2
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/05/2002
 *
 *-----------------------------------------------------------------
 */
int32			//CIFSTATUS 
cifstrans2_replyOpen2(cifs_operation	*opp,
					  cifs_transac2		*transacp	/* ... */)
{
	//TODO

	return (CIFSTATUS_OK);
} /* cifstrans2_replyOpen2() */


//************** Transaction2 Find First2/Next2 ****************************

/*-----------------------------------------------------------------
 * Name			: cifstrans2_findFirst2
 *
 * Description	: TRANS2 FIND FIRST2
 * 				  Begin search for files.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/31/2002
 *
 *-----------------------------------------------------------------
 */
int32			// CIFS_SENDRESPNOW or CIFS_DONTSENDRESP
cifstrans2_findFirst2(cifs_operation	*opp,
					  cifs_transac2		*transacp,
					  uchar8			*parp,
					  ushort16			pbsCnt,
					  uchar8			*datp,
					  ushort16			dbsCnt)
{
	int32					rc = CIFSTATUS_OK;
	ushort16				searchAttr, merCnt, infoLevel, flags;
//	boolean					closeAftThis, closeIfEnd, retResumeKey;
//	boolean					searchFmPrevEnd, bkupIntent;
//	uint32					searchStorageType;
	uchar8					*bp;

	cifs_search				*searchp;
	boolean					searchDir;	// TRUE if search the whole directory
										// FALSE if look at only one file
	fs_AttrType				attrType;
	eee_desc_t				*MCLeDescp=NULL; // MULTI COMPONENT LOOKUP response
	eee_desc_t				*RDPeDescp=NULL; // READDIRPLUS response
	fs_MsgMclResp			*mclResp;
	fs_MsgReadDirPlusResp	*readdirplusResp;
	fs_Attr					*fsAttrp;
	boolean					eos;		// End of search reached or not
	

	//Check share access
	rc = cifs_secCheckShareAccess(opp->siup, SEC_FILE_READ_ATTRIBUTES);
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);


	searchAttr		= GETLE2(parp, 0);	// Search attributes
	merCnt			= GETLE2(parp, 2);	// Maximum number of entries to return
	flags			= GETLE2(parp, 4);	// Additional infomation
//	closeAftThis	= SMB_TESTBIT2(parp, 4, CIFS_FIND_CLOSE_AFTER_THIS);
//	closeIfEnd		= SMB_TESTBIT2(parp, 4, CIFS_FIND_CLOSE_IF_END);
//	retResumeKey	= SMB_TESTBIT2(parp, 4, CIFS_FIND_RET_RESUME_KEY);
//	searchFmPrevEnd	= SMB_TESTBIT2(parp, 4, CIFS_FIND_FROM_PREV_END);
//	bkupIntent		= SMB_TESTBIT2(parp, 4, CIFS_FIND_BACKUP_INTENT);
	infoLevel		= GETLE2(parp, 6);

	CIFS_PRINT(CIFS_DBGLV_DBG, ("cifstrans2_findFirst2: "
				"searchAttr = 0x%x, merCnt = 0x%x, flags = 0x%x, "
				"InfoLevel = 0x%x,", searchAttr, merCnt, flags, infoLevel));

	transacp->flags |= CIFS_TRANS2_FIND_FIRST;

	//--------------------
	// Validate info level
	//--------------------
	switch(infoLevel)
	{
		case CIFS_INFOLV_STANDARD:
		case CIFS_INFOLV_FIND_FILE_DIR:
		case CIFS_INFOLV_FIND_FILE_FULL_DIR:
		case CIFS_INFOLV_FIND_FILE_NAMES:
		case CIFS_INFOLV_FIND_FILE_BOTH_DIR:
			break;

		case CIFS_INFOLV_QUERY_EA_SIZE:
		case CIFS_INFOLV_QUERY_EAS_FROM_LIST:
		case CIFS_INFOLV_FIND_FILE_UNIX:
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifstrans2_findFirst2: "
						"Information level %d unsupported", infoLevel));
			cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_NOT_SUPPORTED);
			break;

		default:
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifstrans2_findFirst2: "
						"Unknown information level %d", infoLevel));
			cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_INVALID_LEVEL);
	}
	transacp->infoLevel = infoLevel;

	//----------------------------------------------------
	// Create a new search for the associated share in use
	//----------------------------------------------------
	searchp = cifs_addSearch(opp);
	if (searchp == NULL)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifstrans2_findFirst2: "
					"no memory for new search"));
		cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_INSUFF_SERVER_RESOURCES);
	}

	searchp->count = merCnt;
	searchp->flags = flags;
	searchp->attributes = searchAttr;
	searchp->infoLevel = infoLevel;
	searchp->cookieVerif = 0;
	searchp->cookie = 0;

	transacp->searchp = searchp;

	// Start of path name in source SMB resquest message
	bp = parp + 12;

	//-----------------------------------------------------------
	// Send MULTI COMPONENT LOOKUP request to FS to get file info
	//-----------------------------------------------------------
	attrType = FS_POST_ATTRS;
	rc = cifs_fsMCLReqFromMsg(opp,
							  NULL,
							  opp->reqMsg.smbp,
							  bp,
							  searchp->pattern,	//obtain search pattern
							  FS_POST_ATTRS);
	if (rc != CIFSTATUS_OK)
	{
		cifs_errResp(opp, 0, 0, rc);
		goto cifstrans2_findFirst2_done;
	}

	if (cifs_uniContainsWildcard(searchp->pattern) == TRUE)
	{
		// To search the whole directory
		searchDir = TRUE;
	}
	else
	{
		// To look at one file
		searchDir = FALSE;
	}

	//-------------------------------------------------
	// Wait for MULTI COMPONENT LOOKUP response from FS
	//-------------------------------------------------
	rc = cifs_waitFsResp(FS_REQ_MCL, (void **)&mclResp, &MCLeDescp);
	if (rc != CIFSTATUS_OK)
	{
		rc = CIFSTATUS_UNSUCCESSFUL;
		cifs_errResp(opp, 0, 0, rc);
		goto cifstrans2_findFirst2_done;
	}

	if (mclResp->hdr.fsStatus == FS_ERR_NONE)
	{
		// File exists. Got file attribute.	
		fs_Attr *fsAttrp = &(mclResp->attr[0]);
		rc = cifstrans2_replyFind(opp,
								  transacp,
								  fsAttrp,		// Attribute of file
								  NULL,
								  NULL,
								  &eos);
		if (rc != CIFSTATUS_OK)
		{
			cifs_errResp(opp, 0, 0, rc);
			goto cifstrans2_findFirst2_done;
		}
	}
	else if ((mclResp->hdr.fsStatus == FS_ERR_NOENT) && (searchDir == TRUE))
	{
		// Directory exists. Will search the whole directory.
		memcpy(&searchp->parentHandle, &mclResp->handle, sizeof(fs_Handle));
		//-------------------------------
		// Send READDRIPLUS request to FS
		//-------------------------------
		//TODO - need to add new params for searchAttr and searchPattern
		rc = cifs_fsReaddirplusReq(opp,
									searchp,
									FS_POST_ATTRS,
									transacp->maxDataCnt);
		if (rc != CIFSTATUS_OK)
		{
			cifs_errResp(opp, 0, 0, rc);
			goto cifstrans2_findFirst2_done;
		}
	}
	else
	{
		rc = CIFSTATUS_UNSUCCESSFUL;
		cifs_returnFsError(opp, mclResp->hdr.fsStatus);
		goto cifstrans2_findFirst2_done;
	}

	if (searchDir == FALSE)
	{
		// Done
		goto cifstrans2_findFirst2_done;
	}

	//--------------------------------------
	// Wait for READDIRPLUS response from FS
	//--------------------------------------
	rc = cifs_waitFsResp(FS_REQ_READDIRPLUS,
						 (void **)&readdirplusResp,
						 &RDPeDescp);
	if (rc != CIFSTATUS_OK)
	{
		rc = CIFSTATUS_UNSUCCESSFUL;
		cifs_errResp(opp, 0, 0, rc);
		goto cifstrans2_findFirst2_done;
	}

	if (readdirplusResp->hdr.fsStatus == FS_ERR_NONE)
	{
		// READDIRPLUS succeeded
		rc = cifstrans2_replyFind(opp,
								  transacp,
								  NULL,
								  readdirplusResp,
								  NULL,
								  &eos);
		if (rc != CIFSTATUS_OK)
		{
			cifs_errResp(opp, 0, 0, rc);
			goto cifstrans2_findFirst2_done;
		}
	}
	else
	{
		rc = CIFSTATUS_UNSUCCESSFUL;
		cifs_returnFsError(opp, readdirplusResp->hdr.fsStatus);
	}

cifstrans2_findFirst2_done:

	// Free MULTI COMPONENT LOOKUP response
	if (MCLeDescp)
		eee_freePacket(MCLeDescp);
	// Free READDIRPLUS response
	if (RDPeDescp)
		eee_freePacket(RDPeDescp);

	// Close search if close search flags are set, or if there are errors
	if ((flags & CIFS_FIND_CLOSE_AFTER_THIS) ||
		((flags & CIFS_FIND_CLOSE_IF_END) && eos) ||
		(rc != CIFSTATUS_OK))
	{
		cifs_closeSearchByPtr(opp->siup, searchp); 
	}

	if (rc != CIFSTATUS_OK)
		return (CIFS_SENDRESPNOW);
	else
		return (CIFS_DONTSENDRESP);
} /* cifstrans2_findFirst2() */


/*-----------------------------------------------------------------
 * Name			: cifstrans2_findNext2
 *
 * Description	: TRANS2 FIND NEXT2
 * 				  Resume search for files.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/31/2002
 *
 *-----------------------------------------------------------------
 */
int32			// CIFS_SENDRESPNOW or CIFS_DONTSENDRESP
cifstrans2_findNext2(cifs_operation	*opp,
					cifs_transac2	*transacp,
					uchar8			*parp,
					ushort16		pbsCnt,
					uchar8			*datp,
					ushort16		dbsCnt)
{
	int32					rc;
	ushort16				sid, merCnt, infoLevel, flags;
	uint32					resumeKey;
	uchar8					*bp;
	uni16					fileName[CIFS_S_FILESTR+1];
	cifs_search				*searchp;
	eee_desc_t				*eDescp = NULL;
	fs_MsgReadDirPlusResp	*readdirplusResp;
	boolean					eos;		// End of search reached or not


	sid			= GETLE2(parp, 0);		// Search handle
	merCnt		= GETLE2(parp, 2);		// Maximum number of entries to return
	infoLevel	= GETLE2(parp, 4);		// Information level
	resumeKey	= GETLE4(parp, 6);		// ResumeKey
	flags		= GETLE2(parp, 10);
	
	CIFS_PRINT(CIFS_DBGLV_DBG, ("cifstrans2_findNext2: "
				"sid = 0x%x, merCnt = 0x%x, flags = 0x%x, " 
				"InfoLevel = 0x%x, resumeKey = 0x%x", 
				sid, merCnt, flags, infoLevel, resumeKey));

	transacp->flags |= CIFS_TRANS2_FIND_NEXT;

	//-------------------------------------------------------------
	// Look for the search data structure associated with previous
	// TRANS2_FIND_FIRST2 request
	//-------------------------------------------------------------
	searchp = cifs_findSearch(opp, sid);
	if (searchp == NULL)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifstrans2_findNext2: "
					"Search handle doesn't match the previous find first request"));
		cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_UNSUCCESSFUL);
	}
	if (searchp->infoLevel != infoLevel)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifstrans2_findNext2: "
					"Information level doesn't match "
					"the previous find first request"));
		cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_UNSUCCESSFUL);
	}
	transacp->infoLevel = infoLevel;

	// Update the search data structure
	searchp->count = merCnt;
	searchp->flags = flags;

	transacp->searchp = searchp;

	//-------------------------------
	// Determine how to do the search
	//---------------------------------
	bp = parp + 12;
	cifs_getSmbUniStr(fileName, bp, unisizeof(fileName), opp->reqMsg.smbp);

	IFDBG(CIFS_DBGLV_DBG2)
	{
		printf("cifstrans2_findNext2: searchp->lastFileName = \n");
		memDump(fileName, unistrsize(searchp->lastFileName));

		printf("cifstrans2_findNext2: fileName = \n");
		memDump(fileName, unistrsize(fileName));
	}

	if ((flags & CIFS_FIND_FROM_PREV_END) ||
		(unistrcmp(fileName, searchp->lastFileName) == 0))
	{
		// Flag bit set for continuing search from previous ending place.
		// Or, if the resume file name in request match the file name
		// of last entry in previous response.
	
		// Continue search from previous ending place
		// Don't care about resume file name
		fileName[0] = 0;
	}
	else
	{
		// Start from the beginning to search the directory
		searchp->cookie = 0;
	}

	//-------------------------------
	// Send READDRIPLUS request to FS
	//-------------------------------
	//TODO - need to add new params for searchAttr and searchPattern
	rc = cifs_fsReaddirplusReq(opp,
								searchp,
								FS_POST_ATTRS,
								transacp->maxDataCnt);
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);

	//--------------------------------------
	// Wait for READDIRPLUS response from FS
	//--------------------------------------
	rc = cifs_waitFsResp(FS_REQ_READDIRPLUS,
						(void **)&readdirplusResp,
						&eDescp);
	if (rc != CIFSTATUS_OK)
	{
		cifs_errResp(opp, 0, 0, CIFSTATUS_UNSUCCESSFUL);
		return (CIFS_SENDRESPNOW);
	}

	if (readdirplusResp->hdr.fsStatus == FS_ERR_NONE)
	{
		// READDIRPLUS succeeded
		rc = cifstrans2_replyFind(opp,
								  transacp,
								  NULL,
								  readdirplusResp,
								  fileName,
								  &eos);
		cifsm_testCIFSTATUS_ifErrFreeDescAndReturn(opp, rc, eDescp);
	}
	else
		cifsm_testFsStatus_ifErrReturn(opp, readdirplusResp->hdr.fsStatus, eDescp);

	// Free READDIRPLUS response
	if (eDescp)
		eee_freePacket(eDescp);

	// Close search if close search flags are set, or if there are errors
	if ((flags & CIFS_FIND_CLOSE_AFTER_THIS) ||
		((flags & CIFS_FIND_CLOSE_IF_END) && eos))
	{
		cifs_closeSearchByPtr(opp->siup, searchp); 
	}

	return (CIFS_DONTSENDRESP);
} /* cifstrans2_findNext2() */

/*-----------------------------------------------------------------
 * Name			: cifst2_fillSearchEntry
 *
 * Description	: Fill in one search entry in SMB TRANS2 FIND FIRST2
 *				  or FIND NEXT2 response message.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/27/2002
 *
 *-----------------------------------------------------------------
 */
short16		// Number of bytes filled, or -1 if there is error
cifst2_fillSearchEntry(cifs_operation	*opp,
						cifs_transac2	*transacp,
						uchar8			*msgBuf,
						fs_Attr			*fsAttrp,
						uni16			*fileName,
						uint32			resumeKey,
						uchar8			**namePos,
						ushort16		maxBytes)
{
	cifs_search		*searchp = transacp->searchp;
	short16			dataCnt;

	switch(transacp->infoLevel)
	{
	case CIFS_INFOLV_STANDARD:
		dataCnt = cifst2_fillFindStandard(opp,
										  msgBuf,
										  fsAttrp,
										  fileName,
										  namePos,
										  maxBytes);
		break;

	//case CIFS_INFOLV_QUERY_EA_SIZE:
	//case CIFS_INFOLV_QUERY_EAS_FROM_LIST:
	//	break;

	case CIFS_INFOLV_FIND_FILE_DIR:
		dataCnt = cifst2_fillFindFileDir(opp,
										 transacp,
										 msgBuf,
										 fsAttrp,
										 fileName,
										 namePos,
										 maxBytes);
		break;

	case CIFS_INFOLV_FIND_FILE_FULL_DIR:
		dataCnt = cifst2_fillFindFileFullDir(opp,
											transacp,
											msgBuf,
											fsAttrp,
											fileName,
											namePos,
											maxBytes);
		break;

	case CIFS_INFOLV_FIND_FILE_NAMES:
		dataCnt = cifst2_fillFindFileNames(opp,
											transacp,
											msgBuf,
											fileName,
											namePos,
											maxBytes);
		break;

	case CIFS_INFOLV_FIND_FILE_BOTH_DIR:
		dataCnt = cifst2_fillFindFileBothDir(opp,
											transacp,
											msgBuf,
											fsAttrp,
											fileName,
											namePos,
											maxBytes);
		break;

	//case CIFS_INFOLV_FIND_FILE_UNIX:
	//	break;
	}
	return (dataCnt);
} /* cifst2_fillSearchEntry() */

/*-----------------------------------------------------------------
 * Name			: cifst2_calFindEntriesSize
 *
 * Description	: Calculate the estimated entries size in bytes in the 
 *				  Find First/Next response.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/30/2002
 *
 *-----------------------------------------------------------------
 */
ushort16		// Number of bytes of one entry
cifst2_calFindEntriesSize(ushort16	infoLevel,
			   			ushort16	entryCnt,
						boolean		unicode)
{
	uint32	i, entrySize = 0, entrySizeWithName;

	i = 0;
	while(cifstransac2_findrsp_entry_size[i].infoLevel != -1)
	{
		if (cifstransac2_findrsp_entry_size[i].infoLevel == infoLevel)
		{
			entrySize = cifstransac2_findrsp_entry_size[i].entrySize;
			break;
		}
		i++;
	}

	if (unicode)
		entrySizeWithName = entrySize + 
							CIFS_TRANS2_FINDRSP_AVERAGE_NAME_LEN*2;
	else
		entrySizeWithName = entrySize + 
							CIFS_TRANS2_FINDRSP_AVERAGE_NAME_LEN;

	return (entrySizeWithName * entryCnt);
} /* cifst2_calFindEntriesSize() */

/*-----------------------------------------------------------------
 * Name			: cifst2_getMoreFileEntries
 *
 * Description	: Call FS READDIRPLUS and get more file entries
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 03/30/2002
 *
 *-----------------------------------------------------------------
 */
int32			// CIFSATUS
cifst2_getMoreFileEntries(cifs_operation		*opp,
						  cifs_search			*searchp,
						  uint32				maxBytes,
						  eee_desc_t			**eDescp,
						  fs_MsgReadDirPlusResp	**readdirplusResp,
						  boolean				*exitThread)
{
	int32	rc;

	// Send READDRIPLUS request to FS
	rc = cifs_fsReaddirplusReq(opp,
								searchp,
								FS_POST_ATTRS,
								maxBytes);
	if (rc != CIFSTATUS_OK)
		return (rc);

	// Wait for READDIRPLUS response from FS
	rc = cifs_waitFsRespAndKill(FS_REQ_READDIRPLUS,
								(void **)readdirplusResp,
								eDescp);
	if (rc != CIFSTATUS_OK)
	{
		if (rc == CIFS_THREADEVT_KILL)
			*exitThread = TRUE;
		rc = CIFSTATUS_UNSUCCESSFUL;
		return (rc);
	}

	// READDIRPLUS succeeded
	if ((*readdirplusResp)->hdr.fsStatus == FS_ERR_NONE)
	{
		rc = CIFSTATUS_OK;
	}
	else
	{
		uchar8 class = 0;
    	ushort16 err = 0;

		rc = cifs_fsToCifsErrMapping((*readdirplusResp)->hdr.fsStatus, 
									 &class, 
									 &err);
		eee_freePacket(*eDescp);
	}
	return (rc);
} /* cifst2_getMoreFileEntries() */

/*-----------------------------------------------------------------
 * Name			: cifst2_cifst2_fillSearchDir
 *
 * Description	: Fill in search entres in SMB TRANS2 FIND FIRST2
 *				  or FIND NEXT2 response message.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 04/02/2002
 *
 *-----------------------------------------------------------------
 */
short16 		// Number of bytes of data filled, or -1 if there is error
cifst2_fillSearchDir(cifs_operation			*opp,
					 cifs_transac2			*transacp,
					 uchar8					*msgBuf,
					 fs_MsgReadDirPlusResp	*dirInfop,
					 uni16					*resumeFileName,
					 ushort16				entryCntTotal,
					 ushort16				maxBytes,
					 ushort16				*entryCnt,
					 uchar8					**lastNamePos,
					 boolean				*eof)
{
	int32			rc;
	cifs_search		*searchp = transacp->searchp;
	boolean			checkResumeFileName;
	ushort16		dataCnt;
	short16			currDataCnt;
	boolean         outOfSpace, fsEof, exitThread;
	ushort16		entryCntThis;
	fs_DirEntryPlus	*direntPlusp;
	fs_Attr			*fsAttrp;
	uni16			fileName[CIFS_S_FILESTR + 1];
	uint32			fileNameLen;
	uint32			resumeKey;
	eee_desc_t      *eDescp;

	IFDBG(CIFS_DBGLV_DBG2)
	{
		if (resumeFileName)
		{
			printf("cifst2_fillSearchDir: resumeFileName = \n");
			memDump(fileName, unistrsize(resumeFileName));
		}
	}

	if (resumeFileName && unistrlen(resumeFileName))
		checkResumeFileName = TRUE;
	else
		checkResumeFileName = FALSE;

	fsEof = *eof;
	outOfSpace = FALSE;
	eDescp = NULL;

	dataCnt = 0;
	*entryCnt = 0;
	*lastNamePos = msgBuf;

	while(*entryCnt < entryCntTotal)
	{
		direntPlusp = &dirInfop->dirEntry[0];
		entryCntThis = 0;
		while((entryCntThis < dirInfop->numEntriesThis) &&
			  (*entryCnt < entryCntTotal))
		{
			fsAttrp = &direntPlusp->attr;
			strUtf82Uni(fileName, direntPlusp->dirEntry.name,
						direntPlusp->dirEntry.nameLen, &fileNameLen,
						CIFS_S_FILESTR*2, TRUE);

			IFDBG(CIFS_DBGLV_DBG2)
			{
				printf("cifst2_fillSearchDir: fileName = \n");
				memDump(fileName, unistrsize(fileName));
			}

			if (checkResumeFileName == TRUE)
			{
				// If we haven't found the resume file, skip this entry.
				// Otherwise, we start to fill entries after this entry.
				if (unistrcmp(resumeFileName, fileName) == 0)
					checkResumeFileName = FALSE;
				goto cifst2_fillSearchDir_continue;
			}

			//TODO: Do some filtering based on search pattern?
			//      Such as, >>>>>>>>">>> filters out "." and ".."

			resumeKey = direntPlusp->dirEntry.cookie;

			// Fill in one entry
			currDataCnt = cifst2_fillSearchEntry(opp,
												transacp,
												msgBuf + dataCnt,
												fsAttrp,
												fileName,
												resumeKey,
												lastNamePos,
												maxBytes - dataCnt);
			if (currDataCnt < 0)
			{	
				// Reach max data bytes to return
				// Stop, and don't send the entry just processed
				outOfSpace = TRUE;
				break;
			}

			dataCnt += currDataCnt;
			(*entryCnt)++;

cifst2_fillSearchDir_continue:
			searchp->cookie = direntPlusp->dirEntry.cookie;
			entryCntThis++;
			FS_READDIRPLUS_ADV_ENTRY_PTR(direntPlusp);
		} /* while(entryCntThis) */

		//------------------------------------
		// Set cookieVerif and eof flag before
		// packet descriptor become invalid
		//------------------------------------
		searchp->cookieVerif = dirInfop->cookieVerif;

		// Set eof flag to tell the caller whether the entries
		// returned reach the end of search or not.
		if ((fsEof == FALSE) ||
			(entryCntThis < dirInfop->numEntriesThis) ||
			(outOfSpace == TRUE))
			*eof = FALSE;
		else
			*eof = TRUE;


		// Clean eDescp
		if (eDescp)
			eee_freePacket(eDescp);

		if ((outOfSpace == TRUE) || 
			(*entryCnt == entryCntTotal) ||
			(fsEof == TRUE))
			// Reach max data bytes to return or max entries to return
			// Or reach eof of directory.
			break;

		// Call READDIRPLUS again
		exitThread = FALSE;
		rc = cifst2_getMoreFileEntries(opp,
										searchp,
										maxBytes - dataCnt,
										&eDescp,
										&dirInfop,
										&exitThread);
		if (rc != CIFSTATUS_OK)
		{
			dataCnt = -1;
			goto cifst2_fillSearchDir_done;
		}

		if (dirInfop->eof)
		{
			// Search of directory reach the end.
			// Don't need to call FS any more.
			fsEof = TRUE;
		}
	} /* while(*entryCnt < entryCntTotal) */

cifst2_fillSearchDir_done:

	if (exitThread == TRUE)
	{
		// Deallocate memory
		if (transacp->internalDatap != NULL)
		{
			CIFS_RAM_DEALLOC(transacp->internalDatap);
			transacp->internalDatap = NULL;
		}
		thread_exit();
		return (-1);
	}
	else
		return (dataCnt);
} /* cifst2_fillSearchDir() */

/*-----------------------------------------------------------------
 * Name			: cifstrans2_replyFind
 *
 * Description	: Reply to TRANS2 FIND FIRST2 or FIND NEXT2
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/04/2002
 *
 *-----------------------------------------------------------------
 */
int32			// CIFSATUS
cifstrans2_replyFind(cifs_operation			*opp,
					cifs_transac2			*transacp,
					fs_Attr					*fsAttrp,
					fs_MsgReadDirPlusResp	*dirInfop,
					uni16					*resumeFileName,
					boolean					*eos)
{
	int32			rc = CIFSTATUS_OK;
	cifs_search		*searchp = transacp->searchp;
	uchar8			*odatap, *oparap;
	ushort16		estDataCnt;
	short16			dataCnt;
	ushort16		entryCntTotal, entryCnt;
	boolean			fsEof;				// FS search reach eof or not
	ushort16		offset, maxBytes;
    uchar8			*lastNamePos;


	transacp->swsCnt	= 0;							 // No setup words
	transacp->tpbsCnt	= MIN(transacp->maxParaCnt, 10); // 10 parameter bytes
	transacp->rpbsCnt	= transacp->tpbsCnt;
	if (transacp->maxParaCnt < 10)
	{
		CIFS_PRINT(CIFS_DBGLV_WARN, ("cifstrans2_replyFind: "
					"Maximum number of parameter bytes allowed "
					"to return is less than 10"));
	}

	//----------------
	//Initialize reply
	//----------------
	if (cifstrans2_replyInit(opp, transacp) == TRUE)
		return (CIFSTATUS_OK);		// No response required

	//------------------------------------
	// Check total entries count to return
	//------------------------------------
	if (fsAttrp)
	{
		// Return only one entry in response
		entryCntTotal = 1;
		fsEof = TRUE;
	}
	else
	{
		// Return entries in the directory in response
		if (dirInfop->eof || dirInfop->numEntriesThis >= searchp->count)
		{
			// Number of entries in response
			entryCntTotal = MIN(dirInfop->numEntriesThis, searchp->count);
			fsEof = TRUE;
		}
		else
		{
			// Need to call READDIRPLUS again
			entryCntTotal = searchp->count;
			fsEof = FALSE;
		}
	}

	//
	// Offset (from where smb header starts) to data parameter
	//
	offset = transacp->datap - opp->respMsg.smbp;

	//
	// Pointer to outgoing buffer where data bytes start
	//
	odatap = transacp->datap;

	//-------------------------------------
	// Calculate estimated data bytes count
	//-------------------------------------
	estDataCnt = 0;
	if (fsEof == TRUE)
	{
		estDataCnt = cifst2_calFindEntriesSize(transacp->infoLevel, 
												entryCntTotal,
				(GETLE2(opp->reqMsg.smbp, SMBO_FLG2) & SMB_FLG2_UNICODE) != 0);

	}

	//------------------------------------------------------------------
	// Allocate memory if estimated number of data bytes is more than 
	// one packet buffer, or we need to call READDIRPLUS more than once.
	//------------------------------------------------------------------
	if ((fsEof == FALSE) ||
		(fsEof == TRUE && (CIFS_S_BUFFER < offset + estDataCnt)))
	{
		// Allocate memory of max data bytes that can be sent
#ifdef CIFSTEST
		CIFS_RAM_ALLOC(transacp->internalDatap, uchar8, 5000);
#else
		CIFS_RAM_ALLOC(transacp->internalDatap, uchar8, transacp->maxDataCnt);
#endif
		if (transacp->internalDatap == NULL)
		{
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifstrans2_replyFind: no memory"));
			return (CIFSTATUS_INSUFF_SERVER_RESOURCES);
		}

		// Pointer to the allocated buffer, used to store data bytes to be sent
		odatap = transacp->internalDatap;
	}

	// Set memory maxBytes so we won't write data beyond that
	if (transacp->internalDatap)
		maxBytes = transacp->maxDataCnt;
	else
		maxBytes = CIFS_S_BUFFER - offset;

	maxBytes = MIN(transacp->maxDataCnt, maxBytes);

	//-------------------
	// Fill in data bytes
	//-------------------

	if (fsAttrp)
	{
		//-----------------
		// Retrun one entry
		//-----------------
		dataCnt = cifst2_fillSearchEntry(opp,
										transacp,
										odatap,
										fsAttrp,
										transacp->searchp->pattern,
										0,
										&lastNamePos,
										maxBytes);
		entryCnt = 1;
		*eos = TRUE;
	}
	else
	{
		//--------------------------------
		// Return entries in the directory
		//--------------------------------
		dataCnt = cifst2_fillSearchDir(opp,
										transacp,
										odatap,
										dirInfop,
										resumeFileName,
										entryCntTotal,
										maxBytes,
										&entryCnt,
										&lastNamePos,
										&fsEof);
		*eos = fsEof;
	}

	if (dataCnt <= 0)
	{
		rc = CIFSTATUS_UNSUCCESSFUL;
		goto cifstrans2_replyFind_done;
	}
	else 
		rc = CIFSTATUS_OK;

	transacp->tdbsCnt = dataCnt;
	transacp->rdbsCnt = transacp->tdbsCnt;

	if ((transacp->maxReplySize - offset) < dataCnt)
	{
		transacp->flags |= CIFS_TRANS2_FLAG_MULTI_MSG;
	}

	cifs_getSmbUniStr(	searchp->lastFileName,
						lastNamePos,
						unisizeof(searchp->lastFileName),
						opp->respMsg.smbp);

	//-----------------------
	//Fill in parameter bytes
	//-----------------------
	oparap = transacp->parap;
	if (transacp->flags & CIFS_TRANS2_FIND_FIRST)
	{
		// Fill Sid field
		PUTLE2(oparap, 0, transacp->searchp->findHandle);
		oparap += 2;
	}

	// Fill SearchCount
	PUTLE2(oparap, 0, entryCnt);
	// Fill EndOfSearch
	PUTLE2(oparap, 2, *eos?1:0);
	// Fill EaErrorOffset;
	PUTLE2(oparap, 4, 0);
	// Fill LastNameOffset
	PUTLE2(oparap, 6, (ushort16)(lastNamePos - odatap));

	//----------------------
	//Build response message
	//----------------------
	cifstrans2_reply(opp, transacp);

cifstrans2_replyFind_done:

	// Deallocate memory
	if (transacp->internalDatap != NULL)
	{
		CIFS_RAM_DEALLOC(transacp->internalDatap);
		transacp->internalDatap = NULL;
	}

	return (rc);
} /* cifstrans2_replyFind */


//************ Transaction2 Query File System Info **************************

/*-----------------------------------------------------------------
 * Name			: cifstrans2_queryFsInfo
 *
 * Description	: TRANS2 QUERY RS INFO
 * 				  Get filesystem infomation.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/31/2002
 *
 *-----------------------------------------------------------------
 */
int32			// CIFS_SENDRESPNOW or CIFS_DONTSENDRESP
cifstrans2_queryFsInfo(cifs_operation	*opp,
						cifs_transac2	*transacp,
						uchar8			*parp,
						ushort16		pbsCnt,
						uchar8			*datp,
						ushort16		dbsCnt)
{
	int32			rc;
	ushort16		infoLevel = GETLE2(parp, 0);
	fs_Handle		shareHandle;
	esm_event_t		evt;


	//Check share access
	//TODO - Is SEC_SYSTEM_SECURITY_ACCESS right one or not?
	rc = cifs_secCheckShareAccess(opp->siup, SEC_SYSTEM_SECURITY_ACCESS);
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);


	CIFS_PRINT(CIFS_DBGLV_DBG, ("cifstrans2_queryFsInfo: "
				"InfoLevel = 0x%x", infoLevel));

	switch(infoLevel)
	{
		case CIFS_FSINFOLV_ALLOCATION:
		case CIFS_FSINFOLV_VOLUME:
		case CIFS_FSINFOLV_QUERY_VOLUME:
		case CIFS_FSINFOLV_QUERY_SIZE:
		case CIFS_FSINFOLV_QUERY_DEVICE:
		case CIFS_FSINFOLV_QUERY_ATTR:
			break;

		default:
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifstrans2_queryFsInfo: "
						"Unknown information level %d", infoLevel));
			cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_unknownlvl,
							CIFSTATUS_INVALID_LEVEL);
			return (CIFS_SENDRESPNOW);
	}
	transacp->infoLevel = infoLevel;

	shareHandle = opp->siup->sharep->handle;

	// TODO - Call FS to get info. 
	//        Need to combine fs_MsgFsinfoReq and fs_MsgStatReq

	//--------------------------
	// Wait for response from FS
	//--------------------------
	thread_yield(&evt, CIFS_INTMSG_TIMEOUT_FS);
	/*
	if (evt.event == )
	{
		if ()
		{
		}
		else
			rc = cifstrans2_replyQueryFsInfo(opp, dirInfop);
	}
	else
	{
		// TODO call cifs_eventCommon, which should call cifs_errResp
		rc = NFX_ERR;
	}
	*/

	if (evt.event != THREAD_EVENT_TIMEOUT)
	{
		eee_freePacket((eee_desc_t*)evt.p1.u.p);
	}

	if (rc == NFX_ERR)
	{
		return (CIFS_SENDRESPNOW);
	}
	else
		return (CIFS_DONTSENDRESP);
} /* cifstrans2_queryFsInfo() */


/*-----------------------------------------------------------------
 * Name			: cifstrans2_replyQueryFsInfo
 *
 * Description	: Reply to TRANS2 QUERY RS INFO
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/05/2002
 *
 *-----------------------------------------------------------------
 */
int32			//CIFSTATUS
cifstrans2_replyQueryFsInfo(cifs_operation	*opp,
							cifs_transac2   *transacp/* ... */)
{
	int32			dataCnt;
	ushort16		offset, maxBytes;

	transacp->swsCnt	= 0;	// No setup words
	transacp->tpbsCnt	= 0;	// No parameter bytes
	transacp->rpbsCnt	= 0;

	//----------------
	//Initialize reply
	//----------------
	if (cifstrans2_replyInit(opp, transacp) == TRUE)
		return (CIFSTATUS_OK);		// No response required

	//
	// Offset (from where smb header starts) to data parameter
	//
	offset = transacp->datap - opp->respMsg.smbp;

	//We assume the whole message would fit in one packet buffer.
	//Here we add protection just in case the message is larger
	//than one buffer size.
	maxBytes = CIFS_S_BUFFER - offset;
	maxBytes = MIN(transacp->maxDataCnt, maxBytes);

	//------------------
	//Fill in data bytes
	//------------------
	switch(transacp->infoLevel)
	{
		case CIFS_FSINFOLV_ALLOCATION:
			// TODO
			dataCnt = 11111111;
			break;

		case CIFS_FSINFOLV_VOLUME:
			// TODO
			dataCnt = 11111111;
			break;

		case CIFS_FSINFOLV_QUERY_VOLUME:
			// TODO
			dataCnt = 11111111;
			break;

		case CIFS_FSINFOLV_QUERY_SIZE:
			// TODO
			dataCnt = 11111111;
			break;

		case CIFS_FSINFOLV_QUERY_DEVICE:
			// TODO
			dataCnt = 11111111;
			break;

		case CIFS_FSINFOLV_QUERY_ATTR:
			// TODO
			dataCnt = 11111111;
			break;
	}
	transacp->tdbsCnt = dataCnt;
	transacp->rdbsCnt = transacp->tdbsCnt;

	//----------------------
	//Build response message
	//----------------------
	cifstrans2_reply(opp, transacp);

	return (CIFSTATUS_OK);
} /* cifstrans2_replyQueryFsInfo() */


//************ Transaction2 Query Path/File Info **************************

/*-----------------------------------------------------------------
 * Name			: cifstrans2_queryPathInfo
 *
 * Description	: TRANS2 QUERY PATH INFO
 * 				  Get information about a named file or directory.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/31/2002
 *
 *-----------------------------------------------------------------
 */
int32			// CIFS_SENDRESPNOW or CIFS_DONTSENDRESP
cifstrans2_queryPathInfo(cifs_operation	*opp,
						cifs_transac2	*transacp,
						uchar8			*parp,
						ushort16		pbsCnt,
						uchar8			*datp,
						ushort16		dbsCnt)
{
	int32				rc;
	ushort16			infoLevel;
	uchar8				*bp;
	fs_MsgMclResp		*mclResp;
	fs_AttrType			attrType;
	fs_Attr				*fsAttrp;
	eee_desc_t			*eDescp;


	//Check share access
	rc = cifs_secCheckShareAccess(opp->siup, SEC_FILE_READ_ATTRIBUTES);
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);


	// Get infomation level and validate it
	infoLevel = GETLE2(parp, 0);

	CIFS_PRINT(CIFS_DBGLV_DBG, ("cifstrans2_queryPathInfo: "
				"InfoLevel = 0x%x", infoLevel));

	switch(infoLevel)
	{
		case CIFS_INFOLV_STANDARD:
		case CIFS_INFOLV_IS_NAME_VALID:
		case CIFS_INFOLV_QUERY_FIEL_BASIC:
		case CIFS_INFOLV_QUERY_FILE_STANDARD:
		case CIFS_INFOLV_QUERY_FILE_NAME:
		case CIFS_INFOLV_QUERY_FILE_ALL:
		case CIFS_INFOLV_QUERY_FILE_ALT_NAME:
		case CIFS_INFOLV_QUERY_FILE_COMPRESSION:
			break;

		case CIFS_INFOLV_QUERY_EA_SIZE:
		case CIFS_INFOLV_QUERY_EAS_FROM_LIST:
		case CIFS_INFOLV_QUERY_ALL_EAS:
		case CIFS_INFOLV_QUERY_FILE_EA:
		case CIFS_INFOLV_QUERY_FILE_STREAM:
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifstrans2_queryPathInfo: "
						"Information level 0x%x unsupported", infoLevel));
			cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_NOT_SUPPORTED);
			break;

		default:
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifstrans2_queryPathInfo: "
						"Unknown information level 0x%x", infoLevel));
			cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_INVALID_LEVEL);
	}
	transacp->infoLevel = infoLevel;

	// Get file or directory name
	bp = parp + 6;	//Point to file name in souce SMB message
	transacp->pathNamep = bp;

	// Treat special infoLevel
	if (infoLevel == CIFS_INFOLV_IS_NAME_VALID ||
		infoLevel == CIFS_INFOLV_QUERY_FILE_NAME)
	{
		rc = cifstrans2_replyQueryPathFileInfo(opp, transacp, NULL);
		cifsm_testCIFSTATUS_ifErrReturn(opp, rc);
	}

	//------------------------------------------------------------------
	// Send MULTI COMPONENT LOOKUP request to FS to get file information 
	//------------------------------------------------------------------
	//XXX - We may not have to check access permission to get file info.
	//      A "Query path info" trace shows that even the user doesn't
	//      have read attributes permission to a file, the file info is 
	//      still returned.
	attrType = FS_POST_ATTRS;
	//TODO: Ask for file's security descriptor only when info level is 
	//      QUERY_FILE_ALL in order to check caller's access to the file
	//if (infoLevel == CIFS_INFOLV_QUERY_FILE_ALL)
		//attrType |= FS_SD
	rc = cifs_fsMCLReqFromMsg(opp, NULL, opp->reqMsg.smbp, bp, NULL, attrType);
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);

	//-------------------------------------------------
	// Wait for MULTI COMPONENT LOOKUP response from FS
	//------------------------------------------------- 
	rc = cifs_waitFsResp(FS_REQ_MCL, (void **)&mclResp, &eDescp);
	if (rc != CIFSTATUS_OK)
		rc = CIFSTATUS_UNSUCCESSFUL;
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);

	if (mclResp->hdr.fsStatus == FS_ERR_NONE)
	{
		// Got file attribute
		fsAttrp = &(mclResp->attr[0]);
		rc = cifstrans2_replyQueryPathFileInfo(opp, transacp, fsAttrp);
		cifsm_testCIFSTATUS_ifErrFreeDescAndReturn(opp, rc, eDescp);
	}
	else
		cifsm_testFsStatus_ifErrReturn(opp, mclResp->hdr.fsStatus, eDescp);

	//TODO also get stateful attributes

	// Free MULTI COMPONENT LOOKUP response
	eee_freePacket(eDescp);

	return (CIFS_DONTSENDRESP);
} /* cifstrans2_queryPathInfo() */


/*-----------------------------------------------------------------
 * Name			: cifstrans2_queryFileInfo
 *
 * Description	: TRANS2 QUERY FILE INFO
 * 				  Get information about a handle.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/31/2002
 *
 *-----------------------------------------------------------------
 */
int32			// CIFS_SENDRESPNOW or CIFS_DONTSENDRESP
cifstrans2_queryFileInfo(cifs_operation	*opp,
						cifs_transac2	*transacp,
						uchar8			*parp,
						ushort16		pbsCnt,
						uchar8			*datp,
						ushort16		dbsCnt)
{
	int32			rc;
	ushort16		fid;
	ushort16		infoLevel;
	fs_AttrType		attrType;
	cifs_fileOpen	*fop;
	fs_Attr			*fsAttrp;
	eee_desc_t		*eDescp;


	//Check share access
	rc = cifs_secCheckShareAccess(opp->siup, SEC_FILE_READ_ATTRIBUTES);
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);


	fid = GETLE2(parp, 0);
	infoLevel = GETLE2(parp, 2);

	CIFS_PRINT(CIFS_DBGLV_DBG, ("cifstrans2_queryFileInfo: "
				"InfoLevel = 0x%x, fid = %d", infoLevel, fid));

	// Get fileOpen
	fop = cifs_findFileOpenByFid(opp->siup, fid);
	if (fop == NULL)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifstrans2_queryFileInfo: Bad Fid %d", fid));
		cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_INVALID_HANDLE);
	}

	switch(infoLevel)
	{
		case CIFS_INFOLV_STANDARD:
		case CIFS_INFOLV_IS_NAME_VALID:
		case CIFS_INFOLV_QUERY_FIEL_BASIC:
		case CIFS_INFOLV_QUERY_FILE_STANDARD:
		case CIFS_INFOLV_QUERY_FILE_NAME:
		case CIFS_INFOLV_QUERY_FILE_ALL:
		case CIFS_INFOLV_QUERY_FILE_ALT_NAME:
		case CIFS_INFOLV_QUERY_FILE_COMPRESSION:
			break;

		case CIFS_INFOLV_QUERY_EA_SIZE:
		case CIFS_INFOLV_QUERY_EAS_FROM_LIST:
		case CIFS_INFOLV_QUERY_ALL_EAS:
		case CIFS_INFOLV_QUERY_FILE_EA:
		case CIFS_INFOLV_QUERY_FILE_STREAM:
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifstrans2_queryFileInfo: "
						"Information level 0x%x unsupported", infoLevel));
			cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_NOT_SUPPORTED);
			break;

		default:
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifstrans2_queryFileInfo: "
						"Unknown information level 0x%x", infoLevel));
			cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_INVALID_LEVEL);
	}
	transacp->infoLevel = infoLevel;
	transacp->fop = fop;
	transacp->pathNamep = NULL;

	//Treat special infoLevel
	if (infoLevel == CIFS_INFOLV_IS_NAME_VALID ||
		infoLevel == CIFS_INFOLV_QUERY_FILE_NAME)
	{
		rc = cifstrans2_replyQueryPathFileInfo(opp, transacp, NULL);
		cifsm_testCIFSTATUS_ifErrReturn(opp, rc);
	}

    //------------------------------------------------------------
	// Send GETATTR request to FS and FSSF to get file information 
	//------------------------------------------------------------
	attrType = FS_POST_ATTRS;
	//TODO: Ask for file's security descriptor only when info level is 
	//      QUERY_FILE_ALL in order to check caller's access to the file 
	//if (infoLevel == CIFS_INFOLV_QUERY_FILE_ALL)
		//attrType |= FS_SD
	//rc = cifs_fsGetAttrReq(opp, opp->reqMsg.smbp, handle, attrType);
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);


	//-----------------------------------
	// Wait for response from FS and FSSF
	//-----------------------------------
	//TODO...

	return (CIFS_DONTSENDRESP);
} /* cifstrans2_queryFileInfo() */


/*-----------------------------------------------------------------
 * Name			: cifstrans2_replyQueryPathFileInfo
 *
 * Description	: Reply to TRANS2 QUERY PATH/FILE INFO
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/05/2002
 *
 *-----------------------------------------------------------------
 */
int32			//CIFSTATUS
cifstrans2_replyQueryPathFileInfo(cifs_operation	*opp,
								  cifs_transac2		*transacp,
								  fs_Attr			*fsAttrp)
{
	int32		rc;
	uni16		pathName[CIFS_S_PATHSTR+1];
	short16		dataCnt;
	ushort16	offset, maxBytes;

	if (fsAttrp->attrType != FS_POST_ATTRS)
		return (CIFSTATUS_UNSUCCESSFUL);

	transacp->swsCnt			= 0;		// No setup words
	if (transacp->infoLevel > 0x100)
	{
		// For the information level greater than 0x100, the transaction 
		// response has 1 parameter word, which should be ignored by the 
		// client. See 4.2.16 in CIFS 0.99.
		transacp->tpbsCnt	= MIN(transacp->maxParaCnt, 2);	// 2 parameter bytes
		transacp->rpbsCnt	= transacp->tpbsCnt;
	}
	else
	{
		transacp->tpbsCnt	= 0;		// No parameter bytes
		transacp->rpbsCnt	= 0;
	}

	//----------------
	//Initialize reply
	//----------------
	if (cifstrans2_replyInit(opp, transacp) == TRUE)
		return (CIFSTATUS_OK);		// No response required

	//
	// Offset (from where smb header starts) to data parameter
	//
	offset = transacp->datap - opp->respMsg.smbp;

	//We assume the whole message would fit in one packet buffer.
	//Here we add protection just in case the message is larger
	//than one buffer size.
	maxBytes = CIFS_S_BUFFER - offset;
	maxBytes = MIN(transacp->maxDataCnt, maxBytes);


	if (transacp->tpbsCnt == 2)
		PUTLE2(transacp->parap, 0, 0);

	//Get UNICODE path name
	if (transacp->pathNamep)
	{
		//SMB request is get PATH Info
		cifs_getSmbUniStr(	pathName,
							transacp->pathNamep,
							unisizeof(pathName),
							opp->reqMsg.smbp);
	}
	else
	{
		//SMB request is get FILE Info
		unistrcpy(pathName, transacp->fop->pathName);
	}
	
	//------------------
	//Fill in data bytes
	//------------------
	switch(transacp->infoLevel)
	{
		case CIFS_INFOLV_STANDARD:
			dataCnt = cifst2_fillFileStandard(transacp->datap,
											  fsAttrp,
											  maxBytes);
			break;

		//case CIFS_INFOLV_QUERY_EA_SIZE:
		//	break;

		//case CIFS_INFOLV_QUERY_EAS_FROM_LIST:
		//	break;

		//case CIFS_INFOLV_QUERY_ALL_EAS:
		//	break;

		case CIFS_INFOLV_IS_NAME_VALID:
		{
			//Validate syntax of path name
			rc = cifs_checkPathSyntax(pathName);
			if (rc != CIFSTATUS_OK)
				return (rc);
			else	
				//Syntax is OK. No data in the response
				dataCnt = 0;
			break;
		}

		case CIFS_INFOLV_QUERY_FIEL_BASIC:
			dataCnt = cifst2_fillFileQueryFileBasic(opp,
													transacp->datap,
													fsAttrp,
													maxBytes);
			break;

		case CIFS_INFOLV_QUERY_FILE_STANDARD:
			dataCnt = cifst2_fillFileQueryFileStandard(opp,
														transacp->datap,
														fsAttrp,
														maxBytes);
			break;

		//case CIFS_INFOLV_QUERY_FILE_EA:
		//	break;

		case CIFS_INFOLV_QUERY_FILE_NAME:
		{
			//Put file name in the response
			uint32	pathNameLen;
			dataCnt = cifs_putUniStr(transacp->datap + 4,
									 pathName,
									 opp->respMsg.smbp,
								   	 &pathNameLen,
									 maxBytes);
			if (dataCnt >= 0)
				PUTLE4(transacp->datap, 0, pathNameLen);
			break;
		}

		case CIFS_INFOLV_QUERY_FILE_ALL:
			dataCnt = cifst2_fillFileQueryFileAll(opp,
												  transacp->datap,
												  fsAttrp,
 												  pathName,
												  maxBytes);
			break;

		//case CIFS_INFOLV_QUERY_FILE_ALT_NAME:
		//	break;

		//case CIFS_INFOLV_QUERY_FILE_STREAM:
		//	break;

		case CIFS_INFOLV_QUERY_FILE_COMPRESSION:
			// TODO
			dataCnt = 1111;
			break;
	}	

	if (dataCnt < 0)
		return (CIFSTATUS_UNSUCCESSFUL);

	transacp->tdbsCnt = dataCnt;
	transacp->rdbsCnt = transacp->tdbsCnt;

	//----------------------
	//Build response message
	//----------------------
	cifstrans2_reply(opp, transacp);

	return (CIFSTATUS_OK);
} /* cifstrans2_replyQueryPathFileInfo() */


//************** Transaction2 Set Path/File Info **************************

/*-----------------------------------------------------------------
 * Name			: cifstrans2_setPathInfo
 *
 * Description	: TRANS2 SET PATH INFO
 *				  Set information about a named file or directory.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/31/2002
 *
 *-----------------------------------------------------------------
 */
int32			// CIFS_SENDRESPNOW or CIFS_DONTSENDRESP
cifstrans2_setPathInfo(cifs_operation	*opp,
						cifs_transac2	*transacp,
						uchar8			*parp,
						ushort16		pbsCnt,
						uchar8			*datp,
						ushort16		dbsCnt)
{
	int32			rc;
	ushort16		infoLevel;
	uchar8			*bp;
	fs_Attr			fsAttr;


	//Check share access
	rc = cifs_secCheckShareAccess(opp->siup, SEC_FILE_WRITE_ATTRIBUTES);
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);


	// Get information level
	infoLevel = GETLE2(parp, 0);
	CIFS_PRINT(CIFS_DBGLV_DBG, ("cifstrans2_setPathInfo: "
				"InfoLevel = 0x%x", infoLevel));

	// Get file or directory name
	bp = parp + 6;
	transacp->pathNamep = bp;

	// Get pointer to the beginning of file info in SMB request message
	if ((GETLE2(opp->reqMsg.smbp, SMBO_FLG2) & SMB_FLG2_UNICODE) != 0)
	{
		// UNICODE path name in source message
		bp += ((address_t)bp) % 2;  //Unicode string are aligned on word
		bp += unistrsize((uni16 *)bp);
	}
	else
	{
		// ASCII path name in source message
		bp += strlen(bp) + 1;
	}

	//TODO - Need to check access permission for write attributes?

	switch(infoLevel)
	{
		case CIFS_INFOLV_STANDARD:
			rc= cifst2_getFileStandard(opp, &fsAttr, bp);
			break;

		case CIFS_INFOLV_QUERY_EA_SIZE:
		case CIFS_INFOLV_QUERY_ALL_EAS:
		case CIFS_INFOLV_SET_UNIX_BASIC:
		case CIFS_INFOLV_SET_UNIX_LINK:
		case CIFS_INFOLV_SET_UNIX_HLINK:
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifstrans2_setPathInfo: "
						"Information level %d unsupported", infoLevel));
			cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_NOT_SUPPORTED);
			break;

		default:
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifstrans2_SetPathInfo: "
						"Unknown information level %d", infoLevel));
			cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_INVALID_HANDLE);
	}
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);
	transacp->infoLevel = infoLevel;


	// TODO - Call FSSF to set info

	// Prepare response message
	cifstrans2_replySetPathInfo(opp, transacp);

	 return (CIFS_DONTSENDRESP);
} /* cifstrans2_setPathInfo() */


/*-----------------------------------------------------------------
 * Name			: cifstrans2_replySetPathInfo
 *
 * Description	: Reply to TRANS2 SET PATH INFO
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/05/2002
 *
 *-----------------------------------------------------------------
 */
int32
cifstrans2_replySetPathInfo(cifs_operation	*opp,
							cifs_transac2	*transacp /* ... */)
{
	int32			dataCnt;

	transacp->swsCnt	= 0;	// No setup words
	transacp->tpbsCnt	= 0;	// No paremeter bytes
	transacp->rpbsCnt	= 0;
	transacp->tdbsCnt	= 0;	// No data bytes
	transacp->rdbsCnt	= 0;

	// Initialize reply
	if (cifstrans2_replyInit(opp, transacp) == TRUE)
		return (CIFSTATUS_OK);		// No response required

	// Build response message
	cifstrans2_reply(opp, transacp);

	return (CIFSTATUS_OK);
} /* cifstrans2_replySetPathInfo() */


/*-----------------------------------------------------------------
 * Name			: cifstrans2_setFileInfo
 *
 * Description	: TRANS2 SET FILE INFO
 * 				  Set information by handle.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/31/2002
 *
 *-----------------------------------------------------------------
 */
int32			// CIFS_SENDRESPNOW or CIFS_DONTSENDRESP
cifstrans2_setFileInfo(cifs_operation	*opp,
						cifs_transac2	*transacp,
						uchar8			*parp,
						ushort16		pbsCnt,
						uchar8			*datp,
						ushort16		dbsCnt)
{
	int32			rc;
	ushort16		fid;
	ushort16		infoLevel;
	cifs_fileOpen	*fop;
	uchar8			*bp;
	fs_Attr			fsAttr;


	//Check share access
	rc = cifs_secCheckShareAccess(opp->siup, SEC_FILE_WRITE_ATTRIBUTES);
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);


	// Get infomation level
	fid = GETLE2(parp, 0);
	infoLevel = GETLE2(parp, 2);
	CIFS_PRINT(CIFS_DBGLV_DBG, ("cifstrans2_setFileInfo: "
				"InfoLevel = 0x%x", infoLevel));

	// Get fileOpen
	fop = cifs_findFileOpenByFid(opp->siup, fid);
	if (fop == NULL)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifstrans2_setFileInfo: Bad Fid %d", fid));
		cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_INVALID_HANDLE);
	}

	//TODO - Need to check access permission for write attributes?

	// Get pointer to the beginning of file info in SMB request message
	bp = parp + 6;

	switch(infoLevel)
	{
		case CIFS_INFOLV_STANDARD:
			rc= cifst2_getFileStandard(opp, &fsAttr, bp);
			break;

		case CIFS_INFOLV_SET_BASIC:
			rc= cifst2_getFileBasic(opp, &fsAttr, bp);
			break;

		case CIFS_INFOLV_SET_DISPOSITION:
			rc= cifst2_getFileDisposition(opp, &fsAttr, bp);
			break;

		case CIFS_INFOLV_SET_ALLOCATION:
			rc= cifst2_getFileAllocation(opp, &fsAttr, bp);
			break;

		case CIFS_INFOLV_SET_END_OF_FILE:
			rc= cifst2_getFileEndOfFile(opp, &fsAttr, bp);
			break;

		case CIFS_INFOLV_QUERY_EA_SIZE:
		case CIFS_INFOLV_SET_UNIX_BASIC:
		case CIFS_INFOLV_SET_UNIX_LINK:
		case CIFS_INFOLV_SET_UNIX_HLINK:
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifstrans2_setFileInfo: "
						"Information level %d unsupported", infoLevel));
			cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_NOT_SUPPORTED);
			break;

		default:
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifstrans2_SetFileInfo: "
					"Unknown information level %d", infoLevel));
			cifsm_testCIFSTATUS_ifErrReturn(opp, CIFSTATUS_INVALID_HANDLE);
	}
	cifsm_testCIFSTATUS_ifErrReturn(opp, rc);
	transacp->infoLevel = infoLevel;

	// TODO - Call FSSF to set info

	// Prepare response message
	cifstrans2_replySetFileInfo(opp, transacp);

	return (CIFS_DONTSENDRESP);
} /* cifstrans2_setFileInfo() */


/*-----------------------------------------------------------------
 * Name			: cifstrans2_replySetFileInfo
 *
 * Description	: Reply to TRANS2 SET FILE INFO
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/05/2002
 *
 *-----------------------------------------------------------------
 */
int32 cifstrans2_replySetFileInfo(cifs_operation *opp, cifs_transac2 *transacp)
{
	int32			dataCnt;

	transacp->swsCnt		= 0;		// No setup words
	if (transacp->infoLevel > 0x100 && transacp->infoLevel < 0x200)
	{
		// For the information level greater than 0x100 and below
		// 0x200, the transaction response has 1 parameter word,
		// which should be ignored by the client. See 4.2.19 in CIFS 0.99.
		transacp->tpbsCnt	= MIN(transacp->maxParaCnt, 2);	// 2 parameter bytes
		transacp->rpbsCnt	= transacp->tpbsCnt;
	}
	else
	{
		transacp->tpbsCnt	= 0;	// No paremeter bytes
		transacp->rpbsCnt	= 0;
	}
	transacp->tdbsCnt		= 0;	// No data bytes
	transacp->rdbsCnt		= 0;

	// Initialize reply
	if (cifstrans2_replyInit(opp, transacp) == TRUE)
		return (CIFSTATUS_OK);		// No response required

	if (transacp->tpbsCnt == 2)
		PUTLE2(transacp->parap, 0, 0);

	// Build response message
	cifstrans2_reply(opp, transacp);

	return (CIFSTATUS_OK);
} /* cifstrans2_replySetFileInfo() */


/*-----------------------------------------------------------------
 * Name			: cifsm_transac2
 *
 * Description	: transaction2 (0x32) 
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 01/29/2002
 *
 *-----------------------------------------------------------------
 */
int32			// CIFS_SENDRESPNOW or CIFS_DONTSENDRESP
cifsm_transac2(cifs_client		*clp,
				cifs_session	*ssp, 
				cifs_shareInUse	*siup,
				cifs_operation	*opp,
				uchar8			*icmdp,
				uchar8			*ocmdp)
{
	cifs_transac2	transac;
	uchar8			*parp = NULL, *datp = NULL;
	ushort16		setup0 = 0xFF;
	ushort16		tpbsCnt, tdbsCnt, mpbrCnt, mdbrCnt, mswrCnt;
	ushort16		pbsCnt, pbsOff, dbsCnt, dbsOff, swsCnt;
	boolean			discTid, noAnswerReq;
	uint32			timeout;	//TODO
	ushort16		maxReplySize;
	int32			result;

	memset(&transac, 0, sizeof(cifs_transac2));
	transac.flags |= CIFS_TRANS2_FLAG_VALID;

	// Decode word fields in request message
	tpbsCnt		= GETLE2(icmdp, SMBCO_WP0);	// Total param bytes being sent
	tdbsCnt		= GETLE2(icmdp, SMBCO_WP1);	// Total data bytes being sent
	mpbrCnt		= GETLE2(icmdp, SMBCO_WP2);	// Max # of param bytes to return
	mdbrCnt		= GETLE2(icmdp, SMBCO_WP3);	// Max # of data bytes to return
	mswrCnt		= GETLE2(icmdp, SMBCO_WP4) & 0xFF;
											// Max # of setup words to return
	discTid		= SMB_TESTBIT2(icmdp, SMBCO_WP5, 0); // Disconnect TID
	noAnswerReq = SMB_TESTBIT2(icmdp, SMBCO_WP5, 1); // One/Two way transaction
	timeout		= GETLE4(icmdp, SMBCO_WP6);	// Timeout to completion
	pbsCnt		= GETLE2(icmdp, SMBCO_WP9);	// # of param bytes in this buffer
	pbsOff		= GETLE2(icmdp, SMBCO_WP10); // Offset fm header to param bytes
	dbsCnt		= GETLE2(icmdp, SMBCO_WP11); // # of data bytes in this buffer
	dbsOff		= GETLE2(icmdp, SMBCO_WP12); // Offset from header to data bytes
	swsCnt		= GETLE2(icmdp, SMBCO_WP13); // Setup words count

	if (pbsCnt > tpbsCnt || dbsCnt > tdbsCnt)
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_transac2: "
					"transaction params error"));
		cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_error, CIFSTATUS_UNSUCCESSFUL);
		return (CIFS_SENDRESPNOW);
	}

	// Get setup word, setup[0], the value of which is subcommand code
	if (swsCnt != 0)
	{
		if (swsCnt != 1)
		{
			CIFS_PRINT(CIFS_DBGLV_INFO, ("cifsm_transac2: "
						"There are %d setup words. "
						"Ignore the setup words except the first one", 
						swsCnt));
		}
		setup0 = GETLE2(icmdp, SMBCO_WP14);

		// Check subcommand
		if (setup0 >= CIFS_S_TRANS2_SUBCMD)
		{
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_transac2: "
						"Invalid subcommand 0x%x in SMB_COM_TRANSACTION2", 
						setup0));
			cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_error,
							CIFSTATUS_UNSUCCESSFUL);
			return (CIFS_SENDRESPNOW);
		}
		else if (cifs_msg_transac2_subcmd[setup0].procSubcmd == NULL)
		{
			CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_transac2: "
						"Unsupported subcommand 0x%x in SMB_COM_TRANSACTION2", 
						setup0));
			cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_nosupport,
							CIFSTATUS_NOT_SUPPORTED);
			return (CIFS_SENDRESPNOW);
		}
	}
	else
	{
		CIFS_PRINT(CIFS_DBGLV_ERR, ("cifsm_transac2: "
					"No setup words in the request"));
		cifs_errResp(opp, CIFS_ERRSRV, CIFSERR_error, CIFSTATUS_UNSUCCESSFUL);
		return (CIFS_SENDRESPNOW);
	}

	if (tpbsCnt != 0)
		parp = opp->reqMsg.smbp + pbsOff;
	if (tdbsCnt != 0)
		datp = opp->reqMsg.smbp + dbsOff;

	if (pbsCnt < tpbsCnt || dbsCnt < tdbsCnt)
	{
		// TODO
		// More parameter bytes or data bytes to come. Do the followings
		// 1. update data structure stored in session
		// 2. create interim reponse (only once, to primary request)
		// 3. if need to send interim response, return (CIFS_SENDRESPNOW);
		//    otherwise, return (CIFS_DONTSENDRESP);
	}

	// Set values in operation request's transaction2 field
	transac.subCmd = setup0;

	if (discTid)
		transac.flags |= CIFS_TRANS2_FLAG_CLOSE_TID;
	if (noAnswerReq)
		transac.flags |= CIFS_TRANS2_FLAG_NO_ANSWER;

	transac.maxParaCnt = mpbrCnt;
	transac.maxDataCnt = mdbrCnt;
	transac.maxSetupCnt = mswrCnt;

	//Data in each reponse SMB is expected to be truncated to the maximum 
	//number of 512 byte blocks which will fit (starting at a 32 bit maxBytes)
	//in the negociated buffer size with the odd bytes remaining (if any) in 
	//the final buffer (CIFS 0.99, paragraph 3.14)
	//maxReplySize is the maximum size of SMB message
	//TODO - maxReplySize may also be restricted to MTU of TCP
	maxReplySize = MIN(opp->clp->maxBufSize, CIFS_S_BUFFER);
	maxReplySize -= maxReplySize % 512;
#ifdef CIFSTEST
	//maxReplySize = 120;
	//maxReplySize = 250;
#endif
	transac.maxReplySize = maxReplySize;

	// Call subcommand
	result = (*cifs_msg_transac2_subcmd[setup0].procSubcmd)(opp, 
															&transac,
															parp, 
															pbsCnt, 
															datp, 
															dbsCnt);
	if ((result == CIFS_SENDRESPNOW) && noAnswerReq == TRUE)
	{
		CIFS_PRINT(CIFS_DBGLV_DBG3, ("cifsm_transac2: To close operation..."));
		// We can reach here only when there are errors during
		// processing subcommands.
		cifs_closeOperation(opp);
		result = CIFS_DONTSENDRESP;
	}

	return (result);
} /* cifsm_transac2() */


/*-----------------------------------------------------------------
 * Name			: cifsm_findClos2
 *
 * Description	: Find Close2 (0x34) 
 * 				  Close a search started by TRANS2_FIND_FIRST2 request.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 02/06/2002
 *
 *-----------------------------------------------------------------
 */
int32			// CIFS_SENDRESPNOW or CIFS_DONTSENDRESP
cifsm_findClos2(cifs_client			*clp,
				cifs_session		*ssp, 
				cifs_shareInUse		*siup,
				cifs_operation		*opp,
				uchar8				*icmdp,
				uchar8				*ocmdp)
{
	ushort16	sid;
	cifs_search	*searchp;

	sid = GETLE2(icmdp, SMBCO_WP0);

	cifs_closeSearchBySid(opp->siup, sid);

	// TODO - send anything to FSSF?

	// Build response message. Word count and byte count are both 0.
	cifs_setCmdWordByteCnt(&opp->respMsg, 0, 0, TRUE);

	return(CIFS_SENDRESPNOW);
} /* cifsm_findClos2() */
