/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : ea.h
 *
 * Description  : Main header file of the Eagle Agent module.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */
#ifndef _EA_H_
#define _EA_H_

/*------------------------------------------------------------------------------
 * Include files
 *------------------------------------------------------------------------------
 */

#include <stdio.h>      // fprintf(), stderr
#include <stdlib.h>     // exit()
#include <string.h>     // memset()

#include "nfx-incl.h"
#include "exports/anpssc-api.h"
#include "exports/evm-api.h"
#include "exports/fs-api.h"
#include "exports/elog-api.h"
#include "exports/ncm-api.h"
#include "ea-api.h"

/*------------------------------------------------------------------------------
 * Macros and constants.
 *------------------------------------------------------------------------------
 */

#define APPNAME_NCM		0
#define APPNAME_EVM		1
#define APPNAME_FS		2
#define APPNAME_CM		3
#define APPNAME_MOUNTD  4
#define APPNAME_TOTAL   5


#define	MAX_EA_SESSIONS		256
#define EA_DBG_PRINT(_par_) E_LOG(class_1, LOG_DEBUG, eaAppId, 0, 0, 0, _par_)
#define EA_ERR_PRINT(_par_) E_LOG(class_1, LOG_ERR  , eaAppId, 0, 0, 0, _par_)
#define EA_FCT_PRINT(_par_) E_LOG(class_1, LOG_DEBUG, eaAppId, 0, 0, 0, _par_)
#define EA_ALERT_PRINT(_par_) E_LOG(class_1, LOG_ALERT, eaAppId, 0, 0, 0, _par_)

#define EA_ASSERT(_expr)                                      \
    if (!(_expr))                                             \
    {                                                         \
        printf("%s:%d: ASSERT FAILED: %s in %s\n",            \
            __FILE__, __LINE__, #_expr, __FUNCTION__);        \
        exit(NFX_ERR);                                        \
    }

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

/* EA's file descriptor */
extern int eaAppId;
extern const char toAppName[APPNAME_TOTAL][8];

/*------------------------------------------------------------------------------
 * Data structures definition
 *------------------------------------------------------------------------------
 */

/* Info needed to reply to a sender. */
typedef struct
{
    char    name[256];
    uint32  cpu;
    uint32  slot;
    uint32  plane;
} ea_senderInfo_t;

/* Basic message structure. */
typedef struct
{
    char    buf[60000];  // optimize!!!
    uint32  len;
} ea_msg_t;

typedef struct
{
	char	ncm[NCM_MSGHDR_LEN];
    char    buf[60000];  // optimize!!!
} ea_clusmsg_t;

/* Control block for the file system.
 */
typedef struct
{
    ea_msgHdr_t *       hdr;                // header of EV request
    uint64              volSize;            // Volume size in bytes
    evm_VolConfig_t *   vol;                // Volume
    
    char            newFsysName[VOL_MAX_NAME];  // New file system name    
    uint64          newHardQuota;               // Hard Quota
    uint64          newMinAutoGrow;             // Min auto grow
    uint32          newHighWaterMark;           // High water mark
	uint32			newSoftQuota;				// Soft Quota
} ea_fsysCb_t;

/* Control block for the file set.
 */
typedef struct
{
    ea_msgHdr_t *       hdr;                // header of EV request

    char                path[FS_PATH_MAX];  // Path
    boolean             quotaOn;            // Quota (on/off)
    uint64              maxNbFile;          // Max No. Files 
    uint64              maxSpace;           // Max Space 
    uint64              maxIops;            // Max IOPS 
    uint64              cacheThreshold;     // Cache Threshold
    evm_VolConfig_t *   vol;                // Volume
} ea_fsetCb_t;

/*------------------------------------------------------------------------------
 * Prototype of global functions.
 *------------------------------------------------------------------------------
 */

// ea-processEvmMsg.c

void
ea_processEvmMsg(ea_msg_t *          msg,
                 ea_senderInfo_t *   sender);

// ea-processEvMsg.c

void
ea_processEvMsg(ea_msg_t *          msg,
                ea_senderInfo_t *   sender);

// ea-processCliMsg.c

void
ea_processCliMsg(ea_msg_t *          msg,
                 ea_senderInfo_t *   sender);

void
ea_sg_init(void);

void
ea_sg_cleanup(void);

// ea-common.c

void
ea_sendMsg(void *   msg,
           size_t   len,
           char *   dest);

void
ea_sendEvNtfy(void *    msg,
              size_t    len);

#ifdef ISABELLE // code to delete
void
ea_sendEvRsp(uint32 result,
             char * output,
             uint32 userInfo);
#endif // ISABELLE

uint32 isLocalNodeName(char *nodeName);

#ifdef REMOVED_BY_SHYHPEI
void
ea_prepClusterReq(void **realReq, uint32 *reqLen,
					char *reqbuf,
					char *node,
					char **toApp,
					uint32 *slot, uint32 *cpu);

int32
ea_getClusterRsp(char *node, uint32 *rspSize, char *rspBuf, void **realRsp);
#endif

int
ea_sendClusterAgileMsg(int sock,
	ea_clusmsg_t *req, int reqLen,
	char *toApp, uint32 cpu, uint32 slot, uint32 plane,
	struct timeval *timeout,
	ea_clusmsg_t *rsp, int rspLen,
	char *node);

int
checkClusterAgileDest(char *node, uint32 cpu, uint32 slot);


// ea.c

int32
ea_crtVol(char *node,
			ea_fsysCb_t * cb,
			char *        err);

int32
ea_delVol(char *node,
			ea_fsysCb_t * cb,
			char *        err);

int32
ea_renameVol(char *node,
				ea_fsysCb_t *  cb,
				char *         err);

int32
ea_modVol(char *node,
			ea_fsysCb_t * cb,
			char *        err);

int32
ea_takeoverVol(char *node, ea_fsysCb_t *cb, char *err, uint32 delay);

int32
ea_onVol(char *node,
			ea_fsysCb_t *  cb,
			char *         err);

int32
ea_offVol(char *node,
			ea_fsysCb_t * cb,
			char *        err);

int32
ea_getClusterVolIdNode(evm_VolConfig_t *vol, char *node, char *err);

int32
ea_getVolId(char *              node,
			evm_VolConfig_t *   vol,
			char *              err);

int32
ea_getNodeVolIdFromVolName(char *volname, char *node, evm_volId_t *volId);

int32
ea_getVolId2(char *node, char *volname, evm_volId_t *volId);


int32
ea_getVolCfg(char *node, evm_VolConfig_t *vol);

ea_fsysList_t *
ea_getVolIdList(char *node);

ea_fsysList_t *
ea_getVolList(char *node);

int32
ea_crtFsys(char *node, ea_fsysCb_t * cb);

int32
ea_modFsys(char *node, ea_fsysCb_t * cb);

int32
ea_mntFsys(char *node, ea_fsysCb_t * cb);

int32
ea_umntFsys(char *node, ea_fsysCb_t * cb);

int32
ea_crtFset(char *node, ea_fsetCb_t * cb);

int32
ea_delFset(char *node, ea_fsetCb_t * cb);

int32
ea_setFsetQuota(char *node, ea_fsetCb_t * cb, ea_quotaCfg_t *quotaCfg);

int32
ea_onFsetQuota(char *node, ea_fsetCb_t * cb);

int32
ea_offFsetQuota(char *node, ea_fsetCb_t * cb);

ea_fsetList_t *
ea_getFsetIdList(char * node);

fs_MsgReportQuotaResp_t *
ea_fsetGetQuota(char *node, ea_fsetId_t *fsetId);

int32 ea_getFsysInfo (char *node,
					evm_volId_t volId, 
					uint32 *state,
                       uint64 *usedSize,
					   uint32 *highWaterMark);

evm_lvStrg_t *
ea_getVolLunList(char *node, evm_volId_t volId, uint32 *numStrg);

int32 
ea_fsysIsExported(char *node, evm_volId_t volId);

uint64 
ea_getFcPortWwn(char *node, uint32 slot, uint32 port);

/*----------------------------------------------------------------------------*/

#endif // _EA_H_
