/*
 *-----------------------------------------------------------------
 *
 * Name:	eee-poll.h
 *
 * RCS:  $Id: eee-poll.h,v 1.4 2002/09/27 22:31:25 rick Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	Polling routines
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/10/01
 *
 *-----------------------------------------------------------------
 */

#ifndef _EEE_POLL_H
#define _EEE_POLL_H

#define EEE_HIGH_PRIORITY		1
#define EEE_LOW_PRIORITY		2

#define POLLING_CAN_CONTINUE(x)	((x)--)

/*-----------------------------------------------------------------
 * Name:	eee_registerPollingFunc(priority, polling_func, cb)
 *
 * Description:	Polling functions will be called periodically.
 *				Drivers should use polling functions to process receive
 *				rings, transmit cleanup, etc.
 *				The polling function will be called with the user-supplied
 *				control block, and a timer reference value.  The timer
 *				reference will be utilized via some EEE layer macros.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/9/01
 *
 *-----------------------------------------------------------------
 */
int32						/* returns handle for polling function or
							 * EEE_ERR_POLLING_TABLE_FULL
							 */
eee_registerPollingFunc(
	uint32	priority,			/* EEE_HIGH_PRIORITY, EEE_LOW_PRIORITY */
	void	(*polling_func)(		/* function to be called as part of
									 * EEE layer polling loop
									 */
				void	* cb,		/* user defined control block for
									 * polling function
									 */
				uint32	timer_ref	/* some indication of how much time
									 * this routine can use the CPU, or
									 * how many packets to process in a
									 * loop - supplied by the EEE layer
									 */
			),
	void	* cb
);

/*-----------------------------------------------------------------
 * Name:	eee_unregisterPollingFunc(priority, handle)
 *
 * Description:	Basically provided for completeness.  A polling function
 *				can be removed from polling loop.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/9/01
 *
 *-----------------------------------------------------------------
 */
int32						/* returns NFX_OK,
							 * EEE_ERR_POLLING_FUNC_INVALID
							 */
eee_unregisterPollingFunc(
	int32	handle				/* from eee_registerPollingFunc()
								 * return code
								 */
);

/*-----------------------------------------------------------------
 * Name:	eee_poll()
 *
 * Description: Main polling loop
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/10/01
 *
 *-----------------------------------------------------------------
 */
#if !defined(_KERNEL)
void
eee_poll(uint32 num_loops);
#endif

#ifdef LINUX_TEST
/*-----------------------------------------------------------------
 * eee_addPollIndicationFd (int fd)
 *
 * In the linux EEE simulation environment, this additional
 * registration should be made by those modules that register
 * a polling function. The caller supplies a file descriptor that
 * indicates when the polling function needs to be called.  The
 * polling function is called when a read on the file descriptor
 * would not block (ie. there is data available).  This is a simple
 * mechanism that allows the EEE linux process to block in the
 * EEE kernel instead of running continuously and soaking up CPU
 * cycles mercilessly.  An example of its use would be that whenever
 * something is added to a queue that the polling function would
 * drain, the fd is written (typically the other end of a socket),
 * and the polling function reads it when removing the enqueued item.
 *
 * Registered polling functions that do not register an fd to be
 * watched will still have their registered functions called, but
 * the resulting delays may be unacceptable to the application.
 *-----------------------------------------------------------------
 */
void
eee_addPollIndicationFd(int fd);
#endif /* LINUX_TEST */

#endif	/* _EEE_POLL_H */

