/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : el-tcp-server.c
 *
 * Description  : Eagle Link (EL) on Eagle Vision.
 *
 * Eagle Link is the interface between Eagle Vision and the NFX.
 * In a first instance, the interface will be TCP/IP.
 * Eagle Vision will be the TCP server, whereas NFXs will act as TCP client.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

#include <stdio.h>      // fprintf(), stderr
#include <stdlib.h>     // exit()
#include <string.h>     // memset()
#include <unistd.h>     // close()
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#include "nfx-incl.h"

#include "el-tcp-common.h"
#include "exports/lnm-api.h"
#include "el-api.h"
#include "sim-ev.h"

/*------------------------------------------------------------------------------
 * Prototypes of local functions.
 *------------------------------------------------------------------------------
 */

static void
els_terminate(int sig);

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

int tcpFd;
static int listenfd;

/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : els_terminate()
 *
 * Description  : Signal handler which provides orderly termination.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
els_terminate(int sig)

{

    printf("close\n");
    
    /* Close sockets.
     */
    close(tcpFd);
    close(listenfd);
    
    exit(0);
}

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : main()
 *
 * Description  : 
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

int
main(int        argc,
     char **    argv)

{
    struct sockaddr_in servaddr;    /* incoming connection address */
    int maxfd;                      /* maximum fd (optimization for select) */
    int maxi;                       /* number of clients ??? */
    int client[FD_SETSIZE];         /* client connection array */
    fd_set allset;                  /* fd set we're interested in */
    int i;                          /* looping index */

    /* Sets the procedure to follow in case of request for program termination
     * or CTRL-C interrupt.
     */
    signal(SIGTERM, els_terminate);
    signal(SIGINT, els_terminate);

    /* Open a TCP socket (an Internet stream socket).
     */
    if ((listenfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
        perror("socket()");
        return(errno);
    }

    /* Bind our local address so that the client can send to us.
     */    
    bzero(&servaddr, sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servaddr.sin_port = htons(SERV_TCP_PORT);
  
    if (bind(listenfd, (struct sockaddr *) &servaddr, sizeof(servaddr)) < 0)
    {
        perror("bind()");
        return(errno);
    }

    /* Set socket to passive mode and ask kernel to buffer upto LISTENQ (inet.h)
     */
    if (listen(listenfd, LISTENQ) < 0)
    {
        perror("listen()");
        return(errno);
    }

    /* Initializations
     */
    maxfd = MAX(listenfd, fileno(stdin));  /* no connections yet; only listen to servaddr and stdin */
    maxi = -1;	     /* index into client[] array */
    
    /* Initialize client connection array (-1 indicates available entry) */
    for (i = 0; i < FD_SETSIZE; i++)
    {
        client[i] = -1;
        if (i == 0)
        {
            tcpFd = client[i];
        }
    } // for

    /* allset is a file descriptor set storing the fd's we're interested in */
    FD_ZERO(&allset);
    FD_SET(listenfd, &allset);
    FD_SET(fileno(stdin), &allset);

    printf("Waiting for connections on port %d\n", SERV_TCP_PORT);
  
    /* Loop forever (daemons never die!)
     */
    while (TRUE)
    {
        int nready;           /* number of file descriptors ready */
        fd_set rset = allset; /* copy allset (structure assignment) */

        /* Wait for one of the file descriptors in rset to be ready for reading
         * (no timeout).
         */
        if ((nready = select(maxfd+1, &rset, NULL, NULL, NULL)) < 1)
        {
            perror("select()");
            exit(errno);
        }

#ifndef ISABELLE
        printf("I received something on one of my sockets...\n");
#endif // ISABELLE

        /* Is this a new client connection ?
         */
        if (FD_ISSET(listenfd, &rset))
        {
            int clilen;
            int connfd;
            struct sockaddr_in cliaddr;
      
            /* Accept the client connection
             */
            clilen = sizeof(cliaddr);
            connfd= accept(listenfd, (struct sockaddr *) &cliaddr, &clilen);
            if (connfd < 0)
            {
                perror("accept()");
                exit(errno);
            }

            printf("New client: %s, port %d; Assigning fd#%d\n",
                   inet_ntoa(cliaddr.sin_addr),
                   ntohs(cliaddr.sin_port),
                   connfd);

            /* Save the client descriptor in the first available (-1) position
             * in the client array.
             */
            for (i = 0; i < FD_SETSIZE; i++)
            {
                if (client[i] < 0)
                {
                    client[i] = connfd;	/* save descriptor */
                    if (i == 0)
                    {
                        tcpFd = client[i];
                    }
                    break;
                }
            } // for

            if (i == FD_SETSIZE)
            {
                fprintf(stderr, "Error: too many clients");
                exit(i);
            }

            /* Add the new descriptor to set (maintain maxfd for select) (*/
            FD_SET(connfd, &allset);
            if (connfd > maxfd)
            {
                maxfd = connfd;
            }
            /* maintain the maximum index indicator for the client[] array */
            if (i > maxi)
            {
                maxi = i;
            }
            --nready; /* one file descriptor processed */
        }

        /* Is there data in standard in ?
         */
        if (FD_ISSET(fileno(stdin), &rset))
        {
            char line[512];
            
            if (fgets(line, 512, stdin) == NULL)
            {
                printf("error\n");
            }
            else
            {
                int argc = 0;
                char * argv[10];
                char * linetmp;
                char name[10] = "simev";

                argv[0] = name;
                
                linetmp = (char *) malloc(512);
                strcpy(linetmp, line);
                for (argc = 1; strcmp(linetmp, "") && argc < 10; argc++)
                {
                    argv[argc] = linetmp;
                    if (index(argv[argc], ' ') != NULL)
                    {
                        linetmp =  index(linetmp, ' ') + 1;
                        *index(argv[argc], ' ') = 0;
                    }
                    else if (index(argv[argc], '\n') != NULL)
                    {
                        linetmp = index(linetmp, '\n');
                        *index(argv[argc], '\n') = 0;
                    }

                } // for
                printf("argc = %d\n", argc);
                
                el_processCli(argc, argv);
            }
        }
        
        /* While there are remaining fds ready to read,
         * and clients in the array...
         */
        for (i = 0; ((nready > 0) && (i <= maxi)); i++)
        {
            int sockfd = client[i];
            if ((sockfd >= 0) && (FD_ISSET(sockfd, &rset)))
            {
                int rcvBufLen;
                char rcvBuf[MAXBUF];

                printf("message received from EA\n");

                /* Read received message.
                 */
                rcvBufLen = recvfrom(sockfd, (void *) &rcvBuf, MAXBUF,
                                     0, NULL, NULL);
                EL_ASSERT(rcvBufLen > 0);

                el_processEaMsg(rcvBuf, rcvBufLen);

                --nready; /* on file descriptor processed */
            }
        } // for
    } // while (TRUE)

    exit(0);
}

/*----------------------------------------------------------------------------*/
