// vi:ts=4
/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : event-as.c
 *
 * Description  : Email notification.
 *
 * Created by   : Isabelle Cnudde
 *				  Pankaj Gupta
 *
 * Date Created : 12-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

#ifdef SSC

#include "nfx-incl.h"

#include <stdio.h>      // fprintf(), stderr
#include <stdlib.h>     // exit()
#include <string.h>     // memset()
#include <unistd.h>     // close()

#include <fcntl.h>
#include <netinet/in.h>

#include <sys/time.h>
#include <sys/socket.h>
#include <signal.h>
#include <errno.h>

#include "exports/anpssc-api.h"
#include "exports/event-api.h"
#include "exports/ncm-api.h"
#include "exports/opt-api.h"
#include "exports/elog-api.h"
#include "exports/fs-api.h"
#include "event-as.h"

/*------------------------------------------------------------------------------
 * Constants and macros.
 *------------------------------------------------------------------------------
 */

#define MAILSENDER 		"autosupport-alpha@agilestorage.com"
#define AS_CRON_PROG		"/usr/local/agile/bin/ascron"
#define NFX_CRONTAB_FILE	"/var/cron/tabs/root"
#define NFX_CRONTAB_TMPFILE	"/var/cron/tabs/.root.tmp"

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */
extern char *eventTypeStr [];
extern char *eventStateStr [];
extern char *volEvtStr [];
extern int myappid;
extern event_stats_t eventStats;

/* 
 * E-mail configuration (autosupport).
 */
extern as_config_t   asCfg;

/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */
void as_addCrontabEntry(void);
void as_deleteCrontabEntry(void);

/*------------------------------------------------------------------------------
 * Name         : create_smtp_hdr()
 *
 * Description  : 
 *
 * Created by   : 
 *
 * Date Created : 10-SEP-2001
 *
 *------------------------------------------------------------------------------
 */
static void
create_smtp_hdr (FILE *f, char *to, char *subject)
{
	fprintf (f, "To: %s\n", to);
	fprintf (f, "From: %s\n", MAILSENDER);
	fprintf (f, "Subject: %s\n\n", subject);
}

/*------------------------------------------------------------------------------
 * Name         : ems_sendMail()
 *
 * Description  : Send mail.
 *
 * Created by   : Isabelle Cnudde
 *				  Pankaj Gupta
 *
 * Date Created : 10-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ems_sendMail(char * subject,
             char * to,
             char * body,
             char * attach)
{
	char *smtpheader;
	int hdr_fd;
	FILE *hdr_f;

    EVENT_DBG_PRINT(("> ems_sendMail()\n"));

    EVENT_DBG_PRINT(("asCfg.enable: %d\n", asCfg.enable));
    EVENT_DBG_PRINT(("to: %s\n", to));

   	if ((asCfg.enable == TRUE) && (strcmp(to, "")))
    {
        char    command[512];

		smtpheader = strdup (NFX_TMP_DIR"smtpXXXXX.hdr");
		hdr_fd = mkstemp (smtpheader);
		hdr_f = fdopen (hdr_fd, "w+");

		create_smtp_hdr (hdr_f, to, subject);

		fclose (hdr_f);
		close (hdr_fd);

        EVENT_DBG_PRINT(("Going to send email...\n"));
	
		sprintf (command, 
		        "/bin/cat %s %s %s | /usr/sbin/sendmail -f %s -i -- %s", 
		                   smtpheader, body, attach ? attach : "", 
		                   MAILSENDER, to);

		system (command);

		unlink (smtpheader);
		free (smtpheader);
    }
    else
    {
        EVENT_DBG_PRINT(("Didn't send email...\n"));
    }
    
    EVENT_DBG_PRINT(("< ems_sendMail()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : appAddDetails()
 *
 * Description  : This is the place where any event specific info can
 *                be added by applications/modules who want more
 *                info to be added in the auto support email.
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 10-SEP-2001
 *
 *------------------------------------------------------------------------------
 */
void
appAddDetails(FILE *fd, nfx_eventMsg_t *evt)
{
	switch(evt->eventType)
	{
		case NFX_EVENT_NODE:
		break;

		case NFX_EVENT_PCC:
		break;

		case NFX_EVENT_SCC:
		break;

		case NFX_EVENT_CARD:
		break;

		case NFX_EVENT_CPU:
		break;

		case NFX_EVENT_VOLUME:
		break;

		case NFX_EVENT_IP_IF:
		break;

		case NFX_EVENT_PORT:
		break;

		case NFX_EVENT_NFS:
		break;

		case NFX_EVENT_ROUTE:
		break;

        case NFX_EVENT_QUOTA:
       {
            struct quotaEvent *qevt = &evt->param.quotaEvent;

            fprintf(fd, "\tVolume name: %s\n", qevt->volName);
            fprintf(fd, "\tFileset id : %d\n", qevt->fileSetId);
            switch (qevt->quotaType)    {
                case FS_QUOTA_USER:
                    fprintf(fd, "\tUser id    : %d\n", qevt->id);
                    break;
                    
                case FS_QUOTA_GROUP:
                    fprintf(fd, "\tGroup id   : %d\n", qevt->id);
                    break;
            }
            if (qevt->notify & FS_EXCEED_BLOCK) {
                fprintf(fd, "\tSize       : ");
            } else {
                fprintf(fd, "\tFiles      : ");
            }
            if (qevt->notify & FS_EXCEED_HARDLIMIT) {
                fprintf(fd, "exceeded Hardlimit ");
            }
            if (qevt->notify & FS_EXCEED_SOFTLIMIT) {
                fprintf(fd, "exceeded Softlimit ");
            }
            if (qevt->notify & FS_EXCEED_TIMELIMIT) {
                fprintf(fd, "exceeded Timelimit ");
            }
            fprintf(fd, "\n");

            break;
        }
        
		case NFX_EVENT_MKFS:
		break;

		default:
    		EVENT_ERR_PRINT(("appAddDetails: Unknown event type %d\n",
							evt->eventType));
	}
}

/*------------------------------------------------------------------------------
 * Name         : ems_sendAlarmMail()
 *
 * Description  : Send event that require corrective action from the operator.
 *
 * - send email with the event and all configuration info.
 * - send short email with a summary of the event.
 *
 * Created by   : Isabelle Cnudde
 *				  Pankaj Gupta
 *
 * Date Created : 10-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

int
ems_sendAlarmMail(nfx_eventMsg_t * evt, char *app)
{
    char    tmpfile[100];
    char    subject[100];
	int		tfd;
    FILE *  fd;

	if (asCfg.enable == FALSE)
	{
		EVENT_DBG_PRINT(("Autosupport not enabled\n"));
		return NFX_ERR;
	}

    EVENT_DBG_PRINT(("> ems_sendAlarmMail()\n"));

    /* 
	 * We can filter out here to send email only
	 * for certain events and state
     */

	if (evt)
		sprintf(subject, "NFX %s event", eventTypeStr[evt->eventType]);
	else
		sprintf(subject, "General AutoSupport Report");

    EVENT_DBG_PRINT(("subject is %s\n", subject));

    /* 
	 * Create temporary file.
     */
    strcpy(tmpfile, NFX_TMP_DIR);
    strcat(tmpfile, "email_XXXXXX");
    tfd = mkstemp(tmpfile);
    EVENT_DBG_PRINT(("tmpfile = %s\n", tmpfile));
    fd = fdopen(tfd, "w+");
    if (fd != NULL)
    {
        /* 
		 * Overwrite the file.
         */
        fprintf(fd, "Autosupport mail:\n");
		if (evt)
		{
        	fprintf(fd, "\tNFX Node   : %s\n", evt->nodeName);
        	fprintf(fd, "\tEvent      : %s\n", eventTypeStr[evt->eventType]);
        	fprintf(fd, "\tApplication: %s\n", app);
        	fprintf(fd, "\tMessage    : %s\n", evt->reason);
			appAddDetails(fd, evt);
        	fprintf(fd, "-----\n");
		}

		/*
		 * TBD: Add event statistics here
		 */

        fprintf(fd, "-----\n");
        fprintf(fd, "THIS MAIL HAS BEEN AUTOMATICALLY GENERATED.\n");
        fprintf(fd, "PLEASE DO NOT REPLY.\n");
        fprintf(fd, "-----\n");


        /* Closing file should flush buffers
         */
        fclose(fd);
        
    	/* 
		 * Send email with the configuration file attached.
     	*/
    	ems_sendMail(subject, asCfg.to, tmpfile, NFX_CONFIG_FILE);
    
    	/* 
		 * Send a short note email.
     	*/
    	ems_sendMail(subject, asCfg.noteto, tmpfile, NULL);

    }
    else
    {
        perror("fopen() failed\n");
    }
	close (tfd);

#ifdef DEBUG
{
static char cmdBuf[128];
	printf("Email start >>>>>>>>>>>>>>>>>>\n");
	sprintf(cmdBuf, "cat %s", tmpfile);
	system(cmdBuf); 
	printf("Email end   >>>>>>>>>>>>>>>>>>\n");
}    
#endif

   	/* 
	 * Delete the temporary file.
	 */
	unlink (tmpfile);

    EVENT_DBG_PRINT(("< ems_sendAlarmMail()\n"));
    return NFX_OK;
}

/*------------------------------------------------------------------------------
 * Name         : ems_logEvent()
 *
 * Description  : Put the event info in local auto support log file
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 
 *
 *------------------------------------------------------------------------------
 */

void
ems_logEvent(nfx_eventMsg_t *evt)
{
	struct timeval now;
	struct in_addr in1;
	struct in_addr in2;

	EVENT_DBG_PRINT(("> ems_logEvent()\n"));

	switch(evt->eventType)
	{
		case NFX_EVENT_NODE:
		case NFX_EVENT_PCC:
		case NFX_EVENT_SCC:
			in1.s_addr = evt->param.nodeEvent.mgmt_ip[0];
			in2.s_addr = evt->param.nodeEvent.mgmt_ip[1];
            E_LOG(class_3, info_s, myappid, 0, 0, 0,
                  ("Process-EVENT %s: Mgmt Port %s %s State %s", 
				   eventTypeStr[evt->eventType],
				   strdup((char *)inet_ntoa(in1)), 
				   strdup((char *)inet_ntoa(in2)),
				   eventStateStr[evt->param.nodeEvent.state]
				  ));
			break;

		case NFX_EVENT_CARD:
            E_LOG(class_3, info_s, myappid, 0, 0, 0,
                  ("Process-EVENT %s: Slot %d Type %d State %s", 
				   eventTypeStr[evt->eventType],
				   evt->param.cardEvent.slot,
				   evt->param.cardEvent.cardType,
				   eventStateStr[evt->param.cardEvent.state]
				  ));
			break;
		
		case NFX_EVENT_CPU:
            E_LOG(class_3, info_s, myappid, 0, 0, 0,
                  ("Process-EVENT %s: Slot %d CPU %d State %s", 
				   eventTypeStr[evt->eventType],
				   evt->param.cpuEvent.slot,
				   evt->param.cpuEvent.cpu,
				   eventStateStr[evt->param.cpuEvent.state]
				  ));
			break;


		case NFX_EVENT_VOLUME:
            E_LOG(class_3, info_s, myappid, 0, 0, 0,
                  ("Process-EVENT %s: %s (%016llx) Event %s State %s", 
				   eventTypeStr[evt->eventType],
				   strlen(evt->param.volEvent.volName)?
				   		evt->param.volEvent.volName:" ",
				   evt->param.volEvent.volId.whole_thing,
				   volEvtStr[evt->param.volEvent.event],
				   eventStateStr[evt->param.volEvent.state]
				  ));
			break;

		case NFX_EVENT_IP_IF:
			in1.s_addr = evt->param.ipIfEvent.ipa;
            E_LOG(class_3, info_s, myappid, 0, 0, 0,
                  ("Process-EVENT %s: IP %s Port %s State %s", 
				   eventTypeStr[evt->eventType],
				   inet_ntoa(in1),
				   evt->param.ipIfEvent.portName,
				   eventStateStr[evt->param.ipIfEvent.state]
				  ));
			break;

		case NFX_EVENT_PORT:
            E_LOG(class_3, info_s, myappid, 0, 0, 0,
                  ("Process-EVENT %s: Slot %d CPU %d Id %d (0x%x) State %s", 
				   eventTypeStr[evt->eventType],
				   EEE_GET_SLOT_NUM(evt->param.portEvent.portAddr),
				   EEE_GET_DEST_CPU(evt->param.portEvent.portAddr),
				   EEE_GET_PORT_ID(evt->param.portEvent.portAddr),
				   evt->param.portEvent.portAddr,
				   eventStateStr[evt->param.portEvent.state]));
			break;

		case NFX_EVENT_NFS:
			in1.s_addr = evt->param.exportEvent.expAddr;
            E_LOG(class_3, info_s, myappid, 0, 0, 0,
                  ("Process-EVENT %s: %s (%016llx) IP %s state %s", 
				   eventTypeStr[evt->eventType],
				   evt->param.exportEvent.expPath,
				   evt->param.exportEvent.volId.whole_thing,
				   inet_ntoa(in1),
				   eventStateStr[evt->param.exportEvent.state]
				  ));
			break;

		case NFX_EVENT_ROUTE:
			in1.s_addr = evt->param.ipRteEvent.destNet;
			in2.s_addr = evt->param.ipRteEvent.gateWay;
            E_LOG(class_3, info_s, myappid, 0, 0, 0,
                  ("Process-EVENT %s: Net %s Mask 0x%x Gateway %s State %s", 
				   eventTypeStr[evt->eventType],
				   strdup((char *)inet_ntoa(in1)), 
				   evt->param.ipRteEvent.destMask,
				   strdup((char *)inet_ntoa(in2)),
				   eventStateStr[evt->param.ipRteEvent.state]
				  ));
			break;
			break;

		case NFX_EVENT_QUOTA:
            E_LOG(class_3, info_s, myappid, 0, 0, 0,
                  ("Process-EVENT %s: %s (%016llx) FsetId 0x%x Type %d Id %d Notify %d", 
				   eventTypeStr[evt->eventType],
				   evt->param.quotaEvent.volName,
				   evt->param.quotaEvent.volId.whole_thing,
				   evt->param.quotaEvent.fileSetId,
				   evt->param.quotaEvent.quotaType,
				   evt->param.quotaEvent.id,
				   evt->param.quotaEvent.notify));
			break;

		case NFX_EVENT_MKFS:
            E_LOG(class_3, info_s, myappid, 0, 0, 0,
                  ("Process-EVENT %s: Vol id %016llx", 
				   eventTypeStr[evt->eventType],
				   evt->param.mkfsEvent.volId.whole_thing));
			break;


		default:
			E_LOG(class_3, info_s, myappid, 0, 0, 0,
              ("Process-EVENT %s:", eventTypeStr[evt->eventType]));
			return;
	}

	/*
	 * Update event statistics
	 */
	gettimeofday(&now, NULL);
	eventStats.eventRcvd[evt->eventType]++;
	eventStats.lastEventTime[evt->eventType] = now.tv_sec;

	EVENT_DBG_PRINT(("< ems_logEvent()\n"));
	return;
}

/*------------------------------------------------------------------------------
 * Name         : ems_sendTrap()
 *
 * Description  : Send SNMP trap.
 *
 * Created by   : Isabelle Cnudde
 *				  Pankaj Gupta
 *
 * Date Created : 10-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

void
ems_sendTrap(nfx_eventMsg_t *evt)
{
	EVENT_DBG_PRINT(("> ems_sendTrap()\n"));
	
	EVENT_DBG_PRINT(("ems_sendTrap Not yet implemented!\n"));

	EVENT_DBG_PRINT(("< ems_sendTrap()\n"));
	return;
}


/*------------------------------------------------------------------------------
 * Name         : as_addCrontabEntry()
 *
 * Description  : Add an entry in the cron tab file
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 18-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
void
as_addCrontabEntry(void)
{
	char cronCmdLine[1024];
	char lclBuf[256];
	int32 needComma = 0;
	int32 idx;
	FILE *fd;

	/*
	 * Open the crontab file in append mode
	 */
	fd = fopen (NFX_CRONTAB_FILE, "a");
	if (!fd)
	{
		EVENT_ERR_PRINT((
			"as_addCrontabEntry: Could not open %s for cron config", \
					NFX_CRONTAB_FILE));
		return;
	}

	/*
	 * Prepare the cron tab entry in correct format
	 */

	memset(cronCmdLine, 0, sizeof(cronCmdLine));
	memset(lclBuf, 0, sizeof(lclBuf));

	/* 
	 * Minute field 
	 */
	needComma = 0;
	if (asCfg.minBitMap)
	{
		for (idx = 0; idx <= 59; idx++)
		{
			if (GET_BIT(asCfg.minBitMap, idx))
			{
				if (needComma)
					strcat(cronCmdLine, ",");
				sprintf(lclBuf, "%d", idx);
				strcat(cronCmdLine, lclBuf);
				needComma = 1;
			}
		}
	}
	else
	{
		strcat(cronCmdLine, "*");
	}
	strcat(cronCmdLine, "\t");

	/* 
	 * Hour field 
	 */
	needComma = 0;
	if (asCfg.hourBitMap)
	{
		for (idx = 0; idx <= 23; idx++)
		{
			if (GET_BIT(asCfg.hourBitMap, idx))
			{
				if (needComma)
					strcat(cronCmdLine, ",");
				sprintf(lclBuf, "%d", idx);
				strcat(cronCmdLine, lclBuf);
				needComma = 1;
			}
		}
	}
	else
	{
		strcat(cronCmdLine, "*");
	}
	strcat(cronCmdLine, "\t");

	/* 
	 * Date field 
	 */
	needComma = 0;
	if (asCfg.dateBitMap)
	{
		for (idx = 1; idx <= 31; idx++)
		{
			if (GET_BIT(asCfg.dateBitMap, idx))
			{
				if (needComma)
					strcat(cronCmdLine, ",");
				sprintf(lclBuf, "%d", idx);
				strcat(cronCmdLine, lclBuf);
				needComma = 1;
			}
		}
	}
	else
	{
		strcat(cronCmdLine, "*");
	}
	strcat(cronCmdLine, "\t");

	/* 
	 * Month field 
	 */
	needComma = 0;
	if (asCfg.monthBitMap)
	{
		for (idx = 1; idx <= 12; idx++)
		{
			if (GET_BIT(asCfg.monthBitMap, idx))
			{
				if (needComma)
					strcat(cronCmdLine, ",");
				sprintf(lclBuf, "%d", idx);
				strcat(cronCmdLine, lclBuf);
				needComma = 1;
			}
		}
	}
	else
	{
		strcat(cronCmdLine, "*");
	}
	strcat(cronCmdLine, "\t");

	/* 
	 * Day field 
	 */
	needComma = 0;
	if (asCfg.dayBitMap)
	{
		for (idx = 0; idx <= 7; idx++)
		{
			if (GET_BIT(asCfg.dayBitMap, idx))
			{
				if (needComma)
					strcat(cronCmdLine, ",");
				sprintf(lclBuf, "%d", idx);
				strcat(cronCmdLine, lclBuf);
				needComma = 1;
			}
		}
	}
	else
	{
		strcat(cronCmdLine, "*");
	}
	strcat(cronCmdLine, "\t");

	/*
	 * Put program name
	 */
	strcat(cronCmdLine, AS_CRON_PROG);

	/*
	 * Write entry in the cron tab file
	 */
	fprintf(fd, "%s\n", cronCmdLine);
	fclose(fd);
}

/*------------------------------------------------------------------------------
 * Name         : as_deleteCrontabEntry()
 *
 * Description  : Update the cron tab file
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 18-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
void
as_deleteCrontabEntry(void)
{
	char cronCmdLine[1024];
	char patternBuf[256];
	char lclBuf[256];
	char cmdBuf[256];
	FILE *fd;
	FILE *fdTmp;

	EVENT_DBG_PRINT(("Enter as_deleteCrontabEntry"));
	fd = fopen (NFX_CRONTAB_FILE, "r");
	if (!fd)
	{
		EVENT_ERR_PRINT((
			"as_deleteCrontabEntry: Could not open %s for cron config",
					NFX_CRONTAB_FILE));
		return;
	}

	fdTmp = fopen (NFX_CRONTAB_TMPFILE, "w");
	if (!fdTmp)
	{
		EVENT_ERR_PRINT((
			"as_deleteCrontabEntry: Could not open %s for cron config",
					NFX_CRONTAB_TMPFILE));
		return;
	}

	memset(cronCmdLine, 0, sizeof(cronCmdLine));
	memset(patternBuf, 0, sizeof(patternBuf));
	memset(lclBuf, 0, sizeof(lclBuf));
	memset(cmdBuf, 0, sizeof(cmdBuf));

	/*
	 * Put program name
	 */
	strcat(patternBuf, AS_CRON_PROG);

	/*
	 * Keep reading each line and writing it back
	 * Remove the line which has pattern matching for 
	 * auto support entry to be deleted
	 */
	 while(fgets(cronCmdLine, 1024, fd) != NULL)
	 {
		if (strstr(cronCmdLine, patternBuf) != NULL)
			continue;
		fputs(cronCmdLine, fdTmp);
		memset(cronCmdLine, 0, sizeof(cronCmdLine));
	 }

	 fclose(fd);
	 fclose(fdTmp);

	 sprintf(cmdBuf, "mv %s %s", NFX_CRONTAB_TMPFILE, NFX_CRONTAB_FILE);
	 system(cmdBuf);
}



/*----------------------------------------------------------------------------*/

#endif	/* SSC */
