/*------------------------------------------------------------------------------
 * RCS:  $Id: evm-api.c,v 1.48 2002/12/10 23:30:39 isabellec Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name     : evm-utils.c
 *
 * Description  : functions that are available to other modules
 *                eg init, EEE rcv message for evm, etc
 *
 * Created by   : Lan Chin
 *                Isabelle Cnudde
 *
 * Date Created : 28-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

#include "nfx-incl.h"
#include "exports/debug.h"
#include "exports/hash-api.h"
#include "evm-lv.h"
#include "evm-hash.h"
#include "evm-debug.h"
#include "evm-io.h"
#include "evm-queue.h"

/*------------------------------------------------------------------------------
 * Macros.
 *------------------------------------------------------------------------------
 */

#define MIN_PACKET_LEN (64 - sizeof(enet_hdr_t) - sizeof(agile_hdr_t) - 4)
NFX_ASSERT_STATIC(MIN_PACKET_LEN == 28);

#define EXPECTED_PACKET_LEN(_type_) \
    ((sizeof(_type_) >= MIN_PACKET_LEN)? sizeof(_type_) : MIN_PACKET_LEN)


/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

evm_param_t evmParam;
evm_stats_t evm_stats;

uint32      evmSrvrPort;

int32       evmAppId;
int32       sacAppId;
int32       sbmAppId;
int32       eaAppId;

#ifdef SY_IO_PERFORMANCE_TEST
int32		perfAppId;
extern void evm_perf_rcvMsg(eee_descPtr_t edesc);
#endif

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : evm_rcvMsg
 *
 * Description  : Receives unicast messages sent to evm.
 *
 * Created by   : Greg Seibert
 *                Lan Chin
 *                Isabelle Cnudde
 *
 * Date Created : 14-AUG-2001
 *
 *------------------------------------------------------------------------------
 */

void
evm_rcvMsg (eee_descPtr_t edesc)
    
{
	uint32          msgLen = 0;
	char *          msgBuf = NULL;
	uint32 *        msgType;
	evm_MsgStat_t   stat_msg;
    eee_descBuf_t * bd = NULL;
    boolean         freeMsgBuf = TRUE;
    int32           rc, numBufs;
    uint32          srcPort, destPort;

    numBufs = EEE_GET_NUM_BUFS(edesc);
    srcPort = edesc->hdr.src_port;
    destPort = edesc->hdr.dest_port;
    if (numBufs == 1)
    {
        // Single buffer message, don't copy to flat buffer.
        if (edesc->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER)
            bd = &(edesc->bd[0]);
        else
            bd = &(edesc->bd[1]);
        msgBuf = bd->buf;
        msgLen = bd->len;
        
        // Done with the descriptor.
        eee_deallocateDesc(edesc);
    }
    else
    {
        // Copy the multi-buffer message to a
        // contiguous malloc buffer and free the
        // received packet.
        msgLen = EEE_GET_PKT_LEN(edesc);
        EVM_RAM_ALLOC(msgBuf, char, msgLen);
        if (msgBuf == NULL)
        {
            evm_trace(EVM_TRACE_ERRORS, "No memory, dropping packet.\n");
            eee_freePacket(edesc);
            return;
        }
        rc = eee_copyPacketToLocalBuffer(msgBuf,
                                         edesc,
                                         msgLen,
                                         0);
        if (rc != msgLen)
        {
            evm_trace(EVM_TRACE_ERRORS, "eee_copyPacketToLocal() failed.\n");
            eee_freePacket(edesc);
            return;
        }

        // Done with the packet.
        eee_freePacket(edesc);
    }

	msgType = (int *)msgBuf;
    
	bzero(&stat_msg, sizeof(stat_msg));
	stat_msg.hdr.msgType = EVM_MSG_STAT;
	stat_msg.hdr.clientPrivate = ((evm_msgHdr_t *)msgBuf)->clientPrivate;
	stat_msg.stat = NFX_OK;

	switch (*msgType)
    {
    case EVM_LVACT_CREATELV:
    {
        evm_MsgCreateLV_t * create_msg;

        NFX_ASSERT(msgLen >= EXPECTED_PACKET_LEN(evm_MsgCreateLV_t));
        
        create_msg = (evm_MsgCreateLV_t *)msgBuf;
        evm_trace(EVM_TRACE_SRVR_MSG,
                  "EVM SRVR: EVM_LVACT_CREATELV for VolId %llx",
                  create_msg->lVol.whole_thing);
        NFX_SPIN_LOCK(&evmVolLock);
        if (evm_lvExists(create_msg->lVol, NULL))
        {
            NFX_SPIN_UNLOCK(&evmVolLock);
            stat_msg.stat = EVM_VOL_ALREADY_EXIST;
        }
        else
        {
            NFX_SPIN_UNLOCK(&evmVolLock);
            if (evm_createVol(create_msg) == NFX_ERR)
            {
                stat_msg.stat = NFX_ERR;
            }
        }
        eee_sendMessage(&stat_msg, sizeof(stat_msg), srcPort, destPort);
        break;
    } // case EVM_LVACT_CREATELV
    
    case EVM_LVACT_ADDSTRG:
    {
        evm_MsgAddStrg_t    *add_msg;
        
        NFX_ASSERT(msgLen == EXPECTED_PACKET_LEN(evm_MsgAddStrg_t));
        
        add_msg = (evm_MsgAddStrg_t *) msgBuf;
        evm_trace(EVM_TRACE_SRVR_MSG,
                  "EVM SRVR: EVM_LVACT_ADDSTRG VolId %llx",
                  add_msg->lVol.whole_thing);
        evm_trace(EVM_TRACE_DETAIL,
                  "EVM SRVR DETAIL: for ADDSTRG VolId %llx auSize %d",
                  add_msg->lVol.whole_thing, add_msg->strg.auSize);
        evm_trace(EVM_TRACE_DETAIL,
                  "                 scsi %llx startBlk %d",
                  add_msg->strg.strgId, add_msg->strg.startBlk);
        if (evm_addLun(add_msg) == NFX_ERR)
        {
            stat_msg.stat = NFX_ERR;
        }
        eee_sendMessage (&stat_msg, sizeof(stat_msg), 
                         srcPort, destPort);
        break;
    }
    
    case EVM_LVACT_OFFLINELV:
    {
        evm_MsgOffLinLV_t   *off_msg;
        
        NFX_ASSERT(msgLen == EXPECTED_PACKET_LEN(evm_MsgOffLinLV_t));
        
        off_msg = (evm_MsgOffLinLV_t *) msgBuf;
        evm_trace(EVM_TRACE_SRVR_MSG,
                  "EVM SRVR: EVM_LVACT_OFFLINELV VolId %lld",
                  off_msg->lVol.whole_thing);
        if (evm_offlineLV(off_msg->lVol, TRUE) == NFX_ERR)
        {
            stat_msg.stat = NFX_ERR;
        }
        eee_sendMessage(&stat_msg, sizeof(stat_msg), srcPort, destPort);
        break;
    }

    case EVM_LVACT_VOLLIST_REQ:
    {
        evm_MsgReqVolList_t *   req = (evm_MsgReqVolList_t *) msgBuf;
        uint32                  rspSize;
        evm_MsgRspVolList_t *   rsp = NULL;
        uint32                  nbVol = 0;
        evm_volId_t *           ids;

        NFX_ASSERT(msgLen == EXPECTED_PACKET_LEN(evm_MsgReqVolList_t));
        
        evm_trace(EVM_TRACE_SRVR_MSG, "EVM SRVR: EVM_REQ_VOL_LIST");

        EVM_RAM_ALLOC(ids, evm_volId_t, MAX_IDS * sizeof(evm_volId_t));
        nbVol = evm_getListVol(ids);

        rspSize = sizeof(evm_MsgRspVolList_t) +
            (((nbVol == 0)? 0 : (nbVol - 1)) * sizeof(evm_volId_t));
        EVM_RAM_ALLOC_ZERO(rsp, evm_MsgRspVolList_t, rspSize);
        
        rsp->hdr.msgType = EVM_LVACT_VOLLIST_RSP;
        rsp->hdr.clientPrivate = req->hdr.clientPrivate;
        rsp->nbVol = nbVol;
        memcpy((void *) rsp->ids, (const void *) ids,
               nbVol * sizeof(evm_volId_t));
        eee_sendMessage(rsp, rspSize, srcPort, destPort);

        EVM_RAM_DEALLOC(rsp);
        EVM_RAM_DEALLOC(ids);
        break;
    }
    
    case EVM_STAT_REQ:
    {
        evm_MsgStatReq_t *  req = (evm_MsgStatReq_t *) msgBuf;
        evm_MsgStatRsp_t    rsp;
        
        NFX_ASSERT(msgLen == EXPECTED_PACKET_LEN(evm_MsgStatReq_t));
        evm_trace(EVM_TRACE_SRVR_MSG, "EVM SRVR: EVM_STAT_REQ");
        rsp.hdr.msgType = EVM_STAT_RSP;
        rsp.hdr.clientPrivate = req->hdr.clientPrivate;
        NFX_SPIN_LOCK(&evmStatLock);
		evm_stats.count_per_usec = count_per_usec;
        memcpy(&(rsp.stats), &evm_stats , sizeof(evm_stats_t));
        if (req->reset == TRUE)
        {
            bzero((void *) &evm_stats,
                  sizeof(evm_stats) - sizeof(evm_queueCnt_t));
        }
        NFX_SPIN_UNLOCK(&evmStatLock);
        eee_sendMessage (&rsp, sizeof(evm_MsgStatRsp_t), srcPort, destPort);
        break;
    }
    
    /* I/O 
     */
    
    case SCSI_MID_READ_RSP:
    {
        NFX_ASSERT(msgLen >= EXPECTED_PACKET_LEN(scsi_readRspMsg_t));
        
        evm_trace(EVM_TRACE_DETAIL,
                  "EVM DETAIL: SCSI_MID_READ_RSP received");
        
        evm_io_scsiReadRsp(msgBuf, msgLen, srcPort);

        // Don't free the msgBuf, we'll pass it on to the client.
        freeMsgBuf = FALSE;
        NFX_ASSERT(numBufs == 1); // Client assumes its a system buf, not
                                  // a malloc buf.
        break;
    } // case SCSI_MID_READ_RSP
    
    case SCSI_MID_WRITE_RSP:
    {
        scsi_writeRspMsg_t *wp = (scsi_writeRspMsg_t *) msgBuf;

        NFX_ASSERT(msgLen == EXPECTED_PACKET_LEN(scsi_writeRspMsg_t));
        evm_trace(EVM_TRACE_DETAIL,
                  "EVM DETAIL: SCSI_MID_WRITE_RSP received");
        evm_io_scsiWriteRsp(wp, srcPort);
        
        break;
    } // case SCSI_MID_WRITE_RSP
    
    case SCSI_MID_ERR_RSP:
    {
        scsi_errRspMsg_t *  ep = (scsi_errRspMsg_t *) msgBuf;
         
        NFX_ASSERT(msgLen == EXPECTED_PACKET_LEN(scsi_errRspMsg_t));
        
        evm_trace(EVM_TRACE_ERRORS,
                  "EVM ERROR: SCSI_MID_ERR_RSP eCode %d cHandle %d",
                  ep->eCode, ep->hdr.cHandle);

        evm_io_scsiErrRsp(ep);
        
        break;
    } // case SCSI_MID_ERR_RSP
  
    case EVM_MSG_STAT:
    {
        evm_trace(EVM_TRACE_DETAIL, "EVM DETAIL: EVM_MSG_STAT");
        // ignore message
        break;
    } // case EVM_CONFIG_NTFY

    default:
    {
        evm_trace(EVM_TRACE_ERRORS, "EVM ERROR: "
                  "unknown message type %d from 0x%x to 0x%x",
                  *msgType, srcPort, destPort);
    }
	} // switch
    

    if (freeMsgBuf)
    {
        if (numBufs == 1)
        {
            EVM_DEALLOC_BUF(msgBuf);
        }
        else
        {
            EVM_RAM_DEALLOC(msgBuf);
        }
    }
}

/*-----------------------------------------------------------------
 * Name: evm_rcvMulticast
 *
 * Description: Receives multicast messages sent to evm.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 14 Aug 2001
 *
 *-----------------------------------------------------------------
 */

void
evm_rcvMulticast(eee_descPtr_t desc)
{
    return;
} /* evm_rcvMulticast */

/*------------------------------------------------------------------------------
 * Name: evm_IOReq
 *
 * Description: Is the client's interface for logical volume I/O.
 *
 * Created by:  Greg Seibert
 *
 * Date Created: 23 Aug 2001
 *
 *------------------------------------------------------------------------------
 */

int32
evm_IOReq (evm_IOReq_t * req,
           void (*cmplt)(evm_IOStat_t *stat))
{
    return(evm_io_req(req, cmplt, FALSE));
} /* evm_IOReq */

/*------------------------------------------------------------------------------
 * Name         : evm_IOMap
 *
 * Description  : Map a logical IO to a physical one.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 16-JAN-2003
 *
 *------------------------------------------------------------------------------
 */

int32
evm_IOMap(evm_volId_t       volId,          // input
          uint64            volOffset,      // input
          uint32            volLen,         // input (nb 512bytes blocks)
          scsi_id_t *       lunId,          // output
          int32 *           lunHandle,      // output
          uint32 *          lunOffset,      // output
          uint32 *          lunLen)         // output (nb 512bytes blocks)
    
{
    evm_vol_t * vol = NULL;
    evm_lun_t * lun = NULL;
    
    vol = evm_getVol(volId);
    if (vol == NULL) return(NFX_ERR);
    
    lun = evm_getPhys(volOffset, volLen, vol, lunOffset, lunLen, FALSE);
    if (lun == NULL) return(NFX_ERR);
    
    *lunId = lun->id;
    *lunHandle = lun->openHandle;
 
    return(NFX_OK);
}

/*------------------------------------------------------------------------------
 * Name         : evm_receiveEvent
 *
 * Description  : Receives events.
 *
 * Created by   : Isabelle
 *
 * Date Created : 25-SEP-2002
 *
 *------------------------------------------------------------------------------
 */

void
evm_receiveEvent(nfx_eventMsg_t * evt)

{
    switch(evt->eventType)
    {
    case NFX_EVENT_PCC:
    {
        if (evt->param.nodeEvent.state == EVENT_STATE_DOWN)
        {
            evm_startLookingForTakeover();
        }
        else
        {
            evm_stopLookingForTakeover();
        }
        break;
    } // case NFX_EVENT_PCC
    
    case NFX_EVENT_CPU:
    {
        /* Only care about SSC events.
         */
        if ((evt->param.cpuEvent.slot == 0) &&
            (evt->param.cpuEvent.cpu == 0))
        {
            if (evt->param.cpuEvent.state == EVENT_STATE_DOWN)
            {
                evm_startLookingForTakeover();
            }
            else
            {
                // do nothing
            }
        }
        break;
    } // case NFX_EVENT_CPU
    
    default:
    {
        evm_trace(EVM_TRACE_ERRORS,
                  "Received an event (%d) for which it didn't subscribed!\n",
                  evt->eventType);
        NFX_ASSERT(FALSE);
    }
    } // switch

    return;
}

/*------------------------------------------------------------------------------
 * Name         : evm_postEvt
 *
 * Description  : Send an event.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 29-MAY-2002
 *
 *------------------------------------------------------------------------------
 */

void
evm_postEvt(uint32      event,
            evm_volId_t volId,
            int			state)
    
{
    nfx_eventMsg_t      evtMsg;
    
    /* Fill event.
     */
    bzero((void *) &evtMsg, sizeof(nfx_eventMsg_t));
    evtMsg.msgType = NFX_EVENT_POST;
    //evtMsg.nodeName;
    evtMsg.needAck = FALSE;
    //evtMsg.status;
    evtMsg.eventScope = NFX_EVENT_LOCAL;
    evtMsg.eventType = NFX_EVENT_VOLUME;
    //eventType.reason;
    evtMsg.param.volEvent.event = event;
    evtMsg.param.volEvent.volId = volId;
    evtMsg.param.volEvent.state = state;
    
    /* Post event.
     */
    event_postEvent(&evtMsg);

    return;
}

/*------------------------------------------------------------------------------
 * Name         : evm_initModule
 *
 * Description  : Called by EEE to get evm_srvr ready to go
 *
 * Created by   : Greg Seibert
 *                Lan Chin
 *                Isabelle Cnudde
 *
 * Date Created : 14-AUG-2001
 *
 *------------------------------------------------------------------------------
 */

int32
evm_initModule(void)
    
{
    int32   rc = NFX_OK;

    evm_trace(EVM_TRACE_FUNC, "EVM CALL: evm_initModule()\n");

    /* Initialize EVM tunables.
     */
	eee_msToRTC(EVM_IO_TIMEOUT, &evmParam.ioTimeout);
    evmParam.maxIoRetry = EVM_MAX_IO_RETRY;
    evmParam.maxIo = EVM_MAX_IO;
    
    /* Register our message handler.
     */
    evmAppId = eee_allocateApplicationId ("evm");
    if (evmAppId == EEE_ERR_INVALID_APP_NAME)
    {
        goto evm_initModule_err;
    }
    rc = eee_registerReceiveFunc (evmAppId, evm_rcvMsg, evm_rcvMulticast);
    if (rc != NFX_OK)
    {
        goto evm_initModule_err;
    }
   
    /* Register to events.
     */
    event_registerClient(NFX_EVENT_PCC, evm_receiveEvent);
    event_registerClient(NFX_EVENT_CPU, evm_receiveEvent);
    
#ifdef SY_IO_PERFORMANCE_TEST
    perfAppId = eee_allocateApplicationId ("evm_perf");
    if (perfAppId == EEE_ERR_INVALID_APP_NAME)
        evm_trace(EVM_TRACE_ERRORS, "EVM ERR: unable to register evm_perf");

    rc = eee_registerReceiveFunc(perfAppId, evm_perf_rcvMsg, evm_rcvMulticast);
    if (rc != NFX_OK)
        evm_trace(EVM_TRACE_ERRORS,
                  "EVM ERR: unable to register evm_perf recv function");
#endif // SY_IO_PERFORMANCE_TEST

    /* Initialize the EVM server port address.
     */
    evmSrvrPort = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
                                            nfxMyCPUId,
                                            evmAppId,
                                            EEE_MGMT, 
                                            NFX_COS_HIGH, 
                                            PORT_BEST_FIT);
    
    /* Initialize spin locks for the global EVM statistics.
     */
    NFX_SPIN_INIT(&evmStatLock, SPINLOCK_LEVEL_UNKNOWN);

    /* Get application ids of SAC, SBM and EA.
     */
    sacAppId = eee_getApplicationId("sac");
    sbmAppId = eee_getApplicationId("sbm");
    eaAppId = eee_getApplicationId("ea");

    /* Initialize the LV module.
     */
    rc = evm_lv_init();
    if (rc != NFX_OK)
    {
        goto evm_initModule_err;
    }
    
    /* Initialize the I/O module.
     */
    rc = evm_io_init();    

    return(rc);

evm_initModule_err:

    return(NFX_ERR);

}

/*----------------------------------------------------------------------------*/
