/*------------------------------------------------------------------------------
 * RCS:  $Id: evm-debug.c,v 1.51 2002/12/10 23:30:39 isabellec Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : evm-debug.c
 *
 * Description  : EVM srvr debug functions
 *
 * Created by   : Lan Chin
 *                Isabelle Cnudde
 *
 * Date Created : 28-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

#include "exports/elog-api.h"
#include "evm-lv.h"
#include "evm-debug.h"
#include "evm-test.h"

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

extern evm_param_t evmParam;

#define EVM_TRACE_STRING_MAX    128
#define EVM_MAX_METHODS         5
#define EVM_INVALID_METHOD      4

#ifdef DEBUG
int32 evm_traceLevel = EVM_TRACE_DEBUG;
boolean evm_mwchksum_on = FALSE; //TRUE;
boolean evm_mrchksum_on = FALSE; //TRUE;
#else
int32 evm_traceLevel = EVM_TRACE_DEFAULT;
boolean evm_mwchksum_on = FALSE;
boolean evm_mrchksum_on = FALSE;
#endif

/* Spinlock for the global statistics.
 */
nfx_spinlock_t  evmStatLock;

/*------------------------------------------------------------------------------
 * Local function prototypes.
 *------------------------------------------------------------------------------
 */

static int32
evm_adminTraceCmd(int32     argc,
                  char8 *   argv[]);

static void
evm_displayOneLV(evm_vol_t *lv);

static int32
evm_displayAssets(int32     argc,
                  char8 *   argv[]);

static int32
evm_adminLVCmd(int32    argc,
               char8 *  argv[]);

static int32
evm_adminStatCmd(int32      argc,
                 char8 *    argv[]);

#ifdef DEBUG
static int32
evm_tstamp(int32      argc,
           char8 *    argv[]);
#endif // DEBUG

static int32
evm_adminChksum(int32     argc,
                char8 *   argv[]);


/*------------------------------------------------------------------------------
 * Name         : evm_trace
 *
 * Description  : trace function for the EVM server module
 *
 * Created by   : Lan Chin
 *                Isabelle Cnudde
 *
 * Date Created : 28-SEP-2001
 *------------------------------------------------------------------------------
 */

void
evm_trace(int32 level, char *fmt, ...)
    
{
    if (evm_traceLevel != 0)
    {
        va_list ap;
        char    trace_buf[EVM_TRACE_STRING_MAX];
        
        if (evm_traceLevel & EVM_TRACE_MASK)
        {
            if (((evm_traceLevel & EVM_TRACE_IO) && (level & EVM_TRACE_IO)) ||
                ((evm_traceLevel & EVM_TRACE_FUNC) && (level & EVM_TRACE_FUNC)) ||
                ((evm_traceLevel & EVM_TRACE_SRVR_MSG) && (level & EVM_TRACE_SRVR_MSG)) ||
                ((evm_traceLevel & EVM_TRACE_EEE) && (level & EVM_TRACE_EEE)) ||
                ((evm_traceLevel & EVM_TRACE_DETAIL) && (level & EVM_TRACE_DETAIL)) ||
                ((evm_traceLevel & EVM_TRACE_MEMORY) && (level & EVM_TRACE_MEMORY)))
            {
                va_start(ap, fmt);
                vsnprintf(trace_buf, EVM_TRACE_STRING_MAX, fmt, ap);
                E_LOG(class_1, LOG_DEBUG, evmAppId, 0, 0, 0,
                      ("EVM server(debug): %s", trace_buf));
                va_end(ap);
            }
        }
        else if (level & EVM_TRACE_MASK)
        {
            return;
        }
        
        if ((evm_traceLevel & EVM_TRACE_ERRORS) && (level & EVM_TRACE_ERRORS))
        {
            va_start(ap, fmt);
            vsnprintf(trace_buf, EVM_TRACE_STRING_MAX, fmt, ap);
            E_LOG(class_1, LOG_ERR, evmAppId, 0, 0, 0,
                  ("EVM server(error): %s", trace_buf));
            va_end(ap);
        }
        else if ((evm_traceLevel & EVM_TRACE_INFO) && (level & EVM_TRACE_INFO))
        {
            va_start(ap, fmt);
            vsnprintf(trace_buf, EVM_TRACE_STRING_MAX, fmt, ap);
            E_LOG(class_1, LOG_INFO, evmAppId, 0, 0, 0,
                  ("EVM server(info): %s", trace_buf));
            va_end(ap);
        }
    }    
    return;
}

/*-----------------------------------------------------------------
 * Name:        evm_adminTraceCmd
 *
 * Description: evmadmin UI command to enable or disable trace
 *
 *  agrv[0] -> trace [enable|disable] ....
 *  argc >= 2
 *
 * Command combinations:
 *
 *  trace enable: enable debug trace level
 *  trace disable: revert back to error trace level
 *  trace enable all: turn on all trace levels
 *  trace enable io: add io trace level
 *  trace enable sm: add sm (state machine) trace level
 *  trace enable io sm func: add io, sm and func (function call) trace levels
 *  etc etc for func,msg,detail,mem
 *  trace disable io: turn off io trace level
 *  etc etc etc
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static int32
evm_adminTraceCmd(int32     argc,
                  char8 *   argv[])
    
{
    char8   *cmd, *opt;
    int32   enable;

    if (argc < 2)
    {
        printf("USAGE: evmadmin trace [enable|disable] "
               "[err|info|io|func|msg|detail|mem|host|all]\n");
        return NFX_ERR;
    }
    cmd = argv[1];
    argv++; // skip to enable or disable
    if (strcmp(cmd, "enable") == 0)
    {
        enable = 1;
    }
    else if (strcmp(cmd, "disable") == 0)
    {
        enable = 0;
    }
    else
    {
        printf("USAGE: evmadmin trace [enable|disable]\n");
        return NFX_ERR;
    }
    /*
     * no specific level given, just use defaults
     */
    if (argc == 2)
    {
        if (enable)
        {
#ifdef DEBUG
            evm_traceLevel = EVM_TRACE_DEBUG;
#else
            evm_traceLevel = EVM_TRACE_DEFAULT;
#endif // DEBUG
            printf("Traces enabled for debug (error + info).\n");
        }
        else
        {
            evm_traceLevel = 0;
            printf("Traces disabled.\n");
        }
        return NFX_OK;
    }
    /*
     * turn on of off a specific level
     */
    argc -= 2;      // skip "trace [enable|disable]"
    argv++;         // skip to levels
    while (argc--)
    {
        opt = argv[0];
        argv++;     // go to next level
        if (strcmp(opt, "err") == 0)
        {
            if (enable) evm_traceLevel |= EVM_TRACE_ERRORS;
            else        evm_traceLevel &= ~EVM_TRACE_ERRORS;
        }
        else if (strcmp(opt, "info") == 0)
        {
            if (enable) evm_traceLevel |= EVM_TRACE_INFO;
            else        evm_traceLevel &= ~EVM_TRACE_INFO;
        }
        else if (strcmp(opt, "io") == 0)
        {
            if (enable) evm_traceLevel |= EVM_TRACE_IO;
            else        evm_traceLevel &= ~EVM_TRACE_IO;
        }
        else if (strcmp(opt, "func") == 0)
        {
            if (enable) evm_traceLevel |= EVM_TRACE_FUNC;
            else        evm_traceLevel &= ~EVM_TRACE_FUNC;
        }
        else if (strcmp(opt, "msg") == 0)
        {
            if (enable) evm_traceLevel |= EVM_TRACE_SRVR_MSG;
            else        evm_traceLevel &= ~EVM_TRACE_SRVR_MSG;
        }
        else if (strcmp(opt, "eee") == 0)
        {
            if (enable) evm_traceLevel |= EVM_TRACE_EEE;
            else        evm_traceLevel &= ~EVM_TRACE_EEE;
        }
        else if (strcmp(opt, "detail") == 0)
        {
            if (enable) evm_traceLevel |= EVM_TRACE_DETAIL;
            else        evm_traceLevel &= ~EVM_TRACE_DETAIL;
        }
        else if (strcmp(opt, "mem") == 0)
        {
            if (enable) evm_traceLevel |= EVM_TRACE_MEMORY;
            else        evm_traceLevel &= ~EVM_TRACE_MEMORY;
        }
        else if (strcmp(opt, "hist") == 0)
        {
            if (enable) evm_traceLevel |= EVM_TRACE_HISTORY;
            else        evm_traceLevel &= ~EVM_TRACE_HISTORY;
        }
        else if (strcmp(opt, "all") == 0)
        {
            if (enable) evm_traceLevel = EVM_TRACE_ALL;
            else        evm_traceLevel = 0;
        }
        else
        {
            printf("USAGE: evmadmin trace [enable|disable] "
                   "[err|info|io|func|msg|eee|detail|mem|hist|all]\n");
            return NFX_ERR;
        }
    }
    printf("New trace level 0x%x\n", evm_traceLevel);
    return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name:        evm_displayOneLV, evm_displayAssets
 *
 * Description: called to display physical storage assets for LV
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static void
evm_displayOneLV(evm_vol_t *  lv)
    
{
    int32               i = 0;
    evm_lun_t * lun = NULL;
    //evm_assetMatrix_t * mp;
    //evm_assetCell_t *   ap;

    printf("volId 0x%llx: VolSize %lld\n",
           lv->id.whole_thing, lv->size);
    //mp = lv->assets;
    
    DLL_ITERATE(lun, &(lv->lunList), entry)
        //while (mp)
    {
        printf("\t[%d] -> scsi id 0x%llx offStart %lld end %lld nBlks %lld\n",
               i, lun->id, lun->volOffsetStart, lun->volOffsetEnd, lun->nBlks);
        i++;
    } // DLL_ITERATE
    printf("\n");
} /* evm_displayOneLV */

/*----------------------------------------------------------------------------*/

static int32
evm_displayAssets(int32     argc,
                  char8 *   argv[])
    
{
    int32       rc = NFX_OK;
    int32       i = 0;
    int32       lvIndex;
    evm_vol_t * lv;

    NFX_SPIN_LOCK(&evmVolLock);
    
    if (argc == 2)
    {
        lvIndex = atoi(argv[1]);
        DLL_ITERATE(lv, &evm_volList, entry)
        {
            if (lvIndex == 0)
                break;
            lvIndex--;
        }
        if (lv == NULL)
        {
            printf("\nNo logical volume found for index %d\n", lvIndex);
            rc = NFX_ERR;
            goto evm_displayAssets_exit;
        }
        printf("[%d] ", lvIndex);
        evm_displayOneLV(lv);
        goto evm_displayAssets_exit;
    }
    /* display all */
    DLL_ITERATE(lv, &evm_volList, entry)
    {
        printf("[%d] ", i++);
        evm_displayOneLV(lv);
    }

  evm_displayAssets_exit:
    
    NFX_SPIN_UNLOCK(&evmVolLock);
    return(rc);

} /* evm_displayAssets */

/*-----------------------------------------------------------------
 * Name:        evm_adminTraceCmd
 *
 * Description: evmadmin UI command to show configured logical volumes
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static int32
evm_adminLVCmd(int32    argc,
               char8 *  argv[])
    
{
    char8   *cmd;

    if (argc < 2)
    {
        printf("USAGE: evmadmin lv show [vol-id]\n");
        return NFX_ERR;
    }
    
    cmd = argv[1];
    if (strcmp(cmd, "show") == 0)
    {
        evm_displayAssets(--argc, ++argv);
    }
    else
    {
        printf("USAGE: evmadmin lv show [vol-id]\n");
        return NFX_ERR;
    }
    
    return NFX_OK;

} /* evm_adminLVCmd */

/*------------------------------------------------------------------------------
 * Name:        evm_adminTraceCmd
 *
 * Description: evmadmin UI command to show or clear evm global statistics
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *------------------------------------------------------------------------------
 */

void
evm_displayStats(void)
    
{
    int slot, cpu;
    
    NFX_SPIN_LOCK(&evmStatLock);
    printf("Number of calls to evm_IOReq ................... %lld\n",
           evm_stats.nIOReqs);
    printf("Number evm_IOReq user callbacks ................ %lld\n",
           evm_stats.nIOCompletes);
    printf("Meta read:\n");
    printf("     Number of requests ........................ %lld\n",
           evm_stats.nMReads);
    printf("     Number of retries ......................... %lld\n",
           evm_stats.nMReadsRetry);
    printf("     Number of responses ....................... %lld\n",
           evm_stats.nScsiMReadRsps);
    printf("     Number of bytes ........................... %lld\n",
           evm_stats.nBlkMRead * EVM_LBLK_SIZE);
    printf("Meta write:\n");
    printf("     Number of requests ........................ %lld\n",
           evm_stats.nMWrites);
    printf("     Number of retries ......................... %lld\n",
           evm_stats.nMWritesRetry);
    printf("     Number of responses ....................... %lld\n",
           evm_stats.nScsiMWriteRsps);
    printf("     Number of bytes ........................... %lld\n",
           evm_stats.nBlkMWrite * EVM_LBLK_SIZE);
    printf("User read:\n");
    printf("     Number of requests ........................ %lld\n",
           evm_stats.nUReads);
    printf("     Number of retries ......................... %lld\n",
           evm_stats.nUReadsRetry);
    printf("     Number of responses ....................... %lld\n",
           evm_stats.nScsiUReadRsps);
    printf("     Number of bytes ........................... %lld\n",
           evm_stats.nBlkURead * EVM_LBLK_SIZE);
    printf("User write:\n");
    printf("     Number of requests ........................ %lld\n",
           evm_stats.nUWrites);
    printf("     Number of retries ......................... %lld\n",
           evm_stats.nUWritesRetry);
    printf("     Number of responses ....................... %lld\n",
           evm_stats.nScsiUWriteRsps);
    printf("     Number of bytes ........................... %lld\n",
           evm_stats.nBlkUWrite * EVM_LBLK_SIZE);
    printf("\n");
    printf("Number evm_IOReq rejected ...................... %lld\n",
           evm_stats.nIORejected);
    printf("Number of SCSI device errors ................... %lld\n",
           evm_stats.nScsiDevErrs);
    printf("Number of errors ............................... %lld\n",
           evm_stats.nErrors);
    printf("Number of IO Timeouts .......................... %lld\n",
           evm_stats.nTimeouts);
    printf("Number of EEE Tx Queue Full .................... %lld\n",
           evm_stats.nEEETxQueueFull);
    printf("Number of stale IO responses ................... %lld\n",
           evm_stats.nStales);
    printf("Number of bad packet length .................... %lld\n",
           evm_stats.nBadPacketLength);
    printf("\n");
    printf("Number of memory allocations ................... %lld\n",
           evm_stats.nAllocs);
    printf("Number of memory deallocations ................. %lld\n",
           evm_stats.nDeallocs);
    printf("Number of timer allocations .................... %lld\n",
           evm_stats.nTimerCreates);
    printf("Number of timer deallocations .................. %lld\n",
           evm_stats.nTimerDestroys);
    printf("\n");
    printf("Meta I/O queue:\n");
    printf("     Number of free elements ................... %d\n",
           evm_stats.qCnt.metaFree);
    printf("     Number of waiting physical requests ....... %d (got %d)\n",
           evm_stats.qCnt.metaPioWait, evm_stats.nMetaMaxPioWait);
    printf("     Number of pending physical requests ....... %d (got %d)\n",
           evm_stats.qCnt.metaPioPend, evm_stats.nMetaMaxPioPend);
    printf("User I/O queue:\n");
    printf("     Number of free elements ................... %d\n",
           evm_stats.qCnt.userFree);
    printf("     Number of waiting physical requests ....... %d (got %d)\n",
           evm_stats.qCnt.userPioWait, evm_stats.nUserMaxPioWait);
    printf("     Number of pending physical requests ....... %d (got %d)\n",
           evm_stats.qCnt.userPioPend, evm_stats.nUserMaxPioPend);
    printf("Total I/O queues (meta + user):\n");
    printf("     Max. number of waiting physical requests .. %d\n",
           evm_stats.nMaxPioWait);
    printf("     Max. number of pending physical requests .. %d\n",
           evm_stats.nMaxPioPend);
    printf("\n");
#ifdef DEBUG
    printf("SCSI queue depth ............................... %lld (got %lld)\n",
           (evm_stats.numQueueDepth == 0)? 0 :
           (evm_stats.sumQueueDepth/evm_stats.numQueueDepth),
           evm_stats.maxQueueDepth);
    printf("\n");
#endif // DEBUG
    printf("IO response time (in msec):\n");
    printf("more than  %7d    %7d    %7d    %7d    "
           "%7d    %7d        maxTime    avgTime    \n",
           EVM_AMSEC, EVM_BMSEC, EVM_CMSEC, EVM_DMSEC, EVM_EMSEC, EVM_FMSEC);
    printf("-----------------------------------------------------------"
           "--------------------------------------------\n");
    printf("META READ .... %-10lld %-10lld %-10lld %-10lld %-10lld %-10lld "
           "%-10lld %-10lld\n",
           evm_stats.MReadTime.moreThanAmsec,
           evm_stats.MReadTime.moreThanBmsec,
           evm_stats.MReadTime.moreThanCmsec,
           evm_stats.MReadTime.moreThanDmsec,
           evm_stats.MReadTime.moreThanEmsec,
           evm_stats.MReadTime.moreThanFmsec,
           eee_rtcToMS(&evm_stats.MReadTime.maxTime),
           (evm_stats.MReadTime.numTime == 0)? 0:
           (uint64)eee_rtcDivMS(&evm_stats.MReadTime.sumTime, evm_stats.MReadTime.numTime));
    printf("META WRITE ... %-10lld %-10lld %-10lld %-10lld %-10lld %-10lld "
           "%-10lld %-10lld\n",
           evm_stats.MWriteTime.moreThanAmsec,
           evm_stats.MWriteTime.moreThanBmsec,
           evm_stats.MWriteTime.moreThanCmsec,
           evm_stats.MWriteTime.moreThanDmsec,
           evm_stats.MWriteTime.moreThanEmsec,
           evm_stats.MWriteTime.moreThanFmsec,
           eee_rtcToMS(&evm_stats.MWriteTime.maxTime),
           (evm_stats.MWriteTime.numTime == 0)? 0:
           (uint64)eee_rtcDivMS(&evm_stats.MWriteTime.sumTime, evm_stats.MWriteTime.numTime));
    printf("USER READ .... %-10lld %-10lld %-10lld %-10lld %-10lld %-10lld "
           "%-10lld %-10lld\n",
           evm_stats.UReadTime.moreThanAmsec,
           evm_stats.UReadTime.moreThanBmsec,
           evm_stats.UReadTime.moreThanCmsec,
           evm_stats.UReadTime.moreThanDmsec,
           evm_stats.UReadTime.moreThanEmsec,
           evm_stats.UReadTime.moreThanFmsec,
           eee_rtcToMS(&evm_stats.UReadTime.maxTime),
           (evm_stats.UReadTime.numTime == 0)? 0:
           (uint64)eee_rtcDivMS(&evm_stats.UReadTime.sumTime, evm_stats.UReadTime.numTime));
    printf("USER WRITE ... %-10lld %-10lld %-10lld %-10lld %-10lld %-10lld "
           "%-10lld %-10lld\n",
           evm_stats.UWriteTime.moreThanAmsec,
           evm_stats.UWriteTime.moreThanBmsec,
           evm_stats.UWriteTime.moreThanCmsec,
           evm_stats.UWriteTime.moreThanDmsec,
           evm_stats.UWriteTime.moreThanEmsec,
           evm_stats.UWriteTime.moreThanFmsec,
           eee_rtcToMS(&evm_stats.UWriteTime.maxTime),
           (evm_stats.UWriteTime.numTime == 0)? 0:
           (uint64)eee_rtcDivMS(&evm_stats.UWriteTime.sumTime, evm_stats.UWriteTime.numTime));
    printf("\n");
    for (slot = 0; slot < 1; slot++)
    {
        for (cpu = 0; cpu < 4; cpu++)
        {
            printf("Number of IOs exchanged with FC-%d cpu %d\n", slot, cpu);
            printf("\tmeta read req/rsp .................... %lld / %lld\n",
                   evm_stats.ioPerFc[slot][cpu].metaReadReq,
                   evm_stats.ioPerFc[slot][cpu].metaReadRsp);
            printf("\tmeta write req/rsp ................... %lld / %lld\n",
                   evm_stats.ioPerFc[slot][cpu].metaWriteReq,
                   evm_stats.ioPerFc[slot][cpu].metaWriteRsp);
            printf("\tuser read req/rsp .................... %lld / %lld\n",
                   evm_stats.ioPerFc[slot][cpu].userReadReq,
                   evm_stats.ioPerFc[slot][cpu].userReadRsp);
            printf("\tuser write req/rsp ................... %lld / %lld\n",
                   evm_stats.ioPerFc[slot][cpu].userWriteReq,
                   evm_stats.ioPerFc[slot][cpu].userWriteRsp);
        } // for (cpu)
    } // for (slot)
    printf("\n");
    NFX_SPIN_UNLOCK(&evmStatLock);

} /* evm_displayStats */

/*----------------------------------------------------------------------------*/

static int32
evm_adminStatCmd(int32      argc,
                 char8 *    argv[])
    
{
    char8       *cmd;

    if (argc < 2)
    {
        printf("USAGE: evmadmin stat {show | clear}\n");
        return NFX_ERR;
    }
    
    cmd = argv[1];
    if (strcmp(cmd, "show") == 0)
    {
        evm_displayStats();
    }
    else if (strcmp(cmd, "clear") == 0)
    {
        NFX_SPIN_LOCK(&evmStatLock);
        bzero((void *) &evm_stats, sizeof(evm_stats) - sizeof(evm_queueCnt_t));
        printf("Done.\n");
        NFX_SPIN_UNLOCK(&evmStatLock);
    }
    else
    {
        printf("USAGE: evmadmin stat {show | clear}\n");
        return NFX_ERR;
    }
    return NFX_OK;

} /* evm_adminStatCmd */

#ifdef DEBUG
/*------------------------------------------------------------------------------
 * Name         : evm_drop
 *
 * Description  : EVM UI commands (evmadmin) see sm-ui/cmdtable.c
 *                Requests EVM server to drop some IO requests.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-MAY-2002
 *------------------------------------------------------------------------------
 */

evm_dropData_t  evm_dropData = EVM_DROPDATA_NONE;
evm_dropIo_t    evm_dropIo = EVM_DROPIO_NONE;
uint32          evm_dropWhen = 0;
uint32          evm_dropNb = 0;

uint32          evm_nbReq = 0;
uint32          evm_dropCnt = 0;

static int32
evm_drop(int32      argc,
         char8 *    argv[])
    
{
    if (argc < 2)
        goto evm_drop_printUsage;

    if (!strcmp(argv[1], "stop"))
    {
        if (argc != 2)
            goto evm_drop_printUsage;
        
        evm_dropData = EVM_DROPDATA_NONE;
        evm_dropIo = EVM_DROPIO_NONE;
        evm_dropWhen = 0;
        evm_dropNb = 0;
     }
    else
    {
        if (argc != 5)
            goto evm_drop_printUsage;
        
        if (!strcmp(argv[1], "user"))
            evm_dropData = EVM_DROPDATA_USER;
        else if (!strcmp(argv[1], "meta"))
            evm_dropData = EVM_DROPDATA_META;
        else if (!strcmp(argv[1], "both"))
            evm_dropData = EVM_DROPDATA_BOTH;
        else
            goto evm_drop_printUsage;
        
        if (!strcmp(argv[2], "read"))
            evm_dropIo = EVM_DROPIO_READ;
        else if (!strcmp(argv[2], "write"))
            evm_dropIo = EVM_DROPIO_WRITE;
        else if (!strcmp(argv[2], "both"))
            evm_dropIo = EVM_DROPIO_BOTH;
        else
        {
            evm_dropData = EVM_DROPDATA_NONE;
            goto evm_drop_printUsage;
        }

        evm_dropWhen = atoi(argv[3]);
        evm_dropNb = atoi(argv[4]);
    }
    return NFX_OK;

  evm_drop_printUsage:

    printf("USAGE: evmadmin drop {stop | <data> <io> <when> <nb>}\n");
    printf("       <data> : {user|meta|both}\n");
    printf("       <io>   : {read|write|both}\n");
    printf("       <when> : start dropping after how many requests\n");
    printf("       <nb>   : number of requests to drop\n");

    return NFX_ERR;

}
#endif // DEBUG

/*------------------------------------------------------------------------------
 * Name         : evm_assert
 *
 * Description  : EVM UI commands (evmadmin) see sm-ui/cmdtable.c
 *                Requests EVM server to assert or not on error detection.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 07-JUN-2002
 *------------------------------------------------------------------------------
 */

#ifdef DEBUG
//boolean evmDoAssert = TRUE;
boolean evmDoAssert = FALSE;
#else
boolean evmDoAssert = FALSE;
#endif // DEBUG

static int32
evm_assert(int32      argc,
           char8 *    argv[])
    
{
    if (argc != 2)
        goto evm_assert_printUsage;

    if (!strcmp(argv[1], "on"))
    {
        evmDoAssert = TRUE;
        printf("EVM server assert turned ON.\n");
    }
    else if (!strcmp(argv[1], "off"))
    {
        evmDoAssert = FALSE;
        printf("EVM server assert turned OFF.\n");
    }
    else if (!strcmp(argv[1], "show"))
    {
        printf("EVM server assert is turned %s\n", (evmDoAssert)? "ON":"OFF");
    }
    else
    {
        goto evm_assert_printUsage;
    }
    return(NFX_OK);

  evm_assert_printUsage:

    printf("USAGE: evmadmin assert {show|on|off}\n");
    return(NFX_ERR);

}

/*------------------------------------------------------------------------------
 * Name         : evm_tstamp
 *
 * Description  : EVM UI commands (evmadmin) see sm-ui/cmdtable.c
 *                Requests EVM server to timestamp or not the volume.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 08-JUL-2002
 *------------------------------------------------------------------------------
 */

boolean evmDoTstamp = FALSE;

#ifdef DEBUG

static int32
evm_tstamp(int32      argc,
           char8 *    argv[])
    
{
    if (argc != 2)
        goto evm_tstamp_printUsage;

    if (!strcmp(argv[1], "on"))
    {
        //evmDoTstamp = TRUE;
        evm_startLookingForTakeover();
        printf("Volume timestamping is ON.\n");
    }
    else if (!strcmp(argv[1], "off"))
    {
        //evmDoTstamp = FALSE;
        evm_stopLookingForTakeover();
        printf("Volume timestamping is OFF.\n");
    }
    else if (!strcmp(argv[1], "show"))
    {
        printf("Volume timestamping is %s\n", (evmDoTstamp)? "ON":"OFF");
    }
    else
    {
        goto evm_tstamp_printUsage;
    }
    return(NFX_OK);

  evm_tstamp_printUsage:

    printf("USAGE: evmadmin tstamp {show|on|off}\n");
    return(NFX_ERR);

}
#endif // DEBUG

/*-----------------------------------------------------------------
 * Name:        evm_adminChksum
 *-----------------------------------------------------------------
 */

static int32
evm_adminChksum(int32     argc,
                char8 *   argv[])
    
{
    if (argc == 1)
    {
        printf("USAGE: evmadmin chksum [show|on|off]\n");
    }
	else
	{
		if (strcmp(argv[1], "on") == 0)
        {
			evm_mwchksum_on = TRUE;
			evm_mrchksum_on = TRUE;
        }
		else if (strcmp(argv[1], "off") == 0)
        {
			evm_mwchksum_on = FALSE;
			evm_mrchksum_on = FALSE;
        }
		else if (strcmp(argv[1], "show") == 0)
        {
            printf("EVM server meta write checksum is %s\n",
                   (evm_mwchksum_on == TRUE)? "ON" : "OFF");
            printf("EVM server meta read checksum is %s\n",
                   (evm_mrchksum_on == TRUE)? "ON" : "OFF");
        }
		else
		{
			printf("USAGE: evmadmin chksum [show|on|off]\n");
			return NFX_ERR;
		}
	}

    return NFX_OK;
}

/*------------------------------------------------------------------------------
 * Name:        evm_admin
 *
 * Description: EVM UI commands (evmadmin) see sm-ui/cmdtable.c
 *
 *              available only via FP simulator
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *------------------------------------------------------------------------------
 */

int32
evm_admin(int32 argc, char8 *argv[])
    
{
    int32   rc;
    char8   *cmd;

    if (argc < 2)
    {
        goto evm_admin_printUsage;
    }
    
    cmd = argv[1];
    if (strcmp(cmd, "stat") == 0)
    {
        rc = evm_adminStatCmd(--argc, ++argv);
    }
    else if (strcmp(cmd, "lv") == 0)
    {
        rc = evm_adminLVCmd(--argc, ++argv);
    }
    else if (strcmp(cmd, "trace") == 0)
    {
        rc = evm_adminTraceCmd(--argc, ++argv);
    }
    else if (strcmp(cmd, "test") == 0)
    {
        rc = evm_test(--argc, ++argv);
    }
    else if (strcmp(cmd, "assert") == 0)
    {
        rc = evm_assert(--argc, ++argv);
    }
    else if (strcmp(cmd, "chksum") == 0)
    {
        rc = evm_adminChksum(--argc, ++argv);
    }
#ifdef DEBUG
    else if (strcmp(cmd, "drop") == 0)
    {
        rc = evm_drop(--argc, ++argv);
    }
    else if (strcmp(cmd, "tstamp") == 0)
    {
        rc = evm_tstamp(--argc, ++argv);
    }
#endif // DEBUG
#ifdef SY_IO_PERFORMANCE_TEST
    else if (strcmp(cmd, "perf_evm") == 0)
    {
        rc = evm_perf(--argc, ++argv);
    }
    else if (strcmp(cmd, "perf_eee") == 0)
    {
        rc = evm_perfeee(--argc, ++argv);
    }
    else if (strcmp(cmd, "perf_evm_prof") == 0)
    {
        rc = evm_perf_prof(--argc, ++argv);
    }
#endif // SY_IO_PERFORMANCE_TEST
    else if (strcmp(cmd, "iotimeout") == 0)
    {
        if (argc == 3)
			eee_msToRTC(atoi(argv[2]) * 1000, &evmParam.ioTimeout);
        else
            printf("ioTimeout = %lld\n", eee_rtcToSec(&evmParam.ioTimeout));
    }
    else if (strcmp(cmd, "maxioretry") == 0)
    {
        if (argc == 3)
            evmParam.maxIoRetry = atoi(argv[2]);
        else
            printf("maxIoRetry = %d\n", evmParam.maxIoRetry);
    }
    else if (strcmp(cmd, "maxio") == 0)
    {
        if (argc == 3)
            evmParam.maxIo = atoi(argv[2]);
        else
            printf("maxIo = %d\n", evmParam.maxIo);
    }
    else
    {
        goto evm_admin_printUsage;
    }
    
    return(rc);
    
  evm_admin_printUsage:

#ifdef SY_IO_PERFORMANCE_TEST
    printf("USAGE: evmadmin <stat|lv|trace|test|drop|assert|tstamp|chksum|"
           "iotimeout|maxioretry|maxio>\n");
    printf("PERFORMANCE TESTING USAGE: "
           "evmadmin <perf_evm|perf_eee|perf_evm_prof>\n");
#else
#ifdef DEBUG
    printf("USAGE: evmadmin <stat|lv|trace|test|drop|assert|chksum|"
           "iotimeout|maxioretry|maxio|tstamp>\n");
#else // DEBUG
    printf("USAGE: evmadmin <stat|lv|trace|test|assert|chksum|"
           "iotimeout|maxioretry|maxio>\n");
#endif // DEBUG
#endif
    return(NFX_ERR);    
}

/*----------------------------------------------------------------------------*/
