/*------------------------------------------------------------------------------
 * RCS:  $Id: evm-msg.c,v 1.42 2002/12/10 23:30:39 isabellec Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : evm-msg.c
 *
 * Description  : functions to send EEE packets to SBM/SAC and SCSI
 *
 * Created by   : Lan Chin
 *                Isabelle Cnudde
 *
 * Date Created : 9/28/2001
 *
 *------------------------------------------------------------------------------
 */

#include "nfx-incl.h"
#include "exports/cod-api.h"
#include "exports/debug.h"
#include "evm-msg.h"
#include "evm-hash.h"
#include "evm-debug.h"

#ifdef DEBUG
extern evm_dropData_t   evm_dropData;
extern evm_dropIo_t     evm_dropIo;
extern uint32           evm_dropWhen;
extern uint32           evm_dropNb;
extern uint32           evm_nbReq;
extern uint32           evm_dropCnt;
extern int32			evm_mwchksum_on;
#endif // DEBUG

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : evm_sendScsiMsg
 *
 * Description  : 
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 19-AUG-2002
 *------------------------------------------------------------------------------
 */

int32
evm_sendScsiMsg(evm_lioCb_t *   lcb,
                evm_pioCb_t *   pcb)
    
{
    int32               rc;
    int32               totalSize, msgSize, bufCount;
    eee_descPtr_t       msgp=NULL;
    scsi_msgHdr_t       *hp=NULL;
    scsi_readReqMsg_t   *rp=NULL;
    scsi_writeReqMsg_t  *wp=NULL;
    
#ifdef DEBUG
    evm_nbReq++;
    if ((evm_dropData & EVM_DROPDATA_META) && (evm_nbReq >= evm_dropWhen))
    {
        if (evm_dropCnt >= evm_dropNb)
        {
            evm_dropCnt = 0;
            evm_nbReq = 0;
        }
        else
        {
            if (((evm_dropIo & EVM_DROPIO_READ) &&
                 (lcb->type == m_read)) ||
                ((evm_dropIo & EVM_DROPIO_WRITE) &&
                 (lcb->type == m_write)))
            {
                evm_trace(EVM_TRACE_ERRORS, "EVM[%d]: dropping request...",
                          pcb->handle.whole);
                evm_dropCnt++;
                return NFX_OK;
            }
        }
    }
#endif // DEBUG

    /*
     * allocate eee message descriptor and scsi message
     */
    msgp = eee_allocateDesc(EEE_SHARED_POOL, NFX_COS_HIGH);
    if (msgp == NULL)
    {
        evm_trace(EVM_TRACE_ERRORS,
                  "[%d]: cannot allocate EEE descriptor",
                  pcb->handle.whole);
        goto evm_sendScsiMsg_err;
    }
    msgp->hdr.attr = 0;
    msgp->hdr.next = NULL;
    msgp->hdr.offset = 0;
    /*
     * we can allocate memory for the SCSI message from
     * the shared pool or use a malloc'd buffer
     * eee will know what to do with it (see eee_deallocateBuffer)
     */
    switch (lcb->type)
    {
    case m_read:
    case u_read:
    {
        msgSize = sizeof(scsi_readReqMsg_t);
        hp = (scsi_msgHdr_t *) eee_allocateBuffer(EEE_SHARED_POOL,
                                                  EDESC_BD_SMALL, 0);
        if (hp == NULL)
        {
            evm_trace(EVM_TRACE_ERRORS,
                      "[%d]: cannot allocate buffer",
                      pcb->handle.whole);
            goto evm_sendScsiMsg_err;
        }
        
        // Apply an offset into the buffer for transport
        // alignment requirements.
        (uint64) (address_t) hp += (MALLOC_ALIGNED_SIZE * 2);
        (char *)hp -= sizeof(scsi_readReqMsg_t);
        msgp->bd[1].buf = (void *) hp;
        msgp->bd[1].len = msgSize;
        msgp->bd[1].control = EDESC_BD_CMD_IN_USE;
        
        rp = (scsi_readReqMsg_t *) hp;
        bufCount = 1;
        totalSize = msgSize;

        /* Fill in the read request 
         */
        rp->transSeqId = 0;
        rp->blkOffset = pcb->offset;
        rp->blkCnt = pcb->size;
        rp->attributes = lcb->req.attributes;
        rp->dataDest = 0;

        hp->mid = SCSI_MID_READ_REQ;

        break;
    }
            
    case m_write:
    {
        void    *bufBase;

        msgSize = sizeof(scsi_writeReqMsg_t);
        if (eee_isUltraBuf(pcb->dataHandle.dataPtr, &bufBase))
        {
            NFX_ASSERT(pcb->dataHandle.dataPtr - msgSize >= bufBase);

            // Caller's supplied buffer is an ULTRA buffer and
            // contains enough space for prepending scsi header.

            hp = (scsi_msgHdr_t *) pcb->dataHandle.dataPtr;
            (char *)hp -= sizeof(scsi_writeReqMsg_t);

            msgp->bd[1].buf = (void *) hp;
            msgp->bd[1].len = msgSize + (pcb->size * EVM_LBLK_SIZE);
            msgp->bd[1].control = EDESC_BD_CMD_IN_USE | EDESC_BD_CMD_NO_DEALLOC;
            bufCount = 1;
        }
        else
        {
            VERIFY(((uint64) (address_t) pcb->dataHandle.dataPtr &
                    (MALLOC_ALIGNED_SIZE - 1)) == 0);
            msgp->bd[2].buf = pcb->dataHandle.dataPtr;
            msgp->bd[2].len = pcb->size * EVM_LBLK_SIZE; 
            msgp->bd[2].control = EDESC_BD_CMD_IN_USE | EDESC_BD_CMD_NO_DEALLOC;

            hp = (scsi_msgHdr_t *) eee_allocateBuffer(EEE_SHARED_POOL,
                                                      EDESC_BD_SMALL, 0);
            if (hp == NULL)
            {
                evm_trace(EVM_TRACE_ERRORS,
                          "[%d]: cannot allocate buffer",
                          pcb->handle.whole);
                goto evm_sendScsiMsg_err;
            }

            // Apply an offset into the buffer for transport
            // alignment requirements.
            (uint64) (address_t) hp += (MALLOC_ALIGNED_SIZE * 2);
            (char *)hp -= sizeof(scsi_writeReqMsg_t);

            msgp->bd[1].buf = (void *) hp;
            msgp->bd[1].len = msgSize;
            msgp->bd[1].control = EDESC_BD_CMD_IN_USE;
            bufCount = 2;
        }

        /* Fill in the write request
         */
        wp = (scsi_writeReqMsg_t *) hp;
        wp->transSeqId = 0;
        wp->blkOffset = pcb->offset;
        wp->blkCnt = pcb->size;
        wp->attributes = lcb->req.attributes;
        wp->checkSum = pcb->checksum;
#ifdef DEBUG
        if ((pcb->nbRetry > 0) && evm_mwchksum_on)
        {
            NFX_ASSERT(pcb->checksum ==
					   uint32_checksum(pcb->dataHandle.dataPtr,
									   pcb->size * EVM_LBLK_SIZE,
									   0));
        }
#endif // DEBUG
        totalSize = msgSize + (pcb->size * EVM_LBLK_SIZE);

        hp->mid = SCSI_MID_WRITE_REQ;

        break;
    }
    default:
        evm_trace(EVM_TRACE_ERRORS,
                  "[%d]: invalid SCSI message type %d",
                  pcb->handle.whole, lcb->type);
        goto evm_sendScsiMsg_err;
    } // switch

    /*
     * fill in the common eee and scsi headers
     * first buffer left unused in case EEE needs it
     * second has the scsi message
     */
    hp->cHandle = pcb->handle.whole;

    rc = sdma_handleToPath(pcb->dev->openHandle, &(hp->path),
                           &(msgp->hdr.dest_port));
    if (rc == -1)
    {
        evm_trace(EVM_TRACE_ERRORS, "[%d]: sdma_handleToPath(%x) failed",
                  pcb->handle.whole, pcb->dev->openHandle);
        goto evm_sendScsiMsg_err;        
    }
    evm_trace(EVM_TRACE_IO, "EVM IO[%d]: "
              "SCSI -> sdma handle 0x%x path 0x%x dest_port 0x%x",
              pcb->handle.whole, pcb->dev->openHandle, hp->path,
              msgp->hdr.dest_port);
    if (lcb->type == u_read)
    {
        // response should be sent to SBM, not me
        msgp->hdr.src_port = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
                                                       NFX_NFP_TXRX_ANY,
                                                       sacAppId,
                                                       EEE_SAC,
                                                       NFX_COS_HIGH,
                                                       PORT_BEST_FIT);
    }
    else
    {
        msgp->hdr.src_port = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
                                                       nfxMyCPUId,
                                                       evmAppId,
                                                       EEE_USER_DATA,
                                                       NFX_COS_HIGH,
                                                       PORT_BEST_FIT);
    }
    msgp->hdr.control = (bufCount << EDESC_BUFCNT_SHIFT) | totalSize;

    rc = eee_forwardPacket(msgp);
    switch (rc)
    {
        case EEE_DROP_PACKET:
            /*
             * The packet was dropped by either the driver or
             * the EEE layer.  The descriptor has been freed.
             * The caller should make no further references to the packet.
             */
            msgp = NULL;   // set to NULL so we won't free
            hp = NULL;
            evm_trace(EVM_TRACE_ERRORS,
                      "[%d]: "
                      "EEE_DROP_PACKET from EEE forwardPacket",
                      pcb->handle.whole);
            EVM_STAT_EEEDROPPACKET;
            goto evm_sendScsiMsg_err;
            
        case EEE_TX_QUEUE_FULL:
            /*
             * The packet could not be queued due a tx queue full
             * condition.  The caller can take whatever action
             * it wants with the packet, since it is still the owner.
             */
            evm_trace(EVM_TRACE_ERRORS,
                      "[%d]: EEE_TX_QUEUE_FULL from EEE forwardPacket",
                      pcb->handle.whole);
            EVM_STAT_EEETXQUEUEFULL;
            goto evm_sendScsiMsg_err;
            
        case EEE_FORWARDED_PACKET:
            /*
             * The packet was queued for transmit successfully.
             * The caller should not free the packet (or touch
             * the packet in any way), as it is now owned by someone else.
             */
            evm_trace(EVM_TRACE_EEE,
                      "EVM EEE[%d]: SCSI msg type FORWARDED",
                      pcb->handle.whole);
            break;
            
        default:
            evm_trace(EVM_TRACE_ERRORS,
                      "[%d]: invalid rc %d from EEE forwardPacket",
                      pcb->handle.whole);
            break;
    }

#ifdef DEBUG
    switch (lcb->type)
    {
    case u_read:
        ATOMIC_INC_DWORD(&(evm_stats.ioPerFc[0][EEE_GET_DEST_CPU(msgp->hdr.dest_port)].userReadReq), 1);
        break;
        
    case m_read:
        ATOMIC_INC_DWORD(&(evm_stats.ioPerFc[0][EEE_GET_DEST_CPU(msgp->hdr.dest_port)].metaReadReq), 1);
        break;
        
    case m_write:
        ATOMIC_INC_DWORD(&(evm_stats.ioPerFc[0][EEE_GET_DEST_CPU(msgp->hdr.dest_port)].metaWriteReq), 1);
        break;
    } // switch
#endif // DEBUG
    return NFX_OK;

evm_sendScsiMsg_err:

	evm_trace(EVM_TRACE_ERRORS, ": "
              "evm_sendScsiMsg: Error Processing\n");
    /*
     * something went wrong
     * free up any allocated resources before returning err
     */
    if (msgp)
    {
 		eee_freePacket(msgp);
    }
       
    return NFX_ERR;

}

/*------------------------------------------------------------------------------
 * Name         : evm_sendSbmScsiWriteMsg
 *
 * Description  : called to format and send the
 *                CMD_TO_MCPU_SEND_CONTROL_AND_DATA_TO_SNIM for a
 *                EVM USER WRITE request
 *
 * Created by   : Lan Chin
 *
 * Date Created : 9/28/01
 *------------------------------------------------------------------------------
 */

int32
evm_sendSbmScsiWriteMsg(evm_lioCb_t *   lcb,
                        evm_pioCb_t *   pcb)
    
{
    int32                           rc;
    int32                           sMsgSize;       // SCSI msg size
    int32                           totalLen;       // SBM+SCSI msg size
    eee_descPtr_t                   msgp=NULL;      // EEE descriptor
    sbm_sendControlAndSACData_t *   sbmMsg=NULL;    // SBM WRITE msg
    scsi_writeReqMsg_t          *   wp;             // SCSI write msg

#ifdef DEBUG
    evm_nbReq++;
    if ((evm_dropData & EVM_DROPDATA_USER) && (evm_nbReq >= evm_dropWhen))
    {
        if (evm_dropCnt >= evm_dropNb)
        {
            evm_dropCnt = 0;
            evm_nbReq = 0;
        }
        else
        {
            if (evm_dropIo & EVM_DROPIO_WRITE)
            {
                evm_trace(EVM_TRACE_ERRORS, "EVM[%d]: dropping request...",
                          pcb->handle.whole);
                evm_dropCnt++;
                return NFX_OK;
            }
        }
    }
#endif // DEBUG

    sMsgSize = sizeof(scsi_writeReqMsg_t);
    totalLen = sizeof(sbm_sendControlAndSACData_t) + sMsgSize;
    /*
     * allocate eee message descriptor and a buffer
     * for the SBM/SCSI message
     */
    msgp = eee_allocateDesc(EEE_SHARED_POOL, NFX_COS_HIGH);
    if (msgp == NULL)
    {
        evm_trace(EVM_TRACE_ERRORS,
                  "[%d]: cannot allocate EEE descriptor",
                  pcb->handle.whole);
        goto evm_sendSbmScsiWriteMsg_err;
    }
    msgp->hdr.attr = 0;
    msgp->hdr.next = NULL;
    msgp->hdr.offset = 0;

    /* Fill in the EEE header
     * destination port is the SBM on MCPU
     */
    msgp->hdr.dest_port = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
                                                    NFX_NFP_TXRX_ANY,
                                                    sbmAppId,
                                                    EEE_CONTROL,
                                                    NFX_COS_HIGH,
                                                    PORT_BEST_FIT);
    msgp->hdr.src_port = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
                                                   nfxMyCPUId,
                                                   evmAppId,
                                                   EEE_USER_DATA,
                                                   NFX_COS_HIGH,
                                                   PORT_BEST_FIT);

    msgp->hdr.control = ((1 << EDESC_BUFCNT_SHIFT) | totalLen);
    sbmMsg = (sbm_sendControlAndSACData_t *)
        eee_allocateBuffer(EEE_SHARED_POOL,
                           EEE_BUF_SMALL,
                           NFX_COS_HIGH);
    if (sbmMsg == NULL)
    {
        evm_trace(EVM_TRACE_ERRORS,
                  "[%d]: cannot allocate SBM/SCSI message",
                  pcb->handle.whole);
        goto evm_sendSbmScsiWriteMsg_err;
    }
    msgp->bd[1].buf = (void *) sbmMsg;
    msgp->bd[1].len = totalLen;
    msgp->bd[1].control = EDESC_BD_CMD_IN_USE;
    
    /* Fill in the SBM message
     * dest_port filled in below
     */
    sbmMsg->cmd = CMD_TO_MCPU_SEND_CONTROL_AND_DATA_TO_SNIM;
    sbmMsg->source_port = msgp->hdr.src_port;
    sbmMsg->msg_len = sMsgSize;
    sbmMsg->COD = pcb->dataHandle.codHandle;
    sbmMsg->length = pcb->size * EVM_LBLK_SIZE;
    sbmMsg->offset = 0; // TODO: need offset in cod
    
    /* SCSI message follows the SBM message
     */
    wp = (scsi_writeReqMsg_t *) &sbmMsg->msg;
    
    /* Fill in the SCSI WRITE message
     */
    wp->hdr.mid = SCSI_MID_WRITE_REQ;
    wp->hdr.cHandle = pcb->handle.whole;
    rc = sdma_handleToPath(pcb->dev->openHandle, &(wp->hdr.path),
                           &sbmMsg->dest_port);
    if (rc == -1)
    {
        evm_trace(EVM_TRACE_ERRORS, "[%d]: sdma_handleToPath() failed",
                  pcb->handle.whole);
        goto evm_sendSbmScsiWriteMsg_err;        
    }
    wp->attributes = lcb->req.attributes;
    wp->blkOffset = pcb->offset;
    wp->blkCnt = pcb->size;
    wp->checkSum = pcb->checksum;
    wp->transSeqId = 0;
    evm_trace(EVM_TRACE_IO, "EVM IO[%d]: "
              "SBM/SCSI-WRITE -> sdma handle 0x%x path 0x%x dest_port 0x%x",
              pcb->handle.whole, pcb->dev->openHandle,
              wp->hdr.path, sbmMsg->dest_port);
    
    rc = eee_forwardPacket(msgp);
    switch (rc)
    {
    case EEE_DROP_PACKET:
    {
        /*
         * The packet was dropped by either the driver or
         * the EEE layer.  The descriptor has been freed.
         * The caller should make no further references to the packet.
         */
        msgp = NULL;   // set to NULL so we won't free
        evm_trace(EVM_TRACE_ERRORS,
                  "[%d]: EEE_DROP_PACKET from EEE forwardPacket",
                  pcb->handle.whole);
        EVM_STAT_EEEDROPPACKET;
        goto evm_sendSbmScsiWriteMsg_err;
    }
            
    case EEE_TX_QUEUE_FULL:
    {
        /*
         * The packet could not be queued due a tx queue full
         * condition.  The caller can take whatever action
         * it wants with the packet, since it is still the owner.
         */
        evm_trace(EVM_TRACE_ERRORS,
                  "[%d]: EEE_TX_QUEUE_FULL from EEE forwardPacket",
                  pcb->handle.whole);
        EVM_STAT_EEETXQUEUEFULL;
        goto evm_sendSbmScsiWriteMsg_err;
    }
    
    case EEE_FORWARDED_PACKET:
    {
        /*
         * The packet was queued for transmit successfully.
         * The caller should not free the packet (or touch
         * the packet in any way), as it is now owned by someone else.
         */
        evm_trace(EVM_TRACE_EEE,
                  "EVM EEE[%d]: SCSI/SBM msg type FORWARDED",
                  pcb->handle.whole);
        break;
    }
    
    default:
        break;
    } // switch

#ifdef DEBUG
    ATOMIC_INC_DWORD(&(evm_stats.ioPerFc[0][EEE_GET_DEST_CPU(sbmMsg->dest_port)].userWriteReq), 1);
#endif // DEBUG
    return NFX_OK;

evm_sendSbmScsiWriteMsg_err:

    /* Something went wrong
     * free up any allocated resources before returning err
     */
    if (msgp)
    {
        eee_freePacket(msgp);
    }

    return NFX_ERR;

}

/*------------------------------------------------------------------------------
 * Name         : evm_sendCreateCodMsg
 *
 * Description  : called to format and send CMD_TO_MCPU_MANIPULATE_CREATE
 *
 *                used by evm-test.c since normally the caller
 *                to evm_IOReq() will supply the cod
 *
 * Note         : use only for test purpose
 *
 * Created by   : Lan Chin
 *
 * Date Created : 10/01/01
 *------------------------------------------------------------------------------
 */

int32
evm_sendCreateCodMsg(nfx_cod_t *    codp,
                     uint32         key,
                     uint32         size,
                     uint32         fill)
    
{
    int32           rc;
    sbm_createCOD_t *sp;
    uint32 dst = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
                                           NFX_NFP_TXRX_ANY,
                                           sbmAppId,
                                           EEE_CONTROL,
                                           NFX_COS_HIGH,
                                           PORT_BEST_FIT);
    uint32 src = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
                                           nfxMyCPUId,
                                           evmAppId,
                                           EEE_CONTROL,
                                           NFX_COS_HIGH,
                                           PORT_BEST_FIT);
    EVM_RAM_ALLOC(sp,sbm_createCOD_t, sizeof(sbm_createCOD_t));
    if (sp == NULL)
    {
        evm_trace(EVM_TRACE_ERRORS,
                  "[]: cannot allocate RAM");
        return NFX_ERR;
    }
    sp->cmd = CMD_TO_MCPU_MANIPULATE_CREATE;
    sp->COD = COD_GET_NEXT();
    sp->len = size * EVM_LBLK_SIZE;
    sp->fill = fill;
    rc = eee_sendMessage((void *) sp, sizeof(sbm_createCOD_t), dst, src);
    if (rc != NFX_OK)
    {
        rc = NFX_ERR;
    }
    else
    {
        *codp = sp->COD;
    }

    /*
     * free message buffer before returning
     * eee_sendMessage() does not
     */
    EVM_RAM_DEALLOC(sp);
    return rc;

} /* evm_sendCreateCodMsg */


/*------------------------------------------------------------------------------
 * Name         : evm_sendFreeCodMsg
 *
 * Description  : called to format and send CMD_TO_MCPU_MANIPULATE_FREE
 *
 *                used by evm-test.c since normally the caller
 *                to evm_IOReq() will supply the cod
 *
 * Note         : use only for test purpose
 *
 * Created by   : Lan Chin
 *
 * Date Created : 10/01/01
 *------------------------------------------------------------------------------
 */

int32
evm_sendFreeCodMsg(nfx_cod_t cod)
    
{
    int32           rc;
    sbm_freeCOD_t   *sp;
    uint32 dst = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
                                           NFX_NFP_TXRX_ANY,
                                           sbmAppId,
                                           EEE_CONTROL,
                                           NFX_COS_HIGH,
                                           0);
    uint32 src = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
                                           nfxMyCPUId,
                                           evmAppId,
                                           EEE_CONTROL,
                                           NFX_COS_HIGH,
                                           PORT_BEST_FIT);
    EVM_RAM_ALLOC(sp, sbm_freeCOD_t, sizeof(sbm_freeCOD_t));
    if (sp == NULL)
    {
        evm_trace(EVM_TRACE_ERRORS, "[]: cannot allocate RAM");
        return NFX_ERR;
    }
    sp->cmd = CMD_TO_MCPU_MANIPULATE_FREE;
    sp->COD = cod;
    rc = eee_sendMessage((void *) sp, sizeof(sbm_freeCOD_t), dst, src);
    if (rc != NFX_OK)
    {
        rc = NFX_ERR;
    }
    /*
     * free message buffer before returning
     * eee_sendMessage() does not
     */
    EVM_RAM_DEALLOC(sp);
    return rc;
}

/*----------------------------------------------------------------------------*/
