/*------------------------------------------------------------------------------
 * RCS:  $Id: evm-prof.c,v 1.7 2002/10/22 01:30:32 jong Exp $
 * Copyright (C) 2002, Agile Storage, Inc.
 *
 * Name		 : evm-prof.c
 *
 * Description  : 
 *
 * Created by   : Lan Chin
 *				Isabelle Cnudde
 *              Shyh-Pei Yen
 *
 * Date Created : 25-JUN-2002
 *
 *-----------------------------------------------------------------
 */

#include "nfx-incl.h"
#include "exports/hash-api.h"
#include "exports/timer-api.h"
#include "evm-lv.h"
#include "evm-msg.h"
#include "evm-debug.h"

#ifdef SY_IO_PERFORMANCE_TEST

/* this skip the whole performance testing code if not defined... */

#define USE_NEW_COD_EVERYTIME


/*------------------------------------------------------------------------------
 * Type definition.
 *------------------------------------------------------------------------------
 */

typedef struct
{
	int32		key;	// hash key
	int32		vidx;	// volume index for this test (always start from 0)
	int32		id;		// test id
	evm_pReq_t	type;	// request type
	evm_IOReq_t	req;
	char		*buf;	// write data
	int32		len;	// nbytes of data
	ushort16	chksum;	// calculated write checksum
	eeeGMT_t	ts;		// timestamp at request sent
} evm_perf_cb_t;

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

static int32   evm_srvr_test_init = 0;

#define EVM_SRVR_TEST_HASH_BUCKETS  255

static address_t	evm_srvr_test_hash;
static int32		evm_srvr_test_seqno = 1;

#define EVM_MAX_NUM_OF_LVS		10
#define EVM_MAX_NUM_OF_TESTS	10000

#define EVM_PERF_FIXED		0
#define EVM_PERF_SEQ		1
#define EVM_PERF_SKIPSEQ	2
#define EVM_PERF_RAND		3
#define EVM_PERF_LOOPBACK	4


/* command line parameters */
static uint32			totalvols;
static int32			perf_type;		/* m_write/m_read/u_write/u_read */
#if 0
static int32			perf_respt;		/* for checking respone time only */
#endif
static evm_volOffset_t	perf_offset;	/* LBA offset */
static int32			perf_nBlks;		/* total blocks to read/write */
static int32			perf_opoffset;	/* fixed|skip|seq|rand */
static uint32			perf_nreqs;		/* total requests to send out for each volume */
static uint32			perf_Qlen;		/* keep request queue to this length for each volume */
static evm_vol_t        *perf_lvp[EVM_MAX_NUM_OF_LVS];	/* LUN volume array */
static uint32			perf_fill;

/* stats keeper */
static uint32		nextid[EVM_MAX_NUM_OF_LVS];
static uint32		totalrsps[EVM_MAX_NUM_OF_LVS];


/*------------------------------------------------------------------------------
 * Local function prototypes.
 *------------------------------------------------------------------------------
 */

static uint32
evm_srvr_test_hashFunc(uint32 key);

static int32
evm_srvr_test_hashCompare(evm_perf_cb_t *cb, int32	compare_key);

static void
evm_srvr_test_hashRelease(evm_perf_cb_t *cb);

static evm_perf_cb_t *
evm_createPerfCb(void);

static void
evm_deletePerfCb(evm_perf_cb_t *cb);

static ushort16
evm_doChksum(ushort16 *buf, int32 len);

static int32
evm_doNextWrite(evm_perf_cb_t *cb);

static int32
evm_doWrite(int32			vidx,
			int32			n,
			evm_pReq_t		type,
			evm_vol_t	    *lv,
			evm_volOffset_t	offset,
			evm_volSize_t	size,
			boolean			doUltra);


/*-----------------------------------------------------------------
 * Name:		evm_srvr_test_hashFunc
 *
 * Description: hash function for the general hash routine
 *			  (see hash-api.h)
 *
 *			  called to return the key or hash bucket
 *			  for the test hash
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static uint32
evm_srvr_test_hashFunc(uint32 key)
	
{
	return (key % EVM_SRVR_TEST_HASH_BUCKETS);
} /* evm_srvr_test_hashFunc */

/*-----------------------------------------------------------------
 * Name:		evm_srvr_test_hashCompare
 *
 * Description: compare function for the general hash routine
 *			  (see hash-api.h)
 *
 *			  called to check for a key match for the test hash
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static int32
evm_srvr_test_hashCompare(evm_perf_cb_t *cb, // data stored in hash
						 int32 compare_key)  // data to compare with
{

	if (cb->key == compare_key)
	{
		/* return 0 for a match */
		return 0;
	}

	return 1;

} /* evm_srvr_test_hashCompare */

/*-----------------------------------------------------------------
 * Name:		evm_srvr_test_hashRelease
 *
 * Description: release function for the general hash routine
 *			  (see hash-api.h)
 *
 * Note:		only called by hash module if the entire
 *			  hash table is destoryed !! 
 *			  will NOT be called for a hash_delete()
 *
 *			  called to deallocate memory (if necc) for a node
 *			  in the IO request hash
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static void
evm_srvr_test_hashRelease(evm_perf_cb_t * cb)
	
{
	/*
	 * we have to deallocate the write buffer as well
	 * since evm will set eee control to EDESC_BD_CMD_NO_DEALLOC
	 */
	if (cb->buf)
	{
		void *bufBase;

		if (eee_isUltraBuf(cb->buf, &bufBase))
		{
			EVM_DEALLOC_BUF(cb->buf);
		}
		else
		{
			EVM_RAM_DEALLOC(cb->buf);
		}
	}
	EVM_RAM_DEALLOC(cb);

	return;

} /* evm_srvr_test_hashRelease */

/*-----------------------------------------------------------------
 * Name:		evm_createPerfCb
 *
 * Description: called to create a test control block
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static evm_perf_cb_t *
evm_createPerfCb(void)
	
{
	int32		   rc;
	evm_perf_cb_t   *cb;

	EVM_RAM_ALLOC_ZERO(cb, evm_perf_cb_t, sizeof(evm_perf_cb_t));
	if (cb == NULL)
		return NULL;

	cb->key = evm_srvr_test_seqno++;
	/* a signature for evm-io to send qdepth response */
	cb->req.clientPrivate = cb->key;
	rc = hash_insert(evm_srvr_test_hash, (void *) ((uintptr_t)cb->key),
					  (void *) ((uintptr_t)cb));
	if (rc == NFX_ERR)
	{
		evm_srvr_test_hashRelease(cb);
		return NULL;
	}

	return cb;

} /* evm_createPerfCb */

/*-----------------------------------------------------------------
 * Name:		evm_deletePerfCb
 *
 * Description: called to delete a test control block
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static void
evm_deletePerfCb(evm_perf_cb_t * cb)
	
{

	hash_delete(evm_srvr_test_hash, (void *) ((uintptr_t)cb->key),
				(void *) ((uintptr_t)cb->key));
	evm_srvr_test_hashRelease(cb);

} /* evm_deletePerfCb */

/*-----------------------------------------------------------------
 * Name:		evm_doChksum
 *
 * Description: called to calculate checksum
 *			  based on standard TCP/IP checksum algorithm
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static ushort16
evm_doChksum(ushort16 * buf,
			 int32	  len)
	
{
	ushort16	sum=0;
	int32	   n=len;

	while (n--)
		sum += *buf++;

	while (sum > 0xffff)
		sum = (sum & 0xffff) + (sum >> 16l);

	sum = (~sum & 0xffff);
	return sum;

} /* evm_doChksum */


/*-----------------------------------------------------------------
 * Name:		evm_writeComplete
 *
 * Description: test completion routine for evm_IOReq
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static void
evm_writeComplete(evm_IOStat_t * stat)
{
	int32			i, vidx;
	int32			rc, key;
	evm_perf_cb_t	*cb;
	eeeGMT_t		now;

	key = stat->clientPrivate;
	cb = (evm_perf_cb_t *) hash_lookup(evm_srvr_test_hash,
									   (void *) ((uintptr_t)key),
									   (void *) ((uintptr_t)key));
	if (cb == NULL)
	{
		evm_trace(EVM_TRACE_ERRORS, "\nEVM PERF IO: write cb ERROR, cannot find perf cb for key %d\n",
			   key);
		return;
	}

	if (stat->status != EVM_STAT_NO_ERROR)
	{
		evm_trace(EVM_TRACE_ERRORS,
			"\nEVM performance test done due to request failed: %d\n", stat->status);
		evm_deletePerfCb(cb);
		return;
	}

	vidx = cb->vidx;
	totalrsps[vidx]++;

#ifdef USE_NEW_COD_EVERYTIME
	/* free the current user write COD */
	if (cb->type == u_write)
	{
        evm_sendFreeCodMsg(cb->req.wDataHandle.codHandle);
        cb->req.wDataHandle.codHandle = NULL;
	}
#endif

	if (totalrsps[vidx] >= perf_nreqs)
	{
		evm_trace(EVM_TRACE_ERRORS, "EVM performance test for vol[%d] done", vidx);
		evm_deletePerfCb(cb);
		for (i = 0; i < totalvols; i++)
		{
			if (totalrsps[vidx] < perf_nreqs)
				break;
		}
		if (i >= totalvols)
		{
			eee_TimeGMT(&now);
			evm_trace(EVM_TRACE_ERRORS, "All done at %d:%d", now.sec, now.usec);
		}
		return;
	}

	/* issue next write request */
	if (nextid[vidx] < perf_nreqs)
	{
		rc = evm_doNextWrite(cb);
		if (rc == NFX_ERR)
		{
			evm_trace(EVM_TRACE_ERRORS, "\nERROR[%d]: cannot do subsequent META WRITE for offset %llx nBlks %lld\n",
				cb->id, cb->req.offset, cb->req.reqSize);
		}
	}
	else
	{
		/* since we already sent enough requests */
		/* we'll no long need to use this cb     */
		if ((cb->type == u_write) && cb->req.wDataHandle.codHandle)
		{
			evm_sendFreeCodMsg(cb->req.wDataHandle.codHandle);
			cb->req.wDataHandle.codHandle = NULL;
		}
		evm_deletePerfCb(cb);
	}

	return;

} /* evm_writeComplete */


/*-----------------------------------------------------------------
 * Name:		evm_doNextWrite
 *
 * Description: called to test EVM srvr meta or user write IO
 *
 * Created by : Shyh-Pei Yen
 *
 * Date Created: 6/25/02
 *-----------------------------------------------------------------
 */

static int32
evm_doNextWrite(evm_perf_cb_t *cb)
{
	int32		   rc;
	evm_volOffset_t	offset;

	if (perf_opoffset == EVM_PERF_SEQ)
		offset = cb->req.offset - (cb->id * cb->req.reqSize);
	else if (perf_opoffset == EVM_PERF_SKIPSEQ)
		offset = cb->req.offset - (cb->id * (cb->req.reqSize+2));
	cb->id = nextid[cb->vidx]++;
	if (perf_opoffset == EVM_PERF_SEQ)
		cb->req.offset = offset + (cb->id * cb->req.reqSize);
	else if (perf_opoffset == EVM_PERF_SEQ)
		cb->req.offset = offset + (cb->id * (cb->req.reqSize+2));

#ifdef USE_NEW_COD_EVERYTIME
	if (cb->type == u_write)
	{
		/* allocate another COD for next write */
		rc = evm_sendCreateCodMsg(&cb->req.wDataHandle.codHandle,
								  cb->key, cb->req.reqSize, perf_fill);
		if (rc == NFX_ERR)
		{
			/* callback evm_writeComplete deletes cb if err */
			evm_trace(EVM_TRACE_ERRORS, "EVM perf userwr: can't create cod for next write\n");
			return NFX_ERR;
		}
	}
#endif

	rc = evm_IOReq(&cb->req, evm_writeComplete);
	if (rc == NFX_ERR)
	{
		evm_trace(EVM_TRACE_ERRORS, "\nEVM TEST IO[%d]: evmIOReq/WRITE FAILED\n", cb->id);
		return NFX_ERR;
	}

	return NFX_OK;

} /* evm_doNextWrite */


/*-----------------------------------------------------------------
 * Name:		evm_doWrite
 *
 * Description: called to unit test EVM srvr meta or user WRITE
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static int32
evm_doWrite(int32				vidx,
			int32				n,
			evm_pReq_t			type,
			evm_vol_t		    *lv,
			evm_volOffset_t		offset,
			evm_volSize_t		size,
			boolean				doUltra)
{
	int32		   rc;
	evm_perf_cb_t   *cb;
	char			*buf;

	cb = evm_createPerfCb();
	if (cb == NULL)
		return NFX_ERR;

	//cb->type = type;
	cb->req.reqType = EVM_WRITE_REQ;
	cb->req.dataType = (type == m_write) ? EVM_META_DATA_REQ : EVM_USER_DATA_REQ;
	cb->req.id = lv->id;
	cb->req.offset = offset;
	cb->req.reqSize = size;
	if (perf_opoffset == EVM_PERF_LOOPBACK)
	{
		cb->req.attributes = SCSI_MSG_ATTR_LOOPBACK;
	}
	cb->id = n;
	cb->vidx = vidx;
	cb->len = size * EVM_LBLK_SIZE;
	perf_fill = (cb->key % EVM_SRVR_TEST_HASH_BUCKETS);

	if (type == m_write)
	{
		if (doUltra)
		{
			NFX_ASSERT(EEE_BUFSIZE_ULTRA > 8192);
			if (cb->len > 8192)
			{
				evm_trace(EVM_TRACE_ERRORS, "evmadmin: truncating to max ultra buf 8192\n");
				cb->len = 8192;
			}
			buf = eee_allocateBuffer(EEE_SHARED_POOL, EEE_BUF_ULTRA, NFX_COS_HIGH);
			// Preserve the unused leader for packet headers.
			buf += EEE_BUFSIZE_ULTRA - 8192;
		}
		else
		{
			EVM_RAM_ALLOC_ALIGNED(buf, char, cb->len);
		}
		if (buf == NULL)
		{
			evm_trace(EVM_TRACE_ERRORS, "\nEVM perf metawr[%d]: no mem for write buf\n", n);
			evm_deletePerfCb(cb);
			return NFX_ERR;
		}
		memset((void *) buf, perf_fill, cb->len);
		{
			int i;
			//evm_trace(EVM_TRACE_ERRORS, "len = %d\n", cb->len);
			for (i = 0; i < cb->len; i++)
			{
				buf[i] = i & 0xff;
			} // for
		}
		cb->chksum = evm_doChksum((ushort16 *) buf, cb->len);
		cb->buf = buf;

		cb->req.wDataHandle.dataPtr = buf;
	}
	else
	{
		/*
		 * user write, allocate COD and perf_fill with pattern
		 */
		cb->buf = NULL;
		rc = evm_sendCreateCodMsg(&cb->req.wDataHandle.codHandle,
								  cb->key, size, perf_fill);
		if (rc == NFX_ERR)
		{
			/* callback evm_writeComplete deletes cb if err */
			evm_trace(EVM_TRACE_ERRORS, "\nEVM PERF USER IO[%d]: cannot create cod\n", n);
			return NFX_ERR;
		}
	}

	/* put timestamp into cb */
	//eee_TimeGMT(&cb->ts);

	rc = evm_IOReq(&cb->req, evm_writeComplete);
	if (rc == NFX_ERR)
	{
		/* callback evm_writeComplete deletes cb if err */
		evm_trace(EVM_TRACE_ERRORS, "\nEVM PERF IO[%d]: evmIOReq/WRITE FAILED\n", n);
		return NFX_ERR;
	}

	return NFX_OK;

} /* evm_doWrite */


/*-----------------------------------------------------------------
 * Name:		evm_perf_usage
 *-----------------------------------------------------------------
 */
static void
evm_perf_usage()
{
	evm_trace(EVM_TRACE_ERRORS,
			"USAGE: evmadmin perf_evm_prof {metawr} {fixed|seq|skip|loopback} "
			"<offset> <nsectors> <totalreqs> <num-of-vol> <Qnreqs>");
	evm_trace(EVM_TRACE_ERRORS,
			"       evmadmin perf_evm_prof where");
	return;
}


/*-----------------------------------------------------------------
 * Name:		evm_perf
 *
 * Description: called to performance test EVM srvr IO
 * Created by : Lan Chin
 *				Shyh-Pei Yen
 * Date Created: 6/26/02
 *-----------------------------------------------------------------
 */

int32
evm_perf_prof(int32 argc, char8 *argv[])
{
	int32			rc, argi;
	int32			i, j;
	int32			lvToIdx;
	evm_vol_t       *lv;
	boolean			doUltra = FALSE;
	eeeGMT_t		now;

	if (evm_srvr_test_init == 0)
	{
		evm_srvr_test_init = 1;
		evm_srvr_test_hash = hash_create(EVM_SRVR_TEST_HASH_BUCKETS,
										 evm_srvr_test_hashFunc,
										 evm_srvr_test_hashCompare,
										 evm_srvr_test_hashRelease);
	}

	/* check number of arguments */
	if (argc < 2)
	{
		evm_perf_usage();
		return NFX_ERR;
	}

	/* NOW PARSE THE PARAMETERS */
	argi = 1;

	/* ==================================== */
	/* what type of operation? meta write   */
	/* or meta read or user write or user   */
	/* read?                                */
	if (strcmp(argv[argi], "where") == 0)
	{
		for (i = 0; i < totalvols; i++)
			evm_trace(EVM_TRACE_ERRORS, "Vol[%d] is currently at %d", i, nextid[i]);
		return NFX_OK;
	}
	else if (strcmp(argv[argi], "metawr") == 0)
	{
		/* check number of arguments */
		if (argc < 8)
		{
			evm_perf_usage();
			return NFX_ERR;
		}

		perf_type = m_write;
		doUltra = FALSE;
	}
	else if (strcmp(argv[argi], "userwr") == 0)
	{
		/* check number of arguments */
		if (argc < 8)
		{
			evm_perf_usage();
			return NFX_ERR;
		}

		perf_type = u_write;
		doUltra = FALSE;
	}
	else
	{
		evm_perf_usage();
		return NFX_ERR;
	}
	argi++;

	/* initialized variables and counters */
	for (i=0; i<EVM_MAX_NUM_OF_LVS; i++)
	{
		perf_lvp[i] = 0;
		nextid[i] = 0;
		totalrsps[i] = 0;
	}
	totalvols = 0;

	/* ==================================== */
	/* is it a fixed or sequential or       */
	/* random offset for each read/write?   */
	if (strcmp(argv[argi], "fixed") == 0)
		perf_opoffset = EVM_PERF_FIXED;
	else if (strcmp(argv[argi], "seq") == 0)
		perf_opoffset = EVM_PERF_SEQ;
#if 0
/* not yet supported */
	else if (strcmp(argv[argi], "rand") == 0)
		perf_opoffset = EVM_PERF_RAND;
#endif
	else if (strcmp(argv[argi], "skip") == 0)
		perf_opoffset = EVM_PERF_SKIPSEQ;
	else if (strcmp(argv[argi], "loopback") == 0)
		perf_opoffset = EVM_PERF_LOOPBACK;
	else
	{
		evm_perf_usage();
		return NFX_ERR;
	}
	argi++;

	/* ==================================== */
	/* get volume offset and number of      */
	/* sectors to be read/write             */
	perf_offset = atol(argv[argi++]);
	perf_nBlks = atol(argv[argi++]);
	if (perf_nBlks % 2)
	{
		evm_trace(EVM_TRACE_ERRORS, "\nERROR: scsi checksum reqires even number block io\n");
		return NFX_ERR;
	}

	/* ==================================== */
	/* total requests to send out           */
	perf_nreqs = atoi(argv[argi++]);

	/* ==================================== */
	/* get the volumes (from and to indexes */
	lvToIdx = atoi(argv[argi++]);
	if (lvToIdx <= 0)
	{
		evm_perf_usage();
		return NFX_ERR;
	}

	NFX_SPIN_LOCK(&evm_lv_lock);
	DLL_ITERATE(lv, &evm_volList, entry)
	{
		if (lv == NULL)
			break;
		if (lvToIdx > 0)
			lvToIdx--;
		else
			break;
		perf_lvp[i] = lv;
		totalvols++;
	} // DLL_ITERATE
	NFX_SPIN_UNLOCK(&evm_lv_lock);
	if (lvToIdx > 0)
	{
		evm_trace(EVM_TRACE_ERRORS,
			"No. of volumes specified exceeds max %d volumes this test supported.",
			EVM_MAX_NUM_OF_LVS);
	}

	/* no volume? */
	if (perf_lvp[0] == NULL)
	{
		evm_trace(EVM_TRACE_ERRORS, "No logical volume found...\n");
		return NFX_ERR;
	}

	/* ==================================== */
	/* how many requests to sent to queue   */
	/* (maintain that queue length until    */
	/* all requests are sent)               */
	perf_Qlen = atol(argv[argi++]);

	if (perf_nreqs < perf_Qlen)
		perf_Qlen = perf_nreqs;

	eee_TimeGMT(&now);
	evm_trace(EVM_TRACE_ERRORS, "Starting at %d:%d", now.sec, now.usec);

	if ((perf_type == m_write) || (perf_type == u_write))
	{
		for (i=0; i<totalvols; i++)
		{
			evm_volOffset_t	offset1 = perf_offset;

			evm_trace(EVM_TRACE_ERRORS,
					"evmadmin performance test: volID=%llx nsectors=%d totalreqs=%d offset=%lld Qnreqs=%d",
					perf_lvp[i]->id.whole_thing, perf_nBlks, perf_nreqs, offset1, perf_Qlen);
			nextid[i] = perf_Qlen;
			for (j=0; j<perf_Qlen; j++)
			{
				rc = evm_doWrite(i, j, perf_type, perf_lvp[i], offset1, perf_nBlks, doUltra);
				if (rc == NFX_ERR)
				{
					evm_trace(EVM_TRACE_ERRORS,
							"\nERROR[%d]: cannot do META WRITE for vol %d offset %llx nBlks %lld\n",
							i, offset1, perf_nBlks);
				}
				if (perf_opoffset == EVM_PERF_SEQ)
					offset1 += perf_nBlks;
				if (perf_opoffset == EVM_PERF_SKIPSEQ)
					offset1 += perf_nBlks + 2;
			}
		}
	}

	return NFX_OK;

} /* evm_perf */

/*----------------------------------------------------------------------------*/

#endif // SY_IO_PERFORMANCE_TEST
