/*------------------------------------------------------------------------------
 * RCS:  $Id: evm-test.c,v 1.18 2002/10/22 01:30:32 jong Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : evm-test.c
 *
 * Description  : 
 *
 * Created by   : Lan Chin
 *                Isabelle Cnudde
 *
 * Date Created : 28-SEP-2001
 *
 *-----------------------------------------------------------------
 */

#include "nfx-incl.h"
#include "exports/hash-api.h"
#include "evm-lv.h"
#include "evm-msg.h"
#include "evm-test.h"

#ifdef ISABELLE // not used
/*------------------------------------------------------------------------------
 * Type definition.
 *------------------------------------------------------------------------------
 */

typedef struct
{
    int32       key;        // hash key
    int32       id;         // test id
    evm_pReq_t  type;       // request type
    evm_IOReq_t req;
    char        *buf;       // write data
    int32       len;        // nbytes of data
    ushort16    chksum;     // calculated write checksum
} evm_test_cb_t;

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

static int32   evm_srvr_test_init = 0;

#define EVM_SRVR_TEST_HASH_BUCKETS  255

static address_t   evm_srvr_test_hash;
static int32   evm_srvr_test_seqno = 1;

/*------------------------------------------------------------------------------
 * Local function prototypes.
 *------------------------------------------------------------------------------
 */

static uint32
evm_srvr_test_hashFunc(uint32 key);

static int32
evm_srvr_test_hashCompare(evm_test_cb_t *   cb,
                          int32             compare_key);

static void
evm_srvr_test_hashRelease(evm_test_cb_t * cb);

static evm_test_cb_t *
evm_createTestCb(void);

static void
evm_deleteTestCb(evm_test_cb_t * cb);

static ushort16
evm_doChksum(ushort16 * buf,
             int32      len);

static void
evm_readComplete(evm_IOStat_t * stat);

static int32
evm_doRead(evm_test_cb_t * cb);

static void
evm_writeComplete(evm_IOStat_t * stat);


static int32
evm_doWrite(int32               n,
            evm_pReq_t          type,
            evm_vol_t *   lv,
            evm_volOffset_t     offset,
            evm_volSize_t       size,
            boolean             doUltra);

/*-----------------------------------------------------------------
 * Name:        evm_srvr_test_hashFunc
 *
 * Description: hash function for the general hash routine
 *              (see hash-api.h)
 *
 *              called to return the key or hash bucket
 *              for the test hash
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static uint32
evm_srvr_test_hashFunc(uint32 key)
    
{

    return (key % EVM_SRVR_TEST_HASH_BUCKETS);

} /* evm_srvr_test_hashFunc */

/*-----------------------------------------------------------------
 * Name:        evm_srvr_test_hashCompare
 *
 * Description: compare function for the general hash routine
 *              (see hash-api.h)
 *
 *              called to check for a key match for the test hash
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static int32
evm_srvr_test_hashCompare(evm_test_cb_t *cb, // data stored in hash
                         int32 compare_key)  // data to compare with
{

    if (cb->key == compare_key)
    {
        /* return 0 for a match */
        return 0;
    }

    return 1;

} /* evm_srvr_test_hashCompare */

/*-----------------------------------------------------------------
 * Name:        evm_srvr_test_hashRelease
 *
 * Description: release function for the general hash routine
 *              (see hash-api.h)
 *
 * Note:        only called by hash module if the entire
 *              hash table is destoryed !! 
 *              will NOT be called for a hash_delete()
 *
 *              called to deallocate memory (if necc) for a node
 *              in the IO request hash
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static void
evm_srvr_test_hashRelease(evm_test_cb_t * cb)
    
{
    /*
     * we have to deallocate the write buffer as well
     * since evm will set eee control to EDESC_BD_CMD_NO_DEALLOC
     */
    if (cb->buf)
    {
        void *bufBase;

        if (eee_isUltraBuf(cb->buf, &bufBase))
        {
            EVM_DEALLOC_BUF(cb->buf);
        }
        else
        {
            EVM_RAM_DEALLOC(cb->buf);
        }
    }
    EVM_RAM_DEALLOC(cb);

    return;

} /* evm_srvr_test_hashRelease */

/*-----------------------------------------------------------------
 * Name:        evm_createTestCb
 *
 * Description: called to create a test control block
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static evm_test_cb_t *
evm_createTestCb(void)
    
{
    int32           rc;
    evm_test_cb_t   *cb;

    EVM_RAM_ALLOC_ZERO(cb, evm_test_cb_t, sizeof(evm_test_cb_t));
    if (cb == NULL)
    {
        return NULL;
    }
    cb->key = evm_srvr_test_seqno++;
    cb->req.clientPrivate = cb->key;
    rc = hash_insert(evm_srvr_test_hash,
					 (void *) ((uintptr_t)cb->key),
					 (void *) ((uintptr_t)cb));
    if (rc != NFX_OK)
    {
        evm_srvr_test_hashRelease(cb);
        return NULL;
    }

    return cb;

} /* evm_createTestCb */

/*-----------------------------------------------------------------
 * Name:        evm_deleteTestCb
 *
 * Description: called to delete a test control block
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static void
evm_deleteTestCb(evm_test_cb_t * cb)
    
{

    hash_delete(evm_srvr_test_hash, (void *) ((uintptr_t)cb->key),
				(void *) ((uintptr_t)cb->key));
    evm_srvr_test_hashRelease(cb);

} /* evm_deleteTestCb */

/*-----------------------------------------------------------------
 * Name:        evm_doChksum
 *
 * Description: called to calculate checksum
 *              based on standard TCP/IP checksum algorithm
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static ushort16
evm_doChksum(ushort16 * buf,
             int32      len)
    
{
    ushort16    sum=0;
    int32       n=len;

    while (n--)
    {
        sum += *buf++;
    }
    while (sum > 0xffff)
    {
        sum = (sum & 0xffff) + (sum >> 16l);
    }
    sum = (~sum & 0xffff);
    return sum;

} /* evm_doChksum */

/*-----------------------------------------------------------------
 * Name:        evm_readComplete
 *
 * Description: test completion routine for evm_IOReq Read
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static void
evm_readComplete(evm_IOStat_t * stat)
    
{
    int32           key;
    ushort16        chksum;
    uchar8          *bufp;
    evm_test_cb_t   *cb;

    key = stat->clientPrivate;
    cb = (evm_test_cb_t *) hash_lookup(evm_srvr_test_hash,
                                       (void *) ((uintptr_t)key),
									   (void *) ((uintptr_t)key));
    if (cb == NULL)
    {
        return;
    }
    /*
     * if read ok, calculate checksum and make sure
     * it matches with the write
     */
    if (stat->status != EVM_STAT_NO_ERROR)
    {
        evm_trace(EVM_TRACE_DETAIL, "\nEVM TEST IO[%d]: ERROR for READ COMPLETION status %d\n",
               cb->id, stat->status);
#ifdef ISABELLE // debug only
		{
			evm_trace(EVM_TRACE_DETAIL, "Print 512 first bytes read...\n");
			bufp = stat->rDataHandle.dataPtr;
			bmc12500Eth_printBuf(bufp, cb->len);
		}
#endif // ISABELLE
        evm_deleteTestCb(cb);
        return;
    }
    if (cb->type == m_write)
    {
		int i;
		boolean readOk = TRUE;

        /*
         * calculate read result with original checksum
         */
        bufp = stat->rDataHandle.dataPtr;
		
		for (i = 0; i < cb->len; i++)
		{
			if (bufp[i] != (i & 0xff))
			{
				evm_trace(EVM_TRACE_DETAIL, "failed at %d, expected %02x, got %02x\n", 
					   i, (i & 0xff), bufp[i]);
				//bmc12500Eth_printBuf(bufp, cb->len);
				readOk = FALSE;
			}
		} // for

		evm_trace(EVM_TRACE_DETAIL, "EVM TEST META IO[%d]: READ PASS\n", cb->key);
#ifdef NO_SUCH_DEF
        chksum = evm_doChksum((ushort16 *) bufp, cb->len);
        if (chksum != cb->chksum)
		{
		}
#endif 
		if (readOk == FALSE)
        {
            evm_trace(EVM_TRACE_DETAIL, "EVM TEST META IO[%d]: READ FAILED\n", cb->key);
            evm_trace(EVM_TRACE_DETAIL, "    status %d chksum 0x%x (calculated 0x%x)\n",
                   stat->status, chksum, cb->chksum);
        }
        else
        {
            evm_trace(EVM_TRACE_DETAIL, "EVM TEST META IO[%d]: READ PASS\n", cb->key);
        }
        EVM_DEALLOC_BUF(stat->rDataHandle.dataPtr);
    }
    else // u_read
    {
        /*
         * watchout:
         * with sbm-app-rx.c 1.8,
         * we do not get an ACK for the retrieve sac msg
         * so just assume read user data is ok
         */
        evm_trace(EVM_TRACE_DETAIL, "EVM TEST USER IO[%d]: READ PASS\n", cb->key);
        evm_sendFreeCodMsg(cb->req.wDataHandle.codHandle); // free write COD
        evm_sendFreeCodMsg(stat->rDataHandle.codHandle);   // free read COD
#if NO_SUCH_DEF
        rc = evm_sendRetrieveSacData(cb->req.wDataHandle.codHandle,
                                     cb->len,
                                     cb->req.offset,
                                     cb->key,
                                     (uint32) cb->buf);
        if (rc == NFX_OK)
        {
            evm_trace(EVM_TRACE_DETAIL, "EVM TEST USER IO[%d]: READ COMPLETE, issued SAC read\n",
                   cb->key);
        }
        else
        {
            /*
             * for user tests, we allocate the cod, so we need to free
             */
            evm_trace(EVM_TRACE_DETAIL, "EVM TEST USER IO[%d]: READ COMPLETE, "
                   "error retreiving SAC data\n",
                   cb->key);
            evm_sendFreeCodMsg(cb->req.wDataHandle.codHandle);
        }
#endif
    }
    
    evm_deleteTestCb(cb);
} /* evm_readComplete */

/*-----------------------------------------------------------------
 * Name:        evm_doRead
 *
 * Description: called to unit test EVM srvr meta or user read IO
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static int32
evm_doRead(evm_test_cb_t * cb)
    
{
    int32           rc;

    /*
     * just fill in req type, leave rest alone
     */
    cb->req.reqType = EVM_READ_REQ;
    rc = evm_IOReq(&cb->req, evm_readComplete);
    if (rc != NFX_OK)
    {
        evm_trace(EVM_TRACE_DETAIL, "\nEVM TEST IO[%d]: evmIOReq/READ FAILED\n", cb->id);
        evm_deleteTestCb(cb);
        return NFX_ERR;
    }
    evm_trace(EVM_TRACE_DETAIL, "EVM TEST IO[%d]: evmIOReq/READ OK\n", cb->id);
    evm_trace(EVM_TRACE_DETAIL, "                 offset %lld size %lld\n",
           cb->req.offset, cb->req.reqSize);

    return NFX_OK;

} /* evm_doRead */

/*-----------------------------------------------------------------
 * Name:        evm_writeComplete
 *
 * Description: test completion routine for evm_IOReq
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static void
evm_writeComplete(evm_IOStat_t * stat)
    
{
    int32           rc, key;
    evm_test_cb_t   *cb;

    key = stat->clientPrivate;
    cb = (evm_test_cb_t *) hash_lookup(evm_srvr_test_hash,
                                       (void *) ((uintptr_t)key),
									   (void *) ((uintptr_t)key));
    if (cb == NULL)
    {
        evm_trace(EVM_TRACE_DETAIL, "\nEVM TEST IO: write cb ERROR, cannot find test cb for key %d\n",
               key);
        return;
    }
    /*
     * if write ok, read back the data
     */
    if (stat->status != EVM_STAT_NO_ERROR)
    {
        evm_trace(EVM_TRACE_DETAIL, "\nEVM TEST IO[%d]: ERROR for WRITE COMPLETION status %d\n",
               cb->id, stat->status);
        evm_deleteTestCb(cb);
        return;
    }
    rc = evm_doRead(cb);
    if (rc != NFX_OK)
    {
        evm_trace(EVM_TRACE_DETAIL, "\nEVM TEST META IO[%d]: ERROR for READ BACK\n", cb->id);
        /* callback evm_readComplete deletes cb if err */
        //evm_deleteTestCb(cb);
        return;
    }
    evm_trace(EVM_TRACE_DETAIL, "EVM TEST IO[%d]: WRITE PASSED, issued READ BACK\n", cb->id);

} /* evm_writeComplete */


/*-----------------------------------------------------------------
 * Name:        evm_doWrite
 *
 * Description: called to unit test EVM srvr meta or user WRITE
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

static int32
evm_doWrite(int32               n,
            evm_pReq_t          type,
            evm_vol_t *   lv,
            evm_volOffset_t     offset,
            evm_volSize_t       size,
            boolean             doUltra)
    
{
    int32           rc;
    uint32          fill;
    evm_test_cb_t   *cb;
    char            *buf;

    cb = evm_createTestCb();
    if (cb == NULL)
    {
        return NFX_ERR;
    }
    cb->type = type;
    cb->req.reqType = EVM_WRITE_REQ;
    cb->req.dataType = (type == m_write) ? EVM_META_DATA_REQ : EVM_USER_DATA_REQ;
    cb->req.id = lv->id;
    cb->req.offset = offset;
    cb->req.reqSize = size;
    cb->id = n;
    cb->len = size * EVM_LBLK_SIZE;
    fill = (cb->key % EVM_SRVR_TEST_HASH_BUCKETS);
    if (doUltra)
    {
        NFX_ASSERT(EEE_BUFSIZE_ULTRA > 8192);
        if (cb->len > 8192)
        {
            evm_trace(EVM_TRACE_DETAIL, "evmadmin:  max ultra buf is 8192, truncating test\n");
            cb->len = 8192;
        }
        buf = eee_allocateBuffer(EEE_SHARED_POOL, EEE_BUF_ULTRA, NFX_COS_HIGH);
        // Preserve the unused leader for packet headers.
        buf += EEE_BUFSIZE_ULTRA - 8192;
    }
    else
    {
        EVM_RAM_ALLOC_ALIGNED(buf, char, cb->len);
    }
    if (buf == NULL)
    {
        evm_trace(EVM_TRACE_DETAIL, "\nEVM TEST META IO[%d]: no memory for buffer\n", n);
        evm_deleteTestCb(cb);
        return NFX_ERR;
    }
    memset((void *) buf, fill, cb->len);
	{
		int i;
		//evm_trace(EVM_TRACE_DETAIL, "len = %d\n", cb->len);
		for (i = 0; i < cb->len; i++)
		{
			buf[i] = i & 0xff;
		} // for
	}
    cb->chksum = evm_doChksum((ushort16 *) buf, cb->len);
    cb->buf = buf;
    if (type == m_write)
    {
        cb->req.wDataHandle.dataPtr = buf;
    }
    else
    {
        /*
         * user write, allocate COD and fill with pattern
         */
        rc = evm_sendCreateCodMsg(&cb->req.wDataHandle.codHandle,
                                  cb->key, size, fill);
        if (rc != NFX_OK)
        {
            /* callback evm_writeComplete deletes cb if err */
            evm_trace(EVM_TRACE_DETAIL, "\nEVM TEST USER IO[%d]: cannot create cod\n", n);
            evm_deleteTestCb(cb);
            return NFX_ERR;
        }
    }
    rc = evm_IOReq(&cb->req, evm_writeComplete);
    if (rc != NFX_OK)
    {
        /* callback evm_writeComplete deletes cb if err */
        evm_trace(EVM_TRACE_DETAIL, "\nEVM TEST IO[%d]: evmIOReq/WRITE FAILED\n", n);
        evm_deleteTestCb(cb);
        return NFX_ERR;
    }
    if (type == m_write)
    {
        evm_trace(EVM_TRACE_DETAIL, "EVM TEST META IO[%d]: evmIOReq/WRITE OK\n", n);
        evm_trace(EVM_TRACE_DETAIL, "                      offset %lld size %lld (chksum %0x len %d)\n",
               offset, size, cb->chksum, cb->len);
    }
    else
    {
        evm_trace(EVM_TRACE_DETAIL, "EVM TEST USER IO[%d]: evmIOReq/WRITE OK\n", n);
        evm_trace(EVM_TRACE_DETAIL, "                      cod %llx offset %lld size %lld (chksum %0x len %d)\n",
               cb->req.wDataHandle.codHandle,
               offset, size, cb->chksum, cb->len);
    }
    return NFX_OK;

} /* evm_doWrite */


/*-----------------------------------------------------------------
 * Name:        evm_test
 *
 * Description: called to test EVM srvr IO
 *              see sm-ui/cmdtable.c
 *
 *              available only via FP simulator
 *
 * Created by : Lan Chin
 *
 * Date Created: 9/28/01
 *-----------------------------------------------------------------
 */

int32
evm_test(int32 argc, char8 *argv[])
{
    int32           rc;
    int32           i, nTests;
    int32           lvIndex;
    evm_pReq_t      type;
    evm_volOffset_t offset=0;
    evm_volSize_t   size;
    evm_vol_t *lv;
    boolean         doUltra = FALSE;

    if (evm_srvr_test_init == 0)
    {
        evm_srvr_test_init = 1;
        evm_srvr_test_hash = hash_create(EVM_SRVR_TEST_HASH_BUCKETS,
                                         evm_srvr_test_hashFunc,
                                         evm_srvr_test_hashCompare,
                                         evm_srvr_test_hashRelease);
    }
    if (argc < 5)
    {
        evm_trace(EVM_TRACE_DETAIL, "\nUSAGE: evmadmin test {user | meta | uuser | umeta} "
               "<vol-index> <nblks> <ntimes> [offset]\n");
        return NFX_ERR;
    }
    if (strcmp(argv[1], "user") == 0)
    {
        type = u_write;
        doUltra = FALSE;
    }
    else if (strcmp(argv[1], "meta") == 0)
    {
        type = m_write;
        doUltra = FALSE;
    }
    else if (strcmp(argv[1], "uuser") == 0)
    {
        type = u_write;
        doUltra = TRUE;
    }
    else if (strcmp(argv[1], "umeta") == 0)
    {
        type = m_write;
        doUltra = TRUE;
    }
    NFX_SPIN_LOCK(&evm_lv_lock);
    //lv = evm_lvs_tbl;
    lvIndex = atoi(argv[2]);
    DLL_ITERATE(lv, &evm_volList, entry)
        {
            if (lvIndex == 0)
                break;
            lvIndex--;
        }
    if (lvIndex != 0)
    {
        evm_trace(EVM_TRACE_DETAIL,
                  "\nNo logical volume found for index %d\n", lvIndex);
        NFX_SPIN_UNLOCK(&evm_lv_lock);
        return NFX_ERR;
    }
    NFX_SPIN_UNLOCK(&evm_lv_lock);
    size = atol(argv[3]);
    if (size % 2)
    {
        evm_trace(EVM_TRACE_DETAIL,
                  "\nERROR: scsi checksum reqires even number block io\n");
        return NFX_ERR;
    }
    nTests = atoi(argv[4]);
    if (argc == 6)
    {
        offset = atol(argv[5]);
    }
    evm_trace(EVM_TRACE_DETAIL, "evmadmin test: %llx %lld %d %lld\n",
            lv->id.whole_thing, size, nTests, offset);
    for (i=0; i<nTests; i++)
    {
        rc = evm_doWrite(i, type, lv, offset, size, doUltra);
        if (rc != NFX_OK)
        {
            evm_trace(EVM_TRACE_DETAIL, "\nERROR[%d]: cannot do META WRITE for offset %llx size %lld\n",
                    i, offset, size);
        }
        offset += size;
    }
    evm_trace(EVM_TRACE_DETAIL, "\nEVM TEST META IO: %d tests started\n", i);
    return NFX_OK;

} /* evm_test */

#else // ISABELLE
int32
evm_test(int32 argc, char8 *argv[])
{
    return(NFX_OK);
}
#endif // ISABELLE
/*----------------------------------------------------------------------------*/
