/*
 *-----------------------------------------------------------------
 *
 * Name:
 *
 * RCS:  $Id: crashdump-api.c,v 1.26 2002/11/26 22:29:17 rick Exp $
 *
 * Copyright (C) 2002, Agile Storage, Inc.
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/7/02
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <sys/mips.h>
#include <sys/amd29f040.h>
#include <exports/crashdump-api.h>
#include <exports/timer-api.h>
#include "crashdump.h"
#include <exports/eee-api.h>
#include <exports/eee.h>
#include <exports/pte-api.h>
#include <exports/malloc-api.h>
#include <exports/libc.h>

#include <sys/exec_elf.h>

static uint8 * exc_format_buf;

crashdump_t		temp_cd CPU_PRIVATE;

#if defined(NFP_FP)

#include <exports/esm-api.h>

extern char _end[];
extern char _start[];

extern void *malloc_base;

extern struct core_file *core_open(void);
extern int core_file_seek(struct core_file *file, uint64 pos);
extern int core_file_write(struct core_file *file, uchar8 *data, uint64 length);
extern int core_file_close(struct core_file *file);

extern nfx_spinlock_t	prom_lock;

#if !defined(LARGE_MEMORY)
#define MALLOC_SIZE (64*1024*1024)
#endif

enum core_segments
{
    CORESEG_TEXT_DATA = 1,
    CORESEG_MALLOC,
    CORESEG_SHMEM,
    CORESEG_TXRX,
    CORESEG_LAST
};

#if defined(LARGE_MEMORY)
#if defined(USE_BSS_THREAD_STACKS)
#define MAX_CORE_SEGS (CORESEG_LAST)
#else
#define MAX_CORE_SEGS (CORESEG_LAST + ESM_NUM_STACKS)
#endif
#else
#define MAX_CORE_SEGS (CORESEG_LAST)
#endif

#define ELF_PAGESIZE 4096


static char notebuf[4096];
static char corebuf[MAX_CORE_SEGS * sizeof(Elf64_Phdr) + ELF_PAGESIZE * 2];
static Elf64_Phdr pheaders[MAX_CORE_SEGS];

static int writing_core;

#if defined(LARGE_MEMORY)
/*-----------------------------------------------------------------
 * Name :	core_seg_start()
 *
 * Description: Return the virtual address of the start of the core
 * segment
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/27/02
 *
 *----------------------------------------------------------------- */
static uint64
core_seg_start(int seg)
{
    switch (seg) {
    case CORESEG_TEXT_DATA:
        return (uint64)&_start;
    case CORESEG_MALLOC:
        return get_memory_virt_base(MMU_MALLOC_REGION);
    case CORESEG_SHMEM:
        return eee.shmem_base;
    case CORESEG_TXRX:
        return get_memory_virt_base(MMU_SM_TXRX_REGION);
#if defined(USE_BSS_THREAD_STACKS)
	default:
		return 0;
#else
    default:
		/* Thread stacks */
		return get_memory_virt_base(MMU_THREADSTACK_REGION) +
			(uint64)(seg - CORESEG_LAST) * (uint64)PTE_ENTRY_SIZE;
#endif
    }
}

/*-----------------------------------------------------------------
 * Name :	core_seg_size()
 *
 * Description: Return the size of the core segment
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/27/02
 *
 *-----------------------------------------------------------------
 */
static uint64
core_seg_size(int seg)
{
    switch (seg) {
    case CORESEG_TEXT_DATA:
        return (((u_int64_t)&_end - (u_int64_t)&_start + ELF_PAGESIZE - 1)
                & ~(ELF_PAGESIZE - 1));
    case CORESEG_MALLOC:
        return get_memory_region_size(MMU_MALLOC_REGION);
    case CORESEG_SHMEM:
        return eee.sm_size;
    case CORESEG_TXRX:
        return get_memory_region_size(MMU_SM_TXRX_REGION);
#if defined(USE_BSS_THREAD_STACKS)
	default:
		return 0;
#else		
    default:
		/* Thread stacks */
		return ESM_STACK_SIZE;
#endif		
    }
}
#else

/*-----------------------------------------------------------------
 * Name :	core_seg_start()
 *
 * Description: Return the virtual address of the start of the core
 * segment
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/27/02
 *
 *----------------------------------------------------------------- */
static uint64
core_seg_start(enum core_segments seg)
{
    switch (seg) {
    case CORESEG_TEXT_DATA:
        return (uint64)&_start;
    case CORESEG_MALLOC:
        return malloc_base;
    case CORESEG_SHMEM:
        return eee.shmem_base;
    case CORESEG_TXRX:
        return 0x20000000;
    default:
        panic("core_seg_start: %d\n", seg);
    }
}

/*-----------------------------------------------------------------
 * Name :	core_seg_size()
 *
 * Description: Return the size of the core segment
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/27/02
 *
 *-----------------------------------------------------------------
 */
static uint64
core_seg_size(enum core_segments seg)
{
    switch (seg) {
    case CORESEG_TEXT_DATA:
        return (((u_int64_t)&_end - (u_int64_t)&_start + ELF_PAGESIZE - 1)
                & ~(ELF_PAGESIZE - 1));
    case CORESEG_MALLOC:
        return MALLOC_SIZE;
    case CORESEG_SHMEM:
        return eee.sm_size;
    case CORESEG_TXRX:
        return 0x4000000;
    default:
        panic("core_seg_size: %d\n", seg);
    }
}

#endif


/*-----------------------------------------------------------------
 * Name :	core_size()
 *
 * Description: Calculate the size of the core dump
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/27/02
 *
 *-----------------------------------------------------------------
 */
uint64
core_size(void)
{
    int seg;
	/* ELF Header + registers */
    uint64 size = sizeof(Elf64_Phdr) * MAX_CORE_SEGS + 2 * ELF_PAGESIZE;
    for (seg = CORESEG_TEXT_DATA; seg < MAX_CORE_SEGS; ++seg) {
        size += core_seg_size(seg);
    }
    return size;
}


/*-----------------------------------------------------------------
 * Name :	fill_seg()
 *
 * Description: Initialize elf program header for the segment number
 * segno.
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	06/27/02
 *
 *-----------------------------------------------------------------
 */
static void
fill_seg(int segno, Elf64_Phdr *hdrs)
{
    Elf64_Phdr *seg = &hdrs[segno];
    seg->p_type = PT_LOAD;
    seg->p_flags = PF_R | PF_W;
    seg->p_vaddr = core_seg_start(segno);
    seg->p_paddr = core_seg_start(segno);
    seg->p_filesz = core_seg_size(segno);
    seg->p_memsz = core_seg_size(segno);
    seg->p_align = 4096;
}


/*-----------------------------------------------------------------
 * Name :    write_core()
 *
 * Description: Write the core dump. The core writes should multiple
 * of 2*1024.
 *
 * Created by:  Maxim Kozlovsky
 * 
 * Date Created:  06/07/02
 * 
 *-----------------------------------------------------------------
 */
void
write_core()
{
    int segs = 0;
    Elf64_Ehdr elf;
    Elf64_Phdr *registers;
    u_int64_t offset;
    struct core_file *core_file;
    int i;
    int error;
    char *notebufptr = notebuf;
    Elf64_Note *prstatus_note;
    int gpregsz;

	if (writing_core)
		return;

	writing_core = 1;

    watchdog_disable();
    
    /* Registers */
    prstatus_note = (Elf64_Note *)notebufptr;
    prstatus_note->namesz = (sizeof("CORE") + 1 + 3) & ~3;
    prstatus_note->descsz = 0;
    prstatus_note->type = NT_PRSTATUS;
    notebufptr += sizeof(*prstatus_note);
    strcpy(notebufptr, "CORE");
    notebufptr += prstatus_note->namesz;
    gpregsz = 32 * sizeof(cpureg_t);
    memcpy(notebufptr, new_cd->regs, gpregsz);
    prstatus_note->descsz += gpregsz;
    notebufptr += gpregsz;

    memcpy(notebufptr, &new_cd->regs[R_STATUS], sizeof(uint64));
    prstatus_note->descsz += sizeof(uint64);
    notebufptr += sizeof(uint64);

    memcpy(notebufptr, &new_cd->regs[R_LO], sizeof(uint64));
    prstatus_note->descsz += sizeof(uint64);
    notebufptr += sizeof(uint64);

    memcpy(notebufptr, &new_cd->regs[R_HI], sizeof(uint64));
    prstatus_note->descsz += sizeof(uint64);
    notebufptr += sizeof(uint64);

    memcpy(notebufptr, &new_cd->regs[R_BADVA], sizeof(uint64));
    prstatus_note->descsz += sizeof(uint64);
    notebufptr += sizeof(uint64);

    memcpy(notebufptr, &new_cd->regs[R_CAUSE], sizeof(uint64));
    prstatus_note->descsz += sizeof(uint64);
    notebufptr += sizeof(uint64);

    memcpy(notebufptr, &new_cd->regs[R_EPC], sizeof(uint64));
    prstatus_note->descsz += sizeof(uint64);
    notebufptr += sizeof(uint64);

    registers = &pheaders[segs];
    ++segs;
    registers->p_type = PT_NOTE;
    registers->p_flags = PF_R | PF_X | PF_W;
    registers->p_vaddr = (u_int64_t)0;
    registers->p_paddr = (u_int64_t)0;
    registers->p_memsz = 0;
    registers->p_align = 0;
    registers->p_filesz = sizeof(*prstatus_note) + prstatus_note->namesz +
        prstatus_note->descsz;

    for (i = segs; i < MAX_CORE_SEGS; ++i) {
        fill_seg(i, pheaders);
    }
    pheaders[CORESEG_TEXT_DATA].p_flags |= PF_X;
    segs = i;

    elf.e_ident[EI_MAG0] = ELFMAG0;
    elf.e_ident[EI_MAG1] = ELFMAG1;
    elf.e_ident[EI_MAG2] = ELFMAG2;
    elf.e_ident[EI_MAG3] = ELFMAG3;
    elf.e_ident[EI_CLASS] = ELFCLASS64;
    elf.e_ident[EI_DATA] = ELFDATA2LSB;
    elf.e_ident[EI_VERSION] = EV_CURRENT;
    memset(elf.e_ident + EI_PAD, 0, EI_NIDENT - EI_PAD);
    
    elf.e_type = ET_CORE;
    elf.e_machine = EM_NONE;
    elf.e_version = EV_CURRENT;
    elf.e_phoff = sizeof(elf);
    elf.e_shoff = 0;
    elf.e_flags = 0;
    elf.e_ehsize = sizeof(elf);
    elf.e_phentsize = sizeof(Elf64_Phdr);
    elf.e_phnum = segs;
    elf.e_shentsize = 0;
    elf.e_shnum = 0;
    elf.e_shstrndx = 0;

    printf("Writing core file\n");

    core_file = core_open();
    if (core_file != NULL) {
        char *bufptr = corebuf;
        memset(corebuf, 0, sizeof(corebuf));
        /* Write the header */

        memcpy(bufptr, &elf, sizeof(elf));
        bufptr += sizeof(elf);
        
        offset = sizeof(elf) + sizeof(Elf64_Phdr) * segs;

        /* Calculate the section offsets and write the program section
           headers and notes section */
        for (i = 0; i < segs; ++i) {
            pheaders[i].p_offset = offset;
            memcpy(bufptr, &pheaders[i], sizeof(Elf64_Phdr));
            bufptr += sizeof(Elf64_Phdr);
            offset += pheaders[i].p_filesz;
			offset = (offset + ELF_PAGESIZE - 1) & ~(ELF_PAGESIZE - 1);
        }

        memcpy(bufptr, notebuf, pheaders[0].p_filesz);

        core_file_seek(core_file, 0);
		printf("writing seg %d off %llx sz %llx\n", 0, 0LL,
			   pheaders[1].p_offset);
        core_file_write(core_file, corebuf, pheaders[1].p_offset);

        /* Write the rest of the sections */
        error = 0;
        for (i = 1; !error && i < segs; ++i) {
            u_int64_t seek_pos = pheaders[i].p_offset;
            if (!(error = core_file_seek(core_file, seek_pos))) {
                printf("writing seg %d off %llx sz %llx va %llx\n", i,
                       seek_pos, pheaders[i].p_filesz,
					   pheaders[i].p_vaddr);
                error = core_file_write(core_file,
										(uchar8 *)pheaders[i].p_vaddr,
                                        pheaders[i].p_filesz);
                if (error) {
                    printf("Failed to write data %d\n", i);
                }
            }
            else {
                printf("seek %lld failed\n", seek_pos);
            }
        }

        if (!error) {
            error = core_file_close(core_file);
        }

        if (error == 0) {
            printf("Core written successfully\n");
        }
        else {
            printf("Core write failed\n");
        }
    }
    else {
        printf("Can not open core file\n");
    }
}   



void
crashdump_test_core()
{
    write_core();
}
#else
void
crashdump_test_core(void)
{
}

void
write_core(void)
{
}
#endif



/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/8/02
 *
 *-----------------------------------------------------------------
 */
void
exc_crashdump_set_stacktop(
	address_t	stacktop
)
{
	crashdump_t	* cptr = new_cd;

	cptr->stacktop = stacktop;
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description: returns non-zero, if a crashdump is present in the prom
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/7/02
 *
 *-----------------------------------------------------------------
 */
int32
exc_crashdump_is_present()
{
	crashdump_get(&saved_cd);

	return (crashdump_valid(&saved_cd));
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description: save a crashdump from a CPU exception
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/7/02
 *
 *-----------------------------------------------------------------
 */
void
exc_crashdump_except_save(
	uchar8	* str
)
{
	crashdump_t	* cptr = new_cd;

	if (str) {
		strncpy(cptr->crash_string, str, CRASHDUMP_STRING_LEN);
		cptr->flags |= CRASHDUMP_FLAGS_EXCEPT_STRING;
	}

	if (crashdump_getTimeStamp(cptr->crash_time))
		cptr->flags |= CRASHDUMP_FLAGS_CRASH_TIME;

	crashdump_save_regs(cptr);
	crashdump_save_stack(cptr);

	cptr->flags &= ~CRASHDUMP_FLAGS_CAUSE_MASK;
	cptr->flags |= CRASHDUMP_FLAGS_CAUSE_EXCEPT;

	sprintf(cptr->crash_time, "%s", eee_rtcStr(NULL));
	cptr->flags |= CRASHDUMP_FLAGS_CRASH_TIME;

	sprintf(cptr->start_time, "%s", eee_humandate(bootGMTTime.sec));
	cptr->flags |= CRASHDUMP_FLAGS_START_TIME;

	cptr->checksum = crashdump_checksum(cptr);

	NFX_SPIN_LOCK(&prom_lock);
	if (!exc_crashdump_is_present())
		crashdump_save(cptr);
	NFX_SPIN_UNLOCK(&prom_lock);
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description: save a crashdump from a software panic
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/7/02
 *
 *-----------------------------------------------------------------
 */
void
exc_crashdump_panic_save(
	uchar8	* str
)
{
	crashdump_t	* cptr = new_cd;

	if (str) {
		strncpy(cptr->crash_string, str, CRASHDUMP_STRING_LEN);
		cptr->flags |= CRASHDUMP_FLAGS_PANIC_STRING;
	}

	if (crashdump_getTimeStamp(cptr->crash_time))
		cptr->flags |= CRASHDUMP_FLAGS_CRASH_TIME;

	crashdump_save_regs(cptr);
	crashdump_save_stack(cptr);

	cptr->flags &= ~CRASHDUMP_FLAGS_CAUSE_MASK;
	cptr->flags |= CRASHDUMP_FLAGS_CAUSE_PANIC;

	sprintf(cptr->crash_time, "%s", eee_rtcStr(NULL));
	cptr->flags |= CRASHDUMP_FLAGS_CRASH_TIME;

	sprintf(cptr->start_time, "%s", eee_humandate(bootGMTTime.sec));
	cptr->flags |= CRASHDUMP_FLAGS_START_TIME;

	cptr->checksum = crashdump_checksum(cptr);

	NFX_SPIN_LOCK(&prom_lock);
	if (!exc_crashdump_is_present())
		crashdump_save(cptr);
	NFX_SPIN_UNLOCK(&prom_lock);

}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description: save a crashdump from a watchdog timeout - NMI
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/7/02
 *
 *-----------------------------------------------------------------
 */
void
exc_crashdump_nmi_save()
{
	crashdump_t	* cptr = new_cd;

	strncpy(cptr->crash_string, "Watchdog Timeout NMI", CRASHDUMP_STRING_LEN);
	cptr->flags |= CRASHDUMP_FLAGS_NMI_STRING;

	if (crashdump_getTimeStamp(cptr->crash_time))
		cptr->flags |= CRASHDUMP_FLAGS_CRASH_TIME;

	crashdump_save_regs(cptr);
	crashdump_save_stack(cptr);

	cptr->flags &= ~CRASHDUMP_FLAGS_CAUSE_MASK;
	cptr->flags |= CRASHDUMP_FLAGS_CAUSE_NMI;

	sprintf(cptr->crash_time, "%s", eee_rtcStr(NULL));
	cptr->flags |= CRASHDUMP_FLAGS_CRASH_TIME;

	sprintf(cptr->start_time, "%s", eee_humandate(bootGMTTime.sec));
	cptr->flags |= CRASHDUMP_FLAGS_START_TIME;

	cptr->checksum = crashdump_checksum(cptr);

	NFX_SPIN_LOCK(&prom_lock);
	if (!exc_crashdump_is_present())
		crashdump_save(cptr);
	NFX_SPIN_UNLOCK(&prom_lock);
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/7/02
 *
 *-----------------------------------------------------------------
 */
void
exc_crashdump_remove()
{
	NFX_SPIN_LOCK(&prom_lock);

    memset(&saved_cd, 0xff, sizeof(crashdump_t));
    from_programbytes(FROM_CRASHDUMP_OFFSET, (void *)&saved_cd, sizeof(crashdump_t));
    memset(&saved_cd, 0, sizeof(crashdump_t));

	NFX_SPIN_UNLOCK(&prom_lock);
}

static void
print_format_buf(
    uint8 * bp,
    uint32  len 
        )
{
#define BUFSIZE 127
	char print_buf[BUFSIZE + 1];
	int off = 0;
	while (off < len) {
		int print_len = MIN(len - off, BUFSIZE);
		memcpy(print_buf, bp + off, print_len);
		print_buf[print_len] = 0;
		printf(print_buf);
		off += print_len;
	}
}

/*-----------------------------------------------------------------
 * Name: format_crashdump
 *
 * Description: format the crashdump
 *
 * Created by:  John Hahn
 *
 * Date Created:	6/4/02
 *
 *-----------------------------------------------------------------
 */
static uint32
format_crashdump(
	crashdump_t	* cptr,
    uint8       * format_buf,
    boolean     iflag
)
{
    uint32 rv;
    uchar8 * bp;

    bp = format_buf;
    bp[0] = '\0';       // start it off as a null string
    rv = 0;

	if (cptr->signature == CRASHDUMP_SIGNATURE) {
		if (crashdump_valid(cptr)) {
			if (cptr->format == 1) {
				bp += snprintf(bp, exc_fblen(bp), "\nCrashdump:\n");
				bp += snprintf(bp, exc_fblen(bp), "----------\n");

				if (cptr->flags & CRASHDUMP_FLAGS_CAUSE_EXCEPT)
					bp += snprintf(bp, exc_fblen(bp), "Exception : %s\n", cptr->crash_string);
				else if (cptr->flags & CRASHDUMP_FLAGS_CAUSE_PANIC)
					bp += snprintf(bp, exc_fblen(bp), "Panic : %s\n", cptr->crash_string);
				else if (cptr->flags & CRASHDUMP_FLAGS_CAUSE_NMI)
					bp += snprintf(bp, exc_fblen(bp), "NMI : %s\n", cptr->crash_string);

				if (cptr->flags & CRASHDUMP_FLAGS_RT_VER)
					bp += snprintf(bp, exc_fblen(bp), "Image Version : %s\n", cptr->rt_version);
				else
					bp += snprintf(bp, exc_fblen(bp), "Image Version : Unavailable\n");

				if (cptr->flags & CRASHDUMP_FLAGS_PROM_VER)
					bp += snprintf(bp, exc_fblen(bp), "PROM Version  : %s\n", cptr->prom_version);
				else
					bp += snprintf(bp, exc_fblen(bp), "PROM Version  : Unavailable\n");

				if (cptr->flags & CRASHDUMP_FLAGS_START_TIME)
					bp += snprintf(bp, exc_fblen(bp), "Boot Time  : %s\n", cptr->start_time);
				else
					bp += snprintf(bp, exc_fblen(bp), "Boot Time  : Unavailable\n");

				if (cptr->flags & CRASHDUMP_FLAGS_CRASH_TIME)
					bp += snprintf(bp, exc_fblen(bp), "Crash Time : %s\n", cptr->crash_time);
				else
					bp += snprintf(bp, exc_fblen(bp), "Crash Time : Unavailable\n");

				if (cptr->flags & CRASHDUMP_FLAGS_REGS) {
					bp += snprintf(bp, exc_fblen(bp), "Regs  :\n");
					bp += crashdump_displayRegs(cptr, bp);
				}
				else
					bp += snprintf(bp, exc_fblen(bp), "Regs  : Unavailable\n");


				if (cptr->flags & CRASHDUMP_FLAGS_STACK) {
					bp += crashdump_backtrace(cptr, bp);
					bp += snprintf(bp, exc_fblen(bp), "Stack (%d words) :\n",
                            cptr->stacksize);

                    if (iflag) {
                        rv = (uint32) (bp - format_buf);
                        print_format_buf(format_buf, rv);
                        bp = format_buf;
                        bp[0] = '\0';       // start it off as a null string
                    }
					bp += crashdump_displayRawStack(cptr, bp, iflag);
				}
				else
					bp += snprintf(bp, exc_fblen(bp), "Stack : Unavailable\n");

				bp += snprintf(bp, exc_fblen(bp), "----------\n\n");
			}
			else
				bp += snprintf(bp, exc_fblen(bp), "Invalid Crashdump format, %d\n",
                        cptr->format);
		}
		else
			bp += snprintf(bp, exc_fblen(bp), "Invalid Crashdump checksum\n");
	}
	else
		bp += snprintf(bp, exc_fblen(bp), "Invalid Crashdump Signature\n");

    if (iflag && (bp != format_buf)) {
        print_format_buf(format_buf, (uint32)(bp - format_buf));
    }

    rv += (uint32) (bp - format_buf);

    return(rv);
}

/*-----------------------------------------------------------------
 * Name:
 *
 * Description: view a crashdump and stacktrace
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/7/02
 *
 *-----------------------------------------------------------------
 */
int32
exc_crashdump_view(
	uint32	which
)
{
	crashdump_t	*   cptr;
    uint32          format_len;
    int32           rv;

	if (which == EXC_CD_SAVED) {
		NFX_SPIN_LOCK(&prom_lock);
		crashdump_get(&temp_cd);
		NFX_SPIN_UNLOCK(&prom_lock);
		cptr = &temp_cd;
	}
	else if (which == EXC_CD_PENDING) {
		cptr = new_cd;
	}
    else
        return  NFX_ERR;

    exc_format_buf = (uchar8 *) eee_ramAlloc(EXE_CD_FORMAT_BUF_ALLOC_SIZE);
    if (exc_format_buf == NULL) {
        printf("crashdump: cannot allocate memory\n");
        return(NFX_ERR);
    }
    bzero(exc_format_buf, EXE_CD_FORMAT_BUF_ALLOC_SIZE);

    format_len = format_crashdump(cptr, exc_format_buf, TRUE /*interactive?*/);
    if (format_len > EXE_CD_FORMAT_BUF_SIZE) {
        printf("crashdump: out of format buffer space truncated output\n");
        printf("           need=%d, have=%d\n", 
                format_len, EXE_CD_FORMAT_BUF_SIZE);
        rv = NFX_ERR;
    }
    else {
        rv = NFX_OK;
    }

/*
    printf("exc_crashdump_store: format buf@%p len=%d remain=%d\n",
            exc_format_buf, format_len, exc_fblen(&exc_format_buf[format_len]));
*/

    eee_ramDealloc(exc_format_buf);
    return(rv);
}


#if !defined(SSC_MGMT)
/*-----------------------------------------------------------------
 * Name :	crashdump_save_rcv()
 *
 * Description: receive function for crashsave application
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	09/02/02
 *
 *-----------------------------------------------------------------
 */
static void
crashdump_save_rcv(eee_descPtr_t desc)
{
	eee_unregisterReceiveFunc(eee_getApplicationId("crashsave"));
	if (*((uint32 *)edesc_firstBuf(desc)->buf)) {
		exc_crashdump_remove();
	}
	eee_freePacket(desc);
}
#endif


/*-----------------------------------------------------------------
 * Name: exc_crashdump_store
 *
 * Description: store off the crashdump to SSC fs 
 *
 * Created by:    John Hahn
 *
 * Date Created:    6/4/02
 *
 *-----------------------------------------------------------------
 */
int32
exc_crashdump_store(uint32 which)
{
#if !defined(SSC_MGMT)
	crashdump_t	*   cptr;
    uint32          format_len;
    int32           rv;

	if (which == EXC_CD_SAVED) {
		NFX_SPIN_LOCK(&prom_lock);
		crashdump_get(&temp_cd);
		NFX_SPIN_UNLOCK(&prom_lock);
		cptr = &temp_cd;
	}
	else if (which == EXC_CD_PENDING) {
		cptr = new_cd;
	}
    else
        return NFX_ERR;

	if (crashdump_valid(cptr)) {
		uint32 appid = eee_getApplicationId("crashsave");
		uint32 dp, sp;
		exc_format_buf = (uchar8 *) eee_ramAlloc(EXE_CD_FORMAT_BUF_ALLOC_SIZE);
			
		if (exc_format_buf == NULL) {
			printf("crashdump: cannot allocate memory\n");
			return(NFX_ERR);
		}
		bzero(exc_format_buf, EXE_CD_FORMAT_BUF_ALLOC_SIZE);

		format_len = format_crashdump(cptr, exc_format_buf, FALSE /*interactive?*/);
		if (format_len > EXE_CD_FORMAT_BUF_SIZE) {
			printf("crashdump: out of format buffer space truncated output\n");
			printf("           need=%d, have=%d\n", 
				   format_len, EXE_CD_FORMAT_BUF_SIZE);
			format_len = EXE_CD_FORMAT_BUF_SIZE;
		}

		eee_registerReceiveFunc(appid, crashdump_save_rcv,
								crashdump_save_rcv);
		dp = PACK_UNICAST_PORT_ADDRESS(0, 0, appid,
									   EEE_MGMT, NFX_COS_HIGH, 0);
		sp = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId, nfxMyCPUId, appid,
									   EEE_MGMT, NFX_COS_HIGH, 0);
		eee_sendMessage(exc_format_buf, format_len, dp, sp);
		eee_ramDealloc(exc_format_buf);
		rv = NFX_OK;

/*
  printf("exc_crashdump_store: format buf@%p len=%d remain=%d\n",
  exc_format_buf, format_len, exc_fblen(&exc_format_buf[format_len]));
*/
	}
	else {
		rv = NFX_ERR;
	}
    return(rv);
#else
	return NFX_ERR;
#endif
}


/*-----------------------------------------------------------------
 * Name: exc_fblen
 *
 * Description: space left in crashdump format storage buffer
 *
 * Created by:    John Hahn
 *
 * Date Created:    6/4/02
 *
 *-----------------------------------------------------------------
 */
uint32
exc_fblen(uint8 * bp)
{
    uint8 * end_p;
    uint32 rv;


    end_p = exc_format_buf + EXE_CD_FORMAT_BUF_SIZE;
    if ((address_t)bp >= (address_t)end_p)
        rv = 0;
    else
        rv = (uint32)((address_t)end_p -(address_t)bp);

    return(rv);
}
