/*
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * $Id: emapleaf.c,v 1.78 2002/12/04 20:01:09 bob Exp $
 */

#if !defined(SSC)

#include "btree.h"
#include "emap.h"

static void fs_emapleaf_removeent(fs_Context *context, fs_Buf *buf, int i);

static void
emapleaf_check(fs_Inode *inode, fs_Buf *buf)
{
#ifdef FS_DEBUG_EMAP_LEAF_CHECK
	uint32 index[FS_EMAP_EXTENTS_MAX];
	uint32 magic = 0xCAFEBABE;
	int max = EMAP_IS_COMPACT(inode) ? FS_EMAP_COMPACT_EXTENTS_MAX : FS_EMAP_EXTENTS_MAX;
	int i;

	/*
	 * Verify that the index maps every entry somewhere.
	 */
	for (i = 0; i < max; i++)
		index[i] = magic;

	for (i = 0; i < max; i++) {
		assert(get_index(inode, buf, i) < max);
		index[get_index(inode, buf, i)] = i;
	}

	for (i = 0; i < max; i++)
		assert(index[i] != magic);

	/*
	 * Verify that each extent does not overlap with its
	 * neighbors.
	 */
	for (i = 0; i < get_nkeys(buf) - 1; i++) {
		assert(get_offset(inode, buf, i) + fs_extentLength(get_extent(inode, buf, i))
				<= get_offset(inode, buf, i+1));
	}
#endif 
}

static boolean
query_dispose(fs_Context *context, fs_Buf *buf, fs_Bt_eeksearch *r, char *fmt, ...)
{
	va_list ap;
	char msg[256];

	va_start(ap, fmt);
	vsprintf(msg, fmt, ap);
	va_end(ap);

    fs_lockExclusive(context, context->eek.lock, NULL);
    context->eek.errcount++;
    fs_unlockExclusive(context, context->eek.lock);

    if (r->fixIt) {
        if (fs_eek_query(context, "(I#%lld) emapleaf %lld: %s.\nDispose?",
                            buf->inum, fs_blockNumber(buf->block), msg)) {
            r->status = FS_BT_EEK_DISPOSE;
            return TRUE;
        }
    } else {
	    fs_eek_print_status(context, FALSE,
            "%s: (I#%lld) emapleaf %lld: %s\n", 
            context->fileSystem->super->mountName,
		    buf->inum, fs_blockNumber(buf->block), msg);
    }
    fs_bufReleaseExclusive(context, buf);
    return FALSE;
}

/*
 * Return TRUE if an emapleaf is okay.
 * FALSE otherwise
 */
static boolean
fs_emapleaf_eek_check(fs_Context *context, fs_Inode *inode, fs_Buf *buf, fs_Bt_eeksearch *r)
{
	int max = EMAP_IS_COMPACT(inode) ? FS_EMAP_COMPACT_EXTENTS_MAX : FS_EMAP_EXTENTS_MAX;
	uint32 index[FS_EMAP_EXTENTS_MAX];
	uint32 magic = 0xCAFEBABE;
	int nkeys = get_nkeys(buf);
	int i;

	if (nkeys > max) {
		return query_dispose(context, buf, r, "nkeys %d > %d", nkeys, max);
	}

	/*
	 * Verify that the index maps every entry somewhere.
	 */
	for (i = 0; i < max; i++)
		index[i] = magic;

	for (i = 0; i < max; i++) {
		if (get_index(inode, buf, i) >= max) {
			return query_dispose(context, buf, r,
				            "index[%d]: %d >= %d",
				            i, get_index(inode, buf, i), max);
		}
		if (index[get_index(inode, buf, i)] != magic) {
			return query_dispose(context, buf, r,
				            "index[%d] and index[%d] conflict",
				            i, get_index(inode, buf, i), max);
		}
		index[get_index(inode, buf, i)] = i;
	}

	/*
	 * Verify that each extent does not overlap with its
	 * neighbors.
	 */
	for (i = 0; i < nkeys - 1; i++) {
		if (get_offset(inode, buf, i) + fs_extentLength(get_extent(inode, buf, i)) > get_offset(inode, buf, i+1)) {
			return query_dispose(context, buf, r,
				            "extents %d and %d overlap %lld:(%d,%lld), %lld:(%d, %lld)",
				            i, i+1,
				            get_offset(inode, buf, i),
				            fs_extentLength(get_extent(inode, buf, i)),
				            fs_blockNumber(fs_extentStartBlock(get_extent(inode, buf, i))),
				            get_offset(inode, buf, i+1),
				            fs_extentLength(get_extent(inode, buf, i+1)),
				            fs_blockNumber(fs_extentStartBlock(get_extent(inode, buf, i+1))));
        }
	}

	/*
	 * Verify that each extent has a valid block number
	 */
	for (i = 0; i < nkeys; i++) {
		fs_Extent e = get_extent(inode, buf, i);
		if (fs_extentLength(e) + fs_blockNumber(fs_extentStartBlock(e)) >
		    context->fileSystem->params.totalBlocks) {
			return query_dispose(context, buf, r,
				            "extent %d has invalid blocknum: (%d,%lld)",
				            i,
				            fs_extentLength(e),
				            fs_blockNumber(fs_extentStartBlock(e)));
        }
	}

	return TRUE;
}

/*
 * Scan a dirleaf for a given name, returns an index into the
 * leaf and whether or not the name was found. If the name
 * is not found, the index points to where an insertion would
 * occur.
 */
static boolean
fs_emapleaf_scan(fs_Inode *inode, fs_Buf *buf, int *index, uint64 offset)
{
	int nkeys = get_nkeys(buf);
	int bot, top;
	int i;

	bot = 0;
	top = nkeys;
	i = 0;

	FS_ASSERT((unsigned)nkeys <= FS_EMAP_EXTENTS_MAX);

	/*
	 * Binary search to a likely starting point.
	 */
	while (bot < top) {
		uint64 offset_i; // cache of the ith offset in the leaf

		i = (top + bot) / 2;
		offset_i = get_offset(inode, buf, i);
		if (offset == offset_i)
			break;
		if (offset < offset_i)
			top = i-1;
		else
			bot = i+1;
	}

	/*
	 * Move forward over small keys.
	 */
	while (i < nkeys && offset > get_offset(inode, buf, i))
		i++;

	/*
	 * Back up over larger keys.
	 */
	while (i > 0
               && offset < (get_offset(inode, buf, i - 1)
                             + fs_extentLength(get_extent(inode, buf, i - 1))))
        {
		--i;
        }

	*index = i;

	return i < nkeys && offset >= get_offset(inode, buf, i)
			    && offset < (get_offset(inode, buf, i)
                                         + fs_extentLength(get_extent(inode, buf, i)));
}

/*
 * Add an entry to a dirleaf, assumes that any B-tree splitting has already
 * been done and that there is room in the page.
 */
static void
fs_emapleaf_addent(fs_Context *context, fs_Inode *inode, fs_Buf *buf, int i, uint64 offset, fs_Extent extent)
{
	fs_Emapleaf *l = buf->u.emapleaf;
	fs_Extent *extentp;
	uint64 *offsetp;

	emapleaf_check(inode, buf);

	/*
	 * First check for merges between this extent and the extents
	 * to the left and right of the insertion point.
	 */

	/*
	 * Check for a merge to the right -- if this is successful, then
	 * check for a merge to the left: this double merge would cause
	 * the leaf to shrink by an element.
	 */
	if (i != l->nkeys) {
		offsetp = get_offset_ref(inode, buf, i);
		extentp = get_extent_ref(inode, buf, i);
		if (fs_areAdjacent(extent, *extentp, offset, *offsetp)) {
			// printf("merging on the right\n");
			*offsetp = offset;
			*extentp = fs_mergeExtents(extent, *extentp);
			TOUCH_ELEM(buf, extentp);
			TOUCH_ELEM(buf, offsetp);

			/*
			 * Now that the new extent has been merged with
			 * the extent on the rhs, check to see whether
			 * it might "fill a hole" and merge with an
			 * extent to the left as well. In that case,
			 * grow the extent and remove the extent to
			 * the left.
			 */

            /*
             * call fs_areAdjacent w/ the new merged extent
             * since we need to check for length overflow in a 3 way merge
             */

			if (i != 0 &&
                fs_areAdjacent(get_extent(inode, buf, i-1), *extentp,
						       get_offset(inode, buf, i-1), *offsetp)) {
				*offsetp = get_offset(inode, buf, i-1);
				*extentp = fs_mergeExtents(get_extent(inode, buf, i-1), *extentp);
				TOUCH_ELEM(buf, extentp);
				TOUCH_ELEM(buf, offsetp);
				fs_emapleaf_removeent(context, buf, i-1);
			}
			return;
		}
	}

	/*
	 * A separate check for a merge to the left is performed when
	 * the first check for a merge to the right fails.
	 */
	if (i != 0) {
		fs_Extent extent_left = get_extent(inode, buf, i-1);
		uint64 offset_left = get_offset(inode, buf, i-1);

		if (fs_areAdjacent(extent_left, extent, offset_left, offset)) {
			// printf("merging on the left\n");
			extentp = get_extent_ref(inode, buf, i-1);
			*extentp = fs_mergeExtents(*extentp, extent);
			TOUCH_ELEM(buf, extentp);
			return;
		}
	}

	/*
	 * If we're inserting, move any entries to the right over by one.
	 *
	 * XXX: HACK: l->index array is at the same offset for compact and regular leaves.
	 */
	if (i != l->nkeys) {
		int new = l->index[l->nkeys];
		memmove(&l->index[i+1], &l->index[i], (l->nkeys - i) * sizeof *l->index);
		l->index[i] = new;
	}

	l->nkeys++;
	assert(l->nkeys <= (EMAP_IS_COMPACT(inode)?FS_EMAP_COMPACT_EXTENTS_MAX:FS_EMAP_EXTENTS_MAX));

	extentp = get_extent_ref(inode, buf, i);
	offsetp = get_offset_ref(inode, buf, i);

	*extentp = extent;
	*offsetp = offset;

	TOUCH(buf, &l->index[i], (l->nkeys - i) * sizeof *l->index);
	TOUCH_ELEM(buf, &l->nkeys);
	TOUCH_ELEM(buf, extentp);
	TOUCH_ELEM(buf, offsetp);
}

/*
 * Remove the i'th (in the index heap) directory entry from a page.
 */
static void
fs_emapleaf_removeent(fs_Context *context, fs_Buf *buf, int i)
{
	fs_Emapleaf *l = buf->u.emapleaf;
	int old = l->index[i];
	
	/*
	 * If we're not at the end of the array, move over any elements to the
	 * right left by one.
	 */
	assert(l->nkeys > 0);
	if (i != l->nkeys-1) {
		memmove(&l->index[i], &l->index[i+1], (l->nkeys - (i+1)) * sizeof *l->index);
		l->index[--l->nkeys] = old;
		TOUCH(buf, &l->index[i], (l->nkeys+1 - i) * sizeof *l->index);
	} else {
		--l->nkeys;
	}

	TOUCH_ELEM(buf, &l->nkeys);
}

void
fs_emapleaf_init_leaf(fs_Emapleaf *l)
{
	int i;

	for (i = 0; i < FS_EMAP_EXTENTS_MAX; i++)
		l->index[i] = i;
	l->nkeys = 0;
}

void
fs_emapleaf_init(fs_Context *context, fs_Buf *buf)
{
	fs_Emapleaf *l = buf->u.emapleaf;

	fs_emapleaf_init_leaf(l);

	FS_BUF_ZERO_ALL(buf);
	TOUCH_ELEM(buf, &l->nkeys);
	TOUCH(buf, l->index, sizeof l->index);
}

void
fs_emapleaf_compact_init(fs_Context *context, fs_Buf *buf)
{
	fs_EmapleafCompact *l = buf->u.emapleafCompact;
	int i;

	for (i = 0; i < FS_EMAP_COMPACT_EXTENTS_MAX; i++)
		l->index[i] = i;
	l->nkeys = 0;

	FS_BUF_ZERO_PART(buf, 0, FS_COMPACT_PAGE_SIZE);
	TOUCH_ELEM(buf, &l->nkeys);
	TOUCH(buf, l->index, sizeof l->index);
}

int emap_splits;

static fs_Bt_split
fs_emapleaf_split(fs_Context *context, fs_Inode *inode, fs_Buf *buf, int pivot)
{
	fs_Emapleaf *l = buf->u.emapleaf;
	uint16 swap[FS_EMAP_EXTENTS_MAX];
	fs_Emapleaf *l2;
	fs_Bt_split p;
	fs_Buf *buf2;
	int i;

	emap_splits++;

	FS_ASSERT(!EMAP_IS_COMPACT(inode));

	/*
	 * Allocate a new buffer to store half the entries.
	 */
	buf2 = B_CREATE(inode);
	l2 = buf2->u.emapleaf;
	p.buf = buf2;

	/*
	 * Initialize its index heap.
	 */
	fs_emapleaf_init(context, buf2);

	/*
	 * Split the page, where the promoted key is defined as the
	 * pivot's offset + extent length - 1. That way, the Btree
	 * condition of descending to the left on "<=" is preserved.
	 */
	p.key = get_offset(inode, buf, pivot)
                + fs_extentLength(get_extent(inode, buf, pivot)) - 1;

	/*
	 * First copy the keys less or equal to the pivot into l2.
	 */
	for (i = 0; i <= pivot; i++) {
		l2->offsets[i] = get_offset(inode, buf, i);
		l2->extents[i] = get_extent(inode, buf, i);
		l2->nkeys++;
	}

	l2->nkeys = pivot+1;
	l->nkeys -= l2->nkeys;

	TOUCH(buf2, &l2->offsets[0], (pivot+1) * sizeof *l2->offsets);
	TOUCH(buf2, &l2->extents[0], (pivot+1) * sizeof *l2->extents);
	TOUCH_ELEM(buf2, &l2->nkeys);
	TOUCH_ELEM(buf, &l->nkeys);

	/*
	 * Do memmoves to make "l"s entries move to the front of its
	 * index.
	 */
	memcpy(swap, &l->index[0], (pivot+1) * sizeof *l->index);
	memmove(&l->index[0], &l->index[pivot+1], l->nkeys * sizeof *l->index);
	memcpy(&l->index[l->nkeys], swap, (pivot+1) * sizeof *l->index);

	TOUCH(buf, l->index, sizeof l->index);

	return p;
}

void
fs_emapleaf_compact_expand(
    fs_Context *context,
    fs_Inode *inode,
    fs_Buf *buf)
{
	fs_Emapleaf *newLeaf;
	fs_EmapleafCompact *compact = buf->u.emapleafCompact;
	int i;

	// printf("expanding %lld\n", buf->page);

	newLeaf = fs_ramAlloc(FS_RAM_READ_DATA, PAGE_SIZE);
	bzero(newLeaf, PAGE_SIZE);
	fs_emapleaf_init_leaf(newLeaf);

	// Copy the entries.
	newLeaf->nkeys = compact->nkeys;
	for (i = 0; i < compact->nkeys; i++) {
		newLeaf->offsets[i] = get_offset(inode, buf, i);
		newLeaf->extents[i] = get_extent(inode, buf, i);
	}

	// We know its a compact leaf, but its possible
	// that the underlying area is actually an ultra
	// buf, so check it and do the right dealloc.
	if (fs_bufIsCompact(buf))
    {
 		fs_ramDealloc(FS_RAM_COMPACT_LEAF, compact);
	}
    else
	{
        fs_ramDealloc(FS_RAM_READ_DATA, compact);
    }
    
	buf->u.emapleaf = newLeaf;

	fs_bufLeafExpanded(context, buf);
	
	TOUCH_ELEM(buf, &newLeaf->nkeys);
	TOUCH(buf, newLeaf->index, sizeof newLeaf->index);
	TOUCH(buf, newLeaf->offsets, newLeaf->nkeys * sizeof *newLeaf->offsets);
	TOUCH(buf, newLeaf->extents, newLeaf->nkeys * sizeof *newLeaf->extents);

	inode->dinode.attrs &= ~FS_IFLG_COMPACT_EMAP;
	fs_inodeWrite(context, inode, INO_WRITE_LOG);

	emapleaf_check(inode, buf);
}

fs_Bt_insert
fs_emapleaf_insert(fs_Context *context,
		   fs_Inode *inode,
		   fs_Buf *buf,
		   int level,
		   uint64 key,
		   void *extentp0,
		   boolean split_allowed,
		   fs_Bt_insert_dispatch *insert)
{
	fs_Extent extent = *(fs_Extent *)extentp0;
	uint64 offset = key;
	boolean found;
	fs_Bt_insert p;
	int i;

	// printf("insert offset %lld, inum %lld\n", key, inode->inum);

	p.blknum = 0;
	found = fs_emapleaf_scan(inode, buf, &i, offset);
	assert(!found);

	if (EMAP_IS_COMPACT(inode) && get_nkeys(buf) == FS_EMAP_COMPACT_EXTENTS_MAX) {
		fs_emapleaf_compact_expand(context, inode, buf);
	}

	if (get_nkeys(buf) == FS_EMAP_EXTENTS_MAX) {
		fs_Bt_split sp;
		if (!split_allowed) {
			fs_bufReleaseExclusive(context, buf);
			p.status = FS_BT_NEEDS_SPLIT;
			return p;
		}

		sp = fs_emapleaf_split(context, inode, buf, get_nkeys(buf) / 2);
		if (key <= sp.key) {
			buf = sp.buf;
		}
		p.key = sp.key;
		p.blknum = fs_blockNumber(sp.buf->block);
		fs_emapleaf_scan(inode, buf, &i, offset);
	}

	p.status = FS_BT_INSERT_OK;
	fs_emapleaf_addent(context, inode, buf, i, offset, extent);

	return p;
}

fs_Bt_remove
fs_emapleaf_remove(fs_Context *context,
		   fs_Inode *inode,
		   fs_Buf *buf,
		   int level,
		   uint64 key,
		   void *arg0,
		   boolean dispose_allowed,
		   fs_Bt_remove_dispatch *remove)

{
	uint64 offset = key;
	fs_Bt_remove r;
	boolean found;
	int i;

	r.status = FS_BT_REMOVE_OK;
	found = fs_emapleaf_scan(inode, buf, &i, offset);
	assert(found);

	if (get_nkeys(buf) == 1 && !dispose_allowed) {
		fs_bufReleaseExclusive(context, buf);
		r.status = FS_BT_REMOVE_NEEDS_DISPOSAL;
		return r;
	}

	r.nblocks = fs_extentLength(get_extent(inode, buf, i));
	fs_emapleaf_removeent(context, buf, i);

	if (get_nkeys(buf) == 0) {
		B_FREE(buf);
		r.status = FS_BT_REMOVE_OK_EMPTY;
	}

	return r;
}

uint64
fs_emapleaf_search(fs_Context *context,
		   fs_Inode *inode,
		   fs_Buf *buf,
		   int level,
		   uint64 key,
		   void *pair0,
		   fs_Bt_search_dispatch *search,
                   boolean exclusive)
{
	fs_Emappair *p = pair0;
	uint64 offset = key;
	boolean found;
	int i;

	found = fs_emapleaf_scan(inode, buf, &i, offset);

	/*
	 * Return the matching extent but also the matching offset
	 * because the caller might have called with an offset matching
	 * the middle of the extent.
	 */
	if (found) {
		p->offset = get_offset(inode, buf, i);
		p->extent = get_extent(inode, buf, i);
	}

	fs_bufRelease(context, buf);
	return 0;
}

/*
 * Shrink the extentmap B-tree by one page. Call into the block allocator
 * to free each extent one at a time.
 */
fs_Bt_remove
fs_emapleaf_shrink(fs_Context *context,
		   fs_Inode *inode,
		   fs_Buf *buf,
		   int level,
		   uint64 key,
		   void *arg0,
		   boolean dispose_allowed,
		   fs_Bt_remove_dispatch *remove)

{
	fs_Bt_remove r;
	int num_blks_done;
#define BLKS_PER_PASS 64

	r.status = FS_BT_REMOVE_OK_EMPTY;
	r.nblocks = 0;

	assert(dispose_allowed);

#if 0 // XXX: for now, limit the size of transactions by only doing BLKS_PER_PASS blocks
	for (i = 0, num_blks_done = 0; i < get_nkeys(buf); i++) {
		fs_Extent extent = get_extent(inode, buf, i);
		r.nblocks += fs_extentLength(extent);
		num_blks_done += fs_extentLength(extent);
		fs_deallocateExtent(context, extent);

		if ((num_blks_done % BLKS_PER_PASS) == 0) {
			esm_event_t event;
			// Yield the CPU for a bit, this can consume
			// a great deal of log space.
			event.event = FS_THREAD_EVENT_SYNC;
			rc = esm_deliverEvent(context->smHandle, &event);
			if (rc != NFX_OK) {
				fs_error(context, FS_EEE_ERR,
					 "fs_emapleaf_shrink");
			}
			esm_yield(&event, 10);
			assert(event.event == FS_THREAD_EVENT_SYNC);
		}
	}
#else
	num_blks_done = 0;
	while (get_nkeys(buf) > 0 && num_blks_done < BLKS_PER_PASS) {
		fs_Extent extent = get_extent(inode, buf, get_nkeys(buf) - 1);
        uint32 length = fs_extentLength(extent);

        //printf("shrink found blk %lld,%d offset %lld\n",
        //  fs_blockNumber(fs_extentStartBlock(extent)),
        //  fs_extentLength(extent),
        //  get_offset(inode, buf, get_nkeys(buf) - 1));

        if (length + num_blks_done > BLKS_PER_PASS)
        {
            fs_Extent *extentp = get_extent_ref(inode, buf, get_nkeys(buf) - 1);
            fs_Extentpair pair = fs_splitExtent(extent, length -
                        (BLKS_PER_PASS - num_blks_done));

            // deallocate this piece
            extent = pair.right;
            length = fs_extentLength(extent);

            // re-insert this piece
            *extentp = pair.left;
			TOUCH_ELEM(buf, extentp);
        }
        else
        {
            fs_emapleaf_removeent(context, buf, get_nkeys(buf) - 1);
        }

		r.nblocks += length;
		num_blks_done += length;
		fs_deallocateExtent(context, extent);
	}

	assert(r.nblocks > 0); // make sure some work got done

	if (get_nkeys(buf) == 0) {
		B_FREE(buf);
	} else {
		r.status = FS_BT_REMOVE_OK;
	}
	return r;
#endif

	B_FREE(buf);
	return r;
}

fs_Bufpair
fs_emapleaf_cleave(fs_Context *context,
		   fs_Inode *inode,
		   fs_Buf *buf,
		   int level,
		   uint64 key,
		   void *arg0,
		   fs_Bt_cleave_dispatch *cleave)
{
	fs_Emapleaf *l;
	fs_Bt_split p;
	uint64 offset = key;
	boolean found;
	fs_Bufpair cleavepair;
	int i;

	found = fs_emapleaf_scan(inode, buf, &i, offset);

	/*
	 * If the given offset is <= the smallest offset in the leaf,
	 * or > the biggest offset in the leaf, return an empty buffer
	 * since no split is necessary.
	 */
	if (i == 0) {
		if (!found || offset == get_offset(inode, buf, i)) {
			cleavepair.left = 0;
			cleavepair.right = fs_blockNumber(buf->block);
			fs_bufReleaseExclusive(context, buf);
			return cleavepair; // an empty buffer on the left.
		}
	} else if (i == get_nkeys(buf)) {
		cleavepair.right = 0;
		cleavepair.left = fs_blockNumber(buf->block);
		fs_bufReleaseExclusive(context, buf);
		return cleavepair; // an empty buffer on the right.
	}

    if (EMAP_IS_COMPACT(inode)) {
        fs_emapleaf_compact_expand(context, inode, buf);
    }
	l = buf->u.emapleaf;

	/*
	 * If the offset exactly matches the start of an extent or falls
	 * into a hole, then the leaf must be split on the previous extent.
	 * Otherwise, an extent itself must be split up into two pieces and
	 * added back in to each new leaf.
	 */
	if (!found || offset == get_offset(inode, buf, i)) {
		p = fs_emapleaf_split(context, inode, buf, i - 1);
	} else {
		fs_Extent extent = get_extent(inode, buf, i);
		uint64 old_offset = get_offset(inode, buf, i);
		fs_Extentpair sp;

		sp = fs_splitExtent(extent, offset - old_offset);

		fs_emapleaf_removeent(context, buf, i);

		p = fs_emapleaf_split(context, inode, buf, i - 1);

		fs_emapleaf_addent(context, inode, p.buf, i, old_offset, sp.left);
		fs_emapleaf_addent(context, inode, buf, 0, offset, sp.right);
	}

	cleavepair.right = fs_blockNumber(buf->block);
	cleavepair.left = fs_blockNumber(p.buf->block);

	return cleavepair;
}

/*
 * Set blocks allocated by extents in an emap leaf into a bitmap for eek.
 */
fs_Bt_eeksearch
fs_emapleaf_eek_mark(fs_Context *context,
		 fs_Inode *inode,
		 fs_Buf *buf,
		 int level,
		 uint64 key,
		 void *eek0,
		 fs_Bt_eek_dispatch *disp)
{
	fs_Bt_eek *eek = eek0;
	uint8 *bitmap = eek->bitmap;
	fs_Bt_eeksearch r;
	int nkeys = get_nkeys(buf);
	int i;

	r.key = 0;
	r.status = FS_BT_EEK_OK;
    r.fixIt = TRUE;
	
	if (!(eek->marktype&FS_EEK_MARK_EXTENTS)
	 || !fs_emapleaf_eek_check(context, inode, buf, &r)) {
		return r;
	}

	/*
	 * Buf may have been zeroed in eek_check; return early w/disposal.
	 */
	if (r.status == FS_BT_EEK_DISPOSE) {
        fs_eek_unmark(context, bitmap, fs_blockNumber(buf->block));
        fs_bufFree(context, buf, BUF_DO_DEALLOCATE|BUF_DONT_LOG_FLUSH);
		return r;
	}

	for (i = 0; i < nkeys; i++) {
		fs_Extent extent = get_extent(inode, buf, i);
		int nblocks = fs_extentLength(extent);
		int j;

		for (j = 0; j < nblocks; j++) {
			fs_Block block = fs_extentBlock(extent, j);
			uint64 blocknum = fs_blockNumber(block);

			if (!fs_eek_mark(context, "(I#%lld) emapleaf block number", bitmap, blocknum, inode->inum)) {
                /*
                 * duplicate block in buf; return early w/disposal 
                 */
                if (!query_dispose(context, buf, &r, "Dispose?")) {
                    return r;
                }
                if (r.status == FS_BT_EEK_DISPOSE) {
		            fs_eek_unmark(context, bitmap, fs_blockNumber(buf->block));
                    fs_bufFree(context, buf, BUF_DO_DEALLOCATE|BUF_DONT_LOG_FLUSH);
		            return r;
                }
            }
		}
	}

	fs_bufReleaseExclusive(context, buf);
	return r; /* continue the search at the next page */
}

void
fs_emap_walk_extent(
	fs_Context *context,
	fs_Inode *inode,
	fs_Extent extent,
	uint64 offset,
	boolean (*worker)(fs_Context *context, fs_Inode *inode, fs_WalkResult *result, void *arg0),
	void *arg0)
{
    fs_WalkResult result;

    result.type = FS_WALK_EXTENT_TYPE;
    result.extent = extent;
    result.offset = offset;
    worker(context, inode, &result, arg0);
}

uint64
fs_emapleaf_walk(fs_Context *context,
		 fs_Inode *inode,
		 fs_Buf *buf,
		 int level,
		 uint64 key,
		 void *walk0,
		 fs_Bt_search_dispatch *search,
                 boolean exclusive)
{
	fs_Bt_walk *walk = walk0;
	fs_Bt_eeksearch r;
	int nkeys = get_nkeys(buf);
	int i;

	r.key = 0;
	r.status = FS_BT_EEK_OK;
    r.fixIt = FALSE;
	if (!fs_emapleaf_eek_check(context, inode, buf, &r)) {
		return 0;
	}
	
	for (i = 0; i < nkeys; i++) {
		fs_Extent extent = get_extent(inode, buf, i);
		fs_emap_walk_extent(context, inode, extent, get_offset(inode, buf, i), walk->worker, walk->arg);
	}

	fs_bufReleaseExclusive(context, buf);
	return 0; /* continue the search at the next page */
}

#endif // SSC
