/*
 *------------------------------------------------------------------------
 * File:        chash-api.h
 *
 * RCS: $Id: chash-api.h,v 1.8 2001/11/06 19:39:04 bob Exp $
 * 
 * Purpose:
 *      API for hash tables for simple chained hash tables
 *
 * File History:
 *      
 *
 *------------------------------------------------------------------------
 */

#if !defined(CHASH_API_H)
#define CHASH_API_H

#include "fs-primitive.h"

/*
 *------------------------------------------------------------------------
 * Function:    chash_create
 * 
 * Description:
 *      Create a chained hash. passed hash, compare and destroy functions
 *      returns hid >=0 if okay, -1 in case of failure
 *
 * Possible extensions:
 *      Moduke owner, id, name Used to describe/debug the hash table.
 * 
 * Function History:
 *      
 *
 *------------------------------------------------------------------------
 */
uint32 chash_create (uint32 array_elements, 
                     uint32 (*hash_func) (const void *key),
                     int32 (*keycmp_func) (const void *key1, const void *key2),
                     void (*destruct_func) (void *key, void *obj));

/*
 *------------------------------------------------------------------------
 * Function:    chash_destroy
 * 
 * Description:
 *      Destroy a chained hash.
 * 
 * Function History:
 *      
 *
 *------------------------------------------------------------------------
 */
void chash_destroy (uint32 hid);

/*
 *------------------------------------------------------------------------
 * Function:    chash_insert
 * 
 * Description:
 *      Insert a node into the chained hash.
 *      return 0 if OK, -1 in case of failure
 * 
 * Function History:
 *      
 *
 *------------------------------------------------------------------------
 */
int32 chash_insert (uint32 hid, void *key, void *obj);

/*
 *------------------------------------------------------------------------
 * Function:    chash_remove
 * 
 * Description:
 *      Remove a node from the chained hash.
 *      return 0 if OK, -1 in case of failure
 * 
 * Function History:
 *      
 *
 *------------------------------------------------------------------------
 */
int32 chash_remove (uint32 hid, void *key);

/*
 *------------------------------------------------------------------------
 * Function:    chash_find
 * 
 * Description:
 *      Find a node in the chained hash.
 *      return the obj, or NULL if not found
 * 
 * Function History:
 *      
 * It is up to the caller the lock the object !!!!
 *
 *------------------------------------------------------------------------
 */
void *chash_find (uint32 hid, void *key);

/*
 *------------------------------------------------------------------------
 * Function:    chash_reset
 * 
 * Description:
 *      Reset the iterator for this chained hash.
 * 
 * Function History:
 *      
 *
 *------------------------------------------------------------------------
 */
int32 chash_reset (uint32 hid);

/*
 *------------------------------------------------------------------------
 * Function:    chash_next
 * 
 * Description:
 *      Get the next value, move the iterator
 * 
 * Function History:
 *      
 *
 *------------------------------------------------------------------------
 */
int32 chash_next (uint32 hid, void **keyp, void **objp);

/*
 *------------------------------------------------------------------------
 * Function:    chash_stat
 * 
 * Description:
 *      View the statistics for given hash table
 * 
 * Function History:
 *      
 *
 *------------------------------------------------------------------------
 */
int32 chash_stat (uint32 hid);

/*
 *------------------------------------------------------------------------
 * Function:    chash_stat_nbr
 * 
 * Description:
 *      returns the statistics for given hash table
 *      array_sz: the size of the array
 *      elements: the current total number of elements/nodes in the table 
 *      curr_max_size: the current max on one row
 *      max_size: the overall max once on one row
 * 
 * Function History:
 *      
 *
 *------------------------------------------------------------------------
 */
int32
chash_stat_nbr (
    uint32 hid, 
    uint32 *array_sz, 
    uint32 *elements, 
    uint32 *curr_max_size,
    uint32 *max_size);


#endif // CHASH_API_H
