/*
 * Name:  fs-admin.c
 *
 * RCS:  $Id: fs-admin.c,v 1.57 2002/12/09 19:00:31 lanc Exp $
 *
 * Description:  Handlers for miscellaneous administrative/control requests.
 *
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#if !defined(SSC)

#include "fs-includes.h"


/**
 * Thread for creating a snapshot.
 */
int32 fs_snapCreateThread(
    fs_Context *context)
{
    fs_MsgSnapCreateReq *req = (fs_MsgSnapCreateReq *) context->reqHdr;
    return fs_snapCreate(context, req->snapDirName);
}


/**
 * Thread for removing a snapshot.
 */
int32 fs_snapRemoveThread(
    fs_Context *context)
{
    fs_MsgSnapRemoveReq *req = (fs_MsgSnapRemoveReq *) context->reqHdr;
    return fs_snapRemove(context, req->snapDirName);
}


/**
 * Thread for reverting to a snapshot.
 */
int32 fs_snapRevertThread(
    fs_Context *context)
{
    fs_MsgSnapRevertReq *req = (fs_MsgSnapRevertReq *) context->reqHdr;
    int snapnum;

    switch (req->type) {
        case FS_SNAP_REVERT_NAME:
            snapnum = fs_getSnapNum(context, req->snapname);
            if (snapnum == -1)
                return FS_ERR_SNAP_NO_EXIST;
            break;
        case FS_SNAP_REVERT_ID:
            // convert to internal snapnum
            snapnum = req->snapnum + FS_MAX_USER_SNAPSHOTS;
            break;
        case FS_SNAP_REVERT_NEWEST:
            snapnum = fs_snapGetNewest(context);
            if (snapnum == -1)
                return FS_ERR_SNAP_NO_EXIST;
            break;
    }
    return fs_snapRevert(context, snapnum);
}


/**
 * Thread for setting snapshot schedule.
 */
int32 fs_snapSchedThread(
    fs_Context *context)
{
    fs_MsgSnapSchedReq *req = (fs_MsgSnapSchedReq *) context->reqHdr;
    return fs_snapSetSched(context, &req->sched, TRUE);
}


/**
 * Thread for enabling or disabling the snapshot feature.
 */
int32 fs_snapStateThread(
    fs_Context *context)
{
    fs_MsgSnapStateReq *req = (fs_MsgSnapStateReq *) context->reqHdr;
    fs_snapState(context, req->enable);
    return NFX_OK;
}


/**
 * Thread for enabling or disabling the snapshot feature.
 */
int32 fs_snapPinThread(
    fs_Context *context)
{
    fs_MsgSnapPinReq *req = (fs_MsgSnapPinReq *) context->reqHdr;
    int32 snapnum;

    // don't allow pinning the reserved names

    if (FS_SNAP_RESERVE_NAME(req->snapname)) {
        return FS_ERR_SNAP_NOT_ENABLED;
    }

    snapnum = fs_getSnapNum(context, req->snapname);
    if (snapnum < 0) {
        return FS_ERR_SNAP_NO_EXIST;
    }
    if (req->pin)
        return fs_snapPin(context, snapnum);
    else
        return fs_snapUnpin(context, snapnum);
}


/**
 * Thread for renaming snapshots.
 */
int32 fs_snapRenameThread(
    fs_Context *context)
{
    fs_MsgSnapRenameReq *req = (fs_MsgSnapRenameReq *) context->reqHdr;
    return fs_snapRename(context, req->oldname, req->newname);
}


/**
 * Thread for snapshot admin.
 */
int32 fs_snapAdminThread(
    fs_Context *context)
{
    fs_FileSystem   *fileSystem;

    fs_MsgSnapAdminReq *req = (fs_MsgSnapAdminReq *) context->reqHdr;
    if (req->type == FS_SNAP_ADMIN_SCHED) {
        fs_snapSched(context, &req->data.time);
    }
    if (req->type == FS_SNAP_ADMIN_SET_INTERNAL) {
        fileSystem = fs_fileSystemForVolume(req->data.snapInternals.volId);
        if (fileSystem == NULL) {
            return NFX_ERR;
        }
        context->volId = req->data.snapInternals.volId;
        context->fileSystem = fileSystem;
        fs_snapSetInternal(context,
                           req->data.snapInternals.maxInternalSnaps,
                           req->data.snapInternals.internalTimeout,
                           FALSE);
    }
    return NFX_OK;
}


/**
 * Thread for snapshot mgmt show.
 */
int32 fs_snapShowThread(
    fs_Context *context)
{
    fs_MsgSnapShowReq *req = (fs_MsgSnapShowReq *) context->reqHdr;
    fs_MsgSnapShowResp *resp;
    fs_FileSystem *fileSystem;

    // setup response
    resp = (fs_MsgSnapShowResp *) context->respHdr;
    resp->type = req->type;
    resp->volId = req->volId;

    // setup context with vol and fs

    fileSystem = fs_fileSystemForVolume(req->volId);
    if (!fileSystem)
        return NFX_ERR;
    context->volId = req->volId;
    context->fileSystem = fileSystem;

    switch (req->type) {
        case FS_SNAP_SHOW_SCHED:
            if (fs_snapGetSched(context, &resp->sched) == NFX_ERR) {
                return NFX_ERR;
            }
            break;
        case FS_SNAP_SHOW_USAGE:
            resp->totalBlks = fileSystem->super->numBlks;
            resp->blksUsed = fileSystem->super->blksUsed;
            resp->nUsers = 0;
            fs_snapGetInfo(context, &resp->user[0], &resp->nUsers, FALSE);
            resp->nInternals = 0;
            fs_snapGetInfo(context, &resp->internal[0], &resp->nInternals, TRUE);
            break;
        default:
            return NFX_ERR;
    }
    context->respSize = sizeof(fs_MsgSnapShowResp);
    return NFX_OK;
}


/**
 * Thread framework for file system initialization (makefs).
 */
int32 fs_makeFsThread(fs_Context *context)
{
    fs_MsgMakeFsReq *req = (fs_MsgMakeFsReq *)  context->reqHdr;

    FS_TRACE(context, FS_TRACE_REQ, ("makefs blocks %lld blockSize %d",
             req->numBlks, req->blkSize));

    context->respSize = sizeof (fs_MsgMakeFsResp);

    if (fs_fileSystemForVolume(context->volId) != NULL)
    {
        fs_warn(context, FS_ALREADY_MOUNTED_ERR, "fs_makeFsThread");
        return FS_ERR_ALREADY_MOUNTED;
    }

    if (fs_formatFileSystem(context, req->numBlks,
                            req->blkSize, req->ext_log))
    {
        return NFX_OK;
    }
    else
    {
        return NFX_ERR;
    }

} /* fs_makeFsThread */

/**
 * Extent map debug test thread
 */
int32
fs_emapdebugThread(fs_Context *context)
{
    //fs_MsgEmapdebugReq *req = (fs_MsgEmapdebugReq *) context->reqHdr;

    FS_TRACE(context, FS_TRACE_REQ, ("emapdebug"));

    context->respSize = sizeof (fs_MsgEmapdebugResp);

    fs_emapdebug(context);

    fs_logCommitEntry(context);

    return NFX_OK;
}

/**
 * Drive eek from the shell
 */
int32
fs_eekThread(fs_Context *context)
{
    enum {FS_EEK_OK, FS_EEK_WRONG_STATE, FS_EEK_STILL_BUDDY};
    int rc = FS_EEK_OK;
	char *state;
	fs_MsgEekReq *eekreq = (void *)context->reqHdr;
    fs_FileSystem *fileSystem;

	FS_TRACE(context, FS_TRACE_REQ, ("eek"));
    context->eek.defaultReply = eekreq->default_reply;
    context->eek.sendToSsc = TRUE;
    // context->eek.nthreads = eekreq->nthreads;
    // context->eek.phase = eekreq->phase;

	fileSystem = fs_fileSystemForVolume(context->volId);
	if (fileSystem == NULL) {
		printf(
            "eek on unmounted filesystem (0x%llx) -- running interactively\n",
                context->volId.whole_thing);
		context->eek.interactive = 1;
		fs_doMount(context, (void *)&eekreq->mountreq);
	    ((fs_MsgEekResp *) context->respHdr)->quit = TRUE;
	    ((fs_MsgEekResp *) context->respHdr)->has_status = FALSE;
	} else {
        // PR #2116 don't allow eek until log replay is complete ie must be online
        if ((fileSystem->state != FS_VOL_STATE_ONLINE) &&
            (fileSystem->state != FS_VOL_STATE_ONLINE_RDONLY)) {
            switch (fileSystem->state) {
                case FS_VOL_STATE_OFFLINE:
                    state = "offline";
                    break;
                case FS_VOL_STATE_PAUSED:
                    state = "paused";
                    break;
                case FS_VOL_STATE_MOUNTING:
                    state = "mounting";
                    break;
                case FS_VOL_STATE_UNMOUNTING:
                    state = "umounting";
                    break;
                case FS_VOL_STATE_FAILING:
                    state = "failing";
                    break;
                case FS_VOL_STATE_BACKUP_RESTORE:
                    state = "backup-restore";
                    break;
                default:
                    state = "unknown";
                    break;
            }
            rc = FS_EEK_WRONG_STATE;
        }
        if (rc == FS_EEK_OK) {
            // PR #2126 also until buddy recal is complete
            NFX_SPIN_LOCK(&context->fileSystem->spin);
            rc = (context->fileSystem->buddyThreadHandle != 0) ? FS_EEK_STILL_BUDDY: FS_EEK_OK;
            NFX_SPIN_UNLOCK(&context->fileSystem->spin);
        }
        switch (rc) {
            case FS_EEK_OK:
		        fs_lockExclusive(context, &context->fileSystem->accessLock, NULL);
		        fs_lockExclusive(context, &context->fileSystem->modifyLock, NULL);

                // eek reads inodes from buffer cache and not inode cache
                // so sync everything to disk first so eek will see the latest

                fs_syncThread(context);

                // non interactive if the file system is mounted
                // since eek does not log and any modifications
                // will not be guaranteed if a crash occurs
		        fs_eek(context);
		        fs_unlockExclusive(context, &context->fileSystem->modifyLock);
		        fs_unlockExclusive(context, &context->fileSystem->accessLock);
	            ((fs_MsgEekResp *) context->respHdr)->quit = TRUE;
	            ((fs_MsgEekResp *) context->respHdr)->has_status = FALSE;
                break;
            case FS_EEK_WRONG_STATE:
                sprintf(((fs_MsgEekResp *) context->respHdr)->msg,
                        "Cannot run eek with filesystem (%s) in state %s",
                        fileSystem->super->mountName, state);
	            ((fs_MsgEekResp *) context->respHdr)->has_status = TRUE;
	            ((fs_MsgEekResp *) context->respHdr)->quit = TRUE;
                break;
            case FS_EEK_STILL_BUDDY:
                sprintf(((fs_MsgEekResp *) context->respHdr)->msg,
                        "Cannot run eek on filesystem (%s): buddy recal thread still running",
                        fileSystem->super->mountName);
	            ((fs_MsgEekResp *) context->respHdr)->has_status = TRUE;
	            ((fs_MsgEekResp *) context->respHdr)->quit = TRUE;
                break;
        }
	}

	((fs_MsgEekResp *) context->respHdr)->has_query = FALSE;
	context->respSize = sizeof(fs_MsgEekResp);
	return NFX_OK;
}


/**
 * Handle the checkfs request.
 */
int32
fs_checkFsThread(
    fs_Context      *context)
{
    return NFX_ERR;
}


/**
 * Obtain a filehandle for the specified pathname.
 *
 *  !!! This routine is destructive to pathName !!!
 */
int32
fs_getHandleFromPath(
    fs_Context  *context,
    char        *pathName,
    fs_Handle   *handle)
{
    fs_Inode            *parentInode, *inode = NULL;
    char                *sepPtr;
    uint64              inum;

    inode = I_GET_SHARED(0, FS_ROOT_INODE);
    
    if (inode == NULL)
    {
        printf("getRoot: failed to get root inode!\n");
        return NFX_ERR;
    }

    //printf("getRoot for path %s\n", pathName);

    while (*pathName)
    {
        // Eat leading '/'
        while (*pathName && *pathName == '/')
            pathName += 1;
    
        if (!(*pathName))
        {
            break;
        }
    
        sepPtr = strchr(pathName, '/');
        if (sepPtr)
            *sepPtr = '\0';

        parentInode = inode;

        //printf("looking up pathName %s len: %d\n",
        //                      pathName, strlen(pathName));

        // Disallow mounting of .snapshot dirs.  TODO - allow this?
        if (!strcmp(pathName, ".snapshot"))
        {
            // TODO - syslog this?
            fs_inodeRelease(context, parentInode);
            return FS_ERR_ACCESS;
        }

        inum = fs_dir_lookupent(context,
                                parentInode,
                                pathName,
                                strlen(pathName));
        if (inum == 0)
        {
            // Not found.
            fs_inodeRelease(context, parentInode);
            //printf("inode not found for path component\n");
            return NFX_ERR;
        }

        inode = I_GET_SHARED((parentInode->inum == FS_SNAPSHOT_DIR_INODE)
                                ? FS_HANDLE_GET_SNAP(inum)
                                : parentInode->snapnum,
                             inum);
        if (!inode)
        {
            FS_ASSERT(FALSE);
            fs_inodeRelease(context, parentInode);
            return NFX_ERR;
        }

        if (!fs_permissionOkay(context, inode, FS_PERM_LOOKUP))
        {
            fs_inodeRelease(context, parentInode);
            fs_inodeRelease(context, inode);
            //printf("no permission for path component\n");
            return FS_ERR_ACCESS;
        }

        fs_inodeRelease(context, parentInode);
		if (!sepPtr)
			break;
        pathName = sepPtr + 1;
    }
    
    fs_handleInit(context->fileSystem,
                  handle,
                  inode->snapnum,
                  inode->inum,
                  inode->dinode.genNum,
                  inode->inum);

    fs_inodeRelease(context, inode);

    return NFX_OK;

} /* fs_getHandleFromPath */


/**
 * Handle the GET_ROOT request.
 */
int32
fs_getRootThread(
    fs_Context          *context)
{
    fs_MsgGetRootReq    *req;
    fs_MsgGetRootResp   *resp;
	char				*pathName = "/";

    req = (fs_MsgGetRootReq *) context->reqHdr;
    resp = (fs_MsgGetRootResp *) context->respHdr;
    context->respSize = sizeof(fs_MsgGetRootResp);

	if (req->pathLen > 0)
		pathName = req->pathName;

    return fs_getHandleFromPath(context, pathName, &resp->dirHandle);

} /* fs_getRootThread */


/**
 * Handle the SYNC request.
 */
int32
fs_syncThread(
    fs_Context      *context)
{
    // Sync the inode cache.
    // This will block until completed.
    fs_inodeSyncCache(context);

    // Sync the buffer cache.
    // This will block until completed.
    fs_bufSyncWritebackLists(context);

    // Sync the inode cache.
    // This will block until completed.
    fs_inodeSyncCache(context);

    // Now do it again, since the first one
    // might have created additional dirty
    // bufs doing deferred allocation.
    fs_bufSyncWritebackLists(context);

//    context->respSize = sizeof(fs_MsgRespHdr);

    return NFX_OK;

} /* fs_syncThread */


/** 
 * Handle the FLUSH request.
 */
int32
fs_flushThread(
    fs_Context      *context)
{
    fs_inodeCacheFlush(context);
    fs_bufCacheFlush(context, NULL);

    context->respSize = sizeof(fs_MsgRespHdr);

    return NFX_OK;

} /* fs_flushThread */


/**
 * Handle the SHOW_LOCKED request.
 */
int32
fs_showLockedThread(
    fs_Context      *context)
{
    fs_inodeShowLocked();
    
    return NFX_OK;

} /* fs_showLockedThread */


/**
 * Thread to hold the file system modification lock while it is frozen.
 */
static void
fs_freezerThread(
    esm_handle_t    handle,
    void            *arg)
{
    fs_Context      *context = arg;
    esm_event_t     event;

    fs_freeze(context);

    // notify the user
    printf("volume 0x%llx frozen\n", context->volId.whole_thing);

    // wait for thaw
    fs_yield(context, &event, FS_YIELD_TIME_ADMIN, FS_YIELD_TYPE_ADMIN);
    if (event.event != FS_THREAD_EVENT_THAW) {
        fs_error(context, FS_INTERNAL_ERR,
                 "fs_freezerThread: event %d", event.event);
    }
    assert(event.p1.u.p == NULL);
    assert(event.p2.u.p == NULL);

    fs_thaw(context);

    // we're back
    printf("volume 0x%llx thawed\n", context->volId.whole_thing);

    fs_threadTerminate(context);
}


/**
 * Quiesce a filesystem until a thaw request comes in.
 */
int32 fs_freezeThread(
    fs_Context *context)
{
    fs_FileSystem *fs = context->fileSystem;

    fs_lockExclusive(context, &fs->freezeLock, NULL);

    if (fs->freezeHandle != 0) {
        fs_warn(context, FS_INTERNAL_ERR, "already frozen");
        fs_unlockExclusive(context, &fs->freezeLock);
        return NFX_ERR;
    }

    fs->freezeHandle = fs_threadFork(context, fs_freezerThread, TRUE, FALSE,
                                     FS_DAEMON_TYPE_FREEZER, NULL);

    fs_unlockExclusive(context, &fs->freezeLock);
    return NFX_OK;
}


/**
 * Unquiesce a filesystem which has been frozen.
 */
int32 fs_thawThread(
    fs_Context *context)
{
    int rc;
    fs_FileSystem *fs = context->fileSystem;
    esm_handle_t handle;
    esm_event_t event;

    fs_lockExclusive(context, &fs->freezeLock, NULL);

    handle = fs->freezeHandle;
    if (handle == 0) {
        fs_warn(context, FS_INTERNAL_ERR, "not frozen");
        fs_unlockExclusive(context, &fs->freezeLock);
        return NFX_ERR;
    }

    fs->freezeHandle = 0;
    event.event = FS_THREAD_EVENT_THAW;
    event.p1.u.p = NULL;
    event.p2.u.p = NULL;
    rc = thread_event(handle, &event);
    if (rc != NFX_OK)
        fs_error(context, FS_EEE_ERR, "thawing freezer thread");

    fs_unlockExclusive(context, &fs->freezeLock);
    return NFX_OK;
}


/**
 * Handle the TX_COMPL notification.
 */
int32
fs_txComplInstantiate(
	const fs_ReqTypeTbl *reqType,
    const fs_MsgReqHdr  *reqHdr,
    int32               bufType,
    uint32              srcAddr)
{
    fs_MsgTransmitComplete  *req;
	esm_handle_t			handle;
    esm_event_t             event;
	int32 					rc;

    req  = (fs_MsgTransmitComplete *) reqHdr;
	handle = req->fsPrivate;
	FS_TRACE(NULL, FS_TRACE_REQ, ("smHandle %llx txcompl msg rcvd",
			(uint64)((uintptr_t)handle)));
    event.event = FS_THREAD_EVENT_TX_COMPL;
    event.p1.u.p = req;
    event.p2.u.w = bufType;
	rc = thread_event(handle, &event);
	if (rc != NFX_OK)
	{
		fs_warn(NULL, FS_EVENT_DELIVERY_FAIL, "txcompl event");
	}

	return NFX_OK;

} /* fs_txComplInstantiate */


#ifndef SBM_REV_2
/**
 * Handle the SAC_STORE notification.
 */
int32
fs_sacStoreInstantiate(
    sbm_ack_msg_t   *msg,
    int32           bufType,
    uint32          srcAddr)
{
    int32               rc;

    rc = fs_cod_write_confirm(msg->msg.cod_rcv.cod);
    if (rc != NFX_OK)
    {
        UNREACHED(); // TODO - anything we can/should do here?
    }
    if (bufType & FS_CONTEXT_SHARED_REQUEST)
    {
        eee_deallocateBuffer(msg);
    }
    else
    {
#if defined(FS_STANDALONE)
        fs_ramDealloc(FS_RAM_MCPU, msg);
#else
        fs_ramDealloc(FS_RAM_REQ_MSG, msg);
#endif
    }

    return rc;

} /* fs_sacStoreInstantiate */
#endif

/**
 * ???
 */
#if defined(FS_STANDALONE)
int32
mcpu_sim_sbm_show_sum (void);
#endif 

#endif // SSC
