// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=4 shiftwidth=4 expandtab:

/*
 * Name:  fs-bitmap.c
 *
 * RCS:  $Id: fs-bitmap.c,v 1.14 2002/10/28 21:50:38 bob Exp $
 *
 * Description:  Interface for bitmap file utilities
 * 
 * Created By:  Paul Haahr <haahr@agilestorage.com>
 *
 * Date Created:  2001-08-05
 */

#include "fs-includes.h"


// TODO - use abstract word type, to support 64-bits on 64-bit machines?


/**
 * Create a bitmap location.
 */
static inline fs_BitmapLoc fs_makeBitmapLoc(
    uint64 page,
    uint32 word,
    uint32 mask)
{
    fs_BitmapLoc loc;
    loc.page = page;
    loc.word = word;
    loc.mask = mask;
    return loc;
}


/**
 * Return true if this bitmap location is valid.
 */
inline boolean fs_bitmapLocIsValid(fs_BitmapLoc loc)
{
    return loc.mask != 0;
}


/**
 * Convert an index number into a location in a bitmap file.
 */
fs_BitmapLoc fs_bitmapIndexToLoc(
    fs_Context *context,
    uint64     index)
{
    fs_Parameters *params = &context->fileSystem->params;
    uint64 word = index / BITS_PER_WORD;
    uint32 bit  = index % BITS_PER_WORD;

    fs_BitmapLoc loc;

    assert(index != FS_INVALID_BITMAP_INDEX);
    
    loc.page = word / params->wordsPerBlock;    // TODO - shift
    loc.word = word % params->wordsPerBlock;    // TODO - mask
    loc.mask = (1UL << bit);

    //printf("bitmapIndexToLoc(%lld) = { %ld, %ld, %#lx }\n",
    //       index, loc.page, loc.word, loc.mask);
    return loc;
}


/**
 * Convert a location in a bitmap file into an index.
 */
uint64 fs_bitmapLocToIndex(
    fs_Context   *context,
    fs_BitmapLoc loc)
{
    fs_Parameters *params = &context->fileSystem->params;

    assert(fs_bitmapLocIsValid(loc));
    return ((((loc.page * params->wordsPerBlock) + loc.word) * BITS_PER_WORD)
            + log2_32(loc.mask));
}


/**
 * Return the bitmap index for the specified page, word, and (single bit) mask.
 */
static uint64 fs_makeBitmapIndex(
    fs_Context *context,
    uint64     page,
    uint32     word,
    uint32     mask)
{
    return fs_bitmapLocToIndex(context, fs_makeBitmapLoc(page, word, mask));
}


/**
 * Return true if the specified inode is a bitmap file.
 */
static boolean fs_inumIsBitmap(
    fs_Inumber inum)
{
    return ((inum == FS_FULL_INODE_BLOCK_INODE) ||
            (inum == FS_DIRECTORY_IN_INODE_BLOCK_INODE));
}


/**
 * Return true if the specified inode contains a bitmap file.
 *
 * TODO - mark this in the inode, rather than using a predefined set
 */
static boolean fs_inodeIsBitmap(
    fs_Inode *inode)
{
    return fs_inumIsBitmap(inode->inum);
}


/**
 * Log a change to the specified bitmap location, which is known to be in
 * the buffer backed by the given buffer.
 *
 * The buffer must be locked exclusively when this function is called.
 */
static void fs_bitmapLogWrite(
    fs_Context   *context,
    fs_Buf       *buffer,
    fs_BitmapLoc loc,
    uint32       mask,
    boolean      isSet,
    fs_BitmapOperation *debugOp)
{
    assert(buffer->page == loc.page);
    assert((buffer->snapshots & 1) != 0); // TODO - symbolic
    assert(fs_inumIsBitmap(buffer->inum));
    fs_bufWrite(context,
                buffer,
                loc.word * sizeof (uint32),
                sizeof (uint32),
                FS_BUF_WRITE_LOG
                | (isSet
                   ? FS_BUF_WRITE_LOG_BIT_SET
                   : FS_BUF_WRITE_LOG_BIT_RESET),
                &mask,
                debugOp);
}


/**
 * Turn on the bit in the specified bitmap file for the specified
 * location.
 */
void fs_bitmapSetLoc(
    fs_Context   *context,
    fs_Inode     *bitmap,
    fs_BitmapLoc loc,
    fs_BitmapOperation *debugOp)
{
    fs_Buf *buffer = O_GET_EXCL(bitmap, loc.page);
    uint32 word = buffer->u.word[loc.word];
    if ((word & loc.mask) == 0) {
        buffer->u.word[loc.word] |= loc.mask;
        fs_bitmapLogWrite(context, buffer, loc, loc.mask, TRUE, debugOp);
    }
    fs_bufReleaseExclusive(context, buffer);
}


/**
 * Turn off the bit in the specified bitmap file for the specified
 * location.
 */
void fs_bitmapClearLoc(
    fs_Context   *context,
    fs_Inode     *bitmap,
    fs_BitmapLoc loc,
    fs_BitmapOperation *debugOp)
{
    fs_Buf *buffer = O_GET_EXCL(bitmap, loc.page);
    uint32 word = buffer->u.word[loc.word];
    if ((word & loc.mask) != 0) {
        buffer->u.word[loc.word] |= loc.mask;
        fs_bitmapLogWrite(context, buffer, loc, loc.mask, FALSE, debugOp);
    }
    fs_bufReleaseExclusive(context, buffer);
}


/**
 * Turn on the bit in the specified bitmap file for the specified index.
 */
void fs_bitmapSet(
    fs_Context *context,
    fs_Inode   *bitmap,
    uint64     index,
    fs_BitmapOperation *debugOp)
{
    fs_bitmapSetLoc(context, bitmap, fs_bitmapIndexToLoc(context, index),
                    debugOp);
}


/**
 * Turn off the bit in the specified bitmap file for the specified index.
 */
void fs_bitmapClear(
    fs_Context *context,
    fs_Inode   *bitmap,
    uint64     index,
    fs_BitmapOperation *debugOp)
{
    fs_bitmapClearLoc(context, bitmap, fs_bitmapIndexToLoc(context, index),
                      debugOp);
}


/**
 * Clear the bit in the specified bitmap file for the specified index,
 * if it is set.  Returns true if the bit had been set.
 */
boolean fs_bitmapClearIfSet(
    fs_Context *context,
    fs_Inode   *bitmap,
    uint64     index,
    fs_BitmapOperation *debugOp)
{
    fs_BitmapLoc loc = fs_bitmapIndexToLoc(context, index);
    fs_Buf *buffer = O_GET_EXCL(bitmap, loc.page);
    uint32 word;

    assert(fs_inodeIsBitmap(bitmap));

    word = buffer->u.word[loc.word];
    if ((word & loc.mask) == 0) {
        fs_bufReleaseExclusive(context, buffer);
        return FALSE;
    }
    else {
        buffer->u.word[loc.word] = word &~ loc.mask;
        fs_bitmapLogWrite(context, buffer, loc, loc.mask, FALSE, debugOp);
        fs_bufReleaseExclusive(context, buffer);
        return TRUE;
    }
}


/**
 * Set all bits in the bitmap from the specified index and after, to the
 * end of the page.
 *
 * If I were more careful in the bitmap search functions, this wouldn't
 * be necessary.  But it's useful from a belt-and-suspenders perspective.
 */
void fs_bitmapSetBitsAfter(
    fs_Context *context,
    fs_Inode   *bitmap,
    uint64     index)
{
    fs_Parameters *params = &context->fileSystem->params;
    fs_BitmapLoc loc = fs_bitmapIndexToLoc(context, index);
    uint32 wordsPerBlock = params->blockSize / sizeof (uint32);
    fs_Buf *buffer = O_GET_EXCL(bitmap, loc.page);
    uint32 word;

    assert(fs_inodeIsBitmap(bitmap));

    buffer->u.word[loc.word] |= ~(loc.mask - 1);

    for (word = loc.word + 1; word < wordsPerBlock; word++) {
        assert(buffer->u.word[word] == 0);
        buffer->u.word[word] = -1; // all ones
    }

    FS_BUF_WRITE(buffer,
                 loc.word * sizeof (uint32),
                 (wordsPerBlock - loc.word) * sizeof (uint32));
}


/**
 * Clear all bits in the bitmap from the specified index and after, to the
 * end of the page.
 */
void fs_bitmapClearBitsAfter(
    fs_Context *context,
    fs_Inode   *bitmap,
    uint64     index)
{
    fs_Parameters *params = &context->fileSystem->params;
    fs_BitmapLoc loc = fs_bitmapIndexToLoc(context, index);
    uint32 wordsPerBlock = params->blockSize / sizeof (uint32);
    fs_Buf *buffer = O_GET_EXCL(bitmap, loc.page);

    assert(fs_inodeIsBitmap(bitmap));

    buffer->u.word[loc.word] &= (loc.mask - 1);

    bzero((void *) &(buffer->u.word[loc.word + 1]),
          (wordsPerBlock - loc.word - 1) * sizeof (uint32));

    FS_BUF_WRITE(buffer,
                 loc.word * sizeof (uint32),
                 sizeof (uint32));

    fs_logAddDataRecord(context, buffer,
                        (loc.word + 1) * sizeof (uint32),
                        (wordsPerBlock - loc.word - 1) * sizeof (uint32),
                        0, // not used
                        FS_LOG_ADD_FLAG_ZERO);

    FS_DBG_UPDATE_SHADOW(buffer,
                        (loc.word + 1) * sizeof (uint32),
                        (wordsPerBlock - loc.word - 1) * sizeof (uint32));
}


/**
 * A list of bitmap locations.
 */
struct fs_BitmapLocList {
    fs_BitmapLoc loc;
    fs_BitmapLocList *next;
};


/**
 * Create an entry in a list of bitmap locations.
 */
fs_BitmapLocList *fs_makeBitmapLocList(
    fs_Context *context,
    fs_BitmapLocList *next,
    uint64 index)
{
    fs_BitmapLocList *list = fs_contextMemAlloc(context, sizeof *list);
    list->loc  = fs_bitmapIndexToLoc(context, index);
    list->next = next;
    return list;
}


/**
 * Mask out the values from a bitmap location list from the specified
 * word at the specified page and index.
 */
static inline uint32 fs_bitmapLocListMaskOut(
    fs_Context *context,
    fs_BitmapLocList *exclude,
    uint32 page,
    uint32 index,
    uint32 value)
{
    fs_BitmapLocList *list;
    for (list = exclude; value != 0 && list != NULL; list = list->next) {
        if (list->loc.page == page && list->loc.word == index)
            value &= ~list->loc.mask;
    }
    return value;
}


/**
 * A specification for searching pages of a bitmap.
 *
 * When using this abstraction, we are searching for a one bit in the
 * words returned by the getWord function.
 */
typedef struct {
    /** Return the number of pages in the bitmap. */
    uint64 (*getPageCount)(
        fs_Context *context,
        void *arg);
    /** Begin searching on the specified page. */
    void (*beginPage)(
        fs_Context *context,
        uint64 page,
        void *arg);
    /** End searching on the specified page. */
    void (*endPage)(
        fs_Context *context,
        void *arg);
    /** Function which will return the specified word of the bitmap, where
     *  we are looking for a one bit. */
    uint32 (*getWord)(
        fs_Context *context,
        int word,
        void *arg);
    /** Pointer to data used by the getWord function. */
    void *arg;
} fs_BitmapSearcher;


/**
 * Return the number of pages in an inode.
 */
static uint64 fs_bitmapGetPageCount(
    fs_Context *context,
    fs_Inode *inode)
{
    return ceilingDiv64(inode->dinode.size,
                        context->fileSystem->params.blockSize);
}


/**
 * Choose a one bit close to the specified mask bit from the specified
 * word.
 */
static inline uint64 fs_bitmapFindBitInWord(
    fs_Context *context,
    uint32 word,
    uint32 target)
{
    assert(word != 0);
    assert(leastBit(target) == target); // target is a power of two

    if ((word & target) != 0)
        return target;
                    
    else {
        int j;
        for (j = 1;; j++) {
            uint32 hi = (target << j);
            uint32 lo = (target >> j);
            assert(j < BITS_PER_WORD);
            if ((word & hi) != 0) {
                return hi;
            }
            if ((word & lo) != 0) {
                return lo;
            }
        }
    }
}


/**
 * Search a specified page of a bitmap using the specified bitmap search
 * mechanism.  Returns the index, or the invalid sentinel if none was found.
 */
static uint64 fs_bitmapSearchPage(
    fs_Context *context,
    fs_BitmapLoc start,
    fs_BitmapSearcher *searcher,
    fs_BitmapLocList *excluded)
{
    uint32 i;
    uint32 wordsPerPage
        = context->fileSystem->params.blockSize / sizeof (uint32);

    assert(wordsPerPage == leastBit(wordsPerPage)); // is a power of two

    for (i = 0; i < wordsPerPage; i++) {
        int32 half = (i >> 1);
        int32 delta = isEven(i) ? half : -(half + 1);
        uint32 index = (start.word + delta) & (wordsPerPage - 1);

        uint32 word = searcher->getWord(context, index, searcher->arg);
        uint32 adjusted =
            fs_bitmapLocListMaskOut(
                context, excluded, start.page, index, word);
        
        if (adjusted != 0) {
            uint32 bit = fs_bitmapFindBitInWord(context, adjusted, start.mask);
            assert(bit != 0);
            assert(bit == leastBit(bit));
            return fs_makeBitmapIndex(context, start.page, index, bit);
        }
    }

    return FS_INVALID_BITMAP_INDEX;
}


/**
 * Search a bitmap.
 */
static uint64 fs_bitmapSearch(
    fs_Context *context,
    fs_BitmapSearcher *searcher,
    uint64 startIndex,
    fs_BitmapLocList *excluded)
{
    fs_BitmapLoc start = fs_bitmapIndexToLoc(context, startIndex);
    uint64 result;

    searcher->beginPage(context, start.page, searcher->arg);
    result = fs_bitmapSearchPage(context, start, searcher, excluded);
    searcher->endPage(context, searcher->arg);

    if (result == FS_INVALID_BITMAP_INDEX) {
        // Crab walk forwards and backwards
        uint64 pages = searcher->getPageCount(context, searcher->arg);
        uint64 i;
        uint64 limit = MAX(start.page, pages - start.page);

        for (i = 1; i < limit; i++) {
            if (start.page + i < pages) {
                fs_BitmapLoc loc = start;
                loc.page = start.page + i;
                searcher->beginPage(context, loc.page, searcher->arg);
                result = fs_bitmapSearchPage(context, loc, searcher, excluded);
                searcher->endPage(context, searcher->arg);
                if (result != FS_INVALID_BITMAP_INDEX)
                    break;
            }
            if (i < start.page) {
                fs_BitmapLoc loc = start;
                loc.page = start.page - i;
                searcher->beginPage(context, loc.page, searcher->arg);
                result = fs_bitmapSearchPage(context, loc, searcher, excluded);
                searcher->endPage(context, searcher->arg);
                if (result != FS_INVALID_BITMAP_INDEX)
                    break;
            }
        }
    }

    return result;
}


/**
 * Argument for bitmap searcher with a single bitmap.
 */
typedef struct {
    fs_Inode *inode;
    fs_Buf *buf;
} fs_BitmapSearch1Arg;


/**
 * Begin page function for single-bitmap searches.
 */
static uint64 fs_bitmapSearch1GetPageCount(
    fs_Context *context,
    void *varg)
{
    fs_BitmapSearch1Arg *arg = varg;
    return fs_bitmapGetPageCount(context, arg->inode);
}


/**
 * Begin page function for single-bitmap searches.
 */
static void fs_bitmapSearch1BeginPage(
    fs_Context *context,
    uint64 page,
    void *varg)
{
    fs_BitmapSearch1Arg *arg = varg;
    assert(arg->buf == NULL);
    arg->buf = O_GET_SHARED(arg->inode, page);
    assert(arg->buf != NULL);
}


/**
 * End page function for single-bitmap searches.
 */
static void fs_bitmapSearch1EndPage(
    fs_Context *context,
    void *varg)
{
    fs_BitmapSearch1Arg *arg = varg;
    assert(arg->buf != NULL);
    fs_bufRelease(context, arg->buf);
    arg->buf = NULL;
}


/**
 * Argument for bitmap searcher with two bitmaps.
 */
typedef struct {
    fs_Inode *inode1;
    fs_Inode *inode2;
    fs_Buf *buf1;
    fs_Buf *buf2;
} fs_BitmapSearch2Arg;


/**
 * Begin page function for single-bitmap searches.
 */
static uint64 fs_bitmapSearch2GetPageCount(
    fs_Context *context,
    void *varg)
{
    fs_BitmapSearch2Arg *arg = varg;
    return fs_bitmapGetPageCount(context, arg->inode1);
}


/**
 * Begin page function for double-bitmap searches.
 */
static void fs_bitmapSearch2BeginPage(
    fs_Context *context,
    uint64 page,
    void *varg)
{
    fs_BitmapSearch2Arg *arg = varg;
    assert(arg->buf1 == NULL);
    assert(arg->buf2 == NULL);
    arg->buf1 = O_GET_SHARED(arg->inode1, page);
    arg->buf2 = O_GET_SHARED(arg->inode2, page);
    assert(arg->buf1 != NULL);
    assert(arg->buf2 != NULL);
}


/**
 * End page function for double-bitmap searches.
 */
static void fs_bitmapSearch2EndPage(
    fs_Context *context,
    void *varg)
{
    fs_BitmapSearch2Arg *arg = varg;
    assert(arg->buf1 != NULL);
    assert(arg->buf2 != NULL);
    fs_bufRelease(context, arg->buf1);
    fs_bufRelease(context, arg->buf2);
    arg->buf1 = NULL;
    arg->buf2 = NULL;
}


/**
 * Search function for bitmapFindClear.  Returns a complemented value,
 * because we're looking for clear (zero) bits.
 */
static uint32 fs_bitmapFindClearGetWord(
    fs_Context *context,
    int word,
    void *varg)
{
    fs_BitmapSearch1Arg *arg = varg;
    return ~arg->buf->u.word[word];
}


/**
 * Search for a bitmap location which is not set, near the specified
 * start index, in the specified bitmap file.
 */
uint64 fs_bitmapFindClear(
    fs_Context       *context,
    fs_Inode         *bitmap,
    uint64           startIndex,
    fs_BitmapLocList *excluded)
{
    fs_BitmapSearcher searcher;
    fs_BitmapSearch1Arg arg;
    
    assert(fs_inodeIsBitmap(bitmap));

    arg.inode = bitmap;
    arg.buf = NULL;

    searcher.getPageCount = fs_bitmapSearch1GetPageCount;
    searcher.beginPage    = fs_bitmapSearch1BeginPage;
    searcher.endPage      = fs_bitmapSearch1EndPage;
    searcher.getWord      = fs_bitmapFindClearGetWord;
    searcher.arg = &arg;

    return fs_bitmapSearch(context, &searcher, startIndex, excluded);
}


/**
 * Search function for bitmapFindClear2.  Returns a complemented value,
 * because we're looking for clear (zero) bits.
 */
static uint32 fs_bitmapFindClear2GetWord(
    fs_Context *context,
    int word,
    void *varg)
{
    fs_BitmapSearch2Arg *arg = varg;
    return ~(arg->buf1->u.word[word] | arg->buf2->u.word[word]);
}


/**
 * Search for an inode bitmap location which is not set in the union of
 * the two bitmap files, near the given starting index.
 */
uint64 fs_bitmapFindClear2(
    fs_Context       *context,
    fs_Inode         *bitmap1,
    fs_Inode         *bitmap2,
    uint64           startIndex,
    fs_BitmapLocList *excluded)
{
    fs_BitmapSearcher searcher;
    fs_BitmapSearch2Arg arg;
    
    assert(fs_inodeIsBitmap(bitmap1));
    assert(fs_inodeIsBitmap(bitmap2));

    bzero(&arg, sizeof arg);
    arg.inode1 = bitmap1;
    arg.inode2 = bitmap2;

    searcher.getPageCount = fs_bitmapSearch2GetPageCount;
    searcher.beginPage    = fs_bitmapSearch2BeginPage;
    searcher.endPage      = fs_bitmapSearch2EndPage;
    searcher.getWord      = fs_bitmapFindClear2GetWord;
    searcher.arg = &arg;

    return fs_bitmapSearch(context, &searcher, startIndex, excluded);
}
