// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=4 shiftwidth=4 expandtab:

/*
 * Name:  fs-bitmap.h
 *
 * RCS:  $Id: fs-bitmap.h,v 1.4 2002/04/01 18:57:01 bobonfrogger Exp $
 *
 * Description:  Interface for bitmap file utilities
 * 
 * Created By:  Paul Haahr <haahr@agilestorage.com>
 *
 * Date Created:  2001-08-05
 */

#ifndef FS_BITMAP_H
#define FS_BITMAP_H


/**
 * The location of a bit in a bitmap file, broken down into page of the
 * file, page within that block, and bit within word.
 */
typedef struct {
    /** The page (block) within the bitmap file. */
    uint64 page;
    /** The word within the page. */
    uint32 word;
    /** A bit mask within the word. */
    uint32 mask;
} fs_BitmapLoc;


/**
 * An invalid bitmap location.
 *
 * TODO - non-GNUC version
 */
#define FS_INVALID_BITMAP_LOC  ((fs_BitmapLoc) { ~0, ~0, 0 })


/**
 * An invalid index.
 */
#define FS_INVALID_BITMAP_INDEX (~0ULL)


/**
 * Return true if the given bitmap location is valid.
 *
 * TODO - inline
 */
extern boolean fs_bitmapLocIsValid(
    fs_BitmapLoc loc);


/**
 * Convert an index number into a location in a bitmap file.
 */
extern fs_BitmapLoc fs_bitmapIndexToLoc(
    fs_Context *context,
    uint64     index);


/**
 * Convert a location in a bitmap file into an index.
 */
extern uint64 fs_bitmapLocToIndex(
    fs_Context   *context,
    fs_BitmapLoc loc);


/**
 * Turn on the bit in the specified bitmap file for the specified index.
 */
extern void fs_bitmapSet(
    fs_Context *context,
    fs_Inode   *bitmap,
    uint64     index,
    fs_BitmapOperation *debugOp);


/**
 * Turn off the bit in the specified bitmap file for the specified index.
 */
extern void fs_bitmapClear(
    fs_Context *context,
    fs_Inode   *bitmap,
    uint64     index,
    fs_BitmapOperation *debugOp);


/**
 * Turn on the bit in the specified bitmap file for the specified
 * location.
 */
extern void fs_bitmapSetLoc(
    fs_Context   *context,
    fs_Inode     *bitmap,
    fs_BitmapLoc loc,
    fs_BitmapOperation *debugOp);

/**
 * Turn off the bit in the specified bitmap file for the specified
 * location.
 */
extern void fs_bitmapClearLoc(
    fs_Context   *context,
    fs_Inode     *bitmap,
    fs_BitmapLoc loc,
    fs_BitmapOperation *debugOp);


/**
 * Clear the bit in the specified bitmap file for the specified index,
 * if it is set.  Returns true if the bit had been set.
 */
extern boolean fs_bitmapClearIfSet(
    fs_Context *context,
    fs_Inode   *bitmap,
    uint64     index,
    fs_BitmapOperation *debugOp);


/**
 * Set all bits in the bitmap from the specified index and after, to the
 * end of the page.
 */
extern void fs_bitmapSetBitsAfter(
    fs_Context *context,
    fs_Inode   *bitmap,
    uint64     index);


/**
 * Clear all bits in the bitmap from the specified index and after, to the
 * end of the page.
 */
extern void fs_bitmapClearBitsAfter(
    fs_Context *context,
    fs_Inode   *bitmap,
    uint64     index);


/**
 * Search for a bitmap index for which the bit is not set, near the
 * specified start index, in the specified bitmap file.
 */
extern uint64 fs_bitmapFindClear(
    fs_Context       *context,
    fs_Inode         *bitmap,
    uint64           startIndex,
    fs_BitmapLocList *excluded);


/**
 * Search for a bitmap index for which the bit is not set, near the
 * specified start index, in the union of the two specified bitmap files.
 */
extern uint64 fs_bitmapFindClear2(
    fs_Context       *context,
    fs_Inode         *bitmap1,
    fs_Inode         *bitmap2,
    uint64           startIndex,
    fs_BitmapLocList *excluded);

/**
 * Create an entry in a list of bitmap locations.
 */
extern fs_BitmapLocList *fs_makeBitmapLocList(
    fs_Context *context,
    fs_BitmapLocList *next,
    uint64 index);


#endif // FS_BITMAP_H
