// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=4 shiftwidth=4 expandtab:

/*
 * Name :  fs-buddy.c
 *
 * RCS:  $Id: fs-buddy.c,v 1.41 2002/11/12 19:28:44 janv Exp $
 *
 * Description:  A binary buddy scheme for allocation
 *
 * Created By:  Paul Haahr <haahr@agilestorage.com>
 *
 * Date Created:  2002-01-17
 */


#if !defined(SSC)

#include "fs-includes.h"

/*
 * The binary buddy allocator assigns each disk block to a contiguous
 * "region," which contains 32K blocks (based on an 8K block size).  Free
 * blocks and extents for each region are described on a single page (the
 * buddy region page) in a binary buddy scheme.
 *
 * All extents in the binary buddy scheme are power-of-two sized;  the
 * level of an extent is defined as the base 2 logarithm of its size.
 *
 * All extents which are larger than one block are divided into two
 * sub-extents.  These two sub-extents are "buddies," and the original
 * extent is their "parent."  One extent -- the "maximal extent" -- covers
 * the entire region and has no defined parent or buddy.
 *
 * The first bit in a region page is unused (it must be zero), which makes
 * the math on indexes work out better.  The second bit in a region page
 * describes the maximal extent -- a one indicates that it is free, and a
 * zero indicates that it is in use.  The next two bits describe the two
 * halves of the region, with a one bit indicating both that the region is
 * free and that the containing larger extent is not marked free.  This
 * continues all the way down to the individual block level.
 *
 * Note that in this representation, a region page with all bits off
 * describes a region with no free blocks.  This is convenient for
 * initialization.
 *
 * Given this layout, the index of the buddy of a given extent is the
 * index of the original extent with the low bit inverted (i.e., xor'ed
 * with one).  The parent of an extent can be found by shifting the index
 * once to the right.
 *
 * Each region page which is loaded in the file system is summarized by an
 * in-core count summary, which is owned by the buffer.  To modify the
 * summaries, you must hold the buffer's lock exclusively.
 *
 * The region pages should usually be acquired exclusive, and modified
 * with the FS_BUF_WRITE_LOG_BIT_SET or _RESET flags when calling
 * fs_bufWrite, to avoid two-phase locking problems.
 */

enum {
    /** Log2(maximum size of an extent in a region). */
    MAX_LEVELS = FS_MAX_BUDDY_LEVELS,
    /** Number of blocks in a region page. */
    BLOCKS_PER_REGION = FS_BLOCKS_PER_BUDDY_REGION,
    /** Log2(number of bits in a word) */
    WORD_SHIFT = 5,
};

// These constants are dependent on the block size.  That dependency
// should just be eliminated.  Instead, I just check it here.

NFX_ASSERT_STATIC(BLOCKS_PER_REGION == (BITS_PER_BYTE * FS_BLOCK_SIZE / 2));

// Also check that the assumption that the number of regions we'll see
// fits in a 32-bit quantity is safe.

NFX_ASSERT_STATIC(((1ULL << 50) / (FS_BLOCK_SIZE * BLOCKS_PER_REGION))
                  < (1ULL << 32));

// And ensure that our definition of word is still sensible

NFX_ASSERT_STATIC(BITS_PER_WORD == (1 << WORD_SHIFT));


/**
 * A summary of the number of free blocks in a region.
 */
struct fs_RegionSummary {
    /**
     * Counts for each size of extent, indexed by log2(size).
     *
     * TODO - these could be packed denser, at some cost for access
     */
    uint16 freeCount[MAX_LEVELS + 1];
};

/*---------------------------------------------------------------------------->  * Local functions prototype.
 *---------------------------------------------------------------------------->  */
 
static uint32 fs_buddyFindFreeExtentAtLevel(
    fs_Context *context,
    fs_Buf *buf,
    int level,
    uint32 targetOffset);

static uint32 fs_buddyFindFreeExtentAtLevelBytewise(
    fs_Context *context,
    fs_Buf *buf,
    int level,
    uint32 targetOffset);

static uint32 fs_buddyFindFreeExtentAtLevelBitwise(
    fs_Context *context,
    fs_Buf *buf,
    int level,
    uint32 targetOffset);

static void
fs_buddyGetSummary(fs_Context * context,
                   fs_Buf *     buf);

static boolean
fs_buddyValidateRegion(fs_Context * context,
                       uint32       region);

static int32
fs_buddyRecalculateRegion(fs_Context *  context,
                          fs_Buf *      buddyRegion,
                          uint32        region);


/**
 * Return the region number for a given block.
 */
inline uint32 fs_blockNumberBuddyRegion(
    fs_Context *context,
    uint64 blockno)
{
    return blockno >> MAX_LEVELS;
}


/**
 * Return the region number for a given block.
 */
inline uint32 fs_blockBuddyRegion(
    fs_Context *context,
    fs_Block block)
{
    return fs_blockNumberBuddyRegion(context, fs_blockNumber(block));
}


/**
 * Return the block number within a region for the specified block number.
 */
inline uint32 fs_blockNumberBuddyOffset(
    fs_Context *context,
    uint64 blockno)
{
    return blockno & ((1 << MAX_LEVELS) - 1);
}


/**
 * Return the block number within a region for a given block.
 */
inline uint32 fs_blockBuddyOffset(
    fs_Context *context,
    fs_Block block)
{
    return fs_blockNumberBuddyOffset(context, fs_blockNumber(block));
}


/**
 * Compute the index (in bits) within the buffer for the extent of the
 * given length, specified by its level (logarithm), starting at the
 * specified block.
 */
inline uint32 fs_buddyExtentIndex(
    fs_Context *context,
    int level,
    uint32 startBlock)
{
    assert(level <= MAX_LEVELS);
    assert((startBlock & ((1 << level) - 1)) == 0);
    return (1 << (MAX_LEVELS - level)) + (startBlock >> level);
}


/**
 * Compute the start block for the specified index of the specified
 * level.
 */
inline uint32 fs_buddyIndexToStartBlock(
    fs_Context *context,
    int level,
    uint32 index)
{
    uint32 base = (1 << (MAX_LEVELS - level));
    uint32 offset = index - base;

    assert(base <= index);
    assert(offset < base);

    return offset << level;
}


/**
 * Return true if the bit at the specified index within the buffer is set.
 *
 * TODO - move this (and the next two functions) to fs-buf.c or fs-bitmap.c
 */
inline boolean fs_bufBitIsSet(
    fs_Context *context,
    fs_Buf *buf,
    uint32 index)
{
    int word = index / BITS_PER_WORD;
    int mask = 1 << (index & (BITS_PER_WORD - 1));

    return (buf->u.word[word] & mask) != 0;
}


/**
 * Set the specified bit within the buffer and log the change.
 *
 * The release flag should be set to allow the buffer to be release
 * immediately after setting the bit.
 */
static void fs_bufBitSet(
    fs_Context *context,
    fs_Buf *buf,
    uint32 index,
    boolean release,
    fs_Extent extent)
{
    fs_BitmapOperation debugOp;
    int word = index / BITS_PER_WORD;
    int mask = 1 << (index & (BITS_PER_WORD - 1));
    union {
        uchar8 bytes[BITS_PER_WORD / BITS_PER_BYTE];
        uint32 mask;
    } bitmap;
#ifdef ISABELLE // under construction
    int level = MAX_LEVELS - log2_32(powerOfTwoFloor(index));
#endif // ISABELLE    

    assert((buf->u.word[word] & mask) == 0); // TODO - allow unconditional set?

    debugOp.u.extent = extent;
    debugOp.op = BITMAP_OP_DEALLOC;

    bitmap.mask = mask;
    buf->u.word[word] |= mask;
#ifndef BUDDY_INCORE
#ifdef FS_BUDDY_LOG
    fs_bufWrite(context,
                buf,
                word * sizeof (*buf->u.word),
                sizeof (*buf->u.word),
                FS_BUF_WRITE_LOG
                | FS_BUF_WRITE_LOG_BIT_SET
                | (release ? 0 : FS_BUF_LOG_ADD_BUF),
                bitmap.bytes, &debugOp);
#else
    if (release)
        FS_BUF_WRITE_NO_LOG(buf,
                            word * sizeof (*buf->u.word),
                            sizeof (*buf->u.word));
    else
        FS_BUF_WRITE_ADD_LOCK(buf,
                              word * sizeof (*buf->u.word),
                              sizeof (*buf->u.word));
#endif
#endif // BUDDY_INCORE
#ifdef ISABELLE // under construction
    if (level > 0)
    {
        fs_modBitsAtLevel0(context, buf, index, level, 1, release);
    }
#endif // ISABELLE
}


/**
 * Set the specified bit within the buffer and log the change.
 *
 * The release flag should be set to allow the buffer to be release
 * immediately after setting the bit.
 */
static void fs_bufBitClear(
    fs_Context *context,
    fs_Buf *buf,
    uint32 index,
    boolean release,
    fs_Extent extent)
{
    fs_BitmapOperation debugOp;
    uint32 word = index / BITS_PER_WORD;
    uint32 mask = 1 << (index & (BITS_PER_WORD - 1));
    union {
        uchar8 bytes[BITS_PER_WORD / BITS_PER_BYTE];
        uint32 mask;
    } bitmap;
#ifdef ISABELLE // under construction
    int level = MAX_LEVELS - log2_32(powerOfTwoFloor(index));
#endif // ISABELLE    

    assert((buf->u.word[word] & mask) != 0); // TODO - allow unconditional?

    debugOp.u.extent = extent;
    if (fs_extentLength(extent) != 0)
    {
        debugOp.op = BITMAP_OP_ALLOC;
    }
    else
    {
        debugOp.op = BITMAP_OP_MERGE;
    }
    bitmap.mask = mask;
    buf->u.word[word] &= ~mask;
#ifndef BUDDY_INCORE
#ifdef FS_BUDDY_LOG
    fs_bufWrite(context,
                buf,
                word * sizeof (*buf->u.word),
                sizeof (*buf->u.word),
                FS_BUF_WRITE_LOG
                | FS_BUF_WRITE_LOG_BIT_RESET
                | (release ? 0 : FS_BUF_LOG_ADD_BUF),
                bitmap.bytes, &debugOp);
#else
    if (release)
        FS_BUF_WRITE_NO_LOG(buf,
                            word * sizeof (*buf->u.word),
                            sizeof (*buf->u.word));
    else
        FS_BUF_WRITE_ADD_LOCK(buf,
                              word * sizeof (*buf->u.word),
                              sizeof (*buf->u.word));
#endif
#endif // BUDDY_INCORE
#ifdef ISABELLE // under construction
    if (level > 0)
    {
        fs_modBitsAtLevel0(context, buf, index, level, 0, release);
    }
#endif // ISABELLE
}

#ifdef ISABELLE // under construction
/**
 */
static void
fs_modBitsAtLevel0(fs_Context * context,
                   fs_Buf *     buf,
                   uint32       index,
                   int          level,
                   int          bit,    // 0 or 1
                   boolean      release)
    
{
    uint32  len = 1 << level;
    uint32  startIdx = (1 << MAX_LEVELS) +
        len *(index - (1 << log2_32(powerOfTwoFloor(index))));
    uint32  startWord = startIdx / BITS_PER_WORD;
    uint32  startBit = startIdx % BITS_PER_WORD;
    uint32  mask;
    uint32  wordIdx;

    FS_ASSERT(level > 0);
    printf("fs_modBitsAtLevel0(%d/%d)\n", startWord, len);

    if ((startBit + len) < BITS_PER_WORD)
    {
        mask = ((1 << len) - 1) << (BITS_PER_WORD - len - startBit);
        FS_ASSERT(((buf->u.word[startWord] & mask) == 0) == bit);
        buf->u.word[startWord] ^= mask;
        len = 0;
        printf("fs_modBitsAtLevel0(%d/0x%x)\n", startWord, mask);
    }
    else
    {
        mask = (1 << (BITS_PER_WORD - startBit)) - 1;
        FS_ASSERT(((buf->u.word[startWord] & mask) == 0) == bit);
        buf->u.word[startWord] ^= mask;
        len -= BITS_PER_WORD - startBit;
        printf("fs_modBitsAtLevel0(%d/0x%x)\n", startWord, mask);
    }

    wordIdx = startWord;
    while (len >= BITS_PER_WORD)
    {
        wordIdx++;
        mask = -1;
        FS_ASSERT(((buf->u.word[wordIdx] & mask) == 0) == bit);
        buf->u.word[wordIdx] ^= mask;
        len -= BITS_PER_WORD;
        printf("fs_modBitsAtLevel0(%d/0x%x)\n", wordIdx, mask);
    } // while

    FS_ASSERT((0 <= len) && (len < BITS_PER_WORD));
    if (len > 0)
    {
        wordIdx++;
        mask = ((1 << len) - 1) << (BITS_PER_WORD - len);
        FS_ASSERT(((buf->u.word[wordIdx] & mask) == 0) == bit);
        buf->u.word[wordIdx] ^= mask;
        printf("fs_modBitsAtLevel0(%d/0x%x)\n", wordIdx, mask);
    }

#ifndef BUDDY_INCORE
    if (release)
        FS_BUF_WRITE_NO_LOG(buf,
                            startWord * sizeof (*buf->u.word),
                            (wordIdx - startWord) * sizeof (*buf->u.word));
    else
        FS_BUF_WRITE_ADD_LOCK(buf,
                              startWord * sizeof (*buf->u.word),
                              (wordIdx - startWord) * sizeof (*buf->u.word));
#endif // BUDDY_INCORE
}
#endif // ISABELLE under construction


/**
 * Return the binary buddy region page for the specified region, locked
 * exclusive if the specified flag is true.
 */
#ifdef BUDDY_INCORE
fs_Buf *
fs_buddyGetRegionBuf(fs_Context *   context,
                     uint32         region,
                     boolean        exclusive)
{
    fs_Inode *  inode = I_GET_SHARED(0, FS_FREE_BUDDY_TREE_INODE);
    fs_Buf *    buf = NULL;
    
    FS_ASSERT(region < context->fileSystem->params.buddyRegionCount);
    
    if (context->fileSystem->buddies[region].summary == NULL)
    {
        fs_Buf *    buddyRegion = O_GET_EXCL(inode, region); // O_CREATE???

        //printf("fs_buddyGetRegionBuf(%d)\n", region);
        
        /* Recalculate the buddy region before returning it for use.
         */
        fs_buddyRecalculateRegion(context, buddyRegion, region);

        /* Calculate buddy summary.
         */
        fs_buddyGetSummary(context, buddyRegion);
        
        fs_bufReleaseExclusive(context, buddyRegion);
    }
    
    /* Get the buffer of the buddy region.
     */
    buf = (exclusive? O_GET_EXCL(inode, region) : O_GET_SHARED(inode, region));
    FS_ASSERT(buf != NULL);
        
    fs_inodeRelease(context, inode);
    return(buf);
}
#else // BUDDY_INCORE
static fs_Buf *
fs_buddyGetRegionBuf(fs_Context *   context,
                     uint32         region,
                     boolean        exclusive)
{
    boolean     done;
    fs_Inode *  inode = NULL;
    fs_Buf *    buf = NULL;

    /* Recalculate if necc the buddy region before returning it for use.
     */
    assert(region < context->fileSystem->params.buddyRegionCount);
    done = fs_buddyValidateRegion(context, region);

    /* Get the buffer of the buddy region.
     */
    inode = I_GET_SHARED(0, FS_FREE_BUDDY_TREE_INODE);
    buf = (exclusive? O_GET_EXCL(inode, region) : O_GET_SHARED(inode, region));
    fs_inodeRelease(context, inode);

    /* Calculate summary if not already done.
     */
    if (context->fileSystem->buddies[buf->page].summary == NULL)
    {
        fs_buddyGetSummary(context, buf);
    }
    
    return(buf);
}
#endif // BUDDY_INCORE


/**
 * Compute the region summary for the specified buf.
 * And indicates if the region is full or not.
 */
static fs_RegionSummary *
fs_buddyComputeSummary(fs_Context * context,
                       fs_Buf *     buf,
                       boolean *    buddyIsFull)

{
    fs_RegionSummary * summary = fs_ramAlloc(FS_RAM_REGION_SUMMARY,
                                             sizeof *summary);
    uint32  idx = (1 << (MAX_LEVELS - 2)) - 1; // last byte in buddy
    int     level;
    
    /* Assume that the buddy is fill, until proven otherwise.
     */
    *buddyIsFull = TRUE;

    /* Count bits byte per byte for levels 0 to 12.
     * at least 1 byte at level
     */
    for (level = 0; level <= (MAX_LEVELS - 3); level++)
    {
        int i, count = 0;
        
        for (i = 0; i < (1 << (MAX_LEVELS - 3 - level)); i++)
        {
            count += fs_onesPerByte[buf->u.bytes[idx]];
            idx--;
        } // for
        summary->freeCount[level] = count;
        if (summary->freeCount[level] > 0) {*buddyIsFull = FALSE;}
    } // for

    /* For the last 3 levels, the info is in a single byte.
     * Let's take 3 special cases.
     */
    summary->freeCount[MAX_LEVELS - 2] = fs_onesPerByte[buf->u.bytes[idx] & 0xf0];
    if (summary->freeCount[MAX_LEVELS - 2] > 0) {*buddyIsFull = FALSE;}
    
    summary->freeCount[MAX_LEVELS - 1] = fs_onesPerByte[buf->u.bytes[idx] & 0x0c];
    if (summary->freeCount[MAX_LEVELS - 1] > 0) {*buddyIsFull = FALSE;}

    summary->freeCount[MAX_LEVELS] = fs_onesPerByte[buf->u.bytes[idx] & 0x02];
    if (summary->freeCount[MAX_LEVELS] > 0) {*buddyIsFull = FALSE;}

    return(summary);
}


#ifdef FS_DEBUG_BUDDY_SUMMARY
/**
 * Check the summary on a buddy page.
 */
static void fs_buddyCheckSummary(
    fs_Context *context,
    fs_Buf *buf)
{
    int i;
    fs_RegionSummary *newSummary;
    fs_RegionSummary *oldSummary = buf->buddySummary;
    
    if (oldSummary == NULL)
        return;
    
    if (!fs_lockOwnedExclusive(context, &buf->lock)) {
        // can't reliably do this if someone else might change the buffer
        return;
    }

    newSummary = fs_buddyComputeSummary(context, buf);

    for (i = 0; i <= MAX_LEVELS; i++) {
        assert(newSummary->freeCount[i] == oldSummary->freeCount[i]);
    }
    
    fs_ramDealloc(FS_RAM_REGION_SUMMARY, newSummary);
}
#endif


/**
 * Return the region summary for a given region.
 * Update the buddies in filesystem if neeeded.
 */
static void
fs_buddyGetSummary(fs_Context * context,
                   fs_Buf *     buf)
    
{
    fs_buddy_t *    buddy = &(context->fileSystem->buddies[buf->page]);
    boolean         buddyIsFull = TRUE;

    if (buddy->summary == NULL)
    {
        NFX_SPIN_LOCK(&(buddy->spin));
        if (buddy->summary == NULL)
        {
            buddy->summary = fs_buddyComputeSummary(context, buf,
                                                    &buddyIsFull);
            if (buddyIsFull)
            {
                buddy->flags |= FS_BUDDY_FULL;
            }
            else
            {
                buddy->flags &= ~FS_BUDDY_FULL;
            }
        }
        NFX_SPIN_UNLOCK(&(buddy->spin));
    }
    return;
}


/**
 * Return TRUE if the region is full, FALSE otherwise.
 */
static boolean
fs_buddyIsFull(fs_RegionSummary * summary)
{
    boolean buddyIsFull = TRUE;
    int     level;

    for (level = 0; level <= MAX_LEVELS; level++)
    {
        if (summary->freeCount[level] > 0)
        {
            buddyIsFull = FALSE;
            break;
        }
    } // for

    return(buddyIsFull);
}

/**
 * Decompose a (len, block-with-region) pair over the binary buddy extents
 * in the specified region, passing the decomposed extents to the specified
 * function.
 */
static void fs_buddyDecomposeExtentInRegion(
    fs_Context *context,
    fs_Buf *buf,
    boolean release,
    uint32 region,
    uint32 length,
    uint32 startBlock,
    void (*func)(
        fs_Context *context,
        fs_Buf *buf,
        boolean release,
        uint16 length,
        uint32 startBlock,
        void *arg),
    void *arg)
{
    int remaining = length;
    int nextStart = startBlock;

    assert(length > 0);
    assert(buf->page == region);

    if (startBlock > 0) {
        int nextLen = leastBit(startBlock);
        while (nextLen < remaining) {
            assert(nextLen > 0);
            func(context,
                 buf,
                 release,
                 nextLen,
                 nextStart,
                 arg);
            remaining -= nextLen;
            nextStart += nextLen;
            nextLen = leastBit(nextStart);
        }
    }

    while (remaining > 0) {
        int nextLen = powerOfTwoFloor(remaining);
        assert(nextLen > 0);
        func(context,
             buf,
             release,
             nextLen,
             nextStart,
             arg);
        remaining -= nextLen;
        nextStart += nextLen;
    }

    assert(remaining == 0);
}


/**
 * Decompose a set of blocks to the regions that contain it and the
 * bitlevel and start bits for each section.
 */
static void fs_buddyDecomposeBlocks(
    fs_Context *context,
    boolean release,
    uint64 length,
    fs_Block firstBlock,
    void (*func)(
        fs_Context *context,
        fs_Buf *buf,
        boolean release,
        uint16 length,
        uint32 startBlock,
        void *arg),
    void *arg)
{
    uint64 len = length;
    uint64 blockno = fs_blockNumber(firstBlock);

    while (len > 0) {
        uint32 region = fs_blockNumberBuddyRegion(context, blockno);
        uint32 startBlock = fs_blockNumberBuddyOffset(context, blockno);
        uint32 blocks = MIN(len, BLOCKS_PER_REGION - startBlock);

        fs_Buf *buf = fs_buddyGetRegionBuf(context, region, TRUE);

        fs_buddyDecomposeExtentInRegion(
            context, buf, release, region, blocks, startBlock, func, arg);
        //fs_buddyCheckSummary(context, buf);

        if (release || !fs_bufWasModified(context, buf))
            fs_bufReleaseExclusive(context, buf);
        else
            fs_bufDirty(context, buf);

        len -= blocks;
        blockno += blocks;
    }

    assert(len == 0);
    assert(blockno == fs_blockNumber(firstBlock) + length);
}


/**
 * Decompose an extent to the regions that contain it and the bitlevel and
 * start bits for each section.
 *
 * When calls are made to the inner function, the buffer is locked
 * exclusively;  if the release flag is true, the buffers are released
 * before this function returns.  Otherwise they should be held until
 * commit time.
 *
 * TODO - exclusive flag?
 */
static void fs_buddyDecomposeExtent(
    fs_Context *context,
    boolean release,
    fs_Extent extent,
    void (*func)(
        fs_Context *context,
        fs_Buf *buf,
        boolean release,
        uint16 length,
        uint32 startBlock,
        void *arg),
    void *arg)
{
    fs_buddyDecomposeBlocks(context,
                            release,
                            fs_extentLength(extent),
                            fs_extentStartBlock(extent),
                            func,
                            arg);
}


/**
 * Add a single binary buddy extent to the tree.
 */
static void fs_buddyAddExtentWorker(
    fs_Context *context,
    fs_Buf *buf,
    boolean release,
    uint16 length,
    uint32 startBlock,
    void *arg)
{
    fs_buddy_t *        buddy = &(context->fileSystem->buddies[buf->page]);
    int                 level = log2_32(length);
    uint32              index = fs_buddyExtentIndex(context, level, startBlock);
    fs_RegionSummary *  summary = NULL;
    uint16 *            count = NULL;
    fs_Extent           extent;

    assert(level <= MAX_LEVELS);
    assert(!fs_bufBitIsSet(context, buf, index));
    assert(length != 0);

    fs_buddyGetSummary(context, buf);
    summary = buddy->summary;
    count = summary->freeCount;

    extent = fs_makeExtent(length, fs_makeBlock(startBlock + (buf->page *
                                                FS_BLOCKS_PER_BUDDY_REGION)));
    NFX_SPIN_LOCK(&(buddy->spin));
    while (level < MAX_LEVELS)
    {
        uint32 buddy = index ^ 1;
        if (!fs_bufBitIsSet(context, buf, buddy))
            break;

        // merge and promote
        fs_bufBitClear(context, buf, buddy, release,
                        fs_makeExtent(0,
                             fs_makeBlock(fs_buddyIndexToStartBlock(context,
                                                               level, buddy) +
                                  (buf->page * FS_BLOCKS_PER_BUDDY_REGION))));
        assert(count[level] > 0);
        count[level] -= 1;
        level += 1;
        index >>= 1;
        assert(!fs_bufBitIsSet(context, buf, index));
    } // while

    fs_bufBitSet(context, buf, index, release, extent);
    count[level] += 1;

    assert(count[level] < (1 << (MAX_LEVELS - level))
           || (count[level] == 1 && level == MAX_LEVELS));

    /* The buddy summary has been modified, update the buddyFull info.
     */
    // buddy->flags &= ~FS_BUDDY_FULL;
    if (fs_buddyIsFull(summary))
    {
        FS_ASSERT(FALSE);
        buddy->flags |= FS_BUDDY_FULL;
    }
    else
    {
        buddy->flags &= ~FS_BUDDY_FULL;
    }

    NFX_SPIN_UNLOCK(&(buddy->spin));
    return;
}


/**
 * Recalculate's version of fs_buddyAddExtentWorker:
 *   Add a single binary buddy extent to the tree.
 *   The difference is that we don't update the summary when
 *   this region is being recalculated.
 */
static void fs_buddyRecalAddExtentWorker(
    fs_Context *context,
    fs_Buf *buf,
    boolean release,
    uint16 length,
    uint32 startBlock,
    void *arg)
{
    int level = log2_32(length);
    uint32 index = fs_buddyExtentIndex(context, level, startBlock);
    fs_Extent extent;

    assert(level <= MAX_LEVELS);
    assert(!fs_bufBitIsSet(context, buf, index));
    assert(length != 0);

    extent = fs_makeExtent(length, fs_makeBlock(startBlock + (buf->page *
                                                FS_BLOCKS_PER_BUDDY_REGION)));
    while (level < MAX_LEVELS) {
        uint32 buddy = index ^ 1;
        if (!fs_bufBitIsSet(context, buf, buddy))
            break;

        // merge and promote
        fs_bufBitClear(context, buf, buddy, release,
            fs_makeExtent(0, fs_makeBlock(fs_buddyIndexToStartBlock(context, level, buddy) +
                                          (buf->page * FS_BLOCKS_PER_BUDDY_REGION))));
        level += 1;
        index >>= 1;
        assert(!fs_bufBitIsSet(context, buf, index));
    }

    fs_bufBitSet(context, buf, index, release, extent);
}


/**
 * Decompose an extent down to its one-block descendents.  If any
 * descendents are already free, adds the rest to the tree and returns
 * true.  Otherwise, returns false.
 */
static boolean fs_buddyMaybeAddExtentChildren(
    fs_Context *context,
    fs_Buf *buf,
    boolean release,
    int level,
    uint32 index,
    void *arg)
{
    
    if (level == 0)
        return fs_bufBitIsSet(context, buf, index);

    else {
        int next = level - 1;
        uint32 leftIndex = (index << 1);
        boolean left
            = fs_buddyMaybeAddExtentChildren(
                context, buf, release, next, leftIndex, arg);
        boolean right
            = fs_buddyMaybeAddExtentChildren(
                context, buf, release, next, leftIndex + 1, arg);

        if (left) {
            if (!right) {
                fs_buddyAddExtentWorker(
                    context, buf, release, 1 << next,
                    fs_buddyIndexToStartBlock(context, next, leftIndex + 1),
                    arg);
            }
            return TRUE;
        }

        if (right) {
            assert(!left);
            fs_buddyAddExtentWorker(
                    context, buf, release, 1 << next,
                    fs_buddyIndexToStartBlock(context, next, leftIndex),
                    arg);
            return TRUE;
        }

        return FALSE;
    }
}


/**
 * Recalculate's version of fs_buddyMaybeAddExtentChildren():
 *   Call buddyRecalAddExtentWorker instead so summary will not be updated
 *   while this region is being validated.
 */
static boolean fs_buddyRecalAddExtentChildren(
    fs_Context *context,
    fs_Buf *buf,
    boolean release,
    int level,
    uint32 index,
    void *arg)
{

    if (level == 0)
        return fs_bufBitIsSet(context, buf, index);

    else {
        int next = level - 1;
        uint32 leftIndex = (index << 1);
        boolean left
            = fs_buddyRecalAddExtentChildren(
                context, buf, release, next, leftIndex, arg);
        boolean right
            = fs_buddyRecalAddExtentChildren(
                context, buf, release, next, leftIndex + 1, arg);

        if (left) {
            if (!right) {
                fs_buddyRecalAddExtentWorker(
                    context, buf, release, 1 << next,
                    fs_buddyIndexToStartBlock(context, next, leftIndex + 1),
                    arg);
            }
            return TRUE;
        }

        if (right) {
            assert(!left);
            fs_buddyRecalAddExtentWorker(
                    context, buf, release, 1 << next,
                    fs_buddyIndexToStartBlock(context, next, leftIndex),
                    arg);
            return TRUE;
        }

        return FALSE;
    }
}


/**
 * Add a single binary buddy extent to the tree, if it's not already in
 * there.
 */
static void fs_buddyMaybeAddExtentWorker(
    fs_Context *context,
    fs_Buf *buf,
    boolean release,
    uint16 length,
    uint32 startBlock,
    void *arg)
{
    int level = log2_32(length);
    uint32 index = fs_buddyExtentIndex(context, level, startBlock);
    int i;

    // check if this or any ancestor extent is free

    for (i = 0; (level + i) <= MAX_LEVELS; i++) {
        if (fs_bufBitIsSet(context, buf, index >> i))
            return;
    }

    // check descendents

    if (!fs_buddyMaybeAddExtentChildren(
            context, buf, release, level, index, arg))
    {
        fs_buddyAddExtentWorker(
            context, buf, release, length, startBlock, arg);
    }
}


/**
 * Recal's version of fs_buddyMaybeAddExtentWorker():
 *   Call fs_buddyRecalAddExtentWorker instead
 */
static void fs_buddyRecalAddExtent(
    fs_Context *context,
    fs_Buf *buf,
    boolean release,
    uint16 length,
    uint32 startBlock,
    void *arg)
{
    int level = log2_32(length);
    uint32 index = fs_buddyExtentIndex(context, level, startBlock);
    int i;

    // check if this or any ancestor extent is free

    for (i = 0; (level + i) <= MAX_LEVELS; i++) {
        if (fs_bufBitIsSet(context, buf, index >> i))
            return;
    }

    // check descendents

    if (!fs_buddyRecalAddExtentChildren(
            context, buf, release, level, index, arg))
    {
        fs_buddyRecalAddExtentWorker(
            context, buf, release, length, startBlock, arg);
    }
}


/**
 * Add an extent to the binary buddy tree.  The binary buddy region buffers
 * are released after the operation if the release flag is true;  otherwise,
 * they will be held until the transaction commits.
 */
void fs_buddyAddExtent(
    fs_Context *context,
    fs_Extent extent,
    boolean release)
{
    fs_buddyDecomposeExtent(
        context,
        release,
        extent,
        fs_buddyAddExtentWorker,
        NULL);
}


/**
 * Add an extent to the binary buddy tree, if it's not already in there.
 *
 * This is similar to fs_buddyAddExtent, except that it's robust in the
 * presence of redundant adds;  that robustness costs performance, so
 * this routine takes time proportional to the size of the extent.
 */
void fs_buddyMaybeAddExtent(
   fs_Context *context,
   fs_Extent extent,
   boolean release)
{
    fs_buddyDecomposeExtent(
        context,
        release,
        extent,
        fs_buddyMaybeAddExtentWorker,
        NULL);
    FS_DBG_SHADOW_FREE_EXTENT(extent);
}


/**
 * Return a level (i.e., log2(size)) within the region summary to
 * allocate from, given the specified target level.  Returns -1 if
 * no level has an appropriate extent.
 *
 * The heuristic here is that it's mildly better to take a big extent and
 * chop it up than to take a smaller extent and only return that.  This
 * loop lookup both up and down in level, in a particular order, for that
 * reason.
 */
static int fs_buddyFindFreeLevelInSummary(
    fs_Context *context,
    fs_RegionSummary *summary,
    int level)
{
    int i;
    int limit = MAX(level, MAX_LEVELS - level);

    for (i = 0; i <= limit; i++) {
        // check the target and above
        if (level + i <= MAX_LEVELS) {
            if (summary->freeCount[level + i] > 0)
                return level + i;
        }

        // check below
        if (i < level) {
            if (summary->freeCount[level - i - 1] > 0)
                return level - i - 1;
        }
    }

    return -1;
}


/**
 * Find and extract an extent of the specified level.  Returns the
 * offset within the region of the start of the extent.
 *
 * Since this function should only be called if the summary says there
 * is a free extent of this size, fails if no extent is found.
 */
static uint32 fs_buddyFindFreeExtentAtLevel(
    fs_Context *context,
    fs_Buf *buf,
    int level,
    uint32 targetOffset)
{
    if (level > (MAX_LEVELS - 6)) // less 2 words at this level
    {
        return(fs_buddyFindFreeExtentAtLevelBytewise(context, buf,
                                                     level, targetOffset));
    }
    else // at least 2 words at this level
    {
        /* Let's look word wise for free extents.
         */
        uint32 base = (1 << (MAX_LEVELS - level)) / BITS_PER_WORD;
        uint32 targetIndex = (targetOffset >> level) / BITS_PER_WORD;
        uint32 limit = MAX(targetIndex, base - targetIndex);
        uint32 i, newTargetOffset;
    
        for (i = 0; i < limit; i++)
        {
            // target and after
            if (targetIndex + i < base)
            {
                int index = base + targetIndex + i;
                if (buf->u.word[index] != 0)
                {
                    if (i == 0)
                    {
                        newTargetOffset = targetOffset;
                    }
                    else
                    {
                        newTargetOffset = (((targetIndex + i) * BITS_PER_WORD) +
                                           (BITS_PER_WORD / 2)) << level;
                    }
                    return(fs_buddyFindFreeExtentAtLevelBytewise(context, buf,
                                                                 level,
                                                                 newTargetOffset));
                }
            }
            
            // before target
            if (i < targetIndex)
            {
                int index = base + targetIndex - i - 1;
                if (buf->u.word[index] != 0)
                {
                    newTargetOffset = (((targetIndex - i - 1) * BITS_PER_WORD) +
                                       (BITS_PER_WORD / 2)) << level;
                    return(fs_buddyFindFreeExtentAtLevelBytewise(context, buf,
                                                                 level,
                                                                 newTargetOffset));
                }
            }
        } // for
    }

    fs_error(context, FS_INTERNAL_ERR, "no free extent despite summary");
    UNREACHED();
}

static uint32 fs_buddyFindFreeExtentAtLevelBytewise(
    fs_Context *context,
    fs_Buf *buf,
    int level,
    uint32 targetOffset)
{
    if (level > (MAX_LEVELS - 4)) // less than 2 bytes at this level
    {
        return(fs_buddyFindFreeExtentAtLevelBitwise(context, buf,
                                                    level, targetOffset));
    }
    else // at least 2 bytes at this level
    {
        /* Let's look byte wise for free extents.
         */
        uint32 base = (1 << (MAX_LEVELS - level)) / BITS_PER_BYTE;
        uint32 targetIndex = (targetOffset >> level) / BITS_PER_BYTE;
        uint32 limit = MAX(targetIndex, base - targetIndex);
        uint32 i, newTargetOffset;
    
        for (i = 0; i < limit; i++)
        {
            // target and after
            if (targetIndex + i < base)
            {
                int index = base + targetIndex + i;
                if (buf->u.bytes[index] != 0)
                {
                    if (i == 0)
                    {
                        newTargetOffset = targetOffset;
                    }
                    else
                    {
                        newTargetOffset = (((targetIndex + i) * BITS_PER_BYTE) +
                                           (BITS_PER_BYTE / 2)) << level;
                    }
                    return(fs_buddyFindFreeExtentAtLevelBitwise(context, buf, level,
                                                                newTargetOffset));
                }
            }
            
            // before target
            if (i < targetIndex)
            {
                int index = base + targetIndex - i - 1;
                if (buf->u.bytes[index] != 0)
                {
                    newTargetOffset = (((targetIndex - i - 1) * BITS_PER_BYTE) +
                                       (BITS_PER_BYTE / 2)) << level;
                    return(fs_buddyFindFreeExtentAtLevelBitwise(context, buf, level,
                                                                newTargetOffset));
                }
            }
        } // for
    }

    fs_error(context, FS_INTERNAL_ERR, "no free extent despite summary");
    UNREACHED();
}


static uint32 fs_buddyFindFreeExtentAtLevelBitwise(
    fs_Context *context,
    fs_Buf *buf,
    int level,
    uint32 targetOffset)
{
    fs_Extent extent;
    uint32 base = 1 << (MAX_LEVELS - level);
    uint32 targetIndex = targetOffset >> level;
    uint32 limit = MAX(targetIndex, base - targetIndex);
    uint32 i, offset;

    for (i = 0; i < limit; i++)
    {
        // target and after
        if (targetIndex + i < base) {
            int index = base + targetIndex + i;
            if (fs_bufBitIsSet(context, buf, index)) {
                offset = (index - base) << level;
                extent = fs_makeExtent(1 << level, fs_makeBlock(
                              (buf->page * BLOCKS_PER_REGION) + offset));
                fs_bufBitClear(context, buf, index, TRUE, extent);
                return offset;
            }
        }

        // before target
        if (i < targetIndex) {
            int index = base + targetIndex - i - 1;
            if (fs_bufBitIsSet(context, buf, index)) {
                offset = (index - base) << level;
                extent = fs_makeExtent(1 << level, fs_makeBlock(
                              (buf->page * BLOCKS_PER_REGION) + offset));
                fs_bufBitClear(context, buf, index, TRUE, extent);
                return offset;
            }
        }
    }

    fs_error(context, FS_INTERNAL_ERR, "no free extent despite summary");
    UNREACHED();
}


/**
 * Find an extent in the specified region.  If possible, we look for an
 * extent of a similar size and similar offset within the block, but we'll
 * return anything if we miss those.  Removes the extent if one was found.
 *
 * (Perhaps we should try to stay close to the size on a first pass, and
 * do a second pass over all regions if that fails?)
 *
 * Returns a zero length extent if no such extent was found in the region.
 */
static fs_Extent
fs_buddyFindExtentInRegion(fs_Context * context,
                           uint32       region,
                           uint16       targetLevel,
                           uint32       targetOffset,
                           uint16       len)
    
{
    fs_buddy_t *        buddy = &(context->fileSystem->buddies[region]);
    fs_Buf *            buf = NULL;
    fs_RegionSummary *  summary = NULL;
    int                 level;
    boolean             targetIsFree = FALSE;
    uint32              offset;

    /* Check if there is at least one free block into the buddy region.
     */
    if ((buddy->summary != NULL) && (buddy->flags & FS_BUDDY_FULL))
    {
        return(fs_makeExtent(0, fs_makeBlock(0)));
    }
    else
    {
        buf = fs_buddyGetRegionBuf(context, region, TRUE);
        summary = buddy->summary;
    }
        
    /* Check if the block at the targetOffset is free.
     * Return with the level where to find it.
     */
    for (level = MIN(log2_32(powerOfTwoCeiling(len)), MAX_LEVELS);
         level <= MAX_LEVELS; level++)
    {
        uint32 index = (1 << (MAX_LEVELS - level)) + (targetOffset >> level);
        if (fs_bufBitIsSet(context, buf, index))
        {
            targetIsFree = TRUE;
            break;
        }
    } // for

    if (targetIsFree)
    {
        /* Get the 1st block (= offset) of the level,
         * and reserve the whole level.
         * Assert that the target block is in the level.
         * The level may be too small for all the blocks we need,
         * so reculate the length.
         */
        offset = fs_buddyFindFreeExtentAtLevel(context, buf, level,
                                               targetOffset);
        FS_ASSERT(offset == ((targetOffset >> level) << level));
        if (((targetOffset + len) > (offset + (1 << level))) && (len != 1))
        {
            len = (offset + (1 << level)) - targetOffset;
        }
        offset = targetOffset;
    }
    else
    {
        /* The target is not free, find a free level the closest to the
         * target level.
         * Reserve the whole level, and return the 1st block (= offset).
         */
        level = fs_buddyFindFreeLevelInSummary(context, summary, targetLevel);
        if (level < 0)
        {
            fs_bufReleaseExclusive(context, buf);
            return fs_makeExtent(0, fs_makeBlock(0));
        }
        offset = fs_buddyFindFreeExtentAtLevel(context, buf, level,
                                               targetOffset);
    }

    {
        // At this stage, we're guaranteed to find an extent, assuming
        // the summary is correct.
        
        uint64 page = buf->page;
        uint16 usedLen;

        assert(summary->freeCount[level] > 0);
        NFX_SPIN_LOCK(&(buddy->spin));
        summary->freeCount[level] -= 1;
        if (fs_buddyIsFull(summary))
        {
            buddy->flags |= FS_BUDDY_FULL;
        }
        else
        {
            buddy->flags &= ~FS_BUDDY_FULL;
        }
        NFX_SPIN_UNLOCK(&(buddy->spin));
        usedLen = MIN(len, 1 << level);
            
        // Return any excess here before unlocking the buf.
        if ((1 << level) > len)
        {
            int base;
            base = (offset >> level) << level;

            if (offset == base)
            {
                /* The extent starts at the beginning of the level,
                 * return the rigth side of the level.
                 */
                fs_buddyDecomposeExtentInRegion
                    (context, buf, TRUE, region,
                     (1 << level) - usedLen,
                     offset + usedLen,
                     fs_buddyAddExtentWorker, NULL);
            }
            else
            {
                /* The extent doesn't start at the beginning of the level,
                 * return the left and rigth side of the level.
                 */
                
                int left = offset - base;
                FS_ASSERT(left > 0);
                
                if (left != 0)
                {
                    fs_buddyDecomposeExtentInRegion
                        (context, buf, TRUE, region,
                         left,
                         base,
                         fs_buddyAddExtentWorker, NULL);
                }
                if (((1 << level) - usedLen - left) != 0)
                {
                    fs_buddyDecomposeExtentInRegion
                        (context, buf, TRUE, region,
                         (1 << level) - usedLen - left,
                         offset + usedLen,
                         fs_buddyAddExtentWorker, NULL);
                }
            }
        }
        fs_bufReleaseExclusive(context, buf);
        
        return(fs_makeExtent(usedLen,
                             fs_makeBlock((page *
                                           BLOCKS_PER_REGION) + offset)));
    }
}


/**
 * Find an extent in the binary buddy tree near the specified block,
 * and remove the extent from the tree.  The system will try to find
 * an extent at least as long as the requested amount, but it may return
 * something shorter.
 *
 * Returns a zero length extent if no such extent was found.
 */
fs_Extent
fs_buddyFindExtent(fs_Context * context,
                   uint16       len,
                   fs_Block     target,
                   boolean  isFirstFileExtent)
    
{
    fs_Extent   extent;
    uint32      region = fs_blockBuddyRegion(context, target);
    uint16      level = MIN(log2_32(powerOfTwoCeiling(len)), MAX_LEVELS);

    assert(len > 0);

#ifdef NEW_ALLOC_POLICY
    /* There is nothing magic about 4, could be something different.
     * Should be about the same size as the average file size.
     */
    if (isFirstFileExtent == TRUE)
        level = 4;  // ~130kbytes
#endif // NEW_ALLOC_POLICY
    
    extent = fs_buddyFindExtentInRegion(context,
                                        region,
                                        level,
                                        fs_blockBuddyOffset(context, target), 
                                        len);
    
#if defined(FS_DEBUG_BUDDY_CHECK_REFBYTE)
    {
        int i;
        for (i = 0; i < fs_extentLength(extent); i++) {
            fs_RefByte byte
                 = fs_refGetByte(context, fs_extentBlock(extent, i));
            if (byte != 0) {
                fs_error(
                    context, FS_INTERNAL_ERR, 
                    "allocating block %lld (%d of extent) with refByte %x",
                    fs_blockNumber(fs_extentBlock(extent, i)), i, byte);
            }
        }
    }
#endif
    
    if (fs_extentLength(extent) > 0)
    {
        return(extent);
    }
    else if (context->fileSystem->params.buddyRegionCount > 1)
    {
        /* The target region is full and so are probably the regions around it.
         * Instead of looking around the target region for another free region,
         * let's take randomly a region (!= target region) and start the
         * search from there.
         */
        uint32 i;
        uint32 count = context->fileSystem->params.buddyRegionCount;
        uint32 limit;
#ifdef NEW_ALLOC_POLICY
        uint32 newRegion = region;
#else // NEW_ALLOC_POLICY
        uint32 newRegion = (uint32) fs_random64(context, count);

        while (newRegion == region)
        {
            newRegion = (uint32) fs_random64(context, count);
        }
#endif // NEW_ALLOC_POLICY
        
        limit = MAX(newRegion, count - newRegion);
        
        // walk forwards and backwards in regions
#ifdef NEW_ALLOC_POLICY
        for (i = 1; i < limit; i++)
        {
            // forwards
            if (newRegion + i < count)
            {
                extent = fs_buddyFindExtentInRegion(context,
                                                    newRegion + i,
                                                    level,
                                                    0,
                                                    len);
                if (fs_extentLength(extent) > 0)
                    return extent;
            }

            // backwards
            if (i <= newRegion)
            {
                extent = fs_buddyFindExtentInRegion(context,
                                                    newRegion - i,
                                                    level,
                                                    BLOCKS_PER_REGION - 1,
                                                    len);
                if (fs_extentLength(extent) > 0)
                    return extent;
            }
        } // for
#else // NEW_ALLOC_POLICY
        for (i = 0; i < limit; i++)
        {
            // forwards
            if ((newRegion != region) && (newRegion + i < count))
            {
                extent = fs_buddyFindExtentInRegion(context,
                                                    newRegion + i,
                                                    level,
                                                    0,
                                                    len);
                if (fs_extentLength(extent) > 0)
                    return extent;
            }

            // backwards
            if ((newRegion != region) && (i <= newRegion))
            {
                extent = fs_buddyFindExtentInRegion(context,
                                                    newRegion - i,
                                                    level,
                                                    BLOCKS_PER_REGION - 1,
                                                    len);
                if (fs_extentLength(extent) > 0)
                    return extent;
            }
        } // for
#endif // NEW_ALLOC_POLICY
        FS_ELOG(context, LOG_ERR, FS_TRACE_REQ,
                ("fs_buddyFindExtent: no space, "
                 "level %d region %d count %d limit %d newRegion %d",
                 level, region, count, limit, newRegion));
        FS_ASSERT(FALSE);
        // fs_volumeException()
        return(fs_makeExtent(0, fs_makeBlock(0)));
    }
    else
    {
        FS_ELOG(context, LOG_ERR, FS_TRACE_REQ,
                ("fs_buddyFindExtent: no space, level %d region %d",
                 level, region));
        FS_ASSERT(FALSE);
        // fs_volumeException()
        return(fs_makeExtent(0, fs_makeBlock(0)));
    }
}


/**
 * Get bit for buddy region in bitmap
 */
inline uint32 fs_buddyGetRegionBit(
        fs_Context *context,
        uint32 region)
{
    uint32 bit;

    NFX_SPIN_LOCK(&context->fileSystem->buddyBitmapSpin);
    assert(region < context->fileSystem->buddyRegions);
    bit = (context->fileSystem->buddyBitmap[region >> 3] >> (region & 0x7)) & 0x1;
    NFX_SPIN_UNLOCK(&context->fileSystem->buddyBitmapSpin);
    return bit;
}


/**
 * Set bit for buddy region in bitmap
 */
inline uint32 fs_buddySetRegionBit(
        fs_Context *context,
        uint32 region)
{
    uint8 bit = 1 << (region & 0x7);
    uint8 *map;
    int32 oldBit;

    NFX_SPIN_LOCK(&context->fileSystem->buddyBitmapSpin);
    assert(region < context->fileSystem->buddyRegions);
    map = &context->fileSystem->buddyBitmap[region >> 3];
    oldBit = ((*map & bit) != 0);
    *map |= bit;
    NFX_SPIN_UNLOCK(&context->fileSystem->buddyBitmapSpin);
    return oldBit;
}


/**
 * Initialize the buddy region bitmap to have nRegions
 * If setValid, set the entire bitmap to have valid region bits
 * Otherwise, invalidate all region bits
 */
void fs_buddyInitBitmap(
        fs_Context *context,
        fs_FileSystem *fs,
        uint32 nRegions,
        boolean setValid)
{

    NFX_SPIN_LOCK(&fs->buddyBitmapSpin);

    // resize buddy region bitmap if necc
    if ((fs->buddyRegions != nRegions) && fs->buddyBitmap) {
        FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
                ("buddy init bitmap: redo with new max regions %d old %d",
                 nRegions, fs->buddyRegions));
        fs_ramDealloc(FS_RAM_CORE_MISC, fs->buddyBitmap);
    }
    fs->buddyRegions = nRegions;
    fs->buddyBitmap = fs_ramAlloc(FS_RAM_CORE_MISC,
                                  ceilingDiv64(fs->buddyRegions, BITS_PER_BYTE));
    if (setValid) {
        // Set buddy region bitmap valid
        memset(fs->buddyBitmap, 0xff,
                ceilingDiv64(fs->buddyRegions, BITS_PER_BYTE));
        FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
                ("buddy init bitmap: bitmap valid nRegions %d",
                 fs->buddyRegions));
    } else {
        // Invalidate buddy region bitmap and start background thread to recalculate
        bzero(fs->buddyBitmap, ceilingDiv64(fs->buddyRegions, BITS_PER_BYTE));
        FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
                ("buddy init bitmap: bitmap invalid nRegions %d",
                 fs->buddyRegions));
    }

    NFX_SPIN_UNLOCK(&fs->buddyBitmapSpin);
    return;
}


/**
 * Recalculate's version of fs_buddyDecomposeBlocks
 * The difference is a O_GET_EXCL is caled since the region does not need
 * to be validated while it is being validated (otherwise infinite recursion !!)
 */
static void fs_buddyRecalAddBlocks(
    fs_Context *context,
    uint64 length,
    fs_Block firstBlock,
    void (*func)(
        fs_Context *context,
        fs_Buf *buf,
        boolean release,
        uint16 length,
        uint32 startBlock,
        void *arg),
    void *arg)
{
    uint64 len = length;
    uint64 blockno = fs_blockNumber(firstBlock);
    fs_Inode *buddy = I_GET_SHARED(0, FS_FREE_BUDDY_TREE_INODE);

    while (len > 0) {
        uint32 region = fs_blockNumberBuddyRegion(context, blockno);
        uint32 startBlock = fs_blockNumberBuddyOffset(context, blockno);
        uint32 blocks = MIN(len, BLOCKS_PER_REGION - startBlock);
        fs_Buf *buf = O_GET_EXCL(buddy, region); // buddy region will be in cache

        fs_buddyDecomposeExtentInRegion(
            context, buf, TRUE, region, blocks, startBlock, func, arg);

        if (!fs_bufWasModified(context, buf))
            fs_bufReleaseExclusive(context, buf);
        else
            fs_bufDirty(context, buf);

        len -= blocks;
        blockno += blocks;
    }

    assert(len == 0);
    assert(blockno == fs_blockNumber(firstBlock) + length);
    fs_inodeRelease(context, buddy);

    //FS_DBG_SHADOW_FREE_EXTENT(extent);
}


/**
 * Add the blocks between the first specified block (inclusive) and the
 * final specified block (exclusive) to the binary buddy tree.
 * Call recalculate's version to add blocks without validating region.
 */
void fs_buddyAddBlocks(
    fs_Context *context,
    fs_Block first,
    fs_Block last,
    boolean release)
{
    fs_buddyRecalAddBlocks(
        context,
        fs_blockNumber(last) - fs_blockNumber(first),
        first,
        fs_buddyAddExtentWorker,
        NULL);
}


/**
 * Recalculate the n'th region of buddy
 */
static int32
fs_buddyRecalculateRegion(fs_Context *  context,
                          fs_Buf *      buddyRegion,
                          uint32        region)
    
{
    uint64 startBlock, endBlock;
    uint64 startRefPage, pageno;
    uint64 b, firstBlock, lastBlock, maxBlock;
    uint64 used = 0;
    uint32 nRefPages = (FS_BLOCKS_PER_BUDDY_REGION / FS_BLOCK_SIZE);
    fs_Parameters * params = &context->fileSystem->params;
    fs_Inode *      refInode = I_GET_SHARED(0, FS_REF_COUNT_INODE);

    assert(fs_lockOwnedExclusive(context, &buddyRegion->lock));

    startBlock = region * FS_BLOCKS_PER_BUDDY_REGION;
    endBlock = startBlock + FS_BLOCKS_PER_BUDDY_REGION;
    startRefPage = startBlock >> params->blockBytesShift;

    if (endBlock > params->totalBlocks)
        endBlock = params->totalBlocks;

    // zero out first
    bzero(buddyRegion->u.data, FS_BLOCK_SIZE);
#ifndef BUDDY_INCORE
    FS_BUF_WRITE_NO_LOG(buddyRegion, 0, FS_BLOCK_SIZE);
#endif // BUDDY_INCORE

    // get all the refbytes for this buddy region
    for (pageno = 0;
         pageno < nRefPages && startBlock < endBlock;
         pageno++, startBlock += FS_BLOCK_SIZE) {

        fs_Buf *refBuf = O_GET_SHARED(refInode, startRefPage+pageno);
        fs_RefByte *ref = refBuf->u.refBytes;

        if ((startBlock + FS_BLOCK_SIZE) > endBlock)
            maxBlock = endBlock % FS_BLOCK_SIZE;
        else
            maxBlock = FS_BLOCK_SIZE;

        for (b = 0; b < maxBlock; ) {
            // skip non zero in use reference bytes

            if (ref[b]) {
                b++;
                used++;
            } else {

                // find first and last blocks to add to buddy free tree

                int32 len = 1;    // found the firstBlock, start with next
                firstBlock = startBlock + b++;
                while ((b < maxBlock) && (b < endBlock)) {
                    if (ref[b++]) {
                        used++;
                        break;
                    }
                    len++;
                }
                lastBlock = firstBlock + len - 1;
                //printf("%s region %d len %d start block %lld\n",
                //      context->fileSystem->super->mountName, region,
                //      len, firstBlock);
                fs_buddyRecalAddBlocks(context, len, fs_makeBlock(firstBlock),
                                       fs_buddyRecalAddExtent, NULL);
            }
        }
        fs_bufRelease(context, refBuf);
    }

    fs_inodeRelease(context, refInode);

    //printf("buddy recal region: %s %d done\n", 
    //      context->fileSystem->super->mountName, region);

    // validate the buddy region bit
    fs_buddySetRegionBit(context, region);

    return NFX_OK;
}


/**
 * Check if a buddy region needs to be recalcuated
 * Return TRUE if it was recalculated
 * FALSE if it was already validated
 */
static boolean
fs_buddyValidateRegion(fs_Context * context,
                       uint32       region)
    
{
    fs_Inode *inode = I_GET_SHARED(0, FS_FREE_BUDDY_TREE_INODE);
    fs_Buf *buddyRegion;

    // check if buddy region needs to be recalculated

    if (!fs_buddyGetRegionBit(context, region)) {
        buddyRegion = O_GET_EXCL(inode, region);

        // check again
        if (!fs_buddyGetRegionBit(context, region)) {
            fs_buddyRecalculateRegion(context, buddyRegion, region);
            fs_bufReleaseExclusive(context, buddyRegion);
            fs_inodeRelease(context, inode);
            return TRUE;
        }
        fs_bufReleaseExclusive(context, buddyRegion);
    }
    fs_inodeRelease(context, inode);
    return FALSE;
}


/**
 * Thread to recalculate buddy regions.
 * Fired off after log replay to start rebuilding buddy.
 */
void
fs_buddyRecalThread(esm_handle_t    handle,
                    void *          arg)
    
{
    fs_Context *    context = arg;
    fs_FileSystem * fs = context->fileSystem;
    uint32  r;
    boolean done;

    // PR #1789: acquire global access lock
    fs_lockShared(context, &context->fileSystem->accessLock, NULL);

    FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
            ("buddy recal thread: buddyRegions %d",
             fs->buddyRegions));

    // recalculate all regions
    for (r = 0; r < fs->buddyRegions; r++)
    {
        /* Recalculate the buddy region from the ref. bytes.
         */
        done = fs_buddyValidateRegion(context, r);

        /* Compute the summary of this buddy.
         */
        if (context->fileSystem->buddies[r].summary == NULL)
        {
            fs_Inode *  inode = I_GET_SHARED(0, FS_FREE_BUDDY_TREE_INODE);
            fs_Buf *    buf = O_GET_EXCL(inode, r);
            fs_inodeRelease(context, inode);
            fs_buddyGetSummary(context, buf);
            fs_bufReleaseExclusive(context, buf);
        }
    } // for (r)

    // release global access lock
    fs_unlockShared(context, &context->fileSystem->accessLock);

#if defined(FS_DEBUG_BUDDY_RECAL)
    fs_lockExclusive(context, &context->fileSystem->accessLock, NULL);
    fs_lockExclusive(context, &context->fileSystem->modifyLock, NULL);

    context->eek.squelchSscPrintf = TRUE;
    FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
            ("Performing mini-eek"));

    rc = fs_eek_ref_vs_buddy(context);

    fs_unlockExclusive(context, &context->fileSystem->modifyLock);
    fs_unlockExclusive(context, &context->fileSystem->accessLock);
    if (rc != NFX_OK) {
        FS_ELOG(context, LOG_ERR, FS_TRACE_REQ,
                ("mini-eek FAILED!!!"));
        assert(TRUE);
    }
#endif

    FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
            ("buddy recal thread: exiting"));

    // indicate that the recalculation is done by clearing the buddy
    // thread handle in the in core filesystem
    // if unmount was done before all regions are recalculated, it must
    // be marked dirty so the recalculations can be done at the next mount

    NFX_SPIN_LOCK(&context->fileSystem->spin);
    context->fileSystem->buddyThreadHandle = 0;
    NFX_SPIN_UNLOCK(&context->fileSystem->spin);

    fs_threadTerminate(context);
}


//   E E K


/**
 * Check that the specified extent (and, recursively, its children)
 * are free, for eek.  Print warnings about free children of free
 * extents.
 */
static boolean fs_buddyEekCheckFreeExtent(
    fs_Context *context,
    fs_Buf *buf,
    uint64 base,
    int freeLevel,
    int freeOffset,
    int level,
    uint32 offset)
{
    int index = (1 << (MAX_LEVELS - level)) + (offset >> level);

    assert(level >= 0);
    assert((offset & ((1 << level) - 1)) == 0);

    if (fs_bufBitIsSet(context, buf, index)) {
        FS_ELOG(context, LOG_ERR, FS_TRACE_REQ,
                ("ERROR: free extent %lld ... %lld"
                 " covered by free extent %lld %lld",
                 base + offset,
                 base + offset + (1 << level) - 1,
                 base + freeOffset,
                 base + freeOffset + (1 << freeLevel) - 1));
        return FALSE;
    }
    
    if (level > 0) {
        int next = level - 1;
        if (!fs_buddyEekCheckFreeExtent(context, buf, base, freeLevel, freeOffset,
                                   next, offset))
            return FALSE;
        if (!fs_buddyEekCheckFreeExtent(context, buf, base, freeLevel, freeOffset,
                                   next, offset + (1 << next)))
            return FALSE;
    }
    return TRUE;
}


/**
 * Mark the specified extent (and, recursively, its children) for eek.
 * The extent is identified by its level and offset within the region.
 */
static boolean fs_buddyEekMarkExtent(
    fs_Context *context,
    fs_Buf *buf,
    uint8 *bitmap,
    uint64 base,
    int level,
    uint32 offset)
{
    int index = (1 << (MAX_LEVELS - level)) + (offset >> level);

    assert(level >= 0);
    assert((offset & ((1 << level) - 1)) == 0);

    if (fs_bufBitIsSet(context, buf, index)) {
        uint32 i;
        uint32 len = 1 << level;

        //printf("free: %lld ... %lld @ %d (%d)\n",
        //       base + offset, base + offset + len - 1, level, len);
        
	//
	// XXX: if this fails, best to recompute the buddy tree
	// from scratch.
	//
        for (i = 0; i < len; i++) {
            if (!fs_eek_mark(context, "buddy tree", bitmap, base + offset + i))
                return FALSE;
        }
        if (level > 0) {
            int next = level - 1;
            if (!fs_buddyEekCheckFreeExtent(context, buf, base, level, offset,
                                       next, offset))
                return FALSE;
            if (!fs_buddyEekCheckFreeExtent(context, buf, base, level, offset,
                                       next, offset + (1 << next)))
                return FALSE;
        }
    }
    else if (level > 0) {
        int next = level - 1;
	    boolean ok;
        ok = fs_buddyEekMarkExtent(context, buf, bitmap, base,
                              next, offset);
	    if (!ok)
		    return FALSE;
        return fs_buddyEekMarkExtent(context, buf, bitmap, base,
                                    next, offset + (1 << next));
    }
    return TRUE;
}


/**
 * Mark the blocks on a region page of the binary buddy tree for eek.
 *
 * TODO - check region summary?
 */
boolean fs_buddyEekMarkRegion(
    fs_Context *context,
    fs_Buf *buf,
    uint8 *bitmap,
    uint64 baseBlock)
{
    return fs_buddyEekMarkExtent(context, buf, bitmap, baseBlock, MAX_LEVELS, 0);
}

/* Test if a block is not yet allocated.
 */
boolean
fs_isBlockFree(fs_Context * context,
               fs_Block     block)
{
    fs_RefByte refByte = fs_refGetByte(context, block);
    return(fs_refbyteIsLive(refByte));
}



boolean
fs_verifyBuddyCount(fs_Context * context, int recalc)
{
    int loop, loop1;
    uint64 free, tot = 0 ;
    fs_Inode *inode;
    
    printf ("fs_verifyBuddyCount blks used %lld tot %lld - inodes used %lld tot %lld - buddy %d\n",
            context->fileSystem->super->blksUsed, 
            context->fileSystem->super->numBlks,
            context->fileSystem->super->inodesUsed, 
            context->fileSystem->params.maxInodeNumber,
            context->fileSystem->params.buddyRegionCount);

    tot = 0;
    for (loop = 0 ; loop < context->fileSystem->params.buddyRegionCount; loop++)
    {
        free = 0;
        printf ("region %d - %d - ", loop, context->fileSystem->buddies[loop].flags);
        if (context->fileSystem->buddies[loop].summary)
        {
            for (loop1 = 0 ; loop1 < 16; loop1++)
            {
                printf ("%d ",
                        ((uint16*)context->fileSystem->buddies[loop].summary)[loop1]);
                free += ( ((uint16*)context->fileSystem->buddies[loop].summary)[loop1] * 
                          (1 << loop1));
            }
        }
        tot += free;
        printf (" - free %lld\n", free);
    }
    printf ("blks used %lld free %lld tot %lld (%lld)\n",
            context->fileSystem->super->blksUsed,
            tot,
            context->fileSystem->super->numBlks,
            context->fileSystem->super->numBlks - context->fileSystem->super->blksUsed - tot);

    if ( ! recalc)
    {
        return (TRUE);
    }

    inode = I_GET_SHARED(0, FS_FREE_BUDDY_TREE_INODE);
    for (loop = 0 ; loop < context->fileSystem->params.buddyRegionCount; loop++)
    {
        fs_Buf *buddyRegion;
        
        printf ("region %d - %d\n", loop, fs_buddyGetRegionBit(context, loop));

        buddyRegion = O_GET_EXCL(inode, loop);
        fs_buddyRecalculateRegion(context, buddyRegion, loop);
        fs_bufReleaseExclusive(context, buddyRegion);
    }
    fs_inodeRelease(context, inode);

    tot = 0;
    for (loop = 0 ; loop < context->fileSystem->params.buddyRegionCount; loop++)
    {
        free = 0;
        printf ("region %d - %d - ", loop, context->fileSystem->buddies[loop].flags);
        if (context->fileSystem->buddies[loop].summary)
        {
            for (loop1 = 0 ; loop1 < 16; loop1++)
            {
                printf ("%d ",
                        ((uint16*)context->fileSystem->buddies[loop].summary)[loop1]);
                free += ( ((uint16*)context->fileSystem->buddies[loop].summary)[loop1] * 
                          (1 << loop1));
            }
        }
        tot += free;
        printf (" - free %lld\n", free);
    }
    printf ("blks used %lld free %lld tot %lld (%lld)\n",
            context->fileSystem->super->blksUsed,
            tot,
            context->fileSystem->super->numBlks,
            context->fileSystem->super->numBlks - context->fileSystem->super->blksUsed - tot);
	return TRUE;
} 

#endif
