/*
 * Name:  fs-buf.h
 *
 * RCS:  $Id: fs-buf.h,v 1.131 2002/12/04 22:51:43 isabellec Exp $
 *
 * Description:  Definitions for fs buffer cache and rw locks.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#if !defined(FS_BUF_H)
#define FS_BUF_H


/**
 * Sentinel value for a buffer that's not in the page cache.
 */
#define FS_BUF_NO_PAGE      (-1)

/**
 * The flags field in a buffer header.
 */
typedef enum fs_bufFlag
{
    FS_BUF_DIRTY        = 0x00000001,   // Data needs writeback.
    FS_BUF_IO_BUSY      = 0x00000002,   // I/O in progress on buffer.
    FS_BUF_TYPE_USER    = 0x00000004,   // Buffer is for user data.
    FS_BUF_TYPE_META    = 0x00000008,   // Buffer is for meta data.
    FS_BUF_COW_UNK      = 0x00000010,   // Buffer COW state unknown
    FS_BUF_COW          = 0x00000020,   // Buffer is COW
    FS_BUF_INVALID      = 0x00000040,   // Buf no longer contains valid data.
    FS_BUF_TRANS_WRITE  = 0x00000080,   // First write for transaction occured.
    FS_BUF_FREE_DEFER   = 0x00000100,   // Buf is pending deallocation.
    FS_BUF_COPIED       = 0x00000200,   // Buf was copied on a write
    FS_BUF_IS_CLONE     = 0x00000400,   // Buf is a clone for writeback.
    FS_BUF_MRD          = 0x00000800,   // Buf is on MRD list.
    FS_BUF_WAIT_LOGBUF  = 0x00001000,   // Buf I/O pending logbuf flush.
    FS_BUF_FREE_LOGBUF  = 0x00002000,   // Buf free pending logbuf flush.
    FS_BUF_LOG_REPLAY   = 0x00004000,   // Buf read in during log replay time.
    FS_BUF_READ_VERIFY  = 0x00008000,   // Buf needs read verify (FS_DEBUG)
    FS_BUF_READ_VERIFIED= 0x00010000,   // Buf read verify done (FS_DEBUG)
    FS_BUF_WAITING_MRD  = 0x00020000,   // Buf is on waiting MRD list.
    FS_BUF_WAIT_CLONE   = 0x00040000,   // Buf writeback is waiting for
                                        // clone to complete
    FS_BUF_WB_STARTING  = 0x00080000,   // Buf writeback starting
    FS_BUF_INUM_CNT     = 0x00100000,   // Buf is counted in per inum stats
    FS_BUF_DIRECTORY    = 0x00200000,   // Buf is a directory
    FS_BUF_COMPACT      = 0x00400000,   // Buf is in compact leaf form
} fs_BufFlag;

// #define FS_DEBUG_BUF_HISTORY

#ifdef FS_DEBUG_BUF_HISTORY
#define FS_DEBUG_BUF_HISTORY_MAX   16

typedef enum fs_bufHistory
{
    FS_BUF_HIST_NON_INIT,
    FS_BUF_HIST_BUFLOCK,         // Rx
    FS_BUF_HIST_BUFLOCK_EVENT,   // Sent
    FS_BUF_HIST_BUFLOG,          // Rx
    FS_BUF_HIST_BUFLOG_EVENT,    // Sent
    FS_BUF_HIST_IODONE,
    FS_BUF_HIST_IODONE_RETURN,
    FS_BUF_HIST_CLONE_ORIG_WB,
    FS_BUF_HIST_CLONE_ORIG_REL,
    FS_BUF_HIST_IODONE_CLEAR,
    FS_BUF_HIST_SET_WB_STARTING,
    FS_BUF_HIST_SET_WB_STARTING_WAITING_MRD,
    FS_BUF_HIST_SET_WB_STARTING_WB_LIST,
    FS_BUF_HIST_SET_IO_BUSY,
    FS_BUF_HIST_DESTROY,
    FS_BUF_HIST_CLONE,
    FS_BUF_HIST_CLONED,
    FS_BUF_HIST_ALLOC,
    FS_BUF_HIST_ALLOC_REUSE,
    FS_BUF_HIST_CLONE_WAIT,
    FS_BUF_HIST_CLONE_WAIT_LOG,
    FS_BUF_HIST_FREE_DEFER_RESET_DIRTY,
    FS_BUF_HIST_SET_DIRTY,
    FS_BUF_HIST_SET_DIRTY_MRD,
    FS_BUF_HIST_RESET_DIRTY_IO_BUSY,
    FS_BUF_HIST_RESET_COMPACT,
    FS_BUF_HIST_SET_COMPACT,
    FS_BUF_HIST_FREE_DEFER,
    FS_BUF_HIST_SET_INVALID,
    FS_BUF_HIST_FREE_DEFER_COMMIT,

    FS_BUF_HIST_INSERT_MRD,
    FS_BUF_HIST_REMOVE_MRD_WB,
    FS_BUF_HIST_REMOVE_W_MRD_WB,
    FS_BUF_HIST_REMOVE_L_MRD_WB,
    FS_BUF_HIST_REMOVE_MRD_WB_2,
    FS_BUF_HIST_REMOVE_L_MRD_WB_2,
    FS_BUF_HIST_INSERT_W_MRD_WB_2,
    FS_BUF_HIST_REMOVE_MRD_FREE,
    FS_BUF_HIST_REMOVE_W_MRD_FREE,
    FS_BUF_HIST_REMOVE_L_MRD_FREE,
    FS_BUF_HIST_REMOVE_MRD_DIRTY,
    FS_BUF_HIST_REMOVE_W_MRD_DIRTY,
    FS_BUF_HIST_REMOVE_L_MRD_DIRTY,
    FS_BUF_HIST_INSERT_L_MRD_DIRTY,
    FS_BUF_HIST_REMOVE_PAGE,
    FS_BUF_HIST_INSERT_PAGE,
    FS_BUF_HIST_REMOVE_PAGE_DIRTY,
    FS_BUF_HIST_INSERT_PAGE_DIRTY,
    FS_BUF_HIST_INSERT_PAGE_GET,


    FS_BUF_HIST_MAX
} fs_BufHistory;

#endif


/**
 * A buffer header user-type component.
 */
typedef struct fs_BufUser
{
    nfx_cod_t           codHandle;
    int32               codOffset;
    int32               codLength;
} fs_BufUser;

/**
 * Inner node of a B-tree.
 *
 * Any key[i] has a left pointer left[i] such that any nodes in left[i]
 * have keys <= key[i]. The rightmost pointer is always stored in
 * fs_Btree.right, regardless of the number of keys.
 */
#define FS_BTREE_MAXKEYS \
    ((PAGE_SIZE/(sizeof (fs_Extent) + sizeof (uint64) + sizeof (uint16))) - 1)

typedef struct fs_Btree
{
    uint16  index[FS_BTREE_MAXKEYS];    /* index heap */
    int16  level;                       /* level in tree of node  */
    int16  nkeys;                       /* number of keys in node */

    uint64  keys[FS_BTREE_MAXKEYS];     /* keys used to order Btree */

    uint64  left[FS_BTREE_MAXKEYS];     /* left pointers */
    uint64  right;                      /* rightmost pointer */

} fs_Btree;



/*
 * Directory page structures:
 *
 * !!!!!!!! NOTE !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 * !!                                                                !!
 * !!  The fields within the 3 structures are positioned             !!
 * !!  such that we can do some dirty casting between them.          !!
 * !!  The compact form overlays the long form completely, and       !!
 * !!  the len and cont fields overlay eachother between the         !!
 * !!  long and continuation.  Don't goof with this!                 !!
 * !!                                                                !!
 * !!!!!!!! NOTE !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
 *
 * 40 byte long-form directory entry.
 */
typedef struct fs_Dirent {
	uint16 len;
    uint16 aliasSeq;                    // 8.3 ~sequence number
#define FS_DIR_ALIAS_SEQ_MASK   0x7f    // the number after the tilde
#define FS_DIR_ALIAS_SEQ_MAX    99      // the highest we'll create
#define FS_DIR_ALIAS_SEQ_CHARS_MAX  2   // up to 2 decimal digits
#define FS_DIR_ALIAS_SEQ_RAND   FS_DIR_ALIAS_SEQ_MAX + 1 // alias is in random num format
#define FS_DIR_ALIAS_SEQ_FLG_MASK   0xf800  
#define FS_DIR_ALIAS_SEQ_FLG_CASE   0x0800 // case collision exists/existed
#define FS_DIR_ALIAS_SEQ_FLG_ILCH   0x1000 // longname contains illegal chars
#define FS_DIR_ALIAS_SEQ_FLG_DUPE   0x2000 // longname is/was dup of another entry's alias 
    uint32 aliasCksumRand;  // 8.3 cksum or random num
	uint64 inum;
	char name[16];
    uint32 hash;
	uint16 spare;
	uint16 cont;
} fs_Dirent;
NFX_ASSERT_STATIC(sizeof(fs_Dirent) == 40);

typedef struct fs_Dircont {
	uint16 len;
	char name[36];
	uint16 cont;
} fs_Dircont;
NFX_ASSERT_STATIC(sizeof(fs_Dircont) == sizeof(fs_Dirent));

#define FS_DIRCONT_EOL 0xf00f
#define FS_DIRFREE_EOL 0xf00d
#define FS_DIRCONT_XXX 0xf0ee

/*
 * 32 byte short-form directory entry.
 */
typedef struct fs_DirentCompact {
	uint16 len;
    uint16 aliasSeq;        // 8.3 ~sequence number
    uint32 aliasCksumRand;  // 8.3 cksum or random num
    uint64 inum;
	char name[16];
} fs_DirentCompact;
NFX_ASSERT_STATIC(sizeof(fs_DirentCompact) == 32);

enum {
    DIRLEAF_COMPACT_MAGIC = 0xfcea,
    DIRLEAF_LONG_MAGIC = 0xebbb
};

#define FS_DIRLEAF_ENTRIES_MAX ((PAGE_SIZE - 2*(sizeof (uint16)) - sizeof (uint64)) / sizeof (fs_Dirent))
typedef struct fs_Dirleaf {
	uint64 parent_inum;
    uint16 magic;
	uint16 count;
	uint16 freelist;
	union {
		fs_Dirent ent;
		fs_Dircont cont;
	} u[FS_DIRLEAF_ENTRIES_MAX];
} fs_Dirleaf;

#define FS_DIRLEAF_COMPACT_ENTRIES_MAX  ((FS_COMPACT_PAGE_SIZE - sizeof (uint16) - sizeof (uint64)) / sizeof (fs_DirentCompact))
typedef struct fs_DirleafCompact {
	uint64 parent_inum;
    uint16 magic;
	uint16 count;
	fs_DirentCompact entries[FS_DIRLEAF_COMPACT_ENTRIES_MAX];
} fs_DirleafCompact;
NFX_ASSERT_STATIC(sizeof(fs_DirleafCompact) <= FS_COMPACT_PAGE_SIZE);


/**
 * Directory indexes are a hack to provide for stable dirent
 * indexes to be used as cookies for NFS readdir.  Without them,
 * we are unable to ensure consistent readdir results with the
 * many ill-behaved NFS clients that treat cookies as fixed
 * directory offsets for a dirent. Ideally, we should be able to
 * use cookies derived from our btree-based structures, but since
 * they compact and do not remain stable for a given dirent, the
 * clients that do readdir(0)/remove/readdir(n) can miss dirents
 * and things like 'rm -rf <dir>' tend to fail with "directory
 * not empty" results.
 */


/**
 * A single dirent entry on an indexer page.
 */
typedef struct fs_DirIndexEntry
{
    uint32  hashcode;
    uint8   leafIndex;
    uint8   isLast;
    uint8   spare[2];

} fs_DirIndexEntry;

/**
 * Map this many directory indexes per indexer page.  We leave
 * some holes in the namespace (unused index numbers due to other
 * info on each page), but there's plenty to spare.
 */
#define FS_DIR_INDEX_MAP_PER_PAGE   (PAGE_SIZE / sizeof(fs_DirIndexEntry))



#define FS_DIR_INDEX_SUMMARY_WORDS  56
#define FS_DIR_INDEX_SUMMARY_BITS   (FS_DIR_INDEX_SUMMARY_WORDS * BITS_PER_WORD)
#define FS_DIR_INDEX_ADMIN_WORDS    8
#define FS_DIR_INDEX_ENTRY_MAX                                      \
    ((PAGE_SIZE -                                                   \
      ((FS_DIR_INDEX_SUMMARY_WORDS * sizeof(uint32)) +              \
       (FS_DIR_INDEX_ADMIN_WORDS * sizeof(uint32)))) /              \
     sizeof (fs_DirIndexEntry))
// This many indexes are reserved, which includes space for
// '.', '..', and a few for other names we might want to generate.
// For simplicity, we skip this many indexes at the start of each
// page of the index file.
#define FS_DIR_INDEX_PAGE_INDEX_START     \
      (FS_DIR_INDEX_MAP_PER_PAGE - FS_DIR_INDEX_ENTRY_MAX)
NFX_ASSERT_STATIC((FS_DIRLEAF_ENTRIES_MAX + FS_DIR_INDEX_PAGE_INDEX_START) <= \
                        FS_DIR_INDEX_ENTRY_MAX);

typedef struct fs_DirIndexPageField
{
    uint32  summary[FS_DIR_INDEX_SUMMARY_WORDS];
    uint32  admin[FS_DIR_INDEX_ADMIN_WORDS];
// First slot of admin field stores the number of indexes
// used on the page.
#define FS_DIR_INDEX_ADMIN_NUM_USED     0
    fs_DirIndexEntry    entries[FS_DIR_INDEX_ENTRY_MAX];
} fs_DirIndexPageField;

typedef struct fs_DirIndexPage
{
    union {
        fs_DirIndexPageField    field;
        fs_DirIndexEntry        entry[FS_DIR_INDEX_MAP_PER_PAGE];
    } u;

} fs_DirIndexPage;

/**
 * File inode leaf page, an array of file byte
 * offsets and corresponding extents.
 */
#define FS_EMAP_EXTENTS_MAX (PAGE_SIZE /        \
                    (sizeof (fs_Extent) + sizeof (uint64) + sizeof (uint16)))

#define FS_EMAP_COMPACT_EXTENTS_MAX (FS_COMPACT_PAGE_SIZE /        \
                    (sizeof (fs_Extent) + sizeof (uint64) + sizeof (uint16)))

#define TYPEDEF_EMAPLEAF(_type_, _max_)\
typedef struct _type_\
{\
    int16       nkeys;            /* number of keys in node */ \
    uint16      index[_max_];     /* index heap */ \
\
    uint64      offsets[_max_];   /* keys used to order Btree */ \
    fs_Extent   extents[_max_];   /* left pointers */ \
\
} _type_

TYPEDEF_EMAPLEAF(fs_Emapleaf, FS_EMAP_EXTENTS_MAX);
TYPEDEF_EMAPLEAF(fs_EmapleafCompact, FS_EMAP_COMPACT_EXTENTS_MAX);

/**
 * A page of on-disk inodes.
 */
#define FS_DINODE_PAGE_MAX (PAGE_SIZE/sizeof(fs_Dinode))
typedef struct fs_DinodePage
{
    fs_Dinode   dinode[FS_DINODE_PAGE_MAX];

} fs_DinodePage;


/**
 * Common header for on-disk security descriptors in the
 * security file.
 */
struct fs_SecurityDescHdr
{
    uint16  attr;
#define FS_SDESC_TYPE_MASK  0x0003
    uint16  len;    // Length of the descriptor including this header,
                    // not counting continuation entries.
    uint32  cont;   // Offset at which continuation entry exists
                    // for this descriptor.
    uint32  cksum;  // Checksum over the descriptor including this header,
                    // not counting continuation entries.
    uint32  uQtId;  // index in the user quota file (0 -> uninitialized)
    uint32  gQtId;  // index in the group quota file (0 -> uninitialized)

};


/**
 * An on-disk NFS security descriptor.
 */
struct fs_SecurityDescNfs
{
    fs_SecurityDescHdr  hdr;
    uint32              uid;
    uint32              gid;
    uint16              mode;		// see fs-api.h
    uint16              pad[1];

};

NFX_ASSERT_STATIC(FS_BLOCK_SIZE % sizeof(fs_SecurityDescNfs) == 0);
#define FS_SDESC_NFS_PER_PAGE   (FS_BLOCK_SIZE / sizeof(fs_SecurityDescNfs))

typedef struct fs_SecurityDescPageNfs fs_SecurityDescPageNfs;
struct fs_SecurityDescPageNfs
{
    fs_SecurityDescNfs  nfs[FS_SDESC_NFS_PER_PAGE]; 
};


/**
 * An on-disk CIFS security descriptor.
 */
struct fs_SecurityDescCifs
{
    fs_SecurityDescHdr  hdr;
    uint64              incorePtr;  // used in-core only, ptr to var len data
    
    // TBD

};

/**
 * On0disk Fileset Page
 */

typedef struct fs_FileSetPage
{
    fs_FileSetDiskRecordPadded recs[FS_FILESET_RECS_PER_PAGE];
} fs_FileSetPage;

/**
 *  On-disk Quota Page
 */  
typedef struct fs_QuotaPage
{
    fs_QuotaDiskRecordPadded  recs[FS_QUOTA_RECS_PER_PAGE]; 
}fs_QuotaPage;


/**
 * A buffer header.
 */
struct fs_Buf
{
    fs_FileSystem       *fileSystem;    // Filesystem buf associated with.
    fs_Block            block;          // Physical volume block location.
    uint64              inum;           // Inode number for the block.
    uint64              page;           // Offset, in blocks, within inode.
    uint64              snapshots;      // Bitset of snapshots this block is in
    fs_Lock             lock;           // Lock for the buffer.
    fs_Buf              *clone;         // Ptr to clone of this buf, or in the 
                                        // clone ptr back to orig.
    fs_LogSegmentDeps   logSegDep;      // Log dependencies
volatile fs_BufFlag     flags;          // Buffer/block state flags.
    uint16              logRefCount;    // Count of committed transactions in
                                        // log buffers not yet on disk that
                                        // have mods for this buf.
    int16               headLogSeg;     // Log segment this buf first dirties.
    fs_LogId            zeroFillLogId;  // Log id of transaction to zero-fill
                                        // this buf.
    nfx_spinlock_t      spin;
    
    void (*cleanupFunction)(fs_Context *context, fs_Buf *buf);
                                        // function called when buf is delisted

    DLL_ENTRY(fs_Buf)    blkNumEntry;   // Overflow chain for hashed by
                                        // blkNum list.
    DLL_ENTRY(fs_Buf)    pageHashEntry; // Entry for inode/page hash chain.
    DLL_ENTRY(fs_Buf)    pageEntry;     // Entry for inode page header chain.
    DLL_ENTRY(fs_Buf)    pageDirtyEntry;// Dirty entry for inode page header
                                        // chain.
    DLL_ENTRY(fs_Buf)    tmpPageEntry;  // Tmp entry for inode page header
                                        // chain in fs_bufFreeInode.
    DLL_ENTRY(fs_Buf)    mruEntry;      // Most Recently Used (MRU) entry.
    DLL_ENTRY(fs_Buf)    mrdEntry;      // MRD (dirtied) list.  Aids write
                                        // scheduling for log truncaction.
#ifndef NEW_ZF
    DLL_ENTRY(fs_Buf)    zeroFillEntry; // Buf was zero-filled during replay.
#endif // NEW_ZF
    DLL_HDR(fs_LogUndoSummary)
                         undoHdr;       // List of undo'able records.
    DLL_HDR(fs_LogUndoSummary)
                         candoHdr;      // List of undo'able records that are
                                        // ready for being made permanent.
    union
    {
        fs_BufUser          user;
        void                *rawdata;
        char                *data;
        fs_RefByte          *refBytes;
        fs_Btree            *btree;
        fs_Dirleaf          *dirleaf;
        fs_DirleafCompact   *dirleafCompact;
        fs_DirIndexPage     *dirIndexPage;
        fs_Emapleaf         *emapleaf;
        fs_EmapleafCompact  *emapleafCompact;
        fs_DinodePage       *dinode;
        fs_SecurityDescPageNfs  *sdnfs;
        fs_FileSetPage      *fset;
        fs_QuotaPage        *quota;
        uint8               *heap1;
        uint16              *heap2;
        uint32              *word;
        uint8               *bytes;
        uint16              *shorts;
        uint32              *longs;
        uint64              *longlongs;
    } u;

#if defined(FS_DEBUG_SHADOW_BUF)
    char                *shadow;      // Shadow area for data, used to compare
#endif                                // logged writes to live buffer data.

#ifdef FS_DEBUG_BUF_HISTORY
    nfx_spinlock_t      spinHistory;
    int                 indexHistory;
    char                history[FS_DEBUG_BUF_HISTORY_MAX];
    char                cpuHistory[FS_DEBUG_BUF_HISTORY_MAX];
    fs_BufFlag          flagsHistory[FS_DEBUG_BUF_HISTORY_MAX];
    void                *mrdEntryNextHistory[FS_DEBUG_BUF_HISTORY_MAX];
    void                *mrdEntryPrevHistory[FS_DEBUG_BUF_HISTORY_MAX];
#endif

};


/**
 * A header for a per-inum chain of bufs by page.
 */
typedef struct fs_BufPageHdr fs_BufPageHdr;
struct fs_BufPageHdr
{
    fs_FileSystem       *fileSystem;
    uint64              inum;
    DLL_ENTRY(fs_BufPageHdr)
                        inumEntry;  // Overflow chain for hashed by inum list.

    DLL_HDR(fs_Buf)     bufChainHdr;// Header of page chain of bufs.
    DLL_HDR(fs_Buf)     bufDirtyHdr;// Header of page chain of dirty bufs.
    uint32              dirtyCount; // Count of dirty pages for this inode.
    DLL_HDR(fs_Waiter)  waiterHdr;  // contexts waiting for I/O to complete.
};


/**
 * Buffer locking modes.
 */
typedef enum fs_BufferMode {
    /** I can share access with others. */
    BUF_GET_MODE_SHARED = 1,
    /** I need exclusive access. */
    BUF_GET_MODE_EXCL   = 2
} fs_BufferMode;


/**
 * Buffer get options.  (Not declared as an enum type, because they can be
 * combined.)
 */
enum {
    /** Sleep wait vs. return immediately. */
    BUF_GET_OPT_SLEEP = 0x01,
    /** Instantiate new block (don't read). */
    BUF_GET_OPT_NEW = 0x02,
    /** Is user buf (vs. metadata buf). */
    BUF_GET_OPT_USER = 0x04,
    /** Just return NULL if buffer is not already cached. */
    BUF_GET_OPT_LOOKUP = 0x08,
    /** Defer allocation of block number. */
    BUF_GET_OPT_DEFER_ALLOC = 0x10,
    /** Return NULL if someone else has the buffer locked. */
    BUF_GET_OPT_TRY_LOCK = 0x20,
};


/**
 * Buffer free options.
 */
enum {
	BUF_DO_DEALLOCATE   = 0x01,
    BUF_IS_REPLAY       = 0x02,
    BUF_DONT_LOG_FLUSH  = 0x04,
};


/**
 * Very ugly sentinel value that isn't NULL for reporting a trylock failure.
 * Only reported when BUF_GET_OPT_TRY_LOCK is used.
 */
#define FS_BUF_TRY_LOCK_FAILED   ((fs_Buf *) -1)



#ifdef FS_DEBUG_BUF_CKSUM

/**
 * Used to track cksum (in-core only) of written meta bufs.
 */

struct fs_BufCksum
{
// if adding data, please verify to update FS_BUF_CKSUM_ENTRIES_PER_MEM_BLOCK

    uint64      cksum;
    fs_Block    block;
    uint32      lastWriteTime;
    DLL_ENTRY(fs_BufCksum)
                dllEntry;
};


// For the meta buffer Cksum calculations, 
// allocate big blocks of memory and 
// keep a global pool of free entries

#define FS_BUF_CKSUM_ENTRIES_PER_MEM_BLOCK   10000
// total size for malloc, big enough to hit cont mem
// with sizeof(fs_BufCksum) == 40 --> 400K

#define FS_BUF_CKSUM_MAXNBR_OF_BLOCKS        10
// Define an absolute max nbr of entries (== memory) we can use for this
// This is global for ALL filesystems !!!!!
// with sizeof(fs_BufCksum) == 40 --> 400K * 10 = 4M

typedef struct fs_BufCksumFreePool
{
    nfx_spinlock_t         spin;
    DLL_HDR(fs_BufCksum)   freeList;
    uint32                 nbrOfMemBlocks;
    uint32                 nbrOnFreeList;
    uint32                 nbrInUse;
} fs_BufCksumFreePool;

#endif // FS_DEBUG_BUF_CKSUM


extern uint32  fs_diskWritebackType[FS_MAX_META_INODE];
extern uint32  fs_transWriteType[FS_MAX_META_INODE];


/**
 * Called to get (by block number) and lock a block in the buffer cache.
 */
extern fs_Buf *fs_bufGetBlock(
    fs_Context      *context,
    fs_Inode        *inode,
    fs_Block        block,
    int32           mode,
    uint32          option);


/**
 * Return the buffer for the specified (inode, page) pair, locked with the
 * specified mode.
 */
extern fs_Buf *fs_bufGetPage(
    fs_Context      *context,
    fs_Inode        *inode,
    uint64          page,
    int32           mode,
    uint32          option,
    evm_volSize_t   readLen);


/**
 * Search for a buffer containing the specified (inode, page) pair and
 * return its block number, or an invalid block if the page is not found
 * in the buffer cache.
 *
 * The buffer is not held locked on return from this routine, so it is
 * possible that the block will change before the information is used.
 * Thus, this information should only be used in an advisory sense, and
 * not for, say, reading a block from disk.
 */
fs_Block fs_bufLookupBlockByPage(
    fs_Context *context,
    fs_Inode   *inode,
    uint64     page);


/**
 * Called to unlock and release an fs block, held with a shared lock, back
 * to the cache.
 */
extern void fs_bufRelease(
    fs_Context  *context,
    fs_Buf      *buf);

/**
 * Called to unlock and release an fs block, held with a exclusive lock,
 * back to the cache.
 *
 * Should normally only be used if buffer was NOT modified.  If buffer was
 * modified, FS_BUF_WRITE() should be used (eventually followed by
 * fs_logCommitEntry()).
 */
extern void fs_bufReleaseExclusive(
    fs_Context  *context,
    fs_Buf      *buf);


/**
 * Buffer write options.
 */
enum {
    /** Log the write operation. */
    FS_BUF_WRITE_LOG = 0x1,
    /** Add the buf to the log summary if this is the first write for this
     *  transaction. */
    FS_BUF_LOG_ADD_BUF = 0x2,
    /** Buffer contents are zero.  Use this to record a compact,
     *  zero-filled log entry. */
    FS_BUF_WRITE_ZERO = 0x4,
    /** Record a bitmap set log entry for the write */
    FS_BUF_WRITE_LOG_BIT_SET = 0x8,
    /** Record a bitmap reset log entry for the write */
    FS_BUF_WRITE_LOG_BIT_RESET = 0x10,
    /** Add an undo entry to the buf */
    FS_BUF_WRITE_ADD_UNDO = 0x40,
};

/**
 * Called to indicate that a buf has been written.
 * Takes care of all COW and logging concerns.
 */
extern void fs_bufWrite(
    fs_Context  *context,
    fs_Buf      *buf,       
    int32       offset,     // Byte offset where write started (n/a for
                            // user (cod) bufs.)
    int32       length,     // Length of write in bytes (n/a for user
                            // (cod) bufs.)
    uint32      flags,      // Optional flags.
    void        *bitmap,    // Optional bitmap for log record.
    fs_BitmapOperation *debugOp  // Optional for bitmap op debugging.
);


/**
 * Mark a buffer as dirty if it is not already so.
 */
extern void fs_bufDirty(
    fs_Context *context,
    fs_Buf *buf);


/**
 * Called at transaction commit to commit changes changes
 * to a buf and release it.
 */
extern void
fs_bufCommit(
    fs_Context  *context,
    fs_Buf      *buf
);


/**
 * Print all locked buffers in the system.
 */
extern void fs_bufShowLocked();


/**
 * Instantiate the writeback state machines for the volume in the
 * specified context.
 */
extern void fs_bufStartWritebackDaemon(
    fs_Context *context);


/**
 * Destroy the writeback daemon state machines for the
 * specified filesystem and wait until they're actually
 * stopped.
 */
extern void fs_bufStopWritebackDaemonWait(
    fs_Context  *context,
    boolean     isForce);


/**
 * Signal the writeback daemon to wakeup for the specified buflist.
 * Has no effect if the writeback daemon is currently active for
 * a given bufflist.
 */
enum {
    FS_WB_WAKEUP_INVALID,
    FS_WB_WAKEUP_PAGE_THRESH,
    FS_WB_WAKEUP_ALL,
    FS_WB_WAKEUP_PAGE_SYNC,
    FS_WB_WAKEUP_SYNC,
    FS_WB_WAKEUP_INO_SYNC,
    FS_WB_WAKEUP_REPLAY,
    FS_WB_WAKEUP_LOG,
    FS_WB_WAKEUP_WAITER,
    FS_WB_WAKEUP_PERIODIC,
    FS_WB_WAKEUP_RESCHED,
    FS_WB_WAKEUP_PAD,
    FS_WB_WAKEUP_MAX,
};
// Keep WB_WAKEUP_MAX even, otherwise a padding problem
// in the cachestats resp can occur.
NFX_ASSERT_STATIC(FS_WB_WAKEUP_MAX % 2 == 0);

extern int32 fs_bufWbWakeupCount[FS_BUFLIST_MAX][FS_WB_WAKEUP_MAX];
extern uint32 fs_userBufWbMaxPerLoop;
extern uint32 fs_metaBufWbMaxPerLoop;

extern void fs_bufSignalWriteback(
    fs_Context  *context,
    int32       type,
    int32       eventType);


/**
 * Signal the writeback daemon to wakeup for a specified buffer
 * list and wait for it to complete the IO.
 */
extern void fs_bufSyncWriteback(
    fs_Context  *context,
    int32       type);


/**
 * Signal the writeback daemon to write all the buffer lists for the
 * current context, and wait until it's completed all of them.
 */
extern void fs_bufSyncWritebackLists(
    fs_Context *context);


/**
 * Perform allocation on the specified buf.
 */
extern void
fs_bufAllocBuf (
    fs_Context      *context,
    fs_Inode        *inode,
    fs_Buf          *buf,
	boolean			commitInode);


/**
 * Remove the log dependencies associated with buf.
 */
extern void fs_bufRemoveLogDeps(
    fs_Context      *context,
    fs_Buf          *buf);


/**
 * Start I/O on all dirty user bufs for the specified inode, and block
 * until the I/O is completed.
 */
extern void
fs_bufSyncInodeBufs(
    fs_Context         *context,
    fs_Inode           *inode);


/**
 * Start I/O on all dirty user bufs for the specified inode, and block
 * until the I/O is completed, or return FALSE if the I/O doesn't
 * complete within the specified timeout period.
 */
extern boolean
fs_bufSyncInodeBufsAndTimeout(
    fs_Context         *context,
    fs_Inode           *inode,
    int32              timeoutSecs);


/**
 * Called to get the copy-on-write state for a block.  Returns TRUE if the
 * block is COW.
 */
extern boolean fs_blockIsCopyOnWrite(
    fs_Context *context,
    fs_Block   block);


/**
 * Called to get the copy-on-write state for a buffer.  Returns TRUE if
 * buf is COW.
 */
extern boolean fs_bufIsCopyOnWrite(
    fs_Context      *context,
    fs_Buf          *buf);


/**
 * Called to free the resources (and disk block) associated with a buf.
 */
extern void fs_bufFree(
    fs_Context  *context,
    fs_Buf      *bufp,
    int32		option);


/**
 * Per filesystem buf alloc thread.
 */
void
fs_bufAllocThread(
    esm_handle_t    handle,
    void            *arg);


/**
 * Destroy the contents of the buffer cache for the
 * specified volume without trying to sync to disk.
 */
extern void
fs_bufCacheDestroyVolume(
    fs_Context      *context,
    fs_FileSystem   *fileSystem);


/**
 * Free all the buf's for the specified inode starting with the
 * specified page.  Don't deallocate their disk space.  This is
 * used during file truncation to clear the buffer cache for a file.
 */
void
fs_bufFreeInode(
    fs_Context  *context,
    fs_Inode    *inode,
    uint64      startPage,
    boolean     isReplay);


/**
 * Copy a metadata buffer on write.  Does eager allocation.
 */
extern void fs_bufCopyOnWriteMetadata(
    fs_Context  *context,
    fs_Inode    *inode,
    fs_Buf      *buf);


/**
 * Called to flush all non-dirty buffers from the cache. 
 */
extern void fs_bufCacheFlushSnap(
    fs_Context  *context,
    fs_Inode    *inode,
    int32       snapnum);


/**
 * Called to flush all non-dirty buffers from the cache.
 */
static inline void
fs_bufCacheFlush(
    fs_Context  *context,
    fs_Inode    *inode)
{
    fs_bufCacheFlushSnap(context, inode, 0);
}


/**
 * Adds an entry to the page cache.
 * Must be called with buflist spinlock held.
 */
extern void fs_bufPageCacheInsert(
    fs_Context  *context,
    fs_Buf      *buf,
    fs_BufList  *bufList,
    fs_Inode    *inode);


// If not found in cache return NULL vs. reading it in.
#define O_LOOKUP_SHARED(_inode_,_page_)         \
    fs_bufGetPage(    context,                  \
                        (_inode_),              \
                        (_page_),               \
                        BUF_GET_MODE_SHARED,    \
                        BUF_GET_OPT_SLEEP   |   \
                            BUF_GET_OPT_LOOKUP, \
                        0)

// If not found in cache return NULL vs. reading it in.
#define O_LOOKUP_EXCL(_inode_,_page_)           \
    fs_bufGetPage(    context,                  \
                        (_inode_),              \
                        (_page_),               \
                        BUF_GET_MODE_EXCL,      \
                        BUF_GET_OPT_SLEEP   |   \
                            BUF_GET_OPT_LOOKUP, \
                        0)

// Get meta buffer with shared lock, sleep if necessary.
#define O_GET_SHARED(_inode_,_page_)            \
    fs_bufGetPage(    context,                  \
                        (_inode_),              \
                        (_page_),               \
                        BUF_GET_MODE_SHARED,    \
                        BUF_GET_OPT_SLEEP,      \
                        FS_BLOCK_SIZE)

// Get meta buffer with exclusive lock, sleep if necessary.
#define O_GET_EXCL(_inode_,_page_)              \
    fs_bufGetPage(    context,                  \
                        (_inode_),              \
                        (_page_),               \
                        BUF_GET_MODE_EXCL,      \
                        BUF_GET_OPT_SLEEP,      \
                        FS_BLOCK_SIZE)

// Create meta buffer and hold exclusive lock, sleep if necessary.
// Previous contents of the block are lost. A block number will be
// allocated prior to returning unless BUF_GET_OPT_DEFER_ALLOC is set,
// in which case the allocation will be deferred until the buf is written
// out to disk.
#define O_CREATE(_inode_,_page_)                \
    fs_bufGetPage(    context,                  \
                        (_inode_),              \
                        (_page_),               \
                        BUF_GET_MODE_EXCL,      \
                        BUF_GET_OPT_SLEEP |     \
                            BUF_GET_OPT_NEW,    \
                        0)

#define O_CREATE_SHARED(_inode_,_page_)             \
    fs_bufGetPage(    context,                      \
                        (_inode_),                  \
                        (_page_),                   \
                        BUF_GET_MODE_SHARED,        \
                        BUF_GET_OPT_SLEEP |         \
                            BUF_GET_OPT_NEW,        \
                        0)

// If not found in cache return NULL vs. reading it in.
#define O_USER_LOOKUP_SHARED(_inode_,_page_)        \
    fs_bufGetPage(    context,                      \
                        (_inode_),                  \
                        (_page_),                   \
                        BUF_GET_MODE_SHARED,        \
                        BUF_GET_OPT_SLEEP   |       \
                            BUF_GET_OPT_LOOKUP  |   \
                            BUF_GET_OPT_USER,       \
                        0)

// If not found in cache return NULL vs. reading it in.
#define O_USER_LOOKUP_EXCL(_inode_,_page_)          \
    fs_bufGetPage(    context,                      \
                        (_inode_),                  \
                        (_page_),                   \
                        BUF_GET_MODE_EXCL,          \
                        BUF_GET_OPT_SLEEP   |       \
                            BUF_GET_OPT_LOOKUP  |   \
                            BUF_GET_OPT_USER,       \
                        0)

// Get user buffer with shared lock, sleep if necessary.
#define O_USER_GET_SHARED(_inode_,_page_,_readLen_) \
    fs_bufGetPage(    context,                      \
                        (_inode_),                  \
                        (_page_),                   \
                        BUF_GET_MODE_SHARED,        \
                        BUF_GET_OPT_SLEEP |         \
                            BUF_GET_OPT_USER,       \
                        (_readLen_))

// Get user buffer with exclusive lock, sleep if necessary.
#define O_USER_GET_EXCL(_inode_,_page_,_readLen_)   \
    fs_bufGetPage(    context,                      \
                        (_inode_),                  \
                        (_page_),                   \
                        BUF_GET_MODE_EXCL,          \
                        BUF_GET_OPT_SLEEP |         \
                            BUF_GET_OPT_USER,       \
                        (_readLen_))

// Create user buffer and hold exclusive lock, sleep if necessary.
#define O_USER_CREATE(_inode_,_page_)                   \
    fs_bufGetPage(    context,                          \
                        (_inode_),                      \
                        (_page_),                       \
                        BUF_GET_MODE_EXCL,              \
                        BUF_GET_OPT_SLEEP |             \
                            BUF_GET_OPT_NEW |           \
                            BUF_GET_OPT_USER |          \
                            BUF_GET_OPT_DEFER_ALLOC,    \
                        0)

// If not found in cache return NULL vs. reading it in.
#define B_LOOKUP_SHARED(_inode_, _block_)                       \
    fs_bufGetBlock( context,                                    \
                    (_inode_),                                  \
                    (_block_),                                  \
                    BUF_GET_MODE_SHARED,                        \
                    BUF_GET_OPT_SLEEP |                         \
                        BUF_GET_OPT_LOOKUP)

// If not found in cache return NULL vs. reading it in.
#define B_LOOKUP_EXCL(_inode_, _block_)                         \
    fs_bufGetBlock( context,                                    \
                    (_inode_),                                  \
                    (_block_),                                  \
                    BUF_GET_MODE_EXCL,                          \
                    BUF_GET_OPT_SLEEP |                         \
                        BUF_GET_OPT_LOOKUP)

// Get meta buffer with shared lock, sleep if necessary.
#define B_GET_SHARED(_inode_, _block_)                          \
    fs_bufGetBlock( context,                                    \
                    (_inode_),                                  \
                    (_block_),                                  \
                    BUF_GET_MODE_SHARED,                        \
                    BUF_GET_OPT_SLEEP)

// Get meta buffer with exclusive lock, sleep if necessary.
#define B_GET_EXCL(_inode_,_block_)                             \
    fs_bufGetBlock( context,                                    \
                    (_inode_),                                  \
                    (_block_),                                  \
                    BUF_GET_MODE_EXCL,                          \
                    BUF_GET_OPT_SLEEP)

// Create meta buffer and hold exclusive lock, sleep if necessary.
// Previous contents of the block are lost.  Because the block is
// FS_INVALID_BLOCK, a block will be allocated prior to returning.
#define B_CREATE(_inode_)                                       \
    fs_bufGetBlock( context,                                    \
                    (_inode_),                                  \
                    FS_INVALID_BLOCK,                           \
                    BUF_GET_MODE_EXCL,                          \
                    BUF_GET_OPT_SLEEP |                         \
                        BUF_GET_OPT_NEW)

// Get user buffer with exclusive lock
// If not found in cache return NULL vs. reading it in.
#define B_USER_LOOKUP_EXCL(_inode_,_blknum_)                    \
    fs_bufGetBlock( context,                                    \
                    (_inode_),                                  \
                    (_blknum_),                                 \
                    BUF_GET_MODE_EXCL,                          \
                    BUF_GET_OPT_SLEEP |                         \
                        BUF_GET_OPT_LOOKUP |                    \
                        BUF_GET_OPT_USER)

// Get user buffer with shared lock, sleep if necessary.
#define B_USER_GET_SHARED(_inode_,_blknum_)                     \
    fs_bufGetBlock( context,                                    \
                    (_inode_),                                  \
                    (_blknum_),                                 \
                    BUF_GET_MODE_SHARED,                        \
                    BUF_GET_OPT_SLEEP |                         \
                        BUF_GET_OPT_USER)

// Get user buffer with exclusive lock, sleep if necessary.
#define B_USER_GET_EXCL(_inode_,_blknum_)                       \
    fs_bufGetBlock( context,                                    \
                    (_inode_),                                  \
                    (_blknum_),                                 \
                    BUF_GET_MODE_EXCL,                          \
                    BUF_GET_OPT_SLEEP |                         \
                        BUF_GET_OPT_USER)

// Create user buffer and hold exclusive lock, sleep if necessary.
// Previous contents of the block are lost.  Because the block is
// FS_INVALID_BLOCK, a block will be allocated prior to returning.
#define B_USER_CREATE(_iptr)                                    \
    fs_bufGetBlock( context,                                    \
                    (_inode_),                                   \
                    FS_INVALID_BLOCK,                           \
                    BUF_GET_MODE_EXCL,                          \
                    BUF_GET_OPT_SLEEP |                         \
                        BUF_GET_OPT_NEW |                       \
                        BUF_GET_OPT_USER)

#define B_FREE(_bufp_)      \
    fs_bufFree(context, (_bufp_), BUF_DO_DEALLOCATE);

#define FS_BUF_WRITE(_bufp_,_offset_,_length_)          \
    fs_bufWrite(    context,                            \
                    (_bufp_),                           \
                    (_offset_),                         \
                    (_length_),                         \
                    FS_BUF_WRITE_LOG |                  \
                        FS_BUF_LOG_ADD_BUF,             \
                    NULL,								\
                    NULL)

#define FS_BUF_WRITE_NO_LOG(_bufp_,_offset_,_length_)   \
    fs_bufWrite(    context,                            \
                    (_bufp_),                           \
                    (_offset_),                         \
                    (_length_),                         \
                    0,                                  \
                    NULL,								\
                    NULL)

#define FS_BUF_WRITE_ADD_LOCK(_bufp_,_offset_,_length_) \
    fs_bufWrite(    context,                            \
                    (_bufp_),                           \
                    (_offset_),                         \
                    (_length_),                         \
                    FS_BUF_LOG_ADD_BUF,                 \
                    NULL,                               \
                    NULL)


// This doesn't actually write the buf.  Its analogous to
// FS_BUF_WRITE, but instead records a compact zero-fill log
// entry.
#define FS_BUF_ZERO_ALL(_bufp_)                         \
    fs_bufWrite(    context,                            \
                    (_bufp_),                           \
                    0,                                  \
                    FS_BLOCK_SIZE,                      \
                    FS_BUF_WRITE_LOG |                  \
                        FS_BUF_WRITE_ZERO |             \
                        FS_BUF_LOG_ADD_BUF,             \
                    NULL,								\
                    NULL)

// This doesn't actually write the buf.  Its analogous to
// FS_BUF_WRITE, but instead records a compact zero-fill log
// entry.
#define FS_BUF_ZERO_PART(_bufp_,_offset_,_len_)         \
    fs_bufWrite(    context,                            \
                    (_bufp_),                           \
                    (_offset_),                         \
                    (_len_),                            \
                    FS_BUF_WRITE_LOG |                  \
                        FS_BUF_WRITE_ZERO |             \
                        FS_BUF_LOG_ADD_BUF,             \
                    NULL,								\
                    NULL)

/**
 * Initialize the writeback buffer lists.
 */
extern void fs_bufInitWritebackLists(
    fs_FileSystem *fileSystem);


/**
 * Return the pageHdr for the specified buf.
 */
extern fs_BufPageHdr *
fs_bufPageHdrForInum(
    fs_FileSystem   *fileSystem,
    int32           bufListType,
    uint64          inum);

/**
 * Return true if the buf has been dirtied by the current transaction.
 */
extern boolean
fs_bufWasModified(
    fs_Context  *context,
    fs_Buf      *buf);

extern void
fs_bufDirtyForList(
    fs_Context  *context,
    fs_Buf      *buf,
    fs_BufList  *bufList,
    uint32      bufWriteFlags);

void
fs_bufAllocForCow(
	fs_Context	*context,
	fs_Buf		*buf);

int32 
fs_bufHashVolBlock(
    evm_volId_t volId,
    fs_Block    block);
int32
fs_bufHashVolInum(
    evm_volId_t volId,
    uint64      inum);
int32
fs_bufHashVolInumPage(
    evm_volId_t volId,
    uint64      inum,
    uint64      page);

/**
 * Initialize the buffer cache.
 */
extern void
fs_bufCacheInit();


/**
 * Increment the log buffer reference
 * count on the specified buf.
 */
extern void
fs_bufLogRefIncr(
    fs_Context  *context,
    fs_Buf      *buf);


/**
 * Decrement the log buffer reference
 * count on the specified buf.
 */
extern void
fs_bufLogRefDecr(
    fs_Context  *context,
    fs_Buf      *buf);

/**
 * Search for a buffer with the specified inum and page.
 * Must be called with the buflist spinlock held.
 * Return NULL if the buffer is not found.
 */
extern fs_Buf *
fs_bufFindByPage(
    fs_Context *context,
    uint64     inum,
    int32      snapnum,
    fs_BufList *bufList,
    uint64     page);


#ifdef FS_DEBUG_BUF_CKSUM

/*
 * Release checksum for a given block
 */

extern  void
fs_bufReleaseCksumForBlock( fs_Context  *context, fs_Block    block);

/**
 * Release all bufCksum struct for given filesystem
 */
extern void
fs_bufReleaseAllCksum(
    fs_Context  *context);


/**
 * Compute and store (in-core only) a checksum over
 * the meta buf.
 */
extern void
fs_bufComputeCksum(
    fs_Context  *context,
    fs_Block    block,
    void        *dataPtr);

/**
 * Compare the previously stored (in-core) cksum,
 * if one exists, to the just read meta buf.
 */
extern void
fs_bufCompareCksum(
    fs_Context  *context,
    fs_Block    block,
    void        *dataPtr,
    fs_Buf      *buf,
    char        *msg,
    boolean     clearBusy);

#endif // FS_DEBUG_BUF_CKSUM


/**
 * Allocate for a list of buffers.
 */
void
fs_bufAllocDoBufs(
    fs_Context *context,
    fs_Inode   *inode,
    int        numInBufs,
    fs_Buf     **bufs,
	boolean    commitInode);


#define FS_MAX_READ_CLUSTER 64  // Maximum clustering for read ahead

/*
 * Asynchronously read ahead the pages
 */

extern int32
fs_bufGetPageAsync(
    fs_Context      *context,
    fs_Inode        *inode,
    uint64          page,
    uint32          len,
    uint32          type);
extern int32
fs_bufGetBlockAsync(
    fs_Context      *context,
    fs_Inode        *inode,
    uint64          blk,
    uint32          len,
    uint32          type);

extern void
fs_bufLookupBlock(fs_Context *  context,
                  uint64        inum,
                  uint64        page,
                  fs_Block      block,
                  uint32        type,
                  boolean       sync);

/**
 * Start IO writeback on the specified inode's dirty
 * pages if the number of dirty pages exceeds a
 * threshold.
 */
extern void
fs_bufStartIoThreshold(
    fs_Context  *context,
    fs_Inode    *inode,
    int32       bufListType);


/**
 * Flush the buffers that have been deallocated since reallocation
 * of the same block will produce "duplicate cache insert".
 * Called during snapshot remove for blocks that are deallocated
 * (vs flushing entire buffer cache).
 */
extern void
fs_bufCacheFlushBlocks(
    fs_Context *context,
    uint64 startBlock,
    uint32 len);


/**
 * Enable COW mechanism after a snapshot operation (create).
 */
extern void fs_bufCacheEnableCow(
    fs_Context *context);


/**
 * Remove a cow buffer from its old cache location and re-insert it
 */
extern void
fs_bufCopyOnWriteReInsertBuf(
    fs_Context  *context,
    fs_Buf      *buf,
    fs_Block    oldBlock);

/**
 * Set the buf onto the dirty list for the specified
 * log segment.
 */
extern void
fs_bufSetDirtySegment(
    fs_Context  *context,
    void        *obj,
    int32       segNum,
	int32		offset);


/**
 * Convert any cando records to permanent, and return
 * the number of records converted.
 */
int32
fs_bufConvertCando(
    fs_Context  *context,
    fs_Buf      *buf);


/**
 * Return meta cache itype for buf.
 */
inline int32
fs_bufGetItype(
    fs_Buf  *buf);


/**
 * Update the cache stats to reflect the specified buf
 * expanding from a compact leaf form to full sized.
 */
void
fs_bufLeafExpanded(
    fs_Context  *context,
    fs_Buf      *buf);


/**
 * Update the cache stats to reflect the specified buf
 * shrinking from full sized to a compact leaf form.
 */
void
fs_bufLeafShrunk(
    fs_Context  *context,
    fs_Buf      *buf,
    boolean     isDir);


/**
 * Return TRUE if the specified buf's data area is
 * a compact allocation.
 */
inline boolean
fs_bufIsCompact(
    fs_Buf  *buf);


/**
 * find who consumes most memory in the cache
 * and free the resources
 *
 * Caller indicates which spinlock it owns.
 */
void
fs_bufFindAndFreeResource(
    fs_Context          *context,
    int                 haveMetaSpinLock,
    int                 haveUserSpinLock,
    int                 haveInodeSpinLock,
    int32               reduceCount);


/**
 * Make a copy-on-write copy of the block at the specified number.
 */
extern uint64 fs_make_block_copy(
        fs_Context *context,
        fs_Inode *inode,
        uint64 blknum);

/**
 * Set the upper limit on the cache sizes.
 */
extern void
fs_bufSetCacheLimits();

#endif /* FS_BUF_H */
