/*
 * Name:  fs-context.h
 *
 * RCS:  $Id: fs-context.h,v 1.64 2002/12/10 17:27:07 bob Exp $
 *
 * Description:  Definitions for thread/context related stuff.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#if !defined(FS_CONTEXT_H)
#define FS_CONTEXT_H


/* 
 * Initial size of the shared locks list attached to each context.
 * Will grow when needed.
 */ 
enum { SHARED_LOCK_LIST_INIT = 16 };


/**
 * For collecting CPU yield/run stats.
 */
typedef struct fs_RecvReqStats
{
    uint64  num[FS_REQ_MAX];
} fs_RecvReqStats;

typedef struct fs_ReplyStats
{
    uint64  ok[FS_REQ_MAX];
    uint64  err[FS_REQ_MAX];
} fs_ReplyStats;

typedef struct fs_RecvNfsReqStats
{
    uint64  num[FS_REQ_NFS_MAX];
} fs_RecvNfsReqStats;

enum {
    FS_YIELD_TYPE_ON_CPU    = 0,   // Not yielding (ie. running!)
    FS_YIELD_TYPE_IO_UR     = 1,   // Waiting for IO completion, user read
    FS_YIELD_TYPE_IO_UW     = 2,   // Waiting for IO completion, user write
    FS_YIELD_TYPE_IO_MR     = 3,   // Waiting for IO completion, meta read
    FS_YIELD_TYPE_IO_MW     = 4,   // Waiting for IO completion, meta write
    FS_YIELD_TYPE_IO_LOG    = 5,   // Waiting for IO completion, log (RD + WR)
    FS_YIELD_TYPE_LOCK      = 6,   // Waiting for a lock.
    FS_YIELD_TYPE_TXCOMPL   = 7,   // Waiting for TX_COMPL arrival.
    FS_YIELD_TYPE_COD       = 8,   // Waiting for COD arrival.
    FS_YIELD_TYPE_ADMIN     = 9,   // Misc yields in system.
    FS_YIELD_TYPE_SLEEP     = 10,   // Sleep type.
    FS_YIELD_TYPE_TOTAL     = 11,   // total elapsed time
    FS_YIELD_TYPE_IO_TR     = 12,   // Tape Reads
    FS_YIELD_TYPE_IO_TW     = 13,   // Tape Writes
    FS_YIELD_TYPE_IO_COMMIT = 14,   // Waiting for IO completion, commit
    FS_YIELD_TYPE_REPLAY    = 15,
    FS_YIELD_TYPE_MAX       = 16 
};

typedef struct fs_YieldStats
{
    eeeRTC_t	num[FS_REQ_NFS_MAX + 2 + FS_DAEMON_TYPE_MAX][FS_YIELD_TYPE_MAX];
} fs_YieldStats;

/**
 * Yield timeouts.
 *
 * TODO - reduce these artificially high values once the
 * agile box starts behaving better.
 */
enum {
    FS_YIELD_TIME_FOREVER   = 0,
    FS_YIELD_TIME_IO        = 75,   // EVM retries: 1 retry at 30 secs + our
									// own safety margin.
    FS_YIELD_TIME_LOCK      = 600,  // could be long waits on access lock
                                    // during eek or snapshots
    FS_YIELD_TIME_TXCOMPL   = 30,
    FS_YIELD_TIME_COD       = 30,
    FS_YIELD_TIME_ADMIN     = 60,
};


/**
 * File system thread events.
 */
enum {
    FS_THREAD_EVENT_START                   = 1,
    FS_THREAD_EVENT_IO_DONE                 = 2,
    FS_THREAD_EVENT_LOCK                    = 3,
    FS_THREAD_EVENT_COD_CONFIRMED           = 4,
    FS_THREAD_EVENT_TX_COMPL                = 5,
    FS_THREAD_EVENT_LOG_COMPL               = 6,
    FS_THREAD_EVENT_SYNC                    = 7,
    FS_THREAD_EVENT_THAW                    = 8,
    FS_THREAD_EVENT_DUMP_REST_IOCOMPLETE    = 9,
    FS_THREAD_EVENT_TERMINATE               = 10,
    FS_THREAD_EVENT_EEKCONT                 = 11,
    FS_THREAD_EVENT_EXCEPTION               = 12,
    FS_THREAD_EVENT_YIELD_WAKE              = 13,
    FS_THREAD_EVENT_READ_PROCEED            = 14,
    FS_THREAD_EVENT_YIELD_SLEEP_WAKEUP      = 15,
    FS_THREAD_EVENT_DUMP_FH_COMPLETE        = 16,
    FS_THREAD_EVENT_WORK_DONE               = 17,
    FS_THREAD_EVENT_SCAN                    = 18,
    FS_THREAD_EVENT_REPLAY                  = 19,
#ifdef FS_HIST_WAIT
    FS_THREAD_EVENT_SEND_HIST               = 20
#endif
};


struct fs_ReqTypeTbl;

/**
 * File system request context.
 *
 * This provides space for holding global context related
 * to the processing of a specific filesystem request.
 * It ties together the request message, the request type
 * specific state structure, the response message, and bits
 * and pieces that help avoid passing many arguments to
 * functions during request processing.
 */
struct fs_Context
{
    evm_volId_t     volId;              // Volume this request is for.
    esm_handle_t    smHandle;           // State machine handle.
    fs_FileSystem   *fileSystem;        // In-core filesystem struct
    const fs_MsgReqHdr *reqHdr;            // Original request msg.
	const struct fs_ReqTypeTbl * const reqType;
	uint32          srcAddr;            // Packed src address of requestor.
    int32           daemonType;         // For threadFork'd threads.
    fs_MsgRespHdr   *respHdr;           // Ptr to response msg area.
    int32           respSize;           // Actual size of response msg to
                                        // send.
    int32           initRespSize;       // DEBUG - TODO remove this
    void            *reqInfo;           // Req type dependent
                                        // processing info.
    fs_LogSummary   *logSummary;        // In-core summary info for log entry.
    boolean         isThread;           // This context is for an ESM thread.
    DLL_ENTRY(fs_Context)
                    fsEntry;            // Entry for the fs list of 
                                        // system (not user) threads.
    fs_LockWait     lockWait;           // Structure for adding this context
                                        // to a lock wait list.
    fs_Arena        *arena;             // Temporary memory allocation arena.
    uint64          randomSeed;         // state for random number generator
    eeeRTC_t		startTime;          // Time we started this request.
    eeeRTC_t		tmpTime;            // Timestamp for tracking yield times.
    boolean         use83Names;         // Client uses 8.3 filenames.
    boolean         caseInsensitive;    // Client ignores case in filenames.
    boolean         hashCollides;       // Last dir lookup encountered name
                                        // hash collision - used by next addent
    uint16          flags;              // (See below.)
    fs_Eek  	    eek;

    fs_ExtentVector *pendingFreeExtents; // extents to free at commit time

    /** Certain contexts (eg. wb daemon) can avoid lock order checks */
    boolean         ignoreLockOrder;

    /** The list of locks in the chain of owned locks. */
    DLL_HDR(fs_Lock) ownerList;
    
    /** list of locks and acquisition counts for shared locks. */
    fs_SharedLockRef *sharedLocksPtr;
    int               sharedLockCount;    //nb of distinct shared locks held
    int               sharedLockSize;     //current list size
#ifdef FS_DEBUG_LOCK
    fs_SharedLockRef  sharedLocks[SHARED_LOCK_LIST_INIT];     
#endif

    /** To protect from concurrent access to the owned and shared lock list. */
    nfx_spinlock_t  lockListSpin;

#ifdef FS_DEBUG_LOCK_ORDER
    /** Highest held lock in lockorder heirarchy */
    int32           lockHeight;

    /** Are we holding a leaf lock? */
    void			*leafHeld;

	/** Highest held reserved inode in lockorderheirarchy */
	uint32			maxReservedInodeHeight;

	/** Parent inum of currently requested inode, if known */
	uint64			lockOrderParentInum;
#endif  // FS_DEBUG_LOCK_ORDER
    void            *extraArg;  // for thread fork
};


/**
 * Flags for contexts.
 */
enum fs_ContextFlags {
    /** Request buffer lives in shared memory. */
    FS_CONTEXT_SHARED_REQUEST = 0x1,
    /** Response buffer lives in shared memory. */
    FS_CONTEXT_SHARED_RESPONSE = 0x2,
	/** No space condition exists when this request was accepted */
	FS_CONTEXT_ENOSPC_CONDITION = 0x4,
    /** Context acquired the file system modify lock. */
    FS_CONTEXT_MODIFY_LOCK_ACQUIRED = 0x8,
    /** Context should acquire the file system modify lock. */
    FS_CONTEXT_MODIFY_LOCK_NEEDED = 0x10,
    /** This context has incremented the nbrOngoingTransactions counter */
    FS_CONTEXT_TRANSACTION_COUNTED = 0x020,
    /** Context acquired the file system access lock. */
    FS_CONTEXT_ACCESS_LOCK_ACQUIRED = 0x40,
    /** Context allow lock access starvation */
    FS_CONTEXT_ALLOW_LOCK_STARVATION = 0x80,
    /** Context should acquire the file system access lock. */
    FS_CONTEXT_ACCESS_LOCK_NEEDED = 0x100,
    /** Disallow exclusive locks to time out for locks held by this context **/
    FS_CONTEXT_DISALLOW_EXCL_LOCK_TIMEOUT = 0x200
 };


/**
 * Create the in-core structure for a given file system, which is specified
 * by the given context's volume.
 */
extern fs_FileSystem *fs_makeFileSystem(
    fs_Context    *context,
    fs_SuperBlock *super);


/**
 * Nfs statistics
 */
enum {
    FS_NFS_TIME_1MSEC     = 0,
    FS_NFS_TIME_3MSEC     = 1,
    FS_NFS_TIME_5MSEC     = 2,
    FS_NFS_TIME_7MSEC     = 3,
    FS_NFS_TIME_10MSEC    = 4,
    FS_NFS_TIME_15MSEC    = 5,
    FS_NFS_TIME_GT15MSEC  = 6,
    FS_NFS_TIME_MAX       = 14
};

typedef struct fs_NfsStats
{
#ifdef FS_DEBUG_NFS_DELTA_BUCKET
    uint64  num[FS_REQ_NFS_MAX][FS_NFS_TIME_MAX];
#endif  // FS_DEBUG_NFS_DELTA_BUCKET
    uint32  errs[FS_REQ_NFS_MAX];
} fs_NfsStats;


/*
 * Request type flags. 
 */
enum {
    FS_REQ_FLAGS_LOG            = 0x0001,   // Log the operation.
    FS_REQ_FLAGS_ATTRS          = 0x0002,   // Alloc context space for attrs
                                            // indicated in request message.
    FS_REQ_FLAGS_CLIENT         = 0x0004,   // Request is from an fs client.
    FS_REQ_FLAGS_SPACE          = 0x0008,   // Request may use disk space.
    FS_REQ_FLAGS_MODIFY         = FS_CONTEXT_MODIFY_LOCK_NEEDED,
											// Obtain the global modify lock in
                                            // shared mode.
    FS_REQ_FLAGS_INODE          = 0x0020,   // Request may use an inode.
    FS_REQ_FLAGS_NO_FS_OK       = 0x0040,   // Request doesn't require a
                                            // mounted filesystem.
    FS_REQ_FLAGS_STARV          = FS_CONTEXT_ALLOW_LOCK_STARVATION,
											// Request allows lock starvation
                                            // (do not keep track of locks).
    FS_REQ_FLAGS_NO_ACCESS_LOCK = FS_CONTEXT_ACCESS_LOCK_NEEDED,
											// Request doesn't require
                                            // the fileSystem access lock.
    FS_REQ_FLAGS_NO_EXCL_LOCK_TIMEOUT = FS_CONTEXT_DISALLOW_EXCL_LOCK_TIMEOUT,
                                    // Disallow lock timeout on locks
                                    // held in exclusive mode by this context
};

/*-----------------------------------------------------------------------------
 * Request type table.  Use FS_REQ_* to index into this table
 * to obtain reqType specific defintions.
 */
typedef struct fs_ReqTypeTbl
{
    int32       (* instantiateFunc)             // Function to instantiate.
					(const struct fs_ReqTypeTbl *reqType,
					 const fs_MsgReqHdr   *reqHdr,
                     int32          bufType,
                     uint32  		srcAddr);
    evm_volId_t (* extractVolIdFunc)            // Function to extract volId.
                    (const fs_MsgReqHdr   *reqHdr);
    int32       (* threadFunc)
                    (fs_Context     *context);  // Function to perform request
                                                // operation for those that
                                                // are run as a thread.
    void        (* respFunc)                    // Function for normal
                    (fs_Context     *context,   // response.
                     int32          status);
    void        (* ncRespFunc)                  // Function for NULL context
                    (const fs_MsgReqHdr   *reqHdr,    // response.
                     int32          bufType,
                     uint32  		srcAddr,
                     fs_Status      status);
    void        (* clnupFunc)                   // Function for context
                    (fs_Context     *context,
                     int	       doLockTerm);// cleanup.
    char        *name;                          // Request operation name.
    short       reqInfoSize;                    // Additional context alloc
                                                // size.
    short       respAreaSize;                   // Size to prealloc for
                                                // response message.
    int32       flags;                          // Request type flags, see
                                                // above.
} fs_ReqTypeTbl;

extern const fs_ReqTypeTbl fs_reqTypeTbl[FS_REQ_MAX];

extern fs_NfsStats     fs_nfsStats;
extern uint32          fs_currentActiveReqs;
extern fs_YieldStats   fs_yieldStats;
extern fs_RecvReqStats fs_recvReqStats;
extern fs_ReplyStats   fs_replyStats;
extern nfx_spinlock_t  fs_statsReqRcvdSpinId;
extern int32           fs_nbrOfContexts;
extern int32           fs_maxNbrOfContexts;
extern uint32          fs_maxActiveReqs;

extern int32           fs_nfsStatsUsecs[FS_NFS_TIME_MAX];

/**
 * Instantiation function for threads.
 */
extern int32 fs_threadInstantiate(
	const fs_ReqTypeTbl	*reqType,
    const fs_MsgReqHdr        *reqHdr,
    int32               bufType,
    uint32		        srcAddr);

enum
{
    // Start the thread with the message processing function
    // from the fs_reqTypeTbl.
    FS_THREAD_FS_REQ_MSG = 1,

    // Start the thread with the function passed in explicitly.
    FS_THREAD_GENERIC
};

/**
 * Handle a received message for the filesystem.  It expects a flat
 * contiguous buffer holding one of the messages defined in our API.
 */
extern void fs_receiveMsg(
    void            *msgBuf,
    uint32			src,
    int32           msgLen,
    int32           bufType);


/**
 * Instantiation functions call this to setup the context block for a
 * filesystem request.
 */
extern int32 fs_contextInit(
	const fs_ReqTypeTbl	*reqType,
    const fs_MsgReqHdr    *reqHdr,
    int32           bufType,
    uint32			srcAddr,
    fs_Context      **context);


/**
 * Send the response for the thread if applicable.
 */
void
fs_threadRespond(
    fs_Context      *context,
    int32           retStatus);


/**
 * Send an error back to the requestor.  This is used when we fail
 * to create a context for handling the request (typically due to
 * the volume not being available).
 */
void
fs_nullContextResponse(
    const fs_MsgReqHdr    *reqHdr,
    int32           bufType,
    uint32			srcAddr,
    fs_Status       status);


/**
 * Send ndmp response from context.
 */
void
fs_ndmpResponse(
    fs_Context  *context,
    int32       status);

/**
 * Send an error back to the ndmp requestor.  This is used when we fail
 * to create a context for handling the request (typically due to
 * the volume not being available).
 */
void
fs_nullContextSendNdmpResponse(
    const fs_MsgReqHdr    *reqHdr,
    int32           bufType,
    uint32          srcAddr,
    fs_Status       status);


/**
 * Print the contents of an fs_Context.
 */
void
fs_contextDump(
    fs_Context  *context);

/**
 * Wake up a daemon with the given event, setting a flag value to 1 in
 * the process.  Returns TRUE of the daemon was signalled, false otherwise.
 */
boolean
fs_daemonWakeup(
    fs_Context *context,
    nfx_spinlock_t *spin,
    esm_handle_t handle,
    int32 *flag,
    esm_event_t event);


/**
 * Notify the specified waiter that its wait is over.
 */
void
fs_notifyWaiter(
    fs_Context  *context,
    void        *waitList,
    fs_Waiter   *waiter);


/**
 * Initialize a handle for the specified file in the filesystem in the
 * specified context.
 */
void
fs_handleInit(
    fs_FileSystem   *fileSystem,
    fs_Handle       *handle,
    uint8           snap,
    uint64          inum,
    uint32          gen,
    uint64          mountPointInum);


//  A P P L I C A T I O N   I D S

extern int32 fs_applicationId;
extern int32 fs_srcPortId;
extern int32 fs_efsSscAppId;
extern int32 fs_nfsAppId;
extern int32 fs_mcpuAppId;
extern int32 fs_mcpuDestPortId;


/**
 * Create a new thread, for the same file system as the specified context,
 * running the specified start function.  Returns the new thread's handle.
 */
extern esm_handle_t
fs_threadFork(
    fs_Context  *context,
    void        (*func)(esm_handle_t, void *),
    boolean     addToActiveList,
    boolean     allowStarvation,
    int32       daemonType,
    void        *extraArg
);

/**
 * Terminate a thread created with fs_threadFork.
 */
extern void
fs_threadTerminate(
    fs_Context  *context
);


/**
 * Default cleanup function for thread contexts.
 */
void
fs_threadContextDestroy(
    fs_Context      *context,
    int		         doLockTerm
);


/**
 * Cleanup function for NDMP related thread contexts.
 */
void
fs_ndmpContextDestroy(
    fs_Context  *context
);


/**
 * Free the response buffer for the specified context.  If the response
 * has already been freed, this is a no-op.
 */
void
fs_contextFreeResponse(
    fs_Context      *context);


/**
 * Yield the processor.  If event is supplied, event.type will be checked
 * against the wakup event when delivered.  Time (in seconds) is passed through
 * to esm_yield() for timeout warning.  Type indicates what the yield is
 * waiting for for instrumentation purposes only.
 */
extern void
fs_yield(
    fs_Context  *context,
    esm_event_t *event,
    int32       time,
    int32       type);


/**
 * Yield the processor for the specified period.
 */
extern void
fs_sleep(
	fs_Context	*context,
	int32		msecs);

extern void
fs_threadContextDtor(
	void *);

extern void
fs_threadContextCtor(
	void *,
	uint32);

extern void
fs_threadContextReinit(
	fs_Context *);
	
/**
 * Handle the lookup request.  This function first tries the
 * dnlc, and if it fails, it creates a thread to do the rest.
 */
extern int32
fs_lookupInstantiate(
	const fs_ReqTypeTbl *reqType,
    const fs_MsgReqHdr  *reqHdr,
    int32               bufType,
    uint32              srcAddr);

#endif /* FS_CONTEXT_H */
