/*
 * Name:    fs-dirhash.c
 *
 * Copyright (C) 2002, Agile Storage Inc.
 *
 * RCS:     $Id: fs-dirhash.c,v 1.5 2002/12/07 00:46:05 bob Exp $
 *
 * Description:  Directory hash cache.
 */

#include "fs-includes.h"


/*
 * Size a directory must be before "owning" a hash table.
 */
#define MIN_DIRECTORY_HASH_SIZE (PAGE_SIZE)

/*
 * Hash tables are 4k (power of 2 is useful here).
 */
#define TABLE_SIZE 4096

/*
 * The number of entries in a hash table
 */
#define MAX_HASH_ENTRIES (TABLE_SIZE / 2)

/*
 * Threshold before rehashing (includes dead entries)
 */
#define REHASH_THRESHOLD ((3*TABLE_SIZE) / 4)

/*
 * Marker for end of "hash chain"
 */
#define SENTINEL	0xff

/*
 * The number of directory entries on a single page.
 */
#define ENTRIES_PER_PAGE (PAGE_SIZE / sizeof (fs_Dirent))
NFX_ASSERT_STATIC(ENTRIES_PER_PAGE == 204);

/*
 * The number of entries forced into a single cluster; determined
 * by the size of the hash table address space.
 */
#define ENTRIES_PER_CLUSTER 4

/*
 * The number of directory pages which fall into one
 * hash table.
 */
#define PAGES_PER_TABLE 5

#define CLUSTERS_PER_PAGE (ENTRIES_PER_PAGE/ENTRIES_PER_CLUSTER)
NFX_ASSERT_STATIC(CLUSTERS_PER_PAGE*PAGES_PER_TABLE - 1 < SENTINEL);

#define ENTRIES_PER_TABLE (ENTRIES_PER_PAGE * PAGES_PER_TABLE)

typedef struct fs_Dirhash_table fs_Dirhash_table;
struct fs_Dirhash_table {
	DLL_ENTRY(fs_Dirhash_table) hashEntry;
	DLL_ENTRY(fs_Dirhash_table) mruEntry;
	int32   refcnt;
	uint32  hash;

	evm_volId_t volId;
	uint64 inum;
	uint8 snapnum;
	uint8 isInited;
	uint8 *table;
	int count;
	uint32 genNum;
	fs_Dirhash_table *next;
};

/*
 * TODO: make this tunable/dynamic somehow. Also have a way to
 * evict the cache when memory needs to be reclaimed.
 */
enum {
	FS_DIRHASH_CACHE_PAGE_USE_MAX = 2048,    // Max number of table pages in cache
	FS_DIRHASH_CACHE_HASH_MAX = 512,
	FS_DIRHASH_CACHE_HASH_SHIFT = 9
};

struct {
	nfx_spinlock_t spin;
	DLL_HDR(fs_Dirhash_table) hashHdr[FS_DIRHASH_CACHE_HASH_MAX];
	DLL_HDR(fs_Dirhash_table) mruHdr;
	int32 numUsed;   // number of directories in cache
	int32 pagesUsed; // total number of allocated pages in cache
} fs_dirhash_cache;


/**
 * Initialize the specified hash table from the specified
 * directory.  The hash table has already been appropriately
 * sized for the directory.  This function can block while
 * reading the directory, but the table is left in the un-inited
 * state to prevent other threads from trying to access it or
 * instantiate the same directory in the cache while this is
 * on-going.
 */
static void
hash_directory(fs_Context *context, fs_Inode *inode, fs_Dirhash_table *tpFirst)
{
	fs_Dirhash_table *tp = NULL;
	uint64 pagecount = ceilingDiv64(inode->dinode.size, PAGE_SIZE);
	uint64 page;
	uint8 *table = NULL;
	uint32 cluster = 0;

	for (page = 0; page < pagecount; page++) {
		fs_Buf *buf;
		fs_Dirleaf *d;
		int i;

		if (page % PAGES_PER_TABLE == 0) {
			cluster = 0;
			tp = (tp == NULL) ? tpFirst : tp->next;
			table = tp->table;
			memset(table, SENTINEL, TABLE_SIZE);
			tp->count = 0;
		}

		buf = O_GET_SHARED(inode, page);
		d = buf->u.dirleaf;

		for (i = 0; i < FS_DIRLEAF_ENTRIES_MAX; i++) {
			if (d->u[i].ent.len != 0) {
				uint32 hash = d->u[i].ent.hash % TABLE_SIZE;
				while (table[hash] != SENTINEL)
					hash = (hash + 1) % TABLE_SIZE;
				table[hash] = cluster;
				tp->count++;

                if ((hash = d->u[i].ent.aliasCksumRand) != 0) {
                    hash %= TABLE_SIZE;
				    while (table[hash] != SENTINEL)
					    hash = (hash + 1) % TABLE_SIZE;
				    table[hash] = cluster;
				    tp->count++;
                }
			}
			if ((i%ENTRIES_PER_CLUSTER) == (ENTRIES_PER_CLUSTER-1))
				cluster++;
		}
		fs_bufRelease(context, buf);
	}

	tpFirst->isInited = 1;
}

/*
 * Rehash one table only, from start_page.  This function could
 * block while reading the directory, but the directory is held
 * exclusive, so there is no worry of another context attempting
 * to read the uninitialized table.
 */
static void
rehash(fs_Context *context, fs_Inode *inode, uint64 start_page, fs_Dirhash_table *tp)
{
	uint64 pagecount = ceilingDiv64(inode->dinode.size, PAGE_SIZE);
	uint64 max_page;
	uint64 page;
	uint8 *table;
	uint32 cluster = 0;

	assert(fs_lockOwnedExclusive(context, &inode->lock));

	table = tp->table;

    start_page -= start_page % PAGES_PER_TABLE;
    max_page = start_page + PAGES_PER_TABLE;
	max_page = MIN(max_page, pagecount);

	tp->count = 0;
	memset(table, SENTINEL, TABLE_SIZE);
	for (page = start_page; page < max_page; page++) {
		fs_Buf *buf = O_GET_SHARED(inode, page);
		fs_Dirleaf *d = buf->u.dirleaf;
		int i;

		for (i = 0; i < FS_DIRLEAF_ENTRIES_MAX; i++) {
			if (d->u[i].ent.len != 0) {
				uint32 hash = d->u[i].ent.hash % TABLE_SIZE;
				while (table[hash] != SENTINEL)
					hash = (hash + 1) % TABLE_SIZE;
				table[hash] = cluster;
				tp->count++;

                if ((hash = d->u[i].ent.aliasCksumRand) != 0) {
                    hash %= TABLE_SIZE;
				    while (table[hash] != SENTINEL)
					    hash = (hash + 1) % TABLE_SIZE;
				    table[hash] = cluster;
				    tp->count++;
                }
			}
			if ((i%ENTRIES_PER_CLUSTER) == (ENTRIES_PER_CLUSTER-1))
				cluster++;
		}
		fs_bufRelease(context, buf);
	}
}


/**
 * Allocate and init a new table page.
 */
static fs_Dirhash_table *
alloc_table(fs_Context *context)
{
	fs_Dirhash_table *tp = fs_ramAlloc(FS_RAM_DIRHASH_TP, sizeof *tp);
	tp->table = fs_ramAlloc(FS_RAM_DIRHASH_TABLE, TABLE_SIZE);
	memset(tp->table, SENTINEL, TABLE_SIZE);
	tp->next = NULL;
	tp->count = 0;
	tp->isInited = 0;
	tp->refcnt = 0;
	fs_dirhash_cache.pagesUsed++;
	return tp;
}


/**
 * Deallocate the specified table.
 *
 * Must be called with the spinlock held.
 */
static void
dealloc_table(fs_Dirhash_table *tp)
{
	while (tp != NULL) {
		fs_Dirhash_table *tpnext = tp->next;

		fs_ramDealloc(FS_RAM_DIRHASH_TABLE, tp->table);
		fs_ramDealloc(FS_RAM_DIRHASH_TP, tp);
		tp = tpnext;
		assert(fs_dirhash_cache.pagesUsed > 0);
		--fs_dirhash_cache.pagesUsed;
	}
}


/**
 * Find and return the table for the specified inode if it
 * exists, or instantiate it if necessary.  This call must
 * be followed by a call to dirhash_put_table when the calling
 * context is finished with it.
 */
static fs_Dirhash_table *
fs_dirhash_get_table(
	fs_Context *context,
	fs_Inode *inode,
	boolean isRemove)
{
	uint64 tablecount;
	fs_Dirhash_table *tp;
	uint32 hash;

	hash = context->volId.details.u.parts.clusterUnique +
					  (inode->inum << 8) + inode->snapnum;
	hash += (hash >> FS_DIRHASH_CACHE_HASH_SHIFT);
	hash %= FS_DIRHASH_CACHE_HASH_MAX;

	/*
	 * Look for a matching tp in the chain. If it's found, move
	 * it to the front of its own chain.
	 */
	NFX_SPIN_LOCK(&fs_dirhash_cache.spin);
	DLL_ITERATE(tp, &fs_dirhash_cache.hashHdr[hash], hashEntry)
	{
		assert(tp->hash == hash);

		if (context->volId.whole_thing == tp->volId.whole_thing &&
			inode->inum == tp->inum &&
			inode->snapnum == tp->snapnum)
		{
			if (!tp->isInited) {
				// Another context is currently initing the
				// table for this inode, force this one to
				// the slow path to avoid initialization race.
				NFX_SPIN_UNLOCK(&fs_dirhash_cache.spin);
				FS_ASSERT(!isRemove);
				return NULL;
			}

			FS_ASSERT(tp->genNum == inode->dinode.genNum);

			if (!DLL_IS_ISOLATED(tp, mruEntry)) {
				DLL_REMOVE(&fs_dirhash_cache.mruHdr, tp, mruEntry);
			}
			else {
				assert(tp->refcnt > 0);
			}
			DLL_MOVE_TO_HEAD(&fs_dirhash_cache.hashHdr[hash], tp, hashEntry);
			tp->refcnt += 1;
			NFX_SPIN_UNLOCK(&fs_dirhash_cache.spin);
			return tp;
		}
	}

    // We didn't find it.  If its too small, don't
    // instantiate it.
	if (inode->dinode.size < MIN_DIRECTORY_HASH_SIZE)
    {
		NFX_SPIN_UNLOCK(&fs_dirhash_cache.spin);
		return NULL;
    }

	/*
	 * Evict a directory from the list if necessary.
	 */
	while (fs_dirhash_cache.pagesUsed >= FS_DIRHASH_CACHE_PAGE_USE_MAX)
	{
		DLL_REMOVE_TAIL(&fs_dirhash_cache.mruHdr, tp, mruEntry);
		if (tp == NULL)
		{
			// This could happen if there is a huge table in
			// use on another cpu.
			NFX_SPIN_UNLOCK(&fs_dirhash_cache.spin);
			return NULL;
		}

		if (!DLL_IS_ISOLATED(tp, hashEntry)) {
			DLL_REMOVE(&fs_dirhash_cache.hashHdr[tp->hash], tp, hashEntry);
		}

		assert(fs_dirhash_cache.numUsed > 0);
		--fs_dirhash_cache.numUsed;

		dealloc_table(tp);
	}

	/*
	 * Allocate a new chain of tp's for this directory.
	 */
	tp = NULL;
	tablecount = ceilingDiv64(inode->dinode.size, PAGE_SIZE*PAGES_PER_TABLE);
	do {
		fs_Dirhash_table *new = alloc_table(context);
		new->volId = context->volId;
		new->inum = inode->inum;
		new->snapnum = inode->snapnum;
		new->genNum = inode->dinode.genNum;
		new->next = tp;
		tp = new;
	} while (--tablecount > 0);

	/*
	 * Insert into hash list.
	 */
	DLL_ISOLATE_ELEM(tp, hashEntry);
	DLL_ISOLATE_ELEM(tp, mruEntry);
	tp->hash = hash;
	tp->refcnt = 1;
	DLL_INSERT_HEAD(&fs_dirhash_cache.hashHdr[hash], tp, hashEntry);
	fs_dirhash_cache.numUsed++;
	NFX_SPIN_UNLOCK(&fs_dirhash_cache.spin);

	hash_directory(context, inode, tp);

	return tp;
}


/**
 * Release the specified table, putting onto the MRU free
 * list if the reference count goes to 0.
 */
static void
fs_dirhash_put_table(
	fs_Context *context,
	fs_Dirhash_table *tp)
{
	assert(tp->refcnt > 0);
	NFX_SPIN_LOCK(&fs_dirhash_cache.spin);
	tp->refcnt -= 1;
	if (tp->refcnt == 0) {
		assert(DLL_IS_ISOLATED(tp, mruEntry));
		DLL_INSERT_HEAD(&fs_dirhash_cache.mruHdr, tp, mruEntry);
	}
	NFX_SPIN_UNLOCK(&fs_dirhash_cache.spin);
}


/**
 * Initialize the dirhash cache.
 */
void
fs_dirhash_cache_init()
{
	int i;

	for (i = 0; i < FS_DIRHASH_CACHE_HASH_MAX; i++) {
		DLL_INIT(&fs_dirhash_cache.hashHdr[i]);
	}

	DLL_INIT(&fs_dirhash_cache.mruHdr);
	NFX_SPIN_INIT(&fs_dirhash_cache.spin, SPINLOCK_LEVEL_UNKNOWN);
	fs_dirhash_cache.numUsed = 0;
	fs_dirhash_cache.pagesUsed = 0;
}


/**
 * Lookup the specified entry in the specified directory.
 * If the table doesn't exist, return UNKNOWN, otherwise
 * return FOUND or NOT_FOUND.
 */
int32
fs_dirhash_lookup(
	fs_Context *context,
	fs_Inode *inode,
    fs_DirSearchCb *cb)
{
	fs_Dirhash_table *tpFirst, *tp = fs_dirhash_get_table(context, inode, FALSE);
	uint64 page_base = 0;
	uint32 h;

	if (tp == NULL)
		return FS_DIRHASH_UNKNOWN;

	/*
	 * Cluster numbers in a table are relative to the pages which the table
	 * addresses; hence we keep track of the start page for that table in
	 * a variable called "page_base". It's added to the page number to
	 * arrive at the absolute page number in the directory.
	 *
	 * Similarly, within a page a cluster number is used to arrive at
	 * the starting index for that particular set of directory entries
	 * belonging to that cluster. We call this number "index_base".
	 */
	tpFirst = tp;
	while (tp != NULL) {
		uint8 *table = tp->table;

		for (h = cb->hash % TABLE_SIZE; table[h] != SENTINEL; h = (h + 1) % TABLE_SIZE)
		{
			uint32 page = (table[h] * ENTRIES_PER_CLUSTER) / ENTRIES_PER_PAGE;
			uint32 index_base = (table[h] * ENTRIES_PER_CLUSTER) % ENTRIES_PER_PAGE;
			fs_Buf *buf = O_GET_SHARED(inode, page + page_base);
			fs_Dirleaf *d = buf->u.dirleaf;
			int i;

			for (i = 0; i < ENTRIES_PER_CLUSTER; i++) {
                int direntIndex;

                direntIndex = index_base + i;

                if (d->u[direntIndex].ent.len == 0)
                    continue;

				if (fs_dir_namecmp(context, inode, d, &d->u[direntIndex].ent, cb) == 0)
				{
					cb->page = page + page_base;
					cb->index = direntIndex;
					cb->inum = d->u[direntIndex].ent.inum;
					fs_bufRelease(context, buf);
					fs_dirhash_put_table(context, tpFirst);
					return FS_DIRHASH_FOUND;
				}
			}
			fs_bufRelease(context, buf);
		}

		tp = tp->next;
		page_base += PAGES_PER_TABLE;
	}

	fs_dirhash_put_table(context, tpFirst);
	return FS_DIRHASH_NOT_FOUND;
}


/**
 * Insert the specified entry into the dirhash cache.
 */
void
fs_dirhash_insert(
	fs_Context *context,
	fs_Inode *inode,
	uint64 page,
	int index,
	uint32 hash)
{
	fs_Dirhash_table *tpFirst, *tp = fs_dirhash_get_table(context, inode, FALSE);
	uint64 start_page = page;
	uint64 page_base = page / PAGES_PER_TABLE;
	uint64 cluster;
	uint8 *table;

	if (tp == NULL)
		return;

	/*
	 * We're relying on the fact that the inode is
	 * held exclusive to prevent concurrency issues
	 * with the table, so check it here.
	 */
	assert(fs_lockOwnedExclusive(context, &inode->lock));

	page %= PAGES_PER_TABLE;
	tpFirst = tp;
	while (page_base != 0) {
		/*
		 * When we're extending the directory into a new page,
		 * create a new hash table.
		 */
		if (tp->next == NULL) {
			FS_ASSERT(page_base == 1);
			tp->next = alloc_table(context);
		    tp->next->volId = context->volId;
		    tp->next->inum = inode->inum;
		    tp->next->snapnum = inode->snapnum;
		    tp->next->genNum = inode->dinode.genNum;
			rehash(context, inode, start_page, tp->next);
		}
		tp = tp->next;
		--page_base;
	}

	cluster = (page*ENTRIES_PER_PAGE + index) / ENTRIES_PER_CLUSTER;
	table = tp->table;

	for (hash %= TABLE_SIZE; table[hash] != SENTINEL; hash = (hash + 1) % TABLE_SIZE)
		;

	table[hash] = cluster;

	if (++tp->count > REHASH_THRESHOLD)
		rehash(context, inode, start_page, tp);

	fs_dirhash_put_table(context, tpFirst);
}


/**
 * Remove the specified entry from the dirhash cache.
 */
void
fs_dirhash_remove(
	fs_Context *context,
	fs_Inode *inode,
	uint64 page,
	int index,
	uint32 hash)
{
	fs_Dirhash_table *tpFirst, *tp = fs_dirhash_get_table(context, inode, TRUE);
	uint64 page_base = page / PAGES_PER_TABLE;
	uint8 *table;
	uint64 cluster;

	if (tp == NULL)
		return;

	/*
	 * We're relying on the fact that the inode is
	 * held exclusive to prevent concurrency issues
	 * with the table, so check it here.
	 */
	assert(fs_lockOwnedExclusive(context, &inode->lock));

	page %= PAGES_PER_TABLE;
	tpFirst = tp;
	while (page_base != 0) {
		tp = tp->next;
		FS_ASSERT(tp != NULL);
		--page_base;
	}

	cluster = (page*ENTRIES_PER_PAGE + index) / ENTRIES_PER_CLUSTER;
	table = tp->table;

	for (hash %= TABLE_SIZE; table[hash] != cluster; hash = (hash + 1) % TABLE_SIZE)
		FS_ASSERT(table[hash] != SENTINEL);

	/*
	 * Remove an entry from the hash table only if it
	 * doesn't break the linear probe.
	 */
	if (table[(hash + 1) % TABLE_SIZE] == SENTINEL) {
		table[hash] = SENTINEL;
		--tp->count;
	}

	fs_dirhash_put_table(context, tpFirst);
}


/**
 * Iterate over the directory visiting each
 * entry with a matching hash with iter_fun().
 *
 * Return TRUE if the iteration could be carried
 * out, FALSE otherwise.
 */
boolean
fs_dirhash_iterate(
    fs_Context      *context,
    fs_Inode        *inode,
    fs_DirSearchCb  *cb,
    void            (*iter_fun)(
            fs_Inode    *inode,
            fs_Dirent   *dirent,
            fs_DirSearchCb  *cb))
{
	fs_Dirhash_table *tpFirst, *tp = fs_dirhash_get_table(context, inode, FALSE);
	uint32 page_base = 0;
	uint32 h;
#define EPT_NUM_DWORDS (ENTRIES_PER_TABLE / 64) + 1
#define EPT_NUM_BYTES (ENTRIES_PER_TABLE / 8) + 1
    union {
        uint64  dwords[EPT_NUM_DWORDS];
        uchar8  bytes[EPT_NUM_BYTES];
    } visited;

	if (tp == NULL)
		return FALSE;

	tpFirst = tp;
	while (tp != NULL)
    {
        int dw;
		uint8 *table = tp->table;

        for (dw = 0; dw < EPT_NUM_DWORDS; dw++)
            visited.dwords[dw] = 0;

		for (h = cb->hash % TABLE_SIZE; table[h] != SENTINEL; h = (h + 1) % TABLE_SIZE)
		{
			uint32 page = (table[h] * ENTRIES_PER_CLUSTER) / ENTRIES_PER_PAGE;
			uint32 index_base = (table[h] * ENTRIES_PER_CLUSTER) % ENTRIES_PER_PAGE;
			fs_Buf *buf = O_GET_SHARED(inode, page + page_base);
			fs_Dirleaf *d = buf->u.dirleaf;
			int i;

			for (i = 0; i < ENTRIES_PER_CLUSTER; i++)
            {
                int indexByte, indexBit;

                if (d->u[index_base+i].ent.len == 0)
                    continue;

                // Visit each entry only once.
                indexByte = ((page * ENTRIES_PER_PAGE) + index_base + i) / 8;
                indexBit = 1 << (((page * ENTRIES_PER_PAGE) + index_base + i) % 8);
                if (visited.bytes[indexByte] & indexBit)
                    continue;
                visited.bytes[indexByte] |= indexBit;

				if (fs_dir_hashcmp(&d->u[index_base+i].ent, cb->hash, cb->nameType) == 0)
				{
					cb->page = page_base + page;
					cb->index = index_base + i;
					cb->inum = d->u[index_base+i].ent.inum;

                    iter_fun(inode, &d->u[index_base+i].ent, cb);
				}
			}
			fs_bufRelease(context, buf);
		}

		tp = tp->next;
		page_base += PAGES_PER_TABLE;
	}

	fs_dirhash_put_table(context, tpFirst);
	return TRUE;
#undef EPT_NUM_DWORDS
#undef EPT_NUM_BYTES
}


/**
 * Flush the dirhash cache for the specified directory.
 */
extern void
fs_dirhash_destroy_directory(
	fs_Context *context, fs_Inode *inode)
{
	fs_Dirhash_table *tp = fs_dirhash_get_table(context, inode, TRUE);

    if (tp == NULL)
        return;

	NFX_SPIN_LOCK(&fs_dirhash_cache.spin);
	DLL_REMOVE(&fs_dirhash_cache.hashHdr[tp->hash], tp, hashEntry);
	if (!DLL_IS_ISOLATED(tp, mruEntry))
		DLL_REMOVE(&fs_dirhash_cache.mruHdr, tp, mruEntry);
	--fs_dirhash_cache.numUsed;

	dealloc_table(tp);
	NFX_SPIN_UNLOCK(&fs_dirhash_cache.spin);
}


/**
 * Flush the dirhash cache for the specified filesystem.
 */
extern void
fs_dirhash_destroy_volume(
	fs_FileSystem   *fileSystem)
{
	fs_Dirhash_table *tp, *tpNext;
	int i;

	NFX_SPIN_LOCK(&fs_dirhash_cache.spin);
	for (i = 0; i < FS_DIRHASH_CACHE_HASH_MAX; i++)
	{
		DLL_ITERATE_NEXT(tp, tpNext, &fs_dirhash_cache.hashHdr[i], hashEntry)
		{
			if (fileSystem->volId.whole_thing == tp->volId.whole_thing)
			{
				DLL_REMOVE(&fs_dirhash_cache.hashHdr[i], tp, hashEntry);
				if (!DLL_IS_ISOLATED(tp, mruEntry))
					DLL_REMOVE(&fs_dirhash_cache.mruHdr, tp, mruEntry);
				--fs_dirhash_cache.numUsed;

				dealloc_table(tp);
			}
		}
	}
	NFX_SPIN_UNLOCK(&fs_dirhash_cache.spin);
}


/**
 * Flush the dirhash cache for the specified filesystem,snapnum.
 */
extern void
fs_dirhash_destroy_snapnum(
	fs_FileSystem   *fileSystem,
	int32           snapnum)
{
	fs_Dirhash_table *tp, *tpNext;
	int i;

	NFX_SPIN_LOCK(&fs_dirhash_cache.spin);
	for (i = 0; i < FS_DIRHASH_CACHE_HASH_MAX; i++)
	{
		DLL_ITERATE_NEXT(tp, tpNext, &fs_dirhash_cache.hashHdr[i], hashEntry)
		{
			if (fileSystem->volId.whole_thing == tp->volId.whole_thing &&
				snapnum == tp->snapnum)
			{
				DLL_REMOVE(&fs_dirhash_cache.hashHdr[i], tp, hashEntry);
				if (!DLL_IS_ISOLATED(tp, mruEntry))
					DLL_REMOVE(&fs_dirhash_cache.mruHdr, tp, mruEntry);
				--fs_dirhash_cache.numUsed;

				dealloc_table(tp);
			}
		}
	}
	NFX_SPIN_UNLOCK(&fs_dirhash_cache.spin);
}

