/*
 * Name		: fs-dumprestore.h
 * RCS		: $ID$
 * Description	: Dump/Restore format and related definations
 * Created by	: Rahul Khona
 * Date Created	: 8/10/2001
 */

#ifndef _FS_DUMPRESTORE_H_
#define _FS_DUMPRESTORE_H_


#define FS_DR_TBSIZE        1024        // size of tape record
#define FS_DR_TBSHIFT       10

#define FS_DR_FULL          1           // full dump
#define FS_DR_INCR          2           // incremental dump
#define FS_DR_PARTIAL_FULL  3           // partial full dump
#define FS_DR_PARTIAL_INCR  4           // partial incremental dump (NOT USED)

#define FS_DR_VERSION       2           // format version

#define FS_DR_CHECKSUM      0xfaafbeefeefbfaaf

/*
 *  Operations that need to be done by restore to perform incremental restore
 */

#define FS_DR_CREATE_FILE   1       // create a file
#define FS_DR_CREATE_SYML   2       // create a symlink
#define FS_DR_CREATE_BLOCK  3       // create a block device
#define FS_DR_CREATE_CHAR   4       // create a char device
#define FS_DR_CREATE_SOCK   5       // create a socket
#define FS_DR_CREATE_FIFO   6       // create a named pipe
#define FS_DR_CREATE_DIR    7       // create a dir
#define FS_DR_CREATE_FSET   8       // create a fileset
#define FS_DR_CREATE_LINK   9       // create a hardlink
#define FS_DR_RENAME        10      // rename
#define FS_DR_REMOVE        11      // remove
#define FS_DR_REMOVE_DIR    12      // rmdir
#define FS_DR_DATA          13      // Data for file
#define FS_DR_SET_FQUOTA    14      // set fileset quota
#define FS_DR_SET_UQUOTA    15      // set user quota
#define FS_DR_SET_GQUOTA    16      // set qroup quota
#define FS_DR_DIR_END       17      // End of directory. This is really
                                    // a nop, but used to indicate
                                    // that directory processing is done
#define FS_DR_TAPE_START    18
#define FS_DR_NAME_ONLY     19      // Indicates the record contains
                                    // full pathname only, following
                                    // records would contain data

#define FS_DR_TAPE_END      20

#define FS_DR_TEMP_NAME     1       // Name in this header is to be created
                                    // moved to a temp dir and on completion
                                    // of processing of tape, move it
                                    // to the final destination
#define FS_DR_LINK_TEMP     2       // for links, the target is already
                                    // created as temp
#define FS_DR_PART_PARENT   4       // for directories, is a parent of the
                                    // partial tree dumped
#define FS_DR_83_SOURCE     8       // Object name or source in case
                                    // of rename and target in case of link 
                                    // has 83 name
#define FS_DR_83_TARGET     16      // Rename target or link source has 83 name

typedef struct fs_DRAttr    fs_DRAttr_t;

struct fs_DRAttr {
    union {
        uint64              size;       // size of the file
        struct {
            uint32          major;
            uint32          minor;
        } devNum;
    } sizeDev;
    uint16              mode;       // mode bits         
    uint16              simpleAttrs; // simple attrs         
    fs_Time             ctime;      // last inode change time
    fs_Time             mtime;      // last time file was modified
    fs_Time             atime;      // last time file was accessed
    uint32              uid;        // user id
    uint32              gid;        // group id
    uint32              pad;        // top align to 64 bits boundary
};


#define FS_DR_START 1
#define FS_DR_END   2

typedef struct fs_DumpGenericHeader fs_DumpGenericHeader_t;
struct fs_DumpGenericHeader {
    uint64                  headerId;       // 1 per object
    uint64                  headerSubId;    // If header spans multiple blocks
    uint64                  headerBlocks;   // number of header blocks 
                                            // including this to follow
    uint32                  dataExtents;    // Extents defined in this header
    uint8                   flags;          // Any additional info
                                            // header
    uint8                   op;             // operation to perform
    uint16                  version;        // format version
    fs_DRAttr_t             attr;          // attributes of file
};                          /* 68 bytes */

#define FS_DR_RESERVE    99
#define FS_DR_MAGIC         0xeac1eac1eac1eac1LL
typedef struct fs_DumpStartHeader {
    fs_DumpGenericHeader_t    header;         // generic header
    uint64                  magic;          // magic number
    uint64                  snapshot;       // snapshot name
    uint64                  prevSnapshot;   // previous snapshot if any
    fs_Time                 dumpTime;       // dump time
    fs_Time                 prevDumpTime;   // previous dump time
    uint8                   level;          // dump level
    uint8                   type;           // dump type
    uint8                   pathSeparator;  // path seperator
    uint8                   flags;          // flags if any
    uint64                  volid;          // volid                    
    char                    volname[64];    // volume name if any
    char                    host[32];       // host name/ip addr
    uint64                  maxInodes;      // maximum number of inodes
                                            // used for full dump
    uint64                  checkSum;       // checksum
    uint64                  reserve[FS_DR_RESERVE];
} fs_DumpStartHeader_t;                     /* 1024 bytes */

typedef struct fs_DumpStartHeader  fs_DumpEndHeader_t;

typedef struct fs_DumpExtent {
    uint64                  offset;
    uint64                  len;
} fs_DumpExtent_t;

#define FS_DR_MAXEXTENTS 59
typedef struct fs_DumpDataExtension {
    fs_DumpGenericHeader_t    header;         // header
    fs_DumpExtent_t           extents[FS_DR_MAXEXTENTS];  
} fs_DumpDataExtension_t;

typedef struct fs_DumpOperation {
    fs_DumpGenericHeader_t      header;
    union {
            char                shortNames[22]; // source followed by target
            struct {
                    uint16      sourceSeqId;    // 8.3 seqid for source
                    uint16      targetSeqId;    // 8.3 seqid for target
                    uint32      sourceRand;     // source random num
                    uint32      targetRand;     // source random num
            } shortNameAliases;
    } u;
    uint16                      sourceNameLen;
    uint16                      targetNameLen;
    uint16                      len;        // Total length of name records
                                            // in this header
} fs_DumpOperation_t;

typedef struct fs_DumpHeader {
    union {
        char                dummy[FS_DR_TBSIZE];
        fs_DumpStartHeader_t    start;
        fs_DumpEndHeader_t      end;
        fs_DumpDataExtension_t      extension;
        fs_DumpOperation_t      operation;
    } u;
} fs_DumpHeader_t;
#endif /* _FS_DUMPRESTORE_H_ */
