/*
 * Name		: fs-filehistory.c
 * RCS		: $Id: fs-filehistory.c,v 1.8 2002/11/20 01:04:44 rahulk Exp $
 * Description	: File History implementation for Incremental backup
 * Created by	: Rahul Khona
 * Date Created	: 9/19/2001
 */
#include "fs-includes.h"
#include <exports/ndmp-fs-api.h>


/*
 * Insert a child for
 * a given directory node
 */

void
fs_ndmpAddNode(
        fs_Context              *context,
        fs_DumpContext_t        *dumpContext,
        int64                   parentBlock,
        int64                   parentBoff,
        fs_Inode                *inode,
        fs_Inumber              parent,
        char                    *newName,
        int                     newNameLen,
        int64                   *newBlock,
        uint16                  *newBoff
)
{
        fs_Inode                *nodeFile = NULL;
        fs_Inode                *nameFile = NULL;
        fs_Buf                  *nodeBuf = NULL;
        fs_Buf                  *nameBuf = NULL;
        fs_Buf                  *siblingBuf = NULL;
        fs_Buf                  *buf = NULL;
        int64                   nodeFileSize;
        int64                   nameFileSize;
        uint16                  boff, nBoff;
        uint64                  offset;
        int64                   nameBlock, nodeBlock;
        fs_NdmpHistoryNode_t    *historyNode = NULL, *siblingNode = NULL;
        fs_NdmpHistoryNode_t    *node;
        uint16                  nameLen;
        uint64                  unstableId;

        FS_ASSERT(inode != NULL && (inode->inum == FS_ROOT_INODE ||
                                    (parent != 0)));
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("In fs_ndmpAddNode for inum %lld parent %lld",
                  inode->inum, parent ? parent : -1));
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("In fs_ndmpAddNode for name %s", newName));
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("In fs_ndmpAddNode for parent Block %lld parent boff %d",
                  parentBlock, (int)parentBoff));
        /*
         * Add an entry to the end of the node and names files
         * and update the last child to make the new node as the
         * sibling. Then update the last pointer
         */

        //Get fs modify lock for the transaction
        fs_lockShared(context, &context->fileSystem->modifyLock, NULL);
        
        nodeFile = I_GET_EXCL(0, dumpContext->nodeInum);
        nameFile = I_GET_EXCL(0, dumpContext->nameInum);
        FS_ASSERT(nodeFile != NULL && nameFile != NULL);
        nodeFileSize = nodeFile->dinode.size;
        nameFileSize = nameFile->dinode.size;

        // Begin Transaction
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("ndmpAddNode: begin trasnaction"));
        fs_logBeginTransaction(context);

        /*
         * First get the parent node
         */
        if (inode->inum != FS_ROOT_INODE) {
                FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                         ("ndmpAddNode: Getting parentBlocks for non root"));
                FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                         ("ndmpAddNode: Parent block %lld parent boff %d",
                          parentBlock, (int)parentBoff));
                buf = O_GET_EXCL( nodeFile, parentBlock);
                FS_ASSERT(buf != NULL);
                node = (fs_NdmpHistoryNode_t *)(buf->u.data + parentBoff);
                FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                         ("parent inum %lld inode inum %lld",
                          parent, node->inum));
                FS_ASSERT(node != NULL && node->inum == parent);
        }

        /*
         * First fill in the name in the name file
         */
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("ndmpAddNode: Prepare to fill in name"));

        nBoff = nameFileSize & (FS_BLOCK_SIZE - 1);
        offset = nameFileSize;
        nameBlock = offset >> FS_BLOCK_SHIFT;
        nameLen = newNameLen + 1;
        if ((FS_BLOCK_SIZE - nBoff) > nameLen) {
                FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                         ("ndmpAddNode: name will not fit in buffer"));
                nBoff = 0;
                offset = (offset + FS_BLOCK_SIZE - 1) & (~(FS_BLOCK_SIZE - 1));
                nameBlock = offset >> FS_BLOCK_SHIFT;
                nameBuf = O_GET_EXCL(nameFile, nameBlock);
                if (nameBuf == NULL) {
                        nameBuf = O_CREATE(nameFile, nameBlock);
                }
                FS_ASSERT(nameBuf != NULL);
                bzero(nameBuf->u.data, FS_BLOCK_SIZE);
        } else {
                FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                         ("ndmpAddNode: Get buffer for name"));
                nameBuf = O_GET_EXCL(nameFile, nameBlock);
                if (nameBuf == NULL) {
                        nameBuf = O_CREATE(nameFile, nameBlock);
                }
        }
        bcopy(newName, &nameBuf->u.data[nBoff], nameLen);
        nameFile->dinode.size = (offset + nBoff + nameLen);
        fs_inodeWrite(context, nameFile, INO_WRITE_LOG);

        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("ndmpAddNode: Prepare to write node"));

        offset = nodeFileSize;
        nodeBlock = offset >> FS_BLOCK_SHIFT;
        boff = nodeFileSize & (FS_BLOCK_SIZE - 1);
        if (boff >= FS_MAX_NDMP_BOFF_PER_BLOCK) {

                FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                         ("ndmpAddNode: node does not fit in block"));

                /*
                 * Round up to the next block
                 */
                boff = 0;
                offset = (offset + FS_BLOCK_SIZE - 1) & (~(FS_BLOCK_SIZE - 1));
                nodeBlock = offset >> FS_BLOCK_SHIFT;
                nodeBuf = O_GET_EXCL(nodeFile, offset >> FS_BLOCK_SHIFT);
                if (nodeBuf == NULL) {
                        nodeBuf = O_CREATE(nodeFile, offset >> FS_BLOCK_SHIFT);
                }
                FS_ASSERT(nodeBuf != NULL);
                bzero(nodeBuf->u.data, FS_BLOCK_SIZE);
        } else {
                FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                         ("ndmpAddNode: Get block from disk"));
                nodeBuf = O_GET_EXCL(nodeFile, offset >> FS_BLOCK_SHIFT);
                if (nodeBuf == NULL) {
                        nodeBuf = O_CREATE(nodeFile, offset >> FS_BLOCK_SHIFT);
                }
        }

        historyNode = (fs_NdmpHistoryNode_t *)(nodeBuf->u.data + boff);
        nodeFile->dinode.size = (offset + boff +
                                  sizeof (fs_NdmpHistoryNode_t));

        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("ndmpAddNode: Filling history node"));

        /*
         * Fill the history node
         */

        historyNode->inum = inode->inum;
        //historyNode->size = inode->inum;
        historyNode->size = inode->dinode.size;
        historyNode->uid = inode->sd->u.nfs.uid;
        historyNode->gid = inode->sd->u.nfs.gid;
        historyNode->atime = inode->dinode.atime;
        historyNode->mtime = inode->dinode.mtime;
        historyNode->ctime = inode->dinode.ctime;
        historyNode->linkCnt = inode->dinode.linkCnt;
        //historyNode->mode = inode->dinode.attrs & FS_IMODE_MASK;
        historyNode->mode = inode->dinode.attrs & FS_IFMT_MASK;
        historyNode->mode |= inode->sd->u.nfs.mode & FS_IMODE_MASK;

        /*
         * Fill in the pointers
         */
        historyNode->nameBlock = nameBlock;
        historyNode->nameBoff = nBoff;
        historyNode->nameLen = nameLen;

        historyNode->sibling = FS_NULL_DISKPTR_BLOCK;
        historyNode->firstChildBlock = FS_NULL_DISKPTR_BLOCK;
        historyNode->lastChildBlock = FS_NULL_DISKPTR_BLOCK;
        historyNode->parent = parent;

        if (inode->inum != FS_ROOT_INODE) {
                /*
                 * Now update the parent
                 */


                if (node->lastChildBlock != FS_NULL_DISKPTR_BLOCK) {
                        siblingBuf = O_GET_EXCL(nodeFile, node->lastChildBlock);
                        FS_ASSERT(siblingBuf != NULL);
                        siblingNode =
                            (fs_NdmpHistoryNode_t *)(siblingBuf->u.data +
                                                        node->lastChildBoff);
                        siblingNode->sibling = nodeBlock;
                        siblingNode->siblingBoff = boff;
                } else {
                        node->firstChildBlock = nodeBlock;
                        node->firstChildBoff = boff;
                }
                node->lastChildBlock = nodeBlock;
                node->lastChildBoff = boff;
        }


        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("ndmpAddNode: Writing meta data"));
        fs_inodeWrite(context, nodeFile, INO_WRITE_LOG);
        FS_BUF_WRITE_NO_LOG(nodeBuf, 0, FS_BLOCK_SIZE);
        FS_BUF_WRITE_NO_LOG(nameBuf, 0, FS_BLOCK_SIZE);
        if (inode->inum != FS_ROOT_INODE) {
                FS_BUF_WRITE_NO_LOG(buf, 0, FS_BLOCK_SIZE);
                if (siblingBuf) {
                        FS_BUF_WRITE_NO_LOG(siblingBuf, 0, FS_BLOCK_SIZE);
                }
        }

        // commit transaction. Do it Asynchronously so as to not
        // impact any performance
        unstableId = nodeFile->inum;
        vl_addUnstableId(context, &unstableId);
        fs_logCommit(context, FS_LOG_WRITE_COMMIT_UNSTABLE, unstableId);

        /*
         * Release the buffers as transaction will not release them
         */

        fs_bufReleaseExclusive(context, nameBuf);
        fs_bufReleaseExclusive(context, nodeBuf);
        if (inode->inum != FS_ROOT_INODE) {
                fs_bufReleaseExclusive(context, buf);
                if (siblingBuf) {
                        fs_bufReleaseExclusive(context, siblingBuf);
                }
        }

        //release fs modify lock 
        fs_unlockShared(context, &context->fileSystem->modifyLock);
        
        //Make sure we leave enough space in the log
        fs_makeSureLogIsNotFull(context);

        //Grow the file system if needed
        if (fs_growIfNeeded(context)
         || !fs_haveMininodesAvail(context->fileSystem))    {
             fs_sleep (context, 0);
        }
        
        *newBlock = nodeBlock;
        *newBoff = boff;
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("In fs_ndmpAddNode for inode %lld parent %lld",
                  inode->inum, parent));
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("In fs_ndmpAddNode for node Block %lld node boff %d",
                  nodeBlock, boff));
}

/*
 * Walk the History generalized list and calling the action routine
 * for each of the node. The list is always walked in pre-order.
 */

void
fs_walkHistoryNodes(
        fs_Context          *context,
        fs_DumpContext_t    *dumpContext,
        void                (*func)(fs_Context *,
                                    fs_DumpContext_t *,
                                    fs_Inode *, int64, uint16)
)
{
        fs_NdmpHistoryNodeQueue_t  *entry, *tentry;
        fs_Inode                   *nodeFile;
        fs_Buf                     *nodeBuf;
        fs_NdmpHistoryNode_t       *node;
        int64                      block;
        uint16                     boff;
        fs_NdmpHistoryNodeQueueHead_t   hdr1;
        fs_NdmpHistoryNodeQueueHead_t   *hdr = &hdr1;

        bzero(hdr, sizeof (fs_NdmpHistoryNodeQueueHead_t));
        DLL_INIT(hdr);
        FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                 ("In fs_walkHistoryNodes"));
        nodeFile = I_GET_SHARED(0, dumpContext->nodeInum);
        FS_ASSERT(nodeFile != NULL);

        if (nodeFile->dinode.size == 0) {
                fs_inodeRelease(context, nodeFile);
                return;
        }
        tentry = fs_ramAlloc(FS_RAM_FILE_HISTORY_QUEUE,
                            sizeof (fs_NdmpHistoryNodeQueue_t));
        FS_ASSERT(tentry != NULL);
        bzero(tentry, sizeof (fs_NdmpHistoryNodeQueue_t));
        tentry->block = 0;
        tentry->boff = 0;

        /*
         * Write history for root
         */

        func(context, dumpContext, nodeFile, 0, 0);
        DLL_INSERT_TAIL(hdr, tentry, link);
        //DLL_ITERATE(entry, hdr, link) {
loop:
        DLL_REMOVE_HEAD(hdr, entry, link);
        while (entry != NULL) {
                fs_yieldPeriodic(context);
                nodeBuf = O_GET_SHARED(nodeFile, entry->block);
                FS_ASSERT(nodeBuf != NULL);
                node = (fs_NdmpHistoryNode_t *)(nodeBuf->u.data +
                                                    entry->boff);
                /*
                 * Call the func for each of the children and add
                 * each of the children to this list
                 */

                block = node->firstChildBlock;
                boff = node->firstChildBoff;
                fs_bufRelease(context, nodeBuf);
                while (block != FS_NULL_DISKPTR_BLOCK) {
                        fs_yieldPeriodic(context);
                        func(context, dumpContext, nodeFile, block, boff);

                        /*
                         * Add this node to the queue
                         */

                        tentry = fs_ramAlloc(FS_RAM_FILE_HISTORY_QUEUE,
                                            sizeof (fs_NdmpHistoryNodeQueue_t));
                        FS_ASSERT(tentry != NULL);
                        bzero(tentry, sizeof (fs_NdmpHistoryNodeQueue_t));
                        tentry->block = block;
                        tentry->boff = boff;
                        DLL_INSERT_TAIL(hdr, tentry, link);

                        nodeBuf = O_GET_SHARED(nodeFile, block);
                        FS_ASSERT(nodeBuf != NULL);
                        node = (fs_NdmpHistoryNode_t *)(nodeBuf->u.data +
                                                            boff);
                        block = node->sibling;
                        boff = node->siblingBoff;
                        fs_bufRelease(context, nodeBuf);

                } 
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                     ("In fs_walkHistory freeing entry"));
                fs_ramDealloc(FS_RAM_FILE_HISTORY_QUEUE, entry);
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
                     ("In fs_walkHistory freed entry"));
                //entry = NULL;
                //DLL_REMOVE_HEAD(hdr, entry, link);
                goto loop;
        }
        fs_inodeRelease(context, nodeFile);
}

/*
 * For a given object send the file histories dirs
 */

void
fs_fileHistoryProcessDirs(
                fs_Context          *context,
                fs_DumpContext_t    *dumpContext,
                fs_Inode            *nodeFile,
                int64               block,
                uint16              boff
)
{
    fs_NdmpHistoryNode_t    *node;
    fs_Inode                *nameFile;
    fs_Buf                  *nodeBuf;
    fs_Buf                  *nameBuf;
    char                    *namePtr;
    int     rc;

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("In fs_fileHistoryProcess for block %lld boff %d",
              block, boff));

    if (!dumpContext->doFileHistory) {
        rc =  NFX_OK;
        goto out;
    }

    if (dumpContext->abort) {
        rc = NFX_OK;
        goto out;
    }

    if (nodeFile->dinode.size == 0) {

            /*
             * This file system was not backed up as nothing had changed
             */
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("In fs_fileHistoryProcess no data in node file"));
            rc = NFX_OK;
            goto out;
    }
    nameFile = I_GET_SHARED(0, dumpContext->nameInum);
    FS_ASSERT(nameFile != NULL);

    nodeBuf = O_GET_SHARED(nodeFile, block);
    FS_ASSERT(nodeBuf != NULL);
    node = (fs_NdmpHistoryNode_t *)(nodeBuf->u.data + boff);
    nameBuf = O_GET_SHARED(nameFile, node->nameBlock);
    FS_ASSERT(nameBuf != NULL);
    namePtr = nameBuf->u.data + node->nameBoff;
#ifndef FS_STANDALONE
    fs_ndmpDumpDirHistory(context, dumpContext, node->parent,
                          node->inum, namePtr, node->nameLen);
    if (FS_I_IS_DIR(node->mode & FS_IFMT_MASK)) {
            if (node->inum == FS_ROOT_INODE) {
                    fs_ndmpDumpDirHistory(context, dumpContext, node->inum,
                                  node->inum, "..", 2);
            } else {
                    fs_ndmpDumpDirHistory(context, dumpContext, node->inum,
                                  node->inum, ".", 1);
                    fs_ndmpDumpDirHistory(context, dumpContext, node->inum,
                                  node->parent, "..", 2);
            }
    }
#endif

   /*
    * Send the filehistory info as we have everything we need now
    */

    fs_bufRelease(context, nodeBuf);
    fs_bufRelease(context, nameBuf);
    fs_inodeRelease(context, nameFile);
    fs_sleep(context, 0);
out:
    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("In fs_fileHistoryProcess returning"));
    return;
}

/*
 * For a given object send the file histories dirs
 */

void
fs_fileHistoryProcessNodes(
                fs_Context          *context,
                fs_DumpContext_t    *dumpContext,
                fs_Inode            *nodeFile,
                int64               block,
                uint16              boff
)
{
    fs_NdmpHistoryNode_t    *node;
    fs_Buf                  *nodeBuf;
    int     rc;

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("In fs_fileHistoryProcess for block %lld boff %d",
              block, boff));

    if (!dumpContext->doFileHistory) {
        rc =  NFX_OK;
        goto out;
    }

    if (dumpContext->abort) {
        rc = NFX_OK;
        goto out;
    }

    if (nodeFile->dinode.size == 0) {

            /*
             * This file system was not backed up as nothing had changed
             */
            FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
             ("In fs_fileHistoryProcess no data in node file"));
            rc = NFX_OK;
            goto out;
    }
    nodeBuf = O_GET_SHARED(nodeFile, block);
    FS_ASSERT(nodeBuf != NULL);
    node = (fs_NdmpHistoryNode_t *)(nodeBuf->u.data + boff);
    fs_ndmpDumpNodeHistoryNodeBuf(context, dumpContext, node);

    fs_bufRelease(context, nodeBuf);
    //fs_sleep(context, 0);
out:
    return;
}

/*
 *  Create the file history files that will be required during
 *  dump. The files are always created under a special directory
 *  
 *  TODO - protect against dump and mount when all the objects
 *  are being removed
 */
void
fs_createFileHistories(
        fs_Context          *context,
        fs_DumpContext_t    *dumpContext
)
{
    fs_Inode  *nodeFile;
    fs_Inode  *nameFile;
    fs_Inode  *dumpTempDir;
    char      buf[256];

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE, ("In fs_createFileHistory"));

    //Get file system modify lock
    fs_lockShared(context, &context->fileSystem->modifyLock, NULL);
    
    dumpTempDir = I_GET_EXCL(0, FS_TEMP_DUMP_SWAP);
    FS_ASSERT(dumpTempDir != NULL);
    fs_logBeginTransaction(context);
    sprintf(buf, "%d.%d.%s", fs_getTimeOfDaySecs(), dumpContext->snapnum,
            "NODE");
    nodeFile = fs_allocateFileInode(context, dumpTempDir, 0777, 0, 0, 0);
    nodeFile->dinode.size = 0;
    nodeFile->dinode.piu.parentInum = FS_TEMP_DUMP_SWAP;
    FS_ASSERT(nodeFile != NULL);
    fs_dir_addent(context, dumpTempDir, buf, strlen(buf), nodeFile->inum);

    sprintf(buf, "%d.%d.%s", fs_getTimeOfDaySecs(), dumpContext->snapnum,
            "NAME");
    nameFile = fs_allocateFileInode(context, dumpTempDir, 0777, 0, 0, 0);
    FS_ASSERT(nodeFile != NULL);
    nameFile->dinode.size = 0;
    nameFile->dinode.piu.parentInum = FS_TEMP_DUMP_SWAP;
    fs_dir_addent(context, dumpTempDir, buf, strlen(buf), nameFile->inum);

    dumpTempDir->dinode.atime = fs_getTimeOfDay();
    dumpTempDir->dinode.mtime = dumpTempDir->dinode.atime;
    dumpTempDir->dinode.ctime = dumpTempDir->dinode.atime;
    fs_inodeWrite(context, dumpTempDir, INO_WRITE_LOG);
    fs_inodeWrite(context, nodeFile, INO_WRITE_LOG);
    fs_inodeWrite(context, nameFile, INO_WRITE_LOG);

    dumpContext->nameInum = nameFile->inum;
    dumpContext->nodeInum = nodeFile->inum;

    fs_logCommitEntry(context);

    //Release file system modify lock
    fs_unlockShared(context, &context->fileSystem->modifyLock);
    
    //Make sure we leave enough space in the log
    fs_makeSureLogIsNotFull(context);

    //Grow the file system if needed
    if (fs_growIfNeeded(context)
     || !fs_haveMininodesAvail(context->fileSystem))    {
         fs_sleep (context, 0);
    }
        
}

/*
 * Remove all the files in the dumpTempDir for the particular dump
 * Context or all the files for all the dumps.
 */

void
fs_cleanFileHistories(
    fs_Context          *context,
    fs_DumpContext_t    *dumpContext
)
{
    fs_Inode        *dumpTempDir, *inode;
    fs_DirEntry     *dirents, *dirp;
    int             eof = 0, i;
    uint32          numEntries = 0;
    uint64          cookie = 0;

    FS_TRACE(context, FS_TRACE_DUMP_RESTORE,
            ("In fs_cleanFileHistory with dumpContext %p", dumpContext));

    /*
     * Iterate throught the dump temp directory and remove
     * all the files that are requested to be removed
     */
    
    //Get file system modify lock
    fs_lockShared(context, &context->fileSystem->modifyLock, NULL);

    dirents = fs_ramAlloc(FS_RAM_DUMP_DIRENT, FS_MAXPATHLEN);
retry:
    dumpTempDir = I_GET_EXCL(0, FS_TEMP_DUMP_SWAP);
    FS_ASSERT(dumpTempDir != NULL);
    eof = 0;
    while (!eof) {
            dirp = dirents;
            fs_dir_readdir(context, dumpTempDir, dirp, FS_MAXPATHLEN + 1,
                            cookie, FALSE, &numEntries, &eof);
            fs_inodeReleaseExclusive(context, dumpTempDir);
            for (i = 0; i < numEntries; i++) {
                    cookie = dirp->cookie;
                    if (strcmp(dirp->name, ".") == 0) {
                        FS_READDIR_ADV_ENTRY_PTR(dirp);
                        continue;
                    }
                    if (strcmp(dirp->name, "..") == 0) {
                        FS_READDIR_ADV_ENTRY_PTR(dirp);
                        continue;
                    }
                    if (strcmp(dirp->name, ".snapshot") == 0) {
                        FS_READDIR_ADV_ENTRY_PTR(dirp);
                        continue;
                    }
                    if (strcmp(dirp->name, "lost+found") == 0) {
                        FS_READDIR_ADV_ENTRY_PTR(dirp);
                        continue;
                    }
                    if (dumpContext == NULL ||
                        dumpContext->nodeInum == dirp->inode ||
                        dumpContext->nameInum == dirp->inode) {

                        dumpTempDir = I_GET_EXCL(0, FS_TEMP_DUMP_SWAP);
                        inode = I_GET_EXCL(0, dirp->inode);

                        fs_logBeginTransaction(context);
                        
                        fs_truncateFile(context, inode, 0, FALSE, NULL);

                        fs_deallocateInode(context, inode);
                        fs_dir_removeent(context, dumpTempDir, dirp->name,
                                         dirp->nameLen - 1);

                        inode->dinode.attrs = 0;
                        inode->dinode.linkCnt--;
                        fs_inodeWrite(context, inode, INO_WRITE_LOG);
                        dumpTempDir->dinode.ctime = fs_getTimeOfDay();
                        dumpTempDir->dinode.mtime = dumpTempDir->dinode.ctime;
                        fs_inodeWrite(context, dumpTempDir, INO_WRITE_LOG);
                        fs_logCommitEntry(context);
                        
                        //Make sure we leave enough space in the log
                        fs_makeSureLogIsNotFull(context);

                        goto retry;
                    }
                    FS_READDIR_ADV_ENTRY_PTR(dirp);

            }
            dumpTempDir = I_GET_EXCL(0, FS_TEMP_DUMP_SWAP);

    }
    fs_ramDealloc(FS_RAM_DUMP_DIRENT, dirents);
    fs_inodeReleaseExclusive(context, dumpTempDir);

    //Release file system modify lock
    fs_unlockShared(context, &context->fileSystem->modifyLock);
}
