/*------------------------------------------------------------------------------
 * RCS:  $Id: fs-log-replay.c,v 1.49 2002/12/10 17:46:08 janv Exp $
 *
 * Name         : fs-log-replay.c
 *
 * Description  : Filesystem log replay.
 * 
 * Created By   : Bob Fozard
 *
 * Date Created : 3/20/2001
 *
 *------------------------------------------------------------------------------
 */

#if !defined(SSC)

/*------------------------------------------------------------------------------
 * Header files.
 *------------------------------------------------------------------------------
 */
#include "fs-includes.h"
#include "fs-log-private.h"

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */
// volume log cache
extern fs_vlcCache vlcCache;

/*------------------------------------------------------------------------------
 * Clean up any remaining logSummaries that we might
 * have allocated during replay.
 */
PRIVATE void
fs_logReplayDiscard(
    fs_Context      *context,
    fs_LogFsRoot    *logRoot,
    fs_LogReplay    *replay)
{
    int32               hash;
    fs_LogSummary       *logSummary, *logSummaryNext;

    for (hash = 0; hash < FS_LOG_SUMM_HASH_MAX; hash++)
    {
        DLL_ITERATE_NEXT(logSummary, logSummaryNext,
                         &replay->summHdr[hash], hashEntry)
        {
            DLL_REMOVE(&replay->summHdr[hash], logSummary, hashEntry);

            fs_ramDealloc(FS_RAM_LOG_SUMM, logSummary);
        }
    }

} /* fs_logReplayDiscard */

/*------------------------------------------------------------------------------
 * Find the buf in cache or retrieve from disk.
 */
void
fs_logReplayFindBuf(
    fs_Context      *context,
    fs_LogReplay    *replay,
    uint64          inum, 
    uint64          page,
    fs_Block        block,
    boolean         isMeta,
    boolean         doRead,
    fs_Buf          **bufp)
{
    fs_BufList      *bufList;
    fs_Buf          *buf = NULL;
    int32           bufListWhich;

    *bufp = NULL;

    LOG_TRACE(("replay findBuf inum %lld page %lld block %lld", 
                  inum, page, fs_blockNumber(block))); 

    FS_ASSERT (fs_blockIsValid(block));

    if (isMeta)
    {
        bufList = &context->fileSystem->bufList[FS_BUFLIST_META];
        bufListWhich = FS_BUFLIST_META;
    }
    else
    {
        bufList = &context->fileSystem->bufList[FS_BUFLIST_USER];
        bufListWhich = FS_BUFLIST_USER;
    }

    if (isMeta)
    {
        if (doRead)
        {
            buf = B_GET_EXCL(NULL, block);
        }
        else
        {
            buf = B_LOOKUP_EXCL(NULL, block);
            if (buf == NULL)
            {
                buf = fs_bufGetBlock(context, NULL, block,
                                     BUF_GET_MODE_EXCL,
                                     BUF_GET_OPT_SLEEP | BUF_GET_OPT_NEW);
            }
        }
    }
    else
    {
        if (doRead)
        {
            buf = B_USER_GET_EXCL(NULL, block);
        }
        else
        {
            buf = B_USER_LOOKUP_EXCL(NULL, block);
            if (buf == NULL)
            {
                buf = fs_bufGetBlock(context, NULL, block,
                                     BUF_GET_MODE_EXCL,
                                     BUF_GET_OPT_SLEEP | BUF_GET_OPT_NEW |
                                        BUF_GET_OPT_USER);
            }
        }
    }

    FS_ASSERT (buf != NULL);
    buf->inum = inum;
    buf->page = page;
    buf->snapshots = 1;
    buf->flags |= FS_BUF_LOG_REPLAY;

    LOG_TRACE(("replay findBuf got buf %p", buf));

    // insert into page cache if necc
    // the page cache needs to be repopulated for
    // an inode flush record and for a truncate after log replay
    // fs_bufFreeInode() which will flush by page only

    if (page != -1)
    {
        fs_Buf *pbuf;

        NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
        pbuf = fs_bufFindByPage(context, inum, 0, bufList, page);
        if (pbuf == NULL)
        {
            // add it to the page cache
            fs_bufPageCacheInsert(context, buf, bufList, NULL);

            // need to bump up the page counter here if dirty

            if (buf->flags & FS_BUF_DIRTY)
            {
                fs_BufPageHdr   *pageHdr;

                // Has already been dirtied, but wasn't on page list
                // before, so increment pageHdr dirty count here.

                pageHdr = fs_bufPageHdrForInum(buf->fileSystem,
                                               bufList->type, buf->inum);
                FS_ASSERT(pageHdr != NULL);
                FS_ASSERT(!DLL_IS_ISOLATED(buf, mrdEntry));

                if (DLL_IS_ISOLATED(buf, pageDirtyEntry))
                {
                    // insert to dirty page hdr
                    DLL_INSERT_HEAD(&pageHdr->bufDirtyHdr, buf, pageDirtyEntry);
                    pageHdr->dirtyCount += 1;
                }
            }
        }
        NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);
    }

#ifndef NEW_ZF
    if (bufListWhich == FS_BUFLIST_USER)
    {
#ifdef NEW_ZF
        FS_ASSERT(DLL_IS_ISOLATED(buf, zeroFillEntry));
#endif // NEW_ZF
        
        NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
        if (!DLL_IS_ISOLATED(buf, zeroFillEntry))
        {
            // This transaction is referencing a buf that was
            // previously zero-filled during this replay.  Remove
            // it from the zeroFill list so it won't be flushed
            // from the cache when we're done with replay.
    
            //printf("removing buf %p inum %lld page %lld from zeroFillHdr\n",
            //  buf, buf->inum, buf->page);
            DLL_REMOVE(&replay->zeroFillHdr, buf, zeroFillEntry);
        }
        NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);
    }
#endif // NEW_ZF

    *bufp = buf;
}

/*------------------------------------------------------------------------------
 * Dump information from the specified bitmap record.
 */
void
fs_logDumpBitmapRec(
    fs_LogRecord    *record,
    uint64          inum)
{
    if (inum == FS_INODE_INODE)
    {
      printf("%d %s inum %lld\n",
          record->hdr.id,
          record->u.bitmap.debugOp.op == BITMAP_OP_ALLOC ? "alloc" : "free ",
          record->u.bitmap.debugOp.u.inum);
    }
    if (inum == FS_REF_COUNT_INODE ||
        inum == FS_FREE_BUDDY_TREE_INODE)
    {
        char *opStr[4] = {"unk", "alloc", "free ", "merge"};
        printf("%d %s %s blk %lld len %d\n",
                record->hdr.id,
                inum == FS_REF_COUNT_INODE ? "ref" : "bud",
                opStr[record->u.bitmap.debugOp.op],
                fs_blockNumber(fs_extentStartBlock(
                    record->u.bitmap.debugOp.u.extent)),
                    fs_extentLength(record->u.bitmap.debugOp.u.extent));
    
        //if (fs_extentLength(record->u.bitmap.debugOp.u.extent) == 0)
        //{
            //printf(" b %lld o %d l %d ",
             //fs_blockNumber(record->u.bitmap.blkId.block),
                   //record->u.bitmap.offset,
                   //record->u.bitmap.length);
        //}
    }
}

/*------------------------------------------------------------------------------
 * Mark the specified buf dirty, add to the dirty list if
 * needed.
 */
void
fs_logReplayMarkBufDirty(
    fs_Context  *context,
    fs_BufList  *bufList, 
    fs_Buf      *buf)
{
    if ((buf->flags & FS_BUF_DIRTY) == 0)
    {
        int32 itype;

        // mark DIRTY
        LOG_TRACE(("mrd buf %p blk %lld", buf, fs_blockNumber(buf->block)));

        FS_ASSERT(DLL_IS_ISOLATED(buf, mrdEntry));

        buf->flags |= FS_BUF_DIRTY;

        NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);

        DLL_INSERT_HEAD(&bufList->mrdHdr, buf, mrdEntry);
        buf->flags |= FS_BUF_MRD;

        itype = fs_bufGetItype(buf);
        bufList->count.dirtyBufs[itype] += 1;
        fs_bufListGlobal[bufList->type].count.dirtyBufs[itype] += 1;
        bufList->count.dirtyBufs[FS_BUF_CACHE_ITYPE_TOTAL] += 1;
        fs_bufListGlobal[bufList->type].count.dirtyBufs[FS_BUF_CACHE_ITYPE_TOTAL] += 1;

        if (!DLL_IS_ISOLATED(buf, pageEntry))
        {
            fs_BufPageHdr   *pageHdr;
            
            pageHdr = fs_bufPageHdrForInum(buf->fileSystem,
                                           bufList->type, buf->inum);
            FS_ASSERT(pageHdr != NULL);
    
            if (DLL_IS_ISOLATED(buf, pageDirtyEntry))
            {
                // insert to dirty page hdr
                DLL_INSERT_HEAD(&pageHdr->bufDirtyHdr, buf, pageDirtyEntry);
                pageHdr->dirtyCount += 1;
            }
        }
    
        NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);
    
        if (fs_bufListGlobal[bufList->type].count.dirtyBufs[itype] >
            fs_bufListGlobal[bufList->type].count.dirtyBufLimit[itype])
        {
            fs_bufSignalWriteback(context, bufList->type, FS_WB_WAKEUP_REPLAY);
        }
    }

} /* fs_logReplayMarkBufDirty */

/*------------------------------------------------------------------------------
 * Perform replay of the specified metadata record.
 */
PRIVATE int32
fs_logReplayMetaRecord(
    fs_Context      *context,
    fs_LogReplay    *replay,
    fs_LogFsRoot    *logRoot,
    fs_LogSummary   *logSummary,
    fs_LogRecord    *record,
    void            **objectPtr)
{
    char            *newDatap = NULL;
    int32           offset, length, i;
    uint64          srcBlock = 0;
    uint64          inum = 0, page;
    fs_Block        block;
    fs_Buf          *buf = NULL;
    fs_BufList      *bufList;
    boolean         isMeta = TRUE;
    boolean         doRead = TRUE;

    switch(record->hdr.recType)
    {
    case FS_LOG_REC_META_ONE:
        LOG_TRACE(("replay M1 id %d inum %lld page %lld block %lld "
                   "offset %d len %d",
                   record->hdr.id,
                   record->u.metaOne.blkId.inum,
                   record->u.metaOne.blkId.page,
                   fs_blockNumber(record->u.metaOne.blkId.block),
                   record->u.metaOne.offset,
                   record->u.metaOne.length));
        block = record->u.metaOne.blkId.block;
        inum = record->u.metaOne.blkId.inum;
        page = record->u.metaOne.blkId.page;
        offset = record->u.metaOne.offset;
        length = record->u.metaOne.length;
        newDatap = record->u.metaOne.data;
        break;

    case FS_LOG_REC_META_TWO:
        LOG_TRACE(("replay M2 id %d sb %lld inum %lld page %lld block %lld "
                   "offset %d len %d",
                   record->hdr.id,
                   record->u.metaTwo.srcBlkNum,
                   record->u.metaTwo.dstBlkId.inum,
                   record->u.metaTwo.dstBlkId.page,
                   fs_blockNumber(record->u.metaTwo.dstBlkId.block),
                   record->u.metaTwo.offset,
                   record->u.metaTwo.length));
        srcBlock = record->u.metaTwo.srcBlkNum;
        block = record->u.metaTwo.dstBlkId.block;
        inum = record->u.metaTwo.dstBlkId.inum;
        page = record->u.metaTwo.dstBlkId.page;
        offset = record->u.metaTwo.offset;
        length = record->u.metaTwo.length;
        newDatap = record->u.metaTwo.data;
        break;

#ifdef NEW_ZF
    case FS_LOG_REC_META_ZERO:
        LOG_TRACE(("replay META_ZERO %d, inum %lld page %lld block %lld "
                   "offset %d len %d",
                   record->hdr.id,
                   record->u.metaZero.blkId.inum,
                   record->u.metaZero.blkId.page,
                   fs_blockNumber(record->u.metaZero.blkId.block),
                   record->u.metaZero.offset,
                   record->u.metaZero.length));
        
        block = record->u.metaZero.blkId.block;
        inum = record->u.metaZero.blkId.inum;
        page = record->u.metaZero.blkId.page;
        offset = record->u.metaZero.offset;
        length = record->u.metaZero.length;
        isMeta = TRUE;
        newDatap = NULL;
        if (length == FS_BLOCK_SIZE)
        {
            // If we're going to fill the entire block
            // don't read it from disk, its a waste of
            // time, and could also cause on-disk checksum
            // to fail, in the event we've never written
            // the block before.
            doRead = FALSE;
        }
        break;
#else // NEW_ZF
    case FS_LOG_REC_ZERO_FILL:
        LOG_TRACE( ("replay Z %d, inum %lld page %lld block %lld offset %d "
                    "len %d isMeta %d",
                    record->hdr.id,
                    record->u.zero.blkId.inum,
                    record->u.zero.blkId.page,
                    fs_blockNumber(record->u.zero.blkId.block),
                    record->u.zero.offset,
                    record->u.zero.length,
                    record->u.zero.isMeta));
        
        block = record->u.zero.blkId.block;
        inum = record->u.zero.blkId.inum;
        page = record->u.zero.blkId.page;
        offset = record->u.zero.offset;
        length = record->u.zero.length;
        isMeta = record->u.zero.isMeta;
        newDatap = NULL;
        if (length == FS_BLOCK_SIZE)
        {
            // If we're going to fill the entire block
            // don't read it from disk, its a waste of
            // time, and could also cause on-disk checksum
            // to fail, in the event we've never written
            // the block before.
            doRead = FALSE;
        }
        break;
#endif // NEW_ZF

    case FS_LOG_REC_BITMAP:

        LOG_TRACE(
            ("replay B %d, inum %lld page %lld block %lld offset %d "
             "length %d  %s",
             record->hdr.id,
             record->u.bitmap.blkId.inum,
             record->u.bitmap.blkId.page,
             fs_blockNumber(record->u.bitmap.blkId.block),
             record->u.bitmap.offset,
             record->u.bitmap.length,
             ((record->u.bitmap.op == FS_LOG_REC_BITMAP_SET) ? "SET" : "RESET")));

        block = record->u.bitmap.blkId.block;
        inum = record->u.bitmap.blkId.inum;
        page = record->u.bitmap.blkId.page;
        offset = record->u.bitmap.offset;
        length = record->u.bitmap.length;
        newDatap = NULL;
        //fs_logDumpBitmapRec(record, inum);
        break;

    default:
        // Bad record.  This was just checked, something is seriously wrong.
        fs_error(context, FS_INTERNAL_ERR, "replay logrec");
        return NFX_ERR;
    } // switch

    if (offset >= FS_BLOCK_SIZE)
    {
        fs_warn(context, FS_LOG_ENTRY_ERR, "offset");
        return NFX_ERR;
    }
    if (length > FS_BLOCK_SIZE)
    {
        fs_warn(context, FS_LOG_ENTRY_ERR, "rec length");
        return NFX_ERR;
    }
    if (offset + length > FS_BLOCK_SIZE)
    {
        fs_warn(context, FS_LOG_ENTRY_ERR, "offset+length");
        return NFX_ERR;
    }
    if (srcBlock >= context->fileSystem->super->numBlks)
    {
        fs_warn(context, FS_LOG_ENTRY_ERR, "srcBlk >= numblks");
        return NFX_ERR;
    }
    if (fs_blockIsValid(block) &&
        (fs_blockNumber(block) >= context->fileSystem->super->numBlks))
    {
        fs_warn(context, FS_LOG_ENTRY_ERR, "dstBlk >= numblks");
        return NFX_ERR;
    }

    fs_logReplayFindBuf(context,
                        replay,
                        inum,
                        page,
                        block,
                        isMeta,
                        doRead,
                        &buf);

    if (buf == NULL)
        return NFX_OK;

    if (isMeta)
    {
        bufList = &context->fileSystem->bufList[FS_BUFLIST_META];
    }
    else
    {
        bufList = &context->fileSystem->bufList[FS_BUFLIST_USER];
    }

    switch(record->hdr.recType)
    {
    case FS_LOG_REC_META_ONE:
    case FS_LOG_REC_META_TWO:
        memcpy ((buf->u.data + offset), newDatap, length);

        break;
#ifdef NEW_ZF
    case FS_LOG_REC_META_ZERO:
        memset ((buf->u.data + offset), 0, length);
        break;
#else // NEW_ZF
    case FS_LOG_REC_ZERO_FILL:
        if (isMeta)
        {
            memset ((buf->u.data + offset), 0, length);
        }
        else
        {
            if (length == FS_BLOCK_SIZE)
            {
                FS_ASSERT(offset == 0);
                FS_ASSERT(doRead == FALSE);

                if (buf->u.user.codHandle)
                {
                    fs_cod_free_cod(buf->u.user.codHandle);
                }

                buf->u.user.codHandle = fs_cod_get_new_cod();
                buf->u.user.codLength = FS_BLOCK_SIZE;
                buf->u.user.codOffset = 0;
                fs_cod_create_null(buf->u.user.codHandle,
                                   FS_BLOCK_SIZE, 0);
            }
            else
            {
                fs_cod_fill_null(&buf->u.user.codHandle, length, offset);
            }

            if (!DLL_IS_ISOLATED(buf, zeroFillEntry))
            {
                FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                        ("FS_DBG: dup zero-fill buf %p inum %lld page %lld",
                         buf, inum, page));
            }
            else
            {
                //printf(
                //    "inserting buf %p inum %lld page %lld on zero-fill hdr\n",
                //       buf, buf->inum, buf->page);
                NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
                DLL_INSERT_TAIL(&replay->zeroFillHdr, buf, zeroFillEntry);
                NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);
            }
        }
        break;
#endif // NEW_ZF
    case FS_LOG_REC_BITMAP:
        for (i = 0; i < length; i++)
        {
            if (record->u.bitmap.op == FS_LOG_REC_BITMAP_SET)
            {
                buf->u.bytes[record->u.bitmap.offset + i] |=
                    record->u.bitmap.bits[i];
            }
            else
            {
                buf->u.bytes[record->u.bitmap.offset + i] &= 
                    ~record->u.bitmap.bits[i];
            }
        }
        break;
    } // switch

    fs_logReplayMarkBufDirty(context, bufList, buf);

    if (isMeta)
    {
        FS_DBG_UPDATE_SHADOW(buf, offset, length);
    }

    *objectPtr = buf;
    
    if (buf != NULL)
    {
        if (!(buf->flags & FS_BUF_TRANS_WRITE))
        {
            fs_logAddBufLock(context, logSummary, buf);
            buf->flags |= FS_BUF_TRANS_WRITE;
        }
    }
    fs_lockSetDirty(context, &buf->lock);

    {
        int i, n;
        
        for (i = 0, n = fs_superBlockCount(); i < n; i++)
        {
            if (fs_blockNumber(block) == fs_blockNumber(fs_superBlockBlock(i)))
            {
                fs_SuperBlock * super = (fs_SuperBlock*)buf->u.data;
                uint32          savedCrc = super->crc32;

                FS_ASSERT(isMeta);
                FS_ASSERT(record->hdr.recType == FS_LOG_REC_META_ONE);
                FS_ASSERT(offset == 0);
                FS_ASSERT(length == sizeof(fs_SuperBlock));

                super->crc32 = 0;
                super->crc32 = fs_checksum(super, sizeof(fs_SuperBlock), ~0);
                if (savedCrc == super->crc32)
                {
                    // update the incore superblock 
                    NFX_SPIN_LOCK(&context->fileSystem->superLock);
                    memcpy (context->fileSystem->super, buf->u.data, length);
                    context->fileSystem->params = 
                        fs_computeSizeParameters(context->fileSystem->super);
                    NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
                }
                else
                {
                    fs_warn(context, FS_LOG_ENTRY_ERR, "dstBlk >= numblks");
                    return NFX_ERR;
                }
            }
        } // for
    }

    return (NFX_OK);

} /* fs_logReplayMetaRecord */

/*------------------------------------------------------------------------------
 * Perform replay of the specified userdata record.
 */
PRIVATE int32
fs_logReplayUserRecord(
    fs_Context      *context,
    fs_LogReplay    *replay,
    fs_LogFsRoot    *logRoot,
    fs_LogSummary   *logSummary,
    fs_LogRecord    *record,
    void            **objectPtr)
{
    fs_Buf          *buf = NULL;
    fs_BufList      *bufList;
    
    LOG_TRACE(("replay U %d, inum %lld page %lld block %lld, cod 0x%llx",
              record->hdr.id,
              record->u.user.blkId.inum,
              record->u.user.blkId.page,
              fs_blockNumber(record->u.user.blkId.block),
              record->u.user.codHandle));

    // This function is left around for possible future
    // support of an NVRAM log device.  Not currently
    // supported, and very probably broken in various
    // ways. -- bob.
    //
    FS_ASSERT(FALSE);

    fs_logReplayFindBuf (context,
                         replay,
                         record->u.user.blkId.inum,
                         record->u.user.blkId.page,
                         record->u.user.blkId.block,
                         FS_BUF_TYPE_USER,
                         TRUE,
                         &buf);

    if (buf == NULL)
        return NFX_OK;

    buf->u.user.codHandle = record->u.user.codHandle;
    buf->u.user.codLength = record->u.user.codLength;
    buf->u.user.codOffset = 0; // Todo Verify

    // mark COD as being in use 
    // mark DIRTY
    bufList = &context->fileSystem->bufList[FS_BUFLIST_USER];

    fs_logReplayMarkBufDirty(context, bufList, buf);

    *objectPtr = buf;
    
    if (buf != NULL)
    {
        if (!(buf->flags & FS_BUF_TRANS_WRITE))
        {
            fs_logAddBufLock(context, logSummary, buf);
            buf->flags |= FS_BUF_TRANS_WRITE;
        }
    }
    fs_lockSetDirty(context, &buf->lock);

    return NFX_OK;

}  // fs_logReplayUserRecord

/*------------------------------------------------------------------------------
 * Replay the specified alloc record.
 */
PRIVATE int32
fs_logReplayAllocRecord(
    fs_Context      *context,
    fs_LogReplay    *replay,
    fs_LogFsRoot    *logRoot,
    fs_LogSummary   *logSummary,
    fs_LogRecord    *record,
    void            **objectPtr)
{
    fs_Buf          *buf = NULL;
    fs_BufList      *bufList;

    LOG_TRACE(("replay A %d, inum %lld page %lld block %lld isCow %d "
               "srcBlk %lld",
              record->hdr.id,
              record->u.alloc.blkId.inum,
              record->u.alloc.blkId.page,
              fs_blockNumber(record->u.alloc.blkId.block),
              record->u.alloc.isCow,
              record->u.alloc.oldBlkNum));

//    printf ("replay A %d, inum %lld page %lld block %lld\n",
//              record->hdr.id, record->u.alloc.blkId.inum,
//              record->u.alloc.blkId.page, fs_blockNumber(record->u.alloc.blkId.block));

    FS_ASSERT(fs_blockIsValid(record->u.alloc.blkId.block));

    if (record->u.alloc.isMeta)
        bufList = &context->fileSystem->bufList[FS_BUFLIST_META];
    else
        bufList = &context->fileSystem->bufList[FS_BUFLIST_USER];

    // If COW, get the original or source block and replace with new.
    if (record->u.alloc.isCow)
    {
        fs_Buf      *newBuf;
        fs_Block    srcBlock;

        // need to flush new blk that was read in during readahead scan
        // to avoid duplicate blk insert via fs_bufCopyOnWriteReInsertBuf

        if (record->u.alloc.isMeta)
        {
            newBuf = B_LOOKUP_EXCL(NULL, record->u.alloc.blkId.block);
        }
        else
        {
            newBuf = B_USER_LOOKUP_EXCL(NULL, record->u.alloc.blkId.block);
        }
        if (newBuf)
        {
            fs_bufFree(context, newBuf, BUF_IS_REPLAY);
        }

        srcBlock = fs_makeBlock(record->u.alloc.oldBlkNum);
        FS_ASSERT(fs_blockIsValid(srcBlock));
        fs_logReplayFindBuf(context,
                            replay,
                            record->u.alloc.blkId.inum,
                            record->u.alloc.blkId.page,
                            srcBlock,
                            record->u.alloc.isMeta,
                            TRUE,
                            &buf);
        FS_ASSERT(buf);

        // Remove old dependencies.
        if (buf->logSegDep.count > 0)
        {
            fs_logRemoveDeps(context, &buf->logSegDep, &buf->spin);
        }

        buf->block = record->u.alloc.blkId.block;

        // remove the buffer from its old cache location and re-insert it
        fs_bufCopyOnWriteReInsertBuf(context, buf, srcBlock);
    }
    else
    {
        fs_logReplayFindBuf(context,
                            replay,
                            record->u.alloc.blkId.inum,
                            record->u.alloc.blkId.page,
                            record->u.alloc.blkId.block,
                            record->u.alloc.isMeta,
                            TRUE,
                            &buf);
    }

    // mark DIRTY

    fs_logReplayMarkBufDirty(context, bufList, buf);

    *objectPtr = buf;
    
    if (buf != NULL)
    {
        if (!(buf->flags & FS_BUF_TRANS_WRITE))
        {
            fs_logAddBufLock(context, logSummary, buf);
            buf->flags |= FS_BUF_TRANS_WRITE;
        }
    }
    fs_lockSetDirty(context, &buf->lock);

    return NFX_OK;
} /* fs_logReplayAllocRecord */

/*------------------------------------------------------------------------------
 *
 */
PRIVATE int32
fs_logReplayInodeRecord(
    fs_Context      *context,
    fs_LogFsRoot    *logRoot,
    fs_LogSummary   *logSummary,
    fs_LogRecord    *record,
    void            **objectPtr)
{
    fs_Inode        *inode;
    
    LOG_TRACE(("replay I %d, inum %lld, snapnum %d",
                record->hdr.id,
                record->u.inode.inum,
                record->u.inode.snapnum));

    inode = fs_inodeInsertForReplay(context, record->u.inode.snapnum,
                                    record->u.inode.inum,
                                    (FS_LOCK_MODE_EXCL|FS_LOCK_MODE_WAIT));
    if (inode == NULL)
    {
        *objectPtr = NULL;
        return NFX_ERR;
    }
    memcpy(&inode->dinode, &record->u.inode.dinode, sizeof(fs_Dinode));
    fs_inodeWrite(context, inode, 0);
 
    if (!(inode->flags & INO_FLAGS_TRANS_LOCK))
    {
        fs_logAddInodeLock(context, logSummary, inode);
        inode->flags |= INO_FLAGS_TRANS_LOCK;
    }
    *objectPtr = inode;

    return NFX_OK;

} /* fs_logReplayInodeRecord */

/*------------------------------------------------------------------------------
 *
 */
PRIVATE int32
fs_logReplayInodeFlushRecord(
    fs_Context      *context,
    fs_LogFsRoot    *logRoot,
    fs_LogSummary   *logSummary,
    fs_LogRecord    *record,
    void            **objectPtr)
{
    fs_Inode        *inode;
    
    LOG_TRACE(("replay IF %d, inum %lld",
                record->hdr.id,
                record->u.inodeFlush.inum));

    inode = fs_inodeInsertForReplay(context, 0, record->u.inodeFlush.inum,
                                    FS_LOCK_MODE_EXCL|FS_LOCK_MODE_WAIT);
    if (inode == NULL)
    {
        // A flush record for an inode not in the cache.
        // Ignore it.  TODO - is this a bug or a feature?!?!
    }
    else
    {
        // Remove dependencies
        if (inode->logSegDep.count > 0)
        {
            fs_logRemoveDeps(context, &inode->logSegDep, NULL);
        }

        // Clear the buffer cache of all pages for this
        // inode.
        fs_bufFreeInode(context, inode, 0, TRUE);

        // Now remove the inode itself from the cache.
        fs_inodeRemoveFromCache(context, inode, TRUE);
    }

    *objectPtr = NULL;
    return NFX_OK;

} /* fs_logReplayInodeFlushRecord */

/*------------------------------------------------------------------------------
 *
 */
PRIVATE int32
fs_logReplayBufFlushRecord(
    fs_Context      *context,
    fs_LogFsRoot    *logRoot,
    fs_LogSummary   *logSummary,
    fs_LogRecord    *record,
    void            **objectPtr)
{
    
    LOG_TRACE(("replay BF %d, inum %lld, page %lld, blknum %lld",
                record->hdr.id,
                record->u.bufFlush.inum,
                record->u.bufFlush.startPage,
                fs_blockNumber(record->u.bufFlush.block)));

    if (fs_blockIsValid(record->u.bufFlush.block))
    {
        fs_Buf      *buf;
        fs_Block    block = record->u.bufFlush.block;

        if (record->u.bufFlush.bufList == FS_BUFLIST_META)
        {
            buf = B_LOOKUP_EXCL(NULL, block);
        }
        else
        {
            buf = B_USER_LOOKUP_EXCL(NULL, block);
        }

        if (buf != NULL)
        {
            fs_bufFree(context, buf, BUF_IS_REPLAY);
        }
    }
    else
    {
        // Not currently supported (possible future NVRAM
        // log feature).
    }

    *objectPtr = NULL;
    return NFX_OK;

} /* fs_logReplayBufFlushRecord */

/*------------------------------------------------------------------------------
 *
 */
extern void
fs_logPrintStats(fs_LogStats *stats, uint32 logLen)
{
    printf("  Nbr of Meta 1 entries          %12d\n", stats->nbrMetaOne);
    printf("  Nbr of Meta 2 entries          %12d\n", stats->nbrMetaTwo);
    printf("  Nbr of User   entries          %12d\n", stats->nbrUser);
    printf("  Nbr of Alloc  entries          %12d\n", stats->nbrAlloc);
    printf("  Nbr of Zero U entries          %12d\n", stats->nbrZeroUser);
    printf("  Nbr of Zero M entries          %12d\n", stats->nbrZeroMeta);
    printf("  Nbr of Remove entries          %12d\n", stats->nbrUserRm);
    printf("  Nbr of Bitmap entries          %12d\n", stats->nbrBitmap);
    printf("  Nbr of Inode  entries          %12d\n", stats->nbrInode);
    printf("  Nbr of InodeFlush  entries     %12d\n", stats->nbrInodeFlush);
    printf("  Nbr of BufFlush  entries       %12d\n", stats->nbrBufFlush);
    printf("  Nbr of Commit entries          %12d\n", stats->nbrCommit);
    printf("  Nbr of Done entries            %12d\n", stats->nbrMarkDone);
    printf("  Nbr of EOL entries             %12d\n", stats->nbrEOL);
    printf("  Nbr of extra bitmap creations  %12d\n", stats->nbrExtraBitmapsCreate);
    if (stats->nbrCommit)
    {
        printf("  AVG/MAX nbr of entries per transaction %4d %4d\n", 
               stats->nbrEntries / stats->nbrCommit, stats->nbrMaxTransaction);
        if (stats->nbrEntries)
            printf("  AVG/MAX bytes per record               %4d %4d\n",  
                   logLen / stats->nbrEntries, stats->MaxRecLen);
    }
#ifdef NEW_ZF
    FS_ASSERT(stats->nbrMarkDone == 0);
#endif // NEW_ZF
}

/*------------------------------------------------------------------------------
 * Called to perform some basic sanity check for fs_LogRecord
 */
int32
fs_logVerifyRecord(fs_LogRecord *record)
{
    fs_Block    block = fs_makeBlock(0);
    boolean     checkBlock = FALSE;
    uint32      offset = 0, length = 0;
    boolean     checkOffLen = FALSE;

    if (record->hdr.id < FS_LOG_ID_MIN ||
        record->hdr.id > FS_LOG_ID_MAX)
    {
        return NFX_ERR;
    }

    switch (record->hdr.recType)
    {
        case FS_LOG_REC_META_ONE:
            block = record->u.metaOne.blkId.block;
            checkBlock = TRUE;
            length = record->u.metaOne.length;
            offset = record->u.metaOne.offset;
            checkOffLen = TRUE;
            break;
        case FS_LOG_REC_META_TWO:
            if (!fs_blockIsValid(record->u.metaTwo.dstBlkId.block))
            {
                printf(
                    "log: id %d invalid dst block %lld for metaTwo record\n",
                        record->hdr.id,
                        fs_blockNumber(record->u.metaTwo.dstBlkId.block));
                return NFX_ERR;
            }
            if (!fs_blockIsValid(fs_makeBlock(record->u.metaTwo.srcBlkNum)))
            {
                printf(
                   "log: id %d invalid src block %lld for metaTwo record\n",
                        record->hdr.id, record->u.metaTwo.srcBlkNum);
                return NFX_ERR;
            }
            length = record->u.metaTwo.length;
            offset = record->u.metaTwo.offset;
            checkOffLen = TRUE;
            break;
        case FS_LOG_REC_USER:
            block = record->u.user.blkId.block;
            checkBlock = TRUE;
            break;
        case FS_LOG_REC_ALLOC:
            block = record->u.alloc.blkId.block;
            checkBlock = TRUE;
            if (record->u.alloc.isCow)
            {
                if (!fs_blockIsValid(fs_makeBlock(record->u.alloc.oldBlkNum)))
                {
                    printf("log: id %d invalid old cow block %lld\n",
                            record->hdr.id, record->u.alloc.oldBlkNum);
                    return NFX_ERR;
                }
                if (fs_blockNumber(record->u.alloc.blkId.block) ==
                        record->u.alloc.oldBlkNum)
                {
                    printf(
               "log: id %d new and src block %lld same for cow alloc record\n",
                            record->hdr.id, record->u.alloc.oldBlkNum);
                    return NFX_ERR;
                }
            }
            break;
#ifdef NEW_ZF
    case FS_LOG_REC_META_ZERO:
        block = record->u.metaZero.blkId.block;
        checkBlock = TRUE;
        length = record->u.metaZero.length;
        offset = record->u.metaZero.offset;
        checkOffLen = TRUE;
        break;
    case FS_LOG_REC_USER_ZERO:
        // don't check anything for this record
        break;
#else // NEW_ZF
        case FS_LOG_REC_ZERO_FILL:
            block = record->u.zero.blkId.block;
            checkBlock = TRUE;
            length = record->u.zero.length;
            offset = record->u.zero.offset;
            checkOffLen = TRUE;
            break;
#endif // NEW_ZF
        case FS_LOG_REC_BITMAP:
            if ((record->u.bitmap.op != FS_LOG_REC_BITMAP_SET) &&
                (record->u.bitmap.op != FS_LOG_REC_BITMAP_RESET))
            {
                printf("log: id %d invalid bitmap operation %d\n",
                        record->hdr.id, record->u.bitmap.op);
                return NFX_ERR;
            }
            block = record->u.bitmap.blkId.block;
            checkBlock = TRUE;
            length = record->u.bitmap.length;
            offset = record->u.bitmap.offset;
            checkOffLen = TRUE;
            break;
#ifndef NEW_ZF
        case FS_LOG_REC_MARK_DONE:
#endif // NEW_ZF
        case FS_LOG_REC_INODE:
        case FS_LOG_REC_INODE_FLUSH:
            break;
        case FS_LOG_REC_BUF_FLUSH:
            // block may be invalid for fs_bufFreeInode
            break;
        case FS_LOG_REC_COMMIT:
            break;
        default:
            return NFX_ERR;
    }

    if (checkBlock && !fs_blockIsValid(block))
    {
        printf("log: id %d invalid block %lld for record type %d\n",
               record->hdr.id, fs_blockNumber(block), record->hdr.recType);
        return NFX_ERR;
    }
    if (checkOffLen && ((length + offset) > FS_BLOCK_SIZE))
    {
        printf("log: id %d invalid length %d offset %d for recrod type %d\n",
                record->hdr.id, length, offset, record->hdr.recType);
        return NFX_ERR;
    }

    return NFX_OK;
}

/*------------------------------------------------------------------------------
 * Called to increment fs_LogStats based on record type
 */
void
fs_logIncStats(
    fs_LogRecord    *record,
    int32           recLen,
    fs_LogStats     *stats)
{
    switch(record->hdr.recType)
    {
    case FS_LOG_REC_META_ONE:
        stats->nbrMetaOne++;
        break;
    case FS_LOG_REC_META_TWO:
        stats->nbrMetaTwo++;
        break;
    case FS_LOG_REC_USER:
        stats->nbrUser++;
        break;
    case FS_LOG_REC_ALLOC:
        stats->nbrAlloc++;
        break;
#ifdef NEW_ZF
    case FS_LOG_REC_META_ZERO:
        stats->nbrZeroMeta++;
        break;
    case FS_LOG_REC_USER_ZERO:
        stats->nbrZeroUser++;
        break;
#else // NEW_ZF
    case FS_LOG_REC_ZERO_FILL:
        if (record->u.zero.isMeta)
            stats->nbrZeroMeta++;
        else
            stats->nbrZeroUser++;
        break;
#endif // NEW_ZF
    case FS_LOG_REC_BITMAP:
        stats->nbrBitmap++;
        break;
#ifndef NEW_ZF
    case FS_LOG_REC_MARK_DONE:
        stats->nbrMarkDone++;
        break;
#endif // NEW_ZF
    case FS_LOG_REC_INODE:
        stats->nbrInode++;
        break;
    case FS_LOG_REC_INODE_FLUSH:
        stats->nbrInodeFlush++;
        break;
    case FS_LOG_REC_BUF_FLUSH:
        stats->nbrBufFlush++;
        break;
    case FS_LOG_REC_COMMIT:
        stats->nbrCommit++;
        if (stats->nbrCurrTransaction > stats->nbrMaxTransaction)
            stats->nbrMaxTransaction = stats->nbrCurrTransaction;
        stats->nbrCurrTransaction = 0;
        break;
    } // switch

    stats->nbrCurrTransaction++;
    stats->nbrEntries++;
    if (recLen > stats->MaxRecLen)
        stats->MaxRecLen = recLen;
}

/*------------------------------------------------------------------------------
 * Remove the specified logSummary from the replay.
 */
PRIVATE void
fs_logRemoveReplaySummary(
    fs_LogReplay    *replay,
    fs_LogSummary   *logSummary)
{
    uint32          hash = FS_LOG_SUMMARY_HASH(logSummary->id);

    DLL_REMOVE(&replay->summHdr[hash], logSummary, hashEntry);
}

/*------------------------------------------------------------------------------
 * Return the logSummary for the specified transaction
 * id, allocating a new one if necessary.
 */
fs_LogSummary *
fs_logGetReplaySummary(
    fs_LogReplay    *replay,
    fs_LogId        id)
{
    fs_LogSummary   *logSummary = NULL;
    uint32          hash = FS_LOG_SUMMARY_HASH(id);

    DLL_ITERATE(logSummary, &replay->summHdr[hash], hashEntry)
    {
        if (logSummary->id == id)
            break;
    }

    if (logSummary == DLL_NIL)
    {
        // Allocate log summary for the id.
        logSummary = fs_ramAlloc(FS_RAM_LOG_SUMM, sizeof(fs_LogSummary));
        bzero(logSummary, sizeof(fs_LogSummary));
        logSummary->id = id;
        DLL_INIT(&logSummary->undoHdr);

        DLL_INSERT_HEAD(&replay->summHdr[hash], logSummary, hashEntry);
    }

    return logSummary;
}

/*------------------------------------------------------------------------------
 * Discard the specified replay record.
 */
PRIVATE void
fs_logReplayDiscardRecord(
    fs_Context      *context,
    fs_LogReplay    *replay,
    fs_LogRecHdr    *record,
    uint32          offset,
    int32           length)
{
    fs_LogFsRoot        *logRoot = context->fileSystem->logRoot;
    fs_LogSummary       *logSummary;
    fs_LogObjectSumm    objSumm;

    logSummary = fs_logGetReplaySummary(replay, record->id);
    context->logSummary = logSummary;

    LOG_TRACE(("replay discard id %d type %d off %d len %d",
              record->id,
              record->recType,
              offset,
              length));

    if (offset == logRoot->head)
    {
        LOG_TRACE(("discard head id %d type %d",
                    record->id, record->recType));
    }

    if (record->recType == FS_LOG_REC_COMMIT)
    {
        // Trash the summary.
        
        fs_logRemoveReplaySummary(replay, logSummary);
        fs_ramDealloc(FS_RAM_LOG_SUMM, logSummary);
        context->logSummary = NULL;
    }

    objSumm.objectPtr = NULL;
    objSumm.objectType = FS_LOG_DEP_NONE;
    objSumm.offset = offset;
    objSumm.length = length;

    fs_logMarkActive(context, &objSumm);
}

/*------------------------------------------------------------------------------
 * Replay the specified COMMIT record.
 */
PRIVATE void
fs_logReplayCommitRecord(
    fs_Context      *context,
    fs_LogReplay    *replay,
    fs_LogSummary   *logSummary,
    fs_LogRecHdr    *record,
    uint32          offset,
    int32           length)
{
    fs_LogFsRoot        *logRoot = context->fileSystem->logRoot;
    fs_LogObjectSumm    objSumm;

    fs_logRemoveReplaySummary(replay, logSummary);

    fs_logCommitInodes(context, logSummary);
    fs_logCommitSummary(context, logRoot, logSummary);

    objSumm.objectPtr = NULL;
    objSumm.objectType = FS_LOG_DEP_NONE;
    objSumm.offset = offset;
    objSumm.length = length;

    fs_logMarkActive(context, &objSumm);

    // Take this opportunity to sync out dirty inodes if necessary
    if (fs_inodePercentAvail() < fs_incoreTune->inodePctAvailLoWtrMk)
    {
        fs_inodeSyncCacheSpawnNew(context);
    }
}

#ifdef NEW_ZF
/*------------------------------------------------------------------------------
 * Build a list of all pages to user data to zero fill.
 */
PRIVATE void
fs_logReplayUserZeroRecord(fs_Context *         context,
                           fs_LogRecord *       record,
                           fs_LogObjectSumm *   objSumm)

{
    if (record->u.userZero.zero == TRUE)
    {
        // add pages to list
        fs_zfAddPage(context,
                     record->u.userZero.blkId.inum,
                     record->u.userZero.blkId.page,
                     record->u.userZero.offset,
                     record->u.userZero.nbPage,
                     record, objSumm);
    }
    else
    {
        // remove pages from list
        fs_zfDelPage(context,
                     record->u.userZero.blkId.inum,
                     record->u.userZero.blkId.page,
                     TRUE);
        fs_logMarkActive(context, objSumm);
    }
    
    return;
}
#endif // NEW_ZF

/*------------------------------------------------------------------------------
 * Replay the specified log record.
 */
PRIVATE int32
fs_logReplayRecord(
    fs_Context      *context,
    fs_LogReplay    *replay,
    fs_LogRecord    *record,
    uint32          offset,
    int32           length)
{
    fs_LogSummary       *logSummary;
    fs_LogObjectSumm    objSumm;
    fs_LogFsRoot        *logRoot = context->fileSystem->logRoot;
    boolean             addDep = FALSE;
    int32               rc = NFX_OK;

    logSummary = fs_logGetReplaySummary(replay, record->hdr.id);
    context->logSummary = logSummary;

    objSumm.offset = offset;
    objSumm.length = length;
    objSumm.objectType = FS_LOG_DEP_NONE;
    objSumm.objectPtr = NULL;

    switch(record->hdr.recType)
    {
    case FS_LOG_REC_COMMIT:
        fs_logReplayCommitRecord(context, replay, logSummary,
                                 (fs_LogRecHdr *) record,
                                 offset, length);
        // Nothing more to do here for COMMIT.
        return NFX_OK;
    case FS_LOG_REC_ALLOC:
        rc = fs_logReplayAllocRecord(context, replay, logRoot, logSummary,
                                     record, &objSumm.objectPtr);
        objSumm.objectType = FS_LOG_DEP_BUF;
        addDep = TRUE;
        break;
    case FS_LOG_REC_USER:
        rc = fs_logReplayUserRecord (context, replay, logRoot, logSummary,
                                     record, &objSumm.objectPtr);
        objSumm.objectType = FS_LOG_DEP_BUF;
        addDep = TRUE;
        break;
    case FS_LOG_REC_META_ONE:
    case FS_LOG_REC_META_TWO:
#ifdef NEW_ZF
    case FS_LOG_REC_META_ZERO:
#else // NEW_ZF
    case FS_LOG_REC_ZERO_FILL:
#endif // NEW_ZF
    case FS_LOG_REC_BITMAP:
        rc = fs_logReplayMetaRecord(context, replay, logRoot, logSummary,
                                    record, &objSumm.objectPtr);
        objSumm.objectType = FS_LOG_DEP_BUF;
        addDep = TRUE;

        //if(record->u.bitmap.blkId.inum == FS_REF_COUNT_INODE ||
        //   record->u.bitmap.blkId.inum == FS_FREE_BUDDY_TREE_INODE ||
        //   record->u.bitmap.blkId.inum == FS_INODE_INODE)
        // {
        //     printf("num %d offset %d len %d\n",
        //        record->hdr.recNum, offset, length); 
        //}
        break;
        
#ifdef NEW_ZF
    case FS_LOG_REC_USER_ZERO:
    {
        fs_logReplayUserZeroRecord(context, record, &objSumm);
        rc = NFX_OK;
        break;
    }
#endif // NEW_ZF
        
    case FS_LOG_REC_INODE:
        rc = fs_logReplayInodeRecord(context, logRoot, logSummary,
                                     record, &objSumm.objectPtr);
        if (rc != NFX_OK)
        {
            // Inode not found, trash this record
            rc = NFX_OK;
        }
        else
        {
            objSumm.objectType = FS_LOG_DEP_INODE;
            addDep = TRUE;
        }
        break;
    case FS_LOG_REC_INODE_FLUSH:
        rc = fs_logReplayInodeFlushRecord(context, logRoot, logSummary,
                                          record, &objSumm.objectPtr);
        break;
    case FS_LOG_REC_BUF_FLUSH:
        rc = fs_logReplayBufFlushRecord(context, logRoot, logSummary,
                                        record, &objSumm.objectPtr);
        break;
#ifndef NEW_ZF
    case FS_LOG_REC_MARK_DONE:
        rc = NFX_OK;
        break;
#endif // NEW_ZF
    default:
        // Bad record. something is seriously wrong.
        fs_warn(context, FS_INTERNAL_ERR,
                "replay logrec (type %d)", record->hdr.recType);
        return NFX_ERR;
    }

    if (rc != NFX_OK)
    {
        fs_warn(context, FS_INTERNAL_ERR, "replay logrec (rc %d)", rc);
        return rc;
    }
    /*
      if (offset == logRoot->head)
      {
      printf("Replay Head Id %d Type %d AddDep %d objectPtr %p\n", 
      record->hdr.id, record->hdr.recType, addDep, objSumm.objectPtr);
      }
    */
    
#ifdef NEW_ZF
    if (record->hdr.recType != FS_LOG_REC_USER_ZERO)
    {
#endif // NEW_ZF
    if (addDep)
    {
        fs_logAddDep(context, &record->hdr, &objSumm, FALSE);
    }
    else
    {
        fs_logMarkActive(context, &objSumm);
    }
#ifdef NEW_ZF
    }
#endif // NEW_ZF

    return NFX_OK;

} // fs_logReplayRecord

/*------------------------------------------------------------------------------
 * Don't replay yet the specified log record,
 * but read the meta data it will need so there will be in the cache.
 */
PRIVATE int32
fs_logPrepareReplayRecord(fs_Context *      context,
                          fs_LogRecord *    record,
                          boolean           sync)
{
    int32       rc = NFX_OK;
    uint64      inum;
    uint64      page;
    fs_Block    block;
    uint32      type;
    boolean     read = TRUE;

    switch(record->hdr.recType)
    {
    case FS_LOG_REC_COMMIT:
    {
        // don't need any meta data
        read = FALSE;
        break;
    }
    case FS_LOG_REC_ALLOC:
    {
        inum = record->u.alloc.blkId.inum;
        page = record->u.alloc.blkId.page;
        if (record->u.alloc.isCow)
            block =  fs_makeBlock(record->u.alloc.oldBlkNum);
        else
            block = record->u.alloc.blkId.block;
        type = FS_BUFLIST_META;
        break;
    }
    case FS_LOG_REC_USER:
    {
        inum = record->u.user.blkId.inum;
        page = record->u.user.blkId.page;
        block = record->u.user.blkId.block;
        type = FS_BUFLIST_USER;
        break;
    }
    case FS_LOG_REC_META_ONE:
    {
        inum = record->u.metaOne.blkId.inum;
        page = record->u.metaOne.blkId.page;
        block = record->u.metaOne.blkId.block;
        type = FS_BUFLIST_META;
        break;
    }        
    case FS_LOG_REC_META_TWO:
    {
        inum = record->u.metaTwo.dstBlkId.inum;
        page = record->u.metaTwo.dstBlkId.page;
        block = record->u.metaTwo.dstBlkId.block;
        type = FS_BUFLIST_META;
        break;
    }        
#ifdef NEW_ZF
    case FS_LOG_REC_USER_ZERO:
    case FS_LOG_REC_META_ZERO:
    {
        read = FALSE;
        break;
    }        
#else // NEW_ZF
    case FS_LOG_REC_ZERO_FILL:
    {
        inum = record->u.zero.blkId.inum;
        page = record->u.zero.blkId.page;
        block = record->u.zero.blkId.block;
        type = FS_BUFLIST_META;
        break;
    }        
#endif // NEW_ZF
    case FS_LOG_REC_BITMAP:
    {
        inum = record->u.bitmap.blkId.inum;
        page = record->u.bitmap.blkId.page;
        block = record->u.bitmap.blkId.block;
        type = FS_BUFLIST_META;
        break;
    }              
    case FS_LOG_REC_INODE:
    case FS_LOG_REC_INODE_FLUSH:
    {
        read = FALSE;
        break;
    }        
    case FS_LOG_REC_BUF_FLUSH:
    {
        inum = record->u.bufFlush.inum;
        page = record->u.bufFlush.startPage;
        block = record->u.bufFlush.block;
        type = record->u.bufFlush.bufList;
        break;
    }              
#ifndef NEW_ZF
    case FS_LOG_REC_MARK_DONE:
    {
        read = FALSE;
        break;
    }        
#endif // NEW_ZF
    default:
    {
        read = FALSE;
        rc = NFX_ERR;
    }
    } // switch

    if (read == TRUE)
    {
        fs_bufLookupBlock(context, inum, page, block, type, sync);
    }
    
    return(rc);
}

/*------------------------------------------------------------------------------
 * Release memory allocated for replay bitmap.
 */
void
fs_logReleaseReplayBitmap(
    fs_Context               *context,
    fs_LogReplayBitmap       *bitmap)
{
    fs_LogReplayExtraBitmap *extra, *nextExtra;

    if (bitmap->bits)
    {
        fs_ramDealloc(FS_RAM_CORE_MISC, bitmap->bits);
        bitmap->bits = NULL;
    }

    DLL_ITERATE_NEXT(extra, nextExtra, &bitmap->extra, entry)
        {
            DLL_REMOVE(&bitmap->extra, extra, entry);
            fs_ramDealloc(FS_RAM_CORE_MISC, extra);
        };

}

/*------------------------------------------------------------------------------
 * Initialize the bitmap for recording which transactions
 * will be replayed.
 */
void
fs_logInitReplayBitmap(
    fs_Context               *context,
    fs_LogReplayBitmap       *bitmap)
{
    uint32  numBits;
    uint32  numBytes;
    
    bzero(bitmap, sizeof *bitmap);

    DLL_INIT(&(bitmap->extra));

    // Simply allocate enough bits to cover the maximum
    // log size with nothing but commit records in it.
    // This would not be a valid log, since there must be
    // something other than just commits of nothing, but
    // this number will serve to provide for handling
    // all the ids in our circular bitmap.
    numBits = (FS_LOG_MAX_NBR_BLOCKS * FS_BLOCK_SIZE) / sizeof(fs_LogRecHdr);

    numBytes = numBits / BITS_PER_BYTE;

    bitmap->bits = fs_ramAlloc(FS_RAM_CORE_MISC, numBytes);
    bzero(bitmap->bits, numBytes);

    bitmap->numBits = numBits;
    bitmap->firstId = FS_LOG_ID_INVALID;

    bitmap->maxRevIndex = numBits;

    FS_ASSERT(FS_LOG_ID_MAX > bitmap->numBits);

    //printf("initReplayBitmap numBytes %d numBits %d nbrCommits %d\n",
    //       numBytes, bitmap->numBits, logInfo->nbrCommits);
}

/*------------------------------------------------------------------------------
 * Return the bit index in the replayBitmap for the
 * specified id.
 */

#define FS_LOG_INDEX_NOT_IN_BITMAP     -2

PRIVATE uint32
fs_logReplayBitmapGetIndex(
    fs_Context               *context,
    fs_LogReplayBitmap       *bitmap,
    fs_LogId                 id)
{
    uint32  index, zeroIndex;

    FS_ASSERT(id >= FS_LOG_ID_MIN &&
              id <= FS_LOG_ID_MAX);

    if (bitmap->firstId == FS_LOG_ID_INVALID)
    {
        // Initialize our point of reference in the bitmap.
        //printf("setup bitmap->firstId to %d\n", id);
        bitmap->firstId = id;
        return 0;
    }
    
    if (id == bitmap->firstId)
    {
        index = 0;
    }
    else 
    if (id > bitmap->firstId)
    {
        index = id - bitmap->firstId;

        if (index >= bitmap->maxRevIndex)
        {
            if (bitmap->firstId >= bitmap->numBits)
            {
                // collison
                index = FS_LOG_INDEX_NOT_IN_BITMAP ;
            }
            else
            {
                zeroIndex = bitmap->numBits - bitmap->firstId;
               
                if (zeroIndex <= bitmap->maxForwIndex)
                {
                    // collison
                    index = FS_LOG_INDEX_NOT_IN_BITMAP ;
                }
                else
                if ((FS_LOG_ID_MAX - id) >= (zeroIndex - bitmap->maxForwIndex))
                {
                    // collison
                    index = FS_LOG_INDEX_NOT_IN_BITMAP ;
                }
                else
                {
                    index = zeroIndex - (FS_LOG_ID_MAX - id);

                    FS_ASSERT(index > bitmap->maxForwIndex);
                
                    if (index < bitmap->maxRevIndex)
                    {
                        bitmap->maxRevIndex = index;
                    }
                }
            }
        }
        else
        {
            if (index > bitmap->maxForwIndex)
            {
                bitmap->maxForwIndex = index;
            }
        }
    }
    else
    {

        index = bitmap->firstId - id;

        if (index > bitmap->numBits)
        {
            // check the forward wrap around case
        
            index = (FS_LOG_ID_MAX - bitmap->firstId) + id ;
            if (index >= bitmap->maxRevIndex)
            {
                // collison, out of available bitmap
                index = FS_LOG_INDEX_NOT_IN_BITMAP ;
            }
            else
            {
                if (index > bitmap->maxForwIndex)
                {
                    bitmap->maxForwIndex = index;
                }
            }
        }
        else
        {
            index = bitmap->numBits - (bitmap->firstId - id) ;
        
            if ((index == 0) || (index <= bitmap->maxForwIndex))
            {
                // collison, not enough space
                index = FS_LOG_INDEX_NOT_IN_BITMAP ;
            }
            else
            {
                if (index < bitmap->maxRevIndex)
                {
                    bitmap->maxRevIndex = index;
                }
            }
        }
    }
    FS_ASSERT((index == FS_LOG_INDEX_NOT_IN_BITMAP) ||
              (index < bitmap->numBits));
    //printf("id %d index %d\n", id, index);
    return index;
}

/*------------------------------------------------------------------------------
 * Return the bit index in the extra Bitmap for the
 * specified id.
 * create the extra bitmap if needed.
 */
PRIVATE uint32
fs_logReplayGetExtraBitmapIndex(
    fs_Context               *context,
    fs_LogReplayBitmap       *bitmap,
    fs_LogId                 id,
    fs_LogReplayExtraBitmap  **extraBitmap)
{
    uint32                    bitmapId;
    uint32                    bitIndex;
    fs_LogReplayExtraBitmap   *extra;

    FS_ASSERT(FS_LOG_EXTRA_BITMAP >= 3);

    context->fileSystem->logRoot->logStats.nbrExtraBitmapsAccess++;

    bitmapId = id >> FS_LOG_EXTRA_BITMAP ;
    bitIndex = id & ((1 << FS_LOG_EXTRA_BITMAP) - 1);

    DLL_ITERATE(extra, &bitmap->extra, entry)
        {
            if (extra->bitmapId == bitmapId)
            {
                *extraBitmap = extra;
                return(bitIndex);
            }
        }
    
    // not found in list, create extra bitmap entry
    extra = fs_ramAlloc(FS_RAM_CORE_MISC, sizeof(fs_LogReplayExtraBitmap));

    bzero(extra, sizeof(fs_LogReplayExtraBitmap));
    extra->bitmapId = bitmapId;
    
    DLL_INSERT_HEAD(&bitmap->extra, extra, entry);

    context->fileSystem->logRoot->logStats.nbrExtraBitmapsCreate++;

    // printf("logReplay: Create extra bitmap (log id %d)\n", id);

    *extraBitmap = extra;
    return(bitIndex);
}

/*------------------------------------------------------------------------------
 * Return TRUE if the specified id's bit is set
 * in the replayBitmap, FALSE otherwise.
 */
PRIVATE boolean
fs_logReplayBitmapIsSet(
    fs_Context               *context,
    fs_LogReplayBitmap       *bitmap,
    fs_LogId                 id)
{
    uint32 bitIndex = fs_logReplayBitmapGetIndex(context, bitmap, id);
    uint32 byteIndex = bitIndex / BITS_PER_BYTE;
    uchar8 mask = 1 << (bitIndex % BITS_PER_BYTE);

    if (bitIndex == FS_LOG_INDEX_NOT_IN_BITMAP)
    {
        fs_LogReplayExtraBitmap  *extraBitmap;
        
        bitIndex = fs_logReplayGetExtraBitmapIndex(context, bitmap, id, &extraBitmap);
        byteIndex = bitIndex / BITS_PER_BYTE;
        mask = 1 << (bitIndex % BITS_PER_BYTE);
        
        return (extraBitmap->bits[byteIndex] & mask);
    }
    else
    {
        return (bitmap->bits[byteIndex] & mask);
    }
}

/*------------------------------------------------------------------------------
 * Update the replay bitmap boundaries to ensure a
 * bit is reserved for the specified transaction id.
 */
PRIVATE void
fs_logReplayBitmapReserve(
    fs_Context               *context,
    fs_LogReplayBitmap       *bitmap,
    fs_LogId                 id)
{
    // uint32 bitIndex = 
    fs_logReplayBitmapGetIndex(context, bitmap, id);

    // No need to reserve space for the extra bitmaps
    // Only needed for commit (done) records
}

/*------------------------------------------------------------------------------
 * Update the replay bitmap to reflect commit of
 * the specified transaction id.
 */
PRIVATE void
fs_logReplayBitmapCommit(
    fs_Context               *context,
    fs_LogReplayBitmap       *bitmap,
    fs_LogId                 id)
{
    uint32 bitIndex = fs_logReplayBitmapGetIndex(context, bitmap, id);
    uint32 byteIndex = bitIndex / BITS_PER_BYTE;
    uchar8 mask = 1 << (bitIndex % BITS_PER_BYTE);

    if (bitIndex == FS_LOG_INDEX_NOT_IN_BITMAP)
    {
        fs_LogReplayExtraBitmap  *extraBitmap;
        
        bitIndex = fs_logReplayGetExtraBitmapIndex(context, bitmap, id, &extraBitmap);
        byteIndex = bitIndex / BITS_PER_BYTE;
        mask = 1 << (bitIndex % BITS_PER_BYTE);
        
        extraBitmap->bits[byteIndex] |= mask;
    }
    else
    {
        bitmap->bits[byteIndex] |= mask;
    }

    //printf("commit id %d\n", id);
}

#ifndef NEW_ZF
/*------------------------------------------------------------------------------
 * Update the replay bitmap to reflect done (ie. don't replay)
 * of the specified transaction id.
 */
PRIVATE void
fs_logReplayBitmapDone(
    fs_Context               *context,
    fs_LogReplayBitmap       *bitmap,
    fs_LogId                 id)
{
    uint32 bitIndex = fs_logReplayBitmapGetIndex(context, bitmap, id);
    uint32 byteIndex = bitIndex / BITS_PER_BYTE;
    uchar8 mask = 1 << (bitIndex % BITS_PER_BYTE);

    if (bitIndex == FS_LOG_INDEX_NOT_IN_BITMAP)
    {
        fs_LogReplayExtraBitmap  *extraBitmap;
        
        bitIndex = fs_logReplayGetExtraBitmapIndex(context, bitmap,
                                                   id, &extraBitmap);
        byteIndex = bitIndex / BITS_PER_BYTE;
        mask = 1 << (bitIndex % BITS_PER_BYTE);
        
        extraBitmap->bits[byteIndex] &= ~mask;
    }
    else
    {
        bitmap->bits[byteIndex] &= ~mask;
    }

    //printf("ignore id %d\n", id);
}
#endif // NEW_ZF

/*------------------------------------------------------------------------------
 * Perform replay action for the specified record.
 */
PRIVATE int32
fs_logActionReplay(
    fs_Context      *context,
    fs_LogReplay    *replay,
    fs_LogRecHdr    *record,
    uint32          offset,
    int32           length,
    uint32          *nbrCommits)
{
    fs_LogFsRoot        *logRoot = context->fileSystem->logRoot;
    int32               rc;

    LOG_TRACE(("action replay rec id %d type %d offset %d len %d",
               record->id, record->recType, offset, length));

    if (record->recType == FS_LOG_REC_COMMIT)
    {
        (*nbrCommits)++;
    }
    
    if (logRoot->nextId <= record->id)
        logRoot->nextId = record->id + 1 ;
    if (logRoot->nextId >= FS_LOG_ID_MAX)
        logRoot->nextId = FS_LOG_ID_MIN;

    if (!fs_logReplayBitmapIsSet(context, &replay->bitmap, record->id))
    {
        // Transaction is not committed or marked done.
        // Add this record to the trash.
        fs_logReplayDiscardRecord(context, replay, record, offset, length);

        return NFX_OK;
    }

    rc = fs_logReplayRecord(context, replay, (fs_LogRecord *) record,
                            offset, length);
        
    return rc;
}

/*------------------------------------------------------------------------------
 * Scan through the entire log for the specified action.
 */
int32
fs_logScan( fs_Context *    context,
            fs_LogFsRoot *  logRoot,
            fs_LogReplay *  replay,
            fs_LogAction    action)
{
    int32               rc;
    fs_FileSystem *     fs = context->fileSystem;
    fs_LogInfo *        logInfo = &(fs->logRoot->logInfo);
    fs_LogDevTypeTable *devTypeTable = context->fileSystem->logDevTypeTable;
    fs_LogStats         logStats, *stats = &logStats;
    fs_LogRecord *      record = NULL;
    int32               totalRead = 0;
    int32               maxReadSize, bytesRead, bytesLeft, recLen = 0;
    uint32              logSize = logInfo->logSize;
    uint32              offset, recOffset, minRead;
    uchar8 *            datap;
    uint32              nbrCommits = 0;
    uint32              readLoop = 0; // only for protect againts infinite loops
    vl_VolBufRoot *     volBufRoot = fs->logRoot->volBufRoot;
    vl_BlockCache *     blockCache = &volBufRoot->blockCache;
    eeeRTC_t            beginTime, delta;
  
    bzero (stats, sizeof(fs_LogStats));
#ifdef ISABELLE // debug
    bzero(&(blockCache->stats), sizeof(blockCache->stats));
#endif // ISABELLE
    
    switch (action)
    {
        case FS_LOG_ACTION_DONE: 
            fs_logInitReplayBitmap(context, &replay->bitmap);
            // fall thru to set init logRoot

        case FS_LOG_ACTION_REPLAY:
            logRoot->headPtr = 0;
            logRoot->nextId = FS_LOG_ID_MIN;
            logRoot->currLen = 0;
            logRoot->head = logInfo->head;
            logRoot->tail = logRoot->head;
            if (logRoot->head % logRoot->segmentSize != 0)
            {
                // Set this segment to active in order for
                // correct activation/truncation to occur.
                logRoot->logSeg[logRoot->head /
                                logRoot->segmentSize].active = TRUE;
                logRoot->logSeg[logRoot->head /
                                logRoot->segmentSize].head = logRoot->head;
                //printf("Replay set segNum %d to active head %d\n",
                //logRoot->head / logRoot->segmentSize, logRoot->head);
            }
            break;
    } // switch

    bytesRead = logInfo->readChunkSize;  // default number of bytes to read
    offset = logInfo->head;              // start at the head pointer

    eee_rtc(&beginTime);

    while (totalRead <= logSize)
    {
        //printf("READ BLOCK OFFSET (%d/%d)\n", offset, bytesRead);
        // read chunk of log data from device
        rc = (devTypeTable->logRead)(context,
                                     &datap,
                                     offset,
                                     &bytesRead,
                                     logInfo->tail);
        if (rc == NFX_ERR)
        {
            FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                    ("logScan: ERROR from logRead(offset %d bytesRead %d) "
                     "totalRead %d", offset, bytesRead, totalRead));
            goto replayErr;
        }

        // check for end of log
        if (bytesRead == 0)
        {
            // verify logRoot->tail matches logInfo
            switch (action)
            {
            case FS_LOG_ACTION_REPLAY:
            case FS_LOG_ACTION_DONE:
                if (logRoot->tail != logInfo->tail)
                {
                    FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                            ("Log replay early termination, logRoot->tail"
                             " %d logInfo.tail %d action %d",
                             logRoot->tail, logInfo->tail, action));
                    stats->nbrEOL++;
                    goto replayErr;
                }
                break;
            } // switch
            
            LOG_TRACE(("scan: no more log data (totalRead %d)", totalRead));
            stats->nbrEOL++;
            goto replayDone;
        }

        // readLoop check
        // should not exceed 3 reads:
        //   record hdr, hdr again if not enough, record hdr plus meta data
        readLoop++;
        if (readLoop > 2)
        {
            FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                    ("logScan: record not available (length %d id %d type %d)",
                     recLen, record->hdr.id, record->hdr.recType));
            goto replayErr;
        }

        bytesLeft = bytesRead;
        maxReadSize = bytesRead;
        minRead = sizeof(fs_LogRecHdr);

        LOG_TRACE(("scan: logRead() offset %d bytesRead %d",
                   offset, bytesRead));

        // read until bytesLeft cannot satisfy minimum log hdr size
        while (minRead <= bytesLeft)
        {
            // set next record ptr using datap, bytesRead and bytesLeft
            record = (fs_LogRecord *) (datap + (maxReadSize - bytesLeft));

            // check for EOL, only for NVRAM
            // TBD: can NVRAM scan for EOL and return
            // correct bytesRead instead ??
            if (logRoot->logDevType == FS_LOG_DEVTYPE_NVRAM &&
                record->hdr.id == FS_LOG_ID_EOL)
            {
                LOG_TRACE(("scan: done, nvram EOL found (totalRead %d)",
                           totalRead));

                if (action == FS_LOG_ACTION_DONE || 
                    action == FS_LOG_ACTION_REPLAY)
                {
                    FS_LOG_INCR_PTR(logSize, logRoot->tail, FS_LOG_ID_BYTES);
                    logRoot->currLen += FS_LOG_ID_BYTES;
                }
                stats->nbrEOL++;
                goto replayDone;
            }

            // check if we have enough bytes for this record type
            minRead = fs_logRecSize(context, record->hdr.recType);
            if (!minRead)
            {
                // Bad record, treat it as end of log.
                FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                        ("logScan: bad record id %d type %d EOL",
                         record->hdr.id, record->hdr.recType));
                goto replayErr;
            }

            if (bytesLeft < minRead)
            {
                // we haven't read the complete record,
                // read another chunk of log
                bytesRead = logInfo->readChunkSize;
                LOG_TRACE(("scan: bytesLeft %d does not satisfy min %d "
                           "(readLoop %d)", bytesLeft, minRead, readLoop));
                break;
            }

            // recalculate record length
            // metaOne and metaTwo have variable length fields
            // and determine if we need to read another chunk of data
            recLen = fs_logGetRecLen(context, record);
            if (recLen == 0)
            {
                FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                        ("logScan: bad record length %d id %d type %d",
                          recLen, record->hdr.id, record->hdr.recType));
                goto replayErr;
            }
            
            if (logInfo->maxSize && (recLen >= logInfo->maxSize))
            {
                // this is an absolute max of a record
                // If we are here, we assume the log is corrupt.
                // remark: we have not yet checked the ck_sum
                FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                        ("logScan: bad record length %d id %d type %d",
                          recLen, record->hdr.id, record->hdr.recType));
                goto replayErr;
            }

            if (recLen > bytesLeft)
            {
                // we haven't read the complete record,
                // read another chunk of log
                bytesRead = recLen;
                LOG_TRACE(("scan: bytesLeft %d does not satisfy rec len %d "
                           "(readLoop %d)", bytesLeft, recLen, readLoop));
                break;
            }

            // Do some sanity checks on the record.
            rc = fs_logVerifyRecord(record);
            if (rc == NFX_ERR)
            {
                FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                        ("logScan: invalid rec content for rec id %d type %d "
                         "len %d",
                         record->hdr.id, record->hdr.recType, recLen));
                goto replayErr;
            }

            // if here, a complete valid log record has been read
            // we have enough for this record, read default value next time
            LOG_TRACE(("scan: complete recLen %d", recLen));
            bytesRead = logInfo->readChunkSize;

            switch (action)
            {
                case FS_LOG_ACTION_DONE:
                    //printf("actDone recid %d rectyp %d xi %d xid %d\n",
                    //record->hdr.id, record->hdr.recType, transactionEntry,
                    //transactionList[transactionEntry]);
      
                    // update logRoot's currLen and tail
                    logRoot->currLen += recLen;
                    recOffset = logRoot->tail;
                    FS_LOG_INCR_PTR(logInfo->logSize, logRoot->tail, recLen);

                    if (record->hdr.recType == FS_LOG_REC_COMMIT)
                    {
                        fs_logReplayBitmapCommit(context, 
                                                 &replay->bitmap,
                                                 record->hdr.id);
                    }
#ifndef NEW_ZF
                    else if (record->hdr.recType == FS_LOG_REC_MARK_DONE)
                    {
                        fs_logReplayBitmapDone(context, 
                                               &replay->bitmap,
                                               record->u.markDone.logId);
                    }
#endif // NEW_ZF
                    else
                    {
                        fs_logReplayBitmapReserve(context, 
                                                  &replay->bitmap,
                                                  record->hdr.id);
                    }

                    /* If we are replaying, start reading asynchronously
                     * the meta data so they will be in the cache for
                     * the next pass.
                     * If there isn't enough resources do a synchronous read,
                     * so all the previous async. read are back.
                     */
                    if (logRoot->skipReplay == FALSE)
                    {
                        if (fs_ioPending < FS_RDWR_MAX/2)
                            fs_logPrepareReplayRecord(context, record, FALSE);
                        else
                            fs_logPrepareReplayRecord(context, record, TRUE);
                    }
                    
                    break;

                case FS_LOG_ACTION_REPLAY:
                    // PR #298 need to update currLen before replay
                    logRoot->currLen += recLen;
                    
                    recOffset = logRoot->tail;
                    FS_LOG_INCR_PTR(logInfo->logSize, logRoot->tail, recLen);

                    rc = fs_logActionReplay(context, replay, 
                                            (fs_LogRecHdr *) record,
                                            recOffset, recLen,
                                            &nbrCommits); 
                    if (rc != NFX_OK)
                    {
                        FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                                ("logScan: replay ERROR for rec id %d "
                                 "type %d len %d",
                                 record->hdr.id, record->hdr.recType, recLen));
                        goto replayErr;
                    }

                    break;

                default:
                    fs_error(context, FS_LOG_ENTRY_ERR,
                             "bad action type: %d", action);
            } // switch (action)

            // decrement bytesLeft by record length
            // reset readLoop and increment stats
            bytesLeft -= recLen;
            readLoop = 0;
            fs_logIncStats(record, recLen, stats);
            LOG_TRACE(("scan: bytesLeft %d", bytesLeft));

            // TBD: not necc

            switch (logRoot->logDevType)
            {
            case FS_LOG_DEVTYPE_SELF:
            case FS_LOG_DEVTYPE_VOLID:
                if (logRoot->tail == logInfo->tail)
                {
                    stats->nbrEOL++;
                    LOG_TRACE(("scan: EOL logInfo.tail same as "
                               "logRoot->tail %d", logRoot->tail));
                    // next read should return 0
                    //goto replayDone;
                }
                break;
            } // switch (logRoot->logDevType)

            eee_rtcDuration(&beginTime, &delta);
            if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
            {
                fs_sleep(context, 0);
                eee_rtc(&beginTime);
            }
        } // while (minRead <= bytesLeft)

        FS_LOG_INCR_PTR(logInfo->logSize, offset, (maxReadSize - bytesLeft));
        totalRead += (maxReadSize - bytesLeft);
        LOG_TRACE(("scan: thisRead %d totalRead %d next offset %d",
                   (maxReadSize - bytesLeft), totalRead, offset));
    } // while (totalRead <= logSize)

    // TODO - validate this....
    // If we get here, we read the entire log device
    // but never detected an end of log.  This should
    // be a very rare event, and not necessarily an
    // indicator of a bad log.
    LOG_TRACE(("scan: EOL condition not detected"));

replayDone:

    NFX_SPIN_LOCK(&(blockCache->spin));
    blockCache->replay = FALSE;
    NFX_SPIN_UNLOCK(&(blockCache->spin));    

#ifdef ISABELLE // debug
    NFX_SPIN_LOCK(&(blockCache->spin));
    FS_ELOG(context, LOG_INFO, FS_TRACE_LOG,
            ("nbMetaReadAhead %lld nbMetaRead %lld "
             "nbLogReadAhead %lld nbYield %lld avgYieldTime %lld usec",
             blockCache->stats.nbMetaReadAhead, blockCache->stats.nbMetaRead,
             blockCache->stats.nbReadAhead, blockCache->stats.nbYield,
             (blockCache->stats.nbYield == 0)? 0 :
             (blockCache->stats.totalYieldTime / blockCache->stats.nbYield)));
    NFX_SPIN_UNLOCK(&(blockCache->spin));    
#endif // ISABELLE

/*
  #ifdef NEW_ZF // debug
  // Display the list of user data page to zero
  if (action == FS_LOG_ACTION_REPLAY)
  fs_zfDisplayList(context);
  #endif // NEW_ZF
  
  #ifdef FS_DEBUG
  if (action == FS_LOG_ACTION_REPLAY)
  {
  printf("%s: log replay stats:\n",
  context->fileSystem->super->mountName);
  fs_logPrintStats(stats, logRoot->currLen);
  }
  #endif // FS_DEBUG
*/
    if (action == FS_LOG_ACTION_REPLAY)
    {
        FS_ELOG(context, LOG_INFO, FS_TRACE_LOG,
                ("replay successfull logHead %d logTail %d currLen %d "
                 "nbrCommits %d",
                 logRoot->head, logRoot->tail, logRoot->currLen, nbrCommits));
    }
    
    return NFX_OK;

replayErr:

    // check the status of the header, and elog/trace some data
    FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
            ("replay error nbrCommits: %d", nbrCommits));

    return NFX_ERR;
}

#ifndef NEW_ZF
/*------------------------------------------------------------------------------
 * Remove unreferenced zero-filled bufs from the
 * buffer cache.
 */
PRIVATE void
fs_logReplayFlushZeroFilled(
    fs_Context      *context,
    fs_LogFsRoot    *logRoot,
    fs_LogReplay    *replay)
{
    fs_BufList  *bufList = &context->fileSystem->bufList[FS_BUFLIST_USER];

    // Scan the zero fill buf list.  All remaining entries must
    // be removed from the buffer cache.  Remaining entries are
    // those that we had logged (and replayed) zero-fill records
    // for but no subsequent transaction was logged that referenced
    // them.  We remove them from the cache to avoid the possible
    // duplicate cache entry that would occur if a new transaction
    // attempted to create the same inum,page.
    if (!DLL_IS_EMPTY(&replay->zeroFillHdr))
    {
        fs_Buf      *buf;
        
        // Allocate a temporary logSummary to facilitate bufFree
        // trying to add the buf locks.
        fs_logBeginTransaction(context);
    
        NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
        while (!DLL_IS_EMPTY(&replay->zeroFillHdr))
        {
            boolean locked;

            DLL_REMOVE_HEAD(&replay->zeroFillHdr, buf, zeroFillEntry);
            locked =
                fs_trylockExclusiveSleep(context, &buf->lock,
                                         &fs_bufListGlobal[bufList->type].spin);
            NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);

            if (locked)
            {
                // Okay to read inodes from disk as needed now,
                // since we've completed replaying all records.
                fs_Inode *inode = I_GET_SHARED(0, buf->inum);

                if (!inode)
                {
                    fs_bufReleaseExclusive(context, buf);
                    NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
                    continue;
                }

                if (inode->dinode.size <= ((buf->page) * FS_BLOCK_SIZE) ||
                    !fs_blockIsValid(fs_emap_lookup_block(context, inode,
                                                          buf->page)))
                {
                    //printf("flush zero-filled buf %p inum %lld page %lld\n",
                    //       buf, buf->inum, buf->page);
                    fs_bufFree(context, buf, BUF_IS_REPLAY);
                    fs_bufCommit(context, buf);
                    context->logSummary->bufCount = 0;
                }
                else
                {
                    //printf("keep zero-filled buf %p inum %lld page %lld\n",
                    //       buf, buf->inum, buf->page);
                    fs_bufReleaseExclusive(context, buf);
                }
                fs_inodeRelease(context, inode);
            }
            NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
        }
        NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);

        // Free the temporary logSummary.
        fs_ramDealloc(FS_RAM_LOG_SUMM, context->logSummary);
    }
} /* fs_logReplayFlushZeroFilled */
#endif // NEW_ZF

/*------------------------------------------------------------------------------
 *
 */
static void
fs_logCleanupAfterReplay(
    fs_Context      *context,
    fs_LogFsRoot    *logRoot,
    fs_LogReplay    *replay)
{
    fs_logReleaseReplayBitmap(context, &replay->bitmap);
        
    fs_logReplayDiscard(context, logRoot, replay);

    context->logSummary = NULL;

#ifndef NEW_ZF
    // Remove unreferenced zero-filled bufs from the
    // buffer cache.
    fs_logReplayFlushZeroFilled(context, logRoot, replay);
    FS_ASSERT(DLL_IS_EMPTY(&replay->zeroFillHdr));
#endif // NEW_ZF
    
    context->logSummary = NULL;

    fs_ramDealloc(FS_RAM_CORE_MISC, replay);
    logRoot->replay = NULL;
    logRoot->inReplay = FALSE;

}

/*------------------------------------------------------------------------------
 * Perform log replay.
 *
 * Phase 0: get the head and tail, check if log can be replayed at this time.
 * Phase 1: read the whole log in memory
 * Phase 2: calculate the bitmap and trigger meta read
 * Phase 3: replay
 */
EXPORTED int32
fs_logReplay(fs_Context *   context,
             boolean        skipReplay,
             fs_LogFsRoot * logRoot)
{
    fs_FileSystem *         fs = context->fileSystem;
    fs_LogDevTypeTable *    devTypeTable = fs->logDevTypeTable;
    vl_VolBufRoot *         volBufRoot = fs->logRoot->volBufRoot;
    vl_BlockCache *         blockCache = &(volBufRoot->blockCache);
    fs_LogReplay *          replay;
    uint32                  i, rc;
    fs_LogInfo *            logInfo = &(fs->logRoot->logInfo);
    int32                   currLogSize = 0;
#ifdef ISABELLE // test only
    eeeRTC_t t1, t2, t3, t4, t5;
    eee_rtc(&t1);
#endif // ISABELLE        

    FS_ELOG(context, LOG_INFO, FS_TRACE_LOG, ("replay: start"));
    
    /* Init replay structure.
     */
    replay = fs_ramAlloc(FS_RAM_CORE_MISC, sizeof *replay);
    for (i = 0; i < FS_LOG_SUMM_HASH_MAX; i++)
    {
        DLL_INIT(&replay->summHdr[i]);
    }
#ifndef NEW_ZF
    DLL_INIT(&replay->zeroFillHdr);
#endif // NEW_ZF
    logRoot->replay = replay;
    logRoot->inReplay = TRUE;
    logRoot->skipReplay = skipReplay;
    
    /* Get and initialize some logRoot values.
     */
    bzero (logInfo, sizeof(fs_LogInfo));
    rc = (devTypeTable->logReadInit)(context, logInfo);
    if (rc == NFX_ERR)
    {
        FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                ("replay: cannot get log info (rc: %d)\n", rc));
        rc = NFX_ERR;
        goto fs_logReplay_exit;
    }
    fs->logSize = logInfo->logSize;

    /* Initialize the volume cache.
     */
    vl_clearBlockCache(context);
    fs_logVolInitCache(blockCache);
    blockCache->entrySize = VLC_BLOCK_SIZE;
    blockCache->replay = TRUE;

    /* Check if enough place in cache to do this replay
     * if not, queue it
     */
    if ((logRoot->logDevType == FS_LOG_DEVTYPE_SELF) ||
        (logRoot->logDevType == FS_LOG_DEVTYPE_VOLID))
    {
        if (logInfo->tail >= logInfo->head)
            currLogSize = logInfo->tail - logInfo->head;
        else
            currLogSize = logInfo->logSize - logInfo->head + logInfo->tail;
        FS_ASSERT(currLogSize <= VLC_MAX_SIZE);
        
        NFX_SPIN_LOCK(&(vlcCache.spin));
        if (currLogSize > (vlcCache.size - vlcCache.used))
        {
            // queue the replay
            esm_event_t wakeEvent;
            fs_vlcMount * vlcFsys = fs_ramAlloc(FS_RAM_VLC,
                                                sizeof(fs_vlcMount));
            FS_ASSERT(vlcFsys != NULL);
            vlcFsys->logSize = currLogSize;
            vlcFsys->smHandle = context->smHandle;
            DLL_ISOLATE_ELEM(vlcFsys, nextInLine);
            DLL_INSERT_TAIL(&(vlcCache.pendingMountList), vlcFsys, nextInLine);
            FS_ELOG(context, LOG_INFO, FS_TRACE_LOG,
                    ("replay: filesystem queued for replay"));
            NFX_SPIN_UNLOCK(&(vlcCache.spin));
            fs_yield(context, &wakeEvent,
                     FS_YIELD_TIME_FOREVER, FS_YIELD_TYPE_REPLAY);
            if (!((wakeEvent.event == FS_THREAD_EVENT_REPLAY) &&
                  (wakeEvent.p1.u.p == NULL) && (wakeEvent.p2.u.p == NULL)))
            {
                fs_error(NULL, FS_INTERNAL_ERR,
                         "Received bad event %d %p %p "
                         "expected %d (nil) (nil)",
                         wakeEvent.event, wakeEvent.p1.u.p,
                         wakeEvent.p2.u.p, FS_THREAD_EVENT_REPLAY);
            }
            NFX_SPIN_LOCK(&(vlcCache.spin));
            FS_ASSERT(currLogSize <= vlcCache.used);
            //DLL_REMOVE(&(vlcCache.pendingMountList), vlcFsys, nextInLine);
            //fs_ramDealloc(FS_RAM_VLC, vlcFsys);
            FS_ELOG(context, LOG_INFO, FS_TRACE_LOG,
                    ("replay: filesystem dequeued and replay now"));
        }
        else
        {
            //FS_ELOG(context, LOG_INFO, FS_TRACE_LOG, ("replay now"));
            vlcCache.used += currLogSize;
            FS_ASSERT(vlcCache.used <= vlcCache.size);
        }
        NFX_SPIN_UNLOCK(&(vlcCache.spin));
    }    
    
#ifdef ISABELLE // test only
    eee_rtc(&t2);
#endif // ISABELLE

    /* Phase 1:
     * Read ahead the whole log.
     */
    FS_ELOG(context, LOG_INFO, FS_TRACE_LOG, ("replay: phase 1"));
    if ((logRoot->logDevType == FS_LOG_DEVTYPE_SELF) ||
        (logRoot->logDevType == FS_LOG_DEVTYPE_VOLID))
    {
        fs_logVolReadCompleteLog(context, logInfo);
    }

#ifdef ISABELLE // test only
    eee_rtc(&t3);
#endif // ISABELLE

    /* Phase 2: 
     * Scan for committed/done transactions.  All transactions
     * that were committed and not marked done will be indicated
     * in the replay bitmap.
     * At the same time start reading the meta needed for replay.
     */
    FS_ELOG(context, LOG_INFO, FS_TRACE_LOG, ("replay: phase 2"));
    rc = fs_logScan(context, logRoot, replay, FS_LOG_ACTION_DONE);
    if ((rc == NFX_OK) &&
        (logRoot->head == logRoot->tail) &&  (logRoot->currLen == 0))
    {
        FS_ELOG(context, LOG_INFO, FS_TRACE_LOG,
                ("replay: successfull [Empty Log]"));
        fs_logCleanupAfterReplay(context, logRoot, replay);
        goto fs_logReplay_exit;
    }

    if (skipReplay)
    {
        if (rc == NFX_OK)
        {
            FS_ELOG(context, LOG_INFO, FS_TRACE_LOG,
                    ("Scanning the log was successfull "));
        }
        else
        {
            FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                    ("Scanning the log failed "));
        }
        FS_ELOG(context, LOG_INFO, FS_TRACE_LOG, ("Skip the replay "));
        fs_logCleanupAfterReplay(context, logRoot, replay);
        goto fs_logReplay_exit;
    }

    if (rc != NFX_OK)
    {
        FS_ELOG(context, LOG_ERR, FS_TRACE_LOG,
                ("Error in scanning the log"));
        fs_logCleanupAfterReplay(context, logRoot, replay);
        goto fs_logReplay_exit;
    }
    
#ifdef ISABELLE // test only
    eee_rtc(&t4);
#endif // ISABELLE

    /* Phase 3:
     * Replay all records for transactions indicated in the replay bitmap.
     */
    FS_ELOG(context, LOG_INFO, FS_TRACE_LOG, ("replay: phase 3"));
    rc = fs_logScan(context, logRoot, replay, FS_LOG_ACTION_REPLAY);
    fs_logCleanupAfterReplay(context, logRoot, replay);

    fs_logReplay_exit:
    
    (devTypeTable->logReadDone)(context, NULL);
    
#ifdef ISABELLE // debug
    {
        eeeRTC_t diffTot, diff1, diff2, diff3, diff4;
        
        eee_rtc(&t5);
        FS_ELOG(context, LOG_INFO, FS_TRACE_LOG, ("replay: end"));
        eee_rtcSub(&t5, &t1, &diffTot);
        eee_rtcSub(&t2, &t1, &diff1);
        eee_rtcSub(&t3, &t2, &diff2);
        eee_rtcSub(&t4, &t3, &diff3);
        eee_rtcSub(&t5, &t4, &diff4);
        FS_ELOG(context, LOG_INFO, FS_TRACE_LOG,
                ("replay: time %lld (%lld + %lld + %lld + %lld) msec",
                 eee_rtcToMS(&diffTot),
                 eee_rtcToMS(&diff1), eee_rtcToMS(&diff2),
                 eee_rtcToMS(&diff3), eee_rtcToMS(&diff4)));
    }
#endif // ISABELLE

    /* Replay the waiting filesystem if any.
     */
    if ((logRoot->logDevType == FS_LOG_DEVTYPE_SELF) ||
        (logRoot->logDevType == FS_LOG_DEVTYPE_VOLID))
    {
        fs_vlcMount * currFsys = NULL;
        fs_vlcMount * nextFsys = NULL;
        
        NFX_SPIN_LOCK(&(vlcCache.spin));
        vlcCache.used -= currLogSize;
        FS_ASSERT(vlcCache.used >= 0);
        DLL_ITERATE_NEXT(currFsys, nextFsys,
                         &(vlcCache.pendingMountList), nextInLine)
        {
            if (currFsys->logSize <= (vlcCache.size - vlcCache.used))
            {
                // remove this replay from the pending list and wake it up
                esm_event_t event;
                int32       rc = NFX_OK;

                DLL_REMOVE(&(vlcCache.pendingMountList), currFsys, nextInLine);
                
                vlcCache.used += currFsys->logSize;
                FS_ASSERT(vlcCache.used <= vlcCache.size);
                
                event.event = FS_THREAD_EVENT_REPLAY;
                event.p1.u.p = NULL;
                event.p2.u.p = NULL;
                rc = thread_highPriorityEvent(currFsys->smHandle, &event);
                FS_ASSERT(rc == NFX_OK);             

                fs_ramDealloc(FS_RAM_VLC, currFsys);
            }
        } // DLL_ITERATE_NEXT
        NFX_SPIN_UNLOCK(&(vlcCache.spin));        
    }

    return rc;
} /* fs_logReplay */

/*----------------------------------------------------------------------------*/
#endif // SSC
