// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=4 shiftwidth=4 expandtab:
/*
 * Name:  fs-mkfs.c
 *
 * RCS:  $Id: fs-mkfs.c,v 1.115 2002/12/05 21:08:57 bob Exp $
 *
 * Description:  file system formatting
 *
 * Created By:  Paul Haahr <haahr@agilestorage.com>
 *
 * Date Created:  2001-09-03
 */

#if !defined(SSC)

#include "fs-includes.h"


#define TRACE_SPACE(args)   FS_TRACE(context, FS_TRACE_SPACE, args)


#define FS_MKFS_QUEUE_SIZE  THREAD_QUEUE_MAX


struct fs_BootAllocator
{
    /** Lock for this structure. */
    fs_Lock lock;

    /** The inode for the inode file. */
    fs_Inode *inodeInode;
    /** The inode for the root directory. */
    fs_Inode *rootInode;
    /** The inode for the ref count file. */
    fs_Inode *refbytesInode;
    /** The inode for the security file. */
    fs_Inode *securityInode;
    /** The inode for the full-inode-block summary file. */
    fs_Inode *fullInodeSummaryInode;
    /** The inode for the directory-in-inode-block summary file. */
    fs_Inode *directoryInodeSummaryInode;
    /** The inode for the binary buddy region file. */
    fs_Inode *buddyRegionInode;

    /** The inode for the fileset info file */
    fs_Inode *filesetInfoInode;
    /** Default User quota Inode */
    fs_Inode *defUserQuotaInode;
    /** Default group quota Inode */
    fs_Inode *defGroupQuotaInode;

    /** This inode is used for table of links in the filesystem */
    fs_Inode *linkInode;

    /** This inode is used to keep host temporary roots during restore */
    fs_Inode *tempDirInode;

    /** Dump swap dir */
    fs_Inode *tempDumpSwap;

    /** Audit file for CIFS auditing records */
    fs_Inode *auditInode;

    /** Extended attributes file */
    fs_Inode *eattrInode;

    /** Refbytes overflow file */
    fs_Inode *refOverflowInode;

    /** This inode is used for empty snapshot directory entries */
    fs_Inode *nullInode;

    /** Inode number for the lost+found directory */
    uint64   lostFoundInum;

    /** Blocks used in the live file system. */
    fs_BlockSet *used;
    /** Extra metadata allocated by the front-door allocation interface. */
    fs_BlockSet *more;
    /** Blocks allocated by the front-door allocation interface that should
     *  be marked as never being snapshotted. */
    fs_BlockSet *neverSnap;

    /** Block just prior to first block available to allocator. */
    fs_Block beginBlock;
    /** Total block count. */
    uint64 numBlocks;
};


/**
 * Allocate and return an extent of count blocks.  Simply
 * allocates starting at the first available block.
 */
static fs_Extent
fs_mkfsAllocateNextExtent(
    fs_Context       *context,
    fs_BootAllocator *boot,
    uint32           count)
{
    fs_Block    first, nextFree;
    uint32      numDone;

    numDone = 0;
    first = nextFree = fs_blockSetNextMissing(boot->used, boot->beginBlock);

    while (fs_blockIsValid(nextFree) &&
           fs_blockNumber(nextFree) < boot->numBlocks &&
           numDone < count)
    {
        numDone += 1;
        fs_blockSetAdd(boot->used, nextFree);
        nextFree = fs_blockSetNextMissing(boot->used, nextFree);
    }

    if (numDone != count)
    {
        FS_ELOG(context, LOG_ERR, FS_TRACE_SPACE,
                  ("fs_mkfsAllocateNextExtent failed to allocate extent"));
        fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
    }

    //printf("mkfsAllocateNextExtent %lld,%d\n",
    //        fs_blockNumber(first), count);

    return fs_makeExtent(count, first);
}


/**
 * Allocate an extent with the boot allocator.
 */
fs_Extent
fs_bootAllocateExtent(
    fs_Context  *context,
    fs_Inode    *inode,
    uint16      len)
{
    fs_BootAllocator    *boot = context->fileSystem->bootstrap;
    fs_Extent           extent;
    fs_Block            start;
    boolean             noSnap;
    uint64              inum;
    int                 i;

    noSnap = (inode != NULL &&
         (inode->dinode.attrs & FS_IFLG_NO_SNAP) != 0);
    inum = inode != NULL ? inode->inum : 0;

    fs_lockExclusive(context, &boot->lock, NULL);
    extent = fs_mkfsAllocateNextExtent(context, boot, len);

    start = fs_extentStartBlock(extent);
    for (i = 0; i < len; i++)
    {
        fs_Block block = fs_makeBlock(fs_blockNumber(start) + i);
        fs_blockSetAdd(boot->more, block);
        if (noSnap)
            fs_blockSetAdd(boot->neverSnap, block);
    }
    fs_unlockExclusive(context, &boot->lock);

    TRACE_SPACE(("bootAllocateExtent(%d) -> (%lld ... +%d) inum %lld",
                 len, fs_blockNumber(start), i, inum));
    
    //printf("bootAllocateExtent inum %lld want %d got %lld,%d\n",
    //        inum, len, fs_blockNumber(start), i);

    return extent;
}


/**
 * Commit our transaction log and (optionally) prepare a
 * new transaction log and sync the buffer cache.
 */
static void
fs_mkfsCommit(
    fs_Context *context,
    boolean    more)
{
    TRACE_SPACE(("mkfsCommit more %s", more ? "TRUE" : "FALSE"));

    // commit the log
    fs_logCommitEntry(context);

    // Let i/o completions take place, relinquish the processor

    fs_sleep(context, 0);

    // re-initialize the log if we need another transaction
    if (more)
    {
        fs_logBeginTransaction(context);
    }
    else
    {
        // sync the inode and buffer caches
        fs_syncThread(context);
    }
}


/**
 * Create the initial buddy bits freelist.
 */
static void fs_mkfsInitBuddyRegions(
    fs_Context       *context,
    fs_BootAllocator *boot)
{
    fs_Block    nextUsed;
    fs_Block    nextFree;
    fs_BlockSet *used = boot->used;
    uint64      last = context->fileSystem->params.totalBlocks;
    
    TRACE_SPACE(("mkfsInitBuddyRegions"));

    // disable boot allocation -- from here on, use the buddy system
    
    context->fileSystem->bootstrap = NULL;

    // traverse the used blocks, adding the missing extents to the buddy tree

    for (nextFree = fs_blockSetFirstMissing(used);
         fs_blockIsValid(nextFree) && fs_blockNumber(nextFree) < last;
         nextFree = fs_blockSetNextMissing(used, nextUsed))
    {
        nextUsed = fs_blockSetNext(used, nextFree);
        if (fs_blockIsValid(nextUsed)) {
            fs_buddyAddBlocks(context, nextFree, nextUsed, TRUE);
        }
        else {
            fs_buddyAddBlocks(context, nextFree, fs_makeBlock(last), TRUE);
            break;
        }
        fs_mkfsCommit(context, TRUE);
    }
}


/*
 * Write the superblocks to disk asynchronously and transactionally.
 */
void
fs_writeSuperblocksTrans(
    fs_Context      *context,
    fs_SuperBlock   *super)
{
    int i, n;

    for (i = 0, n = fs_superBlockCount(); i < n; i++) {
        fs_Buf *buf = B_GET_EXCL(NULL, fs_superBlockBlock(i));
        NFX_SPIN_LOCK(&context->fileSystem->superLock);
        super->crc32 = 0 ;
        super->crc32 = fs_checksum(super, sizeof *super, ~0);
        memcpy(buf->u.data, super, sizeof *super);
        NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
        FS_BUF_WRITE(buf, 0, sizeof *super);
    }
}


/**
 * Write the superblocks to disk synchronously, non-transactionally.
 */
void
fs_writeSuperblocksSync(
    fs_Context      *context,
    fs_SuperBlock   *super)
{
    int        i, n;
    fs_RdwrReq *rdwr;
    fs_SuperBlock   *mysuper;

    mysuper = fs_ramAlloc(FS_RAM_SUPER, super->blkSize);

    NFX_SPIN_LOCK(&context->fileSystem->superLock);
    super->crc32 = 0 ;
    super->crc32 = fs_checksum(super, sizeof(fs_SuperBlock), ~0);
    memcpy(mysuper, super, sizeof *super);
    NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
    for (i = 0, n = fs_superBlockCount(); i < n; i++)
    {
#ifdef FS_DEBUG_BUF_CKSUM
        fs_bufComputeCksum(context, fs_superBlockBlock(i), mysuper);
#endif

        rdwr = fs_rdwrAlloc();
        fs_volWriteBlock(context, rdwr, fs_superBlockBlock(i),
                         mysuper);
#ifdef FS_DEBUG_BUF_CKSUM
        fs_bufCompareCksum(context, fs_superBlockBlock(i), mysuper, NULL,
                            "write super sync done", TRUE);
#endif
        fs_rdwrDealloc(rdwr);
    }
    fs_ramDealloc(FS_RAM_SUPER, mysuper);
}


/**
 * Target number of blocks in the file system per inode.  This is used to
 * compute where the inode blocks are scattered in the system.  The target
 * can be changed for a file system, but it will eventually leave data
 * blocks further from their inodes than optimal.  Picking too low a
 * number will waste a little space.
 */
enum { FS_INITIAL_BLOCK_INODE_RATIO = 16 };


/**
 * Create the initial superblocks.
 */
void
fs_mkfsInitializeSuperblock(
    uint64          totalBlocks,
    fs_SuperBlock   *super,
    int32           blockSize)
{
    fs_Time now = fs_getTimeOfDay();

    bzero(super, blockSize);

    super->magic      = FS_SUPER_MAGIC;
    super->version    = FS_SUPER_CURRENT_VERSION;
    super->flags      = FS_SUPER_FLAG_MKFS; // flag incomplete in case we abort
    super->blkSize    = blockSize;
    super->numBlks    = totalBlocks;
    super->blockInodeRatio = FS_INITIAL_BLOCK_INODE_RATIO;
    super->mkfsTime   = now.secs;
    super->logInitTimeSec   = now.secs;
    super->logInitTimeNsec  = now.nsecs;
    super->snapshots  = 1;
    super->blksUsed   = 0;
    super->inodesUsed = 0;
    super->inodeFileBlks = 1;   // count the snodes (page 0)
    super->inodeFileSize = FS_BLOCK_SIZE;
    super->codePage = FS_CODEPAGE_DEFAULT;
    super->locale = FS_LOCALE_DEFAULT;
    
    strcpy(super->mountName, "(mkfs)");
}


/**
 * Create a file with the specified inode number containing the blocks
 * from the specified extent.
 */
static fs_Inode *
fs_mkfsCreateFileFromExtent(
    fs_Context  *context,
    fs_Extent   *extent,
    fs_Inumber  inum,
    uint16      format,
    uint32      linkCnt)
{
    fs_Parameters   *params = &context->fileSystem->params;
    fs_Inode        *inode = fs_inodeInstantiateMetanode(context, 0, inum);
    fs_Block        startBlock;
    uint32          extentLength;

    extentLength = extent ? fs_extentLength(*extent) : 0;
    startBlock = extent ? fs_extentStartBlock(*extent) : FS_INVALID_BLOCK;

    inode->dinode.attrs = format;
    inode->dinode.linkCnt = linkCnt;
    inode->dinode.size = extentLength * params->blockSize;

    // These inodes are never snapped.
    if (inum == FS_REF_COUNT_INODE || 
        inum == FS_FREE_BUDDY_TREE_INODE || 
        inum == FS_FULL_INODE_BLOCK_INODE ||
        inum == FS_DIRECTORY_IN_INODE_BLOCK_INODE)
    {
        inode->dinode.attrs |= FS_IFLG_NO_SNAP;
    }

    fs_emaproot_init(context, inode);
    if (extentLength > 0)
    {
        TRACE_SPACE(("mkfs emap_add inum %lld startblk %lld len %d",
                    inum, fs_blockNumber(startBlock), extentLength));
        fs_emap_add(context, inode, 0, *extent);
        inode->dinode.numblocks += fs_extentLength(*extent);
    }

    fs_inodeWrite(context, inode, INO_WRITE_LOG);

    return inode;
}


/**
 * Initialize the security descriptor for the
 * specified inode.
 */
static void
fs_mkfsSetInodeSecurityDesc(
    fs_Context  *context,
    fs_Inode    *inode,
    uint32      uid,
    uint32      gid,
    uint16      perm)
{
    fs_SecurityDescNfs  nfs;

    nfs.hdr.attr = FS_SDESC_TYPE_NFS;
    nfs.hdr.len = sizeof nfs;
    nfs.hdr.cont = 0;
    nfs.hdr.uQtId = nfs.hdr.gQtId = 0;
    nfs.uid = uid;
    nfs.gid = gid;
    nfs.mode = (perm & FS_IMODE_MASK);
    nfs.pad[0] = 0;

    fs_securityDescSet(context, inode, &nfs.hdr);
}


/**
 * A block of zero bytes.
 */
static char *zeroes;


/**
 * Fill all of the blocks of a file with zeros.
 */
void
fs_mkfsZeroFile(
    fs_Context 	  *context,
    fs_WriteQueue *queue,
    fs_Inode      *inode,
    uint64	  startPage)
{
    uint32 blockSize = context->fileSystem->params.blockSize;
    uint64 blocks = inode->dinode.size / blockSize;
    uint64 offset;

    assert(blockSize <= FS_MAX_BLOCK_SIZE);

    for (offset = startPage; offset < blocks; offset++)
    {
        fs_Block block = fs_emap_lookup_block(context, inode, offset);
        assert(fs_blockIsValid(block));
        fs_queueWriteBlock(queue, block, blockSize, zeroes);
    }

    fs_writeQueueSync(queue);
}


/**
 * Set the reference bits for the specified inode as allocator metadata to
 * indicate that it's in use and doesn't need to be copied-on-write.
 */
static void fs_markMetaDataRefs(
    fs_Context *context,
    fs_Inode   *inode,
    uint64     startPage,
    boolean    commit)
{
    uint32 blockSize = context->fileSystem->params.blockSize;
    uint64 blocks = inode->dinode.size / blockSize;
    uint64 offset;

    for (offset = startPage; offset < blocks; offset++)
    {
        fs_Block block = fs_emap_lookup_block(context, inode, offset);
        fs_refByteModifyBlock(context, fs_refbyteSetMeta, block);
        if (commit && offset && offset % 8 == 0)
        {
            // Don't hog too much of the buffer
            // cache.
            fs_mkfsCommit(context, TRUE);
        }
    }
}


/**
 * Mark the blocks in the specified block set as live in the file system.
 */
static void fs_mkfsMarkBlockSetLive(
    fs_Context  *context,
    fs_BlockSet *blocks)
{
    fs_Block block;

    for (block = fs_blockSetFirst(blocks);
         fs_blockIsValid(block);
         block = fs_blockSetNext(blocks, block))
    {
        fs_refByteModifyBlock(context, fs_refbyteForceLive, block);
    }
}


/**
 * Mark the blocks in the specified block set as never snapshot.
 */
static void fs_mkfsMarkBlockSetNeverSnap(
    fs_Context  *context,
    fs_BlockSet *blocks)
{
    fs_Block block;

    for (block = fs_blockSetFirst(blocks);
         fs_blockIsValid(block);
         block = fs_blockSetNext(blocks, block))
    {
        fs_refByteModifyBlock(context, fs_refbyteDisableSnapshots, block);
    }
}


/**
 * Initialize the bootstrap allocator.
 */
static void fs_mkfsInitBoot(
    fs_Context       *context,
    fs_BootAllocator *boot,
    fs_Block         beginBlock,
    uint64           numBlocks)
{
    TRACE_SPACE(("mkfsInitBoot"));

    bzero(boot, sizeof *boot);

    fs_lockInit(&boot->lock, FS_LOCK_TYPE_BOOT);

    boot->used      = fs_makeBlockSet(context);
    boot->more      = fs_makeBlockSet(context);
    boot->neverSnap = fs_makeBlockSet(context);

    boot->beginBlock = beginBlock;
    boot->numBlocks = numBlocks;
}




/**
 * Mount the file system enough to get some work done during formatting.
 */
static fs_FileSystem *
fs_mkfsMount(
    fs_Context       *context,
    fs_SuperBlock    *super,
    fs_BootAllocator *boot)
{
    int32 rc;
    fs_FileSystem       *fs;

    TRACE_SPACE(("mkfsMount"));

    // initialize our context

    fs = context->fileSystem = fs_makeFileSystem(context, super);
    if (fs == NULL)
    {
        return NULL;
    }
    fs->bootstrap = boot;

    // Add ourselves to active thread list for the fs to
    // ensure exception during mkfs is properly handled.
    NFX_SPIN_LOCK(&fs->spin);
    DLL_INSERT_TAIL(&fs->activeThreadHdr, context, fsEntry);
    NFX_SPIN_UNLOCK(&fs->spin);

    // No logging needed during the mkfs
    fs->logRoot->logDevType = FS_LOG_DEVTYPE_NULL;
    fs->logDevTypeTable = &fs_logDevTypeTable[FS_LOG_DEVTYPE_NULL];
    fs->persistentUserBufCache = FALSE;

    /*
     * Start the writeback daemons.
     */
    fs_bufStartWritebackDaemon(context);
    
    rc = fs_logInitDevice(context);
    if (rc != NFX_OK) {
        fs_warn(context, FS_INTERNAL_ERR,
                 "failed to init log device: %d", rc);
        return NULL;
    }

    fs_logBeginTransaction(context);

    return fs;
}


/**
 * Zero the first block of the inode file in each snapshot.
 */
static void fs_mkfsZeroMetaInodes(
    fs_Context *context)
{
    int i;
    uint32 blockSize = context->fileSystem->params.blockSize;

    TRACE_SPACE(("mkfsZeroMetaInodes"));

    for (i = 0; i < FS_MAX_SNAPSHOTS; i++) {
        fs_Block block = fs_firstInodeBlock(i);
        fs_Buf *buf = fs_bufGetBlock(context, NULL, block, BUF_GET_MODE_EXCL,
                                     BUF_GET_OPT_SLEEP | BUF_GET_OPT_NEW);
        bzero(buf->u.data, blockSize);
        FS_BUF_ZERO_ALL(buf);
    }
}


/**
 * Create and fill in the inodes used for allocation structures.
 */
static void
fs_mkfsCreateMetaInodes(
    fs_Context          *context,
    fs_BootAllocator    *boot,
    fs_Extent           *refbytes,
    fs_Extent           *buddies,
    fs_Extent           *inosum,
    fs_Extent           *dirsum)
{
    fs_Extent       extent;

    TRACE_SPACE(("mkfsCreateMetaInodes"));

    // create the inode file
    extent = fs_makeExtent(1, fs_makeBlock(1));
    boot->inodeInode = 
        fs_mkfsCreateFileFromExtent(context, &extent,
                                    FS_INODE_INODE,
                                    FS_IFMT_REG, 0);
    fs_mkfsSetInodeSecurityDesc(context, boot->inodeInode, 0, 0, 0700);

    // create refbytes file
    boot->refbytesInode =
        fs_mkfsCreateFileFromExtent(context, refbytes,
                                    FS_REF_COUNT_INODE,
                                    FS_IFMT_REG, 0);
    fs_mkfsSetInodeSecurityDesc(context, boot->refbytesInode,
                                0, 0, 0700);

    // create buddy bits file
    boot->buddyRegionInode =
        fs_mkfsCreateFileFromExtent(context, buddies,
                                    FS_FREE_BUDDY_TREE_INODE,
                                    FS_IFMT_REG, 0);
    fs_mkfsSetInodeSecurityDesc(context, boot->buddyRegionInode,
                                0, 0, 0700);

    // create inode summary file
    boot->fullInodeSummaryInode =
        fs_mkfsCreateFileFromExtent(context, inosum,
                                    FS_FULL_INODE_BLOCK_INODE,
                                    FS_IFMT_REG, 0);
    fs_mkfsSetInodeSecurityDesc(context, boot->fullInodeSummaryInode,
                                0, 0, 0700);

    // create directory summary file
    boot->directoryInodeSummaryInode =
        fs_mkfsCreateFileFromExtent(context, dirsum,
                                    FS_DIRECTORY_IN_INODE_BLOCK_INODE,
                                    FS_IFMT_REG, 0);
    fs_mkfsSetInodeSecurityDesc(context, boot->directoryInodeSummaryInode,
                                0, 0, 0700);

    // create fileset info file
    boot->filesetInfoInode = 
        fs_mkfsCreateFileFromExtent(context, NULL,
                                    FS_FILESET_INFO_INODE,
                                    FS_IFMT_REG, 1);
    fs_mkfsSetInodeSecurityDesc(context, boot->filesetInfoInode,
                                0, 0, 0700);

    // create user quota file
    boot->defUserQuotaInode = 
        fs_mkfsCreateFileFromExtent(context, NULL,
                                    FS_FILESET_USER_QUOTA_INODE,
                                    FS_IFMT_QUOT, 1);
    fs_mkfsSetInodeSecurityDesc(context, boot->defUserQuotaInode,
                                0, 0, 0700);

    // create group quota file
    boot->defGroupQuotaInode = 
        fs_mkfsCreateFileFromExtent(context, NULL,
                                    FS_FILESET_GROUP_QUOTA_INODE,
                                    FS_IFMT_QUOT, 1);
    fs_mkfsSetInodeSecurityDesc(context, boot->defGroupQuotaInode,
                                0, 0, 0700);

    // create link file
    boot->linkInode = 
        fs_mkfsCreateFileFromExtent(context, NULL,
                                    FS_LINK_INODE,
                                    FS_IFMT_REG, 1);
    fs_mkfsSetInodeSecurityDesc(context, boot->linkInode,
                                0, 0, 0700);

    // Initialize Temp Dir Inode
    boot->tempDirInode = fs_inodeInstantiateMetanode(context, 0,
                                                     FS_TEMPDIR_INODE);
    fs_dirroot_init(context, boot->tempDirInode, FS_ROOT_INODE);
    boot->tempDirInode->dinode.attrs =  FS_IFMT_DIR | FS_IFLG_COMPACT_DIR;
    boot->tempDirInode->dinode.linkCnt = 2;
    fs_inodeWrite(context, boot->tempDirInode, 0);
    fs_mkfsSetInodeSecurityDesc(context, boot->tempDirInode,
                                0, 0, 0777);
    
    // Initialize Temp Dir Inode
    boot->tempDumpSwap = fs_inodeInstantiateMetanode(context, 0,
                                                     FS_TEMP_DUMP_SWAP);
    fs_dirroot_init(context, boot->tempDumpSwap, FS_ROOT_INODE);
    boot->tempDumpSwap->dinode.attrs =  FS_IFMT_DIR | FS_IFLG_COMPACT_DIR;
    boot->tempDumpSwap->dinode.linkCnt = 2;
    fs_inodeWrite(context, boot->tempDumpSwap, 0);
    fs_mkfsSetInodeSecurityDesc(context, boot->tempDumpSwap,
                                0, 0, 0777);

    // Initialize null inode
    boot->nullInode = fs_inodeInstantiateMetanode(context, 0, FS_NULL_INODE);
    boot->nullInode->dinode.attrs = FS_IFMT_REG;
    boot->nullInode->dinode.linkCnt = 1;
    fs_emaproot_init(context, boot->nullInode);
    boot->nullInode->dinode.size = 0;
    fs_inodeWrite(context, boot->nullInode, 0);
    fs_mkfsSetInodeSecurityDesc(context, boot->nullInode,
                                0, 0, 0700);

    // Initialize audit file
    boot->auditInode = fs_inodeInstantiateMetanode(context, 0, FS_AUDIT_INODE);
    boot->auditInode->dinode.attrs = FS_IFMT_REG;
    boot->auditInode->dinode.linkCnt = 1;
    fs_emaproot_init(context, boot->auditInode);
    boot->auditInode->dinode.size = 0;
    fs_inodeWrite(context, boot->auditInode, 0);
    fs_mkfsSetInodeSecurityDesc(context, boot->auditInode,
                                0, 0, 0700);

    // Initialize extended attributes file
    boot->eattrInode = fs_inodeInstantiateMetanode(context, 0,
                                                   FS_EXTENDED_ATTR_INODE);
    boot->eattrInode->dinode.attrs = FS_IFMT_REG;
    boot->eattrInode->dinode.linkCnt = 1;
    fs_emaproot_init(context, boot->eattrInode);
    boot->eattrInode->dinode.size = 0;
    fs_inodeWrite(context, boot->eattrInode, 0);
    fs_mkfsSetInodeSecurityDesc(context, boot->eattrInode,
                                0, 0, 0700);

    // Initialize ref overflow inode
    boot->refOverflowInode = fs_inodeInstantiateMetanode(context, 0,
                                                   FS_REF_OVERFLOW_INODE);
    boot->refOverflowInode->dinode.attrs = FS_IFMT_REG;
    boot->refOverflowInode->dinode.linkCnt = 1;
    fs_emaproot_init(context, boot->refOverflowInode);
    boot->refOverflowInode->dinode.size = 0;
    fs_inodeWrite(context, boot->refOverflowInode, 0);
    fs_mkfsSetInodeSecurityDesc(context, boot->refOverflowInode,
                                0, 0, 0700);

    FS_TRACE(context, FS_TRACE_SPACE, ("linkInode is %lld size is %lld",
            boot->linkInode->inum, boot->linkInode->dinode.size));
    FS_TRACE(context, FS_TRACE_SPACE, ("tempDirInode is %lld size is %lld",
            boot->tempDirInode->inum, boot->tempDirInode->dinode.size));
}


/**
 * Zero initialize space and inode allocation files.
 */
static void
fs_mkfsZeroMetaData(
    fs_WriteQueue    *queue,
    fs_BootAllocator *boot)
{
    fs_Context *context = queue->context;

    FS_TRACE(context, FS_TRACE_SPACE, ("mkfsZeroMetaData"));

    fs_mkfsZeroFile(context, queue, boot->refbytesInode, 0);
    fs_mkfsZeroFile(context, queue, boot->buddyRegionInode, 0);
    fs_mkfsZeroFile(context, queue, boot->fullInodeSummaryInode, 0);
    fs_mkfsZeroFile(context, queue, boot->directoryInodeSummaryInode, 0);
}


/**
 * Acquire exclusive locks on the metadata inodes used for allocation
 * strucutures, and store the resultant inodes in the bootstrap object.
 */
static void fs_mkfsAcquireMetaInodes(
    fs_Context       *context,
    fs_BootAllocator *boot)
{
    TRACE_SPACE(("mkfsAcquireMetaInodes"));

    boot->inodeInode
        = I_GET_EXCL(0, FS_INODE_INODE);
    boot->refbytesInode
        = I_GET_EXCL(0, FS_REF_COUNT_INODE);
    boot->buddyRegionInode
        = I_GET_EXCL(0, FS_FREE_BUDDY_TREE_INODE);
    boot->securityInode
        = I_GET_EXCL(0, FS_SECURITY_INODE);
    boot->fullInodeSummaryInode
        = I_GET_EXCL(0, FS_FULL_INODE_BLOCK_INODE);
    boot->directoryInodeSummaryInode
        = I_GET_EXCL(0, FS_DIRECTORY_IN_INODE_BLOCK_INODE);
    boot->filesetInfoInode 
        = I_GET_EXCL(0, FS_FILESET_INFO_INODE);
    boot->defUserQuotaInode 
        = I_GET_EXCL(0, FS_FILESET_USER_QUOTA_INODE);
    boot->defGroupQuotaInode 
        = I_GET_EXCL(0, FS_FILESET_GROUP_QUOTA_INODE);
    boot->linkInode
        = I_GET_EXCL(0, FS_LINK_INODE);
    boot->tempDirInode
        = I_GET_EXCL(0, FS_TEMPDIR_INODE);
    boot->tempDumpSwap
        = I_GET_EXCL(0, FS_TEMP_DUMP_SWAP);
    boot->nullInode
        = I_GET_EXCL(0, FS_NULL_INODE);

    assert(boot->inodeInode != NULL);
    assert(boot->refbytesInode != NULL);
    assert(boot->buddyRegionInode != NULL);
    assert(boot->securityInode != NULL);
    assert(boot->fullInodeSummaryInode != NULL);
    assert(boot->directoryInodeSummaryInode != NULL);
    assert(boot->filesetInfoInode != NULL);
    assert(boot->defUserQuotaInode != NULL);
    assert(boot->defGroupQuotaInode != NULL);
    assert(boot->linkInode != NULL);
    assert(boot->tempDirInode != NULL);
    assert(boot->nullInode != NULL);
}


/**
 * Create the root directory.
 */
static void fs_mkfsCreateRootDirectory(
    fs_Context       *context,
    fs_BootAllocator *boot)
{
    TRACE_SPACE(("mkfsCreateRootDirectory"));

    boot->rootInode = fs_inodeInstantiateMetanode(context, 0, FS_ROOT_INODE);
    fs_dirroot_init(context, boot->rootInode, FS_ROOT_INODE);

    /*
     * By default only write access only to root for root directory
     */

    boot->rootInode->dinode.attrs = FS_IFMT_DIR | FS_IFLG_COMPACT_DIR;
    boot->rootInode->dinode.linkCnt = 2;
    fs_mkfsSetInodeSecurityDesc(context, boot->rootInode,
                                0, 0, 0755);
    fs_inodeWrite(context, boot->rootInode, INO_WRITE_LOG);
}


/**
 * Create the trashcan directory.
 */
static void fs_mkfsCreateTrashDirectory(
    fs_Context       *context)
{
    fs_Inode *inode;

    TRACE_SPACE(("mkfsCreateTrashDirectory"));

    inode = fs_inodeInstantiateMetanode(context, 0, FS_TRASHCAN_INODE);
    fs_dirroot_init(context, inode, FS_ROOT_INODE);
    inode->dinode.attrs = FS_IFMT_DIR | FS_IFLG_COMPACT_DIR;
    inode->dinode.linkCnt = 2;
    fs_mkfsSetInodeSecurityDesc(context, inode,
                                0, 0, 0777);
    fs_inodeWrite(context, inode, INO_WRITE_LOG);
}


/**
 * Create the system directory.
 */
static void fs_mkfsCreateSystemDirectory(
    fs_Context       *context)
{
    fs_Inode *inode;

    TRACE_SPACE(("mkfsCreateSystemDirectory"));

    inode = fs_inodeInstantiateMetanode(context, 0, FS_SYSTEM_HIDDEN_INODE);
    fs_dirroot_init(context, inode, FS_ROOT_INODE);
    inode->dinode.attrs = FS_IFMT_DIR | FS_IFLG_COMPACT_DIR;
    inode->dinode.linkCnt = 2;
    fs_mkfsSetInodeSecurityDesc(context, inode,
                                0, 0, 0755);
    fs_inodeWrite(context, inode, INO_WRITE_LOG);
}


/**
 * Create the cifs shares directory.
 */
static void fs_mkfsCreateCifsSharesDirectory(
    fs_Context       *context)
{
    fs_Inode *inode;

    TRACE_SPACE(("mkfsCreateCifsSharesDirectory"));

    inode = fs_inodeInstantiateMetanode(context, 0, FS_CIFS_SHARES_INODE);
    fs_dirroot_init(context, inode, FS_ROOT_INODE);
    inode->dinode.attrs = FS_IFMT_DIR | FS_IFLG_COMPACT_DIR;
    inode->dinode.linkCnt = 2;
    fs_mkfsSetInodeSecurityDesc(context, inode,
                                0, 0, 0755);
    fs_inodeWrite(context, inode, INO_WRITE_LOG);
}


/**
 * Create the /lost+found directory.
 */
static void fs_mkfsCreateLostFoundDirectory(
    fs_Context *context,
    fs_BootAllocator *boot)
{
    fs_Inode *lf;
    fs_Inode *root = boot->rootInode;
    char *const name = "lost+found";

    TRACE_SPACE(("mkfsCreateLostFoundDirectory"));
    lf = fs_allocateDirectoryInode(context, root, 0700, 0, 0,
                                   root->dinode.fsetNum);
    fs_dir_addent(context, root, name, strlen(name), lf->inum);
    (void)fs_addLink(context, 0, lf, root->inum);
    ++root->dinode.linkCnt;
    fs_inodeWrite(context, root, INO_WRITE_LOG);
    boot->lostFoundInum = lf->inum;
    fs_mkfsSetInodeSecurityDesc(context, lf,
                                0, 0, 0700);
    fs_inodeWrite(context, lf, INO_WRITE_LOG);
}


/**
 * Create the /.snapshot directory.
 */
static void fs_mkfsCreateSnapshotDirectory(
    fs_Context *context,
    fs_BootAllocator *boot)
{
    fs_Inode *inode;
    fs_Time now = fs_getTimeOfDay();

    TRACE_SPACE(("mkfsCreateSnapshotDirectory"));

    inode = fs_inodeInstantiateMetanode(context, 0, FS_SNAPSHOT_DIR_INODE);
    fs_dirroot_init(context, inode, FS_ROOT_INODE);
    inode->dinode.attrs = FS_IFMT_DIR | FS_IFLG_COMPACT_DIR;
    inode->dinode.linkCnt = 2;
    inode->dinode.fsetNum = 0;
    inode->dinode.ctime = now;
    inode->dinode.mtime = now;
    inode->dinode.atime = now;
    fs_mkfsSetInodeSecurityDesc(context, inode,
                                0, 0, 0777);
    fs_inodeWrite(context, inode, INO_WRITE_LOG);
}


/*
 * Create the intial fileset.
 * Do we need to account for root inode?
 */

static void fs_mkfsCreateDefaultFileset(
    fs_Context *context,
    fs_BootAllocator *boot)
{
    FS_ASSERT(fs_lockOwnedExclusive(context, &boot->rootInode->lock));
    FS_ASSERT(fs_lockOwnedExclusive(context, &boot->defUserQuotaInode->lock));
    FS_ASSERT(fs_lockOwnedExclusive(context, &boot->defGroupQuotaInode->lock));
    FS_TRACE(context, FS_TRACE_REQ, ("def User Quota Inode is %lld\n",
             boot->defUserQuotaInode->inum));
    FS_TRACE(context, FS_TRACE_REQ, ("def Group Quota Inode is %lld\n",
             boot->defGroupQuotaInode->inum));
    fs_makeDefaultFileSet(context, context->fileSystem, NULL, boot->rootInode,
                          boot->filesetInfoInode);
    fs_inodeReleaseExclusive(context, boot->defUserQuotaInode);
    fs_inodeReleaseExclusive(context, boot->defGroupQuotaInode);
    FS_TRACE(context, FS_TRACE_REQ, ("returned  from mkfileset"));
}


/**
 * Mark the metadata corresponding to the metadata in the refcount and
 * inode summary files.
 */
static void fs_mkfsMarkMetaData(
    fs_Context       *context,
    fs_BootAllocator *boot)
{
    int i, n;
    fs_Parameters *p = &context->fileSystem->params;

    TRACE_SPACE(("mkfsMarkMetaData"));

    for (i = 0, n = fs_superBlockCount(); i < n; i++)
    {
        fs_refByteModifyBlock(context, fs_refbyteSetMeta,
                              fs_superBlockBlock(i));
    }
    fs_refByteModify(context, fs_refbyteSetMeta, FS_MAX_SNAPSHOTS,
                     fs_firstInodeBlock(0));

    fs_markMetaDataRefs(context, boot->refbytesInode, 0, TRUE);
    fs_markMetaDataRefs(context, boot->buddyRegionInode, 0, TRUE);
    fs_markMetaDataRefs(context, boot->fullInodeSummaryInode, 0, TRUE);
    fs_markMetaDataRefs(context, boot->directoryInodeSummaryInode, 0, TRUE);

    // mark past the end of the file system in the ref counts
    fs_refByteModifyPastEnd(context, fs_refbyteSetMeta, p);

    // mark the first bits in the inode summary files
    for (i = 0, n = 0; n < FS_MAX_RESERVED_INODE; i++, n += FS_INODES_PER_PAGE)
    {
        fs_bitmapSet(context, boot->fullInodeSummaryInode, i, NULL);
        fs_bitmapSet(context, boot->directoryInodeSummaryInode, i, NULL);
    }

    // mark past the end of the file system in the inode summaries
    fs_bitmapSetBitsAfter(context,
                          boot->fullInodeSummaryInode,
                          p->maxInodeBlocks);
    fs_bitmapSetBitsAfter(context,
                          boot->directoryInodeSummaryInode,
                          p->maxInodeBlocks);

    fs_mkfsCommit(context, TRUE);
}


/**
 * Initialize timestamps for the metadata inodes.
 */
static void fs_mkfsInitTimeStamps(
    fs_Context       *context,
    fs_BootAllocator *boot)
{
    fs_Time now = fs_getTimeOfDay();

    TRACE_SPACE(("mkfsInitTimeStamps"));

#define UPDATE(_inode_) do {                            \
        fs_Inode *__inode = boot->_inode_;              \
        __inode->dinode.ctime = now;                    \
        __inode->dinode.mtime = now;                    \
        __inode->dinode.atime = now;                    \
        fs_inodeWrite(context, __inode, INO_WRITE_LOG); \
    } while (0)
    
    UPDATE(inodeInode);
    UPDATE(refbytesInode);
    UPDATE(securityInode);
    UPDATE(buddyRegionInode);
    UPDATE(fullInodeSummaryInode);
    UPDATE(directoryInodeSummaryInode);
    UPDATE(rootInode);
    UPDATE(filesetInfoInode);
    UPDATE(defUserQuotaInode);
    UPDATE(defGroupQuotaInode);
    UPDATE(linkInode);
    UPDATE(tempDirInode);
    UPDATE(tempDumpSwap);
    UPDATE(nullInode);

#undef  UPDATE
}


/**
 * Initialize the security file.
 */
static void fs_mkfsInitSecurityFile(
    fs_Context       *context,
    fs_BootAllocator *boot)
{
    fs_SecurityDescNfs  *nfs;
    fs_Buf              *buf;

    // Create the security file

    boot->securityInode = fs_inodeInstantiateMetanode(context, 0,
                                        FS_SECURITY_INODE);
    boot->securityInode->dinode.attrs = FS_IFMT_REG;
    boot->securityInode->dinode.linkCnt = 1;
    fs_emaproot_init(context, boot->securityInode);
    boot->securityInode->dinode.size = 0;
    fs_inodeWrite(context, boot->securityInode, 0);

    buf = O_CREATE(boot->securityInode, 0);
    boot->securityInode->dinode.size = FS_BLOCK_SIZE;
    bzero(buf->u.data, FS_BLOCK_SIZE);
    FS_BUF_ZERO_ALL(buf);

    nfs = &buf->u.sdnfs->nfs[0];

    nfs->hdr.attr = FS_SDESC_TYPE_NFS;
    nfs->hdr.len = sizeof(fs_SecurityDescNfs);
    nfs->uid = 0;
    nfs->gid = 0;
    nfs->mode = 0700;
    nfs->hdr.cksum = uint32_checksum(nfs, sizeof *nfs, 0);

    nfs += 1;
    nfs->hdr.attr = FS_SDESC_TYPE_NFS;
    nfs->hdr.len = sizeof(fs_SecurityDescNfs);
    nfs->uid = 0;
    nfs->gid = 0;
    nfs->mode = 0755;
    nfs->hdr.cksum = uint32_checksum(nfs, sizeof *nfs, 0);

    nfs += 1;
    nfs->hdr.attr = FS_SDESC_TYPE_NFS;
    nfs->hdr.len = sizeof(fs_SecurityDescNfs);
    nfs->uid = 0;
    nfs->gid = 0;
    nfs->mode = 0777;
    nfs->hdr.cksum = uint32_checksum(nfs, sizeof *nfs, 0);

    FS_BUF_WRITE(buf, 0, 3 * sizeof *nfs);

}


#ifdef NOTDEF // debug only
#ifdef FS_STANDALONE
/**
 * Dump all the refbytes blocks to standard output.
 */
static void fs_refbytesDump(
    fs_Context *context,
    fs_BootAllocator *boot,
    int blockSize)
{
    uint64 i, n;
    for (i = 0, n = boot->refbytesInode->dinode.size / blockSize;
         i < n;
         i++)
    {
        char name[100];
        fs_Buf *buf;
        sprintf(name, "mkfs refBlock %lld", i);
        buf = O_GET_SHARED(boot->refbytesInode, i);
        memdump(name, buf->u.refBytes, blockSize);
        fs_bufRelease(context, buf);
    }
}

/**
 * Write a non-zero pattern.
 */
static void fs_mkfsVolumeInit(
    fs_Context *context,
    uint64      numBlocks,
    uint32      blockSize)
{
    int         fd;
    int32       i, num;
    int32       *blkPtr = malloc(FS_BLOCK_SIZE);
    extern char *volName[FS_SIM_VOL_MAX + 1];
   
    if (blkPtr == NULL)
    {
        fs_error(NULL, FS_SIM_OPEN_ERR, "blkPtr == NULL");
    }

    fd = open(  volName[context->volId.whole_thing],
                O_RDWR | O_CREAT, S_IREAD | S_IWRITE);
    if (fd <= 0)
    {
        perror("open");
        fs_error(NULL, FS_SIM_OPEN_ERR, "fs_mkfsVolumeInit");
    }

    for (i = 0; i < FS_BLOCK_SIZE / sizeof(*blkPtr); i++)
    {
        *(blkPtr + i) = 0x6c6c6c6c;
    }

    for (i = 0; i < numBlocks; i += 1)
    {
        num = write(fd, blkPtr, blockSize);
        if (num != blockSize)
        {
            perror("write");
            fs_error(NULL, FS_SIM_ARGS_ERR, "fs_mkfsVolumeInit");
        }
    }

    free(blkPtr);
}

#else

/**
 * Write a non-zero pattern.
 */
static void fs_mkfsVolumeInit(
    fs_Context *context,
    uint64      numBlocks,
    uint32      blockSize)
{
    fs_WriteQueue   queue;
    char            *initData;
    int32           i;

    fs_initWriteQueue(context, &queue, FS_MKFS_QUEUE_SIZE);
    initData = fs_ramAlloc(FS_RAM_CORE_MISC, blockSize + MALLOC_ALIGNED_SIZE);
    initData += (MALLOC_ALIGNED_SIZE - 1);
    (address_t) initData &= ~(MALLOC_ALIGNED_SIZE - 1);

    memset(initData, 0x6c, blockSize);
        
    for (i = 0; i < numBlocks; i++)
    {
        fs_Block block = fs_makeBlock(i);
        fs_queueWriteBlock(&queue, block, blockSize, initData);

        if (i && i % FS_MKFS_QUEUE_SIZE == 0)
            fs_writeQueueSync(&queue);
    }

    fs_writeQueueSync(&queue);

    fs_ramDealloc(FS_RAM_CORE_MISC, initData);
}
#endif
#endif //  NOTDEF (debug only)

/*
 * Smaller filesystems need a larger relative log space
 * to keep up with the potential workload.  There's probably
 * a better way to do this - TODO.
 */
// in 8K blocks
#define FS_LOG_FSSIZE_SM                    (20 * 1024) // Small filesystem
#define FS_LOG_FSSIZE_MD              (2 * 1024 * 1024) // Medium filesystem
#define FS_LOG_FSSIZE_LG            (100 * 1024 * 1024) // Large filesystem

// percentage of blocks reserved for logging  
#define FS_LOG_SELF_RESERVED_SM       50   // 2.0 %
#define FS_LOG_SELF_RESERVED_MD      200   // 0.5 %
#define FS_LOG_SELF_RESERVED_LG      500   // 0.2 %
#define FS_LOG_SELF_RESERVED_XL     1000   // 0.1 %


void
fs_mkfsGetLogSize(uint64 *logSize, uint64 *numBlocks, uint32 logDevType)
{
    uint64 logReserveSize, logNbrBlocks;
    
    *logSize = 0;

    // Reserve space for logging
    switch (logDevType)
    {
    case FS_LOG_DEVTYPE_NVRAM:
        // Todo: 
        *logSize = 2 * 1024 * 1024;  // 2M bytes
        break;
    case FS_LOG_DEVTYPE_VOLID:
        // logSize: Must be defined in the mount req command
        *logSize = 0;
        break;
    case FS_LOG_DEVTYPE_SELF:
        if (*numBlocks <= FS_LOG_FSSIZE_SM)
            logReserveSize = FS_LOG_SELF_RESERVED_SM;
        else if (*numBlocks <= FS_LOG_FSSIZE_MD)
            logReserveSize = FS_LOG_SELF_RESERVED_MD;
        else if (*numBlocks <= FS_LOG_FSSIZE_LG)
            logReserveSize = FS_LOG_SELF_RESERVED_LG;
        else
            logReserveSize = FS_LOG_SELF_RESERVED_XL;

        logNbrBlocks = *numBlocks / logReserveSize;
        if (logNbrBlocks < FS_LOG_MIN_NBR_BLOCKS)
            logNbrBlocks = FS_LOG_MIN_NBR_BLOCKS;
        if (logNbrBlocks > FS_LOG_MAX_NBR_BLOCKS)
            logNbrBlocks = FS_LOG_MAX_NBR_BLOCKS;

        // reserve in this filesystem a log area at the end
        *numBlocks -= logNbrBlocks;

        // The log start in the next block
        logNbrBlocks--;
        // calculate the payload log size in bytes (subtract all overhead)
        *logSize = logNbrBlocks * FS_BLOCK_SIZE -
            sizeof(vl_LogHdr) -
            sizeof(vl_BufDiskHdr) * logNbrBlocks * FS_BLOCK_SIZE / VL_BUFFER_SIZE;
        break;
    case FS_LOG_DEVTYPE_NULL:
    default:
        break;
    }

    // Make sure its an even multiple of segments.
    *logSize = (*logSize / FS_LOG_MAX_SEGMENTS) * FS_LOG_MAX_SEGMENTS;
} 


#if !defined(FS_STANDALONE)
/**
 * Notify listners that a volume has been formatted, so that
 * they can cleanup their states. This is required by ndmp to
 * clean up its dump dates files
 */
static void
fs_sendMkfsEvent(evm_volId_t volId)
{
    nfx_eventMsg_t          msg;

    msg.eventType = NFX_EVENT_MKFS;
    msg.param.mkfsEvent.volId = volId;

    event_postEvent(&msg);
}
#endif // FS_STANDALONE

/**
 * Create a file system on the volume specified by the context argument,
 * with the given number of blocks and block size.
 */
boolean
fs_formatFileSystem(
    fs_Context  *context,
    uint64      numBlocks,
    uint32      blockSize,
    boolean     extLog)
{
    fs_BootAllocator boot;
    fs_SuperBlock   *super = fs_ramAlloc(FS_RAM_SUPER, blockSize);
    fs_WriteQueue   queue;
    fs_Parameters   params;
    fs_Extent       inosumExtent;
    fs_Extent       dirsumExtent;
    fs_Extent       buddiesExtent;
    fs_Extent       refbytesExtent;
    uint32          logDevType;
    uint64          logSize;
    uint64          adjustedNumBlocks = numBlocks;
    char            *zeroesMalloc;
    int             i, loop;

    //fs_mkfsVolumeInit(context, numBlocks, blockSize);

    FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
              ("Mounting for mkfs"));

    // Allocate a max block size chunk of zeroes, and ensure
    // its aligned such that the sibyte won't choke on it.
    zeroes = zeroesMalloc = fs_ramAlloc(FS_RAM_CORE_MISC, FS_MAX_BLOCK_SIZE +
                                                    MALLOC_ALIGNED_SIZE);
    zeroes += (MALLOC_ALIGNED_SIZE - 1);
    (address_t) zeroes &= ~(MALLOC_ALIGNED_SIZE - 1);
    bzero(zeroes, FS_MAX_BLOCK_SIZE);

    if (extLog) {
        // can later be set with the mount cmd to the correct value
        logDevType = FS_LOG_DEVTYPE_NVRAM;
    }
    else {
        logDevType = FS_LOG_DEVTYPE_SELF;       // default
    }
    fs_mkfsGetLogSize(&logSize, &adjustedNumBlocks, logDevType);

    fs_initWriteQueue(context, &queue, FS_MKFS_QUEUE_SIZE);

    // phase 0: get something mounted
    //printf("vol %llx phase 0: get something mounted\n",
    //    context->volId.whole_thing);

    fs_mkfsInitBoot(context, &boot,
                    fs_makeBlock(fs_blockNumber(fs_superBlockBlock(1))),
                    adjustedNumBlocks);

    fs_mkfsInitializeSuperblock(adjustedNumBlocks, super, blockSize);

    params = fs_computeSizeParameters(super);

    //fs_printParameters(&params);

    // set default snap state
    fs_snapSetDefaultState(&super->snapState);

    // default values, Todo use optional parameters
    // super->logDevType = FS_LOG_DEVTYPE_SELF;
    super->logSize = (uint32)logSize;
    super->logSpaceSize = (uint32)logSize;
    super->logDevType = logDevType;
    super->logStartOffset = (adjustedNumBlocks + 1) * params.blockSize;

    if (fs_mkfsMount(context, super, &boot) == NULL)
    {
        fs_ramDealloc(FS_RAM_CORE_MISC, zeroesMalloc);
        fs_ramDealloc(FS_RAM_SUPER, super);
        return FALSE;
    }
    fs_writeSuperblocksSync(context, super);

    fs_queueWriteBlock(&queue, fs_makeBlock(adjustedNumBlocks - 1), blockSize,
                       zeroes);
    fs_writeQueueSync(&queue);

    // place the superblocks
    for (i = 0; i < fs_superBlockCount(); i++) {
        fs_Block block = fs_superBlockBlock(i);
        fs_blockSetAdd(boot.used, block);
    }

    // place the meta inode blocks
    for (i = 0; i < FS_MAX_SNAPSHOTS; i++) {
        fs_Block block = fs_firstInodeBlock(i);
        fs_blockSetAdd(boot.used, block);
    }
    
    /* Initialize the in-core buddies.
     */
    context->fileSystem->buddies = fs_ramAlloc(FS_RAM_CORE_MISC,
                                               (params.buddyRegionCount *
                                                sizeof(fs_buddy_t)));
    for (loop = 0; loop < params.buddyRegionCount; loop++)
    {
        NFX_SPIN_INIT(&(context->fileSystem->buddies[loop].spin),SPINLOCK_LEVEL_UNKNOWN);
        context->fileSystem->buddies[loop].flags = 0;
        context->fileSystem->buddies[loop].summary = NULL;
    } // for

    // place the refbyte blocks
    refbytesExtent = fs_mkfsAllocateNextExtent(context,
                                               &boot,
                                               params.refbytesBlockCount);

    // place the buddy bits blocks
    buddiesExtent = fs_mkfsAllocateNextExtent(context,
                                              &boot,
                                              params.buddyRegionCount);

    // place the directory summary blocks
    dirsumExtent = fs_mkfsAllocateNextExtent(context,
                                             &boot,
                                             params.inodeSummaryCount);

    // place the inode summary blocks
    inosumExtent = fs_mkfsAllocateNextExtent(context,
                                             &boot,
                                             params.inodeSummaryCount);

    // phase 1: erase first inode blocks
    //printf("vol %llx phase 1: erase first inode blocks\n",
    //            context->volId.whole_thing);
    fs_mkfsZeroMetaInodes(context);
    fs_mkfsCommit(context, TRUE);

    //printf("vol %llx init security file\n", context->volId.whole_thing);
    fs_mkfsInitSecurityFile(context, &boot);

    // phase 2: initialize meta inodes
    //printf("vol %llx phase 2: initialize meta inodes\n",
    //            context->volId.whole_thing);
    fs_mkfsCreateMetaInodes(context, &boot, &refbytesExtent, &buddiesExtent,
                            &dirsumExtent, &inosumExtent);

    fs_mkfsCommit(context, TRUE);

    // phase 3: complete the file system metadata
    //printf("vol %llx phase 3: complete the file system metadata\n",
    //        context->volId.whole_thing);
    fs_mkfsAcquireMetaInodes(context, &boot);
    fs_mkfsZeroMetaData(&queue, &boot);

    fs_mkfsCreateRootDirectory(context, &boot);
    fs_mkfsCreateTrashDirectory(context);
    fs_mkfsCreateSystemDirectory(context);
    fs_mkfsCreateCifsSharesDirectory(context);

    fs_mkfsMarkMetaData(context, &boot);
    fs_mkfsInitBuddyRegions(context, &boot);
    fs_mkfsMarkBlockSetLive(context, boot.more);
    fs_mkfsMarkBlockSetNeverSnap(context, boot.neverSnap);
    fs_mkfsAcquireMetaInodes(context, &boot);
    boot.rootInode = I_GET_EXCL(0, FS_ROOT_INODE);
    fs_mkfsInitTimeStamps(context, &boot);

    // phase 4: gewgaws and ornaments
    //printf("vol %llx phase 4: gewgaws and ornaments\n",
    //            context->volId.whole_thing);

    boot.filesetInfoInode = I_GET_EXCL(0, FS_FILESET_INFO_INODE);
    boot.defUserQuotaInode = I_GET_EXCL(0, FS_FILESET_USER_QUOTA_INODE);
    boot.defGroupQuotaInode = I_GET_EXCL(0, FS_FILESET_GROUP_QUOTA_INODE);

    //printf("vol %llx create snapshot dir\n", context->volId.whole_thing);
    fs_mkfsCreateSnapshotDirectory(context, &boot);

    //printf("vol %llx create default fset\n", context->volId.whole_thing);
    fs_mkfsCreateDefaultFileset(context, &boot);

    //printf("vol %llx create lost+found\n", context->volId.whole_thing);
    fs_mkfsCreateLostFoundDirectory(context, &boot);

    //printf("vol %llx final commit\n", context->volId.whole_thing);
    fs_mkfsCommit(context, FALSE);

    FS_TRACE(context, FS_TRACE_REQ, ("commited mkfs data"));

    //fs_refBytePrintAllocated(context);

    // Mark the superblocks clean.
    super->flags &= ~FS_SUPER_FLAG_MKFS;

    // Unmount the volume
    fs_unmountThread(context);

    // Notify ndmp and other listners who want to monitor mkfs on
    // volumes.

#if !defined(FS_STANDALONE)
    // printf("invoking fs_sendMkfsEvent\n");
    fs_sendMkfsEvent(context->volId);
#endif

    fs_ramDealloc(FS_RAM_CORE_MISC, zeroesMalloc);

    return TRUE;
}


/**
 * Update the allocator shadow bitmap for the
 * specified extent (called after growfs has
 * extended the filesystem).
 */
static void
fs_growfsUpdateShadow(
    void        *arg,
    fs_Extent   extent)
{
#ifdef FS_DEBUG_ALLOC
    fs_Context  *context = arg;
    FS_DBG_SHADOW_ALLOC_EXTENT(extent);
#endif
}


/**
 * Grow the filesystem to newNumBlks.
 */
static void
fs_growfs(
    fs_Context  *context,
    uint64      newNumBlks)
{
    fs_Inode		*refInode = I_GET_EXCL(0, FS_REF_COUNT_INODE);
    fs_Inode		*buddyInode = I_GET_EXCL(0, FS_FREE_BUDDY_TREE_INODE);
    fs_SuperBlock 	newSuper, *super = context->fileSystem->super;
    fs_Parameters 	newParams, params = context->fileSystem->params;
    fs_Extent           refbytesExtent, buddiesExtent;
    fs_WriteQueue       queue;
    fs_BootAllocator    boot;
    uint64              firstFreeBlknum;
    uint64 		numNewRefBlocks, numNewBuddyBlocks;
    int 		i;
    char *              zeroesMalloc;
    fs_FileSystem *     fs = context->fileSystem;
    fs_Buf *            lastBuddy = NULL;

    FS_TRACE(context, FS_TRACE_SPACE,
             ("growfs worker currNumBlks %lld newNumBlks %lld",
	     super->numBlks, newNumBlks));
    
    FS_ASSERT(newNumBlks > super->numBlks);

    // Allocate a max block size chunk of zeroes, and ensure
    // its aligned such that the sibyte won't choke on it.
    zeroes = zeroesMalloc = fs_ramAlloc(FS_RAM_CORE_MISC, FS_MAX_BLOCK_SIZE +
                                                    MALLOC_ALIGNED_SIZE);
    zeroes += (MALLOC_ALIGNED_SIZE - 1);
    (address_t) zeroes &= ~(MALLOC_ALIGNED_SIZE - 1);
    bzero(zeroes, FS_MAX_BLOCK_SIZE);

    fs_mkfsInitBoot(context, &boot, fs_makeBlock(super->numBlks - 1),
                    newNumBlks);
    context->fileSystem->bootstrap = &boot;

    // Resize up the superblock and FS params into the new
    // temporaries "newSuper" and "newParams".
    newSuper = *super;
    newSuper.numBlks = newNumBlks;
    newParams = fs_computeSizeParameters(&newSuper);
    //fs_printParameters(&newParams);

    /* Make sure the latest buddy region is in the cache.
     * Need it before adding new blocks to it.
     */
    lastBuddy = fs_buddyGetRegionBuf(context,
                                     fs->params.buddyRegionCount - 1,
                                     TRUE);
    
    /* Grow the in-core buddies.
     */
    if (newParams.buddyRegionCount > fs->params.buddyRegionCount)
    {
        int             loop;
        fs_buddy_t *    newBuddies = fs_ramAlloc(FS_RAM_CORE_MISC,
                                                 (newParams.buddyRegionCount *
                                                  sizeof(fs_buddy_t)));
        
        memcpy(newBuddies, fs->buddies,
               fs->params.buddyRegionCount*sizeof(fs_buddy_t));
        for (loop = fs->params.buddyRegionCount;
             loop < newParams.buddyRegionCount; loop++)
        {
            NFX_SPIN_INIT(&(newBuddies[loop].spin),SPINLOCK_LEVEL_UNKNOWN);
            newBuddies[loop].flags = 0;
            newBuddies[loop].summary = NULL;
        } // for
        NFX_SPIN_LOCK(&fs->spin);
        fs_ramDealloc(FS_RAM_CORE_MISC, fs->buddies);
        fs->buddies = newBuddies;
        NFX_SPIN_UNLOCK(&fs->spin);
    }

    // Initialize the space in the new region without
    // regard for transactional recovery, since the filesystem
    // doesn't yet know of this new space.

    
    numNewRefBlocks = newParams.refbytesBlockCount -
                        params.refbytesBlockCount;
    if (numNewRefBlocks > 0)
    {
        // Place the new refbytes.
        refbytesExtent = fs_mkfsAllocateNextExtent(context,
                                                   &boot,
                                                   numNewRefBlocks);
    }
    numNewBuddyBlocks = newParams.buddyRegionCount -
                          params.buddyRegionCount;
    if (numNewBuddyBlocks > 0)
    {
        // Place the new buddy blocks.
        buddiesExtent = fs_mkfsAllocateNextExtent(context,
                                                  &boot,
                                                  numNewBuddyBlocks);
    }

    fs_initWriteQueue(context, &queue, FS_MKFS_QUEUE_SIZE);

    // Zero the new pages for the refcounts file.
    for (i = 0; i < numNewRefBlocks; i++)
    {
        fs_Block    block = fs_extentBlock(refbytesExtent, i);
        fs_queueWriteBlock(&queue, block, params.blockSize, zeroes);
    }

    // Zero the new pages for the buddy bits file.
    for (i = 0; i < numNewBuddyBlocks; i++)
    {
        fs_Block    block = fs_extentBlock(buddiesExtent, i);
        fs_queueWriteBlock(&queue, block, params.blockSize, zeroes);
    }

    fs_writeQueueSync(&queue);




    // The remainder we'll do transactionally and via the
    // buffer cache, since we'll start  modifying the
    // existing filesystem metadata.



    // Mark as available the new blocks that fall into the
    // existing refbytes last block.  Do this prior to updating
    // the in-core superblock with the new size, otherwise the
    // refbyte code will incorrectly subtract these blocks from
    // super->blksUsed.
    fs_refByteModifyPastEnd(context, fs_refbyteZero, &params);
    
    // Update incore params and incore and ondisk super.
    // Refresh our local copy of the superblock once we
    // have the spinlock held, just in case the used block/inode
    // counter daemons were running and its been updated since
    // we last grabbed it.
    NFX_SPIN_LOCK(&context->fileSystem->superLock);
    super->numBlks = newNumBlks;
    context->fileSystem->params = newParams;
    NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
    fs_writeSuperblocksTrans(context, super);

    // Add the new pages to the files.  Any space required to
    // support the additions will come from the new space,
    // currently managed by the boot allocator.
    if (numNewRefBlocks > 0)
    {
        fs_emap_add(context, refInode, params.refbytesBlockCount,
                    refbytesExtent);
        refInode->dinode.numblocks += fs_extentLength(refbytesExtent);
    }
    if (numNewBuddyBlocks > 0)
    {
        fs_emap_add(context, buddyInode, params.buddyRegionCount,
                    buddiesExtent);
        buddyInode->dinode.numblocks += fs_extentLength(buddiesExtent);
    }

    // Set new inode sizes.
    refInode->dinode.size = newParams.refbytesBlockCount *
                                  newParams.blockSize;
    buddyInode->dinode.size = newParams.buddyRegionCount *
                                  newParams.blockSize;

    // Mark past the new end of the file system in the ref counts.
    fs_refByteModifyPastEnd(context, fs_refbyteSetMeta, &newParams);

    // Mark the refbytes for the new pages as metadata.
    fs_markMetaDataRefs(context, refInode, params.refbytesBlockCount, FALSE);
    fs_markMetaDataRefs(context, buddyInode, params.buddyRegionCount, FALSE);

    // Mark refbytes for blocks in boot.more and boot.neverSnap
    fs_mkfsMarkBlockSetLive(context, boot.more);
    fs_mkfsMarkBlockSetNeverSnap(context, boot.neverSnap);

    // Add remaining free blocks to the buddy file.
    firstFreeBlknum = fs_blockNumber(fs_blockSetNextMissing(boot.used,
                                                            boot.beginBlock));
    for (i = 0; i < newNumBlks - firstFreeBlknum;  )
    {
        fs_Block    block;
        uint64      len;

        block = fs_makeBlock(firstFreeBlknum + i);
        len = MIN(EXTENT_LENGTH_MAX, (newNumBlks - firstFreeBlknum) - i);

        //printf("growfs adding extent %lld,%lld to buddy tree\n",
        //       fs_blockNumber(block), len);

        fs_buddyAddBlocks(context, block, fs_makeBlock(firstFreeBlknum + i + len), TRUE);

        i += len;
    }

    // reinitialize the buddy bitmap
    fs_buddyInitBitmap(context, context->fileSystem,
                       context->fileSystem->params.buddyRegionCount, TRUE);

    // Update the incore allocation shadow bitmaps.
    fs_reallocShadow(context, context->fileSystem);
    fs_blockSetVisitExtents(boot.used, context, fs_growfsUpdateShadow);

    // Revert to real allocator now.
    context->fileSystem->bootstrap = NULL;

    fs_inodeWrite(context, refInode, INO_WRITE_LOG);
    fs_inodeWrite(context, buddyInode, INO_WRITE_LOG);

    //Grow the inode and summary files.
    fs_growInodeFile(context, &params, &newParams);

    fs_ramDealloc(FS_RAM_CORE_MISC, zeroesMalloc);

    fs_bufReleaseExclusive(context, lastBuddy);

} /* fs_growfs */


/**
 * Drive growfs
 */
int32
fs_growfsThread(fs_Context *context)
{
    fs_SuperBlock       *super = context->fileSystem->super;
    uint64               newLogSize, newNumBlks, currNumBlks;
    int32                loop, rc;
    fs_LogDevTypeTable  *devTypeTable = context->fileSystem->logDevTypeTable;
    fs_Time              now = fs_getTimeOfDay();
    boolean              logEmpty;
#define FS_MAX_WAIT_FOR_LOG_EMPTY  20
    
    context->respSize = sizeof (fs_MsgGrowfsResp);
    newNumBlks = ((fs_MsgGrowfsReq *)context->reqHdr)->newNumBlks;
    currNumBlks = super->numBlks;

    fs_mkfsGetLogSize(&newLogSize, &newNumBlks, super->logDevType);

    FS_TRACE(context, FS_TRACE_SPACE,
             ("growfs currNumBlks %lld newNumBlks %lld",
	     currNumBlks, newNumBlks));
    
    if (newNumBlks <= currNumBlks)
    {
        //fs_warn(context, FS_INTERNAL_ERR,
        //  "growfs called with no increase over current blkcnt (%lld/%lld)",
        //        newNumBlks, currNumBlks);
	    return NFX_OK;
    }

#define FS_MIN_GROWFS_BLKCNT    64
    if (newNumBlks - currNumBlks < FS_MIN_GROWFS_BLKCNT)
    {
        // The process of growing the fs may itself require
        // space, so ensure some minimum growth amount.
        fs_warn(context, FS_INTERNAL_ERR,
                "growfs requires minimum of %d blocks new space",
                FS_MIN_GROWFS_BLKCNT);
        return NFX_ERR;
    }

    if (newNumBlks > FS_MAX_BLOCKS_SUPPORTED)
    {
        fs_warn(context, FS_INTERNAL_ERR,
                "growfs exceeds max block count supported (%lld/%lld)",
                newNumBlks, FS_MAX_BLOCKS_SUPPORTED);
        return NFX_ERR;
    }

    FS_TRACE(context, FS_TRACE_SPACE,
             ("growfs log adjusted newNumBlks %lld",
	     newNumBlks));

    fs_freeze(context);

    // Ensure the log is emptied by syncing, then relocate
    // and re-init the log in its new location.
    fs_syncThread(context);
    fs_syncThread(context);
    FS_ASSERT(FS_LOG_IS_EMPTY(context->fileSystem->logRoot));

    // Pause here for a bit to allow log IO completion
    // events to be handled before proceeding to terminate
    // the logging daemon.
    logEmpty = fs_waitLogEmpty(context);
    if (!logEmpty)
    {
        fs_warn(context, FS_INTERNAL_ERR,
                "growfs, sync, flush but still something in Log");
        fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
        return NFX_ERR;
    }

    fs_logStopDaemon(context, FALSE);

    super->logStartOffset = (newNumBlks + 1) *
                             context->fileSystem->params.blockSize;

    context->fileSystem->logSize = (uint32)newLogSize;
    super->logSize = (uint32)newLogSize;
    super->logSpaceSize = (uint32)newLogSize;
    super->logInitTimeSec = now.secs;
    super->logInitTimeNsec = now.nsecs;
    
    // Shift the previous grow info.
    for (loop = (FS_SUPER_NBR_MOUNTTIME - 1); loop > 0; loop--)
    {
        super->growTime[loop] = super->growTime[loop - 1];
    }
    super->nbrGrow++;

    rc = fs_logInitDevice(context);
    if (rc != NFX_OK)
    {
        fs_warn(context, FS_INTERNAL_ERR,
                 "failed to init log device: %d", rc);

        fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
        return NFX_ERR;
    }
    fs_logInitLogSegments(context, context->fileSystem->logRoot, context->fileSystem->logSize);

    // Pause here for a bit to allow log init completion
    logEmpty = 0 ;
    while ( ! devTypeTable->logIsEmpty(context))
    {
        if (logEmpty++ > FS_MAX_WAIT_FOR_LOG_EMPTY)
        {
            fs_warn(context, FS_INTERNAL_ERR,
                    "failed to init new log");

            fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
            return NFX_ERR;
        }
        
        fs_sleep(context, 50);
    }

    // Update the on-disk superblock to reflect
    // the new log location.
    fs_writeSuperblocksSync(context, super);

// Do the grow in steps to avoid very large transactions.
#define FS_GROW_STEP_BLKS   (FS_BLOCKS_PER_BUDDY_REGION * 16)

    while (currNumBlks < newNumBlks)
    {
        uint64  remain;

        currNumBlks = MIN(newNumBlks, currNumBlks + FS_GROW_STEP_BLKS);
        remain = newNumBlks - currNumBlks;
        if (remain < FS_GROW_STEP_BLKS)
        {
            // If the next iteration would be short, lets do it
            // all now, since we don't want to trip over ourselves
            // doing a really short grow on the last iteration.
            currNumBlks = newNumBlks;
        }
        fs_logBeginTransaction(context);
        fs_growfs(context, currNumBlks);
        fs_logCommitEntry(context);
    }

    // Update the on-disk superblock to reflect
    // the new numBlks (inodesUsed and blksUsed)
    fs_writeSuperblocksSync(context, super);

    // Calculate the min free blocks for current
    // conditions.
    fs_recalcMinFreeBlks(context);

    fs_thaw(context);
 
    FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
            ("fs_growfsThread exiting (blks used %lld tot %lld - inodes used %lld tot %lld)",
             context->fileSystem->super->blksUsed, 
             context->fileSystem->super->numBlks,
             context->fileSystem->super->inodesUsed, 
             context->fileSystem->params.maxInodeNumber));
    
    return NFX_OK;
}

#endif // SSC







