/*
 * Name:  fs-rdwr.h
 *
 * RCS:  $Id: fs-rdwr.h,v 1.30 2002/12/10 17:27:07 bob Exp $
 *
 * Description:  Definitions for doing reads/writes on a volume.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#if !defined(FS_RDWR_H)
#define FS_RDWR_H


/**
 * All IO's must be modulo this size.
 */
#define FS_MIN_IO_SIZE  1024


/**
 * Issue no more than this many total
 * IO requests concurrently.
 */
#define FS_RDWR_MAX     8192


extern nfx_spinlock_t   fs_rdwrSpin;
extern fs_RdwrReq  *fs_rdwrStart;
extern fs_RdwrReq  *fs_rdwrEnd;

extern nfx_spinlock_t   fs_ioSpin;
extern int32            fs_ioPending;
extern int32            fs_ioWritesPending;
extern int32            fs_ioMaxWritesPending;
extern int32            fs_ioReadsPending;
extern int32            fs_ioMaxReadsPending;
extern uint32           fs_ioTotalMetaWrites;
extern uint32           fs_ioTotalMetaReads;
extern uint32           fs_ioTotalMetaReadAheads;
extern uint32           fs_ioTotalUserWrites;
extern uint32           fs_ioTotalUserReads;
extern uint32           fs_ioTotalUserReadAheads;
extern uint32           fs_ioTotalLogWrites;
extern uint32           fs_ioTotalLogReads;

/*-----------------------------------------------------------------------------
 * fs_RdwrReq - Callers allocate and pass this as argument to
 *              fs_volReadWrite.  It is returned as event data
 *              upon I/O completion with status and datap (for
 *              reads) updated.
 */
struct fs_RdwrReq
{
    esm_handle_t    smHandle;       // Handle of state machine to send
                                    // completion event to.
    int32           doneEvent;      // Event to deliver upon completion.
    fs_FileSystem   *fileSystem;    // Which filesystem this request is for.
    void            *private;       // Caller's private pointer.
    int32           flags;
#define FS_RDWR_FLAGS_SYNC  0x0001  // Yield CPU and do not return to caller
                                    // until I/O completes.  Only valid for
                                    // sleep-able tasks.
#define FS_RDWR_FLAGS_FREE  0x0002  // rdwr struct is on free list
    evm_IOReq_t     evmReq;         // Volume Mgr's request struct.
    evm_IOStat_t    evmStat;        // Volume Mgr's status struct.
    eeeRTC_t        timestamp_start;  // timestamp at monent send to EVM
    eeeRTC_t        timestamp_return;  // timestamp at io completion
    DLL_ENTRY(fs_RdwrReq)   dllEntry; // For chaining requests together.
    DLL_ENTRY(fs_RdwrReq)   retryEntry; // link to the list of requests to retry
};


/**
 * Initialize fs_RdwrReq stack.
 */
extern void
fs_rdwrInit();


/**
 * Allocate an fs_RdwrReq.
 */
extern fs_RdwrReq *
fs_rdwrAlloc();


/**
 * Deallocate an fs_RdwrReq.
 */
extern void
fs_rdwrDealloc(
    fs_RdwrReq  *rdwr);

/*
 */
extern void
fs_volReadWrite(
    fs_Context      *context,
    fs_RdwrReq      *req);


/**
 * Convenience function for the typical meta data read request case.
 */
extern void fs_volRead(
    fs_Context      *context,
    fs_RdwrReq      *reqp,
    evm_volOffset_t offset,
    evm_volSize_t   size);


/**
 * Convenience function for the typical meta data block read request case.
 */
extern void fs_volReadBlock(
    fs_Context      *context,
    fs_RdwrReq      *reqp,
    fs_Block        block);


/**
 * Convenience function for the typical user data read request case.
 */
extern void fs_volReadUser(
    fs_Context      *context,
    fs_RdwrReq      *reqp,
    evm_volOffset_t offset,
    evm_volSize_t   size);


/**
 * Convenience function for the typical user data block read request case.
 */
extern void fs_volReadUserBlock(
    fs_Context      *context,
    fs_RdwrReq      *reqp,
    fs_Block        block);


/**
 * Convenience function for an async meta data read request.
 */
extern void fs_volReadAsync(
    fs_Context      *context,
    fs_RdwrReq      *reqp,
    evm_volOffset_t offset,
    evm_volSize_t   size,
    int32           evt,
    uint64          attributes,
    void            *priv);


/**
 * Convenience function for an async meta data block read request.
 */
extern void fs_volReadAsyncBlock(
    fs_Context      *context,
    fs_RdwrReq      *reqp,
    fs_Block        block,
    int32           evt,
    void            *priv);


/**
 * Convenience function for the typical meta data write request case.
 */
extern void fs_volWrite(
    fs_Context      *context,
    fs_RdwrReq      *reqp,
    evm_volOffset_t offset,
    evm_volSize_t   size,
    void            *ptr);


/**
 * Convenience function for the typical meta data block write request case.
 */
extern void fs_volWriteBlock(
    fs_Context      *context,
    fs_RdwrReq      *reqp,
    fs_Block        block,
    void            *ptr);


/**
 * Convenience function for the typical user data write request case.
 */
extern void fs_volWriteUser(
    fs_Context      *context,
    fs_RdwrReq      *reqp,
    evm_volOffset_t offset,
    evm_volSize_t   size,
    nfx_cod_t       cod);


/**
 * Convenience function for the typical user data block write request case.
 */
extern void fs_volWriteUserBlock(
    fs_Context      *context,
    fs_RdwrReq      *reqp,
    fs_Block        block,
    nfx_cod_t       cod);


/**
 * Convenience function for an async meta data write request.
 */
extern void fs_volWriteAsync(
    fs_Context      *context,
    fs_RdwrReq      *reqp,
    esm_handle_t    handle,
    evm_volOffset_t offset,
    evm_volSize_t   size,
    void            *ptr,
    int32           evt,
    uint64          attributes,
    void            *priv);


/**
 * Convenience function for an async meta data block write request.
 */
extern void fs_volWriteAsyncBlock(
    fs_Context      *context,
    fs_RdwrReq      *reqp,
    esm_handle_t    handle,
    fs_Block        block,
    void            *ptr,
    int32           evt,
    void            *priv);


/**
 * Convenience function for an async user data write request.
 */
extern void fs_volWriteAsyncUser(
    fs_Context      *context,
    fs_RdwrReq      *reqp,
    esm_handle_t    handle,
    evm_volOffset_t offset,
    evm_volSize_t   size,
    nfx_cod_t       cod,
    int32           evt,
    void            *priv);


/**
 * Convenience function for an async user data block write request.
 */
extern void fs_volWriteAsyncUserBlock(
    fs_Context      *context,
    fs_RdwrReq      *reqp,
    esm_handle_t    handle,
    fs_Block        block,
    nfx_cod_t       cod,
    int32           len,
    int32           evt,
    void            *priv);


/**
 * Clean up rdwr from error return.
 */
extern void
fs_ioDoneError(
    fs_RdwrReq  *rdwr);

/**
 * Mark all the rdwr for this context to point to a NULL filesystem.
 * All pending IO (which are guaranteed by EVM server to arrive) cannot
 * use the filesystem.
 * This will be deallocated from this point on.
 */
void
fs_rdwrMArkFsysDestroyed(fs_FileSystem * fs);

//   W R I T E   Q U E U E S


/**
 * A queue of outstanding write requests.
 */
typedef struct fs_WriteQueue
{
    /** Context for the write requests. */
    fs_Context *context;
    /** Head for dll of unscheduled requests. */
    DLL_HDR(fs_RdwrReq) dllHdr;
    /** Maximum outstanding requests. */
    int max;
    /** Current number of entries queued (unscheduled). */
    int enqueued;
    /** Current number of entries in I/O. */
    int ioPending;

} fs_WriteQueue;


/**
 * Initialize a write queue for a given number of outstanding requests.
 *
 * Once a request from a write queue is inuse, it may not be
 * waited for individually.  The only available operation is to wait
 * for the entire queue to be written.
 */
extern void
fs_initWriteQueue(
    fs_Context      *context,
    fs_WriteQueue   *queue,
    int             size);


/**
 * Wait for all outstanding writes in the specified queue to complete.
 */
extern void
fs_writeQueueSync(
    fs_WriteQueue   *queue);


/**
 * Queue writing meta data.
 */
extern void
fs_queueWrite(
    fs_WriteQueue   *queue,
    evm_volOffset_t offset,
    evm_volSize_t   size,
    void            *data);


/**
 * Queue writing a meta data block.
 */
extern void
fs_queueWriteBlock(
    fs_WriteQueue   *queue,
    fs_Block        block,
    int32           blockSize,
    void            *data);


/**
 * Wait for all pending IO requests to complete.
 */
void
fs_volWaitPendingIO(
    fs_Context  *context,
    boolean     isForce);

/*
 * Read ahead related routines
 */
extern void
fs_volReadAheadUser(
    fs_Context      *context,
    fs_RdwrReq      *req,
    evm_volOffset_t offset,
    evm_volSize_t   size,
    esm_handle_t    raDoneHandle);

extern void
fs_volReadAhead(
    fs_Context      *context,
    fs_RdwrReq      *req,
    evm_volOffset_t offset,
    evm_volSize_t   size,
    esm_handle_t    raDoneHandle);

extern void
fs_volReadAheadUserBlock(
    fs_Context      *context,
    fs_RdwrReq      *req,
    fs_Buf          *buf,
    esm_handle_t    raDoneHandle);
extern void
fs_volReadAheadBlock(
    fs_Context      *context,
    fs_RdwrReq      *req,
    fs_Buf          *buf,
    esm_handle_t    raDoneHandle);

/**
 * List of IO requests to retry.
 */
typedef struct
{
    DLL_HDR(fs_RdwrReq) dllHdr;
    uint32              timerHandle;
    boolean             timerSignal;

} fs_Io2retryQueue;
extern fs_Io2retryQueue fs_io2retryQ[EVM_DATA_TYPE_MAX];


/**
 * Flush the IO's on the retry queue for the
 * specified filesystem.
 */
extern void
fs_rdwrDestroyRetryQueue(
    fs_Context  *context);


#endif /* FS_RDWR_H */
