// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=4 shiftwidth=4 expandtab:

/*
 * Name:  fs-ref.c
 *
 * RCS:  $Id: fs-ref.c,v 1.34 2002/11/27 17:39:26 janv Exp $
 *
 * Description:  Reference bytes.
 *
 * Created By:  Paul Haahr <haahr@agilestorage.com>
 *
 * Date Created:  2001-08-26
 */

#if !defined(SSC)

#include "fs-includes.h"

/**
 * Return a reference byte with the given is-live flag, never-snapshot flag,
 * and reference count.
 */
fs_RefByte fs_makeRefByte(
    boolean isLive,
    boolean neverSnapshot,
    unsigned snapCount)
{
    assert(!neverSnapshot || isLive);
    assert(!neverSnapshot || (snapCount == 0));
    assert(snapCount < FS_REFBYTE_SNAP_MAX);

    return ((isLive ? FS_REFBYTE_IS_LIVE : 0)
            | (neverSnapshot ? FS_REFBYTE_NEVER_SNAPSHOT : 0)
            | (snapCount << FS_REFBYTE_SNAP_COUNT_SHIFT));
}


/**
 * Return true if the specified reference byte indicates a live block.
 */
boolean fs_refbyteIsLive(fs_RefByte refbyte)
{
    return (refbyte & FS_REFBYTE_IS_LIVE) != 0;
}


/**
 * Return true if the specified reference byte indicates a block that
 * should never be in a snapshot.
 */
boolean fs_refbyteNeverSnapshot(fs_RefByte refbyte)
{
    return (refbyte & FS_REFBYTE_NEVER_SNAPSHOT) != 0;
}


/**
 * Return the snapshot count from the specified reference byte.
 */
unsigned fs_refbyteSnapCount(fs_RefByte refbyte)
{
    // TODO - elide the mask because the shift makes it unnecessary?
    return ((refbyte & FS_REFBYTE_SNAP_COUNT_MASK)
            >> FS_REFBYTE_SNAP_COUNT_SHIFT);
}

/**
 * Return a zero refbyte, used by the Modify routine's users.
 */
fs_RefByte fs_refbyteZero(fs_RefByte refbyte)
{
    return 0;
}

/**
 * Return the specified reference byte with the live bit set.
 */
fs_RefByte fs_refbyteSetLive(fs_RefByte refbyte)
{
    assert(!fs_refbyteIsLive(refbyte));
    return (refbyte | FS_REFBYTE_IS_LIVE);
}


/**
 * Return the specified reference byte with the live bit set.  Does not
 * check whether the live bit was previously set.
 */
fs_RefByte fs_refbyteForceLive(fs_RefByte refbyte)
{
    return (refbyte | FS_REFBYTE_IS_LIVE);
}


/**
 * Return a reference byte with the live and never-snapshot flags true
 * and the reference count zero.  The previous value for the byte should
 * be zero.
 */
fs_RefByte fs_refbyteSetMeta(fs_RefByte refbyte)
{
    assert(refbyte == 0);
    return fs_makeRefByte(TRUE, TRUE, 0);
}


/**
 * Return the specified reference byte with the live bit clear.
 */
fs_RefByte fs_refbyteClearLive(fs_RefByte refbyte)
{
    assert(fs_refbyteIsLive(refbyte));
    assert(!fs_refbyteNeverSnapshot(refbyte));
    return (refbyte &~ FS_REFBYTE_IS_LIVE);
}


/**
 * Return the specified reference byte modified so that it can be
 * added back into the allocator's structures.
 *
 * In practice, this is a partial function that always returns 0.
 * However, it asserts that the block had been live and no snapshots
 * are using it.
 */
fs_RefByte fs_refbyteDeallocate(fs_RefByte refbyte)
{
    assert(fs_refbyteIsLive(refbyte));
    assert(fs_refbyteSnapCount(refbyte) == 0);
    return 0;
}


/**
 * Return the specified reference byte with the never-snapshot bit set.
 */
fs_RefByte fs_refbyteDisableSnapshots(fs_RefByte refbyte)
{
    assert(!fs_refbyteNeverSnapshot(refbyte));
    assert(fs_refbyteIsLive(refbyte));
    assert(fs_refbyteSnapCount(refbyte) == 0);
    return (refbyte | FS_REFBYTE_NEVER_SNAPSHOT);
}


/**
 * Return the specified reference byte with the never-snapshot bit clear.
 */
fs_RefByte fs_refbyteEnableSnapshots(fs_RefByte refbyte)
{
    assert(fs_refbyteNeverSnapshot(refbyte));
    return (refbyte &~ FS_REFBYTE_NEVER_SNAPSHOT);
}


/**
 * Return the specified reference byte with the snapshot reference count
 * set to the specified value.
 */
fs_RefByte fs_refbyteSetSnapCount(
    fs_RefByte refbyte,
    unsigned snapCount)
{
    assert(snapCount < FS_REFBYTE_SNAP_MAX);
    assert(snapCount == 0 || !fs_refbyteNeverSnapshot(refbyte));
    return ((refbyte &~ FS_REFBYTE_SNAP_COUNT_MASK)
            | (snapCount << FS_REFBYTE_SNAP_COUNT_SHIFT));
}


/**
 * Return the specified reference byte with the snapshot reference count
 * incremented by one.
 */
fs_RefByte fs_refbyteIncrSnapCount(fs_RefByte refbyte)
{
    assert(fs_refbyteSnapCount(refbyte) + 1 < FS_REFBYTE_SNAP_MAX);
    assert(!fs_refbyteNeverSnapshot(refbyte));
    return (refbyte + (1U << FS_REFBYTE_SNAP_COUNT_SHIFT));
}


/**
 * Return the specified reference byte with the snapshot reference count
 * decremented by one.
 */
fs_RefByte fs_refbyteDecrSnapCount(fs_RefByte refbyte)
{
    assert(fs_refbyteSnapCount(refbyte) > 0);
    return (refbyte - (1U << FS_REFBYTE_SNAP_COUNT_SHIFT));
}


/**
 * Return the reference byte for the specified block.
 */
fs_RefByte fs_refGetByte(fs_Context *context, fs_Block block)
{
    fs_Parameters *params = &context->fileSystem->params;
    uint64 pageno = fs_blockNumber(block) >> params->blockBytesShift;
    uint32 offset = fs_blockNumber(block) &  params->blockBytesMask;

    fs_Inode *refInode = I_GET_SHARED(0, FS_REF_COUNT_INODE);
    fs_Buf *buf = O_GET_SHARED(refInode, pageno);

    fs_RefByte byte = buf->u.refBytes[offset];

    fs_bufRelease(context, buf);
    fs_inodeRelease(context, refInode);

    return byte;
}

/**
 * Increment or decrement the superblock blocks used count. If the
 * blocks used count is being recalculated after log replay, then
 * this routine will return early if the refbytes page being modified
 * falls after the pages recalculated so far.
 */
void
fs_refByteModifyBlksused(
    fs_Context *context,
    uint64 pageno,
    fs_RefByte oldByte,
    fs_RefByte newByte)
{
    /*
     * Increment blocks used when the new byte is nonzero and the
     * old byte is zero -- decrement when the reverse occurs. For
     * any other change, the block allocation state remains unchanged.
     */
    if (newByte != 0 && oldByte == 0) {
        NFX_SPIN_LOCK(&context->fileSystem->superLock);
        context->fileSystem->super->blksUsed++;
        NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
        context->logSummary->blksUsed++;
    } else if (newByte == 0 && oldByte != 0) {
        NFX_SPIN_LOCK(&context->fileSystem->superLock);
        --context->fileSystem->super->blksUsed;
        NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
        --context->logSummary->blksUsed;
    } else {
        /* No change -- e.g., byte added to snapshot. */
    }
}


/**
 * Returns TRUE if specified transform affects bitfields
 * in a refbyte (vs. the counter).
 */
static boolean fs_refByteTransformIsBitField(
    fs_RefByte (*transform)(fs_RefByte))
{
    return (transform == fs_refbyteClearLive ||
            transform == fs_refbyteSetLive ||
            transform == fs_refbyteForceLive ||
            transform == fs_refbyteEnableSnapshots ||
            transform == fs_refbyteDisableSnapshots ||
            transform == fs_refbyteDeallocate ||
            transform == fs_refbyteSetMeta);
}


/**
 *  Obtain the buffer for the specified page of the refbyte
 *  file in the lock mode appropriate for the transform.
 */
static inline fs_Buf *fs_refByteGetBuffer(
    fs_Context  *context,
    fs_RefByte (*transform)(fs_RefByte),
    fs_Inode    *inode,
    uint64      pageno)
{
    return O_GET_EXCL(inode, pageno);
}


/**
 *  Release the specified refbyte buffer in the mode
 *  appropriate for the transform.
 */
static inline void fs_refByteReleaseBuffer(
    fs_Context  *context,
    fs_RefByte (*transform)(fs_RefByte),
    fs_Buf      *buffer)
{
    fs_bufReleaseExclusive(context, buffer);
}


/**
 * Static bitmap array for marking refbytes live.
 */
static uchar8 fs_refbyteLogLiveBits[FS_LOG_BITMAP_MAX_BYTES] = {
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE,
    FS_REFBYTE_IS_LIVE
};


/**
 * Static bitmap array for marking refbytes never snapshot.
 */
uchar8 fs_refbyteLogNeverSnapshotBits[FS_LOG_BITMAP_MAX_BYTES] = {
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_NEVER_SNAPSHOT,
};


/**
 * Static bitmap array for marking refbytes never snapshot.
 */
uchar8 fs_refbyteLogBothBits[FS_LOG_BITMAP_MAX_BYTES] = {
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
    FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT,
};


/**
 * Make log entry for refbyte changes.
 */
void fs_refByteLog(
    fs_Context  *context,
    fs_RefByte (*transform)(fs_RefByte),
    fs_Buf      *buffer,
    int32       offset,
    int32       length)
{
    int32   i;

    // If we have modified only the live or snapshot bit(s), we
    // make a bitmap operation log entry and release the lock on
    // the buf.  Releasing the lock now vs. at commit time provides
    // for greater concurrency, and since we're creating an operation
    // log (vs. recording the new data like other log entries), we
    // provide for the system being able to undo our changes if a
    // writeback of the buf needs to occur before our transaction
    // commits.
    if (fs_refByteTransformIsBitField(transform))
    {
        uint32 bitmapFlag;
        char *bitmap;
        fs_BitmapOperation debugOp;
        uint64 blockNum = (buffer->page * FS_BLOCK_SIZE) + offset;
        
        debugOp.u.extent = fs_makeExtent(length, fs_makeBlock(blockNum));

        if (transform == fs_refbyteClearLive) {
            bitmapFlag = FS_BUF_WRITE_LOG_BIT_RESET;
            bitmap = fs_refbyteLogLiveBits;
            debugOp.op = BITMAP_OP_DEALLOC;
        }
        else if (transform == fs_refbyteEnableSnapshots) {
            bitmapFlag = FS_BUF_WRITE_LOG_BIT_RESET;
            bitmap = fs_refbyteLogNeverSnapshotBits;
        }
        else if (transform == fs_refbyteDisableSnapshots) {
            bitmapFlag = FS_BUF_WRITE_LOG_BIT_SET;
            bitmap = fs_refbyteLogNeverSnapshotBits;
        }
        else if (transform == fs_refbyteDeallocate) {
            bitmapFlag = FS_BUF_WRITE_LOG_BIT_RESET;
            bitmap = fs_refbyteLogBothBits;
            debugOp.op = BITMAP_OP_DEALLOC;
        }
        else if (transform == fs_refbyteSetMeta) {
            bitmapFlag = FS_BUF_WRITE_LOG_BIT_SET;
            bitmap = fs_refbyteLogBothBits;
        }
        else {
            assert(transform == fs_refbyteSetLive
                   || transform == fs_refbyteForceLive);
            bitmapFlag = FS_BUF_WRITE_LOG_BIT_SET;
            bitmap = fs_refbyteLogLiveBits;
            debugOp.op = BITMAP_OP_ALLOC;
        }

        for (i = 0; i < length; i += FS_LOG_BITMAP_MAX_BYTES)
        {
            fs_bufWrite(context,
                        buffer,
                        offset + i,
                        MIN(FS_LOG_BITMAP_MAX_BYTES, length - i),
                        FS_BUF_WRITE_LOG | bitmapFlag,
                        bitmap, &debugOp);
        }

        fs_bufReleaseExclusive(context, buffer);
    }
    else
    {
        fs_bufWrite(context,
                    buffer,
                    offset,
                    length,
                    FS_BUF_WRITE_LOG | FS_BUF_LOG_ADD_BUF,
                    NULL, NULL);
    }
}

    
/**
 * Modify the reference bytes within the specified page of the reference
 * count file, starting at the given byte, for the given count.
 *
 * Returns true if the transformation caused a change to occur, and
 * false otherwise.
 *
 * Assumes a shared lock is held on the reference count inode, which
 * is passed in as an argument.
 */
static boolean fs_refByteModifyPage(
    fs_Context *context,
    fs_RefByte (*transform)(fs_RefByte),
    fs_Inode   *refInode,
    uint64     pageno,
    uint32     start,
    uint32     count)
{
    uint32 offset;
    uint32 end = start + count;
    uint32 firstChange = end;
    uint32 lastChange = start;

    fs_Buf *buffer = fs_refByteGetBuffer(context, transform, refInode, pageno);
    fs_RefByte *bytes = buffer->u.refBytes;
    fs_SuperBlock *super = context->fileSystem->super;

    // printf("fs_refByteModifyPage(%lld, %ld, %ld)\n", pageno, start, count);
#ifdef ISABELLE_STANDALONE
    printf("fs_refByteModifyPage(%5lld, %5d, %5d)\n", pageno, start, count);
#endif // ISABELLE_STANDALONE

    for (offset = start; offset < end; offset++) {
        fs_RefByte oldByte = bytes[offset];
        fs_RefByte newByte = transform(oldByte);
        if (newByte != oldByte) {
	    if (offset < firstChange)
		firstChange = offset;
	    lastChange = offset;
	    if (offset + (pageno*FS_BLOCK_SIZE) < super->numBlks)
	        fs_refByteModifyBlksused(context, pageno, oldByte, newByte);
        }
        bytes[offset] = newByte;
    }

    if (firstChange <= lastChange) {
        fs_refByteLog(context, transform, buffer, firstChange, 
                      lastChange - firstChange + 1);
        return TRUE;
    }
    else {
        fs_refByteReleaseBuffer(context, transform, buffer);
        return FALSE;
    }
}



/**
 * Modify the reference bytes for all blocks in the range specified by the
 * given block count and start position, according to the specified
 * transformation function.
 *
 * Returns true if the transformation caused a change to occur, and
 * false otherwise.
 *
 * Assumes that no locks are held.  (Which might be rather unlikely.)
 */
boolean fs_refByteModify(
    fs_Context *context,
    fs_RefByte (*transform)(fs_RefByte),
    uint32     count,
    fs_Block   block)
{
    boolean modified;
    fs_Parameters *params = &context->fileSystem->params;
    fs_Inode *refInode = I_GET_SHARED(0, FS_REF_COUNT_INODE);

    uint32 blockSize   = params->blockSize;
    uint64 startBlock  = fs_blockNumber(block);
    uint64 endBlock    = startBlock + count;
    uint64 startPage   = startBlock >> params->blockBytesShift;
    uint64 endPage     = endBlock   >> params->blockBytesShift;
    uint64 startByte   = startBlock &  params->blockBytesMask;
    uint64 endByte     = endBlock   &  params->blockBytesMask;

    // printf("fs_refByteModify(%ld, %lld)\n", count, fs_blockNumber(block));

    assert(count > 0);

#ifdef ISABELLE_STANDALONE
    printf("fs_refByteModify(%5lld, %5d)\n", fs_blockNumber(block), count);
#endif // ISABELLE_STANDALONE

    if (endByte == 0) {
        endByte = blockSize;
        endPage--;
    }
    assert(startPage <= endPage);
    assert(0 <= startByte && startByte < blockSize);
    assert(0 < endByte && endByte <= blockSize);

    if (startPage == endPage) {
        assert(endByte > startByte);
        modified = fs_refByteModifyPage(context,
                                        transform,
                                        refInode,
                                        startPage,
                                        startByte,
                                        endByte - startByte);
    }

    else {
        uint64 page;

        modified = fs_refByteModifyPage(context,
                                        transform,
                                        refInode,
                                        startPage,
                                        startByte,
                                        blockSize - startByte);

        for (page = startPage + 1; page < endPage; page++) {
            if (fs_refByteModifyPage(context,
                                     transform,
                                     refInode,
                                     page,
                                     0,
                                     blockSize))
            {
                modified = TRUE;
            }
        }

        if (fs_refByteModifyPage(context,
                                 transform,
                                 refInode,
                                 endPage,
                                 0,
                                 endByte))
        {
            modified = TRUE;
        }
    }

    fs_inodeRelease(context, refInode);
    return modified;
}


/**
 * Modify the reference byte for the given block number, according
 * to the given transformation function.
 *
 * Returns true if the transformation caused a change to occur, and
 * false otherwise.
 *
 * Assumes that no locks are held.  (Which might be rather unlikely.)
 */
boolean fs_refByteModifyBlock(
    fs_Context *context,
    fs_RefByte (*transform)(fs_RefByte),
    fs_Block block)
{
    fs_Parameters *params = &context->fileSystem->params;
    uint64 pageno = fs_blockNumber(block) >> params->blockBytesShift;
    uint32 offset = fs_blockNumber(block) &  params->blockBytesMask;

    fs_Inode *refInode = I_GET_SHARED(0, FS_REF_COUNT_INODE);
    fs_Buf *buffer = fs_refByteGetBuffer(context, transform, refInode, pageno);
    fs_RefByte *bytes = buffer->u.refBytes;

    fs_RefByte oldByte = bytes[offset];
    fs_RefByte newByte = transform(oldByte);

    assert(refInode != NULL);
    assert(fs_blockNumber(block) < params->totalBlocks);

    if (newByte == oldByte)
    {
        fs_refByteReleaseBuffer(context, transform, buffer);
        fs_inodeRelease(context, refInode);
        return FALSE;
    }

    bytes[offset] = newByte;

    if (fs_blockNumber(block) < context->fileSystem->super->numBlks)
        fs_refByteModifyBlksused(context, pageno, oldByte, newByte);
    fs_refByteLog(context, transform, buffer, offset, 1);
    fs_inodeRelease(context, refInode);

    return TRUE;
}


/**
 * Modify the reference byte for all teh blocks in the specified extent,
 * according to the given transformation function.
 *
 * Returns true if the transformation caused a change to occur, and
 * false otherwise.
 *
 * Assumes that no locks are held.  (Which might be rather unlikely.)
 */
boolean fs_refByteModifyExtent(
    fs_Context *context,
    fs_RefByte (*transform)(fs_RefByte),
    fs_Extent  extent)
{
    assert(fs_blockNumber(fs_extentBlock(extent, fs_extentLength(extent) - 1))
           < context->fileSystem->params.totalBlocks);

#ifdef ISABELLE_STANDALONE
    printf("fs_refByteModifyExtent(%5lld, %5d) ok\n",
           fs_blockNumber(fs_extentStartBlock(extent)), fs_extentLength(extent));
#endif // ISABELLE_STANDALONE
    return fs_refByteModify(context,
                            transform,
                            fs_extentLength(extent),
                            fs_extentBlock(extent, 0));
}


/**
 * Apply the given transform to the refbytes past the end of the filesystem.
 * (Used by growfs and mkfs).
 */
void fs_refByteModifyPastEnd(
    fs_Context *context,
    fs_RefByte (*transform)(fs_RefByte),
    fs_Parameters *p)
{
    uint64 pastEnd = (p->refbytesBlockCount * p->blockSize) - p->totalBlocks;

    if (pastEnd > 0)
    {
        FS_TRACE(context, FS_TRACE_SPACE,
                 ("%lld ref bytes past end of file system", pastEnd));

        fs_refByteModify(context, transform, pastEnd,
                         fs_makeBlock(p->totalBlocks));
    }
}


/**
 * Count all allocated blocks, according to the reference bytes.
 *
 * The result is stored in the superblock as a total count of used
 * blocks.
 */
void
fs_refByteCountThread(
    esm_handle_t    handle,
    void            *arg)
{
    fs_Context      *context = arg;
    fs_Parameters   *params = &context->fileSystem->params;
    fs_Inode        *refInode;
    fs_Buf          *buffer;
    uint64          block;
    uint64          pageno = 0;
    uint64          blksUsed;

    // PR #1789: acquire global access lock
    fs_lockShared(context, &context->fileSystem->accessLock, NULL);

    // FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
    //        ("fs_refByteCountThread: started ref byte recalc thread."));

    refInode = I_GET_SHARED(0, FS_REF_COUNT_INODE);
    buffer = O_GET_EXCL(refInode, pageno);

    blksUsed = 0;

    for (block = 0; block < params->totalBlocks && buffer != NULL; )
    {
        fs_RefByte *bytes = buffer->u.refBytes;
        uint64 blocksUsed = 0;
        uint64 max_block;
        uint64 i;
        
        if (block + params->blockSize > params->totalBlocks)
            max_block = params->totalBlocks % params->blockSize;
        else
            max_block = params->blockSize;
        
        for (i = 0; i < max_block; i++)
            blocksUsed += (bytes[i] != 0);

	    //printf("adding %lld blocks page %lld\n", blocksUsed, pageno);

        block += params->blockSize;
        pageno = block >> params->blockBytesShift;

        blksUsed += blocksUsed;

        fs_bufReleaseExclusive(context, buffer);
        buffer = O_GET_EXCL(refInode, pageno);
    }

    if (buffer != NULL)
        fs_bufReleaseExclusive(context, buffer);
    
    fs_inodeRelease(context, refInode);

    // release global access lock
    fs_unlockShared(context, &context->fileSystem->accessLock);

    NFX_SPIN_LOCK(&context->fileSystem->superLock);
    //if (blksUsed != context->fileSystem->super->blksUsed)
    {
        FS_ELOG(context, LOG_INFO, FS_TRACE_REQ,
                ("fs_refByteCountThread recalc %lld (calc %lld) tot %lld - %lld%%",
                 context->fileSystem->super->blksUsed, 
                 blksUsed,
                 context->fileSystem->super->numBlks,
                 100LL * context->fileSystem->super->blksUsed / 
                   context->fileSystem->super->numBlks));
    }
    NFX_SPIN_UNLOCK(&context->fileSystem->superLock);

    fs_threadTerminate(context);
}


/**
 * Return the reference byte for the specified block number.
 */
static fs_RefByte fs_refByteGet(
    fs_Context *context,
    fs_Inode   *refInode,
    uint64     blockno)
{
    fs_Parameters *params = &context->fileSystem->params;
    fs_Buf *buffer
        = O_GET_SHARED(refInode, blockno >> params->blockBytesShift);
    fs_RefByte *bytes = buffer->u.refBytes;
    fs_RefByte result = bytes[blockno & params->blockBytesMask];
    fs_bufRelease(context, buffer);
    return result;
}


/**
 * Print all allocated blocks, according to the reference bytes.
 *
 * Contiguous ranges of bytes with the same reference byte value are
 * printed as one unit.  Each block or set of blocks is followed by
 * the following annotations:
 *
 *   !    block is not in the live file system
 *   *    block is non-snapshotted metadata
 *   (n)  block is in n snapshots
 */

void fs_refBytePrintAllocated(
    fs_Context *context)
{
    fs_Parameters *params = &context->fileSystem->params;
    fs_Inode *refInode = I_GET_SHARED(0, FS_REF_COUNT_INODE);
    uint64 block;

    for (block = 0; block < params->totalBlocks; ) {
        fs_RefByte byte = fs_refByteGet(context, refInode, block);

#if !defined(FS_STANDALONE)
            // yield the CPU to prevent watchdog timeouts
            if (block % 500 == 499)
                fs_sleep(context, 10);
#endif

        if (byte == 0)
            block++;
        else {
            uint64 startBlock = block;
            uint64 endBlock;
            
            while (block < params->totalBlocks
                   && fs_refByteGet(context, refInode, block) == byte)
            {
                block++;
            }
            endBlock = block - 1;

            if (startBlock == endBlock)
                printf("%lld", startBlock);
            else
                printf("%lld-%lld", startBlock, endBlock);

            if (!fs_refbyteIsLive(byte))
                printf("!");
            if (fs_refbyteNeverSnapshot(byte))
                printf("*");
            if (fs_refbyteSnapCount(byte) != 0)
                printf("(%d)", fs_refbyteSnapCount(byte));
            printf("\n");
        }
    }

    fs_inodeRelease(context, refInode);
}

#endif // SSC

