/*
 * Name:  fs-sim.h
 *
 * RCS:  $Id: fs-sim.h,v 1.53 2002/12/10 22:25:39 bobonfeline Exp $
 *
 * Description:  Definitions for providing a UNIX based Eagle simulatator for
 *               filesystem testing.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#if !defined(FS_SIM_H)
#define FS_SIM_H

#if defined(NFX_SPIN_INIT) && defined(NFX_SMP)
#undef NFX_SPIN_INIT
#undef NFX_SPIN_LOCK
#undef NFX_SPIN_TRYLOCK
#undef NFX_SPIN_UNLOCK
#endif

extern pthread_mutexattr_t fs_simMutexAttr;

#define FS_SIM_SPIN_MAGIC 0xd00d1eab

#define NFX_SPIN_INIT(_lockp_,_level_)                                  \
{                                                                       \
    if (pthread_mutex_init(&(_lockp_)->mutex, &fs_simMutexAttr) != 0) { \
        perror("pthread_mutex_init");                                   \
        exit(2);                                                        \
    }                                                                   \
    (_lockp_)->magic = FS_SIM_SPIN_MAGIC;                               \
}

#define NFX_SPIN_LOCK(_lockp_)                                      \
{                                                                   \
    FS_ASSERT((_lockp_)->magic == FS_SIM_SPIN_MAGIC);               \
    if (pthread_mutex_lock(&(_lockp_)->mutex) != 0) {               \
        perror("pthread_mutex_lock");                               \
        exit(2);                                                    \
    }                                                               \
}

#define NFX_SPIN_TRYLOCK(_lockp_)                                   \
{                                                                   \
    int rc;                                                         \
    FS_ASSERT((_lockp_)->magic == FS_SIM_SPIN_MAGIC);               \
    if ((rc = pthread_mutex_trylock(&(_lockp_)->mutex)) == 0)       \
        1;                                                          \
    else if (rc == EBUSY)                                           \
        0;                                                          \
    else {                                                          \
        perror("pthread_mutex_trylock");                            \
        exit(2);                                                    \
    }                                                               \
}

#define NFX_SPIN_UNLOCK(_lockp_)                                    \
{                                                                   \
    FS_ASSERT((_lockp_)->magic == FS_SIM_SPIN_MAGIC);               \
    if (pthread_mutex_unlock(&(_lockp_)->mutex) != 0) {             \
        perror("pthread_mutex_unlock");                             \
        exit(2);                                                    \
    }                                                               \
}

#define FS_SIM_NFS_THREAD_MAX   8

/*-----------------------------------------------------------------
 * Name:    eee_TimeGMT()
 *
 * Description: Returns the current GMT time in epoch since 1 Jan 1970.
 *
 *-----------------------------------------------------------------
 */

#define SEC_PER_MIN    60
#define SEC_PER_HOUR   (SEC_PER_MIN * 60)
#define SEC_PER_DAY    (SEC_PER_HOUR * 24)

/*-----------------------------------------------------------------------------
 * Simulated e-layer messaging stuff.
 */
enum {
    FS_SIM_MSG_ADDR_FP_0    = 0,
    FS_SIM_MSG_ADDR_FP_1    = 1,
    FS_SIM_MSG_ADDR_SSC     = 2,
    FS_SIM_MSG_ADDR_FSSF    = 3,
    FS_SIM_MSG_ADDR_NFS     = 4,
    FS_SIM_MSG_ADDR_MCPU    =  FS_SIM_MSG_ADDR_NFS + FS_SIM_NFS_THREAD_MAX,
};
int32
fs_simSendMsg(
    int32       src,
    int32       dest,
    char        *msg,
    int32       len
);

#ifndef MALLOC_ALIGNED_SIZE
#define MALLOC_ALIGNED_SIZE  8
#endif

nfx_cod_t
fs_sim_get_new_cod(int id);

typedef struct fs_SimMsgQueueEntry fs_SimMsgQueueEntry;
struct fs_SimMsgQueueEntry
{
    void                    *msg;
    DLL_ENTRY(fs_SimMsgQueueEntry)
                            dllEntry;
};

typedef struct fs_SimMsgQueueHead fs_SimMsgQueueHead;
struct fs_SimMsgQueueHead
{
    DLL_HDR(fs_SimMsgQueueEntry)
                            dllHdr;
    pthread_mutex_t         mutex;
    pthread_cond_t          condVar;
};   

/*-----------------------------------------------------------------------------
 * Simulated volume manager stuff.
 */
#define FS_SIM_VOL_MAX      3

/*-----------------------------------------------------------------------------
 * NFS daemon.
 */
// #define FS_SIM_MSG_BUF_LEN      9000
#define FS_SIM_MSG_BUF_LEN      (1024 * 64)

extern struct sockaddr_in fs_simMyNfsIntSockFsAddr;
extern int32 fs_simMyNfsIntSocket;
extern int32 fs_simMyNfsIntSockIndex;

extern int32 fs_simNfsIntPorts[FS_SIM_NFS_THREAD_MAX];
extern int32 fs_simNfsIntSockets[FS_SIM_NFS_THREAD_MAX];

typedef struct fs_SimNfsThread
{
    pthread_t           thread;
    pthread_cond_t      condVar;
    fs_SimMsgQueueHead  rxQueue;
} fs_SimNfsThread;
extern fs_SimNfsThread fs_simNfsThread[FS_SIM_NFS_THREAD_MAX];

#define FS_XLATE_REQ_PROTO(op)      \
    int32                           \
    fs_sim##op##Req2(               \
        void    *req,               \
        void    *arg,               \
        int32   *len,               \
        void    *buf                \
    );
#define FS_XLATE_RESP_PROTO(op)     \
    void *                          \
    fs_sim##op##Resp2(              \
        void    *fsRresp            \
    );
#define FS_XLATE_FREE_RESP_PROTO(op)\
    void                            \
    fs_simFree##op##Resp2(          \
        void    *resp               \
    );

#define FS_XLATE_PROTOS(op)         \
    FS_XLATE_REQ_PROTO(op)          \
    FS_XLATE_RESP_PROTO(op)         \
    FS_XLATE_FREE_RESP_PROTO(op)

FS_XLATE_PROTOS(Null);
FS_XLATE_PROTOS(GetAttr);
FS_XLATE_PROTOS(SetAttr);
FS_XLATE_PROTOS(GetRoot);
FS_XLATE_PROTOS(Lookup);
FS_XLATE_PROTOS(ReadLink);
FS_XLATE_PROTOS(Read);
FS_XLATE_PROTOS(Bad);
FS_XLATE_PROTOS(Write);
FS_XLATE_PROTOS(Create);
FS_XLATE_PROTOS(Remove);
FS_XLATE_PROTOS(Rename);
FS_XLATE_PROTOS(Link);
FS_XLATE_PROTOS(Symlink);
FS_XLATE_PROTOS(Mkdir);
FS_XLATE_PROTOS(Rmdir);
FS_XLATE_PROTOS(Readdir);
FS_XLATE_PROTOS(Statfs);

void
fs_simFreeResp2(
    void    *resp
);

#define FS_XLATE_REQ_PROTO_3(op)    \
    int32                           \
    fs_sim##op##Req3(               \
        void    *req,               \
        void    *arg,               \
        int32   *len,               \
        void    *buf                \
    );
#define FS_XLATE_RESP_PROTO_3(op)   \
    void *                          \
    fs_sim##op##Resp3(              \
        void    *fsRresp            \
    );
#define FS_XLATE_FREE_RESP_PROTO_3(op)\
    void                            \
    fs_simFree##op##Resp3(          \
        void    *resp               \
    );

#define FS_XLATE_PROTOS_3(op)         \
    FS_XLATE_REQ_PROTO_3(op)          \
    FS_XLATE_RESP_PROTO_3(op)         \
    FS_XLATE_FREE_RESP_PROTO_3(op)

FS_XLATE_PROTOS_3(Null);
FS_XLATE_PROTOS_3(GetAttr);
FS_XLATE_PROTOS_3(SetAttr);
FS_XLATE_PROTOS_3(Lookup);
FS_XLATE_PROTOS_3(Access);
FS_XLATE_PROTOS_3(ReadLink);
FS_XLATE_PROTOS_3(Read);
FS_XLATE_PROTOS_3(Write);
FS_XLATE_PROTOS_3(Create);
FS_XLATE_PROTOS_3(Mkdir);
FS_XLATE_PROTOS_3(Symlink);
FS_XLATE_PROTOS_3(Mknod);
FS_XLATE_PROTOS_3(Remove);
FS_XLATE_PROTOS_3(Rmdir);
FS_XLATE_PROTOS_3(Rename);
FS_XLATE_PROTOS_3(Link);
FS_XLATE_PROTOS_3(Readdir);
FS_XLATE_PROTOS_3(ReadDirPlus);
FS_XLATE_PROTOS_3(Fsstat);
FS_XLATE_PROTOS_3(Fsinfo);
FS_XLATE_PROTOS_3(PathConf);
FS_XLATE_PROTOS_3(Commit);

void
fs_simFreeResp3(
    void    *resp
);


/*-----------------------------------------------------------------------------
 * Simulated e-layer state machine stuff.
 */
extern pthread_key_t fs_simKeyEsmHandle;
extern pthread_key_t fs_simKeyCpuId;

#define ESM_Q_MAX   4096

typedef struct esm_instance_t esm_instance_t;
struct esm_instance_t {
#define ESM_NAME_MAX    32
    char                        name[ESM_NAME_MAX];
    esm_table_t                 *table;
    ushort16                    currentState;
    ushort16                    flags;
#define ESM_CAN_YIELD        0x0001  // this type of ESM can call
                                     // esm_yield()
#define ESM_YIELD            0x0002  // this ESM instance is in
                                     // esm_yield()
#define ESM_RUNNING          0x0004  // CAN_YIELD ESM thread has
                                     // been created
#define ESM_IS_BUSY          0x0008  // ESM is busy with an executive
#define ESM_DESTROY          0x0010  // Deferred destroy
    esm_handle_t                handle;
    void                        *userInfo;
    esm_event_t                 eventQueue[ESM_Q_MAX];
    int32                       qload;
    int32                       qunload;
    int32                       qcount;
    DLL_ENTRY(esm_instance_t)   readyQueueEntry;
    esm_event_t                 currentEvent;
    int32                       threadNum;
};

#define FS_SIM_NUM_FP_EXEC_THREADS      2
#define FS_SIM_ESM_MAX                  1024
extern esm_instance_t        *fs_simEsmInstance[FS_SIM_ESM_MAX];
extern nfx_spinlock_t           fs_simEsmInstLock;
extern int32        fs_simEsmNextEntry;
extern ushort16 fs_simEsmGeneration;

#define FS_SIM_ESM_CREATE_HANDLE(_handlep_, _index_)              \
    fs_simEsmGeneration++;                                        \
    if (fs_simEsmGeneration == 0) fs_simEsmGeneration = 1;        \
    (_handlep_) = ((fs_simEsmGeneration) << 16) | (_index_)

#define FS_SIM_ESM_GET_HANDLE_INDEX(_handle_)   \
    ((_handle_) & 0xffff)

extern esm_handle_t fs_simEsmCurrHandle[FS_SIM_NUM_FP_EXEC_THREADS];
extern esm_instance_t *fs_simEsmCurrInst[FS_SIM_NUM_FP_EXEC_THREADS];
extern pthread_key_t  fs_simKeyCpuId;

/*-----------------------------------------------------------------------------
 * Simulated embedded executive
 */
#define panic fs_panic

#define FS_SIM_NUM_FP_CPUS      2
typedef struct fs_SimExec
{
    pthread_t           thread;
    pthread_cond_t      condVar;
    pthread_mutex_t     mutex;  // NOTE: The FP execs share a single mutex
} fs_SimExec;
extern pthread_mutex_t  fs_simFpExecMutex;
extern fs_SimExec fs_simFpExec[FS_SIM_NUM_FP_EXEC_THREADS];
extern fs_SimExec fs_simSscExec;

extern pthread_mutex_t  fs_simFpExecMutex;
extern fs_SimExec fs_simFpExec[FS_SIM_NUM_FP_EXEC_THREADS];
extern fs_SimExec fs_simSscExec;

typedef struct fs_SimThreadArg fs_SimThreadArg;
struct fs_SimThreadArg {
    int32               cpuId;
    esm_handle_t        handle;
    esm_event_t         event;
    esm_instance_t      *smInst;
    esm_actionFunc_t    *func;
    int32               threadNum;
    DLL_ENTRY(fs_SimThreadArg)
                        dllEntry;
};

typedef struct fs_SimReqThread
{
    pthread_t           thread;
    pthread_cond_t      condVar;
    fs_SimThreadArg     threadArg;
#define FS_SIM_THREAD_BUSY   0
#define FS_SIM_THREAD_AVAIL  1
    int32               state;
    jmp_buf             jmpBuf;

} fs_SimReqThread;
extern int32 fs_simNumReqThreadsWaiting;
extern int32 fs_simNumReqThreadsCreated;
#define FS_SIM_EXEC_REQ_THREADS_MAX  32
extern fs_SimReqThread fs_simReqThread[FS_SIM_EXEC_REQ_THREADS_MAX];
extern pthread_key_t  fs_simKeyReqThreadNum;

typedef struct fs_SimThreadJobQueueHead
{
    DLL_HDR(fs_SimThreadArg)
                    dllHdr;
} fs_SimThreadJobQueueHead;
// queue of pending yielding esm (threads) requests
extern fs_SimThreadJobQueueHead fs_simThreadJobQueue;

// queue of received filesystem requests
extern fs_SimMsgQueueHead   fs_simFpRxQueue[FS_SIM_NUM_FP_EXEC_THREADS];

// queue of received ssc responses
extern fs_SimMsgQueueHead   fs_simSscRxQueue;

/*-----------------------------------------------------------------------------
 * Debugging
 */
extern void memdump(
    const char *name,
    const void *memory,
    uint32 size);
#define MEMDUMP(p) memdump(#p, p, sizeof *p)

/*-----------------------------------------------------------------------------
 * Misc
 */
extern fs_MsgReqHdr *fs_shell_parsecmd(char *input,
							   fs_MsgReqHdr *(**resp)(int stat, fs_MsgRespHdr *));

extern uint64 sbm_totalSacSize();

extern void *
fs_simAlloc(int32 numBytes);

extern void
fs_simDealloc(void *ptr);


extern unsigned long atolx (char8 *p);
extern char8 *  strnchr (char8 *str, char8 c, int32 n);
extern char *   strnrchr (char *str, char c, int n);

/*-----------------------------------------------------------------------------
 * Simulated real time clock layer
 */
#define USEC_PER_SEC  1000000
#define USEC_PER_MSEC 1000
#define MSEC_PER_SEC  1000
#define TIMER_TIC_MS			10		/* granularity of tic - 10 ms */
#define SEC_PER_MIN    60
#define SEC_PER_HOUR   (SEC_PER_MIN * 60)
#define SEC_PER_DAY    (SEC_PER_HOUR * 24)

typedef struct {
	int32 sec;
	int32 usec;
} eeeGMT_t;

extern uint64 count_per_usec;

/*-----------------------------------------------------------------
 * Name:    eee_TimeGMT()
 *
 * Description: Returns the current GMT time in epoch since 1 Jan 1970.
 *
 *-----------------------------------------------------------------
 */
static inline void
eee_TimeGMT (eeeGMT_t *t)
{
    struct timeval  now;

    gettimeofday(&now, NULL);
    t->sec = now.tv_sec;
    t->usec = now.tv_usec;
    
    return ;
}

/* input: seconds since  1 Jan 1970
   output "Mon Oct  8 11:21:00 2001" */

char *
eee_humandate(uint32 time);

extern eeeRTC_t eeeLastRTC;
extern eeeRTC_t sscRTC;
extern eeeGMT_t bootGMTTime;
extern eeeGMT_t sscGMTTime;
#if defined(NFX_SMP)
extern nfx_spinlock_t wait_time_lock;
#endif

extern void watchdog_kick(void);

/*-----------------------------------------------------------------
 * Name:    eee_rtcSSCTimeInitted()
 *
 * Description: This routine returns whether we have initialized the
 *              the real time clock services.
 *
 *-----------------------------------------------------------------
 */
static inline int32
eee_rtcSSCTimeInitted(void)
{
    return (sscRTC != 0);
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcSSCTimeInit()
 *
 * Description: This routine initializes the real time clock's sscTime info.
 *				This can be called multiple times if the sscTime
 *				changes.
 *
 *-----------------------------------------------------------------
 */
static inline void
eee_rtcSSCTimeInit(
	eeeGMT_t sscTime)
{
    struct timeval  now;

	NFX_SPIN_LOCK(&wait_time_lock);
	sscGMTTime = sscTime;
    gettimeofday(&now, NULL);
    sscRTC = (eeeRTC_t) now.tv_usec + ((eeeRTC_t) now.tv_sec * USEC_PER_SEC);
    NFX_SPIN_UNLOCK(&wait_time_lock);
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtc()
 *
 * Description: Returns the real time clock since boot up.
 *              This routine cannot be called at interrupt level.
 *
 * Arguments:
 *     clock          -  This is the returned clock value
 *-----------------------------------------------------------------
 */
static inline void
eee_rtc(eeeRTC_t *clock)
{
    struct timeval  now;

    gettimeofday(&now, NULL);
    *clock = (eeeRTC_t) now.tv_usec + ((eeeRTC_t) now.tv_sec * USEC_PER_SEC);
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcZero()
 *
 * Description: Returns a pointer to a zero real time clock value.
 *
 *-----------------------------------------------------------------
 */
static inline eeeRTC_t *
eee_rtcZero(void)
{
	static eeeRTC_t zTime = 0;
    return &zTime;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcStart()
 *
 * Description: This starts an interval timer using the real time clock.
 *
 * Arguments:
 *     interval       -  This is the interval timer.
 *-----------------------------------------------------------------
 */
static inline void
eee_rtcStart(eeeRTC_t *clock)
{
    eee_rtc(clock);
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcReset()
 *
 * Description: This resets an interval timer.
 *
 * Arguments:
 *     interval       -  This is the interval timer.
 *-----------------------------------------------------------------
 */
static inline void
eee_rtcReset(eeeRTC_t *interval)
{
    eee_rtc(interval);
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcAdd()
 *
 * Description: This returns sums the given intervals.
 *
 * Arguments:
 *     interval1      -  A interval to sum.
 *     interval2      -  A interval to sum.
 *     sum            -  The sum
 * Return Value:
 *     NFX_OK         -  Success
 *     NFX_ERR        -  Overflow
 *-----------------------------------------------------------------
 */
static inline int32
eee_rtcAdd(eeeRTC_t *interval1,
		   eeeRTC_t *interval2,
		   eeeRTC_t *sum)
{
    *sum = *interval1 + *interval2;
    if (*sum >= *interval1) {
        return NFX_OK;
    }
    // overflow
    return NFX_ERR;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcIncr()
 *
 * Description: This increments the given interval by the given delta time.
 *              If the operation would result in an overflow, no changes
 *				are made.
 *
 * Arguments:
 *     interval       -  A interval to increment.
 *     delta          -  A delta time to add
 * Return Value:
 *     NFX_OK         -  Success
 *     NFX_ERR        -  Overflow
 *-----------------------------------------------------------------
 */
static inline int32
eee_rtcIncr(eeeRTC_t *interval,
			const eeeRTC_t *delta)
{
    eeeRTC_t sum;

    sum = *interval + *delta;
    if (sum >= *interval) {
		*interval = sum;
		return NFX_OK;
    }
    // overflow
    return NFX_ERR;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcDivMS()
 *
 * Description: This returns the given interval divided by an integer
 *				the results are in msec.
 *
 * Arguments:
 *     interval       -  A interval to divide.
 *     divisor        -  The divisor
 *-----------------------------------------------------------------
 */
static inline float
eee_rtcDivMS(eeeRTC_t *interval,
			 uint64 divisor)
{
    return((float)*interval / ((float)divisor * USEC_PER_MSEC));
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcDiv()
 *
 * Description: This returns the given interval divided by an integer.
 *				No checks for a divide by zero are done.
 *
 * Arguments:
 *     interval       -  A interval to divide, in place.
 *     divisor        -  The divisor
 *-----------------------------------------------------------------
 */
static inline void
eee_rtcDiv(eeeRTC_t *interval,
		   const uint64 divisor)
{
	NFX_ASSERT(divisor != 0);
    *interval /= divisor;
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcSub()
 *
 * Description: This returns the difference in the intervals.
 *
 * Arguments:
 *     interval1      -  A interval.
 *     interval2      -  The interval to subtract.
 *     diff           -  The difference
 * Return Value:
 *     NFX_OK         -  Success
 *     NFX_ERR        -  Underflow
 *-----------------------------------------------------------------
 */
static inline int32
eee_rtcSub(const eeeRTC_t *interval1,
		   const eeeRTC_t *interval2,
		   eeeRTC_t *diff)
{
    *diff = *interval1 - *interval2;
    if (*interval1 >= *interval2) {
        return NFX_OK;
    }
    // underflow
    return NFX_ERR;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcDuration()
 *
 * Description: This returns the interval since invoking eee_rtcStart
 *
 * Arguments:
 *     interval       -  This is the interval timer.
 *     duration       -  This is the returned duration.
 *-----------------------------------------------------------------
 */
static inline void
eee_rtcDuration(const eeeRTC_t *interval,
				eeeRTC_t *duration)
{
    eeeRTC_t now;
    int32 underflow;

    eee_rtc(&now);
    underflow = eee_rtcSub(&now, interval, duration);

    // Catch an underflow
    NFX_ASSERT(underflow == NFX_OK);

    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcDurationTV()
 *
 * Description: This returns the interval since invoking eee_rtcStart
 *              in timeval form.
 *
 * Arguments:
 *     interval       -  This is the interval timer.
 *     duration       -  This is the returned duration.
 *-----------------------------------------------------------------
 */
static inline void
eee_rtcDurationTV(const eeeRTC_t *interval,
				  eeeGMT_t *duration)
{
    eeeRTC_t diff;

    eee_rtcDuration(interval, &diff);
    duration->sec = diff / USEC_PER_SEC;
	duration->usec = (int32)(diff % USEC_PER_SEC);
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcDurationSec()
 *
 * Description: This returns the interval since invoking eee_rtcStart
 *              in seconds form.
 *
 * Arguments:
 *     interval       -  This is the interval timer.
 *-----------------------------------------------------------------
 */
static inline uint64
eee_rtcDurationSec(const eeeRTC_t *interval)
{
    eeeRTC_t diff;

    eee_rtcDuration(interval, &diff);
    return(diff / USEC_PER_SEC);
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcGMT()
 *
 * Description: Returns the current GMT.
 *
 * Arguments:
 *     clock          -  This is the returned clock value
 *-----------------------------------------------------------------
 */
static inline void
eee_rtcGMT(eeeGMT_t *clock)
{
    eeeGMT_t duration;

    // sscGMTTime roughly corresponds to sscRTC.
	NFX_ASSERT(eee_rtcSSCTimeInitted());

	NFX_SPIN_LOCK(&wait_time_lock);
    eee_rtcDurationTV(&sscRTC, &duration);
    clock->sec = sscGMTTime.sec + duration.sec;
    clock->usec = sscGMTTime.usec + duration.usec;
	while (clock->usec >= USEC_PER_SEC) {
		clock->usec -= USEC_PER_SEC;
        clock->sec += 1;
	}
	NFX_SPIN_UNLOCK(&wait_time_lock);
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcToTV()
 *
 * Description: This converts a real time clock interval into
 *              timeval form.
 *
 * Arguments:
 *     interval       -  This is the interval in rtc form.
 *     duration       -  This is the returned duration in timeval form.
 *-----------------------------------------------------------------
 */
static inline void
eee_rtcToTV(const eeeRTC_t *interval,
            eeeGMT_t *duration)
{
	duration->usec = (int32)(*interval % USEC_PER_SEC);

	if (*interval >= USEC_PER_SEC) {
        duration->sec = (int32)(*interval / USEC_PER_SEC);
	}
    else {
        duration->sec = 0;
    }
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcToGMT()
 *
 * Description: Returns the time in GMT for the given real time clock
 *				value.
 *
 * Arguments:
 *     rtcClock       -  The clock value in rtc form
 *     clock          -  This is the returned clock value
 *-----------------------------------------------------------------
 */
static inline void
eee_rtcToGMT(const eeeRTC_t *rtcClock,
			 eeeGMT_t *clock)
{
    eeeGMT_t duration;
	eeeRTC_t delta;

    // sscGMTTime roughly corresponds to sscRTC.
	NFX_ASSERT(eee_rtcSSCTimeInitted());

	NFX_SPIN_LOCK(&wait_time_lock);
    if (eee_rtcSub(rtcClock, &sscRTC, &delta) == NFX_OK) {
		eee_rtcToTV(&delta, &duration);
		clock->sec = sscGMTTime.sec + duration.sec;
		clock->usec = sscGMTTime.usec + duration.usec;
		while (clock->usec >= USEC_PER_SEC) {
			clock->usec -= USEC_PER_SEC;
			clock->sec += 1;
		}
	}
	else {
		// Just return the boot time.
		*clock = sscGMTTime;
	}
	NFX_SPIN_UNLOCK(&wait_time_lock);
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcCompare()
 *
 * Description: This returns compares two intervals.
 *
 * Arguments:
 *     interval1      -  A pointer to an interval.
 *     interval2      -  A pointer to an interval.
 * Return Value:
 *     0              -  They are equal
 *     1              -  The first interval is larger
 *    -1              -  The second interval is larger
 *-----------------------------------------------------------------
 */
static inline int32
eee_rtcCompare(const eeeRTC_t * interval1,
			   const eeeRTC_t * interval2)
{
	if (*interval1 > *interval2) {
        return 1;
    }
    if (*interval1 < *interval2) {
        return -1;
    }
    return 0;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcToUsec()
 *
 * Description: This converts a real time clock interval into
 *              usec form.
 *
 * Arguments:
 *     interval       -  This is the interval in rtc form.
 *-----------------------------------------------------------------
 */
static inline uint64
eee_rtcToUsec(const eeeRTC_t *interval)
{
    return (*interval);
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcToMS()
 *
 * Description: This converts a real time clock interval into
 *              ms form.
 *
 * Arguments:
 *     interval       -  This is the interval in rtc form.
 *-----------------------------------------------------------------
 */
static inline uint64
eee_rtcToMS(const eeeRTC_t *interval)
{
    return (*interval / 1000);
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcToSec()
 *
 * Description: This converts a real time clock interval into
 *              seconds form.
 *
 * Arguments:
 *     interval       -  This is the interval in rtc form.
 *-----------------------------------------------------------------
 */
static inline uint64
eee_rtcToSec(const eeeRTC_t *interval)
{
    return (*interval / 1000000);
}

/*-----------------------------------------------------------------
 * Name:    eee_tvToRTC()
 *
 * Description: This converts a timeval into
 *              real time clock interval form.
 *
 * Arguments:
 *     interval       -  This is the interval in timeval form.
 *     duration       -  This is the returned duration in rtc form.
 *-----------------------------------------------------------------
 */
static inline void
eee_tvToRTC(const eeeGMT_t *interval,
            eeeRTC_t *duration)
{
    *duration = (uint64)interval->usec +
		((uint64)interval->sec * USEC_PER_SEC);
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_usecToRTC()
 *
 * Description: This converts a number of usec into
 *              real time clock interval form.
 *
 * Arguments:
 *     interval       -  This is the interval in usec form.
 *     duration       -  This is the returned duration in rtc form.
 *-----------------------------------------------------------------
 */
static inline void
eee_usecToRTC(const uint64 interval,
			  eeeRTC_t *duration)
{
	*duration = interval; 
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_msToRTC()
 *
 * Description: This converts a number of ms into
 *              real time clock interval form.
 *
 * Arguments:
 *     interval       -  This is the interval in ms form.
 *     duration       -  This is the returned duration in rtc form.
 *-----------------------------------------------------------------
 */
static inline void
eee_msToRTC(const uint32 interval,
			eeeRTC_t *duration)
{
	*duration = interval * USEC_PER_MSEC;
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcStr()
 *
 * Description: Returns the human readable date for the given real
 *				time clock value.
 *
 * Arguments:
 *     clock          -  An optional rtc value.  Current time is
 *						 the default.
 * Return Value:
 *     date string    -  if sscTime is available
 *     "unknown"      -  if sscTime is not available
 *
 *-----------------------------------------------------------------
 */
static inline char *
eee_rtcStr(eeeRTC_t *clock)
{
	if (eee_rtcSSCTimeInitted()) {
		eeeGMT_t now;

		if (clock == NULL) {
			eee_rtcGMT(&now);
		} else {
			eee_rtcToGMT(clock, &now);
		}
		return eee_humandate(now.sec);
	} else {
		return "0";
	}
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcInit()
 *
 * Description: Inits the real time clock.
 *
 *-----------------------------------------------------------------
 */
static inline void
eee_rtcInit(void)
{
    return;
}

#endif /* FS_SIM_H */
