// -*- tab-width: 8; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=8 shiftwidth=4 expandtab:

/*
 * Name:  fs-skip-heap.c
 *
 * RCS:  $Id: fs-skip-heap.c,v 1.21 2002/01/23 23:48:20 haahr Exp $
 *
 * Description:  Skip list implementation of intra-block sorted sets.
 * 
 * Created By:  Paul Haahr <haahr@agilestorage.com>
 *
 * Date Created:  2001-08-05
 */

#if !defined(SSC)

#include "fs-includes.h"


enum {
    MAX_LEVEL = 13 // == log2(8192)
};


/**
 * Compute the level of the specified slot.  Uses a binary distribution.
 */
static int slotLevel(int slot)
{
    int n = log2_32(leastBit(slot)) + 1;
    if (n <= MAX_LEVEL)
        return n;
    else
        return MAX_LEVEL;
}


/**
 * A skip-heap template.
 *
 * The heaps are laid out within blocks with space for the payload objects
 * first, followed by the reserved data structures, followed by the links.
 */
struct fs_SkipHeapTemplate {
    int (*compare)(void *payload1, void *payload2);

    int blockSize;
    int payloadSize;
    int indexSize;
    int reserveSize;
    int capacity;
    int level;

    // All offsets are in index-sized words from the start of the buffer
    int countOffset;
    int freeHeadOffset;
    int freeTailOffset;
    int headerOffset;
    struct {
        uint8 level;
        uint16 offset;
    } link[1];    // variable-sized, one per payload
};


// reserved slots needed in a skip heap
//   freeHead
//   freeTail
//   size
//   slot for one level in the header

enum {
    RESERVED = 4
};


/**
 * Compute the size of the pieces of a skip heap, based on the given block
 * size, payload size, and index size.
 */
static void computeSizes(
    fs_SkipHeapTemplate *template,
    int blockSize,
    int payloadSize,
    int reserveSize,
    int indexSize)
{
    int i;
    int used = 0;
    int available = blockSize - reserveSize;

    template->blockSize   = blockSize;
    template->payloadSize = payloadSize;
    template->indexSize   = indexSize;
    template->reserveSize = reserveSize;
    template->level       = 1;

    used += indexSize * RESERVED;

    for (i = 1;; i++) {
        int level = slotLevel(i);
        int size = payloadSize + level * indexSize;
        if (level > template->level)
            size += (level - template->level) * indexSize;

        if (used + size >= available)
            break;

        used += size;
        if (level > template->level)
            template->level = level;
    }

    template->capacity = i - 1;
}


/**
 * Create a skip-heap template for a specified block size, payload size,
 * and comparison function.
 */
fs_SkipHeapTemplate *fs_makeSkipHeapTemplate(
    int blockSize,
    int payloadSize,
    int reserveSize,
    int (*compare)(void *payload1, void *payload2))
{
    fs_SkipHeapTemplate prototype;
    fs_SkipHeapTemplate *t;
    int i, offset;
    boolean force2Byte = (blockSize / payloadSize) > (1 << BITS_PER_BYTE);

    if (!force2Byte)
        computeSizes(&prototype, blockSize, payloadSize, reserveSize, 1);
    if (force2Byte || prototype.capacity > ((1 << BITS_PER_BYTE) - 1))
        computeSizes(&prototype, blockSize, payloadSize, reserveSize, 2);

    t = fs_ramAlloc(FS_RAM_SKIPHEAP,
                    offsetof(fs_SkipHeapTemplate, link[prototype.capacity]));
    assert(t != NULL);
    memcpy(t, &prototype, sizeof prototype);

    offset = alignto(t->capacity * payloadSize, t->indexSize) / t->indexSize;
    t->countOffset    = offset++;
    t->freeHeadOffset = offset++;
    t->freeTailOffset = offset++;
    t->headerOffset   = offset;
    offset += t->level;

    for (i = 0; i < t->capacity; i++) {
        int level = slotLevel(i + 1);
        t->link[i].level  = level;
        t->link[i].offset = offset;
        offset += level;
    }

#if 0
#define P(p) printf("# %-12s %4d\n", #p ":", t->p)
    P(blockSize);
    P(payloadSize);
    P(indexSize);
    P(reserveSize);
    P(capacity);
    P(level);
    P(countOffset);
    P(freeHeadOffset);
    P(freeTailOffset);
    P(headerOffset);
#undef P
    for (i = 0; i < t->capacity; i++)
        printf("# link[%d] offset %d level %d\n",
               i, t->link[i].offset, t->link[i].level);
#endif
    
    assert(offset * t->indexSize <= blockSize - reserveSize);

    t->compare = compare;

    return t;
}


/**
 * Return the capacity of heaps with the specified template.
 */
int fs_skipHeapTemplateCapacity(
    fs_SkipHeapTemplate *template)
{
    return template->capacity;
}


/**
 * Return the block size of heaps with the specified template.
 */
int fs_skipHeapTemplateBlockSize(
    fs_SkipHeapTemplate *template)
{
    return template->blockSize;
}


/**
 * Return the payload size of heaps with the specified template.
 */
int fs_skipHeapTemplatePayloadSize(
    fs_SkipHeapTemplate *template)
{
    return template->payloadSize;
}


/**
 * Create a skip heap in the specified context, using the specified buffer
 * to contain its data and the specified template for its layout.
 */
fs_SkipHeap fs_makeSkipHeap(
    fs_Context *context,
    fs_SkipHeapTemplate *template,
    fs_Buf *buffer)
{
    struct fs_SkipHeap heap;

    assert(template != NULL);
    assert(buffer != NULL);
    
    heap.context  = context;
    heap.template = template;
    heap.buffer   = buffer;
    return heap;
}


/**
 * Return a pointer to the client-reserved space in the skip heap.
 */
void *fs_skipHeapGetReserve(
    fs_SkipHeap *heap)
{
    fs_SkipHeapTemplate *t = heap->template;
    return &heap->buffer->u.data[t->blockSize - t->reserveSize];
}


/**
 * List terminators for the two possible index sizes.
 */
enum { NIL1 = 0xff, NIL2 = 0xffff };


/**
 * GCC-specific attribute for variables that may go unused without
 * generating a warning.
 */
#ifdef __GNUC__
#define MAYBE_UNUSED __attribute__ ((unused))
#else
#define MAYBE_UNUSED
#endif


/**
 * Execute a body of code with the appropriate local definitions for
 * either 1- or 2-byte indices.  Requires that the variable heap be
 * bound to the skip heap to be operated upon.
 *
 * Inside the body, the following local variables are bound:
 *
 *   context - the operation context the heap was created for
 *   t       - the template describing the heap
 *   data    - the buffer's data, as either 1- or 2-byte unsigned values
 *   NIL     - the list terminator for this type of heap
 *
 * In addition, the SET_DATA macro will work inside this context.
 */
#define SWITCH(body)                                            \
    do {                                                        \
        fs_SkipHeapTemplate *t = heap->template;                \
        fs_Context *context MAYBE_UNUSED = heap->context;       \
        if (t->indexSize == 2) {                                \
            uint16 *data MAYBE_UNUSED = heap->buffer->u.heap2;  \
            const uint16 NIL MAYBE_UNUSED = NIL2;               \
            body;                                               \
        }                                                       \
        else {                                                  \
            uint8 *data MAYBE_UNUSED = heap->buffer->u.heap1;   \
            const uint8 NIL MAYBE_UNUSED = NIL1;                \
            body;                                               \
        }                                                       \
    } while (0)


/**
 * Modify the data in the heap.
 */
#define SET_DATA(_offset_, _value_)                                          \
    do {                                                                     \
        data[_offset_] = (_value_);                                          \
        FS_BUF_WRITE(heap->buffer, (_offset_) * t->indexSize, t->indexSize); \
    } while (0)


/**
 * Initialize a block containing a skip heap.
 */
void fs_skipHeapInit(
    fs_SkipHeap *heap)
{
    // TODO - need to avoid the fs_bufWrite operations here, by making
    // simple log entries for initializing skip heaps.

    SWITCH({
        int i;
        SET_DATA(t->countOffset,    0);
        SET_DATA(t->freeHeadOffset, 0);
        SET_DATA(t->freeTailOffset, t->capacity - 1);
        for (i = 0; i < t->level; i++)
            SET_DATA(t->headerOffset + i, NIL);
        for (i = 0; i < t->capacity - 1; i++)
            SET_DATA(t->link[i].offset, i + 1);
        SET_DATA(t->link[t->capacity - 1].offset, NIL);
    });
}

/**
 * Modify the data in the heap without writing the log (used
 * only at mkfs time).
 */
#define SET_DATA_NO_LOG(_offset_, _value_)                                   \
    do {                                                                     \
        data[_offset_] = (_value_);                                          \
        FS_BUF_WRITE_NO_LOG(heap->buffer, (_offset_) * t->indexSize,         \
                            t->indexSize);                                   \
    } while (0)

/**
 * Initialize a block containing a skip heap.
 */
void fs_skipHeapInitNoLog(
    fs_SkipHeap *heap)
{
    SWITCH({
        int i;
        SET_DATA_NO_LOG(t->countOffset,    0);
        SET_DATA_NO_LOG(t->freeHeadOffset, 0);
        SET_DATA_NO_LOG(t->freeTailOffset, t->capacity - 1);
        for (i = 0; i < t->level; i++)
            SET_DATA_NO_LOG(t->headerOffset + i, NIL);
        for (i = 0; i < t->capacity - 1; i++)
            SET_DATA_NO_LOG(t->link[i].offset, i + 1);
        SET_DATA_NO_LOG(t->link[t->capacity - 1].offset, NIL);
    });
}



//   S I Z I N G


/**
 * Return the number of values currently stored in the given skip heap.
 * The count includes both value and continuation slots.
 */
int fs_skipHeapCount(
    fs_SkipHeap *heap)
{
    SWITCH(return data[t->countOffset]);
}


/**
 * Return the total number of slots in the given skip heap.
 */
int fs_skipHeapCapacity(
    fs_SkipHeap *heap)
{
    return heap->template->capacity;
}


/**
 * Return the number of free slots in the given skip heap.
 */
int fs_skipHeapRemaining(
    fs_SkipHeap *heap)
{
    return fs_skipHeapCapacity(heap) - fs_skipHeapCount(heap);
}


/**
 * Return true if the given skip heap is empty;  that is, it contains no
 * entries yet.
 */
boolean fs_skipHeapIsEmpty(
    fs_SkipHeap *heap)
{
    return (fs_skipHeapCount(heap) == 0);
}


/**
 * Return true if the given skip heap is full;  that is, an attempt to
 * add a new value or acquire a free slot.
 */
boolean fs_skipHeapIsFull(
    fs_SkipHeap *heap)
{
    return (fs_skipHeapRemaining(heap) == 0);
}



//   S L O T S


/**
 * Acquire a slot of the specified skip heap, for storing an added value
 * or for a continuation, and return the index of that slot.
 */
int fs_skipHeapAcquireIndex(
    fs_SkipHeap *heap)
{
    SWITCH({
        int index = data[t->freeHeadOffset];
        if (index == NIL) {
            return -1;
        }
        else {
            int next = data[t->link[index].offset];
            assert(data[t->countOffset] < t->capacity);
            SET_DATA(t->freeHeadOffset, next);
            if (next == NIL) {
                assert(data[t->freeTailOffset] == index);
                SET_DATA(t->freeTailOffset, NIL);
            }
            SET_DATA(t->countOffset, data[t->countOffset] + 1);
            return index;
        }
    });
}


/**
 * Release a slot back to the heap's list of free elements.
 */
void fs_skipHeapReleaseIndex(
    fs_SkipHeap *heap,
    int index)
{
    SWITCH({
        int prev = data[t->freeTailOffset];
        SET_DATA(t->freeTailOffset, index);
        if (prev == NIL) {
            assert(data[t->freeHeadOffset] == NIL);
            SET_DATA(t->freeHeadOffset, index);
        }
        else {
            assert(data[t->link[prev].offset] == NIL);
            SET_DATA(t->link[prev].offset, index);
        }
        SET_DATA(t->countOffset, data[t->countOffset] - 1);
        SET_DATA(t->link[index].offset, NIL);
    });

    fs_skipHeapCheck(heap);
}


/**
 * Translate a pointer to a slot in the given skip heap into an index
 * within the heap.
 */
int fs_skipHeapSlotToIndex(
    fs_SkipHeap *heap,
    void *slot)
{
    fs_SkipHeapTemplate *t = heap->template;
    int offset = ((unsigned char *) slot) - heap->buffer->u.heap1;
    unsigned index = offset / t->payloadSize;
    assert(index < t->capacity);
    assert((offset % t->payloadSize) == 0);
    return index;
}


/**
 * Translate an index within the given skip heap into an pointer to the
 * corresponding slot in the heap.  Returns NULL if the index is not
 * valid index for the heap.
 */
void *fs_skipHeapIndexToSlot(
    fs_SkipHeap *heap,
    int index)
{
    fs_SkipHeapTemplate *t = heap->template;
    if ((unsigned) index >= t->capacity)
        return NULL;
    return &heap->buffer->u.heap1[t->payloadSize * index];
}


/**
 * Acquire a slot of the specified skip heap, for storing an added value
 * or for a continuation, and return the index of that slot.
 */
void *fs_skipHeapAcquireSlot(
    fs_SkipHeap *heap)
{
    int index = fs_skipHeapAcquireIndex(heap);
    if (index < 0)
        return NULL;
    else {
        SWITCH({
            SET_DATA(t->link[index].offset, NIL);
            return fs_skipHeapIndexToSlot(heap, index);
        });
    }
}


/**
 * Release a slot back to the heap's list of free elements.
 */
void fs_skipHeapReleaseSlot(
    fs_SkipHeap *heap,
    void *slot)
{
    fs_skipHeapReleaseIndex(heap, fs_skipHeapSlotToIndex(heap, slot));
}



//   B A S I C   O P E R A T I O N S


/**
 * Search for the specified value in the specified skip heap.  If the
 * value is found, returns a pointer to the slot containing it.
 */
void *fs_skipHeapGet(
    fs_SkipHeap *heap,
    void *value)
{
    SWITCH({
        int i;
        int links = t->headerOffset;
        int limit = NIL;
        for (i = t->level - 1; i >= 0; i--) {
            int next;
            for (;
                 (next = data[links + i]) != limit;
                 links = t->link[next].offset)
            {
                void *slot = fs_skipHeapIndexToSlot(heap, next);
                int compare = t->compare(slot, value);
                if (compare == 0)
                    return slot;
                if (compare > 0) {
                    limit = next;
                    break;
                }
            }
        }
        return NULL;
    });
}


/**
 * Search for a value near the specified value in the specified skip heap.
 * If the skip heap is non-empty, returns a pointer to the slot containing
 * the nearest value; otherwise, returns NULL.
 *
 * This function depends on a comparison function which returns relative
 * distance values.
 */
void *fs_skipHeapFindNearest(
    fs_SkipHeap *heap,
    void *value)
{
    SWITCH({
        int i;
        int links = t->headerOffset;
        int limit = NIL;
        int prev = NIL;

        for (i = t->level - 1; i >= 0; i--) {
            int next;
            for (;
                 (next = data[links + i]) != limit;
                 links = t->link[next].offset)
            {
                void *slot = fs_skipHeapIndexToSlot(heap, next);
                int compare = t->compare(slot, value);
                if (compare == 0)
                    return slot;
                if (compare > 0) {
                    limit = next;
                    break;
                }
                prev = next;
            }
        }
        if (limit == NIL)
            return fs_skipHeapIndexToSlot(heap, prev);
        else if (prev == NIL)
            return fs_skipHeapIndexToSlot(heap, limit);
        else {
            void *prevSlot = fs_skipHeapIndexToSlot(heap, prev);
            void *nextSlot = fs_skipHeapIndexToSlot(heap, limit);
            int prevcmp = t->compare(prevSlot, value);
            int nextcmp = t->compare(nextSlot, value);
            // printf("prevcmp %d -nextcmp %d prev %d next %d\n",
            //       prevcmp, -nextcmp, *(int *)prevSlot, *(int *)nextSlot);
            // negative comparisons used to avoid two's complement overflow
            if (prevcmp > -nextcmp)
                return prevSlot;
            if (prevcmp < -nextcmp)
                return nextSlot;
            // arbitrary from here on in
            return ((prev + limit) & 1) ? prevSlot : nextSlot;
        }
    });
}



/**
 * Create a slot for the specified value in the skip heap, but do nothing
 * to actually insert the value.  This function returns NULL if the value
 * was already present in the heap or if there is no space for the new
 * entry.
 *
 * The caller must properly fill in the returned slot before making
 * any further calls using the skip heap, except that slots for the
 * continuation structures may be acquired before the value is filled
 * in.  It is the caller's responsibility to touch the buffer with
 * fs_bufWrite when making chaassert(data[t->countOffset] <= t->capacity);nges to the returned slot.
 */
void *fs_skipHeapInsert(
    fs_SkipHeap *heap,
    void *value)
{
    SWITCH({
        int i;
        int update[MAX_LEVEL];
        int prev  = NIL;
        int links = t->headerOffset;
        int limit = NIL;
        int index;
        int level;

        for (i = t->level - 1; i >= 0; i--) {
            int next;
            for (;
                 (next = data[links + i]) != limit;
                 prev = next, links = t->link[next].offset)
            {
                void *slot = fs_skipHeapIndexToSlot(heap, next);
                int compare = t->compare(slot, value);
                if (compare == 0)
                    return slot;
                if (compare > 0) {
                    limit = next;
                    break;
                }
            }
            update[i] = prev;
        }

        index = fs_skipHeapAcquireIndex(heap);
        if (index < 0)
            return NULL;
        level = t->link[index].level;
        links = t->link[index].offset;

        for (i = 0; i < level; i++) {
            int prevIndex;
            if (update[i] == NIL)
                prevIndex = t->headerOffset + i;
            else
                prevIndex = t->link[update[i]].offset + i;
            SET_DATA(links + i, data[prevIndex]);
            SET_DATA(prevIndex, index);
        }

        return fs_skipHeapIndexToSlot(heap, index);
    });
}



/**
 * Add the specified value to the specified skip heap.  Returns true if
 * the value was successfully inserted, or false if there was no space
 * for the new value in the skip heap.  If an entry equal to this value
 * already exists in the heap it is replaced.
 */
boolean fs_skipHeapAdd(
    fs_SkipHeap *heap,
    void *value)
{
    //printf("entering skipHeapAdd\n");
    
    void *slot = fs_skipHeapInsert(heap, value);
    fs_Context *context = heap->context;

    //printf("slot = %p\n", slot);

    if (slot == NULL)
        return FALSE;

    memcpy(slot, value, heap->template->payloadSize);
    FS_BUF_WRITE(heap->buffer, ((uint8 *) slot) - heap->buffer->u.heap1,
                 heap->template->payloadSize);
    return TRUE;
}


/**
 * Remove the entry with the specified value from the skip heap.  Returns
 * true if the value was present in the heap, and false otherwise.
 */
boolean fs_skipHeapRemove(
    fs_SkipHeap *heap,
    void *value)
{
    SWITCH({
        int i;
        int update[MAX_LEVEL];
        int prev  = NIL;
        int links = t->headerOffset;
        int limit = NIL;
        int index = NIL;
        int level;

        for (i = t->level - 1; i >= 0; i--) {
            int next;
            for (;
                 (next = data[links + i]) != limit;
                 prev = next, links = t->link[next].offset)
            {
                void *slot = fs_skipHeapIndexToSlot(heap, next);
                int compare = t->compare(slot, value);
                if (compare == 0) {
                    index = next;
                    limit = next;
                    break;
                }
                if (compare > 0) {
                    limit = next;
                    break;
                }
            }
            update[i] = prev;
        }

        if (index == NIL)
            return FALSE;
        level = t->link[index].level;
        links = t->link[index].offset;

        fs_skipHeapCheck(heap);
      
        for (i = 0; i < level; i++) {
            int prevIndex;
            if (update[i] == NIL)
                prevIndex = t->headerOffset + i;
            else
                prevIndex = t->link[update[i]].offset + i;
            SET_DATA(prevIndex, data[links + i]);
        }

        fs_skipHeapReleaseIndex(heap, index);

        return TRUE;
    });
}


//   I T E R A T I O N


/**
 * Return the first element in the specified skip heap, in the ordering
 * defined by the comparison function.  Returns NULL for an empty heap.
 */
void *fs_skipHeapFirst(
    fs_SkipHeap *heap)
{
    SWITCH({
        int index = data[t->headerOffset];
        return fs_skipHeapIndexToSlot(heap, index);
    });
}


/**
 * Return the element following the specified element in the skip heap,
 * in the ordering defined by the comparison function.  Returns NULL
 * at the end of the heap.
 */
void *fs_skipHeapNext(
    fs_SkipHeap *heap,
    void *prev)
{
    SWITCH({
        int prevIndex = fs_skipHeapSlotToIndex(heap, prev);
        int index = data[t->link[prevIndex].offset];
        return fs_skipHeapIndexToSlot(heap, index);
    });
}


/**
 * Set the next link for the specified slot to point at the specified next
 * slot.  Both slots must have been allocated for continuations, though
 * the second slot may be null.  It is an unchecked error to use this
 * function if either slot is part of the base skip-list was not allocated
 * as a continuation.
 *
 * Use fs_skipHeapNext to find the next entry.
 */
void fs_skipHeapSetNext(
    fs_SkipHeap *heap,
    void *slot,
    void *next)
{
    SWITCH({
        int index = fs_skipHeapSlotToIndex(heap, slot);
        int nextIndex;
        if (next == NULL)
            nextIndex = NIL;
        else
            nextIndex = fs_skipHeapSlotToIndex(heap, next);
        SET_DATA(t->link[index].offset, nextIndex);
    });
}


/**
 * Compare two unsigned 32-bit values, returning zero if they are equal, a
 * negative value if the first is less than the second, and a positive
 * value if the first is greater.  The magnitude of the returned result is
 * proportional to the distance between the two, clamped at 2^31-1.
 */
int fs_uint32compare(void *payload1, void *payload2)
{
    const uint32 BIG = 0x7fffffff;
    uint32 i1 = *(int *) payload1;
    uint32 i2 = *(int *) payload2;
    if (i1 < i2) {
        uint32 delta = i2 - i1;
        return -MIN(delta, BIG);
    }
    if (i1 > i2) {
        unsigned delta = i1 - i2;
        return MIN(delta, BIG);
    }
    return 0;
}



//   S A N I T Y   C H E C K S


/**
 * Check that a skip heap is well-formed and obeys all the representation
 * invariants.
 */
void fs_skipHeapCheck(
    fs_SkipHeap *heap)
{
    enum { USED, FREE, UNVISITED } state[heap->template->capacity];

    SWITCH({
        int i;
        int count = 0;
        int first = data[t->headerOffset];
        int firstFree = data[t->freeHeadOffset];
        int lastFree = NIL;
        int index;
        int update[MAX_LEVEL];

        for (i = 0; i < t->capacity; i++)
            state[i] = UNVISITED;

        assert(t->level <= MAX_LEVEL);
        assert(data[t->countOffset] <= t->capacity);

        for (i = 0; i < t->level; i++) {
            update[i] = data[t->headerOffset + i];
        }

        for (index = first; index != NIL; index = data[t->link[index].offset])
        {
            void *slot = fs_skipHeapIndexToSlot(heap, index);
            assert(slot != NULL);
            assert(index < t->capacity);
            assert(state[index] == UNVISITED);
            state[index] = USED;
            for (i = 0; i < t->link[index].level; i++) {
                int next = data[t->link[index].offset + i];
                if (update[i] != next) {
                    assert(update[i] == index);
                    update[i] = next;
                }
            }
            count++;
        }

        // TODO: cope with continuation-allocated slots
        assert(count == data[t->countOffset]);

        for (i = 0; i < t->level; i++) {
            assert(update[i] == NIL);
        }

        for (index = firstFree; index != NIL; index = data[t->link[index].offset])
        {
            assert(index < t->capacity);
            assert(state[index] == UNVISITED);
            state[index] = FREE;
            lastFree = index;
        }
        assert(lastFree == data[t->freeTailOffset]);

        for (i = 0; i < t->capacity; i++) {
            assert(state[i] != UNVISITED);
        }
    });
}
#endif // SSC
