/*
 * Name:  vollog.h
 *
 * RCS:  $Id: fs-vol-log.h,v 1.56 2002/12/10 19:34:28 bob Exp $
 *
 * Description:  Definitions for the volume logging
 *
 * Created By:  Jan Vanhoutte
 *
 * Date Created:  11/10/2001
 */


#if ! defined _FS_VOL_LOG_H_
#define _FS_VOL_LOG_H_

#define FS_LOG2VOLOFFSET(_logOffset_, _volOffset_)                          \
{                                                                           \
    evm_volOffset_t vOffset, bOffset;                                       \
    uint32          nbOffBuf;                                               \
                                                                            \
    nbOffBuf = ((_logOffset_ + sizeof (vl_LogHdr)) /                        \
                 (volBufRoot->bufSize - sizeof (vl_BufDiskHdr)) + 1);       \
    vOffset = (volBufRoot->startOffset + (evm_volOffset_t) _logOffset_ +    \
                 sizeof(vl_LogHdr) + sizeof(vl_BufDiskHdr) * nbOffBuf);     \
    bOffset = (vOffset / volBufRoot->bufSize) * volBufRoot->bufSize;        \
    _volOffset_ = (((evm_volOffset_t) bOffset / volBufRoot->bufSize) *      \
                   volBufRoot->bufSize);                                    \
}


enum {
    VL_BUFFER_SIZE        = 1024,
//    VL_BUFFER_SIZE        = 8192,    // This size must be tuned for maximum performance
                                     // The optimal size to write continous to disk
                                     // The size for writing to the log 

    VL_IO_DONE_ARR_MAX    = 16,     // the max nbr of entries, if VL_BUFFER_SIZE
                                     // (8192 - sizeof(vl_BufDiskHdr)) / 
                                     //   (sizeof(fs_LogRecDone) + sizeof(fs_LogRecHdr))
                                     // (minunum one record (smallest size) + commit)

   VL_IO_DONE_ARR_EXP     = (64 - VL_IO_DONE_ARR_MAX),    // In the uncommon situation that the fixed io done
                                    // array is not sufficient we have space for
                                    // expansion.

    VL_BUFFER_LOG_HWM     = 256,     // High watermark number of buffers in the volume log
                                     // before the write thread blocks
    VL_BUFFER_LOG_MAX     = 1024     // An abs max value for vol log buffers
                                     // The incoretune and fs_incoreEnoughResource defines 
                                     // how nuch must be available.
};

#define VL_MAX_BUF_ID  (1<<30)
#define VL_BUF_ID_WRAP (VL_MAX_BUF_ID/2)

/* 
 * This header indicates that the log volume is formatted and
 * contain 0 or more valid records. Replay is possible.
 */

typedef struct vl_LogHdr
{
    uint32             magic;
#define VL_LOG_MAGIC        0xd11fbead
    uint32             version;
#define VL_LOG_VERSION_ONE  0x00000001
    uint32             timeStamp;       // timestamp set at mkfs time
    uint32             logInitTimeSec;  // time (secs - nsec) log was last initialized
	uint32             logInitTimeNsec;
    uint32             logSize;         // Payload Size in bytes of total vol log
    uint32             bufSize;         // Size in bytes of vol log buffers
    uint32             cksum;
} vl_LogHdr;


typedef struct vl_BufDiskHdr vl_BufDiskHdr;

struct vl_BufDiskHdr
{
    uint32             magic;            // magic, timestamp, and cksum (from vl_LogHdr)
    uint32             timeStamp;        //  are used to verify the validity of the log buf
    uint32             logInitTimeSec;   // time (secs - nsec) log was last initialized
	uint32             logInitTimeNsec;

    uint32             ioWriteTime;  // Time when last write io was done 

    uint32             bufId;        // Increasing id nbr to find the most recent vol buf
    uint32             cksum;        // Protect the whole buf including all records.
                                     // (this avoids to cal/ver the cksum for each record)
    uint32             size;         // The used size in bytes in this vol log buf
                                     // including this header (size protected by cksum)
    uint32             head;         // The head of the vol log.
                                     // (No header in beginning of the log)
    uint32             tail;         // The tail of the vol log.
                                     // (No header in beginning of the log)
    uint64             blksUsed;     // blocks in use
                                     // The valid (commited) value for this Buf
    uint64             inodesUsed;   // inodes in use
                                     // The valid (commited) value for this Buf
};

// logSize = (bufSize - sizeof(bufDiskHdr)) * nbrOfBuf - sizeof (logHdr);
#define  VL_GET_NBR_OF_BUF(_logSize_, _bufSize_)                            \
         ((((_logSize_ - 1) + sizeof(vl_LogHdr)) / (_bufSize_ - sizeof (vl_BufDiskHdr))) + 1)



/*-----------------------------------------------------------------------------
 * Holds a list of reference to trigger when volume logging is finished 
 * for a given log buffer
 * An array and next pointer to facilitate expanding the number 
 */
typedef struct vl_BufIODone
{
    uint32          state;
#define VL_IO_DONE_STATE_FREE            0    // bzero
#define VL_IO_DONE_STATE_READY           1    // FILE_STABLE, and ready for vol log io
#define VL_IO_DONE_STATE_IO              2    // vol log io ongoing
#define VL_IO_DONE_STATE_UNSTABLE_READY  3    // UNSTABLE, and ready for vol log io
#define VL_IO_DONE_STATE_UNSTABLE_IO     4    // vol log io ongoing

    union
    {
        esm_handle_t    smHandle;             // State machine handle.
        uint64          inum;                 // 64-bit inode number
    } u;

    uint32          nbrBufDep;                // number of buffers in dependency list
#define VL_DEF_NBR_DEP_BUF               5
    fs_Buf          *buf[VL_DEF_NBR_DEP_BUF]; // use this during decr if bufList is NULL
    fs_Buf          **bufList;
    
    uint32          nbrInodeDep;              // number of inodes in dependency list
#define VL_DEF_NBR_DEP_INODE             4
    union
    {
        fs_Inode        *inode[VL_DEF_NBR_DEP_INODE];
        fs_Inode        **inodeList;         // use malloc if nbrInodeDep > VL_DEF_NBR_DEP_INODE
    } u_inode;                     
    DLL_HDR(fs_LogUndoSummary)  undoHdr;     // list of undo records to convert
                                             // to cando when the buf reaches
                                             // the disk
} vl_BufIODone;


typedef struct vl_unstableFileHandle
{
    uint32                nbrEntries;   // Number of transaction entries in the log, waiting to be 
                                        // written to disk
    DLL_HDR(fs_Waiter)    waiterHdr;    // sm of the nfs Commit threads,  waiting for 
                                        // log to complete 
                                        // severaL nfs commits can come in simultaneously

    uint64                inum;         // The unique key: 64-bit inode number
} vl_unstableFileHandle;


/*-----------------------------------------------------------------------------
 * The in core data structure for each log buffer
 */

typedef struct vl_BufCoreHdr vl_BufCoreHdr;

struct vl_BufCoreHdr
{
    nfx_spinlock_t     idSpin;       // Lock around the private part of the buff
    uint32             state;
#define  VL_BUF_STATE_IO        1
#define  VL_BUF_STATE_CURR      2  
#define  VL_BUF_STATE_READY     3
#define  VL_BUF_STATE_IO_CURR   4
#define  VL_BUF_STATE_IO_READY  5
    
#define  VL_BUF_IOFINISHED_AND_QUEUED        1
#define  VL_BUF_IOFINISHED_NOT_SET           2
    uint32             ioFinished;   // Boolean to indicate that the IO of this buf 
                                     // done, but one or more of the previous buf not
    uint32             size;         // The used size in bytes in this vol log buf
                                     // including diskHdr header size
                                     // When the buf is in IO state,
                                     // we cannot overwrite the diskHdr itself. 

    uint32             currCommits;  // Always incrementing, to know if commit was 
                                     // was done since last IO
    uint32             currIo;       // Number of actual disk I/Os performed on this buffer.
    uint32             head;         // The head of the vol log.
                                     // The valid value for this Buf
    uint32             tail;         // The tail of the vol log.
                                     // The valid value for this Buf
    uint32             nbrCommits;   // Gives the indication how complete the
                                     // replay was (in case of no EOL or CRC error)
                                     // The valid value for this Buf
    uint64             blksUsed;     // blocks in use
                                     // The valid (commited) value for this Buf
    uint64             inodesUsed;   // inodes in use
                                     // The valid (commited) value for this Buf

    uint32             currNbr;      // index in the log, to calculate the offset

    uint32             ioDoneUsed;   // nbr of ioDone entries in use
    vl_BufIODone       ioDone[VL_IO_DONE_ARR_MAX];
    vl_BufIODone       *ioDoneExp;   // Incase the above array is not sufficient, then
                                     // memory for expansion.

}; 

#define VL_IO_DONE_EXP_SIZE \
                (sizeof(vl_BufIODone) * VL_IO_DONE_ARR_EXP)



typedef struct vl_Disk_Buffer
{
    vl_BufDiskHdr          diskHdr;
    uchar8                 data[VL_BUFFER_SIZE - sizeof(vl_BufDiskHdr)];  
                                     // Data to be written to disk
} vl_Disk_Buffer;


typedef struct vl_VolBuf vl_VolBuf;

struct vl_VolBuf
{
#define VL_VOL_LOG_BUF_STATE_FREE        1
#define VL_VOL_LOG_BUF_STATE_ALLOCATED   3
    uint32                 state;
    
    DLL_ENTRY(vl_VolBuf)   entry;    // Chain entry for the buffers.
    vl_BufCoreHdr          coreHdr;  // Private in core data

    vl_Disk_Buffer         *diskLogBuf;   // header and data (Ultra buf)
};


typedef struct vl_Stats
{
    uint64  nbrWrites ;
    uint64  totalTime ;              // usec
    uint32  min;
    uint32  max;
    // something as standard deviation
} vl_Stats;


/*-----------------------------------------------------------------------------
 * VLC - Volume Log Cache
 */

#define VLC_BLOCK_SIZE      8192
#define VLC_LOGBUF_SIZE     1024

/* Global cache.
 */

#define VLC_MAX_SIZE        (FS_LOG_MAX_NBR_BLOCKS * FS_BLOCK_SIZE)
NFX_ASSERT_STATIC(VLC_MAX_SIZE == (64*1024*1024));
#define VLC_HASH_BUCKETS    8000    // VLC_MAX_SIZE/sizeof(ultrabuf) = 64M/8k
#define VLC_HASH_SHIFT      13

typedef struct fs_vlcMount fs_vlcMount;
struct fs_vlcMount
{
    uint32                  logSize;    // size of the log to replay
    esm_handle_t            smHandle;   // handle of the mount thread
    DLL_ENTRY(fs_vlcMount)  nextInLine;
};

typedef struct fs_vlcCache
{
    nfx_spinlock_t          spin;       // spin lock for this global cache
    uint32                  hashTab;    // the cache is hashed
    int32                   size;       // size of the cache (VLC_MAX_SIZE)
    int32                   used;       // used size
    DLL_HDR(fs_vlcMount)    pendingMountList;
} fs_vlcCache;

/* Cache per volume.
 */

// how many log buffers do we read in one IO (8 * 1024bytes)
#define VL_NB_READ_BUF      8

typedef struct fs_vlcEntry fs_vlcEntry;
struct fs_vlcEntry
{
    evm_volId_t             volId;      // hash key: volume id
    uint16                  page;       // hash key: offset/entrySize
    uint16                  valid;      // TRUE if pointing to received data
    uint16                  reading;    // TRUE when block is being read...
    uint16                  nbBuf;      // how many log buf in the ultrabuf
    char *                  ultrabuf;   // THE data
    evm_volOffset_t         offset;     // offset of the data
    esm_handle_t            guyWaiting; // handle of waiting thread if any
    DLL_ENTRY(fs_vlcEntry)  entry;
};

typedef struct vl_BlockCache
{
    nfx_spinlock_t          spin;       // spin lock for the volume cache
    boolean                 validCache; // TRUE if the cache is valid
    boolean                 replay;     // TRUE if replay is started
    uint16                  entrySize;  // data size in ultrabuf
    DLL_HDR(fs_vlcEntry)    hdr;        // list of all data realted to the vol
#ifdef ISABELLE // debug
    struct
    {
        uint64              nbReadAhead;
        uint64              nbYield;
        uint64              totalYieldTime;
        uint64              nbMetaReadAhead;
        uint64              nbMetaRead;
    } stats;
#endif // ISABELLE
} vl_BlockCache;

/*-----------------------------------------------------------------------------
 * The root of the data structures for volume logging
 */

typedef struct fs_vl_stats
{
    int32              currNbrBuf;    // Current number of buffer
                                      // all are in ready state, except last
    int32              currNbrIO;     // Current number of buffer in IO (VL_BUFFER_IO_MAX)
                                      // (initiate parallel writes to disk)
    int32              maxNbrBuf;
    int32              maxNbrIO;
    uint32             totalNbrIO;

#ifdef FS_DEBUG_VOL_LOG_BUCKETS
#define FS_VL_STATS_NBR_DEP_BUCKETS  6
    int32              nbrDepBuf[FS_VL_STATS_NBR_DEP_BUCKETS];
    int32              nbrDepInode[FS_VL_STATS_NBR_DEP_BUCKETS];
    int32              nbrDepUndo[FS_VL_STATS_NBR_DEP_BUCKETS];
    int32              maxDepBuf;
    int32              maxDepInode;
    int32              maxDepUndo;    
#endif // FS_DEBUG_VOL_LOG_BUCKETS

} fs_vl_stats;


typedef struct vl_VolBufRoot
{
    // this is the data area used while writing into the vol buffers
    // When ready for I/O this info is copied to the real vl_BufHdr
    nfx_spinlock_t     idSpin;       // Lock around nextId.

    uint32             bufSize;      // Size in bytes of vol log buffers
    // See context->fileSystem->logSize for Payload Size in bytes of total vol log

    DLL_HDR(vl_VolBuf) hdr;          // LL of all Volume buffers 
    
    int32              wakeupSignalled;

    boolean            postponedLogWrite;  // indicating that the curr buf is postponed
    boolean            forceWrite;         // indicating that we need a forced log write
                                     // If curr buf is in IO, when done, start
                                     //   a new IO immediate
    
    fs_vl_stats        nbr;

    uint32             magic;        // magic, timestamp, and cksum
    uint32             timeStamp;    //  are used to verify the validity of the log buf
    uint32             logInitTimeSec;  // time (secs - nsec) log was last initialized
	uint32             logInitTimeNsec;

    uint32             currBufId;    // Increasing id nbr to find the most recent vol buf [0..MAXINT]
    uint32             currBufNbr;   // index in the log, to calculate the offset [0..logsize/bufsize]
    uint32             currCommits;  // Always incrementing, to know if commit
                                     // was done since last IO

    uint32             head;         // The head of the vol log.
                                     // (No header in beginning of the log)
    esm_handle_t       ioWrHandle;   // Vol log IO state machine handle.
    fs_Context         *ioWrContext; // Vol log IO state machine context ptr.

    fs_Context         *ioRdContext; // Read context, only needed to read another volid than FS.
    
    evm_volOffset_t    startOffset;

    vl_BlockCache      blockCache;

    char               *tmpBuffer;
    
    nfx_spinlock_t     hidIdSpin;       // Lock around the hash table
    uint32             fileHandleHid ;
    vl_Disk_Buffer     *copyOfCurrentBuf;

} vl_VolBufRoot;




//	struct timeval now;
//	struct timeval start;

// (void)gettimeofday(&now, (struct timezone *)NULL);

// delay = (now.tv_sec - start.tv_sec) * 1000000 +
//         (now.tv_usec - start.tv_usec) ;


/*
 * volume log function prototypes
 */

/*-----------------------------------------------------------------------------
 * fs_logVolWriteDevHdr
 *
 * Description: Called to write the log header to the log device.
 *
 * Created By:  Jan Vanhoutte
 *
 * Date Created:  3/20/2001
 */

extern int32
fs_logVolWriteDevHdr (
    fs_Context      *context,
    int32           operation,
    uint32          count);


/*-----------------------------------------------------------------------------
 * fs_logReadVolLogHdr
 *
 * Description: Called to find and verify the volume log header
 *              See init to write the header
 *
 * Created By:  Jan Vanhoutte
 *
 * Date Created:  3/20/2001
 */

extern int32
fs_logReadVolLogHdr (
    fs_Context      *context,
    fs_LogFsRoot    *logRoot,
    uint32          *logSize,
    uint32          *bufSize);

/*-----------------------------------------------------------------------------
 * fs_logReadVol
 *
 * Description: Called to read something in the volume logger
 *              No path of return, if it fails
 *
 * Created By:  Jan Vanhoutte
 *
 * Date Created:  9/20/2001
 */

extern uchar8 *
fs_logReadVol(
    fs_Context      *context,
    fs_LogFsRoot    *logRoot,
    uint32          offset,
    uint32          *size) ;



/*-----------------------------------------------------------------------------
 * fs_logVolWrite
 *
 * Description: Called to write something in the Volume log file
 *              No path of return, if it fails
 *
 * Created By:  Jan Vanhoutte
 *
 * Date Created:  9/20/2001
 */
extern void
fs_logVolWrite(
    fs_Context          *context,
    uint32              *offset,
    uchar8              *datap[LOG_IOVEC_MAX],
    uint32              lenp[LOG_IOVEC_MAX],
    fs_LogStableWrite   stable,
    uint64              inum);


/*-----------------------------------------------------------------------------
 * fs_logVolInitDevice
 *
 * Description: Called to initialize the volume log device.
 *
 * Created By:  Jan Vanhoutte
 *
 * Date Created:  3/20/2001
 */

extern int32
fs_logVolInitDevice(
    fs_Context      *context) ;


/**
 * the volume log daemon and state machine functions
 *
 */

extern void
vl_startIODaemon(
    fs_Context *context);

extern void 
vl_stopIODaemon(
    fs_Context *context,
    boolean     isForce);

/*----------------------------------------------------------------
 *
 * Vol logging need to lock the read cache while reading the volume
 * During normal operation, the volume is only written, and the cache
 * is invalid.
 * During replay (wyith the lock on), no writing is done. 
 * Only during debug, while reading the log data, the lock will
 * temporary block writing. 
 *
 */

extern void
vl_volLock (
    fs_Context      *context);

extern void
vl_volUnLock (
    fs_Context      *context);


/*----------------------------------------------------------------
 *
 */

EXPORTED void
vl_addUnstableId(
    fs_Context         *context,
    uint64             *inum);


EXPORTED void
vl_removeUnstableId(
    fs_Context         *context,
    uint64             *inum);

EXPORTED void
vl_checkBufIODone (
    fs_Context         *context,
    fs_Inode           *inode,
    uint64             *inum);

EXPORTED void
vl_waitInodeStable(
    fs_Context         *context,
    uint64             inum);

EXPORTED void
vl_addUserBuf (
    fs_Context        *context,
    fs_Inode          *inode,
    fs_Buf            *buf,
    fs_WriteStableHow stableHow);

EXPORTED void
vl_removeUserBuf (
    fs_Context        *context,
    fs_Buf            *buf);


extern int32
fs_logVolCommitTrans(
    fs_Context          *context);

extern int32
fs_logVolCommitWrite(
    fs_Context          *context,
    fs_Inode            *inode,
    fs_WriteStableHow   *stableHow);

extern int32
fs_logVolSyncInodeBuf(
    fs_Context  *context,
    fs_Inode    *inode);

extern int32
fs_logVolSyncInodeLog(
    fs_Context  *context,
    uint64      inum);

extern int32
fs_logVolReadInit(
    fs_Context  *context,
    fs_LogInfo  *info);

extern int32
fs_logVolRead(
    fs_Context  *context,
    uchar8      **datap,
    uint32      offset,
    uint32      *bytesRead,
    uint32      tail);

extern void
fs_logVolReadDone(
    fs_Context  *context,
    uchar8      *datap);

extern void
fs_vl_showStats (
    fs_Context      *context);

extern void
fs_vl_getStats (
    fs_Context      *context,
    boolean         isClear,
    fs_vl_stats     *msg);

extern int32
fs_logVolIsEmpty(
    fs_Context  *context);

extern void
fs_signalVolLogWrite (
    fs_Context      *context);

extern void
fs_logVolInitCache(vl_BlockCache * blockCache);

extern void
fs_logVolReadAhead(fs_Context *     context,
                   vl_VolBufRoot *  volBufRoot,
                   evm_volOffset_t  readOffset,
                   uint32           nbBuf);

extern void
fs_logVolReadCompleteLog(fs_Context *   context,
                         fs_LogInfo *   logInfo);
void
fs_vlcInit(void);
uint32
fs_vlcHash(const void * key);
int32
fs_vlcCompare(const void * key1,
              const void * key2);
void
fs_vlcRelease(void * key,
              void * obj);
fs_vlcEntry *
fs_vlcGetEntry(fs_Context *     context,
               evm_volOffset_t  offset);
fs_vlcEntry *
fs_vlcAddEntry(fs_Context *     context,
               evm_volOffset_t  offset,
               uint32           nbBuf);
void
fs_vlcDelEntry(fs_Context *     context,
               evm_volOffset_t  offset);
extern void
vl_clearBlockCache(fs_Context * context);

#endif // _FS_VOL_LOG_H_
