/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: read.c
 *
 * RCS:  
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: read data from a device.
 *
 *-----------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <nfx-incl.h>
#include <nfx-types.h>
#include <exports/eee-api.h>
#include <exports/scsi-api.h>
#include <exports/anpssc-api.h>
#include <exports/sdm-api.h>
#include <exports/evm-api.h>

#include "fsdb-context.h"






#define BLOCK_SIZE	512



extern int32 
fsdb_get_volid(
    char *id_str, 
    evm_volId_t *volId)
{
    evm_MsgGetId_t      msg;
    evm_MsgStat_t       respMsg;
    int                 rc;
    struct timeval      timeout = { 5, 0};

    msg.hdr.msgType = EVM_VOL_GET_ID;
    strncpy(msg.name, id_str, VOL_MAX_NAME);

    rc = sendAgileMsg(-1,     // system assigned tmp socket
                      &msg,
                      sizeof(evm_MsgGetId_t),
                      "evm",
                      NFX_SSC_UNIX,
                      SSC_SLOT, 
                      MGMTPLANE,
                      &timeout,
                      &respMsg,
                      sizeof(evm_MsgStat_t));
    if (rc < 0) {
        perror("sendAgileMessage");
        fprintf(stderr, "Failed to communicate with EVM\n");
        fprintf(stderr, "Failed to obtain volume ID\n");
        return NFX_ERR;
    }
    if (rc == 0) {
        fprintf(stderr, "Timed out waiting for EVM response\n");
        fprintf(stderr, "Failed to obtain volume ID\n");
        return NFX_ERR;
    }

    if (respMsg.stat != NFX_OK)
    {
        fprintf(stderr, "EVM returned error status (%d)\n", respMsg.stat);
        fprintf(stderr, "Failed to obtain volume ID\n");
        return NFX_ERR;
    }

    volId->whole_thing = respMsg.payload.lVolId.whole_thing;

    //printf("returning volid %lld\n", volId->whole_thing);
    return NFX_OK;
}

typedef struct
{
    char    buf[EVM_MAX_BUF_SIZE];
    uint32  len;
} msg_t;

extern int32
fsdb_get_evm_vol_cfg(
    evm_volId_t          volId,
    fsdb_VolumeInfo      *info)
{
    evm_MsgVolGetAttrib_t   evmReq;
    evm_MsgVolCreate_t      evmRsp;
    evm_MsgGetStrg_t        getReq;
    msg_t                   msg;
    evm_MsgLVStrg_t        *getRsp = (evm_MsgLVStrg_t *) msg.buf;
    evm_VolConfig_t        *cfg;
    int                     rc, loop;
    struct timeval          timeout = { 5, 0};
    uint64                  volSize = 0;

    /* Reset memory for request and response.
     */
    bzero((void *) &evmReq, sizeof(evmReq));
    bzero((void *) &evmRsp, sizeof(evmRsp));

    /* Fill request.
     */
    evmReq.hdr.msgType = EVM_VOL_GET_ATTRIB;
    evmReq.hdr.clientPrivate = 0;   // not used
    memcpy((void *) &(evmReq.id), (const void *) &(volId), sizeof(evm_volId_t));

    /* Send request to EVM.
     */
    rc = sendAgileMsg(-1,
                      &evmReq, 
                      sizeof(evmReq), 
                      "evm", 
                      NFX_SSC_UNIX, 
                      SSC_SLOT, 
                      MGMTPLANE,
                      &timeout,
                      &evmRsp, 
                      sizeof(evmRsp));
    
    if (rc < 0) 
    {
        perror("sendAgileMessage");
        fprintf(stderr, "Failed to communicate with EVM\n");
        fprintf(stderr, "Failed to get vol attributes\n");
        return NFX_ERR;
    }
    if (rc == 0) 
    {
        fprintf(stderr, "Timed out waiting for EVM response\n");
        fprintf(stderr, "Failed to get vol attributes\n");
        return NFX_ERR;
    }
	if ((rc < sizeof(evm_MsgVolCreate_t)) ||
        (evmRsp.hdr.msgType != EVM_VOL_CONFIG)) 
    {
        fprintf(stderr, "Received wrong EVM response\n");
        fprintf(stderr, "Failed to get vol attributes\n");
        return NFX_ERR;
    }

    // printf("vol 0x%llx is %s\n", volId.whole_thing,
    //       (evmRsp.config.onLine ? "online" : "offline"));
   
    if (volId.whole_thing != evmRsp.config.id.whole_thing)
    {
        fprintf(stderr, "Received wrong EVM response\n");
        fprintf(stderr, "Failed to get vol attributes\n");
        return NFX_ERR;
    }

    cfg = &evmRsp.config;

    info->volId.whole_thing = cfg->id.whole_thing;
    strcpy(info->mountName, cfg->name);
    info->currSize = cfg->currSize  * EVM_LBLK_SIZE;
    info->evmState = evmRsp.state;
    info->evmOnLine = cfg->onLine;
    
    printf("\nLogical volume name: %s, id: 0x%llx\n",
           cfg->name, cfg->id.whole_thing);
    printf("  current size (MB): %lld, minautogw (MB): %lld\n",
           cfg->currSize * EVM_LBLK_SIZE / EVM_MEGA_BYTES,
           cfg->minAutoGrow * EVM_LBLK_SIZE / EVM_MEGA_BYTES);
    printf("  hard quota (MB): %lld, soft quota (MB): %d\n",
           cfg->maxSize * EVM_LBLK_SIZE / EVM_MEGA_BYTES,
           cfg->softQuota);
    printf("  nim: %d, online: %s\n",
           cfg->onNIM, cfg->onLine?"Yes":"No");
    if (evmRsp.state != EVM_VOL_READY)
    {
        printf("  *** volume missing LUNs, will not be put online ***\n");
    }
    
    bzero((void *) &getReq, sizeof(evm_MsgGetStrg_t));
    bzero((void *) &msg, sizeof(msg));

    getReq.hdr.msgType = EVM_GET_LV_STRG;
    getReq.hdr.clientPrivate = 0;
    getReq.id = cfg->id;

    msg.len = sendAgileMsg(-1, 
                           &getReq, 
                           sizeof(evm_MsgGetStrg_t),
                           "evm", 
                           NFX_SSC_UNIX, 
                           SSC_SLOT, 
                           MGMTPLANE,
                           &timeout,
                           msg.buf, 
                           sizeof(msg.buf));
    if (rc < 0) 
    {
        perror("sendAgileMessage");
        fprintf(stderr, "Failed to communicate with EVM\n");
        fprintf(stderr, "Failed to get vol attributes\n");
        return NFX_ERR;
    }
    if (rc == 0) 
    {
        fprintf(stderr, "Timed out waiting for EVM response\n");
        fprintf(stderr, "Failed to get vol attributes\n");
        return NFX_ERR;
    }
	if (getRsp->hdr.msgType != EVM_LV_STRG_CFG) 
    {
        fprintf(stderr, "Received wrong EVM response\n");
        fprintf(stderr, "Failed to get vol attributes\n");
        return NFX_ERR;
    }

	// if (getRsp->strg.num >=  FSDB_MAX_NBR_LUN) 
	if (getRsp->nbStrg >=  FSDB_MAX_NBR_LUN) 
    {
        fprintf(stderr, "fsdb has not diminsionned for so many LUN's (%d)\n",
                getRsp->nbStrg);
        fprintf(stderr, "Failed to get vol attributes\n");
        return NFX_ERR;
    }

    info->nbrLuns = getRsp->nbStrg;

    for (loop = 0; loop < info->nbrLuns ; loop++)
    {
#ifdef OLD_EVM
        sdm_devStatusReqMsg_t msg;
        sdm_devStatusRspMsg_t devStatus;

        printf("  - lun #%d\n", loop + 1);
        printf("    scsi id: 0x%llx\n"
               "    size: %lld bytes, startBlk: %d\n",
               getRsp->strg.au[loop].strgId, 
               getRsp->strg.au[loop].auSize * EVM_LBLK_SIZE,
               getRsp->strg.au[loop].startBlk);
        info->lunInfo[loop].scsiId = getRsp->strg.au[loop].strgId;
        info->lunInfo[loop].size = getRsp->strg.au[loop].auSize * EVM_LBLK_SIZE;
        volSize += info->lunInfo[loop].size;
        info->lunInfo[loop].startBlk = getRsp->strg.au[loop].startBlk;
        info->lunInfo[loop].nbr = loop;

        msg.hdr.mid = SDM_MID_DEV_STATUS_REQ;
        msg.id      = getRsp->strg.au[loop].strgId;

        memset(&devStatus, 0, sizeof(sdm_devStatusRspMsg_t));
        rc = sendAgileMsg(-1, 
                          &msg,
                          sizeof(msg),
                          "sdm",
                          NFX_SSC_UNIX, 
                          SSC_SLOT, 
                          MGMTPLANE,
                          &timeout,
                          &devStatus, 
                          sizeof(sdm_devStatusRspMsg_t));

        if (rc < 0) 
        {
            perror("sendAgileMessage");
            fprintf(stderr, "Failed to communicate with SDM\n");
            fprintf(stderr, "Failed to obtain DEV info\n");
            return NFX_ERR;
        }
        if (rc == 0) 
        {
            fprintf(stderr, "Timed out waiting for SDM response\n");
            fprintf(stderr, "Failed to obtain DEV info\n");
            return NFX_ERR;
        }
        if((devStatus.hdr.mid != SDM_MID_DEV_STATUS_RSP) ||
           (devStatus.eCode == SDM_FAIL))
        {
            fprintf(stderr, "Error: Invalid Response Received from SDM: 0x%x\n", devStatus.hdr.mid);
            fprintf(stderr, "Failed to obtain DEV info\n");
            return NFX_ERR;
        }

        info->lunInfo[loop].sdmState = devStatus.paths[0].state;
        info->lunInfo[loop].wwn = devStatus.paths[0].wwn;
        info->lunInfo[loop].lun = devStatus.paths[0].lun;

        info->lunInfo[loop].sdmHandle = fsdb_sdm_ui_getSdmHandle (
            info->lunInfo[loop].wwn,
            info->lunInfo[loop].lun);

        printf("    devName %-39s  model: %-16s\n"
               "    WWN:    0x%016llx  lun: %03d\n"
               "    %s  raidLevel: %04d\n", 
               devStatus.st.name,
               devStatus.st.model,
               devStatus.paths[0].wwn,
               devStatus.paths[0].lun,
               devStatus.paths[0].state == SDM_PATH_STATE_ACTIVE ? "OPENED":"CLOSED",
               devStatus.raidLevel);
#else // OLD_EVM

        printf("  - lun #%d\n", loop + 1);

        printf("    scsi id: 0x%llx\n"
               "    size: %lld bytes, startBlk: %d\n",
               getRsp->strg[loop].au.strgId, 
               getRsp->strg[loop].au.auSize * EVM_LBLK_SIZE,
               getRsp->strg[loop].au.startBlk);

        info->lunInfo[loop].scsiId = getRsp->strg[loop].au.strgId;
        info->lunInfo[loop].sdmHandle = getRsp->strg[loop].au.openHandle;
        info->lunInfo[loop].startBlk = getRsp->strg[loop].au.startBlk;
        info->lunInfo[loop].size = getRsp->strg[loop].au.auSize * EVM_LBLK_SIZE;
        volSize += info->lunInfo[loop].size;
        info->lunInfo[loop].nbr = loop;

        // ?? info->lunInfo[loop].sdmState = 
        info->lunInfo[loop].wwn = getRsp->strg[loop].wwn;
        info->lunInfo[loop].lun = getRsp->strg[loop].lun;

        printf("    name %s\n"
               "    WWN:    0x%016llx  lun: %03d\n",
               // ?? devStatus.devName,
               // ?? devStatus.modelName,
               getRsp->strg[loop].name,
               getRsp->strg[loop].wwn,
               getRsp->strg[loop].lun);
               // ?? devStatus.paths[0].state == SDM_PATH_STATE_ACTIVE ? "OPENED":"CLOSED",
               // ?? devStatus.raidLevel);

#endif

    }

    if (volSize != info->currSize)
    {
        printf("Volume size %lld != sum of LUN size %lld\n", volSize, info->currSize);
        printf("Ignore ... \n");
        if (volSize > info->currSize)
        {
            info->currSize = volSize;
        }
    }

    return NFX_OK;
}




extern int32    // sdmHandle
fsdb_sdm_ui_getSdmHandle( 
    int64 wwn, 
    int32 lun )
{
    int                       rc;
    uint32                    sdmHandle = SDM_INVALID_HANDLE;
    char                	  respbuff[1024];
    sdm_getSdmHandleReqMsg_t  msg;
    sdm_getSdmHandleRspMsg_t *rsp;
    struct timeval            timeout = { 5, 0};

	msg.hdr.mid = SDM_MID_GET_SDMHANDLE_REQ;
    msg.wwn     = wwn;
    msg.lun     = lun;

    rc = sendAgileMsg(-1, 
                      &msg, 
                      sizeof(msg),
                      "sdm", 
                      NFX_SSC_UNIX,
                      SSC_SLOT, 
                      MGMTPLANE,
                      &timeout,
                      respbuff, 
                      sizeof(respbuff));

    if (rc < 0) 
    {
        perror("sendAgileMessage");
        fprintf(stderr, "Failed to communicate with SDM\n");
        fprintf(stderr, "Failed to obtain sdmHanble\n");
        return NFX_ERR;
    }
    if (rc == 0) 
    {
        fprintf(stderr, "Timed out waiting for SDM response\n");
        fprintf(stderr, "Failed to obtain sdmHandle\n");
        return NFX_ERR;
    }

    rsp = (sdm_getSdmHandleRspMsg_t *)respbuff;

    if( rsp->hdr.mid != SDM_MID_GET_SDMHANDLE_RSP ) 
        printf("Error: Invalid Response Received: %x\n", rsp->hdr.mid);
    else 
    if( rsp->rCode == SDM_FAIL ) 
		printf("\nDevice Not Open: 0x%016llx/%d\n\n", wwn, lun);
	else 
		sdmHandle = rsp->sdmHandle;

	return sdmHandle;

} // sdm_ui_getSdmHandle



extern int32   // NFX_OK, with data, else NFX_ERR
fsdb_sdm_disk_read(
    uint32     handle, 
    uint64     offset,        // offset on the LUN
    uint32     byteCount,     // nbr of bytes to read (multiple of 1024 (2 sdm blocks)
    char       *data)         // mem location to copy data (min byte_count size)
{
	int					   blockCount, lba, dlen, rspLen;
	int 				   rc;
	sdm_readDataReqMsg_t   msg;
	sdm_readDataRspMsg_t  *rsp;
    struct timeval         timeout = { 5, 0};

	if (handle == SDM_INVALID_HANDLE)
    {
        printf("Invalid sdm handle 0x%x\n", handle);
        return NFX_ERR;
    }

    if (byteCount % (2 * SDM_BLOCK_SIZE)) 
    {
        printf("Byte count %d must be a multiple of %d\n", blockCount, 2 * SDM_BLOCK_SIZE);
        return NFX_ERR;
    }

    blockCount = byteCount / SDM_BLOCK_SIZE;

    if (offset % (2 * SDM_BLOCK_SIZE)) 
    {
        printf("Offset %lld must be a multiple of %d\n", offset, 2 * SDM_BLOCK_SIZE);
        return NFX_ERR;
    }

    lba = (int)(offset / SDM_BLOCK_SIZE);

	dlen   = blockCount * SDM_BLOCK_SIZE;
	rspLen = sizeof(sdm_readDataRspMsg_t) + dlen;
	rsp    = (sdm_readDataRspMsg_t *)malloc( rspLen ); 

	memset(&msg, 0, sizeof(sdm_readDataReqMsg_t));
	msg.hdr.mid   = SDM_MID_READ_REQ;
	msg.handle    = handle;
	msg.blkOffset = lba;
	msg.blkCnt    = blockCount;
	msg.blkSize   = SDM_BLOCK_SIZE;
	//msg.dataDest  = dataAddr;
	msg.dataDest  = 0;
	
	rc = sendAgileMsg(-1, 
                      (char *)&msg, 
                      sizeof(sdm_readDataReqMsg_t),
                      "sdm", 
                      NFX_SSC_UNIX, 
                      SSC_SLOT, 
                      MGMTPLANE,
                      &timeout,
                      (char*)rsp, 
                      rspLen);

    if (rc < 0) {
        perror("sendAgileMessage");
        fprintf(stderr, "Failed to communicate with SDM\n");
        fprintf(stderr, "Failed to read data\n");
        free(rsp);
        return NFX_ERR;
    }
    if (rc == 0) {
        fprintf(stderr, "Timed out waiting for SDM response\n");
        fprintf(stderr, "Failed to read data\n");
        free(rsp);
        return NFX_ERR;
    }
	if( rsp->hdr.mid != SDM_MID_READ_RSP ) 
    {
		printf("Error: Invalid Response Received from SDM: 0x%x\n", rsp->hdr.mid);
        free(rsp);
		return NFX_ERR;
	}

    if( rsp->eCode != SDM_OK ) 
    {
        printf("DISK READ ERROR: Handle: 0x%x code 0x%x\n", handle, rsp->eCode);
        free(rsp);
		return NFX_ERR;
    }

    if( rsp->dataLen != byteCount) 
    {
        printf("DISK READ ERROR: request to read %d bytes, received %d bytes\n", byteCount, rsp->dataLen);
        free(rsp);
		return NFX_ERR;
    }

    // printf("\nDISK READ DONE: Bytes: %d, COD: 0x%llx, LBA: %d\n\n", rsp->dataLen, rsp->codHandle, lba);
    memcpy (data, (char *)rsp->data, rsp->dataLen);
	free(rsp);

    return NFX_OK;

} /* sbm_disk_read */


extern int32   // NFX_OK, with data, else NFX_ERR
fsdb_sdm_disk_write(
    uint32     handle, 
    uint64     offset,        // offset on the LUN
    uint32     byteCount,     // nbr of bytes to read (multiple of 1024 (2 sdm blocks)
    char       *data)         // mem location to copy data (min byte_count size)
{
	int					   blockCount, lba, dlen, reqLen;
	int 				   rc;
	sdm_writeDataReqMsg_t *req;
	sdm_writeDataRspMsg_t  rsp;
    struct timeval         timeout = { 5, 0};

	if (handle == SDM_INVALID_HANDLE)
    {
        printf("Invalid sdm handle 0x%x\n", handle);
        return NFX_ERR;
    }

    if (byteCount % (2 * SDM_BLOCK_SIZE)) 
    {
        printf("Byte count %d must be a multiple of %d\n", blockCount, 2 * SDM_BLOCK_SIZE);
        return NFX_ERR;
    }

    blockCount = byteCount / SDM_BLOCK_SIZE;

    if (offset % (2 * SDM_BLOCK_SIZE)) 
    {
        printf("Offset %lld must be a multiple of %d\n", offset, 2 * SDM_BLOCK_SIZE);
        return NFX_ERR;
    }

    lba = (int)(offset / SDM_BLOCK_SIZE);

	dlen   = blockCount * SDM_BLOCK_SIZE;
	reqLen = sizeof(sdm_writeDataReqMsg_t) + byteCount;
	req    = (sdm_writeDataReqMsg_t *)malloc(reqLen); 

	memset(req, 0, sizeof(sdm_writeDataReqMsg_t));
	req->hdr.mid   = SDM_MID_WRITE_REQ;
	req->handle    = handle;
	req->blkOffset = lba;
	req->blkCnt    = blockCount;
	req->blkSize   = SDM_BLOCK_SIZE;
    req->attributes = SCSI_MSG_ATTR_IGNORE_WRITE_GATE;
	//req->dataDest  = dataAddr;
    memcpy (req->data, data, byteCount);
	
	rc = sendAgileMsg(-1, 
                      (char *)req,
                      reqLen,
                      "sdm", 
                      NFX_SSC_UNIX, 
                      SSC_SLOT, 
                      MGMTPLANE,
                      &timeout,
                      (char*)&rsp, 
                      sizeof(sdm_writeDataRspMsg_t));

	free(req);

    if (rc < 0) {
        perror("sendAgileMessage");
        fprintf(stderr, "Failed to communicate with SDM\n");
        fprintf(stderr, "Failed to write data\n");
        return NFX_ERR;
    }
    if (rc == 0) {
        fprintf(stderr, "Timed out waiting for SDM response\n");
        fprintf(stderr, "Failed to write data\n");
        return NFX_ERR;
    }
	if( rsp.hdr.mid != SDM_MID_WRITE_RSP ) 
    {
		printf("Error: Invalid write Response Received from SDM: 0x%x\n", rsp.hdr.mid);
		return NFX_ERR;
	}

    if( rsp.eCode != SDM_OK ) 
    {
        printf("DISK WRITE ERROR: Handle: 0x%x code 0x%x\n", handle, rsp.eCode);
		return NFX_ERR;
    }

    return NFX_OK;

} /* sbm_disk_write */



char evm_raid_level_strings[7][32] =
{
    "raid 0",
    "raid 1",
    "raid 2",
    "raid 3",
    "invalid raid level",
    "raid 5",
    "none"
};

extern int32
fsdb_evm_read_label(
    fsdb_VolumeInfo      *info,
    int32                lunNbr,
    fsdb_LunLabelAction  action)
{
    uint64          offset;
    evm_strgCfg_t  *odCfg;
    evm_mgrArea_t  *mgr;
    evm_tkoArea_t  *tko;
    uint32          cksum;
    int             i;

    offset = info->lunInfo[lunNbr].size;

    if (fsdb_sdm_disk_read(info->lunInfo[lunNbr].sdmHandle, 
                           offset,
                           FS_BLOCK_SIZE,   // EVM_LUN_AREA,
                           info->data) != NFX_OK)
    {
        printf ("No valid data \n\n");
        return NFX_ERR;
    }

    odCfg = (evm_strgCfg_t *)info->data;
    mgr = (evm_mgrArea_t *)&(info->data[4*SDM_BLOCK_SIZE]);
    tko = (evm_tkoArea_t *)&(info->data[6*SDM_BLOCK_SIZE]);

    if (action == FSDB_LUNLABEL_WRITE_MANAGEMENT)
    {
        //strcpy(mgr->mgrId.cluster, "fsdb");
        mgr->mgrId.cluster[233]++;
        fsdb_sdm_disk_write(info->lunInfo[lunNbr].sdmHandle, 
                            offset,
                            FS_BLOCK_SIZE,   // EVM_LUN_AREA,
                            info->data);
    }
    else
    if (action == FSDB_LUNLABEL_WRITE_TAKEOVER)
    {
        struct timeval now;

        gettimeofday(&now, NULL);
        // strcpy(tko->tkoId.cluster, "fsdb");
        // tko->timestamp = now.tv_sec;
        // mgr->mgrId.cluster[233]++;
        tko->timestamp++;

        fsdb_sdm_disk_write(info->lunInfo[lunNbr].sdmHandle, 
                            offset,
                            FS_BLOCK_SIZE,   // EVM_LUN_AREA,
                            info->data);
    }

    printf("*** LUN label ***\n");
    printf("magic:              0x%x\n", odCfg->magic);
    printf("version:            %d (0x%x)\n", odCfg->version, odCfg->version);
    if ((cksum = uint32_checksum (odCfg, sizeof(evm_strgCfg_t) - sizeof(uint32), 0)) == odCfg->cksum)
        printf("cksum OK            0x%x\n", odCfg->cksum);
    else
        printf("WRONG checksum      0x%x (must be 0x%x)\n", 
               odCfg->cksum, cksum);

    printf("volId:              %lld (0x%llx)\n", odCfg->id.whole_thing, odCfg->id.whole_thing);
    printf("name:               '%s'\n", odCfg->name);
    printf("currSize:           %lld\n", odCfg->currSize);
    printf("hardQuota:          %lld\n", odCfg->hardQuota);
    printf("softQuota:          %d\n", odCfg->softQuota);
    printf("minAutoGrow:        %lld\n", odCfg->minAutoGrow);
    
    printf("raidLevel:          %s\n", (odCfg->raidLevel == EVM_RAID_ANY) ?
           "no raid level" : evm_raid_level_strings[odCfg->raidLevel]);
    printf("minLunSize:         %d\n", odCfg->minLunSize);
    printf("maxLunSize:         %d\n", odCfg->maxLunSize);
    //printf("\twwn: 0x%llx\n", odCfg->wwn);
    printf("raidController:     %s\n", odCfg->raidCtrl);
    printf("nbVendor:           %d\n", odCfg->nbVendor);
    for (i = 0; i < odCfg->nbVendor; i++)
    {
        printf("vendor/model        %s/%s\n",
               (strlen(odCfg->vendorInfo[i].vendor) ?
                odCfg->vendorInfo[i].vendor : "no vendor name"),
               (strlen(odCfg->vendorInfo[i].model) ?
                odCfg->vendorInfo[i].model : "no model name"));
    } 

    printf("allocUnit:          %d\n", odCfg->allocUnit);
    printf("lblk:               %lld\n", odCfg->lblk);
    printf("stBlk:              %d\n", odCfg->stBlk);
    printf("nBlks:              %d\n", odCfg->nBlks);
    printf("last label mod.     %s", ctime((time_t *)&odCfg->timestamp));

    printf("\n");

    printf("*** MANAGEMENT AREA ****\n");
    printf("magic:              0x%x\n", mgr->magic);
    printf("mgrId cluster name  %s\n", mgr->mgrId.cluster);
    printf("mgrId node name     %s\n", mgr->mgrId.node);
    printf("mgrId chassisId     %d (0x%x)\n", mgr->mgrId.chassisId, mgr->mgrId.chassisId);
    printf("timestamp:          %s", ctime((time_t *)&mgr->timestamp));
    printf("cksum               0x%x\n", mgr->cksum);
    printf("\n");

    printf("*** TAKEOVER AREA ****\n");
    printf("magic:              0x%x\n", tko->magic);
    printf("mgrId cluster name  %s\n", tko->tkoId.cluster);
    printf("mgrId node name     %s\n", tko->tkoId.node);
    printf("mgrId chassisId     %d (0x%x)\n", tko->tkoId.chassisId, tko->tkoId.chassisId);
    printf("timestamp:          %s", ctime((time_t *)&tko->timestamp));
    printf("cksum               0x%x\n", tko->cksum);
    printf("\n");

    return NFX_OK;

}


