/*------------------------------------------------------------------------------
 * RCS:  $Id: fssf-api.c,v 1.18 2002/12/10 22:16:22 bobonfeline Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : fssf-api.c
 *
 * Description  : Entry module of the File System State-Full (FSSF).
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Header files.
 *------------------------------------------------------------------------------
 */

#include "fssf.h"

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

/* FSSF application id.
 * FSSF and FS port addresses.
 */
int32           fssfAppId = 0;
uint32          fssfPort = 0;
uint32          fsPort = 0;

/* Spinlock for the global variables.
 */
nfx_spinlock_t  fssf_stats_lock;
nfx_spinlock_t  fssf_tab_lock;

/* FSSF statistics.
 */
fssf_stats_t    fssfStats;

/* Open file(s).
 * Hash table.
 */
uint32          fileTab = 0;

/* Client with open file(s).
 * Hash table
 */
uint32          ownerTab = 0;

/* Open state(s).
 * Indexed table.
 */
fssf_openTab_t  openTab;

/* Lock state(s).
 * Indexed table.
 */
fssf_lockTab_t  lockTab;

/*------------------------------------------------------------------------------
 * Function prototypes.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : fssf_rcvMsg
 *
 * Description  : Receives unicast message.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_rcvMsg(eee_descPtr_t edesc)
    
{
	uint32          msgLen;
	char *          msgBuf;
	uint32 *        msgType;
    eee_descBuf_t * bd;
    boolean         freeMsgBuf = TRUE;
    int32           rc, numBufs;
    uint32          srcPort;//, destPort;

#ifndef FS_STANDALONE
    numBufs = EEE_GET_NUM_BUFS(edesc);
    srcPort = edesc->hdr.src_port;
    //msg.destPort = edesc->hdr.dest_port;
    if (numBufs == 1)
    {
        // Single buffer message, don't copy to flat buffer.
        if (edesc->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER)
            bd = &(edesc->bd[0]);
        else
            bd = &(edesc->bd[1]);
        msgBuf = bd->buf;

        // Done with the descriptor.
        eee_deallocateDesc(edesc);
    }
    else
    {
        // Copy the multi-buffer message to a
        // contiguous malloc buffer and free the
        // received packet.
        msgLen = EEE_GET_PKT_LEN(edesc);
        msgBuf = (char *) fssf_ramAlloc(msgLen);
        if (msgBuf == NULL)
        {
            FSSF_ERR_PRINT(("FSSF: no mem, dropping packet"));
            eee_freePacket(edesc);
            return;
        }
        rc = eee_copyPacketToLocalBuffer(msgBuf, edesc, msgLen, 0);
        if (rc != msgLen)
        {
            FSSF_ERR_PRINT(("FSSF: copyPacketToLocal failed"));
            eee_freePacket(edesc);
            return;
        }

        // Done with the packet.
        eee_freePacket(edesc);
    }
#endif // FS_STANDALONE

    /* Process the message.
     */
    fssf_processMsg(msgBuf, srcPort);

#ifndef FS_STANDALONE
    if (freeMsgBuf)
    {
        if (numBufs == 1)
            eee_deallocateBuffer(msgBuf);
        else
            fssf_ramDealloc(msgBuf);
    }
#endif // FS_STANDALONE
    return;
}

/*------------------------------------------------------------------------------
 * Name         : fssf_processMsg
 *
 * Description  : Process received message.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_processMsg(char *  msgBuf,
                uint32  srcPort)
    
{
    uint32 * msgType = (uint32 *) msgBuf;
    
    /* Get the type of the message and dispatch it.
     */
    switch(*msgType)
    {

    /* Requests from client.
     * ---------------------
     */
    case FSSF_CLOSE_REQ:
    {
        fssf_cb_t * cb;
        
        FSSF_DBG_PRINT(("FSSF_CLOSE_REQ"));
        cb = (fssf_cb_t *) fssf_ramAllocZero(sizeof(fssf_cb_t));
        cb->sender = srcPort;
        cb->msgBuf = msgBuf;
        cb->threadHandle = thread_create(cb);
        NFX_ASSERT(cb->threadHandle != 0);
        NFX_ASSERT(thread_start(&(cb->threadHandle),
                                fssf_processCloseReq) == NFX_OK);
        break;
    } // case FSSF_CLOSE_REQ

    case FSSF_CREATE_REQ:
    {
        FSSF_DBG_PRINT(("FSSF_CREATE_REQ"));
        //fssf_processCreateReq((fssf_createReq_t *) msgBuf);
        break;
    } // case FSSF_CREATE_REQ

    case FSSF_GETATTR_REQ:
    {
        fssf_cb_t * cb;
        
        FSSF_DBG_PRINT(("FSSF_GETATTR_REQ"));
        cb = (fssf_cb_t *) fssf_ramAllocZero(sizeof(fssf_cb_t));
        cb->sender = srcPort;
        cb->msgBuf = msgBuf;
        cb->threadHandle = thread_create(cb);
        NFX_ASSERT(cb->threadHandle != 0);
        NFX_ASSERT(thread_start(&(cb->threadHandle),
                                fssf_processGetattrReq) == NFX_OK);
        break;
    } // case FSSF_GETATTR_REQ

    case FSSF_LOCK_REQ:
    {
        FSSF_DBG_PRINT(("FSSF_LOCK_REQ"));
        //fssf_processLockReq((fssf_lockReq_t *) msgBuf);
        break;
    } // case FSSF_LOCK_REQ

    case FSSF_OPEN_REQ:
    {
        fssf_cb_t * cb;
        
        FSSF_DBG_PRINT(("FSSF_OPEN_REQ"));
        cb = (fssf_cb_t *) fssf_ramAllocZero(sizeof(fssf_cb_t));
        cb->sender = srcPort;
        cb->msgBuf = msgBuf;
        cb->threadHandle = thread_create(cb);
        NFX_ASSERT(cb->threadHandle != 0);
        NFX_ASSERT(thread_start(&(cb->threadHandle),
                                fssf_processOpenReq) == NFX_OK);
        break;
    } // case FSSF_OPEN_REQ

    case FSSF_READ_REQ:
    {
        fssf_cb_t * cb;

        FSSF_DBG_PRINT(("FSSF_READ_REQ"));
        cb = (fssf_cb_t *) fssf_ramAllocZero(sizeof(fssf_cb_t));
        cb->sender = srcPort;
        cb->msgBuf = msgBuf;
        cb->threadHandle = thread_create(fssf_processReadReq);
        NFX_ASSERT(cb->threadHandle != 0);
        NFX_ASSERT(thread_start(&(cb->threadHandle),
                                fssf_processReadReq) == NFX_OK);
        break;
    } // case FSSF_READ_REQ

    case FSSF_REMOVE_REQ:
    {
        FSSF_DBG_PRINT(("FSSF_REMOVE_REQ"));
        //fssf_processRemoveReq((fssf_removeReq_t *) msgBuf);
        break;
    } // case FSSF_REMOVE_REQ

    case FSSF_SETATTR_REQ:
    {
        FSSF_DBG_PRINT(("FSSF_SETATTR_REQ"));
        //fssf_processSetattrReq((fssf_setattrReq_t *) msgBuf);
        break;
    } // case FSSF_SETATTR_REQ

    case FSSF_UNLOCK_REQ:
    {
        FSSF_DBG_PRINT(("FSSF_UNLOCK_REQ"));
        //fssf_processUnlockReq((fssf_unlockReq_t *) msgBuf);
        break;
    } // case FSSF_UNLOCK_REQ

    case FSSF_WRITE_REQ:
    {
        FSSF_DBG_PRINT(("FSSF_WRITE_REQ"));
        //fssf_processWriteReq((fssf_writeReq_t *) msgBuf);
        break;
    } // case FSSF_WRITE_REQ

    /* Responses from EFS.
     * -------------------
     */
    case FS_REQ_CREATE:
    {
        int32           rc = NFX_OK;
        fs_MsgRespHdr * hdr = (fs_MsgRespHdr *) msgBuf;
        esm_event_t     evt;
        
        FSSF_DBG_PRINT(("FS_REQ_CREATE"));
        FSSF_STAT_FSCREATERSP;
        
        bzero((void *) &evt, sizeof(esm_event_t));
        evt.event = FS_REQ_CREATE;
        evt.p1.u.p = (void *) msgBuf;
        rc = thread_event(hdr->clientPrivate, &evt);
        NFX_ASSERT(rc == NFX_OK);
        break;
    } // case FS_REQ_CREATE
    
    case FS_REQ_LOOKUP:
    {
        int32           rc = NFX_OK;
        fs_MsgRespHdr * hdr = (fs_MsgRespHdr *) msgBuf;
        esm_event_t     evt;
        
        FSSF_DBG_PRINT(("FS_REQ_LOOKUP"));
        FSSF_STAT_FSCREATERSP;
        
        bzero((void *) &evt, sizeof(esm_event_t));
        evt.event = FS_REQ_LOOKUP;
        evt.p1.u.p = (void *) msgBuf;
        rc = thread_event(hdr->clientPrivate, &evt);
        NFX_ASSERT(rc == NFX_OK);
        break;
    } // case FS_REQ_LOOKUP
    
    default:
    {
        FSSF_ERR_PRINT(("FSSF: Unknown message (%d)!\n", *msgType));
        break;
    } // default
    
    } // switch

}

/*------------------------------------------------------------------------------
 * Name         : fssf_rcvMulticast
 *
 * Description  : Receives multicast messages.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_rcvMcastMsg(eee_descPtr_t desc)
    
{
    // ignore message
    return;
}

/*------------------------------------------------------------------------------
 * Name         : fssf_init
 *
 * Description  : Initialize the FSSF module.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 04-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int32
fssf_init(void)
    
{
    int32   rc = NFX_OK;
    
    FSSF_FCT_PRINT(("FSSF: > main()\n"));

#ifndef FS_STANDALONE
    /* Initialize our appllication id and port address.
     * Register our message handler.
     */
    fssfAppId = eee_allocateApplicationId ("fssf");
    if (fssfAppId == EEE_ERR_INVALID_APP_NAME)
    {
        rc = NFX_ERR;
        goto fssf_init_exit;
    }
    
    fssfPort = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId, nfxMyCPUId,
                                         fssfAppId,
                                         0, 0, 0);
    fsPort = PACK_UNICAST_PORT_ADDRESS(nfxMySlotId, nfxMyCPUId,
                                       eee_allocateApplicationId ("fs"),
                                       0, 0, 0);
    
    rc = eee_registerReceiveFunc(fssfAppId, fssf_rcvMsg, fssf_rcvMcastMsg);
    if (rc != NFX_OK)
    {
        goto fssf_init_exit;
    }
#endif // FS_STANDALONE
    
    /* Initialize the statistics.
     */
    bzero((void *) &fssfStats, sizeof(fssf_stats_t));
    NFX_SPIN_INIT(&fssf_stats_lock, SPINLOCK_LEVEL_UNKNOWN);
    
    /* Initialize the state-full structures.
     * TODO!!! initialize less and grow later depending on available memory
     */
    fssf_initFileTab(FSSF_HASH_BUCKETS_OPEN_FILE);
    fssf_initOwnerTab(FSSF_HASH_BUCKETS_OWNER);
    fssf_initOpenTab(FSSF_MIN_OPEN_STATE);
    fssf_initLockTab(FSSF_MIN_LOCK_STATE);
   
  fssf_init_exit:
    return(rc);
}

/*----------------------------------------------------------------------------*/
