/*------------------------------------------------------------------------------
 * RCS:  $Id: fssf-open.c,v 1.1 2002/03/27 00:09:48 isabellec Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : fssf-open.c
 *
 * Description  : File System State-Full (FSSF)
 *                Operations on the openTab.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 26-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

#include "fssf.h"
#include "fssf-hash.h"

/*------------------------------------------------------------------------------
 * Local function prototypes.
 *------------------------------------------------------------------------------
 */

static void
fssf_growOpenTab(int growSize);

static fssf_state_t *
fssf_allocOpenState(void);

static void
fssf_freeOpenState(fssf_state_t *   state);

/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : fssf_growOpenTab
 *
 * Description  : Grow the open states table.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 10-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static void
fssf_growOpenTab(int growSize)

{
    int             i;
    fssf_state_t *  tab = NULL;
    fssf_state_t ** ptab = NULL;
    
    NFX_SPIN_LOCK(&fssf_tab_lock);
    tab = (fssf_state_t *) fssf_ramAllocZero(growSize * sizeof(fssf_state_t ));
    ptab = (fssf_state_t **) fssf_ramAllocZero((openTab.size + growSize) *
                                               sizeof(fssf_state_t *));
    for (i = 0; i < openTab.size; i++)
    {
        ptab[i] = openTab.tab[i];
    } // for
    for (i = openTab.size; i < (openTab.size + growSize); i++)
    {
        tab[i].stateid.parts.cid = 0x01234567; // TODO!!!
        tab[i].stateid.parts.idx = i - openTab.size;
        ptab[i] = &tab[i - openTab.size];
        DLL_INSERT_TAIL(&(openTab.freeList), ptab[i], freeState);
    } // for
    fssf_ramDealloc(openTab.tab);
    openTab.tab = ptab;
    openTab.size = openTab.size + growSize;
    NFX_SPIN_UNLOCK(&fssf_tab_lock);
    return;
}

/*------------------------------------------------------------------------------
 * Name         : fssf_allocOpenState
 *
 * Description  : Return a free state.
 *                If no more free states and still possibility to grow, do it.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 16-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static fssf_state_t *
fssf_allocOpenState(void) 
    
{
    fssf_state_t *  state = NULL;

    DLL_REMOVE_HEAD(&(openTab.freeList), state, freeState);
    if ((state == NULL) && (openTab.size < FSSF_MAX_OPEN_STATE))
    {
        fssf_growOpenTab(FSSF_MIN_OPEN_STATE);
        DLL_REMOVE_HEAD(&(openTab.freeList), state, freeState);
        NFX_ASSERT(state == NULL);
    }
    
    return(state);
}
    
/*------------------------------------------------------------------------------
 * Name         : fssf_freeOpenState
 *
 * Description  : Free a used state.
 *                Increment the sequence number within the state id.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 16-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

static void
fssf_freeOpenState(fssf_state_t *   state) 
    
{
    state->stateid.parts.seq++;
    DLL_INSERT_TAIL(&(openTab.freeList), state, freeState);
    return;
}
    
/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : fssf_initOpenTab
 *
 * Description  : Initialize the open states table.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 10-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_initOpenTab(int initSize)

{
    int             i;
    fssf_state_t *  tab = NULL;
    fssf_state_t ** ptab = NULL;
        
    NFX_SPIN_LOCK(&fssf_tab_lock);
    tab = (fssf_state_t *) fssf_ramAllocZero(initSize * sizeof(fssf_state_t ));
    ptab = (fssf_state_t **) fssf_ramAllocZero(initSize *
                                               sizeof(fssf_state_t *));
    DLL_INIT(&(openTab.freeList));
    for (i = 0; i < initSize; i++)
    {
        tab[i].stateid.parts.cid = 0x01234567; // TODO!!!
        tab[i].stateid.parts.idx = i;
        ptab[i] = &tab[i];
        DLL_INSERT_TAIL(&(openTab.freeList), ptab[i], freeState);
    } // for
    openTab.size = initSize;
    openTab.tab = ptab;
    NFX_SPIN_UNLOCK(&fssf_tab_lock);
    return;
}

/*------------------------------------------------------------------------------
 * Name         : fssf_addOpenState
 *
 * Description  : Get a entry in the table for this new open file.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 16-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_addOpenState(fssf_state_t **   state)  // output
    
{
    *state = fssf_allocOpenState();
    return;
}

/*------------------------------------------------------------------------------
 * Name         : fssf_delOpenState
 *
 * Description  : Remove a state from openTab.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 18-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void
fssf_delOpenState(fssf_state_t **   state)  // input/output
    
{
    fssf_freeOpenState(*state);
    *state = NULL;
    return;
}

/*------------------------------------------------------------------------------
 * Name         : fssf_getState
 *
 * Description  : Get a entry in the table for this state file.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 18-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

fssf_status_t
fssf_getState(fssf_stateid_t *  stateid,    // input
              fssf_state_t **   openState)  // output
    
{
    int32   rc = NFX_OK;

    if ((stateid->parts.cid == 0x01234567) && // TODO!!!
        (0 <= stateid->parts.idx) && (stateid->parts.idx < openTab.size))
    {
        *openState = openTab.tab[stateid->parts.idx];
        if (((*openState)->fileShare == NULL) && ((*openState)->owner == NULL))
        {
            *openState = NULL;
            rc = FSSF_ERR_STALE_STATEID;
        }
    }
    else
    {
        rc = FSSF_ERR_BAD_STATEID;
    }
    
    return(rc);
}

/*----------------------------------------------------------------------------*/
