/*
 *-----------------------------------------------------------------
 *
 * Name:	pte.h
 *
 * RCS:  $Id: pte-api.h,v 1.2 2002/07/25 19:43:55 rick Exp $
 *
 * Copyright (C) 2002, Agile Storage, Inc.
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/20/02
 *
 *-----------------------------------------------------------------
 */

#ifndef _PTE_API_H
#define _PTE_API_H

#define MMU_NUM_TABLES			2

#define MMU_MALLOC_REGION		1
#define MMU_SM_REGION			2

extern address_t    mmu_phys_offset_start[MMU_NUM_TABLES+1];
extern address_t    mmu_phys_offset_end[MMU_NUM_TABLES+1];
extern address_t    mmu_phys_start[MMU_NUM_TABLES+1];
extern address_t    mmu_phys_end[MMU_NUM_TABLES+1];
extern address_t    mmu_virt_start[MMU_NUM_TABLES+1];
extern address_t    mmu_virt_end[MMU_NUM_TABLES+1];

#if 0
/*------------------------------------------------------------------
 *
 * The pte_t strcture's size of 32 bytes is tightly coupled with xtlb refill
 * handler, which calculates table offset by multiplying index * 32, which
 * is the size of the pte_t structure.  If you change this, you'll also need
 * to change the multipler in xtlb refill, and it needs to be an exponent of
 * 2 in size, so that the refill handler can do its calculations with only one
 * processor register.
 *
 *------------------------------------------------------------------*/
typedef struct pte_struct {
	address_t	tlb_lo0_phys;
	address_t	tlb_lo1_phys;
	uint32		pgsize;
	uint32		cattr;
	address_t	res;
} pte_t;

/*------------------------------------------------------------------
 *
 * Please note that these are tightly coupled with the xtlb refill handler.
 * So you will want to understand what you're doing before changing them.
 * They have dependencies and ramafications on:
 *		memory usage
 *		addressing
 *		TLB entry sizes
 *
 *------------------------------------------------------------------*/
#define PTE_ENTRY_SIZE		0x800000			/* 8 MB per entry */
#define PTE_MAX_MEMORY		0x100000000			/* 4 GB of memory/region */

/*------------------------------------------------------------------
 *
 * All regions are justified to 8 MB (since each TLB has 2 pages of either
 *	4 MB or 16 MB).  Anything smaller than 4 MB per TLB page is completely
 *	inefficient for normally embedded operation.  It seemed wasteful to
 *	justify spaces on 32 MB boundary, so I compromised with 8 MB.
 *
 * We're going to use bits [39:36] of virtual address to determine region:
 *
 *		10_0000_0000 = Malloc base
 *		20_0000_0000 = TXRX shared memory base
 *		30_0000_0000 = SAC base (in TXRX memory)
 *		40_0000_0000 = FP shared memory base
 *		50_0000_0000 = PCI
 *
 * In the refill exception handler, we get the region by shifting the XCONTEXT
 * register (which holds bits 39:13 of the requested virtual address), and
 * masking to get bits 39:36.  Those bits are the region, and that region
 * specifies which mmu table we need to use to get the new tlb page entries.
 *
 *		lm_table = malloc PTEs
 *		sm_txrx_table = TXRX shared memory PTEs
 *		sac_table = sac memory PTEs
 *		sm_fp_table = FP shared memory PTEs
 *		pci_table = PCI PTEs
 *
 * We support up to 4 GB of addressable memory in each region.  We limit regions
 * to 4 GB to minimize the mmu table sizes.
 *
 *		4 GB / 8 MB = 512 entries per table
 *		512 * 32 B = 16 KB per table
 *		16 KB * 5 tables = 80 KB
 *
 * If we supported 16 GB/region:
 *		16 GB/8 MB = 8K entries * 32 B/entry = 256 KB/table * 5 = 1.25 MB
 *
 * Also, there is checking in the refill handler to verify that bits 36:32 are
 * 0 (within 4 GB of the memory region base).
 *
 *------------------------------------------------------------------*/

#define MAX_PTE_TABLE_ENTRIES	(PTE_MAX_MEMORY/PTE_ENTRY_SIZE)

extern pte_t	lm_table[MAX_PTE_TABLE_ENTRIES];
extern pte_t	sm_txrx_table[MAX_PTE_TABLE_ENTRIES];
extern pte_t	sm_fp_table[MAX_PTE_TABLE_ENTRIES];
extern pte_t	sac_table[MAX_PTE_TABLE_ENTRIES];
extern pte_t	pci_table[1];

extern uint64		mmu_table_size[MMU_NUM_TABLES+1];
extern address_t	mmu_phys_offset_start[MMU_NUM_TABLES+1];
extern address_t	mmu_phys_offset_end[MMU_NUM_TABLES+1];
extern address_t	mmu_phys_start[MMU_NUM_TABLES+1];
extern address_t	mmu_phys_end[MMU_NUM_TABLES+1];
extern address_t	mmu_virt_start[MMU_NUM_TABLES+1];
extern address_t	mmu_virt_end[MMU_NUM_TABLES+1];

extern pte_t		* mmu_tptr[MMU_NUM_TABLES+1];

/*-----------------------------------------------------------------
 * Name:	pte_setup()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/20/02
 *
 *-----------------------------------------------------------------
 */
void
pte_setup(
    uint32      region,         /*  MMU_MALLOC_REGION,
                                    MMU_SM_TXRX_REGION,
                                    MMU_SM_FP_REGION,
                                    MMU_SAC_REGION,
                                    MMU_PCI_REGION,
                                    MMU_WKA_REGION,
                                    MMU_PCI_MBOX_REGION
                                 */
    uint64      phys_offset,    /* offset from 0, assuming contiguous memory */
    address_t   virt,           /* virtual address */
    uint64      size            /* of entire region, in bytes */
);

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/20/02
 *
 *-----------------------------------------------------------------
 */
address_t
virt_to_phys(
	address_t
);

/*-----------------------------------------------------------------
 * Name:
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/20/02
 *
 *-----------------------------------------------------------------
 */
address_t
phys_to_virt(
	address_t
);

/*-----------------------------------------------------------------
 * Name:	get_memory_region_size()
 *
 * Description: returns size of requested memory region, in bytes
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/22/02
 *
 *-----------------------------------------------------------------
 */
uint64
get_memory_region_size(
	uint32	region_id		/*  MMU_MALLOC_REGION,
								MMU_SM_TXRX_REGION,
								MMU_SAC_REGION,
								MMU_SM_FP_REGION,
								MMU_WKA_REGION,
							*/
);

/*-----------------------------------------------------------------
 * Name:	get_memory_virt_base()
 *
 * Description:	returns virtual base address of requested region
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/22/02
 *
 *-----------------------------------------------------------------
 */
address_t
get_memory_virt_base(
	uint32	region_id		/*  MMU_MALLOC_REGION,
								MMU_SM_TXRX_REGION,
								MMU_SAC_REGION,
								MMU_SM_FP_REGION,
								MMU_WKA_REGION,
							*/
);

/*-----------------------------------------------------------------
 * Name:	get_memory_phys_base()
 *
 * Description:
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/22/02
 *
 *-----------------------------------------------------------------
 */
address_t
get_memory_phys_base(
	uint32	region_id		/*  MMU_MALLOC_REGION,
								MMU_SM_TXRX_REGION,
								MMU_SAC_REGION,
								MMU_SM_FP_REGION,
								MMU_WKA_REGION,
							*/
);

#endif

#endif	/* !_PTE_API_H */

