/*
 *-----------------------------------------------------------------
 *
 * Name:	vs-hash-api.c
 *
 * RCS:  $Id: $
 *
 * Copyright (C) 2003, Agile Storage Inc.
 *
 * Description:	Virtual server hash table implementation.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	01/15/03
 *
 *-----------------------------------------------------------------
 */

/*
#include <nfx-incl.h>
#include <exports/hash-api.h>
#include <exports/malloc-api.h>
*/
#include <exports/vs-hash-api.h>
#include "vs-hash.h"

/*-----------------------------------------------------------------
 * Name:	vs_hash_create(num_buckets, hash, compare)
 *
 * Description:	Allocate memory for a hash table
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	01/15/03
 *
 *-----------------------------------------------------------------
 */
address_t			/* returns hash table handle, or NULL */
vs_hash_create(
	uint32	num_buckets,
	uint32	(* hash)(
				void	* key
			),
	int32	(* compare)(
				vs_mem_index_t	ht_data,
				vs_mem_index_t	search_data
			),
	void 	(* release)(
				vs_mem_index_t	data
			)
)
{
	vs_ht_t	* htp;
	uint32	i;

	htp = (vs_ht_t *)eee_ramAlloc(sizeof(vs_ht_t));
	htp->head = (vs_ht_item_t **)
						eee_ramAlloc(num_buckets * sizeof(vs_ht_item_t *));

	htp->hash = hash;
	htp->compare = compare;
	htp->release = release;
	for (i=0; i< num_buckets; i++)
		htp->head[i] = NULL;

	htp->num_buckets = num_buckets;
	NFX_SPIN_INIT(&htp->ht_spin, SPINLOCK_LEVEL_UNKNOWN);

	return ((address_t)htp);

}	/* vs_hash_create */

/*-----------------------------------------------------------------
 * Name:	vs_hash_destroy(htp)
 *
 * Description:	Destroy a hash table, freeing memory allocated by hash
 *				table routines.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	01/15/03
 *
 *-----------------------------------------------------------------
 */
void
vs_hash_destroy(
	address_t	htp
)
{
	vs_ht_item_t	* np;
	uint32			i;

	for (i=0; i<((vs_ht_t *)htp)->num_buckets; i++) {
		np = ((vs_ht_t *)htp)->head[i];
		while (np != NULL) {
			vs_ht_item_t	* temp;

			temp = np;
			np = np->next;
			if (((vs_ht_t *)htp)->release)
				(*((vs_ht_t *)htp)->release)(temp->data);
			eee_ramDealloc(temp);
		}
	}

	eee_ramDealloc(((vs_ht_t *)htp)->head);
	eee_ramDealloc((void *)htp);

}	/* vs_hash_destroy */

static inline int
vs_do_compare(struct vs_ht *ht, vs_ht_item_t *np, vs_mem_index_t data)
{
	return ht->compare ? ht->compare(np->data, data) == 0 : 
						 VS_MEM_IDX_IS_EQUAL(np->data, data);
}


/*-----------------------------------------------------------------
 * Name:	vs_hash_insert(htp, key, data)
 *
 * Description:	Insert item into hash table.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	01/15/03
 *
 *-----------------------------------------------------------------
 */
int32						/* return NFX_OK or VS_HT_ERR_DATA_IN_USE */
vs_hash_insert(
	address_t		htp,			/* hash table */
	void			* key,			/* key for hashing function */
	vs_mem_index_t	data			/* data structure index to insert */
)
{
	vs_ht_item_t	* np;
	uint32			hashval;
	int				rc = VS_HT_ERR_DATA_IN_USE;
	struct vs_ht 	* ht = (struct vs_ht *)htp;

	hashval = (*ht->hash)(key);
	if (hashval >= ht->num_buckets)
		panic("hashval(0x%x) > num_buckets(0x%x)\n",
			  hashval, ht->num_buckets);

	NFX_SPIN_LOCK(&ht->ht_spin);
	for (np = ht->head[(int)hashval]; np != NULL; np = np->next) {
		if (vs_do_compare(ht, np, data)) {
			break;
		}
	}
	if (np == NULL) {
		np = vs_hash_getNode();
		np->next = ht->head[(int)hashval];
		np->data = data;
		ht->head[hashval] = np;
		rc = NFX_OK;
	}
	NFX_SPIN_UNLOCK(&ht->ht_spin);

	return rc;

}	/* vs_hash_insert */


/*-----------------------------------------------------------------
 * Name:	vs_hash_lookup(ht, key, data)
 *
 * Description:	Find an item in the hash table.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	01/15/03
 *
 *-----------------------------------------------------------------
 */
vs_mem_index_t		/* returns data structure index stored in table,
					   or zero-value index */
vs_hash_lookup(
	address_t		htp,			/* hash table */
	void			* key,			/* key for hashing function */
	vs_mem_index_t	data			/* data structure index to find */
)
{
	vs_ht_item_t	* np;
	uint32			hashval;
	struct vs_ht 	* ht = (struct vs_ht *)htp;
	vs_mem_index_t	temp, zero;

	hashval = ht->hash(key);
	if (hashval >= ht->num_buckets)
		panic("hashval(0x%x) > num_buckets(0x%x)\n",
			  hashval, ht->num_buckets);
	
	NFX_SPIN_LOCK(&ht->ht_spin);
	for (np = ht->head[(int)hashval]; np != NULL; np = np->next) {
		if (vs_do_compare(ht, np, data))
			break;
	}
	VS_MEM_IDX_SET_NULL(zero);
	temp = np ? np->data : zero;
	NFX_SPIN_UNLOCK(&ht->ht_spin);

	return temp;

}	/* vs_hash_lookup */

/*-----------------------------------------------------------------
 * Name:	vs_hash_delete(ht, key, data)
 *
 * Description:	Remove an item from a hash table.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	01/15/01
 *
 *-----------------------------------------------------------------
 */
vs_mem_index_t		/* returns index to data structure, or zero-value index */
vs_hash_delete(
	address_t		htp,			/* hash table */
	void			* key,			/* key for hashing function */
	vs_mem_index_t	data			/* data structure index to find & delete */
)
{
	vs_ht_item_t	* np;
	vs_ht_item_t	**prev;
	uint32			hashval;
	struct vs_ht 	*ht = (struct vs_ht *)htp;
	vs_mem_index_t	temp;

	hashval = (*ht->hash)(key);
	if (hashval >= ht->num_buckets)
		panic("hashval(0x%x) > num_buckets(0x%x)\n",
										hashval, ht->num_buckets);
	NFX_SPIN_LOCK(&ht->ht_spin);

	prev = &ht->head[(int)hashval];
	np = ht->head[(int)hashval];
	while (np != NULL && vs_do_compare(ht, np, data) == 0) {
		prev = &np->next;
		np = np->next;
	}
	if (np != NULL) {
		*prev = np->next;
		temp = np->data;
		vs_hash_freeNode(np);
	}
	else {
		VS_MEM_IDX_SET_NULL(temp);
	}
	
	NFX_SPIN_UNLOCK(&ht->ht_spin);
	return temp;
}	/* vs_hash_delete */

