/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: ipmd-stats.c
 *
 * RCS:  $Id: ipmd-stats.c,v 1.5 2002/09/23 16:56:32 swang Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: process stats 
 *
 * Created by:	Su Wang
 *
 * Date Created: 02/11/2002
 *
 *-----------------------------------------------------------------
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/if_ether.h>

#include <nfx-error.h>
#include <nfx-types.h>
#include <nfx-defs.h>
#include <exports/anpssc-api.h>
#include <exports/ipmd-lib.h>
#include <exports/ipm-api.h>
#include "../sm-ipm/ipm.h"
#include "ipmd-stats.h"

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_FindConn
 *
 * Description: Find a connection in table.
 *
 * Created by:	Su Wang
 *
 * Date Created: 12/13/2001
 *
 *-----------------------------------------------------------------
 */
static ipm_inconn_t*
ipmd_FindConn(ipm_conn_t*		connp, /* record to be searched */
			  ipm_inconn_t		*connroot, /* root of record chain */
			  cmp_func			cmp)
{
  ipm_inconn_t			*connaddr;
  int					rc;
  
  for (connaddr = connroot; connaddr;/* end of table */
	   connaddr = connaddr->next)
	{
	  rc = (*cmp)(connp, &(connaddr->conn));
	  if (rc == 0)
		return connaddr;
	  else if (rc < 0)
		break;
	}
  
  return NULL;
}

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_CmpConn
 *
 * Description: compare the record of two conn. record in snmp order.
 *
 * Created by:	Su Wang
 *
 * Date Created: 12/13/2001
 *
 *-----------------------------------------------------------------
 */
static int						/* -1 less, 0 equal, 1 large */
ipmd_CmpConn(void		*in1, 
			 void		*in2)
{
	/* do not cmp ipm_conn_t->state */
	return memcmp(in1, in2, (sizeof(ipm_conn_t)-4));
}

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_CpyConnToLocal
 *
 * Description:
 *
 * Created by:	Su Wang
 *
 * Date Created: 12/13/2001
 *
 *-----------------------------------------------------------------
 */
static void
ipmd_CpyConnToLocal(ipm_conn_t	*connp, 
				    wi_t		*wi)
{
  char	*ptr;

  ptr = (char*)(wi->cp_ptr);
  memcpy(ptr, connp, sizeof(ipm_conn_t));
  
  /* move along the ptrs and counter */
  wi->cp_ptr = (void*)(ptr + sizeof(ipm_conn_t));
  wi->cp_cnt--;
}

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_WalkConn
 *
 * Description: walk through connection record, performing given func.
 *
 * Created by:	Su Wang
 *
 * Date Created: 12/13/2001
 *
 *-----------------------------------------------------------------
 */
static void
ipmd_WalkConn(ipm_inconn_t	*connroot,
			  walk_func		walk, /* function executed at each node */
			  wi_t			*wi)  /* input parameter for func (*walk) */
{
  ipm_inconn_t *connaddr;
  
  for (connaddr = connroot; (connaddr && (wi->cp_cnt > 0));
	   connaddr = connaddr->next)
	  (*walk)(&(connaddr->conn), wi);
  
  return;
}


/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_InsConn
 *
 * Description: insert conn. to UPD/TCP conn table. 
 *
 * *connroot: the offset to start of shmaddr that represents
 *			  the said conn. table.
 *
 * Created by:	Su Wang
 *
 * Date Created: 12/13/2001
 *
 *-----------------------------------------------------------------
 */
#define IPM_DUP	(NFX_ERR - 1)
static int32					/* NFX_OK, NFX_ERR, IPM_DUP */
ipmd_InsConn(ipm_conn_t		*connp, 
			 ipm_inconn_t	**connroot, 
			 cmp_func		cmp)
{
  ipm_inconn_t			*connaddr=0, *prev_connaddr=0, *new_connaddr; 
  int					rc;
  

  /* find the place for insertion */
  for (connaddr = *connroot; connaddr; connaddr = connaddr->next) {

	  rc = (*cmp)(connp, &(connaddr->conn));
	  
	  if (rc == 0) {
		memcpy(&(connaddr->conn), connp, sizeof(ipm_conn_t));
		return IPM_DUP;			/* conn. already exist */
	  }
	  else if (rc < 0)
		break;					/* find the place */

	  prev_connaddr = connaddr;
  }

  /* allocate and cpy */
  IPM_ALLOC_CONN(new_connaddr)

  if (new_connaddr == 0)
	  return NFX_ERR;

  memcpy(&(new_connaddr->conn), connp, sizeof(ipm_conn_t));

  if (prev_connaddr == NULL)	/* inserting element is the first */
	{
	  new_connaddr->next = *connroot;
	  *connroot = new_connaddr;
	}
  else							/* insert in the middle */
	{
	  new_connaddr->next = prev_connaddr->next;
	  prev_connaddr->next = new_connaddr;
	}

  return NFX_OK;
} 


/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_DelConn
 *
 * Description: delete conn. from UPD/TCP conn table. 
 *
 * *connroot: the offset to start of shmaddr that represents
 *			  the said conn. table.
 *
 * Created by:	Su Wang
 *
 * Date Created: 12/13/2001
 *
 *-----------------------------------------------------------------
 */
static int32					/* NFX_OK, NFX_ERR */
ipmd_DelConn(ipm_conn_t			*connp, 
			 ipm_inconn_t		**connroot, 
			 cmp_func			cmp)
{
  ipm_inconn_t *connaddr = 0, *prev_connaddr=0;
  int			rc;

  /* find the node to delete */
  for (connaddr=*connroot; connaddr; connaddr = connaddr->next) {
	
	  rc = (*cmp)(connp, &(connaddr->conn));
	  
	  if (rc == 0)				/* node found */
		{
		  if (prev_connaddr == NULL) /* delete first node */ 
			*connroot = connaddr->next;
		  else					/* in the midle */
			prev_connaddr->next = connaddr->next; 

		  IPM_FREE_CONN(connaddr)
		  return NFX_OK;
		}
	  else if (rc < 0)			/* list in ascending order, nothing found */
		break;

	  prev_connaddr = connaddr;
  }

  return NFX_ERR;
}





/* make stats memory static for now */
static	uint64	stats_buf[IPM_STATSMEM_SIZE/sizeof(uint64)];
ipm_root_t		*ipmd_stats_root = (ipm_root_t*)stats_buf;


/*-----------------------------------------------------------------
 *
 * Name: ipmd_InitStatsMem
 *
 * Description: init stats memory space.
 *		
 *      ipm_root_t;
 *		ipm_card_t * NFX_NUM_SLOTS_PER_CHASSIS;
 * 		IPM_MAX_CONN_ONCHASSIS*ipm_inconn_t;
 *
 *
 *	for each valid card:
 *	    one or more ipm_ifstat_t
 *		ipm_ipstat_t
 *		ipm_icpstat_t
 *		ipm_tcpstat_t
 *		ipm_udpstat_t
 *
 * Created by:	Su Wang
 *
 * Date Created: 12/11/2001
 *
 *-----------------------------------------------------------------
 */
int32					/* NFX_OK, NFX_ERR */
ipmd_InitStatsMem()
{
	ipm_card_t		*cardaddr;
	int				i;
	ipm_inconn_t	*connaddr;
	char			*tmp_ptr;
  
	/* initialize structures */
	bzero(ipmd_stats_root, IPM_STATSMEM_SIZE);
	
	/* root_t */
	cardaddr = ipmd_stats_root->cardp = 
		(ipm_card_t*)(ipmd_stats_root + 1);


	/* initialize free conn stack, which is shared by all cards */
	ipmd_stats_root->firstconnp = ipmd_stats_root->freeconnp = 
		(ipm_inconn_t*)(cardaddr + NFX_NUM_SLOTS_PER_CHASSIS);
  
	ipmd_stats_root->freeconncnt = IPM_MAX_CONN_ONCHASSIS;
  

  
	/* chain all free conns */
	for (i=0, connaddr = ipmd_stats_root->freeconnp;
		 i < (IPM_MAX_CONN_ONCHASSIS - 1); i++, connaddr++) {
			/* do not setup the last element */
			connaddr->next = connaddr + 1;
	}
	ipmd_stats_root->lastconnp = connaddr;
	

	/* card configurations */
	tmp_ptr = (char*)(ipmd_stats_root->lastconnp + 1);
	cardaddr++;					/* reserve first card for ssc */
  
	cardaddr->card = 1;			/* first txrx */

	cardaddr->ifcnt = IPM_LINK_MAX;
	for (i = 0; i< IPM_LINK_MAX; i++) {
		cardaddr->ifstatp[i] = (ipm_ifstat_t*)tmp_ptr;
		tmp_ptr += sizeof(ipm_ifstat_t);
	}

	cardaddr->ipstatp = (ipm_ipstat_t*)tmp_ptr;
	tmp_ptr += sizeof(ipm_ipstat_t);
								/* ptr beyond ipstat */

	cardaddr->icmpstatp = (ipm_icpstat_t*)tmp_ptr;
	tmp_ptr += sizeof(ipm_icpstat_t);
								/* ptr beyond icmp */

	cardaddr->tcpstatp = (ipm_tcpstat_t*)tmp_ptr;
	tmp_ptr += sizeof(ipm_tcpstat_t);
								/* ptr beyond tcptat */

	cardaddr->udpstatp = (ipm_udpstat_t*)tmp_ptr;
	tmp_ptr += sizeof(ipm_udpstat_t);
								/* ptr beyond udptat */
								/* ready for next valid card */  

//    printf("stats mem initialized total size %x\n", tmp_ptr - (char*)stats_buf);
	/* finish off */
//	ipmd_PrintStatsMem();

	return NFX_OK;
}


/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_PrintStatsMem
 *
 * Description: print root section of ipm shared memory
 *
 * Created by:	Su Wang
 *
 * Date Created: 11/28/2001
 *
 *-----------------------------------------------------------------
 */
void
ipmd_PrintStatsMem()
{
  ipm_card_t	*cardaddr;
  int			i;

  printf("size of structures\n");
  printf("root    :		%d\n", sizeof(ipm_root_t));
  printf("card    :		%d\n", sizeof(ipm_card_t));
  printf("ifstat  :		%d\n", sizeof(ipm_ifstat_t));
  printf("ipstat  :		%d\n", sizeof(ipm_ipstat_t));		
  printf("icmpstat:		%d\n", sizeof(ipm_icpstat_t));
  printf("tcpstat :		%d\n", sizeof(ipm_tcpstat_t));
  printf("udpstat :		%d\n", sizeof(ipm_udpstat_t));
  printf("conn    :		%d\n", sizeof(ipm_inconn_t));
  printf("\n");

  printf("mem start:	%x\n", (uint32)ipmd_stats_root);
  printf("card array:	%x\n", (uint32)ipmd_stats_root->cardp);
  printf("fconn     :	%x\n", (uint32)ipmd_stats_root->freeconnp);
  printf("tcpconn	:	%x\n", (uint32)ipmd_stats_root->tcpconnp);
  printf("udpconn	:	%x\n", (uint32)ipmd_stats_root->udpconnp);
  printf("tcp conn #:	%d		udp con #:   %d\n", 
		 ipmd_stats_root->tcpcnt, ipmd_stats_root->udpcnt);
  printf("\n\n");

  printf("stats ptr per card \n");
  printf("--------------------------------------------------------\n");
  printf("card	ifcnt	ifstatp	ipstatp	icmpsp	tcpstp	udpstp	\n");
  printf("--------------------------------------------------------\n");
  cardaddr = ipmd_stats_root->cardp;
  for (i=0; i<NFX_NUM_SLOTS_PER_CHASSIS; i++)
	{
	  printf("%-8x%-8x%-8x%-8x%-8x%-8x%-8x\n",
			 cardaddr->card,
			 cardaddr->ifcnt,
			 (uint32)cardaddr->ifstatp[0],
			 (uint32)cardaddr->ipstatp,
			 (uint32)cardaddr->icmpstatp,
			 (uint32)cardaddr->tcpstatp,
			 (uint32)cardaddr->udpstatp);
	  cardaddr++;
	}
}



/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_ChkStatsMem
 *
 * Description: validate stats memory.
 *
 * Created by:	Su Wang
 *
 * Date Created: 12/13/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK, NFX_ERR */
ipmd_ChkStatsMem(char	*result)
{
  ipm_inconn_t	*udpconnaddr, *tcpconnaddr, *freeconnaddr;
  int			i, count;

  // ipmd_PrintStatsMem();
  
  /* validate free connection */
  count = 0;
  for (i=0, freeconnaddr = ipmd_stats_root->freeconnp;
	   i< ipmd_stats_root->freeconncnt; i++) {

	  /* list too short */
	  if (!freeconnaddr) {
		  sprintf(result, "free conn. list and cnt does not match, count %d, list head %x, actual cnt %d\n",
				 ipmd_stats_root->freeconncnt, (uint32)ipmd_stats_root->freeconnp, count);
		  return NFX_ERR;
	  }

	  if (freeconnaddr->flags & IPM_NODE_INUSE) {
		  sprintf(result, "free conn. node in use, connp %x\n", (uint32)freeconnaddr);
		  return NFX_ERR;
	  }

	  freeconnaddr = freeconnaddr->next;
	  count++;
  }
  
  /* list too long */
  if (freeconnaddr) {
	  sprintf(result, "free conn. list and cnt does not match, count %d, list head %x\n",
			 ipmd_stats_root->freeconncnt, (uint32)ipmd_stats_root->freeconnp);
	  return NFX_ERR;
  }

  /* validate tcp connections */
  count = 0;
  for (i=0, tcpconnaddr = ipmd_stats_root->tcpconnp;
	   i< ipmd_stats_root->tcpcnt; i++) {
	  if (!tcpconnaddr) {
		  sprintf(result, "tcp conn. list and cnt does not match, count %d, list head %x, actul cnt %d\n",
				 ipmd_stats_root->tcpcnt, (uint32)ipmd_stats_root->tcpconnp, count);
		  return NFX_ERR;
	  }

	  if (!(tcpconnaddr->flags & IPM_NODE_INUSE)) {
		  sprintf(result, "tcp conn, node not in use connp %x\n", (uint32)tcpconnaddr);
		  return NFX_ERR;
	  }

	  tcpconnaddr = tcpconnaddr->next;
	  count++;
  }
  
  if (tcpconnaddr) {
	  sprintf(result, "tcp conn. list and cnt does not match, count %d, list head %x\n",
			 ipmd_stats_root->tcpcnt, (uint32)ipmd_stats_root->tcpconnp);
	  return NFX_ERR;
  }

  /* validate udp connections */
  count = 0;
  for (i=0, udpconnaddr = ipmd_stats_root->udpconnp;
	   i< ipmd_stats_root->udpcnt; i++) {
	  
	  if (!udpconnaddr) {
		  sprintf(result, "udp conn. list and cnt does not match, count %d, list head %x, actual cnt %d\n",
				 ipmd_stats_root->udpcnt, (uint32)ipmd_stats_root->udpconnp, count);
		  return NFX_ERR;
	  }

	  if (!(udpconnaddr->flags & IPM_NODE_INUSE)) {
		  sprintf(result, "udpp conn, node not in use connp %x\n", (uint32)udpconnaddr);
		  return NFX_ERR;
		}


	  udpconnaddr =  udpconnaddr->next;
	  count++;
  }
  
  if (udpconnaddr)
	{
	  sprintf(result, "udp conn. list and cnt does not match, count %d, list head %x\n",
			 ipmd_stats_root->udpcnt, (uint32)ipmd_stats_root->udpconnp);
	  return NFX_ERR;
	}

  sprintf(result, "validate statsmem OK\n");
  return NFX_OK;
}


/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_ProcConnMsg
 *
 * Description: process connection update messages
 *
 * Created by:	Su Wang
 *
 * Date Created: 12/12/2001
 *
 *-----------------------------------------------------------------
 */
int32 
ipmd_ProcConnMsg(uint32			mtype, 
				 uint32			sslot,
				 int			count, 
				 ipm_mconn_t	*ipm_connp)
{
  ipm_inconn_t	*connroot;
  int			i;
  uint32		*conns;			/* num. of connections in chain */

  /* which record we want to modify */
  if (mtype == IPM_MSG_UDP_CONN) {
	  connroot = ipmd_stats_root->udpconnp;
	  conns = &(ipmd_stats_root->udpcnt);
  }
  else	{
	  connroot = ipmd_stats_root->tcpconnp;
	  conns = &(ipmd_stats_root->tcpcnt);
  }

  
  /* walk through each msg on list */
  for (i = 0; i < count; ipm_connp++, i++) {
	  if ((ipm_connp->cmd == IPM_ADD) || 
		 (ipm_connp->cmd == IPM_CHGE_STATE)) {
		  if (ipmd_InsConn(&(ipm_connp->conn), &connroot, ipmd_CmpConn)
			  == NFX_OK)
			(*conns)++;
	  }
	  else {
		  if (ipmd_DelConn(&(ipm_connp->conn), &connroot, ipmd_CmpConn)
			  == NFX_OK)
			  (*conns)--;
	  }
  }

  /* update connroot as they might be changed */
  if (mtype == IPM_MSG_UDP_CONN)
	  ipmd_stats_root->udpconnp = connroot;
  else
	  ipmd_stats_root->tcpconnp = connroot;

  return NFX_OK;
}

/* 
 *-----------------------------------------------------------------
 *
 * Name: ipmd_ProcStatMsg
 *
 * Description: cpy rcved msg to shared memory
 *
 * Created by:	Su Wang
 *
 * Date Created: 12/12/2001
 *
 *-----------------------------------------------------------------
 */
int32
ipmd_ProcStatMsg(int mtype, void *msg, uint32 sslot)
{
  int			size;
  int			type, slot, link;
  ipm_card_t	*cardaddr;
  void			*srcaddr, *destaddr;

  srcaddr = msg;

  /* which slot this stats update come from */
  cardaddr = (ipmd_stats_root->cardp) + sslot;

  switch(mtype)
	{
	case IPM_MSG_INTF_STATS:
	  {
		ipm_ifstat_t		*ifsp;

		size = sizeof(ipm_ifstat_t);
		ifsp = (ipm_ifstat_t*)msg;
		/* get interface */
		if ((ipmd_ParseIfName(ifsp->ifi_ifname, &type, &slot, &link) != NFX_OK) ||
			(type != IPM_INTF_TYPE_ETH)) {
			IPM_LOGERR(IPM_ERR_STATS);
			return NFX_ERR;
		}
		
		/* find the memory space responding to stats of this interface */
		if (cardaddr->ifcnt <= link)  { /* invalid link num. */
			IPM_LOGERR(IPM_ERR_STATS);
			return NFX_ERR;
		}

		destaddr = cardaddr->ifstatp[link];
	  }
	  break;

	case IPM_MSG_IP_STATS:
	  {
		ipm_ipstat_t		*ipsp;

		size = sizeof(ipm_ipstat_t);
		ipsp = (ipm_ipstat_t*)msg;

		/* cardaddr not initialized, must be a msg */
		if (!cardaddr->ipstatp) {
			IPM_LOGERR(IPM_ERR_STATS);
			return NFX_ERR;
		}

		destaddr = cardaddr->ipstatp;
	  }
	  break;

	  
	case IPM_MSG_ICMP_STATS:
	  {
		ipm_icpstat_t		*icmpsp;

		size = sizeof(ipm_icpstat_t);
		icmpsp = (ipm_icpstat_t*)msg;

		if (!cardaddr->icmpstatp) {
			IPM_LOGERR(IPM_ERR_STATS);
			return NFX_ERR;
		}
		
		destaddr = cardaddr->icmpstatp;
	  }
	  break;

	case IPM_MSG_TCP_STATS:
	  {
		ipm_tcpstat_t		*tcpsp;

		size = sizeof(ipm_tcpstat_t);
		tcpsp = (ipm_tcpstat_t*)msg;

		if (!cardaddr->tcpstatp) {
			IPM_LOGERR(IPM_ERR_STATS);
			return NFX_ERR;
		}

		destaddr = cardaddr->tcpstatp;
	  }
	  break;	  


	case IPM_MSG_UDP_STATS:
	  {
		ipm_udpstat_t		*udpsp;
		
		size = sizeof(ipm_udpstat_t);
		udpsp = (ipm_udpstat_t*)msg;

		if (!cardaddr->udpstatp) {
			IPM_LOGERR(IPM_ERR_STATS);
			return NFX_ERR;
		}
		
		destaddr = cardaddr->udpstatp;
	  }
	  break;

	default:
		IPM_LOGERR(IPM_ERR_STATS);
		return NFX_ERR;
	}

  /* destaddr, srcaddr already setup */
  memcpy(destaddr, srcaddr, size);
  return NFX_OK;
}

/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetTxRxIfStats
 *
 * Description: get stats associated with a device. 
 *				device format: [name]:[slot]/[link] i.e."eth:1/1" 
 *				if device is NULL, cpy all ifstats in
 *				buff in consecutive ipm_ifstat_t.
 *
 * Created by:	Su Wang
 *
 * Date Created: 12/12/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR, IPM_GET_NO_BUF */
ipmd_GetTxRxIfStats(char*			ifname,	
					ipm_ifstat_t	*ifstat, /* usr provided buf */
					int				bufsize, /* buffer size */
					int				*count,	/* number of record in ifstat */
					int				*needed) /* buffer size needed */
	
{
	
  ipm_card_t	*cardaddr;
  ipm_ifstat_t	*ifsp;
  int			rc = IPM_GET_OK;
  int			type, slot, link;
  
  

	
  if (ifname) {					/* single device */

		if (ipmd_ParseIfName(ifname, &type, &slot, &link)) {
			rc = IPM_GET_ERR;
			goto out;
		}

		if (bufsize < sizeof(ipm_ifstat_t)) {
		  rc = IPM_GET_NO_BUF;
		  *needed = sizeof(ipm_ifstat_t);
		  goto out;
		}
	  
		cardaddr = (ipmd_stats_root->cardp) + slot;
	  
		/* invalid link */
		if ((!cardaddr->ifcnt) || (cardaddr->ifcnt <= link)) {
			rc = IPM_GET_ERR;
			goto out;
		}
		
		ifsp = cardaddr->ifstatp[link];
		/* stats have been received */
		if (ifsp->ifi_ifname[0] == '\0')
			goto out;

		memcpy(ifstat, ifsp, sizeof(ipm_ifstat_t));
		(*count)++;
		goto out;
  }
  else {							/* get all devices */
	
	  for (slot = 0, cardaddr = ipmd_stats_root->cardp; 
		   slot <NFX_NUM_SLOTS_PER_CHASSIS; slot++, cardaddr++) {
		  
		  /* invalid card */
		  if (cardaddr->ifcnt == 0)
			continue;

		  for (link = 0; link < cardaddr->ifcnt; link++) {
			  ifsp = cardaddr->ifstatp[link];


			  /* make sure update have been received */
			  if (ifsp->ifi_ifname[0] == '\0')
				  continue;
			  
			  if (bufsize < sizeof(ipm_ifstat_t)) {
				  rc = IPM_GET_NO_BUF;
				  /* guess needed space */
				  *needed = IPM_LINK_MAX*
					  NFX_NUM_SLOTS_PER_CHASSIS*sizeof(ipm_ifstat_t);
				  goto out;
			  }
			  memcpy(ifstat, ifsp, sizeof(ipm_ifstat_t));
			  bufsize -= sizeof(ipm_ifstat_t);
			  ifstat++;
			  (*count)++;
			}
	  }
  }

 out:
  return rc;
}

/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetTxRxCardStats
 *
 * Description: core function to get various card stats
 *				if card is -1, cpy ip stats of all cards in
 *				buff in consecutive ipm_ipstat_t.
 *
 * params:
 *		cardoff:  field ptr offset to start of ipm_shmcard_t  
 *		fieldsize: size of requested stats structure 
 *
 * Created by:	Su Wang
 *
 * Date Created: 12/12/2001
 *
 *-----------------------------------------------------------------
 */
static int32					/* IPM_GET_OK, IPM_GET_ERR, IPM_GET_NO_BUF */
ipmd_GetTxRxCardStats(int			cardoff,
					  int			fieldsize,
					  int32			card,	
					  char			*buf, 
					  int			bufsize, /* buffer size */
					  int			*count,	/* number of record in ipstat */
					  int			*needed	/* buff size needed */
				)
{
  ipm_card_t	*cardaddr;
  int			slot;
  void			*field;
  int			rc = IPM_GET_OK;


  for (slot = 0; slot <NFX_NUM_SLOTS_PER_CHASSIS; slot++) {

	  /* card does not matching */
	  if ((card != -1) && (card != slot))
		  continue;

	  cardaddr = (ipmd_stats_root->cardp)+slot;

	  /* get ptr to stats structure */
	  field = *(void**)(((char*)cardaddr) + cardoff);
	  
	  /* invalid card */
	  if (!field)
		  continue;
	  
	  /* no update rcved */
	  if (*(uint32*)field == 0)
		  continue;
	  
	  /* enough given buf */
	  if (bufsize < fieldsize) {
		  rc = IPM_GET_NO_BUF;
		  *needed = NFX_NUM_SLOTS_PER_CHASSIS * fieldsize;
		  goto out;
	  }
	  
	  memcpy(buf, field, fieldsize);
	  bufsize -= fieldsize;
	  buf = buf + fieldsize;
	  (*count)++;
	}

 out:
  return rc;
}


/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetTxRxIpStats
 *
 * Description: get ip stats  associated with a card. 
 *				if card is -1, cpy ip stats of all cards in
 *				buff in consecutive ipm_ipstat_t.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/23/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR, IPM_GET_NO_BUF */
ipmd_GetTxRxIpStats(int32			card,	
					void			*ipstat, 
					int				bufsize, /* buffer size */
					int				*count,	/* number of record in ipstat */
					int				*needed) /* buff size needed */

{
	
	return ipmd_GetTxRxCardStats(CARD_IPS_OFFSET, sizeof(ipm_ipstat_t),
								 card, (char*)ipstat, bufsize, count, needed);
}

/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetTxRxIcmpStats
 *
 * Description: get icmp stats  associated with a card. 
 *				if card is -1, cpy ip stats of all cards in
 *				buff in consecutive ipm_icpstat_t.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/23/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR, IPM_GET_NO_BUF */
ipmd_GetTxRxIcmpStats(int32				card,	
					  void				*icmpstat, 
					  int				bufsize, /* buffer size */
					  int				*count,	/* number of record */
					  int			    *needed) /* buff size needed */
{

	return ipmd_GetTxRxCardStats(CARD_ICMPS_OFFSET, sizeof(ipm_icpstat_t),
								 card, (char*)icmpstat, bufsize, count, needed);
}


/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetTxRxTcpStats
 *
 * Description: get Tcp stats  associated with a card. 
 *				if card is -1, cpy ip stats of all cards in
 *				buff in consecutive ipm_tcpstat_t.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/23/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR, IPM_GET_NO_BUF */
ipmd_GetTxRxTcpStats(int32			card,	
					 void			*tcpstat, 
					 int			bufsize,
					 int			*count,/* number of record */
					 int			*needed)	/* buff size needed */
				
{
	return ipmd_GetTxRxCardStats(CARD_TCPS_OFFSET, sizeof(ipm_tcpstat_t),
								 card, (char*)tcpstat, bufsize, count, needed); 
}


/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetTxRxUdpStats
 *
 * Description: get udp stats  associated with a card. 
 *				if card is -1, cpy ip stats of all cards in
 *				buff in consecutive ipm_udpstat_t.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/23/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR, IPM_GET_NO_BUF */
ipmd_GetTxRxUdpStats(int32			card,	
					 void			*udpstat, 
					 int			bufsize,
					 int			*count,/* number of record */
					 int			*needed)	/* buff size needed */
				
{
	return ipmd_GetTxRxCardStats(CARD_UDPS_OFFSET, sizeof(ipm_udpstat_t),
								 card, (char*)udpstat, bufsize, count, needed); 
}


/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetTxRxConnTable
 *
 * Description: core function to get system wide connection table.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/23/2001
 *
 *-----------------------------------------------------------------
 */
static int32					/* IPM_GET_OK, IPM_GET_ERR, IPM_GET_NO_BUF */
ipmd_GetTxRxConnTable(int			proto, /* tcp/udp */
					  ipm_conn_t	*connp,	/* buffer provided by user */
					  int			bufsize, /* buffer size */
					  int			index, /* where to start */
					  int			incount, /* how many user desired, -1 until end */
					  int			*count,/* number of record */
					  int			*needed) /* buff size needed */
{
  ipm_inconn_t	*connroot;
  int			rc = IPM_GET_OK, conn_cnt;
  
  /* which connection table */
  if (proto == IPM_IP_PROTO_TCP) {
	  conn_cnt = ipmd_stats_root->tcpcnt;
	  connroot = ipmd_stats_root->tcpconnp;
  }
  else	{
	  conn_cnt = ipmd_stats_root->udpcnt;
	  connroot = ipmd_stats_root->udpconnp;
  }

  /* advance to desired localtion */
  while ((index > 0)) {
	  
	  /* skip ones that we don't need, and index less
	   * than chain length
	   */
	  index--;
	  connroot = connroot->next;
	
	  /* reaching end of chain */
	  if (--conn_cnt <= 0) {
		  rc = IPM_GET_ERR;
		  goto out;
	  }
  }
  
  /* determine how many records need to 
   * copied 
   */
  if ((incount > 0) && (conn_cnt > incount))
	  conn_cnt = incount;
  
  /* make sure buffer size is sufficient */
  if (bufsize < conn_cnt * sizeof(ipm_conn_t)) {
	  *needed = conn_cnt * sizeof(ipm_conn_t);
	  *count = conn_cnt;
	  rc = IPM_GET_NO_BUF;
	  goto out;
  }
  else	{						/* cpy to buf provided */
	  wi_t	wi;
	  
	  wi.cp_ptr = (void*)connp;
	  wi.cp_cnt = conn_cnt;
	  
	  ipmd_WalkConn(connroot, ipmd_CpyConnToLocal, &wi);
	  *count = conn_cnt;
  }

out:
  return rc;
}


/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetTxRxTcpConnTable
 *
 * Description: get system wide TCP connection table.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/23/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR, IPM_GET_NO_BUF */
ipmd_GetTxRxTcpConnTable(ipm_conn_t		*tcpconn,
						 int			bufsize,
						 int			index, /* where to start */
						 int			incount, /* how many user desired */
						 int			*count,/* number of record */
						 int			*needed) /* buff size needed */
					 
{
  return ipmd_GetTxRxConnTable(IPM_IP_PROTO_TCP,
							   tcpconn, bufsize, index, incount, count, needed);
}


/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetTxRxUDPConnTable
 *
 * Description: get system wide UDP connection table.
 *
 * Created by:	Su Wang
 *
 * Date Created: 10/23/2001
 *
 *-----------------------------------------------------------------
 */
int32							/* IPM_GET_OK, IPM_GET_ERR, IPM_GET_NO_BUF */
ipmd_GetTxRxUdpConnTable(ipm_conn_t		*udpconn,
						 int			bufsize,
						 int			index, /* where to start */
						 int			incount, /* how many user desired */
						 int			*count,/* number of record */
						 int			*needed	/* buff size needed */
					 )
{
  return ipmd_GetTxRxConnTable(IPM_IP_PROTO_UDP,
						   udpconn, bufsize, index, incount, count, needed);
}

/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetTxRxTcpConn
 *
 * Description: find a TCP connection in record.
 *
 * Created by:	Su Wang
 *
 * Date Created: 01/07/2002
 *
 *-----------------------------------------------------------------
 */
int32						/* IPM_GET_OK, IPM_GET_ERR */
ipmd_GetTxRxTcpConn(ipm_conn_t*	connp, ipm_conn_t* result)
{
  int32			rc;
  ipm_inconn_t	*in_connp = 0;
  

  /* find element */
  in_connp = ipmd_FindConn(connp, ipmd_stats_root->tcpconnp, ipmd_CmpConn);
  
  if (in_connp) {
	  memcpy(result, &(in_connp->conn), sizeof(ipm_conn_t));
	  rc = IPM_GET_OK;
  }
  else
	  rc = IPM_GET_ERR;
  
  return rc;
}

/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetTxRxUdpConn
 *
 * Description: find a UDP connection in record.
 *
 * Created by:	Su Wang
 *
 * Date Created: 01/07/2002
 *
 *-----------------------------------------------------------------
 */
int32						/* IPM_GET_OK, IPM_GET_ERR */
ipmd_GetTxRxUdpConn(ipm_conn_t*	connp, ipm_conn_t* result)
{
  int32			rc;
  ipm_inconn_t	*in_connp = 0;
  

  /* find element */
  in_connp = ipmd_FindConn(connp, ipmd_stats_root->udpconnp, ipmd_CmpConn);
  
  if (in_connp) {
	  memcpy(result, &(in_connp->conn), sizeof(ipm_conn_t));
	  rc = IPM_GET_OK;
  }
  else
	  rc = IPM_GET_ERR;
  
  return rc;
}

/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetTxRxTcpNxtConn
 *
 * Description: find next connection in TCP record.
 *
 * Created by:	Su Wang
 *
 * Date Created: 01/07/2002
 *
 *-----------------------------------------------------------------
 */
int32						/* IPM_GET_OK, IPM_GET_ERR */
ipmd_GetTxRxTcpNxtConn(ipm_conn_t*	connp, ipm_conn_t* result)
{
  int32			rc;
  ipm_inconn_t	*in_connp = 0;
  
  if (!connp)	
	  /* Null ptr, give first record */
	  in_connp = ipmd_stats_root->tcpconnp;
  else {
	  /* find element */
	  in_connp = ipmd_FindConn(connp, ipmd_stats_root->tcpconnp, ipmd_CmpConn);
	  /* advance to next record */
	  in_connp = in_connp->next;
  }
  
  
  if (in_connp) {
	  memcpy(result, &(in_connp->conn), sizeof(ipm_conn_t));
	  rc = IPM_GET_OK;
  }
  else
	  rc = IPM_GET_ERR;
  
  return rc;
}

/*-----------------------------------------------------------------
 *
 * Name: ipmd_GetTxRxUdpNxtConn
 *
 * Description: find next connection in Udp record.
 *
 * Created by:	Su Wang
 *
 * Date Created: 01/07/2002
 *
 *-----------------------------------------------------------------
 */
int32						/* IPM_GET_OK, IPM_GET_ERR */
ipmd_GetTxRxUdpNxtConn(ipm_conn_t*	connp, ipm_conn_t* result)
{
  int32			rc;
  ipm_inconn_t	*in_connp = 0;
  

  if (!connp)					
	  /* Null ptr, give first record */
	  in_connp = ipmd_stats_root->udpconnp;
  else {
	  /* find element */
	  in_connp = ipmd_FindConn(connp, ipmd_stats_root->udpconnp, ipmd_CmpConn);
  	  /* advance to next record */
	  in_connp = in_connp->next;
  }
  
  if (in_connp) {
	  memcpy(result, &(in_connp->conn), sizeof(ipm_conn_t));
	  rc = IPM_GET_OK;
  }
  else
	  rc = IPM_GET_ERR;
  
  return rc;
}
