/*
 *------------------------------------------------------------------------------
 *
 * Name:	ispfc_aen_handler.c
 *
 * RCS:		$Id: ispfc_aen_handler.c,v 1.9 2002/10/25 22:24:36 raster Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	QLogic 2300 FC HAL driver AEN handler code
 *
 * Created by:	Radek Aster
 *
 * Date Created: 6/1/01
 *
 *------------------------------------------------------------------------------
 */

#include "ispfc_private.h"
#include "ispfc_aen_handler.h"
#include "ispfc_mbox.h"
#include "../sm-eee/eee-test.h"

void ispfc_default_aen_handler(ispfc_cb_t *, uint32);

/* SYSDVT */
extern sysdvt_qlogic_t  *qdvt;

/*******************************************************************************
 * ispfc_register_aen_handler -
 *
 *   Register a handler for a particular AEN code. There can only be one
 *   per code.
 ******************************************************************************/
void ispfc_register_aen_handler(ispfc_cb_t *cb,
				uint32 aen_code,
				void (*aen_handler)(ispfc_cb_t *, void *, uint32),
				void * aen_cookie,
				uint32 aen_arg)
{
	ispfc_aen_handler_t *	aen_h = (ispfc_aen_handler_t *)cb->aen_handler[aen_code & 0xFF];

	VERIFY(aen_h->valid == 0);
	aen_h->aen_code = aen_code;
	aen_h->aen_handler = aen_handler;
	aen_h->aen_cookie = aen_cookie;
	aen_h->aen_arg = aen_arg;
	++aen_h->valid;
}

void ispfc_deregister_aen_handler(ispfc_cb_t *cb,
				  uint32 aen_code)
{
	ispfc_aen_handler_t *	aen_h = (ispfc_aen_handler_t *)cb->aen_handler[aen_code & 0xFF];

	VERIFY(aen_h->valid != 0);

	aen_h->aen_code = 0;
	aen_h->aen_handler = NULL;
	aen_h->aen_cookie = NULL;
	aen_h->aen_arg = 0;
	aen_h->valid = 0;
}
				
void ispfc_reset_aen_handlers(ispfc_cb_t *cb)
{
	int i;

	for (i = 0; i < sizeof(cb->aen_handler)/sizeof(void *); ++i) {
		bzero(cb->aen_handler[i], sizeof(ispfc_aen_handler_t));
	}
}

void ispfc_aen_handler(ispfc_cb_t *cb, uint32 aen_code)
{
	ispfc_aen_handler_t *	aen_h;

	VERIFY((aen_code & ~ISPFC_CS_AEN_STATUS) < AEN_HANDLERS);

	FC_DBG(1, ("%s: ispfc_aen_handler: aen_code 0x%x\n", 
		   cb->debug_name, aen_code));

	aen_h = (ispfc_aen_handler_t *)cb->aen_handler[aen_code & 0xFF];
	if (aen_h->valid) 
		(*aen_h->aen_handler)(cb, aen_h->aen_cookie, aen_h->aen_arg);
	else
		ispfc_default_aen_handler(cb, aen_code);
}

void ispfc_default_aen_handler(ispfc_cb_t *cb, uint32 aen_code)
{
	switch(aen_code) {
	case ISPFC_CS_SYSTEM_ERROR:	/* Fatal firmware error */
	case ISPFC_CS_REQ_XFR_ERROR:	/* Error while DMAing from request queue */
	case ISPFC_CS_RSP_XFR_ERROR:	/* Error while DMAing to response queue */
		FC_DBG(0, ("%s: Error: ispfc_default_aen_handler: aen_code 0x%x\n", cb->debug_name, aen_code));
		qdvt->qlogic_FatalErr ++;
		SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
		break;

	case ISPFC_CS_LIP_OCCURRED:	/* LIP happened */
	case ISPFC_CS_LIP_RESET:	/* LIP-reset happened */
	case ISPFC_CS_LIP_F8_RECVD:	/* LIP-F8 happened */
		if ((LINK_STATE(cb) == ISPFC_CLS_STATE_UP) ||
		    (LINK_STATE(cb) == ISPFC_CLS_STATE_LIP_OCCURRED) ||
		    (LINK_STATE(cb) == ISPFC_CLS_STATE_LIP_GENERATED && aen_code == ISPFC_CS_LIP_OCCURRED)) {
			SET_LINK_STATE(cb, ISPFC_CLS_STATE_LIP_OCCURRED);
		}
		else
		if (LINK_STATE(cb) == ISPFC_CLS_STATE_REESTABLISHING)
			cb->link_state |= ISPFC_CLF_NESTED_LIP;
		break;

	case ISPFC_CS_LIP_FAILURE:
		FC_DBG(0, ("%s: ispfc_aen_handler: LIP failure detected -- 0x%x\n", 
			   cb->debug_name, ISPFC_REG_READ(cb, ISPFC_REG_MBOX1)));
		break;
		    
	case ISPFC_CS_PNT_TO_PNT_CONNECTION: /* Link connected in point-to-point mode */
		cb->link_state |= ISPFC_CLF_POINT_TO_POINT_OK;

	case ISPFC_CS_NSDB_CHANGED:	/* NameServer State-Change-Notification received */
		if ((LINK_STATE(cb) == ISPFC_CLS_STATE_UP) ||
		    (LINK_STATE(cb) == ISPFC_CLS_STATE_NSCN_OCCURRED) ||
		    (LINK_STATE(cb) == ISPFC_CLS_STATE_NSCN_GENERATED && aen_code == ISPFC_CS_NSDB_CHANGED)) {
			SET_LINK_STATE(cb, ISPFC_CLS_STATE_NSCN_OCCURRED);
		}
		else
		if (LINK_STATE(cb) == ISPFC_CLS_STATE_REESTABLISHING)
			cb->link_state |= ISPFC_CLF_NESTED_NSCN;
		break;

	case ISPFC_CS_LOOP_UP:		/* Loop/Link transitioned from down to up */
		if (LINK_STATE(cb) == ISPFC_CLS_STATE_DOWN) {
			FC_DBG(0, ("%s: ispfc_default_aen_handler: Loop going from down to up\n", cb->debug_name));
			SET_LINK_STATE(cb, ISPFC_CLS_STATE_RESET_REQUESTED);
		}	
		break;

	case ISPFC_CS_LOOP_DOWN:	/* Loop/Link transitioned from up to down */
		if (LINK_STATE(cb) == ISPFC_CLS_STATE_UP)
			SET_LINK_STATE(cb, ISPFC_CLS_STATE_DOWN);
		break;

	default:
		VERIFY(0);
	}
}

