/*
 *------------------------------------------------------------------------------
 *
 * Name:	ispfc_scsi.c
 *
 * RCS:		$Id: ispfc_scsi.c,v 1.49 2002/10/25 22:24:36 raster Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	QLogic 2300 FC HAL driver SCSI command transport code
 *
 * Created by:	Radek Aster
 *
 * Date Created: 6/5/01
 *
 *------------------------------------------------------------------------------
 */

#if defined(LINUX_TEST)
#include <stdlib.h>
#endif

#include "ispfc_private.h"
#include "ispfc_iocb.h"
#include "ispfc_scsi.h"
#include "ispfc_mbox.h"
#include "ispfc_ct.h"
#include "ispfc_device.h"
#include "../sm-scsi/scsi-dev-cmds.h"

#if defined(ISPFC_SIM)
char *simdir = NULL;

void ispfc_sim_scsi_execute(ispfc_cb_t *, ispfc_dev_t *, scsi_desc_t *);
void ispfc_sim_scsi_execute_disk_inquiry(ispfc_cb_t *, ispfc_dev_t *, scsi_desc_t *);
void ispfc_sim_scsi_execute_disk_read_capacity(ispfc_cb_t *, ispfc_dev_t *, scsi_desc_t *);
void ispfc_sim_scsi_execute_disk_read(ispfc_cb_t *, ispfc_dev_t *, scsi_desc_t *);
void ispfc_sim_scsi_execute_disk_write(ispfc_cb_t *, ispfc_dev_t *, scsi_desc_t *);
#if (ISPFC_SIM == ISPFC_SIM_LINUX_FILE)
void ispfc_sim_scsi_execute_tape_inquiry(ispfc_cb_t *, ispfc_dev_t *, scsi_desc_t *);
void ispfc_sim_scsi_execute_tape_mode_sense(ispfc_cb_t *, ispfc_dev_t *, scsi_desc_t *);
void ispfc_sim_scsi_execute_tape_mode_select(ispfc_cb_t *, ispfc_dev_t *, scsi_desc_t *);
void ispfc_sim_scsi_execute_tape_read(ispfc_cb_t *, ispfc_dev_t *, scsi_desc_t *);
void ispfc_sim_scsi_execute_tape_write(ispfc_cb_t *, ispfc_dev_t *, scsi_desc_t *);
void ispfc_sim_scsi_execute_tape_rewind(ispfc_cb_t *, ispfc_dev_t *, scsi_desc_t *);
void ispfc_sim_scsi_execute_tape_load_unload(ispfc_cb_t *, ispfc_dev_t *, scsi_desc_t *);
void ispfc_sim_scsi_execute_tape_write_fm(ispfc_cb_t *, ispfc_dev_t *, scsi_desc_t *);
void ispfc_sim_scsi_execute_tape_space(ispfc_cb_t *, ispfc_dev_t *, scsi_desc_t *);
void ispfc_sim_scsi_execute_tape_erase(ispfc_cb_t *, ispfc_dev_t *, scsi_desc_t *);
void ispfc_sim_scsi_execute_tape_read_position(ispfc_cb_t *, ispfc_dev_t *, scsi_desc_t *);

void ispfc_sim_tape_directory_flush_to_file(ispfc_dev_t *, int);
void ispfc_sim_tape_directory_sync_to_file(ispfc_dev_t *);
void ispfc_sim_tape_add_directory_entry(ispfc_dev_t *, int);
void ispfc_sim_tape_erase_to_eod(ispfc_dev_t *);
uint32 ispfc_sim_tape_bytecount_at_eod(ispfc_cb_t *, ispfc_dev_t *);
void ispfc_sim_tape_rewind(ispfc_dev_t *);
#endif /* ISPFC_SIM == ISPFC_SIM_LINUX_FILE */
#endif /* ISPFC_SIM */

int32 
ispfc_scsi_send(scsi_desc_t *sd)
{
	ispfc_dev_t *	dev;
	ispfc_cb_t *	cb;

	cb = hal_dt_entry_get()->hal_handle;
	VERIFY(cb);

	/* 
	 * Cleanse return status 
	 */
	SD_SET_DRV_STATUS(sd, 0);
	SD_SET_DEV_STATUS(sd, 0);
	sd->status &= ~SD_DEV_SENSE_VALID;

	if (cb->dev_count == 0 ||
	    (dev = ispfc_hash_table_obj_find(cb->dev_handle_hash, sd->halHandle)) == NULL || 
	    DEV_STATE(dev) == ISPFC_DS_STATE_INVALID) {
		SD_SET_DRV_STATUS(sd, SD_DRV_STATUS_INVALID_DEVICE);
		SCSI_RECEIVE(sd);
		return(NFX_ERR);
	}

	/* If express and link not up, immediately kill I/O */
	if ((sd->control & SD_FLAG_EXPRESS) && !IS_LINK_UP(cb)) {
		SD_SET_DRV_STATUS(sd, SD_DRV_STATUS_LINK_DOWN);
		SCSI_RECEIVE(sd);
		return(NFX_ERR);
	}
	
	sd->state = SCSI_DESC_STATE_WAITING;
	eee_rtc(&sd->first_issued_timestamp);
	sd->current_to_retry_count = 0;

	/* 
	 * Stick descriptor on device's waiting queue; it will be
	 * picked up and dispatched by the device management handler.  
	 */
	ispfc_queue_desc_on_dev_waiting_queue(cb, dev, sd);

	return(NFX_OK);
}


int32
ispfc_scsi_dispatch_desc(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
	scsi_desc_t *	sdc;

	sd->current_timeout_count = sd->timeOutCnt;

	cb->adap_frames_in++;		/* RSP frame */
	cb->adap_frames_out++;		/* CDB frame */

	switch(sd->cdb[0]) {
	case SCSI_CMD_READ_6:
	case SCSI_CMD_READ_10:
		for (sdc = sd; sdc; sdc = sdc->chain) {
			uint32	byte_count;
			uint32	frame_count;

			VERIFY(sdc->status & SD_DEV_DATA_IN_EXPECT);
			VERIFY(sdc->expectDataLen > 0);
			if (sdc->eDesc == NULL)
				sdc->eDesc = ispfc_edesc_allocate(sdc);
			VERIFY(sdc->eDesc);
			ispfc_edesc_set_len(sdc->eDesc, sdc->expectDataLen);
			
			if (sdc->cdb[0] == SCSI_CMD_READ_6) {
				cb->adap_read_6_cmd_cnt++;
				byte_count = 
					(sdc->cdb[2] << 16) | 
					(sdc->cdb[3] <<  8) | 
					sdc->cdb[4];
			}
			else {
				cb->adap_read_10_cmd_cnt++;
				byte_count = 
					(sdc->cdb[7] << 8) | 
					 sdc->cdb[8];
				byte_count *= 
					(sdc->control & SD_FLAG_EMBEDDED_CS) ? BLOCKSIZE_520 : BLOCKSIZE_512;
			}
			frame_count = byte_count / cb->fw_init_cb->max_frame_len;
			cb->adap_frames_in += frame_count;
		} /* for (sdc) */

		break;

	case SCSI_CMD_WRITE_6:
	case SCSI_CMD_WRITE_10:
		for (sdc = sd; sdc; sdc = sdc->chain) {
			uint32	byte_count;
			uint32	frame_count;

			VERIFY(sdc->status & SD_DEV_DATA_OUT_EXPECT);
			VERIFY(sdc->eDesc);

			if (sdc->cdb[0] == SCSI_CMD_WRITE_6) {
				cb->adap_write_6_cmd_cnt++;
				byte_count = 
					(sdc->cdb[2] << 16) | 
					(sdc->cdb[3] <<  8) | 
					sdc->cdb[4];
			}
			else {
				cb->adap_write_10_cmd_cnt++;
				byte_count = 
					(sdc->cdb[7] << 8) | 
					 sdc->cdb[8];
				byte_count *= 
					(sdc->control & SD_FLAG_EMBEDDED_CS) ? BLOCKSIZE_520 : BLOCKSIZE_512;
			}
			frame_count = byte_count / cb->fw_init_cb->max_frame_len;
			cb->adap_frames_out += frame_count;
		}
		break;

	default:
		if (sd->status & SD_DEV_DATA_IN_EXPECT) {
			VERIFY(sd->expectDataLen > 0);
			if (sd->eDesc == NULL)
				sd->eDesc = ispfc_edesc_allocate(sd);
			VERIFY(sd->eDesc);
			ispfc_edesc_set_len(sd->eDesc, sd->expectDataLen);
			cb->adap_frames_in++;
		}
		else
		if (sd->status & SD_DEV_DATA_OUT_EXPECT) {
			VERIFY(sd->eDesc);
			cb->adap_frames_out++;
		}
		cb->adap_other_cmd_cnt++;
		break;
	}

#ifdef DEBUG
	if (ispfc_dbg_level >= 5) {
		for (sdc = sd; sdc; sdc = sdc->chain) {
			if (sdc == sd)
				P("ispfc_scsi_dispatch_desc: (sd %p)\n", sdc);
			else
				P("                          (sd %p)\n", sdc);

			scsi_displayDesc(sdc);
		} /* for (sdc) */
	}
#endif

#if !defined(ISPFC_SIM)
	if (ispfc_send_scsi_desc(cb, dev, sd) != NFX_OK)
		return(NFX_ERR);
#else
	{
		eeeGMT_t delta;

		/* Simulate the response occuring in up to 131 ms. */
		eee_rtc(&sd->sim_response_timestamp);
		eee_rtcToTV(&sd->sim_response_timestamp, &delta);
		delta.usec += delta.usec & 0x1FFFF;
		if (delta.usec > USEC_PER_SEC) {
			delta.usec -= USEC_PER_SEC;
			delta.sec += 1;
		}
		eee_tvToRTC(&delta, &sd->sim_response_timestamp);
	}
	ispfc_sim_scsi_execute(cb, dev, sd);
#endif

	ispfc_queue_desc_on_dev_busy_queue(cb, dev, sd);

	return(NFX_OK);
}

void
ispfc_dispose_of_scsi_desc(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd, int retry)
{
	eeeGMT_t delta;

	/* 
	 * Check time-to-live and kill the command if it's exceeded.
	 */

	eee_rtcDurationTV(&sd->first_issued_timestamp, &delta);
	if (delta.sec > sd->maxTimeToLive) {
		FC_DBG(1, ("%s: MaxTimeToLive abort (%p) phys_dev_id 0x%x, handle %d, "
			   "cdb %s, scsi_status 0x%x, retryCnt %d\n",
			   cb->debug_name, sd, dev->phys_dev_id, dev->dev_handle,
			   ispfc_scsi_cdb_to_string(sd->cdb),
			   sd->status,
			   sd->retryCnt));
		SD_SET_DRV_STATUS(sd, SD_DRV_STATUS_ABORTED);
		SCSI_RECEIVE(sd);
		return;
	}

	/*
	 * If express, or 
	 *    not scheduled for retry, or
	 *    retries exhausted 
	 * ....... terminate.
	 */
	if ((sd->control & SD_FLAG_EXPRESS) || !retry || sd->retryCnt == 0) {
		if (sd->control & SD_FLAG_EXPRESS) {
			FC_DBG(1, ("%s: Express abort (%p) phys_dev_id 0x%x, handle %d, "
				   "cdb %s, scsi_status 0x%x, retryCnt %d\n",
				   cb->debug_name, sd, dev->phys_dev_id, dev->dev_handle,
				   ispfc_scsi_cdb_to_string(sd->cdb),
				   sd->status,
				   sd->retryCnt));
			SD_SET_DRV_STATUS(sd, SD_DRV_STATUS_ABORTED);
		}
		SCSI_RECEIVE(sd);
		return;
	}
	    
	/* 
	 * Cleanse return status 
	 */
	SD_SET_DRV_STATUS(sd, 0);
	SD_SET_DEV_STATUS(sd, 0);
	sd->status &= ~SD_DEV_SENSE_VALID;
	    
	/* Stick descriptor back on waiting queue */
	ispfc_queue_desc_on_dev_waiting_queue(cb, dev, sd);
}

#ifdef ISPFC_SIM
#define SIM_SENSE_LEN	16
byte illegal_opcode_sense[SIM_SENSE_LEN] = {
	0xF0, 0x00, 0x05, 0x00, 
	0x00, 0x00, 0x00, 0x0a, 
	0x00, 0x00, 0x00, 0x00, 
	0x20, 0x00, 0x00, 0x00};
byte illegal_lun_sense[SIM_SENSE_LEN] = {
	0xF0, 0x00, 0x05, 0x00, 
	0x00, 0x00, 0x00, 0x0a, 
	0x00, 0x00, 0x00, 0x00, 
	0x25, 0x00, 0x00, 0x00};
byte illegal_sili_sense[SIM_SENSE_LEN] = {
	0xF0, 0x00, 0x40, 0x00, 
	0x00, 0x00, 0x00, 0x0a, 
	0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00};
byte block_mismatch_sense[SIM_SENSE_LEN] = {
	0xF0, 0x00, 0x05, 0x00, 
	0x00, 0x00, 0x00, 0x0a, 
	0x00, 0x00, 0x00, 0x00, 
	0x81, 0x00, 0x00, 0x00};
byte blank_check_sense[SIM_SENSE_LEN] = {
	0xF0, 0x00, 0x08, 0x00, 
	0x00, 0x00, 0x00, 0x0a, 
	0x00, 0x00, 0x00, 0x00, 
	0x00, 0x05, 0x00, 0x00};
byte filemark_check_sense[SIM_SENSE_LEN] = {
	0xF0, 0x00, 0x80, 0x00, 
	0x00, 0x00, 0x00, 0x0a, 
	0x00, 0x00, 0x00, 0x00, 
	0x00, 0x01, 0x00, 0x00};
byte ili_check_sense[SIM_SENSE_LEN] = {
	0xF0, 0x00, 0x20, 0x00, 
	0x00, 0x00, 0x00, 0x0a, 
	0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00};
byte illegal_mode_page_sense[SIM_SENSE_LEN] = {
	0xF0, 0x00, 0x05, 0x00, 
	0x00, 0x00, 0x00, 0x0a, 
	0x00, 0x00, 0x00, 0x00, 
	0x24, 0x00, 0x00, 0x00};
byte peot_sense[SIM_SENSE_LEN] = {
	0xF0, 0x00, 0x4d, 0x00, 
	0x00, 0x00, 0x00, 0x0a, 
	0x00, 0x00, 0x00, 0x00, 
	0x00, 0x02, 0x00, 0x00};
byte leot_sense[SIM_SENSE_LEN] = {
	0xF0, 0x00, 0x40, 0x00, 
	0x00, 0x00, 0x00, 0x0a, 
	0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00};
byte illegal_count_sense[SIM_SENSE_LEN] = {
	0xF0, 0x00, 0x05, 0x00, 
	0x00, 0x00, 0x00, 0x0a, 
	0x00, 0x00, 0x00, 0x00, 
	0x24, 0x00, 0x00, 0x00};
byte illegal_code_sense[SIM_SENSE_LEN] = {
	0xF0, 0x00, 0x05, 0x00, 
	0x00, 0x00, 0x00, 0x0a, 
	0x00, 0x00, 0x00, 0x00, 
	0x24, 0x00, 0x00, 0x00};
byte eod_found_sense[SIM_SENSE_LEN] = {
	0xF0, 0x00, 0x08, 0x00, 
	0x00, 0x00, 0x00, 0x0a, 
	0x00, 0x00, 0x00, 0x00, 
	0x00, 0x05, 0x00, 0x00};
byte pbot_found_sense[SIM_SENSE_LEN] = {
	0xF0, 0x00, 0x40, 0x00, 
	0x00, 0x00, 0x00, 0x0a, 
	0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00};
byte filemark_found_sense[SIM_SENSE_LEN] = {
	0xF0, 0x00, 0x80, 0x00, 
	0x00, 0x00, 0x00, 0x0a, 
	0x00, 0x00, 0x00, 0x00, 
	0x00, 0x01, 0x00, 0x00};
byte invalid_field_in_cdb_sense[SIM_SENSE_LEN] = {
	0xF0, 0x00, 0x05, 0x00, 
	0x00, 0x00, 0x00, 0x0a, 
	0x00, 0x00, 0x00, 0x00, 
	0x24, 0x00, 0x00, 0x00};
byte data_protect_sense[SIM_SENSE_LEN] = {
	0xF0, 0x00, 0x07, 0x00, 
	0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00, 
	0x00, 0x00, 0x00, 0x00};

void 
ispfc_sim_scsi_execute(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
	uint32	lun;

#if (ISPFC_SIM == ISPFC_SIM_LINUX_FILE)
	/* Initialize simdir */
	if (simdir == NULL)
		simdir = getenv("SIMDIR");
#endif

	lun = (sd->path & SCSI_PATH_LUN_MASK) >> SCSI_PATH_LUN_SHIFT;
	if (lun != 0)
		goto unsupported_lun;

	if (dev->sim_dev_type == 0) {
		switch(sd->cdb[0]) {
		case SCSI_CMD_TEST_UNIT_READY:
			break;

		case SCSI_CMD_INQUIRY:
			ispfc_sim_scsi_execute_disk_inquiry(cb, dev, sd);
			break;

		case SCSI_CMD_READ_CAPACITY:
			ispfc_sim_scsi_execute_disk_read_capacity(cb, dev, sd);
			break;

		case SCSI_CMD_READ_6:
		case SCSI_CMD_READ_10:
			ispfc_sim_scsi_execute_disk_read(cb, dev, sd);
			break;

		case SCSI_CMD_WRITE_6:
		case SCSI_CMD_WRITE_10:
			ispfc_sim_scsi_execute_disk_write(cb, dev, sd);
			break;
		
		default:
			FC_DBG(0, ("%s: phys_dev_id 0x%x: Unsupported disk SCSI command %s\n",
				   cb->debug_name, dev->phys_dev_id, ispfc_scsi_cdb_to_string(sd->cdb)));
			goto unsupported_command;
		} 
	}
#if (ISPFC_SIM == ISPFC_SIM_LINUX_FILE)
	else {
		switch(sd->cdb[0]) {
		case SCSI_CMD_TEST_UNIT_READY:
			break;

		case SCSI_CMD_INQUIRY:
			ispfc_sim_scsi_execute_tape_inquiry(cb, dev, sd);
			break;
				
		case SCSI_CMD_MODE_SELECT_6:
			ispfc_sim_scsi_execute_tape_mode_select(cb, dev, sd);
			break;

		case SCSI_CMD_MODE_SENSE_6:
			ispfc_sim_scsi_execute_tape_mode_sense(cb, dev, sd);			
			break;

		case SCSI_CMD_READ_6:
			ispfc_sim_scsi_execute_tape_read(cb, dev, sd);
			break;

		case SCSI_CMD_WRITE_6:
			ispfc_sim_scsi_execute_tape_write(cb, dev, sd);
			break;

		case SCSI_CMD_REWIND_UNIT:
			ispfc_sim_scsi_execute_tape_rewind(cb, dev, sd); 
			break;

		case SCSI_CMD_LOAD_UNLOAD:
			ispfc_sim_scsi_execute_tape_load_unload(cb, dev, sd); 
			break;

		case SCSI_CMD_WRITE_FILEMARKS:
			ispfc_sim_scsi_execute_tape_write_fm(cb, dev, sd);
			break;

		case SCSI_CMD_SPACE:
			ispfc_sim_scsi_execute_tape_space(cb, dev, sd);
			break;

		case SCSI_CMD_ERASE:
			ispfc_sim_scsi_execute_tape_erase(cb, dev, sd);
			break;

		case SCSI_CMD_READ_POSITION:
			ispfc_sim_scsi_execute_tape_read_position(cb, dev, sd);
			break;
			
		default:
			FC_DBG(0, ("%s: phys_dev_id 0x%x: Unsupported tape SCSI command %s\n",
				   cb->debug_name, dev->phys_dev_id, ispfc_scsi_cdb_to_string(sd->cdb)));
                        goto unsupported_command;
		}
	}
#endif /* ISPFC_SIM == ISPFC_SIM_LINUX_FILE */
		
	return;

 unsupported_command:
	SD_SET_DEV_STATUS(sd, SCSI_DEV_STATUS_CHECK);
	sd->senseData = eee_ramAlloc(SIM_SENSE_LEN);
	bzero(sd->senseData, SIM_SENSE_LEN);
	SD_SET_SENSE(sd, illegal_opcode_sense, SIM_SENSE_LEN);

 unsupported_lun:
	SD_SET_DEV_STATUS(sd, SCSI_DEV_STATUS_CHECK);
	sd->senseData = eee_ramAlloc(SIM_SENSE_LEN);
	bzero(sd->senseData, SIM_SENSE_LEN);
	SD_SET_SENSE(sd, illegal_lun_sense, SIM_SENSE_LEN);
}

/* 
 ************************************************
 * DISK COMMANDS 
 ************************************************
 */
void
ispfc_sim_scsi_execute_disk_inquiry(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
	eee_desc_t *	ed = sd->eDesc;
	eee_descBuf_t *	bd = &ed->bd[1];
	byte *		buf = bd->buf;
	byte *		cdb = sd->cdb;
	uint32     	sim_xfer_len;

	if (cdb[1] & SCSI_INQUIRY_EVPD) {
		/*
		 * Handle VPD pages 
		 */
		switch (cdb[2]) {
		case SCSI_INQUIRY_SERIALNO_PAGE_CODE:
			buf[0] = SCSI_DTYPE_DIRECT;
			buf[1] = SCSI_INQUIRY_SERIALNO_PAGE_CODE;
			buf[2] = 0;
			buf[3] = 8;
			memcpy(&buf[4], dev->scsi_serialno, 8);	
			sim_xfer_len = 12;
			break;

		default:
			VERIFY(0);
		}
	}
	else {
		/*
		 * "Standard" inquiry
		 */
		buf[0] = SCSI_DTYPE_DIRECT;
		buf[1] = 0;

		buf[2] = 2;
		buf[3] = 2;
		buf[4] = 0;
		buf[5] = 30;
		buf[6] = (0) ? 0x40 : 0; /* Supports SES */
		buf[6] |= 0x10;		 /* Dual ported device */
		buf[6] |= (0) ? 0x20 : 0x00; /* Port A/B */
		buf[7] = 0x02;
		memcpy(&buf[8], dev->scsi_vendor_id, 8);
		memcpy(&buf[16], dev->scsi_product_id, 16);
		memcpy(&buf[32], dev->scsi_revision, 4);
		sim_xfer_len = 36;
	}
	ispfc_edesc_set_len(ed, sim_xfer_len);
	sd->status |= SD_DEV_DATA_VALID;
}

void
ispfc_sim_scsi_execute_disk_read_capacity(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
	eee_desc_t *		ed = sd->eDesc;
	eee_descBuf_t *		bd = &ed->bd[1];
	byte *			buf = bd->buf;
	byte *			cdb = sd->cdb;
	uint32			sim_xfer_len;
	scsi_read_capacity_t *	cap = (scsi_read_capacity_t *)buf;
	
	cap->last_block = dev->capacity - 1;
	inplace_htonl(&cap->last_block);
	cap->block_length = dev->sector_size;
	inplace_htonl(&cap->block_length);
	sim_xfer_len = 8;

	ispfc_edesc_set_len(ed, sim_xfer_len);
	sd->status |= SD_DEV_DATA_VALID;
}

void
ispfc_sim_scsi_execute_disk_read(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
	scsi_desc_t *		sdc;
	int			i;
	eee_desc_t *		ed;
	eee_descBuf_t *		bd;
	eee_descBuf_t *		cs_bd;
	byte *			data_p;
	byte *			cs_p;
	int32			sim_xfer_len;
	int32			bufcnt;
	int32			bytecnt;
	uint32			lba;
	ushort16		lbc;
	int32			requested_data_len;
	int32			actual_data_len;
	int32			data_in_block;
	int32			data_in_buf;
	int32			cs_blockcnt;
#if (ISPFC_SIM == ISPFC_SIM_RAM_DISK)
	char *			ram_buf_p;
#endif

	if (sd->cdb[0] == SCSI_CMD_READ_6) {
		lba = (sd->cdb[1] << 16) + (sd->cdb[2] << 8) + sd->cdb[3];
		lbc = sd->cdb[4];
	}
	else {
		memcpy(&lba, &sd->cdb[2], 4);
		inplace_ntohl(&lba);
		memcpy(&lbc, &sd->cdb[7], 2);
		inplace_ntohs(&lbc);
	}

	/* 
	 * Check addressing bounds - for now just
	 * panic, later return appropriate SCSI error.
	 */
	if (lba + lbc > dev->capacity) {
		panic("Out of bounds on read from phys_dev 0x%x: lba %d, lbc %d, capacity %d\n",
		      dev->phys_dev_id, lba, lbc, dev->capacity);
	}

	/* "Seek" to start_sector */
#if (ISPFC_SIM == ISPFC_SIM_LINUX_FILE)
	if (lseek(dev->fd, lba * dev->sector_size, SEEK_SET) == -1)
		panic("lseek failed in %s: %s", 
		      "ispfc_sim_scsi_execute_disk_read", strerror(errno));
#elif (ISPFC_SIM == ISPFC_SIM_RAM_DISK)
	ram_buf_p = dev->ram_disk_buf + lba * dev->sector_size;
#endif
	sim_xfer_len = lbc * dev->sector_size;

	for (sdc = sd; sdc; sdc = sdc->chain) {
		if (sdc->cdb[0] == SCSI_CMD_READ_6)
			lbc = sdc->cdb[4];
		else {
			memcpy(&lbc, &sdc->cdb[7], 2);
			inplace_ntohs(&lbc);
		}

		ed = sdc->eDesc;			
		while (ed) {
			bufcnt = (ed->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;
			bytecnt = ed->hdr.control & EDESC_TOTAL_LEN;
			bd = &ed->bd[0];
			if (!(bd->control & EDESC_BD_CMD_IN_USE))
				++bd;
			VERIFY(bd->control & EDESC_BD_CMD_IN_USE);
			VERIFY(bd->buf);
			data_in_buf = bd->len;
			data_p = bd->buf;
			if (sdc->control & SD_FLAG_EMBEDDED_CS) {
				cs_blockcnt = (ed->hdr.control & EDESC_TOTAL_LEN) / (ECSUM_BLOCKSIZE + ECSUM_LEN); 
				cs_bd = bd + bufcnt - 1;
				VERIFY(cs_bd->control & EDESC_BD_CMD_IN_USE);
				VERIFY(cs_bd->buf);
				cs_p = cs_bd->buf + cs_bd->len - cs_blockcnt * ECSUM_LEN;
				data_in_block = ECSUM_BLOCKSIZE;
			}
			else
				data_in_block = bytecnt;

			while (bytecnt) {
				/* Read user data */
				while (data_in_block) {
					requested_data_len = MIN(data_in_block, data_in_buf);
					while (requested_data_len) {
#if (ISPFC_SIM == ISPFC_SIM_LINUX_FILE)
						actual_data_len = read(dev->fd, data_p, requested_data_len);
						if (actual_data_len < 0)
							panic("read failed in %s: %s", 
							      "ispfc_sim_scsi_execute_disk_read", strerror(errno));
#elif (ISPFC_SIM == ISPFC_SIM_RAM_DISK)
						memcpy(data_p, ram_buf_p, requested_data_len);
						ram_buf_p += requested_data_len;
						actual_data_len = requested_data_len;
#endif
						requested_data_len -= actual_data_len;
						sim_xfer_len -= actual_data_len;
						bytecnt -= actual_data_len;
						data_in_block -= actual_data_len;
						data_in_buf -= actual_data_len;
						data_p += actual_data_len;
					}
					if (data_in_buf == 0 && data_in_block > 0) {
						++bd;
						VERIFY(bd->control & EDESC_BD_CMD_IN_USE);
						VERIFY(bd->buf);
						data_in_buf = bd->len;
						data_p = bd->buf;
					}
				}

				if (!(sdc->control & SD_FLAG_EMBEDDED_CS))
					continue;
				else
					data_in_block = ECSUM_BLOCKSIZE;

				/* Read checksum */
				requested_data_len = ECSUM_LEN;
				while (requested_data_len) {
#if (ISPFC_SIM == ISPFC_SIM_LINUX_FILE)
					actual_data_len = read(dev->fd, cs_p, requested_data_len);
					if (actual_data_len < 0)
						panic("read failed in %s: %s",
						      "ispfc_sim_scsi_execute_disk_read", strerror(errno));
#elif (ISPFC_SIM == ISPFC_SIM_RAM_DISK)
					memcpy(cs_p, ram_buf_p, requested_data_len);
					ram_buf_p += requested_data_len;
					actual_data_len = requested_data_len;
#endif
					requested_data_len -= actual_data_len;
					sim_xfer_len -= actual_data_len;
					bytecnt -= actual_data_len;
					cs_p += actual_data_len;
				}
			} /* while (bytecnt) */

			if (ed->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER) {
				ed = (eee_desc_t *)ed->bd[EEE_MAX_BUFS-1].buf;
				VERIFY(ed);
			}
			else 
				ed = NULL;
		} /* while (ed) */
	} /* for (sdc) */
	if (sim_xfer_len > 0)
		panic("Data underflow on read to phys_dev %x.\n", dev->phys_dev_id);
	else
	if (sim_xfer_len < 0)
		panic("Data overflow on read to phys_dev %x.\n", dev->phys_dev_id);
}

void
ispfc_sim_scsi_execute_disk_write(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
	scsi_desc_t *		sdc;
	int			i;
	eee_desc_t *		ed;
	eee_descBuf_t *		bd;
	eee_descBuf_t *		cs_bd;
	byte *			data_p;
	byte *			cs_p;
	int32			sim_xfer_len;
	int32			bufcnt;
	int32			bytecnt;
	uint32			lba;
	ushort16		lbc;
	int32			requested_data_len;
	int32			actual_data_len;
	int32			data_in_block;
	int32			data_in_buf;
	int32			cs_blockcnt;
#if (ISPFC_SIM == ISPFC_SIM_RAM_DISK)
	char *			ram_buf_p;
#endif

	if (sd->cdb[0] == SCSI_CMD_WRITE_6) {
		lba = (sd->cdb[1] << 16) + (sd->cdb[2] << 8) + sd->cdb[3];
		lbc = sd->cdb[4];
	}
	else {
		memcpy(&lba, &sd->cdb[2], 4);
		inplace_ntohl(&lba);
		memcpy(&lbc, &sd->cdb[7], 2);
		inplace_ntohs(&lbc);
	}

	/* 
	 * Check addressing bounds - for now just
	 * panic, later return appropriate SCSI error.
	 */
	if (lba + lbc > dev->capacity) {
		panic("Out of bounds on write to phys_dev 0x%x: lba %d, lbc %d, capacity %d\n",
		      dev->phys_dev_id, lba, lbc, dev->capacity);
	}

	/* "Seek" to start_sector */
#if (ISPFC_SIM == ISPFC_SIM_LINUX_FILE)
	if (lseek(dev->fd, lba * dev->sector_size, SEEK_SET) == -1)
		panic("lseek failed in %s: %s", 
		      "ispfc_sim_scsi_execute_disk_write", strerror(errno));
#elif (ISPFC_SIM == ISPFC_SIM_RAM_DISK)
	ram_buf_p = dev->ram_disk_buf + lba * dev->sector_size;
#endif
	sim_xfer_len = lbc * dev->sector_size;

	for (sdc = sd; sdc; sdc = sdc->chain) {
		ed = sdc->eDesc;
		while (ed) {
			bufcnt = (ed->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;
			bytecnt = ed->hdr.control & EDESC_TOTAL_LEN;
			bd = &ed->bd[0];
			if (!(bd->control & EDESC_BD_CMD_IN_USE))
				++bd;
			VERIFY(bd->control & EDESC_BD_CMD_IN_USE);
			VERIFY(bd->buf);
			data_in_buf = bd->len;
			data_p = bd->buf;
			if (sdc->control & SD_FLAG_EMBEDDED_CS) {
				cs_blockcnt = (ed->hdr.control & EDESC_TOTAL_LEN) / (ECSUM_BLOCKSIZE + ECSUM_LEN);
				cs_bd = bd + bufcnt - 1;
				VERIFY(cs_bd->control & EDESC_BD_CMD_IN_USE);
				VERIFY(cs_bd->buf);
				cs_p = cs_bd->buf + cs_bd->len - cs_blockcnt * CSUM_LEN;
				data_in_block = ECSUM_BLOCKSIZE;
			}
			else
				data_in_block = bytecnt;

			while (bytecnt) {
				/* Write user data */
				while (data_in_block) {
					requested_data_len = MIN(data_in_block, data_in_buf);
					while (requested_data_len) {
#if (ISPFC_SIM == ISPFC_SIM_LINUX_FILE)
						actual_data_len = write(dev->fd, data_p, requested_data_len);
						if (actual_data_len < 0)
							panic("write failed in %s: %s", 
							      "ispfc_sim_scsi_execute_disk_write", strerror(errno));
#elif (ISPFC_SIM == ISPFC_SIM_RAM_DISK)
						memcpy(ram_buf_p, data_p, requested_data_len);
						ram_buf_p += requested_data_len;
						actual_data_len = requested_data_len;
#endif
						requested_data_len -= actual_data_len;
						sim_xfer_len -= actual_data_len;
						bytecnt -= actual_data_len;
						data_in_block -= actual_data_len;
						data_in_buf -= actual_data_len;
						data_p += actual_data_len;
					}
					if (data_in_buf == 0 && data_in_block > 0) {
						++bd;
						VERIFY(bd->control & EDESC_BD_CMD_IN_USE);
						VERIFY(bd->buf);
						data_in_buf = bd->len;
						data_p = bd->buf;
					}
				}
				if (!(sdc->control & SD_FLAG_EMBEDDED_CS))
					continue;
				else
					data_in_block = ECSUM_BLOCKSIZE;

				/* Write checksum */
				requested_data_len = ECSUM_LEN;
				while (requested_data_len) {
#if (ISPFC_SIM == ISPFC_SIM_LINUX_FILE)
					actual_data_len = write(dev->fd, cs_p, requested_data_len);
					if (actual_data_len < 0)
						panic("write failed in %s: %s",
						      "ispfc_sim_scsi_execute_disk_write", strerror(errno));
#elif (ISPFC_SIM == ISPFC_SIM_RAM_DISK)
					memcpy(ram_buf_p, cs_p, requested_data_len);
					ram_buf_p += requested_data_len;
					actual_data_len = requested_data_len;
#endif
					requested_data_len -= actual_data_len;
					sim_xfer_len -= actual_data_len;
					bytecnt -= actual_data_len;
					cs_p += actual_data_len;
				}
			} /* while (bytecnt) */

			if (ed->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER) {
				ed = (eee_desc_t *)ed->bd[EEE_MAX_BUFS-1].buf;
				VERIFY(ed);
			}
			else 
				ed = NULL;
		} /* while (ed) */
	} /* for (sdc) */
	if (sim_xfer_len > 0)
		panic("Data underflow on write to phys_dev %x.\n", dev->phys_dev_id);
	else
	if (sim_xfer_len < 0)
		panic("Data overflow on write to phys_dev %x.\n", dev->phys_dev_id);
}

/* 
 ************************************************
 * TAPE COMMANDS 
 ************************************************
 */
void
ispfc_sim_scsi_execute_tape_inquiry(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
	eee_desc_t *	ed = sd->eDesc;
	eee_descBuf_t *	bd = &ed->bd[1];
	byte *		buf = bd->buf;
	byte *		cdb = sd->cdb;
	uint32     	sim_xfer_len;

	if (cdb[1] & SCSI_INQUIRY_EVPD) {
		/*
		 * Handle VPD pages 
		 */
		switch (cdb[2]) {
		case SCSI_INQUIRY_SERIALNO_PAGE_CODE:
			buf[0] = SCSI_DTYPE_DIRECT;
			buf[1] = SCSI_INQUIRY_SERIALNO_PAGE_CODE;
			buf[2] = 0;
			buf[3] = 8;
			memcpy(&buf[4], dev->scsi_serialno, 8);	
			sim_xfer_len = 12;
			break;

		default:
			VERIFY(0);
		}
	}
	else {
		/*
		 * "Standard" inquiry
		 */
		buf[0] = SCSI_DTYPE_SEQUENTIAL;
		buf[1] = 0;

		buf[2] = 2;
		buf[3] = 2;
		buf[4] = 0;
		buf[5] = 30;
		memcpy(&buf[8], dev->scsi_vendor_id, 8);
		memcpy(&buf[16], dev->scsi_product_id, 16);
		memcpy(&buf[32], dev->scsi_revision, 4);
		sim_xfer_len = 36;
	}
	ispfc_edesc_set_len(ed, sim_xfer_len);
	sd->status |= SD_DEV_DATA_VALID;
}

void
ispfc_sim_scsi_execute_tape_mode_sense(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
#define MODE_DESC_LEN 8
	byte *			cdb = &sd->cdb[0];
	int			page_code = cdb[2] & 0x3F;
	int			page_len = cdb[4];
	scsi_mode_header_t *	mode_header;
	uchar8 *		mode_desc;
	scsi_mode_page_desc_t *	mode_page;
	eee_descBuf_t *		bd;
	uint32			sim_xfer_len = 0;

	VERIFY(sd->eDesc);
	bd = &sd->eDesc->bd[0];
	if (!(bd->control & EDESC_BD_CMD_IN_USE))
		++bd;
	VERIFY(bd->control & EDESC_BD_CMD_IN_USE);

	mode_header = (scsi_mode_header_t *)bd->buf;
	mode_desc = (uchar8 *)bd->buf + sizeof(scsi_mode_header_t);
	mode_page = (scsi_mode_page_desc_t *)(mode_desc + MODE_DESC_LEN);

	bzero(bd->buf, page_len);

	switch(page_code) {
	case SCSI_MPAGE_COMPRESSION:
		sim_xfer_len = sizeof(scsi_mode_header_t) + MODE_DESC_LEN + 2 + sizeof(scsi_mode_compress_mode_t);

		/*
		 * Set up header(s)
		 */
		mode_header->length = sim_xfer_len - 1;
		mode_header->flag |= (dev->write_protected ? 0x80 : 0x00);
		mode_header->block_desc_len = MODE_DESC_LEN;
		*(uint32 *)mode_desc = dev->max_bytes_to_eod / 
			(dev->fixed_blocksize > 0 ? dev->fixed_blocksize : ISPFC_SIM_TAPE_DEFAULT_BLOCKSIZE)
			- 1;
		inplace_ntohl((uint32 *)mode_desc);
		*((uint32 *)mode_desc + 1) = dev->fixed_blocksize;
		inplace_ntohl((uint32 *)mode_desc + 1);		

		mode_page->code = page_code;
		mode_page->len = mode_header->length - 
			sizeof(scsi_mode_header_t) -
			MODE_DESC_LEN - 2;

		mode_page->u.compress_mode.compression_enab =
			SCSI_CM_COMPRESSION_CAPABLE |
			(dev->compression_enabled ? SCSI_CM_ENABLE_COMPRESSION : 0);
		break;
		
	default:
		SD_SET_DEV_STATUS(sd, SCSI_DEV_STATUS_CHECK);
		sd->senseData = eee_ramAlloc(SIM_SENSE_LEN);
		bzero(sd->senseData, SIM_SENSE_LEN);
		SD_SET_SENSE(sd, illegal_mode_page_sense, SIM_SENSE_LEN);
	}

	if (sim_xfer_len) {
		ispfc_edesc_set_len(sd->eDesc, sim_xfer_len);
		sd->status |= SD_DEV_DATA_VALID;
	}
}

void
ispfc_sim_scsi_execute_tape_mode_select(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
#define MODE_DESC_LEN 8
	int			page_code;
	scsi_mode_header_t *	mode_header;
	uchar8 *		mode_desc;
	scsi_mode_page_desc_t *	mode_page;
	eee_descBuf_t *		bd;

	VERIFY(sd->eDesc);
	bd = &sd->eDesc->bd[0];
	if (!(bd->control & EDESC_BD_CMD_IN_USE))
		++bd;
	VERIFY(bd->control & EDESC_BD_CMD_IN_USE);

	mode_header = (scsi_mode_header_t *)bd->buf;
	mode_desc = (uchar8 *)bd->buf + sizeof(scsi_mode_header_t);
	mode_page = (scsi_mode_page_desc_t *)(mode_desc + MODE_DESC_LEN);

	page_code = mode_page->code;

	switch(page_code) {
	case SCSI_MPAGE_COMPRESSION:
		dev->fixed_blocksize = *((uint32 *)mode_desc + 1);
		inplace_htonl(&dev->fixed_blocksize);
		dev->compression_enabled = ((mode_page->u.compress_mode.compression_enab & SCSI_CM_ENABLE_COMPRESSION) != 0);
		break;

	default:
		SD_SET_DEV_STATUS(sd, SCSI_DEV_STATUS_CHECK);
		sd->senseData = eee_ramAlloc(SIM_SENSE_LEN);
		bzero(sd->senseData, SIM_SENSE_LEN);
		SD_SET_SENSE(sd, illegal_mode_page_sense, SIM_SENSE_LEN);
	}
}

#if (ISPFC_SIM == ISPFC_SIM_LINUX_FILE)
void
ispfc_sim_scsi_execute_tape_read(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
	int		count = (sd->cdb[2] << 16) | (sd->cdb[3] << 8) | sd->cdb[4];
	int		fixed = (uint)sd->cdb[1] & 0x01;
	int		sili = (uint)sd->cdb[1] & 0x02;
	int		i;
	int		req_len, cur_len;
	int		total_requested_bytes, total_xferred_bytes;
	int		ili = 0;
	int		req_num_blocks;
	int		req_block_size;
	int		residue;
	eee_desc_t *	ed;
	eee_descBuf_t * bd;
	char *		p;
	int		bufcnt;
	int		bd_xfer_residue, block_xfer_residue;

	/*
	 * Can't have SILI and fixed on at the same time
	 */
	if (sili && fixed) {
		SD_SET_DEV_STATUS(sd, SCSI_DEV_STATUS_CHECK);
		sd->senseData = eee_ramAlloc(SIM_SENSE_LEN);
		bzero(sd->senseData, SIM_SENSE_LEN);
		SD_SET_SENSE(sd, illegal_sili_sense, SIM_SENSE_LEN);
		return;
	}

	if (fixed && dev->fixed_blocksize == 0) {
		SD_SET_DEV_STATUS(sd, SCSI_DEV_STATUS_CHECK);
		sd->senseData = eee_ramAlloc(SIM_SENSE_LEN);
                bzero(sd->senseData, SIM_SENSE_LEN);
                SD_SET_SENSE(sd, block_mismatch_sense, SIM_SENSE_LEN);
                return;
        }

	if (!fixed) {
	   	req_num_blocks = 1;
		req_block_size = count;
	}
	else {
	   	req_num_blocks = count;
		req_block_size = dev->fixed_blocksize;
	}		
	total_requested_bytes = req_num_blocks * req_block_size;
	total_xferred_bytes = 0;

	VERIFY(sd->status & SD_DEV_DATA_IN_EXPECT);
	VERIFY(sd->expectDataLen >= total_requested_bytes);

	if (count == 0)
		goto read_done;

	dev->dev_state |= ISPFC_DF_TAPE_MOTION_WITHOUT_WRITE;

	/*
	 * Open directory file if we need to.
	 */
	if (dev->current_dir == NULL)
		ispfc_sim_tape_directory_sync_to_file(dev);

	/* 
	 * Open data file if we need to.
	 */
	if (dev->current_data_fd == -1) {
		int offset = 0;
		
		sprintf(dev->current_data_filename, ISPFC_SIM_TAPE_DATA_FMT, simdir ? simdir : "/tmp", 
			dev->dev_portname.u.wwn, dev->current_fileno);
		if ((dev->current_data_fd = open(dev->current_data_filename, O_RDWR)) == -1)
			dev->current_data_fd = open(dev->current_data_filename, O_RDWR | O_CREAT | O_TRUNC, 0644);
		if (dev->current_data_fd == -1)
			panic("open failed in %s on %s: %s",
			      "ispfc_sim_scsi_execute_tape_read", dev->current_data_filename, strerror(errno));

		/* Seek to right offset */
		for (i = 0; i < dev->current_blockno; ++i)
			offset += dev->current_dir[i];
		if (lseek(dev->current_data_fd, offset, SEEK_SET) == -1)
			panic("lseek failed in %s: %s",
			      "ispfc_sim_scsi_execute_tape_read", strerror(errno));
	}

	VERIFY(dev->current_blockno <= dev->current_dir_size);

	VERIFY(ed = sd->eDesc);
	bufcnt = (ed->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;
	bd = &ed->bd[0];
	if (!(bd->control & EDESC_BD_CMD_IN_USE))
		++bd;
	p = bd->buf;
	bd_xfer_residue = bd->len;

	while (req_num_blocks) {

		/*
		 * Check if next read will hit the file mark or EOD.
		 */
		if (dev->current_blockno == dev->current_dir_size) {
			/* 
			 * Is the next file available? If not, then we
			 * hit EOD, else we hit a filemark
			 */
			ispfc_sim_tape_directory_flush_to_file(dev, 1);
			if (dev->current_data_fd != -1) {
				close(dev->current_data_fd);
				dev->current_data_fd = -1;
			}

			sprintf(dev->current_dir_filename, ISPFC_SIM_TAPE_DIR_FMT, simdir ? simdir : "/tmp", 
				dev->dev_portname.u.wwn, dev->current_fileno+1);
			if ((dev->current_dir_fd = open(dev->current_dir_filename, O_RDWR)) == -1) {
				/* Hit EOD */
				FC_DBG(1, ("%s: phys_dev_id 0x%x (file %d, block %d) Hit EOD.\n", 
					   cb->debug_name, dev->phys_dev_id, dev->current_fileno, dev->current_blockno));

				SD_SET_DEV_STATUS(sd, SCSI_DEV_STATUS_CHECK);
				sd->senseData = eee_ramAlloc(SIM_SENSE_LEN);
				bzero(sd->senseData, SIM_SENSE_LEN);
				SD_SET_SENSE(sd, blank_check_sense, SIM_SENSE_LEN);
				residue = (fixed ? req_num_blocks : req_block_size);
				((scsi_sense_t *)sd->senseData)->info[0] = (residue & 0xFF000000) >> 24;
				((scsi_sense_t *)sd->senseData)->info[1] = (residue & 0x00FF0000) >> 16;
				((scsi_sense_t *)sd->senseData)->info[2] = (residue & 0x0000FF00) >>  8;
				((scsi_sense_t *)sd->senseData)->info[3] = (residue & 0x000000FF);
				goto read_done;
			}
			else {
				/* Hit filemark */
				FC_DBG(1, ("%s: phys_dev_id 0x%x (file %d, block %d) Hit filemark.\n", 
					   cb->debug_name, dev->phys_dev_id, dev->current_fileno, dev->current_blockno));
				/* Read over filemark */
				close(dev->current_dir_fd);
				dev->current_dir_fd = -1;
				++dev->current_fileno;
				dev->current_blockno = 0;
				
				SD_SET_DEV_STATUS(sd, SCSI_DEV_STATUS_CHECK);
				sd->senseData = eee_ramAlloc(SIM_SENSE_LEN);
                                bzero(sd->senseData, SIM_SENSE_LEN);
                                SD_SET_SENSE(sd, filemark_check_sense, SIM_SENSE_LEN);
				residue = (fixed ? req_num_blocks : req_block_size); 
				((scsi_sense_t *)sd->senseData)->info[0] = (residue & 0xFF000000) >> 24;
				((scsi_sense_t *)sd->senseData)->info[1] = (residue & 0x00FF0000) >> 16;
				((scsi_sense_t *)sd->senseData)->info[2] = (residue & 0x0000FF00) >>  8;
				((scsi_sense_t *)sd->senseData)->info[3] = (residue & 0x000000FF);
				goto read_done;
			}
		}

		/* 
		 * Check if block size requested equals block on tape.
		 */
		if (dev->current_dir[dev->current_blockno] != req_block_size) {
			FC_DBG(1, ("%s: phys_dev_id 0x%x (file %d, block %d) ILI.\n", 
				   cb->debug_name, dev->phys_dev_id, dev->current_fileno, dev->current_blockno));
			++ili;
			if (!sili) {
				SD_SET_DEV_STATUS(sd, SCSI_DEV_STATUS_CHECK);
				sd->senseData = eee_ramAlloc(SIM_SENSE_LEN);
                                bzero(sd->senseData, SIM_SENSE_LEN);
                                SD_SET_SENSE(sd, ili_check_sense, SIM_SENSE_LEN);
				residue = (fixed ? 
					   req_num_blocks : 
					   req_block_size - dev->current_dir[dev->current_blockno]);
				((scsi_sense_t *)sd->senseData)->info[0] = (residue & 0xFF000000) >> 24;
                                ((scsi_sense_t *)sd->senseData)->info[1] = (residue & 0x00FF0000) >> 16;
                                ((scsi_sense_t *)sd->senseData)->info[2] = (residue & 0x0000FF00) >>  8;
                                ((scsi_sense_t *)sd->senseData)->info[3] = (residue & 0x000000FF);
			}
		}

		block_xfer_residue = MIN(req_block_size, dev->current_dir[dev->current_blockno]);

		/*
		 * Read the data.
		 */
		while (block_xfer_residue) {
			req_len = MIN(bd_xfer_residue, block_xfer_residue);
			VERIFY(req_len > 0);
			while (req_len) {
				cur_len = read(dev->current_data_fd, p, req_len);
				if (cur_len < 0)
					panic("read failed in %s: %s", "ispfc_sim_scsi_execute_tape_read", strerror(errno));
				req_len -= cur_len;
				total_xferred_bytes += cur_len;
				block_xfer_residue -= cur_len;
				bd_xfer_residue -= cur_len;
				if (bd_xfer_residue)
					p += cur_len;
				else {
					if (--bufcnt == 0 && block_xfer_residue > 0) {
						VERIFY(ed->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER);
						ed = (eee_desc_t *)ed->bd[EEE_MAX_BUFS-1].buf;
						VERIFY(ed);
						bufcnt = (ed->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;
						bd = &ed->bd[0];
						if (!(bd->control & EDESC_BD_CMD_IN_USE))
							++bd;
						VERIFY((block_xfer_residue == 0) || ((bd->control & EDESC_BD_CMD_IN_USE) && bd->buf));
						p = bd->buf;
						bd_xfer_residue = bd->len;
					}
					else {
						++bd;
						VERIFY((block_xfer_residue == 0) || ((bd->control & EDESC_BD_CMD_IN_USE) && bd->buf));
						p = bd->buf;
						bd_xfer_residue = bd->len;
					}
				}
			}
		}

		++dev->current_blockno;
		--req_num_blocks;

		if (ili) {
			/* If ILI, then close file since we may be off */
			close(dev->current_data_fd);
                        dev->current_data_fd = -1;
			if (!sili)
				goto read_done;
		}
	} /* while (req_num_blocks) */

 read_done:
	ispfc_edesc_set_len(sd->eDesc, total_xferred_bytes);
	if (total_xferred_bytes)
		sd->status |= SD_DEV_DATA_VALID;
	return;
}

void
ispfc_sim_scsi_execute_tape_write(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
	int		count = (sd->cdb[2] << 16) | (sd->cdb[3] << 8) | sd->cdb[4];
	int		fixed = (uint)sd->cdb[1] & 0x01;
	int		i;
	eee_desc_t *	ed;
	eee_descBuf_t * bd;
        char *          p;
	int		bufcnt;
	int		req_len;
	int		cur_len;
	int		total_bytecount = 0;
	
	if (count == 0)
		return;

	if (dev->write_protected) {
		SD_SET_DEV_STATUS(sd, SCSI_DEV_STATUS_CHECK);
		sd->senseData = eee_ramAlloc(SIM_SENSE_LEN);
                bzero(sd->senseData, SIM_SENSE_LEN);
                SD_SET_SENSE(sd, data_protect_sense, SIM_SENSE_LEN);
                return;
	}

	if (fixed && dev->fixed_blocksize == 0) {
		SD_SET_DEV_STATUS(sd, SCSI_DEV_STATUS_CHECK);
		sd->senseData = eee_ramAlloc(SIM_SENSE_LEN);
                bzero(sd->senseData, SIM_SENSE_LEN);
                SD_SET_SENSE(sd, block_mismatch_sense, SIM_SENSE_LEN);
                return;
	}

	if (dev->dev_state & ISPFC_DF_TAPE_MOTION_WITHOUT_WRITE) {
		ispfc_sim_tape_erase_to_eod(dev);
		dev->dev_state &= ~ISPFC_DF_TAPE_MOTION_WITHOUT_WRITE;
		dev->bytes_at_eod = ispfc_sim_tape_bytecount_at_eod(cb, dev);
	}

	/*
	 * Open directory file if we need to.
	 */
	if (dev->current_dir == NULL) {
		ispfc_sim_tape_directory_sync_to_file(dev);
	}

	/* 
	 * Open data file if we need to.
	 */
	if (dev->current_data_fd == -1) {
		int offset = 0;

		sprintf(dev->current_data_filename, ISPFC_SIM_TAPE_DATA_FMT, simdir ? simdir : "/tmp",
                        dev->dev_portname.u.wwn, dev->current_fileno);
		if ((dev->current_data_fd = open(dev->current_data_filename, O_RDWR)) == -1) {
			dev->current_data_fd = open(dev->current_data_filename, O_RDWR | O_CREAT | O_TRUNC, 0644);
			if (dev->current_data_fd == -1)
				panic("open failed in %s on %s: %s",
				      "ispfc_sim_scsi_execute_tape_write", dev->current_data_filename, strerror(errno));
		}

		/* Seek to right offset */
		for (i = 0; i < dev->current_blockno; ++i)
			offset += dev->current_dir[i];
		if (lseek(dev->current_data_fd, offset, SEEK_SET) == -1)
			panic("lseek failed in %s: %s",
			      "ispfc_sim_scsi_execute_tape_write", strerror(errno));
	}

	/* Check for PEOT, defined as 120% of max. */
	if (dev->bytes_at_eod >= ISPFC_SIM_TAPE_MAX_BYTES * 120 / 100) {
		SD_SET_DEV_STATUS(sd, SCSI_DEV_STATUS_CHECK);
		sd->senseData = eee_ramAlloc(SIM_SENSE_LEN);
		bzero(sd->senseData, SIM_SENSE_LEN);
		SD_SET_SENSE(sd, peot_sense, SIM_SENSE_LEN);
		return;
	}

	/*
	 * Write the data first, and deal with updating the directory later.
	 */
	VERIFY(ed = sd->eDesc);
	while (ed) {
		bufcnt = (ed->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;
		bd = &ed->bd[0];
		if (!(bd->control & EDESC_BD_CMD_IN_USE))
			++bd;
		p = bd->buf;

		for (i = 0; i < bufcnt; ++i, ++bd) {
			VERIFY(bd->control & EDESC_BD_CMD_IN_USE);
			req_len = bd->len;
			p = bd->buf;
			while (req_len) {
				cur_len = write(dev->current_data_fd, p, req_len);
				if (cur_len < 0)
					panic("write failed in %s: %s", "ispfc_sim_scsi_execute_tape_write", strerror(errno));
				req_len -= cur_len;
				p += cur_len;
			}
			total_bytecount += bd->len;
		}
		if (ed->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER)
			ed = (eee_desc_t *)ed->bd[EEE_MAX_BUFS-1].buf;
		else 
			ed = NULL;
	}

	/* 
	 * Write one single variable length block or multiple fixed length blocks.
	 */
	if (!fixed) {
		VERIFY(count == total_bytecount);
		++dev->current_blockno;
		dev->bytes_at_eod += total_bytecount;
		ispfc_sim_tape_add_directory_entry(dev, total_bytecount);
	}
	else {
		VERIFY(count * dev->fixed_blocksize == total_bytecount);
		for (i = 0; i < count; ++i) {
			++dev->current_blockno;
			dev->bytes_at_eod += total_bytecount;
			ispfc_sim_tape_add_directory_entry(dev, dev->fixed_blocksize);
		}
	}

	/* 
	 * Check for LEOT, defined as 100% of max.
	 */
	if (dev->bytes_at_eod >= ISPFC_SIM_TAPE_MAX_BYTES) {
		SD_SET_DEV_STATUS(sd, SCSI_DEV_STATUS_CHECK);
		sd->senseData = eee_ramAlloc(SIM_SENSE_LEN);
                bzero(sd->senseData, SIM_SENSE_LEN);
                SD_SET_SENSE(sd, leot_sense, SIM_SENSE_LEN);
                return;
	}
}

void
ispfc_sim_scsi_execute_tape_rewind(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
	ispfc_sim_tape_rewind(dev);

	FC_DBG(1, ("%s: phys_dev_id 0x%x (file %d, block %d) Rewinding ....\n", 
		   cb->debug_name, dev->phys_dev_id, dev->current_fileno, dev->current_blockno));

	dev->dev_state |= ISPFC_DF_TAPE_MOTION_WITHOUT_WRITE;
}

void 
ispfc_sim_scsi_execute_tape_load_unload(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
	return;
}

void
ispfc_sim_scsi_execute_tape_write_fm(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
	uchar8 *	cdb = &sd->cdb[0];
	int		count = (cdb[2] << 16) | (cdb[3] << 8) | cdb[4];

	if (count == 0)
		return;

	if (cdb[2] & 0x80)
		goto illegal_count;

	if (dev->write_protected) {
		SD_SET_DEV_STATUS(sd, SCSI_DEV_STATUS_CHECK);
		sd->senseData = eee_ramAlloc(SIM_SENSE_LEN);
                bzero(sd->senseData, SIM_SENSE_LEN);
                SD_SET_SENSE(sd, data_protect_sense, SIM_SENSE_LEN);
                return;
	}

	/*
	 * Close the current data file if it's open
	 */
	if (dev->current_data_fd != -1) {
		close(dev->current_data_fd);
		dev->current_data_fd = -1;
	}

	/*
	 * Flush the current dir file and close it
	 */
	ispfc_sim_tape_directory_flush_to_file(dev, 1);

	if (dev->dev_state & ISPFC_DF_TAPE_MOTION_WITHOUT_WRITE) {
		ispfc_sim_tape_erase_to_eod(dev);
		dev->dev_state &= ~ISPFC_DF_TAPE_MOTION_WITHOUT_WRITE;
		dev->bytes_at_eod = ispfc_sim_tape_bytecount_at_eod(cb, dev);
	}

	FC_DBG(1, ("%s: phys_dev_id 0x%x (file %d, block %d) Writing %d filemarks.\n", 
		   cb->debug_name, dev->phys_dev_id, dev->current_fileno, dev->current_blockno, count));

	while (count--) {
		ispfc_sim_tape_directory_flush_to_file(dev, 1);

		dev->current_blockno = 0;
		dev->current_fileno++;

		ispfc_sim_tape_directory_sync_to_file(dev);
	}

	return;

 illegal_count:
	SD_SET_DEV_STATUS(sd, SCSI_DEV_STATUS_CHECK);
	sd->senseData = eee_ramAlloc(SIM_SENSE_LEN);
	bzero(sd->senseData, SIM_SENSE_LEN);
	SD_SET_SENSE(sd, illegal_count_sense, SIM_SENSE_LEN);
	return;
}

void
ispfc_sim_scsi_execute_tape_space(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
	uchar8 *	cdb = &sd->cdb[0];
	int		code = cdb[1] & 0x07;
	int		count = (cdb[2] << 16) | (cdb[3] << 8) | cdb[4];
	int		tmp_fileno;
	char		tmp_filename[128];
	int		tmp_fd;
	int		residue;

	/* Sign extend the count */
	if (cdb[2] & 0x80)
		count |= ((int)-1 & ~0xFFFFFF);

	dev->dev_state |= ISPFC_DF_TAPE_MOTION_WITHOUT_WRITE;
	
	if (dev->current_data_fd != -1) {
		close(dev->current_data_fd);
		dev->current_data_fd = -1;
	}
	ispfc_sim_tape_directory_flush_to_file(dev, 1);

	switch(code) {
	case 0:			/* Space fixed/variable length records */
		FC_DBG(1, ("%s: phys_dev_id 0x%x (file %d, block %d) Spacing %d blocks.\n", 
			   cb->debug_name, dev->phys_dev_id, dev->current_fileno, dev->current_blockno, count));
		if (count == 0)
		    return;
		if (count > 0) {
			ispfc_sim_tape_directory_sync_to_file(dev);
			while (count) {
				if (dev->current_blockno == dev->current_dir_size) {
					/* EOD or filemark hit? Check if next file is available */
					tmp_fileno = dev->current_fileno+1;
					sprintf(tmp_filename, ISPFC_SIM_TAPE_DIR_FMT, simdir ? simdir : "/tmp", 
						dev->dev_portname.u.wwn, tmp_fileno);
					if ((tmp_fd = open(tmp_filename, O_RDWR)) == -1) {
						/* We hit EOD */
						residue = count;
						goto eod_found;
					}
					else {
						/* We hit a filemark */
						close(tmp_fd);
						residue = count;
						goto filemark_found;
					}
				}
				++dev->current_blockno;
				--count;
			}
		}
		else {
			while (count) {
				if (dev->current_blockno == 0) {
					/* BOT or filemark hit? Check if current file is 0 */
					if (dev->current_fileno == 0) {
						/* We hit BOT */
						residue = -count;
						goto pbot_found;
					}
					else {
						/* We hit a filemark */
                                                residue = -count;
                                                goto filemark_found;
					}
				}
				--dev->current_blockno;
                                ++count;
			}
		}
		break;

	case 1:			/* Space filemarks */
		FC_DBG(1, ("%s: phys_dev_id 0x%x (file %d, block %d) Spacing %d filemarks.\n", 
			   cb->debug_name, dev->phys_dev_id, dev->current_fileno, dev->current_blockno, count));
		if (count == 0)
		    return;
		if (count > 0) {
			while (count) {
				/*
				 * Is next file available?
				 */
				tmp_fileno = dev->current_fileno+1;
				sprintf(tmp_filename, ISPFC_SIM_TAPE_DIR_FMT, simdir ? simdir : "/tmp",
					dev->dev_portname.u.wwn, tmp_fileno);
				if ((tmp_fd = open(tmp_filename, O_RDWR)) == -1) 
					break;
				close(tmp_fd);
				dev->current_fileno = tmp_fileno;
				--count;
			}

			ispfc_sim_tape_directory_sync_to_file(dev);
			residue = count;

			/* Did we hit EOD before we found the file we're looking for? */
			if (residue == 0) {
				/* Seek to beginning of file */
				dev->current_blockno = 0;
			}
			else {
				/* Seek to end of file */
				dev->current_blockno = dev->current_dir_size;
				residue = count;
				goto eod_found;
			}
		}
		else {
			/* Will we hit PBOT if we try to go back as far as requested? */
			if (count + dev->current_fileno < 0) {
				dev->current_fileno = 0;
				residue = -(count + dev->current_fileno);
			}
			else {
				dev->current_fileno += count;
                                residue = 0;
			} 

                        ispfc_sim_tape_directory_sync_to_file(dev);

			/* Did we hit PBOT? */
			if (residue == 0) {
				/* Seek to end of requested file */
				dev->current_blockno = dev->current_dir_size;
			}
			else {
				/* Seek to beginning of first file */
				dev->current_blockno = 0;
                                goto pbot_found;
			}
		}
		break;

	default:
		SD_SET_DEV_STATUS(sd, SCSI_DEV_STATUS_CHECK);
		sd->senseData = eee_ramAlloc(SIM_SENSE_LEN);
		bzero(sd->senseData, SIM_SENSE_LEN);
		SD_SET_SENSE(sd, illegal_code_sense, SIM_SENSE_LEN);
	}

	return;


 eod_found:
	SD_SET_DEV_STATUS(sd, SCSI_DEV_STATUS_CHECK);
	sd->senseData = eee_ramAlloc(SIM_SENSE_LEN);
	bzero(sd->senseData, SIM_SENSE_LEN);
	SD_SET_SENSE(sd, eod_found_sense, SIM_SENSE_LEN);
	((scsi_sense_t *)sd->senseData)->info[0] = (residue & 0xFF000000) >> 24;
	((scsi_sense_t *)sd->senseData)->info[1] = (residue & 0x00FF0000) >> 16;
	((scsi_sense_t *)sd->senseData)->info[2] = (residue & 0x0000FF00) >>  8;
	((scsi_sense_t *)sd->senseData)->info[3] = (residue & 0x000000FF);
	return;

 pbot_found:
	SD_SET_DEV_STATUS(sd, SCSI_DEV_STATUS_CHECK);
	sd->senseData = eee_ramAlloc(SIM_SENSE_LEN);
	bzero(sd->senseData, SIM_SENSE_LEN);
	SD_SET_SENSE(sd, pbot_found_sense, SIM_SENSE_LEN);
	((scsi_sense_t *)sd->senseData)->info[0] = (residue & 0xFF000000) >> 24;
	((scsi_sense_t *)sd->senseData)->info[1] = (residue & 0x00FF0000) >> 16;
	((scsi_sense_t *)sd->senseData)->info[2] = (residue & 0x0000FF00) >>  8;
	((scsi_sense_t *)sd->senseData)->info[3] = (residue & 0x000000FF);
	return;

 filemark_found:
	SD_SET_DEV_STATUS(sd, SCSI_DEV_STATUS_CHECK);
	sd->senseData = eee_ramAlloc(SIM_SENSE_LEN);
	bzero(sd->senseData, SIM_SENSE_LEN);
	SD_SET_SENSE(sd, filemark_found_sense, SIM_SENSE_LEN);
	((scsi_sense_t *)sd->senseData)->info[0] = (residue & 0xFF000000) >> 24;
	((scsi_sense_t *)sd->senseData)->info[1] = (residue & 0x00FF0000) >> 16;
	((scsi_sense_t *)sd->senseData)->info[2] = (residue & 0x0000FF00) >>  8;
	((scsi_sense_t *)sd->senseData)->info[3] = (residue & 0x000000FF);
	return;
}

void
ispfc_sim_scsi_execute_tape_erase(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
	ispfc_sim_tape_rewind(dev);

	FC_DBG(1, ("%s: phys_dev_id 0x%x (file %d, block %d) Erasing tape.\n", 
		   cb->debug_name, dev->phys_dev_id, dev->current_fileno, dev->current_blockno));

	ispfc_sim_tape_erase_to_eod(dev);
}

void 
ispfc_sim_scsi_execute_tape_read_position(ispfc_cb_t *cb, ispfc_dev_t *dev, scsi_desc_t *sd)
{
	scsi_rpos_long_desc_t *	data;
	eee_descBuf_t *		bd;

	/* We only handle TCLP and long returns */
	if ((sd->cdb[1] & SCSI_RPOS_CONTROL_LONG) &&
	    (sd->cdb[1] & SCSI_RPOS_CONTROL_TCLP)) {
		VERIFY(sd->eDesc);
		bd = &sd->eDesc->bd[0];
		if (!(bd->control & EDESC_BD_CMD_IN_USE))
			++bd;
		VERIFY(bd->control & EDESC_BD_CMD_IN_USE);
		data = (scsi_rpos_long_desc_t *)bd->buf;
		bzero(data, sizeof(scsi_rpos_long_desc_t));

		if (dev->current_blockno == 0 && dev->current_fileno == 0)
			data->flags |= SCSI_RPOS_FLAGS_BOP;
		data->block_number[1] = dev->current_blockno;
		inplace_htonl(&data->block_number[1]);
		data->file_number[1] = dev->current_fileno;
		inplace_htonl(&data->file_number[1]);

		ispfc_edesc_set_len(sd->eDesc, sizeof(scsi_rpos_long_desc_t));
		sd->status |= SD_DEV_DATA_VALID;
	}
	else {
		SD_SET_DEV_STATUS(sd, SCSI_DEV_STATUS_CHECK);
		sd->senseData = eee_ramAlloc(SIM_SENSE_LEN);
		bzero(sd->senseData, SIM_SENSE_LEN);
		SD_SET_SENSE(sd, invalid_field_in_cdb_sense, SIM_SENSE_LEN);
		return;
	}
}

/* 
 ************************************************
 * TAPE SUPPORT ROUTINES
 ************************************************
 */
void
ispfc_sim_tape_directory_flush_to_file(ispfc_dev_t *dev, int free)
{
	int	req_len, len;
	char *	p;

	if (dev->current_dir_dirty) {
		if (dev->current_dir_fd == -1) {
			sprintf(dev->current_dir_filename, ISPFC_SIM_TAPE_DIR_FMT, simdir ? simdir : "/tmp", 
				dev->dev_portname.u.wwn, dev->current_fileno);
			if ((dev->current_dir_fd = open(dev->current_dir_filename, O_RDWR)) == -1) {
				dev->current_dir_fd = open(dev->current_dir_filename, O_RDWR | O_CREAT | O_TRUNC, 0644);
				if (dev->current_dir_fd == -1)
					panic("open failed in %s on %s: %s\n", 
					      "ispfc_sim_tape_directory_flush_to_file", dev->current_dir_filename, 
					      strerror(errno));
			}
		}

		if (lseek(dev->current_dir_fd, 0, SEEK_SET) == -1)
			panic("lseek failed in %s: %s", 
			      "ispfc_sim_tape_directory_flush_to_file", strerror(errno));

		if (ftruncate(dev->current_dir_fd, 0) == -1)
			panic("ftruncate failed in %s: %s", 
			      "ispfc_sim_tape_directory_flush_to_file", strerror(errno));

		req_len = dev->current_dir_size * sizeof(tape_dir_ent_t);
		p = (char *)dev->current_dir;
		while (req_len) {
			len = write(dev->current_dir_fd, p, req_len);
			if (len < 0)
				panic("write failed in %s: %s", 
				      "ispfc_sim_tape_directory_flush_to_file", strerror(errno));
			req_len -= len;
			p += len;
		}
		dev->current_dir_dirty = 0;
	}
	if (dev->current_dir_fd != -1) {
		close(dev->current_dir_fd);
		dev->current_dir_fd = -1;
	}
	if (free && dev->current_dir) {
		eee_ramDealloc(dev->current_dir);
		dev->current_dir = NULL;
	}
}

void
ispfc_sim_tape_directory_sync_to_file(ispfc_dev_t *dev)
{
	int	file_len, req_len, len;
	char *	p;

	VERIFY(dev->current_dir_dirty == 0);

	if (dev->current_dir_fd == -1) {
		sprintf(dev->current_dir_filename, ISPFC_SIM_TAPE_DIR_FMT, simdir ? simdir : "/tmp", 
			dev->dev_portname.u.wwn, dev->current_fileno);

		if ((dev->current_dir_fd = open(dev->current_dir_filename, O_RDWR)) == -1) {
			dev->current_dir_fd = open(dev->current_dir_filename, O_RDWR | O_CREAT | O_TRUNC, 0644);
			if (dev->current_dir_fd == -1)
				panic("open failed in %s on %s: %s\n",
				      "ispfc_sim_tape_directory_sync_to_file", dev->current_dir_filename,
				      strerror(errno));
		}
	}

	if (lseek(dev->current_dir_fd, 0, SEEK_SET) == -1)
		panic("lseek failed in %s: %s", 
		      "ispfc_sim_tape_directory_sync_to_file", strerror(errno));
	file_len = lseek(dev->current_dir_fd, 0, SEEK_END);
	if (file_len == -1)
		panic("lseek failed in %s: %s",
                      "ispfc_sim_tape_directory_sync_to_file", strerror(errno));
	if (file_len % sizeof(tape_dir_ent_t))
		panic("Illegal directory entry file length (%s) %d\n", 
		      dev->current_dir_filename, file_len);

	/* Currently we don't expect to be merging in-core and on-file directories. */
	VERIFY(dev->current_dir == NULL);

	if (file_len == 0)
		dev->current_dir_alloced_size = 10;
	else
		dev->current_dir_alloced_size = 2 * file_len / sizeof(tape_dir_ent_t);
	dev->current_dir = eee_ramAlloc(dev->current_dir_alloced_size * sizeof(tape_dir_ent_t));
	VERIFY(dev->current_dir);

	if (lseek(dev->current_dir_fd, 0, SEEK_SET) == -1)
                panic("lseek failed in %s: %s",
                      "ispfc_sim_tape_directory_sync_to_file", strerror(errno));
	
	req_len = file_len;
	p = (char *)dev->current_dir;
	while (req_len) {
		len = read(dev->current_dir_fd, p, req_len);
		if (len < 0)
			panic("read failed in %s: %s",
			      "ispfc_sim_tape_directory_sync_to_file", strerror(errno));
		req_len -= len;
		p += len;
	}
	dev->current_dir_size = file_len / sizeof(tape_dir_ent_t);
	dev->current_dir_dirty = 0;
	close(dev->current_dir_fd);
	dev->current_dir_fd = -1;
}

void
ispfc_sim_tape_add_directory_entry(ispfc_dev_t *dev, int blocksize)
{
	tape_dir_ent_t *	new_directory;
	
	/*
	 * Dynamically double size of incore directory if we've run out of space
	 */
	if (dev->current_dir_size == dev->current_dir_alloced_size) {
		dev->current_dir_alloced_size = dev->current_dir_alloced_size * 2;
		new_directory = eee_ramAlloc(dev->current_dir_alloced_size * sizeof(tape_dir_ent_t));
		VERIFY(new_directory);
		memcpy(new_directory, dev->current_dir, dev->current_dir_alloced_size * sizeof(tape_dir_ent_t));
		eee_ramDealloc(dev->current_dir);
		dev->current_dir = new_directory;
	}
	dev->current_dir[dev->current_dir_size++] = blocksize;
	dev->current_dir_dirty++;
}

void
ispfc_sim_tape_erase_to_eod(ispfc_dev_t *dev)
{
	int	file_no;
	int	fd;
	char	filename[128];
	int	i;
	int	bytecount = 0;

	/*
	 * Close the current data file if it's open
	 */
	if (dev->current_data_fd != -1) {
		close(dev->current_data_fd);
		dev->current_data_fd = -1;
	}

	/*
	 * Truncate current data file to "eod"
	 */
	if (dev->current_dir == NULL)
		ispfc_sim_tape_directory_sync_to_file(dev);
	sprintf(filename, ISPFC_SIM_TAPE_DATA_FMT, simdir ? simdir : "/tmp", 
		dev->dev_portname.u.wwn, dev->current_fileno);
	fd = open(filename, O_RDWR);
	if (fd > 0) {
		for (i = 0; i < dev->current_blockno; ++i)
			bytecount += dev->current_dir[i];
		if (ftruncate(fd, bytecount) == -1)
			panic("ftruncate failed in %s: %s",
			      "ispfc_sim_tape_erase_to_eod", strerror(errno));
		close(fd);
	}

	/*
	 * Flush the current dir file and close it
	 */
	dev->current_dir_size = dev->current_blockno;
	++dev->current_dir_dirty;
	ispfc_sim_tape_directory_flush_to_file(dev, 1);

	/*
	 * Now go delete all files after this point, i.e. erase
	 * remainder of tape
	 */
	for (file_no = dev->current_fileno+1 ;; ++file_no) {
		/* If DIR file exists, remove it and .... */
		sprintf(filename, ISPFC_SIM_TAPE_DIR_FMT, simdir ? simdir : "/tmp", 
			dev->dev_portname.u.wwn, file_no);
		if ((fd = open(filename, O_RDWR)) == -1)
			break;
		close(fd);
		unlink(filename);

		/* remove data file */
		sprintf(filename, ISPFC_SIM_TAPE_DATA_FMT, simdir ? simdir : "/tmp",
			dev->dev_portname.u.wwn, file_no);
		unlink(filename);
	}
}

uint32
ispfc_sim_tape_bytecount_at_eod(ispfc_cb_t *cb, ispfc_dev_t *dev)
{
	uint32	bytecount = 0;
	int	old_file_no;
	int	i;

	/*
	 * Flush out current directory
	 */
	old_file_no = dev->current_fileno;
	ispfc_sim_tape_directory_flush_to_file(dev, 1);

	for (dev->current_fileno = 0; ; ++dev->current_fileno) {
		sprintf(dev->current_dir_filename, ISPFC_SIM_TAPE_DIR_FMT, simdir ? simdir : "/tmp",
			dev->dev_portname.u.wwn, dev->current_fileno);
                if ((dev->current_dir_fd = open(dev->current_dir_filename, O_RDWR)) == -1)
                        break;
                close(dev->current_dir_fd);
		dev->current_dir_fd = -1;
		ispfc_sim_tape_directory_sync_to_file(dev);
		for (i = 0; i < dev->current_dir_size; ++i)
			bytecount += dev->current_dir[i];
		if (dev->current_dir) {
			eee_ramDealloc(dev->current_dir);
			dev->current_dir = NULL;
		}
	}

	/*
	 * Restore current directory
	 */
	dev->current_fileno = old_file_no;
	ispfc_sim_tape_directory_sync_to_file(dev);
	FC_DBG(1, ("%s: phys_dev 0x%x (file %d, block %d) %d bytes at EOD.\n",
		   cb->debug_name, dev->phys_dev_id, 
		   dev->current_fileno, dev->current_blockno, bytecount));
	return(bytecount);
}

void
ispfc_sim_tape_rewind(ispfc_dev_t *dev)
{
	/*
	 * Close the current data file if it's open
	 */
	if (dev->current_data_fd != -1) {
		close(dev->current_data_fd);
		dev->current_data_fd = -1;
	}

	/*
	 * Flush the current dir file and close it
	 */
	ispfc_sim_tape_directory_flush_to_file(dev, 1);

	dev->current_blockno = 0;
	dev->current_fileno = 0;
}
#endif /* ISPFC_SIM == ISPFC_SIM_LINUX_FILE */

#endif /* ISPFC_SIM */
