/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        lock-api.h
 *
 * RCS:  $Id: $
 *
 * Copyright (C) 2002, ClariStor Inc.
 *
 * Description: Byte range lock API functions definitions.
 *
 * Created by:  Xiaobing Zhang
 *
 * Date Created:    11/23/2002
 *
 *-----------------------------------------------------------------
 */
#ifndef _LOCK_API_H
#define _LOCK_API_H

#include <exports/timer-api.h>

/*
 * Lock owner type
 */
#define LOCK_OWNER_NLM   0
#define LOCK_OWNER_CIFS  1
#define LOCK_OWNER_NFSV4 2
#define LOCK_OWNER_DUMMY 3

/*
 * Lock type
 */
#define LOCK_TYPE_EXCL   0       /*Exclusive lock*/
#define LOCK_TYPE_SHAR   1       /*Shared lock*/
#define LOCK_TYPE_DUMMY  2       /*Don't care the lock type*/

/*
 * Lock state
 */
#define LOCK_STATE_GRANTED      0   /*Lock is granted*/
#define LOCK_STATE_DENIED       1   /*Lock is denied*/
#define LOCK_STATE_BLOCKED      2   /*Lock is blocked*/

/*
 * Lock error code
 * TODO: may change value
 */
#define LOCK_ERR_GENERAL                    NFX_ERR
#define LOCK_ERR_INVALID_INPUT				NFX_ERR-1
#define LOCK_ERR_NOT_FOUND					NFX_ERR-2
#define LOCK_ERR_UNSUPPORTED_OWNER_TYPE     NFX_ERR-3
#define LOCK_ERR_REACH_LIMIT                NFX_ERR-4
#define LOCK_ERR_INSUFFICIENT_MEMORY        NFX_ERR-5
#define LOCK_ERR_DEADLOCK                   NFX_ERR-6

/*-----------------------
 * Lock range information
 *-----------------------
 */
#define LOCK_MIN_OFFSET     0
#define LOCK_MAX_OFFSET     0xFFFFFFFFFFFFFFFF
#define LOCK_MAX_LENGTH     (LOCK_MAX_OFFSET - LOCK_MIN_OFFSET + 1)

/*----------------
 * Data structures
 *----------------
 */

/*
 * Lock request.
 * This data structure is used by APIs to pass various lock information
 * as parameters.
 */

typedef uint32  ip_index_t;

typedef struct lock_requestCifsOwnerInfo {
    uint16     tid;
    uint16     fid;
    uint16     pid;
} lock_requestCifsOwnerInfo_t;

typedef struct lock_requestNlmOwnerInfo {
    uint32     clientState;
	ip_index_t clientIpAddr;
    int32      pid;
    uint16     callerNameLen;
    uint16     ownerNameLen;
    char       *callerName;
    char       *ownerName;
} lock_requestNlmOwnerInfo_t;

typedef struct lock_request {
    uint8      lockType;      /*LOCK_TYPE_EXCL or LOCK_TYPE_SHAR*/
    uint8      ownerType;     /*LOCK_OWNER_NLM, LOCK_OWNER_CIFS, or
                                LOCK_OWNER_NFSV4*/
    union
    {
        lock_requestCifsOwnerInfo_t cifsOwner;
        lock_requestNlmOwnerInfo_t  nlmOwner;
    } owner;
    eeeRTC_t   timeout;      /*Time when the lock timeout if it is pending.
                               This is in real time clock form. A timeout value
                               of infinite indicates to wait as long as it
                               takes. A timeout value of 0 indicates not to
                               wait.*/
    uint64     offset;
    uint64     len;
    vs_mem_index_t ulPrivate;/*Pointer to context passed by the protocol 
                               layer. Will be used during callback. For NLM,
                               it points to a nlm_request_t structure.*/
    vs_mem_index_t fcb;      /*Type: fcb_t. Used by NLM*/
} lock_request_t;

/*--------------------------------------------------------------------------
 *    Several API functions defined later have a parameter of type
 *  lock_request_t. For functions, we expect different fields of the above
 *  structure to be filled. These functions include,
 *
 *      1. lock_byteRangeLockReq
 *      2. lock_unlockSingleReq
 *      3. lock_cancelLockReq
 *      4. lock_testLockReq
 *      5. lock_checkLockForWriteReq
 *      6. lock_checkLockForReadReq
 *
 *    The following table shows what fields are required for each function.
 *
 *      TABLE 1. Required input parameters for different functions.
 *      =============================================================
 *       Functions:       |   1   |   2   |   3   |  4  |  5  |  6  |
 *                        | ----- | ----- | ----- | --- | --- | --- |
 *                        | C | N | C | N | C | N |  N  |  C  |  C  |
 *                        | | | L | I | L | I | L |  L  |  I  |  I  |
 *                        | F | M | F | M | F | M |  M  |  F  |  F  |
 *                        | S |   | S |   | S |   |     |  S  |  S  |
 *      =============================================================
 *      lockType          | x   x | x   x | x   x |  x  |     |     |
 *      ------------------|       |       |       |     |     |     |
 *      ownerType         | x   x | x   x | x   x |     |     |     |
 *      ------------------|       |       |       |     |     |     |
 *      cifs tid          | x     | x     | x     |     |  x  |  x  |
 *      ------------------|       |       |       |     |     |     |
 *      cifs fid          | x     | x     | x     |     |  x  |  x  |
 *      ------------------|       |       |       |     |     |     |
 *      cifs pid          | x     | x     | x     |     |  x  |  x  |
 *      ------------------|       |       |       |     |     |     |
 *      nlm clientState   |     x |       |       |     |     |     |
 *      ------------------|       |       |       |     |     |     |
 *      nlm clientIpAddr  |     x |     x |     x |  x  |     |     |
 *      ------------------|       |       |       |     |     |     |
 *      nlm pid           |     x |     x |     x |  x  |     |     |
 *      ------------------|       |       |       |     |     |     |
 *      nlm callerNameLen |     x |       |       |     |     |     |
 *      ------------------|       |       |       |     |     |     |
 *      nlm callerName    |     x |       |       |     |     |     |
 *      ------------------|       |       |       |     |     |     |
 *      nlm ownerNameLen  |     x |       |       |     |     |     |
 *      ------------------|       |       |       |     |     |     |
 *      nlm ownerName     |     x |       |       |     |     |     |
 *      ------------------|       |       |       |     |     |     |
 *      timeout           | x   x |       |       |     |     |     |
 *      ------------------|       |       |       |     |     |     |
 *      offset            | x   x | x   x | x   x |  x  |  x  |  x  |
 *      ------------------|       |       |       |     |     |     |
 *      len               | x   x | x   x | x   x |  x  |  x  |  x  |
 *      ------------------|       |       |       |     |     |     |
 *      ulPrivate         | x   x |     x |     x |     |     |     |
 *      ------------------|       |       |       |     |     |     |
 *      fcb               |     x |       |       |     |     |     |
 *      =============================================================
 *
 *--------------------------------------------------------------------------
 */

/*
 * Result of NLM lock test request
 */
typedef struct lock_testResult {
	uint8	lockState;		/*LOCK_STATE_GRANTED - if the lock can be granted.
							  LOCK_STATE_DENIED - if there is a conflict lock.*/

	/*The following fields are vaild if lockState is LOCK_STATE_DENIED.
	 *They identify the conflicting lock and its owner.*/
	uint8	lockType;		/*LOCK_TYPE_EXCL or LOCK_TYPE_SHAR*/
	int32	pid;
	uint64	offset;
	uint64	len;
	uint16	ownerNameLen;
	char	*ownerName;
} lock_testResult_t;

/*--------------------
 * Function prototypes
 *--------------------
 */

/*-----------------------------------------------------------------
 * Name:        lock_initFileLockReq
 *
 * Description:
 *
 *     Allocate and initialize a new lock_fileLock_t structure.
 *
 * Arguments:    
 * 
 *     fileLockIdxp - Pointer to the index to the newly allocated
 *                    lock_fileLock_t structure.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If failed.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/23/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_initFileLockReq(vs_mem_index_t *fileLockIdxp);

/*-----------------------------------------------------------------
 * Name:         lock_uninitFileLockReq
 *
 * Description:
 *
 *     Uninitialize an existing lock_fileLock_t structure.
 *
 * Arguments:    
 * 
 *     fileLockIdxp - Pointer to the index to a lock_fileLock_t structure.
 *                    If the data structure is uninit successfully, the
 *                    index becomes zero on return.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If failed.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/02/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_uninitFileLockReq(vs_mem_index_t *fileLockIdxp);

/*-----------------------------------------------------------------
 * Name:         lock_byteRangeLockReq
 *
 * Description:
 *
 *     Lock a byte range.
 *
 * Arguments:    
 * 
 *     fileLockIdx - Index to a lock_fileLock_t structure.
 *
 *     lockReqp - Pointer to data structure that contains lock information
 *                as input arguments for byte range lock request.
 *
 * Return value:
 *
 *     LOCK_STATE_GRANTED, LOCK_STATE_DENIED, or LOCK_STATE_BLOCKED -
 *                                                                  If succeed.
 *
 *     lock error code - If failed.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/02/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_byteRangeLockReq(vs_mem_index_t fileLockIdx,
				      lock_request_t *lockReqp);

/*-----------------------------------------------------------------
 * Name:         lock_unlockSingleReq
 *
 * Description:
 *
 *     Unlock a single byte range.
 *
 * Arguments:    
 * 
 *     fileLockIdx - Index to a lock_fileLock_t structure.
 *
 *     lockReqp - Pointer to data structure that contains lock information
 *                as input arguments for unlock a single byte range request.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If failed.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/04/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_unlockSingleReq(vs_mem_index_t fileLockIdx,
				     lock_request_t *lockReqp);

/*-----------------------------------------------------------------
 * Name:         lock_unlockAllByFidReq
 *
 * Description:
 *
 *     Release all locks held on an open file handle (CIFS only).
 *
 * Arguments:    
 * 
 *     fileLockIdx - Index to a lock_fileLock_t structure.
 *  
 *     fid - File ID, representing an open file handle.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If failed.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/06/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_unlockAllByFidReq(vs_mem_index_t fileLockIdx,
					   uint16         fid);

/*-----------------------------------------------------------------
 * Name:         lock_unlockAllByKeyReq
 *
 * Description:
 *
 *     Release all locks held on an open file handle that
 *     match a key (CIFS only).
 *
 * Arguments:    
 * 
 *     fileLockIdx - Index to a lock_fileLock_t structure.
 *
 *     fid - File ID, representing an open file handle.
 *
 *     pid - Process ID.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/06/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_unlockAllByKeyReq(vs_mem_index_t fileLockIdx,
					   uint16         fid,
					   uint16         pid);

/*-----------------------------------------------------------------
 * Name:         lock_unlockAllByIpAddrReq
 *
 * Description:
 *
 *     Release all locks held by a client (NLM only).
 *
 * Arguments:    
 * 
 *     ipAddr - Client IP address.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *  
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/26/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_unlockAllByIpAddrReq(vs_mem_index_t    fileLockIdx,
						  ip_index_t        clientIpAddr);
                                                        
/*-----------------------------------------------------------------
 * Name:         lock_cancelLockReq
 *
 * Description:
 *
 *     Cancel a pending lock.
 *
 * Arguments:    
 * 
 *     fileLockIdx - Index to a lock_fileLock_t structure.
 *
 *     lockReqp - Pointer to data structure that contains lock information
 *                as input arguments for cancel lock request.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/06/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_cancelLockReq(vs_mem_index_t fileLockIdx,
				   lock_request_t *lockReqp);
 
/*-----------------------------------------------------------------
 * Name:         lock_testLockReq
 *
 * Description:
 *
 *     Test to see whether a given lock can be granted or not (NLM only).
 *
 * Arguments:    
 * 
 *     fileLockIdx - Index to a lock_fileLock_t structure.
 *
 *     lockReqp - Pointer to data structure that contains lock information
 *                as input arguments for test lock request.
 *
 *     testResultp - This is output. It contains the pointer to a structure
 *                   that contains testing result.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/27/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_testLockReq(vs_mem_index_t		fileLockIdx,
				 lock_request_t		*lockReqp,
				 lock_testResult_t	**testResultp);

/*-----------------------------------------------------------------
 * Name:         lock_checkLockForWriteReq
 *
 * Description:
 *
 *     Called for write access to a range of bytes (CIFS only).
 *     Called by CIFS write to see if there is a conflicting mandatory lock.
 *
 * Arguments:    
 *  
 *     fileLockIdx - Index to a lock_fileLock_t structure.
 *
 *     lockReqp - Pointer to data structure that contains lock information
 *                as input arguments for request of checking write access
 *                to a range of bytes.
 *
 * Return value:
 *
 *     NFX_OK - If write access is allowed.
 *
 *     NFX_ERR - If write access is denied.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/06/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_checkLockForWriteReq(vs_mem_index_t fileLockIdx,
						  lock_request_t *lockReqp);

/*-----------------------------------------------------------------
 * Name:         lock_checkLockForReadReq
 *
 * Description:
 *
 *     Called for read access to a range of bytes (CIFS only).
 *     Called by CIFS read to see if there is a conflicting mandatory lock.
 *
 * Arguments:    
 * 
 *     fileLockIdx - Index to a lock_fileLock_t structure.
 *
 *     lockReqp - Pointer to data structure that contains lock information
 *                as input arguments for request of checking read access
 *                to a range of bytes.
 *
 * Return value:
 *
 *     NFX_OK - If read access is allowed.
 *
 *     NFX_ERR - If read access is denied.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/06/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_checkLockForReadReq(vs_mem_index_t fileLockIdx,
						 lock_request_t *lockReqp);

/*-----------------------------------------------------------------
 * Name:         lock_areThereCurrentFileLocksReq
 *
 * Description:
 *
 *     Check if there are any locks currently assigned to a file or not.
 *
 * Arguments:    
 * 
 *     fileLockIdx - Index to a lock_fileLock_t structure.
 *
 * Return value:
 *
 *     TRUE - There are lock currently assigned to a file.
 *
 *     FALSE - No lock currently assigned to a file.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/27/2002
 *
 *-----------------------------------------------------------------
 */
boolean
lock_areThereCurrentFileLocksReq(vs_mem_index_t fileLock);

/*-----------------------------------------------------------------
 * Name:         lock_getZeroTimeout
 *
 * Description:
 *
 *     Called by the CIFS or NLM module to get a 0 value timeout
 *     in real time clock form.
 *
 * Arguments:    
 * 
 *     rtcTimeout - Zero timeout in real time clock form.
 *
 * Return value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/21/2003
 *
 *-----------------------------------------------------------------
 */
void
lock_getZeroTimeout(eeeRTC_t	*rtcTimeout);

/*-----------------------------------------------------------------
 * Name:         lock_getIndefiniteTimeout
 *
 * Description:
 *
 *     Called by the CIFS or NLM module to get a timeout value
 *     in real time clock form to represent indefinite timeout.
 *
 * Arguments:    
 * 
 *     rtcTimeout - Indefinite timeout in real time clock form.
 *
 * Return value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 01/21/2003
 *
 *-----------------------------------------------------------------
 */
void
lock_getIndefiniteTimeout(eeeRTC_t	*rtcTimeout);

#endif /* _LOCK_API_H */
