/* vi:ts=4
 *
 *-----------------------------------------------------------------
 *
 * Name:			lock-comp.c
 *
 * RCS:  $Id: $
 *
 * Copyright (C) 2002, ClariStor Inc.
 *
 * Description: Functions related to lock ranges comparision, 
 *              locks conflict checking, read/write access checking,
 *              etc.
 *
 * Created by:		Xiaobing Zhang
 *
 * Date Created:	12/03/2002
 *
 *-----------------------------------------------------------------
 */

#include "lock-incls.h"

/*-----------------------------------------------------------------
 * Name:         lock_rangeComp
 *
 * Description:
 *
 *     Compare positions of two locked ranges.
 *
 *   The two ranges, lock1 and lock2, may overlap, or one range has
 *   lower address in the file than the other. If they overlap, the
 *   result is 0; if lock1's position is lower than lock2's, the result
 *   is -1; otherwise the result is 1.
 *
 *   We define the following,
 *      uint64  s1;    //Start of lock1's lock range
 *      uint64  e1;    //End of lock1's lock range
 *      uint64  l1;    //Length of lock1's lock range
 *      uint64  s2;    //Start of lock2's lock range
 *      uint64  e2;    //End of lock2's lock range
 *      uint64  l2;    //Length of lock2's lock range
 *
 *   Note: The length of an NLM lock is always greater than 0.
 *
 *   Here are the rules for comparing two lock ranges:
 *      =============================================================
 *        l1   |  l2   |  return -1  |   return 0     |   return 1
 *      =======+=======+=============+================+==============
 *   1.  != 0  | != 0  |  (e1 < s2)  |  (s1 <= e2) && |  (e2 < s1)
 *             |       |             |  (s2 <= e1)    |
 *      -------+-------+-------------+----------------+--------------
 *      CIFS specific, when lock length is 0:
 *   2.   = 0  |  = 0  |  (s1 < s2)  | never overlap  |  (s1 >= s2)
 *      -------+-------+-------------+----------------+--------------
 *   3.   = 0  | != 0  |  (s1 <= s2) |  (s1 > s2) &&  |  (s1 >= e2)
 *             |       |             |  (s1 < e2)     |
 *      -------+-------+-------------+----------------+--------------
 *   4.  != 0  |  = 0  |  (e1 <= s2) |  (s2 > s1) &&  |  (s1 >= s2)
 *             |       |             |  (s2 < e1)     |
 *      -------+-------+-------------+----------------+--------------
 *
 *
 * Arguments:
 *
 *     start1 - Lock1's offset.
 *
 *     length1 - Lock1's length.
 *
 *     start2 - Lock2's offset.
 *
 *     length2 - Lock2's length.
 *
 * Return value:
 *
 *     -1 - If lock1's region has lower address than lock2's region.
 *      0 - If lock1's region and lock2's region overlap.
 *      1 - If lock1's region has higher address than lock2's region.
 *      2 - If there is error.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 11/25/2002
 *
 *-----------------------------------------------------------------
 */
int8
lock_rangeComp(uint64	start1,
			   uint64	length1,
			   uint64	start2,
			   uint64	length2)
{
	uint64  end1, end2;

	LOCK_GET_LOCK_RANGE_END(start1, length1, end1);
	LOCK_GET_LOCK_RANGE_END(start2, length2, end2);

	if ((length1 != 0) && (length2 != 0))
	{
		/*
		 * Rule 1
		 */
		if (end1 < start2)
			return (-1);
		else if (end2 < start1)
			return (1);
		else
			/* Overlap */
			return (0);
	}

	if ((length1 == 0) && (length2 == 0))
	{
		/*
		 * Rule 2
		 */
		if (start1 < start2)
			return (-1);
		else
			return (1);
	}
	else if (length1 == 0)
	{
		/*
		 * Rule 3
		 */
		if (start1 <= start2)
			return (-1);
		else if (start1 >= end2)
			return (1);
		else
			/* Overlap */
			return (0);
	}
	else if (length2 == 0)
	{
		/*
		 * Rule 4
		 */
		if (end1 <= start2)
			return (-1);
		else if (start1 >= start2)
			return (1);
		else
			/* Overlap */
			return (0);
	}
	else
		return (2);
} /* lock_rangeComp() */

/*-----------------------------------------------------------------
 * Name:         lock_lockInfoComp
 *
 * Description:
 *
 *     Compare two locks given searchMask.
 *     If searchMask is 0, only check overlapping lock ranges.
 *
 * Arguments:    
 * 
 *     lockInfoIdx1 - Index to lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                    structure. It has lock info of first lock.
 *
 *     lockInfop1 - Pointer to the above lock info structure.
 *
 *         One of the above two paramenters is valid, not both.
 *
 *     lockInfoIdx2 - Index to lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                    structure. It has lock info of second lock.
 *
 *     lockOwnerType1 - The type of owner of the first lock.
 *
 *     lockOwnerType1 - The type of owner of the first lock.
 *
 *     lockOffset1 - The first lock's offset.
 *
 *     lockOffset2 - The second lock's offset.
 *
 *     lockLen1 - The first lock's length.
 *
 *     locklen2 - The second lock's length.
 *
 *     searchMask - The criteria based on which we compare two locks.
 *
 *     specialCompResult - 1 or -1. Comparison result if the lock
 *                         ranges overlap but the criteria is not
 *                         satisfied.
 *
 * Return value:
 *
 *     -1 - If inNode(or inLockInfo) < root of new tree.
 *      0 - If inNode(or inLockInfo) == root of new tree.
 *      1 - If inNode(or inLockInfo) > root of new tree.
 *      2 - If there is error.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/03/2002
 *
 *-----------------------------------------------------------------
 */
int8
lock_lockInfoComp(vs_mem_index_t	lockInfoIdx1,
				  void				*lockInfop1,
				  vs_mem_index_t	lockInfoIdx2,
				  uint8				lockOwnerType1,
			   	  uint8				lockOwnerType2,
				  uint64			lockOffset1,
				  uint64			lockOffset2,
				  uint64			lockLen1,
				  uint64			lockLen2,
				  uint16			searchMask,
				  int8				specialCompResult)
{
	int8			result;

	/*-----------------------------------
	 * Check if two ranges overlap or not
	 *-----------------------------------
	 */
	if ((searchMask & LOCK_SEARCH_SAME_RANGE) &&
		(lockLen2 == 0) &&
	   	(lockLen2 == 0) &&
		(lockOffset1 == lockOffset2))
	{
		/* Special treatement for two CIFS 0-length locks */
		result = 0;
	}
	else
	{
		result = lock_rangeComp(lockOffset1,
							    lockLen1,
							    lockOffset2,
							    lockLen2);
	}

	if (result != 0)
		return (result);

	/*---------------------------------------
	 * Further compare two overlapping ranges
	 *---------------------------------------
	 */
	if (searchMask & LOCK_SEARCH_SAME_RANGE)
	{
		/* Search for a lock in the tree
		 * so that two locks have the same range */
		if (!((lockOffset1 == lockOffset2) && (lockLen1 == lockLen2)))
			result = 1;
	}

	if (searchMask & LOCK_SEARCH_DIFF_RANGE)
	{
		/* Search for a lock in the tree
		 * so that two locks do not have the same range */
		if ((lockOffset1 == lockOffset2) && (lockLen1 == lockLen2))
			result = specialCompResult;
	}

	if (searchMask & LOCK_SEARCH_SAME_OWNER)
	{
		/* Search for a lock in the tree
		 * so that two locks have the same owner */
		if (!((lockOwnerType1 == lockOwnerType2) &&
			  (lock_haveSameOwner(lockInfoIdx1,
								  lockInfop1,
								  lockInfoIdx2,
								  lockOwnerType1,
								  lockOwnerType2) == TRUE)))
			result = specialCompResult;
	}

	if (searchMask & LOCK_SEARCH_DIFF_OWNER)
	{
		/* Search for a lock in the tree
		 * so that two locks have different owners */
		if ((lockOwnerType1 == lockOwnerType2) &&
			(lock_haveSameOwner(lockInfoIdx1,
							   	lockInfop1,
							   	lockInfoIdx2,
								lockOwnerType1,
								lockOwnerType2) == TRUE))
			result = specialCompResult;
	}

	if (searchMask & LOCK_SEARCH_MANDATORY_LOCK)
	{
		/* Search for mandatory lock in a tree */
		if (!LOCK_IS_MANDATORY(lockOwnerType2))
			result = specialCompResult;
	}

	if (searchMask & LOCK_SEARCH_NONZERO_LEN)
	{
		/* Search for lock with non-zero length in a tree */
		if (lockLen2 == 0)
			result = specialCompResult;
	}

	/*TODO: Add more searchMask checking if needed.*/

	return (result);
} /* lock_lockInfoComp() */

/*-----------------------------------------------------------------
 * Name:         lock_twoLockInfoCompFunc
 *
 * Description:
 *
 *     Compare lock ranges of given two lockInfo structures.
 *
 * Arguments:    
 * 
 *     lockInfoIdx1 - Index to lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                    structure. It has lock info of first lock.
 *
 *     lockInfop1 - Pointer to the above lock info structure.
 *
 *     lockOwnerType1 - The type of owner of the first lock.
 *
 *         As to the above three arguments, the caller either,
 *             - provides lockInfoIdx1 and
 *               leaves lockInfop1 and lockOwnerType1 invalid, or
 *             - provides lockInfop1 and lockOwnerType1 and
 *               leaves lockInfoIdx1 invalid.
 *
 *     lockInfoIdx2 - Index to lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                    structure. It has lock info of second lock.
 *
 *     searchMask - The criteria based on which we compare two locks.
 *
 *     specialCompResult - 1 or -1. Comparison result if the lock
 *                         ranges overlap but the criteria is not
 *                         satisfied.
 *
 * Return value:
 *
 *     -1 - If inNode(or inLockInfo) < root of new tree.
 *      0 - If inNode(or inLockInfo) == root of new tree.
 *      1 - If inNode(or inLockInfo) > root of new tree.
 *      2 - If there is error.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/16/2002
 *
 *-----------------------------------------------------------------
 */
int8
lock_twoLockInfoCompFunc(vs_mem_index_t	lockInfoIdx1,
						 void			*lockInfop1,
						 uint8			lockOwnerType1,
						 vs_mem_index_t	lockInfoIdx2,
						 uint16			searchMask,
						 int8			specialCompResult)
{
	int32		rc;
	int8		result;
	uint8		lockOwnerType2;
	uint64		lockOffset1, lockOffset2;
	uint64		lockLen1, lockLen2;

	/*----------------
	 * Get lock ranges
	 *----------------
	 */
	rc = lock_getLockInformationFromLockInfo(lockInfoIdx1,
											 lockInfop1,
											 &lockOwnerType1,
											 &lockOffset1,
											 &lockLen1);
	if (rc != NFX_OK)
	{
		return (2);
	}

	rc = lock_getLockInformationFromLockInfo(lockInfoIdx2,
											 NULL,
											 &lockOwnerType2,
											 &lockOffset2,
											 &lockLen2);
	if (rc != NFX_OK)
	{
		return (2);
	}

	/*------------------
	 * Compare two locks
	 *------------------
	 */
	result = lock_lockInfoComp(lockInfoIdx1,
							   lockInfop1,
							   lockInfoIdx2,
							   lockOwnerType1,
							   lockOwnerType2,
							   lockOffset1,
							   lockOffset2,
							   lockLen1,
							   lockLen2,
							   searchMask,
							   specialCompResult);

	LOCK_PRINTF(LOCK_DEBUG2, ("lock_twoLockInfoCompFunc: result = %d", result));
	return (result);
} /* lock_twoLockInfoCompFunc() */

/*-----------------------------------------------------------------
 * Name:         lock_sharedTwoNodesCompFunc
 *
 * Description:
 *
 *     Compare lock ranges of two nodes in a shared lock btree.
 *
 * Arguments:    
 * 
 *     inShNodep - Pointer to input node.
 *
 *     treeShNodep - Pointer to a node in the tree.
 *
 *     searchMask - The criteria based on which we compare two nodes.
 *
 *     specialCompResult - 1 or -1. Comparison result if the lock
 *                         ranges overlap but the criteria is not
 *                         satisfied.
 *
 * Return value:
 *
 *     -1 - If inNode(or inLockInfo) < root of new tree.
 *      0 - If inNode(or inLockInfo) == root of new tree.
 *      1 - If inNode(or inLockInfo) > root of new tree.
 *      2 - If there is error.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/03/2002
 *
 *-----------------------------------------------------------------
 */
int8
lock_sharedTwoNodesCompFunc(lock_shNode_t	*inShNodep,
							lock_shNode_t	*treeShNodep,
							uint16			searchMask,
							int8			specialCompResult)	
{
	int32			rc;
	int8			result;
	uint64			inLockOffset, treeLockOffset;
	uint64			inLockLen, treeLockLen;
	uint8			dummyLockOwnerType;
	vs_mem_index_t	dummyLockInfoIdx;

	/*--------------------
	 * Get input lock info
	 *--------------------
	 */
	rc = lock_getLockInformationFromShNode(inShNodep,
										   &inLockOffset,
										   &inLockLen);
	if (rc != NFX_OK)
	{
		return (2);
	}

	/*-------------------------
	 * Get info of lock in tree
	 *-------------------------
	 */
	rc = lock_getLockInformationFromShNode(treeShNodep,
				   						   &treeLockOffset,
										   &treeLockLen);
	if (rc != NFX_OK)
	{
		return (2);
	}

	/*------------------
	 * Compare two nodes
	 *-------------------
	 */
	/* Make sure searchMask doesn't require lock owner checking */
	if ((searchMask & LOCK_SEARCH_SAME_OWNER) ||
		(searchMask & LOCK_SEARCH_DIFF_OWNER) ||
		(searchMask & LOCK_SEARCH_MANDATORY_LOCK))
	{
		LOCK_PRINTF(LOG_WARNING, ("lock_sharedTwoNodesCompFunc: "
								"Will not check lock owner."));
		searchMask |= ~(LOCK_SEARCH_SAME_OWNER |
					   	LOCK_SEARCH_DIFF_OWNER |
						LOCK_SEARCH_MANDATORY_LOCK);
	}

	VS_MEM_IDX_SET_NULL(dummyLockInfoIdx);
	dummyLockOwnerType = LOCK_OWNER_DUMMY;
	result = lock_lockInfoComp(dummyLockInfoIdx,
							   NULL,
							   dummyLockInfoIdx,
							   dummyLockOwnerType,
							   dummyLockOwnerType,
							   inLockOffset,
							   treeLockOffset,
							   inLockLen,
							   treeLockLen,
							   searchMask,
							   specialCompResult);
	if (result != 0)
	{
		uint64  inLockEnd, treeLockEnd;

		LOCK_GET_LOCK_RANGE_END(inLockOffset, inLockLen, inLockEnd);
		LOCK_GET_LOCK_RANGE_END(treeLockOffset, treeLockLen, treeLockEnd);

		if ((treeLockLen == 0) || (inLockLen == 0))
		{
			/*----------------------------------------------------------
			 * Special checking for
			 *   - the node with only zero-length locks
			 *   - the zero-length input node
			 *----------------------------------------------------------
			 */
			/* If the 0-length inNode is on the edge of the treeNode's range,
			 * or vise versa, result is 0.
			 */
			if ((inLockOffset == treeLockOffset) ||
				(inLockEnd == treeLockEnd))
					result = 0;

			/*But if we are searching for nodes with different range, and
			 *the two 0-length nodes here have same offset, result is not 0
			 */
			if ((searchMask & LOCK_SEARCH_DIFF_RANGE) &&
				(treeLockLen == 0) && (inLockLen == 0) &&
				(treeLockOffset == inLockOffset))
					result = specialCompResult;
		}
	}

	return (result);
} /* lock_sharedTwoNodesCompFunc() */

/*-----------------------------------------------------------------
 * Name:         lock_sharedLockInfoAndNodeCompFunc
 *
 * Description:
 *
 *     Compare lock ranges of a lock and a node in a shared lock btree.
 *
 * Arguments:    
 * 
 *     inLockInfop - Pointer to a lock_cifsLockInfo_t or 
 *                   lock_nlmLockInfo_t structure. It contains lock
 *                   information, which we compare a tree node with.
 *
 *     inLockOwnerType - Type of lock owner.
 *                       Value is LOCK_OWNER_CIFS or LOCK_OWNER_NLM.
 *
 *     treeShNodep - Pointer to a node in the tree.
 *
 *     searchMask - The criteria based on which we compare the lockInfo
 *                  with a node in the tree.
 *
 *     specialCompResult - 1 or -1. Comparison result if the lock
 *                         ranges overlap but the criteria is not
 *                         satisfied.
 *
 *     foundShLockpp - This is output. The value is pointer to the
 *                     found matching shared lock if search goes into
 *                     the linked lock list.
 *
 *     prevShLockIdxp - This is output. The value is index to the lock
 *                      before the found one in the list.
 *
 * Return value:
 *
 *     -1 - If inNode(or inLockInfo) < root of new tree.
 *      0 - If inNode(or inLockInfo) == root of new tree.
 *      1 - If inNode(or inLockInfo) > root of new tree.
 *      2 - If there is error.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/03/2002
 *
 *-----------------------------------------------------------------
 */
int8
lock_sharedLockInfoAndNodeCompFunc(void				*inLockInfop,
								   uint8			inLockOwnerType,
								   lock_shNode_t	*treeShNodep,
								   uint16			searchMask,
								   int8				specialCompResult,
								   lock_shLock_t	**foundShLockpp,
								   vs_mem_index_t	*prevShLockIdxp)
{
	int32			rc;
	int8			result, result2;
	uint8			treeLockOwnerType, dummyLockOwnerType;
	uint64			inLockOffset, treeLockOffset;
	uint64			inLockLen, treeLockLen;
	uint16			searchMask2;
	vs_mem_index_t	shLockIdx;
	lock_shLock_t	*shLockp;
	vs_mem_index_t	dummyLockInfoIdx;

	*foundShLockpp = NULL;
	VS_MEM_IDX_SET_NULL(*prevShLockIdxp);

	VS_MEM_IDX_SET_NULL(dummyLockInfoIdx);

	if (searchMask & LOCK_SEARCH_MANDATORY_LOCK)
	{
		if (treeShNodep->lockCntAndHoley.mandatoryLockCnt == 0)
			return (1);
	}

	/*--------------------
	 * Get input lock info
	 *--------------------
	 */
	rc = lock_getLockInformationFromLockInfo(dummyLockInfoIdx,
											 inLockInfop,
											 &inLockOwnerType,
											 &inLockOffset,
											 &inLockLen);
	if (rc != NFX_OK)
		return (2);

	/*-------------------------------
	 * Get lock range of node in tree
	 *-------------------------------
	 */
	rc = lock_getLockInformationFromShNode(treeShNodep,
										   &treeLockOffset,
										   &treeLockLen);
	if (rc != NFX_OK)
	{
		return (2);
	}

	/*---------------------------------
	 * Compare the lockInfo with shNode
	 *---------------------------------
	 */
	searchMask2 = searchMask;

	if (searchMask2 & LOCK_SEARCH_IN_SHARED_LIST)
	{
		/* At this moment, only check if lockInfo and shNode have
		 * overlapping ranges or not.
		 * Other checking will be done when comparing the lockInfo
		 * with each lock in the lock list of the found node.*/
		searchMask2 = 0;
	}
	else if ((searchMask2 & LOCK_SEARCH_SAME_OWNER) ||
			 (searchMask2 & LOCK_SEARCH_DIFF_OWNER) ||
			 (searchMask2 & LOCK_SEARCH_MANDATORY_LOCK))
	{
		/* Make sure searchMask2 doesn't require lock owner checking */
		LOCK_PRINTF(LOG_WARNING, ("lock_sharedLockInfoAndNodeCompFunc: "
								"Will not check lock owner."));
		searchMask2 |= ~(LOCK_SEARCH_SAME_OWNER |
					     LOCK_SEARCH_DIFF_OWNER |
						 LOCK_SEARCH_MANDATORY_LOCK);
	}

	dummyLockOwnerType = LOCK_OWNER_DUMMY;
	result = lock_lockInfoComp(dummyLockInfoIdx,
							   inLockInfop,
							   dummyLockInfoIdx,
							   inLockOwnerType,
							   dummyLockOwnerType,
							   inLockOffset,
							   treeLockOffset,
							   inLockLen,
							   treeLockLen,
							   searchMask2,
							   specialCompResult);
	if (result != 0)
	{
		uint64	inLockEnd, treeLockEnd;

		LOCK_GET_LOCK_RANGE_END(inLockOffset, inLockLen, inLockEnd);
		LOCK_GET_LOCK_RANGE_END(treeLockOffset, treeLockLen, treeLockEnd);

		if ((treeLockLen == 0) || (inLockLen == 0))
		{
			/*----------------------------------------------------------
			 * Special checking for
			 *   - the node with only zero-length locks
			 *   - the zero-length input lock
			 *----------------------------------------------------------
			 */
			/* If
			 *   -the only locks in the node and the input lock are
			 *    both zero-lengh and have same offset, or
			 *   -If the 0-length lock is on the edge of the node's range, or
			 *   -If the 0-length node is on the edge of the input lock's range,
			 * then the result is 0.
			 */
			if ((inLockOffset == treeLockOffset) ||
				(inLockEnd == treeLockEnd))
				result = 0;
		}

		if (result != 0)
			return (result);
	}
	
	if (searchMask & LOCK_SEARCH_IN_SHARED_LIST)
	{
		/*--------------------------------------------------
		 * Compare the lockInfo with each shLock in the list
		 *--------------------------------------------------
		 */
		result = 1;

		shLockIdx = treeShNodep->lockList;
		while (!VS_MEM_IDX_IS_NULL(shLockIdx))
		{
			shLockp = (lock_shLock_t *)vs_mem_idxToPtr(shLockIdx);
			if (shLockp == NULL)
			{
				LOCK_PRINTF(LOG_ERR, ("lock_sharedLockInfoAndNodeCompFunc: "
							"Couldn't find shLock data structure."));
				return (2);
			}

			/* Get lock info of the shared lock */
			rc = lock_getLockInformationFromLockInfo(shLockp->info,
													 NULL,
													 &treeLockOwnerType,
													 &treeLockOffset,
													 &treeLockLen);

			/* Compare two locks */
			result2 = lock_lockInfoComp(dummyLockInfoIdx,
										inLockInfop,
										shLockp->info,
										inLockOwnerType,
										treeLockOwnerType,
										inLockOffset,
										treeLockOffset,
										inLockLen,
										treeLockLen,
										searchMask,
										specialCompResult);
			if (result2 == 0)
			{
				/* Found the overlapping lock */
				result = 0;
				*foundShLockpp = shLockp;
				break;
			}

			*prevShLockIdxp = shLockIdx;
			shLockIdx = shLockp->link;
		} /* while() */
	}

	return (result);
} /* lock_sharedLockInfoAndNodeCompFunc() */

/*-----------------------------------------------------------------
 * Name:         lock_exclCompFunc
 *
 * Description:
 *
 *     Compare lock ranges of an input lock and an node in 
 *     an exclusive lock btree.
 *
 * Arguments:    
 * 
 *     inExLockp - Pointer to the input node.
 *
 *     inLockInfop - Pointer to a lock_cifsLockInfo_t or 
 *                   lock_nlmLockInfo_t structure. It contains lock
 *                   information, which we compare a tree node with.
 *
 *     inLockOwnerType - Type of lock owner.
 *                       Value is LOCK_OWNER_CIFS or LOCK_OWNER_NLM.
 *
 *     useNode - TRUE if inShNodep is valid.
 *               FALSE if inLockInfop and inLockOwnerType are valid.
 *
 *     treeExLockp - Pointer to a node in the tree.
 *
 *     searchMask - The criteria based on which we compare two nodes
 *                  or compare the lockInfo with a node in the tree.
 *
 *     specialCompResult - 1 or -1. Comparison result if the lock
 *                         ranges overlap but the criteria is not
 *                         satisfied.
 *
 * Return value:
 *
 *     -1 - If inNode(or inLockInfo) < root of new tree.
 *      0 - If inNode(or inLockInfo) == root of new tree.
 *      1 - If inNode(or inLockInfo) > root of new tree.
 *      2 - If there is error.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/03/2002
 *
 *-----------------------------------------------------------------
 */
int8
lock_exclCompFunc(lock_exLock_t	*inExLockp,
				  void			*inLockInfop,
				  uint8			inLockOwnerType,
				  boolean       useNode,
				  lock_exLock_t *treeExLockp,
				  uint16        searchMask,
				  int8			specialCompResult)
{
	int32			rc;
	int8			result;
	vs_mem_index_t	inLockInfoIdx;
	uint8			treeLockOwnerType;
	uint64			inLockOffset, treeLockOffset;
	uint64			inLockLen, treeLockLen;

	/*---------------------
	 * Get input lock range
	 *---------------------
	 */
	if (useNode == TRUE)
	{
		inLockInfoIdx = inExLockp->info;
	}
	else
		VS_MEM_IDX_SET_NULL(inLockInfoIdx);

	rc = lock_getLockInformationFromLockInfo(inLockInfoIdx,
											 inLockInfop,
											 &inLockOwnerType,
											 &inLockOffset,
											 &inLockLen);
	if (rc != NFX_OK)
	{
		return (2);
	}

	/*--------------------------
	 * Get range of lock in tree
	 *--------------------------
	 */
	rc = lock_getLockInformationFromLockInfo(treeExLockp->info,
											 NULL,
											 &treeLockOwnerType,
											 &treeLockOffset,
											 &treeLockLen);
	if (rc != NFX_OK)
	{
		return (2);
	}

	/*------------------
	 * Compare two locks
	 *------------------
	 */
	result = lock_lockInfoComp(inLockInfoIdx,
							   inLockInfop,
							   treeExLockp->info,
							   inLockOwnerType,
							   treeLockOwnerType,
							   inLockOffset,
							   treeLockOffset,
							   inLockLen,
							   treeLockLen,
							   searchMask,
							   specialCompResult);

	if ((result == 0) && (useNode == TRUE) && (inLockLen == 0))
	{
		/*------------------------------------
		 * Special processing of 0-length lock
		 *------------------------------------
		 */
		/* It is possible that the tree contains more than one 0-length
		 * locks with same offset and same owner. When we find such a
		 * lock and want to delete it, this piece of code makes sure
		 * lock_btree_delete will delete the right one.
		 * This can happen when we are trying to delete locks whose
		 * owner match a fid or key.
		 */
		if (!VS_MEM_IDX_IS_EQUAL(inExLockp->info, treeExLockp->info))
			result = -1; /*XXX: Should be -1 or not?*/
	}
	else if ((result != 0) && (inLockLen == 0) && (treeLockLen == 0) &&
			  (searchMask & LOCK_SEARCH_EXCL_BOTH_0LEN))
		result = 0;

	LOCK_PRINTF(LOCK_DEBUG2, ("lock_exclCompFunc: result = %d", result));
	return (result);
} /* lock_exclCompFunc() */

/*-----------------------------------------------------------------
 * Name:         lock_sharedCompFunc
 *
 * Description:
 *
 *     Compare lock ranges of an input lock and a node in a shared
 *     lock btree.
 *
 * Arguments:    
 * 
 *     inShNodep - Pointer to the input node.
 *
 *     inLockInfop - Pointer to a lock_cifsLockInfo_t or 
 *                   lock_nlmLockInfo_t structure. It contains lock
 *                   information, which we compare a tree node with.
 *
 *     inLockOwnerType - Type of lock owner.
 *                       Value is LOCK_OWNER_CIFS or LOCK_OWNER_NLM.
 *
 *     useNode - TRUE if inShNodep is valid.
 *               FALSE if inLockInfop and inLockOwnerType are valid.
 *
 *     treeShNodep - Pointer to a node in the tree.
 *
 *     searchMask - The criteria based on which we compare two nodes
 *                  or compare the lockInfo with a node in the tree.
 *
 * Return value:
 *
 *     -1 - If inNode(or inLockInfo) < root of new tree.
 *      0 - If inNode(or inLockInfo) == root of new tree.
 *      1 - If inNode(or inLockInfo) > root of new tree.
 *      2 - If there is error.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/03/2002
 *
 *-----------------------------------------------------------------
 */
int8
lock_sharedCompFunc(lock_shNode_t	*inShNodep,
					void			*inLockInfop,
					uint8			inLockOwnerType,
					boolean         useNode,
					lock_shNode_t   *treeShNodep,
					uint16          searchMask,
					int8			specialCompResult)	
{
	int8			result;

	if (useNode == TRUE)
	{
		/* Compare two shared lock node */
		result = lock_sharedTwoNodesCompFunc(inShNodep,
											 treeShNodep,
											 searchMask,
											 specialCompResult);
	}
	else
	{
		lock_shLock_t	*dummyShLockp;
		vs_mem_index_t	dummyShLockIdx;

		/* Compare lockInfo with a shared lock node */
		result = lock_sharedLockInfoAndNodeCompFunc(
											inLockInfop,
											inLockOwnerType,
											treeShNodep,
											searchMask,
											specialCompResult,
											&dummyShLockp,		/*Don't care*/
											&dummyShLockIdx);	/*Don't care*/
	}

	LOCK_PRINTF(LOCK_DEBUG2, ("lock_sharedCompFunc: result = %d", result));
	return (result);
} /* lock_sharedCompFunc() */

/*-----------------------------------------------------------------
 * Name:         lock_compFunc
 *
 * Description:
 *
 *     Compare lock ranges of an input lock and a node in a btree.
 *
 * Arguments:    
 * 
 *     inNodep - Pointer to the input node.
 *
 *     inLockInfop - Pointer to a lock_cifsLockInfo_t or 
 *                   lock_nlmLockInfo_t structure. It contains lock
 *                   information, which we compare a tree node with.
 *
 *     inLockOwnerType - Type of lock owner.
 *                       Value is LOCK_OWNER_CIFS or LOCK_OWNER_NLM.
 *
 *     useNode - TRUE if inNodep is valid.
 *               FALSE if inLockInfop and inLockOwnerType are valid.
 *
 *     treeNodep - Pointer to a node in the tree.
 *
 *     searchMask - The criteria based on which we compare two nodes
 *                  or compare the lockInfo with a node in the tree.
 *
 *     specialCompResult - 1 or -1. Comparison result if the lock
 *                         ranges overlap but the criteria is not
 *                         satisfied.
 *
 *     treeType - Exclusive or shared lock tree.
 *                Value is LOCK_TYPE_EXCL or LOCK_TYPE_SHAR.
 *
 * Return value:
 *
 *     -1 - If inNode(or inLockInfo) < root of new tree.
 *      0 - If inNode(or inLockInfo) == root of new tree.
 *      1 - If inNode(or inLockInfo) > root of new tree.
 *      2 - If there is error.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/03/2002
 *
 *-----------------------------------------------------------------
 */
int8
lock_compFunc(void         *inNodep,
			  void         *inLockInfop,
			  uint8        inLockOwnerType,
			  boolean      useNode,	
			  void         *treeNodep,
			  uint16       searchMask,
			  int8         specialCompResult,
			  uint8        treeType)
{
	int8	result;

	if (treeType == LOCK_TYPE_EXCL)
	{
		lock_exLock_t	*inExLockp, *treeExLockp;

		inExLockp = (lock_exLock_t *)inNodep;
		treeExLockp = (lock_exLock_t *)treeNodep;
		result = lock_exclCompFunc(inExLockp, inLockInfop, inLockOwnerType,
								   useNode, treeExLockp,
								   searchMask, specialCompResult);
	}
	else
	{
		lock_shNode_t	*inShNodep, *treeShNodep;

		inShNodep = (lock_shNode_t *)inNodep;
		treeShNodep = (lock_shNode_t *)treeNodep;
		result = lock_sharedCompFunc(inShNodep, inLockInfop, inLockOwnerType,
					   				 useNode, treeShNodep,
									 searchMask, specialCompResult);
	}

	return (result);
} /* lock_compFunc() */

/*-----------------------------------------------------------------
 * Name:         lock_checkUnlockedRange
 *
 * Description:
 *
 *     Check whether a lock range partially remains locked 
 *     if an unlock range overlaps with the lock range (NLM only).
 *
 *    The diagrams below shows all the possible relationships between
 *    the lock and unlock ranges.
 *
 *                                                 The remainnig range:
 *    (1).
 *      lock range:     *-----------*       ===>      *--*     *--*
 *      unlock range:      *-----*              (lower part)  (upper part)
 *
 *    (2).
 *      lock range:        *-----*          ===>      nothing
 *      unlock range:   *-----------*
 *
 *      lock range:       *--------*        ===>      nothing
 *      unlock range:     *--------*
 *
 *      lock range:        *-----*          ===>      nothing
 *      unlock range:      *--------*
 *
 *      lock range:        *-----*          ===>      nothing
 *      unlock range:   *--------*
 *
 *    (3).
 *      lock range:      *--------*         ===>      *--*
 *      unlock range:       *--------*             (lower part)
 *
 *      lock range:      *--------*         ===>      *--*
 *      unlock range:       *-----*                (lower part)
 *
 *    (4).
 *      lock range:        *--------*       ===>                *--*
 *      unlock range:   *--------*                            (upper part)
 *
 *      lock range:        *--------*       ===>                *--*
 *      unlock range:      *-----*                            (upper part)
 *
 *
 *     Define the following,
 *         slock   : start of lock range
 *         elock   : end of lock range
 *         sunlock : start of unlock range
 *         eunlock : end of unlock range
 *
 *           ==================================================
 *                             | lower part of | upper part of
 *                             | lock range    | lock range
 *                             | remains       | remains
 *           ===================================================
 *      (1).  slock < sunlock  |      > 0      |      > 0
 *            elock > eunlock  |               | 
 *           ---------------------------------------------------
 *      (2).  slock >= sunlock |      = 0      |      = 0
 *            elock <= eunlock |               | 
 *           ---------------------------------------------------
 *      (3).  slock < sunlock  |      > 0      |      = 0
 *            elock <= eunlock |               | 
 *           ---------------------------------------------------
 *      (4).  slock >= sunlock |      = 0      |      > 0
 *            elock > eunlock  |               | 
 *           ---------------------------------------------------
 *
 * Arguments:    
 * 
 *     lockInfop - Pointer to a lock_cifsLockInfo_t or 
 *                 lock_nlmLockInfo_t structure. It contains a
 *                 locked byte range.
 *
 *     unlockInfop - Pointer to a lock_cifsLockInfo_t or 
 *                  lock_nlmLockInfo_t structure. It contains a
 *                  byte range to be unlocked.
 *
 *     lowerOffsetp - This is output. The offset of lower part of
 *                    remaining unlocked range.
 *
 *     lowerLenp - This is output. The length of lower part of
 *                 remaining unlocked range.
 *
 *     upperOffsetp - This is output. The offset of lower part of
 *                    remaining unlocked range.
 *
 *     upperLenp - This is output. The length of lower part of
 *                 remaining unlocked range.
 *
 * Return value:
 *
 *     None. 
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/19/2002
 *-----------------------------------------------------------------
 */
void
lock_checkUnlockedRange(lock_nlmLockInfo_t	*lockInfop,
						lock_nlmLockInfo_t	*unlockInfop,
						uint64				*lowerOffsetp,
						uint64				*lowerLenp,	
						uint64				*upperOffsetp,
						uint64				*upperLenp)
{
	uint64	sLock, sUnlock, eLock, eUnlock;

	sLock = lockInfop->offset;
	LOCK_GET_LOCK_RANGE_END(lockInfop->offset, lockInfop->len, eLock);
	sUnlock = unlockInfop->offset;
	LOCK_GET_LOCK_RANGE_END(unlockInfop->offset, unlockInfop->len, eUnlock);

	*lowerOffsetp = 0;
	*lowerLenp = 0;
	*upperOffsetp = 0;
	*upperLenp = 0;

	if ((sLock < sUnlock) && (eLock > eUnlock))
	{
		/* Two sub lock ranges remain */
		*lowerOffsetp = sLock;
		*lowerLenp = sUnlock - sLock;
		*upperOffsetp = eUnlock + 1;
		*upperLenp = eLock - eUnlock;
	}
	else if ((sLock >= sUnlock) && (eLock <= eUnlock))
	{
		/* The whole lock range is unlocked */
	}
	else if ((sLock < sUnlock) && (eLock <= eUnlock))
	{
		/* Lower part of lock range remain */
		*lowerOffsetp = sLock;
		*lowerLenp = sUnlock - sLock;
	}
	else if ((sLock >= sUnlock) && (eLock > eUnlock))
	{
		/* Upper part of lock range remain */
		*upperOffsetp = eUnlock + 1;
		*upperLenp = eLock - eUnlock;
	}
} /* lock_checkUnlockedRange */

/*-----------------------------------------------------------------------------
      Define rules for granting, denying, or blocking a new lock if it 
	overlaps with one in the exclusive or/and shared tree.

      Logic table:
      Explanation of each column in the table:

        protocol:      The protocol of the lock to be added.
        add lock type: The type of the lock to be added.
        existing lock type: The type of the existing lock being compared with
                            the lock to be added.
        owner:         Whether the two locks have the same owner.
        block:         Whether the lock to be added wants to wait if the lock
                       cannot be granted.
        result:        Indicate whether the lock can be added or not if the
                       condition in that row is satisfied.

      If conditions represented by two or more rows are satisfied at the same
    time, and not all the results are GRANTED, then the return value should
    be either DENIED or BLOCKED.

          ======================================================
                    | add    | existing|       |       |
                    | lock   | lock    |       |       |
           protocol | type   | type    | owner | block |  result
          ======================================================
        1.   CIFS   | excl   | excl    | same  | yes   | BLOCKED
        2.   CIFS   | excl   | excl    | same  | no    | DENIED
        3.   CIFS   | excl   | excl    | diff  | yes   | BLOCKED
        4.   CIFS   | excl   | excl    | diff  | no    | DENIED
      
        5.   CIFS   | excl   | shared  | same  | yes   | BLOCKED
        6.   CIFS   | excl   | shared  | same  | no    | DENIED
        7.   CIFS   | excl   | shared  | diff  | yes   | BLOCKED
        8.   CIFS   | excl   | shared  | diff  | no    | DENIED

        9.   CIFS   | shared | excl    | same  | yes   | GRANTED
        10.  CIFS   | shared | excl    | same  | no    | GRANTED
        11.  CIFS   | shared | excl    | diff  | yes   | BLOCKED
        12.  CIFS   | shared | excl    | diff  | no    | DENIED

        13.  CIFS   | shared | shared  | same  | yes   | GRANTED
        14.  CIFS   | shared | shared  | same  | no    | GRANTED
        15.  CIFS   | shared | shared  | diff  | yes   | GRANTED
        16.  CIFS   | shared | shared  | diff  | no    | GRANTED

        17.  NLM    | excl   | excl    | same  | yes   | GRANTED
        18.  NLM    | excl   | excl    | same  | no    | GRANTED
        19.  NLM    | excl   | excl    | diff  | yes   | BLOCKED
        20.  NLM    | excl   | excl    | diff  | no    | DENIED

        21.  NLM    | excl   | shared  | same  | yes   | GRANTED
        22.  NLM    | excl   | shared  | same  | no    | GRANTED
        23.  NLM    | excl   | shared  | diff  | yes   | BLOCKED
        24.  NLM    | excl   | shared  | diff  | no    | DENIED

        25.  NLM    | shared | excl    | same  | yes   | GRANTED
        26.  NLM    | shared | excl    | same  | no    | GRANTED
        27.  NLM    | shared | excl    | diff  | yes   | BLOCKED
        28.  NLM    | shared | excl    | diff  | no    | DENIED

        29.  NLM    | shared | shared  | same  | yes   | GRANTED
        30.  NLM    | shared | shared  | same  | no    | GRANTED
        31.  NLM    | shared | shared  | diff  | yes   | GRANTED
        32.  NLM    | shared | shared  | diff  | no    | GRANTED
          ======================================================

    We implement the rule using a 5-dimensional array of 2x2x2x2x2:
	    array[protocol][addLockType][treeType][owner][block]
	where,
	    CIFS   = 0
		NLM    = 1
        excl   = 0
		shared = 1
		same   = 0
		diff   = 1
		yes    = 0
		no     = 1

 *-----------------------------------------------------------------------------
 */

/* Definition of array size is in lock-comp.h */

/* Define the array */
uint8 lock_conflictRule[LOCK_CONFLICT_RULE_ARRAY_S_PROTOCOL]\
                       [LOCK_CONFLICT_RULE_ARRAY_S_ADDLOCKTYPE]\
                       [LOCK_CONFLICT_RULE_ARRAY_S_TREETYPE]\
                       [LOCK_CONFLICT_RULE_ARRAY_S_OWNER]\
                       [LOCK_CONFLICT_RULE_ARRAY_S_BLOCK];

/*-----------------------------------------------------------------
 * Name:         lock_conflictRuleInit
 *
 * Description:
 *
 *     Initialize the lock conflict rule array.
 *
 * Arguments:    
 * 
 *     None.
 *
 * Return value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/03/2002
 *
 *-----------------------------------------------------------------
 */
void
lock_conflictRuleInit()
{
	int32	CIFS   = 0;
	int32	NLM    = 1;
	int32	excl   = 0;
	int32	shared = 1;
	int32	same   = 0;
	int32	diff   = 1;
	int32	yes    = 0;
	int32	no     = 1;

/*1*/  lock_conflictRule[CIFS][excl][excl][same][yes]     = LOCK_STATE_BLOCKED;
/*2*/  lock_conflictRule[CIFS][excl][excl][same][no]      = LOCK_STATE_DENIED;
/*3*/  lock_conflictRule[CIFS][excl][excl][diff][yes]     = LOCK_STATE_BLOCKED;
/*4*/  lock_conflictRule[CIFS][excl][excl][diff][no]      = LOCK_STATE_DENIED;
/*5*/  lock_conflictRule[CIFS][excl][shared][same][yes]   = LOCK_STATE_BLOCKED;
/*6*/  lock_conflictRule[CIFS][excl][shared][same][no]    = LOCK_STATE_DENIED;
/*7*/  lock_conflictRule[CIFS][excl][shared][diff][yes]   = LOCK_STATE_BLOCKED;
/*8*/  lock_conflictRule[CIFS][excl][shared][diff][no]    = LOCK_STATE_DENIED;
/*9*/  lock_conflictRule[CIFS][shared][excl][same][yes]   = LOCK_STATE_GRANTED;
/*10*/ lock_conflictRule[CIFS][shared][excl][same][no]    = LOCK_STATE_GRANTED;
/*11*/ lock_conflictRule[CIFS][shared][excl][diff][yes]   = LOCK_STATE_BLOCKED;
/*12*/ lock_conflictRule[CIFS][shared][excl][diff][no]    = LOCK_STATE_DENIED;
/*13*/ lock_conflictRule[CIFS][shared][shared][same][yes] = LOCK_STATE_GRANTED;
/*14*/ lock_conflictRule[CIFS][shared][shared][same][no]  = LOCK_STATE_GRANTED;
/*15*/ lock_conflictRule[CIFS][shared][shared][diff][yes] = LOCK_STATE_GRANTED;
/*16*/ lock_conflictRule[CIFS][shared][shared][diff][no]  = LOCK_STATE_GRANTED;
/*17*/ lock_conflictRule[NLM][excl][excl][same][yes]      = LOCK_STATE_GRANTED;
/*18*/ lock_conflictRule[NLM][excl][excl][same][no]       = LOCK_STATE_GRANTED;
/*19*/ lock_conflictRule[NLM][excl][excl][diff][yes]      = LOCK_STATE_BLOCKED;
/*20*/ lock_conflictRule[NLM][excl][excl][diff][no]       = LOCK_STATE_DENIED;
/*21*/ lock_conflictRule[NLM][excl][shared][same][yes]    = LOCK_STATE_GRANTED;
/*22*/ lock_conflictRule[NLM][excl][shared][same][no]     = LOCK_STATE_GRANTED;
/*23*/ lock_conflictRule[NLM][excl][shared][diff][yes]    = LOCK_STATE_BLOCKED;
/*24*/ lock_conflictRule[NLM][excl][shared][diff][no]     = LOCK_STATE_DENIED;
/*25*/ lock_conflictRule[NLM][shared][excl][same][yes]    = LOCK_STATE_GRANTED;
/*26*/ lock_conflictRule[NLM][shared][excl][same][no]     = LOCK_STATE_GRANTED;
/*27*/ lock_conflictRule[NLM][shared][excl][diff][yes]    = LOCK_STATE_BLOCKED;
/*28*/ lock_conflictRule[NLM][shared][excl][diff][no]     = LOCK_STATE_DENIED;
/*29*/ lock_conflictRule[NLM][shared][shared][same][yes]  = LOCK_STATE_GRANTED;
/*30*/ lock_conflictRule[NLM][shared][shared][same][no]   = LOCK_STATE_GRANTED;
/*31*/ lock_conflictRule[NLM][shared][shared][diff][yes]  = LOCK_STATE_GRANTED;
/*32*/ lock_conflictRule[NLM][shared][shared][diff][no]   = LOCK_STATE_GRANTED;
 
} /* lock_conflictRuleInit() */

/*-----------------------------------------------------------------
 * Name:         lock_checkConflict
 *
 * Description:
 *
 *     Check if the new lock conflicts with the existing locks or not.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     lockInfop - Pointer to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                 data structure, which contains information of new lock.
 *
 *     lockOwnerType - The type of owner (LOCK_OWNER_CIFS or LOCK_OWNER_NLM)
 *                     of the new lock.
 *
 *     lockType - Exclusive or shared lock (LOCK_TYPE_EXCL or LOCK_TYPE_SHAR).
 *
 *     blockIfConflict - TRUE if the new lock is blocked in case there is
 *                            conflict.
 *                       FALSE if retrun failure immediately in case there
 *                             is conflict.
 *
 *     confLockIdxp - This is output. Pointer to the index of a
 *                    lock_exLock_t or lock_shLock_t structure.
 *                    It is the first found conflicting lock in the tree.
 *
 *     exclResultp - This is output. It contains result of conflict checking
 *                   in exclusive tree.
 *
 *     sharResultp - This is output. It contains result of conflict checking
 *                   in shared tree.
 *
 * Return value:
 *
 *     LOCK_STATE_GRANTED, LOCK_STATE_DENIED, or LOCK_STATE_BLOCKED -
 *                                                                  If succeed.
 *
 *     lock error code - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/04/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_checkConflict(lock_fileLock_t	*fileLockp,
				   void				*lockInfop,
				   uint8			lockOwnerType,
				   uint8			lockType,
				   boolean			blockIfConflict,
				   vs_mem_index_t	*confLockIdxp,
				   int32			*exclResultp,
				   int32			*sharResultp)
{
	uint16			searchMask;
	lock_exLock_t	*exLockp;
	lock_shNode_t	*shNodep;
	vs_mem_index_t	prevShLockIdx;
	lock_shLock_t	*shLockp, *prevShLockp;
	int32			exclResult, sharResult, result;

	int32 protocol, addlocktype, treetype, owner, block; /*Array dimentions*/

	if ((fileLockp == NULL) || (lockInfop == NULL))
	{
		LOCK_PRINTF(LOG_ERR, ("lock_checkConflict: NULL input pointer."));
		return (LOCK_ERR_GENERAL);
	}

	VS_MEM_IDX_SET_NULL(*confLockIdxp);

	protocol = (lockOwnerType == LOCK_OWNER_CIFS) ? 0 : 1;
	addlocktype = (lockType == LOCK_TYPE_EXCL) ? 0 : 1;
	block = (blockIfConflict == TRUE) ? 0: 1;

	/*--------------------------------------------
	 * Check conflict with the exclusive lock tree
	 *--------------------------------------------
	 */
	treetype = 0;	/*Exlcusive lock tree*/
	exclResult = LOCK_STATE_GRANTED;

	/*
	 * Check overlapping locks who have the same owners
	 */
	searchMask = LOCK_SEARCH_SAME_OWNER;
	owner = 0;
	exLockp = lock_findExclLockByLockInfo(
									&(fileLockp->exclLockTree),
									lockInfop,
									lockOwnerType,
									searchMask);
	if (exLockp != NULL)
	{
		exclResult = lock_conflictRule[protocol]\
									  [addlocktype]\
									  [treetype]\
									  [owner]\
									  [block];
		if (exclResult != LOCK_STATE_GRANTED)
		{
			*confLockIdxp = fileLockp->exclLockTree;
		}
	}

	/*
	 * Check overlapping locks who have different owners
	 */
	searchMask = LOCK_SEARCH_DIFF_OWNER;
	owner = 1;
	exLockp = lock_findExclLockByLockInfo(
									&(fileLockp->exclLockTree),
									lockInfop,
									lockOwnerType,
									searchMask);
	if (exLockp != NULL)
	{
		result = lock_conflictRule[protocol]\
								  [addlocktype]\
								  [treetype]\
								  [owner]\
								  [block];
		if (result != LOCK_STATE_GRANTED)
		{
			exclResult = result;
			if (VS_MEM_IDX_IS_NULL(*confLockIdxp))
				*confLockIdxp = fileLockp->exclLockTree;
		}
	}

	/*-----------------------------------------
	 * Check conflict with the shared lock tree
	 *-----------------------------------------
	 */
	treetype = 1;	/*Shared lock tree*/
	sharResult = LOCK_STATE_GRANTED;

	/* There is no need to check shared lock tree 
	 * if a shared lock is to be added.
	 */
	if (lockType == LOCK_TYPE_EXCL)
	{
		/*TODO: Improvement:
		 * If new lock is CIFS exclusive lock, search only once.
		 * searchMask is 0.
		 */

		/*
		 * Check overlapping locks who have the same owners
		 */
		searchMask = LOCK_SEARCH_SAME_OWNER | LOCK_SEARCH_IN_SHARED_LIST;
		owner = 0;
		shLockp = lock_findSharedLockByLockInfo(
								&(fileLockp->sharedLockTree),
								lockInfop,
								lockOwnerType,
								searchMask,
								&shNodep,
								&prevShLockIdx);
		if (shLockp != NULL)
		{
			sharResult = lock_conflictRule[protocol]\
										  [addlocktype]\
										  [treetype]\
										  [owner]\
										  [block];
			if (sharResult != LOCK_STATE_GRANTED)
			{
				if (VS_MEM_IDX_IS_NULL(prevShLockIdx))
					*confLockIdxp = shNodep->lockList;
				else
				{
					prevShLockp = (lock_shLock_t *)vs_mem_idxToPtr(
															prevShLockIdx);
					if (prevShLockp == NULL)
						return (LOCK_ERR_GENERAL);
					*confLockIdxp = prevShLockp->link;
				}
			}
		}
		/*
		 * Check overlapping locks who have different owners
		 */
		searchMask = LOCK_SEARCH_DIFF_OWNER | LOCK_SEARCH_IN_SHARED_LIST;
		owner = 1;
		shLockp = lock_findSharedLockByLockInfo(
								&(fileLockp->sharedLockTree),
								lockInfop,
								lockOwnerType,
								searchMask,
								&shNodep,
								&prevShLockIdx);
		if (shLockp != NULL)
		{
			result = lock_conflictRule[protocol]\
									  [addlocktype]\
									  [treetype]\
									  [owner]\
									  [block];
			if (result != LOCK_STATE_GRANTED)
			{
				sharResult = result;
				if (VS_MEM_IDX_IS_NULL(*confLockIdxp))
				{
					if (VS_MEM_IDX_IS_NULL(prevShLockIdx))
						*confLockIdxp = shNodep->lockList;
					else
					{
						prevShLockp = (lock_shLock_t *)vs_mem_idxToPtr(
																prevShLockIdx);
						if (prevShLockp == NULL)
							return (LOCK_ERR_GENERAL);
						*confLockIdxp = prevShLockp->link;
					}
				}
			}
		}
	} /* if (lockType == LOCK_TYPE_EXCL) */

	*exclResultp = exclResult;
	*sharResultp = sharResult;

	if (exclResult == LOCK_STATE_GRANTED)
		return (sharResult);
	else
		return (exclResult);
} /* lock_checkConflict() */

/*-----------------------------------------------------------------------------
      Define rules for checking whether read or write access to a range of
	types are allowed. Used only by CIFS.

      Logic table:
      Explanation of each column in the table:

        existing mandatory lock type: The type of the existing mandatory lock
                                      being compared with the given lock range.
        wr/rd:     Check for read or write access.
        owner:     Whether the two lock ranges have the same owner.
        result:    Indicate whether the lock range can be accessed or not.

      If conditions represented by two or more rows are satisfied at the same
    time, and not all results are NFX_OK, then the return value should be
    NFX_ERR.

            ===================================
             existing |       |       |
             mandatory|       |       |
             lock type| wr/rd | owner | result
            ===================================
          1.   excl   |   wr  | same  |  OK
          2.   excl   |   wr  | diff  |  ERR

          3.   excl   |   rd  | same  |  OK
          4.   excl   |   rd  | diff  |  ERR

          5.   shared |   wr  | same  |  ERR
          6.   shared |   wr  | diff  |  ERR

          7.   shared |   rd  | same  |  OK
          8.   shared |   rd  | diff  |  OK

    We implement the rule using a 3-dimensional array of 2x2x2:
        array[lockType][wrrd][owner]
    where,
        excl   = 0
        shared = 1
		wr     = 0
		rd     = 1
        same   = 0
        diff   = 1

 *-----------------------------------------------------------------------------
 */

/* Definition of array size */
#define LOCK_ACCESS_RULE_ARRAY_S_LOCKTYPE	2
#define LOCK_ACCESS_RULE_ARRAY_S_WRRD		2
#define LOCK_ACCESS_RULE_ARRAY_S_OWNER		2

/* Define the array */
int32 lock_accessRule[LOCK_ACCESS_RULE_ARRAY_S_LOCKTYPE]\
					 [LOCK_ACCESS_RULE_ARRAY_S_WRRD]\
					 [LOCK_ACCESS_RULE_ARRAY_S_OWNER];

/*-----------------------------------------------------------------
 * Name:         lock_accessRuleInit
 *
 * Description:
 *
 *     Initialize the lock access rule array.
 *
 * Arguments:    
 * 
 *     None.
 *
 * Return value:
 *
 *     None.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/07/2002
 *
 *-----------------------------------------------------------------
 */
void
lock_accessRuleInit(void)
{
    int32 excl   = 0;	/*LOCK_TYPE_EXCLUSIVE*/
    int32 shared = 1;	/*LOCK_TYPE_SHARED*/
	int32 wr     = 0;
	int32 rd     = 1;
    int32 same   = 0;
    int32 diff   = 1;


	/*1.*/   lock_accessRule[excl][wr][same]   = NFX_OK;
    /*2.*/   lock_accessRule[excl][wr][diff]   = NFX_ERR;
    /*3.*/   lock_accessRule[excl][rd][same]   = NFX_OK;
    /*4.*/   lock_accessRule[excl][rd][diff]   = NFX_ERR;
    /*5.*/   lock_accessRule[shared][wr][same] = NFX_ERR;
    /*6.*/   lock_accessRule[shared][wr][diff] = NFX_ERR;
    /*7.*/   lock_accessRule[shared][rd][same] = NFX_OK;
    /*8.*/   lock_accessRule[shared][rd][diff] = NFX_OK;

} /* lock_conflictRuleInit() */

/*-----------------------------------------------------------------
 * Name:         lock_checkAccess
 *
 * Description:
 *
 *     Check if read/write access to a byte range is ok or not (CIFS only).
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     cifsLockInfop - Pointer to a cifsLockInfo data structure, which
 *                     contains byte range information.
 *
 *     writeAccess - TRUE if checking write access.
 *                   FALSE if checking read access.
 *
 * Return value:
 *
 *     NFX_OK - If access is allowed.
 *
 *     NFX_ERR - If access is denied.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/04/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_checkAccess(lock_fileLock_t		*fileLockp,
				 lock_cifsLockInfo_t	*cifsLockInfop,
				 boolean				writeAccess)
{
	uint16			searchMask;
	lock_exLock_t	*exLockp;
	lock_shLock_t	*shLockp;
	lock_shNode_t	*shNodep;
	uint8			locktype, wrrd, owner;
	int32			lastGrantedResult, exclResult, sharResult, result;
	void			*lastGrantedLockp;
	vs_mem_index_t	lastGrantedLockInfoIdx, dummyLockInfoIdx;
	int8			comp;
	vs_mem_index_t	dummyShLockIdx;

	if (writeAccess == TRUE)
		wrrd = 0;	/*Write access*/
	else
		wrrd = 1;	/*Read access*/

	/*----------------------------------------------------------
	 * Look at the last granted lock. See if the same lock owner
	 * is checking read/write access to the same byte range as
	 * the last ranged lock. 
	 *----------------------------------------------------------
	 */
	/* Get lockInfo of last granted lock */
	lastGrantedLockp = lock_getGrantedLock(fileLockp->lastGrantedLock, &locktype);
	if (lastGrantedLockp != NULL)
	{
		if (locktype == LOCK_TYPE_EXCLUSIVE)
		{
			exLockp = (lock_exLock_t *)lastGrantedLockp;
			lastGrantedLockInfoIdx = exLockp->info;
		}
		else if (locktype == LOCK_TYPE_SHARED)
		{
			shLockp = (lock_shLock_t *)lastGrantedLockp;
			lastGrantedLockInfoIdx = shLockp->info;
		}
		else
			return (NFX_ERR);

		/*
		 * Check if two byte ranges are the same and have the same owners
		 */
		searchMask = LOCK_SEARCH_SAME_RANGE | LOCK_SEARCH_SAME_OWNER;
		owner = 0;		/*Same Owner*/

		VS_MEM_IDX_SET_NULL(dummyLockInfoIdx);
		comp = lock_twoLockInfoCompFunc(dummyLockInfoIdx,
										(void *)cifsLockInfop,
										LOCK_OWNER_CIFS,
										lastGrantedLockInfoIdx,
										searchMask,
										LOCK_SPECIAL_COMP_RESULT);
		if (comp == 0)
		{
			/* The locks match */
			lastGrantedResult = lock_accessRule[locktype][wrrd][owner];
			return (lastGrantedResult);
		}
	}

	/*------------------------------------------------------
	 * Look at the lowest mandatory lock offset.
	 * See if read/write access to the byte range lower than
	 * the lowest mandatory lock offset. 
	 *------------------------------------------------------
	 */
	{
		uint64 end;

		LOCK_GET_LOCK_RANGE_END(cifsLockInfop->offset, cifsLockInfop->len, end);

		if (end < fileLockp->lowestMandatoryLockOffset)
			return (NFX_OK);
	}

	/*------------------------------------
	 * Check access to exclusive lock tree
	 *------------------------------------
	 */
	locktype = 0;	/*Exclusive lock*/
	exclResult = NFX_OK;

	/*
	 * Check non-zero length overlapping locks who have the same owners
	 */
	searchMask = LOCK_SEARCH_SAME_OWNER | LOCK_SEARCH_NONZERO_LEN;
	owner = 0;		/*Same Owner*/
	exLockp = lock_findExclLockByLockInfo(
									&(fileLockp->exclLockTree),
									(void *)cifsLockInfop,
									LOCK_OWNER_CIFS,
									searchMask);
	if (exLockp != NULL)
	{
		exclResult = lock_accessRule[locktype][wrrd][owner];
	}

	/*
	 * Check non-zero length overlapping locks who have different owners
	 */
	searchMask = LOCK_SEARCH_DIFF_OWNER |
		   		 LOCK_SEARCH_MANDATORY_LOCK |
				 LOCK_SEARCH_NONZERO_LEN;
	owner = 1;	/*Different owner*/
	exLockp = lock_findExclLockByLockInfo(
									&(fileLockp->exclLockTree),
									(void *)cifsLockInfop,
									LOCK_OWNER_CIFS,
									searchMask);
	if (exLockp != NULL)
	{
		result = lock_accessRule[locktype][wrrd][owner];
		if (result != NFX_OK)
			exclResult = result;
	}

	/*---------------------------------
	 * Check access to shared lock tree
	 *---------------------------------
	 */
	if (wrrd == 1)
	{
		/* Checking read access. Always OK */
		sharResult = NFX_OK;
	}
	else
	{
		/* Checking write access.
		 * Not OK if there is an non-zero length overlapping lock */
		searchMask = LOCK_SEARCH_IN_SHARED_LIST |
					 LOCK_SEARCH_MANDATORY_LOCK |
					 LOCK_SEARCH_NONZERO_LEN;
		shLockp = lock_findSharedLockByLockInfo(
									&(fileLockp->sharedLockTree),
									(void *)cifsLockInfop,
									LOCK_OWNER_CIFS,
									searchMask,
									&shNodep,
									&dummyShLockIdx);
		if (shLockp == NULL)
			sharResult = NFX_OK;
		else
			sharResult = NFX_ERR;
	}

	if ((exclResult == NFX_ERR) || (sharResult == NFX_ERR))
		return (NFX_ERR);
	else
		return (NFX_OK);
} /* lock_checkAccess() */

