/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        lock-mem.c
 *
 * RCS:  $Id: $
 *
 * Copyright (C) 2002, ClariStor Inc.
 *
 * Description: Byte range lock memory management functions.
 *
 * Created by:  Xiaobing Zhang
 *
 * Date Created:    11/09/2002
 *
 *-----------------------------------------------------------------
 */

#include "lock-incls.h"

/*-----------------------------------------------------------------
 * Name:        lock_memAllocCifsLockOwner
 *
 * Description:
 *  
 *     Allocate and initialize a structure of type lock_cifsLockOwner_t.
 *  
 * Arguments:    
 *      
 *     tid - CIFS tree connection.
 *      
 *     fid - Open file handle.
 *      
 *     pid - Process ID.
 *      
 *     lockOwnerIdxp - This is output. It is pointer to the index to
 *                     the newly allocated lock_cifsLockOwner_t structure.
 *  
 * Return value:
 *      
 *     Pointer to the allocated lock_cifsLockOwner_t - If succeed.
 * 
 *     NULL - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 11/23/2002
 *
 *-----------------------------------------------------------------
 */
lock_cifsLockOwner_t *  
lock_memAllocCifsLockOwner(uint16			tid,
						   uint16			fid,
						   uint16			pid,
						   vs_mem_index_t	*lockOwnerIdxp) /*Return the index*/
{
	lock_cifsLockOwner_t    *lockOwnerp;

	/*-----------------------------------------------
	 * Allocate a lock_cifsLockOwner_t data structure
	 *-----------------------------------------------
	 */
	lockOwnerp = (lock_cifsLockOwner_t *)vs_mem_alloc(0, //FIXME: srvrIdx
													  vs_mem_dt_cifsLockOwner,
													  lockOwnerIdxp);
	if (lockOwnerp == NULL)
	{
		lock_globalCfgp->errCounters.insufficientMemory++;
		return (NULL);
	}

	/*------------------------
	 * Init the data structure
	 *------------------------
	 */
	lockOwnerp->header.ownerType = LOCK_OWNER_CIFS;
	lockOwnerp->header.lockCount = 0;
	lockOwnerp->header.inList = 0;
	lockOwnerp->tid = tid;
	lockOwnerp->fid = fid;
	lockOwnerp->pid = pid;

	/*-----------------------------
	 * Update debug global counters
	 *-----------------------------
	 */
	lock_globalCfgp->debugCounters.cifsLockOwners++;
	lock_globalCfgp->debugCounters.allocCifsLockOwner++;

	return (lockOwnerp);
} /* lock_memAllocCifsLockOwner() */

/*-----------------------------------------------------------------
 * Name:        lock_memFreeCifsLockOwner
 *
 * Description:
 *  
 *     Free an existing structure of type lock_cifsLockOwner_t.
 *  
 * Arguments:    
 *      
 *     lockOwnerIdx - Index to lock_cifsLockOwner_t structure.
 *      
 * Return value:
 *      
 *     NFX_OK - If succeed.
 * 
 *     NFX_ERR - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 01/06/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_memFreeCifsLockOwner(vs_mem_index_t	lockOwnerIdx)
{
	int32	rc;

	/*------------
	 * Free memory
	 *------------
	 */
	rc = vs_mem_free(lockOwnerIdx);
	if (rc == NFX_OK)
	{
		/*----------------------------
		 * Update global debug counter
		 *----------------------------
		 */
		lock_globalCfgp->debugCounters.cifsLockOwners--;
		lock_globalCfgp->debugCounters.freeCifsLockOwner++;
	}

	return (rc);

} /* lock_memFreeCifsLockOwner() */

/*-----------------------------------------------------------------
 * Name:        lock_memAllocNlmLockClient
 *
 * Description:
 *  
 *     Allocate and initialize a structure of type lock_nlmLockOwner_t.
 *  
 * Arguments:    
 *      
 *     clientIpAddr - NLM client IP address.
 *      
 *     state - Client's Status Monitor state number.
 *      
 *     pid - Lock owner process ID.
 *      
 *     callerNameLen - The number of bytes in the client machine's name.
 *      
 *     callerName - Name of the client machine.
 *      
 *     lockClientIdxp - This is output. It is pointer to the index to
 *                      the newly allocated lock_nlmLockClient_t structure.
 *  
 * Return value:
 *      
 *     Pointer to the allocated lock_nlmLockClient_t - If succeed.
 * 
 *     NULL - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 11/23/2002
 *
 *-----------------------------------------------------------------
 */
lock_nlmLockClient_t *
lock_memAllocNlmLockClient(ip_index_t		clientIpAddr,
						   int32			state,
						   int32			pid,
						   uint16			callerNameLen,
						   char				*callerName,
						   vs_mem_index_t	*lockClientIdxp)
{
	lock_nlmLockClient_t	*lockClientp;

	/*-----------------------------------------------
	 * Allocate a lock_nlmLockClient_t data structure
	 *-----------------------------------------------
	 */
	lockClientp = (lock_nlmLockClient_t *)vs_mem_alloc(0, //FIXME: srvrIdx
													   vs_mem_dt_nlmLockClient,
													   lockClientIdxp);
	if (lockClientp == NULL)
	{
		lock_globalCfgp->errCounters.insufficientMemory++;
		return (NULL);
	}

	/*------------------------
	 * Init the data structure
	 *------------------------
	 */
	lockClientp->header.ownerCount = 0;
	lockClientp->header.inList = 0;
	lockClientp->state = state;
	VS_MEM_IDX_SET_NULL(lockClientp->volumeList);
	lockClientp->clientIpAddr = clientIpAddr;
	lockClientp->callerNameLen = callerNameLen;
	LOCK_SHMEM_ALLOC(lockClientp->callerName,
					 LOCK_VSERVER_IDX_ANY, /*FIXME: What value should it be?*/
					 char,
					 callerNameLen);
	if (lockClientp->callerName == NULL)
	{
		vs_mem_free(*lockClientIdxp);
		lock_globalCfgp->errCounters.insufficientMemory++;
		return (NULL);
	}
	memcpy(lockClientp->callerName, callerName, callerNameLen);

	/*----------------------------
	 * Update global debug counter
	 *----------------------------
	 */
	lock_globalCfgp->debugCounters.nlmLockClients++;
	lock_globalCfgp->debugCounters.allocNlmLockClient++;

	return (lockClientp);
} /* lock_memAllocNlmLockClient() */

/*-----------------------------------------------------------------
 * Name:        lock_memFreeNlmLockClient
 *
 * Description:
 *  
 *     Free an existing structure of type lock_nlmLockClient_t.
 *  
 * Arguments:    
 *      
 *     lockClientIdx - Index to lock_nlmLockClient_t structure.
 *      
 *     lockClientp - Pointer to lock_nlmLockClient_t structure.
 *                   This may be NULL.
 *  
 * Return value:
 *      
 *     NFX_OK - If succeed.
 * 
 *     NFX_ERR - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 01/06/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_memFreeNlmLockClient(vs_mem_index_t		lockClientIdx,
						  lock_nlmLockClient_t	*lockClientp)
{
	int32	rc;

	/*-----------------
	 * Find lock client 
	 *-----------------
	 */
	if (lockClientp == NULL)
	{
		lockClientp = (lock_nlmLockClient_t *)vs_mem_idxToPtr(lockClientIdx);
		if (lockClientp == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_memFreeNlmLockClient: "
										"NULL pointer to lockClient."));
			return (NFX_ERR);
		}
	}

	/*------------
	 * Free memory
	 *------------
	 */
	LOCK_SHMEM_DEALLOC(lockClientp->callerName);
	rc = vs_mem_free(lockClientIdx);
	if (rc == NFX_OK)
	{
		/*----------------------------
		 * Update global debug counter
		 *----------------------------
		 */
		lock_globalCfgp->debugCounters.nlmLockClients--;
		lock_globalCfgp->debugCounters.freeNlmLockClient++;
	}

	return (rc);
} /* lock_memFreeNlmLockClient() */

/*-----------------------------------------------------------------
 * Name:        lock_memAllocNlmLockOwner
 *
 * Description:
 *  
 *     Allocate and initialize a structure of type lock_nlmLockOwner_t.
 *  
 * Arguments:    
 *      
 *     lockClientIdx - Index to the lock_nlmLockClient_t structure,
 *                     which contains lock client information.
 *
 *     lockClientp - Pointer to the lock_nlmLockClient_t structure,
 *                   which contains lock client information.
 *
 *     pid - Process ID.
 *      
 *     ownerNameLen - The number of bytes in the lock owner's name.
 *      
 *     ownerName - Name of the lock owner.
 *      
 *     fcb - Index to the file control block. The lock owner
 *           owns lock(s) on this file.
 *
 *     lockOwnerIdxp - This is output. It is pointer to the index to
 *                     the newly allocated lock_nlmLockOwner_t structure.
 *  
 * Return value:
 *      
 *     Pointer to the allocated lock_nlmLockOwner_t - If succeed.
 * 
 *     NULL - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 11/23/2002
 *
 *-----------------------------------------------------------------
 */
lock_nlmLockOwner_t *  
lock_memAllocNlmLockOwner(vs_mem_index_t		lockClientIdx,
						  lock_nlmLockClient_t	*lockClientp,
						  int32					pid,
						  uint16				ownerNameLen,
						  char					*ownerName,
						  vs_mem_index_t		fcb,
						  vs_mem_index_t		*lockOwnerIdxp)
{
	lock_nlmLockOwner_t    *lockOwnerp;

	/*----------------------------------------------
	 * Allocate a lock_nlmLockOwner_t data structure
	 *----------------------------------------------
	 */
	lockOwnerp = (lock_nlmLockOwner_t *)vs_mem_alloc(0, //FIXME: srvrIdx
				                            	     vs_mem_dt_nlmLockOwner,
													 lockOwnerIdxp);
	if (lockOwnerp == NULL)
	{
		lock_globalCfgp->errCounters.insufficientMemory++;
		return (NULL);
	}

	/*------------------------
	 * Init the data structure
	 *------------------------
	 */
	lockOwnerp->header.ownerType = LOCK_OWNER_NLM;
	lockOwnerp->header.lockCount = 0;
	lockOwnerp->header.inList = 0;
	lockOwnerp->lockClient = lockClientIdx;
	lockOwnerp->pid = pid;
	VS_MEM_IDX_SET_NULL(lockOwnerp->volumeList);
	VS_MEM_IDX_SET_NULL(lockOwnerp->dependencyList);
	lockOwnerp->ownerNameLen = ownerNameLen;
	LOCK_SHMEM_ALLOC(lockOwnerp->ownerName,
					 LOCK_VSERVER_IDX_ANY, /*FIXME: What value should it be?*/
					 char,
					 ownerNameLen);
	if (lockOwnerp->ownerName == NULL)
	{
		vs_mem_free(*lockOwnerIdxp);
		lock_globalCfgp->errCounters.insufficientMemory++;
		return (NULL);
	}
	memcpy(lockOwnerp->ownerName, ownerName, ownerNameLen);

    /*---------------------------------------
	 * Update lock owner count in lock client
	 *---------------------------------------
	 */
	lockClientp->header.ownerCount++;

    /*TODO: cluster NLM support.
	 * Update the lock owner count in a nlmRecOnVolCB
	 * in volumeList in lockClientp.
	 *
	 * The volume represented by the nlmRecOnVolCB is the volume
	 * that has the file represented by the fcb.
	 *
	 * If this lock client record is not stored on this volume,
	 * save this record on the volume.
	 */

	/*----------------------------
	 * Update global debug counter
	 *----------------------------
	 */
	lock_globalCfgp->debugCounters.nlmLockOwners++;
	lock_globalCfgp->debugCounters.allocNlmLockOwner++;

	return (lockOwnerp);
} /* lock_memAllocNlmLockOwner() */

/*-----------------------------------------------------------------
 * Name:        lock_memFreeNlmLockOwner
 *
 * Description:
 *  
 *     Free an existing structure of type lock_nlmLockOwner_t.
 *  
 * Arguments:    
 *      
 *     lockOwnerIdx - Index to lock_nlmLockOwner_t structure.
 *      
 *     lockOwnerp - Pointer to lock_nlmLockOwner_t structure.
 *                  This may be NULL.
 *  
 * Return value:
 *      
 *     NFX_OK - If succeed.
 * 
 *     NFX_ERR - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 01/06/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_memFreeNlmLockOwner(vs_mem_index_t			lockOwnerIdx,
						 lock_nlmLockOwner_t	*lockOwnerp)
{
	int32					rc;
	lock_nlmLockClient_t	*lockClientp;

	/*----------------
	 * Find lock owner
	 *----------------
	 */
	if (lockOwnerp == NULL)
	{
		lockOwnerp = (lock_nlmLockOwner_t *)vs_mem_idxToPtr(lockOwnerIdx);
		if (lockOwnerp == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_memFreeNlmLockOwner: "
										"NULL pointer to lockOwner."));
			return (NFX_ERR);
		}
	}

	/*-----------------------------------------
	 * Find lock client, and update owner count
	 *-----------------------------------------
	 */
	lockClientp = (lock_nlmLockClient_t *)vs_mem_idxToPtr(lockOwnerp->lockClient);
	if (lockClientp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_memFreeNlmLockOwner: "
										"NULL pointer to lockOwner."));
		return (NFX_ERR);
	}
	lockClientp->header.ownerCount--;

	if ((lockClientp->header.ownerCount == 0) &&
		(lockClientp->header.inList == 0))
	{
		/*------------------------------------------------------
		 * Put the lock client structure in a list to be deleted
		 * a few minutes later.
		 *------------------------------------------------------
		 */
		rc = lock_insertOwnerClientToTimeoutList(lockOwnerp->lockClient);
		lockClientp->header.inList = 1;
	}

	/*------------
	 * Free memory
	 *------------
	 */
	LOCK_SHMEM_DEALLOC(lockOwnerp->ownerName);
	rc = vs_mem_free(lockOwnerIdx);
	if (rc == NFX_OK)
	{
		/*----------------------------
		 * Update global debug counter
		 *----------------------------
		 */
		lock_globalCfgp->debugCounters.nlmLockOwners--;
		lock_globalCfgp->debugCounters.freeNlmLockOwner++;
	}

	return (rc);
} /* lock_memFreeNlmLockOwner() */

/*-----------------------------------------------------------------
 * Name:        lock_memAllocDependency
 *
 * Description:
 *  
 *     Allocate and initialize a structure of type lock_dependency_t.
 *  
 * Arguments:    
 *      
 *     lockOwnerIdx - Index to lock_cifsLockOwner_t or lock_nlmLockOwner_t
 *                    structure.
 *      
 *     dependencyIdxp - This is output. It is pointer to the index to
 *                      the newly allocated lock_dependency_t structure.
 *  
 * Return value:
 *      
 *     Pointer to the allocated lock_dependency_t - If succeed.
 * 
 *     NULL - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 01/07/2003
 *
 *-----------------------------------------------------------------
 */
lock_dependency_t *
lock_memAllocDependency(vs_mem_index_t	lockOwnerIdx,
						vs_mem_index_t	*dependencyIdxp)
{
	lock_dependency_t	*dependencyp;

	if (VS_MEM_IDX_IS_NULL(lockOwnerIdx))
		return (NULL);

	/*----------------------------------------------
	 * Allocate a lock_dependency_t data structure
	 *----------------------------------------------
	 */
	dependencyp = (lock_dependency_t *)vs_mem_alloc(0,			//FIXME
							   						vs_mem_dt_lockDependency,
													dependencyIdxp);
	if (dependencyp == NULL)
	{
		lock_globalCfgp->errCounters.insufficientMemory++;
		return (NULL);
	}

	/*------------------------
	 * Init the data structure
	 *------------------------
	 */
	VS_MEM_IDX_SET_NULL(dependencyp->next);
	dependencyp->blockingOwner = lockOwnerIdx;
	dependencyp->blockedLockCount = 1;
	
	/*----------------------
	 * Update global counter
	 *----------------------
	 */
	lock_globalCfgp->debugCounters.allocDependency++;

	return (dependencyp);
} /* lock_memAllocDependency() */

/*-----------------------------------------------------------------
 * Name:        lock_memFreeDependency
 *
 * Description:
 *  
 *     Free an existing structure of type lock_dependency_t.
 *  
 * Arguments:    
 *      
 *     dependencyIdx - Index to lock_dependency_t structure.
 *  
 * Return value:
 *      
 *     NFX_OK - If succeed.
 * 
 *     NFX_ERR - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 01/07/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_memFreeDependency(vs_mem_index_t	dependencyIdx)
{
	int32					rc;

	/*------------
	 * Free memory
	 *------------
	 */
	rc = vs_mem_free(dependencyIdx);
	if (rc == NFX_OK)
	{
		/* Update global debug counter */
		lock_globalCfgp->debugCounters.freeDependency++;
	}

	return (rc);
} /* lock_memFreeDependency() */

/*-----------------------------------------------------------------
 * Name:        lock_memAllocOwnerTimeout
 *
 * Description:
 *  
 *     Allocate and initialize a structure of type lock_ownerTimeout_t.
 *  
 * Arguments:    
 *      
 *     ownerClientIdx - Index to lock_cifsLockOwner_t, lock_nlmLockClient_t,
 *                      or lock_nlmLockOwner_t structure.
 *      
 *     ownerTimeoutIdxp - This is output. It is pointer to the index to
 *                        the newly allocated lock_ownerTimeout_t structure.
 *  
 * Return value:
 *      
 *     Pointer to the allocated lock_ownerTimeout_t - If succeed.
 * 
 *     NULL - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 01/21/2003
 *
 *-----------------------------------------------------------------
 */
lock_ownerTimeout_t *
lock_memAllocOwnerTimeout(vs_mem_index_t	ownerClientIdx,
						  vs_mem_index_t	*ownerTimeoutIdxp)
{
	int32					rc;
	lock_ownerTimeout_t		*ownerTimeoutp;
	eeeRTC_t				timeout;

	if (VS_MEM_IDX_IS_NULL(ownerClientIdx))
		return (NULL);

	/*----------------------------------------------
	 * Allocate a lock_ownerTimeout_t data structure
	 *----------------------------------------------
	 */
	ownerTimeoutp = (lock_ownerTimeout_t *)vs_mem_alloc(
													0,			//FIXME
							   						vs_mem_dt_lockOwnerTimeout,
													ownerTimeoutIdxp);
	if (ownerTimeoutp == NULL)
	{
		lock_globalCfgp->errCounters.insufficientMemory++;
		return (NULL);
	}

	/*------------------------
	 * Init the data structure
	 *------------------------
	 */
	VS_MEM_IDX_SET_NULL(ownerTimeoutp->next);
	ownerTimeoutp->owner_client = ownerClientIdx;
	
	/* Assign timeout value */
	eee_rtcStart(&timeout);
	rc = eee_rtcIncr(&timeout, &lock_timeTenMinutes);
	if (rc != NFX_OK)
		return (NULL);
	ownerTimeoutp->timeout = timeout;

	/*----------------------
	 * Update global counter
	 *----------------------
	 */
	lock_globalCfgp->debugCounters.allocOwnerTimeout++;

	return (ownerTimeoutp);
} /* lock_memAllocOwnerTimeout() */

/*-----------------------------------------------------------------
 * Name:        lock_memFreeOwnerTimeout
 *
 * Description:
 *  
 *     Free an existing structure of type lock_ownerTimeout_t.
 *  
 * Arguments:    
 *      
 *     ownerTimeoutIdx - Index to lock_ownerTimeout_t structure.
 *  
 * Return value:
 *      
 *     NFX_OK - If succeed.
 * 
 *     NFX_ERR - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 01/21/2003
 *
 *-----------------------------------------------------------------
 */
int32
lock_memFreeOwnerTimeout(vs_mem_index_t	ownerTimeoutIdx)
{
	int32					rc;

	/*------------
	 * Free memory
	 *------------
	 */
	rc = vs_mem_free(ownerTimeoutIdx);
	if (rc == NFX_OK)
	{
		/* Update global debug counter */
		lock_globalCfgp->debugCounters.freeOwnerTimeout++;
	}

	return (rc);
} /* lock_memFreeOwnerTimeout() */

/*-----------------------------------------------------------------
 * Name:        lock_memAllocCifsLockInfo
 *
 * Description:
 *  
 *     Allocate and initialize a structure of type lock_cifsLockInfo_t.
 *  
 * Arguments:    
 *      
 *     cifsLockOwnerIdx - Index to lock_cifsLockOwner_t structure.
 *                        This is the owner of the lock.
 *      
 *     cifsLockOwnerp - Pointer to lock_cifsLockOwner_t structure.
 *                      This is the owner of the lock.
 *      
 *     offset - Byte range lock offset.
 *
 *     len - Byte range lock length.
 *
 *     cifsLockInfoIdxp - This is output. It is pointer to the index to
 *                     the newly allocated lock_cifsLockInfo_t structure.
 *  
 * Return value:
 *      
 *     Pointer to the allocated lock_cifsLockInfo_t - If succeed.
 * 
 *     NULL - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 11/23/2002
 *
 *-----------------------------------------------------------------
 */
lock_cifsLockInfo_t *
lock_memAllocCifsLockInfo(vs_mem_index_t		cifsLockOwnerIdx,
						  lock_cifsLockOwner_t	*cifsLockOwnerp,
						  uint64				offset,
						  uint64				len,
						  vs_mem_index_t		*cifsLockInfoIdxp)
{
	lock_cifsLockInfo_t	*cifsLockInfop;

	if (cifsLockOwnerp == NULL)
	{
		cifsLockOwnerp = (lock_cifsLockOwner_t *)
								vs_mem_idxToPtr(cifsLockOwnerIdx);
		if (cifsLockOwnerp == NULL)
			return (NULL);
	}

	/*----------------------------------------------
	 * Allocate a lock_cifsLockInfo_t data structure
	 *----------------------------------------------
	 */
	cifsLockInfop = (lock_cifsLockInfo_t *)vs_mem_alloc(0,			//FIXME
							   							vs_mem_dt_cifsLockInfo,
														cifsLockInfoIdxp);
	if (cifsLockInfop == NULL)
	{
		lock_globalCfgp->errCounters.insufficientMemory++;
		return (NULL);
	}

	/*------------------------
	 * Init the data structure
	 *------------------------
	 */
	cifsLockInfop->owner = cifsLockOwnerIdx;
	cifsLockInfop->offset = offset;
	cifsLockInfop->len = len;
	
	/*--------------------------------
	 * Update lock count in lock owner
	 *--------------------------------
	 */
	cifsLockOwnerp->header.lockCount++;

	/*-------------------------
	 * Update global lock count
	 *-------------------------
	 */
	lock_globalCfgp->lockCount++;

	/*----------------------------
	 * Update global debug counter
	 *----------------------------
	 */
	lock_globalCfgp->debugCounters.allocCifsLockInfo++;

	return (cifsLockInfop);
} /* lock_memAllocCifsLockInfo() */

/*-----------------------------------------------------------------
 * Name:        lock_memFreeCifsLockInfo
 *
 * Description:
 *  
 *     Free an existing structure of type lock_cifsLockInfo_t.
 *  
 * Arguments:    
 *      
 *     lockInfoIdx - Index to lock_cifsLockInfo_t structure.
 *  
 *     lockInfop - Pointer to the lock_cifsLockInfo_t structure.
 *                 This may be NULL.
 *  
 * Return value:
 *      
 *     NFX_OK - If succeed.
 * 
 *     NFX_ERR - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 11/23/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_memFreeCifsLockInfo(vs_mem_index_t			lockInfoIdx,
						 lock_cifsLockInfo_t	*lockInfop)
{
	int32					rc;
	lock_cifsLockOwner_t	*lockOwnerp;

	/*---------------
	 * Find lock info
	 *---------------
	 */
	if (lockInfop == NULL)
	{
		lockInfop = (lock_cifsLockInfo_t *)vs_mem_idxToPtr(lockInfoIdx);
		if (lockInfop == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_memFreeCifsLockInfo: "
								"NULL pointer to lockInfo."));
			return (NFX_ERR);
		}
	}

	/*---------------------------------------
	 * Find lock owner, and update lock count
	 *---------------------------------------
	 */
	lockOwnerp = (lock_cifsLockOwner_t *)vs_mem_idxToPtr(lockInfop->owner);
	if (lockOwnerp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_memFreeCifsLockInfo: "
							"NULL pointer to lockOwner."));
		return (NFX_ERR);
	}
	lockOwnerp->header.lockCount--;

	if ((lockOwnerp->header.lockCount == 0) &&
		(lockOwnerp->header.inList == 0))
	{
		/*-----------------------------------------------------
		 * Put the lock owner structure in a list to be deleted
		 * a few minutes later.
		 *-----------------------------------------------------
		 */
		rc = lock_insertOwnerClientToTimeoutList(lockInfop->owner);
		lockOwnerp->header.inList = 1;
	}

	/*-------------------------
	 * Update global lock count
	 *-------------------------
	 */
	lock_globalCfgp->lockCount--;

	/*------------
	 * Free memory
	 *------------
	 */
	rc = vs_mem_free(lockInfoIdx);
	if (rc == NFX_OK)
	{
		/*----------------------------
		 * Update global debug counter
		 *----------------------------
		 */
		lock_globalCfgp->debugCounters.freeCifsLockInfo++;
	}

	return (rc);
} /* lock_memFreeCifsLockInfo() */

/*-----------------------------------------------------------------
 * Name:        lock_memAllocNlmLockInfo
 *
 * Description:
 *  
 *     Allocate and initialize a structure of type lock_nlmLockInfo_t.
 *  
 * Arguments:    
 *      
 *     nlmLockOwnerIdx - Index to lock_nlmLockOwner_t structure.
 *                       This is the owner of the lock.
 *      
 *     nlmLockOwnerp - Pointer to lock_nlmLockOwner_t structure.
 *                      This is the owner of the lock.
 *      
 *     offset - Byte range lock offset.
 *
 *     len - Byte range lock length.
 *
 *     fcb - Index to control block of the file that has the lock on it.
 *
 *     nlmLockInfoIdxp - This is output. It is pointer to the index to
 *                       the newly allocated lock_nlmLockInfo_t structure.
 *  
 * Return value:
 *      
 *     Pointer to the allocated lock_nlmLockInfo_t - If succeed.
 * 
 *     NULL - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 11/23/2002
 *
 *-----------------------------------------------------------------
 */
lock_nlmLockInfo_t *
lock_memAllocNlmLockInfo(vs_mem_index_t			nlmLockOwnerIdx,
						 lock_nlmLockOwner_t	*nlmLockOwnerp,
						 uint64					offset,
						 uint64					len,
						 vs_mem_index_t			fcb,
						 vs_mem_index_t			*nlmLockInfoIdxp)
														/*Return the index*/
{
	lock_nlmLockInfo_t	*nlmLockInfop;

	if (nlmLockOwnerp == NULL)
	{
		nlmLockOwnerp = (lock_nlmLockOwner_t *)
								vs_mem_idxToPtr(nlmLockOwnerIdx);
		if (nlmLockOwnerp == NULL)
			return (NULL);
	}

	/*---------------------------------------------
	 * Allocate a lock_nlmLockInfo_t data structure
	 *---------------------------------------------
	 */
	nlmLockInfop = (lock_nlmLockInfo_t *)vs_mem_alloc(0,			//FIXME
							   						  vs_mem_dt_nlmLockInfo,
													  nlmLockInfoIdxp);
	if (nlmLockInfop == NULL)
	{
		lock_globalCfgp->errCounters.insufficientMemory++;
		return (NULL);
	}

	/*------------------------
	 * Init the data structure
	 *------------------------
	 */
	nlmLockInfop->owner = nlmLockOwnerIdx;
	nlmLockInfop->offset = offset;
	nlmLockInfop->len = len;
	nlmLockInfop->fcb = fcb;
	nlmLockInfop->idxInRecFile = 0;
	
	/*--------------------------------
	 * Update lock count in lock owner
	 *--------------------------------
	 */
	nlmLockOwnerp->header.lockCount++;

	/*TODO: cluster NLM support.
	 * Update the lock count in a nlmRecOnVolCB in
	 * volumeList in nlmLockOwnerp.
	 *
	 * The volume represented by the nlmRecOnVolCB is the volume
	 * that has the file represented by the fcb.
	 *
	 * If this lock owner record is not stored on this volume,
	 * save this record on the volume.
	 */

	/*-------------------------
	 * Update global lock count
	 *-------------------------
	 */
	lock_globalCfgp->lockCount++;

	/*----------------------------
	 * Update global debug counter
	 *----------------------------
	 */
	lock_globalCfgp->debugCounters.allocNlmLockInfo++;

	return (nlmLockInfop);
} /* lock_memAllocNlmLockInfo() */

/*-----------------------------------------------------------------
 * Name:        lock_memFreeNlmLockInfo
 *
 * Description:
 *  
 *     Free an existing structure of type lock_nlmLockInfo_t.
 *  
 * Arguments:    
 *      
 *     lockInfoIdx - Index to lock_nlmLockInfo_t structure.
 *  
 *     lockInfop - Pointer to the lock_nlmLockInfo_t structure.
 *                 This may be NULL.
 *  
 * Return value:
 *      
 *     NFX_OK - If succeed.
 * 
 *     NFX_ERR - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 11/23/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_memFreeNlmLockInfo(vs_mem_index_t		lockInfoIdx,
						lock_nlmLockInfo_t	*lockInfop)
{
	int32					rc;
	lock_nlmLockOwner_t		*lockOwnerp;

	/*---------------
	 * Find lock info
	 *---------------
	 */
	if (lockInfop == NULL)
	{
		lockInfop = (lock_nlmLockInfo_t *)vs_mem_idxToPtr(lockInfoIdx);
		if (lockInfop == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_memFreeNlmLockInfo: "
								"NULL pointer to lockInfo."));
			return (NFX_ERR);
		}
	}

	/*---------------------------------------
	 * Find lock owner, and update lock count
	 *---------------------------------------
	 */
	lockOwnerp = (lock_nlmLockOwner_t *)vs_mem_idxToPtr(lockInfop->owner);
	if (lockOwnerp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_memFreeNlmLockInfo: "
							"NULL pointer to lockOwner."));
		return (NFX_ERR);
	}
	lockOwnerp->header.lockCount--;

	/*TODO: cluster NLM support.
	 * Decrement the lock count in a nlmRecOnVolCB in
	 * volumeList in nlmLockOwnerp.
	 *
	 * The volume represented by the nlmRecOnVolCB is the volume
	 * that has the file represented by the fcb.
	 *
	 * If the lock count in the nlmRecOnVolCB becomes 0,
	 * remove this owner record from the volume.
	 * And decrement the owner count in a lock_nlmRecOnVolCB in
	 * volumeList in nlmLockClientp. The lock owner is from this client.
	 * If the owner count in the lock_nlmRecOnVolCB becomes 0,
	 * remove this client record from the volume.
	 */

	if ((lockOwnerp->header.lockCount == 0) &&
		(lockOwnerp->header.inList == 0))
	{
		/*-----------------------------------------------------
		 * Put the lock owner structure in a list to be deleted
		 * a few minutes later.
		 *-----------------------------------------------------
		 */
		rc = lock_insertOwnerClientToTimeoutList(lockInfop->owner);
		lockOwnerp->header.inList = 1;
	}

	/*-------------------------
	 * Update global lock count
	 *-------------------------
	 */
	lock_globalCfgp->lockCount--;

	/*------------
	 * Free memory
	 *------------
	 */
	rc = vs_mem_free(lockInfoIdx);
	if (rc == NFX_OK)
	{
		/*----------------------------
		 * Update global debug counter
		 *----------------------------
		 */
		lock_globalCfgp->debugCounters.freeNlmLockInfo++;
	}

	return (rc);
} /* lock_memFreeNlmLockInfo() */

/*-----------------------------------------------------------------
 * Name:        lock_memFreeLockInfo
 *
 * Description:
 *  
 *     Free an existing structure of type lock_cifsLockInfo_t
 *     or lock_nlmLockInfo_t.
 *  
 * Arguments:    
 *      
 *     lockInfoIdx - Index to lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                   structure.
 *  
 *     lockInfop - Pointer to the lock_cifsLockInfo_t
 *                 or lock_nlmLockInfo_t structure.
 *                 This may be NULL.
 *  
 * Return value:
 *      
 *     NFX_OK - If succeed.
 * 
 *     NFX_ERR - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 11/23/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_memFreeLockInfo(vs_mem_index_t	lockInfoIdx,
					 void			*lockInfop)
{
	int32	rc;
	uint8	ownerType;
	
	ownerType = lock_getOwnerTypeByLockInfoIdx(lockInfoIdx);
	
	switch (ownerType)
	{
		case LOCK_OWNER_CIFS:
			rc = lock_memFreeCifsLockInfo(lockInfoIdx, 
										  (lock_cifsLockInfo_t *)lockInfop);
			break;

		case LOCK_OWNER_NLM:
		{
			boolean	delete;

			/*Remove the NLM lock record from the stable storage if necessary*/
			rc = lock_nlmClusRemoveLock(lockInfoIdx,
										(lock_nlmLockInfo_t *)lockInfop,
										&delete);
			if (rc != NFX_OK)
				return (rc);

			if (delete == TRUE)
			{
				rc = lock_memFreeNlmLockInfo(lockInfoIdx,
							   				 (lock_nlmLockInfo_t *)lockInfop);
			}
		}
			break;

		default:
			break;
	}

	return (rc);
} /* lock_memFreeLockInfo() */

/*-----------------------------------------------------------------
 * Name:        lock_memAllocExLock
 *
 * Description:
 *  
 *     Allocate and initialize a structure of type lock_exLock_t.
 *  
 * Arguments:    
 *      
 *     lockInfoIdx - Index to lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                   structure. The allocated exclusive lock structure
 *                   has this lock info.
 *  
 *     exLockIdxp - This is output. It is pointer to the index to
 *                  the allocated structure.
 *  
 * Return value:
 *      
 *     Pointer to the allocated lock_exLock_t structure - If succeed.
 * 
 *     NULL - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 11/23/2002
 *
 *-----------------------------------------------------------------
 */
lock_exLock_t *
lock_memAllocExLock(vs_mem_index_t	lockInfoIdx,
					vs_mem_index_t	*exLockIdxp)
{
	lock_exLock_t	*exLockp;

	/*----------------------------------------
	 * Allocate a lock_exLock_t data structure
	 *----------------------------------------
	 */
	exLockp = (lock_exLock_t *)vs_mem_alloc(0,			//FIXME
				   							vs_mem_dt_exLock,
											exLockIdxp);
	if (exLockp == NULL)
	{
		lock_globalCfgp->errCounters.insufficientMemory++;
		return (NULL);
	}

	/*------------------------
	 * Init the data structure
	 *------------------------
	 */
	VS_MEM_IDX_SET_NULL(exLockp->parent);
	VS_MEM_IDX_SET_NULL(exLockp->child[0]);
	VS_MEM_IDX_SET_NULL(exLockp->child[1]);
	exLockp->info = lockInfoIdx;

	/*----------------------
	 * Update global counter
	 *----------------------
	 */
	lock_globalCfgp->debugCounters.allocExLock++;

	return (exLockp);
} /* lock_memAllocExLock */

/*-----------------------------------------------------------------
 * Name:        lock_memFreeExLock
 *
 * Description:
 *  
 *      Free an existing structure of type lock_exLock_t.
 *  
 * Arguments:    
 *
 *     exLockIdx - Index to the lock_exLock_t structure to be freed.
 *                 This index must be valid.
 *  
 *     exLockp - Pointer to the lock_exLock_t structure to be freed.  
 *               The value of this argument can be NULL.  
 *      
 * Return value:
 *      
 *     NFX_OK - If succeed.
 * 
 *     NFX_ERR - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 11/23/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_memFreeExLock(vs_mem_index_t	exLockIdx,
				   lock_exLock_t	*exLockp)
{
	int32				rc;

	/*------------
	 * Find exLock
	 *------------
	 */
	if (exLockp == NULL)
	{
		exLockp = (lock_exLock_t *)vs_mem_idxToPtr(exLockIdx);
		if (exLockp == NULL)
		{
			return (NFX_ERR);
		}
	}

	/*------------------------------
	 * Free the associated lock info
	 *------------------------------
	 */
	rc = lock_memFreeLockInfo(exLockp->info, NULL);
	if (rc != NFX_OK)
		return (NFX_ERR);

	/*------------
	 * Free memory
	 *------------
	 */
	rc = vs_mem_free(exLockIdx);
	if (rc == NFX_OK)
	{
		/* Update global debug counter */
		lock_globalCfgp->debugCounters.freeExLock++;
	}

	return (rc);
} /* lock_memFreeExLock() */

/*-----------------------------------------------------------------
 * Name:        lock_memAllocShLock
 *
 * Description:
 *  
 *     Allocate and initialize a structure of type lock_shLock_t.
 *  
 * Arguments:    
 *      
 *     lockInfoIdx - Index to lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                   structure. The allocated shared lock structure
 *                   has this lock info.
 *  
 *     shLockIdxp - This is output. It is pointer to the index to
 *                  the allocated structure.
 *  
 * Return value:
 *      
 *     Pointer to the allocated lock_shLock_t structure - If succeed.
 * 
 *     NULL - If fail.
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 12/14/2002
 *
 *-----------------------------------------------------------------
 */
lock_shLock_t *
lock_memAllocShLock(vs_mem_index_t	lockInfoIdx,
					vs_mem_index_t	*shLockIdxp)
{
	lock_shLock_t	*shLockp;

	/*----------------------------------------
	 * Allocate a lock_shLock_t data structure
	 *----------------------------------------
	 */
	shLockp = (lock_shLock_t *)vs_mem_alloc(0,			//FIXME
				   							vs_mem_dt_shLock,
											shLockIdxp);
	if (shLockp == NULL)
	{
		lock_globalCfgp->errCounters.insufficientMemory++;
		return (NULL);
	}

	/*------------------------
	 * Init the data structure
	 *------------------------
	 */
	VS_MEM_IDX_SET_NULL(shLockp->link);
	shLockp->info = lockInfoIdx;

	/*----------------------
	 * Update global counter
	 *----------------------
	 */
	lock_globalCfgp->debugCounters.allocShLock++;

	return (shLockp);
} /* lock_memAllocShLock */

/*-----------------------------------------------------------------
 * Name:        lock_memFreeShLock
 *
 * Description:
 *  
 *      Free an existing structure of type lock_shLock_t.
 *  
 * Arguments:    
 *
 *     shLockIdx - Index to the lock_shLock_t structure to be freed.
 *                 This index must be valid.
 *  
 *     shLockp - Pointer to the lock_shLock_t structure to be freed.  
 *               The value of this argument can be NULL.  
 *      
 * Return value:
 *      
 *     NFX_OK - If succeed.
 * 
 *     NFX_ERR - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 12/14/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_memFreeShLock(vs_mem_index_t	shLockIdx,
				   lock_shLock_t	*shLockp)
{
	int32				rc;

	/*------------
	 * Find shLock
	 *------------
	 */
	if (shLockp == NULL)
	{
		shLockp = (lock_shLock_t *)vs_mem_idxToPtr(shLockIdx);
		if (shLockp == NULL)
		{
			return (NFX_ERR);
		}
	}

	/*------------------------------
	 * Free the associated lock info
	 *------------------------------
	 */
	rc = lock_memFreeLockInfo(shLockp->info, NULL);
	if (rc != NFX_OK)
		return (NFX_ERR);

	/*------------
	 * Free memory
	 *------------
	 */
	rc = vs_mem_free(shLockIdx);
	if (rc == NFX_OK)
	{
		/* Update global debug counter */
		lock_globalCfgp->debugCounters.freeShLock++;
	}

	return (rc);
} /* lock_memFreeShLock() */

/*-----------------------------------------------------------------
 * Name:        lock_memAllocShNode
 *
 * Description:
 *  
 *     Allocate and initialize a structure of type lock_shNode_t.
 *  
 * Arguments:    
 *      
 *     shLockIdx - Index to lock_shLock_t structure. This shared lock
 *                 will become the first lock in the lock list of newly
 *                 allocated shared lock node.
 *      
 *     shLockp - Pointer to the above lock_shLock_t structure.
 *  
 *     shNodeIdxp - This is output. It is pointer to the index to
 *                  the allocated structure.
 *  
 * Return value:
 *      
 *     Pointer to the allocated lock_shNode_t structure - If succeed.
 * 
 *     NULL - If fail.
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 12/14/2002
 *
 *-----------------------------------------------------------------
 */
lock_shNode_t *
lock_memAllocShNode(vs_mem_index_t	shLockIdx,
					lock_shLock_t	*shLockp,
					vs_mem_index_t	*shNodeIdxp)
{
	lock_shNode_t	*shNodep;
	uint8			shLockOwnerType;
	uint64			offset, len;

	/*----------------------------------------
	 * Allocate a lock_shNode_t data structure
	 *----------------------------------------
	 */
	shNodep = (lock_shNode_t *)vs_mem_alloc(0,			//FIXME
				   							vs_mem_dt_shNode,
											shNodeIdxp);
	if (shNodep == NULL)
	{
		lock_globalCfgp->errCounters.insufficientMemory++;
		return (NULL);
	}

	/*------------------------
	 * Init the data structure
	 *------------------------
	 */
	bzero(shNodep, sizeof(lock_shNode_t));
	shNodep->lockList = shLockIdx;
	shNodep->lockListTail = shLockIdx;

	/* Get offset and len of shLock */
	lock_getLockInformationFromLockInfo(shLockp->info,
					                    NULL,
										&shLockOwnerType,
										&offset,
										&len);
	LOCK_GET_LOCK_RANGE_END(offset, len, shNodep->end);

	LOCK_SHNODE_INC_COUNT(shNodep, shLockOwnerType);

	/*----------------------
	 * Update global counter
	 *----------------------
	 */
	lock_globalCfgp->debugCounters.allocShNode++;

	return (shNodep);
} /* lock_memAllocShNode */

/*-----------------------------------------------------------------
 * Name:        lock_memFreeShNode
 *
 * Description:
 *  
 *      Free an existing structure of type lock_shNode_t.
 *  
 * Arguments:    
 *
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     shLockIdx - Index to the lock_shNode_t structure to be freed.
 *                 This index must be valid.
 *  
 *     shLockp - Pointer to the lock_shNode_t structure to be freed.  
 *               The value of this argument can be NULL.  
 *
 *     freeShLock - This is meaningful for a shared node. 
 *                  TRUE if the associated shared locks are also to be freed.
 * Return value:
 *      
 *     NFX_OK - If succeed.
 * 
 *     NFX_ERR - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 12/14/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_memFreeShNode(lock_fileLock_t	*fileLockp,
				   vs_mem_index_t	shNodeIdx,
				   lock_shNode_t	*shNodep,
				   boolean			freeShLock)
{
	int32			rc;
	vs_mem_index_t	shLockIdx;
	lock_shLock_t	*shLockp;
	vs_mem_index_t	lockInfoIdx;
	uint8			lockOwnerType;

	/*------------
	 * Find shNode
	 *------------
	 */
	if (shNodep == NULL)
	{
		shNodep = (lock_shNode_t *)vs_mem_idxToPtr(shNodeIdx);
		if (shNodep == NULL)
		{
			return (NFX_ERR);
		}
	}

	/*--------------------------------
	 * Free the associated linked list
	 *--------------------------------
	 */
	if (freeShLock == TRUE)
	{
		shLockIdx = shNodep->lockList;
		while(!VS_MEM_IDX_IS_NULL(shLockIdx))
		{
			/* Get current shLock structure in the list*/
			shLockp = (lock_shLock_t *)vs_mem_idxToPtr(shLockIdx);
			if (shLockp == NULL)
				return (NFX_ERR);

			/* Get lock owner type */
			lockInfoIdx = lock_getLockInfoIdxByLockIdx(shLockIdx);
			lockOwnerType = lock_getOwnerTypeByLockInfoIdx(lockInfoIdx);

			/*--------------
			 * Free the lock 
			 *--------------
			 */
			rc = lock_memFreeShLock(shLockIdx, shLockp);
			if (rc != NFX_OK)
				return (NFX_ERR);

			/*-----------------------
			 * Decrement nlmLockCount
			 *-----------------------
			 */
			if (lockOwnerType == LOCK_OWNER_NLM)
				lock_decNlmLockCntOfFile(fileLockp);

			/*----------------------------
			 * Update global lock counters
			 *----------------------------
			 */
			LOCK_DEC_LOCK_COUNTERS(
						lockOwnerType,
						lock_globalCfgp->lockCounters.grantedSharedCifsLocks,
						lock_globalCfgp->lockCounters.grantedSharedNlmLocks);

			shLockIdx = shLockp->link;
		}
	}

	/*------------
	 * Free memory
	 *------------
	 */
	rc = vs_mem_free(shNodeIdx);
	if (rc == NFX_OK)
	{
		/* Update global debug counter */
		lock_globalCfgp->debugCounters.freeShNode++;
	}

	return (rc);
} /* lock_memFreeShNode() */

/*-----------------------------------------------------------------
 * Name:        lock_memAllocPendingLock
 *
 * Description:
 *  
 *     Allocate and initialize a structure of type lock_pendingLock_t.
 *  
 * Arguments:    
 *      
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     lockInfoIdx - Index to lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                   structure. The allocated pending lock structure
 *                   has this lock info.
 *  
 *     lockType - The pending lock is shared or exclusive.
 *
 *     timeout - Time, in real time clock form, to cancel the lock if it is 
 *               blocked by other locks.
 *  
 *     ulPrivate - The lock request structure passed from upper (protocol)
 *                 layer.
 *
 *     pendingLockIdxp - This is output. It is pointer to the index to
 *                       the allocated structure.
 *  
 * Return value:
 *      
 *     Pointer to the allocated lock_pendingLock_t structure - If succeed.
 * 
 *     NULL - If fail.
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 12/16/2002
 *
 *-----------------------------------------------------------------
 */
lock_pendingLock_t *
lock_memAllocPendingLock(lock_fileLock_t	*fileLockp,
						 vs_mem_index_t		lockInfoIdx,
						 uint8				lockType,
						 eeeRTC_t			timeout,
						 vs_mem_index_t		ulPrivate,
						 vs_mem_index_t		*pendLockIdxp)
{
	vs_mem_index_t		fileLockIdx;
	lock_pendingLock_t	*pendLockp;

	/*---------------------------------------------
	 * Allocate a lock_pendingLock_t data structure
	 *---------------------------------------------
	 */
	pendLockp = (lock_pendingLock_t *)vs_mem_alloc(0,			//FIXME
				   								   vs_mem_dt_pendingLock,
												   pendLockIdxp);
	if (pendLockp == NULL)
	{
		lock_globalCfgp->errCounters.insufficientMemory++;
		return (NULL);
	}

	/*----------------
	 * Get fileLockIdx
	 *----------------
	 */
	fileLockIdx = vs_mem_ptrToIdx(lockInfoIdx.srvrIdx,
								  vs_mem_dt_fileLock,
								  (void *)fileLockp);
	if (!VS_MEM_IDX_IS_VALID(fileLockIdx))
		return (NULL);

	/*------------------------
	 * Init the data structure
	 *------------------------
	 */
	bzero(pendLockp, sizeof(lock_pendingLock_t));
	pendLockp->fileLock = fileLockIdx;
	pendLockp->ulPrivate = ulPrivate;
	pendLockp->info = lockInfoIdx;
	pendLockp->timeout = timeout;
	pendLockp->type = lockType;

	/*----------------------
	 * Update global counter
	 *----------------------
	 */
	lock_globalCfgp->debugCounters.allocPendingLock++;

	return (pendLockp);
} /* lock_memAllocPendingLock */

/*-----------------------------------------------------------------
 * Name:        lock_memFreePendingLock
 *
 * Description:
 *  
 *      Free an existing structure of type lock_pendingLock_t.
 *  
 * Arguments:    
 *
 *     pendLockIdx - Index to the lock_pendingLock_t structure to be freed.
 *                   This index must be valid.
 *  
 *     pendLockp - Pointer to the lock_pendingLock_t structure to be freed.  
 *                 The value of this argument can be NULL.  
 *
 *     delLockInfo - TRUE if the assocated lockInfo structure is to be deleted.
 *                   FALSE if not to delete the associatd lockInfo structure.
 *      
 * Return value:
 *      
 *     NFX_OK - If succeed.
 * 
 *     NFX_ERR - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 12/16/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_memFreePendingLock(vs_mem_index_t	  pendLockIdx,
					   lock_pendingLock_t *pendLockp,
					   boolean			  delLockInfo)
{
	int32			rc;

	/*------------
	 * Free memory
	 *------------
	 */
	rc = vs_mem_free(pendLockIdx);
	if (rc == NFX_OK)
	{
		/* Update global debug counter */
		lock_globalCfgp->debugCounters.freePendingLock++;
	}

	if (delLockInfo == TRUE)
	{
		if (pendLockp == NULL)
		{
			pendLockp = (lock_pendingLock_t *)vs_mem_idxToPtr(pendLockIdx);
			if (pendLockp == NULL)
				return (NFX_ERR);
		}
		rc = lock_memFreeLockInfo(pendLockp->info, NULL);
	}

	return (rc);
} /* lock_memFreePendingLock() */

/*-----------------------------------------------------------------
 * Name:        lock_memAllocFileLock
 *
 * Description:
 *  
 *     Allocate and initialize a structure of type lock_fileLock_t.
 *  
 * Arguments:    
 *      
 *     fileLockIdxp - This is output. It is pointer to the index to
 *                    the newly allocated lock_fileLock_t structure.
 *  
 * Return value:
 *      
 *     Pointer to the allocated lock_fileLock_t - If succeed.
 * 
 *     NULL - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 11/23/2002
 *
 *-----------------------------------------------------------------
 */
lock_fileLock_t *   
lock_memAllocFileLock(vs_mem_index_t	*fileLockIdxp)
{
	lock_fileLock_t	*fileLockp;

	/*------------------------------------------
	 * Allocate a lock_fileLock_t data structure
	 *------------------------------------------
	 */
	fileLockp = (lock_fileLock_t *)vs_mem_alloc(0, 		//FIXME: srvrIdx
												vs_mem_dt_fileLock,
												fileLockIdxp);
	if (fileLockp == NULL)
	{
		lock_globalCfgp->errCounters.insufficientMemory++;
		return (NULL);
	}

	/*------------------------
	 * Init the data structure
	 *------------------------
	 */
	fileLockp->lowestMandatoryLockOffset = LOCK_MAX_LOWEST_MAND_OFFSET;
	VS_MEM_IDX_SET_NULL(fileLockp->exclLockTree);
	VS_MEM_IDX_SET_NULL(fileLockp->sharedLockTree);
	VS_MEM_IDX_SET_NULL(fileLockp->pendingLockList);
	VS_MEM_IDX_SET_NULL(fileLockp->lastGrantedLock);
	fileLockp->nlmLockCount = 0;

	/*----------------------
	 * Update global counter
	 *----------------------
	 */
	lock_globalCfgp->debugCounters.allocFileLock++;

	return (fileLockp);
} /* lock_memAllocFileLock() */

/*-----------------------------------------------------------------
 * Name:        lock_memFreeFileLock
 *
 * Description:
 *  
 *     Free a data structure of type lock_fileLock_t.
 *  
 * Arguments:    
 *      
 *     fileLockIdxp - Pointer to the index to the lock_fileLock_t
 *                    structure to be freed.
 *                    This index must be valid.
 *  
 *     fileLockp - Pointer to the lock_fileLock_t structure to be freed.  
 *                 The value of this argument can be NULL.  
 *  
 * Return value:
 *      
 *     NFX_OK - If succeed.
 * 
 *     NFX_ERR - If fail. 
 * 
 * Created by:   Xiaobing Zhang
 *      
 * Date Created: 11/23/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_memFreeFileLock(vs_mem_index_t		*fileLockIdxp,
					 lock_fileLock_t	*fileLockp)
{
	int32	rc;

	/*---------------------------------
	 * Find the fileLock data structure
	 *---------------------------------
	 */
	if (fileLockp == NULL)
	{
		fileLockp = (lock_fileLock_t *)vs_mem_idxToPtr(*fileLockIdxp);
		if (fileLockp == NULL)
			return (NFX_ERR);
	}

	/*-------------------------------
	 * Delete the exclusive lock tree 
	 *-------------------------------
	 */
	rc = lock_btree_deleteTree(fileLockp, LOCK_TYPE_EXCL);
	if (rc != NFX_OK)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_memFreeFileLock: "
								"Failed to delete the exclusive lock tree."));
		return (NFX_ERR);
	}

	/*----------------------------
	 * Delete the shared lock tree 
	 *----------------------------
	 */
	rc = lock_btree_deleteTree(fileLockp, LOCK_TYPE_SHAR);
	if (rc != NFX_OK)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_memFreeFileLock: "
								"Failed to delete the shared lock tree."));
		return (NFX_ERR);
	}

	/*-----------------------------
	 * Delete the pending lock list 
	 *-----------------------------
	 */
	rc = lock_deletePendingList(&(fileLockp->pendingLockList));
	if (rc != NFX_OK)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_memFreeFileLock: "
								"Failed to delete the pending lock list."));
		return (NFX_ERR);
	}

	/*---------------------------------
	 * Free the fileLock data structure
	 *---------------------------------
	 */
	rc = vs_mem_free(*fileLockIdxp);
	if (rc != NFX_OK)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_memFreeFileLock: "
								"Failed to free the fileLock structure."));
		return (NFX_ERR);
	}
	VS_MEM_IDX_SET_NULL(*fileLockIdxp);

	/* Update glogal debug counter */
	lock_globalCfgp->debugCounters.freeFileLock++;

	return (NFX_OK);
} /* lock_memFreeFileLock() */


