/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        lock-proc.c
 *
 * RCS:  $Id: $
 *
 * Copyright (C) 2002, ClariStor Inc.
 *
 * Description: Process lock/unlock/cancel lock requests.
 *
 * Created by:  Xiaobing Zhang
 *
 * Date Created:    12/03/2002
 *
 *-----------------------------------------------------------------
 */

#include "lock-incls.h"

/*-----------------------------------------------------------------
 * Name:         lock_addExclLock
 *
 * Description:
 *
 *     Add a lock to the exclusive lock tree.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     newLockInfoIdx - Index to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                      data structure, which contains information of granted
 *                      exclusive lock.
 *
 *     newLockInfop - Pointer to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                    data structure, which contains information of granted
 *                    exclusive lock.
 *
 *     newLockOwnerType - The type of owner (LOCK_OWNER_CIFS or LOCK_OWNER_NLM)
 *                        of the granted exclusive lock.
 *
 *         As to the above three arguments, the caller either,
 *             - provides newLockInfoIdx and
 *               leaves newLockInfop and newLockOwnerType invalid, or
 *             - provides newLockInfop and newLockOwnerType and
 *               leaves newLockInfoIdx invalid.
 *
 *     exLockIdxp - This is output. It contains index of lock_exLock_t data
 *                  structure, which is added to the exclusive lock tree.
 *  
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/04/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_addExclLock(lock_fileLock_t	*fileLockp,
				 vs_mem_index_t		newLockInfoIdx,
				 void				*newLockInfop,
				 uint8				newLockOwnerType,
				 vs_mem_index_t		*exLockIdxp)
{
	int32			rc;
	vs_mem_index_t	*rootIdxp;
	vs_mem_index_t	lockInfoIdx;
	void			*lockInfop;
	lock_exLock_t	*exLockp;
	uint16			searchMask = 0; /*Look for locks of overlapping ranges*/

	if (!VS_MEM_IDX_IS_NULL(newLockInfoIdx))
	{
		/* Find the existing lockInfo data structue */
		newLockInfop = lock_getLockInfo(newLockInfoIdx, &newLockOwnerType);
		if (newLockInfop == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_addExclLock: "
							"Failed to find the lockInfo data structure."));
			return (NFX_ERR);
		}
	}

	rootIdxp = &(fileLockp->exclLockTree);

	/*--------------------------------------------------------------------
	 * If the new lock overlaps with any existing locks of the same owner,
	 * merge them.
	 *--------------------------------------------------------------------
	 */
	rc = lock_mergeExLocks(fileLockp,
						   newLockInfop,
						   newLockOwnerType,
						   exLockIdxp,
						   &exLockp);

	if (rc != NFX_OK)
		return (rc);

	if (exLockp != NULL)
	{
		/* Merged the new lock with existing locks */

		if (!VS_MEM_IDX_IS_NULL(newLockInfoIdx))
		{
			/* Since the lock is merged with other locks,
			 * the lockInfo structure holding info of this lock
			 * shold be deleted
			 */
			rc = lock_memFreeLockInfo(newLockInfoIdx, newLockInfop);
		}
		return (NFX_OK);
	}

	/*------------------------------
	 * Add a exLock node to the tree
	 *------------------------------
	 */
	if (!VS_MEM_IDX_IS_NULL(newLockInfoIdx))
	{
		lockInfoIdx = newLockInfoIdx;
	}
	else
	{
		/* Create a lockInfo data structue */
		lockInfop = lock_createLockInfo(newLockInfop,
					   					newLockOwnerType,
									   	&lockInfoIdx);
		if (lockInfop == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_addExclLock: "
							"Failed to create a lockInfo data structure."));
			return (NFX_ERR);
		}
	}

	/* Create a lock_exLock_t data structure */
	exLockp = (lock_exLock_t *)lock_memAllocExLock(lockInfoIdx, exLockIdxp);
	if (exLockp == NULL)
	{
		lock_memFreeLockInfo(lockInfoIdx, lockInfop);
		return (NFX_ERR);
	}

	LOCK_PRINTF(LOG_DEBUG, ("lock_addExclLock: "
				"exLockIdx(%d, %d, %d)",
				exLockIdxp->srvrIdx, exLockIdxp->dataType, exLockIdxp->idx));

	/* Add the lock to the exclusive lock tree */
	rc = lock_btree_insert(rootIdxp,
						   *exLockIdxp,
						   (lock_btNode_t *)exLockp,
						   searchMask,
						   LOCK_TYPE_EXCL);
	if (rc != NFX_OK)
		lock_memFreeExLock(*exLockIdxp, exLockp);
	else
	{
		/*-----------------------
		 * Increment nlmLockCount
		 *-----------------------
		 */
		if (newLockOwnerType == LOCK_OWNER_NLM)
			lock_incNlmLockCntOfFile(fileLockp);

		/*----------------------------
		 * Update global lock counters
		 *----------------------------
		 */
		LOCK_INC_LOCK_COUNTERS(
					newLockOwnerType,
				   	lock_globalCfgp->lockCounters.grantedExclCifsLocks,
				   	lock_globalCfgp->lockCounters.grantedExclNlmLocks);

		/*------------------------------------------------------------
		 * Save the NLM lock record on the stable storage if necessary
		 *------------------------------------------------------------
		 */
		if (newLockOwnerType == LOCK_OWNER_NLM)
		{
			rc = lock_nlmClusSaveLock(lockInfoIdx,
									  lockInfop,
									  LOCK_TYPE_EXCL,
									  LOCK_STATE_GRANTED);
			/*TODO: If rc is NFX_ERR, remove the lock? */
		}
	}

	return (rc);
} /* lock_addExclLock() */

/*-----------------------------------------------------------------
 * Name:         lock_addSharedLock
 *
 * Description:
 *
 *     Add a lock to the shared lock tree.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     newLockInfoIdx - Index to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                      data structure, which contains information of granted
 *                      shared lock.
 *
 *     newLockInfop - Pointer to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                    data structure, which contains information of granted
 *                    shared lock.
 *
 *     newLockOwnerType - The type of owner (LOCK_OWNER_CIFS or LOCK_OWNER_NLM)
 *                        of the granted shared lock.
 *
 *         As to the above three arguments, the caller either,
 *             - provides newLockInfoIdx and
 *               leaves newLockInfop and newLockOwnerType invalid, or
 *             - provides newLockInfop and newLockOwnerType and
 *               leaves newLockInfoIdx invalid.
 *
 *     shLockIdxp - This is output. It contains index of lock_shLock_t data
 *                  structure, which is added to the shared lock tree.
 *  
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/14/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_addSharedLock(lock_fileLock_t	*fileLockp,
				   vs_mem_index_t	newLockInfoIdx,
				   void				*newLockInfop,
				   uint8			newLockOwnerType,
				   vs_mem_index_t	*shLockIdxp)
{
	int32			rc;
	vs_mem_index_t	*rootIdxp;
	uint16			searchMask;
	vs_mem_index_t	lockInfoIdx;
	void			*lockInfop;
	lock_shLock_t	*shLockp;
	vs_mem_index_t	shNodeIdx;
	lock_shNode_t	*shNodep;
	
	rootIdxp = &(fileLockp->sharedLockTree);

	/*-------------------------------------------------------------
	 * Create a lock_shLock_t data structue to be added in the tree
	 *-------------------------------------------------------------
	 */
	if (!VS_MEM_IDX_IS_NULL(newLockInfoIdx))
	{
		/* Find the existing lockInfo data structue */
		lockInfoIdx = newLockInfoIdx;
		lockInfop = lock_getLockInfo(lockInfoIdx, &newLockOwnerType);
		if (lockInfop == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_addSharedLock: "
							"Failed to find lockInfo data structure."));
			return (NFX_ERR);
		}
	}
	else
	{
		/* Create a new lockInfo data structue */
		lockInfop = lock_createLockInfo(newLockInfop,
										newLockOwnerType,
									   	&lockInfoIdx);
		if (lockInfop == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_addSharedLock: "
							"Failed to create a lockInfo data structure."));
			return (NFX_ERR);
		}
	}

	/* Create a lock_shLock_t data structue */
	shLockp = (lock_shLock_t *)lock_memAllocShLock(lockInfoIdx, shLockIdxp);
	if (shLockp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_addSharedLock: "
						"Failed to allocate a shLock data structure."));
		if (VS_MEM_IDX_IS_NULL(newLockInfoIdx))
		{
			/* Free the allocated lockInfo structure */
			lock_memFreeLockInfo(lockInfoIdx, lockInfop);
		}
		return (NFX_ERR);
	}

	LOCK_PRINTF(LOG_DEBUG, ("lock_addSharedLock: "
				"shLockIdx(%d, %d, %d)",
				shLockIdxp->srvrIdx, shLockIdxp->dataType, shLockIdxp->idx));

	/*------------------------------------------------------------
	 * Check if the new lock overlaps with an existing node or not
	 *------------------------------------------------------------
	 */
	searchMask = 0;
	shNodep = lock_findSharedNodeByLockInfo(rootIdxp,
											lockInfop,
											newLockOwnerType,
											searchMask);
	if (shNodep != NULL)
	{
		/*---------------------------------------------------
		 * Overlap with an existing node.
		 * Add the new lock to the linked list of found node.
		 *---------------------------------------------------
		 */
		rc = lock_addSharedLockToNode(shNodep,
									  *shLockIdxp,
									  shLockp,
									  lockInfoIdx,
									  lockInfop,
									  newLockOwnerType);
		if (rc != NFX_OK)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_addSharedLock: "
						"Failed to add the lock to the shared lock list."));
			lock_memFreeShLock(*shLockIdxp, shLockp);
			return (NFX_ERR);
		}

		/*--------------------------------------------
		 * Merge the overlapping nodes if there is any
		 *--------------------------------------------
		 */
		shNodep = lock_mergeShNodes(fileLockp, shNodep);

		/*-----------------------------------------------------------
		 * Try to split the node if it is "holey".
		 * Note: After adding the new lock, the previous "holey" node
		 *       may become contiguous.
		 *-----------------------------------------------------------
		 */
		if (shNodep->lockCntAndHoley.holey)
		{
			rc = lock_splitSharedNode(fileLockp, *rootIdxp, shNodep);
			if (rc != NFX_OK)
			{
				LOCK_PRINTF(LOG_ERR, ("lock_addSharedLock: "
							"Error in splitting the \"holey\" shared node."));
			}
		}
	}
	else
	{
		/*--------------------------------------------
		 * Insert a new shNode in the shared lock tree
		 *--------------------------------------------
		 */	
		shNodep = lock_memAllocShNode(*shLockIdxp,
									  shLockp,
									  &shNodeIdx);
		if (shNodep == NULL)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_addSharedLock: "
							"Failed to allocate a shNode data structure."));
			lock_memFreeShLock(*shLockIdxp, shLockp);
			return (NFX_ERR);
		}

		searchMask = 0;
		rc = lock_btree_insert(rootIdxp,
							   shNodeIdx,
							   (lock_btNode_t *)shNodep,
							   searchMask,
							   LOCK_TYPE_SHAR);
		if (rc != NFX_OK)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_addSharedLock: "
							"Failed to insert a shNode into shared lock tree."));
			lock_memFreeShNode(fileLockp, shNodeIdx, shNodep, TRUE);
			return (NFX_ERR);
		}
	}

	/*-----------------------
	 * Increment nlmLockCount
	 *-----------------------
	 */
	if (newLockOwnerType == LOCK_OWNER_NLM)
		lock_incNlmLockCntOfFile(fileLockp);

	/*----------------------------
	 * Update global lock counters
	 *----------------------------
	 */
	LOCK_INC_LOCK_COUNTERS(
					newLockOwnerType,
				   	lock_globalCfgp->lockCounters.grantedSharedCifsLocks,
				   	lock_globalCfgp->lockCounters.grantedSharedNlmLocks);

	/*------------------------------------------------------------
	 * Save the NLM lock record on the stable storage if necessary
	 *------------------------------------------------------------
	 */
	if (newLockOwnerType == LOCK_OWNER_NLM)
	{
		rc = lock_nlmClusSaveLock(lockInfoIdx,
								  lockInfop,
								  LOCK_TYPE_SHAR,
								  LOCK_STATE_GRANTED);
		/*TODO: If rc is NFX_ERR, remove the lock? */
	}

	return (rc);
} /* lock_addSharedLock() */

/*-----------------------------------------------------------------
 * Name:         lock_addGrantedLock
 *
 * Description:
 *
 *     Add a granted lock to exclusive or shared lock tree.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     newLockInfoIdx - Index to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                      data structure, which contains information of granted
 *                      lock.
 *
 *     newLockInfop - Pointer to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                    data structure, which contains information of granted
 *                    lock.
 *
 *     newLockOwnerType - The type of owner (LOCK_OWNER_CIFS or LOCK_OWNER_NLM)
 *                        of the granted lock.
 *
 *         As to the above three arguments, the caller either,
 *             - provides newLockInfoIdx and
 *               leaves newLockInfop and newLockOwnerType invalid, or
 *             - provides newLockInfop and newLockOwnerType and
 *               leaves newLockInfoIdx invalid.
 *
 *     newLockType - The type of granted lock (LOCK_TYPE_EXCL or
 *                   LOCK_TYPE_SHAR).
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/16/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_addGrantedLock(lock_fileLock_t *fileLockp,
					vs_mem_index_t	newLockInfoIdx,
					void			*newLockInfop,
					uint8			newLockOwnerType,
					uint8			newLockType)
{
	int32			rc;
	vs_mem_index_t	lockIdx;

	if (fileLockp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_addGrantedLock: "
								"NULL pointer to fileLock."));
		return (NFX_ERR);
	}

	if (VS_MEM_IDX_IS_NULL(newLockInfoIdx) && (newLockInfop == NULL))
	{
		LOCK_PRINTF(LOG_ERR, ("lock_addGrantedLock: "
								"NULL pointer to newLockInfo."));
		return (NFX_ERR);
	}

	switch (newLockType)
	{
		case LOCK_TYPE_EXCL:
			/*---------------------------------------
			 * Insert the lock to exclusive lock tree
			 *---------------------------------------
			 */
			rc = lock_addExclLock(fileLockp,
								  newLockInfoIdx,
						   		  newLockInfop,
								  newLockOwnerType,
								  &lockIdx);
			if (rc != NFX_OK)
			{
				LOCK_PRINTF(LOG_ERR, ("lock_addGrantedLock: "
							"Failed to add lock to exclusive lock tree."));
				return (NFX_ERR);
			}

			break;

		case LOCK_TYPE_SHAR:
			/*------------------------------------
			 * Insert the lock to shared lock tree
			 *------------------------------------
			 */
			rc = lock_addSharedLock(fileLockp,
									newLockInfoIdx,
									newLockInfop,
									newLockOwnerType,
									&lockIdx);
			if (rc != NFX_OK)
			{
				LOCK_PRINTF(LOG_ERR, ("lock_addGrantedLock: "
							"Failed to add lock to shared lock tree."));
				return (NFX_ERR);
			}

			break;

		default:
			LOCK_PRINTF(LOG_ERR, ("lock_addGrantedLock: "
											"Unsupported lock type."));
			return (NFX_ERR);
			break;
	}

    /*--------------------------------------------------------------------
	 * Update lowestMandatoryLockOffset field in lock_fileLock_t structure
	 *--------------------------------------------------------------------
	 */
	if (LOCK_IS_MANDATORY(newLockOwnerType))
	{
		lock_cifsLockInfo_t	*cifsLockInfop;

		if (!VS_MEM_IDX_IS_NULL(newLockInfoIdx))
			cifsLockInfop = (lock_cifsLockInfo_t *)vs_mem_idxToPtr(
															newLockInfoIdx);
		else
			cifsLockInfop = (lock_cifsLockInfo_t *)newLockInfop;

		/*We only need to decrease the value of lowestMandatoryLockOffset*/
		if (fileLockp->lowestMandatoryLockOffset > cifsLockInfop->offset)
			fileLockp->lowestMandatoryLockOffset = cifsLockInfop->offset;
	}

    /*----------------------------------------------------------
	 * Update lastGrantedLock field in lock_fileLock_t structure
	 *----------------------------------------------------------
	 */
	fileLockp->lastGrantedLock = lockIdx;

	return (NFX_OK);
} /* lock_addGrantedLock() */

/*-----------------------------------------------------------------
 * Name:         lock_addPendingLock
 *
 * Description:
 *
 *     Add a lock to the pending lock list.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     newLockInfop - Pointer to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                    data structure, which contains information of pending 
 *                    lock.
 *
 *     newLockOwnerType - The type of owner (LOCK_OWNER_CIFS or LOCK_OWNER_NLM)
 *                        of the pending lock.
 *
 *     newLockType - The type of granted lock (LOCK_TYPE_EXCL or
 *                   LOCK_TYPE_SHAR).
 *
 *     newLockTimeout - Time to cancel the pending lock. An indefinite timeout
 *                      value indicates to wait as long as it takes.
 *
 *     newLockUlPrivate - Private information of upper layer.
 *  
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/16/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_addPendingLock(lock_fileLock_t	*fileLockp,
					void			*newLockInfop,
					uint8			newLockOwnerType,
					uint8			newLockType,
					eeeRTC_t		newLockTimeout,
					vs_mem_index_t	newLockUlPrivate)
{
	int32				rc;
	vs_mem_index_t		*pendingListHeadp;
	vs_mem_index_t		lockInfoIdx;
	void				*lockInfop;
	vs_mem_index_t		pendLockIdx;
	lock_pendingLock_t	*pendLockp;

	pendingListHeadp = &(fileLockp->pendingLockList);

	/*-------------------------------------------------------------------
	 * Create a lock_pendingLock_t data structure to be added to the list
	 *-------------------------------------------------------------------
	 */
	/* Create a lockInfo data structue */
	lockInfop = lock_createLockInfo(newLockInfop,
				   					newLockOwnerType,
								   	&lockInfoIdx);
	if (lockInfop == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_addPendingLock: "
							"Failed to create a lockInfo data structure."));
		return (LOCK_ERR_GENERAL);
	}

	/* Create a lock_pendingLock_t data structue */
	pendLockp = (lock_pendingLock_t *)lock_memAllocPendingLock(
															fileLockp,
															lockInfoIdx,
															newLockType,
															newLockTimeout,
															newLockUlPrivate,
															&pendLockIdx);
	if (pendLockp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_addPendingLock: "
						"Failed to allocate a pendingLock data structure."));
		lock_memFreeLockInfo(lockInfoIdx, lockInfop);
		return (LOCK_ERR_INSUFFICIENT_MEMORY);
	}

	LOCK_PRINTF(LOG_DEBUG, ("lock_addPendingLock: "
				"pendLockIdx(%d, %d, %d)",
				pendLockIdx.srvrIdx, pendLockIdx.dataType, pendLockIdx.idx));

	/*------------------------------------------------------------
	 * Insert the new pending lock to the file's pending lock list
	 *------------------------------------------------------------
	 */
	rc = lock_insertPendLockToLocalList(pendingListHeadp,
										pendLockIdx,
										pendLockp);
	if (rc != NFX_OK)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_addPendingLock: "
							"Failed to add lock to local pending list."));
		lock_memFreePendingLock(pendLockIdx, pendLockp, TRUE);
		return (LOCK_ERR_GENERAL);
	}
	else
	{	/*----------------------------
		 * Update global lock counters
		 *----------------------------
		 */
		if (newLockType == LOCK_TYPE_EXCL)
		{
			LOCK_INC_LOCK_COUNTERS(
					newLockOwnerType,
				   	lock_globalCfgp->lockCounters.pendingExclCifsLocks,
				   	lock_globalCfgp->lockCounters.pendingExclNlmLocks);
		}
		else if (newLockType == LOCK_TYPE_SHAR)
		{
			LOCK_INC_LOCK_COUNTERS(
					newLockOwnerType,
				   	lock_globalCfgp->lockCounters.pendingSharedCifsLocks,
				   	lock_globalCfgp->lockCounters.pendingSharedNlmLocks);
		}

		/*------------------------------------------------------------
		 * Save the NLM lock record on the stable storage if necessary
		 *------------------------------------------------------------
		 */
		if (newLockOwnerType == LOCK_OWNER_NLM)
		{
			/*Save the NLM lock record on the stable storage if necessary*/
			rc = lock_nlmClusSaveLock(lockInfoIdx,
									  lockInfop,
									  newLockType,
									  LOCK_STATE_BLOCKED);
			/*TODO: If rc is NFX_ERR, remove the lock? */
		}
	}

	/*------------------------------------------------------------
	 * Insert the new pending lock to the global pending lock list
	 * if the lock is to wait for a definite length of time.
	 *------------------------------------------------------------
	 */
	if ((eee_rtcCompare(&newLockTimeout, &lock_timeZero) > 0) &&
		(eee_rtcCompare(&newLockTimeout, &lock_timeIndefinite) != 0))
	{
		rc = lock_insertPendLockToGlobalList(pendLockIdx, pendLockp);
		if (rc != NFX_OK)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_addPendingLock: "
								"Failed to add lock to global pending list."));
		}
	}

	return (NFX_OK);
} /* lock_addPendingLock() */

/*-----------------------------------------------------------------
 * Name:         lock_procLock
 *
 * Description:
 *
 *     Process lock request.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     lockInfop - Pointer to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                 data structure, which contains information of requested
 *                 lock.
 *
 *     lockOwnerType - The type of owner (LOCK_OWNER_CIFS or LOCK_OWNER_NLM)
 *                     who requests the lock.
 *
 *     lockType - Exclusive or shared lock (LOCK_TYPE_EXCL or LOCK_TYPE_SHAR).
 *
 *     newLockTimeout - Time to cancel the lock if the lock is blocked by
 *                      other locks. An indefinite timeout value indicates
 *                      to wait as long as it takes. A value of 0 indicates
 *                      not to wait.
 *
 *     ulPrivate - Private information of upper layer.
 *  
 * Return value:
 *
 *     LOCK_STATE_GRANTED, LOCK_STATE_DENIED, or LOCK_STATE_BLOCKED -
 *                                                                  If succeed.
 *
 *     lock error code - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/03/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_procLock(lock_fileLock_t *fileLockp,
			  void            *lockInfop,     /*Type:lock_cifsLockInfo_t or
												     lock_nlmLockInfo_t*/
			  uint8           lockOwnerType,
			  uint8			  lockType,
			  eeeRTC_t		  timeout,
			  vs_mem_index_t  ulPrivate)
{
	int32			rc, exclResult, sharResult, result;
	boolean			block;
	vs_mem_index_t	dummyLockInfoIdx, dummyLockIdx;
	vs_mem_index_t	lockOwnerIdx;

	block = (eee_rtcCompare(&timeout, &lock_timeZero) != 0) ? TRUE : FALSE;
	result = lock_checkConflict(fileLockp,
								lockInfop,
								lockOwnerType,
								lockType,
								block,
								&dummyLockIdx,
								&exclResult,
								&sharResult);

	VS_MEM_IDX_SET_NULL(dummyLockInfoIdx);

	if (result == LOCK_STATE_GRANTED)
	{
		/*-------------------------------------------------
		 * Insert the lock to exclusive or shared lock tree
		 *-------------------------------------------------
		 */
		rc = lock_addGrantedLock(fileLockp,
								 dummyLockInfoIdx,
								 lockInfop,
								 lockOwnerType,
								 lockType);
		if (rc != NFX_OK)
			return (LOCK_ERR_GENERAL);
	}
	else if (result == LOCK_STATE_BLOCKED)
	{
		vs_mem_index_t		dependencyListIdx;
		lock_dependency_t	*dependencyListp;

		if (lockOwnerType == LOCK_OWNER_NLM)
		{
			boolean			checkExclTree = FALSE;
		    boolean			checkSharedTree = FALSE;
			boolean			deadlock;

			if (exclResult == LOCK_STATE_BLOCKED)
				checkExclTree = TRUE;

			if (sharResult == LOCK_STATE_BLOCKED)
				checkSharedTree = TRUE;

			/*---------------------------------------------------
			 * Create a list of lock owners that block this lock.
			 *---------------------------------------------------
			 */
			dependencyListp = lock_createDependencyListOfPendLock(
														fileLockp,
														lockInfop,
														lockType,
														checkExclTree,
														checkSharedTree,
														&dependencyListIdx);
			if (dependencyListp == NULL)
			{
				LOCK_PRINTF(LOG_ERR, ("lock_procLock: "
					"Failed to create dependency list of the blocked lock."));
				return (LOCK_STATE_DENIED);
			}

			/* Get lockOwnerIdx */
			lockOwnerIdx = lock_getLockOwnerIdxByLockInfo(dummyLockInfoIdx,
														  lockInfop,
														  lockOwnerType);
			if (VS_MEM_IDX_IS_NULL(lockOwnerIdx))
			{
				LOCK_PRINTF(LOG_ERR, ("lock_procLock: "
					"Failed to get the index to the pending lock's owner."));
				return (LOCK_STATE_DENIED);
			}

			/*----------------
			 * Detect deadlock
			 *----------------
			 */
		 	deadlock = lock_detectDeadlock(dependencyListp, lockOwnerIdx);
		 	if (deadlock == TRUE)
			{
				/*----------------------
				 * Update global counter
				 *----------------------
				 */
				lock_globalCfgp->debugCounters.detectedDeadlocks++;

		 		return (LOCK_STATE_DENIED);
			}
		} /* if (lockOwnerType) */

		/*-----------------------------
		 * Add the lock to pending list
		 *-----------------------------
		 */
		rc = lock_addPendingLock(fileLockp,
								 lockInfop,
								 lockOwnerType,
								 lockType,
								 timeout,
								 ulPrivate);
		if (rc != NFX_OK)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_procLock: "
								"Failed to add lock to pending list."));
			return (LOCK_STATE_DENIED);
		}

		if (lockOwnerType == LOCK_OWNER_NLM)
		{
			/*---------------------------
			 * Update the dependency list
			 *---------------------------
			 */
			lock_incDependencyListOfBlockedOwner(lockOwnerIdx,
												 dependencyListIdx,
												 dependencyListp);
		}
	}

	return (result);
} /* lock_procLock() */

/*-----------------------------------------------------------------
 * Name:         lock_delCifsExclLock
 *
 * Description:
 *
 *     Delete a CIFS lock from the exclusive lock tree.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     lockInfop - Pointer to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                 data structure, which contains information of lock
 *                 to be deleted.
 *
 *     exLockp - Pointer to the exclusive lock to be deleted. 
 *  
 *         Either lockInfop or exLockp is valid, but not both.
 *
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/18/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_delCifsExclLock(lock_fileLock_t		*fileLockp,
					 lock_cifsLockInfo_t	*lockInfop,
					 lock_exLock_t			*exLockp)
{
	int32			rc;
	vs_mem_index_t	*rootIdxp;
	uint16			searchMask;

	rootIdxp = &(fileLockp->exclLockTree);

	searchMask = (LOCK_SEARCH_SAME_RANGE | LOCK_SEARCH_SAME_OWNER);

	if (exLockp == NULL)
	{
		/*----------------------------------------------
		 * Search for the lock to be deleted in the tree
		 * based on given lockInfo.
		 *----------------------------------------------
		 */
		exLockp = lock_findExclLockByLockInfo(rootIdxp,
											  lockInfop,
											  LOCK_OWNER_CIFS,
											  searchMask);
		if (exLockp == NULL)
		{
			/* There is no match for CIFS lock */
			return (NFX_ERR);
		}
	}

	/*------------------------------
	 * Delete the lock from the tree
	 *------------------------------ 
	 */
	rc = lock_delBtNode(fileLockp,
						(lock_btNode_t *)exLockp,
						LOCK_TYPE_EXCL,
						TRUE,
						TRUE);

	return (rc);
} /* lock_delCifsExclLock() */

/*-----------------------------------------------------------------
 * Name:         lock_delNlmExclLock
 *
 * Description:
 *
 *     Delete an NLM lock from the exclusive lock tree.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     lockInfop - Pointer to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                 data structure, which contains information of lock
 *                 to be deleted.
 *
 *     exLockp - Pointer to the exclusive lock to be deleted. 
 *  
 *         Either lockInfop or exLockp is valid, but not both.
 *
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/18/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_delNlmExclLock(lock_fileLock_t		*fileLockp,
					lock_nlmLockInfo_t	*lockInfop,
					lock_exLock_t		*exLockp)
{
	int32			rc;
	vs_mem_index_t	*rootIdxp;
	uint16			searchMask;
	boolean			firstTime;
	uint8			deleteFlag;

	rootIdxp = &(fileLockp->exclLockTree);

	if (exLockp)
	{
		/*----------------------
		 * Delete the given lock
		 *----------------------
		 */
		rc = lock_delBtNode(fileLockp,
							(lock_btNode_t *)exLockp,
							LOCK_TYPE_EXCL,
							TRUE,
							TRUE);
		return (rc);
	}

	/*----------------------------------------------------------
	 * Delete lock ranges that overlap with the given lock range
	 *----------------------------------------------------------
	 */
	searchMask = LOCK_SEARCH_SAME_OWNER;

	firstTime = TRUE;
	while (1)
	{
		/*----------------------------------------------
		 * Search for the lock to be deleted in the tree
		 * based on given lockInfo.
		 *----------------------------------------------
		 */
		exLockp = lock_findExclLockByLockInfo(rootIdxp,
											  lockInfop,
											  LOCK_OWNER_NLM,
											  searchMask);
		if (exLockp == NULL)
		{
			if (firstTime == TRUE)
				/* There is no match for NLM lock */
				return (NFX_ERR);
			else
				/* There is no more match for NLM lock */
				break;
		}
		firstTime = FALSE;

		/*--------------------------------------------------
		 * Split or modify the lock range of the found lock,
		 * a portion of the range is unlocked.
		 *--------------------------------------------------
		 */
		rc = lock_splitNlmLock(fileLockp,
							   exLockp->info,
							   lockInfop,
							   LOCK_TYPE_EXCL,
							   &deleteFlag);
		if (rc != NFX_OK)
			return (NFX_ERR);

		if (deleteFlag == LOCK_SPLIT_LOCK_FLAG_DELETE)
		{
			/*----------------------
			 * Delete the found lock
			 *----------------------
			 */
			rc = lock_delBtNode(fileLockp,
								(lock_btNode_t *)exLockp,
								LOCK_TYPE_EXCL,
								TRUE,
								TRUE);
			if (rc != NFX_OK)
				return (NFX_ERR);
		}
	} /*while(1)*/

	return (NFX_OK);
} /* lock_delNlmExclLock() */

/*-----------------------------------------------------------------
 * Name:         lock_delExclLock
 *
 * Description:
 *
 *     Delete a lock from the exclusive lock tree.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     lockInfop - Pointer to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                 data structure, which contains information of lock
 *                 to be deleted.
 *
 *     lockOwnerType - The type of owner (LOCK_OWNER_CIFS or LOCK_OWNER_NLM)
 *                     of the lock to be deleted.
 *
 *     exLockp - Pointer to the exclusive lock to be deleted. 
 *  
 *         As to the above three arguments, the caller either,
 *             - provides lockInfop and lockOwnerType and
 *               leaves exLockp invalid, or
 *             - provides exLockp and
 *               leaves lockInfop and lockOwnerType invalid.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/04/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_delExclLock(lock_fileLock_t	*fileLockp,
				 void				*lockInfop,
				 uint8				lockOwnerType,
				 lock_exLock_t		*exLockp)
{
	int32			rc;

	if (exLockp)
	{
		/* Get lock owner type */
		lockOwnerType = lock_getOwnerTypeByLockInfoIdx(exLockp->info);
	}

	switch (lockOwnerType)
	{
		case LOCK_OWNER_CIFS:
			rc = lock_delCifsExclLock(fileLockp,
									  lockInfop,
									  exLockp);
			break;

		case LOCK_OWNER_NLM:
			rc = lock_delNlmExclLock(fileLockp,
									 lockInfop,
									 exLockp);
			break;

		default:
			rc = NFX_ERR;
			break;
	}

	if (rc != NFX_OK)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_delExclLock: "
								"Failed to delete the exclusive lock."));
	}

	return (rc);
} /* lock_delExclLock() */

/*-----------------------------------------------------------------
 * Name:         lock_delCifsSharedLock
 *
 * Description:
 *
 *     Delete a CIFS lock from the shared lock tree.
 *
 * Arguments:
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     lockInfop - Pointer to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                 data structure, which contains information of lock
 *                 to be deleted.
 *
 *     shLockp - Pointer to the shared lock to be deleted. 
 *  
 *         Either lockInfop or exLockp is valid, but not both.
 *
 *     shNodep - Pointer to the node the deleted shared lock belonged to.
 *               Must be provided if shLockp is provided.
 *
 *     prevShLockIdx - Index to the lock before shLockp in the list.
 *                     Must be provided if shLockp is provided.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/19/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_delCifsSharedLock(lock_fileLock_t	*fileLockp,
					   void				*lockInfop,
					   lock_shLock_t	*shLockp,
					   lock_shNode_t	*shNodep,
					   vs_mem_index_t	prevShLockIdx)
{
	int32			rc;
	vs_mem_index_t	*rootIdxp;
	uint16			searchMask;

	rootIdxp = &(fileLockp->sharedLockTree);

	searchMask = (LOCK_SEARCH_SAME_RANGE |
				  LOCK_SEARCH_SAME_OWNER |
				  LOCK_SEARCH_IN_SHARED_LIST);

	if (shLockp == NULL)
	{
		/*----------------------------------------------
		 * Search for the lock to be deleted in the tree
		 * based on given lockInfo.
		 *----------------------------------------------
		 */
		shLockp = lock_findSharedLockByLockInfo(rootIdxp,
												lockInfop,
												LOCK_OWNER_CIFS,
												searchMask,
												&shNodep,
												&prevShLockIdx);
		if (shLockp == NULL)
		{
			/* There is no match for CIFS lock */
			LOCK_PRINTF(LOG_ERR, ("lock_delCifsSharedLock: "
							"There is no match for the CIFS lock."));
			return (NFX_ERR);
		}
	}

	/*------------------------------
	 * Delete the lock from the tree
	 *------------------------------ 
	 */
	rc = lock_delSharedLockFromNode(fileLockp,
									&shNodep,
									prevShLockIdx,
									shLockp,
									LOCK_OWNER_CIFS,
									TRUE);
	if (rc != NFX_OK)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_delCifsSharedLock: "
						"Error in deleting the shLock from shared lock tree."));
		return (NFX_ERR);
	}

	if (shNodep != NULL)
	{
		/*--------------------------------------------
		 * Update "end" and "holey" fields of the node 
		 *--------------------------------------------
		 */
		rc = lock_updateShNodeEndAndHoley(shNodep);
		if (rc != NFX_OK)
			return (NFX_ERR);
	}

	return (NFX_OK);
} /* lock_delCifsSharedLock() */

/*-----------------------------------------------------------------
 * Name:         lock_delNlmSharedLock
 *
 * Description:
 *
 *     Delete an NLM lock from the shared lock tree.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     lockInfop - Pointer to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                 data structure, which contains information of lock
 *                 to be deleted.
 *
 *     shLockp - Pointer to the shared lock to be deleted. 
 *  
 *         Either lockInfop or exLockp is valid, but not both.
 *
 *     shNodep - Pointer to the node the deleted shared lock belonged to.
 *               Must be provided if shLockp is provided.
 *
 *     prevShLockIdx - Index to the lock before shLockp in the list.
 *                     Must be provided if shLockp is provided.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/19/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_delNlmSharedLock(lock_fileLock_t	*fileLockp,
					  void				*lockInfop,	
					  lock_shLock_t		*shLockp,
					  lock_shNode_t		*shNodep,
					  vs_mem_index_t	prevShLockIdx)
{
	int32				rc;
	vs_mem_index_t		*rootIdxp;
	uint16				searchMask;
	boolean				nodeFirstTime;
	vs_mem_index_t		shLockIdx;
	vs_mem_index_t		dummyLockInfoIdx;
	int8				comp;
	uint8				deleteFlag;
	lock_shLock_t		*prevShLockp;

	rootIdxp = &(fileLockp->sharedLockTree);

	if (shLockp)
	{
		/*------------------------------
		 * Delete the lock from the tree
		 *------------------------------ 
		 */
		rc = lock_delSharedLockFromNode(fileLockp,
										&shNodep,
										prevShLockIdx,
										shLockp,
										LOCK_OWNER_NLM,
										TRUE);
		if (rc != NFX_OK)
		{
			LOCK_PRINTF(LOG_ERR, ("lock_delNlmSharedLock: "
						"Error in deleting the shLock "
						"from shNode in shared lock tree."));
			return (NFX_ERR);
		}

		if (shNodep != NULL)
		{
			/*--------------------------------------------
			 * Update "end" and "holey" fields of the node 
			 *--------------------------------------------
			 */
			rc = lock_updateShNodeEndAndHoley(shNodep);
		}
		return (rc);
	}

	/*----------------------------------------------------------
	 * Delete lock ranges that overlap with the given lock range
	 *----------------------------------------------------------
	 */
	searchMask = LOCK_SEARCH_SAME_OWNER | LOCK_SEARCH_IN_SHARED_LIST;

	VS_MEM_IDX_SET_NULL(dummyLockInfoIdx);

	nodeFirstTime = TRUE;
	while (1)
	{
		/*-------------------------------------
		 * Search for the overlapping node 
		 * in the tree based on given lockInfo.
		 *-------------------------------------
		 */
		shNodep = lock_findSharedNodeByLockInfo(rootIdxp,
												lockInfop,
												LOCK_OWNER_NLM,
												searchMask);
		if (shNodep == NULL)
		{
			if (nodeFirstTime == TRUE)
				/* There is no match for NLM lock */
				return (NFX_ERR);
			else
			{
				/* There is no more match for NLM lock */
				break;
			}
		}
		nodeFirstTime = FALSE;

		/*----------------------------------------------------------------
		 * Go through the shared lock list of the found overlapping shNode
		 *----------------------------------------------------------------
		 */
		VS_MEM_IDX_SET_NULL(prevShLockIdx);
		shLockIdx = shNodep->lockList;
		while (!VS_MEM_IDX_IS_NULL(shLockIdx))
		{
			shLockp = (lock_shLock_t *)vs_mem_idxToPtr(shLockIdx);
			if (shLockp == NULL)
				return (NFX_ERR);

			/*---------------------------
			 * Check the overlapping lock
			 *---------------------------
			 */
			comp = lock_twoLockInfoCompFunc(dummyLockInfoIdx,
											lockInfop,
											LOCK_OWNER_NLM,
											shLockp->info,
											searchMask,
											LOCK_SPECIAL_COMP_RESULT);
			if (comp > 0)
			{
				/* Go to next lock to see if there is a match */
				prevShLockIdx = shLockIdx;
				shLockIdx = shLockp->link;
				continue;
			}
			else if (comp < 0)
			{
				/* There is no more match for NLM lock on this node */
				break;
			}
			
			/*--------------------------------------------------
			 * Split or modify the lock range of the found lock,
			 * a portion of the range is unlocked.
			 *--------------------------------------------------
			 */
			rc = lock_splitNlmLock(fileLockp,
								   shLockp->info,
								   lockInfop,
								   LOCK_TYPE_SHAR,
								   &deleteFlag);
			if (rc != NFX_OK)
				return (NFX_ERR);

			if (deleteFlag != LOCK_SPLIT_LOCK_FLAG_KEEP)
			{
				/*-------------------------------------------
				 * Delete/remove the found lock from the list
				 *-------------------------------------------
				 */
				boolean		deleteLock = TRUE;

				if (deleteFlag == LOCK_SPLIT_LOCK_FLAG_DELETE_ADD)
					deleteLock = FALSE;

				rc = lock_delSharedLockFromNode(fileLockp,
												&shNodep,
												prevShLockIdx,
												shLockp,
												LOCK_OWNER_NLM,
												deleteLock);
				if (rc != NFX_OK)
				{
					LOCK_PRINTF(LOG_ERR, ("lock_delNlmSharedLock: "
								"Error in deleting the shLock "
								"from shNode in shared lock tree."));
					return (NFX_ERR);
				}
				if (shNodep == NULL)
				{
					/* The shNode is deleted after all the locks are deleted */
					break;
				}

				if (deleteFlag == LOCK_SPLIT_LOCK_FLAG_DELETE_ADD)
				{
					/* This is because the lock offset has changed */
					rc = lock_addSharedLockToNode(shNodep,
												  shLockIdx,
												  shLockp,
												  shLockp->info,
												  NULL,
												  LOCK_OWNER_NLM);
					if (rc != NFX_OK)
					{
						LOCK_PRINTF(LOG_ERR, ("lock_delNlmSharedLock: "
									"Error in adding the shLock "
									"to shNode in shared lock tree."));
						return (NFX_ERR);
					}
				}

				/*---------------------------------
				 * Move on to next lock in the list
				 *---------------------------------
				 */
				if (VS_MEM_IDX_IS_NULL(prevShLockIdx))
					shLockIdx = shNodep->lockList;
				else
				{
					prevShLockp = (lock_shLock_t *)vs_mem_idxToPtr(prevShLockIdx);
					if (prevShLockp == NULL)
						return (NFX_ERR);

					shLockIdx = prevShLockp->link;
				}
				continue;
			}
	
			prevShLockIdx = shLockIdx;
			shLockIdx = shLockp->link;
		} /* while(shLockIdx) */

		if (shNodep != NULL)
		{
			/*--------------------------------------------
			 * Update "end" and "holey" fields of the node 
			 *--------------------------------------------
			 */
			rc = lock_updateShNodeEndAndHoley(shNodep);
			if (rc != NFX_OK)
				return (NFX_ERR);
		}
	} /* while(1) */

	return (NFX_OK);
} /* lock_delNlmSharedLock() */

/*-----------------------------------------------------------------
 * Name:         lock_delSharedLock
 *
 * Description:
 *
 *     Delete a lock from the shared lock tree.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     lockInfop - Pointer to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                 data structure, which contains information of lock
 *                 to be deleted.
 *
 *     lockOwnerType - The type of owner (LOCK_OWNER_CIFS or LOCK_OWNER_NLM)
 *                     of the lock to be deleted.
 *
 *     shLockp - Pointer to the lock to be deleted. 
 *  
 *         As to the above three arguments, the caller either,
 *             - provides lockInfop and lockOwnerType and
 *               leaves shLockp invalid, or
 *             - provides shLockp and
 *               leaves lockInfop and lockOwnerType invalid.
 *
 *     shNodep - Pointer to the node the deleted shared lock belonged to.
 *               Must be provided if shLockp is provided.
 *
 *     prevShLockIdx - Index to the lock before shLockp in the list.
 *                     Must be provided if shLockp is provided.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/16/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_delSharedLock(lock_fileLock_t	*fileLockp,
				   void				*lockInfop,
				   uint8			lockOwnerType,
				   lock_shLock_t	*shLockp,
				   lock_shNode_t	*shNodep,
				   vs_mem_index_t	prevShLockIdx)
{
	int32			rc;

	if (shLockp)
	{
		/* Get lock owner type */
		lockOwnerType = lock_getOwnerTypeByLockInfoIdx(shLockp->info);
	}

	switch (lockOwnerType)
	{
		case LOCK_OWNER_CIFS:
			rc = lock_delCifsSharedLock(fileLockp,
										lockInfop,
										shLockp,
										shNodep,
										prevShLockIdx);
			break;

		case LOCK_OWNER_NLM:
			rc = lock_delNlmSharedLock(fileLockp,
									   lockInfop,
									   shLockp,
									   shNodep,
									   prevShLockIdx);
			break;

		default:
			rc = NFX_ERR;
			break;
	}

	if (rc != NFX_OK)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_delSharedLock: "
								"Failed to delete the shared lock."));
	}

	return (rc);
} /* lock_delSharedLock() */

/*-----------------------------------------------------------------
 * Name:         lock_procUnlockSingle
 *
 * Description:
 *
 *     Process unlock a single lock request.
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     lockInfop - Pointer to a lock_cifsLockInfo_t or lock_nlmLockInfo_t
 *                 data structure, which contains information of lock to
 *                 be unlocked.
 *
 *     lockOwnerType - The type of owner (LOCK_OWNER_CIFS or LOCK_OWNER_NLM)
 *                     who requests the lock to be unlocked.
 *
 *     lockType - Exclusive or shared lock (LOCK_TYPE_EXCL or LOCK_TYPE_SHAR).
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/04/2002
 *
 *-----------------------------------------------------------------
 */
int32 
lock_procUnlockSingle(lock_fileLock_t *fileLockp,
			  		  void            *lockInfop,  /*Type:lock_cifsLockInfo_t or
												          lock_nlmLockInfo_t*/
			  		  uint8           lockOwnerType,
					  uint8			  lockType)
{
	int32			rc;
	vs_mem_index_t	dummyShLockIdx;

	switch (lockType)
	{
		case LOCK_TYPE_EXCL:
			/*-----------------------------------------
			 * Delete the lock from exclusive lock tree
			 *-----------------------------------------
			 */
			rc = lock_delExclLock(fileLockp,
						   		  lockInfop,
								  lockOwnerType,
								  NULL);
			break;

		case LOCK_TYPE_SHAR:
			/*--------------------------------------
			 * Delete the lock from shared lock tree
			 *--------------------------------------
			 */
			VS_MEM_IDX_SET_NULL(dummyShLockIdx);
			rc = lock_delSharedLock(fileLockp,
									lockInfop,
									lockOwnerType,
									NULL,
									NULL,
									dummyShLockIdx);
			break;

		default:
			break;
	}

	if (rc != NFX_OK)
		return (rc);

	/*--------------------------------------------------------------------
	 * Update the dependency lists of the owners who own the pending locks
	 * that are blocked by the unlocked byte range.
	 *--------------------------------------------------------------------
	 */
	if (lockOwnerType == LOCK_OWNER_NLM)
	{
		/*TODO. Improvement: The current implementation does the updating
		 *                   every time when a lock is deleted no matter 
		 *                   there are pending locks waiting for its release
		 *                   or not.
		 *                   The updating process may only be called when
		 *                   we are deleting a lock which blocks other locks.
		 */
		rc = lock_decDependencyListOfBlockedOwners(fileLockp,
												   lockInfop,
												   lockType);
	}

	/*-------------------------------------------------------
	 * Check if any locks in the pending list can be granted.
	 * If yes, grant the pending locks.
	 *-------------------------------------------------------
	 */
	rc = lock_grantPendingLocks(fileLockp);
	if (rc != NFX_OK)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_procUnlockSingle: "
							"Error in granting pending locks."));
	}

	/*-------------------------------------------------
	 * Reset lowestMandatoryLockOffset field to default
	 * if all locks are removed.
	 *-------------------------------------------------
	 */
	if (VS_MEM_IDX_IS_NULL(fileLockp->exclLockTree) &&
		VS_MEM_IDX_IS_NULL(fileLockp->sharedLockTree))
		fileLockp->lowestMandatoryLockOffset = LOCK_MAX_LOWEST_MAND_OFFSET;

	return (rc);
} /* lock_procUnlockSingle() */

/*-----------------------------------------------------------------
 * Name:         lock_unlockExclAll
 *
 * Description:
 *
 *     Release all the exclusive locks that match the key (CIFS)
 *     or client IP address (NLM).
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     key - Combination of fid and pid.
 *
 *     clientIpAddr - Client IP address.
 *
 *     searchMask - The criteria based on which we search for a lock.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/26/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_unlockExclAll(lock_fileLock_t	*fileLockp,
				   uint32			key,
				   ip_index_t		clientIpAddr,
				   uint16			searchMask)
{
	int32			rc, rcFinal = NFX_OK;
	vs_mem_index_t	exLockIdx;
	lock_exLock_t	*exLockp, *nextExLockp;
	boolean			matchResult;
	vs_mem_index_t	dummyLockOwnerIdx;

	VS_MEM_IDX_SET_NULL(dummyLockOwnerIdx);

	/*---------------------------------------------------------------------
	 * Go through the exclusive lock tree,
	 * and delete all the locks that matche the given key or client IP addr
	 *---------------------------------------------------------------------
	 */
	exLockp = (lock_exLock_t *)
			                lock_btree_getNextNode(fileLockp->exclLockTree,
												   TRUE,
												   &exLockIdx);
	while (exLockp)
	{
		/* Get next exclusive lock in the tree */
		nextExLockp = (lock_exLock_t *)
							lock_btree_getNextNode(exLockIdx,
												   FALSE,
												   &exLockIdx);

		/* Check if current exclusive lock matches the key or clientIpAddr */
		matchResult = lock_match(exLockp->info,
								 dummyLockOwnerIdx,
								 TRUE, /*Use lockInfo*/
								 key,
								 clientIpAddr,
								 searchMask);
		if (matchResult == TRUE)
		{
			/* Found a matching lock */
			rc = lock_delExclLock(fileLockp,
								  NULL,
								  LOCK_OWNER_DUMMY,
								  exLockp);
			if (rc != NFX_OK)
				rcFinal = rc;
		}
		exLockp = nextExLockp;
	}

	return (rcFinal);
} /*lock_unlockExclAll() */

/*-----------------------------------------------------------------
 * Name:         lock_unlockSharedAll
 *
 * Description:
 *
 *     Release all the shared locks that match the key (CIFS)
 *     or client IP address (NLM).
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     key - Combination of fid and pid.
 *
 *     clientIpAddr - Client IP address.
 *
 *     searchMask - The criteria based on which we search for a lock.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/26/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_unlockSharedAll(lock_fileLock_t	*fileLockp,
					 uint32				key,
					 ip_index_t			clientIpAddr,
					 uint16				searchMask)
{
	int32			rc, rcFinal = NFX_OK;
	vs_mem_index_t	shNodeIdx, shLockIdx, prevShLockIdx;
	lock_shNode_t	*shNodep, *nextShNodep;
	lock_shLock_t	*shLockp;
	boolean			matchResult;
	vs_mem_index_t	dummyLockOwnerIdx;

	VS_MEM_IDX_SET_NULL(dummyLockOwnerIdx);

	/*---------------------------------------------------------------------
	 * Go through the shared lock tree
	 * and delete all the locks that matche the given key or client IP addr
	 *---------------------------------------------------------------------
	 */
	shNodep = (lock_shNode_t *)
			                lock_btree_getNextNode(fileLockp->sharedLockTree,
												   TRUE,
												   &shNodeIdx);
	while (shNodep)
	{
		/* Get next shared node in the tree */
		nextShNodep = (lock_shNode_t *)
							lock_btree_getNextNode(shNodeIdx,
												   FALSE,
												   &shNodeIdx);

		/*---------------------------------------------
		 * Go through each shared lock in the lock list
		 *---------------------------------------------
		 */
		VS_MEM_IDX_SET_NULL(prevShLockIdx);
		shLockIdx = shNodep->lockList;
		while (!VS_MEM_IDX_IS_NULL(shLockIdx))
		{
			shLockp = (lock_shLock_t *)vs_mem_idxToPtr(shLockIdx);
			if (shLockp == NULL)
			{
				LOCK_PRINTF(LOG_ERR, ("lock_unlockSharedAll: "
									"Couldn't find shLock data structure."));
				return (NFX_ERR);
			}

			/* Check if current shared lock matches the key or clientIpAddr */
			matchResult = lock_match(shLockp->info,
									 dummyLockOwnerIdx,
									 TRUE, /*Use lockInfo*/
									 key,
									 clientIpAddr,
									 searchMask);
			if (matchResult == TRUE)
			{
				/* Delete the lock */

				shLockIdx = shLockp->link;	/*Prepare for going to next lock*/

				rc = lock_delSharedLock(fileLockp,
										NULL,
										LOCK_OWNER_DUMMY,
										shLockp,
										shNodep,
										prevShLockIdx);
				if (rc != NFX_OK)
					rcFinal = rc;
				else
					continue;
			}

			prevShLockIdx = shLockIdx;
			shLockIdx = shLockp->link;
		}
		shNodep = nextShNodep;
	}

	return (rcFinal);
} /*lock_unlockSharedAll() */

/*-----------------------------------------------------------------
 * Name:         lock_lock_deletePendingAll
 *
 * Description:
 *
 *     Delete all the pending locks that match the key (CIFS)
 *     or client IP address (NLM).
 *
 * Arguments:    
 * 
 *     fileLockp - Pointer to a lock_fileLock_t structure.
 *
 *     key - Combination of fid and pid.
 *
 *     clientIpAddr - Client IP address.
 *
 *     searchMask - The criteria based on which we search for a lock.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/26/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_deletePendingAll(lock_fileLock_t	*fileLockp,
					  uint32			key,
					  ip_index_t		clientIpAddr,
					  uint16			searchMask)
{
	int32						rc;
	lock_delPendLocksParam_t	param;

	/*---------------------------------------------------------------------
	 * Go through the pending list
	 * and delete all the locks that matche the given key or client IP addr
	 *---------------------------------------------------------------------
	 */
	param.type = LOCK_DELPENDLOCKS_TYPE_UNLOCKALL;
	param.u.delAll.key = key;
	param.u.delAll.clientIpAddr = clientIpAddr;
	param.u.delAll.searchMask = searchMask;

	rc = lock_deletePendingLocksFromLocalList(fileLockp, param);

	return (rc);
} /* lock_deletePendingAll() */

/*-----------------------------------------------------------------
 * Name:         lock_procUnlockAll
 *
 * Description:
 *
 *     Release all the locks that match the key (CIFS)
 *     or client IP address (NLM).
 *
 * Arguments:    
 * 
 *     fileLockIdx - Index to a lock_fileLock_t structure.
 *
 *     key - Combination of fid and pid.
 *
 *     clientIpAddr - Client IP address.
 *
 *     searchMask - The criteria based on which we search for a lock.
 *
 * Return value:
 *
 *     NFX_OK - If succeed.
 *
 *     NFX_ERR - If fail.
 *
 * Created by:   Xiaobing Zhang
 *
 * Date Created: 12/06/2002
 *
 *-----------------------------------------------------------------
 */
int32
lock_procUnlockAll(vs_mem_index_t	fileLockIdx,
				   uint32			key,
				   ip_index_t		clientIpAddr,
				   uint16			searchMask)
{
	int32			rc, rcFinal;
	lock_fileLock_t	*fileLockp;

	/*---------------------------
	 * Get the fileLock structure
	 *---------------------------
	 */
	fileLockp = (lock_fileLock_t *)vs_mem_idxToPtr(fileLockIdx);
	if (fileLockp == NULL)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_procUnlockAll: "
								"Couldn't find fileLock structrue."));
		return (NFX_ERR);
	}

	/*---------------------------------------------------------------------
	 * Go through the exclusive lock tree,
	 * and delete all the locks that matche the given key or client IP addr
	 *---------------------------------------------------------------------
	 */
	rcFinal = lock_unlockExclAll(fileLockp, key, clientIpAddr, searchMask);
	if (rcFinal != NFX_OK)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_procUnlockAll: "
								"Error in deleting exclusive locks."));
	}

	/*---------------------------------------------------------------------
	 * Go through the shared lock tree
	 * and delete all the locks that matche the given key or client IP addr
	 *---------------------------------------------------------------------
	 */
	rc = lock_unlockSharedAll(fileLockp, key, clientIpAddr, searchMask);
	if (rc != NFX_OK)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_procUnlockAll: "
								"Error in deleting shared locks."));
		rcFinal = rc;
	}

	/*---------------------------------------------------------------------
	 * Go through the pending list
	 * and delete all the locks that matche the given key or client IP addr
	 *---------------------------------------------------------------------
	 */
	rc = lock_deletePendingAll(fileLockp, key, clientIpAddr, searchMask);
	if (rc != NFX_OK)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_procUnlockAll: "
								"Error in deleting pending locks."));
		rcFinal = rc;
	}

	/*-----------------------------------------------------------------
	 * Update the dependency lists, which contain nodes for owners with
	 * the "clientIpAddr". Remove these nodes from the list.
	 *-----------------------------------------------------------------
	 */
	if (searchMask & LOCK_SEARCH_SAME_IPADDR)
		rc = lock_delOwnersFmDependencyListOfBlockedOwners(fileLockp,
														   clientIpAddr);

	/*------------------------------------------------------------
	 * Go through the pending list to see any locks can be granted
	 *------------------------------------------------------------
	 */
	rc = lock_grantPendingLocks(fileLockp);
	if (rc != NFX_OK)
	{
		LOCK_PRINTF(LOG_ERR, ("lock_procUnlockAll: "
							"Error in granting pending locks."));
		rcFinal = rc;
	}

	/*-------------------------------------------------
	 * Reset lowestMandatoryLockOffset field to default
	 * if all locks are removed.
	 *-------------------------------------------------
	 */
	if (VS_MEM_IDX_IS_NULL(fileLockp->exclLockTree) &&
		VS_MEM_IDX_IS_NULL(fileLockp->sharedLockTree))
		fileLockp->lowestMandatoryLockOffset = LOCK_MAX_LOWEST_MAND_OFFSET;

	return (rcFinal);
} /* lock_procUnlockAll() */


