/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: luc.c
 *
 * RCS:  $Id: luc-cam.c,v 1.9 2002/10/22 01:30:38 jong Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Contains Look Up Classifier CAM (Lara LNI7020) related 
 *				routines. LUC-CAM Interface registers are used to 
 *				program LNI7020 (read/write LNI7020 registers & 
 *				data/mask arrays). 
 *				TCP/UDP flow entries and IP IF with or w/o VLAN are 
 *				stored in the data & mask arrays in the CAM for lookup. 
 *				There is NO SRAM attached to the CAM. The usrInfo
 *				associated with each flow (flow ptr) is stored in
 *				lookup table in local SDRAM at same index.
 * 				CAM returns index of matching entry. LUC puts index
 *				in usrInfo field of Receive Complete Descriptor.
 *				LUC driver uses index to retrieve assoc. usrInfo.
 *
 * Created by:	Shaun Hui
 *
 * Date Created: 7/01/2001
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include "../sm-bsd-sys/flags.h"
#include "../sm-bsd-sys/macros.h"
#include "luc-cam.h"

#define   LUC_VALIDATE	0
#define	  LUC_CAM_PRINTF(a)

luc_camCfg_t	luc_camCfg;

/*-----------------------------------------------------------------
 * Name: luc_camInit
 *
 * Description: Initialize LUC-CAM interface, LNI7020 CAM & related
 *				lookup table, bit maps etc.
 *
 * Created by : Shaun Hui
 * 
 * Date Created: 7/01/01
 *-----------------------------------------------------------------
 */
void 
luc_camInit( void )
{
	int32			idx;
	uint32			camCmd;
	uint32			regIdx;

/*
	luc_camCfg.camBase = (luc_camIfReg_t *)CAM_IF_REG_BASE;
*/
	luc_camCfg.camBase = (luc_camIfReg_t *)eee_ramAlloc(sizeof(luc_camIfReg_t));

//printf("luc_camInit 1\n");
	/*-------------------------------
	 * Issue software reset on 
	 * CAM command register. 
	 * Bits [67:32] are reserved (0).
	 *-------------------------------
	 */
	luc_camWrite( CAM_IFCMD_BLK_REG, CAM_CMD_IDX, 1, 
				  CAM_CMD_SRST, 0, 0, 0, 0);

//printf("luc_camInit 2\n");
	luc_camDelay( 100 );

	/*-------------------------------
	 * Configure CAM:
	 * Device enable.
	 * TLSZ = 00 ( 1 CAM) 
	 * Last Device (only 1 CAM)
	 * Last Device on SRAM (only 1 CAM)
	 * 4 x 4K x 136.
	 * Bits [67:32] are reserved (0).
	 *-------------------------------
	 */
	camCmd = CAM_CMD_DEVE | CAM_CMD_TLSZ_1 | CAM_CMD_HLAT |
			 CAM_CMD_LDEV | CAM_CMD_LRAM | CAM_CMD_CFG_136;
	luc_camWrite( CAM_IFCMD_BLK_REG, CAM_CMD_IDX, 1,
				  camCmd, 0, 0, 0,0);

//printf("luc_camInit 3\n");
	/*-------------------------------
	 * Clear Data & Mask arrays
	 *-------------------------------
	 */

	/* clear flow table in cam */ 
	for (idx = 0; idx < CAM_MAX_FLOW_ENTRY; idx++)
	{
		/*-------------------------------
		 * Clear 16K x 136bits data array.
		 * Lower 68 bits in odd index.
		 * Upper 68 bits in even index.
		 *-------------------------------
		 */
		luc_camWrite( CAM_IFCMD_BLK_DATA, idx, 0, 
					  0, 0, 0, 0, 0);

		/*-------------------------------
		 * Clear 16K x 136bits mask array.
		 * Lower 68 bits in odd index.
		 * Upper 68 bits in even index.
		 *-------------------------------
		 */
		luc_camWrite( CAM_IFCMD_BLK_MASK, idx, 0, 
					  0, 0, 0, 0, 0);
	}

	/* clear interface table in cam */
	for (idx = CAM_IPIF_BASE_IDX; idx <
			 CAM_IPIF_BASE_IDX +CAM_MAX_IPIF_ENTRY; idx++)
	{
		/*-------------------------------
		 * Clear 16K x 136bits data array.
		 * Lower 68 bits in odd index.
		 * Upper 68 bits in even index.
		 *-------------------------------
		 */
		luc_camWrite( CAM_IFCMD_BLK_DATA, idx, 0,
					  0, 0, 0, 0, 0);

		/*-------------------------------
		 * Clear 16K x 136bits mask array.
		 * Lower 68 bits in odd index.
		 * Upper 68 bits in even index.
		 *-------------------------------
		 */
		luc_camWrite( CAM_IFCMD_BLK_MASK, idx, 0,
					  0, 0, 0, 0, 0);
	}

//printf("luc_camInit 4\n");
	/*-------------------------------
	 * Clear Lookup Table contents. 
	 *-------------------------------
	 */
	for (idx = 0; idx < CAM_MAX_FLOW_ENTRY; idx++)
	{
		luc_camCfg.lkupTbl[idx] = NULL;
	}

	for (idx = 0; idx < CAM_MAX_IPIF_ENTRY; idx++)
	{
		luc_camCfg.lkupIpTbl[idx] = NULL;
	}	

	/*-------------------------------
	 * Mark all flow entries as 'free'.
	 *-------------------------------
	 */
	for (idx = 0; idx < CAM_MAX_FLOW_BITMAP; idx++)
	{
		luc_camCfg.flowBitMap[idx] = 0xFFFFFFFF;
	}

	/*-------------------------------
	 * Mark all IP IF entries as 'free'.
	 *-------------------------------
	 */
	for (idx = 0; idx < CAM_MAX_IPIF_BITMAP; idx++)
	{
		luc_camCfg.ipIfBitMap[idx] = 0xFFFFFFFF;
	}
//printf("luc_camInit 5\n");

}

/*-----------------------------------------------------------------
 * Name: luc_camWrite
 *
 * Description: Write CAM register or Data/Mask array entry on given 
 *				index & values.
 *
 * Created by : Shaun Hui
 * 
 * Date Created: 7/01/01
 *-----------------------------------------------------------------
 */
int32
luc_camWrite(int32	blk,		/* regster, data or mask array */
			 int32	idx,		/* register, data or mask array index */
			 int	host,		/* is data in host order? */
			 uint32 data_w0,
			 uint32 data_w1,
			 uint32 data_w2,
			 uint32 data_w3,
			 uint32 data_w4)
{
	volatile luc_camIfReg_t	*pCamIf;
	int32	 rc;

	/*-------------------------------
	 * Make sure blk & idx are valid.
	 *-------------------------------
	 */
	rc = luc_camCheck(blk, idx);	
	if (rc != NFX_OK)
		return(rc);

	pCamIf = luc_camCfg.camBase;

	/*-------------------------------
	 * Write 168 bits of array entry
	 *-------------------------------
	 */
	if (host) {
		pCamIf->camDataW[0] = htonl(data_w0);
		pCamIf->camDataW[1] = htonl(data_w1);
		pCamIf->camDataW[2] = htonl(data_w2);
		pCamIf->camDataW[3] = htonl(data_w3);
		pCamIf->camDataW[4] = htonl(data_w4);
	}
	else {
		pCamIf->camDataW[0] = data_w0;
		pCamIf->camDataW[1] = data_w1;
		pCamIf->camDataW[2] = data_w2;
		pCamIf->camDataW[3] = data_w3;
		pCamIf->camDataW[4] = data_w4;	
	}

	/*-------------------------------
	 * Set CAM reg, Data/Mask array index.
	 *-------------------------------
	 */
	pCamIf->camIdx = htonl(idx);

	/*-------------------------------
	 * Issue CAM IF command for:
	 * Write to Reg/Data/Mask Block.
	 *-------------------------------
	 */
again:
	pCamIf->camCmd = htonl(CAM_IFCMD_CMD_WR | blk);

	/* read status */
	if (!(pCamIf->lucStatus & htonl(CAM_STATUS_OK))) {
//		LUC_CAM_PRINTF(("write failed try again\n"));
		
		/* clear luc_status register */
		pCamIf->lucStatusClr = htonl((CAM_STATUS_CYC|CAM_STATUS_NCLR));
//		goto again;
	}

	/* write successful, clr status register */
	pCamIf->lucStatusClr = htonl(CAM_STATUS_OK);
	

	return(NFX_OK);
}

/*-----------------------------------------------------------------
 * Name: luc_camRead
 *
 * Description: Read CAM register or Data/Mask array entry on given 
 *				index. 
 *
 * Created by : Shaun Hui
 * 
 * Date Created: 7/01/01
 *-----------------------------------------------------------------
 */	
int32
luc_camRead(int32	blk, 		/* register, data or mask array */
			int32	idx, 		/* register, data or mask array index */
			int		host,		/* data in host order or network order */
			uint32	*data_r0,	
			uint32	*data_r1,
			uint32	*data_r2,
			uint32	*data_r3,
			uint32	*data_r4)
{
	volatile luc_camIfReg_t	*pCamIf;
	int32	 rc;
	
	/*-------------------------------
	 * Make sure blk & idx are valid.
	 *-------------------------------
	 */
	rc = luc_camCheck(blk, idx);	
	if (rc != NFX_OK)
		return(rc);

	pCamIf = luc_camCfg.camBase;

	/*-------------------------------
	 * Set CAM reg. index.
	 *-------------------------------
	 */
	pCamIf->camIdx = htonl(idx);

	/*-------------------------------
	 * Issue CAM IF command for:
	 * Read from Reg/Data/Mask Block.
	 *-------------------------------
	 */
again:
	pCamIf->camCmd = htonl(CAM_IFCMD_CMD_RD | blk);

	/* read status */
	if (!(pCamIf->lucStatus & htonl(CAM_STATUS_OK))) {
		LUC_CAM_PRINTF(("read failed try again\n"));
		
		/* clear luc_status register */
		pCamIf->lucStatusClr = htonl(CAM_STATUS_CYC|CAM_STATUS_NCLR);
//		goto again;
	}

	/* read result */
	if (host) {
		*data_r0 = ntohl(pCamIf->camDataR[0]);
		*data_r1 = ntohl(pCamIf->camDataR[1]);
		*data_r2 = ntohl(pCamIf->camDataR[2]);
		*data_r3 = ntohl(pCamIf->camDataR[3]);
		*data_r4 = ntohl(pCamIf->camDataR[4]);	
	}
	else {
		*data_r0 = pCamIf->camDataR[0];
		*data_r1 = pCamIf->camDataR[1];
		*data_r2 = pCamIf->camDataR[2];
		*data_r3 = pCamIf->camDataR[3];
		*data_r4 = pCamIf->camDataR[4];
	}

	/* clear status register */
	pCamIf->lucStatusClr = htonl(CAM_STATUS_OK);

	return(NFX_OK);
}

/*-----------------------------------------------------------------
 * Name: luc_camFind
 *
 * Description: search CAM Data/Mask array entry on given 
 *				data.. 
 *
 * Created by :  Su Wang
 * 
 * Date Created: 03/14/02
 *-----------------------------------------------------------------
 */	
int32							/* cam ifx, or -1 err */
luc_camFind(int32	blk, 		/* register, data or mask array */
			int		host,		/* data in host order or network order */
			uint32	data_w0,	
			uint32	data_w1,
			uint32	data_w2,
			uint32	data_w3,
			uint32	data_w4)
{
	volatile luc_camIfReg_t	*pCamIf;
	int32	 rc = -1;
	
	/*-------------------------------
	 * Make sure blk & idx are valid.
	 *-------------------------------
	 */
	if (blk != CAM_IFCMD_BLK_DATA)
		return rc;
	
	pCamIf = luc_camCfg.camBase;

	/*-------------------------------
	 * Write 168 bits of array entry
	 *-------------------------------
	 */
	if (host) {
		pCamIf->camDataW[0] = htonl(data_w0);
		pCamIf->camDataW[1] = htonl(data_w1);
		pCamIf->camDataW[2] = htonl(data_w2);
		pCamIf->camDataW[3] = htonl(data_w3);
		pCamIf->camDataW[4] = htonl(data_w4);
	}
	else {
		pCamIf->camDataW[0] = data_w0;
		pCamIf->camDataW[1] = data_w1;
		pCamIf->camDataW[2] = data_w2;
		pCamIf->camDataW[3] = data_w3;
		pCamIf->camDataW[4] = data_w4;	
	}

	
again:
	pCamIf->camCmd = htonl(CAM_IFCMD_CMD_SR | blk);

	/* read status */
	if (!(pCamIf->lucStatus & htonl(CAM_STATUS_OK))) {
		LUC_CAM_PRINTF(("write failed try again\n"));
		
		/* clear luc_status register */
		pCamIf->lucStatusClr = htonl((CAM_STATUS_CYC|CAM_STATUS_NCLR));
//		goto again;
	}

	/* search result returned */
	rc = ntohl(pCamIf->camIdxS) & 0xffff;
	
	if (rc & CAM_IDXS_HIT)
		rc &= ~CAM_IDXS_HIT;
	else
		rc = -1;
	
	/* clr status register */
	pCamIf->lucStatusClr = htonl(CAM_STATUS_OK);
	

	return(rc);
}

/*-----------------------------------------------------------------
 * Name: luc_camAlloc
 *
 * Description: Allocate an CAM data array entry. Seperate bit map
 *				tables are used for Flows & IP IF. Each bit represents
 *				an entry.
 *
 * Created by : Shaun Hui
 * 
 * Date Created: 7/01/01
 *
 * Returns: Index to data array or no resource error. (see luc-cam.h)
 *-----------------------------------------------------------------
 */
int32 
luc_camAlloc(int32 type)
{
	uint32	*pBitMap;
	uint32	bitMap;
	uint32	bitIdx;
	uint32	byteIdx;
	int32	baseIdx;
	int32	byteCnt;

	/*-------------------------------
	 * Get start of bit map of table.
	 *-------------------------------
	 */
	switch (type)
	{
		case	CAM_FLOW_TBL_ID:
				pBitMap = &luc_camCfg.flowBitMap[0];
				baseIdx = CAM_FLOW_BASE_IDX;
				byteCnt = CAM_MAX_FLOW_BITMAP;
				break;

		case	CAM_IPIF_TBL_ID:
				pBitMap = &luc_camCfg.ipIfBitMap[0];
				baseIdx = CAM_IPIF_BASE_IDX;
				byteCnt = CAM_MAX_IPIF_BITMAP;
				break;

		default:
				return( LUC_CAM_ERR_BAD_TBL);
	}

	/*-------------------------------
	 * Each set bit represents a free entry.
	 *-------------------------------
	 */
    for (byteIdx=0; byteIdx < byteCnt; byteIdx++)
	{
		/*-------------------------------
		 * Skip this byte if no free entry.
		 *-------------------------------
		 */
		bitMap = pBitMap[byteIdx];
		if (!bitMap)
			continue;

		/*-------------------------------
		 * Find a free entry.
		 *-------------------------------
		 */
		for (bitIdx=0; bitIdx < BITS_PER_WORD; bitIdx++)
		{
			if (bitMap & (1 << bitIdx))
			{
				/*-------------------------------
				 * Clear the bit to mark entry used.
				 *-------------------------------
				 */
				bitMap &= ~(1 << bitIdx);
				pBitMap[byteIdx] = bitMap;

				/*-------------------------------
				 * Return index based on byte & bit pos.
				 *-------------------------------
				 */
				return( baseIdx + (byteIdx*BITS_PER_WORD) + bitIdx );
			}
		}
	}

	return (LUC_CAM_ERR_NO_RSRC);
}

/*-----------------------------------------------------------------
 * Name: luc_camFree
 *
 * Description: Free a CAM data array entry. 
 *
 * Created by : Shaun Hui
 * 
 * Date Created: 7/01/01
 *
 * Returns: NFX_OK or error. (see luc-cam.h)
 *-----------------------------------------------------------------
 */
static int32 
luc_camFree( int32 type, int32 idx )
{
	uint32	*pBitMap;
	uint32	bitMap;
	uint32	bitIdx;
	uint32	byteIdx;
	int32	baseIdx;

	/*-------------------------------
	 * Get start of bit map of table.
	 *-------------------------------
	 */
	switch (type)
	{
		case	CAM_FLOW_TBL_ID:
				pBitMap = &luc_camCfg.flowBitMap[0];
				baseIdx = CAM_FLOW_BASE_IDX;
				if ((idx < baseIdx) || 
					(idx > (baseIdx + CAM_MAX_FLOW_ENTRY)))
					return (LUC_CAM_ERR_BAD_IDX);
				break;

		case	CAM_IPIF_TBL_ID:
				pBitMap = &luc_camCfg.ipIfBitMap[0];
				baseIdx = CAM_IPIF_BASE_IDX;
				if ((idx < baseIdx) || 
					(idx > (baseIdx + CAM_MAX_IPIF_ENTRY)))
					return (LUC_CAM_ERR_BAD_IDX);
				break;

		default:
				return( LUC_CAM_ERR_BAD_TBL);
	}

	/*-------------------------------
	 * make idx relative to base.
	 *-------------------------------
	 */
	idx    -= baseIdx;
	bitIdx  = idx & WORD_MASK;
	byteIdx = idx / BITS_PER_WORD;
	bitMap  = pBitMap[byteIdx];

	/*-------------------------------
	 * Check if bit is set (free).
	 *-------------------------------
	 */
	if (bitMap & (1 << bitIdx))
	{
		printf("Free twice!\n");
		return( LUC_CAM_ERR_DUP_FREE);
	}
	else
	{
		/*-------------------------------
		 * Set bit to mark entry as free.
		 *-------------------------------
		 */
		bitMap |= (1 << bitIdx);
		pBitMap[byteIdx] = bitMap;
	}
	return (NFX_OK);
}



/*-----------------------------------------------------------------
 * Name: luc_camAddFlow
 *
 * Description: Add a TCP/UDP Flow entry into CAM lookup.
 *
 * Created by : Shaun Hui
 * 
 * Date Created: 7/01/01
 * 
 * Returns: Index to CAM data array or error. (see luc-cam.h)
 *-----------------------------------------------------------------
 */
int32 
luc_camAddFlow( int32  prot,       /* protocol: UDP/TCP */
				uint32 sip,        /* source IP */
				uint32 dip,        /* destination IP */
				uint32 sport,      /* source port */
				uint32 dport)      /* destination port */
                     
{
	int32	idx;
	uint32	data[5];
	uint32	mask[5];
	int		rc;

	/*-------------------------------
	 * Allocate a new entry
	 *-------------------------------
	 */
	idx = luc_camAlloc( CAM_FLOW_TBL_ID );

#if LUC_VALIDATE
	/* make sure content is empty */
	luc_camRead(CAM_IFCMD_BLK_DATA, idx, 0, 
				&data[0], &data[1], &data[2], &data[3], &data[4]);
	
    if (data[0] || data[1] || data[2] || data[3] || data[4])
		LUC_CAM_PRINTF(("%s: content is not empty\n", __FUNCTION__));

#endif /* LUC_VALIDATE */
	
	if (idx < 0)
		return( LUC_CAM_ERR_NO_RSRC );

	/*-------------------------------
	 * 
	 * BIG Endian mode
	 *
	 * 31                                0 
	 * +---------------------------------+ 
	 * |     src     IP                  |  data[0]
	 * +---------------------------------+ 
	 *
	 * 63                               32 
	 * +---------------------------------+ 
	 * |     dest    IP                  | 
	 * +---------------------------------+  data[1]
	
	 * 95                                64  
	 * +----------------+----------------+ 
	 * |     dport      |     sport      |  data[2]
	 * +----------------+----------------+ 
	 *
	 * 127                  103          96
	 * +----------------------+-+--------+ 
	 * |            0s        |   prot   | 
	 * +----------------------+-+--------+  data[3]
	 *
	 *                       136      128  
	 *                         +---------+ 
	 *                         | 1|      | 
	 *                         +---------+  data[4]
	 *
	 *-------------------------------
	 */

	/* src and dest ip, they are network order already */
	data[0] = sip;
	data[1] = dip;
	mask[0] = 0xFFFFFFFF;
	mask[1] = 0xFFFFFFFF;

	/* local and remote port */
	data[2] = dport | (sport <<16);/* move sport to high address space */
	mask[2] = 0xFFFFFFFF;

	/* protocol */
	data[3] = htonl((prot & 0xff));
	mask[3] = htonl(0x000000FF);	


	/* indicate ft table entry */
	data[4] = htonl(0x80);

	/* write to resigters */
	luc_camWrite( CAM_IFCMD_BLK_DATA, idx, 0,
				  data[0], data[1], data[2], data[3], data[4]);

	luc_camWrite( CAM_IFCMD_BLK_MASK, idx, 0,
				  mask[0], mask[1], mask[2], mask[3], mask[4]);

#if LUC_VALIDATE
	/* make sure whatever written can be retrived */
	
	rc = luc_camFind(CAM_IFCMD_BLK_DATA, 0, 
					 data[0], data[1], data[2], data[3], data[4]);
	
	if (rc != idx)
		LUC_CAM_PRINTF(("%s: index no match, rc=%d, idx=%d\n",
						__FUNCTION__, rc, idx));

	luc_camRead(CAM_IFCMD_BLK_DATA, idx, 0,
				&data[0], &data[1], &data[2], &data[3], &data[4]);

	LUC_CAM_PRINTF(("%s: read data %x %x %x %x %x\n",
					__FUNCTION__, data[0], data[1], data[2], data[3]));
	
#endif /* VALIDATE */

	return (idx - CAM_FLOW_BASE_IDX);
}

/*-----------------------------------------------------------------
 * Name: luc_camDelFlow
 *
 * Description: Delete a TCP/UDP Flow entry from CAM lookup.
 *
 * Created by : Shaun Hui
 * 
 * Date Created: 7/01/01
 * 
 * Returns: NFX_OK or error. (see luc-cam.h)
 *-----------------------------------------------------------------
 */
int32 
luc_camDelFlow(int32 idx ) /* data array index */
{
	int32	rc;
	uint32	data[5];
	
	/* get absolute index value */
	idx += CAM_FLOW_BASE_IDX;


	/*-------------------------------
	 * Free CAM bit map.
	 *-------------------------------
	 */
	rc = luc_camFree( CAM_FLOW_TBL_ID, idx );

	if (rc != NFX_OK)
		return rc;

#if LUC_VALIDATE
	/* make sure content not empty */
	luc_camRead(CAM_IFCMD_BLK_DATA, idx, 0, 
				&data[0], &data[1], &data[2], &data[3], &data[4]);
	
    if (data[0] && data[1] && data[2] && data[3])
		LUC_CAM_PRINTF(("%s: content is empty\n", __FUNCTION__));

#endif /* LUC_VALIDATE */

	/*-------------------------------
	 * Clear CAM data array content.
	 *-------------------------------
	 */
	luc_camWrite(CAM_IFCMD_BLK_DATA, idx, 0,
				 0, 0, 0, 0, 0);


	luc_camWrite(CAM_IFCMD_BLK_MASK, idx, 0,
				 0, 0, 0, 0, 0);

	/*-------------------------------
	 * Clear lookup table entry.
	 *-------------------------------
	 */
	idx -= CAM_FLOW_BASE_IDX;
	luc_camCfg.lkupTbl[idx] = NULL;

	/*-------------------------------
	 * Delete simulated entry also.
	 *-------------------------------
	 */
	luc_simCamDel( idx );

	return (NFX_OK);
}



/*-----------------------------------------------------------------
 * Name: luc_camCheck
 *
 * Description: Check CAM block id belongs to register or Data/Mask 
 *				array entry and respective index is in range.
 *
 * Created by : Shaun Hui
 * 
 * Date Created: 7/01/01
 * 
 * Returns: NFX_OK or error. (see luc-cam.h)
 *-----------------------------------------------------------------
 */
int32
luc_camCheck(uint32 blk, uint32 idx)
{
	switch(blk)
	{
		case	CAM_IFCMD_BLK_DATA:
		case	CAM_IFCMD_BLK_MASK:
				if (!((idx < CAM_MAX_FLOW_ENTRY) || /* flow table */
					((idx >= CAM_IPIF_BASE_IDX) && /* ipif table */
					(idx < CAM_IPIF_BASE_IDX + CAM_MAX_IPIF_ENTRY))))
				{
					printf("CAM Error: Array index %d too big.\n", idx);
					return (LUC_CAM_ERR_BAD_IDX);
				}
				break;

		case	CAM_IFCMD_BLK_REG:
				if (idx > CAM_MAX_REG_IDX)
				{
					printf("CAM Error: Register index too big.\n");
					return (LUC_CAM_ERR_BAD_IDX);
				}
				break;
		default:
				printf("CAM Error: unknown block type.\n");
				return (LUC_CAM_ERR_BAD_BLK);
	}
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name: luc_camDelay
 *
 * Description: Perform delay loop by reading LUC-CAM IF status
 *				register.
 *
 *-----------------------------------------------------------------
 */
void 
luc_camDelay( uint32 delay )
{
	uint32	i;
	uint32	value;
	volatile uint32	*pReg;

// #if defined(LUC_HARDWARE)
//	while (!(luc_camCfg.camBase->lucStatus & htonl(CAM_STATUS_INIT)))
//		i = 1;
// #else
	for (i=0; i < delay; i++)
		value = luc_camCfg.camBase->lucStatus; 
// #endif /* LUC_HARDWARE */
	return;
}

/*-----------------------------------------------------------------
 * Name: luc_camGetUsrInfo
 *
 * Description: Return content of lookup table of a given index.
 *
 *-----------------------------------------------------------------
 */
void *
luc_camGetUsrInfo( int32 idx /* data array index */)
{
	return (luc_camCfg.lkupTbl[idx]);
}

/*-----------------------------------------------------------------
 * Name: luc_camSetUsrInfo
 *
 * Description: Set content of lookup table of a given index.
 *
 *-----------------------------------------------------------------
 */
void 
luc_camSetUsrInfo(int32 idx, void	*usrInfo)
{
	/*-------------------------------
	 * Write user info in lookup table.
	 * LUC will give us index in RCR
	 * for a rx pkt with matching flow.
	 * Driver looks up usrInfo using idx.
	 *-------------------------------
	 */
	luc_camCfg.lkupTbl[idx] = usrInfo;
}

void
luc_camTest(int argc, char	*argv[])
{
	uint32		root;
	int			reps;
	int			i;
	uint32		srcAddr, dstAddr;
	ushort16	srcPort, dstPort;
	uchar8		proto;
	int			index, rc;
	uint32		data[5];
	eeeRTC_t	tm;
	
	eee_rtc(&tm);
	root = (uint32)(eee_rtcToUsec(&tm) & 0xFFFFFFFFLL);

	reps = atoi(argv[1]);
	
	for (i=0; i< reps; i++) {
		
		/* assgin values */
		srcAddr = root;
		root += 128;
		dstAddr =  root;
		root += 128;
		srcPort = root & 0xffff;
		root += 128;
		dstPort = root & 0xffff;
		root += 128;
		proto = root * 0xff;
		
		/* perform operations */
		printf("flow entry added: src=%x dst=%x sport=%x dport=%x prot=%x\n",
			   ntohl(srcAddr), ntohl(dstAddr), 
			   ntohs(srcPort), ntohs(dstPort), proto);
		
		index = luc_camAddFlow(proto, srcAddr, dstAddr, srcPort, dstPort);
		
		if (index < 0) {
			printf("flow add falled \n");
			continue;
		}
		else 
			printf("flow add worked, index = %d\n", index);
		
		/* get absolute index */
		index += CAM_FLOW_BASE_IDX;
		
		printf("\n read index %d\n", index);
		if (luc_camRead(CAM_IFCMD_BLK_DATA, index, 0, &data[0], 
						&data[1], &data[2], &data[3], &data[4]) != NFX_OK) {
			printf("read flow index %x failed\n");
			continue;
		}
		else 
			printf("read flow worked: data4=%x data3=%x data2=%x data1=%x data0=%x\n",
				   ntohl(data[4]), ntohl(data[3]), ntohl(data[2]),
				   ntohl(data[1]), ntohl(data[0]));
		

		printf("\n search for index \n");
		rc = luc_camFind(CAM_IFCMD_BLK_DATA, 0, data[0], data[1], data[2],
						 data[3], data[4]);
		
		if (rc != (index - CAM_FLOW_BASE_IDX)) {
			printf("search flow failed: data4=%x data3=%x data2=%x data1=%x data0=%x\n",
				   ntohl(data[4]), ntohl(data[3]), ntohl(data[2]),
				   ntohl(data[1]), ntohl(data[0]));
			continue;
		}
		else
			printf("search flow worked: indx = %d, data4=%x data3=%x data2=%x data1=%x data0=%x\n",
				   rc, ntohl(data[4]), ntohl(data[3]), ntohl(data[2]),
				   ntohl(data[1]), ntohl(data[0]));


		/* removing flow */
		if (luc_camDelFlow((index - CAM_FLOW_BASE_IDX)) != NFX_OK) {
			printf("remove flow failed: index =%d\n", index);
			continue;
		}
		else
			printf("remove flow worked: index =%d\n", index);	

		/* verify flow has been removed */
		rc = luc_camFind(CAM_IFCMD_BLK_DATA, 0, data[0], data[1], data[2],
						 data[3], data[4]);
		
		if (rc != -1 ) 
			printf("flow remove verify failed: indx = %d,\n", index);
		else
			printf("flow remove verify worked: indx = %d,\n", index);
	}

	return;
}

void
luc_camPrintRegs()
{
	luc_camIfReg_t	*pCamIf;
	
	pCamIf = luc_camCfg.camBase;

	printf("lucCmd:        %x\n", pCamIf->lucCmd);
	printf("lucStatus:     %x\n", pCamIf->lucStatus);
	printf("lucStatusClr:  %x\n", pCamIf->lucStatus);
	printf("camDataW_0:    %x\n", pCamIf->camDataW[0]);
	printf("camDataW_1:    %x\n", pCamIf->camDataW[1]);
	printf("camDataW_2:    %x\n", pCamIf->camDataW[2]);
	printf("camDataW_3:    %x\n", pCamIf->camDataW[3]);
	printf("camDataW_4:    %x\n", pCamIf->camDataW[4]);
	printf("\n\n");
	printf("camCmd:        %x\n", pCamIf->camCmd);
	printf("camIdx:        %x\n", pCamIf->camIdx);
	printf("camDataR_0:    %x\n", pCamIf->camDataR[0]);
	printf("camDataR_1:    %x\n", pCamIf->camDataR[1]);
	printf("camDataR_2:    %x\n", pCamIf->camDataR[2]);
	printf("camDataR_3:    %x\n", pCamIf->camDataR[3]);
	printf("camDataR_4:    %x\n", pCamIf->camDataR[4]);
	printf("camIdxS:       %x\n", pCamIf->camIdxS);
	

	return;
}



#if xxxx
/*-----------------------------------------------------------------
 * Name: luc_camAddIpIf
 *
 * Description: Add a IP IF/VLAN entry into CAM lookup.
 *
 * Created by : Shaun Hui
 * 
 * Date Created: 7/01/01
 * 
 * Returns: Index to CAM data array or error. (see luc-cam.h)
 *-----------------------------------------------------------------
 */
int32 
luc_camAddIpIf( int32 link, int32 vlan, uint32 ip )
{
	int32	idx;
	uint32	vlanFlag;
	uint32	data[6];
	uint32	mask[6];
	
	/*-------------------------------
	 * Allocate a new entry
	 *-------------------------------
	 */
	idx = luc_camAlloc( CAM_IPIF_TBL_ID );
	if (idx < 0)
		return( LUC_CAM_ERR_NO_RSRC );

	/*-------------------------------
	 *
	 * 31                               0  
	 * +---------------------------------+ 
	 * |            IP  Address          |  data[0]: IP address
	 * +---------------------------------+ 
	 *
	 * 63                              32 
	 * +---------------+-+--+------------+ 
	 * |               |f| l|  vlan id   |  data[1]: flag, link, vlanid
	 * +---------------+-+--+------------+ 
	 *
	 * 95                              64 
	 * +---------------------------------+ 
	 * |                0s               | data[2]: 0
	 * +---------------------------------+ 
	 *
	 * 127                             96 
	 * +---------------------------------+ 
	 * |                0s               | data[3]: 0
	 * +---------------------------------+ 
	 *
	 *                       136      128  
	 *                         +---------+ data[4]: 0
	 *                         |         | data[5]: 0
	 *                         +---------+ 
	 *
	 *-------------------------------
	 */

	/*-------------------------------
	 * Check if entry is VLAN or IP.
	 *-------------------------------
	 */
	vlanFlag = (vlan ? CAM_VLAN_FLAG : 0);

	/*-------------------------------
	 * Generate lower 68 data bits.
	 *-------------------------------
	 */
	data[0] = ip;
	data[1] = vlanFlag | ((link & 0x00000003) << 12) | vlan;
	data[2] = 0;

	/*-------------------------------
	 * Generate upper  68 data bits.
	 *-------------------------------
	 */
	data[3] = 0;
	data[4] = 0;
	data[5] = 0;

	/*-------------------------------
	 * Generate lower 68 mask bits.
	 *-------------------------------
	 */
	mask[0] = 0xFFFFFFFF;
	mask[1] = 0x00007FFF;
	mask[2] = 0x00000000;

	/*-------------------------------
	 * Generate upper 68 mask bits.
	 *-------------------------------
	 */
	mask[3] = 0;
	mask[4] = 0;
	mask[5] = 0x0000000F;	/* last 4 bit table ID */

	/*-------------------------------
	 * Odd index contains lower 68 bit.
	 * Even index contains upper 68 bits.
	 *-------------------------------
	 */
	luc_camWrite( CAM_IFCMD_BLK_DATA, idx+1, data[0], data[1], data[2]);
	luc_camWrite( CAM_IFCMD_BLK_DATA, idx,   data[3], data[4], data[5]);

	luc_camWrite( CAM_IFCMD_BLK_MASK, idx+1, mask[0], mask[1], mask[2]);
	luc_camWrite( CAM_IFCMD_BLK_MASK, idx,   mask[3], mask[4], mask[5]);

	return (idx);
}

/*-----------------------------------------------------------------
 * Name: luc_camDelIpIf
 *
 * Description: Delete a IP IP/VLAN entry from CAM lookup.
 *
 * Created by : Shaun Hui
 * 
 * Date Created: 7/01/01
 * 
 *-----------------------------------------------------------------
 */
int32 
luc_camDelIpIf( int32 idx /* data array index */)
{
	
	/*-------------------------------
	 * Clear data array content.
	 *-------------------------------
	 */
	luc_camWrite( CAM_IFCMD_BLK_DATA, idx+1, 0, 0, 0);
	luc_camWrite( CAM_IFCMD_BLK_DATA, idx,   0, 0, 0);

	luc_camFree( CAM_IPIF_TBL_ID, idx );

}
#endif /* xxxx */
