/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        mnt-msg.c
 *
 * RCS:  $Id: mnt-msg.c,v 1.50 2002/11/05 20:17:42 shaun Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:
 *      Contains MOUNT Procedures and File System interface functions.
 * Created by: Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
#ifdef LINUX_TEST
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif	//LINUX_TEST
#include <nfx-incl.h>
#include <exports/nfs-api.h>

#define	MNT_DBG_ON


extern uint32 trace_enable;
#define     MNT_DBG_PRINTF  if (trace_enable) printf

/*-----------------------------------------------------------------
 * Name:       	mnt_msgBuildFSHdr 
 *
 * Description:
 *		Build a file system rquest message header for message that will
 *		be sent to the file system.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
void
mnt_msgBuildFSHdr(fs_MsgReqHdr *fsMsgReqHdrp,   /*File System message req hdr*/
		uint32 reqType,                         /*Req message type*/		 
		uint32 attrFlag, 						/*request attributes flag*/
		mnt_request_t *mntReqp)
{
	rpc_authSysInfo_t   	*authSysInfop;
	rpcCallHdr_t            *rpcCallHdrp;
	int32					i;

	authSysInfop = NULL;
	rpcCallHdrp = mntReqp->rpcCallHdrp;
	if (rpcCallHdrp)
		authSysInfop = &rpcCallHdrp->authFlavor.authSys;
	fsMsgReqHdrp->requestType = reqType;
	fsMsgReqHdrp->attrType = attrFlag;
	if ( ( authSysInfop == NULL) || 
		(authSysInfop && (authSysInfop->authtype != PRC_AUTH_SYS)) )
	{
		/*
		 *Authentication type is not AUTH_SYS.
		 */
		fsMsgReqHdrp->uid = 0;
		fsMsgReqHdrp->gid = 0;
		fsMsgReqHdrp->numGids = 0;
		for (i = 0; (i < FS_REQ_GIDS_MAX); i++)
			fsMsgReqHdrp->gids[i] = 0;
	}
	else
	{
		/*
		 *Authentication type is AUTH_SYS
		 */
		fsMsgReqHdrp->uid = authSysInfop->userId;
		fsMsgReqHdrp->gid = authSysInfop->groupId;
		if (authSysInfop->auxGidCount > FS_REQ_GIDS_MAX)
			fsMsgReqHdrp->numGids = FS_REQ_GIDS_MAX;
		else
			fsMsgReqHdrp->numGids = authSysInfop->auxGidCount;
		//printf("mnt_msgBuildFSHdr: No GIDs: %d\n", fsMsgReqHdrp->numGids);
		for (i = 0; (i < fsMsgReqHdrp->numGids); i++) 
		{
			fsMsgReqHdrp->gids[i] = authSysInfop->auxGids[i];
			//printf("Gid[%d]: %d\n", i, fsMsgReqHdrp->gids[i]);
		}
	}
	fsMsgReqHdrp->proto = FS_PROTO_NFS_V3; 
	if (rpcCallHdrp)
	{
		fsMsgReqHdrp->proto = 
			(rpcCallHdrp->progVer == NFS_VER2)? FS_PROTO_NFS_V2:
			(rpcCallHdrp->progVer == NFS_VER3)? FS_PROTO_NFS_V3: FS_PROTO_NFS_V4;
	}
	fsMsgReqHdrp->clientPrivate = mntReqp;
	MNT_DBG_PRINTF("mnt_msgBuildFSHdr: fsMsgReqHdrp->clientPrivate: %p\n", 
					fsMsgReqHdrp->clientPrivate);	


	return;
}

/*-----------------------------------------------------------------
 * Name:       	mnt_msgMapMountReq 
 *
 * Description:
 * Map Mount request received from client to a file 
 * system control message.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
int32
mnt_msgMapMountReq(mnt_request_t *mntReqp,
		evm_volId_t volId)
{
	nfp_fileSysCBPtr_t  fileSysCBp;
	uchar8				*buf;
	fs_MsgGetRootReq	*fsGetRootReqp;
	char				*pathNamep;
	int32 				msgLen;
	uint32				attrFlag;
	int32				volNameLen;

	MNT_DBG_PRINTF("mnt_msgMapMountReq: Enter\n");
	/*
	 *Allocate buffer to send request to file system
	 */
	fileSysCBp = mntReqp->fileSysCBp;
	msgLen = sizeof(fs_MsgGetRootReq) + fileSysCBp->nfp_nameLen + 1;
	buf = (uchar8 *)eee_ramAlloc(msgLen);
	if (buf == NULL)
		return (NFX_ERR);
	//Increment debug counter
	mnt_globalCfgp->mnt_debugCounters.mnt_eeeAllocCnt1++;
	memset(buf,0, msgLen);

	fsGetRootReqp = (fs_MsgGetRootReq *)buf;
	/* 
	 *Fill in File System (FS) request header msg
	 */ 
	attrFlag = 0;
	mnt_msgBuildFSHdr(&fsGetRootReqp->hdr, FS_REQ_GET_ROOT, attrFlag, mntReqp);

	fsGetRootReqp->volId.whole_thing = volId.whole_thing;
		
	fsGetRootReqp->pathLen = 0;
	volNameLen = 0;
	/*
	 *Get volume name
	 */
	mnt_getVolumeName(&fsGetRootReqp->pathName[0], 
			&volNameLen,
			&fileSysCBp->nfp_name[0], fileSysCBp->nfp_nameLen);
	MNT_DBG_PRINTF("%s: volname: %s Id: %llx \n",
		 __FUNCTION__, &fileSysCBp->nfp_name[0], fsGetRootReqp->volId.whole_thing);
	/*
	 *See if this is a root volume name
	 *or this Mount Req contains a sub-directory pathname
	 */
	if (fileSysCBp->nfp_nameLen != volNameLen);
	{
		pathNamep = &fileSysCBp->nfp_name[0] + volNameLen;
		fsGetRootReqp->pathLen = fileSysCBp->nfp_nameLen - volNameLen;
		memcpy(&fsGetRootReqp->pathName[0], pathNamep, 
				fsGetRootReqp->pathLen);
		fsGetRootReqp->pathName[fsGetRootReqp->pathLen] = '\0';
		MNT_DBG_PRINTF("mnt_msgMapMountReq: pathName: %s len: %d\n",
				fsGetRootReqp->pathName, fsGetRootReqp->pathLen);
	}

	mntReqp->fileMsgp = (char *)buf;
	mntReqp->fileMsgLen = msgLen;

	MNT_DBG_PRINTF("mnt_msgMapMountReq: Exit\n");

    return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:       	mnt_msgMountResp 
 *
 * Description:
 *	Map file system control message to a Mount reply message.
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
int32	/*NFX_OK or NFX_ERR*/
mnt_msgMountResp(
	uint32 status,							/*operation result status*/
	fs_MsgGetRootResp *fsGetRootRespMsgp,	/*file system resp. message*/
	char *txMsgp,							/*current pointer to transmit buffer*/
	int32 *txMsgLen,						/*current length of message*/
	mnt_request_t *mntReqp					/*Mount request block*/
)
{
	nfp_clientCBPtr_t		clientCBp;
	rpcCallHdr_t            *rpcCallHdrp;
	char                    *msgp;
	nfp_fileSysCBPtr_t      fileSysCBp;
	int32					length;

	MNT_DBG_PRINTF("mnt_msgMountResp: enter\n");
	rpcCallHdrp = mntReqp->rpcCallHdrp;
	length = 0;
	msgp = txMsgp;
	/*
	 *Set status
	 */
	*(uint32 *)msgp = status;
	MNT_DBG_PRINTF("mnt_msgMountResp: status: %d\n", status);

	msgp += 4;
	*txMsgLen += 4;

	MNT_DBG_PRINTF("%s: fileHandle: volId: %llx inum: %llx mntPnt: %llx ver: %d\n",
		__FUNCTION__, fsGetRootRespMsgp->dirHandle.volId.whole_thing,
				fsGetRootRespMsgp->dirHandle.inum, 
				fsGetRootRespMsgp->dirHandle.mntPnt,
				rpcCallHdrp->progVer);
	length = nfs_storeFhandle(rpcCallHdrp->progVer, 
								&fsGetRootRespMsgp->dirHandle, msgp);

	msgp += length;
	*txMsgLen += length;

	length = mnt_storeAuthFlavor(msgp);
	msgp += length;
	*txMsgLen += length;
	
	clientCBp = (nfp_clientCBPtr_t)mntReqp->clientCBp;
	if (nfp_validateClientCB(clientCBp) != NFX_OK)
		panic ("mnt_msgMountResp: Error-Bad Client Control Block!\n");
	fileSysCBp = (nfp_fileSysCBPtr_t)mntReqp->fileSysCBp;
	if (fileSysCBp && (fileSysCBp->nfp_flag & NFP_ENTRY_INUSED) )
	{
		memcpy(&fileSysCBp->nfp_fileHandle, &fsGetRootRespMsgp->dirHandle,
					sizeof(fs_Handle));	
		fileSysCBp->nfp_flag |= NFP_ENTRY_CACHE_FHANDLE;
//		MNT_DBG_PRINTF("mnt_msgMountResp: cache Root filehandle\n");
	}
	MNT_DBG_PRINTF("mnt_msgMountResp: exit\n");
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name: mnt_msgSendToFS
 * Description:
 *      Send a MOUNT control message to the file system for processing.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32		/*NFX_OK or NFX_ERR*/
mnt_msgSendToFs(mnt_request_t *mntReqp)
{
    uint32              srcPort;
    uint32              destPort;
	int32				rc;
	MNT_DBG_PRINTF("mnt_msgSendToFs: Enter\n");

#if defined(FS_STANDALONE)
	sim_sendMsgToFS((fs_MsgReqHdr *)mntReqp->fileMsgp, mntReqp->fileMsgLen, MNT_PROG);
#else	/*FS_STANDALONE*/
	srcPort = PACK_UNICAST_PORT_ADDRESS(
                                nfxMySlotId,
								nfxMyCPUId,
								nfs_globalCfgp->eee_nfsAppId,
								EEE_CONTROL,
								NFX_COS_HIGH,
								0);
	/*
	MNT_DBG_PRINTF("MNT TXRX : src app_id (%d) src slot (%d) src cpu (%d)\n", 
                        EEE_GET_APP_ID(srcPort),
                        EEE_GET_SLOT_NUM(srcPort),
                        EEE_GET_DEST_CPU(srcPort));
	*/
	destPort = PACK_UNICAST_PORT_ADDRESS(
								nfxMySlotId,
								NFX_NFP_FP2,
								nfs_globalCfgp->eee_efsAppId,	
								EEE_CONTROL,
								NFX_COS_HIGH,
								0);
	rc = eee_sendMessage((uchar8 *)mntReqp->fileMsgp, mntReqp->fileMsgLen,
						destPort, srcPort);
	if (rc != NFX_OK) 
	{
		printf("mnt_msgSendToFs: rc: %d\n", rc);
//		return (NFX_ERR);
	}
#endif	/*FS_STANDALONE*/

	/*
	 *Free local buffer after send  to file system
	 */
	if (eee_ramDealloc ((uchar8 *)mntReqp->fileMsgp) != NFX_OK )
	{
		printf("nfs_msgSendToFs: Failure to deallocate buf!\n");
		return (NFX_ERR);
	}
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	mnt_globalCfgp->mnt_debugCounters.mnt_totalNumOfReqXmtToFs++;

	//increment debug counter
	mnt_globalCfgp->mnt_debugCounters.mnt_eeeDeallocCnt1++;
	mntReqp->fileMsgp = NULL;
	mntReqp->fileMsgLen = 0;

	MNT_DBG_PRINTF("mnt_msgSendToFs: Exit\n");

	return (NFX_OK);	

}

/*-----------------------------------------------------------------
 * Name: mnt_msgXmtPkt
 *
 * Description:
 *      Transmit a MOUNT reply pkt to the client.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32		
mnt_msgXmtPkt(pkt_desc_t *tplXmitPktDescp, mnt_request_t *mntReqp)
{

	MNT_DBG_PRINTF("mnt_msgXmtPkt: tplXmitPktDescp: %08x mntReqp: %08x\n",
					tplXmitPktDescp, mntReqp);

	mntReqp->pktDescp =	(void *)tplXmitPktDescp;
	tplXmitPktDescp->appInfo = mntReqp;
	if (trace_enable != 0)
		mnt_xmtDecodePkt(mntReqp);

	/* 
	 * Perform Host to Net conversion.
	 */
	xdr_swap(tplXmitPktDescp->pHdr, tplXmitPktDescp->hdrSize, 1);

	if (tpl_xmtPkt (mntReqp->tplConnp, tplXmitPktDescp) != NFX_OK)
	{
		MNT_DBG_PRINTF("mnt_msgXmtPkt: tpl_xmtPkt failed!\n");		
		mnt_globalCfgp->mnt_debugCounters.mnt_tplXmtSendErr++;
		return (NFX_ERR);
	}	
	mnt_globalCfgp->mnt_debugCounters.mnt_tplXmtPktSend++;
	return (NFX_OK);
}

/*
 *------------------------------------------------------------------------------
 * Function Name:       mnt_msgSendErrResp
 *
 * Description:
 *  This function is call when an error occurred during processing of the
 *  client's request.  This function is call to send an error response
 *      back to the client.
 *
 * Created by:          Joseph Vu
 *
 * Created:             5/28/2000
 *
 *------------------------------------------------------------------------------
 */
int32	/* NFX_OK or NFX_ERR*/
mnt_msgSendErrResp (mnt_request_t *mntReqp, int32 status)
{
	nfp_clientCBPtr_t       clientCBp;
	pkt_desc_t              *tplXmitPktDescp;
	char                    *txMsgp;
	int32                   txMsgLen;

	MNT_DBG_PRINTF("mnt_msgSendErrResp:  Enter\n");

	if (mntReqp == NULL)
		return (NFX_ERR);
	clientCBp = mntReqp->clientCBp;
	if (nfp_validateClientCB(clientCBp) != NFX_OK)
	{
		/* Shaun 11/05/02	#2676
		 * Return error instead of panic.
		 */
		return (NFX_ERR);
	}

	if (tpl_validateConnCb( mntReqp->tplConnp ) != NFX_OK)
	{
		MNT_DBG_PRINTF("mnt_msgSendErrResp: Bad TPL PConn ID (%p)!\n",mntReqp->tplConnp );
		return (NFX_ERR);
	} 
	MNT_DBG_PRINTF("mnt_msgSendErrResp: tcpConnp: %p\n",mntReqp->tplConnp ); 
	tplXmitPktDescp = TPL_ALLOC_PACKET(mntReqp->tplConnp, 0);
	if (tplXmitPktDescp == NULL)
	{
		//panic ("mnt_msgSendErrResp: Failure to alloc TPL packet!\n");
		nfs_globalCfgp->nfs_debugPktCounters.tpl_noPktRes++;
		return (NFX_ERR);
	}
	txMsgp = (char *)tplXmitPktDescp->pHdr;
	txMsgLen = 0;
	if (mntReqp->rpcCallHdrp == NULL)
	{
		goto mnt_msgSendErrResErr;				
	}
	/*
	 *Start with RPC reply header
	 */
	txMsgLen += nfs_msgRpcReply(txMsgp, mntReqp->rpcCallHdrp->rpc_xid,
				RPC_REPLY_MSG_ACCEPTED, RPC_REPLY_SUCCESS);
	txMsgp += txMsgLen;
	/*
	 *Set Status
	 */
	*(uint32 *)txMsgp = status;
	txMsgLen += 4;

	/*
	 *Set packet desc hdr size
	 */
	tplXmitPktDescp->hdrSize = txMsgLen;
	if (mnt_msgXmtPkt(tplXmitPktDescp, mntReqp) == NFX_OK)
	{
		/*
		 *Set request state to in-progress
		 */
		mntReqp->reqState = MNT_REQ_REPLIED;
	}
	else
	{
		//Calling function shall call mnt_freeAll()
		//to free mntReqp
		goto mnt_msgSendErrResErr;				
	}
	return (NFX_OK);

mnt_msgSendErrResErr:
	TPL_FREE_PACKET(mntReqp->tplConnp, tplXmitPktDescp);
	return (NFX_ERR);
}

/*
 *----------------------------------------------------------------
 * Function Name:      mmt_msgRcv 
 *
 * Description: 
 * This function is registered with the EEE layer to receive MOUNTD SSC
 * response messages. 
 *
 * Created by:        	Joseph Vu 
 *
 * Created:      	5/28/2000
 * 
 *
 *---------------------------------------------------------------
 */
void mnt_msgRcv(eee_desc_t *eDescp )
{
	nfs_request_t           *nfsReqp;
	eee_descBuf_t           *bd;
	mnt_infoResp_t          *mntInfoResp;
	mnt_infoEntry_t         *mtnInfoEntryp;
	char                    *exportInfop;
	nfp_fileSysCBPtr_t      fileSysCBp;
	mnt_request_t           *mntReqp;
	int32                   msgLen;
	int32					msgType;

	
	if (eDescp == NULL)
	{
		return;
	}
	bd = (eDescp->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER) ?
			&eDescp->bd[0] : &eDescp->bd[1];
	if (bd == NULL)
	{
		eee_freePacket(eDescp);
		return;
	}

	msgType = *(uint32 *)bd->buf;	
	MNT_DBG_PRINTF("mnt_msgRcv: msgType: %d\n", msgType);
	if ((msgType >= MNT_EXPORT_ENABLED_REQ) &&
			 (msgType <= MNT_MSG_MAX))
	{
		//Share enable/disable request
		return (mnt_getMountdReq(eDescp));
	}

	mntInfoResp = (mnt_infoResp_t *)(bd->buf);
	msgLen = eDescp->hdr.control & EDESC_TOTAL_LEN;
	mtnInfoEntryp = &mntInfoResp->mtnInfoEntry[0];

	MNT_DBG_PRINTF("mnt_msgRcv: msgLen: %d pathName: %s pathNameLen: %d volId: %lld hostAddr: %08x\n",
				msgLen,
				mtnInfoEntryp->pathName, 
				mtnInfoEntryp->pathNameLen,
				mtnInfoEntryp->volId.whole_thing,
				mtnInfoEntryp->hostAddr);

	exportInfop = &mtnInfoEntryp->pathName[0] + mtnInfoEntryp->pathNameLen;
	MNT_DBG_PRINTF("mnt_msgRcv: exportInfo: %s exportLen: %d\n",
					exportInfop, mtnInfoEntryp->exportInfoLen);	

	MNT_DBG_PRINTF("mnt_msgRcv: status: %d clientPrivate: %p\n", 
			mntInfoResp->msgHdr.status, mntInfoResp->msgHdr.private);

	MNT_DBG_PRINTF("mnt_msgRcv: MOUNTD response export option mask: %08x\n",
			mtnInfoEntryp->exportOptionMask);

	if (mntInfoResp->msgHdr.msgType == MNT_INFO_RESP_BY_ID)
		mnt_globalCfgp->mnt_debugCounters.mnt_rxRespByIdFromSsc++;
	else if (mntInfoResp->msgHdr.msgType == MNT_INFO_RESP)
		mnt_globalCfgp->mnt_debugCounters.mnt_rxRespByNameFromSsc++;

	mntReqp = (mnt_request_t *)mntInfoResp->msgHdr.private;
	if (nfp_validateMntReqCB(mntReqp) != NFX_OK )
	{
		printf("mnt_msgRcv: Invalid MNT Client Control Block!\n");
		//luc_panic();
		eee_freePacket(eDescp);
		return;
	}
	MNT_DBG_PRINTF("mnt_msgRcv: NFS share state: %s\n", 
		(mtnInfoEntryp->state == MNT_SHARE_EXPORT_DISABLED)? "disable": "enable");

	/*
	 *Get file system control bock
	 */
	fileSysCBp = mntReqp->fileSysCBp;

	if ( (mtnInfoEntryp->volId.whole_thing <= 0) ||
		 (mntInfoResp->msgHdr.status != NFX_OK) ||
		 (nfp_validateFileSysCB(fileSysCBp) != NFX_OK) ||
         (!(fileSysCBp->nfp_flag & NFP_ENTRY_INUSED)) ||
		 (mtnInfoEntryp->state == MNT_SHARE_EXPORT_DISABLED)   )
	{
		/*Either Volume/filesystem is not ready 
		 *or NFS share is in disable state
		 *Return error = NFSERR_IO (5)to client.
		 *Do appropriate clean up.
		 */
		if ( fileSysCBp && 
			(mntInfoResp->msgHdr.status ==  MOUNTD_ERR_BAD_VOL_ID) &&
			(mtnInfoEntryp->volId.whole_thing <= 0) )
		{

			nfs_delFileSysCBFromList(fileSysCBp, mntReqp->clientCBp);
			nfp_deallocFileSysCB(fileSysCBp);
		}		
		mnt_globalCfgp->mnt_statCounters.mnt_errIO++;
		nfsReqp = mntReqp->nfsReqp;
		if (nfp_validateNfsReqCB(nfsReqp) == NFX_OK )
		{
			if ( (nfsReqp->reqFlag & NFS_REQ_FILESYS_NOT_READY) &&
				 (nfsReqp->rpcCallHdrp) )
			{
				MNT_DBG_PRINTF("mnt_msgRcv: Volume not ready. nfsReqp: %p!\n", nfsReqp);
				//Free RX Pkt Descp     
				if ( (nfsReqp->rxPktDesc != NULL) &&
					(pkt_validateDesc(nfsReqp->rxPktDesc , PKT_CHK_DUP ) == NFX_OK) )
				{
					PKT_FREE_PACKET(nfsReqp->rxPktDesc);
					nfsReqp->rxPktDesc = NULL;
				}
				nfsReqp->reqState = NFS_REQ_DONE;
				nfs_msgFreeAll(nfsReqp, 0);
				mnt_freeAll(mntReqp);
				eee_freePacket(eDescp);
				return;
			}
			nfsReqp->reqState = NFS_REQ_DONE;
			nfs_msgFreeAll(nfsReqp, 0);
		}
		MNT_DBG_PRINTF("mnt_msgRcv: Volume not ready.  Send err resp mntReqp: %p!\n", mntReqp);

		if (mnt_msgSendErrResp(mntReqp, NFSERR_IO ) != NFX_OK)
		{
			mnt_freeAll(mntReqp);
		}
		eee_freePacket(eDescp);

		return;
	}

	//Check for duplicate Filesys Control Block
	if ( fileSysCBp && 
			(mntInfoResp->msgHdr.msgType == MNT_INFO_RESP_BY_ID) ) 
	{
		nfp_fileSysCBPtr_t      duplicateFileSysCBp;

		//This should happen only when server restarted

		duplicateFileSysCBp = mnt_isFileSysDuplicate(fileSysCBp, mtnInfoEntryp);
		if (duplicateFileSysCBp)
		{
			nfs_delFileSysCBFromList(fileSysCBp, mntReqp->clientCBp);
			nfp_deallocFileSysCB(fileSysCBp);
			fileSysCBp = duplicateFileSysCBp;
		}
		else
		{
			/*
	 	 	 *Saved volume name
	 	 	 */
			memcpy(&fileSysCBp->nfp_name[0], &mtnInfoEntryp->pathName[0], 
				mtnInfoEntryp->pathNameLen);
			fileSysCBp->nfp_nameLen = mtnInfoEntryp->pathNameLen;
		}
	}	
	fileSysCBp->laddr = mtnInfoEntryp->localIpAddr;
	fileSysCBp->nfp_fileHandle.volId = mtnInfoEntryp->volId;
	fileSysCBp->protocol.nfp_nfsFileSysInfo.exportInfoLen = 
		mtnInfoEntryp->exportInfoLen;
	memcpy(&fileSysCBp->protocol.nfp_nfsFileSysInfo.exportInfoStr[0], 
			exportInfop, 
			mtnInfoEntryp->exportInfoLen);

	/*
	 *Set default export option: 
	 *rw: read-write 
	 *root_squash:  request from root clients (UID=0) 
	 *are mapped to nobody user and
	 *group ID so they will only have privilleges associated 
	 *with each other.
	 */
	mnt_storeExportOpt(fileSysCBp, mtnInfoEntryp);
	MNT_DBG_PRINTF("mnt_msgRcv:  default export mask: %08x\n", 
					fileSysCBp->defaultExpMask);
	/*
	 *set nfs share NFP flag to correct state 
	 */
	fileSysCBp->nfp_flag &= ~NFP_SHARE_INIT_STATE;	
	fileSysCBp->nfp_flag |= NFP_SHARE_ENABLED;
	MNT_DBG_PRINTF("fileSysCBp->nfp_flag: %08x\n", 
				fileSysCBp->nfp_flag);

	switch (mntInfoResp->msgHdr.msgType)
	{

		case MNT_INFO_RESP:
		/*
	 	 *Send Get Root Message to the file system to get
	 	 *get the root directory handle
	 	 */
		if (mnt_msgMapMountReq(mntReqp,  mtnInfoEntryp->volId) == NFX_ERR)
		{
			nfs_errMsg(mntReqp->rpcCallHdrp->rpc_xid, mntReqp->rpcCallHdrp->rpc_proc,
				NFSERR_MAP_FS_MSG);
			mnt_freeAll(mntReqp);
			break;	
		}

		if (mnt_msgSendToFs(mntReqp) == NFX_ERR)
		{
			nfs_errMsg(mntReqp->rpcCallHdrp->rpc_xid, mntReqp->rpcCallHdrp->rpc_proc,
					NFSERR_SENDMSG_TO_FS);
			mnt_freeAll(mntReqp);			
		}
		break;

		case MNT_INFO_RESP_BY_ID:
#ifdef	NOTUSED
		/*
		 *Saved volume name
		 */
		memcpy(&fileSysCBp->nfp_name[0], &mtnInfoEntryp->pathName[0], 
				mtnInfoEntryp->pathNameLen);
		fileSysCBp->nfp_nameLen = mtnInfoEntryp->pathNameLen;
#endif	//NOTUSED
		/*
		 *Process NFS request
		 */
		nfs_tplProcessNfsReq(mntReqp->nfsReqp);

		mnt_freeAll(mntReqp);
		break;

	}

		
	eee_freePacket(eDescp);

	return;
}


/*-----------------------------------------------------------------
 * Name:	mnt_msgFpRcv 
 *
 * Description:
 *     MOUNT/FP receive function. 
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
void mnt_msgFpRcv( eee_desc_t *eDescp )
{
	mnt_pktCounters_t	*txPktStatp;
	nfp_clientCBPtr_t	clientCBp;
	mnt_request_t    	*mntReqp;
	fs_MsgRespHdr       *fsMsgRespHdrp;
	pkt_desc_t          *tplXmitPktDescp;
	char                *txMsgp;
	eee_descBuf_t       *bd;
	int32               txMsgLen;
	int32				status;
	int32				version;

	MNT_DBG_PRINTF("mnt_msgFpRcv: Enter-eDescp: %08x\n",eDescp);


	bd = (eDescp->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER) ?
		&eDescp->bd[0] : &eDescp->bd[1];

	if (bd == NULL)
	{
		eee_freePacket(eDescp);
		return;	
	}

	fsMsgRespHdrp = (fs_MsgRespHdr   *) (bd->buf);
	mnt_globalCfgp->mnt_debugCounters.mnt_totalNumOfRespRcvdFromFs++;

	mntReqp = (mnt_request_t *)fsMsgRespHdrp->clientPrivate;

	if (nfp_validateMntReqCB(mntReqp) != NFX_OK )
	{
		panic("mnt_msgFpRcv: Bad Client Control Block!\n");
	
		//luc_panic();	
		eee_freePacket(eDescp);
		return;	
	}
	/*
	 *Check for the Admin Get fileHandle Request
	 *originate from MOUNTD SSC 
	 */
	if (mntReqp->reqState == MNT_REQ_GET_FHANDLE)
	{
		MNT_DBG_PRINTF("mnt_msgFpRcv: MNT_REQ_GET_FHANDLE Resp:\n");
		mnt_sendFileHandleResp(eDescp);	
		eee_freePacket(eDescp);
		return;	
	}
	version = mntReqp->rpcCallHdrp->progVer;

	if ((fsMsgRespHdrp->fsStatus == FS_ERR_NOT_MOUNTED) ||
       	(fsMsgRespHdrp->fsStatus == FS_ERR_NO_RESOURCE))
	{
		mnt_freeAll(mntReqp);
		eee_freePacket(eDescp);
		mnt_globalCfgp->mnt_debugCounters.mnt_fsErrNoResource++;
		//Increment SNMP counters
		if (version == MNT_VER2)
			nfs_globalCfgp->nfs_v2XmtPktCounters.nfs_fsResourceErr++;
		else
			nfs_globalCfgp->nfs_v3XmtPktCounters.nfs_fsResourceErr++;

		return;
	}


	if ( (fsMsgRespHdrp->fsStatus < FS_ERR_NONE) || 
		(fsMsgRespHdrp->fsStatus > FS_ERR_SERVERFAULT) )
	{
		if (version >= MNT_VER3)
			status = NFSERR_SERVERFAULT;
		else
			status = NFSERR_ACCES;
	}
	else
		status = fsMsgRespHdrp->fsStatus;

	
	MNT_DBG_PRINTF("mnt_msgFpRcv: status :%08x\n", status);
	MNT_DBG_PRINTF("mnt_msgFpRcv: mntReqp:%p tplConnp: %p\n", 
					mntReqp, mntReqp->tplConnp);

	clientCBp = mntReqp->clientCBp;

	/*
	 *Allocate packet descriptor for transmission
	 */
	tplXmitPktDescp = TPL_ALLOC_PACKET(mntReqp->tplConnp, 0);
	if (tplXmitPktDescp == NULL)
	{
		//panic ("mnt_msgFpRcv: Failure to allocate TPL packet!\n");
		nfs_globalCfgp->nfs_debugPktCounters.tpl_noPktRes++;
		mnt_freeAll(mntReqp);
		eee_freePacket(eDescp);
		return;
	}
	txMsgp = (char *)tplXmitPktDescp->pHdr;
	txMsgLen = 0;
	/*
	 *Build RPC reply message to send to the client
	 */
    txMsgLen += nfs_msgRpcReply(txMsgp, mntReqp->rpcCallHdrp->rpc_xid,
                RPC_REPLY_MSG_ACCEPTED, RPC_REPLY_SUCCESS);
    txMsgp += txMsgLen;

	txPktStatp = &mnt_globalCfgp->mnt_xmtPktCounters;

	switch (fsMsgRespHdrp->requestType)
	{
		case FS_REQ_GET_ROOT:
		{
			fs_MsgGetRootResp   *fsGetRootRespMsgp;

			fsGetRootRespMsgp = (fs_MsgGetRootResp *)bd->buf;
			mnt_msgMountResp (status, fsGetRootRespMsgp,
							txMsgp, &txMsgLen, mntReqp);

			txPktStatp->mnt_addMnt++;
			break;
		}
#ifdef	NOTUSED
		case FS_REQ_MOUNT:

			break;
		case FS_REQ_UNMOUNT:	
			break;
#endif	//NOTUSED
		default:
			TPL_FREE_PACKET(mntReqp->tplConnp, tplXmitPktDescp);
			mnt_freeAll(mntReqp);
			eee_freePacket(eDescp);
			return;

	}
	/*
	 *Set packet desc hdr size
	 */
	tplXmitPktDescp->hdrSize = txMsgLen;

	if (mnt_msgXmtPkt(tplXmitPktDescp, mntReqp) == NFX_OK)
	{
		/*
		 *Set request state to in-progress
		 */
		mntReqp->reqState =  MNT_REQ_REPLIED;
	}
	else
	{
		TPL_FREE_PACKET(mntReqp->tplConnp, tplXmitPktDescp);
		mnt_freeAll(mntReqp);
	}

	eee_freePacket(eDescp);
	MNT_DBG_PRINTF("mnt_msgFpRcv: Exit\n");

}


#if defined(FS_STANDALONE)	
	/*-------------------------------------------------
	 *Use to for testing with File System via Sockets
	 *----------------------------------------------
	 */
#define FS_PAD_64(_what_)                                       \
    ((int32)(_what_)) += (sizeof(uint64) - 1);          \
    ((int32)(_what_)) &= ~(sizeof(uint64) - 1);

#define FS_SIM_MSG_BUF_LEN				9000
#define FS_SIM_NFS_INT_NFS_PORT         3005
#define FS_SIM_NFS_INT_FS_PORT          3055
struct sockaddr_in  IntSockFsAddr;
int32               mySocket;
fs_Handle			globalRootFhandle;

sim_msgBuildFSHdr(fs_MsgReqHdr *fsMsgReqHdrp,   /*File System message req hdr*/
        uint32 reqType,                         /*Req message type*/
        uint32 attrFlag)                        /*request attributes flag*/
{
    rpc_authSysInfo_t       *authSysInfop;
    rpcCallHdr_t            *rpcCallHdrp;

    fsMsgReqHdrp->requestType = reqType;
    fsMsgReqHdrp->attrType = attrFlag;
    fsMsgReqHdrp->uid = 0; 
    fsMsgReqHdrp->gid = 0;
    /*
     *TODO: Check with Bob F. regarding the proto field for MOUNT
     */
    fsMsgReqHdrp->proto = (rpcCallHdrp->progVer == NFS_VER2)? FS_PROTO_NFS_V2:
            (rpcCallHdrp->progVer == NFS_VER3)? FS_PROTO_NFS_V3: FS_PROTO_NFS_V4;
    fsMsgReqHdrp->clientPrivate = (void *)0;


    return;
}

int32
sim_mntMsgMapMountReq()
{
    uchar8              *buf;
    fs_MsgGetRootReq    *fsGetRootReqp;
    int32               msgLen;
    uint32              attrFlag;

    MNT_DBG_PRINTF("mnt_msgMapMountReq: Enter\n");
    /*
     *Allocate buffer to send request to file system
     */
    msgLen = sizeof(fs_MsgGetRootReq);
    buf = (uchar8 *)eee_ramAlloc(msgLen);		//JVU: THIS FUNCTION IS NOT
												//CURRENTLY USED
    if (buf == NULL)
        return (NFX_ERR);
    memset(buf,0, msgLen);

    fsGetRootReqp = (fs_MsgGetRootReq *)buf;
    /*
     *Fill in File System (FS) request header msg
     */
    attrFlag = 0;
    sim_msgBuildFSHdr(&fsGetRootReqp->hdr, FS_REQ_GET_ROOT, attrFlag);

    fsGetRootReqp->volId.whole_thing = 1;       /*DEBUG ONLY: for testing with Bob's FsSim*/

	sim_sendMsgToFS((fs_MsgReqHdr *)buf, msgLen, MNT_PROG);
    MNT_DBG_PRINTF("mnt_msgMapMountReq: Exit\n");

    return (NFX_OK);
}


int32
sim_bindToFs()
{
    struct hostent      *host;
    struct sockaddr_in  sin;

	printf("sim_bindToFs: enter\n");
    // Create the internal socket.
    mySocket = socket(AF_INET, SOCK_DGRAM, 0);
    if (mySocket < 0)
    {
        perror("int socket");
        exit(2);
    }
    // Bind the nfs end of the internal socket.
    bzero((char *) &sin, sizeof(struct sockaddr_in));
    sin.sin_family=AF_INET;
    sin.sin_port = htons(FS_SIM_NFS_INT_NFS_PORT);
    host = gethostbyname("localhost");
    if (!host) {
        perror ("gethostbyname");
        exit(2);
    }
    memcpy (&sin.sin_addr, host->h_addr, host->h_length);
    if (bind(mySocket,(struct sockaddr *)&sin,sizeof(sin)) != 0) {
        perror("bind int nfs socket");
        exit(2);
    }
    // Create the address to send to for the internal socket (the
    // fs end of the socket).
    bzero((char *) &IntSockFsAddr, sizeof(struct sockaddr_in));
    IntSockFsAddr.sin_family=AF_INET;
    IntSockFsAddr.sin_port = htons(FS_SIM_NFS_INT_FS_PORT);
    host = gethostbyname("localhost");
    if (!host) {
        perror ("gethostbyname");
        exit(2);
    }
    memcpy (&IntSockFsAddr.sin_addr, host->h_addr, host->h_length);
	printf("sim_bindToFs: exit\n");

}

void sim_processFSResponseMntMsg (fs_MsgRespHdr *fsMsgRespHdrp)
{
   	mnt_pktCounters_t   *txPktStatp;
    nfp_clientCBPtr_t   clientCBp;
    mnt_request_t       *mntReqp;
    pkt_desc_t          *tplXmitPktDescp;
    char                *txMsgp;
    int32               txMsgLen;
    int32               status;
	fs_MsgGetRootResp   *fsGetRootRespMsgp;


	printf("sim_processFSResponseMsg: status for FS: %d\n", fsMsgRespHdrp->fsStatus);
    if (fsMsgRespHdrp->fsStatus >= FS_ERR_DEFERRED)
        status = NFSERR_SERVERFAULT;
    else
        status = fsMsgRespHdrp->fsStatus;

	fsGetRootRespMsgp = (fs_MsgGetRootResp *)fsMsgRespHdrp;
	if (fsMsgRespHdrp->requestType == FS_REQ_GET_ROOT)
	{
		printf("sim_processFSResponseMntMsg: dirHandle.volId: %ld\n", 
								fsGetRootRespMsgp->dirHandle.volId.whole_thing);
		printf("sim_processFSResponseMntMsg: inum: %ld\n", fsGetRootRespMsgp->dirHandle.inum);
		printf("sim_processFSResponseMntMsg: gen: %d\n", fsGetRootRespMsgp->dirHandle.gen);
		memcpy(&globalRootFhandle, &fsGetRootRespMsgp->dirHandle, sizeof(fs_Handle));
	}

    MNT_DBG_PRINTF("sim_processFSResponseMntMsg: Exit\n");
}
int32 sim_sendMsgToFS(fs_MsgReqHdr *fsReq, uint32 length, uint32 protocol)
{
	int32				rc;
	char                *respBuffPtr, respBuf[FS_SIM_MSG_BUF_LEN + 8];
	eee_desc_t 			eDescriptor;
	eee_descBuf_t  		*bd;
	eee_desc_t  		*eDescp;

	// Use 8-byte aligned req/resp bufs
	respBuffPtr = respBuf;
	FS_PAD_64(respBuffPtr);

	fsReq->sockIndex = 5;
    // Send the request.
    rc = sendto(    mySocket,
                    fsReq,
                    length,
                    0,
                    (struct sockaddr *) &IntSockFsAddr,
                    sizeof(struct sockaddr_in));

    if (rc != length)
    {
        perror("sendto");
		/*
        FS_ASSERT(FALSE);
		*/
		printf("ERROR: sendto!\n");
    }


    // Wait for the response.
    rc = recv(  mySocket,
                respBuffPtr,
                FS_SIM_MSG_BUF_LEN,
                0);
    if (rc == -1)
    {
        perror("recv");
		printf("ERROR: recv!\n");
		/*
        FS_ASSERT(FALSE);
		*/
    }

	eDescp = &eDescriptor;
	bd = &eDescp->bd[1];
	bd->buf = respBuffPtr;
	bd->len = FS_SIM_MSG_BUF_LEN;

	if (protocol == MNT_PROG)
	{
		mnt_msgFpRcv( eDescp );
	}
	else
	{
		nfs_msgRcv( eDescp );
	}		
	
}

#endif	/*defined(FS_STANDALONE)*/

