/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        rpc-auth.c
 *
 * RCS:  $Id: rpc-auth.c,v 1.14 2002/09/13 00:19:20 joseph Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description: RPC Authentication 
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/18/01
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <exports/rpc.h>
#include <exports/evm-api.h>
#include <exports/fs-api.h>

uint64				mnt_rxAuthTypeErr = 0;

rpc_globalCfg_t *rpc_globalCfgp;	//RPC global cfg

/*-----------------------------------------------------------------
 * Name:   	rpc_extractAuthSysInfo  
 *
 * Description:
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
int32	/*length = number of bytes*/
rpc_extractAuthSysInfo (char *msgp, 		
			rpc_authSysInfo_t *authSysp
)
{
	int32			i;
	int32			length;
	uint32			nameLen;
	uint32 			fieldLen;
	uint32			*msg32p;
	char			*msg8p;

	msg32p = (uint32 *)msgp;

	/*
	 *Parse the credential field
	 */	
	authSysp->authtype = ntohl(*msg32p++);
	fieldLen = 4;

	if ((authSysp->authtype != PRC_AUTH_SYS) &&
		(authSysp->authtype != RPC_AUTH_NONE) )	
	{
		mnt_rxAuthTypeErr++;
		return (NFX_ERR);
	}

	/*
	 *Get authentication field len
	 */
	authSysp->authlen = ntohl(*msg32p++);
	fieldLen += 4;

	/*
	 *Check for AUTH_NONE or AUTH_NULL Verifier field
	 */
	if (authSysp->authtype == RPC_AUTH_NONE) 
	{
		uint32	 	verifier_authFlavor;
		uint32		verifier_authLen;

		verifier_authFlavor = ntohl(*msg32p++);
		verifier_authLen = ntohl(*msg32p++);
		fieldLen += 8;
		return (fieldLen);
	}


	/*
	 *Handle AUTH_SYS Type
	 */

	/*
	 *Skip stamp field
	 */
	msg32p++;
	msg8p = (char *)msg32p;		
	
#ifdef NOTUSED
	/*
	 *Extract hostname
	 */
	length = nfs_extractString(msg8p, &hostName, &nameLen, FS_NAME_MAX);
	if (length == NFX_ERR)
	{
		return (fieldLen);
	}
#endif

	/* Shaun 08/10/02
	 *
	 * Performance tuning:
	 * Skip hostname instead of extracting it.
	 * Saved call to memcpy.
	 */

	/*
	 * len of hostname + len field
	 */
    length = ntohl(*msg32p) + 4;

	/*
	 * Make length multiple of 4.
	 */
    if (length & 0x03)
    {
        length >>= 2;
        length++;
        length <<= 2;
    }

	msg8p += length;

	msg32p = (uint32 *)msg8p;

	authSysp->userId  = ntohl(*msg32p++);
	authSysp->groupId = ntohl(*msg32p++);

	/*
	 *Extract GIDs
	 */
	authSysp->auxGidCount = ntohl(*msg32p++);
	if ( (authSysp->auxGidCount < 0) ||
		(authSysp->auxGidCount >= NFS_AUTH_SYS_GID_COUNT) )
		authSysp->auxGidCount = 0;
	for (i = 0; i < authSysp->auxGidCount; i++)
		authSysp->auxGids[i] = ntohl(*msg32p++);

	fieldLen += authSysp->authlen;

	/*
	 *Accomodate field length for AUTH_SYS verifier field
	 *of Auth None Flavor & length field
	 */
	fieldLen += 8;

	return (fieldLen);	

}


/*-----------------------------------------------------------------
 * Name:   	rpc_getHdrSize 
 *  
 * Description:
 *  Get length of RPC header.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
uint32	/*length = Number of bytes in the RPC Call header*/
rpc_getHdrSize(rpcCallHdr_t *pktRpcCallHdrp)
{
	rpc_authSysInfo_t	authSys;
	uint32				rpcHdrSize;
	int32				fieldLen;
	char				*msgp;

	/*
	 *Fix length for XID, Direction, RPC Ver, RPC Program No., Program Ver and Procedure No.
	 */
	rpcHdrSize = 24;
	msgp = (char *)pktRpcCallHdrp + rpcHdrSize;
	fieldLen = rpc_extractAuthSysInfo(msgp, &authSys);
	if (fieldLen == NFX_ERR)
		return (NFX_ERR);

	rpcHdrSize += fieldLen;	
	
	return (rpcHdrSize);
}
/*-----------------------------------------------------------------
 * Name:   	rpc_processAuth 
 *  
 * Description:
 * 	Process RPC header information. 
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
rpcCallHdr_t *rpc_processAuth(rpcCallHdr_t	*pktRpcCallHdrp, uint32 *rpcHdrSize)
{
	char			*msgp;
	rpcCallHdr_t	*newRpcCallp;
	uint32			length;
	uint32			fieldLen;

	if (pktRpcCallHdrp == NULL)
		return (NULL);
	length = 0;

	newRpcCallp = (rpcCallHdr_t *) eee_ramAlloc(sizeof(rpcCallHdr_t));
	if (newRpcCallp == NULL)
		return (NULL);

	newRpcCallp->rpc_xid = pktRpcCallHdrp->rpc_xid;
	newRpcCallp->rpc_direction = pktRpcCallHdrp->rpc_direction;
	newRpcCallp->rpc_rpcVer = pktRpcCallHdrp->rpc_rpcVer;
	newRpcCallp->rpc_prog = pktRpcCallHdrp->rpc_prog;
	newRpcCallp->progVer = pktRpcCallHdrp->progVer;
	newRpcCallp->rpc_proc = pktRpcCallHdrp->rpc_proc;

	length = 24;
	msgp = (char *)pktRpcCallHdrp + length;

	fieldLen = rpc_extractAuthSysInfo(msgp, &newRpcCallp->authFlavor.authSys);
	if (fieldLen == NFX_ERR)
	{
		goto rpc_processAuthErr;	
	}	
	length += fieldLen;

	*rpcHdrSize = length;

	return (newRpcCallp);

rpc_processAuthErr:
	/*
	 *Free alloc memory
	 */
	eee_ramDealloc (newRpcCallp);
	return (NULL);
}

/*-----------------------------------------------------------------
 * Name:  rpc_processReplyHdr  
 *
 * Description:
 * 	Process RPC reply header received from the client. 
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/02
 *
 *-----------------------------------------------------------------
 */
int32  //length in byte of RPC reply hdr or NFX_ERR
rpc_processReplyHdr (char *rpcHdrp)
{
	int32			rpcReplyHdrLen;
	uint32			replyXid, replyDir, replyStatus;
	uint32			*rpcReplyHdrp;

	rpcReplyHdrLen = 0;

//TODO:  Define C template for RPC reply hdr for better 
//Access
	//Format or RPC reply message
	//Get Xid, reply (1), reply status
	
	rpcReplyHdrp = (uint32 *)rpcHdrp;
	replyXid =  *rpcReplyHdrp++;
	replyDir =  *rpcReplyHdrp++;
	replyStatus = *rpcReplyHdrp++;
	rpcReplyHdrLen = 12;

	//Check reply status
	switch (replyStatus)
	{
		case RPC_REPLY_MSG_ACCEPTED:
		{
			uint32		authType, authLen;
			uint32		acceptStatus;
			uint32		progVerLow, progVerHigh;	

			//Check verifier
			//currently only accept AUTH_NULL type
			authType = *rpcReplyHdrp++;
			authLen = *rpcReplyHdrp++;
			rpcReplyHdrLen += 8;

			if ( (authType != RPC_AUTH_NONE) ||
				(authLen != 0) )
			{
				//Error: not AUTH_NULL authentication
				rpc_globalCfgp->stats.rxReplyAuthErr++;	
				return (NFX_ERR);
			}

			//Get RPC accept status
			acceptStatus =  *rpcReplyHdrp++;
			rpcReplyHdrLen += 4;

			switch (acceptStatus)
			{
				case RPC_REPLY_SUCCESS:
					rpc_globalCfgp->stats.rxReplySuccess++;
					return (rpcReplyHdrLen);
				case RPC_REPLY_PROG_MISMATCH:
					progVerLow = *rpcReplyHdrp++;
					progVerHigh = *rpcReplyHdrp++;
					rpcReplyHdrLen += 8;
					rpc_globalCfgp->stats.rxReplyProgMismatchErr++;
					return (NFX_ERR);
				case RPC_REPLY_PROC_UNAVAIL:
					rpc_globalCfgp->stats.rxReplyProcUnavailErr++;
					return (NFX_ERR);
				case RPC_REPLY_GARBAGE_ARGS:
					rpc_globalCfgp->stats.rxReplyGarArgErr++;	
					return (NFX_ERR);
				case RPC_REPLY_SYSTEM_ERR:
					rpc_globalCfgp->stats.rxReplySysErr++;
					return (NFX_ERR);
				default: 
					rpc_globalCfgp->stats.rxReplyGenErr++;
					return (NFX_ERR);
			}
			break;
		}			
		case RPC_REPLY_MSG_DENIED:
		{
			uint32		rejectedReply;

			rejectedReply = *rpcReplyHdrp++;
			rpcReplyHdrLen += 4;
			switch (rejectedReply)
			{
				uint32		rpcVerLow, rpcVerHigh;
				uint32		authErrType;

				case RPC_REPLY_RPC_MISMATCH:
					rpc_globalCfgp->stats.rxReplyRpcMismatchErr++;
					rpcVerLow = *rpcReplyHdrp++;
					rpcVerHigh = *rpcReplyHdrp++;
					rpcReplyHdrLen += 8;
					return (NFX_ERR);	
				case RPC_REPLY_AUTH_ERROR:
					authErrType = *rpcReplyHdrp++;
					rpcReplyHdrLen += 4;
					if ( (authErrType < RPC_AUTH_MIN_ERR) ||
						(authErrType > PRC_AUTH_MAX_ERR) )
						authErrType = RPC_AUTH_FAILED;
					rpc_globalCfgp->stats.rxAuthErrType[authErrType] += 1;
					return (NFX_ERR);	
				default:
					rpc_globalCfgp->stats.rxReplyGenErr++;
			}
			return (NFX_ERR);

		}

		default:
			rpc_globalCfgp->stats.rxInvalidReplyStatErr++;
			return (NFX_ERR);	
	}

	return (NFX_ERR);	
}


/*-----------------------------------------------------------------
 * Name:   	rpc_createHdrInfo 
 *
 * Description:
 *  Create RPC header information.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
rpcCallHdr_t *rpc_createHdrInfo(rpcCallHdr_t  *pktRpcCallHdrp)
{
	rpcCallHdr_t	*newRpcCallp;
	uint32			length;
//	uint32			fieldLen;

	if (pktRpcCallHdrp == NULL)
		return (NULL);
	length = 0;
	newRpcCallp = (rpcCallHdr_t *) eee_ramAlloc(sizeof(rpcCallHdr_t));
	if (newRpcCallp == NULL)
		return (NULL);

	newRpcCallp->rpc_xid = pktRpcCallHdrp->rpc_xid;
	newRpcCallp->rpc_direction = pktRpcCallHdrp->rpc_direction;
	newRpcCallp->rpc_rpcVer = pktRpcCallHdrp->rpc_rpcVer;
	newRpcCallp->rpc_prog = pktRpcCallHdrp->rpc_prog;
	newRpcCallp->progVer = pktRpcCallHdrp->progVer;
	newRpcCallp->rpc_proc = pktRpcCallHdrp->rpc_proc;

	memcpy(&newRpcCallp->authFlavor.authSys, &pktRpcCallHdrp->authFlavor.authSys, sizeof(rpc_authSysInfo_t));

	return (newRpcCallp);

}


/*-----------------------------------------------------------------
 * Name:   rpc_displayStats	
 *
 * Description:
 *  Display RPC stats.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/02
 *
 *-----------------------------------------------------------------
 */

void
rpc_displayStats(void)
{

	printf("RPC Error counters\n");
	printf("----------------------------\n");
	printf("RX RPC Reply OK:                  %lld\n", rpc_globalCfgp->stats.rxReplySuccess);
	printf("RX RPC Reply AUTH ERR:            %lld\n", rpc_globalCfgp->stats.rxReplyAuthErr);
	printf("RX RPC Reply PROG MISMATH ERR:    %lld\n", rpc_globalCfgp->stats.rxReplyProgMismatchErr);
	printf("RX RPC Reply PROC UNAVAIL ERR:    %lld\n", rpc_globalCfgp->stats.rxReplyProcUnavailErr);
	printf("RX RPC Reply GARB ARG ERR:        %lld\n", rpc_globalCfgp->stats.rxReplyGarArgErr); 
	printf("RX RPC Reply System ERR:          %lld\n", rpc_globalCfgp->stats.rxReplySysErr); 
	printf("RX RPC Reply Generic ERR:         %lld\n", rpc_globalCfgp->stats.rxReplyGenErr); 
	printf("RX RPC Reply RPC mismatch ERR:    %lld\n", rpc_globalCfgp->stats.rxReplyRpcMismatchErr); 
	printf("RX RPC Invalid Reply status ERR:  %lld\n", rpc_globalCfgp->stats.rxInvalidReplyStatErr); 
	printf("RX RPC Reply AUTH OK:             %lld\n", rpc_globalCfgp->stats.rxAuthErrType[RPC_AUTH_OK]); 
	printf("RX RPC Reply AUTH Bad Credential: %lld\n", rpc_globalCfgp->stats.rxAuthErrType[RPC_AUTH_BADCRED]); 
	printf("RX RPC Reply AUTH Rej Credential: %lld\n", rpc_globalCfgp->stats.rxAuthErrType[RPC_AUTH_REJECTEDCRED]); 
	printf("RX RPC Reply AUTH Bad Verifier:   %lld\n", rpc_globalCfgp->stats.rxAuthErrType[RPC_AUTH_BADVERF]); 
	printf("RX RPC Reply AUTH Rej Verifer:    %lld\n", rpc_globalCfgp->stats.rxAuthErrType[RPC_AUTH_REJECTEDVERF]); 
	printf("RX RPC Reply AUTH Too Weak:       %lld\n", rpc_globalCfgp->stats.rxAuthErrType[RPC_AUTH_TOOWEAK]); 
	printf("RX RPC Reply AUTH Invalid Resp:   %lld\n", rpc_globalCfgp->stats.rxAuthErrType[RPC_AUTH_INVALIDRESP]); 
	printf("RX RPC Reply AUTH Failed:         %lld\n", rpc_globalCfgp->stats.rxAuthErrType[RPC_AUTH_FAILED]); 
	printf("RX RPC Reply AUTH KERB Generic:   %lld\n", rpc_globalCfgp->stats.rxAuthErrType[RPC_AUTH_KERB_GENERIC]); 
	printf("RX RPC Reply AUTH Timer Expire:   %lld\n", rpc_globalCfgp->stats.rxAuthErrType[RPC_AUTH_TIMEEXPIRE]); 
	printf("RX RPC Reply AUTH TKT File:       %lld\n", rpc_globalCfgp->stats.rxAuthErrType[RPC_AUTH_TKT_FILE]); 
	printf("RX RPC Reply AUTH Decode:         %lld\n", rpc_globalCfgp->stats.rxAuthErrType[RPC_AUTH_DECODE]); 
	printf("RX RPC Reply AUTH Net Addr:       %lld\n", rpc_globalCfgp->stats.rxAuthErrType[RPC_AUTH_NET_ADDR]); 
	printf("RX RPCSEC Reply GSS CRED Problem: %lld\n", rpc_globalCfgp->stats.rxAuthErrType[RPC_RPCSEC_GSS_CREDPROBLEM]); 
	printf("RX RPCSEC Reply GSS CTX Problem:  %lld\n", rpc_globalCfgp->stats.rxAuthErrType[RPC_RPCSEC_GSS_CTXPROBLEM]); 

	return;
}

/*-----------------------------------------------------------------
 * Name:   rpc_init	
 *
 * Description:
 *  Initialize RPC. 
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        7/28/02
 *
 *-----------------------------------------------------------------
 */

int32
rpc_init(void)
{
	printf("rpc_init\n");
    //Alloc & Initialize NLM global Config
    rpc_globalCfgp =
            (rpc_globalCfg_t *)eee_ramAlloc(sizeof(rpc_globalCfg_t) );

    if (rpc_globalCfgp == NULL)
        panic ("%s: Failure to alloc RPC global Cfg!\n",__FUNCTION__);


    bzero(rpc_globalCfgp, sizeof(rpc_globalCfg_t));

	return (NFX_OK);
}
