/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        ndmp-config.c
 *
 * RCS:  $Id: ndmp-config.c,v 1.22 2002/08/28 23:40:18 anhv Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	Process all NDMP_CONFIG_xxx commands sent by the
 *				NDMP client.  These commands are ways for the client
 *				to learn what we as a server support in terms of 
 *				tape devices, tape libraries, what type of FS we
 *				have, what o/s we're running on, etc.
 *
 * Created by: Anh Vu
 *
 * Date Created:        9/06/01
 *
 *-----------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/utsname.h>
#include <unistd.h>
#include <string.h>
#include <nfx-types.h>
#include <exports/eee-api.h>
#include <exports/elog-api.h>
#include <exports/anpssc-api.h>
#include <exports/eee-api.h>
#include <exports/scsi-api.h>
#include <exports/sdm-api.h>
#include "ndmp.h"
#include "ndmp-ssc.h"
#include "ndmp-debug.h"
#include <nfx-incl.h>
#include <exports/tape-api.h>

char    tbuf[256];

/*-----------------------------------------------------------------
 * Name: ndmp_procCfgGetHostInfoReq 
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procCfgGetHostInfoReq(ndmpSess_t *sess)
{
	struct utsname 				buf;
	ndmp_cfg_get_host_info_rsp 	rsp;
	

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procCfgGetHostInfoReq: Begin.\n"));

	/* Init response */
	memset(&rsp, 0, sizeof(rsp));
	rsp.error = NDMP_NO_ERR;

	if (uname(&buf)) 
	{
        E_LOG(class_1, err_s, eee_getApplicationId("ndmp"), 0, 0, 0, 
			("Error processing CFG_GET_HOST_INFO req.\n"));
		rsp.error = NDMP_UNDEFINED_ERR;
	} else {	

		rsp.hostname = buf.nodename;
		rsp.os_type = buf.sysname;
		rsp.os_vers = buf.version;	/* right now, the field buf.release has
									 * "2.2.17-8" for linux sim, don't know 
									 * what'll be later, version, however,
									 * contains "#1 Fri Nov 17 16:2225 EST 2000"
									 * I know, strange, but this is what uname
									 * yield, we can put 'release' for os_vers
									 * in the response later if that makes more sense
									 */
		sprintf(tbuf, "%lx", gethostid());
		rsp.hostid = tbuf;
	}

   	return (ndmp_rspClient(sess, &rsp));

} /* ndmp_procCfgGetHostInfoReq */


/*-----------------------------------------------------------------
 * Name: ndmp_procCfgGetSrvrInfoReq 
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procCfgGetSrvrInfoReq(ndmpSess_t *sess)
{
	ndmp_cfg_get_srvr_info_rsp 	rsp;
	ndmp_AuthTyp 				auth_type_supported[3];	

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procCfgGetSrvrInfoReq: Begin.\n"));
	/* Init response */
	memset(&rsp, 0, sizeof(rsp));
	rsp.error = NDMP_NO_ERR;

	rsp.vendorName 	= AGILE_VENDOR_NAME;
	rsp.prodName 	= AGILE_PRODUCT_NAME;
	rsp.revNum 		= AGILE_PROD_VERSION;
	auth_type_supported[0] = NDMP_AUTHNONE;
	auth_type_supported[1] = NDMP_AUTHTXT;
	auth_type_supported[2] = NDMP_AUTHMD5;
	rsp.authTyp.authTyp_len = 3;
	rsp.authTyp.authTyp_val = auth_type_supported;

   	return (ndmp_rspClient(sess, &rsp));

} /* ndmp_procCfgGetSrvrInfoReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procCfgGetConnTypeReq 
 *
 * Description:		Client wants to know a list of data connection
 *				types we support.  Give it to him.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procCfgGetConnTypeReq(ndmpSess_t *sess)
{
	ndmp_cfg_get_conn_type_rsp rsp;
	//ndmp_AddrTyp	addr_typ_supported[2];	
	ndmp_AddrTyp	addr_typ_supported[1];	

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procCfgGetConnTypeReq: Begin.\n"));

	/* Init response */
	memset(&rsp, 0, sizeof(rsp));
	rsp.error = NDMP_NO_ERR;

	addr_typ_supported[0] = NDMP_ADDR_LOCAL;
	//addr_typ_supported[1] = NDMP_ADDR_TCP;
	//rsp.addrTypes.addrTypes_len = 2;
	rsp.addrTypes.addrTypes_len = 1;
	rsp.addrTypes.addrTypes_val = addr_typ_supported;

   	return (ndmp_rspClient(sess, &rsp));

} /* ndmp_procCfgGetConnTypeReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procCfgGetAuthAttrReq 
 *
 * Description:		Client tells us what type of authentication 
 *				he will use and we reply w/ the attribute for that
 * 				authentication type.  The only attribute exists til
 *				this day for authentication is the challenge string
 *				for MD5 auth type.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procCfgGetAuthAttrReq(ndmpSess_t *sess)
{
	ndmp_cfg_get_auth_attr_req *req;
	ndmp_cfg_get_auth_attr_rsp rsp;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procCfgGetAuthAttrReq: Begin.\n"));
	/* get request */
	req = (ndmp_cfg_get_auth_attr_req *) sess->msgRcvd.content;
	
	/* Init response */
	memset(&rsp, 0, sizeof(rsp));
	rsp.error = NDMP_NO_ERR;

	if ((req->authTyp != NDMP_AUTHNONE) && (req->authTyp != NDMP_AUTHTXT) && 
			(req->authTyp != NDMP_AUTHMD5)) 
	{
        E_LOG(class_1, err_s, eee_getApplicationId("ndmp"), 0, 0, 0, 
			("Process CONFIG_GET_AUTH_ATTR Req:  Auth type %d not supported.", req->authTyp));
		rsp.error = NDMP_ILLEGAL_ARGS_ERR;

	} else if (req->authTyp == NDMP_AUTHMD5)
	{
		rsp.srvrAttr.authTyp = req->authTyp;

		/* Challenge string is derived from a random number */
		getChallengStr(rsp.srvrAttr.ndmp_AuthAttr_u.challenge);
		memcpy(sess->challenge, rsp.srvrAttr.ndmp_AuthAttr_u.challenge, NDMP_CHALLENGE_LEN);
	}
   	return (ndmp_rspClient(sess, &rsp));

} /* ndmp_procCfgGetAuthAttrReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procCfgGetButypeInfoReq 
 *
 * Description:		
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procCfgGetButypeInfoReq(ndmpSess_t *sess)
{
	ndmp_butype_info			bu_info;	
	ndmp_cfg_get_butype_info_rsp 	cl_rsp;
	ndmp_pval					env_lst[NDMP_CFG_BUTYP_NUM_ENV];

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procCfgGetButypeInfoReq: Begin.\n"));

	/* Init client's response */
	memset(&cl_rsp, 0, sizeof(cl_rsp));
	memset(&bu_info, 0, sizeof(ndmp_butype_info));
	cl_rsp.error = NDMP_NO_ERR;
	
	/* We only support one type Backup type at the moment */
	bu_info.butype_name = NDMP_EFX_BKUP_TYP_NAME;
	bu_info.defEnv.defEnv_len = NDMP_CFG_BUTYP_NUM_ENV;
	bu_info.defEnv.defEnv_val = env_lst;
	bu_info.attrs = (NDMP_BUTYPE_BACKUP_FILE_HISTORY |
					NDMP_BUTYPE_RECOVER_FILELIST |
					NDMP_BUTYPE_BACKUP_INCREMENTAL |
					NDMP_BUTYPE_RECOVER_INCREMENTAL |
					NDMP_BUTYPE_BACKUP_UTF8 |
					NDMP_BUTYPE_RECOVER_UTF8);
	 
	memset(env_lst, 0, NDMP_CFG_BUTYP_NUM_ENV * sizeof(ndmp_pval));

	/* We advertise the defaults for these environment variables */
	env_lst[0].name = "FILESYSTEM";			// file system to bkup/recover
	env_lst[0].value = "";
	env_lst[1].name = "DIRECT";				// direct access not supported
	env_lst[1].value = "n";
	env_lst[2].name = "RECURSIVE";			// processing subdirs
	env_lst[2].value = "y";
	env_lst[3].name = "TYPE";				// bkup type
	env_lst[3].value = NDMP_EFX_BKUP_TYP_NAME;
	env_lst[4].name = "HIST";				// type of file history generated
	env_lst[4].value = "f";
	env_lst[5].name = "PATHNAME_SEPARATOR";
	env_lst[5].value = "/";
	env_lst[6].name = "LEVEL";				// dump level
	env_lst[6].value = "0-9";
	env_lst[7].name = "EXTRACT";			// y means -x option is used
											// for extraction
	env_lst[7].value = "y";
	env_lst[8].name = "UPDATE";				// update dumpdate file
	env_lst[8].value = "y";
	env_lst[9].name = "FILES";				// list of files to backup
	env_lst[9].value = "";

#if 0
	env_lst[10].name = "PREFIX";			// path of where to backup
	env_lst[10].value = "";
	env_lst[11].name = "CWD";				// same as prefix
	env_lst[11].value = "";
#endif
	cl_rsp.error = NDMP_NO_ERR;
	cl_rsp.butype_info.butype_info_len = 1;
	cl_rsp.butype_info.butype_info_val = &bu_info;

   	return (ndmp_rspClient(sess, &cl_rsp));

} /* ndmp_procCfgGetButypeInfoReq */

/*-----------------------------------------------------------------
 * Name:	ndmp_sendCfgFSInfoRsp 
 *
 * Description:	Form the Cfg FS Info Response to NDMP client
 *				from the info we've collected from the FS
 *				for all of the FS in the box.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/29/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_sendCfgFSInfoRsp(ndmpSess_t *sess)
{                                      
	ndmp_cfg_get_fs_info_rsp	cl_rsp;
	ndmp_fs_info				*fs_info_lst = NULL;	// list of NDMP fs info to be sent to client
	fs_Stat						*fs_lst = NULL;			// fs info we've collected from FS
	int 						i, num_fs = sess->tot_numFs, error;

	/* Init our response to NDMP client */
	memset(&cl_rsp, 0, sizeof(cl_rsp));
	cl_rsp.error = NDMP_NO_ERR;
	error = NFX_OK;

	// prepare list we'll send
    NDMP_SSC_ALLOC(fs_info_lst, ndmp_fs_info, (num_fs*sizeof (ndmp_fs_info)));
    NFX_ASSERT(fs_info_lst != NULL);

	memset(fs_info_lst, 0, (num_fs * sizeof (ndmp_fs_info)));
	fs_lst = sess->fsInfoLst;			// list we've accumulated 
	cl_rsp.fs_info.fs_info_len= num_fs;
	cl_rsp.fs_info.fs_info_val= fs_info_lst;
	for (i = 0; i < num_fs; i++)
	{
		fs_info_lst[i].invalid = 0;	
		fs_info_lst[i].fs_type = NDMP_EFX_FS_TYPE_NAME;
        NDMP_SSC_ALLOC(fs_info_lst[i].fs_logical_device, char, NDMP_EFX_FS_LOG_DEV_LEN);
        if (fs_info_lst[i].fs_logical_device == NULL)
        {
            NDMP_DBG_PRINTF(D_INFO, ("ndmp_sendCfgFSInfoRsp:  Malloc fs_logical_dev error\n"));
			error = NFX_ERR;
			goto errhdling;
        }
		// This will be the Volume Id used as FS name
		fs_info_lst[i].fs_logical_device = sess->vol_lst[i].name;
		fs_info_lst[i].fs_physical_device = "";	
		fs_info_lst[i].total_size.low = (fs_lst[i].fsBlockSize * fs_lst[i].numBlks); 
		fs_info_lst[i].total_size.high = (fs_lst[i].fsBlockSize * fs_lst[i].numBlks) >> 32; 
		fs_info_lst[i].used_size.low = 
						(fs_lst[i].fsBlockSize * (fs_lst[i].numBlks - fs_lst[i].numBlksAvail));	
		fs_info_lst[i].used_size.high = 
						(fs_lst[i].fsBlockSize * (fs_lst[i].numBlks - fs_lst[i].numBlksAvail)) >> 32;	
		fs_info_lst[i].avail_size.low = fs_lst[i].fsBlockSize * fs_lst[i].numBlksAvail;	
		fs_info_lst[i].avail_size.high = (fs_lst[i].fsBlockSize * fs_lst[i].numBlksAvail) >> 32;	
		fs_info_lst[i].total_inodes.low = fs_lst[i].numInodes;	
		fs_info_lst[i].total_inodes.high = fs_lst[i].numInodes >> 32;	
		fs_info_lst[i].used_inodes.low = fs_lst[i].numInodes - fs_lst[i].numInodesAvail;	
		fs_info_lst[i].used_inodes.high = (fs_lst[i].numInodes - fs_lst[i].numInodesAvail) >> 32;	
		fs_info_lst[i].fs_status = NDMP_EFX_FS_TYPE_STATUS;
		fs_info_lst[i].fs_env.fs_env_len = NDMP_NUM_FS_ENV;	
        NDMP_SSC_ALLOC(fs_info_lst[i].fs_env.fs_env_val, ndmp_pval,
                                                         (NDMP_NUM_FS_ENV * sizeof(ndmp_pval)));
        if (fs_info_lst[i].fs_env.fs_env_val == NULL)
		{
            NDMP_DBG_PRINTF(D_INFO, ("ndmp_sendCfgFSInfoRsp:  Malloc fs_logical_dev error\n"));
			error = NFX_ERR;
			goto errhdling;
		}		
		memset(fs_info_lst[i].fs_env.fs_env_val, 0, NDMP_NUM_FS_ENV * sizeof(ndmp_pval));
		fs_info_lst[i].fs_env.fs_env_val[0].name = "LOCAL";
		fs_info_lst[i].fs_env.fs_env_val[0].value = "y";
		fs_info_lst[i].fs_env.fs_env_val[1].name = "TYPE";
		fs_info_lst[i].fs_env.fs_env_val[1].value = NDMP_EFX_FS_TYPE_NAME;
		fs_info_lst[i].fs_env.fs_env_val[2].name = "AVAILABLE_BACKUP";
		fs_info_lst[i].fs_env.fs_env_val[2].value = NDMP_EFX_BKUP_TYP_NAME;	/* "efx_dump" */
		fs_info_lst[i].fs_env.fs_env_val[3].name = "AVAILABLE_RECOVERY";
		fs_info_lst[i].fs_env.fs_env_val[3].value = NDMP_EFX_BKUP_TYP_NAME;
	}
	/* Send Client a respond w/ FS stats */
   	error = (ndmp_rspClient(sess, &cl_rsp));

errhdling:
	// Free the list that we've been preparing to send
	if (fs_info_lst)
	{
		for (i = 0; i < num_fs; i++)
		{
			if (fs_info_lst[i].fs_env.fs_env_val)
                NDMP_SSC_FREE((fs_info_lst[i].fs_env.fs_env_val));
		}
        NDMP_SSC_FREE(fs_info_lst);
	}

	// Free the accumulated FS list acquired straight from FS
	if (sess->fsInfoLst)
        NDMP_SSC_FREE(sess->fsInfoLst);

	if (sess->vol_lst)
        NDMP_SSC_FREE(sess->vol_lst);

	// If there's error w/ mallocing
	if (error == NFX_ERR)
	{
		// Send error to NDMP Client
		cl_rsp.error =  NDMP_UNDEFINED_ERR;
   		ndmp_rspClient(sess, &cl_rsp);
	}
    return error;  
} /* ndmp_sendCfgFSInfoRsp */

/*-----------------------------------------------------------------
 * Name:    ndmp_procFSInfoRsp
 *
 * Description: Got a response to our FS INFO request.
 *              Gather its info into our list of FS info.
 *				If not done w/ our list of FS, send another request
 *				for the next FS volId.
 *				When done, we'll form a response to Client
 *				for his CfgFSInfo Request.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/29/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procFSInfoRsp(ndmpSess_t *sess, char8 *rcvbuf)
{
	fs_MsgStatVolResp *fs_rsp = (fs_MsgStatVolResp *) rcvbuf;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procFSInfoRsp: Begin.\n"));

	sess->ndmp_ssc_chld_msgStats[FS_REQ_FSSTAT_VOL_RSP_RCVD]++;
	// Collect the info and store it into our sess memory.
	memcpy(&sess->fsInfoLst[sess->nextFsIdx-1], &fs_rsp->stat, sizeof(fs_Stat));	
	
	NDMP_DBG_PRINTF(D_INFO, ("VolId %llx\n", sess->vol_lst[sess->nextFsIdx-1].id.whole_thing));
	NDMP_DBG_PRINTF(D_INFO, ("fsBlockSize %d\n", fs_rsp->stat.fsBlockSize));
	NDMP_DBG_PRINTF(D_INFO, ("numBlks %lld\n", fs_rsp->stat.numBlks));
	NDMP_DBG_PRINTF(D_INFO, ("numBlksAvail %lld\n", fs_rsp->stat.numBlksAvail));
	NDMP_DBG_PRINTF(D_INFO, ("numBlksAvailUser %lld\n", fs_rsp->stat.numBlksAvailUser));
	NDMP_DBG_PRINTF(D_INFO, ("numInodes %lld\n", fs_rsp->stat.numInodes));
	NDMP_DBG_PRINTF(D_INFO, ("numInodesAvail %lld\n", fs_rsp->stat.numInodesAvail));
	NDMP_DBG_PRINTF(D_INFO, ("numInodesAvailUser %lld\n", fs_rsp->stat.numInodesAvailUser));
	NDMP_DBG_PRINTF(D_INFO, ("readsPerSec %d\n", fs_rsp->stat.readsPerSec));
	NDMP_DBG_PRINTF(D_INFO, ("writesPerSec %d\n", fs_rsp->stat.writesPerSec));
	NDMP_DBG_PRINTF(D_INFO, ("avgReadSize %d\n", fs_rsp->stat.avgReadSize));
	NDMP_DBG_PRINTF(D_INFO, ("avgWriteSize %d\n", fs_rsp->stat.avgWriteSize));

 	return NFX_OK;

} /* ndmp_procFSInfoRsp */

/*-----------------------------------------------------------------
 * Name:    ndmp_sendFSInfoReq
 * 
 * Description: Send a FS INFO request using the VolId in our Vol list.
 *				If there's an error send the request, we'll send
 *				an error response to NDMP client.
 * 
 * Created by:  Anh Vu
 * 
 * Date Created:        10/29/01
 *
 *-----------------------------------------------------------------
 */
int
ndmp_sendFSInfoReq(ndmpSess_t *sess)
{
    fs_MsgStatVolReq            fs_req;
    ndmp_cfg_get_fs_info_rsp    cl_rsp;
	int	i;
	struct timeval to;
	char	respbuff[2000];

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_sendFSInfoReq: Begin.\n"));

	 
	for (i=0; i < sess->tot_numFs; i++)
	{
		bzero(respbuff, sizeof(respbuff));                                           
    	// Send the next FS Info request to FS
    	fs_req.hdr.requestType = FS_REQ_FSSTAT_VOL;
    	fs_req.hdr.len = sizeof(fs_MsgStatVolReq);
    	fs_req.hdr.clientPrivate = (uint64) sess->sessId;
    	fs_req.volId.whole_thing = sess->vol_lst[sess->nextFsIdx].id.whole_thing;
   
		to.tv_sec = 10;
		to.tv_usec = 0;

		sess->ndmp_ssc_chld_msgStats[FS_REQ_FSSTAT_VOL_REQ_SNT]++;
    	if (sendAgileMsg (-1, &fs_req, sizeof(fs_MsgStatVolReq), 
                             "efs", NFX_NFP_FP2, 1, MGMTPLANE,  // 1 or NIM_FP?
                             &to,  respbuff, sizeof(respbuff)) <=0)
    	{
        	NDMP_DBG_PRINTF(D_INFO, ("ndmp_sendFSInfoReq:  Agile message sent for FS_REQ_FSSTAT_VOL failed\n"));

        	// Send error response to NDMP Client
        	cl_rsp.error =  NDMP_UNDEFINED_ERR;
    		return (ndmp_rspClient(sess, &cl_rsp));
    	}
    	sess->nextFsIdx ++; 
		ndmp_procFSInfoRsp(sess, respbuff);
	}

	return NFX_OK;
} /* ndmp_sendFSInfoReq */


/*-----------------------------------------------------------------
 * Name:    ndmp_procVolLstRsp
 *
 * Description: Now that we have a list of FS (volume ids)
 *              from EVM, send a request to FS
 *              on NNIM to collect status of the first FS
 *              given the volume id from the vol_lst returned
 *              from EVM.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/29/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procVolLstRsp(ndmpSess_t *sess, char8 *rcvbuf)
{
	int i;
   	evm_MsgVolNameList_t *vol_lst = (evm_MsgVolNameList_t *)rcvbuf; 

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procVolLstRsp: EVM reports %d volumes.\n", vol_lst->num));

	sess->ndmp_ssc_chld_msgStats[EVM_VOL_GET_NAME_LST_RSP_RCVD]++;

    /* Allocate space for all of FS info in our system,
	 * will free when we're done replying FS status
	 */
    NDMP_SSC_ALLOC(sess->fsInfoLst, fs_Stat, (vol_lst->num * sizeof(fs_Stat)));
    NFX_ASSERT(sess->fsInfoLst != NULL);

    sess->nextFsIdx = 0;
    sess->tot_numFs = sess->curr_numFs = vol_lst->num;

    NDMP_SSC_ALLOC(sess->vol_lst, vol_entry_t, (vol_lst->num * sizeof(vol_entry_t)));
    NFX_ASSERT(sess->vol_lst != NULL);

	memcpy( sess->vol_lst, vol_lst->s, (vol_lst->num * sizeof(evm_volId_t)));

	for (i = 0; i < vol_lst->num; i++)
	{
		sess->vol_lst[i].id.whole_thing = vol_lst->s[i].id.whole_thing;
		strcpy(sess->vol_lst[i].name, vol_lst->s[i].name);
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_procVolLstRsp: VolId[%d] = 0x%llx.\n", i+1, vol_lst->s[i].id.whole_thing));
	}
    return NFX_OK;

} /* ndmp_procVolLstRsp */

/*-----------------------------------------------------------------
 * Name: ndmp_procCfgGetFsInfoReq 
 *
 * Description:		
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procCfgGetFsInfoReq(ndmpSess_t *sess)
{
	evm_MsgVolGetNameList_t 	vollst_req;
	ndmp_cfg_get_fs_info_rsp    cl_rsp;
	char	respbuff[1024];
	struct timeval	to;
	

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procCfgGetFsInfoReq: Begin.\n"));

	/* Get a list of volume ids and their names */
    memset (&vollst_req, 0, sizeof(evm_MsgVolGetNameList_t));
    vollst_req.hdr.msgType = EVM_VOL_GET_NAME_LIST;
    vollst_req.hdr.clientPrivate = (int64) sess->sessId;
 
    to.tv_sec = 5;
    to.tv_usec = 0;     	 

	sess->ndmp_ssc_chld_msgStats[EVM_VOL_GET_NAME_LST_REQ_SNT]++;
    if (sendAgileMsg (-1, &vollst_req, sizeof(evm_MsgVolGetNameList_t),
                             "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                             &to, respbuff, sizeof(respbuff)) <= 0)
	{
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_procCfgGetFsInfoReq:  Agile message sent to EVM_VOL_GET_NAME_LIST failed\n"));

		// Send error to NDMP Client
		cl_rsp.error =  NDMP_UNDEFINED_ERR;
    	if (ndmpXmtMsg(sess, NDMP_MSG_RSP, sess->msgRcvd.hdr.msgId, 
				sess->msgRcvd.hdr.xmtSeq, NDMP_NO_ERR, (void *)&cl_rsp) < 0)
    	{
        	E_LOG(class_1, err_s, eee_getApplicationId("ndmp"), 0, 0, 0, 
						("Error sending CFG_GET_FS_INFO rsp."));
		}
        return NFX_ERR; 
	}
	/* Get a list of volumes from EVM */
	ndmp_procVolLstRsp(sess, respbuff);

	/* Query FS stats for each of those volumes reported by EVM */
	ndmp_sendFSInfoReq(sess);

	/* Formulate & send reply to NDMP client */
	ndmp_sendCfgFSInfoRsp(sess);

	return NFX_OK;

} /* ndmp_procCfgGetFsInfoReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procCfgGetTapeInfoReq 
 *
 * Description:		
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procCfgGetTapeInfoReq(ndmpSess_t *sess)
{
	tape_get_devlist_req_msg_t	tape_req;
	ndmp_cfg_get_tape_info_rsp	cl_rsp;
	char	respbuff[8192];
	struct timeval	to;

	
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procCfgGetTapeInfoReq: Begin.\n"));
	

	/* Ask for a list of devices from Tape-driver */
	memset(&tape_req, 0, sizeof(tape_req));
	tape_req.hdr.mid = TAPE_MID_GET_DEVLIST_REQ;

	to.tv_sec = 5;
    to.tv_usec = 0;

	/* Query tape driver */
	if (sendAgileMsg(-1, &tape_req, sizeof(tape_req),
                        "tape-driver", 0, 0, 0,  /* to which applic */
                        &to, respbuff, sizeof(respbuff)) <= 0)
    {
        NDMP_DBG_PRINTF(D_INFO, ("Agile message send failed\n"));
        if (ndmpXmtMsg(sess, NDMP_MSG_RSP, sess->msgRcvd.hdr.msgId, 
				sess->msgRcvd.hdr.xmtSeq, NDMP_NO_ERR, (void *) &cl_rsp) < 0)
        {
            E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
                    ("ndMsgtoTapeDrv:  Error sending rsp to client, msgid %d.\n", sess->msgRcvd.hdr.msgId));
 
        }
        return NFX_ERR;
    } 
	sess->ndmp_ssc_chld_msgStats[TAPE_GET_DEVLST_REQ_SNT]++;
	ndmp_procTapeGetDevLstRsp(sess, respbuff) ;	
		
#if 0 //Treat it serially
	/* Query tape driver */
    return (sndMsgtoTapeDrv(sess, &tape_req, sizeof(tape_req), &cl_rsp));
#endif 

	return NFX_OK;

}   /* ndmp_procCfgGetTapeInfoReq */ 

/*-----------------------------------------------------------------
 * Name: ndmp_procCfgGetScsiInfoReq
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procCfgGetScsiInfoReq(ndmpSess_t *sess)
{
    mc_get_devlist_req_msg_t  	mc_req;
    ndmp_cfg_get_scsi_info_rsp  cl_rsp;
   
    NDMP_DBG_PRINTF(D_INFO, ("ndmp_procCfgGetScsiInfoReq: Begin.\n"));
   

    /* Ask for a list of devices from Tape-driver */
    memset(&mc_req, 0, sizeof(mc_req));
    mc_req.hdr.mid = MC_MID_GET_DEVLIST_REQ;

    /* Query tape driver */
	sess->ndmp_ssc_chld_msgStats[MC_GET_DEVLST_REQ_SNT]++;
    return (sndMsgtoTapeDrv(sess, &mc_req, sizeof(mc_req), &cl_rsp));

}   /* ndmp_procCfgGetScsiInfoReq */

/*=================================================================
 * 
 * 				V2 handling functions
 *
 *=================================================================
 */
/*-----------------------------------------------------------------
 * Name: ndmpv2_procCfgGetHostInfoReq 
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int
ndmpv2_procCfgGetHostInfoReq(ndmpSess_t *sess)
{
	struct utsname		buf;
	ndmpv2_cfg_get_host_info_rsp 	rsp;
	ndmp_AuthTyp		auth_type_supported[3];
	

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procCfgGetHostInfoReq: Begin.\n"));

	/* Init response */
	memset(&rsp, 0, sizeof(rsp));
	rsp.error = NDMP_NO_ERR;

	if (uname(&buf)) 
	{
        E_LOG(class_1, err_s, eee_getApplicationId("ndmp"), 0, 0, 0, 
			("Error processing CFG_GET_HOST_INFO req.\n"));
		rsp.error = NDMP_UNDEFINED_ERR;
	} else {	

		rsp.hostname = buf.nodename;
		rsp.os_type = buf.sysname;
		rsp.os_vers = buf.version;	/* right now, the field buf.release has
									 * "2.2.17-8" for linux sim, don't know 
									 * what'll be later, version, however,
									 * contains "#1 Fri Nov 17 16:2225 EST 2000"
									 * I know, strange, but this is what uname
									 * yield, we can put 'release' for os_vers
									 * in the response later if that makes more sense
									 */
		sprintf(tbuf, "%lx", gethostid());
		rsp.hostid = tbuf;

		/* Fill in auth type */
		auth_type_supported[0] = NDMP_AUTHNONE;
		auth_type_supported[1] = NDMP_AUTHTXT;
		auth_type_supported[2] = NDMP_AUTHMD5;

		rsp.authTyp.authTyp_len = 3;
		rsp.authTyp.authTyp_val = auth_type_supported;
	}

   	return (ndmp_rspClient(sess, &rsp));

} /* ndmpv2_procCfgGetHostInfoReq */

/*-----------------------------------------------------------------
 * Name: ndmpv2_procCfgGetButypeAttrReq 
 *
 * Description:		
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int
ndmpv2_procCfgGetButypeAttrReq(ndmpSess_t *sess)
{
	ndmpv2_cfg_get_butype_attr_req 	*req = 
		(ndmpv2_cfg_get_butype_attr_req *) sess->msgRcvd.content;
	ndmpv2_cfg_get_butype_attr_rsp 	rsp;

	NDMP_DBG_PRINTF(D_INFO, ("ndmpv2_procCfgGetButypeAttrReq: Begin.\n"));

	/* Init client's response */
	bzero(&rsp, sizeof(rsp));

	/* we only support 'dump' */
	if ( ! (!strcmp(req->butype_name, "dump")) ||
			(!strcmp(req->butype_name, "DUMP")) )
		rsp.error = NDMP_ILLEGAL_ARGS_ERR;
	else 
		rsp.error = NDMP_NO_ERR;
	
	/* We only support one type Backup type at the moment */
	rsp.attrs = ( NDMP_BUTYPE_NO_BACKUP_FILELIST |
					NDMP_BUTYPE_NO_RECOVER_FHINFO |
					NDMP_BUTYPE_NO_RECOVER_INC_ONLY);
	 

   	return (ndmp_rspClient(sess, &rsp));

} /* ndmpv2_procCfgGetButypeAttrReq */
