/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        ndmp-info.c
 *
 * RCS:  $Id: 
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:		Send messages such as logs, file histories, and
 *					notifications to the NDMP client.  We are relaying
 *					these msgs to the client for the Data/Mover on 
 *					the NNIM.
 *
 * Created by: Anh Vu
 *
 * Date Created:        11/15/01
 *
 *-----------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <nfx-types.h>
#include <exports/eee-api.h>
#include <exports/elog-api.h>
#include <exports/anpssc-api.h>
#include "ndmp.h"
#include "ndmp-debug.h"
#include "ndmp-fs-api.h"
#include "ndmp-ssc.h"
#include <nfx-incl.h>

int ndmpv2_sndLogDbg(	ndmpSess_t *sess, int32 logTyp, char *msg);
int ndmpv2_mkFHReq(ndmpSess_t *sess,  char8 *rcvbuf);
//int fill_dir_entry(ndmpSess_t *sess, ndmp_dir *de, char *fname, fh_dir_rec_t *f_data);
int ndmp_mkFHReq(ndmpSess_t *sess,  char8 *rcvbuf);

#define DOT_FILE	"."
#define DOT_DOT_FILE ".."

char buf[200];

/*-----------------------------------------------------------------
 * Name: ndmp_procNotifHaltedMsg 
 *
 * Description:	Send a NOTIFY_HALTED msg (for Data or Mover) to 
 *				NDMP client.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        11/15/01
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procNotifHaltedMsg(ndmpSess_t *sess, char *rcvbuf, ndmp_src_t *src)
{
	ndmp_NFXRsp_t *rsp = (ndmp_NFXRsp_t *)rcvbuf;
    ndmp_notify_data_halted_req *d_req = (ndmp_notify_data_halted_req *) rsp->content;
    ndmp_notify_mover_halted_req *m_req = (ndmp_notify_mover_halted_req *) rsp->content;
	void	*req;
	uint32	m_typ, *msgtyp = (uint32 *)rcvbuf;

	
	if (*msgtyp == NDMP_MID_NOTIFY_DATA_HALTED)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_procNotifHaltedMsg:  msgtyp = NDMP_MID_NOTIFY_DATA_HALTED, halt_reason %s.\n", decode_data_halt_reas(d_req->reason)));
		d_req->text_reason = "";
		req = (void*) d_req;
		m_typ = NDMP_NOTIFY_DATA_HALTED;
	} else {
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_procNotifHaltedMsg:  msgtyp = NDMP_NOTIFY_MOVER_HALTED, halt_reason %s.\n", decode_mover_halt_reas(m_req->reason)));
		if (rsp->len)
		{	
			m_req->text_reason = (char *) (rsp->content + rsp->len);
			NDMP_DBG_PRINTF(D_INFO, ("txt_reason %s\n", m_req->text_reason));
		} else {
		    m_req->text_reason = "";
		}
		req = (void*) m_req;
		m_typ = NDMP_NOTIFY_MOVER_HALTED;
	}	
    if (ndmpXmtMsg(sess, NDMP_MSG_REQ, m_typ, 0, NDMP_NO_ERR, (void *) req) < 0)
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_procNotifHaltedMsg:  Error sending %d req.\n", *msgtyp));
        E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
                ("Error sending %d req.", *msgtyp));
    }
    return NFX_OK;

} /* ndmp_procNotifHaltedMsg */

/*-----------------------------------------------------------------
 * Name: 			ndmpv2_sndLogDbg
 *
 * Description:		Send a v2 LOG DEBUG msg to NDMP client.
 *
 * Created by:  	Anh Vu
 *
 * Date Created:   	04/11/02
 *
 *-----------------------------------------------------------------
 */
int
ndmpv2_sndLogDbg(	ndmpSess_t *sess, 
				int32 logTyp, char *msg)
{
	ndmpv2_log_dbg_req	log_dbg;
	int32				rc = NFX_OK;

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_sndLogDbg:  %s.\n", msg));

	bzero(&log_dbg, sizeof(ndmpv2_log_dbg_req));
   	log_dbg.msg = msg;

    if (ndmpXmtMsg(sess, NDMP_MSG_REQ, NDMP_LOG_DEBUG, 0, NDMP_NO_ERR, &log_dbg) < 0)
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_sndLogDbg:  Error sending NDMP_LOG_DBG req.\n"));
        E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
                ("Error sending NDMP_LOG_DBG req."));
		rc = NFX_ERR;
    }

	return rc;
} /* ndmpv2_sndLogDbg */


/*-----------------------------------------------------------------
 * Name: ndmp_sndLogMsg
 *
 * Description:	Send a log msg from SSC to NDMP client.
 *				Arg 'msgid' is for future use.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        01/08/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_sndLogMsg(	ndmpSess_t *sess, 
				int32 logTyp, 
				char *msg, int32 msgid)
{
	ndmp_log_msg_req	log_msg;
	ndmpv2_log_log_req	log_log;
	int32				rc = NFX_OK, log_msg_typ;
	void				*log_req;

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_sndLogMsg:  %s.\n", msg));

	if (sess->protoVer == NDMP_VER_2)
	{
		bzero(&log_log, sizeof(ndmpv2_log_log_req));
    	log_log.entry = msg;

	} else /* v3 & v4 should be same */
	{
		bzero(&log_msg, sizeof(ndmp_log_msg_req));
    	log_msg.log_type = logTyp;
    	log_msg.entry = msg;
    	log_msg.message_id = msgid;
            	/* could also be used to refer msg in reference doc
             	 * if we want to here we stick in FS status for now
             	 * which follows file_name_len field in blob.
             	 */
	}
	log_req = (sess->protoVer == NDMP_VER_2)? (void *)&log_log: (void *)&log_msg;
	log_msg_typ = (sess->protoVer == NDMP_VER_2)? NDMP_LOG_LOG : NDMP_LOG_MESSAGE;
	
    if (ndmpXmtMsg(sess, NDMP_MSG_REQ, log_msg_typ, 0, NDMP_NO_ERR, log_req) < 0)
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_sndLogMsg:  Error sending %s req.\n",
						(sess->protoVer == NDMP_VER_2)? "NDMP_LOG_LOG" : "NDMP_LOG_MESSAGE"));
        E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
                	("Error sending %s req.", 
					(sess->protoVer == NDMP_VER_2)? "NDMP_LOG_LOG" : "NDMP_LOG_MESSAGE"));
		rc = NFX_ERR;
    }

	return rc;
} /* ndmp_sndLogMsg */

/*-----------------------------------------------------------------
 * Name: ndmp_procLogFile 
 *
 * Description:	Send a LOG_FILE request to NDMP client.  This
 *				msg indicates the file/dir that was or was not
 *				recovered by the FS restore program.  These msgs
 *				are generated only for file	paths indicated in 
 *				the file_list that was tagged on with the restore 
 *				command from NDMP client.  If FS send us a log txt
 *				within this log file msg, we send the log txt
 *				msg separately.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        01/08/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procLogFile(ndmpSess_t *sess, char *rcvbuf, ndmp_src_t *src)
{
	ndmp_ssc_send_log_req_msg_t *req = (ndmp_ssc_send_log_req_msg_t *)rcvbuf;
	ndmp_log_file_req		log_file;
	ndmpv2_log_file_req		v2_log_file;
	void					*log_req;
	int32					error;


	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procLogFile:  FS stat = %s\n", 
			(req->status == FS_ERR_NONE)? "NDMP_NO_ERR":
            (req->status == FS_ERR_NOENT)? "NDMP_FILE_NOT_FOUND_ERR" :
            "NDMP_UNDEFINED_ERR"));
	
	error = (req->status == FS_ERR_NONE)? NDMP_NO_ERR:
			(req->status == FS_ERR_NOENT)? NDMP_FILE_NOT_FOUND_ERR :
			NDMP_UNDEFINED_ERR;    

	if (sess->protoVer == NDMP_VER_2)
	{
		log_req = (void *) &v2_log_file;
		v2_log_file.error = error;
		v2_log_file.name =  req->blob;
	} else {
		log_req = (void *) &log_file;
		log_file.error = error;
		log_file.name =  req->blob;
	}	
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procLogFile:  FS file = %s\n", req->blob));

	if (ndmpXmtMsg(sess, NDMP_MSG_REQ, NDMP_LOG_FILE, 0, NDMP_NO_ERR, log_req) < 0)
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_procLogFile:  Error sending NDMP_LOG_FILE req.\n"));
        E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0, 
				("Error sending NDMP_LOG_FILE req."));
    }

	/* If there's log_txt sent together w/ this log_file msg, it's
	 * likely an error log.
	 */
	if (req->log_txt_len)
	{
		return ndmp_sndLogMsg(sess, NDMP_LOG_ERROR, 
						&req->blob[req->file_name_len], 0);
	}

	return NFX_OK;	

} /* ndmp_procLogFile */

/*-----------------------------------------------------------------
 * Name: ndmp_procLogMsg 
 *
 * Description:	Send a LOG_MSG to NDMP client.  This is when FS
 *				asks us to forward a log message to NDMP client
 *				pertaining to the bkup/restore operation.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        11/15/01
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procLogMsg(ndmpSess_t *sess, char *rcvbuf, ndmp_src_t *src)
{
	ndmp_ssc_send_log_req_msg_t *req = (ndmp_ssc_send_log_req_msg_t *)rcvbuf;
	int32				msgTyp, log_type;


	msgTyp = req->log_msgTyp;
	log_type = (msgTyp == NDMP_LOG_MSG_TYP_INFO)? NDMP_LOG_NORMAL:
			 		((msgTyp == NDMP_LOG_MSG_TYP_ERR) || (msgTyp == NDMP_LOG_MSG_TYP_PANIC))? 
					NDMP_LOG_ERROR :
                   	(msgTyp == NDMP_LOG_MSG_TYP_WARNING)? 
					NDMP_LOG_WARNING: NDMP_LOG_DEBUG; 

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procLogMsg:  type %s, text %s\n", 
								(log_type == NDMP_LOG_NORMAL)? "normal":
								(log_type == NDMP_LOG_ERROR)? "error":
								(log_type == NDMP_LOG_WARNING)? "warning":
									"debug", &req->blob[req->file_name_len]));

	if (sess->protoVer == NDMP_VER_2)
	{
		if (log_type == NDMP_LOG_DEBUG)
		{
			return ndmpv2_sndLogDbg(sess, NDMP_DBG_DIAG_INFO, 
					&req->blob[req->file_name_len]);
		}
	} else {
		return ndmp_sndLogMsg(sess, log_type, &req->blob[req->file_name_len], 0);
	}
	return NFX_OK;

} /* ndmp_procLogMsg */

#if 1 // this one is v2 DIR/NODE 
/*-----------------------------------------------------------------
 * Name: 			ndmpv2_mkFHReq 
 *
 * Description:		Send a v2 File History of DIR/NODE type msg to NDMP 
 * 					client.  Do this on	behalf of the NNIM.
 *
 * Created by:  	Anh Vu
 *
 * Date Created:	11/15/01
 *
 *-----------------------------------------------------------------
 */
int
ndmpv2_mkFHReq(ndmpSess_t *sess,  char8 *rcvbuf)
{
	ndmp_NFX_FHReq_t 	*req = (ndmp_NFX_FHReq_t *) rcvbuf;
	fs_DirHistory_t 	*f_ddata = (fs_DirHistory_t *) req->content;
	fs_NodeHistory_t	*f_ndata = (fs_NodeHistory_t *) req->content;
	ndmpv2_fh_add_dir_req	fd_msg;
	ndmpv2_fh_add_node_req	fn_msg;
	ndmpv2_dir 			*d_lst = NULL;
	ndmpv2_node			*n_lst = NULL;
	int32				i; 
	ndmpv2_file_stat	*fstat;	
	char				*ptr;

	NDMP_DBG_PRINTF(D_INFO, ("ndmpv2_mkFHReq:  # of files rcvd = %d.\n", req->num_files));

 	if (req->mid == NDMP_MID_FH_ADD_DIR)
 	{

    	/******* DIR **********/

    	NDMP_SSC_ALLOC(d_lst, ndmpv2_dir, req->num_files * sizeof(ndmpv2_dir));
    	if (d_lst == NULL)
		{
			NDMP_DBG_PRINTF(D_INFO, ("ndmpv2_mkFHReq:  Malloc error d_lst, send log to client.\n"));
			sprintf(buf, "SessId %d:  File History cannot be generated.  "
						"Server internal error, out of memory.\n", sess->sessId);
			ndmp_sndLogMsg(sess, NDMP_LOG_WARNING, buf, 0);
			return NFX_OK;
		}
		memset(d_lst, 0, req->num_files * sizeof(ndmpv2_dir));
		fd_msg.dirs.dirs_len = req->num_files; 

		f_ddata = (fs_DirHistory_t *) req->content;

    	NDMP_DBG_PRINTF(D_INFO, ("req->num_files received = %d\n", req->num_files));

		for (i= 0; i< req->num_files ; i++)
		{
        	NDMP_DBG_PRINTF(D_INFO, ("file_name %s\nfile_name len %d\nnode id %lld\nparent node id %lld\n", 
				f_ddata->name, strlen(f_ddata->name), f_ddata->inum, f_ddata->parent));

			d_lst[i].name = f_ddata->name;
			d_lst[i].node = (u_long) f_ddata->inum;	
			d_lst[i].parent = (u_long) f_ddata->parent;	
			ptr = (char *)f_ddata;
			ptr = (char *)
					(((unsigned long) ptr + sizeof(fs_DirHistory_t) + f_ddata->len + 7) & ~7); 
			f_ddata = (fs_DirHistory_t *) ptr;
    	}

		fd_msg.dirs.dirs_val = d_lst;

		/* Send the ADD_DIR msg */
		if (ndmpXmtMsg(sess, NDMP_MSG_REQ, NDMP_FH_ADD_UNIX_DIR, 0, NDMP_NO_ERR, (void *) &fd_msg) < 0)
    	{
        	NDMP_DBG_PRINTF(D_INFO, ("ndmpv2_mkFHReq:  Error sending NDMP_FH_ADD_UNIX_DIR req.\n"));
        	E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0, 
						("Error sending NDMP_FH_ADD_UNIX_DIR req.\n"));
			sprintf(buf, "SessId %d:  File History cannot be generated.  "
						"Server internal error in sending DIR msg.\n", sess->sessId);
	    	ndmp_sndLogMsg(sess, NDMP_LOG_WARNING, buf, 0);
    	}
		NDMP_SSC_FREE(d_lst);

 	} else {
    /******* NODE **********/

		/* Formulate NODE File History message for x number of files */
		fn_msg.nodes.nodes_len = req->num_files; 

		NDMP_DBG_PRINTF(D_INFO, ("num files = %d\n", req->num_files));
    	NDMP_SSC_ALLOC(n_lst, ndmpv2_node, req->num_files * sizeof(ndmpv2_node));
		if (n_lst == NULL)
		{
			NDMP_DBG_PRINTF(D_INFO, ("ndmpv2_mkFHReq:  Malloc error n_lst\n"));
			sprintf(buf, "SessId %d:  File History cannot be generated.  "
						"Server internal error, out of memory.\n", sess->sessId);
	    	ndmp_sndLogMsg(sess, NDMP_LOG_WARNING, buf, 0);
        	return NFX_OK;
		}
		memset(n_lst, 0, (req->num_files * sizeof(ndmpv2_node)));

		f_ndata = (fs_NodeHistory_t *) req->content;

		for (i= 0; i< req->num_files; i++, f_ndata++)
		{
        	NDMP_DBG_PRINTF(D_INFO, ("\nnode id %lld\n" 
				"file type %d\nmtime %d\natime %d\nctime %d\nowner %d\ngroup %d\nfattr %d\nsize %llu\n"
				"links %d\n---------\n",
				f_ndata->inum, f_ndata->file_Type, f_ndata->mtime.secs, 
                f_ndata->atime.secs, f_ndata->ctime.secs, f_ndata->owner, 
                f_ndata->group, f_ndata->fattr, f_ndata->size, f_ndata->links));
		    n_lst[i].node = (u_long) f_ndata->inum;
		    fstat = &n_lst[i].stat;
     	    fstat->ftype = f_ndata->file_Type;
            fstat->mtime = f_ndata->mtime.secs;
            fstat->atime = f_ndata->atime.secs;
            fstat->ctime = f_ndata->ctime.secs;
            fstat->owner = f_ndata->owner;
            fstat->group = f_ndata->group;
            fstat->fattr = f_ndata->fattr;
            fstat->size.low = (u_long) f_ndata->size;
            fstat->size.high = (u_long) (f_ndata->size >> 32);
            fstat->fh_info.low = 0;             // data stream position for file, used for DAR only
            fstat->fh_info.high = 0;            // we don't support DAR at the moment
		}	
		fn_msg.nodes.nodes_val = n_lst;

		NDMP_DBG_PRINTF(D_INFO, ("Sending NODE msg\n"));
		if (ndmpXmtMsg(sess, NDMP_MSG_REQ, NDMP_FH_ADD_UNIX_NODE, 0, NDMP_NO_ERR, (void *) &fn_msg) < 0)
    	{
        	NDMP_DBG_PRINTF(D_INFO, ("ndmpv2_mkFHReq:  Error sending NDMP_FH_ADD_UNIX_NODE req.\n"));
        	E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0, 
						("Error sending NDMP_FH_ADD_UNIX_NODE req.\n"));
			sprintf(buf, "SessId %d:  File History cannot be generated.  "
						"Server internal error in sending NODE msg.\n", sess->sessId);
	    	ndmp_sndLogMsg(sess, NDMP_LOG_WARNING, buf, 0);
    	} 
    	NDMP_SSC_FREE(n_lst);
 	}

	return NFX_OK;

} /* ndmpv2_mkFHReq */
#endif

#if 0 //v2 FILE type fh
/*-----------------------------------------------------------------
 * Name: 			ndmpv2_mkFHReq 
 *
 * Description:		Send a v2 File History of FILE type msg to NDMP 
 *                  client.  Do this on	behalf of the NNIM.
 *
 * Created by:  	Anh Vu
 *
 * Date Created:	11/15/01
 *
 *-----------------------------------------------------------------
 */
int
ndmpv2_mkFHReq(ndmpSess_t *sess,  char8 *rcvbuf)
{
	ndmp_NFX_FHReq_t 	*req = (ndmp_NFX_FHReq_t *) rcvbuf;
	fh_rec_t 			*f_data = (fh_rec_t *) req->content;
	ndmpv2_fh_add_file_req 	fh_msg; 
	ndmpv2_file_stat  	*fstat;
	int32				i; 
	ndmpv2_file 		*f_lst;
	int32				whole_fn_len;
	char				*fname_ptr;
	

	/* Formulate File History message for x number of files */
	fh_msg.files.files_len = req->num_files; 
	NDMP_DBG_PRINTF(D_INFO, ("files_len = %d\n", req->num_files));
    NDMP_SSC_ALLOC(f_lst, ndmpv2_file, (req->num_files * sizeof(ndmpv2_file)));
	if (f_lst == NULL)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmpv2_mkFHReq:  Malloc error f_lst\n"));
		return NFX_ERR;
	}
	memset(f_lst, 0, ((req->num_files) * sizeof(ndmpv2_file)));

	for (i= 0; i< req->num_files; i++, f_data++)
	{
		fname_ptr = (char *) (req->content + f_data->fname_offset);

		/* Bakbone requested that we return '/vol1' always */
		/* Append the volname in front of each file name */		

		/* If filename is a '/', then just return '/volname */
		if (!strcmp(fname_ptr, "/"))
		{
			whole_fn_len = strlen(sess->Serv.Data.fs_volname) +  2;  //for BakBone
		} else {
			/* else return '/volname/fn' */
			whole_fn_len = strlen(sess->Serv.Data.fs_volname) + f_data->fname_len + 2;  //for BakBone
		}
        NDMP_SSC_ALLOC(f_lst[i].name, char, whole_fn_len);
		if (f_lst[i].name == NULL)
        {
            NDMP_DBG_PRINTF(D_INFO, ("ndmp_procFileHistMsg:  Malloc error unix_name.\n"));
            return NFX_ERR;
        }

		/* Always prepend '/volname' */
		bzero(f_lst[i].name, whole_fn_len);
		strcpy(f_lst[i].name, "/");
		strcpy(f_lst[i].name + 1, sess->Serv.Data.fs_volname);

		/* If filename is not a '/', append '/volname' to file name */
		if (strcmp(fname_ptr, "/") != 0)
		{
			strcpy(f_lst[i].name + strlen(sess->Serv.Data.fs_volname) + 1, fname_ptr);	
		}
		
        NDMP_DBG_PRINTF(D_INFO, ("file_name %s\n", f_lst[i].name));
        NDMP_DBG_PRINTF(D_INFO, ("file_name len %d\n", strlen(f_lst[i].name)));

		fstat = &f_lst[i].stat;
        fstat->ftype = f_data->file_type;
        fstat->mtime = f_data->mtime.secs;
        fstat->atime = f_data->atime.secs;
        fstat->ctime = f_data->ctime.secs;
        fstat->owner = f_data->owner;
        fstat->group = f_data->group;
        fstat->fattr = f_data->fattr;
        fstat->size.low = f_data->size;
        fstat->size.high = f_data->size >> 32;
        fstat->fh_info.low = 0;             // data stream position for file, used for DAR only
        fstat->fh_info.high = 0;            // we don't support DAR at the moment

    	NDMP_DBG_PRINTF(D_INFO, ("file type %d\n", f_data->file_type));
    	NDMP_DBG_PRINTF(D_INFO, ("mtime %d\n", f_data->mtime.secs));
    	NDMP_DBG_PRINTF(D_INFO, ("atime %d\n", f_data->atime.secs));
    	NDMP_DBG_PRINTF(D_INFO, ("ctime %d\n", f_data->ctime.secs));
    	NDMP_DBG_PRINTF(D_INFO, ("owner %d\n", f_data->owner));
    	NDMP_DBG_PRINTF(D_INFO, ("group %d\n", f_data->group));
    	NDMP_DBG_PRINTF(D_INFO, ("fattr %d\n", f_data->fattr));
		
	}	

	fh_msg.files.files_val = f_lst;
	if (ndmpXmtMsg(sess, NDMP_MSG_REQ, NDMP_FH_ADD_UNIX_PATH, 0, NDMP_NO_ERR, (void *) &fh_msg) < 0)
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmpv2_mkFHReq:  Error sending NDMP_FH_ADD_FILE req.\n"));
        E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0, 
						("Error sending NDMP_FH_ADD_UNIX_PATH req.\n"));
    }

	for (i = 0; i<req->num_files; i++)
	{
		if (f_lst[i].name) 
			NDMP_SSC_FREE (f_lst[i].name);
	}

	NDMP_SSC_FREE (f_lst);

	return NFX_OK;

} /* ndmpv2_mkFHReq */
#endif

#if 1  //v3 DIR/NODE fh
/*-----------------------------------------------------------------
 * Name: 			ndmp_mkFHReq 
 *
 * Description:		Send a v3 File History msg to NDMP client.  Do this on
 *					behalf of the NNIM.  This sends FH as DIR/NODE types.
 *
 * Created by:  	Anh Vu
 *
 * Date Created:	11/15/01
 *
 *-----------------------------------------------------------------
 */
int
ndmp_mkFHReq(ndmpSess_t *sess,  char8 *rcvbuf)
{
	ndmp_NFX_FHReq_t 	*req = (ndmp_NFX_FHReq_t *) rcvbuf;
	fs_DirHistory_t		*f_ddata = (fs_DirHistory_t *) req->content;
	fs_NodeHistory_t	*f_ndata = (fs_NodeHistory_t *) req->content;
	ndmp_fh_add_dir_req	fd_msg;
	ndmp_fh_add_node_req	fn_msg;
	ndmp_dir 			*d_lst = NULL;
	ndmp_node			*n_lst = NULL;
	int32				i;
	ndmp_file_stat		*fstat;	
	char *ptr;


	NDMP_DBG_PRINTF(D_INFO, ("ndmp_mkFHReq: # of files received = %d\n", req->num_files));

    if (req->mid == NDMP_MID_FH_ADD_DIR)
    {
        /******* DIR **********/

        NDMP_SSC_ALLOC(d_lst, ndmp_dir, req->num_files  * sizeof(ndmp_dir));
        if (d_lst == NULL)
	    {
		    NDMP_DBG_PRINTF(D_INFO, ("ndmp_mkFHReq:  Malloc error d_lst, send log to client.\n"));
		    sprintf(buf, "SessId %d:  File History cannot be generated.  "
						"Server internal error, out of memory.\n", sess->sessId);
		    ndmp_sndLogMsg(sess, NDMP_LOG_WARNING, buf, 0);
		    return NFX_OK;
	    }
	    memset(d_lst, 0, req->num_files  * sizeof(ndmp_dir));
	    fd_msg.dirs.dirs_len = req->num_files ; 

	    f_ddata = (fs_DirHistory_t *) req->content;

        NDMP_DBG_PRINTF(D_INFO, ("req->num_files received = %d\n", req->num_files));

	    for (i= 0; i< req->num_files ; i++)
	    {
            NDMP_DBG_PRINTF(D_INFO, ("file_name %s\nfile_name len %d\nnode id %lld\nparent node id %lld\n",
                f_ddata->name, strlen(f_ddata->name), f_ddata->inum, f_ddata->parent));

	        NDMP_SSC_ALLOC(d_lst[i].names.names_val, ndmp_file_name, sizeof(ndmp_file_name));
        	if (d_lst[i].names.names_val == NULL)
	        {
		        NDMP_DBG_PRINTF(D_INFO, ("fill_dir_entry:  Malloc error names_val, send log to client.\n"));
		        sprintf(buf, "SessId %d:  File History cannot be generated.  "
					"Server internal error, out of memory.\n", sess->sessId);
                ndmp_sndLogMsg(sess, NDMP_LOG_WARNING, buf, 0);
		        goto mkFHReq_free_dlst;
	        }
	        d_lst[i].names.names_len = 1;
	        d_lst[i].names.names_val->fs_type =  NDMP_FS_UNIX;
	        d_lst[i].names.names_val->ndmp_file_name_u.unix_name = f_ddata->name;
        	d_lst[i].node.low = (u_long) f_ddata->inum;	
        	d_lst[i].node.high = (u_long) (f_ddata->inum >> 32);	
        	d_lst[i].parent.low = (u_long) f_ddata->parent;	
        	d_lst[i].parent.high = (u_long) (f_ddata->parent >> 32);	
   	        NDMP_DBG_PRINTF(D_INFO, ("ndmp_mkFHReq:  file_name %s\n", 
							d_lst[i].names.names_val->ndmp_file_name_u.unix_name));
			ptr = (char *)f_ddata;
			ptr = (char *)
					(((unsigned long) ptr + sizeof(fs_DirHistory_t) + f_ddata->len + 7) & ~7); 
			f_ddata = (fs_DirHistory_t *) ptr;
        }

	    fd_msg.dirs.dirs_val = d_lst;

	    /* Send the ADD_DIR msg */
	    NDMP_DBG_PRINTF(D_INFO, ("ndmp_mkFHReq:  Sending DIR msg\n"));
	    if (ndmpXmtMsg(sess, NDMP_MSG_REQ, NDMP_FH_ADD_DIR, 0, NDMP_NO_ERR, (void *) &fd_msg) < 0)
        {
            NDMP_DBG_PRINTF(D_INFO, ("ndmp_mkFHReq:  Error sending NDMP_FH_ADD_UNIX_DIR req.\n"));
            E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0, 
						("Error sending NDMP_FH_ADD_DIR req.\n"));
		    sprintf(buf, "SessId %d:  File History cannot be generated.  "
						"Server internal error in sending DIR msg.\n", sess->sessId);
		    ndmp_sndLogMsg(sess, NDMP_LOG_WARNING, buf, 0);
        }

mkFHReq_free_dlst:
	    if (d_lst)
	    {
		    for (i = 0; i< req->num_files; i++)
			    NDMP_SSC_FREE (d_lst[i].names.names_val);
		    NDMP_SSC_FREE (d_lst);
	    }

    } else {

        /******* NODE **********/

	    /* Formulate NODE File History message for x number of files */
	    fn_msg.nodes.nodes_len = req->num_files; 

        NDMP_SSC_ALLOC(n_lst, ndmp_node, req->num_files * sizeof(ndmp_node));
	    if (n_lst == NULL)
	    {
		    NDMP_DBG_PRINTF(D_INFO, ("ndmp_mkFHReq:  Malloc error n_lst\n"));
		    sprintf(buf, "SessId %d:  File History cannot be generated.  "
						"Server internal error, out of memory.\n", sess->sessId);
            ndmp_sndLogMsg(sess, NDMP_LOG_WARNING, buf, 0);
            return NFX_OK;
	    }
	    memset(n_lst, 0, (req->num_files * sizeof(ndmp_node)));

	    f_ndata = (fs_NodeHistory_t *) req->content;

	    for (i= 0; i< req->num_files; i++, f_ndata++)
	    {
            NDMP_DBG_PRINTF(D_INFO, ("\nnode id %lld\n"
                    "file type %d\nmtime %d\natime %d\nctime %d\nowner %d\ngroup %d\nfattr %d\nsize %llu\n"
                    "links %d\n---------\n",
                    f_ndata->inum, f_ndata->file_Type, f_ndata->mtime.secs,
                    f_ndata->atime.secs, f_ndata->ctime.secs, f_ndata->owner,
                    f_ndata->group, f_ndata->fattr, f_ndata->size, f_ndata->links));

    		n_lst[i].node.low = (u_long) f_ndata->inum;
    		n_lst[i].node.high = (u_long) (f_ndata->inum >> 32);
            n_lst[i].fh_info.low = 0;             // data stream position for file, used for DAR only
            n_lst[i].fh_info.high = 0;            // we don't support DAR at the moment

		    n_lst[i].stats.stats_len = 1;

		    NDMP_SSC_ALLOC(n_lst[i].stats.stats_val, ndmp_file_stat, sizeof(ndmp_file_stat));
		    if (n_lst[i].stats.stats_val == NULL)
		    {
			    NDMP_DBG_PRINTF(D_INFO, ("ndmp_mkFHReq:  Malloc error NODE stats_val.\n"));
			    sprintf(buf, "SessId %d:  File History cannot be generated.  "
						"Server internal error, out of memory.\n", sess->sessId);
	            ndmp_sndLogMsg(sess, NDMP_LOG_WARNING, buf, 0);
			    goto mkFHReq_free_nlst;
		    }

		    fstat = n_lst[i].stats.stats_val;
            fstat->invalid = 0;          // we support all (atime, ctime, and group)
            fstat->fs_type = NDMP_FS_UNIX;       // always a unix file system type
     	    fstat->ftype = f_ndata->file_Type;
            fstat->mtime = f_ndata->mtime.secs;
            fstat->atime = f_ndata->atime.secs;
            fstat->ctime = f_ndata->ctime.secs;
            fstat->owner = f_ndata->owner;
            fstat->group = f_ndata->group;
            fstat->fattr = f_ndata->fattr;
            fstat->size.low = (u_long) f_ndata->size;
            fstat->size.high = (u_long) (f_ndata->size >> 32);
            fstat->links = f_ndata->links;

	    }	

	    fn_msg.nodes.nodes_val = n_lst;

	    NDMP_DBG_PRINTF(D_INFO, ("Sending NODE msg\n"));
	    if (ndmpXmtMsg(sess, NDMP_MSG_REQ, NDMP_FH_ADD_NODE, 0, NDMP_NO_ERR, (void *) &fn_msg) < 0)
        {
            NDMP_DBG_PRINTF(D_INFO, ("ndmp_mkFHReq:  Error sending NDMP_FH_ADD_NODE req.\n"));
            E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0, 
						("Error sending NDMP_FH_ADD_NODE req.\n"));
		    sprintf(buf, "SessId %d:  File History cannot be generated.  "
						"Server internal error in sending NODE msg.\n", sess->sessId);
            ndmp_sndLogMsg(sess, NDMP_LOG_WARNING, buf, 0);
        } 

mkFHReq_free_nlst:
	    if (n_lst)
	    {
		    for (i = 0; i< req->num_files; i++)
			    NDMP_SSC_FREE (n_lst[i].stats.stats_val);

		    NDMP_SSC_FREE (n_lst);
	    }
    }
	return NFX_OK;


} /* ndmp_mkFHReq */
#endif 

#if 0// this one works as v3 FILE type fh
/*-----------------------------------------------------------------
 * Name: 		ndmp_mkFHReq 
*
 *
 * Description:	Send a v3 File History msg to NDMP client.  Do this on
 *				behalf of the NNIM.  This sends FH as FILE type.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        11/15/01
 *
 *-----------------------------------------------------------------
 */
int
ndmp_mkFHReq(ndmpSess_t *sess,  char8 *rcvbuf)
{
	ndmp_NFX_FHReq_t *req = (ndmp_NFX_FHReq_t *) rcvbuf;
	int32			i, whole_fn_len;
	ndmp_file_stat  *fstat;
	ndmp_fh_add_file_req 	fh_msg; 
	fh_rec_t 		*f_data = (fh_rec_t *) req->content;
	ndmp_file 		*f_lst;
	char			*fname_ptr;
	

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procFileHistMsg.\n"));


	/* Formulate File History message for x number of files */
	fh_msg.files.files_len = req->num_files;
    NDMP_SSC_ALLOC(f_lst, ndmp_file, req->num_files * sizeof(ndmp_file));
	if (f_lst == NULL)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_procFileHistMsg:  Malloc error f_lst\n"));
		return NFX_ERR;
	}
	memset(f_lst, 0, (req->num_files * sizeof(ndmp_file)));
	for (i= 0; i< req->num_files; i++, f_data++)
	{
		f_lst[i].names.names_len = 1;		// only 1 name for this file
        NDMP_SSC_ALLOC(f_lst[i].names.names_val, ndmp_file_name,
                                                 sizeof(ndmp_file_name));
		if (f_lst[i].names.names_val == NULL)
		{
			NDMP_DBG_PRINTF(D_INFO, ("ndmp_procFileHistMsg:  Malloc error names_val\n"));
			return NFX_ERR;
		}
		f_lst[i].names.names_val->fs_type =  NDMP_FS_UNIX; 

		fname_ptr = (char *) (req->content + f_data->fname_offset);
		if (!strcmp(fname_ptr, "/"))
		{
			whole_fn_len = strlen(sess->Serv.Data.fs_volname) +  2;  //for BakBone
		} else {
			whole_fn_len = strlen(sess->Serv.Data.fs_volname) + f_data->fname_len + 2;  //for BakBone
		}
        NDMP_SSC_ALLOC(f_lst[i].names.names_val->ndmp_file_name_u.unix_name,
                       char, whole_fn_len);
		if (f_lst[i].names.names_val->ndmp_file_name_u.unix_name == NULL)
        {
            NDMP_DBG_PRINTF(D_INFO, ("ndmp_procFileHistMsg:  Malloc error unix_name.\n"));
            return NFX_ERR;
        }
		bzero(f_lst[i].names.names_val->ndmp_file_name_u.unix_name, whole_fn_len);

		strcpy(f_lst[i].names.names_val->ndmp_file_name_u.unix_name, "/");
		strcpy(f_lst[i].names.names_val->ndmp_file_name_u.unix_name + 1,
				sess->Serv.Data.fs_volname);

		if (strcmp(fname_ptr, "/") != 0)
		{
			strcpy(f_lst[i].names.names_val->ndmp_file_name_u.unix_name +
				strlen(sess->Serv.Data.fs_volname) + 1, fname_ptr);	
		}
		f_lst[i].stats.stats_len = 1;  // only 1 file attrib for this file

        NDMP_DBG_PRINTF(D_INFO, ("file_name %s\n", 
				f_lst[i].names.names_val->ndmp_file_name_u.unix_name));
        NDMP_DBG_PRINTF(D_INFO, ("file_name len %d\n", 
				strlen(f_lst[i].names.names_val->ndmp_file_name_u.unix_name)));

        NDMP_SSC_ALLOC(f_lst[i].stats.stats_val, ndmp_file_stat,
                                                 sizeof(ndmp_file_stat));
		if (f_lst[i].stats.stats_val == NULL)
		{
			NDMP_DBG_PRINTF(D_INFO, ("ndmp_procFileHistMsg:  Malloc error stats_val.\n"));
			return NFX_ERR;
		}

		fstat = f_lst[i].stats.stats_val;
        fstat->invalid = 0;          // we support all (atime, ctime, and group)
        fstat->fs_type = NDMP_FS_UNIX;       // always a unix file system type
        fstat->ftype = f_data->file_type;
        fstat->mtime = f_data->mtime.secs;
        fstat->atime = f_data->atime.secs;
        fstat->ctime = f_data->ctime.secs;
        fstat->owner = f_data->owner;
        fstat->group = f_data->group;
        fstat->fattr = f_data->fattr;
        fstat->size.low = f_data->size;
        fstat->size.high = f_data->size >> 32;
        fstat->links = f_data->links;

        // fill the rest of ndmp_file struc
        f_lst[i].node.low = f_data->inode;
        f_lst[i].node.high = f_data->inode >> 32;
        f_lst[i].fh_info.low = 0;             // data stream position for file, used for DAR only
        f_lst[i].fh_info.high = 0;            // we don't support DAR at the moment
		
	}	
	fh_msg.files.files_val = f_lst;
	if (ndmpXmtMsg(sess, NDMP_MSG_REQ, NDMP_FH_ADD_FILE, 0, NDMP_NO_ERR, (void *) &fh_msg) < 0)
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_procFileHistMsg:  Error sending NDMP_FH_ADD_FILE req.\n"));
        E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0, 
						("Error sending NDMP_FH_ADD_FILE req.\n"));
    }
	for (i = 0; i<req->num_files; i++)
	{
		if (f_lst[i].names.names_val) 
		{
			if (f_lst[i].names.names_val->ndmp_file_name_u.unix_name)
				NDMP_SSC_FREE (f_lst[i].names.names_val->ndmp_file_name_u.unix_name);
			NDMP_SSC_FREE (f_lst[i].names.names_val);
		}
		if (f_lst[i].stats.stats_val) 
			NDMP_SSC_FREE (f_lst[i].stats.stats_val);
	}
	NDMP_SSC_FREE (f_lst);
	return NFX_OK;

} /* ndmp_mkFHReq */
#endif

/*-----------------------------------------------------------------
 * Name: ndmp_procDirFileHistMsg 
 *
 * Description:	Send a v2 or v3 File History msg to NDMP client.  
 *				Do this on behalf of the NNIM.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        11/15/01
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procFileHistMsg(ndmpSess_t *sess,  char8 *rcvbuf, ndmp_src_t *src)
{

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procDirFileHistMsg.\n"));

	if (sess->protoVer == NDMP_VER_2)
		ndmpv2_mkFHReq(sess, rcvbuf);
	else
		ndmp_mkFHReq(sess, rcvbuf);

	return NFX_OK;
}
/*
 *-----------------------------------------------------------------
 * Name: ndmp_updateEnvReq 
 *
 * Description: Update env. for dump requests. Necessary to inform
 *              the client about change in dump level.
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 08/06/02
 *
 *-----------------------------------------------------------------
 */

int
ndmp_updateEnvReq(ndmpSess_t *sess,  char8 *rcvbuf, ndmp_src_t *src)
{
    ndmp_NFXRsp_t *rsp = (ndmp_NFXRsp_t *)rcvbuf;
    dmpdate_t     *ddrec = (dmpdate_t *)rsp->content;
    Data_SSC_t *data = &sess->Serv.Data;
    int32 i;
    char  err[200];

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_updateEnvReq\n"));

    for (i = 0; i < data->env_len; i++)
    {
        if (!strcmp("LEVEL", data->env_lst[i].name))
            if (atoi(data->env_lst[i].value) != ddrec->dump_level)
            {
                /* Set the level */
                NDMP_DBG_PRINTF(D_INFO, ("ndmp_updateEnvReq: change level from %s to %d\n", data->env_lst[i].value, ddrec->dump_level));
                bzero(err, sizeof(char)*200);
                sprintf(err, "Changing dump level from %s to %d",data->env_lst[i].value, ddrec->dump_level);
                sprintf(data->env_lst[i].value,"%d",ddrec->dump_level);
                ndmp_sndLogMsg(sess, NDMP_LOG_WARNING, err, 0);
                break;
            }
    }

    /*
     * This is a temporary entry for dumpdate
     */
    ddrec->curr_bkup_ind = 1;
    ndmp_setDumpDate(sess, rcvbuf, src);
    memcpy(&data->dd, ddrec, sizeof(dmpdate_t));

    return NFX_OK;
}
/*
 *-----------------------------------------------------------------
 * Name: ndmp_addEnvReq 
 *
 * Description: Add a new env. var to our list.  These env vars
 *              will be returned to us by the client on restore.
 *
 * Created by:  Anh Vu 
 *
 * Date Created: 08/12/02
 *
 *-----------------------------------------------------------------
 */

int
ndmp_addEnvReq(ndmpSess_t *sess,  char8 *rcvbuf, ndmp_src_t *src)
{
#if 0
    ndmp_NFXRsp_t *rsp = (ndmp_NFXRsp_t *)rcvbuf;
    new_env_t *envrec = (new_env_t *)rsp->content;
    Data_SSC_t *data = &sess->Serv.Data;
    int32 i;

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_addEnvReq\n"));

	/* Right now we only have 1 extra ENV var that we add to list,
	 * so this env var will go at end of env_lst.
	 */
	i = data->env_len;

	NDMP_SSC_ALLOC(data->env_lst[i].name, char, (envrec->name_len)+1);
	NFX_ASSERT(data->env_lst[i].name != NULL);
	strcpy(data->env_lst[i].name, envrec->data);
					    
	NDMP_SSC_ALLOC(data->env_lst[i].value, char, (envrec->val_len)+1);
	NFX_ASSERT(data->env_lst[i].value != NULL);
	strcpy(data->env_lst[i].value, &envrec->data[envrec->name_len+1]);

	for (i = 0; i< data->srvr_env_len; i++)
    {
        printf("ndmp_addEnvReq:  [%d].name = %s\n", i, data->env_lst[i].name);
	    printf("ndmp_addEnvReq:  [%d].value = %s\n", i, data->env_lst[i].value);
	}
#endif
    return NFX_OK;
}

