/* 
 *-----------------------------------------------------------------
 *
 * Name:        ndmp-mover-func.c
 *
 * RCS:  $Id: ndmp-mover-func.c,v 1.59 2002/12/06 01:47:54 anhv Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description: 	Process the Mover Server state machine events
 *					for NDMP v2 and v3.  The only difference between
 *					the 2 versions for Mover Server is that there
 *					is no MOVER_CONNECT msg (i.e. no CONNECT event)
 *					in v2.  So we use the same state machine
 *					for maximum common code use.  We are guaranteed
 *					that we won't be getting MOVER_CONNECT event
 *					for v2 because this control msg is intercepted
 *					on the SSC first and will be marked as invalid.
 *
 * Created by: 		Anh Vu
 *
 * Date Created:	9/06/01
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <exports/eee-api.h>
#include <exports/malloc-api.h>
#include <exports/timer-api.h>
#include <exports/cm-api.h>
#include "ndmp.h"
#include "ndmp-fs-api.h"
#include "ndmp-mover-sm.h"
#include "ndmp-data-sm.h"
#include "ndmp-data-v2-sm.h"
#include <exports/fs-ndmp-api.h>
#include "ndmp-debug.h"

#define MAX_SCSIIO_REQ 64

extern ndmp_NNIMCtrl_t ndmp_NNIMCtrl;
extern int ndmp_fp_msgStats[];

// performance meas
extern eeeRTC_t time_wr_req;
extern eeeRTC_t time_wr_rsp;
extern eeeRTC_t time_rd_req;
extern uint64   pm_ioCookie;
int32 ndmp_notify_data_srvr(NdmpNIMService_t *servp, void *data, int32 status, uint64 ioCookie);
int ndmp_sndIOReqstoSCSI(NdmpNIMService_t *servp, int32 *nextState);

/*
 *------------------------------------------------------------------------------
 * Function Name:   dbg_print_rec
 *
 * Description:     Debugging func to print record of data to/from tape.
 *
 * Created by:      Anh Vu
 * Data Created:    
 * Returns:
 *------------------------------------------------------------------------------
 */
void
dbg_print_rec (uchar8 *ptr, int32 len)
{
	int i;

    NDMP_DBG_PRINTF(D_DBG, ("--------------------------------------------------------------------------\n"));
    if (len)
    {
        for (i = 0; i < len; ++i)
        {
            if ((i % 16) == 0)
                NDMP_DBG_PRINTF(D_DBG, ("\n[0x%04x]: ", i));
            NDMP_DBG_PRINTF(D_DBG, ("%02x ", ptr[i]));
        }
        NDMP_DBG_PRINTF(D_DBG, ("\n"));
    }
    NDMP_DBG_PRINTF(D_DBG, ("\n"));
    NDMP_DBG_PRINTF(D_DBG, ("--------------------------------------------------------------------------\n"));
	return;
}
/*
 *------------------------------------------------------------------------------
 * Function Name:   mover_decode_event
 *
 * Description:    
 *
 * Created by:      Anh Vu
 *
 * Data Created:    10/01/02
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
char *
mover_decode_event (int event)
{
	switch (event)
	{
		case NDMP_MOVER_EVENT_LISTEN: 	return "listen";
		case NDMP_MOVER_EVENT_CONNECT: 	return "connect";
		case NDMP_MOVER_EVENT_CONNECTED: return "connected";
		case NDMP_MOVER_EVENT_SET_INFO: return "set_info";
		case NDMP_MOVER_EVENT_WRITE_DATA: return "write_data";
		case NDMP_MOVER_EVENT_WRITE_DONE: 	return "write_done";
		case NDMP_MOVER_EVENT_READ_DATA: 	return "read_data";
		case NDMP_MOVER_EVENT_READ_DONE:  	return "read_done";
		case NDMP_MOVER_EVENT_PAUSE_INDIC: return "pause_indic";
		case NDMP_MOVER_EVENT_CONTINUE:return "continue";
		case NDMP_MOVER_EVENT_ABORT:		return "abort";
		case NDMP_MOVER_EVENT_STOP:		return "stop";
		case NDMP_MOVER_EVENT_GET_STATE: return "get_state"; 
		case NDMP_MOVER_EVENT_CONN_CLOSED: 	return "conn_closed";
		case NDMP_MOVER_EVENT_CLOSE: return "close"; 
		case NDMP_MOVER_EVENT_TAPE_OPS_DONE: return "tape_ops_done"; 
		case NDMP_MOVER_EVENT_INTERNAL_ERR: return "internal_err"; 
		case NDMP_MOVER_EVENT_CONN_ERR: return "conn_err"; 
		default:						return 	"unknown";
	}
}	

/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_chk_precond
 *
 * Description:    Check for valid recsiz, window len and window offset
 *                  pg 152 v4 spec.
 *
 * Created by:      Anh Vu
 *
 * Data Created:    09/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int32
ndmp_chk_precond (ndmp_mover_mode mode, Mover_t *mover)
{

    /* Record size must be set prior to transitioning out of IDLE state,
     * pg 152 v4 spec
     */
#ifdef NDMP_V4		// V3 does not have these restrictions, will fail Connectathon tests
    if (mover->recSiz == 0) 
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_chk_precond:  RecSiz, has to be set "
				"by client prior to processing LISTEN, CONN, & CONT cmds\n"));
        return NFX_ERR;
    }
    if (!mover->wdwLen) 
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_chk_precond:  Windown Length has to "
				"be set by client prior to processing LISTEN, CONN, & CONT cmds\n"));
        return NFX_ERR;
    }

    /* if we're in bkup operation, wdw len has to be all ones or
     * multiple of recsiz. p159, v4.  Mode_read means read from 
	 * DATA server or data connection.
    */
    if (mode == NDMP_MOVER_MODE_READ)  /* bkup ops */
    {
        /* Window LENGTH has to be max OR multiple of record size */
        if ((mover->wdwLen != NDMP_MOVER_MAX_WDWLEN) && (mover->wdwLen % mover->recSiz)) 
		{
            NDMP_DBG_PRINTF(D_INFO, ("ndmp_chk_precond:  Bkup ops, invalid "
					"window LEN %lld\n", mover->wdwLen));
            return NFX_ERR;
        }
    } else /* mode = WRITE, ie. recovery ops */
    {
        /* Window OFFSET has to be 0 OR multiple of record size */
        if ((mover->wdwOffset) && (mover->wdwOffset % mover->recSiz))
        {
            NDMP_DBG_PRINTF(D_INFO, ("ndmp_chk_precond:  Recov ops, "
					"invalid window OFFSET %lld\n", mover->wdwOffset));
            return NFX_ERR;
        }
    }
#endif /* NDMP_V4 */

    return NFX_OK;

} /* ndmp_chk_precond */
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_mover_release_resource
 *
 * Description:     Device init state machine initialization function.
 *
 * Created by:      Anh Vu
 *
 * Data Created:    09/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
void
ndmp_mover_release_resource(Mover_t *mover)
{
	if (mover == NULL)
		return;

    while (mover->ioPendQ.ioReqCnt)
        ndmp_delReqFromQueue(mover);

	return;
}
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_init
 *
 * Description:     Device init state machine initialization function.
 *
 * Created by:      Anh Vu
 *
 * Data Created:    09/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
void
ndmp_Mover_init(Mover_t *mover)
{
    mover->mode = -1;      /* current mode */
    mover->pauseReas = NDMP_MOVER_PAUSE_NA;      /* current reason */
    mover->haltReas = NDMP_MOVER_HALT_NA;        /* current reason */
    mover->dataWritten = 0;                      /* total written to tape */
    mover->seekPos = 0;                          /* current seek position */
    mover->wdwOffset = -1;                       /* valid data window begin */
    mover->wdwLen = -1;                          /* valid data window length */
    mover->curpos = 0;                         	 /* current data stream pos */
    mover->recSiz = 0;                         	 /* tape I/O record size */
    mover->recNum = 0;                           /* current record num */
    mover->listenSock = -1;                      /* data conn listen socket */
    mover->connSock = -1;                        /* data conn socket */
    mover->rIndex = -1;                          /* buffer read  index */
    mover->wIndex = -1;                          /* buffer write index */
	mover->re_request_rd = FALSE;				 /* to reissue SCSI read after PAUSE->ACTIVE state */
	mover->re_request_rd_ioCookie = 0;				 /* to reissue SCSI read after PAUSE->ACTIVE state */
	mover->re_request_wr = FALSE;				 /* to reissue SCSI write request after PAUSE->ACTIVE */
	mover->re_request_wr_ioCookie = 0;				 /* to reissue SCSI write request after PAUSE->ACTIVE */
	mover->nextIoCookieSend = 1;				/* for restore */

} /* ndmp_Mover_init */
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_create
 *
 * Description:     init function for Mover state machine
 *
 * Created by:      Anh Vu
 *
 * Data Created:    09/17/01 
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int32
ndmp_Mover_create(esm_handle_t   handle,     /* handle for this sm inst    */
                 int32          nextState,  /* next state from table      */
                 esm_event_t    thisEvent,  /* event driving this action  */
                 void           *userInfo)  /* ctrl blk linked to sm      */
{
    NdmpNIMService_t *servp = (NdmpNIMService_t *) userInfo;
	Mover_t		  *mover = &servp->Mover;	

	ndmp_Mover_init(mover);

	//NDMP_DBG_PRINTF(D_INFO, ("Ndmp_Mover_create:  End.  handle %d, "
	//		"sessid %d\n", handle, servp->sessId));
    return(nextState);

} /* ndmp_Mover_create */

/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_listen
 *
 * Description:		The client passes control of tape drive to Mover once this
 *					cmd is issued.   Client can't do further Tape requests
 *					til Mover halts or pauses after this cmd. (p. 20 v4)
 *					In response to this command, we are to provide the address
 *					with which the client will tell the remote server to connect.
 *					We checked for the following on the SSC side already:
 *						- DEV_NOT_OPEN_ERR
 *						- NDMP_PERMISSION_ERR (based on mode indic. by client,
 *							check for tape opened for rd/wr permission).
 *						- NDMP_ILLEGAL_ARGS (invalid addr type) 
 *
 * Created by:      Anh Vu
 *
 * Data Created:    9/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_listen(esm_handle_t handle,
                    int32       nextState, /* next state from table       */
                    esm_event_t thisEvent, /* event driving this action   */
                    void        *userInfo)  /* ctrl blk linked to sm          */
{
    ndmp_NFXRsp_t 			*rsp = NULL;
    ndmp_mover_listen_rsp   *listen_rsp;
    ndmp_NFXReq_t 			*req = (ndmp_NFXReq_t *) thisEvent.p1.u.p;
    ndmp_mover_listen_req   *listen_req = (ndmp_mover_listen_req *) req->content;
    NdmpNIMService_t 		*servp = (NdmpNIMService_t *) userInfo;
    Mover_t     			*mover = &servp->Mover;
    uint32 					dst_port = servp->ssc_src_port, msglen;
	int32					rc, newState = nextState;


    //NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_listen:  Begin, sessid %d, esm_hdl %d\n", servp->sessId, handle));

	/* check for valid recsiz, window len and window offset
	 * pg 152 v4 spec 
	 */
	
    NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_listen:  Sessid %d, recSiz %d, "
			"wdwLen 0x%llx\n", servp->sessId, mover->recSiz, mover->wdwLen));
	if (ndmp_chk_precond(listen_req->mode, mover) == NFX_ERR)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_listen:  Precondion check error. "
				"sessid %d, esm_hdl %d\n", servp->sessId, handle));
		E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
				("ndmp_Mover_listen:  Precondion check error. "
				"sessid %d, esm_hdl %d\n", servp->sessId, handle));
		ndmp_sendCmdRsp(req, dst_port, NDMP_ILLEGAL_STATE_ERR, 0);
		newState = esm_getCurrState(handle);
		goto ndmp_MoverListExit;
	}

    /*
     * Create response message and send it.
     */
	msglen = sizeof(ndmp_NFXRsp_t) + sizeof(ndmp_mover_listen_rsp) - 1;
    NDMP_FP_ALLOC(rsp, ndmp_NFXRsp_t, msglen, "nfxrsp LISTEN");
    if (rsp == NULL)
	{
		E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
				("ndmp_Mover_listen:  Malloc error. "
				"sessid %d, esm_hdl %d\n", servp->sessId, handle));
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_listen:  Malloc error. "
				"sessid %d, esm_hdl %d\n", servp->sessId, handle));
		ndmp_sendCmdRsp(req, dst_port, NDMP_UNDEFINED_ERR, 0);
        newState = esm_getCurrState(handle);
		goto ndmp_MoverListExit;
	}
	listen_rsp = (ndmp_mover_listen_rsp *) rsp->content;

    rsp->hdr.sessId  = req->hdr.sessId;
    rsp->hdr.mid     = req->hdr.mid;
    rsp->ndmp_mid    = req->ndmp_mid;
	memcpy(&rsp->cl_hdr, &req->cl_hdr, sizeof(ndmp_Hdr));

	listen_rsp->error = NDMP_NO_ERR;			

	switch (listen_req->addr_type)
	{
		case NDMP_ADDR_LOCAL:
			listen_rsp->data_conn_addr.addr_type = NDMP_ADDR_LOCAL;
			break;
		case NDMP_ADDR_TCP:
#if 0
			port = getFreePort();
			address = getlocalIPAddr();
			ndmp_tpl_bindReq(port);
			listen_rsp.data_conn_addr.addr_type = NDMP_ADDR_TCP;
			listen_rsp.data_conn_addr.ndmp_addr_u.tcp_addr.ip_addr = address;
			listen_rsp.data_conn_addr.ndmp_addr_u.tcp_addr.port = port;
#endif
			break;
	}
    rc = eee_sendMessage((void *) rsp, msglen, dst_port, ndmp_NNIMCtrl.src_port);
	if (rc != NFX_OK)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_listen: eee_sendMessage failed! sessid %d\n", servp->sessId));
	    E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
	           ("ndmp_Mover_listen: eee_sendMessage failed! sessid %d\n", servp->sessId));
	} else {
#ifdef NFX_SMP 
		ATOMIC_INC_INT(&ndmp_fp_msgStats[SSC_MOVER_LISTEN_RSP_SNT], 1);
#else
		ndmp_fp_msgStats[SSC_MOVER_LISTEN_RSP_SNT]++;
#endif
	}	
    mover->mode = listen_req->mode;           /* current mode: RD/WR */
    mover->addr_type = listen_req->addr_type; /* LOCAL/TCP */          

	if (!mover->tape_hdl)
	{
		/* we piggy-backed the tape device handle */		
		mover->tape_hdl = req->userInfo;	  
	}

    NDMP_FP_FREE(rsp, "nfxrsp LISTEN");

ndmp_MoverListExit:

	/* dealloc what we malloced in msgp in ndmp_NNIMReceiveMsg() */	
    NDMP_FP_FREE(req, "msgp LISTEN");

    return newState;

} /* ndmp_Mover_listen */
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_conn
 *
 * Description:		We get here because of the MOVER_CONNECT_EVENT, which
 *					can be from the client or from the Data server (for local
 *					bkup case).  The client passes control of tape drive to 
 *					Mover once this	cmd is issued.   Client can't do further 
 *					Tape requests til Mover halts or pauses after this cmd. 
 *					(p. 20 v4).  Client is telling Mover to connect to the 
 *					port advertised	by a remote NDMP server.
 *
 * Created by:      Anh Vu
 *
 * Data Created:    9/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_conn(esm_handle_t handle,
                    int32       nextState, /* next state from table       */
                    esm_event_t thisEvent, /* event driving this action   */
                    void        *userInfo)  /* ctrl blk linked to sm          */
{
    ndmp_NFXReq_t 	*req = (ndmp_NFXReq_t *) thisEvent.p1.u.p;
    ndmp_mover_conn_req   *conn_req = (ndmp_mover_conn_req *) req->content;
    NdmpNIMService_t 	*servp = (NdmpNIMService_t *) userInfo;
    Mover_t     	*mover = &servp->Mover;
    int32 			dst_port = servp->ssc_src_port, newState, rc;
    esm_event_t     esmEvent;

    //NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_conn:  Begin, sessid %d, esm_hdl %d\n", 
	//				servp->sessId, handle));

    /* check for valid recsiz, window len and window offset
     * pg 152 v4 spec
     */
	newState = nextState;

    if (ndmp_chk_precond(conn_req->mode, mover) == NFX_ERR)
    {
        ndmp_sendCmdRsp(req, dst_port, NDMP_ILLEGAL_STATE_ERR, 0);
        newState = esm_getCurrState(handle);
		goto ndmp_MoverConnExit;
    }

    switch (conn_req->addr.addr_type)
    {
        case NDMP_ADDR_LOCAL:
			/* Data service has to be listening for Mover to connect to, yeh? */
			if (esm_getCurrState(mover->data_smHdl) != NDMP_DATA_STATE_LISTEN)	
			{
        		ndmp_sendCmdRsp(req, dst_port, NDMP_ILLEGAL_STATE_ERR, 0);
        		newState = esm_getCurrState(handle);
				goto ndmp_MoverConnExit;
			}
			/* send event to signal Data to be in CONNECTED mode as well */
            bzero(&esmEvent, sizeof(esm_event_t));
			esmEvent.event = NDMP_DATA_EVENT_CONNECTED;
            rc = esm_deliverEvent(mover->data_smHdl, &esmEvent);

            if (rc == NFX_ERR)
            {
                NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_conn: ERROR, cannot post "
						"event %d to SM hdl %d",
                      	NDMP_DATA_EVENT_CONNECTED, mover->data_smHdl));
                E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
						("ndmp_Mover_conn: ERROR, cannot post event %d to SM hdl %d",
                      	NDMP_DATA_EVENT_CONNECTED, mover->data_smHdl));
        		newState = esm_getCurrState(handle);
				goto ndmp_MoverConnExit;
            }
			/* Go straight to active since we don't have to
			 * set up a remote data connection.
			 */
			newState = NDMP_MOVER_STATE_ACTIVE;
            break;
        case NDMP_ADDR_TCP:
#if 0
			rc = tpl_openConnReq( &mover->pBcb,
					    &mover->Conn_connId,	/* connId when we're connecting */
						conn_req->addr.ndmp_addr_u.tcp_addr.ip_addr,
						conn_req->addr.ndmp_addr_u.tcp_addr.port );
			if (rc != NFX_OK)
			{
				NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_conn:  tpl_openConnReq "
						"failed w/ %d, sessid %d, handle %d\n", rc, servp->sessId, handle));
        		ndmp_sendCmdRsp(req, dst_port, NDMP_CONNECT_ERR, 0);
        		return esm_getCurrState(handle);
			}
            break;
#endif
		default:
			// shouldn't be here, already handled over SSC
    		NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_conn:  Invalid addr_type %d "
						"received.\n", conn_req->addr.addr_type));
    		E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
						("ndmp_Mover_conn:  Invalid addr_type %d "
						"received.\n", conn_req->addr.addr_type));
    }
	ndmp_sendCmdRsp(req, dst_port, NDMP_NO_ERR, 0);

    mover->mode = conn_req->mode;           /* current mode */
    mover->addr_type = conn_req->addr.addr_type;

	if (!mover->tape_hdl)
	{
		/* we piggy-backed the tape device handle */		
		mover->tape_hdl = req->userInfo;	  
	}

ndmp_MoverConnExit:

	/* free what was originally msgp in ndmp_NNIMReceiveMsg() */	
    NDMP_FP_FREE(req, "msgp MOVER_CONN");

    return newState;

} /* ndmp_Mover_conn */

/*
 *------------------------------------------------------------------------------
 * Function Name:   fill_Mover_get_state_info
 *
 * Description:
 *
 * Created by:      Anh Vu
 *
 * Data Created:    3/21/02
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
fill_Mover_get_state_info(Mover_t *mover, ndmp_mover_get_state_rsp *gs_rsp, esm_handle_t handle)
                    
{
	int32   cur_st;

    //NDMP_DBG_PRINTF(D_INFO, ("fill_Mover_get_state_info:  Begin, esm_hdl %d\n", handle));

	gs_rsp->error	= NDMP_NO_ERR;
	cur_st =  esm_getCurrState(handle);
	gs_rsp->state	= (cur_st == NDMP_MOVER_STATE_IDLE)? NDMP_MOVER_ST_IDLE:
			(cur_st == NDMP_MOVER_STATE_LISTEN)? NDMP_MOVER_ST_LISTEN:
			(cur_st == NDMP_MOVER_STATE_ACTIVE)? NDMP_MOVER_ST_ACTIVE:
			(cur_st == NDMP_MOVER_STATE_PAUSED)? NDMP_MOVER_ST_PAUSED: NDMP_MOVER_ST_HALTED;
	gs_rsp->pause_reason = mover->pauseReas;
	gs_rsp->halt_reason = mover->haltReas;
	gs_rsp->recsiz = mover->recSiz;
	gs_rsp->recnum = mover->recNum;
	gs_rsp->data_written.low = mover->dataWritten;
	gs_rsp->data_written.high = mover->dataWritten >> 32;
	gs_rsp->seek_pos.low = mover->seekPos;
	gs_rsp->seek_pos.high = mover->seekPos >> 32;
	gs_rsp->bytes_left_to_read.low = mover->bytesLeftToRead;
	gs_rsp->bytes_left_to_read.high = mover->bytesLeftToRead >> 32;
	gs_rsp->window_offset.low = mover->wdwOffset;
	gs_rsp->window_offset.high = mover->wdwOffset >> 32;
	gs_rsp->window_len.low = mover->wdwLen;
	gs_rsp->window_len.high = mover->wdwLen >> 32;
	gs_rsp->data_conn_addr.addr_type = mover->dataAddr.addr_type;

	NDMP_DBG_PRINTF(D_INFO, ("Mover Get State:\n"
			"---------------\nState              = %s\n" 
			"Pause Reason       = %s\nHalt Reason        = %s\n" 
			"Recsiz             = %d\n" "RecNum             = %d\n" 
			"Data written       = %lld\n" "Seek position      = %lld\n"
			"Bytes_left_to_read = %lld\n" "Window_offset      = %lld\n"
			"Window_len         = %lld\n" "Addr type          = %s\n", 
			decode_mover_state(cur_st),
			decode_mover_pause_reas(mover->pauseReas), 
			decode_mover_halt_reas(mover->haltReas), 
			mover->recSiz, mover->recNum, mover->dataWritten, mover->seekPos, 
			mover->bytesLeftToRead, mover->wdwOffset, mover->wdwLen,
			(mover->dataAddr.addr_type == NDMP_ADDR_LOCAL)? "local": "tcp"));

	if (mover->dataAddr.addr_type == NDMP_ADDR_TCP)
	{
		gs_rsp->data_conn_addr.ndmp_addr_u.tcp_addr.ip_addr = 
				mover->dataAddr.ndmp_addr_u.tcp_addr.ip_addr;
		gs_rsp->data_conn_addr.ndmp_addr_u.tcp_addr.port = 
				mover->dataAddr.ndmp_addr_u.tcp_addr.port;
	}	

	return NFX_OK;

} /* fill_Mover_get_state_info */

/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_getState
 *
 * Description:		Process MOVER_GET_STATE request for both v2 and v3.
 *					The difference between the structures of the 2 versions
 *					is the additional field in v3 (data_conn_addr).  So we
 *					call fill_Mover_get_state_info() to fill in everything
 *					and just extract that field out for v2.
 *
 * Created by:      Anh Vu
 *
 * Data Created:    9/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_getState(esm_handle_t handle,
                    int32       nextState, /* next state from table       */
                    esm_event_t thisEvent, /* event driving this action   */
                    void        *userInfo)  /* ctrl blk linked to sm          */
{
	ndmp_NFXReq_t   			*req = (ndmp_NFXReq_t *) thisEvent.p1.u.p;
    ndmp_NFXRsp_t 				*rsp = NULL;
	ndmp_mover_get_state_rsp 	gs_rsp;
    NdmpNIMService_t 			*servp = (NdmpNIMService_t *) userInfo;
	Mover_t		  				*mover = &servp->Mover;	
	int32           			rc, dst_port = servp->ssc_src_port, msglen, 
								newState = nextState, gs_len; 
	

    //NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_getState:  Begin, sessid %d, "
	//			"esm_hdl %d\n", servp->sessId, handle));

	gs_len = (servp->version == NDMP_VER_2) ? sizeof(ndmpv2_mover_get_state_rsp): 
			sizeof(ndmp_mover_get_state_rsp);

	fill_Mover_get_state_info(mover, &gs_rsp, handle);

	msglen = sizeof(ndmp_NFXRsp_t) +  gs_len;
	NDMP_FP_ALLOC(rsp, ndmp_NFXRsp_t, msglen, "nfxrsp MOVER_GS");
	if (rsp == NULL)
	{
		ndmp_sendCmdRsp(req, dst_port, NDMP_UNDEFINED_ERR, 0);
        newState = esm_getCurrState(handle);
		goto ndmp_MoverConnExit;
	}
    rsp->hdr.sessId  = req->hdr.sessId;
    rsp->hdr.mid     = req->hdr.mid;
    rsp->ndmp_mid    = req->ndmp_mid;
	memcpy(&rsp->cl_hdr, &req->cl_hdr, sizeof(ndmp_Hdr));

	if (servp->version == NDMP_VER_2)
	{ 
		memcpy(rsp->content, (char *) &gs_rsp, (sizeof(gs_rsp) - sizeof(ndmp_addr))); 		
	} else {
		memcpy(rsp->content, (char *) &gs_rsp, sizeof(gs_rsp));
	}	

    rc = eee_sendMessage((void *)rsp, msglen, dst_port, ndmp_NNIMCtrl.src_port);
	if (rc != NFX_OK)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_getState: eee_sendMessage failed! sessid %d\n", servp->sessId));
	    E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
	           ("ndmp_Mover_getState: eee_sendMessage failed! sessid %d\n", servp->sessId));
	} else {
#ifdef NFX_SMP 
		ATOMIC_INC_INT(&ndmp_fp_msgStats[SSC_MOVER_GET_STATE_RSP_SNT], 1);
#else
		ndmp_fp_msgStats[SSC_MOVER_GET_STATE_RSP_SNT]++;
#endif
	}
	NDMP_FP_FREE(rsp, "nfxrsp MOVER_GS");

ndmp_MoverConnExit:
	NDMP_FP_FREE(req, "msgp MOVER_GS");    /* it was originally msgp in ndmp_NNIMReceiveMsg() */
    return newState;

} /* ndmp_Mover_getState */
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_setInfo
 *
 * Description:		Perform SET_RECSIZE and SET_WINDOW request
 *
 * Created by:      Anh Vu
 *
 * Data Created:    9/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_setInfo(esm_handle_t handle,
                    int32       nextState, /* next state from table       */
                    esm_event_t thisEvent, /* event driving this action   */
                    void        *userInfo)  /* ctrl blk linked to sm          */
{
    ndmp_NFXReq_t 		*req = (ndmp_NFXReq_t *) thisEvent.p1.u.p;
	ndmp_mover_set_record_size_req	*setRec_req = 
					(ndmp_mover_set_record_size_req *) req->content;
	ndmp_mover_set_window_req 		*setWind_req =
					(ndmp_mover_set_window_req *) req->content; 
    NdmpNIMService_t 	*servp = (NdmpNIMService_t *) userInfo;
    Mover_t     		*mover = &servp->Mover;
    int32 				dst_port = servp->ssc_src_port;
	u_long				recSiz;
	uint64				wdLen, wdOffset;

    //NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_setInfo:  Begin, sessid %d, "
	//			"esm_hdl %d\n", servp->sessId, handle));

	switch (req->hdr.mid)
	{
	  	case NDMP_MID_MOVER_SET_RECORD_SIZE:
			if (esm_getCurrState(handle) != NDMP_MOVER_STATE_IDLE)
			{
				ndmp_sendCmdRsp(req, dst_port, NDMP_ILLEGAL_STATE_ERR, 0);
				goto ndmp_Mover_setInfo_exit;
			}

			/* check for valid  record size, must be a multiple of tape block size */
			recSiz = setRec_req->len;

   			//NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_setInfo:  recsiz received %ul.\n", recSiz));

			// we are always operating in VARIABLE block mode

			/* tape is operating in VARIABLE blk size mode, which will be the
		 	 * size of recsize.  Send msg to Tape service of the variable block size 
		 	 */
			mover->recSiz = recSiz;
#ifdef NFX_SMP
			NFX_SPIN_LOCK(&servp->Data.spin);
			servp->Data.recSiz = recSiz;
			NFX_SPIN_UNLOCK(&servp->Data.spin);
#endif


#if 0 //SYNCSORT
			if (setRec_req->len <= NDMP_MAX_MOVR_RECSIZ)
			{			
				ndmp_sendCmdRsp(req, dst_port, NDMP_NO_ERR, recSiz);
			}
#endif
			ndmp_sendCmdRsp(req, dst_port, NDMP_NO_ERR, recSiz);
			
	  		break;

	  	case NDMP_MID_MOVER_SET_WINDOW:

			// This is for V3.  For V4, it has to be in IDLE & PAUSED */
			if ( (esm_getCurrState(handle) != NDMP_MOVER_STATE_LISTEN) && 
				 (esm_getCurrState(handle) != NDMP_MOVER_STATE_PAUSED) )
			{
				ndmp_sendCmdRsp(req, dst_port, NDMP_ILLEGAL_STATE_ERR, 0);
				goto ndmp_Mover_setInfo_exit;
			}
			wdLen = ((uint64)setWind_req->length.high << 32) + setWind_req->length.low;		
			wdOffset = ((uint64)setWind_req->offset.high << 32) + setWind_req->offset.low;		

#if NDMP_V4 // This restriction is for V4, we are currently V3.  Leave this in for later - avu
			/* record size must be set prior to setting window size */
			if (mover->recSiz == -1)
			{
				ndmp_sendCmdRsp(req, dst_port, NDMP_ILLEGAL_STATE_ERR, 0);
				goto ndmp_Mover_setInfo_exit;
			} 
			/* Window LENGTH has to be nonzero OR  
			 * window len + offset not causes overflow */	
			// do we care overflow here?
			//if ( !wdLen ||  ((wdLen + wdOffset) > NDMP_MOVER_MAX_WDLEN) )
			if (!wdLen)
			{
				ndmp_sendCmdRsp(req, dst_port, NDMP_ILLEGAL_ARGS_ERR, 0);
				goto ndmp_Mover_setInfo_exit;
			}
#endif
			ndmp_sendCmdRsp(req, dst_port, NDMP_NO_ERR, 0);
			mover->wdwLen = wdLen;
			mover->wdwOffset = mover->curpos = wdOffset;
#if NDMP_V4
			mover->recNum = wdOffset/(mover->recSiz);
#endif
			break;
		default:
			break;
	}
	
ndmp_Mover_setInfo_exit:
    NDMP_FP_FREE(req, "msgp SETINFO");
    return nextState;

} /* ndmp_Mover_setInfo */
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_readData
 * 
 * Description:     Client tells Mover to start reading from tape and
 *                  put backup stream onto data connection (Remote Restore)
 *                  or to FS (local Restore).  Issued
 *                  during RECOVERY operation.  Multiple READs at one time are
 *                  not allowed.  
 *                  For LOCAL restore, this read request comes from FS itself,
 *					and relayed by Data server.
 * 
 * Created by:      Anh Vu
 * 
 * Data Created:    9/17/01
 * 
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_readData(esm_handle_t handle,
                    int32       nextState,  // next state from table    
                    esm_event_t thisEvent,  // event driving this action
                    void        *userInfo)  // ctrl blk linked to sm
{
    NdmpNIMService_t        *servp = (NdmpNIMService_t *) userInfo;
    Mover_t                 *mover = &servp->Mover;
    int32                   i, dst_port = servp->ssc_src_port;
    ndmp_Err                error = NDMP_NO_ERR;
    uint64                  offset, len;         
								// where to begin reading in tape
    ndmp_NFXReq_t           *req = (ndmp_NFXReq_t *) thisEvent.p1.u.p;
    ndmp_fs_info_t          *rd_info = (ndmp_fs_info_t *) thisEvent.p2.u.p;
    ndmp_mover_read_req     *rd_req = (ndmp_mover_read_req *) req->content;
	int32					newState = nextState;
	uint64					ioCookie = rd_info->ioCookie;
#if NDMP_V4
	esm_event_t				esmEvent;
#endif

    //NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_readData:  Begin, sessid %d, "
	//		"esm_hdl %d\n", servp->sessId, handle));

    /* A read's in progress or different mode was set during data conn
     * establishment,  return error.  Mode_write means write to data conn.
     */

	if (mover->addr_type != NDMP_ADDR_LOCAL)
	{
		if ((mover->bytesLeftToRead) || (mover->mode != NDMP_MOVER_MODE_WRITE))
		{ 
			error = NDMP_ILLEGAL_STATE_ERR;
		}
		/* Reply to NDMP client of this READ request & continue */
		ndmp_sendCmdRsp(req, dst_port, error , 0);		
	}
	offset = ((uint64) rd_req->offset.high << 32) + rd_req->offset.low;
	len = ((uint64) rd_req->length.high << 32) + rd_req->length.low;

	if (len < mover->recSiz)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_readData:  len to read %lld < recSiz %d\n", len, mover->recSiz));
		E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
						("ndmp_Mover_readData:  len to read %lld < recSiz %d\n", len, mover->recSiz));
		newState = esm_getCurrState(handle);
        NDMP_FP_FREE(req, "msgp MOVER_READ_DATA");

		return newState;
	}
    NDMP_FP_FREE(req, "msgp MOVER_READ_DATA");
    rd_req = NULL;  // Don't use the pointers from here.

    mover->maxIoCookieReq = ioCookie + rd_info->numIOs;
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_readData:  ioCookie rcvd %lld, numIOs %d, maxIoCookie = %lld\n",
					ioCookie, rd_info->numIOs, mover->maxIoCookieReq));

#if NDMP_V4 /*--------------------------------*/
	/* If offset to read is not within current window,
	 * send Pause event to send NOTIFY_MOVER_PAUSE to client.  It will 
	 * proceed to PAUSED state.  
	 */ 
	
	if ((offset > mover->wdwOffset + mover->wdwLen) ||
		(offset < mover->wdwOffset))
	{

		/* Save request so we can reissue it in case we get into PAUSED state
	 	 * for some reason due to this read request.
	 	 */
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_readData:  Send pause Seek.  Read not w/in "
				"cur window.  Offset = %lld, Mover wdwoffset = %lld, "
				"Mover wdwLen = %lld.\n", offset, mover->wdwOffset, mover->wdwLen));
		mover->re_request_rd = TRUE;

		mover->pauseReas =  NDMP_MOVER_PAUSE_SEEK;
		mover->seekPos = offset;		// 1st byte that client asks Mover to move data to data conn.
        bzero(&esmEvent, sizeof(esm_event_t));
        esmEvent.event = NDMP_MOVER_EVENT_PAUSE_INDIC;
        esm_deliverEvent(handle, &esmEvent);

		return newState;
	}

	/* Otherwise, we are w/in the window,  read a 
	 * tape record at a time.
	 */
	mover->bytesLeftToRead = len;

#endif /* NDMP_V4 --------------------------------*/

	/* Tell SCSI to read the data for us */
	if (ndmp_perf_meas)
	{
		eeeRTC_t duration;

		eee_rtcDuration(&time_rd_req, &duration);
		pm_ioCookie = ioCookie + rd_info->numIOs;
   		printf("NDMP:  Time start SCSI Read request = %lld, ioCookie rsp to end %llu\n", 
			   eee_rtcToUsec(&time_rd_req), pm_ioCookie);
		printf("NDMP:  Time taken for SCSI Read request = %lld\n",
			   eee_rtcToUsec(&duration));
	}
	if (!mover->wait_to_pause)
	{
		for (i = mover->nextIoCookieSend;
         		i < mover->maxIoCookieReq && mover->scsiioPendReq < MAX_SCSIIO_REQ;
         		i++)
		{
			if (ndmp_sendSCSIRdWrCmd(SCSI_MID_READ_REQ, i, servp, 0) == NFX_ERR)
			{
				ndmp_notify_data_srvr(servp, NULL, NDMP_STAT_IO_ERR, i);

				/* Close the connection and send an HALTED_NOTIFICATION */
				ndmp_close_conn(servp->sessId, mover, NDMP_MOVER_HALT_INTERNAL_ERROR, 
							servp->ssc_src_port, "read i/o to SCSI failed");

            	/* Force the state machine to go to halted state */
            	newState = NDMP_MOVER_STATE_HALTED;
				break;

			} else {
#ifdef NFX_SMP 
				ATOMIC_INC_INT(&ndmp_fp_msgStats[SCSI_RD_DATA_REQ_SNT], 1);
#else
				ndmp_fp_msgStats[SCSI_RD_DATA_REQ_SNT]++;
#endif
            	mover->scsiioPendReq++;
		    }
		}
    	mover->nextIoCookieSend = i;
	}
	NDMP_FP_FREE(rd_info, "rd_info MOVER_READ_DATA");

	return newState;

} /* ndmp_Mover_readData */


#if 0
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_readData
 *
 * Description:		Client tells Mover to start reading from tape and
 *					put backup stream onto data connection (Remote Restore)
 *					or to FS (local Restore).  Issued
 *					during RECOVERY operation.  Multiple READs at one time are
 *					not allowed.  
 *					For LOCAL restore, this read request could be from FS itself.
 *
 * Created by:      Anh Vu
 *
 * Data Created:    9/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_readData(esm_handle_t handle,
                    int32       nextState, 	// next state from table      
                    esm_event_t thisEvent,  // event driving this action   
                    void        *userInfo)  // ctrl blk linked to sm          
{
    ndmp_mover_read_req   	*rd_req = req->content;
    NdmpNIMService_t   		*servp = (NdmpNIMService_t *) userInfo;
    Mover_t         		*mover = &servp->Mover;
    int32           		dst_port = servp->ssc_src_port, count = 0, msglen;
	ndmp_err				error = NDMP_NO_ERR;
	int64					offset;			// where to begin reading in tape
	ndmp_NFXReq_t   		*rd_req = (ndmp_NFXReq_t *) thisEvent.p1.u.p;
	ndmp_NFXReq_t			*tp_sp_req = NULL;		// send msg to SSC to ask Tape driver
											// to position tape.
	ndmp_tape_mtio_req		*op_req;
	ndmp_tape_mtio_op		tape_op;
    esm_event_t             esmEvent;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_read:  Begin, sessid %d, "
				"esm_hdl %d\n", servp->sessId, handle));

	/* A read's in progress or different mode was set during data conn
	 * establishment,  return error.  Mode_write means write to data conn. 
	 */

	if (mover->bytesLeftToRead) 
	{ 
		error = NDMP_READ_IN_PROGRESS_ERR;
	}
	if (mover->mode != NDMP_MOVER_MODE_WRITE)
	{
		error = NDMP_ILLEGAL_STATE_ERR;
	}
	/* Reply to NDMP client of this READ request */
	ndmp_sendCmdRsp(rd_req, dst_port, error , 0);		

	offset = rd_req->offset;
	len = rd_req->length;


	/* If offset to read is not within current window,
	 * send Pause event to send NOTIFY_MOVER_PAUSE to client.  It will 
	 * proceed to PAUSED state.  
	 */ 
	
	if ((offset > mover->wdwOffset + mover->wdwLen) ||
		(offset < mover->wdwOffset))
	{
		mover->pauseReas =  NDMP_MOVER_PAUSE_SEEK;
		mover->seekPos = offset;
        esmEvent.event = NDMP_MOVER_EVENT_PAUSE_INDIC;
        esmEvent.p1.u.p = NULL;
        esmEvent.p2.u.p = NULL;
        esm_deliverEvent(handle, &esmEvent);
    	return nextState;
	}
	/* Otherwise, we are w/in the window, we shall position
     * the tape to the nearest record boundary prior to the 
	 * desired offset to read. 
	 */

	/* See if we have to Forward Space Record(s) */
	if (offset >= (mover->currpos + mover->recSiz))
	{
		/* Note: if offset is w/in this record, it's ok */

		tape_op = NDMP_MTIO_FSR;
		/* figure out how many records we suppose to forward space */
		count = (offset - mover->currpos)/mover->recSiz;	

		/* Set our current tape position to be at a Record boundary */
		mover->currpos += mover->recSiz * count;
					
	} else if (offet <mover->currpos)	
	{
		/* or Backward Space Record(s) */
	
		 tape_op = NDMP_MTIO_BSR;

		/* figure out how many records we suppose to backward space */
		tmp = (mover->currPos - offset);
		count = (tmp/mover->recSiz) + (tmp % mover->recSiz)? 1: 0;
		mover->currpos -= mover->recSiz * count; 
	} else {
		/* Offset is w/in current record.
		 * Set current tape position to Record boundary
		 */
		mover->currpos -= mover->currpos % mover->recSiz;
	}
	if (count)
	{
		/* Send message to Tape Server to move tape position */
		msglen = sizeof(ndmp_NFXReq_t) + sizeof(ndmp_tape_mtio_req);
        NDMP_FP_ALLOC(tp_sp_req, ndmp_NFXReq_t, msglen, "nfxreq");
        if (tp_sp_req == NULL)
		{
			NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_read:  Malloc erro, "
					"sessid %d, esm_hdl %d\n", servp->sessId, handle));

			/* Reply to NDMP client of this READ request */
			ndmp_sendCmdRsp(rd_req, dst_port, NDMP_UNDEFINED_ERR , 0);		

			// should we abort everything here?
			return esm_getCurrState(handle);
		}
		tp_sp_req->hdr.mid = NDMP_MID_MOVER_TAPE_SPACE_REQ;
		tp_sp_req->hdr.sessId = req->hdr.sessId;
		op_req = (ndmp_tape_mtio_req *) tp_sp_req->content;
		op_req->tape_op = tape_op;
		op_req->count = count;

		rc = eee_sendMessage((void *) tp_sp_req, msglen, dst_port, ndmp_NNIMCtrl.src_port);	
		if (rc != NFX_OK)
		{
			NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_read: eee_sendMessage failed! sessid %d\n", servp->sessId));
	    	E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
	           		("ndmp_Mover_read: eee_sendMessage failed! sessid %d\n", servp->sessId));
		} else {
#ifdef NFX_SMP 
			ATOMIC_INC_INT(ndmp_fp_msgStats[SSC_MOVER_RD_RSP_SNT], 1);
#else
			ndmp_fp_msgStats[SSC_MOVER_RD_RSP_SNT]++;
#endif
		}
        NDMP_FP_FREE(tp_sp_req, "msgp MOVER_READ");

		return esm_getCurrState(handle);
	} else {
		/* Tell SCSI to read the data for us */
		ndmp_sendSCSIRdWrCmd(SCSI_MID_READ_REQ, servp);
	}	
    return nextState;

} /* ndmp_Mover_read */
#endif //0
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_notify_data_srvr
 *
 * Description:     
 *
 * Created by:      Anh Vu
 *
 * Data Created:    11/14/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int32
ndmp_notify_data_srvr(NdmpNIMService_t *servp, void *data, int32 status, uint64 ioCookie)
{
	Mover_t *mover = &servp->Mover;
	ndmp_fs_info_t			*fsInfo;
    esm_event_t             esmEvent;

    bzero(&esmEvent, sizeof(esm_event_t));
    esmEvent.event = (servp->version == NDMP_VER_2)? NDMP_DATA_V2_EVENT_IO_DONE : 
							NDMP_DATA_EVENT_IO_DONE;
    esmEvent.p1.u.p = data;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_notify_data_srvr: status %s, ioCookie %lld\n", 
				(status==NDMP_STAT_IO_ERR)? "NDMP_STAT_IO_ERR":"NDMP_STAT_IO_OK", ioCookie));

	/* Notify Data server a read/write request is done */
	NDMP_FP_ALLOC(fsInfo, ndmp_fs_info_t, sizeof(ndmp_fs_info_t), "notify_data_srvr fs_info");
    if (fsInfo != NULL)
    {
	    fsInfo->fs_status = status;
	    fsInfo->ioCookie = ioCookie;
        esmEvent.p2.u.p = fsInfo;
    }
    if (esm_deliverEvent(mover->data_smHdl, &esmEvent) != NFX_OK)
    {
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_notify_data_srvr: ERROR, cannot "
			"post event NDMP_DATA_EVENT_IO_DONE to SM hdl %d, sessId %d", 
			mover->data_smHdl, servp->sessId));
		E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
				("ndmp_notify_data_srvr: ERROR, cannot "
			"post event NDMP_DATA_EVENT_IO_DONE to SM hdl %d, sessId %d", 
			mover->data_smHdl, servp->sessId));
        NDMP_FP_FREE(fsInfo, "notify_data_srvr fs_info");
		return NFX_ERR;
    }
	return NFX_OK;
}
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_rdDataDone
 * 
 * Description:     Read the data from the tape (Local Restore) or the 
 *					data connection	(Remote Restore) and hand it over to FS.
 *					For Local Restore:
 *					-----------------
 *					A response from SCSI driver corresponding to
 *					us issueing it a rd request will invokes this routine.
 *					We only handle local for now.
 *					
 *					For Remote Restore:
 *					------------------
 *					A TPL callback will invoke this function.
 * 
 * Created by:      Anh Vu
 * 
 * Data Created:    10/17/01
 * 
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_rdDataDone(esm_handle_t handle,
                    int32       nextState,  // next state from table
                    esm_event_t thisEvent,  // event driving this action
                    void        *userInfo)  // ctrl blk linked to sm
{   
    NdmpNIMService_t        *servp = (NdmpNIMService_t *) userInfo;
    Mover_t                 *mover = &servp->Mover;
#if 0 // EEE_BUFF_REST
    scsi_readRspMsg_t       *s_msg = (scsi_readRspMsg_t *) thisEvent.p1.u.p;
#endif
    ndmp_NFXScsiRdRsp       *s_msg = (ndmp_NFXScsiRdRsp *) thisEvent.p1.u.p;
	uint64					ioCookie = (uint64) thisEvent.p2.u.dw;
	int32					i;
    esm_event_t             esmEvent;
    char                    reason[256];


#ifdef NFX_SMP 
	ATOMIC_INC_INT(&ndmp_fp_msgStats[SCSI_RD_DATA_RSP_RCVD], 1);
#else
	ndmp_fp_msgStats[SCSI_RD_DATA_RSP_RCVD]++;
#endif
	bzero(reason, sizeof(reason));
    bzero(&esmEvent, sizeof(esm_event_t));

	if (s_msg == NULL)
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_rdDataDone: Rd rsp is NULL, sessId %d, ioCookie %lld\n", 
								servp->sessId, ioCookie));
        E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
						("ndmp_Mover_rdDataDone: SCSI rd msg is NULL, sessId %d, ioCookie %lld\n", 
						servp->sessId, ioCookie));
		strcpy(reason, "internal err, NULL SCSI rsp");
        goto rd_datadone_err_condition;
    }

	if (ndmp_perf_meas)
	{
		eeeRTC_t duration;

		eee_rtcDuration(&time_rd_req, &duration);
		if (ioCookie == pm_ioCookie)
		{
   			printf("NDMP:  Time start SCSI Read rsp = %lld\n",
				   eee_rtcToUsec(&time_rd_req));
   			printf("NDMP:  Time taken for SCSI Read request, ending w ioCookie %llu = %lld\n", 
				   pm_ioCookie,
				   eee_rtcToUsec(&duration));
		}
	}
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_rdDataDone:  sessid %d, esm_hdl %d, ecode %s\n"
		"Data Length  : %d\n", servp->sessId, handle, decode_scsi_ecode(s_msg->eCode),
		s_msg->dataLen));

    mover->scsiioPendReq--;
	if (mover->wait_to_pause == TRUE)
	{
		if (mover->scsiioPendReq == 0)
    	{
	    	/* Send a PAUSE INDIC event */
        	esmEvent.event = NDMP_MOVER_EVENT_PAUSE_INDIC;
        	esm_deliverEvent(servp->Mover.my_smHdl, &esmEvent);
		}
        free_eee_buf_chain(s_msg->eBufptr);
		NDMP_FP_FREE(s_msg, "msgp scsi READ rsp");

		return nextState;
    }
    switch (s_msg->eCode)
    {
        case SCSI_MSG_ECODE_OK:
        case SCSI_MSG_ECODE_ILI:

            /* Hand the data over to Data server who
			 * will either gives it to data connection (for remote restore)
			 * or FS (for local restore)
			 */

            mover->lastIoCookieRsp = ioCookie;
			mover->bytesLeftToRead -= s_msg->dataLen; // update w/ # bytes read	
			mover->curpos += s_msg->dataLen;
            if (s_msg->dataLen == 0)
            {
                NDMP_DBG_PRINTF(D_INFO, ("SCSI return code OK, but data len 0, sessid %d, ioCookie %lld\n",
								servp->sessId, ioCookie));
                E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
								("SCSI return code OK, but data len 0, sessid %d, ioCookie %lld\n",
								servp->sessId, ioCookie));
				strcpy(reason, "Rd rsp with 0 data_len");
				break;  	//to process error case
            }

#ifdef DBG_BKUP_DATA
			NDMP_DBG_PRINTF(D_DBG, ("ndmp_Mover_rdDataDone:  This is what we get from SCSCI read:\n"));
			//dbg_print_rec(((char *)(s_msg) + sizeof(scsi_readRspMsg_t)), s_msg->dataLen);
#endif

			/* Note on ILI return code:  If requested size > actual blk size, SCSI will
			 * return actual blksiz, residue has to be calculated by NDMP if it needs it.
			 * If requested size < actual blk size, SCSI will return requested size.
			 * In any case residue is not calculated for you by SCSI.		As of 01/03/02
			 */

			/* Now tell Data server of our result */
			if (ndmp_notify_data_srvr(servp, s_msg, NDMP_STAT_IO_OK, ioCookie) == NFX_ERR)
            {
                free_eee_buf_chain(s_msg->eBufptr);
        		NDMP_FP_FREE(s_msg, "scsi s_msg rsp");
            }
			/* Issue some more i/o requests */
			for (i = mover->nextIoCookieSend;
         		i < mover->maxIoCookieReq && mover->scsiioPendReq < MAX_SCSIIO_REQ;
         		i++)
			{
				if (ndmp_sendSCSIRdWrCmd(SCSI_MID_READ_REQ, i, servp, 0) == NFX_ERR)
				{
					ndmp_notify_data_srvr(servp, NULL, NDMP_STAT_IO_ERR, i);

					/* Close the connection and send an HALTED_NOTIFICATION */
					ndmp_close_conn(servp->sessId, mover, NDMP_MOVER_HALT_INTERNAL_ERROR, 
							servp->ssc_src_port, "read i/o to SCSI failed");

            		/* Force the state machine to go to halted state */
            		nextState = NDMP_MOVER_STATE_HALTED;
					return nextState;
				} else {
#ifdef NFX_SMP 
					ATOMIC_INC_INT(&ndmp_fp_msgStats[SCSI_RD_DATA_REQ_SNT], 1);
#else
					ndmp_fp_msgStats[SCSI_RD_DATA_REQ_SNT]++;
#endif
            		mover->scsiioPendReq++;
				}
			}
    		mover->nextIoCookieSend = i;

			return nextState;

        case SCSI_MSG_ECODE_FM_ENCOUNTERED:
        case SCSI_MSG_ECODE_MEDIUM_ERROR:
        case SCSI_MSG_ECODE_END_OF_MEDIUM:

			NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_rdDataDone:  Got error %s from SCSI, sessid %d, ioCookie %lld\n",  
						decode_scsi_ecode(s_msg->eCode), servp->sessId, ioCookie));
			E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
						("ndmp_Mover_rdDataDone:  Got error %s from SCSI, sessid %d, ioCookie %lld\n",  
						decode_scsi_ecode(s_msg->eCode), servp->sessId, ioCookie));
            /* Queue up the Pause Indication event to our Mover state machine */
           	mover->pauseReas = 
					(s_msg->eCode == SCSI_MSG_ECODE_END_OF_MEDIUM)?
                    	NDMP_MOVER_PAUSE_EOM: 
					(s_msg->eCode == SCSI_MSG_ECODE_MEDIUM_ERROR)?
						NDMP_MOVER_PAUSE_MEDIA_ERROR : 
						NDMP_MOVER_PAUSE_EOF;
			
			/* Make sure we reissue the read once we get out of PAUSED state */
            if (mover->re_request_rd != TRUE)
            {
                mover->nextIoCookieSend = ioCookie;
			    mover->re_request_rd = TRUE;
			    mover->re_request_rd_ioCookie = ioCookie;
            }
            if (mover->scsiioPendReq == 0)
            {
                /* Send a PAUSE INDIC event */
                esmEvent.event = NDMP_MOVER_EVENT_PAUSE_INDIC;
                esm_deliverEvent(servp->Mover.my_smHdl, &esmEvent);
            }
			mover->wait_to_pause = TRUE;

			/* Don't reply to FS yet, just resend the IO read request once
			 * NDMP client signals us to CONTINUE.  We'll enter a PAUSED state
			 * after the above event is sent.
			 */
            free_eee_buf_chain(s_msg->eBufptr);
            NDMP_FP_FREE(s_msg, "msgp scsi READ rsp");

			return nextState;

		default:
            mover->lastIoCookieRsp = ioCookie;
			strcpy(reason, decode_scsi_ecode(s_msg->eCode));

			NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_rdDataDone:  Invalid SCSI "
						"return code %s, sessid %d, ioCookie %lld\n", 	
						reason, servp->sessId, ioCookie));
			E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
							("ndmp_Mover_rdDataDone:  Invalid SCSI "
						"return code %s, sessid %d, ioCookie %lld\n", 	
						reason, servp->sessId, ioCookie));
    }

rd_datadone_err_condition:

	/* Handle error condition */
    NDMP_FP_FREE(s_msg, "msgp scsi rsp");

	/* Close the connection and send a HALTED_NOTIFICATION */
	ndmp_close_conn(servp->sessId, mover, 
				NDMP_MOVER_HALT_INTERNAL_ERROR, servp->ssc_src_port, reason);

    /* Force the state machine to go to halted state */
    nextState = NDMP_MOVER_STATE_HALTED;

	/* We'll proceed and reply to FS of this bad error so it can clean up */
	ndmp_notify_data_srvr(servp, NULL, NDMP_STAT_IO_ERR, ioCookie); 


	return nextState;

} /* ndmp_Mover_rdDataDone */
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_conn_close
 *
 * Description:     If doing a backup, flush data to tape if there's any and then sends
 *					the client a notify_mover_halted msg.
 *                  This func can be signalled by the Data Server for Local
 *                  backup or by TPL for remote backup, i.e. the Data server
 *					is done and closes the Data Connection to Mover.
 *
 * Created by:      Anh Vu
 *
 * Data Created:    11/14/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_conn_close(esm_handle_t handle,
                    int32       nextState, /* next state from table       */
                    esm_event_t thisEvent, /* event driving this action   */
                    void        *userInfo)  /* ctrl blk linked to sm          */
{
    NdmpNIMService_t 	*servp = (NdmpNIMService_t *) userInfo;
    Mover_t     		*Mover = &servp->Mover;
	int32            	flush_flag = thisEvent.p2.u.w;

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_conn_close:  Begin, sessid %d, "
					"esm_hdl %d, flush flag = %d\n", 
					servp->sessId, handle, flush_flag));

	/*
	 * Create a NOTIFY_MOVER_HALTED message and send it.
	 */
	if (ndmp_close_conn(servp->sessId, Mover, 
		NDMP_MOVER_HALT_CONNECT_CLOSED, servp->ssc_src_port, NULL) == NFX_ERR)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_conn_close:  Close connection error, sessId %d\n", servp->sessId));
		E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
					("ndmp_Mover_conn_close:  Close connection error, sessId %d\n", servp->sessId));
		return esm_getCurrState(handle);
	}

	return nextState;

} /* ndmp_Mover_conn_close */

//if DMP_PERF
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_sndIOReqstoSCSI
 *
 * Description:		Send up to max that scsiio can handle the requests to SCSI.
 *
 * Created by:      Anh Vu
 *
 * Data Created:    09/21/02
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_sndIOReqstoSCSI(NdmpNIMService_t *servp, int32 *nextState)
{
	Mover_t 			*Mover = &servp->Mover;
	ndmp_ioReqNode_t  	*req_node; 
	int 				i;
#if 0
    printf("ndmp_sndIOReqstoSCSI:  nextIoCookieSend = %d, maxIoCookieReq = %d, "
                    "Mover->scsiioPendReq = %d\n", Mover->nextIoCookieSend, 
                    Mover->maxIoCookieReq, Mover->scsiioPendReq);
#endif
    if (ndmp_perf_meas)
    {
        Mover->wr_pend += Mover->scsiioPendReq;
    }
	for (i = Mover->nextIoCookieSend;
			(i < Mover->maxIoCookieReq) && (Mover->scsiioPendReq < MAX_SCSIIO_REQ); i++)
	{
		/* remove the pending i/o from the q, and send it */
		req_node = ndmp_getNextFromQ(Mover);	//get next req from head of q to send 
		if (req_node == NULL)
			break;

       	if (ndmp_sendSCSIRdWrCmd(SCSI_MID_WRITE_REQ, req_node->ioCookie, servp, req_node->buf) 
						== NFX_ERR)
		{
			ndmp_notify_data_srvr(servp, NULL, NDMP_STAT_IO_ERR, i);

			/* Close the connection and send an HALTED_NOTIFICATION */
			ndmp_close_conn(servp->sessId, Mover, NDMP_MOVER_HALT_INTERNAL_ERROR, 
							servp->ssc_src_port, "read i/o to SCSI failed");

            /* Force the state machine to go to halted state */
            *nextState = NDMP_MOVER_STATE_HALTED;
			return NFX_ERR;

		} else {

#ifdef DBG_BKUP_DATA
       		NDMP_DBG_PRINTF(D_DBG, ("This is what we write to SCSCI :\n"));
			dbg_print_rec(req_node->buf, Mover->recSiz);
#endif
#ifdef NFX_SMP 
			ATOMIC_INC_INT(&ndmp_fp_msgStats[SCSI_WR_DATA_REQ_SNT], 1);
#else
			ndmp_fp_msgStats[SCSI_WR_DATA_REQ_SNT]++;
#endif
			Mover->scsiioPendReq++;
       		NDMP_DBG_PRINTF(D_INFO, ("Send to SCSIIO request ioCookie %lld, Pendq = %d\n", 
									req_node->ioCookie, Mover->scsiioPendReq));
	        if (ndmp_perf_meas)
	        {
                eee_TimeGMT(&req_node->sent_time); 
            }
		}
	}
#if 0
   		printf("NDMP:  Sent %d io Write Reqs (%d -> %d).\n", (i - Mover->nextIoCookieSend),
                        Mover->nextIoCookieSend, i-1);
        printf("Time start SCSI requests = %d\n", time_wr_req.usec);
#endif
	Mover->nextIoCookieSend = i;

	return NFX_OK;

} //ndmp_sndIOReqstoSCSI 

//end DMP_PERF

/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_writeData
 *
 * Description:		Write to tape device for Local Backup and to the
 *					data connection for Remote Backup.  This func
 *					can be signalled by the FS for the initial write
 *					of a buffer.  Or by the processing of write completion
 *					signalled by SCSI when we need to do multiple SCSI writes
 *					for each FS write request.
 *
 * Created by:      Anh Vu
 *
 * Data Created:    10/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_writeData(esm_handle_t handle,
                    int32       nextState, /* next state from table       */
                    esm_event_t thisEvent, /* event driving this action   */
                    void        *userInfo)  /* ctrl blk linked to sm          */
{
    ndmp_write_data_req_msg_t 		*msgp = (ndmp_write_data_req_msg_t *) thisEvent.p1.u.p;
    NdmpNIMService_t 	*servp = (NdmpNIMService_t *) userInfo;
    Mover_t     		*Mover = &servp->Mover;
	uint64				ioCookie = msgp->ioCookie;

	/* NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_writeData:  Begin, sessid %d, "
				"esm_hdl %d, ioCookie = %lld, buf = %p\n", servp->sessId, handle,
                ioCookie, msgp->buf));
	*/

    if (Mover->addr_type ==  NDMP_ADDR_LOCAL)
    {
		/* Add the request to queue */
		Mover->maxIoCookieReq = ioCookie + 1;
	    ndmp_addReqToQueue(Mover, ioCookie, msgp->buf);	//add req to end of pending q
        NDMP_FP_FREE(msgp, "msgp WRITE_DATA");

		//We like to wait for all responses from  SCSIIO to come back before
		//we go to PAUSED state (once we get EOM or Media Err), so don't send more
		//to SCSIIO if we're in waiting-to-pause mode.
		if (!Mover->wait_to_pause)
		{
	 		/* Remove a request from the pending q and send it to SCSI to write to Tape.  */
			ndmp_sndIOReqstoSCSI(servp, &nextState);
		}
	}

    return nextState;
} /* ndmp_Mover_writeData */


/*------------------------------------------------------------------------------*/
#if 0	/* Save the below func, we may need for Remote case or when window
		 * feature is used.  This has headachy algo that deserves saving.
		 * Anh - 9/19/02
		 */
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_writeData
 *
 * Description:		Write to tape device for Local Backup and to the
 *					data connection for Remote Backup.  This func
 *					can be signalled by the FS for the initial write
 *					of a buffer.  Or by the processing of write completion
 *					signalled by SCSI when we need to do multiple SCSI writes
 *					for each FS write request.
 *
 * Created by:      Anh Vu
 *
 * Data Created:    10/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_writeData(esm_handle_t handle,
                    int32       nextState, /* next state from table       */
                    esm_event_t thisEvent, /* event driving this action   */
                    void        *userInfo)  /* ctrl blk linked to sm          */
{
    ndmp_write_data_req_msg_t 		*msgp = (ndmp_write_data_req_msg_t *) thisEvent.p1.u.p;
	char		*fs_buf;
    NdmpNIMService_t *servp = (NdmpNIMService_t *) userInfo;
    Mover_t     *Mover = &servp->Mover;
    uint32 		spaceleft, 
				fs_bufSiz = (uint32) thisEvent.p2.u.w;	// current FS buffer len to write
	uint64		ioCookie = msgp->ioCookie;


	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_writeData:  Begin, sessid %d, "
				"esm_hdl %d\n", servp->sessId, handle));

	// If it's a local backup, we write this to tape.  
	// If it's a remote backup, we write this to the TCP connection

	if (fs_bufSiz == NDMP_FS_WR_BUFSIZ)
	{
		/* First time process write for this FS write request 
		 * Subsequent time is due to tape recsiz < NDMP_FS_WR_BUFSIZ
		 */
		//fs_buf = msgp + sizeof(ndmp_fsMsgHdr_t) + sizeof(void *);
		fs_buf = msgp->buf;
		/* If there's data left over in OUR fs_buf, write it to our writeBuf */
		if (Mover->fs_bufSiz)
		{
			memcpy(Mover->writeBuf + Mover->wIndex, 
							Mover->fsBuf + Mover->fs_rIndex, Mover->fs_bufSiz); 
			Mover->wIndex += Mover->fs_bufSiz;
			Mover->fs_rIndex = 0;
		}
		/* Copy fs buf data from the msg to our own space */
		memcpy (Mover->fsBuf, fs_buf, NDMP_FS_WR_BUFSIZ);
		Mover->fs_bufSiz = fs_bufSiz;
        NDMP_FP_FREE(msgp, "msgp WRITE_DATA");
	} else {
		fs_buf = (char *)thisEvent.p1.u.p;
	}
	if (Mover->recSiz >= fs_bufSiz)
	{
    	/* Write fs_bufSiz (NDMP_FS_WR_BUFSIZ = 1k) data to our 
		 * buffer if it fits a recSiz, else
     	 * break it up.   Collect recSiz to write to tape.  Figure out
	 	 * space left in our local buffer (mover->writeBuf).
     	*/
    	spaceleft = Mover->recSiz - Mover->wIndex;
    	if (Mover->recSiz >= Mover->wIndex + fs_bufSiz)
    	{
        	/* this FS buf will fit into our local tape record buffer */

        	memcpy(Mover->writeBuf + Mover->wIndex, Mover->fsBuf, fs_bufSiz);
        	Mover->wIndex += fs_bufSiz;  // keep track of where to write nxt
			Mover->fs_rIndex = 0;				//nothing left in FS buf
			Mover->fs_bufSiz = 0;
    	} else {
        	/* Fill the tape record w/ whatever space left and keep
         	 * the remaining data in our fs_buf for next time.
         	 */
        	memcpy( Mover->writeBuf + Mover->wIndex, Mover->fsBuf, spaceleft);
        	Mover->wIndex += spaceleft;
			Mover->fs_rIndex += spaceleft;				//update where to read next
			Mover->fs_bufSiz = fs_bufSiz - spaceleft;	//what's left in FS buf
    	}   
	} else  {
		/* Tape record size (recSiz) is smaller than fs_bufSiz, we'll have 
		 * to do multiple writes in order to complete FS's request. 
		 */	
		memcpy(Mover->writeBuf, fs_buf, Mover->recSiz);
       	Mover->wIndex += Mover->recSiz;
		Mover->fs_rIndex += Mover->recSiz;
		Mover->fs_bufSiz = fs_bufSiz - Mover->recSiz;
		
	}

	/* Our local tape record buffer is not full yet, notify FS that
 	 * this WRITE req is temporarily done so he can issue another.
	 */
    if (Mover->wIndex < Mover->recSiz)
	{

	    // If it's LOCAL bkup, we signal Data Server, otherwise, TCP TPL layer will
    	// handle the flow control.
    	if (Mover->addr_type ==  NDMP_ADDR_LOCAL)
    	{
			if (ndmp_notify_data_srvr(servp, NULL, NDMP_STAT_IO_OK, ioCookie)
							== NFX_ERR)
			{
				return esm_getCurrState(handle);
        	}
    	}

	} else if (Mover->wIndex == Mover->recSiz)
    {   
    	/* Otherwise, we've accumulated one tape record. 
	 	 * Therefore issue a SCSI write request to Tape.
         * Fire off the SCSI WRITE request 
		 */
        if (ndmp_sendSCSIRdWrCmd(SCSI_MID_WRITE_REQ, ioCookie, servp, buf?) == NFX_ERR)
		{
			ndmp_notify_data_srvr(servp, NULL, NDMP_STAT_IO_ERR, ioCookie);

			/* Close the connection and send an HALTED_NOTIFICATION */
            ndmp_close_conn(servp->sessId, Mover, NDMP_MOVER_HALT_INTERNAL_ERROR, 
							servp->ssc_src_port, "write i/o to SCSI failed");

            /* Force the state machine to go to halted state */
            nextState = NDMP_MOVER_STATE_HALTED;
		}
#ifdef NFX_SMP 
		ATOMIC_INC_INT(ndmp_fp_msgStats[SCSI_WR_DATA_REQ_SNT], 1);
#else
		ndmp_fp_msgStats[SCSI_WR_DATA_REQ_SNT]++;
#endif

#ifdef DBG_BKUP_DATA
        NDMP_DBG_PRINTF(D_DBG, ("ndmp_Mover_writeData:  This is what we write to SCSCI :\n"));
		dbg_print_rec(Mover->writeBuf, Mover->recSiz);
#endif

    }
    return nextState;

} /* ndmp_Mover_writeData */
#endif //0

/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_wrDataDone
 *
 * Description:     This is to process the completion of writing Data
 *					to either the data connection (for Remote BAckup) or
 *					to the tape device (for Local Backup).  
 *
 *					For Local Backup, it processes the SCSI Write Response msg. 
 *					It will issue another write to the tape device
 *					if need be, otherwise it will send FS an IO done
 *					message so FS can give it more data to backup.	
 *					If SCSI returns an error, it will send an event to send a PAUSE
 *					Indication to the client who will eventually send us an
 *					abort msg.
 *
 * Created by:      Anh Vu
 *
 * Data Created:    9/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_wrDataDone(esm_handle_t handle,
                    int32       nextState, /* next state from table       */
                    esm_event_t thisEvent, /* event driving this action   */
                    void        *userInfo)  /* ctrl blk linked to sm          */
{
    scsi_writeRspMsg_t  *s_msg = (scsi_writeRspMsg_t *)thisEvent.p1.u.p;
	uint64				ioCookie = (uint64) thisEvent.p2.u.dw;
    NdmpNIMService_t 	*servp = (NdmpNIMService_t *) userInfo;
    Mover_t     		*Mover = &servp->Mover;
	int32				stat = NDMP_STAT_IO_OK;
    esm_event_t         esmEvent;
    char                reason[256];

    
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_wrDataDone:  SCSI ecode = %s, sessid %d, esm_hdl %d\n",
        (s_msg->eCode == SCSI_MSG_ECODE_OK)? "ok":
        (s_msg->eCode == SCSI_MSG_ECODE_MEDIUM_ERROR)? "medium_err":
        (s_msg->eCode == SCSI_MSG_ECODE_END_OF_MEDIUM)? "eom": "unknown", servp->sessId, handle));

    bzero(&esmEvent, sizeof(esm_event_t));
#ifdef NFX_SMP 
	ATOMIC_INC_INT(&ndmp_fp_msgStats[SCSI_WR_DATA_RSP_RCVD], 1);
#else
	ndmp_fp_msgStats[SCSI_WR_DATA_RSP_RCVD]++;
#endif
	Mover->scsiioPendReq--;
	if (ndmp_perf_meas)
	{
		    eee_TimeGMT(&Mover->ioPendQ.head->rcv_time);	  
   		    Mover->delta_time += (uint64) ((Mover->ioPendQ.head->rcv_time.usec) - 
                    (Mover->ioPendQ.head->sent_time.usec));
            (Mover->io_wr_done) ++;
#if 0
        printf("NDMP:  Got rsp for ioCookie %lld, time = %d\n", ioCookie, time_wr_rsp.usec);
   		printf("NDMP:  Time taken for SCSI Write req from last time queued = %d\n", 
                        time_wr_rsp.usec - time_wr_req.usec);
#endif
    }
	if (Mover->wait_to_pause == TRUE)
	{
		/* Don't remove any requests from the q cause this
		 * is not an OK response, we should wait til we have received
		 * all response before we send the pause.  We should pick
		 * up from the q the last non-ok response to resend.  
		 *
		 * There's no guarantee that after we get a MEDIA_ERR for a rsp that 
		 * subsequent rsps will also be MEDIA_ERR, we know this is true for EOM.  
		 * So it's better to handle this upfront and ignore subsequent
		 * responses that followed an EOM/MEDIA_ERR rsp.
		 */
		if (Mover->scsiioPendReq == 0)
		{
       		esmEvent.event = NDMP_MOVER_EVENT_PAUSE_INDIC;
       		esm_deliverEvent(servp->Mover.my_smHdl, &esmEvent);
			Mover->re_request_wr = TRUE;	  /* Mark so when we CONTINUE, 
										  	   * we will reissue the write  */
			Mover->nextIoCookieSend = Mover->ioPendQ.head->ioCookie;	  
			Mover->ioPendQ.curr = NULL;			/* so when we continue, it'll get the head of q */	
			E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
					("ndmp_Mover_wrDataDone : Send Mover_Pause, next request to reiisue has ioCookie %lld\n", Mover->nextIoCookieSend));
			NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_wrDataDone:  sciioPendReq = 0, dataWritten %lld (Mb)\n", 
								Mover->dataWritten/1000000));
		}
		NDMP_FP_FREE(s_msg, "s_msg WRDATADONE");
		return nextState;
	}

    switch (s_msg->eCode)
    {

        case SCSI_MSG_ECODE_OK:
            // see if we have to do another write, if not, tell FS we're
            // done.
			Mover->wIndex = 0;

           	Mover->dataWritten += Mover->recSiz;
            Mover->recNum = (Mover->dataWritten)/(Mover->recSiz);

			/* Got a good response for this request, remove it from q */
			ndmp_delReqFromQueue(Mover);

           	break;
 
		/* For v3, w/ MEDIA_ERR, we go to PAUSE state, 
		 * for v4, it'll go to HALTED state.
		 */
        case SCSI_MSG_ECODE_MEDIUM_ERROR:
            NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_wrDataDone : SCSI return "
					"SCSI_MSG_ECODE_MEDIUM_ERROR %d, sessid %d\n", 
					s_msg->eCode, servp->sessId));
            E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
					("ndmp_Mover_wrDataDone : SCSI return "
					"SCSI_MSG_ECODE_MEDIUM_ERROR %d, sessid %d\n", 
					s_msg->eCode, servp->sessId));
        case SCSI_MSG_ECODE_END_OF_MEDIUM:
 
            if (s_msg->eCode == SCSI_MSG_ECODE_END_OF_MEDIUM)
            {
                NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_wrDataDone : SCSI "
					"return SCSI_MSG_ECODE_END_OF_MEDIUM %d, ioCookie %lld, pendQ len %d, sessid %d\n", 
					s_msg->eCode, ioCookie, Mover->scsiioPendReq, servp->sessId));
                E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
					("ndmp_Mover_wrDataDone : SCSI "
					"return SCSI_MSG_ECODE_END_OF_MEDIUM %d, ioCookie %lld, pendQ len %d, sessid %d\n", 
					s_msg->eCode, ioCookie, Mover->scsiioPendReq, servp->sessId));

				Mover->wIndex = 0;
            }

            /* Queue up the Pause Indication event to our Mover state machine */
			Mover->wait_to_pause = TRUE;
			Mover->pauseReas = (s_msg->eCode == SCSI_MSG_ECODE_END_OF_MEDIUM)? 
					NDMP_MOVER_PAUSE_EOM: NDMP_MOVER_PAUSE_MEDIA_ERROR;
            if (Mover->scsiioPendReq == 0)
            {
                esmEvent.event = NDMP_MOVER_EVENT_PAUSE_INDIC;
                esm_deliverEvent(servp->Mover.my_smHdl, &esmEvent);
                Mover->re_request_wr = TRUE;  /* Mark so when we CONTINUE, 
                                               * we will reissue the write  
                                               */
                Mover->nextIoCookieSend = Mover->ioPendQ.head->ioCookie;
                Mover->ioPendQ.curr = NULL;  /* so when we continue, it'll
                                              * get the head of q */
                E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
                        ("ndmp_Mover_wrDataDone : Send Mover_Pause, next request to reiisue has ioCookie %lld\n", Mover->nextIoCookieSend));
                NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_wrDataDone:  sciioPendReq = 0, dataWritten %lld (Mb)\n",
                                    Mover->dataWritten/1000000));
            }

            NDMP_FP_FREE(s_msg, "s_msg WRDATADONE");

			/* Don't tell FS it's IO error cause FS will abort the whole ops,
			 * we'll signal FS when client tells us to continue.   
			 */

			return nextState;

		default:
			stat = NDMP_STAT_IO_ERR;	
			Mover->wIndex = 0;

            bzero(reason, sizeof(reason));
            strcpy(reason, decode_scsi_ecode(s_msg->eCode));

			/* Close the connection and send an HALTED_NOTIFICATION */
            ndmp_close_conn(servp->sessId, Mover, 
				NDMP_MOVER_HALT_INTERNAL_ERROR, servp->ssc_src_port, reason);

            /* Force the state machine to go to halted state */
            nextState = NDMP_MOVER_STATE_HALTED;

            NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_wrDataDone : Invalid "
				"SCSI return %s, dataWritten %lld (Mb), ioCookie %lld, sessid %d\n", 
				reason, Mover->dataWritten/1000000, ioCookie, servp->sessId));
            E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
				("ndmp_Mover_wrDataDone : Invalid SCSI return %s, "
				 "dataWritten %lld (Mb), ioCookie %lld, sessid %d\n", 
				reason, Mover->dataWritten/1000000, ioCookie, servp->sessId));

            /* We'll proceed and reply to FS of this bad error so it can clean up */
			
	}
	/* If it's LOCAL bkup, we signal Data Server, otherwise, 
	 * TCP TPL layer will handle the flow control.
	 * Also if we signal Data server about this IO completion
	 * and it happens to get an op_comp from FS and moves
	 * to HALTED state, nothing (noop) will be done.
	 */
	if (Mover->addr_type ==  NDMP_ADDR_LOCAL)
	{
		/* Send reply to FS for its request that SCSI just responded above */
		if (ndmp_notify_data_srvr(servp, NULL, stat, ioCookie) == NFX_ERR)
		{
            NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_wrDataDone : Error "
				"in notifying Data server of an IO rsp, sessid %d\n", servp->sessId));
            E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
				("ndmp_Mover_wrDataDone : Error "
				"in notifying Data server of an IO rsp, sessid %d\n", servp->sessId));
		}
		if (s_msg->eCode == SCSI_MSG_ECODE_OK)
		{
			/* Now send the rest of requests in our q to SCSI */
			ndmp_sndIOReqstoSCSI(servp, &nextState);
		}
	}
    NDMP_FP_FREE(s_msg, "s_msg WRDATADONE");

    return nextState;

} /* ndmp_Mover_wrDataDone */

/*  
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_qReq
 *
 * Description:     Queue the request that we receive from FS during the paused
 *                  state.
 *
 * Created by:      Nitin Jain
 *
 * Data Created:    09/18/02
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_qReq(esm_handle_t handle,
                int32       nextState, /* next state from table       */
                esm_event_t thisEvent, /* event driving this action   */
                void        *userInfo)  /* ctrl blk linked to sm          */
{
    NdmpNIMService_t        *servp = (NdmpNIMService_t *) userInfo;
    Mover_t                 *mover = &servp->Mover;
    ndmp_NFXReq_t           *req = (ndmp_NFXReq_t *) thisEvent.p1.u.p;
    ndmp_fs_info_t          *rd_info = (ndmp_fs_info_t *) thisEvent.p2.u.p;

    if (mover->addr_type ==  NDMP_ADDR_LOCAL)
    {
		if (mover->mode == NDMP_MOVER_MODE_READ)
		{
			//Backup		
    		ndmp_write_data_req_msg_t 	*msgp = 
					(ndmp_write_data_req_msg_t *) thisEvent.p1.u.p;
						
			/* Add the request to queue */
			mover->maxIoCookieReq = msgp->ioCookie + 1;
	    	ndmp_addReqToQueue(mover, msgp->ioCookie, msgp->buf);	//add req to end of pending q

		} else if (mover->mode == NDMP_MOVER_MODE_WRITE)
		{
			//Restore
        	mover->maxIoCookieReq = rd_info->ioCookie + rd_info->numIOs;
        	NDMP_DBG_PRINTF(D_INFO, ("Req. saved in PAUSED state. IoCookie -> lastRsp %lld," 
						"nextSend %lld, maxReq %lld\n", 
						mover->lastIoCookieRsp, mover->nextIoCookieSend, mover->maxIoCookieReq));
		}
    }
    NDMP_FP_FREE(req, "NFXReq");
    NDMP_FP_FREE(rd_info, "rd_info");

    return nextState;
} /* ndmp_Mover_qReq */

/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_save
 *
 * Description:
 *
 * Created by:      Anh Vu
 *
 * Data Created:    9/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_save(esm_handle_t handle,
                    int32       nextState, /* next state from table       */
                    esm_event_t thisEvent, /* event driving this action   */
                    void        *userInfo)  /* ctrl blk linked to sm          */
{
#if 0
    ndmp_NFXRsp_t rsp;
    ndmp_NFXReq_t *msg = (ndmp_NFXReq_t *) thisEvent.p1.u.p;
    NdmpNIMService_t *servp = (NdmpNIMService_t *) userInfo;
    Mover_t     *mover = &servp->Mover;
    int32 		dst_port = thisEvent.p2.u.w;
	ndmp_Err	err;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_save:  Begin, sessid %d, esm_hdl %d\n", servp->sessId, handle));
#endif
    return nextState;

} /* ndmp_Mover_save */
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_continue
 *
 * Description:
 *
 * Created by:      Anh Vu
 *
 * Data Created:    9/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_continue(esm_handle_t handle,
                    int32       nextState, /* next state from table       */
                    esm_event_t thisEvent, /* event driving this action   */
                    void        *userInfo)  /* ctrl blk linked to sm          */
{
    ndmp_NFXReq_t 		*req = (ndmp_NFXReq_t *) thisEvent.p1.u.p;
    NdmpNIMService_t 	*servp = (NdmpNIMService_t *) userInfo;
    Mover_t     		*mover = &servp->Mover;
    int32 				dst_port = servp->ssc_src_port, i;


	/* Check if valid mover window was set or not */
	if (ndmp_chk_precond(mover->mode, mover) == NFX_ERR)
	{
		ndmp_sendCmdRsp(req, dst_port, NDMP_ILLEGAL_STATE_ERR, 0);
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_continue:  Illegate state error, sessId %d",
				servp->sessId));
		E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
				("ndmp_Mover_continue:  Illegate state error, sessId %d", servp->sessId));
		return esm_getCurrState(handle);
	}
	ndmp_sendCmdRsp(req, dst_port, NDMP_NO_ERR, 0); 
    NDMP_FP_FREE(req, "msgp CONTINUE");

	if (mover->re_request_rd == TRUE)
	{
		mover->re_request_rd = FALSE;
		mover->re_request_rd_ioCookie = 0;

	    for (i = mover->nextIoCookieSend;
             i < mover->maxIoCookieReq && mover->scsiioPendReq < MAX_SCSIIO_REQ;
             i++)
	    {
		    if (ndmp_sendSCSIRdWrCmd(SCSI_MID_READ_REQ, i, servp, 0) == NFX_ERR)
		    {
			    ndmp_notify_data_srvr(servp, NULL, NDMP_STAT_IO_ERR, i);

				/* Close the connection and send an HALTED_NOTIFICATION */
				ndmp_close_conn(servp->sessId, mover, NDMP_MOVER_HALT_INTERNAL_ERROR, 
							servp->ssc_src_port, "read i/o to SCSI failed");

            	/* Force the state machine to go to halted state */
            	nextState = NDMP_MOVER_STATE_HALTED;

				return nextState;
		    } else {
#ifdef NFX_SMP 
				ATOMIC_INC_INT(&ndmp_fp_msgStats[SCSI_RD_DATA_REQ_SNT], 1);
#else
				ndmp_fp_msgStats[SCSI_RD_DATA_REQ_SNT]++;
#endif
                mover->scsiioPendReq++;
		    }
	    }
        mover->nextIoCookieSend = i;
	}

	if (mover->re_request_wr == TRUE)	
	{
        E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
					("ndmp_Mover_continue : Reissue request with ioCookie %lld\n", mover->nextIoCookieSend));
		ndmp_sndIOReqstoSCSI(servp, &nextState);
		mover->re_request_wr = FALSE;
	}
    NDMP_DBG_PRINTF(D_INFO, ("Mover_continue:  PendIO's %d, "
			"nextSend %lld, maxReq %lld, sessId %d\n", mover->scsiioPendReq, 
			mover->nextIoCookieSend, mover->maxIoCookieReq, servp->sessId));

    return nextState;

} /* ndmp_Mover_continue */
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_pause
 *
 * Description:
 *
 * Created by:      Anh Vu
 *
 * Data Created:    9/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_pause(esm_handle_t handle,
                    int32       nextState, /* next state from table       */
                    esm_event_t thisEvent, /* event driving this action   */
                    void        *userInfo)  /* ctrl blk linked to sm          */
{
    NdmpNIMService_t    *servp = (NdmpNIMService_t *) userInfo;
    Mover_t             *mover = &servp->Mover;
    int32               rc, dst_port = servp->ssc_src_port, msglen; 
    ndmp_NFXRsp_t       *rsp = NULL;
    ndmp_notify_mover_paused_req *pause_req;
 
    /*
     * Create a pause notification message and send it.
     */
	msglen = sizeof (ndmp_NFXRsp_t) + sizeof (ndmp_notify_mover_paused_req) - 1;
    NDMP_FP_ALLOC(rsp, ndmp_NFXRsp_t, msglen, "nfxrsp pause");
    if (rsp == NULL)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_pause:  Malloc err, "
				"sessid %d, esm_hdl %d\n", servp->sessId, handle));
		E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
				("ndmp_Mover_pause:  Malloc err, "
				"sessid %d, esm_hdl %d\n", servp->sessId, handle));
		return esm_getCurrState(handle);
	}
	pause_req = (ndmp_notify_mover_paused_req *) rsp->content;
 
    rsp->hdr.sessId   = servp->sessId;
    rsp->hdr.mid      = NDMP_MID_NOTIFY_MOVER_PAUSED;
    rsp->ndmp_mid     = NDMP_NOTIFY_MOVER_PAUSED;
    rsp->userInfo     = 0;
    pause_req->reason = mover->pauseReas;
    pause_req->seek_position.low = mover->seekPos;
    pause_req->seek_position.high = mover->seekPos >> 32;
 
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_pause:  sessid %d, mover pause %s\n", 
							servp->sessId, decode_mover_pause_reas(pause_req->reason)));
	
    /* Now send a NOTIFY_MOVER_HALTED  w/ "haltreas" reason */
    rc = eee_sendMessage((void *) rsp, msglen, dst_port, ndmp_NNIMCtrl.src_port);  
	if (rc != NFX_OK)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_pause: eee_sendMessage failed! sessid %d\n", servp->sessId));
	    E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
	           ("ndmp_Mover_pause: eee_sendMessage failed! sessid %d\n", servp->sessId));
	} else {
#ifdef NFX_SMP 
		ATOMIC_INC_INT(&ndmp_fp_msgStats[SSC_NOTIFY_MOVER_PAUSED_REQ_SNT], 1);
#else
		ndmp_fp_msgStats[SSC_NOTIFY_MOVER_PAUSED_REQ_SNT]++;
#endif
        E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
					("ndmp_Mover_pause : Send Mover_Pause, next request to reiisue has ioCookie %lld\n", mover->nextIoCookieSend));
	}

	mover->wait_to_pause = FALSE;			//we're pausing, so it's ok to release this flag

    NDMP_FP_FREE(rsp, "nfxrsp pause");

    return nextState;

} /* ndmp_Mover_pause */
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_spTapeDone
 *
 * Description:
 *
 * Created by:      Anh Vu
 *
 * Data Created:    9/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_spTapeDone(esm_handle_t handle,
                    int32       nextState, /* next state from table       */
                    esm_event_t thisEvent, /* event driving this action   */
                    void        *userInfo)  /* ctrl blk linked to sm          */
{

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_spTapeDone:  Malloc err, esm_hdl %d\n", handle));
	return nextState;
}

/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_close_conn
 *
 * Description:		caused by the reception of the NDMP_ABORT, NDMP_CLOSE msgs 
 *					from the client, and by Mover_conn_close (when data server 
 *					closes the data connection).  It will send NOTIFY_MOVER_HALTED
 *					msg to client.
 *
 * Created by:      Anh Vu
 *
 * Data Created:    9/17/01
*
 * Returns:
 *------------------------------------------------------------------------------
 */
int32
ndmp_close_conn(int32 sessId, 
				Mover_t *mover, 
				ndmp_mover_halt_reason halt_reas, 
				int32 dst_port, char *reason)
{
    ndmp_NFXRsp_t           *rsp = NULL;
    ndmp_notify_mover_halted_req   	*halt_req;
	int32					rc, msglen;
	char					*txt_reas;

    /*
     * Create a halt notification message and send it.
     */

	msglen = sizeof( ndmp_NFXRsp_t) + sizeof(ndmp_notify_mover_halted_req) +
             ((reason)? strlen(reason)+1 : 0);
    NDMP_FP_ALLOC(rsp, ndmp_NFXRsp_t, msglen, "nfxrsp halted");
    if (rsp == NULL)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_close_conn:  Out of memory, sessId %d\n", sessId));
		E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
				("ndmp_close_conn:  Out of memory, sessId %d\n", sessId));
		return NFX_ERR;			
	}
	halt_req = (ndmp_notify_mover_halted_req *) rsp->content;

    rsp->hdr.sessId  = sessId;
    rsp->hdr.mid     = NDMP_MID_NOTIFY_MOVER_HALTED;
    rsp->ndmp_mid    = NDMP_NOTIFY_MOVER_HALTED;
	rsp->userInfo 	 = 0;
	halt_req->reason = halt_reas;
    if (reason)
	{
		rsp->len = sizeof(ndmp_notify_mover_halted_req);  // offset to text_reason
		txt_reas = (char *)(rsp->content +sizeof(ndmp_notify_mover_halted_req));
	    strcpy(txt_reas, reason);		
	} 

    /* Now send a NOTIFY_MOVER_HALTED  w/ "haltreas" reason */
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_close_conn:  Sending Notify_Mover_halted. Halt_reason:  %s\n", 
							decode_mover_halt_reas(halt_reas)));
	E_LOG(class_1, info_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
			("ndmp_close_conn:  Sending Notify_Mover_halted. Halt_reason:  %s\n", 
							decode_mover_halt_reas(halt_reas)));
    rc = eee_sendMessage((void *)rsp, msglen, dst_port, ndmp_NNIMCtrl.src_port);
	if (rc != NFX_OK)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_close_conn: eee_sendMessage failed! sessid %d\n", sessId));
	    E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
	           ("ndmp_close_conn: eee_sendMessage failed! sessid %d\n", sessId));
	} else {
#ifdef NFX_SMP 
		ATOMIC_INC_INT(&ndmp_fp_msgStats[SSC_NOTIFY_MOVER_HALTED_REQ_SNT], 1);
#else
		ndmp_fp_msgStats[SSC_NOTIFY_MOVER_HALTED_REQ_SNT]++;
#endif
	}

    /* If we had advertised our port to the remote server,
     * close it now. */
    if (mover->listenConnId != -1)
    {
		//tpl_closeConnReq(mover->listenConnId);  /* connection id */
		mover->listenConnId = -1;
    }
    /* If we were provided data connection port to connect to, close it now */
    if (mover->connConnId != -1)
    {
		//tpl_closeConnReq(mover->connConnId);  /* connection id */
		mover->connConnId = -1;
    }

    mover->haltReas = halt_reas;
    NDMP_FP_FREE(rsp, "nfxrsp halted");
	return NFX_OK;
} /* ndmp_close_conn */
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_abort
 *
 * Description:
 *
 * Created by:      Anh Vu
 *
 * Data Created:    9/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_abort(esm_handle_t handle,
                    int32       nextState, /* next state from table       */
                    esm_event_t thisEvent, /* event driving this action   */
                    void        *userInfo)  /* ctrl blk linked to sm          */
{
    ndmp_NFXReq_t       *req = (ndmp_NFXReq_t *) thisEvent.p1.u.p;
    NdmpNIMService_t    *servp = (NdmpNIMService_t *) userInfo;
    Mover_t             *mover = &servp->Mover;
    int32               dst_port = servp->ssc_src_port;


	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_abort:  Begin, sessid %d, "
				"esm_hdl %d\n", servp->sessId, handle));
	
	/* Send a response to this abort cmd */
	ndmp_sendCmdRsp(req, dst_port, NDMP_NO_ERR, 0);

	/* Close any data connection we have and send a HALT notification to client */
	ndmp_close_conn(req->hdr.sessId, mover, NDMP_MOVER_HALT_ABORTED, dst_port, NULL);
		

#if 0	//Not quite sure if this is necessary since according to spec,  3/21/02
		//NDMP client always sends DATA_ABORT to Data server first which will 
		//clean up Mover as well (for local bkup/restore case)
	if (servp->Mover.addr_type == NDMP_ADDR_LOCAL)
	{
		/* If we are doing a write/read for Data server, tell it we're done so
	 	 * it can cancel FS.
	 	 */
		esmEvent.event = (servp->version == NDMP_VER_2)? 
						NDMP_DATA_V2_EVENT_IO_DONE : NDMP_DATA_EVENT_IO_DONE;
    	esmEvent.p2.u.w = NDMP_STAT_IO_ERR;
    	esmEvent.p1.u.p = NULL;
    	if (esm_deliverEvent(mover->data_smHdl, &esmEvent) != NFX_OK)
    	{
        	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_abort: ERROR, cannot "
					"post event NDMP_DATA_EVENT_IO_DONE to SM hdl %d",
                    mover->data_smHdl));
    	}
	}
#endif

	/* release resources */

	/* free what was originally the msgp in ndmp_NNIMReceiveMsg() */	
    NDMP_FP_FREE(req, "msgp ABORT");

    return nextState;

} /* ndmp_Mover_abort */
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_stop
 *
 * Description:
 *
 * Created by:      Anh Vu
 *
 * Data Created:    9/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_stop(esm_handle_t handle,
                    int32       nextState, /* next state from table       */
                    esm_event_t thisEvent, /* event driving this action   */
                    void        *userInfo)  /* ctrl blk linked to sm          */
{

    ndmp_NFXReq_t 		*msg = (ndmp_NFXReq_t *) thisEvent.p1.u.p;
    NdmpNIMService_t 	*servp = (NdmpNIMService_t *) userInfo;
    Mover_t     		*mover = &(servp->Mover);
    int32				dst_port = servp->ssc_src_port;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_stop:  Begin, sessid %d, "
						"esm_hdl %d\n", servp->sessId, handle));

	if ((mover->re_request_rd == TRUE) || (mover->re_request_wr == TRUE))
	{
		/* We have not reply to FS of its READ or WRITE request 
		 * yet and now NDMP client asks us to stop, let FS know 
		 * so it can clean up.  Only one of these condition
		 * can be true (recovery or bkup).
		 */ 

		ndmp_notify_data_srvr(servp, NULL, NDMP_STAT_IO_ERR, 0);
		mover->re_request_rd = FALSE;
		mover->re_request_wr = FALSE;	//For write, when we get EOM, it means data was written
										//no need to re_request the write.
	}

	/* reset all mover state variables */
	ndmp_Mover_init(mover);

	/* release all resources */
	ndmp_mover_release_resource(mover);
	
    /* Send a response to this stop cmd */
    ndmp_sendCmdRsp(msg, dst_port, NDMP_NO_ERR, 0);

    NDMP_FP_FREE(msg, "msgp STOP");
    return nextState;

} /* ndmp_Mover_stop */
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_cl_close
 *
 * Description:		Client sends msg to request a close on data connection.
 *					This function will therefore be invoked.
 *
 * Created by:      Anh Vu
 *
 * Data Created:    9/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_cl_close(esm_handle_t handle,
                    int32       nextState, /* next state from table       */
                    esm_event_t thisEvent, /* event driving this action   */
                    void        *userInfo)  /* ctrl blk linked to sm          */
{
	ndmp_NFXReq_t		*req = (ndmp_NFXReq_t *)thisEvent.p1.u.p;
    NdmpNIMService_t    *servp = (NdmpNIMService_t *) userInfo;
    Mover_t             *mover = &servp->Mover;
    int32               dst_port = servp->ssc_src_port;


	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_cl_close:  Begin, sessid %d, "
				"esm_hdl %d\n", servp->sessId, handle));

    /* Send a response to this Close cmd */
    ndmp_sendCmdRsp(req, dst_port, NDMP_NO_ERR, 0);

	/* Close current data connection */
	ndmp_close_conn(req->hdr.sessId, mover, NDMP_MOVER_HALT_CONNECT_CLOSED, dst_port, NULL);

    NDMP_FP_FREE(req, "msgp CLIENT CLOSE");
    return nextState;

} /* ndmp_Mover_cl_close */

/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_noop
 *
 * Description:
 *
 * Created by:      Anh Vu
 *
 * Data Created:    9/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_noop(esm_handle_t handle,
                    int32       nextState, /* next state from table       */
                    esm_event_t thisEvent, /* event driving this action   */
                    void        *userInfo)  /* ctrl blk linked to sm          */
{
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_noop:  Begin, esm_hdl %d\n", handle));

    return nextState;

} /* ndmp_Mover_noop */
/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_connected
 *
 * Description:
 *
 * Created by:      Anh Vu
 *
 * Data Created:    9/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_connected(esm_handle_t handle,
                    int32       nextState, /* next state from table       */
                    esm_event_t thisEvent, /* event driving this action   */
                    void        *userInfo)  /* ctrl blk linked to sm          */
{
	NdmpNIMService_t   *servp = (NdmpNIMService_t *) userInfo;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_connected:  Begin, esm_hdl %d\n", handle));

	if (servp->Mover.addr_type == NDMP_ADDR_LOCAL)
	{
		return nextState;
	}
    return nextState;

} /* ndmp_Mover_connected */

/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_conn_err
 *
 * Description:
 *
 * Created by:      Anh Vu
 *
 * Data Created:    9/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_conn_err(esm_handle_t handle,
                    int32       nextState, /* next state from table       */
                    esm_event_t thisEvent, /* event driving this action   */
                    void        *userInfo)  /* ctrl blk linked to sm          */
{
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_conn_err:  Begin, esm_hdl %d\n", handle));
    return nextState;

} /* ndmp_Mover_conn_err */

/*
 *------------------------------------------------------------------------------
 * Function Name:   ndmp_Mover_err
 *
 * Description:		Handle ILLEGAL_STATE error
 *
 * Created by:      Anh Vu
 *
 * Data Created:    9/17/01
 *
 * Returns:
 *------------------------------------------------------------------------------
 */
int
ndmp_Mover_err(esm_handle_t handle,
                    int32       nextState, /* next state from table       */
                    esm_event_t thisEvent, /* event driving this action   */
                    void        *userInfo)  /* ctrl blk linked to sm          */
{
	ndmp_NFXReq_t *msg = (ndmp_NFXReq_t *) thisEvent.p1.u.p;
    NdmpNIMService_t *servp = (NdmpNIMService_t *) userInfo;
    Mover_t     		*mover = &servp->Mover;
	int32 		  dst_port = servp->ssc_src_port, currState;
	int32		  errCode;


	currState = esm_getCurrState(handle);

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_err:  sessid %d, event %s, cur_state %s\n",
				servp->sessId, mover_decode_event(thisEvent.event), decode_mover_state(currState)));
	E_LOG(class_1, err_s, ndmp_NNIMCtrl.appId, 0, 0, 0,
				("ndmp_Mover_err:  sessid %d, event %s, cur_state %s\n",
				servp->sessId, mover_decode_event(thisEvent.event), decode_mover_state(currState)));

	errCode = NDMP_ILLEGAL_STATE_ERR;
	
	switch (thisEvent.event) 
	{
		case NDMP_MOVER_EVENT_LISTEN:
		case NDMP_MOVER_EVENT_CONNECT:
			if (currState != NDMP_MOVER_STATE_IDLE) 			 
				ndmp_sendCmdRsp(msg, dst_port, errCode, 0);
			break;
		case NDMP_MOVER_EVENT_CONNECTED:
			if ((currState != NDMP_MOVER_STATE_IDLE) ||
				(currState != NDMP_MOVER_STATE_LISTEN)) 			 
				ndmp_sendCmdRsp(msg, dst_port, errCode, 0);
			break;
		case NDMP_MOVER_EVENT_SET_INFO:
			if ((currState != NDMP_MOVER_STATE_IDLE) && 
					(currState != NDMP_MOVER_STATE_PAUSED))			 
				ndmp_sendCmdRsp(msg, dst_port, errCode, 0);
			break;
		case NDMP_MOVER_EVENT_READ_DATA:
			if (currState != NDMP_MOVER_STATE_ACTIVE && mover->addr_type != NDMP_ADDR_LOCAL)
				ndmp_sendCmdRsp(msg, dst_port, errCode, 0);
			break;
		case NDMP_MOVER_EVENT_CONTINUE:
			if (currState != NDMP_MOVER_STATE_PAUSED) 			 
				ndmp_sendCmdRsp(msg, dst_port, errCode, 0);
			break;
		case NDMP_MOVER_EVENT_ABORT:
			if (currState == NDMP_MOVER_STATE_IDLE) 			 
				ndmp_sendCmdRsp(msg, dst_port, errCode, 0);
			break;
		case NDMP_MOVER_EVENT_STOP:
			if (currState != NDMP_MOVER_STATE_HALTED) 			 
				ndmp_sendCmdRsp(msg, dst_port, errCode, 0);
			break;
		case NDMP_MOVER_EVENT_CLOSE:
			if (currState != NDMP_MOVER_STATE_PAUSED) 			 
				ndmp_sendCmdRsp(msg, dst_port, errCode, 0);
			break;
		case NDMP_MOVER_EVENT_TAPE_OPS_DONE:
			if (currState != NDMP_MOVER_STATE_ACTIVE) 			 
				/* ... */;
			break;
        case NDMP_MOVER_EVENT_INTERNAL_ERR:
            if ((currState == NDMP_MOVER_STATE_ACTIVE) ||
             	(currState == NDMP_MOVER_STATE_PAUSED) ||
                (currState == NDMP_MOVER_STATE_HALTED))
            {
                /* .... */
            }
            break;
        case NDMP_MOVER_EVENT_CONN_ERR:
            if ((currState == NDMP_MOVER_STATE_PAUSED) ||
                (currState == NDMP_MOVER_STATE_ACTIVE))
            {
                /* .... */
            }
            break;
        default:
			NDMP_DBG_PRINTF(D_INFO, ("ndmp_Mover_err:  Invalid event %s, sessId %d\n", 
					mover_decode_event(thisEvent.event), servp->sessId));
	}
    NDMP_FP_FREE(msg, "msgp MOVER_ERR");
    return nextState;

} /* ndmp_Mover_err */

/*-----------------------------------------------------------------
 * Name:        ndmp_Mover_bad_state
 *
 * Description: sm action function to handle bad state transitions
 *
 * created by : Anh Vu
 *
 * Date Created: 10/28/01
 *-----------------------------------------------------------------
 */

int32
ndmp_Mover_bad_state(esm_handle_t handle,
                    int32       nextState,
                    esm_event_t thisEvent,
                    void        * userInfo)
{

    NDMP_DBG_PRINTF(D_INFO, ("NDMP Mover SM[%d]: BAD STATE transition event %d nextState %d",
              handle, thisEvent.event, nextState));

    return nextState;

} /* ndmp_Data_bad_state */
