/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        ndmp-mover.c
 *
 * RCS:  $Id: ndmp-mover.c,v 1.15 2002/04/13 00:30:16 anhv Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description: Process all NDMP_MOVER_xxx commands sent by the
 *              NDMP client.  These commands are ways for the client
 *              to 
 *
 * Created by: Anh Vu
 *
 * Date Created:        9/06/01
 *
 *-----------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/utsname.h>
#include <fcntl.h>
#include <string.h>
#include <nfx-incl.h>
#include <exports/elog-api.h>
#include <exports/anpssc-api.h>
#include "ndmp.h"
#include "ndmp-ssc.h"
#include "ndmp-debug.h"


/*-----------------------------------------------------------------
 * Name: ndmp_ListenConnErrChk
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_ListenConnErrChk(ndmpSess_t *sess, int32 mode, int32 addr_type)
{
	
	Tape_t              *tapeSrv = &(sess->Serv.Tape);


    /* Handle some of the error handling here if we can */
	if ((addr_type != NDMP_ADDR_LOCAL) && (addr_type != NDMP_ADDR_TCP))
    {
        E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
						("ndmp_procMoverListen: Invalid addr_type"));
        return (NDMP_ILLEGAL_ARGS_ERR);
    }
	if ((mode != NDMP_MOVER_MODE_READ) && (mode != NDMP_MOVER_MODE_WRITE))
	{
        E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
					("ndmp_procMoverListen: Invalid r/w Mode"));
        return (NDMP_ILLEGAL_ARGS_ERR);
	}
    if ( tapeSrv->hdl == 0)
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_procMoverConn: no Tape device Open.\n"));
        E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
					("ndmp_procMoverListen: no tape device open."));
        return (NDMP_DEV_NOT_OPEN_ERR);
    } 
	if ( mode == NDMP_MOVER_MODE_READ &&
                tapeSrv->mode == TAPE_MODE_RD_ONLY)
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_procMoverConn: Incorrect mode for Read only device\n"));
        E_LOG(class_1, err_s, eee_getApplicationId("ssc_ndmp"), 0, 0, 0,
					("ndmp_procMoverConn: Incorrect mode for Read only device"));
        return (NDMP_PERMISSION_ERR);
    } 
	return NDMP_NO_ERR;	
} /* ndmp_ListenConnErrChk */

/*-----------------------------------------------------------------
 * Name: ndmp_procMoverGetStateReq
 *
 * Description:     Process NDMP_MOVER_GET_STATE req.  Pass it to NNIM.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procMoverGetStateReq(ndmpSess_t *sess)
{
   
    NDMP_DBG_PRINTF(D_INFO, ( "ndmp_procMoverGetState:\n"));
    ndmp_fwdMsgtoNNIM(sess, NDMP_MID_MOVER_GET_STATE, 0);

	return NFX_OK;

} /* ndmp_procMoverGetState */

/*-----------------------------------------------------------------
 * Name: ndmp_procMoverListenReq
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procMoverListenReq(ndmpSess_t *sess)
{
	ndmp_mover_listen_rsp	cl_rsp;
	ndmp_mover_listen_req	*cl_req	= (ndmp_mover_listen_req *) sess->msgRcvd.content;
	
	NDMP_DBG_PRINTF(D_INFO, ( "ndmp_procMoverListen: mode:%s type:%s.\n",
		  cl_req->mode == NDMP_MOVER_MODE_READ ? "READ" : "WRITE",
		  cl_req->addr_type == NDMP_ADDR_LOCAL ? "LOCAL" :
		  cl_req->addr_type == NDMP_ADDR_TCP ? "TCP" : "undefined"));

	memset((void*)&cl_rsp, 0, sizeof(cl_rsp));

	/* Handle some of the error handling here if we can */
	cl_rsp.error = ndmp_ListenConnErrChk(sess, cl_req->mode, cl_req->addr_type);
	
	if (cl_rsp.error != NDMP_NO_ERR)
	{
        return  ndmp_rspClient(sess, &cl_rsp);
	}
	/* Forward MOVER_LISTEN request to Mover on N_NIM */
	ndmp_fwdMsgtoNNIM(sess, NDMP_MID_MOVER_LISTEN, sess->Serv.Tape.hdl);

	return NFX_OK;
} /* ndmp_procMoverListen */

/*-----------------------------------------------------------------
 * Name: ndmp_procMoverConnReq
 *
 * Description:     
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procMoverConnReq(ndmpSess_t *sess)
{
    ndmp_mover_conn_rsp   cl_rsp;
    ndmp_mover_conn_req   *cl_req = (ndmp_mover_conn_req *) sess->msgRcvd.content;

    NDMP_DBG_PRINTF(D_INFO, ( "ndmp_procMoverConn: mode:%s type:%s.\n",
          cl_req->mode == NDMP_MOVER_MODE_READ ? "READ" : "WRITE",
          cl_req->addr.addr_type == NDMP_ADDR_LOCAL ? "LOCAL" :
          cl_req->addr.addr_type == NDMP_ADDR_TCP ? "TCP" : "undefined"));

    memset((void*)&cl_rsp, 0, sizeof(cl_rsp));

    /* Handle some of the error handling here if we can */
	cl_rsp.error = ndmp_ListenConnErrChk(sess, cl_req->mode, cl_req->addr.addr_type);

    if (cl_rsp.error != NDMP_NO_ERR)
    {
        return  ndmp_rspClient(sess, &cl_rsp);
    }
    /* Forward MOVER_CONNECT request to Mover on N_NIM */
	ndmp_fwdMsgtoNNIM(sess, NDMP_MID_MOVER_CONNECT, sess->Serv.Tape.hdl);
    return NFX_OK; 

} /* ndmp_procMoverConnReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procMoverContReq
 *
 * Description:     
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procMoverContReq(ndmpSess_t *sess)
{
 
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procMoverCont.\n"));
 
    /* Forward MOVER_CONTINUE request to Mover on N_NIM */
	ndmp_fwdMsgtoNNIM(sess, NDMP_MID_MOVER_CONTINUE, 0); 
	return NFX_OK;
}	
/*-----------------------------------------------------------------
 * Name: ndmp_procMoverAbortReq
 *
 * Description:     
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procMoverAbortReq(ndmpSess_t *sess)
{
    NDMP_DBG_PRINTF(D_INFO, ("ndmp_procMoverAbort.\n"));
 
    /* Forward MOVER_ABORT request to Mover on N_NIM */
	ndmp_fwdMsgtoNNIM(sess, NDMP_MID_MOVER_ABORT, 0); 
	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: ndmp_procMoverStopReq
 *
 * Description:     
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procMoverStopReq(ndmpSess_t *sess)
{
	
	NDMP_DBG_PRINTF(D_INFO, ("ndmpdMoverStop.\n"));

    /* Forward MOVER_STOP request to Mover on N_NIM */
	ndmp_fwdMsgtoNNIM(sess, NDMP_MID_MOVER_STOP, 0); 
	return NFX_OK;
	
}
/*-----------------------------------------------------------------
 * Name: ndmp_procMoverSetRecSizReq
 *
 * Description:    Forward Set Record Size message to NNIM.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procMoverSetRecSizReq(ndmpSess_t *sess)
{
    ndmp_mover_set_record_size_req  *cl_req =
                (ndmp_mover_set_record_size_req *) sess->msgRcvd.content;
    ndmp_mover_set_record_size_rsp     cl_rsp;

#if SYNCSORT
	char8		buf[100];
#endif 

    NDMP_DBG_PRINTF(D_INFO, ("ndmpdMoverSetRecordSize: len:%ld.\n", cl_req->len));

    memset((void*)&cl_rsp, 0, sizeof(cl_rsp));

#if 0
	// Don't need tape to be open to establish record size. - per connectathon test
    cl_rsp.error = NDMP_DEV_NOT_OPEN_ERR;
    if (ndmp_chkDevOpen(sess, &cl_rsp, TAPE_DEV) != NFX_OK)
        return NFX_ERR;
#endif
#if 0 // for SYNCSORT
	//currently fails Connectathon if we don't accept Client's recsiz - 1/24/02
	// Legato just gives warning - 2/15/02
    if (cl_req->len > NDMP_MAX_MOVR_RECSIZ)
    {
    	NDMP_DBG_PRINTF(D_INFO, ("ndmpdMoverSetRecordSize: len > max:%ld.\n", cl_req->len));
		bzero((void*)&cl_rsp, sizeof(ndmp_mover_set_record_size_rsp));
        cl_rsp.error = NDMP_ILLEGAL_ARGS_ERR;
        ndmp_rspClient(sess, &cl_rsp);

		sprintf(buf, "Mover Record Size is reset to %d.\n", NDMP_MAX_MOVR_RECSIZ);
		ndmp_sndLogMsg( sess, NDMP_LOG_WARNING, buf, 0); 

    } 
#endif
    /* Forward MOVER_SET_RECSIZ request to Mover on N_NIM */
    ndmp_fwdMsgtoNNIM(sess, NDMP_MID_MOVER_SET_RECORD_SIZE, 0);
    return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: ndmp_procMoverSetWdw
 *
 * Description:     
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procMoverSetWdwReq(ndmpSess_t *sess)
{

    ndmp_mover_set_window_req  	*cl_req = 
				(ndmp_mover_set_window_req *) sess->msgRcvd.content;
#if 0
    ndmp_mover_set_window_rsp     cl_rsp;
#endif

	NDMP_DBG_PRINTF(D_INFO, ("ndmpdMoverSetWindow: offset low:%ld high %ld,  length low:%ld high %ld.\n",
		  cl_req->offset.low, cl_req->offset.high, cl_req->length.low, cl_req->length.high));

#if 0
	bzero((void*)&cl_rsp, sizeof(ndmp_mover_set_window_rsp)); 

	// Don't need tape to be open to establish record size. - connectathon test 
    cl_rsp.error = NDMP_DEV_NOT_OPEN_ERR;
    if (ndmp_chkDevOpen(sess, &cl_rsp, TAPE_DEV) != NFX_OK)
        return NFX_ERR;
#endif

    /* Forward MOVER_SET_WINDOW request to Mover on N_NIM */
	ndmp_fwdMsgtoNNIM(sess, NDMP_MID_MOVER_SET_WINDOW, 0); 
	return NFX_OK;
} /* ndmp_procMoverSetWdwReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procMoverRdReq
 *
 * Description:   Check if tape device is opened first, if not respond
 *					w/ error.  Otherwise, forward msg to NNIM.  
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procMoverReadReq(ndmpSess_t *sess)
{
    ndmp_mover_read_rsp   	cl_rsp;
    ndmp_mover_read_req   	*cl_req = (ndmp_mover_read_req *) sess->msgRcvd.content;

	
	NDMP_DBG_PRINTF(D_INFO, ("ndmpdMoverReadReq: offset low: %ld high: %ld, len low:%ld high: %ld.\n",
		  cl_req->offset.low, cl_req->offset.high, cl_req->length.low, cl_req->length.high));

	memset((void*)&cl_rsp, 0, sizeof(cl_rsp));
	
    cl_rsp.error = NDMP_DEV_NOT_OPEN_ERR;                                            	
	if (ndmp_chkDevOpen(sess, &cl_rsp, TAPE_DEV) != NFX_OK)
        return NFX_ERR;

	/* Forward MOVER_READ cmd to Mover on NNIM */
	ndmp_fwdMsgtoNNIM(sess, NDMP_MID_MOVER_READ, 0);		
	return NFX_OK;
} /* ndmp_procMoverReadReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procMoverCloseReq
 *
 * Description:    Forward Close Data Connection msg to NNIM. 
 *
 * Created by:  Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
int32
ndmp_procMoverCloseReq(ndmpSess_t *sess)
{
	
	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procMoverCloseReq.\n"));

	/* Forward MOVER_CLOSE cmd to Mover on NNIM */
	ndmp_fwdMsgtoNNIM(sess, NDMP_MID_MOVER_CLOSE, 0);
	return NFX_OK;
}

