/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        ndmp-srvr-ui.c
 *
 * RCS:  $Id: 
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:		Contain functions to process UI msgs for
 *					the parent NDMP server (who accepts new
 *					session connection) to forward to
 *					the children to process.  Each child represents
 *					an NDMP server session.	
 *
 * Created by: Anh Vu
 *
 * Date Created:        02/23/02
 *
 *-----------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>  
#include <nfx-types.h>
#include <exports/eee-api.h>
#include <exports/elog-api.h>
#include <exports/anpssc-api.h>
#include "ndmp.h"
#include "ndmp-ssc.h"
#include "ndmp-debug.h"
#include "ndmp-fs-api.h"
#include <nfx-incl.h>

extern int ndmp_enabled;
extern int agileFd;
extern sessId_entry_t  sessId_Tbl[NDMP_MAX_SESS]; 
extern int ndmp_hi_ver;
extern int ndmp_ssc_par_msgStats[];

int send_rsp_to_UI(void *rsp, int rsp_len, int32 mysock, char *app, uint32 cpu, uint32 slot);
int ndmp_sndUIReq2FP(ndmpSess_t *sess, uint32 msgId, ndmp_Req_info_t *rq_info);
int32 free_req_info(ndmp_Req_info_t * rq_info);
ndmp_Req_info_t *save_req_info(char *app, int32 cpu, int32 slot, int32 mid);
int send_UIStatRsp(ndmpSess_t *sess, ndmp_Req_info_t* rq_info, int stat);
int send_UIChldStatisticRsp(ndmpSess_t *sess, ndmp_Req_info_t* rq_info, int stat);


/* We support the following:
 * ------------------------
 * - ndmp disable
 * - ndmp enable
 * - ndmp show   active		(show all active sessions, displaying sessids)
 * - ndmp show   status  <-a | -s id> (or <all|sessid>) 
 * - ndmp show   detail  <-a | -s id> (or <all|sessid>)
 * - ndmp delete session <-a | -s id> (or <all|sessid>)
 * - ndmp set version <num> 
 * - ndmp debug level <num> <-a | -s id>
 */

/* ================================================================
 * These are funcs called by parent NDMP server to forward UI
 * msgs to its children sessions that it has forked.
 * ================================================================
 */
/*-----------------------------------------------------------------
 * Name: send_ui_rsp
 *
 * Description:  Parent responds to UI process of what it knows.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
int
send_rsp_to_UI(void *rsp, int rsp_len, int32 mysock, char *app, uint32 cpu, uint32 slot)
{
    int rc;

    NDMP_DBG_PRINTF(D_INFO, ("send_rsp_to UI: " 
			"%d bytes to app %s, cpu %d, slot %d\n",
           	rsp_len, app, cpu, slot));
   
    rc = sendAgileMsg(mysock,
              rsp, rsp_len,
              app, cpu, slot, 0,	/* app, cpu, slot, plane */
              NULL,					/* timeout */
              NULL, 0);				/* rcvbuf, rcvlen) */

	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: ndmp_procChildReport
 *
 * Description:	Process Child Report message.  Msg for child process
 *				to inform parent process of its application name
 *				so it can keep track to send children msgs later on.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/23/02
 *
 *-----------------------------------------------------------------
 */
int ndmp_procChildReport(char *rcvbuf, ndmp_src_t *src)
{
	int i;
	ndmp_NFXSndRptReq_t *req = (ndmp_NFXSndRptReq_t *) rcvbuf;

	/* Capture child's application name so we can send
	 * UI msgs to him later on.
	 */

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procChildReport:\n"));
	for (i = 0; i < NDMP_MAX_SESS; i++)
    {
        if (sessId_Tbl[i].in_use == TRUE)
        {
			if (req->sessId == sessId_Tbl[i].sessId)
			{
				strcpy(sessId_Tbl[i].app_name, src->app);
			}
            NDMP_DBG_PRINTF(D_INFO, ("%2d          %s              %d          %s\n",
                i, inet_ntoa(sessId_Tbl[i].from.sin_addr), 
				req->sessId, sessId_Tbl[i].app_name));
        }
	}
	return NFX_OK;
}
/*-----------------------------------------------------------------
 * Name: ndmp_procChildSndStats
 *
 * Description:	Process Child SndStats message.  Msg for child process
 *				to inform parent process of its stats before it dies
 *				so parent can keep an accumulative count of all stats.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/23/02
 *
 *-----------------------------------------------------------------
 */
int ndmp_procChildSndStats(char *rcvbuf, ndmp_src_t *src)
{
	int i;
	ndmp_NFXSndStatsReq_t *req = (ndmp_NFXSndStatsReq_t *) rcvbuf;

	/* Capture child's stats */

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procChildSndStats:\n"));
	for (i = 0; i < (MAX_SSC_CHLD_STAT) ; i++)
    {
		ndmp_ssc_par_msgStats[i] += req->statistic[i];
	}
	return NFX_OK;
}
/*-----------------------------------------------------------------
 * Name: ndmp_procSetVerReq
 *
 * Description:	Process set version command
 *
 * Created by:  Anh Vu
 *
 * Date Created:        05/6/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procSetVerReq(char *rcvbuf, ndmp_src_t *src)
{
	ndmp_NFXSetVerReq_t *req = (ndmp_NFXSetVerReq_t *) rcvbuf;
	ndmp_NFXSetVerRsp_t rsp;
	char ver[3];

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procSetVerReq:  ver = %d\n", req->version));

	bzero(&rsp, sizeof(ndmp_NFXSetVerRsp_t));
	rsp.mid = NDMP_MID_UI_SET_VERSION;
	rsp.stat = NFX_OK;
	
	/* Is user asking for valid supported version ? */
	if ((req->version != NDMP_VER_3) && (req->version != NDMP_VER_2))
	{
		rsp.stat = NFX_ERR;
	} else {
		if (ndmp_hi_ver == req->version)
			rsp.already_so = TRUE;
		else 
			ndmp_hi_ver = req->version;
	}
    send_rsp_to_UI(&rsp, sizeof(rsp), agileFd, src->app, src->cpu, src->slot);

	bzero(ver, sizeof(ver));
	sprintf(ver, "%d", ndmp_hi_ver);
	ndmp_regSetOption(OPTIONS_NDMP_HI_VER, ver);

	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: ndmp_procShowStateReq
 *
 * Description:	Process Show state NDMP command
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procShowStateReq(char *rcvbuf, ndmp_src_t *src)
{
	ndmp_NFXShowStateRsp_t rsp;

	bzero(&rsp, sizeof(ndmp_NFXShowStateRsp_t));
	rsp.mid = NDMP_MID_UI_SHOW_STATE;
	rsp.state = ndmp_enabled;
	rsp.version = ndmp_hi_ver;
	rsp.trace = ndmp_trace_msg;
	rsp.tracefile = ndmp_trace_files;
	rsp.restoreMode = ndmp_restore_mode;
	send_rsp_to_UI(&rsp, sizeof(rsp), agileFd, src->app, src->cpu, src->slot);

	return NFX_OK;
} /* ndmp_procShowStateReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procClearStatisticReq
 *
 * Description:	Responding to NDMP_MID_UI_SHOW_PAR_STAT(istic)S  for
 *				parent NDMP server.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        07/30/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procClearStatisticReq(char *rcvbuf, ndmp_src_t *src)
{      

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procClearStatisticReq:\n"));

	bzero(ndmp_ssc_par_msgStats, MAX_SSC_TOT_STAT *sizeof(int));
	
	return NFX_OK;
} /* ndmp_procShowStatisticReq */ 

/*-----------------------------------------------------------------
 * Name: ndmp_procShowStatisticReq
 *
 * Description:	Responding to NDMP_MID_UI_SHOW_PAR_STAT(istic)S  for
 *				parent NDMP server.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        07/30/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procShowStatisticReq(char *rcvbuf, ndmp_src_t *src)
{      
	ndmp_NFXShowParStatsRsp_t	 rsp;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procShowStatisticReq:\n"));

	bzero(&rsp, sizeof(ndmp_NFXShowParStatsRsp_t));
	rsp.mid = NDMP_MID_UI_SHOW_PAR_STATS;
	memcpy(rsp.statistic, &ndmp_ssc_par_msgStats,
				(MAX_SSC_TOT_STAT) *sizeof(int));
	
	send_rsp_to_UI(&rsp, sizeof(rsp), agileFd, src->app, src->cpu, src->slot);


	return NFX_OK;
} /* ndmp_procShowStatisticReq */ 

/*-----------------------------------------------------------------
 * Name: ndmp_procEnaDisNdmpReq
 *
 * Description:	Process Disable/Enable NDMP command
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procEnaDisNdmpReq(char *rcvbuf, ndmp_src_t *src)
{
	ndmp_NFXGenericUIReq_t *req = (ndmp_NFXGenericUIReq_t *) rcvbuf;
	ndmp_NFXDisEnaRsp_t rsp;
	int					i;

	bzero(&rsp, sizeof(ndmp_NFXDisEnaRsp_t));
	switch(req->mid)
	{
		case NDMP_MID_UI_ENABLE_NDMP:
			rsp.mid = NDMP_MID_UI_ENABLE_NDMP;
			if (ndmp_enabled)
				rsp.already_so = TRUE;
			else {
				ndmp_enabled = 1;
			}
			break;
		case NDMP_MID_UI_DISABLE_NDMP:
			rsp.mid = NDMP_MID_UI_DISABLE_NDMP;
			if (!ndmp_enabled)
			{
				/* Ndmp is already disabled */
				rsp.already_so = TRUE;
			} else {
				for (i = 0; i < NDMP_MAX_SESS; i++)
    			{
        			if (sessId_Tbl[i].in_use == TRUE)
        			{
						/* There are sessions still running, 
				 	 	 * new requests will be rejected
				     	 */
						rsp.running_sess = TRUE;
						break;
        			}
    			}	 
				ndmp_enabled = 0;
			}
			break;
		default:
			NDMP_DBG_PRINTF(D_INFO, ("ndmp_procEnaDisNdmpReq: invalid req msgid %d\n", req->mid));
	}	
	ndmp_regSetOption(OPTIONS_NDMP_ENBL_MODE, (ndmp_enabled)?"1":"0");

	send_rsp_to_UI(&rsp, sizeof(rsp), agileFd, src->app, src->cpu, src->slot);

	return NFX_OK;
} /* ndmp_procEnaDisNdmpReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procGetSessTblReq
 *
 * Description:	Parent replies UI process with the table of Session Ids.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procGetSessTblReq(char *rcvbuf, ndmp_src_t *src)
{
	ndmp_NFXGenericUIReq_t  *req = (ndmp_NFXGenericUIReq_t *)rcvbuf;
	ndmp_NFXGetTblRsp_t rsp;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_procGetSessTblReq:\n"));

   
	bzero(&rsp, sizeof(ndmp_NFXGetTblRsp_t));
    rsp.mid = req->mid;
    bcopy(&sessId_Tbl, &rsp.tbl, sizeof(sessId_Tbl));

    send_rsp_to_UI(&rsp, sizeof(rsp), agileFd, src->app, src->cpu, src->slot);

	return NFX_OK;
} /* ndmp_procGetSessTblReq */

/*-----------------------------------------------------------------
 * Name: ndmp_procSetDbgLevel
 *
 * Description:	Set debug level in the parent so successive
 *				processes will take this debug level when it
 *				starts.	
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procSetDbgLevel(char *rcvbuf, ndmp_src_t *src)
{

	ndmp_NFXSetDebugReq_t 	*req = (ndmp_NFXSetDebugReq_t *) rcvbuf;
	ndmp_NFXSetDebugRsp_t	rsp;

	printf("ndmp_procSetDbgLevel:  Parent debug level set to %d\n",
		req->dbg_level);

	ndmp_dbg_level = req->dbg_level;  

	rsp.mid = NDMP_MID_UI_SET_DBG_LEVEL;
    rsp.stat = NFX_OK;

    send_rsp_to_UI(&rsp, sizeof(rsp), agileFd,
            src->app, src->cpu, src->slot);

	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: ndmp_procSetTraceFlag
 *
 * Description: Set trace flag in the parent so successive
 *              processes will take this flag when they fork. 
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 08/20/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procSetTraceFlag(char *rcvbuf, ndmp_src_t *src)
{
    ndmp_NFXSMsgTraceReq_t *req = (ndmp_NFXSMsgTraceReq_t *)rcvbuf;
    ndmp_NFXSMsgTraceRsp_t rsp;

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_procSetTraceFlag: Flag (%s)\n", (req->flag)?"ON":"OFF"));

    ndmp_trace_msg = req->flag;
    ndmp_regSetOption(OPTIONS_NDMP_TRACE_MODE, (ndmp_trace_msg)?"1":"0");
    rsp.mid = NDMP_MID_UI_SET_TRACE_ON_OFF;
    rsp.stat = NFX_OK;

    send_rsp_to_UI(&rsp, sizeof(ndmp_NFXSMsgTraceRsp_t), agileFd,
                   src->app, src->cpu, src->slot);

    return NFX_OK;
}
/*-----------------------------------------------------------------
 * Name: ndmp_procSetTraceFiles
 *
 * Description: Set # of trace files to retain in the system.
 * 				We're setting in the parent so successive
 *              processes will take this flag when they fork. 
 *
 * Created by:  Anh Vu
 *
 * Date Created: 10/11/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procSetTraceFiles(char *rcvbuf, ndmp_src_t *src)
{
    ndmp_NFXSMsgTraceReq_t *req = (ndmp_NFXSMsgTraceReq_t *)rcvbuf;
    ndmp_NFXSMsgTraceRsp_t rsp;
	char numfil[20];

    NDMP_DBG_PRINTF(D_INFO, ("ndmp_procSetTraceFiles: Num of Files (%d)\n", req->num_files));

    ndmp_trace_files = req->num_files;
	bzero(numfil, sizeof(numfil));
	sprintf(numfil, "%d", ndmp_trace_files);
    ndmp_regSetOption(OPTIONS_NDMP_TRACE_DURATION, numfil);
    rsp.mid = NDMP_MID_UI_SET_TRACE_FILES;
    rsp.stat = NFX_OK;

    send_rsp_to_UI(&rsp, sizeof(ndmp_NFXSMsgTraceRsp_t), agileFd,
                   src->app, src->cpu, src->slot);

    return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: ndmp_procSetRestoreMode
 *
 * Description: Set the trace mode.
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 12/01/02
 *
 *-----------------------------------------------------------------
 */
int ndmp_procSetRestoreMode(char *rcvbuf, ndmp_src_t *src)
{
    ndmp_NFXSetRestoreModeReq_t *req = (ndmp_NFXSetRestoreModeReq_t *)rcvbuf;
    ndmp_NFXSetRestoreModeRsp_t rsp;

    NDMP_DBG_PRINTF(D_INFO,("ndmp_procSetRestoreMode: Flag (%d)\n", req->flag));

    ndmp_restore_mode = req->flag;
    ndmp_regSetOption(OPTIONS_NDMP_RESTORE_MODE, (ndmp_restore_mode)?"1":"0");
    rsp.mid = NDMP_MID_UI_SET_RESTORE_MODE;
    rsp.stat = NFX_OK;

    send_rsp_to_UI(&rsp, sizeof(ndmp_NFXSetRestoreModeRsp_t), agileFd,
                   src->app, src->cpu, src->slot);

    return NFX_OK;
}


/* ================================================================
 * These are funcs called by the child NDMP server to get the
 * necessary info and return to UI process?
 * ================================================================
 */
/*-----------------------------------------------------------------
 * Name: ndmp_ChildRpt2Parent
 *
 * Description:		Child report to parent of its appid so
 *				Parent can forward UI msgs to it later.  Parent's
 *				app name is "ssc_ndmp".
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/23/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_ChildRpt2Parent(ndmpSess_t *sess)
{

	ndmp_NFXSndRptReq_t req;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_ChildRpt2Parent: Sess [%d] Socket [%d]\n", sess->sess_idx, sess->Agilesock));
	
	bzero(&req, sizeof(ndmp_NFXSndRptReq_t));
	req.mid = NDMP_MID_CHILD_REPORT;
	req.sessId = sess->sessId;

	/* Report to parent of our application name by 
	 * sending this msg.
	 */
    NDMP_DBG_PRINTF(D_INFO, ("ndmp_ChildRpt2Parent : Sending message to ssc_ndmp\n"));
   	if (sendAgileMsg(sess->Agilesock, &req, sizeof(ndmp_NFXSndRptReq_t),
                       "ssc_ndmp", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, /* to which applic */
                       NULL, NULL, 0) < 0)
	{   	
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_ChildRpt2Parent: Agile message send failed\n"));
   	}
	return NFX_ERR;
}
/*-----------------------------------------------------------------
 * Name: ndmp_ChildSndStats2Parent
 *
 * Description:		Child send its stats to parent when it exits
 *				so Parent can keep an accumulative stats of msgs
 *				sent and received between boards.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        08/2/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_ChildSndStats2Parent(ndmpSess_t *sess)
{

	ndmp_NFXSndStatsReq_t req;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_ChildSndStats2Parent: Sess [%d] Socket [%d]\n", sess->sess_idx, sess->Agilesock));
	
	bzero(&req, sizeof(ndmp_NFXSndStatsReq_t));
	req.mid = NDMP_MID_CHILD_SND_STATS;
	req.sessId = sess->sessId;
	memcpy(req.statistic, sess->ndmp_ssc_chld_msgStats, 
				MAX_SSC_CHLD_STAT *sizeof(int));
	
	/* Send to parent of our stats of msgs snt & rcvd between FP or 
	 * other process w/in SSC to NDMP server.
	 */
    NDMP_DBG_PRINTF(D_INFO, ("ndmp_ChildSndStats2Parent : Sending message to ssc_ndmp\n"));
   	if (sendAgileMsg(sess->Agilesock, &req, sizeof(ndmp_NFXSndStatsReq_t),
                       "ssc_ndmp", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, /* to which applic */
                       NULL, NULL, 0) < 0)
	{   	
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_ChildStats2Parent: Agile message send failed\n"));
   	}
	return NFX_ERR;
}

/*-----------------------------------------------------------------
 * Name: free_req_info
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/23/02
 * 
 *-----------------------------------------------------------------
 */
int32 
free_req_info(ndmp_Req_info_t * rq_info)
{
    NDMP_SSC_FREE(rq_info);
    return NFX_OK;
}   

/*-----------------------------------------------------------------
 * Name: save_req_info
 *
 * Description:
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/23/02
 *
 *-----------------------------------------------------------------
 */
ndmp_Req_info_t *
save_req_info(char *app, int32 cpu, int32 slot, int32 mid)
{
	ndmp_Req_info_t *req_info = NULL;

    NDMP_SSC_ALLOC(req_info, ndmp_Req_info_t, sizeof(ndmp_Req_info_t));
	bzero(req_info, sizeof(ndmp_Req_info_t));

	req_info->msgId = mid;
    strcpy(req_info->src.app, app);
    req_info->src.cpu = cpu;
    req_info->src.slot = slot;

	return req_info;
}

/*-----------------------------------------------------------------
 * Name: send_UIChldStatisticRsp
 *
 * Description:	We send back the statistics. 
 *
 * Created by:  Anh Vu
 *
 * Date Created: 07/29/02
 *
 *-----------------------------------------------------------------
 */
int
send_UIChldStatisticRsp(ndmpSess_t *sess, ndmp_Req_info_t* rq_info, int stat)
{
	ndmp_NFXShowChildStatsRsp_t  c_rsp;

	c_rsp.mid = rq_info->msgId;
	c_rsp.sessId = sess->sessId;

	if (stat == NFX_OK)
	{
		c_rsp.stat = NFX_OK;
		memcpy(c_rsp.statistic, sess->ndmp_ssc_chld_msgStats, 
					MAX_SSC_CHLD_STAT *sizeof(int));
	} else {
		c_rsp.stat = NFX_ERR;
	}			

	send_rsp_to_UI(&c_rsp, sizeof(ndmp_NFXShowChildRsp_t), sess->Agilesock,
			rq_info->src.app, rq_info->src.cpu, rq_info->src.slot);
	free_req_info(rq_info);	

	return NFX_OK;

} /* send_UIChldStatisticRsp */ 

/*-----------------------------------------------------------------
 * Name: send_UIStatRsp
 *
 * Description:	We send back the Sess_t struc which contains all
 *				info needed to display any type of statuses.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/23/02
 *
 *-----------------------------------------------------------------
 */
int
send_UIStatRsp(ndmpSess_t *sess, ndmp_Req_info_t* rq_info, int stat)
{
	ndmp_NFXShowChildRsp_t  c_rsp;
	ndmp_Session_t *r_sess;

	c_rsp.mid = rq_info->msgId;

	if (stat == NFX_OK)
	{
		c_rsp.stat = NFX_OK;

		/* General info */
		r_sess = &c_rsp.sess;
		r_sess->authed = sess->authed;
		r_sess->auth_typ = sess->auth_typ;
		r_sess->protoVer = sess->protoVer;
		r_sess->sessId = sess->sessId;
		r_sess->trace = ndmp_trace_msg;
		r_sess->trace_files = ndmp_trace_files;
		memcpy(&r_sess->host_ipaddr, &sess->host_ipaddr, 
					sizeof(struct sockaddr_in));
		memcpy(r_sess->start_time, sess->start_time, sizeof(sess->start_time));
		/* Tape info */
		r_sess->Tape.hdl = sess->Serv.Tape.hdl;
		r_sess->Tape.mode = sess->Serv.Tape.mode;
		r_sess->Tape.status = sess->Serv.Tape.status;
		memcpy(r_sess->Tape.devname, sess->Serv.Tape.devname, sizeof(sess->Serv.Tape.devname));
		memcpy(r_sess->Tape.phys_devname, sess->Serv.Tape.phys_devname, sizeof(sess->Serv.Tape.phys_devname));

		/* SCSI info */
		r_sess->SCSI.hdl = sess->Serv.SCSI.hdl;
		memcpy(r_sess->SCSI.devname, sess->Serv.SCSI.devname, sizeof(sess->Serv.SCSI.devname));
		memcpy(r_sess->SCSI.phys_devname, sess->Serv.SCSI.phys_devname, sizeof(sess->Serv.SCSI.phys_devname));
	
		/* Mover info */
		r_sess->Mover.state = sess->Serv.Mover.mover_state.state;
		r_sess->Mover.pause_reason = sess->Serv.Mover.mover_state.pause_reason;
		r_sess->Mover.halt_reason = sess->Serv.Mover.mover_state.halt_reason;
		r_sess->Mover.recsiz = sess->Serv.Mover.mover_state.recsiz;
		r_sess->Mover.recnum = sess->Serv.Mover.mover_state.recnum;
		r_sess->Mover.data_written = 	
			((uint64)sess->Serv.Mover.mover_state.data_written.high << 32) + 
			sess->Serv.Mover.mover_state.data_written.low;
		r_sess->Mover.seek_pos = 	
			((uint64)sess->Serv.Mover.mover_state.seek_pos.high << 32) + 
			sess->Serv.Mover.mover_state.seek_pos.low;
		r_sess->Mover.bytes_left_to_read = 	
			((uint64)sess->Serv.Mover.mover_state.bytes_left_to_read.high << 32) + 
			sess->Serv.Mover.mover_state.bytes_left_to_read.low;
		r_sess->Mover.window_offset = 	
			((uint64)sess->Serv.Mover.mover_state.window_offset.high << 32) + 
			sess->Serv.Mover.mover_state.window_offset.low;
		r_sess->Mover.window_len = 	
			((uint64)sess->Serv.Mover.mover_state.window_len.high << 32) + 
			sess->Serv.Mover.mover_state.window_len.low;
		r_sess->Mover.addr_type = sess->Serv.Mover.mover_state.data_conn_addr.addr_type;	
		r_sess->Mover.mode = sess->Serv.Mover.mode;

		/* Data info */
		r_sess->Data.state = sess->Serv.Data.data_state.state;
		r_sess->Data.op = sess->Serv.Data.data_state.op;
		r_sess->Data.halt_reason = sess->Serv.Data.data_state.halt_reason;
		r_sess->Data.bytes_processed = 	
			((uint64)sess->Serv.Data.data_state.bytes_processed.high << 32) + 
			sess->Serv.Data.data_state.bytes_processed.low;
		r_sess->Data.read_length = 	
			((uint64)sess->Serv.Data.data_state.read_length.high << 32) + 
			sess->Serv.Data.data_state.read_length.low;
		r_sess->Data.read_offset = 	
			((uint64)sess->Serv.Data.data_state.read_offset.high << 32) + 
			sess->Serv.Data.data_state.read_offset.low;
		r_sess->Data.addr_type = sess->Serv.Data.data_state.data_conn_addr.addr_type;	

	} else {
		c_rsp.stat = NFX_ERR;
	}			

	send_rsp_to_UI(&c_rsp, sizeof(ndmp_NFXShowChildRsp_t), sess->Agilesock,
			rq_info->src.app, rq_info->src.cpu, rq_info->src.slot);
	free_req_info(rq_info);	

	return NFX_OK;

} /* send_UIStatRsp */ 

/*-----------------------------------------------------------------
 * Name: ndmp_procUIDMGetStateRsp
 *
 * Description:	Process reply from FP to our UI request for Data & 
 *				Mover Get State info.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/23/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_procUIDMGetStateRsp(ndmpSess_t *sess, char* rcvbuf, ndmp_src_t *src)
{
	Mover_SSC_t     *mover = &sess->Serv.Mover;	
    Data_SSC_t      *data = &sess->Serv.Data;
   	ndmp_NFXUI_GetState_Rsp_t	*rsp = (ndmp_NFXUI_GetState_Rsp_t *)rcvbuf;
	ndmp_Req_info_t	*rq_info;
	int				stat = NFX_OK;

	if (rsp->error != NDMP_UNDEFINED_ERR)
	{
		/* Save mover_get_state info */
		bcopy(&rsp->mover_state, &mover->mover_state, sizeof(ndmp_mover_get_state_rsp));
		mover->mode = rsp->mover_mode;

    	/* Save data_get_state info */
    	bcopy(&rsp->data_state, &data->data_state, sizeof(ndmp_data_get_state_rsp));
		rq_info = (ndmp_Req_info_t *)rsp->hdr.clientHdl;
	} else {
		stat = NFX_ERR;
	}
	send_UIStatRsp(sess, rq_info, stat);

	return NFX_OK;
} /* ndmp_procUIDMGetStateRsp */

/*-----------------------------------------------------------------
 * Name: 	ndmp_sndUIReq2FP
 *
 * Description:	Send off request to FP.  Don't wait, we shall treat
 *				this asynchronously.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/23/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_sndUIReq2FP(ndmpSess_t *sess, uint32 msgId, ndmp_Req_info_t *rq_info)
{
	ndmp_NFXReq_t    ndmp_req;
	int32			 rc = NFX_OK;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_sndUIReq2FP.\n"));

    bzero((void*)&ndmp_req, sizeof(ndmp_NFXReq_t));
    ndmp_req.hdr.mid        = msgId;
    ndmp_req.hdr.sessId     = sess->sessId;
    ndmp_req.hdr.clientHdl  = (uint32) rq_info; 	/* saved request info so we know who to reply bk */
 
    if (sendAgileMsg (sess->Agilesock, &ndmp_req, sizeof(ndmp_NFXReq_t),
                        "ndmp",       /* NDMP on the N-NIM */
                         NFX_NFP_FP2, 1, MGMTPLANE, /* cpu, slot, plane */
                   		 NULL, NULL, 0) < 0)
    {
        NDMP_DBG_PRINTF(D_INFO, ("ndmp_sndUIReq2FP:  Agile message send to FP failed\n"));
        rc = NFX_ERR;
    }               
	return rc;
}  

/*-----------------------------------------------------------------
 * Name: ndmp_Child_ProcShowStatusReq
 *
 * Description:	Responding to NDMP_MID_UI_SHOW_STATUS or 
 *				NDMP_MID_SHOW_DETAIL request from UI process.
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/23/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_Child_ProcShowStatusReq(ndmpSess_t *sess, char *rcvbuf, ndmp_src_t *src)
{      
	ndmp_Req_info_t	*rq_info;
	ndmp_NFXShowReq_t *req = (ndmp_NFXShowReq_t *) rcvbuf;


	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Child_ProcShowStatusReq:\n"));

	rq_info = save_req_info(src->app, src->cpu, src->slot, req->mid);

	/* Check whether UI really want to send to this session */
	if (req->sessId != sess->sessId)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_Child_ProcShowStatusReq:  Invalid sessId received %d from UI.  My sessId %d\n", req->sessId, sess->sessId));

		send_UIStatRsp(sess, rq_info, NFX_ERR);
	}
	if (sess->authed)
	{
		ndmp_sndUIReq2FP(sess, NDMP_MID_UI_DM_GET_STATE, rq_info);
	} else
		send_UIStatRsp(sess, rq_info, NFX_OK);

	return NFX_OK;
}
/*-----------------------------------------------------------------
 * Name: ndmp_Child_ProcShowStatisticReq
 *
 * Description:	Responding to NDMP_MID_UI_SHOW_STAT(istic)S 
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/23/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_Child_ProcShowStatisticReq(ndmpSess_t *sess, char *rcvbuf, ndmp_src_t *src)
{      
	ndmp_Req_info_t	*rq_info;
	ndmp_NFXShowReq_t *req = (ndmp_NFXShowReq_t *) rcvbuf;


	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Child_ProcShowStatisticReq:\n"));

	rq_info = save_req_info(src->app, src->cpu, src->slot, req->mid);

	/* Check whether UI really want to send to this session */
	if (req->sessId != sess->sessId)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_Child_ProcShowStatisticReq:  Invalid sessId received %d from UI.  My sessId %d\n", req->sessId, sess->sessId));

		send_UIChldStatisticRsp(sess, rq_info, NFX_ERR);
	} else
		send_UIChldStatisticRsp(sess, rq_info, NFX_OK);

	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name: ndmp_Child_ProcDelSessReq
 *
 * Description:	Probes info on a session
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_Child_ProcDelSessReq(ndmpSess_t *sess, char *rcvbuf, ndmp_src_t *src)
{
    ndmp_NFXDelSessReq_t   *req = (ndmp_NFXDelSessReq_t *)rcvbuf;
	ndmp_NFXDelSessRsp_t   rsp;
	char	buf[200];

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Child_ProcDelSessReq:\n"));

	sprintf(buf, "Session %d is aborted on NDMP server : %s\n", sess->sessId, (strlen(req->reason))?req->reason:"");
    ndmp_sndLogMsg( sess, NDMP_LOG_WARNING, buf, 0);
	E_LOG(class_1, err_s, eee_getApplicationId("ndmp"), 0, 0, 0, (buf));

	if (sess->wait_tape_open_rsp)
	{
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_Child_ProcDelSessReq:  Set conn_await_close TRUE\n"));
		sess->conn_await_close = TRUE;
	} else {
		NDMP_DBG_PRINTF(D_INFO, ("ndmp_Child_ProcDelSessReq:  Set conn_closed TRUE\n"));
		sess->conn_closed = TRUE;
	}

    rsp.mid = NDMP_MID_UI_DEL_SESS;
    rsp.stat = NFX_OK;

    send_rsp_to_UI(&rsp, sizeof(ndmp_NFXDelSessRsp_t), sess->Agilesock,
            src->app, src->cpu, src->slot);

    /*
     * Delete the dumpdate record
     */
    if (strlen(sess->Serv.Data.dd.snapshot))
    {
        char dumpLink[1024], dumpFileName[1024];
        int32 flen;

        bzero(dumpLink, sizeof(char)*1024);
        bzero(dumpFileName, sizeof(char)*1024);
        sprintf(dumpLink, "%s%lld.dmp", dumpDateDir, sess->Serv.Data.fs_volId);
        if (!((flen = readlink(dumpLink, dumpFileName, 1024)) == -1))
            ndmp_deleteDumpRecord(dumpFileName, sess->Serv.Data.dd.snapshot);
        bzero(&sess->Serv.Data.dd, sizeof(sess->Serv.Data.dd));
    }

	return NFX_OK;
}
/*-----------------------------------------------------------------
 * Name: ndmp_Child_ProcSetDbgLevel
 *
 * Description:	
 *
 * Created by:  Anh Vu
 *
 * Date Created:        02/12/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_Child_ProcSetDbgLevel(ndmpSess_t *sess, char *rcvbuf, ndmp_src_t *src)
{

	ndmp_NFXSetDebugReq_t 	*req = (ndmp_NFXSetDebugReq_t *) rcvbuf;
	ndmp_NFXSetDebugRsp_t	rsp;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Child_ProcSetDbgLevel:  sesid %d set to level %d\n",
		sess->sessId, req->dbg_level));

	ndmp_dbg_level = req->dbg_level;  

	rsp.mid = NDMP_MID_UI_SET_DBG_LEVEL;
    rsp.stat = NFX_OK;
    rsp.sessId = sess->sessId;

    send_rsp_to_UI(&rsp, sizeof(ndmp_NFXSetDebugRsp_t), sess->Agilesock,
            src->app, src->cpu, src->slot);

	return NFX_OK;
}
/*-----------------------------------------------------------------
 * Name: ndmp_Child_ProcSetTraceFlag
 *
 * Description:	
 *
 * Created by:  Anh Vu
 *
 * Date Created:        10/05/02
 *
 *-----------------------------------------------------------------
 */
int
ndmp_Child_ProcSetTraceFlag(ndmpSess_t *sess, char *rcvbuf, ndmp_src_t *src)
{

	ndmp_NFXSMsgTraceReq_t 	*req = (ndmp_NFXSMsgTraceReq_t *) rcvbuf;
	ndmp_NFXSMsgTraceRsp_t	rsp;

	NDMP_DBG_PRINTF(D_INFO, ("ndmp_Child_ProcSetTraceFlag:  sesid %d set to  %s\n",
		sess->sessId, req->flag?"ON":"OFF"));

	ndmp_trace_msg = req->flag;  

	rsp.mid = NDMP_MID_UI_SET_TRACE_ON_OFF;
    rsp.stat = NFX_OK;
    rsp.sessId = sess->sessId;

    send_rsp_to_UI(&rsp, sizeof(ndmp_NFXSMsgTraceRsp_t), sess->Agilesock,
            src->app, src->cpu, src->slot);

	return NFX_OK;
}

