/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        ndmp-table.c
 *
 * RCS:  $Id: ndmp-table.c,v 1.15 2002/10/10 01:15:08 anhv Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:    Cmd table for NDMP v3 messages received from NDMP
 *                client.  
 *
 * Created by: Anh Vu
 *
 * Date Created:        8/20/01
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <nfx-types.h>  
#include "ndmp.h"
#include "ndmp-ssc.h"

#define NO_AUTH_REQ 0
#define AUTH_REQ    1


/* Possible msgs received on SSC side from NDMP client. */

/* Each message received is associated w/ a function to process.
 * If there is no function for the message received, then that message
 * is replied w/ NOT_SUPPORTED_ERR.  The client could be sending us
 * a wrong message, for ex: Notify or Log messages, which are messages
 * that we xmit only, not receive.
 */

/* MsgEntry has :
 * - msgid
 * - authReqd flag
 * - xdr decode func for requests
 * - xdr encode func for responses
 * - func to process Requests
 * - msg name
 * - stat_idx_snt to FP or other process on SSC
 */
ndmpMsgEnt_t ndmpMsgTbl[] =
{
    /***** Management service commands *****/
    /* Connect interface commands */
    {
        NDMP_CONNECT_OPEN, NO_AUTH_REQ, xdr_ndmp_conn_open_req, 
        xdr_ndmp_conn_open_rsp, ndmp_procConnOpenReq, "NDMP_CONNECT_OPEN",
        prt_xdr_ndmp_conn_open_req, prt_xdr_ndmp_conn_open_rsp, -1
    },
    {
        NDMP_CONNECT_CLOSE, NO_AUTH_REQ, NULL, NULL, ndmp_procConnCloseReq,
        "NDMP_CONNECT_CLOSE",
        NULL, NULL, -1
    },
    {
        NDMP_CONNECT_CLIENT_AUTH, NO_AUTH_REQ, xdr_ndmp_conn_client_auth_req, 
        xdr_ndmp_conn_client_auth_rsp, ndmp_procConnClientAuthReq,
        "NDMP_CONNECT_CLIENT_AUTH",
        prt_xdr_ndmp_conn_client_auth_req, prt_xdr_ndmp_conn_client_auth_rsp, -1
    },
    {
        NDMP_CONNECT_SERVER_AUTH, NO_AUTH_REQ, xdr_ndmp_conn_srvr_auth_req, 
        xdr_ndmp_conn_srvr_auth_rsp, ndmp_procConnSrvrAuthReq,
        "NDMP_CONNECT_SERVER_AUTH",
        prt_xdr_ndmp_conn_srvr_auth_req, prt_xdr_ndmp_conn_srvr_auth_rsp, -1
    },

    /* Configuration interface commands */
    {
        NDMP_CONFIG_GET_HOST_INFO, NO_AUTH_REQ, NULL,
        xdr_ndmp_cfg_get_host_info_rsp, ndmp_procCfgGetHostInfoReq,
        "NDMP_CONFIG_GET_HOST_INFO",
        NULL, prt_xdr_ndmp_cfg_get_host_info_rsp, -1
    },
    {
        NDMP_CONFIG_GET_SERVER_INFO, NO_AUTH_REQ, NULL,
        xdr_ndmp_cfg_get_srvr_info_rsp, ndmp_procCfgGetSrvrInfoReq,
        "NDMP_CONFIG_GET_SERVER_INFO",
        NULL, prt_xdr_ndmp_cfg_get_srvr_info_rsp, -1
    },
    {
        NDMP_CONFIG_GET_CONNECTION_TYPE, NO_AUTH_REQ, NULL,
        xdr_ndmp_cfg_get_conn_type_rsp, 
        ndmp_procCfgGetConnTypeReq, "NDMP_CONFIG_GET_CONNECTION_TYPE",
        NULL, prt_xdr_ndmp_cfg_get_conn_type_rsp, -1
    },
    {
        NDMP_CONFIG_GET_AUTH_ATTR, NO_AUTH_REQ, xdr_ndmp_cfg_get_auth_attr_req,
        xdr_ndmp_cfg_get_auth_attr_rsp, ndmp_procCfgGetAuthAttrReq,
        "NDMP_CONFIG_GET_AUTH_ATTR",
        prt_xdr_ndmp_cfg_get_auth_attr_req, prt_xdr_ndmp_cfg_get_auth_attr_rsp, -1
    },
    {
        NDMP_CONFIG_GET_BUTYPE_INFO,  NO_AUTH_REQ, NULL,
        xdr_ndmp_cfg_get_butype_info_rsp, 
        ndmp_procCfgGetButypeInfoReq, "NDMP_CONFIG_GET_BUTYPE_INFO",
        NULL, prt_xdr_ndmp_cfg_get_butype_info_rsp, -1
    },
    {
        NDMP_CONFIG_GET_FS_INFO, NO_AUTH_REQ, NULL,
        xdr_ndmp_cfg_get_fs_info_rsp, 
        ndmp_procCfgGetFsInfoReq, "NDMP_CONFIG_GET_FS_INFO",
        NULL, prt_xdr_ndmp_cfg_get_fs_info_rsp, -1
    },
    {
        NDMP_CONFIG_GET_TAPE_INFO, NO_AUTH_REQ, NULL,
        xdr_ndmp_cfg_get_tape_info_rsp, 
        ndmp_procCfgGetTapeInfoReq, "NDMP_CONFIG_GET_TAPE_INFO",
        NULL, prt_xdr_ndmp_cfg_get_tape_info_rsp, TAPE_GET_DEVLST_REQ_SNT
    },
    {
        NDMP_CONFIG_GET_SCSI_INFO, NO_AUTH_REQ, NULL,
        xdr_ndmp_cfg_get_scsi_info_rsp,
        ndmp_procCfgGetScsiInfoReq, "NDMP_CONFIG_GET_SCSI_INFO",
        NULL, prt_xdr_ndmp_cfg_get_scsi_info_rsp, MC_GET_DEVLST_REQ_SNT
    },

    /* Notify interface commands */
    {
        NDMP_NOTIFY_DATA_HALTED, AUTH_REQ, xdr_ndmp_notify_data_halted_req,
        NULL, NULL, "NDMP_NOTIFY_DATA_HALTED",
        prt_xdr_ndmp_notify_data_halted_req, NULL, -1
    },
    {
        NDMP_NOTIFY_CONNECTED, NO_AUTH_REQ, xdr_ndmp_notify_conn_req,
        NULL, NULL, "NDMP_NOTIFY_CONNECTED",
        prt_xdr_ndmp_notify_conn_req, NULL, -1
    },
    {
        NDMP_NOTIFY_MOVER_HALTED, AUTH_REQ, xdr_ndmp_notify_mover_halted_req,
        NULL, NULL, "NDMP_NOTIFY_MOVER_HALTED",
        prt_xdr_ndmp_notify_mover_halted_req, NULL, -1
    },
    {
        NDMP_NOTIFY_MOVER_PAUSED, AUTH_REQ, xdr_ndmp_notify_mover_paused_req,
        NULL, NULL, "NDMP_NOTIFY_MOVER_PAUSED",
        prt_xdr_ndmp_notify_mover_paused_req, NULL, -1
    },
    {
        NDMP_NOTIFY_DATA_READ, AUTH_REQ, xdr_ndmp_notify_data_read_req,
        NULL, NULL, "NDMP_NOTIFY_DATA_READ",
        prt_xdr_ndmp_notify_data_read_req, NULL, -1
    },

    /* Log interface commands */
    {
        NDMP_LOG_FILE, AUTH_REQ, xdr_ndmp_log_file_req,
        NULL, NULL, "NDMP_LOG_FILE",
        prt_xdr_ndmp_log_file_req, NULL, -1
    },
    {
        NDMP_LOG_MESSAGE, AUTH_REQ, xdr_ndmp_log_msg_req,
        NULL, NULL, "NDMP_LOG_MESSAGE",
        prt_xdr_ndmp_log_msg_req, NULL, -1
    },

    /* File History interface commands */
    {
        NDMP_FH_ADD_FILE, AUTH_REQ, xdr_ndmp_fh_add_file_req, NULL, NULL, 
        "NDMP_FH_ADD_FILE", NULL, NULL, -1
    },
    {
        NDMP_FH_ADD_DIR, AUTH_REQ, xdr_ndmp_fh_add_dir_req, NULL, NULL,
        "NDMP_FH_ADD_DIR", NULL, NULL, -1
    },
    {
        NDMP_FH_ADD_NODE, AUTH_REQ, xdr_ndmp_fh_add_node_req, NULL, NULL,
        "NDMP_FH_ADD_NODE", NULL, NULL, -1
    },
    /***** SCSI service commands *****/
    {
        NDMP_SCSI_OPEN, AUTH_REQ, xdr_ndmp_scsi_open_req,
        xdr_ndmp_scsi_open_rsp, ndmp_procScsiOpenReq, "NDMP_SCSI_OPEN",
        prt_xdr_ndmp_scsi_open_req, prt_xdr_ndmp_scsi_open_rsp, MC_OPEN_REQ_SNT
    },
    {
        NDMP_SCSI_CLOSE, AUTH_REQ, NULL, xdr_ndmp_scsi_close_rsp, 
        ndmp_procScsiCloseReq, "NDMP_SCSI_CLOSE",
        NULL, prt_xdr_ndmp_scsi_close_rsp, MC_CLOSE_REQ_SNT
    },
    {
        NDMP_SCSI_EXECUTE_CDB, AUTH_REQ, xdr_ndmp_exec_cdb_req,
        xdr_ndmp_exec_cdb_rsp, ndmp_procScsiExecCdbReq, "NDMP_SCSI_EXECUTE_CDB",
        prt_xdr_ndmp_scsi_exec_cdb_req, prt_xdr_ndmp_scsi_exec_cdb_rsp,
		MC_PASSTHRU_REQ_SNT
    },
    {
        NDMP_SCSI_GET_STATE, AUTH_REQ, NULL, xdr_ndmp_scsi_get_state_rsp, 
        ndmp_procScsiGetStateReq, "NDMP_SCSI_GET_STATE",
        NULL, prt_xdr_ndmp_scsi_get_state_rsp, -1
    },
    {
        NDMP_SCSI_RESET_DEVICE, AUTH_REQ, NULL, xdr_ndmp_scsi_reset_dev_rsp, 
        ndmp_procScsiResetDevReq, "NDMP_SCSI_RESET_DEVICE",
        NULL, prt_xdr_ndmp_scsi_reset_dev_rsp, -1
    },
    {
        NDMP_SCSI_RESET_BUS, AUTH_REQ, NULL, xdr_ndmp_scsi_reset_bus_rsp, 
        ndmp_procScsiResetBusReq, "NDMP_SCSI_RESET_BUS",
        NULL, prt_xdr_ndmp_scsi_reset_bus_rsp, -1
    },
    {
        NDMP_SCSI_SET_TARGET, AUTH_REQ, NULL, xdr_ndmp_scsi_set_target_rsp, 
        ndmp_procScsiSetTargetReq, "NDMP_SCSI_SET_TARGET",
        NULL, prt_xdr_ndmp_scsi_set_target_rsp, -1
    },
    /***** Tape service commands *****/
    {
        NDMP_TAPE_OPEN, AUTH_REQ, xdr_ndmp_tape_open_req,
        xdr_ndmp_tape_open_rsp, 
        ndmp_procTapeOpenReq, "NDMP_TAPE_OPEN",
        prt_xdr_ndmp_tape_open_req, prt_xdr_ndmp_tape_open_rsp,
		TAPE_OPEN_REQ_SNT
    },
    {
        NDMP_TAPE_CLOSE, AUTH_REQ, NULL, xdr_ndmp_tape_close_rsp, 
        ndmp_procTapeCloseReq, "NDMP_TAPE_CLOSE",
        NULL, prt_xdr_ndmp_tape_close_rsp, TAPE_CLOSE_REQ_SNT
    },
    {
        NDMP_TAPE_GET_STATE, AUTH_REQ, NULL, xdr_ndmp_tape_get_state_rsp, 
        ndmp_procTapeGetStateReq, "NDMP_TAPE_GET_STATE",
        NULL, prt_xdr_ndmp_tape_get_state_rsp, TAPE_GET_STATE_REQ_SNT
    },
    {
        NDMP_TAPE_MTIO, AUTH_REQ, xdr_ndmp_tape_mtio_req,
        xdr_ndmp_tape_mtio_rsp, 
        ndmp_procTapeMtioReq, "NDMP_TAPE_MTIO",
        prt_xdr_ndmp_tape_mtio_req, prt_xdr_ndmp_tape_mtio_rsp,
		-1
    },
    {
        NDMP_TAPE_WRITE, AUTH_REQ, xdr_ndmp_tape_write_req,
        xdr_ndmp_tape_write_rsp, 
        ndmp_procTapeWriteReq, "NDMP_TAPE_WRITE",
        prt_xdr_ndmp_tape_write_req, prt_xdr_ndmp_tape_write_rsp,
		TAPE_WR_REQ_SNT
    },
    {
        NDMP_TAPE_READ, AUTH_REQ, xdr_ndmp_tape_read_req,
        xdr_ndmp_tape_read_rsp, 
        ndmp_procTapeReadReq, "NDMP_TAPE_READ",
        prt_xdr_ndmp_tape_read_req, prt_xdr_ndmp_tape_read_rsp,
		TAPE_RD_REQ_SNT
    },
    {
        NDMP_TAPE_EXECUTE_CDB, AUTH_REQ, xdr_ndmp_exec_cdb_req,
        xdr_ndmp_exec_cdb_rsp, 
        ndmp_procTapeExecCdbReq, "NDMP_TAPE_EXECUTE_CDB",
        prt_xdr_ndmp_tape_exec_cdb_req, prt_xdr_ndmp_tape_exec_cdb_rsp,
		TAPE_PASSTHRU_REQ_SNT
    },
    /***** Data service commands *****/
    {
        NDMP_DATA_GET_STATE, AUTH_REQ, NULL, xdr_ndmp_data_get_state_rsp, 
        ndmp_procDataGetStateReq, "NDMP_DATA_GET_STATE",
        NULL, prt_xdr_ndmp_data_get_state_rsp, FP_DATA_GET_STATE_REQ_SNT
    },
    {
        NDMP_DATA_START_BACKUP, AUTH_REQ, xdr_ndmp_data_start_bkup_req,
        xdr_ndmp_data_start_bkup_rsp, 
        ndmp_procDataStartBkupReq, "NDMP_DATA_START_BACKUP",
        prt_xdr_ndmp_data_start_bkup_req, prt_xdr_ndmp_data_start_bkup_rsp,
		FP_DATA_START_BKUP_REQ_SNT
    },
    {
        NDMP_DATA_START_RECOVER, AUTH_REQ, xdr_ndmp_data_start_recover_req,
        xdr_ndmp_data_start_recover_rsp,
        ndmp_procDataStartRecovReq, "NDMP_DATA_START_RECOVER",
        prt_xdr_ndmp_data_start_recover_req,prt_xdr_ndmp_data_start_recover_rsp,
		FP_DATA_START_RST_REQ_SNT
    },
    {
        NDMP_DATA_ABORT, AUTH_REQ, NULL, xdr_ndmp_data_abort_rsp,
        ndmp_procDataAbortReq, "NDMP_DATA_ABORT",
        NULL, prt_xdr_ndmp_data_abort_rsp, FP_DATA_ABORT_REQ_SNT
    },
    {
        NDMP_DATA_GET_ENV, AUTH_REQ, NULL, xdr_ndmp_data_get_env_rsp,
        ndmp_procDataGetEnvReq, "NDMP_DATA_GET_ENV",
        NULL, prt_xdr_ndmp_data_get_env_rsp, FP_DATA_GET_ENV_REQ_SNT
    },
    {
        NDMP_DATA_STOP, AUTH_REQ, NULL, xdr_ndmp_data_stop_rsp,
        ndmp_procDataStopReq, "NDMP_DATA_STOP",
        NULL, prt_xdr_ndmp_data_stop_rsp, FP_DATA_STOP_REQ_SNT
    },
    {
        NDMP_DATA_LISTEN, AUTH_REQ, xdr_ndmp_data_listen_req,
        xdr_ndmp_data_listen_rsp, ndmp_procDataListenReq, "NDMP_DATA_LISTEN",
        prt_xdr_ndmp_data_listen_req, prt_xdr_ndmp_data_listen_rsp, FP_DATA_LISTEN_REQ_SNT
    },
    {
        NDMP_DATA_CONNECT, AUTH_REQ, xdr_ndmp_data_conn_req,
        xdr_ndmp_data_conn_rsp, ndmp_procDataConnReq, "NDMP_DATA_CONNECT",
        prt_xdr_ndmp_data_conn_req, prt_xdr_ndmp_data_conn_rsp, FP_DATA_CONN_REQ_SNT
    },


    /***** Mover service commands *****/
    {
        NDMP_MOVER_GET_STATE, AUTH_REQ, NULL, xdr_ndmp_mover_get_state_rsp, 
        ndmp_procMoverGetStateReq, "NDMP_MOVER_GET_STATE",
        NULL, NULL, FP_MOVER_GET_STATE_REQ_SNT
    },
    {
        NDMP_MOVER_LISTEN, AUTH_REQ, xdr_ndmp_mover_listen_req,
        xdr_ndmp_mover_listen_rsp, ndmp_procMoverListenReq, "NDMP_MOVER_LISTEN",
        prt_xdr_ndmp_mover_listen_req, prt_xdr_ndmp_mover_listen_rsp, FP_MOVER_LISTEN_REQ_SNT
    },
    {
        NDMP_MOVER_CONTINUE, AUTH_REQ, NULL, xdr_ndmp_mover_cont_rsp, 
        ndmp_procMoverContReq, "NDMP_MOVER_CONTINUE",
        NULL, prt_xdr_ndmp_mover_cont_rsp, FP_MOVER_CONT_REQ_SNT
    },
    {
        NDMP_MOVER_ABORT, AUTH_REQ, NULL, xdr_ndmp_mover_abort_rsp, 
        ndmp_procMoverAbortReq, "NDMP_MOVER_ABORT",
        NULL, prt_xdr_ndmp_mover_abort_rsp, FP_MOVER_ABORT_REQ_SNT
    },
    {
        NDMP_MOVER_STOP, AUTH_REQ, NULL, xdr_ndmp_mover_stop_rsp, 
        ndmp_procMoverStopReq, "NDMP_MOVER_STOP",
        NULL, prt_xdr_ndmp_mover_stop_rsp, FP_MOVER_STOP_REQ_SNT
    },
    {
        NDMP_MOVER_SET_WINDOW, AUTH_REQ, xdr_ndmp_mover_set_window_req,
        xdr_ndmp_mover_set_window_rsp, ndmp_procMoverSetWdwReq,
        "NDMP_MOVER_SET_WINDOW",
        prt_xdr_ndmp_mover_set_window_req, prt_xdr_ndmp_mover_set_window_rsp, 
		FP_MOVER_SET_WDW_REQ_SNT
    },
    {
        NDMP_MOVER_READ, AUTH_REQ, xdr_ndmp_mover_read_req,
        xdr_ndmp_mover_read_rsp, 
        ndmp_procMoverReadReq, "NDMP_MOVER_READ",
        prt_xdr_ndmp_mover_read_req, prt_xdr_ndmp_mover_read_rsp, FP_MOVER_READ_REQ_SNT
    },
    {
        NDMP_MOVER_CLOSE, AUTH_REQ, NULL, xdr_ndmp_mover_close_rsp, 
        ndmp_procMoverCloseReq, "NDMP_MOVER_CLOSE",
        NULL, prt_xdr_ndmp_mover_close_rsp, FP_MOVER_CLOSE_REQ_SNT
    },
    {
        NDMP_MOVER_SET_RECORD_SIZE, AUTH_REQ,
        xdr_ndmp_mover_set_record_size_req,
        xdr_ndmp_mover_set_record_size_rsp, ndmp_procMoverSetRecSizReq,
        "NDMP_MOVER_SET_RECORD_SIZE",
        prt_xdr_ndmp_mover_set_record_size_req,
        prt_xdr_ndmp_mover_set_record_size_rsp, FP_MOVER_SET_RECSIZ_REQ_SNT
    },
    {
        NDMP_MOVER_CONNECT, AUTH_REQ, xdr_ndmp_mover_conn_req,
        xdr_ndmp_mover_conn_rsp, 
        ndmp_procMoverConnReq, "NDMP_MOVER_CONNECT",
        prt_xdr_ndmp_mover_conn_req, prt_xdr_ndmp_mover_conn_rsp, FP_MOVER_CONN_REQ_SNT
    },
    {
        0, 0, 0, 0, 0, 0, 0, 0, 0
    }
};

