/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * RCS:  $Id: nb-msg.c,v 1.1.1.1 2002/04/23 17:52:14 laurentb Exp $
 *
 * Name			: nb-msg.c
 *
 * Description	: NetBIOS message encoding and decoding processing.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 11/07/2001
 *
 *-----------------------------------------------------------------
 */

#include "nb-api.h"
#include "nb-ns.h"
#include "nb-ss.h"
#include "nb-ds.h"
#include "nb-msg.h"

#define IFDBG(lvl)	if (lvl <= nb_debugLevel)

#define NB_BOOLSTR(b) ((b) ? "Yes" : "No")		// True or False

typedef struct nb_msgString
{
	uint16	val;
	uchar8	*str;
} nb_msgString_t;

static nb_msgString_t nb_opcodeStrings[] = {
				{NB_NS_OPCODE_QUERY, "Query"},
				{NB_NS_OPCODE_REG, "Registration"},
				{NB_NS_OPCODE_RELEASE, "Release"},
				{NB_NS_OPCODE_WACK, "WACK"},
				{NB_NS_OPCODE_REFRESH_8, "Refresh"},
				{NB_NS_OPCODE_REFRESH_9, "Refresh"},
				{-1, 0,}
};

static nb_msgString_t nb_nodeTypeStrings[] = {
				{NB_NS_NBFLAG_BNODE, "B node"},
				{NB_NS_NBFLAG_PNODE, "P node"},
				{NB_NS_NBFLAG_MNODE, "M node"},
				{NB_NS_NBFLAG_HNODE, "H node"},
				{-1, 0}
};

static nb_msgString_t nb_sessionTypeStrings[] = {
				{NB_SS_TYPE_MSG, "SESSION MESSAGE"},
				{NB_SS_TYPE_REQ, "SESSION REQUEST"},
				{NB_SS_TYPE_POSRSP, "POSITIVE SESSION RESPONSE"},
				{NB_SS_TYPE_NEGRSP, "NEGATIVE SESSION RESPONSE"},
				{NB_SS_TYPE_RETRSP, "RETARGET SESSION RESPONSE"},
				{NB_SS_TYPE_KEEPALIVE, "SESSION KEEP ALIVE"},
				{-1, 0}
};

static nb_msgString_t nb_sessionErrCodeStrings[] = {
		{NB_SS_ERR_NOT_LISTEN_CALLEDNAME, "Not listening on called name"},
		{NB_SS_ERR_NOT_LISTEN_CALLINGNAME, "Not listening on calling name"},
		{NB_SS_ERR_CALLEDNAME_NOT_PRESENT, "Called name not present"},
		{NB_SS_ERR_INSUFFICIENT_RSRC, "Called name present, but insufficient resources"},
		{NB_SS_ERR_UNSPECIFIED, "Unspecified error"},
		{-1, 0}
};

static nb_msgString_t nb_datagramTypeStrings[] = {
		{NB_DS_DGM_TYPE_DIR_UNIQUE, "Direct-unique datagram"},
		{NB_DS_DGM_TYPE_DIR_GROUP, "Direct-group datagram"},
		{NB_DS_DGM_TYPE_BROADCAST, "Broadcast datagram"},
		{NB_DS_DGM_TYPE_ERR, "Datagram error"},
		{NB_DS_DGM_TYPE_QUERY_REQ, "Datagram query request"},
		{NB_DS_DGM_TYPE_POS_QUERY_RSP, "Datagram positive query response"},
		{NB_DS_DGM_TYPE_NEG_QUERY_RSP, "Datagram negative query response"},
		{-1, 0}
};

static nb_msgString_t nb_datagramNodeTypeStrings[] = {
		{NB_DS_NODE_TYPE_B, "B node"},
		{NB_DS_NODE_TYPE_P, "P node"},
		{NB_DS_NODE_TYPE_H, "H node"},
		{NB_DS_NODE_TYPE_NBDD, "NBDD"},
		{-1, 0}
};

extern void memDump(void *address, uint32 size);
extern void strtoupper (char8 *p);

/*-----------------------------------------------------------------------------
 * Name			: nb_getMsgString
 *
 * Description	: Get NetBIOS message string depending on the integer value 
 * 				  of the message. 
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 11/24/2001
 *-----------------------------------------------------------------------------
 */
static uchar8 *		// Message string
nb_getMsgString(nb_msgString_t	*msgStrings,	// Message strings array
				uint16			val,			// Integer for the string 
				uchar8			*unknownStr)	// String for invalid value
{
	nb_msgString_t	*msgString;
	uint32			i;

	for (i = 0; msgStrings[i].str != 0; i++)
	{
		msgString = &msgStrings[i];
		if (val == msgString->val)
			return (msgString->str);
	}
	return (unknownStr);
}

/*-----------------------------------------------------------------------------
 * Name			: nb_getNameWithType
 *
 * Description	: Get NetBIOS name with type.
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 11/24/2001
 *-----------------------------------------------------------------------------
 */
void
nb_getNameWithType(nb_name_t *name, uchar8 *nameStr)
{
	if (name->name == NULL)
	{
		*nameStr = 0;
		return;
	}
	sprintf(nameStr, "%s<%02X>", name->name, name->type);
	return;
}

/*-----------------------------------------------------------------------------
 * Name			: nb_displayRsrcMsg
 *
 * Description	: Display name service resource record message for debugging.
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 11/24/2001
 *-----------------------------------------------------------------------------
 */
void
nb_displayRsrcMsg(nb_nsResource_t *rsrcP)
{
	uint8	nbFlags;
	uchar8	nameStr[NB_S_NAME + 10];

	nb_getNameWithType(&rsrcP->rrName, nameStr);
	printf("    rr name = %s\n", nameStr);
	printf("    type = 0x%X, class = 0x%X\n", rsrcP->rrType, rsrcP->rrClass);
	printf("    ttl = %d (seconds)\n", (int32)(rsrcP->ttl / 1000));
	printf("    length = %d\n", rsrcP->rdLength);

	if (rsrcP->rdLength == 0 || rsrcP->rdata == NULL)
		return;

	printf("    Node flags: %s, group = %s\n",
			nb_getMsgString(nb_nodeTypeStrings,
							rsrcP->rdata[0],
							"Unknown node type"),
			NB_BOOLSTR((rsrcP->rdata[0] & NB_NS_NBFLAG_GROUP)));

	printf("    rdata+2:\n");
	memDump(&(rsrcP->rdata[2]), (rsrcP->rdLength) - 2);

	return; 
}

/*-----------------------------------------------------------------------------
 * Name			: nb_displayDecodedNsMsg
 *
 * Description	: Display decoded NetBIOS name service message for debugging.
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 11/24/2001
 *-----------------------------------------------------------------------------
 */
void
nb_displayDecodedNsMsg(nb_nsPacket_t	*nsPktP)
{
	nb_nsHeader_t	*headerP;
	nb_nsQuestion_t	*questionP;
	nb_nsResource_t	*rsrcP;
	uchar8			nameStr[NB_S_NAME + 10];

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_displayDecodedNsMsg()"));

	printf("\n----- NetBIOS Name Service Message -----\n");

	/*-------------
	 * Print header
	 *-------------
	 */
	headerP = &nsPktP->header;
	printf("Header:\n");
	printf("    Transaction ID = 0x%X, opcode = %s, response = %s\n",
			headerP->nameTransID,
			nb_getMsgString(nb_opcodeStrings,
							headerP->opcode,
							"Unknown opcode"),
			NB_BOOLSTR(headerP->rsp));
	printf("    Flags: broadcast: %s\n", NB_BOOLSTR(headerP->nmFlags.bcast));
	printf("           recursion available: %s\n", NB_BOOLSTR(headerP->nmFlags.recAval));
	printf("           recursion desired: %s\n", NB_BOOLSTR(headerP->nmFlags.recDesired));
	printf("           truncation: %s\n", NB_BOOLSTR(headerP->nmFlags.trunc));
	printf("           authoritative answer: %s\n", NB_BOOLSTR(headerP->nmFlags.auth));
	printf("    rcode = 0x%x\n", headerP->rcode);
	printf("    Question count = %d, Answer count = %d\n",
			headerP->qdCount, headerP->anCount);
	printf("    Authority count = %d, Additional count = %d\n",
			headerP->nsCount, headerP->arCount);

	/*-----------------------
	 * Print question section
	 *-----------------------
	 */
	if (headerP->qdCount)
	{
		questionP = &nsPktP->question;
		printf("Question section:\n");
		nb_getNameWithType(&questionP->qName, nameStr);
		printf("    name = %s\n", nameStr);
		printf("    type = 0x%X, class = 0x%X\n", questionP->qType, questionP->qClass);
	}

	/*-------------------------------------
	 * Print answer resource record section
	 *-------------------------------------
	 */
	rsrcP = nsPktP->answer;
	if (rsrcP && headerP->anCount)
	{
		printf("Answer record section:\n");
		nb_displayRsrcMsg(rsrcP);
	}

	/*----------------------------------------
	 * Print authority resource record section
	 *----------------------------------------
	 */
	rsrcP = nsPktP->authority;
	if (rsrcP && headerP->nsCount)
	{
		printf("Authority record section:\n");
		nb_displayRsrcMsg(rsrcP);
	}

	/*-----------------------------------------
	 * Print additional resource record section
	 *-----------------------------------------
	 */
	rsrcP = nsPktP->additional;
	if (rsrcP && headerP->arCount)
	{
		printf("Additional record section:\n");
		nb_displayRsrcMsg(rsrcP);
	}

	printf("------------ End of Message ------------\n\n");

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_displayDecodedNsMsg()"));

	return;
} /* nb_displayDecodedNsMsg() */

/*-----------------------------------------------------------------------------
 * Name			: nb_displayDecodedSsMsg
 *
 * Description	: Display decoded NetBIOS session service message for debugging.
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 12/04/2001
 *-----------------------------------------------------------------------------
 */
void
nb_displayDecodedSsMsg(uchar8	*buf)
{
	int32		currOffset;
	uchar8		ssnType;
	uint32		ssnLen, nameLen;
	nb_name_t	name;
	uchar8		nameStr[NB_S_NAME + 10];

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_displayDecodedSsMsg()"));

	ssnType = (uchar8)buf[0];
	ssnLen = NB_SS_GETLEN(buf);

	printf("\n----- NetBIOS Session Service Message -----\n");

	printf("    Type = 0x%x (%s)\n", ssnType,
									 nb_getMsgString(nb_sessionTypeStrings,
													 ssnType,
													 "Unknown session type"));
	printf("    Total session packet length = %d\n", ssnLen);
	
	/* Displaying different msg depending on session type */
	switch(ssnType)
	{
		case NB_SS_TYPE_MSG:
			printf("    (%d bytes of user data)\n", ssnLen);
			break;

		case NB_SS_TYPE_REQ:
			/* Get called name */
			currOffset = 0;
			nameLen = nb_parseNameMsg(buf + 4, currOffset, ssnLen, &name);
			if (nameLen == 0)
			{
				NB_PRINT(NB_DBGLV_ERR, ("nb_displayDecodedSsMsg: "
							"Error in parsing called name"));
				break;
			}
			nb_getNameWithType(&name, nameStr);
			printf("    Called name = %s\n", nameStr);

			/* Get calling name */
			currOffset += nameLen;
			nameLen = nb_parseNameMsg(buf + 4, currOffset, ssnLen, &name);
			if (nameLen == 0)
			{
				NB_PRINT(NB_DBGLV_ERR, ("nb_displayDecodedSsMsg: "
							"Error in parsing calling name"));
				break;
			}
			nb_getNameWithType(&name, nameStr);
			printf("    Calling name = %s\n", nameStr);
			break;

		case NB_SS_TYPE_NEGRSP:
			printf("    Error = 0x%x (%s)\n",
							buf[4],
							nb_getMsgString(nb_sessionErrCodeStrings,
											buf[4],
											"Unknown error"));
			break;

		case NB_SS_TYPE_RETRSP:
			printf("    IP address = 0x%x\n", GETBE4(buf, 4));
			printf("    port = 0x%x\n", GETBE2(buf, 8));
			break;

		case NB_SS_TYPE_POSRSP:
		case NB_SS_TYPE_KEEPALIVE:
			break;
	}
	printf("------------- End of Message -------------\n\n");

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_displayDecodedSsMsg()"));
	return;
} /* nb_displayDecodedSsMsg() */

/*-----------------------------------------------------------------------------
 * Name			: nb_displayDecodedDsMsg
 *
 * Description	: Display decoded NetBIOS datagram service message
 *				  for debugging.
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 12/07/2001
 *-----------------------------------------------------------------------------
 */
void
nb_displayDecodedDsMsg(uchar8	*buf)
{
	uchar8		dgmType, nodeType;
	uint32		dgmLen, currOffset, nameLen;
	nb_name_t	name;
	uchar8		nameStr[NB_S_NAME + 10];

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_displayDecodedDsMsg()"));

	printf("\n----- NetBIOS Datagram Service Message -----\n");

	/* Print datagram type */
	dgmType = (uchar8)buf[0];
	printf("    Type = 0x%x (%s)\n", dgmType,
									 nb_getMsgString(nb_datagramTypeStrings,
													 dgmType,
													 "Unknown datagram type"));
	/* Print node type */
	nodeType = (buf[1] >> 2 ) & 0x03;
	printf("    Flags: Node type = %s\n", 
							nb_getMsgString(nb_datagramNodeTypeStrings,
											nodeType,
											"Unknows node type"));
	printf("           First packet = %s\n", NB_BOOLSTR(buf[1] & 0x02));
	printf("           More to follow = %s\n", NB_BOOLSTR(buf[1] & 0x01));
	

	/* Print datagram ID */
	printf("    Datagram ID = 0x%x\n", GETBE2(buf, 2));

	/* Print source IP address and port */
	printf("    Source IP address = 0x%x\n", GETBE4(buf, 4));
	printf("    Source port = %d\n", GETBE2(buf, 8));

	/* Print datagram length (including names) */
	dgmLen = GETBE2(buf, 10);
	printf("    Total datagram length (including names) = %d\n", dgmLen);

	/* Print packet offset */
	printf("    Packet offset = %d\n", GETBE2(buf, 12));

	currOffset = 14;

	/* Print source name */
	nameLen = nb_parseNameMsg(buf, currOffset, dgmLen, &name);
	if (nameLen == 0)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_displayDecodedDsMsg: "
					"Error in parsing source name"));
		return;
	}
	nb_getNameWithType(&name, nameStr);
	printf("    Source name = %s\n", nameStr);

	currOffset += nameLen;

	/* Print destination name */
	nameLen = nb_parseNameMsg(buf, currOffset, dgmLen, &name);
	if (nameLen == 0)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_displayDecodedDsMsg: "
					"Error in parsing destination name"));
		return;
	}
	nb_getNameWithType(&name, nameStr);
	printf("    destination name = %s\n", nameStr);

	/* Print user data */
	printf("    (%d bytes of user data)\n", dgmLen - (NB_S_COMPRESSED_NAME * 2));

	printf("--------------- End of Message ---------------\n\n");

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_displayDecodedDsMsg()"));
	return;
} /* nb_displayDecodedDsMsg() */

/*-----------------------------------------------------------------------------
 * Name			: nb_displayEncodedMsg
 *
 * Description	: Display encoded NetBIOS message for debugging.
 *				  It displays the content of the message buffer.
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 11/24/2001
 *-----------------------------------------------------------------------------
 */
void 
nb_displayEncodedMsg(uchar8 *buf, uint32 length)
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_displayEncodedMsg()"));

	memDump((void *)buf, length);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_displayEncodedMsg()"));
	return;
} /* nb_displayEncodedMsg() */


/*-----------------------------------------------------------------------------
 * Name			: nb_buildNameMsg
 *
 * Description	: Build a compressed NetBIOS name and put it in buffer.
 *				  TODO: The scope is not considered yet.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/15/2001
 *-----------------------------------------------------------------------------
 */
uint32			// Buffer length for NetBIOS name
nb_buildNameMsg(uchar8		*buf,	 // Buffer to store compressed NetBIOS name
				nb_name_t	*nbName) // Name
{
	uint32		i, len = 0;
	char		name[NB_S_NAME + 1];

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_buildNameMsg()"));

	if (strcmp(nbName->name, "*") == 0)
	{
		/* Special case for wildcard name */
		memset(name, 0, NB_S_NAME + 1);		// Fill the array with 0
		name[0] = '*';
	}
	else
	{
		strcpy(name, nbName->name);

		/* Convert the name string to upper case */
		strtoupper(name);

		/* Fill in the trailing part with " "(SPACE) */
		memset(name + strlen(name), ' ', NB_S_NAME + 1);

	}
	name[NB_S_NAME] = (uchar8) nbName->type;

	*buf = 32;				// NetBIOS name length.
	len = 34;				// NetBIOS name length plus 2

	for (i = 0; i < NB_S_NAME + 1; i++)
	{
		*(buf + 1 + 2*i) = 'A' + ((name[i] >> 4) & 0x0F);
		*(buf + 2 + 2*i) = 'A' + (name[i] & 0x0F);
	}	
	*(buf + len - 1) = 0;
	
	NB_PRINT(NB_DBGLV_DBG2, ("< nb_buildNameMsg()"));

	return (len);
} /* nb_buildNameMsg() */

/*-----------------------------------------------------------------------------
 * Name			: nb_buildRsrcMsg 
 *
 * Description	: Build a resource record and put it in buffer.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/15/2001
 *-----------------------------------------------------------------------------
 */
uint32			// Buffer length
nb_buildRsrcMsg(uchar8			*buf,	// Buffer to store resource record
				nb_nsResource_t	*rsrcP,	// Resource record
				boolean			namePointer)	// If TURE, put name pointer
												// instead of name in RR_NAME
{
	uint32		len = 0;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_buildRsrcMsg()"));

	if (namePointer == TRUE)
	{
		uint16	qNameOffset = 12;

		*buf = (0xC0 | (qNameOffset >> 8 & 0xFF));
		*(buf + 1) = (qNameOffset & 0xFF);
		len = 2;
	}
	else
	{
		len = nb_buildNameMsg(buf, &rsrcP->rrName);
	}
	
	PUTBE2(buf, len, rsrcP->rrType);
	PUTBE2(buf, len + 2, rsrcP->rrClass);
	PUTBE4(buf, len + 4, (rsrcP->ttl / 1000));	// Convert millisecond to second
	PUTBE2(buf, len + 8, rsrcP->rdLength);
	len += 10;

	memcpy(buf + len, rsrcP->rdata, rsrcP->rdLength);

	len += rsrcP->rdLength;

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_buildRsrcMsg()"));

	return (len);
} /* nb_buildRsrcMsg() */

/*-----------------------------------------------------------------------------
 * Name			: nb_buildNsMsg
 *
 * Description	: Build a NetBIOS Name Service message to transmit.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/15/2001
 *-----------------------------------------------------------------------------
 */
uint32			// Length of name service message
nb_buildNsMsg(uchar8		*buf,	 // Buffer for name service msg to be sent
			  nb_nsPacket_t	*nsPktP) // Name service packet structure
{
	nb_nsHeader_t	*header;
	uint32			offset = 0, len;
	uint16			tmp = 0;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_buildNsMsg()"));

	/*-----------
	 * Set header
	 *-----------
	 */
	header = &nsPktP->header;
	
	/* Set transaction ID */
	PUTBE2(buf, offset, header->nameTransID);
	offset += 2;

	/* Set response flag, opcode, nm_flags, and rcode */
	if (header->rsp == TRUE)
		tmp |= 0x8000;

	tmp |= ((header->opcode & 0x000F) << 11);

	if (header->nmFlags.auth == TRUE && header->rsp == TRUE)
		tmp |= (1 << 10);
	if (header->nmFlags.trunc == TRUE)
		tmp |= (1 << 9);
	if (header->nmFlags.recDesired == TRUE)
		tmp |= (1 << 8);
	if (header->nmFlags.recAval == TRUE && header->rsp == TRUE)
		tmp |= (1 << 7);
	if (header->nmFlags.bcast == TRUE)
		tmp |= (1 << 4);

	tmp |= (header->rcode & 0x000F);

	PUTBE2(buf, offset, tmp);
	offset += 2;

	/* Set qdcount, ancount, nscount, arcount */
	PUTBE2(buf, offset, header->qdCount);
	PUTBE2(buf, offset + 2, header->anCount);
	PUTBE2(buf, offset + 4, header->nsCount);
	PUTBE2(buf, offset + 6, header->arCount);
	offset += 8;

	/*---------------------
	 * Set question section
	 *---------------------
	 */
	if (header->qdCount)
	{
		/* Only support (qdCount <= 1) */
		len = nb_buildNameMsg(buf + offset, &nsPktP->question.qName);
		offset += len;

		PUTBE2(buf, offset, nsPktP->question.qType);
		PUTBE2(buf, offset + 2, nsPktP->question.qClass);
		offset += 4;
	}

	/*---------------------------
	 * Set answer resource record
	 *---------------------------
	 */
	if (header->anCount)
	{
		/* Only support (anCount <= 1) */
		len = nb_buildRsrcMsg(buf + offset, nsPktP->answer, FALSE);
		offset += len;
	}
	
	/*------------------------------
	 * Set authority resource record
	 *------------------------------
	 */
	if (header->nsCount)
	{
		/* Only support (nsCount <= 1) */
		len = nb_buildRsrcMsg(buf + offset, nsPktP->authority, FALSE);
		offset += len;
	}
	
	/*-------------------------------
	 * Set additional resource record
	 *-------------------------------
	 */
	if (header->arCount)
	{
		boolean	namePointer;

		/* Only support (arCount <= 1) */
		/* For name registration, release, and refresh request, 
		 * the RR_NAME field of additional resource record 
		 * contains compressed name pointer.
		 */
		if (header->rsp == FALSE && header->arCount == 1 &&
			(header->opcode == NB_NS_OPCODE_REG ||
			 header->opcode == NB_NS_OPCODE_RELEASE ||
			 header->opcode == NB_NS_OPCODE_REFRESH_8 ||
			 header->opcode == NB_NS_OPCODE_REFRESH_9))
		{
			namePointer = TRUE;
		}
		else
		{
			namePointer = FALSE;
		}
		len = nb_buildRsrcMsg(buf + offset, nsPktP->additional, namePointer);
		offset += len;
	}
	
	IFDBG(NB_DBGLV_DBG)
	{
		nb_displayDecodedNsMsg(nsPktP);
		nb_displayEncodedMsg(buf, offset);
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_buildNsMsg()"));

	return (offset);
} /* nb_buildNsMsg() */


/*-----------------------------------------------------------------------------
 * Name			: nb_parseNameMsg
 *
 * Description	: Parse a compressed NetBIOS name.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/15/2001
 *-----------------------------------------------------------------------------
 */
uint32			// Return number of bytes taken by the name. 0 if invalid.
nb_parseNameMsg(uchar8		*buf,		// Name service msg received
				int32		offset,		// Offset of compressed name
				uint32		length,		// Length of name service message 
				nb_name_t	*nbNameP)
{
	boolean	namePointer = FALSE;
	uint32	retOffset = 0;
	uint32	loopCnt = 0;
	uint32	nameLen1, nameLen2 = 0;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_parseNameMsg()"));

	if (length - offset < 2)
		return (0);

	/*---------------------------
	 * Parse name pointer, if any
	 *---------------------------
	 */
	while ((buf[offset] & 0x0C) == 0x0C)
	{
		if (namePointer == FALSE)
			retOffset += 2;
		namePointer = TRUE;
		offset = ((buf[offset] & ~0x0C) << 8) | buf[offset + 1];
		if (loopCnt++ == 10 || offset < 0 || offset > length - 2)
		{
			return (0);
		}
	}

	/*----------------------
	 * Parse compressed name
	 *----------------------
	 */
	nameLen1 = buf[offset];

	if (!nameLen1)
		return (0);

	if ((nameLen1 & 0xC0) || offset + nameLen1 + 2 > length)
	{
		return (0);
	}

	memset(nbNameP->name, 0, sizeof(nb_name_t));

	if (namePointer == FALSE)
		retOffset += nameLen1 + 2;

	offset++;

	/* Get name string and type */
	while(nameLen1 > 0)
	{
		uchar8	c1, c2;
		c1 = buf[offset++] - 'A';
		c2 = buf[offset++] - 'A';
		if ((c1 & 0xF0) || (c2 & 0xF0) || (nameLen2 > NB_S_NAME + 1))
			return (0);
		nbNameP->name[nameLen2++] = (c1 << 4) | c2;
		nameLen1 -= 2;
	}
	
	if (nameLen2 == NB_S_NAME + 1)
	{
		/* Get name type */
		nbNameP->type = nbNameP->name[NB_S_NAME] & 0xFF;

		/* Remove trailing spaces */
		nbNameP->name[NB_S_NAME] = 0;
		nameLen2 = NB_S_NAME - 1;
		while (nameLen2 && nbNameP->name[nameLen2] == ' ')
			nbNameP->name[nameLen2--] = 0;
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_parseNameMsg()"));

	return (retOffset);
} /* nb_parseNameMsg() */

/*-----------------------------------------------------------------------------
 * Name			: nb_parseRsrcMsg
 *
 * Description	: Parse a resource record.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/15/2001
 *-----------------------------------------------------------------------------
 */
int32			// offset
nb_parseRsrcMsg(uchar8			*buf,	// Name service msg received
				uint32			offset,	// Offset of resource record
				uint32			length,	// Length of name service message
				nb_nsResource_t	**rsrcP)
{
	uint32	nameLen;
	uint32	newOffset;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_parseRsrcMsg()"));

	/* Alloc memory to store resource record */
	NB_RAM_ALLOC_ZERO(*rsrcP, nb_nsResource_t, sizeof(nb_nsResource_t));
	if (*rsrcP == NULL)
		return (0);

	/* Parse compressed name corresponding to this resource record */
	nameLen = nb_parseNameMsg(buf, offset, length, &((*rsrcP)->rrName));
	newOffset = offset + nameLen;

	if (nameLen == 0 || newOffset + 10 > length)
	{
		eee_ramDealloc((uchar8 *)*rsrcP);
		return (0);
 	}

	(*rsrcP)->rrType = GETBE2(buf, newOffset);
	(*rsrcP)->rrClass = GETBE2(buf, newOffset + 2);
	(*rsrcP)->ttl = (GETBE4(buf, newOffset + 4)) * 1000; // Convert second to millisecond
	(*rsrcP)->rdLength = GETBE2(buf, newOffset + 8);
	newOffset += 10;

	if (((*rsrcP)->rdLength > NB_S_MAX_IP_DATAGRAM) ||
		 (*rsrcP)->rdLength + newOffset > length)
	{
		eee_ramDealloc(*rsrcP);
		return (0);
	}

	memcpy((*rsrcP)->rdata, buf + newOffset, (*rsrcP)->rdLength);

	newOffset += (*rsrcP)->rdLength;

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_parseRsrcMsg()"));

	return (newOffset);
} /* nb_parseRsrcMsg */

/*-----------------------------------------------------------------------------
 * Name			: nb_parseNsMsg
 *
 * Description	: Parse a NetBIOS Name Service message received.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/15/2001
 *-----------------------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
nb_parseNsMsg(uchar8		*buf,		// Name service msg received
			  uint32		length,		// Length of message to be parsed
			  nb_nsPacket_t	*nsPktP)	// Name service packet
{
	nb_nsHeader_t	*header;
	int32			offset = 0, nameLen;
	uint16			tmp = 0;
	
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_parseNsMsg()"));

	if (length < 12)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_parseNsMsg: Invalid name service message"));
		return (NFX_ERR);	// Name service header is incomplete
	}

	/*-------------
	 * Parse header
	 *-------------
	 */
	header = &nsPktP->header;

	/* Get transaction ID */
	header->nameTransID = GETBE2(buf, offset);
	offset += 2;

	/* Get response flag, opcode, nm_flags, and rcode */
	tmp = GETBE2(buf, offset);

	header->rsp = (tmp & 0x8000) ? TRUE : FALSE;
	header->opcode = (tmp >> 11) & 0x000F;
	header->nmFlags.auth = ((tmp >> 10) & 1) ? TRUE : FALSE;
	header->nmFlags.trunc = ((tmp >> 9) & 1) ? TRUE : FALSE;
	header->nmFlags.recDesired = ((tmp >> 8) & 1) ? TRUE : FALSE;
	header->nmFlags.recAval = ((tmp >> 7) & 1) ? TRUE : FALSE;
	header->nmFlags.bcast = ((tmp >> 4) & 1) ? TRUE : FALSE;
	header->rcode = tmp & 0x000F;
	offset += 2;

	/* Get qdcount, ancount, nscount, arcount */
	header->qdCount = GETBE2(buf, offset);
	header->anCount = GETBE2(buf, offset + 2);
	header->nsCount = GETBE2(buf, offset + 4);
	header->arCount = GETBE2(buf, offset + 6);
	offset += 8;

	/*---------------------
	 * Get question section
	 *---------------------
	 */
	if (header->qdCount)
	{
		nameLen = nb_parseNameMsg(buf, offset, length, &nsPktP->question.qName); 
		if (!nameLen)
		{
			NB_PRINT(NB_DBGLV_ERR, ("nb_parseNsMsg: Invalid NetBIOS name"));
			return (NFX_ERR);
		}

		if (length - (offset + nameLen) < 4)
		{
			NB_PRINT(NB_DBGLV_ERR, ("nb_parseNsMsg: Invalid name service message"));
			return (NFX_ERR);
		}
		offset += nameLen;
		nsPktP->question.qType = GETBE2(buf, offset);
		nsPktP->question.qClass = GETBE2(buf, offset + 2);
		offset += 2;  
	}

	/*---------------------------
	 * Get answer resource record
	 *---------------------------
	 */
	if (header->anCount)
	{
		offset = nb_parseRsrcMsg(buf, offset, length, &nsPktP->answer);
		if (!offset)
		{
			NB_PRINT(NB_DBGLV_ERR, ("nb_parseNsMsg: "
						"Invalid answer resource record"));
			return (NFX_ERR);
		}
	}

	/*------------------------------
	 * Get authority resource record
	 *------------------------------
	 */
	if (header->nsCount)
	{
		offset = nb_parseRsrcMsg(buf, offset, length, &nsPktP->authority);
		if (!offset)
		{
			NB_PRINT(NB_DBGLV_ERR, ("nb_parseNsMsg: "
						"Invalid authority resource record"));
			return (NFX_ERR);
		}
	}

	/*-------------------------------
	 * Get additional resource record
	 *-------------------------------
	 */
	if (header->arCount)
	{
		offset = nb_parseRsrcMsg(buf, offset, length, &nsPktP->additional);
		if (!offset)
		{
			NB_PRINT(NB_DBGLV_ERR, ("nb_parseNsMsg: "
						"Invalid additional resource record"));
			return (NFX_ERR);
		}
	}

	IFDBG(NB_DBGLV_DBG)
	{
		nb_displayEncodedMsg(buf, length);
		nb_displayDecodedNsMsg(nsPktP);
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_parseNsMsg()"));

	return (NFX_OK);
} /* nb_parseNsMsg() */

/*-----------------------------------------------------------------------------
 * Name			: nb_buildDsMsg
 *
 * Description	: Build a NetBIOS Datagram Service message to transmit.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/07/2001
 *-----------------------------------------------------------------------------
 */
uint32			// Length of datagram service message
nb_buildDsMsg(uchar8		*buf,	 // Buffer for datagram msg to be sent
			  nb_dsPacket_t	*dsPktP, // Datagram service packet structure
			  uint16		msgType) // Datagram message type
{
	nb_dsHeader_t	*headerP;
	uint32			offset, len;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_buildDsMsg()"));

	/*-----------
	 * Set header
	 *-----------
	 */
	headerP = &dsPktP->header;
	
	/* Set message type */
	buf[0] = msgType;

	/* Set flags */
	buf[1] = 0;
	if (headerP->flags.more == TRUE)
		buf[1] |= 0x01;
	if (headerP->flags.first == TRUE)
		buf[1] |= 0x02;
	buf[1] |= ((headerP->flags.nodeType & 0x03) << 2);

	offset = 2;

	/* Set datagram ID */
	PUTBE2(buf, offset, headerP->dgmID);
	offset += 2;

	/* Set srouce IP address and port */
	PUTBE4(buf, offset, headerP->srcIP);
	PUTBE2(buf, offset + 4, headerP->srcPort);
	offset += 6;
	
	/* Set datagram length */
	PUTBE2(buf, offset, headerP->dgmLen);
	offset += 2;

	/* Set offset */
	PUTBE2(buf, offset, headerP->offset);
	offset += 2;

	/* Set source name */
	len = nb_buildNameMsg(buf + offset, &dsPktP->srcName);
	if (len != NB_S_COMPRESSED_NAME)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_buildDsMsg: "
					"Incorrect length of source NetBIOS name"));
		panic("nb_buildDsMsg: Incorrect length of source NetBIOS name\n");
	}
	offset += len;

	/* Set destination name */
	len = nb_buildNameMsg(buf + offset, &dsPktP->destName);
	if (len != NB_S_COMPRESSED_NAME)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_buildDsMsg: "
					"Incorrect length of destination NetBIOS name"));
		panic("nb_buildDsMsg: Incorrect length of destination NetBIOS name\n");
	}
	offset += len;

	IFDBG(NB_DBGLV_DBG)
	{
		/*
		nb_displayDecodedDsMsg(dsPktP);
		*/
		nb_displayDecodedDsMsg(buf);
		nb_displayEncodedMsg(buf, offset + dsPktP->dataLen);
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_buildDsMsg()"));

	return (offset);
} /* nb_buildNsMsg() */


