/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * RCS:  $Id: nb-ns.c,v 1.1.1.1 2002/04/23 17:52:14 laurentb Exp $
 *
 * Name			: nb-ns.c
 *
 * Description	: Contains NetBIOS name service functions.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 11/07/2001
 *
 *-----------------------------------------------------------------
 */

#include "nb-api.h"
#include "nb-ns.h"
#include "nb-ds.h"
#include "nb-tpl.h"
#include "nb-util.h"
#include "nb-msg.h"


/*-----------------------------------------------------------------------------
 * Name			: nb_nsCreateReqPkt
 *
 * Description	: Create and initialize a name service request packet
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/14/2001
 *----------------------------------------------------------------------------- 
 */
nb_nsPacket_t *	// Ptr to name service packet or NULL if failed
nb_nsCreateReqPkt(nb_name_t	*name,		// Name related to the request
				  boolean	bcast)		// Broadcast the message or not
{
	nb_nsPacket_t	*nsPktP;
	nb_nsHeader_t	*nsHdrP;
	nb_nsQuestion_t	*nsQuestionP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsCreateReqPkt()"));

	/* Alloc memory for Name Service packet */
	NB_RAM_ALLOC_ZERO(nsPktP, nb_nsPacket_t, (sizeof(nb_nsPacket_t)));
	if (nsPktP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsCreateReqPkt: "
					"Failed to alloc memory for name service packet"));
		return (NULL);
	}

	/* Set header */
	nsHdrP = &nsPktP->header;
	nsHdrP->nameTransID = nb_genXid();
	nsHdrP->rsp = FALSE;

	nsHdrP->nmFlags.bcast = bcast;
	nsHdrP->nmFlags.recAval = FALSE;
	nsHdrP->nmFlags.trunc = FALSE;
	nsHdrP->nmFlags.auth = FALSE;

	nsHdrP->rcode = 0;
	nsHdrP->qdCount = 1;
	nsHdrP->anCount = 0;
	nsHdrP->nsCount = 0;

	/* Set question section */
	nsQuestionP = &nsPktP->question;
	memcpy(&nsQuestionP->qName, name, sizeof(nb_name_t));
	nsQuestionP->qType = NB_NS_QTYPE_NB;
	nsQuestionP->qClass = NB_NS_RRCLASS_IN;

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsCreateReqPkt()"));

	return (nsPktP);
} /* nb_nsCreateReqPkt() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsDeletePkt
 *
 * Description	: Delete a name service packet. Free the memory.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/14/2001
 *----------------------------------------------------------------------------- 
 */
void
nb_nsDeletePkt(nb_nsPacket_t *nsPktP)
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsDeletePkt()"));

	if (nsPktP == NULL)
		return;

	if (nsPktP->answer)
		eee_ramDealloc(nsPktP->answer);

	if (nsPktP->authority)
		eee_ramDealloc(nsPktP->authority);

	if (nsPktP->additional)
		eee_ramDealloc(nsPktP->additional);

	eee_ramDealloc(nsPktP);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsDeletePkt()"));

	return;
} /* nb_nsDeletePkt() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsSetNBFlags
 *
 * Description	: Set NB_FLAGS field of the Resource Record RDATA filed 
 * 				  for RR_TYPE of "NB".
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/14/2001
 *----------------------------------------------------------------------------- 
 */
void
nb_nsSetNBFlags (uchar8	*rdataP,		// Ptr to RDATA buffer
				 uchar8	gFlag,			// Group name flag
				 uchar8	nodeType)		// Node type
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsSetNBFlags()"));

	if (rdataP == NULL)
		return;
	*rdataP = gFlag | nodeType;
	*(rdataP+1) = '\0';

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsSetNBFlags()"));
	return;
} /* nb_nsSetNBFlags() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsGetNBFlags
 *
 * Description	: Get NB_FLAGS field of the Resource Record RDATA filed 
 * 				  for RR_TYPE of "NB".
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/14/2001
 *----------------------------------------------------------------------------- 
 */
uint8
nb_nsGetNBFlags (uchar8 *rdata)
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsGetNBFlags()"));
	return (*rdata & NB_NS_NBFLAG_MASK);
} /* nb_nsGetNBFlags() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsIsGroup
 *
 * Description	: Check if the group/unique characteristic of a NetBIOS
 *				  name service message is a group or not.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/14/2001
 *----------------------------------------------------------------------------- 
 */
boolean
nb_nsIsGroup(uchar8 *rdata)
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsIsGroup()"));
	return((nb_nsGetNBFlags(rdata) & NB_NS_NBFLAG_GROUP) ? TRUE : FALSE);
} /* nb_nsIsGroup */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsCreateAddRsrc
 *
 * Description	: Create and initialize additional resource record 
 * 				  in request packet.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/14/2001
 *----------------------------------------------------------------------------- 
 */
int32			// NFX_OK or NFX_ERR
nb_nsCreateAddRsrc(nb_nsPacket_t *nsPktP,	// Name service packet
				   uint32		 ipAddr,	// IP address of the name to
											// register, refresh, or release
											// In network byte order.
				   boolean		 group)
{
	nb_nsResource_t	*addRsrcP;
	uchar8			nodeType;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsCreateAddRsrc()"));

	if (nsPktP == NULL)
		return (NFX_ERR);

	/* Alloc memory for additional resource record */
	NB_RAM_ALLOC_ZERO(addRsrcP, nb_nsResource_t, sizeof(nb_nsResource_t));
	if (addRsrcP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsCreateAddRsrc: "
					"Failed to alloc memory for additional resource record"));
		return (NFX_ERR);
	}

	nsPktP->additional = addRsrcP;

	addRsrcP->rrName = nsPktP->question.qName;
	addRsrcP->rrType = NB_NS_RRTYPE_NB;
	addRsrcP->rrClass = NB_NS_RRCLASS_IN;

	if (nsPktP->header.nmFlags.bcast == TRUE)
	{
		addRsrcP->ttl = NB_NS_INFINITE_TTL;
		nodeType = NB_NS_NBFLAG_BNODE;
	}
	else
	{
		addRsrcP->ttl = NB_NS_TTL;
		nodeType = NB_NS_NBFLAG_HNODE;
	}

	addRsrcP->rdLength = 6;

	/* Set NB_FLAGS in RDATA field */
	nb_nsSetNBFlags(addRsrcP->rdata,
					group ? NB_NS_NBFLAG_GROUP : NB_NS_NBFLAG_UNIQUE,
					nodeType);

	/* Set IP address of the name */
	memcpy(&(addRsrcP->rdata[2]), &ipAddr, 4);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsCreateAddRsrc()"));

	return (NFX_OK);
} /* nb_nsCreateAddRsrc() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsClearRD
 *
 * Description	: Clear "Recursion Desired" field in name service packet header
 *				  The purpose is to convert Name Registration Request packet
 *				  to Name Overwirte Demand packet during the last step of name
 *				  registration process in a broadcast netowrk.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/14/2001
 *----------------------------------------------------------------------------- 
 */
void
nb_nsClearRD(nb_nsPacket_t *nsPktP, pkt_desc_t *nsXmtPktP)
{
	uchar8	*buf;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsClearRD()"));

	nsPktP->header.nmFlags.recDesired = FALSE;

	buf = nsXmtPktP->pHdr;
	buf[2] &= 0xFE;

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsClearRD()"));
} /* nb_nsClearRD() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsCreateNameRegPkt
 *
 * Description	: Create and initialize a name registration packet
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/14/2001
 *----------------------------------------------------------------------------- 
 */
nb_nsPacket_t *		// Prt to name service packet
nb_nsCreateNameRegPkt(nb_srvCB_t	*nbSrvCbP, 	// Prt to NetBIOS srv CB
					  nb_name_t		*reqName,
					  boolean		bcast,
					  boolean		group)
{
	int32			rc;
	nb_nsPacket_t	*nameRegPktP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsCreateNameRegPkt()"));

	nameRegPktP = nb_nsCreateReqPkt(reqName, bcast);
	if (nameRegPktP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsCreateNameRegPkt: "
					"Failed to create name service packet"));
		return (NULL);
	}

	nameRegPktP->header.opcode = NB_NS_OPCODE_REG;
	nameRegPktP->header.nmFlags.recDesired = TRUE;
	nameRegPktP->header.arCount = 1;

	rc = nb_nsCreateAddRsrc(nameRegPktP, nbSrvCbP->regCb.cifsSrvIpAddr, group);
	if (rc != NFX_OK)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsCreateNameRegPkt: "
					"Failed to create additional resource record"));
		nb_nsDeletePkt(nameRegPktP);
		return (NULL);
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsCreateNameRegPkt()"));

	return (nameRegPktP);
} /* nb_nsCreateNameRegPkt() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsCreateNameQueryPkt
 *
 * Description	: Create and initialize a name query packet
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/14/2001
 *----------------------------------------------------------------------------- 
 */
nb_nsPacket_t *		// Prt to name service packet
nb_nsCreateNameQueryPkt(nb_srvCB_t	*nbSrvCbP,	// Prt to NetBIOS srv CB
						nb_name_t	*reqName,
						boolean		bcast)
{
	int32			rc;
	nb_nsPacket_t	*nameQueryPktP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsCreateNameQueryPkt()"));

	nameQueryPktP = nb_nsCreateReqPkt(reqName, bcast);
	if (nameQueryPktP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsCreateNameQueryPkt: "
					"Failed to create name service packet"));
		return (NULL);
	}

	nameQueryPktP->header.opcode = NB_NS_OPCODE_QUERY;
	nameQueryPktP->header.nmFlags.recDesired = TRUE;
	nameQueryPktP->header.arCount = 0;

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsCreateNameQueryPkt()"));

	return (nameQueryPktP);
} /* nb_nsCreateNameQueryPkt() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsCreateNameReleasePkt
 *
 * Description	: Create and initialize a name release packet
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/22/2001
 *----------------------------------------------------------------------------- 
 */
nb_nsPacket_t *		// Prt to name service packet
nb_nsCreateNameReleasePkt(nb_srvCB_t	*nbSrvCbP, 	// Prt to NetBIOS srv CB
						  nb_name_t		*reqName,
						  boolean		bcast,
						  boolean		group)
{
	int32			rc;
	nb_nsPacket_t	*nameReleasePktP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsCreateNameReleasePkt()"));

	nameReleasePktP = nb_nsCreateReqPkt(reqName, bcast);
	if (nameReleasePktP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsCreateNameReleasePkt: "
					"Failed to create name service packet"));
		return (NULL);
	}

	nameReleasePktP->header.opcode = NB_NS_OPCODE_RELEASE;
	nameReleasePktP->header.nmFlags.recDesired = FALSE;
	nameReleasePktP->header.arCount = 1;

	rc = nb_nsCreateAddRsrc(nameReleasePktP,
							nbSrvCbP->regCb.cifsSrvIpAddr,
							group);
	if (rc != NFX_OK)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsCreateNameReleasePkt: "
					"Failed to create additional resource record"));
		nb_nsDeletePkt(nameReleasePktP);
		return (NULL);
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsCreateNameReleasePkt()"));

	return (nameReleasePktP);
} /* nb_nsCreateNameReleasePkt() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsCreateNameRefreshPkt
 *
 * Description	: Create and initialize a name refresh packet
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/17/2001
 *----------------------------------------------------------------------------- 
 */
nb_nsPacket_t *		// Prt to name service packet
nb_nsCreateNameRefreshPkt(nb_srvCB_t	*nbSrvCbP, 	// Prt to NetBIOS srv CB
						  nb_name_t		*reqName,
						  boolean		bcast,
						  boolean		group)
{
	int32			rc;
	nb_nsPacket_t	*nameRefreshPktP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsCreateNameRefreshPkt()"));

	nameRefreshPktP = nb_nsCreateReqPkt(reqName, bcast);
	if (nameRefreshPktP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsCreateNameRefreshPkt: "
					"Failed to create name service packet"));
		return (NULL);
	}

	nameRefreshPktP->header.opcode = NB_NS_OPCODE_REFRESH_8;
	nameRefreshPktP->header.nmFlags.recDesired = FALSE;
	nameRefreshPktP->header.arCount = 1;

	rc = nb_nsCreateAddRsrc(nameRefreshPktP,
							nbSrvCbP->regCb.cifsSrvIpAddr,
							group);
	if (rc != NFX_OK)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsCreateNameRefreshPkt: "
					"Failed to create additional resource record"));
		nb_nsDeletePkt(nameRefreshPktP);
		return (NULL);
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsCreateNameRefreshPkt()"));

	return (nameRefreshPktP);
} /* nb_nsCreateNameRefreshPkt() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsSendRequest
 *
 * Description	: Send a name service request.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/14/2001
 *----------------------------------------------------------------------------- 
 */
int32			// NFX_OK or NFX_ERR
nb_nsSendRequest(nb_reqCB_t	*nbReqCbP,	// Prt to NetBIOS srv CB
				 nb_name_t	*reqName,
				 uint32		raddr,
				 uint16		rport,
				 uint16		reqType,	// Request type
				 boolean	group)
{
	nb_srvCB_t		*nbSrvCbP = NULL;
	nb_nsPacket_t	*nsPktP = NULL;
	pkt_desc_t		*xmtPktDescP = NULL;
	boolean			bcast;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsSendRequest()"));

	if (nbReqCbP == NULL)
		return (NFX_ERR);

	nbSrvCbP = nbReqCbP->nbSrvCbP;

	/*-----------------------------------
	 * Create NetBIOS name service packet
	 *-----------------------------------
	 */
	bcast = (reqType & NB_REQ_TO_WINS) ? FALSE : TRUE;
	switch (reqType & ~NB_REQ_TO_WINS)
	{
		case NB_REQ_TYPE_NAME_REG:
			nsPktP = nb_nsCreateNameRegPkt(nbSrvCbP, reqName, bcast, group);
			if (nsPktP == NULL)
				return (NFX_ERR);
			break;

		case NB_REQ_TYPE_NAME_QUERY:
			nsPktP = nb_nsCreateNameQueryPkt(nbSrvCbP, reqName, bcast);
			if (nsPktP == NULL)
				return (NFX_ERR);
			break;

		case NB_REQ_TYPE_NAME_RELEASE:
			nsPktP = nb_nsCreateNameReleasePkt(nbSrvCbP, reqName, bcast, group);
			if (nsPktP == NULL)
				return (NFX_ERR);
			break;

		case NB_REQ_TYPE_NAME_REFRESH:
			nsPktP = nb_nsCreateNameRefreshPkt(nbSrvCbP, reqName, bcast, group);
			if (nsPktP == NULL)
				return (NFX_ERR);
			break;

		default:
			break;
	}

	/*----------------------------------
	 * Prepare the packet to be sent out
	 *----------------------------------
	 */
	xmtPktDescP = nb_createXmtPktSP((void *)nsPktP, NB_NS, NULL);
	if (xmtPktDescP == NULL)
	{
		nb_nsDeletePkt(nsPktP);
		return (NFX_ERR);
	}

	/*---------------------
	 * Send out the message 
	 *---------------------
	 */
	nb_xmtPktSP(nbSrvCbP->nameService.bsdSockP, xmtPktDescP, raddr, rport);

	/*----------------------
	 * Update the request CB
	 *----------------------
	 */
	nbReqCbP->reqID = nsPktP->header.nameTransID;
	nbReqCbP->reqPktP = (void *)nsPktP;
	nbReqCbP->reqXmtPktP = xmtPktDescP;
	nbReqCbP->raddr = raddr;
	nbReqCbP->rport = rport;

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsSendRequest()"));

	return (NFX_OK);
} /* nb_nsSendRequest() */


/*-----------------------------------------------------------------------------
 * Name			: nb_nsRequest
 *
 * Description	: Function for outgoing NetBIOS name service request
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/12/2001
 *----------------------------------------------------------------------------- 
 */
nb_reqCB_t *	// Ptr to request CB if succeed; NULL if fail
nb_nsRequest(nb_srvCB_t	*nbSrvCbP,	// Prt to NetBIOS srv CB
			 nb_name_t	*reqName,	// Name related to the request
			 uint16		reqType,	// Request type
			 boolean	group)
{
	int32		rc = NFX_OK;
	uint32		raddr;
	uint16		rport;
	nb_reqCB_t	*reqCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsRequest()"));

	/*----------------------------------
	 * Create a request CB in server CB
	 *----------------------------------
	 */
	reqCbP = nb_createReqCb(nbSrvCbP, reqType);
	if (reqCbP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsRequest: Failed to setup a request CB"));
		return (NULL);
	}

	/*-------------------------------
	 * Get remote IP address and port
	 *-------------------------------
	 */
	if (!(reqType & NB_REQ_TO_WINS))
	{
		/* Broadcast */
		raddr = nbSrvCbP->regCb.cifsSrvIpAddr | ~(nbSrvCbP->regCb.subnetMask);
	}
	else
	{
		/* Unicast */
		raddr = nbSrvCbP->regCb.winsSrvIpAddr;
	}

	rport = NB_NS_UDP_PORT;

	/*---------------------------------
	 * Send NetBIOS name service packet
	 *---------------------------------
	 */
	rc = nb_nsSendRequest(reqCbP, reqName, raddr, rport, reqType, group);
	if (rc != NFX_OK)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsRequest: "
					"Failed to send name service request"));
		goto nb_nsRequestErr;
	}

	/*---------------------------------------------------
	 * Processing specific to requests of different types
	 *---------------------------------------------------
	 */
	switch (reqType & ~NB_REQ_TO_WINS)
	{
		case NB_REQ_TYPE_NAME_REG:
			/* Change server name state */
			if (nb_isNameEqual(reqName, &nbSrvCbP->regCb.cifsSrvName) == TRUE)
			{
				nbSrvCbP->srvNameState = NB_NAME_STATE_REGISTERING;
			}
			else
				nbSrvCbP->domNameState = NB_NAME_STATE_REGISTERING;
			break;

		case NB_REQ_TYPE_NAME_RELEASE:
			/* Change server name state */
			nbSrvCbP->srvNameState = NB_NAME_STATE_DEREGISTERING;
			break;

		case NB_REQ_TYPE_NAME_QUERY:
		case NB_REQ_TYPE_NAME_REFRESH:
			break;

		default:
			break;
	}

	/*----------------------------------
	 * Create a retry timer and start it
	 *----------------------------------
	 */
#ifndef FS_STANDALONE
	reqCbP->reqTimerHdl = eee_timerCreate(reqCbP->retryInterval,
										  EEE_TIMER_REPEAT);
	if (reqCbP->reqTimerHdl == TIMER_ERR_NO_TIMER_AVAILABLE)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsRequest: "
					"Failed to create name registration request timer"));
		goto nb_nsRequestErr;
	}

	rc = eee_timerStart(reqCbP->reqTimerHdl,	// Timer handle
						nb_nsReqTimeout,		// Timeout handler function
						(void *)reqCbP,			// Ptr to Request CB
					   	0, 0, 0);
	if (rc != NFX_OK)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsRequest: "
					"Failed to start name registration request timer"));
		goto nb_nsRequestErr;
	}
#endif //FS_STANDALONE

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsRequest()"));
	return (reqCbP);

nb_nsRequestErr:
	nb_deleteReqCb(reqCbP);
	return (NULL);
} /* nb_nsRequest() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsLookupNameInCache
 *
 * Description	: Search for a name in name query CB list.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/20/2001
 *----------------------------------------------------------------------------- 
 */
int32			// NFX_OK or NFX_ERR
nb_nsLookupNameInCache(nb_srvCB_t	*nbSrvCbP,		// Ptr to server CB
					   nb_name_t	*nbName,		// Name to search
					   nb_queryCB_t **nbQueryCbP)	// Ptr to query CB on return
{
	int32			rc;
	nb_queryCB_t	*queryCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsLookupNameInCache()"));

	rc = NFX_ERR;
	*nbQueryCbP = NULL;

	queryCbP = nbSrvCbP->queryCbP;
	while(queryCbP)
	{
		if (nb_isNameEqual(nbName, &queryCbP->name) == FALSE)
		{
			queryCbP = queryCbP->nextP;
		}
		else
		{
			/* Found the name */
			*nbQueryCbP = queryCbP;
			rc = NFX_OK;
			break;
		}
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsLookupNameInCache()"));

	return (rc);
} /* nb_nsLookupNameInCache() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsProcWackRsp
 *
 * Description	: Process incoming NetBIOS WACK response messages.
 *				  Stop and destroy current retry timer, create and start
 *				  a new one with timeout ttl given in WACK response message.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/17/2001
 *----------------------------------------------------------------------------- 
 */
int32			// NFX_OK or NFX_ERR
nb_nsProcWackRsp(nb_nsPacket_t	*nsInRspPktP,
				 nb_reqCB_t		*nsReqCbP)
{
	int32	rc;
	uint64	ttl;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsProcWackRsp()"));

	nb_timerStop(&nsReqCbP->reqTimerHdl);
	nb_timerDestroy(&nsReqCbP->reqTimerHdl);

	nsReqCbP->retryCnt = 0;

	ttl = nsInRspPktP->answer->ttl;
	if (ttl == 0)
	{
		/* WINS doesn't know how long it will take to process the request */

		/* TODO- Then how long should we wait? 
		 * The current implementation doesn't start another timer. 
		 */
		return (NFX_OK);
	}

#ifndef FS_STANDALONE
	nsReqCbP->reqTimerHdl = eee_timerCreate(ttl, EEE_TIMER_ONCE);
	if (nsReqCbP->reqTimerHdl == TIMER_ERR_NO_TIMER_AVAILABLE)
	{ 
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcWackRsp: "
					"Failed to create WACK wait timer"));
		return (NFX_ERR);
	}

	rc = eee_timerStart(nsReqCbP->reqTimerHdl,	// Timer handle
						nb_nsReqTimeout,		// Timeout handler function
						(void *)nsReqCbP,		// Ptr to Request CB
						0, 0, 0);
	if (rc != NFX_OK)
	{ 
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcWackRsp: "
					"Failed to create WACK wait timer"));
		return (NFX_ERR);
	}
#endif //FS_STANDALONE

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsProcWackRsp()"));
	return (NFX_OK);
} /* nb_nsProcWackRsp() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsProcNameRegRsp
 *
 * Description	: Process incoming NetBIOS name registration response messages.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/22/2001
 *----------------------------------------------------------------------------- 
 */
int32			// NFX_OK or NFX_ERR
nb_nsProcNameRegRsp(nb_nsPacket_t	*nsInRspPktP,
				    nb_reqCB_t		*nsReqCbP)
{
	int32			rc = NFX_OK;
	nb_nsPacket_t	*nsOutReqPktP;
	nb_name_t		*questionName, *answerName;
	nb_reqCB_t		*refreshReqCbP;
	nb_srvCB_t		*nbSrvCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsProcNameRegRsp()"));

	if (nsInRspPktP == NULL || nsReqCbP == NULL)
		return (NFX_ERR);

	nsOutReqPktP = (nb_nsPacket_t *)nsReqCbP->reqPktP;
	nbSrvCbP = nsReqCbP->nbSrvCbP;

	/*------------
	 * Verify name
	 *------------
	 */
	questionName = &nsOutReqPktP->question.qName;
	answerName = &nsInRspPktP->answer->rrName;

	if (answerName->name == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcNameRegRsp: Answer name is NULL"));
		return (NFX_ERR);
	}

	if (nb_isNameEqual(questionName, answerName) == FALSE)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcNameRegRsp: "
					"Question name (%s) and answer name (%s) doesn't match",
					questionName->name, answerName->name));
		return (NFX_ERR);
	}

	if (nsInRspPktP->header.rcode != 0)
	{
		/* Name registration failed */
		nb_nsProcNameRegResult(nsReqCbP, nsInRspPktP, NFX_ERR);
	}
	else
	{
		/* Positive or WACK response */

		if (nsOutReqPktP->header.nmFlags.bcast == TRUE)
		{
			/*----------
			 * Broadcast
			 *----------
			 */
			/* B node shouldn't get this. Ignore */
		}
		else if (nsInRspPktP->header.opcode == NB_NS_OPCODE_WACK)
		{
			/*-----------------------------
			 * WINS server wants us to wait
			 *-----------------------------
			 */

			/* Stop and destroy current retry timer,
			 * create and start a new one with timeout ttl */
			rc = nb_nsProcWackRsp(nsInRspPktP, nsReqCbP);
		}
		else if (nsInRspPktP->header.nmFlags.recAval == 0)
		{
			/* Received END-NODE CHALLENGE NAME REGISTRATION RESPONSE */
			/* TODO -
			 * When a client receives an END-NODE CHALLENGE, its duty is 
			 * to query the owner (the owner's IP address will be in the 
			 * ANSWER_RECORD.RDATA.NB_ADDRESS field) to see if the owner 
			 * still wants the name. If the owner does not respond, or if 
			 * it replies with a NEGATIVE NAME QUERY RESPONSE, then the name 
			 * is available and the requester may send a NAME UPDATE REQUEST 
			 * to the NBNS. The NBNS will blindly trust the requester, change 
			 * the entry, and reply with a POSITIVE NAME REGISTRATION RESPONSE. 
			 * The NAME UPDATE REQUEST is the same as the unicast 
			 * NAME REGISTRATION REQUEST except that the RD bit is clear 
			 * (Recursion Desired is 0).
			 *
			 * There is nothing to stop a client from skipping the name query 
			 * and sending the update message to the NBNS, effectively 
			 * stealing the name. This is why the RFCs use the term 
			 * non-secured when describing this mechanism.
			 */
		}
		else
		{
			/*-----------------------------
			 * Unicast registration succeed 
			 *-----------------------------
			 */
			if (nbSrvCbP->currNodeType == NB_NS_NBFLAG_BNODE)
			{
				/* Make sure our node type is H-node.
				 * The reason the node type is re-set here is that 
				 * the previous node type may be switched to B-node
				 * when the WINS server was down. */
				nbSrvCbP->currNodeType = NB_NS_NBFLAG_HNODE;
			}

			/* Send name refresh request after ttl timeout */
			nbSrvCbP->srvNameTTL = nsInRspPktP->answer->ttl;
			if (nsInRspPktP->answer->ttl != 0)
			{
				refreshReqCbP = nb_nsRequest(
								nbSrvCbP,
								&(nbSrvCbP->regCb.cifsSrvName),
								(NB_REQ_TYPE_NAME_REFRESH | NB_REQ_TO_WINS),
								nb_nsIsGroup(nsOutReqPktP->additional->rdata));
			}
			nb_nsProcNameRegResult(nsReqCbP, NULL, NFX_OK);
		}
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsProcNameRegRsp()"));

	return (rc);
} /* nb_nsProcNameRegRsp() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsProcNameQueryRsp
 *
 * Description	: Process incoming NetBIOS name query response messages.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/22/2001
 *----------------------------------------------------------------------------- 
 */
int32			// NFX_OK or NFX_ERR
nb_nsProcNameQueryRsp(nb_incomingMsgCB_t	*inMsgCbP,
					  nb_nsPacket_t			*nsInRspPktP,
					  nb_reqCB_t			*nsReqCbP)
{
	nb_nsPacket_t	*nsOutReqPktP;
	nb_name_t		*questionName, *answerName;
	nb_queryCB_t	*queryCbP;
	int32			rc = NFX_OK;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsProcNameQueryRsp()"));

	if (nsInRspPktP == NULL || nsReqCbP == NULL)
		return (NFX_ERR);

	nsOutReqPktP = (nb_nsPacket_t *)nsReqCbP->reqPktP;

	/*------------
	 * Verify name
	 *------------
	 */
	questionName = &nsOutReqPktP->question.qName;
	answerName = &nsInRspPktP->answer->rrName;

	if (answerName->name == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcNameQueryRsp: Answer name is NULL"));
		return (NFX_ERR);
	}

	if (nb_isNameEqual(questionName, answerName) == FALSE)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcNameQueryRsp: "
					"Question name (%s) and answer name (%s) doesn't match",
					questionName->name, answerName->name));
		return (NFX_ERR);
	}

	queryCbP = nsReqCbP->queryCbP;
	if (queryCbP == NULL)
		return (NFX_ERR);

	if (nsOutReqPktP->header.nmFlags.bcast == TRUE)
	{
		/*----------
		 * Broadcast
		 *----------
		 */
		if (nsInRspPktP->header.rcode == 0)
		{
			/* Positive response */

			queryCbP->rspMsgCnt++;

			if (queryCbP->rspMsgCnt <= 1)
			{
				/*-------------------------------------
				 * This is the first response recieved.
				 *-------------------------------------
				 */

				/* Set address and port of node that sent authoritative rsp */
				queryCbP->raddr = inMsgCbP->raddr;
				queryCbP->rport = inMsgCbP->rport;

				/* Set group flag from response packet */
				queryCbP->group = nb_nsIsGroup(nsInRspPktP->answer->rdata);

				/* Process positive response result */
				nb_nsProcNameQueryResult(nsReqCbP, nsInRspPktP, NFX_OK);

#ifndef FS_STANDALONE
				/* Create and start conlict timer */
				queryCbP->conflTimerHdl = eee_timerCreate(NB_NS_CONFLICT_TIMER,
														  EEE_TIMER_ONCE);
				if (queryCbP->conflTimerHdl == TIMER_ERR_NO_TIMER_AVAILABLE)
				{
					NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcNameQueryRsp: "
								"Failed to create conflict timer"));
				}
				rc = eee_timerStart(queryCbP->conflTimerHdl,
									nb_nsConflTimeout,
									(void *)queryCbP,
								   	0, 0, 0);
				if (rc != NFX_OK)
				{
					NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcNameQueryRsp: "
								"Failed to start conflict timer"));
				}
#endif //FS_STANDALONE
			}
			else
			{
				/*------------------------------ 
				 * Respons(es) already received.
				 * Check for conflict.
				 *------------------------------
				 */
				boolean		duplicated = FALSE, group = TRUE;
				uint32		*state;

				/* Verify whether it is from the same node that
				 * sent authoritative response.
				 */
				if (inMsgCbP->raddr == queryCbP->raddr &&
					inMsgCbP->rport == queryCbP->rport)
					duplicated = TRUE;

				/* Check if group/unique characteristic of the authoritative
				 * response or this response if "unique"
				 */
				if (queryCbP->group == FALSE ||
					nb_nsIsGroup(nsInRspPktP->answer->rdata) == FALSE)
					group = FALSE;

				if (duplicated == FALSE && group == FALSE)
				{
					/* There is name conflict. */
					state = nb_getSrvNameState(inMsgCbP->nbSrvCbP, questionName);
					*state = NB_NAME_STATE_CONFLICT;

					/* TODO - Inform CIFS module of the name conflict */
				}
				else
				{
					/* No conflict. Duplicated response. */
					queryCbP->rspMsgCnt--;
				}
			}
		}
		else
		{
			/* Negative response. B node shouldn't get this. Ignore. */
		}
	}
	else
	{
		/*--------
		 * Unicast
		 *--------
		 */

		queryCbP->rspMsgCnt++;

		if (queryCbP->rspMsgCnt > 1)
		{
			/* Already called the indication function */
			return (NFX_OK);
		}
		else if (nsInRspPktP->header.opcode == NB_NS_OPCODE_WACK)
		{
			/*-----------------------------
			 * WINS server wants us to wait
			 *-----------------------------
			 */

			/* Stop and destroy current retry timer,
			 * create and start a new one with timeout ttl */
			rc = nb_nsProcWackRsp(nsInRspPktP, nsReqCbP);
			queryCbP->rspMsgCnt--;
		}
		else if (nsInRspPktP->header.rcode == 0)
		{
			if (nsInRspPktP->header.nmFlags.auth == FALSE &&
				nsInRspPktP->header.nmFlags.recAval == FALSE)
			{
				/* Received a REDIRECT NAME QUERY RESPONSE */
				/* TODO - process it */
			}
			else
			{
				/* Process positive response result */
				nb_nsProcNameQueryResult(nsReqCbP, nsInRspPktP, NFX_OK);
			}
		}
		else
		{
			/* Process negative response result */
			nb_nsProcNameQueryResult(nsReqCbP, nsInRspPktP, NFX_ERR);
		}
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsProcNameQueryRsp()"));

	return (rc);
} /* nb_nsProcNameQueryRsp() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsProcNameReleaseRsp
 *
 * Description	: Process incoming NetBIOS name release response messages.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/22/2001
 *----------------------------------------------------------------------------- 
 */
int32			// NFX_OK or NFX_ERR
nb_nsProcNameReleaseRsp(nb_nsPacket_t	*nsInRspPktP,
						nb_reqCB_t		*nsReqCbP)
{
	int32			rc;
	nb_nsPacket_t	*nsOutReqPktP;
	nb_name_t		*questionName, *answerName;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsProcNameReleaseRsp()"));

	if (nsInRspPktP == NULL || nsReqCbP == NULL)
		return (NFX_ERR);

	nsOutReqPktP = (nb_nsPacket_t *)nsReqCbP->reqPktP;

	/*------------
	 * Verify name
	 *------------
	 */
	questionName = &nsOutReqPktP->question.qName;
	answerName = &nsInRspPktP->answer->rrName;

	if (answerName->name == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcNameReleaseRsp: Answer name is NULL"));
		return (NFX_ERR);
	}

	if (nb_isNameEqual(questionName, answerName) == FALSE)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcNameReleaseRsp: "
					"Question name (%s) and answer name (%s) doesn't match",
					questionName->name, answerName->name));
		return (NFX_ERR);
	}

	if (nsOutReqPktP->header.nmFlags.bcast == TRUE)
	{
		/*----------
		 * Broadcast
		 *----------
		 */
		/* B node shouldn't get name release response. Ignore */
	}
	else
	{
		/*--------
		 * Unicast
		 *--------
		 */
		if (nsInRspPktP->header.rcode != 0)
		{
			/* Name release failed. We are not allowed to release the name. */
			nb_nsProcNameReleaseResult(nsReqCbP, NFX_ERR);
		}
		else if (nsInRspPktP->header.opcode == NB_NS_OPCODE_WACK)
		{
			/* WINS server wants us to wait */

			/* Stop and destroy current retry timer,
			 * create and start a new one with timeout ttl */
			rc = nb_nsProcWackRsp(nsInRspPktP, nsReqCbP);
		}
		else
		{
			/* Name release succeed */
			nb_nsProcNameReleaseResult(nsReqCbP, NFX_OK);
		}
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsProcNameReleaseRsp()"));

	return (rc);
} /* nb_nsProcNameReleaseRsp() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsProcResponse
 *
 * Description	: Process incoming NetBIOS response messages.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/22/2001
 *----------------------------------------------------------------------------- 
 */
void
nb_nsProcResponse(nb_incomingMsgCB_t	*inMsgCbP,
				  nb_nsPacket_t			*nsInRspPktP)
{
	int32		rc = NFX_OK;
	uint16		xid;		// Transaction ID of response packet
	nb_reqCB_t	*nsReqCbP = NULL;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsProcResponse()"));


	/*---------------------------------------------------------
	 * Look for the corresponding request CB in request CB list
	 *---------------------------------------------------------
	 */
	xid = nsInRspPktP->header.nameTransID;
	nsReqCbP = nb_getReqCb(inMsgCbP->nbSrvCbP, xid);
	if (nsReqCbP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcResponse: "
					"No matching transaction ID"));
		return;
	}
	if ((nsInRspPktP->header.opcode != NB_NS_OPCODE_WACK) &&
		(((nb_nsPacket_t *)nsReqCbP->reqPktP)->header.opcode != nsInRspPktP->header.opcode)) 
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcResponse: "
					"Request and response types don't match"));
		return;
	}

	/*-------------------------
	 * Validate received packet
	 *-------------------------
	 */
	switch (nsInRspPktP->header.opcode)
	{
		case NB_NS_OPCODE_REG:
		case NB_NS_OPCODE_REFRESH_8:
		case NB_NS_OPCODE_REFRESH_9:
			if (nsInRspPktP->header.anCount == 0)
			{
				NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcResponse: "
							"Invalid REGISTRATION/REFRESH packet"));
				rc = NFX_ERR;
			}
			break;

		case NB_NS_OPCODE_QUERY:
			if (nsInRspPktP->header.anCount != 0 &&
				nsInRspPktP->header.anCount != 1 )
			{
				NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcResponse: "
							"Invalid QUERY packet"));
				rc = NFX_ERR;
			}
			break;

		case NB_NS_OPCODE_RELEASE:
			if (nsInRspPktP->header.anCount == 0)
			{
				NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcResponse: "
							"Invalid RELEASE packet"));
				rc = NFX_ERR;
			}
			break;

		case NB_NS_OPCODE_WACK:
			if (nsInRspPktP->header.anCount != 1)
			{
				NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcResponse: "
							"Invalid WAK packet"));
				rc = NFX_ERR;
			}
			break;

		default:
			NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcResponse: "
						"Opcode %d is not supported",
						nsInRspPktP->header.opcode));
			rc = NFX_ERR;
			break;
	}
	if (rc != NFX_OK)
		return;

	/* If incoming packet is the response to a unicast request,
	 * remove the corresponding broadcast request if any.
	 * This is becasue the WINS server is up again and we
	 * don't need to broadcast any more.
	 */
	if ((nsReqCbP->reqType & NB_REQ_TO_WINS) && nsReqCbP->bcastReqCbP)
	{
		nb_deleteReqCb(nsReqCbP->bcastReqCbP);
	}

	/*---------------------
	 * Process the response
	 *---------------------
	 */
	switch (nsReqCbP->reqType & ~NB_REQ_TO_WINS)
	{
		case NB_REQ_TYPE_NAME_REG:
		case NB_REQ_TYPE_NAME_REFRESH:
			rc = nb_nsProcNameRegRsp(nsInRspPktP, nsReqCbP);
			break;

		case NB_REQ_TYPE_NAME_QUERY:
			rc = nb_nsProcNameQueryRsp(inMsgCbP, nsInRspPktP, nsReqCbP);
			break;

		case NB_REQ_TYPE_NAME_RELEASE:
			rc = nb_nsProcNameReleaseRsp(nsInRspPktP, nsReqCbP);
			break;

		default:
			break;
	}

	/*-----------------------------
	 * Delete request control block
	 *-----------------------------
	 */
	if (nsInRspPktP->header.opcode != NB_NS_OPCODE_WACK)
		nb_deleteReqCb(nsReqCbP);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsProcResponse()"));

} /* nb_nsProcResponse() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsCreateRspPkt
 *
 * Description	: Create a name service response packet
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/22/2001
 *----------------------------------------------------------------------------- 
 */
void
nb_nsCreateRspPkt(nb_nsRspParam_t	*rspParamP,
				  nb_nsPacket_t		*nsOutRspPktP)
{	
	nb_nsPacket_t		*nsInReqPktP	= rspParamP->nsInReqPktP;
	uint16				rcode			= rspParamP->rcode;
	uint16				reqType			= rspParamP->reqType;
	uint16				opcode			= rspParamP->opcode;
	uint32				ttl				= rspParamP->ttl;
	uint8				*rdata			= rspParamP->rdata;
	uint16				dataLen			= rspParamP->dataLen;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsCreateRspPkt()"));

	switch(reqType)
	{
		case NB_REQ_TYPE_NAME_REG:
		case NB_REQ_TYPE_NAME_REFRESH:
		case NB_REQ_TYPE_NAME_QUERY:
			nsOutRspPktP->header.nmFlags.recDesired = TRUE;
			nsOutRspPktP->header.nmFlags.recAval = TRUE;
			break;

		case NB_REQ_TYPE_NAME_RELEASE:
			nsOutRspPktP->header.nmFlags.recDesired = FALSE;
			nsOutRspPktP->header.nmFlags.recAval = FALSE;
			break;

		default:
			/* TODO: Other request types are not supported at present */
			break;
	}

	/*-----------
	 * Set header
	 *-----------
	 */
	nsOutRspPktP->header.nameTransID = nsInReqPktP->header.nameTransID;
	nsOutRspPktP->header.opcode = nsInReqPktP->header.opcode;
	nsOutRspPktP->header.rsp = TRUE;
	nsOutRspPktP->header.nmFlags.bcast = FALSE;
	nsOutRspPktP->header.nmFlags.trunc = FALSE;
	nsOutRspPktP->header.nmFlags.auth = TRUE;
	nsOutRspPktP->header.rcode = rcode;
	nsOutRspPktP->header.qdCount = 0;
	nsOutRspPktP->header.anCount = 1;
	nsOutRspPktP->header.nsCount = 0;
	nsOutRspPktP->header.arCount = 0;

	/*----------------------------
	 * Question section is invalid
	 *----------------------------
	 */
	memset(&nsOutRspPktP->question, 0, sizeof(nb_nsQuestion_t));

	/*---------------------------
	 * Set answer resource record
	 *---------------------------
	 */
	memset(nsOutRspPktP->answer, 0, sizeof(nb_nsResource_t));
	nsOutRspPktP->answer->rrName = nsInReqPktP->question.qName;
	nsOutRspPktP->answer->rrType = nsInReqPktP->question.qType;
	nsOutRspPktP->answer->rrClass = nsInReqPktP->question.qClass;
	nsOutRspPktP->answer->ttl = ttl;

	if (rdata && dataLen)
	{
		nsOutRspPktP->answer->rdLength = dataLen;
		memcpy(nsOutRspPktP->answer->rdata, rdata, dataLen);
	}

	/*------------------------------------------------------
	 * Authority and additional resource records are invalid
	 *------------------------------------------------------
	 */
	nsOutRspPktP->authority = NULL;
	nsOutRspPktP->additional = NULL;

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsCreateRspPkt()"));
	return;
} /* nb_nsCreateRspPkt() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsResponse
 *
 * Description	: Function for outgoing NetBIOS name service response
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/22/2001
 *----------------------------------------------------------------------------- 
 */
int32			// NFX_OK or NFX_ERR
nb_nsResponse(nb_incomingMsgCB_t	*inMsgCbP,
			  nb_nsRspParam_t		*rspParamP)
{
	nb_nsPacket_t		nsOutRspPkt;		// Outgoing response packet
	nb_nsResource_t		answer;
	pkt_desc_t			*xmtPktDescP = NULL;
	uint32				raddr;
	uint16				rport;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsResponse()"));

	/*-----------------------
	 * Create response packet
	 *-----------------------
	 */
	nsOutRspPkt.answer = &answer;
	nb_nsCreateRspPkt(rspParamP, &nsOutRspPkt);

	/*----------------------------------
	 * Prepare the packet to be sent out
	 *----------------------------------
	 */
	xmtPktDescP = nb_createXmtPktSP((void *)&nsOutRspPkt, NB_NS, NULL);
	if (xmtPktDescP == NULL)
		return (NFX_ERR);

	/*---------------------
	 * Send out the message 
	 *---------------------
	 */
	raddr = inMsgCbP->raddr;
	rport = inMsgCbP->rport;

	nb_xmtPktSP(inMsgCbP->nbSrvCbP->nameService.bsdSockP,
				xmtPktDescP,
				raddr,
				rport);

	PKT_FREE_PACKET(xmtPktDescP);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsResponse()"));
	return (NFX_OK);
} /* nb_nsResponse() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsProcNameRegReq
 *
 * Description	: Process incoming NetBIOS name registration request messages.
 * 				  Send negative response for unicast request.
 * 				  Send negative response if the CIFS server owns the name.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/22/2001
 *----------------------------------------------------------------------------- 
 */
void
nb_nsProcNameRegReq(nb_incomingMsgCB_t	*inMsgCbP,
					nb_nsPacket_t		*nsInReqPktP)
{
	int32				rc;
	uint8				rdata[6];
	boolean				bcast;
	nb_name_t			*reqName;
	nb_nsRspParam_t		rspParam;

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsProcNameRegReq()"));

	/* Partially initialize parameters for nb_nsResponse() */
	rspParam.nsInReqPktP = nsInReqPktP;
	rspParam.reqType = NB_REQ_TYPE_NAME_REG;
	rspParam.opcode = NB_NS_OPCODE_REG;
	rspParam.ttl = 0;
	rspParam.dataLen = 6;

	/* Set NB_FLAGS in rdata*/
	nb_nsSetNBFlags(rdata,
					NB_NS_NBFLAG_UNIQUE,
					inMsgCbP->nbSrvCbP->currNodeType);
	/* Set IP address of the name in rdata */
	memcpy(&rdata[2], &inMsgCbP->nbSrvCbP->regCb.cifsSrvIpAddr, 4);
	rspParam.rdata = rdata;

	bcast = nsInReqPktP->header.nmFlags.bcast;
	if (bcast == FALSE)
	{
		/* The CIFS server shouldn't get unicast name registration packet.
		 * Send negative response.
		 */
		rspParam.rcode = NB_NS_FMT_ERR;
		nb_nsResponse(inMsgCbP, &rspParam);
		return;
	}

	reqName = &nsInReqPktP->question.qName;
	
	/* Lookup the server name from server list */
	rc = nb_lookupSrvName(reqName);
	if (rc == NFX_OK)
	{
		/* One of CIFS servers has registered the name */
		/* Unique name.
		 * Send negative response.
		 */
		rspParam.rcode = NB_NS_ACT_ERR;
		nb_nsResponse(inMsgCbP, &rspParam);
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsProcNameRegReq()"));
	return;
} /* nb_nsProcNameRegReq() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsProcNameRefreshReq
 *
 * Description	: Process incoming NetBIOS name refresh request messages.
 * 				  Send negative name registration response for unicast request.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/28/2001
 *----------------------------------------------------------------------------- 
 */
void
nb_nsProcNameRefreshReq(nb_incomingMsgCB_t	*inMsgCbP,
						nb_nsPacket_t		*nsInReqPktP)
{
	int32				rc;
	uint8				rdata[6];
	boolean				bcast;
	nb_name_t			*reqName;
	nb_nsRspParam_t		rspParam;

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsProcNameRefreshReq()"));

	/* Partially initialize parameters for nb_nsResponse() */
	rspParam.nsInReqPktP = nsInReqPktP;
	rspParam.reqType = NB_REQ_TYPE_NAME_REG;
	rspParam.opcode = NB_NS_OPCODE_REG;
	rspParam.ttl = 0;
	rspParam.dataLen = 6;

	/* Set NB_FLAGS in rdata*/
	nb_nsSetNBFlags(rdata,
					NB_NS_NBFLAG_UNIQUE,
					inMsgCbP->nbSrvCbP->currNodeType);
	/* Set IP address of the name in rdata */
	memcpy(&rdata[2], &inMsgCbP->nbSrvCbP->regCb.cifsSrvIpAddr, 4);
	rspParam.rdata = rdata;

	bcast = nsInReqPktP->header.nmFlags.bcast;
	if (bcast == FALSE)
	{
		/* The CIFS server shouldn't get unicast name refresh packet.
		 * Send negative response.
		 */
		rspParam.rcode = NB_NS_FMT_ERR;
		nb_nsResponse(inMsgCbP, &rspParam);
		return;
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsProcNameRefreshReq()"));
	return;
} /* nb_nsProcNameRefreshReq() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsProcNameQueryReq
 *
 * Description	: Process incoming NetBIOS name query request messages.
 * 				  Send positive response if the CIFS server owns the name.
 * 				  Never send negative response to broadcast request.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/22/2001
 *----------------------------------------------------------------------------- 
 */
void
nb_nsProcNameQueryReq(nb_incomingMsgCB_t	*inMsgCbP,
					  nb_nsPacket_t			*nsInReqPktP)
{
	uint8				rdata[6];
	nb_name_t			*reqName, *myName;
	nb_nsRspParam_t		rspParam;
	uint32				myIpAddr;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsProcNameQueryReq()"));

	/* Partially initialize parameters for nb_nsResponse() */
	rspParam.nsInReqPktP = nsInReqPktP;
	rspParam.reqType = NB_REQ_TYPE_NAME_QUERY;
	rspParam.opcode = NB_NS_OPCODE_QUERY;

	reqName = &nsInReqPktP->question.qName;
	myName = &(inMsgCbP->nbSrvCbP)->regCb.cifsSrvName;
	myIpAddr = (inMsgCbP->nbSrvCbP)->regCb.cifsSrvIpAddr;

	if (!strcmp(reqName->name, "*") ||
		nb_isNameEqual(reqName, myName) == TRUE)
	{
		/* The query name is wildcard or the CIFS server owns the name. 
		 * Send positive response. */
		rspParam.rcode = 0;
		rspParam.ttl = NB_NS_TTL;
		rspParam.dataLen = 6;

		/* Set NB_FLAGS in rdata*/
		nb_nsSetNBFlags(rdata,
						NB_NS_NBFLAG_UNIQUE,
						inMsgCbP->nbSrvCbP->currNodeType);
		/* Set IP address of the name in rdata */
		memcpy(&rdata[2], &myIpAddr, 4);
		rspParam.rdata = rdata;
	}
	else if (nsInReqPktP->header.nmFlags.bcast == FALSE)
	{
		/* Send negative response to unicast name questy request */

		/* Comment from samba's code:
		 * From observation, if a unicast packet is sent to a 
		 * non-WINS server with the recursion desired bit set
		 * then never send a negative response.
		 */
		if (nsInReqPktP->header.nmFlags.recDesired == TRUE)
			return;

		rspParam.rcode = NB_NS_NAM_ERR;
		rspParam.ttl = 0;
		rspParam.rdata = NULL;
		rspParam.dataLen = 0;
	}
	else
	{
		/* Don't send negative response to broadcast name query request */
		return;
	}

	nb_nsResponse(inMsgCbP, &rspParam);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsProcNameQueryReq()"));
	return;
} /* nb_nsProcNameQueryReq() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsProcNodeStatusReq
 *
 * Description	: Process incoming NetBIOS node status request messages.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/26/2001
 *----------------------------------------------------------------------------- 
 */
void
nb_nsProcNodeStatusReq(nb_incomingMsgCB_t	*inMsgCbP,
					   nb_nsPacket_t		*nsInReqPktP)
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsProcNodeStatusReq()"));

	/* TODO */

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsProcNodeStatusReq()"));
} /* nb_nsProcNodeStatusReq() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsProcNameReleaseReq
 *
 * Description	: Process incoming NetBIOS name release request messages.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/22/2001
 *----------------------------------------------------------------------------- 
 */
void
nb_nsProcNameReleaseReq(nb_incomingMsgCB_t	*inMsgCbP,
						nb_nsPacket_t		*nsInReqPktP)
{
	uint32				rc;
	uint8				rdata[6];
	boolean				bcast, group;
	nb_name_t			*reqName;
	nb_nsRspParam_t		rspParam;
	nb_queryCB_t		*queryCbP;
	nb_reqCB_t			*reqCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsProcNameReleaseReq()"));

	/* Partially initialize parameters for nb_nsResponse() */
	rspParam.nsInReqPktP = nsInReqPktP;
	rspParam.reqType = NB_REQ_TYPE_NAME_RELEASE;
	rspParam.opcode = NB_NS_OPCODE_RELEASE;
	rspParam.ttl = 0;
	rspParam.dataLen = 6;

	/* Set NB_FLAGS in rdata*/
	nb_nsSetNBFlags(rdata,
					NB_NS_NBFLAG_UNIQUE,
					inMsgCbP->nbSrvCbP->currNodeType);
	/* Set IP address of the name in rdata */
	memcpy(&rdata[2], &inMsgCbP->nbSrvCbP->regCb.cifsSrvIpAddr, 4);
	rspParam.rdata = rdata;


	bcast = nsInReqPktP->header.nmFlags.bcast;
	if (bcast == FALSE)
	{
		/* The CIFS server shouldn't get unicast name release packet.
		 * Send negative response.
		 */
		rspParam.rcode = NB_NS_FMT_ERR;
		nb_nsResponse(inMsgCbP, &rspParam);
		return;
	}

	group = nb_nsIsGroup(nsInReqPktP->additional->rdata);
	if (group == TRUE)
	{
		/* Ingore */
		return;
	}

	reqName = &nsInReqPktP->question.qName;

	/* Check if someone is releasing the name of one of CIFS servers */
	rc = nb_lookupSrvName(reqName);
	if (rc == NFX_OK)
	{
		 /* Send negative response. */
		rspParam.rcode = NB_NS_ACT_ERR;
		nb_nsResponse(inMsgCbP, &rspParam);
		return;
	}

	/* Lookup the name in cache */
	rc = nb_nsLookupNameInCache(inMsgCbP->nbSrvCbP, reqName, &queryCbP);
	if (rc == NFX_OK)
	{
		/* Name is found in cache. Remove the entry. */
		reqCbP = queryCbP->nbReqCbP;
		nb_deleteQueryCb(queryCbP);
		nb_deleteReqCb(reqCbP);
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsProcNameReleaseReq()"));

	return;
} /* nb_nsProcNameReleaseReq() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsProcRequest
 *
 * Description	: Process incoming NetBIOS request messages.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/22/2001
 *----------------------------------------------------------------------------- 
 */
void
nb_nsProcRequest(nb_incomingMsgCB_t	*inMsgCbP,
				 nb_nsPacket_t		*nsInReqPktP)
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsProcRequest()"));

	switch (nsInReqPktP->header.opcode)
	{
		case NB_NS_OPCODE_REG:
			if (nsInReqPktP->header.qdCount == 0 ||
				nsInReqPktP->header.arCount == 0 )
			{
				NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcRequest: "
							"Invalid REGISTRATION packet"));
			}
			else
			{
				nb_nsProcNameRegReq(inMsgCbP, nsInReqPktP);
			}
			break;

		case NB_NS_OPCODE_QUERY:
			if (nsInReqPktP->header.qdCount == 0 ||
				(nsInReqPktP->question.qType != NB_NS_QTYPE_NB &&
				 nsInReqPktP->question.qType != NB_NS_QTYPE_NBSTAT))
			{
				NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcRequest: "
							"Invalid QUERY packet"));
			}
			else if (nsInReqPktP->question.qType == NB_NS_QTYPE_NB)
			{
				nb_nsProcNameQueryReq(inMsgCbP, nsInReqPktP);
			}
			else
			{
				nb_nsProcNodeStatusReq(inMsgCbP, nsInReqPktP);
			}
			break;

		case NB_NS_OPCODE_RELEASE:
			if (nsInReqPktP->header.qdCount == 0 ||
				nsInReqPktP->header.arCount == 0)
			{
				NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcRequest: "
							"Invalid RELEASE packet"));
			}
			else
			{
				nb_nsProcNameReleaseReq(inMsgCbP, nsInReqPktP);
			}
			break;

		case NB_NS_OPCODE_REFRESH_8:
		case NB_NS_OPCODE_REFRESH_9:
				/* CIFS server is not supposed to receive a refresh request */
				NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcRequest: "
							"Received a REFRESH packet"));
				nb_nsProcNameRefreshReq(inMsgCbP, nsInReqPktP);
			break;

		default:
			NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcRequest: "
						"Opcode %d is not supported",
						nsInReqPktP->header.opcode));
			break;
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsProcRequest()"));
	return;
} /* nb_nsProcRequest */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsRcvMsg
 *
 * Description	: Process incoming NetBIOS name service messages.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/22/2001
 *----------------------------------------------------------------------------- 
 */
int32			// TPL_FREE_PKT or TPL_HOLD_PKT
nb_nsRcvMsg(nb_incomingMsgCB_t	*inMsgCbP,
			pkt_desc_t			*rcvPktDescP)	// Contains incoming message
{
	int32			rc;
	nb_srvCB_t		*nbSrvCbP = NULL;
	uchar8			*inBuf;		// Ptr to received message
	uint32			inLen;		// Length of the received message
	nb_nsPacket_t	*nsInPktP = NULL;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsRcvMsg()"));

	if (inMsgCbP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsRcvMsg: "
					"NULL pointer to incoming message control block"));
		return (TPL_FREE_PKT);
	}

	nbSrvCbP = inMsgCbP->nbSrvCbP;
	if (nbSrvCbP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsRcvMsg: "
					"NULL pointer to server control block"));
		return (TPL_FREE_PKT);
	}
	
	if (rcvPktDescP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsRcvMsg: NULL packet descriptor"));
		return (TPL_FREE_PKT);
	}
	inBuf = (uchar8 *)rcvPktDescP->pHdr;
	inLen = rcvPktDescP->hdrSize;

	/*-------------------------------------
	 * Alloc memory for name service packet
	 *--------------------------------------
	 */
	NB_RAM_ALLOC_ZERO(nsInPktP, nb_nsPacket_t, sizeof(nb_nsPacket_t));
	if (nsInPktP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsRcvMsg: "
					"Failed to alloc memory for name service packet"));
		return (TPL_FREE_PKT);
	}

	/*------------------------------------
	 * Parse received name service message
	 *------------------------------------
	 */
	rc = nb_parseNsMsg(inBuf, inLen, nsInPktP);
	if (rc != NFX_OK)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsRcvMsg: "
					"Encountered error when parsing incoming message"));
		eee_ramDealloc(nsInPktP);
		return (TPL_FREE_PKT);
	}

	/*-------------------------------------
	 * Process received name service packet
	 *--------------------------------------
	 */
	if (nsInPktP->header.rsp == TRUE)
	{
		/* Received a response message */
		nb_nsProcResponse(inMsgCbP, nsInPktP);
	}
	else
	{
		/* Received a request message */
		if (nbSrvCbP->srvNameState != NB_NAME_STATE_REGISTERED ||
			nbSrvCbP->domNameState != NB_NAME_STATE_REGISTERED)
		{
			NB_PRINT(NB_DBGLV_ERR, ("nb_nsRcvMsg: "
						"Server is not registered yet"));
			return (TPL_FREE_PKT);
		}

		nb_nsProcRequest(inMsgCbP, nsInPktP);
	}

	/* Free nsInPktP */
	eee_ramDealloc(nsInPktP);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsRcvMsg()"));

	return(TPL_FREE_PKT);
} /* nb_nsRcvMsg() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsProcNameRegResult
 *
 * Description	: Process name registration result.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/20/2001
 *----------------------------------------------------------------------------- 
 */
void
nb_nsProcNameRegResult(nb_reqCB_t		*nbReqCbP,	// Request CB
					   nb_nsPacket_t	*rspPktP,	// Response packet
					   int32			result)		// NFX_OK or NFX_ERR
{
	nb_srvCB_t	*nbSrvCbP;
	nb_name_t	*reqName;
	uint32		*state;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsProcNameRegResult()"));

	if (nbReqCbP == NULL)
		return;

	/* Stop and destroy timer */
	nb_timerStop(&nbReqCbP->reqTimerHdl);
	nb_timerDestroy(&nbReqCbP->reqTimerHdl);

	nbSrvCbP = nbReqCbP->nbSrvCbP;
	reqName = &((nb_nsPacket_t *)(nbReqCbP->reqPktP))->question.qName;

	/* Get registration state of CIFS server name or domain name */
	state = nb_getSrvNameState(nbSrvCbP, reqName);

	if (result == NFX_OK)
	{
		/* Succeeded */
		if (*state != NB_NAME_STATE_REGISTERED)
		{
			*state = NB_NAME_STATE_REGISTERED;
			(nbSrvCbP->regCb.ul_nameRegCfmInd)(nbSrvCbP->regCb.cifsSrvCbP, result);
			/* Start name list timer */
			nb_nsStartNameListTimer(nbSrvCbP);
		}
	}
	else
	{
		/* Failed */
		if (rspPktP == NULL)
			return;

		if (*state != NB_NAME_STATE_UNREGISTERED)
		{
			if (rspPktP->header.rcode == NB_NS_CFT_ERR)
			{
				/* Received a NAME CONFLICT DEMAIN */
				*state = NB_NAME_STATE_CONFLICT;
			}
			else
			{
				*state = NB_NAME_STATE_UNREGISTERED;
			}

			NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcNameRegResult: "
						"Failed to register name %s ", reqName->name));
			(nbSrvCbP->regCb.ul_nameRegCfmInd)(nbSrvCbP->regCb.cifsSrvCbP, result);
		}
	}

#if 0
	if (*state != NB_NAME_STATE_REGISTERED)
	{
		/* Registratoin is not successful yet */
		if (result == NFX_OK)
		{
			/* Succeeded */
			*state = NB_NAME_STATE_REGISTERED;
		}
		else
		{
			/* Failed */
			*state = NB_NAME_STATE_UNREGISTERED;
			NB_PRINT(NB_DBGLV_ERR, ("nb_nsProcNameRegResult: "
						"Failed to register name %s", reqName));
		}

		(nbSrvCbP->regCb.ul_nameRegCfmInd)(nbSrvCbP->regCb.cifsSrvCbP, result);
	}
#endif

	/* Remove request control block */
	if (nbReqCbP->bcastReqCbP)
	{
		/* This is a unicast request, 
		 * and has a corresponding pending broadcast */
		nbReqCbP->bcastReqCbP->ucastReqCbP = NULL;
	}

	if (nbReqCbP->ucastReqCbP)
	{
		/* This is a broadcast request, 
		 * and has a corresponding pending unicast */
		nbReqCbP->ucastReqCbP->bcastReqCbP = NULL;
	}

	nb_deleteReqCb(nbReqCbP);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsProcNameRegResult()"));

	return;
} /* nb_nsProcNameRegResult() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsProcNameQueryResult
 *
 * Description	: Process name query result.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/20/2001
 *----------------------------------------------------------------------------- 
 */
void
nb_nsProcNameQueryResult(nb_reqCB_t		*nbReqCbP,	// Request CB
				 		 nb_nsPacket_t	*rspPktP,	// Response packet
						 int32			result)		// NFX_OK or NFX_ERR
{
	nb_queryCB_t	*queryCbP;
	uint8			*rdata;
	uint32			i;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsProcNameQueryResult()"));

	if (nbReqCbP == NULL || nbReqCbP->queryCbP == NULL)
		return;

	/* Stop and destroy timer */
	nb_timerStop(&nbReqCbP->reqTimerHdl);
	nb_timerDestroy(&nbReqCbP->reqTimerHdl);

	queryCbP = nbReqCbP->queryCbP;
	if (result == NFX_OK)
	{
		/* Succeeded */
		if (rspPktP == NULL)
		{
			/* This shouldn't happen */
			panic("nb_nsProcNameQueryResult: NULL ptr to response packet");
			return;
		}

		queryCbP->state = NB_NAME_STATE_VALID;
		queryCbP->ttl = rspPktP->answer->ttl;
		queryCbP->ipAddrCnt = (rspPktP->answer->rdLength) / 6;

		rdata = rspPktP->answer->rdata;

		for(i = 0; i < queryCbP->ipAddrCnt; i++)
		{
			/* Get IP address(es) from packet */
			memcpy(&queryCbP->ipAddr[i], rdata + 2, 4);

			rdata = rdata + 6;
		}
	}
	else
	{
		/* Failed */
		queryCbP->ipAddrCnt = 0;
	}

	(queryCbP->nb_nsNameQueryInd)(queryCbP->ipAddrCnt,
								  queryCbP->ipAddr,
								  queryCbP->userInfo,
								  result);

	/* If name query failed, remove request control block */
	if (result != NFX_OK)
		nb_deleteReqCb(nbReqCbP);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsProcNameQueryResult()"));

	return;
} /* nb_nsProcNameQueryResult() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsProcNameReleaseResult
 *
 * Description	: Process name release result.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/22/2001
 *----------------------------------------------------------------------------- 
 */
void
nb_nsProcNameReleaseResult(nb_reqCB_t	*nbReqCbP,	// Request CB
						   int32		result)		// NFX_OK or NFX_ERR
{	
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsProcNameReleaseResult()"));

	if (nbReqCbP == NULL)
		return;

	/* Stop and destroy timer */
	nb_timerStop(&nbReqCbP->reqTimerHdl);
	nb_timerDestroy(&nbReqCbP->reqTimerHdl);

	if (nbReqCbP->nbSrvCbP->srvNameState == NB_NAME_STATE_UNREGISTERED)
	{
		/* Server name is already released */
	}
	else if (result == NFX_OK)
	{
		/* Succeeded */
		nbReqCbP->nbSrvCbP->srvNameState = NB_NAME_STATE_UNREGISTERED;
		nb_srvReleaseReq(nbReqCbP->nbSrvCbP, NULL);

		/* Remove request control block */
		if (nbReqCbP->bcastReqCbP)
		{
			/* This is a unicast request, 
		 	* and has a corresponding pending broadcast */
			nbReqCbP->bcastReqCbP->ucastReqCbP = NULL;
		}

		if (nbReqCbP->ucastReqCbP)
		{
			/* This is a broadcast request, 
		 	* and has a corresponding pending unicast */
			nbReqCbP->ucastReqCbP->bcastReqCbP = NULL;
		}

		nb_deleteReqCb(nbReqCbP);
	}
	else
	{
		/* TODO - Handle result == NFX_ERR case */
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsProcNameReleaseResult()"));

	return;
} /* nb_nsProcNameReleaseResult() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsReqTimeout
 *
 * Description	: Name service request timeout handler function.
 *				  If retry count is not 0, send out request again.
 *				  Otherwise, stop and destroy timer.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/12/2001
 *----------------------------------------------------------------------------- 
 */
int32			// NFX_OK or NFX_ERR
nb_nsReqTimeout(int32 timerHdl, void* param1, 
				void* param2, void* param3, void* param4)
{
	nb_reqCB_t		*reqCbP, *newReqCbP;
	nb_srvCB_t		*nbSrvCbP;
	nb_nsPacket_t	*reqPktP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsReqTimeout()"));

	reqCbP = (nb_reqCB_t *)param1;
	if (reqCbP == NULL)
		return (NFX_ERR);

	if (reqCbP->reqTimerHdl != timerHdl)
		return (NFX_ERR);

	nbSrvCbP = reqCbP->nbSrvCbP;
	reqPktP = (nb_nsPacket_t *)(reqCbP->reqPktP);

	if (reqCbP->retryCnt)
	{
		if (reqCbP->reqType == NB_REQ_TYPE_NAME_REG &&
			reqCbP->retryCnt == 1)
		{
			/* Prepare to send NAME OVERWIRTE DEMAN for
			 * broadcast name registration */
			nb_nsClearRD(reqPktP, reqCbP->reqXmtPktP);
		}

		/* Transmit the request message again */
		nb_xmtPktSP(nbSrvCbP->nameService.bsdSockP,
					reqCbP->reqXmtPktP,
					reqCbP->raddr,
					reqCbP->rport);
		reqCbP->retryCnt--;
	}
	else
	{
		switch(reqCbP->reqType & ~NB_REQ_TO_WINS)
		{
			case NB_REQ_TYPE_NAME_REG:
			case NB_REQ_TYPE_NAME_REFRESH:
				if (!(reqCbP->reqType & NB_REQ_TO_WINS))
				{
					/* Broadcast name registration succeeded */
					/* Name refresh won't come to here */
					if ((reqCbP->reqType & ~NB_REQ_TO_WINS) == NB_REQ_TYPE_NAME_REG)
						nb_nsProcNameRegResult(reqCbP, NULL, NFX_OK);
				}
				else
				{
					if (nbSrvCbP->currNodeType != NB_NS_NBFLAG_BNODE)
					{
						/* WINS server is down. Switch to B node type */
						nbSrvCbP->currNodeType = NB_NS_NBFLAG_BNODE;

						/* Send a new name registration request as B-node */
						newReqCbP = nb_nsRequest(
									nbSrvCbP,
									&nbSrvCbP->regCb.cifsSrvName,
									NB_REQ_TYPE_NAME_REG,
									nb_nsIsGroup(reqPktP->additional->rdata));
						reqCbP->bcastReqCbP = newReqCbP;
						newReqCbP->ucastReqCbP = reqCbP;
					}

					/* Keep trying to contact with WINS */

					/* Reset the retry count and double the interval between 
					 * retries (maximum at 5 minutes).
					 */
					reqCbP->retryCnt = NB_N_UCAST_REQ_RETRY;
					reqCbP->retryInterval *= 2;
					if (reqCbP->retryInterval > (5 * 60 * 1000))
						reqCbP->retryInterval = (5 * 60 * 1000);
				}
				break;

			case NB_REQ_TYPE_NAME_QUERY:
				if (!(reqCbP->reqType & NB_REQ_TO_WINS))
				{
					/* Boradcast name query failed */
					nb_nsProcNameQueryResult(reqCbP, NULL, NFX_ERR);
				}
				else
				{
					/* WINS server may be down.
					 * Send a new name query request as B-node.
					 */
					nb_queryCB_t	*newQueryCbP;

					newReqCbP = nb_nsRequest(nbSrvCbP,
											 &(reqPktP->question.qName),
											 NB_REQ_TYPE_NAME_QUERY,
											 FALSE);	// Unique name

					/* Create a name query CB in server CB */
					newQueryCbP = nb_createQueryCb(newReqCbP);
					if (newQueryCbP == NULL)
					{
						nb_deleteReqCb(newReqCbP);
					}
					else
					{
						newQueryCbP->userInfo = reqCbP->queryCbP->userInfo;
						newQueryCbP->name = reqCbP->queryCbP->name;
						newQueryCbP->nb_nsNameQueryInd = reqCbP->queryCbP->nb_nsNameQueryInd;
					}

					/* Delete the unicast name query request */
					nb_deleteReqCb(reqCbP);
				}
				break;

			case NB_REQ_TYPE_NAME_RELEASE:
				if (!(reqCbP->reqType & NB_REQ_TO_WINS))
				{
					/* Broadcast name release succeeded */
					nb_nsProcNameReleaseResult(reqCbP, NFX_OK);
				}
				else
				{
					/* WINS server may be down.
					 * Send a new name release request as B-node */
					newReqCbP = nb_nsRequest(
									nbSrvCbP,
									&nbSrvCbP->regCb.cifsSrvName,
									NB_REQ_TYPE_NAME_RELEASE,
									nb_nsIsGroup(reqPktP->additional->rdata));
					reqCbP->bcastReqCbP = newReqCbP;
					newReqCbP->ucastReqCbP = reqCbP;

					/* Keep trying to contact with WINS */

					/* Reset the retry count and double the interval between 
					 * retries (maximum at 5 minutes).
					 */
					reqCbP->retryCnt = NB_N_UCAST_REQ_RETRY;
					reqCbP->retryInterval *= 2;
					if (reqCbP->retryInterval > (5 * 60 * 1000))
						reqCbP->retryInterval = (5 * 60 * 1000);
				}
				break;

			default:
				break;
		}
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsReqTimeout()"));

	return (NFX_OK);
} /* nb_nsReqTimeout() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsConflTimeout
 *
 * Description	: Name conflict timeout handler function.
 *				  When timeout, time is stopped and destroied. 
 *				  Request control block is deleted so that any further
 *				  response messages are ignored.
 *				  If the query receives more than one response message,
 *				  there is conflict. The name being queried is removed
 *				  from cache.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/28/2001
 *----------------------------------------------------------------------------- 
 */
int32			// NFX_OK or NFX_ERR
nb_nsConflTimeout(int32 timerHdl, void* param1, 
				  void* param2, void* param3, void* param4)
{
	nb_queryCB_t	*queryCbP;
	nb_reqCB_t		*reqCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsConflTimeout()"));

	queryCbP = (nb_queryCB_t *)param1;
	if (queryCbP == NULL)
		return (NFX_ERR);

	if (queryCbP->conflTimerHdl != timerHdl)
		return (NFX_ERR);

	/* Stop and destroy timer */
	nb_timerStop(&queryCbP->conflTimerHdl);
	nb_timerDestroy(&queryCbP->conflTimerHdl);

	/* Remove request CB */
	nb_deleteReqCb(queryCbP->nbReqCbP);

	if (queryCbP->rspMsgCnt > 1)
	{
		/*------------------
		 * There is conflict
		 *------------------
		 */
		/* Remove query CB */
		nb_deleteQueryCb(queryCbP);
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsConflTimeout()"));

	return(NFX_OK);
} /* nb_nsConflTimeout() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsNameQueryReq
 *
 * Description	: Process an outgoing NetBIOS name query request.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/20/2001
 *-----------------------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR 
nb_nsNameQueryReq(nb_srvCB_t	*nbSrvCbP,		// NetBIOS server CB ptr
				  nb_name_t		name,			// Name to query
				  boolean		group,			// Unique or group name
				  void			*userInfo,		// Ptr to user specific info. 
			 									// Given back to caller in 
												// callback function.
				  void (*nb_nsNameQueryInd) (	// Callback function 
						    uint16 ipAddrCnt,		// # of IP addr of the name
							uint32 *ipAddrs,		// Prt to IP addresses
							void *userInfo,			// Ptr to user specific info
							int32 status)			// NFX_OK or NFX_ERR
				)
{
	int32			rc;
	uint16			reqType;
	nb_reqCB_t		*reqCbP = NULL;
	nb_queryCB_t	*queryCbP = NULL;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsNameQueryReq()"));

	if (nbSrvCbP->srvNameState != NB_NAME_STATE_REGISTERED ||
		nbSrvCbP->domNameState != NB_NAME_STATE_REGISTERED)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsNameQueryReq: "
					"Server %s is not registered",
					nbSrvCbP->regCb.cifsSrvName.name));
		return (NFX_ERR);
	}

	/*-------------------------
	 * Lookup the name in cache
	 *-------------------------
	 */
	rc = nb_nsLookupNameInCache(nbSrvCbP, &name, &queryCbP);
	if (rc == NFX_OK)
	{
		/* Name is found in cache */
		if (queryCbP->state == NB_NAME_STATE_VALID)
		{
			/* Name is valid */
			(nb_nsNameQueryInd) (queryCbP->ipAddrCnt,
								 queryCbP->ipAddr,
								 userInfo,
								 NFX_OK);
		}
		else
		{
			/* Name is being queried. Waiting for the response. */
		}
		return (NFX_OK);
	}

	/*--------------------------------------------
	 * Create a name query request and send it out
	 *--------------------------------------------
	 */
	reqType = NB_REQ_TYPE_NAME_QUERY;
	if (nbSrvCbP->currNodeType != NB_NS_NBFLAG_BNODE)
		reqType |= NB_REQ_TO_WINS;
	reqCbP = nb_nsRequest(nbSrvCbP, &name, reqType, group);
	if (reqCbP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsNameQueryReq: "
					"Failed in sending name query request"));
		return (NFX_ERR);
	}

	/*------------------------------------
	 * Create a name query CB in server CB 
	 *------------------------------------
	 */
	queryCbP = nb_createQueryCb(reqCbP);
	if (queryCbP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsNameQueryReq: "
					"Failed in creating name query CB"));
		nb_deleteReqCb(reqCbP);
		return(NFX_ERR);
	}

	queryCbP->userInfo = userInfo;
	queryCbP->name = name;
	queryCbP->nb_nsNameQueryInd = nb_nsNameQueryInd;

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsNameQueryReq()"));

	return (NFX_OK);
} /* nb_nsNameQueryReq() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsStartNameListTimer
 *
 * Description	: Start the timer for TTL of every name in the name list.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 11/20/2001
 *-----------------------------------------------------------------------------
 */
uint32
nb_nsStartNameListTimer(nb_srvCB_t	*nbSrvCbP)
{
	int32	rc;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsStartNameListTimer()"));

#ifndef FS_STANDALONE
	nbSrvCbP->nameListTimerHdl = eee_timerCreate(NB_NAME_LIST_TIMEOUT,
												 EEE_TIMER_REPEAT);
	if (nbSrvCbP->nameListTimerHdl == TIMER_ERR_NO_TIMER_AVAILABLE)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsStartNameListTimer: Failed to create timer"));
		return (NFX_ERR);
	}
	rc = eee_timerStart(nbSrvCbP->nameListTimerHdl,	// Timer handle
						nb_nsNameListTimeout,		// Timeout handler function
						(void *)nbSrvCbP,			// Ptr to server CB
						0, 0, 0);
	if (rc != NFX_OK)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_nsStartNameListTimer: Failed to start timer"));
	}
#endif //FS_STANDALONE

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsStartNameListTimer()"));
	return (rc);
} /* nb_nsStartNameListTimer() */

/*-----------------------------------------------------------------------------
 * Name			: nb_nsNameListTimeout 
 *
 * Description	: Name list timeout handler function.
 *				  Update TTL of a name. If TTL becomes 0 or negative,
 *				  remove the name from list.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/28/2001
 *----------------------------------------------------------------------------- 
 */
int32			// NFX_OK or NFX_ERR
nb_nsNameListTimeout(int32 timerHdl, void* param1, 
					 void* param2, void* param3, void* param4)
{
	nb_srvCB_t		*srvCbP;
	nb_queryCB_t	*queryCbP, *nextQueryCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_nsNameListTimeout()"));

	srvCbP = (nb_srvCB_t *)param1;
	if (srvCbP == NULL)
		return (NFX_ERR);

	if (srvCbP->nameListTimerHdl != timerHdl)
		return (NFX_ERR);

	/* Update ttl field of each name in the list (list of query CBs) */
	queryCbP = srvCbP->queryCbP;
	while(queryCbP)
	{
		nextQueryCbP = queryCbP->nextP;

		if (queryCbP->state != NB_NAME_STATE_VALID)
		{
			/* Name query is not done yet. Do nothing */
		}
		else if (queryCbP->ttl <= NB_NAME_LIST_TIMEOUT)
		{
			/* Remove the name from list */
			nb_deleteQueryCb(queryCbP);
		}
		else
		{
			/* Update TTL */
			queryCbP->ttl -= NB_NAME_LIST_TIMEOUT;
		}

		queryCbP = nextQueryCbP;
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_nsNameListTimeout()"));
	return (NFX_OK);
} /* nb_nsNameListTimeout() */

