/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * RCS:  $Id: nb-ss.c,v 1.1.1.1 2002/04/23 17:52:14 laurentb Exp $
 *
 * Name			: nb-ss.c
 *
 * Description	: Contains NetBIOS session service functions.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 11/07/2001
 *
 *-----------------------------------------------------------------
 */

#include "nb-api.h"
#include "nb-ns.h"
#include "nb-ss.h"
#include "nb-ds.h"
#include "nb-tpl.h"
#include "nb-util.h"
#include "nb-msg.h"

#define IFDBG(lvl)	if (lvl <= nb_debugLevel)

/*-----------------------------------------------------------------------------
 * Name			: nb_ssKeepAlive
 *  
 * Description	: Reset and restart keep-alive timer.
 *  
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/01/2001
 *-----------------------------------------------------------------------------
 */ 
void
nb_ssKeepAlive(nb_ssnCB_t *ssnCbP)
{
	int32	rc;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_ssKeepAlive"));

#ifndef FS_STANDALONE
	if (ssnCbP->kaTimerHdl == TIMER_ERR_NO_TIMER_AVAILABLE)
		return;

	nb_timerStop(&ssnCbP->kaTimerHdl);
	rc = eee_timerStart(ssnCbP->kaTimerHdl,
						nb_ssKaTimeout,			// Timeout handler function
						(void *)ssnCbP,			// Prt to session CB
						0, 0, 0);
	if (rc != NFX_OK)
	{
		NB_PRINT(NB_DBGLV_WARN, ("nb_ssKeepAlive: "
					"Failed to start keep alive timer"));
		return;
	}
#endif //FS_STANDALONE

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_ssKeepAlive"));
} /* nb_ssKeepAlive() */

/*-----------------------------------------------------------------------------
 * Name			: nb_ssKaTimeout
 *  
 * Description	: Keep-alive timeout handler function.
 * 				  It sends a keep-alive packet.
 *  
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/01/2001
 *-----------------------------------------------------------------------------
 */ 
int32			// NFX_OK or NFX_ERR
nb_ssKaTimeout(int32 timerHdl, void* param1,
			   void* param2, void* param3, void* param4)
{
	int32		rc;
	nb_ssnCB_t	*ssnCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_ssKaTimeout"));

	ssnCbP = (nb_ssnCB_t *)param1;
	if (ssnCbP == NULL)
		return (NFX_ERR);

	if (ssnCbP->kaTimerHdl != timerHdl)
		return (NFX_ERR);

	/* Send a keep-alive packet */
	rc = nb_ssXmtMsg(ssnCbP,
					 NULL,
					 NB_SS_TYPE_KEEPALIVE,
					 0);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_ssKaTimeout"));
	return (rc);
} /* nb_ssKaTimeout() */

/*-----------------------------------------------------------------------------
 * Name			: nb_ssXmtMsg
 *  
 * Description	: Send outgoing NetBIOS session service messages.
 *  
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/01/2001
 *-----------------------------------------------------------------------------
 */ 
int32			// NFX_OK or NFX_ERR
nb_ssXmtMsg(nb_ssnCB_t	*ssnCbP,
			pkt_desc_t	*xmtPktDescP,
			uint32		msgType,
	   		int32		errCode)
{
	int32	rc;
	uchar8	*outBuf;
	uint32	outLen = 0;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_ssXmtMsg()"));

	/*-------------------------------------------------------
	 * Allocate a packet descriptor from TPL.
	 * (If a CIFS message is to be sent, xmtPktDescP points 
	 * to packet descriptor that contains the CIFS message.)
	 *-------------------------------------------------------
	 */
	if (xmtPktDescP == NULL)
	{
#ifndef NBTEST
		xmtPktDescP = TPL_ALLOC_PACKET(ssnCbP->sessionService.tplConnCbP);
#else
		xmtPktDescP = PKT_ALLOC_PACKET(1);
#endif /* NBTEST */
		if (xmtPktDescP == NULL)
		{
			NB_PRINT(NB_DBGLV_ERR, ("nb_ssXmtMsg: "
						"Failed to allocate a packet descriptor"));
			return (NFX_ERR);
		}
		xmtPktDescP->hdrSize = 0;
	}

	outBuf = xmtPktDescP->pHdr;
	outLen = xmtPktDescP->hdrSize;

	/* Set packet type field */
	outBuf[0] = msgType & 0xFF;

	switch (msgType)
	{
		case NB_SS_TYPE_MSG:
			break;

		case NB_SS_TYPE_REQ:
			break;

		case NB_SS_TYPE_POSRSP:
			outLen = 0;
			break;

		case NB_SS_TYPE_NEGRSP:
			outLen = 1;

			/* Set error code */
			PCAR(outBuf, 4, errCode);
			break;

		case NB_SS_TYPE_RETRSP:
			break;

		case NB_SS_TYPE_KEEPALIVE:
			outLen = 0;
			break;

		default:
			break;
	}

	/* Set length field */
	NB_SS_SETLEN(outBuf, outLen);

	/* Set packet size */
	xmtPktDescP->hdrSize = outLen + 4;

	IFDBG(NB_DBGLV_DBG)
	{
		nb_displayDecodedSsMsg(outBuf);
		nb_displayEncodedMsg(outBuf, xmtPktDescP->hdrSize);
	}

	/* Tansmit the packet via fast path */
#ifndef NBTEST
	rc = tpl_xmtPkt(ssnCbP->sessionService.tplConnCbP, xmtPktDescP);
#else
	printf("NBTESTING ... call tpl_xmtPkt ...");
	if (msgType == NB_SS_TYPE_MSG)
		(*ssnCbP->nbSrvCbP->regCb.ul_ssnXmtCmpInd)(ssnCbP->cifsSsnCbP, 
												   xmtPktDescP->appInfo,
												   xmtPktDescP);
	rc = NFX_OK;
#endif /* NBTEST */
	if (rc != NFX_OK)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_ssXmtMsg: "
					"Failed in transmitting packet via TPL"));

		/* TODO: Handle failure case */

		return (NFX_ERR);
	}

	/* Reset and restart keep-alive timer */
	nb_ssKeepAlive(ssnCbP);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_ssXmtMsg()"));
	return (NFX_OK);
} /* nb_ssXmtMsg() */

/*-----------------------------------------------------------------------------
 * Name			: nb_ssProcUserData
 *  
 * Description	: Process incoming NetBIOS session user data, assuming
 *				  one CIFS message can be sent via several TCP packets.
 *  
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/01/2001
 *-----------------------------------------------------------------------------
 */ 
int32			// TPL_HOLD_PKT
nb_ssProcUserData(nb_ssnCB_t	*ssnCbP,		// Ptr to session CB
				  pkt_desc_t	*rcvPktDescP)	// Contains incoming message
{
	int32				rc;
	uchar8				*inBuf;		// Ptr to received session message 
	nb_ssnFragUdCB_t	*fragUdP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_ssProcUserData()"));

	rcvPktDescP->pNext = NULL;
	inBuf = (uchar8 *)rcvPktDescP->pHdr;

	fragUdP		= &(ssnCbP->fragUd);	// Session user data control block

	if (fragUdP->firstP == NULL)
	{
		/* This is the first NetBIOS packet for the CIFS message */
		/* Parse session message length */
		fragUdP->firstP = fragUdP->currP = rcvPktDescP;
		fragUdP->totalLen = NB_SS_GETLEN(inBuf);
		fragUdP->currLen = rcvPktDescP->hdrSize - NB_SS_HEADER_LEN;

		NB_PRINT(NB_DBGLV_DBG2, ("nb_ssProcUserData: "
					"First packet. totalLen = %d, currLen = %d",
					fragUdP->totalLen, fragUdP->currLen));
	}
	else
	{
		(fragUdP->currP)->pNext = rcvPktDescP;
		fragUdP->currP = rcvPktDescP;
		fragUdP->currLen += rcvPktDescP->hdrSize;
		NB_PRINT(NB_DBGLV_DBG2, ("nb_ssProcUserData: "
					"Next packet. hdrSize = %d, currLen = %d",
					rcvPktDescP->hdrSize, fragUdP->currLen));
	}

	if (fragUdP->totalLen <= fragUdP->currLen)
	{
		/* All the CIFS message, excpet CIFS user data, has been received */
		/* Send "receive packet indication" to CIFS */
		(ssnCbP->nbSrvCbP)->regCb.ul_rcvPktInd(
											ssnCbP->cifsSsnCbP,
											fragUdP->firstP,
											(fragUdP->firstP)->pHdr + NB_SS_HEADER_LEN,
											fragUdP->totalLen);

		if (fragUdP->totalLen < fragUdP->currLen)
		{
			/* Another CIFS message is sent within this packet */
			panic("nb_ssProcUserData: Two NetBIOS messages in one TCP packet.");
#if 0
			uchar8		*newP;
			pkt_desc_t	*newPktDescP;
	   
			/* Allocate a new packet descriptor */
			newPktDescP = TPL_ALLOC_PACKET(ssnCbP->sessionService.tplConnCbP);
			if (newPktDescP == NULL)
				return(TPL_HOLD_PKT);

			newP = rcvPktDescP->pHdr + 
				   rcvPktDescP->hdrSize - 
				   (fragUdP->currLen - fragUdP->totalLen);
			memcpy(newPktDescP->pHdr, newP, 
							(fragUdP->currLen - fragUdP->totalLen));

			/* Clear the fragmented user data control block
			 * for a new CIFS message. */
			memset(fragUdP, 0, sizeof(nb_ssnFragUdCB_t));

			rc = nb_ssProcUserData(ssnCbP, newPktDescP);
#endif
		}
		else
		{
			/* Clear the fragmented user data control block
			 * for a new CIFS message. */
			memset(fragUdP, 0, sizeof(nb_ssnFragUdCB_t));
		}
	}
	else
	{
		/* Waiting for more data to come to complete one CIFS message */
	}

	/* Reset and restart keep-alive timer */
	nb_ssKeepAlive(ssnCbP);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_ssProcUserData()"));

	return(TPL_HOLD_PKT);
} /* nb_ssProcUserData() */

/*-----------------------------------------------------------------------------
 * Name			: nb_ssProcEstablishReq
 *  
 * Description	: Process incoming NetBIOS session establishment request.
 *  
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/01/2001
 *-----------------------------------------------------------------------------
 */ 
int32			// TPL_FREE_PKT
nb_ssProcEstablishReq(nb_ssnCB_t	*ssnCbP,		// Ptr to session CB
					  pkt_desc_t	*rcvPktDescP)	// Contains incoming message
{
	int32			rc;
	uchar8			*inBuf;				// Ptr to received session message
	uint32			ssnLen;				// Session message length
	int32			currOffset;
	uint32			calledNameLen;
	nb_name_t		calledName;
	uint32			callingNameLen;
	nb_name_t		callingName;
	nb_name_t		*myName;
	nb_srvCB_t		*nbSrvCbP;
	uint32			msgType;
	int32			errCode;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_ssProcEstablishReq()"));

	nbSrvCbP = ssnCbP->nbSrvCbP;
	inBuf = (uchar8 *)rcvPktDescP->pHdr;

	/* Parse session message length */
	ssnLen = NB_SS_GETLEN(inBuf);

	/* Parse called name */
	currOffset = 4;
	calledNameLen = nb_parseNameMsg(inBuf, currOffset, ssnLen, &calledName);
	if (calledNameLen == 0)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_ssProcEstablishReq: "
					"Error in parsing called name"));

		/* Send negative response */
		msgType = NB_SS_TYPE_NEGRSP;
		errCode = NB_SS_ERR_CALLEDNAME_NOT_PRESENT;

		/* Delete connedtion */
		nb_ssnEndReq(ssnCbP);

		return (TPL_FREE_PKT);
	}

	/* Compare called name with CIFS server's name */
	myName = &(nbSrvCbP->regCb.cifsSrvName);
	if (nb_isNameEqual(&calledName, myName) == FALSE)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_ssProcEstablishReq: "
					"Called name is not CIFS server's name"));

		/* Send negative response */
		msgType = NB_SS_TYPE_NEGRSP;
		errCode = NB_SS_ERR_NOT_LISTEN_CALLEDNAME;

		/* Delete connedtion */
		nb_ssnEndReq(ssnCbP);

		return (TPL_FREE_PKT);
	}

	/* Parse calling name */
	currOffset += calledNameLen;
	callingNameLen = nb_parseNameMsg(inBuf, currOffset, ssnLen, &callingName);
	if (calledNameLen == 0)
	{
		NB_PRINT(NB_DBGLV_WARN, ("nb_ssProcEstablishReq: "
					"Error in parsing calling name or no calling name"));
	}

	/* Update session control block */
	if (callingNameLen)
		ssnCbP->rName = callingName;

	/* Send "session establishment indication" to CIFS */
	rc = nbSrvCbP->regCb.ul_ssnInd(nbSrvCbP->regCb.cifsSrvCbP,
								   ssnCbP,
								   &(ssnCbP->cifsSsnCbP));
	if (rc == NFX_OK)
	{
		/* Send positive response */
		ssnCbP->state = NB_SS_IN_ESTABLISHING;
		msgType = NB_SS_TYPE_POSRSP;
		errCode = 0;
	}
	else
	{
		/* CIFS layer rejects the session */
		/* Send negative response */
		ssnCbP->state = NB_SS_INIT;
		msgType = NB_SS_TYPE_NEGRSP;
		errCode = NB_SS_ERR_INSUFFICIENT_RSRC;
	}

	rc = nb_ssXmtMsg(ssnCbP, NULL, msgType, errCode);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_ssProcEstablishReq()"));

	return (TPL_FREE_PKT);
} /* nb_ssProcEstablishReq() */

/*-----------------------------------------------------------------------------
 * Name			: nb_ssRcvMsg
 *  
 * Description	: Process incoming NetBIOS session service messages.
 *  
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/01/2001
 *-----------------------------------------------------------------------------
 */ 
int32			// TPL_FREE_PKT or TPL_HOLD_PKT
nb_ssRcvMsg(nb_ssnCB_t	*ssnCbP,		// Ptr to session CB
			pkt_desc_t	*rcvPktDescP)	// Contains incoming message
{
	int32			rc;
	uchar8			*inBuf;		// Ptr to received message
	uchar8			ssnType;	// Session packet type

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_ssRcvMsg()"));

	inBuf = (uchar8 *)rcvPktDescP->pHdr;

	if (ssnCbP->fragUd.firstP != NULL)
	{
		/* This packet contains partial user data of a NetBIOS message */

		IFDBG(NB_DBGLV_DBG)
		{
			nb_displayEncodedMsg(inBuf, rcvPktDescP->hdrSize);
		}

		rc = nb_ssProcUserData(ssnCbP, rcvPktDescP);
		return (rc);
	}

#if 0
	/* Check size of the received packet against the length field in message */
	if (rcvPktDescP->hdrSize != NB_SS_GETLEN(inBuf) + 4)
	{
		NB_PRINT(NB_DBGLV_WARN, ("nb_ssProcEstablishReq: "
					"Size of received message is different from "
					"the original mesage length"));
	}
#endif

	IFDBG(NB_DBGLV_DBG)
	{
		nb_displayEncodedMsg(inBuf, rcvPktDescP->hdrSize);
		nb_displayDecodedSsMsg(inBuf);
	}

	/*------------------------
	 * Get session packet type
	 *------------------------
	 */
	ssnType = (uchar8)inBuf[0];
	switch(ssnType)
	{
		case NB_SS_TYPE_MSG:
			rc = nb_ssProcUserData(ssnCbP, rcvPktDescP);
			break;

		case NB_SS_TYPE_REQ:
			rc = nb_ssProcEstablishReq(ssnCbP, rcvPktDescP);

			break;
		case NB_SS_TYPE_POSRSP:
			break;

		case NB_SS_TYPE_NEGRSP:
			break;

		case NB_SS_TYPE_RETRSP:
			break;

		case NB_SS_TYPE_KEEPALIVE:
			/* Ignore */
			break;

		default:
			break;
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_ssRcvMsg()"));

	return (rc);
} /* nb_ssRcvMsg() */


/*-----------------------------------------------------------------------------
 * Name			: nb_ssnSetupReq
 *
 * Description	: Process a NetBIOS session establishment request. 
 * 				  Called by CIFS.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/01/2001
 *-----------------------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
nb_ssnSetupReq (nb_srvCB_t	*nbSrvCbP,		// NetBIOS server CB ptr
				uint32		raddr,			// Destination IP address
				uint16		rport,			// Destination port
				void		*cifsSsnCbP,	// CIFS session CB ptr. Given
											// back to caller in callback.
				nb_ssnCB_t	**nbSsnCbP,		// NetBIOS session CB ptr
												// upon return
				void (*ul_ssnSetupCfmInd) (		// Callback function
							void *cifsSrvCbP,	// CIFS server CB ptr
							void *cifsSsnCbP,	// CIFS session CB prt
							int32 status)		// NFX_OK or NFX_ERR

			   )
{
	return (NFX_OK);
} /* nb_ssnSetupReq() */


/*-----------------------------------------------------------------------------
 * Name			: nb_ssnXmtPktReq
 *
 * Description	: Process a NetBIOS session send message request. 
 *                Called by CIFS. 
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/01/2001
 *-----------------------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
nb_ssnXmtPktReq(nb_ssnCB_t	*nbSsnCbP,		// NetBIOS session CB ptr
				void		*cifsXmtCbP,	// CIFS xmt CB ptr
				pkt_desc_t	*msgBufferP,	// Msg to be sent buffer ptr
				uint32		smbMsgLen)		// SMB message length
{
	int32		rc;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_ssnXmtPktReq()"));

	if (nbSsnCbP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_ssnXmtPktReq: Pointer to session CB is NULL"));
		return (NFX_ERR);
	}

	/* Check if session is established */
	if (nbSsnCbP->state != NB_SS_ESTABLISHED)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_ssnXmtPktReq: Session is not established"));
		return (NFX_ERR);
	};

	msgBufferP->hdrSize = smbMsgLen;
	msgBufferP->appInfo = (void *)cifsXmtCbP;
	rc = nb_ssXmtMsg(nbSsnCbP,
					 msgBufferP,
					 NB_SS_TYPE_MSG,
					 0);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_ssnXmtPktReq()"));
	return (rc);
} /* nb_ssnXmtPktReq() */

/*-----------------------------------------------------------------------------
 * Name			: nb_ssnEndReq
 *
 * Description	: Process a NetBIOS session termination request.
 *
 * Created by	:Xiaobing Zhang
 *
 * Date created	: 12/01/2001
 *-----------------------------------------------------------------------------
 */
void
nb_ssnEndReq(nb_ssnCB_t	*nbSsnCbP)		// NetBIOS session CB ptr
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_ssnEndReq()"));

	if (nbSsnCbP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_ssnEndReq: Pointer to session CB is NULL"));
		return;
	}

#ifndef NBTEST
	tpl_delConnReq(nbSsnCbP->sessionService.tplConnCbP);
#else
	printf("nb_ssnEndReq: ... call tpl_delConnReq() ...");
	nb_tplSsDelConnInd(nbSsnCbP, nbSsnCbP->sessionService.tplConnCbP);
#endif /* NBTEST */

	/* TODO: Start SSN_CLOSE_TIMEOUT timer? */

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_ssnEndReq()"));
	return;
} /* nb_ssnEndReq() */

/*-----------------------------------------------------------------------------
 * Name			: nb_ssnGetRemoteIpAddr
 *
 * Description	: Get IP address of remote side of a session. Called by CIFS.
 *
 * Created by	:Xiaobing Zhang
 *
 * Date created	: 12/01/2001
 *-----------------------------------------------------------------------------
 */
uint32			// Remote IP address
nb_ssnGetRemoteIpAddr(nb_ssnCB_t	*nbSsnCbP)	// NetBIOS session CB ptr
{
	return (nbSsnCbP->rIpAddr);
} /* nb_ssnGetRemoteIpAddr() */

/*-----------------------------------------------------------------------------
 * Name			: nb_ssnGetRemotePort
 *
 * Description	: Get port of remote side of a session. Called by CIFS.
 *
 * Created by	:Xiaobing Zhang
 *
 * Date created	: 12/01/2001
 *-----------------------------------------------------------------------------
 */
uint16			// Remote port
nb_ssnGetRemotePort(nb_ssnCB_t	*nbSsnCbP)	// NetBIOS session CB ptr
{
	return (nbSsnCbP->rPort);
} /* nb_ssnGetRemotePort() */

/*-----------------------------------------------------------------------------
 * Name			: nb_ssnGetRemoteName
 *
 * Description	: Get NetBIOS name of remote side of a session. Called by CIFS.
 *
 * Created by	:Xiaobing Zhang
 *
 * Date created	: 11/12/2001
 *-----------------------------------------------------------------------------
 */
nb_name_t		// Remote name
nb_ssnGetRemoteName(nb_ssnCB_t	*nbSsnCbP)	// NetBIOS session CB ptr
{
	return (nbSsnCbP->rName);
} /* nb_ssnGetRemoteName() */

/*-----------------------------------------------------------------------------
 * Name			: nb_ssnGetLocalName
 *
 * Description	: Get local NetBIOS name in a session request. Called by CIFS.
 *
 * Created by	:Xiaobing Zhang
 *
 * Date created	: 11/12/2001
 *-----------------------------------------------------------------------------
 */
nb_name_t		// Local name 
nb_ssnGetLocalName(nb_ssnCB_t	*nbSsnCbP)	// NetBIOS session CB ptr
{
	return (nbSsnCbP->lName);
} /* nb_ssnGetLocalName() */

