/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * RCS:  $Id: nb-util.c,v 1.1.1.1 2002/04/23 17:52:14 laurentb Exp $
 *
 * Name			: nb-util.c
 *
 * Description	: Contains  NetBIOS utility functions.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date Created	: 11/07/2001
 *
 *-----------------------------------------------------------------
 */

#include "nb-api.h"
#include "nb-ns.h"
#include "nb-ss.h"
#include "nb-ds.h"
#include "nb-util.h"
#include "nb-msg.h"

#ifdef FS_STANDALONE
void
strtoupper (char8 *p)
{
	if (!p)
		return;

	for (; *p; p++)
		*p = toupper (*p);
} /* strtoupper() */
#endif //FS_STANDALONE

/*-----------------------------------------------------------------------------
 * Name			: nb_genXid
 *
 * Description	: Generates a unique transaction ID
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 11/07/2001
 *-----------------------------------------------------------------------------
 */
static uint16 nb_XID = 0;

uint16
nb_genXid(void)
{
	nb_XID++;
	if (nb_XID == 0xFFFF)
		nb_XID = 0;

	return nb_XID;
}

/*-----------------------------------------------------------------------------
 * Name			: nb_isNameEqual
 *
 * Description	: Compare two NetBIOS names. The comparison is case insensitive.
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 11/24/2001
 *-----------------------------------------------------------------------------
 */
boolean			// TRUE or FALSE
nb_isNameEqual(nb_name_t *name1, nb_name_t *name2)
{
	char nameStr1[NB_S_NAME + 1], nameStr2[NB_S_NAME + 1];

	if (name1->type == name2->type)
	{	
		strcpy(nameStr1, name1->name);
		strtoupper(nameStr1);
		strcpy(nameStr2, name2->name);
		strtoupper(nameStr2);

		if (!strcmp(nameStr1, nameStr2))
			return (TRUE);
		else
			return (FALSE);
	}
	else
		return (FALSE);
}

/*-----------------------------------------------------------------------------
 * Name			: nb_lookupSrvName
 *
 * Description	: Lookup server name from server list
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 11/24/2001
 *-----------------------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
nb_lookupSrvName(nb_name_t	*name)
{
	nb_srvCB_t	*nbSrvCbP;

	nbSrvCbP = nb_globalCfgP->servers;
	while(nbSrvCbP)
	{
		if (nb_isNameEqual(name, &nbSrvCbP->regCb.cifsSrvName) == TRUE)
			return (NFX_OK);
	}
	return (NFX_ERR);
} /* nb_lookupSrvName() */

/*-----------------------------------------------------------------------------
 * Name			: nb_getSrvNameState
 *
 * Description	: Get registration state of CIFS server name or domain name
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 12/17/2001
 *-----------------------------------------------------------------------------
 */
uint32 *		// Pointer to the state
nb_getSrvNameState(nb_srvCB_t *srvCbP, nb_name_t *name)
{
	if (nb_isNameEqual(name, &srvCbP->regCb.cifsSrvName) == TRUE)
		return (&srvCbP->srvNameState);
	else
		return (&srvCbP->domNameState);
} /* nb_getSrvNameState() */

/*-----------------------------------------------------------------------------
 * Name			: nb_addServerCb
 *
 * Description	: Add a server control block to list
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 11/20/2001
 *-----------------------------------------------------------------------------
 */
void
nb_addServerCb(nb_srvCB_t	**srvCbsHeadP,	// Ptr to head of server CB list
			   nb_srvCB_t	*srvCbP)		// Ptr to server CB to be added
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_addServerCb()"));

	if (*srvCbsHeadP != NULL)
	{
		srvCbP->nextP = *srvCbsHeadP;
		(*srvCbsHeadP)->prevP = srvCbP;
	}
	else
	{
		srvCbP->nextP = NULL;
	}
	srvCbP->prevP = NULL;

	*srvCbsHeadP = srvCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_addServerCb()"));
	return;
} /* nb_addServerCb() */

/*-----------------------------------------------------------------------------
 * Name			: nb_removeServerCb
 *
 * Description	: Remove a server control block from list
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 11/20/2001
 *-----------------------------------------------------------------------------
 */
void
nb_removeServerCb(nb_srvCB_t	**srvCbsHeadP,	// Ptr to head of server CB list
				  nb_srvCB_t	*srvCbP)	// Ptr to server CB to be removed
{
	NB_PRINT(NB_DBGLV_DBG2, ("> nb_removeServerCb()"));

	if (srvCbP->prevP != NULL)
	{
		(srvCbP->prevP)->nextP = srvCbP->nextP;
	}
	else
	{
		(*srvCbsHeadP)->nextP = NULL;
	}
	
	if (srvCbP->nextP != NULL)
		(srvCbP->nextP)->prevP = srvCbP->prevP;

	srvCbP->prevP = NULL;
	srvCbP->nextP = NULL;

	if (nb_globalCfgP->servers == srvCbP)
		nb_globalCfgP->servers = NULL;

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_removeServerCb()"));
	return;
} /* nb_removeServerCb() */

/*-----------------------------------------------------------------------------
 * Name			: nb_isValidServer
 *
 * Description	: Check if a server is in the server list
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 11/20/2001
 *-----------------------------------------------------------------------------
 */
int32			// NFX_OK or NFX_ERR
nb_isValidServer(nb_srvCB_t	*nbSrvCbP)
{
	nb_srvCB_t	*srvCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_isValidServer()"));

	srvCbP = nb_globalCfgP->servers;
	while(srvCbP)
	{
		if (srvCbP == nbSrvCbP)
			return (NFX_OK);
		srvCbP = srvCbP->nextP;
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_isValidServer()"));

	return (NFX_ERR);
} /* nb_isValidServer() */

/*-----------------------------------------------------------------------------
 * Name			: nb_createQueryCb
 *
 * Description	: Create a name query control block related to a server
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 11/20/2001
 *-----------------------------------------------------------------------------
 */
nb_queryCB_t *	// Ptr to query control block
nb_createQueryCb(nb_reqCB_t	*nbReqCbP)	// Ptr to NetBIOS name query request CB
{
	nb_queryCB_t	*queryCbP, *queryCbsHeadP;
	nb_srvCB_t		*nbSrvCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_createQueryCb()"));

	if (nbReqCbP == NULL)
		return (NULL);

	nbSrvCbP = nbReqCbP->nbSrvCbP;
	if (nbSrvCbP == NULL)
		return (NULL);

	/* Allocate a query CB */
    queryCbP = nb_allocQueryCb();
	if (queryCbP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_createQueryCb: "
					"Fail to alloc a query control block"));
		return (NULL);
	}
	memset(queryCbP, 0, sizeof(nb_queryCB_t));

	nbReqCbP->queryCbP = queryCbP;

	/* Add the query CB to the list in the server CB */
	queryCbsHeadP = nbSrvCbP->queryCbP;
    if (queryCbsHeadP != NULL)
	{
		queryCbP->nextP = queryCbsHeadP;
		queryCbsHeadP->prevP = queryCbP;
	}
	else
	{
		queryCbP->nextP = NULL;
	}
	queryCbP->prevP = NULL;

	nbSrvCbP->queryCbP = queryCbP;
	queryCbP->nbSrvCbP = nbSrvCbP;

	/* Init the query CB */
	queryCbP->state = NB_NAME_STATE_QUERING;
	queryCbP->group = FALSE;
	queryCbP->rspMsgCnt = 0;
	queryCbP->nbReqCbP = nbReqCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_createQueryCb()"));

	return (queryCbP);
} /* nb_createQueryCb() */

/*-----------------------------------------------------------------------------
 * Name			: nb_deleteQueryCb
 *
 * Description	: Delete a query control block
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 11/15/2001
 *-----------------------------------------------------------------------------
 */
void
nb_deleteQueryCb(nb_queryCB_t	*nbQueryCbP)
{
	nb_srvCB_t	*nbSrvCbP;
	nb_reqCB_t	*nbReqCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_deleteQueryCb()"));

	if (nbQueryCbP == NULL)
		return;

	/* Remove the query CB from the query CB list in the server CB */
	nbSrvCbP = nbQueryCbP->nbSrvCbP;
	if (nbQueryCbP->prevP)
		(nbQueryCbP->prevP)->nextP = nbQueryCbP->nextP;
	else
		nbSrvCbP->queryCbP = nbQueryCbP->nextP;

	if (nbQueryCbP->nextP)
		(nbQueryCbP->nextP)->prevP = nbQueryCbP->prevP;

	/* Set pointer to the query CB in request CB to NULL */
	nbReqCbP = nbQueryCbP->nbReqCbP;
	if (nbReqCbP)
		nbReqCbP->queryCbP = NULL;

	/* Free the query CB */
	nb_freeQueryCb(nbQueryCbP);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_deleteQueryCb()"));
	return;
} /* nb_deleteQueryCb() */


/*-----------------------------------------------------------------------------
 * Name			: nb_createReqCb
 *
 * Description	: Create a request control block related to a server
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 11/15/2001
 *-----------------------------------------------------------------------------
 */
nb_reqCB_t *	// Ptr to request control block
nb_createReqCb(nb_srvCB_t	*nbSrvCbP,	// Prt to NetBIOS srv CB
			   uint16		reqType)
{
	nb_reqCB_t	*reqCbP, *reqCbsHeadP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_createReqCb()"));

	if (nbSrvCbP == NULL)
		return (NULL);

	/* Allocate a request CB */
    reqCbP = nb_allocReqCb();
	if (reqCbP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_createReqCb: "
					"Fail to alloc a request control block"));
		return (NULL);
	}
	memset(reqCbP, 0, sizeof(nb_reqCB_t));

	/* Add the request CB to the list in the server CB */
	reqCbsHeadP = nbSrvCbP->reqCbP;
    if (reqCbsHeadP != NULL)
	{
		reqCbP->nextP = reqCbsHeadP;
		reqCbsHeadP->prevP = reqCbP;
	}
	else
	{
		reqCbP->nextP = NULL;
	}
	reqCbP->prevP = NULL;

	nbSrvCbP->reqCbP = reqCbP;
	reqCbP->nbSrvCbP = nbSrvCbP;

	/* Init the request CB */
	reqCbP->reqType = reqType;	
	if (!(reqType & NB_REQ_TO_WINS))
	{
		/* Broadcast */
		reqCbP->retryCnt = NB_N_BCAST_REQ_RETRY;
		reqCbP->retryInterval = NB_BCAST_REQ_RETRY_TIMEOUT_MS;
	}
	else
	{
		/* Unicast */
		reqCbP->retryCnt = NB_N_UCAST_REQ_RETRY;

		if ((reqType & ~NB_REQ_TO_WINS) == NB_REQ_TYPE_NAME_REFRESH)
		{
			/* For name refresh request, retry interval is ttl in
			 * received positive name registration response message.
			 */
			reqCbP->retryInterval = nbSrvCbP->srvNameTTL;
		}
		else
			reqCbP->retryInterval = NB_UCAST_REQ_RETRY_TIMEOUT;
	}
	reqCbP->reqTimerHdl = 0;

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_createReqCb()"));

	return (reqCbP);
} /* nb_createReqCb() */

/*-----------------------------------------------------------------------------
 * Name			: nb_deleteReqCb
 *
 * Description	: Delete a request control block
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 11/15/2001
 *-----------------------------------------------------------------------------
 */
void
nb_deleteReqCb(nb_reqCB_t	*nbReqCbP)
{
	nb_srvCB_t	*nbSrvCbP;
	nb_reqCB_t	*reqCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_deleteReqCb()"));

	if (nbReqCbP == NULL || nbReqCbP->nbSrvCbP == NULL)
		return;

	nbSrvCbP = nbReqCbP->nbSrvCbP;

	/* Remove the request CB from the request CB list in the server CB */
	if (nbReqCbP->prevP)
		(nbReqCbP->prevP)->nextP = nbReqCbP->nextP;
	else
		nbSrvCbP->reqCbP = nbReqCbP->nextP;

	if (nbReqCbP->nextP)
		(nbReqCbP->nextP)->prevP = nbReqCbP->prevP;

	/* Free memory allocated to store packet */
	switch (nbReqCbP->reqType & ~NB_REQ_TO_WINS)
	{
		case NB_REQ_TYPE_NAME_REG:
		case NB_REQ_TYPE_NAME_QUERY:
		case NB_REQ_TYPE_NAME_RELEASE:
		case NB_REQ_TYPE_NAME_REFRESH:
			nb_nsDeletePkt((nb_nsPacket_t *)nbReqCbP->reqPktP);
			break;

		case NB_REQ_TYPE_SESSION:
			break;

		case NB_REQ_TYPE_DATAGRAM:
			break;

		default:
			break;
	}

	/* Free memory allocated to store transmitted message */
	if (nbReqCbP->reqXmtPktP)
#ifndef NBTEST
		PKT_FREE_PACKET(nbReqCbP->reqXmtPktP);
#endif /* NBTEST */

	/* Stop and destroy timer */
	nb_timerStop(&nbReqCbP->reqTimerHdl);
	nb_timerDestroy(&nbReqCbP->reqTimerHdl);

	/* Delete related query CB if the name is being queried */
	if (nbReqCbP->queryCbP)
	{
		if (nbReqCbP->queryCbP->state == NB_NAME_STATE_QUERING)
		{
			nb_deleteQueryCb(nbReqCbP->queryCbP);
		}
		else
			(nbReqCbP->queryCbP)->nbReqCbP = NULL;
	}

	nbReqCbP->nbSrvCbP = NULL;

	/* Free the request CB itself */
	nb_freeReqCb(nbReqCbP);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_deleteReqCb()"));
	return;
} /* nb_deleteReqCb() */

/*-----------------------------------------------------------------------------
 * Name			: nb_getReqCb
 *
 * Description	: Get a request control block that has give Xid
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 11/15/2001
 *-----------------------------------------------------------------------------
 */
nb_reqCB_t *	// Ptr to request CB
nb_getReqCb(nb_srvCB_t	*nbSrvCbP,
			uint16		xid)
{
	nb_reqCB_t	*nbReqCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_getReqCb()"));

	/* Search all request CBs of the server */
	nbReqCbP = nbSrvCbP->reqCbP;
	while(nbReqCbP)
	{
		if (nbReqCbP->reqID == xid)
			return (nbReqCbP);
		nbReqCbP = nbReqCbP->nextP;
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_getReqCb()"));
	return (NULL);
} /* nb_getReqCb() */

/*-----------------------------------------------------------------------------
 * Name			: nb_createSsnCb
 *
 * Description	: Create a session control block related to a server
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 12/01/2001
 *-----------------------------------------------------------------------------
 */
nb_ssnCB_t *	// Ptr to session control block
nb_createSsnCb(nb_srvCB_t	*nbSrvCbP)	// Prt to NetBIOS srv CB
{
	nb_ssnCB_t	*ssnCbP, *ssnCbsHeadP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_createSsnCb()"));

	if (nbSrvCbP == NULL)
		return (NULL);

	/* Allocate a session CB */
    ssnCbP = nb_allocSsnCb();
	if (ssnCbP == NULL)
	{
		NB_PRINT(NB_DBGLV_ERR, ("nb_createSsnCb: "
					"Fail to alloc a session control block"));
		return (NULL);
	}
	memset(ssnCbP, 0, sizeof(nb_ssnCB_t));

	/* Add the session CB to the list in the server CB */
	ssnCbsHeadP = nbSrvCbP->ssnCbP;
    if (ssnCbsHeadP != NULL)
	{
		ssnCbP->nextP = ssnCbsHeadP;
		ssnCbsHeadP->prevP = ssnCbP;
	}
	else
	{
		ssnCbP->nextP = NULL;
	}
	ssnCbP->prevP = NULL;

	nbSrvCbP->ssnCbP = ssnCbP;
	ssnCbP->nbSrvCbP = nbSrvCbP;

	/* Init the session CB */
	ssnCbP->lName = nbSrvCbP->regCb.cifsSrvName;
	ssnCbP->state = NB_SS_INIT;

#ifndef FS_STANDALONE
	/* Create keep-alive timer */
	ssnCbP->kaTimerHdl = eee_timerCreate(NB_SS_KEEP_ALIVE_TIMEOUT,
										 EEE_TIMER_REPEAT);
	if (ssnCbP->kaTimerHdl == TIMER_ERR_NO_TIMER_AVAILABLE)
	{
		NB_PRINT(NB_DBGLV_WARN, ("nb_createSsnCb: "
					"Failed to create keep-alive timer"));
	}
#endif //FS_STANDALONE

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_createSsnCb()"));

	return (ssnCbP);
} /* nb_createSsnCb() */

/*-----------------------------------------------------------------------------
 * Name			: nb_deleteSsnCb
 *
 * Description	: Delete a session control block
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 12/01/2001
 *-----------------------------------------------------------------------------
 */
void
nb_deleteSsnCb(nb_ssnCB_t	*nbSsnCbP)
{
	nb_srvCB_t	*nbSrvCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_deleteSsnCb()"));

	if (nbSsnCbP == NULL)
		return;

	nbSrvCbP = nbSsnCbP->nbSrvCbP;

	/* Remove the session CB from the session CB list in the server CB */
	if (nbSsnCbP->prevP)
		(nbSsnCbP->prevP)->nextP = nbSsnCbP->nextP;
	else
		nbSrvCbP->ssnCbP = nbSsnCbP->nextP;

	if (nbSsnCbP->nextP)
		(nbSsnCbP->nextP)->prevP = nbSsnCbP->prevP;

#ifndef FS_STANDALONE
	/* Destroy keep-alive timer */
	if (nbSsnCbP->kaTimerHdl != TIMER_ERR_NO_TIMER_AVAILABLE)
	{
		nb_timerStop(&nbSsnCbP->kaTimerHdl);
		nb_timerDestroy(&nbSsnCbP->kaTimerHdl);
	}
#endif //FS_STANDALONE

	/* Free the request CB itself */
	nb_freeSsnCb(nbSsnCbP);

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_deleteSsnCb()"));
	return;
} /* nb_deleteSsnCb() */

/*-----------------------------------------------------------------------------
 * Name			: nb_getSsnCb
 *
 * Description	: Get a session control block that has give 
 * 				  remote IP address and port 
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 12/01/2001
 *-----------------------------------------------------------------------------
 */
nb_ssnCB_t *	// Ptr to session CB
nb_getSsnCb(nb_srvCB_t	*srvCbP,
			uint32		raddr,			// Remote IP address
			uint16		rport)			// Remote port
{
	nb_ssnCB_t	*ssnCbP;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_getSsnCb()"));

	/* Search all session CBs of the server */
	ssnCbP = srvCbP->ssnCbP;
	while(ssnCbP)
	{
		if (ssnCbP->rIpAddr == raddr && ssnCbP->rPort == rport)
			return (ssnCbP);
		ssnCbP = ssnCbP->nextP;
	}

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_getSsnCb()"));
	return (NULL);
} /* nb_getSsnCb() */

/*-----------------------------------------------------------------------------
 * Name			: nb_createXmtPktSP
 *
 * Description	: Prepare the packet to be sent via slow path.
 *
 * Created by	: Xiaobing Zhang
 * 
 * Date Created	: 11/15/2001
 *-----------------------------------------------------------------------------
 */
pkt_desc_t *	// Prt to packet descriptor containing message to be sent
nb_createXmtPktSP(void			*pktP,		// NetBIOS service packet
				  uint32		svcType,	// NetBIOS service type
				  pkt_desc_t	*xmtPktDescP)	// Preallocated packet desc
												// It may be NULL
{
	uchar8			*nbBufP;
	uint32			xmtMsgLen;
	boolean			prealloc = TRUE;;

	NB_PRINT(NB_DBGLV_DBG2, ("> nb_createXmtPktSP()"));

	/*----------------------------------------------------------------
	 * Allocate a packet from TPL to be used to store outgoing packet.
	 *----------------------------------------------------------------
	 */
	if (xmtPktDescP == NULL)
	{
		prealloc = FALSE;
		xmtPktDescP = PKT_ALLOC_PACKET(1);
		if (xmtPktDescP == NULL)
		{
			NB_PRINT(NB_DBGLV_ERR, ("nb_createXmtPktSP: Failed to alloc a packet"));
			return (NULL);
		}
	}
	nbBufP = (uchar8 *)xmtPktDescP->pHdr;

	/*-------------------------------------------------------
	 * Build service message to be sent based on service type
	 *-------------------------------------------------------
	 */
	switch(svcType)
	{
		case NB_NS:
			xmtMsgLen = nb_buildNsMsg(nbBufP, (nb_nsPacket_t *)pktP);
			break;

		case NB_SS:
			break;

		case NB_DS:
			xmtMsgLen = nb_buildDsMsg(nbBufP,
									  (nb_dsPacket_t *)pktP,
									  NB_DS_DGM_TYPE_DIR_GROUP);
			break;

		default:
			break;
	}

	if (prealloc == FALSE)
		xmtPktDescP->hdrSize = xmtMsgLen;

	NB_PRINT(NB_DBGLV_DBG2, ("< nb_createXmtPktSP()"));

	return (xmtPktDescP);
} /* nb_createXmtPktSP() */

/*-----------------------------------------------------------------------------
 * Name			: nb_allocOutgoingPkt
 *
 * Description	: Calls TPL packet allocation function to allocate a packet 
 * 				  buffer. Called by CIFS.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/03/01
 *-----------------------------------------------------------------------------
 */
pkt_desc_t *	// PKT descriptor or NULL
nb_allocOutgoingPkt (nb_ssnCB_t	*nbSsnCbP,		// NetBIOS session CB ptr
			 									// For datagram, it is NULL.
					 void		**smbMsgHdrP)	// Ptr to location where
												// the SMB message should start.
{
	pkt_desc_t	*newPktDescP;

	if (nbSsnCbP != NULL)
	{
		/* Allocate pkt descriptor for session */
#ifdef LINUX_TEST
		newPktDescP = TPL_ALLOC_PACKET(nbSsnCbP->sessionService.tplConnCbP);
#else
		newPktDescP = TPL_ALLOC_PACKET(nbSsnCbP->sessionService.tplConnCbP, 0);
#endif		
		if (newPktDescP == NULL)
			return (NULL);
		*smbMsgHdrP = newPktDescP->pHdr + NB_SS_HEADER_LEN;
	}
	else
	{
		/* Allocate pkt descriptor for datagram */

		newPktDescP = PKT_ALLOC_PACKET(1);
		if (newPktDescP == NULL)
			return (NULL);

		*smbMsgHdrP = newPktDescP->pHdr +
					  14 +						// Length of datagram header
					  NB_S_COMPRESSED_NAME * 2;	// Length of 2 NetBIOS names
	}

	newPktDescP->pNext = NULL;

	return (newPktDescP);
} /* nb_allocOutgoingPkt() */

/*-----------------------------------------------------------------------------
 * Name			: nb_freeOutgoingPkt
 *
 * Description	: Calls TPL packet free function to free a packet buffer.
 * 				  Called by CIFS.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/03/01
 *-----------------------------------------------------------------------------
 */
int32			// NFX_OK or TPL error
nb_freeOutgoingPkt (nb_ssnCB_t	*nbSsnCbP,	// NetBIOS session CB ptr
											// For datagram, it is NULL
					pkt_desc_t	*pktDescP)	// Packet descriptor to be freed
{
	pkt_desc_t	*currP, *nextP;
   	
	currP = nextP = pktDescP;

	if (nbSsnCbP != NULL)
	{
		/* This is a session packet */
		return (TPL_FREE_PACKET(nbSsnCbP->sessionService.tplConnCbP, currP));
	}
	else
	{
		/* This is a datagram packet */
#ifndef NBTEST
		return (PKT_FREE_PACKET(currP));
#else
		return (NFX_OK);
#endif /* NBTEST */
	}
} /* nb_freeOutgoingPkt() */

/*-----------------------------------------------------------------------------
 * Name			: nb_freeIncomingPkt
 *
 * Description	: Calls PKT or TPL packet free function to free a 
 *				  packet buffer that stores incoming packet. Called by CIFS.
 *
 * Created by	: Xiaobing Zhang
 *
 * Date created	: 12/03/01
 *-----------------------------------------------------------------------------
 */
int32			// NFX_OK or PKT error
nb_freeIncomingPkt (nb_ssnCB_t	*nbSsnCbP,	// NetBIOS session CB ptr
											// For datagram, it is NULL
					pkt_desc_t	*pktDescP)	// Packet descriptor to be freed
{
	int32		rc;
	pkt_desc_t	*currP, *nextP;
   	
	currP = pktDescP;
	nextP = currP->pNext;

	//Free all the buffers in the chain
	while(currP)
	{
		if (nbSsnCbP != NULL)
		{
			/* This is a session packet */
			rc = TPL_FREE_PACKET(nbSsnCbP->sessionService.tplConnCbP, currP);
		}
		else
		{
			/* This is a datagram packet */
			rc = PKT_FREE_PACKET(currP);
		}
		currP = nextP;
		nextP = currP->pNext;
	}
	return (rc);
} /* nb_freeIncomingPkt() */


int32 nb_timerStop(int32 *tmr_handle)
{
#ifndef FS_STANDALONE
	if (*tmr_handle != TIMER_ERR_INVALID_HANDLE)
	{
		return(eee_timerStop(*tmr_handle));
	}
#endif //FS_STANDALONE
	/* Already destroyed */
	return NFX_OK;
}

int32 nb_timerDestroy(int32 *tmr_handle)
{
	int32	rc;
#ifndef FS_STANDALONE
	if (*tmr_handle != TIMER_ERR_INVALID_HANDLE)
	{
		rc = eee_timerDestroy(*tmr_handle);
		*tmr_handle = TIMER_ERR_INVALID_HANDLE;
		return rc;
	}
#endif //FS_STANDALONE
	/* Already destroyed */
	return NFX_OK;
}

