/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        nfs-nfp.c
 *
 * RCS:  $Id: nfs-nfp.c,v 1.16 2002/10/01 00:39:16 shaun Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:
 *  This file contains interface functions between NFS and NFP.
 *
 * Created by: Joseph Vu
 *
 * Date Created:        7/28/01
 *
 *-----------------------------------------------------------------
 */
#ifdef	NFP_TXRX
#include <nfx-incl.h>
#include <exports/nfs-api.h>

/*-----------------------------------------------------------------
 * Name:        nfs_addNode
 *
 * Description:
 *    Allocate a generic node to hold a control block.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
nfp_nodePtr_t nfs_addNode(void *CtrlBlkp,   /*Control block to be added*/
            int32 nodeType)                 /*Control block node type*/
{
    nfp_nodePtr_t   addNodep;
    /*
     *Alloc node to store entry
     */
    addNodep = nfp_allocNode();
    if (addNodep == NULL)
    {
        return (NULL);
    }
    addNodep->nfp_nodeType = nodeType;
    addNodep->nfp_entryCBp = CtrlBlkp;
    addNodep->nfp_flag |= NFP_ENTRY_INUSED;
    addNodep->nfp_prevp = NULL;
    addNodep->nfp_nextp = NULL;
    return (addNodep);
}


/*-----------------------------------------------------------------
 * Name:        nfs_addCBToList
 *
 * Description:
 *     Add a specified Control Block (CB) into a active list.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32   /*
         *NFX_OK:  Entry has been added to the list
         *NFX_ERR:  Failure to add entry into the list
         */
nfs_addCBToList(nfp_list_t *nodeListp, void *CtrlBlkp, uint32 nodeType)
{
    nfp_nodePtr_t   addNodep;

    addNodep = nfs_addNode(CtrlBlkp, nodeType);
    if (addNodep == NULL)
        return (NFX_ERR);
    if (nodeListp->nfp_headp == NULL)
    {
        /*
         *Add to empty list
         */
        nodeListp->nfp_headp = addNodep;
        nodeListp->nfp_tailp = addNodep;
    }
    else
    {
        /*
         *Add to head of the list
         */
        nodeListp->nfp_headp->nfp_prevp = addNodep;
        addNodep->nfp_nextp =  nodeListp->nfp_headp;
        nodeListp->nfp_headp  = addNodep;
    }

	/* Shaun 08/10/2002
	 *
	 * Performance tunning:
	 * Store node ptr in REQ CB.
	 * Used for delele.
	 */
	if (nodeType == NFP_REQCB_NODE)
	{
		((nfs_request_t *)CtrlBlkp)->pNode = addNodep;
	}

    nodeListp->nfp_reqCnt++;

    return (NFX_OK);

}

/*-----------------------------------------------------------------
 * Name:        nfs_delReqCBFromList
 *
 * Description:
 *  Delete Request Control Block (RCB) from RCB list in the
 *  Client Control Block.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfs_delReqCBFromList(nfs_request_t *nfsReqp)
{
    nfp_list_t          *reqListp;

    reqListp = nfs_globalCfgp->pReqCbList;

    return (nfs_delCBFromList(reqListp, (void *)nfsReqp, NFP_REQCB_NODE));
}

/*-----------------------------------------------------------------
 * Name:   	nfs_delFileSysCB 
 *
 * Description:
 * 
 *  Delete connection indication from the transport layer to delete
 *  an existing connection.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
void nfs_delFileSysCB(nfp_clientCBPtr_t   clientCBp)
{
    uint32                  count;
    nfp_nodePtr_t           nodep;
    nfp_fileSysCBPtr_t      entryFileSysp;
	nfp_list_t 				*nodeListp;


	nodeListp = &clientCBp->nfp_fileSysList;
    count = nodeListp->nfp_reqCnt;
    for (nodep = nodeListp->nfp_headp; (nodep != NULL && count > 0);
        nodep = nodep->nfp_nextp, count--)
    {
		if (nfp_validateNode(nodep) != NFX_OK)
			continue;
        entryFileSysp = (nfp_fileSysCBPtr_t)nodep->nfp_entryCBp;
        if ( (nfp_validateFileSysCB(entryFileSysp) != NFX_OK) ||
			(!(entryFileSysp->nfp_flag & NFP_ENTRY_INUSED)) )
        {
            continue;
        }
		nfs_delFileSysCBFromList(entryFileSysp, clientCBp);
    }
}


/*-----------------------------------------------------------------
 * Name:        nfs_findReqCBNode
 *
 * Description:
 *      Find the node contains the input Request Control Block (RCB)
 *  from the input list.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
nfp_nodePtr_t nfs_findReqCBNode(nfp_list_t *nodeListp, nfs_request_t *nfsReqp)
{
    uint32                  count;
    nfs_request_t           *entryReqp;
    nfp_nodePtr_t           nodep;

    count = nodeListp->nfp_reqCnt;
    for (nodep = nodeListp->nfp_headp; (nodep != NULL && count > 0);
            nodep = nodep->nfp_nextp, count--)
    {
        entryReqp = (nfs_request_t *)nodep->nfp_entryCBp;
        if ((nfsReqp->rpcCallHdrp && entryReqp->rpcCallHdrp) &&
			(nfsReqp->rpcCallHdrp->rpc_xid == entryReqp->rpcCallHdrp->rpc_xid) &&
            (nfsReqp->rpcCallHdrp->rpc_proc == entryReqp->rpcCallHdrp->rpc_proc))
        {
            return (nodep);
        }
    }
    return (NULL);
}

/*-----------------------------------------------------------------
 * Name:        nfs_findNlmReqCBNode
 *
 * Description:
 *      Find the node contains the input NLM Request Control Block (RCB)
 *  from the input list.
 *
 * Created by:  Xiaobing Zhang
 *
 * Date Created:        7/19/02
 *
 *-----------------------------------------------------------------
 */
nfp_nodePtr_t nfs_findNlmReqCBNode(nfp_list_t *nodeListp, nlm_request_t *nlmReqp)
{
    uint32                  count;
    nlm_request_t           *entryReqp;
    nfp_nodePtr_t           nodep;

    count = nodeListp->nfp_reqCnt;
    for (nodep = nodeListp->nfp_headp; (nodep != NULL && count > 0);
            nodep = nodep->nfp_nextp, count--)
    {
        entryReqp = (nlm_request_t *)nodep->nfp_entryCBp;
        if ((nlmReqp->rpcCallHdrp && entryReqp->rpcCallHdrp) &&
			(nlmReqp->rpcCallHdrp->rpc_xid == entryReqp->rpcCallHdrp->rpc_xid) &&
            (nlmReqp->rpcCallHdrp->rpc_proc == entryReqp->rpcCallHdrp->rpc_proc))
        {
            return (nodep);
        }
    }
    return (NULL);
}

/*-----------------------------------------------------------------
 * Name:        nfs_findFileSysCBNode
 *
 * Description:
 *  Find the node that contains specified File System Control Block
 *  from the input list.
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
nfp_nodePtr_t nfs_findFileSysCBNode(nfp_list_t *nodeListp,  /*list*/
        nfp_fileSysCBPtr_t fileSysCBp   /*fileSysCB to be retrieved*/
)
{
    uint32                  count;
    nfp_nodePtr_t           nodep;
    nfp_fileSysCBPtr_t      entryFileSysp;

	if ( (nodeListp == NULL) ||
		 (fileSysCBp == NULL) )
		return (NULL);
		
    count = nodeListp->nfp_reqCnt;
    for (nodep = nodeListp->nfp_headp; (nodep != NULL && count > 0);
        nodep = nodep->nfp_nextp, count--)
    {
        entryFileSysp = (nfp_fileSysCBPtr_t)nodep->nfp_entryCBp;
		if ( (nfp_validateFileSysCB(entryFileSysp) != NFX_OK) ||
			(!(entryFileSysp->nfp_flag & NFP_ENTRY_INUSED)) )
		{
			continue;
		}
        if ( (entryFileSysp == fileSysCBp) &&
			(entryFileSysp->nfp_fileHandle.volId.whole_thing ==
                            fileSysCBp->nfp_fileHandle.volId.whole_thing) &&
			(entryFileSysp->nfp_fileHandle.mntPnt == fileSysCBp->nfp_fileHandle.mntPnt ) )
            return (nodep);
    }
    return (NULL);

}
/*-----------------------------------------------------------------
 * Name:       nfs_findFileCBNode
 *
 * Description:
 * Find the node that contains specified File Control Block from the input list.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
nfp_nodePtr_t
nfs_findFileCBNode(nfp_list_t *nodeListp,   /*list*/
            nfp_fileCBPtr_t fileCBp         /*fileCB to be retrieved*/
)
{
    uint32                  count;
    nfp_nodePtr_t           nodep;
    nfp_fileCBPtr_t         entryFileCBp;

    count = nodeListp->nfp_reqCnt;
    for (nodep = nodeListp->nfp_headp; (nodep != NULL && count > 0);
        nodep = nodep->nfp_nextp, count--)
    {
        entryFileCBp = (nfp_fileCBPtr_t)nodep->nfp_entryCBp;
        if ((entryFileCBp == fileCBp) &&
            (entryFileCBp->nfp_fileHandle.inum == fileCBp->nfp_fileHandle.inum) )
            return (nodep);
        }
        return (NULL);
}

/*-----------------------------------------------------------------
 * Name:        nfs_findClientCBNode
 *
 * Description:
 *  Find the node that contains the specified Client Control Block from
 *  the input list.
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
nfp_nodePtr_t
nfs_findClientCBNode (nfp_list_t *nodeListp,
            nfp_clientCBPtr_t clientCBp
)
{
    uint32                  count;
    nfp_nodePtr_t           nodep;
    nfp_clientCBPtr_t       entryClientCBp;

    count = nodeListp->nfp_reqCnt;
    for (nodep = nodeListp->nfp_headp; (nodep != NULL && count > 0);
        nodep = nodep->nfp_nextp, count--)
    {
        entryClientCBp = (nfp_clientCBPtr_t)nodep->nfp_entryCBp;
        if ((entryClientCBp == clientCBp) &&
            (entryClientCBp->nfp_connId2 == clientCBp->nfp_connId2) )
        return (nodep);
    }
    return (NULL);

}

/*-----------------------------------------------------------------
 :          nfs_delNode
 *
 * Description:
 *      Deletes a node from the input list.
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32 nfs_delNode(nfp_list_t *nodeListp, nfp_nodePtr_t delNodep)
{
    nfp_nodePtr_t           tempNodep;

    if ( (delNodep == NULL) || (nodeListp == NULL) )
        return (NFX_ERR);


    if (nodeListp->nfp_reqCnt == 1)
    {
        nfp_deallocNode(delNodep);
        nodeListp->nfp_reqCnt -= 1;
        nodeListp->nfp_headp = NULL;
        nodeListp->nfp_tailp = NULL;
        return (NFX_OK);
    }
    if (delNodep->nfp_prevp == NULL)
    {
        /*
         *We're at the head
         */
        nodeListp->nfp_headp = delNodep->nfp_nextp;     /*update head*/
        nodeListp->nfp_headp->nfp_prevp = NULL;         /*update next mode*/
    }
    else if (delNodep->nfp_nextp == NULL)
    {
        /*
         *We're at the tail
         */
        tempNodep = delNodep->nfp_prevp;                /*get the previous node*/
        tempNodep->nfp_nextp = NULL;
        nodeListp->nfp_tailp = tempNodep;               /*update tail*/
    }
    else
    {
        /*
         *We're are in the list
         */
        tempNodep = delNodep->nfp_prevp;
        tempNodep->nfp_nextp = delNodep->nfp_nextp;
        tempNodep = delNodep->nfp_nextp;
        tempNodep->nfp_prevp = delNodep->nfp_prevp;
    }

    nfp_deallocNode(delNodep);
    nodeListp->nfp_reqCnt -= 1;
    return (NFX_OK);

}
/*-----------------------------------------------------------------
 * Name:      nfs_delCBFromList
 *
 * Description:
 *      Delete a Control Block entry from the input list.
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32 nfs_delCBFromList(
        nfp_list_t *nodeListp,
        void *CtrlBlkp,                 /*Control Block to be deleted*/
        int32 nodeType                 /*Type of node or Control Block to be deleted*/
)
{
    nfp_nodePtr_t           delNodep;

    /*
     *Look for entry to remove from the list
     */
    delNodep = NULL;
    switch (nodeType)
    {
        case NFP_CLIENTCB_NODE:                 /*ClientCB entry*/
            delNodep = nfs_findClientCBNode(nodeListp, (nfp_clientCBPtr_t)CtrlBlkp);
            break;
        case NFP_FILESYSCB_NODE:                /*FileSysCB entry*/
            delNodep = nfs_findFileSysCBNode(nodeListp,(nfp_fileSysCBPtr_t)CtrlBlkp);
            break;

        case NFP_FILECB_NODE:                   /*FileCB entry*/
            delNodep = nfs_findFileCBNode(nodeListp, (nfp_fileCBPtr_t)CtrlBlkp);
            break;
        case NFP_REQCB_NODE:                    /*RequestCB entry*/

			/* Shaun 08/10/02
			 *
			 * Performance tuning:
			 * 1 to 1 relationship of reqCb to nfp_node
			 */
            delNodep = ((nfs_request_t *)CtrlBlkp)->pNode;
            ((nfs_request_t *)CtrlBlkp)->pNode = NULL;

            break;
        case NFP_NLMREQCB_NODE:                 /*NLM RequestCB entry*/
            delNodep = nfs_findNlmReqCBNode(nodeListp, (nlm_request_t *)CtrlBlkp);
            break;
        case NFP_DIRCB_NODE:                    /*Directory entry*/
            delNodep = nfs_findFileCBNode(nodeListp, (nfp_fileCBPtr_t)CtrlBlkp);
            break;
    }
    if (delNodep == NULL)
    {
        return (NFX_ERR);
    }

    return (nfs_delNode(nodeListp, delNodep));
}

/*-----------------------------------------------------------------
 * Name:        nfs_addFileSysCBToList
 *
 * Description:
 *      Add File System Control Block (FSCB) into active file system
 * list mounted by a client.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32   /*
         *NFX_OK or NFX_ERR
         */
nfs_addFileSysCBToList (nfp_fileSysCBPtr_t fileSysCBp,  /*file system control block*/
        fs_Handle *fhandlep,            /*file handle*/
        nfp_clientCBPtr_t clientCBp     /*client control block*/
)
{
    if ( (fileSysCBp == NULL) || (clientCBp == NULL) )
        return (NFX_ERR);

    /*
     *Add file system control block
     */
    nfp_addFileSysCB (fileSysCBp, fhandlep);

    /*
     *Add file system control block into Client Control Block into
     *list active file system list used by a client
     */
    nfs_addCBToList(&clientCBp->nfp_fileSysList, (void *)fileSysCBp, NFP_FILESYSCB_NODE);

    /*
     *Add clientCB into list of active clients that are accesing this
     *file system
     */
    nfs_addCBToList(&fileSysCBp->nfp_clientList, (void *)clientCBp, NFP_CLIENTCB_NODE);
    return (NFX_OK);

}
/*-----------------------------------------------------------------
 * Name:        nfs_delFileSysCBFromList
 *
 * Description:
 *
 *  Delete File System Control Block (FSCB) from the Client Control
 *  Block list.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32 nfs_delFileSysCBFromList(
        nfp_fileSysCBPtr_t fileSysCBp,      /*fileSysCB to be removed*/
        nfp_clientCBPtr_t clientCBp     /*Client Control Block*/
)
{

    if ( (fileSysCBp == NULL) || clientCBp == NULL)
        return (NFX_ERR);

    /*
     *Remove fileSysCB from list in Client Control Block
     */
    nfs_delCBFromList(&clientCBp->nfp_fileSysList,
            (void *)fileSysCBp, NFP_FILESYSCB_NODE);
#ifdef  NOTUSED
    /* 
     *Remove clientCB from list in the file system control block
     */
    nfs_delCBFromList(&fileSysCBp->nfp_clientList,
            (void *)clientCBp, NFP_CLIENTCB_NODE);
#endif
    return (NFX_OK);

}

/*-----------------------------------------------------------------
 * Name:        nfs_addFileCBToList
 *
 * Description:
 *  Add File Control Block (fileCB) into skip lists, open files
 * by a client (clientCB) and open files under its parent file system (fileSysCB).
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfs_addFileCBToList(nfp_fileCBPtr_t fileCBp, fs_Handle *fhandlep,
            nfp_clientCBPtr_t clientCBp)
{
    if  ((fileCBp == NULL) || (clientCBp == NULL) )
        return (NFX_ERR);

    if (nfp_addFileCB(fileCBp, fhandlep) != NFX_OK)
    {
        return (NFX_ERR);
    }
    /*
     *Add fileCB into open files list by a client (ClientCB)
     */
    nfs_addCBToList(&clientCBp->nfp_fileList, 
					(void *)fileCBp, NFP_FILECB_NODE);


    return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:        nfsDelAllFileCBFromDir
 *
 * Description:
 *  Delete all File Control Blocks from the input directory.
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32 nfsDelAllFileCBFromDir(nfp_list_t *fileListp
)
{
    nfp_nodePtr_t           delNodep;
    nfp_fileCBPtr_t         fileCBp;
    int32                   count;

	if (fileListp == NULL)
		return (NFX_ERR);
    //fileListp = &dirFileCBp->nfp_dirFileList;
    count = fileListp->nfp_reqCnt;
    for (delNodep = fileListp->nfp_headp;  
		((delNodep != NULL) && (count > 0));
        delNodep = delNodep->nfp_nextp, count--)
    {
        fileCBp = (nfp_fileCBPtr_t)delNodep->nfp_entryCBp;
        if (fileCBp == NULL)
            continue;
        nfs_delNode(fileListp, delNodep);
    }

    return (NFX_OK);
}
/*-----------------------------------------------------------------
 * Name:        nfs_delAllClientCBFromFile
 *
 * Description:
 *      Delete all Client Control Blocks from input File Control Block.
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32 nfs_delAllClientCBFromFile(nfp_fileCBPtr_t fileCBp        /*File control block*/
)
{
    int32                   count;
    nfp_nodePtr_t           delNodep;
    nfp_clientCBPtr_t       clientCBp;
    nfp_list_t              *clientListp;

    if (fileCBp == NULL)
        return (NFX_ERR);

    clientListp = &fileCBp->nfp_clientList;
    count = clientListp->nfp_reqCnt;
    for (delNodep = clientListp->nfp_headp; ((delNodep != NULL) && (count > 0));
            delNodep = delNodep->nfp_nextp, count--)
    {
        clientCBp = (nfp_clientCBPtr_t)delNodep->nfp_entryCBp;
        if (clientCBp == NULL)
            continue;
        nfs_delNode(clientListp, delNodep);
    }
    return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:        nfs_delAllClientCBFromFileSys
 *
 * Description:
 *      Delete all Client Control Blocks from input File System Control Block.
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfs_delAllClientCBFromFileSys(nfp_fileSysCBPtr_t fileSysCBp)
{
    int32                   count;
    nfp_nodePtr_t           delNodep;
    nfp_clientCBPtr_t       clientCBp;
    nfp_list_t              *clientListp;

    if (fileSysCBp == NULL)
        return (NFX_ERR);

    clientListp = &fileSysCBp->nfp_clientList;
    count = clientListp->nfp_reqCnt;
    for (delNodep = clientListp->nfp_headp ; ((delNodep != NULL) && (count > 0));
        delNodep = delNodep->nfp_nextp, count--)
    {
        clientCBp  = (nfp_clientCBPtr_t)delNodep->nfp_entryCBp;
        if (clientCBp == NULL)
            continue;
        nfs_delNode(clientListp, delNodep);
    }

    return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:        nfs_delAllFileCBFromFileSys
 *
 * Description:
 *      Delete all File Control Blocks from input File System Control Block.
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfs_delAllFileCBFromFileSys(nfp_fileSysCBPtr_t fileSysCBp)
{

    int32                   count;
    nfp_nodePtr_t           delNodep;
    nfp_fileCBPtr_t         fileCBp;
    nfp_list_t              *fileListp;

    if (fileSysCBp == NULL)
        return (NFX_ERR);

    fileListp = &fileSysCBp->nfp_fileList;
    count = fileListp->nfp_reqCnt;

    for (delNodep = fileListp->nfp_headp ; ((delNodep != NULL) && (count > 0));
        delNodep = delNodep->nfp_nextp, count--)
    {
        fileCBp = (nfp_fileCBPtr_t)delNodep->nfp_entryCBp;
        if (fileCBp == NULL)
            continue;
        nfs_delNode(fileListp, delNodep);
    }

    return (NFX_OK);

}
/*-----------------------------------------------------------------
 * Name:        nfs_delFileCBFromList
 *
 * Description:
 *  This function should be call when the file is remove to delete
 *  the File Control Block (fileCB) from clientCB, file system and
 *  file list.
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32 nfs_delFileCBFromList(
        nfp_fileCBPtr_t parentDirFileCBp,   /*parent dir fileCB*/
        nfp_fileCBPtr_t fileCBp,            /*fileCB to be removed*/
        nfp_clientCBPtr_t clientCBp         /*Client Control Block*/
)
{
    fs_Handle   *fhandlep;
	nfs_fileInfo_t          *nfs_fileInfop;

    if  (fileCBp == NULL) 
        return (NFX_ERR);

    fhandlep = &fileCBp->nfp_fileHandle;
    /*
     *Removes all clientCB from the list in the File Control Block
     *De-alloc node(s) that were previously allocated
     */
#ifdef	NOTUSED
    if (fileCBp->nfp_clientList.nfp_reqCnt != 0)
	{
		//printf("nfs_delFileCBFromList 1\n");
        nfs_delAllClientCBFromFile(fileCBp);
	}
#endif	//NOTUSED
    /*
     *delete all fileCB  from file list if type is directory
     *De-alloc node(s) that were previously allocated
     */
    if (fileCBp->nfp_dirFileList.nfp_reqCnt != 0)
	{
		//printf("nfs_delFileCBFromList 2\n");
        nfsDelAllFileCBFromDir(&fileCBp->nfp_dirFileList);
	}
    if (fileCBp->nfp_dirCacheFileList.nfp_reqCnt != 0)
	{
		//printf("nfs_delFileCBFromList 2.5\n");
        nfsDelAllFileCBFromDir(&fileCBp->nfp_dirCacheFileList);
	}
	
    /*
     *Remove fileCB from file list in ClientCB
     */
#ifdef	NOTUSED
	if (clientCBp)
		nfs_delCBFromList(&clientCBp->nfp_fileList,
            (void *)fileCBp, NFP_FILECB_NODE);
#endif
	nfs_fileInfop = &fileCBp->protocol.nfp_nfsFileInfo;
    if (nfs_fileInfop && nfs_fileInfop->nfs_namep != NULL)
    {
#ifdef  NFS_USE_SHMEM
        eee_deallocateBuffer(nfs_fileInfop->nfs_namep);
#else
        eee_ramDealloc((uchar8 *)nfs_fileInfop->nfs_namep);
#endif
		nfs_fileInfop->nfs_nameLen = 0;
		nfs_fileInfop->nfs_namep = NULL;		

    }
	if (nfs_fileInfop && nfs_fileInfop->nfs_fileNamep)
	{
#ifdef  NFS_USE_SHMEM
		eee_deallocateBuffer(nfs_fileInfop->nfs_fileNamep);
#else
		eee_ramDealloc((uchar8 *)nfs_fileInfop->nfs_fileNamep);
#endif
		nfs_fileInfop->nfs_filenameLen = 0;
		nfs_fileInfop->nfs_fileNamep = NULL;		
	}
	if (nfp_validateFileCB(parentDirFileCBp) == NFX_OK)
	{
    	if (parentDirFileCBp->nfp_dirCacheFileList.nfp_reqCnt != 0)
    	{
			//printf("nfs_delFileCBFromList 4\n");
        	nfs_delCBFromList(&parentDirFileCBp->nfp_dirCacheFileList,
                                (void *)fileCBp, NFP_FILECB_NODE);
    	}

		if (parentDirFileCBp->nfp_dirFileList.nfp_reqCnt != 0)
		{
			//printf("nfs_delFileCBFromList 5\n");
			nfs_delCBFromList(&parentDirFileCBp->nfp_dirFileList,
					(void *)fileCBp, NFP_FILECB_NODE);
		}
	}
    /*
     *-Remove fileCB from list
     *-Deallocate File Control Block
     */
    if (nfp_delFileCB (&fileCBp->nfp_fileHandle) != NFX_OK)
    {
        return (NFX_ERR);
    }
    return (NFX_OK);

}

/*-----------------------------------------------------------------
 * Name:        nfs_delFileCB
 *
 * Description:
 *      Delete and de-allocate all memory associate with this File
 *      Control BLock (FileCB) list.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32   /*
         *NFX_OK or NFX_ERR
         */
nfs_delFileCB(nfp_list_t *fileListp)
{
    nfp_nodePtr_t           currp;
    nfp_fileCBPtr_t         fileCBp;
    int32                   count;


    if (fileListp->nfp_headp == NULL)
        return (NFX_ERR);
    count = fileListp->nfp_reqCnt;

    for (currp = fileListp->nfp_headp; ((currp != NULL) && (count > 0));
                    currp = currp->nfp_nextp, count--)
    {
        fileCBp = (nfp_fileCBPtr_t)currp->nfp_entryCBp;
        if (fileCBp == NULL)
            continue;
        if (fileCBp->protocol.nfp_nfsFileInfo.nfs_namep != NULL)
        {
#ifdef  NFS_USE_SHMEM
            eee_deallocateBuffer(fileCBp->protocol.nfp_nfsFileInfo.nfs_namep);
#else
            eee_ramDealloc((uchar8 *)fileCBp->protocol.nfp_nfsFileInfo.nfs_namep);
#endif
			fileCBp->protocol.nfp_nfsFileInfo.nfs_namep = NULL;
			fileCBp->protocol.nfp_nfsFileInfo.nfs_nameLen = 0;
        }
        nfp_delFileCB (&fileCBp->nfp_fileHandle);
    }
    return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:        nfs_delReqCB
 *
 * Description:
 *      Delete and de-allocate all memory associate with this Request
 *  Control BLock (RCB) list.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32   /*
         *NFX_OK or NFX_ERR
         */
nfs_delReqCB(nfp_list_t *reqListp)
{
    nfp_nodePtr_t                   currp;
    uint32                          count;
    nfs_request_t                   *nfsReqp;

    if (reqListp->nfp_headp == NULL)
        return (NFX_ERR);
    count = reqListp->nfp_reqCnt;

    for (currp = reqListp->nfp_headp; 
		 ((currp != NULL) && (count > 0));
         currp = currp->nfp_nextp, count--)
    {
        nfsReqp = (nfs_request_t *)currp->nfp_entryCBp;
        if (nfsReqp == NULL)
            continue;

		/*
		 * Shaun 04/03/02
		 * Bug #362, #419
		 *
		 * Only delete completed requests.
		 * Pending request will be deleted
		 * upon tx complete.
		 * 
		 */
		if (nfsReqp->reqState != NFS_REQ_DONE)
            continue;

        /*
         *Free all memory corresponding to this request
         */
        nfs_msgFreeAll(nfsReqp, TRUE);
    }
    return (NFX_OK);
}
#endif	//NFP_TXRX
