/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:        nfsv2-proc.c
 *
 * RCS:  $Id: nfsv2-proc.c,v 1.5 2002/10/04 01:29:25 shaun Exp $
 *
 * Copyright (C) 2002, Agile Storage Inc.
 *
 * Description: Contains NFS V2 Procedures.
 *
 * Created by: Joseph Vu 
 *
 * Date Created:	5/28/01
 *
 * History: 8/27/02 - Shaun splitted nfs-proc.c into nfsv3-proc.c
 *                    nfsv2-proc.c & nfs-acl.c
 *
 *-----------------------------------------------------------------
 */
#ifdef NFP_TXRX

#include <nfx-incl.h>
#include <exports/nfs-api.h>
#include "nfsv2-proc.h"


#define	NFS_DBG_COD 	1


pkt_desc_t *nfs_tplAllocPkt( nfp_clientCBPtr_t clientCBp, 
							nfs_request_t 		*nfsReqp,
							int32				usrDataLen);


/*-----------------------------------------------------------------
 * Name:	nfsv2_procRxNullReq 
 *
 * Description: Handles NFS V2 NULL Requests.
 *
 *-----------------------------------------------------------------
 */
int32
nfsv2_procRxNullReq(nfs_request_t  *nfsReqp)
{
	uint32				txMsgLen;
	char				*txMsgp;
	nfp_clientCBPtr_t	clientCBp;
	pkt_desc_t 			*tplXmitPktDescp;
    nfs_clientInfo_t    *clientInfop;


#ifdef PKT_USE_TIME_STAMP
    pkt_setTimeStamp(nfsReqp->pTime, PKT_TIME_IDX_FS_REQ);
    pkt_setTimeStamp(nfsReqp->pTime, PKT_TIME_IDX_FS_RSP);
#endif

	nfsReqp->reqType = NFS_NULL;

	/*
	 *Increment NULL receive statistic counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.null++;
	nfs_globalCfgp->nfs_v2RxPktCounters.null++; 

	clientCBp = nfsReqp->clientCBp;
	clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
	clientInfop->nfs_v2RxPktCounters.null++;


	/*
	 *Alloc TPL packet to send
	 */
	tplXmitPktDescp = (pkt_desc_t *)nfs_tplAllocPkt(clientCBp, nfsReqp, 0);

	if (tplXmitPktDescp == NULL)
	{
		//printf ("nfs_procRxNullReq: Failure to alloc TPL packet!\n");
		nfs_globalCfgp->nfs_debugPktCounters.tpl_noPktRes++;
		return (NFSERR_ACCES);
	}


	txMsgp = (char *)tplXmitPktDescp->pHdr;
	txMsgLen = 0;
	/*
	 *Start with RPC reply header
	 */
	txMsgLen += nfs_msgRpcReply(txMsgp, 
								nfsReqp->rpcCallHdrp->rpc_xid,
                                RPC_REPLY_MSG_ACCEPTED, 
								RPC_REPLY_SUCCESS);
	txMsgp += txMsgLen;

	/*
	 *Set packet desc hdr size
	 */
	tplXmitPktDescp->hdrSize = txMsgLen;
	if (nfs_msgXmtPkt(tplXmitPktDescp, nfsReqp) != NFX_OK)
	{
		//Error:  set reqState to DONE so that
		//nfs_reqTimeoutHandler() will free nfsReqp
		nfsReqp->pktDescp = (void *)tplXmitPktDescp;
		nfsReqp->reqState = NFS_REQ_DONE;
	}

	nfs_globalCfgp->nfs_xmtPktCounters.null++;
	nfs_globalCfgp->nfs_v2XmtPktCounters.null++; 

	clientInfop->nfs_v2XmtPktCounters.null++;

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name: nfsv2_procRxGetAttrReq 
 *
 * Description: Handle NFS V2 Get Attribute Requests.
 *
 *-----------------------------------------------------------------
 */
int32
nfsv2_procRxGetAttrReq(nfs_request_t *nfsReqp)
{
    fs_MsgGetAttrReq    *fsGetAttrReqp;
    nfs_clientInfo_t    *clientInfop;
    nfp_clientCBPtr_t   clientCBp;
	char				*nfs_msgp;

	nfsReqp->reqType = NFS_GETATTR;

	/*
	 *Increment GetAttr receive statistic counter
	 */	
	nfs_globalCfgp->nfs_rxPktCounters.getAttr++;
	nfs_globalCfgp->nfs_v2RxPktCounters.getAttr++;

	/*
	 *Get start of the GetAttr request
	 */
	nfs_msgp = nfsReqp->procStart;


	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_ACCES);

	/*
	 *Map a Getattr procedure to File System Request message.
	 *Send request message to the File System 
	 *to retrieve the attributes 
	 */

    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v2RxPktCounters.getAttr++;

    fsGetAttrReqp = (fs_MsgGetAttrReq  *)nfsReqp->fileMsgp;

    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(	&fsGetAttrReqp->hdr,
            			FS_REQ_GET_ATTR, 
						FS_POST_ATTRS, 
						nfsReqp);

    /*
     *Fill in filehandle of the requested attributes
     */
    fsGetAttrReqp->fileHandle = nfsReqp->fileHandle;

    nfsReqp->fileMsgLen = sizeof(fs_MsgGetAttrReq);

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.getAttr++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;

	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_ACCES);
	}
	
	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name: 		nfsv2_procRxSetAttrReq 
 *
 * Description: Handle NFS V2 Set Attribute Requests.
 *
 *-----------------------------------------------------------------
 */
int32 	
nfsv2_procRxSetAttrReq(nfs_request_t *nfsReqp)
{
    fs_MsgSetAttrReq    *fsSetAttrReqp;
    nfs_clientInfo_t    *clientInfop;
    nfp_clientCBPtr_t   clientCBp;
	char                *nfs_msgp;
	int32				length;
	uint32				atimeType;
	uint32				mtimeType;

	nfsReqp->reqType = NFS_SETATTR;

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_ACCES);

    fsSetAttrReqp = (fs_MsgSetAttrReq *)nfsReqp->fileMsgp;

	/*
	 *Increment SetAttr Rx stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.setAttr++;
	nfs_globalCfgp->nfs_v2RxPktCounters.setAttr++;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;

	/*
	 *Get filehandle length
	 */
	length = NFS_V2_FHANDLE_LEN;

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;


	length = nfs_extractSetattr(nfsReqp->rpcCallHdrp->progVer, 
								nfs_msgp, 
								&fsSetAttrReqp->attr, 
								&atimeType, 
								&mtimeType);
	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Map a SetAttr request to File System message
	 *Send request message to File System for processing
	 */

    clientCBp = (nfp_clientCBPtr_t)nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v2RxPktCounters.setAttr++;


    fsSetAttrReqp->checkCtime.secs = 0;
    fsSetAttrReqp->checkCtime.nsecs = 0;

    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(	&fsSetAttrReqp->hdr, 
						FS_REQ_SET_ATTR, 
						FS_POST_ATTRS, 
						nfsReqp);

    /*
     *The filehandle of file/dir to set attributes
     */
    fsSetAttrReqp->fileHandle = nfsReqp->fileHandle;

    nfsReqp->fileMsgLen = sizeof(fs_MsgSetAttrReq);


	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.setAttr++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_ACCES);
	}
	return (NFS_OK);

}

/*-----------------------------------------------------------------
 * Name: nfsv2_procRxLookupReq 
 *
 * Description: Handle NFS V2 Lookup Requests.
 *
 *-----------------------------------------------------------------
 */
int32
nfsv2_procRxLookupReq (nfs_request_t  *nfsReqp)
{
    fs_MsgLookupReq     *fsLookupReqp;
    nfs_clientInfo_t    *clientInfop;
	nfp_clientCBPtr_t	clientCBp;
	char                *nfs_msgp;
	int32				length;
	ushort16			nameLen;

	nfsReqp->reqType = NFS_LOOKUP;

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_ACCES);

    fsLookupReqp = (fs_MsgLookupReq *)nfsReqp->fileMsgp;

	/*
	 *Increment Lookup Rx stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.lookUp++;
	nfs_globalCfgp->nfs_v2RxPktCounters.lookUp++;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;

	length = NFS_V2_FHANDLE_LEN;

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Extract the filename
	 */
	length = nfs_extractString(nfs_msgp,
								&fsLookupReqp->fileName[0], 
								&nameLen, 
								FS_NAME_MAX);
	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		eee_deallocateBuffer( nfsReqp->fileMsgp );
		return (NFSERR_NAMETOL);
	}

	/*
	 *Map a Lookup request to File System message
	 *Send request message to File System for processing
	 */	

    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v2RxPktCounters.lookUp++;

 
    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(&fsLookupReqp->hdr, 
						FS_REQ_LOOKUP,
                    	FS_POST_ATTRS, 
						nfsReqp);
 
    /*
     *The handle of parent directory.
     */ 
    fsLookupReqp->parentHandle = nfsReqp->fileHandle;

    /*
     *Length of name in bytes
     *Name to lookup
     */
    fsLookupReqp->nameLen = nameLen;
    fsLookupReqp->fileName[nameLen] = '\0';
    
    nfsReqp->fileMsgLen = sizeof(fs_MsgLookupReq) + nameLen + 1;


	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.lookUp++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_ACCES);
	} 
	return (NFS_OK);
}


/*-----------------------------------------------------------------
 * Name: 		nfsv2_procRxReadLinkReq 
 *
 * Description: Handle NFS V2 Readlink requests.
 *
 *-----------------------------------------------------------------
 */
int32
nfsv2_procRxReadLinkReq(nfs_request_t *nfsReqp)
{
    fs_MsgReadLinkReq   *fsReadlinkReqp;
    nfs_clientInfo_t    *clientInfop;
	nfp_clientCBPtr_t	clientCBp;
	char				*nfs_msgp;

	nfsReqp->reqType = NFS_READLINK;

	/*
	 *Increment Readlink receive statistic counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.readLink++;
	nfs_globalCfgp->nfs_v2RxPktCounters.readLink++;

	/*
	 *Get start of the Readlink request
	 */
	nfs_msgp = nfsReqp->procStart;

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_ACCES);

    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v2RxPktCounters.readLink++;

    fsReadlinkReqp = (fs_MsgReadLinkReq *)nfsReqp->fileMsgp;

    /*
     *Fill in File System (FS) request header message
     */
    nfs_msgBuildFSHdr(&fsReadlinkReqp->hdr, FS_REQ_READLINK, 0, nfsReqp);

    /*
     *Filehandle of the requested attributes
     */
    fsReadlinkReqp->fileHandle = nfsReqp->fileHandle;

    nfsReqp->fileMsgLen = sizeof(fs_MsgReadLinkReq);

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.readLink++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;

	/*
	 *Send request to file system
	 */
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_ACCES);
	}

	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name:		nfsv2_procRxReadReq  
 *
 * Description: Handle NFS V2 Read Requests.
 *
 *-----------------------------------------------------------------
 */
int32
nfsv2_procRxReadReq (nfs_request_t  *nfsReqp)
{
    fs_MsgReadReq       *readReqp;
    nfs_clientInfo_t    *clientInfop;
    nfp_clientCBPtr_t   clientCBp;
	int32				length;
	uint64				offset;
    uint32          	v2Offset;
	uint32				numOfBytes;
	char				*nfs_msgp;

	nfsReqp->reqType = NFS_READ;

	/*
	 *Increment Lookup Rx stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.read++;
	nfs_globalCfgp->nfs_v2RxPktCounters.read++;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;
    length = NFS_V2_FHANDLE_LEN;

	/*
	 *Update message pointer
	 */
	nfs_msgp += length;

	/*
	 *Extract position offset within the file at which the read is to begin
	 */
	length = nfs_extractLongVal(nfs_msgp, &v2Offset);
	offset = v2Offset;

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Extract number of bytes of data that are to be read
	 */
	length = nfs_extractLongVal(nfs_msgp, &numOfBytes);

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_ACCES);

	/*
	 *Map a Read request to File System message
	 *Send a request message to File System for processing
	 */

    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v2RxPktCounters.read++;


    readReqp = (fs_MsgReadReq *)nfsReqp->fileMsgp;

    /*
     *Fill in File System request header message
     */
    nfs_msgBuildFSHdr(&readReqp->hdr, FS_REQ_READ, FS_POST_ATTRS, nfsReqp);

    /*
     * Fill in filehandle of file to be read,
     * file offset and number of bytes to read
     */
    readReqp->fileHandle = nfsReqp->fileHandle;
    readReqp->fileOffset = offset;
    readReqp->readLength = numOfBytes;

    nfsReqp->fileMsgLen = sizeof(fs_MsgReadReq);


	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.read++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_ACCES);
	}

	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name:       	nfs_procRxWriteReq 
 *
 * Description:
 *
 * Procedure WRITE data to a file.
 *
 * Created by:  Joseph Vu
 *
 * Date Created:        5/28/01
 *
 *-----------------------------------------------------------------
 */
int32
nfsv2_procRxWriteReq(nfs_request_t  *nfsReqp)
{
    fs_MsgWriteReq      *writeReqp;
    nfs_clientInfo_t    *clientInfop;
    nfp_clientCBPtr_t   clientCBp;

	int32           codIdx;
	int32			length;	
	int32			padSize;	
	uint32			beginOffset;
	uint64			fileOffset;
    uint32          v2FileOffset;
	uint32			numOfBytes;
	char			*nfs_msgp;
	pkt_desc_t 		*descp;

	nfsReqp->reqType = NFS_WRITE;

	/*
	 *Increment Write receive statistic counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.write++;	
	nfs_globalCfgp->nfs_v2RxPktCounters.write++ ;

	/*
	 *Get start of the Write request
	 */
	nfs_msgp = nfsReqp->procStart;
    length = NFS_V2_FHANDLE_LEN;

	nfs_msgp += length;

	length = nfs_extractLongVal(nfs_msgp, &beginOffset);
	nfs_msgp += length;

	length = nfs_extractLongVal(nfs_msgp, &v2FileOffset);
    fileOffset = v2FileOffset;
	nfs_msgp += length;

	length = nfs_extractLongVal(nfs_msgp, &numOfBytes);
	nfs_msgp += length;

    /*
     * length field of opague data
     */
    length = nfs_extractLongVal(nfs_msgp, &numOfBytes);
    nfs_msgp += length;


	descp = (pkt_desc_t      *)nfsReqp->rxPktDesc;
	if (!(descp->flags & PKT_DESC_INUSE))
	{
		printf("%s: bad nfsReqp->rxPktDesc: %p\n", 
				__FUNCTION__, nfsReqp->rxPktDesc);	
		luc_panic();
	}

    /* Shaun 09/12/02  Fix #2180
     * Exclude NFS pad in data send to SAC.
     */
    padSize = (4 - (numOfBytes & 0x03)) & 0x03;

#ifdef SBM_REV2
    nfsReqp->nbrUsedCod = luc_sndGmcData( nfsReqp->rxPktDesc, 
										  nfs_msgp, 
										  numOfBytes, 
										  fileOffset, 
										  padSize, 
										  &nfsReqp->cod[0]);
#else

    nfsReqp->nbrUsedCod = luc_sndGmcData( nfsReqp->rxPktDesc, 
						nfs_msgp, numOfBytes, &nfsReqp->cod[0]);
#endif


	if (nfsReqp->nbrUsedCod == 0)
		return (NFSERR_ACCES);

#ifdef	NFS_DBG_COD

	if (nfs_dbgCod( nfsReqp, numOfBytes) != NFX_OK)
		luc_panic();

#endif	//NFS_DBG_COD

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_ACCES);

	/*
	 *Map a Write request to File System message
	 *Send a request message to File System for processing
	 */

    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v2RxPktCounters.write++;

    writeReqp = (fs_MsgWriteReq  *)nfsReqp->fileMsgp;

    memset(writeReqp, 0, sizeof(fs_MsgWriteReq));

    /*
     *Fill in File System header message
     */
    nfs_msgBuildFSHdr(&writeReqp->hdr, FS_REQ_WRITE, FS_POST_ATTRS, nfsReqp);

    /*
     *Fill in the filehandle for the file to which data to be written
     */
    writeReqp->fileHandle = nfsReqp->fileHandle;

    writeReqp->fileOffset = fileOffset;
    writeReqp->writeLength = numOfBytes;    //Total number of bytes
                                            //write req

    /*
     *Fill in COD for write buffer
     */
    for (codIdx =0; codIdx < nfsReqp->nbrUsedCod; codIdx++)
    {
        writeReqp->cod[codIdx].cod = nfsReqp->cod[codIdx].cod;
        writeReqp->cod[codIdx].offset = nfsReqp->cod[codIdx].offset;
        writeReqp->cod[codIdx].length = nfsReqp->cod[codIdx].length;
    }
    writeReqp->nbrUsedCod = nfsReqp->nbrUsedCod;

    nfsReqp->fileMsgLen = sizeof(fs_MsgWriteReq);


	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.write++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_ACCES);
	}

	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name: 		nfsv2_procRxCreateReq 
 *
 * Description: Handle NFS V2 Create Requests.
 *
 *-----------------------------------------------------------------
 */
int32 
nfsv2_procRxCreateReq (nfs_request_t  *nfsReqp)
{
    fs_MsgCreateReq     *fsCreateReqp;
    nfs_clientInfo_t    *clientInfop;
    nfp_clientCBPtr_t   clientCBp;
	char				*nfs_msgp;
	ushort16			nameLen;
	int32				length;
	uint64				verifier;
	uint32				mode;
	uint32				atimeType;
	uint32				mtimeType;

	nfsReqp->reqType = NFS_CREATE;

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_ACCES);

    fsCreateReqp = (fs_MsgCreateReq *)nfsReqp->fileMsgp;

	/*
	 *Increment Lookup Rx stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.create++;	
	nfs_globalCfgp->nfs_v2RxPktCounters.create++;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;
    length = NFS_V2_FHANDLE_LEN;

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Extract the filename
	 */
	length = nfs_extractString(	nfs_msgp, 
								&fsCreateReqp->fileName[0], 
								&nameLen, 
								FS_NAME_MAX);
	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		eee_deallocateBuffer( nfsReqp->fileMsgp );
		return (NFSERR_NAMETOL);
	}

    fsCreateReqp->nameLen = nameLen;
    fsCreateReqp->fileName[nameLen] = '\0';

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	verifier = 0;
	mode = 0;	
	length = nfs_extractSetattr(nfsReqp->rpcCallHdrp->progVer, 
								nfs_msgp, 
								&fsCreateReqp->u.attr, 
								&atimeType, 
								&mtimeType);


	/*
	 *Map a NFS Create file request to file system control message
	 */

    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v2RxPktCounters.create++;


    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(	&fsCreateReqp->hdr, 
						FS_REQ_CREATE, 
						FS_POST_ATTRS, 
						nfsReqp);

    /*
     *Fill in the parentHandle
     */
    fsCreateReqp->parentHandle = nfsReqp->fileHandle;

    /*
     *Set the mode field (Unchecked, Guarded, or Exclusive)
     */
    fsCreateReqp->flags = mode;

    nfsReqp->fileMsgLen = sizeof (fs_MsgCreateReq) + nameLen + 1;



	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.create++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_ACCES);
	}
	
	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name: nfsv2_procRxMkdirReq
 *
 * Description: Handle NFS V2 Mkdir Requests.
 *
 *-----------------------------------------------------------------
 */
int32	
nfsv2_procRxMkdirReq(nfs_request_t  *nfsReqp)
{
    fs_MsgMkdirReq      *fsMkdirReqp;
    nfs_clientInfo_t    *clientInfop;
    nfp_clientCBPtr_t   clientCBp;
	char				*nfs_msgp;
	int32				length;
	ushort16			nameLen;
	uint32				atimeType;
	uint32				mtimeType;
	
	nfsReqp->reqType = NFS_MKDIR;

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_ACCES);

    fsMkdirReqp = (fs_MsgMkdirReq  *)nfsReqp->fileMsgp;

	/*
	 *Increment Lookup Rx stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.mkdir++;	
	nfs_globalCfgp->nfs_v2RxPktCounters.mkdir++ ;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;
    length = NFS_V2_FHANDLE_LEN;

	/*
	 *Update message pointer
	 */
	nfs_msgp += length;

	/*
	 *Extract sub-directory name
	 */
	length = nfs_extractString(nfs_msgp, 
								&fsMkdirReqp->dirName[0], 
								&nameLen, 
								FS_NAME_MAX);
	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		eee_deallocateBuffer( nfsReqp->fileMsgp );
		return (NFSERR_NAMETOL);
	}

    fsMkdirReqp->nameLen = nameLen;
    fsMkdirReqp->dirName[nameLen] = '\0';

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	length = nfs_extractSetattr(nfsReqp->rpcCallHdrp->progVer, 
								nfs_msgp, 
								&fsMkdirReqp->dirAttr, 
								&atimeType, 
								&mtimeType);

	nfs_msgp += length;

	/*
	 *Map a NFS MKDIR request to File System message
	 *Send request message to File System for processing
	 */
    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v2RxPktCounters.mkdir++;

    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(	&fsMkdirReqp->hdr, 
						FS_REQ_MKDIR,
                    	FS_POST_ATTRS, 
						nfsReqp);

    /*
     *Fill in filehandle of parent dir
     */
    fsMkdirReqp->parentHandle = nfsReqp->fileHandle;

    nfsReqp->fileMsgLen = sizeof (fs_MsgMkdirReq) + nameLen + 1;


	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.mkdir++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_ACCES);
	}

	return (NFS_OK);
}


/*-----------------------------------------------------------------
 * Name:		nfsv2_procRxSymlinkReq 
 *
 * Description: Handle NFS V2 Symlink Requests.
 *
 *-----------------------------------------------------------------
 */
int32	
nfsv2_procRxSymlinkReq(nfs_request_t  *nfsReqp)
{
    fs_MsgSymLinkReq    *fsSymlinkReqp;
    char                *fsLinkNamep;
    char                *fsLinkToNamep;
    nfs_clientInfo_t    *clientInfop;
    nfp_clientCBPtr_t   clientCBp;
	char				*nfs_msgp;
	int32				length;
	fs_Attr				fsAttr;
	char				linkName[FS_NAME_MAX];
	ushort16			linkNameLen;
	char				linkText[FS_PATH_MAX];
	ushort16			linkTextLen;
	uint32				atimeType;
	uint32				mtimeType;

	nfsReqp->reqType = NFS_SYMLINK;

	/*
	 *Increment Symlink Rx stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.symlink++;	
	nfs_globalCfgp->nfs_v2RxPktCounters.symlink++;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;
	length = NFS_V2_FHANDLE_LEN;

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Get name string that is to be associated 
	 *with the created symbolic link 
	 */
	length = nfs_extractString(	nfs_msgp,
								&linkName[0], 
								&linkNameLen, 
								FS_NAME_MAX);
	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		return (NFSERR_NAMETOL);
	}
	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;


	/*
	 *Get string of symbolic data 
	 */
	length = nfs_extractString(	nfs_msgp, 
								&linkText[0], 
								&linkTextLen, 
								FS_PATH_MAX);
	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		return (NFSERR_NAMETOL);
	}
	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Get initial attributes of the symbolic link
	 */
	length = nfs_extractSetattr(nfsReqp->rpcCallHdrp->progVer,
								nfs_msgp, 
								&fsAttr, 
								&atimeType, 
								&mtimeType);
	if (length == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_ACCES);
	}

	/* Shaun 08/18/02
	 *
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_ACCES);

	/*
	 *Map a Symlink request to File System message
	 *Send request message to File System for processing
	 */
    clientCBp = (nfp_clientCBPtr_t)nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v2RxPktCounters.symlink++;

    /*
     *Allocate buffer to send request to file system
     */

    fsSymlinkReqp = (fs_MsgSymLinkReq *)nfsReqp->fileMsgp;

    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(&fsSymlinkReqp->hdr, FS_REQ_SYMLINK, 0, nfsReqp);

    /*
     *Fill in filehandle of parent directory for linkName
     */
    fsSymlinkReqp->parentHandle =  nfsReqp->fileHandle;

    /*
     *Set length in bytes
     */
    fsSymlinkReqp->linkNameLen = linkNameLen;
    fsSymlinkReqp->linkToNameLen = linkTextLen;

    /*
     *Fill in linkName & linkTest
     */
    fsLinkNamep = (char *)fsSymlinkReqp->linkName;
    fsLinkToNamep = fsLinkNamep + linkNameLen + 1;


    memcpy(fsLinkNamep, linkName, linkNameLen);
    fsSymlinkReqp->linkName[linkNameLen] = '\0';
    memcpy(fsLinkToNamep, linkText, linkTextLen);
    fsLinkToNamep[linkTextLen] = '\0';

    nfsReqp->fileMsgLen = sizeof(fs_MsgSymLinkReq) + 
							linkNameLen + 1 + linkTextLen + 1;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.symlink++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_ACCES);
	}

	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name: nfsv2_procRxRemoveReq 
 *
 * Description: Handle NFS V2 Remove Requests.
 *
 *-----------------------------------------------------------------
 */
int32
nfsv2_procRxRemoveReq(nfs_request_t  *nfsReqp)
{
    fs_MsgRemoveReq     *fsRemoveReqp;
    nfs_clientInfo_t    *clientInfop;
    nfp_clientCBPtr_t   clientCBp;  
	ushort16			nameLen;
	int32               length;
	char                *nfs_msgp;


	nfsReqp->reqType = NFS_REMOVE;

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_ACCES);

    fsRemoveReqp = (fs_MsgRemoveReq *)nfsReqp->fileMsgp;

	/*
	 *Increment Lookup Rx stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.remove++;;
	nfs_globalCfgp->nfs_v2RxPktCounters.remove++ ;
	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.remove++;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;

    length = NFS_V2_FHANDLE_LEN;

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Extract the filename or sub-dir to be removed
	 */
	length = nfs_extractString(nfs_msgp,
								&fsRemoveReqp->name[0], 
								&nameLen, 
								FS_NAME_MAX);
	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		eee_deallocateBuffer( nfsReqp->fileMsgp );
		return (NFSERR_NAMETOL);
	}

    fsRemoveReqp->nameLen = nameLen;
    fsRemoveReqp->name[nameLen] = '\0';

	/*
	 *Map a Remove file request to File System message
	 *Send a request message to File System for processing
	 */

    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v2RxPktCounters.remove++;


    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(	&fsRemoveReqp->hdr,
            			FS_REQ_REMOVE,
			    		FS_POST_ATTRS,
           			 	nfsReqp);
 
    /*
     *Fill in filehandle for the directory from
     *which entry is to be removed
     */
    fsRemoveReqp->parentHandle = nfsReqp->fileHandle;

    nfsReqp->fileMsgLen = sizeof(fs_MsgRemoveReq) + nameLen + 1;

	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_ACCES);
	}

	return (NFS_OK);
}


/*-----------------------------------------------------------------
 * Name: nfsv2_procRxRmdirReq 
 *
 * Description: Handle NFS V2 Rmdir Requests.
 *
 *-----------------------------------------------------------------
 */
int32
nfsv2_procRxRmdirReq(nfs_request_t  *nfsReqp)
{
    fs_MsgRmdirReq      *fsRmdirReqp;
    nfs_clientInfo_t    *clientInfop;
    nfp_clientCBPtr_t   clientCBp;
	ushort16			nameLen;
	int32               length;
	char                *nfs_msgp;


	nfsReqp->reqType = NFS_RMDIR;

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_ACCES);

    fsRmdirReqp = (fs_MsgRmdirReq *)nfsReqp->fileMsgp;

	/*
	 *Increment Lookup Rx stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.rmdir++;	
	nfs_globalCfgp->nfs_v2RxPktCounters.rmdir++ ;
	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.rmdir++;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;

    length = NFS_V2_FHANDLE_LEN;

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Extract the filename or sub-dir to be removed
	 */
	length = nfs_extractString(nfs_msgp,
								&fsRmdirReqp->name[0], 
								&nameLen, 
								FS_NAME_MAX);
	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		return (NFSERR_NAMETOL);
	}

    fsRmdirReqp->nameLen = nameLen;
    fsRmdirReqp->name[nameLen] = '\0';

    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v2RxPktCounters.rmdir++;
   

    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(	&fsRmdirReqp->hdr,
            			FS_REQ_RMDIR, 
						FS_POST_PAR_ATTRS, 
						nfsReqp);

    /*
     *Fill in filehandle for the directory from which entry is to be removed
     */
    fsRmdirReqp->parentHandle = nfsReqp->fileHandle;


    nfsReqp->fileMsgLen = sizeof(fs_MsgRmdirReq) + nameLen + 1;


	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_ACCES);
	}


	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:		nfsv2_procRxRenameReq 
 *
 * Description: Handle NFS V2 Rename Requests.
 *
 *-----------------------------------------------------------------
 */
int32
nfsv2_procRxRenameReq(nfs_request_t  *nfsReqp)
{
    fs_MsgRenameReq     *fsRenameReqp;
    nfs_clientInfo_t    *clientInfop;
    nfp_clientCBPtr_t   clientCBp;
    char                *fsFromNamep;
    char                *fsToNamep;
	char				*nfs_msgp;
	char				fromName[FS_NAME_MAX];
	ushort16			fromNameLen;
	char				toName[FS_NAME_MAX];
	ushort16			toNameLen;
	int32				length;


	nfsReqp->reqType = NFS_RENAME;

	nfs_globalCfgp->nfs_rxPktCounters.rename++;;
	nfs_globalCfgp->nfs_v2RxPktCounters.rename++;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;
    length = NFS_V2_FHANDLE_LEN;

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Extract filename to be renamed
	 */
	length = nfs_extractString(	nfs_msgp,
								&fromName[0], 
								&fromNameLen, 
								FS_NAME_MAX);
	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		return (NFSERR_NAMETOL);
	}
	nfs_msgp += length;

	/*
	 *Get  filehandle of the destination (todir) directory
	 */
	length = nfs_extractFhandle(nfsReqp->rpcCallHdrp->progVer,
								nfs_msgp, 
								&nfsReqp->toFileHandle);

	if (length == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errStaleFhandle++;
		return (NFSERR_STALE);
	}
	nfs_msgp += length;

	/*
	 *Extract new name of the object (toname)
	 */
	length = nfs_extractString( nfs_msgp, 
								&toName[0],  
								&toNameLen,  
								FS_NAME_MAX);
	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		return (NFSERR_NAMETOL);
	}


	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_ACCES);

	/*
	 *Map a rename request to file system control message
	 *Send a request to file system for proecessing
	 */

    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v2RxPktCounters.rename++;
    
    fsRenameReqp = (fs_MsgRenameReq *)nfsReqp->fileMsgp;

    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(	&fsRenameReqp->hdr, 
						FS_REQ_RENAME,
                        FS_POST_ATTRS, 
						nfsReqp);
    /*
     *Fill in the from directory filehandle
     */
    fsRenameReqp->fromDirHandle = nfsReqp->fileHandle;
    
    /*
     *Fill in the to directory filehandle
     */
    fsRenameReqp->toDirHandle = nfsReqp->toFileHandle;
    
    fsRenameReqp->fromNameLen = fromNameLen;
    fsRenameReqp->toNameLen = toNameLen;
    
    /*
     *Fill in fromName and toName
     */
    fsFromNamep = (char *)fsRenameReqp->fromName;
    fsToNamep = fsFromNamep + fromNameLen + 1;      

    memcpy(fsFromNamep, fromName, fromNameLen);
    fsRenameReqp->fromName[fromNameLen] = '\0';
    memcpy(fsToNamep, toName, toNameLen);
    fsToNamep[toNameLen] = '\0';


    nfsReqp->fileMsgLen = sizeof(fs_MsgRenameReq) + fromNameLen + 1
                                                  + toNameLen + 1 ;


	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.rename++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_ACCES);
	}
	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name:		nfsv2_procRxLinkReq 
 * 
 * Description: Handle NFS V2 Link Requests.
 *
 *-----------------------------------------------------------------
 */
int32	
nfsv2_procRxLinkReq(nfs_request_t  *nfsReqp)
{
    fs_MsgHardLinkReq   *fsHardLinkReqp;
    nfs_clientInfo_t    *clientInfop;
    nfp_clientCBPtr_t   clientCBp;
	char                *nfs_msgp;
	int32               length;
	ushort16            linkNameLen;

	nfsReqp->reqType = NFS_LINK;

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_ACCES);

    fsHardLinkReqp = (fs_MsgHardLinkReq *)nfsReqp->fileMsgp;

	/*
	 *Increment Link procedure Rx stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.link++;
	nfs_globalCfgp->nfs_v2RxPktCounters.link++;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;
	length = NFS_V2_FHANDLE_LEN;

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Get filehandle of for the directory in 
	 *which the link is to be created
	 */
	length = nfs_extractFhandle(nfsReqp->rpcCallHdrp->progVer,
								nfs_msgp, 
								&nfsReqp->toFileHandle);
	if (length == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errStaleFhandle++;
		return (NFSERR_STALE);
	}

	/*
	 *Update message pointer & message length
	 */
	nfs_msgp += length;

	/*
	 *Get the name to be associated with the created link
	 */
	length = nfs_extractString( nfs_msgp,
    							&fsHardLinkReqp->linkName[0],
								&linkNameLen, 
								FS_NAME_MAX);
	if (length < 0)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errNameTooLong++;
		return (NFSERR_NAMETOL);
	}
    fsHardLinkReqp->linkName[linkNameLen] = '\0';
    fsHardLinkReqp->nameLen = linkNameLen;

	/*
	 *Map a NFS LINK request to to file system message.
	 *Send request message to file system
	 */
    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v2RxPktCounters.link++;
    
    
    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(	&fsHardLinkReqp->hdr, 
						FS_REQ_HARDLINK,
                        0, 
						nfsReqp);
    /*
     *Fill in filehandle for file to link to
     */
    fsHardLinkReqp->fileHandle = nfsReqp->fileHandle;

    /*
     *Fill in filehandle for the parent dir for link
     */
    fsHardLinkReqp->parentDirHandle = nfsReqp->toFileHandle;

    nfsReqp->fileMsgLen = sizeof(fs_MsgHardLinkReq) + linkNameLen + 1;

	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.link++;
	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;

	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_ACCES);
	}

	return (NFS_OK);
}

/*-----------------------------------------------------------------
 * Name: 		nfsv2_procRxReaddirReq 
 *
 * Description: Handle NFS V2 Readdir Requests.
 *
 *-----------------------------------------------------------------
 */
int32
nfsv2_procRxReaddirReq(nfs_request_t  *nfsReqp)
{
    fs_MsgReadDirReq        *fsReaddirReqp;
    nfs_clientInfo_t        *clientInfop;
    nfp_clientCBPtr_t       clientCBp;
	int32					length;
	uint64					cookie;
	uint64					cookieVerf;
	uint32					replyMaxSize;
	char					*nfs_msgp;

	nfsReqp->reqType = NFS_READDIR;

	/*
	 *Increment Lookup Rx stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.readdir++;
	nfs_globalCfgp->nfs_v2RxPktCounters.readdir++ ;
	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.readdir++;


	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;
    length = NFS_V2_FHANDLE_LEN;

	/*
	 *Update message pointer
	 */
	nfs_msgp += length;

	/*
	 *Get cookie & cookie verifier.
	 *This should be set 0 on first read request.  
	 *On subsequent requests,
	 *it should be value as returned by the server
	 */

	length = nfs_extractReaddirCookie(	nfsReqp->rpcCallHdrp->progVer, 
										nfs_msgp, 
										&cookie, 
										&cookieVerf);
	if (length == NFX_ERR)
	{
		return (NFSERR_ACCES);
	}
			
	nfs_msgp += length;

	/*
	 *Get count field, the max size of the 
	 *reply arguments, in bytes.
	 */
	length = nfs_extractLongVal(nfs_msgp, &replyMaxSize);

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_ACCES);

	/*
	 *Map a NFS READDIR request to file system control message
	 */

    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v2RxPktCounters.readdir++;

    fsReaddirReqp = (fs_MsgReadDirReq *)nfsReqp->fileMsgp;

    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(&fsReaddirReqp->hdr, 
						FS_REQ_READDIR, 
						FS_POST_ATTRS, 
						nfsReqp);

    /*
     *Fill in filehandle of the directory
     */
    fsReaddirReqp->fileHandle = nfsReqp->fileHandle;

    /*
     *Fill in cookie & cookie verifier
     */
    fsReaddirReqp->cookie = cookie;
    fsReaddirReqp->cookieVerif =  cookieVerf;

    /*
     *Fill in count, max size of the reply arguments, in bytes
     */
    fsReaddirReqp->maxBytes = replyMaxSize;

    nfsReqp->fileMsgLen = sizeof(fs_MsgReadDirReq);

	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_ACCES);
	}

	return (NFS_OK);
}


/*-----------------------------------------------------------------
 * Name:       	nfsv2_procRxStatfsReq 
 *
 * Description: Handle NFS V2 StatFS Requests.
 *
 *-----------------------------------------------------------------
 */
int32 
nfsv2_procRxStatfsReq(nfs_request_t  *nfsReqp)
{
    fs_MsgStatReq       *fsStatfsReqp;
    nfs_clientInfo_t    *clientInfop;
    nfp_clientCBPtr_t   clientCBp;
	int32				length;
	char				*nfs_msgp;


	nfsReqp->reqType = NFS_STATFS;

	/*
	 *Increment Rx Stat counter
	 */
	nfs_globalCfgp->nfs_rxPktCounters.statfs++;	
	nfs_globalCfgp->nfs_v2RxPktCounters.statfs++ ;

	/*
	 *Get start of the request message
	 */
	nfs_msgp = nfsReqp->procStart;
	length = NFS_V2_FHANDLE_LEN;

	/* 
 	 * Allocate msg buff to FS.
 	 */
	nfsReqp->fileMsgp =  eee_allocateBuffer(EEE_SHARED_POOL, 
											EEE_BUF_LARGE, 
											NFX_COS_HIGH);
	if (nfsReqp->fileMsgp == NULL)
		return (NFSERR_ACCES);

	/*
	 * Map a STATFS request to file system 
	 */
    clientCBp = nfsReqp->clientCBp;
    clientInfop = &clientCBp->protocol.nfp_nfsClientInfo;
    clientInfop->nfs_v2RxPktCounters.statfs++;

    fsStatfsReqp = (fs_MsgStatReq *)nfsReqp->fileMsgp;

    /*
     *Fill in File System (FS) request header msg
     */
    nfs_msgBuildFSHdr(	&fsStatfsReqp->hdr,
                    	FS_REQ_STAT, 
						0, 
						nfsReqp);
    /*
     *Fill in filehandle of the requested attributes
     */
    fsStatfsReqp->fsHandle = nfsReqp->fileHandle;

    /*
     *Store our reqCB for FS to return in response message
     */
    nfsReqp->fileMsgLen = sizeof(fs_MsgStatReq);


	nfs_globalCfgp->nfs_debugPktCounters.nfs_PktFsXmtCounters.statfs++;

	nfs_globalCfgp->nfs_totalNumOfReqXmtToFs++;
	if (nfs_msgSendToFs(nfsReqp) == NFX_ERR)
	{
		nfs_globalCfgp->nfs_statCounters.nfs_errGeneric++;
		return (NFSERR_ACCES);
	}

	return (NFS_OK);	
}

int32
nfsv2_procRxNoOpReq(nfs_request_t  *nfsReqp)
{

    return (NFX_OK);
}



#endif  //NFP_TXRX
