/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:			nsm-init.c
 *
 * RCS:  $Id: nsm-init.c,v 1.2 2002/10/24 21:12:45 isabellec Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:
 *	Contain Network Status Monitor (NSM) initialize functions.
 *
 * Created by:		Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 *
 *-----------------------------------------------------------------
 */
#ifdef NFP_TXRX

#include "nsm.h"

nsm_globalCfg_t *nsm_globalCfgp;		//NSM global cfg
uint32			nsm_traceEnable = 0;
boolean			nsm_firstMsgToThread = TRUE;

/*-----------------------------------------------------------------
 * Name:		nsm_initSrcDstPorts
 *
 * Description: NSM initialize source and destination ports to
 *              other modules.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 *
 *-----------------------------------------------------------------
 */
void
nsm_initSrcDstPorts()
{
	int32   eee_evmAppId, eee_efsAppId;

	//------------------------------------------------
	// Set management and control source ports for NSM
	//------------------------------------------------
	nsm_globalCfgp->mgmtSrcPort = 
		PACK_UNICAST_PORT_ADDRESS(
						nfxMySlotId,
						nfxMyCPUId,
						nsm_globalCfgp->eee_nsmAppId,
						EEE_MGMT,
						NFX_COS_HIGH,
						0);

	nsm_globalCfgp->ctrlSrcPort = 
		PACK_UNICAST_PORT_ADDRESS(
						nfxMySlotId,
						nfxMyCPUId,
						nsm_globalCfgp->eee_nsmAppId,
						EEE_CONTROL,
						NFX_COS_HIGH,
						0);

	//-------------------------------------
	// Set destinattion port to file system
	//-------------------------------------

	// Get file system App Id
	eee_efsAppId = eee_getApplicationId("efs");
	if (eee_efsAppId == EEE_ERR_INVALID_APP_NAME)
		panic("%s: Failure to get EFS Application ID!\n", __FUNCTION__);

	// Pack destination port
	nsm_globalCfgp->efsDstPort = 
		PACK_UNICAST_PORT_ADDRESS(
							nfxMySlotId,
							NFX_NFP_FP2,
							eee_efsAppId,
							EEE_CONTROL,
							NFX_COS_HIGH,
							0);
	
	//----------------------------------------
	// Set destinattion port to volume manager
	//----------------------------------------

	// Get file system App Id
	eee_evmAppId = eee_getApplicationId("evm"); 
	if (eee_evmAppId == EEE_ERR_INVALID_APP_NAME)
		panic("%s: Failure to get EVM Application ID!\n", __FUNCTION__);

	// Pack destination port
	nsm_globalCfgp->evmDstPort =
		PACK_UNICAST_PORT_ADDRESS(
	 						nfxSscSlotId,
							NFX_SSC_UNIX,
							eee_evmAppId,
							EEE_MGMT,
							NFX_COS_HIGH,
							PORT_BEST_FIT);

	 return;
} // nsm_initSrcDstPorts()

/*-----------------------------------------------------------------
 * Name:		nsm_init
 *
 * Description: NSM initialize function on the TX/RX CPU.
 *
 * Created by:	Xiaobing Zhang
 *
 * Date Created:	08/01/2002
 *
 *-----------------------------------------------------------------
 */
int32
nsm_init(void)
{
	//-------------------------------------
	// Alloc & Initialize NSM global Config
	//-------------------------------------
	NSM_SHMEM_ALLOC(nsm_globalCfgp, nsm_globalCfg_t, sizeof(nsm_globalCfg_t));
	if (nsm_globalCfgp == NULL)
		panic ("%s: Failure to alloc NSM global Cfg!\n",__FUNCTION__);

	bzero(nsm_globalCfgp, sizeof(nsm_globalCfg_t));

	//-------------------------
	// Bind NSM server with TPL
	//-------------------------
	if (nsm_tplBindReq (0) != NFX_OK)
	{
		panic ("%s: Bind NSM with TPL failed!\n", __FUNCTION__);
	}

	//-------------------------
	// Bind Portmapper with TPL
	//-------------------------
	if (nsm_portMapBindReq (0) != NFX_OK)
	{
		panic ("%s: Bind Portmapper with TPL failed!\n", __FUNCTION__);
	}

	//-----------------------------
	// Register with EEE and app Id
	//-----------------------------
	nsm_globalCfgp->eee_nsmAppId = 
		eee_allocateApplicationId("nsm_srvr");
	if (nsm_globalCfgp->eee_nsmAppId == EEE_ERR_INVALID_APP_NAME)
	{
		panic("%s: Failure to alloc NSM Application ID!\n", __FUNCTION__);
	}

	if (eee_registerReceiveFunc(nsm_globalCfgp->eee_nsmAppId, 
				nsm_msgRcv, NULL) != NFX_OK)
	{
		panic("%s: Register nsm_msgRcv() failed!\n", __FUNCTION__);
	}

	//----------------------------------------
	// Initialize source and destination ports
	//----------------------------------------
	nsm_initSrcDstPorts();

	//------------------------------
	// Initialize client hash tables
	//------------------------------
	nsm_globalCfgp->nsm_clientHash = hash_create(NSM_HASH_CLI_BUCKETS,
												nsm_clientHashFunc,
												nsm_clientHashCompare,
												nsm_clientHashRelease);
	if (nsm_globalCfgp->nsm_clientHash == NULL)
		panic ("%s: Failure to allocate client hash table!\n", __FUNCTION__);

	nsm_globalCfgp->nsm_volumeHash = hash_create(NSM_HASH_VOL_BUCKETS,
												nsm_volumeHashFunc,
												nsm_volumeHashCompare,
												nsm_volumeHashRelease);
	if (nsm_globalCfgp->nsm_volumeHash == NULL)
		panic ("%s: Failure to allocate volume hash table!\n", __FUNCTION__);

#ifdef NOTUSED
	//----------------------------
	// Set and start request timer
	//----------------------------
	nsm_globalCfgp->reqTimeoutHandle = 
			eee_timerCreate(NSM_REQ_TIMEOUT_PERIOD, EEE_TIMER_REPEAT);

	if (nsm_globalCfgp->reqTimeoutHandle == TIMER_ERR_NO_TIMER_AVAILABLE)
		panic ("%s: Failure to create NSM request timer!\n", __FUNCTION__);

	rc = eee_timerStart(nsm_globalCfgp->reqTimeoutHandle,
					(void *)nsm_reqTimeoutHandler, 0, 0, 0, 0);
	if (rc != NFX_OK)
		panic("%s: Failure to start NSM request timer!\n", __FUNCTION__);
#endif //NOTUSED

	//--------------------------------------------------
	// Alloc memory for the read buffer and write buffer
	//--------------------------------------------------
	NSM_SHMEM_ALLOC(nsm_globalCfgp->readBuf.buf, char, NSM_FILEBUF_SIZE); 
	if (nsm_globalCfgp->readBuf.buf == NULL)
		panic("%s: Failure to alloc memory for file buffer!\n", __FUNCTION__);

	NSM_SHMEM_ALLOC(nsm_globalCfgp->writeBuf.buf, char, NSM_FILEBUF_SIZE); 
	if (nsm_globalCfgp->writeBuf.buf == NULL)
		panic("%s: Failure to alloc memory for file buffer!\n", __FUNCTION__);

	//---------------------------------------
	// Create thread to handle NSM processing 
	//---------------------------------------
	nsm_globalCfgp->threadId = thread_create(NULL);
	thread_start(&(nsm_globalCfgp->threadId), nsm_process);

	return (NFX_OK);
} // nsm_init()


#endif	//NFP_TXRX
