/*
 * Name:  nvram.c
 *
 * RCS:  $Id: nvram.c,v 1.24 2002/12/10 22:26:09 bobonfeline Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:  NVRAM module implementation.
 * 
 * Created By:  Bob Fozard
 *              Jan Vanhoutte
 *
 * Date Created:  8/20/2001
 */


#if !defined(SSC)

// to compile  use 
// -D STANDALONE
#ifdef FS_STANDALONE
#define STANDALONE
#endif

#if defined(STANDALONE)
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
// #include <assert.h>
#include <unistd.h>

#endif


#include <nfx-incl.h>
#include "nvram-api.h"
#include "nvram.h"

#if defined(STANDALONE)
#include <exports/eee-desc.h>
#include <exports/eee-app.h>
#include <exports/malloc-api.h>
#endif

#include <exports/elog-api.h>

#if defined(FS_STANDALONE)
#ifdef NFX_SPIN_INIT
#undef NFX_SPIN_INIT
#undef NFX_SPIN_LOCK
#undef NFX_SPIN_UNLOCK
#define NFX_SPIN_INIT(_lockp_,_level_)      pthread_mutex_init(&(_lockp_)->mutex, NULL)
#define NFX_SPIN_LOCK(_lockp_)      pthread_mutex_lock(&(_lockp_)->mutex)
#define NFX_SPIN_UNLOCK(_lockp_)    pthread_mutex_unlock(&(_lockp_)->mutex)
#endif
#endif // FS_STANDALONE

nfx_spinlock_t	nvram_lock;

#if defined(STANDALONE)
int32
nvram_restart_from_file();
#endif

uint32 
nvram_activate_modification (nvram_modification_t *mod_p);


/*
 * internal nvram data
 */

static nvram_header_t *nvram_hdr_p = NULL;


#define NVRAM_GET_SEGMENT_P(_seg_) (nvram_segment_t *)((nvram_segment_t *)(nvram_hdr_p + 1) + _seg_)

#define NVRAM_GET_MOD_P() (nvram_modification_t *)(NVRAM_GET_SEGMENT_P(nvram_hdr_p->segments))

#define NVRAM_GET_BLOCK_P(_block_) (nvram_block_t *)((char *)(NVRAM_GET_MOD_P() + 1) + _block_ * nvram_hdr_p->blockSize)

int genNum;

/*
 * Some definitions for tracing/error/warning
 */

int nvram_appl_id = 1 ;
int cookie = 0 ;
int code = 0 ;

#define NVRAM_TRACE_MNGT     0
#define NVRAM_TRACE_WRITE    1
#define NVRAM_TRACE_READ     2

#define NVRAM_TRACE(_app_type_ , _code_ , _parameters_)                 \
  E_LOG(class_1, LOG_DEBUG, nvram_appl_id, _app_type_, 0, _code_,       \
        _parameters_) ;

#define NVRAM_WARNING(_app_type_ , _code_ , _parameters_)               \
  E_LOG(class_1, LOG_WARNING, nvram_appl_id, _app_type_, 0, _code_,     \
        _parameters_) ;

#define NVRAM_ERROR(_app_type_ , _code_ , _parameters_)                 \
  E_LOG(class_1, LOG_ERR, nvram_appl_id, _app_type_, 0, _code_,         \
        _parameters_) ;


int32
nvram_init_e_log ()
{
#if defined(STANDALONE)
  nvram_appl_id = 24;
#else
  nvram_appl_id = eee_allocateApplicationId("nvram");
  // nvram_appl_id = eee_getAppIdFromName("nvram");
#endif

  NVRAM_TRACE (NVRAM_TRACE_MNGT, 0, ("nvram started"));

  return (NFX_OK);
}



/* 
 * Name: ck_sum
 *
 * Description: calculate/verify the checksum
 *
 *   Input: 
 *   - start ptr (must be int aligned !!)
 *   - end ptr (not used) (must be int aligned !!)
 *   - ptr to cksum
 *   Output 
 *   - if cksum == NULL : just return the result
 *     else  update the check sum
 *   - result == 0 if OK
 *
 *   To verify: call with cksum NULL
 *   To set: init (*cksum) to 0, call the function
 *           with arg *cksum set
 *
 *
 * Created by: Jan Vanhoutte
 *
 * Date Created: 8/20/2001
 *
 */
static int ck_sum (void *start, void *end, int *cksum)
{
  int *st = start;
  int *en = end;
  int sum = 0;
  int result;
  
  while (st != en) {
	sum += *st;
	st++;
  }

  if (cksum == NULL)
	{
	  result = - sum ;
	}
  else
	{
	  result = sum + *cksum;
	  *cksum = - sum ;
	}

  return (result) ;  
}

/*
 * Name: nvram_check_header
 *
 * Description: verify if the header is OK
 *
 * Created by: Jan Vanhoutte
 *
 * Date Created: 8/20/2001
 *
 *
 */
int32
nvram_check_header(nvram_header_t *hdr_p)
{
  if (hdr_p == NULL)
	{
	  NVRAM_WARNING (NVRAM_TRACE_MNGT, 0, ("nvram_check_header: NULL"));	  
	  return (NFX_ERR);
	}
  
  if (hdr_p->magic != NVRAM_HDR_MAGIC)
	{
      // OK at reboot time.
	  NVRAM_TRACE (NVRAM_TRACE_MNGT, 0, ("nvram_check_header: wrong MAGIC"));	  
	  return (NFX_ERR);
	}
  
  if (hdr_p->version != NVRAM_VERSION_ONE)
	{
	  NVRAM_WARNING (NVRAM_TRACE_MNGT, 0, ("nvram_check_header: wrong version"));	  
	  return (NFX_ERR);
	}
  
  if (ck_sum (hdr_p, &(hdr_p->spare), NULL) != 0)
	{
	  NVRAM_WARNING (NVRAM_TRACE_MNGT, 0, ("nvram_check_header: wrong Checksum"));
	  return (NFX_ERR);
	}
  
  return (NFX_OK);
}

/*
 * Name: nvram_check_segment
 *
 * Description: verify if the segment is OK
 *
 * Created by: Jan Vanhoutte
 *
 * Date Created: 8/20/2001
 *
 *
 */
int32
nvram_check_segment(nvram_segment_t *sgm_p)
{
  int nbr ;
  int block;
  nvram_block_t *blk_p ;
  
  if (nvram_hdr_p == NULL)
  {
	  return (NFX_ERR);
  }
  
  if (sgm_p == NULL)
	{
	  NVRAM_WARNING (NVRAM_TRACE_MNGT, 0, ("nvram_check_segment: NULL"));
	  return (NFX_ERR);
	}
  
  if (sgm_p->magic != NVRAM_SEGMENT_MAGIC)
	{
	  NVRAM_WARNING (NVRAM_TRACE_MNGT, 0, ("nvram_check_segment: wrong MAGIC"));
	  return (NFX_ERR);
	}
  
  if (ck_sum (sgm_p, &(sgm_p->spare), NULL) != 0)
	{
	  NVRAM_WARNING (NVRAM_TRACE_MNGT, 0, ("nvram_check_segment: wrong checksum"));	  
	  return (NFX_ERR);
	}
  
  if (sgm_p->numBlks)
	{
	  if (sgm_p->numBlks > nvram_hdr_p->blocks)
		{
		  NVRAM_WARNING (NVRAM_TRACE_MNGT, 0, ("nvram_check_segment: wrong number of blocks %d %d", 
                         sgm_p->numBlks, nvram_hdr_p->blocks));
		  return (NFX_ERR);
		}
	  
	  block = sgm_p->startBlk ;
	  for (nbr = 0 ; nbr < sgm_p->numBlks ; nbr++)
		{
		  blk_p = NVRAM_GET_BLOCK_P(block) ;
		  
		  if (blk_p->magic != NVRAM_BLOCK_MAGIC)
			{
			  NVRAM_WARNING (NVRAM_TRACE_MNGT, 0, ("nvram_check_segment: wrong block MAGIC"));
			  return (NFX_ERR);
			}
		  
		  if (ck_sum (blk_p, &(blk_p->data), NULL) != 0)
			{
			  NVRAM_WARNING (NVRAM_TRACE_MNGT, 0, ("nvram_check_segment: wrong block checksum"));
			  return (NFX_ERR);
			}
		  
		  block = blk_p->next_block;
		}

	  if (block != NVRAM_END_OF_BLOCK)
		{
		  NVRAM_WARNING (NVRAM_TRACE_MNGT, 0, ("nvram_check_segment: wrong end of block"));
		  return (NFX_ERR);
		}
	  
	}

  return (NFX_OK);
}

/*
 * Name: nvram_print_segment
 *
 * Description: print the info for the segment
 *
 * Created by: Jan Vanhoutte
 *
 * Date Created: 8/20/2001
 *
 *
 */
int32
nvram_print_segment(nvram_segment_t *sgm_p)
{
  int nbr ;
  int block;
  nvram_block_t *blk_p ;
  
  if (nvram_hdr_p == NULL)
  {
	  return (NFX_ERR);
  }
  
  if (sgm_p == NULL)
	{
	  NVRAM_WARNING (NVRAM_TRACE_MNGT, 0, ("nvram_check_segment: NULL"));
	  return (NFX_ERR);
	}
  
  if (sgm_p->magic != NVRAM_SEGMENT_MAGIC)
	{
	  NVRAM_WARNING (NVRAM_TRACE_MNGT, 0, ("nvram_check_segment: wrong MAGIC"));
	  return (NFX_ERR);
	}
  
  if (ck_sum (sgm_p, &(sgm_p->spare), NULL) != 0)
	{
	  NVRAM_WARNING (NVRAM_TRACE_MNGT, 0, ("nvram_check_segment: wrong checksum"));	  
	  return (NFX_ERR);
	}
  
  if (sgm_p->numBlks)
	{
      printf ("Number of blocks %d\n", sgm_p->numBlks);
  
	  if (sgm_p->numBlks > nvram_hdr_p->blocks)
		{
		  NVRAM_WARNING (NVRAM_TRACE_MNGT, 0, ("nvram_check_segment: wrong number of blocks %d %d", 
                         sgm_p->numBlks, nvram_hdr_p->blocks));
		  return (NFX_ERR);
		}
	  
	  block = sgm_p->startBlk ;
	  for (nbr = 0 ; nbr < sgm_p->numBlks ; nbr++)
		{
          printf (" %d ", block);
		  blk_p = NVRAM_GET_BLOCK_P(block) ;
		  
		  if (blk_p->magic != NVRAM_BLOCK_MAGIC)
			{
			  NVRAM_WARNING (NVRAM_TRACE_MNGT, 0, ("nvram_check_segment: wrong block MAGIC"));
			  return (NFX_ERR);
			}
		  
		  if (ck_sum (blk_p, &(blk_p->data), NULL) != 0)
			{
			  NVRAM_WARNING (NVRAM_TRACE_MNGT, 0, ("nvram_check_segment: wrong block checksum"));
			  return (NFX_ERR);
			}
		  
		  block = blk_p->next_block;
		}

	  if (block != NVRAM_END_OF_BLOCK)
		{
		  NVRAM_WARNING (NVRAM_TRACE_MNGT, 0, ("nvram_check_segment: wrong end of block"));
		  return (NFX_ERR);
		}
      printf ("   OK\n");
	}

  return (NFX_OK);
}

void
nvram_print_all()
{
  int nbr, rc;
  nvram_segment_t *sgm_p;

  if (nvram_hdr_p == NULL)
  {
	  return ;
  }
  
  for (nbr = 0 ; nbr < nvram_hdr_p->segments ; nbr++)
  {
    printf ("Segment %d ", nbr);
    sgm_p = NVRAM_GET_SEGMENT_P(nbr) ;
    rc = nvram_print_segment (sgm_p);
    if (rc != NFX_OK)
        printf ("Problems in segment %d\n", nbr);
  }
  printf ("\n");
}

int nvram_ever_used = 0;


/*
 * Name: nvram_init
 *
 * Description:  see nvram-api.h
 *
 * Created by: Jan Vanhoutte
 *
 * Date Created: 8/20/2001
 *
 *
 */
int32
nvram_init (void)
{
  void *nvram_p;
  int nbr;
  nvram_segment_t *sgm_p;
  nvram_modification_t *mod_p;

  nvram_init_e_log();

  NFX_SPIN_INIT(&nvram_lock,SPINLOCK_LEVEL_UNKNOWN);

#if defined(STANDALONE)
  genNum = random() & 0x0000FFFF;
  nvram_p = (void*)eee_ramAlloc(NVRAM_NVRAM_SIZE * 1024) ;
#else
  // Todo: genNum = some value random() & 0x0000FFFF;
  genNum = random() & 0x0000FFFF;
  // Todo: nvram_p = begin of the NVRAM ;
  // Find the NVRAM begin address, and format the appropriate memory
  // As long as we don't have a way to look for it,
  //   just set to NULL
  nvram_p = NULL;
#endif

  nvram_hdr_p = (nvram_header_t *)nvram_p ;

#if defined(STANDALONE)
  if (nvram_restart_from_file() != NFX_OK)
	//	goto nvram_reformat;
	; // just for test, keep checking the NVRAM crc
  
#endif

  NFX_SPIN_LOCK(&nvram_lock);

  if (nvram_check_header(nvram_hdr_p) != NFX_OK)
	{

	  NVRAM_TRACE (NVRAM_TRACE_MNGT, 0, ("nvram_init: need to reformat the NVRAM"));

#if defined(STANDALONE)
	  goto nvram_reformat;
#else
	  // Todo: send msg to SSC, asking for the NVRAM config and wait until SSC comes along to init us;
	  // Todo: nvram_hdr_p = NULL ;
	  goto nvram_reformat;
#endif
	}
  else
	{
	  for (nbr = 0 ; nbr < nvram_hdr_p->segments ; nbr++)
		{
		  sgm_p = NVRAM_GET_SEGMENT_P(nbr) ;

		  if (nvram_check_segment(sgm_p) != NFX_OK)
			{
			  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_init: segment %d is NOK", nbr));
			  goto nvram_reformat;
			}
		}
	}

  // Now activate the changes if any !!
  mod_p = NVRAM_GET_MOD_P();
  if ((mod_p->magic == NVRAM_MODIFICATION_MAGIC) &&
      (ck_sum (mod_p, &(mod_p->spare), NULL) == 0))
  {
      if (nvram_activate_modification (mod_p) != NFX_OK)
      {
          NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_allocate: the activate_modification failed: Panic"));
      }
  }
  // Erase Modification record
  bzero (mod_p, sizeof(nvram_modification_t)); 



  NFX_SPIN_UNLOCK(&nvram_lock);

  return NFX_OK;

nvram_reformat:

#if defined(STANDALONE)

  NFX_SPIN_UNLOCK(&nvram_lock);
  nvram_format (NVRAM_MODE_STANDALONE,
                4096 + 40, // NVRAM_NVRAM_SIZE, // in KB // + overhead
                128, // NVRAM_SEGMENTS_MAX,
                128 * 1024); // NVRAM_BLOCK_SIZE);
  NFX_SPIN_LOCK(&nvram_lock);

  return NFX_OK;

#else

  // No need to allocate the memory now, wait for real need  
  // Find the NVRAM begin address, and format the appropriate memomry
  // As long as we don't have a way to look for it,
  //   just return error
  
  return NFX_ERR;

#endif
  
} /* nvram_init */


/*-----------------------------------------------------------------------------
 * nvram_format
 *
 * Description:  see nvram-api.h
 *
 */
int32
nvram_format (
    uint32      mode,
    uint32      nvramSize,
    uint32      segments,
    uint32      blockSize)
{
  void *nvram_p;
  uint64 tmp ;
  int nbr;
  nvram_segment_t *sgm_p;
  uint32 nbr_of_blocks;
  nvram_block_t *blk_p = NULL;

  // This rewrite the complete NVRAM
  // Avoid all access !! 
  // Blocking time is not critical, should only happen at startup
  // If crash during format, should be 99.99 % protected by nvram_hdr_p->magic

  NVRAM_TRACE (NVRAM_TRACE_MNGT, 0, ("nvram_format: mode = %s, Size = %d KB, %d segments, blockSize = %d B",
			   (mode == NVRAM_MODE_CLUSTERED) ? "Clustered":"Standalone", 
			   nvramSize, segments, blockSize));

  if (blockSize < (8 * 1024))
	{
	  NVRAM_TRACE (NVRAM_TRACE_MNGT, 0, ("nvram_format: blockSize = %d, must be min: %d",
										 blockSize, 8 * 1024));
	  return (NFX_ERR);
	}
  
  NFX_SPIN_LOCK(&nvram_lock);

#if defined(STANDALONE)
  if (nvram_hdr_p)
	eee_ramDealloc (nvram_hdr_p);
  nvram_p = (void*)eee_ramAlloc(nvramSize * 1024) ;

#else
  if (nvram_hdr_p)
	eee_ramDealloc (nvram_hdr_p);
  nvram_p = (void*)eee_ramAlloc(nvramSize * 1024) ;

  // Todo: nvram_p = begin of the NVRAM ;
#endif

  nvram_hdr_p = (nvram_header_t *)nvram_p ;

  // if we crash during format, force a reformat, magic == 0
  bzero (nvram_hdr_p, sizeof (nvram_header_t));
  //nvram_hdr_p->magic = 0 ;

  // The first one is used for the free block list
  // Allocate one more
  segments += 1 ;

  blockSize += sizeof(nvram_block_t);
  
  tmp = (uint64)nvramSize * 1024 ;
  tmp = tmp - sizeof (nvram_header_t) - segments * sizeof (nvram_segment_t) - sizeof (nvram_modification_t);
  nbr_of_blocks = (uint32)(tmp / (blockSize + sizeof(nvram_block_t))) ;

  nvram_hdr_p->version = NVRAM_VERSION_ONE ;
  nvram_hdr_p->mode = mode ;
  nvram_hdr_p->nvramSize = nvramSize ;
  nvram_hdr_p->segments = segments ;
  nvram_hdr_p->blockSize = blockSize ;
  nvram_hdr_p->blocks = nbr_of_blocks ;
  nvram_hdr_p->timeStamp = 0 ;  // Todo

  // init all segments to: 'not in use'
  for (nbr = 0 ; nbr < segments ; nbr++)
	{
	  sgm_p = NVRAM_GET_SEGMENT_P(nbr) ;

	  bzero (sgm_p, sizeof (nvram_segment_t));

	  sgm_p->magic = NVRAM_SEGMENT_MAGIC ;
	  sgm_p->clientId = NVRAM_CLIENT_NONE ;
	  // sgm_p->cksum = 0 ;
	  ck_sum (sgm_p, &(sgm_p->spare), &(sgm_p->cksum));
	  // NVRAM_TRACE (NVRAM_TRACE_MNGT, 0, ("FORMAT sgm %3d 0x%8x", nbr, (int)sgm_p));
	}

  // init all blocks, and use the first segment as the free block list
  for (nbr = 0 ; nbr < nbr_of_blocks ; nbr++)
	{
	  blk_p = NVRAM_GET_BLOCK_P(nbr) ;
	  
	  bzero (blk_p, sizeof (nvram_block_t));
	  // -- or -- bzero (blk_p, blockSize);

	  blk_p->magic = NVRAM_BLOCK_MAGIC ;
	  blk_p->next_block = nbr + 1 ;
	  // blk_p->cksum = 0 ;
	  ck_sum (blk_p, &(blk_p->data), &(blk_p->cksum));
	  // NVRAM_TRACE (NVRAM_TRACE_MNGT, 0, ("FORMAT blk %3d 0x%8x", nbr, (int)blk_p));
	}
  blk_p->next_block = NVRAM_END_OF_BLOCK ;
  blk_p->cksum = 0 ;
  ck_sum (blk_p, &(blk_p->data), &(blk_p->cksum));

  // use the first segment to store the list of free blocks
  sgm_p = NVRAM_GET_SEGMENT_P(0) ;
  sgm_p->clientId = NVRAM_CLIENT_INTERNAL ;
  sgm_p->startBlk = 0;
  sgm_p->numBlks = nbr_of_blocks;
  sgm_p->cksum = 0 ;
  ck_sum (sgm_p, &(sgm_p->spare), &(sgm_p->cksum));

  nvram_hdr_p->cksum = 0 ;
  nvram_hdr_p->magic = NVRAM_HDR_MAGIC ;
  ck_sum (nvram_hdr_p, &(nvram_hdr_p->spare), &(nvram_hdr_p->cksum));
  
  // Just for FUN
  // VERIFY ( ! (ck_sum (nvram_hdr_p, &(nvram_hdr_p->spare), NULL)));

  // Now is the NVRAM dev valid 

  NFX_SPIN_UNLOCK(&nvram_lock);

  return NFX_OK;

} /* nvram_format */


/*
 * Name: nvram_activate_modification
 *
 * Description: Activate the logged modification into the real segments/blocks
 *
 * Created by: Jan Vanhoutte
 *
 * Date Created: 8/20/2001
 *
 *
 */
uint32 
nvram_activate_modification (nvram_modification_t *mod_p)
{
  nvram_segment_t *sgm_p;
  nvram_segment_t *free_sgm_p;
  nvram_block_t *blk_p ;

  // must be locked by caller !!

  NVRAM_TRACE (NVRAM_TRACE_MNGT, 0, ("nvram_activate_modification"));

  if (nvram_check_header(nvram_hdr_p) != NFX_OK)
	{
	  // request REFORMAT ???
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_activate_modification: wrong nvram_hdr_p"));
	  return NFX_ERR;
	}

  if (mod_p->state != NVRAM_MODIFICATION_PROGRESS)
	{
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_activate_modification: wrong modification state"));
	  // invalid modification record
	  bzero (mod_p, sizeof(nvram_modification_t)); 
	  return NFX_ERR;
	}

  if (mod_p->magic != NVRAM_MODIFICATION_MAGIC)
	{
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_activate_modification: wrong modification magic"));
	  // invalid modification record
	  bzero (mod_p, sizeof(nvram_modification_t)); 
	  return NFX_ERR;
	}

  if (ck_sum (mod_p, &(mod_p->spare), NULL) != 0)
	{
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_activate_modification: wrong modification record"));
	  // invalid modification record
	  bzero (mod_p, sizeof(nvram_modification_t)); 
	  return NFX_ERR;
	}

  free_sgm_p = sgm_p = NVRAM_GET_SEGMENT_P(0) ;
  if (sgm_p == NULL)
	{
	  // PANIC 
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_activate_modification: wrong free segment_p"));
	  return NFX_ERR;
	}
  if (free_sgm_p->magic != NVRAM_SEGMENT_MAGIC)
	{
	  // PANIC 
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_activate_modification: wrong free segment_p magic"));
	  return NFX_ERR;
	}
  if (free_sgm_p->clientId != NVRAM_CLIENT_INTERNAL)
	{
	  // PANIC 
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_activate_modification: wrong free segment_p client Id"));
	  return NFX_ERR;
	}
  
  sgm_p = NVRAM_GET_SEGMENT_P(mod_p->segment) ;

  if (sgm_p->magic != NVRAM_SEGMENT_MAGIC)
	{
	  // PANIC 
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_activate_modification: wrong segment_p magic %d", mod_p->segment));
	  return NFX_ERR;
	}

  if (mod_p->block != NVRAM_END_OF_BLOCK)
	{
	  blk_p = NVRAM_GET_BLOCK_P(mod_p->block) ;
	  blk_p->next_block = mod_p->next_block ;
	  blk_p->cksum = 0 ;
	  ck_sum (blk_p, &(blk_p->data), &(blk_p->cksum));
	}
  
  free_sgm_p->numBlks = mod_p->free_sgm_new_numBlks ;
  free_sgm_p->startBlk = mod_p->free_sgm_new_startBlk ;
  // free_sgm_p->timeStamp = ToDo ;
  free_sgm_p->cksum = 0 ;
  ck_sum (free_sgm_p, &(free_sgm_p->spare), &(free_sgm_p->cksum));

  sgm_p->clientId = mod_p->sgm_clientId ;
  sgm_p->clientCookie = mod_p->sgm_clientCookie ;
  sgm_p->startBlk = mod_p->sgm_startBlk ;
  sgm_p->numBlks = mod_p->sgm_numBlks ;
  sgm_p->Handle = mod_p->sgm_Handle ;
  // sgm_p->timeStamp = ToDo ;
  sgm_p->cksum = 0 ;
  ck_sum (sgm_p, &(sgm_p->spare), &(sgm_p->cksum));

  return (NFX_OK);
}


/*-----------------------------------------------------------------------------
 * nvram_allocate
 *
 * Description:  Allocate an NVRAM segment.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
int32
nvram_allocate(
    nvram_clientId_t    clientId,
    nvram_cookie_t      cookie,
    int32               size,
    nvram_handle_t      *handlep)
{
  int nbr ;
  int found = -1 ;
  nvram_segment_t *sgm_p = NULL;
  nvram_segment_t *free_sgm_p;
  int nbr_of_blocks;
  int block;
  nvram_block_t *blk_p ;
  nvram_modification_t *mod_p;
  
  // This does write, in case of crash: TBI
  // Avoid that someone else write to nvram 
  // Max blocking time is loop of segments (512) and shuffle free blocks

  nvram_ever_used = 1;
  
  NVRAM_TRACE (NVRAM_TRACE_MNGT, 0, ("nvram_allocate  %d %lld %d", clientId, cookie, size));

  *handlep = NVRAM_INVALID_HANDLE;
  
  NFX_SPIN_LOCK(&nvram_lock);

  if (nvram_check_header(nvram_hdr_p) != NFX_OK)
	{
	  // request REFORMAT ???
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_allocate: wrong nvram header"));
	  NFX_SPIN_UNLOCK(&nvram_lock);
	  return NFX_ERR;
	}

  nbr_of_blocks = size / (nvram_hdr_p->blockSize - sizeof(nvram_block_t)) ;
  if (nbr_of_blocks == 0)
	nbr_of_blocks = 1;

  free_sgm_p = NVRAM_GET_SEGMENT_P(0) ;
  if (free_sgm_p->clientId != NVRAM_CLIENT_INTERNAL)
	{
	  // PANIC 
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_allocate: wrong free segment clientid"));
	  NFX_SPIN_UNLOCK(&nvram_lock);
	  return NFX_ERR;
	}
  
  // Verify that the clientId and Cookie does not yet exist !!
  for (nbr = 1 ; nbr < nvram_hdr_p->segments ; nbr++)
	{
	  sgm_p = NVRAM_GET_SEGMENT_P(nbr) ;
	  if ((sgm_p->clientId == clientId) &&
		  (sgm_p->clientCookie == cookie) &&
		  (nvram_check_segment(sgm_p) == NFX_OK))
		{
		  // NVRAM segment exist already, return Handle
		  // What if it exist, but wrong cksum ????

		  NVRAM_WARNING (NVRAM_TRACE_MNGT, 0, ("nvram_allocate: clientID and cookie exist already"));

		  if (nbr_of_blocks != sgm_p->numBlks)
			{
			  // What to do ????
			  // Print warning 
			  // Increase size could be done if blocks available
			  // Decrease ?
			  // Or just reject, Reason code ??
			  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_allocate: reallocate with a different size %d %d", 
                           nbr_of_blocks, sgm_p->numBlks));
			  NFX_SPIN_UNLOCK(&nvram_lock);
			  return (NFX_ERR);
			}

		  *handlep = sgm_p->Handle;
		  NFX_SPIN_UNLOCK(&nvram_lock);
		  return (NFX_OK);
		}
	}

  // Find a free segment
  for (nbr = 1 ; nbr < nvram_hdr_p->segments ; nbr++)
	{
	  sgm_p = NVRAM_GET_SEGMENT_P(nbr) ;
	  if ((sgm_p->clientId == NVRAM_CLIENT_NONE) &&
		  (nvram_check_segment(sgm_p) == NFX_OK))
		{
		  // found a valid free segment, now initialize !!
		  found = nbr ;
		  break ;
		}
	}

  if (found == -1)
	{
	  // No more free segments available
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_allocate: no more segments available"));
	  NFX_SPIN_UNLOCK(&nvram_lock);
	  return NFX_ERR;
	}

  nbr_of_blocks = ((size - 1) / (nvram_hdr_p->blockSize - sizeof(nvram_block_t))) + 1 ;

  if (free_sgm_p->numBlks < nbr_of_blocks)
	{
	  // no more free blocks available
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_allocate: no enough free blocks available %d %d", 
                   free_sgm_p->numBlks, nbr_of_blocks));
	  NFX_SPIN_UNLOCK(&nvram_lock);
	  return NFX_ERR;
	}

  block = free_sgm_p->startBlk ;
  blk_p = NVRAM_GET_BLOCK_P(block) ;
  for (nbr = 0 ; nbr < (nbr_of_blocks - 1) ; nbr++)
	{
	  block = blk_p->next_block;
	  blk_p = NVRAM_GET_BLOCK_P(block) ;
	}

  mod_p = NVRAM_GET_MOD_P();

  if (mod_p->state != NVRAM_MODIFICATION_NONE)
	{
	  // already some modification ongoing, should never happpen
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_allocate: modification already ongoing"));
	  NFX_SPIN_UNLOCK(&nvram_lock);
	  return NFX_ERR;
	}
  

  // Everything is ready
  // Store all changes in modification record

  mod_p->block = block;
  mod_p->next_block = NVRAM_END_OF_BLOCK ;

  mod_p->free_sgm_new_numBlks = free_sgm_p->numBlks - nbr_of_blocks ;
  mod_p->free_sgm_new_startBlk = blk_p->next_block ;
  // free_sgm_p->timeStamp = ToDo ;

  mod_p->segment = found ;
  mod_p->sgm_clientId = clientId ;
  mod_p->sgm_clientCookie = cookie ;
  mod_p->sgm_startBlk = free_sgm_p->startBlk ;
  mod_p->sgm_numBlks = nbr_of_blocks ;
  mod_p->sgm_Handle = found + (genNum << 16) ;
  // mod_p->sgm_timeStamp = ToDo ;

  mod_p->magic = NVRAM_MODIFICATION_MAGIC;
  mod_p->state = NVRAM_MODIFICATION_PROGRESS;
  mod_p->cksum = 0;
  ck_sum (mod_p, &(mod_p->spare), &(mod_p->cksum));

  // From here on we can recover from the modification record

  // Now activate all changes
  if (nvram_activate_modification (mod_p) != NFX_OK)
	{
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_allocate: the activate_modification failed: Panic"));
	  bzero (mod_p, sizeof(nvram_modification_t)); 
	  NFX_SPIN_UNLOCK(&nvram_lock);
	  return (NFX_ERR);
	}

  // Changes succesfull done
  // Erase Modification record
  bzero (mod_p, sizeof(nvram_modification_t)); 

  // OK, all changes done succesfully

  *handlep = sgm_p->Handle ;  

  NFX_SPIN_UNLOCK(&nvram_lock);

  return NFX_OK;

} /* nvram_allocate */

/*-----------------------------------------------------------------------------
 * nvram_deallocate
 *
 * Description:  Deallocate an NVRAM segment.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
int32
nvram_deallocate(
    nvram_clientId_t    clientId,
    nvram_handle_t      handlep)
{
  nvram_segment_t *sgm_p;
  nvram_segment_t *free_sgm_p;
  int segment;
  int block;
  int nbr;
  nvram_block_t *blk_p = NULL;
  nvram_modification_t *mod_p;
  
  NVRAM_TRACE (NVRAM_TRACE_MNGT, 0, ("nvram_deallocate"));

  NFX_SPIN_LOCK(&nvram_lock);

  if (nvram_check_header(nvram_hdr_p) != NFX_OK)
	{
	  // request REFORMAT ???
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_deallocate: wrong nvram header"));
	  NFX_SPIN_UNLOCK(&nvram_lock);
	  return NFX_ERR;
	}

  free_sgm_p = NVRAM_GET_SEGMENT_P(0) ;
  if (free_sgm_p->clientId != NVRAM_CLIENT_INTERNAL)
	{
	  // PANIC 
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_deallocate: wrong free segment clientid "));
	  NFX_SPIN_UNLOCK(&nvram_lock);
	  return NFX_ERR;
	}

  segment = NVRAM_SEGMENT_HANDLE(handlep);
  if ((segment < 1) ||
	  (segment >= nvram_hdr_p->segments))
	{
	  // wrong Handle
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_deallocate: wrong Handle %d 0x%x", segment, handlep));
	  NFX_SPIN_UNLOCK(&nvram_lock);
	  return NFX_ERR;
	}
  sgm_p = NVRAM_GET_SEGMENT_P(segment) ;
  if (sgm_p->Handle != handlep)
	{
	  // wrong Handle, stale
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_deallocate: wrong Handle 0x%x 0x%x", sgm_p->Handle, handlep));
	  NFX_SPIN_UNLOCK(&nvram_lock);
	  return NFX_ERR;
	}
  if (sgm_p->clientId != clientId)
	{
	  // Not owner
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_deallocate: wrong clientId %d %d", sgm_p->clientId, clientId));
	  NFX_SPIN_UNLOCK(&nvram_lock);
	  return NFX_ERR;
	}

  if (nvram_check_segment(sgm_p) != NFX_OK)
	{
	  // request REFORMAT ???
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_deallocate: wrong segment check"));
	  NFX_SPIN_UNLOCK(&nvram_lock);
	  return NFX_ERR;	  
	}

  mod_p = NVRAM_GET_MOD_P();
  
  if (mod_p->state != NVRAM_MODIFICATION_NONE)
	{
	  // already some modification ongoing, should never happpen
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_deallocate: modification already ongoing"));
	  NFX_SPIN_UNLOCK(&nvram_lock);
	  return NFX_ERR;
	}

  // go to last block of free list
  block = free_sgm_p->startBlk ;
  if (free_sgm_p->numBlks > 0)
	{
	  for (nbr = 0 ; nbr < (free_sgm_p->numBlks - 1) ; nbr++)
		{
		  blk_p = NVRAM_GET_BLOCK_P(block) ;
		  block = blk_p->next_block;
		}
	  blk_p = NVRAM_GET_BLOCK_P(block) ;
	  if (blk_p->next_block !=  NVRAM_END_OF_BLOCK)
		{
		  // something wrong in free list 
		  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_deallocate: free block list problem"));
		  NFX_SPIN_UNLOCK(&nvram_lock);
		  return NFX_ERR;
		}
	}

  // Everything is ready
  // Store all changes in modification record

  mod_p->block = block;  // could be NVRAM_END_OF_BLOCK
  mod_p->next_block = sgm_p->startBlk ;

  mod_p->free_sgm_new_numBlks = free_sgm_p->numBlks + sgm_p->numBlks ;
  if (block == NVRAM_END_OF_BLOCK)  // free list was empty
	mod_p->free_sgm_new_startBlk = sgm_p->startBlk ;
  else
	mod_p->free_sgm_new_startBlk = free_sgm_p->startBlk ;
  // free_sgm_p->timeStamp = ToDo ;

  mod_p->segment = segment ;
  mod_p->sgm_clientId = NVRAM_CLIENT_NONE ;
  mod_p->sgm_clientCookie = -1 ;
  mod_p->sgm_startBlk = NVRAM_END_OF_BLOCK ;
  mod_p->sgm_numBlks = 0 ;
  mod_p->sgm_Handle = NVRAM_INVALID_HANDLE ;
  // mod_p->sgm_timeStamp = ToDo ;

  mod_p->magic = NVRAM_MODIFICATION_MAGIC;
  mod_p->state = NVRAM_MODIFICATION_PROGRESS;
  mod_p->cksum = 0;
  ck_sum (mod_p, &(mod_p->spare), &(mod_p->cksum));

  // From here on we can recover from the modification record

  // Now activate all changes
  if (nvram_activate_modification (mod_p) != NFX_OK)
	{
	  // PANIC, forget the whole thing
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_deallocate: the activate_modification failed: Panic"));
	  bzero (mod_p, sizeof(nvram_modification_t)); 
	  NFX_SPIN_UNLOCK(&nvram_lock);
	  return (NFX_ERR);
	}

  // Changes succesfull done
  // Erase Modification record
  bzero (mod_p, sizeof(nvram_modification_t)); 

  // OK, all changes done succesfully

  NFX_SPIN_UNLOCK(&nvram_lock);

  return NFX_OK;

} /* nvram_deallocate */

/*-----------------------------------------------------------------------------
 * nvram_find
 *
 * Description:  Find a previously allocated NVRAM segment.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
int32
nvram_find(
    nvram_clientId_t    clientId,
    nvram_cookie_t      cookie,
    nvram_handle_t      *handlep)
{
  int nbr ;
  nvram_segment_t *sgm_p;

  // This does not write, so no problem with a crash
  // Avoid that someone else write to nvram 
  // Max blocking time is loop of segments (512)

  nvram_ever_used = 1;
  
  NFX_SPIN_LOCK(&nvram_lock);

  if (nvram_check_header(nvram_hdr_p) != NFX_OK)
	{
	  // request REFORMAT ???
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_find: wrong nvram header"));
	  NFX_SPIN_UNLOCK(&nvram_lock);
	  return NFX_ERR;
	}

  for (nbr = 1 ; nbr < nvram_hdr_p->segments ; nbr++)
	{
	  sgm_p = NVRAM_GET_SEGMENT_P(nbr) ;
  
	  if ((sgm_p->clientId == clientId) &&
		  (sgm_p->clientCookie == cookie) &&
		  (nvram_check_segment(sgm_p) == NFX_OK))
		{
		  *handlep = sgm_p->Handle;
		  NVRAM_TRACE (NVRAM_TRACE_MNGT, 0, ("nvram_find:  %d %lld : handle 0x%x", clientId, cookie, sgm_p->Handle));
		  NFX_SPIN_UNLOCK(&nvram_lock);
		  return (NFX_OK);
		}
	}
  
  *handlep = NVRAM_INVALID_HANDLE;
  
  NVRAM_TRACE (NVRAM_TRACE_MNGT, 0, ("nvram_find: Not found, %d %lld ",  clientId, cookie));

  NFX_SPIN_UNLOCK(&nvram_lock);
  return NFX_ERR;

} /* nvram_find */

/*-----------------------------------------------------------------------------
 * nvram_read
 *
 * Description:   Read from an NVRAM segment.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
int32
nvram_read(
    nvram_rw_t    *req)
{
    int segment ;
    nvram_segment_t *sgm_p;
    int block;
    nvram_block_t *blk_p ;
    int vec;
    int dataSize;
    int dataOffset ; // within the block

    // NVRAM_TRACE (NVRAM_TRACE_READ, 0, ("nvram_read"));

    NFX_SPIN_LOCK(&nvram_lock);

    if (nvram_check_header(nvram_hdr_p) != NFX_OK)
    {
        // request REFORMAT ???
        NVRAM_ERROR (NVRAM_TRACE_READ, 0, ("nvram_read: wrong nvram_hdr_p"));
        NFX_SPIN_UNLOCK(&nvram_lock);
        return NFX_ERR;
    }

    segment = NVRAM_SEGMENT_HANDLE(req->handle);

    if ((segment < 1) ||
        (segment >= nvram_hdr_p->segments))
    {
        // wrong Handle
        NVRAM_ERROR (NVRAM_TRACE_READ, 0, ("nvram_read: wrong Handle %d 0x%x", segment, req->handle));
        NFX_SPIN_UNLOCK(&nvram_lock);
        return NFX_ERR;
    }

    sgm_p = NVRAM_GET_SEGMENT_P(segment);

    if (sgm_p->Handle != req->handle)
    {
        // wrong Handle, stale
        NVRAM_ERROR (NVRAM_TRACE_READ, 0, ("nvram_read: wrong Handle 0x%x 0x%x", sgm_p->Handle, req->handle));
        NFX_SPIN_UNLOCK(&nvram_lock);
        return NFX_ERR;
    }

    if (nvram_check_segment(sgm_p) != NFX_OK)
    {
        // request REFORMAT ???
        NVRAM_ERROR (NVRAM_TRACE_READ, 0, ("nvram_read: wrong segment check"));
        NFX_SPIN_UNLOCK(&nvram_lock);
        return NFX_ERR;	  
    }

    block = sgm_p->startBlk ;
    blk_p = NVRAM_GET_BLOCK_P(block);
    dataSize = nvram_hdr_p->blockSize - sizeof(nvram_block_t) ;
    dataOffset = req->offset; // within the block
  
    if (req->offset >= (sgm_p->numBlks * dataSize))
    {
        // fs_error(NULL, FS_LOG_WRITE_ERR, "offset");
        NVRAM_ERROR (NVRAM_TRACE_READ, 0, ("nvram_read: wrong offset %d %d", req->offset, (sgm_p->numBlks * dataSize)));
        NFX_SPIN_UNLOCK(&nvram_lock);
        return NFX_ERR;	  
    }

    // find block with given offset 
    while (dataOffset >= dataSize)
    {
        dataOffset -= dataSize;
        block = blk_p->next_block;
        if (block == NVRAM_END_OF_BLOCK)
        {
            // fs_error(NULL, FS_SIM_LSEEK_ERR, "fpnv_read");
            NVRAM_ERROR (NVRAM_TRACE_READ, 0, ("nvram_read: wrap problem %d", dataOffset));
            NFX_SPIN_UNLOCK(&nvram_lock);
            return NFX_ERR;	  
        }
        else
        {
            blk_p = NVRAM_GET_BLOCK_P(block);
            NVRAM_TRACE (NVRAM_TRACE_READ, 0, ("nvram_read: block %3d o %6d", block, dataOffset));
        }
    }  

    for (vec = 0; vec < req->numVecs; vec++)
    {
        int bytesTodo;
        char *datap;

        datap = req->datap[vec];
        bytesTodo = req->lenp[vec];

        NVRAM_TRACE (NVRAM_TRACE_READ, 0, ("nvram_read: o %6d %6d %6d v %3d b %3d", 
                                           req->offset, dataOffset, dataSize, vec, bytesTodo));

        if (bytesTodo < (dataSize - dataOffset))
        {
            memcpy(datap, &(blk_p->data[dataOffset]), bytesTodo);

            dataOffset += bytesTodo;
        }
        else
        {
            // not enough data in this block, read part and go to next block
            int this_bytes ;

            while (bytesTodo > 0)
            {
                NVRAM_TRACE (NVRAM_TRACE_READ, 0, ("nvram_read: part 1 %6d %6d", dataOffset, dataSize));

                this_bytes = MIN((dataSize - dataOffset), bytesTodo);

                memcpy(datap, &(blk_p->data[dataOffset]), this_bytes);
                datap += this_bytes;
                bytesTodo -= this_bytes ;
                dataOffset += this_bytes ;

                if (bytesTodo > 0)
                {
                    block = blk_p->next_block;
                    if (block == NVRAM_END_OF_BLOCK)
                    {
                        if (req->doWrap)
                        {
                            dataOffset = req->wrapReserve;
                            block = sgm_p->startBlk ;
                            blk_p = NVRAM_GET_BLOCK_P(block);
                            NVRAM_TRACE (NVRAM_TRACE_READ, 0, ("nvram_read: part 1 wrap block %3d o %6d", 
                                                               block, dataOffset));
                        }
                        else
                        {
                            // fs_error(NULL, FS_SIM_LSEEK_ERR, "fpnv_read");
                            NVRAM_ERROR (NVRAM_TRACE_READ, 0, ("nvram_read: wrap problem 1"));
                            NFX_SPIN_UNLOCK(&nvram_lock);
                            return NFX_ERR;	  
                        }
                    }
                    else
                    {
                        blk_p = NVRAM_GET_BLOCK_P(block);
                        dataOffset = 0 ;
                        NVRAM_TRACE (NVRAM_TRACE_WRITE, 0, ("nvram_read: part 1 block %3d o %6d", 
                                                            block, dataOffset));
                    }
                }
            }
        }
	}
  
    NFX_SPIN_UNLOCK(&nvram_lock);

    return NFX_OK;

} /* nvram_read */

/*-----------------------------------------------------------------------------
 * nvram_write
 *
 * Description:   Write to an NVRAM segment.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
int32
nvram_write(
    nvram_rw_t    *req)
{
    int segment ;
    nvram_segment_t *sgm_p;
    int block;
    nvram_block_t *blk_p ;
    int vec;
    int dataSize;
    int dataOffset ; // within the block

    // NVRAM_TRACE (NVRAM_TRACE_WRITE, 0, ("nvram_write"));

    NFX_SPIN_LOCK(&nvram_lock);

    if (nvram_check_header(nvram_hdr_p) != NFX_OK)
	{
        // request REFORMAT ???
        NVRAM_ERROR (NVRAM_TRACE_WRITE, 0, ("nvram_write: wrong nvram_hdr_p "));
        NFX_SPIN_UNLOCK(&nvram_lock);
        return NFX_ERR;
	}

    segment = NVRAM_SEGMENT_HANDLE(req->handle);

    if ((segment < 1) ||
        (segment >= nvram_hdr_p->segments))
	{
        // wrong Handle
        NVRAM_ERROR (NVRAM_TRACE_WRITE, 0, ("nvram_write: wrong Handle %d 0x%x", segment, req->handle));
        NFX_SPIN_UNLOCK(&nvram_lock);
        return NFX_ERR;
	}

    sgm_p = NVRAM_GET_SEGMENT_P(segment);

    if (sgm_p->Handle != req->handle)
	{
        // wrong Handle, stale
        NVRAM_ERROR (NVRAM_TRACE_WRITE, 0, ("nvram_write: wrong Handle 0x%x 0x%x", sgm_p->Handle, req->handle));
        NFX_SPIN_UNLOCK(&nvram_lock);
        return NFX_ERR;
	}

    if (nvram_check_segment(sgm_p) != NFX_OK)
	{
        // request REFORMAT ???
        NVRAM_ERROR (NVRAM_TRACE_WRITE, 0, ("nvram_write: wrong segment check"));
        NFX_SPIN_UNLOCK(&nvram_lock);
        return NFX_ERR;	  
	}

    block = sgm_p->startBlk ;
    blk_p = NVRAM_GET_BLOCK_P(block);
    dataSize = nvram_hdr_p->blockSize - sizeof(nvram_block_t) ;
    dataOffset = req->offset;

    if (req->offset >= (sgm_p->numBlks * dataSize))
    {
        // fs_error(NULL, FS_LOG_WRITE_ERR, "offset");
        NVRAM_ERROR (NVRAM_TRACE_WRITE, 0, ("nvram_write: wrong offset %d %d", req->offset, (sgm_p->numBlks * dataSize)));
        NFX_SPIN_UNLOCK(&nvram_lock);
        return NFX_ERR;	  
    }

    // find block with given offset 
    while (dataOffset >= dataSize)
	{
        dataOffset -= dataSize;
        block = blk_p->next_block;
        if (block == NVRAM_END_OF_BLOCK)
		{
            NVRAM_ERROR (NVRAM_TRACE_WRITE, 0, ("nvram_write: wrap problem"));
            NFX_SPIN_UNLOCK(&nvram_lock);
            return NFX_ERR;	  
		}
        else
		{
            blk_p = NVRAM_GET_BLOCK_P(block);
            NVRAM_TRACE (NVRAM_TRACE_WRITE, 0, ("nvram_write: block %3d o %6d", block, dataOffset));
		}
	}


    for (vec = 0; vec < req->numVecs; vec++)
	{
        int bytesTodo;
        char *datap;
	  
		if (req->lenp[vec] == 0)
			continue;

        datap = req->datap[vec];
        bytesTodo = req->lenp[vec];

        NVRAM_TRACE (NVRAM_TRACE_WRITE, 0, ("nvram_write: o %6d %6d %6d v %3d b %3d", 
                                            req->offset, dataOffset, dataSize, vec, bytesTodo));

        if (bytesTodo < (dataSize - dataOffset))
		{
            memcpy (&(blk_p->data[dataOffset]), datap, bytesTodo);

            dataOffset += bytesTodo;
		}
        else
		{
            // not enough place in this block, write part and go to next block
            int this_bytes ;

            while (bytesTodo > 0)
            {
                NVRAM_TRACE (NVRAM_TRACE_WRITE, 0, ("nvram_write: part 1 %6d %6d", 
                                                    dataOffset, dataSize));
                this_bytes = MIN((dataSize - dataOffset), bytesTodo);

                memcpy (&(blk_p->data[dataOffset]), datap, this_bytes);

                datap += this_bytes;
                bytesTodo -= this_bytes ;
                dataOffset += this_bytes;
                
                if (bytesTodo > 0)
                {
                    block = blk_p->next_block;
                    if (block == NVRAM_END_OF_BLOCK)
                    {
                        if (req->doWrap)
                        {
                            dataOffset = req->wrapReserve;
                            block = sgm_p->startBlk ;
                            blk_p = NVRAM_GET_BLOCK_P(block);
                            NVRAM_TRACE (NVRAM_TRACE_WRITE, 0, ("nvram_write: part 1 wrap block %3d o %6d",
                                                                block, dataOffset));
                        }
                        else
                        {
                            // fs_error(NULL, FS_SIM_LSEEK_ERR, "fpnv_read");
                            NVRAM_ERROR (NVRAM_TRACE_WRITE, 0, ("nvram_write: wrap problem 1"));
                            NFX_SPIN_UNLOCK(&nvram_lock);
                            return NFX_ERR;	  
                        }
                    }
                    else
                    {
                        blk_p = NVRAM_GET_BLOCK_P(block);
                        dataOffset = 0 ;
                        NVRAM_TRACE (NVRAM_TRACE_WRITE, 0, ("nvram_write: part 1 block %3d o %6d", 
                                                            block, dataOffset));
                    }
                }
            }
		}
	}

    NFX_SPIN_UNLOCK(&nvram_lock);

    return NFX_OK;

} /* nvram_write */




#if defined(STANDALONE)

int32
nvram_restart_from_file()
{
  int fd = 0;
  int numRead;
  int tot_numRead = 0;
  int nbr, nbr_1;
  nvram_segment_t *sgm_p;
  int block;
  nvram_block_t *blk_p ;
  nvram_modification_t *mod_p;
  nvram_header_t nvram_hdr;  
  int free_blocks = 0;
  int segments = 0;
  int used_blocks = 0;
  int mod_record = 0;

  // NVRAM_TRACE (NVRAM_TRACE_MNGT, 0, ("nvram_restart_from_file"));

  NFX_SPIN_LOCK(&nvram_lock);

  // whatever the state, read again, as long as the pointer exist
  if (nvram_hdr_p == NULL)
	{
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_restart_from_file: nvram_hdr_p == NULL"));
	  NFX_SPIN_UNLOCK(&nvram_lock);
	  return NFX_ERR;
	}

  fd = open("FS_EAGLE_SIM_NVRAM", O_RDONLY);
  if (fd > 0)
	{
	  numRead = read(fd, &nvram_hdr, sizeof(nvram_header_t));
	  if (numRead != sizeof(nvram_header_t))
		{
		  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_restart_from_file: read header failed"));
		  goto nvram_restart_cleanup;
		}
	  tot_numRead += numRead ;

	  NVRAM_TRACE (NVRAM_TRACE_MNGT, 0, ("nvram_restart_from_file: found an NVRAM system of %d KB", nvram_hdr.nvramSize));

#if defined(STANDALONE)
	  eee_ramDealloc (nvram_hdr_p);
	  nvram_hdr_p = (void*)eee_ramAlloc(nvram_hdr.nvramSize * 1024) ;
#endif

	  memcpy (nvram_hdr_p, &nvram_hdr, sizeof(nvram_header_t));
  
	  for (nbr = 0 ; nbr < nvram_hdr_p->segments ; nbr++)
		{
		  sgm_p = NVRAM_GET_SEGMENT_P(nbr) ;

		  numRead = read(fd, sgm_p, sizeof(nvram_segment_t));
		  if (numRead != sizeof(nvram_segment_t))
			{
			  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_restart_from_file: read segment %d failed", nbr));
			  goto nvram_restart_cleanup;
			}
		  tot_numRead += numRead ;
		  segments++;
		}

	  // READ BLOCKS !!!!
	  nbr = 0 ;
	  sgm_p = NVRAM_GET_SEGMENT_P(nbr) ;
	  block = sgm_p->startBlk ;
	  for (nbr_1 = 0 ; nbr_1 < sgm_p->numBlks ; nbr_1++)
		{
		  blk_p = NVRAM_GET_BLOCK_P(block) ;
		  numRead = read(fd, &(blk_p->next_block), sizeof(uint32));
		  if (numRead != sizeof(uint32))
			{
			  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_read_from_file: read block %d %d failed", nbr, nbr_1));
			  goto nvram_restart_cleanup;
			}
		  tot_numRead += numRead ;
		  free_blocks++;
		  blk_p->magic = NVRAM_BLOCK_MAGIC ;
		  blk_p->cksum = 0 ;
		  ck_sum (blk_p, &(blk_p->data), &(blk_p->cksum));

		  block = blk_p->next_block;
		}
	  
	  for (nbr = 1 ; nbr < nvram_hdr_p->segments ; nbr++)
		{
		  sgm_p = NVRAM_GET_SEGMENT_P(nbr) ;

		  block = sgm_p->startBlk ;
		  for (nbr_1 = 0 ; nbr_1 < sgm_p->numBlks ; nbr_1++)
			{
			  blk_p = NVRAM_GET_BLOCK_P(block) ;
			  numRead = read(fd, blk_p, nvram_hdr_p->blockSize);
			  if (numRead != nvram_hdr_p->blockSize)
				{
				  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_read_from_file: read block %d %d failed, %d %d %d", 
													 nbr, nbr_1, sgm_p->numBlks, numRead, nvram_hdr_p->blockSize));
				  goto nvram_restart_cleanup;
				}
			  tot_numRead += numRead ;
			  used_blocks++ ;
			  block = blk_p->next_block;
			}
		}

	  mod_p = NVRAM_GET_MOD_P();
	  numRead = read(fd, mod_p, sizeof(nvram_modification_t));
	  if (numRead == sizeof(nvram_modification_t))
		{
		  NVRAM_TRACE (NVRAM_TRACE_MNGT, 0, ("nvram_read_from_file: read modification record"));
		  tot_numRead += numRead ;
          mod_record = 1 ;
        }
	  else
		bzero(mod_p, sizeof(nvram_modification_t));

	  close (fd);

	}

  NFX_SPIN_UNLOCK(&nvram_lock);

  NVRAM_TRACE (NVRAM_TRACE_MNGT, 0, ("nvram_read_from_file: read %d Bytes, %d segments, %d free %d used blocks, mod %d", 
			   tot_numRead, segments, free_blocks, used_blocks, mod_record));
printf ("nvram_read_from_file: read %d Bytes, %d segments, %d free %d used blocks, mod %d\n", 
        tot_numRead, segments, free_blocks, used_blocks, mod_record);


  return (NFX_OK);

nvram_restart_cleanup:

  if (fd)
	close (fd);

  return (NFX_ERR);

}

int32
nvram_write_to_file()
{
  int fd = 0;
  int openFlags = 0;
  int numWritten;
  int tot_numWritten = 0;
  int nbr, nbr_1;
  nvram_segment_t *sgm_p;
  int block;
  nvram_block_t *blk_p ;
  nvram_modification_t *mod_p;

  int free_blocks = 0;
  int segments = 0;
  int used_blocks = 0;
  int mod_record = 0;

  static int write_once = 0 ;

  // NVRAM_TRACE (NVRAM_TRACE_MNGT, 0, ("nvram_write_to_file"));

  NFX_SPIN_LOCK(&nvram_lock);

  if (write_once)
	{
	  NVRAM_TRACE (NVRAM_TRACE_MNGT, 0, ("nvram_write_to_file: was already done"));
	  NFX_SPIN_UNLOCK(&nvram_lock);
	  return (0);
	}

  if (nvram_check_header(nvram_hdr_p) != NFX_OK)
	{
	  // request REFORMAT ???
	  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_write_to_file: wrong nvram_hdr_p"));
	  NFX_SPIN_UNLOCK(&nvram_lock);
	  return NFX_ERR;
	}

  fd = open("FS_EAGLE_SIM_NVRAM",  openFlags | O_RDWR | O_CREAT, S_IREAD | S_IWRITE);

  if (fd > 0)
	{
	  numWritten = write(fd, nvram_hdr_p, sizeof(nvram_header_t));
	  if (numWritten != sizeof(nvram_header_t))
		{
		  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_write_to_file: write header failed"));
		  goto nvram_write_cleanup;
		}
	  tot_numWritten += numWritten;
	  
	  for (nbr = 0 ; nbr < nvram_hdr_p->segments ; nbr++)
		{
		  sgm_p = NVRAM_GET_SEGMENT_P(nbr) ;

		  numWritten = write(fd, sgm_p, sizeof(nvram_segment_t));
		  if (numWritten != sizeof(nvram_segment_t))
			{
			  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_write_to_file: write segment %d failed", nbr));
			  goto nvram_write_cleanup;
			}
		  tot_numWritten += numWritten;
		  segments++;
		}

	  // WRITE BLOCKS !!!!
	  nbr = 0 ;
	  sgm_p = NVRAM_GET_SEGMENT_P(nbr) ;
	  // write only the free list pointers free blocks !!!
	  block = sgm_p->startBlk ;
	  for (nbr_1 = 0 ; nbr_1 < sgm_p->numBlks ; nbr_1++)
		{
		  blk_p = NVRAM_GET_BLOCK_P(block) ;
		  numWritten = write(fd, &(blk_p->next_block), sizeof(uint32));
		  if (numWritten != sizeof(uint32))
			{
			  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_write_to_file: write block %d %d failed %d", 
                                                 nbr, nbr_1, numWritten));              
			  goto nvram_write_cleanup;
			}
		  tot_numWritten += numWritten;
		  block = blk_p->next_block;
		  free_blocks++;
		}

	  for (nbr = 1 ; nbr < nvram_hdr_p->segments ; nbr++)
		{
		  sgm_p = NVRAM_GET_SEGMENT_P(nbr) ;
		  block = sgm_p->startBlk ;
		  for (nbr_1 = 0 ; nbr_1 < sgm_p->numBlks ; nbr_1++)
			{
			  blk_p = NVRAM_GET_BLOCK_P(block) ;
			  numWritten = write(fd, blk_p, nvram_hdr_p->blockSize);
			  if (numWritten != nvram_hdr_p->blockSize)
				{
				  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_write_to_file: write block %d %d failed, %d %d %d", 
													 nbr, nbr_1, sgm_p->numBlks, numWritten, nvram_hdr_p->blockSize));
				  goto nvram_write_cleanup;
				}
			  tot_numWritten += numWritten;
			  used_blocks++ ;
			  block = blk_p->next_block;
			}
		}

	  mod_p = NVRAM_GET_MOD_P();
      if ((mod_p->magic == NVRAM_MODIFICATION_MAGIC) &&
          (ck_sum (mod_p, &(mod_p->spare), NULL) == 0))
		{
		  NVRAM_TRACE (NVRAM_TRACE_MNGT, 0, ("nvram_write_to_file: write modification record "));
		  numWritten = write(fd, mod_p, sizeof(nvram_modification_t));
		  if (numWritten != sizeof(nvram_modification_t))
			{
			  NVRAM_ERROR (NVRAM_TRACE_MNGT, 0, ("nvram_write_to_file: write modification record failed"));
			  goto nvram_write_cleanup;
			}
		  tot_numWritten += numWritten;
          mod_record = 1 ;
		}

	  close (fd);

	  write_once = 1 ;

	}

  NFX_SPIN_UNLOCK(&nvram_lock);

  NVRAM_TRACE (NVRAM_TRACE_MNGT, 0, ("nvram_write_to_file: write %d Bytes, %d segments, %d free %d used blocks, mod %d", 
			   tot_numWritten, segments, free_blocks, used_blocks, mod_record));
printf ("nvram_write_to_file: write %d Bytes, %d segments, %d free %d used blocks, mod %d\n", 
        tot_numWritten, segments, free_blocks, used_blocks, mod_record);

  return (NFX_OK);

nvram_write_cleanup:

  if (fd)
	close (fd);

  return (NFX_ERR);
}

#endif // STANDALONE

#endif // !SSC
