// vi:ts=4
/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : opt-api.h
 *
 * Description  : Interface and API of the registry manager (REGM).
 *
 *
 * Agile configuration file
 * ------------------------
 *
 * All agile specific information is stored in one file
 * (/usr/local/agile/agile.conf).
 * This file is in ascii format and managed by the registry manager.
 *
 * The syntax of a line in the file is: <option>=<value>.
 * Where:
 *     <option>     <keyword>.<module_name>.<...>
 *     <keyword>    options|state
 *     options      Defines the configuration info and can be view by operator
 *     state        Allows modules to store some persistent info like state,
 *                  this is hiden to the operator.
 *     <...>        Module specific syntax.
 *     <value>      Value of the option. Could ne anything but
 *                  shall not contain '=', tabulations or spaces.
 *
 * A line shall always terminate with a '\n'.
 * Line starting with '#' will be ignored and removed.
 *
 *
 * Options command
 * ---------------
 *
 * The options command allows the operator to view the configuration and
 * to modify it.
 *
 * options
 *     Display all the options and their values.
 * options <option_prefix>
 *     Display all the options (and their values) matching the string
 *     option_prefix.
 * options <option> <value>
 *     Modify the value of the (fully specified) option.
 *
 *
 * Registry manager (regm)
 * -----------------------
 *
 * The registry manager offers the following interfaces:
 *
 * 1. At any time, a module (XYZ) can get any information stored
 *    in the registry file.
 *
 *    The module just need to specify the option or set of options it wants.
 *
 *    The registry manager will return all the matching options and
 *    their value (if any).
 * 
 *       XYZ                             REGM
 *        |        GET_OPTION_REQ         |
 *        |------------------------------>|
 *        |                               |
 *        |        GET_OPTION_RSP         |
 *        |<------------------------------|
 *        |                               |
 *
 * 2. A module (XYZ) can add/modify options information in the registry file.
 * 
 *       XYZ                             REGM
 *        |        SET_OPTION_REQ         |
 *        |------------------------------>|
 *        |                               |
 *        |        SET_OPTION_RSP         |
 *        |<------------------------------|
 *        |                               |
 *
 *
 * If you need to add an option into the file...
 * ---------------------------------------------
 *
 * 1. Modify agile.conf
 *    Add the option in the file following the file syntax.
 *
 * 2. Modify opt-api.h (this file)
 *    Add a #define with the syntax of the new option.
 *
 * 3. Modify opt-routines.c
 *    If it's a new module, add the module name into the global 'cmds' table,
 *    and create a new routine to handle the options set command.
 *    If the module already exist, modify the corresponding routine to handle
 *    the new option.
 *
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-SEP-2001
 *
 *------------------------------------------------------------------------------
 */
#ifndef _OPT_API_H_
#define _OPT_API_H_

/*------------------------------------------------------------------------------
 * Include files
 *------------------------------------------------------------------------------
 */

#include "nfx-incl.h"

/*------------------------------------------------------------------------------
 * Constants and macros
 *------------------------------------------------------------------------------
 */

#define OPT_MAX_OPTION_LEN      128 // to be optimized
#define MAX_VALUE_LEN       256 // to be optimized


#define OPTIONS_AUTOSUPPORT_ALL         "options.autosupport.*"
#define OPTIONS_AUTOSUPPORT_ENABLE      "options.autosupport.enable"
#define OPTIONS_AUTOSUPPORT_TO          "options.autosupport.to"
#define OPTIONS_AUTOSUPPORT_NOTETO      "options.autosupport.noteto"
#define OPTIONS_AUTOSUPPORT_SCHEDULE    "options.autosupport.schedule"

#define OPTIONS_ELOG_ALL                "options.elog.*"
#define OPTIONS_ELOG_ENABLE             "options.elog.enable"
#define OPTIONS_ELOG_DISPLAY            "options.elog.display"
#define OPTIONS_ELOG_FACILITY           "options.elog.facility"
#define OPTIONS_ELOG_HOST               "options.elog.host"
#define OPTIONS_ELOG_LEVEL              "options.elog.level"

#define OPTIONS_NDMP_AUTH_CLIENT_ID     "options.ndmp.auth.client"
#define OPTIONS_NDMP_AUTH_CLIENT_ALL    "options.ndmp.auth.client.*"

#define OPTIONS_NDMP_AUTH_SERVER_ID     "options.ndmp.auth.server"
#define OPTIONS_NDMP_AUTH_SERVER_ALL    "options.ndmp.auth.server.*"

#define OPTIONS_NDMP_HI_VER		        "options.ndmp.hi.ver"
#define OPTIONS_NDMP_ENBL_MODE	        "options.ndmp.enable.mode"
#define OPTIONS_NDMP_RESTORE_MODE       "options.ndmp.restore.mode"
#define OPTIONS_NDMP_TRACE_MODE         "options.ndmp.trace.mode"
#define OPTIONS_NDMP_TRACE_DURATION     "options.ndmp.trace.duration"

#define OPTIONS_NVRAM_MODE              "options.nvram.mode"
#define OPTIONS_NVRAM_SIZE              "options.nvram.size"
#define OPTIONS_NVRAM_SEGMENT           "options.nvram.segment"
#define OPTIONS_NVRAM_BLOCKSIZE         "options.nvram.blocksize"

#define OPTIONS_AV_ALL         			"options.av.*"
#define OPTIONS_AV_LOGIN				"options.av.login"
#define OPTIONS_AV_PASSWORD				"options.av.password"
#define OPTIONS_AV_IP					"options.av.ip"

#define OPTIONS_EVM_VOL_SEQ             "evm.vol.seq"
#define OPTIONS_EVM_SPOOL_SEQ           "evm.spool.seq"
#define OPTIONS_SPM_POOL_SEQ            "spm.pool.seq"

#define OPTIONS_EVM_IO_PENDING_QSIZE    "options.evm.io.pending.qsize"
#define OPTIONS_EVM_IO_WAITING_QSIZE    "options.evm.io.waiting.qsize"

#define OPTIONS_CLUSTER_NAME			"cluster.name"
#define OPTIONS_CLUSTER_IP				"cluster.ip"
#define OPTIONS_CLUSTER_LV				"cluster.lv"
#define OPTIONS_CLUSTER_NFS				"cluster.nfs"
#define OPTIONS_SLAVE_ALL				"slave.*"
#define OPTIONS_CLUSTER_NODE_TYPE		"cluster.nodetype"
#define OPTIONS_CLUSTER_MEMBER			"cluster.member"
#define OPTIONS_NODE_TYPE				"node.type"

#define OPTIONS_TAPE_ALIAS 				"options.tape.alias"
#define OPTIONS_NTP_MODE 				"ntp.mode"

#define OPTIONS_NAMESERVICE 			"nameservice"

#define OPTIONS_IPM_INTF_MTU			"ipm.if.mtu"
#define OPTIONS_IPM_INTF_STATE			"ipm.if.state"
#define OPTIONS_IPM_INTF_AUTONEG		"ipm.if.autoneg"
#define OPTIONS_IPM_VIF_CONF			"ipm.vif"


/*------------------------------------------------------------------------------
 * Type definitions.
 *------------------------------------------------------------------------------
 */

typedef char opt_option_t[OPT_MAX_OPTION_LEN];
typedef char opt_value_t[MAX_VALUE_LEN];

/* Registry binary entry.
 */
typedef struct
{
    opt_option_t    option;
    opt_value_t     value;
} opt_entry_t;

/*------------------------------------------------------------------------------
 * Header of the message.
 *------------------------------------------------------------------------------
 */

/* Message type
 */
typedef enum
{
    OPT_GET_OPTION_REQ  = REGM_MSG_BASE,
    OPT_GET_OPTION_RSP,
    
    OPT_SET_OPTION_REQ,
    OPT_SET_OPTION_RSP,

	OPT_GET_BIGOPT_REQ,
	OPT_GET_BIGOPT_RSP,
	OPT_SET_BIGOPT_REQ,
	OPT_SET_BIGOPT_RSP

} opt_msgType_t;

/* User info.
 * Client private data that can be use as a cookie.
 */
typedef uint32  opt_userInfo_t;

/* Header of the message.
 */
typedef struct
{
    opt_msgType_t   msgType;    // the message type SHALL always be first
    opt_userInfo_t  userInfo;
} opt_msgHdr_t;

/*------------------------------------------------------------------------------
 * Body of the message.
 *------------------------------------------------------------------------------
 */

/* Allow any modules to get configuration info.
 *
 * To get info, the user shall specify the option (wilcard can be used):
 * "options.*"              value of all options
 * "options.nis.*"          value of all NIS options
 * "options.nfs.tcp.enable" value of specific NFS option
 */
typedef struct
{
    opt_option_t    option;     // which option to get?
} opt_getOptionReq_t;

typedef struct
{
    int             nbEntry;    // length of the  variable length array
    opt_entry_t     entry[1];   // variable length array of {option, values}
} opt_getOptionRsp_t;

/* Allow any modules to set configuration info.
 */
typedef struct
{
    boolean     del;    // TRUE if to remove the option, FALSE otherwise
    opt_entry_t entry;  // {option, values} to add/modify in registry
} opt_setOptionReq_t;

typedef struct
{
    int     result;     // NFX_OK, NFX_ERR
} opt_setOptionRsp_t;

/*----------------------------------------------------------------------------*/

// Things to make message formation easier

typedef struct {
	opt_msgHdr_t		hdr;		// OPT_SET_OPTION_REQ
	opt_setOptionReq_t	setReq;
} regmSetReq_t;

typedef struct {
	opt_msgHdr_t		hdr;		// OPT_GET_OPTION_REQ
	opt_getOptionReq_t	getReq;
} regmGetReq_t;

typedef struct {
	opt_msgHdr_t		hdr;		// OPT_SET_OPTION_RSP or OPT_SET_BIGOPT_RSP
	opt_setOptionRsp_t	setRsp;
} regmSetRsp_t, regmBigSetRsp_t;

typedef struct {
	opt_msgHdr_t		hdr;		// OPT_GET_OPTION_RSP
	opt_getOptionRsp_t	getRsp;
} regmGetRsp_t;

// And their big brothers

typedef struct {
	opt_msgHdr_t		hdr;		// OPT_SET_BIGOPT_REQ
	int					del;		// Delete 
	char				opt_val[1];	// Pair of NULL terminated strings
									// in the case of del being true, one 
									// string
} regmBigSetReq_t;

typedef struct {
	opt_msgHdr_t		hdr;		// OPT_GET_BIGOPT_REQ
	char				opt[1];		// Single NULL terminated string
} regmBigGetReq_t;

typedef struct {
	opt_msgHdr_t		hdr;		// OPT_GET_BIGOPT_RSP
	int					nbEntry;	// Number of option/value pairs that follow
	char				opt_val[1];	// Pairs of NULL terminated strings
} regmBigGetRsp_t;

/*------------------------------------------------------------------------------
 * Function prototypes
 *------------------------------------------------------------------------------
 */

extern int32 getRegistryOption(char *option, char *value);
extern int32 setRegistryOption(char *option, char *value);

/*
 * How to use these puppys -
 *
 * list = get_cfg_val(key) returns a NULL-terminated list of option/value
 *  pairs.The list and the strings referenced therein are allocated from heap
 *  (malloc()) and MUST NOT BE MODIFIED or free_cfg_val() will not be able
 *  to return the memory once you're done.
 *
 * free_cfg_val(list) frees up the memory allocated by get_cfg_val()
 *
 * put_cfg_val(obj) takes a single pair of option/value string and adds that
 *  to the configuration database.
 *
 * del_cfg_obj(obj) takes a single string and removes it and its value from
 *  the database.
 *
 * Examples:
 *  
 *  Given agile.conf looks like:
 *  cluster.lv.0000beefbabedead=online:puppeteer
 *  cluster.lv.000000000abcdef0=online:puppeteer
 *  
 *  char **lvs = get_cfg_val ("cluster.lv.*"); // All the LVs in this cluster
 *  Will result:
 *  lvs[0] = "cluster.lv.0000beefbabedead"
 *  lvs[1] = "online:puppeteer"
 *  lvs[2] = "cluster.lv.000000000abcdef0"
 *  lvs[3] = "online:puppeteer"
 *  lvs[4] = NULL
 *  
 *  Want to remove the second one?
 *   del_cfg_obj (&lvs[2])
 *
 *  When you're done, free the memory by:
 *   free_cfg_val (lvs);
 */
extern char **get_cfg_val (char *key);
extern void free_cfg_val (char **list);
extern int put_cfg_val (char **objs);
extern int del_cfg_obj (char **objs);

/* How to determine if you are running on the PCC */

extern int node_is_PCC(void);


extern int ensure_fs_writable(const char *path);
extern void restore_fs_permissions(const char *path);
extern int modify_readonly_file(const char *path,
								void *fun, /* int (*)(const char *, void *) */
								void *arg);
extern int make_allfs_writable(void);
extern int string_to_file(const char *path, char *str);

/* How you update the hosts file when changing node name */

extern int redo_hosts (char *hostname, char *newname);

#endif // _OPT_API_H_
