/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : opt-routines.c
 *
 * Description  : Contain all routines to handle to handle the options.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

#include "opt.h"
#include "opt-api.h"
#include "exports/elog-api.h"

extern struct timeval waitRegmTo;

/*------------------------------------------------------------------------------
 * Prototype of local functions.
 *------------------------------------------------------------------------------
 */

static int
opt_setOption(char * option,
              char * value);

/*------------------------------------------------------------------------------
 * Prototype of global functions.
 *------------------------------------------------------------------------------
 */

void
opt_xxx(char * option,
        char * value);

void
opt_autosupport(char * option,
                char * value);

void
opt_elog(char * option,
        char * value);

void
opt_ndmp(char * option,
         char * value);

/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */

opt_cmd_t cmds[] =
{
	{"admin",           opt_xxx},
	{"auditlog",        opt_xxx},
	{"autologout",      opt_xxx},
	{"autosupport",     opt_autosupport},
	{"cf",              opt_xxx},
	{"cifs",            opt_xxx},
	{"cifsuser",        opt_xxx},
	{"cksum_offload",   opt_xxx},
	{"console",         opt_xxx},
	{"disk",            opt_xxx},
	{"dns",             opt_xxx},
	{"elog",             opt_elog},
	//{"evm",             opt_evm},
	{"httpd",           opt_xxx},
	{"if",              opt_xxx},
	{"ip",              opt_xxx},
	{"ndmp",            opt_ndmp},
	{"nfs",             opt_xxx},
	{"nis",             opt_xxx},
	{"nvram",           opt_xxx},
	{"package",         opt_xxx},
	{"raid",            opt_xxx},
	{"rsh",             opt_xxx},
	{"service",         opt_xxx},
	{"snmp",            opt_xxx},
	{"sysconfig",       opt_xxx},
	{"system",          opt_xxx},
	{"thresholds",      opt_xxx},
	{"timed",           opt_xxx},
	{"trusted",         opt_xxx},
	{"udp_lg_dgram",    opt_xxx},
	{"useradmin",       opt_xxx},
	{"vol",             opt_xxx},
	{"wafl",            opt_xxx},
	{0, 0}
};

/*------------------------------------------------------------------------------
 * Local functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : opt_setOption()
 *
 * Description  : Set the option in the registry.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

static int
opt_setOption(char * option,
              char * value)
    
{
    opt_msgHdr_t        setReqHdr;
    opt_setOptionReq_t  setReqBody;
    opt_msg_t       req;
    opt_msg_t       rsp;
    opt_msgHdr_t *  hdr;
    void *          bufp = NULL;
    int             result;

    OPT_DBG_PRINT(("> opt_setOption()\n"));   

    /* Fill the request.
     */
    setReqHdr.msgType = OPT_SET_OPTION_REQ;
    setReqHdr.userInfo = 0; // not used
    
    setReqBody.del = FALSE;
    strcpy(setReqBody.entry.option, option);
    strcpy(setReqBody.entry.value, value);
    
    /* 
     */
    bufp = &req.buf;
    bzero(bufp,
          (sizeof(opt_msgHdr_t) +
           sizeof(opt_setOptionReq_t)));
    
    memcpy(bufp, (void *) &setReqHdr, sizeof(opt_msgHdr_t));
    bufp += sizeof(opt_msgHdr_t);
    
    memcpy(bufp, (void *) &setReqBody, sizeof(opt_setOptionReq_t));
    bufp += sizeof(opt_setOptionReq_t);

    req.len = bufp - (void *) req.buf;

    /* Send the request to the registry manager located on the SSC.
     */
    rsp.len = sendAgileMsg(-1,
                           req.buf, req.len,
                           "regm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                           &waitRegmTo,
                           rsp.buf, sizeof(rsp.buf));
    
    /* Receive response from the registry manager.
     */
    OPT_DBG_PRINT(("nb bytes received = %d\n", rsp.len));
    if (rsp.len == -1)
    {
		perror("sendAgileMsg failed");
		return(NFX_ERR);
    }

    /* Read the message header and check if it's expected response.
     */
    bufp = &rsp.buf;
    hdr = (opt_msgHdr_t *) rsp.buf;
    bufp += sizeof(opt_msgHdr_t);
    if (hdr->msgType != OPT_SET_OPTION_RSP)
    {
        OPT_ERR_PRINT(("Message received from REGM is not the expected one!\n"));
		return(NFX_ERR);
    }
    
    /* Read result.
     */
    memcpy((void *) &result, bufp, sizeof(int));
    bufp += sizeof(int);
    OPT_DBG_PRINT(("result returned = %d\n", result));
    
    OPT_DBG_PRINT(("< opt_setOption()\n"));   
    return(result);
}

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : opt_xxx()
 *
 * Description  : Example of routine.
 * Set the option into the registry.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 03-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

void
opt_xxx(char * option,
        char * value)
    
{
    OPT_DBG_PRINT(("opt_xxx\n"));   
    OPT_DBG_PRINT(("\toption = %s\n", option));   
    OPT_DBG_PRINT(("\tvalue = %s\n", value));

    opt_setOption(option, value);
    return;
}

/*------------------------------------------------------------------------------
 * Name         : opt_autosupport()
 *
 * Description  : Process 'autosupport' options command.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 11-SEP-2001
 *
 *------------------------------------------------------------------------------
 */


void
opt_autosupport(char * option,
                char * value)
    
{
    /* Save configuration in file.
     */
    opt_setOption(option, value);

    {
        /* Notify log manager.
         */
        char                buf[sizeof(elog_configNtfy_t)];
        elog_configNtfy_t *  msg = (elog_configNtfy_t *) &buf;

        OPT_DBG_PRINT(("Notify event manager.\n"));

        bzero((void *) &buf, sizeof(buf));

        msg->msgType = ELOG_CONFIG_NTFY ;
        strcpy(msg->entry.option, option);
        strcpy(msg->entry.value, value);

        if (sendAgileMsg(-1,
                         buf, sizeof(buf),
                         "elog", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                         NULL, NULL, 0) != 0)
        {
            perror("sendAgileMsg() failed");
        }
    }
    
    return;
}

/*------------------------------------------------------------------------------
 * Name         : opt_elog()
 *
 * Description  : Process ELOG options command.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 11-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

#include "exports/elog-api.h"

void
opt_elog(char * option,
        char * value)
    
{
    /* Notify log manager (ELOG).
     */
    {
        char                buf[sizeof(elog_configNtfy_t)];
        elog_configNtfy_t *  msg = (elog_configNtfy_t *) &buf;

        OPT_DBG_PRINT(("Notify event manager.\n"));

        bzero((void *) &buf, sizeof(buf));

        msg->msgType = ELOG_CONFIG_NTFY ;
        strcpy(msg->entry.option, option);
        strcpy(msg->entry.value, value);

        if (sendAgileMsg(-1,
                         buf, sizeof(buf),
                         "elog", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                         NULL, NULL, 0) != 0)

        {
            perror("sendAgileMsg() failed");
        }
    }
    
    /* Save configuration in file.
     */
    opt_setOption(option, value);

    return;
}

/*------------------------------------------------------------------------------
 * Name         : opt_ndmp()
 *
 * Description  : Process NDMP options command.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 08-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

void
opt_ndmp(char * option,
         char * value)
    
{
    printf("To modify NDMP parameters use the following commands:\n");
    printf("\t- ndmpauth for authentication\n");
    printf("\t- ...\n");
    return;
}

#ifdef ISABELLE // under contruction
/*------------------------------------------------------------------------------
 * Name         : opt_elog()
 *
 * Description  : Process 'evm' options command.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 08-JAN-2002
 *
 *------------------------------------------------------------------------------
 */

#include "exports/evm-api.h"

void
opt_evm(char * option,
        char * value)
    
{
    /* Save configuration in file.
     */
    opt_setOption(option, value);

    {
        /* Notify volume manager.
         */
        char                buf[sizeof(evm_configNtfy_t)];
        evm_configNtfy_t *  msg = (evm_configNtfy_t *) &buf;

        bzero((void *) &buf, sizeof(buf));

        msg->msgType = EVM_CONFIG_NTFY ;
        strcpy(msg->entry.option, option);
        strcpy(msg->entry.value, value);

        if (!strcmp(option, OPTIONS_EVM_IO_PENDING_QSIZE) ||
            !strcmp(option, OPTIONS_EVM_IO_WAITING_QSIZE))
        {
            OPT_DBG_PRINT(("Notify volume manager server.\n"));
            // TODO!!!
            // send to all FP-NIMs
            // do I need to send to both processors???
            if (sendAgileMsg(-1,
                             buf, sizeof(buf),
                             "evm", NFX_NFP_FP2, 1, MGMTPLANE,
                             NULL, NULL, 0) != 0)
            {
                perror("sendAgileMsg() failed");
            }
        }
    }
    
    return;
}
#endif // ISABELLE

/*----------------------------------------------------------------------------*/
