/*
 *-----------------------------------------------------------------
 *
 * Name:	queue-api.h
 *
 * RCS:  $Id: queue-api.h,v 1.12 2002/10/22 01:30:40 jong Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	API function prototypes for general queue operations
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */

#ifndef _QUEUE_API_H
#define _QUEUE_API_H

#include <exports/queue-error.h>

#define Q_MAX_QUEUES		1024
#define Q_NUM_NODES			1024

/*
 * control word definitions for searching functions
 */
#define Q_FIND_ONE					0x1
#define Q_FIND_TWO					0x2
#define Q_FIND_BOTH					0x3
#define Q_INSERT_ORDER_ASCENDING	0x10000000
#define Q_INSERT_ORDER_DESCENDING	0x20000000

/*
 * An entry in a queue is of type q_node_t.  Each node contains two pieces
 * of user data (or pointers to other objects).  Currently only support
 * simple integers and pointers.
 */
typedef struct q_node {
	struct q_node	* next;
	struct q_node	* prev;
	uint64  		item1;	// Can hold any pointer and simple integer
	uint64  		item2;	// Can hold any pointer and simple integer
} q_node_t, * q_nodePtr_t;

typedef int32			(*queue_compare_func)(
										 /* item passed in here.  If function
										  * pointer is NULL, then numerial
										  * comparison will be used.
										  * Should return:
										  *	  0  = items are equal,
										  *	< 0  = item1 is smaller than item2,
										  *	> 0  = item1 is larger than item2,
										  */
						const void * item1,	/* pointer to user data */
						const void * item2	/* pointer to user data */
					);

/*
 * Each queue starts with a head node.  The head node contains the number
 * of nodes in the queue, the internal read pointer for the SuperQueue
 * functions, and a "queue is in use" flag.
 */
typedef struct q_head_node {
	uint32			in_use;
	struct q_node	* read_ptr;
	uint32			size;
	uint32			item1Size;
	uint32			item2Size;
	struct q_node	* next;
	struct q_node	* prev;
	uint32			index;					/* into head_node array */
    void            * owner;
	queue_compare_func compare_func1;
	queue_compare_func compare_func2;
} q_headNode_t, * q_headNodePtr_t;

/*
 * The nodes are allocated at compile time.  This can be changed later.
 */
typedef struct queues {
	q_headNode_t	head_node[Q_MAX_QUEUES];
	q_node_t		nodes[Q_NUM_NODES];
	uint32			num_nodes_avail;
	uint32			num_queues_avail;
	q_nodePtr_t		next_free_node;
#define Q_NO_FREE_QUEUE_INDEX		0xFFFFFFFF
	uint32			next_free_queue;		/* index into head_node array */
} queues_t;

/*-----------------------------------------------------------------
 * Name:	q_createQueue()
 *
 * Description:	Create a new queue.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
extern q_headNodePtr_t			/* returns the queue head pointer or NULL
								 * if a queue couldn't be created
								 */
q_createQueue(
	uint32 item1Size,				/* Size of objects. */
	uint32 item2Size,				/* Size of objects. */
	queue_compare_func compare_func1,/* Comparision routine for item1. */
	queue_compare_func compare_func2/* Comparision routine for item2. */
);

/*-----------------------------------------------------------------
 * Name:	q_destroyQueue()
 *
 * Description:	Destroy a queue.  Further accesses to the queue will fail.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
void
q_destroyQueue(
	q_headNodePtr_t	head_ptr		/* queue head pointer */
);

/*-----------------------------------------------------------------
 * Name:	q_enQueue()
 *
 * Description:	Adds a new node to the 'end' of the queue.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK,
								 * Q_ERR_NO_AVAILABLE_NODE
								 */
q_enQueue(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2			/* pointer to user data */
);

/*-----------------------------------------------------------------
 * Name:	q_deQueue()
 *
 * Description:	removes top entry from the queue
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_EMPTY_QUEUE
								 */
q_deQueue(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2			/* pointer to user data */
);

#define q_isQueueEmpty(_q_ptr)		(((_q_ptr)->size == 0) ? 1 : 0)

/*-----------------------------------------------------------------
 * Name:	q_resetReadPointer()
 *
 * Description:	The internal pointer can be reset back to the head of
 *				the queue with this function.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_QUEUE_EMPTY
								 */
q_resetReadPointer(
	q_headNodePtr_t	head_ptr		/* queue head pointer */
);

/*-----------------------------------------------------------------
 * Name:	q_previewQueueItem()
 *
 * Description:	Reads the user data, but does not alter the read pointer
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/25/01
 *
 *-----------------------------------------------------------------
 */
extern int32					/* NFX_OK
								 * Q_ERR_EMPTY_QUEUE
								 */
q_previewQueueItem(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2			/* pointer to user data */
);

/*-----------------------------------------------------------------
 * Name:	q_findQueueItem()
 *
 * Description:	This function searches the list for a specific item.
 *				If Q_FIND_ONE is used, we search for item1 and return
 *				the found value in item2.  The opposite is done for 
 *				Q_FIND_TWO.  Q_FIND_BOTH searches for both and doesn't
 *				update either item.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
extern int32					/* NFX_OK
								 * Q_ERR_ITEM_NOT_PRESENT
								 */
q_findQueueItem(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2,			/* pointer to user data */
	uint32			control			/* Q_FIND_ONE, Q_FIND_TWO, Q_FIND_BOTH */
);

/*-----------------------------------------------------------------
 * Name:	q_readQueueItem()
 *
 * Description:	The user data within a found node can be read using this
 *				function.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_EMPTY_QUEUE
								 */
q_readQueueItem(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2			/* pointer to user data */
);

/*-----------------------------------------------------------------
 * Name:	q_readNextQueueItem()
 *
 * Description:	The current read pointer can be advanced and the user data
 *				from the next node read.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_EMPTY_QUEUE
								 */
q_readNextQueueItem(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2			/* pointer to user data */
);

/*-----------------------------------------------------------------
 * Name:	q_deleteQueueItem()
 *
 * Description:	This function deletes the node pointed to be the read pointer.
 *				The read pointer was set via the findQueueItem function.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_EMPTY_QUEUE
								 */
q_deleteQueueItem(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2			/* pointer to user data */
);

/*-----------------------------------------------------------------
 * Name:	q_changeQueueItem()
 *
 * Description:	After finding a node with a specific item, the data inside
 *				that node can be changed.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_EMPTY_QUEUE
								 */
q_changeQueueItem(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2			/* pointer to user data */
);

/*-----------------------------------------------------------------
 * Name:	q_insertBeforeItem()
 *
 * Description:	A new node can be inserted into the list before a found node.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_EMPTY_QUEUE
								 */
q_insertBeforeItem(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2			/* pointer to user data */
);
	
/*-----------------------------------------------------------------
 * Name:	q_insertAfterItem()
 *
 * Description:	A new node can be inserted into the list after a found node.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_NO_AVAILABLE_NODE
								 */
q_insertAfterItem(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2			/* pointer to user data */
);
	
/*-----------------------------------------------------------------
 * Name:	q_insertItemInOrder()
 *
 * Description:	This is essentially a combination of q_findQueueItem() and
 *				q_insertBeforeItem().  Either item1 or item2 can be used
 *				as the comparison key, BUT NOT BOTH.
 *				The caller should supply a comparison function which
 *				will be called to compare the node item with the stored item.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */
int32							/* NFX_OK
								 * Q_ERR_NO_AVAILABLE_NODE
								 */
q_insertItemInOrder(
	q_headNodePtr_t	head_ptr,		/* queue head pointer */
	void *  		item1,			/* pointer to user data */
	void *  		item2,			/* pointer to user data */
	uint32			control			/* Q_FIND_ONE or Q_FIND_TWO
									 *			and
									 * Q_INSERT_ORDER_ASCENDING or
									 * Q_INSERT_ORDER_DESCENDING
									 */
);

extern void q_displayStats(void);

/*=================================================================
 * MACROS for managing doubly-linked lists.
 *=================================================================*/

/* Include debug.h for ASSERT macro */
#include <exports/debug.h>

/*-----------------------------------------------------------------
 * List header.
 *-----------------------------------------------------------------
 */
#define DLL_HDR(_type_)                         \
struct                                          \
{                                               \
    _type_ *head;                               \
    _type_ *tail;                               \
}

/*-----------------------------------------------------------------
 * List element.
 *-----------------------------------------------------------------
 */
#define DLL_ENTRY(_type_)                       \
struct                                          \
{                                               \
    _type_ *next;                               \
    _type_ *prev;                               \
}

/*-----------------------------------------------------------------
 * A link to a cell past the start or end of the list.  This value is
 * used in the next cell of the last element of the list and the prev
 * cell of the first element, along with the head and tail pointers
 * of the list header.
 *
 * We should use NULL for this value, but that's reserved for elements
 * which are not on the list.  Thus, a NULL appearing in the head or
 * tail of the list header is always in error.
 *-----------------------------------------------------------------
 */
#define DLL_NIL      ((void *) 1)

/*-----------------------------------------------------------------
 * Init list header.
 *-----------------------------------------------------------------
 */
#define DLL_INIT(_hdr_)                      \
    do {                                     \
        (_hdr_)->head = DLL_NIL;             \
        (_hdr_)->tail = DLL_NIL;             \
    } while (0)

/*-----------------------------------------------------------------
 * Check for empty condition.
 *-----------------------------------------------------------------
 */
#define DLL_IS_EMPTY(_hdr_)                  \
    ((_hdr_)->head == DLL_NIL)

/*-----------------------------------------------------------------
 * Isolate element.
 *-----------------------------------------------------------------
 */
#define DLL_ISOLATE_ELEM(_elem_,linkName)    \
    do {                                     \
        (_elem_)->linkName.next = NULL;      \
        (_elem_)->linkName.prev = NULL;      \
    } while (0)

/*-----------------------------------------------------------------
 * Return true if the element is not on any list
 *-----------------------------------------------------------------
 */
#define DLL_IS_ISOLATED(_elem_, linkName)                   \
    (((_elem_)->linkName.next == NULL) && ((_elem_)->linkName.prev == NULL))

/*-----------------------------------------------------------------
 * Insert element at head of list.
 *-----------------------------------------------------------------
 */
#define DLL_INSERT_HEAD(_hdr_,_elem_,linkName)       \
    do {                                             \
        ASSERT((_hdr_)->head != NULL);               \
        ASSERT((_hdr_)->tail != NULL);               \
        ASSERT((_elem_)->linkName.next == NULL);     \
        ASSERT((_elem_)->linkName.prev == NULL);     \
        (_elem_)->linkName.next = (_hdr_)->head;     \
        (_elem_)->linkName.prev = DLL_NIL;           \
        if ((_hdr_)->tail == DLL_NIL)                \
        {                                            \
            ASSERT((_hdr_)->head == DLL_NIL);        \
            (_hdr_)->tail = (_elem_);                \
        }                                            \
        else                                         \
        {                                            \
            ASSERT((_hdr_)->head != DLL_NIL);        \
            (_hdr_)->head->linkName.prev = (_elem_); \
        }                                            \
        (_hdr_)->head = (_elem_);                    \
        ASSERT((_hdr_)->head != DLL_NIL);            \
        ASSERT((_hdr_)->tail != DLL_NIL);            \
    } while (0)

/*-----------------------------------------------------------------
 * Insert element at tail of list.
 *-----------------------------------------------------------------
 */
#define DLL_INSERT_TAIL(_hdr_,_elem_,linkName)       \
    do {                                             \
        ASSERT((_hdr_)->head != NULL);               \
        ASSERT((_hdr_)->tail != NULL);               \
        ASSERT((_elem_)->linkName.next == NULL);     \
        ASSERT((_elem_)->linkName.prev == NULL);     \
        (_elem_)->linkName.next = DLL_NIL;           \
        (_elem_)->linkName.prev = (_hdr_)->tail;     \
        if ((_hdr_)->head == DLL_NIL)                \
        {                                            \
            ASSERT((_hdr_)->tail == DLL_NIL);        \
            (_hdr_)->head = (_elem_);                \
        }                                            \
        else                                         \
        {                                            \
            ASSERT((_hdr_)->tail != DLL_NIL);        \
            (_hdr_)->tail->linkName.next = (_elem_); \
        }                                            \
        (_hdr_)->tail = (_elem_);                    \
        ASSERT((_hdr_)->head != DLL_NIL);            \
        ASSERT((_hdr_)->tail != DLL_NIL);            \
    } while (0)

/*-----------------------------------------------------------------
 * Remove element from head of list.
 *-----------------------------------------------------------------
 */
#define DLL_REMOVE_HEAD(_hdr_,_ptr_,linkName)                \
    do {                                                     \
        ASSERT((_hdr_)->head != NULL);                       \
        ASSERT((_hdr_)->tail != NULL);                       \
        if ((_hdr_)->head == DLL_NIL)                        \
        {                                                    \
            ASSERT((_hdr_)->tail == DLL_NIL);                \
            (_ptr_) = NULL;                                  \
        }                                                    \
        else                                                 \
        {                                                    \
            ASSERT((_hdr_)->tail != DLL_NIL);                \
            (_ptr_) = (_hdr_)->head;                         \
            (_hdr_)->head = (_ptr_)->linkName.next;          \
            ASSERT((_ptr_)->linkName.next != NULL);          \
            ASSERT((_ptr_)->linkName.prev != NULL);          \
            if ((_hdr_)->head == DLL_NIL)                    \
            {                                                \
                (_hdr_)->tail = DLL_NIL;                     \
            }                                                \
            else                                             \
            {                                                \
                (_hdr_)->head->linkName.prev = DLL_NIL;      \
            }                                                \
            DLL_ISOLATE_ELEM((_ptr_),linkName);              \
        }                                                    \
        ASSERT(((_hdr_)->head != DLL_NIL &&                  \
                (_hdr_)->tail != DLL_NIL) ||                 \
               ((_hdr_)->head == DLL_NIL &&                  \
                (_hdr_)->tail == DLL_NIL));                  \
    } while (0)

/*-----------------------------------------------------------------
 * Remove element from tail of list.
 *-----------------------------------------------------------------
 */
#define DLL_REMOVE_TAIL(_hdr_,_ptr_,linkName)                \
    do {                                                     \
        ASSERT((_hdr_)->head != NULL);                       \
        ASSERT((_hdr_)->tail != NULL);                       \
        if ((_hdr_)->tail == DLL_NIL)                        \
        {                                                    \
            ASSERT((_hdr_)->head == DLL_NIL);                \
            (_ptr_) = NULL;                                  \
        }                                                    \
        else                                                 \
        {                                                    \
            ASSERT((_hdr_)->head != DLL_NIL);                \
            (_ptr_) = (_hdr_)->tail;                         \
            (_hdr_)->tail = (_ptr_)->linkName.prev;          \
            ASSERT((_ptr_)->linkName.next != NULL);          \
            ASSERT((_ptr_)->linkName.prev != NULL);          \
            if ((_hdr_)->tail == DLL_NIL)                    \
            {                                                \
                (_hdr_)->head = DLL_NIL;                     \
            }                                                \
            else                                             \
            {                                                \
                (_hdr_)->tail->linkName.next = DLL_NIL;      \
            }                                                \
            DLL_ISOLATE_ELEM((_ptr_),linkName);              \
        }                                                    \
        ASSERT(((_hdr_)->head != DLL_NIL &&                  \
                (_hdr_)->tail != DLL_NIL) ||                 \
               ((_hdr_)->head == DLL_NIL &&                  \
                (_hdr_)->tail == DLL_NIL));                  \
    } while (0)

/*-----------------------------------------------------------------
 * Insert after element.
 *-----------------------------------------------------------------
 */
#define DLL_INSERT_AFTER(_hdr_,_which_,_elem_,linkName)         \
    do {                                                        \
        ASSERT((_hdr_)->head != NULL);                          \
        ASSERT((_hdr_)->tail != NULL);                          \
        ASSERT((_elem_)->linkName.next == NULL);              	\
        ASSERT((_elem_)->linkName.prev == NULL);              	\
        ASSERT((_which_)->linkName.next != NULL);              	\
        ASSERT((_which_)->linkName.prev != NULL);              	\
        (_elem_)->linkName.next = (_which_)->linkName.next;     \
        (_elem_)->linkName.prev = (_which_);                    \
        if ((_which_)->linkName.next == DLL_NIL)                \
        {                                                       \
            (_hdr_)->tail = (_elem_);                           \
        }                                                       \
        else                                                    \
        {                                                       \
            (_which_)->linkName.next->linkName.prev = (_elem_); \
        }                                                       \
        (_which_)->linkName.next = (_elem_);                    \
        ASSERT((_hdr_)->head != DLL_NIL);                       \
        ASSERT((_hdr_)->tail != DLL_NIL);                       \
    } while (0)

/*-----------------------------------------------------------------
 * Insert before element.
 *-----------------------------------------------------------------
 */
#define DLL_INSERT_BEFORE(_hdr_,_which_,_elem_,linkName)        \
    do {                                                        \
        ASSERT((_hdr_)->head != NULL);                          \
        ASSERT((_hdr_)->tail != NULL);                          \
        ASSERT((_elem_)->linkName.next == NULL);              	\
        ASSERT((_elem_)->linkName.prev == NULL);              	\
        ASSERT((_which_)->linkName.next != NULL);              	\
        ASSERT((_which_)->linkName.prev != NULL);              	\
        (_elem_)->linkName.next = (_which_);                    \
        (_elem_)->linkName.prev = (_which_)->linkName.prev;     \
        if ((_which_)->linkName.prev == DLL_NIL)                \
        {                                                       \
            (_hdr_)->head = (_elem_);                           \
        }                                                       \
        else                                                    \
        {                                                       \
            (_which_)->linkName.prev->linkName.next = (_elem_); \
        }                                                       \
        (_which_)->linkName.prev = (_elem_);                    \
        ASSERT((_hdr_)->head != DLL_NIL);                       \
        ASSERT((_hdr_)->tail != DLL_NIL);                       \
    } while (0)

/*-----------------------------------------------------------------
 * Remove element.
 *-----------------------------------------------------------------
 */
#define DLL_REMOVE(_hdr_,_elem_,linkName)                       \
    do {                                                        \
        ASSERT((_hdr_)->head != NULL);                          \
        ASSERT((_hdr_)->tail != NULL);                          \
        ASSERT((_elem_)->linkName.next != NULL);                \
        ASSERT((_elem_)->linkName.prev != NULL);                \
        if ((_elem_)->linkName.next == DLL_NIL)                 \
        {                                                       \
            (_hdr_)->tail = (_elem_)->linkName.prev;            \
        }                                                       \
        else                                                    \
        {                                                       \
            (_elem_)->linkName.next->linkName.prev =            \
                                (_elem_)->linkName.prev;        \
        }                                                       \
        if ((_elem_)->linkName.prev == DLL_NIL)                 \
        {                                                       \
            (_hdr_)->head = (_elem_)->linkName.next;            \
        }                                                       \
        else                                                    \
        {                                                       \
            (_elem_)->linkName.prev->linkName.next =            \
                                (_elem_)->linkName.next;        \
        }                                                       \
        DLL_ISOLATE_ELEM((_elem_),linkName);                    \
        ASSERT(((_hdr_)->head != DLL_NIL &&                     \
                (_hdr_)->tail != DLL_NIL) ||                    \
               ((_hdr_)->head == DLL_NIL &&                     \
                (_hdr_)->tail == DLL_NIL));                     \
    } while (0)

/*-----------------------------------------------------------------
 * Move element to head of list.
 *-----------------------------------------------------------------
 */
#define DLL_MOVE_TO_HEAD(_hdr_,_elem_,linkName)                 \
    do {                                                        \
        ASSERT((_hdr_)->head != NULL);                          \
        ASSERT((_hdr_)->tail != NULL);                          \
        ASSERT((_elem_)->linkName.next != NULL);              	\
        ASSERT((_elem_)->linkName.prev != NULL);              	\
        if ((_elem_)->linkName.prev != DLL_NIL)                 \
        {                                                       \
            DLL_REMOVE(_hdr_, _elem_, linkName);                \
            DLL_INSERT_HEAD(_hdr_, _elem_, linkName);           \
        }                                                       \
        ASSERT((_hdr_)->head != DLL_NIL);                       \
        ASSERT((_hdr_)->tail != DLL_NIL);                       \
    } while (0)

/*-----------------------------------------------------------------
 * Move element to tail of list.
 *-----------------------------------------------------------------
 */
#define DLL_MOVE_TO_TAIL(_hdr_,_elem_,linkName)                 \
    do {                                                        \
        ASSERT((_hdr_)->head != NULL);                          \
        ASSERT((_hdr_)->tail != NULL);                          \
        ASSERT((_elem_)->linkName.next != NULL);              	\
        ASSERT((_elem_)->linkName.prev != NULL);              	\
        if ((_elem_)->linkName.next != DLL_NIL)                 \
        {                                                       \
            (_elem_)->linkName.next->linkName.prev =            \
                                    (_elem_)->linkName.prev;    \
            if ((_elem_)->linkName.prev == DLL_NIL)             \
            {                                                   \
                (_hdr_)->head = (_elem_)->linkName.next;        \
            }                                                   \
            else                                                \
            {                                                   \
                (_elem_)->linkName.prev->linkName.next =        \
                                    (_elem_)->linkName.next;    \
            }                                                   \
            (_elem_)->linkName.next = DLL_NIL;                  \
            (_elem_)->linkName.prev = (_hdr_)->tail;            \
            (_hdr_)->tail->linkName.next = (_elem_);            \
            (_hdr_)->tail = (_elem_);                           \
        }                                                       \
        ASSERT((_hdr_)->head != DLL_NIL);                       \
        ASSERT((_hdr_)->tail != DLL_NIL);                       \
    } while (0)



/*-----------------------------------------------------------------
 * Iterate through the elements of doubly-linked list.
 * (This macro must be used to control the loop body.)
 *-----------------------------------------------------------------
 */
#define DLL_ITERATE(_ivar_, _hdr_, linkName)    \
    ASSERT((_hdr_)->head != NULL);              \
    ASSERT((_hdr_)->tail != NULL);              \
    for (_ivar_ = (_hdr_)->head;                \
         (_ivar_) != DLL_NIL;                   \
         _ivar_ = _ivar_->linkName.next)

/*-----------------------------------------------------------------
 * Iterate through the elements of doubly-linked list with the next
 * link managed explicitly by the loop.
 * (This macro must be used to control the loop body.)
 *
 * The contortions used to allow dereferencing the next pointer if it's
 * potentially nul -- the combination of ?: with a comma expression --
 * is illegal in most industrial democracies, yet is tolerated in a few
 * primitive societies.
 *-----------------------------------------------------------------
 */
#define DLL_ITERATE_NEXT(_ivar_, _nvar_, _hdr_, linkName)                 \
    ASSERT((_hdr_)->head != NULL);                                        \
    ASSERT((_hdr_)->tail != NULL);                                        \
    for (_ivar_ = (_hdr_)->head;                                          \
         (_ivar_) != DLL_NIL ? ((_nvar_ = _ivar_->linkName.next), 1) : 0; \
         _ivar_ = _nvar_)

/*-----------------------------------------------------------------
 * Iterate through the elements of doubly-linked list
 * starting from the tail and working backward.
 * (This macro must be used to control the loop body.)
 *-----------------------------------------------------------------
 */
#define DLL_ITERATE_BACK(_ivar_, _hdr_, linkName)       \
    ASSERT((_hdr_)->head != NULL);                      \
    ASSERT((_hdr_)->tail != NULL);                      \
    for (_ivar_ = (_hdr_)->tail;                        \
         (_ivar_) != DLL_NIL;                           \
         _ivar_ = _ivar_->linkName.prev)

#endif	/* _QUEUE_API_H */
