/*
 *------------------------------------------------------------------------------
 *
 * Name:	raid-ui.c
 *
 * RCS:		$Id: raid-ui.c,v 1.32 2002/10/23 22:18:02 nitinb Exp $
 *
 * Copyright (C) 2001, 2002 Agile Storage Inc.
 *
 * Description:	RAID config manager UI.
 *
 * Created by:	Radek Aster
 *
 * Date Created: 4/17/02
 *
 *------------------------------------------------------------------------------
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <errno.h>

#include <nfx-incl.h>
#include <nfx-types.h>
#include <exports/eee-api.h>
#include <exports/scsi-api.h>
#include <exports/anpssc-api.h>
#include <exports/scsi-dev-cmds.h>
#include "raid-api.h"
#include "raid-parse-logpage.h"

#define	SECTOR_SIZE		512
#define RAID_UI_DEV_FILE	"/tmp/.raid-ui.devname"

int send_raid_message(void *, int, void *, int *, int);
char *raid_ecode_to_string(uint32);
char *raid_ui_state_to_str(uint32);
char *raid_ui_physdev_state_to_str(uint32);
char *raid_ui_logdev_state_to_str(uint32);
char *raid_ui_raid_level_to_str(uint32);

void raid_ui_set_devname(char **);
int  raid_ui_get_devname(int, char **);
int  raid_ui_is_devname(char *);

int raid_ui_devlist(int, char **, int, ...);
int raid_ui_open(int, char **, int, ...);
int raid_ui_close(int, char **, int, ...);
int raid_ui_ctrl_reset(int, char **, int, ...);
int raid_ui_ctrl_state(int, char **, int, ...);
int raid_ui_ctrl_flush_cache(int, char **, int, ...);
int raid_ui_ctrl_info(int, char **, int, ...);
int raid_ui_ctrl_parm(int, char **, int, ...);
int raid_ui_ctrl_parm_set(int, char **, int, ...);
int raid_ui_physdev(int, char **, int, ...);
int raid_ui_physdev_stats(int, char **, int, ...);
int raid_ui_config_clr(int, char **, int, ...);
int raid_ui_config_save(int, char **, int, ...);
int raid_ui_config_load(int, char **, int, ...);
int raid_ui_lun_list(int, char **, int, ...);
int raid_ui_lun_bg_status(int, char **, int, ...);
int raid_ui_lun_create_simple(int, char **, int, ...);
int raid_ui_lun_create(int, char **, int, ...);
int raid_ui_lun_delete(int, char **, int, ...);
int raid_ui_lun_init(int, char **, int, ...);
int raid_ui_lun_check(int, char **, int, ...);
int raid_ui_lun_physlist(int, char **, int, ...);
int raid_ui_lun_cache(int, char **, int, ...);
int raid_ui_dbg_level(int, char **, int, ...);
int raid_ui_get_sizes(char *devname, uint64 *phySpace, uint64 *freeSpace, uint64 *usableSpace);
int raid_ui_get_logpage(int , char **, int , ...);
int raid_ui_direct_read(int, char **, int, ...); 
int raid_ui_corrupt_bits(int, char **, int, ...);
int raid_ui_get_modepage(int, char **, int, ...);
int raid_ui_update_drv_fw(int, char **,int, ...);

int raid_ui_do_direct_read(int, char **, int, char *, int);
int raid_ui_do_direct_write(int, char **, int, char *, int);
		
typedef struct {
	char *	name;
	int	parm_cnt;	/* Min command-line parameter count */
	int (*func)(int, char **, int, ...);
	int	timeout;
	char *	usage;
} raid_ui_node_t;

raid_ui_node_t raid_ui_root[] = {
	{ "devlist",	      0,  raid_ui_devlist,		5,	  
	  "devlist" },
	{ "open",	      1,  raid_ui_open,			60,	  
	  "open <device-name>" },
	{ "close",	      1,  raid_ui_close,		60,	  
	  "close <device-name>" },
	{ "ctrl_reset",	      1,  raid_ui_ctrl_reset,		60,	  
	  "ctrl_reset <device-name>" },
	{ "ctrl_state",	      1,  raid_ui_ctrl_state,		60,	  
	  "ctrl_state <device-name>" },
	{ "ctrl_flush_cache", 1,  raid_ui_ctrl_flush_cache,	60,	  
	  "ctrl_flush_cache <device-name>" },
	{ "ctrl_info",	      1,  raid_ui_ctrl_info,		60,	  
	  "ctrl_info <device-name>" },
	{ "ctrl_parm",	      1,  raid_ui_ctrl_parm,		60,	  
	  "ctrl_parm <device-name>" },
	{ "ctrl_parm_set",    1,  raid_ui_ctrl_parm_set,	60,	  
	  "ctrl_parm_set <device-name>" },
	{ "physdev",	      1,  raid_ui_physdev,		60,	  
	  "physdev <device-name>" },
	{ "physdev_stats",    1,  raid_ui_physdev_stats,	60,	  
	  "physdev_stats <device-name>" },
	{ "config_clr",       1,  raid_ui_config_clr,		60, 
	  "config_clr <device-name>" },
	{ "config_save",      2,  raid_ui_config_save,		60, 
	  "config_save <device-name> <filename>" },
	{ "config_load",      2,  raid_ui_config_load,		60, 
	  "config_load <device-name> <filename>" },
	{ "lun_list",	      1,  raid_ui_lun_list,		60,	  
	  "lun_list <device-name>" },
	{ "lun_bg_status",    1,  raid_ui_lun_bg_status,	60,	  
	  "lun_bg_status <device-name>" },
	{ "lun_create_simple",5,  raid_ui_lun_create_simple,	60,	  
	  "lun_create_simple <device-name> <lun-count> <raid-type> <size-in-MB> <stripesize-in-KB>" },
	{ "lun_create",       5,  raid_ui_lun_create,		60,	  
	  "lun_create <device-name> <raid-type> <size-in-MB> <stripesize-in-KB> <cid>/<tid> ..." },
	{ "lun_delete",       2,  raid_ui_lun_delete,		60,	  
	  "lun_delete <device-name> <LUN-id>" },
	{ "lun_init",         2,  raid_ui_lun_init,		60,	  
	  "lun_init <device-name> <LUN-id>|ALL|ALLNOT" },
	{ "lun_check",        2,  raid_ui_lun_check,		60,	  
	  "lun_check <device-name> <LUN-id>" },
	{ "lun_physlist",     2,  raid_ui_lun_physlist,		60,	  
	  "lun_physlist <device-name> <LUN-id>" },
	{ "lun_cache",        4,  raid_ui_lun_cache,		60,	  
	  "lun_cache <device-name> <LUN-id> <Read cache ON|OFF> <Write cache ON|OFF>" },
	{ "dbg_level",	      1,  raid_ui_dbg_level,	        5,	  
	  "dbg_level <value> -- set driver debug level" },
	{ "get_logpage",      3,  raid_ui_get_logpage,		60,
	  "get_logpage <device-name> <CID> <TID> [page-code]" },
	{ "direct_read",      4,  raid_ui_direct_read,		60,
	  "direct_read <device-name> <CID> <TID> <lba> [size]" },
	{ "modeselect",	  	  4,  raid_ui_get_modepage,     60,
	  "modeselect <device-name> <CID> <TID> <page-code>" },
	{ "update_drv_fw",	  2,  raid_ui_update_drv_fw,	180,
	  "update_drv_fw <device-name> <firmware-file> [CID] [TID]" },
	{ NULL,		      0,  NULL,				0,	  
	  NULL }
};
	
static void 
print_usage(char *command)
{
	int i;

	for (i = 0; raid_ui_root[i].name; ++i) {
		fprintf(stderr, "%s %s %s\n",
			(i == 0 ? "Usage:" : "      "),
			command, raid_ui_root[i].usage);
	}
}

int
raid_ui_devlist(int argc, char **argv, int timeout, ...)
{
	int			     rc;
	char			     req_msg[1024];
	char			     rsp_msg[16 * 1024];
	int			     req_len = 0;
	int			     rsp_len = 0;
	raid_getListReqMsg_t *	     devlist_req;
	raid_getListRspMsg_t *	     devlist_rsp;
	raid_getInfoReqMsg_t 	     devinfo_req;
	raid_getInfoRspMsg_t 	     devinfo_rsp;
	int			     i;

	devlist_req = (raid_getListReqMsg_t *)req_msg;
	devlist_req->header.mid = RAID_MID_GET_LIST_REQ;
	devlist_req->header.cookie = 1;
			
	req_len = sizeof(raid_getListReqMsg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_raid_message(req_msg, req_len, rsp_msg, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", devlist_req->header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", devlist_req->header.mid);
		return(NFX_ERR);
	}

	devlist_rsp = (raid_getListRspMsg_t *)rsp_msg;
	if (devlist_rsp->header.mid != RAID_MID_GET_LIST_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_GET_LIST_RSP, devlist_rsp->header.mid);
		return(NFX_ERR); 
	}
	if (rsp_len != sizeof(raid_getListRspMsg_t) + sizeof(raid_ctrl_id_t) * (devlist_rsp->numEntries - 1)) {
		printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
		       devlist_req->header.mid, 
		       sizeof(raid_getListRspMsg_t) + sizeof(raid_ctrl_id_t) * (devlist_rsp->numEntries - 1),
		       rsp_len);
		return(NFX_ERR);
	}
	if (devlist_rsp->rCode != RAID_RCODE_SUCCESS) {
		printf("raid devlist failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(devlist_rsp->rCode), devlist_rsp->rCode);
		return(NFX_ERR);
	}

	printf("%d devices found.\n", devlist_rsp->numEntries);
	printf("ID  DEVICE NAME                      WWN              LUN STATE     \n");
	printf("--- -------------------------------- ---------------- --- ---------------\n");
	for (i = 0; i < devlist_rsp->numEntries; ++i) {
		/* XXX get info on each device */
		bzero(&devinfo_req, sizeof(raid_getInfoReqMsg_t));
		devinfo_req.header.mid = RAID_MID_GET_INFO_REQ;
		devinfo_req.header.cookie = 1;
		memcpy(devinfo_req.header.id, devlist_rsp->idList[i], sizeof(raid_ctrl_id_t));
		req_len = sizeof(raid_getInfoReqMsg_t);
		rsp_len = sizeof(raid_getInfoRspMsg_t);
		rc = send_raid_message(&devinfo_req, req_len, &devinfo_rsp, &rsp_len, timeout);
		if (rc < 0) {
			printf("send_raid_message(mid 0x%x) failed.\n", devinfo_req.header.mid);
			return(NFX_ERR);
		}
		else 
			if (rc == 0) {
				printf("send_raid_message(mid 0x%x) timed out.\n", devinfo_req.header.mid);
				return(NFX_ERR);
			}

		if (devinfo_rsp.header.mid != RAID_MID_GET_INFO_RSP) {
			printf("Invalid response mid; expected 0x%x got 0x%x\n",
			       RAID_MID_GET_INFO_RSP, devinfo_rsp.header.mid);
			return(NFX_ERR); 
		}
		if (rsp_len != sizeof(raid_getInfoRspMsg_t)) {
			printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
			       devinfo_req.header.mid, 
			       sizeof(raid_getInfoRspMsg_t), rsp_len);
			return(NFX_ERR);
		}
		if (devinfo_rsp.rCode != RAID_RCODE_SUCCESS) {
			printf("raid getinfo failed: rCode %s(0x%x)\n", 
			       raid_ecode_to_string(devinfo_rsp.rCode), devinfo_rsp.rCode);
			return(NFX_ERR);
		}

		printf("%-3d %-32s %016llx %-3d %-15s\n", i, devlist_rsp->idList[i], 
		       devinfo_rsp.wwn, devinfo_rsp.lun, raid_ui_state_to_str(devinfo_rsp.state));

	} 
	return(NFX_OK);
}

int
raid_ui_open(int argc, char **argv, int timeout, ...)
{
	int			rc;
	char			req_msg[1024];
	char			rsp_msg[1024];
	int			req_len = 0;
	int			rsp_len = 0;
	raid_openReqMsg_t *	open_req;
	raid_openRspMsg_t *	open_rsp;

	open_req = (raid_openReqMsg_t *)req_msg;
	open_req->header.mid = RAID_MID_OPEN_REQ;
	open_req->header.cookie = 1;
	strcpy(open_req->header.id, argv[1]);
			
	req_len = sizeof(raid_openReqMsg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_raid_message(req_msg, req_len, rsp_msg, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", open_req->header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", open_req->header.mid);
		return(NFX_ERR);
	}

	open_rsp = (raid_openRspMsg_t *)rsp_msg;
	if (open_rsp->header.mid != RAID_MID_OPEN_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_OPEN_RSP, open_rsp->header.mid);
		return(NFX_ERR);
	}
	if (rsp_len != sizeof(raid_openRspMsg_t)) {
		printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
		       open_req->header.mid, sizeof(raid_openRspMsg_t), rsp_len);
		return(NFX_ERR);
	}
	if (open_rsp->rCode != RAID_RCODE_SUCCESS) {
		printf("raid open failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(open_rsp->rCode), open_rsp->rCode);
		return(NFX_ERR);
	}

	printf("SCSI ID              : %016llx\n", open_rsp->scsi_id);
	printf("WWN                  : %016llx\n", open_rsp->wwn);
	printf("LUN                  : %d\n", open_rsp->lun);
	return(NFX_OK);
}

int
raid_ui_close(int argc, char **argv, int timeout, ...)
{
	int			rc;
	char			req_msg[1024];
	char			rsp_msg[1024];
	int			req_len = 0;
	int			rsp_len = 0;
	raid_closeReqMsg_t *	close_req;
	raid_closeRspMsg_t *	close_rsp;

	close_req = (raid_closeReqMsg_t *)req_msg;
	close_req->header.mid = RAID_MID_CLOSE_REQ;
	close_req->header.cookie = 1;
	strcpy(close_req->header.id, argv[1]);
			
	req_len = sizeof(raid_closeReqMsg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_raid_message(req_msg, req_len, rsp_msg, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", close_req->header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", close_req->header.mid);
		return(NFX_ERR);
	}

	close_rsp = (raid_closeRspMsg_t *)rsp_msg;
	if (close_rsp->header.mid != RAID_MID_CLOSE_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_CLOSE_RSP, close_rsp->header.mid);
		return(NFX_ERR);
	}
	if (rsp_len != sizeof(raid_closeRspMsg_t)) {
		printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
		       close_req->header.mid, sizeof(raid_closeRspMsg_t), rsp_len);
		return(NFX_ERR);
	}
	if (close_rsp->rCode != RAID_RCODE_SUCCESS) {
		printf("raid close failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(close_rsp->rCode), close_rsp->rCode);
		return(NFX_ERR);
	}

	return(NFX_OK);
}

int 
raid_ui_ctrl_info(int argc, char **argv, int timeout, ...)
{
	int			rc;
	int			req_len = 0;
	int			rsp_len = 0;
	raid_ctrlInfoReqMsg_t	req;
	raid_ctrlInfoRspMsg_t	rsp;
	uint64		freeSpace;
	uint64		phySpace;
	uint64		usableSpace;
	uint64		usedSpace;

	req.header.mid = RAID_MID_CTRL_INFO_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);
			
	req_len = sizeof(raid_ctrlInfoReqMsg_t);
	rsp_len = sizeof(raid_ctrlInfoRspMsg_t);
	rc = send_raid_message(&req, req_len, &rsp, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}

	if (rsp.header.mid != RAID_MID_CTRL_INFO_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_CTRL_INFO_RSP, rsp.header.mid);
		return(NFX_ERR);
	}
	if (rsp.rCode != RAID_RCODE_SUCCESS) {
		printf("raid ctrl_info failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(rsp.rCode), rsp.rCode);
		return(NFX_ERR);
	}

	if(raid_ui_get_sizes(argv[1], &phySpace, &freeSpace, &usableSpace)) {
		return(NFX_ERR);
	}
	usedSpace = usableSpace - freeSpace;

	printf("ctrlType             %s\n", rsp.ctrlType);
	printf("ctrlName             %s\n", rsp.ctrlName);
	printf("serialNumber         %-s\n", rsp.serialNumber);
	printf("vendorName           %s\n", rsp.vendorName);
	printf("hardwareVer          %s\n", rsp.hardwareVer);
	printf("firmwareVer          %s\n", rsp.firmwareVer);
	printf("cacheInMB            %d\n", rsp.cacheInMB);
	printf("memoryInMB           %d\n", rsp.memoryInMB);
	printf("bbuPresent           %d\n", rsp.bbuPresent);
	printf("numDriveChannels     %d\n", rsp.numDriveChannels);
	printf("maxDrivesPerChannel  %d\n", rsp.maxDrivesPerChannel);
	printf("ctrlWwn              %llx\n", rsp.ctrlWwn);
	printf("dualCtrlWwn          %llx\n", rsp.dualCtrlWwn);
	printf("number of LUNs       %d\n", rsp.numLogicalDevices);
	printf("numPhysicalDevices   %d\n", rsp.numPhysicalDevices);
	if(usableSpace > 1000)
		printf("total space          %lld.%03lld GB\n", usableSpace/1000, usableSpace%1000);
	else
		printf("total space          %lld MB\n", usableSpace);
	if(usedSpace > 1000) 
		printf("used space           %lld.%03lld GB\n", usedSpace/1000, usedSpace%1000);
	else
		printf("used space           %lld MB\n", usedSpace);
	if(freeSpace > 1000)
		printf("available space      %lld.%03lld GB\n", freeSpace/1000, freeSpace%1000);
	else
		printf("available space      %lld MB\n", freeSpace);

	return(NFX_OK);
}

int 
raid_ui_ctrl_reset(int argc, char **argv, int timeout, ...)
{
	int			rc;
	int			req_len = 0;
	int			rsp_len = 0;
	raid_ctrlResetReqMsg_t	req;
	raid_ctrlResetRspMsg_t	rsp;

	req.header.mid = RAID_MID_CTRL_RESET_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);
			
	req_len = sizeof(raid_ctrlResetReqMsg_t);
	rsp_len = sizeof(raid_ctrlResetRspMsg_t);
	rc = send_raid_message(&req, req_len, &rsp, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}

	if (rsp.header.mid != RAID_MID_CTRL_RESET_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_CTRL_RESET_RSP, rsp.header.mid);
		return(NFX_ERR);
	}
	if (rsp.rCode != RAID_RCODE_SUCCESS) {
		printf("raid ctrl_reset failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(rsp.rCode), rsp.rCode);
		return(NFX_ERR);
	}

	return(NFX_OK);
}

int 
raid_ui_ctrl_state(int argc, char **argv, int timeout, ...)
{
	int				rc;
	int				req_len = 0;
	int				rsp_len = 0;
	raid_getCtrlStateReqMsg_t	req;
	raid_getCtrlStateRspMsg_t	rsp;

	req.header.mid = RAID_MID_CTRL_GETSTATE_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);
			
	req_len = sizeof(raid_getCtrlStateReqMsg_t);
	rsp_len = sizeof(raid_getCtrlStateRspMsg_t);
	rc = send_raid_message(&req, req_len, &rsp, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}

	if (rsp.header.mid != RAID_MID_CTRL_GETSTATE_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_CTRL_GETSTATE_RSP, rsp.header.mid);
		return(NFX_ERR);
	}
	if (rsp.rCode != RAID_RCODE_SUCCESS) {
		printf("raid ctrl_state failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(rsp.rCode), rsp.rCode);
		return(NFX_ERR);
	}

	printf("Master/Slave State:           %s\n", rsp.masterSlaveState);
	printf("Master/Slave Additional Info: %s\n", rsp.masterSlaveAdditionInfo);
	printf("Partner Status:               %s\n", rsp.parterState);

	return(NFX_OK);
}

int 
raid_ui_ctrl_flush_cache(int argc, char **argv, int timeout, ...)
{
	int				rc;
	int				req_len = 0;
	int				rsp_len = 0;
	raid_flushCtrlCacheReqMsg_t	req;
	raid_flushCtrlCacheRspMsg_t	rsp;

	req.header.mid = RAID_MID_CTRL_FLUCACHE_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);
			
	req_len = sizeof(raid_flushCtrlCacheReqMsg_t);
	rsp_len = sizeof(raid_flushCtrlCacheRspMsg_t);
	rc = send_raid_message(&req, req_len, &rsp, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}

	if (rsp.header.mid != RAID_MID_CTRL_FLUCACHE_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_CTRL_FLUCACHE_RSP, rsp.header.mid);
		return(NFX_ERR);
	}
	if (rsp.rCode != RAID_RCODE_SUCCESS) {
		printf("raid ctrl_flush_cache failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(rsp.rCode), rsp.rCode);
		return(NFX_ERR);
	}

	return(NFX_OK);
}

int 
raid_ui_ctrl_parm(int argc, char **argv, int timeout, ...)
{
	int				rc;
	int				req_len = 0;
	int				rsp_len = 0;
	raid_getCtrlParamReqMsg_t	req;
	raid_getCtrlParamRspMsg_t	rsp;
	raid_ctrlParam_t *		p;

	req.header.mid = RAID_MID_CTRL_GETPARAM_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);
			
	req_len = sizeof(raid_getCtrlParamReqMsg_t);
	rsp_len = sizeof(raid_getCtrlParamRspMsg_t);
	rc = send_raid_message(&req, req_len, &rsp, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}

	if (rsp.header.mid != RAID_MID_CTRL_GETPARAM_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_CTRL_GETPARAM_RSP, rsp.header.mid);
		return(NFX_ERR);
	}
	if (rsp.rCode != RAID_RCODE_SUCCESS) {
		printf("raid ctrl_parm failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(rsp.rCode), rsp.rCode);
		return(NFX_ERR);
	}

	p = &rsp.params;

	printf("Read Ahead Enabled                           %d\n", p->readAhead);
	printf("Reassign Restricted to 1 block Enabled:      %d\n", p->reasRestrict);
	printf("True Data Verification Enabled:              %d\n", p->trueVerify);
	printf("Write Thru Verification Enabled:             %d\n", p->writeVerify);
	printf("Background Init Enabled:                     %d\n", p->bgInit);
	printf("Super Read Ahead Enabled:                    %d\n", p->superReadAhead);
	printf("Coalescing optimization Enabled:             %d\n", p->coalOptimal);
	printf("Auto Rebuild Enabled:                        %d\n", p->autoRebuild);
	printf("Operation Fault Manage Enabled:              %d\n", p->operFault);
	printf("Rebuild Rate:                                %d\n", p->rebuildRate);
	printf("Background Init Rate:                        %d\n", p->initRate);
	printf("Background Consistency Check Rate:           %d\n", p->consistencyCheckRate);
	printf("Start Up Mode:                               %s\n", (p->startupMode == 0 ? "AUTO" : 
								    (p->startupMode == 1 ? "POWER" :
								    (p->startupMode == 2 ? "COMMAND" :
								    "???"))));
	printf("Start Up Number of Devices:                  %d\n", p->startupNumDev);
	printf("Start Up Delay:                              %d\n", p->startupDelay);
	printf("SCSI Start Up Delay:                         %d\n", p->scsiStartupDelay);
	printf("Vendor unique TUR Enabled:                   %d\n", p->vendorTUR);	
	printf("Check Condition for Invalid LUN Disabled:    %d\n", p->disCCInvLun);
	printf("No Pause on controller not ready Enabled:    %d\n", p->noPause);
	printf("Busy on Queue Full Enabled:                  %d\n", p->queueFullBusy);
	printf("Disable busy during Failback Enabled:        %d\n", p->disBusyFailback);
	printf("Node Name retention Enabled:                 %d\n", p->nodeNameRet);
	printf("SAFTE Monitoring of UPS Enabled:             %d\n", p->safteofups);
	printf("FC Topology:                                 %s\n", (p->topology == RAID_TOPO_INACTPORT ? "INACTIVE PORT (0)" :
								    (p->topology == RAID_TOPO_MULTIPORT ? "MULTIPORT (1)" :
								    (p->topology == RAID_TOPO_CLUSTER ? "CLUSTER (2)" :
								    (p->topology == RAID_TOPO_MULTITID ? "MULTI TID (3)" :
								    (p->topology == RAID_TOPO_MASSLAV ? "MASTER/SLAVE (4)" :
								    "???"))))));
	printf("Reset Propagation Enabled:                   %d\n", p->resetProp);
	printf("RS232 port type:                             %s\n", (p->rs232PortType == RAID_PORT_DEBUG ? "DEBUG" :
								    (p->rs232PortType == RAID_PORT_SLPVT100 ? "SLP/VT100" :
								    "???")));
	printf("RS232 port Baud Rate:                        %d\n", p->rs232BaudRate);
	printf("Smart Large Host Transfers Enabled:          %d\n", p->smartLrgTrans);
	printf("FC frame size:                               %d\n", p->frameSize);
	printf("PCI Latency:                                 %s\n", (p->pciLatCtrl == RAID_PLC_LONG ? "LONG" :
								    (p->pciLatCtrl == RAID_PLC_MEDIUM ? "MEDIUM" :
								    (p->pciLatCtrl == RAID_PLC_SHORT ? "SHORT" :
								    "???"))));
	printf("Auto Failback Enabled:                       %d\n", p->autoFailback);
	printf("Force Simplex Enabled:                       %d\n", p->forceSimplex);
	printf("Conservative Cache Mode Enabled:             %d\n", p->conserveCache);
	printf("Simplex no Reset COM Enabled:                %d\n", p->noRSTCOM);
	printf("Back End Traffic Queue Limit:                %d\n", p->queueLimit);
	printf("C0P0 Enabled:                                %d (0x%x)\n", p->ctrl0Port0, p->ctrl0Port0LoopId);
	printf("C0P1 Enabled:                                %d (0x%x)\n", p->ctrl0Port1, p->ctrl0Port1LoopId);
	printf("C1P0 Enabled:                                %d (0x%x)\n", p->ctrl1Port0, p->ctrl1Port0LoopId);
	printf("C1P1 Enabled:                                %d (0x%x)\n", p->ctrl1Port1, p->ctrl1Port1LoopId);
	printf("Debug Dump Enabled:                          %d\n", p->debugDump);
	printf("ROF Interval:                                %s\n", (p->rofInterval == RAID_ROFI_NEVER ? "NEVER" :
								    (p->rofInterval == RAID_ROFI_3min ? "3 MINS" :
								    (p->rofInterval == RAID_ROFI_5min ? "5 MINS" :
								    (p->rofInterval == RAID_ROFI_15min ? "15 MINS" :
								    (p->rofInterval == RAID_ROFI_30min ? "30 MINS" :
								    (p->rofInterval == RAID_ROFI_60min ? "60 MINS" :
								    (p->rofInterval == RAID_ROFI_90min ? "90 MINS" :
								    (p->rofInterval == RAID_ROFI_2hour ? "2 HRS" :
								    (p->rofInterval == RAID_ROFI_4hour ? "4 HRS" :
								    (p->rofInterval == RAID_ROFI_8hour ? "8 HRS" :
								    (p->rofInterval == RAID_ROFI_12hour ? "12 HRS" :
								    (p->rofInterval == RAID_ROFI_16hour ? "16 HRS" :
								    (p->rofInterval == RAID_ROFI_24hour ? "24 HRS" :
								    (p->rofInterval == RAID_ROFI_2day ? "2 DAYS" :
								    (p->rofInterval == RAID_ROFI_4day ? "4 DAYS" :
								    (p->rofInterval == RAID_ROFI_7day ? "7 DAYS" :
								    "???")))))))))))))))));
	printf("ROF Reboot Count:                            %d\n", p->rofReboot);

	return(NFX_OK);
}

int 
raid_ui_ctrl_parm_set(int argc, char **argv, int timeout, ...)
{
	int				rc;
	int				req_len = 0;
	int				rsp_len = 0;
	raid_getCtrlParamReqMsg_t	get_req;
	raid_getCtrlParamRspMsg_t	get_rsp;
	raid_setCtrlParamReqMsg_t	set_req;
	raid_setCtrlParamRspMsg_t	set_rsp;
	raid_ctrlParam_t *		p;

	get_req.header.mid = RAID_MID_CTRL_GETPARAM_REQ;
	get_req.header.cookie = 1;
	strcpy(get_req.header.id, argv[1]);
			
	req_len = sizeof(raid_getCtrlParamReqMsg_t);
	rsp_len = sizeof(raid_getCtrlParamRspMsg_t);
	rc = send_raid_message(&get_req, req_len, &get_rsp, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", get_req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", get_req.header.mid);
		return(NFX_ERR);
	}

	if (get_rsp.header.mid != RAID_MID_CTRL_GETPARAM_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_CTRL_GETPARAM_RSP, get_rsp.header.mid);
		return(NFX_ERR);
	}
	if (get_rsp.rCode != RAID_RCODE_SUCCESS) {
		printf("raid ctrl_parm_set failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(get_rsp.rCode), get_rsp.rCode);
		return(NFX_ERR);
	}

	memcpy(&set_req.params, &get_rsp.params, sizeof(raid_ctrlParam_t));

	/* Set our default parameters */
	p = &set_req.params;
	p->topology = RAID_TOPO_MULTIPORT;
	p->disCCInvLun = 0;

	set_req.header.mid = RAID_MID_CTRL_SETPARAM_REQ;
	set_req.header.cookie = 1;
	strcpy(set_req.header.id, argv[1]);
			
	req_len = sizeof(raid_setCtrlParamReqMsg_t);
	rsp_len = sizeof(raid_setCtrlParamRspMsg_t);
	rc = send_raid_message(&set_req, req_len, &set_rsp, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", set_req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", set_req.header.mid);
		return(NFX_ERR);
	}

	if (set_rsp.header.mid != RAID_MID_CTRL_SETPARAM_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_CTRL_SETPARAM_RSP, set_rsp.header.mid);
		return(NFX_ERR);
	}
	if (set_rsp.rCode != RAID_RCODE_SUCCESS) {
		printf("raid ctrl_parm_set failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(set_rsp.rCode), set_rsp.rCode);
		return(NFX_ERR);
	}

	return(NFX_OK);
}

int 
raid_ui_physdev(int argc, char **argv, int timeout, ...)
{
	int				rc;
	int				req_len = 0;
	int				rsp_len = 0;
	raid_phyDevListReqMsg_t		req;
	raid_phyDevListRspMsg_t	*	rsp;
	char				rspbuf[65536];
	int				i;
	raid_phyDevInfoReqMsg_t		info_req;
	raid_phyDevInfoRspMsg_t		info_rsp;


	req.header.mid = RAID_MID_PHYDEV_LIST_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);
			
	req_len = sizeof(raid_phyDevListReqMsg_t);
	rsp_len = sizeof(rspbuf);
	rc = send_raid_message(&req, req_len, rspbuf, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}
	rsp = (raid_phyDevListRspMsg_t *)rspbuf;
	if (rsp->header.mid != RAID_MID_PHYDEV_LIST_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_PHYDEV_LIST_RSP, rsp->header.mid);
		return(NFX_ERR);
	}
	if (rsp->rCode != RAID_RCODE_SUCCESS) {
		printf("raid physdev failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(rsp->rCode), rsp->rCode);
		return(NFX_ERR);
	}

	printf("CID TID ENC ROW COL RDN  VENDOR   MODEL            VERS SERIAL #         WWN              PHYS SZ ACT SZ  FREE SZ STATE   \n");
	printf("--- --- --- --- --- ---- -------- ---------------- ---- ---------------- ---------------- ------- ------- ------- --------\n");
	for (i = 0; i < rsp->numEntries; ++i) {
		info_req.header.mid = RAID_MID_PHYDEV_INFO_REQ;
		info_req.header.cookie = 1;
		strcpy(info_req.header.id, argv[1]);
		info_req.id.channelNumber = rsp->idList[i].channelNumber;
		info_req.id.targetId = rsp->idList[i].targetId;		
			
		req_len = sizeof(raid_phyDevInfoReqMsg_t);
		rsp_len = sizeof(raid_phyDevInfoRspMsg_t);
		rc = send_raid_message(&info_req, req_len, &info_rsp, &rsp_len, timeout);
		if (rc < 0) {
			printf("send_raid_message(mid 0x%x) failed.\n", info_req.header.mid);
			return(NFX_ERR);
		}
		else 
		if (rc == 0) {
			printf("send_raid_message(mid 0x%x) timed out.\n", info_req.header.mid);
			return(NFX_ERR);
		}
		if (info_rsp.header.mid != RAID_MID_PHYDEV_INFO_RSP) {
			printf("Invalid response mid; expected 0x%x got 0x%x\n",
			       RAID_MID_PHYDEV_INFO_RSP, info_rsp.header.mid);
			return(NFX_ERR);
		}
		if (info_rsp.rCode != RAID_RCODE_SUCCESS) {
			printf("raid physdev failed: rCode %s(0x%x)\n", 
			       raid_ecode_to_string(info_rsp.rCode), info_rsp.rCode);
			return(NFX_ERR);
		}
		printf("%3d %3d %3d %3d %3d %04x %8s %16s %4s %16s %016llx %3d,%03d %3d,%03d %3d,%03d %s\n", 
		       info_rsp.id.channelNumber, info_rsp.id.targetId, 
		       info_rsp.enclosure, info_rsp.row, info_rsp.column,
		       info_rsp.rdn,
		       info_rsp.vendorName, info_rsp.modelName, info_rsp.microcodeVer, info_rsp.serialNumber, info_rsp.wwn,
		       info_rsp.physicalSizeInMB/1000, info_rsp.physicalSizeInMB%1000,
		       info_rsp.usableSizeInMB/1000, info_rsp.usableSizeInMB%1000,
		       info_rsp.freeSizeInMB/1000, info_rsp.freeSizeInMB%1000,
		       raid_ui_physdev_state_to_str(info_rsp.deviceState));
	}

	return(NFX_OK);
}

int 
raid_ui_physdev_stats(int argc, char **argv, int timeout, ...)
{
	int				rc;
	int				req_len = 0;
	int				rsp_len = 0;
	raid_phyDevListReqMsg_t		req;
	raid_phyDevListRspMsg_t	*	rsp;
	char				rspbuf[65536];
	int				i;
	raid_phyDevStatsReqMsg_t	stats_req;
	raid_phyDevStatsRspMsg_t	stats_rsp;


	req.header.mid = RAID_MID_PHYDEV_LIST_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);
			
	req_len = sizeof(raid_phyDevListReqMsg_t);
	rsp_len = sizeof(rspbuf);
	rc = send_raid_message(&req, req_len, rspbuf, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}
	rsp = (raid_phyDevListRspMsg_t *)rspbuf;
	if (rsp->header.mid != RAID_MID_PHYDEV_LIST_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_PHYDEV_LIST_RSP, rsp->header.mid);
		return(NFX_ERR);
	}
	if (rsp->rCode != RAID_RCODE_SUCCESS) {
		printf("raid physdev_stats failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(rsp->rCode), rsp->rCode);
		return(NFX_ERR);
	}

	printf("CID TID UPTIME(mS) READS    WRITES   KB RD    KB WRT   SOFT ERR HARD ERR PART ERR MISC ERR\n");
	printf("--- --- ---------- -------- -------- -------- -------- -------- -------- -------- --------\n");

	for (i = 0; i < rsp->numEntries; ++i) {
		stats_req.header.mid = RAID_MID_PHYDEV_STATS_REQ;
		stats_req.header.cookie = 1;
		strcpy(stats_req.header.id, argv[1]);
		stats_req.id.channelNumber = rsp->idList[i].channelNumber;
		stats_req.id.targetId = rsp->idList[i].targetId;		
		stats_req.reset_on_read = 1;
			
		req_len = sizeof(raid_phyDevStatsReqMsg_t);
		rsp_len = sizeof(raid_phyDevStatsRspMsg_t);
		rc = send_raid_message(&stats_req, req_len, &stats_rsp, &rsp_len, timeout);
		if (rc < 0) {
			printf("send_raid_message(mid 0x%x) failed.\n", stats_req.header.mid);
			return(NFX_ERR);
		}
		else 
		if (rc == 0) {
			printf("send_raid_message(mid 0x%x) timed out.\n", stats_req.header.mid);
			return(NFX_ERR);
		}
		if (stats_rsp.header.mid != RAID_MID_PHYDEV_STATS_RSP) {
			printf("Invalid response mid; expected 0x%x got 0x%x\n",
			       RAID_MID_PHYDEV_STATS_RSP, stats_rsp.header.mid);
			return(NFX_ERR);
		}

		if (stats_rsp.rCode == RAID_RCODE_SUCCESS) {
			printf("%3d %3d %10d %8d %8d %8d %8d %8d %8d %8d %8d\n", 
					stats_rsp.id.channelNumber, stats_rsp.id.targetId, 
		      	 	stats_rsp.numMsecSinceStart,
		       		stats_rsp.numReads, stats_rsp.numWrites,
					stats_rsp.dataReadsInKB, stats_rsp.dataWritesInKB,
					stats_rsp.numSoftErrors, stats_rsp.numHardErrors,
					stats_rsp.numParityErrors, stats_rsp.numMiscErrors);
		}
		else if(stats_rsp.rCode == RAID_RCODE_PHYSDEV_UNCONFIGURED) {
			printf("%3d %3d %50s\n", 
					stats_rsp.id.channelNumber, stats_rsp.id.targetId,
					"Device Unconfigured");
		}

		else {
			printf("raid physdev_stats failed: rCode %s(0x%x)\n", 
			       raid_ecode_to_string(stats_rsp.rCode), stats_rsp.rCode);
			return(NFX_ERR);
		}
	}

	return(NFX_OK);
}

int 
raid_ui_lun_list(int argc, char **argv, int timeout, ...)
{
	int				rc;
	int				req_len = 0;
	int				rsp_len = 0;
	raid_logDevListReqMsg_t		req;
	raid_logDevListRspMsg_t	*	rsp;
	char				rspbuf[65536];
	int				i;
	raid_logDevInfoReqMsg_t		info_req;
	raid_logDevInfoRspMsg_t		info_rsp;


	req.header.mid = RAID_MID_LOGDEV_LIST_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);
			
	req_len = sizeof(raid_logDevListReqMsg_t);
	rsp_len = sizeof(rspbuf);
	rc = send_raid_message(&req, req_len, rspbuf, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}
	rsp = (raid_logDevListRspMsg_t *)rspbuf;
	if (rsp->header.mid != RAID_MID_LOGDEV_LIST_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_LOGDEV_LIST_RSP, rsp->header.mid);
		return(NFX_ERR);
	}
	if (rsp->rCode != RAID_RCODE_SUCCESS) {
		printf("raid lun_list failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(rsp->rCode), rsp->rCode);
		return(NFX_ERR);
	}

	printf("ID   RAID LEVEL                       PHYS SZ ACT SZ  STRP SZ I'D R $ W $ BG TASK STATE      \n");
	printf("---  -------------------------------- ------- ------- ------- --- --- --- ------- -----------\n");

	for (i = 0; i < rsp->numEntries; ++i) {
		info_req.header.mid = RAID_MID_LOGDEV_INFO_REQ;
		info_req.header.cookie = 1;
		strcpy(info_req.header.id, argv[1]);
		info_req.id = rsp->idList[i];
			
		req_len = sizeof(raid_logDevInfoReqMsg_t);
		rsp_len = sizeof(raid_logDevInfoRspMsg_t);
		rc = send_raid_message(&info_req, req_len, &info_rsp, &rsp_len, timeout);
		if (rc < 0) {
			printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
			return(NFX_ERR);
		}
		else 
		if (rc == 0) {
			printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
			return(NFX_ERR);
		}
		if (info_rsp.header.mid != RAID_MID_LOGDEV_INFO_RSP) {
			printf("Invalid response mid; expected 0x%x got 0x%x\n",
			       RAID_MID_LOGDEV_INFO_RSP, info_rsp.header.mid);
			return(NFX_ERR);
		}
		if (info_rsp.rCode != RAID_RCODE_SUCCESS) {
			printf("raid lun_list failed: rCode %s(0x%x)\n", 
			       raid_ecode_to_string(rsp->rCode), rsp->rCode);
			return(NFX_ERR);
		}
		printf("%-4d %-32s %3d,%03d %3d,%03d %5dKB %3s %3s %3s %7s %s\n", 
		       info_rsp.id, raid_ui_raid_level_to_str(info_rsp.raidLevel),
		       info_rsp.physicalSizeInMB/1000, info_rsp.physicalSizeInMB%1000, 
		       info_rsp.actualSizeInMB/1000, info_rsp.actualSizeInMB%1000, 
		       info_rsp.stripeSizeInKB,
		       info_rsp.isInitialized ? "YES" : "NO",
		       info_rsp.readCacheEnabled ? "ON" : "OFF",
		       info_rsp.writeCacheEnabled ? "ON" : "OFF",
		       info_rsp.hasBgTask ? "YES" : "NO",
		       raid_ui_logdev_state_to_str(info_rsp.deviceState));
	}

	return(NFX_OK);
}

int 
raid_ui_lun_bg_status(int argc, char **argv, int timeout, ...)
{
	int				rc;
	int				req_len = 0;
	int				rsp_len = 0;
	raid_bpListReqMsg_t		req;
	raid_bpListRspMsg_t *		rsp;
	char				rspbuf[65536];
	int				i;


	req.header.mid = RAID_MID_LUN_BPLIST_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);
			
	req_len = sizeof(raid_bpListReqMsg_t);
	rsp_len = sizeof(rspbuf);
	rc = send_raid_message(&req, req_len, rspbuf, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}
	rsp = (raid_bpListRspMsg_t *)rspbuf;
	if (rsp->header.mid != RAID_MID_LUN_BPLIST_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_LUN_BPLIST_RSP, rsp->header.mid);
		return(NFX_ERR);
	}
	if (rsp->rCode != RAID_RCODE_SUCCESS) {
		printf("raid lun_bg_status failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(rsp->rCode), rsp->rCode);
		return(NFX_ERR);
	}

	printf("ID CHECK   INIT    REBUILD EXPAND \n");
	printf("-- ------- ------- ------- -------\n");

	for (i = 0; i < rsp->numEntries; ++i) {
		printf("%02x ", rsp->bpList[i].id);
		if (rsp->bpList[i].checkProgress == -1)
			printf("IDLE    ");
		else 
			printf("%3d%c    ", rsp->bpList[i].checkProgress, '%');

		if (rsp->bpList[i].initProgress == -1)
			printf("IDLE    ");
		else 
			printf("%3d%c    ", rsp->bpList[i].initProgress, '%');
		
		if (rsp->bpList[i].rebuildProgress == -1)
			printf("IDLE    ");
		else 
			printf("%3d%c    ", rsp->bpList[i].rebuildProgress, '%');
		
		if (rsp->bpList[i].expandProgress == -1)
			printf("IDLE    ");
		else 
			printf("%3d%c    ", rsp->bpList[i].expandProgress, '%');
		printf("\n");
	}

	return(NFX_OK);
}

int 
raid_ui_config_save(int argc, char **argv, int timeout, ...)
{
	int			rc;
	int			req_len = 0;
	int			rsp_len = 0;
	raid_ctrlConfigReqMsg_t req;
	raid_ctrlConfigRspMsg_t rsp;

	req.header.mid = RAID_MID_CTRL_CFGDATA_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);
	req.type = RAID_CONFIG_SAVE;
	strcpy(req.fileName, argv[2]);
			
	req_len = sizeof(raid_ctrlConfigReqMsg_t);
	rsp_len = sizeof(raid_ctrlConfigRspMsg_t);
	rc = send_raid_message(&req, req_len, &rsp, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}

	if (rsp.header.mid != RAID_MID_CTRL_CFGDATA_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_CTRL_CFGDATA_RSP, rsp.header.mid);
		return(NFX_ERR);
	}
	if (rsp.rCode != RAID_RCODE_SUCCESS) {
		printf("raid config_save failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(rsp.rCode), rsp.rCode);
		return(NFX_ERR);
	}

	return(NFX_OK);
}

int 
raid_ui_config_load(int argc, char **argv, int timeout, ...)
{
	int			rc;
	int			req_len = 0;
	int			rsp_len = 0;
	raid_ctrlConfigReqMsg_t req;
	raid_ctrlConfigRspMsg_t rsp;

	req.header.mid = RAID_MID_CTRL_CFGDATA_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);
	req.type = RAID_CONFIG_LOAD;
	strcpy(req.fileName, argv[2]);
			
	req_len = sizeof(raid_ctrlConfigReqMsg_t);
	rsp_len = sizeof(raid_ctrlConfigRspMsg_t);
	rc = send_raid_message(&req, req_len, &rsp, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}

	if (rsp.header.mid != RAID_MID_CTRL_CFGDATA_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_CTRL_CFGDATA_RSP, rsp.header.mid);
		return(NFX_ERR);
	}
	if (rsp.rCode != RAID_RCODE_SUCCESS) {
		printf("raid config_load failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(rsp.rCode), rsp.rCode);
		return(NFX_ERR);
	}

	return(NFX_OK);
}

int 
raid_ui_config_clr(int argc, char **argv, int timeout, ...)
{
	int			rc;
	int			req_len = 0;
	int			rsp_len = 0;
	raid_ctrlClrCfgReqMsg_t req;
	raid_ctrlClrCfgRspMsg_t rsp;

	req.header.mid = RAID_MID_CTRL_CLRCFG_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);
			
	req_len = sizeof(raid_ctrlClrCfgReqMsg_t);
	rsp_len = sizeof(raid_ctrlClrCfgRspMsg_t);
	rc = send_raid_message(&req, req_len, &rsp, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}

	if (rsp.header.mid != RAID_MID_CTRL_CLRCFG_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_CTRL_CLRCFG_RSP, rsp.header.mid);
		return(NFX_ERR);
	}
	if (rsp.rCode != RAID_RCODE_SUCCESS) {
		printf("raid config_clr failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(rsp.rCode), rsp.rCode);
		return(NFX_ERR);
	}

	return(NFX_OK);
}

int 
raid_ui_lun_create_simple(int argc, char **argv, int timeout, ...)
{
	int				rc;
	int				req_len = 0;
	int				rsp_len = 0;
	raid_createSimpleLunReqMsg_t	req;
	raid_createSimpleLunRspMsg_t	rsp;
	char *				end_ptr;

	req.header.mid = RAID_MID_LUN_CREATE_SIMPLE_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);
	req.lunCount = strtol(argv[2], &end_ptr, 10);
	if (*end_ptr != '\0') {
		printf("raid lun_create_simple failed: Invalid LUN count '%s'.\n", argv[2]);
		return(NFX_ERR);
	}
	req.raidLevel = strtol(argv[3], &end_ptr, 10);
	if (*end_ptr != '\0' ||
	    (req.raidLevel != 0 && req.raidLevel != 3 && req.raidLevel != 5)) {
		printf("raid lun_create_simple failed: Invalid RAID type '%s'.\n", argv[3]);
		return(NFX_ERR);
	}
	req.sizeInMB = strtol(argv[4], &end_ptr, 10);
	if (*end_ptr != '\0') {
		printf("raid lun_create_simple failed: Invalid size '%s'.\n", argv[4]);
		return(NFX_ERR);
	}
	req.stripeSizeInKB = strtol(argv[5], &end_ptr, 10);
	if (*end_ptr != '\0' ||
	    (req.stripeSizeInKB != 8 && req.stripeSizeInKB != 16 && req.stripeSizeInKB != 32 && req.stripeSizeInKB != 64)) {
		printf("raid lun_create_simple failed: Invalid stripe size '%s'.\n", argv[5]);
		return(NFX_ERR);
	}
			
	req_len = sizeof(raid_createSimpleLunReqMsg_t);
	rsp_len = sizeof(raid_createSimpleLunRspMsg_t);
	rc = send_raid_message(&req, req_len, &rsp, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}

	if (rsp.header.mid != RAID_MID_LUN_CREATE_SIMPLE_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_LUN_CREATE_SIMPLE_RSP, rsp.header.mid);
		return(NFX_ERR);
	}
	if (rsp.rCode != RAID_RCODE_SUCCESS) {
		printf("raid lun_create_simple failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(rsp.rCode), rsp.rCode);
		return(NFX_ERR);
	}

	return(NFX_OK);
}

int 
raid_ui_lun_create(int argc, char **argv, int timeout, ...)
{
	int				rc;
	int				physdev_count;
	int				req_len = 0;
	int				rsp_len = 0;
	raid_createLunReqMsg_t *        req;
	raid_createLunRspMsg_t		rsp;
	char *				end_ptr;
	int				i;

	physdev_count = argc - 5;
	req_len = sizeof(raid_createLunReqMsg_t) + physdev_count * sizeof(raid_phydev_id_t);
	req = malloc(req_len);
	bzero(req, req_len);
	
	req->header.mid = RAID_MID_LUN_CREATE_REQ;
	req->header.cookie = 1;
	strcpy(req->header.id, argv[1]);
	req->lundef.raidLevel = strtol(argv[2], &end_ptr, 10);
	if (*end_ptr != '\0' ||
	    (req->lundef.raidLevel != 0 && 
	     req->lundef.raidLevel != 1 && 
	     req->lundef.raidLevel != 3 && 
	     req->lundef.raidLevel != 5)) {
		printf("raid lun_create failed: Invalid RAID type '%s'.\n", argv[2]);
		free(req);
		return(NFX_ERR);
	}
	req->lundef.sizeInMB = strtol(argv[3], &end_ptr, 10);
	if (*end_ptr != '\0') {
		printf("raid lun_create failed: Invalid size '%s'.\n", argv[3]);
		free(req);
		return(NFX_ERR);
	}
	req->lundef.stripeSizeInKB = strtol(argv[4], &end_ptr, 10);
	if (*end_ptr != '\0') {
		printf("raid lun_create failed: Invalid stripe size '%s'.\n", argv[4]);
		free(req);
		return(NFX_ERR);
	}
	req->lundef.cacheEnabled = 1;
	req->lundef.autoInit = 0;
	req->lundef.numOfPhyDevices = physdev_count;

	for (i = 0; i < physdev_count; ++i) {
		req->lundef.phyDevList[i].channelNumber = strtol(argv[5+i], &end_ptr, 10);
		if (*end_ptr != '/') {
			printf("raid lun_create failed: Invalid device definitions '%s'.\n", argv[5+i]);
			free(req);
			return(NFX_ERR);
		}
		++end_ptr;
		req->lundef.phyDevList[i].targetId = strtol(end_ptr, &end_ptr, 10);
		if (*end_ptr != '\0') {
			printf("raid lun_create failed: Invalid device definitions '%s'.\n", argv[5+i]);
			free(req);
			return(NFX_ERR);
		}
	}

	rsp_len = sizeof(raid_createLunRspMsg_t);
	rc = send_raid_message(req, req_len, &rsp, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req->header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req->header.mid);
		return(NFX_ERR);
	}

	if (rsp.header.mid != RAID_MID_LUN_CREATE_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_LUN_CREATE_RSP, rsp.header.mid);
		return(NFX_ERR);
	}
	if (rsp.rCode != RAID_RCODE_SUCCESS) {
		printf("raid lun_create failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(rsp.rCode), rsp.rCode);
		return(NFX_ERR);
	}

	printf("LUN %d successfully created.\n", rsp.id);

	return(NFX_OK);
}

int 
raid_ui_lun_delete(int argc, char **argv, int timeout, ...)
{
	int				rc;
	int				req_len = 0;
	int				rsp_len = 0;
	raid_deleteLunReqMsg_t		req;
	raid_deleteLunRspMsg_t		rsp;
	char *				end_ptr;

	bzero(&req, req_len);
	req.header.mid = RAID_MID_LUN_DELETE_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);
	req.id = strtol(argv[2], &end_ptr, 10);
	if (*end_ptr != '\0') {
		printf("raid lun_delete failed: Invalid LUN ID '%s'.\n", argv[2]);
		return(NFX_ERR);
	}

	req_len = sizeof(raid_deleteLunReqMsg_t);
	rsp_len = sizeof(raid_deleteLunRspMsg_t);
	rc = send_raid_message(&req, req_len, &rsp, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}

	if (rsp.header.mid != RAID_MID_LUN_DELETE_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_LUN_DELETE_RSP, rsp.header.mid);
		return(NFX_ERR);
	}
	if (rsp.rCode != RAID_RCODE_SUCCESS) {
		printf("raid lun_delete failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(rsp.rCode), rsp.rCode);
		return(NFX_ERR);
	}

	printf("LUN deleted successfully.\n");

	return(NFX_OK);
}

int 
raid_ui_lun_init(int argc, char **argv, int timeout, ...)
{
	int				rc;
	int				req_len = 0;
	int				rsp_len = 0;
	raid_initLunReqMsg_t		req;
	raid_initLunRspMsg_t		rsp;
	char *				end_ptr;

	bzero(&req, req_len);
	req.header.mid = RAID_MID_LUN_INIT_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);
	if (strcmp(argv[2], "ALL") == 0)
		req.type = RAID_INITTYPE_ALL;
	else
	if (strcmp(argv[2], "ALLNOT") == 0)
		req.type = RAID_INITTYPE_ALLNOT;
	else {
		req.id = strtol(argv[2], &end_ptr, 10);
		if (*end_ptr != '\0') {
			printf("raid lun_init failed: Invalid LUN ID '%s'.\n", argv[2]);
			return(NFX_ERR);
		}
		req.type = RAID_INITTYPE_ONE;
	}

	req_len = sizeof(raid_initLunReqMsg_t);
	rsp_len = sizeof(raid_initLunRspMsg_t);
	rc = send_raid_message(&req, req_len, &rsp, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}

	if (rsp.header.mid != RAID_MID_LUN_INIT_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_LUN_INIT_RSP, rsp.header.mid);
		return(NFX_ERR);
	}
	if (rsp.rCode != RAID_RCODE_SUCCESS) {
		printf("raid lun_init failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(rsp.rCode), rsp.rCode);
		return(NFX_ERR);
	}

	printf("%d LUN init operations started successfully.\n", rsp.init_started_count);

	return(NFX_OK);
}

int 
raid_ui_lun_check(int argc, char **argv, int timeout, ...)
{
	int				rc;
	int				req_len = 0;
	int				rsp_len = 0;
	raid_checkLunReqMsg_t		req;
	raid_checkLunRspMsg_t		rsp;
	char *				end_ptr;

	bzero(&req, req_len);
	req.header.mid = RAID_MID_LUN_CHECK_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);
	req.id = strtol(argv[2], &end_ptr, 10);
	if (*end_ptr != '\0') {
		printf("raid lun_check failed: Invalid LUN ID '%s'.\n", argv[2]);
		return(NFX_ERR);
	}
	req.type = RAID_CHKTYPE_CHKRES;
	
	req_len = sizeof(raid_checkLunReqMsg_t);
	rsp_len = sizeof(raid_checkLunRspMsg_t);
	rc = send_raid_message(&req, req_len, &rsp, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}

	if (rsp.header.mid != RAID_MID_LUN_CHECK_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_LUN_CHECK_RSP, rsp.header.mid);
		return(NFX_ERR);
	}
	if (rsp.rCode != RAID_RCODE_SUCCESS) {
		printf("raid lun_check failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(rsp.rCode), rsp.rCode);
		return(NFX_ERR);
	}

	printf("LUN check started successfully.\n");

	return(NFX_OK);
}

int 
raid_ui_lun_physlist(int argc, char **argv, int timeout, ...)
{
	int				rc;
	int				req_len = 0;
	int				rsp_len = 0;
	raid_logDevPhyListReqMsg_t	req; 
	raid_logDevPhyListRspMsg_t *    rsp;
	char				rspbuf[65536];
	char *				end_ptr;
	int				i;

	bzero(&req, req_len);
	req.header.mid = RAID_MID_LOGDEV_PHYLIST_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);
	req.id = strtol(argv[2], &end_ptr, 10);
	if (*end_ptr != '\0') {
		printf("raid lun_physlist failed: Invalid LUN ID '%s'.\n", argv[2]);
		return(NFX_ERR);
	}
	
	req_len = sizeof(raid_logDevPhyListReqMsg_t);
	rsp_len = sizeof(rspbuf);
	rc = send_raid_message(&req, req_len, rspbuf, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}

	rsp = (raid_logDevPhyListRspMsg_t *)rspbuf;
	if (rsp->header.mid != RAID_MID_LOGDEV_PHYLIST_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_LOGDEV_PHYLIST_RSP, rsp->header.mid);
		return(NFX_ERR);
	}
	if (rsp->rCode != RAID_RCODE_SUCCESS) {
		printf("raid lun_physlist failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(rsp->rCode), rsp->rCode);
		return(NFX_ERR);
	}


	printf("RDN  CID TID START BLK  USED BLKS  TOTAL BLKS\n");
	printf("---- --- --- ---------- ---------- ----------\n");
	for (i = 0; i < rsp->numDevices; ++i) {
		printf("%04x %3d %3d 0x%08x 0x%08x 0x%08x\n", 
		       rsp->driveList[i].rdn, 
		       rsp->driveList[i].channelNumber,
		       rsp->driveList[i].targetId,
		       rsp->driveList[i].startBlock,
		       rsp->driveList[i].usedBlocks, 
		       rsp->driveList[i].totalBlocks);
	}

	return(NFX_OK);
}

int 
raid_ui_lun_cache(int argc, char **argv, int timeout, ...)
{
	int				rc;
	int				req_len = 0;
	int				rsp_len = 0;
	raid_cacheLunReqMsg_t		req;
	raid_cacheLunRspMsg_t		rsp;
	char *				end_ptr;

	bzero(&req, req_len);
	req.header.mid = RAID_MID_LUN_CACHE_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);

	req.id = strtol(argv[2], &end_ptr, 10);
	if (*end_ptr != '\0') {
		printf("raid lun_cache failed: Invalid LUN ID '%s'.\n", argv[2]);
		return(NFX_ERR);
	}

	if ((strcmp(argv[3], "ON") == 0) || 
	    (strcmp(argv[3], "on") == 0))
		req.rc_enabled = 1;
	else
	if ((strcmp(argv[3], "OFF") == 0) ||
	    (strcmp(argv[3], "off") == 0))
		req.rc_enabled = 0;
	else {
		printf("raid lun_cache failed: Invalid value for read cache '%s'.\n", argv[3]);
		return(NFX_ERR);
	}

	if ((strcmp(argv[4], "ON") == 0) || 
	    (strcmp(argv[4], "on") == 0))
		req.wc_enabled = 1;
	else
	if ((strcmp(argv[4], "OFF") == 0) ||
	    (strcmp(argv[4], "off") == 0))
		req.wc_enabled = 0;
	else {
		printf("raid lun_cache failed: Invalid value for write cache '%s'.\n", argv[4]);
		return(NFX_ERR);
	}

	req_len = sizeof(raid_cacheLunReqMsg_t);
	rsp_len = sizeof(raid_cacheLunRspMsg_t);
	rc = send_raid_message(&req, req_len, &rsp, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}

	if (rsp.header.mid != RAID_MID_LUN_CACHE_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_LUN_CACHE_RSP, rsp.header.mid);
		return(NFX_ERR);
	}
	if (rsp.rCode != RAID_RCODE_SUCCESS) {
		printf("raid lun_cache failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(rsp.rCode), rsp.rCode);
		return(NFX_ERR);
	}

	return(NFX_OK);

}

int 
raid_ui_dbg_level(int argc, char **argv, int timeout, ...)
{
	int				rc;
	char				req_msg[1024];
	char				rsp_msg[1024];
	int				req_len = 0;
	int				rsp_len = 0;
	raid_setDebugLevelReqMsg_t *	req;
	raid_setDebugLevelRspMsg_t *	rsp;

	req = (raid_setDebugLevelReqMsg_t *)req_msg;
	req->header.mid = RAID_MID_SET_DEBUG_LEVEL_REQ;
	req->header.cookie = 1;
	req->debug_level = strtoul(argv[1], NULL, 0);
			
	req_len = sizeof(raid_setDebugLevelReqMsg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_raid_message(req_msg, req_len, rsp_msg, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req->header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req->header.mid);
		return(NFX_ERR);
	}

	rsp = (raid_setDebugLevelRspMsg_t *)rsp_msg;
	if (rsp->header.mid != RAID_MID_SET_DEBUG_LEVEL_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_SET_DEBUG_LEVEL_RSP, rsp->header.mid);
		return(NFX_ERR);
	}
	if (rsp_len != sizeof(raid_setDebugLevelRspMsg_t)) {
		printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
		       req->header.mid, sizeof(raid_setDebugLevelRspMsg_t), rsp_len);
		return(NFX_ERR);
	}
	return(NFX_OK);
}

// checks if a name is valid devname
int raid_ui_is_devname(char *devname)
{
	int				rc;
	char				req_msg[1024];
	char				rsp_msg[16 * 1024];
	int			     	req_len = 0;
	int			     	rsp_len = 0;
	raid_getListReqMsg_t*	     	devlist_req;
	raid_getListRspMsg_t*	     	devlist_rsp;
	int				i;

	if(!devname) return 0;

	devlist_req = (raid_getListReqMsg_t *)req_msg;
	devlist_req->header.mid = RAID_MID_GET_LIST_REQ;
	devlist_req->header.cookie = 1;
			
	req_len = sizeof(raid_getListReqMsg_t);
	rsp_len = sizeof(rsp_msg);
	rc = send_raid_message(req_msg, req_len, rsp_msg, &rsp_len, 60);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", devlist_req->header.mid);
		return(0);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", devlist_req->header.mid);
		return(0);
	}

	devlist_rsp = (raid_getListRspMsg_t *)rsp_msg;
	if (devlist_rsp->header.mid != RAID_MID_GET_LIST_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_GET_LIST_RSP, devlist_rsp->header.mid);
		return(0); 
	}
	if (rsp_len != sizeof(raid_getListRspMsg_t) + sizeof(raid_ctrl_id_t) * (devlist_rsp->numEntries - 1)) {
		printf("Invalid response length on mid 0x%x: expected %d got %d\n", 
		       devlist_req->header.mid, 
		       sizeof(raid_getListRspMsg_t) + sizeof(raid_ctrl_id_t) * (devlist_rsp->numEntries - 1),
		       rsp_len);
		return(0);
	}
	if (devlist_rsp->rCode != RAID_RCODE_SUCCESS) {
		printf("raid devlist failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(devlist_rsp->rCode), devlist_rsp->rCode);
		return(0);
	}

	// check if devname is valid
	for(i=0; i < devlist_rsp->numEntries; i++) {
		if(!strcmp(devname, devlist_rsp->idList[i]))
			return 1;
	}

	return 0;
}

// gets the device name from file
// and sets it in argv[1]
int raid_ui_get_devname(int argc, char **argv)
{
	char *devname;
	FILE *fp;
	
	// dbg_level & devlist have no devname and open requires
	// a devname, so if arg count is incorrect, its an error
	if(!strcmp(argv[1], "dbg_level") ||
	   !strcmp(argv[1], "devlist") ||
	   !strcmp(argv[1], "open")) 
		return 0;

	// check if argv[2] is a devname. If so, then return error since some
	// other arg is missing
	if(argc >= 3 && raid_ui_is_devname(argv[2])) return 0;

	devname = (char *)malloc(256);
	if(!devname) return 0;

	// since argv[2] isn't a devname, look for devname.
	// if found, then overwrite argv[1] with it
	// else return error
	fp = fopen(RAID_UI_DEV_FILE, "r");
	if(!fp) return 0;
	fscanf(fp, "%s", devname);
	fclose(fp);

	// if a close is being issued, then remove file
	if(!strcmp(argv[1], "close")) 
		unlink(RAID_UI_DEV_FILE);

	argv[1] = devname;

	return 1;
}

// saves device name in file
void raid_ui_set_devname(char **argv)
{
	FILE *fp;

	if(!strcmp(argv[1], "close")) {
		unlink(RAID_UI_DEV_FILE);
		return;
	}

	if(!strcmp(argv[1], "dbg_level")) return;
	if(!strcmp(argv[1], "devlist")) return;

	if(!raid_ui_is_devname(argv[2])) return;

	fp = fopen(RAID_UI_DEV_FILE, "w");
	if(!fp) return;
	fprintf(fp, "%s", argv[2]);
	fclose(fp);
}


int 
main(int argc, char **argv)
{
	int	i;
	int	rc;
	int offset = 1;

	if (argc < 2)
		goto arg_error;

	for (i = 0; raid_ui_root[i].name; ++i) {
		if (strcmp(raid_ui_root[i].name, argv[1]) == 0) {
			if (argc-2 < raid_ui_root[i].parm_cnt) {

				// check if missing arg. is devname
				if(argc-1 == raid_ui_root[i].parm_cnt) {
					if(!raid_ui_get_devname(argc, argv))
						goto arg_error;
					offset = 0;
				}
				else 
					goto arg_error;
			}
			else
				raid_ui_set_devname(argv);

			rc = raid_ui_root[i].func(argc-offset, argv+offset, 
						  raid_ui_root[i].timeout);
			return(rc == NFX_OK ? 0 : 1);
		}
	}

arg_error:
	print_usage(argv[0]);
	exit(1);
}

int
send_raid_message(void *req_buf, int req_len, void *resp_buf, int *resp_len, int timeout)
{
	int		stat;
	struct timeval	to;

	to.tv_sec = timeout;
	to.tv_usec = 0;

	stat = sendAgileMsg(-1,
			    req_buf, req_len,
			    "raid-config", 0, 0, 0, &to,
			    resp_buf, *resp_len);
	if (stat > 0)
		*resp_len = stat;
	return(stat);
}

char *
raid_ecode_to_string(uint32 ecode)
{
	switch(ecode) {
	case RAID_RCODE_SUCCESS:		return("SUCCESS");
	case RAID_RCODE_NOT_OPEN:		return("NOT_OPEN");
	case RAID_RCODE_NO_DEVICE:		return("NO_DEVICE");
	case RAID_RCODE_BUSY:			return("BUSY");
	case RAID_RCODE_SDM_ERROR:		return("SDM_ERROR");
	case RAID_RCODE_IO_ERROR:		return("IO_ERROR");
	case RAID_RCODE_INVALID_PARAMETERS:	return("INVALID_PARMS");
	case RAID_RCODE_INVALID_PHYSICAL_DEVICE:return("INVALID_PHYSICAL_DEVICE");
	case RAID_RCODE_INVALID_LUN:		return("INVALID_LUN");
	case RAID_RCODE_INSUFFICIENT_SPACE:	return("INSUFFICIENT_SPACE");
	case RAID_RCODE_TOO_MANY_LUNS:		return("TOO_MANY_LUNS");
	case RAID_RCODE_UNABLE_TO_DELETE_LUN:	return("UNABLE_TO_DELETE_LUN");
	case RAID_RCODE_LUN_IS_BUSY:		return("LUN_IS_BUSY");
	case RAID_RCODE_CANNOT_OPEN_FILE:	return("CANNOT_OPEN_FILE");
	case RAID_RCODE_PARSE_ERROR:		return("PARSE_ERROR");
	case RAID_RCODE_NOT_IMPLEMENTED:	return("NOT_IMPLEMENTED");
	case RAID_RCODE_PHYSDEV_UNCONFIGURED:	return("PHYSICAL_DEVICE_UNCONFIGURED");
	case RAID_RCODE_DISK_UPGRADE_NOTSUPPORTED: return("DISK_UPGRADE_NOT_SUPPORTED");
	default:				return("???");
	}
}

char *
raid_ui_state_to_str(uint32 dev_state)
{
	switch(dev_state) {
	case RAID_DEV_STATE_CLOSED:	return("CLOSED");
	case RAID_DEV_STATE_OPEN_IDLE:	return("OPEN (idle)");
	case RAID_DEV_STATE_OPEN_BUSY:	return("OPEN (busy)");
	case RAID_DEV_STATE_CLOSING:	return("CLOSING");
	default:			return("???");
	}
}

char *
raid_ui_physdev_state_to_str(uint32 physdev_state)
{
	switch(physdev_state) {
	case RAID_PHYDEVSTATE_UNCONFIGURED:		return("Unconfigured");
	case RAID_PHYDEVSTATE_OPTIMAL:			return("Optimal");
	case RAID_PHYDEVSTATE_REBUILD:			return("Rebuilding");
	case RAID_PHYDEVSTATE_FAILED:			return("Offline");
	case RAID_PHYDEVSTATE_CRITICAL:			return("Critical");
	case RAID_PHYDEVSTATE_GLOBHOTSPARE:		return("Glob. HotSpare");
	case RAID_PHYDEVSTATE_PERMHOTSPARE:		return("Dedicated HotSpare");
	default:					return("Unknown");
	}
}

char *
raid_ui_logdev_state_to_str(uint32 physdev_state)
{
	switch(physdev_state) {
	case RAID_LOGDEVSTATE_UNCONFIGURED:		return("Unconfigured");
	case RAID_LOGDEVSTATE_OPTIMAL:			return("Optimal");
	case RAID_LOGDEVSTATE_REBUILDING:		return("Rebuilding");
	case RAID_LOGDEVSTATE_FAILED:			return("Offline");
	case RAID_LOGDEVSTATE_CRITICAL:			return("Critical");
	default:					return("Unknown");
	}
}

char *
raid_ui_raid_level_to_str(uint32 raid_level)
{
	switch (raid_level)
	{
	case 0x0:		return("RAID-0 (striping)");
	case 0x1:		return("RAID-1 (mirroring)");
	case 0x3:		return("RAID-3 Right asymmetric parity");
	case 0x5:		return("RAID-5 Right asymmetric parity");
	case 0x6:		return("RAID-0+1");
	case 0x7:		return("JBOD");
	case 0x8:		return("New Mylex SPAN");
	case 0x9:		return("RAID-3 (Fixed parity)");
	case 0xb:		return("RAID-3 Left asymmetric parity");
	case 0xc:		return("Spanned device");
	case 0xd:		return("RAID-5 Left asymmetric parity");
	case 0xe:		return("Concatentation");
	case 0xf:		return("Physical device");
	default:		return("unknown");
	} 
}

int
raid_ui_get_sizes(char *devname, uint64 *phySpace, uint64 *freeSpace, uint64 *usableSpace)
{
	int				rc;
	int				req_len = 0;
	int				rsp_len = 0;
	raid_phyDevListReqMsg_t		req;
	raid_phyDevListRspMsg_t	*	rsp;
	char				rspbuf[65536];
	int				i;
	raid_phyDevInfoReqMsg_t		info_req;
	raid_phyDevInfoRspMsg_t		info_rsp;
	uint64			freeSpace_l = 0;
	uint64			phySpace_l = 0;
	uint64			usableSpace_l = 0;
	int				timeout = 60;

	*phySpace = *freeSpace = *usableSpace = 0;

	req.header.mid = RAID_MID_PHYDEV_LIST_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, devname);
			
	req_len = sizeof(raid_phyDevListReqMsg_t);
	rsp_len = sizeof(rspbuf);
	rc = send_raid_message(&req, req_len, rspbuf, &rsp_len, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}
	rsp = (raid_phyDevListRspMsg_t *)rspbuf;
	if (rsp->header.mid != RAID_MID_PHYDEV_LIST_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_PHYDEV_LIST_RSP, rsp->header.mid);
		return(NFX_ERR);
	}
	if (rsp->rCode != RAID_RCODE_SUCCESS) {
		printf("raid physdev failed: rCode %s(0x%x)\n", 
		       raid_ecode_to_string(rsp->rCode), rsp->rCode);
		return(NFX_ERR);
	}

	for (i = 0; i < rsp->numEntries; ++i) {
		info_req.header.mid = RAID_MID_PHYDEV_INFO_REQ;
		info_req.header.cookie = 1;
		strcpy(info_req.header.id, devname);
		info_req.id.channelNumber = rsp->idList[i].channelNumber;
		info_req.id.targetId = rsp->idList[i].targetId;		
			
		req_len = sizeof(raid_phyDevInfoReqMsg_t);
		rsp_len = sizeof(raid_phyDevInfoRspMsg_t);
		rc = send_raid_message(&info_req, req_len, &info_rsp, &rsp_len, timeout);
		if (rc < 0) {
			printf("send_raid_message(mid 0x%x) failed.\n", info_req.header.mid);
			return(NFX_ERR);
		}
		else 
		if (rc == 0) {
			printf("send_raid_message(mid 0x%x) timed out.\n", info_req.header.mid);
			return(NFX_ERR);
		}
		if (info_rsp.header.mid != RAID_MID_PHYDEV_INFO_RSP) {
			printf("Invalid response mid; expected 0x%x got 0x%x\n",
			       RAID_MID_PHYDEV_INFO_RSP, info_rsp.header.mid);
			return(NFX_ERR);
		}
		if (info_rsp.rCode != RAID_RCODE_SUCCESS) {
			printf("raid physdev failed: rCode %s(0x%x)\n", 
			       raid_ecode_to_string(info_rsp.rCode), info_rsp.rCode);
			return(NFX_ERR);
		}
		phySpace_l +=  info_rsp.physicalSizeInMB;
		usableSpace_l +=  info_rsp.usableSizeInMB;
		freeSpace_l += info_rsp.freeSizeInMB;
	}

	*phySpace = phySpace_l;
	*usableSpace = usableSpace_l;
	*freeSpace = freeSpace_l;

	return(NFX_OK);
}

int
raid_ui_get_logpage(int argc, char **argv, int timeout, ...)
{
	uint32						page_code = 0;
	raid_phyPassThruReqMsg_t	req;
	raid_phyPassThruRspMsg_t	*rsp;
	char						rspbuf[5000];
	uint32						rsplen = 0;
	int							rc;
	int							internal_page_code = ALL_PAGE_CODES;

	if(argc != 4 && argc != 5 && argc != 6) return 1;

	if(argc == 6) {
		internal_page_code = strtol(argv[5], NULL, 0);
		argc--;
	}
	if(argc == 5) {
		page_code = strtol(argv[4], NULL, 0);
		argc--;
	}

	// prepare the message
	memset((void *)&req, 0, sizeof(req));
	req.header.mid = RAID_MID_PHY_PASSTHRU_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);
	req.channelNumber = atol(argv[2]);
	req.targetId = atol(argv[3]);
	
	switch(page_code) {
		case SCSI_LOGPAGE_CODE_SUPPORTED_PAGES:
			req.cmd = RAID_PASSTHRU_PAGELIST;
			break;
		case SCSI_LOGPAGE_CODE_ERROR_COUNTER_WRITE:
			req.cmd = RAID_PASSTHRU_WRITE_ERROR_PAGE;
			break;
		case SCSI_LOGPAGE_CODE_ERROR_COUNTER_READ:
			req.cmd = RAID_PASSTHRU_READ_ERROR_PAGE;
			break;
		case SCSI_LOGPAGE_CODE_ERROR_COUNTER_READ_REV:
			req.cmd = RAID_PASSTHRU_REV_READ_ERROR_PAGE;
			break;
		case SCSI_LOGPAGE_CODE_ERROR_COUNTER_VERIFY:
			req.cmd = RAID_PASSTHRU_VERIFY_ERROR_PAGE;
			break;
		case SCSI_LOGPAGE_CODE_TEMPERATURE:
			req.cmd = RAID_PASSTHRU_TEMP_PAGE;
			break;
		case SCSI_LOGPAGE_CODE_BUFFER_OVER_UNDER_RUN:
			req.cmd = RAID_PASSTHRU_BUFFER_RUN_PAGE;
			break;
		default:
			printf("Page code 0x%x not implemented/not found\n", page_code);
			return 1;
	}

	// send it now
	rsplen = sizeof(rspbuf);
	rc = send_raid_message(&req, sizeof(req), rspbuf, &rsplen, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}

	rsp = (raid_phyPassThruRspMsg_t *)rspbuf;
	if (rsp->header.mid != RAID_MID_PHY_PASSTHRU_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_PHY_PASSTHRU_RSP, rsp->header.mid);
		return(NFX_ERR);
	}
	if (rsp->rCode != RAID_RCODE_SUCCESS) {
		printf("raid get_logpage failed for dev %s: rCode %s(0x%x)\n", 
		       argv[1], raid_ecode_to_string(rsp->rCode), rsp->rCode);
		return(NFX_ERR);
	}

	switch(page_code) {
		case SCSI_LOGPAGE_CODE_SUPPORTED_PAGES:
			raid_parse_logpage_list_rsp(rsp->dataLen, rsp->data);
			break;

		case SCSI_LOGPAGE_CODE_BUFFER_OVER_UNDER_RUN:
			raid_parse_buffer_run_rsp(rsp->dataLen, rsp->data);
			break;

		case SCSI_LOGPAGE_CODE_ERROR_COUNTER_WRITE:
		case SCSI_LOGPAGE_CODE_ERROR_COUNTER_READ:
		case SCSI_LOGPAGE_CODE_ERROR_COUNTER_READ_REV:
		case SCSI_LOGPAGE_CODE_ERROR_COUNTER_VERIFY:
			raid_parse_logpage_error_rsp(page_code, rsp->dataLen, rsp->data, internal_page_code);
			break;
		case SCSI_LOGPAGE_CODE_TEMPERATURE:
			raid_parse_logpage_temp_rsp(rsp->dataLen, rsp->data);
			break;
	}

	return (NFX_OK);
}

int
raid_ui_corrupt_bits(int argc, char **argv, int timeout, ...)
{
	int		num_bits = 0;
	int		rc;
	char	data[RAID_PASSTHRU_DEFAULT_RD_WR_SIZE];
	char	*st = data;
	int		bit_val;
	uchar8	mask;
	int		i;

	if(argc != 6) return 1;

	num_bits = atol(argv[5]);
	argc--;
	
	// read the sector + ecc
	rc = raid_ui_do_direct_read(argc, argv, timeout, data, sizeof(data));
	if(rc) return 0;

	// corrupt the bits
	for(i=0; i < num_bits; i++) {

		if(i && i % 8 == 0) st++;

		mask = 1 << (i % 8);
		bit_val = (*st >> (i % 8)) & 0x1;
		if(bit_val)
			(*st) &= ~mask;
		else
			(*st) |= mask;
	}
	
	// write the sector + ecc
	rc = raid_ui_do_direct_write(argc, argv, timeout, data, sizeof(data));
	if(rc) return 1;

	printf("%d adjecent bits successfully corrupted\n", num_bits);

	return (NFX_OK);
}

int
raid_ui_direct_read(int argc, char **argv, int timeout, ...)
{
	int		dataLen;
	char	*data;
	int		rc, i;
	
	if(argc != 5 && argc != 6) return 1;
	
	if(argc == 6) {
		dataLen = atol(argv[5]);
		argc--;
	}
	else 
		dataLen = RAID_PASSTHRU_DEFAULT_RD_WR_SIZE;

	data = (char *)malloc(dataLen);
	bzero(data, dataLen);

	// read the sector + ecc
	rc = raid_ui_do_direct_read(argc, argv, timeout, data, dataLen);	
	if(rc) return 1;

	// print the data
	for(i=0; i < dataLen; i++) {
		if(i % 8 == 0) printf("\n");
		if(i == SECTOR_SIZE) printf("\nECC DATA\n");
		printf("0x%02x ", data[i] & 0xFF);
	}
	printf("\n\n");
	
	return (NFX_OK);
}

int
raid_ui_do_direct_read(int argc, char **argv, int timeout, char *rspData, int dataLen)
{
	raid_phyPassThruReqMsg_t	req;
	raid_phyPassThruRspMsg_t	*rsp;
	uint32						rsplen = 0;
	int							rc;

	// prepare the message
	memset((void *)&req, 0, sizeof(req));
	req.header.mid = RAID_MID_PHY_PASSTHRU_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);
	req.channelNumber = atol(argv[2]);
	req.targetId = atol(argv[3]);
	req.cmd = RAID_PASSTHRU_READ_LONG;
	req.lba = strtoq(argv[4], NULL, 0);
	req.dataLen = dataLen;
	
	// send it now
	rsplen = sizeof(raid_phyPassThruRspMsg_t) + req.dataLen - 1;
	rsp = (raid_phyPassThruRspMsg_t *)malloc(rsplen);
	bzero(rsp, rsplen);
	rc = send_raid_message(&req, sizeof(req), (char *)rsp, &rsplen, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}

	if (rsp->header.mid != RAID_MID_PHY_PASSTHRU_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_PHY_PASSTHRU_RSP, rsp->header.mid);
		return(NFX_ERR);
	}
	if (rsp->rCode != RAID_RCODE_SUCCESS) {
		printf("raid direct_read for dev %s failed: rCode %s(0x%x)\n", 
		       argv[1], raid_ecode_to_string(rsp->rCode), rsp->rCode);
		return(NFX_ERR);
	}
	if(rsp->dataLen != req.dataLen) {
		printf("Read Failed: Requested %d bytes, got %d bytes\n",
				req.dataLen, rsp->dataLen);
		return(NFX_ERR);
	}

	printf("Read of %d bytes completed succesdfully\n", req.dataLen);

	memcpy((void *)rspData, rsp->data, dataLen);

	return(NFX_OK);
}

int
raid_ui_do_direct_write(int argc, char **argv, int timeout, char *reqData, int dataLen)
{
	raid_phyPassThruReqMsg_t	*req;
	raid_phyPassThruRspMsg_t	*rsp;
	uint32						rsplen = 0;
	int							rc;
	int							mlen;

	// prepare the message
	mlen = sizeof(raid_phyPassThruReqMsg_t) + dataLen - 1;
	req = (raid_phyPassThruReqMsg_t *)malloc(mlen);
	memset((void *)req, 0, mlen);
	req->header.mid = RAID_MID_PHY_PASSTHRU_REQ;
	req->header.cookie = 1;
	strcpy(req->header.id, argv[1]);
	req->channelNumber = atol(argv[2]);
	req->targetId = atol(argv[3]);
	req->cmd = RAID_PASSTHRU_WRITE_LONG;
	req->lba = strtoq(argv[4], NULL, 0);
	req->dataOut = 1;
	req->dataLen = dataLen;
	memcpy((void *)req->data, reqData, dataLen);
	
	// send it now
	rsplen = sizeof(raid_phyPassThruRspMsg_t);
	rsp = (raid_phyPassThruRspMsg_t *)malloc(rsplen);
	bzero(rsp, rsplen);
	rc = send_raid_message(req, mlen, (char *)rsp, &rsplen, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req->header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req->header.mid);
		return(NFX_ERR);
	}

	if (rsp->header.mid != RAID_MID_PHY_PASSTHRU_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_PHY_PASSTHRU_RSP, rsp->header.mid);
		return(NFX_ERR);
	}
	if (rsp->rCode != RAID_RCODE_SUCCESS) {
		printf("raid direct_write for dev %s failed: rCode %s(0x%x)\n", 
		       argv[1], raid_ecode_to_string(rsp->rCode), rsp->rCode);
		return(NFX_ERR);
	}

	printf("Write of %d bytes completed succesdfully\n", req->dataLen);
	return(NFX_OK);
}


int
raid_ui_get_modepage(int argc, char **argv, int timeout, ...)
{
	uint32						page_code = 0;
	raid_phyPassThruReqMsg_t	req;
	raid_phyPassThruRspMsg_t	*rsp;
	char						rspbuf[5000];
	uint32						rsplen = 0;
	int							rc;
	scsi_mode_header_t			*page_hdr;
	scsi_mode_block_desc_t		*block_desc;
	scsi_mode_page_desc_t		*page_data;
	char						*p;
	int							i;

	if(argc != 5) return 1;
	
	// prepare the message
	memset((void *)&req, 0, sizeof(req));
	req.header.mid = RAID_MID_PHY_PASSTHRU_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, argv[1]);
	req.channelNumber = atol(argv[2]);
	req.targetId = atol(argv[3]);
	page_code = atoi(argv[4]);
	
	req.cmd = RAID_PASSTHRU_MODE_SENSE;
	req.data[0] = page_code;

	// send it now
	rsplen = sizeof(rspbuf);
	rc = send_raid_message(&req, sizeof(req), rspbuf, &rsplen, timeout);
	if (rc < 0) {
		printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
		return(NFX_ERR);
	}
	else 
	if (rc == 0) {
		printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
		return(NFX_ERR);
	}

	rsp = (raid_phyPassThruRspMsg_t *)rspbuf;
	if (rsp->header.mid != RAID_MID_PHY_PASSTHRU_RSP) {
		printf("Invalid response mid; expected 0x%x got 0x%x\n",
		       RAID_MID_PHY_PASSTHRU_RSP, rsp->header.mid);
		return(NFX_ERR);
	}
	if (rsp->rCode != RAID_RCODE_SUCCESS) {
		printf("raid get_logpage failed for dev %s: rCode %s(0x%x)\n", 
		       argv[1], raid_ecode_to_string(rsp->rCode), rsp->rCode);
		return(NFX_ERR);
	}

	page_hdr = (scsi_mode_header_t *)&rsp->data[0];
	block_desc = (scsi_mode_block_desc_t *)(page_hdr + 1);
	page_data = (scsi_mode_page_desc_t *)(block_desc + 1);

	printf("Page hdr:         ");
	p = (char *)page_hdr;
	for (i = 0; i < sizeof(scsi_mode_header_t); ++i, ++p) 
		printf("%02x ", *p & 0xff);
	printf("\n");

	p = (char *)block_desc;
	printf("Block descriptor: ");
	for (i = 0; i < sizeof(scsi_mode_block_desc_t); ++i, ++p) 
		printf("%02x ", *p & 0xff);
	printf("\n");

	printf("Page Code:        %d (0x%x)\n", page_data->code & 0x7F, page_data->code & 0x7F);
	printf("Page Len:         %d\n", page_data->len);
	printf("Page Data:        0x%04x: ", 0);

	p = (char *)page_data->u.data;
	for (i = 0; i < page_data->len; ++i) {
		if (i > 0 && (i % 8) == 0)
			printf("\n                  0x%04x: ", i);
		printf("%02x ", page_data->u.data[i] & 0xff);
	}
	printf("\n");

	return(NFX_OK);
}

int 
raid_ui_update_drv_fw(int argc, char ** argv, int timeout, ...)
{
	raid_updateDrvFwReqMsg_t	req;
	raid_updateDrvFwRspMsg_t	rsp;
	raid_phyDevListReqMsg_t		physListReq;
	raid_phyDevListRspMsg_t	*	physListRsp;
	char						rspbuf[65536];
	int							i;
	struct stat					statBuf;
	int							rc;
	int							rsplen;

	if(argc != 3 && argc != 5) return 1;

	if(stat(argv[2], &statBuf) != 0) {
		printf("stat on Firmware file %s failed: %s\n", argv[2], strerror(errno));
		return (NFX_ERR);
	}

	if(argc == 5) { // if CID/TID is specified
		physListRsp = (raid_phyDevListRspMsg_t *)rspbuf;
		physListRsp->numEntries = 1;
		physListRsp->idList[0].channelNumber = atol(argv[3]);
		physListRsp->idList[0].targetId = atol(argv[4]);
	}

	else {
			
		// get list of all cid/tid pairs for this device
		physListReq.header.mid = RAID_MID_PHYDEV_LIST_REQ;
		physListReq.header.cookie = 1;
		strcpy(physListReq.header.id, argv[1]);
			
		rsplen = sizeof(rspbuf);
		rc = send_raid_message(&physListReq, sizeof(raid_phyDevListReqMsg_t), rspbuf, &rsplen, timeout);
		if (rc < 0) {
			printf("send_raid_message(mid 0x%x) failed.\n", physListReq.header.mid);
			return(NFX_ERR);
		}
		else 
		if (rc == 0) {
			printf("send_raid_message(mid 0x%x) timed out.\n", physListReq.header.mid);
			return(NFX_ERR);
		}
		physListRsp = (raid_phyDevListRspMsg_t *)rspbuf;
		if (physListRsp->header.mid != RAID_MID_PHYDEV_LIST_RSP) {
			printf("Invalid response mid; expected 0x%x got 0x%x\n",
			       RAID_MID_PHYDEV_LIST_RSP, physListRsp->header.mid);
			return(NFX_ERR);
		}
		if (physListRsp->rCode != RAID_RCODE_SUCCESS) {
			printf("raid physdev failed: rCode %s(0x%x)\n", 
			       raid_ecode_to_string(physListRsp->rCode), physListRsp->rCode);
			return(NFX_ERR);
		}
	}
	
	// now upgrade each drive
	for (i = 0; i < physListRsp->numEntries; ++i) {
		// prepare the message
		memset((void *)&req, 0, sizeof(req));
		req.header.mid = RAID_MID_UPDATE_DRV_FW_REQ;
		req.header.cookie = 1;
		strcpy(req.header.id, argv[1]);
		req.channelNumber = physListRsp->idList[i].channelNumber;
		req.targetId = physListRsp->idList[i].targetId;
		strncpy(req.filePath, argv[2], sizeof(req.filePath));

		printf("Updating firmware on cid/tid %d/%d...\n", req.channelNumber, req.targetId);

		// send it now
		memset((void *)&rsp, 0, sizeof(rsp));
		rsplen = sizeof(rsp);
		rc = send_raid_message((void *)&req, sizeof(req), &rsp, &rsplen, timeout);
		if (rc < 0) {
			printf("send_raid_message(mid 0x%x) failed.\n", req.header.mid);
			return(NFX_ERR);
		}
		else 
		if (rc == 0) {
			printf("send_raid_message(mid 0x%x) timed out.\n", req.header.mid);
			return(NFX_ERR);
		}
		if (rsp.header.mid != RAID_MID_UPDATE_DRV_FW_RSP) {
			printf("Invalid response mid; expected 0x%x got 0x%x\n",
		    	   RAID_MID_UPDATE_DRV_FW_RSP, rsp.header.mid);
			return(NFX_ERR);
		}
		if (rsp.rCode != RAID_RCODE_SUCCESS) {
			printf("raid update_drv_fw failed for dev %s: rCode %s(0x%x)\n\n", 
			       argv[1], raid_ecode_to_string(rsp.rCode), rsp.rCode);
			continue;
		}
	
		printf("Firmware successfully updated on %s cid/tid %d/%d\n\n", 
				argv[1], req.channelNumber, req.targetId);
	}

	return(NFX_OK);
}
