/*
 *-----------------------------------------------------------------
 *
 * Name:	sbm-hash.c
 *
 * RCS:  $Id: sbm-hash.c,v 1.7 2002/08/20 20:23:22 rick Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description: Functions that support the hash table mapping of COD
 *				handles to HMT indicies.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/4/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>

#ifdef SBM_REV2
#error Compiled with SBM_REV2
#endif

#ifdef NFP_TXRX

#include <exports/hash-api.h>
#include <exports/sbm-api.h>
#include "sbm.h"
#include "../sm-hash/hash.h"

/*-----------------------------------------------------------------
 * Name:	sbm_hashInit()
 *
 * Description:	called to allocate data structures and create the hash
 *				table for COD handle to HMT index mapping.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/4/01
 *
 *-----------------------------------------------------------------
 */
void
sbm_hashInit()
{
	uint32	i;

	sbm.sbm_hash = hash_create(	sbm.num_sac_blocks/4,
									sbm_hashFunc,
									sbm_hashCODCompare,
									sbm_hashRelease);

	sbm.hash_node = (sbm_hash_node_t *)eee_ramAllocZero(
								sbm.num_sac_blocks * sizeof(sbm_hash_node_t));

	sbm.hash_node_stack.stack = (sbm_hash_node_t **)
									eee_ramAllocZero(sbm.num_sac_blocks *
													sizeof(sbm_hash_node_t *));

	sbm.hash_node_stack.top = sbm.num_sac_blocks;

	for (i=0; i<sbm.num_sac_blocks; i++) {
		sbm_hashNodeStackPush(&(sbm.hash_node[i]));
	}

}	/* sbm_hashInit() */

/*-----------------------------------------------------------------
 * Name:	sbm_hashNodeStackPush(node_addr)
 *
 * Description:	put an SBM hash node back onto the free stack
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/4/01
 *
 *-----------------------------------------------------------------
 */
void
sbm_hashNodeStackPush(
	sbm_hash_node_t		* node_addr
)
{
	if (sbm.hash_node_stack.top == 0)
		panic("SBM hash stack full\n");
	else {
		sbm.hash_node_stack.top--;
		sbm.hash_node_stack.stack[sbm.hash_node_stack.top] = node_addr;
	}
	
}	/* sbm_hashNodeStackPush() */

/*-----------------------------------------------------------------
 * Name:	sbm_hashNodeStackPop()
 *
 * Description:	allocate an SBM hash node from the free stack
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/4/01
 *
 *-----------------------------------------------------------------
 */
sbm_hash_node_t		*
sbm_hashNodeStackPop()
{
	sbm_hash_node_t		* ret;

	if (sbm.hash_node_stack.top == sbm.num_sac_blocks)
		panic("SBM hash stack empty\n");
	else {
		ret = sbm.hash_node_stack.stack[sbm.hash_node_stack.top];
		sbm.hash_node_stack.top++;
		return (ret);
	}

}	/* sbm_hashNodeStackPop() */

/*-----------------------------------------------------------------
 * Name:	sbm_hashFunc()
 *
 * Description:	hashing function supplied to the general hash routines.
 *				This function will be called to return the hash bucket from
 *				the given key.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/4/01
 *
 *-----------------------------------------------------------------
 */
uint32
sbm_hashFunc(
	uint32	key
)
{
/*
	printf("key = 0x%x\n", key);
*/

	return (key % (sbm.num_sac_blocks/4));

}	/* sbm_hashFunc() */

/*-----------------------------------------------------------------
 * Name:	sbm_hashCODCompare()
 *
 * Description: comparision function supplied to the general hash routines.
 *				This function will be called to check if the search data
 *				matches the node data.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/4/01
 *
 *-----------------------------------------------------------------
 */
int32
sbm_hashCODCompare(
	sbm_hash_node_t	* ht_data,
	sbm_hash_node_t	* search_data
)
{
/*
	printf("comparing 0x%llx with 0x%llx\n", ht_data->COD, search_data->COD);
*/
	if (ht_data->COD == search_data->COD)
		return 0;
	else
		return 1;

}	/* sbm_hashCODCompare() */

/*-----------------------------------------------------------------
 * Name:	sbm_hashRelease()
 *
 * Description:	freeing function supplied to the general hash routines.
 *				Will free an SBM hash node back to the free stack.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	4/4/01
 *
 *-----------------------------------------------------------------
 */
void
sbm_hashRelease(
	sbm_hash_node_t	* data
)
{
	sbm_hashNodeStackPush(data);

}	/* sbm_hashRelease() */

#endif	/* NFP_TXRX */

