/*  FileName: devO_sm_funcs.c
 *
 *  Description: Contains action functions for the SCSI device state machine.
 *
 *	Created:	28 March 2001
 *
 *  Author:		Frank Salzmann
 *
 *  Copyright (c)  Agile Storage, Inc. 2001
 */			 

#include "scsi.h"
#include <exports/hal.h>
#include <exports/malloc-api.h>

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devO_create
 *
 * Description:     Called when new instance of device state machine is created.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 25, 2001
 *					
 * Returns:			nextState
 *------------------------------------------------------------------------------
 */
int
scsi_devO_create(esm_handle_t	handle,		/* handle for this sm instance */
				  int32			nextState,	/* next state from table	   */
				  esm_event_t	thisEvent,	/* event driving this action   */
				  void			*userInfo)	/* ctrl blk linked to sm	   */
				
{
	scsi_dev_t *pdev;

	/* Based on the device type plug in the proper cmd table */
	pdev = (scsi_dev_t *)userInfo;

	switch( pdev->type ) {

		case SCSI_DEVTYPE_TAPE:
		case SCSI_DEVTYPE_MEDIA_CHANGER:
			pdev->cmdTable 		 = scsi_tapeCmds;
			pdev->maxDispatchCnt = 1;

			// Disable SCSI level coallescing and sorting on tape drives 
			pdev->flags.fAllowSort  = 0;
			pdev->flags.fAllowChain = 0;
			pdev->flags.fSimEarlyEOM = 0;
			pdev->flags.fLocalCSVerify = 0;
			break;

		case SCSI_DEVTYPE_DISK:
		default:
			pdev->cmdTable 		 = scsi_diskCmds;
			pdev->maxDispatchCnt = scsi_scsi_getDeviceMaxTags(pdev);
			if (pdev->maxDispatchCnt < 0)
				pdev->maxDispatchCnt = SCSI_DISPATCH_CNT_DISK_DEFAULT;
			pdev->flags.fAllowChain = 1;
			pdev->flags.fAllowSort = 1;
			pdev->flags.fLocalCSVerify = 0;	/* Turn off local verification by default */
			break;

            //printf("scsi_devO_create: pdev wwn 0x%llx, type %d -- ignoring\n", pdev->wwn, pdev->type);
			/* VERIFY(0); */
	}
	return nextState;
} /* scsi_devO_create */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devO_procReq
 *
 * Description:     Handle SCSI device cmd requests while in the OPEN state.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 6, 2001
 *					
 * Returns:			nextState 
 *------------------------------------------------------------------------------
 */
int
scsi_devO_procReq(esm_handle_t	handle,		/* handle for this sm instance */
				  int32			nextState,	/* next state from table	   */
				  esm_event_t	thisEvent,	/* event driving this action   */
				  void			*userInfo)	/* ctrl blk linked to sm	   */
				
{
	int32		 rc;
	uchar8		 opCode;
	scsi_dev_t   *pDev;
	scsi_desc_t	 *sd;
	scsi_cmd_t   *cTab;

	scsi_debug("scsi_devO_procReq: [Enter]", SCSI_DBL_3);

	sd   = (scsi_desc_t *)thisEvent.p1.u.p;
	pDev = (scsi_dev_t *)userInfo;
	
	/*
	 * Call function to handle this particular command.
	 */
	cTab = pDev->cmdTable;
	if(!cTab) {
		scsi_sendCmdErrRsp(sd->eDesc->hdr.src_port,
						   sd->cHandle, SCSI_MSG_ECODE_NOT_SUPPORTED);
		scsi_deallocDesc(sd);
	}
		
	opCode = sd->control & SD_OPCODE_MASK; 

	if( !cTab[opCode].req_func ) {
		scsi_sendCmdErrRsp(sd->eDesc->hdr.src_port,
						   sd->cHandle, SCSI_MSG_ECODE_INVALID_MSG);
		scsi_deallocDesc(sd);
	}

	rc = (*cTab[opCode].req_func)(pDev, sd);
	if( rc != NFX_OK ) {
		scsi_sendCmdErrRsp(sd->eDesc->hdr.src_port, sd->cHandle, rc);
		scsi_deallocDesc(sd);
	}

	/* Enqueue descriptor */
	rc = scsi_addToReqQ(sd, pDev);

	/* Look for something to do */
	scsi_sendNextCmd(pDev);

	scsi_debug("scsi_devO_procReq: [Leave]", SCSI_DBL_3);
	return nextState;
} /* scsi_devO_procReq */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devO_procGenIOTimer
 *
 * Description:     Handle SCSI device general IO queue timer expiration while 
 *                  in the OPEN state.
 *
 * Created by:		Radek Aster
 *
 * Data Created:	August 27, 2001
 *					
 * Returns:			nextState 
 *------------------------------------------------------------------------------
 */
int
scsi_devO_procGenIOTimer(esm_handle_t	handle,		/* handle for this sm instance */
						int32			nextState,	/* next state from table	   */
						esm_event_t		thisEvent,	/* event driving this action   */
						void			*userInfo)	/* ctrl blk linked to sm	   */
				
{
	scsi_dev_t	*pdev = (scsi_dev_t *)userInfo;

	scsi_debug("scsi_devO_procGenIOTimer: [Enter]", SCSI_DBL_3);

	/* All descriptors on general IO queue should be dispatched ASAP */
	pdev->genio_q_immediate_schedule_count = scsi_queue_len(pdev->genio_q);

	/* Look for something to do */
	scsi_sendNextCmd(pdev);

	scsi_debug("scsi_devO_procGenIOTimer: [Leave]", SCSI_DBL_3);
	return nextState;
} /* scsi_devO_procReadTimer */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devO_procRsp
 *
 * Description:     Handle SCSI device command responses while in the OPEN 
 *					mode.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 11, 2001
 *					
 * Returns:			ESM_SCSI_STATE_ONLINE 
 *------------------------------------------------------------------------------
 */
int
scsi_devO_procRsp(esm_handle_t	handle,	    /* handle for this sm inst    */
		   	      int32			nextState,	/* next state from table	  */
				  esm_event_t	thisEvent,	/* event driving this action  */
				  void			*userInfo)	/* ctrl blk linked to sm	  */
				
{
	int32				data_failed_count = 0;
	int32				opCode;
	int32				nState = nextState; 
	uint64				cod;
	uint32				ecode = SCSI_MSG_ECODE_OK;
	eee_desc_t *		ed_chain;
	scsi_cmd_t *		cTab;
	scsi_desc_t	*		sd;
	scsi_dev_t	*		pdev;
	uint32				total_csum, fragment_csum, csum;
	scsi_mp_desc_t *	smpd = NULL;

	scsi_debug("scsi_devO_procRsp: [Enter]", SCSI_DBL_3);

	pdev = (scsi_dev_t *)userInfo;
	sd = (scsi_desc_t *)thisEvent.p1.u.p;

	/* 
	 * Attempt to lower no-reorder barrier.
	 */
	if (sd->control & SD_FLAG_NOREORDER) {
		VERIFY(pdev->noreorder_barrier == 1);
		--pdev->noreorder_barrier;
	}

	for (; sd; sd = sd->chain) {

		/*
		 * Get the message type
		 */
		opCode = sd->control & SD_OPCODE_MASK;			

		/* 
		 * Process the [device type dependant] error code, if any. 
		 */
		cTab = pdev->cmdTable;
		if (cTab[opCode].err_func)
		    ecode = (*cTab[opCode].err_func)(pdev, sd);

		/*
		 * Call proper message receive function to further
		 * digest the message.
		 */
		switch(opCode)
		{
		case SD_OPCODE_READ:
			pdev->readRspCnt++;
			if( scsi_pkt_trace ) {
				printf("SCSI Read: PktCtrl: %x \n", sd->eDesc->hdr.control);
				dp83820_printBuf( sd->eDesc->bd[1].buf, (sd->eDesc->bd[1].len > 64) ? 64 : sd->eDesc->bd[1].len, 0);
			}

			// if this was a huge io, then process it accordingly
			if(sd->control & SD_FLAG_HUGE_IO) {
				VERIFY(pdev->hugeio_state->callback);
				pdev->hugeio_state->callback(sd, pdev, ecode);
				break;
			}

			/*
			 * Map back to logical address/count.
			 */
			if (pdev->type == SCSI_DEVTYPE_DISK && !(sd->control & SD_FLAG_PHY_OP)) {
				/* 
				 * XXX Partitioning mapping will take place here. For now just
				 * offset. 
				 */
				sd->lba -= pdev->offset;
			}
			if (pdev->flags.fAllowBlock_CS && !(sd->control & SD_FLAG_PHY_OP))
				scsi_map_physical_to_logical(sd->lba, sd->lbCnt, &sd->lba, &sd->lbCnt);

			/* Don't attempt to transmit data if there is none to transmit */
			if (sd->eDesc && (sd->eDesc->hdr.control & EDESC_TOTAL_LEN)) {

				/*
				 * Extract checksum and verify valid blockno 
				 */
				if (pdev->flags.fAllowBlock_CS && !(sd->control & SD_FLAG_PHY_OP)) {
					uint32		 blockno;
					scsi_extract_block_checksum(sd->lba, sd->lbCnt, sd->eDesc, &total_csum, &blockno);
					if (total_csum && blockno != sd->lba) {
						pdev->readBECnt++;
						E_LOG (class_1, err_s, eee_getApplicationId("scsi"), 0, 0, 0, 
							   ("scsi: block ID verification failed on read request: "
								"device %016llx; lba %d, lbCnt %d, got %d.", 
								pdev->id, sd->lba, sd->lbCnt, blockno));
						printf("ERROR: Wrong blockno in CS block on read from device %016llx; lba %d, got %d\n",
							   pdev->id, sd->lba, blockno);
						if (scsi_debug_drop == 1) {
							NFX_ASSERT(FALSE);
						}
						ecode = SCSI_MSG_ECODE_READ_CHECKSUM_ERR;
					}

					/* 
					 * Verify checksum, if valid
					 */
					if (pdev->flags.fLocalCSVerify && total_csum) {
						csum = 0;
						ed_chain = sd->eDesc;
						while (ed_chain) {
							fragment_csum = scsi_compute_checksum(ed_chain);
							csum = uint32_checksum_add(csum, fragment_csum);
							if (ed_chain->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER)
								ed_chain = (eee_desc_t *)ed_chain->bd[EEE_MAX_BUFS-1].buf;
							else
								ed_chain = NULL;
						}

						if (csum != total_csum) {
							pdev->readCSECnt++;
							E_LOG (class_1, err_s, eee_getApplicationId("scsi"), 0, 0, 0,
								   ("scsi: checksum verification failed on read request: "
									"device %016llx; lba %d, lbCnt %d: got 0x%x, expected 0x%x.",
									pdev->id, sd->lba,  sd->lbCnt, total_csum, csum));
							printf("ERROR: Checksum mismatch on read from device %016llx; lba %d lbCnt %d: "
								   "got 0x%x, expected 0x%x\n", 
								   pdev->id, sd->lba, sd->lbCnt, total_csum, csum);
							if (scsi_debug_drop == 1) {
								NFX_ASSERT(FALSE);
							}
							ecode = SCSI_MSG_ECODE_READ_CHECKSUM_ERR;
						}
					}
				}
				else
					total_csum = 0;

				if (sd->eDesc->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER) { 
					smpd = scsi_MPEntryNew(pdev, sd->cHandle);
					VERIFY(smpd);
					ed_chain = sd->eDesc;
					for (smpd->maxSequenceId = 1; 
						 ed_chain->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER; 
						 ++smpd->maxSequenceId)
						ed_chain = (eee_desc_t *)ed_chain->bd[EEE_MAX_BUFS-1].buf;
					smpd->no_guaranteed_delivery = (sd->control & SD_FLAG_NO_GUAR_DEL);
					smpd->firstEDesc = sd->eDesc;
					smpd->accumulated_checksum = total_csum;
					smpd->eCode = ecode;
					sd->eDesc = NULL;
					smpd->private = sd;
					sd->control |= SD_FLAG_NO_DEALLOC;
				}
				else
					smpd = NULL;
				
				/* Multi-packet sequence ? */
				if (smpd == NULL) {
					if (sd->dataDest) {
						scsi_getCodHandle(&cod);
						scsi_sendReadData(pdev, sd, cod);
						if (sd->eDesc) 
							++data_failed_count;
						scsi_sendReadRsp(sd, (data_failed_count ? SCSI_MSG_ECODE_NO_RESOURCE : ecode), total_csum, cod);
					}
					else {
						scsi_sendReadRspWithData(sd, ecode, total_csum, 0, 0);
					}
				}
				else {
					uint64	data_offset = 0;
					VERIFY(sd->dataDest == 0);

					smpd->transactionId = scsi_getTransactionId();
					smpd->lastSequenceId = 1;

					ed_chain = smpd->firstEDesc;
					while (ed_chain) {
						/* Compute fragment checksum, if overall checksum is valid. */
						if (smpd->accumulated_checksum)
							csum = scsi_compute_checksum(ed_chain);
						sd->eDesc = scsi_copyEEEDesc(ed_chain);
						scsi_sendReadRspWithData(sd, smpd->eCode, csum,
												 SCSI_MSG_TRANSSEQID_PACK(smpd->transactionId,
																		  smpd->maxSequenceId,
																		  smpd->lastSequenceId), 
												 &data_offset);
						++smpd->lastSequenceId;

						if (ed_chain->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER) {
							VERIFY(!(ed_chain->bd[EEE_MAX_BUFS-1].control & EDESC_BD_CMD_IN_USE));
							ed_chain = (eee_desc_t *)ed_chain->bd[EEE_MAX_BUFS-1].buf;
						}
						else 
							ed_chain = NULL;
					} /* while (ed_chain) */
					if (smpd->no_guaranteed_delivery) {
						/* Free scsi desc */
						sd->eDesc = smpd->firstEDesc;
						sd->control &= ~SD_FLAG_NO_DEALLOC;

						/* Free smpd context */
						smpd->private = NULL;
						scsi_MPEntryDestroy(smpd);
					}
					else {
						scsi_start_ack_timer(smpd);
					}
				}
			}
			else {
				scsi_sendReadRsp(sd, ecode, 0, 0);
			}
			break;

		case SD_OPCODE_WRITE:
			pdev->writeRspCnt++;
			if( EEE_GET_SLOT_NUM(sd->rspDest) == 0 )
				scsiCtrl->stats.writeUserRspCnt_ssc++; 
			else
				scsiCtrl->stats.writeUserRspCnt_fp++; 

			// if this was a huge io, then process it accordingly
			if(sd->control & SD_FLAG_HUGE_IO) {
				VERIFY(pdev->hugeio_state->callback);
				pdev->hugeio_state->callback(sd, pdev, ecode);
				break;
			}

			/*
			 * Map back to logical address/count.
			 */
			if (pdev->type == SCSI_DEVTYPE_DISK && !(sd->control & SD_FLAG_PHY_OP)) {
				/* 
				 * XXX Partitioning mapping will take place here. For now just
				 * offset. 
				 */
				sd->lba -= pdev->offset;
			}
			if (pdev->flags.fAllowBlock_CS && !(sd->control & SD_FLAG_PHY_OP))
				scsi_map_physical_to_logical(sd->lba, sd->lbCnt, &sd->lba, &sd->lbCnt);

			scsi_sendWriteRsp(sd, ecode);
			break;

		case SD_OPCODE_RELAY:
			pdev->relayRspCnt++;
			scsiCtrl->stats.relayRspCnt++; 
			scsi_sendRelayRsp(sd);
			break;

		default:
			break;
		} /* end switch */

		/*
		 * Dealloc the sd and if present the eDesc
		 */
		if(!(sd->control & SD_FLAG_HUGE_IO))
			scsi_deallocDesc( sd );
	} /* for (sd) */

	if( pdev->flags.fCloseMode ) {
		if( pdev->dispatchCnt == 0 )
			nState = SCSI_DEVO_STATE_CLOSED;
		else
			nState = SCSI_DEVO_STATE_CLOSING;
	}
	else {
		/* Look for something to do */
		scsi_sendNextCmd(pdev);
	}
		
	scsi_debug("scsi_devO_procRsp: [Leave]", SCSI_DBL_3);
	return nState;
} /* scsi_devO_procRsp */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devO_procAck
 *
 * Description:  	
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 25, 2001
 *					
 * Returns:			
 *------------------------------------------------------------------------------
 */
int
scsi_devO_procAck(esm_handle_t handle,	  /* handle for this sm instance  */
				  int32		  nextState,  /* next state from table	      */
				  esm_event_t thisEvent,  /* event driving this action    */
				  void		  *userInfo)  /* ctrl blk linked to sm	      */
{

	scsi_debug("[scsi_devO_procAck]", SCSI_DBL_3);
	return nextState;

} /* scsi_devO_procAck */
/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devO_discAck
 *
 * Description:  	
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 25, 2001
 *					
 * Returns:			
 *------------------------------------------------------------------------------
 */
int
scsi_devO_discAck(esm_handle_t handle,	  /* handle for this sm instance  */
				  int32		  nextState,  /* next state from table	      */
				  esm_event_t thisEvent,  /* event driving this action    */
				  void		  *userInfo)  /* ctrl blk linked to sm	      */
{

	scsi_debug("[scsi_devO_discAck]", SCSI_DBL_3);
	return nextState;

} /* scsi_devO_discAck */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devO_rejectReq
 *
 * Description:  	Will reject an ULP request being sent to a SCSI device.   
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	March 26, 2001
 *					
 * Returns:			ESM_SCSI_STATE_ONLINE 
 *------------------------------------------------------------------------------
 */
int
scsi_devO_rejectReq(esm_handle_t handle,	/* handle for this sm instance  */
					int32		 nextState,	/* next state from table		*/
					esm_event_t	 thisEvent,	/* event driving this action	*/
					void		 *userInfo)	/* ctrl blk linked to sm		*/
{
	scsi_dev_t   *pDev;
	scsi_desc_t	 *sd;

	scsi_debug("scsi_devO_rejectReq: [Enter]", SCSI_DBL_3);

	sd   = (scsi_desc_t *)thisEvent.p1.u.p;
	pDev = (scsi_dev_t *)userInfo;

	scsi_sendCmdErrRsp(sd->eDesc->hdr.src_port, 
					   sd->cHandle,
					   SCSI_MSG_ECODE_DEV_CLOSED);
	scsi_deallocDesc(sd);

	scsi_debug("scsi_devO_rejectReq: [Leave]", SCSI_DBL_3);
	return nextState;
} /* scsi_devO_rejectReq */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devO_discRsp
 *
 * Description:  	Will discard a response from a scsi device.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	March 26, 2001
 *					
 * Returns:			ESM_SCSI_STATE_OFFLINE 
 *------------------------------------------------------------------------------
 */
int
scsi_devO_discRsp(esm_handle_t	handle,		/* handle for this sm instance  */
				  int32		    nextState,	/* next state from table		*/
				  esm_event_t	thisEvent,	/* event driving this action	*/
				  void			*userInfo)	/* ctrl blk linked to sm		*/
{

	return nextState;
} /* scsi_devO_discResp */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devO_close
 *
 * Description:  	Will remove a scsi device from service.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	March 26, 2001
 *					
 * Returns:			ESM_SCSI_STATE_NULL 
 *------------------------------------------------------------------------------
 */
int
scsi_devO_close(esm_handle_t handle,		/* handle for this sm instance  */
			    int32		 nextState,	/* next state from table		*/
			    esm_event_t	 thisEvent,	/* event driving this action	*/
			    void		 *userInfo)	/* ctrl blk linked to sm		*/
{
	int32  nState;
	uint32 cHandle, rspAddr;
	scsi_dev_t	*pdev;

	scsi_debug("scsi_devO_close: [Enter]", SCSI_DBL_3);
	pdev 	= (scsi_dev_t *)userInfo;
	cHandle = (uint32)thisEvent.p1.u.w;
	rspAddr = (uint32)thisEvent.p2.u.w;

	pdev->flags.fOpenMode  = 0;
	pdev->flags.fCloseMode = 1;

	/* NOTE: For now I will do a brute force close, in the future
	 * it might be nice to let the pending requests complete. 
	 */

	/* Blow away pending requests. */
	scsi_flushReqQ(pdev);
		
	if( pdev->dispatchCnt == 0 ) {
		if( rspAddr )
			scsi_sendCloseRsp(cHandle, SCSI_MSG_ECODE_OK, rspAddr);
		nState = SCSI_DEVO_STATE_CLOSED; 
	}
	else {
		/* Wait for the HAL to send up responses for any commands
		 * that have been dispatched to the device.
		 * Even if the device has gone away, the HAL timeout mechanisim
		 * will generate a response to the SCSI layer.
		 */
		nState = nextState; 
	}

	// Set flag that will cause this state machine instance to be deleted
	// the next time the event queue for this sm is serviced.
	esm_destroy(pdev->smHandle);

	scsi_debug("scsi_devO_close: [Leave]", SCSI_DBL_3);
	return nState;
} /* scsi_devO_close */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devO_delete
 *
 * Description:  	Will remove all data structures  and state machines
 *					associated with the device.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	August 29, 2001
 *					
 * Returns:			
 *------------------------------------------------------------------------------
 */
int
scsi_devO_delete(esm_handle_t handle,		/* handle for this sm instance  */
			    int32		 nextState,	/* next state from table		*/
			    esm_event_t	 thisEvent,	/* event driving this action	*/
			    void		 *userInfo)	/* ctrl blk linked to sm		*/
{
	uint32 cHandle, rspAddr;
	scsi_dev_t	*pdev;

	scsi_debug("scsi_devO_delete: [Enter]", SCSI_DBL_3);

	pdev 	= (scsi_dev_t *)userInfo;
	cHandle = (uint32)thisEvent.p1.u.w;
	rspAddr = (uint32)thisEvent.p2.u.w;

	pdev->flags.fOpenMode  = 0;
	pdev->flags.fCloseMode = 1;

	/* NOTE: For now I will do a brute force removal, in the future
	 * it might be nice to let the pending requests complete. 
	 */

	/* Blow away pending requests. */
	scsi_flushReqQ(pdev);
		
	// Set flag that will cause this state machine instance to be deleted
	// the next time the event queue for this sm is serviced.
	esm_destroy(pdev->smHandle);

	// Delete the device control block.
	scsi_deallocDevice( pdev );

	scsi_debug("scsi_devO_delete: [Leave]", SCSI_DBL_3);
	return nextState;
} /* scsi_devO_delete */
/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devO_bad_state
 *
 * Description:  	Handles state machine exceptions.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	March 26, 2001
 *					
 * Returns:			ESM_SCSI_STATE_NULL 
 *------------------------------------------------------------------------------
 */
int
scsi_devO_bad_state(esm_handle_t handle,	/* handle for this sm instance  */
					int32		 nextState,	/* next state from table		*/
		   	   		esm_event_t	 thisEvent,	/* event driving this action	*/
			   		void		 *userInfo)	/* ctrl blk linked to sm		*/
{
	return nextState;
} /* scsi_devO_bad_state */


/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devO_procWriteGateTimer
 *
 * Description: Handle SCSI device write gate timer expiration. Abort
 *              all queued writes and tell lower layer to abort all
 *              pending writes.
 *
 * Created by:		Radek Aster
 *
 * Data Created:	Sept. 10, 2002
 *					
 * Returns:			nextState 
 *------------------------------------------------------------------------------
 */
int
scsi_devO_procWriteGateTimer(esm_handle_t	handle,		/* handle for this sm instance */
							 int32			nextState,	/* next state from table	   */
							 esm_event_t	thisEvent,	/* event driving this action   */
							 void			*userInfo)	/* ctrl blk linked to sm	   */
				
{
	scsi_dev_t	* pdev = (scsi_dev_t *)userInfo;
	scsi_desc_t * sd;
	scsi_desc_t * sd_head;
	scsi_desc_t * sd_tail;

	VERIFY(pdev->write_gate_timer_running == 0);

	/* 
	 * For each of the 3 queues (hiPri, GenIO and other) pull all off
	 * queue, filter and abort writes and put everthing else back on
	 * again. 
	 */

	/* (1) HIPRI queue */
	sd_head = sd_tail = NULL;
	while ((sd = scsi_dequeue(pdev->hiprio_q))) {
		sd->next = sd->prev = NULL;
		if ((sd->control & SD_OPCODE_MASK) != SD_OPCODE_WRITE) {
			if (sd_head) {
				sd_tail->next = sd;
				sd_tail = sd;
			}
			else
				sd_head = sd_tail = sd;
		}
		else {
			scsi_sendCmdAbortedRsp(sd->rspDest, SCSI_MID_WRITE_RSP, sd->cHandle, SCSI_MSG_ECODE_WRITE_GATE_CLOSED);
			scsi_deallocDesc(sd);
		}
	} /* while (pdev->hiprio_q) */
	while ((sd = sd_head)) {
		sd_head = sd->next;
		sd->next = sd->prev = NULL;
		scsi_enqueue(pdev->hiprio_q, sd);
	}

	/* (2) GENIO queue */
	sd_head = sd_tail = NULL;
	while ((sd = scsi_dequeue(pdev->genio_q))) {
		sd->next = sd->prev = NULL;
		if ((sd->control & SD_OPCODE_MASK) != SD_OPCODE_WRITE) {
			if (sd_head) {
				sd_tail->next = sd;
				sd_tail = sd;
			}
			else
				sd_head = sd_tail = sd;
		}
		else {
			scsi_sendCmdAbortedRsp(sd->rspDest, SCSI_MID_WRITE_RSP, sd->cHandle, SCSI_MSG_ECODE_WRITE_GATE_CLOSED);
			scsi_deallocDesc(sd);
		}
	} /* while (pdev->genio_q) */
	while ((sd = sd_head)) {
		sd_head = sd->next;
		sd->next = sd->prev = NULL;
		scsi_enqueue(pdev->genio_q, sd);
	}

	/* (3) OTHER queue */
	sd_head = sd_tail = NULL;
	while ((sd = scsi_dequeue(pdev->other_q))) {
		sd->next = sd->prev = NULL;
		if ((sd->control & SD_OPCODE_MASK) != SD_OPCODE_WRITE) {
			if (sd_head) {
				sd_tail->next = sd;
				sd_tail = sd;
			}
			else
				sd_head = sd_tail = sd;
		}
		else {
			scsi_sendCmdAbortedRsp(sd->rspDest, SCSI_MID_WRITE_RSP, sd->cHandle, SCSI_MSG_ECODE_WRITE_GATE_CLOSED);
			scsi_deallocDesc(sd);
		}
	} /* while (pdev->other_q) */
	while ((sd = sd_head)) {
		sd_head = sd->next;
		sd->next = sd->prev = NULL;
		scsi_enqueue(pdev->other_q, sd);
	}

	/*
	 * Tell lower layer to kill all pending writes.
	 */
#if 0
	{
		hal_dispatch_entry_t *dte = hal_dt_entry_get();

		(*dte->hal_ioctl)(HAL_IOCTL_COMMON_KILL_WRITES, pdev->llHandle, pdev->lun);
	}
#endif

	return nextState;
} /* scsi_devO_procReadTimer */

