/*  FileName: hds9910_sm_func.c
 *
 *  Description: Contains action functions for the device init state machine.
 *
 *	Created:	18 May 2001
 *
 *  Author:		Frank Salzmann
 *
 *  Copyright (c)  Agile Storage, Inc. 2001
 */			 

#include "scsi.h"

/*
 *------------------------------------------------------------------------------
 * Function Name:	hds9910_I_create
 *
 * Description:     Device init state machine initialization function.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 18, 2001
 *					
 * Returns:			
 *------------------------------------------------------------------------------
 */
int32
hds9910_I_create(esm_handle_t	handle,		/* handle for this sm inst    */
				 int32			nextState,	/* next state from table	  */
				 esm_event_t	thisEvent,	/* event driving this action  */
				 void			*userInfo)	/* ctrl blk linked to sm  	  */
{
	int32		nState;

	scsi_desc_t *sd;
	scsi_dev_t *pdev;

	scsi_debug("[hds9910_I_create] Enter\n", SCSI_DBL_3);

	pdev = (scsi_dev_t *)userInfo;
	sd = (scsi_desc_t *)scsi_allocDesc();

	if( sd == (scsi_desc_t *)NULL )
	{
		scsi_err( SCSI_ERR_SD_ALLOC | SCSI_ERR_LEVEL_0, 
				  "[scsi_devI_create]");
		return SCSI_ERR_SD_ALLOC;
	}

	pdev->flags.fInitMode = 1;
	scsi_sendInquiry(pdev, sd); 

	scsi_debug("[hds9910_I_create] Leave\n", SCSI_DBL_3);	
	return(nextState);
} /* hds9910_I_create */

/*
 *------------------------------------------------------------------------------
 * Function Name:	hds9910_I_procInq
 *
 * Description:     Called when a INQUIRY response event is queued to the 
 *					state machine.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 18, 2001
 *					
 * Returns:			
 *------------------------------------------------------------------------------
 */
int
hds9910_I_procInq(esm_handle_t	handle,		/* handle for this sm inst    */
				  int32			nextState,	/* next state from table	  */
				  esm_event_t	thisEvent,	/* event driving this action  */
				  void			*userInfo)	/* ctrl blk linked to sm  */
				
{
	int			rc;
	scsi_dev_t*  pdev  	= (scsi_dev_t *)userInfo;
	scsi_desc_t* sd		= (scsi_desc_t *)thisEvent.p1.u.p;

	scsi_debug("[hds9910__procInq] Enter", SCSI_DBL_3);

	/*
	 * Call function to parse the INQUIRY response.
	 */
	rc = hds9910_parseInqRsp(pdev,sd);

	if( rc == SCSI_ERR_GENERIC ) {

		/*
		 * Dealloc the sd and if present the eDesc
		 */
		scsi_deallocDesc(sd);
		esm_destroy( pdev->smHandle );
		pdev->smHandle = 0;
		scsi_deallocDevice( pdev );
		return(nextState);
	}

	/*
	 * Release inquiry response packet descriptor.
	 */
	if( sd->eDesc)
		eee_freePacket(sd->eDesc);
	sd->eDesc = 0;

	/*
	 * Get SCSI descriptor ready for reuse.
	 */
	sd->status     = 0;
	sd->retryCnt   = 0;
	sd->timeOutCnt = 0;
	memset(&sd->cdb, 0, 16);

	/* 
	 * Perform next initialization step based on device type.
	 */
	//printf("scsi_dev_procInq: vendor %s, product %s, path 0x%x, scsi_id 0x%016llx, SCSI Type %d\n",
		  //pdev->vendor, pdev->model, pdev->path, pdev->id, pdev->type);
	switch(pdev->type)
	{
		case SCSI_DEVTYPE_DISK:
			scsi_sendReadCapacity(pdev, sd);
			break;

		default:
        {
			// Set flag to delete this state machine instance.
			// the next time the event queue for this sm is serviced.
			esm_destroy(pdev->smHandle);

			// Delete the device control block.
			scsi_deallocDevice( pdev );

			// Delete the scsi descriptor
			scsi_deallocDesc(sd);
		}
	}// End Switch 
	scsi_debug("[hds9910__procInq] Leave", SCSI_DBL_3);
	return HDS9910_I_STATE_RCAP_WAIT;
} /* hds9910_I_procInq */

/*
 *------------------------------------------------------------------------------
 * Function Name:	hds9910_parseInqRsp
 *
 * Description: Will parse the response to the
 *				SCSI INQUERY command returned by 
 *  			the SCSI target device.
 *
 * Returns: SCSI_GENERIC_ERROR | SCSI_OK
 *
 *------------------------------------------------------------------------------
 */
int
hds9910_parseInqRsp( scsi_dev_t *pdev, scsi_desc_t *sd )
{
	uint32 lun;
	uchar8 *rspBuf;
	uchar8 devType;
	uchar8 *mfgStr, *productIdStr;
	uchar8 ldev[5];
	uint64 wwn;
	
	/*
	 * If device returns an inquiry error then assume 
	 * no LUN is  present so abort the whole process 
	 * and remove the device.
	 */
	if( (sd->status & SD_DEV_STATUS_MASK) != SCSI_DEV_STATUS_GOOD )
		return SCSI_ERR_GENERIC;

	/*
	 * Get pointer to buffer containing the inquiry data.
	 */
	rspBuf = sd->eDesc->bd[1].buf;

	if( (rspBuf[0] >> 5) == 3 )		/* No Dev supported on this LUN */
		return SCSI_ERR_GENERIC;

	mfgStr   = &rspBuf[8];
	productIdStr = &rspBuf[16];
	
	/* devDbEntry = scsi_getDevDB(mfgStr, modelStr); */

	strncpy(pdev->vendor, mfgStr, 8);	
	strncpy(pdev->productId, productIdStr, 16);
	pdev->type = rspBuf[0] & 0x1F;
	
	//-- Extract the Hitachi vendor specific logical device id.
	ldev[0] = rspBuf[44];
	ldev[1] = rspBuf[45];
	ldev[2] = rspBuf[46];
	ldev[3] = rspBuf[47];
	ldev[4] = '\0';
	lun = atox(ldev);

	wwn = SCSI_ID_GET_WWN(pdev->id);
	wwn &= 0xFFFFFFFFFFFFFF00ll;
	SCSI_ID_SET_WWN(pdev->id, wwn);
	SCSI_ID_SET_LUN(pdev->id, lun);
	/*
	 * Device is ready to go so set flag.
	 */
	pdev->flags.fDeviceReady = 1;

	return SCSI_OK;
} /* hds9910_parseInqRsp */

/*
 *------------------------------------------------------------------------------
 * Function Name:	hds9910_I_procRCap
 *
 * Description:     Called when a READ_CAPACITY response event is queued 
 *					to the state machine.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 19, 2001
 *					
 * Returns:			nextState 
 *------------------------------------------------------------------------------
 */
int
hds9910_I_procRCap(esm_handle_t     handle,		/* handle for this sm inst    */
					  int32         nextState,	/* next state from table	  */
					  esm_event_t	thisEvent,	/* event driving this action  */
					  void          *userInfo)	/* ctrl blk linked to sm  */
				
{
	scsi_debug("[hds9910_I_procRCap] Enter", SCSI_DBL_3);
	 scsi_dev_procRCap( nextState,
						(scsi_dev_t *)userInfo, 
						(scsi_desc_t *)thisEvent.p1.u.p);
	scsi_debug("[hds9910_I_procRCap] Leave", SCSI_DBL_3);
	return HDS9910_I_STATE_ACK_WAIT;
} /* hds9910_I_procRCap */

/*
 *------------------------------------------------------------------------------
 * Function Name:	hds9910_I_procAck
 *
 * Description:  	
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 25, 2001
 *					
 * Returns:			
 *------------------------------------------------------------------------------
 */
int
hds9910_I_procAck(esm_handle_t handle,	  /* handle for this sm instance  */
				  int32		  nextState,  /* next state from table	      */
				  esm_event_t thisEvent,  /* event driving this action    */
				  void		  *userInfo)  /* ctrl blk linked to sm	      */
{
	scsi_dev_t *pdev;

	scsi_debug("[hds9910_I_procAck] Enter", SCSI_DBL_3);

	pdev = (scsi_dev_t *)userInfo;

	/* Initialization steps are complete. Terminate this state machine
	 * instance and place device into the close mode.
	 */
	pdev->flags.fInitMode  = 0;
	pdev->flags.fOpenMode  = 0;
	pdev->flags.fCloseMode = 1;

	esm_destroy( pdev->smHandle );
	pdev->smHandle = 0;
	scsi_debug("[hds9910_I_procAck] Leave", SCSI_DBL_3);
	return nextState;
} /* hds9910_I_procAck */

/*
 *------------------------------------------------------------------------------
 * Function Name:	hds9910_I_ignoreEvt
 *
 * Description:  	
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 25, 2001
 *					
 * Returns:			
 *------------------------------------------------------------------------------
 */
int
hds9910_I_ignoreEvt(esm_handle_t handle,	   
					int32		nextState, /* next state from table	      */
					esm_event_t	thisEvent, /* event driving this action   */
					void		*userInfo)  /* ctrl blk linked to sm	      */
{
	scsi_debug("[hds9910_I_ignoreEvt]", SCSI_DBL_3);
	printf("thisEvent: %d\n",thisEvent.event);
	return nextState;
} /* hds9910_I_ingnoreEvt */

/*
 *------------------------------------------------------------------------------
 * Function Name:	hds9910_I_bad_state
 *
 * Description:  	
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 25, 2001
 *					
 * Returns:			
 *------------------------------------------------------------------------------
 */
int
hds9910_I_bad_state(esm_handle_t handle,	   
					int32		nextState, /* next state from table	      */
					esm_event_t	thisEvent, /* event driving this action   */
					void		*userInfo)  /* ctrl blk linked to sm	      */
{
	scsi_debug("[hds9910_I_bad_state]", SCSI_DBL_3);
	return nextState;
} /* hds9910_I_bad_state */

/*
 *------------------------------------------------------------------------------
 * Function Name:	hds9910_I_procDel
 *
 * Description:  	
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 25, 2001
 *					
 * Returns:			
 *------------------------------------------------------------------------------
 */
int
hds9910_I_procDel(esm_handle_t handle,	   
					int32		nextState, /* next state from table	      */
					esm_event_t	thisEvent, /* event driving this action   */
					void		*userInfo)  /* ctrl blk linked to sm	      */
{
	int i;
	int32  nState;
	uint32 cHandle, rspAddr;
	scsi_dev_t	*pdev;

	pdev 	= (scsi_dev_t *)userInfo;

	/* Blow away pending requests. */
	scsi_flushReqQ(pdev);
		
	// Set flag that will cause this state machine instance to be deleted
	// the next time the event queue for this sm is serviced.
	esm_destroy(pdev->smHandle);

	// Delete the device control block.
	scsi_deallocDevice( pdev );

	return nextState;
} /* hds9910_I_procDel */

hds9910_parseReportLunRsp(scsi_dev_t *pdev, 
						  scsi_desc_t *sd, 
						  int32 *lunCnt, 
						  int32 **luns)
{
	int32   i,j,tmp, x;
	uchar8 *buf;
	
	scsi_debug("[scsi_parseReportLunRsp] Enter\n", SCSI_DBL_3);
	/*
	 * Get pointer to buffer containing the inquiry data.
	 */
	buf = sd->eDesc->bd[1].buf;

	tmp = ((buf[0] << 24) |
		   (buf[1] << 16) |
		   (buf[2] <<  8) |
		   (buf[3]));
	
	
	*luns = (int32 *)eee_ramAlloc(sizeof(uint32) * (tmp/8)); 
	if( !*luns ) {
		scsi_err( SCSI_ERR_MALLOC | SCSI_ERR_LEVEL_0, 
				  "[hds9910_parseReportLunRsp]");
		return NFX_ERR;
	}

	if( tmp )
		*lunCnt = tmp / 8;

	for(i=j=0; i < tmp; i+=8, j++) {
		x = ((buf[8+i] << 24) |
			 (buf[9+i] << 16) |
			 (buf[10+i] <<  8)|
			 (buf[11+i]));
		(*luns)[j] = x >> 16;
	}
	scsi_debug("[scsi_parseReportLunRsp] Leave\n", SCSI_DBL_3);
	return SCSI_OK;
} /* hds9910_parseReportLunRsp */

/*
 *------------------------------------------------------------------------------
 * Function Name:	hds9910_procRLun
 *
 * Description:     Called when a REPORT_LUN response event is queued 
 *					to the state machine.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	October 24, 2001
 *					
 * Returns:			nextState 
 *------------------------------------------------------------------------------
 */
int
hds9910_procRLun(int32		   nextState,	/* next state from table	  */
				  scsi_dev_t*  pdev,
				  scsi_desc_t* sd)
{
	int32		i, lunCnt, *luns;
	int32		myLunId, cmdStatus;
	uint32		llHandle;
	scsi_sense_t *sense;
	char        vendor[9];
	scsi_id_t   id;
	

	llHandle = pdev->llHandle;
	id = pdev->id;
	strcpy(vendor, pdev->vendor);

	//scsi_debug("[scsi_dev_procRLun] Enter", SCSI_DBL_3);
	//printf("[hds9910_procRLun] Enter");

	/* 
	 * Check command completion status.
	 */
	cmdStatus = sd->status & SD_DEV_STATUS_MASK;
	switch( cmdStatus )
	{
		case SCSI_DEV_STATUS_CHECK:
		case SCSI_DEV_STATUS_CMD_TERMINATED:
			if( sd->status & SD_DEV_SENSE_VALID ) {

				sense = (scsi_sense_t *)sd->senseData;

				/* Look at sense data */
				switch( sense->senseKey )
				{
					case SCSI_SENSE_KEY_NOT_READY:
					case SCSI_SENSE_KEY_HARDWARE_ERROR:
						/* Fatal Error unable to access device */
						scsi_deallocDesc( sd );
						break;

					case SCSI_SENSE_KEY_ILLEGAL_REQUEST:
						/* 
						 * Device does not support this command, so
						 * probe for n LUN's and see what comes
						 * back.
						 */ 
						for(i=1; i < SCSI_MAX_LUN_PROBE_CNT; i++)
							scsi_newDevice( SCSI_LLNOTIFY_FC,
											llHandle,
											SCSI_ID_GET_WWN(id),
											0, 0, i, vendor);
						scsi_deallocDesc(sd);
						break; 

					case SCSI_SENSE_KEY_NO_SENSE:
					case SCSI_SENSE_KEY_RECOVERED_ERROR:
					default:
						/* problem not fatal, fall through */
						break;
				}
			}
			//			scsi_sendAddDevMsg(pdev);
			break;

		case SCSI_DEV_STATUS_GOOD:
		case SCSI_DEV_STATUS_CONDITION_MET:
			/* Parse the REPORT_LUN response. */
			lunCnt = 0;
			luns   = (int32 *)NULL;
			hds9910_parseReportLunRsp(pdev,sd, &lunCnt, &luns);


			// Set flag that will cause this state machine instance to be deleted
			// the next time the event queue for this sm is serviced.
			esm_destroy(pdev->smHandle);

			// Delete the device control block.
			scsi_deallocDevice( pdev );

			// Get rid of response SD.
			scsi_deallocDesc( sd );

			/* For each LUN that is reported create SM instance. */
			myLunId = SCSI_ID_GET_LUN(id);
			//printf("**** Luns Found: %d ****** \n", lunCnt);
			for(i=0; i < lunCnt; i++) {

				scsi_newDevice( SCSI_LLNOTIFY_FC,
								llHandle,
								pdev->wwn,
								(uint64)0,
								0,
								*(luns+i), 
								vendor);

			}

			if( luns && (lunCnt > 0))
				eee_ramDealloc(luns);
			
			break;

		case SCSI_DEV_STATUS_ACA_ACTIVE:
			break;

		default:
			scsi_deallocDesc( sd );
	}
	// scsi_debug("[hds9910_procRLun] Leave", SCSI_DBL_3);
	//printf("[hds9910_procRLun] Leave");
	return nextState;
} /* hds9910_procRLun */
