/*  File Name  :	scsi-api.h
 *
 *  Description:    Contains macros, definitions, types and prototypes exported
 *  				to outside modules that need to interact
 *  				with the SCSI module.
 *
 *  Created    :    03 March 2001
 *
 *  Author	   :    Frank Salzmann
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *
 */

#ifndef SCSI_API_H_INCLUDED

#define SCSI_API_H_INCLUDED

#include <nfx-incl.h>
#include <exports/eee-api.h>

#define NFX_COMMON_BLKSZ		512

//
//-- Defines to allow SCSI application to propagate device state 
//-- up to the SDM running on the SSC.
//
#define SCSI_DEV_STATE_INIT			0
#define SCSI_DEV_STATE_OPEN			1
#define SCSI_DEV_STATE_CLOSED		2

/*
 *===================================================================
 * SCSI Peripheral Device Type Codes
 * As listed in ANSI X3.301-1997
 *===================================================================
 */
#define SCSI_DEVTYPE_DISK			0x0
#define SCSI_DEVTYPE_TAPE			0x1
#define SCSI_DEVTYPE_PRINTER		0x2
#define SCSI_DEVTYPE_PROCESSOR		0x3
#define SCSI_DEVTYPE_WORM			0x4
#define SCSI_DEVTYPE_CDROM			0x5
#define SCSI_DEVYTPE_SCANNER		0x6
#define SCSI_DEVTYPE_OPTICAL_DISK	0x7
#define SCSI_DEVTYPE_MEDIA_CHANGER	0x8
#define	SCSI_DEVTYPE_COMM_DEVICE	0x9
#define SCSI_DEVTYPE_PRINTER_PRE1	0xA		/* Printer pre-processor dev */
#define SCSI_DEVTYPE_PRINTER_PRE2	0xB		/* Printer pre-processor dev */
#define SCSI_DEVTYPE_RAID			0xC		/* SCSI-3 only				 */
#define SCSI_DEVTYPE_SES			0xD
#define SCSI_DEVTYPE_UNKNOWN		0x1F

/*
 *----------------------------------
 * SCSI Device Id
 *----------------------------------
 */
typedef uint64 scsi_id_t;

#define SCSI_ID_WWN_MASK	0x0000FFFFFFFFFFFFll
#define SCSI_ID_LUN_MASK	0x0FFF000000000000ll

#define SCSI_ID_LUN_SHIFT	48

#if 0
#define SCSI_ID_GET_WWN(_id_) ((_id_) & SCSI_ID_WWN_MASK)
#define SCSI_ID_GET_LUN(_id_)	\
    (((_id_) & SCSI_ID_LUN_MASK) >> SCSI_ID_LUN_SHIFT)

#define SCSI_ID_SET_WWN(_id_, _x_) { \
	(_id_) &= ~SCSI_ID_WWN_MASK;			  \
	(_id_) |= ((_x_) & SCSI_ID_WWN_MASK); \
}
#define SCSI_ID_SET_LUN(_id_, _x_)	  { \
    uint64 _xx_;							\
    _xx_ = _x_;								\
	(_id_) &= ~SCSI_ID_LUN_MASK;			\
    (_id_) |= (((_xx_) << SCSI_ID_LUN_SHIFT) & SCSI_ID_LUN_MASK); \
}
#endif

/*
 *----------------------------------
 * SCSI Device Path 
 *----------------------------------
 */
typedef uint32 scsi_path_t;

#define SCSI_PATH_HANDLE_MASK	0x0000FFFF
#define SCSI_PATH_LUN_MASK	 	0x00FF0000
#define SCSI_PATH_PORT_MASK	 	0x0F000000
#define SCSI_PATH_SLOT_MASK	 	0xF0000000

#define SCSI_PATH_LUN_SHIFT		16
#define SCSI_PATH_PORT_SHIFT	24
#define SCSI_PATH_SLOT_SHIFT	28

#define SCSI_PATH_GET_HANDLE(_p_) ((_p_) & SCSI_PATH_HANDLE_MASK)
#define SCSI_PATH_GET_LUN(_p_)	\
    (((_p_) & SCSI_PATH_LUN_MASK) >> SCSI_PATH_LUN_SHIFT)
#define SCSI_PATH_GET_SLOT(_p_)	\
    (((_p_) & SCSI_PATH_SLOT_MASK) >> SCSI_PATH_SLOT_SHIFT)
#define SCSI_PATH_GET_PORT(_p_)	\
    (((_p_) & SCSI_PATH_PORT_MASK) >> SCSI_PATH_PORT_SHIFT)

#define SCSI_PATH_SET_HANDLE(_p_, _x_) (_p_) |= ((_x_) & SCSI_PATH_HANDLE_MASK)
#define SCSI_PATH_SET_LUN(_p_, _x_)	 \
    (_p_) |= (((_x_) << SCSI_PATH_LUN_SHIFT) & SCSI_PATH_LUN_MASK)
#define SCSI_PATH_SET_SLOT(_p_, _x_)	  \
    (_p_) |= (((_x_) << SCSI_PATH_SLOT_SHIFT) & SCSI_PATH_SLOT_MASK)
#define SCSI_PATH_SET_PORT(_p_, _x_)	  \
    (_p_) |= (((_x_) << SCSI_PATH_PORT_SHIFT) & SCSI_PATH_PORT_MASK)


/*
 *==============================================================================
 * SCSI Descriptor
 *==============================================================================
 */

typedef struct _scsi_desc
{
	uint32		control;	
#define SD_OPCODE_MASK		0x000000FF
#define SD_OPCODE_RELAY		0x00000000
#define SD_OPCODE_READ		0x00000001
#define SD_OPCODE_WRITE		0x00000002
#define SD_OPCODE_INQUIRY	0x00000003
#define SD_OPCODE_RCAP		0x00000004
#define SD_OPCODE_RLUN		0x00000005

#define SD_FLAG_MASK 		0x000FFF00
#define SD_FLAG_NOREORDER	0x00000100
#define SD_FLAG_NOCOALLESCE	0x00000200
#define SD_FLAG_EMBEDDED_CS	0x00000400
#define SD_FLAG_BLOCK_CS	0x00000800
#define SD_FLAG_PHY_OP		0x00001000
#define SD_FLAG_EXPRESS		0x00002000
#define SD_FLAG_HUGE_IO		0x00004000
#define SD_FLAG_NO_DEALLOC	0x00008000
#define SD_FLAG_NO_GUAR_DEL	0x00010000
#define SD_CMDID_MASK		0xFF000000
#define SD_CMDID_SHIFT		24

	struct _scsi_dev_entry *pdev;		/* Pointer to device control struct */
	scsi_path_t	path;					/* SCSI device id					*/
	ushort16	halHandle;				/* HAL device handle				*/
	uint32 		cHandle;				/* Upper Layer Protocol handle		*/
	uint32		status;					/* Cmd completion/error status		*/
#define SD_DEV_SENSE_VALID		0x80000000
#define SD_DEV_DATA_VALID		0x40000000
#define SD_DEV_DATA_IN_EXPECT	0x20000000  /* Expect Data From device		*/
#define SD_DEV_DATA_OUT_EXPECT	0x10000000  /* Send Data To device			*/
#define SD_DEV_STATUS_MASK 		0x000000FF
#define SD_DRV_STATUS_MASK		0x0000FF00
#define SD_DEV_STATUS_SHIFT		0
#define SD_DRV_STATUS_SHIFT		8
 	uchar8      *senseData;				/* Device auto-sense data			*/
	uint32		expectDataLen;			/* Length of data expected from dev	*/
	uint32		lba;					/* Read/Write Logical Block Addr	*/
	uint32		lbCnt;					/* Read/Write Logical Block Count	*/
	uint32		lbCntChained;			/* Read/Write Logical Block Count	*/
	uint32		dataDest;				/* Read data destination			*/
	uint32		rspDest;				/* Response Message Destination		*/
	uint32		tmrHandle;				/* Request Timer Handle				*/
	uchar8		cdb[16];				/* SCSI command descriptor block	*/
	eee_desc_t	*eDesc;					/* Pointer to eee_descriptor		*/

	struct _scsi_desc *chain;			/* Used for coallescing descriptors */

	struct _scsi_desc *next;			/* Pointer to next descriptor		*/
	struct _scsi_desc *prev;			/* Pointer to prev descriptor       */
    eeeRTC_t	queued_timestamp;		/* Time when this descriptor queued */

	//-- Timestamping code.
	eeeRTC_t	cmdTS;					// track total time between req and rsp
	eeeRTC_t	toFC;					// track time for each leg of the journy
	eeeRTC_t	fromFC;					// track time for each leg of the journy

	/* These fields used primariy by HAL driver*/
	uint32		retryCnt;				/* Number of times retry attempted	*/
	int32		timeOutCnt;				/* Number of seconds before timeout */
	uint32      maxTimeToLive;			/* Num secs desc can live	120sec	*/

	/* HAL private */
	byte		state;		/* WAITING, BUSY or TIMED OUT */ 
	uint32		current_timeout_count;
	short16		current_to_retry_count;
	eeeRTC_t	first_issued_timestamp;
#if 1
    eeeRTC_t	sim_response_timestamp;
#endif
} scsi_desc_t;

#define SD_DRV_STATUS_ADAPTER_ERROR				1
#define SD_DRV_STATUS_TRANSPORT_ERROR			2
#define SD_DRV_STATUS_EXCESSIVE_LINK_ERRORS		3
#define SD_DRV_STATUS_EXCESSIVE_TIMEOUTS		4
#define SD_DRV_STATUS_INVALID_DEVICE			5
#define SD_DRV_STATUS_DEVICE_BUSY				6
#define SD_DRV_STATUS_LINK_DOWN					7
#define SD_DRV_STATUS_RESOURCE_EXHAUSTED		8
#define SD_DRV_STATUS_ABORTED					9

#define SD_SET_DRV_STATUS(_d_, _v_) { \
		(_d_)->status &= ~SD_DRV_STATUS_MASK; \
		(_d_)->status |= (((_v_) << SD_DRV_STATUS_SHIFT) & SD_DRV_STATUS_MASK); \
	}
#define SD_SET_DEV_STATUS(_d_, _v_) { \
		(_d_)->status &= ~SD_DEV_STATUS_MASK; \
		(_d_)->status |= (((_v_) << SD_DEV_STATUS_SHIFT) & SD_DEV_STATUS_MASK); \
	}
#define SD_SET_SENSE(_d_, _src_, _len_) { \
		(_d_)->status |= SD_DEV_SENSE_VALID; \
		memcpy((_d_)->senseData, (_src_), _len_); \
	}

#define SENSE_INFO(_p_) \
		(*(uint32 *)((_p_)[0]) << 24 | \
		 *(uint32 *)((_p_)[1]) << 16 | \
		 *(uint32 *)((_p_)[2]) <<  8 | \
		 *(uint32 *)((_p_)[3]) <<  0) 

#define SCSI_SET_CMDID( _x_, _val_ )    \
       _x_ |= ((_val_ << SD_CMDID_SHIFT) & SD_CMDID_MASK) \

#define SCSI_GET_CMDID( _x_ )   \
       ((_x_ & SD_CMDID_MASK) >> SD_CMDID_SHIFT) \

#define BLOCKSIZE_512					512
#define BLOCKSIZE_520					520
#define BLOCKSIZE_1024					1024
#define BLOCKSIZE_8192					8192

/*
 * Definitions for embedded (E) checksumming. Each physical sector is
 * formatted to 520 BPS, leaving 16 bytes of checksum for every 1024
 * bytes of data.  
 */
#define ECSUM_BLOCKSIZE					BLOCKSIZE_1024
#define ECSUM_BLOCK_SPAN				2
#define ECSUM_LEN						16

/*
 * Definitions for block (B) checksumming. A physical sector is
 * reserved for every 16 data sectors.  
 */
#define BCSUM_BLOCKSIZE					BLOCKSIZE_8192
#define BCSUM_BLOCK_SPAN				16
#define BCSUM_LEN						512

typedef struct {
		uint32    magic;		// used to validate checkSum and blockNo
		uint32	  checkSum;		// CRC over data
		uint32	  blockNo;		// source/destination block number
		uint32	  pad;
} scsi_cs_t;

#define SCSI_CS_MAGIC_REV_1		0x43535631
#define SCSI_CS_MAGIC_INVALID	0xdeadd00d

/*
 *==============================================================================
 * API Function Prototypes
 *==============================================================================
 */

/*
 *-------------------------------------------------------------------------
 * LL API FUNCTION: scsi_llDevNotify
 *-------------------------------------------------------------------------
 */ 
#define SCSI_LLNOTIFY_FC	0		/* Fibre Channel Transport Layer Type	 */
#define SCSI_LLNOTIFY_IP	1		/* IP Transport Layer				     */

#define SCSI_LLNOTIFY_DEV_DETECTED	0
#define SCSI_LLNOTIFY_DEV_REMOVED	1
#define SCSI_LLNOTIFY_DEV_PATH_ERR	2
#define SCSI_LLNOTIFY_REBALANCE		3

int 
scsi_llDeviceNotify( uint32   event,
				  uint32   tType,	   /* SCSI_LLNOTIFY_FC | SCSI_LLNOTIFY_IP */
				  ushort16 devHandle,  /* Dev handle used by lower layer drvr */
				  uint64   devWWN,
				  uint64   channelWWN,
				  char *   channelName,
				  uint32   lun
				  );

/*
 *----------------------------------
 * SCSI definition for timing stats
 *----------------------------------
 */


#define SCSI_RSP_BUCKET_1   100000    // usec
#define SCSI_RSP_BUCKET_2   300000    // usec
#define SCSI_RSP_BUCKET_3   400000    // usec
#define SCSI_RSP_BUCKET_4   900000    // usec
#define SCSI_RSP_BUCKET_5  1350000    // usec
#define SCSI_RSP_BUCKET_6  3200000    // usec

#define SCSI_COALLESCE_STATS_BUCKETS	16

typedef struct
{
    uint64  moreThanBucket1;
    uint64  moreThanBucket2;
    uint64  moreThanBucket3;
    uint64  moreThanBucket4;
    uint64  moreThanBucket5;
    uint64  moreThanBucket6;
    eeeRTC_t  maxTime;
    eeeRTC_t  sumTime;
    uint64  numTime;
} scsi_timeTab_t;

typedef struct {
	uint64	wwn;
	uint32	lun;
	uint32	readReqCnt;
	uint64 readByteCnt;
	uint64 readByteLast;
	uint32 readBytePerSecCur;
	uint32 readBytePerSecMax;
	uint32 readRspCnt;
	uint32 readRspLast;
	uint32 readIoPerSecCur;
	uint32 readIoPerSecMax;
	uint32 readCSECnt;
	uint32 readBECnt;
	uint32 writeSub8kCnt;
	uint32 writeSub8kPaddedCnt;
	uint32 writeReqCnt;
	uint64 writeByteCnt;
	uint64 writeByteLast;
	uint32 writeBytePerSecCur;
	uint32 writeBytePerSecMax;
	uint32 writeRspCnt;
	uint32 writeRspLast;
	uint32 writeIoPerSecCur;
	uint32 writeIoPerSecMax;
	uint32 writeCSECnt;	
	uint32 relayReqCnt;
	uint32 relayRspCnt;
	uint32 devErrCnt;
	uint32 genio_q_len;
	uint32 hiprio_q_len;
	uint32 other_q_len;
	uint32 r_chain_stats[SCSI_COALLESCE_STATS_BUCKETS+1];
	uint32 w_chain_stats[SCSI_COALLESCE_STATS_BUCKETS+1];
} dev_stats_t;


typedef struct _scsi_stats 
{
	uint32 readUserReqCnt_ssc;
	uint32 readUserRspCnt_ssc;
	uint64 readUserBytes_ssc;
	uint32 readUserReqCnt_fp;
	uint32 readUserRspCnt_fp;
	uint64 readUserBytes_fp;
	eeeRTC_t readUserTotalTime;
	eeeRTC_t readUserFCTime;
	uint32 writeUserReqCnt_ssc;
	uint32 writeUserRspCnt_ssc;
	uint64 writeUserBytes_ssc;
	uint32 writeUserReqCnt_fp;
	uint32 writeUserRspCnt_fp;
	uint64 writeUserBytes_fp;
	eeeRTC_t writeUserTotalTime;
	eeeRTC_t writeUserFCTime;
	uint32 readMetaReqCnt_ssc;
	uint32 readMetaRspCnt_ssc;
	uint64 readMetaBytes_ssc;
	uint32 readMetaReqCnt_fp;
	uint32 readMetaRspCnt_fp;
	uint64 readMetaBytes_fp;
	eeeRTC_t readMetaTotalTime;
	eeeRTC_t readMetaFCTime;
	uint32 writeMetaReqCnt;
	uint32 writeMetaRspCnt;
	eeeRTC_t writeMetaTotalTime;
	eeeRTC_t writeMetaFCTime;
	uint32 relayCnt;
	uint32 relayRspCnt;
	eeeRTC_t relayTotalTime;
	eeeRTC_t relayFCTime;
	uint32 openReqCnt;
	uint32 closeReqCnt;
	uint32 syncReqCnt;
	uint32 statusReqCnt;
	uint32 ioctlReqCnt;
	uint32 statsReqCnt;
	uint32 deleteReqCnt;
	uint32 addReqCnt;
	uint32 invalidMsgCnt;
	uint32 sscMsgSentCnt;
	uint32 sscMsgQueueSize;
	scsi_timeTab_t  totReadDataRspBuckets;
	scsi_timeTab_t  totWriteRspBuckets;
	scsi_timeTab_t  fcReadDataRspBuckets;
	scsi_timeTab_t  fcWriteRspBuckets;
} scsi_stats_t;

#include <exports/scsi-msg.h>

int scsi_llReceive(scsi_desc_t *);
int scsi_llDeviceNotify(uint32, uint32, ushort16, uint64, uint64, char *, uint32);
void scsi_displayDesc(scsi_desc_t *);

#endif  /* SCSI_API_H_INCLUDED */
