/*  FileName: scsi-ll.c
 *
 *  Description: Contains routines that provide the SCSI API to lower layer 
 *				 modules such as the fibre channel and iSCSI drivers.
 *
 *	Created:	17 May	2001
 *
 *  Author:		Frank Salzmann
 *
 *  Copyright (c)  Agile Storage, Inc. 2001
 */			 

#include "scsi.h"
#include <exports/hal.h>


int dp83820_printBuf(char *buf, int len, int fDecode);
/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_llDeviceNotify
 *
 * Description: 	
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 17, 2001
 *
 * Returns:         NFX_OK | NFX_ERR 
 *
 *------------------------------------------------------------------------------
 */
int 
scsi_llDeviceNotify(uint32 event,
				 	uint32 tType,		/* Transport Id 					*/
				 	ushort16 llHandle,	/* Transport driver device handle 	*/
				 	uint64 devWWN,		/* Device WWN						*/
				 	uint64 chnWWN,		/* Channel port WWN					*/
				 	char * chnName,		/* Channel electrical name			*/
					uint32 lun)
{
	switch(event) 
	{
		case SCSI_LLNOTIFY_DEV_DETECTED:
#if SDM_DEV_INIT
			scsi_newDevInit(tType, llHandle, devWWN, chnWWN, chnName,lun);
#else
			scsi_newDevice(tType, llHandle, devWWN, chnWWN, chnName,lun,"Null");
#endif
			break;

		case SCSI_LLNOTIFY_DEV_REMOVED:
			scsi_deleteDevicell(tType, llHandle, devWWN, lun);

			break;

		case SCSI_LLNOTIFY_DEV_PATH_ERR:
			scsi_pathErr(llHandle, devWWN);
			break;

		case SCSI_LLNOTIFY_REBALANCE:
			scsi_sendRebalanceMsg();
			break;
	}
	return(NFX_OK);
} /* scsi_llDeviceNotify */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_dispatchCmd
 *
 * Description: 	Used by the scsi driver to send a command to  a device via 
 *					the sevice delivery	sub-system. It will tag the command with
 * 					and id and place it in the cmdOutTable. Then the cmd will 
 *					be sent to the transport layer.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 05, 2001
 *
 * Returns:         NFX_OK | NFX_ERR 
 *
 *------------------------------------------------------------------------------
 */
int 
scsi_dispatchCmd( scsi_desc_t *sd, scsi_dev_t *pDev )
{
#ifdef SCSI_TIMESTAMP
	eeeRTC_t t;
#endif
	scsi_desc_t *sd_tmp;
    hal_dispatch_entry_t *dte = hal_dt_entry_get();

	//scsi_debug("scsi_dispatchCmd", SCSI_DBL_3);

	sd->halHandle = pDev->llHandle;
	
	if( scsi_pkt_trace ) {
		int i;
		printf("\n================Dispatch Cmd Trace ===========\n");
		printf("\nSCSI CDB: ");
		for(i=0; i < 16; i++)
			printf("%02x ", sd->cdb[i]);
		printf("\n");

		if( (sd->control & SD_OPCODE_MASK) == SD_OPCODE_WRITE ) {
			printf("SCSI Write Data: PktCtrl: %x", sd->eDesc->hdr.control);
			dp83820_printBuf( sd->eDesc->bd[1].buf, 64, 0);
		}
		else
		if( (sd->status & SD_DEV_DATA_OUT_EXPECT) ) {
			printf("SCSI Cmd Data Out: PktCtrl: %x", sd->eDesc->hdr.control);
			dp83820_printBuf( sd->eDesc->bd[1].buf, sd->eDesc->bd[1].len, 0);
		}
	}
#ifdef SCSI_TIMESTAMP
	eee_rtc(&t);
	for(sd_tmp = sd; sd_tmp; sd_tmp = sd_tmp->chain ) {
		sd_tmp->toFC = t;
	}
#endif

	/* If tape I/O, reject if at EOM */
	if (pDev->type == SCSI_DEVTYPE_TAPE && 
		(sd->control & SD_OPCODE_MASK) == SD_OPCODE_WRITE && 
		pDev->eom) {
			scsi_sendCmdAbortedRsp(sd->rspDest, 
								   SCSI_MID_WRITE_RSP,
								   sd->cHandle,
								   SCSI_MSG_ECODE_END_OF_MEDIUM);
			scsi_deallocDesc(sd);
			return(NFX_OK);
	}

	/*
	 * Call transport layer specific request function.
	 */
	(*dte->hal_scsi_send)(sd);
	pDev->dispatchCnt++;

	return( NFX_OK );
} /* scsi_dispatchCmd */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_addToDispatchQ
 *
 * Description: 	Saves command sent to device. Returns cmdId.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 04, 2001
 *					
 *------------------------------------------------------------------------------
 */
int
scsi_addToDispatchQ(scsi_dev_t *pdev, scsi_desc_t *sd)
{
	int i, fsave, rc;

	fsave = 0;
	for(i=0; i < SCSI_MAX_CMD_PER_DEV; i++) {
		if( pdev->dispatchQ[i] )
			continue;
		pdev->dispatchQ[i] = sd;
		fsave = 1;
		break;
	}
	if( fsave ) {
		SCSI_SET_CMDID(sd->control, i);
		rc = NFX_OK;
	}
	else {
		scsi_err( SCSI_ERR_GENERIC | SCSI_ERR_LEVEL_0, 
				  "scsi_saveCmdOut: dispatchQ Full");
		rc = SCSI_ERR_GENERIC;
	}
	
	return(rc);
} /* scsi_addToDispatchQ */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_llReceive
 *
 * Description: 	Called by the sevice delivery sub-system to send command 
 *					status and data to the scsi driver.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	March 23, 2001
 *
 * Returns:         NFX_OK | NFX_ERR 
 *
 *------------------------------------------------------------------------------
 */
int 
scsi_llReceive( scsi_desc_t *sd )
{
	scsi_desc_t *sd_tmp;
	scsi_dev_t  *pdev;
#ifdef SCSI_TIMESTAMP
	eeeRTC_t    t;
#endif

	//scsi_debug("[scsi_llReceive] Enter", SCSI_DBL_3);
	/*
	 * Get pointer to SCSI Device Context Block
	 */
	pdev = sd->pdev;

#ifdef SCSI_TIMESTAMP
	eee_rtc(&t);
	for(sd_tmp = sd; sd_tmp; sd_tmp = sd_tmp->chain ) {
		sd_tmp->fromFC = t;
	}
#endif

	if( scsi_pkt_trace ) {
		printf("\n================ Scsi LL Receive Trace ===========\n");
		printf("\nSCSI Dev Return Status: %x\n", sd->status);
		if( (sd->status & SD_DEV_DATA_VALID) ) {
			printf("SCSI Cmd Data In: PktCtrl: %x", sd->eDesc->hdr.control);
			dp83820_printBuf( sd->eDesc->bd[1].buf, sd->eDesc->bd[1].len, 0);
		}
	}

	//-- Decrement number of outstanding cmds for this device.
	pdev->dispatchCnt--;

	/*
	 * Pass sd to processing function based on current 
	 * device mode.
	 */
	if( pdev->flags.fOpenMode )
		scsi_openModeProc(pdev, sd);
	else
	if( pdev->flags.fInitMode ) {
#if SDM_DEV_INIT
		scsi_devInitModeProc(pdev, sd);
#else
		scsi_initModeProc(pdev, sd);
#endif
	}
	else
		scsi_deallocDesc(sd);

	//scsi_debug("[scsi_llReceive] Leave", SCSI_DBL_3);
	return(NFX_OK); 
} /* scsi_llReceive */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_openModeProc
 *
 * Description:		Sends cmd response event to the dev open mode 
 *					state machine.	
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 19, 2001
 *
 * Returns:         NFX_OK | NFX_ERR 
 *
 *------------------------------------------------------------------------------
 */
int
scsi_openModeProc(scsi_dev_t *pdev, scsi_desc_t *sd)
{
    esm_event_t esmEvent;

	// scsi_debug("[scsi_openModeProc] Enter", SCSI_DBL_3);

    esmEvent.event = SCSI_DEVO_EVENT_REQ_DONE;
    esmEvent.p1.u.p = sd;
    esmEvent.p2.u.p = NULL;

    esm_deliverEvent(pdev->smHandle, &esmEvent);

	// scsi_debug("[scsi_openModeProc] Leave", SCSI_DBL_3);
	return NFX_OK;
} /* scsi_openModeProc */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_initModeProc
 *
 * Description: 	Sends cmd response event to the device init mode 
 *					state machine.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 19, 2001
 *					
 *------------------------------------------------------------------------------
 */
int
scsi_initModeProc(scsi_dev_t *pdev, scsi_desc_t *sd)
{
#if 0 //-- NO LONGER USED
	int32       opCode;
	int32       event;
    esm_event_t esmEvent;

	//scsi_debug("[scsi_initModeProc] Enter", SCSI_DBL_3);
	opCode = sd->control & SD_OPCODE_MASK;

	switch( opCode )
	{
		case SD_OPCODE_INQUIRY:
			event = SCSI_DEVI_EVENT_INQ_RSP;
			break;

		case SD_OPCODE_RCAP:
			event = SCSI_DEVI_EVENT_RCAP_RSP;
			break;

		case SD_OPCODE_RLUN:
			event = SCSI_DEVI_EVENT_RLUN_RSP;
			break;
			
		default:
			event = 999;
	}

    esmEvent.event = event;
    esmEvent.p1.u.p = sd;
    esmEvent.p2.u.p = NULL;
    esm_deliverEvent(pdev->smHandle, &esmEvent);

	//scsi_debug("[scsi_initModeProc] Leave", SCSI_DBL_3);
#endif
	return NFX_OK;
} /* scsi_initModeProc */



/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_remCmdOut
 *
 * Description: 	Removes command from the command out table.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 04, 2001
 *					
 *------------------------------------------------------------------------------
 */
int
scsi_remFromDispatchQ(scsi_dev_t *pdev, scsi_desc_t *sd)
{
	int cmdId;

	cmdId = SCSI_GET_CMDID(sd->control);
	if( pdev->dispatchQ[cmdId] != sd ) {
		scsi_err( SCSI_ERR_GENERIC | SCSI_ERR_LEVEL_0, 
				  "scsi_remCmdOut: SCSI descriptor mismatch.");
		return(SCSI_ERR_GENERIC);
	}
	else {
		pdev->dispatchQ[cmdId] = (scsi_desc_t *)NULL;
		pdev->dispatchCnt--;
	}
	return(NFX_OK);
} /* scsi_remCmdOut */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_sendNextCmd
 *
 * Description: 	Sends the next command waiting on the command queue.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 24, 2001
 *					
 *------------------------------------------------------------------------------
 */
int
scsi_sendNextCmd(scsi_dev_t *pDev)
{
    scsi_desc_t *	sd = NULL;
    scsi_desc_t *	sd_next;
    scsi_desc_t *	sd_chain_tail;
	uint32			chain_len;
	scsi_cmd_t *	cTab;
	uchar8			opCode;

	/*
	 * Rules of dispatch:
	 *
	 * (1) No more than pDev->maxDispatchCnt descriptors outstanding
	 *     at any time, unless "barrier is up", in which case the
	 *     device is single threaded.
	 *
	 * (2) Arriving read or write descriptors with the NOREORDER or
	 *     HIPRIORITY attributes are queued in LILO order on the
	 *     "hiprio" queue. Read or write descriptors with neither the
	 *     NOREORDER nor the HIPRIORITY attributes are sorted into the
	 *     "genio" queue. All other arriving descriptors are queued in
	 *     LILO order onto the "other" queue.
	 *
	 * (3) The "hiprio" queue takes precedence over all other
	 *     queues. Descriptors are dispatched off it first and an
	 *     attempt is made to coallesce descriptors of the same type.
	 *
	 * (4) If the "hiprio" queue is empty, a descriptor is dispatched
	 *     off the general "genio", if either (a) the queue length
	 *     exceeds the hiwater mark, or (b) the stall dispatch timer
	 *     has popped. In this case the queue is emptied, or a close
	 *     approximation thereof.
	 *
	 * (5) A descriptor is dispatched off the "other" queue if nothing
	 *     else can be found to be dispatched.
	 *
	 * (6) XXXX TODO: Starvation timers not yet implemented.  */

    while( pDev->dispatchCnt < pDev->maxDispatchCnt ) {

		/* HIPRIO queue always takes precedence */
		if ((sd = scsi_queue_head(pDev->hiprio_q)) &&
			(!(sd->control & SD_FLAG_NOREORDER) || pDev->noreorder_barrier < MAX_NOREORDER_BARRIER_COUNT)) {
			sd = scsi_dequeue(pDev->hiprio_q);
			VERIFY(sd->chain == NULL);

			/* See if we can chain IOs */
			sd->lbCntChained = sd->lbCnt;
			sd_chain_tail = sd;
			chain_len = 0;
			while (!(sd->control & SD_FLAG_NOCOALLESCE) &&
				   (sd_next = scsi_queue_head(pDev->hiprio_q)) &&
				   !(sd_next->control & SD_FLAG_NOCOALLESCE) &&
				   (sd_next->control & SD_OPCODE_MASK) == (sd->control & SD_OPCODE_MASK) &&
				   (sd_next->control & SD_FLAG_NOREORDER) == (sd->control & SD_FLAG_NOREORDER) &&
				   sd_next->lba == sd_chain_tail->lba + sd_chain_tail->lbCnt &&
				   chain_len < SCSI_MAX_CHAIN_LEN) {
			    sd_chain_tail->chain = scsi_dequeue(pDev->hiprio_q);
				sd_chain_tail = sd_chain_tail->chain;
				VERIFY(sd_chain_tail->chain == NULL);
				sd->lbCntChained += sd_chain_tail->lbCnt;
				++chain_len;
			}

			/* Recompute CDB and length as necesary */
			if (sd->chain) {
				cTab = pDev->cmdTable;
				VERIFY(cTab);
				opCode = sd->control & SD_OPCODE_MASK; 
				VERIFY(opCode == SD_OPCODE_READ || opCode == SD_OPCODE_WRITE);
				(*cTab[opCode].req_func)(pDev, sd);
			}

			/* Raise no-reorder barrier if necessary */
			if (sd->control & SD_FLAG_NOREORDER)
				++pDev->noreorder_barrier;

			/* Update stats */
			if (chain_len--) {
				if ((sd->control & SD_OPCODE_MASK) == SD_OPCODE_READ)
					++pDev->r_chain_stats[chain_len >= SCSI_COALLESCE_STATS_BUCKETS ? 
										 SCSI_COALLESCE_STATS_BUCKETS :
										 chain_len];
				else
					++pDev->w_chain_stats[chain_len >= SCSI_COALLESCE_STATS_BUCKETS ?
										 SCSI_COALLESCE_STATS_BUCKETS :
										 chain_len];
			}
		}

		else
			
		/* Has general IO queue reached hi-water mark? */
		/* Has general IO queue stall timer popped? */
#if (SCSI_GENIO_TIMER_DURATION == 0)
		    if (scsi_queue_len(pDev->genio_q)) {
#else
			if (scsi_queue_len(pDev->genio_q) >= pDev->genio_q_hiwater_mark ||
				pDev->genio_q_immediate_schedule_count) {
#endif
			sd = scsi_dequeue(pDev->genio_q);
			VERIFY(sd->chain == NULL);
			if (pDev->genio_q_immediate_schedule_count)
				--pDev->genio_q_immediate_schedule_count;

			/* See if we can chain IOs */
			sd->lbCntChained = sd->lbCnt;
			sd_chain_tail = sd;
			chain_len = 0;
			while (!(sd->control & SD_FLAG_NOCOALLESCE) &&
				   (sd_next = scsi_queue_head(pDev->genio_q)) &&
				   !(sd_next->control & SD_FLAG_NOCOALLESCE) &&
				   (sd_next->control & SD_OPCODE_MASK) == (sd->control & SD_OPCODE_MASK) &&
				   sd_next->lba == sd_chain_tail->lba + sd_chain_tail->lbCnt &&
				   chain_len < SCSI_MAX_CHAIN_LEN) {
			    sd_chain_tail->chain = scsi_dequeue(pDev->genio_q);
				sd_chain_tail = sd_chain_tail->chain;
				VERIFY(sd_chain_tail->chain == NULL);
				if (pDev->genio_q_immediate_schedule_count)
					--pDev->genio_q_immediate_schedule_count;
				sd->lbCntChained += sd_chain_tail->lbCnt;
				++chain_len;
			}

			/* Recompute CDB and length as necesary */
			if (sd->chain) {
				cTab = pDev->cmdTable;
				VERIFY(cTab);
				opCode = sd->control & SD_OPCODE_MASK; 
				VERIFY(opCode == SD_OPCODE_READ || opCode == SD_OPCODE_WRITE);
				(*cTab[opCode].req_func)(pDev, sd);
			}

			/* Update stats */
			if (chain_len--) {
				if ((sd->control & SD_OPCODE_MASK) == SD_OPCODE_READ)
					++pDev->r_chain_stats[chain_len >= SCSI_COALLESCE_STATS_BUCKETS ? 
										 SCSI_COALLESCE_STATS_BUCKETS :
										 chain_len];
				else
					++pDev->w_chain_stats[chain_len >= SCSI_COALLESCE_STATS_BUCKETS ?
										 SCSI_COALLESCE_STATS_BUCKETS :
										 chain_len];
			}
		}

		else

		/* Something on "other" queue? */
		if (scsi_queue_len(pDev->other_q)) {
			sd = scsi_dequeue(pDev->other_q);
		}

		else 

		/* Nothing to do for now .... */
		break;

		scsi_dispatchCmd(sd, pDev);
	
    }
    return(0);
} /* scsi_sendNextCmd */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_addToReqQ
 *
 * Description: 	Places command descriptor on device request queue.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 24, 2001
 *					
 *------------------------------------------------------------------------------
 */
int
scsi_addToReqQ(scsi_desc_t *sd, scsi_dev_t *pdev)
{
	if ((sd->control & SD_OPCODE_MASK) == SD_OPCODE_READ || 
		(sd->control & SD_OPCODE_MASK) == SD_OPCODE_WRITE) {
		if ((sd->control & SD_FLAG_NOREORDER) || pdev->flags.fAllowSort == 0)
			scsi_enqueue(pdev->hiprio_q, sd);
		else
			scsi_enqueue(pdev->genio_q, sd);
	}
	else 
		scsi_enqueue(pdev->other_q, sd);
	
	return 0;
} /* scsi_addToReqQ */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_genIOQueueTimerHandler
 *
 * Description: 	SCSI general IO Queue stall timer handler.
 * 				
 * Created by:		Radek Aster
 *
 * Data Created:	August 27, 2001
 *					
 *------------------------------------------------------------------------------
 */
int
scsi_genIOQueueTimerHandler(scsi_dev_t *pdev)
{
    esm_event_t esmEvent;

	//	scsi_debug("[scsi_genIOQueueTimerHandler] Enter", SCSI_DBL_0);

    esmEvent.event = SCSI_DEVO_EVENT_GENIO_TIMER;
    esmEvent.p1.u.p = NULL;
    esmEvent.p2.u.p = NULL;
    esm_deliverEvent(pdev->smHandle, &esmEvent);

	//scsi_debug("[scsi_genIOQueueTimerHandler] Leave", SCSI_DBL_0);

	return(NFX_OK);
}

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_getDeviceMaxTags
 *
 * Description: 	Returns the max number of queueable tags, as per the 
 *                  HAL driver.
 * 				
 * Created by:		Radek Aster
 *
 * Data Created:	August 29, 2001
 *					
 *------------------------------------------------------------------------------
 */
int
scsi_scsi_getDeviceMaxTags(scsi_dev_t *pdev)
{
    hal_dispatch_entry_t *		dte = hal_dt_entry_get();
	hal_adapter_info_t			info;
	int							rc;

	rc = (*dte->hal_ioctl)(HAL_IOCTL_COMMON_GET_ADAPTER_INFO, &info);

	return(rc == NFX_OK ? info.max_tags_per_device : rc);
}
