/*
 *------------------------------------------------------------------------------
 *
 * Name:	scsi-queue.c
 *
 * RCS:		$Id: scsi-queue.c,v 1.8 2002/11/13 00:55:45 nitinb Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	Implementation of SCSI descriptor queues.
 *
 * Created by:	Radek Aster
 *
 * Date Created: 8/22/2001
 *
 *------------------------------------------------------------------------------
 */

#include "scsi.h"
#include <exports/malloc-api.h>

int32 scsi_queue_private_timer_handler(int32, scsi_q_t *);

void
scsi_enqueue_simple_lilo(scsi_q_simple_t *sq, scsi_desc_t *sd)
{
	if (sq->q_head == NULL) {
		sq->q_head = sd;
		sd->next = sd->prev = sd;
	}
	else {
		sq->q_head->prev->next = sd;
		sd->prev = sq->q_head->prev;
		sd->next = sq->q_head;
		sq->q_head->prev = sd;
	}
	++sq->q_len;
	++sq->q_enqueues;
}

void 
scsi_enqueue_simple_sorted(scsi_q_simple_t *sq, scsi_desc_t *sd)
{
	scsi_desc_t *	p;

	if (sq->q_head == NULL) {
                sq->q_head = sd;
                sd->next = sd->prev = sd;
        }
	else {
		if (sd->lba < sq->q_head->lba) {
			sq->q_head->prev->next = sd;
			sd->prev = sq->q_head->prev;
			sd->next = sq->q_head;
			sq->q_head->prev = sd;
			sq->q_head = sd;
		}
		else {
#ifdef Q_USE_LOCALITY
			if (sq->last_insertion_point && sd->lba >= sq->last_insertion_point->lba)
				p = sq->last_insertion_point;
			else
				p = sq->q_head;
#else
			p = sq->q_head;
#endif
			/* Find insertion point */
			while (sd->lba >= p->next->lba && p->next != sq->q_head) {
				p = p->next;
				++sq->q_traversals;
			}
			/* Insert between p and p->next */
			sd->prev = p;
			sd->next = p->next;
			p->next->prev = sd;
			p->next = sd;
		}
#ifdef Q_USE_LOCALITY
		sq->last_insertion_point = sd;
#endif
	}
	++sq->q_len;
	++sq->q_enqueues;
}

scsi_desc_t * 
scsi_dequeue_simple(scsi_q_simple_t *sq)
{
	scsi_desc_t *	sd;

	if (sq->q_head == NULL)
		return(NULL);

	sd = sq->q_head;
	sq->q_head = sd->next;

	sd->prev->next = sd->next;
	sd->next->prev = sd->prev;
	sd->next = sd->prev = NULL;

	if (sq->q_head == sd)
		sq->q_head = NULL;

	--sq->q_len;
	++sq->q_dequeues;

#ifdef Q_USE_LOCALITY
	if (sq->last_insertion_point == sd)
		sq->last_insertion_point = sq->q_head;
#endif

	return(sd);
}

void 
scsi_enqueue(scsi_q_t *q, scsi_desc_t *sd)
{
	eee_rtc(&sd->queued_timestamp);

	if (q->q_type == Q_TYPE_SORTED) {
		if (q->sorted_q_b->q_head == NULL || sd->lba >= q->sorted_q_b->q_head->lba) {
			/* printf("Enqueuing onto SORTED Q B: sd 0x%x\n", sd); */
			scsi_enqueue_simple_sorted(q->sorted_q_b, sd);
		}
		else {
			/* printf("Enqueuing onto SORTED Q A: sd 0x%x\n", sd); */
			scsi_enqueue_simple_sorted(q->sorted_q_a, sd);
		}
	}
	else {
		/* printf("Enqueuing onto LILO Q: sd 0x%x\n", sd); */
		scsi_enqueue_simple_lilo(q->lilo_q, sd);
	}

	if (q->q_timer_duration)
		scsi_queue_timer_restart(q);

	/* scsi_queue_dump(q); */
}

scsi_desc_t * 
scsi_dequeue(scsi_q_t *q)
{
	scsi_desc_t *	sd;

	if (q->q_type == Q_TYPE_SORTED) {
		if (q->sorted_q_a->q_head == NULL && q->sorted_q_b->q_head == NULL)
			sd = NULL;
		else {
			sd = scsi_dequeue_simple(q->sorted_q_b);
			if (q->sorted_q_b->q_head == NULL) {
				scsi_q_simple_t *	tmp = q->sorted_q_b;
				q->sorted_q_b = q->sorted_q_a;
				q->sorted_q_a = tmp;
				/* printf("Q A <-> B swapski\n"); */
			}
		}
		/* printf("Dequeuing from SORTED Q: sd 0x%x\n", sd);  */
	}
	else {
		sd = scsi_dequeue_simple(q->lilo_q);
		/* printf("Dequeuing from LILO Q: sd 0x%x\n", sd); */
	}

	if (scsi_queue_len(q) == 0)
		scsi_queue_timer_stop(q);

	/* scsi_queue_dump(q); */
	return(sd);
}

scsi_desc_t *
scsi_queue_head(scsi_q_t *q)
{
	scsi_desc_t *   sd;

	if (q->q_type == Q_TYPE_SORTED) {
		if (q->sorted_q_a->q_head == NULL && q->sorted_q_b->q_head == NULL)
			sd = NULL;
		else
			sd = q->sorted_q_b->q_head;
	}
	else
		sd = q->lilo_q->q_head;
	return(sd);
}

int32
scsi_queue_len(scsi_q_t *q)
{
	return( ((q->lilo_q) ? q->lilo_q->q_len : 0) + 
		((q->sorted_q_a) ? q->sorted_q_a->q_len : 0) + 
		((q->sorted_q_b) ? q->sorted_q_b->q_len : 0) );
}

void
scsi_queue_timer_restart(scsi_q_t *q)
{
	int32	rc;

	if (q->q_timer_running) {
		rc = eee_timerStop(q->q_timer_handle);
		if (rc != NFX_OK) {
			printf("scsi_queue_timer_restart: eee_timerStop failed; handle 0x%x, rc 0x%x\n",
			       q->q_timer_handle, rc);
			VERIFY(rc == NFX_OK);
		}
	}
	rc = eee_timerStart(q->q_timer_handle, (void *)scsi_queue_private_timer_handler, q, NULL, NULL, NULL);
	if (rc != NFX_OK) {
		printf("scsi_queue_timer_restart: eee_timerStart failed; handle 0x%x, rc 0x%x\n",
		       q->q_timer_handle, rc);
		VERIFY(rc == NFX_OK);
	}
	q->q_timer_running = 1;
}

void
scsi_queue_timer_stop(scsi_q_t *q)
{
	int32	rc;

	if (q->q_timer_running) {
		rc = eee_timerStop(q->q_timer_handle);
		VERIFY(rc == NFX_OK);
	}
	q->q_timer_running = 0;
}

int32
scsi_queue_private_timer_handler(int32 timer_handle, scsi_q_t *q)
{
	VERIFY(q->q_timer_handle == timer_handle);
	VERIFY(q->q_timer_running);

	q->q_timer_running = 0;

	(*q->q_timer_callback)(q->q_cookie);
	return(NFX_OK);
}

scsi_q_t * 
scsi_queue_create(uint32 q_type, int32 timer_duration, int32 (*callback)(void *), void *cookie)
{
	scsi_q_t *	q;

	if (q_type != Q_TYPE_SORTED && q_type != Q_TYPE_LILO)
		return(NULL);

	q = (scsi_q_t *)eee_ramAlloc(sizeof(scsi_q_t));
	VERIFY(q);
	bzero(q, sizeof(scsi_q_t));

	q->q_type = q_type;
	q->q_timer_duration = timer_duration;
	q->q_timer_callback = callback;
	q->q_cookie = cookie;

	q->lilo_q = (scsi_q_simple_t *)eee_ramAlloc(sizeof(scsi_q_simple_t));
	VERIFY(q->lilo_q);
	bzero(q->lilo_q, sizeof(scsi_q_simple_t));

	if (q_type == Q_TYPE_SORTED) {
		q->sorted_q_a = (scsi_q_simple_t *)eee_ramAlloc(sizeof(scsi_q_simple_t));
		VERIFY(q->sorted_q_a);
		bzero(q->sorted_q_a, sizeof(scsi_q_simple_t));
		q->sorted_q_b = (scsi_q_simple_t *)eee_ramAlloc(sizeof(scsi_q_simple_t));
		VERIFY(q->sorted_q_b);
		bzero(q->sorted_q_b, sizeof(scsi_q_simple_t));
	}
	
	if (q->q_timer_duration > 0) {
		q->q_timer_handle = eee_timerCreate(q->q_timer_duration, EEE_TIMER_ONCE);
		VERIFY(q->q_timer_handle >= 0);
	}
	else
		q->q_timer_handle = -1;
						    
	return(q);
}

void 
scsi_queue_destroy(scsi_q_t *q)
{
	int32	rc;

	if( !q )
		return;

	if (q->q_timer_duration) {
		if (q->q_timer_running) {
			rc = eee_timerStop(q->q_timer_handle);
			VERIFY(rc == NFX_OK);
		}
		rc = eee_timerDestroy(q->q_timer_handle);
		VERIFY(rc == NFX_OK);
	}

	if (q->lilo_q)
		eee_ramDealloc(q->lilo_q);
	if (q->sorted_q_a)
		eee_ramDealloc(q->sorted_q_a);
	if (q->sorted_q_b)
		eee_ramDealloc(q->sorted_q_b);
	eee_ramDealloc(q);
}

void
scsi_queue_dump(scsi_q_t *q)
{
	void scsi_queue_simple_dump(scsi_q_simple_t *sq, char *prefix)
		{
			int		i;
			scsi_desc_t *	sd;

			printf("%s len %d, traversals %d, enqueues %d, dequeues %d", 
			       prefix, 
			       sq->q_len, sq->q_traversals, sq->q_enqueues, sq->q_dequeues);

#ifdef Q_USE_LOCALITY
			printf(", l_i_p 0x%p", sq->last_insertion_point);
#endif
			i = 0; 
			if (NULL != sq->q_head) {
				sd = sq->q_head;
				do {
					if ((i++ % 4) == 0)
						printf("\n%s ", prefix);
					printf("0x%p/%d/%d ", sd, sd->lba,
                           (uint32)eee_rtcToUsec(&sd->queued_timestamp) / USEC_PER_TICK);
					sd = sd->next;
				} while (sd != sq->q_head);
			}
			printf("\n");
		}

	char *scsi_queue_type(scsi_q_t *q)
		{
			static char	type[16];

			switch(q->q_type) {
			case Q_TYPE_SORTED:
				sprintf(type, "SORTED"); break;
			case Q_TYPE_LILO:
				sprintf(type, "LILO"); break;
			default:
				VERIFY(0);
			}
			return(type);
		}

	printf("SCSI Queue 0x%p: type %s\n", q, scsi_queue_type(q));
	if (q->lilo_q)
		scsi_queue_simple_dump(q->lilo_q,	"       LILO:");
	if (q->sorted_q_a)
		scsi_queue_simple_dump(q->sorted_q_a,	"   SORTED_A:");
	if (q->sorted_q_b)
		scsi_queue_simple_dump(q->sorted_q_b,  "   SORTED_B:");
}

