/*
 *------------------------------------------------------------------------------
 *
 * Name:	scsi-test.c
 *
 * RCS:		$Id: scsi-test.c,v 1.22 2002/08/31 03:14:48 raster Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	Implementation of SCSI level test harness.
 *
 * Created by:	Radek Aster
 *
 * Date Created: 8/23/2001
 *
 *------------------------------------------------------------------------------
 */

#include "scsi.h"
#include "../sm-eee/eee.h"		/* XXX */
#include <exports/malloc-api.h>

#ifdef SCSI_TEST

#define DEVICE_TEST_TYPE	SCSI_DEVTYPE_DISK
/* #define DEVICE_TEST_TYPE	SCSI_DEVTYPE_TAPE */

#if defined(DEVICE_TEST_TYPE) && (DEVICE_TEST_TYPE == SCSI_DEVTYPE_DISK)
/* For DISKS */
#define DISK_IO_SIZE		(16 * 512)
#define MAX_DESC_PER_DEV	32
/* #define SEQUENTIAL */
#define READ_ONLY
/* #define WRITE_ONLY */
/* #define NO_REORDER_WRITES */
#define IMMEDIATE_RESTART
#endif

#if defined(DEVICE_TEST_TYPE) && (DEVICE_TEST_TYPE == SCSI_DEVTYPE_TAPE)
/* For TAPES */
#define MINIMUM_TAPE_IO_SIZE	(1024 * 32)
#define MAXIMUM_TAPE_IO_SIZE	(1024 * 256)
#define MAX_DESC_PER_DEV	1
#define TAPE_IGNORE_MBS		128
#define TAPE_MINIMUM_TIME	60
#define TAPE_MINIMUM_MBS	1000
/* #define WRITE_ONLY */
#define IMMEDIATE_RESTART
#endif

typedef enum {
	NEW,
	OPEN_PENDING,
	REWIND_PENDING,
	OPEN,
	REMOVING
} dev_state_t;

typedef struct _cmd_desc_ {
	int			index;
	int			lba;
	int			lbc;
	int			read;
	struct test_dev *	dev;
	struct _cmd_desc_ *	next;
} cmd_desc_t;

typedef struct test_dev {
	scsi_id_t		dev_id;
	scsi_path_t		dev_path;
	uint32			scsi_type;
	uint32			capacity;
	uint32			sector_size;

	uint32			activity[MAX_DESC_PER_DEV];
	dev_state_t		state;
	cmd_desc_t *		cmds;
	uint32			start_timestamp;
	int			count;
	int			ignore_count;
	int			r_count;
	int			w_count;
	int			miscompares;
	struct test_dev *	next;

#ifdef NO_REORDER_WRITES
	uint32			write_order[MAX_DESC_PER_DEV];
	uint32			out_index;
	uint32			in_index;
#endif
} test_dev_t;

test_dev_t *dev_head = NULL;
int32	    elapsed_seconds = 0;

void        scsi_test_dispatch(test_dev_t *, cmd_desc_t *);
void        scsi_test_dispatch_disk(test_dev_t *, cmd_desc_t *);
void        scsi_test_dispatch_tape(test_dev_t *, cmd_desc_t *);
eee_desc_t *scsi_test_edesc_allocate(int32);
void        scsi_test_precharge(eee_desc_t *, uint32, uint32, uint32);
int         scsi_test_verify(eee_desc_t *, uint32, uint32, uint32);
void        scsi_test_notify_disk(scsi_msgHdr_t *);
void        scsi_test_notify_tape(scsi_msgHdr_t *);

void 
scsi_test_daemon(void *cookie)
{
	test_dev_t *	dev;
	test_dev_t *	prev_dev;
	cmd_desc_t *	cd;

	for (dev = dev_head; dev; dev = dev->next) {
		switch(dev->state) {
		case NEW: 
#if !SDM_DEV_INIT
			{
				scsi_openReqMsg_t *	req;
				eee_desc_t *		ed;
				
				ed = scsi_test_edesc_allocate(sizeof(scsi_openReqMsg_t));
				req = (scsi_openReqMsg_t *)ed->bd[1].buf;
				req->hdr.mid = SCSI_MID_OPEN_REQ;
				req->hdr.cHandle = (uint32)dev;
				req->hdr.path = dev->dev_path;

				dev->state = OPEN_PENDING;
				scsi_receiveMsg(ed);
			}
#else
#if (DEVICE_TEST_TYPE == SCSI_DEVTYPE_TAPE)
			{
				eee_desc_t *		ed;
				scsi_relayReqMsg_t *	req;

				printf("SCSI_TEST: Rewinding device 0x%llx\n", dev->dev_id);
				ed = scsi_test_edesc_allocate(sizeof(scsi_relayReqMsg_t));
				req = (scsi_relayReqMsg_t *)ed->bd[1].buf;
				bzero(req, sizeof(scsi_relayReqMsg_t));
				req->hdr.mid = SCSI_MID_RELAY_REQ;
				req->hdr.cHandle = (uint32)dev->cmds;
				req->hdr.path = dev->dev_path;
				req->cdb[0] = SCSI_CMD_REWIND_UNIT;
				dev->state = REWIND_PENDING;
				scsi_receiveMsg(ed);
			}
#else
			dev->state = OPEN;
#endif
#endif
			break;

		case OPEN_PENDING:
			break;

		case REWIND_PENDING:
			break;

		case OPEN:
			while ((cd = dev->cmds)) {
				dev->cmds = cd->next;
				cd->next = NULL;

				scsi_test_dispatch(dev, cd);
			}
			break;

		case REMOVING:
			while ((cd = dev->cmds)) {
				dev->cmds = cd->next;
				cd->next = NULL;
				eee_ramDealloc(cd);
				dev->count--;
			}
			
			if (dev->count == 0) {
				if (dev == dev_head) {
					dev_head = dev->next;
				}
				else {
					for (prev_dev = dev_head; prev_dev->next != dev; prev_dev = prev_dev->next)
						;
					prev_dev->next = dev->next;
				}

				eee_ramDealloc(dev);
			}
			break;
		}
	}
}

void 
scsi_test_timer(int32 timer_handle, void *arg)
{
#if defined(DEVICE_TEST_TYPE)
        int		w_count, r_count, miscompares;
	test_dev_t *	dev;
#if DEVICE_TEST_TYPE == SCSI_DEVTYPE_TAPE
	int		dev_count = 0;
#endif

	elapsed_seconds += SCSI_TEST_TIMER_PERIOD;

	r_count = w_count = miscompares = 0;	

#if DEVICE_TEST_TYPE == SCSI_DEVTYPE_TAPE
	dev_count = 0;
	for (dev = dev_head; dev; dev = dev->next) {
		uint32		seconds;
		uint64		mbs_r, mbs_w;

		if (dev->state == OPEN && elapsed_seconds - dev->start_timestamp) {
			seconds = elapsed_seconds - dev->start_timestamp;
			r_count += dev->r_count;
			w_count += dev->w_count;
			miscompares += dev->miscompares;

			mbs_r = dev->r_count * dev->sector_size / 1000 / 100;
			mbs_w = dev->w_count * dev->sector_size / 1000 / 100;
			printf("R=%d.%dMB(%d)@%d.%dMB/s:"
			       "W=%d.%dMB(%d)@%d.%dMB/s ",
			       (uint32)(mbs_r/10), (uint32)(mbs_r%10), 
			       dev->r_count,
			       (uint32)(mbs_r/seconds/10), (uint32)(mbs_r/seconds%10), 
			       (uint32)(mbs_w/10), (uint32)(mbs_w%10), 
			       dev->w_count,
			       (uint32)(mbs_w/seconds/10), (uint32)(mbs_w/seconds%10));
			++dev_count;
		}
	}
	if (dev_count)
		printf("\n");
#endif /* DEVICE_TEST_TYPE == SCSI_DEVTYPE_TAPE */
#if DEVICE_TEST_TYPE == SCSI_DEVTYPE_DISK
	for (dev = dev_head; dev; dev = dev->next) {
		if (dev->state == OPEN) {
			r_count += dev->r_count;
			w_count += dev->w_count;
			miscompares += dev->miscompares;
			printf("%d(%d IOPS)/%d(%d IOPS)/%d ", 
			       dev->r_count, dev->r_count/elapsed_seconds, 
			       dev->w_count, dev->w_count/elapsed_seconds, 
			       dev->miscompares);
		}
	}
	printf(": %d(%d IOPS)/%d(%d IOPS)/%d\n", 
	       r_count, r_count/elapsed_seconds,
	       w_count, w_count/elapsed_seconds,
	       miscompares);
#endif /* DEVICE_TEST_TYPE == SCSI_DEVTYPE_DISK */
#endif /* DEVICE_TEST_TYPE */
}

void
scsi_test_dispatch(test_dev_t *dev, cmd_desc_t *cd)
{
	switch(dev->scsi_type) {
#if DEVICE_TEST_TYPE == SCSI_DEVTYPE_DISK
	case SCSI_DEVTYPE_DISK:
		scsi_test_dispatch_disk(dev, cd);
		break;
#endif

#if DEVICE_TEST_TYPE == SCSI_DEVTYPE_TAPE
	case SCSI_DEVTYPE_TAPE:
		scsi_test_dispatch_tape(dev, cd);
		break;
#endif

	default:
		VERIFY(0);
	}
}

#if DEVICE_TEST_TYPE == SCSI_DEVTYPE_DISK
void
scsi_test_dispatch_disk(test_dev_t *dev, cmd_desc_t *cd)
{
	eee_desc_t *	ed;

	if (cd->read) {
		scsi_readReqMsg_t *	req; 

		/* Fire off a read */
		ed = scsi_test_edesc_allocate(sizeof(scsi_readReqMsg_t));
		req = (scsi_readReqMsg_t *)ed->bd[1].buf;
		bzero(req, sizeof(scsi_readReqMsg_t));
		req->hdr.mid = SCSI_MID_READ_REQ;
		req->hdr.cHandle = (uint32)cd;
		req->hdr.path = dev->dev_path;

		req->blkOffset = cd->lba;
		req->blkCnt = cd->lbc;
		req->dataDest = 0;
#if 0
		printf("SCSI_TEST: Reading:D: path 0x%x, index %d, LBA %d, LBC %d\n",
		       dev->dev_path, cd->index, cd->lba, cd->lbc);
#endif
	}
	else {
		scsi_writeReqMsg_t *	req; 
		
 		/* Fire off a write */
		ed = scsi_test_edesc_allocate(sizeof(scsi_writeReqMsg_t) + DISK_IO_SIZE);
		req = (scsi_writeReqMsg_t *)ed->bd[1].buf;
		bzero(req, sizeof(scsi_writeReqMsg_t));
		req->hdr.mid = SCSI_MID_WRITE_REQ;
		req->hdr.cHandle = (uint32)cd;
		req->hdr.path = dev->dev_path;
#ifdef NO_REORDER_WRITES
		req->attributes = SCSI_MSG_ATTR_NO_REORDER;
#endif
		req->blkOffset = cd->lba;
		req->blkCnt = cd->lbc;

#ifndef WRITE_ONLY
		/* Precharge write buffer */
		scsi_test_precharge(ed, cd->index, cd->lba, sizeof(scsi_writeReqMsg_t));
#endif

#if 0
		printf("SCSI_TEST: Writing:D: path 0x%x, index %d, LBA %d, LBC %d\n",
		       dev->dev_path, cd->index, cd->lba, cd->lbc);
#endif

#ifdef NO_REORDER_WRITES
		VERIFY(dev->write_order[dev->out_index] == -1);
		dev->write_order[dev->out_index] = req->blkOffset;
		dev->out_index++;
		if (dev->out_index == MAX_DESC_PER_DEV)
			dev->out_index = 0;
#endif
	}
	scsi_receiveMsg(ed);
}
#endif

#if DEVICE_TEST_TYPE == SCSI_DEVTYPE_TAPE
void
scsi_test_dispatch_tape(test_dev_t *dev, cmd_desc_t *cd)
{
	eee_desc_t *	ed;

	if (cd->read) {
		scsi_readReqMsg_t *	req; 

		/* Fire off a read */
		ed = scsi_test_edesc_allocate(sizeof(scsi_readReqMsg_t));
		req = (scsi_readReqMsg_t *)ed->bd[1].buf;
		bzero(req, sizeof(scsi_readReqMsg_t));
		req->hdr.mid = SCSI_MID_READ_REQ;
		req->hdr.cHandle = (uint32)cd;
		req->hdr.path = dev->dev_path;
		req->blkCnt = dev->sector_size;
#if 0
		printf("SCSI_TEST: Reading:T: path 0x%x\n", dev->dev_path);
#endif
	}
	else {
		scsi_writeReqMsg_t *	req; 
		
		/* Fire off a write */
		ed = scsi_test_edesc_allocate(sizeof(scsi_writeReqMsg_t) + dev->sector_size);
		req = (scsi_writeReqMsg_t *)ed->bd[1].buf;
		bzero(req, sizeof(scsi_writeReqMsg_t));
		req->hdr.mid = SCSI_MID_WRITE_REQ;
		req->hdr.cHandle = (uint32)cd;
		req->hdr.path = dev->dev_path;
		req->blkCnt = dev->sector_size;

		/* Precharge write buffer */
		/* scsi_test_precharge(ed, cd->index, cd->lba, sizeof(scsi_writeReqMsg_t)); */
#if 0
		printf("SCSI_TEST: Writing:T: path 0x%x\n", dev->dev_path);
#endif
	}
	scsi_receiveMsg(ed);
}
#endif

void 
scsi_test_remove_device_notify(scsi_id_t id, scsi_path_t path)
{
	test_dev_t *	dev;

#if 1
	printf("SCSI_TEST: Remove device: id 0x%llx, path 0x%x\n", id, path);
#endif

	for (dev = dev_head; dev; dev = dev->next) {
		if (dev->dev_id == id) {
			dev->state = REMOVING;
		}
	}
}

void 
scsi_test_new_device_notify(scsi_id_t id, scsi_path_t path, uchar8 scsi_type, uint32 capacity, uint32 sector_size)
{
	test_dev_t *	dev;
	int		i;
	cmd_desc_t *	cd;
	static int tapes_found = 0;

#if !defined(DEVICE_TEST_TYPE)
	printf("SCSI_TEST: DEVICE_TEST_TYPE not defined ... ignoring device\n");
	return;
#endif

	if (scsi_type != DEVICE_TEST_TYPE ||
	    id & 0x0000FFFFFFFFFFFFLL != 0x00000080e5127100LL) {	/* Xyratex 2GB */
		/* 0x000000e00201f5c6LL DLT8000 in rack 5 */
		/* 0x000500e0020204f7LL SDLT in L40 */
		/* 0x000200e0020204f7LL LTO in L40 */
		/* printf("SCSI_TEST: Ignore device of type %d\n", scsi_type); */
		return;
	}

	if (scsi_type == SCSI_DEVTYPE_TAPE && tapes_found++) {
		printf("Tapes Found %d\n", tapes_found);
		return;
	}

#if 1
	printf("SCSI_TEST: New device: id 0x%llx, path 0x%x, type %d, capacity %d, sector_size %d\n", 
	id, path, scsi_type, capacity, sector_size);
#endif

	dev = (test_dev_t *)eee_ramAlloc(sizeof(test_dev_t));
	VERIFY(dev);
	bzero(dev, sizeof(test_dev_t));
	dev->next = dev_head;
	dev_head = dev;

	dev->state = NEW;

	dev->dev_id = id;
	dev->dev_path = path;
	dev->scsi_type = scsi_type;


	switch (scsi_type) {
#if DEVICE_TEST_TYPE == SCSI_DEVTYPE_DISK
	case SCSI_DEVTYPE_DISK:
		dev->capacity = capacity;
		dev->sector_size = sector_size;

		/* Allocate command descriptors */
		for (i = 0; i < MAX_DESC_PER_DEV; ++i) {
			cd = (cmd_desc_t *)eee_ramAlloc(sizeof(cmd_desc_t));
			VERIFY(cd);
			bzero(cd, sizeof(cmd_desc_t));
			
			cd->next = dev->cmds;
			dev->cmds = cd;
			dev->count++;

			cd->index = MAX_DESC_PER_DEV - 1 - i + (nfxMyCPUId * MAX_DESC_PER_DEV);
			cd->dev = dev;
			cd->lbc = DISK_IO_SIZE / dev->sector_size;
#ifdef READ_ONLY
			cd->read = 1;
#else
			cd->read = 0;
#endif

#if defined(SEQUENTIAL)
			cd->lba = cd->index * cd->lbc; 
			if (nfxMyCPUId > 0) {
				printf("Cannot run on nfxMyCPUId == %d with SEQUENTIAL\n", nfxMyCPUId);
				VERIFY(0);
			}
#else
			cd->lba = cd->index * (dev->capacity / MAX_DESC_PER_DEV / NFX_FCNIM_NUM_CPUS);
#endif
#ifdef NO_REORDER_WRITES
			dev->write_order[i] = -1;
#endif
		}
		break;
#endif

#if DEVICE_TEST_TYPE == SCSI_DEVTYPE_TAPE
	case SCSI_DEVTYPE_TAPE:
		dev->sector_size = MINIMUM_TAPE_IO_SIZE;

		/* Allocate command descriptors */
		cd = (cmd_desc_t *)eee_ramAlloc(sizeof(cmd_desc_t));
		VERIFY(cd);
		bzero(cd, sizeof(cmd_desc_t));
			
		dev->cmds = cd;
		dev->count++;

		cd->index = 0;

		cd->dev = dev;
		cd->read = 0;

		break;
#endif
	default:
		/* Ignore all other device types */
		;
	}
}

void 
scsi_test_notify(scsi_msgHdr_t *msg)
{
	test_dev_t *	dev;
	cmd_desc_t *	cd;

	switch(msg->mid) {
	case SCSI_MID_OPEN_RSP:		
		dev = (test_dev_t *)msg->cHandle;
		break;

	case SCSI_MID_RELAY_RSP:
		cd = (cmd_desc_t *)msg->cHandle;
		dev = cd->dev;
		break;

	case SCSI_MID_READ_RSP:
	case SCSI_MID_WRITE_RSP:
		dev = ((cmd_desc_t *)msg->cHandle)->dev;
		break;

	default:
		VERIFY(0);
	}

	switch(dev->scsi_type) {
#if DEVICE_TEST_TYPE == SCSI_DEVTYPE_DISK
	case SCSI_DEVTYPE_DISK:
		scsi_test_notify_disk(msg);
		break;
#endif

#if DEVICE_TEST_TYPE == SCSI_DEVTYPE_TAPE
	case SCSI_DEVTYPE_TAPE:		
		scsi_test_notify_tape(msg);
		break;
#endif

	default:
		VERIFY(0);
	}
}

#if DEVICE_TEST_TYPE == SCSI_DEVTYPE_DISK
void 
scsi_test_notify_disk(scsi_msgHdr_t *msg)
{
	test_dev_t *	dev;
	cmd_desc_t *	cd;

	switch(msg->mid) {
	case SCSI_MID_OPEN_RSP:
		{
			scsi_openRspMsg_t *	o_msg = (scsi_openRspMsg_t *)msg;

			dev = (test_dev_t *)msg->cHandle;
			VERIFY(o_msg->eCode == NFX_OK);
			VERIFY(dev->state == OPEN_PENDING);
			dev->state = OPEN;
		}
		break;

	case SCSI_MID_READ_RSP:
		{
			scsi_readRspMsg_t *	r_msg = (scsi_readRspMsg_t *)msg;

			cd = (cmd_desc_t *)msg->cHandle;
			dev = cd->dev;
			VERIFY(dev->state == OPEN || dev->state == REMOVING);

			if (r_msg->eCode != NFX_OK)
				printf("SCSI_TEST: READ_RSP: eCode = 0x%x: lba %d, lbc %d\n", 
				       r_msg->eCode, cd->lba, cd->lbc);

#if !defined(READ_ONLY)
			if (r_msg->eCode == NFX_OK) {
				/* Verify read contents */
				if (scsi_test_verify((eee_desc_t *)((uint32)(r_msg->codHandle)), 
						     cd->index, cd->lba, sizeof(scsi_readRspMsg_t)))
					++dev->miscompares;
			}
#endif
			++dev->r_count;
			++dev->activity[cd->index - (nfxMyCPUId * MAX_DESC_PER_DEV)];

			/* Bump LBA and setup for next write */
#ifdef SEQUENTIAL
#if 1
			cd->lba += cd->lbc * MAX_DESC_PER_DEV;
#endif
			if (cd->lba + cd->lbc > dev->capacity)
				cd->lba = cd->index * cd->lbc;
#else
			cd->lba += cd->lbc;
			if (cd->lba > ((cd->index + 1) * (dev->capacity / MAX_DESC_PER_DEV / NFX_FCNIM_NUM_CPUS)) - cd->lbc)
				cd->lba = cd->index * (dev->capacity / MAX_DESC_PER_DEV / NFX_FCNIM_NUM_CPUS);
#endif

#ifdef READ_ONLY
			cd->read = 1;
#else
			cd->read = 0;
#endif

#ifndef IMMEDIATE_RESTART
			cd->next = dev->cmds;
			dev->cmds = cd;
#else
			if (dev->state == OPEN)
				scsi_test_dispatch(dev, cd);
			else {
				cd->next = dev->cmds;
				dev->cmds = cd;
			}				
#endif
		}
		break;

	case SCSI_MID_WRITE_RSP:
		{
			scsi_writeRspMsg_t *	w_msg = (scsi_writeRspMsg_t *)msg;
			
			cd = (cmd_desc_t *)msg->cHandle;
			dev = cd->dev;
			VERIFY(dev->state == OPEN || dev->state == REMOVING);

			if (w_msg->eCode != NFX_OK)
				printf("SCSI_TEST: WRITE_RSP: eCode = 0x%x: lba %d, lbc %d\n", 
				       w_msg->eCode, cd->lba, cd->lbc);

#ifdef NO_REORDER_WRITES
			if (dev->write_order[dev->in_index] == -1) {
				printf("FATAL: in_index %d, expected %d, got %d\n",
				       dev->in_index, 
				       dev->write_order[dev->in_index],
                                       w_msg->blkOffset);
				VERIFY(0);
			}
			if (dev->write_order[dev->in_index] != w_msg->blkOffset) {
				static oto_count = 0;
				printf("XXX OTO[%d] write[%d]: expected %d, got %d\n", 
				       oto_count, dev->in_index, dev->write_order[dev->in_index],
				       w_msg->blkOffset);
				if (++oto_count > 1000) {
					ispfc_trace_dump(64);
					VERIFY(0);
				}
			}
			dev->write_order[dev->in_index] = -1;
			dev->in_index++;
			if (dev->in_index == MAX_DESC_PER_DEV)
				dev->in_index = 0;			
#endif /* NO_REORDER_WRITES */

			++dev->w_count;
			++dev->activity[cd->index - (nfxMyCPUId * MAX_DESC_PER_DEV)];

#ifndef WRITE_ONLY
			/* Setup for next read */
			cd->read = 1;
#else
#ifdef SEQUENTIAL
#if 1
			cd->lba += cd->lbc * MAX_DESC_PER_DEV;
#endif
			if (cd->lba + cd->lbc > dev->capacity)
				cd->lba = cd->index * cd->lbc;
#else
			cd->lba += cd->lbc;
			if (cd->lba > ((cd->index + 1) * (dev->capacity / MAX_DESC_PER_DEV / NFX_FCNIM_NUM_CPUS)) - cd->lbc)
				cd->lba = cd->index * (dev->capacity / MAX_DESC_PER_DEV / NFX_FCNIM_NUM_CPUS);
#endif
			
#endif

#ifndef IMMEDIATE_RESTART
			cd->next = dev->cmds;
			dev->cmds = cd;
#else
			if (dev->state == OPEN)
				scsi_test_dispatch(dev, cd);
			else {
				cd->next = dev->cmds;
				dev->cmds = cd;
			}				
#endif
		}
		break;

	default:
		VERIFY(0);
	}
}
#endif

#if DEVICE_TEST_TYPE == SCSI_DEVTYPE_TAPE
void 
scsi_test_notify_tape(scsi_msgHdr_t *msg)
{
	test_dev_t *	dev;
	cmd_desc_t *	cd;
	uint64		mbs_written;
	uint64		mbs_read;

	switch(msg->mid) {
	case SCSI_MID_OPEN_RSP:
		{
			scsi_openRspMsg_t *	o_msg = (scsi_openRspMsg_t *)msg;
			eee_desc_t *		ed;
			scsi_relayReqMsg_t *	req;

			dev = (test_dev_t *)msg->cHandle;
			VERIFY(o_msg->eCode == NFX_OK);
			VERIFY(dev->state == OPEN_PENDING);
			
			printf("SCSI_TEST: Rewinding device 0x%llx\n", dev->dev_id);
			ed = scsi_test_edesc_allocate(sizeof(scsi_relayReqMsg_t));
			req = (scsi_relayReqMsg_t *)ed->bd[1].buf;
			bzero(req, sizeof(scsi_relayReqMsg_t));
			req->hdr.mid = SCSI_MID_RELAY_REQ;
			req->hdr.cHandle = (uint32)dev->cmds;
			req->hdr.path = dev->dev_path;
			req->cdb[0] = SCSI_CMD_REWIND_UNIT;
			dev->state = REWIND_PENDING;
			scsi_receiveMsg(ed);
		}
		break;

	case SCSI_MID_RELAY_RSP:
		{
			scsi_relayRspMsg_t *	r_msg = (scsi_relayRspMsg_t *)msg;

			VERIFY(r_msg->cdb[0] == SCSI_CMD_REWIND_UNIT);
			cd = (cmd_desc_t *)msg->cHandle;
			dev = cd->dev;
			VERIFY(dev->scsi_type == SCSI_DEVTYPE_TAPE);
			if (r_msg->status != 0) {
				printf("SCSI_TEST: Rewind status for device 0x%llx is 0x%x\n", dev->dev_id, r_msg->status);
			}
                        VERIFY(r_msg->status == 0);
                        VERIFY(dev->state == REWIND_PENDING);
			dev->state = OPEN;

			/* 
			 * For timing purposes, ignore the first
			 * "write-buffer-full" worth of writes ... that's how
			 * long it takes for the tape to get
			 * moving. Assuming write buffer is 128MB.
			 */
			dev->ignore_count = TAPE_IGNORE_MBS*1024*1024/dev->sector_size;

			printf("SCSI_TEST: Rewind done on device 0x%llx: %s with blocksize %d\n", 
			       dev->dev_id, cd->read ? "Reading" : "Writing", dev->sector_size);
		}
		break;

	case SCSI_MID_READ_RSP:
		{
			scsi_readRspMsg_t *	r_msg = (scsi_readRspMsg_t *)msg;
			eee_desc_t *		ed;
			scsi_relayReqMsg_t *	req;

			if (SCSI_SEQUENCE_ID(r_msg->transSeqId) < SCSI_SEQUENCE_MAX(r_msg->transSeqId))
				break;

			cd = (cmd_desc_t *)msg->cHandle;
			dev = cd->dev;
			VERIFY(dev->state == OPEN || dev->state == REMOVING);

			if (r_msg->eCode != NFX_OK)
				printf("SCSI_TEST: READ_RSP: eCode = 0x%x\n", r_msg->eCode);

			if (dev->ignore_count >= 0) {
				dev->ignore_count--;
				dev->r_count = 0;
				dev->w_count = 0;
				dev->start_timestamp = elapsed_seconds;
			}
			else {
				++dev->r_count;
				++dev->activity[cd->index];
			}

			mbs_read = dev->r_count * dev->sector_size / 1000 / 1000;
			if (mbs_read > TAPE_MINIMUM_MBS && (elapsed_seconds - dev->start_timestamp) > TAPE_MINIMUM_TIME) {
				cd->next = dev->cmds;
				dev->cmds = cd;

				printf("SCSI_TEST: Rewinding device 0x%llx\n", dev->dev_id);
				ed = scsi_test_edesc_allocate(sizeof(scsi_relayReqMsg_t));
				req = (scsi_relayReqMsg_t *)ed->bd[1].buf;
				bzero(req, sizeof(scsi_relayReqMsg_t));
				req->hdr.mid = SCSI_MID_RELAY_REQ;
				req->hdr.cHandle = (uint32)cd;
				req->hdr.path = dev->dev_path;
				req->cdb[0] = SCSI_CMD_REWIND_UNIT;
				dev->state = REWIND_PENDING;
				scsi_receiveMsg(ed);

				if (dev->sector_size >= MAXIMUM_TAPE_IO_SIZE) {
					dev->sector_size = MINIMUM_TAPE_IO_SIZE;
					cd->read = 0;
				}
				else {
					dev->sector_size *= 2;
					cd->read = 0;
				}
			}
			else {
#ifndef IMMEDIATE_RESTART
				cd->next = dev->cmds;
				dev->cmds = cd;
#else
				if (dev->state == OPEN)
					scsi_test_dispatch(dev, cd);
				else {
					cd->next = dev->cmds;
					dev->cmds = cd;
				} 
#endif
			}
		}
		break;

	case SCSI_MID_WRITE_RSP:
		{
			scsi_writeRspMsg_t *	w_msg = (scsi_writeRspMsg_t *)msg;
			eee_desc_t *		ed;
			scsi_relayReqMsg_t *	req;

			cd = (cmd_desc_t *)msg->cHandle;
			dev = cd->dev;
			VERIFY(dev->state == OPEN || dev->state == REMOVING);

			if (w_msg->eCode != NFX_OK)
				printf("SCSI_TEST: WRITE_RSP: eCode = 0x%x\n", w_msg->eCode);

			if (dev->ignore_count >= 0) {
				dev->ignore_count--;
				dev->r_count = 0;
				dev->w_count = 0;
				dev->start_timestamp = elapsed_seconds;
			}
			else {
				++dev->w_count;
				++dev->activity[cd->index];
			}

			mbs_written = dev->w_count * dev->sector_size / 1000 / 1000;
			if (mbs_written > TAPE_MINIMUM_MBS && (elapsed_seconds - dev->start_timestamp) > TAPE_MINIMUM_TIME) {
				cd->next = dev->cmds;
				dev->cmds = cd;

				printf("SCSI_TEST: Rewinding device 0x%llx\n", dev->dev_id);
				ed = scsi_test_edesc_allocate(sizeof(scsi_relayReqMsg_t));
				req = (scsi_relayReqMsg_t *)ed->bd[1].buf;
				bzero(req, sizeof(scsi_relayReqMsg_t));
				req->hdr.mid = SCSI_MID_RELAY_REQ;
				req->hdr.cHandle = (uint32)cd;
				req->hdr.path = dev->dev_path;
				req->cdb[0] = SCSI_CMD_REWIND_UNIT;
				dev->state = REWIND_PENDING;
				scsi_receiveMsg(ed);

#ifdef WRITE_ONLY
				if (dev->sector_size >= MAXIMUM_TAPE_IO_SIZE)
					dev->sector_size = MINIMUM_TAPE_IO_SIZE;
				else 
					dev->sector_size *= 2;
#else
				cd->read = 1;
#endif
			}
			else {
#ifndef IMMEDIATE_RESTART
				cd->next = dev->cmds;
				dev->cmds = cd;
#else
				if (dev->state == OPEN)
					scsi_test_dispatch(dev, cd);
				else {
					cd->next = dev->cmds;
					dev->cmds = cd;
				} 
#endif
			}
		}
		break;

	default:
		VERIFY(0);
	}
}
#endif

eee_desc_t *  
scsi_test_edesc_allocate(int32 byte_count)
{
	eee_desc_t *	ed;
	eee_desc_t *	first_ed = NULL;
	eee_descBuf_t * bd;
	int		requested_count;
	int		alloced_count = 0;
	int		buf_count;
	int		buf_byte_count;
	int		packet_len = 8192 + 128;
	
	while (alloced_count < byte_count) {
		if (first_ed == NULL) {
			first_ed = ed = eee_allocateDesc(EEE_SHARED_POOL, 0);
		}
		else {
			VERIFY(!(ed->bd[EEE_MAX_BUFS-1].control & EDESC_BD_CMD_IN_USE));
			ed->bd[EEE_MAX_BUFS-1].buf = (uchar8 *)eee_allocateDesc(EEE_SHARED_POOL, 0);
			ed->hdr.control |= EDESC_COMMAND_CHAIN_LAST_BUFFER;
			ed = (eee_desc_t *)ed->bd[EEE_MAX_BUFS-1].buf;
		}
		VERIFY(ed);
		bzero(ed, sizeof(eee_desc_t));

		requested_count = MIN(byte_count - alloced_count, packet_len);

		buf_byte_count = 0;
		buf_count = 0;
		bd = &ed->bd[1];
		while (requested_count > 0) {
			bd->control = EDESC_BD_CMD_IN_USE;
			bd->len = MIN(requested_count, packet_len);
			bd->len = MIN(bd->len, EEE_BUFSIZE_ULTRA);
			bd->control |= EDESC_BD_ULTRA;
			bd->buf = eee_allocateBuffer(EEE_SHARED_POOL, bd->control & EDESC_BD_BUFSIZE_MASK, 0);
			VERIFY(bd->buf);
			requested_count -= bd->len;
			alloced_count += bd->len;
			buf_byte_count += bd->len;
			++bd;
			++buf_count;
		}
		ed->hdr.control = ((buf_count << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK);
		ed->hdr.control |= (buf_byte_count & EDESC_TOTAL_LEN);
	}
	VERIFY(alloced_count >= byte_count);
	return(first_ed);
}

void
scsi_test_precharge(eee_desc_t *ed_head, uint32 tag0, uint32 tag1, uint32 buf1_offset)
{
	int		i, j;
	int		byte_count = 0;
	eee_desc_t *	ed;
	eee_descBuf_t *	bd;
	uint32 *	buf;
	uint32		bufcnt;
	int32		buf1_index = 1;
	int32		did_header = 0;
	uint32		prefix = ((tag0 & 0xFF) << 24) | ((tag1 & 0xFF) << 16);

	j = 0;

	ed = ed_head;
	while (ed) {
		bd = &ed->bd[0];
		if (!(bd->control & EDESC_BD_CMD_IN_USE)) {
			++bd;
			buf1_index = 0;
		}
		bufcnt = (ed->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;
		for (i = 0; i < bufcnt; ++i, ++bd) {
			if (bd->control & EDESC_BD_CMD_IN_USE) {
				buf = (uint32 *)((char *)bd->buf + ((ed == ed_head && i == buf1_index) ? buf1_offset : 0));
				/* Precharge the whole buffer with the pattern */
				if (!did_header) {
					*buf++ = tag0;
					*buf++ = tag1;
					*buf++ = tag0;
					*buf++ = tag1;
					did_header = 1;
					j = 16;
				}
				else
					j = 0;
			for (; j < bd->len - ((ed == ed_head && i == buf1_index) ? buf1_offset : 0); j += 4)
				*buf++ = prefix + byte_count + j;
			byte_count += bd->len - ((ed == ed_head && i == buf1_index) ? buf1_offset : 0);
			}
		} /* for */
		if (ed->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER)
			ed = (eee_desc_t *)ed->bd[EEE_MAX_BUFS-1].buf;
		else
			ed = NULL;
	} /* while */
}

int
scsi_test_verify(eee_desc_t *ed, uint32 tag0, uint32 tag1, uint32 buf1_offset)
{
	int		i, j, k;
	int		byte_count = 0;
	eee_descBuf_t *	bd;
	uint32 *	buf;
	uint32		bufcnt = (ed->hdr.control & EDESC_BUFCNT_MASK) >> EDESC_BUFCNT_SHIFT;
	int32           buf1_index = 1;
	int32		did_header = 0;
	int		first_miscompare = 1;
	uint32		prefix = ((tag0 & 0xFF) << 24) | ((tag1 & 0xFF) << 16);	
	int		rc = 0;

	void print_miscompare(eee_desc_t *ed, int bd_index, char *buf, int offset, uint32 expected, uint32 actual)
		{
			if (first_miscompare) {
				eee_desc_t *ed_chain = ed;
				while (ed_chain) {
					printf("      eDesc (0x%x): hdr.control 0x%x\n", ed_chain, ed_chain->hdr.control);
					for (k = 0; k < EEE_MAX_BUFS; ++k) 
						printf("             bd[%2d] control 0x%x, buf 0x%x, len %d\n",
						       k, ed_chain->bd[k].control, ed_chain->bd[k].buf, ed_chain->bd[k].len);
					if (ed_chain->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER)
						ed_chain = (eee_desc_t *)ed_chain->bd[EEE_MAX_BUFS-1].buf;
					else
						ed_chain = NULL;
				} /* while (ed_chain) */
				first_miscompare = 0;
				printf("Data miscompare: ed 0x%x, bd_index %d", ed, bd_index);
				++rc;
			}
			printf("\t 0x%x[offset 0x%04x] (0x%08x) != (0x%08x)\n",
			       buf, offset, actual, expected);
			/* VERIFY(0); */
		}

	bd = &ed->bd[0];
	if (!(bd->control & EDESC_BD_CMD_IN_USE)) {
		++bd;
		buf1_index = 0;
        }

	for (i = 0; i < bufcnt; ++i, ++bd) {
		if (bd->control & EDESC_BD_CMD_IN_USE) {
			buf = (uint32 *)((char *)bd->buf + ((i == buf1_index) ? buf1_offset : 0));
			if (!did_header) {
				if (*buf != tag0) print_miscompare(ed, i, (char *)buf, 0,  tag0, *buf); buf++;
				if (*buf != tag1) print_miscompare(ed, i, (char *)buf, 4,  tag1, *buf); buf++;
				if (*buf != tag0) print_miscompare(ed, i, (char *)buf, 8,  tag0, *buf); buf++;
				if (*buf != tag1) print_miscompare(ed, i, (char *)buf, 12, tag1, *buf); buf++;
				did_header = 1;
				j = 16;
			}
			else 
				j = 0;
			for (; j < bd->len - ((i == buf1_index) ? buf1_offset : 0); j += 4) {
				if (*buf != prefix + byte_count + j)
					print_miscompare(ed, i, (char *)buf, j, prefix + byte_count + j, *buf); 
				buf++;
			}
			byte_count += bd->len - ((i == buf1_index) ? buf1_offset : 0);
		}
	}
	return(rc);
}

#endif /* SCSI_TEST */
