/*  FileName: scsi.c
 *
 *  Description: Contains routines that provide the SCSI API to upper layer 
 *				 modules. As well as the SCSI statemachine code.
 *
 *	Created:	20 February	2001
 *
 *  Author:		Frank Salzmann
 *
 *  Copyright (c)  Agile Storage, Inc. 2001
 *
 *  RCS: $Id: scsi.c,v 1.54 2002/11/19 20:38:29 nitinb Exp $
 */			 

#include "scsi.h"
#include <exports/eee-desc.h>

int scsi_gatherStats();

uint32         scsiHistoryCurEntry = 0;
scsi_history_t scsiHistory[SCSI_MAX_HISTORY];

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_initModule
 *
 * Description: 	Initializes the SCSI layer module.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	March 6, 2001
 * 					
 * Returns: 
 *
 *------------------------------------------------------------------------------
 */
int32
scsi_initModule()
{
	char			*ptr;
	int32			 rc;
	uint32			 i, size;
	scsi_desc_t		*sd;
	scsi_mp_desc_t	*smpd;

	/*
	 * Initialize SCSI error and debug levels
	 */
	for(i=0; i < SCSI_MAX_ERR_LEVEL; i++)
		scsiErrLevelCtrl[i] = SCSI_ERR_LEVEL_ENABLED;

	for(i=0; i < SCSI_MAX_DEBUG_LEVEL; i++)
		scsiDebugLevelCtrl[i] = 0;

	/*
	 * Allocate the SCSI module control block. 
	 */
	size = sizeof(scsi_ctrl_t);
	ptr = (char *)eee_ramAlloc(size);
	if( !ptr )
	{
		scsi_err( SCSI_ERR_MALLOC | SCSI_ERR_LEVEL_0, 
				  "SCSI module control block");
		return SCSI_ERR_MALLOC;
	}
	scsiCtrl = (scsi_ctrl_t *)ptr;

	/*
	 * Initialize SCSI device and control tables  
	 */
	memset( scsiCtrl, 0, sizeof(scsi_ctrl_t));
	memset( scsiDevList, 0, sizeof(scsiDevList));
	memset( scsiHistory, 0, sizeof(scsiHistory));

	/*
	 *--------------------------------------
	 * Allocate memory for  SCSI descriptors
	 *--------------------------------------
	 */
	size = sizeof(scsi_desc_t) * SCSI_DESC_COUNT;
	ptr = (char *)eee_ramAlloc(size);
	if( !ptr )
	{
		scsi_err( SCSI_ERR_MALLOC | SCSI_ERR_LEVEL_0, 
				  "SCSI DESCRIPTORS");
		return SCSI_ERR_MALLOC;
	}
	else
	{
		memset(ptr, 0, size);
		/*
		 *-------------------------------------------
		 * Create the descriptor stack
		 *-------------------------------------------
		 */
		sd = (scsi_desc_t *)ptr;
		scsiCtrl->descCnt  = 0;
		scsiCtrl->descStk = &scsiCtrl->descArray[SCSI_DESC_COUNT];
		for( i=0; i < SCSI_DESC_COUNT; i++ )
		{
			scsi_deallocDesc(sd++);
		}
	}

	/*
	 *--------------------------------------
	 * Allocate memory for MP descriptors
	 *--------------------------------------
	 */
	size = sizeof(scsi_mp_desc_t) * SCSI_MP_DESC_COUNT;
	ptr = (char *)eee_ramAlloc(size);
	if( !ptr )
	{
		scsi_err( SCSI_ERR_MALLOC | SCSI_ERR_LEVEL_0, 
				  "SCSI MP DESCRIPTORS");
		return SCSI_ERR_MALLOC;
	}
	else
	{
		memset(ptr, 0, size);
		/*
		 *-------------------------------------------
		 * Create the MP descriptor stack
		 *-------------------------------------------
		 */
		smpd = (scsi_mp_desc_t *)ptr;
		scsiCtrl->mpDescCnt  = 0;
		scsiCtrl->mpDescStk = &scsiCtrl->mpDescArray[SCSI_MP_DESC_COUNT];
		for( i=0; i < SCSI_MP_DESC_COUNT; i++ )
		{
			scsi_deallocMPDesc(smpd++);
		}
	}

	/*
	 *-----------------------------------------------------
	 * Allocate one ULTRA buffer for padding padded writes.
	 *-----------------------------------------------------
	 */
	scsiCtrl->writePadBuffer = eee_allocateBuffer(EEE_SHARED_POOL, EDESC_BD_ULTRA, 0);
	bzero(scsiCtrl->writePadBuffer, EEE_BUFSIZE_ULTRA);
	CACHE_FLUSH_STRUCT(scsiCtrl->writePadBuffer, EEE_BUFSIZE_ULTRA);

	/*
	 *--------------------------------------
	 * Register with EEE and get app id
	 *--------------------------------------
	 */
	rc = eee_allocateApplicationId("scsi");
	if( rc == NFX_ERR )
	{
		scsi_err( SCSI_ERR_GENERIC | SCSI_ERR_LEVEL_0, 
				  "Allocate app id failed.");
		return NFX_ERR;
	}
	scsiCtrl->appId = rc;
	scsiCtrl->scsiDataAddr = PACK_UNICAST_PORT_ADDRESS( nfxMySlotId, 
														nfxMyCPUId,
														scsiCtrl->appId,
														EEE_USER_DATA,
														NFX_COS_HIGH,
														0);

	rc = eee_registerReceiveFunc( scsiCtrl->appId,	/* App Id		 */
								  scsi_receiveMsg,	/* Ucast Rx Func */
								  NULL);			/* Mcast Rx Func */
	if( rc == NFX_ERR )
	{
		scsi_err( SCSI_ERR_GENERIC | SCSI_ERR_LEVEL_0, 
				  "Register rx func failed.");
		return NFX_ERR;
	}

	/*
	 *---------------------------
	 * Initialize the cod handle 
	 *---------------------------
	scsiCodHandle = (nfxMySlotId << 28) | (nfxMyCPUId << 24) | 0;
	 */
	cod_init();

	/*
	 *----------------------------------
	 * Initialize the device cmd tables.
	 *----------------------------------
	 */
	scsi_diskInitCmds();
	scsi_tapeInitCmds();

#ifdef SCSI_TEST
    rc = eee_registerPollingFunc(EEE_HIGH_PRIORITY, (void *)scsi_test_daemon, NULL);
	if (rc < 0) {
		scsi_err( SCSI_ERR_GENERIC | SCSI_ERR_LEVEL_0,
			  "Register scsi_test_daemon failed.");
                return NFX_ERR;
        }
	rc =  eee_timerCreate(SCSI_TEST_TIMER_PERIOD * 1000, EEE_TIMER_REPEAT);
	VERIFY(rc >= 0);
	rc = eee_timerStart(rc, (void *)scsi_test_timer, NULL, NULL, NULL, NULL);
	VERIFY(rc == NFX_OK);
#endif

	//-- Create and start link timer.
    scsiCtrl->statsTmr = eee_timerCreate(500,		// 500 millisec
										 EEE_TIMER_REPEAT);

    if(scsiCtrl->statsTmr == TIMER_ERR_NO_TIMER_AVAILABLE) {
        printf("SCSI: Could not allocate statistics timer.\n");
    }
    else {
        eee_timerStart(scsiCtrl->statsTmr, 
					   scsi_gatherStats,
                       NULL, NULL, NULL, NULL);
    }

	//-- Create and start SSC msg timer
	scsi_msg_queue_timer_create(SCSI_SDM_POLL_TIMER_PERIOD, scsi_sdm_poll_handler);
	scsi_msg_queue_timer_restart();
	
	//
	//-- In order to start off fresh, we will send a message up to the
	//-- SDM to tell it to delete any stale device entries that it might
	//-- have that are associated with this slot/port.
	scsi_sendSdmDelAllMsg();

	return NFX_OK;
} /* scsi_initModule */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_newDevice
 *
 * Description: 	This function is called	when a new storage device 
 *					is discovered. This is where the device control structure
 *					and state machine instance are created.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	March 6, 2001
 *
 * Returns:		    NFX_OK | NFX_ERR
 *
 *------------------------------------------------------------------------------
 */
int
scsi_newDevice( uint32   tType,		// SCSI_LL_FC | SCSI_LL_IP
				ushort16 llHandle,	// Lower layer device handle
				uint64 	 devWWN,
				uint64   chnWWN,
				char *   chnName,	// Egress switch, egress port wwn.
				uint32   lun, 
				char *   vendor)
{
#if 0
	uint32 rc;
	scsi_dev_t *pdev;
	
	SCSI_FUNC_TRACE("[scsi_newDevice] Enter");

	//printf("tType: %x, llHandle: %x, devWWN: %llx, chnWWN: %llx, chName: %s, lun: %x\n", tType, llHandle, devWWN, chnWWN, chnName, lun);

	//-- See if the device already exists.
	if(scsi_devCheckExist(devWWN, lun, llHandle)) 
		return(NFX_OK);

	//-- Allocate a new device control block
	pdev = (scsi_dev_t *)scsi_allocDevice(lun, llHandle);
	if( !pdev )
		return(NFX_ERR);

	//-- Save WWN
	pdev->wwn    = devWWN;
	pdev->chnWWN = chnWWN;

	//-- Create DEV_ID 
	SCSI_ID_SET_WWN(pdev->id, devWWN);
	SCSI_ID_SET_LUN(pdev->id, lun);

	//-- Create SCSI device state machine instance
	if( !strncmp(vendor, "HITACHI", 7) ) {
		rc = hds9910_I_sm_create( &pdev->smHandle, pdev );
	}
	else {
		rc = scsi_devI_sm_create( &pdev->smHandle, pdev );
	}

	if( rc != NFX_OK ) {

		//-- State machine instance not created so..... 
		scsi_deallocDevice(pdev);        /* free dev ctrl block memory */
		scsi_err( SCSI_ERR_SM_CREATE | SCSI_ERR_LEVEL_0,   /* complain */
				  "func: scsi_newDevice");
		return NFX_ERR;
	}

	SCSI_FUNC_TRACE("[scsi_newDevice] Leave");
#else
	printf("[scsi_newDevice] Old Fucntion, Should not be here!");
#endif
	return( NFX_OK );
} // scsi_newDevice

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_openDevice
 *
 * Description: 	This function is called	when a storage device 
 *					is opened. This is where the device control structure
 *					and state machine instance are created.
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	March 6, 2001
 *
 * Returns:		    NFX_OK | NFX_ERR
 *
 *------------------------------------------------------------------------------
 */
int
scsi_openDevice(scsi_path_t path)
{
	int		   rc;
	scsi_dev_t *pdev;
	
	pdev  = scsi_getDevice(path);
	if( !pdev ) {
		return SCSI_MSG_ECODE_NOSUCH_DEV;
	}

	if( pdev->flags.fOpenMode == 1 )
		return SCSI_MSG_ECODE_OK;
	
	// Destroy init statemachine instance.  ::Note::Will be removed once
	// SM code is fixed to allow a statemachine to terminate itself.
	// esm_destroy(pdev->smHandle);	

	// Create SCSI device state machine instance
	rc = scsi_devO_sm_create( &pdev->smHandle, pdev );
	if( rc != NFX_OK )
	{
		/* State machine instance not created so..... */
		scsi_err( SCSI_ERR_SM_CREATE | SCSI_ERR_LEVEL_0,   /* complain */
				  "SCSI OPEN DEVICE");
		return SCSI_MSG_ECODE_NO_RESOURCE;
	}
	else {
		//		pdev->flags.fInitMode  = 0;
		pdev->flags.fCloseMode = 0;
		pdev->flags.fOpenMode  = 1;
	}

	return SCSI_MSG_ECODE_OK;
} /* scsi_openDevice */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_deleteDevice
 *
 * Description: 	This function is called by the Fibre Channel layer when a 
 *					storage	device is removed from the network. 
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	March 6, 2001
 *					
 * Returns:		SCSI_OK
 *				SCSI_ERROR
 *
 *------------------------------------------------------------------------------
 */
int
scsi_deleteDevicell(uint32   tType,		/* SCSI_LL_FC | SCSI_LL_IP		*/
					ushort16 llHandle,	/* Lower layer device handle 	*/
					uint64   devWWN,
					uint32   lun)  
{
	uint32      i;
	uint32		event;
	scsi_dev_t *pdev;
    esm_event_t esmEvent;

	scsi_debug("scsi_devO_deleteDevice: [Enter]", SCSI_DBL_3);

	for( i=0; i < SCSI_MAX_DEV_PER_PORT; i++ ) {
		if( (pdev = scsiDevList[i]) ) {
			if( pdev->llHandle == llHandle) {

				// Send message to the SDM to remove the device/path.
				scsi_sendRemDevMsg( pdev );

				// If the device is in the open mode then send message 
				// to the state machine to remove the device. Otherwise 
				// just deallocate it.
				if( pdev->smHandle ) {					 

#if SDM_DEV_INIT
					event =  SCSI_DEVO_EVENT_DELETE;
#else
					if( pdev->flags.fInitMode )
						event =  SCSI_DEVI_EVENT_DELETE; 
					else 
						event =  SCSI_DEVO_EVENT_DELETE;
#endif
                    esmEvent.event = event;
                    esmEvent.p1.u.w = 0;
                    esmEvent.p2.u.w = 0;
                    esm_deliverEvent(pdev->smHandle, &esmEvent);
				}
				else {
					scsi_deallocDevice( pdev );
				}
			}
		}
	}

	scsi_debug("scsi_devO_deleteDevice: [Leave]", SCSI_DBL_3);
	return( NFX_OK );
} /* scsi_deleteDevice */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_allocDevice
 *
 * Description: 	Allocates a device control block.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 23, 2001
 *					
 * Returns: 
 *
 *------------------------------------------------------------------------------
 */
scsi_dev_t *
scsi_allocDevice(uint32 lun, ushort16 llHandle)
{
	uint32 align_diff;
	uint32 i, size;
	scsi_dev_t *pdev;

	SCSI_FUNC_TRACE("[scsi_allocDevice] Enter");

	if( scsiCtrl->nDevices >= SCSI_MAX_DEV_PER_PORT )
		return(0);

	size = sizeof( scsi_dev_t );
	pdev = (scsi_dev_t *)eee_ramAlloc(size + 8);
	if( !pdev ) {
		scsi_err( SCSI_ERR_MALLOC | SCSI_ERR_LEVEL_0, 
				  "DEVICE CONTROL BLOCK");
		return 0;
	}
	else {
		bzero(pdev, size + 8);
		align_diff = (uint32)pdev & 0x00000007;

		if( align_diff )
            align_diff = 8 - align_diff;
        (uint32)pdev += align_diff;
	}
	
	for( i=1; i < SCSI_MAX_DEV_PER_PORT; i++ ) {

		if( scsiDevList[i] )
			continue;

		scsiDevList[i] = pdev;
		break;
	}

	pdev->llHandle = llHandle;
	strcpy(pdev->vendor,"unknown");
	strcpy(pdev->model,"unknown");
	strcpy(pdev->productId, "unknown");
	pdev->tape_iocookie = 1;

	/*
	 * Create DEV_PATH 
	 */
	SCSI_PATH_SET_LUN(pdev->path, lun);
	SCSI_PATH_SET_SLOT(pdev->path, nfxMySlotId);
	SCSI_PATH_SET_PORT(pdev->path, nfxMyCPUId);
	SCSI_PATH_SET_HANDLE(pdev->path, i);	

	/*
	 * Create device queues 
	 */
	pdev->genio_q = scsi_queue_create(Q_TYPE_SORTED, SCSI_GENIO_TIMER_DURATION, (void *)scsi_genIOQueueTimerHandler, pdev);
	pdev->genio_q_hiwater_mark = SCSI_GENIO_HIWATER_MARK;
	pdev->hiprio_q = scsi_queue_create(Q_TYPE_LILO, 0, NULL, NULL);
	pdev->other_q = scsi_queue_create(Q_TYPE_LILO, 0, NULL, NULL);

#if 0 //-- Turn this off for now: FSDEBUG
	/*
	 * Create write gate timer
	 */
	scsi_write_gate_timer_create(pdev);
#endif

	scsiCtrl->nDevices++;
	SCSI_FUNC_TRACE("[scsi_allocDevice] Leave");
	return(pdev);
} /* scsi_allocDevice */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_deallocDevice
 *
 * Description: 	Deallocates a device control block.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 31, 2001
 *					
 * Returns: 
 *
 *------------------------------------------------------------------------------
 */
int
scsi_deallocDevice( scsi_dev_t *pdev )
{
	uint32 i;

	SCSI_FUNC_TRACE("[scsi_deallocDevice] Enter");

	if( !pdev )
		return 0;

	for( i=1; i < SCSI_MAX_DEV_PER_PORT; i++ ) {
		if( scsiDevList[i] == pdev ) {
			scsiDevList[i] = 0;
			scsiCtrl->nDevices--;

			/* Destroy device queues */
			scsi_queue_destroy(pdev->genio_q);
			scsi_queue_destroy(pdev->hiprio_q);
			scsi_queue_destroy(pdev->other_q);
			
			/* Empty tape io queue */
			if(pdev->type == SCSI_DEVTYPE_TAPE) {

				eee_desc_t	*eDesc;
				scsi_tape_io_q_t *hd = (scsi_tape_io_q_t *)pdev->tape_io_q;
				scsi_tape_io_q_t *tmp_msg;

				while(hd) {
					eDesc = (eee_desc_t *)hd->msg;
					eee_freePacket(eDesc);
					tmp_msg = hd;
					hd = hd->next;
					eee_ramDealloc(tmp_msg);
				}
			}
#if 0 //-- Turn this off for now: FSDEBUG
			/* Destroy write gate timer */
			scsi_write_gate_timer_destroy(pdev);
#endif

			eee_ramDealloc(pdev);
			break;
		}
	}
	SCSI_FUNC_TRACE("[scsi_deallocDevice] Leave");
	return(NFX_OK);
} /* scsi_deallocDevice */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_getDevice
 *
 * Description: 	Using devId gets a pointer to a SCSI device context block.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	March 7, 2001
 *					
 * Returns: 
 *
 *------------------------------------------------------------------------------
 */
scsi_dev_t*
scsi_getDevice( scsi_path_t path )
{
	uint32 handle;
	scsi_dev_t *pdev;

	handle = SCSI_PATH_GET_HANDLE(path);

	if( handle < 0 || handle > SCSI_MAX_DEV_PER_PORT )
		pdev = 0;
	else 
		pdev = scsiDevList[SCSI_PATH_GET_HANDLE(path)];

	return pdev;
} /* scsi_getDevice */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_devCheckExist
 *
 * Description: 	Checks for the existence of a device control block
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	June 04, 2001
 *					
 * Returns: 
 *
 *------------------------------------------------------------------------------
 */
scsi_dev_t *
scsi_devCheckExist( uint64 devWWN, uint32 lun, uint32 llHandle )
{
	uint32 i;
	scsi_dev_t *pdev;

	SCSI_FUNC_TRACE("[scsi_devCheckExist] Enter");
	//printf("wwn: %llx, lun: %d, handle: %x\n", devWWN, lun, llHandle);
	//devWWN &= 0x0000FFFFFFFFFFFFLL;

	/* printf("WWN: %llx, LUN: %d\n", devWWN, lun); */

	for( i=0; i < SCSI_MAX_DEV_PER_PORT; i++ ) {

		if( (pdev = scsiDevList[i]) ) {
			if( (pdev->wwn == devWWN) && 
				(pdev->lun == lun)    &&
				(pdev->llHandle == llHandle) ) {
				//printf("SCSI: Device Already Exists: wwn 0x%llx, lun %d, llHandle 0x%x\n", devWWN, lun, llHandle);
				return(pdev);
			}
		}
	}

	SCSI_FUNC_TRACE("[scsi_devCheckExist] Leave");
	return(0);
} /* scsi_devCheckExist */

int 
scsi_showDeviceList()
{
	int32 i;

	for(i=0; i < SCSI_MAX_DEV_PER_PORT; i++) {
		scsi_showDevice( scsiDevList[i] );
	}

	return 0;
} /* scsi_showDeviceList */

int
scsi_showDevice( scsi_dev_t *pdev )
{
	char devMode[80];
	int i;

	if( pdev ) {
		printf("id           :   %016llx :\n", pdev->id);
		printf("path         :\t\t%08x : ", pdev->path);
		printf("slot: %d, port: %d, lun: %d, handle: %x\n", 
			   SCSI_PATH_GET_SLOT(pdev->path),
			   SCSI_PATH_GET_PORT(pdev->path),
			   SCSI_PATH_GET_LUN(pdev->path),
			   SCSI_PATH_GET_HANDLE(pdev->path));
		printf("llHandle     :\t\t%08x :\n",pdev->llHandle);
		printf("type         :\t\t%08x :\n",pdev->type);
		printf("blk cnt      :\t\t%10d :\n",pdev->blocks);
		printf("blk size     :\t\t%08d :\n",pdev->blockSize);
		printf("ansi ver     :\t\t%08x :\n",pdev->ansiVersion);
		printf("vendor       :   %16s :\n",pdev->vendor);
		printf("model        :   %16s :\n",pdev->model);
		printf("productId    :   %16s :\n",pdev->productId);
		if (pdev->type == SCSI_DEVTYPE_DISK)
			sprintf(devMode, "%s/%s/%s/%s",
					( ( pdev->flags.fOpenMode ) ? "OPEN" :
					  ( pdev->flags.fCloseMode ) ? "CLOSED" :
					  ( pdev->flags.fInitMode ) ? "INIT" :
					  "UNKNOWN" ),
					(pdev->flags.fAllowSort ? "SORT" : "NOSORT"),
					(pdev->flags.fAllowChain ? "CHAIN" : "NOCHAIN"),
					(pdev->flags.fAllowEmbedded_CS ? 
					 (pdev->flags.fLocalCSVerify ? "CS=EMBEDDED,WR_VERF,RD_VERF" : "CS=EMBEDDED,NO_VERF") : 
					(pdev->flags.fAllowBlock_CS ? 
					 (pdev->flags.fLocalCSVerify ? "CS=BLOCK,WR_VERF,RD_VERF" : "CS=BLOCK,NO_VERF") : 
					 (pdev->flags.fLocalCSVerify ? "CS=NONE,WR_VERF" : "CS=NONE,NO_VERF"))));
		else
			sprintf(devMode, "%s/%s/%s",
					( ( pdev->flags.fOpenMode ) ? "OPEN" :
					  ( pdev->flags.fCloseMode ) ? "CLOSED" :
					  ( pdev->flags.fInitMode ) ? "INIT" :
					  "UNKNOWN" ),
					(pdev->flags.fAllowSort ? "SORT" : "NOSORT"),
					(pdev->flags.fAllowChain ? "CHAIN" : "NOCHAIN"));
		printf("dev mode                :   %16s :\n", devMode);
		printf("sm Handle               :\t\t%08x :\n",pdev->smHandle);
		printf("next Evt                :\t\t%08x :\n",pdev->nextEvent);
		printf("cur blk                 :\t\t%08d :\n",pdev->curBlock);
		printf("max dispatch            :\t\t%08x :\n",pdev->maxDispatchCnt);
		printf("dispatch cnt            :\t\t%08x :\n",pdev->dispatchCnt);
		printf("dispatch Q              :\t\t%p :\n"  ,pdev->dispatchQ);
		printf("read req cnt            :\t\t%08d :\n",pdev->readReqCnt);
		printf("read rsp cnt            :\t\t%08d :\n",pdev->readRspCnt);
		printf("sub-8k write cnt        :\t\t%08d :\n",pdev->writeSub8kCnt);
		printf("sub-8k padded write cnt :\t\t%08d :\n",pdev->writeSub8kPaddedCnt);
		printf("write req cnt           :\t\t%08d :\n",pdev->writeReqCnt);
		printf("write rsp cnt           :\t\t%08d :\n",pdev->writeRspCnt);
		printf("relay req cnt           :\t\t%08d :\n",pdev->relayReqCnt);
		printf("relay rsp cnt           :\t\t%08d :\n",pdev->relayRspCnt);
		printf("dev err cnt             :\t\t%08d :\n",pdev->devErrCnt);
		printf("read CSE cnt            :\t\t%08d :\n",pdev->readCSECnt);
		printf("read BE cnt             :\t\t%08d :\n",pdev->readBECnt);
		printf("write CSE cnt           :\t\t%08d :\n",pdev->writeCSECnt);
		printf("nacks Received          :\t\t%08d :\n",pdev->nacksReceived);
		printf("nacks Transmitted       :\t\t%08d :\n",pdev->nacksTransmitted);
		for (i = 0; i < SCSI_COALLESCE_STATS_BUCKETS+1; ++i) {
		    if ((i % 4) == 0) {
				if (i == 0)
					printf("R chain                 :\t\t"); 
				else 
					printf("\n                        :\t\t"); 			    
		    }
		    printf("%08d ", pdev->r_chain_stats[i]);
		}
		printf("\n");
		for (i = 0; i < SCSI_COALLESCE_STATS_BUCKETS+1; ++i) {
		    if ((i % 4) == 0) {
				if (i == 0)
					printf("W chain                  :\t\t"); 
				else 
					printf("\n                         :\t\t"); 			    
		    }
		    printf("%08d ", pdev->w_chain_stats[i]);
		}
		printf("\n");
		printf("GENQ hiwat               :\t\t%08d\n",pdev->genio_q_hiwater_mark);
		printf("GENQ tmr(mS)             :\t\t%08d\n",SCSI_GENIO_TIMER_DURATION);
		printf("GENQ len                 :\t\t%08d\n",scsi_queue_len(pdev->genio_q));
		printf("GENQ imm. cnt            :\t\t%08d\n",pdev->genio_q_immediate_schedule_count);
		printf("HIPRIOQ len              :\t\t%08d\n",scsi_queue_len(pdev->hiprio_q));
		printf("NOREORDER BAR            :\t\t%08d\n",pdev->noreorder_barrier);
		printf("OTHERQ len               :\t\t%08d\n",scsi_queue_len(pdev->other_q));
		printf("=======================================\n");
	}
	return 0;
} /* scsi_showDevice */

int
scsi_showCtrl()
{
	printf("[SCSI Control Block] Slot: %d, CPU: %d\n", nfxMySlotId, nfxMyCPUId);
	printf("=======================================\n\n");
	printf("AppId               : %08x\n", scsiCtrl->appId);
	printf("Active Device Cnt   : %08d\n", scsiCtrl->nDevices);
	printf("Free SCSI Desc Cnt  : %08d\n", scsiCtrl->descCnt);
	printf("SCSI Desc Stack	    : %p\n",   scsiCtrl->descStk);
	printf("\n");
	return 0;
}  /* scsi_showCtrl */

int
scsi_showDevReqQ(scsi_path_t path)
{
	scsi_dev_t  *pdev;

	printf("[Device Req Q] Slot: %d, CPU: %d, Handle: %d\n", 
		   nfxMySlotId, nfxMyCPUId, SCSI_PATH_GET_HANDLE(path));
	printf("================================================\n\n");
	pdev = scsi_getDevice(path);
	if( !pdev ) {
		printf("Device Not Found.\n");
		return 0;
	}
	printf("general I/O Q:\n");
	scsi_queue_dump(pdev->genio_q);
	printf("hi priority ordered Q:\n");
	scsi_queue_dump(pdev->hiprio_q);
	printf("other Q:\n");
	scsi_queue_dump(pdev->other_q);
	return 0;
}  /* scsi_showCtrl */

int
scsi_setTracing()
{
	char   cmd[80];
	uint32 sel;

	printf("Func Tracing: %s\n", scsi_func_trace ? "ENABLED":"DISABLE");
	printf("Func Trace Dis/Ena [0|1]: ");
#ifdef LINUX_TEST
	fflush(0);
#endif /* LINUX_TEST */
	get_cmd(cmd);
	sscanf(cmd, "%x", &sel);
	switch(sel) {
		case 0:
			scsi_func_trace = 0;
			break;
		case 1:
			scsi_func_trace = 1;
			break;
		default:
			printf("Invalid Selection: %d\n",sel);
	}
	printf("Func Tracing: %s\n", scsi_func_trace ? "ENABLED":"DISABLE");

	printf("Pkt Tracing: %s\n", scsi_pkt_trace ? "ENABLED":"DISABLE");
	printf("Pkt Trace Dis/Ena [0|1]: ");
#ifdef LINUX_TEST
	fflush(0);
#endif /* LINUX_TEST */
	get_cmd(cmd);
	sscanf(cmd, "%x", &sel);

	switch(sel) {
		case 0:
			scsi_pkt_trace = 0;
			break;
		case 1:
			scsi_pkt_trace = 1;
			break;
		default:
			printf("Invalid Selection: %d\n",sel);
	}
	printf("Pkt Tracing: %s\n", scsi_pkt_trace ? "ENABLED":"DISABLE");

	return 0;
} // scsi_setTracing

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_allocDesc
 *
 * Description: 	Allocate a SCSI descriptor
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	March 26, 2001
 *					
 * Returns: 		scsi_desc_t *
 *
 *------------------------------------------------------------------------------
 */
scsi_desc_t *
scsi_allocDesc(void)
{
	scsi_desc_t *sd;

	if( scsiCtrl->descCnt > 0 )
	{
		sd = *(scsiCtrl->descStk);
		scsiCtrl->descStk++;
		scsiCtrl->descCnt--;
		memset(sd,0,sizeof(scsi_desc_t));
	}
	else
	{
		sd = (scsi_desc_t *)NULL;
	}

	return sd;
} /* scsi_allocDesc */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_deallocDesc
 *
 * Description: 	Deallocates a SCSI descriptor
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	March 26, 2001
 *					
 * Returns: 
 *
 *------------------------------------------------------------------------------
 */
void
scsi_deallocDesc( scsi_desc_t *sd )
{
    eee_desc_t *ed;

	if (sd->control & SD_FLAG_NO_DEALLOC)
		return;

	/*
	 * Free the eDesc(s) and associated buffers.
	 */
	if( sd->eDesc ) {
		while( (ed = sd->eDesc) ) {
			if (ed->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER)
				sd->eDesc = (eee_desc_t *)ed->bd[EEE_MAX_BUFS-1].buf;
			else 
				sd->eDesc = NULL;
			ed->hdr.control &= ~EDESC_COMMAND_CHAIN_LAST_BUFFER;
			eee_freePacket(ed);
		}
	}

	if( sd->senseData )
		eee_ramDealloc( sd->senseData );

	if( (scsiCtrl->descCnt + 1) > SCSI_DESC_COUNT )
	{
		/* 
		 * oops something went wrong. We have more descriptors
		 * than we started off with.
		 */
		printf("SCSI Error: Bonus SCSI Descriptors Detected\n");
	}
	else
	{
		scsiCtrl->descStk--;
		*(scsiCtrl->descStk) = sd;
		scsiCtrl->descCnt++;
	}
	return;
} /* scsi_deallocDesc */ 

void
scsi_displayDesc(scsi_desc_t *sd)
{
	int		i;
	eee_desc_t *    ed;

	printf("      control 0x%x, halHandle 0x%x, cHandle 0x%x, status 0x%x\n",
	       sd->control, sd->halHandle, sd->cHandle, sd->status);
	printf("      expectDataLen %d, lba %d, lbCnt %d, lbCntChained %d, retryCnt %d, timeOutCnt %d\n",
	       sd->expectDataLen, sd->lba, sd->lbCnt, sd->lbCntChained, sd->retryCnt, sd->timeOutCnt);
	printf("      rspDest 0x%x, dataDest 0x%x\n", sd->rspDest, sd->dataDest);
	printf("      cdb ");
	for (i = 0; i < 16; ++i)
		printf("%02x ", sd->cdb[i]);
	printf("\n");
	if (sd->eDesc) {
		ed = sd->eDesc;
		while (ed) {
			eee_descDisplay(ed);
			if (ed->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER)
				ed = (eee_desc_t *)ed->bd[EEE_MAX_BUFS-1].buf;
			else
				ed = NULL;
		} /* while (ed) */
	}
}

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_allocMPDesc
 *
 * Description: 	Allocate a SCSI MP descriptor
 * 				
 * Created by:		Radek Aster
 *
 * Data Created:	Dec 30, 2001
 *					
 * Returns: 		scsi_mp_desc_t *
 *
 *------------------------------------------------------------------------------
 */
scsi_mp_desc_t *
scsi_allocMPDesc(void)
{
	scsi_mp_desc_t *smpd;

	if( scsiCtrl->mpDescCnt > 0 )
	{
		smpd = *(scsiCtrl->mpDescStk);
		scsiCtrl->mpDescStk++;
		scsiCtrl->mpDescCnt--;
		memset(smpd, 0, sizeof(scsi_mp_desc_t));
	}
	else
		smpd = (scsi_mp_desc_t *)NULL;
	
	return smpd;
} /* scsi_allocMPDesc */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_deallocMPDesc
 *
 * Description: 	Deallocates a SCSI MP descriptor
 * 				
 * Created by:		Radek Aster
 *
 * Data Created:	Dec 30, 2001
 *					
 * Returns: 
 *
 *------------------------------------------------------------------------------
 */
void
scsi_deallocMPDesc( scsi_mp_desc_t *smpd )
{
	if( (scsiCtrl->mpDescCnt + 1) > SCSI_MP_DESC_COUNT )
	{
		/* 
		 * oops something went wrong. We have more descriptors
		 * than we started off with.
		 */
		printf("SCSI Error: Bonus SCSI MP Descriptors Detected\n");
	}
	else
	{
		scsiCtrl->mpDescStk--;
		*(scsiCtrl->mpDescStk) = smpd;
		scsiCtrl->mpDescCnt++;
	}
	return;
} /* scsi_deallocMPDesc */ 


scsi_mp_desc_t *
scsi_MPEntryNew(scsi_dev_t *pdev, uint32 tag)
{
    int				i;
    scsi_mp_desc_t *smpd;

    /* VERIFY no duplicate tags */
    VERIFY(scsi_MPEntryLookup(pdev, tag) == NULL);
    
    /* Find an unused slot */
    for (i = 0; i < SCSI_MAX_MP_PER_DEV; ++i) {
		if (pdev->activeMP[i] == NULL) {
			smpd = pdev->activeMP[i] = scsi_allocMPDesc();
			VERIFY(smpd);
			bzero(smpd, sizeof(scsi_mp_desc_t));
			smpd->tag = tag;
			smpd->pdev = pdev;
			smpd->ack_timer_handle = -1;
			smpd->exch_timer_handle = -1;
			return(smpd);
		}
	}
    return(NULL);
    
}

scsi_mp_desc_t *
scsi_MPEntryLookup(scsi_dev_t *pdev, uint32 tag)
{
    int i;

    for (i = 0; i < SCSI_MAX_MP_PER_DEV; ++i) {
		if (pdev->activeMP[i] && pdev->activeMP[i]->tag == tag) {
			VERIFY(pdev->activeMP[i]->pdev == pdev);
			return(pdev->activeMP[i]);
		}
    }
    return(NULL);
}

void
scsi_MPEntryDestroy(scsi_mp_desc_t *smpd)
{
    int					i;
	scsi_dev_t *		pdev;

	VERIFY(scsi_MPEntryLookup(smpd->pdev, smpd->tag) == smpd);
	VERIFY(smpd->ack_timer_handle == -1);
	VERIFY(smpd->exch_timer_handle == -1);
	VERIFY(smpd->private == NULL);

	pdev = (scsi_dev_t *)smpd->pdev;

    for (i = 0; i < SCSI_MAX_MP_PER_DEV; ++i) {
		if (pdev->activeMP[i] == smpd) {
			pdev->activeMP[i] = NULL;
			scsi_deallocMPDesc(smpd);
		}
	}
}

void
scsi_destroyDescChain(eee_desc_t *ed)
{
    eee_desc_t *		ed_chain = ed;

	while ((ed = ed_chain)) {
		if (ed_chain->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER) {
			VERIFY(!(ed_chain->bd[EEE_MAX_BUFS-1].control & EDESC_BD_CMD_IN_USE));
			ed_chain = (eee_desc_t *)ed_chain->bd[EEE_MAX_BUFS-1].buf;
		}
		else
			ed_chain = NULL;
        eee_freePacket(ed);		
	}
    return;
}

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_scsi_getDevHandle
 *
 * Description: 	Returns handle for new device.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 10, 2001
 *					
 * Returns: 
 *
 *------------------------------------------------------------------------------
 */
int
scsi_getDevHandle()
{
	return 0;
} /* scsi_getDevHandle */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_getDevDB
 *
 * Description: 	Consults SCSI device database for info on a specific 
 *					device.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 14, 2001
 *					
 * Returns: 
 *
 *------------------------------------------------------------------------------
 */
int *  
scsi_getDevDB(char *mfgStr, char *modelStr)
{
#if 0   /**** TODO *****/
	int32        found=0;
	scsi_devDB_t 	*devDb;

	devDb = &scsi_devDB[0];
	for( ; devDB; devDB++) {

	}

	if( found ) {

	}
	else {
		return((scsi_devDB_t *)NULL);
	}
#endif   /**** TODO *****/
	return NFX_OK;
} /* scsi_getDevDB */



/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_getCodHandle
 *
 * Description: 	Allocates a COD handle.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	June 06, 2001
 *					
 *------------------------------------------------------------------------------
 */
int
scsi_getCodHandle(uint64 *codHandle)
{
	//*codHandle = 0;
	*codHandle = COD_GET_NEXT();
	return NFX_OK;
} /* scsi_getCodHandle */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_flushReqQ
 *
 * Description: 	Flushs all of the pending commands on a devices request
 * 					queue and sends a reject command back to the originator.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	June 11, 2001
 *					
 *------------------------------------------------------------------------------
 */
int
scsi_flushReqQ(scsi_dev_t *pdev)
{
	scsi_desc_t *sd;

	while( (sd = scsi_dequeue(pdev->genio_q)) ) {
	    scsi_sendCmdErrRsp(sd->rspDest, 
						   sd->cHandle,
						   SCSI_MSG_ECODE_DEV_CLOSED);
		scsi_deallocDesc(sd);
	}
	while( (sd = scsi_dequeue(pdev->hiprio_q)) ) {
	    scsi_sendCmdErrRsp(sd->rspDest, 
						   sd->cHandle,
						   SCSI_MSG_ECODE_DEV_CLOSED);
		scsi_deallocDesc(sd);
	}
	while( (sd = scsi_dequeue(pdev->other_q)) ) {
	    scsi_sendCmdErrRsp(sd->rspDest, 
						   sd->cHandle,
						   SCSI_MSG_ECODE_DEV_CLOSED);
		scsi_deallocDesc(sd);
	}
	return 0;
} /* scsi_flushReqQ */

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_gatherStats
 *
 * Description: 	
 *
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 05, 2001
 *
 * Returns:         NFX_OK | NFX_ERR 
 *
 *------------------------------------------------------------------------------
 */
int32
scsi_gatherStats( int32 tmr,				/* Timer Handle */
				  void *user1,
				  void *user2,
				  void *user3,
				  void *user4
				 )
{
	uint32 i, diff;
	uint64 diff1;
	scsi_dev_t *pdev;

	for(i=0; i < SCSI_MAX_DEV_PER_PORT; i++) {
		if( (pdev = scsiDevList[i]) ) {
			if( pdev->flags.fInitMode == 0 ) {

				diff = pdev->readRspCnt - pdev->readRspLast;
				pdev->readRspLast = pdev->readRspCnt;
				if( diff )
					pdev->readIoPerSecCur = diff * 2;

				diff = pdev->writeRspCnt - pdev->writeRspLast;
				pdev->writeRspLast = pdev->writeRspCnt;
				if( diff )
					pdev->writeIoPerSecCur = diff * 2;

				if( pdev->readIoPerSecCur > pdev->readIoPerSecMax )
					pdev->readIoPerSecMax  = pdev->readIoPerSecCur; 
				if( pdev->writeIoPerSecCur > pdev->writeIoPerSecMax )
					pdev->writeIoPerSecMax  = pdev->writeIoPerSecCur; 

				//-- Look for counter wrap condition.
				if( pdev->readByteCnt >= pdev->readByteLast ) {
					diff1 = pdev->readByteCnt - pdev->readByteLast;
					pdev->readByteLast = pdev->readByteCnt;
					if( diff1 )
						pdev->readBytePerSecCur = diff1 * 2;

					if( pdev->readBytePerSecCur > pdev->readBytePerSecMax )
						pdev->readBytePerSecMax  = pdev->readBytePerSecCur; 
				}
				else {
					pdev->readByteLast = pdev->readByteCnt;
				}

				//-- Look for counter wrap condition.
				if( pdev->writeByteCnt >= pdev->writeByteLast ) {
					diff1 = pdev->writeByteCnt - pdev->writeByteLast;
					pdev->writeByteLast = pdev->writeByteCnt;
					if( diff1 )
						pdev->writeBytePerSecCur = diff1 * 2;

					if( pdev->writeBytePerSecCur > pdev->writeBytePerSecMax )
						pdev->writeBytePerSecMax  = pdev->writeBytePerSecCur; 
				}
				else {
					pdev->writeByteLast = pdev->writeByteCnt;
				}

			}
		}
	}

	return NFX_OK;
} 

int
scsi_write_gate_timer_handler(int32 timer_handle, scsi_dev_t *pdev)
{
    esm_event_t esmEvent;

	VERIFY(timer_handle == pdev->write_gate_timer_handle);
	
	pdev->write_gate_timer_running = 0;

    esmEvent.event = SCSI_DEVO_EVENT_WRITE_GATE_TIMER;
    esmEvent.p1.u.p = NULL;
    esmEvent.p2.u.p = NULL;
    esm_deliverEvent(pdev->smHandle, &esmEvent);

	return(NFX_OK);
}

void
scsi_write_gate_timer_create(scsi_dev_t *pdev)
{
	pdev->write_gate_timer_handle = eee_timerCreate(SCSI_DEFAULT_WRITE_GATE_PERIOD, EEE_TIMER_ONCE);
	VERIFY(pdev->write_gate_timer_handle >= 0);
	pdev->write_gate_timer_running = 0;
}

void
scsi_write_gate_timer_destroy(scsi_dev_t *pdev)
{
	int32		rc;

	if (pdev->write_gate_timer_running)
		scsi_write_gate_timer_stop(pdev);

	rc = eee_timerDestroy(pdev->write_gate_timer_handle);
	VERIFY(rc == NFX_OK);
	pdev->write_gate_timer_handle = -1;
}

void
scsi_write_gate_timer_restart(scsi_dev_t *pdev)
{
	int32		rc;

	if (pdev->write_gate_timer_running) {
		rc = eee_timerStop(pdev->write_gate_timer_handle);
		VERIFY(rc == NFX_OK);
	}
	rc = eee_timerStart(pdev->write_gate_timer_handle, (void *)scsi_write_gate_timer_handler, pdev, NULL, NULL, NULL);
	VERIFY(rc == NFX_OK);
	pdev->write_gate_timer_running = 1;
}

void
scsi_write_gate_timer_stop(scsi_dev_t *pdev)
{
	int32		rc;

	if (pdev->write_gate_timer_running) {
		rc = eee_timerStop(pdev->write_gate_timer_handle);
		VERIFY(rc == NFX_OK);
	}
	pdev->write_gate_timer_running = 0;
}


eee_desc_t *
scsi_copyEEEDesc(eee_desc_t *ed_src)
{
	eee_desc_t *		ed_copy;
	eee_descBuf_t *		bd_src;
	eee_descBuf_t *		bd_dest;
	int					buf_count;
	int					buf_byte_count;
	int					i;

	ed_copy = eee_allocateDesc(EEE_SHARED_POOL, 0);
	VERIFY(ed_copy);
	bzero(ed_copy, sizeof(eee_desc_t));
	buf_count = EEE_GET_NUM_BUFS(ed_src);
	buf_byte_count = EEE_GET_PKT_LEN(ed_src);
	EEE_SET_NUM_BUFS(ed_copy, buf_count);
	EEE_SET_PKT_LEN(ed_copy, buf_byte_count);
	bd_src = &ed_src->bd[0];
	bd_dest = &ed_copy->bd[0];
	if (!(bd_src->control & EDESC_BD_CMD_IN_USE)) {
		++bd_src;
		++bd_dest;
	}
	for (i = 0; i < buf_count; ++i, ++bd_src, ++bd_dest) {
		VERIFY(bd_src->control & EDESC_BD_CMD_IN_USE);
		bd_dest->control = bd_src->control;
		bd_dest->control |= EDESC_BD_CMD_NO_DEALLOC;
		bd_dest->len = bd_src->len;
		bd_dest->buf = bd_src->buf;
	}
	return(ed_copy);
}

int32 
scsi_ack_timeout_handler(int32 timer_handle, scsi_mp_desc_t * smpd)
{
	int32				rc;
	scsi_desc_t *		sd;
	eee_desc_t *		ed_chain;
	uint32				csum;

	VERIFY(timer_handle = smpd->ack_timer_handle);
	rc = eee_timerDestroy(smpd->ack_timer_handle);
	VERIFY(rc == NFX_OK);
	smpd->ack_timer_handle = -1;

	sd = (scsi_desc_t *)smpd->private;
	SCSI_DBG(1, ("scsi_ack_timeout_handler: smpd %p, path 0x%x, cHandle 0x%x\n", 
				 smpd, sd->path, sd->cHandle));

	if (smpd->nack_received) {
		/* Retransmit read sequence */
		smpd->transactionId = scsi_getTransactionId();
		smpd->lastSequenceId = 1;

		ed_chain = smpd->firstEDesc;
		while (ed_chain) {
			uint64	data_offset = 0;
			
			/* Compute fragment checksum, if overall checksum is valid. */
			if (smpd->accumulated_checksum)
				csum = scsi_compute_checksum(ed_chain);
			sd->eDesc = scsi_copyEEEDesc(ed_chain);
			scsi_sendReadRspWithData(sd, smpd->eCode, csum,
									 SCSI_MSG_TRANSSEQID_PACK(smpd->transactionId,
															  smpd->maxSequenceId,
															  smpd->lastSequenceId),
									 &data_offset);
			++smpd->lastSequenceId;

			if (ed_chain->hdr.control & EDESC_COMMAND_CHAIN_LAST_BUFFER) {
				VERIFY(!(ed_chain->bd[EEE_MAX_BUFS-1].control & EDESC_BD_CMD_IN_USE));
				ed_chain = (eee_desc_t *)ed_chain->bd[EEE_MAX_BUFS-1].buf;
			}
			else 
				ed_chain = NULL;
		} /* while (ed_chain) */
		smpd->nack_received = 0;
		scsi_start_ack_timer(smpd);
	}
	else {
		/* Free scsi desc */
		sd->eDesc = smpd->firstEDesc;
		sd->control &= ~SD_FLAG_NO_DEALLOC;
		scsi_deallocDesc(sd);

		/* Free smpd context */
		smpd->private = NULL;
		scsi_MPEntryDestroy(smpd);
	}

	return(NFX_OK);
}

int32 
scsi_exch_timeout_handler(int32 timer_handle, scsi_mp_desc_t * smpd)
{
	int32		rc;

	VERIFY(timer_handle = smpd->exch_timer_handle);

	SCSI_DBG(1, ("scsi_exch_timeout_handler: smpd %p, cHandle 0x%x\n", smpd, smpd->tag));
	
	rc = eee_timerDestroy(smpd->exch_timer_handle);
	VERIFY(rc == NFX_OK);
	smpd->exch_timer_handle = -1;

	if (smpd->no_guaranteed_delivery)
		scsi_sendCmdErrRsp(smpd->firstEDesc->hdr.src_port, smpd->tag, SCSI_MSG_ECODE_INVALID_TXN_SEQ);
	else {
		scsi_sendAck(smpd->firstEDesc->hdr.src_port, smpd->tag, 1);
		++((scsi_dev_t *)(smpd->pdev))->nacksTransmitted;
	}
	scsi_destroyDescChain(smpd->firstEDesc);
	scsi_MPEntryDestroy(smpd);
	
	return(NFX_OK);
}

void
scsi_start_ack_timer(scsi_mp_desc_t * smpd)
{
	int32		rc;

	smpd->ack_timer_handle = eee_timerCreate(SCSI_ACK_TIMEOUT, EEE_TIMER_ONCE);
	VERIFY(smpd->ack_timer_handle >= 0);
	rc = eee_timerStart(smpd->ack_timer_handle, (void *)scsi_ack_timeout_handler,
						smpd, NULL, NULL, NULL);
	VERIFY(rc == NFX_OK);
}

void
scsi_stop_ack_timer(scsi_mp_desc_t * smpd)
{
	int32		rc;

	rc = eee_timerStop(smpd->ack_timer_handle);
	VERIFY(rc == NFX_OK);
	rc = eee_timerDestroy(smpd->ack_timer_handle);
	VERIFY(rc == NFX_OK);
	smpd->ack_timer_handle = -1;
}

void
scsi_start_exch_timer(scsi_mp_desc_t * smpd)
{
	int32		rc;

	smpd->exch_timer_handle = eee_timerCreate(SCSI_EXCH_TIMEOUT, EEE_TIMER_ONCE);
	VERIFY(smpd->exch_timer_handle >= 0);
	rc = eee_timerStart(smpd->exch_timer_handle, (void *)scsi_exch_timeout_handler,
						smpd, NULL, NULL, NULL);
	VERIFY(rc == NFX_OK);
}

void
scsi_stop_exch_timer(scsi_mp_desc_t * smpd)
{
	int32		rc;

	rc = eee_timerStop(smpd->exch_timer_handle);
	VERIFY(rc == NFX_OK);
	rc = eee_timerDestroy(smpd->exch_timer_handle);
	VERIFY(rc == NFX_OK);
	smpd->exch_timer_handle = -1;
}

/*
 * SCSI message queue timer functions
 */ 

// create msg queue timer
void
scsi_msg_queue_timer_create(int interval, scsi_timer_func_t t_func)
{
	scsiCtrl->sscMsgs.q_timer_handle = eee_timerCreate(SCSI_MSG_QUEUE_TIMER_PERIOD, EEE_TIMER_ONCE);
	VERIFY(scsiCtrl->sscMsgs.q_timer_handle != TIMER_ERR_NO_TIMER_AVAILABLE);
	scsiCtrl->sscMsgs.q_timer_running = 0;
	scsiCtrl->sscMsgs.q_timer_handler = t_func;
}

// destroy msg queue timer
void
scsi_msg_queue_timer_destroy()
{
	int32		rc;

	if (scsiCtrl->sscMsgs.q_timer_running)
		scsi_msg_queue_timer_stop();

	rc = eee_timerDestroy(scsiCtrl->sscMsgs.q_timer_handle);
	VERIFY(rc == NFX_OK);
	scsiCtrl->sscMsgs.q_timer_handle = -1;
}

// stop msg queue timer
void
scsi_msg_queue_timer_stop()
{
	int rc;

	rc = eee_timerStop(scsiCtrl->sscMsgs.q_timer_handle);
	VERIFY(rc == NFX_OK);
	scsiCtrl->sscMsgs.q_timer_running = 0;
}

void
scsi_msg_queue_timer_func_call()
{
	scsiCtrl->sscMsgs.q_timer_handler(scsiCtrl->sscMsgs.q_timer_handle, NULL, NULL, NULL, NULL);
}

// restart msg queue timer
void
scsi_msg_queue_timer_restart()
{
	int32		rc;

	if (scsiCtrl->sscMsgs.q_timer_running) {
		rc = eee_timerStop(scsiCtrl->sscMsgs.q_timer_handle);
		VERIFY(rc == NFX_OK);
	}
	rc = eee_timerStart(scsiCtrl->sscMsgs.q_timer_handle, (void *)scsiCtrl->sscMsgs.q_timer_handler, NULL, NULL, NULL, NULL);
	VERIFY(rc == NFX_OK);
	scsiCtrl->sscMsgs.q_timer_running = 1;
}

// sdm poll timer handler function
int32
scsi_sdm_poll_handler(int32 timer_handle, void *n1, void *n2, void *n3, void *n4)
{
	VERIFY(timer_handle == scsiCtrl->sscMsgs.q_timer_handle);

	// send null msg to sdm
	scsi_sendNullMsg();

	// restart timer
	scsiCtrl->sscMsgs.q_timer_running = 0;
	scsi_msg_queue_timer_restart();

	return NFX_OK;
}

// msg queue timer handler function
int32 
scsi_msg_q_timer_handler(int32 timer_handle, void *n1, void *n2, void *n3, void *n4)
{
	int rc;
	scsi_msg_queue_t *q_msg;
	
	VERIFY(timer_handle == scsiCtrl->sscMsgs.q_timer_handle);

	// send messages until there are no more messages to send or the receiver
	// does not have buffer space
	while(1) {
		q_msg = scsi_msg_queue_dequeue();
		if(!q_msg) break;
		
		// check if receiver has sufficient space in its buffer
		if(SCSI_EEE_CHAR_SPACE(q_msg->num_bytes) >= scsiCtrl->sscMsgs.char_space ||
			SCSI_EEE_MBUF_SPACE(q_msg->num_bytes) >= scsiCtrl->sscMsgs.mbuf_space) {
				scsi_msg_queue_enqueue_hd(q_msg);	// no space available, enqueue msg to head of queue	
				scsiCtrl->sscMsgs.no_space_cnt++;
				//printf("DEBUG: No buffer space available to send messages to SSC\n");
				if(scsiCtrl->sscMsgs.no_space_cnt >= SCSI_MAX_SSC_NO_SPACE_INTERVALS) {
					scsiCtrl->sscMsgs.q_timer_running = 0;
					printf("No buffer space available to send messages to SSC...SDM/SSC probably down\n");
					scsi_mark_sdm_down();
					return NFX_OK;
				}
				break;	
		}
		scsiCtrl->sscMsgs.no_space_cnt = 0;

		// send the message
		rc = eee_sendMessage((void *)q_msg->msg, q_msg->num_bytes, q_msg->dest_port, q_msg->source_port);	
		if(!rc) {
			scsiCtrl->stats.sscMsgSentCnt++;
			scsiCtrl->sscMsgs.char_space -= SCSI_EEE_CHAR_SPACE(q_msg->num_bytes);
			scsiCtrl->sscMsgs.mbuf_space -= SCSI_EEE_MBUF_SPACE(q_msg->num_bytes);
			//printf("SSC msg send : char_space = %d mbuf_space = %d datalen=%d\n", 
			//		scsiCtrl->sscMsgs.char_space, scsiCtrl->sscMsgs.mbuf_space, 
			//		(q_msg->num_bytes > 28) ? q_msg->num_bytes : 28);
		} 
		else {
			printf("eee_sendMessage failed rc=%d datalen=%d\n", rc, q_msg->num_bytes);
		}

		// free message
		eee_ramDealloc(q_msg->msg);
		eee_ramDealloc(q_msg);
	}

	// restart timer
	scsiCtrl->sscMsgs.q_timer_running = 0;
	scsi_msg_queue_timer_restart();

	return NFX_OK;
}

// queues a msg to head of queue
void
scsi_msg_queue_enqueue_hd(scsi_msg_queue_t *q_msg)
{
	VERIFY(q_msg);

	if(!scsiCtrl->sscMsgs.q_hd) {
		scsiCtrl->sscMsgs.q_hd = scsiCtrl->sscMsgs.q_tl = q_msg;
	}
	else {
		q_msg->next = scsiCtrl->sscMsgs.q_hd;
		scsiCtrl->sscMsgs.q_hd = q_msg;
	}
	scsiCtrl->stats.sscMsgQueueSize++;
}

// queue a msg to tail of message queue
void
scsi_msg_queue_enqueue(scsi_msg_queue_t *q_msg)
{
	VERIFY(q_msg);

	if(!scsiCtrl->sscMsgs.q_hd) {
		scsiCtrl->sscMsgs.q_hd = scsiCtrl->sscMsgs.q_tl = q_msg;
	}
	else {
		scsiCtrl->sscMsgs.q_tl->next = q_msg;
		scsiCtrl->sscMsgs.q_tl = q_msg;
	}
	scsiCtrl->stats.sscMsgQueueSize++;
}

// dequeue a msg from head of message queue
scsi_msg_queue_t* 
scsi_msg_queue_dequeue()
{
	scsi_msg_queue_t *q_msg = NULL;
	
	if(scsiCtrl->sscMsgs.q_hd) {
		q_msg = scsiCtrl->sscMsgs.q_hd;

		VERIFY(q_msg->msg);

		scsiCtrl->sscMsgs.q_hd = scsiCtrl->sscMsgs.q_hd->next;
		scsiCtrl->stats.sscMsgQueueSize--;
	}

	return q_msg;
}

// marks SSC as down, cleans all queues and starts sdm polling 
void
scsi_mark_sdm_down()
{
	scsi_msg_queue_t *q_msg;

	// stop data send timer
	scsi_msg_queue_timer_destroy();
	
	// reset variables
	scsiCtrl->sscMsgs.sdm_id = 0;
	scsiCtrl->sscMsgs.no_space_cnt = 0;
	scsiCtrl->sscMsgs.char_space = 0;
	scsiCtrl->sscMsgs.mbuf_space = 0;

	// empty out the message queue
	while((q_msg = scsi_msg_queue_dequeue())) {
		eee_ramDealloc(q_msg->msg);
		eee_ramDealloc(q_msg);
	}

	//-- Create and start SSC msg timer
	scsi_msg_queue_timer_create(SCSI_SDM_POLL_TIMER_PERIOD, scsi_sdm_poll_handler);
	scsi_msg_queue_timer_restart();
}
