/*  File Name  :	scsi.h
 *
 *  Description:	Contains defines, macros and structures for managing 
 *					SCSI module. 
 *
 *  Created    :    22 February 2001
 *
 *  Author	   :    Frank Salzmann
 *
 *  Copyright(c)  Agile Storage, Inc. 2001
 *
 */
#ifndef SCSI_H_INCLUDED
#define SCSI_H_INCLUDED


#include <nfx-incl.h>
#include <exports/eee-api.h>
#include <exports/cm-api.h>
#include <exports/timer-api.h>
#include <exports/malloc-api.h>
#include <exports/libc.h>
#include <exports/esm-api.h>
#include <exports/cod-api.h>
#include <exports/debug.h>
#include <exports/elog-api.h>

#define SDM_DEV_INIT	1
#define SCSI_TIMESTAMP	1

typedef struct _gmc_msg_hdr {
	uint32 		cmd;
	nfx_cod_t 	cod;
} gmcMsgHdr_t;

typedef struct _scsi_cod {
	uint32 hi;
	uint32 lo;
} scsi_cod_t;

typedef struct _emac_ {
	uchar8 mac[6];
} eMac_t;

#include "scsi-api.h"
#include "scsi-queue.h"
#include "scsi-dev-cmds.h"
#include "scsi-dev.h"
#include "scsi-err.h"
#include "scsi-debug.h"

/* #define SCSI_TEST */

#ifdef SCSI_TEST
#include "scsi-test.h"
#endif

#include "devO_sm.h"

#define SCSI_TODO	0

#define SCSI_5SEC_TMR		5000
#define SCSI_CMD_MAX_RETRY	3

/*
 * Number of SCSI Descriptors allocated.
 */
#define SCSI_DESC_COUNT		4096		

/*
 * Number of SCSI-Multi-Packet Descriptors allocated.
 */
#define SCSI_MP_DESC_COUNT	1024

/*
 * SCSI Debug Printf Macro
 */
#define SCSI_FUNC_TRACE(_str_)  if(scsi_func_trace)	printf(_str_)	

#define SCSI_ACK_TIMEOUT	(5 * 1000)
#define SCSI_EXCH_TIMEOUT	(3 * 1000) /* Must be smaller than ACK_TIMEOUT */
/*
 * SCSI Module Control Block
 */
//-- SCSI message queue
typedef struct _scsi_msg_queue {
	uint32	num_bytes;
	uint32	dest_port;
	uint32	source_port;
	struct _scsi_msg_queue *next;
	uchar8*	msg;
} scsi_msg_queue_t;

typedef int32	(*scsi_timer_func_t)(int32, void *, void *, void *, void *);

// Details about messages sent to SSC
typedef struct _scsi_ssc_msgs {
	int32				q_timer_handle;		/* handle to msg queue timer */
	int32				q_timer_running;	/* flag to indicate running timer */
	scsi_msg_queue_t*	q_hd;				/* head of msg queue */
	scsi_msg_queue_t*	q_tl;				/* tail of msg queue */
	scsi_timer_func_t	q_timer_handler;	/* timer callback function */
	int32				sdm_id;				/* sdm id, 0 if sdm is down */
	uint32				no_space_cnt;		/* number of timer intervals msg could be sent 
											   because of no buffer space */
	uint32				char_space;			/* bytes available */
	uint32				mbuf_space;			/* mbuf bytes available */	
} scsi_ssc_msgs_t;

// SCSI Tape message queue
typedef struct _scsi_tape_io_q {
	uint64	iocookie;
	void*	msg;
	struct _scsi_tape_io_q *next;
} scsi_tape_io_q_t;

typedef struct _scsi_ctrl
{
	uint32				scsiDataAddr;
	int32				appId;							/* SCSI application id		  */
	int32				nDevices;						/* Num active devices		  */
	uint32				descCnt;		          		/* Num of available SCSI desc */
	scsi_desc_t **		descStk;					   	/* Ptr to SCSI desc stack     */
	scsi_desc_t	*		descArray[SCSI_DESC_COUNT];   	/* Stack of SCSI desc         */
	uint32				mpDescCnt;		          		/* Num of available SCSI desc */
	scsi_mp_desc_t **	mpDescStk;
	scsi_mp_desc_t *	mpDescArray[SCSI_MP_DESC_COUNT];
	void *				writePadBuffer;
	scsi_stats_t		stats;
	int32				statsTmr;
	scsi_ssc_msgs_t		sscMsgs;						/* details abt. msgs being sent to SSC */
} scsi_ctrl_t;

#define SCSI_MAX_HISTORY	1024

//-- SCSI Req/Rsp History 
typedef struct _scsi_history {
	scsi_id_t id;
	uint32    op;
	uint32    lba;
	uint32    lbCnt;
	uint32    cHandle;
} scsi_history_t;

#define SCSI_EEE_CHAR_SPACE(x)	(((x) >= 28) ? ((x)+20) : (28+20))
#define SCSI_EEE_MBUF_SPACE(x)	(((x)/1024 + 1) * 2048 + ((x)/1024 + 3) * 128)

#define SCSI_SDM_POLL_TIMER_PERIOD		1000	/* in millisecs */
#define SCSI_MSG_QUEUE_TIMER_PERIOD		100		/* in millisecs */	
#define SCSI_MAX_SSC_NO_SPACE_INTERVALS	150		/* 150*100 i.e. 15 secs */

/*
 * External Variable Declarations
 */
extern uint32 scsi_func_trace;
extern uint32 scsi_pkt_trace;
extern scsi_ctrl_t *scsiCtrl;
extern nfx_cod_t	scsiCodHandle;
extern int scsiErrLevelCtrl[SCSI_MAX_ERR_LEVEL];
extern int scsiDebugLevelCtrl[SCSI_MAX_ERR_LEVEL];
extern scsi_dev_t *scsiDevList[SCSI_MAX_DEV_PER_PORT];
extern scsi_cmd_t scsi_diskCmds[];
extern scsi_cmd_t scsi_tapeCmds[];
extern scsi_history_t  scsiHistory[SCSI_MAX_HISTORY];
extern uint32 scsiHistoryCurEntry;
extern uint32 scsi_debug_drop;
extern uint32 scsi_history_enable;


/*
 * Function Protypes
 */
void scsi_receiveMsg(eee_desc_t *);
scsi_dev_t *scsi_getDevice(scsi_path_t);
scsi_dev_t *scsi_allocDevice(uint32, ushort16);


/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_addToReqQ
 *
 * Description: 	Places command descriptor on device request queue.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 24, 2001
 *					
 *------------------------------------------------------------------------------
 */
extern int
scsi_addToReqQ(scsi_desc_t *sd, scsi_dev_t *pdev);

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_sendNextCmd
 *
 * Description: 	Sends the next command waiting on the command queue.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	April 24, 2001
 *					
 *------------------------------------------------------------------------------
 */
extern int
scsi_sendNextCmd(scsi_dev_t *pDev);

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_getCodHandle
 *
 * Description: 	Allocates a COD handle.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	June 06, 2001
 *					
 *------------------------------------------------------------------------------
 */
extern int
scsi_getCodHandle(uint64 *codHandle);

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_deallocDevice
 *
 * Description: 	Deallocates a device control block.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	May 31, 2001
 *					
 * Returns: 
 *
 *------------------------------------------------------------------------------
 */
extern int
scsi_deallocDevice( scsi_dev_t *pdev );

/*
 *------------------------------------------------------------------------------
 * Function Name:	scsi_flushReqQ
 *
 * Description: 	Flushs all of the pending commands on a devices request
 * 					queue and sends a reject command back to the originator.
 * 				
 * Created by:		Frank Salzmann
 *
 * Data Created:	June 11, 2001
 *					
 *------------------------------------------------------------------------------
 */
extern int
scsi_flushReqQ(scsi_dev_t *pdev);

scsi_desc_t *scsi_allocDesc(void);
void scsi_deallocDesc(scsi_desc_t *);
void scsi_displayDesc(scsi_desc_t *);
scsi_mp_desc_t *scsi_allocMPDesc(void);
void scsi_deallocMPDesc(scsi_mp_desc_t *);
scsi_mp_desc_t *scsi_MPEntryNew(scsi_dev_t *, uint32);
scsi_mp_desc_t *scsi_MPEntryLookup(scsi_dev_t *, uint32);
void scsi_MPEntryDestroy(scsi_mp_desc_t *);
void scsi_destroyDescChain(eee_desc_t *eDesc);
ushort16 scsi_getTransactionId(void);
scsi_dev_t *scsi_devCheckExist( uint64 devWWN, uint32 lun, uint32 llHandle );
int scsi_newDevice(uint32, ushort16, uint64, uint64, char *, uint32, char *);
int scsi_genIOQueueTimerHandler(scsi_dev_t *);
int scsi_scsi_getDeviceMaxTags(scsi_dev_t *);

int scsi_sendCmdErrRsp(int32, uint32, uint32);
int scsi_sendCmdAbortedRsp(int32, uint32, uint32, uint32);
int scsi_sendAck(int32, uint32, int);
void   scsi_insert_embedded_checksum(scsi_dev_t *, scsi_desc_t *, uint32);
int    scsi_verify_embedded_checksum(scsi_dev_t *, scsi_desc_t *);
void   scsi_insert_block_checksum(uint32, uint32, uint32, eee_desc_t *);
void   scsi_extract_block_checksum(uint32, uint32, eee_desc_t *, uint32 *, uint32 *);
uint32 scsi_compute_checksum(eee_desc_t *);
uint32 scsi_add_checksums(uint32, uint32);
int    scsi_map_logical_to_physical(uint32, uint32, uint32 *, uint32 *);
int    scsi_map_physical_to_logical(uint32, uint32, uint32 *, uint32 *);
void   scsi_add_padding(eee_desc_t *, uint32);

int    scsi_write_gate_timer_handler(int32, scsi_dev_t *);
void   scsi_write_gate_timer_create(scsi_dev_t *);
void   scsi_write_gate_timer_destroy(scsi_dev_t *);
void   scsi_write_gate_timer_restart(scsi_dev_t *);
void   scsi_write_gate_timer_stop(scsi_dev_t *);


int scsi_diskInitCmds();
int scsi_tapeInitCmds();

uint32 scsi_swap32(uint32 val);
uint64 scsi_swap64(uint64 val);
 
int scsi_procRelayReq(eee_desc_t *, scsi_msgHdr_t *);
int scsi_procReadReq(eee_desc_t *, scsi_msgHdr_t *);
int scsi_procWriteReq(eee_desc_t *, scsi_msgHdr_t *);
int scsi_procAck(eee_desc_t *, scsi_msgHdr_t *);
int scsi_procOpenMsg(eee_desc_t *, scsi_openReqMsg_t *);
int scsi_procCloseMsg(eee_desc_t *, scsi_msgHdr_t *);
int scsi_procSDMAck(eee_desc_t *, scsi_msgHdr_t *);
int scsi_procSyncMsg(eee_desc_t *, scsi_msgHdr_t *);
int scsi_procDevResetMsg(eee_desc_t *, scsi_msgHdr_t *);
int scsi_procDevStatusMsg(eee_desc_t *, scsi_msgHdr_t *);
int scsi_procIoctlMsg(eee_desc_t *, scsi_msgHdr_t *);
int scsi_procDevStatsMsg(eee_desc_t *, scsi_msgHdr_t *);
int scsi_procDelMsg(eee_desc_t *, scsi_msgHdr_t *);
int scsi_procAddMsg(eee_desc_t *, scsi_addReqMsg_t *);
int scsi_procDevControlMsg(eee_desc_t *, scsi_msgHdr_t *);
int scsi_sendCmdErrRsp(int32, uint32, uint32);
int scsi_procDelDevMsg(eee_desc_t *, scsi_msgHdr_t *);
int scsi_deallocDevice( scsi_dev_t *pdev );
int scsi_sendSdmDelAllMsg();
int scsi_sendRemDevMsg(scsi_dev_t *pdev);
int scsi_showDevice( scsi_dev_t *pdev );
int scsi_openDevice(scsi_path_t path);
int scsi_sendOpenRsp( uint32 cHandle, uint32 eCode, uint32 dest );
int scsi_sendCloseRsp( uint32 cHandle, uint32 eCode, uint32 dest );
int scsi_sendNewDevMsg(scsi_dev_t *pdev);
int scsi_sendAddDevMsg(scsi_dev_t *pdev);
int scsi_sendReadRspWithData( scsi_desc_t *sd, uint32 eCode, uint32 csum, uint32 transSeqId, uint64 *dataOffset );
int scsi_parseInqRsp( scsi_dev_t *pdev, scsi_desc_t *sd );
int scsi_newDevInit(uint32 tType,ushort16 llHandle,uint64 devWWN,uint64 chnWWN, char *chnName, uint32 lun);
int scsi_deleteDevicell(uint32 tType,ushort16 llHandle,uint64 devWWN,uint32 lun); 
int32 scsi_pathErr(ushort16 llHandle, uint64 devWWN);
int scsi_sendRebalanceMsg();
int scsi_openModeProc(scsi_dev_t *pdev, scsi_desc_t *sd);
int scsi_devInitModeProc(scsi_dev_t *pdev, scsi_desc_t *sd);
int scsi_dispatchCmd( scsi_desc_t *sd, scsi_dev_t *pDev );  
int scsi_sendInquiry(scsi_dev_t *pdev, scsi_desc_t *sd);
int scsi_sendToSscApp( uchar8 *msg, uint32 msgSize, char *app);
int scsi_sendWriteRsp( scsi_desc_t *sd, uint32 eCode );
int scsi_sendPathErrMsg(scsi_dev_t * pdev);
int scsi_addGmcHdr(eee_desc_t *edesc, nfx_cod_t cod);
int scsi_ui_clearStats();
int scsi_addToReqQ(scsi_desc_t *sd, scsi_dev_t *pdev);
int scsi_sendNextCmd(scsi_dev_t *pDev);
int scsi_getCodHandle(uint64 *codHandle);
int scsi_sendReadData( scsi_dev_t *pDev, scsi_desc_t *sd, uint64 cod );
int scsi_sendReadRsp( scsi_desc_t *sd, uint32 eCode, uint32 csum, uint64 codHandle);
int scsi_sendRelayRsp( scsi_desc_t *sd );
int scsi_flushReqQ(scsi_dev_t *pdev);
int scsi_procClearStatsMsg(eee_desc_t *, scsi_msgHdr_t *);
int scsi_procGetStatsMsg(eee_desc_t *, scsi_msgHdr_t *);
int scsi_procWindowUpdateMsg(eee_desc_t *);
int scsi_sendNullMsg(void);
int scsi_procNullMsgRsp(eee_desc_t *, scsi_msgHdr_t *); 
int scsi_procHugeioReq(eee_desc_t *, scsi_msgHdr_t *);
eee_desc_t *scsi_copyEEEDesc(eee_desc_t *);
int32 scsi_ack_timeout_handler(int32, scsi_mp_desc_t *);
int32 scsi_exch_timeout_handler(int32, scsi_mp_desc_t *);
void scsi_start_ack_timer(scsi_mp_desc_t *);
void scsi_stop_ack_timer(scsi_mp_desc_t *);
void scsi_start_exch_timer(scsi_mp_desc_t *);
void scsi_stop_exch_timer(scsi_mp_desc_t *);
int	scsi_start_huge_io(scsi_dev_t *, uint32, uint32, uint32, int, scsi_hugeio_callback_t, void *);
int	scsi_stop_huge_io(scsi_dev_t *);
int scsi_huge_io_done(scsi_desc_t *, scsi_dev_t *, int);
int scsi_huge_io_done_ssc(scsi_desc_t *, scsi_dev_t *, int);
int scsi_do_write_verify(scsi_dev_t *, uint32, uint32);
int32 scsi_sendMessage(uchar8 *, uint32, uint32, uint32);
void scsi_mark_sdm_down( void );
void scsi_msg_queue_timer_create();
void scsi_msg_queue_timer_destroy();
void scsi_msg_queue_timer_stop();
void scsi_msg_queue_timer_func_call();
void scsi_msg_queue_timer_restart();
void scsi_msg_queue_enqueue(scsi_msg_queue_t *);
void scsi_msg_queue_enqueue_hd(scsi_msg_queue_t *);
scsi_msg_queue_t* scsi_msg_queue_dequeue();
int32 scsi_msg_q_timer_handler(int32, void *, void *, void *, void *);
int32 scsi_sdm_poll_handler(int32, void *, void *, void *, void *);
scsi_dev_t* scsi_ui_findDevice(uint64 wwn, uint32 lun);

int dp83820_printBuf(char *buf, int len, int fDecode); 

#endif  /* SCSI_H_INCLUDED */
