/* vi:ts=4
 *------------------------------------------------------------------------------
 *
 * Name: storage.c
 *
 * RCS:  $Id: storage.c,v 1.68 2002/11/23 00:26:17 nitinb Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:  Provides user-interface funtions for storage the sub-system.
 *
 * Created by:  Frank Salzmann
 *
 * Date Created: August 31, 2001
 *
 *------------------------------------------------------------------------------
 */
#include <unistd.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h>
#include <errno.h>

#include <nfx-incl.h>
#include <nfx-types.h>
#include <exports/eee-api.h>
#include <exports/timer-api.h>
#include <exports/anpssc-api.h>
#include <exports/scsi-api.h>
#include <exports/scsi-msg.h>
#include <exports/hal.h>
#include <exports/raid-api.h>
#include "../sm-scsi/scsi-dev-cmds.h"
#include "sdm-api.h"
#include "sdm-db.h"
#include "sdm.h"
#include "sdm-conf.h"

#define MAX_DISPLAY_LINE 20
#define MAX_ARG_LEN	32
#define LEAF	0
#define BRANCH	1

#define RAID_UI_EXEC	"/usr/local/agile/bin/raid-ui"

typedef struct ui_node {
	char *name;
	int   type;				// UI_BRANCH_NODE | UI_LEAF_NODE
	int   arg_cnt;			// Minimum argument count for leaf node
	struct ui_node *branch;
	int (*cmd_func)();
	char *usage;
} ui_node_t;

// User-Interface Command Function Prototypes
int sdm_ui_usage(char *, ui_node_t *);
int sdm_ui_parse(char *, ui_node_t *, int, char **);
int sdm_ui_getDevInfo(uint64, sdm_devInfoRspMsg_t *);
int sdm_ui_devClose(int, char **);
int sdm_ui_devName(int, char **);
int sdm_ui_devOpen(int, char **);
int sdm_ui_devClust(int, char **);
int sdm_ui_devSorting(int, char **);
int sdm_ui_devCSVerify(int, char **);
int sdm_ui_viewList(int, char **);
int sdm_ui_viewIds(int, char **);
int sdm_ui_viewDetail(int, char **);
int sdm_ui_viewDevStats(int, char **);
int sdm_ui_viewDevPath(int, char **);
int sdm_ui_viewPortStats(int, char **);
int sdm_ui_viewPortDetail(int, char **);
int sdm_ui_viewPortTopology(int, char **);
int sdm_ui_scsiTest(int, char **);
int sdm_ui_scsiInquiry(int, char **);
int sdm_ui_scsiModeSense(int, char **);
int sdm_ui_scsiModeSelect(int, char **);
int sdm_ui_scsiFormat(int, char **);
int sdm_ui_scsiResvIn(int , char **);
int sdm_ui_scsiResvOut(int , char **);
int sdm_ui_controlDevReset(int, char **);
int sdm_ui_controlPortReset(int, char **);
int sdm_ui_controlPortDisable(int, char **);
int sdm_ui_controlPortEnable(int, char **);
int sdm_ui_controlPortRescan(int, char **);
int sdm_ui_controlPortTrace(int, char **);
int sdm_ui_controlPortBypass(int, char **);
int sdm_ui_controlPortUnBypass(int, char **);
int sdm_ui_controlPortLip(int, char **);
int sdm_ui_getDevStatus(unsigned long long, sdm_devStatusRspMsg_t *);
uint32 sdm_ui_getSdmHandle(int64, int32, char *);
scsi_id_t sdm_ui_getSCSIId(int64, int32, char *);
char *sdm_ui_portTypeToStr(uint32);
int sdm_ui_portIsAlive(int32, int32);
int sdm_ui_devDiscover(int, char **);
int sdm_ui_parsePortString(char *, uint32 *, uint32 *);
int sdm_strToLower( char *, char *);
int sdm_ui_lunMaskShow(int, char **);
int sdm_ui_clearStats(int, uint32, uint32, char *);
int sdm_ui_clearDevStats(int, char **);
int sdm_ui_clearScsiStats(int, char **);
int sdm_ui_clearFcStats(int, char **);
int sdm_ui_clearAllStats(int, char **);
int sdm_ui_viewScsiStats(int, char **);
int sdm_ui_getStats(sdm_getStatsReqMsg_t *, char *, int);
void sdm_ui_printDevIoStats(int32, sdm_getStatsRspMsg_t *);
void sdm_ui_printDevByteStats(int32, sdm_getStatsRspMsg_t *);
int sdm_ui_compare_paths(const void *, const void *);
int sdm_ui_hugeio(int, char **);
int sdm_ui_diskCheckError(int, char **);
int sdm_ui_mylexDiskCheckError(char *);
int sdm_ui_mylexGetDiskList(char *, raid_phyDevListRspMsg_t *, int);

//-- View Port Commands
ui_node_t sdm_ui_viewPort[] = {
  {"detail",	 LEAF, 1,0,sdm_ui_viewPortDetail, "view port detail <port identifier>" },
  {"stats",		 LEAF, 1,0,sdm_ui_viewPortStats,  "view port stats <port identifier>" },
  {"topology",	 LEAF, 1,0,sdm_ui_viewPortTopology,  "view port topology <port identifier>" },
  {0, 0, 0, 0, 0, 0}
};
//  {"detail",     LEAF,  1,  0, sdm_ui_viewDetail,  "view dev detail <wwn> <lun>"},

// View Device Commands
ui_node_t sdm_ui_viewDev[] = {
  {"detail",     LEAF,  1,  0, sdm_ui_viewDetail,  "view dev detail <name>"},
  {"path",		 LEAF,  1,  0, sdm_ui_viewDevPath, "view dev path <devname>"  },
  {"stats",		 LEAF,  1,  0, sdm_ui_viewDevStats,"view dev stats <devname> | <port-identifier>" },
  {0, 0, 0, 0, 0, 0}
};

// View Commands
ui_node_t sdm_ui_view[] = {
  {"dev",		 BRANCH,2,  sdm_ui_viewDev,  0, 0},
  {"list",		 LEAF,  1,  0, sdm_ui_viewList,  "view list (disk|tape|all) [-p PORT] [-v VENDOR]"},
  {"ids",		 LEAF,  1,  0, sdm_ui_viewIds,   "view ids  (disk|tape|all) [-p PORT] [-v VENDOR]"},
  {"port",		 BRANCH,2,  sdm_ui_viewPort, 0, 0},
  {"scsi_stats", LEAF,	1,	0, sdm_ui_viewScsiStats, "view scsi_stats <port identifier>"		  },
  {0, 0, 0, 0, 0, 0}
};

// SCSI Commands
ui_node_t sdm_ui_scsi[] = {
  {"test",		 LEAF,  1,  0, sdm_ui_scsiTest,			"scsi test <devname>" },
  {"inquiry",    LEAF,  2,  0, sdm_ui_scsiInquiry,		"scsi inquiry <devname> <pagecode>"},
  {"modesense",  LEAF,  2,  0, sdm_ui_scsiModeSense,	"scsi modesense <devname> <pagecode>"},
  {"modeselect", LEAF,  4,  0, sdm_ui_scsiModeSelect,	"scsi modeselect <devname> <pagecode> <byteoffset> <value>"},
  {"format",     LEAF,  2,  0, sdm_ui_scsiFormat,		"scsi format <devname> <sector-size>"},
  {"resvIn",	 LEAF,  1,  0, sdm_ui_scsiResvIn,		"scsi resvIn  <devname>"}, 
  //{"resvOut",	 LEAF,  5,  0, sdm_ui_scsiResvOut,		"scsi resvOut <wwn> <lun> (reg|resv|rel|clr|preempt) <key> <serviceKey>"}, 
  {"resvOut",	 LEAF,  3,  0, sdm_ui_scsiResvOut,		"scsi resvOut <devname> (reg|resv|rel|clr|preempt) <key>"}, 
  {0, 0, 0, 0, 0, 0}
};

// Port CONTROL Commands
ui_node_t sdm_ui_controlPort[] = {
  {"reset",		 LEAF,  1,  0, sdm_ui_controlPortReset,	"control port reset <port identfier>" },
  {"enable",	 LEAF,  1,  0, sdm_ui_controlPortEnable,"control port enable <port identifier>" },
  {"disable",	 LEAF,  1,  0, sdm_ui_controlPortDisable,"control port disable <port identifier>" },
  {"rescan",	 LEAF,  1,  0, sdm_ui_controlPortRescan,"control port rescan <port identifier>" },
  {"trace",		 LEAF,	2,	0, sdm_ui_controlPortTrace,"control port trace <port identifier>> on|off|clear|get" },
  {"bypass",	 LEAF,	2,	0, sdm_ui_controlPortBypass,"control port bypass <port identifier> <dev_wwn>" },
  {"unbypass",	 LEAF,	2,	0, sdm_ui_controlPortUnBypass,"control port unbypass <port identifier> <channel_wwn>" },
  {"lip",		 LEAF,	4,	0, sdm_ui_controlPortLip,"control port lip <port identifier> <channel_wwn> <lip_byte_3> <lip_byte_4>" },
  {0, 0, 0, 0, 0, 0}
}; 

// Device CONTROL Commands
ui_node_t sdm_ui_controlDev[] = {
  {"reset",	 LEAF,  1,  0, sdm_ui_controlDevReset,	"control dev reset <devname>" },
  {0, 0, 0, 0, 0, 0}
}; 

// CONTROL Commands
ui_node_t sdm_ui_control[] = {
  {"dev", 	 BRANCH,2,  sdm_ui_controlDev,  0, 0},
  {"port", 	 BRANCH,2,  sdm_ui_controlPort, 0, 0},
  {0, 0, 0, 0, 0, 0}
};

// LUN MASKING Commands
ui_node_t sdm_ui_lunMask[] = {
  {"show",		LEAF,  0,  0, sdm_ui_lunMaskShow,		"lun mask show" },
  {0, 0, 0, 0, 0}
};

// LUN Commands
ui_node_t sdm_ui_lun[] = {
  {"mask",		  BRANCH,2, sdm_ui_lunMask, 0, 0},
  {0, 0, 0, 0, 0}
};

//
// Clear commands
ui_node_t sdm_ui_clear[] = {
  {"scsi_stats",	LEAF, 1, 0, sdm_ui_clearScsiStats, "clear scsi_stats <port identifier>" },
  {"dev_stats", 	LEAF, 0, 0, sdm_ui_clearDevStats,  "clear dev_stats [ devname ]"	},
  {"fc_stats",		LEAF, 1, 0, sdm_ui_clearFcStats,   "clear fc_stats <port identifier>"   },
  {"allstats",		LEAF, 0, 0, sdm_ui_clearAllStats,  "clear allstats"						},
  {0, 0, 0, 0, 0}
};

// Disk Check Commands
ui_node_t sdm_ui_diskCheck[] = {
  {"error",			LEAF, 1, 0, sdm_ui_diskCheckError,  "disk check error <devname>"		},
  {0, 0, 0, 0, 0}
};

// Disk Commands
ui_node_t sdm_ui_disk[] = {
  {"check",			BRANCH,2, sdm_ui_diskCheck, 0, 0},
  {0, 0, 0, 0, 0}
};

// Top Level Command Table
ui_node_t sdm_ui_root[] = {
  {"close",       LEAF,  1, 0, sdm_ui_devClose,    "close <devname>"                  },
  {"clusteredIO", LEAF,  2, 0, sdm_ui_devClust,    "clusteredIO <devname> <ena|dis>"  },
  {"sorting",     LEAF,  2, 0, sdm_ui_devSorting,  "sorting <devname> <ena|dis>"      },
  {"CSVerify",    LEAF,  2, 0, sdm_ui_devCSVerify, "CSVerify <devname> <ena|dis>"     },
  {"name",        LEAF,  2, 0, sdm_ui_devName,     "name <wwn> <name-string>"         },
  {"open",        LEAF,  1, 0, sdm_ui_devOpen,     "open <devname> [-s slot] [-p port] [-i path]" },
  {"discover",    LEAF,  1, 0, sdm_ui_devDiscover, "discover <wwn> "                  },
  {"view",   	  BRANCH,2, sdm_ui_view,    0,	0   							      },
  {"scsi",   	  BRANCH,2, sdm_ui_scsi,    0, 	0   							      },
  {"control",	  BRANCH,2,	sdm_ui_control,	0,	0   							      },
  { "lun",		  BRANCH,2, sdm_ui_lun, 0, 0										  },
  { "clear",	  BRANCH,2, sdm_ui_clear, 0, 0									 	  },	
  { "hugeio",	  LEAF,	 4, 0, sdm_ui_hugeio,		"hugeio <devname> <read|write> <lba> <size-in-KB> [count]" },
  { "disk",	  	  BRANCH,2, sdm_ui_disk, 0, 0									 	  },	
  {0, 0, 0, 0, 0, 0}
};

/* Declare globals used by eee_rtc calls.  The actual values will be returned
 * as part of the responses.
 */
uint64 count_per_usec = 1;

int
main( int argc, char *argv[] )
{
	if( argc == 1 ) {
		printf("\n");
		sdm_ui_usage(argv[0], &sdm_ui_root[0]);
		printf("\n");
		return(0);
	}
	sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
	return 0;
} // main

int
sdm_ui_usage( char *prg_name, ui_node_t *nd )
{
	ui_node_t *node;

	node = nd;
	while(node->name) {
		if( node->type == LEAF )
			printf("Usage: %s %s\n", prg_name, node->usage);
		else
			sdm_ui_usage(prg_name, node->branch);
		node = node++;
	} // while
	return 0;
} // sdm_ui_usage

int
sdm_ui_parse(char *prg_name, ui_node_t *nd, int argc, char *argv[] )
{
	int rc;
	char *tok;
	ui_node_t *node;
	
	node = nd;
	tok = argv[0];

	if( tok ) {
		while( node->name ) {
			if( !(strcmp(tok, node->name)) ) {
				if( node->type == LEAF ) {
					if( --argc >= node->arg_cnt ) {
						rc = node->cmd_func(argc, &argv[1]);
					    if( rc < 0 )
							printf("Usage: %s %s\n", prg_name, node->usage);
					}
					else
						printf("Usage: %s %s\n", prg_name, node->usage);
				}
				else {
						sdm_ui_parse(prg_name, node->branch, --argc, &argv[1]);
				}
				return 0;
			}
			else {
				node++;
			}
		} // while
	}
	printf("Command Not Found.\n");
	return 0;
} // sdm_ui_parse


/*-----------------------------------------------------------------
 * Name: bad_stat
 *
 * Description: General check of sendAgileMsg() returned status
 *
 * Created by:  Greg Seibert
 *
 * Date Created: 20 June 2001
 *
 *-----------------------------------------------------------------
 */
static int bad_stat (int stat)
{
    if (stat < 0) {
        perror ("Send to sdm_cfgd failed");
        return 1;
    }
    else if (!stat) {
        fprintf (stderr, "Timed out waiting for response\n");
        return 1;
    }
    return 0;
}

/*-----------------------------------------------------------------
 * Name: sdm_ui_devClose
 *
 * Description: Send device close message to SCSI device.
 *
 * Created by:  Frank Salzmann
 *
 * Date Created: August 31, 2001
 *
 *-----------------------------------------------------------------
 */
int
sdm_ui_devClose(int argc, char *argv[])
{
    int                 stat;
    int                 rc;
    char                respbuff[1024];
    sdm_closeIdReqMsg_t	msg;
    sdm_closeRspMsg_t	*rsp;
	scsi_id_t			id;

	if( argc != 1 )
		return -1;

	id = sdm_ui_getSCSIId(0, 0, argv[0]);

    msg.hdr.mid = SDM_MID_CLOSE_ID_REQ;
    msg.id = id;

    stat = sendAgileMsg(-1, &msg, sizeof(msg),
                        "sdm", 0, 0, 0,
                        NULL,
                        respbuff, sizeof(respbuff));

    if( bad_stat(stat) ) {
        printf("Create failed\n");
        rc = 1;
    }
    rsp = (sdm_closeRspMsg_t *)respbuff;

    if( rsp->hdr.mid != SDM_MID_DEV_CLOSE_RSP ) {
        printf("Error: Invalid Response Received: %x\n\n", rsp->hdr.mid);
        rc = 1;
    }
    else {
        if( rsp->rCode == SDM_FAIL ) {
            printf("\nDevice Close Failed: %s\n\n", argv[0]);
            rc = 1;
        }
        else {
            printf("\nDevice Closed: %s\n\n", argv[0]);
            rc = 0;
        }
    }
    return rc;

} // sdm_ui_devClose

/*-----------------------------------------------------------------
 * Name: sdm_ui_devOpen
 *
 * Description: Send device open message to SCSI device.
 *
 * Created by:  Frank Salzmann
 *
 * Date Created: August 31, 2001
 *
 *-----------------------------------------------------------------
 */
int
sdm_ui_devOpen(int argc, char *argv[])
{
    int                 stat;
    int                 rc;
    char                respbuff[1024];
    sdm_openReqMsg_t    msg;
    sdm_openRspMsg_t   *rsp;
	scsi_id_t			id;
	uint32				slot = 0;
	uint32				port = 0;
	scsi_path_t			path = 0;

	if( argc != 1 && argc != 3 && argc != 5)
		return -1;

	msg.flags = 0;
	rc = 3;
	while(rc <= argc) {
		if(!strcmp(argv[rc-2], "-s")) {
			slot = strtol(argv[rc-1], NULL, 10);
		}
		else if(!strcmp(argv[rc-2], "-p")) {
			port = strtol(argv[rc-1], NULL, 10);
			msg.flags |= SDM_OPEN_USE_PORT;
		}
		else if(!strcmp(argv[rc-2], "-i")) {
			if(argv[rc-1][0] == '0' && (argv[rc-1][1] == 'x' || argv[rc-1][1] == 'X'))
				path = strtol(&argv[rc-1][2], NULL, 16);
			else
				path = strtol(argv[rc-1], NULL, 16);
			msg.flags |= SDM_OPEN_USE_PATH;
		}
		else return -1;

		rc += 2;
	}

	// either port or path should be specified, not both ! 
	if((msg.flags & SDM_OPEN_USE_PORT) && (msg.flags & SDM_OPEN_USE_PATH))
		return -1;

	id = sdm_ui_getSCSIId(0, 0, argv[0]);

	msg.hdr.mid = SDM_MID_DEV_OPEN_REQ;
    msg.id      = id;
	msg.slot    = slot;
	msg.port	= port;
	msg.path	= path;

    stat = sendAgileMsg(-1, &msg, sizeof(msg),
                        "sdm", 0, 0, 0,
                        NULL,
                        respbuff, sizeof(respbuff));

    if( bad_stat(stat) ) {
        printf("Create failed\n");
        rc = 1;
    }
    rsp = (sdm_openRspMsg_t *)respbuff;

    if( rsp->hdr.mid != SDM_MID_DEV_OPEN_RSP ) {
        printf("Error: Invalid Response Received: %x\n", rsp->hdr.mid);
        rc = 1;
    }
    else {
        if( rsp->rCode == SDM_FAIL ) {
            printf("\nErr: Device Not Found: %s\n\n", argv[0]);
            rc = 1;
        }
		else
		if( rsp->rCode == SDM_TO_MANY_OPENS ) {
            printf("\nErr: Device Already Open: %s\n\n", argv[0]);
            rc = 1;
        }
		else
		if( rsp->rCode == SDM_INVALID_PARM ) {
			printf("\nErr: Invalid parameters passed in open to device %s\n\n", argv[0]);
			rc = 1;
		}
        else {
            printf("\nDevice Opened: %s\n\n", argv[0]);
            rc = 0;
        }
    }
	return 0;
} // sdm_ui_devOpen

/*-----------------------------------------------------------------
 * Name: sdm_ui_devName
 *
 * Description: Assigns a name to a storage device.
 *
 * Created by:  Frank Salzmann
 *
 * Date Created: August 31, 2001
 *
 *-----------------------------------------------------------------
 */
int
sdm_ui_devName(int argc, char *argv[])
{
	printf("dev name: argc: %d, argv: %s\n", argc, argv[0] );

	return 0;
} // sdm_ui_devName

/*-----------------------------------------------------------------
 * Name: sdm_ui_devClust
 *
 * Description: Enables/Disables I/O clustering on a storage device.
 *
 * Created by:  Frank Salzmann
 *
 * Date Created: August 31, 2001
 *
 *-----------------------------------------------------------------
 */
int
sdm_ui_devClust(int argc, char *argv[])
{
    sdm_controlReqMsg_t  msg;
    sdm_controlRspMsg_t  rsp;
    int                  rc;
	int					 enable;
	uint32				 sdmHandle;

#undef CMD		
#define CMD "clusteredIO"

	if (argc != 2) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return(1);
	}

	if(strcmp(argv[1], "ena") == 0)
		enable = 1;
	else
	if(strcmp(argv[1], "dis") == 0)
		enable = 0;
	else {
		printf("%s failed: Invalid parameter '%s'\n", CMD, argv[1]);
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return(1);
	}

	sdmHandle = sdm_ui_getSdmHandle(0, 0, argv[0]);
	if (sdmHandle == 0)
		return(1);
	
	bzero(&msg, sizeof(sdm_controlReqMsg_t));
	msg.hdr.mid = SDM_MID_DEV_CONTROL_REQ;
	msg.handle  = sdmHandle;
	msg.control_enable_mask = SDM_DEV_CONTROL_ENABLE_MASK_COALLESCING;
	msg.control_mask = enable ? SDM_DEV_CONTROL_ENABLE_COALLESCING : SDM_DEV_CONTROL_DISABLE_COALLESCING;
	
	rc = sendAgileMsg(-1, &msg, sizeof(sdm_controlReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  &rsp, sizeof(sdm_controlRspMsg_t));
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("%s failed: timeout\n", CMD);
		return(1);
	}

	if (rsp.hdr.mid != SDM_MID_DEV_CONTROL_RSP) {
		printf("%s failed: Invalid Response Received: %x\n", CMD, rsp.hdr.mid);
		return(1);
	}
	else 
	if (rsp.eCode != SDM_OK) {
		printf("%s failed: SDM eCode 0x%x\n", CMD, rsp.eCode);
		return(1);
	}
	return 0;
} // sdm_ui_devClust

/*-----------------------------------------------------------------
 * Name: sdm_ui_devSorting
 *
 * Description: Enables/Disables disk I/O sorting.
 *
 * Created by:  Frank Salzmann
 *
 * Date Created: August 31, 2001
 *
 *-----------------------------------------------------------------
 */
int
sdm_ui_devSorting(int argc, char *argv[])
{
    sdm_controlReqMsg_t  msg;
    sdm_controlRspMsg_t  rsp;
    int                  rc;
	int					 enable;
	uint32				 sdmHandle;

#undef CMD	
#define CMD "sorting"

	if (argc < 2) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return(1);
	}

	if(strcmp(argv[1], "ena") == 0)
		enable = 1;
	else
	if(strcmp(argv[1], "dis") == 0)
		enable = 0;
	else {
		printf("%s failed: Invalid parameter '%s'\n", CMD, argv[1]);
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return(1);
	}

	sdmHandle = sdm_ui_getSdmHandle(0, 0, argv[0]);
	if (sdmHandle == 0)
		return(1);
	
	bzero(&msg, sizeof(sdm_controlReqMsg_t));
	msg.hdr.mid = SDM_MID_DEV_CONTROL_REQ;
	msg.handle  = sdmHandle;
	msg.control_enable_mask = SDM_DEV_CONTROL_ENABLE_MASK_SORTING;
	msg.control_mask = enable ? SDM_DEV_CONTROL_ENABLE_SORTING : SDM_DEV_CONTROL_DISABLE_SORTING;
	
	rc = sendAgileMsg(-1, &msg, sizeof(sdm_controlReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  &rsp, sizeof(sdm_controlRspMsg_t));
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("%s failed: timeout\n", CMD);
		return(1);
	}

	if (rsp.hdr.mid != SDM_MID_DEV_CONTROL_RSP) {
		printf("%s failed: Invalid Response Received: %x\n", CMD, rsp.hdr.mid);
		return(1);
	}
	else 
	if (rsp.eCode != SDM_OK) {
		printf("%s failed: SDM eCode 0x%x\n", CMD, rsp.eCode);
		return(1);
	}
	return 0;
} // sdm_ui_devSorting

int 
sdm_ui_devCSVerify(int argc, char *argv[])
{
    sdm_controlReqMsg_t  msg;
    sdm_controlRspMsg_t  rsp;
    int                  rc;
	int					 enable;
	uint32				 sdmHandle;

#undef CMD		
#define CMD "CSVerify"

	if (argc < 2) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return(1);
	}

	if(strcmp(argv[1], "ena") == 0)
		enable = 1;
	else
	if(strcmp(argv[1], "dis") == 0)
		enable = 0;
	else {
		printf("%s failed: Invalid parameter '%s'\n", CMD, argv[1]);
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return(1);
	}

	sdmHandle = sdm_ui_getSdmHandle(0, 0, argv[0]);
	if (sdmHandle == 0)
		return(1);
	
	bzero(&msg, sizeof(sdm_controlReqMsg_t));
	msg.hdr.mid = SDM_MID_DEV_CONTROL_REQ;
	msg.handle  = sdmHandle;
	msg.control_enable_mask = SDM_DEV_CONTROL_ENABLE_MASK_CS_VERIFY;
	msg.control_mask = enable ? SDM_DEV_CONTROL_ENABLE_CS_VERIFY : SDM_DEV_CONTROL_DISABLE_CS_VERIFY;
	
	rc = sendAgileMsg(-1, &msg, sizeof(sdm_controlReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  &rsp, sizeof(sdm_controlRspMsg_t));
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("%s failed: timeout\n", CMD);
		return(1);
	}

	if (rsp.hdr.mid != SDM_MID_DEV_CONTROL_RSP) {
		printf("%s failed: Invalid Response Received: %x\n", CMD, rsp.hdr.mid);
		return(1);
	}
	else 
	if (rsp.eCode != SDM_OK) {
		printf("%s failed: SDM eCode 0x%x\n", CMD, rsp.eCode);
		return(1);
	}
	return 0;
} // sdm_ui_devCSVerify

char *devType_str[] = {"DISK",		"TAPE",
					   "PRINTER",	"PROCESSOR",
					   "WORM",		"CDROM",
					   "SCANNER",	"OPTICAL_DISK",
					   "CHANGER",	"COMM_DEV",
					   "PRINTER1",	"PRINTER2",
					   "RAID",		"SES",
					   "UNKNOWN",	"UNKNOWN1"
};


//-----------------------------------------------------------------
// Name: sdm_ui_viewList
//
// Description: Displays a list of attached storage devices.
//
// Created by:  Frank Salzmann
//
// Date Created: August 31, 2001
//
//     Modified: March 29, 2002:  Changed display to show 
//                                device name. Reorded fields.
//-----------------------------------------------------------------

#if SDM_DEV_INIT

int
sdm_ui_viewList(int argc, char *argv[])
{
	int                   i, j, cnt;
	int					  pathCnt, lineCnt;
    int                   stat;
    int                   rc;
    int                   num_entries;
	char                 *state;
    char                  respbuff[65536];
	char                  argBuf[MAX_ARG_LEN], c;
	sdm_devStatusRspMsg_t devInfo;
    sdm_devListReqMsg_t   msg;
    sdm_devListRspMsg_t   *rsp;
	int					  openIndex = 0;

	// printf("view list: argc: %d, argv: %s\n", argc, argv[0] );
	// if(argc != 1)
	// 	return -1;

	memset(&msg, 0, sizeof(msg));
    msg.hdr.mid = SDM_MID_DEV_LIST_REQ;

	memset(argBuf, 0, MAX_ARG_LEN);
	sdm_strToLower( argv[0], argBuf );

	if(!strcmp(argBuf, "disk")) 
		msg.type = SDM_DEVLIST_ALL_DISK;
	else
	if(!strcmp(argBuf, "tape")) 
		msg.type = SDM_DEVLIST_TAPE;
	else  
	if(!strcmp(argBuf, "all")) 
		msg.type = SDM_DEVLIST_ALL;
	else 
        return -1;
	msg.port = SDM_DEVLIST_PORTALL;

    if (argc > 1)
    {
        cnt = 3;
        if (argc >= cnt && !strcmp(argv[cnt-2], "-p"))
        {
            msg.port = strtol(argv[cnt-1], NULL, 10);
            cnt += 2;
            if (msg.port < 0 || msg.port > 3)
            {
                printf ("Port # are 0-3\n");
                return -1;
            }
        }
        if (argc >= cnt && !strcmp(argv[cnt-2], "-v"))
        {
            strcpy(msg.vendorName, argv[cnt-1]);
            cnt += 2;
        }
        if (cnt == 3)
            return -1;
    }

    stat = sendAgileMsg(-1, &msg, sizeof(msg),
                        "sdm", 0, 0, 0,
                        NULL,
                        respbuff, sizeof(respbuff));

    if( bad_stat(stat) ) {
        return -1;
    }
    rsp = (sdm_devListRspMsg_t *)respbuff;

    if( rsp->hdr.mid != SDM_MID_DEV_LIST_RSP ) {
        printf("Error: Invalid Response Received: %x\n", rsp->hdr.mid);
        rc = 1;
    }
    else {
        num_entries = rsp->numEntries;
        if( num_entries == 0 )
            printf("Device List is empty.\n");
        else {
			printf("\n");
			printf("DEV   DEV NAME                                                 TYPE       MODEL             PORT  STATE      RAID\n");
			printf("----  -------------------------------------------------------  ---------  ----------------  ----  ---------  ----\n");
            lineCnt = 2;
            for( i=0; i < num_entries; i++, lineCnt++ ) {
				sdm_ui_getDevStatus(rsp->idList[i], &devInfo);
				pathCnt = devInfo.pathCnt;
				if( devInfo.state == SDM_DEV_STATE_INIT )
					state = "INIT";
				else if(devInfo.state == SDM_DEV_STATE_INIT_OPENED)
					state = "INIT-OPEN";
				else {
					// get open path index, if device is open
					for(j=0; j < devInfo.pathCnt; j++)
						if(devInfo.paths[j].state == SDM_PATH_STATE_ACTIVE) {
							openIndex = j;
							break;
						}
					state = devInfo.paths[openIndex].state == SDM_PATH_STATE_ACTIVE ? "OPEN":"CLOSED";
				}
                printf("%-4d  %-55s  %-9s  %-16s  %-4d  %-9s  %4d\n", 
					   i+1,
					   devInfo.st.name,
					   devType_str[devInfo.devType],
					   devInfo.st.model,
					   SCSI_PATH_GET_PORT(devInfo.paths[openIndex].path),
					   state,
					   devInfo.raidLevel);
                if (lineCnt >= MAX_DISPLAY_LINE)
                {
                    printf ("[Enter or q-Quit]: ");
                    c = getchar();
                    if (c != '\n' && c != ' ')
                    {
                        getchar();
                        if (c == 'q' || c == 'Q')
                            break;
                    }
                    lineCnt = 0;
                }
            }
        }
        rc = 0;
    }
    return rc;
} // sdm_ui_viewList

#else // SDM_DEV_INIT

int
sdm_ui_viewList(int argc, char *argv[])
{
 int                 i;
    int                 stat;
    int                 rc;
    int                 num_entries;
    char                respbuff[1024];
	sdm_devInfoRspMsg_t devInfo;
    sdm_devListReqMsg_t msg;
    sdm_devListRspMsg_t *rsp;

	// printf("view list: argc: %d, argv: %s\n", argc, argv[0] );

	memset(&msg, 0, sizeof(msg));
    msg.hdr.mid = SDM_MID_DEV_LIST_REQ;
	msg.port = SDM_DEVLIST_PORTALL;

	if(!strcmp(argv[0], "disk"))
		msg.type = SDM_DEVLIST_DISK;
	else
	if(!strcmp(argv[0], "tape"))
		msg.type = SDM_DEVLIST_TAPE;
	else  
		msg.type = SDM_DEVLIST_ALL;

    stat = sendAgileMsg(-1, &msg, sizeof(msg),
                        "sdm", 0, 0, 0,
                        NULL,
                        respbuff, sizeof(respbuff));

    if( bad_stat(stat) ) {
        //printf("Create failed\n");
        exit(-1);
    }
    rsp = (sdm_devListRspMsg_t *)respbuff;


    if( rsp->hdr.mid != SDM_MID_DEV_LIST_RSP ) {
        printf("Error: Invalid Response Received: %x\n", rsp->hdr.mid);
        rc = 1;
    }
    else {
        num_entries = rsp->numEntries;
        if( num_entries == 0 )
            printf("Device List is empty.\n");
        else {
			printf("\n");
			printf("DEV   TYPE       VENDOR    MODEL             STATE\n");
			printf("----  ---------  --------  ----------------  ------\n");
            for( i=0; i < num_entries; i++ ) {
				sdm_ui_getDevInfo(rsp->idList[i], &devInfo);
                printf("%04d  %-9s  %-8s  %-16s  %s\n", 
					   i+1, 
					   devType_str[devInfo.devType],
					   devInfo.vendorName,
					   devInfo.modelName,
					   devInfo.state==SDM_DEV_STATE_CLOSED ? "CLOSED":"OPEN");
            }
        }
        rc = 0;
    }
    return rc;
} // sdm_ui_viewList
#endif // SDM_DEV_INIT

/*-----------------------------------------------------------------
 * Name: sdm_ui_viewIds
 *
 * Description: Displays a cross ref list of 
 *              WWN, LUN, SCSI_ID_T and OPEN SDM_HANDLE.
 *
 * Created by:  Frank Salzmann
 *
 * Date Created: February 13, 2001
 *
 *-----------------------------------------------------------------
 */
int
sdm_ui_viewIds(int argc, char *argv[])
{
    int                 i, cnt;
    int                 stat;
    int                 rc, lineCnt;
    int                 num_entries;
    char                respbuff[1024], c;
	sdm_devStatusRspMsg_t devStatus;
    sdm_devListReqMsg_t msg;
    sdm_devListRspMsg_t *rsp;

	// printf("view list: argc: %d, argv: %s\n", argc, argv[0] );

	memset(&msg, 0, sizeof(msg));
    msg.hdr.mid = SDM_MID_DEV_LIST_REQ;
	msg.port = SDM_DEVLIST_PORTALL;

	if(!strcmp(argv[0], "disk"))
		msg.type = SDM_DEVLIST_DISK;
	else
	if(!strcmp(argv[0], "tape"))
		msg.type = SDM_DEVLIST_TAPE;
	else  
		msg.type = SDM_DEVLIST_ALL;

    if (argc > 1)
    {
        cnt = 3;
        if (argc >= cnt && !strcmp(argv[cnt-2], "-p"))
        {
            msg.port = strtol(argv[cnt-1], NULL, 10);
            cnt += 2;
            if (msg.port < 0 || msg.port > 3)
            {
			    printf ("Port # are 0-3\n");
                return -1;
            }
        }
        if (argc >= cnt && !strcmp(argv[cnt-2], "-v"))
        {
            strcpy(msg.vendorName, argv[cnt-1]);
            cnt += 2;
        }
        if (cnt == 3)
            return -1;
    }

    stat = sendAgileMsg(-1, &msg, sizeof(msg),
                        "sdm", 0, 0, 0,
                        NULL,
                        respbuff, sizeof(respbuff));

    if( bad_stat(stat) ) {
        //printf("Create failed\n");
        exit(-1);
    }
    rsp = (sdm_devListRspMsg_t *)respbuff;


    if( rsp->hdr.mid != SDM_MID_DEV_LIST_RSP ) {
        printf("Error: Invalid Response Received: %x\n", rsp->hdr.mid);
        rc = 1;
    }
    else {
        num_entries = rsp->numEntries;
        if( num_entries == 0 )
            printf("Device List is empty.\n");
        else {
			printf("\n");
			printf("WWN                 LUN  SCSI_ID             SDM_HANDLE  VENDOR    STATE\n");
			printf("------------------  ---  ------------------  ----------  --------  -----\n");
            lineCnt = 2;
            for( i=0; i < num_entries; i++, lineCnt++ ) {
				sdm_ui_getDevStatus(rsp->idList[i], &devStatus);
                printf("0x%016llx  %03d  0x%016llx  0x%08x  %-8s  %s\n", 
					   devStatus.wwn,
					   devStatus.lun,
					   rsp->idList[i],
					   devStatus.state==SDM_DEV_STATE_CLOSED ? 0 : devStatus.handles[0],
					   devStatus.st.vendor,
					   devStatus.state==SDM_DEV_STATE_CLOSED ? "CLOSED":"OPEN");                if (lineCnt >= MAX_DISPLAY_LINE)
                {
                    printf ("[Enter or q-Quit]: ");
                    c = getchar();
                    if (c != '\n' && c != ' ')
                    {
                        getchar();
                        if (c == 'q' || c == 'Q')
                            break;
                    }
                    lineCnt = 0;
                }
            }
        }
        rc = 0;
    }
    return rc;
} // sdm_ui_viewIds

int
sdm_ui_getDevInfo(unsigned long long id, sdm_devInfoRspMsg_t *devInfo)
{
    int                 stat;
    int                 rc;
    sdm_devInfoReqMsg_t msg;

	//printf("View Device Info: %llx\n", id);

    msg.hdr.mid = SDM_MID_DEV_INFO_REQ;
    msg.id      = id;

	memset(devInfo, 0, sizeof(sdm_devInfoRspMsg_t));
    stat = sendAgileMsg(-1, &msg, sizeof(msg),
                        "sdm", 0, 0, 0,
                        NULL,
                        devInfo, sizeof(sdm_devInfoRspMsg_t));

    if( bad_stat(stat) ) {
        printf("Create failed\n");
        rc = 1;
    }

    if( devInfo->hdr.mid != SDM_MID_DEV_INFO_RSP ) {
        printf("Error: Invalid Response Received: %x\n", devInfo->hdr.mid);
        rc = 1;
    }
    else {
        if( devInfo->rc == SDM_FAIL ) {
            printf("Device Not Found: %llx\n", id);
            rc = 1;
        }
        else {
        }
    }
    return rc;
} // sdm_ui_getDevInfo

int
sdm_ui_getDevStatus(unsigned long long id, sdm_devStatusRspMsg_t *devStatus)
{
    int                 stat;
    int                 rc;
    sdm_devStatusReqMsg_t msg;


    msg.hdr.mid = SDM_MID_DEV_STATUS_REQ;
    msg.id      = id;

	memset(devStatus, 0, sizeof(sdm_devStatusRspMsg_t));
    stat = sendAgileMsg(-1, &msg, sizeof(msg),
                        "sdm", 0, 0, 0,
                        NULL,
                        devStatus, sizeof(sdm_devStatusRspMsg_t));

    if( bad_stat(stat) ) {
        printf("Create failed\n");
        rc = 1;
    }

    if( devStatus->hdr.mid != SDM_MID_DEV_STATUS_RSP ) {
        printf("Error: Invalid Response Received: %x\n", devStatus->hdr.mid);
        rc = 1;
    }
    else {
        if( devStatus->eCode == SDM_FAIL ) {
            printf("Device Not Found: %llx\n", id);
            rc = 1;
        }
        else {
        }
    }
    return rc;
} // sdm_ui_getDevStatus

/*-----------------------------------------------------------------
 * Name: sdm_ui_viewDetail
 *
 * Description: Displays dynamic information about a device.
 *
 * Created by:  Frank Salzmann
 *
 * Date Created: August 31, 2001
 *
 *-----------------------------------------------------------------
 */
static char devStatesStrings[][10] = { "DOWN", "ERROR", "CLOSED", "OPEN" };
int
sdm_ui_viewDetail(int argc, char *argv[])
{
    int                   i;
    int                   stat;
    int                   rc;
    char                  respbuff[1024];
	scsi_id_t			  id;
    sdm_devStatusReqMsg_t msg;
    sdm_devStatusRspMsg_t *rsp;

	if( argc != 1 )
		return -1;

	id = sdm_ui_getSCSIId(NULL, NULL, argv[0]);

    msg.hdr.mid = SDM_MID_DEV_STATUS_REQ;
    msg.id      = id;

    stat = sendAgileMsg(-1, &msg, sizeof(msg),
                        "sdm", 0, 0, 0,
                        NULL,
                        respbuff, sizeof(respbuff));

    if( bad_stat(stat) ) {
        printf("Create failed\n");
        rc = 1;
    }
    rsp = (sdm_devStatusRspMsg_t *)respbuff;

    if( rsp->hdr.mid != SDM_MID_DEV_STATUS_RSP ) {
        printf("Error: Invalid Response Received: %x\n", rsp->hdr.mid);
        rc = 1;
    }
    else {
        if( rsp->eCode == SDM_FAIL ) {
            printf("Device Not Found:  %s\n", argv[0]);
            rc = 1;
        }
        else {
            printf("\n");
            printf("NAME              : %s\n", rsp->st.name);
            printf("VENDOR            : %s\n", rsp->st.vendor);
            printf("MODEL             : %s\n", rsp->st.model);
			printf("PRODUCT ID        : %s\n", rsp->st.productId);
			printf("TYPE              : %s\n", devType_str[rsp->devType]);
			if (rsp->devType == SCSI_DEVTYPE_DISK) {
				uint64 size = ((uint64)rsp->blockCnt * (uint64)rsp->blockSize)/1000/1000;
				printf("BLOCK CNT         : %d\n", rsp->blockCnt);
				printf("BLOCK SIZE        : %d\n", rsp->blockSize);
				if(size > 1000) 
					printf("CAPACITY          : %lld.%lld GB\n", size/1000, size%1000);
				else
					printf("CAPACITY          : %lld MB\n", size);
			}
			else
			if (rsp->devType == SCSI_DEVTYPE_TAPE) {
				printf("POSITION          : %sblockno %d, fileno %d\n",
					   rsp->tape_position_bom ? "BOM, " :
					   rsp->tape_position_eom ? "EOM, " : 
					   "",
					   rsp->tape_position_blockno,
					   rsp->tape_position_fileno);
			}
            printf("STATE             : %s\n",
					rsp->state == SDM_DEV_STATE_CLOSED ? "CLOSED" : 
					((rsp->state == SDM_DEV_STATE_INIT) ? "INIT" : "OPEN"));
            printf("PATH COUNT        : %d\n", rsp->pathCnt);
            printf("PATH LIST         : PORT  STATE   PATH-ID      WWN                 LUN\n");

			// sort the paths by port number
			qsort((void *)rsp->paths, rsp->pathCnt, sizeof(sdm_path_t), sdm_ui_compare_paths);

            for(i=0; i < rsp->pathCnt; i++)
                printf("%-19s %-4d  %-6s  0x%-9x  0x%-16llx  %d\n", 
						" ", SCSI_PATH_GET_PORT(rsp->paths[i].path), 
						devStatesStrings[rsp->paths[i].state],
						rsp->paths[i].path,
						rsp->paths[i].wwn, rsp->paths[i].lun);
            printf("\n");
            printf("OPEN HANDLE COUNT : %d\n", rsp->handleCnt);
            printf("OPEN HANDLE LIST  :");
            for(i=0; i < rsp->handleCnt; i++)
                printf(" 0x%08x", rsp->handles[i]);
            printf("\n");

            rc = 0;
        }
    }
    return rc;
} // sdm_ui_viewDetail

int
sdm_ui_getStats(sdm_getStatsReqMsg_t *msg, char *respbuff, int bufsize)
{
    int                   stat;
    sdm_getStatsRspMsg_t *rsp;

    stat = sendAgileMsg(-1, msg, sizeof(sdm_getStatsReqMsg_t),
                        "sdm", 0, 0, 0,
                        NULL,
                        respbuff, bufsize);

    if( bad_stat(stat) ) {
        printf("Socket Create failed\n");
		return 1;
    }

    rsp = (sdm_getStatsRspMsg_t *)respbuff;

    if( rsp->hdr.mid != SDM_MID_GET_STATS_RSP ) {
        printf("\nError: Invalid Response Received: %x\n\n", rsp->hdr.mid);
		return 1;
    }
    else 
    if( rsp->eCode == SDM_FAIL ) {
		return 1;
	}
    else 
    if( rsp->eCode == SDM_NOT_OPEN ) {
		printf("\nDevice Not Open: %s\n\n", msg->devname);
		return 1;
	}

	return 0;
}

void
sdm_ui_printDevIoStats(int32 port, sdm_getStatsRspMsg_t *rsp)
{
	if(port >= 0)
		printf("%-4d ", port);
	printf("0x%016llx:%03d  %6u  %6u  %6u  %6u  %6u  %6u  %8u  %8u  %8u\n", 
					   rsp->stats.dev_stats.wwn,
					   rsp->stats.dev_stats.lun,
					   rsp->stats.dev_stats.readRspCnt,
					   rsp->stats.dev_stats.readIoPerSecCur,
					   rsp->stats.dev_stats.readIoPerSecMax,
					   rsp->stats.dev_stats.writeRspCnt,
					   rsp->stats.dev_stats.writeIoPerSecCur,
					   rsp->stats.dev_stats.writeIoPerSecMax,
					   rsp->stats.dev_stats.genio_q_len,
					   rsp->stats.dev_stats.hiprio_q_len,
					   rsp->stats.dev_stats.other_q_len);
}

void
sdm_ui_printDevByteStats(int32 port, sdm_getStatsRspMsg_t *rsp)
{
	if(port >= 0)
		printf("%-4d ", port);
	printf("0x%016llx:%03d  %16llu  %8u  %8u  %16llu  %8u  %8u\n", 
					   rsp->stats.dev_stats.wwn,
					   rsp->stats.dev_stats.lun,
					   rsp->stats.dev_stats.readByteCnt,
					   rsp->stats.dev_stats.readBytePerSecCur,
					   rsp->stats.dev_stats.readBytePerSecMax,
					   rsp->stats.dev_stats.writeByteCnt,
					   rsp->stats.dev_stats.writeBytePerSecCur,
					   rsp->stats.dev_stats.writeBytePerSecMax);
}

/*----------------------------------------------------
 * Name: sdm_ui_compare_paths
 * 
 * Description: Compares paths based on port numbers
 * 
 * Created By: Nitin Bahadur
 * 
 * Date Created: Oct 14, 2002
 * 
 *----------------------------------------------------
 */ 
int sdm_ui_compare_paths(const void *x, const void *y)
{
	scsi_path_t path1, path2;

	path1 = ((sdm_path_t *)x)->path;
	path2 = ((sdm_path_t *)y)->path;

	return (SCSI_PATH_GET_PORT(path1) - SCSI_PATH_GET_PORT(path2));
}

/*-----------------------------------------------------------------
 * Name: sdm_ui_viewDevStats
 *
 * Description: Displays device I/O statistics
 *
 * Created by:  Frank Salzmann
 *
 * Date Created: January 17, 2002
 *
 *-----------------------------------------------------------------
 */
int
sdm_ui_viewDevStats(int argc, char *argv[])
{
    int                   rc;
    sdm_getStatsReqMsg_t msg;
    char                  devlist_respbuff[65535];
    char                  respbuff[1024];
    sdm_devListReqMsg_t   devlist_msg;
    sdm_devListRspMsg_t   *devlist_rsp;
	int					  i, num_entries, stat;
	uint32				  slot, port;

	if( argc != 1)
		return 1;

	memset((void *)&msg, 0, sizeof(msg));
    msg.hdr.mid = SDM_MID_GET_STATS_REQ;
	msg.type = SDM_STATS_TYPE_DEV;

	rc = sdm_ui_parsePortString(argv[0], &slot, &port);

	// if device name was specified
	if(rc != NFX_OK) {
    	sdm_devStatusRspMsg_t status_rsp;
		scsi_id_t			  id;

		// get list of paths for this device
		id = sdm_ui_getSCSIId(NULL, NULL, argv[0]);
		if(sdm_ui_getDevStatus(id, &status_rsp) != 0)
			return 1;

		// sort the paths based on port
		qsort((void *)&status_rsp.paths, status_rsp.pathCnt, sizeof(sdm_path_t), sdm_ui_compare_paths);

		msg.subtype = SDM_STATS_SUBTYPE_USE_NAME | SDM_STATS_SUBTYPE_USE_PORT;
		strncpy(msg.devname, argv[0], sizeof(msg.devname));

		printf("\n");
		printf("                             Read    Cur Rd  Max Rd  Write   Cur Wr  Max Wr  Gen IO Q  HiPriO Q   Other Q\n"); 
		printf("PORT WWN/LUN                 Ops     Op/Sec  Op/Sec  Ops     Op/Sec  Op/Sec   Depth     Depth      Depth \n");
		printf("---- ----------------------  ------  ------  ------  ------  ------  ------  --------  --------  --------\n");
		for(i=0; i < status_rsp.pathCnt; i++) {
			msg.port = SCSI_PATH_GET_PORT(status_rsp.paths[i].path);
			rc = sdm_ui_getStats(&msg, respbuff, sizeof(respbuff));
			if(rc) continue;
			sdm_ui_printDevIoStats(msg.port, (sdm_getStatsRspMsg_t *)respbuff);
		}

		printf("\n");
		printf("                             Read              Cur Rd    Max Rd    Write             Cur Wr    Max Wr  \n");
		printf("PORT WWN/LUN                 Bytes             Byte/Sec  Byte/Sec  Bytes             Byte/Sec  Byte/Sec\n");
		printf("---- ----------------------  ----------------  --------  --------  ----------------  --------  --------\n");
		for(i=0; i < status_rsp.pathCnt; i++) {
			msg.port = SCSI_PATH_GET_PORT(status_rsp.paths[i].path);
			rc = sdm_ui_getStats(&msg, respbuff, sizeof(respbuff));
			if(rc) continue;
			sdm_ui_printDevByteStats(msg.port, (sdm_getStatsRspMsg_t *)respbuff);
		}

		return 0;
	}

	// port identifier was specified
	msg.subtype = SDM_STATS_SUBTYPE_USE_ID | SDM_STATS_SUBTYPE_USE_PORT;
	msg.port = port;

	// get list of devices
	memset(&devlist_msg, 0, sizeof(devlist_msg));
    devlist_msg.hdr.mid = SDM_MID_DEV_LIST_REQ;
	devlist_msg.type = SDM_DEVLIST_ALL;
	devlist_msg.port = port;
    stat = sendAgileMsg(-1, &devlist_msg, sizeof(devlist_msg),
                        "sdm", 0, 0, 0,
                        NULL,
                        devlist_respbuff, sizeof(devlist_respbuff));

    if( bad_stat(stat) ) {
        return -1;
    }
    devlist_rsp = (sdm_devListRspMsg_t *)devlist_respbuff;

    if( devlist_rsp->hdr.mid != SDM_MID_DEV_LIST_RSP ) {
        printf("Error: Invalid Response Received: %x\n", devlist_rsp->hdr.mid);
		return 0;
    }
	num_entries = devlist_rsp->numEntries;
	if( num_entries == 0 ) {
		printf("No entries in device list\n");
		return 0;
	}

	// for each device, get its stats
	printf("\n");
	printf("                        Read    Cur Rd  Max Rd  Write   Cur Wr  Max Wr  Gen IO Q  HiPriO Q   Other Q\n"); 
	printf("WWN/LUN                 Ops     Op/Sec  Op/Sec  Ops     Op/Sec  Op/Sec   Depth     Depth      Depth \n");
	printf("----------------------  ------  ------  ------  ------  ------  ------  --------  --------  --------\n");

    for( i=0; i < num_entries; i++ ) {
		msg.id = devlist_rsp->idList[i];
		rc = sdm_ui_getStats(&msg, respbuff, sizeof(respbuff));
		if(rc) continue;
		sdm_ui_printDevIoStats(-1, (sdm_getStatsRspMsg_t *)respbuff);
	}


	// for each device, get stats again
	printf("\n");
	printf("                        Read              Cur Rd    Max Rd    Write             Cur Wr    Max Wr  \n");
	printf("WWN/LUN                 Bytes             Byte/Sec  Byte/Sec  Bytes             Byte/Sec  Byte/Sec\n");
	printf("----------------------  ----------------  --------  --------  ----------------  --------  --------\n");

    for( i=0; i < num_entries; i++ ) {
		msg.id = devlist_rsp->idList[i];
		rc = sdm_ui_getStats(&msg, respbuff, sizeof(respbuff));
		if(rc) continue;
		sdm_ui_printDevByteStats(-1, (sdm_getStatsRspMsg_t *)respbuff);
	}

    return 0;
} // sdm_ui_viewDevStats

/*-----------------------------------------------------------------
 * Name: sdm_ui_viewDevPath
 *
 * Description: Displays physical path to the device.
 *
 * Created by:  Frank Salzmann
 *
 * Date Created: January 17, 2002
 *
 *-----------------------------------------------------------------
 */
int
sdm_ui_viewDevPath(int argc, char *argv[])
{
	printf("Command Not Implemented.\n");
	return 0;
} // sdm_ui_viewDevPath

int
sdm_ui_scsiTest(int argc, char *argv[])
{
    sdm_relayCdbReqMsg_t  msg;
    sdm_relayCdbRspMsg_t  *rsp;
    char                  respbuff[1024];
    int                   rc;
	int					  scsi_status;
	int					  i;
	uint32 				  sdmHandle;

#undef CMD
#define CMD "scsi test" 

	if( argc != 1) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return -1;
	}

	//-- Get open sdm_devHandle from the SDM.
	sdmHandle = sdm_ui_getSdmHandle(0, 0, argv[0]);
	if( sdmHandle == 0 ) 
		return 1;	

	bzero(&msg, sizeof(sdm_relayCdbReqMsg_t));
	msg.hdr.mid = SDM_MID_RELAY_CDB_REQ;
	msg.handle  = sdmHandle;
	msg.cdb[0]  = SCSI_CMD_TEST_UNIT_READY;
	
	rc = sendAgileMsg(-1, &msg, sizeof(sdm_relayCdbReqMsg_t),
						"sdm", 0, 0, 0,
						NULL,
						respbuff, sizeof(respbuff));
	if (rc < 0) {
		printf("\n%s failed: sendAgileMsg() failed\n\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("\n%s failed: timeout\n\n", CMD);
		return(1);
	}

	rsp = (sdm_relayCdbRspMsg_t *)respbuff;
	
	if (rsp->hdr.mid != SDM_MID_RELAY_CDB_RSP) {
		printf("\n%s failed: Invalid Response Received: %x\n\n", CMD, rsp->hdr.mid);
		return(1);
	}
	else 
	if (rsp->eCode != SDM_OK) {
			printf("\n%s failed: SDM eCode 0x%x\n\n", CMD, rsp->eCode);
			return(1);
	}
	else 
	if (rsp->status & SD_DRV_STATUS_MASK) {
		printf("\n%s failed: non-zero driver status: 0x%x\n\n", CMD, 
			   (rsp->status & SD_DRV_STATUS_MASK) >> SD_DRV_STATUS_SHIFT);
		return(1);
	}
	else {
		scsi_status = (rsp->status & SD_DEV_STATUS_MASK) >> SD_DEV_STATUS_SHIFT;
		if (scsi_status == SCSI_DEV_STATUS_GOOD) {
			printf("\nDevice is ready!\n\n");
			return(0);
		}
		else {
			printf("\nDevice is not ready: SCSI status 0x%x\n", scsi_status);
			if (scsi_status == SCSI_DEV_STATUS_CHECK) {
				printf("Sense : ");
				for (i = 0; i < sizeof(rsp->senseData); ++i) {
					if (i > 0 && (i % 8) == 0)
						printf("\n      : ");
					printf("%02x ", rsp->senseData[i]);
				}
				printf("\n\n");
			}
			return(0);
		}
	}
} // sdm_ui_scsiTest

int 
sdm_ui_scsiInquiry(int argc, char *argv[])
{
    sdm_relayCdbReqMsg_t  msg;
    sdm_relayCdbRspMsg_t  *rsp;
    char                  respbuff[1024];
    int                   rc;
	int					  scsi_status;
	int					  i;
	uint32 				  sdmHandle;
	int					  page_code = 0;
	char *				  end_ptr;
	
#undef CMD
#define CMD "scsi inquiry" 

	if (argc != 2) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return(1);
	}

	page_code = strtol(argv[1], &end_ptr, 0);
	if (*end_ptr != '\0') {
		printf("%s failed: Invalid page code '%s'\n", CMD, argv[1]);
		return(1);
	}
	sdmHandle = sdm_ui_getSdmHandle(0, 0, argv[0]);
	if (sdmHandle == 0)
		return(1);
	
	bzero(&msg, sizeof(sdm_relayCdbReqMsg_t));
	msg.hdr.mid = SDM_MID_RELAY_CDB_REQ;
	msg.handle  = sdmHandle;
	msg.cdb[0]  = SCSI_CMD_INQUIRY;
	msg.cdb[1]  = page_code ? 1 : 0;
	msg.cdb[2]  = page_code;
	msg.cdb[4]  = 255;
	
	rc = sendAgileMsg(-1, &msg, sizeof(sdm_relayCdbReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  respbuff, sizeof(respbuff));
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("%s failed: timeout\n", CMD);
		return(1);
	}

	rsp = (sdm_relayCdbRspMsg_t *)respbuff;
	
	if (rsp->hdr.mid != SDM_MID_RELAY_CDB_RSP) {
		printf("%s failed: Invalid Response Received: %x\n", CMD, rsp->hdr.mid);
		return(1);
	}
	else 
	if (rsp->eCode != SDM_OK) {
		printf("%s failed: SDM eCode 0x%x\n", CMD, rsp->eCode);
		return(1);
	}
	else 
	if (rsp->status & SD_DRV_STATUS_MASK) {
		printf("%s failed: non-zero driver status: 0x%x\n", CMD, 
			   (rsp->status & SD_DRV_STATUS_MASK) >> SD_DRV_STATUS_SHIFT);
		return(1);
	}
	else {
		scsi_status = (rsp->status & SD_DEV_STATUS_MASK) >> SD_DEV_STATUS_SHIFT;
		if (scsi_status == SCSI_DEV_STATUS_GOOD) {
			char *		data = &rsp->data[0];
			int			i, j;
			char		ch;

			for (i = 0; i < rsp->dataLen; i += 8) {
				printf("\n%04x:\t", i);
				for (j = i; j < (i + 8); ++j)
					printf("%02x ", data[j] & 0xff); 
				printf("    ");
				for (j = i; j < (i + 8); ++j) {
					ch = data[j] & 0xff;
					if ((ch > 0x20) && (ch < 0x7F))
						printf("%c", ch); 
					else
					if (ch == 0) 
						printf("%c",'.'); 
					else
						printf("%c",'_'); 
				}
			}
			printf("\n");

			return(0);
		}
		else {
			printf("%s failed: non-zero SCSI status 0x%x\n", CMD, scsi_status);
			if (scsi_status == SCSI_DEV_STATUS_CHECK) {
				printf("Sense : ");
				for (i = 0; i < sizeof(rsp->senseData); ++i) {
					if (i > 0 && (i % 8) == 0)
						printf("\n      : ");
					printf("%02x ", rsp->senseData[i]);
				}
				printf("\n");
			}
			return(0);
		}
	}
}

int 
sdm_ui_scsiModeSense(int argc, char *argv[])
{
    sdm_relayCdbReqMsg_t  msg;
    sdm_relayCdbRspMsg_t  *rsp;
    char                  respbuff[1024];
    int                   rc;
	int					  scsi_status;
	int					  i;
	uint32 				  sdmHandle;
	int					  page_code;
	char *				  end_ptr;
#undef CMD
#define CMD "scsi modesense"

	if (argc != 2) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return(1);
	}

	page_code = strtol(argv[1], &end_ptr, 0);
	if (*end_ptr != '\0') {
		printf("%s failed: Invalid page code '%s'\n", CMD, argv[1]);
		return(1);
	}
	sdmHandle = sdm_ui_getSdmHandle(0, 0, argv[0]);
	if (sdmHandle == 0)
		return(1);
	
	bzero(&msg, sizeof(sdm_relayCdbReqMsg_t));
	msg.hdr.mid = SDM_MID_RELAY_CDB_REQ;
	msg.handle  = sdmHandle;
	msg.cdb[0]  = SCSI_CMD_MODE_SENSE_6;
	msg.cdb[2]  = page_code;
	msg.cdb[4]  = 255;
	
	rc = sendAgileMsg(-1, &msg, sizeof(sdm_relayCdbReqMsg_t),
						"sdm", 0, 0, 0,
						NULL,
						respbuff, sizeof(respbuff));
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("%s failed: timeout\n", CMD);
		return(1);
	}

	rsp = (sdm_relayCdbRspMsg_t *)respbuff;
	
	if (rsp->hdr.mid != SDM_MID_RELAY_CDB_RSP) {
		printf("%s failed: Invalid Response Received: %x\n", CMD, rsp->hdr.mid);
		return(1);
	}
	else 
	if (rsp->eCode != SDM_OK) {
		printf("%s failed: SDM eCode 0x%x\n", CMD, rsp->eCode);
		return(1);
	}
	else 
	if (rsp->status & SD_DRV_STATUS_MASK) {
		printf("%s failed: non-zero driver status: 0x%x\n", CMD, 
			   (rsp->status & SD_DRV_STATUS_MASK) >> SD_DRV_STATUS_SHIFT);
		return(1);
	}
	else {
		scsi_status = (rsp->status & SD_DEV_STATUS_MASK) >> SD_DEV_STATUS_SHIFT;
		if (scsi_status == SCSI_DEV_STATUS_GOOD) {
			scsi_mode_header_t *		page_hdr;
			scsi_mode_block_desc_t *	block_desc;
			scsi_mode_page_desc_t *		page_data;
			char *						p;
			int							i;

			page_hdr = (scsi_mode_header_t *)&rsp->data[0];
			block_desc = (scsi_mode_block_desc_t *)(page_hdr + 1);
			page_data = (scsi_mode_page_desc_t *)(block_desc + 1);

			printf("Page hdr:         ");
			p = (char *)page_hdr;
			for (i = 0; i < sizeof(scsi_mode_header_t); ++i, ++p) 
				printf("%02x ", *p & 0xff);
			printf("\n");

			p = (char *)block_desc;
			printf("Block descriptor: ");
			for (i = 0; i < sizeof(scsi_mode_block_desc_t); ++i, ++p) 
				printf("%02x ", *p & 0xff);
			printf("\n");

			printf("Page Code:        %d (0x%x)\n", page_data->code & 0x7F, page_data->code & 0x7F);
			printf("Page Len:         %d\n", page_data->len);
			printf("Page Data:        0x%04x: ", 0);

			p = (char *)page_data->u.data;
			for (i = 0; i < page_data->len; ++i) {
				if (i > 0 && (i % 8) == 0)
					printf("\n                  0x%04x: ", i);
				printf("%02x ", page_data->u.data[i] & 0xff);
			}
			printf("\n");

			return(0);
		}
		else {
			printf("%s failed: non-zero SCSI status 0x%x\n", CMD, scsi_status);
			if (scsi_status == SCSI_DEV_STATUS_CHECK) {
				printf("Sense : ");
				for (i = 0; i < sizeof(rsp->senseData); ++i) {
					if (i > 0 && (i % 8) == 0)
						printf("\n      : ");
					printf("%02x ", rsp->senseData[i]);
				}
				printf("\n");
			}
			return(0);
		}
	}
}

int 
sdm_ui_scsiModeSelect(int argc, char *argv[])
{
    sdm_relayCdbReqMsg_t		msense_msg;
    sdm_relayCdbReqMsg_t *		mselect_msg;
    sdm_relayCdbRspMsg_t *		rsp;
    char						respbuff[1024];
    int							rc;
	int							scsi_status;
	int							i;
	uint32						sdmHandle;
	int							page_code;
	int							offset, value;
	scsi_mode_header_t *		page_hdr;
	scsi_mode_block_desc_t *	block_desc;
	scsi_mode_page_desc_t *		page_data;
	char *						end_ptr;

#undef CMD
#define CMD "scsi modeselect"

	if (argc != 4) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return(1);
	}

	page_code = strtol(argv[1], &end_ptr, 0);
	if (*end_ptr != '\0') {
		printf("%s failed: Invalid page code '%s'\n", CMD, argv[2]);
		return(1);
	}
	offset = strtol(argv[2], NULL, 0);
	value = strtol(argv[3], NULL, 0);

	sdmHandle = sdm_ui_getSdmHandle(0, 0, argv[0]);
	if (sdmHandle == 0)
		return(1);
	
	/* 
	 * First read (sense) the page so we can modify it using select.
	 */
	bzero(&msense_msg, sizeof(sdm_relayCdbReqMsg_t));
	msense_msg.hdr.mid = SDM_MID_RELAY_CDB_REQ;
	msense_msg.handle  = sdmHandle;
	msense_msg.cdb[0]  = SCSI_CMD_MODE_SENSE_6;
	msense_msg.cdb[2]  = page_code;
	msense_msg.cdb[4]  = 255;
	
	rc = sendAgileMsg(-1, &msense_msg, sizeof(sdm_relayCdbReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  respbuff, sizeof(respbuff));
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("%s failed: timeout\n", CMD);
		return(1);
	}

	rsp = (sdm_relayCdbRspMsg_t *)respbuff;
	
	if (rsp->hdr.mid != SDM_MID_RELAY_CDB_RSP) {
		printf("%s failed: Invalid Response Received: %x\n", CMD, rsp->hdr.mid);
		return(1);
	}
	else 
	if (rsp->eCode != SDM_OK) {
		printf("%s failed: SDM eCode 0x%x\n", CMD, rsp->eCode);
		return(1);
	}
	else 
	if (rsp->status & SD_DRV_STATUS_MASK) {
		printf("%s failed: non-zero driver status: 0x%x\n", CMD, 
			   (rsp->status & SD_DRV_STATUS_MASK) >> SD_DRV_STATUS_SHIFT);
		return(1);
	}
	else {
		scsi_status = (rsp->status & SD_DEV_STATUS_MASK) >> SD_DEV_STATUS_SHIFT;
		if (scsi_status != SCSI_DEV_STATUS_GOOD) {
			printf("%s failed: non-zero SCSI status 0x%x returned by mode sense.\n", CMD, scsi_status);
			if (scsi_status == SCSI_DEV_STATUS_CHECK) {
				printf("Sense : ");
				for (i = 0; i < sizeof(rsp->senseData); ++i) {
					if (i > 0 && (i % 8) == 0)
						printf("\n      : ");
					printf("%02x ", rsp->senseData[i]);
				}
				printf("\n");
			}
			return(0);
		}
	}

	/* 
	 * Next, copy mode sense data to mode select buffer and mask out
	 * values required before doing mode select.  
	 */
	mselect_msg = (sdm_relayCdbReqMsg_t *)malloc(sizeof(sdm_relayCdbReqMsg_t) + rsp->dataLen);
	bzero(mselect_msg, sizeof(sdm_relayCdbReqMsg_t));
	bcopy(&rsp->data[0], &mselect_msg->data[0], rsp->dataLen);

	page_hdr = (scsi_mode_header_t *)&mselect_msg->data[0];
	block_desc = (scsi_mode_block_desc_t *)(page_hdr + 1);
	page_data = (scsi_mode_page_desc_t *)(block_desc + 1);
	
	page_hdr->length = 0;
	block_desc->blocks[0] = 0;
	block_desc->blocks[1] = 0;
	block_desc->blocks[2] = 0;
	page_data->code &= ~MODE_PAGE_CODE_SAVEABLE;

	/*
	 * Now, insert the user defined values to change
	 */
	if (offset >= page_data->len) {
		printf("Offset (%d) exceed page length (%d).\n", offset, page_data->len);
		return(1);
	}
	printf("Changing byte at offset %d from 0x%x --> 0x%x\n",
		   offset, page_data->u.data[offset], value);
	page_data->u.data[offset] = value;

	/* 
	 * Ship off the mode select command 
	 */
	mselect_msg->hdr.mid = SDM_MID_RELAY_CDB_REQ;
	mselect_msg->handle = sdmHandle;
	mselect_msg->dataLen = sizeof(scsi_mode_header_t) + 
		sizeof(scsi_mode_block_desc_t) + 
		2 + page_data->len;
	
	mselect_msg->cdb[0] = SCSI_CMD_MODE_SELECT_6;
	mselect_msg->cdb[1] = 0x10;		/* PF bit */
	mselect_msg->cdb[4] = mselect_msg->dataLen;

	rc = sendAgileMsg(-1, mselect_msg, sizeof(sdm_relayCdbReqMsg_t) + mselect_msg->dataLen,
					  "sdm", 0, 0, 0,
					  NULL,
					  respbuff, sizeof(respbuff));
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("%s failed: timeout\n", CMD);
		return(1);
	}

	rsp = (sdm_relayCdbRspMsg_t *)respbuff;
	
	if (rsp->hdr.mid != SDM_MID_RELAY_CDB_RSP) {
		printf("%s failed: Invalid Response Received: %x\n", CMD, rsp->hdr.mid);
		return(1);
	}
	else 
	if (rsp->eCode != SDM_OK) {
		printf("%s failed: SDM eCode 0x%x\n", CMD, rsp->eCode);
		return(1);
	}
	else 
	if (rsp->status & SD_DRV_STATUS_MASK) {
		printf("%s failed: non-zero driver status: 0x%x\n", CMD, 
			   (rsp->status & SD_DRV_STATUS_MASK) >> SD_DRV_STATUS_SHIFT);
		return(1);
	}
	else {
		scsi_status = (rsp->status & SD_DEV_STATUS_MASK) >> SD_DEV_STATUS_SHIFT;
		if (scsi_status != SCSI_DEV_STATUS_GOOD) {
			printf("%s failed: non-zero SCSI status 0x%x returned by mode sense.\n", CMD, scsi_status);
			if (scsi_status == SCSI_DEV_STATUS_CHECK) {
				printf("Sense : ");
				for (i = 0; i < sizeof(rsp->senseData); ++i) {
					if (i > 0 && (i % 8) == 0)
						printf("\n      : ");
					printf("%02x ", rsp->senseData[i]);
				}
				printf("\n");
			}
			return(0);
		}
	}

	free(mselect_msg);
	return(0);
}

/*-----------------------------------------------------------------
 * Name: sdm_ui_scsiFormat
 *
 * Description: Send format command to disk device.
 *
 * Created by:  Frank Salzmann
 *
 * Date Created: August 31, 2001
 *
 *-----------------------------------------------------------------
 */
int
sdm_ui_scsiFormat(int argc, char *argv[])
{
    int                   stat;
    int                   rc;
	char				  ans;
    char                  respbuff[1024];
	unsigned int		  mlen;
	unsigned int		  sectSize;
    sdm_relayCdbReqMsg_t  *msg;
    sdm_relayCdbRspMsg_t  *rsp;
	uint32                sdmHandle;	

	if ( argc != 2 ) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return -1;
	}

	sectSize = strtol(argv[1],NULL,10);

	//-- Get open sdm_devHandle from the SDM.
	sdmHandle = sdm_ui_getSdmHandle(0, 0, argv[0]);
	if( sdmHandle == 0 ) 
		return 1;	
	printf("\nWarning: Disk Format Request: Device %s, New Sector Size: %d bytes\n",
		   argv[0], sectSize);
	printf("Do you want to continue? [y|n]: ");
	scanf("%c", &ans);
	if( ans != 'y' ) {
		printf("\nFormat Terminated.\n\n");
		exit(1);
	}
	mlen = sizeof(sdm_relayCdbReqMsg_t) + 12;
	msg = (sdm_relayCdbReqMsg_t *)malloc(mlen);
	memset(msg, 0, mlen);
    msg->hdr.mid = SDM_MID_RELAY_CDB_REQ;
    msg->handle  = sdmHandle;

	// Send Mode Select Command with new sector size information.
	// Ref: ANSI X3.301-1997 Pg. 106-108
	msg->cdb[0]  = SCSI_CMD_MODE_SELECT_6;
	msg->cdb[4]  = 0xC;
	msg->dataLen = 0xC;
	msg->data[3] = 0x8;

	msg->data[9]  = (sectSize >> 16) & 0xFF;
	msg->data[10] = (sectSize >> 8) & 0xFF;
	msg->data[11] = sectSize & 0xFF;

    stat = sendAgileMsg(-1, msg, mlen,
                        "sdm", 0, 0, 0,
                        NULL,
                        respbuff, sizeof(respbuff));

    if( bad_stat(stat) ) {
        printf("Create failed\n");
        rc = 1;
    }
    rsp = (sdm_relayCdbRspMsg_t *)respbuff;

    if( rsp->hdr.mid != SDM_MID_RELAY_CDB_RSP ) {
        printf("Error: Invalid Response Received: %x\n", rsp->hdr.mid);
        rc = 1;
    }
    else {
        if( rsp->eCode == SDM_FAIL ) {
            printf("RELAY CDB ERROR:  %s\n\n", argv[0]);
            rc = 1;
        }
        else {
            //printf("\nMode Select Done\n");

			// Mode Select Completed, now send down the SCSI FORMAT_UNIT
			// command.
			memset(msg, 0, sizeof(sdm_relayCdbReqMsg_t));
			msg->hdr.mid = SDM_MID_RELAY_CDB_REQ;
			msg->handle  = sdmHandle;
			msg->cdb[0]  = SCSI_CMD_FORMAT_UNIT;

			printf("Formating.......");
			stat = sendAgileMsg(-1, msg, sizeof(sdm_relayCdbReqMsg_t),
								"sdm", 0, 0, 0,
								NULL,
								respbuff, sizeof(respbuff));

			if( bad_stat(stat) ) {
				printf("Create failed\n");
				rc = 1;
			}
			rsp = (sdm_relayCdbRspMsg_t *)respbuff;
			if( rsp->hdr.mid != SDM_MID_RELAY_CDB_RSP ) {
				printf("Error: Invalid Response Received: %x\n", rsp->hdr.mid);
				rc = 1;
			}
			else {
				if( rsp->eCode == SDM_FAIL ) {
					printf("Format Request Failed: %s\n", argv[0]);
					rc = 1;
				}
				else {
					printf("Format Complete!: %s\n", argv[0]);
					rc = 0;
				}
			}
		}
	}
    return rc;

} // sdm_ui_scsiFormat

int
sdm_ui_scsiResvIn(int argc, char *argv[])
{
    sdm_relayCdbReqMsg_t  msg;
    sdm_relayCdbRspMsg_t  *rsp;
    char                  respbuff[1024];
    int                   rc;
	int					  scsi_status;
	int					  i;
	uint32 				  sdmHandle;
	uint32                allocLen, val;

#undef CMD
#define CMD "scsi reserve in" 

	if( argc != 1) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return -1;
	}

	//-- Get open sdm_devHandle from the SDM.
	sdmHandle = sdm_ui_getSdmHandle(0, 0, argv[0]);
	if( sdmHandle == 0 ) 
		return 1;	

	allocLen    = 512;

	bzero(&msg, sizeof(sdm_relayCdbReqMsg_t));
	msg.hdr.mid   = SDM_MID_RELAY_CDB_REQ;
	msg.handle    = sdmHandle;
	msg.dataInLen = allocLen;
	msg.cdb[0]    = 0x5E;  		// SCSI PERSISTENT RESERVE IN command
	msg.cdb[1]    = 1;  			// Service Code: Read Reservations
	msg.cdb[7]    = (allocLen & 0xFF00) >> 8;
	msg.cdb[8]    = allocLen & 0xFF;
	
	rc = sendAgileMsg(-1, &msg, sizeof(sdm_relayCdbReqMsg_t),
						"sdm", 0, 0, 0,
						NULL,
						respbuff, sizeof(respbuff));
	if (rc < 0) {
		printf("\n%s failed: sendAgileMsg() failed\n\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("\n%s failed: timeout\n\n", CMD);
		return(1);
	}

	rsp = (sdm_relayCdbRspMsg_t *)respbuff;
	
	if (rsp->hdr.mid != SDM_MID_RELAY_CDB_RSP) {
		printf("\n%s failed: Invalid Response Received: %x\n\n", CMD, rsp->hdr.mid);
		return(1);
	}
	else 
	if (rsp->eCode != SDM_OK) {
			printf("\n%s failed: SDM eCode 0x%x\n\n", CMD, rsp->eCode);
			return(1);
	}
	else 
	if (rsp->status & SD_DRV_STATUS_MASK) {
		printf("\n%s failed: non-zero driver status: 0x%x\n\n", CMD, 
			   (rsp->status & SD_DRV_STATUS_MASK) >> SD_DRV_STATUS_SHIFT);
		return(1);
	}
	else {
		int entryCnt, j;
		scsi_status = (rsp->status & SD_DEV_STATUS_MASK) >> SD_DEV_STATUS_SHIFT;
		if (scsi_status == SCSI_DEV_STATUS_GOOD) {
			printf("\nPersistent Reserve In command accepted!\n\n");
			val  = (rsp->data[0] << 24) & 0xFF000000;
			val |= (rsp->data[1] << 16) & 0x00FF0000;
			val |= (rsp->data[2] <<  8) & 0x0000FF00;
			val |=  rsp->data[3] & 0xFF;
			printf("Generation Count: %d\n", val);
			val  = (rsp->data[4] << 24) & 0xFF000000;
			val |= (rsp->data[5] << 16) & 0x00FF0000;
			val |= (rsp->data[6] <<  8) & 0x0000FF00;
			val |=  rsp->data[7] & 0xFF;
			entryCnt = val / 16;
			printf("Number Of Entries: %d\n", entryCnt);
			for(i=0; i < val; i+=16) {
				printf("Reservation Key: ");
				for(j=8; j < 16; j++) 
					printf("%02x ", rsp->data[j+i]);
				printf("\n");
			}

			return(0);
		}
		else {
			printf("\nDevice is not ready: SCSI status 0x%x\n", scsi_status);
			if (scsi_status == SCSI_DEV_STATUS_CHECK) {
				printf("Sense : ");
				for (i = 0; i < sizeof(rsp->senseData); ++i) {
					if (i > 0 && (i % 8) == 0)
						printf("\n      : ");
					printf("%02x ", rsp->senseData[i]);
				}
				printf("\n\n");
			}
			return(0);
		}
	}
} // sdm_ui_scsiResvIn

#ifdef  RESERVE_RAW 
int
sdm_ui_scsiResvOut(int argc, char *argv[])
{
    sdm_relayCdbReqMsg_t  *msg;
    sdm_relayCdbRspMsg_t  *rsp;
    char                  respbuff[1024];
	uchar8				  svcAction;
	char                  *svcString;
    int                   rc;
	int					  scsi_status;
	int					  i;
	uint64                reserveKey;
	uint64				  serviceKey;
	uint32 				  sdmHandle;
	uint32                allocLen, val;

#undef CMD
#define CMD "scsi reserve out" 

	if( argc != 4) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return -1;
	}

	if( !strcmp( argv[1], "reg") ) {
		svcAction = 0;
		svcString = "Register";
	}
	else
	if( !strcmp( argv[1], "resv") ) {
		svcAction = 1;
		svcString = "Reserve";
	}
	else
	if( !strcmp( argv[1], "rel") ) {
		svcAction = 2;
		svcString = "Release";
	}
	else
	if( !strcmp( argv[1], "clr") ) {
		svcAction = 3;
		svcString = "Clear";
	}
	else
	if( !strcmp( argv[1], "preempt") ) {
		svcAction = 4;
		svcString = "Preempt";
	}
	else {
		printf("Error: Invalid Actioni\n");
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return -1;
	}

	reserveKey = strtoq(argv[2],NULL,0);
	serviceKey = strtoq(argv[3],NULL,0);
		

	//-- Get open sdm_devHandle from the SDM.
	sdmHandle = sdm_ui_getSdmHandle(0, 0, argv[0]);
	if( sdmHandle == 0 ) 
		return 1;	

	allocLen    = 24;

	msg = (sdm_relayCdbReqMsg_t *)malloc(sizeof(sdm_relayCdbReqMsg_t) + allocLen);

	bzero(msg, sizeof(sdm_relayCdbReqMsg_t) + allocLen);
	msg->hdr.mid   = SDM_MID_RELAY_CDB_REQ;
	msg->handle    = sdmHandle;
	msg->dataLen   = allocLen;
	msg->cdb[0]    = 0x5F;  		// SCSI PERSISTENT RESERVE OUT command
	msg->cdb[1]    = svcAction; 	
	msg->cdb[2]    = 1;				// Scope/Type  : LU/Write Exclusive
	msg->cdb[7]    = (allocLen & 0xFF00) >> 8;
	msg->cdb[8]    = allocLen & 0xFF;
	
	//-- Reserve Out Parameter List
	msg->data[0]  = (reserveKey & 0xFF00000000000000) >> 56;
	msg->data[1]  = (reserveKey & 0x00FF000000000000) >> 48;
	msg->data[2]  = (reserveKey & 0x0000FF0000000000) >> 40;
	msg->data[3]  = (reserveKey & 0x000000FF00000000) >> 32;
	msg->data[4]  = (reserveKey & 0x00000000FF000000) >> 24;
	msg->data[5]  = (reserveKey & 0x0000000000FF0000) >> 16;
	msg->data[6]  = (reserveKey & 0x000000000000FF00) >> 8;
	msg->data[7]  = (reserveKey & 0x00000000000000FF);

	msg->data[8]  = (serviceKey & 0xFF00000000000000) >> 56;
	msg->data[9]  = (serviceKey & 0x00FF000000000000) >> 48;
	msg->data[10] = (serviceKey & 0x0000FF0000000000) >> 40;
	msg->data[11] = (serviceKey & 0x000000FF00000000) >> 32;
	msg->data[12] = (serviceKey & 0x00000000FF000000) >> 24;
	msg->data[13] = (serviceKey & 0x0000000000FF0000) >> 16;
	msg->data[14] = (serviceKey & 0x000000000000FF00) >> 8;
	msg->data[15] = (serviceKey & 0x00000000000000FF);

	rc = sendAgileMsg(-1, msg, sizeof(sdm_relayCdbReqMsg_t) + allocLen,
						"sdm", 0, 0, 0,
						NULL,
						respbuff, sizeof(respbuff));
	free( msg );

	if (rc < 0) {
		printf("\n%s failed: sendAgileMsg() failed\n\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("\n%s failed: timeout\n\n", CMD);
		return(1);
	}

	rsp = (sdm_relayCdbRspMsg_t *)respbuff;
	
	if (rsp->hdr.mid != SDM_MID_RELAY_CDB_RSP) {
		printf("\n%s failed: Invalid Response Received: %x\n\n", CMD, rsp->hdr.mid);
		return(1);
	}
	else 
	if (rsp->eCode != SDM_OK) {
			printf("\n%s failed: SDM eCode 0x%x\n\n", CMD, rsp->eCode);
			return(1);
	}
	else 
	if (rsp->status & SD_DRV_STATUS_MASK) {
		printf("\n%s failed: non-zero driver status: 0x%x\n\n", CMD, 
			   (rsp->status & SD_DRV_STATUS_MASK) >> SD_DRV_STATUS_SHIFT);
		return(1);
	}
	else {
		scsi_status = (rsp->status & SD_DEV_STATUS_MASK) >> SD_DEV_STATUS_SHIFT;

		switch( scsi_status ) {

			case SCSI_DEV_STATUS_GOOD: 
				printf("\nPersistent Reserve Out:  Action: %s Completed!\n", svcString);
				break;
			case SCSI_DEV_STATUS_CHECK:
				printf("\nDevice returned: CHECK\n");
				for (i = 0; i < sizeof(rsp->senseData); ++i) {
					if (i > 0 && (i % 8) == 0)
						printf("\n      : ");
					printf("%02x ", rsp->senseData[i]);
				}
				printf("\n\n");
				break;

			case SCSI_DEV_STATUS_BUSY:
				printf("\nDevice returned: BUSY\n");
				break;
			case SCSI_DEV_STATUS_RESERVE_CONFLICT:
				printf("\nDevice returned: RESERVE CONFLICT\n");
				break;
			case SCSI_DEV_STATUS_CMD_TERMINATED:
				printf("\nDevice returned: CMD TERMINATED\n");
				break;
			case SCSI_DEV_STATUS_TASK_SET_FULL:
				printf("\nDevice returned: TASK SET FULL\n");
				break;
			default:
				printf("\nDevice returned: Other Error status\n");
		}
	}
	return 0;
} // sdm_ui_scsiResvOut

#else // RESERVE_RAW

int
sdm_ui_scsiResvOut(int argc, char *argv[])
{
    sdm_reserveReqMsg_t  *msg;
    sdm_reserveRspMsg_t  *rsp;
    char                  respbuff[1024];
	uchar8				  svcAction;
	char                  *svcString;
    int                   rc;
	uint64                key;
	uint32 				  sdmHandle;
	uint32                allocLen;

#undef CMD
#define CMD "scsi reserve out" 

	if( argc != 3) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return -1;
	}

	if( !strcmp( argv[1], "reg") ) {
		svcAction = 0;
		svcString = "Register";
	}
	else
	if( !strcmp( argv[1], "resv") ) {
		svcAction = 1;
		svcString = "Reserve";
	}
	else
	if( !strcmp( argv[1], "rel") ) {
		svcAction = 2;
		svcString = "Release";
	}
	else
	if( !strcmp( argv[1], "clr") ) {
		svcAction = 3;
		svcString = "Clear";
	}
	else
	if( !strcmp( argv[1], "preempt") ) {
		svcAction = 4;
		svcString = "Preempt";
	}
	else {
		printf("Error: Invalid Actioni\n");
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return -1;
	}

	key = strtoq(argv[2],NULL,0);

	//-- Get open sdm_devHandle from the SDM.
	sdmHandle = sdm_ui_getSdmHandle(0, 0, argv[0]);
	if( sdmHandle == 0 ) 
		return 1;	

	allocLen    = 24;

	msg = (sdm_reserveReqMsg_t *)malloc(sizeof(sdm_reserveReqMsg_t));

	bzero(msg, sizeof(sdm_reserveReqMsg_t));
	msg->hdr.mid   = SDM_MID_DEV_RESERVE_REQ;
	msg->handle    = sdmHandle;
	msg->action    = svcAction;
	msg->key       = key;
	
	rc = sendAgileMsg(-1, msg, sizeof(sdm_reserveReqMsg_t),
						"sdm", 0, 0, 0,
						NULL,
						respbuff, sizeof(respbuff));
	free( msg );

	if (rc < 0) {
		printf("\n%s failed: sendAgileMsg() failed\n\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("\n%s failed: timeout\n\n", CMD);
		return(1);
	}

	rsp = (sdm_reserveRspMsg_t *)respbuff;
	
	if (rsp->hdr.mid != SDM_MID_DEV_RESERVE_RSP) {
		printf("\n%s failed: Invalid Response Received: %x\n\n", CMD, rsp->hdr.mid);
		return(1);
	}
	else 
	if (rsp->rCode != SDM_OK) {
		char *causestr;
		if( rsp->cause == SDM_DEV_RSV_CONFLICT )
			causestr = "Reservation Conflict";
		else
		if( rsp->cause == SDM_DEV_RSV_DEV_ERROR ) 
			causestr = "Device Command Error";
		else
		if( rsp->cause == SDM_DEV_RSV_BAD_HANDLE ) 
			causestr = "Bad Device Handle";
		else
			causestr = "Unknown Error";
		printf("\nReserve Action: %s Failed. Cause: %s\n", svcString, causestr );
		return(1);
	}
	else {
		printf("\nPersistent Reserve Out:  Action: %s Completed!\n", svcString);
	}
	return 0;
} // sdm_ui_scsiResvOut
#endif // RESERVE_RAW

/*
 *	"control dev reset <wwn>"
 */
int 
sdm_ui_controlDevReset(int argc, char *argv[])
{
    sdm_devResetReqMsg_t  msg;
    sdm_devResetRspMsg_t  *rsp;
    char                  respbuff[1024];
    int                   rc;
	uint32 				  sdmHandle;

#undef CMD
#define CMD "dev reset"

	if (argc != 1) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return(1);
	}

	sdmHandle = sdm_ui_getSdmHandle(0, 0, argv[0]);
	if (sdmHandle == 0)
		return(1);

	bzero(&msg, sizeof(sdm_devResetReqMsg_t));
	msg.hdr.mid = SDM_MID_DEV_RESET_REQ;
	msg.handle  = sdmHandle;
	
	rc = sendAgileMsg(-1, &msg, sizeof(sdm_devResetReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  respbuff, sizeof(respbuff));
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("%s failed: timeout\n", CMD);
		return(1);
	}

	rsp = (sdm_devResetRspMsg_t *)respbuff;
	
	if (rsp->hdr.mid != SDM_MID_DEV_RESET_RSP) {
		printf("%s failed: Invalid Response Received: %x\n", CMD, rsp->hdr.mid);
		return(1);
	}
	else 
	if (rsp->eCode != SDM_OK) {
			printf("%s failed: SDM eCode %d\n", CMD, rsp->eCode);
			return(1);
	}
	return(0);
} 

/*-----------------------------------------------------------------
 * Name: sdm_ui_viewPortStats
 *
 * Description: Displays storage port statistics
 *
 * Created by:  Frank Salzmann
 *
 * Date Created: January 17, 2002
 *
 *-----------------------------------------------------------------
 */
int
sdm_ui_viewPortStats(int argc, char *argv[])
{
    sdm_ioctlReqMsg_t		msg;
	sdm_ioctlRspMsg_t *		rsp;
    char					respbuff[1024];
    int						rc;
	uint32					slot;
	uint32					port;
	hal_fc_link_stats_t *	link_stats;

#undef CMD
#define CMD "view port stats"

	if (argc < 1) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return(1);
	}

	rc = sdm_ui_parsePortString(argv[0], &slot, &port);
	if (rc != NFX_OK) {
		printf("%s failed: invalid port identifier '%s'\n", CMD, argv[0]);
		return(1);
	}

	if (!sdm_ui_portIsAlive(slot, port)) {
		printf("%s failed: port %d on SP-NIM in slot %d not alive.\n", CMD, port, slot);
		return(1);
	}

	bzero(&msg, sizeof(sdm_ioctlReqMsg_t));
	msg.hdr.mid = SDM_MID_IOCTL_REQ;
	msg.code = HAL_IOCTL_FC_GET_ADAPTER_LINK_STATS;
	msg.slot = slot;
	msg.port = port;
	rc = sendAgileMsg(-1, &msg, sizeof(sdm_ioctlReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  respbuff, sizeof(respbuff));
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("%s failed: timeout\n", CMD);
		return(1);
	}

	rsp = (sdm_ioctlRspMsg_t *)respbuff;
	
	if (rsp->hdr.mid != SDM_MID_IOCTL_RSP) {
		printf("%s failed: Invalid Response Received: %x\n", CMD, rsp->hdr.mid);
		return(1);
	}
	else 
	if (rsp->eCode != SDM_OK) {
			printf("%s failed: SCSI eCode %d\n", CMD, rsp->eCode);
			return(1);
	}
	else
	if (rsp->dataLen < sizeof(hal_fc_link_stats_t)) {
		printf("%s failed: Data Length too short: expected at least %d, got %d\n", 
			   CMD, sizeof(hal_fc_link_stats_t), rsp->dataLen);
		return(1);
	}

	link_stats = (hal_fc_link_stats_t *)rsp->data;

	/* Set the value of the rtc to usec conversion factor as
	 * returned by the server.
	 */
	count_per_usec = link_stats->count_per_usec;

	printf("link failure cnt       : %d\n", link_stats->link_failure);
	printf("loss-of-sync cnt       : %d\n", link_stats->loss_of_sync);
	printf("loss-of-signal cnt     : %d\n", link_stats->loss_of_signal);
	printf("invalid xmit word cnt  : %d\n", link_stats->invalid_xmit_words);
	printf("invalid frame CRC cnt  : %d\n", link_stats->invalid_crc);
	printf("\n");
	printf("frames out cnt         : %lld\n", link_stats->frames_out);
	printf("frames in cnt          : %lld\n", link_stats->frames_in);
	printf("transport error cnt    : %d\n", link_stats->transport_errors);
	printf("data underrun cnt      : %d\n", link_stats->underrun_cnt);
	printf("\n");
	printf("SCSI read 6 cnt        : %d\n", link_stats->read_6_cmd_cnt);
	printf("SCSI read 10 cnt       : %d\n", link_stats->read_10_cmd_cnt);
	printf("SCSI write 6 cnt       : %d\n", link_stats->write_6_cmd_cnt);
	printf("SCSI write 10 cnt      : %d\n", link_stats->write_10_cmd_cnt);
	printf("SCSI other cnt         : %d\n", link_stats->other_cmd_cnt);
	printf("\n");

    printf("SCSI reads              : %d\n", 
           link_stats->userReadFp + link_stats->userReadSsc +
           link_stats->metaReadFp + link_stats->metaReadSsc);
    printf("SCSI writes             : %d\n", 
           link_stats->writeFp + link_stats->writeSsc);
    printf("SCSI reads bytes        : %lld\n", 
           link_stats->readUserBytes_ssc + link_stats->readUserBytes_fp +
           link_stats->readMetaBytes_ssc + link_stats->readMetaBytes_fp);
    printf("SCSI writes bytes       : %lld\n", 
           link_stats->writeUserBytes_ssc + link_stats->writeUserBytes_fp);

    if ((link_stats->userReadFp + link_stats->userReadSsc + link_stats->metaReadFp + link_stats->metaReadSsc) > 0)
    {
        printf("SCSI Avg tot Read Time  : %lld\n",
			   (eee_rtcToUsec(&link_stats->totalTimeUserRead) +
				eee_rtcToUsec(&link_stats->totalTimeMetaRead)) / 
               (link_stats->userReadFp + link_stats->userReadSsc + link_stats->metaReadFp + link_stats->metaReadSsc));
        printf("SCSI Avg FC Read Time   : %lld\n",
			   (eee_rtcToUsec(&link_stats->fcTimeUserRead) +
				eee_rtcToUsec(&link_stats->fcTimeMetaRead)) / 
               (link_stats->userReadFp + link_stats->userReadSsc + link_stats->metaReadFp + link_stats->metaReadSsc));
    }
    if ((link_stats->writeFp + link_stats->writeSsc) > 0)
    {
        printf("SCSI Avg tot Write Time : %lld\n",
			   (eee_rtcToUsec(&link_stats->totalTimeUserWrite) +
				eee_rtcToUsec(&link_stats->totalTimeMetaWrite)) / 
               (link_stats->writeFp + link_stats->writeSsc));
        printf("SCSI Avg FC Write Time  : %lld\n",
			   (eee_rtcToUsec(&link_stats->fcTimeUserWrite) +
				eee_rtcToUsec(&link_stats->fcTimeMetaWrite)) /
               (link_stats->writeFp + link_stats->writeSsc));
    }

	printf("\n");

    if (link_stats->bucketSize1 > 0)
    {
        printf("FC response time (in usec):\n");
        printf("more than      %-7d    %-7d    %-7d    %-7d    "
               "%-7d    %-7d    maxTime    avgTime    \n",
               link_stats->bucketSize1, link_stats->bucketSize2, link_stats->bucketSize3, 
               link_stats->bucketSize4, link_stats->bucketSize5, link_stats->bucketSize6);
        printf("-----------------------------------------------------------"
               "--------------------------------------------\n");
        printf("Read tot...... %-10lld %-10lld %-10lld %-10lld %-10lld %-10lld "
               "%-10lld %-10lld\n",
               link_stats->totReadDataRspBuckets.moreThanBucket1,
               link_stats->totReadDataRspBuckets.moreThanBucket2,
               link_stats->totReadDataRspBuckets.moreThanBucket3,
               link_stats->totReadDataRspBuckets.moreThanBucket4,
               link_stats->totReadDataRspBuckets.moreThanBucket5,
               link_stats->totReadDataRspBuckets.moreThanBucket6,
               eee_rtcToUsec(&link_stats->totReadDataRspBuckets.maxTime),
               (link_stats->totReadDataRspBuckets.numTime == 0)? 0:
               eee_rtcToUsec(&link_stats->totReadDataRspBuckets.sumTime) / link_stats->totReadDataRspBuckets.numTime);
        printf("Read FC ...... %-10lld %-10lld %-10lld %-10lld %-10lld %-10lld "
               "%-10lld %-10lld\n",
               link_stats->fcReadDataRspBuckets.moreThanBucket1,
               link_stats->fcReadDataRspBuckets.moreThanBucket2,
               link_stats->fcReadDataRspBuckets.moreThanBucket3,
               link_stats->fcReadDataRspBuckets.moreThanBucket4,
               link_stats->fcReadDataRspBuckets.moreThanBucket5,
               link_stats->fcReadDataRspBuckets.moreThanBucket6,
               eee_rtcToUsec(&link_stats->fcReadDataRspBuckets.maxTime),
               (link_stats->fcReadDataRspBuckets.numTime == 0)? 0:
               eee_rtcToUsec(&link_stats->fcReadDataRspBuckets.sumTime) / link_stats->fcReadDataRspBuckets.numTime);
        printf("Write tot..... %-10lld %-10lld %-10lld %-10lld %-10lld %-10lld "
               "%-10lld %-10lld\n",
               link_stats->totWriteRspBuckets.moreThanBucket1,
               link_stats->totWriteRspBuckets.moreThanBucket2,
               link_stats->totWriteRspBuckets.moreThanBucket3,
               link_stats->totWriteRspBuckets.moreThanBucket4,
               link_stats->totWriteRspBuckets.moreThanBucket5,
               link_stats->totWriteRspBuckets.moreThanBucket6,
               eee_rtcToUsec(&link_stats->totWriteRspBuckets.maxTime),
               (link_stats->totWriteRspBuckets.numTime == 0)? 0:
               eee_rtcToUsec(&link_stats->totWriteRspBuckets.sumTime) / link_stats->totWriteRspBuckets.numTime);
        printf("Write FC...... %-10lld %-10lld %-10lld %-10lld %-10lld %-10lld "
               "%-10lld %-10lld\n",
               link_stats->fcWriteRspBuckets.moreThanBucket1,
               link_stats->fcWriteRspBuckets.moreThanBucket2,
               link_stats->fcWriteRspBuckets.moreThanBucket3,
               link_stats->fcWriteRspBuckets.moreThanBucket4,
               link_stats->fcWriteRspBuckets.moreThanBucket5,
               link_stats->fcWriteRspBuckets.moreThanBucket6,
               eee_rtcToUsec(&link_stats->fcWriteRspBuckets.maxTime),
               (link_stats->fcWriteRspBuckets.numTime == 0)? 0:
               eee_rtcToUsec(&link_stats->fcWriteRspBuckets.sumTime) / link_stats->fcWriteRspBuckets.numTime);
    }
    
	return 0;
} // sdm_ui_viewPortStats

/*-----------------------------------------------------------------
 * Name: sdm_ui_viewPortDetail
 *
 * Description: Displays detailed information about a storage port.
 *
 * Created by:  Frank Salzmann
 *
 * Date Created: January 17, 2002
 *
 *-----------------------------------------------------------------
 */
int
sdm_ui_viewPortDetail(int argc, char *argv[])
{
    sdm_ioctlReqMsg_t		msg;
	sdm_ioctlRspMsg_t *		adap_info_rsp = 0;
	sdm_ioctlRspMsg_t *		adap_connection_state_rsp = 0;
	sdm_ioctlRspMsg_t *		rsp;
	int						rsp_len;
    int						rc;
	uint32					slot;
	uint32					port;
	hal_adapter_info_t *	info;
	uint32					connection_state;

#undef CMD
#define CMD "view port detail"

	if (argc < 1) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return(1);
	}

	rc = sdm_ui_parsePortString(argv[0], &slot, &port);
	if (rc != NFX_OK) {
		printf("%s failed: invalid port identifier '%s'\n", CMD, argv[0]);
		return(1);
	}
	
	if (!sdm_ui_portIsAlive(slot, port)) {
		printf("%s failed: port %d on SP-NIM in slot %d not alive.\n", CMD, port, slot);
		return(1);
	}

	/*
	 * Get PORT/LINK INFO
	 */
	bzero(&msg, sizeof(sdm_ioctlReqMsg_t));
	msg.hdr.mid = SDM_MID_IOCTL_REQ;
	msg.code = HAL_IOCTL_COMMON_GET_ADAPTER_INFO;
	msg.slot = slot;
	msg.port = port;
	
	rsp_len = sizeof(sdm_ioctlRspMsg_t) + sizeof(hal_adapter_info_t);
	rsp = adap_info_rsp = malloc(rsp_len);
	if (rsp == NULL) {
		rc = 1;
		goto done;
	}
	rc = sendAgileMsg(-1, &msg, sizeof(sdm_ioctlReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  rsp, rsp_len);
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		rc = 1;
		goto done;
	}
	else 
	if (rc == 0) {
		printf("%s failed: timeout\n", CMD);
		rc = 1;
		goto done;
	}
	if (rsp->hdr.mid != SDM_MID_IOCTL_RSP) {
		printf("%s failed: Invalid Response Received: %x\n", CMD, rsp->hdr.mid);
		rc = 1;
		goto done;
	}
	else 
	if (rsp->eCode != SDM_OK) {
			printf("%s failed: SCSI eCode %d\n", CMD, rsp->eCode);
			rc = 1;
			goto done;
	}

	if (adap_info_rsp->dataLen < sizeof(hal_adapter_info_t)) {
		printf("%s failed: Data Length too short: expected at least %d, got %d\n", 
			   CMD, sizeof(hal_adapter_info_t), adap_info_rsp->dataLen);
		rc = 1;
		goto done;
	}
	info = (hal_adapter_info_t *)adap_info_rsp->data;


	/*
	 * Get PORT/LINK CONNECTION STATE
	 */
	bzero(&msg, sizeof(sdm_ioctlReqMsg_t));
	msg.hdr.mid = SDM_MID_IOCTL_REQ;
	msg.code = HAL_IOCTL_FC_CONNECTION_STATE;
	msg.slot = slot;
	msg.port = port;
	
	rsp_len = sizeof(sdm_ioctlRspMsg_t) + sizeof(uint32);
	rsp = adap_connection_state_rsp = malloc(rsp_len);
	if (rsp == NULL) {
		rc = 1;
		goto done;
	}
	rc = sendAgileMsg(-1, &msg, sizeof(sdm_ioctlReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  rsp, rsp_len);
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		rc = 1;
		goto done;
	}
	else 
	if (rc == 0) {
		printf("%s failed: timeout\n", CMD);
		rc = 1;
		goto done;
	}
	if (rsp->hdr.mid != SDM_MID_IOCTL_RSP) {
		printf("%s failed: Invalid Response Received: %x\n", CMD, rsp->hdr.mid);
		rc = 1;
		goto done;
	}
	else 
	if (rsp->eCode != SDM_OK) {
			printf("%s failed: SCSI eCode %d\n", CMD, rsp->eCode);
			rc = 1;
			goto done;
	}

	if (adap_connection_state_rsp->dataLen < sizeof(uint32)) {
		printf("%s failed: Data Length too short: expected at least %d, got %d\n", 
			   CMD, sizeof(uint32), adap_connection_state_rsp->dataLen);
		rc = 1;
		goto done;
	}
	bcopy(adap_connection_state_rsp->data, &connection_state, sizeof(uint32));
	
	printf("adapter name                 : %s\n", info->name);
	printf("firmware revision            : %s\n", info->firmware_rev);
	printf("hardware revision            : %s\n", info->hardware_rev);
	printf("interface type               : ");
	switch(info->interface_type) {
	case HAL_IF_FC_L_PORT:	    printf("L_PORT  "); break;
	case HAL_IF_FC_NL_PORT:	    printf("NL_PORT "); break;
	case HAL_IF_FC_N_PORT:	    printf("N_PORT  "); break;
	case HAL_IF_FC_F_PORT:	    printf("F_PORT  "); break;
	case HAL_IF_FC_FL_PORT:	    printf("FL_PORT "); break;
	default: 
		printf("Unknown port type %d", info->interface_type);
	}
	printf("\n");
	printf("link state                   : ");
	switch(connection_state) {
	case HAL_IOCTL_FC_LINK_STATE_LINK_DOWN:		printf("DOWN"); break;
	case HAL_IOCTL_FC_LINK_STATE_LINK_UP:		printf("UP"); break;
	case HAL_IOCTL_FC_LINK_STATE_LINK_OFFLINE:	printf("DISABLED"); break;
	default: 
		printf("Unknown connection state %d", connection_state);
	}
	printf("\n");
	printf("link speed                   : %d Gbit\n", info->u.fc_data.link_speed);
	printf("Max. SCSI tags per device    : %d\n", info->max_tags_per_device);
	printf("Fibrechannel nodename        : %016llx\n", info->u.fc_data.fc_node_name.u.wwn);
	printf("Max. Fibrechannel frame size : %d\n", info->u.fc_data.fc_packet_size);
	printf("SRAM parity available        : %s\n", info->u.fc_data.sram_parity ? "YES" : "NO");

 done:
	if (adap_info_rsp)
		free(adap_info_rsp);
	if (adap_connection_state_rsp)
		free(adap_connection_state_rsp);
	
	return 0;
} // sdm_ui_viewPortDetail

char *
sdm_ui_portTypeToStr(uint32 porttype) 
{
	switch(porttype) {
	case 0x01:
		return("N_PORT    ");
	case 0x02:
		return("NL_PORT   ");
	case 0x81:
		return("F_PORT    ");
	case 0x82:
		return("FL_PORT   ");
	case 0x84:
		return("E_PORT    ");
	case 0x85:
		return("B_PORT    ");
	default:
		return("UNATTACHED");
	}
}

int
sdm_ui_viewPortTopology(int argc, char *argv[])
{
    sdm_ioctlReqMsg_t		msg;
	sdm_ioctlRspMsg_t *		rsp;
    char					respbuff[1500];
    int						rc;
	uint32					slot;
	uint32					port;
	uchar8 *				topology;
	wwn_t					wwn;
	byte					format_type;
	byte					type;
	byte					ie, ie_count;
	byte					ie_port, ie_port_count;
	byte					aport, aport_count;
	byte					val;
	byte					i, port_count;

#undef CMD
#define CMD "view port topology"

	if (argc < 1) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return(1);
	}

	rc = sdm_ui_parsePortString(argv[0], &slot, &port);
	if (rc != NFX_OK) {
		printf("%s failed: invalid port identifier '%s'\n", CMD, argv[0]);
		return(1);
	}

	if (!sdm_ui_portIsAlive(slot, port)) {
		printf("%s failed: port %d on SP-NIM in slot %d not alive.\n", CMD, port, slot);
		return(1);
	}

	bzero(&msg, sizeof(sdm_ioctlReqMsg_t));
	msg.hdr.mid = SDM_MID_IOCTL_REQ;
	msg.code = HAL_IOCTL_FC_TOPOLOGY;
	msg.slot = slot;
	msg.port = port;
	
	rc = sendAgileMsg(-1, &msg, sizeof(sdm_ioctlReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  respbuff, sizeof(respbuff));
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("%s failed: timeout\n", CMD);
		return(1);
	}

	rsp = (sdm_ioctlRspMsg_t *)respbuff;
	
	if (rsp->hdr.mid != SDM_MID_IOCTL_RSP) {
		printf("%s failed: Invalid Response Received: %x\n", CMD, rsp->hdr.mid);
		return(1);
	}
	else 
	if (rsp->eCode != SDM_OK) {
			printf("%s failed: SCSI eCode %d\n", CMD, rsp->eCode);
			return(1);
	}
	else
	if (rsp->dataLen < 10) {
		printf("%s failed: Data Length too short: expected at least %d, got %d\n", 
			   CMD, 10, rsp->dataLen);
		return(1);
	}

	topology = (char *)rsp->data;

	/* Parse topology buffer */
	printf("SAN topology as seen from port %d of SP-NIM in slot %d: \n", port, slot);

	WWN_SET(&wwn, topology); topology += 8;								     /* get self port name         - 8 bytes */	
	printf("\tMy portname: %016llx\n", wwn.u.wwn);

	format_type = *topology++;										         /* get format type            - 1 byte  */
	if (format_type == 0) {
		printf("\tHost Loop ID:    %d\n", *topology++);                      /* get host loop id           - 1 byte  */
		printf("\tLoop Port Count: %d\n", (port_count = *topology++));       /* get loop port count        - 1 byte  */
		printf("\t               : ");
		for (i = 0; i < port_count; ++i)
			printf("%3d ", *topology++);                                     /* get loop port              - 1 byte  */
		printf("\n");
	}
	else
	if (format_type == 1) {
		printf("\tHost ID:    0x%02x\n", *topology++);						 /* get host loop id           - 1 byte  */
		topology++;															 /* skip loop port count       - 1 byte  */
		topology++;															 /* skip first port            - 1 byte  */
		printf("\tPeer ID:    0x%02x\n", *topology++);						 /* get second port            - 1 byte  */		
	}
	else {
		ie_count = *topology++;												 /* get IE count               - 1 byte  */
		printf("\tIE count: %2d\n", ie_count);

		for (ie = 0; ie < ie_count; ++ie) {
			WWN_SET(&wwn, topology); topology += 8;							 /* get IE nodename            - 8 bytes */
			printf("\tIE[%2d]: nodename %016llx, ", ie, wwn.u.wwn);

			val = *topology++;												 /* get symbolic name len      - 1 byte  */
			printf("symbolic name '%s', ", topology);						 /* get symbolic name          - n bytes */
			topology += val;

			type = *topology++;												 /* get IE type                - 1 byte  */
			printf("type %s, ", sdm_ui_portTypeToStr(type));

			ie_port_count = *topology++;									 /* get port count             - 1 byte  */
			printf("port count %2d\n", ie_port_count);

			for (ie_port = 0; ie_port < ie_port_count; ++ie_port) {
				WWN_SET(&wwn, topology); topology += 8;						 /* get port name              - 8 bytes */
				printf("\t\tPort[%2d]: nodename %016llx, ", ie_port, wwn.u.wwn);
				val = *topology++;											 /* get physical port number   - 1 byte  */
				printf("Physical port number %2d, ", val);
				type = *topology++;                                          /* get port type              - 1 byte  */
				printf("type %s, ", sdm_ui_portTypeToStr(type));
				aport_count = *topology++;									 /* get attached port count    - 1 byte  */
				printf("Attached device count %2d\n", aport_count);
			
				if (aport_count > 0) {
					printf("\t\t\tAttached devices: ");
					for (aport = 0; aport < aport_count; ++aport) {
						WWN_SET(&wwn, topology); topology += 8;				 /* get attached port name     - 8 bytes */
						printf("%016llx ", wwn.u.wwn);
					}
					printf("\n");
				}
			}
		}
	}

	return 0;
} // sdm_ui_viewPortTopology

int 
sdm_ui_controlPortReset(int argc, char *argv[])
{
    sdm_ioctlReqMsg_t		msg;
	sdm_ioctlRspMsg_t *		rsp;
    char					respbuff[1024];
    int						rc;
	uint32					slot;
	uint32					port;

#undef CMD
#define CMD "control port reset"

	if (argc < 1) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return(1);
	}

	rc = sdm_ui_parsePortString(argv[0], &slot, &port);
	if (rc != NFX_OK) {
		printf("%s failed: invalid port identifier '%s'\n", CMD, argv[0]);
		return(1);
	}

	if (!sdm_ui_portIsAlive(slot, port)) {
		printf("%s failed: port %d on SP-NIM in slot %d not alive.\n", CMD, port, slot);
		return(1);
	}

	bzero(&msg, sizeof(sdm_ioctlReqMsg_t));
	msg.hdr.mid = SDM_MID_IOCTL_REQ;
	msg.code = HAL_IOCTL_COMMON_RESET_ADAPTER;
	msg.slot = slot;
	msg.port = port;
	
	rc = sendAgileMsg(-1, &msg, sizeof(sdm_ioctlReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  respbuff, sizeof(respbuff));
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("%s failed: timeout\n", CMD);
		return(1);
	}

	rsp = (sdm_ioctlRspMsg_t *)respbuff;

	if (rsp->hdr.mid != SDM_MID_IOCTL_RSP) {
		printf("%s failed: Invalid Response Received: %x\n", CMD, rsp->hdr.mid);
		return(1);
	}
	else 
	if (rsp->eCode != SDM_OK) {
		printf("%s failed: SCSI eCode %d\n", CMD, rsp->eCode);
		return(1);
	}

	return 0;
}

int 
sdm_ui_controlPortDisable(int argc, char *argv[])
{
    sdm_ioctlReqMsg_t		msg;
	sdm_ioctlRspMsg_t *		rsp;
    char					respbuff[1024];
    int						rc;
	uint32					slot;
	uint32					port;

#undef CMD
#define CMD "control port disable"

	if (argc < 1) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return(1);
	}

	rc = sdm_ui_parsePortString(argv[0], &slot, &port);
	if (rc != NFX_OK) {
		printf("%s failed: invalid port identifier '%s'\n", CMD, argv[0]);
		return(1);
	}

	if (!sdm_ui_portIsAlive(slot, port)) {
		printf("%s failed: port %d on SP-NIM in slot %d not alive.\n", CMD, port, slot);
		return(1);
	}

	bzero(&msg, sizeof(sdm_ioctlReqMsg_t));
	msg.hdr.mid = SDM_MID_IOCTL_REQ;
	msg.code = HAL_IOCTL_COMMON_DISABLE_ADAPTER;
	msg.slot = slot;
	msg.port = port;
	
	rc = sendAgileMsg(-1, &msg, sizeof(sdm_ioctlReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  respbuff, sizeof(respbuff));
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("%s failed: timeout\n", CMD);
		return(1);
	}

	rsp = (sdm_ioctlRspMsg_t *)respbuff;

	if (rsp->hdr.mid != SDM_MID_IOCTL_RSP) {
		printf("%s failed: Invalid Response Received: %x\n", CMD, rsp->hdr.mid);
		return(1);
	}
	else 
	if (rsp->eCode != SDM_OK) {
		printf("%s failed: SCSI eCode %d\n", CMD, rsp->eCode);
		return(1);
	}

	return 0;
}

int 
sdm_ui_controlPortEnable(int argc, char *argv[])
{
    sdm_ioctlReqMsg_t		msg;
	sdm_ioctlRspMsg_t *		rsp;
    char					respbuff[1024];
    int						rc;
	uint32					slot;
	uint32					port;

#undef CMD
#define CMD "control port enable"

	if (argc < 1) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return(1);
	}

	rc = sdm_ui_parsePortString(argv[0], &slot, &port);
	if (rc != NFX_OK) {
		printf("%s failed: invalid port identifier '%s'\n", CMD, argv[0]);
		return(1);
	}

	if (!sdm_ui_portIsAlive(slot, port)) {
		printf("%s failed: port %d on SP-NIM in slot %d not alive.\n", CMD, port, slot);
		return(1);
	}

	bzero(&msg, sizeof(sdm_ioctlReqMsg_t));
	msg.hdr.mid = SDM_MID_IOCTL_REQ;
	msg.code = HAL_IOCTL_COMMON_ENABLE_ADAPTER;
	msg.slot = slot;
	msg.port = port;
	
	rc = sendAgileMsg(-1, &msg, sizeof(sdm_ioctlReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  respbuff, sizeof(respbuff));
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("%s failed: timeout\n", CMD);
		return(1);
	}

	rsp = (sdm_ioctlRspMsg_t *)respbuff;

	if (rsp->hdr.mid != SDM_MID_IOCTL_RSP) {
		printf("%s failed: Invalid Response Received: %x\n", CMD, rsp->hdr.mid);
		return(1);
	}
	else 
	if (rsp->eCode != SDM_OK) {
		printf("%s failed: SCSI eCode %d\n", CMD, rsp->eCode);
		return(1);
	}

	return 0;
}

int 
sdm_ui_controlPortRescan(int argc, char *argv[])
{
    sdm_ioctlReqMsg_t		msg;
	sdm_ioctlRspMsg_t *		rsp;
    char					respbuff[1024];
    int						rc;
	uint32					slot;
	uint32					port;

#undef CMD
#define CMD "control port rescan"

	if (argc < 1) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return(1);
	}

	rc = sdm_ui_parsePortString(argv[0], &slot, &port);
	if (rc != NFX_OK) {
		printf("%s failed: invalid port identifier '%s'\n", CMD, argv[0]);
		return(1);
	}

	if (!sdm_ui_portIsAlive(slot, port)) {
		printf("%s failed: port %d on SP-NIM in slot %d not alive.\n", CMD, port, slot);
		return(1);
	}

	bzero(&msg, sizeof(sdm_ioctlReqMsg_t));
	msg.hdr.mid = SDM_MID_IOCTL_REQ;
	msg.code = HAL_IOCTL_COMMON_RESCAN;
	msg.slot = slot;
	msg.port = port;
	
	rc = sendAgileMsg(-1, &msg, sizeof(sdm_ioctlReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  respbuff, sizeof(respbuff));
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("%s failed: timeout\n", CMD);
		return(1);
	}

	rsp = (sdm_ioctlRspMsg_t *)respbuff;

	if (rsp->hdr.mid != SDM_MID_IOCTL_RSP) {
		printf("%s failed: Invalid Response Received: %x\n", CMD, rsp->hdr.mid);
		return(1);
	}
	else 
	if (rsp->eCode != SDM_OK) {
		printf("%s failed: SCSI eCode %d\n", CMD, rsp->eCode);
		return(1);
	}

	return 0;
}

int 
sdm_ui_controlPortTrace(int argc, char *argv[])
{
	printf("XXX Command Not Implemented.\n");
	return 0;
}

int 
sdm_ui_controlPortBypass(int argc, char *argv[])
{
    sdm_ioctlReqMsg_t		msg;
	sdm_ioctlRspMsg_t *		rsp;
    char					respbuff[1024];
    int						rc;
	uint32					slot;
	uint32					port;
	uint64					wwn;

#undef CMD
#define CMD "control port bypass"

	if (argc < 2) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return(1);
	}

	rc = sdm_ui_parsePortString(argv[0], &slot, &port);
	if (rc != NFX_OK) {
		printf("%s failed: invalid port identifier '%s'\n", CMD, argv[0]);
		return(1);
	}

	wwn = strtoq(argv[1],NULL,16);

	if (!sdm_ui_portIsAlive(slot, port)) {
		printf("%s failed: port %d on SP-NIM in slot %d not alive.\n", CMD, port, slot);
		return(1);
	}

	bzero(&msg, sizeof(sdm_ioctlReqMsg_t));
	msg.hdr.mid = SDM_MID_IOCTL_REQ;
	msg.code = HAL_IOCTL_FC_LOOP_PORT_BYPASS;
	msg.slot = slot;
	msg.port = port;
	msg.arg_count = 1;
	msg.args[0] = wwn;
	
	rc = sendAgileMsg(-1, &msg, sizeof(sdm_ioctlReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  respbuff, sizeof(respbuff));
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("%s failed: timeout\n", CMD);
		return(1);
	}

	rsp = (sdm_ioctlRspMsg_t *)respbuff;

	if (rsp->hdr.mid != SDM_MID_IOCTL_RSP) {
		printf("%s failed: Invalid Response Received: %x\n", CMD, rsp->hdr.mid);
		return(1);
	}
	else 
	if (rsp->eCode != SDM_OK) {
		printf("%s failed: SCSI eCode %d\n", CMD, rsp->eCode);
		return(1);
	}

	return 0;
}

int 
sdm_ui_controlPortUnBypass(int argc, char *argv[])
{
    sdm_ioctlReqMsg_t		msg;
	sdm_ioctlRspMsg_t *		rsp;
    char					respbuff[1024];
    int						rc;
	uint32					slot;
	uint32					port;
	uint64					wwn;

#undef CMD
#define CMD "control port unbypass"

	if (argc < 2) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return(1);
	}

	rc = sdm_ui_parsePortString(argv[0], &slot, &port);
	if (rc != NFX_OK) {
		printf("%s failed: invalid port identifier '%s'\n", CMD, argv[0]);
		return(1);
	}

	wwn = strtoq(argv[1],NULL,16);

	if (!sdm_ui_portIsAlive(slot, port)) {
		printf("%s failed: port %d on SP-NIM in slot %d not alive.\n", CMD, port, slot);
		return(1);
	}

	bzero(&msg, sizeof(sdm_ioctlReqMsg_t));
	msg.hdr.mid = SDM_MID_IOCTL_REQ;
	msg.code = HAL_IOCTL_FC_LOOP_PORT_ENABLE_ALL;
	msg.slot = slot;
	msg.port = port;
	msg.arg_count = 1;
	msg.args[0] = wwn;
	
	rc = sendAgileMsg(-1, &msg, sizeof(sdm_ioctlReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  respbuff, sizeof(respbuff));
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("%s failed: timeout\n", CMD);
		return(1);
	}

	rsp = (sdm_ioctlRspMsg_t *)respbuff;

	if (rsp->hdr.mid != SDM_MID_IOCTL_RSP) {
		printf("%s failed: Invalid Response Received: %x\n", CMD, rsp->hdr.mid);
		return(1);
	}
	else 
	if (rsp->eCode != SDM_OK) {
		printf("%s failed: SCSI eCode %d\n", CMD, rsp->eCode);
		return(1);
	}

	return 0;
}

int
sdm_ui_controlPortLip(int argc, char *argv[])
{
    sdm_ioctlReqMsg_t		msg;
	sdm_ioctlRspMsg_t *		rsp;
    char					respbuff[1024];
    int						rc;
	uint32					slot;
	uint32					port;
	uint64					wwn;
	uint32					lip_byte_3;
	uint32					lip_byte_4;

#undef CMD
#define CMD "control port lip"

	if (argc < 4) {
		sdm_ui_parse(argv[0], &sdm_ui_root[0], --argc, &argv[1]);
		return(1);
	}

	rc = sdm_ui_parsePortString(argv[0], &slot, &port);
	if (rc != NFX_OK) {
		printf("%s failed: invalid port identifier '%s'\n", CMD, argv[0]);
		return(1);
	}
	wwn = strtoq(argv[1],NULL,16);
	lip_byte_3 = strtol(argv[2],NULL,0);
	lip_byte_4 = strtol(argv[3],NULL,0);

	if (!sdm_ui_portIsAlive(slot, port)) {
		printf("%s failed: port %d on SP-NIM in slot %d not alive.\n", CMD, port, slot);
		return(1);
	}

	bzero(&msg, sizeof(sdm_ioctlReqMsg_t));
	msg.hdr.mid = SDM_MID_IOCTL_REQ;
	msg.code = HAL_IOCTL_FC_LIP;
	msg.slot = slot;
	msg.port = port;
	msg.arg_count = 3;
	msg.args[0] = wwn;
	msg.args[1] = lip_byte_3;
	msg.args[2] = lip_byte_4;
	
	rc = sendAgileMsg(-1, &msg, sizeof(sdm_ioctlReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL,
					  respbuff, sizeof(respbuff));
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("%s failed: timeout\n", CMD);
		return(1);
	}

	rsp = (sdm_ioctlRspMsg_t *)respbuff;

	if (rsp->hdr.mid != SDM_MID_IOCTL_RSP) {
		printf("%s failed: Invalid Response Received: %x\n", CMD, rsp->hdr.mid);
		return(1);
	}
	else 
	if (rsp->eCode != SDM_OK) {
		printf("%s failed: SCSI eCode %d\n", CMD, rsp->eCode);
		return(1);
	}

	return 0;
}

/*-----------------------------------------------------------------
 * Name: sdm_ui_getSdmHandle
 *
 * Description: Send WWN/Lun to SDM and get back the open SDM_HANDLE.
 *
 * Created by:  Frank Salzmann
 *
 * Date Created: February 04, 2002
 *
 *-----------------------------------------------------------------
 */
uint32
sdm_ui_getSdmHandle( int64 wwn, int32 lun, char *devname )
{
    int                       stat;
    uint32                    rc=0;
    char                	  respbuff[1024];
    sdm_getSdmHandleReqMsg_t  msg;
    sdm_getSdmHandleRspMsg_t *rsp;

	msg.hdr.mid = SDM_MID_GET_SDMHANDLE_REQ;
    msg.wwn      = wwn;
    msg.lun      = lun;
	if(devname)
		strncpy(msg.devname, devname, sizeof(msg.devname));
	else
		msg.devname[0] = 0;

    stat = sendAgileMsg(-1, &msg, sizeof(msg),
                        "sdm", 0, 0, 0,
                        NULL,
                        respbuff, sizeof(respbuff));

    if( bad_stat(stat) ) {
        printf("Msg Socket Create failed\n");
    }
    rsp = (sdm_getSdmHandleRspMsg_t *)respbuff;

    if( rsp->hdr.mid != SDM_MID_GET_SDMHANDLE_RSP ) 
        printf("Error: Invalid Response Received: %x\n", rsp->hdr.mid);
    else 
    if( rsp->rCode == SDM_FAIL ) 
		printf("\nDevice Not Open: 0x%016llx/%d\n\n", wwn, lun);
	else 
		rc = rsp->sdmHandle;

	return rc;
} // sdm_ui_getSdmHandle


/*-----------------------------------------------------------------
 * Name: sdm_ui_getSCSIId
 *
 * Description: Send WWN/Lun to SDM and get back the corresponding
 *              SCSI ID.
 *
 * Created by:  Radek Aster
 *
 * Date Created: March 19, 2002
 *
 *-----------------------------------------------------------------
 */
scsi_id_t
sdm_ui_getSCSIId( int64 wwn, int32 lun, char *name )
{
    int                       stat;
    scsi_id_t                 rc=0;
    char                	  respbuff[1024];
    sdm_getSCSIIdReqMsg_t     msg;
    sdm_getSCSIIdRspMsg_t *   rsp;

	msg.hdr.mid = SDM_MID_GET_SCSIID_REQ;
    msg.wwn      = wwn;
    msg.lun      = lun;
	if( name ) {
		strcpy(msg.name, name);
	}

    stat = sendAgileMsg(-1, &msg, sizeof(msg),
                        "sdm", 0, 0, 0,
                        NULL,
                        respbuff, sizeof(respbuff));

    if( bad_stat(stat) ) {
        printf("Msg Socket Create failed\n");
    }
    rsp = (sdm_getSCSIIdRspMsg_t *)respbuff;

    if( rsp->hdr.mid != SDM_MID_GET_SCSIID_RSP ) 
        printf("Error: Invalid Response Received: %x\n", rsp->hdr.mid);
    else 
    if( rsp->rCode == SDM_FAIL ) 
		rc = 0;
	else 
		rc = rsp->id;

	return rc;
} // sdm_ui_getSCSIId

int
sdm_ui_portIsAlive(int32 slot, int32 port)
{

    sdm_ioctlReqMsg_t		msg;
	sdm_ioctlRspMsg_t *		rsp;
	int						rsp_len;
    int						rc;
	struct timeval			to;

	to.tv_sec = 5;
	to.tv_usec = 0;
	
	/*
	 * "Ping" the port by getting the PORT/LINK INFO
	 */
	bzero(&msg, sizeof(sdm_ioctlReqMsg_t));
	msg.hdr.mid = SDM_MID_IOCTL_REQ;
	msg.code = HAL_IOCTL_COMMON_GET_ADAPTER_INFO;
	msg.slot = slot;
	msg.port = port;
	
	rsp_len = sizeof(sdm_ioctlRspMsg_t) + sizeof(hal_adapter_info_t);
	rsp = malloc(rsp_len);
	if (rsp == NULL) 
			return(0);

	rc = sendAgileMsg(-1, &msg, sizeof(sdm_ioctlReqMsg_t),
					  "sdm", 0, 0, 0,
					  &to,
					  rsp, rsp_len);
	if (rc <= 0 || 
		rsp->hdr.mid != SDM_MID_IOCTL_RSP ||
		rsp->eCode != SDM_OK) {
			free(rsp);
			return(0);
	}

	free(rsp);

	return(1);
}

/*-----------------------------------------------------------------
 * Name: sdm_ui_devDiscover
 *
 * Description: Send device discover message to SDM.
 *
 * Created by:  Frank Salzmann
 *
 * Date Created: April 09, 2002
 *
 *-----------------------------------------------------------------
 */
int
sdm_ui_devDiscover(int argc, char *argv[])
{
    int                 stat;
	char                *endptr;
    sdm_rediscReqMsg_t  msg;
	uint64 				wwn;

	if( argc != 1 )
		return -1;

	wwn = strtoq(argv[0],&endptr,16);
	if( argv[0] == endptr ) 
		return -1;

	msg.hdr.mid = SDM_MID_REDISCOVER_REQ;
    msg.wwn     = wwn;

    stat = sendAgileMsg(-1, &msg, sizeof(msg),
                        "sdm", 0, 0, 0,
                        NULL, NULL, NULL);

	return 0;
} // sdm_ui_devDiscover


int
sdm_strToLower( char *in_str, char *out_str)
{
	int i, len;
	
	len = strlen( in_str );
	if( len > MAX_ARG_LEN )
		len = MAX_ARG_LEN;

	for( i=0; i < len; i++ )
		out_str[i] = tolower(in_str[i]);

	return(0);
} // sdm_strToLower


/*-----------------------------------------------------------------
 * Name: sdm_ui_parsePortString
 *
 * Description: Parse the port identifier string into slot and port.
 *
 *    FC port identifier strings can take one of two forms
 *      sp.<port> where port is 0, 1, 2 or 3
 *      sp<slot>.<port> where slot is 0-15 and port is 0, 1, 2 or 3.
 *
 * Created by:  Radek Aster
 *
 * Date Created: April 23, 2002
 *
 *-----------------------------------------------------------------
 */
int
sdm_ui_parsePortString(char *str, uint32 *slot, uint32 *port)
{
#define DEFAULT_SLOT 2

	char *new_str;

	if (*str != 's' && *str != 'S')
		return(NFX_ERR);
	++str;
	if (*str != 'p' && *str != 'P')
		return(NFX_ERR);
	++str;

	if (*str == '.') {
		*slot = DEFAULT_SLOT;
		++str;
	}
	else {
		*slot = strtol(str, &new_str, 10);
		if (new_str == str)
			return(NFX_ERR);
		str = new_str;
		if (*str != '.')
			return(NFX_ERR);
		++str;
	}

	*port = strtol(str, &new_str, 10);
	if (new_str == str)
		return(NFX_ERR);

	if (*slot > 15 || *port > 3)
		return(NFX_ERR);

	return(NFX_OK);
}

//-----------------------------------------------------------
// LUN MASKING COMMANDS
// ---------------------------------------------------------

static char mask_names[][6] = { "N/A", "Mask", "Allow" };

int
sdm_ui_lunMaskShow(int argc, char *argv[])
{
	sdm_getMaskListReqMsg_t 	req;
	sdm_getMaskListRspMsg_t*	rsp;
	char						*respbuff;
	int							i, j, k;
	int							stat;

	if(argc != 0) return 1;

	// allocate some memory for the response
	respbuff = (char *)malloc(65535);
	if(!respbuff) {
		printf("Memory allocation failed: %s\n", strerror(errno));
		return errno;
	}

	req.hdr.mid = SDM_MID_DEV_MASKLIST_REQ;

    stat = sendAgileMsg(-1, &req, sizeof(req),
                        "sdm", 0, 0, 0,
                        NULL,
                        respbuff, 65535);

    if( bad_stat(stat) ) {
        printf("Msg Socket Create failed\n");
    }
    rsp = (sdm_getMaskListRspMsg_t *)respbuff;
	if(rsp->hdr.mid != SDM_MID_DEV_MASKLIST_RSP) {
		printf("Expected response 0x%x got 0x%x\n", SDM_MID_DEV_MASKLIST_RSP, rsp->hdr.mid);
		return 1;
	}

	if(rsp->count == 0)
		printf("No entries in LUN MASK list\n");
	else {
		printf("DEVICE NAME                                              WWN                 LUN   ACTION  PORT      SLOT\n");
		printf("-------------------------------------------------------  ------------------  ----  ------  --------  ------\n");
		for(i=0; i < rsp->count; i++) {
			k = 0;
			printf("%-55s  0x%-16llx  %-4d  %-6s  ", rsp->list[i].devname, 
					rsp->list[i].wwn, rsp->list[i].lun, mask_names[rsp->list[i].attr.action]);

			// print the port list
			if(rsp->list[i].attr.port == SDM_CONF_ATTR_MAX_PORT_BIT)
				printf("%-8s  ", "ALL");
			else {
				for(j=0; j < SDM_CONF_ATTR_MAX_PORT; j++)
					if(rsp->list[i].attr.port & (1 << j)) {
						printf("%d,", j);
						k += 2;
					}
				printf("\b   ");
				for(j=k; j < 8; j++)
					printf(" ");
			}

			// print the slot list
			if(rsp->list[i].attr.slot == SDM_CONF_ATTR_MAX_SLOT_BIT)
				printf("ALL\n");
			else {
				for(j=0; j < SDM_CONF_ATTR_MAX_SLOT; j++) {
					if(rsp->list[i].attr.slot & (1 << j))
						printf("%d,", j);
				}
				printf("\b \n");
			}
		} // end of for
	} // end of outer else

	return 0;
} // sdm_ui_lunMaskShow



/*
 * TYPE						SLOT		PORT	DEVNAME
 * SDM_STATS_TYPE_DEV  		0			0		devname 
 * SDM_STATS_TYPE_DEVALL	0			0		NULL
 * SDM_STATS_TYPE_SCSI		SLOT		PORT	NULL
 */ 
int
sdm_ui_clearStats(int type, uint32 slot, uint32 port, char *devname)
{
	sdm_clearStatsReqMsg_t		req;
	int							rc;

	memset((void *)&req, 0, sizeof(req));
	req.hdr.mid = SDM_MID_CLR_STATS_REQ;
	req.type = type;
	switch(type) {
		case SDM_STATS_TYPE_DEV:
			strncpy(req.info.dev_info.devname, devname, 128);
			break;
		case SDM_STATS_TYPE_DEVALL:
			break;
		case SDM_STATS_TYPE_SCSI:
			req.info.scsi_info.slot = slot;
			req.info.scsi_info.port = port;
			break;
		default:
			printf("Unknown clear stat type %d\n", type);
			return 1;
	}

	rc = sendAgileMsg(-1, &req, sizeof(req),
					  "sdm", 0, 0, 0,
					  NULL, NULL, 0);
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		return(1);
	}

	return  0;
} // sdm_ui_clearStats

int
sdm_ui_clearDevStats(int argc, char *argv[])
{
	int		rc;

	if(argc != 0 && argc != 1) return -1;

	if(argc == 1)
		rc = sdm_ui_clearStats(SDM_STATS_TYPE_DEV, 0, 0, argv[0]);
	else
		rc = sdm_ui_clearStats(SDM_STATS_TYPE_DEVALL, 0, 0, NULL);
	if(!rc) {
		  if(argc) printf("Statistics for dev %s cleared !\n", argv[0]);
		  else printf("Statistics for all devices cleared !\n");
	}

	return 0;
} // sdm_ui_clearDevStats


int
sdm_ui_clearScsiStats(int argc, char *argv[])
{
	int		rc;
	int		slot;
	int		port;

	if(argc != 1) return 1;

	rc = sdm_ui_parsePortString(argv[0], &slot, &port);
	if (rc != NFX_OK) {
		printf("sdm_ui_clearScsiStats failed: invalid port identifier '%s'\n", argv[0]);
		return(1);
	}

	if (!sdm_ui_portIsAlive(slot, port)) {
		printf("sdm_ui_clearScsiStats failed: port %d on SP-NIM in slot %d not alive.\n", port, slot);
		return(1);
	}

	rc = sdm_ui_clearStats(SDM_STATS_TYPE_SCSI, slot, port, NULL);
	if(!rc) printf("SCSI statistics for sp%d.%d cleared !\n", slot, port);

	return 0;
} // sdm_ui_clearScsiStats

int
sdm_ui_clearFcStats(int argc, char *argv[])
{
	int 				rc;
	int 				slot;
	int					port;
	sdm_ioctlReqMsg_t 	msg;
	sdm_ioctlRspMsg_t 	*rsp;
	char				respbuff[1024];

#undef CMD		
#define CMD "clear port_stats"

	if(argc != 1) return 1;

	rc = sdm_ui_parsePortString(argv[0], &slot, &port);
	if (rc != NFX_OK) {
		printf("%s failed: invalid port identifier '%s'\n", CMD, argv[0]);
		return(1);
	}

	if (!sdm_ui_portIsAlive(slot, port)) {
		printf("%s failed: port %d on SP-NIM in slot %d not alive.\n", CMD, port, slot);
		return(1);
	}

	bzero(&msg, sizeof(sdm_ioctlReqMsg_t));
	msg.hdr.mid = SDM_MID_IOCTL_REQ;
	msg.code = HAL_IOCTL_FC_CLR_ADAPTER_LINK_STATS;
	msg.slot = slot;
	msg.port = port;

	rc = sendAgileMsg(-1, &msg, sizeof(sdm_ioctlReqMsg_t),
					  "sdm", 0, 0, 0,
					  NULL, respbuff, sizeof(respbuff));
	if (rc < 0) {
		printf("%s failed: sendAgileMsg() failed\n", CMD);
		return(1);
	}
	else 
	if (rc == 0) {
		printf("%s failed: timeout\n", CMD);
		return(1);
	}

	rsp = (sdm_ioctlRspMsg_t *)respbuff;

	if (rsp->hdr.mid != SDM_MID_IOCTL_RSP) {
		printf("%s failed: Invalid Response Received: %x\n", CMD, rsp->hdr.mid);
		return(1);
	}
	else 
	if (rsp->eCode != SDM_OK) {
			printf("%s failed: SCSI eCode %d\n", CMD, rsp->eCode);
			return(1);
	}

	printf("Fibre Channel Stats for sp%d.%d cleared !\n", slot, port);

	return 0;
} // sdm_ui_clearPortStats

int
sdm_ui_clearAllStats(int argc, char *argv[])
{
	static int	slot = 2;
	int 		port;
	char		*arglist[1];
	char		clearargs[1][5];

	if(argc != 0) return 1;

	// clear dev stats
	sdm_ui_clearStats(SDM_STATS_TYPE_DEVALL, 0, 0, NULL);
	printf("\nAll Device stats cleared !\n");
	
	// clear scsi stats
	for(port = 0; port < 4; port++) 
		sdm_ui_clearStats(SDM_STATS_TYPE_SCSI, slot, port, NULL);
	printf("All SCSI stats cleared !\n");
	
	// clear fc stats	
	arglist[0] = clearargs[0];
	for(port = 0; port < 4; port++) {
		sprintf(clearargs[0], "sp%d.%d", slot, port);
		sdm_ui_clearFcStats(1, arglist);
	}

	return 0;
} // sdm_ui_clearAllStats

int
sdm_ui_viewScsiStats(int argc, char *argv[])
{
    sdm_getStatsReqMsg_t	msg;
	sdm_getStatsRspMsg_t	rsp;
	uint32					slot;
	uint32					port;
	int						rc;
	scsi_stats_t 			*stats;
	uint32					tmp;

	if(argc != 1) return 1;

	rc = sdm_ui_parsePortString(argv[0], &slot, &port);
	if(rc != NFX_OK) return 1;
	
	memset((void *)&msg, 0, sizeof(msg));
	msg.hdr.mid = SDM_MID_GET_STATS_REQ;
	msg.type = SDM_STATS_TYPE_SCSI;
	msg.port = port;

	rc = sdm_ui_getStats(&msg, (char *)&rsp, sizeof(rsp));
	if(rc) {
		printf("Failed to get SCSI stats for sp%d.%d\n", slot, port);
		return 0;
	}


	stats = &rsp.stats.scsi_stats;

	printf("\n\n");
	printf("                                   SCSI Statistics: sp%d.%d\n", slot, port);
	printf("                             ======================================\n");
	printf("                  Request   Response   TotalTime   Avg TTime    FC Time    Avg FC Time              \n");
	printf("                  Count     Count        (usec)     (usec)       (usec)       (usec)        Bytes   \n");
	printf("                  -------   --------   ---------   ---------   ---------   -----------   -----------\n");
	tmp = stats->readUserRspCnt_ssc + stats->readUserRspCnt_fp;
	printf("user read(fp) ..  %7d   %8d   %9llu   %9llu   %9llu   %11llu   %11llu\n", 
	  stats->readUserReqCnt_fp, stats->readUserRspCnt_fp, 
	  eee_rtcToUsec(&stats->readUserTotalTime), 
	  tmp ? eee_rtcToUsec(&stats->readUserTotalTime)/tmp : 0,
	  eee_rtcToUsec(&stats->readUserFCTime),
	  tmp ? eee_rtcToUsec(&stats->readUserFCTime)/tmp : 0,
	  stats->readUserBytes_fp);
	printf("user read(ssc)..  %7d   %8d                                                     %11llu\n", 
		   stats->readUserReqCnt_ssc, 
		   stats->readUserRspCnt_ssc,
		   stats->readUserBytes_ssc);
	tmp =stats->readMetaRspCnt_fp + stats->readMetaRspCnt_ssc; 
	printf("meta read(fp)...  %7d   %8d   %9llu   %9llu   %9llu   %11llu   %11llu\n", 
	  stats->readMetaReqCnt_fp, stats->readMetaRspCnt_fp, 
	  eee_rtcToUsec(&stats->readMetaTotalTime), 
	  tmp ? eee_rtcToUsec(&stats->readMetaTotalTime)/tmp : 0,
	  eee_rtcToUsec(&stats->readMetaFCTime),
	  tmp ? eee_rtcToUsec(&stats->readMetaFCTime)/tmp : 0,
	  stats->readMetaBytes_fp);
	printf("meta read(ssc)..  %7d   %8d                                                     %11llu\n", 
		   stats->readMetaReqCnt_ssc, 
		   stats->readMetaRspCnt_ssc,
		   stats->readMetaBytes_ssc);
	tmp = stats->writeUserRspCnt_fp + stats->writeUserRspCnt_ssc ;
	printf("write(fp).......  %7d   %8d   %9llu   %9llu   %9llu   %11llu   %11llu\n", 
	 stats->writeUserReqCnt_fp, stats->writeUserRspCnt_fp, 
	 eee_rtcToUsec(&stats->writeUserTotalTime), 
	 tmp ? eee_rtcToUsec(&stats->writeUserTotalTime)/tmp : 0,
	 eee_rtcToUsec(&stats->writeUserFCTime),
	 tmp ? eee_rtcToUsec(&stats->writeUserFCTime)/tmp : 0,
	 stats->writeUserBytes_fp);
	printf("write(ssc)......  %7d   %8d                                                     %11llu\n", 
		   stats->writeUserReqCnt_ssc, 
		   stats->writeUserRspCnt_ssc,
		   stats->writeUserBytes_ssc);
#if 0
	printf("meta write......  %7d   %8d   %9llu   %9llu   %9llu   %11llu\n", 
	 stats->writeMetaReqCnt, stats->writeMetaRspCnt, 
	 eee_rtcToUsec(&stats->writeMetaTotalTime), 
	 stats->writeMetaRspCnt?eee_rtcToUsec(&stats->writeMetaTotalTime)/stats->writeMetaRspCnt:0,
	 eee_rtcToUsec(&stats->writeMetaFCTime),
	 stats->writeMetaRspCnt ? eee_rtcToUsec(&stats->writeMetaFCTime)/stats->writeMetaRspCnt:0);
#endif
	printf("cmd relay.......  %7d   %8d   %9llu   %9llu   %9llu   %11llu\n", 
	 stats->relayCnt, stats->relayRspCnt, 
	 eee_rtcToUsec(&stats->relayTotalTime), 
	 stats->relayRspCnt?eee_rtcToUsec(&stats->relayTotalTime)/stats->relayRspCnt:0,
	 eee_rtcToUsec(&stats->relayFCTime),
	 stats->relayRspCnt ? eee_rtcToUsec(&stats->relayFCTime)/stats->relayRspCnt:0);
	printf("open............  %7d   %8d   %9d   %9d   %9d   %11d\n", 
	 stats->openReqCnt, 0, 0, 0, 0, 0);
	printf("close...........  %7d   %8d   %9d   %9d   %9d   %11d\n", 
	 stats->closeReqCnt, 0, 0, 0, 0, 0);
	printf("sync............  %7d   %8d   %9d   %9d   %9d   %11d\n", 
	 stats->syncReqCnt, 0, 0, 0, 0, 0);
	printf("status..........  %7d   %8d   %9d   %9d   %9d   %11d\n", 
	 stats->statusReqCnt, 0, 0, 0, 0, 0);
	printf("stats...........  %7d   %8d   %9d   %9d   %9d   %11d\n", 
	 stats->statsReqCnt, 0, 0, 0, 0, 0);
	printf("iotl............  %7d   %8d   %9d   %9d   %9d   %11d\n", 
	 stats->ioctlReqCnt, 0, 0, 0, 0, 0);
	printf("delete..........  %7d   %8d   %9d   %9d   %9d   %11d\n", 
	 stats->deleteReqCnt, 0, 0, 0, 0, 0);
	printf("add.............  %7d   %8d   %9d   %9d   %9d   %11d\n", 
	 stats->addReqCnt, 0, 0, 0, 0, 0);
	printf("invalid.........  %7d   %8d   %9d   %9d   %9d   %11d\n", 
	 stats->invalidMsgCnt, 0, 0, 0, 0, 0);
	printf("\n");
#if 0
	printf("Total Message Count .......... %d\n", 
		   (stats->readUserReqCnt   + stats->readMetaReqCnt   +
			stats->writeUserReqCnt  + stats->writeMetaReqCnt  +
			stats->relayCnt         + stats->openReqCnt       +
			stats->closeReqCnt      + stats->syncReqCnt       +
			stats->statusReqCnt     + stats->statsReqCnt      +
			stats->ioctlReqCnt      + stats->deleteReqCnt     +
			stats->addReqCnt        + stats->invalidMsgCnt));
#endif

    printf("FC response time (in usec):\n");
    printf("more than     %-7d    %-7d    %-7d    %-7d    "
           "%-7d    %-7d     maxTime    avgTime    \n",
           SCSI_RSP_BUCKET_1, SCSI_RSP_BUCKET_2, SCSI_RSP_BUCKET_3, 
           SCSI_RSP_BUCKET_4, SCSI_RSP_BUCKET_5, SCSI_RSP_BUCKET_6);
    printf("-----------------------------------------------------------"
           "--------------------------------------------\n");
    printf("Read tot...... %-10lld %-10lld %-10lld %-10lld %-10lld %-10lld "
           "%-10lld %-10lld\n",
           stats->totReadDataRspBuckets.moreThanBucket1,
           stats->totReadDataRspBuckets.moreThanBucket2,
           stats->totReadDataRspBuckets.moreThanBucket3,
           stats->totReadDataRspBuckets.moreThanBucket4,
           stats->totReadDataRspBuckets.moreThanBucket5,
           stats->totReadDataRspBuckets.moreThanBucket6,
           eee_rtcToUsec(&stats->totReadDataRspBuckets.maxTime),
           (stats->totReadDataRspBuckets.numTime == 0)? 0: 
           eee_rtcToUsec(&stats->totReadDataRspBuckets.sumTime) / stats->totReadDataRspBuckets.numTime);
    printf("Read FC ...... %-10lld %-10lld %-10lld %-10lld %-10lld %-10lld "
           "%-10lld %-10lld\n",
           stats->fcReadDataRspBuckets.moreThanBucket1,
           stats->fcReadDataRspBuckets.moreThanBucket2,
           stats->fcReadDataRspBuckets.moreThanBucket3,
           stats->fcReadDataRspBuckets.moreThanBucket4,
           stats->fcReadDataRspBuckets.moreThanBucket5,
           stats->fcReadDataRspBuckets.moreThanBucket6,
           eee_rtcToUsec(&stats->fcReadDataRspBuckets.maxTime),
           (stats->fcReadDataRspBuckets.numTime == 0)? 0:
           eee_rtcToUsec(&stats->fcReadDataRspBuckets.sumTime) / stats->fcReadDataRspBuckets.numTime);
    printf("Write tot..... %-10lld %-10lld %-10lld %-10lld %-10lld %-10lld "
           "%-10lld %-10lld\n",
           stats->totWriteRspBuckets.moreThanBucket1,
           stats->totWriteRspBuckets.moreThanBucket2,
           stats->totWriteRspBuckets.moreThanBucket3,
           stats->totWriteRspBuckets.moreThanBucket4,
           stats->totWriteRspBuckets.moreThanBucket5,
           stats->totWriteRspBuckets.moreThanBucket6,
           eee_rtcToUsec(&stats->totWriteRspBuckets.maxTime),
           (stats->totWriteRspBuckets.numTime == 0)? 0:
           eee_rtcToUsec(&stats->totWriteRspBuckets.sumTime) / stats->totWriteRspBuckets.numTime);
    printf("Write FC...... %-10lld %-10lld %-10lld %-10lld %-10lld %-10lld "
           "%-10lld %-10lld\n",
           stats->fcWriteRspBuckets.moreThanBucket1,
           stats->fcWriteRspBuckets.moreThanBucket2,
           stats->fcWriteRspBuckets.moreThanBucket3,
           stats->fcWriteRspBuckets.moreThanBucket4,
           stats->fcWriteRspBuckets.moreThanBucket5,
           stats->fcWriteRspBuckets.moreThanBucket6,
           eee_rtcToUsec(&stats->fcWriteRspBuckets.maxTime),
           (stats->fcWriteRspBuckets.numTime == 0)? 0:
           eee_rtcToUsec(&stats->fcWriteRspBuckets.sumTime) / stats->fcWriteRspBuckets.numTime);

	printf("Num Messages Sent to SSC: %d\n", stats->sscMsgSentCnt);
	printf("SSC Messages Queue Size : %d\n", stats->sscMsgQueueSize);

	return 0;
} // sdm_ui_viewScsiStats

int
sdm_ui_hugeio(int argc, char *argv[])
{
	scsi_hugeioReqMsg_t		req;
	scsi_hugeioRspMsg_t		*rsp;
	char					respbuff[1024];
	scsi_id_t				id;
	sdm_devStatusReqMsg_t	statusReq;
	sdm_devStatusRspMsg_t*	statusRsp;
	int						stat;

	if(argc != 4 && argc != 5) return 1;

	// get device detail
	id = sdm_ui_getSCSIId(NULL, NULL, argv[0]);

	statusReq.hdr.mid = SDM_MID_DEV_STATUS_REQ;
	statusReq.id      = id;

	stat = sendAgileMsg(-1, &statusReq, sizeof(statusReq),
						"sdm", 0, 0, 0,
						NULL,
						respbuff, sizeof(respbuff));

	if( bad_stat(stat) ) {
		printf("Create failed\n");
		exit(1);
	}
	statusRsp = (sdm_devStatusRspMsg_t *)respbuff;

	if( statusRsp->hdr.mid != SDM_MID_DEV_STATUS_RSP ) {
		printf("Error: Invalid Response Received: 0x%x\n", statusRsp->hdr.mid);
		exit(1);
	}

	if( statusRsp->eCode == SDM_FAIL ) {
		printf("Device %s Not Found\n", argv[0]);
		exit(1);
	}
	if( statusRsp->handleCnt <= 0) {
		printf("Device %s Not Open\n", argv[0]);
		exit(1);
	}

	// fill request structure
	memset((void *)&req, 0, sizeof(req));
	req.hdr.mid = SCSI_MID_HUGEIO_REQ;
	if(!strcmp(argv[1], "read"))
		req.op = SCSI_HUGEIO_READ;
	else if(!strcmp(argv[1], "write"))
		req.op = SCSI_HUGEIO_WRITE;
	else exit(1);

	req.wwn = statusRsp->wwn;
	req.lun = statusRsp->lun;
	req.lba = atoi(argv[2]);
	req.size = atoi(argv[3]) * 1024;
	if(argc == 5)
		req.count = atoi(argv[4]);
	else
		req.count = 1;

	// send hugeio req. to FC slice and wait for response
	stat = sendAgileMsg(-1, &req, sizeof(req), "scsi", 
						SDM_HANDLE_GET_PORT(statusRsp->handles[0]), NFX_SLOTTYPE_FC, 
						0, NULL,
						respbuff, sizeof(respbuff));
	if(stat != sizeof(scsi_hugeioRspMsg_t)) {
		printf("Msg. send to SCSI failed rc=%d\n", stat);
		exit(1);
	}
	rsp = (scsi_hugeioRspMsg_t *)respbuff;
	if(rsp->hdr.mid != SCSI_MID_HUGEIO_RSP) {
		printf("Error: Invalid response received: 0x%x\n", rsp->hdr.mid);
		exit(1);
	}

	if(!rsp->eCode)
		printf("%s request completed successfully\n", argv[1]);
	else
		printf("Failed to do %s for device %s, eCode = %d\n", argv[1], argv[0], rsp->eCode);

	return 0;
} // sdm_ui_hugeio

int
sdm_ui_diskCheckError(int argc, char *argv[])
{
	char *devname = argv[0];
	
	if(argc != 1) return 1; 

	if(strstr(devname, "MYLEX")) 
		return sdm_ui_mylexDiskCheckError(devname);
	else
		printf("disk check for %s not supported\n", devname);

	return 0;
} // sdm_ui_diskCheck


int
sdm_ui_mylexGetDiskList(char *devname, raid_phyDevListRspMsg_t *rsp, int rspbuf_len)
{
	raid_phyDevListReqMsg_t		req;
	int							ret;

	req.header.mid = RAID_MID_PHYDEV_LIST_REQ;
	req.header.cookie = 1;
	strcpy(req.header.id, devname);

	ret = sendAgileMsg(-1, (void *)&req, sizeof(req), "raid-config", 0, 0, 0, NULL,
						(void *)rsp, rspbuf_len);
	if(ret < 0) {
		printf("sdm_ui_mylexGetDiskList: \n");
		return(NFX_ERR);
	}

	return 0;
}

int
sdm_ui_mylexDiskCheckError(char *devname)
{
	int		ret;
	char	cmd[256];
	char	rspbuf[65535];
	raid_phyDevListRspMsg_t	*diskListRsp;
	int		i;
	FILE	*fp;
	int64	read_error_fixed;
	int64	write_error_fixed;
	int64	read_error_notfixed;
	int64	write_error_notfixed;
	int64	sectors_read;
	int64	sectors_written;
	int		dev_open = 0;
	
	// open the raid device if required
	sprintf(cmd, "%s devlist|grep %s|grep OPEN >/dev/null 2>&1", RAID_UI_EXEC, devname);
	ret = system(cmd);
	if(ret == 0) dev_open = 1;
	else {
		sprintf(cmd, "%s open %s >/dev/null", RAID_UI_EXEC, devname);
		if(system(cmd) != 0) {
			printf("Failed to open raid device %s....Close device if already open\n", devname);
			return 1;		
		}
	}

	// get the list of physical devices for this mylex
	diskListRsp = (raid_phyDevListRspMsg_t *)rspbuf;
	ret = sdm_ui_mylexGetDiskList(devname, diskListRsp, sizeof(rspbuf));
	if(ret) return ret;

	printf("\n%s Read/Write Error Statistics\n\n", devname);
	printf("CID\tTID\tSectors Read\t         Read Errors      \tSectors Written\t         Write Errors\n");
	printf("   \t   \t            \t  Corrected  Not-Corrected\t               \t  Corrected  Not-Corrected\n");
	printf("---\t---\t------------\t  ---------- -------------\t---------------\t  ---------- -------------\n");

	// for each device, get the read and write error pages
	for(i=0; i < diskListRsp->numEntries; i++) {

		// sectors read
		sprintf(cmd, "%s get_logpage %s %d %d %d %d\n",
					RAID_UI_EXEC, devname,
					diskListRsp->idList[i].channelNumber,
					diskListRsp->idList[i].targetId,
					SCSI_LOGPAGE_CODE_ERROR_COUNTER_READ,
					SCSI_ERROR_LOGPAGE_BYTES_PROCESSED);
		if((fp = popen(cmd, "r")) == NULL) {
			printf("Failed to get read-error page for device %s cid/tid %d:%d\n", 
					devname, diskListRsp->idList[i].channelNumber,
					diskListRsp->idList[i].targetId);
			continue;		
		}
		fscanf(fp, "%lld", &sectors_read);
		pclose(fp);
		sectors_read /= 512;

		// corrected read errors
		sprintf(cmd, "%s get_logpage %s %d %d %d %d\n",
					RAID_UI_EXEC, devname, 
					diskListRsp->idList[i].channelNumber,
					diskListRsp->idList[i].targetId,
					SCSI_LOGPAGE_CODE_ERROR_COUNTER_READ,
					SCSI_ERROR_LOGPAGE_CORRECTED_ERRORS);
		if((fp = popen(cmd, "r")) == NULL) {
			printf("Failed to get read-error page for device %s cid/tid %d:%d\n", 
					devname, diskListRsp->idList[i].channelNumber,
					diskListRsp->idList[i].targetId);
			continue;		
		}
		fscanf(fp, "%lld", &read_error_fixed);
		pclose(fp);

		// non-correctable read errors
		sprintf(cmd, "%s get_logpage %s %d %d %d %d\n",
					RAID_UI_EXEC, devname, 
					diskListRsp->idList[i].channelNumber,
					diskListRsp->idList[i].targetId,
					SCSI_LOGPAGE_CODE_ERROR_COUNTER_READ,
					SCSI_ERROR_LOGPAGE_UNCORRECTED_ERRORS);
		if((fp = popen(cmd, "r")) == NULL) {
			printf("Failed to get read-error page for device %s cid/tid %d:%d\n", 
					devname, diskListRsp->idList[i].channelNumber,
					diskListRsp->idList[i].targetId);
			continue;		
		}
		fscanf(fp, "%lld", &read_error_notfixed);
		pclose(fp);

		// sectors written
		sprintf(cmd, "%s get_logpage %s %d %d %d %d\n",
					RAID_UI_EXEC, devname, 
					diskListRsp->idList[i].channelNumber,
					diskListRsp->idList[i].targetId,
					SCSI_LOGPAGE_CODE_ERROR_COUNTER_WRITE,
					SCSI_ERROR_LOGPAGE_BYTES_PROCESSED);	
		if((fp = popen(cmd, "r")) == NULL) {
			printf("Failed to get write-error page for device %s cid/tid %d:%d\n", 
					devname, diskListRsp->idList[i].channelNumber,
					diskListRsp->idList[i].targetId);
			continue;		
		}
		fscanf(fp, "%lld", &sectors_written);
		pclose(fp);
		sectors_written /= 512;
		
		// corrected write errors
		sprintf(cmd, "%s get_logpage %s %d %d %d %d\n",
					RAID_UI_EXEC, devname, 
					diskListRsp->idList[i].channelNumber,
					diskListRsp->idList[i].targetId,
					SCSI_LOGPAGE_CODE_ERROR_COUNTER_WRITE,
					SCSI_ERROR_LOGPAGE_CORRECTED_ERRORS);	
		if((fp = popen(cmd, "r")) == NULL) {
			printf("Failed to get write-error page for device %s cid/tid %d:%d\n", 
					devname, diskListRsp->idList[i].channelNumber,
					diskListRsp->idList[i].targetId);
			continue;		
		}
		fscanf(fp, "%lld", &write_error_fixed);
		pclose(fp);
		
		// non-correctable write errors
		sprintf(cmd, "%s get_logpage %s %d %d %d %d\n",
					RAID_UI_EXEC, devname, 
					diskListRsp->idList[i].channelNumber,
					diskListRsp->idList[i].targetId,
					SCSI_LOGPAGE_CODE_ERROR_COUNTER_WRITE,
					SCSI_ERROR_LOGPAGE_UNCORRECTED_ERRORS);	
		if((fp = popen(cmd, "r")) == NULL) {
			printf("Failed to get write-error page for device %s cid/tid %d:%d\n", 
					devname, diskListRsp->idList[i].channelNumber,
					diskListRsp->idList[i].targetId);
			continue;		
		}
		fscanf(fp, "%lld", &write_error_notfixed);
		pclose(fp);
		
		printf("%d\t%d\t%12lld\t%12lld %12lld\t%15lld\t%12lld %12lld\n", 
					diskListRsp->idList[i].channelNumber, 
					diskListRsp->idList[i].targetId,
					sectors_read,
					read_error_fixed,
					read_error_notfixed,
					sectors_written,
					write_error_fixed,
					write_error_notfixed);
	}
	printf("\n");
	
	// close the raid device, if reqd.
	if(!dev_open) {
		sprintf(cmd, "%s close %s", RAID_UI_EXEC, devname);
		if(system(cmd) != 0) 
			printf("Failed to close raid device %s: Please close manually\n", devname);
	}

	return 0;
	
} // sdm_ui_mylexDiskCheck
