/*------------------------------------------------------------------------------
 * RCS:  $Id: spm-lun.c,v 1.8 2002/09/28 00:59:13 nitinj Exp $
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : spm-lun.c
 *
 * Description  : This file contains the methods used to manipulate
 *                SPM's LUN database. There is a single
 *                instance of these on each SSC.
 *
 * Created by   : Nitin Jain
 * Date Created : 27-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

#include "spm.h"

/*
 * Global variables.
 */

DB *scsiDb;    // Index into spm_lunDbCb_t using SCSI-ID
DB *lunDb;     // Index into spm_lunDbCb_t using Device Name

/*
 * Local function prototypes.
 */

/*
 *------------------------------------------------------------------------------
 * Name         : spm_addLunToLnSdb
 *
 * Description  : Add this LUN to lunDb and scsiDb
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 27-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_addLunToLnSdb(spm_lunAttr_t *attr, int32 poolId,
                        char *node, char *err)
{
    DBT   key;
    DBT   rec;
    DBT   skey;
    DBT   srec;
    spm_lunDbCb_t   *lCfg = NULL;

    SPM_FCT_PRINT(("spm_addLunToLnSdb..."));
    if (!attr)
        return NFX_ERR;
    SPM_DBG_PRINT(("\tDev. : %s", attr->name));
    SPM_DBG_PRINT(("\tSCSI : 0x%llx", attr->scsiId));
    SPM_DBG_PRINT(("\tSize : %lld", attr->size));
    SPM_DBG_PRINT(("\tRAID : %d", attr->raidLevel));
    SPM_DBG_PRINT(("\tPool : %d", poolId));
    SPM_DBG_PRINT(("\t WWN : 0x%llx", attr->wwn));
    SPM_DBG_PRINT(("\t LUN : %d", attr->lun));
    SPM_DBG_PRINT(("\tMake : %s", attr->vendor));
    SPM_DBG_PRINT(("\tModel: %s", attr->model));
    SPM_DBG_PRINT(("\tNode : %s", node));
    SPM_DBG_PRINT(("\tClst : %s", attr->cluster));
    SPM_DBG_PRINT(("\tType : %d", attr->devType));

    /*
     * Add LUN to scsiDb and LunDb
     */
    SPM_ALLOC_ZERO(lCfg, spm_lunDbCb_t, sizeof(spm_lunDbCb_t));
    /*
     * Do not free the pointer. It will be stored in the database.
     * The delete routine will eventually free the memory
     */
    lCfg->magic  = SPM_MAGIC_STRING;
    lCfg->wwn    = attr->wwn;
    lCfg->lun    = attr->lun;
    lCfg->sid    = attr->scsiId;
    lCfg->size   = attr->size;
    lCfg->rL     = attr->raidLevel;
    lCfg->lstate = attr->state;
    lCfg->poolId = poolId;
    if (attr->name && strlen(attr->name)) strcpy(lCfg->dev, attr->name);
    if (attr->vendor && strlen(attr->vendor)) strcpy(lCfg->make, attr->vendor);
    if (attr->model && strlen(attr->model)) strcpy(lCfg->model, attr->model);
    if (node && strlen(node)) strcpy(lCfg->node, node);
    lCfg->volId  = attr->volId;
    if (attr->cluster && strlen(attr->cluster))
        strcpy(lCfg->cluster, attr->cluster);
    else if (lCfg->lstate == EVM_FREE_LUN || lCfg->lstate == EVM_USED_LUN)
        strcpy(lCfg->cluster, currCluster);
    if (attr->devType == SCSI_DEVTYPE_DISK)
        lCfg->devType = SPM_DEV_DISK;
    else if (attr->devType == SCSI_DEVTYPE_TAPE)
        lCfg->devType = SPM_DEV_TAPE;
    else if (attr->devType == SCSI_DEVTYPE_MEDIA_CHANGER)
        lCfg->devType = SPM_DEV_CHANGER;
    else
        lCfg->devType = SPM_DEV_MISC;

    /*
     * Add record to lunDb
     */
    key.data = attr->name;
    key.size = strlen(attr->name);
    rec.data = &lCfg;   // Store the pointer
    rec.size = sizeof(lCfg);
    if(lunDb->put(lunDb, &key, &rec, R_NOOVERWRITE) != 0)
    {
        sprintf(err, "Error putting lunDb rec: %s", strerror(errno));
        SPM_ERR_PRINT(("%s", err));
        return(NFX_ERR);
    }

    /*
     * Add record to scsiDb
     */
    skey.data = &lCfg->sid;
    skey.size = sizeof(scsi_id_t);
    srec.data = &lCfg;   // Store the pointer
    srec.size = sizeof(lCfg);
    if(scsiDb->put(scsiDb, &skey, &srec, R_NOOVERWRITE) != 0)
    {
        sprintf(err, "Error putting scsiDb rec: %s", strerror(errno));
        SPM_ERR_PRINT(("%s", err));
        return(NFX_ERR);
    }

    SPM_FCT_PRINT(("...spm_addLunToLnSdb"));
    return(NFX_OK);
}

/*
 *------------------------------------------------------------------------------
 * Name         : spm_delLunFromLnSdb
 *
 * Description  : Delete this LUN from lunDb and scsiDb
 *
 * Created by   : Nitin Jain
 *
 * Date Created : 27-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_delLunFromLnSdb(spm_lunDbCb_t *lCfg, char *err)
{
    DBT key;

    SPM_FCT_PRINT(("spm_delLunFromLnSdb..."));
    if (!lCfg)
        return NFX_ERR;
    SPM_DBG_PRINT(("\tDev. : %s", lCfg->dev));
    SPM_DBG_PRINT(("\t WWN : 0x%llx", lCfg->wwn));
    SPM_DBG_PRINT(("\t LUN : %d", lCfg->lun));
    SPM_DBG_PRINT(("\tSCSI : 0x%llx", lCfg->sid));
    SPM_DBG_PRINT(("\tSize : %lld", lCfg->size));
    SPM_DBG_PRINT(("\tRAID : %d", lCfg->rL));
    SPM_DBG_PRINT(("\tPool : %d", lCfg->poolId));
    SPM_DBG_PRINT(("\tMake : %s", lCfg->make));
    SPM_DBG_PRINT(("\tModel: %s", lCfg->model));

    key.data = lCfg->dev;
    key.size = strlen(lCfg->dev);

    if (lunDb->del(lunDb, &key, 0) == -1)
    {
        sprintf(err, "Error deleting lunDb record (%s).", strerror(errno));
        SPM_ERR_PRINT(("%s", err));
        return(NFX_ERR);
    }

    key.data = &lCfg->sid;
    key.size = sizeof(scsi_id_t);
    if (scsiDb->del(scsiDb, &key, 0) == -1)
    {
        sprintf(err, "Error deleting scsiDb record (%s).", strerror(errno));
        SPM_ERR_PRINT(("%s", err));
        return(NFX_ERR);
    }

    SPM_FREE(lCfg);  // Free lCfg

    SPM_FCT_PRINT(("...spm_delLunFromLnSdb"));
    return NFX_OK;
}

/*
 * Global functions.
 */

/*
 *------------------------------------------------------------------------------
 * Name         : spm_checkLunInLnSdb
 *
 * Description  : Return 1 if present, 0 if not present.
 *                Check lunDb if wwn or lun is non-zero
 *                Check scsiDb if sid is non-zero
 *
 * Create by    : Nitin Jain
 *
 * Date         : 27-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_checkLunInLnSdb(char *device, scsi_id_t sid)
{
    DBT   key;
    DBT   rec;
    int32 retcode = 0;

    SPM_FCT_PRINT(("spm_checkLunInLnSdb..."));
    if (device)
    {
        key.data = device;
        key.size = strlen(device);
        if (lunDb->get(lunDb, &key, &rec, 0) == 0)
            retcode = 1;
    }
    if (sid)
    {
        key.data = &sid;
        key.size = sizeof(sid);
        if (scsiDb->get(scsiDb, &key, &rec, 0) == 0)
            retcode = 1;
    }

    SPM_DBG_PRINT(("Found : %s", (retcode)?"Yes":"No"));

    SPM_FCT_PRINT(("...spm_checkLunInLnSdb"));
    return retcode;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_getScsiDb_rec
 *
 * Description  : Return if the record is present
 *
 * Create by    : Nitin Jain
 *
 * Date         : 27-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void spm_getScsiDb_rec(scsi_id_t sid, spm_lunDbCb_t **lCfg)
{
    DBT           key;
    DBT           rec;

    SPM_FCT_PRINT(("spm_getScsiDb_rec..."));
    SPM_DBG_PRINT(("\tSCSI-Id : 0x%llx", sid));
    *lCfg = NULL;
    if (sid == 0)
        return;
    key.data = &sid;
    key.size = sizeof(scsi_id_t);
    if (scsiDb->get(scsiDb, &key, &rec, 0) == 0)
    {
        memcpy(lCfg, rec.data, sizeof(lCfg)); // Copy pointer
        if (!lCfg)
            *lCfg = NULL;
        if ((*lCfg)->magic != SPM_MAGIC_STRING)
            *lCfg = NULL;
    }

    SPM_FCT_PRINT(("...spm_getScsiDb_rec"));
    return;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_getLunDb_rec
 *
 * Description  : Return if the record is present
 *
 * Create by    : Nitin Jain
 *
 * Date         : 27-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

void spm_getLunDb_rec(char *device, spm_lunDbCb_t **lCfg)
{
    DBT           key;
    DBT           rec;

    SPM_FCT_PRINT(("spm_getLunDb_rec..."));
    SPM_DBG_PRINT(("\tDevice : %s", device));
    *lCfg = NULL;
    if (!device)
        return;

    key.data = device;
    key.size = strlen(device);
    if (lunDb->get(lunDb, &key, &rec, 0) == 0)
    {
        memcpy(lCfg, rec.data, sizeof(lCfg)); // Copy pointer
        if (!lCfg)
            *lCfg = NULL;
        if ((*lCfg)->magic != SPM_MAGIC_STRING)
            *lCfg = NULL;
    }

    SPM_FCT_PRINT(("...spm_getLunDb_rec"));
    return;
}


/*
 *------------------------------------------------------------------------------
 * Name         : spm_getPoolIdForLun
 *
 * Description  : Return pool-Id for the given LUN
 *                sid == 0 means look for wwn:lun
 *                wwn:lun == 0 means look for sid
 *
 * Create by    : Nitin Jain
 *
 * Date         : 28-MAR-2002
 *
 *------------------------------------------------------------------------------
 */

int32 spm_getPoolIdForLun(char *device, scsi_id_t sid)
{
    DBT           key;
    DBT           rec;
    int32         done = -1;
    spm_lunDbCb_t *lCfg = NULL; 

    SPM_FCT_PRINT(("spm_getPoolIdForLun..."));
    if (device != NULL)
    {
        SPM_DBG_PRINT(("\tDevice : %s", device));
        key.data = device;
        key.size = strlen(device);
        if (!(lunDb->get(lunDb, &key, &rec, 0)))
        {
            memcpy(&lCfg, rec.data, sizeof(lCfg));
            if (lCfg && lCfg->magic == SPM_MAGIC_STRING)
                done = lCfg->poolId;
        }
    }
    else if (sid != 0)
    {
        SPM_DBG_PRINT(("\tSCSI-Id : 0x%llx", sid));
        key.data = &sid;
        key.size = sizeof(sid);
        if (!(scsiDb->get(scsiDb, &key, &rec, 0)))
        {
            memcpy(&lCfg, rec.data, sizeof(lCfg));
            if (lCfg && lCfg->magic == SPM_MAGIC_STRING)
                done = lCfg->poolId;
        }
    }

    SPM_DBG_PRINT(("\tPool-ID : %d", done));
    SPM_FCT_PRINT(("...spm_getPoolIdForLun"));

    return done;
}


/*
 *------------------------------------------------------------------------------
 * Name: init_lundb
 *
 * Description: Open/Initialize the physical storage configuration database
 *
 * Created by:  Greg Seibert
 *
 * Date Created: 02-MAY-2001
 *
 *------------------------------------------------------------------------------
 */

int32 init_LnSDb(void)
{
    lunDb = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR, 0666, DB_BTREE, NULL);
    if (!lunDb )
    {
        SPM_ERR_PRINT(("Error opening lunDb %s", strerror(errno)));
        return NFX_ERR;
    }

    scsiDb = dbopen (NULL, O_CREAT|O_TRUNC|O_RDWR, 0666, DB_BTREE, NULL);
    if (!scsiDb )
    {
        SPM_ERR_PRINT(("Error opening scsiDb %s", strerror(errno)));
        lunDb->close(lunDb);
        return NFX_ERR;
    }

    return NFX_OK;
}


/*
 *------------------------------------------------------------------------------
 * Name: clean_lunDb
 *
 * Description: Clean LUN storage configuration database
 *
 * Created by:  Nitin Jain
 *
 * Date Created: 04-APR-2002
 *
 *------------------------------------------------------------------------------
 */

void clean_LnSDb(void)
{
    DBT key;
    DBT rec;
    spm_lunDbCb_t *lCfg = NULL;
    int32  stat;

    stat = lunDb->seq(lunDb, &key, &rec, R_FIRST);
    while (!stat)
    {
        memcpy(&lCfg, rec.data, sizeof(lCfg));
        if (lCfg && lCfg->magic == SPM_MAGIC_STRING)
            SPM_FREE(lCfg);
        stat = lunDb->seq(lunDb, &key, &rec, R_NEXT);
    }
    lunDb->close(lunDb);
    lunDb = NULL;

    scsiDb->close(scsiDb);
    scsiDb = NULL;

    return;
}

/*----------------------------------------------------------------------------*/
