/*
 * Name:  thread-api.c
 *
 * RCS:  $Id: thread-api.c,v 1.7 2002/11/01 21:46:55 janv Exp $
 *
 * Description:  A thin wrapper around a yielding esm to provide
 *               a thread library.
 *
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#if !defined(SSC)

#include <nfx-incl.h>
#include <exports/esm-api.h>
#include <exports/timer-api.h>
#include <exports/debug.h>

#include "thread-api.h"
#include "thread.h"


/**
 * Create a thread and run the specified function
 * in it with the specified argument.  Returns a
 * non-NULL esm_handle_t for the thread on success,
 * NULL otherwise.  A thread remains in existence
 * until it returns or thread_exit() is used.
 */
esm_handle_t
thread_create(
    // Opaque userInfo to deliver to the thread.
    void            *userInfo)
{
    esm_handle_t    handle;
    int32           rc;

    rc = esmthread_sm_create(&handle, userInfo);
    if (rc != NFX_OK)
    {
        return NULL;
    }

	return handle;

} /* thread_create */


/**
 * Start a thread with the function to run in the thread,
 * once a thread is created.
 * Destroy the thread if it failed.
 */ 
int32
thread_start(
    // The handle to the thread
    esm_handle_t * handle,
    // The function to run in the thread.
    void    (*func)(esm_handle_t    handle,     // The thread's esm_handle_t.
                    void            *userInfo)) // Opaque userInfo.
{   
    esm_event_t     event;
    int32           rc;

    event.event = 0;
    event.p1.u.p = func;
    rc = esm_deliverEvent(*handle, &event); 
    if (rc != NFX_OK)
    {
        esm_destroy(*handle);
    }

    return rc;

} /* thread_start */

/**
 * Yield the cpu back to the executive.
 */
void
thread_yield(
	esm_event_t			*eventp,		// Return my waking event info here,
                                        // NULL if I don't care.
	int32				waitSecs) 		// Deliver TIMEOUT event if I sleep
                                        // longer than this, 0 to let me sleep
                                        // forever.
{
    esm_yield(eventp, waitSecs);

    // Map the esm timeout event to our timeout event number.
    if (eventp && eventp->event == ESM_THREAD_EVENT_TIMEOUT)
    {
        eventp->event = THREAD_EVENT_TIMEOUT;
    }

} /* thread_yield */


/**
 * Signal the sleeping thread to continue.
 */
static int32
thread_timeoutHandler(
    uint32  timer,
    void    *param1,
    void    *param2,
    void    *param3,
    void    *param4)
{
    esm_event_t     event;
    int32           rc;

    event.event = THREAD_EVENT_WAKEUP;
    rc = esm_deliverEvent((address_t) param1, &event);
    if (rc != NFX_OK)
    {
        panic("thread_timerHandler: event delivery failed\n");
    }

    return NFX_OK;

} /* threadTimeoutHandler */


/**
 * Sleep the thread for a while, waking it with
 * a WAKEUP event (unless another event is delivered
 * first) when the specified period has elapsed.
 * Returns NFX_ERR if unable to schedule a timer,
 * NFX_OK otherwise.
 */
int32
thread_sleep(
    esm_event_t         *eventp,        // Return my waking event info here,
                                        // NULL if I don't care.
    int32               waitMsecs)      // Wake me up after this many
                                        // milliseconds (10 msec granularity).
{
    int32        timer = TIMER_ERR_NO_TIMER_AVAILABLE, rc;
    esm_event_t  wakeupEvent ;

    if (waitMsecs > 0)
    {
        timer = eee_timerCreate(waitMsecs, EEE_TIMER_ONCE);
        if (timer == TIMER_ERR_NO_TIMER_AVAILABLE)
        {
            return NFX_ERR;
        }
    
        rc = eee_timerStart(timer, thread_timeoutHandler,
                        (void *) (address_t) esm_getCurrHandle(), NULL, NULL, NULL);
        if (rc != NFX_OK)
        {
            eee_timerDestroy(timer);
            return NFX_ERR;
        }
    }
    else
    {
        // create event 
        wakeupEvent.event = THREAD_EVENT_WAKEUP;
        wakeupEvent.p1.u.dw = 0x222;
        wakeupEvent.p2.u.dw = 0x333;

        // and send event to myself
        // rc = thread_event(esm_getCurrHandle(), &wakeupEvent);
        rc = esm_deliverEvent(esm_getCurrHandle(), &wakeupEvent);
        if (rc != NFX_OK)
        {
            panic("thread_sleep: event delivery failed\n");
            return NFX_ERR;
        }
    }
    

    // No yield timeout here, just depend on the
    // EEE timer to do its thing.
    esm_yield(eventp, 0);

    if (waitMsecs > 0)
    {
        eee_timerStop(timer);
        eee_timerDestroy(timer);
    }
    
	return NFX_OK;

} /* thread_sleep */


/**
 * Enqueue an event to a thread.  A thread can have up to
 * THREAD_EVENT_QUEUE_MAX threads awaiting delivery at any
 * given time.  (While multiple enqueued events are supported,
 * generally, a thread will be sleeping waiting for a specific
 * event, thus only a single event would normally be enqueued
 * to a thread at any given time).  Returns NFX_ERR if unable
 * to enqueue the event, otherwise NFX_OK.
 *
 * The value of event.event is user-defined, however, event
 * numbers greater than or equal to THREAD_EVENT_USER_MAX are
 * reserved.
 */
int32
thread_event(
    esm_handle_t    handle,
    esm_event_t     *eventp)
{
    if (eventp->event >= THREAD_EVENT_USER_MAX)
    {
        return NFX_ERR;
    }

    return esm_deliverEvent(handle, eventp);

} /* thread_event */

int32
thread_highPriorityEvent(
    esm_handle_t    handle,
    esm_event_t     *eventp)
{
    if (eventp->event >= THREAD_EVENT_USER_MAX)
    {
        return NFX_ERR;
    }

    return esm_deliverHighPriorityEvent(handle, eventp);

} /* thread_highPriorityEvent */


/**
 * Perform a long jump of the thread back
 * to the executive.  Use this anywhere in a
 * thread to return to the executive and destroy
 * the thread.  This function does not return.
 */
void
thread_exit()
{
    esm_longJump(esm_getCurrHandle());

} /* thread_exit */


/**
 * Create function for thread state machine instantiation.
 * We don't use it, but the state machine API needs it.
 */
int32
esmthread_create(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *smCookie)
{
    return nextState;

} /* esmthread_create */


/**
 * Start action function for thread state machine.  This is
 * where threads first come to life outside the executive.
 */
int32
esmthread_start(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *userInfo)
{
    void    (*func)() = event.p1.u.p;

    // Call the user's thread function.
    // If it returns, the thread dies.
    (func)(handle, userInfo);

    return nextState;

} /* esmthread_start */


/**
 * This should never happen.
 */
int32
esmthread_bad_state(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *userInfo)
{
    panic("esmthread_bad_state\n");

    return nextState;

} /* esmthread_bad_state */

#endif // !SSC
