/*
 *-----------------------------------------------------------------
 *
 * Name:	timer-api.c
 *
 * RCS:  $Id: timer-api.c,v 1.12 2002/11/23 15:24:18 rick Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	API functions for timer support
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */

#include <nfx-incl.h>
#include <exports/queue-api.h>
#include <exports/timer-api.h>
#include "timer.h"
#include <sys/machine-dep.h>

/*-----------------------------------------------------------------
 * Name:	eee_timerCreate()
 *
 * Description:	Creates a new timer, and returns the handle to the caller.
 *				Futher accesses to this timer will be made via the handle.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/27/01
 *
 *-----------------------------------------------------------------
 */
int32						/* returns timer handle or
							 * TIMER_ERR_NO_TIMER_AVAILABLE
							 */
eee_timerCreate(
	uint32		duration,				/* in milliseconds */
	uint32		type					/* EEE_TIMER_ONCE, EEE_TIMER_REPEAT */
)
{
	timer_block_t	* new_tmr;
	uint32			new_handle = TIMER_ERR_NO_TIMER_AVAILABLE;
	address_t		ra;

	NFX_ASSERT(count_per_usec != 0); /* We haven't initialized yet. */
	ra = (address_t)get_ra();
	if ((!duration) || (duration % TIMER_TIC_MS)) {
		panic("timerCreate must be multiple of %d ms\n", TIMER_TIC_MS);
	}

	NFX_SPIN_LOCK(&timer_spin);
	if (tmr_free_blk) {
		new_tmr = tmr_free_blk;
		tmr_free_blk = new_tmr->usr_p1;			/* get "next" pointer */

		new_handle = new_tmr->tmr_handle;
		eee_msToRTC(duration, &new_tmr->duration);
		new_tmr->type = type;
		new_tmr->owner = ra;

		eee_timerSetAllocState(new_handle, TIMER_STATE_ALLOC);
	}
	NFX_SPIN_UNLOCK(&timer_spin);
	
	return new_handle;
}	/* eee_timerCreate() */

/*-----------------------------------------------------------------
 * Name:	eee_timerStart()
 *
 * Description:	Start the timer associated with the given handle.  When the
 *				timer expires, the callback function will be called with
 *				the user defined parameters.  If the timer is a repeating
 *				timer, after the expiration, it will be reset.  If the timer
 *				is only a one-expiry timer, it will not be reset.
 *
 *				Timers are never automatically destroyed.  Applications that
 *				no longer need a created timer must explicitly destroy the
 *				timer.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/27/01
 *
 *-----------------------------------------------------------------
 */
int32						/* NfX_OK
							 * TIMER_ERR_INVALID_HANDLE
							 * TIMER_ERR_ALREADY_RUNNING
							 * TIMER_ERR_Q_ERROR
							 */
eee_timerStart(
	uint32				tmr_handle,		/* returned from timerCreate() */
	timer_callback_func callback_func,
	void		* usr_param1,			/* passed to callback function */
	void		* usr_param2,
	void		* usr_param3,
	void		* usr_param4
)
{
	timer_block_t	* tmr;
	eeeRTC_t		now;
	int rc = NFX_OK;

	if (tmr_handle >= TIMER_NUM_TIMER_BLOCKS)
		return TIMER_ERR_INVALID_HANDLE;

	NFX_SPIN_LOCK(&timer_spin);

	if (eee_timerGetAllocState(tmr_handle) != TIMER_STATE_ALLOC) {
		rc = TIMER_ERR_INVALID_HANDLE;
	}
	else if (eee_timerGetRunningState(tmr_handle) == TIMER_STATE_RUNNING) {
		rc = TIMER_ERR_ALREADY_RUNNING;
	}
	else {
		tmr = &(tmr_block[(int)tmr_handle]);

		if (tmr->tmr_handle != tmr_handle)
			panic("Error in timer handle\n");

		tmr->callback_func = callback_func;
		tmr->usr_p1 = usr_param1;
		tmr->usr_p2 = usr_param2;
		tmr->usr_p3 = usr_param3;
		tmr->usr_p4 = usr_param4;

		eee_rtc(&now);
		(void)eee_rtcAdd(&tmr->duration, &now, &tmr->tb_expiry_time);

		timer_block_add(tmr);
	}

	NFX_SPIN_UNLOCK(&timer_spin);

	return rc;

}	/* eee_timerStart() */

/*-----------------------------------------------------------------
 * Name:	eee_timerStop()
 *
 * Description:	Stop a running timer.  Stopping a timer does not destroy it,
 *				but rather just keeps it from expiring.  The handle is still
 *				valid for a subsequent timerStart() call.  To remove the handle
 *				from the system, use the timerDestroy() function.
 *
 *				The timerStop() function unhooks the timer block from the
 *				expiring lists.  'alloc' state == TIMER_STATE_ALLOC, but
 *				'running' state = TIMER_STATE_STOPPED.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/27/01
 *
 *-----------------------------------------------------------------
 */
int32						/* NFX_OK
							 * TIMER_ERR_INVALID_HANDLE
							 * TIMER_ERR_NOT_RUNNING
							 */
eee_timerStop(
	uint32		tmr_handle			/* returned from timerCreate() */
)
{
	timer_block_t	* tmr;
	int rc = NFX_OK;

	if (tmr_handle >= TIMER_NUM_TIMER_BLOCKS)
		return TIMER_ERR_INVALID_HANDLE;

	NFX_SPIN_LOCK(&timer_spin);

	if (eee_timerGetAllocState(tmr_handle) != TIMER_STATE_ALLOC) {
		rc = TIMER_ERR_INVALID_HANDLE;
	}
	else if (eee_timerGetRunningState(tmr_handle) != TIMER_STATE_RUNNING) {
		rc = TIMER_ERR_NOT_RUNNING;
	}
	else {
		tmr = &(tmr_block[tmr_handle]);

		if (tmr->tmr_handle != tmr_handle)
			panic("Error in timer handle\n");

		timer_block_del(tmr);
	}
	NFX_SPIN_UNLOCK(&timer_spin);

	return rc;

}	/* eee_timerStop() */

/*-----------------------------------------------------------------
 * Name:	eee_timerDestroy()
 *
 * Description:	removes a timer block from the system.  Further accesses to
 *				the given handle are invalid, and should cause an error code
 *				to be returned.
 *
 *				A timer cannot be destroyed if it has not been previously
 *				stopped or is no longer running (i.e. has expired).
 *
 *				Changes 'alloc' state to TIMER_STATE_FREE.
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
int32						/* NFX_OK
							 * TIMER_ERR_INVALID_HANDLE
							 * TIMER_ERR_TIMER_RUNNING
							 */
eee_timerDestroy(
	int32		tmr_handle			/* returned from timerCreate() */
)
{
	timer_block_t	* tmr;
	int rc = NFX_OK;

	if (tmr_handle >= TIMER_NUM_TIMER_BLOCKS)
		return TIMER_ERR_INVALID_HANDLE;

	NFX_SPIN_LOCK(&timer_spin);

	if (eee_timerGetAllocState(tmr_handle) != TIMER_STATE_ALLOC) {
		rc = TIMER_ERR_INVALID_HANDLE;
	}
	else if (eee_timerGetRunningState(tmr_handle) == TIMER_STATE_RUNNING) {
		rc = TIMER_ERR_TIMER_RUNNING;
	}
	else {
		eee_timerSetAllocState(tmr_handle, TIMER_STATE_FREE);

		/*
		 * Add to free list, usr_p1 = "next"
		 */
		tmr = &(tmr_block[tmr_handle]);

		if (tmr->tmr_handle != tmr_handle)
			panic("Error in timer handle\n");

		tmr->usr_p1 = tmr_free_blk;
		tmr_free_blk = tmr;
	}

	NFX_SPIN_UNLOCK(&timer_spin);

	return rc;
}	/* eee_timerDestroy() */


typedef struct tm tm;
struct tm
{
int     tm_sec;         // seconds  0..59  ..61 (leap)  /
int     tm_min;         // minutes  0..59 
int     tm_hour;        // hours    0..23
int     tm_mday;        // day of the month  1..31
int     tm_mon;         // month    0..11
int     tm_year;        // year     since 1900
int     tm_wday;        // day of the week 0..6; 0 = sunday
int     tm_yday;        // day in the year 0..365
int     tm_isdst;       // daylight saving time 0=no, -1 if not available
};
typedef uint32 time_t;

#define SECS_IN_DAY 	(24*60*60)
#define SECS_IN_HR 		(60*60)
#define SECS_IN_MIN 	60

/*
 * Number of days in each month
 */
int32 month_days[] = {
    31, 
    28, 
    31, 
    30, 
    31, 
    30, 
    31, 
    31, 
    30, 
    31, 
    30, 
    31
};

struct tm date_time ;
int32 sys_zone = 0 ; // GMT 

/* 
 * input  UTC sec since 00:00:00 on January 1, 1970, 
 *                      Coordinated Universal Time (UTC)
 * return local tm struct since 1900 
 */

struct tm *localtime (const time_t *val)
{
    int32 secs ;
    int32 secsLeft;
    int32 secsInYr;
    int32 secsInMonth;
    int32 fullYear;

    /* convert to local, because rtc works in local mode */
    secs = *val + sys_zone ;

    secsLeft = secs;

    date_time.tm_year = 70;
    date_time.tm_mon = 0;
    date_time.tm_mday = 1;
    date_time.tm_hour = 0;
    date_time.tm_min = 0;
    date_time.tm_sec = 0;

    fullYear = 1970;

    /* 
     * Increment year and adjust seconds till number
     * of seconds left is less than in the next year
     */

    while(1)
    {
        if ((fullYear % 4) == 0)
        {
            /* Leap year */
            secsInYr = 366*SECS_IN_DAY;
        }
        else
        {
            secsInYr = 365*SECS_IN_DAY;
        }

        if (secsInYr < secsLeft)
        {
            date_time.tm_year++;
            if (date_time.tm_year > 99)
                date_time.tm_year = 0;
            fullYear++;
            secsLeft -= secsInYr;
        }
        else
        {
            break;
        }
    }

	/* 
     * Increment month and adjust seconds till number
     * of seconds left is less than in the next month
     */
    while(1)
    {
        if(date_time.tm_mon == 1)
        {
            if ((fullYear % 4) == 0)
            {
                secsInMonth = 29*SECS_IN_DAY;
            }
            else
            {
                secsInMonth = month_days[date_time.tm_mon]*SECS_IN_DAY;
            }
        }
        else
        {
            secsInMonth = month_days[date_time.tm_mon]*SECS_IN_DAY;
        }

        if (secsInMonth < secsLeft)
        {
            date_time.tm_mon++;
            secsLeft -= secsInMonth;
        }
        else
        {
            break;
        }
    }	

    /* Set number of days based on seconds remaining */
    date_time.tm_mday += (secsLeft/SECS_IN_DAY);
    secsLeft = secsLeft%SECS_IN_DAY;

    /* Set hour based on seconds remaining */
    date_time.tm_hour = (secsLeft/SECS_IN_HR);
    secsLeft = secsLeft%SECS_IN_HR;

    /* Set minutes based on seconds remaining */
    date_time.tm_min = (secsLeft/SECS_IN_MIN);
    secsLeft = secsLeft%SECS_IN_MIN;

    /* Set seconds based on seconds remaining */
    date_time.tm_sec = secsLeft;

    /* Set weekday based on Number of secs since 1970, updated to local time, 
     * divide by secs/day, +4 (thursday 1 jan 1970), modulo 7: 0 = Sunday */
    date_time.tm_wday = (secs / 3600 / 24 + 4 ) % 7 ;

    date_time.tm_yday = -1 ;    /* not used */
    date_time.tm_isdst = -1 ; 

    return (&date_time) ;
}


/*
 * lib_strbuf.h - definitions for routines which use the common string buffers
 */

/*
 * Sizes of things
 */
#define	LIB_NUMBUFS	20
#define	LIB_BUFLENGTH	80

/*
 * Macro to get a pointer to the next buffer
 */
#define	LIB_GETBUF(buf) \
    do { \
        if (!lib_inited) \
            init_lib(); \
        buf = &lib_stringbuf[lib_nextbuf][0]; \
        if (++lib_nextbuf >= LIB_NUMBUFS) \
            lib_nextbuf = 0; \
    } while (0)

/*
 * lib_strbuf - library string storage
 */

/*
 * Storage declarations
 */
char lib_stringbuf[LIB_NUMBUFS][LIB_BUFLENGTH];
int lib_nextbuf;
int lib_inited = 0;

/*
 * initialization routine. 
 */
static void
init_lib(void)
{
    lib_nextbuf = 0;
    lib_inited = 1;
}

static const char *months[] = {
    "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};
static const char *days[] = {
    "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};

#define	JAN_1970	0x83aa7e80	/* 2208988800 1970 - 1900 in seconds */


/* input: seconds since  1 Jan 1970
   output "Mon Oct  8 11:21:00 GMT 2001" */

char *
eee_humandate(uint32 time)
{
    char *bp;
    struct tm *tm;
    time_t sec;
	
    LIB_GETBUF(bp);

    if (time == 0)
    {
        (void) sprintf(bp, "0") ;
    } else {
        sec = time; // - JAN_1970;
        tm = localtime (&sec);

        (void) sprintf(bp, "%s %s %2d %02d:%02d:%02d GMT %4d",
                       days[tm->tm_wday], months[tm->tm_mon], tm->tm_mday,
                       tm->tm_hour, tm->tm_min, tm->tm_sec,
                       ((tm->tm_year < 70) ? 2000 : 1900) + tm->tm_year);
    }
    return bp;
}

/* input: seconds since  1 Jan 1970
   output "Oct  8 11:21:00" */

char *
eee_humandate_short(uint32 time)
{
    char *bp;
    struct tm *tm;
    time_t sec;
	
    LIB_GETBUF(bp);

    if (time == 0)
    {
        (void) sprintf(bp, "0") ;
    } else {
        sec = time; // - JAN_1970;
        tm = localtime (&sec);

        (void) sprintf(bp, "%s %2d %02d:%02d:%02d",
                       months[tm->tm_mon], tm->tm_mday,
                       tm->tm_hour, tm->tm_min, tm->tm_sec);
    }
    return bp;
}
