/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:	timer-api.h
 *
 * RCS:  $Id: timer-api.h,v 1.20 2002/11/27 22:11:14 tonyw Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	Timer API functions
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/23/01
 *
 *-----------------------------------------------------------------
 */

#ifndef _TIMER_API_H
#define _TIMER_API_H

#include <nfx-incl.h>
#include <exports/timer-error.h>
#include <sys/machine-dep.h>
#include <sys/tx3927.h>
#include <exports/bsp-api.h>
#include <exports/cm-api.h>

extern int32 intr_mode; // Define in assembly code

#ifndef C0_COUNT
#define C0_COUNT 9
#endif

#define EEE_TIMER_ONCE			1
#define EEE_TIMER_REPEAT		2

#define TIMER_TIC_MS			10		/* granularity of tic - 10 ms */
//#define TIMER_TIC_MS			1		/* granularity of tic - 1 ms */

#define NSEC_PER_SEC  1000000000LL
#define USEC_PER_SEC  1000000
#define MSEC_PER_SEC  1000
#define USEC_PER_MSEC 1000
#define NSEC_PER_USEC 1000

#define USEC_PER_TICK (1000 * TIMER_TIC_MS)

/******************************************************************/

// How to tell time in the embedded environment

typedef struct {
	int32 sec;
	int32 usec;
} eeeGMT_t;

#define SEC_PER_MIN    60
#define SEC_PER_HOUR   (SEC_PER_MIN * 60)
#define SEC_PER_DAY    (SEC_PER_HOUR * 24)

/* 
 * input  UTC sec since 00:00:00 on January 1, 1970, 
 *                      Coordinated Universal Time (UTC)
 * output "Mon Oct  8 11:21:00 GMT 2001"
 * output "Oct  8 11:21:00"
 *
 * Todo: set the timezone: today hardcoded to 0 hours = GMT
 */
char *
eee_humandate(uint32 time);

char *
eee_humandate_short(uint32 time);


extern void watchdog_init(void);
extern void watchdog_kick(void);
extern void watchdog_enable(int32);
extern void watchdog_disable(void);
extern void watchdog_stop(void);
extern void watchdog_restart(void);
extern void watchdog_view(void);
extern int32 watchdog_cmd(int32 argc, uchar8 **argv);
#if defined(PROM)
extern int32 get_wd_state(void);
#endif

extern void eee_TimeGMT_wait_valid();

#ifdef TX3927
extern uint32 count_per_usec;
extern uint32 count_per_msec;
extern uint32 count_per_sec;
#else
extern uint64 count_per_usec;
extern uint64 count_per_msec;
extern uint64 count_per_sec;
#endif
extern eeeGMT_t sscGMTTime;
extern eeeGMT_t bootGMTTime;
extern eeeRTC_t eeeLastRTC;
#ifndef TIMER_USE_SB1250_TIMER
extern eeeRTC_t eeeLastRTCPreWrap;
#endif
extern eeeRTC_t sscRTC;
#if defined(NFX_SMP)
extern nfx_spinlock_t wait_time_lock;
extern nfx_spinlock_t rtc_spin_lock;
#endif

/*-----------------------------------------------------------------
 * Name:    eee_rtc()
 *
 * Description: Returns the real time clock since boot up.
 *              This routine cannot be called at interrupt level.
 *
 * Arguments:
 *     clock          -  This is the returned clock value
 *-----------------------------------------------------------------
 */
static inline void
eee_rtc(eeeRTC_t *clock)
{
#ifdef TIMER_USE_SB1250_TIMER

	uint64  usec, prev_usec, prev_ref;

	SPINLOCK_LOCK(&rtc_spin_lock);

	usec = TIMER_REF_SYSTEM_TIC_COUNT - timer_get_count(BSP_TIMER_SYSTEM_TIC);
	prev_usec = eeeLastRTC.val & TIMER_REF_SYSTEM_TIC_COUNT;
	prev_ref = eeeLastRTC.val >> TIMER_REF_SHIFT;

	if (prev_usec <= usec) {
		eeeLastRTC.val = (prev_ref << TIMER_REF_SHIFT) | usec;
	}
	else {
		prev_ref += 1;
		eeeLastRTC.val = (prev_ref << TIMER_REF_SHIFT) | usec;
	}

	*clock = eeeLastRTC;

	SPINLOCK_UNLOCK(&rtc_spin_lock);

#else
    uint32  count;

    // Get the new value of the fast interval timer
#ifdef TX3927
    count = *(volatile uint32 *)TX3927_TMTRR0;
#else
    count = _mips_mfc0(C0_COUNT);
#endif
    if (!intr_mode) {

#ifdef TX3927
        /* The counter for the Toshiba is only 24 bits wide.
           We need to handle the earlier wrap.
         */
#define TX3927_COUNT_MASK      0x00FFFFFF
#define TX3927_COUNT_HIGH_MASK 0xFF000000
#define TX3927_COUNT_HIGH_BIT  0x01000000
        if ((eeeLastRTC.u.low & TX3927_COUNT_MASK) <=
            (count & TX3927_COUNT_MASK)) {
            count = (eeeLastRTC.u.low & TX3927_COUNT_HIGH_MASK) | count;
        }
        else {
            // We have wrapped the count register
            count = (eeeLastRTC.u.low & TX3927_COUNT_HIGH_MASK) +
                (TX3927_COUNT_HIGH_BIT | count);
        }
#endif /* TX3927 */

        if (eeeLastRTC.u.low <= count) {
            eeeLastRTC.u.low = count;
            // Remember the value of the RTC for use in interrupt level
#ifdef TX3927
            eeeLastRTCPreWrap.u.low = eeeLastRTC.u.low;
            eeeLastRTCPreWrap.u.high = eeeLastRTC.u.high;
#else
            eeeLastRTCPreWrap = eeeLastRTC;
#endif
        }
        else
        {
            // We have wrapped around the counter.
            eeeLastRTC.u.high += 1;
            eeeLastRTC.u.low = count;
        }
#ifdef TX3927
        // Important: dummy read only, but necessary to keep timer ticking
        clock->u.low = *(volatile uint32 *)TX3927_TMTISR0;
        clock->u.low = eeeLastRTC.u.low;
        clock->u.high = eeeLastRTC.u.high;
#else
        *clock = eeeLastRTC;
#endif
    } else {

        /* It's important not to modify eeeLastRTC at interrupt level.
         * If we allow that to happen then we need to disable interrupts
         * in the above code path to avoid time moving backwards.  The
         * safest solution is to use eeeLastRTCPreWrap to check for wrap
         * checks while at interrupt level.  This will work as long as
         * the maximum time between calls at interrupt level never exceeds
         * the wrap around time ~10 seconds.
         */

        if (eeeLastRTCPreWrap.u.low <= count) {
            clock->u.high = eeeLastRTCPreWrap.u.high;
        }
        else
        {
            // We have wrapped around the counter.
            clock->u.high = eeeLastRTCPreWrap.u.high + 1;
        }
        clock->u.low = count;
    }
#endif
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcSSCTimeInitted()
 *
 * Description: This routine returns whether we have initialized the
 *              the real time clock services.
 *
 *-----------------------------------------------------------------
 */
static inline int32
eee_rtcSSCTimeInitted(void)
{
#ifdef TX3927
	return (sscRTC.u.low != 0);
#else
    return (sscRTC.val != 0);
#endif
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcSSCTimeInit()
 *
 * Description: This routine initializes the real time clock's sscTime info.
 *				This can be called multiple times if the sscTime
 *				changes.
 *
 *-----------------------------------------------------------------
 */
static inline void
eee_rtcSSCTimeInit(
	eeeGMT_t sscTime)
{
    NFX_ASSERT(intr_mode == 0);
	NFX_SPIN_LOCK(&wait_time_lock);
	eee_rtc(&sscRTC);
	sscGMTTime = sscTime;
	NFX_SPIN_UNLOCK(&wait_time_lock);
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcZero()
 *
 * Description: Returns a pointer to a zero real time clock value.
 *
 *-----------------------------------------------------------------
 */
static inline eeeRTC_t *
eee_rtcZero(void)
{
	static eeeRTC_t zTime = {0};
    return &zTime;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcInf()
 *
 * Description: Returns a pointer to a real time clock value representing
 *              infinite time.
 *
 *-----------------------------------------------------------------
 */
static inline eeeRTC_t *
eee_rtcInf(void)
{
	static eeeRTC_t iTime;
	iTime.val = (uint64)0xFFFFFFFFFFFFFFFFll;
    return &iTime;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcStart()
 *
 * Description: This starts an interval timer using the real time clock.
 *
 * Arguments:
 *     interval       -  This is the interval timer.
 *-----------------------------------------------------------------
 */
static inline void
eee_rtcStart(eeeRTC_t *clock)
{
    eee_rtc(clock);
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcReset()
 *
 * Description: This resets an interval timer.
 *
 * Arguments:
 *     interval       -  This is the interval timer.
 *-----------------------------------------------------------------
 */
static inline void
eee_rtcReset(eeeRTC_t *interval)
{
    eee_rtc(interval);
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcAdd()
 *
 * Description: This returns sums the given intervals.
 *
 * Arguments:
 *     interval1      -  A interval to sum.
 *     interval2      -  A interval to sum.
 *     sum            -  The sum
 * Return Value:
 *     NFX_OK         -  Success
 *     NFX_ERR        -  Overflow
 *-----------------------------------------------------------------
 */
static inline int32
eee_rtcAdd(eeeRTC_t *interval1,
		   eeeRTC_t *interval2,
		   eeeRTC_t *sum)
{
#ifdef TX3927
    sum->u.high = interval1->u.high + interval2->u.high;
    if (sum->u.high >= interval1->u.high) {
        sum->u.low = interval1->u.low + interval2->u.low;
        if (sum->u.low >= interval1->u.low) {
            return NFX_OK;
        }
        else {
            sum->u.high += 1;
            if (sum->u.high == 0)
                return NFX_ERR;
            else
                return NFX_OK;
        }
    }
    return NFX_ERR;
#else
    sum->val = interval1->val + interval2->val;
    if (sum->val >= interval1->val) {
        return NFX_OK;
    }
    // overflow
    return NFX_ERR;
#endif
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcCompare()
 *
 * Description: This returns compares two intervals.
 *
 * Arguments:
 *     interval1      -  A pointer to an interval.
 *     interval2      -  A pointer to an interval.
 * Return Value:
 *     0              -  They are equal
 *     1              -  The first interval is larger
 *    -1              -  The second interval is larger
 *-----------------------------------------------------------------
 */
static inline int32
eee_rtcCompare(const eeeRTC_t * interval1,
			   const eeeRTC_t * interval2)
{
#ifdef TX3927
    if (interval1->u.high == interval2->u.high) {
        if (interval1->u.low > interval2->u.low) {
            return 1;
        }
        if (interval1->u.low < interval2->u.low) {
            return -1;
        }
        return 0;
    }
    else if (interval1->u.high > interval2->u.high) {
        return 1;
    }
    else {
        return -1;
    }
#else
    if (interval1->val > interval2->val) {
        return 1;
    }
    if (interval1->val < interval2->val) {
        return -1;
    }
    return 0;
#endif
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcIncr()
 *
 * Description: This increments the given interval by the given delta time.
 *              If the operation would result in an overflow, no changes
 *				are made.
 *
 * Arguments:
 *     interval       -  A interval to increment.
 *     delta          -  A delta time to add
 * Return Value:
 *     NFX_OK         -  Success
 *     NFX_ERR        -  Overflow
 *-----------------------------------------------------------------
 */
static inline int32
eee_rtcIncr(eeeRTC_t *interval,
			const eeeRTC_t *delta)
{
    eeeRTC_t sum;

    sum.val = interval->val + delta->val;
    if (sum.val >= interval->val) {
		*interval = sum;
        return NFX_OK;
    }
    // overflow
    return NFX_ERR;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcDivMS()
 *
 * Description: This returns the given interval divided by an integer
 *				the results are in msec.
 *
 * Arguments:
 *     interval       -  A interval to divide.
 *     divisor        -  The divisor
 *-----------------------------------------------------------------
 */
static inline float
eee_rtcDivMS(eeeRTC_t *interval,
			 uint64 divisor)
{
    return((float)interval->val / ((float)divisor * count_per_msec));
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcDiv()
 *
 * Description: This returns the given interval divided by an integer.
 *				No checks for a divide by zero are done.
 *
 * Arguments:
 *     interval       -  A interval to divide, in place.
 *     divisor        -  The divisor
 *-----------------------------------------------------------------
 */
static inline void
eee_rtcDiv(eeeRTC_t *interval,
		   const uint64 divisor)
{
	NFX_ASSERT(divisor != 0);
    interval->val /= divisor;
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcSub()
 *
 * Description: This returns the difference in the intervals.
 *
 * Arguments:
 *     interval1      -  A interval.
 *     interval2      -  The interval to subtract.
 *     diff           -  The difference
 * Return Value:
 *     NFX_OK         -  Success
 *     NFX_ERR        -  Underflow
 *-----------------------------------------------------------------
 */
static inline int32
eee_rtcSub(const eeeRTC_t *interval1,
		   const eeeRTC_t *interval2,
		   eeeRTC_t *diff)
{
#ifdef TX3927
    uint32 tmp;

    if (interval1->u.high >= interval2->u.high) {
        diff->u.low = interval1->u.low - interval2->u.low;
        if (interval1->u.low >= interval2->u.low) {
            return NFX_OK;
        }
        else {
            tmp = interval1->u.high - 1;
            if (tmp == 0xffffffff) {
                diff->u.low = 0;
                return NFX_ERR;
            }
            if (tmp >= interval2->u.high) {
                return NFX_OK;
            }
            else {
                diff->u.low = 0;
                return NFX_ERR;
            }
        }
    }
    // underflow
    diff->u.low = 0;
    return NFX_ERR;
#else
    diff->val = interval1->val - interval2->val;
    if (interval1->val >= interval2->val) {
        return NFX_OK;
    }
printf("rtcSub, underflow, val1 = 0x%llx, val2 = 0x%llx\n", interval1->val, interval2->val);
    // underflow
	diff->val = 0;
    return NFX_ERR;
#endif
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcDuration()
 *
 * Description: This returns the interval since invoking eee_rtcStart
 *
 * Arguments:
 *     interval       -  This is the interval timer.
 *     duration       -  This is the returned duration.
 *-----------------------------------------------------------------
 */
static inline void
eee_rtcDuration(const eeeRTC_t *interval,
				eeeRTC_t *duration)
{
    eeeRTC_t now;
    int32 underflow;

	eee_rtc(&now);
    underflow = eee_rtcSub(&now, interval, duration);

    // Catch an underflow
    NFX_ASSERT(underflow == NFX_OK);

    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcDurationTV()
 *
 * Description: This returns the interval since invoking eee_rtcStart
 *              in timeval form.
 *
 * Arguments:
 *     interval       -  This is the interval timer.
 *     duration       -  This is the returned duration.
 *-----------------------------------------------------------------
 */
static inline void
eee_rtcDurationTV(const eeeRTC_t *interval,
				  eeeGMT_t *duration)
{
    eeeRTC_t diff;

    eee_rtcDuration(interval, &diff);
    duration->sec = 0;
	duration->usec = (int32)(diff.val / count_per_usec);

	while (duration->usec >= USEC_PER_SEC) {
		duration->usec -= USEC_PER_SEC;
        duration->sec += 1;
	}
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcDurationSec()
 *
 * Description: This returns the interval since invoking eee_rtcStart
 *              in seconds form.
 *
 * Arguments:
 *     interval       -  This is the interval timer.
 *-----------------------------------------------------------------
 */
static inline uint64
eee_rtcDurationSec(const eeeRTC_t *interval)
{
    eeeRTC_t diff;

    eee_rtcDuration(interval, &diff);
    return(diff.val / count_per_sec);
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcGMT()
 *
 * Description: Returns the current GMT.
 *
 * Arguments:
 *     clock          -  This is the returned clock value
 *-----------------------------------------------------------------
 */
static inline void
eee_rtcGMT(eeeGMT_t *clock)
{
    eeeGMT_t duration;

    // sscGMTTime roughly corresponds to sscRTC.
#ifdef NFX_DEBUG_TIME_USE
	NFX_ASSERT(eee_rtcSSCTimeInitted());
#endif /* NFX_DEBUG_TIME_USE */

	NFX_SPIN_LOCK(&wait_time_lock);
    eee_rtcDurationTV(&sscRTC, &duration);
    clock->sec = sscGMTTime.sec + duration.sec;
    clock->usec = sscGMTTime.usec + duration.usec;
	while (clock->usec >= USEC_PER_SEC) {
		clock->usec -= USEC_PER_SEC;
        clock->sec += 1;
	}
	NFX_SPIN_UNLOCK(&wait_time_lock);
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcToTV()
 *
 * Description: This converts a real time clock interval into
 *              timeval form.
 *
 * Arguments:
 *     interval       -  This is the interval in rtc form.
 *     duration       -  This is the returned duration in timeval form.
 *-----------------------------------------------------------------
 */
static inline void
eee_rtcToTV(const eeeRTC_t *interval,
            eeeGMT_t *duration)
{
    duration->sec = 0;
	duration->usec = (int32)(interval->val / count_per_usec);

	while (duration->usec >= USEC_PER_SEC) {
		duration->usec -= USEC_PER_SEC;
        duration->sec += 1;
	}
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcToGMT()
 *
 * Description: Returns the time in GMT for the given real time clock
 *				value.
 *
 * Arguments:
 *     rtcClock       -  The clock value in rtc form
 *     clock          -  This is the returned clock value
 *-----------------------------------------------------------------
 */
static inline void
eee_rtcToGMT(const eeeRTC_t *rtcClock,
			 eeeGMT_t *clock)
{
    eeeGMT_t duration;
	eeeRTC_t delta;

    /* sscGMTTime roughly corresponds to sscRTC.
	 * Note that sscRTC always moves forward but sscGMTTime may move backwards
	 * in response to an update from the SSC.  Since sscRTC can move forwards,
	 * it is possible that rtcClock is less than sscRTC.
	 */
	NFX_SPIN_LOCK(&wait_time_lock);
#ifdef NFX_DEBUG_TIME_USE
	NFX_ASSERT(eee_rtcSSCTimeInitted());
#endif /* NFX_DEBUG_TIME_USE */
    if (eee_rtcCompare(rtcClock, &sscRTC) >= 0) {
		eee_rtcSub(rtcClock, &sscRTC, &delta);
		eee_rtcToTV(&delta, &duration);
		clock->sec = sscGMTTime.sec + duration.sec;
		clock->usec = sscGMTTime.usec + duration.usec;
		while (clock->usec >= USEC_PER_SEC) {
			clock->usec -= USEC_PER_SEC;
			clock->sec += 1;
		}
	} else {
		eee_rtcSub(&sscRTC, rtcClock, &delta);
		eee_rtcToTV(&delta, &duration);
		clock->sec = sscGMTTime.sec - duration.sec;
		if (sscGMTTime.usec >= duration.usec) {
			clock->usec = sscGMTTime.usec - duration.usec;
		} else {
			--clock->sec;
			clock->usec = (USEC_PER_SEC + sscGMTTime.usec) - duration.usec;
		}
	}
	NFX_SPIN_UNLOCK(&wait_time_lock);
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_TimeGMT()
 *
 * Description: Returns the current GMT time in epoch since 1 Jan 1970.
 *
 *-----------------------------------------------------------------
 */
static inline void
eee_TimeGMT (eeeGMT_t *t)
{
    eee_rtcGMT(t);
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcStr()
 *
 * Description: Returns the human readable date for the given real
 *				time clock value.
 *
 * Arguments:
 *     clock          -  An optional rtc value.  Current time is
 *						 the default.
 * Return Value:
 *     date string    -  if sscTime is available
 *     "unknown"      -  if sscTime is not available
 *
 *-----------------------------------------------------------------
 */
static inline char *
eee_rtcStr(eeeRTC_t *clock)
{
	if (eee_rtcSSCTimeInitted()) {
		eeeGMT_t now;

		if (clock == NULL) {
			eee_rtcGMT(&now);
		} else {
			eee_rtcToGMT(clock, &now);
		}
		return eee_humandate(now.sec);
	} else {
		return "0";
	}
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcToUsec()
 *
 * Description: This converts a real time clock interval into
 *              usec form.
 *
 * Arguments:
 *     interval       -  This is the interval in rtc form.
 *-----------------------------------------------------------------
 */
static inline uint64
eee_rtcToUsec(const eeeRTC_t *interval)
{
    return (interval->val / count_per_usec);
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcToMS()
 *
 * Description: This converts a real time clock interval into
 *              ms form.
 *
 * Arguments:
 *     interval       -  This is the interval in rtc form.
 *-----------------------------------------------------------------
 */
static inline uint64
eee_rtcToMS(const eeeRTC_t *interval)
{
    return (interval->val / count_per_msec);
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcToSec()
 *
 * Description: This converts a real time clock interval into
 *              seconds form.
 *
 * Arguments:
 *     interval       -  This is the interval in rtc form.
 *-----------------------------------------------------------------
 */
static inline uint64
eee_rtcToSec(const eeeRTC_t *interval)
{
    return (interval->val / count_per_sec);
}

/*-----------------------------------------------------------------
 * Name:    eee_tvToRTC()
 *
 * Description: This converts a timeval into
 *              real time clock interval form.
 *
 * Arguments:
 *     interval       -  This is the interval in timeval form.
 *     duration       -  This is the returned duration in rtc form.
 *-----------------------------------------------------------------
 */
static inline void
eee_tvToRTC(const eeeGMT_t *interval,
            eeeRTC_t *duration)
{
    duration->val = ((uint64)interval->usec +
					 ((uint64)interval->sec * USEC_PER_SEC)) * count_per_usec;
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_usecToRTC()
 *
 * Description: This converts a number of usec into
 *              real time clock interval form.
 *
 * Arguments:
 *     interval       -  This is the interval in usec form.
 *     duration       -  This is the returned duration in rtc form.
 *-----------------------------------------------------------------
 */
static inline void
eee_usecToRTC(const uint64 interval,
			  eeeRTC_t *duration)
{
	duration->val = interval * count_per_usec; 
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_msToRTC()
 *
 * Description: This converts a number of ms into
 *              real time clock interval form.
 *
 * Arguments:
 *     interval       -  This is the interval in ms form.
 *     duration       -  This is the returned duration in rtc form.
 *-----------------------------------------------------------------
 */
static inline void
eee_msToRTC(const uint32 interval,
			eeeRTC_t *duration)
{
	duration->val = interval * count_per_msec; 
    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_rtcInit()
 *
 * Description: Inits the real time clock.
 *
 *-----------------------------------------------------------------
 */
static inline void
eee_rtcInit(void)
{
#ifdef TX3927
    eeeLastRTC.u.low = 0;
    eeeLastRTC.u.high = 0;
    eeeLastRTCPreWrap.u.low = eeeLastRTC.u.low;
    eeeLastRTCPreWrap.u.high = eeeLastRTC.u.high;
#else
	eeeLastRTC.val = 0;
#ifndef TIMER_USE_SB1250_TIMER
	eeeLastRTCPreWrap = eeeLastRTC;
#else
	if ((nfxMyCPUId == NFX_NFP_TXRX1) || (nfxMyCPUId == NFX_NFP_FP1)) {
		NFX_SPIN_INIT(&rtc_spin_lock, SPINLOCK_LEVEL_UNKNOWN);
		timer_start_us(BSP_TIMER_SYSTEM_TIC, TIMER_REF_SYSTEM_TIC_COUNT, NULL);
	}
#endif
#endif // TX3927

    return;
}

/*-----------------------------------------------------------------
 * Name:    eee_timerGetCurrTic()
 *
 * Description: Returns the running count of clock ticks since boot up.
 *              Units are centiseconds (100 Hz timer and all).
 *
 *-----------------------------------------------------------------
 */
extern uint32 				/* 1/100 seconds */
eee_timerGetCurrTic(void);

/*-----------------------------------------------------------------
 * Name:	eee_timerCreate()
 *
 * Description:	Creates a new timer, and returns the handle to the caller.
 *				Futher accesses to this timer will be made via the handle.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/27/01
 *
 *-----------------------------------------------------------------
 */
int32						/* returns timer handle or
							 * TIMER_ERR_NO_TIMER_AVAILABLE
							 */
eee_timerCreate(
	uint32		duration,				/* in milliseconds */
	uint32		type					/* EEE_TIMER_ONCE, EEE_TIMER_REPEAT */
);

/*-----------------------------------------------------------------
 * Name:	eee_timerStart()
 *
 * Description:	Start the timer associated with the given handle.  When the
 *				timer expires, the callback function will be called with
 *				the user defined parameters.  If the timer is a repeating
 *				timer, after the expiration, it will be reset.  If the timer
 *				is only a one-expiry timer, it will not be reset.
 *
 *				Timers are never automatically destroyed.  Applications that
 *				no longer need a created timer must explicitly destroy the
 *				timer.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/27/01
 *
 *-----------------------------------------------------------------
 */
typedef int32 (*timer_callback_func)(
	uint32	tmr_handle,
	void	* usr_p1,
	void	* usr_p2,
	void	* usr_p3,
	void	* usr_p4
	);

int32						/* NfX_OK
							 * TIMER_ERR_INVALID_HANDLE
							 * TIMER_ERR_ALREADY_RUNNING
							 */
eee_timerStart(
	uint32		tmr_handle,			/* returned from timerCreate() */
	timer_callback_func callback_func,
	void		* usr_param1,		/* passed to callback function */
	void		* usr_param2,
	void		* usr_param3,
	void		* usr_param4
);

/*-----------------------------------------------------------------
 * Name:	eee_timerStop()
 *
 * Description:	Stop a running timer.  Stopping a timer does not destroy it,
 *				but rather just keeps it from expiring.  The handle is still
 *				valid for a subsequent timerStart() call.  To remove the handle
 *				from the system, use the timerDestroy() function.
 *
 * Created by:	Rick Lund
 *
 * Date Created:	5/27/01
 *
 *-----------------------------------------------------------------
 */
int32						/* NFX_OK
							 * TIMER_ERR_INVALID_HANDLE
							 * TIMER_ERR_NOT_RUNNING
							 */
eee_timerStop(
	uint32		tmr_handle			/* returned from timerCreate() */
);

/*-----------------------------------------------------------------
 * Name:	eee_timerDestroy()
 *
 * Description:	removes a timer block from the system.  Further accesses to
 *				the given handle are invalid, and should cause an error code
 *				to be returned.
 *
 * Created by:	Rick Lund
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
int32						/* NFX_OK
							 * TIMER_ERR_INVALID_HANDLE
							 */
eee_timerDestroy(
	int32		tmr_handle			/* returned from timerCreate() */
);

extern void timer_showStats(void);
extern void eee_timerShowTic(void);

#endif	/* _TIMER_API_H */

