/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: tpl-alloc.c
 *
 * RCS:  $Id: tpl-alloc.c,v 1.10 2002/10/22 22:36:39 swang Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Transport Layer PKT allocation routines.
 *
 * Created by:	Shaun Hui
 *
 * Date Created: 07/01/2001
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <exports/pkt-api.h>
#include <exports/tpl-api.h>
#include "tpl.h"


/*-----------------------------------------------------------------
 * Name:        tpl_allocPkt
 *
 * Description: Allocates pkt from free list associated with connection.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 7/01/01
 *-----------------------------------------------------------------
 */
pkt_desc_t * 		
tpl_allocPkt( void *pConnId , int32 usrDataLen, int type, char *filename, int32 linenum)

{
	tpl_connCb_t	*pConn;
	tpl_stats_t		*pStats;
	pkt_descRing_t	*pRing;
	pkt_desc_t		*pDesc;

	pStats = &pTplCfg->tplStats;
	pConn  = (tpl_connCb_t *)pConnId;


	/*-------------------------------
	 * Check if conn CB is valid.
	 *-------------------------------
	 */
	if ((pConn == NULL) || !(pConn->flags & TPL_CONN_INUSE))
	{
		pStats->cntBadConnErr++;
		return ( NULL);
	}

	/* Su, 08/31/02
	 * allocate pkt from within tpl conn,
	 * if requested buffer not ultra
	 */
	if ((type == TPL_PKT_META) && (usrDataLen > EEE_BUFSIZE_LARGE))
		goto slowway;

#if defined(NFX_SMP)
	/* NCPU, use slowpath to allocate pkts
	 * this is to avoid conflict between NCPU and 
	 * ACPU on connection descriptor ring
	 */
	if (CM_IS_TXRX_NCPU)
		goto slowway;
#endif /* NFX_SMP */	
	
	pRing = &(pConn->txPktRing);
	if (pRing->avail > 0) {
		pDesc = *pRing->rdAddr;
		pRing->avail--;
		pStats->cntAllocPkt++;
		PKT_RING_UPDATE_RD_PTR(pRing);
		pConn->cntOutCons = 0;
		return pDesc;
	}
	
slowway:
	// printf("slow alloc ring size %d\n", pRing->avail);
	
	if (type == TPL_PKT_META) {
		
		/*-------------------------------
		 * su, 06/17/02, 
		 * use userDataLen to identify
		 * types of buffer should be 
		 * allocated. Note: tpl doesn't
		 * allocate pkt chains.
		 *-------------------------------
		 */
		if (usrDataLen < EEE_BUFSIZE_LARGE)
			pDesc = PKT_ALLOC_PACKET( PKT_USR_TPL);
		else if (usrDataLen < EEE_BUFSIZE_ULTRA)
			pDesc = PKT_ALLOC_PKT_BYTYPE( PKT_USR_TPL,
										  EEE_BUF_ULTRA);
		else
		{
				
				/*
				 * requested size is too big
				 */
			pDesc = NULL;
		}
		// printf("meta buffer allocated %lx\n", 
		// (address_t)pDesc);
		
	}
	else
	{
	/*-------------------------------
	 * Make sure conn has enough pkts
	 * to cover the segment/fragment
	 * needed to send all USER DATA.
	 *-------------------------------
	 */
		if (usrDataLen)
		{
			int32 cnt;
			
			cnt = (usrDataLen >> 10) + 1;
			
#ifdef TPL_DBG
			printf("%s: %s line=%d dataSize=%d, cnt=%d avail=%d Result:%s\n", 
				   __FUNCTION__, filename, linenum,
				   usrDataLen, cnt, pRing->avail, 
				   (cnt >= pRing->avail) ? "NO" : "YES");
#endif
			
			pRing  = &pPktCfg->freeDescRing;
			if (cnt >= pRing->avail)
				return NULL;
				
		}
	
		/*-------------------------------
		 * Allocate PKT
		 *-------------------------------
		 */
		pDesc =  PKT_ALLOC_PACKET( PKT_USR_TPL);
	}
	
	
	if (pDesc == NULL)
	{
		pStats->cntNoPktRsrc++;
		return ( NULL);
	}

	/*-------------------------------
	 * Setup PKT headers.
	 *-------------------------------
	 */
	tpl_connSetupPktHdrs(pConn, pDesc );

	/*-------------------------------
	 * Set pkt usr flag and inc. count.
	 *-------------------------------
	 */
	pDesc->flags |= (PKT_DESC_INUSE | PKT_USR_TPL | PKT_FAST_PATH);
	pDesc->pNext = NULL;
	pDesc->pTplNext  = NULL;
	pDesc->pRefDesc  = NULL;
	pDesc->tplProt   = pConn->prot;
	pDesc->allocFile = filename;
	pDesc->allocLine = linenum;

	/* Su, assign connId to pkt */
	pDesc->tplConnId = pConn->connId;
	
	pStats->cntAllocPkt++;

	/* Su, 08/31/02
	 * cnt allocation from pkt pool
	 * we should aim to minimize this count, 
	 * and at the same time constrain buffer
	 * usage.
	 */ 
	pStats->cntOutConnAlloc++;
	pConn->cntOutCons ++;
	
	if (pConn->cntOutCons > pStats->cntOutHWater)
		pStats->cntOutHWater = pConn->cntOutCons;
		
	
	pDesc->link = pConn->linkNum;

	/*-------------------------------
	 * For TCP connections, 
	 * adjust app header to account
	 * for 4 bytes of RPC Marker.
	 *-------------------------------
	 */
	if ((pConn->prot  == TPL_PROT_TCP)  &&
		(pConn->flags &  TPL_CONN_USE_RPC)) 
	{
		pDesc->pHdr += 4;
	}

	return (pDesc);
}
