/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: tpl-conn.c
 *
 * RCS:  $Id: tpl-conn.c,v 1.51 2002/11/01 23:12:12 janv Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Transport Layer Connection CB related routines.
 *
 * Created by:	Shaun Hui
 *
 * Date Created: 07/15/2001
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>
#include <exports/pkt-api.h>
#include <exports/luc-api.h>
#include <exports/tpl-api.h>
#include "../sm-bsd-sys/flags.h"
#include "../sm-bsd-sys/cdefs.h"
#include "../sm-bsd-sys/types.h"
#include "../sm-bsd-sys/socket.h"
#include "../sm-bsd-sys/socketvar.h"
#include "../sm-bsd-net/route.h"
#include "../sm-bsd-netinet/in.h"
#include "../sm-bsd-netinet/in_pcb.h"
#include "../sm-bsd-netinet/ip.h"
#include "../sm-bsd-netinet/tcp.h"
#include "../sm-bsd-netinet/udp.h"
#include "tpl.h"
#include <exports/timer-api.h>
#include <exports/luc-pkthdr.h>
#include "../sm-bsd-api/bsd-vif.h"

#define TPL_CONN_PRINTF(a)	    /* printf a */
#define	TPL_REQ				1			/* Upper Layer Request */
#define	TPL_IND				0			/* Lower Layer Indication */

/*-----------------------------------------------------------------
 * Name:		tpl_connSetupEthHdrs
 *
 * Description: Fill in EtherType header in packet buffer. 
 *
 * Created by : Shaun Hui
 *
 * Date Created: 7/15/01
 *-----------------------------------------------------------------
 */
static int32
tpl_connSetupEthHdr( tpl_connCb_t *pConn, char **pBuf )
{
	pkt_ethHdr_t	*ethHdr;

	/*-------------------------------
	 * Move ptr back by sizeof ether hdr.
	 *-------------------------------
	 */
	*pBuf  -= PKT_ETHER_HDR_SIZE;
	ethHdr  = (pkt_ethHdr_t *) (*pBuf);

	/*-------------------------------
	 * Fill in Src & Dst Mac & type. 
	 *-------------------------------
	 */
    memcpy6(&ethHdr->srcMac[0], &pConn->srcMac[0], 6);
    memcpy6(&ethHdr->dstMac[0], &pConn->dstMac[0], 6);

	ethHdr->type	  = htons(PKT_TYPE_IP);

	return (PKT_ETHER_HDR_SIZE);
}

/*-----------------------------------------------------------------
 * Name:		tpl_connSetupIeeeHdrs
 *
 * Description: Fill in IEEE 802.3 header in packet buffer. 
 *
 * Created by : Shaun Hui
 *
 * Date Created: 7/15/01
 *-----------------------------------------------------------------
 */
static int32
tpl_connSetupIeeeHdr( tpl_connCb_t *pConn, char **pBuf )
{
	pkt_ieeeHdr_t	*ieeeHdr;

	/*-------------------------------
	 * Move ptr back by sizeof 802.3 hdr.
	 *-------------------------------
	 */
	*pBuf  -= PKT_IEEE_HDR_SIZE;
	ieeeHdr = (pkt_ieeeHdr_t *) (*pBuf);

	/*-------------------------------
	 * Fill in Src & Dst Mac Addr.
	 *-------------------------------
	 */
    memcpy6(&ieeeHdr->srcMac[0], &pConn->srcMac[0], 6);
    memcpy6(&ieeeHdr->dstMac[0], &pConn->dstMac[0], 6);

	/*-------------------------------
	 * Fill in SNAP header.
	 *-------------------------------
	 */
	ieeeHdr->len = 0;
	ieeeHdr->dsap = 0xAA;
	ieeeHdr->ssap = 0xAA;
	ieeeHdr->ctrl = 0x03;
	ieeeHdr->oui[0] = 0x00;
	ieeeHdr->oui[1] = 0x00;
	ieeeHdr->oui[2] = 0x00;
	ieeeHdr->type = htons(PKT_TYPE_IP);
				
	return (PKT_IEEE_HDR_SIZE);
}

/*-----------------------------------------------------------------
 * Name:		tpl_connSetupVlanHdrs
 *
 * Description: Fill in VLAN header in packet buffer. 
 *
 * Created by : Shaun Hui
 *
 * Date Created: 7/15/01
 *-----------------------------------------------------------------
 */
static int32
tpl_connSetupVlanHdr( tpl_connCb_t *pConn, char **pBuf )
{
	pkt_vlanHdr_t	*vlanHdr;
	
	/*-------------------------------
	 * Move ptr back by sizeof VLAN hdr.
	 *-------------------------------
	 */
	*pBuf  -= PKT_VLAN_HDR_SIZE;
	vlanHdr  = (pkt_vlanHdr_t *) (*pBuf);

	/*-------------------------------
	 * Fill in Src & Dst Mac Addr.
	 *-------------------------------
	 */
    memcpy6(&vlanHdr->srcMac[0], &pConn->srcMac[0], 6);
    memcpy6(&vlanHdr->dstMac[0], &pConn->dstMac[0], 6);

	/*-------------------------------
	 * Fill in VLAN TPID & VLAN ID.
	 *-------------------------------
	 */
	vlanHdr->tpid = htons(PKT_TYPE_TPID);
	vlanHdr->vid  = htons(pConn->vlanId);
	vlanHdr->type = htons(PKT_TYPE_IP);

	return (PKT_VLAN_HDR_SIZE);
}

/*-----------------------------------------------------------------
 * Name:		tpl_connSetupMacHdrs
 *
 * Description: Fill in MAC header in packet buffer. 
 *
 * Created by : Shaun Hui
 *
 * Date Created: 7/15/01
 *-----------------------------------------------------------------
 */
static int32
tpl_connSetupMacHdr( tpl_connCb_t *pConn, char **pBuf )
{
	int32	hlen;
	
	/*-------------------------------
	 * Fill MAC hdr based on link type.
	 *-------------------------------
	 */
	switch(pConn->linkType)
	{
		case	PKT_LINK_TYPE_IEEE:
				hlen = tpl_connSetupIeeeHdr( pConn, pBuf );
				break;

		case	PKT_LINK_TYPE_VLAN:
				hlen = tpl_connSetupVlanHdr( pConn, pBuf );
				break;

		case	PKT_LINK_TYPE_ETHER:
		default:
				hlen = tpl_connSetupEthHdr( pConn, pBuf );
				break;

	}	

	return (hlen);
}

/*-----------------------------------------------------------------
 * Name:		tpl_connSetupIpHdrs
 *
 * Description: Fill in IP header in packet buffer. 
 *				Assumes no IP options.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 7/15/01
 *-----------------------------------------------------------------
 */
static int32
tpl_connSetupIpHdr( tpl_connCb_t *pConn, char **pBuf )
{

	struct	ip 	*ipHdr;

	/*-------------------------------
	 * Move ptr back by sizeof IP hdr.
	 *-------------------------------
	 */
	*pBuf -= PKT_IP_HDR_SIZE;
	ipHdr  = (struct ip *)(*pBuf);

	/*-------------------------------
	 * Fill in IP header addresses.
	 *-------------------------------
	 */
	ipHdr->ip_v   = PKT_IP_VER;
	ipHdr->ip_hl  = PKT_IP_HDR_SIZE >> 2;
	ipHdr->ip_tos = 0;
	ipHdr->ip_len = 0;
	ipHdr->ip_id  = 0;
	ipHdr->ip_off = 0;
	ipHdr->ip_ttl = IPDEFTTL;
	ipHdr->ip_p   = pConn->prot;
	ipHdr->ip_sum = 0;
	ipHdr->ip_src.s_addr = (pConn->laddr);
	ipHdr->ip_dst.s_addr = (pConn->raddr);
	
	return (PKT_IP_HDR_SIZE);
}

/*-----------------------------------------------------------------
 * Name:		tpl_connSetupTcpHdrs
 *
 * Description: Fill in TCP header in packet buffer. 
 *				Assumes no TCP options.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 7/15/01
 *-----------------------------------------------------------------
 */
static int32
tpl_connSetupTcpHdr( tpl_connCb_t *pConn, char **pBuf )
{
	struct	tcphdr 	*tcpHdr;

	/*-------------------------------
	 * Move ptr back by sizeof TCP hdr.
	 *-------------------------------
	 */
	*pBuf  -= PKT_TCP_HDR_SIZE;
	tcpHdr  = (struct tcphdr *)(*pBuf);
	
	/*-------------------------------
	 * Fill in TCP header ports.
	 *-------------------------------
	 */
	tcpHdr->th_sport = (pConn->lport);
	tcpHdr->th_dport = (pConn->rport);
	tcpHdr->th_seq   = 0;
	tcpHdr->th_ack   = 0;
	tcpHdr->th_off   = 0;
	tcpHdr->th_flags = 0;
	tcpHdr->th_win   = 0;
	tcpHdr->th_sum   = 0;
	tcpHdr->th_urp   = 0;

	return (PKT_TCP_HDR_SIZE);
}

/*-----------------------------------------------------------------
 * Name:		tpl_connSetupUdpHdrs
 *
 * Description: Fill in UDP header in packet buffer.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 7/15/01
 *-----------------------------------------------------------------
 */
static int32
tpl_connSetupUdpHdr( tpl_connCb_t *pConn, char **pBuf )
{
	struct	udphdr 	*udpHdr;

	/*-------------------------------
	 * Move ptr back by sizeof UDP hdr.
	 *-------------------------------
	 */
	*pBuf  -= PKT_UDP_HDR_SIZE;
	udpHdr  = (struct udphdr *)(*pBuf);
	
	/*-------------------------------
	 * Fill in UDP header ports.
	 *-------------------------------
	 */
	udpHdr->uh_sport = (pConn->lport);
	udpHdr->uh_dport = (pConn->rport);
	udpHdr->uh_ulen  = 0;
	udpHdr->uh_sum   = 0;

	return (PKT_UDP_HDR_SIZE);
}

/*-----------------------------------------------------------------
 * Name:		tpl_connSetupPktHdrs
 *
 * Description: Fill in MAC to UDP/TCP headers in packet list 
 *				associated with a TPL Connection.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 7/15/01
 *-----------------------------------------------------------------
 */
void
tpl_connSetupPktHdrs( tpl_connCb_t *pConn, pkt_desc_t *pDesc)
{
	char 	*pBuf;
	char 	*macHdr;
	char 	*ipHdr;
	char 	*tplHdr;
	char 	*sesHdr;
	int32	hlen;
	

	/*-------------------------------
	 * Indent to make room for headers &
	 * align upper layer at word boundary.
	 *-------------------------------
	 */
	pBuf  = (char *)((address_t) pDesc->pHdr & (~0xFF));
	pBuf += PKT_HDR_INDENT;

	/*-------------------------------
	 * Upper layer fills in at pHdr.
	 *-------------------------------
	 */
	pDesc->pHdr = pBuf;	
	sesHdr = pBuf;

	/*-------------------------------
	 * Setup UDP/TCP layer first.
	 *-------------------------------
	 */
	if (pConn->prot == TPL_PROT_UDP)
		hlen = tpl_connSetupUdpHdr( pConn, &pBuf );
	else
		hlen = tpl_connSetupTcpHdr( pConn, &pBuf );
	tplHdr = pBuf;

	/*-------------------------------
	 * Setup IP layer next.
	 *-------------------------------
	 */
	hlen += tpl_connSetupIpHdr( pConn, &pBuf );
	ipHdr = pBuf;

	/*-------------------------------
	 * Setup MAC layer last.
	 *-------------------------------
	 */
	hlen += tpl_connSetupMacHdr( pConn, &pBuf );
	macHdr= pBuf;

	/*-------------------------------
	 * Setup mac, ip & tpl hdr ptrs. 
	 *-------------------------------
	 */
	pDesc->pMac   = macHdr;
	pDesc->pIpHdr = ipHdr;
	pDesc->pTplHdr= tplHdr;

	/*-------------------------------
	 * Sum of MAC to TPL header size.
	 *-------------------------------
	 */
	pDesc->hdrSize = hlen;
	
	/*-------------------------------
	 * Remember the Connection.
	 *-------------------------------
	 */
	pDesc->pTplConn = pConn;
	pDesc->tplProt  = pConn->prot;
}

#if defined(LINK_AGG)
/*-----------------------------------------------------------------
 * Name:		tpl_UpdatePkt
 *
 * Description: 
 *
 * Created by : Shaun Hui
 *
 * Date Created: 7/15/01
 *-----------------------------------------------------------------
 */
void
tpl_UpdatePkt(pkt_desc_t *pDesc, tpl_connCb_t  *pConn, uint32 flags)
{

	pkt_ethHdr_t	*ethHdr;
	tpl_stats_t		*pStats = &pTplCfg->tplStats;
	
		
	if ((flags & TPL_PRING_CHGE_SRCMAC) &&
		!(pConn->flags & TPL_CONN_VIF_LA)) {
		ethHdr = (pkt_ethHdr_t*)pDesc->pMac;
		memcpy(&ethHdr->srcMac[0], &pConn->srcMac[0], 6);
	}

	pDesc->link = pConn->linkNum;
	pStats->cntPktLinkChge ++;
	
	return;
}
#endif /* LINK_AGG */

/*-----------------------------------------------------------------
 * Name:		tpl_connAllocPktRing
 *
 * Description: Allocate Pkt Ring associated with a TPL Connection.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 7/15/01
 *-----------------------------------------------------------------
 */
static int32
tpl_connAllocPktRing( tpl_connCb_t *pConn, int32 pktCnt )
{
	pkt_descRing_t	*pRing;
	pkt_desc_t		*pDesc;

	/*-------------------------------
	 * set up per connection tx pkt ring.
	 *-------------------------------
	 */
	pRing = &pConn->txPktRing;
	pRing->startAddr = &pConn->txPktEntry[0];
	pRing->rdAddr    = pRing->startAddr; 
	pRing->wrAddr    = pRing->startAddr; 
	pRing->endAddr   = &pConn->txPktEntry[pktCnt-1];
	pRing->avail	 = 0;
	
	/*-------------------------------
	 * Allocate free tx pkt.
	 *-------------------------------
	 */
	while (pRing->wrAddr != pRing->endAddr)
    {
		pDesc =  PKT_ALLOC_PACKET( PKT_USR_TPL);
		if (!pDesc)
			break;

		/*-------------------------------
		 * write pkt to wr ptr. update avail.
		 *-------------------------------
		 */
		*(pRing->wrAddr++) = pDesc;
		pRing->avail++;

		/*-------------------------------
		 * Setup local & remote ips & ports
		 *-------------------------------
		 */
		pDesc->laddr = pConn->laddr;
		pDesc->raddr = pConn->raddr;
		pDesc->lport = pConn->lport;
		pDesc->rport = pConn->rport;

		/*-------------------------------
		 * Setup desc flags & pkt header.
		 *-------------------------------
		 */
		tpl_connSetupPktHdrs(pConn, pDesc );
		pDesc->flags |= (PKT_DESC_INUSE | PKT_USR_TPL | PKT_FAST_PATH | PKT_USR_TPL_CONN);
		pDesc->pNext = NULL;
		pDesc->pTplNext  = NULL;
		pDesc->pRefDesc  = NULL;
		pDesc->tplProt   = pConn->prot;
		pDesc->numOfCod  = 0;
		pDesc->allocFile = __FILE__;
		pDesc->allocLine = __LINE__;
		pDesc->link = pConn->linkNum;

		/* Su, assign connection id to pkt */
		pDesc->tplConnId = pConn->connId;
		

		/*-------------------------------
		 * For TCP connections, 
		 * adjust app header to account
		 * for 4 bytes of RPC Marker.
		 *-------------------------------
		 */
		if ((pConn->prot  == TPL_PROT_TCP)  &&
			(pConn->flags &  TPL_CONN_USE_RPC)) {
			pDesc->pHdr += 4;
		}
	}
	
	if (pRing->avail == 0)
		return (TPL_ERR_NO_PKT_RSRC);

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:		tpl_connFreePktRing
 *
 * Description: Free Pkt Ring associated with a TPL Connection.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 7/15/01
 *-----------------------------------------------------------------
 */
static int32
tpl_connFreePktRing( tpl_connCb_t *pConn )
{
	pkt_descRing_t	*pRing;
	pkt_desc_t		*pDesc;
	int32			rc = 0;

	pRing = &pConn->txPktRing;

	TPL_CONN_PRINTF(("%s: lport=%x rport=%d avail=%d\n",
        __FUNCTION__, ntohs(pConn->lport), ntohs(pConn->rport), pRing->avail));


	while (pRing->rdAddr != pRing->wrAddr)
	{
		pDesc = *(pRing->rdAddr);
		pDesc->flags |= PKT_DESC_INUSE;

		rc = PKT_FREE_PACKET( *(pRing->rdAddr) );
		PKT_RING_UPDATE_RD_PTR(pRing);
	}

	pRing->avail = 0;
		
	return (rc);
}

/*-----------------------------------------------------------------
 * Name:		tpl_addConnInd
 *
 * Description: Called by BSD UDP/TCP layer to add a new connection.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 7/15/01
 *-----------------------------------------------------------------
 */
static int32	/* NFX_OK, or err */
tpl_addConnInd( ushort16 	prot, 	/* protocol UDP/TCP */
				uint32		laddr, 	/* local  IP address */
				uint32		raddr,	/* remote IP address */
				ushort16 	lport,	/* local  Port # */
				ushort16 	rport,	/* remote Port # */
				void		*pInp,	/* ptr to in_pcb (bsd code)*/
				void		*pTcp,	/* ptr to tcpcb (bsd code)*/
				void 		*sock,
				void		**ppConn) /* ptr to pConn, ret to caller */
{
	tpl_bindCb_t	*pBind;
	tpl_connCb_t	*pConn;
	tpl_stats_t		*pStats;
    struct inpcb    *pInpcb;
	byte			*pMac;
	int32			camIdx;
	int32			link;
	int32			rc = 0;

	pInpcb = pInp;
	pStats = &pTplCfg->tplStats;

	TPL_CONN_PRINTF(("tpl_addConnInd prot=%d la=%x ra=%x lp=%d rp=%d\n",
				prot, ntohl(laddr), ntohl(raddr), ntohs(lport), ntohs(rport)));

	/*-------------------------------
	 * use pBind passed to bsd during 
	 * tpl_bindReq() 
	 *-------------------------------
	 */
	pBind = *ppConn; 

	/*-------------------------------
	 * Check if protocol & port have 
	 * been bound to TPL.
	 *-------------------------------
	 */
	if (tpl_validateBindCb( pBind ) != NFX_OK)
		return ( TPL_ERR_NOT_BIND );

	/*-------------------------------
	 * Find out link id from Inpcb.
	 *-------------------------------
	 */
	link = bsd_getLinkId( pInpcb );
	if (link < 0)
	{
		pStats->cntNoCamRsrc++;
		return (TPL_ERR_BAD_LINK);
	}

	/*-------------------------------
	 * Allocate a CAM entry.
	 *-------------------------------
	 */
#ifdef TPL_SIM	
	camIdx = luc_simCamAdd(prot, laddr, raddr, lport, rport);
#else
	camIdx = luc_camAddFlow(prot, laddr, raddr, lport, rport);
#endif

	if (camIdx < 0)
	{
		pStats->cntNoCamRsrc++;
		return ( TPL_ERR_NO_CAM_RSRC );
	}

	/*-------------------------------
	 * Allocate tpl_connCb of same idx.
	 *-------------------------------
	 */
	pConn = pTplCfg->connTblBase + camIdx;

	/* Su, assign ConnId */
	pConn->connId = pTplCfg->glConnId ++;
	

	if (pConn->flags & TPL_CONN_INUSE)
	{
		/*-------------------------------
		 * CAM & TPL Connection Out of Sync!
		 *-------------------------------
		 */
		pStats->cntBadConnErr++;
 		return ( TPL_ERR_BAD_CONN );
	}

	/*-------------------------------
	 * Setup tpl conn fields.
	 *-------------------------------
	 */
	pConn->flags   = TPL_CONN_INUSE;
	pConn->camIdx  = camIdx;
	pConn->prot    = prot;
	pConn->laddr   = laddr;
	pConn->raddr   = raddr;
	pConn->lport   = lport;
	pConn->rport   = rport;
	pConn->pBindCb = pBind;
	pConn->pInpCb  = pInp;
	pConn->pTcpCb  = pTcp;
	pConn->pSock   = sock;
	pConn->linkNum = link;
	pConn->linkType= PKT_LINK_TYPE_ETHER;

#if defined(LINK_AGG)
	if (bsd_IsLinkVifLA(luc_getIntfFromLink(link)))
		pConn->flags |= TPL_CONN_VIF_LA;
#endif /* LINK_AGG */

	/* su, 05/13/02, indicate connection has no mac */
	if ((pInpcb->inp_flags & INP_NOMAC)) {
		tpl_pend_t *tplPend;
		
		/* enque a pending pConn */
		tplPend = (tpl_pend_t*)eee_ramAlloc(sizeof(tpl_pend_t));
		tplPend->next = 0;
		tplPend->prev = 0;
		tplPend->pConn = pConn;
		tplPend->expire = 0;	
	
		tplPend->next = pTplCfg->tplPend;
		if (pTplCfg->tplPend)
			pTplCfg->tplPend->prev = tplPend;
		pTplCfg->tplPend = tplPend;
		pConn->flags |= TPL_CONN_NOMAC;

		pTplCfg->numPend++;

		TPL_CONN_PRINTF(("%s, pConn Pending %lx\n", 
						__FUNCTION__, (address_t)pConn));
        /* su, 05/13/02, create timer to purge pending pConn */
        if (!pTplCfg->tplTimer) {
        	pTplCfg->tplTimer = eee_timerCreate(60000, EEE_TIMER_REPEAT);

            if (pTplCfg->tplTimer == TIMER_ERR_NO_TIMER_AVAILABLE)
           		panic("tpl: create timer  failed\n");

            if (eee_timerStart(pTplCfg->tplTimer, tpl_purgeConn, NULL,
           		NULL, NULL, NULL) != NFX_OK)
            	panic("tpl: start timer  failed\n");
 		}   
	}
	pConn->arpTx = 0;
	
	if (pBind->bindFlags & TPL_FLAG_USE_RPC)
		pConn->flags |= TPL_CONN_USE_RPC;

	/*-------------------------------
	 * Fill in Src & Dst MACs & TYPE.
	 *-------------------------------
	 */
	pMac = luc_getMacAddr(link);
    memcpy6(&pConn->srcMac[0], pMac, 6);
    memcpy6(&pConn->dstMac[0], &pInpcb->inp_nhmac[0], 6);


	/*-------------------------------
	 * Alloc assoc. pkt desc list. 
	 * Only allocate if the remote mac
	 * is known
	 *-------------------------------
	 */
	if (!(pConn->flags & TPL_CONN_NOMAC)) {
	
		int pktCnt;
	
		pktCnt = PKT_GET_NUM_AVAIL();
		if (pktCnt >= PKT_LOW_DESC_MARK) {
			pktCnt = (pConn->prot == TPL_PROT_UDP) ? PKT_CONN_UDP_CNT:PKT_CONN_TCP_CNT;
			if (tpl_connAllocPktRing( pConn, pktCnt ) != NFX_OK) {
				pStats->cntNoPktRsrc++;
				TPL_LOGERR(TPL_ERR_NO_RSRC, pConn);
				return TPL_ERR_NO_RSRC;
			}
		}
	}
	

	/*-------------------------------
	 * Upcall app. for new connection.
	 *-------------------------------
	 */
#if defined(NFX_SMP)
	TPL_IPC_SET_CONNSTATE(pConn, TPL_CONN_IPCSTATE_ADDING);
	rc = tpl_ipcSendEvntMsg(TPL_EVNT_ADD_CONN, pConn, 1);
#else
	rc =  (pBind->ul_addConnInd)( pBind, pConn, &pConn->pAppHdl ,raddr, rport);
#endif /* NFX_SMP */
	if ( rc != NFX_OK)
	{
		luc_camDelFlow( camIdx );
		pConn->flags = 0;
		pStats->cntUlAddFail++;
		return ( TPL_ERR_UL_ADD_FAIL );
	}

	/*-------------------------------
	 * Setup Usr Info field in CAM.
	 *-------------------------------
	 */
	luc_camSetUsrInfo( camIdx, pConn );


	/*-------------------------------
	 * Insert new conn at head of list.
	 *-------------------------------
	 */
	pConn->pNext = pBind->pConnList;
	pBind->pConnList = pConn;
	pBind->numConnCb++;

	pStats->cntAddConnInd++;
	pTplCfg->numConnCb++;

	*ppConn = pConn;

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:		tpl_delConn
 *
 * Description: Called by Upper/lower layer to delete a connection.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 7/15/01
 *-----------------------------------------------------------------
 */
static int32 		/* NFX_OK or error */
tpl_delConn( tpl_connCb_t *pConn , int32 dir)
{
	tpl_bindCb_t	*pBind;
	tpl_stats_t		*pStats;
	int32			rc;

	TPL_CONN_PRINTF(("tpl_delConn %s pConn=0x%08x\n", 
					(dir)? "Req":"Ind", pConn));

	pStats = &pTplCfg->tplStats;

	/*-------------------------------
	 * Make sure connCb is still vaild.
	 *-------------------------------
	 */
	if (tpl_validateConnCb( pConn ) != NFX_OK)
		return (TPL_ERR_BAD_CONN);

	/*-------------------------------
	 * Delete corresponding CAM entry.
	 *-------------------------------
	 */
	luc_camDelFlow( pConn->camIdx );

	/*-------------------------------
	 * Make sure binding is still valid.
	 * Upcall app. for delete conn.
	 *-------------------------------
	 */
	pBind = pConn->pBindCb;
	if (tpl_validateBindCb( pBind ) != NFX_OK)
		return (TPL_ERR_BAD_CONN);

#if defined(NFX_SMP)
	TPL_IPC_SET_CONNSTATE(pConn, TPL_CONN_IPCSTATE_DELING);
	rc = tpl_ipcSendEvntMsg(TPL_EVNT_DEL_CONN, pConn, 1);
#else
	rc = (pBind->ul_delConnInd)( pConn->pAppHdl , pConn);
#endif /* NFX_SMP */


	/* Always free pkts in transport layer after 
	 * application has free pkts on its side.
	 * Consistency
	 */


	/* su. 05/13/02, clear arpTx queue */
	/* AR1587, make sure pkt inuse. It may
	 * have been freed by application
	 */
	if (pConn->arpTx) {
		pStats->cntXmtMacQueue--;
		if ((pConn->arpTx->flags & PKT_DESC_INUSE)) 
			TPL_FREE_PACKET(pConn, pConn->arpTx);
	}

	/*-------------------------------
	 * Free assoc. pkt desc list. 
	 *-------------------------------
	 */
	tpl_connFreePktRing( pConn );

	/*-------------------------------
	 * Increment del req/ind counter.
	 *-------------------------------
	 */
	if (dir == TPL_IND)
		pStats->cntDelConnInd++;
	else
		pStats->cntDelConnReq++;

	pTplCfg->numConnCb--;
	pBind->numConnCb--;

	/*-------------------------------
	 * Delete conn from head of list.
	 *-------------------------------
	 */
	if (pBind->pConnList == pConn)
		pBind->pConnList = pConn->pNext;
	else
	{
		tpl_connCb_t	*prevPtr;
		tpl_connCb_t	*currPtr;

		/*-------------------------------
		 * Go thru list to find matching 
		 * conn to be removed.
		 *-------------------------------
		 */
		prevPtr = pBind->pConnList;
		currPtr = prevPtr->pNext;
		while(currPtr)
		{
			if (currPtr == pConn)
			{
				prevPtr->pNext = currPtr->pNext;
				break;
			}
			prevPtr = currPtr;
			currPtr = currPtr->pNext;
		}
	}

	/*-------------------------------
	 * Clear tpl_connCb.
	 *-------------------------------
	 */
	memset(pConn, 0, sizeof(tpl_connCb_t));

	/* Su, 10/02/02 
	 * If actively opened conn, clear BindCB
	 */
	if ((pBind->bindFlags & TPL_FLAG_CLIENT)) {
		if (pBind->pConnList || pBind->numConnCb) {
			/* For actively open Conn, each 
			 * bindCB at most has one pConn associate
			 * to it, and this pConn had been deleted
			 * now.
			 */
			TPL_LOGERR(TPL_ERR_BAD_BIND, pBind);
		}

		/* socket has already been closed */
		pBind->pBsdSock = NULL;
		tpl_unbindReq(pBind);
	}
		
	return (rc);
}


/*-----------------------------------------------------------------
 * Name:		tpl_delConnReq
 *
 * Description: Called by NFS/Mount/CIFS layer to delete a connection.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/21/01
 *-----------------------------------------------------------------
 */
#if defined(NFX_SMP)
int32 		/* NFX_OK or error */
tpl_delConnReqCore( void *connPtr )
#else
int32 		/* NFX_OK or error */
tpl_delConnReq( void *connPtr )
#endif /* NFX_SMP */
{
	tpl_connCb_t	*pConn;

	pConn = connPtr;

    TPL_CONN_PRINTF(("%s: pConn=%x lport=%d rport=%d pSock=%x\n",
					__FUNCTION__, 
					pConn, 
					ntohs(pConn->lport),
					ntohs(pConn->rport), 
					pConn->pSock));

	/*
	 * su, 07/02/02
	 * make sure pConn still in use
	 */
	if (!(pConn->flags & TPL_CONN_INUSE))
		return NFX_OK;
	

	/*-------------------------------
	 * Return, if del is pending.
	 *-------------------------------
	 */
    if (pConn->flags & TPL_CONN_DEL_PENDING)
    {
	    TPL_CONN_PRINTF(("%s: pConn=%x : DEL already pending!\n",
				        __FUNCTION__, pConn ));
        return NFX_OK;
    }

	/*-------------------------------
	 * Set pending flag.
	 * Call BSD to close sock.
	 * 
	 * Will get an delConnInd from BSD
	 * lower socket is closed. 
	 * Perform actual closing then.
	 *-------------------------------
	 */
    pConn->flags |= TPL_CONN_DEL_PENDING;
	if (pConn->pSock)
		bsd_CloseSock( pConn->pSock);

	return NFX_OK;
}

#if defined(NFX_SMP)
/*-----------------------------------------------------------------
 * Name:		tpl_delConnReq
 *
 * Description: Called by NFS/Mount/CIFS layer to delete a connection.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/21/01
 *-----------------------------------------------------------------
 */
int32 		/* NFX_OK or error */
tpl_delConnReq( void *connPtr )
{
	return tpl_ipcSendEvntMsg(TPL_EVNT_CLOSE_REQ, connPtr, 0);
}
#endif /* NFX_SMP */

/*-----------------------------------------------------------------
 * Name:		tpl_delConnId
 *
 * Description: Called by BSD UDP/TCP layer to delete a connection.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 9/21/01
 *-----------------------------------------------------------------
 */
int32 		/* NFX_OK or error */
tpl_delConnInd( void *pConn )
{
	/* Su, 12/27/01 */
	return (tpl_delConn( pConn, TPL_IND) );
}

/*-----------------------------------------------------------------
 * Name:        tpl_bsdConnInd
 *
 * Description: Called by BSD layer to indicate a connection event.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 8/20/01
 *-----------------------------------------------------------------
 */
int
tpl_bsdConnInd(void *sock, void *pcb, int evt, void **pHdl)
{
	struct inpcb 		*pInpcb;
	int32				rc = NFX_OK;
	tpl_connCb_t        *pConn = (tpl_connCb_t*)*pHdl;
	pkt_desc_t			*pDesc;

	pInpcb = pcb;
	switch (evt)
	{
		/*-------------------------------
		 * Add connection event.
		 *-------------------------------
		 */
		case	SO_CONN_UP_IND:
				/*-------------------------------
				 * inpcb has all the info we need
				 * to add a TPL connection.
				 * Preserved NETWORK ORDER on
				 * addresses & ports.
				 *-------------------------------
				 */
				rc = tpl_addConnInd(pInpcb->inp_ip_p,
									(pInpcb->inp_laddr.s_addr),
									(pInpcb->inp_faddr.s_addr),
									(pInpcb->inp_lport),
									(pInpcb->inp_fport),
									pInpcb,
									pInpcb->inp_ppcb,
									sock,
									pHdl);
	
				if (rc != NFX_OK) 
					TPL_LOGERR( rc, *pHdl);
				break;

		/*-------------------------------
		 * Delete connection event.
		 *-------------------------------
		 */
		case	SO_CONN_DOWN_IND:
			    /* su, 07/02/02
				 * set  TPL_CONN_DEL_PENDING
				 * so socket won't be closed twice
				 */
			    pConn->flags |= TPL_CONN_DEL_PENDING;
				bsd_CloseSock(sock); 
				break;

		case	SO_CONN_RTCHG_IND:
				/* su. 04/25/02. route/mac has changed. Need to update
                 */
                //printf("mac updated from %s", bsdPrintMac((char*)pConn->dstMac));
				//printf(" to %s\n", bsdPrintMac((char*)pInpcb->inp_nhmac));
                memcpy6(&pConn->dstMac[0], &pInpcb->inp_nhmac[0], 6);

				/* Su, 09/10/02
				 * now that we have mac, allocate pkt ring
				 */
				if (pConn->flags & TPL_CONN_NOMAC)
				{
					int pktCnt;
					
					pktCnt = PKT_GET_NUM_AVAIL();
					if (pktCnt >= PKT_LOW_DESC_MARK) {
						pktCnt = (pConn->prot == TPL_PROT_UDP) ? PKT_CONN_UDP_CNT:PKT_CONN_TCP_CNT;
						tpl_connAllocPktRing( pConn, pktCnt );
					}
				}

				/* su, 05/13/02. unset NOMAC flags, and send any queue pkt */
				pConn->flags &= ~TPL_CONN_NOMAC;
				if ((pDesc = pConn->arpTx)) {
					pConn->arpTx = 0;

					pTplCfg->tplStats.cntXmtMacQueue--;
					
					/* AR1587. make sure pkt is inuse. 
					 * It may be freed by application.
					 */
					if (pDesc->flags & PKT_DESC_INUSE) {
						TPL_CONN_PRINTF(("%s, send Q pdesc %lx\n", __FUNCTION__, pDesc);
						memcpy6(pDesc->pMac, &pInpcb->inp_nhmac[0], 6));
						
						/* as if pkt is new to tpl */
						pDesc->flags &= ~PKT_USR_TPL_TX;
						tpl_xmtPkt(pConn, pDesc);
					}
				}

				break;
	
    	case	SO_CONN_CLOSE_REQC:
		        rc = tpl_delConnInd( *pHdl );
				if (rc != NFX_OK)
				  TPL_LOGERR( rc, *pHdl);
                break; 		  

		default:
				TPL_LOGERR( TPL_ERR_BAD_EVT, *pHdl );
				break;
	}

    return rc;
}

extern ushort16 ip_id;

/*-----------------------------------------------------------------
 * Name:        tpl_clrIpOffset
 *
 * Description: Clears IP offset, MF & DF flags.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 02/09/02
 *-----------------------------------------------------------------
 */
void tpl_clrIpOffset(pkt_desc_t *pDesc)
{
    struct ip   *ipHdr;

    ipHdr         = (struct ip *)pDesc->pIpHdr;
    ipHdr->ip_off = 0;
}

/*-----------------------------------------------------------------
 * Name:        tpl_setIpHdr
 *
 * Description: Setup IP id, cksum & len.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 10/26/01
 *-----------------------------------------------------------------
 */
void tpl_setIpHdr(tpl_connCb_t *pConn, pkt_desc_t *pDesc)
{
    struct ip   *ipHdr;
	ushort16	ipLen;
	

   	ipLen  = pDesc->hdrSize + PKT_IP_HDR_SIZE;

	/* su, 06/24/02.
	 * add udp hdr length if whole pkt or first
	 * fragment
	 */
	if ((!(pDesc->flags & PKT_IP_FRAGMENT)) ||
		 (pDesc->flags & PKT_IP_FIRST_FRAG))
		ipLen += (pConn->prot==TPL_PROT_UDP) ? PKT_UDP_HDR_SIZE:PKT_TCP_HDR_SIZE;

    ipHdr         = (struct ip *)pDesc->pIpHdr;
   	ipHdr->ip_len = htons(ipLen);
    if (!ipHdr->ip_off)
    	ipHdr->ip_id  = htons(ip_id++);
    ipHdr->ip_sum = 0;

	/* su, 07/12/02, hardware support chksum,
	 *  prom always use slowpath to send.
	 */
#if !defined(PROM)
	if (!isCheetah())
#endif
		ipHdr->ip_sum = htons(tpl_cksumIp( pConn, pDesc ));

}

/* su, 06/11/02, use incremental chksum 
 * we rely on tpl_setUdpHdr is called 
 * consecutively for the fragments belongs to
 * the same pkt.
 */
static uint32 chkSumTotal = 0;
/*-----------------------------------------------------------------
 * Name:        tpl_setUdpHdr
 *
 * Description: Setup UDP cksum & len.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 10/26/01
 *-----------------------------------------------------------------
 */
void tpl_setUdpHdr(tpl_connCb_t *pConn, pkt_desc_t *pDesc)
{
    struct udphdr   *udpHdr;
    struct ip       *ipHdr;
    

    ipHdr = (struct ip *)pDesc->pIpHdr;
	udpHdr = (struct udphdr *)pDesc->pTplHdr;

	if (pDesc->flags & PKT_IP_FRAGMENT) {

		/* su, 07/12/02
		 * hardware support udp fragment chksum
		 */
#if !defined(PROM)
		if (isCheetah())
			return;
#endif 

		/* su, 06/11/02, perform partial chksum 
		 * we don't setup uh_ulen, it is already 
		 * be set in tpl_xmtUdpData().
		 */
		if (pDesc->flags & PKT_IP_FIRST_FRAG) {
			udpHdr->uh_sum  = 0;
		}
		else {					
			/* collapse pHdr and tplHdr
			 * for non first fragments.
			 * update hdrSize also 
			 * for chksum */
			ushort16	offset;
			offset = (address_t)pDesc->pHdr - 
				(address_t)pDesc->pTplHdr;
			pDesc->pHdr = pDesc->pTplHdr;
			pDesc->hdrSize += offset;
		}
			
		chkSumTotal += tpl_cksumTpl( pConn, pDesc );

		/* first fragment, give cksum value */ 
		if (pDesc->flags & PKT_IP_FIRST_FRAG) {
			while (chkSumTotal >> 16)
				chkSumTotal = (chkSumTotal & 0xFFFF) + (chkSumTotal >> 16);
			
			/* ok to use htons, only interested in last two bytes */
			udpHdr->uh_sum = ~htons(chkSumTotal);
			chkSumTotal = 0;	/* ready for next fragmented pkts */
		}
	}
	else
	{		
		/* cksum for whole udp */
		if (ipHdr->ip_off)
			return;
		
		udpHdr->uh_ulen = htons(pDesc->hdrSize + PKT_UDP_HDR_SIZE); 
		udpHdr->uh_sum  = 0;

		/* su, 07/12/02
		 * hardware support udp fragment chksum
		 */
#if !defined(PROM)
		if (!isCheetah())
#endif 
			udpHdr->uh_sum  = htons(tpl_cksumTpl( pConn, pDesc ));
	}
}

/*-----------------------------------------------------------------
 * Name:        tpl_setTcpHdr
 *
 * Description: Setup TCP checksum.
 *
 * Created by : Su Wang
 *
 * Date Created: 01/04/02
 *-----------------------------------------------------------------
 */
void tpl_setTcpHdr(tpl_connCb_t *pConn, pkt_desc_t *pDesc)
{
    struct tcphdr   *tcpHdr;
    struct ip       *ipHdr;

	/* su, 07/12/02, hardware support chksum */
#if !defined(PROM)
	if (isCheetah())
		return;
#endif
	

    ipHdr = (struct ip *)pDesc->pIpHdr;

	/* su, 06/11/02, don't do cksum
	 * if fragment. cksum done by
	 * luc
	 */
    if (ipHdr->ip_off)
		return;

    tcpHdr          = (struct tcphdr *)pDesc->pTplHdr;
    tcpHdr->th_sum  = 0;
    

	tcpHdr->th_sum  = htons(tpl_cksumTpl( pConn, pDesc ));
}

/*-----------------------------------------------------------------
 * Name:        tpl_purgeConn
 *
 * Description: timeout has occurs, remove all invalid pConn.
 *
 * Created by : Su Wang
 *
 * Date Created: 05/13/02
 *-----------------------------------------------------------------
 */
int32 
tpl_purgeConn(uint32 handle, void* parm1, void* parm2, void* parm3, void* parm4)
{
	tpl_connCb_t	*pConn;
	tpl_pend_t		*tplPend, *nextPend;
	int				remove = 0;
	
	tplPend = pTplCfg->tplPend;
	
	while (tplPend) {
		nextPend = tplPend->next;
		
		pConn = tplPend->pConn;
		
		TPL_CONN_PRINTF(("pConn is %lx, flag is %x\n", (address_t)pConn, pConn->flags));
		if ((!(pConn->flags & TPL_CONN_INUSE)) || 
			(pConn->flags & TPL_CONN_DEL_PENDING)) {
			/* pConn is already removed or in the process of being removed,
			 * remove corresponding element for pending queue
			 */
			TPL_CONN_PRINTF(("%s: pConn %lx already removed by app\n", 
				   __FUNCTION__, (address_t)pConn));
			remove = 1;
		}
		else if (!(pConn->flags & TPL_CONN_NOMAC)) {
			/* mac has been rcved, remove from pending queue */
			TPL_CONN_PRINTF(("%s: pConn %lx found mac\n", 
				   __FUNCTION__, (address_t)pConn));
			remove = 1;
		}
		else if (tplPend->expire >=2) {
			/* 2 min expired, purge pConn, and rm from pending queue */
			TPL_CONN_PRINTF(("%s: pConn %lx did find mac, purge ..\n", 
				   __FUNCTION__, (address_t)pConn));
			remove = 1;
#if defined(NFX_SMP)
			tpl_delConnReqCore(pConn);
#else
			tpl_delConnReq(pConn);
#endif /* NFX_SMP */
		}
		else {
			/* wait for next timeout */
			tplPend->expire++;
		}

		/* remove pConn from pending queue */
		if (remove) {
			if (tplPend->prev) {
				tplPend->prev->next = tplPend->next;
				if (tplPend->next)
					tplPend->next->prev = tplPend->prev;
			}
			else {
				pTplCfg->tplPend = tplPend->next;
				if (tplPend->next)
					tplPend->next->prev = NULL;
			}
			pTplCfg->numPend--;
			eee_ramDealloc(tplPend);
		}

		tplPend = nextPend;
	}
	
	return NFX_OK;
}

/*-----------------------------------------------------------------
 * Name:    tpl_checkConnActivity    
 *
 * Description: Check client activity for UDP connections. 
 *
 * Created by : Joseph Vu
 *
 * Date Created: 06/25/02
 *-----------------------------------------------------------------
 */
void 
tpl_checkConnActivity(uint32 tpl_connActivityTimeOut)
{
	eeeGMT_t 		now;
	int32			currTimeInSecs;
	int32			deltaSecs;
	int32			i;
	tpl_connCb_t	*pConn;

	eee_rtcGMT (&now);
	currTimeInSecs = now.sec; 

	pConn = pTplCfg->connTblBase;
	for (i= 0; ( (i < pTplCfg->maxConnCb) && (pConn <= pTplCfg->connTblEnd));
			i++, pConn++)
	{
		/*-----------------------------------
		 *Skip if connection is not inused
		 *or not UDP
		 *----------------------------------
		 */
		if ( ((pConn->flags & TPL_CONN_INUSE) == 0) ||  
				(pConn->prot != TPL_PROT_UDP) || 
				(pConn->last_aTime == 0) )
			continue;

		deltaSecs = currTimeInSecs - pConn->last_aTime;
		/*-----------------------------------
		 *Check for wrap condition
		 *-----------------------------------
		 */
		if (deltaSecs < 0)
		{
			pConn->last_aTime = currTimeInSecs;
			continue;
		}

		if (deltaSecs > tpl_connActivityTimeOut)
		{
			/* this function is called on acpu in smp case */
			tpl_delConnReq(pConn);
		}
	}

}

/*-----------------------------------------------------------------
 * Name:  	tpl_connUpdateActivityTime 
 *
 * Description: Update activity time for a TPL connection. 
 *
 * Created by : Joseph Vu
 *
 * Date Created: 06/25/02
 *-----------------------------------------------------------------
 */
void
tpl_connUpdateActivityTime(void *pTplConn, 
	uint32 activityTime			//Activity time in seconds
)
{
	tpl_connCb_t 	*pConn;

	pConn = (tpl_connCb_t    *)pTplConn;
	pConn->last_aTime = activityTime;
	
	return;
}

