/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: tpl-xmt.c
 *
 * RCS:  $Id: tpl-xmt.c,v 1.36 2002/11/20 18:04:03 swang Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: Transport Layer transmit related routines.
 *
 * Created by:	Shaun Hui
 *
 * Date Created: 07/01/2001
 *
 *-----------------------------------------------------------------
 */
#include <nfx-incl.h>

#include <exports/pkt-api.h>
#include <exports/luc-api.h>		
#include <exports/tpl-api.h>
#include <exports/bsdsock-api.h>
#include <exports/bsdnet-api.h> 
#include "tpl.h"

#include "../sm-bsd-sys/flags.h"
#include "../sm-bsd-sys/cdefs.h"
#include "../sm-bsd-sys/types.h"

#include "../sm-bsd-netinet/in.h"
#include "../sm-bsd-netinet/ip.h"
#include "../sm-bsd-netinet/udp.h"
#include "../sm-bsd-netinet/udp_stats.h"
#include "../sm-bsd-netinet/ip_stats.h"
#include "../sm-bsd-api/bsd-vif.h"

#define TPL_DBG_PRINTF(a)		//  printf a

int32 luc_xmtGmcDataReq(pkt_desc_t *pDesc);

static int32 tpl_xmtTcp(    void *pConnId,	pkt_desc_t *pDesc);
static int32 tpl_xmtUdp(    void *pConnId,	pkt_desc_t *pDesc);

static int32 tpl_xmtTcpPkt( void *pConnId,	pkt_desc_t *pDesc);
static int32 tpl_xmtUdpPkt( void *pConnId,	pkt_desc_t *pDesc);

static int32 tpl_xmtTcpData(void *pConnId,	pkt_desc_t 		*pDesc,
							int32 fragCnt,  luc_fragCb_t 	*fcb);

static int32 tpl_xmtUdpData(void *pConnId,	pkt_desc_t 		*pDesc,
							int32 fragCnt,  luc_fragCb_t 	*fcb);

#define TPL_CHK_PKT_RESOURCE(_pDesc, _fragCnt)		\
{											\
	if (_fragCnt > PKT_GET_NUM_AVAIL()) {	\
		pTplCfg->tplStats.cntNoPktRsrc++;	\
		pDesc->refCnt = 0;					\
		return TPL_ERR_NO_PKT_RSRC;			\
	}										\
}											\

#if defined(NFX_SMP)
/*-----------------------------------------------------------------
 * Name:  		tpl_xmtPktSlow
 *
 * Description: slowpath tx on ACPU
 *
 * Created by : Su Wang
 *
 * Date Created: 12/03/01
 *-----------------------------------------------------------------
 */
int32 		/* NFX_OK or error */
tpl_xmtPktSlow(uint32 raddr, ushort16 rport, pkt_desc_t	*pDesc, 
			   void	*sock, int owner)
{
	/* application want to keep the pkt, make a copy */
	if (owner) {
		pkt_desc_t	*tmpDesc;
		
		if (!(tmpDesc = PKT_ALLOC_PACKET(1)))
			return NFX_ERR;

		memcpy(tmpDesc->pHdr, pDesc->pHdr, pDesc->hdrSize);
		tmpDesc->hdrSize = pDesc->hdrSize;
		tmpDesc->flags = pDesc->flags;
		pDesc = tmpDesc;
	}
	
	pDesc->raddr = raddr;
	pDesc->rport = rport;
	pDesc->pTplConn = sock;

	return tpl_xmtPkt(0, pDesc);
}

/*-----------------------------------------------------------------
 * Name:  		tpl_xmtPkt
 *
 * Description: wrapper on ACPU for tplxmtPktCore
 *
 * Created by : Su Wang
 *
 * Date Created: 12/03/01
 *-----------------------------------------------------------------
 */
int32 		/* NFX_OK or error */
tpl_xmtPkt( void 		*pConnId, 	/* TPL Connection CB */
			pkt_desc_t	*pDesc)		/* packet descriptor */
{
	tpl_ipcq_t	*pktTxQueuePtr = &(pTplCfg->tplIpcQ[TPL_IPC_PKTQ_TX]);

	/* queue pkt onto TPL_IPC_PKTQ_TX */
	if (TPL_IPCQ_FULL(pktTxQueuePtr))
		panic("%s: ipc tx queue full\n", __FUNCTION__);
	
	/* no need to pass on pConnId, it is
	 * in pDesc->pTplConn
	 */
	TPL_IPCQ_WRITE(pktTxQueuePtr, pDesc);
	return NFX_OK;
}
#endif /* NFX_SMP */

/*-----------------------------------------------------------------
 * Name:  		tpl_xmtPktCore
 *
 * Description: "Fast-path" transmit function to be called by 
 *				upper layer for sending a packet.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 7/15/01
 *-----------------------------------------------------------------
 */
#if defined(NFX_SMP)
int32 		/* NFX_OK or error */
tpl_xmtPktCore( void 		*pConnId, 	/* TPL Connection CB */
				pkt_desc_t	*pDesc)		/* packet descriptor */
#else
int32 		/* NFX_OK or error */
tpl_xmtPkt( void 		*pConnId, 	/* TPL Connection CB */
				pkt_desc_t	*pDesc)		/* packet descriptor */
#endif /* NFX_SMP */
{
	tpl_connCb_t	*pConn;
	tpl_stats_t		*pStats;

	pStats = &pTplCfg->tplStats;
	pConn  = (tpl_connCb_t *)pConnId;


#if defined(NFX_SMP)
	/* slowpath pkt, pTplConn is ptr to socket */
	if (pDesc && (!(pDesc->flags & PKT_USR_TPL))) {
		bsd_SendSock(pConnId, pDesc->raddr, pDesc->rport, pDesc->pHdr,
					 BSD_BUF_FLAT, pDesc->hdrSize, 0);
		PKT_FREE_PACKET(pDesc);
		return NFX_OK;
	}
#endif /* NFX_SMP */	


#ifdef PKT_USE_TIME_STAMP
	if (pDesc)
    	pkt_setTimeStamp(pDesc->pTime, PKT_TIME_IDX_TPL_TX);
#endif
	
	/*-------------------------------
	 * Make sure connCb is still vaild.
	 *-------------------------------
	 */
	if (!pConn || ((pConn->flags & TPL_CONN_INUSE) == 0) ||
		(pDesc && (pDesc->tplConnId != pConn->connId)))
	{
		/*-------------------------------
		 * connCb is gone. free pkt back 
		 * to free pkt list.
		 *-------------------------------
		 */

		/* su, TODO don't free the pkt
		 * pkt from application, application
		 * need to free it
		 */
		
	    if (pDesc)  /* Su 12/27, pDesc can be NULL for pure tcp contrl */
			PKT_FREE_PACKET( pDesc );

		pStats->cntBadConnErr++;
		return (TPL_ERR_BAD_CONN);
	}


	/* su, 05/13/02, make sure mac is ready, else
     * send arp request, save pdesc.
     */
    if ((pConn->flags & TPL_CONN_NOMAC)) {
		int		ret;
		uchar8	*tmpChar;

		/* in case pcb->rt is removed 
		 * chk if new route available now 
		 */
		tmpChar = bsd_GetPcbNHMac(pConn->pInpCb);
		if ((bsd_ChkPcbRt(pConn->pInpCb) == NFX_OK) &&
			(*tmpChar != 0xff)) {

			pStats->cntXmtMacUncall++;

			/* update connection mac */
			memcpy(&pConn->dstMac[0], bsd_GetPcbNHMac(pConn->pInpCb), 6);
			
			/* update incoming pkt */
			if (pDesc)
				memcpy(pDesc->pMac,pConn->dstMac, 6);

			/* unset flag */
			pConn->flags &= ~TPL_CONN_NOMAC;
			
			/* release queued pkt back to application
			 * as if it's already been transmitted
			 */
			if ((pConn->arpTx) && pDesc != pConn->arpTx) {
				pStats->cntXmtMacQueue--;
				tpl_xmtCmp(pConn, pConn->arpTx);
			}
		}
		else {
			
			if (!pConn->arpTx)	{	/* arpTx is empty now, queue pDesc */
				pConn->arpTx = pDesc;
				pStats->cntXmtMacQueue++;
				
				/* pkt inside tpl but not send down */
				pDesc->flags |= PKT_USR_TPL_TX;
				ret = NFX_OK;
			}
			else if (pConn->arpTx == pDesc)	{/* same pDesc on arpTx, do onthing */
				ret = NFX_OK;
			}
			else	{				/* arpTx is non-empty, ret err */
				pStats->cntXmtNoMacErr++;
				ret = TPL_ERR_XMT_FAIL;
			}
			
			bsd_SendArp(pConn->pInpCb);
			return ret;
		}
    }


	/*---------------------------------
	 * Su, 07/29/02. Make sure pkt is 
	 * not already in transport layer.
	 * Set flag PKT_USR_TPL_TX
	 *---------------------------------
	 */
	if (pDesc) {
		if (pDesc->flags & PKT_USR_TPL_TX) {
			pStats->cntXmtDup++;
			return NFX_OK;
		}
	
		pDesc->flags |= PKT_USR_TPL_TX;
		pStats->cntXmtDownCall++;
	}

	/*-------------------------------
	 * Call UDP or TCP transmit.
	 *-------------------------------
	 */
	switch (pConn->prot)
	{ 
		case 	TPL_PROT_UDP:
				return (tpl_xmtUdp( pConnId, pDesc));

#ifndef PROM
		case	TPL_PROT_TCP:
				return (tpl_xmtTcp( pConnId, pDesc));
#endif
		
		default:
				TPL_LOGERR( TPL_ERR_BAD_PROT, pDesc );
				return TPL_ERR_BAD_PROT;
	}

}

/*-----------------------------------------------------------------
 * Name:  		tpl_xmtCmp
 *
 * Description: Transmit complete function called by lower layer (LUC)
 *				after sending a packet.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 7/15/01
 *-----------------------------------------------------------------
 */
int32 		/* NFX_OK or error */
tpl_xmtCmp( void 		*pConnId, 	/* TPL Connection CB */
			pkt_desc_t	*pDesc)		/* packet descriptor */
{
	tpl_bindCb_t	*pBind;
	tpl_connCb_t	*pConn;
	tpl_stats_t		*pStats;

	pStats = &pTplCfg->tplStats;
	pConn  = (tpl_connCb_t *)pConnId;

	pStats->cntXmtCmp++;

	/*-------------------------------
	 * Make sure connCb is still vaild.
	 *-------------------------------
	 */
	if (!pConn || ((pConn->flags & TPL_CONN_INUSE) == 0) ||
		(pConn->connId != pDesc->tplConnId))
	{
		/*-------------------------------
		 * connCb is gone. free pkt back 
		 * to free pkt list.
		 *-------------------------------
		 */
		pStats->cntBadConnErr++;

		/*-------------------------------
		 * Su, 07/30/02
		 * free reference pkts even if there is
		 * no associated pkts
		 *--------------------------------
		 */
		if (pDesc->pRefDesc) {
			if (--pDesc->pRefDesc->refCnt > 0)
				return (PKT_HOLD_PKT);
			
			TPL_FREE_PACKET(pConn, pDesc->pRefDesc);
			return (PKT_HOLD_PKT);
		}
		
		return (PKT_FREE_PKT);
	}

	/*-------------------------------
	 * Make sure bindCb is still vaild.
	 *-------------------------------
	 */
	pBind = pConn->pBindCb;
	if ((!pBind) || (!pBind->ul_xmtCmpInd))
	{
		/*-------------------------------
		 * bindCb is gone. free pkt back 
		 * to free pkt list.
		 *-------------------------------
		 */
		pStats->cntBadBindErr++;
		return (PKT_FREE_PKT);
	}


	/*-------------------------------
	 * Handle TCP xmt cmp.
	 *-------------------------------
	 */
	if (pConn->prot == TPL_PROT_TCP)
	{

		/*-------------------------------
		 * pkt in tcp tx queue, dont remove 
		 *-------------------------------
		 */
		if (pDesc->flags & PKT_USR_TCP_TX_Q)
		{
			
			/* Su, 08/14/02. pkt not ready to be
			 * freed yet. Keep stats correct
			 */
			pStats->cntXmtCmp--;

			
			/*-------------------------------
			 * pkt is sent, but dont upcall. 
			 * Instead, upcall app tx complete 
			 * when TCP ACK is rcvd, 
			 *-------------------------------
			 */
			return (PKT_HOLD_PKT);
		}

		/*-------------------------------
		 * pkt is tcp ack (no data), free it.
		 *-------------------------------
		 */
		if (pDesc->flags & PKT_USR_TCP_TX_ACK)	
		 {
			TPL_FREE_PACKET( pConn, pDesc );
			return (PKT_HOLD_PKT);
		}
	}
	
	TPL_DBG_PRINTF(("%s: pDesc=%x refDesc=%x, refCnt=%d\n", 
				__FUNCTION__, pDesc, pDesc->pRefDesc, 
				(pDesc->pRefDesc) ? pDesc->pRefDesc->refCnt : pDesc->refCnt));

	/*-------------------------------
	 * Upcall App when all pkts of
	 * this chain had been sent.
	 *-------------------------------
	 */
	if (pDesc->pRefDesc)
	{
		if (--pDesc->pRefDesc->refCnt > 0) 
			return (PKT_HOLD_PKT);

		pDesc = pDesc->pRefDesc;

	}

	/*-------------------------------
	 * Su, 07/29/02. 
	 * All pkt to application must have
	 * PKT_USR_TPL_TX set.
	 * unset this flag
	 *-------------------------------
	 */
	if (!(pDesc->flags & PKT_USR_TPL_TX)) {
		pStats->cntXmtNoTplUsr++;
		TPL_LOGERR( TPL_ERR_BAD_PKT, pDesc);
		return (PKT_FREE_PKT);
	}
	pDesc->flags &= ~(PKT_USR_TPL_TX | PKT_LUC_THRU);
		

	pStats->cntXmtUpCall++;

	TPL_DBG_PRINTF(("%s: UPCALL pDesc=%x refDesc=%x, refCnt=%d\n", 
			__FUNCTION__, pDesc, pDesc->pRefDesc, 
			(pDesc->pRefDesc) ? pDesc->pRefDesc->refCnt : pDesc->refCnt));

#if defined(NFX_SMP)
	{
		tpl_ipcq_t	*pktTxCmplQueuePtr = 
			&(pTplCfg->tplIpcQ[TPL_IPC_PKTQ_TXCOMP]);
		
		if (TPL_IPCQ_FULL(pktTxCmplQueuePtr))
			panic("%s: ipc tx compl queue full\n", pktTxCmplQueuePtr);
		
		TPL_IPCQ_WRITE(pktTxCmplQueuePtr, pDesc);
	}
#else
	/*-------------------------------
	 * Upcall app. for xmt complete.
	 *-------------------------------
	 */
	if ((pBind->ul_xmtCmpInd)(pConn->pAppHdl, pDesc) == TPL_FREE_PKT)
	{
		/*-------------------------------
		 * App. done with pkt, free pkt 
		 * back to TPL pkt list.
		 *-------------------------------
		 */
		TPL_FREE_PACKET( pConn, pDesc );
	}
#endif /* NFX_SMP */

	/*-------------------------------
	 * Since fast-path pkts are allocated
	 * from TPL free pkt list, 
	 * DO NOT return pkt to PKT free list
	 * until connCb is gone.
	 *-------------------------------
	 */
	return (PKT_HOLD_PKT);
}


#ifndef PROM

static int32 							/* NFX_OK or error */
tpl_xmtTcpPkt( void 		*pConnId, 	/* TPL Connection CB */
			   pkt_desc_t	*pDesc)		/* packet descriptor */
{
	tpl_connCb_t	*pConn;
	tpl_stats_t		*pStats;
	pkt_desc_t		*retp;
    int				ret;
	int				more = 0;


	pStats = &pTplCfg->tplStats;
	pConn  = (tpl_connCb_t *)pConnId;


	TPL_DBG_PRINTF(("%s: 1 pDesc=%x\n", __FUNCTION__, pDesc));
	if (pDesc)				
	{
		/*-------------------------------
		 * Make sure hdr size is ok.
		 *-------------------------------
		 */
		if (pDesc->hdrSize > (PKT_BUFF_SIZE-PKT_BUFF_INDENT))
		{
			TPL_LOGERR( TPL_ERR_BAD_PKT, pDesc);
			return (TPL_ERR_BAD_PKT);
		}
	}


again:

		/*-------------------------------
		 * setup tcp ack & seq etc.
		 *-------------------------------
		 */
		ret = tcp_outputC(pConn->pTcpCb, pConn, pDesc, &retp, 0);

		TPL_DBG_PRINTF(("%s: 2 pDesc=%x retp=%x\n", __FUNCTION__, pDesc, retp));

		switch (ret)
		{
			case	BSD_FAST_ERR:		/* TCP Tx ERR */
			case	BSD_FAST_FULL:		/* TCP Tx Q full */
					TPL_DBG_PRINTF(("%s: ret=%s\n", __FUNCTION__,
									(ret==BSD_FAST_ERR) ? 
									 "BSD_ERR" : "BSD_FULL"));
					if (pDesc) {
						/*-----------------------
						 * Su, 07/30/02
						 * simulate as if packet 
						 * has been transmitted.
						 * This is to maintain a
						 * consistent pkt free 
						 * mechanism
						 *-----------------------
						 */
						pStats->cntXmtTcpErr++;
						tpl_xmtCmp(pConnId,  pDesc );
					}
					return (TPL_ERR_BAD_CONN);

			case	BSD_FAST_QUEUED:	/* Queued for xmt later */
					TPL_DBG_PRINTF(("%s: ret= QUEUED\n", __FUNCTION__));
		  			return NFX_OK;

			case	BSD_FAST_MORE:		/* more to send */
					TPL_DBG_PRINTF(("%s: ret= MORE\n", __FUNCTION__));
		  			more = 1;

			default:					
					pDesc = retp;		/* actual pdesc to be sent */
		}

		/*-------------------------------
		 * pkt was sent & own by driver
		 * don't resend.
		 *-------------------------------
		 */
		if (pDesc->flags & PKT_USR_LUC_TX)
		{
			pStats->cntXmtTcpDup++;
			goto next;			
		}

	/*-------------------------------
	 * Turn on TCP Checksum.
	 *-------------------------------
	 */
	pDesc->flags |= PKT_TCP_CSUM_ENA | PKT_IP_CSUM_ENA | PKT_FAST_PATH;

	pDesc->pktSize = pDesc->dataSize + pDesc->hdrSize + 
					 (pDesc->pHdr - pDesc->pMac);

#if !defined(PROM)
	/*-------------------------------
	 * Check if user data exists. 
	 *-------------------------------
	 */
	/* AR 2550, Su, 10/14/02
	 * if pkt already went through sbm, it 
	 * already has all the content.
	 * It could cause potential crash if FP
	 * already timed out COD
	 */
	if (pDesc->dataSize && (!(pDesc->flags & PKT_LUC_THRU)))
	{
	
#if defined(LINK_AGG)
		/* update pdesc */
		if (pDesc->link != pConn->linkNum)
			tpl_UpdatePkt(pDesc, pConn, TPL_PRING_CHGE_SRCMAC);
#endif /* LINK_AGG */

		/*-------------------------------
 		 * Get it from GMC. luc_rcvGmcData()
		 * will rx Hdr+data pkt & send it.
		 *-------------------------------
		 */
		if (luc_xmtGmcDataReq( pDesc ) != NFX_OK)
		{
			/*-------------------------------
 			 * Error in MCPU sending to GMC.
			 * Need to free PKT 
			 *-------------------------------
			 */
			pStats->cntXmtFail++;
			return TPL_ERR_XMT_FAIL;
		}
		else 
			goto next;			/* process next pkt if any */
	}
#endif 

	/* Su, 10/24 
	 * handle retransmited data pkt, where
	 * hdrsize is used in calculate IP len 
	 */
	if (pDesc->dataSize) {
		pDesc->hdrSize += pDesc->dataSize;
		pDesc->dataSize = 0;
	}


	/*-------------------------------
	 * Set IP len & checksum.
	 *-------------------------------
	 */
	tpl_setIpHdr(pConn, pDesc);


	/*-------------------------------
	 * Set TCP/UDP len & checksum.
	 *-------------------------------
	 */
	tpl_setTcpHdr(pConn, pDesc);


#if defined(LINK_AGG)
		/* update pdesc */
		if (pDesc->link != pConn->linkNum)
			tpl_UpdatePkt(pDesc, pConn, TPL_PRING_CHGE_SRCMAC);
#endif /* LINK_AGG */

	/*-------------------------------
	 * Call LUC to xmt pkt.
	 *-------------------------------
	 */
	if (luc_xmtPktReq( pConn->linkNum, pDesc ) != NFX_OK)
	{
		pStats->cntXmtFail++;
		return TPL_ERR_XMT_FAIL;
	}

next:
	pStats->cntXmtPkt++;

	if (more)					/* su 12/27, only apply to tcp fastpath */
	{
		more = 0;
		pDesc = 0;
		goto again;
	}
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:  		tpl_xmtTcpMeta
 *
 * Description: "Fast-path" transmit function to send hdr
 *				 over TCP transport.
 *
 * Created by : Su Wang
 *
 * Date Created: 06/19/02
 *-----------------------------------------------------------------
 */
static int32
tpl_xmtTcpMeta( void 			*pConn, 
				pkt_desc_t		*pktDesc)
{
    int32           i;
//	ushort16		flags;
	pkt_desc_t		*pDesc = NULL;
	pkt_desc_t		*prevPkt;
	int32			tcpMtuSize;
	int32			totTcpSize;
	int32			tcpSize;
	int32			hdrSize;
	int				fragCnt;
	char			*cpyPtr;


	/* if this is a retransmission,
	 * remove all queued fragments
	 */
	if (pktDesc->pTplNext) {
		TPL_FREE_PACKET(pConn, pktDesc->pTplNext);
		pktDesc->pTplNext = 0;
	}

	/* TODO, this should come from socket, since we
	 * don't know about the mtu of remote 
	 */
	tcpMtuSize	  = luc_ifGetMtuSize( pktDesc->link ) -
		PKT_ETHER_HDR_SIZE - PKT_IP_HDR_SIZE - PKT_TCP_HDR_SIZE;

	/* calculate number of fragments needed,
	 */
	hdrSize = totTcpSize = pktDesc->hdrSize;
	fragCnt = ((totTcpSize + (tcpMtuSize - 1))/ tcpMtuSize);


	/*-----------------------------------------
	 * Su, 07/30/02
	 * check there is enough pkt 
	 *-----------------------------------------
	 */
	TPL_CHK_PKT_RESOURCE(pktDesc, fragCnt);


	pktDesc->refCnt = fragCnt;

	/*-------------------------------
	 * Saved the flags. To be used by
	 * other PKTs.
	 *-------------------------------
	 */
    // flags = (pktDesc->flags & (~PKT_FLAT_BUF));
	//flags =  pktDesc->flags & ~PKT_USR_TPL_CONN;
	
	prevPkt = 0;
	cpyPtr = pktDesc->pHdr;
	for (i=0; i < fragCnt; i++)
	{
		if (i != 0)
			pDesc = TPL_ALLOC_PACKET(pConn, tcpMtuSize);
		else
			pDesc = pktDesc;
		
		if (!pDesc) {
			TPL_DBG_PRINTF(("%s:  No PKT\n", __FUNCTION__));
			luc_panic();
	        return NFX_OK;
		}


		if (hdrSize > tcpMtuSize) {
			tcpSize = tcpMtuSize;
			hdrSize -= tcpMtuSize;
			
		}
		else
			tcpSize = hdrSize;

		/*-------------------------------
		 * account for RPC marker, move
		 * back 4 bytes
		 *-------------------------------
		 */
		if ((i != 0) && 
			(((tpl_connCb_t*)pConn)->flags &  TPL_CONN_USE_RPC)) 
			pDesc->pHdr    -= 4;
		
		/*-------------------------------
		 * Save ref. pkt desc. & use 
		 * original flags.
		 *-------------------------------
		 */
		pDesc->pRefDesc = pktDesc;
		// pDesc->flags    |= flags;

		/*-------------------------------
		 * Setup hdr & pkt sizes.
		 *-------------------------------
		 */
		pDesc->hdrSize  = tcpSize;
		pDesc->pktSize  = pDesc->hdrSize + (pDesc->pHdr - pDesc->pMac);



		/*------------------------------
		 * copy content from original 
		 * buf to new one
		 *------------------------------
		 */
		if (i != 0)
			memcpy(pDesc->pHdr, cpyPtr, tcpSize);
		

		/*------------------------------
		 * advance cpyPtr
		 *------------------------------
		 */
		cpyPtr += tcpSize;
		

		/*-------------------------------
		 * Send TCP segment.
		 *-------------------------------
		 */
		TPL_DBG_PRINTF(("tcp meta send(%lx %d): totsize %d, pkt size %d, hdrsize %d\n",
			   			(address_t)pDesc, i, totTcpSize, pDesc->pktSize, 
			   			pDesc->hdrSize));
		tpl_xmtTcpPkt( pConn, pDesc);

		/* chain all the pkts */
		if (prevPkt) {
			/* chain all fragments
			 * behind the first fragment
			 */
			prevPkt->pTplNext = pDesc;
		}
		
		/*------------------------------
		 * save ptr
		 *------------------------------
		 */
		prevPkt = pDesc;
		prevPkt->pTplNext = 0;
	}

	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name:  		tpl_xmtTcpData
 *
 * Description: "Fast-path" transmit function to send USER DATA
 *				(COD) over TCP transport.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 2/20/02
 *-----------------------------------------------------------------
 */
static int32
tpl_xmtTcpData( void 			*pConn, 
				pkt_desc_t		*pktDesc, 
				int32 			fragCnt, 
				luc_fragCb_t 	*fcb)
{
    int32           i;
//	ushort16		flags;
	pkt_desc_t		*pDesc = NULL;
	pkt_desc_t		*prevPkt;


	
	/* if this is a retransmission,
	 * remove all queued fragments
	 */
	if (pktDesc->pTplNext) {
		TPL_FREE_PACKET(pConn, pktDesc->pTplNext);
		pktDesc->pTplNext = 0;
	}

	/*-----------------------------------------
	 * Su, 07/30/02
	 * check there is enough pkt 
	 *-----------------------------------------
	 */
	TPL_CHK_PKT_RESOURCE(pktDesc, fragCnt);

	pDesc = pktDesc;

	/*-------------------------------
	 * Saved the flags. To be used by
	 * other PKTs.
	 *-------------------------------
	 */
    // flags = pDesc->flags & ~PKT_USR_TPL_CONN;

	for (i=0; i < fragCnt; i++)
	{
		/*-------------------------------
		 * 1st segment has RPC marker & APP hdr.
		 *-------------------------------
		 */
		if (i != 0)
		{
			pDesc->hdrSize  = 0;
			pDesc->pHdr    -= 4;
		}

		/*-------------------------------
		 * Setup USER DATA to be sent.
		 *-------------------------------
		 */
		pDesc->codEntry[0].cod    = fcb->fcbCodEntry[0].cod;
		pDesc->codEntry[0].offset = fcb->fcbCodEntry[0].codOffset;
		pDesc->codEntry[0].length = fcb->fcbCodEntry[0].codLength;

		/*-------------------------------
		 * Save ref. pkt desc. & use 
		 * original flags.
		 *-------------------------------
		 */
		pDesc->pRefDesc = pktDesc;
		//pDesc->flags    |= flags;

        /*-------------------------------
		 * Shaun 09/11/02 bug #2075.
		 *
         * Remember pad size for last fragment.
		 * To be passed into sbm_sendDataToLuc()
		 * & be excluded from cod length check.
		 *
		 * #2458 use padSize from pRefDesc.
         *-------------------------------
         */
        if (i == (fragCnt - 1 ))
        {
            if (pktDesc->padSize)
                pDesc->padSize = pktDesc->padSize;
        }

		/*-------------------------------
		 * Setup data & pkt sizes.
		 *-------------------------------
		 */
	    pDesc->numOfCod = 1;
		pDesc->dataSize = fcb->fcbCodEntry[0].codLength;
		pDesc->pktSize  = pDesc->dataSize + pDesc->hdrSize + 
 						  (pDesc->pHdr - pDesc->pMac);


       	TPL_DBG_PRINTF(("%s: i=%d pDesc=%x pktSize=%d dataSize=%d\n", 
			__FUNCTION__, i, pDesc, pDesc->pktSize, pDesc->dataSize));

		/*-------------------------------
		 * Send TCP segment.
		 *-------------------------------
		 */
		tpl_xmtTcpPkt( pConn, pDesc);
			
		prevPkt = pDesc;

		/*-------------------------------
		 * Last segment, break loop.
		 *-------------------------------
		 */
		if (i == fragCnt-1)
			break;

		/*-------------------------------
		 * Advance frag CB.
		 *-------------------------------
		 */
		fcb++;

		/*-------------------------------
		 * Allocate next pkt desc.
		 *-------------------------------
		 */
        pDesc = TPL_ALLOC_PACKET(pConn, fcb->fcbCodEntry[0].codLength);
        if (!pDesc)
		{
			TPL_DBG_PRINTF(("%s:  No PKT\n", __FUNCTION__));
			luc_panic();
	        break;
		}

		/*-------------------------------
		 * Add new pkt to TPL chain.
		 *-------------------------------
		 */
		prevPkt->pTplNext = pDesc;

	}

	/*-------------------------------
	 * Terminate TPL chain.
	 *-------------------------------
	 */
	pDesc->pTplNext = NULL;

	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:  		tpl_xmtTcp
 *
 * Description: "Fast-path" TCP transmit function 
 *
 * Created by : Shaun Hui
 *
 * Date Created: 2/20/02
 *-----------------------------------------------------------------
 */
static int32 						/* NFX_OK or error */
tpl_xmtTcp( void 		*pConnId, 	/* TPL Connection CB */
			pkt_desc_t	*pDesc)		/* packet descriptor */
{
	tpl_connCb_t	*pConn;
	tpl_stats_t		*pStats;

    luc_fragCb_t    fragCb[LUC_MAX_FRAG_CB];
    int32           fragCnt;

	pStats = &pTplCfg->tplStats;
	pConn  = (tpl_connCb_t *)pConnId;

	/*-------------------------------
     * TCP ACK has no pkt desc.
     *-------------------------------
     */
	if (!pDesc) 
		return (tpl_xmtTcpPkt( pConnId, pDesc));

#if defined(NOT_USED)
	/* Su, 09/03/02
	 * sanity check for nfs potent packets 
	 * remove later before checkin
	 */
	if (pDesc->hdrSize > 1460
		&& (pDesc->flags & PKT_USR_APPL_OWN))
		TPL_LOGERR(TPL_ERR_BAD_PKT, pDesc);
#endif /* NOT_USED */

	/*-------------------------------
     * Setup the RPC marker for 1st 
	 * TCP Segment only.
     *-------------------------------
     */
	if ((pConn->flags & TPL_CONN_USE_RPC) && (pDesc) &&
		(!(pDesc->flags & PKT_RPC_MARKER)))
	{
		uint32	rpcMarker;
		uint32	*bufp;

        rpcMarker = (0x80000000 | (pDesc->hdrSize + pDesc->dataSize));

		/*-------------------------------
	     * Adjust pHdr to include RPC marker.
	     *-------------------------------
    	 */
        pDesc->pHdr    -= 4 ;
        pDesc->hdrSize += 4 ;
        pDesc->flags   |= PKT_RPC_MARKER;

		/*-------------------------------
	     * Write the RPC Maker.
	     *-------------------------------
    	 */
        bufp  = (uint32 *)pDesc->pHdr;
        *bufp = htonl(rpcMarker);
	}

	/*-------------------------------
     * No USER DATA, send pkt.
     *-------------------------------
   	 */
	if (pDesc->numOfCod == 0) {

		/*----------------------------
		 * su, 06/20/02.
		 * send segmented tcp pkts.
		 * if pkt size is larger than
		 * tcp mtu
		 *----------------------------
		 */
		
		int tcpMtuSize;
		
		tcpMtuSize = luc_ifGetMtuSize( pDesc->link ) -
			PKT_ETHER_HDR_SIZE - PKT_IP_HDR_SIZE - PKT_TCP_HDR_SIZE;

		if (tcpMtuSize < pDesc->hdrSize)
			return tpl_xmtTcpMeta(pConnId, pDesc);
		else
			return (tpl_xmtTcpPkt( pConnId, pDesc));
	}
	
	/*-------------------------------
     * Segment USER DATA.
     *-------------------------------
   	 */
	luc_fragDataReq( pDesc, &fragCnt, fragCb );
	luc_showFragCb( fragCnt, fragCb );

	/*-------------------------------
     * Send USER DATA in TCP Segments.
     *-------------------------------
   	 */
	pDesc->refCnt = fragCnt;
	return(tpl_xmtTcpData( pConnId, pDesc, fragCnt, fragCb ));
}

#endif /* #ifndef PROM */

/*-----------------------------------------------------------------
 * Name:  		tpl_xmtUdpMeta
 *
 * Description: "Fast-path" transmit function to send hdr, if 
 *				exceeds interface mtu size. We will not 
 *				support fragmentation of both hdr and cod at the 
 *				same time.
 *
 * Created by : Su Wang
 *
 * Date Created: 06/19/02
 *-----------------------------------------------------------------
 */
static int32
tpl_xmtUdpMeta(void			*pConn,
			  pkt_desc_t	*pktDesc)
{
	pkt_desc_t	*pDesc = NULL;
	int32		fragCnt;
	int32		udpMtuSize;
	int32		totUdpSize;
	int32		udpSize = 0;
	char		*cpyPtr;
	ushort		offset;
	int32		hdrSize;
//	ushort16	flags;
	pkt_desc_t *prevPkt;
    struct ip	*ipHdr;
	int			i;

	extern ushort16 ip_id;

	/* if this is a retransmission,
	 * remove all queued fragments
	 */
	if (pktDesc->pTplNext) {
		TPL_FREE_PACKET(pConn, pktDesc->pTplNext);
		pktDesc->pTplNext = 0;
	}

	udpMtuSize	  = luc_ifGetMtuSize( pktDesc->link ) -
		PKT_ETHER_HDR_SIZE - PKT_IP_HDR_SIZE;
	
	/* save flags for all the fragment */
	//flags = pktDesc->flags & ~PKT_USR_TPL_CONN;

	/* calculate number of fragments needed */
	hdrSize = totUdpSize = pktDesc->hdrSize + PKT_UDP_HDR_SIZE;
	fragCnt = ((int32)((totUdpSize + (udpMtuSize - 1))/udpMtuSize));

	
	/*-----------------------------------------
	 * Su, 07/30/02
	 * check there is enough pkt 
	 *-----------------------------------------
	 */
	TPL_CHK_PKT_RESOURCE(pktDesc, fragCnt);

	// printf("****** fragCount is %d\n", fragCnt);
	
	pktDesc->refCnt = fragCnt;

	prevPkt = 0;
	cpyPtr = pktDesc->pHdr;
	for (i=0; i< fragCnt; i++) {
					
		if (i != 0)
			pDesc = TPL_ALLOC_PACKET(pConn, udpMtuSize);
		else
			pDesc = pktDesc;
		
		if (!pDesc) {
			TPL_DBG_PRINTF(("%s:  No PKT\n", __FUNCTION__));
			luc_panic();
			break;
		}

		
		/* calculate the currect udp size, 
		 * and corresponding offset 
		 */
		offset = (totUdpSize - hdrSize)>>3;
		
		if (hdrSize > udpMtuSize) {
			udpSize = udpMtuSize;
			hdrSize -= udpMtuSize;
			offset |= IP_MF;
		}
		else
			udpSize = hdrSize;
		

			
		/*-------------------------------
         * Set fragment offset.
		 * Same IP ID for all fragments.
         *-------------------------------
         */ 
        ipHdr = (struct ip *)pDesc->pIpHdr;
        ipHdr->ip_off = htons(offset);
        ipHdr->ip_id  = htons(ip_id);
		

		/*-------------------------------
         * 1st fragment, set UDP length.
         *-------------------------------
         */
		if (i == 0) {
            struct  udphdr  *udpHdr;
            ushort16        udpLen;

            udpLen = totUdpSize;
            udpHdr = (struct udphdr *)pDesc->pTplHdr;
            udpHdr->uh_ulen = htons(udpLen);
            udpHdr->uh_sum  = 0;

 			/* su, 06/11/02, set fragment flag */
 			if (ipHdr->ip_off) {
 				pDesc->flags |= PKT_IP_FRAGMENT | PKT_IP_FIRST_FRAG;
			}

			/* copy to first fragment, leave udp header */
			pDesc->hdrSize = udpSize - PKT_UDP_HDR_SIZE;
			udpSize = pDesc->hdrSize;
		}
		else {
		
			/* fragments, move pHdr to tplHdr */
			pDesc->pHdr = pDesc->pTplHdr;
			

			pDesc->hdrSize = udpSize;

			/* su, 06/13/02, set frag flags */
			pDesc->flags |= PKT_IP_FRAGMENT;
		}


				
		/*-------------------------------
		 * Save ref. pkt desc. 
		 *-------------------------------
		 */
		pDesc->pRefDesc = pktDesc;

		/* update pkt size */
		pDesc->pktSize = pDesc->hdrSize + 
			(pDesc->pHdr - pDesc->pMac);

		/* copy content */
		if (i != 0)
			memcpy(pDesc->pHdr, cpyPtr, udpSize);
		
		/* advance cpy ptr */
		cpyPtr += udpSize;
		
				
		/* cannot send first fragment, need to
		 * do cksum
		 */
		if (i != 0) {
			TPL_DBG_PRINTF(("Udp meta send(%lx %d): totsize %d, pkt size %d, hdrsize %d, offset %d\n",
							(address_t)pDesc, i, totUdpSize, pDesc->pktSize, 
							pDesc->hdrSize, (offset&IP_OFFMASK)));
		
			/*-------------------------------
			 * Send UDP pkt or IP fragment.
			 *-------------------------------
			 */
			tpl_xmtUdpPkt( pConn, pDesc);
		}

		/* chain all the pkts */
		if (prevPkt) {
			/* chain all fragments
			 * behind the first fragment
			 */
			prevPkt->pTplNext = pDesc;
		}
			
		prevPkt = pDesc;
		prevPkt->pTplNext = 0;
	}

	/* now send first fragment */
	TPL_DBG_PRINTF(("Udp meta send(%lx %d): totsize %d, pkt size %d, hdrsize %d, offset %d\n",
					(address_t)pDesc, 0, totUdpSize, pDesc->pktSize, 
					pDesc->hdrSize, (offset&IP_OFFMASK)));
	tpl_xmtUdpPkt( pConn, pktDesc);
	
	ip_id++;

	return (NFX_OK);
}


/*-----------------------------------------------------------------
 * Name:  		tpl_xmtUdpData
 *
 * Description: "Fast-path" transmit function to send USER DATA
 *				(COD) over UDP transport.
 *
 * Created by : Shaun Hui
 *
 * Date Created: 2/20/02
 *-----------------------------------------------------------------
 */
static int32
tpl_xmtUdpData( void 			*pConn, 
				pkt_desc_t		*pktDesc, 
				int32 			fragCnt, 
				luc_fragCb_t 	*fcb)
{
    int32           i, j;
    int32           dataSize;
//	ushort16		flags;
	pkt_desc_t		*pDesc;
	pkt_desc_t		*prevPkt;
    struct ip 		*ipHdr;
	pkt_desc_t		*firstPkt, *chainStart;

	extern ushort16 ip_id;

	pDesc = pktDesc;

	/*-------------------------------
	 * Saved the flags. To be used by
	 * other PKTs.
	 *-------------------------------
	 */
    //flags    = pDesc->flags & ~PKT_USR_TPL_CONN;
	dataSize = pDesc->dataSize;

	
	/* su, 06/13/02, need to send frags
	 * in reverse order.
	 */

	if (pktDesc->pTplNext) {
		TPL_FREE_PACKET(pConn, pktDesc->pTplNext);
	}

	/*-----------------------------------------
	 * Su, 07/30/02
	 * check there is enough pkt 
	 *-----------------------------------------
	 */ 
	TPL_CHK_PKT_RESOURCE(pktDesc, fragCnt);
	
	fcb += (fragCnt - 1);
	firstPkt = pktDesc;
	prevPkt = NULL;
	chainStart = NULL;
	for (i = (fragCnt-1); i >=0; i--)
	{

		/* su, allocate new desc for non-first
		 * fragments
		 */
		if (i != 0) {
			pDesc = TPL_ALLOC_PACKET(pConn, fcb->fcbCodEntry[0].codLength);
			if (!pDesc) {
				TPL_DBG_PRINTF(("%s:  No PKT\n", __FUNCTION__));
				luc_panic();
				break;
			}
		}
		else
			pDesc = firstPkt;
				
        /*-------------------------------
         * Set fragment offset & MF bits.
		 * Same IP ID for all fragments.
         *-------------------------------
         */ 
        ipHdr = (struct ip *)pDesc->pIpHdr;
        ipHdr->ip_off = htons(fcb->ipOffset);
        ipHdr->ip_id  = htons(ip_id);
    
        /*-------------------------------
         * 1st fragment, set UDP length.
         *-------------------------------
         */
		if (i == 0)
        {
            struct  udphdr  *udpHdr;
            ushort16        udpLen;

            udpLen = PKT_UDP_HDR_SIZE + pDesc->hdrSize + dataSize;
            udpHdr = (struct udphdr *)pDesc->pTplHdr;
            udpHdr->uh_ulen = htons(udpLen);
            udpHdr->uh_sum  = 0;

 			/* su, 06/11/02, set fragment flag */
 			if (ipHdr->ip_off) {
 				pDesc->flags |= PKT_IP_FRAGMENT | PKT_IP_FIRST_FRAG;
			}
			
		}
		else
		{

	        /*-------------------------------
	         * Excl. UDP & APP Hdrs in fragment. 
	         *-------------------------------
	         */
			pDesc->hdrSize  = 0;

			/* su, 06/19/02, move pHdr to pTplHdr.
			 * so that pktSize can be calculated
			 * correctly
			 */
			pDesc->pHdr = pDesc->pTplHdr;
			
			/* su, 06/13/02, set frag flags */
			pDesc->flags |= PKT_IP_FRAGMENT;
				
		}


        /*-------------------------------
         * Save ref. pkt desc. & use
         * original flags.
         *-------------------------------
         */
		pDesc->pRefDesc = pktDesc;

        /*-------------------------------
		 * Shaun 09/11/02 bug #2075.
		 *
         * Remember pad size for last fragment.
		 * To be passed into sbm_sendDataToLuc()
		 * & be excluded from cod length check.
		 *
		 * #2458 use padSize from pRefDesc.
         *-------------------------------
         */
        if (i == (fragCnt - 1 ))
        {
            if (pktDesc->padSize)
                pDesc->padSize = pktDesc->padSize;
        }

        /*-------------------------------
         * Setup data size.
         *-------------------------------
         */
		pDesc->dataSize = 0;
	    pDesc->numOfCod = fcb->fcbCodNum;
		for (j = 0; j < pDesc->numOfCod; j++) {
			
			pDesc->codEntry[j].cod    = fcb->fcbCodEntry[j].cod;
			pDesc->codEntry[j].offset = fcb->fcbCodEntry[j].codOffset;
			pDesc->codEntry[j].length = fcb->fcbCodEntry[j].codLength;
			pDesc->dataSize += fcb->fcbCodEntry[j].codLength;
		}
		
		/* su, 06/19/02, setup pkt size */
		pDesc->pktSize  = pDesc->dataSize + pDesc->hdrSize + 
			(pDesc->pHdr - pDesc->pMac);

       	TPL_DBG_PRINTF(("%s: i=%d pDesc=%x pktSize=%d dataSize=%d\n", 
			__FUNCTION__, i, pDesc, pDesc->pktSize, pDesc->dataSize));

		
        /*-------------------------------
         * Add new pkt to TPL chain.
         *-------------------------------
         */

		/* su, 06/13/02, little twisted
		 * pkt chain must start with firstPkt,
		 * right? While fragments are sent
		 * in reverse order.
		 */
		if (prevPkt) {
			/* chain non first fragments */
			if (pDesc != firstPkt) 
				prevPkt->pTplNext = pDesc;
			else {
				/* chain all fragments
				 * behind the first fragment
				 */
				prevPkt->pTplNext = 0;
				pDesc->pTplNext = chainStart;
			}
		}
		else {
			chainStart = pDesc;
			pDesc->pTplNext = NULL;	/* in case single fragment */
		}

        /*-------------------------------
         * Send UDP pkt or IP fragment.
         *-------------------------------
         */
		tpl_xmtUdpPkt( pConn, pDesc);
			

		fcb --;
		
		prevPkt = pDesc;
		
	}

	ip_id++;
	return (NFX_OK);
}

/*-----------------------------------------------------------------
 * Name:  		tpl_xmtUdp
 *
 * Description: "Fast-path" Udp transmit function 
 *
 * Created by : Shaun Hui
 *
 * Date Created: 2/20/02
 *-----------------------------------------------------------------
 */
static int32 						/* NFX_OK or error */
tpl_xmtUdp( void 		*pConnId, 	/* TPL Connection CB */
			pkt_desc_t	*pDesc)		/* packet descriptor */
{
	tpl_connCb_t	*pConn;
	tpl_stats_t		*pStats;

    luc_fragCb_t    fragCb[LUC_MAX_FRAG_CB];
    int32           fragCnt;

	/* su, 05/06/02.
	 * incr. udp stats counter 
	 */
	udpstat.udps_opackets++;

	pStats = &pTplCfg->tplStats;
	pConn  = (tpl_connCb_t *)pConnId;


	pDesc->pktSize = pDesc->dataSize + pDesc->hdrSize + 
					 (pDesc->pHdr - pDesc->pMac);

#if defined(NOT_USED)
	/* Su, 09/03/02
	 * sanity check for nfs potent packets 
	 * remove later before checkin
	 */
	if (pDesc->hdrSize > 1472
		&& (pDesc->flags & PKT_USR_APPL_OWN))
		TPL_LOGERR(TPL_ERR_BAD_PKT, pDesc);
#endif /* NOT_USED */

#ifdef PROM

	/*-------------------------------
	 * PROM has No USER DATA. Send pkt. 
	 *-------------------------------
	 */
	return (tpl_xmtUdpPkt( pConnId, pDesc));

#else

	/*-------------------------------
     * No USER DATA, send pkt.
     *-------------------------------
   	 */
	if (pDesc->numOfCod == 0) {

		/*----------------------------
		 * su, 06/20/02.
		 * send fragmented ip pkts.
		 * if pkt size is larger than
		 * udp mtu
		 *----------------------------
		 */	
		
		int	udpMtuSize;
		
		udpMtuSize = luc_ifGetMtuSize( pDesc->link ) -
			PKT_ETHER_HDR_SIZE - PKT_IP_HDR_SIZE;

		if (udpMtuSize < (pDesc->hdrSize + PKT_UDP_HDR_SIZE))
			return tpl_xmtUdpMeta(pConnId, pDesc);
		else
			return (tpl_xmtUdpPkt( pConnId, pDesc));
	}
	
	/*-------------------------------
     * Segment USER DATA.
     *-------------------------------
   	 */
	pDesc->tplProt = TPL_PROT_UDP;
	luc_fragDataReq( pDesc, &fragCnt, fragCb );
	luc_showFragCb( fragCnt, fragCb );

	/* su, 05/06/02.
	 * inc. ip fragmentation cnt 
	 */
	ipstat.ips_fragmented++;
	ipstat.ips_ofragments += fragCnt;

	/*-------------------------------
     * Send USER DATA in TCP Segments.
     *-------------------------------
   	 */
	pDesc->refCnt = fragCnt;
	
	return(tpl_xmtUdpData( pConnId, pDesc, fragCnt, fragCb ));

#endif /* #ifdef PROM */

}

/*-----------------------------------------------------------------
 * Name:  		tpl_xmtUdp
 *
 * Description: "Fast-path" UDP transmit function 
 *
 * Created by : Shaun Hui
 *
 * Date Created: 2/20/02
 *-----------------------------------------------------------------
 */
static int32 							/* NFX_OK or error */
tpl_xmtUdpPkt(void 			*pConnId, 	/* TPL Connection CB */
		      pkt_desc_t	*pDesc)		/* packet descriptor */
{
	tpl_connCb_t	*pConn;
	tpl_stats_t		*pStats;


	pStats = &pTplCfg->tplStats;
	pConn  = (tpl_connCb_t *)pConnId;

	/*-------------------------------
	 * Make sure hdr size is ok.
	 *-------------------------------
	 */
	if (pDesc->hdrSize > (PKT_BUFF_SIZE-PKT_BUFF_INDENT))
	{
		TPL_LOGERR( TPL_ERR_BAD_PKT, pDesc);
		return (TPL_ERR_BAD_PKT);
	}

	pDesc->flags |=  PKT_UDP_CSUM_ENA | PKT_IP_CSUM_ENA | PKT_FAST_PATH;

#if !defined(PROM)
	/*-------------------------------
	 * Check if user data exists. 
	 *-------------------------------
	 */
	if (pDesc->dataSize)
	{
#if defined(LINK_AGG)
		/* update pdesc */
		if (pDesc->link != pConn->linkNum)
			tpl_UpdatePkt(pDesc, pConn, TPL_PRING_CHGE_SRCMAC);
#endif /* LINK_AGG */

		/*-------------------------------
 		 * Get it from GMC. luc_rcvGmcData()
		 * will rx Hdr+data pkt & send it.
		 *-------------------------------
		 */
		if (luc_xmtGmcDataReq( pDesc ) != NFX_OK)
		{
			/*-------------------------------
 			 * Error in MCPU sending to GMC.
			 * Need to free PKT 
			 *-------------------------------
			 */
			pStats->cntXmtFail++;
			return TPL_ERR_XMT_FAIL;
		}
		
		pStats->cntXmtPkt++;
		
		return (NFX_OK);
	}
#endif 

	/*-------------------------------
	 * Set IP len & checksum.
	 *-------------------------------
	 */
	tpl_setIpHdr(pConn, pDesc);


	/*-------------------------------
	 * Set UDP len & checksum.
	 *-------------------------------
	 */
	tpl_setUdpHdr(pConn, pDesc);


#if defined(LINK_AGG)
		/* update pdesc */
		if (pDesc->link != pConn->linkNum)
			tpl_UpdatePkt(pDesc, pConn, TPL_PRING_CHGE_SRCMAC);
#endif /* LINK_AGG */

	/*-------------------------------
	 * Call LUC to xmt pkt.
	 *-------------------------------
	 */
	if (luc_xmtPktReq( pConn->linkNum, pDesc ) != NFX_OK)
	{
		pStats->cntXmtFail++;
		return TPL_ERR_XMT_FAIL;
	}

	pStats->cntXmtPkt++;

	return (NFX_OK);
}
