/*
 *-----------------------------------------------------------------
 *
 * Name:	tssccomm-bsd.c
 *
 * RCS:  $Id: tssccomm-bsd.c,v 1.7 2002/04/29 17:31:05 maximk Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	Communication between Toshiba and SSC
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	01/31/2002
 *
 *-----------------------------------------------------------------
 */


#include <sys/param.h>
#include <sys/mbuf.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/device.h>
#include <net/if.h>
#include <net/if_types.h>
#include <net/netisr.h>
#include <nfx-types.h>
#include "tssccomm.h"
#include <mips64/memconf.h>

#include "bpfilter.h"

#if NBPFILTER
#include <net/bpf.h>
#endif

#if defined(DEBUG)
extern int debug_ethereee;
#define debug(x) { if (debug_ethereee) printf x; }
#else
#define debug(x)
#endif


extern int warm_boot;

void tssccommattach(int);
int tsscioctl(struct ifnet *ifp, u_long cmd, caddr_t data);
int tsscoutput(struct ifnet *ifp, struct mbuf *m, struct sockaddr *dst,
			   struct rtentry *rt);
tsc_message_t **tsc_ssc_save_last_tx_addr(void);
tsc_message_t **tsc_ssc_save_last_rx_addr(void);
int tsc_ssc_check_rings(void);

tsc_sc_t tsc_sc;


tsc_message_t **
tsc_ssc_save_last_tx_addr()
{
	return (tsc_message_t **)(tsc_ssc_rx_ring_addr() + TSC_RING_LENGTH + 10);
}


tsc_message_t **
tsc_ssc_save_last_rx_addr()
{
	return (tsc_ssc_save_last_tx_addr() + 1);
}


void
tsc_copy_message_data(char *dest, data_descriptor_t *data,
				  uint32 offset, uint32 bytecount)
{
	m_copydata(data, offset, bytecount, dest);
}


int
tsc_data_length(data_descriptor_t *data)
{
	return data->m_pkthdr.len;
}


data_descriptor_t *
tsc_make_data_descriptor(struct tsc_sc *sc, tsc_message_t *message)
{
	int totlen = tsc_calculate_packet_length(sc, message);
	struct mbuf *m;
	struct mbuf *top = NULL, **mp = &top;

	MGETHDR(m, M_DONTWAIT, MT_DATA);
	if (m) {
		int len;
		int offset = 0;			/* Offset in copy from */
		
		m->m_pkthdr.rcvif = &sc->sc_if;
		m->m_pkthdr.len = totlen;
		m->m_len = MHLEN;

		while (totlen > 0) {
			if (top != NULL) {
				MGET(m, M_DONTWAIT, MT_DATA);
				if (m == NULL) {
					m_freem(top);
					return NULL;
				}
				m->m_len = MLEN;
			}

			len = totlen;
			if (len >= MINCLSIZE) {
				MCLGET(m, M_DONTWAIT);
				if (m->m_flags & M_EXT)
					m->m_len = len = min(totlen, MCLBYTES);
				else
					len = m->m_len;
			}
			else {
				if (len < m->m_len) {
					if (top == NULL &&
						len + max_linkhdr <= m->m_len)
						m->m_data += max_linkhdr;
					m->m_len = len;
				}
				else
					len = m->m_len;
			}

			tsc_copy_from_packet(mtod(m, char *), sc, &message, &offset, len);
			
			*mp = m;
			mp = &m->m_next;
			totlen -= len;
		}
	}

	return top;
}


extern struct ifqueue eeeintrq;

static void
tsscintr(void *arg)
{
	struct ifnet *ifp = arg;
	tsc_sc_t *sc = (tsc_sc_t *)ifp->if_softc;

	while (tsc_rxpacket_ready(sc)) {
		struct mbuf *m;
		int s = splhigh();

		debug(("tsscintr: rx packet ready\n"));
		m = tsc_do_rxmessage(sc);

#if NBPFILTER > 0
		if (m && ifp->if_bpf)
			bpf_mtap(ifp->if_bpf, m);
#endif		
		if (m) {
			struct ifqueue *inq = &eeeintrq;
			schednetisr(NETISR_EEE);
			if (IF_QFULL(inq)) {
				IF_DROP(inq);
				m_freem(m);
			}
			else {
				IF_ENQUEUE(inq, m);
			}
			ifp->if_ipackets++;
		}
		tsc_rx_cleanup_packet(sc, tsc_last_rx_message(sc));
		*tsc_ssc_save_last_rx_addr() = tsc_last_rx_message(sc);
		splx(s);
	}

	timeout(tsscintr, ifp, 1);
}


int
tsscioctl(struct ifnet *ifp, u_long cmd, caddr_t data)
{
	int error = 0;
	tsc_sc_t *sc = ifp->if_softc;
	
	switch (cmd) {
	case SIOCSIFADDR:
		ifp->if_flags |= IFF_UP;

		if (debug_ethereee) {
			tsc_print_rxqueue(sc);
			tsc_print_txqueue(sc);
		}

		timeout(tsscintr, ifp, 1);

		break;
	default:
		error = EINVAL;
	}

	return error;
}


/*
 * tsc_check_ring - return true if the messages in the ring have
 * valid next indexes
 */
int
tsc_check_ring(tsc_message_t *start, int len)
{
	int i;
	tsc_message_t *msg;
	for (msg = start, i = 0; i < len - 1; ++i, ++msg) {
		debug(("msg=%x i=%d next=%d\n", msg, i,
			   tsc_next_message_index(msg)));
		if (tsc_next_message_index(msg) != i + 1) {
			return 0;
		}
	}
	debug(("msg=%x i=%d next=%d\n", msg, i, tsc_next_message_index(msg)));
	if (tsc_next_message_index(msg) != 0) {
		return 0;
	}
	return 1;
}


/*
 * tsc_ssc_check_rings - return true if the ssc transmit and receive
 * rings are valid
 */
int
tsc_ssc_check_rings(void)
{
	return tsc_check_ring((tsc_message_t *)
						  CACHED_TO_UNCACHED(tsc_ssc_tx_ring_addr()),
						  TSC_RING_LENGTH)
		&& tsc_check_ring((tsc_message_t *)
						  CACHED_TO_UNCACHED(tsc_ssc_rx_ring_addr()),
						  TSC_RING_LENGTH);
}


static void
tsc_ssc_init_tx_ring(tsc_message_t *start)
{
	tsc_init_tx_ring(&tsc_sc, start, TSC_RING_LENGTH);
	*tsc_ssc_save_last_tx_addr() = start;
}


static void
tsc_ssc_init_rx_ring(tsc_message_t *start)
{
	tsc_init_rx_ring(&tsc_sc, start, TSC_RING_LENGTH);
	*tsc_ssc_save_last_rx_addr() = start;
}


void
tssccommattach(int n)
{
	struct ifnet *ifp;
	tsc_sc_t *sc = &tsc_sc;

	ifp = &sc->sc_if;
	strcpy(ifp->if_xname, "tsc0");
	ifp->if_softc = sc;
	ifp->if_mtu = 8192;
	ifp->if_flags = IFF_POINTOPOINT | IFF_SIMPLEX;
	ifp->if_type = IFT_OTHER;
	ifp->if_hdrlen = sizeof(uint32_t);
	ifp->if_addrlen = 0;
	ifp->if_ioctl = tsscioctl;
	ifp->if_output = tsscoutput;

	Mips_HitInvalidateDCache((vm_offset_t)tsc_ssc_tx_ring_addr(),
							 TSC_RING_LENGTH * sizeof(tsc_message_t));
	Mips_HitInvalidateDCache((vm_offset_t)tsc_ssc_tx_ring_addr(),
							 TSC_RING_LENGTH * sizeof(tsc_message_t));

	assert(mem_layout[1].mem_start + mem_layout[1].mem_size
		   <= (unsigned int)(tsc_ssc_tx_ring_addr() + TSC_RING_LENGTH)
		   && (unsigned int)(tsc_ssc_tx_ring_addr() + TSC_RING_LENGTH)
		   < 0x90000000);
	assert(mem_layout[1].mem_start + mem_layout[1].mem_size
		   <= (unsigned int)(tsc_ssc_rx_ring_addr() + TSC_RING_LENGTH)
		   && (unsigned int)(tsc_ssc_rx_ring_addr() + TSC_RING_LENGTH)
		   < 0x90000000);
	printf("tsc0: tx at 0x%x rx at 0x%x\n",
		   tsc_ssc_tx_ring_addr(), tsc_ssc_rx_ring_addr());
	if (!warm_boot) {
		tsc_ssc_init_tx_ring((tsc_message_t *)
							 CACHED_TO_UNCACHED(tsc_ssc_tx_ring_addr()));
		tsc_ssc_init_rx_ring((tsc_message_t *)
							 CACHED_TO_UNCACHED(tsc_ssc_rx_ring_addr()));
	}
	else {
		tsc_reinit_tx_ring(&tsc_sc, 
						   (tsc_message_t *)
						   CACHED_TO_UNCACHED(tsc_ssc_tx_ring_addr()),
						   *tsc_ssc_save_last_tx_addr(),
						   TSC_RING_LENGTH);
		tsc_reinit_rx_ring(&tsc_sc,
						   (tsc_message_t *)
						   CACHED_TO_UNCACHED(tsc_ssc_rx_ring_addr()),
						   *tsc_ssc_save_last_rx_addr(),
						   TSC_RING_LENGTH);
	}

	if_attachhead(ifp);

#if NBPFILTER > 0
	bpfattach(&ifp->if_bpf, ifp, DLT_SHMEM, 0);
#endif	
}


void
tsc_txstart(tsc_sc_t *sc)
{
}


int
tsscoutput(struct ifnet *ifp, struct mbuf *m, struct sockaddr *dst,
		   struct rtentry *rt)
{
	tsc_sc_t *sc = (tsc_sc_t *)ifp->if_softc;
	int rc = ENOBUFS;
	
	debug(("tsscoutput: \n"));

	if ((m->m_flags & M_PKTHDR) == 0) {
		panic("tsscoutput: no header mbuf");
	}
	ifp->if_lastchange = time;
	if (tsc_check_buffers_availability(sc, tsc_data_length(m))) {
		int s = splhigh(); /* spl so the pointer to the last buffer does not
							  get out of sync */
		int put_result;

		put_result = tsc_txput_packet(sc, m);
		assert(put_result != 0);
		*tsc_ssc_save_last_tx_addr() = tsc_last_tx_message(sc);

		splx(s);

#if NBPFILTER > 0
		if (ifp->if_bpf && m)
			bpf_mtap(ifp->if_bpf, m);
#endif		
				
		m_freem(m);
		tsc_txstart(sc);
		ifp->if_opackets++;

		rc = 0;

		if (debug_ethereee) tsc_print_txqueue(sc);
	}
	else {
		printf("tsscoutput: no buffers, packet dropped\n");
	}

	return rc;
}


