/*
 *-----------------------------------------------------------------
 *
 * Name:	tssccomm-eee.c
 *
 * RCS:  $Id: tssccomm-eee.c,v 1.6 2002/10/22 01:30:43 jong Exp $
 *
 * Copyright (C) 2001, Agile Storage Inc.
 *
 * Description:	Communication between Toshiba and SSC
 *
 * Created by:	Maxim Kozlovsky
 *
 * Date Created:	01/31/2002
 *
 *-----------------------------------------------------------------
 */


#include <exports/eee-desc.h>
#include <exports/eee-drv.h>
#include <exports/eee-app.h>
#include <exports/libc.h>
#include <exports/eee-poll.h>
#include "tssccomm.h"

#if defined(DEBUG)
#define debug(x) printf x
#else
#define debug(x)
#endif

extern uint32 ssc_memory_base(void);
extern uint32 ssc_memory_map(void);

struct tsc_sc *tsc_sc_ptr;
struct tsc_sc tsc_sc;

/*
 * tsc_get_sc - return the driver structure for the descriptor
 */
static struct tsc_sc *
tsc_get_sc(eee_desc_t *desc)
{
	return tsc_sc_ptr;
}


/*
 * tsc_alloc_sc - allocate the driver structure
 */
static struct tsc_sc *
tsc_alloc_sc()
{
	return &tsc_sc;
}


/*
 * tsc_copy_message_data - copy bytecount of the data from data descriptor
 * starting at offset to buffer dest
 */
void
tsc_copy_message_data(char *dest, data_descriptor_t *data,
					  uint32 offset, uint32 bytecount)
{
	int bufno = 0;
	eee_desc_t *desc = data;
	int copied_so_far = 0;

	while (offset >= desc->bd[bufno].len) {
		offset -= desc->bd[bufno].len;
	}

	while (copied_so_far < bytecount) {
		eee_descBuf_t *bufdesc = &desc->bd[bufno];
		int copy_len = MIN(bytecount, bufdesc->len - offset);
		memcpy(dest, bufdesc->buf, copy_len);
		dest += copy_len;
		if (offset + copy_len == bufdesc->len) {
			++bufno;
			offset = 0;
		}
		else {
			offset += copy_len;
		}
		copied_so_far += copy_len;
	}
}


/*
 * allocate_buffers - allocate buffers for descriptor desc to store
 * total_len of data
 */
static int
allocate_buffers(eee_desc_t *desc, int total_len)
{
	int allocLen = 0;
	unsigned int allocBuf = 1;	/* Start from first buffer */
	uchar8 *bufptr;

	while (allocLen < total_len
		   && (bufptr =
			   eee_allocateBuffer(EEE_SHARED_POOL, EEE_BUF_LARGE, 0))) {
		eee_descBuf_t *bufdesc = &(desc->bd[allocBuf]);
		bufdesc->len = MIN(EEE_BUFSIZE_LARGE, total_len - allocLen);
		bufdesc->buf = bufptr;
		bufdesc->control = EDESC_BD_CMD_IN_USE | EDESC_BD_LARGE;
		allocLen += bufdesc->len;
		++allocBuf;
	}

	desc->hdr.control = (((allocBuf - 1) << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK);

	return allocLen == total_len;
}


/*
 * copy_to_buffers - copy the data from the message to eee descriptor
 */
static void
copy_to_buffers(eee_desc_t *desc, struct tsc_sc *sc,
			  struct tsc_message *message, int total_len)
{
	eee_descBuf_t *bufdesc = &desc->bd[1];
	int offset = sizeof(agile_hdr_t);
	agile_hdr_t *hdr = (agile_hdr_t *)tsc_message_data(message);

	total_len -= sizeof(agile_hdr_t);
	
	desc->hdr.dest_port = hdr->dest_port;
	desc->hdr.src_port = hdr->src_port;
#if defined(NETEEE_FRAGMENT)
	desc->hdr.offset = hdr->offset;
#endif	
	desc->hdr.control |= total_len;
	
	while (total_len) {
		int copyLen = MIN(total_len, EEE_BUFSIZE_LARGE);
		assert(bufdesc->control & EDESC_BD_CMD_IN_USE);
		tsc_copy_from_packet(bufdesc->buf, sc, &message, &offset, 
							  copyLen);
		total_len -= copyLen;
	}
}


/*
 * tsc_make_data_descriptor - create eee descriptor from the message
 */
data_descriptor_t *
tsc_make_data_descriptor(struct tsc_sc *sc, struct tsc_message *message)
{
	eee_desc_t *desc;
	int total_len;

	assert(tsc_receiver_owns(message) && tsc_is_first_message(message));

	total_len = tsc_calculate_packet_length(sc, message);

	if ((desc = eee_allocateDesc(EEE_SHARED_POOL, NFX_COS_HIGH))
		&& (allocate_buffers(desc, total_len - sizeof(agile_hdr_t)))) {
		copy_to_buffers(desc, sc, message, total_len);
	}
	else {
		eee_deallocateDesc(desc);
		desc = 0;
	}

	return desc;
}


/*
 * tsc_interrupt_ssc - issue an interrupt on SSC
 */
static void
tsc_interrupt_ssc(void)
{
}


/*
 * tsc_rxdone - finish rx processing
 */
static void
tsc_rxdone(struct tsc_sc *sc)
{
	/* Signal SSC that some buffers are free */
	tsc_interrupt_ssc();
}


/*
 * tsc_poll - poll for arriving messages and receive them
 */
static void
tsc_poll(void *cb, uint32 limit)
{
	struct tsc_sc *sc = cb;
	eee_desc_t *edesc;
	
	while (tsc_rxpacket_ready(sc) &&
		   (edesc = tsc_make_data_descriptor(sc, tsc_last_rx_message(sc)))) {
		tsc_rx_cleanup_packet(sc, tsc_last_rx_message(sc));
		tsc_rxdone(sc);
		eee_receivePacket(edesc);
	}
}


/*
 * tsc_txstart - start the transmission
 */
void
tsc_txstart(struct tsc_sc *sc)
{
	tsc_interrupt_ssc();
}


/*
 * tsc_data_length - return the length of data in the descriptor
 */
int
tsc_data_length(data_descriptor_t *data)
{
	return EEE_GET_PKT_LEN(data);
}


/*
 * tsc_prepend_header - add agile header to the descriptor
 */
static eee_desc_t *
tsc_prepend_header(eee_desc_t *edesc)
{
	int len = EEE_GET_PKT_LEN(edesc);
	char *buf;
	int bufcnt = EEE_GET_NUM_BUFS(edesc);

	debug(("tsc_prepend_header\n"));

	assert(!(edesc->hdr.control & EDESC_COMMAND_USE_FIRST_BUFFER));

	len += sizeof(agile_hdr_t);
	++bufcnt;
	if ((buf = eee_allocateBuffer(EEE_SHARED_POOL, EEE_BUF_LARGE, 0))) {
		agile_hdr_t *agile_hdr = (agile_hdr_t *)buf;
		eee_descBuf_t *bufdesc = &edesc->bd[0];
		bufdesc->buf = buf;
		bufdesc->len = sizeof(agile_hdr_t);
		bufdesc->control = EDESC_BD_CMD_IN_USE | EDESC_BD_LARGE;
		edesc->hdr.control = (len & EDESC_TOTAL_LEN)
			| ((bufcnt << EDESC_BUFCNT_SHIFT) & EDESC_BUFCNT_MASK)
			| EDESC_COMMAND_USE_FIRST_BUFFER;
		agile_hdr->dest_port = edesc->hdr.dest_port;
		agile_hdr->src_port = edesc->hdr.src_port;
		agile_hdr->pad = 0;
#if defined(NETEEE_FRAGMENT)
		agile_hdr->offset = edesc->hdr.offset;
#endif		
	}

	return edesc;
}


/*
 * tsc_txUnicast - transmit desciptor
 */
static int
tsc_txUnicast(uint32 dest, eee_desc_t *edesc)
{
	struct tsc_sc *sc = tsc_get_sc(edesc);
	int rc = EEE_TX_QUEUE_FULL;
	eee_desc_t *desc_with_header;
	
	debug(("tsc_txUnicast: dest=%x edesc=%x len=%d\n", dest, edesc,
		   tsc_data_length(edesc)));

	if (tsc_check_buffers_availability(sc, tsc_data_length(edesc) +
									   sizeof(agile_hdr_t))
		&& (desc_with_header = tsc_prepend_header(edesc))) {
		int put_result __attribute__((unused));
		put_result = tsc_txput_packet(sc, desc_with_header);
		assert(put_result != 0);
		eee_freePacket(desc_with_header);
		tsc_txstart(sc);

		rc = NFX_OK;
	}

	return rc;
}


/*
 * tsc_tx_address - address of tx ring
 */
static tsc_message_t *
tsc_toshiba_tx_address()
{
	/* toshiba tx ring is SSC rx ring */
	
	assert((uint32)tsc_ssc_rx_ring_addr() > ssc_memory_base());
	return (tsc_message_t *)((uint32)tsc_ssc_rx_ring_addr() -
							 ssc_memory_base() + ssc_memory_map());
}


/*
 * tsc_rx_address - address of rx ring
 */
static tsc_message_t *
tsc_toshiba_rx_address()
{
	/* toshiba rx ring is SSC tx ring */
	
	assert((uint32)tsc_ssc_tx_ring_addr() > ssc_memory_base());
	return (tsc_message_t *)((uint32)tsc_ssc_tx_ring_addr() -
							 ssc_memory_base() + ssc_memory_map());
}


static tsc_sc_t *
tsc_init_sc()
{
	tsc_sc_t *sc = tsc_alloc_sc();

	tsc_init_tx_ring(sc, tsc_toshiba_tx_address(), TSC_RING_LENGTH);
	tsc_init_rx_ring(sc, tsc_toshiba_rx_address(), TSC_RING_LENGTH);

	tsc_sc_ptr = sc;

	return sc;
}


/*
 * tsc_init - initialize messaging between ssc and toshiba
 */
void
tsc_init()
{
	tsc_sc_t *sc;
	
	printf("initialize messaging with SSC\n");

	sc = tsc_init_sc();
	
	eee_registerTransmitFunc(SSC_PORT_LOCAL_QUEUE,
							 tsc_txUnicast, tsc_txUnicast,
							 8192);
	eee_registerPollingFunc(EEE_HIGH_PRIORITY, tsc_poll, sc);
}

