/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: Cluster Configuration Manager (ccm.c)
 *
 * RCS:  $Id: ccm.c,v 1.72 2002/10/04 02:17:06 gregs Exp $
 *
 * Copyright (C) 2002, Agile Storage, Inc.
 *
 * Description: Maintains the cluster's configuration.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 12 Feb 2002
 *
 *-----------------------------------------------------------------
 */
#include <signal.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>

#include "nfx-defs.h"
#include "nfx-ports.h"

#include "exports/anpssc-api.h"
#include "exports/opt-api.h"
#include "exports/evm-api.h"
#include "exports/ea-api.h"
#include "exports/rpc.h"	// Because mount.h needs it that's why
#include "exports/mount.h"
#include "exports/ipmd-lib.h"
#include "exports/elog-api.h"
#include "exports/event-api.h"
#include "exports/timekeep-api.h"

#include "exports/ncm-api.h"

#include "exports/ccm-api.h"
#include "ccm-loc.h"

static pid_t mypid;
uint32 myappid;

static int we_have_a_pcc = 0;	// Initially, none elected
int we_have_a_scc = 0;			// Initially, none elected
static int appSock;				// ccm registered socket fd

/*-----------------------------------------------------------------
 * Name: termHand
 *
 * Description: Where we go to die
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 13 Feb 2002
 *
 *-----------------------------------------------------------------
 */
static void
termHand (int sig)
{
/*
	This is a normal exit path so clean things up nicely
 */
	E_LOG (class_1, info_s, myappid, 0, 0, 0,
	       ("ccmd normal exit"));
	exit (0);
}

/*-----------------------------------------------------------------
 * Name: abnormal
 *
 * Description: Where we go to die under abnormal conditions
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 13 Feb 2002
 *
 *-----------------------------------------------------------------
 */
static void
abnormal (int sig)
{
/*
	Try to send an event indicating we're dying a painful death
 */
	E_LOG (class_1, debug_s, myappid, 0, 0, 0,
	       ("ccmd taking abnormal abort on signal %d", sig));
	signal (sig, SIG_DFL);		// Let it dump core now
	return;
}

static void
reset_slave_oper_cfg (char *node)
{
	char **slaves;
	char **key;
	char *which;

	if (node) {
		asprintf (&which, "slave.%s.*", node);
	}
	else {	// The entire operational state
/*
	The names will come back with each visible event we get
 */
		char **names = get_cfg_val ("lvname.*");
		if (names) {
			char **whack;
			for (whack = names; *whack; whack += 2) {
				del_cfg_obj (&whack[0]);
			}
			free_cfg_val (names);
		}

		which = strdup ("slave.*");
	}

	slaves = get_cfg_val (which);
	free (which);

	if (!slaves) {
		return;
	}

	for (key = slaves; *key; key +=2)
	{
		del_cfg_obj (key);
	}
	free_cfg_val (slaves);
}

static char *serv_objs[] = 
{"cluster.lv.*", "cluster.nfs.*", "cluster.ip.*", NULL};

static void
whack_cluster_service_cfg (void)
{
	char **key;
	char **objs;
	char **look;

	for (key = serv_objs; *key; key++) {
		objs = get_cfg_val (*key);
		if (objs) {

			for (look = objs; *look; look += 2) {
				del_cfg_obj (&look[0]);
			}

			free_cfg_val (objs);
		}
	}
}

/*-----------------------------------------------------------------
 * Name: syncNode
 *
 * Description: Synchronize configuration with the given node
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 4 Mar 2002
 *
 *-----------------------------------------------------------------
 */
typedef enum {
	pcc,
	scc,
	peon
} my_assignment_t;

static my_assignment_t i_am = peon;
static char *my_pcc_name = NULL;
static int my_pcc_failed = 0;

static void
syncNode (nfx_eventMsg_t *new)
{
	int sz;
	char *membKey;
	char **cfgObj;
	char myname[256];
	int was_pcc;
	int del;
	char *evt_type;
/*
	Boolean indicating node's state
 */
	del = new->param.nodeEvent.state == EVENT_STATE_DOWN;
/*
	Debuggage for later E_LOG()s that print event type
 */
	switch (new->eventType) {
		case NFX_EVENT_PCC:
			evt_type = "pcc";
			break;
		case NFX_EVENT_SCC:
			evt_type = "scc";
			break;
		default:
			evt_type = "node";
	}
/*
	First off, let's see if we should update our node type operational
	assignment.
 */
	was_pcc = node_is_PCC();

	gethostname (myname, sizeof (myname));
	if (!strcasecmp (myname, new->nodeName)) {	// Tis I
		char *opertype[3];

		switch (new->eventType) {
			case NFX_EVENT_PCC:
				opertype[1] = "pcc";
				i_am = pcc;
				break;
			case NFX_EVENT_SCC:
				opertype[1] = "scc";
/*
		When I become SCC, I'll need to clear leftover cluster service
		config objects. The PCC will synchronize with us.
 */
				if (!del) {
					whack_cluster_service_cfg();
				}
				i_am = scc;
				break;
			default:
				opertype[1] = NULL;
				i_am = peon;
		}

		E_LOG (class_6, debug_s, myappid, 0, 0, 0,
		       ("I am now %s", opertype[1] ? opertype[1] : "Nothing"));

		if (opertype[1]) {
			opertype[0] = "node.type";
			opertype[2] = NULL;
			if (!del) {
				put_cfg_val (opertype);
			}
			else {
				del_cfg_obj (opertype);
			}
		}
	}
/*
	I need to record my PCC's "real" name in case I become SCC and he
	fails.
 */
	else if (!del && (new->eventType == NFX_EVENT_PCC)) {
		if (my_pcc_name) {
			free(my_pcc_name);
		}
		my_pcc_name = strdup (new->nodeName);
	}
/*
	See if this node is known. If not, create its root config object
 */
	asprintf (&membKey, "cluster.member.%s", new->nodeName);

	cfgObj = get_cfg_val (membKey);
/*
	Node Failure. Let's see if we need to transfer its services to
	another node.
 */
	if (del) {

		if (cfgObj) {
			del_cfg_obj (cfgObj);
		}
		free (membKey);
		reset_slave_oper_cfg (new->nodeName); // Lose the oper state

		E_LOG (class_6, info_s, myappid, 0, 0, 0, 
		       ("Node %s/%s down", new->nodeName, evt_type));

		if (node_is_PCC()) {
			node_failed (new->nodeName);	// Transfer its services
		}
/*
	When the PCC fails, the SCC needs to ensure its services get taken over.
	It's probable we'll see a node down event for our PCC before we get
	elected to that position. Being the SCC, we have the configuration to
	allow us the ability to reassign its resources when we become PCC.
 */
		else if ((i_am == scc) && my_pcc_name && 
		                          !strcasecmp (new->nodeName, my_pcc_name)) {
			my_pcc_failed = 1;
		}
	} 
	else { 
		E_LOG (class_6, debug_s, myappid, 0, 0, 0, 
		       ("Node %s/%s up", new->nodeName, evt_type));

		if (!cfgObj) {
			cfgObj = malloc (3 * sizeof(char *));
			cfgObj[0] = membKey;
		}
		else {	// Just rewrite it with the updated addresses
			free (cfgObj[1]);
		}

		asprintf (&cfgObj[1], "%08x:%08x", new->param.nodeEvent.mgmt_ip[0], 
		                                   new->param.nodeEvent.mgmt_ip[1]);
		cfgObj[2] = NULL;

		sz = put_cfg_val (cfgObj);

		if (sz) {
			E_LOG (class_1, alert_s, myappid, 0, 0, 0,
			       ("syncNode error putting root obj"));
		}
	}
	free_cfg_val (cfgObj);
/*
	If I just became the PCC, I'll need to ask myself for a config
	update to get the visible LV restored.
 */
	if (!was_pcc && node_is_PCC()) {
		char **cl_ntp;
		char **my_ntp;
		char *tmp;
		char *t2;

		ask_node_for_update (myname);
/*
	When I become PCC, I need to ensure the cluster has a good NTP source
	mode.
 */
		cl_ntp = get_cfg_val ("cluster.ntp.mode");
		my_ntp = get_cfg_val ("ntp.mode");
		if (!cl_ntp) {
/*
	If its not there, use mine
 */
			if (my_ntp) {
				asprintf (&tmp, "cluster.%s", my_ntp[0]);
				free (my_ntp[0]);
				my_ntp[0] = tmp;
				put_cfg_val (my_ntp);
				sync_SCC (my_ntp);
			}
		}
		else {		// Already set so set mine to it
			tmp = cl_ntp[0];
			strsep (&tmp, ".");		// Get rid of "cluster."
			t2 = strdup (tmp);
			free (cl_ntp[0]);
			cl_ntp[0] = t2;

			put_cfg_val (cl_ntp);
			tk_update_ntp_conf();
		}
		free_cfg_val (cl_ntp);
		free_cfg_val (my_ntp);
/*
	Finally, I just became the cluster's PCC. Did the previous PCC fail
	while I was SCC? Only the flag tells us for sure.
 */
		if (my_pcc_failed) {
			node_failed (my_pcc_name);
		}
		my_pcc_failed = 0;
		if (my_pcc_name) {
			free (my_pcc_name);
		}
		my_pcc_name = strdup (myname);
	}
/*
	If I've just gotten a new PCC, I need to use it as the time reference
	source.
 */
	if ((new->eventType == NFX_EVENT_PCC) && 
	    (new->param.nodeEvent.state == EVENT_STATE_UP) &&
	    !node_is_PCC()) {

		tk_set_ntp_broadcast();		// reset server list

		if (new->param.nodeEvent.mgmt_ip[0]) {
			struct in_addr inp;
			inp.s_addr = new->param.nodeEvent.mgmt_ip[0];
			tk_mod_ntp_servers (inet_ntoa(inp), 0);
		}
		if (new->param.nodeEvent.mgmt_ip[1]) {
			struct in_addr inp;
			inp.s_addr = new->param.nodeEvent.mgmt_ip[1];
			tk_mod_ntp_servers (inet_ntoa(inp), 0);
		}

		tk_update_ntp_conf();
	}
/*
	If I'm the PCC and I've gotten a new SCC, I need to update it with
	the cluster's configuration.
 */
	if (node_is_PCC() && (new->eventType == NFX_EVENT_SCC) && 
	    (new->param.nodeEvent.state == EVENT_STATE_UP)) {

		sync_up_SCC (new->nodeName);
	}
}

static void
restore_IP_cfg(char *on_node)
{
	char **cfg_objs;
	char **obj;
	char *key;
	char *ip_addr_str;
	char *mask_str;
	char *node;
	char *port_spec;
	uint32 ip_addr;
	int32 mask;

	cfg_objs = get_cfg_val ("cluster.ip.*");

	if (!cfg_objs) {
		return;			// None there
	}
/*
	Restore only those IP addresses still assigned to this node
 */
	for (obj = cfg_objs; obj[0]; obj +=2) {

		key = obj[0];
		ip_addr_str = &key[strlen("cluster.ip.")];

		ip_addr = inet_addr (ip_addr_str);

		port_spec = obj[1];

		node = strsep (&port_spec, ":");

		if (!strcmp (node, on_node)) {
			mask_str = port_spec;

			port_spec = strsep (&mask_str, "/");

			mask = atoi(mask_str);

			E_LOG (class_1, debug_s, myappid, 0, 0, 0,
			       ("ccm:restoreIP: port %s, addr %x, mask %x, node %s", 
			         port_spec, ip_addr, mask, node));

			if (ipmd_AddIpCluster (port_spec, ip_addr, mask, NULL, 0, node)
			    != NFX_OK) {
				E_LOG (class_1, alert_s, myappid, 0, 0, 0,
				       ("rst_IP_cfg: ipmd_AddIpCluster failed on %s", 
				        port_spec));
			}
/*
	Restore all routes reachable on this network
 */
			restore_rte (ip_addr, mask, node);
		}
	}

	free_cfg_val (cfg_objs);
}

/*-----------------------------------------------------------------
 * Name:export_LV
 *
 * Description:Send info to mountd on node to export said lv
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 6 Mar 2002
 *
 *-----------------------------------------------------------------
 */
#define LV_RESTORING_MAGIC 0xbabedef0
#define MAX_MOUNT_WAIT	1800 // 1/2 hour should be long enough to clean house
#define LV_FAIL_CHECK   10   // Between getting LV fail event and restarting it

typedef struct lv_mounting {
	int					magic;
	char 				*nfs_stuff;
	char				*node;
	evm_volId_t			lv_id;
	time_t 				when;
	struct lv_mounting	*next;
	struct lv_mounting	*prev;
} lv_mounting_t;

static lv_mounting_t *lvs_being_mounted = NULL;

static __inline__ void
free_lv_blk (lv_mounting_t *blk)
{
	free (blk->nfs_stuff);
	free (blk->node);
	blk->magic = 0;
	free (blk);
}

/*
 * Just discard the memory so we don't eat up all heap if things are not
 * working properly.
 */
static void
timeout_LV_restore()
{
	time_t now = time (NULL);
	lv_mounting_t *look, *next;

	for (look = lvs_being_mounted; look; look = next) {
		next = look->next;
		if ((now - look->when) > MAX_MOUNT_WAIT) {
			E_LOG (class_1, err_s, myappid, 0, 0, 0,
			       ("LV restore timed out on node %s, lv %s", look->node,
			        look->nfs_stuff));
			if (look->prev) {
				look->prev->next = look->next;
			}
			else {
				lvs_being_mounted = look->next;
			}
			if (look->next) {
				look->next->prev = look->prev;
			}

			free_lv_blk (look);
		}
	}
}

static int
restore_LV (int appsock, evm_volId_t lv, char *node, char *nfsOpts, int force)
{
	int sz;
	int rc;
	struct timeval waitTime = {60, 0};
	char *volname;
	lv_mounting_t *new_mount;
	int onNIM;

	uint64 msgBuf[1024];		// The space
	ncmMsg_t *ncmMsg = (ncmMsg_t *)msgBuf;
	evm_MsgVolGetAttrib_t *getName = (evm_MsgVolGetAttrib_t *)ncmMsg->appMsg;
	evm_MsgVolCreate_t *evmCfg = (evm_MsgVolCreate_t *)ncmMsg->appMsg;
	ea_movInFsysReq_t *ol_mntMsg = (ea_movInFsysReq_t *)ncmMsg->appMsg;
/*
	It was online there before, let's make sure the IP addresses are 
	configured on "node".
	Forceful restoration will restore the IP config after initiating
	the takeover.
 */
	if (!force) {
		restore_IP_cfg(node);
	}
/*
	We only keep track of the Volume Id. EA wants the volume name. We
	have to ask EVM to translate for us.
 */
	memset (ncmMsg, 0, sizeof(ncmMsg_t));
	ncmMsg->msgType = NCM_SEND;
	strcpy (ncmMsg->appName, "evm");
	strcpy (ncmMsg->nodeName, node);
	ncmMsg->rqResp = 1;

	memset (getName, 0, sizeof(evm_MsgVolGetAttrib_t));
	getName->hdr.msgType = EVM_VOL_GET_ATTRIB;
	getName->hdr.clientPrivate = mypid;
	getName->id = lv;

	sz = (int)((char *)getName - (char *)ncmMsg) + 
	           sizeof(evm_MsgVolGetAttrib_t);

	rc = sendAgileMsg (-1, msgBuf, sz,
	                   "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, &waitTime,
	                   msgBuf, sizeof(msgBuf));

	if ((rc <= sizeof (ncmMsg_t)) || (evmCfg->hdr.msgType != EVM_VOL_CONFIG)) {
		E_LOG (class_1, alert_s, myappid, 0, 0, 0,
		       ("EVM rejected config request for %llx\n", lv.whole_thing));
		return 1;
	}
	volname = strdup (evmCfg->config.name);	// Dup these before we clear msgBuf
	onNIM = evmCfg->config.onNIM;
/*
	Build LV restore state block and push it on the list
 */
	new_mount = malloc (sizeof(lv_mounting_t));
	new_mount->magic = LV_RESTORING_MAGIC;
	new_mount->lv_id = lv;
	new_mount->nfs_stuff = strdup (nfsOpts);
	new_mount->node = strdup (node);
	new_mount->prev = NULL;
	new_mount->next = lvs_being_mounted;
	if (lvs_being_mounted) {
		lvs_being_mounted->prev = new_mount;
	}
	lvs_being_mounted = new_mount;
	new_mount->when = time (NULL);
/*
	Put LV online and mount it using EA as proxy.
 */
	memset (ncmMsg, 0, sizeof(ncmMsg_t));
	ncmMsg->msgType = NCM_SEND;
	strcpy (ncmMsg->appName, "ea");
	strcpy (ncmMsg->nodeName, node);
	ncmMsg->rqResp = 1;

	memset (ol_mntMsg, 0, sizeof(ea_movInFsysReq_t));
	if (force) {
		ol_mntMsg->hdr.msgType = EA_FORCE_MOVIN_FSYS_REQ;
		ol_mntMsg->forceMoveIn = 1;
		ol_mntMsg->forceTime = 100;
	}
	else {
		ol_mntMsg->hdr.msgType = EA_MOVIN_FSYS_REQ;
	}
	ol_mntMsg->hdr.userInfo = (uint32)new_mount;
	strcpy (ol_mntMsg->fsysId.fsysName, volname);
	free (volname);
	ol_mntMsg->fsysId.volId = lv;
	ol_mntMsg->nim = onNIM;

	sz = (int)((char *)ol_mntMsg - (char *)ncmMsg) + 
	           sizeof(ea_movInFsysReq_t);
/*
	Send it off and since it could take quite a bit of time, we'll pick
	up the response later off our registered socket.
 */
	sendAgileMsg (appsock, msgBuf, sz,
	              "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, NULL, NULL, 0);

	E_LOG (class_1, debug_s, myappid, 0, 0, 0,
	       ("Requested EA to %s %s|%llx", 
	     ol_mntMsg->hdr.msgType == EA_MOVIN_FSYS_REQ ? "MOVIN" : "FORCE MOVIN",
	         ol_mntMsg->fsysId.fsysName, ol_mntMsg->fsysId.volId.whole_thing));
	return 0;
}

/*
	Forcably restore an LV
 */
extern void
force_takeover (char *newnode, char *lvstr, char *nfs_opts)
{
	evm_volId_t lv;

	lv.whole_thing = strtoq (lvstr, NULL, 16);

	E_LOG (class_6, debug_s, myappid, 0, 0, 0,
	       ("%s force takeover of %s nfs %s", newnode, lvstr, nfs_opts));

	restore_LV (appSock, lv, newnode, nfs_opts, 1);
}

typedef enum {
	export_ok,
	export_fail,
	export_redundent
} exp_stat_t;

static exp_stat_t
export_this (evm_volId_t lvid, char *nfs, char *onNode)
{
	uint64 msgBuf[1024];		// The space
	ncmMsg_t *ncmMsg = (ncmMsg_t *)msgBuf;
	mnt_infoReq_t *expReq = (mnt_infoReq_t *)ncmMsg->appMsg;
	mnt_infoResp_t *expRsp = (mnt_infoResp_t *)ncmMsg->appMsg;
	struct timeval waitTime = { 10, 0 };

	char *msg;
	char *volname;
	char *optStr;
	char *ops;
	char *nxtOpt;
	exp_stat_t rc = export_ok;
	int sz;
	int sAMrc;

	nxtOpt = nfs;
	E_LOG(class_1, debug_s, myappid, 0, 0, 0,
	      ("about to send exports :%s:", nxtOpt));

	while ((optStr = strsep(&nxtOpt, ";"))) {

		memset (ncmMsg, 0, sizeof(ncmMsg_t));
		ncmMsg->msgType = NCM_SEND;
		strcpy (ncmMsg->appName, "mountd");
		strcpy (ncmMsg->nodeName, onNode);
		ncmMsg->rqResp = 1;

		memset (expReq, 0, sizeof (mnt_infoReq_t));
		expReq->msgHdr.msgType = MNT_EXPORT_REQ;
		expReq->msgHdr.private = mypid;
		expReq->volId = lvid;

		volname = &(optStr[9]); // Name starts after export address:
		ops = index (optStr, '|');

		if (ops) {
			*ops = '\0';
			ops++;
			expReq->optionsLen = strlen(ops);
		}

		expReq->nameLen = strlen (volname);	// Now it's safe to take the len
		memcpy (expReq->pathName, volname, expReq->nameLen);

		if (ops) {
			memcpy (&expReq->pathName[expReq->nameLen], ops, 
			        expReq->optionsLen);
		}

		sz = (int)((char *)expReq - (char *)ncmMsg) + 
		     sizeof (mnt_infoReq_t) + expReq->nameLen + expReq->optionsLen;

		sAMrc = sendAgileMsg (-1, msgBuf, sz,
		                   "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, 
						   &waitTime,
		                   msgBuf, sizeof(msgBuf));

		if ((sAMrc <= sizeof (ncmMsg_t)) || 
		    (expRsp->msgHdr.status != NFX_OK)) {

			switch (expRsp->msgHdr.status) {
				case MOUNTD_ERR_DUPLICATE_EXPORT:
					msg = "MOUNTD_ERR_DUPLICATE_EXPORT";
					goto com;
				case MOUNTD_ERR_MORE_GENERIC_EXPORT_FOUND:
					msg = "MOUNTD_ERR_MORE_GENERIC_EXPORT_FOUND";
					goto com;
				case MOUNTD_ERR_MORE_SPECIFIC_EXPORT_FOUND:
					msg = "MOUNTD_ERR_MORE_SPECIFIC_EXPORT_FOUND";
					goto com;
				case MOUNTD_ERR_CONFLICT_EXPORT:
					msg = "MOUNTD_ERR_CONFLICT_EXPORT";
					goto com;
				case MOUNTD_ERR_MULTIPLE_INTERFACE:
					msg = "MOUNTD_ERR_MULTIPLE_INTERFACE";

com:				rc = export_redundent;
					E_LOG (class_1, debug_s, myappid, 0, 0, 0,
					       ("mountd, sort of, rejected :%s:%s: with %s",
					       volname, ops ? ops : "NULL OPTIONS", msg));
					break;

				default:
					E_LOG (class_1, alert_s, myappid, 0, 0, 0,
							("mountd rejected export request :%s:%s:", 
							volname, ops ? ops : "NULL OPTIONS"));
					rc = export_fail;
			}
		}
		else {
			E_LOG (class_1, debug_s, myappid, 0, 0, 0,
			       ("successful export %s as %s", volname, 
			        ops ? ops : "NULL OPTIONS"));
		}
	}
	return rc;
}

/*
	Got the response from the ea_movInFsysReq_t sent in lv_restore()
 */
static int
ready_to_export (ncmMsg_t *respMsg)
{
	ea_movInFsysRsp_t *ol_mntRsp = (ea_movInFsysRsp_t *)respMsg->appMsg;

	lv_mounting_t *lv_blk;

	int sz;
	char **cfgObj;
/*
	If this is a mount response, make sure the cookie returned is a pointer
	to our control block. If so remove it from the list and proceed.
 */
	if ((ol_mntRsp->hdr.msgType != EA_MOVIN_FSYS_RSP) &&
	    (ol_mntRsp->hdr.msgType != EA_FORCE_MOVIN_FSYS_RSP)) {
		return 1;
	}

	lv_blk = (lv_mounting_t *)(ol_mntRsp->hdr.userInfo);

	if (lv_blk->magic != LV_RESTORING_MAGIC) {
		E_LOG (class_1, alert_s, myappid, 0, 0, 0,
		       ("EA response has invalid private info"));
		return 0;
	}

	if (!lv_blk->prev) {	// It's the head of the list
		lvs_being_mounted = lv_blk->next;
	}
	else {
		lv_blk->prev->next = lv_blk->next;
	}
	if (lv_blk->next) {
		lv_blk->next->prev = lv_blk->prev;
	}

	cfgObj = malloc (3 * sizeof(char *));

	asprintf (&cfgObj[0], "slave.%s.%016llx", lv_blk->node, 
	         lv_blk->lv_id.whole_thing);
	cfgObj[2] = NULL;


    if (ol_mntRsp->result != NFX_OK) {
		E_LOG (class_1, alert_s, myappid, 0, 0, 0,
		       ("EA rejected MOVIN_FSYS request for %llx saying %s",
		        lv_blk->lv_id.whole_thing, ol_mntRsp->output));
		cfgObj[1] = strdup (ol_mntRsp->output);
		lv_failed_movin (lv_blk->lv_id);
	}
/*
	Now we can export "lv" as "nfsOpts". mnt_infoReq_t MNT_EXPORT_REQ
	nfsOpts is "XXXXXXXX:/vol|options[;XXXXXXXX:/vol|options ...]"
 */
	else {
		exp_stat_t happened;

		lv_going_online (lv_blk->lv_id);

		happened = 
		        export_this (lv_blk->lv_id, lv_blk->nfs_stuff, lv_blk->node);

		switch (happened) {
			case export_ok:
				cfgObj[1] = strdup("awaiting NFS export event");
				break;
			case export_fail:
				cfgObj[1] = strdup ("export failure - see log");
				break;
			case export_redundent:
				cfgObj[1] = NULL;
		}
	}

	free_lv_blk (lv_blk);	// Done with the auto restore of this puppy

	if (cfgObj[1]) {
		sz = put_cfg_val (cfgObj);
		if (sz) {
			E_LOG (class_1, alert_s, myappid, 0, 0, 0,
			       ("ready_to_export error putting LV object %s", cfgObj[0]));
		}
	}
	free_cfg_val (cfgObj);

	return 0;
}

/*-----------------------------------------------------------------
 * Name:lvVis
 *
 * Description: Got an LV Visible event message
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 5 Mar 2002
 *
 *-----------------------------------------------------------------
 */
extern void
lvVis (int appsock, char *node, evm_volId_t lv)
{
	char **cfgObj;
	char lvObj[256];
	char *slaveStat;
	int sz;
/*
	Is this LV new to the cluster?
 */
	sprintf (lvObj, "cluster.lv.%016llx", lv.whole_thing);
	cfgObj = get_cfg_val(lvObj);
/*
	Was there some administrative config concerning this LV's
	export state? (If so, then restore it.)
 */
	if (cfgObj) {
		char *onlineNode;

		onlineNode = malloc (strlen("online:") + strlen(node) + 1);
		sprintf (onlineNode, "online:%s", node);
/*
	We only take action to restore the service if we're a duly elected
	PCC or standalone (which by default is acting as a PCC).
 */
		if (node_is_PCC() && we_have_a_pcc &&
		    (!strcasecmp (cfgObj[1], onlineNode) ||
		    lv_can_be_online (lv, node))) {	// This node?
/*
	We should be able to find out if it was previously exported by
	this node and do the magic to restore it thusly.
	Get its export info object.
 */
			free_cfg_val (cfgObj);			// Out with the old
/*
	If we're not the PCC, we don't do anything towards restoring our
	services.
 */
			sprintf (lvObj, "cluster.nfs.%016llx", lv.whole_thing);
			cfgObj = get_cfg_val(lvObj);	// In with the new

			if (cfgObj)	{
				if (!restore_LV (appsock, lv, node, cfgObj[1], 0)) {
					slaveStat = strdup ("remount in progress");
					lv_going_online (lv);
				}
				else {
					slaveStat = strdup ("unable to remount");
				}
			}
			else {
				E_LOG (class_1, warning_s, myappid, 0, 0, 0,
				       ("LV %s assigned to node w/o export info", lvObj));

				remove_failed_lv (lv);	// Don't auto retry this one
				slaveStat = strdup("visible");
			}
		}
		else {
			slaveStat = strdup("visible");
		}
		free(onlineNode);

		free_cfg_val (cfgObj);
	}
/*
	Otherwise, just mark this LV visible by this node
 */
	else {
		slaveStat = strdup("visible");
	}

	cfgObj = malloc (3 * sizeof(char *));

	sz = strlen ("slave.") + strlen (node) + 18; // ".LV<NULL>"
	cfgObj[0] = malloc(sz);
	sprintf (cfgObj[0], "slave.%s.%016llx", node, lv.whole_thing);

	cfgObj[1] = slaveStat;
	cfgObj[2] = NULL;
	sz = put_cfg_val (cfgObj);
	sync_SCC (cfgObj);
	if (sz) {
		E_LOG (class_1, alert_s, myappid, 0, 0, 0,
		       ("lvVis error putting LV object %s", cfgObj[0]));
	}

	free_cfg_val (cfgObj);
}

/*-----------------------------------------------------------------
 * Name: ifCfg
 *
 * Description: A GbE interface has undergone some configuration event
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 6 Mar 2002
 *
 *-----------------------------------------------------------------
 */
static void
ifCfg (nfx_eventMsg_t *msg)
{
	int sz;
	char newKey[256];
	char **cfgObj;
	char **curr;
	struct in_addr addr;
	int slot, port;
	boardType_t board;
/*
	We will ignore the IP address removal events from a node that is in the
	process of shutting down because we don't want to remove its admin state.
	In the event the PCC was performing a take-over, the node's services will
	be changed before shutting it down.
	We also ignore the 10/100 management port address events.
 */
	getPortInfo (msg->param.ipIfEvent.portName, newKey, &board, &slot, &port);

	E_LOG (class_1, debug_s, myappid, 0, 0, 0,
	       ("ip evt: rem %d, node shut %d, node %s port %s addr %08x",
	        msg->param.ipIfEvent.state == EVENT_STATE_REMOVE,
	        is_shutting_down(msg->nodeName), msg->nodeName,
	        msg->param.ipIfEvent.portName, msg->param.ipIfEvent.ipa));

	if ((board == bt_sc) || 
	    ((msg->param.ipIfEvent.state == EVENT_STATE_REMOVE) && 
	    is_shutting_down(msg->nodeName))) {
		return;
	}
/*
	Let's record this as an admin state object
 */
	addr.s_addr = msg->param.ipIfEvent.ipa;
	sprintf (newKey, "cluster.ip.%s", inet_ntoa(addr));
/*
	If it's there, we'll have to rewrite or delete it
 */
	curr = get_cfg_val (newKey);

	if (curr) {
		if (msg->param.ipIfEvent.state == EVENT_STATE_REMOVE) {
			if (strstr (curr[1], msg->param.ipIfEvent.portName)) {	
				del_cfg_obj (curr);
				free (curr[1]);
				curr[1] = NULL;
				sync_SCC (curr);
			}
			else {
				E_LOG (class_1, debug_s, myappid, 0, 0, 0,
				       ("DelReq invalid: %s, is: %s", 
				        msg->param.ipIfEvent.portName, curr[1]));
			}
		}
		free_cfg_val (curr);
	}
/*
	Otherwise, create it anew
 */
	if (msg->param.ipIfEvent.state == EVENT_STATE_UP) {
		cfgObj = malloc (3 * sizeof (char *));
		cfgObj[0] = strdup (newKey);

		sz = strlen (msg->param.ipIfEvent.portName) + 4; // port/xx<NULL>
		cfgObj[1] = malloc (sz);
		sprintf (cfgObj[1], "%s/%d", msg->param.ipIfEvent.portName, 
		         msg->param.ipIfEvent.mask);

		cfgObj[2] = NULL;

		put_cfg_val (cfgObj);
		sync_SCC (cfgObj);

		free_cfg_val (cfgObj);
	}
#if defined (DEBUG)
	else if (!curr) { // Error. Deleting config that doesn't exist
		E_LOG (class_1, debug_s, myappid, 0, 0, 0,
		       ("ifCfg: deleting config %s", msg->param.ipIfEvent.portName));
	}
#endif

	return;
}

/*-----------------------------------------------------------------
 * Name:lvOnline/lvOffline
 *
 * Description: An LV is online/offline. Update cfg
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 6 Mar 2002
 *
 *-----------------------------------------------------------------
 */
static void
lvChgline (char *node, evm_volId_t lv, int del)
{
	char **opt;

	opt = malloc (3 * sizeof (char *));
	opt[0] = malloc (strlen("cluster.lv.") + 17);
	sprintf (opt[0], "cluster.lv.%016llx", lv.whole_thing);

	opt[2] = NULL;

	if (del) {
		asprintf (&opt[1], "offline:%s", node);
	}
	else {
		asprintf (&opt[1], "online:%s", node);
	}

	put_cfg_val (opt);
	sync_SCC (opt);

	free_cfg_val (opt);
}

static int
lv_auto_restoring (evm_volId_t lv)
{
	lv_mounting_t *look;

	for (look = lvs_being_mounted; look; look = look->next) {
		if (lv.whole_thing == look->lv_id.whole_thing) {
			return 1;
		}
	}
	return 0;
}

// LV has been put online
static void
lvOnline (char *node, evm_volId_t lv)
{
	char *nfskey;
	char **nfsopts;

	lvChgline (node, lv, 0);

	if (!lv_auto_restoring (lv)) {
		asprintf (&nfskey, "cluster.nfs.%016llx", lv.whole_thing);

		nfsopts = get_cfg_val (nfskey);
		free (nfskey);

		if (nfsopts) {
			export_this (lv, nfsopts[1], node);
			free_cfg_val (nfsopts);
		}
	}
}

// LV has been taken offline
static void
lvOffline (char *node, evm_volId_t lv)
{
	lvChgline (node, lv, 1);
}

static void
whack_objs (char **them)
{
	char **it;
	for (it = them; *it; it +=2 ) {
		if (it[1]) {
			free (it[1]);
			it[1] = NULL;
		}
		del_cfg_obj (it);
		sync_SCC (it);
	}
}

// LV has been destroyed
static void
lvBuhBye (evm_volId_t lv)
{
	char **cfgObjs;
	char **it;
	char *key;
	char lvstr[17];

	sprintf (lvstr, "%016llx", lv.whole_thing);
/*
	First we remove the cluster's knowledge of this LV
 */
	asprintf (&key, "cluster.lv.%s", lvstr);
	cfgObjs = get_cfg_val (key);
	free (key);
	if (cfgObjs) {
		whack_objs (cfgObjs);
		free_cfg_val (cfgObjs);
	}

	asprintf (&key, "cluster.nfs.%s", lvstr);
	cfgObjs = get_cfg_val (key);
	free (key);
	if (cfgObjs) {
		whack_objs (cfgObjs);
		free_cfg_val (cfgObjs);
	}
/*
	Now, let's clear all the slaves operational state for those who 
	happen to see this LV.
 */
	cfgObjs = get_cfg_val ("slave.*");
	if (cfgObjs) {
		char *scc_syncer[3];
		scc_syncer[1] = scc_syncer[2] = NULL;
		for (it = cfgObjs; *it; it += 2) {
			if (strstr (*it, lvstr)) {
				del_cfg_obj (it);
				scc_syncer[0] = it[0];
				sync_SCC (scc_syncer);
			}
		}
		free_cfg_val (cfgObjs);
	}
/*
	Remove it from our internal list of failed LVs (if indeed it's there)
 */
	remove_failed_lv (lv);
}

/*-----------------------------------------------------------------
 * Name: nfsExport
 *
 * Description: A volume is exported now so make note of it
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 6 Mar 2002
 *
 *-----------------------------------------------------------------
 */
static void
nfsExport (nfx_eventMsg_t *msg)
{
	char **opt;
	char **curr;
	char *expAddr;
	int sz;
	int i;
	int ntok;
	char *look;
	char **tokens;
	char *new_val;
	char *slave_obj;
/*
	We will ignore the NFX unexport events from a node that is in the
	process of shutting down because we don't want to remove its admin state.
	In the event the PCC was performing a take-over, the node's services will
	be changed before shutting it down.
	Likewise, we want to retain the LV's admin state when it fails and we're
	forced to pull its export.
 */
	if ((msg->param.exportEvent.state == EVENT_STATE_REMOVE) && 
	    (is_shutting_down(msg->nodeName) || 
	      lv_has_failed (msg->param.exportEvent.volId))) {
		return;
	}
/*
	We'll need the address as a string for the config db
 */
	asprintf (&expAddr, "%08x:%s", msg->param.exportEvent.expAddr,
	                               msg->param.exportEvent.expPath);

	E_LOG (class_1, debug_s, myappid, 0, 0, 0,
	       ("from nfs path :%s: options :%s:", 
	         msg->param.exportEvent.expPath, msg->param.exportEvent.expOptions));
/*
	If this LV is already exported, just add this instance
 */
	opt = malloc (3 * sizeof (char *));
	asprintf (&(opt[0]), "cluster.nfs.%016llx", 
	                  msg->param.exportEvent.volId.whole_thing);
/*
	What's the new share supposed to say?
 */
	if (msg->param.exportEvent.state == EVENT_STATE_UP) {
		asprintf (&new_val, "%s|%s", expAddr, 
		                    msg->param.exportEvent.expOptions);
	}
	else {
		new_val = NULL;
	}
/*
	We'll be updating the serving node's item too so build its object name
 */
	asprintf (&slave_obj, "slave.%s.%016llx", msg->nodeName,
	          msg->param.exportEvent.volId.whole_thing);
/*
	See if it was there already
 */
	curr = get_cfg_val (opt[0]);

	if (curr) {		// See if this one is in there. We'll replace it 
/*
	First, "tokenize" to gather ALL the exports of this LV. 
	Always has 1 token. Error to be empty.
 */
		for (ntok = 1, look = curr[1]; *look; look++) {
			if (*look == ';') {
				ntok++;
			}
		}
		tokens = malloc ((ntok + 1) * sizeof(char *));
		for (look = curr[1], i = 0; i < ntok; i++) {
			tokens[i] = strtok (look, ";");
			look = NULL;
		}
		tokens[ntok] = NULL;	// End of the token list
/*
	Is this addr:path there?
 */
		for (i = 0; i < ntok; i++) {
			if (strstr(tokens[i], expAddr)) {
				break;
			}
		}
		tokens[i] = new_val;	// Either append or overwrite
/*
	Build the new config option value string with what was there and
	the new export. Notice how the delete has new_val == NULL and thus
	deletes its token. Cool, huh?
 */
		for (sz = 0, i = 0; i <= ntok; i++) {
			if (tokens[i]) {
				sz += (strlen(tokens[i]) + 1);	// plus one for the semicolon
			}
		}

		if (sz) {
			opt[1] = malloc(sz);
			memset (opt[1], 0, sz);

			for (i = 0; i <= ntok; i++) {
				if (tokens[i]) {
					strcat (opt[1], tokens[i]);
					strcat (opt[1], ";");
				}
			}
			*(opt[1] + sz - 1) = '\0';
		}
		else {
			opt[1] = NULL;		// Last export removed
		}

		if (new_val) {
			free (new_val);
		}
		free (tokens);

		free_cfg_val (curr);
	}
/*
	Adding it new
 */
	else if (msg->param.exportEvent.state == EVENT_STATE_UP) {
		opt[1] = new_val;	// free_cfg_val() will return this to heap now
	}
	else {
		opt[1] = NULL;
		E_LOG (class_1, warning_s, myappid, 0, 0, 0,
		       ("Deleting a share on non-exported lv %s,%s", opt[0], expAddr));
	}
	free (expAddr);

	opt[2] = NULL;		// End of list, man

	if (!opt[1]) {		// No value means we've deleted the last share
		if (curr) {		// If this is false, its an error but we should cope
			del_cfg_obj (opt);
			sync_SCC (opt);
			free (opt[0]);
			opt[0] = slave_obj;
			opt[1] = strdup ("visible");
			put_cfg_val (opt);
			sync_SCC (opt);
		}
	}
	else {
		put_cfg_val (opt);
		sync_SCC (opt);
		free (opt[0]);		// Let's update the "slave." too
		opt[0] = slave_obj;
		put_cfg_val (opt);
		sync_SCC (opt);
	}

	free_cfg_val (opt);
}

static int
relay_this (nfx_eventMsg_t *evt, int sz)
{
	uint64 buf[256];
	ncmMsg_t *rly = (ncmMsg_t *)buf;
/*
	When we're not the PCC, we need to relay these events to the "real"
	PCC.
 */
	if (!node_is_PCC() && we_have_a_pcc) {
		rly->msgType = NCM_SEND;
		strcpy (rly->appName, "ccm");
		strcpy (rly->nodeName, "PCC");
		rly->rqResp = 0;
		memcpy (rly->appMsg, evt, sz);

		sendAgileMsg (-1, rly, sz + sizeof (ncmMsg_t) - 1,
		              "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
		              NULL, NULL, 0);
		return 1;
	}
	return 0;
}

/*
	Delete ALL IP addresses from the named node
 */
static void
clear_IP_addrs (char *onNode)
{
	char **ipcfg;
	uint32 ipaddr;
	char *nodespec;
	char **look;
	char *portspec;
	char *tmp;

	ipcfg = get_cfg_val ("cluster.ip.*");
	if (ipcfg) {
		asprintf (&nodespec, "%s:", onNode);

		for (look = ipcfg; *look; look +=2) {

			tmp = look[1];
			portspec = strsep (&tmp, "/");

			if (!strncasecmp (nodespec, portspec, strlen(nodespec))) {

				ipaddr = inet_addr (look[0] + strlen("cluster.ip."));
				
				tmp = strdup (strchr(portspec, ':') + 1);

				if (ipmd_DelIpCluster (tmp, ipaddr, NULL, 0, onNode)) {
					E_LOG (class_1, err_s, myappid, 0, 0, 0,
					       ("clear_IP_addrs: error deleting %x from %s", 
					        ipaddr, portspec));
				}
				free (tmp);
			}
		}

		free (nodespec);
		free_cfg_val (ipcfg);
	}
/*
	Su said we need to do this too in order to close all TCP connections
 */
	ipmd_ShutDownCluster ("Su said we", onNode);
	ipmd_RestartCluster ("need to do this", onNode);
}

static void
new_lvname (char *name, evm_volId_t id)
{
	char *cfgObj[3];

	asprintf (&cfgObj[0], "lvname.%s", name);
	asprintf (&cfgObj[1], "%016llx", id.whole_thing);
	cfgObj[2] = NULL;

	put_cfg_val (cfgObj);

	free (cfgObj[0]);
	free (cfgObj[1]);
}

static void
delete_lvname (char *name)
{
	char *cfgObj[3];

	asprintf (&cfgObj[0], "lvname.%s", name);
	cfgObj[1] = cfgObj[2] = NULL;

	del_cfg_obj (cfgObj);

	sync_SCC (cfgObj);

	free (cfgObj[0]);
}

static void
check_vol_name_change (char *name, evm_volId_t id)
{
	char **lvnames;
	char **look;
	char idstr[17];

	lvnames = get_cfg_val ("lvname.*");
	if (lvnames) {
		sprintf (idstr, "%016llx", id.whole_thing);
		for (look = lvnames; *look; look += 2) {
			if (!strcmp (idstr, look[1])) {
				del_cfg_obj (&look[0]);
				break;
			}
		}
		free_cfg_val (lvnames);

		new_lvname (name, id);
	}
}

static int
I_am_in_a_cluster (void)
{
	char **clusNodeType;

	clusNodeType = get_cfg_val ("cluster.nodetype");
	if (clusNodeType) {
		free_cfg_val (clusNodeType);
		return 1;
	}
	else {
		return 0;
	}
}

static void
I_have_failed (void)
{
	nfx_eventMsg_t node_failed;
/*
	Something is broken such that I cannot continue servicing my assignments.
	I need to notify the PCC of this so it can reassign thoses services.
	If I'm the PCC, the very fact I'm executing this indicates my SSC is
	still alive so I'll reassign my own services. It must have been another
	failure in the system that got me here.
 */
	node_failed.msgType = NFX_EVENT_POST;
	gethostname (node_failed.nodeName, sizeof (node_failed.nodeName));
	node_failed.needAck = 0;
	node_failed.eventScope = NFX_EVENT_GLOBAL;
	node_failed.eventType = NFX_EVENT_NODE;
	strcpy (node_failed.reason, "Node failed");
	node_failed.param.nodeEvent.state = EVENT_STATE_DOWN;

	if (node_is_PCC()) {
		sendAgileMsg (-1, &node_failed, sizeof (node_failed),
		              "ccm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
		              NULL, NULL, 0);
	}
	else {
		relay_this (&node_failed, sizeof (node_failed));
	}
}

/*-----------------------------------------------------------------
 * Name: procEvt
 *
 * Description: Process the event message we got and figure out how
 *              the cluster's configuration must change because of it.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 14 Feb 2002
 *
 *-----------------------------------------------------------------
 */
static int we_are_alive = 1;
static void
procEvt (int appsock, nfx_eventMsg_t *evtMsg, int msgSize)
{
	if (we_are_alive)
	switch (evtMsg->eventType) {
		case NFX_EVENT_PCC:
			we_have_a_pcc = evtMsg->param.nodeEvent.state == EVENT_STATE_UP;
			if (we_have_a_pcc) {
				syncNode (evtMsg);
			}
			break;
		case NFX_EVENT_SCC:
		{
			int node_up = evtMsg->param.nodeEvent.state == EVENT_STATE_UP;

			if (node_is_PCC()) {	// Only PCCs can have SCCs
				we_have_a_scc = node_up;
			}
			else {
				we_have_a_scc = 0;
			}

			if (node_up) {
				syncNode (evtMsg);
			}
			break;
		}
		case NFX_EVENT_NODE:
			syncNode (evtMsg);
/*
		In case this node still has IP addresses configured on its
		GbE ports but no export services, restore them (but only the PCC
		is allowed to do so).
 */
			if ((evtMsg->param.nodeEvent.state == EVENT_STATE_UP) && 
			    node_is_PCC()) {
				restore_IP_cfg (evtMsg->nodeName);
			}
			break;

		case NFX_EVENT_PORT:
			break;

		case NFX_EVENT_VOLUME:
		{
			relay_this (evtMsg, msgSize);

			switch (evtMsg->param.volEvent.event) {
				case EVENT_VOL_ONLINE:
					lvOnline (evtMsg->nodeName, 
					          evtMsg->param.volEvent.volId);
					lv_going_online (evtMsg->param.volEvent.volId);
					break;
				case EVENT_VOL_OFFLINE:	// Only record admin offlines
					if (!is_shutting_down (evtMsg->nodeName) &&
					    !lv_has_failed (evtMsg->param.volEvent.volId)) {
						lvOffline (evtMsg->nodeName, 
						           evtMsg->param.volEvent.volId);
					}
					break;
				case EVENT_VOL_CREATE:
					lvVis (appsock, evtMsg->nodeName, 
				           evtMsg->param.volEvent.volId);
					new_lvname (evtMsg->param.volEvent.volName,
					            evtMsg->param.volEvent.volId);
                    // Remove the volid reservation from the configuration file.
                    {
                        char * oldReserved[3];
                        
                        asprintf(&oldReserved[0], "evm.reserved.volid.%llx",
                                 evtMsg->param.volEvent.volId.whole_thing);
                        oldReserved[1] = NULL;
                        oldReserved[2] = NULL;
                        del_cfg_obj(oldReserved);
                        free(oldReserved[0]);
                    }
                    break;
				case EVENT_VOL_DELETE:
					lvBuhBye (evtMsg->param.volEvent.volId);
					delete_lvname (evtMsg->param.volEvent.volName);
					break;
                case EVENT_VOL_BROKEN:
                case EVENT_VOL_IOERROR:
                case EVENT_VOL_FSYSABORT:
                case EVENT_VOL_CONFLICT:
                case EVENT_VOL_CORRUPT:
					if (!is_shutting_down (evtMsg->nodeName)) {
						lv_just_failed (evtMsg);
					}
					break;
/*
	This should only happen if we get separated from the cluster
	and the PCC commands another node to takeover our services. Let's
	back off everything in this case.
 */
                case EVENT_VOL_TAKEOVER:
					we_are_alive = 0;
					E_LOG(class_1, emerg_s, myappid, 0, 0, 0,
					      ("node %s, LV %016llx taken over", evtMsg->nodeName,
					       evtMsg->param.volEvent.volId.whole_thing));
					clear_IP_addrs (evtMsg->nodeName);
					break;

				case EVENT_VOL_MODIFY:
					check_vol_name_change (evtMsg->param.volEvent.volName,
					                       evtMsg->param.volEvent.volId);
					break;

				default:
					E_LOG(class_1, debug_s, myappid, 0, 0, 0,
					      ("Unhandled vol event: %d", 
					       evtMsg->param.volEvent.event));
			}
			break;
		}

		case NFX_EVENT_IP_IF:
			relay_this (evtMsg, msgSize);
			ifCfg (evtMsg);
			break;

		case NFX_EVENT_ROUTE:
			relay_this (evtMsg, msgSize);
			rte_chg (evtMsg);
			break;

		case NFX_EVENT_NFS:
			relay_this (evtMsg, msgSize);
			nfsExport (evtMsg);
			break;

		case NFX_EVENT_CPU:
			if (evtMsg->param.cpuEvent.state == EVENT_STATE_DOWN) {
				if (I_am_in_a_cluster()) {
					I_have_failed();
				}
			}
			else {
				switch (evtMsg->param.cpuEvent.cpu) {
					case NFX_NFP_TXRX1:
					case NFX_NFP_TXRX2:
						if (node_is_PCC()) {
							restore_IP_cfg (evtMsg->nodeName);
						}
						else {
							relay_this (evtMsg, msgSize);
						}
						break;
					default:
				}
			}
			break;

		case NFX_EVENT_CARD:
			if (evtMsg->param.cardEvent.state == EVENT_STATE_DOWN) {
				if (I_am_in_a_cluster()) {
					I_have_failed();
				}
			}
			break;

		default:
			E_LOG (class_1, debug_s, myappid, 0, 0, 0,
			       ("Invalid event type %d\n", evtMsg->eventType));
			return;
	}
	return;
}

/*-----------------------------------------------------------------
 * Name:initOperCfg
 *
 * Description: Initializes the node's operational configuration
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 7 Mar 2002
 *
 *-----------------------------------------------------------------
 */
static void
initOperCfg()
{
	char *nt[3];
	char **cl_name;
/*
	Clear the operational node assignment and all the slave tags
 */
	nt[0] = "node.type";
	nt[1] = nt[2] = NULL;
	del_cfg_obj (nt);

	reset_slave_oper_cfg (NULL);
/*
	If we're a cluster member, we'll have been assigned a cluster.nodetype.
	Otherwise, we'll make everyone's life easier by setting the
	cluster.name to our hostname.
 */
	cl_name = get_cfg_val ("cluster.name");
	if (!cl_name) {
		cl_name = malloc (3 * sizeof(char *));
		cl_name[0] = strdup ("cluster.name");
		cl_name[1] = malloc(256);
		gethostname (cl_name[1], 256);
		cl_name[2] = 0;
		put_cfg_val (cl_name);
	}

	free_cfg_val (cl_name);
}

extern void
update_pcc_cfg (char *obj, char *val, char *node)
{
	uint64 msgbuf[1024];
	ncmMsg_t *ncm = (ncmMsg_t *)msgbuf;
	ccmLclCfg_t *cfg = (ccmLclCfg_t *)ncm->appMsg;
	int sz;

	ncm->msgType = NCM_SEND;
	strcpy (ncm->appName, "ccm");
	strcpy (ncm->nodeName, "pcc");

	cfg->hdr.msgType = ccmLclCfg;
	strcpy (cfg->name, node);
/*
 * The message is variable length with the opt/value pair being stuck
 * one after the other.
 */
	sz = (char *)cfg - (char *)ncm + sizeof(ccmLclCfg_t) + 
		strlen (obj) + strlen (val) + 1; // and the terminating '\0 of course

	strcpy (cfg->cfg_pair, obj);
	strcpy (cfg->cfg_pair + strlen (obj) + 1, val);

	sendAgileMsg (-1, ncm, sz, "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
	              NULL, NULL, 0);
}

static int
send_operational_state (char *our_name)
{
	char **ol_vol_cfg;
	char **this;
	int rc = 0;
	char *online_here;
	char *failed_here;
	char *nfs_opts;
	char **nfs_obj;
	char **ip_cfg;
/*
 * We need to synchronize our operational state with the PCC. This state
 * includes online volumes, their share parameters, and our GbE addresses.
 */
	ol_vol_cfg = get_cfg_val ("cluster.lv.*");

	if (ol_vol_cfg) {

		asprintf (&online_here, "online:%s", our_name);
		asprintf (&failed_here, "failed:%s", our_name);
/*
 * We're really only interested in the volumes that are/were online here
 */
		for (this = ol_vol_cfg; *this; this +=2 ) {
			if (!strcasecmp (this[1], online_here) ||
			    !strcasecmp (this[1], failed_here)) {

				rc = 1;	// PCC has been notified
				update_pcc_cfg (this[0], this[1], our_name);

				asprintf (&nfs_opts, "cluster.nfs.%s", 
				                      strrchr(this[0], '.') + 1);
				nfs_obj = get_cfg_val (nfs_opts);
				if (nfs_obj) {
					update_pcc_cfg (nfs_obj[0], nfs_obj[1], our_name);
					free_cfg_val (nfs_obj);
				}
				free (nfs_opts);
			}
		}

		free_cfg_val (ol_vol_cfg);
		free (online_here);
		free (failed_here);
	}
/*
 * Now, lets make sure the PCC knows about our IP configuration
 */
	ip_cfg = get_cfg_val ("cluster.ip.*");

	if (ip_cfg) {
		for (this = ip_cfg; *this; this += 2) {
			char *cmpstr = strdup (this[1]);

			*strchr (cmpstr, ':') = '\0';

			if (!strcasecmp (cmpstr, our_name)) {
				rc = 1; // PCC has been notified of us now
				update_pcc_cfg (this[0], this[1], our_name);
			}

			free (cmpstr);
		}

		free_cfg_val (ip_cfg);
	}

	return rc;
}

static void
send_cfg(void)
{
	char **slave_cfg;
	char **this;
	int pcc_told;
	char *my_slave;
	char my_name[256];
	char **lvnames;
	char **daname;
/*
 * Who am I?
 */
	gethostname (my_name, sizeof(my_name));
/*
 * We'll need to send our operational state to the new PCC.
 */
	pcc_told = send_operational_state(my_name);
/*
 * Let the new PCC know what volumes we can see
 */
	asprintf (&my_slave, "slave.%s.*", my_name);

	slave_cfg = get_cfg_val (my_slave);
	free (my_slave);

	if (slave_cfg) {
		lvnames = get_cfg_val ("lvname.*");

		for (this = slave_cfg; *this; this += 2) {
			char *lvstr = strrchr (this[0], '.') + 1;

			update_pcc_cfg (this[0], this[1], my_name);
/*
 * We've got a slave object for this lv, let's send the name too.
 */
			for (daname = lvnames; daname && *daname; daname += 2) {
				if (!strcmp (lvstr, daname[1])) {
					update_pcc_cfg (daname[0], daname[1], my_name);
					break;
				}
			}
		}
		free_cfg_val (slave_cfg);
		free_cfg_val (lvnames);
	}

	pcc_told = pcc_told || slave_cfg;	//Did we send anything to the pcc yet?

	if (!pcc_told) {
	/*
	 * We have to send the PCC a NULL sync message so it will acknowledge 
	 * we're up and running.
	 */
		update_pcc_cfg ("", "", my_name);
	}
}

/*
	Either add, update or delete this config object
 */
static void
update_cfg (ccmCfgSync_t *new_obj_msg)
{
	char *cfg_obj[3];

	cfg_obj[0] = new_obj_msg->cfg_strings;
	cfg_obj[1] = new_obj_msg->cfg_strings + 
	             strlen (new_obj_msg->cfg_strings) + 1;
	cfg_obj[2] = NULL;
	if (!strlen(cfg_obj[1])) {
		del_cfg_obj (cfg_obj);
	}
	else {
		put_cfg_val (cfg_obj);
	}
}

static void
return_PCC_Name (char *app, int cpu, int slot, int plane)
{
	ccmPCCName_t rsp;

	rsp.hdr.msgType = ccmPCCName;

	if (my_pcc_name) {
		strcpy (rsp.PCCName, my_pcc_name);
	}
	else {
		strcpy (rsp.PCCName, "None elected");
	}

	sendAgileMsg (-1, &rsp, sizeof(rsp), app, cpu, slot, plane,
	              NULL, NULL, 0);
}

/*-----------------------------------------------------------------
 * Name: main
 *
 * Description: My Main Man! What it is!
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 13 Feb 2002
 *
 *-----------------------------------------------------------------
 */
extern int
main ()
{
	sigset_t mask;
	int nrecd;
	long long buf[1024];
	char frmapp[256];
	uint32_t frmCPU;
	uint32_t frmSlot;
	uint32_t onPlane;
	struct timeval restore_TO;
/*
	Mask out everything except the signals we plan to handle
 */
	sigfillset (&mask);
	sigdelset (&mask, SIGTERM);
	sigdelset (&mask, SIGBUS);
	sigdelset (&mask, SIGSEGV);
#if defined (DEBUG)			// To make gdb works
	sigdelset (&mask, SIGTRAP);
	sigdelset (&mask, SIGINT);
#endif
	sigprocmask (SIG_SETMASK, &mask, NULL);
	signal (SIGTERM, termHand);
	signal (SIGSEGV, abnormal);
	signal (SIGBUS, abnormal);

	mypid = getpid();		// Who am i?
	myappid = eee_getApplicationId ("ccm");
/*
	Register us
 */
	appSock = regAgileApp ("ccm");
	if (appSock < 0) {
		E_LOG (class_1, err_s, myappid, 0, 0, 0,
		       ("Error registering ccm app"));
		perror ("Error registering ccm app");
		exit(1);
	}
/*
	Reset configuration's operational state
 */
	initOperCfg();
/*
	Tell MCP we're ready to go
 */
	kill (getppid(), SIGUSR1);
/*
	Let's see if we're a stand-alone node which by default is the PCC for
	the "cluster of one".
 */
	we_have_a_pcc = node_is_PCC();
/*
	Forever, we do our thing.
 */
	while (1) {
/*
	We'll timeout the wait for messages faster if we have LVs on the failed
	list that we might be able to restart.
 */
		restore_TO.tv_sec = have_failed_lvs() ? LV_FAIL_CHECK : MAX_MOUNT_WAIT;
		restore_TO.tv_usec = 0;

		nrecd = recvAgileMsg (appSock, buf, sizeof(buf), &restore_TO,
		                      frmapp, &frmCPU, &frmSlot, &onPlane);

		if (nrecd)	// Only if something was received need we look
		switch (*(int32 *)buf) {
/*
	This could be an event indicating some change in the
	cluster's configuration. procEvt() takes care of those that don't
	require responses (i.e. events).
 */
			case NFX_EVENT_POST:
				procEvt (appSock, (nfx_eventMsg_t *)buf, nrecd);
				break;
/*
		We get the shutdown request message for one of two reasons:
		1) sysdown has told us to shutdown its node (done is false)
		2) We got a takeover req and shutdown the old_node prior to
		   transferring the services to the new node. shut_reply() responds
		   to whomever send the shutdown request (us in that case).
 */
			case ccmShutDown:
				if (!((ccmShutdown_Msg_t *)buf)->done) {
					shutdown_node (appSock, frmapp, frmCPU, frmSlot, 
						                  onPlane, (ccmShutdown_Msg_t *)buf);
				}
				else {
					complete_takeover ((ccmShutdown_Msg_t *)buf);
				}
				break;
/*
	CCMs talk amongst themselves when nodes are admitted to the
	cluster to ensure their configurations are synchronized.
	Additionally, the SCC will request the PCC to periodically send
	the administrative state of cluster services.
 */
			case ccmLclCfgRq:		// From CCM on PCC. They want to sync up
				send_cfg();
				break;

			case ccmLclCfg:			// From CCM on slave. Here it is.
				node_up (appSock, (ccmLclCfg_t *)buf);
				break;

			case ccmTakeOver:		// Node Takeover command
				takeover_node_services (appSock, frmapp, frmCPU, frmSlot,
				                        onPlane, (ccmTakeOverRq_t *)buf);
				break;

			case ccmCfgSync:		// PCC sends these to SCC
				update_cfg ((ccmCfgSync_t *)buf);
				break;

			case ccmRqPCCName:
				return_PCC_Name (frmapp, frmCPU, frmSlot, onPlane);
				break; 

			case NCM_RESP:		// Responses we get via ncm
				if (shut_reply ((ncmMsg_t *)buf) &&
					ready_to_export((ncmMsg_t *)buf)) {

					E_LOG (class_1, err_s, myappid, 0, 0, 0,
				       ("unexpected ncm payload received %d", 
			            *(int *)(((ncmMsg_t *)buf)->appMsg)));
				}
				break;

			case NCM_NODE_EVENT:
				if (((ncmnodeEvent_t *)buf)->new) {
					if (node_is_PCC() && we_have_a_pcc) {
						ask_node_for_update (((ncmnodeEvent_t *)buf)->nodeName);
					}
				}
				break;

			default:	// Bogus message direct
				E_LOG (class_1, alert_s, myappid, 0, 0, 0,
			       ("ccmd: unhandled message type %d", *(uint32 *)buf));

		}	// Message processing stuff ends here
/*
	Check if we need to do some timeout stuff
 */
		check_failed_lvs ();

		timeout_LV_restore ();

	}	// Again, and this time with feeling
}
