/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:	lnm-util.c
 *
 * RCS:  $Id: lnm-util.c,v 1.10 2002/09/27 01:00:55 gregs Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:	LNM cluster utility routines
 *
 * Created by:	Pankaj Gupta 
 *
 * Date Created:	11/05/01
 *
 *-----------------------------------------------------------------
 */

#include <stdio.h>      // fprintf(), stderr
#include <stdlib.h>     // exit()
#include <string.h>     // memset()
#include <unistd.h>     // close()

#include <fcntl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <signal.h>
#include <errno.h>
#include <netdb.h>

#include "nfx-incl.h"

#include "exports/anpssc-api.h"
#include "exports/lnm-api.h"
#include "exports/ncm-api.h"
#include "exports/opt-api.h"
#include "exports/elog-api.h"
#include "exports/event-api.h"
#include "lnm-local.h"




/*------------------------------------------------------------------------------
 * Global variables.
 *------------------------------------------------------------------------------
 */
extern	char				myClusterName[LNM_MAX_CLUSTER_NAME];
extern	cluster_nodeInfo_t 	*myClusterNode;
extern	cluster_nodeInfo_t	*myClusterPCC;
extern	time_t				PCC_failed_time;
extern	cluster_nodeInfo_t	*myClusterSCC;
extern	cluster_nodeInfo_t	clusterNodeInfoArray[MAX_NODE_PER_CLUSTER];

struct timeval waitRegmTo = {10, 0};

/*------------------------------------------------------------------------------
 * Prototypes of functions.
 *------------------------------------------------------------------------------
 */
int lnm_setConfig(char * option, char * value);
int lnm_getConfig(char * option);


/*------------------------------------------------------------------------------
 * Name         : cluster_findNode()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
cluster_nodeInfo_t *cluster_findNode(uint32 nodeChassisId)
{
	int idx;
	cluster_nodeInfo_t	*nodeEntry;

	for (idx = 0; idx < MAX_NODE_PER_CLUSTER; idx++)
	{
		nodeEntry = &clusterNodeInfoArray[idx];

		if (!(nodeEntry->useFlag & NODE_ENTRY_ADDED))
			continue;

		if (nodeEntry->nodeChassisId == nodeChassisId)
			return nodeEntry;
	}

	return NULL;
}

/*------------------------------------------------------------------------------
 * Name         : cluster_allocateNode()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
cluster_nodeInfo_t *cluster_allocateNode()
{
	int idx;
	cluster_nodeInfo_t	*nodeEntry;

	for (idx = 0; idx < MAX_NODE_PER_CLUSTER; idx++)
	{
		nodeEntry = &clusterNodeInfoArray[idx];

		if ( !(nodeEntry->useFlag & NODE_ENTRY_ALLOC))
		{
			memset(nodeEntry, 0, sizeof(cluster_nodeInfo_t));
			nodeEntry->useFlag |= NODE_ENTRY_ALLOC;
			return nodeEntry;
		}
	}

	perror("No free entries in clusterNodeInfoArray");
	return NULL;
}

/*------------------------------------------------------------------------------
 * Name         : cluster_deallocateNode()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
void cluster_deallocateNode(cluster_nodeInfo_t *delNodeEntry)
{
	int idx;
	cluster_nodeInfo_t	*nodeEntry;

	if ( !(delNodeEntry->useFlag & NODE_ENTRY_ALLOC))
	{
		perror("Cannot deallocate unallocated entry");
		return;
	}

	for (idx = 0; idx < MAX_NODE_PER_CLUSTER; idx++)
	{
		nodeEntry = &clusterNodeInfoArray[idx];

		if (!(nodeEntry->useFlag & NODE_ENTRY_ALLOC))
			continue;

		if (nodeEntry == delNodeEntry)
		{
			nodeEntry->useFlag &= ~NODE_ENTRY_ALLOC;
			return;
		}
	}

	perror("Tryin to free unallocated node");
	return;
}

/*------------------------------------------------------------------------------
 * Name         : cluster_addNode()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
cluster_nodeInfo_t *cluster_addNode(cluster_nodeInfo_t *newNodeEntry)
{
	int idx;
	cluster_nodeInfo_t	*nodeEntry;

	if ( !(newNodeEntry->useFlag & NODE_ENTRY_ALLOC))
	{
		E_LOG(class_1, debug_s, myAppId, 0, 0, 0,
		      ("Cannot add unallocated entry"));
		return NULL;
	}

	for (idx = 0; idx < MAX_NODE_PER_CLUSTER; idx++)
	{
		nodeEntry = &clusterNodeInfoArray[idx];

		if (!(nodeEntry->useFlag & NODE_ENTRY_ALLOC))
			continue;

		if (nodeEntry == newNodeEntry)
		{
			nodeEntry->useFlag |= NODE_ENTRY_ADDED;
			E_LOG(class_1, info_s, myAppId, 0, 0, 0,
			      ("Adding node for cluster %s chassis %d IP Addr 0x%x\n",
	               myClusterName, nodeEntry->nodeChassisId,
			       nodeEntry->nodeIpa[0]));

			return nodeEntry;
		}
	}

	E_LOG(class_1, debug_s, myAppId, 0, 0, 0,
	      ("Cannot add unallocated entry in clusterNodeInfoArray"));
	return NULL;
}

/*------------------------------------------------------------------------------
 * Name         : cluster_deleteNode()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
void cluster_deleteNode(uint32 nodeChassisId)
{
	int idx;
	cluster_nodeInfo_t	*nodeEntry;

	for (idx = 0; idx < MAX_NODE_PER_CLUSTER; idx++)
	{
		nodeEntry = &clusterNodeInfoArray[idx];

		if (!(nodeEntry->useFlag & NODE_ENTRY_ADDED))
			continue;

		if (nodeEntry->nodeChassisId == nodeChassisId)
		{
			if (myClusterNode == nodeEntry)
				myClusterNode = NULL;

			if (myClusterPCC == nodeEntry)
				myClusterPCC = NULL;

			if (myClusterSCC == nodeEntry)
				myClusterSCC = NULL;

			nodeEntry->useFlag &= ~NODE_ENTRY_ADDED;
			return;
		}
	}

	E_LOG(class_1, debug_s, myAppId, 0, 0, nodeChassisId,
	      ("cluster_delete_node could not find requested node"));
	return;
}

/*------------------------------------------------------------------------------
 * Name         : cluster_changeNodeState()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
void cluster_changeNodeState(cluster_nodeInfo_t* nodeEntry, int newState)
{
	extern char *cnStateStr[];
	int oldStateIsUp;
	int newStateIsUp;

	if (nodeEntry->nodeState == newState)
	{
		E_LOG(class_1, debug_s, myAppId, 0, 0, 0,
		      ("cluster_changeNodeState node %d already in %s state\n",
		      nodeEntry->nodeChassisId, cnStateStr[newState]));
		return;
	}

	E_LOG(class_1, info_s, myAppId, 0, 0, 0,
	      ("Changing node %d state from %s to %s\n",
	       nodeEntry->nodeChassisId, cnStateStr[nodeEntry->nodeState],
	       cnStateStr[newState]));

	/*
	 * Update PCC/SCC pointers if node is giving up that role
	 */
	if ((nodeEntry->nodeState == CN_STATE_PCC) && 
			(newState != CN_STATE_PCC))
	{
		if (myClusterPCC == nodeEntry)
		{
			cluster_postEvent(nodeEntry, NFX_EVENT_PCC, EVENT_STATE_DOWN,
			                  NULL, 0);
			myClusterPCC = NULL;
			PCC_failed_time = time (NULL);
		}
		
	}

	if ((nodeEntry->nodeState == CN_STATE_SCC) && 
			(newState != CN_STATE_SCC))
	{
		if (myClusterSCC == nodeEntry)
		{
			cluster_postEvent(nodeEntry, NFX_EVENT_SCC, EVENT_STATE_DOWN,
			                  NULL, 0);
			myClusterSCC = NULL;
		}
	}

	oldStateIsUp = cluster_nodeIsUp(nodeEntry);

	nodeEntry->nodeState = newState;

	newStateIsUp = cluster_nodeIsUp(nodeEntry);

	if (oldStateIsUp && !newStateIsUp)
	{
	// CCM & NCM on PCC needs to know node went away
		cluster_postEvent(nodeEntry, NFX_EVENT_NODE, EVENT_STATE_DOWN,
		                  nodeEntry->nodeIpa, sizeof(nodeEntry->nodeIpa));
	}
	else if (!oldStateIsUp && newStateIsUp)
	{
	// CCM & NCM on PCC needs to know about new node
		cluster_postEvent(nodeEntry, NFX_EVENT_NODE, EVENT_STATE_UP,
		                  nodeEntry->nodeIpa, sizeof(nodeEntry->nodeIpa));
	}

	if (newState == CN_STATE_PCC)
	{
	// NCMs on all nodes need to know about new PCC
		cluster_postEvent(nodeEntry, NFX_EVENT_PCC, EVENT_STATE_UP,
		                  nodeEntry->nodeIpa, sizeof(nodeEntry->nodeIpa));
		myClusterPCC = nodeEntry;
	}

	if (newState == CN_STATE_SCC)
	{
	// CCM on the SCC needs to keep in sync with CCM on the PCC
		cluster_postEvent(nodeEntry, NFX_EVENT_SCC, EVENT_STATE_UP,
		                  nodeEntry->nodeIpa, sizeof(nodeEntry->nodeIpa));
		myClusterSCC = nodeEntry;
	}
}

/*------------------------------------------------------------------------------
 * Name         : cluster_updateNode()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
void cluster_updateNode(cluster_nodeInfo_t *nodeEntry, 
								cluster_controlMsg_t *ccMsg)
{
	nodeEntry->deadWhen = now + node_timeout;

	/*
	 * Add other fields TBD
	 */
	if (ccMsg)
	{
		nodeEntry->CCMode = ccMsg->msg.nodeHelloMsg.CCMode;
		nodeEntry->preferredPCCMode = ccMsg->msg.nodeHelloMsg.preferredPCCMode;
	}
}

/*------------------------------------------------------------------------------
 * Name         : cluster_findSccCandidate()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
cluster_nodeInfo_t *cluster_findSccCandidate()
{
	int idx;
	cluster_nodeInfo_t	*nodeEntry;

	if (myClusterNode != myClusterPCC)
	{
		E_LOG(class_1, debug_s, myAppId, 0, 0, __LINE__,
		      ("cluster_findSccCandidate:I'm not PCC"));
		return NULL;
	}

	if (myClusterSCC)
	{
		E_LOG(class_1, debug_s, myAppId, 0, 0, __LINE__,
		      ("cluster_findSccCandidate:myClusterSCC already exists"));
		return myClusterSCC;
	}

	for (idx = 0; idx < MAX_NODE_PER_CLUSTER; idx++)
	{
		nodeEntry = &clusterNodeInfoArray[idx];

		if (!(nodeEntry->useFlag & NODE_ENTRY_ADDED))
			continue;

		if (nodeEntry->nodeState == CN_STATE_CC)
			return nodeEntry;
	}

	return NULL;
}


/*------------------------------------------------------------------------------
 * Name         : cluster_nodeIsUp()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
int cluster_nodeIsUp(cluster_nodeInfo_t *nodeEntry)
{

	if ((nodeEntry->nodeState != CN_STATE_INIT) &&
		(nodeEntry->nodeState != CN_STATE_DOWN))
	{
		return 1;
	}

	return 0;
}

/*------------------------------------------------------------------------------
 * Name         : printIpa()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
char *printIpa(int ipaddr)
{
	struct in_addr in;
	in.s_addr = ipaddr;
	return inet_ntoa(in);
}

/*------------------------------------------------------------------------------
 * Name         : get_ip_addr()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 12/10/01
 *
 *------------------------------------------------------------------------------
 */
void get_ip_addr(struct in_addr* inaddr,char* str)
{
	char *sc_name;
	int if_num;
	struct hostent *me;

	sc_name = alloca (strlen (str) + 5);	// .scN\0 "WTF is 5?", you asked.

	for (if_num = 0; if_num < 2; if_num++) {

		sprintf (sc_name, "%s.sc%d", str, if_num + 1);

		me = gethostbyname (sc_name);

		if (me) {
			inaddr[if_num] = *(struct in_addr *)(me->h_addr_list[0]);
		}
		else {
			memset (&(inaddr[if_num]), 0, sizeof (inaddr[if_num]));
		}
	}
}
