/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: Configure name services
 *
 * RCS:  $Id: cfg_nfxnis.c,v 1.11 2002/08/23 21:48:35 gregs Exp $
 *
 * Copyright (C) 2002, Agile Storage, Inc.
 *
 * Description: These are the methods provided by the library to
 *              manipulate the configuration of name services.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 28 Mar 2002
 *
 *-----------------------------------------------------------------
 */
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>

#include "exports/opt-api.h"
#include "exports/anpssc-api.h"
#include "exports/nfxnis-api.h"
#include "nfxnis-loc.h"

#define REGM_TO_SEC 5

static char *service_names[] = { "dns", "yp" };

extern int
nfxnis_rebind_yp (void)
{
	char xxx[100];
	int cfgfd;
	char *cmd;
	char *key;
	char **cfg;
	char **look;
	char *domainstr;
	char *srvrstr;
	char *netstr;
	char *maskstr;
	int mask;
	int net;
	int nbits;
	struct in_addr srvaddr;
	int sz;
	int rc = 0;

	key = malloc (strlen(NAMESERVICE) + 2);	// *\0
	sprintf (key, "%s*", NAMESERVICE);

	cfg = get_cfg_val (key);
	free (key);

	cfgfd = open (NFXNISCFG, O_CREAT|O_TRUNC|O_WRONLY, 0600);

	if (cfg) {

		sz = 0;		// Use this later to determine if there's work to do

		for (look = cfg; *look; look += 2) {

			if (!strncmp (look[1], "yp", 2)) {

				maskstr = look[0] + strlen (NAMESERVICE);
				netstr = strsep (&maskstr, ".");

				net = strtoul (netstr, NULL, 16);
				mask = strtoul (maskstr, NULL, 16);
				mask = ntohl (mask);

				for (nbits = 32; !(mask & 1) && (nbits > 0); 
				     nbits--, mask >>= 1){};

				netstr = inet_net_ntop (AF_INET, &net, nbits, xxx, sizeof(xxx));

				srvrstr = look[1] + 3;
				domainstr = strsep (&srvrstr, ":");

				srvaddr.s_addr = strtoul (srvrstr, NULL, 16);
				srvrstr = inet_ntoa (srvaddr);

				sz = strlen (netstr) + strlen (domainstr) + 
				     strlen (srvrstr) + 4; // net\tdomain\tservaddr\n\0
				cmd = malloc (sz);
				sprintf (cmd, "%s\t%s\t%s\n", netstr, domainstr, srvrstr);
				write (cfgfd, cmd, strlen(cmd));
				free (cmd);
			}
		}
		free_cfg_val (cfg);

		if (sz) {
			sz = strlen (SETDOMPRG) + strlen (NFXNISCFG) + 4; // prg < file\0
			cmd = malloc (sz);
			sprintf (cmd, "%s < %s", SETDOMPRG, NFXNISCFG);
			rc = system (cmd);
			free (cmd);
		}
	}
	close (cfgfd);

	return rc;
}

/*-----------------------------------------------------------------
 * Name: nfxnis_addservice
 *
 * Description: Add a name server
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 28 Mar 2002
 *
 *-----------------------------------------------------------------
 */
extern int
nfxnis_addservice (nfxnis_serv_t type,
                   in_addr_t network, in_addr_t netmask,
                   in_addr_t serv_addr, char *domain)
{
	char **cfg;
	char *key;
	char **look;
	in_addr_t lnet, lmask;
	char *serv_type;
	char *domain_name;
	char *servers;
	int rc = 0;
	char *new_val = NULL;

	key = malloc (strlen(NAMESERVICE) + 2);	// *\0
	sprintf (key, "%s*", NAMESERVICE);

	cfg = get_cfg_val (key);
	free (key);

	if (cfg) {		// Make sure it doesn't collide with something that's there

		for (look = cfg; look[0]; look += 2) {

			lnet = (in_addr_t)strtoul (look[0] + strlen(NAMESERVICE), 
			                           NULL, 16);
			lmask = (in_addr_t)strtoul (look[0] + strlen(NAMESERVICE) + 9, 
			                           NULL, 16);
			servers = look[1];
			serv_type = strsep (&servers, ":");
			domain_name = strsep (&servers, ":");

			if ((lnet & lmask) == (network & netmask)) {
				if (type != dns_name) {
					errno = EEXIST;
					rc = -1;
				}
				else { // Might be able to add another server
/*
		Existing service type had better be "dns" if you want to add another
		server.
 */
					if (strcmp (serv_type, service_names[type])) {
						errno = EEXIST;
						rc = -1;
					}
/*
		Make sure we haven't reached the max number of DNS servers yet.
 */
					else if (strlen(servers) >= 
					         ((NFX_MAX_DNS_SERVERS_PER_DOMAIN * 9) - 1)) {
							errno = EEXIST;
							rc = -1;
					}
					else {
						new_val = malloc (strlen(domain_name) + 
						                strlen (serv_type) +
						                strlen (servers) + 12); // 3*':'addr\0
						sprintf (new_val, "%s:%s:%s:%08x", serv_type, 
						         domain_name, servers, serv_addr);
					}
				}
				break;
			}
		}
		free_cfg_val (cfg);
	}

	if (!rc) { // Need to update with given network/mask
		cfg = malloc (3 * sizeof(char *));
		cfg[2] = NULL;

		cfg[0] = malloc (strlen(NAMESERVICE) + 18); // nnnnnnnn.mmmmmmmm\0
		sprintf (cfg[0], "%s%08x.%08x", NAMESERVICE, network, netmask);

		if (!new_val) {	// Mean it's a new cfg object all together
			cfg[1] = malloc (strlen(service_names[type]) + strlen (domain) +
			                 11); 	// servname:domain:xxxxxxxx\0
			sprintf (cfg[1], "%s:%s:%08x", service_names[type], 
			                               domain, serv_addr);
		}
		else {
			cfg[1] = new_val;
		}
		rc = put_cfg_val (cfg);
/*
	Do we need to update ypbind?
 */
		if (type == nis_name) {
			rc = nfxnis_rebind_yp();
			if (rc) {
				del_cfg_obj (cfg);
				errno = EINVAL;
			}
		}
/*
	If it's the default, let's update /etc/resolv.conf so the
	SSC stuff can use it.
 */
		if (!rc && !network && !netmask) {
			if (ensure_fs_writable("/etc/resolv.conf") == 0) {
				FILE *resolvconf = fopen ("/etc/resolv.conf", "w+");
				if (resolvconf) {
					switch (type) {
					case nis_name:
						fprintf (resolvconf, "lookup yp file\n");
						setdomainname (domain, strlen (domain));
						break;
					case dns_name:
						fprintf (resolvconf, "search %s\n", domain);
						servers = cfg[1];
						strsep (&servers, ":");	// pointing to domain now
						strsep (&servers, ":");	// the first server's address
						while (servers) {
							struct in_addr srvaddr;
							srvaddr.s_addr = strtol (strsep (&servers, ":"), 
							                         NULL, 16);
							fprintf (resolvconf, "nameserver %s\n", 
									 inet_ntoa (srvaddr));
						}
						fprintf (resolvconf, "lookup bind file\n");
					}
					fclose (resolvconf);
				}
				restore_fs_permissions("/etc/resolv.conf");
			}
		}

		free_cfg_val (cfg);
	}
	return rc;
}

/*-----------------------------------------------------------------
 * Name: nfxnis_delservice
 *
 * Description: Delete a name server
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 28 Mar 2002
 *
 *-----------------------------------------------------------------
 */
extern int
nfxnis_delservice (in_addr_t network, in_addr_t netmask)
{
	char *key;
	char **cfg;
	in_addr_t lookie_net;
	in_addr_t net;
	in_addr_t mask;
	char *netstr;
	char *maskstr;
	char **look;
	int is_default = (!network && !netmask);
	int rc = 0;

	key = malloc (strlen(NAMESERVICE) + 2);	// *\0
	sprintf (key, "%s*", NAMESERVICE);

	cfg = get_cfg_val (key);
	free (key);
	if (cfg) {
		lookie_net = network & netmask;

		for (look = cfg; *look; look += 2) {
			netstr = look[0] + strlen (NAMESERVICE);
			maskstr = netstr + 9;
			net = strtoul (netstr, NULL, 16);
			mask = strtoul (maskstr, NULL, 16);

			if ((net & mask) == lookie_net) {
				del_cfg_obj (look);
				break;
			}
		}

		if (!look[0]) {
			rc = -1;
		}
		else if (!strncmp (look[1], "yp", 2)) {
			nfxnis_rebind_yp();
			if (is_default) {
				setdomainname (DEFDOMAINNAME, strlen (DEFDOMAINNAME));
			}
		}

		free_cfg_val (cfg);
	}
	if (is_default) {
		modify_readonly_file("/etc/resolv.conf", unlink, NULL);
	}

	return rc;
}

/*-----------------------------------------------------------------
 * Name: nfxnis_get_cfg
 *
 * Description: Give me the list configured of name servers
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 28 Mar 2002
 *
 *-----------------------------------------------------------------
 */
extern nfxnis_cfg_t **
nfxnis_get_cfg (in_addr_t network, in_addr_t netmask)
{
	char **cfg;
	char *key;
	nfxnis_cfg_t **list = NULL;
	nfxnis_cfg_t **nxt;
	in_addr_t net;
	in_addr_t mask;
	char *netstr;
	char *maskstr;
	char **look;
	in_addr_t lookie_net;
	char *srv_type;
	char *domain;
	char *srvstr;
	int num_srvrs;
	int sz;

	key = malloc (strlen(NAMESERVICE) + 2);	// *\0
	sprintf (key, "%s*", NAMESERVICE);

	cfg = get_cfg_val (key);
	free (key);
/*
	Look through all "nameservice." objects and return those matching the
	given network.
 */
	if (cfg) {
/*
	Let's count the number of configured domains first
 */
		for (num_srvrs = 0,look = cfg; *look; num_srvrs++, look += 2);

		sz = sizeof(nfxnis_cfg_t *) * (num_srvrs + 1);
		list = malloc (sz);
		memset (list, 0, sz);
		nxt = list;

		lookie_net = network & netmask;
/*
	Find out which ones apply and extract their info for return.
 */
		for (look = cfg; *look; look += 2) {
			netstr = look[0] + strlen (NAMESERVICE);
			maskstr = netstr + 9;
			net = strtoul (netstr, NULL, 16);
			mask = strtoul (maskstr, NULL, 16);

			if ((net & netmask) == lookie_net) {
				*nxt = malloc (sizeof(nfxnis_cfg_t));
				memset (*nxt, 0, sizeof(nfxnis_cfg_t));

				(*nxt)->net = net;
				(*nxt)->mask = mask;

				srvstr = look[1];

				srv_type = strsep (&srvstr, ":");
				domain = strsep (&srvstr, ":");

				(*nxt)->domain = strdup (domain);

				if (!strcasecmp(srv_type, service_names[nis_name])) {
					(*nxt)->type = nis_name;
					(*nxt)->num_srvrs = 1;
					(*nxt)->srvrs[0] = strtoul (srvstr, NULL, 16);
				}
				else if (!strcasecmp(srv_type, service_names[dns_name])) {
					int i;
					char *srvr;
					(*nxt)->type = dns_name;
					for (i = 0; srvstr && (i < NFX_MAX_DNS_SERVERS_PER_DOMAIN);
					     i++) {
						srvr = strsep (&srvstr, ":");
						(*nxt)->srvrs[i] = strtoul (srvr, NULL, 16);
					}
					(*nxt)->num_srvrs = i;
				}
				else {
					continue;	// Something's bogus
				}

				nxt++;
			}
		}

		if (nxt == list) {	// None found
			free (list);
			list = NULL;
		}

		free_cfg_val (cfg);
	}

	return list;
}

extern char *
nfxnis_get_dns_srvr (in_addr_t add)
{
	char *key;	
	char **def;
	char **cfg;
	char **look;
	char *netstr;
	char *maskstr;
	in_addr_t net;
	in_addr_t mask;
	char *rc;

	key = malloc (strlen(NAMESERVICE) + 2);	// *\0
	sprintf (key, "%s*", NAMESERVICE);

	cfg = get_cfg_val (key);
	free (key);

	rc = NULL;

	if (cfg) {
/*
	Find out which ones apply and extract their info for return.
 */
		for (def = NULL, look = cfg; *look; look += 2) {
			netstr = look[0] + strlen (NAMESERVICE);
			maskstr = netstr + 9;
			net = strtoul (netstr, NULL, 16);
			mask = strtoul (maskstr, NULL, 16);

			if (!strncmp (look[1], "dns", 3)) {
				if (!mask) {
					def = look;
				}
				else if ((add & mask) == (net & mask)) {
					def = look;	// Found it
					break;
				}
			}
		}

		if (def) {
			rc = strdup (def[1]);
		}

		free_cfg_val (cfg);
	}

	return rc;
}
