/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:
 *
 * RCS:  $Id: cmd_admin.c,v 1.12 2002/10/18 02:19:34 lanc Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:
 *
 * Created by:	Greg Seibert
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#include <sys/types.h>
#include <pwd.h>
#include <readline/history.h>
#include <grp.h>
#include <unistd.h>

#include "zebra.h"
#include "command.h"
#include "log.h"
#include "vtysh.h"
#include "memory.h"

#include "nfxsh.h"

#include "exports/opt-api.h"

#define ALLOW_ADMIN_CMDS 1

#define ADMIN_STR "Administrate system users\n"
#define PRIVS "(view|add|change|full)"
#define PRIVHELP "specifies their privilege level\n"\
		"View Privilege\n"\
		"Add Privilege\n"\
		"Change Privilege\n"\
		"Full Privilege\n"

/*
	In order for ALL BSD user stuff to be happy, we have to limit the
	user name to 8 characters.
 */
#define MAX_USER_NAME 8

static char *privNames[NUMPRIVS] = {"view", "add", "change", "full", "diag"};

#if ALLOW_ADMIN_CMDS
/*-----------------------------------------------------------------
 * Name:getPriv
 *
 * Description:Given the name, get the privilege
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 26 Feb 2002
 *
 *-----------------------------------------------------------------
 */
static priv_t
getPriv (char *pword)
{
	priv_t t;
	int i;
/*
	Match up the privilege name and number
 */
	t = (priv_t)-1;
	for (i = 0; i < NUMPRIVS; i++) {	
		if (!strcmp(pword, privNames[i])) {
			t = i;
			break;
		}
	}
	if (t < 0) {
		fprintf (stderr, "BAD THING: Invalid privilege name: %s\n", pword);
		exit(1);
	}
	return t;
}
#endif

/*-----------------------------------------------------------------
 * Name:isValidAdmin
 *
 * Description: Determines if the given user name is a SANFiler admin.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 26 Feb 2002
 *
 *-----------------------------------------------------------------
 */
static int
isValidAdmin (const char *adminName)
{
	struct passwd *pwent;
	struct group *grent;
	int i;

	pwent = getpwnam (adminName);

	if (!pwent) {
		return 0;
	}

	grent = getgrgid (pwent->pw_gid);

	if (!grent) {
		return 0;
	}

	for (i = 0; i < NUMPRIVS; i++) {
		if (!strcmp(grent->gr_name, groupNames[i])) {
			break;
		}
	}

	return (i != NUMPRIVS);
}

/*
	Command parsers/action functions
 */
DEFUN	(admin_passwd,
		password_cmd,
		"admin password",
		ADMIN_STR
		"Change your password\n")
{
	become_root();
	if (ensure_fs_writable("/etc/") == 0) {
		unbecome_root();
		execute_command ("/usr/bin/passwd", 0, NULL, NULL);
		become_root();
		restore_fs_permissions("/etc/");
	}
	unbecome_root();


	return CMD_SUCCESS;
}

#if ALLOW_ADMIN_CMDS
DEFUN	(admin_add,
		add_cmd,
		"admin add USERNAME -p "PRIVS" -N [REALNAME ...]",
		ADMIN_STR
		"Add admin access\n"
		PRIVHELP
		"New admin name\n"
		"Give them a Real Name\n"
		"This their real name\n")
{
	char *uName = argv[0];
	priv_t uPriv;
	int i;
	char realName[80];
/*
	Make sure the user name isn't too long for BSD to cope with
 */
	if (strlen (uName) > MAX_USER_NAME) {
		fprintf (stderr, 
		   "User name must be no longer than %d characters\n", MAX_USER_NAME);
		return CMD_ERR_FAILED;
	}
/*
	What's the user's privilege?
 */
	uPriv = getPriv (argv[1]);
/*
	If they give us a name, use it
 */
	realName[0] = '\0';
	if (argc < 3) {
		strcat (realName, uName);
	}
	else {
		for (i = 2; (i < argc) && (strlen(realName) < sizeof(realName)); i++) {
			strcat (realName, argv[i]);
			strcat (realName, " ");		// Break up multiple words
		}
	}

	become_root();

	if (ensure_fs_writable("/etc/") == 0) {
		execute_command ("/usr/sbin/useradd", 9, 
						 "-d", "/usr/local/agile/root",
						 "-s", "/usr/local/agile/bin/nfxsh",
						 "-c", realName, 
						 "-g", groupNames[uPriv], uName, NULL);

		execute_command ("/usr/bin/passwd", 1, uName, NULL);
		restore_fs_permissions("/etc/");
	}

	unbecome_root();

	return CMD_SUCCESS;
}

DEFUN	(admin_modifypriv,
		modifypriv_cmd,
		"admin modify USERNAME -p "PRIVS,
		ADMIN_STR
		"Change admin access\n"
		"for this user name\n"
		PRIVHELP)
{
	priv_t uPriv;
	char *uName = argv[0];

	if (!isValidAdmin (uName)) {
		fprintf (stderr, "Invalid admin user name: %s\n", uName);
		return CMD_ERR_FAILED;
	}
/*
	What's the user's new privilege?
 */
	uPriv = getPriv (argv[1]);

	become_root();

	if (ensure_fs_writable("/etc/") == 0) {
		execute_command ("/usr/sbin/usermod", 3, 
						 "-g", groupNames[uPriv], uName, NULL);
		restore_fs_permissions("/etc/");
	}

	unbecome_root();

	return CMD_SUCCESS;
}

DEFUN	(admin_modifyname,
		modifyname_cmd,
		"admin modify USERNAME -N [REALNAME ...]",
		ADMIN_STR
		"Change admin access\n"
		"for this user name\n"
		"New Real Name\n"
		"This is their real name\n")
{
	char *uName = argv[0];
	char realName[80];
	int i;

	if (!isValidAdmin (uName)) {
		fprintf (stderr, "Invalid admin user name: %s\n", uName);
		return CMD_ERR_FAILED;
	}
/*
	If they give us a name, use it
 */
	if (argc < 2) {
		strncpy (realName, uName, sizeof(realName) - 1);
	}
	else {
		realName[0] = '\0';
		for (i = 1; (i < argc) && (strlen(realName) < sizeof(realName)); i++) {
			strcat (realName, argv[i]);
			strcat (realName, " ");		// Break up multiple words
		}
	}

	become_root();

	if (ensure_fs_writable("/etc/") == 0) {
		execute_command ("/usr/sbin/usermod", 3, 
						 "-c", realName, uName, NULL);
		restore_fs_permissions("/etc/");
	}

	unbecome_root();

	return CMD_SUCCESS;
}
#endif

DEFUN	(admin_modifypass,
		modifypass_cmd,
		"admin modify USERNAME -P",
		ADMIN_STR
		"Change admin access\n"
		"for this user name\n"
		"Prompt for a new password\n")
{
	char *uName = argv[0];

	if (!isValidAdmin (uName)) {
		fprintf (stderr, "Invalid admin user name: %s\n", uName);
		return CMD_ERR_FAILED;
	}
/*
	We need to be root to set another's password and nfxsh needs to be setuid
	root for this to work.
 */
	become_root();

	if (ensure_fs_writable("/etc/") == 0) {
		execute_command ("/usr/bin/passwd", 1, uName, NULL);
		restore_fs_permissions("/etc/");
	}

	unbecome_root();

	return CMD_SUCCESS;
}

#if ALLOW_ADMIN_CMDS
DEFUN	(admin_delete,
		delete_cmd,
		"admin delete USERNAME",
		ADMIN_STR
		"Remove admin access\n"
		"Remove this user\n")
{
	char *uName = argv[0];

	if (!isValidAdmin (uName)) {
		fprintf (stderr, "Invalid admin user name: %s\n", uName);
		return CMD_ERR_FAILED;
	}

	become_root();

	if (ensure_fs_writable("/etc/") == 0) {
		execute_command ("/usr/sbin/userdel", 1, uName, NULL);
		restore_fs_permissions("/etc/");
	}

	unbecome_root();

	return CMD_SUCCESS;
}
#endif

static void
displayAdmin (struct passwd *p, int priv)
{
	printf ("Real Name : %s\n", p->pw_gecos);
	printf ("Login Name: %s\n", p->pw_name);
	printf ("Admin Priv: %s\n", privNames[priv]);
	printf ("-----------\n");
}

DEFUN	(admin_show,
		show_cmd,
		"admin show",
		ADMIN_STR
		"Show admin user's info\n")
{
	struct group *ge;
	struct passwd *pe;
	gid_t valGids[NUMPRIVS-1];
	int i;
/*
	Let's get the list of gids corresponding to the visible admin groups
 */
	for (i = 0; i < (NUMPRIVS - 1); i++) {
		ge = getgrnam(groupNames[i]);	// Better not fail here
		valGids[i] = ge->gr_gid;
	}
/*
	Go through all passwd entries displaying only those in an admin group
 */
	setpwent();

	while ((pe = getpwent())) {

		for (i = 0; i < (NUMPRIVS - 1); i++) {
			if (pe->pw_gid == valGids[i]) {
				break;
			}
		}
		if (i < (NUMPRIVS - 1)) {
			displayAdmin (pe, i);
		}
	}

	endpwent();

	return CMD_SUCCESS;
}

void
install_admin (priv_t myPriv, enum node_type node)
{
/*
	Organizing this switch statment such that the higher priority commands
	are first allows us to fall through and pick up the lower priority
	ones at no extra cost.
 */
	switch (myPriv) {
		case diagPriv:
		case fullPriv:
#if ALLOW_ADMIN_CMDS
			install_element (node, &add_cmd);
			install_element (node, &delete_cmd);
			install_element (node, &modifypriv_cmd);
			install_element (node, &modifyname_cmd);
#endif
			install_element (node, &modifypass_cmd);
			install_element (node, &show_cmd);
		case changePriv:
		case addPriv:
		case viewPriv:
/*
	These commands are available in all modes
 */
			install_element (node, &password_cmd);
	}
}
