/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:
 *
 * RCS:  $Id: cmd_cluster.c,v 1.25 2002/10/02 21:32:05 gregs Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:
 *
 * Created by:	Greg Seibert
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#include <sys/types.h>
#include <sys/wait.h>
#include <pwd.h>
#include <readline/history.h>
#include <grp.h>
#include <unistd.h>

#include "zebra.h"
#include "command.h"
#include "log.h"
#include "vtysh.h"
#include "memory.h"

#include "nfxsh.h"

#include "exports/anpssc-api.h"
#include "exports/opt-api.h"
#include "exports/ncm-api.h"
#include "exports/evm-api.h"	// ccm-api.h needs this
#include "exports/ccm-api.h"
#include "exports/licman-api.h"
#include "exports/ipmd-lib.h"
#include "exports/ea-api.h"
#include "exports/rpc.h"  // Because mount.h needs it that's why
#include "exports/mount.h"

#define IGNORE_LICENSES 1		// Won't require licenses for now

#define REGM_TO_SEC	5			// How long we'll wait for regm to respond

#define CLUS_STR "Manage the Cluster\n"
#define CLUS_MOD_STR "Change Cluster Info\n"
#define CLUS_SHOW_STR "Display Cluster Info\n"

extern void
warn_need_reboot(void)
{
	pid_t mcp_pid;
	int status;

	fprintf (stdout, "To affect this change. All services will be restarted.");
	fflush (stdout);

	become_root();
	kill_daemon ("/var/run/mcp.pid");
	kill_daemon ("/var/run/snmpd.pid");
	kill_daemon ("/var/www/logs/httpd.pid");

	if (!(mcp_pid = fork())) {
		execlp ("/usr/local/agile/bin/mcp", "/usr/local/agile/bin/mcp", NULL);
		fprintf (stderr, "Unable to restart services. Call for help\n");
	}
	else {
		waitpid (mcp_pid, &status, 0);	// Reap the zombie
		sleep (5);
		fprintf (stdout, " ... Services restarted\n");
	}
	unbecome_root();
}

extern int
node_has_services (char *name)
{
	char **cfg;
	char **look;
	char *match;
	int rc = 0;

	cfg = get_cfg_val ("cluster.lv.*");
	if (cfg) {
		asprintf (&match, "online:%s", name);
		for (look = cfg; !rc && *look; look += 2) {
			rc = !strcmp (match, look[1]);
		}
		free (match);
		free_cfg_val (cfg);
	}
	return rc;
}

static char *rem_services[] = { "cluster.lv.*", "cluster.nfs.*", 
"cluster.ip.*", NULL};

extern void
remove_cluster_services (void)
{
	char **srvc;
	char **objs;
	char **rem;

	for (srvc = rem_services; *srvc; srvc++) {
		objs = get_cfg_val (*srvc);
		if (objs) {
			for (rem = objs; *rem; rem += 2) {
				del_cfg_obj (rem);
			}
			free_cfg_val (objs);
		}
	}
}

static int
node_in_cluster (void)
{
	char **node_type = get_cfg_val ("cluster.nodetype");
	free_cfg_val (node_type);

	return node_type != NULL;
}
/*-----------------------------------------------------------------
 * Name: modify cluster name
 *
 * Description: Changes/Specifies or deletes the local node's cluster name
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 28 Feb 2002
 *
 *-----------------------------------------------------------------
 */
static int
modClusName (char *name)
{
	char *clus_stuff[3];
	char **clus_role;
	char response[10];
	char us[256];

	gethostname (us, sizeof (us));

	if (node_has_services (us)) {
		fprintf (stdout, "This node still has cluster services assigned\n");
		fprintf (stdout, "You will need to reassign them before the cluster\n");
		fprintf (stdout, "name can be changed\n");
		return CMD_ERR_FAILED;
	}

	fprintf (stdout, "This will remove this node's service configuration\n");
	fprintf (stdout, "Are you absolutely certain you want to do this? y or n ");
	fgets (response, sizeof (response), stdin);

	if (toupper (response[0]) != 'Y') {
		fprintf (stdout, "OK. No changes made\n");
		return CMD_SUCCESS;
	}

	remove_cluster_services();
	fprintf (stdout, "Service configuration irrevocably removed\n");

	clus_stuff[0] = "cluster.name";
	clus_stuff[1] = name;
	clus_stuff[2] = NULL;

	if (name) {
		put_cfg_val (clus_stuff);
/*
	If we've given this node a cluster name AND it hadn't a defined
	role, default it to controller candidate.
*/
		clus_role = get_cfg_val ("cluster.nodetype");
		if (!clus_role) {
			clus_stuff[0] = "cluster.nodetype";
			clus_stuff[1] = "cc";
			clus_stuff[2] = NULL;
			put_cfg_val (clus_stuff);
		}
		else {
			free_cfg_val (clus_role);
		}

	}
	else {	// We're deleting it
		del_cfg_obj (clus_stuff);
	}
	warn_need_reboot();	// Make it so

	return CMD_SUCCESS;
}

DEFUN	(clusDelName,
		clusDelName_cmd,
		"cluster modify clustername -d",
		CLUS_STR
		CLUS_MOD_STR
		"Modify Cluster Name\n"
		"Remove from cluster\n")
{
	return modClusName (NULL);
}

DEFUN	(clusModName,
		clusModName_cmd,
		"cluster modify clustername -c CLUSTERNAME",
		CLUS_STR
		CLUS_MOD_STR
		"Modify Cluster Name\n"
		"Cluster Name change\n"
		"New Name\n")
{
	if (!IGNORE_LICENSES && !licman_is_valid(LIC_CLUSTERING, NFX_SW_VER)) {
		fprintf (stderr, "Clustering license not installed\n");
		return CMD_ERR_FAILED;
	}
	else {
		return modClusName (argv[0]);
	}
}

#define DISABLE_SLAVE_ROLE	// It was decided for first release to not allow

#if defined (DISABLE_SLAVE_ROLE)
#define CLUST_MOD_CMD "cluster modify nodetype -t (pcc|cc|disabled)"
#define CLUST_SLAVE_HELP
#else
define CLUST_MOD_CMD "cluster modify nodetype -t (pcc|cc|slave|disabled)"
#define CLUST_SLAVE_HELP "Cluster slave\n"
#endif

DEFUN	(clusType,
		clusType_cmd,
		CLUST_MOD_CMD,
		CLUS_STR
		CLUS_MOD_STR
		"Modify node type\n"
		"This type\n"
		"Preferred cluster controller\n"
		"Cluster controller candidate\n"
		CLUST_SLAVE_HELP
		"Disabled\n")
{
	char *clus_role[3];
	char us[256];

	gethostname (us, sizeof (us));

	if (node_has_services (us)) {
		fprintf (stdout, "This node still has cluster services assigned\n");
		fprintf (stdout, "You will need to reassign them before this node's\n");
		fprintf (stdout, "role in the cluster can be changed\n");
		return CMD_ERR_FAILED;
	}

	clus_role[0] = "cluster.nodetype";
	clus_role[2] = NULL;

	switch (*argv[0]) {
		case 'p':
			clus_role[1] = "pcc";
			break;
		case 'c':
			clus_role[1] = "cc";
			break;
#if !defined (DISABLE_SLAVE_ROLE)
		case 's':
			clus_role[1] = "slave";
			break;
#endif
		case 'd':
			clus_role[1] = NULL;
			del_cfg_obj (clus_role);
			warn_need_reboot();
			return CMD_SUCCESS;
		default:
			return CMD_ERR_FAILED;
	}

	if (!IGNORE_LICENSES && 
	    !licman_is_valid(LIC_CLUSTERING, NFX_SW_VER)) {
		fprintf (stderr, "Clustering license not installed\n");
		return CMD_ERR_FAILED;
	}

	put_cfg_val (clus_role);

	warn_need_reboot();
	return CMD_SUCCESS;
}

/*-----------------------------------------------------------------
 * Name: clusShowName
 *
 * Description: Shows the cluster name assigned to the local node
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 28 Feb 2002
 *
 *-----------------------------------------------------------------
 */
DEFUN	(clusShowName,
		clusShowName_cmd,
		"cluster show (clustername|nodetype)",
		CLUS_STR
		CLUS_SHOW_STR
		"Show cluster name\n"
		"Show node type\n")
{
	regmGetReq_t getRq;
	long long buf[512];
	regmGetRsp_t *getRsp = (regmGetRsp_t *)buf;
	int rc;
	struct timeval regmTo = {REGM_TO_SEC, 0};
	char *shThang;
	char *shThang1;
	char *attribute;

	if (!node_in_cluster()) {
		fprintf (stderr, "Clustering not enabled on this node\n");
		return CMD_SUCCESS;
	}

	if (*argv[0] == 'c') {
		shThang = "cluster.name";
		shThang1 = NULL;
		attribute = "Cluster name";
	}
	else {
		shThang = "cluster.nodetype";
		shThang1 = "node.type";
		attribute = "Node type";
	}

	memset (&getRq, 0, sizeof(getRq));
	memset (buf, 0, sizeof(buf));

	getRq.hdr.msgType = OPT_GET_OPTION_REQ;
	getRq.hdr.userInfo = getpid();
	strcpy (getRq.getReq.option, shThang);

	rc = sendAgileMsg (-1, &getRq, sizeof(getRq),
		"regm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, &regmTo,
		buf, sizeof(buf));
	
	if (rc <= 0) {
		fprintf (stderr, "Unable to get %s\n", shThang);
		if (rc < 0) {
			perror ("Error is");
		}
		return CMD_ERR_FAILED;
	}
	else if (getRsp->getRsp.nbEntry < 1) {
		fprintf (stderr, "Clustering not enabled on this node\n");
		return CMD_SUCCESS;
	}
	else {
		fprintf (stdout, "Configured %s: %s\n", attribute, 
		         getRsp->getRsp.entry[0].value);
	}

	if (shThang1) {
		memset (&getRq, 0, sizeof(getRq));
		memset (buf, 0, sizeof(buf));

		getRq.hdr.msgType = OPT_GET_OPTION_REQ;
		getRq.hdr.userInfo = getpid();
		strcpy (getRq.getReq.option, shThang1);

		rc = sendAgileMsg (-1, &getRq, sizeof(getRq),
			"regm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, &regmTo,
			buf, sizeof(buf));
	
		if (rc <= 0) {
			fprintf (stderr, "Unable to get %s\n", shThang1);
			if (rc < 0) {
				perror ("Error is");
			}
			return CMD_ERR_FAILED;
		}
		else if (getRsp->getRsp.nbEntry < 1) {
			fprintf (stderr, "Operational node type not yet assigned.\n");
			fprintf (stderr, "Election in progress. Check back in 1 minute.\n");
			return CMD_SUCCESS;
		}

		fprintf (stdout, "Current %s: %s\n", attribute, 
		         getRsp->getRsp.entry[0].value);

	}
	return CMD_SUCCESS;
}

static void
prnt_lv_name (evm_volId_t lv)
{
	evm_MsgVolGetAttrib_t get_name;
	evm_MsgVolCreate_t cfg;
	struct timeval five_sec = {5, 0};
	char *vol_name;

	get_name.hdr.msgType = EVM_VOL_GET_ATTRIB;
	get_name.id = lv;

	if (sendAgileMsg (-1, &get_name, sizeof (get_name),
	                   "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
	                   &five_sec,
	                   &cfg, sizeof(cfg)) != sizeof(cfg)) {
		vol_name = "";
	}
	else if (cfg.hdr.msgType != EVM_VOL_CONFIG) {
		vol_name = "";
	}
	else {
		vol_name = cfg.config.name;
	}
	fprintf (stdout, "%s\n", vol_name);
}

DEFUN	(clusShowVols,
		clusShowVols_cmd,
		"cluster show volumes [-v]",
		CLUS_STR
		CLUS_SHOW_STR
		"Show volumes and exports served by this cluster\n"
		"Be verbose and tell me every little thing you know\n")
{
	char **lvs;
	char **nfs;
	char **vis;
	char *state;
	char *node;
	char *dis;
	char *nxt;
	int verbose = (argc == 1);
	evm_volId_t lvid;
	char **look_lv;
	char **look_nfs;
	char **look_vis;

	if (!node_in_cluster()) {
		fprintf (stderr, "Clustering not enabled on this node\n");
		return CMD_SUCCESS;
	}

	lvs = get_cfg_val ("cluster.lv.*");
	nfs = get_cfg_val ("cluster.nfs.*");
	vis = get_cfg_val ("slave.*");

	if (lvs) {
/*
	For each lv in the cluster, determine what it's state is, where it's
	served, and if verbose its export parameters and which nodes have them
	visible.
 */
		fprintf (stdout, "LVid             State     On Node\n");
		fprintf (stdout, "Volume Name\n");
		if (verbose) {
			fprintf (stdout, " NFS share options\n Inter-node connectivity\n");
		}
		fprintf (stdout, "----------------------------------\n");

		for (look_lv = lvs; *look_lv; look_lv +=2) {
			lvid.whole_thing = strtoq (look_lv[0]+strlen("cluster.lv."),
			                           NULL, 16);
			node = look_lv[1];
			state = strsep (&node, ":");

			fprintf (stdout, "%016llx %-9s %s\n", lvid.whole_thing, 
			                 state, node);
			prnt_lv_name (lvid);

			if (verbose) {
				if (nfs) {
					for (look_nfs = nfs; *look_nfs; look_nfs +=2) {
						if (lvid.whole_thing ==
						    strtoq (look_nfs[0]+strlen ("cluster.nfs."),
						            NULL, 16)) {

							nxt = look_nfs[1];
							do {
								dis = strsep (&nxt, ";");
								*index (dis, '|') = ' ';
								fprintf (stdout, " %s\n", dis+9);
							} while (nxt);
							break;		// Done with this one
						}
					}
				}

				if (vis) {
					for (look_vis = vis; *look_vis; look_vis +=2) {
						dis = rindex(look_vis[0], '.') + 1;
						if (lvid.whole_thing == strtoq (dis, NULL, 16)) {
							dis[-1] = '\0';
							fprintf (stdout, " %s %s\n",
							         look_vis[0]+strlen("slave."), look_vis[1]);
						}
					}
				}
				fprintf (stdout, "\n");
			}
		}
	}
/*
	Clean up and play nice now
 */
	free_cfg_val (lvs);
	free_cfg_val (nfs);
	free_cfg_val (vis);

	return CMD_SUCCESS;
}

DEFUN	(clShNodes,
		clShNodes_cmd,
		"cluster show nodes",
		CLUS_STR
		CLUS_SHOW_STR
		"Show the list of nodes in this cluster\n")
{
	uint64 buf[256];
	ncmRqNodeList_t *rqLst = (ncmRqNodeList_t *)buf;
	ncmNodeList_t *daLst = (ncmNodeList_t *)buf;
	struct timeval to = {10, 0};
	int rc;

	if (!node_in_cluster()) {
		fprintf (stderr, "Clustering not enabled on this node\n");
		return CMD_SUCCESS;
	}

	rqLst->msgType = NCM_RQ_NODE_LIST;
	rc = sendAgileMsg (-1, rqLst, sizeof (ncmRqNodeList_t),
	                   "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, &to,
	                   buf, sizeof(buf));
	if (!rc) {
		fprintf (stderr, "Unable to get node list\n");
		return CMD_ERR_FAILED;
	}

	fprintf (stdout, "List of connected cluster nodes:\n");
	for (rc = 0; rc < daLst->numConnNodes; rc++) {
		fprintf (stdout, "%s\n", daLst->nodeNames[rc]);
	}
	return CMD_SUCCESS;
}

DEFUN	(clShPCC,
		clShPCC_cmd,
		"cluster show pccname",
		CLUS_STR
		CLUS_SHOW_STR
		"Show the cluster PCC node name\n")
{
	ccmRqPCCName_t rq;
	ccmPCCName_t rsp;
	struct timeval to = {10, 0};
	int rc;

	if (!node_in_cluster()) {
		fprintf (stderr, "Clustering not enabled on this node\n");
		return CMD_SUCCESS;
	}

	rq.hdr.msgType = ccmRqPCCName;

	rc = sendAgileMsg (-1, &rq, sizeof(rq),
	                   "ccm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, &to,
	                   &rsp, sizeof(rsp));
	if (!rc) {
		fprintf (stderr, "Unable to get PCC name\n");
		return CMD_ERR_FAILED;
	}
	fprintf (stdout, "PCC node name is: %s\n", rsp.PCCName);
	return CMD_SUCCESS;
}

DEFUN	(nodeTakeOver,
		nodeTakeOver_cmd,
		"cluster takeover OLDNODE NEWNODE",
		CLUS_STR
		"Takeover cluster services\n"
		"From this node\n"
		"And assign those services to this node\n")
{
	ccmTakeOverRq_t take_msg;
	uint64 buf[1024];
	ccmRsp_Msg_t *rsp = (ccmRsp_Msg_t *)buf;
	int rc;
	struct timeval timeout = {120, 0};

	take_msg.hdr.msgType = ccmTakeOver;
	take_msg.hdr.clientPrivate = 0;
	strncpy (take_msg.old_node, argv[0], sizeof(take_msg.old_node));
	strncpy (take_msg.new_node, argv[1], sizeof(take_msg.new_node));

	rc = sendAgileMsg (-1, &take_msg, sizeof (take_msg), 
	                   "ccm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, &timeout,
	                   buf, sizeof(buf));

	if (!rc) {
		fprintf (stderr, "Didn't get a response to takeover command\n");
		rc = CMD_ERR_FAILED;
	}
	else {
		if ((rsp->hdr.msgType == ccmStatRsp) && (rsp->status == NFX_OK)) {
			if (rc > sizeof (ccmRsp_Msg_t)) {
				fprintf (stdout, "%s\n", (char *)rsp + sizeof (ccmRsp_Msg_t));
			}
			rc = CMD_SUCCESS;
		}
		else {
			fprintf (stderr, "Got an error on takeover command: %s\n",
			         (rc > sizeof (ccmRsp_Msg_t)) ? 
			         (char *)rsp + sizeof (ccmRsp_Msg_t) : "");
			rc = CMD_ERR_FAILED;
		}
	}
	
	return rc;
}

static char **
remove_shares (evm_volId_t lv, char *node)
{
	char *share_tag;
	char **cfg;
	char *nfs_opts;
	char *nxt;
	char *path;
	char *laddr;
	uint64 msg_buf[256];
	ncmMsg_t *ncm_msg = (ncmMsg_t *)msg_buf;
	mnt_infoReq_t *umnt_msg = (mnt_infoReq_t *)ncm_msg->appMsg;
	int sz;

	asprintf (&share_tag, "cluster.nfs.%016llx", lv.whole_thing);
	cfg = get_cfg_val (share_tag);
	free (share_tag);

	if (cfg) {
		nfs_opts = strdup (cfg[1]);		// We'll be messing with this copy

		ncm_msg->msgType = NCM_SEND;
		ncm_msg->rqResp = 0;
		strcpy (ncm_msg->appName, "mountd");
		strcpy (ncm_msg->nodeName, node);

		umnt_msg->msgHdr.msgType = MNT_UNEXPORT_REQ;
		umnt_msg->optionsLen = 0;

		for (nxt = nfs_opts; nxt;) {
			path = strsep (&nxt, ";");
			laddr = strsep (&path, ":");
			*index (path, '|') = '\0';

			sz = (char *)umnt_msg - (char *)ncm_msg + 
			     sizeof (mnt_infoReq_t) + strlen (path) + 1;

			strcpy (umnt_msg->pathName, path);
			umnt_msg->nameLen = strlen (path);

			umnt_msg->localIpAddr = strtoul (laddr, NULL, 16);

			fprintf (stdout, "Removed export: %s on %s\n", path, node);

			sendAgileMsg (-1, ncm_msg, sz,
			              "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
			              NULL, NULL, 0);
		}

		free (nfs_opts);
	}
	else {
		fprintf (stderr, "LV %016llx is online but not shared\n", 
		                 lv.whole_thing);
	}
	
	return cfg;
}

static void
move_LV (evm_volId_t lv, char *from, char *to)
{
	uint64 msgbuf[512];
	ncmMsg_t *ncm = (ncmMsg_t *)msgbuf;
	ea_movOutFsysReq_t *mo = (ea_movOutFsysReq_t *)ncm->appMsg;
	ea_movOutFsysRsp_t *mo_rsp = (ea_movOutFsysRsp_t *)ncm->appMsg;
	ea_movInFsysReq_t *mi = (ea_movInFsysReq_t *)ncm->appMsg;
	ea_movInFsysRsp_t *mi_rsp = (ea_movInFsysRsp_t *)ncm->appMsg;
	int sz;
	int rc;
	char **lvnames;
	char **look;
	char *itsname = NULL;
	char idstr[17];
	struct timeval ea_wait = {300, 0}; // cmd_vol.c has it at 300!
	                                   // A bit much for a clean move IMO
/*
	EA seems to need the volume "name" and id. We only have the id.
 */
	lvnames = get_cfg_val ("lvname.*");

	if (!lvnames) {
		fprintf (stderr, "Unable to determine the name of LV %016llx\n",
		         lv.whole_thing);
		return;
	}
	sprintf (idstr, "%016llx", lv.whole_thing);

	for (look = lvnames; *look; look +=2) {
		if (!strncmp (look[1], idstr, 16)) {
			itsname = strdup (look[0] + strlen ("lvname."));
			break;
		}
	}
	free_cfg_val (lvnames);

	if (!itsname) {
		fprintf (stderr, "Unable to determine the name of LV %016llx\n",
		         lv.whole_thing);
		return;
	}
/*
	EA moveout from old node
 */
	memset (ncm, 0, sizeof (ncmMsg_t));
	ncm->msgType = NCM_SEND;
	strcpy (ncm->appName, "ea");
	strcpy (ncm->nodeName, from);
	ncm->rqResp = 1;

	memset (mo, 0, sizeof(ea_movOutFsysReq_t));
	mo->hdr.msgType = EA_MOVOUT_FSYS_REQ;
	strcpy (mo->hdr.node, from);
	mo->fsysId.volId = lv;
	strcpy (mo->fsysId.fsysName, itsname);

	sz = (char *)mo - (char *)ncm + sizeof (ea_movOutFsysReq_t);

	rc = sendAgileMsg (-1, ncm, sz,
	                   "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
	                   &ea_wait, msgbuf, sizeof(msgbuf));

	if ((rc < (sizeof(ncmMsg_t) + sizeof(ea_movOutFsysRsp_t) - 
	          sizeof(long long))) ||
	    (mo_rsp->result != NFX_OK)) {
		fprintf (stderr, "Failed to unmount %s from %s\n", itsname, from);
	}
	else {
		fprintf (stdout, "Unmounted %s from %s\n", itsname, from);
/*
	EA movein on new node
 */
		memset (ncm, 0, sizeof (ncmMsg_t));
		ncm->msgType = NCM_SEND;
		strcpy (ncm->appName, "ea");
		strcpy (ncm->nodeName, to);
		ncm->rqResp = 1;

		memset (mi, 0, sizeof(ea_movInFsysReq_t));
		mi->hdr.msgType = EA_MOVIN_FSYS_REQ;
		strcpy (mi->hdr.node, to);
		mi->fsysId.volId = lv;
		strcpy (mi->fsysId.fsysName, itsname);

		sz = (char *)mi - (char *)ncm + sizeof (ea_movInFsysReq_t);

		rc = sendAgileMsg (-1, ncm, sz,
		                   "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
		                   &ea_wait, msgbuf, sizeof(msgbuf));

		if ((rc < (sizeof(ncmMsg_t) + sizeof(ea_movInFsysRsp_t) - 
		          sizeof(long long))) ||
	    	(mi_rsp->result != NFX_OK)) {
			fprintf (stderr, "Failed to mount %s on %s\n", itsname, to);
		}
		else {
			fprintf (stdout, "Mounted %s on %s\n", itsname, to);
		}
	}

	free (itsname);
	return;
}

static void
add_shares (evm_volId_t lv, char *node, char **shares)
{
	uint64 msgBuf[1024];
	ncmMsg_t *ncmMsg = (ncmMsg_t *)msgBuf;
	mnt_infoReq_t *expReq = (mnt_infoReq_t *)ncmMsg->appMsg;
	mnt_infoResp_t *expRsp = (mnt_infoResp_t *)ncmMsg->appMsg;
	char *optStr;
	char *nxtOpt = shares[1];
	char *volname;
	char *ops;
	int sz;
	int rc;
	struct timeval waitTime = { 10, 0 };
	char *msg;

	if (shares) {
		while ((optStr = strsep(&nxtOpt, ";"))) {
	
			memset (ncmMsg, 0, sizeof(ncmMsg_t));
			ncmMsg->msgType = NCM_SEND;
			strcpy (ncmMsg->appName, "mountd");
			strcpy (ncmMsg->nodeName, node);
			ncmMsg->rqResp = 1;
	
			memset (expReq, 0, sizeof (mnt_infoReq_t));
			expReq->msgHdr.msgType = MNT_EXPORT_REQ;
			expReq->volId = lv;
	
			volname = &(optStr[9]); // Name starts after export address:
			ops = index (optStr, '|');
	
			if (ops) {
				*ops = '\0';
				ops++;
				expReq->optionsLen = strlen(ops);
			}
	
			expReq->nameLen = strlen (volname);
			memcpy (expReq->pathName, volname, expReq->nameLen);
	
			if (ops) {
				memcpy (&expReq->pathName[expReq->nameLen], ops, 
				        expReq->optionsLen);
			}
	
			sz = (int)((char *)expReq - (char *)ncmMsg) + 
			     sizeof (mnt_infoReq_t) + expReq->nameLen + expReq->optionsLen;
	
			rc = sendAgileMsg (-1, msgBuf, sz,
			                   "ncm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE, 
			                   &waitTime,
			                   msgBuf, sizeof(msgBuf));
	
			if ((rc <= sizeof (ncmMsg_t)) || 
			    (expRsp->msgHdr.status != NFX_OK)) {
	
				switch (expRsp->msgHdr.status) {
					case MOUNTD_ERR_DUPLICATE_EXPORT:
						msg = "MOUNTD_ERR_DUPLICATE_EXPORT";
						break;
					case MOUNTD_ERR_MORE_GENERIC_EXPORT_FOUND:
						msg = "MOUNTD_ERR_MORE_GENERIC_EXPORT_FOUND";
						break;
					case MOUNTD_ERR_MORE_SPECIFIC_EXPORT_FOUND:
						msg = "MOUNTD_ERR_MORE_SPECIFIC_EXPORT_FOUND";
						break;
					case MOUNTD_ERR_CONFLICT_EXPORT:
						msg = "MOUNTD_ERR_CONFLICT_EXPORT";
						break;
					case MOUNTD_ERR_MULTIPLE_INTERFACE:
						msg = "MOUNTD_ERR_MULTIPLE_INTERFACE";
						break;
	
					default:
						msg = "mountd rejected export request";
				}
				fprintf (stderr, "Less than optimal result, %s for %s:%s\n",
				         msg, volname, ops ? ops : "NULL OPTIONS");
			}
			else {
				fprintf (stdout, "successful export %s as %s\n", volname, 
				        ops ? ops : "NULL OPTIONS");
			}
		}

		free_cfg_val (shares);
	}
	return;
}

DEFUN	(nodeRestore,
		nodeRestore_cmd,
		"cluster restore NODENAME",
		CLUS_STR
		"Restore services taken over\n"
		"From this node\n")
{
	char **taken_stuff;
	char **look;
	char *srch;
	int len;
	char *ipaddr_str;
	char *port_str;
	uint32 ipaddr;
	uint32 mask;
	char *mask_str;
	char *old_node;
	struct in_addr addr;
/*
	First, find out what, if anything, was taken from this node
 */
	asprintf (&srch, "cluster.taken.%s.*", argv[0]);

	taken_stuff = get_cfg_val (srch);
	free (srch);

	if (!taken_stuff) {
		fprintf (stdout, "%s has nothing to restore\n", argv[0]);
		return CMD_SUCCESS;
	}
/*
	We should make sure the restored node can indeed see the LVs it once
	had before we restore them too. In case the LV no longer exists, just
	pass it over.
 */
	asprintf (&srch, "cluster.taken.%s.lv.", argv[0]);
	len = strlen (srch);

	for (look = taken_stuff; *look; look += 2) {

		if (!strncmp (look[0], srch, len)) {
			char *key;
			char **cfg_obj;
			char *lvstr = look[0] + len;

			asprintf (&key, "cluster.lv.%s", lvstr);
			cfg_obj = get_cfg_val (key);
			free (key);
			if (!cfg_obj) {		// LV no longer exists in cluster
				del_cfg_obj (&look[0]);
			}
			else {
				free_cfg_val (cfg_obj);	// No longer interested in cluster.lv

				asprintf (&key, "slave.%s.%s", argv[0], lvstr);

				cfg_obj = get_cfg_val (key);
				free (key);

				if (!cfg_obj) {		// Yikes!
					fprintf (stdout, "%s can no longer see lv %s\n",
					         argv[0], lvstr);
					fprintf (stdout, "Restoration aborted\n");

					free (srch);
					free_cfg_val (taken_stuff);
					return CMD_ERR_FAILED;
				}
				free_cfg_val (cfg_obj);
			}
		}
	}
	free (srch);
	free_cfg_val (taken_stuff);
/*
	Now, see if anything that was taken still exists
 */
	asprintf (&srch, "cluster.taken.%s.*", argv[0]);

	taken_stuff = get_cfg_val (srch);
	free (srch);

	if (!taken_stuff) {
		fprintf (stdout, "%s has nothing to restore\n", argv[0]);
		return CMD_SUCCESS;
	}
/*
	First, restore its IP addresses
 */
	asprintf (&srch, "cluster.taken.%s.ip.", argv[0]);
	len = strlen (srch);

	for (look = taken_stuff; *look; look += 2) {

		if (!strncmp (look[0], srch, len)) {

// 		Delete address from the old

			ipaddr_str = look[0] + len;
			ipaddr = strtoul (ipaddr_str, NULL, 16);

			mask_str = look[1];
			old_node = strsep (&mask_str, ":");
			port_str = strsep (&mask_str, "/");

			addr.s_addr = ipaddr;
			ipaddr_str = inet_ntoa (addr);

			if (ipmd_DelIpCluster (port_str, ipaddr, NULL, 0, old_node)) {
				fprintf (stderr, "For some reason ipm failed deleting %s ",
				         ipaddr_str);
				fprintf (stderr, "From %s:%s\nCarrying on anyway\n", old_node, 
				         port_str);
			}

			mask = strtoul (mask_str, NULL, 10);

// 		Restore to the new

			if (ipmd_AddIpCluster (port_str, ipaddr, mask, NULL, 0, argv[0])) {
				fprintf (stderr, "For some reason ipm failed adding %s ",
				         ipaddr_str);
				fprintf (stderr, "To %s:%s\nCarrying on anyway\n", argv[0],
				         port_str);
			}

			fprintf (stdout, "Moved %s from %s:%s to %s:%s\n", ipaddr_str,
			         old_node, port_str, argv[0], port_str);

//		Delete "taken" object from config file (no longer taken, right)

			del_cfg_obj (&look[0]);
		}
	}

	free (srch);
/*
	Move the LVs and their exports back to where they were
 */
	asprintf (&srch, "cluster.taken.%s.lv.", argv[0]);
	len = strlen (srch);

	for (look = taken_stuff; *look; look += 2) {

		if (!strncmp (look[0], srch, len)) {

			evm_volId_t lvid;
			char **shares;

			lvid.whole_thing = strtouq (look[0] + len, NULL, 16);
/*
	Remove the nfs exports from the node that took these services
 */
			shares = remove_shares (lvid, look[1]);
/*
	Move the LVs from the taken node to the one we're restoring
 */
			move_LV (lvid, look[1], argv[0]);
/*
	Restore its exports
 */
			add_shares (lvid, argv[0], shares);
/*
	Delete "taken" object from config file.
 */
			del_cfg_obj (&look[0]);
		}
	}

	free (srch);
	free_cfg_val (taken_stuff);

	return CMD_SUCCESS;
}

#undef ENABLE_DISABLE_NODES
#if defined (ENABLE_DISABLE_NODES)
DEFUN	(nodeEnable,
		nodeEnable_cmd,
		"cluster enable NODENAME",
		CLUS_STR
		"Enable this node's participation in the cluster\n"
		"The new node's name\n")
{
	return CMD_ERR_FAILED;
}

DEFUN	(nodeDisable,
		nodeDisable_cmd,
		"cluster disable NODENAME",
		CLUS_STR
		"Disable this node from participating in the cluster\n"
		"The node's name\n")
{
	return CMD_ERR_FAILED;
}
#endif

static void
update_node_failtime (char *timeStr)
{
	lnm_newTimeoutMsg_t tell_lnm;
	char *failtime[3];

	failtime[0] = "cluster.nodefailtime";
	failtime[1] = timeStr;
	failtime[2] = NULL;

	put_cfg_val (failtime);

	tell_lnm.msgType = CLUSTER_NODE_TIMEOUT_MSG;
	tell_lnm.newTimeout = atoi (timeStr);

	sendAgileMsg (-1, &tell_lnm, sizeof(tell_lnm), 
	              "cluster", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
	              NULL, NULL, 0);
}

DEFUN	(setTOtime,
		setTOtime_cmd,
		"cluster modify faildetecttime <10-120>",
		CLUS_STR
		CLUS_MOD_STR
		"Failure detection time\n"
		"How many seconds before declaring node failed\n")
{
	update_node_failtime (argv[0]);

	return CMD_SUCCESS;
}

DEFUN	(shTOtime,
		shTOtime_cmd,
		"cluster show faildetecttime",
		CLUS_STR
		CLUS_SHOW_STR
		"How many seconds before declaring node failed\n")
{
	char **nodefail;
	int detectTime;

	if (!node_in_cluster()) {
		fprintf (stderr, "Clustering not enabled on this node\n");
		return CMD_SUCCESS;
	}

	nodefail = get_cfg_val ("cluster.nodefailtime");
	if (nodefail) {
		detectTime = atoi (nodefail[1]);
		free_cfg_val (nodefail);
	}
	else {
		detectTime = CLUST_NODE_FAIL_TIMEOUT;
	}
	if (detectTime) {
		fprintf (stdout, "Node Failure detect time: %d seconds\n", detectTime);
	}
	else {
		fprintf (stdout, "Node Failure detection disabled\n");
	}

	return CMD_SUCCESS;
}

DEFUN	(takeover_ena_dis,
		takeover_ena_dis_cmd,
		"cluster faildetect (enabled|disabled)",
		CLUS_STR
		"To detect node failures (or not)\n")
{
	char *failtime;

	if (*argv[0] == 'e') {
		asprintf (&failtime, "%d", CLUST_NODE_FAIL_TIMEOUT);
	}
	else {
		failtime = strdup ("0");
	}

	update_node_failtime (failtime);
	free (failtime);

	return CMD_SUCCESS;
}

ALIAS	(shTOtime,
		shtakeover_ena_dis,
		"cluster show faildetect",
		CLUS_STR
		CLUS_SHOW_STR
		"Show state of cluster node failure detection\n")

/*-----------------------------------------------------------------
 * Name:install_cluster
 *
 * Description: Install the cluster commands menu items
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 28 Feb 2002
 *
 *-----------------------------------------------------------------
 */
void
install_cluster (priv_t myPriv, enum node_type node)
{
/*
	Organizing this switch statment such that the higher priority commands
	are first allows us to fall through and pick up the lower priority
	ones at no extra cost.
 */
	switch (myPriv) {
		case diagPriv:
			install_element (node, &setTOtime_cmd);
			install_element (node, &shTOtime_cmd);
		case fullPriv:
			install_element (node, &nodeRestore_cmd);
#if defined (ENABLE_DISABLE_NODES)
			install_element (node, &nodeDisable_cmd);
#endif
		case changePriv:
			install_element (node, &clusModName_cmd);
			install_element (node, &clusDelName_cmd);
			install_element (node, &clusType_cmd);
			install_element (node, &nodeTakeOver_cmd);
			install_element (node, &takeover_ena_dis_cmd);
		case addPriv:
#if defined (ENABLE_DISABLE_NODES)
			install_element (node, &nodeEnable_cmd);
#endif
		case viewPriv:
/*
	These commands are available in all modes
 */
			install_element (node, &clusShowName_cmd);
			install_element (node, &clusShowVols_cmd);
			install_element (node, &clShNodes_cmd);
			install_element (node, &clShPCC_cmd);
			install_element (node, &shtakeover_ena_dis);
	}
}
