/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:
 *
 * RCS:  $Id: cmd_interface.c,v 1.18 2002/11/22 02:51:24 swang Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:
 *
 * Created by:	Greg Seibert
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#include <sys/types.h>
#include <pwd.h>
#include <readline/history.h>
#include <grp.h>
#include <unistd.h>

#include "zebra.h"
#include "command.h"
#include "log.h"
#include "vtysh.h"
#include "memory.h"

#include "nfxsh.h"
#if defined(__OpenBSD__)
#include <netinet/if_ether.h>
#endif /* SSC_mips */
#include "exports/ipmd-lib.h"

uint64		result_buf[IPM_SSC_BUFSIZE];

#define INTF_PRINTF(a)	

/* separete node name from device name */
static void separate_ifname(char *in_ifname, char** node_p, char** ifname_p)
{
	static char	node[64];
	static char	ifname[32];
	char		*str_p;
	int			len;

	for (str_p = in_ifname; *str_p && (*str_p != ':'); str_p++);      
	
	if (*str_p) {				/* node name found */
		len = (int)(str_p - in_ifname);
		strncpy(node, in_ifname, len);
		str_p ++;
		
    }
    else {						/* no node infor found */
		strcpy(node, "local");
		str_p = in_ifname;
	}

	strcpy(ifname, str_p);

	INTF_PRINTF(("interface %s separated node=%s, ifname=%s\n",
				 in_ifname, node, ifname));

	*node_p = node;
	*ifname_p = ifname;
	return;
}

/* convert ascii a.b.c.d/mask string into 
 * network ordered a.b.c.d mask pair
 */  
int32							/* NFX_OK, NFX_ERR */
net_atoipmask(char *addr_str, uint32 *addr, uint32 *mask)
{
	char	*str_p;
	int		mask_len, i, j;
	int		rc = NFX_OK;
	char	mask_in[4];
	char	ip_in[20];
	

	for (str_p = addr_str; *str_p && (*str_p != '/'); str_p++);

	if (!(*str_p))				/* string does not contain '/' */
		return NFX_ERR;
	
	/* break address_str into two separate string */
	*str_p = '\0';
	
	/* get address */
	if (!inet_aton(addr_str, (struct in_addr*)addr)) {
		rc = NFX_ERR;
		goto out;
	}
	
	/* make sure inet_aton() convert corretly */
	sprintf(ip_in, "%d.%d.%d.%d", 
			((uchar8*)(addr))[0],
			((uchar8*)(addr))[1],
			((uchar8*)(addr))[2],
			((uchar8*)(addr))[3]);
	
	if (strcmp(ip_in, addr_str)) {
		rc = NFX_ERR;
		goto out;
	}

	/* get mask len */
	mask_len = atoi((str_p+1));
	if (mask_len < 0 || mask_len>32) {
		rc = NFX_ERR;
		goto out;
	}
	
	/* convert masklen into mask */
	bzero(mask_in, sizeof(mask_in));
	for (i = 0; i < 4; i++) {
		if (mask_len >= 8) {
			mask_in[i] = 0xff;
			mask_len -= 8;
		}
		else {
			for (j=0; j<mask_len; j++)
				mask_in[i] = (mask_in[i]>>1)|0x80;
			
			break;
		}
	}
	
	*mask = *(uint32*)mask_in;

	INTF_PRINTF(("ip is %s,", inet_ntoa( *(struct in_addr*)addr)));
	INTF_PRINTF(("masklen is %d, mask =%s\n", 
				mask_len, inet_ntoa( *(struct in_addr*)mask)));
	
out:
	/* reset *str_p */
	*str_p = '/';
	return rc;
}



/*-----------------------------------------------------------------
 * Name: intf_enable
 *
 * Description: enable an interface
 *
 * Created by:	Su Wang
 *
 * Date Created: 01/28/02
 *
 *-----------------------------------------------------------------
 */
DEFUN	(intf_enable,
		 ena_intf_ena_cmd,
		 "interface enable INTERFACE",
		 "Network Interface commands\n"
		 "Enable interface \n"
		 "Interface name\n")
{

	char	*node, *ifname;
	int32	rc;
		
	separate_ifname(argv[0], &node, &ifname);
	
	rc = ipmd_EnableIntfCluster(ifname, NULL, 0, node);
	
	if ( rc == NFX_OK)
		return CMD_SUCCESS;
	else {
		ipmd_PrntErrMsg(rc, ifname, NULL, 0, 0);
		return CMD_ERR_FAILED;
	}
}

/*-----------------------------------------------------------------
 * Name: intf_disable
 *
 * Description: disable an interface
 *
 * Created by:	Su Wang
 *
 * Date Created: 01/28/02
 *
 *-----------------------------------------------------------------
 */
DEFUN	(intf_disable,
		 ena_intf_dis_cmd,
		 "interface disable INTERFACE",
		 "Network Interface commands\n"
		 "Disable interface\n"
		 "Interface name\n")
{
	char	*node, *ifname;
	int32	rc;
	
	separate_ifname(argv[0], &node, &ifname);

	rc = ipmd_DisableIntfCluster(ifname, NULL, 0, node);
	
	if (rc == NFX_OK)
		return CMD_SUCCESS;
	else {
		ipmd_PrntErrMsg(rc, ifname, NULL, 0, 0);
		return CMD_ERR_FAILED;
	}
}


/*-----------------------------------------------------------------
 * Name: intf_addip
 *
 * Description: add IP to an interface
 *
 * Created by:	Su Wang
 *
 * Date Created: 01/28/02
 *
 *-----------------------------------------------------------------
 */
DEFUN	(intf_addip,
		 ena_intf_addip_cmd,
		 "interface add INTERFACE -a IP_ADDRESS/MASKLEN",
		 "Network interface commands\n"
		 "add IP interface \n"
		 "Interface name\n"
		 "IP address\n"
	     "Interface IP address/masklen\n")
{
	uint32	ip, mask;
	char	*node, *ifname;
	int32	rc;
	
	INTF_PRINTF(("add ip parm1 =%s, parm2 = %s, opt =%s\n", 
		   argv[0], argv[1], optarg));

	separate_ifname(argv[0], &node, &ifname);

	if (net_atoipmask(argv[1], &ip, &mask) != NFX_OK) {
		printf("Incorrect ip/mask_len pair %s, format A.B.C.D/masklen\n",
			   argv[1]);
		return CMD_SUCCESS;
	}

	/* catch obvious errors */
	if (!ip || ip == 0xffffffff ||
		!mask || mask == 0xffffffff) {
		printf("IP/Mask pair %s is not valid\n",argv[1]);
		return CMD_ERR_FAILED;
	}


	rc = ipmd_AddIpCluster(ifname, ip, mask, NULL, 0, node);
	
	if ( rc == NFX_OK)
		return CMD_SUCCESS;
	else {
		ipmd_PrntErrMsg(rc, ifname, NULL, ip, mask);
		return CMD_ERR_FAILED;
	}
}

/*-----------------------------------------------------------------
 * Name: intf_rmip
 *
 * Description: remove IP from an interface
 *
 * Created by:	Su Wang
 *
 * Date Created: 01/28/02
 *
 *-----------------------------------------------------------------
 */
DEFUN	(intf_rmip,
		 ena_intf_rmip_cmd,
		 "interface delete INTERFACE -a A.B.C.D",
		 "Network interface commands\n"
		 "Delete IP interface \n"
		 "Interface name\n"
		 "IP address\n"
		 "Interface IP address\n")		 

{
	uint32	ip;
	char	*node, *ifname;
	int32	rc;
	
	separate_ifname(argv[0], &node, &ifname);

	if (!inet_aton(argv[1], (struct in_addr*)&ip))
		return CMD_ERR_FAILED;

	rc = ipmd_DelIpCluster(ifname, ip, NULL, 0, node);
	
	if (rc == NFX_OK)
		return CMD_SUCCESS;
	else {
		ipmd_PrntErrMsg(rc, ifname, NULL, ip, 0);
		return CMD_ERR_FAILED;
	}
}

/*-----------------------------------------------------------------
 * Name: intf_setmtu
 *
 * Description: ret mtu size
 *
 * Created by:	Su Wang
 *
 * Date Created: 01/28/02
 *
 *-----------------------------------------------------------------
 */
DEFUN	(intf_setmtu,
		 ena_intf_setmtu_cmd,
		 "interface setmtu INTERFACE -s MTUSIZE",
		 "Network interface commands\n"
		 "Set Interface mtu  \n"
		 "Interface name\n"
		 "mtu size\n"
		 "<1500-9600>")
{
	char	*node, *ifname;
	int32	rc;
	
	separate_ifname(argv[0], &node, &ifname);

	rc = ipmd_SetIntfMtuCluster(ifname, atoi(argv[1]), NULL, 0, node);
	
	if (rc == NFX_OK)
		return CMD_SUCCESS;
	else {
		ipmd_PrntErrMsg(rc, ifname, NULL, 0, 0);
		return CMD_ERR_FAILED;
	}
}

/*-----------------------------------------------------------------
 * Name: intf_setautoneg
 *
 * Description: enable/disable autoneg.
 *
 * Created by:	Su Wang
 *
 * Date Created: 05/09/02
 *
 *-----------------------------------------------------------------
 */
DEFUN	(intf_setautoneg,
		 ena_intf_setautoneg_cmd,
		 "interface autoneg INTERFACE -s (enable|disable)",
		 "Network interface commands\n"
		 "Set Ethernet Driver Auto Negotiate Mode\n"
		 "Interface name\n"
		 "enable or disable\n")
{
	char	*node, *ifname;
	int32		rc;

	separate_ifname(argv[0], &node, &ifname);

	if ((strncmp(argv[1], "e", 1) == 0))
		rc = ipmd_EnableIntfAutoCluster(ifname,NULL, 0, node);
	else
		rc = ipmd_DisableIntfAutoCluster(ifname,NULL, 0, node);
	
	if ((rc == NFX_OK))
		return CMD_SUCCESS;
	else {
		ipmd_PrntErrMsg(rc, ifname, NULL, 0, 0);
		return CMD_ERR_FAILED;
	}
}

/*-----------------------------------------------------------------
 * Name: intf_show
 *
 * Description: show ssc interface table.
 *
 * Created by:	Su Wang
 *
 * Date Created: 01/29/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (intf_show,
         vi_intf_show_cmd,
		 "interface show (ip|interface)"
		 "[-a SPEC] [-n NODE]",
		 "Network interface commands\n"
		 "Display interface  \n"
		 "ip table or interface table\n")
{
	int			count=0, need=0;
	ipm_ip_t	*ipm_ipp;
	ipm_if_t	*ipm_ifp;
	char		ip[20], mask[20], mac[20];
	int			c;
	char		node[64], parm2[20];
	int			opt_num = 0;	/* num. of options */
	int			my_argc = 1;
	int32		rc;
	
	/* parse options */
	node[0] = '\0';
	parm2[0] = '\0';
	optind = 1;
	while ((c = getopt(argc, argv, "a:n:")) != -1) {
		switch (c) {
		case 'n':
			strcpy(node, optarg);
			opt_num ++;
			break;
			
		case 'a':
			strcpy(parm2, optarg);
			opt_num ++;
			break;
		}
	}

	
	/* check parameter number matches */
	my_argc += 2* opt_num;
	if (my_argc != argc) {
		printf("incorrect number of parameters, argc %d, in_arg %d\n",
			   argc, my_argc);
		return  CMD_ERR_FAILED;
	}
	
	/* node has not been specified */
	if (strcmp(argv[0], "ip")== 0)	{ /* IP table */
		uint32	addr = 0;

		if (!(*node))
			strcpy(node, "local");

		/* ip specified */
		if (*parm2) 
			if (!inet_aton(parm2, (struct in_addr*)&addr)) {
				printf("incorrect IP address %s\n", parm2);
				return  CMD_ERR_FAILED;
			}
		
		rc = ipmd_GetIpByNameCluster(NULL, (ipm_ip_t*)result_buf, 
									IPM_SSC_BUFSIZE*sizeof(uint64), 
									 &count, &need, node);

		
		if (rc != IPM_GET_OK) {
			ipmd_PrntErrMsg(rc, NULL, NULL, 0, 0);
			return CMD_ERR_FAILED;
		}

		printf("IP Table\n");
		printf("---------------------------------------------------------\n");
		printf("dest              mask              interface\n");
		printf("---------------------------------------------------------\n");
		for (ipm_ipp = (ipm_ip_t*)result_buf; count >0; count--, ipm_ipp++) {

			if (addr && addr != ipm_ipp->ip_addr)
				continue;

			strcpy(ip, inet_ntoa(*((struct in_addr*)&ipm_ipp->ip_addr)));
			strcpy(mask, inet_ntoa(*((struct in_addr*)&ipm_ipp->ip_mask)));
#if defined(LINK_AGG)
			if (ipm_ipp->ip_vifName[0] != '\0')
				printf("%-18s%-18s%s\n", 
					   ip, mask, ipm_ipp->ip_vifName);
			else
#endif /* LINK_AGG */
				printf("%-18s%-18s%s\n", 
					   ip, mask, ipm_ipp->ip_ifname);

		}	
		
	}
	else {
		char *node_p=NULL, *ifname_p=NULL;
		
		/* interface specified */
		if (*parm2)
			separate_ifname(parm2, &node_p, &ifname_p);
		
		/* if node specified by user, use node info */
		if (*node)
			node_p = node;
		
		/* still node not determined */
		if (!node_p)
			node_p = "local";
		
		rc = ipmd_GetIfByNameCluster(ifname_p, (ipm_if_t*)result_buf, 
									IPM_SSC_BUFSIZE*sizeof(uint64), 
									 &count, &need, node_p);
		
		if (rc != IPM_GET_OK) {
			ipmd_PrntErrMsg(rc, NULL, NULL, 0, 0);
			return CMD_ERR_FAILED;
		}		

		if (count == 1) {			/* detailed */
			ipm_ifp = (ipm_if_t*)result_buf;
			printf("Interface     :  %s\n", ipm_ifp->if_name);
			printf("Admin. State  :  %s\n", (ipm_ifp->if_flags & IFF_UP)?"UP":"DOWN");
			printf("Operat State  :  %s\n", (ipm_ifp->if_auxFlags & IPM_AUXIF_RUNNING)?"UP":"DOWN");
			printf("Auto Negotiate:  %s\n", (ipm_ifp->if_auxFlags & IPM_AUXIF_AUTO)?"ENABLED":"DISABLED");
			printf("Mtu Size      :  %u\n", ipm_ifp->if_mtusize);
			strcpy(mac, ether_ntoa((struct ether_addr*)ipm_ifp->if_mac));
			printf("Mac Address   :  %s\n", mac);
		}
		else {
			printf("Interface  Table\n");
			printf("---------------------------------------------------------\n");
			printf("Interface     State      Mac Address          VIF\n");
			printf("---------------------------------------------------------\n");
			for (ipm_ifp = (ipm_if_t*)result_buf; count >0; count--, ipm_ifp++) {
				strcpy(mac, ether_ntoa((struct ether_addr*)ipm_ifp->if_mac));
#if defined(LINK_AGG)
				if (ipm_ifp->if_vifName[0] != '\n')
					printf("%-15s%-10s%-22s%s\n", 
						   ipm_ifp->if_name, 
						   ((ipm_ifp->if_flags & IFF_UP)?"UP":"DOWN"), 
						   mac,
						   ipm_ifp->if_vifName);
				else
#endif /* LINK_AGG */
					printf("%-15s%-12s%s\n", ipm_ifp->if_name, 
						   (ipm_ifp->if_flags & IFF_UP)?"UP":"DOWN", 
						   mac);

			}
		}
	}
	return CMD_SUCCESS;
}

/*-----------------------------------------------------------------
 * Name: ipmd_diag
 *
 * Description: ipmd diag
 *
 * Created by:	Su Wang
 *
 * Date Created: 01/29/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (ipmd_diag,
         ipmd_diag_cmd,
		 "interface ipmd (restart|shutdown|status)",
		 "Network interface commands\n"
		 "ipmd debug\n"
		 "options\n")
{
	if (strcmp(argv[0], "restart") == 0)
		ipmd_Restart(NULL);
	else if (strcmp(argv[0], "shutdown") == 0)
		ipmd_ShutDown(NULL);
	else 
		ipmd_ValidateIpmMem();
	return CMD_SUCCESS;
}


void
install_interface (priv_t myPriv, enum node_type node)
{
	switch (myPriv) {
	case diagPriv:
		install_element(node, &ipmd_diag_cmd);
	case fullPriv:
		install_element(node, &ena_intf_ena_cmd);
		install_element(node, &ena_intf_dis_cmd);
		install_element(node, &ena_intf_setmtu_cmd);
		install_element(node, &ena_intf_setautoneg_cmd);
	case changePriv:
	case addPriv:
		install_element(node, &ena_intf_addip_cmd);
		install_element(node, &ena_intf_rmip_cmd);			

	case viewPriv:
		install_element(node, &vi_intf_show_cmd);
	}
	
}
