/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:
 *
 * RCS:  $Id: cmd_nameservice.c,v 1.10 2002/11/22 21:06:37 maximk Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:
 *
 * Created by:	Greg Seibert
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#include <sys/types.h>
#include <pwd.h>
#include <readline/history.h>
#include <grp.h>
#include <unistd.h>

#include "zebra.h"
#include "command.h"
#include "log.h"
#include "vtysh.h"
#include "memory.h"

#include "nfxsh.h"

#include "exports/nfxnis-api.h"

#define NAMSVCSTR "Host name resolution\n"

#undef ENABLE_DISABLE		// I just don't see a use for them over add/delete

static __inline__ int mknet (char *mskstr, in_addr_t *net)
{
	int mask_bits;
	int tmpmask;

	mask_bits = inet_net_pton (AF_INET, mskstr, net, sizeof (in_addr_t));

	if (mask_bits < 0) {
		perror ("Invalid network specified");
		return CMD_ERR_FAILED;
	}

	for (tmpmask = 0x80000000; mask_bits; mask_bits--) {
		tmpmask >>= 1;
	}
	tmpmask <<= 1;

	return htonl (tmpmask);
}

static __inline__ int
is_valid_srvr_addr (char *ip_addr)
{
	char *this;
	int dot_count;

	for (this = ip_addr, dot_count = 0; *this && (dot_count < 3); this++) {
		if (*this == '.') {
			dot_count++;
		}
	}
	return (dot_count == 3) && *this;
}

DEFUN	(add,
		add_cmd,
		"nameservice add (nis|dns) (default|A.B.C.D/M) -d DOMAINNAME -a A.B.C.D",
		NAMSVCSTR
		"Add this service\n"
		"Network served by NIS/YP name service\n"
		"Network served by DNS name service\n"
		"For the default\n"
		"on this specfic network\n"
		"is assigned domain name\n"
		"NIS domain name or DNS search suffix\n"
		"specify the server\n"
		"the address of the server\n")
{
	nfxnis_serv_t srv_type;
	in_addr_t net;
	in_addr_t mask;
	in_addr_t srvr;
	struct in_addr srvr_addr;
	int rc = CMD_SUCCESS;

	if (*argv[0] == 'n') {
		srv_type = nis_name;
	}
	else {
		srv_type = dns_name;
	}

	if (*argv[1] == 'd') {
		mask = 0;
		net = 0;
	}
	else {
		mask = mknet (argv[1], &net);
	}

	if (!inet_aton (argv[3], &srvr_addr)) {
		fprintf (stderr, "Invalid server address\n");
		rc = CMD_ERR_FAILED;
	}
	else if (!is_valid_srvr_addr (argv[3])) {
			fprintf (stderr, "Invalid server address\n");
			rc = CMD_ERR_FAILED;
	}
	else {
		srvr = srvr_addr.s_addr;
		become_root();	// Need this to ypset
		if (nfxnis_addservice (srv_type, net, mask, srvr, argv[2])) {
			if (errno == EEXIST) {
				fprintf (stderr, 
				         "Name service already defined for that network\n");
			}
			else {
				perror ("Error encountered updating nameservice configuration");
			}
			rc =  CMD_ERR_FAILED;
		}
		unbecome_root();
	}
	return rc;
}

DEFUN	(delete,
		delete_cmd,
		"nameservice delete (default|A.B.C.D/M)",
		NAMSVCSTR
		"Delete the service\n"
		"For the default\n"
		"on this specific network\n")
{
	in_addr_t net;
	in_addr_t mask;
	int rc = CMD_SUCCESS;

	if (*argv[0] == 'd') {
		mask = 0;
		net = 0;
	}
	else {
		mask = mknet (argv[0], &net);
	}

	become_root();
	if (nfxnis_delservice (net, mask)) {
		rc = CMD_ERR_FAILED;
	}
	unbecome_root();

	return rc;
}

#if defined(ENABLE_DISABLE)
DEFUN	(disable,
		disable_cmd,
		"nameservice disable A.B.C.D/M",
		NAMSVCSTR
		"Disable this service\n"
		"on this network\n")
{
	return CMD_SUCCESS;
}

DEFUN	(enable,
		enable_cmd,
		"nameservice enable A.B.C.D/M",
		NAMSVCSTR
		"Enable this service\n"
		"on this network (must have been added already)\n")
{
	return CMD_SUCCESS;
}
#endif

DEFUN	(show,
		show_cmd,
		"nameservice show (all|nis|dns)",
		NAMSVCSTR
		"Show this service\n"
		"All services\n"
		"Just NIS servers\n"
		"Just DNS servers\n")
{
	nfxnis_cfg_t **services;
	int i, j;
	nfxnis_serv_t want = 0;
	struct in_addr inaddr;

	services = nfxnis_get_cfg (0, 0);
	
	if (services) {
		if (argc && (*argv[0] == 'a')) {
			argc = 0;
		}
		else if (argc && (*argv[0] =='n')) {
			want = nis_name;
		}
		else {
			want = dns_name;
		}

		fprintf (stdout, 
		         "Network         Netmask         Type Domain   Servers\n");

		for (i = 0; services[i]; i++) {

			if (!argc || (services[i]->type == want)) {
				if (!services[i]->mask) {
					fprintf (stdout, "%-31s", "default");
				}
				else {
					inaddr.s_addr = services[i]->net;
					fprintf (stdout, "%-15s", inet_ntoa (inaddr));
					inaddr.s_addr = services[i]->mask;
					fprintf (stdout, " %-15s", inet_ntoa (inaddr));
				}
				fprintf (stdout, " %s ", 
				         (services[i]->type == nis_name) ? "NIS" : "DNS");
				fprintf (stdout, " %s", services[i]->domain);

				for (j = 0; j < services[i]->num_srvrs; j++) {
					inaddr.s_addr = services[i]->srvrs[j];
					fprintf (stdout, " %s", inet_ntoa (inaddr));
				}
				fprintf (stdout, "\n");
			}

			free (services[i]->domain);
			free (services[i]);
		}
		free (services);
	}
	else {
		fprintf (stdout, "No name services defined\n");
	}

	return CMD_SUCCESS;
}
DEFUN	(query,
		query_cmd,
		"nameservice query A.B.C.D HOSTNAME",
		NAMSVCSTR
		"Query name service\n"
		"For this network\n"
		"About this hostname\n")
{
	in_addr_t addr;
	struct in_addr inaddr;

	addr = nfxsh_nis_hostAddr (argv[0], argv[1]);

	if (!addr) {
		fprintf (stdout, "No hostname known to the nameserver on that net\n");
	}
	else {
		inaddr.s_addr = addr;
		fprintf (stdout, "It said: %s\n", inet_ntoa (inaddr));
	}
	return CMD_SUCCESS;
}

void
install_nameservice (priv_t myPriv, enum node_type node)
{
/*
	Organizing this switch statment such that the higher priority commands
	are first allows us to fall through and pick up the lower priority
	ones at no extra cost.
 */
	switch (myPriv) {
		case diagPriv:
			install_element (node, &query_cmd);
		case fullPriv:
#if defined(ENABLE_DISABLE)
			install_element (node, &disable_cmd);
#endif
			install_element (node, &delete_cmd);
		case changePriv:
		case addPriv:
			install_element (node, &add_cmd);
#if defined(ENABLE_DISABLE)
			install_element (node, &enable_cmd);
#endif
		case viewPriv:
/*
	These commands are available in all modes
 */
			install_element (node, &show_cmd);
	}
}
