/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:
 *
 * RCS:  $Id: cmd_ndmp.c,v 1.15 2002/12/04 22:31:22 nitinj Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:
 *
 * Created by:	Greg Seibert
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#include <sys/types.h>
#include <pwd.h>
#include <readline/history.h>
#include <grp.h>
#include <unistd.h>
#include "zebra.h"
#include <exports/opt-api.h>
#include <exports/anpssc-api.h>
#include "command.h"
#include "log.h"
#include "vtysh.h"
#include "memory.h"

#include "nfxsh.h"

#define NDMP_CMD_STR "NDMP commands\n"
#define NDMP_SHOW_STR "Display NDMP Info\n"

static int has_digits(char *str);

static int has_digits(char *str)
{
    int  cnt;

    if (str == NULL)
        return 0;

    for (cnt=0; cnt < strlen(str); cnt++)
        if (!isdigit(str[cnt]))
            return 0;
    return 1;
}

/*-----------------------------------------------------------------
 * Name: ndmp_disable
 *
 * Description: NDMP disable 
 *
 * Created by:  Anh Vu
 *
 * Date Created:  03/18/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (ndmp_disable,
         ndmp_disable_cmd,
         "ndmp disable",
         NDMP_CMD_STR
         "Disable the feature\n")
{

    return execute_command("ndmp_ui",
                           1,
                           "disable");
}

/*-----------------------------------------------------------------
 * Name: ndmp_enable
 *
 * Description: NDMP enable 
 *
 * Created by:  Anh Vu
 *
 * Date Created:  03/18/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (ndmp_enable,
         ndmp_enable_cmd,
         "ndmp enable",
         NDMP_CMD_STR
         "Enable the feature\n")
{

    return execute_command("ndmp_ui",
                           1,
                           "enable");
}

/*-----------------------------------------------------------------
 * Name: ndmp_setver
 *
 * Description: Set NDMP version 
 *
 * Created by:  Anh Vu
 *
 * Date Created:  05/06/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (ndmp_setver,
         ndmp_setver_cmd,
         "ndmp set version NUM",
         NDMP_CMD_STR
         "Set a variable\n"
		 "highest NDMP version supported.\n"
		 "Version number {2|3}\n")
{

    if (!has_digits(argv[0]))
    {
        printf ("Error : Version number is not a number\n");
        return CMD_ERR_FAILED;
    }
    return execute_command("ndmp_ui",
                           2,
                           "set_version",
						   argv[0]);
}


/*-----------------------------------------------------------------
 * Name: ndmp_auth_client_add
 *
 * Description: Add or register client username and password
 *				into the Registry Manager so we can 
 *				authenticate the client later.
 *
 * Created by:  Anh Vu
 *
 * Date Created:  03/18/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (ndmp_auth_client_add,
         ndmp_auth_client_add_cmd,
         "ndmp auth client add USERNAME",
         NDMP_CMD_STR
		 "For authentication purposes\n"
		 "Client account\n"
		 "Add new account\n"
         "Account name\n")
{

    return execute_command("ndmpauth",
                           2,
                           "add",
						   argv[0]);
}
/*-----------------------------------------------------------------
 * Name: ndmp_auth_client_del
 *
 * Description: Remover client username and password
 *				from the Registry Manager.
 *
 * Created by:  Anh Vu
 *
 * Date Created:  03/18/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (ndmp_auth_client_del,
         ndmp_auth_client_del_cmd,
         "ndmp auth client del USERNAME",
         NDMP_CMD_STR
		 "For authentication purposes\n"
		 "Client account\n"
		 "Remove existing account\n"
         "Account name\n")
{

    return execute_command("ndmpauth",
                           2,
                           "del",
						   argv[0]);
}
/*-----------------------------------------------------------------
 * Name: ndmp_auth_client_mod
 *
 * Description: Modify client username and password
 *				in the Registry Manager.
 *
 * Created by:  Anh Vu
 *
 * Date Created:  03/18/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (ndmp_auth_client_mod,
         ndmp_auth_client_mod_cmd,
         "ndmp auth client mod USERNAME",
         NDMP_CMD_STR
		 "For authentication purposes\n"
		 "Client account\n"
		 "Modify existing account\n"
         "Account name\n")
{

    return execute_command("ndmpauth",
                           2,
                           "mod",
						   argv[0]);
}
/*-----------------------------------------------------------------
 * Name: ndmp_auth_server_add
 *
 * Description: Register our (the server) username and password
 *				so the client can authenticate us later.
 *
 * Created by:  Anh Vu
 *
 * Date Created:  03/18/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (ndmp_auth_server_add,
         ndmp_auth_server_add_cmd,
         "ndmp auth server add USERNAME",
         NDMP_CMD_STR
		 "For authentication purposes\n"
		 "Server account\n"
		 "Add new account\n"
         "Account name\n")
{

	char8               reqbuf[1024], rspbuf[1024], id[50];
	opt_msgHdr_t        *opt_hdr, *rsp = (opt_msgHdr_t *) rspbuf;
	opt_getOptionRsp_t  *getrsp;
	int32               rc, buflen;
	struct timeval  	to;


	bzero(reqbuf, sizeof(reqbuf));
	opt_hdr = (opt_msgHdr_t *) &reqbuf;
	opt_hdr->msgType = OPT_GET_OPTION_REQ;
	strcpy( reqbuf + sizeof(opt_msgHdr_t), "options.ndmp.auth.server.*");
	buflen = sizeof(opt_msgHdr_t) + sizeof(opt_getOptionRsp_t);

    to.tv_sec = 30;
    to.tv_usec = 0;

    rc = sendAgileMsg(-1, (void *) reqbuf, buflen,
                        "regm", 0, 0, 0, &to,  /* to which applic, cpu, slot, plane, timeout */
                        rspbuf, sizeof(rspbuf));
    if (rc <= 0) {
		printf("NDMP Request failed\n");
        return CMD_ERR_FAILED;
    }

	getrsp = ((opt_getOptionRsp_t *)(rsp + 1));
	if (getrsp->nbEntry >= 1)
	{
		strcpy(id, rindex((char *)getrsp->entry[0].option, '.') + 1);
		printf("There is already a Server entry in the Registry Manager:  %s\n", id);
		printf("Try delete then add a new entry.\n");
		return CMD_SUCCESS;
	}

	return execute_command("ndmpauth",
                           3,
                           "add",
                           "server",
						   argv[0]);
}
/*-----------------------------------------------------------------
 * Name: ndmp_auth_server_del
 *
 * Description: Remove our server username and password
 *				from the registry manager.
 *
 * Created by:  Anh Vu
 *
 * Date Created:  03/18/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (ndmp_auth_server_del,
         ndmp_auth_server_del_cmd,
         "ndmp auth server del USERNAME",
         NDMP_CMD_STR
		 "For authentication purposes\n"
		 "Server account\n"
		 "Remove existing account\n"
         "Account name\n")
{

    return execute_command("ndmpauth",
                           3,
                           "del",
                           "server",
						   argv[0]);
}
/*-----------------------------------------------------------------
 * Name: ndmp_auth_server_mod
 *
 * Description: Modify our server username and password
 *				in the registry manager.
 *
 * Created by:  Anh Vu
 *
 * Date Created:  03/18/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (ndmp_auth_server_mod,
         ndmp_auth_server_mod_cmd,
         "ndmp auth server mod USERNAME",
         NDMP_CMD_STR
		 "For authentication purposes\n"
		 "Server account\n"
		 "Modify existing account\n"
         "Account name\n")
{

    return execute_command("ndmpauth",
                           3,
                           "mod",
                           "server",
						   argv[0]);
}

/*-----------------------------------------------------------------
 * Name: ndmp_show_auth_client
 *
 * Description: Show NDMP authenticated client usernames.
 *
 * Created by:  Anh Vu
 *
 * Date Created:  05/09/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (ndmp_show_auth_client,
         ndmp_show_auth_client_cmd,
         "ndmp show auth client",
         NDMP_CMD_STR
         NDMP_SHOW_STR
		 "Authenticated username\n"
		 "Of the clients\n")
{

    return execute_command("ndmp_ui",
                           1,
                           "show_auth_client");
}
/*-----------------------------------------------------------------
 * Name: ndmp_show_auth_srvr
 *
 * Description: Show NDMP authenticated server username.
 *
 * Created by:  Anh Vu
 *
 * Date Created:  05/09/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (ndmp_show_auth_srvr,
         ndmp_show_auth_srvr_cmd,
         "ndmp show auth server",
         NDMP_CMD_STR
         NDMP_SHOW_STR
		 "Authenticated username\n"
		 "Of the server\n")
{

    return execute_command("ndmp_ui",
                           1,
                           "show_auth_server");
}

/*-----------------------------------------------------------------
 * Name: ndmp_show_state
 *
 * Description: Show NDMP state: enable or disable.  
 *
 * Created by:  Anh Vu
 *
 * Date Created:  04/03/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (ndmp_show_state,
         ndmp_show_state_cmd,
         "ndmp show state",
         NDMP_CMD_STR
         NDMP_SHOW_STR
         "Whether the feature is ENABLED or DISABLED.\n")
{

    return execute_command("ndmp_ui",
                           1,
                           "show_state");
}


/*-----------------------------------------------------------------
 * Name: ndmp_show_active
 *
 * Description: NDMP enable
 *
 * Created by:  Anh Vu
 *
 * Date Created:  03/18/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (ndmp_show_active,
         ndmp_show_active_cmd,
         "ndmp show active",
         NDMP_CMD_STR
         NDMP_SHOW_STR
		 "Currently active sessions Ids.\n")
{

    return execute_command("ndmp_ui",
                           1,
                           "show_active");
}


/*-----------------------------------------------------------------
 * Name: ndmp_show_stat
 *
 * Description: NDMP Show status of one or all sessions 
 *
 * Created by:  Anh Vu
 *
 * Date Created:  03/18/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (ndmp_show_stat_all,
         ndmp_show_stat_all_cmd,
		 "ndmp show status -a",
         NDMP_CMD_STR
         NDMP_SHOW_STR
         "Show NDMP services status for active session(s)\n"
		 "Show all sessions\n")
{

    return execute_command("ndmp_ui",
                           2,
                           "show_status",
						   "all");
}
DEFUN   (ndmp_show_stat_one,
         ndmp_show_stat_one_cmd,
         "ndmp show status -s SESSION-ID",
         NDMP_CMD_STR
         NDMP_SHOW_STR
         "Show NDMP services status for active session(s)\n"
         "Show for a session-id\n")
{
    if (!has_digits(argv[0]))
    {
        printf ("Error : Session-Id is not a number\n");
        return CMD_ERR_FAILED;
    }
    return execute_command("ndmp_ui",
                           2,
                           "show_status",
                           argv[0]);
}

/*-----------------------------------------------------------------
 * Name: ndmp_show_detail
 *
 * Description: NDMP show DETAILED status of one or all sessions 
 *
 * Created by:  Anh Vu
 *
 * Date Created:  03/18/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (ndmp_show_detail_all,
         ndmp_show_detail_all_cmd,
		 "ndmp show detail -a",
         NDMP_CMD_STR
         NDMP_SHOW_STR
         "Show NDMP services detailed status for active session(s)\n"
		 "Show all sessions\n")
{

    return execute_command("ndmp_ui",
                           2,
                           "show_detail",
						   "all");
}
DEFUN   (ndmp_show_detail_one,
         ndmp_show_detail_one_cmd,
         "ndmp show detail -s SESSION-ID",
         NDMP_CMD_STR
         NDMP_SHOW_STR
         "Show NDMP services detailed status for active session(s)\n"
         "Show for a session-id\n")
{

    if (!has_digits(argv[0]))
    {
        printf ("Error : Session-Id is not a number\n");
        return CMD_ERR_FAILED;
    }
    return execute_command("ndmp_ui",
                           2,
                           "show_detail",
                           argv[0]);
}

DEFUN   (ndmp_show_log,
         ndmp_show_log_cmd,
         "ndmp show log [NAME]",
         NDMP_CMD_STR
         NDMP_SHOW_STR
         "Show list of dump files\n"
         "Show details of dump file\n")
{
    char fileName[1024];

    if (argc == 0)
        return (execute_command("/bin/ls", 1, NFX_DUMPDATE_DIR));
    else if (argc == 1)
    {
        bzero(fileName, 1024);
        sprintf(fileName, "%s%s", NFX_DUMPDATE_DIR, argv[0]);
        return (execute_command("/usr/bin/more", 1, fileName));
    }
    return CMD_ERR_FAILED;
}

/*-----------------------------------------------------------------
 * Name: ndmp_del_sess
 *
 * Description: NDMP Show status of one or all sessions 
 *
 * Created by:  Anh Vu
 *
 * Date Created:  03/18/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (ndmp_del_sess_all,
         ndmp_del_sess_all_cmd,
		 "ndmp delete session -a",
         NDMP_CMD_STR
         "Kill NDMP session\n"
		 "Kill all sessions\n")
{

    return execute_command("ndmp_ui",
                           2,
                           "delete_session",
						   "all");
}
DEFUN   (ndmp_del_sess_one,
         ndmp_del_sess_one_cmd,
         "ndmp delete session -s SESSION-ID",
         NDMP_CMD_STR
         "Kill NDMP session\n"
         "Kill a session-id\n")
{

    if (!has_digits(argv[0]))
    {
        printf ("Error : Session-Id is not a number\n");
        return CMD_ERR_FAILED;
    }
    return execute_command("ndmp_ui",
                           2,
                           "delete_session",
                           argv[0]);
}
/*-----------------------------------------------------------------
 * Name: ndmp_trace
 *
 * Description: NDMP trace enable/disable 
 *
 * Created by:  Anh Vu
 *
 * Date Created:  10/10/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (ndmp_trace_all,
         ndmp_trace_all_cmd,
         "ndmp set trace flag FLAG -a",
         NDMP_CMD_STR
         "Set a variable\n"
		 "debug traces of messages between client and server\n"
		 "flag as on or off\n"
		 "{ON|OFF}\n"
		 "For all sessions\n")
{

    return execute_command("ndmp_ui",
                           3,
                           "trace",
						   argv[0],
						   "all");
}
DEFUN   (ndmp_trace_one,
         ndmp_trace_one_cmd,
         "ndmp set trace flag FLAG -s SESSION-ID",
         NDMP_CMD_STR
         "Set a variable\n"
		 "debug traces of messages between client and server\n"
		 "flag as on or off\n"
		 "{ON|OFF}\n"
		 "For a particular active session\n")
{

    if (!has_digits(argv[1]))
    {
        printf ("Error : Session-Id is not a number\n");
        return CMD_ERR_FAILED;
    }
    return execute_command("ndmp_ui",
                           3,
                           "trace",
						   argv[0],
						   argv[1]);
}
DEFUN   (ndmp_trace_file,
         ndmp_trace_file_cmd,
         "ndmp set trace file NUM_DATES",
         NDMP_CMD_STR
         "Set a variable\n"
		 "debug traces of messages between client and server\n"
		 "retain the latest number_of_dates of log files\n"
		 "<1-10>, default is 1.\n")
{

    return execute_command("ndmp_ui",
                           2,
                           "tracefile",
						   argv[0]);
}
DEFUN   (ndmp_restore_path,
         ndmp_restore_path_cmd,
         "ndmp set restore path (ON|OFF)",
         NDMP_CMD_STR
         "Set a variable\n"
		 "Restore full path or eliminate common path\n"
		 "Restore full path or eliminate common path\n"
		 "{ON|OFF} ON - restore full path, OFF - remove most common path. Default is ON\n"
		 "{ON|OFF} ON - restore full path, OFF - remove most common path. Default is ON\n")
{

    return execute_command("ndmp_ui",
                           2,
                           "restore_path",
						   argv[0]);
}
/*-----------------------------------------------------------------
 * Name: ndmp_sync_snap_shot
 *
 * Description: NDMP trace enable/disable 
 *
 * Created by:  Nitin Jain
 *
 * Date Created:  10/11/02
 *
 *-----------------------------------------------------------------
 */
DEFUN(ndmp_sync_snap_shot,
      ndmp_sync_snap_shot_cmd,
      "ndmp sync snapshot",
      NDMP_CMD_STR
      "sync NDMP snapshots with filesystem\n"
      "sync NDMP snapshots with filesystem\n")
{

    /* Clean and sync the snapshots */

    return execute_command("nscd",
                           0);
}

#define NDMP_SCH_CLEAN "[-m MINUTE] [-h HOUR] [-d DATE] [-M MONTH] [-D DAY]\n"

DEFUN(ndmp_schedule_snap_clean,
      ndmp_schedule_snap_clean_cmd,
      "ndmp schedule create [OPT] [OPT] [OPT] [OPT] [OPT] [OPT] [OPT] [OPT] [OPT] [OPT]",
      "ndmp schedule snapshot\n"
      "ndmp schedule snapshot\n"
      "ndmp schedule snapshot create\n"
      NDMP_SCH_CLEAN
      NDMP_SCH_CLEAN
      NDMP_SCH_CLEAN
      NDMP_SCH_CLEAN
      NDMP_SCH_CLEAN
      NDMP_SCH_CLEAN
      NDMP_SCH_CLEAN
      NDMP_SCH_CLEAN
      NDMP_SCH_CLEAN
      NDMP_SCH_CLEAN)
{
    /* Add an entry to the corn tab file */
    return execute_vararg_command("ndmp_ui", argc, "cleanup_sch", argv);
}

DEFUN(ndmp_schedule_snap_clean_show,
      ndmp_schedule_snap_clean_show_cmd,
      "ndmp show schedule",
      "ndmp schedule snapshot\n"
      "ndmp schedule snapshot show\n"
      "ndmp schedule snapshot show\n")
{

    return execute_command("ndmp_ui",
                           1,
                           "cleanup_sch_show");
}
DEFUN(ndmp_schedule_snap_clean_del,
      ndmp_schedule_snap_clean_del_cmd,
      "ndmp schedule delete",
      "ndmp schedule snapshot\n"
      "ndmp schedule snapshot\n"
      "ndmp schedule snapshot delete\n")
{
    /* delete the schedule entry from corn tab file. */
    return execute_command("ndmp_ui",
                           1,
                           "cleanup_sch_del");
}


/*-----------------------------------------------------------------
 * Name:install_ndmp
 *
 * Description: Install the ndmp commands menu items
 *
 * Created by:  Anh Vu
 *
 * Date Created: 03/18/02
 *
 *-----------------------------------------------------------------
 */
void
install_ndmp (priv_t myPriv, enum node_type node)
{
	/* From Gregg:
     * Organize this switch statment such that the higher priority commands
     * are first allowing us to fall through and pick up the lower priority
     * ones at no extra cost.
 	 */
    switch (myPriv) {
        case diagPriv:
            install_element (node, &ndmp_show_log_cmd);
        case fullPriv:
            install_element (node, &ndmp_trace_all_cmd);
            install_element (node, &ndmp_trace_one_cmd);
            install_element (node, &ndmp_trace_file_cmd);
            install_element (node, &ndmp_sync_snap_shot_cmd);
            install_element (node, &ndmp_schedule_snap_clean_del_cmd);
            install_element (node, &ndmp_disable_cmd);
            install_element (node, &ndmp_setver_cmd);
            install_element (node, &ndmp_del_sess_all_cmd);
            install_element (node, &ndmp_del_sess_one_cmd);
            install_element (node, &ndmp_auth_client_add_cmd);
            install_element (node, &ndmp_auth_client_del_cmd);
            install_element (node, &ndmp_auth_client_mod_cmd);
            install_element (node, &ndmp_auth_server_add_cmd);
            install_element (node, &ndmp_auth_server_del_cmd);
            install_element (node, &ndmp_auth_server_mod_cmd);
            install_element (node, &ndmp_schedule_snap_clean_cmd);
            install_element (node, &ndmp_restore_path_cmd);
        case changePriv:
        case addPriv:
            install_element (node, &ndmp_enable_cmd);
        case viewPriv:
			/*
    		 * These commands are available in all modes
 			 */
            install_element (node, &ndmp_schedule_snap_clean_show_cmd);
            install_element (node, &ndmp_show_auth_client_cmd);
            install_element (node, &ndmp_show_auth_srvr_cmd);
            install_element (node, &ndmp_show_active_cmd);
            install_element (node, &ndmp_show_state_cmd);
            install_element (node, &ndmp_show_stat_all_cmd);
            install_element (node, &ndmp_show_stat_one_cmd);
            install_element (node, &ndmp_show_detail_all_cmd);
            install_element (node, &ndmp_show_detail_one_cmd);
    }
}
