/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:
 *
 * RCS:  $Id: cmd_quota.c,v 1.16 2002/10/02 00:52:45 laurentb Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:
 *
 * Created by:	Greg Seibert
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#include <sys/types.h>
#include <nfx-types.h>
#include <pwd.h>
#include <readline/history.h>
#include <grp.h>
#include <unistd.h>

#include "zebra.h"
#include "command.h"
#include "log.h"
#include "vtysh.h"
#include "memory.h"

#include "nfxsh.h"
#include "exports/anpssc-api.h"
#include "exports/evm-api.h"
#include "exports/fs-api.h"

#include "cmd_fsutil.h"

extern int optind;

#define QUOTA_STR   "Quota Administration Commands\n"
#define ID_MAXLEN   11
#define DEF_MAXLEN  9

#define FSCMD       "fscmd"
#define SET_UG_QUOTA    "set_ug_quota"
#define GET_UG_QUOTA    "get_ug_quota"
#define SET_DEFAULT_QUOTA   "set_def_quota"
#define SET_FILESET_QUOTA   "set_fset_quota"
#define GET_FILESET_QUOTA   "get_fset_quota"
#define GET_DEFAULT_QUOTA   "get_fset_quota"
#define REPORT_UG_QUOTA     "report_ug_quota"
#define REPORT_FSET_QUOTA   "report_fset_quota"
#define QUOTA_ON    "quotaon"
#define QUOTA_OFF    "quotaoff"


#define USER            "user"
#define GROUP           "group"
#define DEFAULT         "default"
#define AGGR            "aggr"
#define DEFAULT_USER        "def_user"
#define DEFAULT_GROUP       "def_group"
#define DEFAULT_AGGREGATE   "def_aggr"


#define SET_USER_QUOTA_CMD                                    \
"quota set user VOL FILESET_PATH UID FILE_SOFT_LIMIT " \
" FILE_HARD_LIMIT FILE_TIME_LIMIT SIZE_SOFT_LIMIT SIZE_HARD_LIMIT "\
"SIZE_TIME_LIMIT\n"

#define SET_USER_QUOTA_HELP                                \
"set quota limits for users, groups or filesets\n"\
"set quota limits for a specified user\n"\
"Specify the volume/filesystem for which quota are being set\n"\
"Specify the fileset for which the quota are being set\n"\
"Specify uid of the user\n"\
"Specify number of files soft limit (0: unlimited)\n"\
"Specify number of files hard limit (0: unlimited)\n"\
"Specify time limit in hours for which files used can exceed the soft limit"\
" (0: unlimited)\n"\
"Specify size soft limit in MB (0: unlimited)\n"\
"Specify size hard limit in MB (0: unlimited)\n"\
"Specify time limit in hours for which size can exceed the soft limit"\
" (0: unlimited)\n"

#define SET_USER_QUOTA_2_DEFAULT_CMD                                    \
"quota set user VOL FILESET_PATH UID -d\n"

#define SET_USER_QUOTA_2_DEFAULT_HELP                                \
"set quota limits for users, groups or filesets\n"\
"set quota limits for a specified user\n"\
"Specify the volume/filesystem for which quota are being set\n"\
"Specify the fileset for which the quota are being set\n"\
"Specify uid of the user\n"\
"To use fileset default user quota\n"

#define SET_GROUP_QUOTA_CMD                                     \
"quota set group VOL FILESET_PATH GID FILE_SOFT_LIMIT " \
" FILE_HARD_LIMIT FILE_TIME_LIMIT SIZE_SOFT_LIMIT SIZE_HARD_LIMIT "\
"SIZE_TIME_LIMIT\n"

#define SET_GROUP_QUOTA_HELP                                 \
"set quota limits for users, groups or filesets\n"\
"set quota limits for a specified group\n"\
"Specify the volume/filesystem for which quota are being set\n"\
"Specify the fileset for which the quota are being set\n"\
"Specify gid of the group\n"\
"Specify number of files soft limit (0: unlimited)\n"\
"Specify number of files hard limit (0: unlimited)\n"\
"Specify time limit in hours for which files used can exceed the soft limit"\
" (0: unlimited)\n"\
"Specify size soft limit in MB (0: unlimited)\n"\
"Specify size hard limit in MB (0: unlimited)\n"\
"Specify time limit in hours for which size can exceed the soft limit"\
" (0: unlimited)\n"

#define SET_GROUP_QUOTA_2_DEFAULT_CMD                                     \
"quota set group VOL FILESET_PATH GID -d\n" 

#define SET_GROUP_QUOTA_2_DEFAULT_HELP                                 \
"set quota limits for users, groups or filesets\n"\
"set quota limits for a specified group\n"\
"Specify the volume/filesystem for which quota are being set\n"\
"Specify the fileset for which the quota are being set\n"\
"Specify gid of the group\n"\
"To use fileset default group quota\n"

#define SET_FILESET_QUOTA_CMD                                     \
"quota set fileset VOL FILESET_PATH FILE_SOFT_LIMIT " \
" FILE_HARD_LIMIT FILE_TIME_LIMIT SIZE_SOFT_LIMIT SIZE_HARD_LIMIT "\
"SIZE_TIME_LIMIT\n"

#define SET_FILESET_QUOTA_HELP                                 \
"set quota limits for users, groups or filesets\n"\
"set quota limits for the specified fileset\n"\
"Specify the volume/filesystem for which quota are being set\n"\
"Specify the fileset for which the quota are being set\n"\
"Specify number of files soft limit (0: unlimited)\n"\
"Specify number of files hard limit (0: unlimited)\n"\
"Specify time limit in hours for which files used can exceed the soft limit"\
" (0: unlimited)\n"\
"Specify size soft limit in MB (0: unlimited)\n"\
"Specify size hard limit in MB (0: unlimited)\n"\
"Specify time limit in hours for which size can exceed the soft limit"\
" (0: unlimited)\n"

#define SET_FILESET_QUOTA_2_DEFAULT_CMD                             \
"quota set fileset VOL FILESET_PATH -d"
        
#define SET_FILESET_QUOTA_2_DEFAULT_HELP                            \
"set quota limits for users, groups or filesets\n"\
"set quota limits for the specified fileset\n"\
"Specify the volume/filesystem for which quota are being set\n"\
"Specify the fileset for which the quota are being set\n"\
"To use default quota from the parent fileset\n"

#define SET_DEFAULT_QUOTA_CMD                                     \
"quota set default VOL FILESET_PATH (-u|-g|-f) FILE_SOFT_LIMIT " \
" FILE_HARD_LIMIT FILE_TIME_LIMIT SIZE_SOFT_LIMIT SIZE_HARD_LIMIT "\
"SIZE_TIME_LIMIT\n"

#define SET_DEFAULT_QUOTA_HELP                                      \
"set quota limits for users, groups or filesets\n"\
"Specify the default/prototype quotas\n"\
"Specify the volume/filesystem for which quota are being set\n"\
"Specify the fileset for which the quota are being set\n"\
"To set user default quotas for this fileset\n"\
"To set group default quotas for this fileset\n"\
"To set default quotas for filesets created in this fileset\n"\
"Specify number of files soft limit (0: unlimited)\n"\
"Specify number of files hard limit (0: unlimited)\n"\
"Specify time limit in hours for which files used can exceed the soft limit"\
" (0: unlimited)\n"\
"Specify size soft limit in MB (0: unlimited)\n"\
"Specify size hard limit in MB (0: unlimited)\n"\
"Specify time limit in hours for which size can exceed the soft limit"\
" (0: unlimited)\n"

#define QUOTA_ON_CMD "quota on VOL FILESET_PATH"

#define QUOTA_ON_HELP                                               \
"Turn quota on for the given fileset and volume\n"\
"Specify the volume/filesystem for which quota are being turned on\n"\
"Specify the fileset for which the quota are being turned on\n"

#define QUOTA_OFF_CMD "quota off VOL FILESET_PATH"

#define QUOTA_OFF_HELP                                               \
"Turn quota off for the given fileset and volume\n"\
"Specify the volume/filesystem for which quota are being turned off\n"\
"Specify the fileset for which the quota are being turned off\n"

#define SHOW_USER_QUOTA_CMD                                         \
"quota show user VOL FILESET_PATH UID"

#define SHOW_USER_QUOTA_HELP                                        \
"show quota limits and usage\n"\
"show the limits and usage for a specified user\n"\
"Specify the volume/filesystem for which quota is to be provided\n" \
"Specify the fileset for which the quota is to be provided\n"\
"Specify uid of the user\n"

#define SHOW_GROUP_QUOTA_CMD                                         \
"quota show group VOL FILESET_PATH GID"

#define SHOW_GROUP_QUOTA_HELP                                           \
"show quota limits and usage\n"\
"show the limits and usage for a specified group\n"\
"Specify the volume/filesystem for which quota is to be providedt\n"\
"Specify the fileset for which the quota is to be provided\n"\
"Specify gid of the group\n"

#define SHOW_FILESET_QUOTA_CMD                                         \
"quota show fileset VOL FILESET_PATH"

#define SHOW_FILESET_QUOTA_HELP                                         \
"show quota limits and usage\n"\
"show the limits and usage for a specified fileset\n"\
"Specify the volume/filesystem for which quota is to be provided\n"\
"Specify the fileset for which the quota is to be provided\n"

#define REPORT_USER_QUOTA_CMD                       \
"quota report user VOL FILESET"

#define REPORT_USER_QUOTA_HELP                  \
"report the limits and usage for users or groups or filesets\n"\
"report limits and usage for all the users for a specified fileset\n"\
"Specify the volume/filesystem for which quota is to be provided\n"\
"Specify the fileset for which the quota is to be provided\n"

#define REPORT_GROUP_QUOTA_CMD                       \
"quota report group VOL FILESET"

#define REPORT_GROUP_QUOTA_HELP                  \
"report the limits and usage for users or groups or filesets\n"\
"report limits and usage for all the groups for a specified fileset\n"\
"Specify the volume/filesystem for which quota is to be provided\n"\
"Specify the fileset for which the quota is to be provided\n"

#define REPORT_FILESET_QUOTA_CMD                       \
"quota report fileset VOL"

#define REPORT_FILESET_QUOTA_HELP                  \
"report the limits and usage for users or groups or filesets\n"\
"report limits and usage for all the specified fileset\n"\
"Specify the volume/filesystem for which quota is to be provided\n"

/*
 * Set user quota
 */

#if 0
DEFUN(set_user_quota, set_user_quota_cmd, SET_USER_QUOTA_CMD, 
      QUOTA_STR
      SET_USER_QUOTA_HELP)
{
    int     rc;
    char    *vol = argv[0];
    char    *fset = argv[1];
    char    *id = argv[2];
    char    *is = argv[3];
    char    *ih = argv[4];
    char    *it = argv[5];
    char    *bs = argv[6];
    char    *bh = argv[7];
    char    *bt = argv[8];

    
    /*
     * Invoke fscmd
     */

    rc = execute_command(FSCMD,
                         12,            // argc for this command
                         SET_UG_QUOTA,
                         vol,       // volume id
                         fset,       // fileset
                         USER,          // this is an user related quota
                         id,            // uid
                         is,        // inode soft limit
                         ih,       // inode hard limit
                         it,        // inode time limit
                         bs,        // block soft limit
                         bh,        // block hard limit
                         bt,        // block time limit
                         NULL
                        );
                         
    if (rc) {
            return CMD_ERR_FAILED;
    }
    return CMD_SUCCESS;
}

/*
 * Set user quota
 */

DEFUN(set_user_quota_2_default, set_user_quota_2_default_cmd, 
      SET_USER_QUOTA_2_DEFAULT_CMD, 
      QUOTA_STR
      SET_USER_QUOTA_2_DEFAULT_HELP)
{
    int     rc;
    char    *vol = argv[0];
    char    *fset = argv[1];
    char    *id = argv[2];
    
    /*
     * Invoke fscmd
     */

    rc = execute_command(FSCMD,
                         7,             // argc for this command
                         SET_UG_QUOTA,
                         vol,           // volume id
                         fset,          // fileset
                         USER,          // this is an user related quota
                         id,            // uid
                         "-d",
                         NULL
                        );
                         
    if (rc) {
            return CMD_ERR_FAILED;
    }
    return CMD_SUCCESS;
}

/*
 * Set group quota
 */

DEFUN(set_group_quota, set_group_quota_cmd, SET_GROUP_QUOTA_CMD, 
      QUOTA_STR
      SET_GROUP_QUOTA_HELP)
{
    int     rc;
    char    *vol = argv[0];
    char    *fset = argv[1];
    char    *id = argv[2];
    char    *is = argv[3];
    char    *ih = argv[4];
    char    *it = argv[5];
    char    *bs = argv[6];
    char    *bh = argv[7];
    char    *bt = argv[8];
    
    /*
     * Invoke fscmd
     */

    rc = execute_command(FSCMD,
                         12,        // argc for this command
                         SET_UG_QUOTA,
                         vol,       // volume id
                         fset,      // fileset
                         GROUP,     // this is an user related quota
                         id,        // uid
                         is,        // inode soft limit
                         ih,        // inode hard limit
                         it,        // inode time limit
                         bs,        // block soft limit
                         bh,        // block hard limit
                         bt,        // block time limit
                         NULL
                        );
                         
    if (rc) {
            return CMD_ERR_FAILED;
    }
    return CMD_SUCCESS;
}

/*
 * Set group quota to default
 */

DEFUN(set_group_quota_2_default, set_group_quota_2_default_cmd, 
      SET_GROUP_QUOTA_2_DEFAULT_CMD, 
      QUOTA_STR
      SET_GROUP_QUOTA_2_DEFAULT_HELP)
{
    int     rc;
    char    *vol = argv[0];
    char    *fset = argv[1];
    char    *id = argv[2];
    
    /*
     * Invoke fscmd
     */

    rc = execute_command(FSCMD,
                         7,         // argc for this command
                         SET_UG_QUOTA,
                         vol,       // volume id
                         fset,      // fileset
                         GROUP,     // this is an user related quota
                         id,        // uid
                         "-d",
                         NULL
                        );
                         
    if (rc) {
            return CMD_ERR_FAILED;
    }
    return CMD_SUCCESS;
}

/*
 * Set default quota
 */

DEFUN(set_default_quota, set_default_quota_cmd, SET_DEFAULT_QUOTA_CMD, 
      QUOTA_STR
      SET_DEFAULT_QUOTA_HELP)
{
    int     gFlag = 0, uFlag = 0, fFlag = 0, c;
    int     rc;
    char    buf[DEF_MAXLEN + 1];
    char    *vol = argv[0];
    char    *fset = argv[1];
    char    *is = argv[3];
    char    *ih = argv[4];
    char    *it = argv[5];
    char    *bs = argv[6];
    char    *bh = argv[7];
    char    *bt = argv[8];

    /*
     * parse the options
     */

	optind = 0;
    while ((c = getopt(argc, argv, "guf")) != -1) {

        switch(c) {

        case 'g':
            gFlag++;
            break;

        case 'u':
            uFlag++;
            break;
        case 'f':
            fFlag++;
            break;
        default:
            return CMD_ERR_AMBIGUOUS;
        }
    }

    if ((gFlag + uFlag + fFlag) != 1) {
        return CMD_ERR_AMBIGUOUS;
    }

    if (gFlag) {
        strcpy(buf, GROUP);
    } else if (uFlag) {
        strcpy(buf, USER);
    } else {
        strcpy(buf, AGGR);
    }


    /*
     * Invoke fscmd
     */

    rc = execute_command(FSCMD,
                         11,            // argc
                         SET_DEFAULT_QUOTA,
                         vol,       // volume id
                         fset,       // fileset
                         buf,           // default user or group or aggregate
                         is,        // inode soft limit
                         ih,        // inode hard limit
                         it,        // inode time limit
                         bs,        // block soft limit
                         bh,        // block hard limit
                         bt,        // block time limit
                         NULL
                        );
    if (rc) {
            return CMD_ERR_FAILED;
    }
    return CMD_SUCCESS;
}

/*
 * Set fileset quota
 */

DEFUN(set_fileset_quota, set_fileset_quota_cmd, SET_FILESET_QUOTA_CMD, 
      QUOTA_STR
      SET_FILESET_QUOTA_HELP)
{
    int     rc;


    /*
     * Invoke fscmd
     */

    rc = execute_command(FSCMD,
                         10,            // argc
                         SET_FILESET_QUOTA,
                         argv[0],       // volume id
                         argv[1],       // fileset
                         argv[2],       // inode soft limit
                         argv[3],       // inode hard limit
                         argv[4],       // inode time limit
                         argv[5],       // block soft limit
                         argv[6],       // block hard limit
                         argv[7],       // block time limit
                         NULL);
    if (rc) {
            return CMD_ERR_FAILED;
    }
    return CMD_SUCCESS;
}

/*
 * Set fileset quota to default
 */

DEFUN(  set_fileset_quota_2_default, set_fileset_quota_2_default_cmd, 
        SET_FILESET_QUOTA_2_DEFAULT_CMD, 
        QUOTA_STR
        SET_FILESET_QUOTA_2_DEFAULT_HELP)
{
    int     rc;

    /*
     * Invoke fscmd
     */

    rc = execute_command(FSCMD,
                         5,             // argc
                         SET_FILESET_QUOTA,
                         argv[0],       // volume id
                         argv[1],       // fileset
                         "-d",          // set to default option
                         NULL);
    if (rc) {
        return CMD_ERR_FAILED;
    }
    return CMD_SUCCESS;
}

/*
 * Enable quota
 */

DEFUN(quota_on, quota_on_cmd, QUOTA_ON_CMD,
      QUOTA_STR
      QUOTA_ON_HELP)
{
    int rc;
    /*
     * Invoke fscmd
     */

    rc = execute_command(FSCMD,
                         4,             // argc
                         QUOTA_ON,
                         argv[0],       // volume id
                         argv[1],        // fileset
                         NULL
                        );
                         
    if (rc) {
            return CMD_ERR_FAILED;
    }
    return CMD_SUCCESS;
}

/*
 * Disable quota
 */

DEFUN(quota_off, quota_off_cmd, QUOTA_OFF_CMD,
      QUOTA_STR
      QUOTA_OFF_HELP)
{
    int rc;
    /*
     * Invoke fscmd
     */

    rc = execute_command(FSCMD,
                         4,             // argc
                         QUOTA_OFF,
                         argv[0],       // volume id
                         argv[1],        // fileset
                         NULL
                        );
    if (rc) {
            return CMD_ERR_FAILED;
    }
    return CMD_SUCCESS;
}

/*
 * Show user quota
 */

DEFUN(show_user_quota, show_user_quota_cmd, SHOW_USER_QUOTA_CMD, 
      QUOTA_STR
      SHOW_USER_QUOTA_HELP)
{
    int     rc;
    char    *vol = argv[0];
    char    *fset = argv[1];
    char    *id = argv[2];

    /*
     * Invoke fscmd
     */

    rc = execute_command(FSCMD,
                         6,             // argc
                         GET_UG_QUOTA,
                         vol,       // volume id
                         fset,       // fileset
                         USER,
                         id,
                         NULL
                        );
    if (rc) {
            return CMD_ERR_FAILED;
    }
    return CMD_SUCCESS;
}

/*
 * Show group quota
 */

DEFUN(show_group_quota, show_group_quota_cmd, SHOW_GROUP_QUOTA_CMD, 
      QUOTA_STR
      SHOW_GROUP_QUOTA_HELP)
{
    int     rc;
    char    *vol = argv[0];
    char    *fset = argv[1];
    char    *id = argv[2];
    /*
     * Invoke fscmd
     */

    rc = execute_command(FSCMD,
                         6,             // argc
                         GET_UG_QUOTA,
                         vol,       // volume id
                         fset,       // fileset
                         GROUP,
                         id,
                         NULL
                        );
    if (rc) {
            return CMD_ERR_FAILED;
    }
    return CMD_SUCCESS;
}


/*
 * Show fileset quota
 */

DEFUN(show_fileset_quota, show_fileset_quota_cmd, SHOW_FILESET_QUOTA_CMD, 
      QUOTA_STR
      SHOW_FILESET_QUOTA_HELP)
{
    int     rc;

    /*
     * Invoke fscmd
     */

    rc = execute_command(FSCMD,
                         3,             // argc
                         GET_FILESET_QUOTA,
                         argv[0],       // volume id
                         argv[1],       // fileset
                         NULL
                        );

    if (rc) {
            return CMD_ERR_FAILED;
    }
    return CMD_SUCCESS;
}


DEFUN(report_user_quota, report_user_quota_cmd, REPORT_USER_QUOTA_CMD,
      QUOTA_STR
      REPORT_USER_QUOTA_HELP)
{
    int     rc;
    
    /*
     * Invoke fscmd
     */

    rc = execute_command(FSCMD,
                         4,             // argc
                         REPORT_UG_QUOTA,
                         argv[0],       // volume id
                         argv[1],       // fileset
                         USER,          // user
                         NULL
                        );
    if (rc) {
            return CMD_ERR_FAILED;
    }
    return CMD_SUCCESS;
}

/*
 * Report group quota for a fileset
 */

DEFUN(report_group_quota, report_group_quota_cmd, REPORT_GROUP_QUOTA_CMD,
      QUOTA_STR
      REPORT_GROUP_QUOTA_HELP)
{
    int     rc;
    
    /*
     * Invoke fscmd
     */

    rc = execute_command(FSCMD,
                         4,             // argc
                         REPORT_UG_QUOTA,
                         argv[0],       // volume id
                         argv[1],       // fileset
                         GROUP,         // group
                         NULL
                        );
    if (rc) {
            return CMD_ERR_FAILED;
    }
    return CMD_SUCCESS;
}

/*
 * Report group quota for a fileset
 */

DEFUN(report_fileset_quota, report_fileset_quota_cmd, REPORT_FILESET_QUOTA_CMD,
      QUOTA_STR
      REPORT_FILESET_QUOTA_HELP)
{
    int     rc;

    /*
     * Invoke fscmd
     */

    rc = execute_command(FSCMD,
                         2,             // argc
                         REPORT_FSET_QUOTA,
                         argv[0],       // volume id
                         NULL
                        );
    if (rc) {
            return CMD_ERR_FAILED;
    }
    return CMD_SUCCESS;
}

#endif
void
install_quota (priv_t myPriv, enum node_type node)
{
#if 0
    switch (myPriv) {
        case diagPriv:
        case fullPriv:
        case changePriv:
            install_element(node, &quota_on_cmd);
            install_element(node, &quota_off_cmd);
            install_element(node, &set_user_quota_cmd);
            install_element(node, &set_user_quota_2_default_cmd);
            install_element(node, &set_group_quota_cmd);
            install_element(node, &set_group_quota_2_default_cmd);
            install_element(node, &set_fileset_quota_cmd);
            install_element(node, &set_fileset_quota_2_default_cmd);
            install_element(node, &set_default_quota_cmd);
        case addPriv:
        case viewPriv:
            install_element(node, &show_user_quota_cmd);
            install_element(node, &show_group_quota_cmd);
            install_element(node, &show_fileset_quota_cmd);
            install_element(node, &report_user_quota_cmd);
            install_element(node, &report_group_quota_cmd);
            install_element(node, &report_fileset_quota_cmd);
            break;
    }
#endif
}
