/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:
 *
 * RCS:  $Id: cmd_route.c,v 1.12 2002/11/22 02:51:24 swang Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:
 *
 * Created by:	Greg Seibert
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#include <sys/types.h>
#include <pwd.h>
#include <readline/history.h>
#include <grp.h>
#include <unistd.h>

#include "zebra.h"
#include "command.h"
#include "log.h"
#include "vtysh.h"
#include "memory.h"

#include "nfxsh.h"
#include "exports/ipmd-lib.h"

extern uint64		result_buf[];

/*-----------------------------------------------------------------
 * Name: rt_add
 *
 * Description: add IP Route
 *
 * Created by:	Su Wang
 *
 * Date Created: 01/29/02
 *
 *-----------------------------------------------------------------
 */
DEFUN	(rt_add,
		 ena_rt_add_cmd,
		 "route add (net|default)"
		 "-g A.B.C.D [-a IP_ADDRESS/MASKLEN] [-n NODE]",
		 "Route commands\n"
		 "Add a static route\n"
		 "Net destination\n"
		 "Default need not specify destination\n"
		 "Gateway\n"
		 "Gateway address\n"
		 "Destination network\n"
		 "Network address/masklen\n"
		 "Which node\n"
		 "Node name\n")

{
	
	int			c;
	char		net_str[32], node[64], *gw_str;
	uint32		ip, mask, gw;
	int			opt_num = 0;	/* num. of options */
	int			my_argc = 2;
	int32		rc;

	gw_str = argv[1];

	net_str[0]	= '\0';
	strcpy(node, "local");
	optind = 0;
	while ((c = getopt(argc, argv, "a:n:")) != -1) {
		switch (c) {
		case 'a':
			strcpy(net_str, optarg);
			opt_num ++;
			break;

		case 'n':
			strcpy(node, optarg);
			opt_num ++;
			break;
		}
	}

	/* check parameter number matches */
	my_argc += 2* opt_num;
	if (my_argc != argc) {
		printf("incorrect number of parameters, argc %d, in_arg %d\n",
			   argc, my_argc);
		return  CMD_ERR_FAILED;
	}
	
	/* get ip and mask */
	if (strcmp(argv[0], "net")==0) { /* not a default */
		if (!(*net_str)) {
			printf("network destination is not given\n");
			return CMD_ERR_FAILED;
		}

		/* get ip address and mask */
		if (net_atoipmask(net_str, &ip, &mask) != NFX_OK) {
			printf("bad network destination %s\n", net_str);
			return CMD_ERR_FAILED;
		}
	}
	else {						/* default */
		ip = 0;
		mask = 0;
	}

	/* get gateway */
	if (!inet_aton(gw_str, (struct in_addr*)&gw)) {
		printf("bad gw address %s\n", gw_str);
		return CMD_ERR_FAILED;
	}

	rc = ipmd_AddRtCluster(ip, mask, gw, NULL, 0, node);
	
	if (rc == NFX_OK)
		return CMD_SUCCESS;	
	else {
		ipmd_PrntErrMsg(rc, NULL, NULL, ip, gw);
		return CMD_ERR_FAILED;
	}
}

/*-----------------------------------------------------------------
 * Name: rt_rm
 *
 * Description: remove IP Route
 *
 * Created by:	Su Wang
 *
 * Date Created: 01/29/02
 *
 *-----------------------------------------------------------------
 */
DEFUN	(rt_rm,
		 ena_rt_rm_cmd,
		 "route delete (net|default)"
		 "-g A.B.C.D [-a IP_ADDRESS/MASKLEN] [-n NODE]",
		 "Route commands\n"
		 "delete a static route\n"
		 "Net destination\n"
		 "Default need not specify destination\n"
		 "Gateway\n"
		 "Gateway address\n"
		 "Destination network\n"
		 "Network address/masklen\n"
		 "Which node\n"
		 "Node name\n")


{
	int			c;
	char		net_str[32], node[64], *gw_str;
	uint32		ip, mask, gw;
	int			opt_num = 0;	/* num. of options */
	int			my_argc = 2;
	int32		rc;

	gw_str = argv[1];

	net_str[0]	= '\0';
	strcpy(node, "local");
	optind = 0;
	while ((c = getopt(argc, argv, "a:n:")) != -1) {
		switch (c) {
		case 'a':
			strcpy(net_str, optarg);
			opt_num++;
			break;

		case 'n':
			strcpy(node, optarg);
			opt_num++;
			break;
		}
	}

	/* check parameter number matches */
	my_argc += 2* opt_num;
	if (my_argc != argc) {
		printf("incorrect number of parameters, argc %d, in_arg %d\n",
			   argc, my_argc);
		return  CMD_ERR_FAILED;
	}
	
	/* get ip and mask */
	if (strcmp(argv[0], "net")==0) { /* not a default */
		if (!(*net_str)) {
			printf("network destination is not given\n");
			return CMD_ERR_FAILED;
		}

		/* get ip address and mask */
		if (net_atoipmask(net_str, &ip, &mask) != NFX_OK) {
			printf("bad network destination %s\n", net_str);
			return CMD_ERR_FAILED;
		}
	}
	else {						/* default */
		ip = 0;
		mask = 0;
	}

	
	/* get gateway */
	if (!inet_aton(gw_str, (struct in_addr*)&gw))
		printf("bad gw address %s\n", gw_str);

	rc = ipmd_DelRtCluster(ip, mask, gw, NULL, 0, node);
	
	if (rc == NFX_OK) 
		return CMD_SUCCESS;
	else {
		ipmd_PrntErrMsg(rc, NULL, NULL, ip, gw);
		return CMD_ERR_FAILED;
	}

}


/*-----------------------------------------------------------------
 * Name: rt_show
 *
 * Description: show ssc route table.
 *
 * Created by:	Su Wang
 *
 * Date Created: 01/29/02
 *
 *-----------------------------------------------------------------
 */
DEFUN   (rt_show,
         vi_route_show_cmd,
		 "route show [-n NODE] [-a A.B.C.D/M]",
		 "Route commands\n"
		 "Show route table\n"
		 "Which node\n"
	     "Node name\n"
		 "Specific route\n"
		 "Network destination\n")

{
	int			count=0, need = 0;
	ipm_rt_t	*ipm_rtp;
	char		dest_str[20], mask_str[20], nexthop_str[20];
	char		node[64], net_str[64];
	uint32		ip, mask;
	int			opt_num = 0;	/* num. of options */
	int			my_argc = 0;
	int			i;
	int32		rc;

	
	strcpy(node, "local");
	net_str[0]  = '\0';

	/* copy option ourselves, 
	 * getopt() does not work
	 */
	for (i=0; (i+2) <=argc; i+=2) { /* argc must be even */
		
		if (strcasecmp(argv[i], "-n") == 0) {
			strcpy(node, argv[i+1]);
			opt_num ++;
		}

		if (strcasecmp(argv[i], "-a") == 0) {
			strcpy(net_str, argv[i+1]);
			opt_num ++;
		}
	}
	
	   
	/* check parameter number matches */
	my_argc += 2* opt_num;
	if (my_argc != argc) {
		printf("incorrect number of parameters, argc %d, in_arg %d\n",
			   argc, my_argc);
		return  CMD_ERR_FAILED;
	}

	/* get ip address and mask */
	if (*net_str)
		if (net_atoipmask(net_str, &ip, &mask) != NFX_OK) {
			printf("bad network destination %s\n", net_str);
			return CMD_ERR_FAILED;
		}


	rc = ipmd_GetRtTblCluster((ipm_rt_t*)result_buf, 
							  IPM_SSC_BUFSIZE*sizeof(uint64),
							  &count, &need, node);
	
	if (rc != IPM_GET_OK) {
		ipmd_PrntErrMsg(rc, NULL, NULL, 0, 0);
		return CMD_ERR_FAILED;
	}
	
	printf("Route Table\n");
	printf("---------------------------------------------------------\n");
	printf("dest              mask              nexthop           interface\n");
	printf("---------------------------------------------------------\n");
	for (ipm_rtp = (ipm_rt_t*)result_buf; count >0; count--, ipm_rtp++) {
		
		/* look for specific route */
		if (*net_str && 
			((ip != ipm_rtp->rt_addr) || (mask != ipm_rtp->rt_mask)))
			continue;
		
		
		strcpy(dest_str, inet_ntoa(*((struct in_addr*)&ipm_rtp->rt_addr)));
		strcpy(mask_str, inet_ntoa(*((struct in_addr*)&ipm_rtp->rt_mask)));
		strcpy(nexthop_str, inet_ntoa(*((struct in_addr*)&ipm_rtp->rt_nhop)));
#if defined(LINK_AGG)
		if ((ipm_rtp->rt_vifName[0] != '\0') )
			printf("%-18s%-18s%-18s%s\n", 
				   dest_str, mask_str, nexthop_str, ipm_rtp->rt_vifName);
		else
#endif /* LINK_AGG */
			printf("%-18s%-18s%-18s%s\n", 
				   dest_str, mask_str, nexthop_str, ipm_rtp->rt_ifname);
	}

	return CMD_SUCCESS;
}

void
install_route (priv_t myPriv, enum node_type node)
{
	switch (myPriv) {
	case diagPriv:
	case fullPriv:
		install_element(node, &ena_rt_rm_cmd);
	case changePriv:
	case addPriv:
		install_element(node, &ena_rt_add_cmd);

	case viewPriv:
		install_element(node, &vi_route_show_cmd);	
	}
}

 
