/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:
 *
 * RCS:  $Id: cmd_snapshot.c,v 1.11 2002/12/04 19:13:58 lanc Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:
 *
 * Created by:	Greg Seibert
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#include <sys/types.h>
#include <pwd.h>
#include <readline/history.h>
#include <grp.h>
#include <unistd.h>

#include "zebra.h"
#include "command.h"
#include "log.h"
#include "vtysh.h"
#include "memory.h"

#include "nfxsh.h"

#define SNAPSHOT_STR    "Snapshot administration commands\n"

/*
 * commands and arguments for fscmd
 */
#define FSCMD               "fscmd"
#define SNAP_CREATE_CMD     "snap create"
#define SNAP_REMOVE_CMD     "snap remove"
#define SNAP_SCHED_CMD      "snap_sched"
#define SNAP_SHOW_CMD       "snap_show"
#define SNAP_SHOW_USAGE     "usage"
#define SNAP_SHOW_SCHED     "sched"
#define SNAP_REVERT_CMD     "snap_revert"
#define SNAP_ENABLE_CMD     "snap_state"
#define SNAP_DISABLE_CMD    "snap_state"
#define SNAP_PIN_CMD        "snap_pin"
#define SNAP_UNPIN_CMD      "snap_pin"
#define SNAP_RENAME_CMD     "snap_rename"

/*
 * snarfed from cmd_vol.c
 */
static int has_digits(char *str)
{
    int cnt;

    for (cnt=0; cnt < strlen(str); cnt++)
        if (!isdigit(str[cnt]))
            return 0;
    return 1;
}


/*
 * snapshot create command
 * nfxsh syntax: snapshot create VOL SNAPNAME
 * fscmd syntax: fscmd snap create VOL SNAPNAME
 */
DEFUN   (snap_create,
         snap_create_cmd,
         "snapshot create VOL SNAPNAME",
         SNAPSHOT_STR
         "create a snapshot\n"
         "create on this volume name\n"
         "create with this snapshot name\n")
{
    int rc;

    rc = execute_command(FSCMD,
                         argc+1,
                         SNAP_CREATE_CMD,
                         argv[0],           // VOL
                         argv[1],           // SNAPNAME
                         NULL);

    if (rc) return CMD_ERR_FAILED;
    return CMD_SUCCESS;
}
        

/*
 * snapshot delete command
 * nfxsh syntax: snapshot delete VOL SNAPNAME
 * fscmd syntax: fscmd snap remove VOL SNAPNAME
 */
DEFUN   (snap_remove,
         snap_remove_cmd,
         "snapshot delete VOL SNAPNAME",
         SNAPSHOT_STR
         "delete a snapshot\n"
         "delete from this volume name\n"
         "delete this snapshot name\n")
{
    int rc;

    rc = execute_command(FSCMD,
                         argc+1,
                         SNAP_REMOVE_CMD,
                         argv[0],           // VOL
                         argv[1],           // SNAPNAME
                         NULL);

    if (rc) return CMD_ERR_FAILED;
    return CMD_SUCCESS;
}


/*
 * snapshot disable command
 * nfxsh syntax: snapshot disable VOL
 * fscmd syntax: fscmd snap_state VOL disable
 */
DEFUN   (snap_disable,
         snap_disable_cmd,
         "snapshot disable VOL",
         SNAPSHOT_STR
         "disable snapshot feature\n"
         "disable snapshot feature for volume name\n")
{
    int rc;

    rc = execute_command(FSCMD,
                         3,
                         SNAP_DISABLE_CMD,
                         argv[0],           // VOL
                         "disable",
                         NULL);

    if (rc) return CMD_ERR_FAILED;
    return CMD_SUCCESS;
}


/*
 * snapshot enable command
 * nfxsh syntax: snapshot enable VOL
 * fscmd syntax: fscmd snap_state VOL enable
 */
DEFUN   (snap_enable,
         snap_enable_cmd,
         "snapshot enable VOL",
         SNAPSHOT_STR
         "enable snapshot feature\n"
         "enable snapshot feature for volume name\n")
{
    int rc;

    rc = execute_command(FSCMD,
                         3,
                         SNAP_ENABLE_CMD,
                         argv[0],           // VOL
                         "enable",
                         NULL);

    if (rc) return CMD_ERR_FAILED;
    return CMD_SUCCESS;
}


/*
 * snapshot schedule command
 * nfxsh syntax: snapshot schedule VOL -w MAX_WEEKLY -d MAX_DAILY -h MAX_HOURLY -l HOURS
 *               where HOURS is a list of hours separated by a comma
 *                  eg -l 8,10,12,14,16
 * fscmd syntax: fscmd snap_sched <volname> <max-weekly> <max-daily> <max-hourly> [@1-24]
 */
#define SNAPSHOT_SCHED_CMD "snapshot schedule VOL [-w MAX_WEEKLY] [-d MAX_DAILY] [-h MAX_HOURLY] [-l HOURS]"
DEFUN   (snap_sched,
         snap_sched_cmd,
         SNAPSHOT_SCHED_CMD,
         SNAPSHOT_STR
         "schedule automatic snapshot creation and deletion\n"
         "set this schedule for this volume name\n"
         "[-w MAX_WEEKLY] [-d MAX_DAILY] [-h MAX_HOURLY] [-l HOURS]\n")
{
    int rc;

    if (argc == 3) {
        rc = execute_command(FSCMD,
                            argc+1,
                            SNAP_SCHED_CMD,
                            argv[0],    // VOL
                            argv[1],    // option
                            argv[2],    // value
                            NULL);
    } else if (argc == 5) {
        rc = execute_command(FSCMD,
                            argc+1,
                            SNAP_SCHED_CMD,
                            argv[0],    // VOL
                            argv[1],    // option 1
                            argv[2],    // value 1
                            argv[3],    // option 2
                            argv[4],    // value 2
                            NULL);
    } else if (argc == 7) {
        rc = execute_command(FSCMD,
                            argc+1,
                            SNAP_SCHED_CMD,
                            argv[0],    // VOL
                            argv[1],    // option 1
                            argv[2],    // value 1
                            argv[3],    // option 2
                            argv[4],    // value 2
                            argv[5],    // option 3
                            argv[6],    // value 3
                            NULL);
    } else if (argc == 9) {
        rc = execute_command(FSCMD,
                            argc+1,
                            SNAP_SCHED_CMD,
                            argv[0],    // VOL
                            argv[1],    // option 1
                            argv[2],    // value 1
                            argv[3],    // option 2
                            argv[4],    // value 2
                            argv[5],    // option 3
                            argv[6],    // value 3
                            argv[7],    // option 4
                            argv[8],    // value 4
                            NULL);
    } else {
        printf("invalid number of arguments %d\n", argc);
        printf("usage: %s\n", SNAPSHOT_SCHED_CMD);
        return CMD_ERR_FAILED;
    }

    if (rc) return CMD_ERR_FAILED;
    return CMD_SUCCESS;
}


/*
 * snapshot show command
 * nfxsh syntax: snapshot show VOL -v TYPE
 * fscmd syntax: fscmd snap_show [sched | usage] <volname>
 */
DEFUN   (snap_show,
         snap_show_cmd,
         "snapshot show VOL -v TYPE",
         SNAPSHOT_STR
         "show snapshot schedule or usage\n"
         "show information for this volume name\n"
         "mandatory option to specify show type: schedule or usage\n"
         "schedule to show current snapshot schedule or usage to show current snapshot disk usage\n")
{
    int rc;
    char *fscmdShowType;

    // make nfxsh more flexible by taking more options
    // (for sched and schedule) than fscmd

    if (strcmp(argv[1], "usage") == 0) {
        fscmdShowType = SNAP_SHOW_USAGE;
    } else if (strcmp(argv[1], "schedule") == 0) {
        fscmdShowType = SNAP_SHOW_SCHED;
    } else if (strcmp(argv[1], "sched") == 0) {
        fscmdShowType = SNAP_SHOW_SCHED;
    } else return CMD_ERR_FAILED;

    rc = execute_command(FSCMD,
                         argc+1,
                         SNAP_SHOW_CMD,
                         fscmdShowType,     // TYPE
                         argv[0],           // VOL
                         NULL);

    if (rc) return CMD_ERR_FAILED;
    return CMD_SUCCESS;
}


/*
 * snapshot revert command
 * nfxsh syntax: snapshot revert VOL [-n SNAPNAME] [-i SNAPID]
 * fscmd syntax: fscmd snap_revert <volname> [-n name] [-i snapnum]
 */
DEFUN   (snap_revert,
         snap_revert_cmd,
         "snapshot revert VOL [-n SNAPNAME] [-i SNAPID]",
         SNAPSHOT_STR
         "revert to a snapshot\n"
         "do the revert on this volume name\n"
         "revert to this snapshot name or this internal snap id\n")
{
    extern int optind;
    int rc = CMD_ERR_FAILED;
    char c, vol[256], snaparg[256];
    boolean revertName = FALSE, revertId = FALSE;

    // save vol since argv will be munged by getopt

    strcpy(vol, argv[0]);

    if (argc == 1) {
        goto snap_revert_execute_cmd;
    }

    optind = 0;
    while ((c = getopt(argc, argv, "n:i:")) != -1) {
        switch (c) {
            case ':':
                printf("snapshot revert: missing parameter for option\n");
                return CMD_ERR_FAILED;
            case '?':
                printf("snapshot revert: unknown option\n");
                return CMD_ERR_FAILED;
            case 'n':
                revertName = TRUE;
                strcpy(snaparg, optarg);
                break;
            case 'i':
                if (has_digits(optarg)) {
                    strcpy(snaparg, optarg);
                    revertId = TRUE;
                } else {
                    printf("snapshot revert: -i option requires a snap id (from snapshot show)\n");
                    return CMD_ERR_FAILED;
                }
                break;
        }
    }

    if (revertName) {
        rc = execute_command(FSCMD,
                            argc+1,
                            SNAP_REVERT_CMD,
                            vol,
                            "-n",
                            snaparg,    // SNAPNAME
                            NULL);
    }
    if (revertId) {
        rc = execute_command(FSCMD,
                            argc+1,
                            SNAP_REVERT_CMD,
                            vol,
                            "-i",
                            snaparg,    // SNAPID
                            NULL);
    }

snap_revert_execute_cmd:

    if (argc == 1) {
        rc = execute_command(FSCMD,
                             argc+1,
                             SNAP_REVERT_CMD,
                             vol,
                             NULL);
    }

    if (rc) return CMD_ERR_FAILED;
    return CMD_SUCCESS;
}


/*
 * snapshot pin command
 * nfxsh syntax: snapshot pin VOL
 * fscmd syntax: fscmd snap_pin VOL SNAPNAME pin
 */
DEFUN   (snap_pin,
         snap_pin_cmd,
         "snapshot pin VOL SNAPNAME",
         SNAPSHOT_STR
         "snapshot pin feature\n"
         "unpin snapshot for volume name\n"
         "snapshot name to pin\n")
{
    int rc;

    rc = execute_command(FSCMD,
                         4,
                         SNAP_PIN_CMD,
                         argv[0],           // VOL
                         argv[1],           // SNAPNAME
                         "pin",
                         NULL);

    if (rc) return CMD_ERR_FAILED;
    return CMD_SUCCESS;
}


/*
 * snapshot unpin command
 * nfxsh syntax: snapshot unpin VOL
 * fscmd syntax: fscmd snap_pin VOL SNAPNAME unpin
 */
DEFUN   (snap_unpin,
         snap_unpin_cmd,
         "snapshot unpin VOL SNAPNAME",
         SNAPSHOT_STR
         "snapshot unpin feature\n"
         "unpin snapshot for volume name\n"
         "snapshot name to unpin\n")
{
    int rc;

    rc = execute_command(FSCMD,
                         4,
                         SNAP_PIN_CMD,
                         argv[0],           // VOL
                         argv[1],           // SNAPNAME
                         "unpin",
                         NULL);

    if (rc) return CMD_ERR_FAILED;
    return CMD_SUCCESS;
}


/*
 * snapshot rename command
 * nfxsh syntax: snapshot rename VOL OLDNAME NEWNAME
 * fscmd syntax: fscmd snap_rename VOL OLDNAME NEWNAME
 */
DEFUN   (snap_rename,
         snap_rename_cmd,
         "snapshot rename VOL OLDNAME NEWNAME",
         SNAPSHOT_STR
         "snapshot rename feature\n"
         "rename snapshot for volume name\n"
         "old snapshot name to rename\n"
         "new snapshot name\n")
{
    int rc;

    rc = execute_command(FSCMD,
                         5,
                         SNAP_RENAME_CMD,
                         argv[0],           // VOL
                         argv[1],           // OLDNAME
                         argv[2],           // NEWNAME
                         NULL);

    if (rc) return CMD_ERR_FAILED;
    return CMD_SUCCESS;
}


void
install_snapshot (priv_t myPriv, enum node_type node)
{
    // put the higher privilage commands first and fall through
    // to pick up the lower priority ones 

    switch (myPriv) {
        case diagPriv:
        case fullPriv:

            install_element(node, &snap_unpin_cmd);
            install_element(node, &snap_disable_cmd);
            install_element(node, &snap_remove_cmd);
            install_element(node, &snap_revert_cmd);

        case changePriv:

            install_element(node, &snap_sched_cmd);
            install_element(node, &snap_pin_cmd);
            install_element(node, &snap_rename_cmd);
            install_element(node, &snap_enable_cmd);

        case addPriv:

            install_element(node, &snap_create_cmd);

        case viewPriv:

            install_element(node, &snap_show_cmd);

        break;
    }
}
