/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:
 *
 * RCS:  $Id: cmd_system.c,v 1.46 2002/11/02 22:31:51 maximk Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:
 *
 * Created by:	Greg Seibert
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#include <sys/types.h>
#include <tzfile.h>
#include <pwd.h>
#include <readline/history.h>
#include <grp.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "zebra.h"
#include "command.h"
#include "log.h"
#include "vtysh.h"
#include "memory.h"

#include "nfxsh.h"
#include "exports/opt-api.h"
#include "exports/timekeep-api.h"
#include "exports/evm_com_defs.h"
#include "exports/nfxnis-api.h"
#include <exports/evm-api.h>

#include <exports/anpssc-api.h>

#define CHASSIS_CMD_PROG "/usr/local/agile/bin/chassis"
#define NETSTAT_CMD_PROG "/usr/bin/netstat"

#define SYSCMDS "System commands\n"

extern evm_volId_t nfxsh_get_vol_id(char *name);

in_addr_t
nfxsh_nis_hostAddr(char *exportAddr, char *hostName)
{
	in_addr_t rv;
	become_root();
	rv = nfxnis_hostAddr(exportAddr, hostName);
	unbecome_root();
	return rv;
}


DEFUN	(system_ping,
		ping_cmd,
		"system ping HOSTNAME_OR_IPADDRESS",
		SYSCMDS
		"Send ICMP ECHO requests\n"
		"Destination hostname or address\n")
{
	in_addr_t host_ip;
	struct in_addr ip_addr;
	char *ping_this;

	host_ip = inet_addr (argv[0]);

	if (host_ip == INADDR_NONE) {
		ip_addr.s_addr = nfxsh_nis_hostAddr("0.0.0.0", argv[0]);
		if (!ip_addr.s_addr) {
			fprintf (stderr, "%s not known\n", argv[0]);
			return CMD_ERR_FAILED;
		}
		ping_this = strdup (inet_ntoa(ip_addr));
	}
	else {
		ping_this = strdup (argv[0]);
	}

	execute_command ("/sbin/ping", 1, ping_this, NULL);

	free (ping_this);

	return CMD_SUCCESS;
}

DEFUN	(system_traceroute,
		traceroute_cmd,
		"system traceroute HOSTNAME_OR_IPADDRESS",
		SYSCMDS
		"Trace IP route to destination\n"
		"Destination address or hostname\n")
{
	in_addr_t host_ip;
	struct in_addr ip_addr;
	char *trace_this;

	host_ip = inet_addr (argv[0]);

	if (host_ip == INADDR_NONE) {
		ip_addr.s_addr = nfxsh_nis_hostAddr ("0.0.0.0", argv[0]);
		if (!ip_addr.s_addr) {
			fprintf (stderr, "%s not known\n", argv[0]);
			return CMD_ERR_FAILED;
		}
		trace_this = strdup (inet_ntoa(ip_addr));
	}
	else {
		trace_this = strdup (argv[0]);
	}

	execute_command ("/usr/sbin/traceroute", 1, trace_this, NULL);
	free (trace_this);
	return CMD_SUCCESS;
}

/* Gilbert's aliases */

ALIAS	(system_ping,
		p_ping_cmd,
		"ping HOSTNAME_OR_IPADDRESS",
		"Send ICMP ECHO requests\n"
		"Destination hostname or address\n")

ALIAS	(system_traceroute,
		t_traceroute_cmd,
		"traceroute HOSTNAME_OR_IPADDRESS",
		"Trace IP route to destination\n"
		"Destination address or hostname\n")

DEFUN	(system_halt,
		halt_cmd,
		"system halt",
		SYSCMDS
		"Halts the system\n")
{
	become_root();
	execute_command ("/sbin/halt", 0, NULL, NULL);
	return CMD_SUCCESS;
}

DEFUN	(system_total_reset,
		total_reset_cmd,
		"system reboot",
		SYSCMDS
		"Reboot the system\n")
{
/*
	We need to dismount things before we reset the embedded side
	so let's run the shutdown script first.
 */
	execute_command ("/bin/sh", 1, "/etc/rc.shutdown", NULL);
/*
	Reset the embedded boards
 */
	execute_command(CHASSIS_CMD_PROG, 3, "reboot", "slot", "1", NULL);
	execute_command(CHASSIS_CMD_PROG, 3, "reboot", "slot", "2", NULL);
/*
	Now reboot us
 */
	become_root();
	kill_daemon ("/var/run/mcp.pid");
	execute_command ("/sbin/reboot", 1, "-c", NULL);
	return CMD_SUCCESS;
}

DEFUN	(system_reboot,
		reboot_cmd,
		"system reset ssc [-s]",
		SYSCMDS
		"Resets a card/cpu\n"
		"reset the SSC\n"
		"reboot the SSC and switch active/standby flash\n")
{
	if (argc == 1) {
		if (!strcmp(argv[0], "-s")) {
			become_root();
			execute_command("/sbin/reboot", 1, "-s", NULL);
			unbecome_root();
			return CMD_SUCCESS;
		}
		else {
			return CMD_ERR_FAILED;
		}
	}
	else {
		pid_t twirl = start_twiddle();

		become_root();
		kill_daemon ("/var/run/mcp.pid");
		kill_daemon ("/var/run/snmpd.pid");
		kill_daemon ("/var/www/logs/httpd.pid");
		sleep (5);
		execute_command ("/usr/local/agile/bin/mcp", 0, NULL);
		unbecome_root();
		stop_twiddle (twirl);
	}
	return CMD_SUCCESS;
}


DEFUN (system_crash,
	   crash_cmd,
	   "system modify corevolume [VOLUMENAME]",
	   SYSCMDS
	   "Changes stuff about the system\n"
	   "Set/clear system core dump volume\n"
	)
{
	if (argc == 1) {
		char *objs[2] = { "option.corevolume", argv[0] };
		evm_volId_t volid;
		if (strlen(argv[0]) >= VOL_MAX_NAME - 1) {
			printf("volume name is too long\n");
			return CMD_ERR_FAILED;
		}
		if ((volid = nfxsh_get_vol_id(argv[0])).whole_thing == - 1) {
			printf("%s: volume does not exist\n", argv[0]);
			return CMD_ERR_FAILED;
		}
		return put_cfg_val(objs);
	}
	else {
		char *objs[2] = { "option.corevolume", NULL };
		return del_cfg_obj(objs);
	}
}


#define COREVOL_BLOCKS  ( 2LL * 1024LL * 1024LL * 1024LL ) / EVM_LBLK_SIZE


DEFUN (system_corevol_create,
	   system_corevol_create_cmd,
	   "system createcorevolume VOLUMENAME",
	   SYSCMDS
	   "Create a volume for coredump\n"
	)
{
	if (argc == 1) {
		struct timeval to = { 5, 0 };
		if (strlen(argv[0]) >= VOL_MAX_NAME - 1) {
			printf("volume name is too long\n");
			return CMD_ERR_FAILED;
		}
		else {
			evm_MsgVolCreate_t createReq;
			evm_MsgStat_t createRsp;
			int len;
			bzero((void *) &createReq, sizeof(evm_MsgVolCreate_t));
			createReq.hdr.msgType = EVM_VOL_CREATE;
			createReq.hdr.clientPrivate = getpid();
			createReq.config.id.whole_thing = 0;
			strncpy(createReq.config.name, argv[0], VOL_MAX_NAME);
			createReq.config.type = EVM_TYPE_COREVOL;
			createReq.config.maxSize = 0;
			createReq.config.softQuota = 0;
			createReq.config.minAutoGrow = 0;
			createReq.config.onNIM = 1;
			createReq.config.onLine = FALSE;
			createReq.config.attrs.raidLevel = EVM_RAID_ANY;
			createReq.config.attrs.minLunSize = COREVOL_BLOCKS;
			createReq.config.attrs.maxLunSize = EVM_DONT_CARE;
			createReq.config.attrs.flags = EVM_SAME_VENDOR | EVM_SAME_MODEL;
			strcpy(createReq.config.attrs.raidCtrl, "");
			createReq.config.attrs.num = 0;
			createReq.nbLun = 0;
			if ((len = sendAgileMsg(-1, &createReq, sizeof(evm_MsgVolCreate_t),
									"evm", NFX_SSC_UNIX, 0, 0,
									&to, &createRsp, sizeof(createRsp))) > 0) {
				if (createRsp.stat != NFX_OK) {
					return CMD_ERR_FAILED;
				}
				else {
					return 0;
				}
			}
		}
	}
	return CMD_ERR_FAILED;
}


DEFUN (system_corevol_delete,
	   system_corevol_delete_cmd,
	   "system deletecorevolume VOLUMENAME",
	   SYSCMDS
	   "Delete a coredump volume\n"
	)
{
	if (argc == 1) {
		struct timeval to = { 5, 0 };
		if (strlen(argv[0]) >= VOL_MAX_NAME - 1) {
			printf("volume name is too long\n");
			return CMD_ERR_FAILED;
		}
		else {
			evm_MsgVolDestroy_t delReq;
			evm_volId_t         id;

			char **val = get_cfg_val("option.corevolume");
			/* Remove the corevolume assignment if it is
			   assigned to the volume we are deleting */
			if (val != NULL && !strcmp(val[1], argv[0])) {
				char *objs[2] = { "option.corevolume", NULL };
				del_cfg_obj(objs);
			}
			if (val != NULL)
				free_cfg_val(val);

			/* Delete the volume */
			id = nfxsh_get_vol_id(argv[0]);
			if (id.whole_thing != -1) {
				char buf[32 * 1024];
				int len;
				evm_MsgStat_t * stat_msg;
				bzero(&delReq, sizeof(evm_MsgVolDestroy_t));

				delReq.hdr.msgType = EVM_VOL_DESTROY;
				delReq.hdr.clientPrivate = getpid();
				delReq.id = id;

				len = sendAgileMsg(-1, &delReq,
									   sizeof(evm_MsgVolDestroy_t),
									   "evm",
									   NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
									   &to,
									   buf, sizeof(buf));

				stat_msg = (evm_MsgStat_t *)buf;
				if (len < 0 || stat_msg->stat != NFX_OK)
				{
					printf("Destroy failed.\n");
					return(CMD_ERR_FAILED);
				}
				printf("Done.\n");
				return CMD_SUCCESS;
			}
		}
	}
	return CMD_ERR_FAILED;
}


DEFUN (system_crashauto,
	   crash_auto_cmd,
	   "system modify coredump (on|off)",
	   SYSCMDS
	   "Changes stuff about the system\n"
	   "Change automatic coredump option\n"
	   "Set automatic coredump on\n"
	   "Set automatic coredump off\n"
	)
{
	if (!strcmp("on", argv[0])) {
		char *objs[2] = { "option.coredump", "on" };
		return put_cfg_val(objs);
	}
	else {
		char *objs[2] = { "option.coredump", NULL };
		return del_cfg_obj(objs);
	}
}

DEFUN (system_crash_show,
	   crash_show_cmd,
	   "system show corevolume",
	   SYSCMDS
	   "Show core dump volume setting\n")
{
	char **val = get_cfg_val("option.corevolume");
	if (val == NULL) {
		printf("Core dump volume is not assigned\n");
	}
	else {
		printf("Core dump volume is %s\n", val[1]);
		free_cfg_val(val);
	}
	return CMD_SUCCESS;
}


DEFUN (system_crashauto_show,
	   crashauto_show_cmd,
	   "system show coredump",
	   SYSCMDS
	   "Show crashdump volume setting\n")
{
	char **val = get_cfg_val("option.coredump");
	if (val == NULL) {
		printf("Automatic core dump is off\n");
	}
	else {
		printf("Automatic core dump is on\n");
		free_cfg_val(val);
	}
	return CMD_SUCCESS;
}


DEFUN (system_coredir,
	   system_coredir_cmd,
	   "system modify coresave VOLUME:/DIRECTORY",
	   SYSCMDS
	   "Changes stuff about the system\n"
	   "Set the directory where core dumps are saved\n")
{
	if (argc == 1) {
		char *column = strchr(argv[0], ':');
		if (column != NULL) {
			if (strlen(argv[0]) < 256) {
				char *objs[2] = { "option.coresave", argv[0] };
				evm_volId_t volid;
				char volname[256];
				memcpy(volname, argv[0], column - argv[0]);
				volname[column - argv[0]] = 0;
				if ((volid = nfxsh_get_vol_id(volname)).whole_thing != -1) {
					return put_cfg_val(objs);
				}
				else {
					printf("%s: volume does not exist\n", volname);
					return CMD_ERR_FAILED;
				}
			}
			else {
				printf("path is too long\n");
				return CMD_ERR_FAILED;
			}
		}
		else {
			printf("wrong argument format\n");
			return CMD_ERR_FAILED;
		}
	}
	else {
		return CMD_ERR_FAILED;
	}
}


DEFUN (system_coredir_show,
	   system_coredir_show_cmd,
	   "system show coresave",
	   SYSCMDS
	   "Show the directory where core dumps are saved\n")
{
	char **val = get_cfg_val("option.coresave");
	if (val == NULL) {
		printf("Directory for saving core dumps is not assigned\n");
	}
	else {
		printf("Directory for saving core dumps is %s\n", val[1]);
		free_cfg_val(val);
	}
	return CMD_SUCCESS;
}


DEFUN	(system_reset,
		reset_cmd,
		"system reset (fp|sp)",
		SYSCMDS
		"Resets a card/cpu\n"
		"reset the entire fp\n"
		"reset the entire sp\n")
{
	char	* dotp;

	if ((dotp = strchr(argv[0], '.'))) {
		dotp++;
		if (!(strncmp(argv[0], "fp", 2))) {
			if (execute_command(CHASSIS_CMD_PROG, 3, "reboot", "slot", "1", NULL))
				return CMD_ERR_FAILED;
		}
		else if (!(strncmp(argv[0], "sp", 2))) {
			if (*dotp == '0') {
				if (execute_command(CHASSIS_CMD_PROG, 4, "reboot", "cpu", "2", "0",
																		NULL))
					return CMD_ERR_FAILED;
			}
			else if (*dotp == '1') {
				if (execute_command(CHASSIS_CMD_PROG, 4, "reboot", "cpu", "2", "1",
																		NULL))
					return CMD_ERR_FAILED;
			}
			else if (*dotp == '2') {
				if (execute_command(CHASSIS_CMD_PROG, 4, "reboot", "cpu", "2", "2",
																		NULL))
					return CMD_ERR_FAILED;
			}
			else if (*dotp == '3') {
				if (execute_command(CHASSIS_CMD_PROG, 4, "reboot", "cpu", "2", "3",
																		NULL))
					return CMD_ERR_FAILED;
			}
			else {
				printf("Invalid CPU number, %d\n", atoi(dotp));
				return CMD_ERR_FAILED;
			}
		}
	}
	else {
		if (!(strncmp(argv[0], "fp", 2))) {
			if (execute_command(CHASSIS_CMD_PROG, 3, "reboot", "slot", "1", NULL))
				return CMD_ERR_FAILED;
		}
		else if (!(strncmp(argv[0], "sp", 2))) {
			if (execute_command(CHASSIS_CMD_PROG, 3, "reboot", "slot", "2", NULL))
				return CMD_ERR_FAILED;
		}
	}
	return CMD_SUCCESS;
}

ALIAS	(system_reset,
		reset_reset_cmd,
		"system reset (sp.0|sp.1|sp.2|sp.3)",
		SYSCMDS
		"Resets a card/cpu\n"
		"reset an sp slice\n")

DEFUN	(system_show,
		show_cmd,
		"system show (ipstat|connection|chassis|version|temperature)",
		SYSCMDS
		"Shows stuff about the system\n")
{
	if (*argv[0] == 'i') {
		if (execute_command(NETSTAT_CMD_PROG, 3, "-f", "inet", "-s", NULL))
			return CMD_ERR_FAILED;
	}
	else if (!(strncmp(argv[0], "ch", 2))) {
		if (execute_command(CHASSIS_CMD_PROG, 2, "view", "sys", NULL))
			return CMD_ERR_FAILED;
	}
	else if (*argv[0] == 'c') {
		if (execute_command(NETSTAT_CMD_PROG, 3, "-a", "-f", "inet", NULL))
			return CMD_ERR_FAILED;
	}
	else if (*argv[0] == 'v') {
		if (execute_command(CHASSIS_CMD_PROG, 2, "view", "ver", NULL))
			return CMD_ERR_FAILED;
	}
	else if (*argv[0] == 't') {
		if (execute_command(CHASSIS_CMD_PROG, 2, "view", "temp", NULL))
			return CMD_ERR_FAILED;
	}
	
	return CMD_SUCCESS;
}

DEFUN	(system_upgrade,
		upgrade_cmd,
		"system upgrade [-s] HOST:/PATH",
		SYSCMDS
		"Upgrades the system software\n")
{
	return do_upgrade(argc, argv);
}


DEFUN  (system_version,
		version_cmd,
		"system version",
		SYSCMDS
		"Display the system software version")
{
	return do_show_version(argc, argv);
}

#define TIMECMDS "Controls the system time\n"
#define NTPCMDS "Controls how to use the Network Time Protocol\n"

/* The following was "borrowed" from the OpenBSD date command and to
 * comply with their rules... */
/*
 * Copyright (c) 1985, 1987, 1988, 1993
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#define	ATOI2(ar)	((ar)[0] - '0') * 10 + ((ar)[1] - '0'); (ar) += 2;
#define badformat() { fprintf (stderr, "Bad time format\n"); return; }
static void
setthetime(p)
	register char *p;
{
	register struct tm *lt;
	struct timeval tv;
	char *dot, *t;
	int bigyear;
	int yearset = 0;
	time_t tval = time(NULL);

	for (t = p, dot = NULL; *t; ++t) {
		if (isdigit(*t))
			continue;
		if (*t == '.' && dot == NULL) {
			dot = t;
			continue;
		}
		badformat();
	}

	lt = localtime(&tval);

	if (dot != NULL) {			/* .SS */
		*dot++ = '\0';
		if (strlen(dot) != 2)
			badformat();
		lt->tm_sec = ATOI2(dot);
		if (lt->tm_sec > 61)
			badformat();
	} else
		lt->tm_sec = 0;

	switch (strlen(p)) {
	case 12:				/* cc */
		bigyear = ATOI2(p);
		lt->tm_year = bigyear * 100 - TM_YEAR_BASE;
		yearset = 1;
		/* FALLTHROUGH */
	case 10:				/* yy */
		if (yearset) {
			lt->tm_year += ATOI2(p);
		} else {
			lt->tm_year = ATOI2(p);
			if (lt->tm_year < 69)		/* hack for 2000 */
				lt->tm_year += (2000 - TM_YEAR_BASE);
			else
				lt->tm_year += (1900 - TM_YEAR_BASE);
		}
		/* FALLTHROUGH */
	case 8:					/* mm */
		lt->tm_mon = ATOI2(p);
		if ((lt->tm_mon > 12) || !lt->tm_mon)
			badformat();
		--lt->tm_mon;			/* time struct is 0 - 11 */
		/* FALLTHROUGH */
	case 6:					/* dd */
		lt->tm_mday = ATOI2(p);
		if ((lt->tm_mday > 31) || !lt->tm_mday)
			badformat();
		/* FALLTHROUGH */
	case 4:					/* HH */
		lt->tm_hour = ATOI2(p);
		if (lt->tm_hour > 23)
			badformat();
		/* FALLTHROUGH */
	case 2:					/* MM */
		lt->tm_min = ATOI2(p);
		if (lt->tm_min > 59)
			badformat();
		break;
	default:
		badformat();
	}

	/* convert broken-down time to UTC clock time */
	if ((tval = mktime(lt)) < 0)
		fprintf(stderr, "specified date is outside allowed range\n");

	tv.tv_sec = tval;
	tv.tv_usec = 0;
	if (settimeofday(&tv, NULL))
		perror("settimeofday failed");

}
DEFUN	(system_set_time,
		set_time_cmd,
		"system time set TIME",
		SYSCMDS
		TIMECMDS
		"Set current system time\n"
		"Time of the form [[[[[[cc]yy]mm]dd]HH]MM[.SS]]\n")
{
	become_root();
	setthetime (argv[0]);
	unbecome_root();
	return CMD_SUCCESS;
}

DEFUN	(system_time,
		time_cmd,
		"system time show",
		SYSCMDS
		TIMECMDS
		"Show current time\n")
{
	execute_command ("/bin/date", 0, NULL);

	return CMD_SUCCESS;
}

DEFUN	(system_time_v,
		time_cmd_v,
		"system time show -v",
		SYSCMDS
		TIMECMDS
		"Show current time\n"
		"and include NTP configuration\n")
{
	char **opts;
	execute_command ("/bin/date", 0, NULL);

	opts = get_cfg_val (NTP_MODE);

	if (!opts) {
		fprintf (stdout, "NTP disabled\n");
	}
	else {
		free_cfg_val (opts);
		execute_command ("/usr/local/bin/ntpq", 2, "-crv", "-cpeer", NULL);
	}

	return CMD_SUCCESS;
}

DEFUN	(system_time_zone,
		system_time_zone_cmd,
		"system time zone",
		SYSCMDS
		TIMECMDS
		"Set local time zone\n")
{
	char buf[80];

	fprintf (stdout, "Current time zone is\n");

	system ("/bin/ls -l /etc/localtime | /usr/bin/sed 's;.*zoneinfo/;;'");

	fprintf (stdout, "Do you wish to change it? [yn] ");
	if (!fgets (buf, sizeof(buf), stdin) || strncasecmp ("y", buf, 1)) {
		clearerr (stdin);	// In case they hit ^D
		fprintf (stdout, "Time zone setting unchanged\n");
	}
	else {
		become_root();
		execute_command ("/usr/local/agile/bin/selecttz", 0, NULL);
		unbecome_root();
	}
	return CMD_SUCCESS;
}

DEFUN	(ntp_broadcast,
		ntp_broadcast_cmd,
		"system time ntp broadcast",
		SYSCMDS
		TIMECMDS
		NTPCMDS
		"Synchronize to NTP broadcast packets\n")
{
	tk_set_ntp_broadcast();

	become_root();
	tk_update_ntp_conf();
	unbecome_root();

	return CMD_SUCCESS;
}

DEFUN	(ntp_disable,
		ntp_disable_cmd,
		"system time ntp disable",
		SYSCMDS
		TIMECMDS
		NTPCMDS
		"Disable use of NTP for time synchronization\n")
{
	tk_disable_ntp();

	become_root();
	tk_update_ntp_conf();
	unbecome_root();

	return CMD_SUCCESS;
}

DEFUN	(ntp_server,
		ntp_server_cmd,
		"system time ntp server (-a | -d) A.B.C.D",
		SYSCMDS
		TIMECMDS
		NTPCMDS
		"Specify NTP server address\n"
		"add this server\n"
		"delete this server\n"
		"The NTP server's address\n")
{
	char **opts;
	int rc;
	int del = !strcmp(argv[0], "-d");

	opts = get_cfg_val (NTP_MODE);
/*
	If we've already got servers, add this one to the list
 */
	if (!opts && del) {
		fprintf (stderr, "No NTP servers configured\n");
		return CMD_ERR_FAILED;
	}
	free_cfg_val (opts);

	rc = CMD_SUCCESS;
	if (tk_mod_ntp_servers (argv[1], del)) {
		if (del && (errno == ENOENT)) {
			fprintf (stderr, "%s not a configured server\n", argv[1]);
			rc = CMD_ERR_FAILED;
		}
		else {
			fprintf (stderr, "Error setting NTP configuration\n");
		}
	}
/*
	Commit config changes
 */
	become_root();
	tk_update_ntp_conf();
	unbecome_root();

	return rc;
}

DEFUN	(ShowNodeName,
		ShowNodeName_cmd,
		"system show nodename",
		SYSCMDS
		"Shows stuff about the system\n"
		"Show this node's name\n")
{
	char node_name[256];

	if (gethostname(node_name, sizeof(node_name))) {
		perror ("Error getting node name");
		return CMD_ERR_FAILED;
	}
	else {
		fprintf (stdout, "Node name is: %s\n", node_name);
		return CMD_SUCCESS;
	}
}

DEFUN	(SetNodeName,
		SetNodeName_cmd,
		"system modify nodename NODENAME",
		SYSCMDS
		"Changes stuff about the system\n"
		"Change this node's name\n"
		"This would be the new name\n")
{
	int rc;
	char *myname = argv[0];
	char old_name[256];
	
	gethostname (old_name, sizeof (old_name));

	if (!strlen (myname)) {
		fprintf (stderr, "Invalid node name\n");
		rc = CMD_ERR_FAILED;
	}
	else if (!node_has_services (old_name)) {
		become_root();
		if (sethostname(myname, strlen(myname))) {
			perror ("Error setting node name");
			rc = CMD_ERR_FAILED;
		}
		else {
			if (((rc = modify_readonly_file("/etc/myname", string_to_file,
										   myname)) != CMD_SUCCESS) ||
				((rc = modify_readonly_file("/etc/hosts", redo_hosts,
										   myname))!= CMD_SUCCESS)) {
				perror ("Unable to set node name");
				rc = CMD_ERR_FAILED;
			}
		}
		remove_cluster_services();
		warn_need_reboot();	// Make it so
		unbecome_root();
	}
	else {
		fprintf (stderr, "Not allowed to rename a node while it has\n");
		fprintf (stderr, "services still assigned because its name\n");
		fprintf (stderr, "is used to label storage\n");
		rc = CMD_ERR_FAILED;
	}

	return rc;
}


#define COPYCMDS "System backup commands\n"

DEFUN   (system_copy_init,
		 system_copy_init_cmd,
		 "system copy init",
		 SYSCMDS
		 COPYCMDS
		 "Initialize the secondary local disk\n"
	)
{
	return do_copy_init(argc, argv);
}



DEFUN   (system_copy_all,
		 system_copy_all_cmd,
		 "system copy all [-i]",
		 SYSCMDS
		 COPYCMDS
		 "Save the system on the secondary local disk\n"
		 "Initialize the secondary local disk before copying\n"
	)
{
	int rc = CMD_SUCCESS;
	
	if (argc == 1 && !strcmp("-i", argv[0])) {
		rc = do_copy_init(argc, argv);
	}
	if (rc == CMD_SUCCESS) {
		rc = do_copy_all(argc, argv);
	}
	return rc;
}


#define CONFIGCMDS "Operations on system configuration\n"

DEFUN	(system_copy_config,
		 system_copy_config_cmd,
		 "system config copy",
		 SYSCMDS
		 CONFIGCMDS
		 "Save configuration files on the secondary local disk\n"
	)
{
	return do_copy_config(argc, argv);
}

DEFUN   (system_config_restore,
		 system_config_restore_cmd,
		 "system config restore",
		 SYSCMDS
		 CONFIGCMDS
		 "Restore configuration files from the secondary local disk\n"
	)
{
	return do_restore_config(argc, argv);
}


DEFUN	(system_config_reset,
		 system_config_reset_cmd,
		 "system config reset",
		 SYSCMDS
		 CONFIGCMDS
		 "Reset system configuration to factory defaults\n"
		)
{
	return do_reset_config(argc, argv);
}


DEFUN  (system_prom_upgrade,
		system_prom_upgrade_cmd,
		"system promupgrade SLOT CPU",
		SYSCMDS
		"Upgrade prom for the specified slot and cpu\n"
	)
{
	return do_prom_upgrade(argc, argv);
}

typedef struct cfgaud {
	long long		lvid;
	char			*admin;
	char			*oper;
	char			*nfs;
	struct cfgaud	*next;
} cfgaud_t;

static cfgaud_t *
get_lv (cfgaud_t **da_list, long long lv)
{
	cfgaud_t *lookie;

	for (lookie = *da_list; lookie; lookie = lookie->next) {
		if (lookie->lvid == lv) {
			break;
		}
	}

	if (!lookie) {
		lookie = malloc (sizeof(cfgaud_t));
		memset (lookie, 0, sizeof(cfgaud_t));
		lookie->next = *da_list;
		lookie->lvid = lv;
		*da_list = lookie;
	}
	return lookie;
}

static void
audit_conf (char **lv, char **slave, char **nfs)
{
	cfgaud_t *list = NULL;
	cfgaud_t *some;
	char **look;
	long long an_id;
/*
 * We need to make sure all the visible volumes are accounted for in both
 * the node and nfs assignements.
 * Collect the list of ALL volumes and their assignements.
 */
	if (lv) {
		for (look = lv; *look; look += 2) {
			an_id = strtouq (rindex (look[0], '.') + 1, NULL, 16);
			some = get_lv (&list, an_id);
			some->admin = look[1];
		}
	}

	if (slave) {
		for (look = slave; *look; look += 2) {
			char *node_name;

			an_id = strtouq (rindex (look[0], '.') + 1, NULL, 16);
			some = get_lv (&list, an_id);

			*rindex (look[0], '.') = '\0';
			node_name = index (look[0], '.') + 1;

			if (!some->oper) {
				some->oper = strdup (node_name);
			}
			else {
				char *new_thang;
				asprintf (&new_thang, "%s, %s", some->oper, node_name);
				free (some->oper);
				some->oper = new_thang;
			}
		}
	}

	if (nfs) {
		for (look = nfs; *look; look += 2) {
			an_id = strtouq (rindex (look[0], '.') + 1, NULL, 16);
			some = get_lv (&list, an_id);
			some->nfs = look[1];
		}
	}

	for (some = list; some; some = list) {
		if (some->admin && some->oper && some->nfs) {
			fprintf (stdout, "LV %016llx OK\n", some->lvid);
		}
		else {
			char resp[80];
			fprintf (stdout, "LV %016llx not quite right\n", some->lvid);

			if (some->admin) {
				char *status = some->admin;
				strsep (&status, ":");
				fprintf (stdout, "On node %s it is %s\n", status, some->admin);
			}

			if (some->oper) {
				fprintf (stdout, "Visible on nodes: %s\n", some->oper);
			}

			if (some->nfs) {
				char *volname;
				char *nfsopts;
				char *next_export = some->nfs;
				while (next_export) {
					nfsopts = next_export;
					strsep (&nfsopts, ":");
					volname = strsep (&nfsopts, "|");
					strsep (&next_export, ";");

					fprintf (stdout, "Volume %s exported as %s\n", 
					         volname, nfsopts);
				}
			}

			fputs ("Shall we delete these configuration items? y/[n] ", stdout);
			fgets (resp, sizeof(resp), stdin);

			resp[0] = toupper (resp[0]);
			if (resp[0] == 'Y') {
				char *del_this[3];
				del_this[1] = del_this[2] = NULL;

				asprintf (&del_this[0], "cluster.lv.%016llx", some->lvid);
				del_cfg_obj (del_this);
				free (del_this[0]);

				asprintf (&del_this[0], "cluster.nfs.%016llx", some->lvid);
				del_cfg_obj (del_this);
				free (del_this[0]);
			}
		}

		list = some->next;
		free (some->oper);
		free (some);
	}
}

DEFUN	(system_audit_conf,
		system_audit_conf_cmd,
		"system auditconf",
		SYSCMDS
		"Audit configuration file for inconsistencies\n")
{
	char **lv_objs;
	char **slave_objs;
	char **nfs_objs;

	lv_objs = get_cfg_val ("cluster.lv.*");
	nfs_objs = get_cfg_val ("cluster.nfs.*");
	slave_objs = get_cfg_val ("slave.*");

	audit_conf (lv_objs, slave_objs, nfs_objs);

	free_cfg_val (lv_objs);
	free_cfg_val (slave_objs);
	free_cfg_val (nfs_objs);

	return CMD_SUCCESS;
}


void
install_system (priv_t myPriv, enum node_type node)
{
/*
	Organizing this switch statment such that the higher priority commands
	are first allows us to fall through and pick up the lower priority
	ones at no extra cost.
 */
	switch (myPriv) {
		case diagPriv:
		case fullPriv:
			install_element (node, &halt_cmd);
			install_element (node, &reboot_cmd);
			install_element (node, &total_reset_cmd);
			install_element (node, &reset_cmd);
			install_element (node, &reset_reset_cmd);
			install_element (node, &upgrade_cmd);
			install_element (node, &crash_cmd);
			install_element (node, &crash_auto_cmd);
			install_element (node, &system_coredir_cmd);
			install_element (node, &system_copy_all_cmd);
			install_element (node, &system_copy_config_cmd);
			install_element (node, &system_copy_init_cmd);
			install_element (node, &system_prom_upgrade_cmd);
			install_element (node, &system_config_restore_cmd);
			install_element (node, &system_config_reset_cmd);
			install_element (node, &system_audit_conf_cmd);
			install_element (node, &system_corevol_create_cmd);
			install_element (node, &system_corevol_delete_cmd);
		case changePriv:
			install_element (node, &ntp_broadcast_cmd);
			install_element (node, &ntp_disable_cmd);
			install_element (node, &ntp_server_cmd);
			install_element (node, &system_time_zone_cmd);
			install_element (node, &SetNodeName_cmd);
			install_element (node, &set_time_cmd);
		case addPriv:
		case viewPriv:
/*
	These commands are available in all modes
 */
			install_element (node, &version_cmd);
			install_element (node, &time_cmd);
			install_element (node, &time_cmd_v);
			install_element (node, &ping_cmd);
			install_element (node, &traceroute_cmd);
			install_element (node, &p_ping_cmd);
			install_element (node, &t_traceroute_cmd);
			install_element (node, &ShowNodeName_cmd);
			install_element (node, &show_cmd);
			install_element (node, &crash_show_cmd);
			install_element (node, &crashauto_show_cmd);
			install_element (node, &system_coredir_show_cmd);
	}
}
