/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name:
 *
 * RCS:  $Id: cmd_vol.c,v 1.54 2002/10/29 02:39:27 syen Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description:
 *
 * Created by:    Greg Seibert
 *
 * Date Created:
 *
 *-----------------------------------------------------------------
 */
#include <sys/types.h>
#include <pwd.h>
#include <readline/history.h>
#include <grp.h>
#include <unistd.h>

#include "zebra.h"
#include "command.h"
#include "log.h"
#include "vtysh.h"
#include "memory.h"

#include "nfxsh.h"
#include "nfx-incl.h"
#include "exports/anpssc-api.h"
#include "exports/ea-api.h"
#include "exports/evm-api.h"
#include "exports/fs-opt-api.h"
#include "exports/cfgrules-api.h"

#define VOL_COMMAND_STR  "Volume Commands\n"
#define VOL_CLI_MSG_LEN 6000
#define BYTES_IN_GB     (1024*1024*1024)
#define BYTES_IN_MB     (1024*1024)

// #define LINES_PER_PAGE  18
#define VOL_SUMMARY_RECS 14

/* nbEvmBlocks = nbMBytes * nbBytesInMBytes / nbBytesPerEvmBlock
 * 2^32 = maxValue * 2^10 * 2^10 / 2^9 => (maxValue < 2^21)
 * 2^64 = maxValue * 2^10 * 2^10 / 2^9 => (maxValue < 2^53)
 */
uint32  maxUserLimit_32 = 1 << 21;
uint64  maxUserLimit_64 = 1ULL << 53;

/*
 * Constants.
 */

char *raidStr[] =
{
    "Raid-0", // EVM_RAID0 = 0,      // stripping
    "Raid-1", // EVM_RAID1 = 1,      // mirror
    "Raid-01", // EVM_RAID01 = 2,     // stripe + mirror
    "Raid-3", // EVM_RAID3 = 3,      // raid 3
    "",
    "Raid-5", // EVM_RAID5 = 5,      // raid 5
    "Plain", // EVM_RAID_NONE = 6,  // no raid level, ie plain
    "Any" // EVM_RAID_ANY = 0xffff
};

char *fsStateStr[] =
{
    "Unknown",
    "Mounted",
    "Mounted-RO",
    "NotMounted",
    "Paused",
    "Mounting",
    "Unmounting",
    "Failing",
    "Mounted"
};

// Wait for EA to respond ()
static struct timeval waitEaTo = {300, 0};
static struct timeval evm_to = {60, 0};

/*
 * Data structures definition
 */

typedef struct
{
    char    buf[60000];  // optimize!!!
    uint32  len;
} evm_msg_t;

/*
 * Extern functions.
 */

extern void install_lvol (priv_t myPriv, enum node_type node);

static evm_volId_t get_newVolId(void);

static int32 isValidName(char *str)
{
    char curChar;
    int32  len = 0;
    int32  idx = 0;

    /*
     * A vol name must have len > 0, cannot start with 
     * non alpha-numeric character and can only have 
     * a-z, A-Z or 0-9
     */
    
    if (str == NULL)
        return 0;

    len = strlen(str);

    if (len >= VOL_MAX_NAME)
        return 0;

    for (idx=0; idx < len; idx++)
    {
        curChar = str[idx];

        if ((idx == 0) && !isalnum(curChar))
            return 0;

        if (!isalnum(str[idx]) && (curChar != '-')
                && (curChar != '_'))
        {
            return 0;
        }
    }

    return 1;
}

static int32 has_digits(char *str)
{
    int32  cnt;

    for (cnt=0; cnt < strlen(str); cnt++)
        if (!isdigit(str[cnt]))
            return 0;
    return 1;
}


static int64 display_vol_show_summary(ea_getFsysRsp_t *eaRsp)
{
    int32     cnt = 0, volCnt = 0;
    ea_fsysCfg_t   *fCfg;
    double  curr = 0.0;
    double  used = 0.0;
    double  avail = 0.0;
    double  pused = 0.0;
    double hardQ;
    uint32 softQ;
    char statusStr[24];

    printf("Volume Summary\n\n");

    printf("Name          Status      Current (MB)  HardQuota (MB)  Used (MB)     Avail (MB)    Usage   \n");
    printf("--------------------------------------------------------------------------------------------\n");
    while (cnt < eaRsp->fsysList.nb)
    {
        fCfg = (ea_fsysCfg_t *)&eaRsp->fsysList.fsys[cnt];
        curr = 0.0;
        used = 0.0;
        avail = 0.0;
        pused = 0.0;
        hardQ = 0.0;
        softQ = 0;

        // display only the regular volumes/filesystems
        if (fCfg->type != EVM_TYPE_FSYS)
        {
            cnt++;
            continue;
        }

        if (fCfg->currSize)
        {
            uint64  cSize, uSize, aSize;

            uSize = fCfg->usedSize*(uint64)EFS_BLOCK_SIZE;
            cSize = fCfg->currSize*(uint64)EVM_LBLK_SIZE;
            aSize = cSize - uSize;

            curr = (double)cSize/(double)BYTES_IN_MB;
            used = (double)uSize/(double)BYTES_IN_MB;
            avail = (double)aSize/(double)BYTES_IN_MB;
            pused = (double)((double)uSize*100/(double)cSize);
        }

        if (fCfg->maxSize)
        {
           uint64 mSize;

           mSize = fCfg->maxSize*(uint64)EVM_LBLK_SIZE;
           hardQ = (double)mSize/(double)BYTES_IN_MB;
           softQ = fCfg->softQuota;
        }

        if (fCfg->onlineStatus)
            strcpy(statusStr, fsStateStr[fCfg->fsState]);
        else
            strcpy(statusStr, "Offline");


        printf("%-12s  %-10s  %-12.2f  ", 
                    fCfg->fsysId.fsysName, statusStr, curr);

        if (hardQ)
        {
            printf("%-12.2f    ", hardQ);
        }
        else
        {
            printf("No Limit        ");
        }

        if (fCfg->onlineStatus && fCfg->fsState)
        {
            printf("%-12.2f  %-12.2f  %-5.2f%%", used, avail, pused);
        }
        else
        {
            printf("N/A           N/A           N/A");
        }

        printf("\n");

        volCnt++;
        if (volCnt == VOL_SUMMARY_RECS)
        {
            int32 c;

            printf ("---[more | q-Quit] :");
            c = getchar();
            if (c == 'q' || c == 'Q')
            {
                getchar();
                return 0;
            }
            else
                return fCfg->fsysId.volId.whole_thing;
        }
        cnt++;
    }

    return fCfg->fsysId.volId.whole_thing;
}


static void display_vol_detail(ea_getDetailFsysRsp_t *eaDRsp)
{
    ea_fsysDetail_t     *rsp = (ea_fsysDetail_t *)&eaDRsp->fsysDetail;
    evm_VolConfig_t     *evR = (evm_VolConfig_t *)&rsp->volCfg;
    char statusStr[24];
    int32   cnt = 0;
    double  curr = 0.0;
    double  used = 0.0;
    double  avail = 0.0;
    double  pused = 0.0;
    double  hardQ = 0.0;
    uint32  softQ = 0;
    double  mgrow = 0;
    double  llun = 0.0;
    double  mlun = 0.0;

    if (evR->currSize)
    {
        uint64  cSize, uSize, aSize;

        uSize = rsp->usedSize*(uint64)EFS_BLOCK_SIZE;
        cSize = evR->currSize*(uint64)EVM_LBLK_SIZE;
        aSize = cSize - uSize;

        curr = (double)cSize/(double)BYTES_IN_MB;
        used = (double)uSize/(double)BYTES_IN_MB;
        avail = (double)aSize/(double)BYTES_IN_MB;
        pused = (double)((double)uSize*100/(double)cSize);
    }

    if (evR->maxSize)
    {
        uint64 mSize;

        mSize = evR->maxSize*(uint64)EVM_LBLK_SIZE;
        hardQ = (double)mSize/(double)BYTES_IN_MB;
        softQ = evR->softQuota;
    }
    if (evR->minAutoGrow)
    {
        uint64 gSize;

        gSize = evR->minAutoGrow*(uint64)EVM_LBLK_SIZE;
        mgrow = (double)((double)gSize/(double)BYTES_IN_MB);
    }
    if (evR->attrs.minLunSize != EVM_DONT_CARE)
    {
         uint64  lsize;

         lsize = (uint64)evR->attrs.minLunSize*(uint64)EVM_LBLK_SIZE;
         llun = (double)lsize/(double)BYTES_IN_MB;
    }
    if (evR->attrs.maxLunSize != EVM_DONT_CARE)
    {
         uint64  lsize;

         lsize = (uint64)evR->attrs.maxLunSize*(uint64)EVM_LBLK_SIZE;
         mlun = (double)lsize/(double)BYTES_IN_MB;
    }

    if (evR->onLine)
        strcpy(statusStr, fsStateStr[rsp->fsState]);
    else
        strcpy(statusStr, "Offline");

    printf ("Volume Details\n");
    printf ("--------------\n\n");
    printf ("                 Name : %-s\n", evR->name);
    printf ("                   Id : 0x%llx\n", evR->id.whole_thing);
    switch (evR->type)
    {
    case EVM_TYPE_FSYS:
        printf ("                 Type : filesystem\n");
        break;
    case EVM_TYPE_COREVOL:
        printf ("                 Type : corevolume\n");
        break;
    case EVM_TYPE_MIRROR:
        printf ("                 Type : mirror\n");
        break;
    default:
        printf ("                 Type : unknown\n");
        break;
    } // switch
    printf ("               Status : %-s\n", statusStr);
    printf ("    Current Size (MB) : %-.2f\n", curr);

    if (evR->onLine && rsp->fsState)
        printf ("       Used Size (MB) : %-.2f\n", used);
    else
        printf ("       Used Size (MB) : N/A\n");

    if (evR->onLine && rsp->fsState)
        printf ("  Available Size (MB) : %-.2f\n", avail);
    else
        printf ("  Available Size (MB) : N/A\n");

    if (evR->onLine && rsp->fsState)
        printf ("            Usage (%%) : %-.2f %%\n", pused);
    else
        printf ("            Usage (%%) : N/A\n");

    if (hardQ)
    {
        printf ("      Hard Quota (MB) : %-.2f\n", hardQ);
        printf ("       Soft Quota (%%) : %d %%\n", softQ);
    }
    else
    {
        printf ("      Hard Quota (MB) : No Limit\n");
        printf ("       Soft Quota (%%) : N/A\n");
    }

    if (evR->onLine && rsp->fsState && 
                (rsp->highWaterMark != DISABLE_HIGHWATERMARK))
    {
        printf ("  High Water Mark (%%) : %d %%\n", rsp->highWaterMark);
    }
    else
        printf ("  High Water Mark (%%) : N/A\n");

    printf ("Min. Auto Growth (MB) : %-.2f\n", mgrow);

    printf ("               Vendor : %-s\n",
                (strlen(evR->attrs.vendorInfo[0].vendor)) ?
                                evR->attrs.vendorInfo[0].vendor :"Any");

    printf ("                Model : %-s\n",
                (strlen(evR->attrs.vendorInfo[0].model)) ?
                                evR->attrs.vendorInfo[0].model :"Any");

    printf ("           RAID Level : %-s\n",
                            (evR->attrs.raidLevel == EVM_RAID_NONE)?"Plain":
                            (evR->attrs.raidLevel == EVM_RAID0)?"Raid-0":
                            (evR->attrs.raidLevel == EVM_RAID1)?"Raid-1":
                            (evR->attrs.raidLevel == EVM_RAID01)?"Raid-01":
                            (evR->attrs.raidLevel == EVM_RAID3)?"Raid-3":
                            (evR->attrs.raidLevel == EVM_RAID5)?"Raid-5":
                            (evR->attrs.raidLevel == EVM_RAID_ANY)?"Any":
                            "Unknown");

    if (evR->attrs.minLunSize != EVM_DONT_CARE)
        printf ("   Min. Lun Size (MB) : %-.2f\n", llun);
    else
        printf ("   Min. Lun Size (MB) : No Min\n");

    if (evR->attrs.maxLunSize != EVM_DONT_CARE)
        printf ("   Max. Lun Size (MB) : %-.2f\n", mlun);
    else
        printf ("   Max. Lun Size (MB) : No Max\n");

    //printf ("              Stripes : %-d\n", evR->numStripes);
    //printf ("      Strip Size (KB) : %-d\n", evR->stripeSize/1000);
    //printf ("    Number of Mirrors : %-d\n", evR->numMirrors);

    printf("\n");
    printf ("LUN List (Num Luns: %d)\n", rsp->nbStrg);
    if (rsp->nbStrg > 0)
    {
        uint64   bSize = 0;
        double   _bSize = 0;

        printf ("Idx   Device Name                       Size (MB)\n");
        printf ("-------------------------------------------------\n");
        for (cnt = 0; cnt < rsp->nbStrg; cnt++)
        {
            bSize = rsp->strg[cnt].au.auSize*(uint64)EVM_LBLK_SIZE;
            _bSize = (double)bSize/(double)BYTES_IN_MB;
            printf ("%-4d  %-32s  %.2f\n", cnt+1,
                strlen(rsp->strg[cnt].name)? rsp->strg[cnt].name:"Missing", 
                    _bSize);
        }
        printf ("-------------------------------------------------\n");
    }
    else
        printf ("None\n");

    return;
}

/*------------------------------------------------------------------------------
 * Name         : get_newVolId
 *
 * Description  : Get a new volume id
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 26-JUL-2002
 *
 *------------------------------------------------------------------------------
 */

static evm_volId_t
get_newVolId(void)

{
    evm_MsgGetNewVolId_t    getReq;
    evm_msg_t               msg;
    evm_MsgStat_t *         getRsp = (evm_MsgStat_t *) msg.buf;

    bzero((void *) &getReq, sizeof(evm_MsgGetNewVolId_t));
    bzero((void *) &msg, sizeof(msg));

    getReq.hdr.msgType = EVM_VOL_GET_NEWVOLID;
    getReq.hdr.clientPrivate = getpid();

    msg.len = sendAgileMsg(-1, &getReq, sizeof(evm_MsgGetNewVolId_t),
                           "evm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                           &evm_to, msg.buf, sizeof(msg.buf));

    if ((msg.len != sizeof(evm_MsgStat_t)) || (getRsp->stat != NFX_OK))
    {
        getRsp->payload.lVolId.whole_thing = -1;
    }

    return(getRsp->payload.lVolId);
}

/*
 *------------------------------------------------------------------------------
 * Name         : vol_create
 *
 * Description  : Create a volume and file system.
 *
 * Created by   : Isabelle Cnudde
 * Modified by  : Nitin Jain
 *
 * Date Created : 06-NOV-2001
 *                28-FEB-2002
 *------------------------------------------------------------------------------
 */

#define VOL_CREATE_CMD_OPTIONS "    [-q HARDQUOTA] [-s SOFTQUOTA] [-g MINAUTHGROW] [-h HIGHWATERMARK] [-l MINLUNSIZE] [-u MAXLUNSIZE] [-r RAIDLEVEL] [-R RAIDCONTROLLER] [-d MAKE] [-m MODEL] [-n CLUSTERNODE]\n"

DEFUN(vol_create,
      vol_create_cmd,
      "volume create VOLNAME [OPTIONS ...]",
      VOL_COMMAND_STR
      "Create a new logical volume.\n"
      "Name of the new volume\n"
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS
      VOL_CREATE_CMD_OPTIONS)
{
    ea_addFsysReq_t     eaReq;              
    ea_addFsysRsp_t     eaRsp;
    int                 c;
    extern int          optind;
    int32               rspLen;
    uint64              readVal;
    pid_t           twid;

    /* 
     * Prepare EA request
     */
	memset(&eaReq, 0, sizeof(ea_addFsysReq_t));
	memset(&eaRsp, 0, sizeof(ea_addFsysRsp_t));

    eaReq.hdr.msgType = EA_ADD_FSYS_REQ;
    eaReq.hdr.userInfo = getpid();

    strncpy(eaReq.fsysId.fsysName, argv[0], VOL_MAX_NAME);

    if (!isValidName(eaReq.fsysId.fsysName))
    {
        printf ("Invalid volume name.\n");
        return CMD_ERR_FAILED;
    }
#ifdef NEWVOLID
    eaReq.fsysId.volId = get_newVolId();
    if (eaReq.fsysId.volId.whole_thing == -1)
    {
        printf("Cannot get a new volume id.\n");
        return(CMD_ERR_FAILED);
    }
#else
    eaReq.fsysId.volId.whole_thing = 0;
#endif 
    eaReq.volSize = 0;
    eaReq.maxSize = 0;
    eaReq.softQuota = 0;
    eaReq.minAutoGrow = DEFAULT_MIN_AUTOGROW;
    eaReq.highWaterMark = DISABLE_HIGHWATERMARK;
    eaReq.attr.raidLevel = EVM_RAID_ANY;
    eaReq.attr.minLunSize = EVM_DONT_CARE;
    eaReq.attr.maxLunSize = EVM_DONT_CARE;
    eaReq.attr.num = 1;
    strcpy(eaReq.attr.vendorInfo[0].vendor, "");
    strcpy(eaReq.attr.vendorInfo[0].model, "");
    strcpy(eaReq.attr.raidCtrl, "");
    eaReq.nim = 1;    // N-NIM in simulation

    optind = 0;
    while ((c = getopt (argc, argv, "q:g:s:h:l:u:r:d:m:R:n:")) != -1)
    {
        switch (c)
        {
            case 'q':
                if (!has_digits(optarg))
                {
                    printf ("Hardquota not a number\n");
                    return CMD_ERR_FAILED;
                }
                readVal = strtoq(optarg, NULL, 0);
                if ((readVal == QUAD_MIN) || (readVal == QUAD_MAX) ||
                    (readVal < 0) || (readVal >= maxUserLimit_64))
                {
                    printf ("Invalid number range\n");
                    return CMD_ERR_FAILED;
                }
                eaReq.maxSize = readVal * (BYTES_IN_MB / EVM_LBLK_SIZE);
                //printf("eaReq.hardQuota = %lld\n", eaReq.maxSize);
            break;

            case 'g':
                if (!has_digits(optarg))
                {
                    printf ("Min auto growth not a number\n");
                    return CMD_ERR_FAILED;
                }
                readVal = strtoq(optarg, NULL, 0);
                if ((readVal == QUAD_MIN) || (readVal == QUAD_MAX) ||
                    (readVal < 0) || (readVal >= maxUserLimit_64))
                {
                    printf ("Invalid number range\n");
                    return CMD_ERR_FAILED;
                }
                if (readVal == 0)
                {
                    printf("Min auto growth has to be non-zero\n");
                    return CMD_ERR_FAILED;
                }
                eaReq.minAutoGrow = readVal * (BYTES_IN_MB / EVM_LBLK_SIZE);
                //printf("eaReq.minAutoGrow = %d\n", eaReq.minAutoGrow);
            break;

            case 's':
                eaReq.softQuota = atoi(optarg);
                if (!has_digits(optarg) ||
                    (eaReq.softQuota < 0) ||
                            (eaReq.softQuota > 100))
                {
                    printf ("Invalid Softquota value\n");
                    return(CMD_ERR_FAILED);
                }
                //printf("eaReq.softQuota = %d\n", eaReq.softQuota);
                break;

            case 'h':
                eaReq.highWaterMark = atoi(optarg);
                if (!has_digits(optarg))
                {
                    printf ("High water mark not a number\n");
                    return CMD_ERR_FAILED;
                }
                if (((eaReq.highWaterMark < MIN_HIGHWATERMARK) &&
                     (eaReq.highWaterMark != DISABLE_HIGHWATERMARK)) ||
                     (eaReq.highWaterMark > MAX_HIGHWATERMARK))
                {
                    printf ("Valid High water mark values are 0, 25-100.\n");
                    return CMD_ERR_FAILED;
                }
                //printf("eaReq.highWaterMark = %d\n", eaReq.highWaterMark);
            break;

            case 'l':
                if (!has_digits(optarg))
                {
                    printf ("Min Lun size not a number\n");
                    return CMD_ERR_FAILED;
                }
                readVal = strtol(optarg, NULL, 0);
                if ((readVal == LONG_MIN) || (readVal == LONG_MAX) ||
                    (readVal < 0) || (readVal >= maxUserLimit_32))
                {
                    printf ("Invalid number range\n");
                    return CMD_ERR_FAILED;
                }
                eaReq.attr.minLunSize = readVal * (BYTES_IN_MB/EVM_LBLK_SIZE);
                //printf("eaReq.attr.minLunSize = %d\n", eaReq.attr.minLunSize);
            break;

            case 'u':
                if (!has_digits(optarg))
                {
                    printf ("Max Lun size not a number\n");
                    return CMD_ERR_FAILED;
                }
                readVal = strtol(optarg, NULL, 0);
                if ((readVal == LONG_MIN) || (readVal == LONG_MAX) ||
                    (readVal < 0) || (readVal >= maxUserLimit_32))
                {
                    printf ("Invalid number range\n");
                    return CMD_ERR_FAILED;
                }
                eaReq.attr.maxLunSize = readVal * (BYTES_IN_MB/EVM_LBLK_SIZE);
                //printf("eaReq.attr.maxLunSize = %d\n", eaReq.attr.maxLunSize);
            break;

            case 'r':
                if (strcasecmp(optarg, "Raid-0") == 0)
                    eaReq.attr.raidLevel = EVM_RAID0;
                else if (strcasecmp(optarg, "Raid-1") == 0)
                    eaReq.attr.raidLevel = EVM_RAID1;
                else if (strcasecmp(optarg, "Raid-01") == 0)
                    eaReq.attr.raidLevel = EVM_RAID01;
                else if (strcasecmp(optarg, "Raid-3") == 0)
                    eaReq.attr.raidLevel = EVM_RAID3;
                else if (strcasecmp(optarg, "Raid-5") == 0)
                    eaReq.attr.raidLevel = EVM_RAID5;
#if 0
                else if (strcasecmp(optarg, "Plain") == 0)
                    eaReq.attr.raidLevel = EVM_RAID_NONE;
#endif
                else if (strcasecmp(optarg, "Any") == 0)
                    eaReq.attr.raidLevel = EVM_RAID_ANY;
                else
                {
                    printf ("Invalid Raid Level\n");
                    return (CMD_ERR_FAILED);
                }
                //printf("eaReq.attr.raidLevel = %d\n", eaReq.attr.raidLevel);
            break;

            case 'd':
                eaReq.attr.flags |= EVM_SAME_VENDOR;
                strncpy(eaReq.attr.vendorInfo[0].vendor, optarg, 
                                        EVM_VENDOR_NAME_LEN);
                //printf("vendor = %s\n", eaReq.attr.vendorInfo[0].vendor);
                break;

            case 'm':
                eaReq.attr.flags |= EVM_SAME_MODEL;
                strncpy(eaReq.attr.vendorInfo[0].model, optarg, 
                                        EVM_MODEL_NAME_LEN);
                //printf("model = %s\n", eaReq.attr.vendorInfo[0].model);
                break;

            case 'R':
                strncpy(eaReq.attr.raidCtrl, optarg, SDM_DEV_NAME_LEN);
                //printf("raidCtrl = %s\n", eaReq.attr.raidCtrl);
                break;

            case 'n':
                strncpy(eaReq.hdr.node, optarg, EVM_MAX_NODE_NAME);
                //printf("Destination node = %s\n", eaReq.hdr.node);
            break;

            case '?':
                if (isprint (optopt))
                    fprintf (stderr, "Unknown option `-%c'.\n", optopt);
                else
                    fprintf (stderr, "Unknown option character `\\x%x'.\n", optopt);
            return CMD_ERR_FAILED;

            default:
                fprintf (stderr, "Unknown option `-%c'.\n", optopt);
            return CMD_ERR_FAILED;
        } // switch
    } // while

    if (optind != argc)
    {
        printf("Syntax error\n");
        return CMD_ERR_FAILED;
    }

    /*
     * Send message to EA.
     */
    twid = start_twiddle ();
    rspLen = sendAgileMsg(-1, &eaReq, sizeof(ea_addFsysReq_t),
                          "ea", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                          &waitEaTo, &eaRsp, sizeof(ea_addFsysRsp_t));

    /*
     * Receive response from EA.
     * Display result to operator.
     */
    stop_twiddle (twid);
    if (rspLen < 0)
    {
        perror("Cannot send request...");
        return(CMD_ERR_FAILED);
    }

    if (rspLen < sizeof(ea_addFsysRsp_t))
    {
        fprintf(stderr, "No response from - ea -\n");
    }
    else if (eaRsp.hdr.msgType != EA_ADD_FSYS_RSP)
    {
        fprintf(stderr, "Incorrect message type from - ea -\n");
    }
    else if (eaRsp.hdr.userInfo != getpid())
    {
        fprintf(stderr, "Response from ea : Not for our request\n");
    }
    else
        printf("%s\n", eaRsp.output);

    return(CMD_SUCCESS);
}

/*
 *------------------------------------------------------------------------------
 * Name         : vol_delete
 *
 * Description  : Delete the volume and file system.
 *
 * Created by   : Isabelle Cnudde
 * Modified by  : Nitin Jain
 *
 * Date Created : 06-NOV-2001
 *                28-FEB-2002
 *------------------------------------------------------------------------------
 */

DEFUN(vol_delete,
      vol_delete_cmd,
      "volume delete VOLNAME",
      VOL_COMMAND_STR
      "Delete a volume and its associated file system.\n"
      "Name of an existing volume\n"
      "Node where the volume exists\n")
{
    ea_delFsysReq_t     eaReq;
    ea_delFsysRsp_t     eaRsp;
    int32               rspLen;

    pid_t               twid;

	memset(&eaReq, 0, sizeof(ea_delFsysReq_t));
	memset(&eaRsp, 0, sizeof(ea_delFsysRsp_t));

    /*
     * Fill the request with operator input.
     */
    eaReq.hdr.msgType = EA_DEL_FSYS_REQ;
    eaReq.hdr.userInfo = getpid();
    strncpy(eaReq.fsysId.fsysName, argv[0], VOL_MAX_NAME);

    /*
     * Send message to EA.
     */
    twid = start_twiddle();
    rspLen = sendAgileMsg(-1, &eaReq, sizeof(ea_delFsysReq_t),
                          "ea", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                          &waitEaTo, &eaRsp, sizeof(ea_delFsysRsp_t));

    /*
     * Receive response from EA.
     * Display result to operator.
     */
    if (rspLen < 0)
    {
        perror("Cannot send request...");
        return(CMD_ERR_FAILED);
    }

    if (rspLen < sizeof(ea_delFsysRsp_t))
    {
        stop_twiddle (twid);
        fprintf(stderr, "No response from - ea -\n");
    }
    // NFX_ASSERT(eaMsg.len >= sizeof(ea_msgHdr_t));
    else if (eaRsp.hdr.msgType != EA_DEL_FSYS_RSP)
    {
        stop_twiddle (twid);
        fprintf(stderr, "Incorrect message type from - ea -\n");
    }
    // NFX_ASSERT(eaRsp->hdr.msgType == EA_DEL_FSYS_RSP);
    else {
		// no need to wait anymore as EA's doing that now...
        // sleep (5);  // Wait for ccmd to catch up
        stop_twiddle (twid);
        printf("%s\n", eaRsp.output);
    }

    return(CMD_SUCCESS);
}

/*
 *------------------------------------------------------------------------------
 * Name         : vol_show
 *
 * Description  : Show the volume and file system.
 *
 * Created by   : Isabelle Cnudde
 * Modified by  : Nitin Jain
 *
 * Date Created : 21-NOV-2001
 *                28-FEB-2002
 *------------------------------------------------------------------------------
 */

DEFUN(vol_show,
      vol_show_cmd,
      "volume show [VOLNAME]",
      VOL_COMMAND_STR
      "Show volume associated with VOLNAME.\n"
      "Name of an existing volume\n")
{
    uint32 rspSize;
    uint32 vol_name = 0;
    int32  rcvLen;

    if (argc == 0)
        vol_name = 0;

    else if (argc == 1)
        vol_name = 1;
    else
    {
        printf ("Usage :\n\tvolume show [VOLNAME]\n");
        return CMD_SUCCESS;
    }

    /*
     * Fill the request with operator input.
     */
    if (!vol_name)
    {
    	ea_getFsysReq_t eaReq;
    	ea_getFsysRsp_t *eaRsp;
        int32 getNext = 1;
        int64 volId = 0;

		memset(&eaReq, 0, sizeof(ea_getFsysReq_t));

        rspSize = sizeof(ea_getFsysRsp_t) +
                        (VOL_SUMMARY_RECS - 1)*sizeof(ea_fsysCfg_t);
		eaRsp = (ea_getFsysRsp_t *) malloc(rspSize);

        /* 
         * Prepare EA request
         */
        eaReq.hdr.msgType = EA_GET_FSYS_REQ;
        eaReq.hdr.userInfo = getpid();
        while (getNext)
        {
            eaReq.reqType = EA_REQ_HAS_FSYS_INIT;
            eaReq.num     = VOL_SUMMARY_RECS;

            /*
             * Send message to EA.
             */
			memset(eaRsp, 0, rspSize);

            rcvLen = sendAgileMsg(-1, &eaReq, sizeof(ea_getFsysReq_t),
                                 "ea", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                                  &waitEaTo, eaRsp, rspSize);

            if (rcvLen < 0)
            {
                perror("Cannot send request...");
                return(CMD_ERR_FAILED);
            }

            if (rcvLen < sizeof (ea_getFsysRsp_t))
            {
                fprintf(stderr, "No response from - ea -\n");
            }
            else if (eaRsp->hdr.msgType != EA_GET_FSYS_RSP)
                fprintf(stderr, "Incorrect message type from - ea -\n");
            else if (eaRsp->result == NFX_ERR)
                fprintf(stderr, "%s\n", eaRsp->output);
            else if (eaRsp->fsysList.nb > 0)
                volId = display_vol_show_summary(eaRsp);

            if (volId == 0)
                getNext = 0;
            else
                eaReq.fsysId[0].volId.whole_thing = volId;

            if (eaRsp->fsysList.nb < VOL_SUMMARY_RECS)
                getNext = 0;
        }

		free(eaRsp);
    }
    else
    {
        /* Get detailed information */
		ea_getDetailFsysReq_t eaReq;
		ea_getDetailFsysRsp_t *eaRsp;
        evm_msg_t eaMsg;

		memset(&eaReq, 0, sizeof(ea_getDetailFsysReq_t));

        memset(eaMsg.buf, 0, sizeof(eaMsg.buf));
        eaRsp = (ea_getDetailFsysRsp_t *) eaMsg.buf;

        /* 
         * Prepare EA request
         */
        eaReq.hdr.msgType = EA_GET_DETAIL_FSYS_REQ;
        eaReq.hdr.userInfo = getpid();
        strncpy(eaReq.fsysId.fsysName, argv[0], VOL_MAX_NAME); // Volume name

        /*
         * Send message to EA.
         */
        rcvLen = sendAgileMsg(-1, &eaReq, sizeof(ea_getDetailFsysReq_t),
                             "ea", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                              &waitEaTo, eaRsp, sizeof(eaMsg.buf));

        if (rcvLen < 0)
        {
            perror("Cannot send request...");
            return(CMD_ERR_FAILED);
        }

        if (rcvLen < sizeof(ea_getDetailFsysReq_t))
        {
            fprintf(stderr, "No response from - ea -\n");
        }
        else if (eaRsp->hdr.msgType != EA_GET_DETAIL_FSYS_RSP)
            fprintf(stderr, "Incorrect message type from - ea -\n");
        else if (eaRsp->result == NFX_ERR)
            fprintf(stderr, "%s\n", eaRsp->output);
        else
            display_vol_detail(eaRsp);
    }

    return(CMD_SUCCESS);
}

/*
 *------------------------------------------------------------------------------
 * Name         : vol_modify
 *
 * Description  : Modify the volume.
 *
 * Created by   : Isabelle Cnudde
 * Modified by  : Nitin Jain
 *
 * Date Created : 06-NOV-2001
 *                28-FEB-2002
 *------------------------------------------------------------------------------
 */

#define VOL_MODIFY_CMD_OPTIONS "[-q HARDQUOTA] [-s SOFTQUOTA] [-g MINAUTHGROW] [-h HIGHWATERMARK] [-p NEWVOLNAME]\n"

DEFUN(vol_modify,
      vol_modify_cmd,
      "volume modify VOLNAME [OPTIONS ...]",
      VOL_COMMAND_STR
      "Modify a volume.\n"
      "Name of an existing volume\n"
      VOL_MODIFY_CMD_OPTIONS
      VOL_MODIFY_CMD_OPTIONS
      VOL_MODIFY_CMD_OPTIONS
      VOL_MODIFY_CMD_OPTIONS
      VOL_MODIFY_CMD_OPTIONS
      VOL_MODIFY_CMD_OPTIONS
      VOL_MODIFY_CMD_OPTIONS
      VOL_MODIFY_CMD_OPTIONS
      VOL_MODIFY_CMD_OPTIONS
      VOL_MODIFY_CMD_OPTIONS
      VOL_MODIFY_CMD_OPTIONS
      VOL_MODIFY_CMD_OPTIONS)
{
    ea_modFsysReq_t     eaReq;              
    ea_modFsysRsp_t     eaRsp;
    int                 c;
    extern int          optind;
    int32               rspLen;
    uint64              readVal;

	memset(&eaReq, 0, sizeof(ea_modFsysReq_t));
	memset(&eaRsp, 0, sizeof(ea_modFsysRsp_t));

    /* 
     * Prepare EA request
     */
    eaReq.hdr.msgType = EA_MOD_FSYS_REQ;
    eaReq.hdr.userInfo = getpid();

    strncpy(eaReq.fsysId.fsysName, argv[0], VOL_MAX_NAME);

    optind = 0;
    while ((c = getopt (argc, argv, "q:s:g:h:p:")) != -1)
    {
        switch (c)
        {
            case 'q':
                if (!has_digits(optarg))
                {
                    printf ("Hardquota not a number\n");
                    return CMD_ERR_FAILED;
                }
                SET_FSYS_HQUOTA(eaReq.mask);
                readVal = strtoq(optarg, NULL, 0);
                if ((readVal == QUAD_MIN) || (readVal == QUAD_MAX) ||
                    (readVal < 0) || (readVal >= maxUserLimit_64))
                {
                    printf ("Invalid number range\n");
                    return CMD_ERR_FAILED;
                }
                eaReq.hardQuota = readVal * (BYTES_IN_MB/EVM_LBLK_SIZE);
                //printf("eaReq.hardQuota = %lld\n", eaReq.hardQuota);
            break;

            case 's':
                eaReq.softQuota = atoi(optarg);
                if (!has_digits(optarg) ||
                    (eaReq.softQuota < 0) ||
                    (eaReq.softQuota > 100))
                {
                    printf ("Invalid Softquota value\n");
                    return(CMD_ERR_FAILED);
                }
                SET_FSYS_SQUOTA(eaReq.mask);
                //printf("eaReq.softQuota = %d\n", eaReq.softQuota);
                break;

            case 'g':
                if (!has_digits(optarg))
                {
                    printf ("Min auto growth not a number\n");
                    return CMD_ERR_FAILED;
                }
                SET_FSYS_MINAUTOGROW(eaReq.mask);
                readVal = strtoq(optarg, NULL, 0);
                if ((readVal == QUAD_MIN) || (readVal == QUAD_MAX) ||
                    (readVal < 0) || (readVal >= maxUserLimit_64))
                {
                    printf ("Invalid number range\n");
                    return CMD_ERR_FAILED;
                }
                if (readVal == 0)
                {
                    printf("Min auto growth has to be non-zero.\n");
                    return CMD_ERR_FAILED;
                }
                eaReq.minAutoGrow = readVal * (BYTES_IN_MB/EVM_LBLK_SIZE);
                //printf("eaReq.minAutoGrow = %d\n", eaReq.minAutoGrow);
            break;

            case 'h':
                eaReq.highWaterMark = atoi(optarg);
                if (!has_digits(optarg))
                {
                    printf ("High water mark not a number\n");
                    return CMD_ERR_FAILED;
                }
                if (((eaReq.highWaterMark < MIN_HIGHWATERMARK) &&
                     (eaReq.highWaterMark != DISABLE_HIGHWATERMARK)) ||
                    (eaReq.highWaterMark > MAX_HIGHWATERMARK))
                {
                    printf ("Valid High water mark values are 0, 25-100.\n");
                    return CMD_ERR_FAILED;
                }
                SET_FSYS_HIGHWATERMARK(eaReq.mask);
                //printf("eaReq.highWaterMark = %d\n", eaReq.highWaterMark);
            break;

            case 'p':
                SET_FSYS_NAME(eaReq.mask);
                strncpy(eaReq.newFsysName, optarg, VOL_MAX_NAME);
                if (!isValidName(eaReq.newFsysName))
                {
                    printf ("Invalid volume name\n");
                    return CMD_ERR_FAILED;
                }
                //printf("eaReq.newFsysName = %s\n", eaReq.newFsysName);
                break;

            default:
                break;
        } // switch
    } // while

    if (optind != argc)
    {
        printf("Syntax error\n");
        return CMD_ERR_FAILED;
    }

    /*
     * Send message to EA.
     */
    rspLen = sendAgileMsg(-1, &eaReq, sizeof(ea_modFsysReq_t),
                          "ea", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                          &waitEaTo, &eaRsp, sizeof(ea_modFsysRsp_t));

    /*
     * Receive response from EA.
     * Display result to operator.
     */
    if (rspLen < 0)
    {
        perror("Cannot send request...");
        return(CMD_ERR_FAILED);
    }

    if (rspLen < sizeof(ea_modFsysRsp_t))
    {
        fprintf(stderr, "No response from - ea -\n");
    }
    else if (eaRsp.hdr.msgType != EA_MOD_FSYS_RSP)
    {
        fprintf(stderr, "Incorrect message type from - ea -\n");
    }
    else if (eaRsp.hdr.userInfo != getpid())
    {
        fprintf(stderr, "Response from ea : Not for our request\n");
    }
    else
        printf("%s\n", eaRsp.output);

    return(CMD_SUCCESS);
}

/*
 *------------------------------------------------------------------------------
 * Name         : vol_offline
 *
 * Description  : Move a file system out of the NFX.
 *
 * Created by   : Isabelle Cnudde
 * Modified by  : Nitin Jain
 *
 * Date Created : 07-DEC-2001
 *                28-FEB-2002
 *------------------------------------------------------------------------------
 */


DEFUN(vol_offline,
      vol_offline_cmd,
      "volume offline VOLNAME",
      VOL_COMMAND_STR
      "Bring offline the specified volume.\n"
      "Name of an existing volume\n")
{
    ea_movOutFsysReq_t  eaReq;
    ea_movOutFsysRsp_t  eaRsp;
    int32               rspLen;

	memset(&eaReq, 0, sizeof(ea_movOutFsysReq_t));
	memset(&eaRsp, 0, sizeof(ea_movOutFsysRsp_t));

    /*
     * Fill the request with operator input.
     */
    eaReq.hdr.msgType = EA_MOVOUT_FSYS_REQ;
    eaReq.hdr.userInfo = getpid();
    strncpy(eaReq.fsysId.fsysName, argv[0], VOL_MAX_NAME);

    /*
     * Send message to EA.
     */
    rspLen = sendAgileMsg(-1, &eaReq, sizeof(ea_movOutFsysReq_t),
                          "ea", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                          &waitEaTo, &eaRsp, sizeof(ea_movOutFsysRsp_t));

    /*
     * Receive response from EA.
     * Display result to operator.
     */
    if (rspLen < 0)
    {
        perror("Cannot send request...");
        return(CMD_ERR_FAILED);
    }
    if (rspLen < sizeof(ea_movOutFsysRsp_t))
    {
        fprintf(stderr, "No response from - ea -\n");
    }
    // NFX_ASSERT(eaMsg.len >= sizeof(ea_msgHdr_t));
    else if (eaRsp.hdr.msgType != EA_MOVOUT_FSYS_RSP)
    {
        fprintf(stderr, "Incorrect message type from - ea -\n");
    }
    // NFX_ASSERT(eaRsp->hdr.msgType == EA_MOVOUT_FSYS_RSP);
    else
        printf("%s\n", eaRsp.output);

    return(CMD_SUCCESS);
}

/*
 *------------------------------------------------------------------------------
 * Name         : vol_online
 *
 * Description  : Move a file system in the NFX.
 *
 * Created by   : Isabelle Cnudde
 * Modified by  : Nitin Jain
 *
 * Date Created : 07-DEC-2001
 *                28-FEB-2002
 *------------------------------------------------------------------------------
 */

DEFUN(vol_online,
      vol_online_cmd,
      "volume online VOLNAME",
      VOL_COMMAND_STR
      "Bring the specified volume online.\n"
      "Name of an existing volume\n")
{
    ea_movInFsysReq_t   eaReq;
    ea_movInFsysRsp_t   eaRsp;
    int32               rspLen;

	memset(&eaReq, 0, sizeof(ea_movInFsysReq_t));
	memset(&eaRsp, 0, sizeof(ea_movInFsysRsp_t));

    /*
     * Fill the request with operator input.
     */
    eaReq.hdr.msgType = EA_MOVIN_FSYS_REQ;
    eaReq.hdr.userInfo = getpid();
    strncpy(eaReq.fsysId.fsysName, argv[0], VOL_MAX_NAME);
    eaReq.nim = 1;   // This is a future option

    /* Send message to EA.
     */
    rspLen = sendAgileMsg(-1, &eaReq, sizeof(ea_movInFsysReq_t),
                          "ea", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
                          &waitEaTo, &eaRsp, sizeof(ea_movInFsysRsp_t));

    /* Receive response from EA.
     * Display result to operator.
     */
    if (rspLen < 0)
    {
        perror("Cannot send request...");
        return(CMD_ERR_FAILED);
    }

    if (rspLen < sizeof(ea_movInFsysRsp_t))
    {
        fprintf(stderr, "No response from - ea -\n");
    }
    // NFX_ASSERT(eaMsg.len >= sizeof(ea_msgHdr_t));
    else if (eaRsp.hdr.msgType != EA_MOVIN_FSYS_RSP)
    {
        fprintf(stderr, "Incorrect message type from - ea -\n");
    }
    // NFX_ASSERT(eaRsp.hdr.msgType == EA_MOVIN_FSYS_RSP);
    else
        printf("%s\n", eaRsp.output);

    return(CMD_SUCCESS);
}

void
install_vol (priv_t myPriv, enum node_type node)
{
    switch (myPriv)
    {
        case diagPriv:
             install_lvol(diagPriv, node);
        case fullPriv:
             install_element(node, &vol_delete_cmd);
             install_element(node, &vol_offline_cmd);
             install_element(node, &vol_online_cmd);
        case changePriv:
             install_element(node, &vol_modify_cmd);
        case addPriv:
             install_element(node, &vol_create_cmd);
        case viewPriv:
             install_element(node, &vol_show_cmd);
    }
    return;
}
