/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: menu.c
 *
 * RCS:	 $Id: menu.c,v 1.37 2002/10/29 02:18:13 syen Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: This code was adapted from the GPL Zebra package for use
 *				as the NFX CLI.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 23 Oct 2001
 *
 *-----------------------------------------------------------------
 */
#include <sys/types.h>
#include <sys/wait.h>
#include <pwd.h>
#include <readline/history.h>
#include <grp.h>
#include <unistd.h>

#include "zebra.h"
#include "command.h"
#include "log.h"
#include "vtysh.h"
#include "memory.h"

#include "nfxsh.h"

#include "nfx-defs.h"

#include "exports/opt-api.h"

// These are the group names associated with each of the admin privledges

char *groupNames[] = { "nfxview", 
                       "nfxadd", 
                       "nfxchang", 
                       "nfxfull", 
                       "nfxdiags"};

static
struct cmd_node viewPrivNode = 
{
	VIEW_PRIV,
	"%s view> ",
};

static
struct cmd_node addPrivNode = 
{
	ADD_PRIV,
	"%s add> ",
};

static
struct cmd_node changePrivNode = 
{
	CHANGE_PRIV,
	"%s change> ",
};

static
struct cmd_node fullPrivNode = 
{
	FULL_PRIV,
	"%s full> ",
};

static
struct cmd_node diagPrivNode = 
{
	DIAG_PRIV,
	"%s diag> ",
};

/* Host information structure. */
extern struct host host;


DEFUN	(vtysh_exit_all,
		 vtysh_exit_all_cmd,
		 "exit",
		 "Exit CLI shell\n")
{
	return vtysh_exit (vty);
}


	
DEFUN (show_history,
	   show_history_cmd,
	   "history",
	   "Display the session command history\n")
{
	HIST_ENTRY **lst = history_list();

	while (*lst) {
		printf ("%s\n", (*lst)->line);
		lst++;
	}

	return CMD_SUCCESS;
}

/* Help display function for all node. */
DEFUN (config_list,
	   config_list_cmd,
	   "list",
	   "Print command list\n")
{
	int i;
	struct cmd_node *cnode = vector_slot (cmdvec, vty->node);
	struct cmd_element *cmd;

	for (i = 0; i < vector_max (cnode->cmd_vector); i++)
		if ((cmd = vector_slot (cnode->cmd_vector, i)) != NULL)
			vty_out (vty, "	 %s%s", cmd->string,
					 VTY_NEWLINE);
	return CMD_SUCCESS;
}

void
vtysh_install_default (enum node_type node)
{
	install_element (node, &config_list_cmd);
	install_element (node, &vtysh_exit_all_cmd);
	install_element (node, &show_history_cmd);
}

/*-----------------------------------------------------------------
 * Name: uid manipulation things
 *
 * Description: We need to be able to switch back and forth between
 *              root and user id
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 26 Feb 2002
 *
 *-----------------------------------------------------------------
 */
static uid_t ruid;

static int
progIsSetUID()
{
	if (geteuid()) {
		return 0;
	}
	ruid = getuid();
	setuid(0);
	seteuid(ruid);

	return 1;
}

extern void
become_root()
{
	seteuid(0);
}

extern void
unbecome_root()
{
	seteuid(ruid);
}

static char *thumbs[] = {"/", "-", "\\", "|"};

extern pid_t
start_twiddle (void)
{
	pid_t twid = fork();
	int i;
	int num_thumbs = sizeof (thumbs) / sizeof (*thumbs);

	if (twid) {
		return twid;
	}
	
	signal (SIGUSR2, exit);

	printf ("  "); fflush (stdout);

	for (i = 0;; i = (i + 1) % num_thumbs) {
		printf ("%s", thumbs[i]); fflush (stdout);
		usleep (300000);
	}
}

extern void
stop_twiddle (pid_t twid)
{
	int stat;

	kill (twid, SIGUSR2);
	waitpid (twid, &stat, 0);
	printf ("\n");
}

/*-----------------------------------------------------------------
 * Name: init_menu
 *
 * Description: This is where the initial menu is constructed
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 23 Oct 2001
 *
 *-----------------------------------------------------------------
 */
extern void
init_menu()
{
	gid_t grList[NUMPRIVS];
	gid_t mygid;
	struct group *grInfo;
	int i;
	priv_t myPriv;
	priv_t realPriv;
	char **clus_cfg;
/*
	Collect the group ids for our administrator groups
 */
	for (i = 0; i < NUMPRIVS; i++) {
		grInfo = getgrnam (groupNames[i]);
		if (!grInfo) {
			fprintf (stderr, "System config error: missing admin group %s\n",
			groupNames[i]);
			exit(1);
		}
		grList[i] = grInfo->gr_gid;
	}
	mygid = getgid();	// This user is what?
/*
	Make sure this user is authorized to admin by being in one of the groups.
	We always allow root diag access.
 */
	if (!getuid()) {
		realPriv = myPriv = diagPriv;
	}
	else {
		for (i = 0; (mygid != grList[i]) && (i < NUMPRIVS); i++);

		if (i == NUMPRIVS) {
			fprintf (stderr, "User not authorized to administrate %s\n",
			         PRODUCT_NAME);
			exit(1);
		}
		realPriv = (priv_t)i;

		clus_cfg = get_cfg_val ("cluster.name");
		if (clus_cfg) {
			free_cfg_val (clus_cfg);
			clus_cfg = get_cfg_val ("node.type");
			if (clus_cfg && strcasecmp ("pcc", clus_cfg[1])) {
				myPriv = viewPriv;
				fprintf (stdout, 
				         "Only view commands available on non-PCC nodes\n");
			}
			else {
				myPriv = realPriv;
			}
			free_cfg_val (clus_cfg);
		}
		else {
			myPriv = realPriv;		// Not a clustered box
		}
	}
/*
	Make sure this program was setuid(root) otherwise things later on will
	fail in strange and ugly ways.
 */
	if (!progIsSetUID()) {
		fprintf (stderr, "Installation error - shell must be setuid 0\n");
		exit(1);
	}
/*
	The privilege determine which node we'll be using
 */
	switch (myPriv) {
		case viewPriv:
			install_node (&viewPrivNode, NULL);
			vty->node = VIEW_PRIV;
			break;
		case addPriv:
			install_node (&addPrivNode, NULL);
			vty->node = ADD_PRIV;
			break;
		case changePriv:
			install_node (&changePrivNode, NULL);
			vty->node = CHANGE_PRIV;
			break;
		case fullPriv:
			install_node (&fullPrivNode, NULL);
			vty->node = FULL_PRIV;
			break;
		case diagPriv:
			install_node (&diagPrivNode, NULL);
			vty->node = DIAG_PRIV;
			break;
	}
/*
	Install the basic commands
 */
	vtysh_install_default (vty->node);
/*
	Now install all the commands
 */
	install_admin (myPriv, vty->node);
	install_am (myPriv, vty->node);
	install_arp (myPriv, vty->node);
	install_cluster (realPriv, vty->node);
	install_elog (myPriv, vty->node);
	install_autosupport (myPriv, vty->node);
	install_fileset (myPriv, vty->node);
	install_interface (myPriv, vty->node);
	install_vif (myPriv, vty->node);
	install_license (myPriv, vty->node);
	install_nameservice (myPriv, vty->node);
	install_ndmp (myPriv, vty->node);
	install_nfs (myPriv, vty->node);
	install_port (myPriv, vty->node);
	install_quota (myPriv, vty->node);
	install_rcon (myPriv, vty->node);
	install_route (myPriv, vty->node);
	install_snapshot (myPriv, vty->node);
	install_snmp (myPriv, vty->node);
	install_spool (myPriv, vty->node);
	install_system (myPriv, vty->node);
	install_vol (myPriv, vty->node);
	install_av (myPriv, vty->node);
	install_help (myPriv, vty->node);
	install_raid (myPriv, vty->node);
	install_scsi (myPriv, vty->node);
	install_disk (myPriv, vty->node);
	install_tape (myPriv, vty->node);
	install_stats (myPriv, vty->node);
}
