/* vi:ts=8 sw=2
 *-----------------------------------------------------------------
 *
 * Name: vtysh.c
 *
 * RCS:  $Id: vtysh.c,v 1.12 2002/11/21 17:21:30 janv Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: This code was adapted from the GPL Zebra package for use
 *              as the NFX CLI.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 23 Oct 2001
 *
 *-----------------------------------------------------------------
 */
/* Virtual terminal interface shell.
 * Copyright (C) 2000 Kunihiro Ishiguro
 *
 * This file is part of GNU Zebra.
 *
 * GNU Zebra is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * GNU Zebra is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNU Zebra; see the file COPYING.  If not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  
 */

#include "zebra.h"

#include <sys/un.h>
#include <setjmp.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <sys/stat.h>

#include <readline/readline.h>
#include <readline/history.h>

#include "command.h"
#include "memory.h"
#include "vtysh.h"

/* Struct VTY. */
struct vty *vty;

/* VTY shell pager name. */
char *vtysh_pager_name = NULL;

/* VTY shell client structure. */
struct vtysh_client
{
  int fd;
} vtysh_client[VTYSH_INDEX_MAX];

struct node_list {
	int node;
	struct node_list *prev;
};
static struct node_list *nodeStack = NULL;

static int 
pop_node()
{
  int newNode;
  struct node_list *oldNode;

  if (nodeStack) {
    newNode = nodeStack->node;
    oldNode = nodeStack;
    nodeStack = oldNode->prev;
    free (oldNode);
  }
  else {
    newNode = vty->node;
  }
  return newNode;
}

extern void 
push_node (int newNode)
{
  struct node_list *newOne = malloc (sizeof(struct node_list));

  if (!newOne) {
    fprintf (stderr, "PANIC: error getting memory for node stack\n");
    exit(1);
  }
  newOne->prev = nodeStack;
  newOne->node = vty->node;
  nodeStack = newOne;
  vty->node = newNode;
}


/* When '^Z' is received from vty, move down to the enable mode. */
extern int
vtysh_end (struct vty *vty)
{
  vty->node = pop_node();
  return CMD_SUCCESS;
}

extern void
vtysh_pager_init ()
{
  vtysh_pager_name = getenv ("VTYSH_PAGER");
}

/* Command execution over the vty interface. */
static void
vtysh_execute_func (char *line, int pager)
{
  int ret;
  vector vline;
  struct cmd_element *cmd;
  FILE *fp = NULL;

  /* Split readline string up into the vector */
  vline = cmd_make_strvec (line);

  if (vline == NULL)
    return;

  ret = cmd_execute_command (vline, vty, &cmd);

  cmd_free_strvec (vline);

  switch (ret)
    {
    case CMD_WARNING:
      if (vty->type == VTY_FILE)
	printf ("Warning...\n");
      break;
    case CMD_ERR_AMBIGUOUS:
      printf ("%% Ambiguous command/option.\n");
      break;
    case CMD_ERR_NO_MATCH:
      printf ("%% Unknown command/option.\n");
      break;
    case CMD_ERR_INCOMPLETE:
      printf ("%% Command incomplete.\n");
      break;
    case CMD_ERR_FAILED:
      printf ("%% Command failure.\n");
      break;
    case CMD_SUCCESS_DAEMON:
      {
	if (pager && vtysh_pager_name)
	  {
	    fp = popen ("more", "w");
	    if (fp == NULL)
	      {
		perror ("popen");
		exit (1);
	      }
	  }
	else
	  fp = stdout;

	if (! strcmp(cmd->string,"configure terminal"))
	  {
	      if (cmd->func)
		{
		  (*cmd->func) (cmd, vty, 0, NULL);
		  break;
                }
	  }
	if (cmd->func)
	  (*cmd->func) (cmd, vty, 0, NULL);
      }
    }
  if (pager && vtysh_pager_name && fp)
    {
      if (pclose (fp) == -1)
	{
	  perror ("pclose");
	  exit (1);
	}
      fp = NULL;
    }
}

void
vtysh_execute_no_pager (char *line)
{
  vtysh_execute_func (line, 0);
}

void
vtysh_execute (char *line)
{
  vtysh_execute_func (line, 1);
}

/* Configration make from file. */
int
vtysh_config_from_file (struct vty *vty, FILE *fp)
{
  int ret;
  vector vline;
  struct cmd_element *cmd;

  while (fgets (vty->buf, VTY_BUFSIZ, fp))
    {
      if (vty->buf[0] == '!' || vty->buf[0] == '#')
	continue;

      vline = cmd_make_strvec (vty->buf);

      /* In case of comment line */
      if (vline == NULL)
	continue;

      /* Execute configuration command : this is strict match */
      ret = cmd_execute_command_strict (vline, vty, &cmd);

      /* Try again with setting node to CONFIG_NODE */
      if (ret != CMD_SUCCESS 
	  && ret != CMD_SUCCESS_DAEMON
	  && ret != CMD_WARNING)
	{
	      vtysh_execute ("end");
	      vtysh_execute ("configure terminal");
	      push_node(CONFIG_NODE);
	      ret = cmd_execute_command_strict (vline, vty, &cmd);
	}	  

      cmd_free_strvec (vline);

      switch (ret)
	{
	case CMD_WARNING:
	  if (vty->type == VTY_FILE)
	    printf ("Warning...\n");
	  break;
	case CMD_ERR_AMBIGUOUS:
	  printf ("%% Ambiguous command.\n");
	  break;
	case CMD_ERR_NO_MATCH:
	  printf ("%% Unknown command: %s", vty->buf);
	  break;
	case CMD_ERR_INCOMPLETE:
	  printf ("%% Command incomplete.\n");
	  break;
	case CMD_SUCCESS_DAEMON:
	  {
	    if (cmd->func)
	      (*cmd->func) (cmd, vty, 0, NULL);
	  }
	}
    }
  return CMD_SUCCESS;
}

/* We don't care about the point of the cursor when '?' is typed. */
static int
vtysh_rl_describe ()
{
  int ret;
  int i;
  vector vline;
  vector describe;
  int width;
  struct desc *desc;

  vline = cmd_make_strvec (rl_line_buffer);

  /* In case of '> ?'. */
  if (vline == NULL)
    {
      vline = vector_init (1);
      vector_set (vline, '\0');
    }
  else 
    if (rl_end && isspace ((int) rl_line_buffer[rl_end - 1]))
      vector_set (vline, '\0');

  describe = cmd_describe_command (vline, vty, &ret);

  printf ("\n");

  /* Ambiguous and no match error. */
  switch (ret)
    {
    case CMD_ERR_AMBIGUOUS:
      cmd_free_strvec (vline);
      printf ("%% Ambiguous command.\n");
      rl_on_new_line ();
      return 0;
      break;
    case CMD_ERR_NO_MATCH:
      cmd_free_strvec (vline);
      printf ("%% There is no matched command.\n");
      rl_on_new_line ();
      return 0;
      break;
    }  

  /* Get width of command string. */
  width = 0;
  for (i = 0; i < vector_max (describe); i++)
    if ((desc = vector_slot (describe, i)) != NULL)
      {
	int len;

	if (desc->cmd[0] == '\0')
	  continue;

	len = strlen (desc->cmd);
	if (desc->cmd[0] == '.')
	  len--;

	if (width < len)
	  width = len;
      }

  for (i = 0; i < vector_max (describe); i++)
    if ((desc = vector_slot (describe, i)) != NULL)
      {
	if (desc->cmd[0] == '\0')
	  continue;

	if (! desc->str)
	  printf ("  %-s\n",
		  desc->cmd[0] == '.' ? desc->cmd + 1 : desc->cmd);
	else
	  printf ("  %-*s  %s\n",
		  width,
		  desc->cmd[0] == '.' ? desc->cmd + 1 : desc->cmd,
		  desc->str);
      }

  cmd_free_strvec (vline);
  vector_free (describe);

  rl_on_new_line();

  return 0;
}

/* result of cmd_complete_command() call will be stored here
   and used in new_completion() in order to put the space in
   correct places only */
int complete_status;

static char *
command_generator (char *text, int state)
{
  vector vline;
  static char **matched = NULL;
  static int index = 0;

  /* First call. */
  if (! state)
    {
      index = 0;

      if (vty->node == AUTH_NODE || vty->node == AUTH_ENABLE_NODE)
	return NULL;

      vline = cmd_make_strvec (rl_line_buffer);
      if (vline == NULL)
	return NULL;

      if (rl_end && isspace ((int) rl_line_buffer[rl_end - 1]))
	vector_set (vline, '\0');

      matched = cmd_complete_command (vline, vty, &complete_status);
    }

  if (matched && matched[index])
    return matched[index++];

  return NULL;
}

static char **
new_completion (char *text, int start, int end)
{
  char **matches;

  matches = completion_matches (text, command_generator);

  if (matches)
    {
      rl_point = rl_end;
      if (complete_status == CMD_COMPLETE_FULL_MATCH)
	rl_pending_input = ' ';
    }

  return matches;
}

extern int
vtysh_exit (struct vty *vty)
{
  switch (vty->node)
    {
    case CONFIG_NODE:
      vty->node = ENABLE_NODE;
      break;
    default:
      exit (0);
    }
  return CMD_SUCCESS;
}

/*#define MAX_ARGS 10*/
#define MAX_ARGS 20

/* Execute command in child process. */
int
execute_command (char *command, int argc, ...)
{
  va_list ap;
  int ret;
  pid_t pid;
  int status;
  char *a[MAX_ARGS];
  int i;
  int return_code;
  struct sigaction act, oact;

  /* Block tty stop in child processes so they can not be suspended by ^Z */
  act.sa_handler = SIG_IGN;
  sigfillset(&act.sa_mask);
  act.sa_flags = 0;
  sigaction(SIGTSTP, &act, &oact);

  va_start(ap, argc);

  /* Call fork(). */
  pid = fork ();

  if (pid < 0)
    {
      /* Failure of fork(). */
      fprintf (stderr, "Can't fork: %s\n", strerror (errno));
      exit (1);
    }
  else if (pid == 0)
    {
      /* This is child process. */
/*
  This would be much easier if exec*() accepted an va_list.
 */
      if (argc > MAX_ARGS) {
	fprintf (stderr, "execute_command: got too many arguments\n");
	argc = MAX_ARGS;
      }
      memset (a, 0, sizeof (a));
      for (i = 0; i < argc; i++) {
	a[i] = va_arg (ap, char *);
	if (!a[i]) {
	  break;
	}
      }
      va_end(ap);

      ret = execlp (command, command, a[0], a[1], a[2], a[3], 
                    a[4], a[5], a[6], a[7], a[8], a[9], a[10], a[11],
                    a[12], a[13], a[14], a[15], a[16], a[17], a[18], a[19],
                    NULL);

      /* When execlp suceed, this part is not executed. */
      fprintf (stderr, "Can't execute %s: %s\n", command, strerror (errno));
      exit (1);
    }
  else
    {
      /* This is parent. */
      execute_flag = 1;
      ret = wait4 (pid, &status, 0, NULL);

	  sigaction(SIGTSTP, &oact, NULL);
	  
      execute_flag = 0;
      if (WIFEXITED(status)) {
	return_code = WEXITSTATUS(status);
      } else if (WIFSIGNALED(status)) {
	return_code = WTERMSIG(status);
      } else {
	return_code = -1;
      }
    }
  return return_code;
}

#define MAX_VARARGS	256
int
execute_vararg_command(char *command, int argc, char *sub_cmd, char *argv[])
{
	int ret;
	pid_t pid;
	int status;
  	char *a[MAX_VARARGS+2];
	int i, pos=0;
	int return_code;

	/* Call fork(). */
  	pid = fork();

	if(pid < 0)
	  {
	    /* Failure of fork().*/
	    fprintf (stderr, "execute_vararg_command: Can't fork: %s\n", strerror (errno));
	    exit (1);
	  }
	else if(pid == 0)
	  {
	    /* This is child process. */

	    /* create the array to be passed to execv */

	    a[pos] = (char *)malloc(strlen(command)+1);
	    if(!a[pos]) {
	      fprintf (stderr, "execute_vararg_command: malloc failed: %s\n", strerror (errno));
	      exit (1);
	    }
	    strcpy(a[pos], command);
	    pos++;

	    if(sub_cmd) {
	      a[pos] = (char *)malloc(strlen(sub_cmd)+1);
	      if(!a[pos]) {
	   		  fprintf (stderr, "execute_vararg_command: malloc failed: %s\n", strerror (errno));
			  exit (1);
	      }
	      strcpy(a[pos], sub_cmd);
	      pos++;
	    }

	    for(i=0; i < argc; i++)
	      a[pos++] = argv[i];

	    a[pos] = NULL;

	    /* execute the command now */
	    ret = execvp (command, a);

	    /* When execvp succeeds, this part is not executed. */
		fprintf (stderr, "execute_vararg_command: Can't execute %s: %s\n", command, strerror (errno));
		exit (1);
	  }
	else
	  {
	    /* This is parent. */
	    execute_flag = 1;
	    ret = wait4 (pid, &status, 0, NULL);
	    execute_flag = 0;
	    if (WIFEXITED(status)) {
	      return_code = WEXITSTATUS(status);
	    } else if (WIFSIGNALED(status)) {
	      return_code = WTERMSIG(status);
	    } else {
	      return_code = -1;
	    }
	  }
	return return_code;
}

/* To disable readline's filename completion */
static int
vtysh_completion_entry_fucntion (int ignore, int invoking_key)
{
  return 0;
}

extern void
vtysh_readline_init ()
{
  /* readline related settings. */
  rl_bind_key ('?', vtysh_rl_describe);
  rl_completion_entry_function = vtysh_completion_entry_fucntion;
  rl_attempted_completion_function = (CPPFunction *)new_completion;
  /* do not append space after completion. It will be appended
     in new_completion() function explicitly */
  rl_completion_append_character = '\0';
}

extern char *
vtysh_prompt ()
{
  struct utsname names;
  static char buf[100];
  const char*hostname;
  extern struct host host;

  hostname = host.name;

  if (!hostname)
    {
      uname (&names);
      hostname = names.nodename;
    }

  snprintf (buf, sizeof buf, cmd_prompt (vty->node), hostname);

  return buf;
}

extern void
vtysh_init_vty ()
{
  /* Make vty structure. */
  vty = vty_new ();
  vty->type = VTY_SHELL;
  vty->node = VIEW_NODE;

  /* Initialize commands. */

  cmd_init (0);

}
