/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: ssc-rcon
 *
 * RCS:  $Id: ssc-rcon.c,v 1.7 2002/09/23 17:28:49 gregs Exp $
 *
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Description: This is the "remote console" application that runs on the SSC.
 *              It talks to its peer on the embedded CPU to provide console
 *              I/O for the embedded UI.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 19 Dec 2001
 *
 *-----------------------------------------------------------------
 */
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <neteee/eee.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <unistd.h>
#include <signal.h>

#include "nfx-incl.h"
#include "exports/rcon-api.h"
#include "exports/anpssc-api.h"		// For SSC_SLOT define
#include "exports/eee-error.h"		// Error code definitions

/*
	Now for the Zen of what follows. This application provides the virtual
	console for the NFX embedded UI(s). It will block in select() waiting
	for data to arrive on either stdin or the socket doing the right thing.

	To answer the question, "But how do I tell it which CPU to talk to?"
	Easy, it's on the command line:

		rcon Slot CPU 
 */

static struct sockaddr_eee ourPeer;		// To whom we are speaking
static int Sock;						// Use this to send.
static int myPid;						// Who we are

/*-----------------------------------------------------------------
 * Name: init_Rconn
 *
 * Description: Start the conversation with the embedded bits
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 19 Dec 2001
 *
 *-----------------------------------------------------------------
 */
static int
init_Rconn (int sock, struct sockaddr_eee *peer)
{
	char msgBuff[8192];									// Max possible MTU
	rconStatMsg_t *statMsg = (rconStatMsg_t *)msgBuff;
	rconCntlMsg_t control;
	fd_set selSet;
	struct timeval selTO = {5, 0};	// 5 seconds should be enough
	int rc;
/*
	Send the INIT request to our peer and wait for it to respond.
 */
	memset (&control, 0, sizeof(control));
	control.hdr.msgType = RCON_CNTL;
	control.hdr.clientPrivate = myPid;
	control.action = INIT_CONNECTION;

	if (sendto (sock, &control, sizeof(control), 0, 
	            (struct sockaddr *)peer, sizeof(struct sockaddr_eee)) < 0) {
		perror ("sendto() init connection failed");
		return 1;
	}
	FD_ZERO(&selSet);
	FD_SET (sock, &selSet);

	rc = select (sock + 1, &selSet, 0, 0, &selTO);

	if (rc < 0) {
		perror ("Select error");
		return 1;
	}
	else if (!rc) {
		fprintf (stderr, "Embedded rcon not answering our init request\n");
		return 1;
	}

	rc = recv (sock, msgBuff, sizeof (msgBuff), 0);

	if (rc < sizeof(rconStatMsg_t)) {
		fprintf (stderr, "Init stat message truncated at %d\n", rc);
		if (rc < 0) {
			perror ("recv() error");
		}
		return 1;
	}

	if (statMsg->stat != NFX_OK) {
		fprintf (stderr, "Init request status indicates failure");
		if (rc > sizeof(rconStatMsg_t)) {
			fprintf (stderr, ": %s", statMsg->errMsg);
		}
		fprintf (stderr, "\n");
		return 1;
	}
	return 0;
}

/*-----------------------------------------------------------------
 * Name:termHand
 *
 * Description: Handle termination signals
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 19 Dec 2001
 *
 *-----------------------------------------------------------------
 */
static void
termHand (int signo)
{
	rconCntlMsg_t control;
/*
	Tell our peer to shutdown
 */
	memset (&control, 0, sizeof(control));
	control.hdr.msgType = RCON_CNTL;
	control.hdr.clientPrivate = myPid;
	control.action = DESTROY_CONNECTION;

	if (sendto (Sock, &control, sizeof(control), 0, 
	            (struct sockaddr *)&ourPeer, sizeof(ourPeer)) < 0) {
		perror ("sendto() termination message send failed");
	}
/*
	SIGINT and SIGHUP are normal termination signals
 */
	if ((signo == SIGHUP) || (signo == SIGINT)) {
		exit (0);
	}

	fprintf (stderr, "Abnormal Termination\n");
	exit (1);
}

/*-----------------------------------------------------------------
 * Name: main
 *
 * Description: It all starts here
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 19 Dec 2001
 *
 *-----------------------------------------------------------------
 */
extern int
main (int argc, char *argv[])
{
	struct sockaddr_eee ourAddr;
	int slot;
	int CPU;
	char msgBuff[8192];									// Max possible MTU
	rconDataMsg_t *dataMsg = (rconDataMsg_t *)msgBuff;
	rconCntlMsg_t *cntlMsg = (rconCntlMsg_t *)msgBuff;
	sigset_t sigs;
	int rc;
	fd_set selset;
#define BUFSIZE (sizeof (msgBuff) - sizeof (rconDataMsg_t))

/*
	First things first.
	Let's make sure we have rightous arguments.
 */
	myPid = getpid();
	if (argc != 3) {
		fprintf (stderr, "Usage %s Slot# CPU#\n", argv[0]);
		return 1;
	}
	slot = strtol (argv[1], NULL, 10);
	CPU = strtol (argv[2], NULL, 10);

	if ((slot < NFX_SLOT_MIN) || (slot > NFX_SLOT_MAX)) {
		fprintf (stderr, "Invalid slot number\n");
		return 1;
	}
	if ((CPU < 0) || (CPU >= NFX_MAX_CPUS_PER_SLOT)) {
		fprintf (stderr, "Invalid CPU number\n");
		return 1;
	}

/*
	Set up our peer's address
 */
	memset (&ourPeer, 0, sizeof(ourPeer));
	ourPeer.seee_family = AF_EEE;
	if ((slot == 0) && (CPU == 1))
		ourPeer.seee_bplane = 1;		/* for messaging to toshiba */
	else
		ourPeer.seee_bplane = EEE_CONTROLBPLANE;
	ourPeer.seee_app = eee_getApplicationId("rcon");
	if (ourPeer.seee_app == EEE_ERR_INVALID_APP_NAME) {
		fprintf (stderr, "Unknown application id rcon\n");
		exit (1);
	}
	ourPeer.seee_slot = slot;
	ourPeer.seee_cpu = CPU;
/*
	Let's see if we can establish communications with our peer.
	To do that, we'll create two persistent sockets of the EEE kind.
	"Why not use receiveAgileMsg()?", you ask. Well, to do that we would
	need a unique app id for each instance of rcon on the SSC. Could also
	hack "special behavior" into regAgileApp too, but it's really not worth
	the overhead. This is pretty much a special case.
 */
	memset (&ourAddr, 0, sizeof (ourAddr));
	ourAddr.seee_family = AF_EEE;
	ourAddr.seee_slot = SSC_SLOT;
	ourAddr.seee_cpu = NFX_SSC_UNIX;
	if ((slot == 0) && (CPU == 1))
		ourAddr.seee_bplane = 1;		/* for messaging to toshiba */
	else
		ourAddr.seee_bplane = EEE_CONTROLBPLANE;
	
	if ((Sock = socket(PF_EEE, SOCK_DGRAM, 0)) >= 0) {
		if (bind(Sock, (struct sockaddr *)&ourAddr, sizeof(ourAddr)) < 0) {
			perror ("Error binding Rx socket");
			return 1;
		}
	}
	else {
		perror ("Error creating Rx socket");
		return 1;
	}
/*
	We have sockets! Let's use the Rx one so our peer can record its address
	for later consumption (that's the app id it'll send its stdout data to).
 */
	if (init_Rconn (Sock, &ourPeer)) {
		fprintf (stderr, "Couldn't initialize remote console session\n");
		return 1;
	}
/*
	Set up signal handler(s) to terminate the connection upon exit.
	We'll use SIGHUP and/or SIGINT as "normal" termination. We'll catch
	SIGBUS and SIGSEGV and note it as abnormal.
 */
	sigfillset (&sigs);
	sigdelset (&sigs, SIGHUP);
	sigdelset (&sigs, SIGINT);
	sigdelset (&sigs, SIGBUS);
	sigdelset (&sigs, SIGSEGV);
	sigprocmask (SIG_SETMASK, &sigs, NULL);

	signal (SIGHUP, termHand);
	signal (SIGINT, termHand);
	signal (SIGBUS, termHand);
	signal (SIGSEGV, termHand);
/*
	And cried the Raven, evermore. (Sorry Edgar)
	Display the strings we get from our peer and send the stuff we get from
	stdin to them.
 */
	do {
		FD_ZERO (&selset);
		FD_SET (0, &selset);
		FD_SET (Sock, &selset);

		rc = select (Sock + 1, &selset, NULL, NULL, NULL);

		if (FD_ISSET (0, &selset)) {
			if (fgets (dataMsg->buf, BUFSIZE, stdin)) {

				dataMsg->hdr.msgType = RCON_DATA;
				dataMsg->len = strlen (dataMsg->buf) + 1;

				sendto (Sock, dataMsg, 
				    sizeof (rconDataMsg_t) + strlen(dataMsg->buf), 0,
			        (struct sockaddr *)&ourPeer, sizeof(struct sockaddr_eee));
			}
			else if (feof (stdin)) {    // Ignore EOF char. Exit with ^C
				clearerr (stdin);
			}
			else if (ferror (stdin)) {
				fprintf (stderr, "Unexpected error from stdin\n");
				termHand(-1);   // Let's bail
			}
		}

		if (FD_ISSET (Sock, &selset)) {
			rc = recv (Sock, msgBuff, sizeof (msgBuff), 0);
			if (cntlMsg->hdr.msgType == RCON_CNTL) {
				fprintf (stderr, "Exiting due to request from embedded\n");
				termHand (SIGHUP);
			}
			write (1, dataMsg->buf, strlen (dataMsg->buf));
		}

	} while ( (rc > 0) && (dataMsg->hdr.msgType == RCON_DATA));

	if (rc < 0) {
		perror ("RX recv error");
		return 1;
	}
	else {
		return 0;
	}
}
