/* vi:ts=4
 *-----------------------------------------------------------------
 *
 * Name: RCON access library
 *
 * RCS:  $Id: rconlib.c,v 1.2 2002/09/27 00:22:33 gregs Exp $
 *
 * Copyright (C) 2002, Agile Storage, Inc.
 *
 * Description: Provides the functions to perform rcon sessions in a library.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 20 Sep 2002
 *
 *-----------------------------------------------------------------
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <neteee/eee.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>

#include "nfx-incl.h"
#include "exports/anpssc-api.h"     // For SSC_SLOT define
#include "exports/eee-error.h"      // Error code definitions
#include "exports/rcon-api.h"

static int rcon_sock = -1;	// This is where we keep the session's open fd
static struct sockaddr_eee them;
static int myPid;

static int
init_Rconn (int sock, struct sockaddr_eee *peer)
{
	long long msgBuff[1024];					// Max possible MTU
	rconStatMsg_t *statMsg = (rconStatMsg_t *)msgBuff;
	rconCntlMsg_t control;
	fd_set selSet;
	struct timeval selTO = {5, 0};	// 5 seconds should be enough
	int rc;
/*
	Send the INIT request to our peer and wait for it to respond.
 */
	memset (&control, 0, sizeof(control));
	control.hdr.msgType = RCON_CNTL;
	control.hdr.clientPrivate = myPid;
	control.action = INIT_CONNECTION;

	if (sendto (sock, &control, sizeof(control), 0, 
	            (struct sockaddr *)peer, sizeof(struct sockaddr_eee)) < 0) {
		perror ("sendto() init connection failed");
		return 1;
	}
	FD_ZERO(&selSet);
	FD_SET (sock, &selSet);

	rc = select (sock + 1, &selSet, 0, 0, &selTO);

	if (rc < 0) {
		perror ("Select error");
		return 1;
	}
	else if (!rc) {
		fprintf (stderr, "Embedded rcon not answering our init request\n");
		return 1;
	}

	rc = recv (sock, msgBuff, sizeof (msgBuff), 0);

	if (rc < sizeof(rconStatMsg_t)) {
		fprintf (stderr, "Init stat message truncated at %d\n", rc);
		if (rc < 0) {
			perror ("recv() error");
		}
		return 1;
	}

	if (statMsg->stat != NFX_OK) {
		fprintf (stderr, "Init request status indicates failure");
		if (rc > sizeof(rconStatMsg_t)) {
			fprintf (stderr, ": %s", statMsg->errMsg);
		}
		fprintf (stderr, "\n");
		return 1;
	}
	return 0;
}
/*-----------------------------------------------------------------
 * Name: establish_rcon_session
 *
 * Description: Open a session to the given slot and cpu.
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 20 Sep 2002
 *
 *-----------------------------------------------------------------
 */
static void 
bail_out (int sig)
{
	shutdown_rcon_session ();
	fprintf (stderr, "Exiting on SIGINT\n");
	exit (1);
}

extern int 
establish_rcon_session (int slot, int cpu)
{
	int rc = 0;
	struct sockaddr_eee us;
	struct sigaction nint_action;

	if (rcon_sock == -1) {
		myPid = getpid();
/*
	Set up our peer's address
 */
		memset (&them, 0, sizeof(them));
		them.seee_family = AF_EEE;
		if ((slot == 0) && (cpu == 1)) {
			them.seee_bplane = 1;		/* for messaging to toshiba */
		}
		else {
			them.seee_bplane = EEE_CONTROLBPLANE;
		}
		them.seee_app = eee_getApplicationId("rcon");
		if (them.seee_app == EEE_ERR_INVALID_APP_NAME) {
			fprintf (stderr, "Unknown application id rcon\n");
			return -1;
		}
		them.seee_slot = slot;
		them.seee_cpu = cpu;
/*
	Let's see if we can establish communications with our peer.
	To do that, we'll create two persistent sockets of the EEE kind.
	"Why not use receiveAgileMsg()?", you ask. Well, to do that we would
	need a unique app id for each instance of rcon on the SSC. Could also
	hack "special behavior" into regAgileApp too, but it's really not worth
	the overhead. This is pretty much a special case (yea right!).
 */
		memset (&us, 0, sizeof (us));
		us.seee_family = AF_EEE;
		us.seee_slot = SSC_SLOT;
		us.seee_cpu = NFX_SSC_UNIX;
		if ((slot == 0) && (cpu == 1))
			us.seee_bplane = 1;		/* for messaging to toshiba */
		else
			us.seee_bplane = EEE_CONTROLBPLANE;
	
		if ((rcon_sock = socket(PF_EEE, SOCK_DGRAM, 0)) >= 0) {
			if (bind(rcon_sock, (struct sockaddr *)&us, sizeof(us)) < 0) {
				perror ("Error binding Rx socket");
				rc = -1;
			}
		}
		else {
			perror ("Error creating rcon socket");
			rc = -1;
		}
/*
	We have sockets! Let's use the Rx one so our peer can record its address
	for later consumption (that's the app id it'll send its stdout data to).
 */
		if (!rc && init_Rconn (rcon_sock, &them)) {
			fprintf (stderr, "Couldn't initialize remote console session\n");
			rc = -1;
		}
	}
	else {
		fprintf (stderr, "rcon session currently open\n");
		rc = -1;
	}
/*
 * In case they hit cntrl-C and abort, we need to clean up after them.
 * I'm not going to expect the users of this service to do it.
 */
	nint_action.sa_handler = bail_out;
	nint_action.sa_flags = 0;
	sigemptyset (&nint_action.sa_mask);
	sigaction (SIGINT, &nint_action, NULL);

	return rc;
}

/*-----------------------------------------------------------------
 * Name: send_to_rcon
 *
 * Description: Send this string to the open session. 
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 20 Sep 2002
 *
 *-----------------------------------------------------------------
 */
extern int 
send_to_rcon (char *a_string_to_send)
{
#define MAX_STRING (sizeof(buf) - sizeof (rconDataMsg_t))
	int rc = 0;
	long long buf[1024];	// Basically max MTU
	rconDataMsg_t *dataMsg = (rconDataMsg_t *)buf;
	int len;

	if (rcon_sock != -1) {
		len = strlen (a_string_to_send);
		if (len > MAX_STRING) {
			fprintf (stderr, "send_to_rcon: String too long. Max %d\n", 
		             MAX_STRING);
			rc = -1;
		}
		else {
			strcpy (dataMsg->buf, a_string_to_send);
			dataMsg->hdr.msgType = RCON_DATA;
			dataMsg->hdr.clientPrivate = myPid;
			dataMsg->len = len + 1;	// Don't forget to count the \0

			len += sizeof (rconDataMsg_t);	// \0 counted in sizeof()
			
			rc = sendto (rcon_sock, buf, len, 0, 
		            (struct sockaddr *)&them, sizeof(struct sockaddr_eee));

			if (rc == len) {
				rc = 0;
			}
			else {
				fprintf (stderr, "send_to_rcon: Got %d error sending\n", rc);
				rc = -1;
			}
		}
	}
	else {
		fprintf (stderr, "Unable to send. Session not open\n");
		rc = -1;
	}
	return rc;
}

/*-----------------------------------------------------------------
 * Name: get_from_rcon
 *
 * Description: Get me data from the open session
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 20 Sep 2002
 *
 *-----------------------------------------------------------------
 */
extern int 
get_from_rcon (char *a_place_to_put_the_data, int place_size, int secs_to_wait)
{
	int rc = 0;
	long long buf[1024];	// Basically max MTU
	rconDataMsg_t *dataMsg = (rconDataMsg_t *)buf;
	struct timeval waitThis;
	fd_set selset;

	if (rcon_sock != -1) {
		FD_ZERO (&selset);
		FD_SET (rcon_sock, &selset);
		waitThis.tv_sec = secs_to_wait;
		waitThis.tv_usec = 0;

		rc = select (rcon_sock + 1, &selset, NULL, NULL, &waitThis);
		if (rc > 0) {
			rc = recv (rcon_sock, buf, sizeof(buf), 0);
			if (dataMsg->hdr.msgType == RCON_DATA) {
				strncpy (a_place_to_put_the_data, dataMsg->buf, place_size);
				if (dataMsg->len > place_size) {
					rc = place_size;
				}
				else {
					rc = dataMsg->len + 1;	// Remember the \0
				}
			}
			else {
				fprintf (stderr, "Unexpected message from embedded\n");
				rc = -1;
			}
		}
	}
	else {
		fprintf (stderr, "Unable to get. Session not open\n");
		rc = -1;
	}
	return rc;
}

/*-----------------------------------------------------------------
 * Name: shutdown_rcon_session
 *
 * Description: Opposite of establish
 *
 * Created by:	Greg Seibert
 *
 * Date Created: 20 Sep 2002
 *
 *-----------------------------------------------------------------
 */
extern int 
shutdown_rcon_session (void)
{
	int rc = 0;
	rconCntlMsg_t control;

	if (rcon_sock != -1) {
/*
	Tell our peer to shutdown
 */
		memset (&control, 0, sizeof(control));
		control.hdr.msgType = RCON_CNTL;
		control.hdr.clientPrivate = myPid;
		control.action = DESTROY_CONNECTION;

		if (sendto (rcon_sock, &control, sizeof(control), 0, 
		            (struct sockaddr *)&them, sizeof(them)) < 0) {
			perror ("sendto() termination message send failed");
			rc = -1;
		}
		close (rcon_sock);
		rcon_sock = -1;
	}
	else {
		fprintf (stderr, "rcon session not open\n");
		rc = -1;
	}
	return rc;
}
