/*------------------------------------------------------------------------------
 * Copyright (C) 2001, Agile Storage, Inc.
 *
 * Name         : ea-processEvMsg.c
 *
 * Description  : This module deals with the file sytems and file sets.
 * It processes messages received from Eagle Vision (EV) and CLI.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

#include <unistd.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "ea.h"
#include "exports/rpc.h"
#include "exports/mount.h"
#include "exports/raid-api.h"
#include "exports/spm-api.h"
#include "exports/ncm-api.h"
#include "exports/fs-opt-api.h"
#include <netinet/in.h>
#undef MIN
#undef MAX
#include <arpa/inet.h>

#define MNT_MSG_BUF_LEN 2048
/*------------------------------------------------------------------------------
 * Prototype of local functions.
 *------------------------------------------------------------------------------
 */

static void
ea_processAddFsysReq(ea_addFsysReq_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processModFsysReq(ea_modFsysReq_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processMntFsysReq(ea_mntFsysReq_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processUmntFsysReq(ea_umntFsysReq_t * evReq,
                      ea_senderInfo_t *  sender);

static void
ea_processDelFsysReq(ea_delFsysReq_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processGetFsysReq(ea_getFsysReq_t * evReq,
                      ea_senderInfo_t *  sender);

static void
ea_processGetDetailFsysReq(ea_getDetailFsysReq_t * evReq,
                      ea_senderInfo_t *  sender);

static void
ea_processUsgFsysReq(ea_usgFsysReq_t * evReq,
                      ea_senderInfo_t *  sender);

static void
ea_processMovOutFsysReq(ea_movOutFsysReq_t * evReq,
                        ea_senderInfo_t *    sender);

static void
ea_processMovInFsysReq(ea_movInFsysReq_t * evReq,
                       ea_senderInfo_t *   sender);

static void
ea_processNumFsysReq(ea_numFsysReq_t * evReq,
						ea_senderInfo_t *   sender);

static void
ea_processGrowFsysReq(ea_growFsysReq_t *evReq, 
						ea_senderInfo_t *	sender);

static void
ea_processGrowFsysRsp(ea_growFsysRsp_t *evReq, 
						ea_senderInfo_t *	sender);

static void
ea_processGrowFsysInd(ea_growFsysInd_t *evReq, 
						ea_senderInfo_t *	sender,
                        int32 len);

/* 
 * File set messages
 */
static void
ea_processAddFsetReq(ea_addFsetReq_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processModFsetReq(ea_modFsetReq_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processDelFsetReq(ea_delFsetReq_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processNumFsetReq(ea_numFsetReq_t * evReq, 
                     ea_senderInfo_t * sender);

static void
ea_processGetFsetReq(ea_getFsetReq_t * evReq, 
                     ea_senderInfo_t * sender);

static void
ea_processUsgFsetReq(ea_usgFsetReq_t * evReq, 
                     ea_senderInfo_t * sender);

/*
 * Export related messages
 */

static void
ea_processAddExpReq(ea_addExpReq_t * evReq,
					ea_senderInfo_t * sender);

static void
ea_processModExpReq(ea_modExpReq_t * evReq,
					ea_senderInfo_t * sender);

static void
ea_processDelExpReq(ea_delExpReq_t * evReq,
					ea_senderInfo_t * sender);

static void
ea_processDelPathReq(ea_delPathReq_t * evReq,
					ea_senderInfo_t * sender);

static void
ea_processGetExpReq(ea_getExpReq_t * evReq,
					ea_senderInfo_t * sender);

static void
ea_processGetPathExpReq(ea_getPathExpReq_t * evReq,
					ea_senderInfo_t * sender);

/* 
 * Snap shot messages
 */
static void
ea_processCrtSnapReq(ea_crtSnapReq_t * evReq, 
                     ea_senderInfo_t * sender);

static void
ea_processDelSnapReq(ea_delSnapReq_t * evReq, 
                     ea_senderInfo_t * sender);

static void
ea_processSchSnapReq(ea_schSnapReq_t * evReq, 
                     ea_senderInfo_t * sender);

static void
ea_processGetSchSnapReq(ea_getSchSnapReq_t * evReq, 
                     ea_senderInfo_t * sender);

static void
ea_processGetUsgSnapReq(ea_getUsgSnapReq_t * evReq, 
                     ea_senderInfo_t * sender);

/* 
 * Raid Control messages
 */
static void
ea_processRaidGetListReq(ea_raid_getListReqMsg_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processRaidOpenReq(ea_raid_openReqMsg_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processRaidCloseReq(ea_raid_closeReqMsg_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processRaidCtrlInfoReq(ea_raid_ctrlInfoReqMsg_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processRaidCtrlParamReq(ea_raid_getCtrlParamReqMsg_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processRaidPhyDevListReq(ea_raid_phyDevListReqMsg_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processRaidPhyDevInfoReq(ea_raid_phyDevInfoReqMsg_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processRaidLogDevListReq(ea_raid_logDevListReqMsg_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processRaidLogDevInfoReq(ea_raid_logDevInfoReqMsg_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processRaidCreateLunReq(ea_raid_createLunReqMsg_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processRaidCreateSimpleLunReq(ea_raid_createSimpleLunReqMsg_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processRaidDeleteLunReq(ea_raid_deleteLunReqMsg_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processRaidInitLunReq(ea_raid_initLunReqMsg_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processRaidGetLunMaskReq(ea_raid_getLunMaskReqMsg_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processRaidAddLunMaskReq(ea_raid_addLunMaskReqMsg_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processRaidDelLunMaskReq(ea_raid_delLunMaskReqMsg_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processRaidBpListReq(ea_raid_bpListReqMsg_t * evReq,
                     ea_senderInfo_t * sender);

static void
ea_processRaidCancelBpReq(ea_raid_cancelBpReqMsg_t * evReq,
                     ea_senderInfo_t * sender);

/* 
 * Chassis messages
 */
static void
ea_processGetChassisInfoReq(ea_getChassisInfoReq_t * evReq,
                      ea_senderInfo_t *  sender);

static void
ea_processGetFcPortWwnReq(ea_listFcPortWwnReq_t * evReq,
                      ea_senderInfo_t *  sender);

static void
ea_processLunRediscoverReq(ea_lunRediscoverReq_t * evReq,
                      ea_senderInfo_t *  sender);

static void
ea_processGetPortTopologyReq(ea_getPortTopologyReq_t * evReq,
                      ea_senderInfo_t *  sender);

#define htonll(a) \
	((((a) >> 56) & 0x00000000000000FFLL) | \
 		(((a) >> 40) & 0x000000000000FF00LL) | \
 		(((a) >> 24) & 0x0000000000FF0000LL) | \
 		(((a) >>  8) & 0x00000000FF000000LL) | \
 		(((a) <<  8) & 0x000000FF00000000LL) | \
 		(((a) << 24) & 0x0000FF0000000000LL) | \
 		(((a) << 40) & 0x00FF000000000000LL) | \
 		(((a) << 56) & 0xFF00000000000000LL))
	

/*------------------------------------------------------------------------------
 * Prototype of external functions.
 *------------------------------------------------------------------------------
 */
uint32 ea_childGetLock(char *resName);
void ea_childUnLock(void);


/*------------------------------------------------------------------------------
 * Global variables
 *------------------------------------------------------------------------------
 */
struct timeval waitRcmTo = {60, 0};
extern	struct timeval waitEfsTo;
extern	struct timeval waitCmTo;

int myAsyncSock = 0;

char *fsStateStr[] =
{
    "Unknown",
    "Mounted",
    "Mounted-RO",
    "NotMounted",
    "Paused",
    "Mounting",
    "Unmounting",
    "Failing",
    "Mounted"
};


/*------------------------------------------------------------------------------
 * Action on SPM messages
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : ea_processSpmReq()
 *
 * Description  : Process a request from EV for the SPM
 *                EV --> SPM
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created :
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processSpmReq(ea_msgHdr_t * evReq,
                     ea_senderInfo_t * sender)
{
	spm_msgHdr_t *spmMsg;
	uint32 rspLen;

	/*
	 * Strip off the EA msg header and forward to SPM
 	 * EV --> SPM
	 */
	spmMsg = (spm_msgHdr_t *)(((char *) evReq) + sizeof(ea_msgHdr_t));
	rspLen = evReq->msgLen - sizeof(ea_msgHdr_t);

	/*
	 * EA msg type has to be one of the following
	 */
	switch(evReq->msgType)
	{
		case EA_SPM_GET_TOTAL_POOLS_REQ:
		spmMsg->msgType = SPM_GET_TOTAL_POOLS_REQ;
		break;

		case EA_SPM_GET_POOL_LIST_REQ:
		spmMsg->msgType = SPM_GET_POOL_LIST_REQ;
		break;

		case EA_SPM_GET_TOTAL_LUNS_REQ:
		spmMsg->msgType = SPM_GET_TOTAL_LUNS_REQ;
		break;

		case EA_SPM_GET_LUN_LIST_REQ:
		spmMsg->msgType = SPM_GET_LUN_LIST_REQ;
		break;

		case EA_SPM_MARK_LUN_AS_AGILE_REQ:
		spmMsg->msgType = SPM_MARK_LUN_AS_AGILE_REQ;
		break;

		case EA_SPM_MARK_LUN_AS_OC_REQ:
		spmMsg->msgType = SPM_MARK_LUN_AS_OC_REQ;
		break;

		case EA_SPM_UNMARK_LUN_AS_AGILE_REQ:
		spmMsg->msgType = SPM_UNMARK_LUN_AS_AGILE_REQ;
		break;

		case EA_SPM_RESCAN_LUN_REQ:
		spmMsg->msgType = SPM_RESCAN_LUN_REQ;
		break;

		case EA_SPM_GET_STORAGE_POOL_LIST_REQ:
		spmMsg->msgType = SPM_GET_STORAGE_POOL_LIST_REQ;
		break;

		case EA_SPM_GET_STORAGE_LUN_LIST_REQ:
		spmMsg->msgType = SPM_GET_STORAGE_LUN_LIST_REQ;
		break;

		default:
		EA_ERR_PRINT(("Unexpected EA-SPM message type %d\n", evReq->msgType));
		return;
	}

	EA_DBG_PRINT(("ea_processSpmReq: Sending msg %d len %d to SPM\n",
				spmMsg->msgType, rspLen));

    ea_sendMsg(spmMsg, rspLen, "spm");
}

/*------------------------------------------------------------------------------
 * Name         : ea_processSpmRsp()
 *
 * Description  : Process a response or indication coming from SPM for EV/SRM
 *                SPM --> EV/SRM
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created :
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processSpmRsp(spm_msgHdr_t * spmRsp, uint32 len,
                     ea_senderInfo_t * sender)
{
	ea_msgHdr_t *eaMsg;
	uint32 rspLen;

	/* 
	 * Append EA msg header and forward to AV
     * SPM --> EV/SRM
	 */
	rspLen = len + sizeof(ea_msgHdr_t);

	eaMsg = (ea_msgHdr_t *) malloc(rspLen);
	memset(eaMsg, 0, rspLen);

	strcpy(eaMsg->node, spmRsp->node);

	memcpy((void *)((char *)eaMsg + sizeof(ea_msgHdr_t)),
							spmRsp, len);

    switch (spmRsp->msgType)
    {
		case SPM_GET_TOTAL_POOLS_RSP:
		eaMsg->msgType = EA_SPM_GET_TOTAL_POOLS_RSP;
		break;

        case SPM_GET_POOL_LIST_RSP:
		eaMsg->msgType = EA_SPM_GET_POOL_LIST_RSP;
		break;

        case SPM_GET_TOTAL_LUNS_RSP:
		eaMsg->msgType = EA_SPM_GET_TOTAL_LUNS_RSP;
		break;

        case SPM_GET_LUN_LIST_RSP:
		eaMsg->msgType = EA_SPM_GET_LUN_LIST_RSP;
		break;

        case SPM_MARK_LUN_AS_AGILE_RSP:
		eaMsg->msgType = EA_SPM_MARK_LUN_AS_AGILE_RSP;
		break;

        case SPM_MARK_LUN_AS_OC_RSP:
		eaMsg->msgType = EA_SPM_MARK_LUN_AS_OC_RSP;
		break;

        case SPM_UNMARK_LUN_AS_AGILE_RSP:
		eaMsg->msgType = EA_SPM_UNMARK_LUN_AS_AGILE_RSP;
		break;

        case SPM_RESCAN_LUN_RSP:
		eaMsg->msgType = EA_SPM_RESCAN_LUN_RSP;
		break;

        case SPM_SRM_INIT_IND:
		eaMsg->msgType = EA_SPM_SRM_INIT_IND;
		break;

        case SPM_SRM_NEW_LUN_IND:
		eaMsg->msgType = EA_SPM_SRM_NEW_LUN_IND;
		break;

        case SPM_SRM_DEL_LUN_IND:
		eaMsg->msgType = EA_SPM_SRM_DEL_LUN_IND;
		break;

        case SPM_SRM_LUN_STATE_CHANGE_IND:
		eaMsg->msgType = EA_SPM_SRM_LUN_STATE_CHANGE_IND;
		break;

        case SPM_GET_STORAGE_POOL_LIST_RSP:
		eaMsg->msgType = EA_SPM_GET_STORAGE_POOL_LIST_RSP;
		break;

        case SPM_GET_STORAGE_LUN_LIST_RSP:
		eaMsg->msgType = EA_SPM_GET_STORAGE_LUN_LIST_RSP;
		break;

        default:
        EA_ERR_PRINT(("EA has Unknown message (%d)!\n", spmRsp->msgType));
        break;
	}

	EA_DBG_PRINT(("ea_processSpmRsp: Sending SPM msg %d len %d to AV\n",
				eaMsg->msgType, rspLen));

    ea_sendMsg(eaMsg, rspLen, "ev");
}

/*------------------------------------------------------------------------------
 * Action on a file system.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : ea_processAddFsysReq()
 *
 * Description  : Process the "add file system" request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 16-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processAddFsysReq(ea_addFsysReq_t * evReq,
                     ea_senderInfo_t * sender)

{
    ea_addFsysRsp_t *   eaRsp = NULL;
    ea_fsysCb_t         cb;
    int32               rc = NFX_OK;
	char				needAsyncRsp = 1; // The req needs a response
	uint32				rspLen = 0;
	int32				gotPendingRsp = 0;
	int32				i;
	uint32				localNode = 0;

    EA_FCT_PRINT(("> ea_processAddFsysReq()\n"));

#if 0
{
    int spin_wait = 1;
    while (spin_wait != 0)
        sleep(1);
}
#endif

    /* Initialize the response.
     */
	rspLen = sizeof(ea_addFsysRsp_t);
    eaRsp = (ea_addFsysRsp_t *) malloc(rspLen);
    NFX_ASSERT(eaRsp != NULL);
    bzero((void *) eaRsp, rspLen);
    eaRsp->hdr.msgType = EA_ADD_FSYS_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	strcpy(eaRsp->fsysCfg.fsysId.fsysName, evReq->fsysId.fsysName);

    /* Initialize the control block.
     */
    bzero((void *) &cb, sizeof(ea_fsysCb_t));
    cb.hdr = (ea_msgHdr_t *) evReq;

	localNode = isLocalNodeName(cb.hdr->node);
	if (localNode)
		*(cb.hdr->node) = '\0';	// local request, node should be null str

    cb.volSize = evReq->volSize * EVM_LBLK_SIZE;    // units???
    cb.vol = (evm_VolConfig_t *) malloc(sizeof(evm_VolConfig_t));
    NFX_ASSERT(cb.vol != NULL);
    bzero((void *) cb.vol, sizeof(evm_VolConfig_t));
    strcpy(cb.vol->name, evReq->fsysId.fsysName);
    cb.vol->id.whole_thing = evReq->fsysId.volId.whole_thing;
    cb.vol->maxSize = evReq->maxSize;
	if (evReq->minAutoGrow > 0)
		cb.vol->minAutoGrow = evReq->minAutoGrow;
	else
		cb.vol->minAutoGrow = DEFAULT_MIN_AUTOGROW;
	cb.vol->softQuota = evReq->softQuota;
    cb.vol->onNIM = evReq->nim;
    cb.vol->onLine = TRUE;
    memcpy((void *) &(cb.vol->attrs), (const void *) &(evReq->attr),
									sizeof(evm_strg_attributes_t));

    /* 
	 * Get a lock on this volume
     */
	rc = ea_childGetLock(evReq->fsysId.fsysName);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        sprintf(eaRsp->output, "Volume %s is busy", evReq->fsysId.fsysName);
        goto ea_processAddFsysReq_done_failed;
    }

    /* Create a volume.
     */
    rc = ea_crtVol(cb.hdr->node, &cb, eaRsp->output);
	if (rc == EVM_VOL_GROW_PENDING)
	{
		uint32 			byteRead = 0;
		uint32          rspSize;
		ea_msg_t		evmMsg;
		evm_MsgStat_t 	*evmRsp = NULL;

		gotPendingRsp = 1;

		/*
		 * We do not have enough LUNs. Once
		 * SPM/SRM get us the Luns we will
		 * move further for now return pending
		 * Send a Pending response to requestor
		 */
		eaRsp->result = EA_NFX_PENDING;
		EA_ERR_PRINT(("%s\n", eaRsp->output));
		ea_sendMsg(eaRsp, rspLen, sender->name);

		/*
	 	 * Figure out and remember who has sent you this 
		 * request. If response is pending 
		 *  - CLI will NOT wait for the asynchronous 
		 *  - AV will wait for the asynchronous 
		 * So remember if we need to forward
		 * our async response or not
		 */
		if (strcmp(sender->name, "ev") == 0)
			needAsyncRsp = 1;
		else
			needAsyncRsp = 0;

		rspSize = sizeof(evm_MsgStat_t);
		if (!localNode)
			rspSize += NCM_MSGHDR_LEN;

		/*
		 * Wait for the next Rsp from EVM
		 */
		byteRead = recv(myAsyncSock, evmMsg.buf, rspSize, 0);
		close(myAsyncSock);

		EA_DBG_PRINT(("Received from EVM on myAsyncSock\n"));
		/*
	 	 * Check userInfo and status field to determine
	 	 * what to do with the response
	 	 */

		if (!localNode)
		{
			byteRead -= NCM_MSGHDR_LEN;
			evmRsp = (evm_MsgStat_t *) (evmMsg.buf + NCM_MSGHDR_LEN);
		}
		else
			evmRsp = (evm_MsgStat_t *) evmMsg.buf;

		if (evmRsp->hdr.msgType != EVM_MSG_STAT)
		{
			EA_DBG_PRINT(("Received from EVM on myAsyncSock, msg type missmatch (received %d)\n", evmRsp->hdr.msgType));
			evmRsp->stat = NFX_ERR;
		}
	
		if (byteRead != sizeof(evm_MsgStat_t))
		{
			EA_DBG_PRINT(("Received from EVM on myAsyncSock, msg size missmatch (received len %d)\n", byteRead));
			evmRsp->stat = NFX_ERR;
		}
	
		if (evmRsp->stat != NFX_OK)
		{
			eaRsp->result = evmRsp->stat;
            sprintf(eaRsp->output, "Volume %s cannot be created. No more LUNs.",
                    evReq->fsysId.fsysName);
        	goto ea_processAddFsysReq_done_failed;
		}
	} 
	else if (rc != NFX_OK) 
	{
		/*
		 * Any other error
		 */
        eaRsp->result = rc;
        goto ea_processAddFsysReq_done_failed;
    }

    /* Get volume config.
     */
    rc = ea_getVolCfg(cb.hdr->node, cb.vol);
    if (rc != NFX_OK)
    {
		/*
		 * Error: Offline and Delete the volume
		 */
        eaRsp->result = rc;
        sprintf(eaRsp->output, "ea_getVolCfg() failed - %d", rc);
        goto ea_processAddFsysReq_done_delvol;
    }

    /* Make a file system.
     */
    rc = ea_crtFsys(cb.hdr->node, &cb);
    if ((rc != NFX_OK) && (rc != FS_ERR_ALREADY_MOUNTED))
    {
		/*
		 * Error: Offline and Delete the volume
		 */
        eaRsp->result = rc;
        sprintf(eaRsp->output, "ea_crtFsys() failed - %d", rc);
        goto ea_processAddFsysReq_done_delvol;
    }

    /* Mount the file system.
     */

    rc = ea_mntFsys(cb.hdr->node, &cb);
    if ((rc != FS_IN_PROGRESS) &&
        (rc != NFX_OK) && (rc != FS_ERR_ALREADY_MOUNTED))
    {
		/*
		 * Error: Offline and Delete the volume
		 * Del vol will remove file system too
		 */
        eaRsp->result = rc;
        sprintf(eaRsp->output, "ea_mntFsys() failed - %d", rc);
        goto ea_processAddFsysReq_done_delvol;
    }

    /* 
	 * Set Fsys high water mark
     */
    if (rc != FS_IN_PROGRESS)
	{
		if ((evReq->highWaterMark == DISABLE_HIGHWATERMARK) ||
			(evReq->highWaterMark < MIN_HIGHWATERMARK))
		{
			cb.newHighWaterMark = 100;
		}
		else
			cb.newHighWaterMark = evReq->highWaterMark;

        rc = ea_modFsys(cb.hdr->node, &cb);
        if (rc != NFX_OK)
        {
			/*
		 	 * Error: Offline and Delete the volume
		 	 * Del vol will remove file system too
		 	 */
			ea_umntFsys(cb.hdr->node, &cb);
	
        	eaRsp->result = rc;
			sprintf(eaRsp->output, "Could not set high water mark");
        	goto ea_processAddFsysReq_done_delvol;
        }
	}

	/*
	 * We are all done. Unlock the resources 
	 * before we send back a response
	 */
	ea_childUnLock();

	// Need to check whether volume information is in agile.conf
	// already. If not, need to wait it gets there...
	{
		int count;
		char regstr[512];
		char **value;

		// check for lvname first and get the volume ID string.
		sprintf(regstr, "lvname.%s", cb.vol->name);
		count = 60;		// check for about 60 seconds
		while (count--)
		{
			value = get_cfg_val(regstr);
			if (value != NULL)
			{
				sprintf(regstr, "cluster.lv.%s", value[1]);
				break;
			}
			sleep(1);
		}

		if (count > 0)
		{
			count = 60;		// another 60 seconds
			while (count--)
			{
				value = get_cfg_val(regstr);
				if (value != NULL)
					break;
				sleep(1);
			}
		}
	}

    /* Everything went fine.
	 * Update response fields and send response 
	 * with configuration.
     */
    if (needAsyncRsp)
	{
        eaRsp->result = NFX_OK;

        if (rc == FS_IN_PROGRESS)
        	strcpy(eaRsp->output, "FS mount in progress.");
    	else if (rc == FS_ERR_ALREADY_MOUNTED)
        	strcpy(eaRsp->output, "FS already mounted.");
        else 
        	strcpy(eaRsp->output, "Done.");

    	strcpy(eaRsp->fsysCfg.fsysId.fsysName, cb.vol->name);
   		memcpy(&eaRsp->fsysCfg.fsysId.volId, &cb.vol->id, sizeof(evm_volId_t));
		eaRsp->fsysCfg.onlineStatus = cb.vol->onLine;
		eaRsp->fsysCfg.maxSize = cb.vol->maxSize;
		eaRsp->fsysCfg.currSize = cb.vol->currSize;
		eaRsp->fsysCfg.minAutoGrow = cb.vol->minAutoGrow;
		eaRsp->fsysCfg.softQuota = cb.vol->softQuota;
		eaRsp->fsysCfg.raidLevel = cb.vol->attrs.raidLevel;
		eaRsp->fsysCfg.numVendors = cb.vol->attrs.num;
		for (i = 0; i < cb.vol->attrs.num; i++)
		{
			strcpy(eaRsp->fsysCfg.vendorInfo[i].vendor, 
								cb.vol->attrs.vendorInfo[i].vendor);
			strcpy(eaRsp->fsysCfg.vendorInfo[i].model, 
								cb.vol->attrs.vendorInfo[i].model);
		}
		eaRsp->fsysCfg.mountStatus = 1;
		ea_getFsysInfo(cb.hdr->node, cb.vol->id, &eaRsp->fsysCfg.fsState,
			  &eaRsp->fsysCfg.usedSize, &eaRsp->fsysCfg.highWaterMark);

		ea_sendMsg(eaRsp, rspLen, sender->name);
	}

	// if the is not a request from AV, we'll need
	// to notify AV about this...
	if (strcmp(sender->name, "ev"))
	{
		ea_fsysNtfy_t eaNtfy;

		memset(&eaNtfy, 0, sizeof(ea_fsysNtfy_t));

		eaNtfy.hdr.msgType = EA_FSYS_NTFY;
		eaNtfy.hdr.msgLen = sizeof(ea_fsysNtfy_t);

		eaNtfy.msgType = evReq->hdr.msgType;

    	strcpy(eaNtfy.fsysCfg.fsysId.fsysName, cb.vol->name);
   		memcpy(&eaNtfy.fsysCfg.fsysId.volId, &cb.vol->id, sizeof(evm_volId_t));
		eaNtfy.fsysCfg.onlineStatus = cb.vol->onLine;
		eaNtfy.fsysCfg.maxSize = cb.vol->maxSize;
		eaNtfy.fsysCfg.currSize = cb.vol->currSize;
		eaNtfy.fsysCfg.minAutoGrow = cb.vol->minAutoGrow;
		eaNtfy.fsysCfg.softQuota = cb.vol->softQuota;
		eaNtfy.fsysCfg.raidLevel = cb.vol->attrs.raidLevel;
		eaNtfy.fsysCfg.numVendors = cb.vol->attrs.num;
		for (i = 0; i < cb.vol->attrs.num; i++)
		{
			strcpy(eaNtfy.fsysCfg.vendorInfo[i].vendor, 
								cb.vol->attrs.vendorInfo[i].vendor);
			strcpy(eaNtfy.fsysCfg.vendorInfo[i].model, 
								cb.vol->attrs.vendorInfo[i].model);
		}
		eaNtfy.fsysCfg.mountStatus = 1;
		ea_getFsysInfo(cb.hdr->node, cb.vol->id, &eaNtfy.fsysCfg.fsState,
			  &eaNtfy.fsysCfg.usedSize, &eaNtfy.fsysCfg.highWaterMark);

		ea_sendMsg(&eaNtfy, sizeof(ea_fsysNtfy_t), "ev");
	}

	if (gotPendingRsp)
	{
		EA_ALERT_PRINT((
			"EA completed pending %s request for create volume %s\n",
				needAsyncRsp?"AV":"CLI", evReq->fsysId.fsysName));
	}
    
	goto ea_processAddFsysReq_done_succeed;


ea_processAddFsysReq_done_delvol:

	ea_offVol(cb.hdr->node, &cb, eaRsp->output);
	ea_delVol(cb.hdr->node, &cb, eaRsp->output);

ea_processAddFsysReq_done_failed:

    EA_ERR_PRINT(("%s\n", eaRsp->output));
	ea_childUnLock();
	if (needAsyncRsp)
    	ea_sendMsg(eaRsp, rspLen, sender->name);

ea_processAddFsysReq_done_succeed:

	if (eaRsp)
    	free(eaRsp);

    if (cb.vol)
		free(cb.vol);

    EA_FCT_PRINT(("< ea_processAddFsysReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processModFsysReq()
 *
 * Description  : Process the "modify file system" request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processModFsysReq(ea_modFsysReq_t * evReq,
                     ea_senderInfo_t * sender)

{
    ea_modFsysRsp_t *   eaRsp;
    ea_fsysCb_t         cb;
    int32               rc = NFX_OK;
	int32               i;
    
    EA_FCT_PRINT(("> ea_processModFsysReq()\n"));

    /* Initialize the response.
     */
    eaRsp = (ea_modFsysRsp_t *) malloc(sizeof(ea_modFsysRsp_t));
    NFX_ASSERT(eaRsp != NULL);
    bzero((void *) eaRsp, sizeof(ea_modFsysRsp_t));
    eaRsp->hdr.msgType = EA_MOD_FSYS_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(&eaRsp->fsysCfg.fsysId, &evReq->fsysId, sizeof(ea_fsysId_t));

    /* Initialize the control block.
     */
    bzero((void *) &cb, sizeof(ea_fsysCb_t));
    cb.hdr = (ea_msgHdr_t *) evReq;
    cb.vol = (evm_VolConfig_t *) malloc(sizeof(evm_VolConfig_t));
    NFX_ASSERT(cb.vol != NULL);
    bzero((void *) cb.vol, sizeof(evm_VolConfig_t));
    strcpy(cb.vol->name, evReq->fsysId.fsysName);
    cb.vol->onNIM = 1;

	// Get volume id from volume name, and node name that
	// has this volume online...
	rc = ea_getClusterVolIdNode(cb.vol, cb.hdr->node, eaRsp->output);
	// rc = ea_getVolId(cb.vol, eaRsp->output);
	if (rc != NFX_OK)
	{
		eaRsp->result = rc;
		EA_ERR_PRINT(("%s\n", eaRsp->output));
		ea_sendMsg(eaRsp, sizeof(ea_modFsysRsp_t), sender->name);
		free(cb.vol);
		free(eaRsp);
		return;
    }

    /* Get volume config.
     */
    rc = ea_getVolCfg(cb.hdr->node, cb.vol);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        sprintf(eaRsp->output, "ea_getVolCfg() failed - %d", rc);
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_modFsysRsp_t), sender->name);
		free(cb.vol);
        free(eaRsp);
        return;
    }

	/*
	 * Can't change the name if file system is being
	 * exported
	 */
	if ((IS_FSYS_NAME(evReq->mask)) &&
			ea_fsysIsExported(cb.hdr->node, cb.vol->id))
	{
        eaRsp->result = NFX_ERR;
        sprintf(eaRsp->output, "Operation not allowed on exported volume");
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_modFsysRsp_t), sender->name);
		free(cb.vol);
        free(eaRsp);
        return;
	}

    /* 
	 * Get a lock on this volume
     */
	rc = ea_childGetLock(evReq->fsysId.fsysName);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        sprintf(eaRsp->output, "Volume %s is busy", evReq->fsysId.fsysName);
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_modFsysRsp_t), sender->name);
		free(cb.vol);
        free(eaRsp);
        return;
    }

    if (IS_FSYS_NAME(evReq->mask))
    	strcpy(cb.newFsysName, evReq->newFsysName);
	else
		strcpy(cb.newFsysName, cb.vol->name);

    if (IS_FSYS_HQUOTA(evReq->mask))
		cb.newHardQuota = evReq->hardQuota;
	else
		cb.newHardQuota = cb.vol->maxSize;

    if (IS_FSYS_SQUOTA(evReq->mask))
		cb.newSoftQuota = evReq->softQuota;
	else
		cb.newSoftQuota = cb.vol->softQuota;

	if (IS_FSYS_MINAUTOGROW(evReq->mask))
	{
		if (evReq->minAutoGrow > 0)
			cb.newMinAutoGrow = evReq->minAutoGrow;
		else
			cb.newMinAutoGrow = DEFAULT_MIN_AUTOGROW;
	}
	else
    	cb.newMinAutoGrow = cb.vol->minAutoGrow;

    if (IS_FSYS_HIGHWATERMARK(evReq->mask))
	{
        /* Modify Fsys high water mark
         */
		if ((evReq->highWaterMark == DISABLE_HIGHWATERMARK) ||
			(evReq->highWaterMark < MIN_HIGHWATERMARK))
		{
			cb.newHighWaterMark = 100;
		}
		else
			cb.newHighWaterMark = evReq->highWaterMark;

        rc = ea_modFsys(cb.hdr->node, &cb);
        if (rc != NFX_OK)
        {
            eaRsp->result = rc;
			sprintf(eaRsp->output, "Could not modify high water mark");
            EA_ERR_PRINT(("%s\n", eaRsp->output));
			ea_childUnLock();
            ea_sendMsg(eaRsp, sizeof(ea_modFsysRsp_t), sender->name);
			free(cb.vol);
            free(eaRsp);
            return;
        }
	}

    if (IS_FSYS_NAME(evReq->mask))
    {
        /* Modify volume name.
         */
        rc = ea_renameVol(cb.hdr->node, &cb, eaRsp->output);
        if (rc != NFX_OK)
        {
            eaRsp->result = rc;
            EA_ERR_PRINT(("%s\n", eaRsp->output));
			ea_childUnLock();
            ea_sendMsg(eaRsp, sizeof(ea_modFsysRsp_t), sender->name);
			free(cb.vol);
            free(eaRsp);
            return;
        }
    }

    if ((IS_FSYS_HQUOTA(evReq->mask)) ||
		(IS_FSYS_SQUOTA(evReq->mask)) ||
		(IS_FSYS_MINAUTOGROW(evReq->mask)))
    {
        /* Modify hardquota/softquota/minautogrowth of the volume.
         */
        rc = ea_modVol(cb.hdr->node, &cb, eaRsp->output);
        if (rc != NFX_OK)
        {
            eaRsp->result = rc;
            EA_ERR_PRINT(("%s\n", eaRsp->output));
			ea_childUnLock();
            ea_sendMsg(eaRsp, sizeof(ea_modFsysRsp_t), sender->name);
			free(cb.vol);
            free(eaRsp);
            return;
        }
    }

    /* Everything went fine.
     * Send response with configuration.
     */
    eaRsp->result = NFX_OK;
    strcpy(eaRsp->output, "Done.");
    eaRsp->hdr.msgType = EA_MOD_FSYS_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(&eaRsp->fsysCfg.fsysId, &evReq->fsysId, sizeof(ea_fsysId_t));

	/*
	 * Update the desired fields to new values
	 */
    if (IS_FSYS_NAME(evReq->mask))
    	strcpy(cb.vol->name, cb.newFsysName);

    if (IS_FSYS_HQUOTA(evReq->mask))
		cb.vol->maxSize = cb.newHardQuota;

    if (IS_FSYS_SQUOTA(evReq->mask))
		cb.vol->softQuota = cb.newSoftQuota;

	if (IS_FSYS_MINAUTOGROW(evReq->mask))
        cb.vol->minAutoGrow = cb.newMinAutoGrow;

	/*
	 * Send back latest config in response
	 */
    strcpy(eaRsp->fsysCfg.fsysId.fsysName, cb.vol->name);
	eaRsp->fsysCfg.onlineStatus = cb.vol->onLine;
	eaRsp->fsysCfg.maxSize = cb.vol->maxSize;
	eaRsp->fsysCfg.currSize = cb.vol->currSize;
	eaRsp->fsysCfg.minAutoGrow = cb.vol->minAutoGrow;
	eaRsp->fsysCfg.softQuota = cb.vol->softQuota;
	eaRsp->fsysCfg.raidLevel = cb.vol->attrs.raidLevel;
	eaRsp->fsysCfg.numVendors = cb.vol->attrs.num;
	for (i = 0; i < cb.vol->attrs.num; i++)
	{
		strcpy(eaRsp->fsysCfg.vendorInfo[i].vendor, 
							cb.vol->attrs.vendorInfo[i].vendor);
		strcpy(eaRsp->fsysCfg.vendorInfo[i].model, 
							cb.vol->attrs.vendorInfo[i].model);
	}
	eaRsp->fsysCfg.mountStatus = 1;
	ea_getFsysInfo(cb.hdr->node, cb.vol->id, &eaRsp->fsysCfg.fsState,
			&eaRsp->fsysCfg.usedSize, &eaRsp->fsysCfg.highWaterMark);

	/*
	 * We are all done. Unlock the resources 
	 * before we send back a response
	 */
	ea_childUnLock();
    ea_sendMsg(eaRsp, sizeof(ea_modFsysRsp_t), sender->name);
    free(eaRsp);

	// if the is not a request from AV, we'll need
	// to notify AV about this...
	if (strcmp(sender->name, "ev"))
	{
		ea_fsysNtfy_t eaNtfy;

		memset(&eaNtfy, 0, sizeof(ea_fsysNtfy_t));

		eaNtfy.hdr.msgType = EA_FSYS_NTFY;
		eaNtfy.hdr.msgLen = sizeof(ea_fsysNtfy_t);

		eaNtfy.msgType = evReq->hdr.msgType;

		strcpy(eaNtfy.fsysCfg.fsysId.fsysName, cb.vol->name);
   		memcpy(&eaNtfy.fsysCfg.fsysId.volId, &cb.vol->id, sizeof(evm_volId_t));
		eaNtfy.fsysCfg.onlineStatus = cb.vol->onLine;
		eaNtfy.fsysCfg.maxSize = cb.vol->maxSize;
		eaNtfy.fsysCfg.currSize = cb.vol->currSize;
		eaNtfy.fsysCfg.minAutoGrow = cb.vol->minAutoGrow;
		eaNtfy.fsysCfg.softQuota = cb.vol->softQuota;
		eaNtfy.fsysCfg.raidLevel = cb.vol->attrs.raidLevel;
		eaNtfy.fsysCfg.mountStatus = 1;
		eaNtfy.fsysCfg.numVendors = cb.vol->attrs.num;
		for (i = 0; i < cb.vol->attrs.num; i++)
		{
			strcpy(eaNtfy.fsysCfg.vendorInfo[i].vendor, 
								cb.vol->attrs.vendorInfo[i].vendor);
			strcpy(eaNtfy.fsysCfg.vendorInfo[i].model, 
								cb.vol->attrs.vendorInfo[i].model);
		}
		ea_getFsysInfo(cb.hdr->node, cb.vol->id, &eaNtfy.fsysCfg.fsState,
			  &eaNtfy.fsysCfg.usedSize, &eaNtfy.fsysCfg.highWaterMark);

		ea_sendMsg(&eaNtfy, sizeof(ea_fsysNtfy_t), "ev");
	}

    /* Free the volume configuration.
     */
    free(cb.vol);

    EA_FCT_PRINT(("< ea_processModFsysReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processMntFsysReq()
 *
 * Description  : Process the "mount file system" request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 08-NOV-2001
 *
 * !!!NOTE!!!!  : This function will assume that this volume is already
 *                lvol onlined...
 *------------------------------------------------------------------------------
 */

static void
ea_processMntFsysReq(ea_mntFsysReq_t * evReq,
                     ea_senderInfo_t * sender)

{
    ea_mntFsysRsp_t *   eaRsp;
    int32               rc = NFX_OK;
    ea_fsysCb_t         cb;
    
    EA_FCT_PRINT(("> ea_processMntFsysReq()\n"));

    /* Initialize the response.
     */
    eaRsp = (ea_mntFsysRsp_t *) malloc(sizeof(ea_mntFsysRsp_t));
    NFX_ASSERT(eaRsp != NULL);
    bzero((void *) eaRsp, sizeof(ea_mntFsysRsp_t));
    eaRsp->hdr.msgType = EA_MNT_FSYS_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(&eaRsp->fsysId, &evReq->fsysId, sizeof(ea_fsysId_t));
    
    /* Initialize the control block.
     */
    bzero((void *) &cb, sizeof(ea_fsysCb_t));
    cb.hdr = (ea_msgHdr_t *) evReq;
    cb.vol = (evm_VolConfig_t *) malloc(sizeof(evm_VolConfig_t));
    NFX_ASSERT(cb.vol != NULL);
    bzero((void *) cb.vol, sizeof(evm_VolConfig_t));
    strcpy(cb.vol->name, evReq->fsysId.fsysName);
    cb.vol->onNIM = 1;
    
	// Get volume id from volume name, and node name that
	// has this volume online...
	rc = ea_getClusterVolIdNode(cb.vol, cb.hdr->node, eaRsp->output);
	// rc = ea_getVolId(cb.vol, eaRsp->output);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_mntFsysRsp_t), sender->name);
    	free(cb.vol);
        free(eaRsp);
        return;
    }

    /* Get volume config.
     */
    rc = ea_getVolCfg(cb.hdr->node, cb.vol);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        sprintf(eaRsp->output, "ea_getVolCfg() failed - %d", rc);
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_mntFsysRsp_t), sender->name);
    	free(cb.vol);
        free(eaRsp);
        return;
    }

    /* 
	 * Get a lock on this volume
     */
	rc = ea_childGetLock(evReq->fsysId.fsysName);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        sprintf(eaRsp->output, "Volume %s is busy", evReq->fsysId.fsysName);
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_mntFsysRsp_t), sender->name);
		free(cb.vol);
        free(eaRsp);
        return;
    }

    /* Mount the file system.
     */
    rc = ea_mntFsys(cb.hdr->node, &cb);
    if ((rc != FS_IN_PROGRESS) &&
        (rc != NFX_OK) && (rc != FS_ERR_ALREADY_MOUNTED))
    {
        eaRsp->result = rc;
        sprintf(eaRsp->output, "ea_mntFsys() failed - %d", rc);
        EA_ERR_PRINT(("%s\n", eaRsp->output));
		ea_childUnLock();
        ea_sendMsg(eaRsp, sizeof(ea_mntFsysRsp_t), sender->name);
   	 	free(cb.vol);
        free(eaRsp);
        return;
    }
    
    eaRsp->result = NFX_OK;

    if (rc == FS_IN_PROGRESS)
    	strcpy(eaRsp->output, "FS mount in progress.");
	else if (rc == FS_ERR_ALREADY_MOUNTED)
    	strcpy(eaRsp->output, "FS already mounted.");
    else 
    	strcpy(eaRsp->output, "Done.");

	/*
	 * We are all done. Unlock the resources 
	 * before we send back a response
	 */
	ea_childUnLock();
    ea_sendMsg(eaRsp, sizeof(ea_mntFsysRsp_t), sender->name);
    free(eaRsp);
    
    /* Free the volume configuration.
     */
    free(cb.vol);

    EA_FCT_PRINT(("< ea_processMntFsysReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processUmntFsysReq()
 *
 * Description  : Process the "unmount file system" request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 08-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processUmntFsysReq(ea_umntFsysReq_t * evReq,
                      ea_senderInfo_t *  sender)

{
    ea_umntFsysRsp_t *  eaRsp;
    int32               rc = NFX_OK;
    ea_fsysCb_t         cb;
	uint32 				myFsState = 0;
	uint64				myUsedSize = 0;
	uint32				myHighWaterMark = 0;
    
    EA_FCT_PRINT(("> ea_processUmntFsysReq()\n"));

    /* Initialize the response.
     */
    eaRsp = (ea_umntFsysRsp_t *) malloc(sizeof(ea_umntFsysRsp_t));
    NFX_ASSERT(eaRsp != NULL);
    bzero((void *) eaRsp, sizeof(ea_umntFsysRsp_t));
    eaRsp->hdr.msgType = EA_UMNT_FSYS_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(&eaRsp->fsysId, &evReq->fsysId, sizeof(ea_fsysId_t));
    
    /* Initialize the control block.
     */
    bzero((void *) &cb, sizeof(ea_fsysCb_t));
    cb.hdr = (ea_msgHdr_t *) evReq;
    cb.vol = (evm_VolConfig_t *) malloc(sizeof(evm_VolConfig_t));
    NFX_ASSERT(cb.vol != NULL);
    bzero((void *) cb.vol, sizeof(evm_VolConfig_t));
    strcpy(cb.vol->name, evReq->fsysId.fsysName);
    cb.vol->onNIM = 1;
    
	// Get volume id from volume name, and node name that
	// has this volume online...
	rc = ea_getClusterVolIdNode(cb.vol, cb.hdr->node, eaRsp->output);
	// rc = ea_getVolId(cb.vol, eaRsp->output);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_umntFsysRsp_t), sender->name);
    	free(cb.vol);
        free(eaRsp);
        return;
    }

    /* Get volume config.
     */
    rc = ea_getVolCfg(cb.hdr->node, cb.vol);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        sprintf(eaRsp->output, "ea_getVolCfg() failed - %d", rc);
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_umntFsysRsp_t), sender->name);
    	free(cb.vol);
        free(eaRsp);
        return;
    }

	/*
	 * Make sure this is not exported
	 */
	if(ea_fsysIsExported(cb.hdr->node, cb.vol->id))
	{
        eaRsp->result = NFX_ERR;
        sprintf(eaRsp->output, "Operation not allowed on exported volume");
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_umntFsysRsp_t), sender->name);
		free(cb.vol);
        free(eaRsp);
        return;
	}

	/*
	 * Make sure FS is not in the middle of something
	 */
	ea_getFsysInfo(cb.hdr->node, cb.vol->id,
				&myFsState, &myUsedSize, &myHighWaterMark);

	if ((myFsState == FS_VOL_STATE_MOUNTING) || 
		(myFsState == FS_VOL_STATE_UNMOUNTING) || 
		(myFsState == FS_VOL_STATE_FAILING) ||
		(myFsState == FS_VOL_STATE_PAUSED) ||
		(myFsState == FS_VOL_STATE_BACKUP_RESTORE))
	{
        eaRsp->result = NFX_ERR;
        sprintf(eaRsp->output, "FS state %s. Operation not allowed",
								fsStateStr[myFsState]);
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_umntFsysRsp_t), sender->name);
		free(cb.vol);
        free(eaRsp);
        return;
	}

    /* 
	 * Get a lock on this volume
     */
	rc = ea_childGetLock(evReq->fsysId.fsysName);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        sprintf(eaRsp->output, "Volume %s is busy", evReq->fsysId.fsysName);
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_umntFsysRsp_t), sender->name);
		free(cb.vol);
        free(eaRsp);
        return;
    }

    /* Unmount the file system.
     */
    rc = ea_umntFsys(cb.hdr->node, &cb);
    if ((rc != NFX_OK) && (rc != FS_ERR_NOT_MOUNTED))
    {
        eaRsp->result = rc;
        sprintf(eaRsp->output, "ea_umntFsys() failed - %d", rc);
        EA_ERR_PRINT(("%s\n", eaRsp->output));
		ea_childUnLock();
        ea_sendMsg(eaRsp, sizeof(ea_umntFsysRsp_t), sender->name);
    	free(cb.vol);
        free(eaRsp);
        return;
    }

    eaRsp->result = NFX_OK;
    sprintf(eaRsp->output, "Done.");

	/*
	 * We are all done. Unlock the resources 
	 * before we send back a response
	 */
	ea_childUnLock();
    ea_sendMsg(eaRsp, sizeof(ea_umntFsysRsp_t), sender->name);
    free(eaRsp);

    /* Free the volume configuration.
     */
    free(cb.vol);

    EA_FCT_PRINT(("< ea_processUmntFsysReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processDelFsysReq()
 *
 * Description  : Process the "delete file system" request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 16-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processDelFsysReq(ea_delFsysReq_t * evReq,
                     ea_senderInfo_t * sender)

{
    ea_delFsysRsp_t *   eaRsp = NULL;
    int32               rc = NFX_OK;
    ea_fsysCb_t         cb;
	uint32 				myFsState = 0;
	uint64				myUsedSize = 0;
	uint32				myHighWaterMark = 0;
    
    EA_FCT_PRINT(("> ea_processDelFsysReq()\n"));

    /* Initialize the response.
     */
    eaRsp = (ea_delFsysRsp_t *) malloc(sizeof(ea_delFsysRsp_t));
    NFX_ASSERT(eaRsp != NULL);

    bzero((void *) eaRsp, sizeof(ea_delFsysRsp_t));
    eaRsp->hdr.msgType = EA_DEL_FSYS_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(&eaRsp->fsysId, &evReq->fsysId, sizeof(ea_fsysId_t));
    
    /* Initialize the control block.
     */
    bzero((void *) &cb, sizeof(ea_fsysCb_t));
    cb.hdr = (ea_msgHdr_t *) evReq;
    cb.vol = (evm_VolConfig_t *) malloc(sizeof(evm_VolConfig_t));
    NFX_ASSERT(cb.vol != NULL);
    bzero((void *) cb.vol, sizeof(evm_VolConfig_t));
    strcpy(cb.vol->name, evReq->fsysId.fsysName);
	memcpy(&eaRsp->fsysId, &evReq->fsysId, sizeof(ea_fsysId_t));
    cb.vol->onNIM = 1;

	// Get volume id from volume name, and node name that
	// has this volume online...
	rc = ea_getClusterVolIdNode(cb.vol, cb.hdr->node, eaRsp->output);
	// rc = ea_getVolId(cb.vol, eaRsp->output);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
		goto ea_processDelFsysReq_done;
    }

    /* Get volume config.
     */
    rc = ea_getVolCfg(cb.hdr->node, cb.vol);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        sprintf(eaRsp->output, "ea_getVolCfg() failed - %d", rc);
		goto ea_processDelFsysReq_done;
    }

	/*
	 * Make sure this is not exported
	 */
	if(ea_fsysIsExported(cb.hdr->node, cb.vol->id) == 1)
	{
        eaRsp->result = NFX_ERR;
        sprintf(eaRsp->output, "Operation not allowed on exported volume");
		goto ea_processDelFsysReq_done;
	}

	/*
	 * Make sure FS is not in the middle of something
	 */
	ea_getFsysInfo(cb.hdr->node, cb.vol->id,
				&myFsState, &myUsedSize, &myHighWaterMark);

	if ((myFsState == FS_VOL_STATE_MOUNTING) || 
		(myFsState == FS_VOL_STATE_UNMOUNTING) || 
		(myFsState == FS_VOL_STATE_FAILING) ||
		(myFsState == FS_VOL_STATE_PAUSED) ||
		(myFsState == FS_VOL_STATE_BACKUP_RESTORE))
	{
        eaRsp->result = NFX_ERR;
        sprintf(eaRsp->output, "FS state %s. Operation not allowed",
								fsStateStr[myFsState]);
		goto ea_processDelFsysReq_done;
	}

    /* 
	 * Get a lock on this volume
     */
	rc = ea_childGetLock(evReq->fsysId.fsysName);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        sprintf(eaRsp->output, "Volume %s is busy", evReq->fsysId.fsysName);
		goto ea_processDelFsysReq_done;
    }

    /* Unmount the file system.
     */
    rc = ea_umntFsys(cb.hdr->node, &cb);
    if ((rc != NFX_OK) && (rc != FS_ERR_NOT_MOUNTED))
    {
        eaRsp->result = rc;
        sprintf(eaRsp->output, "ea_umntFsys() failed - %d", rc);
		goto ea_processDelFsysReq_unlock;
    }

    /* Put off-line the volume.
     */
    rc = ea_offVol(cb.hdr->node, &cb, eaRsp->output);
    if ((rc != NFX_OK) && (rc != EVM_VOL_ALREADY_OFFLINE))
    {
    	eaRsp->result = rc;
		goto ea_processDelFsysReq_unlock;
    }

    /* Destroy a volume.
     */
    rc = ea_delVol(cb.hdr->node, &cb, eaRsp->output);
    if (rc != NFX_OK)
    {
    	eaRsp->result = rc;
		goto ea_processDelFsysReq_unlock;
    }
    
    if (eaRsp->result == NFX_OK)
	{
		int count;
		char regstr[512];
		char **value;

		// Need to check whether volume information is removed from
		// agile.conf. If not, need to wait until it got deleted.

		// check for lvname first and get the volume ID string.
		sprintf(regstr, "lvname.%s", cb.vol->name);
		count = 60;		// check for about 60 seconds
		while (count--)
		{
			value = get_cfg_val(regstr);
			if (value == NULL)
				break;
			sleep(1);
		}

    	sprintf(eaRsp->output, "Done.");
	}

	memcpy(&eaRsp->fsysId.volId, &cb.vol->id, sizeof(evm_volId_t));

	// if the is not a request from AV, we'll need
	// to notify AV about this...
	if (strcmp(sender->name, "ev"))
	{
		ea_fsysNtfy_t eaNtfy;

		memset(&eaNtfy, 0, sizeof(ea_fsysNtfy_t));

		eaNtfy.hdr.msgType = EA_FSYS_NTFY;
		eaNtfy.hdr.msgLen = sizeof(ea_fsysNtfy_t);

		eaNtfy.msgType = evReq->hdr.msgType;

   		memcpy(&eaNtfy.fsysCfg.fsysId.volId, &cb.vol->id, sizeof(evm_volId_t));

		ea_sendMsg(&eaNtfy, sizeof(ea_fsysNtfy_t), "ev");
	}

ea_processDelFsysReq_unlock:
	// We are all done. Unlock the resources 
	// before we send back a response
	ea_childUnLock();

ea_processDelFsysReq_done:
    // Free the volume configuration.
	EA_ERR_PRINT(("%s\n", eaRsp->output));
	ea_sendMsg(eaRsp, sizeof(ea_delFsysRsp_t), sender->name);
	free(cb.vol);
	free(eaRsp);

    EA_FCT_PRINT(("< ea_processDelFsysReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processGetFsysReq()
 *
 * Description  : Process the "show all file system" request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 21-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processGetFsysReq(ea_getFsysReq_t * evReq,
                      ea_senderInfo_t *  sender)

{
	int 				i, startIdx, totNum;
	char				*node = NULL;
    ea_fsysList_t *     fsysList = NULL;
    ea_getFsysRsp_t *  	eaRsp;
    uint32              rspLen = sizeof(ea_getFsysRsp_t);
    
    EA_FCT_PRINT(("> ea_processGetFsysReq()\n"));

    /* Initialize the response.
     */
    eaRsp = (ea_getFsysRsp_t *) malloc(rspLen);
    NFX_ASSERT(eaRsp != NULL);
    bzero((void *) eaRsp, rspLen);
    eaRsp->hdr.msgType = EA_GET_FSYS_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;

	// evReq->hdr.node should specify a cluster node
	// name if no volume name is specified and this
	// is a remote node operation...
	if (isLocalNodeName(evReq->hdr.node))
		node = NULL;
	else
		node = evReq->hdr.node;

    // Get list of volumes (id + name).
    fsysList = ea_getVolList(node);
    if (!fsysList)
    {
        eaRsp->result = NFX_ERR;
        sprintf(eaRsp->output, "ea_getVolList() failed");
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, rspLen, sender->name);
        free(eaRsp);
        return;
    }
  
	totNum = fsysList->nb;
  	if ((evReq->reqType == EA_REQ_HAS_FSYS_INIT) &&
		(evReq->fsysId[0].volId.whole_thing != 0))
	{
		for (i = 0; i < fsysList->nb; i++)
		{
			totNum--;
			if (fsysList->fsys[i].fsysId.volId.whole_thing ==
						evReq->fsysId[0].volId.whole_thing)
			{
				startIdx = ++i;
				break;
			}
		}
	}

   	if (evReq->num)
	{
		if (totNum > evReq->num)
			fsysList->nb = evReq->num;
		else
			fsysList->nb = totNum;
	}

    if (fsysList->nb > 1)
    {
        rspLen += (fsysList->nb - 1) * sizeof(ea_fsysCfg_t);
        eaRsp = (ea_getFsysRsp_t *) realloc((void *) eaRsp, rspLen);
    	bzero((void *) eaRsp, rspLen);
    	eaRsp->hdr.msgType = EA_GET_FSYS_RSP;
    	eaRsp->hdr.userInfo = evReq->hdr.userInfo;
    }

	if (evReq->reqType == EA_REQ_HAS_FSYS_INIT)
	{
		eaRsp->fsysList.nb = fsysList->nb;
		for (i = startIdx; i < (startIdx + fsysList->nb); i++)
		{
    		memcpy((void *) &(eaRsp->fsysList.fsys[i-startIdx]), 
					(const void *) &fsysList->fsys[i], sizeof(ea_fsysCfg_t));
		}
	}
	else if (evReq->reqType == EA_REQ_HAS_FSYS_LIST)
	{
		uint32 rspIdx = 0;

		for (i = 0; (i < totNum) && (rspIdx < fsysList->nb); i++)
		{
			if (fsysList->fsys[i].fsysId.volId.whole_thing ==
						evReq->fsysId[rspIdx].volId.whole_thing)
			{
    			memcpy((void *) &(eaRsp->fsysList.fsys[rspIdx]), 
					(const void *) &fsysList->fsys[i], sizeof(ea_fsysCfg_t));
				rspIdx++;
				eaRsp->fsysList.nb++;
			}
		}
	}

    free(fsysList);

    eaRsp->result = NFX_OK;
    sprintf(eaRsp->output, "Done.");
    ea_sendMsg(eaRsp, rspLen, sender->name);
    free(eaRsp);

    EA_FCT_PRINT(("< ea_processGetFsysReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processGetDetailFsysReq()
 *
 * Description  : Show the specified file system in detail
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 21-NOV-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processGetDetailFsysReq(ea_getDetailFsysReq_t * evReq,
                      ea_senderInfo_t *  sender)

{
	evm_lvStrg_t		*strgList = NULL;
    ea_fsysCb_t         cb;
    ea_getDetailFsysRsp_t 	*eaRsp;
    int32               rc = NFX_OK;
    uint32              rspLen = sizeof(ea_getDetailFsysRsp_t);
	int					i;
	uint32				numLuns;
   
    EA_FCT_PRINT(("> ea_processGetDetailFsysReq()\n"));

    /* Initialize the response.
     */
    eaRsp = (ea_getDetailFsysRsp_t *) malloc(rspLen);
    NFX_ASSERT(eaRsp != NULL);
    bzero((void *) eaRsp, rspLen);
    eaRsp->hdr.msgType = EA_GET_DETAIL_FSYS_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;

    /* Initialize the control block.
     */
    bzero((void *) &cb, sizeof(ea_fsysCb_t));
    cb.hdr = (ea_msgHdr_t *) evReq;
    cb.vol = (evm_VolConfig_t *) malloc(sizeof(evm_VolConfig_t));
    NFX_ASSERT(cb.vol != NULL);
    bzero((void *) cb.vol, sizeof(evm_VolConfig_t));
    strcpy(cb.vol->name, evReq->fsysId.fsysName);
    cb.vol->onNIM = 1;

	// Get volume id from volume name, and node name that
	// has this volume online...
	rc = ea_getClusterVolIdNode(cb.vol, cb.hdr->node, eaRsp->output);
	// rc = ea_getVolId(cb.vol, eaRsp->output);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_getDetailFsysRsp_t), sender->name);
		free(cb.vol);
        free(eaRsp);
        return;
    }

    /* Get volume config.
     */
    rc = ea_getVolCfg(cb.hdr->node, cb.vol);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        sprintf(eaRsp->output, "ea_getVolCfg() failed - %d", rc);
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_getDetailFsysRsp_t), sender->name);
		free(cb.vol);
        free(eaRsp);
        return;
    }

	/*
	 * Get LUN list
	 */
	numLuns = 0;
	strgList = ea_getVolLunList(cb.hdr->node, cb.vol->id, &numLuns);
	if (strgList == NULL)
	{
        eaRsp->result = rc;
        sprintf(eaRsp->output, "ea_getVolLunList() failed");
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_getDetailFsysRsp_t), sender->name);
		free(cb.vol);
        free(eaRsp);
        return;
	}

	if (numLuns > 1)
	{
		/*
		 * Need to reallocate bigger memory block
		 * and reinitialize header
		 */
		rspLen += (numLuns - 1) * sizeof(evm_lvStrg_t);
		eaRsp = (ea_getDetailFsysRsp_t *) realloc((void *) eaRsp, rspLen);
		bzero((void *) eaRsp, rspLen);
    	eaRsp->hdr.msgType = EA_GET_DETAIL_FSYS_RSP;
    	eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	}

	/*
	 * Everything went OK
	 * Setup the response and send
	 */
    eaRsp->result = NFX_OK;
    sprintf(eaRsp->output, "Done.");
    memcpy(&eaRsp->fsysDetail.volCfg, cb.vol, sizeof(evm_VolConfig_t));
	if (strgList)
	{
		eaRsp->fsysDetail.nbStrg = numLuns;
		for (i = 0; i < numLuns; i++)
		{
			memcpy(&eaRsp->fsysDetail.strg[i], &strgList[i],
						sizeof(evm_lvStrg_t));
#ifdef EA_DEBUG
			printf("ea_processGetDetailFsysReq packed %d %s 0x%llx %d\n",
					i, eaRsp->fsysDetail.strg[i].name,
					eaRsp->fsysDetail.strg[i].wwn,
					eaRsp->fsysDetail.strg[i].lun,
					eaRsp->fsysDetail.strg[i].au.auSize);
#endif
		}
	}
	eaRsp->fsysDetail.mountStatus = 1;
	ea_getFsysInfo(cb.hdr->node, cb.vol->id,
							&eaRsp->fsysDetail.fsState,
	                        &eaRsp->fsysDetail.usedSize,
							&eaRsp->fsysDetail.highWaterMark);

    ea_sendMsg(eaRsp, rspLen, sender->name);

	free(strgList);
    free(eaRsp);
	free(cb.vol);

    EA_FCT_PRINT(("< ea_processGetDetailFsysReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processUsgFsysReq()
 *
 * Description  : Show the specified file system in detail
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 21-NOV-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processUsgFsysReq(ea_usgFsysReq_t * evReq,
                      ea_senderInfo_t *  sender)

{
	int 				i, rspIdx, numEntry;
	char				*node = NULL;
    ea_fsysList_t *     fsysList;
    ea_usgFsysRsp_t *  	eaRsp;
    uint32              rspLen = sizeof(ea_usgFsysRsp_t);
	uint32				dummyFsState = 0;
	uint32				dummyHighWaterMark = 0;
    
    EA_FCT_PRINT(("> ea_processUsgFsysReq()\n"));

    /* Initialize the response.
     */
    eaRsp = (ea_usgFsysRsp_t *) malloc(rspLen);
    NFX_ASSERT(eaRsp != NULL);
    bzero((void *) eaRsp, rspLen);
    eaRsp->hdr.msgType = EA_USG_FSYS_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;

	if (!isLocalNodeName(evReq->hdr.node))
		node = evReq->hdr.node;

    // Get list of volumes (id + name).
    fsysList = ea_getVolList(node);
    if (!fsysList)
    {
        eaRsp->result = NFX_ERR;
        sprintf(eaRsp->output, "ea_getVolList() failed");
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, rspLen, sender->name);
        free(eaRsp);
        return;
    }

	/*
	 * How many entries to return
	 */
   	if (evReq->num)
	{
		if (fsysList->nb > evReq->num)
			numEntry = evReq->num;
		else
			numEntry = fsysList->nb;
	}
	else
		numEntry = fsysList->nb;

    if (numEntry > 1)
    {
        rspLen += (sizeof(ea_fsysList_t) + (numEntry - 1) * sizeof(fsysUsage_t));
        eaRsp = (ea_usgFsysRsp_t *) realloc((void *) eaRsp, rspLen);
    	bzero((void *) eaRsp, rspLen);
    	eaRsp->hdr.msgType = EA_USG_FSYS_RSP;
    	eaRsp->hdr.userInfo = evReq->hdr.userInfo;
    }

	/*
	 * Get all entries
	 */
	rspIdx = 0;
	for (i = 0; (i < fsysList->nb) && (rspIdx < numEntry); i++)
	{
		if ((fsysList->fsys[i].fsysId.volId.whole_thing ==
					evReq->fsysList[rspIdx].volId.whole_thing) ||
				(evReq->num == 0))
		{
			strcpy(eaRsp->usageList[rspIdx].fsysId.fsysName,
					fsysList->fsys[i].fsysId.fsysName);
			eaRsp->usageList[rspIdx].fsysId.volId.whole_thing =
							fsysList->fsys[i].fsysId.volId.whole_thing;
			ea_getFsysInfo(node, fsysList->fsys[i].fsysId.volId, 
						&dummyFsState, &eaRsp->usageList[rspIdx].usedSize,
						&dummyHighWaterMark);
			eaRsp->usageList[rspIdx].allocatedSize =
						fsysList->fsys[i].currSize;
			rspIdx++;
		}
	}

    free(fsysList);

    eaRsp->result = NFX_OK;
    sprintf(eaRsp->output, "Done.");

	eaRsp->num = rspIdx;
	if (eaRsp->num <= 0)
	{
		rspLen = sizeof(ea_usgFsysRsp_t);
	}
	else
	{
		rspLen = sizeof(ea_usgFsysRsp_t) + 
						(eaRsp->num - 1)*sizeof(fsysUsage_t); 
	}

    ea_sendMsg(eaRsp, rspLen, sender->name);
    free(eaRsp);
}

/*------------------------------------------------------------------------------
 * Name         : ea_processMovOutFsysReq()
 *
 * Description  : Process the "move out file system" request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 07-DEC-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processMovOutFsysReq(ea_movOutFsysReq_t * evReq,
                        ea_senderInfo_t *    sender)

{
    ea_movOutFsysRsp_t *    eaRsp;
    int32                   rc = NFX_OK;
    ea_fsysCb_t             cb;
	uint32 				myFsState = 0;
	uint64				myUsedSize = 0;
	uint32				myHighWaterMark = 0;
    
    EA_FCT_PRINT(("> ea_processMovOutFsysReq()\n"));

    /* Initialize the response.
     */
    eaRsp = (ea_movOutFsysRsp_t *) malloc(sizeof(ea_movOutFsysRsp_t));
    NFX_ASSERT(eaRsp != NULL);
    bzero((void *) eaRsp, sizeof(ea_movOutFsysRsp_t));
    eaRsp->hdr.msgType = EA_MOVOUT_FSYS_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(&eaRsp->fsysId, &evReq->fsysId, sizeof(ea_fsysId_t));
    
    /* Initialize the control block.
     */
    bzero((void *) &cb, sizeof(ea_fsysCb_t));
    cb.hdr = (ea_msgHdr_t *) evReq;
    cb.vol = (evm_VolConfig_t *) malloc(sizeof(evm_VolConfig_t));
    NFX_ASSERT(cb.vol != NULL);
    bzero((void *) cb.vol, sizeof(evm_VolConfig_t));
    strcpy(cb.vol->name, evReq->fsysId.fsysName);
    cb.vol->onNIM = 1;

	// Get volume id from volume name, and node name that
	// has this volume online...
	rc = ea_getClusterVolIdNode(cb.vol, cb.hdr->node, eaRsp->output);
	// rc = ea_getVolId(cb.vol, eaRsp->output);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_movOutFsysRsp_t), sender->name);
		free(cb.vol);
        free(eaRsp);
        return;
    }

    /* Get volume config.
     */
    rc = ea_getVolCfg(cb.hdr->node, cb.vol);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        sprintf(eaRsp->output, "ea_getVolCfg() failed - %d", rc);
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_movOutFsysRsp_t), sender->name);
		free(cb.vol);
        free(eaRsp);
        return;
    }

	/*
	 * Make sure this is not exported
	 */
	if(ea_fsysIsExported(cb.hdr->node, cb.vol->id) == 1)
	{
        eaRsp->result = NFX_ERR;
        sprintf(eaRsp->output, "Operation not allowed on exported volume");
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_movOutFsysRsp_t), sender->name);
		free(cb.vol);
        free(eaRsp);
        return;
	}

	/*
	 * Make sure FS is not in the middle of something
	 */
	ea_getFsysInfo(cb.hdr->node, cb.vol->id,
				&myFsState, &myUsedSize, &myHighWaterMark);

	if ((myFsState == FS_VOL_STATE_MOUNTING) || 
		(myFsState == FS_VOL_STATE_UNMOUNTING) || 
		(myFsState == FS_VOL_STATE_FAILING) ||
		(myFsState == FS_VOL_STATE_PAUSED) ||
		(myFsState == FS_VOL_STATE_BACKUP_RESTORE))
	{
        eaRsp->result = NFX_ERR;
        sprintf(eaRsp->output, "FS state %s. Operation not allowed",
								fsStateStr[myFsState]);
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_movOutFsysRsp_t), sender->name);
		free(cb.vol);
        free(eaRsp);
        return;
	}

    /* 
	 * Get a lock on this volume
     */
	rc = ea_childGetLock(evReq->fsysId.fsysName);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        sprintf(eaRsp->output, "Volume %s is busy", evReq->fsysId.fsysName);
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_movOutFsysRsp_t), sender->name);
		free(cb.vol);
        free(eaRsp);
        return;
    }

    /* Unmount the file system.
     */
    rc = ea_umntFsys(cb.hdr->node, &cb);
    if ((rc != NFX_OK) && (rc != FS_ERR_NOT_MOUNTED))
    {
        eaRsp->result = rc;
        sprintf(eaRsp->output, "ea_umntFsys() failed - %d", rc);
        EA_ERR_PRINT(("%s\n", eaRsp->output));
		ea_childUnLock();
        ea_sendMsg(eaRsp, sizeof(ea_movOutFsysRsp_t), sender->name);
    	free(cb.vol);
        free(eaRsp);
        return;
    }

    /* Put off-line the volume.
     */
    rc = ea_offVol(cb.hdr->node, &cb, eaRsp->output);
    if ((rc != NFX_OK) && (rc != EVM_VOL_ALREADY_OFFLINE))
    {
        EA_ERR_PRINT(("%s\n", eaRsp->output));
    }
   
    eaRsp->result = NFX_OK;
    sprintf(eaRsp->output, "Done.");

	/*
	 * We are all done. Unlock the resources 
	 * before we send back a response
	 */
	ea_childUnLock();
    ea_sendMsg(eaRsp, sizeof(ea_movOutFsysRsp_t), sender->name);
    free(eaRsp);

    /* Free the volume configuration.
     */
    free(cb.vol);
    
    EA_FCT_PRINT(("< ea_processMovOutFsysReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processMovInFsysReq()
 *
 * Description  : Process the "move in file system" request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 07-DEC-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processMovInFsysReq(ea_movInFsysReq_t * evReq,
                       ea_senderInfo_t *   sender)

{
    ea_movInFsysRsp_t * eaRsp;
    int32              	rc = NFX_OK;
    ea_fsysCb_t         cb;
    
    EA_FCT_PRINT(("> ea_processMovInFsysReq()\n"));

    /* Initialize the response.
     */
    eaRsp = (ea_movInFsysRsp_t *) malloc(sizeof(ea_movInFsysRsp_t));
    NFX_ASSERT(eaRsp != NULL);
    bzero((void *) eaRsp, sizeof(ea_movInFsysRsp_t));

	if (evReq->forceMoveIn)
    	eaRsp->hdr.msgType = EA_FORCE_MOVIN_FSYS_RSP;
	else
    	eaRsp->hdr.msgType = EA_MOVIN_FSYS_RSP;

    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(&eaRsp->fsysCfg.fsysId, &evReq->fsysId, sizeof(ea_fsysId_t));
    
    /* Initialize the control block.
     */
    bzero((void *) &cb, sizeof(ea_fsysCb_t));
    cb.hdr = (ea_msgHdr_t *) evReq;
    cb.vol = (evm_VolConfig_t *) malloc(sizeof(evm_VolConfig_t));
    NFX_ASSERT(cb.vol != NULL);
    bzero((void *) cb.vol, sizeof(evm_VolConfig_t));
    strcpy(cb.vol->name, evReq->fsysId.fsysName);

	// Get volume id from volume name, and node name that
	// has this volume online...
	rc = ea_getClusterVolIdNode(cb.vol, cb.hdr->node, eaRsp->output);
	// rc = ea_getVolId(cb.vol, eaRsp->output);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_movInFsysRsp_t), sender->name);
		free(cb.vol);
        free(eaRsp);
        return;
    }

    /* Get volume config.
     */
    rc = ea_getVolCfg(cb.hdr->node, cb.vol);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        sprintf(eaRsp->output, "ea_getVolCfg() failed - %d", rc);
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_movInFsysRsp_t), sender->name);
		free(cb.vol);
        free(eaRsp);
        return;
    }

    /* 
	 * Get a lock on this volume
     */
	rc = ea_childGetLock(evReq->fsysId.fsysName);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        sprintf(eaRsp->output, "Volume %s is busy", evReq->fsysId.fsysName);
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_movInFsysRsp_t), sender->name);
		free(cb.vol);
        free(eaRsp);
        return;
    }

    /* Online volume on FP-NIM
     * Release 1: FP-NIM hardcoded to 1 - Change later!!!
     */
    cb.vol->onNIM = 1;

    if (evReq->forceMoveIn)
	{
    	/* 
		 * Takeover the volume.
     	 */
#ifndef ISABELLE
        EA_ERR_PRINT(("EA is doing takeoverVol\n"));
#endif // ISABELLE
    	rc = ea_takeoverVol(cb.hdr->node, &cb, eaRsp->output, evReq->forceTime);
    	if (rc != NFX_OK)
    	{
        	eaRsp->result = rc;
        	EA_ERR_PRINT(("%s\n", eaRsp->output));
			ea_childUnLock();
        	ea_sendMsg(eaRsp, sizeof(ea_movInFsysRsp_t), sender->name);
			free(cb.vol);
        	free(eaRsp);
        	return;
		}
	}
	else
	{
    	/* 
		 * Put on-line the volume.
     	 */
    	rc = ea_onVol(cb.hdr->node, &cb, eaRsp->output);
    	if ((rc != NFX_OK) && (rc != EVM_VOL_ALREADY_ONLINE))
    	{
        	eaRsp->result = rc;
        	EA_ERR_PRINT(("%s\n", eaRsp->output));
			ea_childUnLock();
        	ea_sendMsg(eaRsp, sizeof(ea_movInFsysRsp_t), sender->name);
			free(cb.vol);
        	free(eaRsp);
        	return;
    	}
	}

    /* Mount the file system.
     */
    rc = ea_mntFsys(cb.hdr->node, &cb);
    if ((rc != FS_IN_PROGRESS) &&
        (rc != NFX_OK) && (rc != FS_ERR_ALREADY_MOUNTED))
    {
		ea_offVol(cb.hdr->node, &cb, eaRsp->output);

        eaRsp->result = rc;
        sprintf(eaRsp->output, "ea_mntFsys() failed - %d", rc);
        EA_ERR_PRINT(("%s\n", eaRsp->output));
		ea_childUnLock();
        ea_sendMsg(eaRsp, sizeof(ea_movInFsysRsp_t), sender->name);
		free(cb.vol);
        free(eaRsp);
        return;
    }

    /* Everything went fine.
	 * Update response fields and send response 
	 * with configuration.
     */
    eaRsp->result = NFX_OK;

    if (rc == FS_IN_PROGRESS)
    	strcpy(eaRsp->output, "Mounting filesystem in progress.");
	else if (rc == FS_ERR_ALREADY_MOUNTED)
    	strcpy(eaRsp->output, "Volume already online.");
    else 
    	strcpy(eaRsp->output, "Done.");

	eaRsp->fsysCfg.onlineStatus = 1;
	eaRsp->fsysCfg.maxSize = cb.vol->maxSize;
	eaRsp->fsysCfg.currSize = cb.vol->currSize;
	eaRsp->fsysCfg.minAutoGrow = cb.vol->minAutoGrow;
	eaRsp->fsysCfg.softQuota = cb.vol->softQuota;
	eaRsp->fsysCfg.raidLevel = cb.vol->attrs.raidLevel;
	strcpy(eaRsp->fsysCfg.vendorInfo[0].vendor, 
						cb.vol->attrs.vendorInfo[0].vendor);
	strcpy(eaRsp->fsysCfg.vendorInfo[0].model, 
						cb.vol->attrs.vendorInfo[0].model);
	eaRsp->fsysCfg.mountStatus = 1;
	ea_getFsysInfo(cb.hdr->node, cb.vol->id,
							&eaRsp->fsysCfg.fsState,
	                        &eaRsp->fsysCfg.usedSize,
							&eaRsp->fsysCfg.highWaterMark);

	/*
	 * We are all done. Unlock the resources 
	 * before we send back a response
	 */
	ea_childUnLock();
    ea_sendMsg(eaRsp, sizeof(ea_movInFsysRsp_t), sender->name);
    free(eaRsp);

    /* Free the volume configuration.
     */
    free(cb.vol);
    
    EA_FCT_PRINT(("< ea_processMovInFsysReq()\n"));
    return;
}


/*------------------------------------------------------------------------------
 * Name         : ea_processNumFsysReq()
 *
 * Description  : Process the "num file system" request.
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created :
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processNumFsysReq(ea_numFsysReq_t * evReq,
						ea_senderInfo_t *   sender)
{
	char				*node = NULL;
    ea_fsysList_t *     fsysList;
    ea_numFsysRsp_t *   eaRsp;
    
    EA_FCT_PRINT(("> ea_processNumFsysReq()\n"));

    /* Initialize the response.
     */
    eaRsp = (ea_numFsysRsp_t *) malloc(sizeof(ea_numFsysRsp_t));
    NFX_ASSERT(eaRsp != NULL);
    bzero((void *) eaRsp, sizeof(ea_numFsysRsp_t));
    eaRsp->hdr.msgType = EA_NUM_FSYS_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;

	if (!isLocalNodeName(evReq->hdr.node))
		node = evReq->hdr.node;

    // Get list of volumes (id + name).
    fsysList = ea_getVolList(node);
    if (!fsysList)
    {
        eaRsp->result = NFX_ERR;
        sprintf(eaRsp->output, "ea_getVolList() failed");
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_numFsysRsp_t), sender->name);
        free(eaRsp);
        return;
    }
  
	eaRsp->totNum = fsysList->nb;
    free(fsysList);

    eaRsp->result = NFX_OK;
    sprintf(eaRsp->output, "Done.");
    ea_sendMsg(eaRsp, sizeof(ea_numFsysRsp_t), sender->name);
    free(eaRsp);

    EA_FCT_PRINT(("< ea_processNumFsysReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processGrowFsysReq()
 *
 * Description  : Pass on grow file system request
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created :
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processGrowFsysReq(ea_growFsysReq_t *evReq, 
						ea_senderInfo_t *	sender)
{
#ifdef EA_SIMULATE_SRM
	ea_growFsysInd_t	evInd;

	printf("ea_processGrowFsysReq: Simulate SRM Send ea_growFsysInd_t\n");

	evInd.hdr.msgType = EA_GROW_FSYS_IND;
	evInd.result = NFX_OK;
	strcpy(evInd.output, "EA sending dummy response");
	strcpy(evInd.node, evReq->node);
	evInd.txnId = evReq->txnId;
	evInd.num = 0;
	strcpy(evInd.devName[0], "IBM_5EG7Q222_0");
	strcpy(evInd.devName[0], "IBM_5EG7Q244_0");

    ea_sendMsg(&evInd, sizeof(ea_growFsysInd_t), "spm");
#else /* EA_SIMULATE_SRM */

	/*
	 * Forward the message via EV link to AV
	 */

	// pad1 is used to indicate whether this is a
	// request sent from SPM or EL. NFX_OK in pad1 would
	// indicate this is from SPM. Otherwise, it should be
	// from EL as an indication of problem sending this
	// request to AV. Upon receiving returned msg from EL,
	// EA should send a response to SPM immediately.
	if (evReq->hdr.pad1 == NFX_OK)
    	ea_sendMsg(evReq, sizeof(ea_growFsysReq_t), "ev");
	else
	{
		ea_growFsysRsp_t evRsp;

    	evRsp.hdr.msgType = EA_GROW_FSYS_RSP;
		evRsp.hdr.msgLen = sizeof(evRsp);
		evRsp.hdr.pad1 = evReq->hdr.pad1;
    	evRsp.hdr.userInfo = evReq->hdr.userInfo;
    	memcpy(evRsp.hdr.node, evReq->hdr.node, EVM_MAX_NODE_NAME);
    	evRsp.result = NFX_ERR;
		evRsp.pad1 = 0;
    	strcpy(evRsp.output, "Request not sent to AV");
    	memcpy(evRsp.node, evReq->node, EVM_MAX_NODE_NAME);
		evRsp.txnId = evReq->txnId;

    	ea_sendMsg(&evRsp, sizeof(ea_growFsysRsp_t), "spm");
	}

#endif /* EA_SIMULATE_SRM */
}

/*------------------------------------------------------------------------------
 * Name         : ea_processGrowFsysRsp()
 *
 * Description  : Pass on grow file system response
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created :
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processGrowFsysRsp(ea_growFsysRsp_t *evReq, 
						ea_senderInfo_t *	sender)
{
	/*
	 * Forward the message to SPM
	 */
    ea_sendMsg(evReq, sizeof(ea_growFsysRsp_t), "spm");
}

/*------------------------------------------------------------------------------
 * Name         : ea_processGrowFsysInd()
 *
 * Description  : Pass on grow file system indication
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created :
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processGrowFsysInd(ea_growFsysInd_t *evReq, 
						ea_senderInfo_t *	sender,
                      int32  len)
{
	/*
	 * Forward the message to SPM
	 */
    ea_sendMsg(evReq, len, "spm");
}

/*------------------------------------------------------------------------------
 * Action on a file set.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : ea_processAddFsetReq()
 *
 * Description  : Process the "add file set" request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processAddFsetReq(ea_addFsetReq_t * evReq,
                     ea_senderInfo_t * sender)

{
    int32          	rc = NFX_OK;
    ea_fsetCb_t     cb;
    ea_addFsetRsp_t eaRsp;
	ea_quotaCfg_t 	quotaBlk;
    
    EA_FCT_PRINT(("> ea_processAddFsetReq()\n"));

    /* Initialize the response.
     */
    bzero((void *) &eaRsp, sizeof(ea_addFsetRsp_t));
    eaRsp.hdr.msgType = EA_ADD_FSET_RSP;
    eaRsp.hdr.userInfo = evReq->hdr.userInfo;
	memcpy(&eaRsp.fsetCfg.fsetId, &evReq->fsetId, sizeof(ea_fsetId_t));
    
    /* Initialize the control block.
     */
    bzero((void *) &cb, sizeof(ea_fsetCb_t));
    cb.hdr = (ea_msgHdr_t *) evReq;
    strcpy(cb.path, evReq->fsetId.path);
    cb.quotaOn = evReq->quotaOn;
    cb.vol = (evm_VolConfig_t *) malloc(sizeof(evm_VolConfig_t));
    NFX_ASSERT(cb.vol != NULL);
    bzero((void *) cb.vol, sizeof(evm_VolConfig_t));
    strcpy(cb.vol->name, evReq->fsetId.fsysId.fsysName);
    cb.vol->onNIM = 1;

	// Get volume id from volume name, and node name that
	// has this volume online...
	rc = ea_getClusterVolIdNode(cb.vol, cb.hdr->node, eaRsp.output);
	// rc = ea_getVolId(cb.vol, eaRsp->output);
    if (rc != NFX_OK)
    {
        eaRsp.result = rc;
        EA_ERR_PRINT(("%s\n", eaRsp.output));
        ea_sendMsg(&eaRsp, sizeof(ea_addFsetRsp_t), sender->name);
		free(cb.vol);
        return;
    }

    /* Get volume config.
     */
    rc = ea_getVolCfg(cb.hdr->node, cb.vol);
    if (rc != NFX_OK)
    {
        eaRsp.result = rc;
        sprintf(eaRsp.output, "ea_getVolCfg() failed - %d", rc);
        EA_ERR_PRINT(("%s\n", eaRsp.output));
        ea_sendMsg(&eaRsp, sizeof(ea_addFsetRsp_t), sender->name);
		free(cb.vol);
        return;
    }

    /* Create a file set.
     */
    rc = ea_crtFset(cb.hdr->node, &cb);
    if (rc != NFX_OK)
    {
        eaRsp.result = rc;
        sprintf(eaRsp.output, "ea_crtFset() failed - %d", rc);
        EA_ERR_PRINT(("%s\n", eaRsp.output));
        ea_sendMsg(&eaRsp, sizeof(ea_addFsetRsp_t), sender->name);
		free(cb.vol);
        return;
    }

    /* Set quota on the file set.
     * Note: maxIops and cacheThreshold not yet implemented in FS.
     */
    bzero((void *) &quotaBlk, sizeof(ea_quotaCfg_t));
	quotaBlk.inodeSoftLimit = evReq->quotaCfg.inodeSoftLimit;
	quotaBlk.inodeHardLimit = evReq->quotaCfg.inodeHardLimit;
	quotaBlk.inodeTimeLimit = evReq->quotaCfg.inodeTimeLimit;
	quotaBlk.blockSoftLimit = evReq->quotaCfg.blockSoftLimit;
	quotaBlk.blockHardLimit = evReq->quotaCfg.blockHardLimit;
	quotaBlk.blockTimeLimit = evReq->quotaCfg.blockTimeLimit;

    rc = ea_setFsetQuota(cb.hdr->node, &cb, &quotaBlk);
    if (rc != NFX_OK)
    {
        eaRsp.result = rc;
        sprintf(eaRsp.output, "ea_setFsetQuota() failed - %d", rc);
        EA_ERR_PRINT(("%s\n", eaRsp.output));
        ea_sendMsg(&eaRsp, sizeof(ea_addFsetRsp_t), sender->name);
		free(cb.vol);
        return;
    }

    /* Turn on/off quota on the file set.
     */
    if (cb.quotaOn == TRUE)
    {
        rc = ea_onFsetQuota(cb.hdr->node, &cb);
    }
    else
    {
        rc = ea_offFsetQuota(cb.hdr->node, &cb);
    }
    if (rc != NFX_OK)
    {
        eaRsp.result = rc;
        sprintf(eaRsp.output, "ea_on/offFsetQuota() failed - %d", rc);
        EA_ERR_PRINT(("%s\n", eaRsp.output));
        ea_sendMsg(&eaRsp, sizeof(ea_addFsetRsp_t), sender->name);
		free(cb.vol);
        return;
    }

	/*
	 * Copy back the configuration in the response
	 */
    eaRsp.result = NFX_OK;
    sprintf(eaRsp.output, "Done.");
	memcpy(&eaRsp.fsetCfg.fsetId.fsysId.volId, &cb.vol->id, 
			sizeof(evm_volId_t));
	strcpy(eaRsp.fsetCfg.fsetId.path, cb.path);
	eaRsp.fsetCfg.quotaOn = cb.quotaOn;
	memcpy(&eaRsp.fsetCfg.quotaCfg, &quotaBlk, sizeof(ea_quotaCfg_t));

    ea_sendMsg(&eaRsp, sizeof(ea_addFsetRsp_t), sender->name);

    /* Free the volume configuration.
     */
    free(cb.vol);

    EA_FCT_PRINT(("< ea_processAddFsetReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processModFsetReq()
 *
 * Description  : Process the "modify file set" request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processModFsetReq(ea_modFsetReq_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32          	rc = NFX_OK;
    ea_fsetCb_t     cb;
    ea_modFsetRsp_t eaRsp;
	ea_quotaCfg_t 	quotaBlk;
    
    EA_FCT_PRINT(("> ea_processModFsetReq()\n"));

    /* Initialize the response.
     */
    bzero((void *) &eaRsp, sizeof(ea_modFsetRsp_t));
    eaRsp.hdr.msgType = EA_MOD_FSET_RSP;
    eaRsp.hdr.userInfo = evReq->hdr.userInfo;
	memcpy(&eaRsp.fsetCfg.fsetId, &evReq->fsetId, sizeof(ea_fsetId_t));
    
    /* Initialize the control block.
     */
    bzero((void *) &cb, sizeof(ea_fsetCb_t));
    cb.hdr = (ea_msgHdr_t *) evReq;
    strcpy(cb.path, evReq->fsetId.path);
    cb.quotaOn = evReq->quotaOn;
    cb.vol = (evm_VolConfig_t *) malloc(sizeof(evm_VolConfig_t));
    NFX_ASSERT(cb.vol != NULL);
    bzero((void *) cb.vol, sizeof(evm_VolConfig_t));
    strcpy(cb.vol->name, evReq->fsetId.fsysId.fsysName);
    cb.vol->onNIM = 1;

	// Get volume id from volume name, and node name that
	// has this volume online...
	rc = ea_getClusterVolIdNode(cb.vol, cb.hdr->node, eaRsp.output);
	// rc = ea_getVolId(cb.vol, eaRsp->output);
    if (rc != NFX_OK)
    {
        eaRsp.result = rc;
        EA_ERR_PRINT(("%s\n", eaRsp.output));
        ea_sendMsg(&eaRsp, sizeof(ea_modFsetRsp_t), sender->name);
		free(cb.vol);
        return;
    }

    /* Get volume config.
     */
    rc = ea_getVolCfg(cb.hdr->node, cb.vol);
    if (rc != NFX_OK)
    {
        eaRsp.result = rc;
        sprintf(eaRsp.output, "ea_getVolCfg() failed - %d", rc);
        EA_ERR_PRINT(("%s\n", eaRsp.output));
        ea_sendMsg(&eaRsp, sizeof(ea_modFsetRsp_t), sender->name);
		free(cb.vol);
        return;
    }

    /* Set quota on the file set.
     * Note: maxIops and cacheThreshold not yet implemented in FS.
     */
    bzero((void *) &quotaBlk, sizeof(ea_quotaCfg_t));
	quotaBlk.inodeSoftLimit = evReq->quotaCfg.inodeSoftLimit;
	quotaBlk.inodeHardLimit = evReq->quotaCfg.inodeHardLimit;
	quotaBlk.inodeTimeLimit = evReq->quotaCfg.inodeTimeLimit;
	quotaBlk.blockSoftLimit = evReq->quotaCfg.blockSoftLimit;
	quotaBlk.blockHardLimit = evReq->quotaCfg.blockHardLimit;
	quotaBlk.blockTimeLimit = evReq->quotaCfg.blockTimeLimit;

    rc = ea_setFsetQuota(cb.hdr->node, &cb, &quotaBlk);
    if (rc != NFX_OK)
    {
        eaRsp.result = rc;
        sprintf(eaRsp.output, "ea_setFsetQuota() failed - %d", rc);
        EA_ERR_PRINT(("%s\n", eaRsp.output));
        ea_sendMsg(&eaRsp, sizeof(ea_modFsetRsp_t), sender->name);
		free(cb.vol);
        return;
    }

    /* Turn on/off quota on the file set.
     */
    if (cb.quotaOn == TRUE)
    {
        rc = ea_onFsetQuota(cb.hdr->node, &cb);
    }
    else
    {
        rc = ea_offFsetQuota(cb.hdr->node, &cb);
    }
    if (rc != NFX_OK)
    {
        eaRsp.result = rc;
        sprintf(eaRsp.output, "ea_on/offFsetQuota() failed - %d", rc);
        EA_ERR_PRINT(("%s\n", eaRsp.output));
        ea_sendMsg(&eaRsp, sizeof(ea_modFsetRsp_t), sender->name);
		free(cb.vol);
        return;
    }

	/*
	 * Copy back the configuration in the response
	 */
    eaRsp.result = NFX_OK;
    sprintf(eaRsp.output, "Done.");
	memcpy(&eaRsp.fsetCfg.fsetId.fsysId.volId, &cb.vol->id, 
			sizeof(evm_volId_t));
	strcpy(eaRsp.fsetCfg.fsetId.path, cb.path);
	eaRsp.fsetCfg.quotaOn = cb.quotaOn;
	memcpy(&eaRsp.fsetCfg.quotaCfg, &quotaBlk, sizeof(ea_quotaCfg_t));

    ea_sendMsg(&eaRsp, sizeof(ea_modFsetRsp_t), sender->name);

    /* Free the volume configuration.
     */
    free(cb.vol);

    EA_FCT_PRINT(("< ea_processModFsetReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processDelFsetReq()
 *
 * Description  : Process the "delete file set" request.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processDelFsetReq(ea_delFsetReq_t * evReq,
                     ea_senderInfo_t * sender)

{
    int32           rc = NFX_OK;
    ea_fsetCb_t     cb;
    ea_delFsetRsp_t eaRsp;
    
    EA_FCT_PRINT(("> ea_processDelFsetReq()\n"));

    /* Initialize the response.
     */
    bzero((void *) &eaRsp, sizeof(ea_delFsetRsp_t));
    eaRsp.hdr.msgType = EA_DEL_FSET_RSP;
    eaRsp.hdr.userInfo = evReq->hdr.userInfo;
	memcpy(&eaRsp.fsetId, &evReq->fsetId, sizeof(ea_fsetId_t));
    
    /* Initialize the control block.
     */
    bzero((void *) &cb, sizeof(ea_fsetCb_t));         
    cb.hdr = (ea_msgHdr_t *) evReq;    
    strcpy(cb.path, evReq->fsetId.path);
    cb.vol = (evm_VolConfig_t *) malloc(sizeof(evm_VolConfig_t));
    NFX_ASSERT(cb.vol != NULL);
    bzero((void *) cb.vol, sizeof(evm_VolConfig_t));
    strcpy(cb.vol->name, evReq->fsetId.fsysId.fsysName);
    cb.vol->onNIM = 1;

	// Get volume id from volume name, and node name that
	// has this volume online...
	rc = ea_getClusterVolIdNode(cb.vol, cb.hdr->node, eaRsp.output);
	// rc = ea_getVolId(cb.vol, eaRsp->output);
    if (rc != NFX_OK)
    {
        eaRsp.result = rc;
        EA_ERR_PRINT(("%s\n", eaRsp.output));
        ea_sendMsg(&eaRsp, sizeof(ea_delFsetRsp_t), sender->name);
		free(cb.vol);
        return;
    }

    /* Get volume config.
     */
    rc = ea_getVolCfg(cb.hdr->node, cb.vol);
    if (rc != NFX_OK)
    {
        eaRsp.result = rc;
        sprintf(eaRsp.output, "ea_getVolCfg() failed - %d", rc);
        EA_ERR_PRINT(("%s\n", eaRsp.output));
        ea_sendMsg(&eaRsp, sizeof(ea_delFsetRsp_t), sender->name);
		free(cb.vol);
        return;
    }

    /* Delete a file set.
     */
    rc = ea_delFset(cb.hdr->node, &cb);
    if (rc != NFX_OK)
    {
        eaRsp.result = rc;
        sprintf(eaRsp.output, "ea_delFset() failed - %d", rc);
        EA_ERR_PRINT(("%s\n", eaRsp.output));
        ea_sendMsg(&eaRsp, sizeof(ea_delFsetRsp_t), sender->name);
		free(cb.vol);
        return;
    }

    eaRsp.result = NFX_OK;
    sprintf(eaRsp.output, "Done.");
	memcpy(&eaRsp.fsetId.fsysId.volId, &cb.vol->id, 
			sizeof(evm_volId_t));
	
    ea_sendMsg(&eaRsp, sizeof(ea_delFsetRsp_t), sender->name);

    /* Free the volume configuration.
     */
    free(cb.vol);

    EA_FCT_PRINT(("< ea_processDelFsetReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processNumFsetReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processNumFsetReq(ea_numFsetReq_t * evReq,
					 ea_senderInfo_t * sender)
{
	char				*node = NULL;
    ea_fsetList_t *     fsetList = NULL;
    ea_numFsetRsp_t *   eaRsp;
    
    EA_FCT_PRINT(("> ea_processNumFsetReq()\n"));

    /* Initialize the response.
     */
    eaRsp = (ea_numFsetRsp_t *) malloc(sizeof(ea_numFsetRsp_t));
    NFX_ASSERT(eaRsp != NULL);
    bzero((void *) eaRsp, sizeof(ea_numFsetRsp_t));
    eaRsp->hdr.msgType = EA_NUM_FSET_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;

	if (!isLocalNodeName(evReq->hdr.node))
		node = evReq->hdr.node;

    // Get list of file sets and ids
    fsetList = ea_getFsetIdList(node);
    if (!fsetList)
    {
        eaRsp->result = NFX_ERR;
        sprintf(eaRsp->output, "ea_getFsetIdList() failed");
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_numFsetRsp_t), sender->name);
        free(eaRsp);
        return;
    }
  
	eaRsp->totNum = fsetList->nb;
    free(fsetList);

    eaRsp->result = NFX_OK;
    sprintf(eaRsp->output, "Done.");
    ea_sendMsg(eaRsp, sizeof(ea_numFsetRsp_t), sender->name);
    free(eaRsp);

    EA_FCT_PRINT(("< ea_processNumFsetReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processGetFsetReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 * IMPORTANT NOTES:
 * 1. If ea_fsetId_t are specified, only fsysId on the same node can be
 * specified at the same time. EA will only talk to the node, on which
 * the fsetIds[0].fsysId.volId (first one) resides, to retrieve the
 * fset info
 * 2. If fsetIds[0].fsysId.volId is specified (non-zero), then this ID
 * will be used to locate the node that has this volume mounted. If this
 * volId is 0, then fsetIds[0].fsysId.fsysName is used to get the volume
 * ID and node name for this volume. If fsysName is null string, and volId
 * is 0, then hdr.node is checked to see if a node name is specified there.
 * If not, local node is assumed...
 *------------------------------------------------------------------------------
 */
static void
ea_processGetFsetReq(ea_getFsetReq_t * evReq,
                     ea_senderInfo_t * sender)
{
	int 					totNum = 0;
	int 					i = 0;
	int 					startIdx = 0;
    ea_fsetList_t	 		*fsetList = NULL;
    ea_getFsetRsp_t			*eaRsp;
    uint32					rspLen = sizeof(ea_getFsetRsp_t);
	uint32 					rspIdx = 0;

	fs_MsgReportQuotaResp_t *fsQuotaRsp;
	fs_QuotaBlk_t			*fsQuotaBlk;
	char					*node;

    
    EA_FCT_PRINT(("> ea_processGetFsetReq()\n"));

    /* Initialize the response.
     */
    eaRsp = (ea_getFsetRsp_t *) malloc(rspLen);
    NFX_ASSERT(eaRsp != NULL);
    bzero((void *) eaRsp, rspLen);
    eaRsp->hdr.msgType = EA_GET_FSET_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;

    // Need to get the node name for this request first...
	node = evReq->hdr.node;
	if (*evReq->hdr.node == '\0')
	{
		if (evReq->fsetIds[0].fsysId.volId.whole_thing != 0)
		{
			fs_opt_get_clus_lvol_online_node(
						evReq->fsetIds[0].fsysId.volId, evReq->hdr.node);
			if (!strcmp(evReq->hdr.node, "local"))
				*(evReq->hdr.node) = '\0';
		}
		else if (*evReq->fsetIds[0].fsysId.fsysName)
		{
			ea_getNodeVolIdFromVolName(evReq->fsetIds[0].fsysId.fsysName,
												evReq->hdr.node, NULL);
		}
		else
			node = NULL;
	}

    // Get list of file sets and ids
    fsetList = ea_getFsetIdList(node);
    if (!fsetList)
    {
        eaRsp->result = NFX_ERR;
        sprintf(eaRsp->output, "ea_getFsetIdList() failed");
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, rspLen, sender->name);
        free(eaRsp);
        return;
    }

	totNum = fsetList->nb;
  	if ((evReq->reqType == EA_REQ_HAS_FSET_INIT) &&
		(evReq->fsetIds[0].fsysId.volId.whole_thing != 0))
	{
		for (i = 0; i < fsetList->nb; i++)
		{
			totNum--;
			if ((fsetList->fset[i].fsetId.fsysId.volId.whole_thing ==
				evReq->fsetIds[0].fsysId.volId.whole_thing) &&
				(strcmp(fsetList->fset[i].fsetId.path, 
							evReq->fsetIds[0].path) == 0))
			{
				startIdx = ++i;
				break;
			}
		}
	}

   	if (evReq->num)
	{
		if (totNum > evReq->num)
			fsetList->nb = evReq->num;
		else
			fsetList->nb = totNum;
	}

    if (fsetList->nb > 1)
    {
        rspLen += (fsetList->nb - 1) * sizeof(ea_fsetCfg_t);
        eaRsp = (ea_getFsetRsp_t *) realloc((void *) eaRsp, rspLen);
    	bzero((void *) eaRsp, rspLen);
    	eaRsp->hdr.msgType = EA_GET_FSET_RSP;
    	eaRsp->hdr.userInfo = evReq->hdr.userInfo;
    }

	if (evReq->reqType == EA_REQ_HAS_FSET_INIT)
	{
		eaRsp->fsetList.nb = fsetList->nb;
		for (i = startIdx; i < (startIdx + fsetList->nb); i++)
		{
			/*
			 * Get Quota and usage info
			 */
			bzero((void *)&eaRsp->fsetList.fset[i-startIdx], 
									sizeof(ea_fsetCfg_t));
			eaRsp->fsetList.fset[i-startIdx].fsetId =
					fsetList->fset[i].fsetId;

			fsQuotaRsp = ea_fsetGetQuota(node, &fsetList->fset[i].fsetId);
			if (fsQuotaRsp)
			{
				fsQuotaBlk = &fsQuotaRsp->responses[0];

				if (fsQuotaBlk->flag & FS_QUOTA_ON)
					eaRsp->fsetList.fset[i-startIdx].quotaOn = 1;
				else
					eaRsp->fsetList.fset[i-startIdx].quotaOn = 0;
	
				/*
			 	 * copy over quota config 
			 	 */
				eaRsp->fsetList.fset[i-startIdx].quotaCfg.inodeSoftLimit =
								fsQuotaBlk->inodeSoftLimit;
				eaRsp->fsetList.fset[i-startIdx].quotaCfg.inodeHardLimit =
								fsQuotaBlk->inodeHardLimit;
				eaRsp->fsetList.fset[i-startIdx].quotaCfg.inodeTimeLimit =
								fsQuotaBlk->inodeTimeLimit;
				eaRsp->fsetList.fset[i-startIdx].quotaCfg.blockSoftLimit =
								fsQuotaBlk->blockSoftLimit;
				eaRsp->fsetList.fset[i-startIdx].quotaCfg.blockHardLimit =
								fsQuotaBlk->blockHardLimit;
				eaRsp->fsetList.fset[i-startIdx].quotaCfg.blockTimeLimit =
								fsQuotaBlk->blockTimeLimit;
	
				/*
			 	 * copy over usage info
			 	 */
				eaRsp->fsetList.fset[i-startIdx].quotaUsage.inodesConsumed =
								fsQuotaBlk->inodesConsumed;
				eaRsp->fsetList.fset[i-startIdx].quotaUsage.blocksConsumed =
								fsQuotaBlk->blocksConsumed;
				eaRsp->fsetList.fset[i-startIdx].quotaUsage.inodeSoftLimitTime =
								fsQuotaBlk->inodeSoftLimitTime;
				eaRsp->fsetList.fset[i-startIdx].quotaUsage.blockSoftLimitTime =
								fsQuotaBlk->blockSoftLimitTime;
	
				free(fsQuotaRsp);
			}
		}
	}
	else if (evReq->reqType == EA_REQ_HAS_FSET_LIST)
	{
		for (i = 0; (i < totNum) && (rspIdx < fsetList->nb); i++)
		{
			if ((fsetList->fset[i].fsetId.fsysId.volId.whole_thing ==
				evReq->fsetIds[rspIdx].fsysId.volId.whole_thing) &&
				(strcmp(fsetList->fset[i].fsetId.path, 
							evReq->fsetIds[rspIdx].path) == 0))
			{
				/*
			 	 * Get Quota and usage info
			 	 */
				bzero((void *)&eaRsp->fsetList.fset[rspIdx], 
										sizeof(ea_fsetCfg_t));
				eaRsp->fsetList.fset[rspIdx].fsetId =
										fsetList->fset[i].fsetId;

				fsQuotaRsp = ea_fsetGetQuota(node, &fsetList->fset[i].fsetId);
				if (fsQuotaRsp)
				{
					fsQuotaBlk = &fsQuotaRsp->responses[0];

					if (fsQuotaBlk->flag & FS_QUOTA_ON)
						eaRsp->fsetList.fset[rspIdx].quotaOn = 1;
					else
						eaRsp->fsetList.fset[rspIdx].quotaOn = 0;
	
					/*
			 	 	 * copy over quota config 
			 	 	 */
					eaRsp->fsetList.fset[rspIdx].quotaCfg.inodeSoftLimit =
									fsQuotaBlk->inodeSoftLimit;
					eaRsp->fsetList.fset[rspIdx].quotaCfg.inodeHardLimit =
									fsQuotaBlk->inodeHardLimit;
					eaRsp->fsetList.fset[rspIdx].quotaCfg.inodeTimeLimit =
									fsQuotaBlk->inodeTimeLimit;
					eaRsp->fsetList.fset[rspIdx].quotaCfg.blockSoftLimit =
									fsQuotaBlk->blockSoftLimit;
					eaRsp->fsetList.fset[rspIdx].quotaCfg.blockHardLimit =
									fsQuotaBlk->blockHardLimit;
					eaRsp->fsetList.fset[rspIdx].quotaCfg.blockTimeLimit =
									fsQuotaBlk->blockTimeLimit;
	
					/*
			 	 	 * copy over usage info
			 	 	 */
					eaRsp->fsetList.fset[rspIdx].quotaUsage.inodesConsumed =
									fsQuotaBlk->inodesConsumed;
					eaRsp->fsetList.fset[rspIdx].quotaUsage.blocksConsumed =
									fsQuotaBlk->blocksConsumed;
					eaRsp->fsetList.fset[rspIdx].quotaUsage.inodeSoftLimitTime =
									fsQuotaBlk->inodeSoftLimitTime;
					eaRsp->fsetList.fset[rspIdx].quotaUsage.blockSoftLimitTime =
									fsQuotaBlk->blockSoftLimitTime;
	
					free(fsQuotaRsp);
				}
				rspIdx++;
			}
		}
		eaRsp->fsetList.nb = rspIdx;
	}

    free(fsetList);

#ifdef EA_DEBUG
	printf("Number of entries in GetFset Rsp %d\n", eaRsp->fsetList.nb);

	for (i = 0; i < eaRsp->fsetList.nb; i++)
	{
		ea_fsetCfg_t *fsetEntry = &eaRsp->fsetList.fset[i];

		printf("\n\nFSET vol %s id %lld\n",
			fsetEntry->fsetId.fsysId.fsysName,
			fsetEntry->fsetId.fsysId.volId.whole_thing);

		printf("path: %s\n", fsetEntry->fsetId.path);

		printf("Quota (on/off) %d\n", fsetEntry->quotaOn);
		printf("inodeSoftLimit %lld\n", fsetEntry->quotaCfg.inodeSoftLimit);
		printf("inodeHardLimit %lld\n", fsetEntry->quotaCfg.inodeHardLimit);
		printf("inodeTimeLimit %lld\n", fsetEntry->quotaCfg.inodeTimeLimit);
		printf("blockSoftLimit %lld\n", fsetEntry->quotaCfg.blockSoftLimit);
		printf("blockHardLimit %lld\n", fsetEntry->quotaCfg.blockHardLimit);
		printf("blockTimeLimit %lld\n", fsetEntry->quotaCfg.blockTimeLimit);

		printf("inodesConsumed %lld\n", fsetEntry->quotaUsage.inodesConsumed);
		printf("blocksConsumed %lld\n", fsetEntry->quotaUsage.blocksConsumed);
		printf("inodeSoftLimitTime %d\n", 
									fsetEntry->quotaUsage.inodeSoftLimitTime);
		printf("blockSoftLimitTime %d\n", 
									fsetEntry->quotaUsage.blockSoftLimitTime);
	}
#endif

    eaRsp->result = NFX_OK;
    sprintf(eaRsp->output, "Done.");

	if (eaRsp->fsetList.nb <= 0)
	{
		rspLen = sizeof(ea_getFsetRsp_t);
	}
	else
	{
		rspLen = sizeof(ea_getFsetRsp_t) + 
					(eaRsp->fsetList.nb - 1)*sizeof(ea_fsetCfg_t); 
	}

    ea_sendMsg(eaRsp, rspLen, sender->name);
    free(eaRsp);

    EA_FCT_PRINT(("< ea_processGetFsetReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processUsgFsetReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 * IMPORTANT NOTES:
 * 1. If ea_fsetId_t are specified, only fsysId on the same node can be
 * specified at the same time. EA will only talk to the node, on which
 * the fsetIds[0].fsysId.volId (first one) resides, to retrieve the
 * fset info
 * 2. hdr.node is checked first to see if a node name is specified there.
 * If not, fsetIds[0].fsysId.volId is checked. If it's non-zero, then this
 * ID will be used to locate the node that has this volume mounted. If
 * it's 0, then fsetIds[0].fsysId.fsysName is used to get the volume
 * ID and node name for this volume. If fsysName is also null string, then
 * local node is used.
 *------------------------------------------------------------------------------
 */
static void
ea_processUsgFsetReq(ea_usgFsetReq_t * evReq,
                     ea_senderInfo_t * sender)
{
	int 					totNum = 0;
	int 					i = 0;
    ea_fsetList_t	 		*fsetList = NULL;
    ea_usgFsetRsp_t			*eaRsp;
    int32					rc = NFX_OK;
	uint32	 				rspIdx;
    uint32					rspLen = sizeof(ea_usgFsetRsp_t);

	fs_MsgReportQuotaResp_t *fsQuotaRsp;
	fs_QuotaBlk_t			*fsQuotaBlk;
	char					*node;
    
    EA_FCT_PRINT(("> ea_processUsgFsetReq()\n"));

    /* Initialize the response.
     */
    eaRsp = (ea_usgFsetRsp_t *) malloc(rspLen);
    NFX_ASSERT(eaRsp != NULL);
    bzero((void *) eaRsp, rspLen);
    eaRsp->hdr.msgType = EA_USG_FSET_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;

    // Need to get the node name for this request first...
	node = evReq->hdr.node;
	if (*evReq->hdr.node == '\0')
	{
		if (evReq->fsetIds[0].fsysId.volId.whole_thing != 0)
		{
			fs_opt_get_clus_lvol_online_node(
						evReq->fsetIds[0].fsysId.volId, evReq->hdr.node);
			if (!strcmp(evReq->hdr.node, "local"))
				*(evReq->hdr.node) = '\0';
		}
		else if (*evReq->fsetIds[0].fsysId.fsysName)
		{
			ea_getNodeVolIdFromVolName(evReq->fsetIds[0].fsysId.fsysName,
												evReq->hdr.node, NULL);
		}
		else
			node = NULL;
	}

    // Get list of file sets and ids
    fsetList = ea_getFsetIdList(node);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        sprintf(eaRsp->output, "ea_getFsetIdList() failed - %d", rc);
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, rspLen, sender->name);
        free(eaRsp);
        return;
    }

	totNum = fsetList->nb;

   	if (evReq->num)
	{
		if (totNum > evReq->num)
			fsetList->nb = evReq->num;
		else
			fsetList->nb = totNum;
	}

    if (fsetList->nb > 1)
    {
        rspLen += (fsetList->nb - 1) * sizeof(fsetUsage_t);
        eaRsp = (ea_usgFsetRsp_t *) realloc((void *) eaRsp, rspLen);
    	bzero((void *) eaRsp, rspLen);
    	eaRsp->hdr.msgType = EA_USG_FSET_RSP;
    	eaRsp->hdr.userInfo = evReq->hdr.userInfo;
    }

	rspIdx = 0;
	for (i = 0; (i < totNum) && (rspIdx < fsetList->nb); i++)
	{
		if (((fsetList->fset[i].fsetId.fsysId.volId.whole_thing ==
			evReq->fsetIds[rspIdx].fsysId.volId.whole_thing) &&
			(strcmp(fsetList->fset[i].fsetId.path, 
						evReq->fsetIds[rspIdx].path) == 0)) ||
					(evReq->num == 0))
		{
			/*
		 	 * Get Quota and usage info
		 	 */
			bzero((void *)&eaRsp->usageList[rspIdx],
									sizeof(fsetUsage_t));

			eaRsp->usageList[rspIdx].fsetId =
							fsetList->fset[i].fsetId;
						
			fsQuotaRsp = ea_fsetGetQuota(node, &fsetList->fset[i].fsetId);
			if (fsQuotaRsp)
			{
				fsQuotaBlk = &fsQuotaRsp->responses[0];

				/*
		 	 	 * copy over usage info
		 	 	 */
				eaRsp->usageList[rspIdx].quotaUsage.inodesConsumed =
								fsQuotaBlk->inodesConsumed;
				eaRsp->usageList[rspIdx].quotaUsage.blocksConsumed =
								fsQuotaBlk->blocksConsumed;
				eaRsp->usageList[rspIdx].quotaUsage.inodeSoftLimitTime =
								fsQuotaBlk->inodeSoftLimitTime;
				eaRsp->usageList[rspIdx].quotaUsage.blockSoftLimitTime =
								fsQuotaBlk->blockSoftLimitTime;
				free(fsQuotaRsp);
			}

			rspIdx++;
		}
	}

    free(fsetList);

    eaRsp->result = NFX_OK;
    sprintf(eaRsp->output, "Done.");

	eaRsp->num = rspIdx;
	if (eaRsp->num <= 0)
	{
		rspLen = sizeof(ea_usgFsetRsp_t);
	}
	else
	{
		rspLen = sizeof(ea_usgFsetRsp_t) + 
						(eaRsp->num - 1)*sizeof(fsetUsage_t); 
	}

    ea_sendMsg(eaRsp, rspLen, sender->name);
    free(eaRsp);

    EA_FCT_PRINT(("< ea_processUsgFsetReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processAddExpReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processAddExpReq(ea_addExpReq_t * evReq,
					ea_senderInfo_t * sender)
{
	int					retStat;
    ea_clusmsg_t		reqBuf;
    ea_clusmsg_t		rspBuf;
	mnt_infoReq_t       *mntInfoReqp;
	mnt_infoResp_t      *mntInfoResp;
	ea_addExpRsp_t		*eaRsp;
	int					rspLen = 0;
	int					mntReqLen = 0;
	int					idx = 0;
	int					numLoop;

	int                 rc;
	struct timeval      timeout = { 10, 0};

    EA_FCT_PRINT(("< ea_processAddExpReq()\n"));

	rspLen = sizeof(ea_addExpRsp_t) + 
				(evReq->expCfg.num - 1)*sizeof(ea_expHost_t);
	eaRsp = (ea_addExpRsp_t *) malloc(rspLen);
	bzero((void *)eaRsp, rspLen);

    eaRsp->hdr.msgType = EA_ADD_EXP_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	eaRsp->fsysId = evReq->fsysId;
	strcpy(eaRsp->expCfg.path.pathName, evReq->expCfg.path.pathName);
	eaRsp->expCfg.shareIP = ntohl(evReq->expCfg.shareIP);
	eaRsp->expCfg.num = 0;

	if (evReq->expCfg.num == 0)
		numLoop = 1;
	else
		numLoop = evReq->expCfg.num;

	retStat = ea_getNodeVolIdFromVolName(evReq->fsysId.fsysName,
									evReq->hdr.node, NULL);
	if (retStat != NFX_OK)
		*(evReq->hdr.node) = '\0';

	for (idx = 0; idx < numLoop; idx++)
	{
		retStat = NFX_OK;

		memset(&reqBuf, 0, sizeof(ea_clusmsg_t));
		mntInfoReqp = (mnt_infoReq_t *) reqBuf.buf;

		/*
		 * Build client:/VolName/PathName
		 */
		if (evReq->expCfg.expHostList[idx].hostIP)
		{
			struct in_addr in;

			in.s_addr = evReq->expCfg.expHostList[idx].hostIP;
			strcpy(mntInfoReqp->pathName, (char *)inet_ntoa(in));
			strcat(mntInfoReqp->pathName, ":");
		}

		strcat(mntInfoReqp->pathName,"/");
		strcat(mntInfoReqp->pathName, evReq->fsysId.fsysName);
		strcat(mntInfoReqp->pathName, evReq->expCfg.path.pathName);

		mntInfoReqp->nameLen = strlen(mntInfoReqp->pathName);

		/*
		 * Build options rw,ro,root_squash,no_root_squash,laddr=shareIP
		 */
		if (evReq->expCfg.expHostList[idx].ro_rw == 1)
			strcat(mntInfoReqp->pathName, "ro");
		else if (evReq->expCfg.expHostList[idx].ro_rw == 2)
			strcat(mntInfoReqp->pathName, "rw");
		else 
			strcat(mntInfoReqp->pathName, "none");
		strcat(mntInfoReqp->pathName, ",");

		if (evReq->expCfg.expHostList[idx].rtsquash == 1)
			strcat(mntInfoReqp->pathName, "root_squash");
		else
			strcat(mntInfoReqp->pathName, "no_root_squash");

		if (evReq->expCfg.shareIP)
		{
			struct in_addr in;
			in.s_addr = evReq->expCfg.shareIP;
			strcat(mntInfoReqp->pathName,",laddr=");
			strcat(mntInfoReqp->pathName, (char *)inet_ntoa(in));
			mntInfoReqp->localIpAddr = evReq->expCfg.shareIP;
		}
		mntInfoReqp->optionsLen = strlen(mntInfoReqp->pathName)
										- mntInfoReqp->nameLen;

		/*
		 * Set the rest of the message
		 */

		mntReqLen = sizeof(mnt_infoReq_t) + mntInfoReqp->nameLen
							+ mntInfoReqp->optionsLen + 2;
		mntInfoReqp->msgHdr.private = 0;
		mntInfoReqp->msgHdr.msgType = MNT_EXPORT_REQ;

		mntInfoResp = (mnt_infoResp_t *) rspBuf.buf;

		rc = ea_sendClusterAgileMsg(-1, &reqBuf, mntReqLen,
				MOUNTD_APP_NAME, NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
				&timeout,
				&rspBuf, sizeof(rspBuf.buf),
				evReq->hdr.node);

		if ( (rc <= 0) || (mntInfoResp->msgHdr.status != NFX_OK) )
		{
			EA_ERR_PRINT(("ea_processAddExpReq: Export req for %s failed\n",
							mntInfoReqp->pathName));	
			retStat = NFX_ERR;
			continue;
		}

		/*
		 * Put the successfully exported entry in response
		 */
		memcpy(&eaRsp->expCfg.expHostList[eaRsp->expCfg.num],
					&evReq->expCfg.expHostList[idx], sizeof(ea_expHost_t));
		eaRsp->expCfg.num++;
	}

	/*
	 * Recompute the actual rsp len
	 */
	if (eaRsp->expCfg.num <= 0)
	{
		rspLen = sizeof(ea_addExpRsp_t);
	}
	else
	{
		rspLen = sizeof(ea_addExpRsp_t) + 
				(eaRsp->expCfg.num - 1)*sizeof(ea_expHost_t);
	}

    eaRsp->result = retStat;
    sprintf(eaRsp->output, "Done.");
    ea_sendMsg(eaRsp, rspLen, sender->name);
    free(eaRsp);

    EA_FCT_PRINT(("> ea_processAddExpReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processModExpReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processModExpReq(ea_modExpReq_t * evReq,
					ea_senderInfo_t * sender)
{
	int					retStat;
    ea_clusmsg_t		reqBuf;
    ea_clusmsg_t		rspBuf;
	mnt_infoReq_t       *mntInfoReqp;
	mnt_infoResp_t      *mntInfoResp;
	ea_modExpRsp_t		*eaRsp;
	int					rspLen = 0;
	int					mntReqLen = 0;
	int					idx = 0;
	int					numLoop;

	int                 rc;
	struct timeval      timeout = { 10, 0};

    EA_FCT_PRINT(("< ea_processModExpReq()\n"));

	rspLen = sizeof(ea_modExpRsp_t) + 
				(evReq->expCfg.num - 1)*sizeof(ea_expHost_t);
	eaRsp = (ea_modExpRsp_t *) malloc(rspLen);
	bzero((void *)eaRsp, rspLen);

    eaRsp->hdr.msgType = EA_MOD_EXP_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	eaRsp->fsysId = evReq->fsysId;
	strcpy(eaRsp->expCfg.path.pathName, evReq->expCfg.path.pathName);
	eaRsp->expCfg.shareIP = htonl(evReq->expCfg.shareIP);
	eaRsp->expCfg.num = 0;

	EA_DBG_PRINT(("ea_processModExpReq num = %d  shareIP = 0x%d\n",
			evReq->expCfg.num, evReq->expCfg.shareIP));

	if (evReq->expCfg.num == 0)
		numLoop = 1;
	else
		numLoop = evReq->expCfg.num;

	retStat = ea_getNodeVolIdFromVolName(evReq->fsysId.fsysName,
									evReq->hdr.node, NULL);
	if (retStat != NFX_OK)
		*(evReq->hdr.node) = '\0';

	for (idx = 0; idx < numLoop; idx++)
	{
		retStat = NFX_OK;

		memset(&reqBuf, 0, sizeof(ea_clusmsg_t));
		mntInfoReqp = (mnt_infoReq_t *) reqBuf.buf;

		/*
		 * Build client:/VolName/PathName
		 */
		if (evReq->expCfg.expHostList[idx].hostIP)
		{
			struct in_addr in;
			in.s_addr = evReq->expCfg.expHostList[idx].hostIP;
			strcpy(mntInfoReqp->pathName, (char *)inet_ntoa(in));
			strcat(mntInfoReqp->pathName,":");
		}

		strcat(mntInfoReqp->pathName,"/");
		strcat(mntInfoReqp->pathName, evReq->fsysId.fsysName);
		strcat(mntInfoReqp->pathName, evReq->expCfg.path.pathName);

		mntInfoReqp->nameLen = strlen(mntInfoReqp->pathName);

		/*
		 * Build options rw,ro,root_squash,no_root_squash,laddr=shareIP
		 */
		if (evReq->expCfg.expHostList[idx].ro_rw == 1)
			strcat(mntInfoReqp->pathName, "ro");
		else if (evReq->expCfg.expHostList[idx].ro_rw == 2)
			strcat(mntInfoReqp->pathName, "rw");
		else 
			strcat(mntInfoReqp->pathName, "none");
		strcat(mntInfoReqp->pathName, ",");

		if (evReq->expCfg.expHostList[idx].rtsquash == 1)
			strcat(mntInfoReqp->pathName, "root_squash");
		else
			strcat(mntInfoReqp->pathName, "no_root_squash");

		if (evReq->expCfg.shareIP)
		{
			struct in_addr in;
			in.s_addr = evReq->expCfg.shareIP;
			strcat(mntInfoReqp->pathName,",laddr=");
			strcat(mntInfoReqp->pathName, (char *)inet_ntoa(in));
		}
		mntInfoReqp->optionsLen = strlen(mntInfoReqp->pathName)
										- mntInfoReqp->nameLen;

		/*
		 * Set the rest of the message
		 */

		mntReqLen = sizeof(mnt_infoReq_t) + mntInfoReqp->nameLen
							+ mntInfoReqp->optionsLen + 2;
		mntInfoReqp->msgHdr.private = 0;
		mntInfoReqp->msgHdr.msgType = MNT_EXPORT_UPDATE;

		mntInfoResp = (mnt_infoResp_t *) rspBuf.buf;

		rc = ea_sendClusterAgileMsg(-1, &reqBuf, mntReqLen,
				MOUNTD_APP_NAME, NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
				&timeout,
				&rspBuf, sizeof(rspBuf.buf),
				evReq->hdr.node);
		if ( (rc <= 0) || (mntInfoResp->msgHdr.status != NFX_OK) )
		{
			EA_ERR_PRINT(("ea_processModExpReq: Export req for %s failed\n",
							mntInfoReqp->pathName));	
			retStat = NFX_ERR;
			continue;
		}

		/*
		 * Put the successfully exported entry in response
		 */
		memcpy(&eaRsp->expCfg.expHostList[eaRsp->expCfg.num],
					&evReq->expCfg.expHostList[idx], sizeof(ea_expHost_t));
		eaRsp->expCfg.num++;
	}

	/*
	 * Recompute the actual rsp len
	 */
	if (eaRsp->expCfg.num <= 0)
	{
		rspLen = sizeof(ea_modExpRsp_t);
	}
	else
	{
		rspLen = sizeof(ea_modExpRsp_t) + 
				(eaRsp->expCfg.num - 1)*sizeof(ea_expHost_t);
	}

    eaRsp->result = retStat;
    sprintf(eaRsp->output, "Done.");
    ea_sendMsg(eaRsp, rspLen, sender->name);
    free(eaRsp);

    EA_FCT_PRINT(("> ea_processModExpReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processDelExpReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processDelExpReq(ea_delExpReq_t * evReq,
					ea_senderInfo_t * sender)
{
	int					retStat;
    ea_clusmsg_t		reqBuf;
    ea_clusmsg_t		rspBuf;
	mnt_infoReq_t       *mntInfoReqp;
	mnt_infoResp_t      *mntInfoResp;
	ea_delExpRsp_t		*eaRsp;
	int					rspLen = 0;
	int					mntReqLen = 0;
	int					idx = 0;

	int                 rc;
	struct timeval      timeout = { 10, 0};

    EA_FCT_PRINT(("< ea_processDelExpReq()\n"));

	rspLen = sizeof(ea_delExpRsp_t) + 
				(evReq->expCfg.num - 1)*sizeof(ea_expHost_t);
	eaRsp = (ea_delExpRsp_t *) malloc(rspLen);
	bzero((void *)eaRsp, rspLen);

    eaRsp->hdr.msgType = EA_DEL_EXP_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	eaRsp->fsysId = evReq->fsysId;
	strcpy(eaRsp->expCfg.path.pathName, evReq->expCfg.path.pathName);
	eaRsp->expCfg.shareIP = evReq->expCfg.shareIP;
	eaRsp->expCfg.num = 0;

	for (idx = 0; idx < evReq->expCfg.num; idx++)
	{
		retStat = NFX_OK;

		memset(&reqBuf, 0, sizeof(ea_clusmsg_t));
		mntInfoReqp = (mnt_infoReq_t *) reqBuf.buf;

		mntInfoReqp->localIpAddr = evReq->expCfg.shareIP;

		/*
		 * Build client:/VolName/PathName
		 */
		if (evReq->expCfg.expHostList[idx].hostIP)
		{
			struct in_addr in;
			in.s_addr = evReq->expCfg.expHostList[idx].hostIP;
			strcpy(mntInfoReqp->pathName, (char *)inet_ntoa(in));
			strcat(mntInfoReqp->pathName,":");
		}

		strcat(mntInfoReqp->pathName,"/");
		strcat(mntInfoReqp->pathName, evReq->fsysId.fsysName);
		strcat(mntInfoReqp->pathName, evReq->expCfg.path.pathName);

		mntInfoReqp->nameLen = strlen(mntInfoReqp->pathName);

		EA_DBG_PRINT(("ea_processDelExpReq for %s\n",
				mntInfoReqp->pathName));

		/*
		 * Set the rest of the message
		 */

		mntReqLen = sizeof(mnt_infoReq_t) + mntInfoReqp->nameLen + 1;
		mntInfoReqp->msgHdr.private = 0;
		mntInfoReqp->msgHdr.msgType = MNT_UNEXPORT_REQ;

		mntInfoResp = (mnt_infoResp_t *) rspBuf.buf;

		rc = ea_sendClusterAgileMsg(-1, &reqBuf, mntReqLen,
				MOUNTD_APP_NAME, NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
				&timeout,
				&rspBuf, sizeof(rspBuf.buf),
				evReq->hdr.node);
		if ( (rc <= 0) || (mntInfoResp->msgHdr.status != NFX_OK) )
		{
			EA_ERR_PRINT(("ea_processDelExpReq: Export req for %s failed\n",
							mntInfoReqp->pathName));	
			retStat = NFX_ERR;
			continue;
		}

		/*
		 * Put the successfully deleted entry in response
		 */
		memcpy(&eaRsp->expCfg.expHostList[eaRsp->expCfg.num],
					&evReq->expCfg.expHostList[idx], sizeof(ea_expHost_t));
		eaRsp->expCfg.num++;
	}

	/*
	 * Recompute the actual rsp len
	 */
	 if (eaRsp->expCfg.num <= 0)
	 {
		rspLen = sizeof(ea_delExpRsp_t);
	 }
	 else
	 {
		rspLen = sizeof(ea_delExpRsp_t) + 
				(eaRsp->expCfg.num - 1)*sizeof(ea_expHost_t);
	 }

    eaRsp->result = retStat;
    sprintf(eaRsp->output, "Done.");
    ea_sendMsg(eaRsp, rspLen, sender->name);
    free(eaRsp);

    EA_FCT_PRINT(("> ea_processDelExpReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processDelPathReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processDelPathReq(ea_delPathReq_t * evReq,
					ea_senderInfo_t * sender)
{
	int					retStat;

    ea_clusmsg_t		getReqBuf;
    ea_clusmsg_t		getRspBuf;
	mnt_infoReq_t       *mntGetInfoReqp;
	mnt_infoResp_t      *mntGetInfoResp;

    ea_clusmsg_t		reqBuf;
    ea_clusmsg_t		rspBuf;
	mnt_infoReq_t       *mntInfoReqp;
	mnt_infoResp_t      *mntInfoResp;

	ea_delPathRsp_t		*eaRsp;
	int					rspLen = 0;
	int					i, j, idx = 0;

	int                 rc;
	struct timeval      timeout = { 10, 0};
	mnt_infoEntry_t		*mtnInfo;
	int					volNameLen;

    EA_FCT_PRINT(("< ea_processDelPathReq()\n"));

	rspLen = sizeof(ea_delPathRsp_t) + 
				(evReq->num - 1)*sizeof(ea_pathId_t);
	eaRsp = (ea_delPathRsp_t *) malloc(rspLen);
	bzero((void *)eaRsp, rspLen);

    eaRsp->hdr.msgType = EA_DEL_PATH_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	eaRsp->fsysId = evReq->fsysId;

	retStat = ea_getNodeVolIdFromVolName(evReq->fsysId.fsysName,
								evReq->hdr.node, &evReq->fsysId.volId);
	if (retStat != NFX_OK)
		*(evReq->hdr.node) = '\0';

	volNameLen = strlen(evReq->fsysId.fsysName);

	// get all pathes for a given volume request
	memset(&getReqBuf, 0, sizeof(ea_clusmsg_t));
	mntGetInfoReqp = (mnt_infoReq_t *) getReqBuf.buf;

	mntGetInfoReqp->msgHdr.msgType = MNT_EXPORT_SHOW;
	mntGetInfoReqp->volId = evReq->fsysId.volId;
	mntGetInfoReqp->nameLen = 0;		// All Paths
	mntGetInfoReqp->cookie = 0;

	while (1)
	{
		mntGetInfoResp = (mnt_infoResp_t *) getRspBuf.buf;

		rc = ea_sendClusterAgileMsg(-1, &getReqBuf, sizeof(mnt_infoReq_t),
					MOUNTD_APP_NAME, NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					&timeout,
					&getRspBuf, sizeof(getRspBuf.buf),
					evReq->hdr.node);

		/* Check response from FS. */
		if ((rc <= 0) || (mntGetInfoResp->msgHdr.status != NFX_OK))
			break;

		/* no more entries? */
		if ((mntGetInfoResp->numOfEntries == 0) ||
			(mntGetInfoReqp->cookie == mntGetInfoResp->cookie))
			break;

		mtnInfo = mntGetInfoResp->mtnInfoEntry;

		for (i = 0; i < mntGetInfoResp->numOfEntries; i ++)
		{
			char *p, tmpbuf[512];

			strncpy(tmpbuf, mtnInfo->pathName, mtnInfo->pathNameLen);
			tmpbuf[mtnInfo->pathNameLen] = '\0';

			for (idx = 0; idx < evReq->num; idx++)
			{
				char clientip[32];
				char pathbuf[512];

				if (strncmp(&tmpbuf[volNameLen],
						evReq->pathList[idx].pathName,
						strlen(evReq->pathList[idx].pathName)))
					continue;

				// need to delete this share

				/* find the client first... */
				p = mtnInfo->pathName + mtnInfo->pathNameLen;

				j = 0;
				while (isdigit(*p) || (*p == '.'))
					clientip[j++] = *p++;
				clientip[j] = '\0';

				if (j == 0)
					sprintf(pathbuf, "/%s", tmpbuf);
				else
					sprintf(pathbuf, "%s:/%s", clientip, tmpbuf);

				/* Fill the request. */
				memset(&reqBuf, 0, sizeof(reqBuf));

				mntInfoReqp = (mnt_infoReq_t *) reqBuf.buf;
				mntInfoReqp->msgHdr.msgType = MNT_UNEXPORT_REQ;
				mntInfoReqp->msgHdr.private = 0;
				mntInfoReqp->volId = evReq->fsysId.volId;
				mntInfoReqp->nameLen = strlen(pathbuf);
				strcpy(mntInfoReqp->pathName, pathbuf);

				if (mtnInfo->localIpAddr)
					mntInfoReqp->localIpAddr = mtnInfo->localIpAddr;

				/* Send message to FS. */
				mntInfoResp = (mnt_infoResp_t *) rspBuf.buf;

				rc = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(mnt_infoReq_t)+mntInfoReqp->nameLen+1,
						MOUNTD_APP_NAME, NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
						&timeout,
						&rspBuf, sizeof(rspBuf.buf),
						evReq->hdr.node);

				if ( (rc <= 0) || (mntInfoResp->msgHdr.status != NFX_OK) )
				{
					EA_ERR_PRINT(("ea_processDelPathReq: Export req for %s failed\n",
									mntInfoReqp->pathName));	
				}

				/*
				 * Put deleted entry in response
				 */
				for (j = 0; j < eaRsp->num; j++)
				{
					if (!strcmp(eaRsp->pathList[j].pathName,
								evReq->pathList[idx].pathName))
						break;
				}
				if (j >= eaRsp->num)
				{
					memcpy(&eaRsp->pathList[eaRsp->num],
							&evReq->pathList[idx], sizeof(ea_pathId_t));
					eaRsp->num++;
				}
				break;
			}

			MNT_INFO_RESP_ADV_ENTRY_PTR(mtnInfo);
		}

		mntGetInfoReqp->cookie = mntGetInfoResp->cookie;
	}

	/*
	 * Recompute the actual rsp len
	 */
	if (eaRsp->num <= 0)
	{
		rspLen = sizeof(ea_delPathRsp_t);
	}
	else
	{
		rspLen = sizeof(ea_delPathRsp_t) + 
				(eaRsp->num - 1)*sizeof(ea_pathId_t);
	}

    eaRsp->result = retStat;
    sprintf(eaRsp->output, "Done.");
    ea_sendMsg(eaRsp, rspLen, sender->name);
    free(eaRsp);

    EA_FCT_PRINT(("> ea_processDelPathReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processGetExpReq()
 *
 * Description  : Get all the exports for the specified filesystem.
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processGetExpReq(ea_getExpReq_t * evReq,
					ea_senderInfo_t * sender)
{
    ea_clusmsg_t		reqBuf;
    ea_clusmsg_t		rspBuf;
	mnt_infoReq_t       *mntInfoReqp;
	mnt_infoResp_t      *mntInfoResp;
	int					idx, rc;
	struct timeval		timeout = { 10, 0};
	mnt_infoEntry_t 	*mtnInfoEntryp;
	ea_getExpRsp_t		*eaRsp;
	int					rspLen = 0;
	int					numOfEntries;
	int					fsysNameLen;
	char            	fsysName[MAX_NAME_LEN];

    EA_FCT_PRINT(("< ea_processGetExpReq()\n"));

	/*
	 * Allocate the EA response message
	 */
	rspLen = sizeof(ea_getExpRsp_t);
	eaRsp = (ea_getExpRsp_t *) malloc(rspLen);
	bzero((void *)eaRsp, rspLen);
    eaRsp->hdr.msgType = EA_GET_EXP_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	eaRsp->fsysId = evReq->fsysId;

	rc = ea_getNodeVolIdFromVolName(evReq->fsysId.fsysName,
									evReq->hdr.node, NULL);
	if (rc != NFX_OK)
		*(evReq->hdr.node) = '\0';

	/*
	 * Setup Mnt request and response message
	 */
	memset(&reqBuf, 0, sizeof(ea_clusmsg_t));
	mntInfoReqp = (mnt_infoReq_t *) reqBuf.buf;
	mntInfoReqp->msgHdr.msgType = MNT_EXPORT_SHOWALL;
	mntInfoReqp->msgHdr.private = 0;

	mntInfoResp = (mnt_infoResp_t *) rspBuf.buf;

	rc = ea_sendClusterAgileMsg(-1, &reqBuf, sizeof(mnt_infoReq_t),
				MOUNTD_APP_NAME, NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
				&timeout,
				&rspBuf, sizeof(rspBuf.buf),
				evReq->hdr.node);

	if ( (rc <= 0) || (mntInfoResp->msgHdr.status != NFX_OK) )
	{
        sprintf(eaRsp->output, 
			"ea_processGetExpReq: Get export failed - %d", rc);
        eaRsp->result = rc;
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_getExpRsp_t), sender->name);
        return;
	}

	numOfEntries = mntInfoResp->numOfEntries;

	if (numOfEntries > 1)
	{
		/*
	 	 * Reallocate EA rsp based on num entries returned
	 	 */
		rspLen = sizeof(ea_getExpRsp_t) +
					(numOfEntries - 1)*sizeof(ea_expCfg_t);
    	eaRsp = (ea_getExpRsp_t *) realloc((void *) eaRsp, rspLen);
		bzero((void *)eaRsp, rspLen);
    	eaRsp->hdr.msgType = EA_GET_EXP_RSP;
    	eaRsp->hdr.userInfo = evReq->hdr.userInfo;
		eaRsp->fsysId = evReq->fsysId;
	}

	mtnInfoEntryp = &mntInfoResp->mtnInfoEntry[0];		
	for (idx=0; idx < numOfEntries; idx++)
	{
		/*
		 * Find matching path entry
		 */
		strcpy(fsysName, evReq->fsysId.fsysName);
		fsysNameLen = strlen(fsysName);

		EA_DBG_PRINT(("Comparing fsysName %s vs mtnInfoEntryp->pathName %s len %d\n", 
					fsysName, mtnInfoEntryp->pathName,
						fsysNameLen));

		if (strncmp(fsysName, mtnInfoEntryp->pathName, fsysNameLen) != 0)
		{
			EA_DBG_PRINT(("ea_processGetExpReq idx = %d did not match\n", idx));
			MNT_INFO_RESP_ADV_ENTRY_PTR(mtnInfoEntryp);
			continue;
		}

		/*
		 * Copy the info in host response
		 */
		eaRsp->expList[eaRsp->num].shareIP = mtnInfoEntryp->localIpAddr;
		// eaRsp->expList[eaRsp->num].shareIP = htonl(mtnInfoEntryp->localIpAddr);
		strncpy(eaRsp->expList[eaRsp->num].path.pathName, 
						&mtnInfoEntryp->pathName[0] + fsysNameLen,
							mtnInfoEntryp->pathNameLen - fsysNameLen);
		eaRsp->expList[eaRsp->num].num = 1;

		if (mtnInfoEntryp->exportOptionMask & MOUNTD_RW_OPT)
			eaRsp->expList[eaRsp->num].expHostList[0].ro_rw = 2;
		else if (mtnInfoEntryp->exportOptionMask & MOUNTD_RO_OPT)
			eaRsp->expList[eaRsp->num].expHostList[0].ro_rw = 1;
		else
			eaRsp->expList[eaRsp->num].expHostList[0].ro_rw = 0;

		if (mtnInfoEntryp->exportOptionMask & MOUNTD_ROOT_SQUASH_OPT)
			eaRsp->expList[eaRsp->num].expHostList[0].rtsquash = 1;
		else
			eaRsp->expList[eaRsp->num].expHostList[0].rtsquash = 0;

		eaRsp->expList[eaRsp->num].expHostList[0].hostIP = 
											mtnInfoEntryp->hostAddr;
											// htonl(mtnInfoEntryp->hostAddr);
		EA_DBG_PRINT(("Adding to response fsys %s path %s Host IP 0x%x\n",
					fsysName, eaRsp->expList[eaRsp->num].path.pathName,
							mtnInfoEntryp->hostAddr));
		eaRsp->num++;

		MNT_INFO_RESP_ADV_ENTRY_PTR(mtnInfoEntryp);
	}	

	/*
	 * Recompute the actual rsp len
	 */
	if (eaRsp->num <= 0)
	{
		rspLen = sizeof(ea_getExpRsp_t);
	}
	else
	{
		rspLen = sizeof(ea_getExpRsp_t) +
					(eaRsp->num - 1)*sizeof(ea_expCfg_t);
	}

	EA_DBG_PRINT(("ea_processGetExpReq returning %d entries\n", eaRsp->num));

    eaRsp->result = NFX_OK;
    sprintf(eaRsp->output, "Done.");
    ea_sendMsg(eaRsp, rspLen, sender->name);

    EA_FCT_PRINT(("> ea_processGetExpReq()\n"));
	return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processGetPathExpReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processGetPathExpReq(ea_getPathExpReq_t * evReq,
					ea_senderInfo_t * sender)
{
    ea_clusmsg_t		reqBuf;
    ea_clusmsg_t		rspBuf;
	mnt_infoReq_t       *mntInfoReqp;
	mnt_infoResp_t		*mntInfoResp;
	int					idx, rc;
	struct timeval		timeout = { 10, 0};
	mnt_infoEntry_t 	*mtnInfoEntryp;
	ea_getPathExpRsp_t	*eaRsp;
	int					rspLen = 0;
	int					numOfEntries;
	int					fsysNameLen;
	char            	pathName[MAX_NAME_LEN];

    EA_FCT_PRINT(("< ea_processGetPathExpReq()\n"));

	/*
	 * Allocate the EA response message
	 */
	rspLen = sizeof(ea_getPathExpRsp_t);
	eaRsp = (ea_getPathExpRsp_t *) malloc(rspLen);
	bzero((void *)eaRsp, rspLen);
    eaRsp->hdr.msgType = EA_GET_PATH_EXP_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	eaRsp->fsysId = evReq->fsysId;

	rc = ea_getNodeVolIdFromVolName(evReq->fsysId.fsysName,
									evReq->hdr.node, NULL);
	if (rc != NFX_OK)
		*(evReq->hdr.node) = '\0';

	memset(&reqBuf, 0, sizeof(ea_clusmsg_t));
	mntInfoReqp = (mnt_infoReq_t *) reqBuf.buf;

	/*
	 * Setup Mnt request and response message
	 */
	mntInfoReqp->msgHdr.msgType = MNT_EXPORT_SHOWALL;
	mntInfoReqp->msgHdr.private = 0;

	mntInfoResp = (mnt_infoResp_t *) rspBuf.buf;

	rc = ea_sendClusterAgileMsg(-1, &reqBuf, sizeof(mnt_infoReq_t),
				MOUNTD_APP_NAME, NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
				&timeout,
				&rspBuf, sizeof(rspBuf.buf),
				evReq->hdr.node);

	if ( (rc <= 0) || (mntInfoResp->msgHdr.status != NFX_OK) )
	{
        sprintf(eaRsp->output, 
			"ea_processGetPathExpReq: Get export failed - %d", rc);
        eaRsp->result = rc;
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_getPathExpRsp_t), sender->name);
        return;
	}

	numOfEntries = mntInfoResp->numOfEntries;

	if (numOfEntries > 1)
	{
		/*
	 	 * Reallocate EA rsp based on num entries returned
	 	 */
		rspLen = sizeof(ea_getPathExpRsp_t) +
					(numOfEntries - 1)*sizeof(ea_expHost_t);
    	eaRsp = (ea_getPathExpRsp_t *) realloc((void *) eaRsp, rspLen);
		bzero((void *)eaRsp, rspLen);
    	eaRsp->hdr.msgType = EA_GET_PATH_EXP_RSP;
    	eaRsp->hdr.userInfo = evReq->hdr.userInfo;
		eaRsp->fsysId = evReq->fsysId;
	}

	mtnInfoEntryp = &mntInfoResp->mtnInfoEntry[0];		
	for (idx=0; idx < numOfEntries; idx++)
	{
		/*
		 * Find matching path entry
		 */
		strcpy(pathName, evReq->fsysId.fsysName);
		fsysNameLen = strlen(evReq->fsysId.fsysName);
		strcat(pathName, evReq->path.pathName);

		EA_DBG_PRINT(("Comparing pathName %s vs mtnInfoEntryp->pathName %s len %d\n", 
					pathName, mtnInfoEntryp->pathName,
						strlen(pathName)));

		if (strncmp(pathName, mtnInfoEntryp->pathName,
						strlen(pathName)) != 0)
		{
			EA_DBG_PRINT(("idx = %d did not match\n", idx));
			MNT_INFO_RESP_ADV_ENTRY_PTR(mtnInfoEntryp);
			continue;
		}

		/*
		 * Copy the info in host response
		 */
		eaRsp->expCfg.shareIP = mtnInfoEntryp->localIpAddr;
		// eaRsp->expCfg.shareIP = htonl(mtnInfoEntryp->localIpAddr);
		strncpy(eaRsp->expCfg.path.pathName, 
					&mtnInfoEntryp->pathName[0] + fsysNameLen,
						mtnInfoEntryp->pathNameLen - fsysNameLen);


		if (mtnInfoEntryp->exportOptionMask & MOUNTD_RW_OPT)
			eaRsp->expCfg.expHostList[eaRsp->expCfg.num].ro_rw = 2;
		else if (mtnInfoEntryp->exportOptionMask & MOUNTD_RO_OPT)
			eaRsp->expCfg.expHostList[eaRsp->expCfg.num].ro_rw = 1;
		else
			eaRsp->expCfg.expHostList[eaRsp->expCfg.num].ro_rw = 0;

		if (mtnInfoEntryp->exportOptionMask & MOUNTD_ROOT_SQUASH_OPT)
			eaRsp->expCfg.expHostList[eaRsp->expCfg.num].rtsquash = 1;
		else
			eaRsp->expCfg.expHostList[eaRsp->expCfg.num].rtsquash = 0;

		eaRsp->expCfg.expHostList[eaRsp->expCfg.num].hostIP = 
											mtnInfoEntryp->hostAddr;
											// htonl(mtnInfoEntryp->hostAddr);

		EA_DBG_PRINT(("Adding to response fsys %s path %s Host IP 0x%x\n",
					evReq->fsysId.fsysName, eaRsp->expCfg.path.pathName,
							mtnInfoEntryp->hostAddr));
		eaRsp->expCfg.num++;

		MNT_INFO_RESP_ADV_ENTRY_PTR(mtnInfoEntryp);
	}	

	/*
	 * Recompute the actual rsp len
	 */
	if (eaRsp->expCfg.num <= 0)
	{
		rspLen = sizeof(ea_getPathExpRsp_t);
	}
	else
	{
		rspLen = sizeof(ea_getPathExpRsp_t) +
					(eaRsp->expCfg.num - 1)*sizeof(ea_expHost_t);
	}

	EA_DBG_PRINT(("ea_processGetPathExpReq returning %d entries\n", eaRsp->expCfg.num));

    eaRsp->result = NFX_OK;
    sprintf(eaRsp->output, "Done.");
    ea_sendMsg(eaRsp, rspLen, sender->name);

    EA_FCT_PRINT(("> ea_processGetPathExpReq()\n"));
	return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processCrtSnapReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processCrtSnapReq(ea_crtSnapReq_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32               	rc = NFX_OK;
    ea_fsysCb_t     		cb;
    ea_crtSnapRsp_t 		eaRsp;

    ea_clusmsg_t			reqBuf;
    ea_clusmsg_t			rspBuf;
	fs_MsgSnapCreateReq 	*fsReq;
	fs_MsgSnapCreateResp 	*fsRsp;
    
    EA_FCT_PRINT(("> ea_processCrtSnapReq()\n"));

    /* Initialize the response.
     */
    bzero((void *) &eaRsp, sizeof(ea_crtSnapRsp_t));
    eaRsp.hdr.msgType = EA_CRT_SNAP_RSP;
    eaRsp.hdr.userInfo = evReq->hdr.userInfo;
	memcpy(&eaRsp.fsysId, &evReq->fsysId, sizeof(ea_fsysId_t));

    /* Initialize the control block.
     */
    bzero((void *) &cb, sizeof(ea_fsysCb_t));
    cb.hdr = (ea_msgHdr_t *) evReq;
    cb.vol = (evm_VolConfig_t *) malloc(sizeof(evm_VolConfig_t));
    NFX_ASSERT(cb.vol != NULL);
    bzero((void *) cb.vol, sizeof(evm_VolConfig_t));
    strcpy(cb.vol->name, evReq->fsysId.fsysName);
    cb.vol->onNIM = 1;

	// Get volume id from volume name, and node name that
	// has this volume online...
	rc = ea_getClusterVolIdNode(cb.vol, cb.hdr->node, eaRsp.output);
	// rc = ea_getVolId(cb.vol, eaRsp->output);
    if (rc != NFX_OK)
    {
        eaRsp.result = rc;
        EA_ERR_PRINT(("%s\n", eaRsp.output));
        ea_sendMsg(&eaRsp, sizeof(ea_crtSnapRsp_t), sender->name);
		free(cb.vol);
        return;
    }

	/*
	 * Check if the destination CPU is up
	 */
	if ((checkClusterAgileDest(cb.hdr->node, NFX_NFP_FP1, 1) != NFX_OK) &&
        (checkClusterAgileDest(cb.hdr->node, NFX_NFP_FP2, 1) != NFX_OK))
	{
        eaRsp.result = NFX_ERR;
    	sprintf(eaRsp.output, "Dest CPU not responding.");
        ea_sendMsg(&eaRsp, sizeof(ea_crtSnapRsp_t), sender->name);
		free(cb.vol);
        return;
	}

	memset(&reqBuf, 0, sizeof(ea_clusmsg_t));
	fsReq = (fs_MsgSnapCreateReq *) reqBuf.buf;

	fsReq->hdr.requestType = FS_REQ_SNAP_CREATE;
	fsReq->hdr.len = sizeof(fs_MsgSnapCreateReq);
	fsReq->volId = cb.vol->id;
	fsReq->nameLen = strlen(evReq->snapName);
	strcpy(fsReq->snapDirName, evReq->snapName);

	/* 
	 * Send request to EFS
	 */
	fsRsp = (fs_MsgSnapCreateResp *) rspBuf.buf;

	rc = ea_sendClusterAgileMsg(-1, &reqBuf, sizeof(fs_MsgSnapCreateReq),
				"efs", NFX_NFP_FP_ANY, 1, MGMTPLANE,
				&waitEfsTo,
				&rspBuf, sizeof(rspBuf.buf),
				cb.hdr->node);

	/* 
	 * Receive response from FS.
	 */
	if ((rc <= 0) || (fsRsp->hdr.fsStatus != NFX_OK))
	{
		free(cb.vol);
		eaRsp.result = NFX_ERR;
    	sprintf(eaRsp.output, "EFS error.");
	}
	else
	{
		eaRsp.result = NFX_OK;
    	sprintf(eaRsp.output, "Done.");
	}

	eaRsp.fsysId.volId = cb.vol->id;
	strcpy(eaRsp.snapName, evReq->snapName);

    ea_sendMsg(&eaRsp, sizeof(ea_crtSnapRsp_t), sender->name);

	EA_FCT_PRINT(("< ea_processCrtSnapReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processDelSnapReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processDelSnapReq(ea_delSnapReq_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32               	rc = NFX_OK;
    ea_fsysCb_t     		cb;
    ea_delSnapRsp_t 		eaRsp;
    ea_clusmsg_t			reqBuf;
    ea_clusmsg_t			rspBuf;
	fs_MsgSnapRemoveReq 	*fsReq;
	fs_MsgSnapRemoveResp 	*fsRsp;
    
    EA_FCT_PRINT(("> ea_processDelSnapReq()\n"));

    /* Initialize the response.
     */
    bzero((void *) &eaRsp, sizeof(ea_delSnapRsp_t));
    eaRsp.hdr.msgType = EA_DEL_SNAP_RSP;
    eaRsp.hdr.userInfo = evReq->hdr.userInfo;
	memcpy(&eaRsp.fsysId, &evReq->fsysId, sizeof(ea_fsysId_t));

    /* Initialize the control block.
     */
    bzero((void *) &cb, sizeof(ea_fsysCb_t));
    cb.hdr = (ea_msgHdr_t *) evReq;
    cb.vol = (evm_VolConfig_t *) malloc(sizeof(evm_VolConfig_t));
    NFX_ASSERT(cb.vol != NULL);
    bzero((void *) cb.vol, sizeof(evm_VolConfig_t));
    strcpy(cb.vol->name, evReq->fsysId.fsysName);
    cb.vol->onNIM = 1;

	// Get volume id from volume name, and node name that
	// has this volume online...
	rc = ea_getClusterVolIdNode(cb.vol, cb.hdr->node, eaRsp.output);
	// rc = ea_getVolId(cb.vol, eaRsp->output);
    if (rc != NFX_OK)
    {
        eaRsp.result = rc;
        EA_ERR_PRINT(("%s\n", eaRsp.output));
        ea_sendMsg(&eaRsp, sizeof(ea_delSnapRsp_t), sender->name);
		free(cb.vol);
        return;
    }

	/*
	 * Check if the destination CPU is up
	 */
	if ((checkClusterAgileDest(cb.hdr->node, NFX_NFP_FP1, 1) != NFX_OK) &&
        (checkClusterAgileDest(cb.hdr->node, NFX_NFP_FP2, 1) != NFX_OK))
	{
        eaRsp.result = NFX_ERR;
    	sprintf(eaRsp.output, "Dest CPU not responding.");
        ea_sendMsg(&eaRsp, sizeof(ea_delSnapRsp_t), sender->name);
		free(cb.vol);
        return;
	}

	memset(&reqBuf, 0, sizeof(ea_clusmsg_t));
	fsReq = (fs_MsgSnapRemoveReq *) reqBuf.buf;

	fsReq->hdr.requestType = FS_REQ_SNAP_REMOVE;
	fsReq->hdr.len = sizeof(fs_MsgSnapRemoveReq);
	fsReq->volId = cb.vol->id;
	fsReq->nameLen = strlen(evReq->snapName);
	strcpy(fsReq->snapDirName, evReq->snapName);

	/* 
	 * Send request to EFS
	 */
	fsRsp = (fs_MsgSnapRemoveResp *) rspBuf.buf;

	rc = ea_sendClusterAgileMsg(-1,
				&reqBuf, sizeof(fs_MsgSnapRemoveReq),
				"efs", NFX_NFP_FP_ANY, 1, MGMTPLANE,
				&waitEfsTo,
				&rspBuf, sizeof(rspBuf.buf),
				cb.hdr->node);

	/* 
	 * Receive response from FS.
	 */
	if ((rc <= 0) || (fsRsp->hdr.fsStatus != NFX_OK))
	{
		free(cb.vol);
		eaRsp.result = NFX_ERR;
    	sprintf(eaRsp.output, "EFS error.");
	}
	else
	{
		eaRsp.result = NFX_OK;
    	sprintf(eaRsp.output, "Done.");
	}

	eaRsp.fsysId.volId = cb.vol->id;
	strcpy(eaRsp.snapName, evReq->snapName);

    ea_sendMsg(&eaRsp, sizeof(ea_delSnapRsp_t), sender->name);

	EA_FCT_PRINT(("< ea_processDelSnapReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processSchSnapReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processSchSnapReq(ea_schSnapReq_t * evReq,
                     ea_senderInfo_t * sender)
{
	uint32					i;
    int32               	rc = NFX_OK;
    ea_fsysCb_t     		cb;
    ea_clusmsg_t			reqBuf;
    ea_clusmsg_t			rspBuf;
	fs_MsgSnapSchedReq 		*fsReq;
	fs_MsgSnapSchedResp 	*fsRsp;
    ea_schSnapRsp_t 		eaRsp;

    
    EA_FCT_PRINT(("> ea_processSchSnapReq()\n"));

    /* Initialize the response.
     */
    bzero((void *) &eaRsp, sizeof(ea_schSnapRsp_t));
    eaRsp.hdr.msgType = EA_SCH_SNAP_RSP;
    eaRsp.hdr.userInfo = evReq->hdr.userInfo;

    /* Initialize the control block.
     */
    bzero((void *) &cb, sizeof(ea_fsysCb_t));
    cb.hdr = (ea_msgHdr_t *) evReq;
    cb.vol = (evm_VolConfig_t *) malloc(sizeof(evm_VolConfig_t));
    NFX_ASSERT(cb.vol != NULL);
    bzero((void *) cb.vol, sizeof(evm_VolConfig_t));
    strcpy(cb.vol->name, evReq->snapSch.fsysId.fsysName);
    cb.vol->onNIM = 1;

	// Get volume id from volume name, and node name that
	// has this volume online...
	rc = ea_getClusterVolIdNode(cb.vol, cb.hdr->node, eaRsp.output);
	// rc = ea_getVolId(cb.vol, eaRsp->output);
    if (rc != NFX_OK)
    {
        eaRsp.result = rc;
        EA_ERR_PRINT(("%s\n", eaRsp.output));
        ea_sendMsg(&eaRsp, sizeof(ea_schSnapRsp_t), sender->name);
		free(cb.vol);
        return;
    }

	/*
	 * Check if the destination CPU is up
	 */
	if ((checkClusterAgileDest(cb.hdr->node, NFX_NFP_FP1, 1) != NFX_OK) &&
        (checkClusterAgileDest(cb.hdr->node, NFX_NFP_FP2, 1) != NFX_OK))
	{
        eaRsp.result = NFX_ERR;
    	sprintf(eaRsp.output, "Dest CPU not responding.");
        ea_sendMsg(&eaRsp, sizeof(ea_schSnapRsp_t), sender->name);
		free(cb.vol);
        return;
	}

	memset(&reqBuf, 0, sizeof(ea_clusmsg_t));
	fsReq = (fs_MsgSnapSchedReq *) reqBuf.buf;

	fsReq->hdr.requestType = FS_REQ_SNAP_SCHED;
	fsReq->hdr.len = sizeof(fs_MsgSnapSchedReq);
	fsReq->volId = cb.vol->id;
	fsReq->sched.maxWeekly = evReq->snapSch.nbWeekSnap; 
	fsReq->sched.maxDaily =  evReq->snapSch.nbDaySnap;
	fsReq->sched.maxHourly =  evReq->snapSch.nbHourSnap;
	for (i = 0; i < FS_SNAP_SCHED_MAX_HOURS; i++)
		fsReq->sched.hours[i] = evReq->snapSch.hourSchedule[i];

	/* 
	 * Send request to EFS
	 */
	fsRsp = (fs_MsgSnapSchedResp *) rspBuf.buf;

	rc = ea_sendClusterAgileMsg(-1, &reqBuf, sizeof(fs_MsgSnapSchedReq),
				"efs", NFX_NFP_FP_ANY, 1, MGMTPLANE,
				&waitEfsTo,
				&rspBuf, sizeof(rspBuf.buf),
				cb.hdr->node);

	/* 
	 * Receive response from FS.
	 */
	if ((rc <= 0) || (fsRsp->hdr.fsStatus != NFX_OK))
	{
		eaRsp.result = NFX_ERR;
    	sprintf(eaRsp.output, "EFS error.");
	}
	else
	{
		eaRsp.result = NFX_OK;
    	sprintf(eaRsp.output, "Done.");
	}

	memcpy(&eaRsp.snapSch, &evReq->snapSch, sizeof(ea_snapSch_t));
	eaRsp.snapSch.fsysId.volId = cb.vol->id;

    ea_sendMsg(&eaRsp, sizeof(ea_schSnapRsp_t), sender->name);

	free(cb.vol);

	EA_FCT_PRINT(("< ea_processSchSnapReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processGetSchSnapReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processGetSchSnapReq(ea_getSchSnapReq_t * evReq,
                     ea_senderInfo_t * sender)
{
	uint32				fsIdx;
    int32               rc = NFX_OK;
    ea_fsysCb_t     	cb;
    ea_clusmsg_t		reqBuf;
    ea_clusmsg_t		rspBuf;
	fs_MsgSnapShowReq 	*fsReq;
	fs_MsgSnapShowResp 	*fsRsp;
							
	ea_snapSch_t 		*snapSchEntry;
    ea_getSchSnapRsp_t 	*eaRsp;
	uint32 				eaRspLen = 0;

    EA_FCT_PRINT(("> ea_processGetSchSnapReq()\n"));

	/*
	 * Max response size is FS_MAX_USER_SNAPSHOTS per fsys times number
	 * of fsys requested. evReq->num cannot be zero
	 */
	eaRspLen = sizeof(ea_getSchSnapRsp_t) + 
					(FS_MAX_USER_SNAPSHOTS - 1)*sizeof(ea_snapSch_t);
	eaRsp = (ea_getSchSnapRsp_t *) malloc(eaRspLen);
	bzero((void *)eaRsp, eaRspLen);
    eaRsp->hdr.msgType = EA_GET_SCH_SNAP_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;


	for (fsIdx = 0; fsIdx < evReq->num; fsIdx++)
	{
    	/* 
	 	 * Initialize the control block.
     	 */
    	bzero((void *) &cb, sizeof(ea_fsysCb_t));
    	cb.hdr = (ea_msgHdr_t *) evReq;
    	cb.vol = (evm_VolConfig_t *) malloc(sizeof(evm_VolConfig_t));
    	NFX_ASSERT(cb.vol != NULL);
    	bzero((void *) cb.vol, sizeof(evm_VolConfig_t));
    	strcpy(cb.vol->name, evReq->fsysId[fsIdx].fsysName);
    	cb.vol->onNIM = 1;
		
		// Get volume id from volume name, and node name that
		// has this volume online...
		rc = ea_getClusterVolIdNode(cb.vol, cb.hdr->node, eaRsp->output);
		// rc = ea_getVolId(cb.vol, eaRsp->output);
    	if (rc != NFX_OK)
			continue;

		/*
		 * Check if the destination CPU is up
		 */
        if ((checkClusterAgileDest(cb.hdr->node, NFX_NFP_FP1, 1) != NFX_OK) &&
            (checkClusterAgileDest(cb.hdr->node, NFX_NFP_FP2, 1) != NFX_OK))
			continue;

		memset(&reqBuf, 0, sizeof(ea_clusmsg_t));
		fsReq = (fs_MsgSnapShowReq *) reqBuf.buf;

		fsReq->hdr.requestType = FS_REQ_SNAP_SHOW;
		fsReq->hdr.len = sizeof(fs_MsgSnapShowReq);
		fsReq->volId = cb.vol->id;
		fsReq->type = FS_SNAP_SHOW_SCHED;

		/* 
	 	 * Send request to EFS
	 	 */
		fsRsp = (fs_MsgSnapShowResp *) rspBuf.buf;

		rc = ea_sendClusterAgileMsg(-1,
						&reqBuf, sizeof(fs_MsgSnapShowReq),
						"efs", NFX_NFP_FP_ANY, 1, MGMTPLANE,
						&waitEfsTo,
						&rspBuf, sizeof(rspBuf.buf),
						cb.hdr->node);
	
		/* 
	 	 * Receive response from FS.
	 	 */
		if ((rc <= 0) || (fsRsp->hdr.fsStatus != NFX_OK))
       		continue;

		snapSchEntry = &eaRsp->snapList[eaRsp->num];

		strcpy(snapSchEntry->fsysId.fsysName, evReq->fsysId[fsIdx].fsysName);
		snapSchEntry->fsysId.volId = cb.vol->id;

		snapSchEntry->nbWeekSnap = fsRsp->sched.maxWeekly;
		snapSchEntry->nbDaySnap = fsRsp->sched.maxDaily;
		snapSchEntry->nbHourSnap = fsRsp->sched.maxHourly;

		memcpy((char *)snapSchEntry->hourSchedule, 
				(char *)fsRsp->sched.hours, 
				sizeof(uint32)*FS_SNAP_SCHED_MAX_HOURS);

		eaRsp->num++;
	}

	eaRsp->result = NFX_OK;
    sprintf(eaRsp->output, "Done.");

	if (eaRsp->num <= 0)
	{
		eaRspLen = sizeof(ea_getSchSnapRsp_t);
	}
	else
	{
		eaRspLen = (sizeof(ea_getSchSnapRsp_t) + 
					(eaRsp->num - 1)*sizeof(ea_snapSch_t));
	}

    ea_sendMsg(eaRsp, eaRspLen, sender->name);

	EA_FCT_PRINT(("< ea_processGetSchSnapReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processGetUsgSnapReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processGetUsgSnapReq(ea_getUsgSnapReq_t * evReq,
                     ea_senderInfo_t * sender)
{
	uint32				i;
    int32               rc = NFX_OK;
    ea_fsysCb_t     	cb;
    ea_clusmsg_t		reqBuf;
    ea_clusmsg_t		rspBuf;
	fs_MsgSnapShowReq 	*fsReq;
	fs_MsgSnapShowResp 	*fsRsp;
    ea_getUsgSnapRsp_t 	*eaRsp;
	uint32 				eaRspLen = 0;

    EA_FCT_PRINT(("> ea_processGetUsgSnapReq()\n"));

	/*
	 * Max response size is FS_MAX_USER_SNAPSHOTS per fsys times number
	 * of fsys requested. evReq->num cannot be zero
	 */
	eaRspLen = sizeof(ea_getUsgSnapRsp_t) + 
					(FS_MAX_USER_SNAPSHOTS - 1)*sizeof(ea_snapSch_t);
	eaRsp = (ea_getUsgSnapRsp_t *) malloc(eaRspLen);
	bzero((void *)eaRsp, eaRspLen);
    eaRsp->hdr.msgType = EA_GET_USG_SNAP_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(&eaRsp->usageList.fsysId, &evReq->fsysId, sizeof(ea_fsysId_t));

    /* 
	 * Initialize the control block.
     */
    bzero((void *) &cb, sizeof(ea_fsysCb_t));
    cb.hdr = (ea_msgHdr_t *) evReq;
    cb.vol = (evm_VolConfig_t *) malloc(sizeof(evm_VolConfig_t));
    NFX_ASSERT(cb.vol != NULL);
    bzero((void *) cb.vol, sizeof(evm_VolConfig_t));
    strcpy(cb.vol->name, evReq->fsysId.fsysName);
    cb.vol->onNIM = 1;

	// Get volume id from volume name, and node name that
	// has this volume online...
	rc = ea_getClusterVolIdNode(cb.vol, cb.hdr->node, eaRsp->output);
	// rc = ea_getVolId(cb.vol, eaRsp->output);
    if (rc != NFX_OK)
    {
        eaRsp->result = rc;
        EA_ERR_PRINT(("%s\n", eaRsp->output));
        ea_sendMsg(eaRsp, sizeof(ea_getUsgSnapRsp_t), sender->name);
        return;
    }

	/*
	 * Check if the destination CPU is up
	 */
	if ((checkClusterAgileDest(cb.hdr->node, NFX_NFP_FP1, 1) != NFX_OK) &&
        (checkClusterAgileDest(cb.hdr->node, NFX_NFP_FP2, 1) != NFX_OK))
	{
        eaRsp->result = NFX_ERR;
    	sprintf(eaRsp->output, "Dest CPU not responding.");
        ea_sendMsg(eaRsp, sizeof(ea_getUsgSnapRsp_t), sender->name);
        return;
	}

	memset(&reqBuf, 0, sizeof(ea_clusmsg_t));
	fsReq = (fs_MsgSnapShowReq *) reqBuf.buf;

	fsReq->hdr.requestType = FS_REQ_SNAP_SHOW;
	fsReq->hdr.len = sizeof(fs_MsgSnapShowReq);
	fsReq->volId = cb.vol->id;
	fsReq->type = FS_SNAP_SHOW_USAGE;
	
	/* 
	 * Send request to EFS
	 */
	fsRsp = (fs_MsgSnapShowResp *) rspBuf.buf;

	rc = ea_sendClusterAgileMsg(-1, &reqBuf, sizeof(fs_MsgSnapShowReq),
				"efs", NFX_NFP_FP_ANY, 1, MGMTPLANE,
				&waitEfsTo,
				&rspBuf, sizeof(rspBuf.buf),
				cb.hdr->node);

	/* 
	 * Receive response from FS.
	 */
	if ((rc <= 0) || (fsRsp->hdr.fsStatus != NFX_OK))
	{
		eaRsp->result = NFX_ERR;
    	sprintf(eaRsp->output, "EFS error.");
		eaRsp->usageList.fsysId.volId = cb.vol->id;
    	ea_sendMsg(eaRsp, sizeof(ea_getUsgSnapRsp_t), sender->name);
       	return;
	}

	// lc: does ea care about internal snapshots ??

	for (i = 0; i < fsRsp->nUsers; i++)
	{
		ea_snapUsg_t *usageEntry = NULL;
		char		 *timeStr;

		/*
		 * put this entry in EA response
		 */

		EA_DBG_PRINT(("ea_processGetUsgSnapReq got rsp for %s\n", 
						fsRsp->user[i].name));

		usageEntry = &eaRsp->usageList.usage[eaRsp->usageList.num];
		strcpy(usageEntry->name, fsRsp->user[i].name);
		usageEntry->pct= (fsRsp->user[i].size * 100) / 
							(fsRsp->user[i].size + fsRsp->totalBlks);
		usageEntry->blks = fsRsp->user[i].size;
		timeStr = ctime((time_t *) &fsRsp->user[i].ctime.secs);
		strcpy(usageEntry->date, timeStr);

		eaRsp->usageList.num++;
	}

	eaRsp->result = NFX_OK;
    sprintf(eaRsp->output, "Done.");
	eaRsp->usageList.fsysId.volId = cb.vol->id;

	if (eaRsp->usageList.num <= 0)
	{
		eaRspLen = sizeof(ea_getUsgSnapRsp_t);
	}
	else
	{
		eaRspLen = (sizeof(ea_getUsgSnapRsp_t) + 
					(eaRsp->usageList.num - 1)*sizeof(ea_snapSch_t));
	}

    ea_sendMsg(eaRsp, eaRspLen, sender->name);

	EA_FCT_PRINT(("< ea_processGetUsgSnapReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processRaidGetListReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processRaidGetListReq(ea_raid_getListReqMsg_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32               	rc = NFX_OK;
	uint32					idx, eaRspLen, rcmReqLen;
    ea_raid_getListRspMsg_t *eaRsp;
	raid_getListReqMsg_t 	*rcmReq;
	ea_msg_t      			rcmMsg;
	raid_getListRspMsg_t 	*rcmRsp = (raid_getListRspMsg_t *) rcmMsg.buf;
    
    EA_FCT_PRINT(("> ea_processRaidGetListReq()\n"));

    /* 
	 * Initialize the response for AV
     */
	eaRspLen = sizeof(ea_raid_getListRspMsg_t); 
	eaRsp = (ea_raid_getListRspMsg_t *) malloc(eaRspLen);
    bzero((void *) eaRsp, eaRspLen);
    eaRsp->hdr.msgType = EA_RAID_MID_GET_LIST_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));

    /* 
	 * Create RCM request
     */
	rcmReqLen = sizeof(raid_getListReqMsg_t);
	rcmReq = (raid_getListReqMsg_t *) malloc (rcmReqLen);
	rcmReq->header.mid = RAID_MID_GET_LIST_REQ;
	memcpy(rcmReq->header.id, evReq->id, RAID_CTRL_ID_LEN);

	/* 
	 * Send request to RCM
	 */
	rc = sendAgileMsg(-1, (void *) rcmReq, rcmReqLen,
					"raid-config", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					&waitRcmTo, rcmRsp, sizeof(rcmMsg.buf));

	/* 
	 * Receive response from RCM
	 */
	if (rc <= 0)
	{
		/* 
		 * No response/Timeout 
		 */
		rcmRsp->rCode = NFX_ERR;
	}

	/*
	 * Copy back response for AV
	 */
	if (rcmRsp->numEntries > 1)
	{
		eaRspLen = sizeof(ea_raid_getListRspMsg_t) +
					(rcmRsp->numEntries - 1)*sizeof(ea_raid_ctrl_id_t); 
		eaRsp = (ea_raid_getListRspMsg_t *) realloc(eaRsp, eaRspLen);
    	bzero((void *) eaRsp, eaRspLen);
    	eaRsp->hdr.msgType = EA_RAID_MID_GET_LIST_RSP;
    	eaRsp->hdr.userInfo = evReq->hdr.userInfo;
		memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));
	}

	eaRsp->rCode = rcmRsp->rCode;
	eaRsp->numEntries = rcmRsp->numEntries;
	for (idx = 0; idx < rcmRsp->numEntries; idx++)
	{
		memcpy(eaRsp->idList[idx], rcmRsp->idList[idx], 
					RAID_CTRL_ID_LEN);
	}

	/*
	 * Send response to AV
	 */
    ea_sendMsg(eaRsp, eaRspLen, sender->name);

	free(rcmReq);
	free(eaRsp);

	EA_FCT_PRINT(("< ea_processRaidGetListReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processRaidOpenReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processRaidOpenReq(ea_raid_openReqMsg_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32               	rc = NFX_OK;
	uint32					eaRspLen, rcmReqLen;
    ea_raid_openRspMsg_t 	*eaRsp;
	raid_openReqMsg_t 		*rcmReq;
	ea_msg_t      			rcmMsg;
	raid_openRspMsg_t 		*rcmRsp = (raid_openRspMsg_t *) rcmMsg.buf;
    
    EA_FCT_PRINT(("> ea_processRaidOpenReq()\n"));

    /* 
	 * Initialize the response for AV
     */
	eaRspLen = sizeof(ea_raid_openRspMsg_t); 
	eaRsp = (ea_raid_openRspMsg_t *) malloc(eaRspLen);
    bzero((void *) eaRsp, eaRspLen);
    eaRsp->hdr.msgType = EA_RAID_MID_OPEN_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));

    /* 
	 * Create RCM request
     */
	rcmReqLen = sizeof(raid_openReqMsg_t);
	rcmReq = (raid_openReqMsg_t *) malloc (rcmReqLen);
	rcmReq->header.mid = RAID_MID_OPEN_REQ;
	memcpy(rcmReq->header.id, evReq->id, RAID_CTRL_ID_LEN);

	/* 
	 * Send request to RCM
	 */
	rc = sendAgileMsg(-1, (void *) rcmReq, rcmReqLen,
					"raid-config", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					&waitRcmTo, rcmRsp, sizeof(rcmMsg.buf));

	/* 
	 * Receive response from RCM
	 */
	if (rc <= 0)
	{
		/* 
		 * No response/Timeout 
		 */
		rcmRsp->rCode = NFX_ERR;
	}

	/*
	 * Copy back response for AV
	 */
	eaRsp->rCode = rcmRsp->rCode;
	eaRsp->lun = rcmRsp->lun;
	eaRsp->scsi_id = rcmRsp->scsi_id;
	eaRsp->wwn = htonll(rcmRsp->wwn);

	/*
	 * Send response to AV
	 */
    ea_sendMsg(eaRsp, eaRspLen, sender->name);

	free(rcmReq);
	free(eaRsp);

	EA_FCT_PRINT(("< ea_processRaidOpenReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processRaidCloseReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processRaidCloseReq(ea_raid_closeReqMsg_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32               	rc = NFX_OK;
	uint32					eaRspLen, rcmReqLen;
    ea_raid_closeRspMsg_t 	*eaRsp;
	raid_closeReqMsg_t 		*rcmReq;
	ea_msg_t      			rcmMsg;
	raid_closeRspMsg_t 		*rcmRsp = (raid_closeRspMsg_t *) rcmMsg.buf;
    
    EA_FCT_PRINT(("> ea_processRaidCloseReq()\n"));

    /* 
	 * Initialize the response for AV
     */
	eaRspLen = sizeof(ea_raid_closeRspMsg_t); 
	eaRsp = (ea_raid_closeRspMsg_t *) malloc(eaRspLen);
    bzero((void *) eaRsp, eaRspLen);
    eaRsp->hdr.msgType = EA_RAID_MID_CLOSE_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));

    /* 
	 * Create RCM request
     */
	rcmReqLen = sizeof(raid_closeReqMsg_t);
	rcmReq = (raid_closeReqMsg_t *) malloc (rcmReqLen);
	rcmReq->header.mid = RAID_MID_CLOSE_REQ;
	memcpy(rcmReq->header.id, evReq->id, RAID_CTRL_ID_LEN);

	/* 
	 * Send request to RCM
	 */
	rc = sendAgileMsg(-1, (void *) rcmReq, rcmReqLen,
					"raid-config", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					&waitRcmTo, rcmRsp, sizeof(rcmMsg.buf));

	/* 
	 * Receive response from RCM
	 */
	if (rc <= 0)
	{
		/* 
		 * No response/Timeout 
		 */
		rcmRsp->rCode = NFX_ERR;
	}

	/*
	 * Copy back response for AV
	 */
	eaRsp->rCode = rcmRsp->rCode;

	/*
	 * Send response to AV
	 */
    ea_sendMsg(eaRsp, eaRspLen, sender->name);

	free(rcmReq);
	free(eaRsp);

	EA_FCT_PRINT(("< ea_processRaidCloseReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processRaidCtrlInfoReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processRaidCtrlInfoReq(ea_raid_ctrlInfoReqMsg_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32               		rc = NFX_OK;
	uint32						eaRspLen, rcmReqLen;
    ea_raid_ctrlInfoRspMsg_t 	*eaRsp;
	raid_ctrlInfoReqMsg_t 		*rcmReq;
	ea_msg_t      				rcmMsg;
	raid_ctrlInfoRspMsg_t 		*rcmRsp = (raid_ctrlInfoRspMsg_t *) rcmMsg.buf;
    
    EA_FCT_PRINT(("> ea_processRaidCtrlInfoReq()\n"));

    /* 
	 * Initialize the response for AV
     */
	eaRspLen = sizeof(ea_raid_ctrlInfoRspMsg_t); 
	eaRsp = (ea_raid_ctrlInfoRspMsg_t *) malloc(eaRspLen);
    bzero((void *) eaRsp, eaRspLen);
    eaRsp->hdr.msgType = EA_RAID_MID_CTRL_INFO_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));

    /* 
	 * Create RCM request
     */
	rcmReqLen = sizeof(raid_ctrlInfoReqMsg_t);
	rcmReq = (raid_ctrlInfoReqMsg_t *) malloc (rcmReqLen);
	rcmReq->header.mid = RAID_MID_CTRL_INFO_REQ;
	memcpy(rcmReq->header.id, evReq->id, RAID_CTRL_ID_LEN+1);
	rcmReq->header.cookie = 1;

	/* 
	 * Send request to RCM
	 */
	rc = sendAgileMsg(-1, (void *) rcmReq, rcmReqLen,
					"raid-config", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					&waitRcmTo, rcmRsp, sizeof(rcmMsg.buf));

	/* 
	 * Receive response from RCM
	 */
	if (rc <= 0)
	{
		/* 
		 * No response/Timeout 
		 */
		rcmRsp->rCode = NFX_ERR;
	}

	/*
	 * Copy back response for AV
	 */
	eaRsp->rCode = rcmRsp->rCode;
	eaRsp->bbuPresent = rcmRsp->bbuPresent;
	memcpy(eaRsp->ctrlType, rcmRsp->ctrlType, RAID_CTRL_TYPE_LEN + 1);
	memcpy(eaRsp->ctrlName, rcmRsp->ctrlName, RAID_CTRL_NAME_LEN + 1);
	memcpy(eaRsp->serialNumber, rcmRsp->serialNumber, RAID_CTRL_SERIAL_LEN + 1);
	memcpy(eaRsp->vendorName, rcmRsp->vendorName, RAID_CTRL_VENDOR_LEN + 1);
	memcpy(eaRsp->hardwareVer, rcmRsp->hardwareVer, RAID_CTRL_HWVER_LEN + 1);
	memcpy(eaRsp->firmwareVer, rcmRsp->firmwareVer, RAID_CTRL_FWVER_LEN + 1);
	eaRsp->cacheInMB = rcmRsp->cacheInMB;
	eaRsp->memoryInMB = rcmRsp->memoryInMB;
	eaRsp->numDriveChannels = rcmRsp->numDriveChannels;
	eaRsp->maxDrivesPerChannel = rcmRsp->maxDrivesPerChannel;
	eaRsp->ctrlWwn = htonll(rcmRsp->ctrlWwn);
	eaRsp->dualCtrlWwn = htonll(rcmRsp->dualCtrlWwn);
	eaRsp->numLogicalDevices = rcmRsp->numLogicalDevices;
	eaRsp->numPhysicalDevices = rcmRsp->numPhysicalDevices;

	/*
	 * Send response to AV
	 */
    ea_sendMsg(eaRsp, eaRspLen, sender->name);

	free(rcmReq);
	free(eaRsp);

	EA_FCT_PRINT(("< ea_processRaidCtrlInfoReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processRaidCtrlParamReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processRaidCtrlParamReq(ea_raid_getCtrlParamReqMsg_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32               			rc = NFX_OK;
	uint32							eaRspLen, rcmReqLen;
    ea_raid_getCtrlParamRspMsg_t	*eaRsp;
	raid_getCtrlParamReqMsg_t 		*rcmReq;
	ea_msg_t      					rcmMsg;
	raid_getCtrlParamRspMsg_t 		*rcmRsp = (raid_getCtrlParamRspMsg_t *) 
															rcmMsg.buf;
    
    EA_FCT_PRINT(("> ea_processRaidCtrlParamReq()\n"));

    /* 
	 * Initialize the response for AV
     */
	eaRspLen = sizeof(ea_raid_getCtrlParamRspMsg_t); 
	eaRsp = (ea_raid_getCtrlParamRspMsg_t *) malloc(eaRspLen);
    bzero((void *) eaRsp, eaRspLen);
    eaRsp->hdr.msgType = EA_RAID_MID_CTRL_GETPARAM_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));

    /* 
	 * Create RCM request
     */
	rcmReqLen = sizeof(raid_getCtrlParamReqMsg_t);
	rcmReq = (raid_getCtrlParamReqMsg_t *) malloc (rcmReqLen);
	rcmReq->header.mid = RAID_MID_CTRL_GETPARAM_REQ;
	memcpy(rcmReq->header.id, evReq->id, RAID_CTRL_ID_LEN);

	/* 
	 * Send request to RCM
	 */
	rc = sendAgileMsg(-1, (void *) rcmReq, rcmReqLen,
					"raid-config", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					&waitRcmTo, rcmRsp, sizeof(rcmMsg.buf));

	/* 
	 * Receive response from RCM
	 */
	if (rc <= 0)
	{
		/* 
		 * No response/Timeout 
		 */
		rcmRsp->rCode = NFX_ERR;
	}

	/*
	 * Copy back response for AV
	 */
	eaRsp->rCode = rcmRsp->rCode;
	eaRsp->params.readAhead = rcmRsp->params.readAhead;
	eaRsp->params.reasRestrict = rcmRsp->params.reasRestrict;
	eaRsp->params.trueVerify = rcmRsp->params.trueVerify;
	eaRsp->params.writeVerify = rcmRsp->params.writeVerify;
	eaRsp->params.bgInit = rcmRsp->params.bgInit;
	eaRsp->params.superReadAhead = rcmRsp->params.superReadAhead;
	eaRsp->params.coalOptimal = rcmRsp->params.coalOptimal;
	eaRsp->params.autoRebuild = rcmRsp->params.autoRebuild;
	eaRsp->params.operFault = rcmRsp->params.operFault;
	eaRsp->params.vendorTUR = rcmRsp->params.vendorTUR;
	eaRsp->params.disCCInvLun = rcmRsp->params.disCCInvLun;
	eaRsp->params.noPause = rcmRsp->params.noPause;
	eaRsp->params.queueFullBusy = rcmRsp->params.queueFullBusy;
	eaRsp->params.disBusyFailback = rcmRsp->params.disBusyFailback;
	eaRsp->params.nodeNameRet = rcmRsp->params.nodeNameRet;
	eaRsp->params.safteofups = rcmRsp->params.safteofups;
	eaRsp->params.resetProp = rcmRsp->params.resetProp;
	eaRsp->params.smartLrgTrans = rcmRsp->params.smartLrgTrans;
	eaRsp->params.autoFailback = rcmRsp->params.autoFailback;
	eaRsp->params.forceSimplex = rcmRsp->params.forceSimplex;
	eaRsp->params.conserveCache = rcmRsp->params.conserveCache;
	eaRsp->params.noRSTCOM = rcmRsp->params.noRSTCOM;
	eaRsp->params.ctrl0Port0 = rcmRsp->params.ctrl0Port0;
	eaRsp->params.ctrl0Port1 = rcmRsp->params.ctrl0Port1;
	eaRsp->params.ctrl1Port0 = rcmRsp->params.ctrl1Port0;
	eaRsp->params.ctrl1Port1 = rcmRsp->params.ctrl1Port1;
	eaRsp->params.ctrl0Port0LoopId = rcmRsp->params.ctrl0Port0LoopId;
	eaRsp->params.ctrl0Port1LoopId = rcmRsp->params.ctrl0Port1LoopId;
	eaRsp->params.ctrl1Port0LoopId = rcmRsp->params.ctrl1Port0LoopId;
	eaRsp->params.ctrl1Port1LoopId = rcmRsp->params.ctrl1Port1LoopId;
	eaRsp->params.debugDump = rcmRsp->params.debugDump;
	eaRsp->params.rebuildRate = rcmRsp->params.rebuildRate;
	eaRsp->params.initRate = rcmRsp->params.initRate;
	eaRsp->params.consistencyCheckRate = rcmRsp->params.consistencyCheckRate;
	eaRsp->params.startupMode = rcmRsp->params.startupMode;
	eaRsp->params.startupNumDev = rcmRsp->params.startupNumDev;
	eaRsp->params.startupDelay = rcmRsp->params.startupDelay;
	eaRsp->params.scsiStartupDelay = rcmRsp->params.scsiStartupDelay;
	eaRsp->params.topology = rcmRsp->params.topology;
	eaRsp->params.rs232PortType = rcmRsp->params.rs232PortType;
	eaRsp->params.rs232BaudRate = rcmRsp->params.rs232BaudRate;
	eaRsp->params.frameSize = rcmRsp->params.frameSize;
	eaRsp->params.pciLatCtrl = rcmRsp->params.pciLatCtrl;
	eaRsp->params.queueLimit = rcmRsp->params.queueLimit;
	eaRsp->params.rofInterval = rcmRsp->params.rofInterval;
	eaRsp->params.rofReboot = rcmRsp->params.rofReboot;

	/*
	 * Send response to AV
	 */
    ea_sendMsg(eaRsp, eaRspLen, sender->name);

	free(rcmReq);
	free(eaRsp);

	EA_FCT_PRINT(("< ea_processRaidCtrlParamReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processRaidPhyDevListReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processRaidPhyDevListReq(ea_raid_phyDevListReqMsg_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32               		rc = NFX_OK;
	uint32						idx, eaRspLen, rcmReqLen;
    ea_raid_phyDevListRspMsg_t 	*eaRsp;
	raid_phyDevListReqMsg_t 	*rcmReq;
	ea_msg_t      				rcmMsg;
	raid_phyDevListRspMsg_t 	*rcmRsp = (raid_phyDevListRspMsg_t *) 
											rcmMsg.buf;
    
    EA_FCT_PRINT(("> ea_processRaidPhyDevListReq()\n"));

    /* 
	 * Initialize the response for AV
     */
	eaRspLen = sizeof(ea_raid_phyDevListRspMsg_t); 
	eaRsp = (ea_raid_phyDevListRspMsg_t *) malloc(eaRspLen);
    bzero((void *) eaRsp, eaRspLen);
    eaRsp->hdr.msgType = EA_RAID_MID_PHYDEV_LIST_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));

    /* 
	 * Create RCM request
     */
	rcmReqLen = sizeof(raid_phyDevListReqMsg_t);
	rcmReq = (raid_phyDevListReqMsg_t *) malloc (rcmReqLen);
	rcmReq->header.mid = RAID_MID_PHYDEV_LIST_REQ;
	memcpy(rcmReq->header.id, evReq->id, RAID_CTRL_ID_LEN);

	/* 
	 * Send request to RCM
	 */
	rc = sendAgileMsg(-1, (void *) rcmReq, rcmReqLen,
					"raid-config", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					&waitRcmTo, rcmRsp, sizeof(rcmMsg.buf));

	/* 
	 * Receive response from RCM
	 */
	if (rc <= 0)
	{
		/* 
		 * No response/Timeout 
		 */
		rcmRsp->rCode = NFX_ERR;
	}

	/*
	 * Copy back response for AV
	 */
	if (rcmRsp->numEntries > 1)
	{
		eaRspLen = sizeof(ea_raid_phyDevListRspMsg_t) +
					(rcmRsp->numEntries - 1)*sizeof(ea_raid_phydev_id_t); 
		eaRsp = (ea_raid_phyDevListRspMsg_t *) realloc(eaRsp, eaRspLen);
    	bzero((void *) eaRsp, eaRspLen);
    	eaRsp->hdr.msgType = EA_RAID_MID_PHYDEV_LIST_RSP;
    	eaRsp->hdr.userInfo = evReq->hdr.userInfo;
		memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));
	}

	eaRsp->rCode = rcmRsp->rCode;
	eaRsp->numEntries = rcmRsp->numEntries;
	for (idx = 0; idx < rcmRsp->numEntries; idx++)
	{
		eaRsp->phyDevList[idx].channelNumber =
						rcmRsp->idList[idx].channelNumber;
		eaRsp->phyDevList[idx].targetId =
						rcmRsp->idList[idx].targetId;
	}

	/*
	 * Send response to AV
	 */
    ea_sendMsg(eaRsp, eaRspLen, sender->name);

	free(rcmReq);
	free(eaRsp);

	EA_FCT_PRINT(("< ea_processRaidPhyDevListReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processRaidPhyDevInfoReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processRaidPhyDevInfoReq(ea_raid_phyDevInfoReqMsg_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32               		rc = NFX_OK;
	uint32						eaRspLen, rcmReqLen;
    ea_raid_phyDevInfoRspMsg_t 	*eaRsp;
	raid_phyDevInfoReqMsg_t 	*rcmReq;
	ea_msg_t      				rcmMsg;
	raid_phyDevInfoRspMsg_t 	*rcmRsp = (raid_phyDevInfoRspMsg_t *) 
											rcmMsg.buf;
    
    EA_FCT_PRINT(("> ea_processRaidPhyDevInfoReq()\n"));

    /* 
	 * Initialize the response for AV
     */
	eaRspLen = sizeof(ea_raid_phyDevInfoRspMsg_t); 
	eaRsp = (ea_raid_phyDevInfoRspMsg_t *) malloc(eaRspLen);
    bzero((void *) eaRsp, eaRspLen);
    eaRsp->hdr.msgType = EA_RAID_MID_PHYDEV_INFO_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));
	eaRsp->pid.channelNumber = evReq->pid.channelNumber;
	eaRsp->pid.targetId = evReq->pid.targetId;

    /* 
	 * Create RCM request
     */
	rcmReqLen = sizeof(raid_phyDevInfoReqMsg_t);
	rcmReq = (raid_phyDevInfoReqMsg_t *) malloc (rcmReqLen);
	rcmReq->header.mid = RAID_MID_PHYDEV_INFO_REQ;
	memcpy(rcmReq->header.id, evReq->id, RAID_CTRL_ID_LEN);
	rcmReq->id.channelNumber = evReq->pid.channelNumber;
	rcmReq->id.targetId = evReq->pid.targetId;

	/* 
	 * Send request to RCM
	 */
	rc = sendAgileMsg(-1, (void *) rcmReq, rcmReqLen,
					"raid-config", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					&waitRcmTo, rcmRsp, sizeof(rcmMsg.buf));

	/* 
	 * Receive response from RCM
	 */
	if (rc <= 0)
	{
		/* 
		 * No response/Timeout 
		 */
		rcmRsp->rCode = NFX_ERR;
	}

	/*
	 * Copy back response for AV
	 */
	eaRsp->rCode = rcmRsp->rCode;
	eaRsp->wwn = htonll(rcmRsp->wwn);
	eaRsp->enclosure = rcmRsp->enclosure;
	eaRsp->row = rcmRsp->row;
	eaRsp->column = rcmRsp->column;
	eaRsp->rdn = rcmRsp->rdn;
	eaRsp->physicalSizeInMB = rcmRsp->physicalSizeInMB;
	eaRsp->usableSizeInMB = rcmRsp->usableSizeInMB;
	eaRsp->freeSizeInMB = rcmRsp->freeSizeInMB;
	eaRsp->sectorSizeInByte = rcmRsp->sectorSizeInByte;
	eaRsp->deviceState = rcmRsp->deviceState;
	eaRsp->portInUse = rcmRsp->portInUse;
	memcpy(eaRsp->vendorName, rcmRsp->vendorName, RAID_PHYDEV_VENDOR_LEN + 1);
	memcpy(eaRsp->modelName, rcmRsp->modelName, RAID_PHYDEV_MODEL_LEN + 1);
	memcpy(eaRsp->serialNumber, rcmRsp->serialNumber, 
												RAID_PHYDEV_SERIAL_LEN + 1);
	memcpy(eaRsp->microcodeVer, rcmRsp->microcodeVer, 
												RAID_PHYDEV_MCVER_LEN + 1);

	/*
	 * Send response to AV
	 */
    ea_sendMsg(eaRsp, eaRspLen, sender->name);

	free(rcmReq);
	free(eaRsp);

	EA_FCT_PRINT(("< ea_processRaidPhyDevInfoReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processRaidLogDevListReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processRaidLogDevListReq(ea_raid_logDevListReqMsg_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32               		rc = NFX_OK;
	uint32						idx, eaRspLen, rcmReqLen;
    ea_raid_logDevListRspMsg_t 	*eaRsp;
	raid_logDevListReqMsg_t 	*rcmReq;
	ea_msg_t      				rcmMsg;
	raid_logDevListRspMsg_t 	*rcmRsp = (raid_logDevListRspMsg_t *) 
											rcmMsg.buf;
    
    EA_FCT_PRINT(("> ea_processRaidLogDevListReq()\n"));

    /* 
	 * Initialize the response for AV
     */
	eaRspLen = sizeof(ea_raid_logDevListRspMsg_t); 
	eaRsp = (ea_raid_logDevListRspMsg_t *) malloc(eaRspLen);
    bzero((void *) eaRsp, eaRspLen);
    eaRsp->hdr.msgType = EA_RAID_MID_LOGDEV_LIST_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));

    /* 
	 * Create RCM request
     */
	rcmReqLen = sizeof(raid_logDevListReqMsg_t);
	rcmReq = (raid_logDevListReqMsg_t *) malloc (rcmReqLen);
	rcmReq->header.mid = RAID_MID_LOGDEV_LIST_REQ;
	memcpy(rcmReq->header.id, evReq->id, RAID_CTRL_ID_LEN);

	/* 
	 * Send request to RCM
	 */
	rc = sendAgileMsg(-1, (void *) rcmReq, rcmReqLen,
					"raid-config", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					&waitRcmTo, rcmRsp, sizeof(rcmMsg.buf));

	/* 
	 * Receive response from RCM
	 */
	if (rc <= 0)
	{
		/* 
		 * No response/Timeout 
		 */
		rcmRsp->rCode = NFX_ERR;
	}

	/*
	 * Copy back response for AV
	 */
	if (rcmRsp->numEntries > 1)
	{
		eaRspLen = sizeof(ea_raid_logDevListRspMsg_t) +
					(rcmRsp->numEntries - 1)*sizeof(ea_raid_logdev_id_t); 
		eaRsp = (ea_raid_logDevListRspMsg_t *) realloc(eaRsp, eaRspLen);
    	bzero((void *) eaRsp, eaRspLen);
    	eaRsp->hdr.msgType = EA_RAID_MID_LOGDEV_LIST_RSP;
    	eaRsp->hdr.userInfo = evReq->hdr.userInfo;
		memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));
	}

	eaRsp->rCode = rcmRsp->rCode;
	eaRsp->numEntries = rcmRsp->numEntries;
	for (idx = 0; idx < rcmRsp->numEntries; idx++)
	{
		eaRsp->logDevIdList[idx].id = rcmRsp->idList[idx];
	}

	/*
	 * Send response to AV
	 */
    ea_sendMsg(eaRsp, eaRspLen, sender->name);

	free(rcmReq);
	free(eaRsp);

	EA_FCT_PRINT(("< ea_processRaidLogDevListReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processRaidLogDevInfoReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processRaidLogDevInfoReq(ea_raid_logDevInfoReqMsg_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32               		rc = NFX_OK;
	uint32						eaRspLen, rcmReqLen;
    ea_raid_logDevInfoRspMsg_t 	*eaRsp;
	raid_logDevInfoReqMsg_t 	*rcmReq;
	ea_msg_t      				rcmMsg;
	raid_logDevInfoRspMsg_t 	*rcmRsp = (raid_logDevInfoRspMsg_t *) 
											rcmMsg.buf;
    
    EA_FCT_PRINT(("> ea_processRaidLogDevInfoReq()\n"));

    /* 
	 * Initialize the response for AV
     */
	eaRspLen = sizeof(ea_raid_logDevInfoRspMsg_t); 
	eaRsp = (ea_raid_logDevInfoRspMsg_t *) malloc(eaRspLen);
    bzero((void *) eaRsp, eaRspLen);
    eaRsp->hdr.msgType = EA_RAID_MID_LOGDEV_INFO_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));
	eaRsp->logDevId.id = evReq->logDevId.id;

    /* 
	 * Create RCM request
     */
	rcmReqLen = sizeof(raid_logDevInfoReqMsg_t);
	rcmReq = (raid_logDevInfoReqMsg_t *) malloc (rcmReqLen);
	rcmReq->header.mid = RAID_MID_LOGDEV_INFO_REQ;
	memcpy(rcmReq->header.id, evReq->id, RAID_CTRL_ID_LEN);
	rcmReq->id = evReq->logDevId.id;

	/* 
	 * Send request to RCM
	 */
	rc = sendAgileMsg(-1, (void *) rcmReq, rcmReqLen,
					"raid-config", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					&waitRcmTo, rcmRsp, sizeof(rcmMsg.buf));

	/* 
	 * Receive response from RCM
	 */
	if (rc <= 0)
	{
		/* 
		 * No response/Timeout 
		 */
		rcmRsp->rCode = NFX_ERR;
	}

	/*
	 * Copy back response for AV
	 */
	eaRsp->rCode = rcmRsp->rCode;
	eaRsp->raidLevel = rcmRsp->raidLevel;
	eaRsp->physicalSizeInMB = rcmRsp->physicalSizeInMB;
	eaRsp->actualSizeInMB = rcmRsp->actualSizeInMB;
	eaRsp->stripeSizeInKB = rcmRsp->stripeSizeInKB;
	eaRsp->cacheLineSizeInKB = rcmRsp->cacheLineSizeInKB;
	eaRsp->deviceState = rcmRsp->deviceState;
	eaRsp->readCacheEnabled = rcmRsp->readCacheEnabled;
	eaRsp->writeCacheEnabled = rcmRsp->writeCacheEnabled;
	eaRsp->hasBgTask = rcmRsp->hasBgTask;

	/*
	 * Send response to AV
	 */
    ea_sendMsg(eaRsp, eaRspLen, sender->name);

	free(rcmReq);
	free(eaRsp);

	EA_FCT_PRINT(("< ea_processRaidLogDevInfoReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processRaidCreateLunReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processRaidCreateLunReq(ea_raid_createLunReqMsg_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32               		rc = NFX_OK;
	uint32						idx, eaRspLen, rcmReqLen;
    ea_raid_createLunRspMsg_t 	*eaRsp;
	raid_createLunReqMsg_t 		*rcmReq;
	ea_msg_t      				rcmMsg;
	raid_createLunRspMsg_t 		*rcmRsp = (raid_createLunRspMsg_t *) 
											rcmMsg.buf;
    
    EA_FCT_PRINT(("> ea_processRaidCreateLunReq()\n"));

    /* 
	 * Initialize the response for AV
     */
	eaRspLen = sizeof(ea_raid_createLunRspMsg_t);
	eaRsp = (ea_raid_createLunRspMsg_t *) malloc(eaRspLen);
    bzero((void *) eaRsp, eaRspLen);
    eaRsp->hdr.msgType = EA_RAID_MID_LUN_CREATE_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));

    /* 
	 * Create RCM request
     */
	rcmReqLen = sizeof(raid_createLunReqMsg_t) +
			(evReq->lundef.numOfPhyDevices - 1)*sizeof(ea_raid_phydev_id_t);
	rcmReq = (raid_createLunReqMsg_t *) malloc (rcmReqLen);
	rcmReq->header.mid = RAID_MID_LUN_CREATE_REQ;
	memcpy(rcmReq->header.id, evReq->id, RAID_CTRL_ID_LEN);

	rcmReq->lundef.raidLevel = evReq->lundef.raidLevel;
	rcmReq->lundef.sizeInMB = evReq->lundef.sizeInMB;
	rcmReq->lundef.stripeSizeInKB = evReq->lundef.stripeSizeInKB;
	rcmReq->lundef.cacheEnabled = evReq->lundef.cacheEnabled;
	rcmReq->lundef.autoInit = evReq->lundef.autoInit;
	rcmReq->lundef.numOfPhyDevices = evReq->lundef.numOfPhyDevices;

	for (idx = 0; idx < evReq->lundef.numOfPhyDevices; idx++)
	{
		rcmReq->lundef.phyDevList[idx].channelNumber = 
				evReq->lundef.phyDevList[idx].channelNumber;
		rcmReq->lundef.phyDevList[idx].targetId = 
				evReq->lundef.phyDevList[idx].targetId;
	}

	/* 
	 * Send request to RCM
	 */
	rc = sendAgileMsg(-1, (void *) rcmReq, rcmReqLen,
					"raid-config", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					&waitRcmTo, rcmRsp, sizeof(rcmMsg.buf));

	/* 
	 * Receive response from RCM
	 */
	if (rc <= 0)
	{
		/* 
		 * No response/Timeout 
		 */
		rcmRsp->rCode = NFX_ERR;
	}

	/*
	 * Copy back response for AV
	 */
	eaRsp->rCode = rcmRsp->rCode;
	eaRsp->logDevId.id = rcmRsp->id;

	/*
	 * Send response to AV
	 */
    ea_sendMsg(eaRsp, eaRspLen, sender->name);

	free(rcmReq);
	free(eaRsp);

	EA_FCT_PRINT(("< ea_processRaidCreateLunReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processRaidCreateSimpleLunReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processRaidCreateSimpleLunReq(ea_raid_createSimpleLunReqMsg_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32               				rc = NFX_OK;
	uint32								eaRspLen, rcmReqLen;
    ea_raid_createSimpleLunRspMsg_t 	*eaRsp;
	raid_createSimpleLunReqMsg_t 		*rcmReq;
	ea_msg_t      						rcmMsg;
	raid_createSimpleLunRspMsg_t 		*rcmRsp = 
											(raid_createSimpleLunRspMsg_t *) 
													rcmMsg.buf;
    
    EA_FCT_PRINT(("> ea_processRaidCreateSimpleLunReq()\n"));

    /* 
	 * Initialize the response for AV
     */
	eaRspLen = sizeof(ea_raid_createSimpleLunRspMsg_t);
	eaRsp = (ea_raid_createSimpleLunRspMsg_t *) malloc(eaRspLen);
    bzero((void *) eaRsp, eaRspLen);
    eaRsp->hdr.msgType = EA_RAID_MID_LUN_CREATE_SIMPLE_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));

    /* 
	 * Create RCM request
     */
	rcmReqLen = sizeof(raid_createSimpleLunReqMsg_t);
	rcmReq = (raid_createSimpleLunReqMsg_t *) malloc (rcmReqLen);
	rcmReq->header.mid = RAID_MID_LUN_CREATE_SIMPLE_REQ;
	memcpy(rcmReq->header.id, evReq->id, RAID_CTRL_ID_LEN);

	rcmReq->lunCount = evReq->lunCount;
	rcmReq->raidLevel = evReq->raidLevel;
	rcmReq->sizeInMB = evReq->sizeInMB;
	rcmReq->stripeSizeInKB = evReq->stripeSizeInKB;

	/* 
	 * Send request to RCM
	 */
	rc = sendAgileMsg(-1, (void *) rcmReq, rcmReqLen,
					"raid-config", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					&waitRcmTo, rcmRsp, sizeof(rcmMsg.buf));

	/* 
	 * Receive response from RCM
	 */
	if (rc <= 0)
	{
		/* 
		 * No response/Timeout 
		 */
		rcmRsp->rCode = NFX_ERR;
	}

	/*
	 * Copy back response for AV
	 */
	eaRsp->rCode = rcmRsp->rCode;

	/*
	 * Send response to AV
	 */
    ea_sendMsg(eaRsp, eaRspLen, sender->name);

	free(rcmReq);
	free(eaRsp);

	EA_FCT_PRINT(("< ea_processRaidCreateSimpleLunReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processRaidDeleteLunReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processRaidDeleteLunReq(ea_raid_deleteLunReqMsg_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32               		rc = NFX_OK;
	uint32						eaRspLen, rcmReqLen;
    ea_raid_deleteLunRspMsg_t 	*eaRsp;
	raid_deleteLunReqMsg_t 		*rcmReq;
	ea_msg_t      				rcmMsg;
	raid_deleteLunRspMsg_t 		*rcmRsp = (raid_deleteLunRspMsg_t *) 
											rcmMsg.buf;
    
    EA_FCT_PRINT(("> ea_processRaidDeleteLunReq()\n"));

    /* 
	 * Initialize the response for AV
     */
	eaRspLen = sizeof(ea_raid_deleteLunRspMsg_t);
	eaRsp = (ea_raid_deleteLunRspMsg_t *) malloc(eaRspLen);
    bzero((void *) eaRsp, eaRspLen);
    eaRsp->hdr.msgType = EA_RAID_MID_LUN_DELETE_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));

    /* 
	 * Create RCM request
     */
	rcmReqLen = sizeof(raid_deleteLunReqMsg_t);
	rcmReq = (raid_deleteLunReqMsg_t *) malloc (rcmReqLen);
	rcmReq->header.mid = RAID_MID_LUN_DELETE_REQ;
	memcpy(rcmReq->header.id, evReq->id, RAID_CTRL_ID_LEN);
	rcmReq->id = evReq->logDevId.id;

	/* 
	 * Send request to RCM
	 */
	rc = sendAgileMsg(-1, (void *) rcmReq, rcmReqLen,
					"raid-config", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					&waitRcmTo, rcmRsp, sizeof(rcmMsg.buf));

	/* 
	 * Receive response from RCM
	 */
	if (rc <= 0)
	{
		/* 
		 * No response/Timeout 
		 */
		rcmRsp->rCode = NFX_ERR;
	}

	/*
	 * Copy back response for AV
	 */
	eaRsp->rCode = rcmRsp->rCode;

	/*
	 * Send response to AV
	 */
    ea_sendMsg(eaRsp, eaRspLen, sender->name);

	free(rcmReq);
	free(eaRsp);

	EA_FCT_PRINT(("< ea_processRaidDeleteLunReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processRaidInitLunReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processRaidInitLunReq(ea_raid_initLunReqMsg_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32               		rc = NFX_OK;
	uint32						eaRspLen, rcmReqLen;
    ea_raid_initLunRspMsg_t 	*eaRsp;
	raid_initLunReqMsg_t 		*rcmReq;
	ea_msg_t      				rcmMsg;
	raid_initLunRspMsg_t 		*rcmRsp = (raid_initLunRspMsg_t *) rcmMsg.buf;
    
    EA_FCT_PRINT(("> ea_processRaidInitLunReq()\n"));

    /* 
	 * Initialize the response for AV
     */
	eaRspLen = sizeof(ea_raid_initLunRspMsg_t);
	eaRsp = (ea_raid_initLunRspMsg_t *) malloc(eaRspLen);
    bzero((void *) eaRsp, eaRspLen);
    eaRsp->hdr.msgType = EA_RAID_MID_LUN_INIT_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));

    /* 
	 * Create RCM request
     */
	rcmReqLen = sizeof(raid_initLunReqMsg_t);
	rcmReq = (raid_initLunReqMsg_t *) malloc (rcmReqLen);
	rcmReq->header.mid = RAID_MID_LUN_INIT_REQ;
	memcpy(rcmReq->header.id, evReq->id, RAID_CTRL_ID_LEN);

	rcmReq->id = evReq->logDevId.id;
	rcmReq->type = evReq->type;

	/* 
	 * Send request to RCM
	 */
	rc = sendAgileMsg(-1, (void *) rcmReq, rcmReqLen,
					"raid-config", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					&waitRcmTo, rcmRsp, sizeof(rcmMsg.buf));

	/* 
	 * Receive response from RCM
	 */
	if (rc <= 0)
	{
		/* 
		 * No response/Timeout 
		 */
		rcmRsp->rCode = NFX_ERR;
	}

	/*
	 * Copy back response for AV
	 */
	eaRsp->rCode = rcmRsp->rCode;
	eaRsp->init_started_count = rcmRsp->init_started_count;

	/*
	 * Send response to AV
	 */
    ea_sendMsg(eaRsp, eaRspLen, sender->name);

	free(rcmReq);
	free(eaRsp);

	EA_FCT_PRINT(("< ea_processRaidInitLunReq()\n"));
}
/*------------------------------------------------------------------------------
 * Name         : ea_processRaidGetLunMaskReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processRaidGetLunMaskReq(ea_raid_getLunMaskReqMsg_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32               		rc = NFX_OK;
	uint32						idx, eaRspLen, rcmReqLen;
    ea_raid_getLunMaskRspMsg_t 	*eaRsp;
	raid_getLunMaskReqMsg_t 	*rcmReq;
	ea_msg_t      				rcmMsg;
	raid_getLunMaskRspMsg_t 	*rcmRsp = (raid_getLunMaskRspMsg_t *) 
													rcmMsg.buf;
    
    EA_FCT_PRINT(("> ea_processRaidGetLunMaskReq()\n"));

    /* 
	 * Initialize the response for AV
     */
	eaRspLen = sizeof(ea_raid_getLunMaskRspMsg_t); 
	eaRsp = (ea_raid_getLunMaskRspMsg_t *) malloc(eaRspLen);
    bzero((void *) eaRsp, eaRspLen);
    eaRsp->hdr.msgType = EA_RAID_MID_LUNMASK_INFO_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));
	eaRsp->logDevId.id = evReq->logDevId.id;
	eaRsp->ctrlNumber = evReq->ctrlNumber;
	eaRsp->portNumber = evReq->portNumber;
	

    /* 
	 * Create RCM request
     */
	rcmReqLen = sizeof(raid_getLunMaskReqMsg_t);
	rcmReq = (raid_getLunMaskReqMsg_t *) malloc (rcmReqLen);
	rcmReq->header.mid = RAID_MID_LUNMASK_INFO_REQ;
	memcpy(rcmReq->header.id, evReq->id, RAID_CTRL_ID_LEN);
	rcmReq->id = evReq->logDevId.id;
	rcmReq->ctrlNumber = evReq->ctrlNumber;
	rcmReq->portNumber = evReq->portNumber;

	/* 
	 * Send request to RCM
	 */
	rc = sendAgileMsg(-1, (void *) rcmReq, rcmReqLen,
					"raid-config", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					&waitRcmTo, rcmRsp, sizeof(rcmMsg.buf));

	/* 
	 * Receive response from RCM
	 */
	if (rc <= 0)
	{
		/* 
		 * No response/Timeout 
		 */
		rcmRsp->rCode = NFX_ERR;
	}

	/*
	 * Copy back response for AV
	 */
	if (rcmRsp->numOfHostWwns > 1)
	{
		eaRspLen = sizeof(ea_raid_getLunMaskRspMsg_t) +
					(rcmRsp->numOfHostWwns - 1)*sizeof(uint64); 
		eaRsp = (ea_raid_getLunMaskRspMsg_t *) realloc(eaRsp, eaRspLen);
    	bzero((void *) eaRsp, eaRspLen);
    	eaRsp->hdr.msgType = EA_RAID_MID_LUNMASK_INFO_RSP;
    	eaRsp->hdr.userInfo = evReq->hdr.userInfo;
		memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));
		eaRsp->logDevId.id = evReq->logDevId.id;
		eaRsp->ctrlNumber = evReq->ctrlNumber;
		eaRsp->portNumber = evReq->portNumber;
	}

	eaRsp->rCode = rcmRsp->rCode;

	eaRsp->numOfHostWwns = rcmRsp->numOfHostWwns;
	for (idx = 0; idx < rcmRsp->numOfHostWwns; idx++)
	{
		eaRsp->hostWwnList[idx] = htonll(rcmRsp->hostWwnList[idx]);
	}

	/*
	 * Send response to AV
	 */
    ea_sendMsg(eaRsp, eaRspLen, sender->name);

	free(rcmReq);
	free(eaRsp);

	EA_FCT_PRINT(("< ea_processRaidGetLunMaskReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processRaidAddLunMaskReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processRaidAddLunMaskReq(ea_raid_addLunMaskReqMsg_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32               		rc = NFX_OK;
	uint32						idx, eaRspLen, rcmReqLen;
    ea_raid_addLunMaskRspMsg_t 	*eaRsp;
	raid_addLunMaskReqMsg_t 	*rcmReq;
	ea_msg_t      				rcmMsg;
	raid_addLunMaskRspMsg_t 	*rcmRsp = (raid_addLunMaskRspMsg_t *) 
											rcmMsg.buf;
    
    EA_FCT_PRINT(("> ea_processRaidAddLunMaskReq()\n"));

    /* 
	 * Initialize the response for AV
     */
	eaRspLen = sizeof(ea_raid_addLunMaskRspMsg_t);
	eaRsp = (ea_raid_addLunMaskRspMsg_t *) malloc(eaRspLen);
    bzero((void *) eaRsp, eaRspLen);
    eaRsp->hdr.msgType = EA_RAID_MID_LUNMASK_ADD_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));
	eaRsp->logDevId.id = evReq->logDevId.id;
	eaRsp->ctrlNumber = evReq->ctrlNumber;
	eaRsp->portNumber = evReq->portNumber;

    /* 
	 * Create RCM request
     */
	rcmReqLen = sizeof(raid_addLunMaskReqMsg_t) +
			(evReq->numOfHostWwns - 1)*sizeof(uint64);
	rcmReq = (raid_addLunMaskReqMsg_t *) malloc (rcmReqLen);
	rcmReq->header.mid = RAID_MID_LUNMASK_ADD_REQ;
	memcpy(rcmReq->header.id, evReq->id, RAID_CTRL_ID_LEN);
	rcmReq->id = evReq->logDevId.id;
	rcmReq->ctrlNumber = evReq->ctrlNumber;
	rcmReq->portNumber = evReq->portNumber;

	rcmReq->numOfHostWwns = evReq->numOfHostWwns;
	for (idx = 0; idx < evReq->numOfHostWwns; idx++)
	{
		rcmReq->hostWwnList[idx] = evReq->hostWwnList[idx];
	}

	/* 
	 * Send request to RCM
	 */
	rc = sendAgileMsg(-1, (void *) rcmReq, rcmReqLen,
					"raid-config", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					&waitRcmTo, rcmRsp, sizeof(rcmMsg.buf));

	/* 
	 * Receive response from RCM
	 */
	if (rc <= 0)
	{
		/* 
		 * No response/Timeout 
		 */
		rcmRsp->rCode = NFX_ERR;
	}

	/*
	 * Copy back response for AV
	 */
	eaRsp->rCode = rcmRsp->rCode;

	/*
	 * Send response to AV
	 */
    ea_sendMsg(eaRsp, eaRspLen, sender->name);

	free(rcmReq);
	free(eaRsp);

	EA_FCT_PRINT(("< ea_processRaidAddLunMaskReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processRaidDelLunMaskReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processRaidDelLunMaskReq(ea_raid_delLunMaskReqMsg_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32               		rc = NFX_OK;
	uint32						idx, eaRspLen, rcmReqLen;
    ea_raid_delLunMaskRspMsg_t 	*eaRsp;
	raid_delLunMaskReqMsg_t 	*rcmReq;
	ea_msg_t      				rcmMsg;
	raid_delLunMaskRspMsg_t 	*rcmRsp = (raid_delLunMaskRspMsg_t *) 
											rcmMsg.buf;
    
    EA_FCT_PRINT(("> ea_processRaidDelLunMaskReq()\n"));

    /* 
	 * Initialize the response for AV
     */
	eaRspLen = sizeof(ea_raid_delLunMaskRspMsg_t);
	eaRsp = (ea_raid_delLunMaskRspMsg_t *) malloc(eaRspLen);
    bzero((void *) eaRsp, eaRspLen);
    eaRsp->hdr.msgType = EA_RAID_MID_LUNMASK_DEL_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));
	eaRsp->logDevId.id = evReq->logDevId.id;
	eaRsp->ctrlNumber = evReq->ctrlNumber;
	eaRsp->portNumber = evReq->portNumber;

    /* 
	 * Create RCM request
     */
	rcmReqLen = sizeof(raid_delLunMaskReqMsg_t) +
			(evReq->numOfHostWwns - 1)*sizeof(uint64);
	rcmReq = (raid_delLunMaskReqMsg_t *) malloc (rcmReqLen);
	rcmReq->header.mid = RAID_MID_LUNMASK_DEL_REQ;
	memcpy(rcmReq->header.id, evReq->id, RAID_CTRL_ID_LEN);
	rcmReq->id = evReq->logDevId.id;
	rcmReq->ctrlNumber = evReq->ctrlNumber;
	rcmReq->portNumber = evReq->portNumber;

	rcmReq->numOfHostWwns = evReq->numOfHostWwns;
	for (idx = 0; idx < evReq->numOfHostWwns; idx++)
	{
		rcmReq->hostWwnList[idx] = evReq->hostWwnList[idx];
	}

	/* 
	 * Send request to RCM
	 */
	rc = sendAgileMsg(-1, (void *) rcmReq, rcmReqLen,
					"raid-config", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					&waitRcmTo, rcmRsp, sizeof(rcmMsg.buf));

	/* 
	 * Receive response from RCM
	 */
	if (rc <= 0)
	{
		/* 
		 * No response/Timeout 
		 */
		rcmRsp->rCode = NFX_ERR;
	}

	/*
	 * Copy back response for AV
	 */
	eaRsp->rCode = rcmRsp->rCode;

	/*
	 * Send response to AV
	 */
    ea_sendMsg(eaRsp, eaRspLen, sender->name);

	free(rcmReq);
	free(eaRsp);

	EA_FCT_PRINT(("< ea_processRaidDelLunMaskReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processRaidBpListReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processRaidBpListReq(ea_raid_bpListReqMsg_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32               	rc = NFX_OK;
	uint32					idx, eaRspLen, rcmReqLen;
    ea_raid_bpListRspMsg_t 	*eaRsp;
	raid_bpListReqMsg_t 	*rcmReq;
	ea_msg_t      			rcmMsg;
	raid_bpListRspMsg_t 	*rcmRsp = (raid_bpListRspMsg_t *) rcmMsg.buf;
    
    EA_FCT_PRINT(("> ea_processRaidBpListReq()\n"));

    /* 
	 * Initialize the response for AV
     */
	eaRspLen = sizeof(ea_raid_bpListRspMsg_t); 
	eaRsp = (ea_raid_bpListRspMsg_t *) malloc(eaRspLen);
    bzero((void *) eaRsp, eaRspLen);
    eaRsp->hdr.msgType = EA_RAID_MID_LUN_BPLIST_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));

    /* 
	 * Create RCM request
     */
	rcmReqLen = sizeof(raid_bpListReqMsg_t);
	rcmReq = (raid_bpListReqMsg_t *) malloc (rcmReqLen);
	rcmReq->header.mid = RAID_MID_LUN_BPLIST_REQ;
	memcpy(rcmReq->header.id, evReq->id, RAID_CTRL_ID_LEN);

	/* 
	 * Send request to RCM
	 */
	rc = sendAgileMsg(-1, (void *) rcmReq, rcmReqLen,
					"raid-config", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					&waitRcmTo, rcmRsp, sizeof(rcmMsg.buf));

	/* 
	 * Receive response from RCM
	 */
	if (rc <= 0)
	{
		/* 
		 * No response/Timeout 
		 */
		rcmRsp->rCode = NFX_ERR;
	}

	/*
	 * Copy back response for AV
	 */
	if (rcmRsp->numEntries > 1)
	{
		eaRspLen = sizeof(ea_raid_bpListRspMsg_t) +
					(rcmRsp->numEntries - 1)*sizeof(ea_raid_backProc_t); 
		eaRsp = (ea_raid_bpListRspMsg_t *) realloc(eaRsp, eaRspLen);
    	bzero((void *) eaRsp, eaRspLen);
    	eaRsp->hdr.msgType = EA_RAID_MID_LUN_BPLIST_RSP;
    	eaRsp->hdr.userInfo = evReq->hdr.userInfo;
		memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));
	}

	eaRsp->rCode = rcmRsp->rCode;
	eaRsp->numEntries = rcmRsp->numEntries;
	for (idx = 0; idx < rcmRsp->numEntries; idx++)
	{
		eaRsp->bpList[idx].logDevId.id = rcmRsp->bpList[idx].id;
		eaRsp->bpList[idx].checkProgress = rcmRsp->bpList[idx].checkProgress;
		eaRsp->bpList[idx].initProgress = rcmRsp->bpList[idx].initProgress;
		eaRsp->bpList[idx].rebuildProgress = rcmRsp->bpList[idx].rebuildProgress;
		eaRsp->bpList[idx].expandProgress = rcmRsp->bpList[idx].expandProgress;
	}

	/*
	 * Send response to AV
	 */
    ea_sendMsg(eaRsp, eaRspLen, sender->name);

	free(rcmReq);
	free(eaRsp);

	EA_FCT_PRINT(("< ea_processRaidBpListReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processRaidCancelBpReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 22-OCT-2001
 *
 *------------------------------------------------------------------------------
 */
static void
ea_processRaidCancelBpReq(ea_raid_cancelBpReqMsg_t * evReq,
                     ea_senderInfo_t * sender)
{
    int32               		rc = NFX_OK;
	uint32						eaRspLen, rcmReqLen;
    ea_raid_cancelBpRspMsg_t 	*eaRsp;
	raid_cancelBpReqMsg_t 		*rcmReq;
	ea_msg_t      				rcmMsg;
	raid_cancelBpRspMsg_t 		*rcmRsp = (raid_cancelBpRspMsg_t *) rcmMsg.buf;
    
    EA_FCT_PRINT(("> ea_processRaidCancelBpReq()\n"));

    /* 
	 * Initialize the response for AV
     */
	eaRspLen = sizeof(ea_raid_cancelBpRspMsg_t); 
	eaRsp = (ea_raid_cancelBpRspMsg_t *) malloc(eaRspLen);
    bzero((void *) eaRsp, eaRspLen);
    eaRsp->hdr.msgType = EA_RAID_MID_LUN_CANCELBP_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;
	memcpy(eaRsp->id, evReq->id, sizeof(ea_raid_ctrl_id_t));

    /* 
	 * Create RCM request
     */
	rcmReqLen = sizeof(raid_cancelBpReqMsg_t);
	rcmReq = (raid_cancelBpReqMsg_t *) malloc (rcmReqLen);
	rcmReq->header.mid = RAID_MID_LUN_CANCELBP_REQ;
	memcpy(rcmReq->header.id, evReq->id, RAID_CTRL_ID_LEN);
	rcmReq->type = evReq->type;

	/* 
	 * Send request to RCM
	 */
	rc = sendAgileMsg(-1, (void *) rcmReq, rcmReqLen,
					"raid-config", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					&waitRcmTo, rcmRsp, sizeof(rcmMsg.buf));

	/* 
	 * Receive response from RCM
	 */
	if (rc <= 0)
	{
		/* 
		 * No response/Timeout 
		 */
		rcmRsp->rCode = NFX_ERR;
	}

	/*
	 * Copy back response for AV
	 */
	eaRsp->rCode = rcmRsp->rCode;

	/*
	 * Send response to AV
	 */
    ea_sendMsg(eaRsp, eaRspLen, sender->name);

	free(rcmReq);
	free(eaRsp);

	EA_FCT_PRINT(("< ea_processRaidCancelBpReq()\n"));
}

/*------------------------------------------------------------------------------
 * Name         : ea_processGetChassisInfoReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 21-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processGetChassisInfoReq(ea_getChassisInfoReq_t * evReq,
                      ea_senderInfo_t *  sender)
{
    ea_getChassisInfoRsp_t 	eaRsp;
	chassisdMsgInfoReq_t	cmReq;
	chassisdMsgInfoResp_t	*cmRsp;
    int32               	rc = NFX_OK;
    
    EA_FCT_PRINT(("> ea_processGetChassisInfoReq()\n"));

    /* Initialize the response.
     */
    bzero((void *) &eaRsp, sizeof(ea_getChassisInfoRsp_t));
    eaRsp.hdr.msgType = EA_GET_CHASSIS_INFO_RSP;
    eaRsp.hdr.userInfo = evReq->hdr.userInfo;

	cmRsp = &eaRsp.chInfo;

	/*
	 * send the request to Chassisd
	 */
	bzero((void *)&cmReq, sizeof(chassisdMsgInfoReq_t));
	cmReq.hdr.mid = CHASSISD_GET_CHASSIS_INFO_REQ;

	rc = sendAgileMsg(-1, &cmReq, sizeof(chassisdMsgInfoReq_t),
	                        "cm", 0, 0, 0,
						  &waitCmTo, cmRsp, sizeof(chassisdMsgInfoResp_t));

	if (rc != NFX_OK)
	{
        eaRsp.result = NFX_ERR;
        sprintf(eaRsp.output, "Error: Chassis manager did not respond.");
	}
	else
	{
    	eaRsp.result = NFX_OK;
    	sprintf(eaRsp.output, "Done.");
	}

    ea_sendMsg(&eaRsp, sizeof(ea_getChassisInfoRsp_t), sender->name);

    EA_FCT_PRINT(("< ea_processGetChassisInfoReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processGetFcPortWwnReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 21-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processGetFcPortWwnReq(ea_listFcPortWwnReq_t * evReq,
                      ea_senderInfo_t *  sender)

{
	int 					idx; 
    ea_listFcPortWwnRsp_t 	*eaRsp;
	uint64					wwn;
    uint32              	rspLen = sizeof(ea_listFcPortWwnRsp_t);
    
    EA_FCT_PRINT(("> ea_processGetFcPortWwnReq()\n"));

    /* Initialize the response.
     */
    if (evReq->num > 1)
	{
        rspLen += (evReq->num - 1) * sizeof(SlotPortWwn_t);
	}
    eaRsp = (ea_listFcPortWwnRsp_t *) malloc(rspLen);
    NFX_ASSERT(eaRsp != NULL);
    bzero((void *) eaRsp, rspLen);
    eaRsp->hdr.msgType = EA_GET_FC_PORT_WWN_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;

	for (idx = 0; idx < evReq->num; idx++)
	{
		wwn = ea_getFcPortWwn(NULL, evReq->portList[idx].slotId,
								evReq->portList[idx].portId);
		if (wwn <= 0)
		{
			EA_DBG_PRINT(("Did not get wwn for slot %d port %d\n",
						evReq->portList[idx].slotId,
						evReq->portList[idx].portId));
			continue;
		}

		EA_DBG_PRINT(("Got wwn %lld for slot %d port %d\n",
					wwn, evReq->portList[idx].slotId,
						evReq->portList[idx].portId));

		eaRsp->portList[eaRsp->num].slotId 
							= evReq->portList[idx].slotId;
		eaRsp->portList[eaRsp->num].portId 
							= evReq->portList[idx].portId;
		eaRsp->portList[eaRsp->num].wwn = wwn;
		eaRsp->num++;
	}

    eaRsp->result = NFX_OK;
    sprintf(eaRsp->output, "Done.");

	if (eaRsp->num <= 0)
	{
		rspLen = sizeof(ea_listFcPortWwnRsp_t);
	}
	else
	{
		rspLen = (sizeof(ea_listFcPortWwnRsp_t) + 
					(eaRsp->num - 1)*sizeof(SlotPortWwn_t));
	}
    ea_sendMsg(eaRsp, rspLen, sender->name);
    free(eaRsp);

    EA_FCT_PRINT(("< ea_processGetFcPortWwnReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processLunRediscoverReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 21-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processLunRediscoverReq(ea_lunRediscoverReq_t * evReq,
                      ea_senderInfo_t *  sender)
{
    ea_lunRediscoverRsp_t 	*eaRsp;
    uint32              	rspLen = sizeof(ea_lunRediscoverRsp_t);
	sdm_rediscReqMsg_t		sdmMsg;
    
    EA_FCT_PRINT(("> ea_processLunRediscoverReq()\n"));

    /* Initialize the response.
     */
    eaRsp = (ea_lunRediscoverRsp_t *) malloc(rspLen);
    NFX_ASSERT(eaRsp != NULL);
    bzero((void *) eaRsp, rspLen);
    eaRsp->hdr.msgType = EA_LUN_REDISCOVER_RSP;
    eaRsp->hdr.userInfo = evReq->hdr.userInfo;

    /* 
	 * Create SDM request
     */
	memset((void *)&sdmMsg, 0, sizeof(sdm_rediscReqMsg_t));
	sdmMsg.hdr.mid = SDM_MID_REDISCOVER_REQ;
	sdmMsg.wwn = htonll(evReq->wwn);

	/* 
	 * Send request to SDM (No Ack from SDM)
	 */
	sendAgileMsg(-1, (void *) &sdmMsg, sizeof(sdm_rediscReqMsg_t),
					"sdm", NFX_SSC_UNIX, SSC_SLOT, MGMTPLANE,
					NULL, NULL, NULL);

    eaRsp->result = NFX_OK;
    sprintf(eaRsp->output, "Done.");

    ea_sendMsg(eaRsp, rspLen, sender->name);
    free(eaRsp);

    EA_FCT_PRINT(("< ea_processLunRediscoverReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Name         : ea_processGetPortTopologyReq()
 *
 * Description  : 
 *
 * Created by   : Pankaj Gupta
 *
 * Date Created : 21-NOV-2001
 *
 *------------------------------------------------------------------------------
 */

static void
ea_processGetPortTopologyReq(ea_getPortTopologyReq_t * evReq,
                      ea_senderInfo_t *  sender)
{
    EA_FCT_PRINT(("> ea_processGetPortTopologyReq()\n"));

	printf("ea_processGetPortTopologyReq: TBD (waiting for SDM update)\n");

    EA_FCT_PRINT(("< ea_processGetPortTopologyReq()\n"));
    return;
}

/*------------------------------------------------------------------------------
 * Global functions.
 *------------------------------------------------------------------------------
 */

/*------------------------------------------------------------------------------
 * Name         : ea_processEvMsg()
 *
 * Description  : Process all messages received from EV.
 *
 * Created by   : Isabelle Cnudde
 *
 * Date Created : 01-SEP-2001
 *
 *------------------------------------------------------------------------------
 */

void
ea_processEvMsg(ea_msg_t *          msg,
                ea_senderInfo_t *   sender)
    
{
    EA_FCT_PRINT(("> ea_processEvMsg()\n"));
                       
    if (msg->len < sizeof(ea_msgHdr_t))
    {
        EA_ERR_PRINT(("Got truncated message of len %d", msg->len));
    }
    else
    {
        ea_msgHdr_t * hdr = (ea_msgHdr_t *) msg->buf;
        
        /* Process the message according to its type.
         */
        switch (hdr->msgType)
        {

		case EA_SPM_GET_TOTAL_POOLS_REQ:
		case EA_SPM_GET_POOL_LIST_REQ:
		case EA_SPM_GET_TOTAL_LUNS_REQ:
		case EA_SPM_GET_LUN_LIST_REQ:
		case EA_SPM_MARK_LUN_AS_AGILE_REQ:
		case EA_SPM_MARK_LUN_AS_OC_REQ:
		case EA_SPM_UNMARK_LUN_AS_AGILE_REQ:
		case EA_SPM_RESCAN_LUN_REQ:
		case EA_SPM_GET_STORAGE_POOL_LIST_REQ:
		case EA_SPM_GET_STORAGE_LUN_LIST_REQ:
		{
			ea_msgHdr_t * req = (ea_msgHdr_t *) &(msg->buf);

			EA_DBG_PRINT(("EA: AV (Req) -> SPM\n"));
			ea_processSpmReq(req, sender);
			break;
		}

		case SPM_GET_TOTAL_POOLS_RSP:
		case SPM_GET_POOL_LIST_RSP:
		case SPM_GET_TOTAL_LUNS_RSP:
		case SPM_GET_LUN_LIST_RSP:
		case SPM_MARK_LUN_AS_AGILE_RSP:
		case SPM_MARK_LUN_AS_OC_RSP:
		case SPM_UNMARK_LUN_AS_AGILE_RSP:
		case SPM_RESCAN_LUN_RSP:
		case SPM_GET_STORAGE_POOL_LIST_RSP:
		case SPM_GET_STORAGE_LUN_LIST_RSP:
		case SPM_SRM_INIT_IND:
		case SPM_SRM_NEW_LUN_IND:
		case SPM_SRM_DEL_LUN_IND:
		case SPM_SRM_LUN_STATE_CHANGE_IND:
		{
			spm_msgHdr_t * spmRsp = (spm_msgHdr_t *) &(msg->buf);

			EA_DBG_PRINT(("EA: SPM (Rsp/Ind) -> AV\n"));
			ea_processSpmRsp(spmRsp, msg->len, sender);
			break;
		}

        case EA_ADD_FSYS_REQ:
        {
            ea_addFsysReq_t * req = (ea_addFsysReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_ADD_FSYS_REQ\n"));
            ea_processAddFsysReq(req, sender);
            break;
        }

                
        case EA_MOD_FSYS_REQ:
        {
            ea_modFsysReq_t * req = (ea_modFsysReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_MOD_FSYS_REQ\n"));
            ea_processModFsysReq(req, sender);
            break;
        }
                
        case EA_MNT_FSYS_REQ:
        {
            ea_mntFsysReq_t * req = (ea_mntFsysReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_MNT_FSYS_REQ\n"));
            ea_processMntFsysReq(req, sender);
            break;
        }
                
        case EA_UMNT_FSYS_REQ:
        {
            ea_umntFsysReq_t * req = (ea_umntFsysReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_UMNT_FSYS_REQ\n"));
            ea_processUmntFsysReq(req, sender);
            break;
        }
                
        case EA_DEL_FSYS_REQ:
        {
            ea_delFsysReq_t * req = (ea_delFsysReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_DEL_FSYS_REQ\n"));
            ea_processDelFsysReq(req, sender);
            break;
        }

        case EA_NUM_FSYS_REQ:
        {
            ea_numFsysReq_t * req = (ea_numFsysReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_NUM_FSYS_REQ\n"));
            ea_processNumFsysReq(req, sender);
            break;
        }
                
        case EA_GET_FSYS_REQ:
        {
            ea_getFsysReq_t * req = (ea_getFsysReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_GET_FSYS_REQ\n"));
            ea_processGetFsysReq(req, sender);
            break;
        }

        case EA_GET_DETAIL_FSYS_REQ:
        {
            ea_getDetailFsysReq_t * req = (ea_getDetailFsysReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_GET_DETAIL_FSYS_REQ\n"));
            ea_processGetDetailFsysReq(req, sender);
            break;
        }

        case EA_USG_FSYS_REQ:
        {
            ea_usgFsysReq_t * req = (ea_usgFsysReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_USG_FSYS_REQ\n"));
            ea_processUsgFsysReq(req, sender);
            break;
        }

        case EA_MOVIN_FSYS_REQ:
        case EA_FORCE_MOVIN_FSYS_REQ:
        {
            ea_movInFsysReq_t * req = (ea_movInFsysReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_MOVIN_FSYS_REQ\n"));
            ea_processMovInFsysReq(req, sender);
            break;
        }

        case EA_MOVOUT_FSYS_REQ:
        {
            ea_movOutFsysReq_t * req = (ea_movOutFsysReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_MOVOUT_FSYS_REQ\n"));
            ea_processMovOutFsysReq(req, sender);
            break;
        }

        case EA_GROW_FSYS_REQ:
        {
            ea_growFsysReq_t * req = (ea_growFsysReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_GROW_FSYS_REQ\n"));
			EA_DBG_PRINT(("ea_processGrowFsysReq: Needed size %lld numVendor %d\n",
					req->neededSize, req->attr.num));

            ea_processGrowFsysReq(req, sender);
            break;
        }

        case EA_GROW_FSYS_RSP:
        {
            ea_growFsysRsp_t * req = (ea_growFsysRsp_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_GROW_FSYS_RSP\n"));
            ea_processGrowFsysRsp(req, sender);
            break;
        }

        case EA_GROW_FSYS_IND:
        {
            ea_growFsysInd_t * req = (ea_growFsysInd_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_GROW_FSYS_IND\n"));
            ea_processGrowFsysInd(req, sender, msg->len);
            break;
        }
                
        case EA_ADD_FSET_REQ:
        {
            ea_addFsetReq_t * req = (ea_addFsetReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_ADD_FSET_REQ\n"));
            ea_processAddFsetReq(req, sender);
            break;
        }
                
        case EA_MOD_FSET_REQ:
        {
            ea_modFsetReq_t * req = (ea_modFsetReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_MOD_FSET_REQ\n"));
            ea_processModFsetReq(req, sender);
            break;
        }
                
        case EA_DEL_FSET_REQ:
        {
            ea_delFsetReq_t * req = (ea_delFsetReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_DEL_FSET_REQ\n"));
            ea_processDelFsetReq(req, sender);
            break;
        }

        case EA_NUM_FSET_REQ:
        {
            ea_numFsetReq_t * req = (ea_numFsetReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_NUM_FSET_REQ\n"));
            ea_processNumFsetReq(req, sender);
            break;
        }

        case EA_GET_FSET_REQ:
        {
            ea_getFsetReq_t * req = (ea_getFsetReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_GET_FSET_REQ\n"));
            ea_processGetFsetReq(req, sender);
            break;
        }

        case EA_USG_FSET_REQ:
        {
            ea_usgFsetReq_t * req = (ea_usgFsetReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_USG_FSET_REQ\n"));
            ea_processUsgFsetReq(req, sender);
            break;
        }

		case EA_ADD_EXP_REQ:
		{
            ea_addExpReq_t * req = (ea_addExpReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_ADD_EXP_REQ\n"));
            ea_processAddExpReq(req, sender);
            break;
		}

		case EA_MOD_EXP_REQ:
		{
            ea_modExpReq_t * req = (ea_modExpReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_MOD_EXP_REQ\n"));
            ea_processModExpReq(req, sender);
            break;
		}

		case EA_DEL_EXP_REQ:
		{
            ea_delExpReq_t * req = (ea_delExpReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_DEL_EXP_REQ\n"));
            ea_processDelExpReq(req, sender);
            break;
		}

		case EA_DEL_PATH_REQ:
		{
            ea_delPathReq_t * req = (ea_delPathReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_DEL_PATH_REQ\n"));
            ea_processDelPathReq(req, sender);
            break;
		}

		case EA_GET_EXP_REQ:
		{
            ea_getExpReq_t * req = (ea_getExpReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_GET_EXP_REQ\n"));
            ea_processGetExpReq(req, sender);
            break;
		}

		case EA_GET_PATH_EXP_REQ:
		{
            ea_getPathExpReq_t * req = (ea_getPathExpReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_GET_PATH_EXP_REQ\n"));
            ea_processGetPathExpReq(req, sender);
            break;
		}

        case EA_CRT_SNAP_REQ:
        {
            ea_crtSnapReq_t * req = (ea_crtSnapReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_CRT_SNAP_REQ\n"));
            ea_processCrtSnapReq(req, sender);
            break;
        }

        case EA_DEL_SNAP_REQ:
        {
            ea_delSnapReq_t * req = (ea_delSnapReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_DEL_SNAP_REQ\n"));
            ea_processDelSnapReq(req, sender);
            break;
        }

        case EA_SCH_SNAP_REQ:
        {
            ea_schSnapReq_t * req = (ea_schSnapReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_SCH_SNAP_REQ\n"));
            ea_processSchSnapReq(req, sender);
            break;
        }

        case EA_GET_SCH_SNAP_REQ:
        {
            ea_getSchSnapReq_t * req = (ea_getSchSnapReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_GET_SCH_SNAP_REQ\n"));
            ea_processGetSchSnapReq(req, sender);
            break;
        }

        case EA_GET_USG_SNAP_REQ:
        {
            ea_getUsgSnapReq_t * req = (ea_getUsgSnapReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_GET_USG_SNAP_REQ\n"));
            ea_processGetUsgSnapReq(req, sender);
            break;
        }

		case EA_RAID_MID_GET_LIST_REQ:
		{
			ea_raid_getListReqMsg_t * req = 
							(ea_raid_getListReqMsg_t *) &(msg->buf);
            EA_DBG_PRINT(("EA_RAID_MID_GET_LIST_REQ\n"));
            ea_processRaidGetListReq(req, sender);
            break;
		}

		case EA_RAID_MID_OPEN_REQ:
		{
			ea_raid_openReqMsg_t * req = 
							(ea_raid_openReqMsg_t *) &(msg->buf);
            EA_DBG_PRINT(("EA_RAID_MID_OPEN_REQ\n"));
            ea_processRaidOpenReq(req, sender);
            break;
		}

		case EA_RAID_MID_CLOSE_REQ:
		{
			ea_raid_closeReqMsg_t * req = 
							(ea_raid_closeReqMsg_t *) &(msg->buf);
            EA_DBG_PRINT(("EA_RAID_MID_CLOSE_REQ\n"));
            ea_processRaidCloseReq(req, sender);
            break;
		}

		case EA_RAID_MID_CTRL_INFO_REQ:
		{
			ea_raid_ctrlInfoReqMsg_t * req = 
							(ea_raid_ctrlInfoReqMsg_t *) &(msg->buf);
            EA_DBG_PRINT(("EA_RAID_MID_CTRL_INFO_REQ\n"));
            ea_processRaidCtrlInfoReq(req, sender);
            break;

		}

		case EA_RAID_MID_CTRL_GETPARAM_REQ:
		{
			ea_raid_getCtrlParamReqMsg_t * req = 
							(ea_raid_getCtrlParamReqMsg_t *) &(msg->buf);
            EA_DBG_PRINT(("EA_RAID_MID_CTRL_GETPARAM_REQ\n"));
            ea_processRaidCtrlParamReq(req, sender);
            break;
		}

		case EA_RAID_MID_PHYDEV_LIST_REQ:
		{
			ea_raid_phyDevListReqMsg_t * req = 
							(ea_raid_phyDevListReqMsg_t *) &(msg->buf);
            EA_DBG_PRINT(("EA_RAID_MID_PHYDEV_LIST_REQ\n"));
            ea_processRaidPhyDevListReq(req, sender);
            break;
		}

		case EA_RAID_MID_PHYDEV_INFO_REQ:
		{
			ea_raid_phyDevInfoReqMsg_t * req = 
							(ea_raid_phyDevInfoReqMsg_t *) &(msg->buf);
            EA_DBG_PRINT(("EA_RAID_MID_PHYDEV_INFO_REQ\n"));
            ea_processRaidPhyDevInfoReq(req, sender);
            break;
		}

		case EA_RAID_MID_LOGDEV_LIST_REQ:
		{
			ea_raid_logDevListReqMsg_t * req = 
							(ea_raid_logDevListReqMsg_t *) &(msg->buf);
            EA_DBG_PRINT(("EA_RAID_MID_LOGDEV_LIST_REQ\n"));
            ea_processRaidLogDevListReq(req, sender);
            break;
		}

		case EA_RAID_MID_LOGDEV_INFO_REQ:
		{
			ea_raid_logDevInfoReqMsg_t * req = 
							(ea_raid_logDevInfoReqMsg_t *) &(msg->buf);
            EA_DBG_PRINT(("EA_RAID_MID_LOGDEV_INFO_REQ\n"));
            ea_processRaidLogDevInfoReq(req, sender);
            break;
		}

		case EA_RAID_MID_LUN_CREATE_REQ:
		{
			ea_raid_createLunReqMsg_t * req = 
							(ea_raid_createLunReqMsg_t *) &(msg->buf);
            EA_DBG_PRINT(("EA_RAID_MID_LUN_CREATE_REQ\n"));
            ea_processRaidCreateLunReq(req, sender);
            break;
		}

		case EA_RAID_MID_LUN_CREATE_SIMPLE_REQ:
		{
			ea_raid_createSimpleLunReqMsg_t * req = 
							(ea_raid_createSimpleLunReqMsg_t *) &(msg->buf);
            EA_DBG_PRINT(("EA_RAID_MID_LUN_CREATE_SIMPLE_REQ\n"));
            ea_processRaidCreateSimpleLunReq(req, sender);
            break;
		}

		case EA_RAID_MID_LUN_DELETE_REQ:
		{
			ea_raid_deleteLunReqMsg_t * req = 
							(ea_raid_deleteLunReqMsg_t *) &(msg->buf);
            EA_DBG_PRINT(("EA_RAID_MID_LUN_DELETE_REQ\n"));
            ea_processRaidDeleteLunReq(req, sender);
            break;
		}

		case EA_RAID_MID_LUN_INIT_REQ:
		{
			ea_raid_initLunReqMsg_t * req = 
							(ea_raid_initLunReqMsg_t *) &(msg->buf);
            EA_DBG_PRINT(("EA_RAID_MID_LUN_INIT_REQ\n"));
            ea_processRaidInitLunReq(req, sender);
            break;
		}

		case EA_RAID_MID_LUNMASK_INFO_REQ:
		{
			ea_raid_getLunMaskReqMsg_t * req = 
							(ea_raid_getLunMaskReqMsg_t *) &(msg->buf);
            EA_DBG_PRINT(("EA_RAID_MID_LUNMASK_INFO_REQ\n"));
            ea_processRaidGetLunMaskReq(req, sender);
            break;
		}

		case EA_RAID_MID_LUNMASK_ADD_REQ:
		{
			ea_raid_addLunMaskReqMsg_t * req = 
							(ea_raid_addLunMaskReqMsg_t *) &(msg->buf);
            EA_DBG_PRINT(("EA_RAID_MID_LUNMASK_ADD_REQ\n"));
            ea_processRaidAddLunMaskReq(req, sender);
            break;
		}

		case EA_RAID_MID_LUNMASK_DEL_REQ:
		{
			ea_raid_delLunMaskReqMsg_t * req = 
							(ea_raid_delLunMaskReqMsg_t *) &(msg->buf);
            EA_DBG_PRINT(("EA_RAID_MID_LUNMASK_DEL_REQ\n"));
            ea_processRaidDelLunMaskReq(req, sender);
            break;
		}

		case EA_RAID_MID_LUN_BPLIST_REQ:
		{
            ea_raid_bpListReqMsg_t * req = 
							(ea_raid_bpListReqMsg_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_RAID_MID_LUN_BPLIST_REQ\n"));
            ea_processRaidBpListReq(req, sender);
            break;
		}

		case EA_RAID_MID_LUN_CANCELBP_REQ:
		{
            ea_raid_cancelBpReqMsg_t * req = 
							(ea_raid_cancelBpReqMsg_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_RAID_MID_LUN_CANCELBP_REQ\n"));
            ea_processRaidCancelBpReq(req, sender);
            break;
		}

        case EA_GET_CHASSIS_INFO_REQ:
        {
            ea_getChassisInfoReq_t * req = 
							(ea_getChassisInfoReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_GET_CHASSIS_INFO_REQ\n"));
            ea_processGetChassisInfoReq(req, sender);
            break;
        }

        case EA_GET_FC_PORT_WWN_REQ:
        {
            ea_listFcPortWwnReq_t * req = (ea_listFcPortWwnReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_GET_FC_PORT_WWN_REQ\n"));
            ea_processGetFcPortWwnReq(req, sender);
            break;
        }

        case EA_LUN_REDISCOVER_REQ:
        {
            ea_lunRediscoverReq_t * req = (ea_lunRediscoverReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_LUN_REDISCOVER_REQ\n"));
            ea_processLunRediscoverReq(req, sender);
            break;
        }

        case EA_GET_PORT_TOPOLOGY_REQ:
        {
            ea_getPortTopologyReq_t * req = 
							(ea_getPortTopologyReq_t *) &(msg->buf);

            EA_DBG_PRINT(("EA_GET_PORT_TOPOLOGY_REQ\n"));
            ea_processGetPortTopologyReq(req, sender);
            break;
        }

        default:
        {
            EA_ERR_PRINT(("EA-EV: Unknown msg type (%d)\n", hdr->msgType));
            break;
        }
        } /* switch */
    }
    
    EA_FCT_PRINT(("< ea_processEvMsg()\n"));
    return;
}

/*----------------------------------------------------------------------------*/
