// -*- tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=4 shiftwidth=4 expandtab:

/*
 * Name:  fs-buf.c
 *
 * RCS:  $Id: fs-buf.c,v 1.381 2002/12/10 17:27:07 bob Exp $
 *
 * Description:  Buffer cache management.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#if !defined(SSC)

#include "fs-includes.h"
#include "fs-wb.h"

// Perform read-after-write verification
// of meta bufs?
#if defined (FS_DEBUG_BUF_CKSUM)
int32 fs_bufDoReadVerify = 1;
#else
int32 fs_bufDoReadVerify = 0;
#endif

/* 
 * buf->spin  and  buf->lock  rules 
 *
 * - Use always the buf->spin to set/reset flags
 *   Also needed if you have the buf->lock exclusive !!!
 *   (the WB daemon set/resets flags without having a shared buf->lock
 *
 * - Always get the fs_bufListGlobal[bufList->type].spin first and
 *   then the buf->spin (if you need both)
 *
 * - use the buf->spin to test/incr/decr the logRefCount
 *
 * - Never use atomic operations on bufs->flags 
 *   atomic and spinlock on the same fields must be used exclusive
 *
 * - no buf->spin lock needed to set flags just after allocating a new buf
 *     while holding the fs_bufListGlobal[listType].spin lock
 *
 * - All DLL list entries must have the fs_bufListGlobal[bufList->type].spin
 *
 * - mrdEntry usage
 *   FS_BUF_MRD         indicates bufList->mrdHdr
 *   FS_BUF_WAITING_MRD indicates bufList->mrdWaitingHdr 
 *   none               indicates bufList->logSegmentHdr[LogSeg]
 *   (re)setting these needs the buf->spin AND the fs_bufListGlobal[bufList->type].spin
 *
 * - FS_BUF_DIRTY need buf->spin and fs_bufListGlobal[bufList->type].spin or fs_lockOwnedExclusive
 *   except Line 4170, 6595, 7346
 *
 * - FS_BUF_TRANS_WRITE need buf->spin and fs_lockOwnedExclusive
 *   except Line 5578 --> move clearing the flag when we give the buf to the WBD 
 *
 */ 

#ifdef FS_DEBUG_BUF_HISTORY
#define FS_DEBUG_BUF_HISTORY_ALLOC_MAX 100
int32 fs_buf_allocate_index = 0 ;
fs_Buf *fs_buf_allocate_ptr[FS_DEBUG_BUF_HISTORY_ALLOC_MAX];
char    fs_buf_allocate_oper[FS_DEBUG_BUF_HISTORY_ALLOC_MAX];
nfx_spinlock_t  spinalloc;
#define FS_DEBUG_BUF_HISTORY_ALLOC           1
#define FS_DEBUG_BUF_HISTORY_DEALLOC         2
#define FS_DEBUG_BUF_HISTORY_ALLOC_CLONE     3
#define FS_DEBUG_BUF_HISTORY_DEALLOC_CLONE   4

#endif

uint32  fs_diskWritebackType[FS_MAX_META_INODE];
uint32  fs_transWriteType[FS_MAX_META_INODE];

int32 fs_bufBufAllocRetries = 0;

#define BUF_TRACE(args)   FS_TRACE(context, FS_TRACE_BUF, args)

/**
 * Forward declarations.
 */
static inline boolean
fs_bufIsValid(
    fs_Buf  *buf);

static fs_Buf *
fs_bufFindAndFreeBuf(
    fs_Context  *context,
    int32       listType,
    int32       itype);


/**
 * The global buffer cache structure.
 */
fs_BufListGlobal  fs_bufListGlobal[FS_BUFLIST_MAX];

int32 fs_bufWbWakeupCount[FS_BUFLIST_MAX][FS_WB_WAKEUP_MAX];


/**
 * Cache hash function for block number list.
 */
int32 
fs_bufHashVolBlock(
    evm_volId_t volId,
    fs_Block    block)
{
#ifdef NOTDEF
    struct {
        evm_volId_t volId;
        fs_Block    block;
    } info;

    info.volId = volId;
    info.block = block;

    return fs_coreHash(&info, sizeof(info) / sizeof(uint64), 0) %
                                FS_BUFLIST_HASH_MAX;
#else
    uint64 hash;

    hash = volId.details.u.parts.clusterUnique + (fs_blockNumber(block) << 5);
    hash += (hash >> FS_BUFLIST_HASH_SHIFT);
    return (hash % FS_BUFLIST_HASH_MAX);

    // return (((volId.details.u.parts.clusterUnique << 12) + fs_blockNumber(block)) % FS_BUFLIST_HASH_MAX);
#endif
}


/**
 * Cache hash function for inum page header.
 */
int32
fs_bufHashVolInum(
    evm_volId_t volId,
    uint64      inum)
{
#ifdef NOTDEF
    struct {
        evm_volId_t volId;
        uint64      inum;
    } info;

    info.volId = volId;
    info.inum = inum;

    return fs_coreHash(&info, sizeof(info) / sizeof(uint64), 0) %
                                FS_BUFLIST_HASH_MAX;
#else
    uint64 hash;

    hash = volId.details.u.parts.clusterUnique + (inum << 5);
    hash += (hash >> FS_BUFLIST_HASH_SHIFT);
    return (hash % FS_BUFLIST_HASH_MAX);

    // return (((volId.details.u.parts.clusterUnique << 12) + inum) % FS_BUFLIST_HASH_MAX);
#endif
}


/**
 * Cache hash function for volume, inum, page.
 */
int32
fs_bufHashVolInumPage(
    evm_volId_t volId,
    uint64      inum,
    uint64      page)
{
#ifdef NOTDEF
    struct {
        evm_volId_t volId;
        uint64      inum;
        uint64      page;
    } info;

    info.volId = volId;
    info.inum = (inum << 6);
    info.page = page;

    return fs_coreHash(&info, sizeof(info) / sizeof(uint64), 0) %
                                FS_BUFLIST_HASH_MAX;
#else
    uint64 hash;

    hash = volId.details.u.parts.clusterUnique + (page << 5) + (inum << 16);
    hash += (hash >> FS_BUFLIST_HASH_SHIFT);
    return (hash % FS_BUFLIST_HASH_MAX);
    
    // return (((volId.details.u.parts.clusterUnique << 12) + (page << 8) + inum) % FS_BUFLIST_HASH_MAX);
#endif
}


/**
 * Deallocate the in-core resources associated with the
 * specified buf.
 */
static void
fs_bufDeallocCoreRes(
    fs_Context  *context,
    fs_Buf      *buf)
{
    if (buf->flags & FS_BUF_TYPE_USER)
    {
        if (buf->u.user.codHandle != 0)
        {
            fs_cod_free_cod(buf->u.user.codHandle);
            buf->u.user.codHandle = 0;
        }
    }
    else
    {
        fs_LogUndoSummary   *cando, *candoNext;
            
        if (buf->u.data)
        {
            if (fs_bufIsCompact(buf))
            {
                fs_ramDealloc(FS_RAM_COMPACT_LEAF, buf->u.data);
            }
            else
            {
                fs_ramDealloc(FS_RAM_READ_DATA, buf->u.data);
            }
            buf->u.data = NULL;
        }

        if (buf->clone != NULL)
        {
            if (buf->clone->u.data)
            {
                fs_ramDealloc(FS_RAM_BUF_CLONE_DATA,
                              buf->clone->u.data);
            }
            fs_ramDealloc(FS_RAM_BUF_CLONE, buf->clone);
#ifdef FS_DEBUG_BUF_HISTORY
            NFX_SPIN_LOCK(&spinalloc);
            fs_buf_allocate_ptr[fs_buf_allocate_index] = buf->clone;
            fs_buf_allocate_oper[fs_buf_allocate_index] = FS_DEBUG_BUF_HISTORY_DEALLOC_CLONE;
            fs_buf_allocate_index = (fs_buf_allocate_index + 1) % FS_DEBUG_BUF_HISTORY_ALLOC_MAX;
            NFX_SPIN_UNLOCK(&spinalloc);
#endif
            buf->clone = NULL;
        }
    
        // clean up the cando records
        NFX_SPIN_LOCK(&buf->spin);
        DLL_ITERATE_NEXT(cando, candoNext, &buf->candoHdr, bufEntry)
        {
            DLL_REMOVE(&buf->candoHdr, cando, bufEntry);
            fs_ramDealloc(FS_RAM_LOG_UNDO_SUMM, cando);
        }
        NFX_SPIN_UNLOCK(&buf->spin);

#if defined(FS_DEBUG_SHADOW_BUF)
        FS_DBG_FREE_SHADOW(buf);
#endif

#ifdef BUDDY_INCORE
        if (buf->inum == FS_FREE_BUDDY_TREE_INODE &&
            buf->page != FS_BUF_NO_PAGE)
        {
            NFX_SPIN_LOCK(&(buf->fileSystem->buddies[buf->page].spin));
            buf->fileSystem->buddies[buf->page].flags = 0;
            if (buf->fileSystem->buddies[buf->page].summary != NULL)
            {
                fs_ramDealloc(FS_RAM_REGION_SUMMARY,
                              buf->fileSystem->buddies[buf->page].summary);
                buf->fileSystem->buddies[buf->page].summary = NULL;
            }
            NFX_SPIN_UNLOCK(&(buf->fileSystem->buddies[buf->page].spin));
        }
#endif // BUDDY_INCORE

    }
}


/**
 * Allocate the data area for the specified buf.  Allocates
 * a compact area if the buf so indicates.
 */
static inline void
fs_bufAllocDataPtr(
    fs_Context  *context,
    fs_Buf      *buf)
{
    FS_ASSERT(buf->flags & FS_BUF_TYPE_META);
    FS_ASSERT(buf->u.data == NULL);

    if (fs_bufIsCompact(buf))
    {
        buf->u.data = fs_ramAlloc(FS_RAM_COMPACT_LEAF, FS_COMPACT_PAGE_SIZE);
    }
    else
    {
        buf->u.data = fs_ramAlloc(FS_RAM_READ_DATA, FS_BLOCK_SIZE);
    }
}


/**
 * Return TRUE if buf is an ultra buf.
 */
static inline boolean
fs_isUltraBuf(
    void    *ptr)
{
#ifdef FS_STANDALONE
    return FALSE;
#else
    void    *bufBase;

    return eee_isUltraBuf(ptr, &bufBase);
#endif
}

/**
 * Attach the supplied data pointer to the specified
 * buf, or if the flag indicates a compact leaf and
 * the pointer is an ultra-buf, then allocate a compact
 * pointer and copy the data into it and free the ultra
 * buf.
 */
static void
fs_bufSetDataPtr(
    fs_Context  *context,
    fs_Buf      *buf,
    void        *ptr)
{
    FS_ASSERT(buf->flags & FS_BUF_TYPE_META);
    FS_ASSERT(buf->u.data == NULL);

    if (fs_bufIsCompact(buf) &&
        fs_isUltraBuf(ptr))
    {
        void *newPtr = fs_ramAlloc(FS_RAM_COMPACT_LEAF, FS_COMPACT_PAGE_SIZE);

#ifdef FS_STANDALONE
        memcpy(newPtr, ptr, FS_COMPACT_PAGE_SIZE);
#else
        local_memcpy_ass(newPtr, ptr, FS_COMPACT_PAGE_SIZE / CACHELINE_SIZE);
#endif

        fs_ramDealloc(FS_RAM_READ_DATA, ptr);

        ptr = newPtr;
    }

    buf->u.data = ptr;
}


/**
 * Return meta cache itype for buf.
 */
inline int32
fs_bufGetItype(
    fs_Buf  *buf)
{
    if (buf->flags & FS_BUF_TYPE_USER)
        return FS_BUF_CACHE_ITYPE_NONDIR;

    switch(buf->flags & (FS_BUF_DIRECTORY | FS_BUF_COMPACT))
    {
    case FS_BUF_DIRECTORY:
        return FS_BUF_CACHE_ITYPE_DIR;
    case FS_BUF_COMPACT:
        return FS_BUF_CACHE_ITYPE_COMP_NONDIR;
    case (FS_BUF_DIRECTORY | FS_BUF_COMPACT):
        return FS_BUF_CACHE_ITYPE_COMP_DIR;
    default:
        return FS_BUF_CACHE_ITYPE_NONDIR;
    }
}


/**
 * Return meta cache itype for inode.
 */
static inline int32
fs_bufGetItypeFromInode(
    fs_Inode    *inode)
{
    if (inode != NULL)
    {
        if (FS_I_IS_DIR(inode->dinode.attrs))
        {
            if (FS_I_IS_COMPACT_DIR(inode->dinode.attrs))
                return FS_BUF_CACHE_ITYPE_COMP_DIR;
            else
                return FS_BUF_CACHE_ITYPE_DIR;
        }
        else if (FS_I_IS_COMPACT_EMAP(inode->dinode.attrs))
            return FS_BUF_CACHE_ITYPE_COMP_NONDIR;
    }

    return FS_BUF_CACHE_ITYPE_NONDIR;
}


/**
 * Return true if the specified block number is marked copy on write.
 */
boolean
fs_blockIsCopyOnWrite(
    fs_Context *context,
    fs_Block   block)
{
    return context->fileSystem->bootstrap == NULL
        && fs_refbyteSnapCount(fs_refGetByte(context, block)) > 0;
}


/**
 * Called to get the copy-on-write state for a buffer.  Returns TRUE if
 * buf is COW.
 *
 * Assumes a lock (shared, at least) is held on the buffer.
 */
boolean
fs_bufIsCopyOnWrite(
    fs_Context      *context,
    fs_Buf          *buf)
{
    // don't copy on write while mkfs is running
    if (context->fileSystem->bootstrap != NULL)
        return FALSE;
    
    // dirty or deferred-allocation buffers are never COW
    if ((buf->flags & FS_BUF_DIRTY) != 0 || !fs_blockIsValid(buf->block))
        return FALSE;

    return fs_blockIsCopyOnWrite(context, buf->block);
}


/**
 * Return the pageHdr for the specified inum.
 */
fs_BufPageHdr *
fs_bufPageHdrForInum(
    fs_FileSystem       *fileSystem,
    int32               bufListType,
    uint64              inum)
{
    fs_BufListGlobal    *bufListGlobal = &fs_bufListGlobal[bufListType];
    fs_BufPageHdr       *pageHdr;
    int32               hash;

    hash = fs_bufHashVolInum(fileSystem->volId, inum);

    DLL_ITERATE(pageHdr,
                &bufListGlobal->bufsInum[hash],
                inumEntry)
    {
        //printf("checking pghdr inum %lld fs %p for inum %lld fs %p\n",
        //  pageHdr->inum, pageHdr->fileSystem, inum, context->fileSystem);

        if (pageHdr->inum == inum &&
            pageHdr->fileSystem == fileSystem)
        {
            return pageHdr;
        }
    }
    return NULL;
}


/**
 * Return the filesystem->buflist for the specified buf.
 */
static inline fs_BufList *
fs_bufListForBuf(
    fs_Buf  *buf)
{
    if (buf->flags & FS_BUF_TYPE_USER)
    {
        return &buf->fileSystem->bufList[FS_BUFLIST_USER];
    }
    else
    {
        return &buf->fileSystem->bufList[FS_BUFLIST_META];
    }
}
        

/** 
 * Decrement the appropriate cache used counts based on
 * the specified buf being removed from the cache.
 */
static void
fs_bufDecrUsedCount(
    fs_BufList  *bufList,
    fs_Buf      *buf)
{
    int32   itype = fs_bufGetItype(buf);

    FS_ASSERT(fs_bufListGlobal[bufList->type].count.usedBufs[itype] > 0);
    FS_ASSERT(bufList->count.usedBufs[itype] > 0);

    //printf("decrused buf %p flags 0x%x itype %d\n",
    //       buf, buf->flags, itype);

    bufList->count.usedBufs[itype] -= 1;
    fs_bufListGlobal[bufList->type].count.usedBufs[itype] -= 1;

    bufList->count.usedBufs[FS_BUF_CACHE_ITYPE_TOTAL] -= 1;
    fs_bufListGlobal[bufList->type].count.usedBufs[FS_BUF_CACHE_ITYPE_TOTAL] -= 1;

#ifdef FS_DEBUG_BUF_CNT_MISS_TYPE
    if (buf->flags & FS_BUF_INUM_CNT)
    {
        NFX_SPIN_LOCK(&buf->spin);
        buf->flags &= ~FS_BUF_INUM_CNT;
        NFX_SPIN_UNLOCK(&buf->spin);
        bufList->count.useCntType[buf->inum]--;
        fs_bufListGlobal[bufList->type].count.useCntType[buf->inum]--;
    }
#endif
}


/** 
 * Increment the appropriate cache used counts based on
 * the specified buf being added to the cache.
 */
static void
fs_bufIncrUsedCount(
    fs_Context  *context,
    fs_BufList  *bufList,
    fs_Buf      *buf)
{
    int32   itype = fs_bufGetItype(buf);

    //printf("incrused buf %p flags 0x%x itype %d\n",
    //       buf, buf->flags, itype);

    // Nothing uses this yet, so if we get this type, something
    // aint right somewhere.
    // FS_ASSERT(itype != FS_BUF_CACHE_ITYPE_COMP_NONDIR);
    ASSERT_LOCK_OWNED(&fs_bufListGlobal[bufList->type].spin);

    bufList->count.usedBufs[itype] += 1;
    fs_bufListGlobal[bufList->type].count.usedBufs[itype] += 1;

    bufList->count.usedBufs[FS_BUF_CACHE_ITYPE_TOTAL] += 1;
    fs_bufListGlobal[bufList->type].count.usedBufs[FS_BUF_CACHE_ITYPE_TOTAL] += 1;

    // If we've exceeded the limit for the new cache, we
    // must recycle and free the lru one.
    while (fs_bufListGlobal[bufList->type].count.usedBufs[itype] >
            fs_bufListGlobal[bufList->type].count.usedBufLimit[itype])
    {
        fs_Buf  *freeBuf = fs_bufFindAndFreeBuf(context, bufList->type, itype);

        if (freeBuf == NULL)
        {
            break;
        }
        
        fs_lockDestroy(context, &freeBuf->lock);
        fs_ramDealloc(FS_RAM_BUF_PTR, freeBuf);
#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_LOCK(&spinalloc);
        fs_buf_allocate_ptr[fs_buf_allocate_index] = freeBuf;
        fs_buf_allocate_oper[fs_buf_allocate_index] = FS_DEBUG_BUF_HISTORY_DEALLOC;
        fs_buf_allocate_index = (fs_buf_allocate_index + 1) % FS_DEBUG_BUF_HISTORY_ALLOC_MAX;
        NFX_SPIN_UNLOCK(&spinalloc);
#endif
    }

#ifdef FS_DEBUG_BUF_CNT_MISS_TYPE
    if (buf->inum && buf->inum < FS_MAX_META_INODE)
    {
        NFX_SPIN_LOCK(&buf->spin);
        buf->flags |= FS_BUF_INUM_CNT;
        NFX_SPIN_UNLOCK(&buf->spin);
        bufList->count.useCntType[buf->inum]++;
        fs_bufListGlobal[bufList->type].count.useCntType[buf->inum]++;
    }
#endif
}


/**
 * Increment the appropriate cache miss counts based on a
 * buf for the specified inode being missed.
 */
static void
fs_bufIncrMissCount(
    fs_BufList  *bufList,
    fs_Inode    *inode)
{
    int32 itype;

    if (bufList->type == FS_BUFLIST_USER)
        itype = FS_BUF_CACHE_ITYPE_NONDIR;
    else
        itype = fs_bufGetItypeFromInode(inode);

    bufList->count.missCount[itype] += 1;
    fs_bufListGlobal[bufList->type].count.missCount[itype] += 1;

#ifdef FS_DEBUG_BUF_CNT_MISS_TYPE
    if (inode != NULL &&
        inode->inum > 0 &&
        inode->inum < FS_MAX_META_INODE)
    {
        bufList->count.missCntType[inode->inum]++;
        fs_bufListGlobal[bufList->type].count.missCntType[inode->inum]++;
    }
#endif
}


/**
 * Update the cache stats to reflect the specified buf
 * expanding from a compact leaf form to full sized.
 */
void
fs_bufLeafExpanded(
    fs_Context  *context,
    fs_Buf      *buf)
{
    fs_BufList  *bufList = fs_bufListForBuf(buf);
    int32       itype;

    NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
    fs_bufDecrUsedCount(bufList, buf);
    
    if (buf->flags & FS_BUF_DIRTY)
    {
        itype = fs_bufGetItype(buf);
        bufList->count.dirtyBufs[itype] -= 1;
        fs_bufListGlobal[bufList->type].count.dirtyBufs[itype] -= 1;

        bufList->count.dirtyBufs[FS_BUF_CACHE_ITYPE_TOTAL] -= 1;
        fs_bufListGlobal[bufList->type].count.dirtyBufs[FS_BUF_CACHE_ITYPE_TOTAL] -= 1;
    }
    NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);

    NFX_SPIN_LOCK(&buf->spin);
    buf->flags &= ~FS_BUF_COMPACT;
#ifdef FS_DEBUG_BUF_HISTORY
    NFX_SPIN_LOCK(&buf->spinHistory);
    buf->history[buf->indexHistory] = FS_BUF_HIST_RESET_COMPACT;
    buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
    buf->flagsHistory[buf->indexHistory] = buf->flags;
    buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
    buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
    buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
    NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
    NFX_SPIN_UNLOCK(&buf->spin);

    // Update itype since we just changed the buf flags.
    itype = fs_bufGetItype(buf);

    NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
    fs_bufIncrUsedCount(context, bufList, buf);

    if (buf->flags & FS_BUF_DIRTY)
    {
        bufList->count.dirtyBufs[itype] += 1;
        fs_bufListGlobal[bufList->type].count.dirtyBufs[itype] += 1;

        bufList->count.dirtyBufs[FS_BUF_CACHE_ITYPE_TOTAL] += 1;
        fs_bufListGlobal[bufList->type].count.dirtyBufs[FS_BUF_CACHE_ITYPE_TOTAL] += 1;
    }

    NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);

    FS_DBG_INIT_SHADOW(buf);
}


/**
 * Update the cache stats to reflect the specified buf
 * shrinking from full sized to a compact leaf form.
 */
void
fs_bufLeafShrunk(
    fs_Context  *context,
    fs_Buf      *buf,
    boolean     isDir)
{
    fs_BufList  *bufList = fs_bufListForBuf(buf);
    int32 itype = fs_bufGetItype(buf);

    //printf("bufshrunk buf %p flags 0x%x itype %d isDir %d\n",
    //  buf, buf->flags, fs_bufGetItype(buf), isDir);

    NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
    fs_bufDecrUsedCount(bufList, buf);
    
    if (buf->flags & FS_BUF_DIRTY)
    {
        bufList->count.dirtyBufs[itype] -= 1;
        fs_bufListGlobal[bufList->type].count.dirtyBufs[itype] -= 1;

        bufList->count.dirtyBufs[FS_BUF_CACHE_ITYPE_TOTAL] -= 1;
        fs_bufListGlobal[bufList->type].count.dirtyBufs[FS_BUF_CACHE_ITYPE_TOTAL] -= 1;
    }
    NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);

    NFX_SPIN_LOCK(&buf->spin);
    buf->flags |= (FS_BUF_COMPACT | (isDir ? FS_BUF_DIRECTORY : 0));
#ifdef FS_DEBUG_BUF_HISTORY
    NFX_SPIN_LOCK(&buf->spinHistory);
    buf->history[buf->indexHistory] = FS_BUF_HIST_SET_COMPACT;
    buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
    buf->flagsHistory[buf->indexHistory] = buf->flags;
    buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
    buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
    buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
    NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
    NFX_SPIN_UNLOCK(&buf->spin);

    // Update itype since we just changed the buf flags.
    itype = fs_bufGetItype(buf);

    NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
    fs_bufIncrUsedCount(context, bufList, buf);

    if (buf->flags & FS_BUF_DIRTY)
    {
        bufList->count.dirtyBufs[itype] += 1;
        fs_bufListGlobal[bufList->type].count.dirtyBufs[itype] += 1;

        bufList->count.dirtyBufs[FS_BUF_CACHE_ITYPE_TOTAL] += 1;
        fs_bufListGlobal[bufList->type].count.dirtyBufs[FS_BUF_CACHE_ITYPE_TOTAL] += 1;
    }

    NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);
}


/**
 * Remove the specified buf from the list.  Caller must
 * hold EXCL lock on buf.  Data pointer or cod are not
 * touched.
 *
 * Caller must hold bufListGlobal spinlock.
 */
static void
fs_bufDelist(
    fs_Context  *context,
    fs_BufList  *bufList,
    fs_Buf      *buf,
    boolean     isForce)
{
    fs_BufListGlobal    *bufListGlobal = &fs_bufListGlobal[bufList->type];
    int32               hash, itype = fs_bufGetItype(buf);
    boolean             didCount = FALSE;

    FS_ASSERT(buf != NULL);
    FS_ASSERT(isForce || (buf->flags & FS_BUF_IO_BUSY) == 0);
    FS_ASSERT(isForce || (buf->flags & FS_BUF_DIRTY) == 0);
    FS_ASSERT(isForce || buf->zeroFillLogId == 0);
    FS_ASSERT(isForce || DLL_IS_ISOLATED(buf, pageDirtyEntry));
    FS_ASSERT(isForce || fs_lockOwnedExclusive(context, &buf->lock));
    ASSERT_LOCK_OWNED(&fs_bufListGlobal[bufList->type].spin);

    if (buf->cleanupFunction != NULL)
    {
        buf->cleanupFunction(context, buf);
        buf->cleanupFunction = NULL;
    }

    if (!DLL_IS_ISOLATED(buf, blkNumEntry))
    {
        FS_ASSERT(fs_blockIsValid(buf->block));
        hash = fs_bufHashVolBlock(buf->fileSystem->volId, buf->block);
        DLL_REMOVE(&bufListGlobal->bufsBlk[hash], buf, blkNumEntry);
#ifdef FS_BUF_BLK_CHECK_HASH_DEPTH
        bufListGlobal->blkDepth[hash]--;
#endif

        if (!didCount)
        {
            fs_bufDecrUsedCount(bufList, buf);
            didCount = TRUE;
        }
    }

    if (!DLL_IS_ISOLATED(buf, pageHashEntry))
    {
        hash = fs_bufHashVolInumPage(buf->fileSystem->volId, buf->inum,
                                     buf->page);
        DLL_REMOVE(&bufListGlobal->bufsPage[hash], buf, pageHashEntry);
#ifdef FS_BUF_PAGE_CHECK_HASH_DEPTH
        bufListGlobal->pageDepth[hash]--;
#endif

        if (!didCount)
        {
            fs_bufDecrUsedCount(bufList, buf);
            didCount = TRUE;
        }
    }

    if (!DLL_IS_ISOLATED(buf, mruEntry))
    {
        DLL_REMOVE(&fs_bufListGlobal[bufList->type].mruHdr[itype], buf, mruEntry);
        fs_bufListGlobal[bufList->type].count.freeBufs[itype] -= 1;

        if (!didCount)
        {
            fs_bufDecrUsedCount(bufList, buf);
            didCount = TRUE;
        }
    }

    if (!DLL_IS_ISOLATED(buf, pageEntry))
    {
        fs_BufPageHdr       *pageHdr;

        pageHdr = fs_bufPageHdrForInum(buf->fileSystem,
                                       bufList->type, buf->inum);
        FS_ASSERT(pageHdr != NULL);

        DLL_REMOVE(&pageHdr->bufChainHdr, buf, pageEntry);
        if (isForce && !DLL_IS_ISOLATED(buf, pageDirtyEntry))
        {
            DLL_REMOVE(&pageHdr->bufDirtyHdr, buf, pageDirtyEntry);
            pageHdr->dirtyCount -= 1;
#ifdef FS_DEBUG_BUF_HISTORY
            NFX_SPIN_LOCK(&buf->spinHistory);
            buf->history[buf->indexHistory] = FS_BUF_HIST_REMOVE_PAGE;
            buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
            buf->flagsHistory[buf->indexHistory] = buf->flags;
            buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
            buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
            buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
            NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
        }

        if (DLL_IS_EMPTY(&pageHdr->bufChainHdr) &&
            DLL_IS_EMPTY(&pageHdr->bufDirtyHdr))
        {
            int32   iHash = fs_bufHashVolInum(buf->fileSystem->volId,
                                              buf->inum);
            // No more bufs for this inode,
            // remove the pageHdr from the
            // inum hash chain.
            DLL_REMOVE(&fs_bufListGlobal[bufList->type].bufsInum[iHash],
                       pageHdr,
                       inumEntry);
#ifdef FS_BUF_INUM_CHECK_HASH_DEPTH
            fs_bufListGlobal[bufList->type].inumDepth[iHash]--;
#endif
            FS_ASSERT(isForce || pageHdr->dirtyCount == 0);
            FS_ASSERT(isForce || DLL_IS_EMPTY(&pageHdr->waiterHdr));
            fs_ramDealloc(FS_RAM_BUF_PAGE_HDR, pageHdr);
        }
    }

#ifndef NEW_ZF
    if (buf->flags & FS_BUF_TYPE_USER)
    {
        if (!DLL_IS_ISOLATED(buf, zeroFillEntry))
        {
            DLL_REMOVE(&buf->fileSystem->logRoot->replay->zeroFillHdr,
                       buf, zeroFillEntry);
        }
    }
#endif // NEW_ZF

} /* fs_bufDelist */


#if defined(FS_DEBUG_BUF_CKSUM)

fs_BufCksumFreePool fs_bufCksumFreePool;
boolean             fs_bufCksumInitialized = FALSE;


/**
 * Compute and store (in-core only) a checksum over
 * the meta buf.
 */
extern void
fs_bufComputeCksum(
    fs_Context  *context,
    fs_Block    block,
    void        *dataPtr)
{
    fs_BufCksum *bufCksum;
    uint32      loop;
    uint64      cksum = fs_coreHash(dataPtr, FS_BLOCK_SIZE /
                                                sizeof(uint64), 0);
    int32       hash = fs_bufHashVolBlock(context->volId, block);

    if (! fs_bufCksumInitialized)
    { 
        NFX_SPIN_INIT(&fs_bufCksumFreePool.spin,SPINLOCK_LEVEL_UNKNOWN);
        NFX_SPIN_LOCK(&fs_bufCksumFreePool.spin);
        
        DLL_INIT(&(fs_bufCksumFreePool.freeList));
        fs_bufCksumFreePool.nbrOfMemBlocks = 0;
        fs_bufCksumFreePool.nbrOnFreeList = 0;
        fs_bufCksumFreePool.nbrInUse = 0 ;
        
        fs_bufCksumInitialized = TRUE; 

        NFX_SPIN_UNLOCK(&fs_bufCksumFreePool.spin);
    }
    
    DLL_ITERATE(bufCksum,
                &context->fileSystem->bufCksumHdr[hash], 
                dllEntry)
    {
        if (fs_blockEq(bufCksum->block, block))
            break;
    }

    if (bufCksum == DLL_NIL)
    {
        NFX_SPIN_LOCK(&fs_bufCksumFreePool.spin);

        if (fs_bufCksumFreePool.nbrOnFreeList > 0)
        {
            // use one from the free list
            DLL_REMOVE_HEAD(&fs_bufCksumFreePool.freeList,
                            bufCksum,
                            dllEntry);
            fs_bufCksumFreePool.nbrOnFreeList--;
            FS_ASSERT(fs_bufCksumFreePool.nbrOnFreeList >= 0);
            fs_bufCksumFreePool.nbrInUse++;
        }
        else
        {
            if (fs_bufCksumFreePool.nbrOfMemBlocks < FS_BUF_CKSUM_MAXNBR_OF_BLOCKS)
            {
                fs_bufCksumFreePool.nbrOfMemBlocks++;

                bufCksum = fs_ramAlloc(FS_RAM_CORE_MISC, 
                                       FS_BUF_CKSUM_ENTRIES_PER_MEM_BLOCK * sizeof(fs_BufCksum));
                
                for (loop = 0; loop < FS_BUF_CKSUM_ENTRIES_PER_MEM_BLOCK; loop++)
                {
                    DLL_ISOLATE_ELEM(bufCksum, dllEntry);
                    DLL_INSERT_TAIL(&fs_bufCksumFreePool.freeList,
                                    bufCksum,
                                    dllEntry);
                    fs_bufCksumFreePool.nbrOnFreeList++;
                    bufCksum++;
                }
                
                // use one from the free list
                DLL_REMOVE_HEAD(&fs_bufCksumFreePool.freeList,
                                bufCksum,
                                dllEntry);
                fs_bufCksumFreePool.nbrOnFreeList--;
                FS_ASSERT(fs_bufCksumFreePool.nbrOnFreeList >= 0);
                fs_bufCksumFreePool.nbrInUse++;
            }
            else
            {
                // not longer allowed to allocate memory for this debugging tool
                NFX_SPIN_UNLOCK(&fs_bufCksumFreePool.spin);
                return;
            }
        }
        
        NFX_SPIN_UNLOCK(&fs_bufCksumFreePool.spin);
        
        bufCksum->block = block;
        bufCksum->lastWriteTime = 0;
        NFX_SPIN_LOCK(&context->fileSystem->spin);
        DLL_INSERT_HEAD(&context->fileSystem->bufCksumHdr[hash],
                        bufCksum,
                        dllEntry);
        NFX_SPIN_UNLOCK(&context->fileSystem->spin);
    }

    // Write-busy bit should be in reset state.
    FS_ASSERT(!(bufCksum->lastWriteTime & 0x80000000));

    bufCksum->cksum = cksum;
    bufCksum->lastWriteTime = fs_getTimeOfDaySecs();

    // Set write-busy bit.
    bufCksum->lastWriteTime |= 0x80000000;
}


/**
 * Compare the previously stored (in-core) cksum,
 * if one exists, to the just read meta buf.
 */
extern void
fs_bufCompareCksum(
    fs_Context  *context,
    fs_Block    block,
    void        *dataPtr,
    fs_Buf      *buf,
    char        *msg,
    boolean     clearBusy)
{
    fs_BufCksum *bufCksum;
    int32   hash = fs_bufHashVolBlock(context->volId, block);

    ///////////////
    //
    // Careful, buf may be NULL (eg. when called from writeSuperSync).
    //
    //////////////

    if (! fs_bufCksumInitialized)
    {
        return;
    }
    
    DLL_ITERATE(bufCksum,
                &context->fileSystem->bufCksumHdr[hash], 
                dllEntry)
    {
        if (fs_blockEq(bufCksum->block, block))
            break;
    }

    if (bufCksum != DLL_NIL)
    {
        uint64  cksum = fs_coreHash(dataPtr, FS_BLOCK_SIZE /
                                               sizeof(uint64), 0);

        if (clearBusy)
        {
            // Clear the write-busy bit.
            FS_ASSERT(bufCksum->lastWriteTime & 0x80000000);
            bufCksum->lastWriteTime &= ~0x80000000;
        }

        // State better be non-write-busy.
        FS_ASSERT(!(bufCksum->lastWriteTime & 0x80000000));
        
        if (cksum != bufCksum->cksum)
        {
            printf(
             "bufCksum error %s (0x%llx/0x%llx) blknum %lld ptr %p buf %p\n",
                   msg, cksum, bufCksum->cksum, fs_blockNumber(block),
                   dataPtr, buf);

            FS_ASSERT(FALSE);
        }
    }
}


/**
 * Return TRUE if an in-core cksum exists for the
 * block specified in buf, FALSE otherwise.
 */
static boolean
fs_bufFindCksum(
    fs_Context  *context,
    fs_Buf      *buf)
{
    fs_BufCksum *bufCksum;
    int32   hash = fs_bufHashVolBlock(context->volId, buf->block);

    if (! fs_bufCksumInitialized)
    {
        return FALSE;
    }
   
    DLL_ITERATE(bufCksum,
                &context->fileSystem->bufCksumHdr[hash],
                dllEntry)
    {
        if (fs_blockEq(bufCksum->block, buf->block))
        {
            return TRUE;
        }
    }

    return FALSE;
}


/**
 * Release bufCksum struct for buf if it exists.
 */
static void
fs_bufReleaseCksum(
    fs_Context  *context,
    fs_Buf      *buf)
{
    fs_BufCksum *bufCksum;
    int32   hash = fs_bufHashVolBlock(context->volId, buf->block);

    if (! fs_bufCksumInitialized)
    {
        return;
    }
    
    DLL_ITERATE(bufCksum,
                &context->fileSystem->bufCksumHdr[hash], 
                dllEntry)
    {
        if (fs_blockEq(bufCksum->block, buf->block))
            break;
    }

    if (bufCksum != DLL_NIL)
    {
        NFX_SPIN_LOCK(&context->fileSystem->spin);
        DLL_REMOVE(&context->fileSystem->bufCksumHdr[hash],
                   bufCksum,
                   dllEntry);
        context->fileSystem->bufCksumNbrEntries[hash]--;
        FS_ASSERT(context->fileSystem->bufCksumNbrEntries[hash] >= 0);
        NFX_SPIN_UNLOCK(&context->fileSystem->spin);

        NFX_SPIN_LOCK(&fs_bufCksumFreePool.spin);
        DLL_INSERT_TAIL(&fs_bufCksumFreePool.freeList,
                   bufCksum,
                   dllEntry);
        fs_bufCksumFreePool.nbrOnFreeList++;
        fs_bufCksumFreePool.nbrInUse--;
        FS_ASSERT(fs_bufCksumFreePool.nbrInUse >= 0);
        NFX_SPIN_UNLOCK(&fs_bufCksumFreePool.spin);
    }
}

/*
 * Deallocation of extents should call this to delete any cached checksum
 * for the blocks
 */

EXPORTED void
fs_bufReleaseCksumForBlock(
    fs_Context  *context,
    fs_Block    block)
{
    fs_BufCksum *bufCksum;
    int32   hash = fs_bufHashVolBlock(context->volId, block);

    if (! fs_bufCksumInitialized)
    {
        return;
    }
    
    DLL_ITERATE(bufCksum,
                &context->fileSystem->bufCksumHdr[hash], 
                dllEntry)
    {
        if (fs_blockEq(bufCksum->block, block))
            break;
    }

    if (bufCksum != DLL_NIL)
    {
        NFX_SPIN_LOCK(&context->fileSystem->spin);
        DLL_REMOVE(&context->fileSystem->bufCksumHdr[hash],
                   bufCksum,
                   dllEntry);
        context->fileSystem->bufCksumNbrEntries[hash]--;
        FS_ASSERT(context->fileSystem->bufCksumNbrEntries[hash] >= 0);
        NFX_SPIN_UNLOCK(&context->fileSystem->spin);

        NFX_SPIN_LOCK(&fs_bufCksumFreePool.spin);
        DLL_INSERT_TAIL(&fs_bufCksumFreePool.freeList,
                   bufCksum,
                   dllEntry);
        fs_bufCksumFreePool.nbrOnFreeList++;
        fs_bufCksumFreePool.nbrInUse--;
        FS_ASSERT(fs_bufCksumFreePool.nbrInUse >= 0);
        NFX_SPIN_UNLOCK(&fs_bufCksumFreePool.spin);
    }
}

/**
 * Release all bufCksum struct for given filesystem
 */
extern void
fs_bufReleaseAllCksum(
    fs_Context  *context)
{
    fs_BufCksum *bufCksum, *bufCksumNext;
    int32       hash;

    if (! fs_bufCksumInitialized)
    {
        return;
    }
    
    for (hash = 0; hash < FS_BUFLIST_HASH_MAX; hash++)
    {
        DLL_ITERATE_NEXT(bufCksum, bufCksumNext,
                         &context->fileSystem->bufCksumHdr[hash],
                         dllEntry)
            {
                NFX_SPIN_LOCK(&context->fileSystem->spin);
                DLL_REMOVE(&context->fileSystem->bufCksumHdr[hash],
                           bufCksum,
                           dllEntry);
                context->fileSystem->bufCksumNbrEntries[hash]--;
                FS_ASSERT(context->fileSystem->bufCksumNbrEntries[hash] >= 0);
                NFX_SPIN_UNLOCK(&context->fileSystem->spin);

                NFX_SPIN_LOCK(&fs_bufCksumFreePool.spin);
                DLL_INSERT_TAIL(&fs_bufCksumFreePool.freeList,
                                bufCksum,
                                dllEntry);
                fs_bufCksumFreePool.nbrOnFreeList++;
                fs_bufCksumFreePool.nbrInUse--;
                FS_ASSERT(fs_bufCksumFreePool.nbrInUse >= 0);
                NFX_SPIN_UNLOCK(&fs_bufCksumFreePool.spin);
            }
    }
}

#endif // FS_DEBUG_BUF_CKSUM


/**
 * Return TRUE if the specified buf is available for
 * recycling.
 *
 * Must be called with the bufListGlobal spinlock held.
 */
static boolean
fs_bufIsAvail(
    fs_Buf      *buf)
{
    const int INUSE_FLAGS = FS_BUF_DIRTY | FS_BUF_IO_BUSY | 
                             FS_BUF_FREE_DEFER | FS_BUF_FREE_LOGBUF;

    return ((buf->flags & INUSE_FLAGS) == 0);
}


/**
 * Scan the specified bufList for an available buffer;  that is, one
 * which can be reallocated for use for a new buffer.
 *
 * Must be called with the bufListGlobal spinlock held.
 *
 * Returns NULL or an available buf fs_lockOwnedExclusive
 */
static fs_Buf *
fs_bufFindAvailable(
    fs_Context      *context,
    int32           listType,
    int32           itype)
{
    fs_BufList  *bufList;
    fs_Buf      *buf;
    boolean     acquired;

    ASSERT_LOCK_OWNED(&fs_bufListGlobal[listType].spin);

    // Just pop the next one off the free list.
    if (!DLL_IS_EMPTY(&fs_bufListGlobal[listType].mruHdr[itype]))
    {
        DLL_REMOVE_TAIL(&fs_bufListGlobal[listType].mruHdr[itype], buf, mruEntry);
        fs_bufListGlobal[listType].count.freeBufs[itype] -= 1;

        FS_ASSERT(!fs_lockHeldShare(context, &buf->lock));
        FS_ASSERT(!fs_lockHeldExclusive(context, &buf->lock));
        FS_ASSERT(fs_bufIsAvail(buf));
        FS_ASSERT(buf->logSegDep.count == 0);

        if (DLL_IS_ISOLATED(buf, blkNumEntry) &&
            DLL_IS_ISOLATED(buf, pageHashEntry))
        {
            // Buf not on any list now, decrement
            // usedCount.
            bufList = fs_bufListForBuf(buf);
            fs_bufDecrUsedCount(bufList, buf);
        }

        // We should not be blocked getting the lock since
        // we hold the cache spinlock and fs_bufIsAvailable()
        // returned TRUE.
        acquired = fs_trylockExclusive(context, &buf->lock);
        FS_ASSERT(acquired);

        fs_bufDeallocCoreRes(context, buf);

        return buf;
    }
    
    return NULL;
}


#ifdef NOTDEF // not currently used
void
fs_dumpBufferCacheStats(
    fs_Context      *context)
{
    int32 itype;

    printf("\nType            Max        Used       DirtyLimit  Dirty        Hits      Misses         COW\n");
    printf("-------------------------------------------------------------------------------\n");

    itype = FS_BUF_CACHE_ITYPE_NONDIR;
    printf("8knondir %10d  %10d  %10d  %10d  %10d  %10d  %10d\n",
           fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimit[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.usedBufs[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufLimit[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufs[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.hitCount[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.missCount[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.cowCount[itype]);

    itype = FS_BUF_CACHE_ITYPE_COMP_NONDIR;
    printf("1knondir %10d  %10d  %10d  %10d  %10d  %10d  %10d\n",
           fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimit[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.usedBufs[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufLimit[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufs[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.hitCount[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.missCount[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.cowCount[itype]);

    itype = FS_BUF_CACHE_ITYPE_DIR;
    printf("8kdir    %10d  %10d  %10d  %10d  %10d  %10d  %10d\n",
           fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimit[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.usedBufs[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufLimit[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufs[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.hitCount[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.missCount[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.cowCount[itype]);

    itype = FS_BUF_CACHE_ITYPE_COMP_DIR;
    printf("1kdir    %10d  %10d  %10d  %10d  %10d  %10d  %10d\n",
           fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimit[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.usedBufs[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufLimit[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufs[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.hitCount[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.missCount[itype],
           fs_bufListGlobal[FS_BUFLIST_META].count.cowCount[itype]);

    itype = FS_BUF_CACHE_ITYPE_NONDIR;
    printf("user     %10d  %10d  %10d  %10d  %10d  %10d  %10d\n",
           fs_bufListGlobal[FS_BUFLIST_USER].count.usedBufLimit[itype],
           fs_bufListGlobal[FS_BUFLIST_USER].count.usedBufs[itype],
           fs_bufListGlobal[FS_BUFLIST_USER].count.dirtyBufLimit[itype],
           fs_bufListGlobal[FS_BUFLIST_USER].count.dirtyBufs[itype],
           fs_bufListGlobal[FS_BUFLIST_USER].count.hitCount[itype],
           fs_bufListGlobal[FS_BUFLIST_USER].count.missCount[itype],
           fs_bufListGlobal[FS_BUFLIST_USER].count.cowCount[itype]);

};
#endif // NOTDEF


/**
 * Allocate a new buf ptr or reclaim the LRU one.
 * Returns buf with EXCL lock held
 *
 * Caller must hold bufListGlobal spinlock.
 */

static fs_Buf *
fs_bufFindAndFreeBuf(
    fs_Context  *context,
    int32       listType,
    int32       itype)
{
    fs_Buf              *buf;

    ASSERT_LOCK_OWNED(&fs_bufListGlobal[listType].spin);

    buf = fs_bufFindAvailable(context, listType, itype);
    if (buf == NULL)
    {
        // Caller must handle this case.
        return NULL;
    }

    FS_ASSERT(buf->logSegDep.count == 0);

    BUF_TRACE(
        ("vol %llx evicting blknum %lld inum %lld page %lld from cache",
         buf->fileSystem->volId.whole_thing,
         fs_blockNumber(buf->block), buf->inum, buf->page));

    FS_ASSERT(fs_lockOwnedExclusive(context, &buf->lock));
    fs_bufDelist(context, &buf->fileSystem->bufList[listType], buf, FALSE);

    return buf;

}

/**
 * free resource for given listype and itype
 *
 * Caller must hold bufListGlobal spinlock.
 * spinlock is per listType: 
 * - user buf can only free user bufs
 * - meta buf can only free meta bufs
 */
static void
fs_bufFreeResource(
    fs_Context          *context,
    int32               listType,
    int32               itype,
    int32               reduceCount)
{
#ifdef FS_DYN_CACHE_SIZE
    fs_BufListGlobal    *bufListGlobal = &fs_bufListGlobal[listType];
    fs_Buf              *freebuf;
    int32               count;

    ASSERT_LOCK_OWNED(&fs_bufListGlobal[listType].spin);

    // the resource usage is over the limit, , 
    // knock some out
    // reduce the used buf limit,
    // reuse a freed buf (if available)

    for (count = 0 ; count < reduceCount; count++)
    {
        if (bufListGlobal->count.usedBufLimit[itype] <= bufListGlobal->count.usedBufLimitLWM[itype])
        {
            // don't drop below this level 
            if (count == 0)
                printf ("fs_bufFreeResource don't drop below usedBufLimitLWM (%d %d %d %d %d)\n", 
                        listType, itype, 
                        bufListGlobal->count.usedBufs[itype], 
                        bufListGlobal->count.usedBufLimit[itype], 
                        bufListGlobal->count.usedBufLimitLWM[itype]) ;
            return;
        }

        freebuf = fs_bufFindAndFreeBuf(context, listType, itype);
   
        if (freebuf == NULL)
        {
            printf ("fs_bufFreeResource could not fs_bufFindAndFreeBuf (%d %d %d %d %d)\n",
                    listType, itype, 
                    bufListGlobal->count.usedBufs[itype], 
                    bufListGlobal->count.usedBufLimit[itype], 
                    bufListGlobal->count.usedBufLimitLWM[itype]) ;
            return;
        }
        
        bufListGlobal->count.usedBufLimit[itype]--;

        // todo decr stats (global and per fs (?)) for this type
   
        fs_lockDestroy(context, &freebuf->lock);
        fs_ramDealloc(FS_RAM_BUF_PTR, freebuf);
#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_LOCK(&spinalloc);
        fs_buf_allocate_ptr[fs_buf_allocate_index] = freebuf;
        fs_buf_allocate_oper[fs_buf_allocate_index] = FS_DEBUG_BUF_HISTORY_DEALLOC;
        fs_buf_allocate_index = (fs_buf_allocate_index + 1) % FS_DEBUG_BUF_HISTORY_ALLOC_MAX;
        NFX_SPIN_UNLOCK(&spinalloc);
#endif
    }
#endif //  FS_DYN_CACHE_SIZE
}

/**
 * find who consumes most memory in the cache
 * and free the resources
 *
 * Caller indicates which spinlock it owns.
 */
void
fs_bufFindAndFreeResource(
    fs_Context          *context,
    int                 haveMetaSpinLock,
    int                 haveUserSpinLock,
    int                 haveInodeSpinLock,
    int32               reduceCount)
{
#ifdef FS_DYN_CACHE_SIZE
    int32               listType = FS_BUFLIST_META;
    int32               itype = FS_BUF_CACHE_ITYPE_NONDIR;
    int32               count;
    int32               count_max = 0;
    int                 gotUserSpinLock = FALSE;
    int                 gotMetaSpinLock = FALSE;
    int                 gotInodeSpinLock = FALSE;    
    
    // first try to cleanup (fragmentation) the cache itself
    cleanup_caches();

    if (((uint32)(eee_ramTotalFlexAvail() / 1024)) > fs_incoreTune->memMinAvailable)
    {
        // cleanup_caches, job well done, no further cleanup needed.
        return;
    }

    if ( ! haveMetaSpinLock)
    {
        if (NFX_SPIN_TRYLOCK(&fs_bufListGlobal[FS_BUFLIST_META].spin))
        {
            gotMetaSpinLock = TRUE;
        }
    }
    if (haveMetaSpinLock || gotMetaSpinLock)
    {
        count = fs_bufListGlobal[listType].count.usedBufs[FS_BUF_CACHE_ITYPE_NONDIR] - 
            fs_bufListGlobal[listType].count.usedBufLimitLWM[FS_BUF_CACHE_ITYPE_NONDIR];
        if (count > count_max)
        {
            itype = FS_BUF_CACHE_ITYPE_NONDIR;
            count_max = count;
        }
    
        count = fs_bufListGlobal[listType].count.usedBufs[FS_BUF_CACHE_ITYPE_DIR] - 
            fs_bufListGlobal[listType].count.usedBufLimitLWM[FS_BUF_CACHE_ITYPE_DIR];
        if (count > count_max)
        {
            itype = FS_BUF_CACHE_ITYPE_DIR;
            count_max = count;
        }
    
        count = fs_bufListGlobal[listType].count.usedBufs[FS_BUF_CACHE_ITYPE_COMP_NONDIR] - 
            fs_bufListGlobal[listType].count.usedBufLimitLWM[FS_BUF_CACHE_ITYPE_COMP_NONDIR];
        if (count > count_max)
        {
            itype = FS_BUF_CACHE_ITYPE_COMP_NONDIR;
            count_max = count;
        }
    
        count = fs_bufListGlobal[listType].count.usedBufs[FS_BUF_CACHE_ITYPE_COMP_DIR] - 
            fs_bufListGlobal[listType].count.usedBufLimitLWM[FS_BUF_CACHE_ITYPE_COMP_DIR];
        if (count > count_max)
        {
            itype = FS_BUF_CACHE_ITYPE_COMP_DIR;
            count_max = count;
        }
    }
    
    if ( ! haveUserSpinLock)
    {
        if (NFX_SPIN_TRYLOCK(&fs_bufListGlobal[FS_BUFLIST_USER].spin))
        {
            gotUserSpinLock = TRUE;
        }
    }
    if (haveUserSpinLock || gotUserSpinLock)
    {
        count = fs_bufListGlobal[FS_BUFLIST_USER].count.usedBufs[FS_BUF_CACHE_ITYPE_NONDIR] - 
            fs_bufListGlobal[FS_BUFLIST_USER].count.usedBufLimitLWM[FS_BUF_CACHE_ITYPE_NONDIR];
        if (count > count_max)
        {
            listType = FS_BUFLIST_USER;
            itype = FS_BUF_CACHE_ITYPE_NONDIR;
            count_max = count;
        }    
    }

    if ( ! haveInodeSpinLock)
    {
        if (NFX_SPIN_TRYLOCK(&fs_inodeCache->spin))
        {
            gotInodeSpinLock = TRUE;
        }
    }
    if (haveInodeSpinLock || gotInodeSpinLock)
    {        
        count = fs_inodeCache->count.numCached - fs_inodeCache->count.numCachedLimitLWM;
        if (count > count_max)
        {
            listType = -1;  // indicating inode 
            count_max = count;
        }
    }

    if (count_max > 0)
    {
        if (listType == FS_BUFLIST_META)
        {
            fs_bufFreeResource (context, listType, itype, reduceCount);
        }
        else 
        if (listType == FS_BUFLIST_USER)
        {
            fs_bufFreeResource (context, listType, itype, reduceCount);
        }
        else
        {
            fs_inodeFreeResource(context, reduceCount);
        }
    }
    else
    {
        if (((uint32)(eee_ramTotalFlexAvail() / 1024)) < (fs_incoreTune->memMinAvailable / 2))
        {
            // getting verry close to running out of mem

            printf ("fs_bufFindAndFreeResource failed to free resources (%lld B %d KB)\n",
                    eee_ramTotalFlexAvail(), fs_incoreTune->memMinAvailable);
            printf (" meta lock  %d %d nondir %d %d dir %d %d compact nondir %d %d dir %d %d\n", 
                    haveMetaSpinLock, gotMetaSpinLock,
                    fs_bufListGlobal[FS_BUFLIST_META].count.usedBufs[FS_BUF_CACHE_ITYPE_NONDIR],
                    fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimitLWM[FS_BUF_CACHE_ITYPE_NONDIR],
                    fs_bufListGlobal[FS_BUFLIST_META].count.usedBufs[FS_BUF_CACHE_ITYPE_DIR],
                    fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimitLWM[FS_BUF_CACHE_ITYPE_DIR],
                    fs_bufListGlobal[FS_BUFLIST_META].count.usedBufs[FS_BUF_CACHE_ITYPE_COMP_NONDIR],
                    fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimitLWM[FS_BUF_CACHE_ITYPE_COMP_NONDIR],
                    fs_bufListGlobal[FS_BUFLIST_META].count.usedBufs[FS_BUF_CACHE_ITYPE_COMP_DIR],
                    fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimitLWM[FS_BUF_CACHE_ITYPE_COMP_DIR]) ;
            printf (" user lock  %d %d used %d %d \n", 
                    haveUserSpinLock, gotUserSpinLock, 
                    fs_bufListGlobal[FS_BUFLIST_USER].count.usedBufs[FS_BUF_CACHE_ITYPE_NONDIR],
                    fs_bufListGlobal[FS_BUFLIST_USER].count.usedBufLimitLWM[FS_BUF_CACHE_ITYPE_NONDIR]);
            printf (" inode lock %d %d used %d %d \n", 
                    haveInodeSpinLock, gotInodeSpinLock,
                    fs_inodeCache->count.numCached,
                    fs_inodeCache->count.numCachedLimitLWM);
        }
        
        //FS_ASSERT(FALSE);
    }
    
    if (gotUserSpinLock)
    {
        NFX_SPIN_UNLOCK(&fs_bufListGlobal[FS_BUFLIST_USER].spin);
    }
    if (gotMetaSpinLock)
    {
        NFX_SPIN_UNLOCK(&fs_bufListGlobal[FS_BUFLIST_META].spin);
    }
    if (gotInodeSpinLock)
    {
        NFX_SPIN_UNLOCK(&fs_inodeCache->spin);
    }
#endif // FS_DYN_CACHE_SIZE
}

/**
 * find who consumes most ultra buffers
 * and free the resources
 *
 * Caller owns the META spin lock.
 * only valid for (itype == FS_BUF_CACHE_ITYPE_NONDIR) ||
 *                (itype == FS_BUF_CACHE_ITYPE_DIR)
 */
static void
fs_bufFindAndFreeUltraBufResource(
    fs_Context          *context,
    int32               reduceCount)
{
#ifdef FS_DYN_CACHE_SIZE
    int32               listType = FS_BUFLIST_META;
    int32               itype = FS_BUF_CACHE_ITYPE_NONDIR;
    int32               count;
    int32               count_max = 0;
    
    ASSERT_LOCK_OWNED(&fs_bufListGlobal[listType].spin);

    count = fs_bufListGlobal[listType].count.usedBufs[FS_BUF_CACHE_ITYPE_NONDIR] - 
        fs_bufListGlobal[listType].count.usedBufLimitLWM[FS_BUF_CACHE_ITYPE_NONDIR];
    if (count > count_max)
    {
        itype = FS_BUF_CACHE_ITYPE_NONDIR;
        count_max = count;
    }
    
    count = fs_bufListGlobal[listType].count.usedBufs[FS_BUF_CACHE_ITYPE_DIR] - 
        fs_bufListGlobal[listType].count.usedBufLimitLWM[FS_BUF_CACHE_ITYPE_DIR];
    if (count > count_max)
    {
        itype = FS_BUF_CACHE_ITYPE_DIR;
        count_max = count;
    }
    
    if (count_max > 0)
    {
        fs_bufFreeResource (context, listType, itype, reduceCount);
    }
    else
    {
        printf ("fs_bufFindAndFreeUltraBufResource failed to free resources\n") ;
        printf("  count %d nondir used %d limit %d dir used %d limit %d reduceCount %d\n",
                count, 
                fs_bufListGlobal[FS_BUFLIST_META].count.usedBufs[FS_BUF_CACHE_ITYPE_NONDIR],
                fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimitLWM[FS_BUF_CACHE_ITYPE_NONDIR],
                fs_bufListGlobal[FS_BUFLIST_META].count.usedBufs[FS_BUF_CACHE_ITYPE_DIR],
                fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimitLWM[FS_BUF_CACHE_ITYPE_DIR],
                reduceCount);
    }
#endif // FS_DYN_CACHE_SIZE
}


/**
 * Allocate or reuse a buffer
 * Check the resources and free them if needed
 * Adjust the usedBufLimit if possible/needed
 *
 * Caller must hold bufListGlobal spinlock.
 *
 * Returns with the buf owned exclusive
 *  or NULL if not found
 */
static fs_Buf *
fs_bufAllocPtr(
    fs_Context          *context,
    int32               listType,
    fs_Inode            *inode,
    uint64              page)
{
    fs_BufListGlobal    *bufListGlobal = &fs_bufListGlobal[listType];
    fs_Buf              *buf;
    int32               itype;
#ifdef FS_DYN_CACHE_SIZE
    boolean             resourceAvail = TRUE;
#endif

    ASSERT_LOCK_OWNED(&fs_bufListGlobal[listType].spin);

    if (listType == FS_BUFLIST_USER)
    {
        itype = FS_BUF_CACHE_ITYPE_NONDIR;
#ifdef FS_DYN_CACHE_SIZE
        if (sbm_sac_info->num_buf_large_avail <= fs_incoreTune->sacMinAvailable)
        {
            fs_bufFreeResource (context, listType, itype, fs_incoreTune->sacFreeCount);
            
            resourceAvail = FALSE;
        }
        
#endif // FS_DYN_CACHE_SIZE
    }
    else
    {
        itype = fs_bufGetItypeFromInode(inode);
#ifdef FS_DYN_CACHE_SIZE
		if ((itype == FS_BUF_CACHE_ITYPE_NONDIR) ||
			(itype == FS_BUF_CACHE_ITYPE_DIR))
		{
			if (eee_getAvailableNbrOfBufs(EEE_BUF_ULTRA) < fs_incoreTune->ultraMinAvailable)
			{
                fs_bufFindAndFreeUltraBufResource (context, fs_incoreTune->bufFreeCount);

				resourceAvail = FALSE;
			}
		}

        if (((uint32)(eee_ramTotalFlexAvail() / 1024)) < fs_incoreTune->memMinAvailable)
        {
            // listType == FS_BUFLIST_META: I have the meta buffer list spinlock 
            fs_bufFindAndFreeResource (context, TRUE, FALSE, FALSE, fs_incoreTune->bufFreeCount);

            resourceAvail = FALSE;
        }
#endif // FS_DYN_CACHE_SIZE
    }
    
#ifdef FS_DYN_CACHE_SIZE
    if (resourceAvail)
    {
        if (bufListGlobal->count.usedBufs[itype] >=
                     bufListGlobal->count.usedBufLimit[itype])
        {
            // OK, increase the used buf limit.
            bufListGlobal->count.usedBufLimit[itype] =
                MIN((bufListGlobal->count.usedBufs[itype] + 1),
                     bufListGlobal->count.usedBufLimitHWM[itype]);

            // todo incr stats for this type
        }
    }
    else
    {
        if (bufListGlobal->count.usedBufs[itype] <
                     bufListGlobal->count.usedBufLimit[itype])
        {
            // OK, decrease the used buf limit.
            bufListGlobal->count.usedBufLimit[itype] =
                MAX((bufListGlobal->count.usedBufs[itype]),
                     bufListGlobal->count.usedBufLimitLWM[itype]);

            // todo incr stats for this type
        }
    }    
#endif // FS_DYN_CACHE_SIZE

    if (bufListGlobal->count.usedBufs[itype] < bufListGlobal->count.usedBufLimit[itype])
    {
        buf = fs_ramAlloc(FS_RAM_BUF_PTR, sizeof(fs_Buf));
#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_LOCK(&spinalloc);
        fs_buf_allocate_ptr[fs_buf_allocate_index] = buf;
        fs_buf_allocate_oper[fs_buf_allocate_index] = FS_DEBUG_BUF_HISTORY_ALLOC;
        fs_buf_allocate_index = (fs_buf_allocate_index + 1) % FS_DEBUG_BUF_HISTORY_ALLOC_MAX;
        NFX_SPIN_UNLOCK(&spinalloc);
#endif
        bzero(buf, sizeof(fs_Buf));
        fs_lockInit(&buf->lock, FS_LOCK_TYPE_BUF);
        NFX_SPIN_INIT(&buf->spin, SPINLOCK_LEVEL_UNKNOWN);
#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_INIT(&buf->spinHistory, SPINLOCK_LEVEL_UNKNOWN);
#endif
        DLL_INIT(&buf->undoHdr);
        DLL_INIT(&buf->candoHdr);
        // Obtain exclusive lock on buf.
        fs_lockExclusive(context, &buf->lock, NULL);

#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_LOCK(&buf->spinHistory);
        buf->history[buf->indexHistory] = FS_BUF_HIST_ALLOC;
        buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
        buf->flagsHistory[buf->indexHistory] = buf->flags;
        buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
        buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
        buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
        NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif

    }
    else
    {
        // This cache has reached its limit, knock something
        // out to make room for the new buf.
        // case of usedBufLimitHWM
        buf = fs_bufFindAndFreeBuf(context, listType, itype);

#ifdef FS_DEBUG_BUF_HISTORY
        if (buf)
        {
            NFX_SPIN_LOCK(&buf->spinHistory);
            buf->history[buf->indexHistory] = FS_BUF_HIST_ALLOC_REUSE;
            buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
            buf->flagsHistory[buf->indexHistory] = buf->flags;
            buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
            buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
            buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
            NFX_SPIN_UNLOCK(&buf->spinHistory);
        }
#endif
    }

    if (buf == NULL)
        return NULL;

    buf->fileSystem = context->fileSystem;
    buf->inum = (inode == NULL) ? 0 : inode->inum; // for lockorder checking
    buf->page = page; // for lockorder checking
    buf->flags = 0;
    buf->headLogSeg = -1;

    if (listType == FS_BUFLIST_META &&
        inode != NULL)
    {
        if (FS_I_IS_DIR(inode->dinode.attrs))
        {
            buf->flags |= FS_BUF_DIRECTORY;
            if (FS_I_IS_COMPACT_DIR(inode->dinode.attrs))
            {
                buf->flags |= FS_BUF_COMPACT;
            }
        }
        else if (page == FS_BUF_NO_PAGE &&
                 FS_I_IS_COMPACT_EMAP(inode->dinode.attrs))
        {
            buf->flags |= FS_BUF_COMPACT;
        }
    }

    return buf;

} /* fs_bufAllocPtr */


/**
 * Return TRUE if the specified buf's data area is
 * a compact allocation.
 */
inline boolean
fs_bufIsCompact(
    fs_Buf  *buf)
{
    return ((buf->flags & FS_BUF_COMPACT) == FS_BUF_COMPACT);
}


/**
 * Return true if the buffer is not marked invalid.
 */
static inline boolean
fs_bufIsValid(
    fs_Buf  *buf)
{
    return ((buf->flags & FS_BUF_INVALID) == 0);
}


/**
 * Search for a buffer with the specified inum and page.
 * Must be called with the buflistGlobal spinlock held.
 * Return NULL if the buffer is not found.
 */
fs_Buf *
fs_bufFindByPage(
    fs_Context *context,
    uint64     inum,
    int32      snapnum,
    fs_BufList *bufList,
    uint64     page)
{
    fs_BufDllHdr    *hdr;
    fs_Buf          *buf;
    uint64          snap = 1ULL << snapnum;
    int32           hash = fs_bufHashVolInumPage(context->volId, inum,
                                                 page);
    //printf("bufFindByPage hash %d inum %lld page %lld\n",
    //  hash, inum, page);

    ASSERT_LOCK_OWNED(&fs_bufListGlobal[bufList->type].spin);

    hdr = &fs_bufListGlobal[bufList->type].bufsPage[hash];
    DLL_ITERATE(buf, hdr, pageHashEntry)
    {
        if (buf->fileSystem == context->fileSystem &&
            buf->inum == inum &&
            buf->page == page &&
            (buf->snapshots & snap) == snap)
        {
            return buf;
        }
    }

    return NULL;
}


/**
 * Search for a buffer containing the specified (inode, page) pair and
 * return its block number, or an invalid block if the page is not found
 * in the buffer cache.
 *
 * The buffer is not held locked on return from this routine, so it is
 * possible that the block will change before the information is used.
 * Thus, this information should only be used in an advisory sense, and
 * not for, say, reading a block from disk.
 */
fs_Block fs_bufLookupBlockByPage(
    fs_Context *context,
    fs_Inode   *inode,
    uint64     page)
{
    fs_BufList      *bufList;
    fs_Buf          *buf;
    fs_Block        block;
    nfx_spinlock_t  *spinlock;
    
    // hack to distinguish metadata buffers from user buffers
    //
    // TODO - this hack isn't guaranteed to produce the correct
    // result - some dynamically allocated inodes are metadata
    // files  (eg. directory hash chains, directory indexer)
    // FIXTHIS!
    if (inode->inum < FS_MAX_RESERVED_INODE)
    {
        bufList = &context->fileSystem->bufList[FS_BUFLIST_META];
        spinlock = &fs_bufListGlobal[FS_BUFLIST_META].spin;
    }
    else
    {
        bufList = &context->fileSystem->bufList[FS_BUFLIST_USER];
        spinlock = &fs_bufListGlobal[FS_BUFLIST_USER].spin;
    }

    NFX_SPIN_LOCK(spinlock);
    buf = fs_bufFindByPage(context, inode->inum, inode->snapnum, bufList, page);

    if (buf != NULL &&
        buf->inum == inode->inum &&
        buf->page == page &&
        (buf->snapshots & 1) &&
        fs_bufIsValid(buf))
    {
        block = buf->block;
        NFX_SPIN_UNLOCK(spinlock);
        return block;
    }

    NFX_SPIN_UNLOCK(spinlock);
    return FS_INVALID_BLOCK;
}


/**
 * Search for a buffer containing the specified block.
 * Return NULL if the buffer is not found.
 *
 * Must hold a spin lock on the specified bufList.
 */
static fs_Buf *
fs_bufFindByBlock(
    fs_Context  *context,
    int32       bufListType,
    fs_Block    block)
{
    fs_BufListGlobal  *bufListGlobal = &fs_bufListGlobal[bufListType];
    fs_Buf  *buf;
    int32   hash = fs_bufHashVolBlock(context->volId, block);

    FS_ASSERT(fs_blockIsValid(block));
    ASSERT_LOCK_OWNED(&fs_bufListGlobal[bufListType].spin);

    DLL_ITERATE(buf, &bufListGlobal->bufsBlk[hash], blkNumEntry)
    {
        if (fs_blockEq(buf->block, block) &&
            buf->fileSystem == context->fileSystem)
        {
            return buf;
        }
    }

    return NULL;
}


/**
 * Acquire a lock on the specified buffer with the specified lock mode.
 */
static boolean
fs_bufLock(
    fs_Context      *context,
    fs_Buf          *buf,
    nfx_spinlock_t  *spin,
    int32           lockMode)
{
    if (lockMode == BUF_GET_MODE_EXCL) {
        return (fs_trylockExclusiveSleep(context, &buf->lock, spin));
    }
    else {
        return (fs_trylockSharedSleep(context, &buf->lock, spin));
    }
}


/**
 * Add the specified buf to the free list
 * if its available for reclamation.
 *
 * Must be called with the bufList global spinlock
 * held.
 */
static void
fs_bufMaybeAddToFreeList(
    fs_Context  *context,
    fs_BufList  *bufList,
    fs_Buf      *buf)
{
    FS_ASSERT(DLL_IS_ISOLATED(buf, mruEntry));
    ASSERT_LOCK_OWNED(&fs_bufListGlobal[bufList->type].spin);

    if (buf->lock.count == 0 &&
        fs_bufIsAvail(buf))
    {
        // Buf is free for reclamation, add it to
        // the free list, or just free it if there
        // isn't any useful data with it.

        FS_ASSERT(!fs_lockHeldShare(context, &buf->lock));
        FS_ASSERT(!fs_lockHeldExclusive(context, &buf->lock));

        if ((bufList->type == FS_BUFLIST_META &&
             buf->u.data == NULL) ||
            (bufList->type == FS_BUFLIST_USER &&
             buf->u.user.codHandle == 0))
        {
            FS_ASSERT(DLL_IS_ISOLATED(buf, blkNumEntry));
            FS_ASSERT(DLL_IS_ISOLATED(buf, pageHashEntry));
            FS_ASSERT(DLL_IS_ISOLATED(buf, pageEntry));
            FS_ASSERT(DLL_IS_ISOLATED(buf, pageDirtyEntry));
#ifndef NEW_ZF
            FS_ASSERT(DLL_IS_ISOLATED(buf, zeroFillEntry));
#endif // NEW_ZF

            fs_ramDealloc(FS_RAM_BUF_PTR, buf);
#ifdef FS_DEBUG_BUF_HISTORY
            NFX_SPIN_LOCK(&spinalloc);
            fs_buf_allocate_ptr[fs_buf_allocate_index] = buf;
            fs_buf_allocate_oper[fs_buf_allocate_index] = FS_DEBUG_BUF_HISTORY_DEALLOC;
            fs_buf_allocate_index = (fs_buf_allocate_index + 1) % FS_DEBUG_BUF_HISTORY_ALLOC_MAX;
            NFX_SPIN_UNLOCK(&spinalloc);
#endif
        }
        else
        {
            int32 itype = fs_bufGetItype(buf);
            DLL_INSERT_HEAD(&fs_bufListGlobal[bufList->type].mruHdr[itype],
                            buf, mruEntry);
            fs_bufListGlobal[bufList->type].count.freeBufs[itype] += 1;
    
            if (DLL_IS_ISOLATED(buf, blkNumEntry) &&
                DLL_IS_ISOLATED(buf, pageHashEntry))
            {
                // Buf not on any other list now, increment
                // usedCount.
                fs_bufIncrUsedCount(context, bufList, buf);
            }
        }
    }
}


/**
 * Search for a buffer containing the specified (inode, page) pair and
 * return it, with a lock held.  Return NULL if the buffer is not found.
 *
 * Caller must hold bufListGlobal spinlock, which is released
 * while acquiring the buffer lock.
 */
static fs_Buf *
fs_bufLockByPage(
    fs_Context *context,
    uint64     inum,
    int32      snapnum,
    fs_BufList *bufList,
    uint64     page,
    int32      lockMode,
    boolean    isTryLock)
{
    ASSERT_LOCK_OWNED(&fs_bufListGlobal[bufList->type].spin);

    while (TRUE)
    {
        int32 itype;
        fs_Buf *buf = fs_bufFindByPage(context, inum, snapnum, bufList, page);
        if (buf == NULL)
        {
            //printf("not found by page\n");
            return NULL;
        }
        BUF_TRACE(("  hit ino %lld page %lld", inum, page));

        itype = fs_bufGetItype(buf);
        bufList->count.hitCount[itype] += 1;
        fs_bufListGlobal[bufList->type].count.hitCount[itype] += 1;
        bufList->count.hitCount[FS_BUF_CACHE_ITYPE_TOTAL] += 1;
        fs_bufListGlobal[bufList->type].count.hitCount[FS_BUF_CACHE_ITYPE_TOTAL] += 1;

        if (isTryLock)
        {
            if (!fs_trylock(context, &buf->lock, lockMode))
            {
                return FS_BUF_TRY_LOCK_FAILED;
            }
        }
        else
        {
            if (!fs_bufLock(context, buf, &fs_bufListGlobal[bufList->type].spin, lockMode))
            {
                return FS_BUF_TRY_LOCK_FAILED;
            }
        }
        if (buf->inum == inum &&
            buf->page == page &&
            (buf->snapshots & (1ULL << snapnum)) &&
            fs_bufIsValid(buf))
        {
            return buf;
        }
        
        // Buf no longer valid, try again.
        BUF_TRACE(("  hit ino %lld page %lld, now invalid",
                  inum, page));
        if (lockMode == FS_LOCK_MODE_EXCL)
            fs_unlockExclusive(context, &buf->lock);
        else
            fs_unlockShared(context, &buf->lock);

        fs_bufMaybeAddToFreeList(context, bufList, buf);
    }
}


/**
 * Search for a buffer containing the specified block and return it, with
 * a lock held.  Return NULL if the buffer is not found.
 *
 * Caller must hold bufListGlobal spinlock, which is released
 * while acquiring the buffer lock.
 */
static fs_Buf *
fs_bufLockByBlock(
    fs_Context *context,
    fs_BufList *bufList,
    fs_Block   block,
    int32      lockMode,
    boolean    isTryLock)
{
    ASSERT_LOCK_OWNED(&fs_bufListGlobal[bufList->type].spin);

    while (TRUE)
    {
        int32 itype;
        fs_Buf *buf = fs_bufFindByBlock(context, bufList->type, block);

        if (buf == NULL)
            return NULL;

        BUF_TRACE(
                 ("  hit blk %lld", fs_blockNumber(block)));

        itype = fs_bufGetItype(buf);
        bufList->count.hitCount[itype] += 1;
        fs_bufListGlobal[bufList->type].count.hitCount[itype] += 1;
        bufList->count.hitCount[FS_BUF_CACHE_ITYPE_TOTAL] += 1;
        fs_bufListGlobal[bufList->type].count.hitCount[FS_BUF_CACHE_ITYPE_TOTAL] += 1;

        if (isTryLock) 
        {
            if (!fs_trylock(context, &buf->lock, lockMode))
            {
                return FS_BUF_TRY_LOCK_FAILED;
            }
        }
        else 
        {
            if (!fs_bufLock(context, buf, &fs_bufListGlobal[bufList->type].spin, lockMode))
            {
                return FS_BUF_TRY_LOCK_FAILED;
            }
        }

        if (fs_blockEq(buf->block, block) &&
            fs_bufIsValid(buf))
        {
            return buf;
        }
        
        BUF_TRACE(("  hit blk %lld, now invalid", fs_blockNumber(block)));
        if (lockMode == FS_LOCK_MODE_EXCL)
            fs_unlockExclusive(context, &buf->lock);
        else
            fs_unlockShared(context, &buf->lock);

        fs_bufMaybeAddToFreeList(context, bufList, buf);
    }
}


/**
 * Adds the buf to the blknum cache.  
 *
 * Must be called with the buflistGlobal spinlock held.
 */
static void
fs_bufBlknumCacheInsert(
    fs_Context  *context,
    fs_Buf      *buf,
    fs_BufList  *bufList,
    fs_Inode    *inode) // not used
{    
    fs_BufListGlobal    *bufListGlobal = &fs_bufListGlobal[bufList->type];
    int32  hash;

    BUF_TRACE(("vol %llx block cache insert buf %p block %lld",
               buf->fileSystem->volId.whole_thing,
               buf, fs_blockNumber(buf->block)));

    FS_ASSERT(fs_blockIsValid(buf->block));
    FS_ASSERT(DLL_IS_ISOLATED(buf, blkNumEntry));
    FS_ASSERT(DLL_IS_ISOLATED(buf, mruEntry));
    ASSERT_LOCK_OWNED(&fs_bufListGlobal[bufList->type].spin);

    // If its not already in the page cache
    // (in which case its already counted),
    // count it.
    if (DLL_IS_ISOLATED(buf, pageHashEntry))
    {
        fs_bufIncrUsedCount(context, bufList, buf);
    }

#if defined(FS_DEBUG_DUP_BUF_BLKNUM_CACHE_INSERT)
    {
        fs_Buf *dbg, *dbgNext;
        
        hash = fs_bufHashVolBlock(context->volId, buf->block);
        DLL_ITERATE_NEXT(dbg, dbgNext,
                         &bufListGlobal->bufsBlk[hash],
                         blkNumEntry)
        {
            if (fs_blockEq(dbg->block, buf->block) &&
                buf->fileSystem == context->fileSystem)
            {
                printf("  trying inum %lld snap 0x%llx (blk %lld flags %x)\n",
                    buf->inum, buf->snapshots, fs_blockNumber(buf->block), buf->flags);
                printf("   found inum %lld snap 0x%llx (blk %lld flags %x)\n",
                    dbg->inum, dbg->snapshots, fs_blockNumber(dbg->block), dbg->flags);
                fs_error(
                    context, FS_INTERNAL_ERR,
                    "duplicate cache insert blk %lld (ino %lld page %lld)\n",
                    fs_blockNumber(buf->block), buf->inum, buf->page);
            }
        }
    }
#endif

    hash = fs_bufHashVolBlock(context->volId, buf->block);
    DLL_INSERT_TAIL(
        &bufListGlobal->bufsBlk[hash],
        buf, blkNumEntry);
#ifdef FS_BUF_BLK_CHECK_HASH_DEPTH
    bufListGlobal->blkDepth[hash]++;
#endif

} /* fs_bufBlknumCacheInsert */


/**
 * Add an entry to the page cache.
 *
 * Must be called with buflistGlobal spinlock held.
 */
void fs_bufPageCacheInsert(
    fs_Context  *context,
    fs_Buf      *buf,
    fs_BufList  *bufList,
    fs_Inode    *inode)
{
    fs_BufPageHdr *pageHdr;
    boolean       found;
    int32         hash;

    BUF_TRACE(
       ("vol %llx ino,page cache insert buf %p ino %lld page %lld snaps %llx",
            buf->fileSystem->volId.whole_thing,
            buf, buf->inum, buf->page, buf->snapshots));

    FS_ASSERT(buf->inum > 0);
    FS_ASSERT(buf->page != FS_BUF_NO_PAGE);
    FS_ASSERT(DLL_IS_ISOLATED(buf, pageEntry));
    FS_ASSERT(DLL_IS_ISOLATED(buf, pageHashEntry));
    FS_ASSERT(DLL_IS_ISOLATED(buf, mruEntry));
    ASSERT_LOCK_OWNED(&fs_bufListGlobal[bufList->type].spin);

    // If its not already in the blknum cache
    // (in which case its already counted),
    // count it.
    if (DLL_IS_ISOLATED(buf, blkNumEntry))
    {
        fs_bufIncrUsedCount(context, bufList, buf);
    }
    
    // Add it to the inum hash chain.
    found = FALSE;
    hash = fs_bufHashVolInum(context->volId, buf->inum);
    DLL_ITERATE(pageHdr,
                &fs_bufListGlobal[bufList->type].bufsInum[hash],
                inumEntry)
    {
        if (pageHdr->fileSystem == context->fileSystem &&
            pageHdr->inum == buf->inum)
        {
            found = TRUE;
            break;
        }
    }
    if (!found)
    {
        // Nothing in the cache for this inum, allocate a
        // new page header.
        pageHdr = fs_ramAlloc(FS_RAM_BUF_PAGE_HDR, sizeof(fs_BufPageHdr));
        pageHdr->inum = buf->inum;
        pageHdr->dirtyCount = 0;
        pageHdr->fileSystem = context->fileSystem;
        DLL_INIT(&(pageHdr->bufChainHdr));
        DLL_INIT(&(pageHdr->bufDirtyHdr));
        DLL_INIT(&(pageHdr->waiterHdr));
        DLL_ISOLATE_ELEM(pageHdr, inumEntry);
        DLL_INSERT_HEAD(&fs_bufListGlobal[bufList->type].bufsInum[hash],
                        pageHdr, inumEntry);
#ifdef FS_BUF_INUM_CHECK_HASH_DEPTH
        fs_bufListGlobal[bufList->type].inumDepth[hash]++;
#endif
    }

#if defined(FS_DEBUG_DUP_BUF_PAGE_CACHE_INSERT)
    {
        fs_Buf *dbg, *dbgNext;

        DLL_ITERATE_NEXT(dbg, dbgNext,
                         &pageHdr->bufChainHdr,
                         pageEntry)
        {
            if (dbg->fileSystem == context->fileSystem &&
                dbg->page == buf->page &&
                (dbg->snapshots & buf->snapshots))
            {
                printf("  trying inum %lld page %lld snap 0x%llx (blk %lld flags %x)\n",
                    buf->inum, buf->page, buf->snapshots, 
                    fs_blockNumber(buf->block), buf->flags);
                printf("   found inum %lld page %lld snap 0x%llx (blk %lld flags %x)\n",
                    dbg->inum, dbg->page, dbg->snapshots, 
                    fs_blockNumber(dbg->block), dbg->flags);
                fs_error(context, FS_INTERNAL_ERR,
                      "duplicate page cache insert\n");
            }
        }
    }
#endif

    DLL_INSERT_HEAD(&pageHdr->bufChainHdr, buf, pageEntry);

    // Add it to the inode/page hash.
    hash = fs_bufHashVolInumPage(context->volId, buf->inum, buf->page);

#if defined(FS_DEBUG_DUP_BUF_PAGE_CACHE_INSERT)
    {
        fs_Buf *dbg, *dbgNext;
    
        DLL_ITERATE_NEXT(dbg, dbgNext,
                         &fs_bufListGlobal[bufList->type].bufsPage[hash],
                         pageHashEntry)
        {
            if (dbg->fileSystem == context->fileSystem &&
                dbg->inum == buf->inum &&
                dbg->page == buf->page &&
                (dbg->snapshots & buf->snapshots))
            {
                printf("  trying inum %lld page %lld snap 0x%llx (blk %lld)\n",
                    buf->inum, buf->page, buf->snapshots, 
                    fs_blockNumber(buf->block));
                printf("   found inum %lld page %lld snap 0x%llx (blk %lld)\n",
                    dbg->inum, dbg->page, dbg->snapshots, 
                    fs_blockNumber(dbg->block));
                fs_error(context, FS_INTERNAL_ERR,
                    "duplicate page hash cache insert\n");
            }
        }
    }
#endif

    DLL_INSERT_HEAD(&fs_bufListGlobal[bufList->type].bufsPage[hash],
                    buf, pageHashEntry);
#ifdef FS_BUF_PAGE_CHECK_HASH_DEPTH
    fs_bufListGlobal[bufList->type].pageDepth[hash]++;
#endif

} /* fs_bufPageCacheInsert */


/**
 * Put the caller to sleep for a bit to wait for
 * bufs to come available.
 */
static void
fs_bufWaitAvail(
    fs_Context  *context,
    int32       *numRetries)
{
    if (!context->isThread)
    {
        printf("buffer cache 100%% busy for non-thread\n");
        fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
    }
    else if (++*numRetries > 1000)
    {
        printf("buffer cache 100%% busy, retries exhausted\n"); 
        fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
    }
    else
    {
        fs_bufBufAllocRetries += 1;
        fs_sleep(context, 10);
    }
}


/**
 * Return the buffer for the specified (inode, page) pair, locked with the
 * specified mode.
 */
fs_Buf *
fs_bufGetPage(
    fs_Context      *context,
    fs_Inode        *inode,
    uint64          page,
    int32           mode,
    uint32          option,
    evm_volSize_t   readLen)
{
    fs_BufList      *bufList;
    fs_RdwrReq      *rdwr;
    uint32          typeBit;
    boolean         isTryLock = ((option & BUF_GET_OPT_TRY_LOCK) != 0);
    nfx_spinlock_t  *spinlock;
    int32           numRetries = 0;

    BUF_TRACE(
        ("vol %llx bget ino %lld snap %d page %lld mode %s opts %#x redLen %d",
               context->fileSystem->volId.whole_thing,
               inode->inum, inode->snapnum,
               page,
               mode == BUF_GET_MODE_EXCL ? "excl" : "shared",
               option,
               (int32) readLen));

    // Async not supported yet.
    FS_ASSERT(option & BUF_GET_OPT_SLEEP);
    // we can not call this while holding the fs_inodeCache spinlock
    ASSERT_LOCK_NOT_OWNED(&fs_inodeCache->spin);

    // Ensure we meet the minimum IO size requirements.
    readLen += FS_MIN_IO_SIZE - 1;
    readLen &= ~(FS_MIN_IO_SIZE - 1);

    if ((option & BUF_GET_OPT_USER) != 0) {
        bufList = &context->fileSystem->bufList[FS_BUFLIST_USER];
        typeBit = FS_BUF_TYPE_USER;
        spinlock = &fs_bufListGlobal[FS_BUFLIST_USER].spin;
    }
    else {
        bufList = &context->fileSystem->bufList[FS_BUFLIST_META];
        typeBit = FS_BUF_TYPE_META;
        spinlock = &fs_bufListGlobal[FS_BUFLIST_META].spin;
    }

#ifdef NEW_ZF
    if (!DLL_IS_EMPTY(&(context->fileSystem->zfInodeHdr)) &&
        (inode->zfflags & INO_FLAGS_UZERO))
    {
        fs_zfInode *    zfInode = NULL;
        fs_zfPage *     zfPage = NULL;

        NFX_SPIN_LOCK(&(context->fileSystem->zfSpin));
        // Check if the inode is still part of the ZF list.
        zfInode = fs_zfGetInode(context, inode->inum);
        if (zfInode != NULL)
        {
            zfPage = fs_zfGetPage(context, inode->inum, page);
            if (zfPage != NULL)
            {
                int32 rc;
                
                DLL_REMOVE(&(zfInode->pageList), zfPage, entry);
                rc = chash_remove(fs_zfPageHashTab, (void *) zfPage);
                FS_ASSERT(rc == NFX_OK);

                // If no more pages for this inode, remove inode from ZF list.
                if (DLL_IS_EMPTY(&(zfInode->pageList)))
                {
                    fs_zfDelInode(context, zfInode->inum);
                }
            }
        }
        NFX_SPIN_UNLOCK(&(context->fileSystem->zfSpin));

        if (zfPage != NULL)
        {
            printf("page %10lld of inum %10lld still need to be zeroed "
                   "-  do it now\n", page, inode->inum);
            
            // Zero the page (inserted in cache...) and free the page.
            fs_zfZeroPage(context, inode, zfPage);
            fs_ramDealloc(FS_RAM_ZF, zfPage);            
        }
    }
#endif // NEW_ZF
    
getPageRetry:

    if (!(option & BUF_GET_OPT_NEW))
    {
        fs_Buf *buf;
        fs_Block block;

        // Search the cache for the requested (inode,page).
        NFX_SPIN_LOCK(spinlock);
        buf = fs_bufLockByPage(context, inode->inum, inode->snapnum,
                               bufList, page, mode, isTryLock);
        if (buf != NULL)
        {
            if (buf == FS_BUF_TRY_LOCK_FAILED)
            {
                NFX_SPIN_UNLOCK(spinlock);
                return NULL;
            }

            // We hit in the cache
            // Remove it from the free list.
            if (!DLL_IS_ISOLATED(buf, mruEntry))
            {
                int32 itype = fs_bufGetItype(buf);

                DLL_REMOVE(&fs_bufListGlobal[bufList->type].mruHdr[itype],
                             buf, mruEntry);
                fs_bufListGlobal[bufList->type].count.freeBufs[itype] -= 1;
            }
            NFX_SPIN_UNLOCK(spinlock);

            FS_ASSERT(context->fileSystem == buf->fileSystem);
            return buf;
        }

        if (option & BUF_GET_OPT_LOOKUP)
        {
            // Didn't find it, return NULL.
            NFX_SPIN_UNLOCK(spinlock);
            BUF_TRACE(("  lookup miss ino %lld page %lld",
                      inode->inum, page));
            return NULL;
        }

        BUF_TRACE(("  miss ino %lld page %lld", inode->inum, page));

        fs_bufIncrMissCount(bufList, inode);

        if (!inode->snapnum)
        {
            // The requested page is for the live fs, so
            // instantiate it and schedule I/O if we have an
            // extent for the page. fs_bufAllocPtr returns with
            // the buf locked excl.
            buf = fs_bufAllocPtr(context, bufList->type, inode, page);
            if (buf == NULL)
            {
                NFX_SPIN_UNLOCK(spinlock);
                fs_bufWaitAvail(context, &numRetries);
                goto getPageRetry;
            }
            buf->fileSystem = context->fileSystem;
            buf->inum = inode->inum;
            buf->page = page;
            buf->snapshots = 1ULL << inode->snapnum;
            buf->flags |= (typeBit | FS_BUF_COW_UNK);
            buf->block = FS_INVALID_BLOCK;

            // Add the entry to the cache to prevent others
            // from instantiating this entry while we're waiting
            // for I/O to complete.
            fs_bufPageCacheInsert(context, buf, bufList, inode);
        }

        NFX_SPIN_UNLOCK(spinlock);

        // Now find its block number.

        block = fs_emap_lookup_block(context, inode, page);

        if (!fs_blockIsValid(block))
        {
            // Must be a hole in the file, return NULL.
            BUF_TRACE(("  file hole ino %lld page %lld",
                      inode->inum, page));

            if (buf != NULL)
            {
                NFX_SPIN_LOCK(spinlock);
                FS_ASSERT(fs_lockOwnedExclusive(context, &buf->lock));
                fs_bufDelist(context, bufList, buf, FALSE);
                NFX_SPIN_UNLOCK(spinlock);
    
                fs_lockDestroy(context, &buf->lock);
                fs_ramDealloc(FS_RAM_BUF_PTR, buf);
#ifdef FS_DEBUG_BUF_HISTORY
                NFX_SPIN_LOCK(&spinalloc);
                fs_buf_allocate_ptr[fs_buf_allocate_index] = buf;
                fs_buf_allocate_oper[fs_buf_allocate_index] = FS_DEBUG_BUF_HISTORY_DEALLOC;
                fs_buf_allocate_index = (fs_buf_allocate_index + 1) % FS_DEBUG_BUF_HISTORY_ALLOC_MAX;
                NFX_SPIN_UNLOCK(&spinalloc);
#endif
            }

            return NULL;
        }

        if (!inode->snapnum)
        {
            // Its for the live fs, we've already instantiated
            // buf in the page cache, now add it to the blknum
            // cache, too, and start the read.
            NFX_SPIN_LOCK(spinlock);
            buf->block = block;
            fs_bufBlknumCacheInsert(context, buf, bufList, inode);
            NFX_SPIN_UNLOCK(spinlock);
        }
        else
        {
            byte    refByte = fs_refGetByte(context, block);
            boolean isLive = FALSE;

            // pages of the snapbits file are never live
            if (inode->inum != FS_SNAPSHOT_BITS_INODE)
                isLive = fs_refbyteIsLive(refByte);

            // Caller wants a snapshot version of the page.
            NFX_SPIN_LOCK(spinlock);
            if (isLive)
            {
                // Ensure we don't collide with an attempt to
                // bring in the same page for the live fs.
                buf = fs_bufLockByPage(context, inode->inum, 0, bufList,
                                       page, mode, isTryLock);
            }
            else
            {
                // Check again, since we let go of the spinlock since
                // our last check.
                buf = fs_bufLockByPage(context, inode->inum, inode->snapnum,
                                       bufList, page, mode, isTryLock);
            }
            if (buf != NULL)
            {
                if (buf == FS_BUF_TRY_LOCK_FAILED)
                {
                    NFX_SPIN_UNLOCK(spinlock);
                    return NULL;
                }
    
                // We hit in the cache
                // Remove it from the free list.
                if (!DLL_IS_ISOLATED(buf, mruEntry))
                {
                    int32 itype = fs_bufGetItype(buf);
    
                    DLL_REMOVE(&fs_bufListGlobal[bufList->type].mruHdr[itype],
                                 buf, mruEntry);
                    fs_bufListGlobal[bufList->type].count.freeBufs[itype] -= 1;
                }

                if (isLive)
                {
                    // Set this snapnum in the buf.
                    buf->snapshots |= 1ULL << inode->snapnum;
                }
    
                NFX_SPIN_UNLOCK(spinlock);
    
                FS_ASSERT(context->fileSystem == buf->fileSystem);

                return buf;
            }

            // Still not in the page cache, check the block cache.

            buf = fs_bufLockByBlock(context, bufList, block, mode, isTryLock);
            if (buf == FS_BUF_TRY_LOCK_FAILED)
            {
                NFX_SPIN_UNLOCK(spinlock);
                return NULL;
            }
            if (buf != NULL)
            {
                // The block was found, which means it's part of the same
                // file, but only currently loaded for a different snapshot,
                // or it was read into the blocknum cache directly, but not
                // the page cache (eg. during replay).
    
                FS_ASSERT(buf->fileSystem == context->fileSystem);
                FS_ASSERT(buf->snapshots != 0);
    
                // Replay should have set these up already.
                FS_ASSERT(buf->inum == inode->inum);
                FS_ASSERT(buf->page == page);
    
                if (DLL_IS_ISOLATED(buf, pageEntry))
                {
                    // Add it to the page cache now.
                    fs_bufPageCacheInsert(context, buf, bufList, inode);
    
                    // Need to bump up the page counter here if dirty
                    if (buf->flags & FS_BUF_DIRTY)
                    {
                        fs_BufPageHdr   *pageHdr;
    
                        // Has already been dirtied, but wasn't on page list
                        // before, so increment pageHdr dirty count here.
    
                        pageHdr = fs_bufPageHdrForInum(buf->fileSystem,
                                                       bufList->type, buf->inum);
                        FS_ASSERT(pageHdr != NULL);
                        FS_ASSERT(!DLL_IS_ISOLATED(buf, mrdEntry));
                        FS_ASSERT(DLL_IS_ISOLATED(buf, pageDirtyEntry));
                        pageHdr->dirtyCount += 1;
                        // insert to dirty page hdr
                        DLL_INSERT_HEAD(&pageHdr->bufDirtyHdr, buf, pageDirtyEntry);
#ifdef FS_DEBUG_BUF_HISTORY
                        NFX_SPIN_LOCK(&buf->spinHistory);
                        buf->history[buf->indexHistory] = FS_BUF_HIST_INSERT_PAGE;
                        buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
                        buf->flagsHistory[buf->indexHistory] = buf->flags;
                        buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
                        buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
                        buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
                        NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
                    }
                }
    
                buf->snapshots |= 1ULL << inode->snapnum;
    
                // Remove it from the free list.
                if (!DLL_IS_ISOLATED(buf, mruEntry))
                {
                    int32 itype = fs_bufGetItype(buf);
    
                    DLL_REMOVE(&fs_bufListGlobal[bufList->type].mruHdr[itype],
                                 buf, mruEntry);
                    fs_bufListGlobal[bufList->type].count.freeBufs[itype] -= 1;
                }
                NFX_SPIN_UNLOCK(spinlock);
    
                return buf;
            }
        
            // Not found in the block cache either.
            // Instantiate it and schedule I/O.
            // fs_bufAllocPtr returns with the buf locked excl.
            buf = fs_bufAllocPtr(context, bufList->type, inode, page);
            if (buf == NULL)
            {
                NFX_SPIN_UNLOCK(spinlock);
                fs_bufWaitAvail(context, &numRetries);
                goto getPageRetry;
            }
            buf->fileSystem = context->fileSystem;
            buf->inum = inode->inum;
            buf->page = page;
            buf->snapshots = 1ULL << inode->snapnum;
            buf->flags |= typeBit;
            if (isLive)
            {
                buf->snapshots |= 1ULL;
                if (fs_refbyteSnapCount(refByte) > 0)
                {
                    buf->flags |= FS_BUF_COW;
                }
            }
            buf->block = FS_INVALID_BLOCK;
    
            // Add the entry to the cache to prevent others
            // from instantiating this entry while we're waiting
            // for I/O to complete.
            fs_bufPageCacheInsert(context, buf, bufList, inode);

            // Now add it to the blknum cache, too.
            buf->block = block;
            fs_bufBlknumCacheInsert(context, buf, bufList, inode);
            NFX_SPIN_UNLOCK(spinlock);
        }

        /*
         * Detect if we are read ahead over aggressively so that
         * buffer cache starts reusing pages before they are actually
         * utilized
         */

        NFX_SPIN_LOCK(&inode->spin);
        if (inode->readAheadState != NULL &&
            inode->readAheadState->raOffset > (page << FS_BLOCK_SHIFT))
        {
            //printf("unexpected ra read at %lld page %lld \n",
            //        inode->raOffset, page<< FS_BLOCK_SHIFT);
            inode->readAheadState->uxRead++;

            /*
             * If have hit the more than 255 unexpected reads between
             * read ahead and read, assume we have taken only 255. Anyway
             * currently we care about 0, 1 or more unexpected reads only.
             */

            if (inode->readAheadState->uxRead == 0) {
                inode->readAheadState->uxRead = ~inode->readAheadState->uxRead;
            }
        }
        NFX_SPIN_UNLOCK(&inode->spin);

        rdwr = fs_rdwrAlloc();
        if (typeBit == FS_BUF_TYPE_USER)
        {
            NFX_SPIN_LOCK(&fs_ioSpin);
            fs_ioTotalUserReads += 1;
            NFX_SPIN_UNLOCK(&fs_ioSpin);

            NFX_SPIN_LOCK(&context->fileSystem->spin);
            context->fileSystem->ioTotalUserReads += 1;
            NFX_SPIN_UNLOCK(&context->fileSystem->spin);

            fs_volReadUser(context, rdwr,
                           fs_blockNumber(buf->block) * FS_BLOCK_SIZE,
                           readLen);
        }
        else
        {
            NFX_SPIN_LOCK(&fs_ioSpin);
            fs_ioTotalMetaReads += 1;
            NFX_SPIN_UNLOCK(&fs_ioSpin);

            NFX_SPIN_LOCK(&context->fileSystem->spin);
            context->fileSystem->ioTotalMetaReads += 1;
            NFX_SPIN_UNLOCK(&context->fileSystem->spin);

            fs_volReadBlock(context, rdwr, buf->block);
        }

        if (typeBit == FS_BUF_TYPE_META)
        {
            // Plug in the just-read meta data.
            FS_ASSERT(buf->u.data == NULL);

            fs_bufSetDataPtr(context, buf, rdwr->evmStat.rDataHandle.dataPtr);

#if defined(FS_DEBUG_SHADOW_BUF)
            // Initialize the shadow buffer.
            FS_DBG_INIT_SHADOW(buf);
#endif

#if defined(FS_DEBUG_BUF_CKSUM)
            fs_bufCompareCksum(context, buf->block, buf->u.data, buf,
                               "get page", FALSE);
#endif
        }
        else
        {
            // Plug in the just-read user data.
            buf->u.user.codHandle = rdwr->evmStat.rDataHandle.codHandle;
            buf->u.user.codLength = rdwr->evmReq.reqSize * EVM_LBLK_SIZE;
            buf->u.user.codOffset = 0;
            FS_ASSERT(buf->u.user.codLength <= FS_BLOCK_SIZE);
        }

        NFX_SPIN_LOCK(&buf->spin);
        buf->flags |= FS_BUF_COW_UNK;
        NFX_SPIN_UNLOCK(&buf->spin);

        fs_rdwrDealloc(rdwr);

        // Demote the lock?
        if (mode == BUF_GET_MODE_SHARED)
        {
            fs_lockDemote(context, &buf->lock, FS_LOCK_MODE_SHARE);
        }

        // TODO - strategy for obtaining COW status?  Perhaps
        // kick off async I/O here and update flags when I/O
        // completes?  - or just wait until someone requests it?

        FS_ASSERT(context->fileSystem == buf->fileSystem);
        return buf;
    }
    else
    {
        fs_Buf *buf;

        // We're creating a new buffer. fs_bufAllocPtr() will obtain
        // a new buf ptr (or a recycled one) and return with it
        // locked EXCL.

        BUF_TRACE(("  create ino %lld page %lld",
                 inode->inum, page));

        NFX_SPIN_LOCK(spinlock);
        buf = fs_bufAllocPtr(context, bufList->type, inode, page);
        if (buf == NULL)
        {
            NFX_SPIN_UNLOCK(spinlock);
            fs_bufWaitAvail(context, &numRetries);
            goto getPageRetry;
        }
        NFX_SPIN_UNLOCK(spinlock);

        buf->fileSystem = context->fileSystem;
        buf->inum = inode->inum;
        buf->page = page;
        buf->snapshots = 1ULL << inode->snapnum;
        buf->flags |= typeBit;

        if (!(option & BUF_GET_OPT_DEFER_ALLOC))
        {
            if (typeBit == FS_BUF_TYPE_META)
            {
                fs_Block block;
                if (inode->inum == FS_INODE_INODE)
                {
                    block = fs_allocateInodeFileBlock(context, inode, page);
                }
                else
                {
                    block = fs_allocateMetadataBlock(context, inode);
                }
                buf->block = block;
                fs_emap_add_block(context, inode, page, block);
                if (inode->inum != FS_INODE_INODE)
                {
                    inode->dinode.numblocks += 1;
                    fs_inodeWrite(context, inode, INO_WRITE_LOG);
                }
            }
            else
            {
                FS_ASSERT(FALSE);
            }
            if (!fs_blockIsValid(buf->block))
            {
                fs_error(context, FS_INTERNAL_ERR, "invalid block allocated");
                return NULL;
            }
        }
        else
        {
            buf->block = FS_INVALID_BLOCK;
        }

        if (typeBit == FS_BUF_TYPE_META)
        {
            // Malloc space for the data.
            fs_bufAllocDataPtr(context, buf);

#if defined(FS_DEBUG_SHADOW_BUF)
            // Initialize the shadow buffer.
            FS_DBG_INIT_SHADOW(buf);
#endif
        }
        else
        {
            // New user buf doesn't get allocated a COD, rather
            // a COD arrival should trigger a bufp alloc, so our
            // caller will supply the COD handle.
        }

        // We hold EXCL access to buf at this point, but its not
        // in the cache.  Just add it to the cache, demote the lock
        // if necessary, and return.
        FS_ASSERT(DLL_IS_ISOLATED(buf, blkNumEntry));
        FS_ASSERT(DLL_IS_ISOLATED(buf, pageEntry));
        FS_ASSERT(DLL_IS_ISOLATED(buf, mruEntry));
        FS_ASSERT(DLL_IS_ISOLATED(buf, mrdEntry));
        FS_ASSERT(fs_lockOwnedExclusive(context, &buf->lock));
    
        NFX_SPIN_LOCK(spinlock);
        //printf("inserting inum %lld page %lld blknum %lld into page cache\n",
        //       buf->inum, buf->page, fs_blockNumber(buf->block));
        fs_bufPageCacheInsert(context, buf, bufList, inode);
        if (fs_blockIsValid(buf->block))
        {
            //printf("  and the blknum cache, too\n");
            // Add it to the blknum cache, too.
            fs_bufBlknumCacheInsert(context, buf, bufList, inode);
        }
        NFX_SPIN_UNLOCK(spinlock);
    
        // Demote the lock?
        if (mode == BUF_GET_MODE_SHARED)
        {
            fs_lockDemote(context, &buf->lock, FS_LOCK_MODE_SHARE);
        }
    
        FS_ASSERT(context->fileSystem == buf->fileSystem);
        return buf;
    }

} /* fs_bufGetPage */



/**
 * Called to get (by blknum) and lock an fs block in the cache.
 */
fs_Buf *
fs_bufGetBlock(
    fs_Context      *context,
    fs_Inode        *inode,
    fs_Block        block,
    int32           mode,
    uint32          option)
{
    fs_FileSystem   *fs = context->fileSystem;
    fs_BufList      *bufList;
    nfx_spinlock_t  *spinlock;
    uint64          inum = (inode == NULL) ? 0 : inode->inum;
    uint32          typeBit;
    int32           numRetries = 0;

    BUF_TRACE(("vol %llx bget %lld ino %lld mode %s opts %#x",
               context->fileSystem->volId.whole_thing,
               fs_blockNumber(block),
               inum,
               mode == BUF_GET_MODE_EXCL ? "excl" : "shared",
               option));

    // Async not supported yet.
    FS_ASSERT(option & BUF_GET_OPT_SLEEP);
    // TryLocks not supported yet
    FS_ASSERT(!(option & BUF_GET_OPT_TRY_LOCK));

    if (fs_blockIsValid(block) &&
        fs_blockNumber(block) >= fs->super->numBlks)
    {
        fs_warn(context, FS_INTERNAL_ERR, "block number too large: %lld",
                 fs_blockNumber(block));
        fs_volumeException(context, context->fileSystem, FS_VOLUME_EXCEPTION);
    }

    if ((option & BUF_GET_OPT_USER) != 0)
    {
        bufList = &fs->bufList[FS_BUFLIST_USER];
        typeBit = FS_BUF_TYPE_USER;
        spinlock = &fs_bufListGlobal[FS_BUFLIST_USER].spin;
    }
    else
    {
        bufList = &fs->bufList[FS_BUFLIST_META];
        typeBit = FS_BUF_TYPE_META;
        spinlock = &fs_bufListGlobal[FS_BUFLIST_META].spin;
    }

getBlockRetry:

    if (!(option & BUF_GET_OPT_NEW))
    {
        fs_Buf *buf;
        fs_RdwrReq *rdwr;

        // Search the cache for the requested blkNum.
        NFX_SPIN_LOCK(spinlock);
        buf = fs_bufLockByBlock(context, bufList, block, mode, FALSE);
        if ((buf != NULL) && (buf != FS_BUF_TRY_LOCK_FAILED))
        {
            // We hit in the cache.
            // Remove it from the free list.
            if (!DLL_IS_ISOLATED(buf, mruEntry))
            {
                int32 itype = fs_bufGetItype(buf);

                DLL_REMOVE(&fs_bufListGlobal[bufList->type].mruHdr[itype],
                             buf, mruEntry);
                fs_bufListGlobal[bufList->type].count.freeBufs[itype] -= 1;
            }
            NFX_SPIN_UNLOCK(spinlock);

            FS_ASSERT(context->fileSystem == buf->fileSystem);
            return buf;
        }

        // Missed in the cache.
        if (option & BUF_GET_OPT_LOOKUP)
        {
            // Didn't find it, return NULL.
            NFX_SPIN_UNLOCK(spinlock);
            BUF_TRACE(("  lookup miss blk %lld",
                fs_blockNumber(block)));
            return NULL;
        }

        BUF_TRACE(("  miss blk %lld", fs_blockNumber(block)));

        fs_bufIncrMissCount(bufList, inode);

        // Didn't find it, instantiate it and schedule I/O.
        // fs_bufAllocPtr returns with the buf locked excl.
        buf = fs_bufAllocPtr(context, bufList->type, inode, FS_BUF_NO_PAGE);
        if (buf == NULL)
        {
            NFX_SPIN_UNLOCK(spinlock);
            fs_bufWaitAvail(context, &numRetries);
            goto getBlockRetry;
        }

        buf->fileSystem = context->fileSystem;
        buf->block = block;
        buf->inum = inum;
        buf->page = FS_BUF_NO_PAGE;
        buf->snapshots = 0;
        buf->flags |= (typeBit | FS_BUF_COW_UNK);

        // Add the entry to the cache to prevent others
        // from instantiating this entry while we're waiting
        // for I/O to complete.
        fs_bufBlknumCacheInsert(context, buf, bufList, inode);

        NFX_SPIN_UNLOCK(spinlock);

        rdwr = fs_rdwrAlloc();
#ifdef ISABELLE // debug
        {
            vl_VolBufRoot * volBufRoot = fs->logRoot->volBufRoot;
            vl_BlockCache * blockCache = &(volBufRoot->blockCache);
            blockCache->stats.nbMetaRead++;
        }
#endif // ISABELLE
        if (typeBit == FS_BUF_TYPE_USER)
        {
            NFX_SPIN_LOCK(&fs_ioSpin);
            fs_ioTotalUserReads += 1;
            NFX_SPIN_UNLOCK(&fs_ioSpin);

            NFX_SPIN_LOCK(&context->fileSystem->spin);
            context->fileSystem->ioTotalUserReads += 1;
            NFX_SPIN_UNLOCK(&context->fileSystem->spin);

            fs_volReadUserBlock(context, rdwr, buf->block);
        }
        else
        {
            NFX_SPIN_LOCK(&fs_ioSpin);
            fs_ioTotalMetaReads += 1;
            NFX_SPIN_UNLOCK(&fs_ioSpin);

            NFX_SPIN_LOCK(&context->fileSystem->spin);
            context->fileSystem->ioTotalMetaReads += 1;
            NFX_SPIN_UNLOCK(&context->fileSystem->spin);

            fs_volReadBlock(context, rdwr, buf->block);
        }

        if (typeBit == FS_BUF_TYPE_META)
        {
            // Plug in the just-read meta data.
            FS_ASSERT(buf->u.data == NULL);

            fs_bufSetDataPtr(context, buf, rdwr->evmStat.rDataHandle.dataPtr);

#if defined(FS_DEBUG_SHADOW_BUF)
            // Initialize the shadow buffer.
            FS_DBG_INIT_SHADOW(buf);
#endif

#if defined(FS_DEBUG_BUF_CKSUM)
            fs_bufCompareCksum(context, buf->block, buf->u.data, buf,
                               "get block", FALSE);
#endif
        }
        else
        {
            // Plug in the just-read user data.
            buf->u.user.codHandle = rdwr->evmStat.rDataHandle.codHandle;
            buf->u.user.codLength = rdwr->evmReq.reqSize * EVM_LBLK_SIZE;
            buf->u.user.codOffset = 0;
            FS_ASSERT(buf->u.user.codLength <= FS_BLOCK_SIZE);
        }

        NFX_SPIN_LOCK(&buf->spin);
        buf->flags |= FS_BUF_COW_UNK;
        NFX_SPIN_UNLOCK(&buf->spin);

        fs_rdwrDealloc(rdwr);

        // Demote the lock?
        if (mode == BUF_GET_MODE_SHARED)
        {
            fs_lockDemote(context, &buf->lock, FS_LOCK_MODE_SHARE);
        }

        FS_ASSERT(context->fileSystem == buf->fileSystem);
        return buf;

        // TODO - strategy for obtaining COW status?  Perhaps
        // kick off async I/O here and update flags when I/O
        // completes?
    }
    else
    {
        fs_Buf *buf;

        // We're creating a new buffer. fs_bufAllocPtr() will obtain
        // a new buf ptr (or a recycled one) and return with it
        // locked EXCL.

        BUF_TRACE(("  create blk %lld inum %lld",
                   fs_blockNumber(block), inode ? inode->inum : 0));

        NFX_SPIN_LOCK(spinlock);
        buf = fs_bufAllocPtr(context, bufList->type, inode, FS_BUF_NO_PAGE);
        if (buf == NULL)
        {
            NFX_SPIN_UNLOCK(spinlock);
            fs_bufWaitAvail(context, &numRetries);
            goto getBlockRetry;
        }
        NFX_SPIN_UNLOCK(spinlock);

        buf->fileSystem = context->fileSystem;
        buf->inum = inum;
        buf->page = FS_BUF_NO_PAGE;
        buf->snapshots = 0;
        buf->flags |= typeBit;

        if (!fs_blockIsValid(block))
        {
            // Call into the allocator, otherwise we have
            // no way of referencing this buf!
            block = fs_allocateMetadataBlock(context, inode);
            if (!fs_blockIsValid(block))
            {
                fs_error(context, FS_INTERNAL_ERR, "alloc blk");
            }
            BUF_TRACE(("  alloc blk %lld", fs_blockNumber(block)));
        }
        buf->block = block;

        if (typeBit == FS_BUF_TYPE_META)
        {
            // Malloc space for the data.
            fs_bufAllocDataPtr(context, buf);

#if defined(FS_DEBUG_SHADOW_BUF)
            // Initialize the shadow buffer.
            FS_DBG_INIT_SHADOW(buf);
#endif
        }
        else
        {
            // New user buf doesn't get allocated a COD, rather
            // a COD arrival should trigger a bufp alloc, so our
            // caller will supply the COD handle.
        }

        // We hold EXCL access to buf at this point, but its not
        // in the cache.  Just add it to the cache, demote the lock
        // if necessary, and return.

        // Now add it to the cache.
        NFX_SPIN_LOCK(spinlock);
        fs_bufBlknumCacheInsert(context, buf, bufList, inode);
        NFX_SPIN_UNLOCK(spinlock);
    
        // Demote the lock?
        if (mode == BUF_GET_MODE_SHARED)
        {
            fs_lockDemote(context, &buf->lock, FS_LOCK_MODE_SHARE);
        }

        FS_ASSERT(context->fileSystem == buf->fileSystem);
        return buf;
    }

} /* fs_bufGetBlock */



/**
 * Called to unlock and release an fs block, held with a shared lock, back
 * to the cache.
 */
void
fs_bufRelease(
    fs_Context  *context,
    fs_Buf      *buf)
{
    fs_BufList      *bufList;
    nfx_spinlock_t  *spinlock;

    BUF_TRACE(("vol %llx brelease buf %p blk %lld ino %lld page %lld", 
               buf->fileSystem->volId.whole_thing,
               buf, fs_blockNumber(buf->block), buf->inum,
               buf->page));
    
    bufList = fs_bufListForBuf(buf);
    spinlock = &fs_bufListGlobal[bufList->type].spin;

    NFX_SPIN_LOCK(spinlock);
    fs_unlockShared(context, &buf->lock);
    fs_bufMaybeAddToFreeList(context, bufList, buf);
    NFX_SPIN_UNLOCK(spinlock);
}


#ifdef FS_DEBUG_CANDO_STATS
int32 fs_ramCandoMaxDealloc = 0;
int32 fs_bufWritebackCandoDeallocMax = 0;
#endif

/**
 * Convert any cando records to permanent, and return
 * the number of records converted.
 */
int32
fs_bufConvertCando(
    fs_Context  *context,
    fs_Buf      *buf)
{
    fs_LogUndoSummary   *cando, *candoNext;
    int32               ramCandoDealloc = 0;

    NFX_SPIN_LOCK(&buf->spin);
    DLL_ITERATE_NEXT(cando, candoNext, &buf->candoHdr, bufEntry)
    {
        // Transition the undo records on the cando
        // list to permanent changes that can be written
        // to disk.
        FS_ASSERT(DLL_IS_ISOLATED(cando, logEntry));
        FS_ASSERT(cando->logId != 0);
        FS_ASSERT(cando->objSumm.objectPtr == buf);

        DLL_REMOVE(&buf->candoHdr, cando, bufEntry);
        NFX_SPIN_UNLOCK(&buf->spin);

        fs_logConvertUndoToPerm(context,
                                cando,
                                &buf->logSegDep);

#ifdef FS_DEBUG_CANDO_STATS
        if (++ramCandoDealloc > fs_ramCandoMaxDealloc)
        {
            fs_ramCandoMaxDealloc = ramCandoDealloc;
            //printf("candomax: %d\n", ramCandoDealloc);
        }
#endif

        fs_ramDealloc(FS_RAM_LOG_UNDO_SUMM, cando);
        NFX_SPIN_LOCK(&buf->spin);
    }
    NFX_SPIN_UNLOCK(&buf->spin);

    return ramCandoDealloc;
}

int32 fs_cloneNoDeps = 0;
int32 fs_clonedBufs = 0;

/**
 * Schedule the write for a buffer.
 */
static int32
fs_bufDoWriteBack(
    fs_Context      *context,
    esm_handle_t    handle,
    fs_Buf          *buf)
{
    fs_BufList *    bufList;
    fs_RdwrReq *    rdwr;
    int32           canDoDealloc = 0;
    evm_volSize_t   size;

#ifdef BUDDY_INCORE
    // it could happen when upgrading from build with the flag off
    // to a build with the flag on
    // FS_ASSERT(buf->inum != FS_FREE_BUDDY_TREE_INODE);
#endif // BUDDY_INCORE        
     
    NFX_SPIN_LOCK(&buf->spin);
    if (buf->logRefCount != 0)
    {
        // Can't write this buf until log buffers
        // holding it have been flushed to disk.
        // Set our flag and wait for the event.
        FS_ASSERT(!(buf->flags & FS_BUF_WAIT_LOGBUF));
        buf->flags |= FS_BUF_WAIT_LOGBUF;
        //printf("LOGBUF_WAIT set %p 0x%x\n", buf, buf->flags);
#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_LOCK(&buf->spinHistory);
        buf->history[buf->indexHistory] = FS_BUF_HIST_CLONE_WAIT_LOG;
        buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
        buf->flagsHistory[buf->indexHistory] = buf->flags;
        buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
        buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
        buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
        NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
        NFX_SPIN_UNLOCK(&buf->spin);
        return canDoDealloc;
    }

    if (buf->clone != NULL)
    {
        // There is already a clone for this buf outstanding,
        // wait until it completes before starting this buf.
        FS_ASSERT(!(buf->flags & FS_BUF_WAIT_CLONE));
        buf->flags |= FS_BUF_WAIT_CLONE;
#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_LOCK(&buf->spinHistory);
        buf->history[buf->indexHistory] = FS_BUF_HIST_CLONE_WAIT;
        buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
        buf->flagsHistory[buf->indexHistory] = buf->flags;
        buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
        buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
        buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
        NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
        NFX_SPIN_UNLOCK(&buf->spin);
        fs_bufRelease(context, buf);
        return canDoDealloc;
    }

    if (!DLL_IS_EMPTY(&buf->candoHdr))
    {
        NFX_SPIN_UNLOCK(&buf->spin);
        canDoDealloc = fs_bufConvertCando(context, buf);
        NFX_SPIN_LOCK(&buf->spin);
    }

    if (!DLL_IS_EMPTY(&buf->undoHdr))
    {
        fs_Buf  *newBuf;
        
        BUF_TRACE(("vol %llx cloning buf %p blknum %lld for writeback",
                   buf->fileSystem->volId.whole_thing,
                   buf, fs_blockNumber(buf->block)));


        FS_ASSERT(buf->flags & FS_BUF_WB_STARTING);

        // Clone the buf.
        newBuf = fs_ramAlloc(FS_RAM_BUF_CLONE, sizeof *newBuf);
#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_LOCK(&spinalloc);
        fs_buf_allocate_ptr[fs_buf_allocate_index] = newBuf;
        fs_buf_allocate_oper[fs_buf_allocate_index] = FS_DEBUG_BUF_HISTORY_ALLOC_CLONE;
        fs_buf_allocate_index = (fs_buf_allocate_index + 1) % FS_DEBUG_BUF_HISTORY_ALLOC_MAX;
        NFX_SPIN_UNLOCK(&spinalloc);
#endif
        bzero(newBuf, sizeof *newBuf);
        fs_lockInit(&newBuf->lock, FS_LOCK_TYPE_BUF);
        NFX_SPIN_INIT(&newBuf->spin,SPINLOCK_LEVEL_UNKNOWN);        
#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_INIT(&newBuf->spinHistory, SPINLOCK_LEVEL_UNKNOWN);
#endif
        fs_lockShared(context, &newBuf->lock, NULL);

#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_LOCK(&newBuf->spinHistory);
        newBuf->history[newBuf->indexHistory] = FS_BUF_HIST_CLONE;
        newBuf->cpuHistory[newBuf->indexHistory] = nfxMyCPUId;
        newBuf->indexHistory = (newBuf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
        NFX_SPIN_UNLOCK(&newBuf->spinHistory);
#endif
#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_LOCK(&buf->spinHistory);
        buf->history[buf->indexHistory] = FS_BUF_HIST_CLONED;
        buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
        buf->flagsHistory[buf->indexHistory] = buf->flags;
        buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
        buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
        buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
        NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
        newBuf->fileSystem = buf->fileSystem;
        newBuf->block = buf->block;
        newBuf->inum = buf->inum;
        newBuf->page = buf->page;
        newBuf->flags = buf->flags;
        newBuf->logSegDep = buf->logSegDep;
        newBuf->clone = buf;
        newBuf->flags |= FS_BUF_IS_CLONE;

        // Copy the data.
        newBuf->u.data = fs_ramAlloc(FS_RAM_BUF_CLONE_DATA, FS_BLOCK_SIZE);
        memcpy(newBuf->u.data, buf->u.data, FS_BLOCK_SIZE);

#if defined(FS_DEBUG_SHADOW_BUF)
        // Make sure shadow compare wont assert.
        newBuf->shadow = newBuf->u.data;
#endif

        // Undo the uncommitted stuff.
        fs_logUndoBuf(context, newBuf, buf);

        // The clone inherits the log deps.
        buf->logSegDep.count = 0;
        buf->logSegDep.u.dword = 0;
        buf->headLogSeg = -1;

        buf->clone = newBuf;
        NFX_SPIN_UNLOCK(&buf->spin);

        BUF_TRACE(("writing clone buf %p logSegDeps 0x%llx",
                   newBuf, newBuf->logSegDep.u.dword));

        // Count these, we may want to handle this
        // differently - no need to write it if there
        // are no deps that can be cleared.
        if (newBuf->logSegDep.u.dword == 0)
            fs_cloneNoDeps += 1;

        fs_clonedBufs += 1;

        // We're done with the original buf.  Put it back
        // on the dirty list.
        bufList = fs_bufListForBuf(buf);

        // We've pushed the oldest dirty stuff into the clone,
        // put the original buf back on the MRD.
        NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
        NFX_SPIN_LOCK(&buf->spin);
        DLL_INSERT_HEAD(&(bufList->mrdHdr), buf, mrdEntry);
        FS_ASSERT(buf->flags & FS_BUF_DIRTY);
        buf->flags |= FS_BUF_MRD;
#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_LOCK(&buf->spinHistory);
        buf->history[buf->indexHistory] = FS_BUF_HIST_INSERT_MRD;
        buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
        buf->flagsHistory[buf->indexHistory] = buf->flags;
        buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
        buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
        buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
        NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
        NFX_SPIN_UNLOCK(&buf->spin);
        NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);


        buf = newBuf;
    }
    else
    {
        NFX_SPIN_UNLOCK(&buf->spin);
    }

    if (buf->inum < FS_MAX_META_INODE)
    {
        fs_diskWritebackType[buf->inum] += 1;
    }

    rdwr = fs_rdwrAlloc();

    NFX_SPIN_LOCK(&buf->spin);

    buf->flags |= FS_BUF_IO_BUSY;

#ifdef FS_DEBUG_BUF_HISTORY
    NFX_SPIN_LOCK(&buf->spinHistory);
    buf->history[buf->indexHistory] = FS_BUF_HIST_SET_IO_BUSY;
    buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
    buf->flagsHistory[buf->indexHistory] = buf->flags;
    buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
    buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
    buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
    NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif

    FS_ASSERT(buf->flags & FS_BUF_WB_STARTING);
    FS_ASSERT((buf->flags & FS_BUF_IS_CLONE) ||
              !DLL_IS_ISOLATED(buf, blkNumEntry));
    FS_ASSERT(fs_blockIsValid(buf->block));

    if (buf->flags & FS_BUF_TYPE_META)
    {
        BUF_TRACE(("vol %llx (%llx,%llx) write meta blk %lld rdwr %p",
                   buf->fileSystem->volId.whole_thing,
                   context->volId.whole_thing,
                   context->fileSystem->volId.whole_thing,
                   fs_blockNumber(buf->block), rdwr));

#if defined(FS_DEBUG_SHADOW_BUF)
        // Compare changes to ensure logged data reflects live state.
        if (buf->u.data != NULL)
        {
            if (buf->flags & FS_BUF_COMPACT)
            {
                FS_DBG_MEMCMP(buf->u.data, buf->shadow, FS_COMPACT_PAGE_SIZE);
            }
            else
            {
                FS_DBG_MEMCMP(buf->u.data, buf->shadow, FS_BLOCK_SIZE);
            }
        }
#endif

#if defined(FS_DEBUG_BUF_CKSUM)
        fs_bufComputeCksum(context, buf->block, buf->u.data);

        if (fs_bufDoReadVerify)
        {
            buf->flags |= FS_BUF_READ_VERIFY;
        }
#endif

        NFX_SPIN_UNLOCK(&buf->spin);

        NFX_SPIN_LOCK(&fs_ioSpin);
        fs_ioTotalMetaWrites += 1;
        NFX_SPIN_UNLOCK(&fs_ioSpin);

        NFX_SPIN_LOCK(&buf->fileSystem->spin);
        buf->fileSystem->ioTotalMetaWrites += 1;
        NFX_SPIN_UNLOCK(&buf->fileSystem->spin);

        if (fs_bufIsCompact(buf))
        { 
            size = FS_COMPACT_PAGE_SIZE;
        }  
        else 
        {
            size = FS_BLOCK_SIZE;
        }
    
        fs_volWriteAsync(context, 
                         rdwr, 
                         handle,
                         fs_blockNumber(buf->block) * FS_BLOCK_SIZE, 
                         size,
                         buf->u.data, 
                         FS_WB_EVENT_IO_DONE, 
                         0, 
                         buf);
    }
    else
    {
#if defined(FS_DEBUG_BUF_CKSUM)
        FS_ASSERT(!fs_bufFindCksum(context, buf));
#endif

        NFX_SPIN_UNLOCK(&buf->spin);

        // Pad each user write to a 1K (FS_MIN_IO_SIZE) boundary.
        FS_ASSERT(buf->u.user.codOffset == 0);
        if (buf->u.user.codLength % FS_MIN_IO_SIZE)
        {
            int32 padLen = FS_MIN_IO_SIZE -
                            buf->u.user.codLength % FS_MIN_IO_SIZE;
            fs_cod_fill_null(&buf->u.user.codHandle,
                             padLen,
                             buf->u.user.codLength);
            buf->u.user.codLength += padLen;
            FS_ASSERT(buf->u.user.codLength <= FS_BLOCK_SIZE);
        }

        BUF_TRACE(("vol %llx write user blk %lld len %d",
                   buf->fileSystem->volId.whole_thing,
                   fs_blockNumber(buf->block), buf->u.user.codLength));

        NFX_SPIN_LOCK(&fs_ioSpin);
        fs_ioTotalUserWrites += 1;
        NFX_SPIN_UNLOCK(&fs_ioSpin);

        NFX_SPIN_LOCK(&buf->fileSystem->spin);
        buf->fileSystem->ioTotalUserWrites += 1;
        NFX_SPIN_UNLOCK(&buf->fileSystem->spin);

        fs_volWriteAsyncUser(context,
                             rdwr,
                             handle,
                             fs_blockNumber(buf->block) * FS_BLOCK_SIZE,
                             buf->u.user.codLength,
                             buf->u.user.codHandle,
                             FS_WB_EVENT_IO_DONE,
                             buf);
    }

    return canDoDealloc;
}



/**
 * Initiate writeback of a dirty buf.
 *
 * Returns TRUE if the writeback daemon must be signaled
 * in order for the write to complete.
 * 
 * Caller must hold bufListGlobal spinlock.
 *
 * If caller holds buf exclusive, it will be demoted to
 * a shared lock and the lock will be transferred to the
 * writeback daemon.
 */
static boolean
fs_bufWriteBackBuf(
    fs_Context      *context,
    fs_BufList      *bufList,
    fs_Buf          *buf)
{
    fs_Context      *wbContext;
    boolean         needWakeup = FALSE;
    boolean         haveLock = FALSE;
    boolean         letDaemonSchedule = FALSE;

    ASSERT_LOCK_OWNED(&fs_bufListGlobal[bufList->type].spin);

    NFX_SPIN_LOCK(&buf->spin);

    wbContext = bufList->wbContext;

    FS_ASSERT(buf->flags & FS_BUF_DIRTY);

    if (!(buf->flags & FS_BUF_WB_STARTING))
    {
        // Writeback hasn't started on it yet.  If we have
        // the lock or can get it without blocking, start
        // writeback on it.

        if (context->fileSystem->ioPendingCount >=
                fs_incoreTune->perFsIoPendingMax ||
            fs_ioWritesPending >=
                fs_incoreTune->writesPendingMax ||
            fs_ioPending >=
                FS_RDWR_MAX)
        {
            // Let the writeback daemon schedule this, we're
            // at our limit already.
            letDaemonSchedule = TRUE;
            
            if (fs_lockOwnedExclusive(context, &buf->lock))
            {
                if (buf->flags & FS_BUF_TRANS_WRITE)
                {
                    fs_logRemoveBufLogSummary(context, buf);
                    buf->flags &= ~FS_BUF_TRANS_WRITE;
                }
                fs_lockSetDirty(context, &buf->lock);
                fs_lockCommit(context, &buf->lock);
            }

            goto noLockCheck;
        }

        if (fs_lockOwnedExclusive(context, &buf->lock))
        {
            if (buf->flags & FS_BUF_TRANS_WRITE)
            {
                fs_logRemoveBufLogSummary(context, buf);
                buf->flags &= ~FS_BUF_TRANS_WRITE;
            }
            fs_lockGiveDemote(context, wbContext, &buf->lock);
            haveLock = TRUE;
        }
        else if (fs_trylockShared(context, &buf->lock))
        {
            fs_lockGiveShared(context, wbContext, &buf->lock);
            haveLock = TRUE;
        }

        if (haveLock)
        {
            if (buf->flags & FS_BUF_WAITING_MRD)
            {
                DLL_REMOVE(&bufList->mrdWaitingHdr, buf, mrdEntry);
                buf->flags &= ~FS_BUF_WAITING_MRD;
#ifdef FS_DEBUG_BUF_HISTORY
                NFX_SPIN_LOCK(&buf->spinHistory);
                buf->history[buf->indexHistory] = FS_BUF_HIST_REMOVE_W_MRD_WB;
                buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
                buf->flagsHistory[buf->indexHistory] = buf->flags;
                buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
                buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
                buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
                NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
            }
            else if (buf->flags & FS_BUF_MRD)
            {
                DLL_REMOVE(&bufList->mrdHdr, buf, mrdEntry);
                buf->flags &= ~FS_BUF_MRD;
#ifdef FS_DEBUG_BUF_HISTORY
                NFX_SPIN_LOCK(&buf->spinHistory);
                buf->history[buf->indexHistory] = FS_BUF_HIST_REMOVE_MRD_WB;
                buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
                buf->flagsHistory[buf->indexHistory] = buf->flags;
                buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
                buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
                buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
                NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
            }
            else
            {
#ifdef FS_DEBUG_BUF_HISTORY
                NFX_SPIN_LOCK(&buf->spinHistory);
                buf->history[buf->indexHistory] = FS_BUF_HIST_REMOVE_L_MRD_WB;
                buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
                buf->flagsHistory[buf->indexHistory] = buf->flags;
                buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
                buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
                buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
                NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
                FS_ASSERT(!DLL_IS_ISOLATED(buf, mrdEntry));
                DLL_REMOVE(&bufList->logSegmentHdr[buf->headLogSeg],
                           buf, mrdEntry);
                buf->headLogSeg = -1;
            }

            buf->flags |= FS_BUF_WB_STARTING;

#ifdef FS_DEBUG_BUF_HISTORY
            NFX_SPIN_LOCK(&buf->spinHistory);
            buf->history[buf->indexHistory] = FS_BUF_HIST_SET_WB_STARTING;
            buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
            buf->flagsHistory[buf->indexHistory] = buf->flags;
            buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
            buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
            buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
            NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif

            NFX_SPIN_UNLOCK(&buf->spin);

            FS_ASSERT(fs_blockIsValid(buf->block));
            fs_bufDoWriteBack(context, bufList->wbHandle, buf);
        }
        else
        {
noLockCheck:
            // Couldn't get the lock, or letting the
            // writeback daemon scheule it, move it to the
            // waiting list and indicate the need to
            // wake up the daemon.

            if (!(buf->flags & FS_BUF_WAITING_MRD))
            {
                if (buf->flags & FS_BUF_MRD)
                {
                    buf->flags &= ~FS_BUF_MRD;
                    DLL_REMOVE(&bufList->mrdHdr, buf, mrdEntry);
#ifdef FS_DEBUG_BUF_HISTORY
                    NFX_SPIN_LOCK(&buf->spinHistory);
                    buf->history[buf->indexHistory] = FS_BUF_HIST_REMOVE_MRD_WB_2;
                    buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
                    buf->flagsHistory[buf->indexHistory] = buf->flags;
                    buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
                    buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
                    buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
                    NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
                }
                else
                {
                    FS_ASSERT(!DLL_IS_ISOLATED(buf, mrdEntry));
                    DLL_REMOVE(&bufList->logSegmentHdr[buf->headLogSeg],
                               buf, mrdEntry);
                    buf->headLogSeg = -1;
#ifdef FS_DEBUG_BUF_HISTORY
                    NFX_SPIN_LOCK(&buf->spinHistory);
                    buf->history[buf->indexHistory] = FS_BUF_HIST_REMOVE_L_MRD_WB_2;
                    buf->flagsHistory[buf->indexHistory] = buf->flags;
                    buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
                    buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
                    buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
                    buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
                    NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
                }

                buf->flags |= (FS_BUF_WAITING_MRD | FS_BUF_WB_STARTING);
                
                DLL_INSERT_HEAD(&bufList->mrdWaitingHdr, buf, mrdEntry);
#ifdef FS_DEBUG_BUF_HISTORY
                NFX_SPIN_LOCK(&buf->spinHistory);
                buf->history[buf->indexHistory] = FS_BUF_HIST_INSERT_W_MRD_WB_2;
                buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
                buf->flagsHistory[buf->indexHistory] = buf->flags;
                buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
                buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
                buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
                NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif

#ifdef FS_DEBUG_BUF_HISTORY
                NFX_SPIN_LOCK(&buf->spinHistory);
                buf->history[buf->indexHistory] = FS_BUF_HIST_SET_WB_STARTING_WAITING_MRD;
                buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
                buf->flagsHistory[buf->indexHistory] = buf->flags;
                buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
                buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
                buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
                NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
            }
            needWakeup = TRUE;

            NFX_SPIN_UNLOCK(&buf->spin);
        }

        BUF_TRACE(("vol %llx start writeback buf %p blk %lld",
                    buf->fileSystem->volId.whole_thing,
                    buf, fs_blockNumber(buf->block)));
    }
    else
    {
        // Writeback already started.  Daemon is either waiting for
        // the lock (we may currently own), or has finished writing it.

        if (fs_lockOwnedExclusive(context, &buf->lock))
        {
            if (buf->flags & FS_BUF_TRANS_WRITE)
            {
                fs_logRemoveBufLogSummary(context, buf);

                buf->flags &= ~FS_BUF_TRANS_WRITE;
                NFX_SPIN_UNLOCK(&buf->spin);

                fs_lockSetDirty(context, &buf->lock);
                fs_lockCommit(context, &buf->lock);
            }
            else
            {
                NFX_SPIN_UNLOCK(&buf->spin);
                fs_unlockExclusive(context, &buf->lock);
            }
        }
        else
        {
            NFX_SPIN_UNLOCK(&buf->spin);
        }
    }

    return needWakeup;

} /* fs_bufWriteBackBuf */


/**
 * Called to unlock and release an fs block, held with a exclusive lock,
 * back to the cache.
 *
 * Should normally only be used if buffer was NOT modified.  If buffer was
 * modified, FS_BUF_WRITE() should be used (eventually followed by
 * fs_logCommitEntry()).
 */
void
fs_bufReleaseExclusive(
    fs_Context  *context,
    fs_Buf      *buf)
{
    fs_BufList      *bufList;
    nfx_spinlock_t  *spinlock;

    BUF_TRACE(("vol %llx brelease excl buf %p blk %lld ino %lld page %lld", 
               buf->fileSystem->volId.whole_thing,
               buf, fs_blockNumber(buf->block), buf->inum,
               buf->page));

    bufList = fs_bufListForBuf(buf);
    spinlock = &fs_bufListGlobal[bufList->type].spin;

    NFX_SPIN_LOCK(spinlock);
    fs_unlockExclusive(context, &buf->lock);
    fs_bufMaybeAddToFreeList(context, bufList, buf);
    NFX_SPIN_UNLOCK(spinlock);
}


/**
 * Decrement the buflist dirty count, notify waiters if needed.
 *
 * Must be called with bufListGlobal spinlock held.
 */
static void
fs_bufDecrDirtyCount(
    fs_Context  *context,
    fs_BufList  *bufList,
    fs_Buf      *buf)
{
    fs_FileSystem   *fileSystem = context->fileSystem;
    fs_Waiter       *waiter, *waiterNext;
    int32           itype = fs_bufGetItype(buf);

    FS_ASSERT(bufList->count.dirtyBufs[itype] > 0);
    ASSERT_LOCK_OWNED(&fs_bufListGlobal[bufList->type].spin);

    //printf("bufList->count.dirtyBufs[%d] %d\n",
    //       itype, bufList->count.dirtyBufs[itype]);

    bufList->count.dirtyBufs[itype] -= 1;
    fs_bufListGlobal[bufList->type].count.dirtyBufs[itype] -= 1;

    bufList->count.dirtyBufs[FS_BUF_CACHE_ITYPE_TOTAL] -= 1;
    fs_bufListGlobal[bufList->type].count.dirtyBufs[FS_BUF_CACHE_ITYPE_TOTAL] -= 1;

    DLL_ITERATE_NEXT(waiter, waiterNext, &bufList->waiterHdr, dllEntry)
    {
        FS_ASSERT(waiter->numWait > 0);
        waiter->numWait -= 1;
        if (waiter->numWait == 0)
        {
            DLL_REMOVE(&bufList->waiterHdr, waiter, dllEntry);
            FS_ASSERT(bufList->numWaiters > 0);
            bufList->numWaiters -= 1;
            fs_notifyWaiter(context, fileSystem, waiter);
            fs_ramDealloc(FS_RAM_WAITER, waiter);
        }
    }

    if (!DLL_IS_ISOLATED(buf, pageDirtyEntry))
    {
        fs_BufPageHdr   *pageHdr;

        pageHdr = fs_bufPageHdrForInum(buf->fileSystem,
                                       bufList->type, buf->inum);
        FS_ASSERT(pageHdr->dirtyCount > 0);
        pageHdr->dirtyCount -= 1;
        // delete from dirty page hdr
        DLL_REMOVE(&pageHdr->bufDirtyHdr, buf, pageDirtyEntry);
#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_LOCK(&buf->spinHistory);
        buf->history[buf->indexHistory] = FS_BUF_HIST_REMOVE_PAGE_DIRTY;
        buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
        buf->flagsHistory[buf->indexHistory] = buf->flags;
        buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
        buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
        buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
        NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif

        DLL_ITERATE_NEXT(waiter, waiterNext, &pageHdr->waiterHdr, dllEntry)
        {
            FS_ASSERT(waiter->numWait > 0);
            waiter->numWait -= 1;
            if (waiter->numWait == 0)
            {
                DLL_REMOVE(&pageHdr->waiterHdr, waiter, dllEntry);
                FS_ASSERT(bufList->numWaiters > 0);
                bufList->numWaiters -= 1;
                fs_notifyWaiter(context, pageHdr, waiter);
                fs_ramDealloc(FS_RAM_WAITER, waiter);
            }
        }
    }

} /* fs_bufDecrDirtyCount */


/**
 * Free the resources and disk block associated with a buf.
 */
void
fs_bufFree(
    fs_Context  *context,
    fs_Buf      *buf,
    int32       option)
{
    fs_FileSystem   *fileSystem;
    fs_BufList      *bufList;
    fs_Block        block;
    uint64          inum, page;
    int32           bufListType;
    boolean         doDeallocate = (option & BUF_DO_DEALLOCATE) ? TRUE : FALSE;
    boolean         isReplay = (option & BUF_IS_REPLAY) ? TRUE : FALSE;
    boolean         addFlushRecord = (option & BUF_DONT_LOG_FLUSH) ? FALSE : TRUE;
    nfx_spinlock_t  *spinlock;

    FS_ASSERT((buf->flags & FS_BUF_IO_BUSY) == 0);
    FS_ASSERT((buf->flags & FS_BUF_FREE_DEFER) == 0);
    FS_ASSERT(fs_lockOwnedExclusive(context, &buf->lock));
    FS_ASSERT(context->fileSystem == buf->fileSystem);

    fileSystem = context->fileSystem;

    bufList = fs_bufListForBuf(buf);
    bufListType = bufList->type;
    spinlock = &fs_bufListGlobal[bufList->type].spin;

    if (bufList->type == FS_BUFLIST_META)
    {
#if defined(FS_DEBUG_BUF_CKSUM)
        if (fs_blockIsValid(buf->block))
        {
            fs_bufReleaseCksum(context, buf);
        }
#endif // FS_DEBUG_BUF_CKSUM
    }
    
    BUF_TRACE(("vol %llx bfree %p  block %lld inum %lld page %lld"
               " flags %#x option %#x",
               buf->fileSystem->volId.whole_thing,
               buf, fs_blockNumber(buf->block), buf->inum, buf->page,
               buf->flags, option));

    if (buf->flags & FS_BUF_DIRTY)
    {
        BUF_TRACE(("  bfree dirty"));
        // If the buf is dirty, we can't simply throw it away,
        // even though it may not be relevant to the current live
        // view of the filesystem.  Until this transaction commits,
        // we cannot remove previous transactions log entries that
        // depend on this buffer, but since we can't write this buffer
        // to disk (its space is being deallocated), the normal method
        // for removing the log dependencies won't happen.  So, we mark
        // it as FREE_DEFER, and clean up the log during commit.
        //
        // Of course, none of this matters if this free is being
        // called during replay, which is simply trying to clear the
        // buf from the cache!

        NFX_SPIN_LOCK(spinlock);
        NFX_SPIN_LOCK(&buf->spin);

        if (!DLL_IS_ISOLATED(buf, mrdEntry))
        {
            // Get the buf off of the dirty list.
            if (buf->flags & FS_BUF_MRD)
            {
                DLL_REMOVE(&bufList->mrdHdr, buf, mrdEntry);
                buf->flags &= ~FS_BUF_MRD;
#ifdef FS_DEBUG_BUF_HISTORY
                NFX_SPIN_LOCK(&buf->spinHistory);
                buf->history[buf->indexHistory] = FS_BUF_HIST_REMOVE_MRD_FREE;
                buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
                buf->flagsHistory[buf->indexHistory] = buf->flags;
                buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
                buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
                buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
                NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
            }
            else if (buf->flags & FS_BUF_WAITING_MRD)
            {
                DLL_REMOVE(&bufList->mrdWaitingHdr, buf, mrdEntry);
                buf->flags &= ~FS_BUF_WAITING_MRD;
#ifdef FS_DEBUG_BUF_HISTORY
                NFX_SPIN_LOCK(&buf->spinHistory);
                buf->history[buf->indexHistory] = FS_BUF_HIST_REMOVE_W_MRD_FREE;
                buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
                buf->flagsHistory[buf->indexHistory] = buf->flags;
                buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
                buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
                buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
                NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
            }
            else
            {
                DLL_REMOVE(&bufList->logSegmentHdr[buf->headLogSeg],
                           buf, mrdEntry);
                buf->headLogSeg = -1;
#ifdef FS_DEBUG_BUF_HISTORY
                NFX_SPIN_LOCK(&buf->spinHistory);
                buf->history[buf->indexHistory] = FS_BUF_HIST_REMOVE_L_MRD_FREE;
                buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
                buf->flagsHistory[buf->indexHistory] = buf->flags;
                buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
                buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
                buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
                NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
            }
        }
        NFX_SPIN_UNLOCK(&buf->spin);

        fs_bufDecrDirtyCount(context, bufList, buf);

        NFX_SPIN_LOCK(&buf->spin);
        buf->flags |= FS_BUF_FREE_DEFER;
        buf->flags &= ~FS_BUF_DIRTY;
#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_LOCK(&buf->spinHistory);
        buf->history[buf->indexHistory] = FS_BUF_HIST_FREE_DEFER_RESET_DIRTY;
        buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
        buf->flagsHistory[buf->indexHistory] = buf->flags;
        buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
        buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
        buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
        NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
        NFX_SPIN_UNLOCK(spinlock);
        NFX_SPIN_UNLOCK(&buf->spin);

        // Make sure the buf is logged so commit will take the appropriate
        // action.
        if (!(buf->flags & FS_BUF_TRANS_WRITE))
        {
            fs_logAddBufLock(context, context->logSummary, buf);
            NFX_SPIN_LOCK(&buf->spin);
            buf->flags |= FS_BUF_TRANS_WRITE;
            NFX_SPIN_UNLOCK(&buf->spin);
        }
    }
    else if (buf->logSegDep.count > 0)
    {
        // If buf was not dirtied, remove any deps here.
        // (eg. during replay, an alloc followed by a buf flush
        // needs to remove the alloc dep)  We do it deferred,
        // since the log deps may not be reachable in the
        // in-core log active list yet (not until after this
        // transaction commits).

        // This case tends to occur when removing a COW file.
        // The btree code causes the copy to occur (which is
        // where the log dependency comes from), but doesn't
        // necessarily write the buf (which is why its not
        // dirty) before freeing it.

        NFX_SPIN_LOCK(&buf->spin);
        if (!(buf->flags & FS_BUF_TRANS_WRITE))
        {
            fs_logAddBufLock(context, context->logSummary, buf);
            buf->flags |= FS_BUF_TRANS_WRITE;
        }

        buf->flags |= FS_BUF_FREE_DEFER;
#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_LOCK(&buf->spinHistory);
        buf->history[buf->indexHistory] = FS_BUF_HIST_FREE_DEFER;
        buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
        buf->flagsHistory[buf->indexHistory] = buf->flags;
        buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
        buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
        buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
        NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
        NFX_SPIN_UNLOCK(&buf->spin);
    }

    block = buf->block;
    inum = buf->inum;
    page = buf->page;

    if ((buf->flags & FS_BUF_TYPE_USER) &&
        buf->zeroFillLogId)
    {
        // Buf has a zero-fill transaction still
        // pending in the log, that can now be
        // completed.
        //printf("fs_bufFree and buf->zeroFillLogId %d\n", buf->zeroFillLogId);
        fs_logComplete(context, buf->zeroFillLogId, buf->inum);
        buf->zeroFillLogId = 0;
    }

    NFX_SPIN_LOCK(spinlock);
    NFX_SPIN_LOCK(&buf->spin);
    fs_bufDelist(context, bufList, buf, FALSE);
    buf->block = FS_INVALID_BLOCK;
    buf->flags |= FS_BUF_INVALID;
#ifdef FS_DEBUG_BUF_HISTORY
    NFX_SPIN_LOCK(&buf->spinHistory);
    buf->history[buf->indexHistory] = FS_BUF_HIST_SET_INVALID;
    buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
    buf->flagsHistory[buf->indexHistory] = buf->flags;
    buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
    buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
    buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
    NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
    NFX_SPIN_UNLOCK(spinlock);
    NFX_SPIN_UNLOCK(&buf->spin);

    // Free the disk block.
    if (fs_blockIsValid(block) &&
        doDeallocate)   
    {
        if (buf->flags & FS_BUF_COW_UNK)
        {
            NFX_SPIN_UNLOCK(&buf->spin);
            if (fs_bufIsCopyOnWrite(context, buf))
                buf->flags |= FS_BUF_COW;
            buf->flags &= ~FS_BUF_COW_UNK;
            NFX_SPIN_UNLOCK(&buf->spin);
        }
        
        if (buf->flags & FS_BUF_COW)
        {
            BUF_TRACE(("vol %llx bfree block %lld inum %lld page %lld COW",
                        buf->fileSystem->volId.whole_thing,
                        fs_blockNumber(block), buf->inum, buf->page));
            FS_ASSERT(fs_blockIsValid(block));

            fs_refByteModifyBlock(context, fs_refbyteClearLive, buf->block);
        }
        else
        {
            BUF_TRACE(("  bfree freeing block %lld", fs_blockNumber(block)));
            fs_deallocateBlock(context, block);
        }
    }

    fs_bufDeallocCoreRes(context, buf);

    if (!(buf->flags & FS_BUF_TRANS_WRITE))
    {
        BUF_TRACE(("  bfree unlocking"));

        // The buf hasn't been logged for this transaction,
        // so we unlock it here.
        
        fs_lockSetDirty(context, &buf->lock);
        fs_bufCommit(context, buf);
    }

    if (!isReplay && addFlushRecord)
    {
        // Add a log record to to indicate that this
        // block was flushed.
        fs_logAddBufFlushRecord(context, bufListType, inum, page, block);
    }

} /* fs_bufFree */


/**
 * Comparison function for ordering by page number.
 */
static int
fs_bufPageCompare(
    const void *p0,
    const void *p1)
{
    const uint64    page0 = *(uint64 *) p0;
    const uint64    page1 = *(uint64 *) p1;

    if (page0 < page1)
        return -1;
    if (page0 > page1)
        return 1;

    FS_ASSERT(FALSE);              // array should not have duplicates
    UNREACHED();
}


/**
 * Free all the buf's for the specified inode starting with the
 * specified page.  Don't deallocate their disk space.  This is
 * used during file truncation to clear the buffer cache for a file,
 * or during log replay to clear the cache.
 */

// #define FS_DEBUG_PRINT_BUF_FREE_INODE     1

#ifdef FS_DEBUG_PRINT_BUF_FREE_INODE
static int32           max_total_cnt1 = 0;
static int32           max_total_cnt2 = 0;
static int32           max_total_cnt3 = 0;
static int32           max_sleep_cnt1 = 0;
static int32           max_sleep_cnt2 = 0;
static int32           max_sleep_cnt3 = 0;
#endif

#define FS_CHECK_CPU_USAGE  100

/**
 * Free the bufs on the specified temp dllHdr.
 */
static void
fs_bufFreeList(
    fs_Context      *context,
    fs_Inode        *inode,
    fs_BufPageHdr   *ownPageHdr,
    boolean         isReplay)
{
    fs_Buf      *buf, *bufNext;
    eeeRTC_t    beginTime, delta;
    int32       loopCnt = 0;
    int32       options = isReplay ? BUF_IS_REPLAY : 0;
#ifdef FS_DEBUG_PRINT_BUF_FREE_INODE
    int32       total_cnt2 = 0, sleep_cnt2 = 0;
#endif

    eee_rtc(&beginTime);

    // We have exclusive lock on the bufs in ownPageHdr, free them.
    DLL_ITERATE_NEXT(buf, bufNext, &ownPageHdr->bufChainHdr, tmpPageEntry)
    {
        //printf("bufFreeList inum %lld page %lld blknum %lld\n",
        //  inode->inum, buf->page, fs_blockNumber(buf->block));

        DLL_REMOVE(&ownPageHdr->bufChainHdr, buf, tmpPageEntry);

        fs_bufFree(context, buf, options);

#ifdef FS_DEBUG_PRINT_BUF_FREE_INODE
        total_cnt2++;
#endif
        if (loopCnt++ > FS_CHECK_CPU_USAGE)
        {
            loopCnt = 0;

            // Check the time but only use up to half of it, since
            // our caller may have hogged it too!
            eee_rtcDuration(&beginTime, &delta);
            /* Note I am doubling delta to avoid a division of
               maxCpuUsageRTC by 2, but the result is the same. */
            (void)eee_rtcIncr(&delta, &delta);
            if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
            {
                fs_sleep(context, 0);

#ifdef FS_DEBUG_PRINT_BUF_FREE_INODE
                sleep_cnt2++;
#endif
                eee_rtc(&beginTime);
            }
        }
    }

#ifdef FS_DEBUG_PRINT_BUF_FREE_INODE
        {
            int updates = 0;

            if (max_total_cnt2 < total_cnt2)
            {
                max_total_cnt2 = total_cnt2;
                updates = 1;
            }
            if (max_sleep_cnt2 < sleep_cnt2)
            {
                max_sleep_cnt2 = sleep_cnt2;
                updates = 1;
            }

//            if (updates) 
            {
                printf("fs_bufFreeList max  total cnt2 %d - sleep cnt2 %d\n",
                       max_total_cnt2, max_sleep_cnt2);
            }
        }
#endif

}


void
fs_bufFreeInode(
    fs_Context  *context,
    fs_Inode    *inode,
    uint64      startPage,
    boolean     isReplay)
{
    fs_BufPageHdr   *pageHdr, ownPageHdr;
    fs_BufList      *bufList;
    fs_Buf          *buf, *bufNext;
#define PAGE_ARRAY_MAX  8
    uint64          pageArray[PAGE_ARRAY_MAX], pageArraySize;
    uint64          *pageArrayMalloc = NULL, *pageArrayPtr;
    int32           type, numMissed;
    int32           options = isReplay ? BUF_IS_REPLAY : 0;
    eeeRTC_t        beginTime, delta;
    int32           loopCnt;
#ifdef FS_DEBUG_PRINT_BUF_FREE_INODE
    int32           total_cnt1 = 0;
    int32           total_cnt3 = 0;
    int32           sleep_cnt1 = 0;
    int32           sleep_cnt3 = 0;
#endif

    FS_ASSERT(fs_lockOwnedExclusive(context, &inode->lock));

bufFreeRestart:
    numMissed = 0;
    pageArrayPtr = pageArray;
    pageArraySize = PAGE_ARRAY_MAX;
    loopCnt = 0;

    eee_rtc(&beginTime);

    // Must check both buflists, since we don't
    // know if this inode is a user or meta file.
    // TODO - shouldn't the inode know this?!!
    for (type = 0; type < FS_BUFLIST_MAX; type++)
    {
        bufList = &context->fileSystem->bufList[type];

        DLL_INIT(&ownPageHdr.bufChainHdr);

        NFX_SPIN_LOCK(&fs_bufListGlobal[type].spin);
        pageHdr = fs_bufPageHdrForInum(inode->fileSystem, type,
                                       inode->inum);
        if (pageHdr == NULL)
        {
            NFX_SPIN_UNLOCK(&fs_bufListGlobal[type].spin);
            continue;
        }

        DLL_ITERATE_NEXT(buf, bufNext, &pageHdr->bufChainHdr, pageEntry)
        {
            if (buf->page >= startPage)
            {
                if (!fs_trylock(context, &buf->lock, FS_LOCK_MODE_EXCL))
                {
                    // Remember the page number, we'll do another lookup
                    // for it after we finish a complete first pass.
                    if (numMissed == pageArraySize)
                    {
                        uint64  *old = pageArrayMalloc;

                        // Allocate additional space.
                        pageArrayMalloc = fs_ramAlloc(FS_RAM_CORE_MISC,
                                    (pageArraySize * 2) * sizeof *pageArray);
                        memcpy(pageArrayMalloc, pageArrayPtr,
                               pageArraySize * sizeof *pageArray);
                        pageArrayPtr = pageArrayMalloc;
                        pageArraySize *= 2;
                        if (old)
                        {
                            fs_ramDealloc(FS_RAM_CORE_MISC, old);
                        }
                    }
                    *(pageArrayPtr + numMissed) = buf->page;
                    numMissed += 1;
                }
                else
                {
                    // Add it to our list of those we own and can
                    // free.
                    DLL_INSERT_TAIL(&ownPageHdr.bufChainHdr, buf, tmpPageEntry);

                    // Now that we have the lock, get it off the free list.
                    if (!DLL_IS_ISOLATED(buf, mruEntry))
                    {
                        int32 itype = fs_bufGetItype(buf);

                        DLL_REMOVE(&fs_bufListGlobal[bufList->type].mruHdr[itype],
                                   buf, mruEntry);
                        fs_bufListGlobal[bufList->type].count.freeBufs[itype] -= 1;
                    }
                }

#ifdef FS_DEBUG_PRINT_BUF_FREE_INODE
                total_cnt1++;
#endif
                if (loopCnt++ > FS_CHECK_CPU_USAGE)
                {
                    loopCnt = 0;

                    // Check the time, but save half of it for the call to
                    // bufFreeList.
                    eee_rtcDuration(&beginTime, &delta);
                    /* Note I am doubling delta to avoid a division of
                       maxCpuUsageRTC by 2, but the result is the same. */
                    (void)eee_rtcIncr(&delta, &delta);
                    if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
                    {
                        // Don't sleep with the spinlock held, so free what
                        // we have (if any), drop the lock, sleep then start
                        // again from the top.
                        NFX_SPIN_UNLOCK(&fs_bufListGlobal[type].spin);
                        fs_bufFreeList(context, inode, &ownPageHdr, isReplay);
                        if (pageArrayMalloc)
                        {
                            fs_ramDealloc(FS_RAM_CORE_MISC, pageArrayMalloc);
                            pageArrayMalloc = NULL;
                        }
                        fs_sleep(context, 0);
#ifdef FS_DEBUG_PRINT_BUF_FREE_INODE
                        sleep_cnt1++;
#endif
                        goto bufFreeRestart;
                    }
                }
            }
        }   
        NFX_SPIN_UNLOCK(&fs_bufListGlobal[type].spin);

        // We have exclusive lock on the bufs in ownPageHdr, free them.
        fs_bufFreeList(context, inode, &ownPageHdr, isReplay);


        // Now traverse the list of those we didn't get locked
        // and if they're still in the cache, block if necessary to
        // obtain their lock and then free them.  Here we must sort
        // them by page first, since otherwise we could deadlock.
        if (numMissed != 0)
        {
            int32   i;

            FS_ASSERT(numMissed <= pageArraySize);

            qsort(pageArrayPtr, numMissed, sizeof *pageArrayPtr,
                  fs_bufPageCompare);

            for (i = 0; i < numMissed; i++)
            {
                buf = NULL;
                if (type == FS_BUFLIST_META)
                {
                    buf = O_LOOKUP_EXCL(inode, *(pageArrayPtr + i));
                }
                else
                {
                    buf = O_USER_LOOKUP_EXCL(inode, *(pageArrayPtr + i));
                }
                if (buf != NULL)
                {
                    fs_bufFree(context, buf, options);
                }

#ifdef FS_DEBUG_PRINT_BUF_FREE_INODE
                total_cnt3++;
#endif
                if (loopCnt++ > FS_CHECK_CPU_USAGE)
                {
                    loopCnt = 0;

                    eee_rtcDuration(&beginTime, &delta);
                    if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
                    {
                        fs_sleep(context, 0);

#ifdef FS_DEBUG_PRINT_BUF_FREE_INODE
                        sleep_cnt3++;
#endif
                        eee_rtc(&beginTime);
                    }
                }
            }
            if (pageArrayMalloc)
            {
                fs_ramDealloc(FS_RAM_CORE_MISC, pageArrayMalloc);
            }
        }
#if defined(FS_DEBUG_CHECK_CACHE_AFTER_FREE)
{
        // verify all pages for this inode have been removed from cache
        int64 p, maxPage = ceilingDiv64(inode->dinode.size, FS_BLOCK_SIZE);

        for (p = startPage; p < maxPage; p++)
        {
            buf = NULL;
            if (type == FS_BUFLIST_META)
            {
                buf = O_LOOKUP_EXCL(inode, p);
            }
            else
            {
                buf = O_USER_LOOKUP_EXCL(inode, p);
            }
            FS_ASSERT(buf == NULL);
        }
        // printf("buf free inode: ok for inum %lld\n", inode->inum);
}
#endif

#ifdef FS_DEBUG_PRINT_BUF_FREE_INODE
        {
            int updates = 0;
            
            if (max_total_cnt1 < total_cnt1)
            {
                max_total_cnt1 = total_cnt1;
                updates = 1;
            }
            if (max_total_cnt3 < total_cnt3)
            {
                max_total_cnt3 = total_cnt3;
                updates = 1;
            }

            if (max_sleep_cnt1 < sleep_cnt1)
            {
                max_sleep_cnt1 = sleep_cnt1;
                updates = 1;
            }
            if (max_sleep_cnt3 < sleep_cnt3)
            {
                max_sleep_cnt3 = sleep_cnt3;
                updates = 1;
            }
        
//            if (updates)
            {
                printf("fs_bufFreeInode max  total cnt %d %d %d - sleep cnt %d %d %d\n",
                       max_total_cnt1, max_total_cnt2, max_total_cnt3, 
                       max_sleep_cnt1, max_sleep_cnt2, max_sleep_cnt3);
                
                printf("fs_bufFreeInode curr total cnt1/3 %d/%d - sleep cnt1/3 %d/%d\n",
                       total_cnt1, total_cnt3, 
                       sleep_cnt1, sleep_cnt3);
            }            
        }
#endif

    }

#undef PAGE_ARRAY_MAX

} /* fs_bufFreeInode */


/**
 * Flush all non-dirty bufs for the specified filesystem
 * from the buffer cache.
 */
void
fs_bufCacheFlushSnap(
    fs_Context  *context,
    fs_Inode    *inode,
    int32       snapnum)
{
    boolean acquired;
    int32   list;

    for (list = 0; list < FS_BUFLIST_MAX; list += 1)
    {
        fs_BufList      *bufList = &context->fileSystem->bufList[list];
        fs_Buf          *buf, *bufNext;
        int32           itype;

        for (itype = 0; itype < FS_BUF_CACHE_ITYPE_MAX; itype++)
        {

            NFX_SPIN_LOCK(&fs_bufListGlobal[list].spin);
            DLL_ITERATE_NEXT(buf, bufNext, &fs_bufListGlobal[list].mruHdr[itype],
                             mruEntry)
            {
                FS_ASSERT(fs_bufIsAvail(buf));

                if (buf->fileSystem != context->fileSystem)
                {
                    continue;
                }

                /*
                 * If snapnum is specified, remove this snap's access
                 * to this buf by turning off it's bit in the snapshots mask.
                 * But keep this buffer for other snapshots unless
                 * there are none left.
                 */
                if (snapnum && (buf->snapshots & (1ULL << snapnum)))
                {
                    buf->snapshots &= ~(1ULL << snapnum);
                    if (buf->snapshots)
                    {
                        // don't flush, still used by other snapshots or live
                        continue;
                    }
                }
 
                /*
                 * If inode is specified, free only those blocks which belong
                 * to this inode
                 */
                if (inode)
                {
                    if (buf->inum != inode->inum ||
                        ((buf->snapshots & (1ULL << inode->snapnum)) == 0))
                    {
                        continue;
                    }
                }
    
                BUF_TRACE(
           ("vol %llx flush buf %p blk %lld ino %lld offset %lld blk:%c off:%c",
                      buf->fileSystem->volId.whole_thing, buf,
                      fs_blockNumber(buf->block),
                      buf->inum, buf->page,
                      DLL_IS_ISOLATED(buf, blkNumEntry) ? 'N' : 'Y',
                      DLL_IS_ISOLATED(buf, pageEntry) ? 'N' : 'Y'));
    
                // We should not be blocked getting the lock since
                // we hold the cache spinlock and the buf is on the
                // free list.
                acquired = fs_trylockExclusive(context, &buf->lock);
                FS_ASSERT(acquired);
    
#if defined(FS_DEBUG_BUF_CKSUM)
                if (list == FS_BUFLIST_META &&
                    fs_blockIsValid(buf->block))
                {
                    fs_bufReleaseCksum(context, buf);
                }
#endif  // FS_DEBUG_BUF_CKSUM

                FS_ASSERT(fs_lockOwnedExclusive(context, &buf->lock));

                fs_bufDelist(context, bufList, buf, FALSE);
                fs_bufDeallocCoreRes(context, buf);

                fs_lockDestroy(context, &buf->lock);
                fs_ramDealloc(FS_RAM_BUF_PTR, buf);
#ifdef FS_DEBUG_BUF_HISTORY
                NFX_SPIN_LOCK(&spinalloc);
                fs_buf_allocate_ptr[fs_buf_allocate_index] = buf;
                fs_buf_allocate_oper[fs_buf_allocate_index] = FS_DEBUG_BUF_HISTORY_DEALLOC;
                fs_buf_allocate_index = (fs_buf_allocate_index + 1) % FS_DEBUG_BUF_HISTORY_ALLOC_MAX;
                NFX_SPIN_UNLOCK(&spinalloc);
#endif
            }
            NFX_SPIN_UNLOCK(&fs_bufListGlobal[list].spin);
        }
    }

} /* fs_bufCacheFlushSnap */


/**
 * Remove the log dependencies associated with buf.
 */
void
fs_bufRemoveLogDeps(
    fs_Context      *context,
    fs_Buf          *buf)
{
    FS_ASSERT(buf->logSegDep.count > 0);
    fs_logRemoveDeps(context, &buf->logSegDep, &buf->spin);
    buf->headLogSeg = -1;

} /* fs_bufRemoveLogDeps */


/**
 * Remove a cow buffer from its old cache location and re-insert it
 */
void
fs_bufCopyOnWriteReInsertBuf(
    fs_Context  *context,
    fs_Buf      *buf,
    fs_Block    oldBlock)
{
    fs_BufList *bufList;
    int32 hash, itype;

    FS_ASSERT(!DLL_IS_ISOLATED(buf, blkNumEntry));

    bufList = fs_bufListForBuf(buf);
    hash = fs_bufHashVolBlock(context->volId, oldBlock);
    NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
    DLL_REMOVE(&fs_bufListGlobal[bufList->type].bufsBlk[hash],
               buf, blkNumEntry);
#ifdef FS_BUF_BLK_CHECK_HASH_DEPTH
    fs_bufListGlobal[bufList->type].blkDepth[hash]--;
#endif
    if (DLL_IS_ISOLATED(buf, pageHashEntry))
    {
        // decrement usedCount before re-inserting back to blknum cache
        fs_bufDecrUsedCount(bufList, buf);
    }
    fs_bufBlknumCacheInsert(context, buf, bufList, NULL);

    itype = fs_bufGetItype(buf);
    bufList->count.cowCount[itype] += 1;
    fs_bufListGlobal[bufList->type].count.cowCount[itype] += 1;
    bufList->count.cowCount[FS_BUF_CACHE_ITYPE_TOTAL] += 1;
    fs_bufListGlobal[bufList->type].count.cowCount[FS_BUF_CACHE_ITYPE_TOTAL] += 1;

    NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);
}

/**
 * Copy a metadata buffer on write, allocating space for the
 * new buf.
 *
 * The buffer must be held exclusive.
 */
void
fs_bufCopyOnWriteMetadata(
    fs_Context  *context,
    fs_Inode    *inode,
    fs_Buf      *buf)
{
    fs_Block oldBlock = buf->block;
    fs_Block newBlock;
    int32 rc;

    FS_ASSERT(fs_lockOwnedExclusive(context, &buf->lock));
    
    if (!fs_bufIsCopyOnWrite(context, buf))
    {
        // Nothing to do, ignore this request.
        return;
    }

    BUF_TRACE(("vol %llx COW metadata inum %lld old %lld new %lld",
                buf->fileSystem->volId.whole_thing,
              buf->inum, fs_blockNumber(oldBlock), fs_blockNumber(newBlock)));

    newBlock = fs_allocateMetadataBlockSpecial(context, inode, oldBlock);
    FS_ASSERT(fs_blockIsValid(newBlock));

    // TODO - maybe defer the clearing of the old block ?
    fs_refByteModifyBlock(context, fs_refbyteClearLive, oldBlock);
    NFX_SPIN_LOCK(&buf->spin);
    buf->block = newBlock;
    buf->flags &= ~(FS_BUF_COW|FS_BUF_COW_UNK);
    buf->snapshots = 1;
    NFX_SPIN_UNLOCK(&buf->spin);

    // Dont use the FS_BUF_COPIED flag because there's no extent map
    // entry to remove.

    // Remove the buffer from its old cache location and re-insert it.
    fs_bufCopyOnWriteReInsertBuf(context, buf, oldBlock);
    rc = fs_logAddAllocRecord(context, buf, TRUE, oldBlock);
    FS_ASSERT(rc == NFX_OK);
}


/**
 * Signal the writeback daemon to wakeup for the specified buflist.
 * Has no effect if the writeback daemon is currently active for
 * the buflist.
 */
void
fs_bufSignalWriteback(
    fs_Context  *context,
    int32       bufListType,
    int32       eventType)
{
    esm_event_t     event;
    fs_BufList      *bufList = &context->fileSystem->bufList[bufListType];

    event.event = FS_WB_EVENT_WAKEUP;
    event.p1.u.p = bufList;
    event.p2.u.w = eventType;
    fs_daemonWakeup(context, &fs_bufListGlobal[bufList->type].spin,
                    bufList->wbHandle, &bufList->wbSignal, event);
}


/**
 * Perform allocation for the specified buf for handling
 * COW case.
 */
void
fs_bufAllocForCow(
    fs_Context  *context,
    fs_Buf      *buf)
{
    fs_Inode    *inode = I_GET_SHARED(0, buf->inum);

    fs_bufAllocBuf(context, inode, buf, FALSE);
    fs_inodeRelease(context, inode);
    
} /* fs_bufAllocForCow */


/**
 * Schedule I/O writeback for all dirty bufs on the
 * specified pageHdr.  Returns TRUE if the writeback
 * daemon must be signalled in order to complete the
 * writes.
 *
 * Must be called with bufListGlobal spinlock held.
 */
static boolean
fs_bufStartIOForPageHdr(
    fs_Context      *context,
    fs_BufList      *bufList, 
    fs_BufPageHdr   *pageHdr)
{
    fs_Buf  *buf, *bufNext;
    boolean needWakeup = FALSE;

    ASSERT_LOCK_OWNED(&fs_bufListGlobal[bufList->type].spin);

    DLL_ITERATE_NEXT(buf, bufNext, &pageHdr->bufDirtyHdr, pageDirtyEntry)
    {
        FS_ASSERT(buf->flags & FS_BUF_DIRTY);

        needWakeup |= fs_bufWriteBackBuf(context, bufList, buf);
    }

    return needWakeup;
}


/**
 * Start IO writeback on the specified inode's dirty
 * pages if the number of dirty pages exceeds a
 * threshold.
 */
void
fs_bufStartIoThreshold(
    fs_Context  *context,
    fs_Inode    *inode,
    int32       type)
{
    fs_BufList      *bufList = &context->fileSystem->bufList[type];
    fs_BufPageHdr   *pageHdr;
    boolean         needWakeup = FALSE;

    if (context->fileSystem->ioPendingCount >=
                    fs_incoreTune->perFsIoPendingMax ||
        fs_ioWritesPending >=
                    fs_incoreTune->writesPendingMax ||
        fs_ioPending >=
                    FS_RDWR_MAX)
    {
        // Sorry, too much pending already, don't start
        // more.
        return;
    }

    NFX_SPIN_LOCK(&fs_bufListGlobal[type].spin);
    pageHdr = fs_bufPageHdrForInum(context->fileSystem, type,
                                    inode->inum);
    if (!pageHdr)
    {
        NFX_SPIN_UNLOCK(&fs_bufListGlobal[type].spin);
        return;
    }

    if (type == FS_BUFLIST_USER)
    {
        if (pageHdr->dirtyCount >= fs_incoreTune->userDirtyThreshold)
        {
            needWakeup = fs_bufStartIOForPageHdr(context, bufList, pageHdr);
        }
    }
    else
    {
        if (pageHdr->dirtyCount >= fs_incoreTune->metaDirtyThreshold)
        {
            needWakeup = fs_bufStartIOForPageHdr(context, bufList, pageHdr);
        }
    }
    
    NFX_SPIN_UNLOCK(&fs_bufListGlobal[type].spin);

    if (needWakeup)
    {
        fs_bufSignalWriteback(context, type, FS_WB_WAKEUP_PAGE_THRESH);
    }
}


/**
 * Signal all filesystem's writeback daemons of the
 * specified type for wakeup.
 */
static void
fs_bufSignalWritebackAll(
    fs_Context  *context,
    int32       type)
{
    esm_event_t     event;
    int32           i, rc;

    event.event = FS_WB_EVENT_WAKEUP;
    event.p2.u.w = FS_WB_WAKEUP_ALL;

    for (i = 0; i < FS_FILE_SYSTEM_HASH_MAX; i++)
    {
        fs_FileSystem   *fs;

        NFX_SPIN_LOCK(&fs_fileSystemTableLock);
        for (fs = fs_fileSystems[i]; fs != NULL; fs = fs->next)
        {
            if (fs->state == FS_VOL_STATE_ONLINE ||
                fs->state == FS_VOL_STATE_PAUSED)
            {
                fs_BufList      *bufList = &fs->bufList[type];

                event.p1.u.p = bufList;
                rc = fs_daemonWakeup(context, &fs_bufListGlobal[bufList->type].spin,
                                     bufList->wbHandle, &bufList->wbSignal, event);

                BUF_TRACE(("vol 0x%llx wb type %d signalled, daemon %s awakened",
                    fs->volId.whole_thing, type, rc == TRUE ? "was" : "was not"));
            }
        }
        NFX_SPIN_UNLOCK(&fs_fileSystemTableLock);
    }
}


/**
 * Mark a buffer as dirty, given the bufList it belongs to.
 */
void
fs_bufDirtyForList(
    fs_Context  *context,
    fs_Buf      *buf,
    fs_BufList  *bufList,
    uint32      bufWriteFlags)
{
    int32   itype = fs_bufGetItype(buf);

    if (!(buf->flags & FS_BUF_DIRTY))
    {
        NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
        NFX_SPIN_LOCK(&buf->spin);
        // Test again with the spin lock held
        if (!(buf->flags & FS_BUF_DIRTY))
        {
            BUF_TRACE(("vol %llx mrd buf %p blk %lld", 
                       buf->fileSystem->volId.whole_thing,
                       buf, fs_blockNumber(buf->block)));

            FS_ASSERT(DLL_IS_ISOLATED(buf, mrdEntry));

            // In the COW case, we may have already set up
            // a log dependency (for the ALLOC record), in
            // which case we need to put it on the correct
            // logSegmentHdr instead of the mrdHdr.

            if (buf->logSegDep.count != 0)
            {
                FS_ASSERT(buf->headLogSeg >= 0 &&
                          buf->headLogSeg < FS_LOG_MAX_SEGMENTS);

                DLL_INSERT_HEAD(&bufList->logSegmentHdr[buf->headLogSeg], buf, mrdEntry);
                buf->flags |= FS_BUF_DIRTY;
#ifdef FS_DEBUG_BUF_HISTORY
                NFX_SPIN_LOCK(&buf->spinHistory);
                buf->history[buf->indexHistory] = FS_BUF_HIST_SET_DIRTY;
                buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
                buf->flagsHistory[buf->indexHistory] = buf->flags;
                buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
                buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
                buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
                NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
            }
            else
            {
                FS_ASSERT(buf->headLogSeg == -1);

                DLL_INSERT_HEAD(&bufList->mrdHdr, buf, mrdEntry);
                buf->flags |= (FS_BUF_DIRTY | FS_BUF_MRD);
#ifdef FS_DEBUG_BUF_HISTORY
                NFX_SPIN_LOCK(&buf->spinHistory);
                buf->history[buf->indexHistory] = FS_BUF_HIST_SET_DIRTY_MRD;
                buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
                buf->flagsHistory[buf->indexHistory] = buf->flags;
                buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
                buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
                buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
                NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
            }

            bufList->count.dirtyBufs[itype] += 1;
            fs_bufListGlobal[bufList->type].count.dirtyBufs[itype] += 1;

            bufList->count.dirtyBufs[FS_BUF_CACHE_ITYPE_TOTAL] += 1;
            fs_bufListGlobal[bufList->type].count.dirtyBufs[FS_BUF_CACHE_ITYPE_TOTAL] += 1;

            if (!DLL_IS_ISOLATED(buf, pageEntry))
            {
                fs_BufPageHdr   *pageHdr;

                pageHdr = fs_bufPageHdrForInum(buf->fileSystem,
                                               bufList->type, buf->inum);
                FS_ASSERT(pageHdr != NULL);
    
                // insert to dirty page hdr
                FS_ASSERT(DLL_IS_ISOLATED(buf, pageDirtyEntry));
                DLL_INSERT_HEAD(&pageHdr->bufDirtyHdr, buf, pageDirtyEntry);
                pageHdr->dirtyCount += 1;
#ifdef FS_DEBUG_BUF_HISTORY
                NFX_SPIN_LOCK(&buf->spinHistory);
                buf->history[buf->indexHistory] = FS_BUF_HIST_INSERT_PAGE_DIRTY;
                buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
                buf->flagsHistory[buf->indexHistory] = buf->flags;
                buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
                buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
                buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
                NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
            }

            NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);
            NFX_SPIN_UNLOCK(&buf->spin);
        
            // Check dirtySpace threshold.
            if (fs_bufListGlobal[bufList->type].count.dirtyBufs[itype] >
                fs_bufListGlobal[bufList->type].count.dirtyBufLimit[itype])
            {
                fs_bufSignalWritebackAll(context, bufList->type);
            }
        }
        else
        {
            NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);
            NFX_SPIN_UNLOCK(&buf->spin);
        }
    }
}


/**
 * Mark a buffer as dirty if it is not already so.
 */
void fs_bufDirty(
    fs_Context  *context,
    fs_Buf      *buf)
{
    if (!(buf->flags & FS_BUF_DIRTY))
    {
        fs_BufList *bufList = fs_bufListForBuf(buf);
        fs_bufDirtyForList(context, buf, bufList, 0);
    }
}


/**
 * Called to indicate that a buf has been modified.
 * Takes care of all COW and logging concerns.
 */
void
fs_bufWrite(
    fs_Context  *context,
    fs_Buf      *buf,       
    int32       offset,     // Byte offset where write started (n/a for
                            // user (cod) bufs.
    int32       length,     // Length of write in bytes (n/a for user
                            // (cod) bufs).
    uint32      flags,      // Optional flags.
    void        *bitmap,    // Optional bitmap for log record.
    fs_BitmapOperation *debugOp) // Optional for bitmap debugging.
{
    fs_BufList  *bufList;
    int32       hash;
    int32       logFlag = 0;
    int32       rc;
    boolean     log = FALSE;
    fs_Block    oldBlock = buf->block; // save for meta2 log record
    boolean     isBitmap = (flags &
                    (FS_BUF_WRITE_LOG_BIT_RESET | FS_BUF_WRITE_LOG_BIT_SET));

    BUF_TRACE(("vol %llx bufwrite %lld inode %lld page %lld offset %d len %d"
               " flags %#x bufflags %#x",
               buf->fileSystem->volId.whole_thing,
               fs_blockNumber(buf->block), buf->inum, buf->page,
               offset, length, flags, buf->flags));

    FS_ASSERT(fs_lockOwnedExclusive(context, &buf->lock));
    FS_ASSERT(offset + length <= FS_BLOCK_SIZE);
    FS_ASSERT(!(buf->flags & FS_BUF_FREE_DEFER));
    FS_ASSERT(context->fileSystem == buf->fileSystem);

    bufList = fs_bufListForBuf(buf);

    if (bufList->type == FS_BUFLIST_META)
    {
        if (buf->inum < FS_MAX_META_INODE)
        {
            NFX_SPIN_LOCK(&fs_bufListGlobal[FS_BUFLIST_META].spin);
            fs_transWriteType[buf->inum] += 1;
            NFX_SPIN_UNLOCK(&fs_bufListGlobal[FS_BUFLIST_META].spin);
        }

#if defined(FS_DEBUG_SHADOW_BUF)
        // Update the shadow buffer.  We'll compare its contents to
        // the live buffer at commit time.
        FS_DBG_UPDATE_SHADOW(buf, offset, length);
#endif
    }
#ifdef NEW_ZF
    else if (!DLL_IS_EMPTY(&(context->fileSystem->zfInodeHdr))) // USER_DATA
    {
        fs_Inode *  inode = I_GET_SHARED(0, buf->inum);
        
        if (inode->zfflags & INO_FLAGS_UZERO)
        {
            NFX_SPIN_LOCK(&(context->fileSystem->zfSpin));
            // Check if the inode is still part of the ZF list.
            if (fs_zfGetInode(context, buf->inum) != NULL)
            {
                // Delete the page from the list of pages to zero fill.
                fs_zfDelPage(context, buf->inum, buf->page, FALSE);
            }
            NFX_SPIN_UNLOCK(&(context->fileSystem->zfSpin));
        }

        // Release the inode
        fs_inodeRelease(context, inode);
    }
#endif // NEW_ZF

    if (buf->flags & FS_BUF_COW_UNK)
    {
        boolean isCow;

        isCow = fs_bufIsCopyOnWrite(context, buf);

        NFX_SPIN_LOCK(&buf->spin);
        if (isCow)
            buf->flags |= FS_BUF_COW;
        buf->flags &= ~FS_BUF_COW_UNK;
        NFX_SPIN_UNLOCK(&buf->spin);
    }

    fs_bufDirtyForList(context, buf, bufList, flags);

    if (flags & FS_BUF_WRITE_LOG)
    {
        if (flags & FS_BUF_WRITE_ZERO)
        {
            logFlag |= FS_LOG_ADD_FLAG_ZERO;
        }
        if (flags & FS_BUF_WRITE_ADD_UNDO)
        {
            logFlag |= FS_LOG_ADD_FLAG_UNDO;
        }

        /*
         * Lock the buffer, but don't log it when eek is running.
         */
         if (!context->eek.inUse)
            log = TRUE;
    }

    if ((flags & FS_BUF_LOG_ADD_BUF) &&
         !(buf->flags & FS_BUF_TRANS_WRITE))
    {
        NFX_SPIN_LOCK(&buf->spin);
        FS_ASSERT (!(buf->flags & FS_BUF_TRANS_WRITE));
        BUF_TRACE(("vol %llx bufwrite log buf blk %lld hdl: %lld",
                   buf->fileSystem->volId.whole_thing,
                   fs_blockNumber(buf->block),
                   (uint64)((uintptr_t)context->smHandle)));
        buf->flags |= FS_BUF_TRANS_WRITE;
        NFX_SPIN_UNLOCK(&buf->spin);
        // First write for this transaction,
        // add the buf to the log summary.
        fs_logAddBufLock(context, context->logSummary, buf);
    }

    if (buf->flags & FS_BUF_COW)
    {
        NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
        NFX_SPIN_LOCK(&buf->spin);
        // Test again with the spin lock held
        if (buf->flags & FS_BUF_COW)
        {
            int32 itype = fs_bufGetItype(buf);
            BUF_TRACE(("vol %llx bufwrite inum %lld block %lld COW",
                       buf->fileSystem->volId.whole_thing,
                       buf->inum, fs_blockNumber(oldBlock)));

            FS_ASSERT(buf->inum != 0 && buf->page != FS_BUF_NO_PAGE);
            FS_ASSERT(fs_blockIsValid(oldBlock));

            hash = fs_bufHashVolBlock(context->volId, oldBlock);

            DLL_REMOVE(&fs_bufListGlobal[bufList->type].bufsBlk[hash],
                        buf, blkNumEntry);
#ifdef FS_BUF_BLK_CHECK_HASH_DEPTH
            fs_bufListGlobal[bufList->type].blkDepth[hash]--;
#endif
            bufList->count.cowCount[itype] += 1;
            fs_bufListGlobal[bufList->type].count.cowCount[itype] += 1;
            bufList->count.cowCount[FS_BUF_CACHE_ITYPE_TOTAL] += 1;
            fs_bufListGlobal[bufList->type].count.cowCount[FS_BUF_CACHE_ITYPE_TOTAL] += 1;
            NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);

            buf->block = FS_INVALID_BLOCK;
            buf->flags &= ~FS_BUF_COW;
            buf->flags |= FS_BUF_COPIED;
            buf->snapshots = 1;
            NFX_SPIN_UNLOCK(&buf->spin);

            // do this as a nested transaction to avoid lock
            // order problems with COW (PR #1582)
            fs_logNestTransaction(context);

            fs_refByteModifyBlock(context, fs_refbyteClearLive, oldBlock);
            fs_bufAllocForCow(context, buf);

            // commit the nested transaction.
            fs_logCommitEntry(context);
        }
        else
        {
            NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);
            NFX_SPIN_UNLOCK(&buf->spin);
        }
    }

    // add log records after new block is known (if COW)
    // new data should be applied to the new block for replay

    if (log)
    {
        if (isBitmap)
        {
            rc = fs_logAddBitmapRecord(context,
                                       buf,
                                       offset,
                                       length,
                                       bitmap,
                                       (flags & FS_BUF_WRITE_LOG_BIT_SET) ?
                                            FS_LOG_REC_BITMAP_SET :
                                            FS_LOG_REC_BITMAP_RESET,
                                       debugOp);
        }
        else
        {
            rc = fs_logAddDataRecord(context,
                                     buf,
                                     offset,
                                     length,
                                     fs_blockNumber(oldBlock),
                                     logFlag);
        }
        if (rc != NFX_OK)
        {
            FS_ASSERT(FALSE);
            return;
        }
    }

    // Take this opportunity to convert any cando records
    // to permanent status.
    fs_bufConvertCando(context, buf);
}


/**
 * Return true if the buf has been dirtied by the current transaction.
 */
boolean
fs_bufWasModified(
    fs_Context  *context,
    fs_Buf      *buf)
{
    return (buf->flags & FS_BUF_TRANS_WRITE);

} /* fs_bufWasModified */


/**
 * Called at transaction commit to commit changes
 * to a buf and release it.
 */
void
fs_bufCommit(
    fs_Context  *context,
    fs_Buf      *buf)
{
    fs_BufList      *bufList;
    nfx_spinlock_t  *spinlock;

    bufList = fs_bufListForBuf(buf);
    spinlock = &fs_bufListGlobal[bufList->type].spin;

    FS_ASSERT(context->fileSystem == buf->fileSystem);
    FS_ASSERT(DLL_IS_ISOLATED(buf, mruEntry));

    if (buf->flags & FS_BUF_FREE_DEFER)
    {
        FS_TRACE(context, FS_TRACE_BUF,
                 ("  deferred buf free %p, rem logdeps", buf));

        NFX_SPIN_LOCK(&buf->spin);
        buf->flags &= ~FS_BUF_FREE_DEFER;
#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_LOCK(&buf->spinHistory);
        buf->history[buf->indexHistory] = FS_BUF_HIST_FREE_DEFER_COMMIT;
        buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
        buf->flagsHistory[buf->indexHistory] = buf->flags;
        buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
        buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
        buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
        NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
        NFX_SPIN_UNLOCK(&buf->spin);

        if (buf->logSegDep.count > 0)
        {
            fs_bufRemoveLogDeps(context, buf);
        }

        NFX_SPIN_LOCK(&buf->spin);
        if (buf->logRefCount != 0)
        {
            // Don't let this buf get reused until the outstanding
            // logBufs have made it to disk.

            FS_ASSERT(fs_lockOwnedExclusive(context, &buf->lock));
            FS_ASSERT(buf->flags & FS_BUF_TYPE_META);

            buf->flags |= FS_BUF_FREE_LOGBUF;
            //printf("LOGBUF_FREE set %p 0x%x\n", buf, buf->flags);
        }
        NFX_SPIN_UNLOCK(&buf->spin);
    }

    if ((buf->flags & FS_BUF_TYPE_USER) &&
        !fs_lockOwnedExclusive(context, &buf->lock))
    {
        // If user buf, and I'm no longer the owner,
        // this indicates that the writeback daemon has taken
        // control (or IO already done).

        // Clear the TRANS_WRITE flag so the next transaction to
        // write the buf will cause it to be added to that transaction's
        // log summary.
        NFX_SPIN_LOCK(&buf->spin);
        buf->flags &= ~FS_BUF_TRANS_WRITE;
        NFX_SPIN_UNLOCK(&buf->spin);

        return;
    }    

#if defined(FS_DEBUG_SHADOW_BUF)
    // Compare changes to ensure logged data reflects live state.
    if (buf->flags & FS_BUF_TYPE_META) 
    {
        // Don't compare bufs for the inode file, the dinode
        // buf was checked when unlocking the inode.
        //
        // Don't compare bufs for the ref count file, since they're
        // modified concurrently.
        if (buf->inum != FS_INODE_INODE &&
            buf->inum != FS_REF_COUNT_INODE &&
            buf->u.data != NULL)
        {
            if (buf->flags & FS_BUF_COMPACT)
            {
                FS_DBG_MEMCMP(buf->u.data, buf->shadow, FS_COMPACT_PAGE_SIZE);
            }
            else
            {
                FS_DBG_MEMCMP(buf->u.data, buf->shadow, FS_BLOCK_SIZE);
            }
        }
    }
#endif

    FS_ASSERT(fs_lockOwnedExclusive(context, &buf->lock));

    // Clear the TRANS_WRITE flag so the next transaction to
    // write the buf will cause it to be added to that transaction's
    // log summary.
    NFX_SPIN_LOCK(&buf->spin);
    buf->flags &= ~FS_BUF_TRANS_WRITE;
    NFX_SPIN_UNLOCK(&buf->spin);

    NFX_SPIN_LOCK(spinlock);
    fs_lockCommit(context, &buf->lock);
    fs_bufMaybeAddToFreeList(context, bufList, buf);
    NFX_SPIN_UNLOCK(spinlock);

} /* fs_bufCommit */


/**
 * Start I/O on all dirty user bufs for the specified inode, and block
 * until the I/O is completed.
 *
 * Any bufs that the current context holds in exclusive mode will
 * be demoted and transferred to the writeback daemon.
 *
 * If timeoutSecs is non-zero, it is the timeout value used in seconds.
 * If the timeout expires before all the I/O is completed, FALSE is
 * returned.  If timeoutSecs is zero, a very conservative timeout value
 * is chosen for the I/O completion, and if it expires the volume will
 * be aborted.
 */
static boolean
fs_bufTrySyncInodeBufs(
    fs_Context  *context,
    fs_Inode    *inode,
    int32       timeoutSecs)
{
    fs_Waiter       *waiter;
    fs_BufPageHdr   *pageHdr;
    esm_event_t     wakeEvent;
    fs_BufList      *bufList;
    int32           type, yieldTimeout;
    boolean         needWakeup;
    int32           numWait;
    
    for (type = 0; type < FS_BUFLIST_MAX; type++)
    {
        waiter = fs_ramAlloc(FS_RAM_WAITER, sizeof *waiter);
        waiter->handle = context->smHandle;
        waiter->event.event  = FS_THREAD_EVENT_SYNC;
        waiter->event.p2.u.p = NULL; 
        bufList = &context->fileSystem->bufList[type];

        NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
        pageHdr = fs_bufPageHdrForInum(context->fileSystem, bufList->type,
                                       inode->inum);

        if (pageHdr == NULL ||
            pageHdr->dirtyCount == 0)
        {
            NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);
            fs_ramDealloc(FS_RAM_WAITER, waiter);
            continue;
        }

        waiter->event.p1.u.p = pageHdr;
        DLL_ISOLATE_ELEM(waiter, dllEntry);
        DLL_INSERT_TAIL(&pageHdr->waiterHdr, waiter, dllEntry);
        numWait = pageHdr->dirtyCount;
        waiter->numWait = numWait;
        waiter->numWaitStart = numWait;
        bufList->numWaiters += 1;

        needWakeup = fs_bufStartIOForPageHdr(context, bufList, pageHdr);

        NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);

        if (needWakeup)
        {
            fs_bufSignalWriteback(context, bufList->type, FS_WB_WAKEUP_PAGE_SYNC);
        }

        if (timeoutSecs == 0)
        {
#define SECS_PER_IO_MULT  10
            yieldTimeout = FS_YIELD_TIME_IO + (SECS_PER_IO_MULT * numWait);

            if (bufList->type == FS_BUFLIST_META)
            { 
                fs_yield(context, &wakeEvent, yieldTimeout, FS_YIELD_TYPE_IO_MW);
            }
            else
            {
                fs_yield(context, &wakeEvent, yieldTimeout, FS_YIELD_TYPE_IO_UW);
            }
        
        }
        else
        {
            fs_yield(context, &wakeEvent, timeoutSecs, FS_YIELD_TYPE_IO_COMMIT);
            if (wakeEvent.event == THREAD_EVENT_TIMEOUT)
            {
                return FALSE;
            }
        }

        // Make sure the event we got is the event
        // we were waiting for!
        if (wakeEvent.event != FS_THREAD_EVENT_SYNC ||
            wakeEvent.p1.u.p != pageHdr)
        {
           fs_error(NULL, FS_INTERNAL_ERR, "bufSyncInodeBufs wrong event");
        }
    }

    return TRUE;
}


/**
 * Start I/O on all dirty user bufs for the specified inode, and block
 * until the I/O is completed.
 *
 * Any bufs that the current context holds in exclusive mode will
 * be demoted and transferred to the writeback daemon.
 */
void
fs_bufSyncInodeBufs(
    fs_Context         *context,
    fs_Inode           *inode)
{
    fs_bufTrySyncInodeBufs(context, inode, 0);
}


/**
 * Start I/O on all dirty user bufs for the specified inode, and block
 * until the I/O is completed.
 *
 * Any bufs that the current context holds in exclusive mode will
 * be demoted and transferred to the writeback daemon.
 *
 * If after the specified timeout period (in seconds) is reached
 * and the I/O is not completed, return FALSE.
 */
boolean
fs_bufSyncInodeBufsAndTimeout(
    fs_Context  *context,
    fs_Inode    *inode,
    int32       timeoutSecs)
{
    return fs_bufTrySyncInodeBufs(context, inode, timeoutSecs);
}


/**
 * Comparison function for ordering bufs by page number.
 */
static int
fs_bufBufPageCompare(
    const void *p0,
    const void *p1)
{
    const fs_Buf    *buf0 = *(fs_Buf **) p0;
    const fs_Buf    *buf1 = *(fs_Buf **) p1;

    if (buf0->page < buf1->page)
        return -1;
    if (buf0->page > buf1->page)
        return 1;

    FS_ASSERT(FALSE);              // array should not have duplicates
    UNREACHED();
}


/**
 * Allocate extents for all the specified buffers, which are blocks of
 * the specified inode.
 */
void
fs_bufAllocDoBufs(
    fs_Context *context,
    fs_Inode   *inode,
    int        numInBufs,
    fs_Buf     **bufs,
    boolean    commitInode) // bind the inode to the transaction
{
    int32           i, rc;
    boolean         isCow;
    fs_Block        oldBlock;
    
    // We have a batch of bufs to allocate for, allocate a log summary
    // if necessary.
    if (context->logSummary == NULL) 
    {
        fs_logBeginTransaction(context);
    }
        
    //printf("inode is %p, inum %lld\n", inode, inode->inum);

    for (i = 0; i < numInBufs; )
    {
        fs_Buf  *buf = bufs[i];

        if (buf == NULL)
        {
            fs_error(context, FS_INTERNAL_ERR, "alloc list");
        }

        //printf("allocing for buf %p page %lld\n", buf, buf->page);

        FS_ASSERT(!(buf->flags & FS_BUF_IO_BUSY));
        FS_ASSERT(buf->lock.mode == FS_LOCK_MODE_EXCL);
        FS_ASSERT(!fs_blockIsValid(buf->block));
    
        BUF_TRACE(("vol %llx defer alloc inum %lld page %lld",
                   buf->fileSystem->volId.whole_thing,
                    buf->inum, buf->page));

        if (buf->flags & FS_BUF_TYPE_META)
        {
            fs_Block block;
            fs_BufList *bufList = fs_bufListForBuf(buf);

            if (buf->inum == FS_INODE_INODE)
            {
                // Allocate a block for the inode file.
                block = fs_allocateInodeFileBlock(context, inode, buf->page);
            }
            else
            {
                // Allocate a block for the metadata file.
                block = fs_allocateMetadataBlock(context, inode);
            }

            buf->block = block;

            // Update the btree.
            // No need to increment inode's numblocks if we are
            // just doing a swap of old with new

            if (buf->flags & FS_BUF_COPIED)
            {
                // PR #1857: save oldBlock/source for cow alloc record
                // before modifying emap
                oldBlock = fs_emap_lookup_block(context, inode, buf->page);
                isCow = TRUE;

                // do this as an "atomic" emap remove/add
                // to avoid lock order problems with COW (PR #1582)
                fs_emap_replace_block(context, inode, buf->page, block);
            }
            else
            {
                fs_emap_add_block(context, inode, buf->page, block);
                inode->dinode.numblocks += 1;
                fs_inodeWrite(context, inode, INO_WRITE_LOG);
                isCow = FALSE;
            }

            // Add alloc record to the log, include cow info
            // TODO - this dies when deferred alloc dies.
            rc = fs_logAddAllocRecord(context, buf, isCow, oldBlock);
            FS_ASSERT(rc == NFX_OK);
            NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
            fs_bufBlknumCacheInsert(context, buf, bufList, inode);
            NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);
            i += 1;
        }
        else
        {
            fs_Extent extent;
            fs_BufList *bufList = fs_bufListForBuf(buf);
            int32 numBlocks;
            int32 numContig = 0, j;

            qsort(bufs, numInBufs, sizeof *bufs, fs_bufBufPageCompare);

            // Find contiguous pages.
            while (i + numContig < numInBufs - 1 &&
                   bufs[i + numContig]->page + 1 ==
                   bufs[i + 1 + numContig]->page)
            {
                //printf("buf %p page %lld has buf %p contig\n",
                //      bufs[i+numContig],
                //      bufs[i+numContig]->page,
                //      bufs[i+numContig+1]);

                numContig += 1;
            }

            //printf("inum %lld %d contig (of %d dirty) from page %lld\n",
            //inode->inum, numContig + 1, numInBufs, buf->page);

            // Allocate for the buf(s).
            extent = fs_allocateExtent(context,
                                       inode,
                                       bufs[i]->page,
                                       numContig + 1,
                                       FALSE);

            //printf("got extent length of %d\n", fs_extentLength(extent));

            numBlocks = fs_extentLength(extent);
            for (j = 0; j < fs_extentLength(extent); j += 1)
            {
                fs_Buf *alloc = bufs[i + j];

                if (alloc->flags & FS_BUF_COPIED)
                {
                    isCow = TRUE;
                    oldBlock = fs_emap_lookup_block(context, inode, alloc->page);
                    BUF_TRACE(("  COW remove old page"));
                    fs_emap_remove_block(context, inode, alloc->page);
                    numBlocks--;    // COW block already counted
                }
                else
                {
                    isCow = FALSE;
                }

                alloc->block = fs_extentBlock(extent, j);
                //printf("got block %lld for buf %p\n",
                //    fs_blockNumber(alloc->block), alloc);

                // Add alloc record to the log, include cow info
                rc = fs_logAddAllocRecord(context, alloc, isCow, oldBlock);
                FS_ASSERT(rc == NFX_OK);

                NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
                fs_bufBlknumCacheInsert(context, alloc, bufList, inode);
                NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);
            }

            // Update the btree. And adjust numblocks

            fs_emap_add(context, inode, buf->page, extent);
            if (numBlocks) {
                inode->dinode.numblocks += numBlocks;
                fs_inodeWrite(context, inode, INO_WRITE_LOG);
            }

            //printf("increment bufs i by %d\n", fs_extentLength(extent));

            i += fs_extentLength(extent);
        }
    } // for (i)

    FS_ASSERT(numInBufs != 0);

    // Now commit the transaction.  The inode may or may not have
    // been touched during emap update, so we need to ensure its
    // added to the log here so it will be unlocked during the
    // commit.
    if (commitInode)
        fs_inodeLockTransaction(context, inode, context->logSummary);

}


/**
 * Signal the writeback daemon to wakeup for a specified buffer
 * list and wait for it to complete the IO for this filesystem.
 */
void fs_bufSyncWriteback(
    fs_Context  *context,
    int32       type)
{
    esm_event_t event;
    fs_Waiter   *waiter = fs_ramAlloc(FS_RAM_WAITER, sizeof *waiter);
    fs_BufList  *bufList = &context->fileSystem->bufList[type];

    NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
    if (bufList->count.dirtyBufs[FS_BUF_CACHE_ITYPE_TOTAL] == 0)
    {
        NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);
        BUF_TRACE(
         ("vol %llx %s writeback returned immediately (nothing to do)",
                 context->volId.whole_thing,
                 bufList->type == FS_BUFLIST_META ? "meta" : "user"));
        fs_ramDealloc(FS_RAM_WAITER, waiter);
        return;
    }

    waiter->handle = context->smHandle;
    waiter->event.event  = FS_THREAD_EVENT_SYNC;
    waiter->event.p1.u.p = context->fileSystem;
    waiter->event.p2.u.p = NULL;

    DLL_ISOLATE_ELEM(waiter, dllEntry);
    DLL_INSERT_TAIL(&bufList->waiterHdr, waiter, dllEntry);
    waiter->numWait = bufList->count.dirtyBufs[FS_BUF_CACHE_ITYPE_TOTAL];
    waiter->numWaitStart = waiter->numWait;
    bufList->numWaiters += 1;
    NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);

    fs_bufSignalWriteback(context, type, FS_WB_WAKEUP_SYNC);

    if (bufList->type == FS_BUFLIST_META)
    { 
        fs_yield(context, &event, FS_YIELD_TIME_IO * 10, FS_YIELD_TYPE_IO_MW);
    }
    else
    {
        fs_yield(context, &event, FS_YIELD_TIME_IO * 10, FS_YIELD_TYPE_IO_UW);
    }

    if (event.event != FS_THREAD_EVENT_SYNC ||
        event.p1.u.p != context->fileSystem)
    {
        fs_error(context, FS_INTERNAL_ERR, "fs_bufSyncWriteback wrong event");
    }
}


/**
 * Signal the writeback daemon to write all the buffer lists for the
 * current context, and wait until it's completed all of them.
 */
void fs_bufSyncWritebackLists(
    fs_Context *context)
{
    int i;
    for (i = 0; i < FS_BUFLIST_MAX; i++) {
        fs_bufSyncWriteback(context, i);
    }
}


/**
 * Initiate writeback of dirty buffers for the indicated list.
 */
#ifdef FS_DEBUG_WRITEBACK_TIMING
eeeRTC_t fs_userBufMaxRuntime;
eeeRTC_t fs_metaBufMaxRuntime;
#endif

uint32 fs_userBufWbMaxPerLoop = 0;
uint32 fs_metaBufWbMaxPerLoop = 0;


static boolean
fs_bufWriteBackList(
    fs_Context      *context,
    fs_BufList      *bufList,
    fs_BufDllHdr    *dllHdr)
{
#define BUF_WB_LOOP     16
    fs_Buf      *buf, *bufs[BUF_WB_LOOP];
    int32       numDone, numInBufs, i;
    int32       candoDealloc = 0;
#ifdef FS_DEBUG_WRITEBACK_TIMING
    eeeRTC_t    startTime;
    eeeRTC_t    delta;
    eeeRTC_t    fs_printRuntimeThreshold;
    static boolean first = TRUE;
    
    if (first) {
        first = FALSE;
        // Do some initialization.
        eee_msToRTC(2000, &fs_printRuntimeThreshold);
        fs_userBufMaxRuntime = *(eee_rtcZero());
        fs_metaBufMaxRuntime = *(eee_rtcZero());
    }
    eee_rtcStart(&startTime);
#endif

    BUF_TRACE(("%s writeback awake %d dirtyBufs %d writesPending",
             bufList->type == FS_BUFLIST_META ? "meta" : "user",
             bufList->count.dirtyBufs[FS_BUF_CACHE_ITYPE_TOTAL],
             fs_ioWritesPending));

    if (fs_ioWritesPending >=
            fs_incoreTune->writesPendingMax ||
        context->fileSystem->ioPendingCount >
            fs_incoreTune->perFsIoPendingMax ||
        fs_ioPending >
            FS_RDWR_MAX)
    {
        if (DLL_IS_EMPTY(dllHdr))
        {
            // Nothing left on the list.
            return FALSE;
        }
        else
        {
            // Indicate need to come back for more.
            return TRUE;
        }
    }

    // The loop grabs a handful of bufs.  We then
    // let go of the spinlock to schedule the I/O
    // requests, and come back to the loop for more
    // if necessary.
    numDone = 0;
    while (TRUE)
    {
        numInBufs = 0;
        
        NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
        while (!DLL_IS_EMPTY(dllHdr))
        {
            DLL_REMOVE_TAIL(dllHdr, buf, mrdEntry);
            NFX_SPIN_LOCK(&buf->spin);
            buf->flags &= ~(FS_BUF_MRD|FS_BUF_WAITING_MRD);
            buf->flags |= FS_BUF_WB_STARTING;
#ifdef FS_DEBUG_BUF_HISTORY
            NFX_SPIN_LOCK(&buf->spinHistory);
            buf->history[buf->indexHistory] = FS_BUF_HIST_SET_WB_STARTING_WB_LIST;
            buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
            buf->flagsHistory[buf->indexHistory] = buf->flags;
            buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
            buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
            buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
            NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
            NFX_SPIN_UNLOCK(&buf->spin);

            if (fs_trylockEnqueue(context, context, &buf->lock,
                                  FS_LOCK_MODE_SHARE, FS_WB_EVENT_BUF_LOCK,
                                  buf, TRUE))
            {
                BUF_TRACE(("vol %llx inbufs buf %p blk %lld",
                           buf->fileSystem->volId.whole_thing,
                           buf, fs_blockNumber(buf->block)));

                FS_ASSERT(buf->flags & FS_BUF_DIRTY);
                FS_ASSERT(!DLL_IS_ISOLATED(buf, blkNumEntry));

                bufs[numInBufs] = buf;
                numInBufs += 1;
            }
            else
            {
#ifdef FS_DEBUG_BUF_HISTORY
                NFX_SPIN_LOCK(&buf->spinHistory);
                buf->history[buf->indexHistory] = FS_BUF_HIST_BUFLOCK_EVENT;
                buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
                buf->flagsHistory[buf->indexHistory] = buf->flags;
                buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
                buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
                buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
                NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
                FS_ASSERT(fs_lockIsLive(context, &buf->lock));
                BUF_TRACE(
                ("vol %llx defer buf %p blk %lld cnt %d md %d fg %d own %lld",
                     buf->fileSystem->volId.whole_thing,
                     buf,
                     fs_blockNumber(buf->block),
                     buf->lock.count, buf->lock.mode,
                     buf->lock.flags, (uint64)((uintptr_t)buf->lock.owner)));
            }
    
            numDone += 1;

            if (numInBufs >= BUF_WB_LOOP)
            {
                break;
            }

            if (fs_ioWritesPending + numInBufs >=
                    fs_incoreTune->writesPendingMax ||
                context->fileSystem->ioPendingCount + numInBufs >=
                    fs_incoreTune->perFsIoPendingMax ||
                fs_ioPending + numInBufs >= 
                    FS_RDWR_MAX)
            {
                break;
            }
        }
        NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);

        for (i = 0; i < numInBufs; i++)
        {
            // Write them out.
            candoDealloc += fs_bufDoWriteBack(context, bufList->wbHandle,
                                              bufs[i]);
        }

#ifdef FS_DEBUG_CANDO_STATS
        if (candoDealloc > fs_bufWritebackCandoDeallocMax)
            fs_bufWritebackCandoDeallocMax = candoDealloc;
#endif

        if (fs_ioWritesPending >= fs_incoreTune->writesPendingMax ||
            context->fileSystem->ioPendingCount >
                fs_incoreTune->perFsIoPendingMax ||
            fs_ioPending >= FS_RDWR_MAX ||
            numDone >= fs_incoreTune->bufWritesPerEvent ||
            
            DLL_IS_EMPTY(dllHdr))
        {
            break;
        }
    }

    if (bufList->type == FS_BUFLIST_USER)
    {
        if (numDone > fs_userBufWbMaxPerLoop)
        {
            fs_userBufWbMaxPerLoop = numDone;
        }
    }
    else
    {
        if (numDone > fs_metaBufWbMaxPerLoop)
        {
            fs_metaBufWbMaxPerLoop = numDone;
        }
    }
    

#ifdef FS_DEBUG_WRITEBACK_TIMING
    eee_rtcDuration(&startTime, &delta);
    if (bufList->type == FS_BUFLIST_USER)
    {
        if (eee_rtcCompare(&delta, &fs_userBufMaxRuntime) > 0)
        {
            fs_userBufMaxRuntime = delta;
        }
    }
    else
    {
        if (eee_rtcCompare(&delta, &fs_metaBufMaxRuntime) > 0)
        {
            fs_metaBufMaxRuntime = delta;
        }
    }
    
    if (eee_rtcCompare(&delta, &fs_printRuntimeThreshold) > 0)
    {
        eeeRTC_t endTime;
        uint32   deltaMS;

        deltaMS = (uint32)eee_rtcToMS(&delta)
        if (numDone > 0)
            printf("fs_bufWriteBackList %s : numdone %d runtime %d avg %d\n", 
                   bufList->type == FS_BUFLIST_META ? "meta" : "user",
                   numDone, deltaMS, deltaMS/numDone);
        else
            printf("fs_bufWriteBackList %s :  numdone %d runtime %d\n", 
                   bufList->type == FS_BUFLIST_META ? "meta" : "user",
                   numDone, deltaMS);
        eee_rtcAdd(&startTime, &delta, &endTime);
        printf("%s\n", eee_humandate((uint32)eee_rtcToSec(&endTime)));
    }
#endif

    if (DLL_IS_EMPTY(dllHdr))
    {
        // Nothing left on the list.
        return FALSE;
    }
    else
    {
        // Indicate need to come back for more.
        return TRUE;
    }

} /* fs_bufWriteBackList */


/**
 * Perform allocation on the specified buf.
 */
void
fs_bufAllocBuf (
    fs_Context      *context,
    fs_Inode        *inode,
    fs_Buf          *buf,
    boolean         commitInode)
{
    FS_ASSERT(fs_lockOwnedExclusive(context, &buf->lock));

    fs_bufAllocDoBufs(context, inode, 1, &buf, commitInode);
}


/**
 * Destroy the specified buf, called only during volume
 * exception. 
 */
static void
fs_bufDestroyBuf(
    fs_Context  *context,
    fs_BufList  *bufList,
    fs_Buf      *buf)
{
    int32   itype = fs_bufGetItype(buf);

#if defined(FS_DEBUG_BUF_CKSUM)
    if (fs_blockIsValid(buf->block))
    {
        fs_bufReleaseCksum(context, buf);
    }
#endif // FS_DEBUG_BUF_CKSUM

    ASSERT_LOCK_OWNED(&fs_bufListGlobal[bufList->type].spin);

    NFX_SPIN_LOCK(&buf->spin);

#ifdef FS_DEBUG_BUF_HISTORY
    NFX_SPIN_LOCK(&buf->spinHistory);
    buf->history[buf->indexHistory] = FS_BUF_HIST_DESTROY;
    buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
    buf->flagsHistory[buf->indexHistory] = buf->flags;
    buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
    buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
    buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
    NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif

    if (buf->flags & FS_BUF_DIRTY)
    {
        // This would normally happen in the ioDone,
        // which will never happen for this buf, so
        // correct the count right here.
        bufList->count.dirtyBufs[itype] -= 1;
        fs_bufListGlobal[bufList->type].count.dirtyBufs[itype] -= 1;

        bufList->count.dirtyBufs[FS_BUF_CACHE_ITYPE_TOTAL] -= 1;
        fs_bufListGlobal[bufList->type].count.dirtyBufs[FS_BUF_CACHE_ITYPE_TOTAL] -= 1;

        if (!DLL_IS_ISOLATED(buf, mrdEntry))
        {
            // Get the buf off of the dirty list.
            if (buf->flags & FS_BUF_MRD)
            {
                DLL_REMOVE(&bufList->mrdHdr, buf, mrdEntry);
                buf->flags &= ~FS_BUF_MRD;
            }
            else if (buf->flags & FS_BUF_WAITING_MRD)
            {
                DLL_REMOVE(&bufList->mrdWaitingHdr, buf, mrdEntry);
                buf->flags &= ~FS_BUF_WAITING_MRD;
            }
            else
            {
                DLL_REMOVE(&bufList->logSegmentHdr[buf->headLogSeg],
                           buf, mrdEntry);
                buf->headLogSeg = -1;
            }
        }
        buf->flags &= ~(FS_BUF_DIRTY | FS_BUF_IO_BUSY);

#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_LOCK(&buf->spinHistory);
        buf->history[buf->indexHistory] = FS_BUF_HIST_RESET_DIRTY_IO_BUSY;
        buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
        buf->flagsHistory[buf->indexHistory] = buf->flags;
        buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
        buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
        buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
        NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
    }

    fs_bufDelist(context, bufList, buf, TRUE);

    NFX_SPIN_UNLOCK(&buf->spin);

    fs_bufDeallocCoreRes(context, buf);

    fs_ramDealloc(FS_RAM_BUF_PTR, buf);
#ifdef FS_DEBUG_BUF_HISTORY
    NFX_SPIN_LOCK(&spinalloc);
    fs_buf_allocate_ptr[fs_buf_allocate_index] = buf;
    fs_buf_allocate_oper[fs_buf_allocate_index] = FS_DEBUG_BUF_HISTORY_DEALLOC;
    fs_buf_allocate_index = (fs_buf_allocate_index + 1) % FS_DEBUG_BUF_HISTORY_ALLOC_MAX;
    NFX_SPIN_UNLOCK(&spinalloc);
#endif
}


/**
 * Destroy the contents of the buffer cache for the
 * specified volume without trying to sync to disk,
 * and without regard to buf locks.
 *
 * This is called when a volume is being forcibly
 * unmounted.
 */
void
fs_bufCacheDestroyVolume(
    fs_Context      *context,
    fs_FileSystem   *fileSystem)
{
    fs_BufList  *bufList;
    fs_Buf      *buf, *bufNext;
    int32       hash, listType, itype;
    int32       totalUsed, destroyed;

    for (listType = 0; listType < FS_BUFLIST_MAX; listType++)
    {
        bufList = &context->fileSystem->bufList[listType];
        NFX_SPIN_LOCK(&fs_bufListGlobal[listType].spin);

        totalUsed = bufList->count.usedBufs[FS_BUF_CACHE_ITYPE_TOTAL];
        destroyed = 0;

        for (itype = 0; itype < FS_BUF_CACHE_ITYPE_MAX; itype++)
        {
            DLL_ITERATE_NEXT(buf, bufNext, &fs_bufListGlobal[listType].mruHdr[itype],
                             mruEntry)
            {
                if (buf->fileSystem == fileSystem)
                {
                    destroyed += 1;
                    //printf("bufCacheDest: freeing buf %p\n", buf);
    
                    fs_bufDestroyBuf(context, bufList, buf);
                    FS_ASSERT(bufList->count.usedBufs[FS_BUF_CACHE_ITYPE_TOTAL] ==
                              totalUsed - destroyed);
                }
            }
        }
    
        for (hash = 0; hash < FS_BUFLIST_HASH_MAX; hash++)
        {
            DLL_ITERATE_NEXT(buf, bufNext,
                             &fs_bufListGlobal[listType].bufsBlk[hash],
                             blkNumEntry)
            {
                if (buf->fileSystem == fileSystem)
                {
                    destroyed += 1;
                    //printf("bufCacheDest: freeing buf %p\n", buf);
    
                    fs_bufDestroyBuf(context, bufList, buf);
                    FS_ASSERT(bufList->count.usedBufs[FS_BUF_CACHE_ITYPE_TOTAL] ==
                              totalUsed - destroyed);
                }
            }
    
            DLL_ITERATE_NEXT(buf, bufNext,
                             &fs_bufListGlobal[listType].bufsPage[hash],
                             pageHashEntry)
            {
                if (buf->fileSystem == fileSystem)
                {
                    destroyed += 1;
                    //printf("bufCacheDest: freeing buf %p\n", buf);
    
                    fs_bufDestroyBuf(context, bufList, buf);
                    FS_ASSERT(bufList->count.usedBufs[FS_BUF_CACHE_ITYPE_TOTAL] ==
                              totalUsed - destroyed);
                }
            }
        }

        NFX_SPIN_UNLOCK(&fs_bufListGlobal[listType].spin);
    }
}


/**
 * Increment the log buffer reference
 * count on the specified buf.
 */
void
fs_bufLogRefIncr(
    fs_Context  *context,
    fs_Buf      *buf)
{
    FS_ASSERT(fs_lockHeldShare(context, &buf->lock) ||
              fs_lockHeldExclusive(context, &buf->lock));
    FS_ASSERT(!(buf->flags & FS_BUF_WAIT_LOGBUF));
    FS_ASSERT(buf->flags & FS_BUF_TYPE_META);

    NFX_SPIN_LOCK(&buf->spin);
    buf->logRefCount += 1;
    //printf("buf %p logRefIncr %d\n", buf, buf->logRefCount);
    NFX_SPIN_UNLOCK(&buf->spin);

} /* fs_bufLogRefIncr */


int32 fs_deferDecrDealloc = 0;

/**
 * Decrement the log buffer reference
 * count on the specified buf.
 */
void
fs_bufLogRefDecr(
    fs_Context  *context,
    fs_Buf      *buf)
{
    fs_BufList  *bufList = &context->fileSystem->bufList[FS_BUFLIST_META];
    boolean     doWriteBackEvent = FALSE;
    boolean     doFree = FALSE;

    // We guarantee that the buf pointer is valid
    // since it is dirty and we won't free a dirty
    // buf until the count goes to 0.
    // 
    // TODO - some way to assert this?
    
    FS_ASSERT(buf->flags & FS_BUF_TYPE_META);

    NFX_SPIN_LOCK(&buf->spin);
    FS_ASSERT(!(buf->flags & FS_BUF_IO_BUSY));
    FS_ASSERT(buf->logRefCount > 0);
    buf->logRefCount -= 1;
    //printf("buf %p logRefDecr %d\n", buf, buf->logRefCount);
    if (buf->logRefCount == 0)
    {
        if (buf->flags & FS_BUF_WAIT_LOGBUF)
        {
            FS_ASSERT(fs_lockOwnedShared(bufList->wbContext, &buf->lock, TRUE));
            FS_ASSERT(!(buf->flags & FS_BUF_FREE_LOGBUF));
            doWriteBackEvent = TRUE;
            //printf("LOGBUF_EVENT %p 0x%x\n", buf, buf->flags);
        }
        else if (buf->flags & FS_BUF_FREE_LOGBUF)
        {
            FS_ASSERT(!(buf->flags & FS_BUF_WAIT_LOGBUF));
            buf->flags &= ~FS_BUF_FREE_LOGBUF;
            doFree = TRUE;
            //printf("LOGBUF_FREE clr %p 0x%x\n", buf, buf->flags);
        }
    }
    NFX_SPIN_UNLOCK(&buf->spin);

    if (doWriteBackEvent)
    {
        esm_event_t     event;
        int32           rc;

        event.event = FS_WB_EVENT_LOG_BUF;
        event.p1.u.p = buf;
        rc = esm_deliverEvent(bufList->wbHandle, &event);
        if (rc != NFX_OK)
        {
            fs_error(context, FS_EVENT_DELIVERY_FAIL,
                     "wb logbuf signal");
        }
#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_LOCK(&buf->spinHistory);
        buf->history[buf->indexHistory] = FS_BUF_HIST_BUFLOG_EVENT;
        buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
        buf->flagsHistory[buf->indexHistory] = buf->flags;
        buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
        buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
        buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
        NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
    }
    else if (doFree)
    {
        NFX_SPIN_LOCK(&fs_bufListGlobal[FS_BUFLIST_META].spin);
        fs_bufMaybeAddToFreeList(context, bufList, buf);
        NFX_SPIN_UNLOCK(&fs_bufListGlobal[FS_BUFLIST_META].spin);
    }

} /* fs_bufLogRefDecr */


/**
 * Set the buf onto the dirty list for the specified
 * log segment.
 */
void
fs_bufSetDirtySegment(
    fs_Context  *context,
    void        *obj,
    int32       segNum,
    int32       offset)
{
    fs_BufList      *bufList;
    nfx_spinlock_t  *spinlock;
    fs_Buf          *buf = obj;

    FS_ASSERT(segNum < FS_LOG_MAX_SEGMENTS);
    FS_ASSERT(buf->headLogSeg == -1 ||
              fs_logSegIsOlder(context->fileSystem->logRoot, segNum, buf->headLogSeg));

    // During COW handling, the log dependency is created
    // before the call to bufWrite is made, so the buf isn't
    // marked dirty yet.  In this case, we don't want to put
    // it on the dirty list, we'll do it when the bufWrite
    // happens, so just remember the segNum for now.
    if (!(buf->flags & FS_BUF_DIRTY))
    {
        //printf("1: offset %d segNum %d\n", offset, segNum);
        NFX_SPIN_LOCK(&buf->spin);
        buf->headLogSeg = segNum;
        NFX_SPIN_UNLOCK(&buf->spin);
        return;
    }

    bufList = fs_bufListForBuf(buf);
    spinlock = &fs_bufListGlobal[bufList->type].spin;

    NFX_SPIN_LOCK(spinlock);
    NFX_SPIN_LOCK(&buf->spin);
    
    if (DLL_IS_ISOLATED(buf, mrdEntry))
    {
        FS_ASSERT((buf->flags & FS_BUF_DIRTY) ||
                  (buf->flags & FS_BUF_FREE_DEFER));
        FS_ASSERT((buf->flags & FS_BUF_WB_STARTING) ||
                  (buf->flags & FS_BUF_FREE_DEFER));

        //printf("2: offset %d segNum %d\n", offset, segNum);
        // Writeback already in progress on this
        // buf, don't put it back on the list.
        NFX_SPIN_UNLOCK(&buf->spin);
        NFX_SPIN_UNLOCK(spinlock);
        return;
    }

    if (buf->flags & FS_BUF_MRD)
    {
        buf->flags &= ~FS_BUF_MRD;
        DLL_REMOVE(&bufList->mrdHdr, buf, mrdEntry);
#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_LOCK(&buf->spinHistory);
        buf->history[buf->indexHistory] = FS_BUF_HIST_REMOVE_MRD_DIRTY;
        buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
        buf->flagsHistory[buf->indexHistory] = buf->flags;
        buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
        buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
        buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
        NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
    }
    else if (buf->flags & FS_BUF_WAITING_MRD)
    {
        buf->flags &= ~FS_BUF_WAITING_MRD;
        DLL_REMOVE(&bufList->mrdWaitingHdr, buf, mrdEntry);
#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_LOCK(&buf->spinHistory);
        buf->history[buf->indexHistory] = FS_BUF_HIST_REMOVE_W_MRD_DIRTY;
        buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
        buf->flagsHistory[buf->indexHistory] = buf->flags;
        buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
        buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
        buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
        NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
    }
    else
    {
        FS_ASSERT(buf->headLogSeg < FS_LOG_MAX_SEGMENTS);

        if (!DLL_IS_ISOLATED(buf, mrdEntry))
        {
            DLL_REMOVE(&bufList->logSegmentHdr[buf->headLogSeg],
                       buf, mrdEntry);
#ifdef FS_DEBUG_BUF_HISTORY
            NFX_SPIN_LOCK(&buf->spinHistory);
            buf->history[buf->indexHistory] = FS_BUF_HIST_REMOVE_L_MRD_DIRTY;
            buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
            buf->flagsHistory[buf->indexHistory] = buf->flags;
            buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
            buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
            buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
            NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
        }
    }

    //printf("3: offset %d segNum %d\n", offset, segNum);
    DLL_INSERT_HEAD(&bufList->logSegmentHdr[segNum], buf, mrdEntry);
    buf->headLogSeg = segNum;
#ifdef FS_DEBUG_BUF_HISTORY
    NFX_SPIN_LOCK(&buf->spinHistory);
    buf->history[buf->indexHistory] = FS_BUF_HIST_INSERT_L_MRD_DIRTY;
    buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
    buf->flagsHistory[buf->indexHistory] = buf->flags;
    buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
    buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
    buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
    NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
    NFX_SPIN_UNLOCK(&buf->spin);
    NFX_SPIN_UNLOCK(spinlock);
}


/**
 * Writeback daemon state machine create action.
 */
int32
fs_wb_create(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *smCookie)
{
    // do nothing here.
    return nextState;

} /* fs_wb_create */


#if !defined(FS_STANDALONE)
/**
 * writeback daemon timer callback 
 * param1 points to the fs_BufList
 */
static int32
fs_writebackDaemonTimeout(
    uint32  timer,
    void    *param1,
    void    *param2,
    void    *param3,
    void    *param4)
{
    fs_BufList      *bufList = param1;

    FS_ASSERT(bufList != NULL);

    if (timer == bufList->tmpTimerHandle)
    {
        FS_ASSERT(bufList->tmpTimerExists);
        eee_timerStop(timer);
        eee_timerDestroy(timer);
        bufList->tmpTimerExists = FALSE;
    }

    fs_bufSignalWriteback(bufList->wbContext, bufList->type, (address_t) param2);
    return NFX_OK;
}
#endif


/**
 * Writeback daemon wants to schedule itself for
 * more cpu time for the specified filesystem/buflist.
 */
static void
fs_wbReschedule(
    fs_Context      *context,
    fs_FileSystem   *fileSystem,
    fs_BufList      *bufList)
{
#if !defined(FS_STANDALONE)
    if (fileSystem->ioPendingCount >
        fs_incoreTune->perFsIoPendingMax)
    {
        // Wait a few msecs, the volume is probably
        // a bit busy at the moment.

        if (!bufList->tmpTimerExists)
        {
            bufList->tmpTimerHandle = eee_timerCreate(
                        fs_incoreTune->wbSleepMsecs, EEE_TIMER_ONCE);
            if (bufList->tmpTimerHandle == TIMER_ERR_NO_TIMER_AVAILABLE)
            {
                fs_error(NULL, FS_RAM_ALLOC_ERR, "create tmp wb timer");
            }
            else
            {
                bufList->tmpTimerExists = TRUE;
                if (eee_timerStart(bufList->tmpTimerHandle,
                                   fs_writebackDaemonTimeout,
                                   (void *) bufList, (void *) FS_WB_WAKEUP_RESCHED,
                                   NULL, NULL) != NFX_OK)
                {
                    fs_error(NULL, FS_INTERNAL_ERR, "start wb timer");
                }
            }
        }

        NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
        bufList->wbSignal = 0;
        NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);
    }
    else
#endif
    {
        esm_event_t event;
        int32       rc;

        BUF_TRACE(("%s writeback rescheduling self immediately",
                bufList->type == FS_BUFLIST_META ? "meta" : "user"));

        event.event = FS_WB_EVENT_WAKEUP;
        event.p1.u.p = bufList;
        event.p2.u.w = FS_WB_WAKEUP_RESCHED;
        rc = esm_deliverEvent(bufList->wbHandle, &event);
        if (rc != NFX_OK)
        {
            fs_error(context, FS_EVENT_DELIVERY_FAIL, "wb signal");
        }
    }
}


/**
 * Return TRUE if a dirty buf is potentially
 * holding up log truncation of the current head
 * log segment.
 */
static boolean
fs_bufBlockingLogTruncation(
    fs_Context      *context,
    fs_FileSystem   *fileSystem,
    fs_BufList      *bufList)
{
    int32       headSeg = fs_logGetHeadSegnum(context, fileSystem);

    if (!DLL_IS_EMPTY(&bufList->logSegmentHdr[headSeg]))
    {
        return TRUE;
    }

    return FALSE;
}


/**
 * Return TRUE if the global buffer cache dirty limits
 * are exceeded.
 */
static inline boolean
fs_bufCacheDirtyOverThreshold(
    int32 type)
{
   return (fs_bufListGlobal[type].count.dirtyBufs[FS_BUF_CACHE_ITYPE_NONDIR] >
            fs_bufListGlobal[type].count.dirtyBufLowThresh[FS_BUF_CACHE_ITYPE_NONDIR] ||

           fs_bufListGlobal[type].count.dirtyBufs[FS_BUF_CACHE_ITYPE_DIR] >
            fs_bufListGlobal[type].count.dirtyBufLowThresh[FS_BUF_CACHE_ITYPE_DIR] ||

           fs_bufListGlobal[type].count.dirtyBufs[FS_BUF_CACHE_ITYPE_COMP_NONDIR] >
            fs_bufListGlobal[type].count.dirtyBufLowThresh[FS_BUF_CACHE_ITYPE_COMP_NONDIR] ||

           fs_bufListGlobal[type].count.dirtyBufs[FS_BUF_CACHE_ITYPE_COMP_DIR] >
           fs_bufListGlobal[type].count.dirtyBufLowThresh[FS_BUF_CACHE_ITYPE_COMP_DIR]);
}


/**
 * Writeback daemon state machine wakeup action.
 */
int32
fs_wb_wakeup(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *smCookie)
{
    fs_Context      *context = smCookie;
    fs_FileSystem   *fileSystem = context->fileSystem;
    fs_BufList      *bufList = event.p1.u.p;
    int32           eventType = event.p2.u.w;
    int32           i, segNum;

    // don't do anyting if already in exception handling
    // since locks are abadoned by terminating threads
    // and fs_bufCacheDestroyVolume will do the mop up
    if (fileSystem->state == FS_VOL_STATE_FAILING)
        return nextState;

    FS_ASSERT(event.p2.u.w < FS_WB_WAKEUP_MAX);

    segNum = fs_logGetHeadSegnum(context, fileSystem);

    NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
    fs_bufWbWakeupCount[bufList->type][eventType] += 1;
    NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);

    //printf("%s wb_wakeup logHeadSeg %d blocking %s aboveThresh %s\n",
    //fileSystem->super->mountName, segNum,
    //fs_bufBlockingLogTruncation(context, fileSystem, bufList) ? "TRUE" : "FALSE",
    //fs_logAboveWritebackThreshold(context, fileSystem) ? "TRUE" : "FALSE");

    if (fs_bufWriteBackList(context, bufList, &bufList->mrdWaitingHdr))
    {
        // need to come back for more urgent IO
        // Schedule myself for some more.
        fs_wbReschedule(context, fileSystem, bufList);
        return nextState;
    }

    if (eventType == FS_WB_WAKEUP_PERIODIC)
    {
        // Always do a little something on the periodic
        // wakeup.
        fs_bufWriteBackList(context, bufList, &bufList->logSegmentHdr[segNum]);
        fs_bufWriteBackList(context, bufList, &bufList->mrdHdr);
    }

    for (i = 0; i < FS_LOG_MAX_SEGMENTS; i++)
    {
        if (bufList->numWaiters == 0 &&

            (!fs_bufBlockingLogTruncation(context, fileSystem, bufList) ||
             !fs_logAboveWritebackThreshold(context, fileSystem)) &&

            !fs_bufCacheDirtyOverThreshold(bufList->type))
        {
            // Nothing more to do, or no urgency to do it
            // right away.  Wait for the next event.
    
            NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
            bufList->wbSignal = 0;
            NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);
    
            BUF_TRACE(("%s writeback going back to sleep",
                    bufList->type == FS_BUFLIST_META ? "meta" : "user"));

            return nextState;
        }

        while (fs_bufWriteBackList(context, bufList,
                                   &bufList->logSegmentHdr[segNum]) != 0)
        {
            // Schedule myself for some more.
            fs_wbReschedule(context, fileSystem, bufList);
            return nextState;
        }

        if (++segNum >= FS_LOG_MAX_SEGMENTS)
        {
            segNum = 0;
        }
    }

    if (bufList->numWaiters != 0 ||
        fs_bufCacheDirtyOverThreshold(bufList->type))
    {
        // We've scheduled everything on the waiting list
        // and the logged lists, that only leaves dirty
        // bufs on the mrd list (should only be those that
        // are never logged or are pending write to the log).
        while (fs_bufWriteBackList(context, bufList, &bufList->mrdHdr) != 0)
        {
            // Schedule myself for some more.
            fs_wbReschedule(context, fileSystem, bufList);
            return nextState;
        }
    }

    NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
    bufList->wbSignal = 0;
    NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);
   
    BUF_TRACE(("%s writeback going back to sleep",
            bufList->type == FS_BUFLIST_META ? "meta" : "user"));

    return nextState;

} /* fs_wb_wakeup */


/**
 * Perform housekeeping for the buf that just completed
 * writeback: Update the dirty counts, buf flags, notify waiters.
 */
static void
fs_bufIoDone(
    fs_Context  *context,
    fs_Buf      *buf)
{
    fs_BufList  *bufList;

    FS_ASSERT(fs_lockOwnedShared(context, &buf->lock, TRUE));
    FS_ASSERT(buf->logSegDep.count == 0);
    FS_ASSERT(buf->flags & FS_BUF_WB_STARTING);

    bufList = fs_bufListForBuf(buf);

    if (buf->flags & FS_BUF_IS_CLONE)
    {
        fs_Buf *origBuf = buf->clone;

        FS_ASSERT(fs_lockOwnedShared(context, &origBuf->lock, TRUE));
        FS_ASSERT(origBuf->flags & FS_BUF_WB_STARTING);

        fs_unlockShared(context, &buf->lock);

        // Free the clone's data and itself.
        fs_ramDealloc(FS_RAM_BUF_CLONE_DATA, buf->u.data);
        fs_ramDealloc(FS_RAM_BUF_CLONE, buf);
#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_LOCK(&spinalloc);
        fs_buf_allocate_ptr[fs_buf_allocate_index] = buf;
        fs_buf_allocate_oper[fs_buf_allocate_index] = FS_DEBUG_BUF_HISTORY_DEALLOC_CLONE;
        fs_buf_allocate_index = (fs_buf_allocate_index + 1) % FS_DEBUG_BUF_HISTORY_ALLOC_MAX;
        NFX_SPIN_UNLOCK(&spinalloc);
#endif
        buf = NULL;

        NFX_SPIN_LOCK(&origBuf->spin);
        origBuf->clone = NULL;

        if (origBuf->flags & FS_BUF_WAIT_CLONE)
        {
#ifdef FS_DEBUG_BUF_HISTORY
            NFX_SPIN_LOCK(&origBuf->spinHistory);
            origBuf->history[origBuf->indexHistory] = FS_BUF_HIST_CLONE_ORIG_WB;
            origBuf->cpuHistory[origBuf->indexHistory] = nfxMyCPUId;
            origBuf->indexHistory = (origBuf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
            NFX_SPIN_UNLOCK(&origBuf->spinHistory);
#endif
            origBuf->flags &= ~FS_BUF_WAIT_CLONE;
            NFX_SPIN_UNLOCK(&origBuf->spin);

            // Start IO on the origBuf again.
            fs_bufDoWriteBack(context, bufList->wbHandle, origBuf);
        }
        else
        {
#ifdef FS_DEBUG_BUF_HISTORY
            NFX_SPIN_LOCK(&origBuf->spinHistory);
            origBuf->history[origBuf->indexHistory] = FS_BUF_HIST_CLONE_ORIG_REL;
            origBuf->cpuHistory[origBuf->indexHistory] = nfxMyCPUId;
            origBuf->indexHistory = (origBuf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
            NFX_SPIN_UNLOCK(&origBuf->spinHistory);
#endif
            origBuf->flags &= ~FS_BUF_WB_STARTING;
            NFX_SPIN_UNLOCK(&origBuf->spin);

            // Release our shared lock on the original
            // buf.
            fs_bufRelease(context, origBuf);
        }
    }
    else
    {
        NFX_SPIN_LOCK(&fs_bufListGlobal[bufList->type].spin);
        NFX_SPIN_LOCK(&buf->spin);
        FS_ASSERT(DLL_IS_ISOLATED(buf, mrdEntry));
        buf->flags &= ~(FS_BUF_DIRTY | FS_BUF_IO_BUSY | FS_BUF_WB_STARTING);
        fs_bufDecrDirtyCount(context, bufList, buf);
#ifdef FS_DEBUG_BUF_HISTORY
        NFX_SPIN_LOCK(&buf->spinHistory);
        buf->history[buf->indexHistory] = FS_BUF_HIST_IODONE_CLEAR;
        buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
        buf->flagsHistory[buf->indexHistory] = buf->flags;
        buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
        buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
        buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
        NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
        NFX_SPIN_UNLOCK(&buf->spin);
        NFX_SPIN_UNLOCK(&fs_bufListGlobal[bufList->type].spin);

        if (buf->zeroFillLogId)
        {
            // Buf has a zero-fill transaction still
            // pending in the log, that can now be
            // completed.
            fs_logComplete(context, buf->zeroFillLogId, buf->inum);
            buf->zeroFillLogId = 0;
        }

        fs_bufRelease(context, buf);
    }

} /* fs_bufIoDone*/


/**
 * Writeback daemon state machine ioDone action.
 */
int32
fs_wb_ioDone(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *smCookie)
{
    fs_Context      *context;
    fs_RdwrReq      *rdwr;
    fs_Buf          *buf;

    context = smCookie;
    rdwr = event.p1.u.p;

    FS_ASSERT(rdwr >= fs_rdwrStart &&
              rdwr < fs_rdwrEnd);

    buf = (fs_Buf *) rdwr->private;

#ifdef FS_DEBUG_BUF_HISTORY
    NFX_SPIN_LOCK(&buf->spinHistory);
    buf->history[buf->indexHistory] = FS_BUF_HIST_IODONE;
    buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
    buf->flagsHistory[buf->indexHistory] = buf->flags;
    buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
    buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
    buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
    NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif

    NFX_SPIN_LOCK(&context->fileSystem->spin);
    context->fileSystem->ioPendingCount -= 1;
    NFX_SPIN_UNLOCK(&context->fileSystem->spin);

    if (context->fileSystem->state == FS_VOL_STATE_FAILING)
    {   
        // don't do anyting if already in exception handling
        // since locks are abadoned by terminating threads
        // and fs_bufCacheDestroyVolume will do the mop up
        fs_ioDoneError(rdwr);
        return nextState;
    }

    if (rdwr->evmStat.status != NFX_OK)
    {
        fs_rdwrDealloc(rdwr);
        printf("fs_wb_ioDone io error returned, aborting\n");
        fs_volumeException(context, context->fileSystem, FS_WRITE_ERR);
        return nextState;
    }

    FS_ASSERT(buf->flags & FS_BUF_DIRTY);
    FS_ASSERT(buf->flags & FS_BUF_IO_BUSY);
    FS_ASSERT(fs_lockOwnedShared(context, &buf->lock, TRUE));
    FS_ASSERT(rdwr->evmStat.status == NFX_OK);

    BUF_TRACE(("vol %llx io_wb_done on blk %lld rdwr %p", 
               buf->fileSystem->volId.whole_thing,
               fs_blockNumber(buf->block), rdwr));

    fs_rdwrDealloc(rdwr);

#if defined(FS_DEBUG_BUF_CKSUM)
    if ((buf->flags & FS_BUF_TYPE_META) &&
        !(buf->flags & FS_BUF_READ_VERIFIED))
    {
        fs_bufCompareCksum(context, buf->block, buf->u.data, buf,
                           "ioDone", TRUE);
    }
#endif // FS_DEBUG_BUF_CKSUM

#if defined (FS_DEBUG_BUF_READ_VERIFY)

    if (fs_bufDoReadVerify &&
        (buf->flags & FS_BUF_READ_VERIFY))
    {
        // Kick off a read request to verify the
        // just written data.
        NFX_SPIN_LOCK(&buf->spin);
        buf->flags &= ~FS_BUF_READ_VERIFY;
        NFX_SPIN_UNLOCK(&buf->spin);

        NFX_SPIN_LOCK(&fs_ioSpin);
        fs_ioTotalMetaReads += 1;
        NFX_SPIN_UNLOCK(&fs_ioSpin);

        NFX_SPIN_LOCK(&buf->fileSystem->spin);
        buf->fileSystem->ioTotalMetaReads += 1;
        NFX_SPIN_UNLOCK(&buf->fileSystem->spin);

        rdwr = fs_rdwrAlloc();
        fs_volReadAsyncBlock(context,
                             rdwr,
                             buf->block,
                             FS_WB_EVENT_READ_VERIFY,
                             buf);
        return nextState;
    }
    NFX_SPIN_LOCK(&buf->spin);
    buf->flags &= ~FS_BUF_READ_VERIFIED;
    NFX_SPIN_UNLOCK(&buf->spin);
#endif  // FS_DEBUG_BUF_READ_VERIFY

    BUF_TRACE(("vol %llx iodone remdep blk %lld", 
               buf->fileSystem->volId.whole_thing,
               fs_blockNumber(buf->block)));

    if (buf->logSegDep.count > 0)
    {
        fs_bufRemoveLogDeps(context, buf);
    }

    fs_bufIoDone(context, buf);

#ifdef FS_DEBUG_BUF_HISTORY
    NFX_SPIN_LOCK(&buf->spinHistory);
    buf->history[buf->indexHistory] = FS_BUF_HIST_IODONE_RETURN;
    buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
    buf->flagsHistory[buf->indexHistory] = buf->flags;
    buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
    buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
    buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
    NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif

    return nextState;

} /* fs_wb_ioDone */


/**
 * Writeback daemon state machine readVerify action.
 */
int32
fs_wb_readVerify(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *smCookie)
{
    fs_Context      *context;
    fs_RdwrReq      *rdwr;
    fs_Buf          *buf;

    context = smCookie;
    rdwr = event.p1.u.p;

    FS_ASSERT(rdwr >= fs_rdwrStart &&
              rdwr < fs_rdwrEnd);

    // don't do anyting if already in exception handling
    // since locks are abadoned by terminating threads
    // and fs_bufCacheDestroyVolume will do the mop up
    if (context->fileSystem->state == FS_VOL_STATE_FAILING)
    {
        NFX_SPIN_LOCK(&context->fileSystem->spin);
        context->fileSystem->ioPendingCount -= 1;
        NFX_SPIN_UNLOCK(&context->fileSystem->spin);

        fs_ioDoneError(rdwr);
        return nextState;
    }

    buf = (fs_Buf *) rdwr->private;

    if (memcmp(buf->u.data, rdwr->evmStat.rDataHandle.dataPtr,
               FS_BLOCK_SIZE) != 0)
    {
        printf("vol %s (0x%llx) fs read verify failed\n",
            buf->fileSystem->super->mountName,
            buf->fileSystem->volId.whole_thing);
        printf("  expected datap %p got datap %p blknum %lld\n",
            buf->u.data, rdwr->evmStat.rDataHandle.dataPtr,
            fs_blockNumber(buf->block));
        FS_ASSERT(FALSE);
    }

    fs_ramDealloc(FS_RAM_READ_DATA, rdwr->evmStat.rDataHandle.dataPtr);
    rdwr->evmStat.rDataHandle.dataPtr = NULL;

    NFX_SPIN_LOCK(&buf->spin);
    buf->flags |= FS_BUF_READ_VERIFIED;
    NFX_SPIN_UNLOCK(&buf->spin);

    event.event = FS_WB_EVENT_IO_DONE;
    return fs_wb_ioDone(handle, nextState, event, smCookie);

} /* fs_wb_readVerify */


/**
 * Writeback daemon state machine logBuf action.
 */
int32
fs_wb_logBuf(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *smCookie)
{
    fs_Context *context = smCookie;
    fs_Buf *buf = event.p1.u.p;

    // don't do anyting if already in exception handling
    // since locks are abadoned by terminating threads
    // and fs_bufCacheDestroyVolume will do the mop up
    if (context->fileSystem->state == FS_VOL_STATE_FAILING)
        return nextState;

#ifdef FS_DEBUG_BUF_HISTORY
    NFX_SPIN_LOCK(&buf->spinHistory);
    buf->history[buf->indexHistory] = FS_BUF_HIST_BUFLOG;
    buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
    buf->flagsHistory[buf->indexHistory] = buf->flags;
    buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
    buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
    buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
    NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif

    FS_ASSERT(fs_lockOwnedShared(context, &buf->lock, TRUE));
    FS_ASSERT(fs_blockIsValid(buf->block));
    FS_ASSERT((buf->flags & FS_BUF_DIRTY) != 0);
    FS_ASSERT(buf->flags & FS_BUF_WB_STARTING);
    FS_ASSERT((buf->flags & FS_BUF_WAIT_LOGBUF) != 0);
    FS_ASSERT((buf->flags & FS_BUF_IO_BUSY) == 0);
    FS_ASSERT(buf->logRefCount == 0);

    BUF_TRACE(("vol %llx wb_logbuf buf %p inum %lld page %lld blk %lld", 
               buf->fileSystem->volId.whole_thing,
               buf, buf->inum, buf->page,
               fs_blockNumber(buf->block)));

    NFX_SPIN_LOCK(&buf->spin);
    buf->flags &= ~FS_BUF_WAIT_LOGBUF;
    //printf("LOGBUF_WAIT clr %p 0x%x\n", buf, buf->flags);
    NFX_SPIN_UNLOCK(&buf->spin);

    fs_bufDoWriteBack(context, context->smHandle, buf);

    return nextState;

} /* fs_wb_logBuf */


/**
 * Writeback daemon state machine bufLock action.
 */
int32
fs_wb_bufLock(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *smCookie)
{
    fs_Context      *context;
    fs_Lock         *lockp;
    fs_Buf          *buf;

    context = smCookie;

    // don't do anyting if already in exception handling
    // since locks are abadoned by terminating threads
    // and fs_bufCacheDestroyVolume will do the mop up

    if (context->fileSystem->state == FS_VOL_STATE_FAILING)
        return nextState;

    lockp = event.p1.u.p;
    buf = event.p2.u.p;

    if (lockp == NULL ||                    // buf destroyed
        (buf->flags & FS_BUF_INVALID) ||    // buf contents no longer valid
        (buf->flags & FS_BUF_DIRTY) == 0)   // buf no longer dirty (reused)
    {
        // Never mind.
        BUF_TRACE(("vol %llx wb buflck buf %p blk %lld flags %x ignore",
                   buf->fileSystem->volId.whole_thing,
                   buf, fs_blockNumber(buf->block), buf->flags));

        if (lockp != NULL)
        {
            fs_bufRelease(context, buf);
        }
        return nextState;
    }

#ifdef FS_DEBUG_BUF_HISTORY
    NFX_SPIN_LOCK(&buf->spinHistory);
    buf->history[buf->indexHistory] = FS_BUF_HIST_BUFLOCK;
    buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
    buf->flagsHistory[buf->indexHistory] = buf->flags;
    buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
    buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
    buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
    NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif

    BUF_TRACE(("vol %llx wb buflck buf %p blk %lld flags %x",
               buf->fileSystem->volId.whole_thing,
               buf, fs_blockNumber(buf->block), buf->flags));

    FS_ASSERT(buf->flags & FS_BUF_DIRTY);
    FS_ASSERT(buf->flags & FS_BUF_WB_STARTING);
    FS_ASSERT((buf->flags & FS_BUF_IO_BUSY) == 0);
    FS_ASSERT(fs_lockOwnedShared(context, &buf->lock, TRUE));
    FS_ASSERT(DLL_IS_ISOLATED(buf, mrdEntry));

    fs_bufDoWriteBack(context, context->smHandle, buf);

    return nextState;

} /* fs_wb_bufLock */


/**
 * Writeback daemon state machine stop action.
 */
int32
fs_wb_stop(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *smCookie)
{
    fs_BufList      *bufList = (void *) event.p1.u.p;
        
#if !defined(FS_STANDALONE)
    if (bufList->tmpTimerExists)
    {
        eee_timerStop(bufList->tmpTimerHandle);
        eee_timerDestroy(bufList->tmpTimerHandle);
        bufList->tmpTimerExists = FALSE;
    }
    if (bufList->wbTimerExists)
    {
        eee_timerStop(bufList->wbTimerHandle);
        eee_timerDestroy(bufList->wbTimerHandle);
        bufList->wbTimerExists = FALSE;
    }
#endif

    fs_wb_sm_destroy(bufList->wbHandle);
    fs_lockContextFree(bufList->wbContext);
    fs_ramDealloc(FS_RAM_CORE_MISC, bufList->wbContext);
    bufList->wbContext = NULL;

    return nextState;

} /* fs_wb_stop */


/**
 * Writeback daemon state machine bad_state action.
 */
int32
fs_wb_bad_state(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *smCookie)
{
    FS_ASSERT(FALSE);

    return nextState;

} /* fs_wb_bad_state */


/**
 * Set the limits on the user cache.
 */
static void
fs_bufSetUserCacheLimit()
{
    uint64  memAvail, memUse;
    int32   usedLimit;

    // Calculate how many user bufs to store in the cache.
#if defined(FS_STANDALONE)
    memAvail = (200 * 1024 * 1024);
#else
    memAvail = sbm_get_sac_size();
#endif
    memUse = (memAvail / 100) * fs_incoreTune->bufUserPctMem;
    usedLimit = memUse / FS_BLOCK_SIZE;
    fs_bufListGlobal[FS_BUFLIST_USER].count.usedBufLimit[FS_BUF_CACHE_ITYPE_NONDIR] = usedLimit;
#ifdef FS_DYN_CACHE_SIZE
    fs_bufListGlobal[FS_BUFLIST_USER].count.usedBufLimitLWM[FS_BUF_CACHE_ITYPE_NONDIR] = usedLimit;
    // HWM not used for FS_BUFLIST_USER, just put to a high value MAX_INT
    fs_bufListGlobal[FS_BUFLIST_USER].count.usedBufLimitHWM[FS_BUF_CACHE_ITYPE_NONDIR] = (1 << 31);
#endif
    fs_bufListGlobal[FS_BUFLIST_USER].count.dirtyBufLimit[FS_BUF_CACHE_ITYPE_NONDIR] =
        (usedLimit / 100) * (100 - fs_incoreTune->bufPctAvailLoWtrMk);
    fs_bufListGlobal[FS_BUFLIST_USER].count.dirtyBufLowThresh[FS_BUF_CACHE_ITYPE_NONDIR] =
        (usedLimit / 100) * (100 - fs_incoreTune->bufPctAvailHiWtrMk);
}


/**
 * Set the limits on the meta cache.
 */
static void
fs_bufSetMetaCacheLimit()
{
    uint64  memAvail, memUse;
    int32   usedLimit, sizeEach;

    // Calculate how many full sized non-dir meta bufs to store in the cache.
#ifdef FS_STANDALONE
    usedLimit = 2000;
#else  // FS_STANDALONE
    usedLimit = ((eee_getTotalNbrOfBufs(EEE_BUF_ULTRA) - FS_BUF_META_RESERVED) / 100) *
                        fs_incoreTune->bufMetaNonDirPctMem;
#if defined(FS_DEBUG_SHADOW_BUF)
    usedLimit = usedLimit / 2;   // include shadow buf
#endif // FS_DEBUG_SHADOW_BUF
#endif // FS_STANDALONE
    fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimit[FS_BUF_CACHE_ITYPE_NONDIR] =
        usedLimit;
#ifdef FS_DYN_CACHE_SIZE
    fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimitLWM[FS_BUF_CACHE_ITYPE_NONDIR] =
        usedLimit;
    fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimitHWM[FS_BUF_CACHE_ITYPE_NONDIR] =
        (1 << 31);  // no limit yet
#endif // FS_DYN_CACHE_SIZE
    fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufLimit[FS_BUF_CACHE_ITYPE_NONDIR] =
        (usedLimit / 100) * (100 - fs_incoreTune->bufPctAvailLoWtrMk);
    fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufLowThresh[FS_BUF_CACHE_ITYPE_NONDIR] =
        (usedLimit / 100) * (100 - fs_incoreTune->bufPctAvailHiWtrMk);
    fs_bufListGlobal[FS_BUFLIST_META].count.minInUse[FS_BUF_CACHE_ITYPE_NONDIR] =
        fs_incoreTune->metaNonDirMinInUse;

    // Calculate how many full sized dir meta bufs to store in the cache.
#ifdef FS_STANDALONE
    usedLimit = 2000;
#else  // FS_STANDALONE
    usedLimit = ((eee_getTotalNbrOfBufs(EEE_BUF_ULTRA) - FS_BUF_META_RESERVED) / 100) *
                        fs_incoreTune->bufMetaDirPctMem;
#ifdef FS_DEBUG_SHADOW_BUF
    usedLimit = usedLimit / 2;   // include shadow buf
#endif // FS_DEBUG_SHADOW_BUF
#endif // FS_STANDALONE
    fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimit[FS_BUF_CACHE_ITYPE_DIR] =
        usedLimit;
#ifdef FS_DYN_CACHE_SIZE
    fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimitLWM[FS_BUF_CACHE_ITYPE_DIR] =
        usedLimit;
    fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimitHWM[FS_BUF_CACHE_ITYPE_DIR] =
        (1 << 31);  // no limit yet
#endif // FS_DYN_CACHE_SIZE
    fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufLimit[FS_BUF_CACHE_ITYPE_DIR] =
        (usedLimit / 100) * (100 - fs_incoreTune->bufPctAvailLoWtrMk);
    fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufLowThresh[FS_BUF_CACHE_ITYPE_DIR] =
        (usedLimit / 100) * (100 - fs_incoreTune->bufPctAvailHiWtrMk);
    fs_bufListGlobal[FS_BUFLIST_META].count.minInUse[FS_BUF_CACHE_ITYPE_DIR] =
        fs_incoreTune->metaDirMinInUse;

    // Calculate how many compact non-dir meta bufs to store in the cache.
#ifdef FS_STANDALONE
    memAvail = (20 * 1024 * 1024);
    sizeEach = FS_COMPACT_PAGE_SIZE;
#else // FS_STANDALONE
    memAvail = eee_ramTotalFlexSize();
    sizeEach = (((FS_COMPACT_PAGE_SIZE - 1) / MM_FLEX_MODULO + 1) * MM_FLEX_MODULO) + MM_FLEX_OVERHEAD ;
#endif // FS_STANDALONE
    memUse = (memAvail / 100) * fs_incoreTune->bufMetaCompNonDirPctMem;
    usedLimit = memUse / sizeEach;
#ifdef FS_DEBUG_SHADOW_BUF
    usedLimit = usedLimit / 2;   // include shadow buf
#endif // FS_DEBUG_SHADOW_BUF
    fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimit[FS_BUF_CACHE_ITYPE_COMP_NONDIR] =
        usedLimit;
#ifdef FS_DYN_CACHE_SIZE
    fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimitLWM[FS_BUF_CACHE_ITYPE_COMP_NONDIR] =
        usedLimit;
    fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimitHWM[FS_BUF_CACHE_ITYPE_COMP_NONDIR] =
        (1 << 31);  // no limit yet
#endif
    fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufLimit[FS_BUF_CACHE_ITYPE_COMP_NONDIR] =
        (usedLimit / 100) * (100 - fs_incoreTune->bufPctAvailLoWtrMk);
    fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufLowThresh[FS_BUF_CACHE_ITYPE_COMP_NONDIR] =
        (usedLimit / 100) * (100 - fs_incoreTune->bufPctAvailHiWtrMk);
    fs_bufListGlobal[FS_BUFLIST_META].count.minInUse[FS_BUF_CACHE_ITYPE_COMP_NONDIR] =
        fs_incoreTune->metaCompNonDirMinInUse;

    // Calculate how many compact dir meta bufs to store in the cache.
#if defined(FS_STANDALONE)
    memAvail = (20 * 1024 * 1024);
    sizeEach = FS_COMPACT_PAGE_SIZE;
#else
    memAvail = eee_ramTotalFlexSize();
    sizeEach = (((FS_COMPACT_PAGE_SIZE - 1) / MM_FLEX_MODULO + 1) * MM_FLEX_MODULO) + MM_FLEX_OVERHEAD ;
#endif
    memUse = (memAvail / 100) * fs_incoreTune->bufMetaCompDirPctMem;
    usedLimit = memUse / sizeEach;
#if defined(FS_DEBUG_SHADOW_BUF)
    usedLimit = usedLimit / 4;   // include shadow buf
#endif
    fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimit[FS_BUF_CACHE_ITYPE_COMP_DIR] =
        usedLimit;
#ifdef FS_DYN_CACHE_SIZE
    fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimitLWM[FS_BUF_CACHE_ITYPE_COMP_DIR] =
        usedLimit;
    fs_bufListGlobal[FS_BUFLIST_META].count.usedBufLimitHWM[FS_BUF_CACHE_ITYPE_COMP_DIR] =
        (1 << 31);  // no limit yet
#endif
    fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufLimit[FS_BUF_CACHE_ITYPE_COMP_DIR] =
        (usedLimit / 100) * (100 - fs_incoreTune->bufPctAvailLoWtrMk);
    fs_bufListGlobal[FS_BUFLIST_META].count.dirtyBufLowThresh[FS_BUF_CACHE_ITYPE_COMP_DIR] =
        (usedLimit / 100) * (100 - fs_incoreTune->bufPctAvailHiWtrMk);
    fs_bufListGlobal[FS_BUFLIST_META].count.minInUse[FS_BUF_CACHE_ITYPE_COMP_DIR] =
        fs_incoreTune->metaCompDirMinInUse;
}


/**
 * Set the upper limit on the cache sizes.
 */
void
fs_bufSetCacheLimits()
{
    fs_bufSetMetaCacheLimit();
    fs_bufSetUserCacheLimit();
}


/**
 * Initialize the global portions of the buffer cache.
 */
void
fs_bufCacheInit()
{
    int         i, j;

#ifdef FS_DEBUG_BUF_HISTORY
    NFX_SPIN_INIT(&spinalloc,SPINLOCK_LEVEL_UNKNOWN);
#endif

    for (i = 0; i < FS_BUFLIST_MAX; i++)
    {
        fs_BufListGlobal *bufListGlobal = &fs_bufListGlobal[i];
        
        NFX_SPIN_INIT(&bufListGlobal->spin,SPINLOCK_LEVEL_UNKNOWN);
        bufListGlobal->type = i;

        for (j = 0; j < FS_BUF_CACHE_ITYPE_MAX; j++)
        {
            DLL_INIT(&bufListGlobal->mruHdr[j]);
        }

        for (j = 0; j < FS_BUFLIST_HASH_MAX; j++)
        {
            DLL_INIT(&bufListGlobal->bufsInum[j]);
            DLL_INIT(&bufListGlobal->bufsBlk[j]);
            DLL_INIT(&bufListGlobal->bufsPage[j]);
        }
#ifdef FS_BUF_BLK_CHECK_HASH_DEPTH
        bzero(bufListGlobal->blkDepth, sizeof(bufListGlobal->blkDepth));
#endif
#ifdef FS_BUF_PAGE_CHECK_HASH_DEPTH
        bzero(bufListGlobal->pageDepth, sizeof(bufListGlobal->pageDepth));
#endif
#ifdef FS_BUF_INUM_CHECK_HASH_DEPTH
        bzero(bufListGlobal->inumDepth, sizeof(bufListGlobal->inumDepth));
#endif
    }

    fs_bufSetCacheLimits();
}


/**
 * Instantiate the writeback daemon state machines for the
 * specified filesystem.
 */
void
fs_bufStartWritebackDaemon(
    fs_Context  *context)
{
    fs_FileSystem   *fileSystem = context->fileSystem;
    int32           i, j;

    for (i = 0; i < FS_BUFLIST_MAX; i++)
    {
        fs_BufList      *bufList = &fileSystem->bufList[i];
        int32           rc;
        
        bufList->type = i;
        bufList->wbContext = fs_ramAlloc(FS_RAM_CORE_MISC, sizeof (fs_Context));
        bzero(bufList->wbContext, sizeof(fs_Context));

        bufList->wbContext->volId = context->volId;
        bufList->wbContext->fileSystem = context->fileSystem;

        if (i == FS_BUFLIST_META)
            bufList->wbContext->daemonType = FS_DAEMON_TYPE_WB_META;
        else
            bufList->wbContext->daemonType = FS_DAEMON_TYPE_WB_USER;

        DLL_INIT(&bufList->waiterHdr);

        DLL_INIT(&bufList->mrdHdr);
        DLL_INIT(&bufList->mrdWaitingHdr);
        for (j = 0; j < FS_LOG_MAX_SEGMENTS; j++)
        {
            DLL_INIT(&bufList->logSegmentHdr[j]);
        }

        //Lock lists init
        bufList->wbContext->flags |= FS_CONTEXT_ALLOW_LOCK_STARVATION;
        fs_lockContextInit(bufList->wbContext, 512);

        FS_LOCK_ORDER_DISABLE(bufList->wbContext);

        rc = fs_wb_sm_create(&bufList->wbHandle, bufList->wbContext);
        if (rc != NFX_OK)
        {
            fs_error(NULL, FS_ESM_CREATE_ERR, "wb esm");
        }
        bufList->wbContext->smHandle = bufList->wbHandle;
        //printf("bufList %d wb daemon handle 0x%x\n",
        //       i, bufList->wbHandle);

#if !defined(FS_STANDALONE)
        bufList->wbTimerHandle = eee_timerCreate(FS_WB_TIMEOUT,
                                                 EEE_TIMER_REPEAT);
        if (bufList->wbTimerHandle == TIMER_ERR_NO_TIMER_AVAILABLE)
        {
            fs_error(NULL, FS_RAM_ALLOC_ERR, "create wb timer");
        }
        else
        {
            bufList->wbTimerExists = TRUE;
            if (eee_timerStart(bufList->wbTimerHandle, fs_writebackDaemonTimeout,
                               (void *) bufList, (void *) FS_WB_WAKEUP_PERIODIC, NULL, NULL) != NFX_OK)
            {
                fs_error(NULL, FS_INTERNAL_ERR, "start wb timer");
            }
        }
#endif
    }
}


/**
 * Destroy the writeback daemon state machines for the
 * specified filesystem and wait until they're actually
 * stopped.
 */
void
fs_bufStopWritebackDaemonWait(
    fs_Context  *context,
    boolean     isForce)
{
    fs_FileSystem   *fileSystem = context->fileSystem;
    esm_event_t     event;
    int32           rc, i, numWaits;

    FS_ASSERT(context->isThread);

    for (i = 0; i < FS_BUFLIST_MAX; i++)
    {
        fs_BufList      *bufList = &fileSystem->bufList[i];

        if (bufList->wbContext != NULL)
        {
            event.event = FS_WB_EVENT_STOP;
            event.p1.u.p = bufList;
            event.p2.u.p = NULL;
            rc = esm_deliverEvent(bufList->wbHandle, &event);
            if (rc != NFX_OK)
            {
                fs_error(context, FS_EEE_ERR, "bufStopWriteback deliver event failed");
            }
        }
    }

// about 60 seconds
#define MAX_WAITS   50
#define WAIT_MSECS  50
    for (i = 0; i < FS_BUFLIST_MAX; i++)
    {
        fs_BufList      *bufList = &fileSystem->bufList[i];

        numWaits = 0;
        while (numWaits++ < MAX_WAITS)
        {
            if (bufList->wbContext != NULL)
            {
                if (numWaits > MAX_WAITS / 2)
                {
                    printf("unmount, still waiting for writeback daemon to terminate\n");
                }
                if (isForce)
                {
                    // We're in exception condition already,
                    // don't call fs_sleep, since it will
                    // call the exception handler.
                    thread_sleep(NULL, numWaits * WAIT_MSECS);
                }
                else
                {
                    fs_sleep(context, numWaits * WAIT_MSECS);
                }
            }
            else
            {
                break;
            }
        }
        if (numWaits >= MAX_WAITS)
        {
            if (isForce)
            {
                fs_error(context, FS_INTERNAL_ERR, "bufStopWriteback failed");
            }
            else
            {
                fs_volumeException(context, context->fileSystem,
                                   FS_VOLUME_EXCEPTION);
            }
        }
    }
#undef MAX_WAITS
#undef WAIT_MSECS

}

/*
 * Asynchronous Get Page routine. This routine will not block for
 * i/o reads of data but can block for i/o completion for intermediate
 * btree nodes. The probability of wait for intermediate nodes is very
 * low, so this should not be an issue.  It returns the number of I/O's
 * started.
 */

int32
fs_bufGetPageAsync(
    fs_Context      *context,
    fs_Inode        *inode,
    uint64          page,
    uint32          len,
    uint32          type        // type of buffer list to operate on
)
{
    fs_BufList      *bufList;
    fs_RdwrReq      *rdwr;
    fs_RdwrReq      *rdwrArray[FS_MAX_READ_CLUSTER];
    uint32          typeBit;
    boolean         isTryLock = 1;  // If buffer is locked by someone else
                                    // assume it is read in
    nfx_spinlock_t  *spinlock;
    fs_Buf          *array[FS_MAX_READ_CLUSTER];
    int             i = 0;
    fs_Buf *buf;
    fs_Block block;
    int     count = 0, numRead = 0;
    esm_handle_t    raDoneHandle;
    int32           numRetries = 0;

    //printf("Read ahead for %lld page %lld len %lld size %lld\n",
    //      inode->inum, page, len, inode->dinode.size);
    FS_ASSERT(type == FS_BUFLIST_USER || type == FS_BUFLIST_META);
    FS_ASSERT(len <= FS_MAX_READ_CLUSTER);

    if (context->fileSystem->ioPendingCount >=
            fs_incoreTune->perFsIoPendingMax ||
        fs_ioPending >=
            FS_RDWR_MAX)
    {
        // Sorry we're already oversubscribed.
        return 0;
    }

    typeBit = (type == FS_BUFLIST_USER) ? FS_BUF_TYPE_USER : FS_BUF_TYPE_META;
    bufList = &context->fileSystem->bufList[type];
    spinlock = &fs_bufListGlobal[type].spin;
    raDoneHandle = bufList->wbHandle;

    while (i < len)
    {
getPageAsyncRetry:
        array[i] = NULL;
        rdwrArray[i] = NULL;
        // Search the cache for the requested (inode,page).
        NFX_SPIN_LOCK(spinlock);

        buf = fs_bufLockByPage(context, inode->inum, inode->snapnum, bufList, page, 
                                FS_LOCK_MODE_SHARE, isTryLock);
        if (buf != NULL)
        {
            if (buf == FS_BUF_TRY_LOCK_FAILED)
            {
                NFX_SPIN_UNLOCK(spinlock);
                i++;
                page++;
                continue;
            }
            // We hit in the cache
            // Remove it from the free list.
            if (!DLL_IS_ISOLATED(buf, mruEntry))
            {
                int32 itype = fs_bufGetItype(buf);

                DLL_REMOVE(&fs_bufListGlobal[bufList->type].mruHdr[itype],
                           buf, mruEntry);
                fs_bufListGlobal[bufList->type].count.freeBufs[itype] -= 1;
            }
            NFX_SPIN_UNLOCK(spinlock);
            FS_ASSERT(context->fileSystem == buf->fileSystem);
            fs_bufRelease(context, buf);
            i++;
            page++;
            continue;
        }

        NFX_SPIN_UNLOCK(spinlock);

        // Didn't find it.  Find its block number

        block = fs_emap_lookup_block(context, inode, page);

        if (!fs_blockIsValid(block))
        {
            // Must be a hole in the file, return NULL.
            BUF_TRACE(("  file hole ino %lld page %lld",
                      inode->inum, page));
            i++;
            page++;
            continue;
        }

        NFX_SPIN_LOCK(spinlock);
        // Since we let go of the bufList lock since our
        // last check in the page cache, we need to do it
        // again.
        buf = fs_bufLockByPage(context, inode->inum, inode->snapnum, bufList, page, 
                                FS_LOCK_MODE_SHARE, isTryLock);
        if (buf != NULL)
        {
            if (buf == FS_BUF_TRY_LOCK_FAILED)
            {
                NFX_SPIN_UNLOCK(spinlock);
                i++;
                page++;
                continue;
            }
            // We hit in the cache
            // Remove it from the free list.
            if (!DLL_IS_ISOLATED(buf, mruEntry))
            {
                int32 itype = fs_bufGetItype(buf);

                DLL_REMOVE(&fs_bufListGlobal[bufList->type].mruHdr[itype],
                           buf, mruEntry);
                fs_bufListGlobal[bufList->type].count.freeBufs[itype] -= 1;
            }
            NFX_SPIN_UNLOCK(spinlock);
            FS_ASSERT(context->fileSystem == buf->fileSystem);
            fs_bufRelease(context, buf);
            i++;
            page++;
            continue;
        }

        // Search the block cache for the buffer

        buf = fs_bufLockByBlock(context, bufList, block, 
                                FS_LOCK_MODE_SHARE, isTryLock);
        if (buf == FS_BUF_TRY_LOCK_FAILED)
        {
            NFX_SPIN_UNLOCK(spinlock);
            i++;
            page++;
            continue;
        }
        if (buf != NULL)
        {
            // The block was found, which means it's part of the same
            // file, but only currently loaded for a different snapshot,
            // or it was read into the blocknum cache directly, but not
            // the page cache (eg. during replay).
            uint64 snap = 1ULL << inode->snapnum;

            FS_ASSERT(buf->fileSystem == context->fileSystem);
            FS_ASSERT(buf->snapshots != 0);

            // Replay should have set these up already.
            FS_ASSERT(buf->inum == inode->inum);
            FS_ASSERT(buf->page == page);

            if (DLL_IS_ISOLATED(buf, pageEntry))
            {
                // Add it to the page cache now.
                fs_bufPageCacheInsert(context, buf, bufList, inode);

                // need to bump up the page counter here if dirty

                if (buf->flags & FS_BUF_DIRTY)
                {
                    fs_BufPageHdr   *pageHdr;

                    // Has already been dirtied, but wasn't on page list
                    // before, so increment pageHdr dirty count here.

                    pageHdr = fs_bufPageHdrForInum(buf->fileSystem,
                                                   bufList->type, buf->inum);
                    FS_ASSERT(pageHdr != NULL);
                    FS_ASSERT(!DLL_IS_ISOLATED(buf, mrdEntry));
                    FS_ASSERT(DLL_IS_ISOLATED(buf, pageDirtyEntry));
                    pageHdr->dirtyCount += 1;
                    // insert to dirty page hdr
                    DLL_INSERT_HEAD(&pageHdr->bufDirtyHdr, buf, pageDirtyEntry);
#ifdef FS_DEBUG_BUF_HISTORY
                    NFX_SPIN_LOCK(&buf->spinHistory);
                    buf->history[buf->indexHistory] = FS_BUF_HIST_INSERT_PAGE_GET;
                    buf->cpuHistory[buf->indexHistory] = nfxMyCPUId;
                    buf->flagsHistory[buf->indexHistory] = buf->flags;
                    buf->mrdEntryNextHistory[buf->indexHistory] = buf->mrdEntry.next;
                    buf->mrdEntryPrevHistory[buf->indexHistory] = buf->mrdEntry.prev;
                    buf->indexHistory = (buf->indexHistory + 1) % FS_DEBUG_BUF_HISTORY_MAX;
                    NFX_SPIN_UNLOCK(&buf->spinHistory);
#endif
                }
            }

            buf->snapshots |= snap;

            // Remove it from the free list.
            if (!DLL_IS_ISOLATED(buf, mruEntry))
            {
                int32 itype = fs_bufGetItype(buf);

                DLL_REMOVE(&fs_bufListGlobal[bufList->type].mruHdr[itype],
                           buf, mruEntry);
                fs_bufListGlobal[bufList->type].count.freeBufs[itype] -= 1;
            }
            NFX_SPIN_UNLOCK(spinlock);

            /*
             * If this was a read ahead request release the buffer
             */

            fs_bufRelease(context, buf);
            i++;
            page++;
            continue;
        }
        
        // Instantiate it and schedule I/O if we have an
        // extent for the page. fs_bufAllocPtr returns with
        // the buf locked excl.
        buf = fs_bufAllocPtr(context, bufList->type, inode, page);
        if (buf == NULL)
        {
            NFX_SPIN_UNLOCK(spinlock);
            fs_bufWaitAvail(context, &numRetries);
            goto getPageAsyncRetry;
        }
        buf->fileSystem = context->fileSystem;
        buf->inum = inode->inum;
        buf->page = page;
        buf->snapshots = 1ULL << inode->snapnum;
        buf->flags |= (typeBit | FS_BUF_COW_UNK);
        buf->block = FS_INVALID_BLOCK;
        array[i] = buf;
        rdwrArray[i] = fs_rdwrAlloc();

        // Add the entry to the cache to prevent others
        // from instantiating this entry while we're waiting
        // for I/O to complete.
        fs_bufPageCacheInsert(context, buf, bufList, inode);

        // Now add it to the blknum cache, too.
        buf->block = block;
        fs_bufBlknumCacheInsert(context, buf, bufList, inode);
        NFX_SPIN_UNLOCK(spinlock);
        i++;
        page++;
        count++;
        continue;
    }
    
    for (i = 0; i < len; i++) {
        if (array[i] == NULL) {
            continue;
        }
        buf = array[i];
        rdwr = rdwrArray[i];
        if (type == FS_BUFLIST_USER)
        {
            NFX_SPIN_LOCK(&fs_ioSpin);
            fs_ioTotalUserReads += 1;
            NFX_SPIN_UNLOCK(&fs_ioSpin);

            NFX_SPIN_LOCK(&context->fileSystem->spin);
            context->fileSystem->ioTotalUserReads += 1;
            NFX_SPIN_UNLOCK(&context->fileSystem->spin);

            fs_lockGiveExclusive(context, bufList->wbContext, &buf->lock);
            fs_volReadAheadUserBlock(context, rdwr, buf, raDoneHandle);
        }
        else
        {
            NFX_SPIN_LOCK(&fs_ioSpin);
            fs_ioTotalMetaReads += 1;
            NFX_SPIN_UNLOCK(&fs_ioSpin);

            NFX_SPIN_LOCK(&context->fileSystem->spin);
            context->fileSystem->ioTotalMetaReads += 1;
            NFX_SPIN_UNLOCK(&context->fileSystem->spin);

            fs_lockGiveExclusive(context, bufList->wbContext, &buf->lock);
            fs_volReadAheadBlock(context, rdwr, buf, raDoneHandle);
        }
        numRead += 1;
    }

    return numRead;

} /* fs_bufGetPageAsync */


/**
 * 
 */
void
fs_bufLookupBlock(fs_Context *  context,
                  uint64        inum,
                  uint64        page,
                  fs_Block      block,
                  uint32        type,
                  boolean       sync)

{
    fs_FileSystem *     fs = context->fileSystem;
    fs_Buf *            buf = NULL;
    fs_BufList *        bufList = &fs->bufList[type];
    esm_handle_t        raDoneHandle = bufList->wbHandle;
    nfx_spinlock_t *    spinlock = &fs_bufListGlobal[type].spin;
    fs_RdwrReq *        rdwr = NULL;
    uint32              typeBit = ((type == FS_BUFLIST_USER)?
                                   FS_BUF_TYPE_USER : FS_BUF_TYPE_META);
    boolean             isTryLock = TRUE;   // If buffer is locked by someone
                                            // else assume it is read in.
    
    //printf("fs_bufLookupBlockAsync(inum %lld page %lld block %lld type %d)\n",
    //inum, page, block.block, type);
    
    NFX_SPIN_LOCK(spinlock);

    // Search the block cache for the buffer.
    buf = fs_bufLockByBlock(context, bufList, block, 
                            FS_LOCK_MODE_SHARE, isTryLock);
    if (buf != NULL)
    {
        // Buf is (or will very soon) be in the cache.
        NFX_SPIN_UNLOCK(spinlock);
        if (buf != FS_BUF_TRY_LOCK_FAILED)
        {
            fs_unlockShared(context, &buf->lock);
        }
        return;
    }
    
    // Instantiate it and schedule I/O if we have an
    // extent for the page. fs_bufAllocPtr returns with
    // the buf locked excl.
    buf = fs_bufAllocPtr(context, bufList->type, NULL, page);
    if (buf == NULL)
    {
        // Don't wait for a buf, will come back later.
        NFX_SPIN_UNLOCK(spinlock);
        return;
    }
    buf->fileSystem = fs;
    buf->inum = inum;
    buf->page = page;
    buf->snapshots = 1;
    buf->flags |= (typeBit | FS_BUF_COW_UNK);
    buf->block = block;
    
    // Add the entry to the cache to prevent others from instantiating
    // this entry while we're waiting for I/O to complete.
    fs_bufBlknumCacheInsert(context, buf, bufList, NULL);

    NFX_SPIN_UNLOCK(spinlock);

    if (!sync)
    {
        // Start the asynchronous read now.
        rdwr = fs_rdwrAlloc();
#ifdef ISABELLE // debug
        {
            vl_VolBufRoot * volBufRoot = fs->logRoot->volBufRoot;
            vl_BlockCache * blockCache = &(volBufRoot->blockCache);
            blockCache->stats.nbMetaReadAhead++;
        }
#endif // ISABELLE
        if (type == FS_BUFLIST_USER)
        {
            NFX_SPIN_LOCK(&fs_ioSpin);
            fs_ioTotalUserReads += 1;
            NFX_SPIN_UNLOCK(&fs_ioSpin);
            
            NFX_SPIN_LOCK(&fs->spin);
            fs->ioTotalUserReads += 1;
            NFX_SPIN_UNLOCK(&fs->spin);
            
            fs_lockGiveExclusive(context, bufList->wbContext, &buf->lock);
            fs_volReadAheadUserBlock(context, rdwr, buf, raDoneHandle);
        }
        else
        {
            NFX_SPIN_LOCK(&fs_ioSpin);
            fs_ioTotalMetaReads += 1;
            NFX_SPIN_UNLOCK(&fs_ioSpin);
            
            NFX_SPIN_LOCK(&fs->spin);
            fs->ioTotalMetaReads += 1;
            NFX_SPIN_UNLOCK(&fs->spin);
            
            fs_lockGiveExclusive(context, bufList->wbContext, &buf->lock);
            fs_volReadAheadBlock(context, rdwr, buf, raDoneHandle);
        }
    }
    else // sync
    {
        // Read synchronous now.
        rdwr = fs_rdwrAlloc();
#ifdef ISABELLE // debug
        {
            vl_VolBufRoot * volBufRoot = fs->logRoot->volBufRoot;
            vl_BlockCache * blockCache = &(volBufRoot->blockCache);
            blockCache->stats.nbMetaRead++;
        }
#endif // ISABELLE
        if (type == FS_BUFLIST_USER)
        {
            NFX_SPIN_LOCK(&fs_ioSpin);
            fs_ioTotalUserReads += 1;
            NFX_SPIN_UNLOCK(&fs_ioSpin);
            
            NFX_SPIN_LOCK(&fs->spin);
            fs->ioTotalUserReads += 1;
            NFX_SPIN_UNLOCK(&fs->spin);
            
            //fs_lockGiveExclusive(context, bufList->wbContext, &buf->lock);
            fs_volReadUserBlock(context, rdwr, buf->block);
            
            // Plug in the just-read user data.
            buf->u.user.codHandle = rdwr->evmStat.rDataHandle.codHandle;
            buf->u.user.codLength = rdwr->evmReq.reqSize * EVM_LBLK_SIZE;
            buf->u.user.codOffset = 0;
            FS_ASSERT(buf->u.user.codLength <= FS_BLOCK_SIZE);
        }
        else
        {
            NFX_SPIN_LOCK(&fs_ioSpin);
            fs_ioTotalMetaReads += 1;
            NFX_SPIN_UNLOCK(&fs_ioSpin);
            
            NFX_SPIN_LOCK(&fs->spin);
            fs->ioTotalMetaReads += 1;
            NFX_SPIN_UNLOCK(&fs->spin);
            
            //fs_lockGiveExclusive(context, bufList->wbContext, &buf->lock);
            fs_volReadBlock(context, rdwr, buf->block);

            // Plug in the just-read meta data.
            FS_ASSERT(buf->u.data == NULL);
            fs_bufSetDataPtr(context, buf, rdwr->evmStat.rDataHandle.dataPtr);
#if defined(FS_DEBUG_SHADOW_BUF)
            // Initialize the shadow buffer.
            FS_DBG_INIT_SHADOW(buf);
#endif            
#if defined(FS_DEBUG_BUF_CKSUM)
            fs_bufCompareCksum(context, buf->block, buf->u.data, buf,
                               "get page", FALSE);
#endif
        }
        
        fs_bufReleaseExclusive(context, buf);
        fs_rdwrDealloc(rdwr);
    }

    return;
} /* fs_bufLookupBlockAsync */


/**
 * Writeback daemon state machine read-ahead done action.
 */
int32
fs_wb_raDone(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *smCookie)
{
    fs_Context      *context;
    fs_RdwrReq      *rdwr;
    fs_Buf          *buf;

    context = smCookie;
    rdwr = event.p1.u.p;

    FS_ASSERT(rdwr >= fs_rdwrStart &&
              rdwr < fs_rdwrEnd);

    buf = (fs_Buf *) rdwr->private;
    context->fileSystem = buf->fileSystem;
    context->volId = buf->fileSystem->volId;

    NFX_SPIN_LOCK(&context->fileSystem->spin);
    context->fileSystem->ioPendingCount -= 1;
    NFX_SPIN_UNLOCK(&context->fileSystem->spin);
    if (buf->flags & FS_BUF_TYPE_META)
    {
        // Plug in the just-read meta data.

#if !defined(FS_STANDALONE)
        // hack to make the mem stats look right
        NFX_SPIN_LOCK(&fs_ramStatsSpinId);
        fs_ramStats[FS_RAM_STATS_ALLOC][FS_RAM_BUF_DATA] += 1;
        NFX_SPIN_UNLOCK(&fs_ramStatsSpinId);
#endif
        FS_ASSERT(buf->u.data == NULL);
        fs_bufSetDataPtr(context, buf, rdwr->evmStat.rDataHandle.dataPtr);

        // Initialize the shadow buffer.
        FS_DBG_INIT_SHADOW(buf);

#if defined(FS_DEBUG_BUF_CKSUM)
        fs_bufCompareCksum(context, buf->block, buf->u.data, buf,
                           "get page", FALSE);
#endif
    }
    else
    {
        // Plug in the just-read user data.
        buf->u.user.codHandle = rdwr->evmStat.rDataHandle.codHandle;
        buf->u.user.codLength = rdwr->evmReq.reqSize * EVM_LBLK_SIZE;
        buf->u.user.codOffset = 0;
        FS_ASSERT(buf->u.user.codLength <= FS_BLOCK_SIZE);
    }

    NFX_SPIN_LOCK(&buf->spin);
    buf->flags |= FS_BUF_COW_UNK;
    NFX_SPIN_UNLOCK(&buf->spin);

#ifdef ISABELLE
    printf("fs_wb_raDone(inum %lld page %lld block %lld)\n",
           buf->inum, buf->page, buf->block.block);
    //printf("RADONE: lock owner %d count %d\n", buf->lock.owner, buf->lock.count);
#endif // ISABELLE

    fs_bufReleaseExclusive(context, buf);

    fs_rdwrDealloc(rdwr);

    return nextState;

} /* fs_wb_raDone */


/**
 * Flush the buffers that have been deallocated since reallocation
 * of the same block will produce "duplicate cache insert".
 * Called during snapshot remove for blocks that are deallocated
 * (vs flushing entire buffer cache).
 */
void
fs_bufCacheFlushBlocks(
    fs_Context *context,
    uint64 startBlock,
    uint32 len)
{
    uint64 block, endBlock = startBlock+len;
    fs_BufList *bufList;
    fs_Buf *buf;
    int32 type;

    //printf("buf flush blocks: %lld to %lld\n", startBlock, endBlock-1);

    for (type = 0; type < FS_BUFLIST_MAX; type++)
    {
        bufList = &context->fileSystem->bufList[type];
        NFX_SPIN_LOCK(&fs_bufListGlobal[type].spin);
        for (block = startBlock; block < endBlock; block++)
        {
            buf = fs_bufFindByBlock(context, type, fs_makeBlock(block));
            if (buf == NULL) 
                continue;
            //printf("  %p ino %lld blk %lld\n",
            //      buf, buf->inum, fs_blockNumber(buf->block));
 
            // if debug is on (== FS_ASSERT) we check if no one has the lock (excl, shared)
            // and fs_bufDelist FS_ASSERT(isForce || fs_lockOwnedExclusive(context, &buf->lock));
            FS_ASSERT(fs_trylockExclusive(context, &buf->lock));

            fs_bufDelist(context, bufList, buf, FALSE);
            fs_bufDeallocCoreRes(context, buf);

#ifdef FS_DEBUG
            // if debug is on, need to destroy the lock
            fs_lockDestroy(context, &buf->lock);
#endif

            fs_ramDealloc(FS_RAM_BUF_PTR, buf);
#ifdef FS_DEBUG_BUF_HISTORY
            NFX_SPIN_LOCK(&spinalloc);
            fs_buf_allocate_ptr[fs_buf_allocate_index] = buf;
            fs_buf_allocate_oper[fs_buf_allocate_index] = FS_DEBUG_BUF_HISTORY_DEALLOC;
            fs_buf_allocate_index = (fs_buf_allocate_index + 1) % FS_DEBUG_BUF_HISTORY_ALLOC_MAX;
            NFX_SPIN_UNLOCK(&spinalloc);
#endif
        }
        NFX_SPIN_UNLOCK(&fs_bufListGlobal[type].spin);
    }

} /* fs_bufCacheFlushBlocks */


/**
 * Enable COW mechanism after a snapshot operation (create).
 * Note that only the FS_BUF_COW_UNK flag is re-enabled and not the
 * FS_BUF_COW flag.  To set the cow flag will require these extra checks:
 *
 *  1. only if block has snapcount > 0
 *  2. only if page != FS_BUF_NO_PAGE (btree blocks are not COW'd)
 *  3. only if inum != FS_SNAPSHOT_BITS_INODE (snapbits pages are not COW'd)
 *
 * So just set the unknown flag and defer checks til fs_bufWrite() time.
 */
void fs_bufCacheEnableCow(fs_Context *context)
{
    fs_BufListGlobal *bufListGlobal;
    int32 listType, hash, itype;
    fs_Buf *buf;

    //printf("buf enable cow:\n");

    for (listType = 0; listType < FS_BUFLIST_MAX; listType++)
    {
        bufListGlobal = &fs_bufListGlobal[listType];
        NFX_SPIN_LOCK(&bufListGlobal->spin);
        for (itype = 0; itype < FS_BUF_CACHE_ITYPE_MAX; itype++)
        {
            DLL_ITERATE(buf, &bufListGlobal->mruHdr[itype], mruEntry)
            {
                if (buf->fileSystem == context->fileSystem)
                {
                    NFX_SPIN_LOCK(&buf->spin);
                    buf->flags |= FS_BUF_COW_UNK;
                    //printf("  mru %p ino %lld blk %lld page %lld flags %x\n",
                    //      buf, buf->inum, fs_blockNumber(buf->block), buf->page, buf->flags);
                    NFX_SPIN_UNLOCK(&buf->spin);
                }
            }
        }

        for (hash = 0; hash < FS_BUFLIST_HASH_MAX; hash++)
        {
            DLL_ITERATE(buf, &fs_bufListGlobal[listType].bufsPage[hash],
                        pageHashEntry)
            {
                if (buf->fileSystem == context->fileSystem)
                {
                    NFX_SPIN_LOCK(&buf->spin);
                    buf->flags |= FS_BUF_COW_UNK;
                    //printf("  page %p ino %lld blk %lld page %lld flags %x\n",
                    //      buf, buf->inum, fs_blockNumber(buf->block), buf->page, buf->flags);
                    NFX_SPIN_UNLOCK(&buf->spin);
                }
            }
        }

        NFX_SPIN_UNLOCK(&bufListGlobal->spin);
    }

} /* fs_bufCacheEnableCow */

int32
fs_bufGetBlockAsync(
    fs_Context      *context,
    fs_Inode        *inode,
    uint64          blk,
    uint32          len,
    uint32          type        // type of buffer list to operate on
)
{
    fs_BufList      *bufList;
    fs_RdwrReq      *rdwr;
    fs_RdwrReq      *rdwrArray[FS_MAX_READ_CLUSTER];
    uint32          typeBit;
    boolean         isTryLock = 1;  // If buffer is locked by someone else
                                    // assume it is read in
    nfx_spinlock_t  *spinlock;
    fs_Buf          *array[FS_MAX_READ_CLUSTER];
    int             i = 0;
    fs_Buf *buf;
    int     count = 0, numRead = 0;
    esm_handle_t    raDoneHandle;
    int32           numRetries = 0;

    //printf("Read ahead for %lld page %lld len %lld size %lld\n",
    //      inode->inum, page, len, inode->dinode.size);
    FS_ASSERT(type == FS_BUFLIST_USER || type == FS_BUFLIST_META);
    FS_ASSERT(len <= FS_MAX_READ_CLUSTER);

    typeBit = (type == FS_BUFLIST_USER) ? FS_BUF_TYPE_USER : FS_BUF_TYPE_META;
    bufList = &context->fileSystem->bufList[type];
    spinlock = &fs_bufListGlobal[type].spin;
    raDoneHandle = bufList->wbHandle;

    while (i < len)
    {
getPageAsyncRetry:
        array[i] = NULL;
        rdwrArray[i] = NULL;
        // Search the cache for the requested (inode,page).
        NFX_SPIN_LOCK(spinlock);

        buf = fs_bufLockByBlock(context, bufList, fs_makeBlock(blk), 
                                FS_LOCK_MODE_SHARE, isTryLock);
        if (buf != NULL)
        {
            if (buf == FS_BUF_TRY_LOCK_FAILED)
            {
                NFX_SPIN_UNLOCK(spinlock);
                i++;
                blk++;
                continue;
            }
            // We hit in the cache
            // Remove it from the free list.
            if (!DLL_IS_ISOLATED(buf, mruEntry))
            {
                int32 itype = fs_bufGetItype(buf);

                DLL_REMOVE(&fs_bufListGlobal[bufList->type].mruHdr[itype],
                           buf, mruEntry);
                fs_bufListGlobal[bufList->type].count.freeBufs[itype] -= 1;
            }
            NFX_SPIN_UNLOCK(spinlock);
            FS_ASSERT(context->fileSystem == buf->fileSystem);
            fs_bufRelease(context, buf);
            i++;
            blk++;
            continue;
        }

        
        // Instantiate it and schedule I/O if we have an
        // extent for the page. fs_bufAllocPtr returns with
        // the buf locked excl.
        buf = fs_bufAllocPtr(context, bufList->type, inode, FS_BUF_NO_PAGE);
        if (buf == NULL)
        {
            NFX_SPIN_UNLOCK(spinlock);
            fs_bufWaitAvail(context, &numRetries);
            goto getPageAsyncRetry;
        }
        buf->fileSystem = context->fileSystem;
        buf->inum = inode->inum;
        buf->page = FS_BUF_NO_PAGE;
        buf->snapshots = 0;
        buf->flags |= (typeBit | FS_BUF_COW_UNK);
        buf->block = fs_makeBlock(blk);
        array[i] = buf;
        rdwrArray[i] = fs_rdwrAlloc();
        fs_bufBlknumCacheInsert(context, buf, bufList, inode);
        NFX_SPIN_UNLOCK(spinlock);
        i++;
        blk++;
        count++;
        continue;
    }
    
    for (i = 0; i < len; i++) {
        if (array[i] == NULL) {
            continue;
        }
        buf = array[i];
        rdwr = rdwrArray[i];
        if (type == FS_BUFLIST_USER)
        {
            NFX_SPIN_LOCK(&fs_ioSpin);
            fs_ioTotalUserReads += 1;
            NFX_SPIN_UNLOCK(&fs_ioSpin);

            NFX_SPIN_LOCK(&context->fileSystem->spin);
            context->fileSystem->ioTotalUserReads += 1;
            NFX_SPIN_UNLOCK(&context->fileSystem->spin);

            fs_lockGiveExclusive(context, bufList->wbContext, &buf->lock);
            fs_volReadAheadUserBlock(context, rdwr, buf, raDoneHandle);
        }
        else
        {
            NFX_SPIN_LOCK(&fs_ioSpin);
            fs_ioTotalMetaReads += 1;
            NFX_SPIN_UNLOCK(&fs_ioSpin);

            NFX_SPIN_LOCK(&context->fileSystem->spin);
            context->fileSystem->ioTotalMetaReads += 1;
            NFX_SPIN_UNLOCK(&context->fileSystem->spin);

            fs_lockGiveExclusive(context, bufList->wbContext, &buf->lock);
            fs_volReadAheadBlock(context, rdwr, buf, raDoneHandle);
        }
        numRead += 1;
    }

    return numRead;

} /* fs_bufGetBlockAsync */


/**
 * Make a copy-on-write copy of the block at the specified number.
 */
uint64 fs_make_block_copy(
        fs_Context *context,
        fs_Inode *inode,
        uint64 blknum)
{
    fs_Buf *buf = B_GET_EXCL(inode, fs_makeBlock(blknum));

    fs_bufCopyOnWriteMetadata(context, inode, buf);
    fs_bufReleaseExclusive(context, buf);
    return fs_blockNumber(buf->block);
}

#endif // SSC
