/*
 * Name:  fs-nfs.c
 *
 * RCS:  $Id: fs-nfs.c,v 1.207 2002/12/10 21:02:18 bob Exp $
 *
 * Description:  Handlers for NFSv2/v3 requests.
 *
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#if !defined(SSC)

#include "fs-includes.h"

// some statistics to check the impact of copy COD's
uint64 fs_cod_nbr_copies = 0 ;
uint64 fs_cod_nbr_bytes = 0 ;
uint64 fs_cod_nbr_overwrite = 0 ;

int32 fs_txComplTimeouts = 0;

char *fs_protoStr[FS_PROTO_MAX + 1] = {
    "undef",
    "nfs_v2",
    "nfs_v3",
    "nfs_v4",
    "cifs"
};


/**
 * In order for us to be able to distinguish between the live
 * version and snapshot versions of a file, and for the client
 * to be able to do the same, and for us to properly handle
 * requests for a ".snapshot" directory and its children, we
 * play a few games with nfs file handles and attributes.  This
 * table attempts to clarify what we're doing.  Basically, we
 * need to ensure that every object has a unique id,fsid pair in
 * order for the client to manage itself, and we of course need
 * the handles we pass out to guide us to the correct fs object,
 * be it a generated object (eg. .snapshot dir) or an on-disk
 * object.

ObjectType         Handle (inum,snap,genNum)  Attributes (id,fsid)
-------------------------------------------------------------------
regular file/dir   inum,snapnum,genNum        inum,vol&snap

.snapshot dir      parentInum,0,magic         snapDirInum,parentInum

snapshot boundary   

if parentParentInum is the same as the mnt point
(ie /mnt/.snapshot/snap1)

                   parentParentInum,snapnum,  snapnumInum, parentParentInum
                   genNum
else
(ie /mnt/dir/.snapshot/snap1, or /dir1/dir2/.snapshot/snap1)
                   parentParentInum,snapnum   inum,vol&snap
                   genNum

 *
 * The magic value in the genNum field is used to identify the handle as
 * pertaining to a ".snapshot" directory for the inum identified in
 * the handle.
 *
 * Another magic value in the genNum field is used to identify that
 * the ".snapshot" directory has been traversed. Normally, the ".snapshot" directory
 * is included only for the mount point.  But for certain shells' (ie CShell)
 * implementation of "pwd", the ".snapshot" directory should also be included
 * in the readdir response for a handle with this magic value set:

        eg cd /mnt/dir1/dir2/.snapshot
                -> lookup dir1
                -> lookup dir2
                -> lookup .snapshot
           pwd
                -> lookup ..
                    return handle with magic2 set
                -> readdir
                     .snapshot needs to be included in this readdir response
                     or the pwd algorithm will fail

ObjectType         Handle (inum,snap,genNum)  Attributes (id,fsid)
-------------------------------------------------------------------
dotdot for         inum,snapnum,magic2        inum,vol&snap
.snapshot

 *
 * Note that a .snapshot dirent object is essentially a link to
 * the specified snapnum version of the directory that its parent
 * was generated for.  For example:
 *
 *   /dir1/dir2/.snapshot/snap1 would "point to" the snap1 version
 *   of /dir1/dir2.
 *
 * Also, to avoid clients from complaining about "circular directory structure"
 * (see PR #2600), use the snapnumInum (64-127) as the id and the
 * parentParentInum as the fsid attributes.
 */

#define FS_HANDLE_IS_SNAP_BOUNDARY(_handle)         \
    ((FS_HANDLE_GET_SNAP(_handle.inum) != 0) &&     \
     (FS_HANDLE_GET_INUM(_handle.inum) == _handle.mntPnt))

/**
 * Return TRUE if the specified handle refers to a .snapshot
 * directory or dirent.
 */
static inline boolean
fs_handleIsSnapMagic(
    fs_Handle   *handle)
{
    return ((handle->gen & FS_INODE_GENNUM_SNAP_MAGIC) ==
                                      FS_INODE_GENNUM_SNAP_MAGIC);
}


/**
 * Return TRUE if the specified handle indicate the .snapshot
 * directory should be included in the readdir response.
 */
static inline boolean
fs_handleIsSnapMagic2(
    fs_Handle   *handle)
{
    return ((handle->gen & FS_INODE_GENNUM_SNAP_MAGIC2) ==
                                      FS_INODE_GENNUM_SNAP_MAGIC2);
}


/**
 * Print the attributes.
 */
void
fs_printAttr(
    fs_Attr     *attr)
{
    printf("fsAttr:\n");
    printf("  type: %d\n", attr->type);
    printf("  mode: %o\n", attr->mode);
    printf("  linkCnt: %d\n", attr->numLinks);
    printf("  uid: %d\n", attr->uid);
    printf("  gid: %d\n", attr->gid);
    printf("  size: %lld\n", attr->size);
    printf("  blockSize: %d\n", attr->blockSize);
    printf("  blocks: %d\n", attr->blocks);
    printf("  fsId: %lld\n", attr->fsId);
    printf("  inode: %lld\n", attr->inode);

} /* fs_printAttr */


/**
 * Initialize a handle for the specified file in the filesystem in the
 * specified context.
 */
void
fs_handleInit(
    fs_FileSystem   *fileSystem,
    fs_Handle       *handle,
    uint8           snap,
    uint64          inum,
    uint32          gen,
    uint64          mountPointInum)
{
#define FS_HANDLE_VERS_CURRENT  0
#define FS_HANDLE_VERS_ZERO     0

    FS_ASSERT(inum < FS_VALID_INODE_LIMIT);

    FS_HANDLE_SET_SNAP(handle->inum, snap);
    FS_HANDLE_SET_INUM(handle->inum, inum);
    FS_HANDLE_SET_VERS(handle->inum, FS_HANDLE_VERS_CURRENT);

    handle->volId    = fileSystem->volId;
    handle->gen      = gen;
    handle->mkfsTime = fileSystem->super->mkfsTime;
    handle->mntPnt   = mountPointInum;

} /* fs_handleInit */


/**
 * Return TRUE if the specified handle is valid for
 * the specified inode, otherwise return FALSE.
 */
boolean
fs_handleIsValid(
    fs_Context  *context,
    fs_Handle   *handle,
    fs_Inode    *inode)
{
    if (!fs_handleIsSnapMagic(handle) &&
        !fs_handleIsSnapMagic2(handle) &&
        inode->dinode.genNum != handle->gen)
    {
        return FALSE;
    }

    if (context->fileSystem->super->mkfsTime != handle->mkfsTime)
        return FALSE;

    if (FS_HANDLE_GET_VERS(handle->inum) != FS_HANDLE_VERS_CURRENT)
    {
        return FALSE;
    }

    return TRUE;

} /* fs_handleIsValid */


/**
 * Release allocated extents for a file.
 * Also handles flushing the buffer cache of any
 * entries for the file.
 */
void
fs_truncateFile(
    fs_Context      		*context,
    fs_Inode        		*inode,
    uint64          		newSize,
	boolean					dontTrashCan,
	fs_QuotaContextList_t	**rlist)
{
    uint64 page;
    uint64 startPage = ceilingDiv64(newSize, FS_BLOCK_SIZE);
    uint64 maxPage = ceilingDiv64(inode->dinode.size, FS_BLOCK_SIZE);
    uint64 blockCount = 0;
    fs_QuotaContextList_t  *list = NULL;
    fs_Inode *trash_inode, *trashcanInode;
#if defined FS_DEBUG_TRASH_EMAP_LOOKUP
    eeeRTC_t beginTime;
#endif

    if (inode->dinode.numblocks == 0)
    {
        inode->dinode.size = newSize;
        fs_inodeWrite(context, inode, INO_WRITE_LOG);
        return;
    }
    //printf("truncate inum %lld from %lld to %lld\n",
    //  inode->inum, inode->dinode.size, newSize);

#ifdef NEW_ZF
    if (!FS_I_IS_DIR(inode->dinode.attrs))
    {
        /* Don't zero fill (cancel previous zero fill records for this inode).
         */
        if (newSize == 0)
        {
            fs_logAddUserZeroRecord(context, inode, 0, 0, 0,
                                    FALSE, FALSE, NULL);
        }
        else
        {
            fs_logAddUserZeroRecord(context, inode,
                                    startPage, 0, (maxPage - startPage),
                                    FALSE, FALSE, NULL);
        }
    }
#endif // NEW_ZF

    // Purge the removed pages from the buffer cache.
    fs_bufFreeInode(context, inode, startPage, FALSE);
    
    if (dontTrashCan ||
		inode->dinode.numblocks < FS_SYNC_TRUNCATE_MAX_BLOCKS)
    {
        blockCount = inode->dinode.numblocks;

    	FS_TRACE(context, FS_TRACE_REQ,
			("ino %lld truncating direct len %lld to len %lld",
             inode->inum, inode->dinode.size, newSize));

		if (newSize == 0)
		{
        	while (fs_emap_shrink(context, inode))
            	;
            inode->dinode.numblocks = 0;
		}
		else
		{
    		for (page = startPage; page < maxPage; page++)
    		{
    			fs_Block block;
		
        		block = fs_emap_lookup_block(context, inode, page);
        		if (fs_blockIsValid(block))
        		{
					fs_emap_remove_block(context, inode, page);
					fs_deallocateBlock(context, block);
                    inode->dinode.numblocks -= 1;
        		}
    		}
		}
        // PR #1826 make sure the inode is dirtied and locked
        // til commit time to ensure for proper lock order:
        // inode should remained lock if btree mod/locked
        fs_inodeWrite(context, inode, INO_WRITE_LOG);
    }
    else
    {
    	FS_TRACE(context, FS_TRACE_REQ,
			("ino %lld truncating with trashcan len %lld to len %lld",
             inode->inum, inode->dinode.size, newSize));

        // Large file, use the trashcan.
        trashcanInode = I_GET_EXCL(0, FS_TRASHCAN_INODE);
        if (newSize == 0)
        {
            trash_inode
                = fs_allocateFileInode(context, trashcanInode, 0700, 0, 0, 0);

	        blockCount = inode->dinode.numblocks;

            // copy extent map root, file size.
            trash_inode->dinode = inode->dinode;

			// set parentInum to FS_TRASHCAN_INODE, so that we don't take this
			// inode into account in quota calculation during quota on
			trash_inode->dinode.piu.parentInum = FS_TRASHCAN_INODE;

            // zero the extent map.
            fs_emaproot_init(context, inode);

            // zero the file size.
            inode->dinode.size = 0;
            inode->dinode.numblocks = 0;
            fs_inodeWrite(context, inode, INO_WRITE_LOG);
            fs_inodeWrite(context, trash_inode, INO_WRITE_LOG);
        }
        else
        {
            uint64 numBlocksBefore, numBlocksAfter;

            numBlocksBefore = inode->dinode.numblocks;
            trash_inode = fs_emap_cleave(context, trashcanInode,
                                         inode, newSize);
            numBlocksAfter = inode->dinode.numblocks;
            blockCount = numBlocksBefore - numBlocksAfter;
        }
    
        /*
         * When adding a file fragment to the trashcan, copy over the original
         * file's inode and generation numbers -- this will be used in the
         * trash process to keep track of the block accounting in the original
         * inode.
         *
         * Dirty trick:  overload the numblocks field in the inode
         *               to store the original inum.
         */
        if (trash_inode != NULL) {
    	    if (newSize == 0) {
		        trash_inode->dinode.numblocks = 0;
	        } else {
		        trash_inode->dinode.numblocks = inode->inum;
		        trash_inode->dinode.genNum = inode->dinode.genNum;
	        }
            fs_inodeWrite(context, trash_inode, INO_WRITE_LOG);
            fs_trashcanAdd(context, trashcanInode, trash_inode);
        }
    
        /*
         * The trashcan inode may or may not have been touched 
         * so we need to ensure its added to the log here so it
         * will be unlocked during the commit.
         */
        fs_inodeLockTransaction(context, trashcanInode, context->logSummary);
    }

#if defined FS_DEBUG_TRASH_EMAP_LOOKUP

    eee_rtc(&beginTime);

    for (page = startPage; page < maxPage; page++)
    {
		fs_Block block;
		eeeRTC_t delta;

        block = fs_emap_lookup_block(context, inode, page);
        if (fs_blockIsValid(block))
        {
            //printf("found extent after truncation page %lld newSize %lld \n",
             //   page, newSize);
            fs_inodeDump("truncation", inode);
            FS_ASSERT(FALSE);
        }

		eee_rtcDuration(&beginTime, &delta);
        if (eee_rtcCompare(&delta, &fs_incoreTune->maxCpuUsageRTC) > 0)
        {
            fs_sleep(context, 0);
            
            eee_rtc(&beginTime);
        }
    }
#endif

    /*
     * Free the quota resources held by these blocks that we truncated
     */
    if (blockCount > 0 && !FS_I_IS_QUOT(inode->dinode.attrs))
    {
        (void)fs_freeQuotaResources(context, context->fileSystem,
                                    inode, blockCount, FS_COUNT_BLOCKS,
                                    inode->sd->u.nfs.uid, inode->sd->u.nfs.gid,
                                    &list);
        if (list)   {
    		if (rlist != NULL)	{
    			*rlist = list;
    		} else {
    	        fs_commitContextList(context, context->fileSystem, list, FALSE);
    	        fs_freeQuotaContextList(context, context->fileSystem, list);
    		}
        }
    }

} /* fs_truncateFile */


/**
 * Return TRUE if name is '.' or '..'
 */
boolean
fs_nameIsDotOrDotDot(
    char    *name)
{
    if (name[0] == '.')
    {
        if ((name[1] == '.' && name[2] == '\0') ||
            name[1] == '\0')
        {
            return TRUE;
        }
    }

    return FALSE;

} /* fs_nameIsDotOrDotDot */


/**
 * Return NFX_OK is the name passed in is valid for
 * creation/removal, else return the error appropriate
 * for the violation.
 */
int32
fs_nameIsValid(
    char    *name,
    int32   length)
{
    if (length <= 0)
        return FS_ERR_ACCESS;

    if (length != strlen(name))
        return FS_ERR_ACCESS;

    if (length > FS_NAME_MAX)
        return FS_ERR_NAMETOOLONG;

    if (fs_nameIsDotOrDotDot(name))
        return FS_ERR_ACCESS;

    if (!strcmp(name, ".snapshot"))
        return FS_ERR_PERM;

    if (!strcmp(name, "lost+found"))
        return FS_ERR_PERM;

    if (strchr(name, '/'))
        return FS_ERR_ACCESS;

    return NFX_OK;

} /* fs_nameIsValid */


/**
 * Return TRUE if the evm_volId_t's match in the
 * two handles.
 */
boolean
fs_handleVolMatch(
    fs_Handle   handle1,
    fs_Handle   handle2)
{
    return (handle1.volId.whole_thing == handle2.volId.whole_thing);

} /* fs_volIdsMatch */


/**
 * Returns TRUE if the current requestor is in
 * the the specified group.
 */
boolean
fs_inThisGroup(
    const fs_MsgReqHdr    *reqHdr,
    uint32          group)
{
    int32   i;

    if (reqHdr->gid == group)
        return TRUE;

    for (i = 0; i < reqHdr->numGids; i++)
    {
        if (reqHdr->gids[i] == group)
        {
            return TRUE;
        }
    }

    return FALSE;
}


/**
 * Return TRUE/FALSE for user/group ownership of the inode.
 */
void
fs_getOwnership(
    const fs_MsgReqHdr    *reqHdr,
    uint32          fileUid,
    uint32          fileGid,
    boolean         *user,
    boolean         *group,
    boolean         *isSuperUser)
{
    if (reqHdr->uid == 0)
    {
        // Root owns it all.
        *isSuperUser = *user = *group = TRUE;
    }
    else
    {
        *isSuperUser = *user = *group = FALSE;
        if (reqHdr->uid == fileUid) 
        {
            *user = TRUE;
        }
        else if (fs_inThisGroup(reqHdr, fileGid))
        {
            *group = TRUE;
        }
    }

} /* fs_getOwnership */


boolean
fs_doPermissionOkay(
    const fs_MsgReqHdr    *reqHdr,
    uint64          inum,
    int32           snapnum,
    uint16          mode,
    uint32          fileUid,
    uint32          fileGid,
    int32           type)
{
    boolean     isOwner, isGroupOwner, isSuperUser;

    fs_getOwnership(reqHdr, fileUid, fileGid, &isOwner, &isGroupOwner, &isSuperUser);

    if (inum == FS_NULL_INODE)
            return FALSE;

    switch(type)
    {
    case FS_PERM_LOOKUP:    // Checks dir perms
        if (isSuperUser) {
            return TRUE;
        }
        if (isOwner &&
                (mode & FS_IMODE_XUSR))
            return TRUE;
        if (isGroupOwner &&
                (mode & FS_IMODE_XGRP))
            return TRUE;
        if (!isOwner && !isGroupOwner &&
                (mode & FS_IMODE_XOTH))
            return TRUE;
        break;
    case FS_PERM_READ:      // Checks file perms
        if (isSuperUser) {
            return TRUE;
        }
        /*
         * Allow read for either the R or X bit, per
         * 8.7.5 of the NFS book.
        if (isOwner &&
                ((inode->sd->u.nfs.mode & FS_IMODE_RUSR) ||
                 (inode->sd->u.nfs.mode & FS_IMODE_XUSR)))
            return TRUE;
		if (isGroupOwner &&
                ((inode->sd->u.nfs.mode & FS_IMODE_RGRP) ||
                 (inode->sd->u.nfs.mode & FS_IMODE_XGRP)))
            return TRUE;
		if (!isOwner && !isGroupOwner &&
            ((inode->sd->u.nfs.mode & FS_IMODE_ROTH) ||
             (inode->sd->u.nfs.mode & FS_IMODE_XOTH)))
            return TRUE;
         */
        
        /* We don't follow the NFS book but NetApp, Solaris, Linux... */
        if (isOwner && (mode & FS_IMODE_RUSR))
            return TRUE;
		if (isGroupOwner && (mode & FS_IMODE_RGRP))
            return TRUE;
		if (!isOwner && !isGroupOwner && 
            (mode & FS_IMODE_ROTH))
            return TRUE;
        break;
    case FS_PERM_SETATTR:   // check inode permissions
        if (snapnum || (inum == FS_SNAPSHOT_DIR_INODE))
            return FALSE;
        if (isSuperUser)
            return TRUE;
        if (isOwner) {
            return TRUE;
		}
        /*  CONTINUE WITH PERM_WRITE CHECK */
    case FS_PERM_WRITE:     // Checks file perms
        if (snapnum)
            return FALSE;
        if (isSuperUser) {
            return TRUE;
        }
        if (isOwner &&
            (mode & FS_IMODE_WUSR)) {
            return TRUE;
		}
		if (isGroupOwner &&
                (mode & FS_IMODE_WGRP)) {
            return TRUE;
		}
		if (!isOwner && !isGroupOwner &&
				 (mode & FS_IMODE_WOTH))
            return TRUE;
        break;
    case FS_PERM_DIRENT:    // Checks dir perms
        if (snapnum || (inum == FS_SNAPSHOT_DIR_INODE))
            return FALSE;
        if (isSuperUser) {
            return TRUE;
        }
        if (isOwner &&
                ((mode & (FS_IMODE_WUSR|FS_IMODE_XUSR)) ==
                         (FS_IMODE_WUSR|FS_IMODE_XUSR)))
            return TRUE;
		if (isGroupOwner &&
                ((mode & (FS_IMODE_WGRP|FS_IMODE_XGRP)) ==
                         (FS_IMODE_WGRP|FS_IMODE_XGRP)))
            return TRUE;
		if (!isOwner && !isGroupOwner &&
            ((mode & (FS_IMODE_WOTH|FS_IMODE_XOTH)) ==
                     (FS_IMODE_WOTH|FS_IMODE_XOTH)))
            return TRUE;
        break;
    case FS_PERM_READDIR:    // Checks dir perms
        if (isSuperUser) {
            return TRUE;
        }
        if (isOwner &&
                ((mode & (FS_IMODE_RUSR|FS_IMODE_XUSR)) ==
                         (FS_IMODE_RUSR|FS_IMODE_XUSR)))
            return TRUE;
		if (isGroupOwner &&
                ((mode & (FS_IMODE_RGRP|FS_IMODE_XGRP)) ==
                         (FS_IMODE_RGRP|FS_IMODE_XGRP)))
            return TRUE;
		if (!isOwner && !isGroupOwner &&
            ((mode & (FS_IMODE_ROTH|FS_IMODE_XOTH)) ==
                     (FS_IMODE_ROTH|FS_IMODE_XOTH)))
            return TRUE;
        break;
    case FS_PERM_EXEC:      // Checks file perms
        if (isSuperUser) {
            return TRUE;
        }
        if (isOwner &&
                (mode & FS_IMODE_XUSR))
            return TRUE;
		if (isGroupOwner &&
                (mode & FS_IMODE_XGRP))
            return TRUE;
		if (!isOwner && !isGroupOwner &&
            (mode & FS_IMODE_XOTH))
            return TRUE;
        break;
    default:
        fs_error(NULL, FS_INTERNAL_ERR, "permissionOkay");
        break;
    }
    return FALSE;
}


/**
 * Return TRUE if user has permission to perform the requested action.
 */
boolean
fs_permissionOkay(
    fs_Context  *context,
    fs_Inode    *inode,             /* file or dir for operation */
    int32       type)               /* FS_PERM_LOOKUP, FS_PERM_READ, ... */
{
    return fs_doPermissionOkay(context->reqHdr, inode->inum, inode->snapnum,
                               inode->sd->u.nfs.mode, inode->sd->u.nfs.uid,
                               inode->sd->u.nfs.gid, type);
} /* fs_permissionOkay */


/**
 * Return the fsid for the given filesystem,snapnum.
 */
static inline uint32
fs_fsIdFromVolSnap(
    evm_volId_t volId,
    int32       snapnum)
{
    volId.details.u.parts.reserved = snapnum;

    return volId.details.u.id;
}


/**
 * Copy the attributes from inode into attr.
 */
void
fs_doGetAttr(
    evm_volId_t     volId,
    fs_Attr         *attr,
    fs_Inode        *inode,
    fs_SnapType     snapType,   // FS_SNAPSHOT_DIR_TYPE or FS_SNAPSHOT_BOUNDARY_TYPE
    uint64          snapInfo,   // snapnumInum if FS_SNAPSHOT_BOUNDARY_TYPE
                                // parentInum if FS_SNAPSHOT_DIR_TYPE
    int32           protocol,
    fs_AttrType     attrType)
{
    fs_Dinode *dinode = &inode->dinode;

    attr->attrType = attrType;

    attr->mode = (dinode->attrs & (FS_IFMT_MASK | FS_IATR_MASK)) |
                      inode->sd->u.nfs.mode;

    if (protocol == FS_PROTO_NFS_V3)
    {
        // For NFSv3, don't include the format bits.
        attr->mode &= ~FS_IFMT_MASK;
    }
    attr->numLinks = dinode->linkCnt;
    attr->uid = inode->sd->u.nfs.uid;
    attr->gid = inode->sd->u.nfs.gid;
    attr->size = dinode->size;
    attr->blockSize = FS_BLOCK_SIZE;

    if (protocol == FS_PROTO_NFS_V2)
    {
        // For NFSv2 it seems that we have to assume a fixed
        // blocksize of 512 bytes.  Used netapp as the reference.
        attr->blocks = dinode->numblocks * FS_BLOCK_SIZE / 512;
    }
    else
    {
        attr->blocks = dinode->numblocks;
    }
    
    if (snapType == FS_SNAPSHOT_BOUNDARY_TYPE)
        attr->fsId = inode->inum;
    else if (snapType == FS_SNAPSHOT_DIR_TYPE)
        attr->fsId = snapInfo;
    else
        attr->fsId = fs_fsIdFromVolSnap(volId, inode->snapnum);

    if (snapType == FS_SNAPSHOT_BOUNDARY_TYPE)
        attr->inode = snapInfo;
    else if (snapType == FS_SNAPSHOT_DIR_TYPE)
        attr->inode = FS_SNAPSHOT_DIR_INODE;
    else
        attr->inode = inode->inum;

    attr->atime = inode->dinode.atime;
    attr->mtime = inode->dinode.mtime;
    attr->ctime = inode->dinode.ctime;

    attr->devNum.major = 0;
    attr->devNum.minor = 0;

    switch(dinode->attrs & FS_IFMT_MASK)
    {
    case FS_IFMT_SOCK:
        attr->type = FS_FTYPE_SOCK;
        break;
    case FS_IFMT_SYML:
        attr->type = FS_FTYPE_LNK;
        break;
    case FS_IFMT_REG:
        attr->type = FS_FTYPE_REG;
        break;
    case FS_IFMT_BLK:
        attr->type = FS_FTYPE_BLK;
        attr->devNum.major = dinode->u.devNum.major;
        attr->devNum.minor = dinode->u.devNum.minor;
        if (protocol == FS_PROTO_NFS_V2)
        {
            // In NFSv2, encode the rdev field also.
            attr->rdev = dinode->u.devNum.minor & 0xff;
            attr->rdev |= (dinode->u.devNum.major & 0xff) << 8;
        }
        break;
    case FS_IFMT_DIR:
        attr->type = FS_FTYPE_DIR;
        break;
    case FS_IFMT_CHR:
        attr->type = FS_FTYPE_CHR;
        attr->devNum.major = dinode->u.devNum.major;
        attr->devNum.minor = dinode->u.devNum.minor;
        if (protocol == FS_PROTO_NFS_V2)
        {
            // In NFSv2, encode the rdev field also.
            attr->rdev = dinode->u.devNum.minor & 0xff;
            attr->rdev |= (dinode->u.devNum.major & 0xff) << 8;
        }

        // Is this really a V2 client created FIFO?
        if (dinode->u.devNum.major == -1 &&
            dinode->u.devNum.minor == -1)
        {
            attr->devNum.major = 0;
            attr->devNum.minor = 0;
            if (protocol == FS_PROTO_NFS_V2)
            {
                attr->rdev = -1;
            }
            if (protocol == FS_PROTO_NFS_V3)
            {
                attr->type = FS_FTYPE_FIFO;
            }
        }
        break;
    case FS_IFMT_FIFO:
        attr->type = FS_FTYPE_FIFO;
        if (protocol == FS_PROTO_NFS_V2)
        {
            // For V2 clients, things need to look a little
            // differently to get these right.
            attr->type = FS_FTYPE_CHR;
            attr->devNum.major = 0;
            attr->devNum.minor = 0;
            attr->rdev = -1;
            attr->mode &= ~FS_IFMT_MASK;
            attr->mode |= FS_IFMT_CHR;
        }
        break;
    default:
        attr->type = FS_FTYPE_NON;
        break;
    }

} /* fs_doGetAttr */


/**
 * Get attributes if this type is requested and update dest
 * ptr/count if needed.
 */
void
fs_getRespAttrs(
    fs_Context  *context,
    fs_Inode    *inode,
    uint64      parentInum, // used if inode->inum == FS_SNAPSHOT_DIR_INODE
    fs_AttrType attrType,
    fs_Attr     **attrpp,
    int32       *numAttrp,
    boolean     incrRespSize)
{
    fs_SnapType snapType = FS_NO_SNAPSHOT_TYPE;

    if (inode->inum == FS_SNAPSHOT_DIR_INODE)
    {
        snapType = FS_SNAPSHOT_DIR_TYPE;
    }

    if (context->reqHdr->attrType & attrType)
    {
        context->respHdr->attrType |= attrType;
        fs_doGetAttr(context->volId, *attrpp, inode, snapType, parentInum,
                     context->reqHdr->proto, attrType);
        if (incrRespSize)
        {
            context->respSize += sizeof(fs_Attr);
            *numAttrp += 1;
            *attrpp += 1;
        }
    }

} /* fs_getRespAttrs */


/**
 * Get attributes depending on the snap info and update dest
 * ptr/count if needed.
 */
void
fs_getSnapRespAttrs(
    fs_Context  *context,
    fs_Inode    *inode,
    fs_SnapType snapType,
    uint64      snapInfo,
    fs_AttrType attrType,
    fs_Attr     **attrpp,
    int32       *numAttrp,
    boolean     incrRespSize)
{
    if (context->reqHdr->attrType & attrType)
    {
        context->respHdr->attrType |= attrType;
        fs_doGetAttr(context->volId, *attrpp, inode, snapType, snapInfo,
                     context->reqHdr->proto, attrType);
        if (incrRespSize)
        {
            context->respSize += sizeof(fs_Attr);
            *numAttrp += 1;
            *attrpp += 1;
        }
    }

} /* fs_getSnapRespAttrs */


/**
 * Set attributes in the inode.  Calls truncation function if
 * necessary.
 */
void
fs_doSetAttr(
    fs_Context  *context,
    fs_Attr     *attr,
    fs_Inode    *inode,
	fs_QuotaContextList_t	**rlist)
{
    fs_SecurityDescNfs  sdNew;
    fs_Dinode           *dinode = &inode->dinode;
    fs_Time             now = fs_getTimeOfDay();
    boolean             doSdSet = FALSE;

    if (attr->mode != -1)
    {
        // Set mode/attribute as requested.
        dinode->attrs &= ~FS_IATR_MASK;
        dinode->attrs |= (attr->mode & FS_IATR_MASK);

        if ((attr->mode & FS_IMODE_MASK) != inode->sd->u.nfs.mode)
        {
            sdNew = inode->sd->u.nfs;
            sdNew.mode = (attr->mode & FS_IMODE_MASK);
            doSdSet = TRUE;
        }
    }
    if (attr->uid != -1)
    {
        if (attr->uid != inode->sd->u.nfs.uid)
        {
            if (!doSdSet)
            {
                sdNew = inode->sd->u.nfs;
            }
            sdNew.uid = attr->uid;
            doSdSet = TRUE;
        }
    }
    if (attr->gid != -1)
    {
        if (attr->gid != inode->sd->u.nfs.gid)
        {
            if (!doSdSet)
            {
                sdNew = inode->sd->u.nfs;
            }
            sdNew.gid = attr->gid;
            doSdSet = TRUE;
        }
    }

    if (doSdSet)
    {
        fs_securityDescSet(context, inode, &sdNew.hdr);
    }

    if (FS_I_IS_REG(dinode->attrs) && attr->size != -1)
    {
        if (attr->size < dinode->size)
        {
            fs_truncateFile(context, inode, attr->size, FALSE, rlist);
        }
        dinode->size = attr->size;
    }
    if (FS_I_IS_BLK(dinode->attrs) || FS_I_IS_CHR(dinode->attrs))
    {
        if (context->reqHdr->proto == FS_PROTO_NFS_V2)
        {
            // In NFSv2, the size field carries the major/minor
            // numbers.
            dinode->u.devNum.major = (attr->size >> 8) & 0xff;
            dinode->u.devNum.minor = attr->size & 0xff;
    
            // If the size field is -1 (ie. invalid), and 
            // the file type desired is a character special dev,
            // they really want a fifo.  This is because the create
            // is overloaded in nfs v2.  So we encode all one's, 
            // return in the rdev field so the client can
            // figure it out.  Remember that size field is only
            // 32 bit wide in nfs v2. 

            if (FS_I_IS_CHR(dinode->attrs) && ((int32)attr->size) == -1)
            {
                dinode->attrs &= ~FS_IFMT_CHR;
                dinode->attrs |= FS_IFMT_FIFO;
                dinode->u.devNum.major = -1;
                dinode->u.devNum.minor = -1;
            }
        }
    }

    // Update atime
    if (attr->atimeType == FS_ATTR_TIME_SET_CLIENT)
    {
        dinode->atime = attr->atime;
    }
    else if (attr->atimeType == FS_ATTR_TIME_SET_SRVR)
    {
        dinode->atime = now;
    }

    // Update mtime
    if (attr->mtimeType == FS_ATTR_TIME_SET_CLIENT)
    {
        dinode->mtime = attr->mtime;
    }
    else if (attr->mtimeType == FS_ATTR_TIME_SET_SRVR)
    {
        dinode->mtime = now;
    }

    // Update ctime
    dinode->ctime = now;
    dinode->version++;

} /* fs_doSetAttr */


/*-----------------------------------------------------------------------------
 * NFS request handlers.
 */


/**
 * Handle the NULL request.
 */
int32
fs_nullThread(
    fs_Context      *context)
{
    context->respSize = sizeof(fs_MsgNullResp);

    return NFX_OK;

} /* fs_nullThread */


/**
 * Handle the GET_ATTR request.
 */
int32
fs_getAttrThread(
    fs_Context      *context)
{
    fs_MsgGetAttrResp   *resp;
    fs_MsgGetAttrReq    *req;
    fs_Inode            *inode;
    fs_SnapType         snapType = FS_NO_SNAPSHOT_TYPE;
    uint64              snapInfo = 0;

    resp = (fs_MsgGetAttrResp *) context->respHdr;
    req = (fs_MsgGetAttrReq *) context->reqHdr;
    context->respSize = sizeof(fs_MsgGetAttrResp);

    FS_TRACE(context, FS_TRACE_REQ,
            ("%s xid 0x%x getattr snap %lld inum %lld gen %d",
            fs_protoStr[context->reqHdr->proto],
            context->reqHdr->clientId,
            FS_HANDLE_GET_SNAP(req->fileHandle.inum),
            FS_HANDLE_GET_INUM(req->fileHandle.inum),
            req->fileHandle.gen));

    if (fs_handleIsSnapMagic(&req->fileHandle))
    {
        inode = I_GET_SHARED(0, FS_SNAPSHOT_DIR_INODE);
    }
    else
    {
        inode = I_GET_SHARED(FS_HANDLE_GET_SNAP(req->fileHandle.inum),
                             FS_HANDLE_GET_INUM(req->fileHandle.inum));
    }
    if (inode == NULL)
    {
        return FS_ERR_STALE;
    }
    if (!fs_handleIsValid(context, &req->fileHandle, inode))
    {
        fs_inodeRelease(context, inode);
        return FS_ERR_STALE;
    }

    resp->hdr.attrType = FS_POST_ATTRS;

    if (inode->inum == FS_SNAPSHOT_DIR_INODE)
    {
        snapType = FS_SNAPSHOT_DIR_TYPE;
        snapInfo = FS_HANDLE_GET_INUM(req->fileHandle.inum);
    }
    else if (FS_HANDLE_IS_SNAP_BOUNDARY(req->fileHandle))
    {
        snapType = FS_SNAPSHOT_BOUNDARY_TYPE;
        snapInfo = FS_HANDLE_GET_SNAP(req->fileHandle.inum) + FS_SNAPSHOT_ID_INODE_FIRST;
    }

    fs_doGetAttr(context->volId, &(resp->attr), inode,
                 snapType, snapInfo,
                 context->reqHdr->proto, FS_POST_ATTRS);

    fs_inodeRelease(context, inode);

    return NFX_OK;

} /* fs_getAttrThread */


/**
 * Handle the SET_ATTR request.
 */
int32
fs_setAttrThread(
    fs_Context      *context)
{
    fs_MsgSetAttrResp   *resp;
    fs_MsgSetAttrReq    *req;
    fs_Inode            *inode;
    fs_Buf              *buf;
    int32               rc;
    fs_QuotaContextList_t   *list1 = NULL, *list2 = NULL;
    fs_QuotaContextList_t   *list3 = NULL, *list4 = NULL;
    fs_QuotaContextList_t   *list5 = NULL, *list6 = NULL;
    fs_QuotaContextList_t   *list7 = NULL, *list8 = NULL;
    fs_QuotaContextList_t   *list9 = NULL;
    fs_Attr             *reqAttrp, *respAttrp;
    uint64              quotaReserved1 = 0, quotaReserved2 = 0;
    uint64              quotaReserved3 = 0, quotaReserved4 = 0;
    uint64              blocks = 0;
    uint64              beforeSize;

    resp = (fs_MsgSetAttrResp *) context->respHdr;
    req = (fs_MsgSetAttrReq *) context->reqHdr;
    context->respSize = sizeof(fs_MsgSetAttrResp);
    reqAttrp = &req->attr;
    respAttrp = &resp->attr[0];
    resp->numAttrs = 0;

    if (context->fileSystem->state == FS_VOL_STATE_ONLINE_RDONLY) {
        return FS_ERR_ROFS;
    }

    FS_TRACE(context, FS_TRACE_REQ,
            ("%s xid 0x%x setattr snap %lld inum %lld gen %d",
            fs_protoStr[context->reqHdr->proto],
            context->reqHdr->clientId,
            FS_HANDLE_GET_SNAP(req->fileHandle.inum),
            FS_HANDLE_GET_INUM(req->fileHandle.inum),
            req->fileHandle.gen));

    if (fs_handleIsSnapMagic(&req->fileHandle))
    {
        return FS_ERR_ACCESS;
    }
    else
    {
        inode = I_GET_EXCL( FS_HANDLE_GET_SNAP(req->fileHandle.inum),
                            FS_HANDLE_GET_INUM(req->fileHandle.inum));
    }
    if (inode == NULL)
    {
        return FS_ERR_STALE;
    }
    if (!fs_handleIsValid(context, &req->fileHandle, inode))
    {
        fs_inodeReleaseExclusive(context, inode);
        return FS_ERR_STALE;
    }

    fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->fileHandle.inum),
                    FS_PRE_ATTRS, &respAttrp, &resp->numAttrs, TRUE);

    if (!fs_permissionOkay(context, inode, FS_PERM_SETATTR))
    {
        fs_inodeReleaseExclusive(context, inode);
        return FS_ERR_ACCESS;
    }

    if (reqAttrp->size != -1 &&
        !fs_permissionOkay(context, inode, FS_PERM_WRITE)) {
        fs_inodeReleaseExclusive(context, inode);
        return FS_ERR_ACCESS;
    }

    if (reqAttrp->uid != -1 &&
        inode->sd->u.nfs.uid != reqAttrp->uid)
    {
        // Only allow root user to change
        // owner of the file.
        if (req->hdr.uid != 0)
        {
            //printf("disallowing uid change\n");
            fs_inodeReleaseExclusive(context, inode);
            return FS_ERR_PERM;
        }
    }
    if (reqAttrp->gid != -1 &&
        inode->sd->u.nfs.gid != reqAttrp->gid)
    {
        boolean rootPriv = (req->hdr.uid == 0 ||
                            fs_inThisGroup(context->reqHdr, 0));

        // Only allow root-priviliged user/group to
        // change group arbitrarily.  Non-root priviliged
        // user can only change group according to its
        // membership.
        if ((!rootPriv &&
             req->hdr.uid != inode->sd->u.nfs.uid) ||
            (!rootPriv && !fs_inThisGroup(context->reqHdr, reqAttrp->gid)))
        {
            //printf("disallowing gid change\n");
            fs_inodeReleaseExclusive(context, inode);
            return FS_ERR_PERM;
        }
    }

    // Todo Verify that check only secs
    if (req->checkCtime.secs &&
        inode->dinode.ctime.secs != req->checkCtime.secs)
    {
        fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->fileHandle.inum),
                        FS_POST_ATTRS, &respAttrp, &resp->numAttrs, TRUE);
        fs_inodeReleaseExclusive(context, inode);
        return FS_ERR_NOT_SYNC;
    }

    /*
     * Determine if we are changing ownership / group of the file
     * we have enough quota for the target. Only root can allow
     * chown of the file, so we don't have to worry about quota failure
     * for chown.
     */

    blocks = inode->dinode.numblocks;

    if (reqAttrp->uid != -1 && reqAttrp->uid != inode->sd->u.nfs.uid) {

        /*
         * Reserve blocks
         */

        if (blocks > 0) {
            quotaReserved1 = fs_getQuotaReservation(context,
                                                    context->fileSystem,
                                                    inode, blocks,
                                                    FS_COUNT_BLOCKS,
                                                    reqAttrp->uid, 0, 1, 
                                                    &list1);
            if (quotaReserved1 != blocks) {
                //we only fail because of no disk space
                FS_ASSERT(quotaReserved1 == 0);
                if (list1 != NULL) {
                    fs_freeQuotaContextList(context, context->fileSystem,list1);
                }
                fs_getRespAttrs(context, inode,
                                FS_HANDLE_GET_INUM(req->fileHandle.inum),
                                FS_POST_ATTRS, &respAttrp, &resp->numAttrs, TRUE);
                fs_inodeReleaseExclusive(context, inode);
                return FS_ERR_NOSPC;
            }
        }

        /*
         * Reserve inodes
         */

        quotaReserved2 = fs_getQuotaReservation(context, context->fileSystem,
                                                inode, 1, FS_COUNT_INODES,
                                                reqAttrp->uid, 0, 1, &list2);
        if (quotaReserved2 != 1) {
            //we only fail because of no disk space
            FS_ASSERT(quotaReserved2 == 0);
            if (list1 != NULL) {
                fs_trimQuotaReservation(context, context->fileSystem, list1,
                                        quotaReserved1);
                fs_freeQuotaContextList(context, context->fileSystem, list1);
            }
            if (list2 != NULL) {
                fs_freeQuotaContextList(context, context->fileSystem, list2);
            }
            fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->fileHandle.inum),
                            FS_POST_ATTRS, &respAttrp, &resp->numAttrs, TRUE);
            fs_inodeReleaseExclusive(context, inode);
            return FS_ERR_NOSPC;
        }
    }
    if (reqAttrp->gid != -1 && reqAttrp->gid != inode->sd->u.nfs.gid) {

        /*
         * Reserve blocks
         */

        if (blocks > 0) {
            quotaReserved3 = fs_getQuotaReservation(context, 
                                                    context->fileSystem,
                                                    inode, blocks,
                                                    FS_COUNT_BLOCKS,
                                                    0, reqAttrp->gid, 1, 
                                                    &list3);
            if (quotaReserved3 != blocks) {
                //we only fail because of no disk space
                FS_ASSERT(quotaReserved3 == 0);
                if (list1 != NULL) {
                    fs_trimQuotaReservation(context, context->fileSystem,
                                            list1, quotaReserved1);
                    fs_freeQuotaContextList(context, context->fileSystem,
                                            list1);
                }
                if (list2 != NULL) {
                    fs_trimQuotaReservation(context, context->fileSystem,
                                            list2, quotaReserved2);
                    fs_freeQuotaContextList(context, context->fileSystem,
                                            list2);
                }
                if (list3 != NULL) {
                    fs_freeQuotaContextList(context, context->fileSystem, 
                                            list3);
                }
                fs_getRespAttrs(context, inode,
                                FS_HANDLE_GET_INUM(req->fileHandle.inum),
                                FS_POST_ATTRS, &respAttrp, &resp->numAttrs, TRUE);
                fs_inodeReleaseExclusive(context, inode);
                return FS_ERR_NOSPC;
            }
        }

        /*
         * Reserve inodes
         */

        quotaReserved4 = fs_getQuotaReservation(context, context->fileSystem,
                                           inode, 1, FS_COUNT_INODES,
                                           0, reqAttrp->gid,  1, &list4);
        if (quotaReserved4 != 1) {
            //we only fail because of no disk space
            FS_ASSERT(quotaReserved4 == 0);
            if (list1 != NULL) {
                fs_trimQuotaReservation(context, context->fileSystem, list1, 
                                        quotaReserved1);
                fs_freeQuotaContextList(context, context->fileSystem, list1);
            }
            if (list2 != NULL) {
                fs_trimQuotaReservation(context, context->fileSystem, list2,
                                        quotaReserved2);
                fs_freeQuotaContextList(context, context->fileSystem, list2);
            }
            if (list3 != NULL) {
                fs_trimQuotaReservation(context, context->fileSystem, list3,
                                        quotaReserved3);
                fs_freeQuotaContextList(context, context->fileSystem, list3);
            }
            if (list4 != NULL) {
                fs_freeQuotaContextList(context, context->fileSystem, list4);
            }
            fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->fileHandle.inum),
                            FS_POST_ATTRS, &respAttrp, &resp->numAttrs, TRUE);
            fs_inodeReleaseExclusive(context, inode);
            return FS_ERR_NOSPC;
        }
    }

    /*
     * Now free quota consumption from original owner/group and root
     * to whom we allocated intermediately
     */

    if (list1 != NULL) {
        fs_freeQuotaResources(context, context->fileSystem, inode,
                              blocks, FS_COUNT_BLOCKS,
                              inode->sd->u.nfs.uid, 0, 
                              &list5);
    }
    if (list2 != NULL) {
        fs_freeQuotaResources(context, context->fileSystem, inode,
                              1, FS_COUNT_INODES,
                              inode->sd->u.nfs.uid, 0, 
                              &list6);
    }
    if (list3 != NULL) {
        fs_freeQuotaResources(context, context->fileSystem, inode,
                              blocks, FS_COUNT_BLOCKS,
                              0, inode->sd->u.nfs.gid,
                              &list7);
    }
    if (list4 != NULL) {
        fs_freeQuotaResources(context, context->fileSystem, inode,
                              1, FS_COUNT_INODES,
                              0, inode->sd->u.nfs.gid, 
                              &list8);
    }


    // Update ctime.
    inode->dinode.ctime = fs_getTimeOfDay();
    
    /*
     * !!!!NOTICE!!!
     *
     * Don't update mtime as it breaks rsync and other
     * applications that rely on mtime not being updated.
     */

    beforeSize = inode->dinode.size;

    fs_doSetAttr(context, reqAttrp, inode, &list9);

     /*
     * Now we have quota reservation and release for all chown/chgrp, commit the
     * quotas
     */

    if (list1 != NULL) {
        fs_commitContextList(context, context->fileSystem, list1, TRUE);
        fs_freeQuotaContextList(context, context->fileSystem, list1);
    }
    if (list2 != NULL) {
        list2->newInode = inode;
        fs_commitContextList(context, context->fileSystem, list2, TRUE);
        fs_freeQuotaContextList(context, context->fileSystem, list2);
    }
    if (list3 != NULL) {
        fs_commitContextList(context, context->fileSystem, list3, TRUE);
        fs_freeQuotaContextList(context, context->fileSystem, list3);
    }
    if (list4 != NULL) {
        list4->newInode = inode;
        fs_commitContextList(context, context->fileSystem, list4, TRUE);
        fs_freeQuotaContextList(context, context->fileSystem, list4);
    }
    if (list5 != NULL) {
        fs_commitContextList(context, context->fileSystem, list5, TRUE);
        fs_freeQuotaContextList(context, context->fileSystem, list5);
    }
    if (list6 != NULL) {
        list6->newInode = inode;
        fs_commitContextList(context, context->fileSystem, list6, TRUE);
        fs_freeQuotaContextList(context, context->fileSystem, list6);
    }
    if (list7 != NULL) {
        fs_commitContextList(context, context->fileSystem, list7, TRUE);
        fs_freeQuotaContextList(context, context->fileSystem, list7);
    }
    if (list8 != NULL) {
        list8->newInode = inode;
        fs_commitContextList(context, context->fileSystem, list8, FALSE);
        fs_freeQuotaContextList(context, context->fileSystem, list8);
    }
    if (list9 != NULL) {
        fs_commitContextList(context, context->fileSystem, list9, TRUE);
        fs_freeQuotaContextList(context, context->fileSystem, list9);
    }

    // If the file was extended, ensure the last page is
    // zeroed out if it now contains a hole.
    if (beforeSize < inode->dinode.size &&
        (beforeSize % FS_BLOCK_SIZE) != 0)
    {
        // - file already in use
        // - file not at block boundary
        // find previuous block and fill with all '0's

        //printf("setattr extend from %lld to %lld\n",
        //    beforeSize, inode->dinode.size);

        buf = O_USER_GET_EXCL (inode, beforeSize / FS_BLOCK_SIZE,
                                (beforeSize % FS_BLOCK_SIZE));
        // It is possible that buf does not exist
        // (set file size with set attr cmd)
        if (buf)
        {
            fs_cod_fill_null (&buf->u.user.codHandle,
                              FS_BLOCK_SIZE - (beforeSize % FS_BLOCK_SIZE),
                              beforeSize % FS_BLOCK_SIZE);

            // still same COD in buf->u.user.codHandle
            buf->u.user.codLength = FS_BLOCK_SIZE;

            FS_BUF_WRITE(buf, 0, FS_BLOCK_SIZE) ;
            fs_logWriteBuf(context, inode, buf, FS_WRITE_STABLE_FILE_SYNC);
        }
    }

    fs_inodeWrite(context, inode, INO_WRITE_LOG);

    fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->fileHandle.inum),
                    FS_POST_ATTRS, &respAttrp, &resp->numAttrs, TRUE);

    // Commit the changes.
    rc = fs_logCommitEntry(context);
    if (rc != NFX_OK)
    {
        rc = FS_ERR_IO;
    }

    return rc;

} /* fs_setAttrThread */


/**
 * Handle the LOOKUP request.
 */
int32
fs_lookupThread(
    fs_Context      *context)
{
    fs_MsgLookupResp    *resp = (fs_MsgLookupResp *) context->respHdr;
    fs_MsgLookupReq     *req = (fs_MsgLookupReq *) context->reqHdr;
    fs_Inode            *inode, *parentInode;
    fs_Attr             *attrp;
    uint64              inum;
    int32               rc;
    uint64              parentInum;
    uint32              parentSnap;
    fs_SnapType         snapType = FS_NO_SNAPSHOT_TYPE;
    uint64              snapInfo = 0;

    if (fs_handleIsSnapMagic(&req->parentHandle))
    {
        parentInum = FS_SNAPSHOT_DIR_INODE;
        parentSnap = 0;
    }
    else
    {
        parentInum = FS_HANDLE_GET_INUM(req->parentHandle.inum);
        parentSnap = FS_HANDLE_GET_SNAP(req->parentHandle.inum);
    }

    context->respSize = sizeof(fs_MsgLookupResp);
    resp->numAttrs = 0;
    attrp = &(resp->attr[0]);
    parentInode = inode = NULL;

    FS_TRACE(context, FS_TRACE_REQ,
             ("%s xid 0x%x lookup snap %d inum %lld gen %d name %s",
              fs_protoStr[context->reqHdr->proto],
              context->reqHdr->clientId,
              parentSnap,
              parentInum,
              req->parentHandle.gen,
              req->fileName));

    if (req->nameLen == 0 || strlen(req->fileName) == 0)
    {
        return FS_ERR_EXIST;
    }

    parentInode = I_GET_SHARED(parentSnap, parentInum);
    if (parentInode == NULL)
    {
        return FS_ERR_STALE;
    }
    if (!fs_handleIsValid(context, &req->parentHandle, parentInode))
    {
        fs_inodeRelease(context, parentInode);
        return FS_ERR_STALE;
    }
    if (!fs_permissionOkay(context, parentInode, FS_PERM_LOOKUP))
    {
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_inodeRelease(context, parentInode);
        return FS_ERR_ACCESS;
    }

    // ".snapshot" dirent is generated, not stored.
    if (parentSnap == 0 &&
        parentInum != FS_SNAPSHOT_DIR_INODE &&       
        !strcmp(req->fileName, ".snapshot"))
    {
        inum = FS_SNAPSHOT_DIR_INODE;
    }
    else
    {
        inum = fs_dir_lookupent(context,
                                parentInode,
                                req->fileName,
                                req->nameLen);
    }

    if (inum == 0)
    {
        // File not found.
        rc = FS_ERR_NOENT;

        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
    }
    else
    {
        uint64 effectiveInum;
        int32 snapNum;

        // File found.
		FS_LOCK_ORDER_SET_PARENT(parentInode->inum);

        // Determine the snapshot number of the inode we're
        // looking for.  If the parent directory is a .snapshot
        // directory, then the snapnum is the inum (minus offset),
        // otherwise the snapnum is the same as the parent dir's
        // snapnum.

        // Determine the effective inum that we're looking for.

        if (parentInum == FS_SNAPSHOT_DIR_INODE)
        {
            if (!strcmp(req->fileName, ".."))
            {
                // ".." entry for a ".snapshot" directory should
                // point to its actual parent (which we've strategically
                // placed in the inum field of the handle given out
                // for the ".snapshot" directory).

                snapNum = 0;
                effectiveInum = FS_HANDLE_GET_INUM(req->parentHandle.inum);
            }
            else if (!strcmp(req->fileName, "."))
            {
                snapNum = 0;
                effectiveInum = parentInum;
            }
            else
            {
                snapNum = inum - FS_SNAPSHOT_ID_INODE_FIRST;
                effectiveInum = FS_HANDLE_GET_INUM(req->parentHandle.inum);
            }
        }
        else
        {
            snapNum = parentSnap;
            effectiveInum = inum;
        }

        // If looking up "..", release the parent to avoid
        // deadlock.
        if (effectiveInum == fs_dirroot_get_parent(context, parentInode))
        {
            fs_inodeRelease(context, parentInode);
            parentInode = NULL;
        }

        inode = I_GET_SHARED(snapNum, effectiveInum);
        if (inode == NULL)
        {
            if (parentInum == FS_SNAPSHOT_DIR_INODE)
            {
                fs_Inode    *nullInode = I_GET_SHARED(snapNum, FS_NULL_INODE);

                // File doesn't exist for the indicated snapshot.
                // Return the NULL inode.

                fs_handleInit(context->fileSystem,
                              &resp->fileHandle,
                              snapNum,
                              FS_NULL_INODE,
                              0,
                              req->parentHandle.mntPnt);

                fs_getRespAttrs(context, nullInode,
                                parentInum, FS_POST_ATTRS,
                                &attrp, &resp->numAttrs, TRUE);
                fs_inodeRelease(context, nullInode);
                rc = NFX_OK;
            }
            else
            {
                rc = FS_ERR_NOENT;
            }
        }
        else
        {
            uint64      handleInum;
            uint32      genNum;

            //printf("lookupThread: snap %lld inum %lld\n",
            //       FS_HANDLE_GET_SNAP(req->parentHandle.inum), inum);

            if (effectiveInum == FS_SNAPSHOT_DIR_INODE)
            {
                // The inum in the handle needs to be the parent inum for the
                // .snapshot dir - essentially a link to the directory
                // contents for the snapshot.
                handleInum = FS_HANDLE_GET_INUM(req->parentHandle.inum);
                genNum = FS_INODE_GENNUM_SNAP_MAGIC;
                snapType = FS_SNAPSHOT_DIR_TYPE;
                snapInfo = parentInum;
            }
            else if (parentInum == FS_SNAPSHOT_DIR_INODE)
            {
                // The inum in the handle needs to be the parent inum for a
                // .snapshot dirent - essentially a link to the directory
                // contents for the snapshot.
                handleInum = FS_HANDLE_GET_INUM(req->parentHandle.inum);
                genNum = inode->dinode.genNum;
                // setup snapshot boundary object attributes
                // if looking up dotdot, it's just the default snapType
                if ((strcmp(req->fileName, "..") != 0) &&
                    (FS_HANDLE_GET_INUM(req->parentHandle.inum) == req->parentHandle.mntPnt))
                {
                    snapType = FS_SNAPSHOT_BOUNDARY_TYPE;
                    snapInfo = inum;
                }
                // setup magic2 bit if looking up parent
                if (!strcmp(req->fileName, "..") &&
                    (FS_HANDLE_GET_INUM(req->parentHandle.inum) != req->parentHandle.mntPnt))
                {
                    genNum = FS_INODE_GENNUM_SNAP_MAGIC2;
                }
            }
            else if (FS_HANDLE_IS_SNAP_BOUNDARY(req->parentHandle) &&
                     !strcmp(req->fileName, ".."))
            {
                // traversing a snapshot boundary on the way up
                // parent should be the .snapshot directory
                handleInum = FS_HANDLE_GET_INUM(req->parentHandle.inum);
                genNum = FS_INODE_GENNUM_SNAP_MAGIC;
                snapNum = 0;
                snapType = FS_SNAPSHOT_DIR_TYPE;
                snapInfo = FS_HANDLE_GET_INUM(req->parentHandle.inum);
            }
            else
            {
                handleInum = effectiveInum;
                genNum = inode->dinode.genNum;
                // check if the effectiveInum will cross the snap boundary
                // ie cd dotdot from /mnt/d0/.snapshot/snap -> /mnt/.snapshot/snap
                if ((snapNum != 0) &&
                    (effectiveInum == req->parentHandle.mntPnt) &&
                    !strcmp(req->fileName, ".."))
                {
                    snapType = FS_SNAPSHOT_BOUNDARY_TYPE;
                    snapInfo = snapNum + FS_SNAPSHOT_ID_INODE_FIRST;
                }
            }

            fs_handleInit(context->fileSystem,
                          &resp->fileHandle,
                          snapNum,
                          handleInum,
                          genNum,
                          req->parentHandle.mntPnt);

            fs_getSnapRespAttrs(context, inode, snapType, snapInfo,
                                FS_POST_ATTRS, &attrp,
                                &resp->numAttrs, TRUE);

            fs_inodeRelease(context, inode);
            inode = NULL;
            rc = NFX_OK;
        }

        if (parentInode == NULL)
        {
            // Re-acquire the parent to get POST attrs.
            parentInode = I_GET_SHARED(parentSnap, parentInum);
            if (parentInode == NULL)
            {
                return FS_ERR_STALE;
            }
            if (!fs_handleIsValid(context, &req->parentHandle, parentInode))
            {
                fs_inodeRelease(context, parentInode);
                return FS_ERR_STALE;
            }
        }

        // set the correct attributes for the parent as well
        snapType = FS_NO_SNAPSHOT_TYPE;
        snapInfo = 0;
        if (fs_handleIsSnapMagic(&req->parentHandle))
        {
            snapType = FS_SNAPSHOT_DIR_TYPE;
            snapInfo = FS_HANDLE_GET_INUM(req->parentHandle.inum);
        }
        else if (FS_HANDLE_IS_SNAP_BOUNDARY(req->parentHandle))
        {
            snapType = FS_SNAPSHOT_BOUNDARY_TYPE;
            snapInfo = FS_HANDLE_GET_SNAP(req->parentHandle.inum) + FS_SNAPSHOT_ID_INODE_FIRST;
        }

        fs_getSnapRespAttrs(context, parentInode, snapType, snapInfo,
                            FS_POST_PAR_ATTRS, &attrp,
                            &resp->numAttrs, TRUE);
    }

    if (inode != NULL)
    {
        fs_inodeRelease(context, inode);
    }
    if (parentInode != NULL)
    {
        fs_inodeRelease(context, parentInode);
    }

    return rc;

} /* fs_lookupThread */


/**
 * Handle the lookup request.  This function first tries the
 * dnlc, and if it fails, it creates a thread to do the rest.
 */
int32
fs_lookupInstantiate(
	const fs_ReqTypeTbl *reqType,
    const fs_MsgReqHdr  *reqHdr,
    int32               bufType,
    uint32              srcAddr)
{
    fs_MsgLookupReq     *req = (void *) reqHdr;
    fs_MsgLookupResp    *resp;
    fs_FileSystem       *fileSystem;
    evm_volId_t         volId;
    eee_descPtr_t       edesc;
    eeeRTC_t            startTime, deltaCpu;
    uint32              genNum;
    int32               rc;
    int                 parentAttrIndex = 1;
    int                 fsStatus;

    eee_rtcStart(&startTime);

    FS_TRACE(NULL, FS_TRACE_REQ,
             ("%s xid 0x%x lookup snap %lld inum %lld gen %d name %s",
              fs_protoStr[reqHdr->proto],
              reqHdr->clientId,
              FS_HANDLE_GET_SNAP(req->parentHandle.inum),
              FS_HANDLE_GET_INUM(req->parentHandle.inum),
              req->parentHandle.gen,
              req->fileName));

    volId = reqType->extractVolIdFunc(reqHdr);
    fileSystem = fs_fileSystemForVolume(volId);

    if (fileSystem == NULL ||
        (fileSystem->state != FS_VOL_STATE_ONLINE &&
         fileSystem->state != FS_VOL_STATE_ONLINE_RDONLY))
    {
        (reqType->ncRespFunc)(reqHdr, bufType, srcAddr, FS_ERR_NOT_MOUNTED);
        // PR 2785
        return NFX_OK;
    }

    // Let the slow path handle the oddball cases.
    if (fs_handleIsSnapMagic(&req->parentHandle) ||
        req->nameLen == 0 ||
        fileSystem->super->mkfsTime != req->parentHandle.mkfsTime ||
        !strcmp(req->fileName, ".snapshot") ||
        fs_nameIsDotOrDotDot(req->fileName))
    {
        return fs_threadInstantiate(reqType, reqHdr, bufType, srcAddr);
    }

#if defined(FS_STANDALONE)
    resp = fs_ramAlloc(FS_RAM_RESP_MSG, sizeof(fs_MsgLookupResp) +
                                           (2 * sizeof(fs_Attr)));
    resp->hdr.sockIndex = req->hdr.sockIndex;
    resp->hdr.nfsReqNum = req->hdr.nfsReqNum;
#else
    resp = eee_allocateBuffer(EEE_SHARED_POOL,
                              EEE_BUF_ULTRA,
                              NFX_COS_HIGH);
#endif
    resp->hdr.requestType = req->hdr.requestType;
    resp->hdr.clientPrivate = req->hdr.clientPrivate;
    resp->hdr.clientId = req->hdr.clientId;

    // Check the dnlc.  This returns with the attributes copied
    // from the inode cache if it hit in the dnlc and was able to
    // get parent and child attributes, otherwise it returns 0.

    // Check the dnlc.
    rc = fs_dnlcLookup(fileSystem, req->hdr.proto, req->parentHandle.inum,
                       req->fileName, req->nameLen, &resp->attr[0], &genNum);
    if (rc == FS_DNLC_UNKNOWN)
    {
        // Cache miss, use the slow path.
        eee_deallocateBuffer(resp);
        return fs_threadInstantiate(reqType, reqHdr, bufType, srcAddr);
    }

    FS_ASSERT(rc == FS_DNLC_EXISTS ||
              rc == FS_DNLC_NOT_EXISTS);

    // Don't forget the permission check.  Parent dir attrs
    // are in [0] if the looked up name does not exist.
    if (rc == FS_DNLC_NOT_EXISTS)
    {
        parentAttrIndex = 0;
    }

    if (!fs_doPermissionOkay(reqHdr,
                             FS_HANDLE_GET_INUM(req->parentHandle.inum),
                             FS_HANDLE_GET_SNAP(req->parentHandle.inum),
                             resp->attr[parentAttrIndex].mode,
                             resp->attr[parentAttrIndex].uid,
                             resp->attr[parentAttrIndex].gid,
                             FS_PERM_LOOKUP))
    {
        fsStatus = resp->hdr.fsStatus = FS_ERR_ACCESS;
        resp->hdr.attrType = FS_POST_PAR_ATTRS;
        resp->hdr.len = sizeof(fs_MsgLookupResp) + sizeof(fs_Attr);
        resp->numAttrs = 1;
        // in this case we are responding with only one set
        // of attributes: the parent
        if (parentAttrIndex != 0)
            resp->attr[0] = resp->attr[1]; 
    }
    else if (rc == FS_DNLC_EXISTS)
    {
        fsStatus = resp->hdr.fsStatus = NFX_OK;
        resp->hdr.attrType = FS_POST_PAR_ATTRS | FS_POST_ATTRS;
        resp->hdr.len = sizeof(fs_MsgLookupResp) + (2 * sizeof(fs_Attr));
        resp->numAttrs = 2;

        fs_handleInit(fileSystem,
                      &resp->fileHandle,
                      FS_HANDLE_GET_SNAP(req->parentHandle.inum),
                      resp->attr[0].inode,
                      genNum,
                      req->parentHandle.mntPnt);
    }
    else
    {
        fsStatus = resp->hdr.fsStatus = FS_ERR_NOENT;
        resp->hdr.attrType = FS_POST_PAR_ATTRS;
        resp->hdr.len = sizeof(fs_MsgLookupResp) + sizeof(fs_Attr);
        resp->numAttrs = 1;
    }

    // Send response

#if defined(FS_STANDALONE)
    {
        int32 cpuId = (int32) pthread_getspecific(fs_simKeyCpuId);
        rc = fs_simSendMsg(cpuId,
                           reqHdr->src,
                           (char *) resp,
                           resp->hdr.len);
    }
#else

    edesc = eee_allocateDesc(EEE_SHARED_POOL, NFX_COS_HIGH);
    if (edesc == NULL)
    {
        fs_volumeException(NULL, fileSystem, FS_VOLUME_EXCEPTION);
        goto freeReq;
    }

    edesc->bd[1].buf = (void *) resp;
    edesc->bd[1].len = resp->hdr.len;
    edesc->bd[1].control = EDESC_BD_ULTRA | EDESC_BD_CMD_IN_USE;

    edesc->hdr.control = (1 << EDESC_BUFCNT_SHIFT) | resp->hdr.len;
    edesc->hdr.src_port = fs_packedPortAddr(nfxMyCPUId, EEE_CONTROL);
    edesc->hdr.dest_port = srcAddr;
    rc = eee_forwardPacket(edesc);
#endif
    if (rc != NFX_OK)
    {
        fs_volumeException(NULL, fileSystem, FS_VOLUME_EXCEPTION);
        goto freeReq;
    }

	FS_TRACE(NULL, FS_TRACE_REQ, ("%s xid 0x%x lookup dnlc hit",
                                  fs_protoStr[reqHdr->proto],
                                  reqHdr->clientId));

    eee_rtcDuration(&startTime, &deltaCpu);

    NFX_SPIN_LOCK(&fileSystem->fs_statsReqRcvdSpinId);
    fileSystem->recvReqStats.num[FS_REQ_LOOKUP] += 1;
    if (fsStatus == NFX_OK)
        fileSystem->replyStats.ok[FS_REQ_LOOKUP] += 1;
    else
        fileSystem->replyStats.err[FS_REQ_LOOKUP] += 1;
    eee_rtcIncr(&fileSystem->yieldStats.num[FS_REQ_LOOKUP]
                                           [FS_YIELD_TYPE_ON_CPU],
                &deltaCpu);
    eee_rtcIncr(&fileSystem->yieldStats.num[FS_REQ_LOOKUP]
                                           [FS_YIELD_TYPE_TOTAL],
                &deltaCpu);
    NFX_SPIN_UNLOCK(&fileSystem->fs_statsReqRcvdSpinId);

    NFX_SPIN_LOCK(&fs_statsReqRcvdSpinId);
    // done by caller : fs_recvReqStats.num[msgType] += 1;
    if (fsStatus == NFX_OK)
        fs_replyStats.ok[FS_REQ_LOOKUP] += 1;
    else
        fs_replyStats.err[FS_REQ_LOOKUP] += 1;

    eee_rtcIncr(&fs_yieldStats.num[FS_REQ_LOOKUP][FS_YIELD_TYPE_ON_CPU],
                &deltaCpu);
    eee_rtcIncr(&fs_yieldStats.num[FS_REQ_LOOKUP][FS_YIELD_TYPE_TOTAL],
                &deltaCpu);
    NFX_SPIN_UNLOCK(&fs_statsReqRcvdSpinId);

 freeReq:

    if (bufType == FS_CONTEXT_SHARED_REQUEST)
    {
        eee_deallocateBuffer((void *) reqHdr);
    }
    else
    {
#if defined(FS_STANDALONE)
        fs_ramDealloc(FS_RAM_MCPU, (void *) reqHdr);
#else
        fs_ramDealloc(FS_RAM_REQ_MSG, (void *) reqHdr);
#endif
    }

	return NFX_OK;

} /* fs_lookupInstantiate */


/**
 * Handle the ACCESS request.
 */
int32
fs_accessThread(
    fs_Context      *context)
{
    fs_MsgAccessResp    *resp;
    fs_MsgAccessReq     *req;
    fs_Inode            *inode;
    fs_Attr             *attrp;
    int32               type;
    fs_SnapType         snapType = FS_NO_SNAPSHOT_TYPE;
    uint64              snapInfo = 0;

    resp = (fs_MsgAccessResp *) context->respHdr;
    req = (fs_MsgAccessReq *) context->reqHdr;
    context->respSize = sizeof(fs_MsgAccessResp);
    attrp = &resp->attr[0];
    resp->accessBits = 0;
    resp->numAttrs = 0;
    context->respSize = sizeof(fs_MsgAccessResp);

    FS_TRACE(context, FS_TRACE_REQ,
            ("%s xid 0x%x access snap %lld inum %lld gen %d",
            fs_protoStr[context->reqHdr->proto],
            context->reqHdr->clientId,
            FS_HANDLE_GET_SNAP(req->fileHandle.inum),
            FS_HANDLE_GET_INUM(req->fileHandle.inum),
            req->fileHandle.gen));

    if (fs_handleIsSnapMagic(&req->fileHandle))
    {
        inode = I_GET_SHARED(0, FS_SNAPSHOT_DIR_INODE);
    }
    else
    {
        inode = I_GET_SHARED(   FS_HANDLE_GET_SNAP(req->fileHandle.inum),
                                FS_HANDLE_GET_INUM(req->fileHandle.inum));
    }
    if (inode == NULL)
    {
        return FS_ERR_STALE;
    }
    if (!fs_handleIsValid(context, &req->fileHandle, inode))
    {
        fs_inodeRelease(context, inode);
        return FS_ERR_STALE;
    }

    if (req->accessBits & FS_ACCESS_READ)
    {
        if (FS_I_IS_DIR(inode->dinode.attrs))
        {
            type = FS_PERM_READDIR;
        }
        else
        {
            type = FS_PERM_READ;
        }
        if (fs_permissionOkay(context, inode, type))
        {
            resp->accessBits |= FS_ACCESS_READ;
        }
    }
    if (req->accessBits & FS_ACCESS_LOOKUP)
    {
        if (FS_I_IS_DIR(inode->dinode.attrs))
        {
            if (fs_permissionOkay(context, inode, FS_PERM_LOOKUP))
            {
                resp->accessBits |= FS_ACCESS_LOOKUP;
            }
        }
    }
    if (req->accessBits & FS_ACCESS_MODIFY)
    {
        if (FS_I_IS_DIR(inode->dinode.attrs))
        {
            type = FS_PERM_DIRENT;
        }
        else
        {
            type = FS_PERM_WRITE;
        }
        if (fs_permissionOkay(context, inode, type))
        {
			//printf("Check for write returned success\n");
            resp->accessBits |= FS_ACCESS_MODIFY;
        }
    }
    if (req->accessBits & FS_ACCESS_EXTEND)
    {
        if (FS_I_IS_DIR(inode->dinode.attrs))
        {
            type = FS_PERM_DIRENT;
        }
        else
        {
            type = FS_PERM_WRITE;
        }
        if (fs_permissionOkay(context, inode, type))
        {
			//printf("Check for extent returned success\n");
            resp->accessBits |= FS_ACCESS_EXTEND;
        }
    }
    if (req->accessBits & FS_ACCESS_DELETE)
    {
        if (FS_I_IS_DIR(inode->dinode.attrs))
        {
			//printf("Check for remove returned success\n");
            if (fs_permissionOkay(context, inode, FS_PERM_DIRENT))
            {
                resp->accessBits |= FS_ACCESS_DELETE;
            }
        }
    }
    if (req->accessBits & FS_ACCESS_EXECUTE)
    {
        if (!FS_I_IS_DIR(inode->dinode.attrs))
        {
            if (fs_permissionOkay(context, inode, FS_PERM_EXEC))
            {
                resp->accessBits |= FS_ACCESS_EXECUTE;
            }
        }
    }

    if (inode->inum == FS_SNAPSHOT_DIR_INODE)
    {
        snapType = FS_SNAPSHOT_DIR_TYPE;
        snapInfo = FS_HANDLE_GET_INUM(req->fileHandle.inum);
    }
    else if (FS_HANDLE_IS_SNAP_BOUNDARY(req->fileHandle))
    {
        snapType = FS_SNAPSHOT_BOUNDARY_TYPE;
        snapInfo = FS_HANDLE_GET_SNAP(req->fileHandle.inum) + FS_SNAPSHOT_ID_INODE_FIRST;
    }

    fs_getSnapRespAttrs(context, inode, snapType, snapInfo,
                        FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);
    fs_inodeRelease(context, inode);

    return NFX_OK;

} /* fs_accessThread */


/**
 * Handle the READ_LINK request.
 */
int32
fs_readLinkThread(
    fs_Context      *context)
{
    fs_MsgReadLinkResp  *resp;
    fs_MsgReadLinkReq   *req;
    fs_Inode            *inode;
    fs_Attr             *attrp;
    fs_Buf              *buf;

    resp = (fs_MsgReadLinkResp *) context->respHdr;
    req = (fs_MsgReadLinkReq *) context->reqHdr;
    context->respSize = sizeof(fs_MsgReadLinkResp);
    attrp = &resp->attr;

    FS_TRACE(context, FS_TRACE_REQ,
            ("%s xid 0x%x readlink snap %lld inum %lld gen %d",
            fs_protoStr[context->reqHdr->proto],
            context->reqHdr->clientId,
            FS_HANDLE_GET_SNAP(req->fileHandle.inum),
            FS_HANDLE_GET_INUM(req->fileHandle.inum),
            req->fileHandle.gen));

    if (fs_handleIsSnapMagic(&req->fileHandle))
    {
        return FS_ERR_INVAL;
    }
    else
    {
        inode = I_GET_SHARED(   FS_HANDLE_GET_SNAP(req->fileHandle.inum),
                                FS_HANDLE_GET_INUM(req->fileHandle.inum));
    }
    if (inode == NULL)
    {
        return FS_ERR_STALE;
    }
    if (!fs_handleIsValid(context, &req->fileHandle, inode))
    {
        fs_inodeRelease(context, inode);
        return FS_ERR_STALE;
    }
    if (!FS_I_IS_SYML(inode->dinode.attrs))
    {
        fs_inodeRelease(context, inode);
        return FS_ERR_INVAL;
    }

    if (inode->dinode.size <= FS_INODE_INLINE_LEN)
    {
        memcpy(resp->linkName, inode->dinode.u.data, inode->dinode.size);
    }
    else
    {
        buf = O_GET_SHARED(inode, 0);
        memcpy(resp->linkName, buf->u.data, inode->dinode.size);
        fs_bufRelease(context, buf);
    }

    resp->linkName[inode->dinode.size] = '\0';
    resp->nameLen = inode->dinode.size;
    context->respSize += resp->nameLen + 1;

    fs_inodeAtimeUpdate(context, inode);
    fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->fileHandle.inum),
                    FS_POST_ATTRS, &attrp, NULL, FALSE); 
    fs_inodeRelease(context, inode);

    return NFX_OK;

} /* readLinkThread */


/**
 * Handle the READ request.
 */
int32
fs_readThread(
    fs_Context      *context)
{
    fs_MsgReadResp  *resp ;
    fs_MsgReadReq   *req ;
    fs_MsgTransmitComplete *txcompl;
    fs_Inode        *inode = NULL;
    fs_Attr         *attrp;
    int32           rc;
    fs_Buf          *buf[NFX_MAX_COD_MSG];
    esm_event_t     wakeEvent;
    uint32          bytesRead, nbrUsedCod, loop;
    uint64          fileOffset;
    uint8           sequential = FALSE;
    boolean         isOwner, isGroupOwner, isSuperUser;

    resp = (fs_MsgReadResp *) context->respHdr;
    req  = (fs_MsgReadReq  *) context->reqHdr;
    context->respSize = sizeof(fs_MsgReadResp);
    attrp = &resp->attr[0];

    FS_TRACE(context, FS_TRACE_REQ,
            ("%s xid 0x%x read snap %lld inum %lld gen %d offset %lld len %d"
             " smHandle 0x%llx",
            fs_protoStr[context->reqHdr->proto],
            context->reqHdr->clientId,
            FS_HANDLE_GET_SNAP(req->fileHandle.inum),
            FS_HANDLE_GET_INUM(req->fileHandle.inum),
            req->fileHandle.gen,
            req->fileOffset,
            req->readLength,
            (uint64)((uintptr_t)context->smHandle)));

    // needed for the Tx Complete message
    memcpy (&(resp->fileHandle), &(req->fileHandle), sizeof(fs_Handle)) ;
    resp->fileOffset = req->fileOffset ;
    resp->nbrUsedCod = 0;
    resp->bytesRead = 0;
    resp->numAttrs = 0;

    if (fs_handleIsSnapMagic(&req->fileHandle))
    {
        if (FS_PROTO_NFS_V2 == req->hdr.proto)
          return FS_ERR_ISDIR;
        else
          return FS_ERR_INVAL;
    }
    else
    {
        inode = I_GET_SHARED(FS_HANDLE_GET_SNAP(req->fileHandle.inum),
                             FS_HANDLE_GET_INUM(req->fileHandle.inum));
    }
    if (inode == NULL) {
          //printf ("fs_readThread: I_GET_SHARED failed\n");
        return FS_ERR_STALE;
    }

	/*
	 *  If restore had dirtied buffers for this file flush them
	 */
	if (inode->flags & INO_FLAGS_RESTORE_IN_PROGRESS) {
            FS_ASSERT(FS_I_IS_REG(inode->dinode.attrs));
		    fs_inodeRelease(context, inode);
			inode = I_GET_EXCL(FS_HANDLE_GET_SNAP(req->fileHandle.inum),
								 FS_HANDLE_GET_INUM(req->fileHandle.inum));
			if (inode == NULL) {
					return FS_ERR_STALE;
			}
			if (inode->flags & INO_FLAGS_RESTORE_IN_PROGRESS) {
					fs_bufSyncInodeBufs(context, inode);
					fs_bufCacheFlush(context, inode);
			}
			fs_lockDemote(context, &inode->lock, FS_LOCK_MODE_SHARE);
	}
    //
    //    printf("Read        inum %8lld off %8lld len %5d vers %d \n",
    //    req->fileHandle.inum, req->fileOffset, req->readLength,
    //    req->hdr.proto+1);

    if (!fs_handleIsValid(context, &req->fileHandle, inode))
    {
        /* release lock (if NO modifications */
        //printf ("fs_readThread: genNum failed\n");
        fs_inodeRelease(context, inode);
        return FS_ERR_STALE;
    }

    if (!FS_I_IS_REG (inode->dinode.attrs)) {
        //printf ("fs_readThread: FS_I_IS_REG failed\n");
        fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->fileHandle.inum),
                        FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);

        if (FS_PROTO_NFS_V2 == req->hdr.proto)
          rc = FS_ERR_ISDIR;
        else
          rc = FS_ERR_INVAL ;
        
        fs_inodeRelease(context, inode) ;

        return rc;
    }


    /*
     * Here we apply the rule that the owner can read
     * regardless of the permissions per 8.7.2 of the NFS
     * book.  We don't do this in permissionOkay, since we
     * want requests such as ACCESS to return failure for
     * this case.
     */
    fs_getOwnership(context->reqHdr, inode->sd->u.nfs.uid, inode->sd->u.nfs.gid,
                    &isOwner, &isGroupOwner, &isSuperUser);
    if (!isOwner && !fs_permissionOkay(context, inode, FS_PERM_READ))
    {
        fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->fileHandle.inum),
                        FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_inodeRelease(context, inode);
        return FS_ERR_ACCESS;
    }

    // printf("read:  file size is %lld\n",
    //        inode->dinode.size);

    fileOffset = req->fileOffset;

    if ((fileOffset + req->readLength) < inode->dinode.size) 
    {
        resp->bytesRead = req->readLength;
        resp->eof = FALSE;
    }
    else 
    {
        if (fileOffset >= inode->dinode.size)
        {
            resp->bytesRead = 0;
        }
        else
        {
            resp->bytesRead = inode->dinode.size - fileOffset;
        }
        resp->eof = TRUE;
    }

    /*
     * Prepare for read ahead. Determine if we are being called sequentially
     * or not. If we are and the i/o on this file is in progress, queue
     * the request so that we don't break read ahead. This should not
     * penalize us as the likekly hood of this read block for i/o is
     * much higher if we have not yet started read ahead, so blocking
     * here will not pay additional penalty and at the same time will
     * allow us to do appropriate read aheads thus reducing over all latency.
     *
     * If the i/o is random, then note so and start async i/o for
     * the entire range and then block for each one of them to finish
     */
    //printf("read offset %lld len %d erOffset %lld\n", req->fileOffset,
    //        req->readLength, inode->readAheadState->erOffset);

    if (! fs_incoreTune->enableReadAhead)
    {
        goto read_proceed;
    }

    if (inode->readAheadState != NULL)
    {
        NFX_SPIN_LOCK(&inode->spin);
        if (inode->readAheadState->erOffset == req->fileOffset)
        {
            fs_PendingRead  *q;
    
            sequential = TRUE;
            inode->readAheadState->nPendingReads++;
            inode->readAheadState->erOffset += req->readLength;
            if (inode->readAheadState->nPendingReads > 1) {
                esm_event_t         event;
    
                /*
                 * Note, we can allocate memory holding spin locks as we
                 * we don't sleep for memalloc. If that changes, we need to
                 * move this outside of the spin lock
                 */
    
                q = fs_ramAlloc(FS_RAM_PENDING_REQUESTS,
                                sizeof (fs_PendingRead));
                ASSERT(q != NULL);
                bzero(q, sizeof (fs_PendingRead));
                q->offset = req->fileOffset;
                q->len = req->readLength;
                q->smHandle = context->smHandle;
                q->sequential = sequential;
                DLL_INSERT_TAIL(&inode->readAheadState->pendingReadList, q, list);
                NFX_SPIN_UNLOCK(&inode->spin);
                fs_yield(context, &event, FS_YIELD_TIME_FOREVER, FS_YIELD_TYPE_IO_UR);
                if (event.event != FS_THREAD_EVENT_READ_PROCEED) {
                    FS_ASSERT(0);
                }
                sequential = q->sequential;
                fs_ramDealloc(FS_RAM_PENDING_REQUESTS, q);
            } else {
                inode->readAheadState->erOffset = req->fileOffset + req->readLength;
                NFX_SPIN_UNLOCK(&inode->spin);
            }
            goto read_proceed;
        } else {
    
            fs_PendingRead  *q = NULL;
            /*
             * Random reads, wake up all the pending reads and and let them
             * be done as random i/o and also setup uxRead to 2
             */
    
            inode->readAheadState->uxRead = 2;
            inode->readAheadState->raSize = 0;
            inode->readAheadState->erOffset = req->fileOffset + req->readLength;
            do {
                DLL_REMOVE_HEAD(&inode->readAheadState->pendingReadList, q, list);
                if (q) {
                    esm_event_t   event;
                    event.event = FS_THREAD_EVENT_READ_PROCEED;
                    q->sequential = FALSE;
                    inode->readAheadState->nPendingReads--;
                    thread_event(q->smHandle, &event);
                    //fs_ramDealloc(FS_RAM_PENDING_REQUESTS, q);
                } else {
                    break;
                }
            } while (TRUE);
            FS_ASSERT(DLL_IS_EMPTY(&inode->readAheadState->pendingReadList));
            FS_ASSERT(inode->readAheadState->nPendingReads == 0 || inode->readAheadState->nPendingReads == 1);
    
            NFX_SPIN_UNLOCK(&inode->spin);
        }
    }
read_proceed:

    for (loop = 0; loop < NFX_MAX_COD_MSG; loop++)
    {
        buf[loop] = NULL;
    }
    bytesRead = 0;
    nbrUsedCod = 0;

    if (!sequential)
    {
        // Start async I/O if the request spans multiple pages.
        if (resp->bytesRead > FS_BLOCK_SIZE ||
            resp->bytesRead > FS_BLOCK_SIZE - (fileOffset % FS_BLOCK_SIZE))
        {
            fs_bufGetPageAsync(context, inode, fileOffset >> FS_BLOCK_SHIFT,
                               MIN(resp->bytesRead >> FS_BLOCK_SHIFT,
                               FS_MAX_READ_CLUSTER), FS_BUFLIST_USER);
        }
    }

    while (bytesRead < resp->bytesRead) 
    {
        uint64 wr_begin;
        uint32 lenThisBuf;
        
        wr_begin = (fileOffset / FS_BLOCK_SIZE) * FS_BLOCK_SIZE ;
        lenThisBuf = MIN (resp->bytesRead - bytesRead,
                          FS_BLOCK_SIZE - (fileOffset % FS_BLOCK_SIZE)) ;

        FS_ASSERT (nbrUsedCod < NFX_MAX_COD_MSG);
#if defined(FS_DEBUG)
        FS_ASSERT (lenThisBuf > 0);
        if ((nbrUsedCod > 0) && ((resp->bytesRead - bytesRead) > FS_BLOCK_SIZE))
        {
            FS_ASSERT (lenThisBuf == FS_BLOCK_SIZE);
        }
#endif
        
        buf[nbrUsedCod] = O_USER_GET_SHARED (inode, wr_begin / FS_BLOCK_SIZE,
                                            MIN(FS_BLOCK_SIZE,
                                              inode->dinode.size - wr_begin));
        resp->cod[nbrUsedCod].length = lenThisBuf;
        resp->cod[nbrUsedCod].offset = fileOffset % FS_BLOCK_SIZE ;

        if (buf[nbrUsedCod])
        {
            resp->cod[nbrUsedCod].cod = buf[nbrUsedCod]->u.user.codHandle ;
        }
        else
        {
            if (fs_nullCOD == -1)
            {
                // Initialize our NULL cod.
                fs_nullCOD = fs_cod_get_new_cod();
                fs_cod_create_null(fs_nullCOD,
                                   FS_BLOCK_SIZE,
                                   0);
            }

            resp->cod[nbrUsedCod].cod = fs_nullCOD;
        }

        FS_TRACE(context, FS_TRACE_COD,
                 ("smHandle %llx fs_readThread cod nbr %d %#llx %s off %d len %d", 
                  (uint64)((uintptr_t)context->smHandle), nbrUsedCod,
				  resp->cod[nbrUsedCod].cod,
                  ((resp->cod[nbrUsedCod].cod == fs_nullCOD) ? "NULL-COD" : ""),
                  resp->cod[nbrUsedCod].offset, resp->cod[nbrUsedCod].length));

        nbrUsedCod++;
        fileOffset += lenThisBuf;
        bytesRead += lenThisBuf;
        if (sequential) {
            fs_readAhead(context, inode, fileOffset - lenThisBuf,
                         lenThisBuf, 0);
        }
    }

    if (inode->readAheadState != NULL && sequential)
    {
        fs_PendingRead  *q = NULL;

        NFX_SPIN_LOCK(&inode->spin);
        inode->readAheadState->nPendingReads--;
        DLL_REMOVE_HEAD(&inode->readAheadState->pendingReadList, q, list);

        /*
         * Wakeup the waiter
         */

        if (q != DLL_NIL && q) {
            esm_event_t   event;
            event.event = FS_THREAD_EVENT_READ_PROCEED;
            thread_event(q->smHandle, &event);
            //fs_ramDealloc(FS_RAM_PENDING_REQUESTS, q);
        }
        NFX_SPIN_UNLOCK(&inode->spin);
    }
    
    fs_inodeAtimeUpdate(context, inode);
    fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->fileHandle.inum),
                    FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);
    fs_inodeRelease(context, inode) ;

	resp->fsPrivate = context->smHandle;
	resp->hdr.fsStatus = NFX_OK;
    resp->nbrUsedCod = nbrUsedCod;
	FS_ASSERT(resp->nbrUsedCod <= NFX_MAX_COD_MSG);
	FS_ASSERT (bytesRead == resp->bytesRead);

    FS_TRACE(context, FS_TRACE_REQ,
             ("read_resp fileoffset %lld len %d nbrCOD %d eof %d smHandle 0x%llx",
              resp->fileOffset,
              resp->bytesRead,
              resp->nbrUsedCod,
              resp->eof,
              (uint64)((uintptr_t)context->smHandle)));

	fs_threadRespond(context, NFX_OK);

#if defined(FS_STANDALONE)
	context->respHdr = NULL;
#else
    fs_contextFreeResponse(context);
#endif

    if (bytesRead > 0)
    {
        // Release the CPU, wait for the TX_COMPL message.
        fs_yield(context, &wakeEvent, FS_YIELD_TIME_TXCOMPL, FS_YIELD_TYPE_TXCOMPL);

        if (wakeEvent.event == THREAD_EVENT_TIMEOUT)
        {
            fs_txComplTimeouts += 1;
            printf("fs readThread timed out waiting for TX_COMPL from TXRX proc\n");
            printf("  vol 0x%llx smHandle 0x%llx clientId 0x%x clientPriv 0x%llx\n",
                context->volId.whole_thing, (uint64)((uintptr_t)context->smHandle),
                req->hdr.clientId, req->hdr.clientPrivate);
        }
        else
        {
            FS_TRACE(context, FS_TRACE_REQ, ("smHandle %llx ++++ %s ++++",
                     (uint64)((uintptr_t)context->smHandle), "txcompl"));
    
            // Make sure the event we got is the event
            // we were waiting for!  If not, return error.
    
            if ((wakeEvent.event != FS_THREAD_EVENT_TX_COMPL) ||
                  (wakeEvent.p1.u.p == NULL))
	        {
                fs_error(context, FS_INTERNAL_ERR, "tx compl event");
	        }
    
	        // Sanity checking on the tx complete message.
	        txcompl = wakeEvent.p1.u.p;
#ifdef NOTDEF
	        if (memcmp(&txcompl->fileHandle, &req->fileHandle,
							        sizeof(req->fileHandle)) != 0 ||
		        txcompl->fileOffset != req->fileOffset ||
		        txcompl->bytesRead != bytesRead)
	        {
		        fs_error(context, FS_INTERNAL_ERR, "tx compl params");
	        }
#endif
    
            // make the statistics correct for this event
            NFX_SPIN_LOCK(&fs_statsReqRcvdSpinId);
            fs_recvReqStats.num[FS_REQ_TX_COMPL] += 1;
            NFX_SPIN_UNLOCK(&fs_statsReqRcvdSpinId);
            NFX_SPIN_LOCK(&context->fileSystem->fs_statsReqRcvdSpinId);
            context->fileSystem->recvReqStats.num[FS_REQ_TX_COMPL] += 1;
#if defined(FS_DEBUG_CHECK_CRASH)
            if (fs_debugCrashEnabled)
                fs_debugCheckCrash(context);
#endif
            NFX_SPIN_UNLOCK(&context->fileSystem->fs_statsReqRcvdSpinId);

	        // Done with the txcompl message, free it.
            if (wakeEvent.p2.u.w & FS_CONTEXT_SHARED_REQUEST)
            {
                eee_deallocateBuffer(txcompl);
            }
            else
            {
                fs_ramDealloc(FS_RAM_REQ_MSG, txcompl);
            }
        }
    }

    // unlock shared bufs done with the fs_txComplThread
    for (loop = 0 ; loop < nbrUsedCod; loop++)
    {
        if (buf[loop])
            fs_bufRelease(context, buf[loop]);
    }
    
    if (bytesRead > 0)
    {
        FS_TRACE(context, FS_TRACE_REQ,
                 ("smHandle %llx ---- %s ---- returns 0",
                 (uint64)((uintptr_t)context->smHandle),
                 "txcompl"));
    }

    return NFX_OK;

} /* fs_readThread */


uint64 fs_stat_write_stable = 0;
uint64 fs_stat_write_un_stable = 0;

/**
 * Handle the WRITE request.
 */
int32
fs_writeThread(
    fs_Context  *context)
{
    fs_MsgWriteResp *       resp = (fs_MsgWriteResp *) context->respHdr;
    fs_MsgWriteReq  *       req = (fs_MsgWriteReq  *) context->reqHdr;
    fs_Inode        *       inode = NULL;
    int32                   rc = FS_ERR_NONE;
    int32                   loop;
    fs_QuotaContextList_t * list = NULL;
    fs_Attr *               attrp = &resp->attr[0];
    fs_WriteStableHow       stableHow;
#ifdef FS_DEBUG_CHECK_CRASH
    extern boolean          fs_debugWriteYield;
#endif
    int                     flush = 0;
    uint64                  origSize;
#define MAX_BUFS    20      // max nb of buffers - what's the optimised value?
    struct
    {
        fs_Buf *    buf;
        boolean     created;
    }           bufs[MAX_BUFS];         // list of buffers
    int         bufsIdx = 0;
    int         nbBufs = 0;
    fs_Buf *    bufToAlloc[MAX_BUFS];   // list of buffers to write
    int         bufToAllocIdx = 0;
    int         nbBufToAlloc = 0;
    fs_Buf *    bufToWrite[MAX_BUFS];   // list of new buffers to allocate
    int         bufToWriteIdx = 0;
    int         nbBufToWrite = 0;
    uint64      quotaReserved = 0;
    int         nbCodsTofree;
    extern uint32   fs_codRxNfs;
    boolean     isOwner, isGroupOwner, isSuperUser;
    
    FS_TRACE(context, FS_TRACE_REQ,
             ("%s xid 0x%x write snap %lld inum %lld gen %d off %lld"
              " len %lld %s nbr cod %d",
              fs_protoStr[context->reqHdr->proto],
              context->reqHdr->clientId,
              FS_HANDLE_GET_SNAP(req->fileHandle.inum),
              FS_HANDLE_GET_INUM(req->fileHandle.inum),
              req->fileHandle.gen,
              req->fileOffset,
              req->writeLength,
              ((req->stableHow == 0) ? "UNSTABLE" : "STABLE"),
              req->nbrUsedCod));

    fs_codRxNfs += req->nbrUsedCod;
    nbCodsTofree = req->nbrUsedCod;

    /* Start to initialize the response.
     */
    context->respSize = sizeof(fs_MsgWriteResp);
    resp->numAttrs = 0;
    
    /* Make sure the file is not read-only.
     */
    if (context->fileSystem->state == FS_VOL_STATE_ONLINE_RDONLY)
    {
        rc = FS_ERR_ROFS;
        goto fs_writeThread_exit;
    }

    /* Check if the write should be commited before responding.
     */
    if ((FS_PROTO_NFS_V3 == req->hdr.proto) &&
        (FS_WRITE_STABLE_UNSTABLE == req->stableHow))
    {
        stableHow = FS_WRITE_STABLE_UNSTABLE;
        fs_stat_write_un_stable++;
    }
    else
    {
        stableHow = FS_WRITE_STABLE_FILE_SYNC;
        fs_stat_write_stable++;
    }

    /* Get the inode.
     */
    if (fs_handleIsSnapMagic(&req->fileHandle))
    {
        rc = FS_ERR_INVAL;
        goto fs_writeThread_exit;
    }
    else
    {
        inode = I_GET_EXCL(FS_HANDLE_GET_SNAP(req->fileHandle.inum),
                           FS_HANDLE_GET_INUM(req->fileHandle.inum));
    }
    if (inode == NULL)
    {
        rc = FS_ERR_STALE;
        goto fs_writeThread_exit;
    }
    origSize = inode->dinode.size;

    /* If restore is in progress on this inode, then
     * we should flush all the cache and invalidate it
     */
    if (inode->flags & INO_FLAGS_RESTORE_IN_PROGRESS)
    {
        fs_bufSyncInodeBufs(context, inode);
        fs_bufCacheFlush(context, inode);
        flush = 1;
    }

    /* Check if the file handle is valid.
     */
    if (!fs_handleIsValid(context, &req->fileHandle, inode))
    {
        rc = FS_ERR_STALE;
        goto fs_writeThread_exit;
    }

    // printf("Write       inum %8lld off %8lld len %5lld vers %d %s\n",
    //   req->fileHandle.inum, req->fileOffset, req->writeLength,
    //   req->hdr.proto+1, (stableHow == 0) ? "UNSTABLE" : "STABLE");

    /* Set the pre-attributes in the response.
     */
    fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->fileHandle.inum),
                    FS_PRE_ATTRS, &attrp, &resp->numAttrs, TRUE);

    /* Check if it's a regular file.
     */
    if (!FS_I_IS_REG(inode->dinode.attrs))
    {
        if (FS_I_IS_DIR(inode->dinode.attrs))
        {
            rc = FS_ERR_ISDIR;
        }
        else
        {
            rc = FS_ERR_STALE;
        }
        goto fs_writeThread_exit;
    }

    /* Check if the permission are fine.
     *
     * Here we apply the rule that the owner can write
     * regardless of the permissions per 8.7.2 of the NFS
     * book.  We don't do this in permissionOkay, since we
     * want requests such as ACCESS to return failure for
     * this case.
     */
    fs_getOwnership(context->reqHdr, inode->sd->u.nfs.uid, inode->sd->u.nfs.gid,
                    &isOwner, &isGroupOwner, &isSuperUser);
    if (!isOwner && !fs_permissionOkay(context, inode, FS_PERM_WRITE))
    {
        fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->fileHandle.inum),
                        FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);
        rc = FS_ERR_ACCESS;
        goto fs_writeThread_exit;
    }

    /* ???
     */
	if (context->flags & FS_CONTEXT_ENOSPC_CONDITION)
    {
        fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->fileHandle.inum),
                        FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);
		rc = FS_ERR_NOSPC;
        goto fs_writeThread_exit;
	}

    /*
    if (req->writeLength > NFS_MAXDATA)
      {
        // short write
        // NFSv3 : Todo : define max length
        req->writeLength = NFS_MAXDATA ;
      }
    */

    /* Let's write the data!
     * Go through all received COD and find their matching buffers or create them.
     * Figure out what we need to allocate and what not.
     * Check if the quota allow this write request.
     * Allocate all new buffers in one go.
     * And write all data.
     */
    if (req->writeLength > 0)
    {
        int         i;
        nfx_cod_t   src_cod;
        uint32      src_cod_offset;
        uint32      src_cod_length;
        nfx_cod_t   next_cod = 0;
        uint32      next_cod_length = 0;
        nfx_cod_t   dst_cod = 0;
        int64       wr_begin_block;
        int64       wr_end_block;
        int64       fs_end_block_in_use;
        uint32      usedCod = 0;
        uint64      bytesWritten = 0;
        uint64      fileOffset = req->fileOffset;

        /* Initialize the lists of buffers.
         */
        for (i = 0; i < MAX_BUFS; i++)
        {
            bufs[i].buf = NULL;
            bufs[i].created = FALSE;
            bufToWrite[i] = NULL;
            bufToAlloc[i] = NULL;
        } // for

        /* Calculate: - the byte offset of first block affected by this write.
         *            - byte offset of last block affected by this write
         *            - byte offset of last block in the file
         */
        wr_begin_block = (fileOffset / FS_BLOCK_SIZE) * FS_BLOCK_SIZE;
        wr_end_block = ((fileOffset + req->writeLength - 1) 
                        / FS_BLOCK_SIZE) * FS_BLOCK_SIZE;
        fs_end_block_in_use = ((inode->dinode.size - 1) 
                               / FS_BLOCK_SIZE) * FS_BLOCK_SIZE;

        /* If file already in use and
         * write in higher block than current file size and
         * file not at block boundary.
         * Find previous block and complete it with zeroes.
         */

        if ((fs_end_block_in_use >= 0) &&
            (wr_begin_block > fs_end_block_in_use) &&
            ((inode->dinode.size % FS_BLOCK_SIZE) != 0))
        {
            bufs[bufsIdx].buf = O_USER_GET_EXCL(inode,
                                        fs_end_block_in_use / FS_BLOCK_SIZE,
                                        inode->dinode.size % FS_BLOCK_SIZE);
            bufs[bufsIdx].created = FALSE;
            
            /* It is possible that buf does not exist.
             * Example: set file size with set attr cmd.
            */
            if (bufs[bufsIdx].buf != NULL)
            {
                fs_cod_fill_null(&bufs[bufsIdx].buf->u.user.codHandle, 
                                 FS_BLOCK_SIZE - (inode->dinode.size %
                                                  FS_BLOCK_SIZE),
                                 inode->dinode.size % FS_BLOCK_SIZE);
                // still same COD in buf->u.user.codHandle
                bufs[bufsIdx].buf->u.user.codLength = FS_BLOCK_SIZE;
#ifdef NEW_ZF
                fs_logAddUserZeroRecord(context, inode,
                                        bufs[bufsIdx].buf->page,
                                        (inode->dinode.size % FS_BLOCK_SIZE),
                                        1, TRUE, FALSE, &(bufs[bufsIdx].buf));
#endif // NEW_ZF
                bufsIdx++;
            }
        }

        /* Go through all the data to write.
         */
        src_cod = req->cod[usedCod].cod;
        src_cod_offset = req->cod[usedCod].offset;
        src_cod_length = req->cod[usedCod].length;
        usedCod++;
        // This case is not yet covered
        FS_ASSERT(src_cod_offset == 0);
        while (bytesWritten < req->writeLength)
        {
            int64   begin_in_block;
            int64   end_in_block;
            
            if (usedCod > req->nbrUsedCod)
            {
                printf("fs_writeThread usedCod %d > req->nbrUsedCod %d "
                       "(%lld %lld)\n", usedCod, req->nbrUsedCod,
                       bytesWritten, req->writeLength);
                FS_ASSERT(FALSE);
            }
            FS_ASSERT (src_cod_length > 0);

            begin_in_block = MAX (0, (fileOffset - wr_begin_block));
            if ((fileOffset + src_cod_length - wr_begin_block) > FS_BLOCK_SIZE)
            {
                end_in_block = FS_BLOCK_SIZE - 1;
                next_cod = fs_cod_get_new_cod ();
                next_cod_length = src_cod_length -
                    (FS_BLOCK_SIZE - begin_in_block);
                src_cod_length = FS_BLOCK_SIZE - begin_in_block;
                // next_cod_offset = 0;
                fs_cod_split(src_cod, (FS_BLOCK_SIZE - begin_in_block),
                             next_cod);
            }
            else 
            {
                end_in_block = fileOffset + src_cod_length - wr_begin_block - 1 ;
            }

            if (wr_begin_block > fs_end_block_in_use)
            {
                /* Write a new block after (current) End of File.
                 */
                if (begin_in_block > 0)
                {
                    /* Merge with NULL COD
                     * or write '0' from 0 to begin_in_block
                     * prepend block with 'begin_in_block' '0's
                     */
                    dst_cod = fs_cod_get_new_cod();
                    fs_cod_create_null(dst_cod, begin_in_block, 0);
                    fs_cod_merge (src_cod, (end_in_block - begin_in_block + 1),
                                  src_cod_offset, dst_cod, begin_in_block);
                    // merge will free the src_cod == cod
                    src_cod = dst_cod ;
                }

                /* A new buffer must be created
                 * Todo: allocate aggressively !!
                 */
                bufs[bufsIdx].buf = O_USER_CREATE(inode, wr_begin_block /
                                                  FS_BLOCK_SIZE);
                bufs[bufsIdx].created = TRUE;
                if (bufs[bufsIdx].buf == NULL)
                {
                    fs_error(context, FS_INTERNAL_ERR,
                             "fs_writeThread O_USER_CREATE");
                }
                bufs[bufsIdx].buf->u.user.codHandle = src_cod ;
                bufs[bufsIdx].buf->u.user.codOffset = 0 ;
                bufs[bufsIdx].buf->u.user.codLength = end_in_block+1;
                FS_ASSERT(bufs[bufsIdx].buf->u.user.codLength <= FS_BLOCK_SIZE);
            }
            else
            {
                /* Merge with existing buffer
                 * Find the existing buf
                 * Everything, up to filesize must be valid in this block
                 */
                bufs[bufsIdx].buf = O_USER_GET_EXCL (inode,
                                            wr_begin_block / FS_BLOCK_SIZE,
                                            MIN(FS_BLOCK_SIZE,
                                                    inode->dinode.size -
                                                        wr_begin_block));
                FS_ASSERT(inode->dinode.size > wr_begin_block);

                bufs[bufsIdx].created = FALSE;
                if (bufs[bufsIdx].buf == NULL)
                {
                    /* If buf is not found, assume FS_BLOCK_SIZE of all '0's
                     * - create NULL cod (create new, append FS_BLOCK_SIZE '0'
                     */
                    dst_cod = fs_cod_get_new_cod();
                    fs_cod_create_null(dst_cod, FS_BLOCK_SIZE, 0);

                    /* A new buffer must be created
                     * Todo: allocate non aggressively just one buf !!
                     */
                    bufs[bufsIdx].buf = O_USER_CREATE(inode, wr_begin_block /
                                                      FS_BLOCK_SIZE);
                    bufs[bufsIdx].created = TRUE;
                    if (bufs[bufsIdx].buf == NULL)
                    {
                        fs_error(context, FS_INTERNAL_ERR,
                                 "fs_writeThread O_USER_CREATE");
                    }

                    fs_cod_merge(src_cod, (end_in_block - begin_in_block + 1),
                                 src_cod_offset, dst_cod, begin_in_block);
                    // merge will free the src_cod == cod
                    bufs[bufsIdx].buf->u.user.codHandle = dst_cod ;
                    bufs[bufsIdx].buf->u.user.codOffset = 0 ;
                    bufs[bufsIdx].buf->u.user.codLength = FS_BLOCK_SIZE;
                }
                else
                {
                    dst_cod = bufs[bufsIdx].buf->u.user.codHandle ;

                    if (begin_in_block < inode->dinode.size % FS_BLOCK_SIZE)
                    {
                        // now the statistics
                        fs_cod_nbr_copies++ ;
                        fs_cod_nbr_bytes += bufs[bufsIdx].buf->u.user.codLength ;
                        fs_cod_nbr_overwrite += (end_in_block - begin_in_block + 1) ;
                        FS_TRACE(context, FS_TRACE_COD,
                                 ("Copy a cod: cnt %lld, bytes %lld, "
                                  "overwrite %lld", fs_cod_nbr_copies,
                                  fs_cod_nbr_bytes, fs_cod_nbr_overwrite));
                    }

                    if (inode->dinode.size < (wr_begin_block + begin_in_block))
                    {
                        /* Fill the gap with '0's  between 
                         * - filesize  
                         * - begin_in_block
                         */
                        // just for debugging ??
                        if ((inode->dinode.size <= wr_begin_block) ||
                            (inode->dinode.size >= (wr_begin_block +
                                                     FS_BLOCK_SIZE)))
                        {
                            printf ("WRITE %#llx %#llx %lld %lld %lld %lld "
                                    "%lld %lld\n", src_cod, dst_cod,
                                    inode->dinode.size, wr_begin_block,
                                    wr_end_block, fs_end_block_in_use, 
                                    begin_in_block, end_in_block);
                            FS_ASSERT(FALSE);
                        }
                    
                        fs_cod_fill_null(&dst_cod, 
                                         begin_in_block - (inode->dinode.size %
                                                           FS_BLOCK_SIZE),
                                         inode->dinode.size % FS_BLOCK_SIZE);
                		bufs[bufsIdx].buf->u.user.codLength = FS_BLOCK_SIZE;
                    }
                    
                    fs_cod_merge(src_cod, (end_in_block - begin_in_block + 1),
                                 src_cod_offset, dst_cod, begin_in_block);
                    // merge will free the src_cod == cod
                    bufs[bufsIdx].buf->u.user.codHandle = dst_cod ;
                    bufs[bufsIdx].buf->u.user.codOffset = 0 ;
                    bufs[bufsIdx].buf->u.user.codLength =
                        MAX(end_in_block+1, bufs[bufsIdx].buf->u.user.codLength);
                    FS_ASSERT(bufs[bufsIdx].buf->u.user.codLength <=
                                FS_BLOCK_SIZE);
                }
            }
            bytesWritten += src_cod_length;
            fileOffset += src_cod_length;
            wr_begin_block = (fileOffset / FS_BLOCK_SIZE) * FS_BLOCK_SIZE;
            // wr_begin_block += FS_BLOCK_SIZE;
            inode->dinode.size = MAX (inode->dinode.size, fileOffset);
            fs_end_block_in_use = ((inode->dinode.size - 1) 
                                   / FS_BLOCK_SIZE) * FS_BLOCK_SIZE;
            if (next_cod == 0)
            {
                src_cod = req->cod[usedCod].cod;
                src_cod_offset = req->cod[usedCod].offset;
                src_cod_length = req->cod[usedCod].length;
                usedCod++;
            }
            else
            {
                // the src_cod was split, now take care of the second part
                src_cod = next_cod;
                src_cod_offset = 0;
                src_cod_length = next_cod_length;
                next_cod = 0;
            }
            bufsIdx++;
        } // while (bytesWritten < req->writeLength)
        nbBufs = bufsIdx;
        FS_ASSERT(nbBufs <= MAX_BUFS);
     
        /* Now cods are attached to buffers. They should not be released directly */
        nbCodsTofree = 0;
                
        /* Initialize the list of buffers to write and allocate.
         * Make sure that there is only one buffer per page.
         */
        for (bufsIdx = 0; bufsIdx < nbBufs; bufsIdx++)
        {
            int     j;
            boolean alreadyInList = FALSE;

            /* Add the current buffer to the list of buffers to write
             * if not already added.
             */
            for (j = 0; j < bufToWriteIdx; j++)
            {
                if (bufToWrite[j]->page == bufs[bufsIdx].buf->page)
                {
                    alreadyInList = TRUE;
                    break;
                }
            } // for (j)
            if (alreadyInList == FALSE)
            {
                bufToWrite[bufToWriteIdx] = bufs[bufsIdx].buf;
                bufToWriteIdx++;
                
                /* If the block doesn't exist yet,
                 * add the current buffer to the list of buffers to allocate.
                 */
                if (!fs_blockIsValid(bufs[bufsIdx].buf->block))
                {
                    bufToAlloc[bufToAllocIdx] = bufs[bufsIdx].buf;
                    bufToAllocIdx++;
                }
            }
        } // for(bufsIdx)
        nbBufToWrite = bufToWriteIdx;
        nbBufToAlloc = bufToAllocIdx;
        FS_ASSERT(nbBufToWrite <= nbBufs);
        FS_ASSERT(nbBufToAlloc <= nbBufToWrite);

        /* Now that we now how much blocks we need to allocate, check quota.
         * If what we need is not available, then fail the write request.
         * Otherwise, allocate all buffers in one go.
         */
        if (nbBufToAlloc != 0)
        {
            uint64 quotaReserved;

            quotaReserved = fs_getQuotaReservation(context,
                                                   context->fileSystem,
                                                   inode, nbBufToAlloc,
                                                   FS_COUNT_BLOCKS,
                                                   inode->sd->u.nfs.uid,
                                                   inode->sd->u.nfs.gid,
                                                   0, &list);
            if (quotaReserved != nbBufToAlloc)
            {
                /* Undo what we did so far.
                 */
                inode->dinode.size = origSize;
                for (bufsIdx = 0; bufsIdx < nbBufs; bufsIdx++)
                {
                    if (bufs[bufsIdx].created == TRUE)
                    {
                        B_FREE(bufs[bufsIdx].buf);
                    }
                    else
                    {
                        fs_bufReleaseExclusive(context, bufs[bufsIdx].buf);
                    }
                } // for(bufsIdx)
                
                fs_getRespAttrs(context, inode,
                                FS_HANDLE_GET_INUM(req->fileHandle.inum),
                                FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);
                rc = FS_ERR_DQUOT;
                goto fs_writeThread_exit;
            }
            
            fs_bufAllocDoBufs(context, inode, nbBufToAlloc,
                              bufToAlloc, TRUE);

#ifdef ISABELLE // test only
            //make sure that all allocated pages are sequential
            {
                uint64  prevPage = bufToAlloc[0]->page;
                
                for (i = 1; i < nbBufToAlloc; i++)
                {
                    FS_ASSERT(bufToAlloc[i]->page == prevPage + 1);
                    prevPage = bufToAlloc[i]->page;
                } // for
            }
#endif // ISABELLE
        }
        
        /*
         * Dirty all the bufs we're modifying.
         */
        for (bufToWriteIdx = 0; bufToWriteIdx < nbBufToWrite; bufToWriteIdx++)
        {
            FS_ASSERT(bufToWrite[bufToWriteIdx] != NULL);
            
            FS_BUF_WRITE(bufToWrite[bufToWriteIdx], 0,
                         bufToWrite[bufToWriteIdx]->u.user.codLength) ;
        } // for (bufToWriteIdx)

        /* Zero fill the newly allocated blocks.
         */
        if ((stableHow == FS_WRITE_STABLE_UNSTABLE) && (nbBufToAlloc > 0) &&
            (context->fileSystem->logRoot->logDevType == FS_LOG_DEVTYPE_VOLID ||
             context->fileSystem->logRoot->logDevType == FS_LOG_DEVTYPE_SELF))
#ifdef NEW_ZF
        {
            uint64 lastPage = origSize / FS_BLOCK_SIZE;
            
            if ((bufToAlloc[0]->page == lastPage + 1) &&
                (origSize % FS_BLOCK_SIZE != 0))
            {
                fs_logAddUserZeroRecord(context, inode,
                                        lastPage,
                                        (origSize % FS_BLOCK_SIZE),
                                        (nbBufToAlloc + 1),
                                        TRUE, FALSE, bufToAlloc);
            }
            else
            {
                fs_logAddUserZeroRecord(context, inode,
                                        bufToAlloc[0]->page,
                                        0,
                                        nbBufToAlloc,
                                        TRUE, FALSE, bufToAlloc);
            }
        }
#else // NEW_ZF
        {
            int32 i;
            
            for (i = 0; i < nbBufToAlloc; i++)
            {
                fs_logZeroFillAndCommit(context, bufToAlloc[i]);
            }
        }
#endif // NEW_ZF

        /* ???
         */
        if ((nbBufToWrite > 0) && (bufToWrite[nbBufToWrite - 1] == NULL))
        {
            fs_error(context, FS_INTERNAL_ERR, "fs_writeThread : No buf");
        }
        
        FS_TRACE(context, FS_TRACE_COD,
                 ("fs-esm fs_writeThread, cod %#llx, "
                  "off %lld len %lld, new file size %lld", 
                  req->cod[usedCod].cod, req->fileOffset,
                  req->writeLength, inode->dinode.size));
    }
    else
    {
        FS_TRACE(context, FS_TRACE_COD,
                 ("fs-esm fs_writeThread, writeLength == 0"));
    }

    /* Set mtime.
     * no need to change the parentInode (xtime)
     */
    if (req->writeLength > 0)
    {
        inode->dinode.ctime = fs_getTimeOfDay();
        inode->dinode.version++;
        inode->dinode.mtime = inode->dinode.ctime;
        inode->dinode.atime = inode->dinode.ctime;
    }
    else
    {
        inode->dinode.mtime = fs_getTimeOfDay();
    }

    /* Set the post-attributes in the response.
     */
    fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->fileHandle.inum),
                    FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);

    /* Write the inode.
     */
    fs_inodeWrite(context, inode, INO_WRITE_LOG);

    /* Commit the quotas
     */
    if ((req->writeLength > 0) && (list != NULL))
    {
        fs_commitContextList(context, context->fileSystem, list, FALSE);   
        fs_freeQuotaContextList(context, context->fileSystem, list);   
    }

#ifdef FS_DEBUG_CHECK_CRASH
    /* This simply causes the thread to yield for a period
     * to allow some log debugging (we cause a crash after
     * this and test replay over uncommitted records).
     */
	if (fs_debugWriteYield)
	{
		fs_debugWriteYield = FALSE;
		printf("Logid %d will yield for 30 seconds (tail %d)...\n",
               context->logSummary->id, context->fileSystem->logRoot->tail);
		fs_logDumpIncoreData(context, context->logSummary);
		fs_sleep(context, 30000);
		printf("Logid %d resuming from yield (tail %d)...\n",
               context->logSummary->id, context->fileSystem->logRoot->tail);
	}
#endif
    
    /* Commit the changes.
     * And yield if needed (depending of stableHow
     * This can overwrite the stableHow (only from UNSTABLE to FILE_SYNC)
     */
    rc = fs_logCommitWrite(context, inode, &req->fileHandle.inum, &stableHow);
    if (rc != NFX_OK)
    {
        /* Release quota list.
         */
        if (list != NULL)
        {
            fs_trimQuotaReservation(context, context->fileSystem, list, quotaReserved);
            fs_freeQuotaContextList(context, context->fileSystem, list);
        }
            
        /* Release the inode.
         */
        if (inode != NULL)
        {
            fs_inodeReleaseExclusive(context, inode);
        }

	    NFX_SPIN_LOCK(&fs_errStatsSpinId);
	    fs_errStats.resource[FS_RESOURCE_IO] += 1;
        context->fileSystem->nfsStats.errs[context->reqHdr->requestType] += 1;
		context->fileSystem->errStats.resource[FS_RESOURCE_IO] += 1;
	    NFX_SPIN_UNLOCK(&fs_errStatsSpinId);

        return FS_ERR_NO_RESOURCE;
    }

    /* Fill the response.
     */
#ifdef ISABELLE // test only!!!
    // so client doesn't send commit
    resp->stableHow = FS_WRITE_STABLE_DATA_SYNC;
#else // ISABELLE
    resp->stableHow = stableHow;
#endif // ISABELLE
    resp->verifier = context->fileSystem->super->mountTime[0];
    resp->count = (uint32)req->writeLength;

    /* If restore is in progress, we should flush the data.
     */
    if (flush)
    {
        inode = I_GET_EXCL(FS_HANDLE_GET_SNAP(req->fileHandle.inum),
                           FS_HANDLE_GET_INUM(req->fileHandle.inum));
        fs_bufSyncInodeBufs(context, inode);
        fs_bufCacheFlush(context, inode);
        fs_inodeReleaseExclusive(context, inode);
    }

fs_writeThread_exit:

    if (rc != FS_ERR_NONE)
    {
        for (loop = 0; loop < nbCodsTofree ; loop++)
        {
            fs_cod_free_cod(req->cod[loop].cod);
        }

        /* Release quota list.
         */
        if (list != NULL)
        {
            fs_trimQuotaReservation(context, context->fileSystem, list,
                                    quotaReserved);
            fs_freeQuotaContextList(context, context->fileSystem, list);
        }
            
        /* Release the inode.
         */
        if (inode != NULL)
        {
            fs_inodeReleaseExclusive(context, inode);
        }
    }

    return(rc);
}

/**
 * Handle the CREATE request.
 */
int32
fs_createThread(
    fs_Context      *context)
{
    fs_MsgCreateResp    *resp;
    fs_MsgCreateReq     *req;
    fs_Inode            *inode, *parentInode;
    fs_Attr             *attrp;
    uint64              inum;
    int32               rc;
    int64		        parentNumBlocks;
    uint16              mode;

    fs_QuotaContextList_t   *list1 = NULL, *list2 = NULL;
    uint64                  quotaReserved = 0;
    fs_Status               status;

    resp = (fs_MsgCreateResp *) context->respHdr;
    req = (fs_MsgCreateReq *) context->reqHdr;

    context->respSize = sizeof(fs_MsgCreateResp);
    resp->numAttrs = 0;

    if (context->fileSystem->state == FS_VOL_STATE_ONLINE_RDONLY) {
        return FS_ERR_ROFS;
    }
    attrp = &(resp->attr[0]);
    inode = NULL;

    FS_TRACE(context, FS_TRACE_REQ,
             ("%s xid 0x%x create snap %lld inum %lld gen %d name %s",
             fs_protoStr[context->reqHdr->proto],
             context->reqHdr->clientId,
             FS_HANDLE_GET_SNAP(req->parentHandle.inum),
             FS_HANDLE_GET_INUM(req->parentHandle.inum),
             req->parentHandle.gen,
             req->fileName));

    if (fs_handleIsSnapMagic(&req->parentHandle))
    {
        return FS_ERR_ACCESS;
    }
    else
    {
        parentInode = I_GET_EXCL(FS_HANDLE_GET_SNAP(req->parentHandle.inum),
                                 FS_HANDLE_GET_INUM(req->parentHandle.inum));
    }
    if (parentInode == NULL)
    {
        return FS_ERR_STALE;
    }
    if (!fs_handleIsValid(context, &req->parentHandle, parentInode))
    {
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_STALE;
    }
    if (!FS_I_IS_DIR(parentInode->dinode.attrs))
    {
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_NOTDIR;
    }

    fs_getRespAttrs(context, parentInode, FS_HANDLE_GET_INUM(req->parentHandle.inum),
                    FS_PRE_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);

    rc = fs_nameIsValid(req->fileName, req->nameLen);

    if (!fs_permissionOkay(context, parentInode, FS_PERM_DIRENT))
    {
        rc = FS_ERR_ACCESS;
    }
    if (context->flags & FS_CONTEXT_ENOSPC_CONDITION)
    {
        rc = FS_ERR_NOSPC;
    }
    if (rc != NFX_OK)
    {
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_inodeReleaseExclusive(context, parentInode);
        return rc;
    }

    inum = fs_dir_lookupent (context,
                                parentInode,
                                req->fileName,
                                req->nameLen);
    if (inum != 0)
    {
        // Name already exists.

        if (req->flags & FS_CREATE_GUARDED)
        {
            fs_getRespAttrs(context, parentInode,
                            FS_HANDLE_GET_INUM(req->parentHandle.inum),
                            FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
            fs_inodeReleaseExclusive(context, parentInode);
            return FS_ERR_EXIST;
        }

        FS_LOCK_ORDER_SET_PARENT(parentInode->inum);
        inode = I_GET_EXCL( FS_HANDLE_GET_SNAP(req->parentHandle.inum),
                            inum);
        if (inode == NULL)
        {
            fs_error(context, FS_INTERNAL_ERR, "i_get");
        }
        else
        {
            if (!FS_I_IS_REG (inode->dinode.attrs))
            {
                fs_getRespAttrs(context, parentInode,
                                FS_HANDLE_GET_INUM(req->parentHandle.inum),
                                FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs,
                                TRUE);
                fs_inodeReleaseExclusive(context, parentInode);
                fs_inodeReleaseExclusive(context, inode);
                return FS_ERR_EXIST;
            }

            if (req->flags & FS_CREATE_EXCLUSIVE)
            {
                char    *verifPtr = (char *) &req->u.verifier;

                if (memcmp(verifPtr,
                           (char *) (&inode->dinode.ctime),
                           sizeof(fs_Time)))
                {
                    fs_getRespAttrs(context, parentInode,
                                    FS_HANDLE_GET_INUM(req->parentHandle.inum),
                                    FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs,
                                    TRUE);
                    fs_inodeReleaseExclusive(context, parentInode);
                    fs_inodeReleaseExclusive(context, inode);
                    return FS_ERR_EXIST;
                }
            }

            // Check write perms on file.
            if (!fs_permissionOkay(context, inode, FS_PERM_WRITE))
            {
                fs_getRespAttrs(context, parentInode,
                                FS_HANDLE_GET_INUM(req->parentHandle.inum),
                                FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs,
                                TRUE);
                fs_inodeReleaseExclusive(context, parentInode);
                fs_inodeReleaseExclusive(context, inode);
                return FS_ERR_ACCESS;
            }

            if (inode->dinode.size > 0)
            {
                fs_truncateFile(context, inode, 0, FALSE, &list1);
                inode->dinode.size = 0;
            }

            inode->dinode.ctime = fs_getTimeOfDay();

            fs_doSetAttr(context, &(req->u.attr), inode, &list2);

            fs_inodeWrite(context, inode, INO_WRITE_LOG);

            fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->parentHandle.inum),
                            FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);

            fs_getRespAttrs(context, parentInode,
                            FS_HANDLE_GET_INUM(req->parentHandle.inum),
                            FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);

            fs_inodeReleaseExclusive(context, parentInode);


            fs_handleInit(context->fileSystem,
                          &resp->fileHandle,
                          FS_HANDLE_GET_SNAP(req->parentHandle.inum),
                          inum,
                          inode->dinode.genNum,
                          req->parentHandle.mntPnt);

            rc = NFX_OK;
        }
    }
    else
    {

        /*
         * We need to create the file. Account against quotas first.
         * We only account for inodes at this point. If the directory
         * needs to be expanded, directory entry add routine will
         * take care of that. We will not enforce quota limits against
         * directory block allocation, but we will certainly charge
         * for it, so that the next time around we prohibit consumption
         */

        quotaReserved = fs_getQuotaReservation(context, context->fileSystem,
                                               parentInode, 1,
                                               FS_COUNT_INODES,
                                               req->hdr.uid, req->hdr.gid,
                                               0, &list1);
        /*
         * We don't have enough quota limits to create the file, return
         * failure
         */

        if (quotaReserved == 0) {

            fs_freeQuotaContextList(context, context->fileSystem, list1);
            fs_getRespAttrs(context, parentInode,
                            FS_HANDLE_GET_INUM(req->parentHandle.inum),
                            FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
            fs_inodeReleaseExclusive(context, parentInode);
            return FS_ERR_DQUOT;
        }
                                               
        
        // File doesn't already exist. Create it.
        rc = NFX_OK;
        mode = (req->u.attr.mode & ~FS_IFMT_MASK) | FS_IFMT_REG;
        inode = fs_allocateFileInode(context, parentInode, mode, req->hdr.uid, req->hdr.gid,
                                     parentInode->dinode.fsetNum);
        if (inode == NULL)
        {
            fs_error(context, FS_INTERNAL_ERR, "ialloc");
        }
        inum = inode->inum;
		if (list1)	{
	        list1->newInode = inode;
		}
        if (req->flags & FS_CREATE_EXCLUSIVE)
        {
            char    *verifPtr = (char *) &req->u.verifier;

            inode->dinode.atime = fs_getTimeOfDay();

            // Borrow ctime for the verifier.
            memcpy(&inode->dinode.ctime,
                   verifPtr,
                    sizeof(inode->dinode.ctime));
        }
        else
        {
            inode->dinode.ctime = fs_getTimeOfDay();
            inode->dinode.mtime = inode->dinode.ctime;
            inode->dinode.atime = inode->dinode.ctime;

            // Set requestors attrs.
            fs_doSetAttr(context, &req->u.attr, inode, &list2);
        }

        fs_inodeWrite(context, inode, INO_WRITE_LOG);

        // Insert dirent.
		parentNumBlocks = parentInode->dinode.numblocks;
        rc = fs_dir_addent( context,
                            parentInode,
                            req->fileName,
                            req->nameLen,
                            inum);
        status = fs_addLink(context, 0, inode, parentInode->inum);
        if (status != FS_ERR_NONE) {
            rc = NFX_ERR;
        }
/* TODO
        if (rc != FS_BT_INSERT_OK &&
            rc != FS_BT_INSERT_DUP)
        {
            fs_error(context, FS_INTERNAL_ERR, "add dirent");
            return FS_ERR_IO;
        }
*/
		//Check if directory size has changed
		parentNumBlocks = parentInode->dinode.numblocks - parentNumBlocks;
		if (parentNumBlocks != 0)	{
			fs_QuotaAdjustmentNoCheck(	context,
										context->fileSystem,
										parentInode,
										NULL,
										parentNumBlocks,
										FS_COUNT_BLOCKS,
										parentInode->sd->u.nfs.uid,
										parentInode->sd->u.nfs.gid,
										&list2);
		}
				
        parentInode->dinode.ctime = fs_getTimeOfDay();
        parentInode->dinode.version++;
        parentInode->dinode.mtime = parentInode->dinode.ctime;

        fs_inodeWrite(context, parentInode, INO_WRITE_LOG);

        fs_handleInit(context->fileSystem,
                      &resp->fileHandle,
                      FS_HANDLE_GET_SNAP(req->parentHandle.inum),
                      inum,
                      inode->dinode.genNum,
                      req->parentHandle.mntPnt);

        fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->parentHandle.inum),
                        FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
    }

    /*
     * If we did any quota reservations, commit them.
	 * We commit the directory size change list first since it is less likely
	 * to exist. Therefore, we prevent the unnecessary hold of quota inodes
     */

    if (list2) {
        fs_commitContextList(context, context->fileSystem, list2, TRUE);
        fs_freeQuotaContextList(context, context->fileSystem, list2);
    }       
    if (list1) {
        fs_commitContextList(context, context->fileSystem, list1, FALSE);
        fs_freeQuotaContextList(context, context->fileSystem, list1);
    }       
    
    rc = fs_logCommitEntry(context);
    if (rc != NFX_OK)
    {
        fs_error(context, FS_INTERNAL_ERR, "commit");
    }

    return NFX_OK;

} /* fs_createThread */


/**
 * Handle the MKDIR request.
 */
int32
fs_mkdirThread(
    fs_Context      *context)
{
    fs_MsgMkdirResp     *resp;
    fs_MsgMkdirReq      *req;
    fs_Inode            *inode, *parentInode;
    fs_Attr             *attrp;
    uint64              inum;
    int32               rc;
    fs_QuotaContextList_t   *list1 = NULL, *list2 = NULL;
    fs_QuotaContextList_t   *list3 = NULL, *list4 = NULL;
    uint64              quotaReserved = 0;
    fs_Status           status;
	int64				parentNumBlocks;
    uint16              mode;

    resp = (fs_MsgMkdirResp *) context->respHdr;
    req = (fs_MsgMkdirReq *) context->reqHdr;
    context->respSize = sizeof(fs_MsgMkdirResp);
    resp->numAttrs = 0;

    if (context->fileSystem->state == FS_VOL_STATE_ONLINE_RDONLY) {
        return FS_ERR_ROFS;
    }
    attrp = &(resp->attr[0]);
    inode = NULL;

    FS_TRACE(context, FS_TRACE_REQ,
             ("%s xid 0x%x mkdir snap %lld inum %lld gen %d name %s",
             fs_protoStr[context->reqHdr->proto],
             context->reqHdr->clientId,
             FS_HANDLE_GET_SNAP(req->parentHandle.inum),
             FS_HANDLE_GET_INUM(req->parentHandle.inum),
             req->parentHandle.gen,
             req->dirName));

    if (fs_handleIsSnapMagic(&req->parentHandle))
    {
        return FS_ERR_ACCESS;
    }
    else
    {
        parentInode = I_GET_EXCL(   FS_HANDLE_GET_SNAP(req->parentHandle.inum),
                                    FS_HANDLE_GET_INUM(req->parentHandle.inum));
    }
    if (parentInode == NULL)
    {
        return FS_ERR_STALE;
    }
    if (!fs_handleIsValid(context, &req->parentHandle, parentInode))
    {
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_STALE;
    }
    if (!FS_I_IS_DIR(parentInode->dinode.attrs))
    {
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_NOTDIR;
    }

    fs_getRespAttrs(context, parentInode, FS_HANDLE_GET_INUM(req->parentHandle.inum),
                    FS_PRE_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);

    if (fs_nameIsDotOrDotDot(req->dirName))
    {
        // Do this first so 'mkdir -p' will behave correctly
        // (on HPUX, for example)
        rc = FS_ERR_EXIST;
    }
    else if ((rc = fs_nameIsValid(req->dirName, req->nameLen)) != NFX_OK)
    {
    }
    else if (parentInode->dinode.linkCnt == FS_INODE_LINK_MAX)
    {
        rc = FS_ERR_MLINK;
    }
    else if (!fs_permissionOkay(context, parentInode, FS_PERM_DIRENT))
    {
        rc = FS_ERR_ACCESS;
    }
	else if (context->flags & FS_CONTEXT_ENOSPC_CONDITION)
    {
		rc = FS_ERR_NOSPC;
	}
    if (rc != NFX_OK)
    {
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_inodeReleaseExclusive(context, parentInode);
        return rc;
    }

    inum = fs_dir_lookupent(    context,
                                parentInode,
                                req->dirName,
                                req->nameLen);
    if (inum != 0)
    {
        // Name already exists.
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_EXIST;
    }
    else
    {
        /*
         * We need to create the file. Account against quotas first.
         * We only account for inodes at this point. If the directory
         * needs to be expanded, directory entry add routine will
         * take care of that. We will not enforce quota limits against
         * directory block allocation, but we will certainly charge
         * for it, so that the next time around we prohibit consumption
         */

        quotaReserved = fs_getQuotaReservation(context, context->fileSystem,
                                               parentInode, 1,
                                               FS_COUNT_INODES,
                                               req->hdr.uid, req->hdr.gid,
                                               0, &list1);
        if (quotaReserved != 0) {
            quotaReserved = fs_getQuotaReservation(context, context->fileSystem,
                                                   parentInode, 1,
                                                   FS_COUNT_BLOCKS,
                                                   req->hdr.uid, req->hdr.gid,
                                                   0, &list2);
        }
        
        /*
         * We don't have enough quota limits to create the directory, return
         * failure
         */

        if (quotaReserved == 0) {
            if (list1)  {
                if (list1->availCount != 0) {
                    fs_trimQuotaReservation(context, context->fileSystem, 
                                            list1, list1->availCount);
                }
                fs_freeQuotaContextList(context, context->fileSystem, list1);
            }
            if (list2)  {
                fs_freeQuotaContextList(context, context->fileSystem, list2);
            }
            fs_getRespAttrs(context, parentInode,
                            FS_HANDLE_GET_INUM(req->parentHandle.inum),
                            FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
            fs_inodeReleaseExclusive(context, parentInode);
            return FS_ERR_DQUOT;
        }
        // Create directory.
        rc = NFX_OK;
        mode = (req->dirAttr.mode & ~FS_IFMT_MASK) | FS_IFMT_DIR;
        inode = fs_allocateDirectoryInode(context,
                                          parentInode,
                                          mode,
                                          req->hdr.uid,
                                          req->hdr.gid,
                                          parentInode->dinode.fsetNum);
        if (inode == NULL)
        {
            fs_error(context, FS_INTERNAL_ERR, "inode alloc");
        }
        inum = inode->inum;
        if (list1)  {
            list1->newInode = inode;
        }

        inode->dinode.ctime = fs_getTimeOfDay();
        inode->dinode.mtime = inode->dinode.ctime;
        inode->dinode.atime = inode->dinode.ctime;

        // Set requestors attrs. -- much of this is done already
        fs_doSetAttr(context, &req->dirAttr, inode, &list3);

        fs_inodeWrite(context, inode, INO_WRITE_LOG);
                
        // Insert dirent.
		parentNumBlocks = parentInode->dinode.numblocks;
        rc = fs_dir_addent( context,
                            parentInode,
                            req->dirName,
                            req->nameLen,
                            inum);
        status = fs_addLink(context, 0, inode, parentInode->inum);
        if (status != FS_ERR_NONE) {
            rc = NFX_ERR;
        }
/* TODO
        if (rc != FS_BT_INSERT_OK &&
            rc != FS_BT_INSERT_DUP)
        {
            fs_error(context, FS_INTERNAL_ERR, "add dirent");
            return FS_ERR_IO;
        }
*/

		//Check if directory size has changed
		parentNumBlocks = parentInode->dinode.numblocks - parentNumBlocks;
		if (parentNumBlocks != 0)	{
			fs_QuotaAdjustmentNoCheck(	context,
										context->fileSystem,
										parentInode,
										NULL,
										parentNumBlocks,
										FS_COUNT_BLOCKS,
										parentInode->sd->u.nfs.uid,
										parentInode->sd->u.nfs.gid,
										&list4);
		}

        // Incr parent link count (for .. of this dir).
        parentInode->dinode.linkCnt += 1;

        parentInode->dinode.ctime = fs_getTimeOfDay();
        parentInode->dinode.version++;
        parentInode->dinode.mtime = parentInode->dinode.ctime;

        fs_inodeWrite(context, parentInode, INO_WRITE_LOG);

        fs_handleInit(context->fileSystem,
                      &resp->dirHandle,
                      FS_HANDLE_GET_SNAP(req->parentHandle.inum),
                      inum,
                      inode->dinode.genNum,
                      req->parentHandle.mntPnt);

        fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->parentHandle.inum),
                        FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
    }

    /*
     * If we did any quota reservations, commit them.
	 * We commit the directory size change list first since it is less likely
	 * to exist. Therefore, we prevent the unnecessary hold of quota inodes
     */
    if (list4) {
        fs_commitContextList(context, context->fileSystem, list4, TRUE);
        fs_freeQuotaContextList(context, context->fileSystem, list4);
    }       
    if (list1) {
        fs_commitContextList(context, context->fileSystem, list1, FALSE);
        fs_freeQuotaContextList(context, context->fileSystem, list1);
    }       
    if (list2) {
        fs_commitContextList(context, context->fileSystem, list2, TRUE);
        fs_freeQuotaContextList(context, context->fileSystem, list2);
    }       
    if (list3) {
        fs_commitContextList(context, context->fileSystem, list3, TRUE);
        fs_freeQuotaContextList(context, context->fileSystem, list3);
    }       

    rc = fs_logCommitEntry(context);
    if (rc != NFX_OK)
    {
        fs_error(context, FS_INTERNAL_ERR, "commit");
    }

    return NFX_OK;

} /* fs_mkdirThread */


/*
 * !!! Ensure that we don't have a maxpath larger than FS_BLOCK_SIZE.
 * This is for simplicity only. The largest path we
 * have seen from clients is 4k which is smaller than FS_BLOCK_SIZE.
 * If we support smaller block size, then we will have to perform
 * multiple allocations or preferably we seperate allocation from
 * buffer cache, so that we can perform allocation upfront of whatever
 * size we want.
 */

NFX_ASSERT_STATIC(FS_PATH_MAX <= FS_BLOCK_SIZE);
/**
 * Handle the SYMLINK request.
 */
int32
fs_symLinkThread(
    fs_Context      *context)
{
    fs_MsgSymLinkResp   *resp;
    fs_MsgSymLinkReq    *req;
    fs_Inode            *inode, *parentInode;
    fs_Attr             *attrp;
    fs_Buf              *buf;
    uint64              inum;
    int32               rc;
    char                *linkToNamePtr;
    fs_QuotaContextList_t   *list1 = NULL;
    fs_QuotaContextList_t   *list2 = NULL;
    fs_QuotaContextList_t   *list3 = NULL;
    uint64                  quotaReserved1 = 0;
    uint64                  quotaReserved2 = 0;
    fs_Status               status;
	int64					parentNumBlocks;

    resp = (fs_MsgSymLinkResp *) context->respHdr;
    req = (fs_MsgSymLinkReq *) context->reqHdr;
    context->respSize = sizeof(fs_MsgSymLinkResp);
    resp->numAttrs = 0;

    if (context->fileSystem->state == FS_VOL_STATE_ONLINE_RDONLY) {
        return FS_ERR_ROFS;
    }
    attrp = &(resp->attr[0]);
    inode = NULL;
    linkToNamePtr = req->linkName + req->linkNameLen + 1;

    FS_TRACE(context, FS_TRACE_REQ,
             ("%s xid 0x%x symlink snap %lld inum %lld gen %d name %s "
              "linkto %s",
             fs_protoStr[context->reqHdr->proto],
             context->reqHdr->clientId,
             FS_HANDLE_GET_SNAP(req->parentHandle.inum),
             FS_HANDLE_GET_INUM(req->parentHandle.inum),
             req->parentHandle.gen,
             req->linkName,
             linkToNamePtr));

    if (fs_handleIsSnapMagic(&req->parentHandle))
    {
        return FS_ERR_ACCESS;
    }
    else
    {
        parentInode = I_GET_EXCL(   FS_HANDLE_GET_SNAP(req->parentHandle.inum),
                                    FS_HANDLE_GET_INUM(req->parentHandle.inum));
    }
    if (parentInode == NULL)
    {
        return FS_ERR_STALE;
    }
    if (!fs_handleIsValid(context, &req->parentHandle, parentInode))
    {
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_STALE;
    }
    if (!FS_I_IS_DIR(parentInode->dinode.attrs))
    {
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_NOTDIR;
    }

    fs_getRespAttrs(context, parentInode, FS_HANDLE_GET_INUM(req->parentHandle.inum),
                    FS_PRE_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);

    rc = fs_nameIsValid(req->linkName, req->linkNameLen);

    if (!fs_permissionOkay(context, parentInode, FS_PERM_DIRENT))
    {
        rc = FS_ERR_ACCESS;
    }
	if (context->flags & FS_CONTEXT_ENOSPC_CONDITION)
    {
		rc = FS_ERR_NOSPC;
	}
    if (rc != NFX_OK)
    {
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_inodeReleaseExclusive(context, parentInode);
        return rc;
    }

    inum = fs_dir_lookupent(    context,
                                parentInode,
                                req->linkName,
                                req->linkNameLen);

    if (inum != 0)
    {
        // Name already exists.
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_EXIST;
    }

    /*
     * We need to allocate an inode and potentially a block.
     * Do the quota reservation upfront.
     */

    quotaReserved1 = fs_getQuotaReservation(context, context->fileSystem,
                                            parentInode, 1, FS_COUNT_INODES,
                                            req->hdr.uid, req->hdr.gid,
                                            0, &list1);
    if (quotaReserved1 == 0) {
        fs_freeQuotaContextList(context, context->fileSystem, list1);
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_DQUOT;
    }

    if (req->linkToNameLen > FS_INODE_INLINE_LEN) {
        quotaReserved2 = fs_getQuotaReservation(context, context->fileSystem,
                                            parentInode, 1, FS_COUNT_BLOCKS,
                                            req->hdr.uid, req->hdr.gid,
                                            0, &list2);
        if (quotaReserved2 == 0) {
            fs_trimQuotaReservation(context, context->fileSystem, list1, 1);
            fs_freeQuotaContextList(context, context->fileSystem, list1);
            fs_freeQuotaContextList(context, context->fileSystem, list2);
            fs_getRespAttrs(context, parentInode,
                            FS_HANDLE_GET_INUM(req->parentHandle.inum),
                            FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
            fs_inodeReleaseExclusive(context, parentInode);
            return FS_ERR_DQUOT;
        }
    }

    rc = NFX_OK;
    inode = fs_allocateFileInode(context,
                                 parentInode,
                                 FS_IFMT_SYML | FS_IMODE_RWXUGO,
                                 req->hdr.uid,
                                 req->hdr.gid,
                                 parentInode->dinode.fsetNum);

    if (inode == NULL)
    {
        fs_error(context, FS_INTERNAL_ERR, "ialloc");
    }
    inum = inode->inum;
    if (list1)  {
        list1->newInode = inode;
    }

    inode->dinode.ctime = fs_getTimeOfDay();
    inode->dinode.mtime = inode->dinode.ctime;
    inode->dinode.atime = inode->dinode.ctime;
    
    // Insert dirent.
	parentNumBlocks = parentInode->dinode.numblocks;
    rc = fs_dir_addent( context,
                        parentInode,
                        req->linkName,
                        req->linkNameLen,
                        inum);
    status = fs_addLink(context, 0, inode, parentInode->inum);
    if (status != FS_ERR_NONE) {
        rc = NFX_ERR;
    }
/* TODO
    if (rc != FS_BT_INSERT_OK &&
        rc != FS_BT_INSERT_DUP)
    {
        fs_error(context, FS_INTERNAL_ERR, "add dirent");
        return FS_ERR_IO;
    }
*/
	//Check if directory size has changed
	parentNumBlocks = parentInode->dinode.numblocks - parentNumBlocks;
	if (parentNumBlocks != 0)	{
		fs_QuotaAdjustmentNoCheck(	context,
									context->fileSystem,
									parentInode,
									NULL,
									parentNumBlocks,
									FS_COUNT_BLOCKS,
									parentInode->sd->u.nfs.uid,
									parentInode->sd->u.nfs.gid,
									&list3);
	}
    parentInode->dinode.ctime = fs_getTimeOfDay();
    parentInode->dinode.version++;
    parentInode->dinode.mtime = parentInode->dinode.ctime;
    fs_inodeWrite(context, parentInode, INO_WRITE_LOG);

    FS_ASSERT(req->linkToNameLen <= FS_PATH_MAX);
    inode->dinode.size = req->linkToNameLen;
    if (inode->dinode.size <= FS_INODE_INLINE_LEN)
    {
        memcpy(inode->dinode.u.data, linkToNamePtr, inode->dinode.size);
    }
    else
    {
        // Allocate a block to store the pathname, its too long
        // to fit in the inode.
        fs_emaproot_init(context, inode);
        buf = O_CREATE(inode, 0);
        memcpy(buf->u.data, linkToNamePtr, inode->dinode.size);
        FS_BUF_ZERO_ALL(buf);
        FS_BUF_WRITE(buf, 0, inode->dinode.size);
    }

    fs_inodeWrite(context, inode, INO_WRITE_LOG);

    fs_handleInit(context->fileSystem,
                  &resp->fileHandle,
                  FS_HANDLE_GET_SNAP(req->parentHandle.inum),
                  inum,
                  inode->dinode.genNum,
                  req->parentHandle.mntPnt);

    fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->parentHandle.inum),
                    FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);
    fs_getRespAttrs(context, parentInode, FS_HANDLE_GET_INUM(req->parentHandle.inum),
                    FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);

    /*
     * If we did any quota reservations, commit them.
	 * We commit the directory size change list first since it is less likely
	 * to exist. Therefore, we prevent the unnecessary hold of quota inodes
     */
    if (list3 != NULL) {
        fs_commitContextList(context, context->fileSystem, list3, TRUE);
        fs_freeQuotaContextList(context, context->fileSystem, list3);
    }
    if (list1 != NULL) {
        fs_commitContextList(context, context->fileSystem, list1, TRUE);
        fs_freeQuotaContextList(context, context->fileSystem, list1);
    }
    if (list2 != NULL) {
        fs_commitContextList(context, context->fileSystem, list2, FALSE);
        fs_freeQuotaContextList(context, context->fileSystem, list2);
    }

    rc = fs_logCommitEntry(context);
    if (rc != NFX_OK)
    {
        fs_error(context, FS_INTERNAL_ERR, "commit");
    }

    return NFX_OK;

} /* fs_symLinkThread */


/**
 * Handle the MKNOD request.
 */
int32
fs_mknodThread(
    fs_Context      *context)
{
    fs_MsgMknodResp     *resp;
    fs_MsgMknodReq      *req;
    fs_Inode            *inode, *parentInode;
    fs_Attr             *attrp;
    uint64              inum;
    int32               rc, mode;
    fs_QuotaContextList_t   *list1 = NULL, *list2 = NULL, *list3 = NULL;
    uint64              quotaReserved1 = 0;
    fs_Status           status;
    int64		parentNumBlocks;

    resp = (fs_MsgMknodResp *) context->respHdr;
    req = (fs_MsgMknodReq *) context->reqHdr;
    resp->numAttrs = 0;

    context->respSize = sizeof(fs_MsgMknodResp);
    if (context->fileSystem->state == FS_VOL_STATE_ONLINE_RDONLY) {
        return FS_ERR_ROFS;
    }
    attrp = &(resp->attr[0]);
    inode = NULL;

    FS_TRACE(context, FS_TRACE_REQ,
             ("%s xid 0x%x mknod snap %lld inum %lld gen %d name %s",
             fs_protoStr[context->reqHdr->proto],
             context->reqHdr->clientId,
             FS_HANDLE_GET_SNAP(req->parentHandle.inum),
             FS_HANDLE_GET_INUM(req->parentHandle.inum),
             req->parentHandle.gen,
             req->fileName));

    if (fs_handleIsSnapMagic(&req->parentHandle))
    {
        return FS_ERR_ACCESS;
    }
    else
    {
        parentInode = I_GET_EXCL(   FS_HANDLE_GET_SNAP(req->parentHandle.inum),
                                    FS_HANDLE_GET_INUM(req->parentHandle.inum));
    }
    if (parentInode == NULL)
    {
        return FS_ERR_STALE;
    }
    if (!fs_handleIsValid(context, &req->parentHandle, parentInode))
    {
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_STALE;
    }
    if (!FS_I_IS_DIR(parentInode->dinode.attrs))
    {
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_NOTDIR;
    }

    fs_getRespAttrs(context, parentInode, FS_HANDLE_GET_INUM(req->parentHandle.inum),
                    FS_PRE_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);

    rc = fs_nameIsValid(req->fileName, req->nameLen);

    if (!fs_permissionOkay(context, parentInode, FS_PERM_DIRENT))
    {
        rc = FS_ERR_ACCESS;
    }
	if (context->flags & FS_CONTEXT_ENOSPC_CONDITION)
    {
		rc = FS_ERR_NOSPC;
	}
    switch(req->attr.type)
    {
    case FS_FTYPE_BLK:
        mode = FS_IFMT_BLK;
        break;
    case FS_FTYPE_CHR:
        mode = FS_IFMT_CHR;
        break;
    case FS_FTYPE_SOCK:
        mode = FS_IFMT_SOCK;
        break;
    case FS_FTYPE_FIFO:
        mode = FS_IFMT_FIFO;
        break;
    default:
        mode = 0;
        rc = FS_ERR_BADTYPE;
    }
    if (rc != NFX_OK)
    {
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_inodeReleaseExclusive(context, parentInode);
        return rc;
    }

    inum = fs_dir_lookupent(    context,
                                parentInode,
                                req->fileName,
                                req->nameLen);

    if (inum != 0)
    {
        // Name already exists.
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_EXIST;
    }

    /*
     * We need to allocate an inode.
     * Do the quota reservation upfront.
     */

    quotaReserved1 = fs_getQuotaReservation(context, context->fileSystem,
                                            parentInode, 1, FS_COUNT_INODES,
                                            req->hdr.uid, req->hdr.gid,
                                            0, &list1);
    if (quotaReserved1 == 0) {
        fs_freeQuotaContextList(context, context->fileSystem, list1);
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_DQUOT;
    }

    inode = fs_allocateFileInode(context, parentInode, mode,
                                 req->hdr.uid, req->hdr.gid,
                                 parentInode->dinode.fsetNum);
    if (inode == NULL)
    {
        fs_error(context, FS_INTERNAL_ERR, "ialloc");
    }
    inum = inode->inum;
    if (list1)  {
        list1->newInode = inode;
    }
    inode->dinode.ctime = fs_getTimeOfDay();
    inode->dinode.mtime = inode->dinode.ctime;
    inode->dinode.atime = inode->dinode.ctime;
    
    // Insert dirent.
	parentNumBlocks = parentInode->dinode.numblocks;
    rc = fs_dir_addent( context,
                        parentInode,
                        req->fileName,
                        req->nameLen,
                        inum);
    status = fs_addLink(context, 0, inode, parentInode->inum);
    if (status != FS_ERR_NONE) {
        rc = NFX_ERR;
    }
/* TODO
    if (rc != FS_BT_INSERT_OK &&
        rc != FS_BT_INSERT_DUP)
    {
        fs_error(context, FS_INTERNAL_ERR, "add dirent");
        return FS_ERR_IO;
    }
*/
	//Check if directory size has changed
	parentNumBlocks = parentInode->dinode.numblocks - parentNumBlocks;
	if (parentNumBlocks != 0)	{
		fs_QuotaAdjustmentNoCheck(	context,
									context->fileSystem,
									parentInode,
									NULL,
									parentNumBlocks,
									FS_COUNT_BLOCKS,
									parentInode->sd->u.nfs.uid,
									parentInode->sd->u.nfs.gid,
									&list2);
	}

    parentInode->dinode.ctime = fs_getTimeOfDay();
    parentInode->dinode.version++;
    parentInode->dinode.mtime = parentInode->dinode.ctime;
    fs_inodeWrite(context, parentInode, INO_WRITE_LOG);

    fs_doSetAttr(context, &(req->attr), inode, &list3);
    inode->dinode.size = 0;

    if (req->attr.type == FS_FTYPE_BLK ||
        req->attr.type == FS_FTYPE_CHR)
    {
        inode->dinode.u.devNum.major = req->devNum.major;
        inode->dinode.u.devNum.minor = req->devNum.minor;
    }

    fs_inodeWrite(context, inode, INO_WRITE_LOG);

    fs_handleInit(context->fileSystem,
                  &resp->fileHandle,
                  FS_HANDLE_GET_SNAP(req->parentHandle.inum),
                  inum,
                  inode->dinode.genNum,
                  req->parentHandle.mntPnt);
    
    fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->parentHandle.inum),
                    FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);
    fs_getRespAttrs(context, parentInode,
                    FS_HANDLE_GET_INUM(req->parentHandle.inum),
                    FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);

    /*
     * If we did any quota reservations, commit them.
	 * We commit the directory size change list first since it is less likely
	 * to exist. Therefore, we prevent the unnecessary hold of quota inodes
     */

    if (list2 != NULL) {
        fs_commitContextList(context, context->fileSystem, list2, TRUE);
        fs_freeQuotaContextList(context, context->fileSystem, list2);
    }
    if (list1 != NULL) {
        fs_commitContextList(context, context->fileSystem, list1, FALSE);
        fs_freeQuotaContextList(context, context->fileSystem, list1);
    }
    if (list3 != NULL) {
        fs_commitContextList(context, context->fileSystem, list3, FALSE);
        fs_freeQuotaContextList(context, context->fileSystem, list3);
    }

    rc = fs_logCommitEntry(context);
    if (rc != NFX_OK)
    {
        fs_error(context, FS_INTERNAL_ERR, "commit");
    }

    return NFX_OK;

} /* fs_mknodThread */


/**
 * Utility function for removing non-directory inodes.
 */
void
fs_removeNonDirNode(
    fs_Context  			*context,
    fs_Inode    			*parentInode,
    fs_Inode    			*inode,
    char        			*name,
    int32       			nameLen,
	fs_QuotaContextList_t	**rlist1,
	fs_QuotaContextList_t	**rlist2,
	fs_QuotaContextList_t	**rlist3,
    boolean     			*wasDestroyed)
{
    int32   rc, destroyInode;
    fs_QuotaContextList_t   *list1 = NULL, *list2 = NULL, *list3 = NULL;
    fs_Status       status;
	int64	parentNumBlocks;

    FS_ASSERT(fs_lockOwnedExclusive(context, &parentInode->lock));


    if (inode->dinode.linkCnt == 1)
    {
        if (FS_I_IS_REG(inode->dinode.attrs))
        {
            if (inode->dinode.size > 0)
{
                fs_truncateFile(context, inode, 0, FALSE, &list1);
            }
        }
        else if (FS_I_IS_SYML(inode->dinode.attrs))
        {
            if (inode->dinode.size > FS_INODE_INLINE_LEN)
            {
                fs_truncateFile(context, inode, 0, FALSE, &list1);
            }
        }


        destroyInode = 1;
    }
    else
    {
        inode->dinode.linkCnt -= 1;
        fs_inodeWrite(context, inode, INO_WRITE_LOG);
        destroyInode = 0;
    }

	parentNumBlocks = parentInode->dinode.numblocks;
    rc = fs_dir_removeent(context,
                          parentInode,
                          name,
                          nameLen);
    status = fs_removeLink(context, 0, inode, parentInode->inum);
    if (status != FS_ERR_NONE) {
        rc = NFX_ERR;
    }
/* TODO
    if (rc != FS_BT_REMOVE_OK &&
        rc != FS_BT_REMOVE_OK_EMPTY)
    {
        fs_error(context, FS_INTERNAL_ERR, "btree remove non-dir ent");
    }
*/
	//Check if directory size has changed
	parentNumBlocks = parentInode->dinode.numblocks - parentNumBlocks;
	if (parentNumBlocks != 0)	{
		fs_QuotaAdjustmentNoCheck(	context,
									context->fileSystem,
									parentInode,
									NULL,
									parentNumBlocks,
									FS_COUNT_BLOCKS,
									parentInode->sd->u.nfs.uid,
									parentInode->sd->u.nfs.gid,
									&list2);
	}

    parentInode->dinode.ctime = fs_getTimeOfDay();
    parentInode->dinode.mtime = parentInode->dinode.ctime;
    parentInode->dinode.version++;
    fs_inodeWrite(context, parentInode, INO_WRITE_LOG);

    if (destroyInode)
    {

        /*
         * Free the quota resources held by this inode
         */

        (void)fs_freeQuotaResources(context, context->fileSystem,
                                     parentInode, 1, FS_COUNT_INODES,
                                     inode->sd->u.nfs.uid,
                                     inode->sd->u.nfs.gid,
                                     &list3);
		
	    /*
	     * If we did any quota reservations, commit them.
		 * We commit the directory size change list first since it is less
		 * likely to exist. Therefore, we prevent the unnecessary hold of
		 * quota inodes
	     */
	    if (list2 != NULL) {
			if (rlist2 != NULL)	{
				*rlist2 = list2;
			} else {
		        fs_commitContextList(context, context->fileSystem, list2, TRUE);
		        fs_freeQuotaContextList(context, context->fileSystem, list2);
			}
	    }
	    if (list1 != NULL) {
			if (rlist1 != NULL)	{
				*rlist1 = list1;
			} else {
		        fs_commitContextList(context, context->fileSystem, list1, TRUE);
		        fs_freeQuotaContextList(context, context->fileSystem, list1);
			}
	    }
        if (list3) {
            list3->newInode = inode;
			if (rlist3 != NULL)	{
				*rlist3 = list3;
			} else {
		        fs_commitContextList(context, context->fileSystem, list3,FALSE);
		        fs_freeQuotaContextList(context, context->fileSystem, list3);
			}
        }
        fs_deallocateInode(context, inode);
        *wasDestroyed = TRUE;
    }
    else
    {
	    if (list2 != NULL) {
			if (rlist2 != NULL)	{
				*rlist2 = list2;
			} else {
		        fs_commitContextList(context, context->fileSystem, list2,TRUE);
		        fs_freeQuotaContextList(context, context->fileSystem, list2);
			}
	    }
	    if (list1 != NULL) {
			if (rlist1 != NULL)	{
				*rlist1 = list1;
			} else {
		        fs_commitContextList(context, context->fileSystem, list1,FALSE);
		        fs_freeQuotaContextList(context, context->fileSystem, list1);
			}
	    }
        *wasDestroyed = FALSE;
    }
}


/**
 * Handle the REMOVE request.
 */
int32
fs_removeThread(
    fs_Context      *context)
{
    fs_MsgRemoveResp    *resp;
    fs_MsgRemoveReq     *req;
    fs_Inode            *inode, *parentInode;
    fs_Attr             *attrp;
    uint64              inum;
    int32               rc;

    resp = (fs_MsgRemoveResp *) context->respHdr;
    req = (fs_MsgRemoveReq *) context->reqHdr;
    resp->numAttrs = 0;
    attrp = &resp->attr[0];

    context->respSize = sizeof(fs_MsgRemoveResp);
    if (context->fileSystem->state == FS_VOL_STATE_ONLINE_RDONLY) {
        return FS_ERR_ROFS;
    }

    FS_TRACE(context, FS_TRACE_REQ,
             ("%s xid 0x%x remove snap %lld inum %lld gen %d name %s",
             fs_protoStr[context->reqHdr->proto],
             context->reqHdr->clientId,
             FS_HANDLE_GET_SNAP(req->parentHandle.inum),
             FS_HANDLE_GET_INUM(req->parentHandle.inum),
             req->parentHandle.gen,
             req->name));

    if (fs_handleIsSnapMagic(&req->parentHandle))
    {
        return FS_ERR_ACCESS;
    }
    else
    {
        parentInode = I_GET_EXCL(   FS_HANDLE_GET_SNAP(req->parentHandle.inum),
                                    FS_HANDLE_GET_INUM(req->parentHandle.inum));
    }
    if (parentInode == NULL)
    {
        return FS_ERR_STALE;
    }
    if (!fs_handleIsValid(context, &req->parentHandle, parentInode))
    {
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_STALE;
    }
    if (!FS_I_IS_DIR(parentInode->dinode.attrs))
    {
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_NOTDIR;
    }

    fs_getRespAttrs(context, parentInode,
                    FS_HANDLE_GET_INUM(req->parentHandle.inum),
                    FS_PRE_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);

    rc = fs_nameIsValid(req->name, req->nameLen);

    if (!fs_permissionOkay(context, parentInode, FS_PERM_DIRENT))
    {
        rc = FS_ERR_ACCESS;
    }
    if (rc != NFX_OK)
    {
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_inodeReleaseExclusive(context, parentInode);
        return rc;
    }

    inum = fs_dir_lookupent(    context,
                                parentInode,
                                req->name,
                                req->nameLen);
    if (inum == 0)
    {
        // File not found.
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_NOENT;
    }
    else
    {
        // Dirent found.
		FS_LOCK_ORDER_SET_PARENT(parentInode->inum);
        inode = I_GET_EXCL( FS_HANDLE_GET_SNAP(req->parentHandle.inum),
                            inum);
        if (inode == NULL)
        {
            fs_error(context, FS_INTERNAL_ERR, "i_get");
        }
        else
        {
            if (FS_I_IS_DIR(inode->dinode.attrs))
            {
                fs_getRespAttrs(context, parentInode,
                                FS_HANDLE_GET_INUM(req->parentHandle.inum),
                                FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs,
                                TRUE);
                fs_inodeReleaseExclusive(context, inode);
                fs_inodeReleaseExclusive(context, parentInode);
                return FS_ERR_ISDIR;
            }

            // TODO - error ret?
            //if (rc != NFX_OK)
            //{
            //  fs_error(context, FS_INTERNAL_ERR, "btree remove");
            //  return FS_ERR_IO;
            //}

            fs_handleInit(context->fileSystem,
                          &resp->fileHandle,
                          FS_HANDLE_GET_SNAP(req->parentHandle.inum),
                          inum,
                          inode->dinode.genNum,
                          req->parentHandle.mntPnt);

            fs_removeNonDirNode(context,
                                parentInode,
                                inode,
                                req->name,
                                req->nameLen,
								NULL,
								NULL,
								NULL,
                                &resp->wasDestroyed);

            fs_getRespAttrs(context, parentInode,
                            FS_HANDLE_GET_INUM(req->parentHandle.inum),
                            FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);

            if (!resp->wasDestroyed)
            {
                // Return the updated attributes for
                // the inode so the NFS server can sync up its
                // attribute cache (link count changed).
                fs_getRespAttrs(context, inode,
                                FS_HANDLE_GET_INUM(req->parentHandle.inum),
                                FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);

            }
#ifdef NEW_ZF
            else
            {
                /* Don't zero fill (cancel previous zero fill records
                 * for this inode).
                 */
                fs_logAddUserZeroRecord(context, inode, 0, 0, 0,
                                        FALSE, FALSE, NULL);
            }
#endif // NEW_ZF
        }
    
        rc = fs_logCommitEntry(context);
        if (rc != NFX_OK)
        {
            fs_error(context, FS_INTERNAL_ERR, "commit");
        }
    }

    return rc;

} /* fs_removeThread */


/**
 * Utility function for removing directory inodes.
 */
void
fs_removeDirNode(
    fs_Context  			*context,
    fs_Inode    			*parentInode,
    fs_Inode    			*inode,
    char        			*name,
    int32       			nameLen,
	fs_QuotaContextList_t	**rlist1,
	fs_QuotaContextList_t	**rlist2,
    fs_QuotaContextList_t   **rlist3)
{
    int32   rc;
    fs_QuotaContextList_t   *list1 = NULL, *list2 = NULL, *list3 = NULL;
    fs_Status               status;
	int64	parentNumBlocks;

    // The proper lock order is inode, then space, and
    // removing the dirent might reclaim space.

	parentNumBlocks = parentInode->dinode.numblocks;
    rc = fs_dir_removeent(  context,
                            parentInode,
                            name,
                            nameLen);
    status = fs_removeLink(context, 0, inode, parentInode->inum);
    if (status != FS_ERR_NONE) {
        rc = NFX_ERR;
    }
/* TODO
    if (rc != FS_BT_REMOVE_OK &&
        rc != FS_BT_REMOVE_OK_EMPTY)
    {
        fs_error(context, FS_INTERNAL_ERR, "btree remove dir ent");
    }
*/

    // Do this before truncating, since it depends on
    // inode size.
    fs_dirhash_destroy_directory(context, inode);

    //Truncate directory file
    fs_truncateFile(context, inode, 0, FALSE, &list1);
 
	//Check if parent directory size has changed
	parentNumBlocks = parentInode->dinode.numblocks - parentNumBlocks;
	if (parentNumBlocks != 0)	{
		fs_QuotaAdjustmentNoCheck(	context,
									context->fileSystem,
									parentInode,
									NULL,
									parentNumBlocks,
									FS_COUNT_BLOCKS,
									parentInode->sd->u.nfs.uid,
									parentInode->sd->u.nfs.gid,
									&list2);
	}
    
    FS_ASSERT(parentInode->dinode.linkCnt > 2);

    //Decrement parent link count (for .. of removed dir).
    parentInode->dinode.linkCnt -= 1;
    parentInode->dinode.ctime = fs_getTimeOfDay();
    parentInode->dinode.mtime = parentInode->dinode.ctime;
    parentInode->dinode.version++;
    fs_inodeWrite(context, parentInode, INO_WRITE_LOG);

    (void)fs_freeQuotaResources(context, context->fileSystem,
                                inode, 1, FS_COUNT_INODES,
                                inode->sd->u.nfs.uid, inode->sd->u.nfs.gid,
                                &list3);
    
   
	/*
	 * We commit the directory size change list first since it is less likely
	 * to exist. Therefore, we prevent the unnecessary hold of quota inodes
	 */ 
    if (list2 != NULL) {
		if (rlist2 != NULL)	{
			*rlist2 = list2;
		} else {
	        fs_commitContextList(context, context->fileSystem, list2, TRUE);
	        fs_freeQuotaContextList(context, context->fileSystem, list2);
		}
    }
    if (list1 != NULL) {
		if (rlist1 != NULL)	{
			*rlist1 = list1;
		} else {
	        fs_commitContextList(context, context->fileSystem, list1, TRUE);
	        fs_freeQuotaContextList(context, context->fileSystem, list1);
		}
    }
    if (list3) {
        list3->newInode = inode;
		if (rlist3 != NULL)	{
			*rlist3 = list3;
		} else {
		    fs_commitContextList(context, context->fileSystem, list3, FALSE);
		    fs_freeQuotaContextList(context, context->fileSystem, list3);
		}
    }
    fs_deallocateInode(context, inode);

} /* fs_removeDirNode */


/**
 * Handle the RMDIR request.
 */
int32
fs_rmdirThread(
    fs_Context      *context)
{
    fs_MsgRmdirResp     		*resp;
    fs_MsgRmdirReq      		*req;
    fs_Inode            		*inode, *parentInode;
    uint64              		inum;
    int32               		rc;
	fs_FileSetIncoreRecord_t	*fset;
	fs_Inode					*fsetInfoInode;
    fs_Attr                     *attrp;
    fs_FileSystem               *core = context->fileSystem;

    resp = (fs_MsgRmdirResp *) context->respHdr;
    req = (fs_MsgRmdirReq *) context->reqHdr;
    resp->numAttrs = 0;
    attrp = &resp->attr[0];

    context->respSize = sizeof(fs_MsgRmdirResp);
    if (core->state == FS_VOL_STATE_ONLINE_RDONLY) {
        return FS_ERR_ROFS;
    }

    FS_TRACE(context, FS_TRACE_REQ,
             ("%s xid 0x%x rmdir snap %lld inum %lld gen %d name %s",
             fs_protoStr[context->reqHdr->proto],
             context->reqHdr->clientId,
             FS_HANDLE_GET_SNAP(req->parentHandle.inum),
             FS_HANDLE_GET_INUM(req->parentHandle.inum),
             req->parentHandle.gen,
             req->name));

    if (fs_handleIsSnapMagic(&req->parentHandle))
    {
        return FS_ERR_ACCESS;
    }
    else
    {
        parentInode = I_GET_EXCL(FS_HANDLE_GET_SNAP(req->parentHandle.inum),
                                 FS_HANDLE_GET_INUM(req->parentHandle.inum));
    }
    if (parentInode == NULL)
    {
        return FS_ERR_STALE;
    }
    if (!fs_handleIsValid(context, &req->parentHandle, parentInode))
    {
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_STALE;
    }
    if (!FS_I_IS_DIR(parentInode->dinode.attrs))
    {
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_NOTDIR;
    }

    fs_getRespAttrs(context, parentInode,
                    FS_HANDLE_GET_INUM(req->parentHandle.inum),
                    FS_PRE_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);

    rc = fs_nameIsValid(req->name, req->nameLen);

    if (!fs_permissionOkay(context, parentInode, FS_PERM_DIRENT))
    {
        rc = FS_ERR_ACCESS;
    }
    if (rc != NFX_OK)
    {
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_inodeReleaseExclusive(context, parentInode);
        return rc;
    }

    inum = fs_dir_lookupent(context,
                            parentInode,
                            req->name,
                            req->nameLen);
    if (inum == 0)
    {
        // Dirent not found.
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_NOENT;
    }
    else
    {
        // Dirent found.
		FS_LOCK_ORDER_SET_PARENT(parentInode->inum);
        inode = I_GET_EXCL( FS_HANDLE_GET_SNAP(req->parentHandle.inum),
                            inum);
        if (inode == NULL)
        {
            fs_error(context, FS_INTERNAL_ERR, "inode not found");
        }
        else
        {
            if (FS_I_IS_DIR(inode->dinode.attrs) == 0)
            {
                fs_getRespAttrs(context, parentInode,
                                FS_HANDLE_GET_INUM(req->parentHandle.inum),
                                FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs,
                                TRUE);
                fs_inodeReleaseExclusive(context, inode);
                fs_inodeReleaseExclusive(context, parentInode);
                return FS_ERR_NOTDIR;
            }

            if (!fs_dir_is_empty(context, inode))
            {
                // Directory not empty.
                fs_getRespAttrs(context, parentInode,
                                FS_HANDLE_GET_INUM(req->parentHandle.inum),
                                FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs,
                                TRUE);
                fs_inodeReleaseExclusive(context, inode);
                fs_inodeReleaseExclusive(context, parentInode);
                return FS_ERR_NOTEMPTY;
            }

			fset = inode->fset = fs_getFset(context, core, inode);

			/*
			 * Remove the fileset if this is the root of the fileset
			 */

			if (inode->fset->cachedRecord.rootInode == inode->inum) {

				/*
				 * If we cannot block out quotaon / quotaoff on this
				 * fileset, then return busy indicating that the
				 * operation should be tried again
				 */

#if 1
                //Check whether quota on is going on
            	NFX_SPIN_LOCK(&fset->spinlock);
                if (fset->incoreFlag & FS_FILESET_QUOTA_ON_THREAD)  {
                    NFX_SPIN_UNLOCK(&fset->spinlock);
                    return FS_ERR_BUSY;
                }
            	NFX_SPIN_UNLOCK(&fset->spinlock);
    
				if (!fs_trylockExclusive(context, &fset->traverseLock)) {
                    fs_getRespAttrs(context, parentInode,
                                    FS_HANDLE_GET_INUM(req->parentHandle.inum),
                                    FS_POST_PAR_ATTRS, &attrp,
                                    &resp->numAttrs, TRUE);
                    fs_inodeReleaseExclusive(context, inode);
                    fs_inodeReleaseExclusive(context, parentInode);
                    return FS_ERR_BUSY;
				}

				/*
				 * From this point ownwards remove operation cannot
				 * fail. Remove the fileset as well
				 */

				fsetInfoInode = I_GET_EXCL(0, FS_FILESET_INFO_INODE);
				fs_removeFileSet(context, core, fset, inode,
								 fsetInfoInode);

                //Decrement the number of filesets
                NFX_SPIN_LOCK(&core->filesetListLock);
                FS_ASSERT(core->filesetsCount != 1)
                core->filesetsCount--;
                NFX_SPIN_UNLOCK(&core->filesetListLock);
                
				/*
				 * If we dirtied the fileset inode in this transaction
				 * then it will be released by commit, otherwise unlock
				 * it now
				 */

                if (!fs_inodeWasModified(context, fsetInfoInode)) {
					fs_inodeReleaseExclusive(context, fsetInfoInode);
				}
#else

                /*
                 * This a root of the fileset. Treat it as root of the
                 * filesystem. Don't allow removal of the root.
                 */
                fs_getRespAttrs(context, parentInode,
                                FS_HANDLE_GET_INUM(req->parentHandle.inum),
                                FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs,
                                TRUE);

                fs_inodeReleaseExclusive(context, inode);
                fs_inodeReleaseExclusive(context, parentInode);
                return FS_ERR_EBUSY;
                
#endif
			}

            fs_removeDirNode(context, parentInode, inode, req->name,
							 req->nameLen, NULL, NULL, NULL);

            fs_getRespAttrs(context, parentInode,
                            FS_HANDLE_GET_INUM(req->parentHandle.inum),
                            FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
        }

        rc = fs_logCommitEntry(context);
        if (rc != NFX_OK)
        {
            fs_error(context, FS_INTERNAL_ERR, "commit");
        }
    }

    return rc;

} /* fs_rmdirThread */


/**
 * Handle the RENAME request.
 */
int32
fs_renameThread(
    fs_Context      *context)
{
    fs_MsgRenameResp    *resp;
    fs_MsgRenameReq     *req;
    fs_Handle           dirHandles[2];
    fs_Inode            *fromInode, *fromDirInode;
    fs_Inode            *toInode, *toDirInode;
    fs_Inode            *inodePtrs[2], **fromDirInodePtr, **toDirInodePtr;
    fs_Attr             *attrp;
    boolean             ownModifyLock = FALSE;
    uint64              inum;
    int32               i, rc, numDirs, madeMods = 0, nameLen;
    char                *fromName, *toName;
    fs_FileSetIncoreRecord_t    *fromFset = NULL, *toFset = NULL;
    fs_QuotaContextList_t   *list1, *list2, *list3, *list4, *list5;
    fs_Status           status;
    boolean             wasDestroyed;
	int64				parentNumBlocks;

	list1 = list2 = list3 = list4 = list5 = NULL;
    resp = (fs_MsgRenameResp *) context->respHdr;
    req = (fs_MsgRenameReq *) context->reqHdr;
    context->respSize = sizeof(fs_MsgRenameResp);
    resp->numAttrs = 0;

    if (context->fileSystem->state == FS_VOL_STATE_ONLINE_RDONLY) {
        return FS_ERR_ROFS;
    }

    attrp = &(resp->attr[0]);

    fromName = req->fromName;
    toName = fromName + req->fromNameLen + 1;
    fromInode = fromDirInode = NULL;
    toInode = toDirInode = NULL;
    inodePtrs[0] = inodePtrs[1] = NULL;

    FS_TRACE(context, FS_TRACE_REQ,
            ("%s xid 0x%x rename snap %lld fmdir inum %lld"
             " todir inum %lld fmname %s toname %s",
             fs_protoStr[context->reqHdr->proto],
             context->reqHdr->clientId,
             FS_HANDLE_GET_SNAP(req->fromDirHandle.inum),
             FS_HANDLE_GET_INUM(req->fromDirHandle.inum),
             FS_HANDLE_GET_INUM(req->toDirHandle.inum),
             fromName,
             toName));

    if (!fs_handleVolMatch(req->fromDirHandle, req->toDirHandle))
    {
        return FS_ERR_STALE;
    }

    if (fs_handleIsSnapMagic(&req->fromDirHandle) ||
        fs_handleIsSnapMagic(&req->toDirHandle))
    {
        return FS_ERR_ACCESS;
    }

    if (FS_HANDLE_GET_INUM(req->fromDirHandle.inum) == 
        FS_HANDLE_GET_INUM(req->toDirHandle.inum))
    {
        dirHandles[0] = req->fromDirHandle;
        fromDirInodePtr = &inodePtrs[0];
        toDirInodePtr = &inodePtrs[0];

        numDirs = 1;    
    }
    else
    {
        // Avoid deadlock, acquire lower numbered directory
        // inode first.
        if (FS_HANDLE_GET_INUM(req->fromDirHandle.inum) <
            FS_HANDLE_GET_INUM(req->toDirHandle.inum))
        {
            dirHandles[0] = req->fromDirHandle;
            fromDirInodePtr = &inodePtrs[0];
            dirHandles[1] = req->toDirHandle;
            toDirInodePtr = &inodePtrs[1];
        }
        else
        {
            dirHandles[0] = req->toDirHandle;
            toDirInodePtr = &inodePtrs[0];
            dirHandles[1] = req->fromDirHandle;
            fromDirInodePtr = &inodePtrs[1];
        }

        numDirs = 2;
    }

renameRestart:
    context->respSize = sizeof(fs_MsgRenameResp);
    context->respHdr->attrType = 0;
    resp->numAttrs = 0;
    attrp = &(resp->attr[0]);
    for (i = 0; i < numDirs; i++)
    {
        inodePtrs[i] = I_GET_EXCL(  FS_HANDLE_GET_SNAP(dirHandles[i].inum),
                                    FS_HANDLE_GET_INUM(dirHandles[i].inum));
        if (inodePtrs[i] == NULL ||
            !fs_handleIsValid(context, &dirHandles[i], inodePtrs[i]))
        {
            rc = FS_ERR_STALE;
            goto renameErr;
        }
        if (!FS_I_IS_DIR(inodePtrs[i]->dinode.attrs))
        {
            rc = FS_ERR_NOTDIR;
            goto renameErr;
        }
    }

    fromDirInode = *fromDirInodePtr;
    toDirInode = *toDirInodePtr;

    fs_getRespAttrs(context, fromDirInode,
                    FS_HANDLE_GET_INUM(req->fromDirHandle.inum),
                    FS_PRE_FPAR_ATTRS, &attrp,
                    &resp->numAttrs, TRUE);
    fs_getRespAttrs(context, toDirInode,
                    FS_HANDLE_GET_INUM(req->toDirHandle.inum),
                    FS_PRE_TPAR_ATTRS, &attrp,
                    &resp->numAttrs, TRUE);

    nameLen = MAX(req->fromNameLen, req->toNameLen);
    if (numDirs == 1 && !strncmp(fromName, toName, nameLen))
    {
        // Nothing to do.
        goto renameDone;
    }

    rc = fs_nameIsValid(fromName, req->fromNameLen);
    if (rc != NFX_OK)
    {
        goto renameErr;
    }

    rc = fs_nameIsValid(toName, req->toNameLen);
    if (rc != NFX_OK)
    {
        goto renameErr;
    }

    for (i = 0; i < numDirs; i++)
    {
        if (!fs_permissionOkay(context, inodePtrs[i], FS_PERM_DIRENT))
        {
            rc = FS_ERR_ACCESS;
            goto renameErr;
        }
    }
	if (context->flags & FS_CONTEXT_ENOSPC_CONDITION)
    {
		rc = FS_ERR_NOSPC;
        goto renameErr;
	}

    // Lookup fromName.
    inum = fs_dir_lookupent(    context,
                                fromDirInode,
                                fromName,
                                req->fromNameLen);  
    if (inum == 0)
    {
        // From name doesn't exist.
        rc = FS_ERR_EXIST;
        goto renameErr;
    }
	FS_LOCK_ORDER_SET_PARENT(fromDirInode->inum);
    fromInode = I_GET_EXCL(0, inum);
    if (fromInode == NULL)
    {
        fs_error(context, FS_INTERNAL_ERR, "iget");
    }

    // Before proceeding, verify we're not trying to move
    // a directory to one of its own subdirs.
    if (FS_I_IS_DIR(fromInode->dinode.attrs))
    {
        fs_Inode    *toDirParent = NULL, *prevToDirParent = NULL;
        uint64      toDirParentInum;

        toDirParentInum = fs_dirroot_get_parent(context, toDirInode);
        while (toDirParentInum != FS_ROOT_INODE)
        {
            if (toDirParentInum == fromInode->inum)
            {
                rc = FS_ERR_NOTSUPP;
                if (toDirParent != NULL)
                    fs_inodeRelease(context, toDirParent);
                goto renameErr;
            }
            prevToDirParent = toDirParent;
            toDirParent = I_GET_TRY_SHARED(0, toDirParentInum);
            if (prevToDirParent != NULL)
            {
                fs_inodeRelease(context, prevToDirParent);
            }
            if (!toDirParent)
            {
                // Someone else has the inode locked.  Avoid deadlock
                // by releasing everything and getting the modify lock
                // exclusive.
                FS_TRACE(context, FS_TRACE_REQ,
                    ("smHandle %llx rename acquiring fs modify lock, restarting",
					 (uint64)((uintptr_t)context->smHandle)));
                if (ownModifyLock)
                {
                    fs_error(context, FS_INTERNAL_ERR,
                             "rename trylock failed with modifyLock held");
                }
                fs_inodeReleaseExclusive(context, fromInode);
                fromInode = NULL;
                for (i = 0; i < 2; i++)
                {
                    if (inodePtrs[i] != NULL)
                    {
                        fs_inodeReleaseExclusive(context, inodePtrs[i]);
                        inodePtrs[i] = NULL;
                    }
                }
                fs_unlockShared(context, &context->fileSystem->modifyLock);
                context->flags &= ~FS_CONTEXT_MODIFY_LOCK_ACQUIRED;
                fs_lockExclusive(context, &context->fileSystem->modifyLock, NULL);
                ownModifyLock = TRUE;
                goto renameRestart;
            }

            toDirParentInum = fs_dirroot_get_parent(context, toDirParent);
        }
		if (toDirParent != NULL)
		{
            fs_inodeRelease(context, toDirParent);
		}
    }
    if (ownModifyLock)
    {
        fs_unlockExclusive(context, &context->fileSystem->modifyLock);
        ownModifyLock = FALSE;
    }

    // Lookup toName.
    inum = fs_dir_lookupent(    context,
                                toDirInode,
                                toName,
                                req->toNameLen);    
    if (inum != 0)
    {
        //printf("toName %s exists\n", toName);
        // To name exists.
		FS_LOCK_ORDER_SET_PARENT(toDirInode->inum);
        toInode = I_GET_EXCL(0, inum);
        if (toInode == NULL)
        {
            fs_error(context, FS_INTERNAL_ERR, "iget");
        }

        if (FS_I_IS_DIR(toInode->dinode.attrs))
        {
            //printf("toName is a dir\n");
            if (!fs_dir_is_empty(context, toInode) ||
                !FS_I_IS_DIR(fromInode->dinode.attrs))
            {
                //printf("toName is not empty or from is not a dir\n");
                rc = FS_ERR_EXIST;
                goto renameErr;
            }
        }
        else if (FS_I_IS_DIR(fromInode->dinode.attrs))
        {
            // Mismatch.
            //printf("mismatch\n");
            rc = FS_ERR_EXIST;
            goto renameErr;
        }

        if (fromInode->inum == toInode->inum)
        {
            //printf("to and from are same inode\n");
            // Nothing to do, they're the same file.
            goto renameDone;
        }
    }

    FS_TRACE(context, FS_TRACE_REQ, ("compare fsets"));

    /*
     * If we are trying to rename an object from 1 fset to another
     * fail it.
     */

     fromFset = fs_getFset(context, context->fileSystem, fromInode);
     if (toInode != NULL) {
         toFset = fs_getFset(context, context->fileSystem, toInode);
     } else {
         toFset = fs_getFset(context, context->fileSystem, toDirInode);
     }

     /*
      * We can access the fset without locking it as the root
      * of the fileset is locked, so remove fileset cannot go through
      */

     if (fromFset != toFset) {
            rc = FS_ERR_PERM;  
            goto renameErr;
     }
    
    FS_TRACE(context, FS_TRACE_REQ, ("from fset root is %lld id is %d",
             fromFset->cachedRecord.rootInode, fromFset->fileSetId));
    FS_TRACE(context, FS_TRACE_REQ, ("to fset root is %lld id is %d",
             toFset->cachedRecord.rootInode, toFset->fileSetId));

    /*
     * Similarly if the toInode is a root of a fileset, return EBUSY.
     * Treat it like a root of a filesystem
     */

    if (toInode)
    {
        // Remove existing toName.
        if (FS_I_IS_DIR(toInode->dinode.attrs))
        {
            //printf("removing existing dir toName %s\n", toName);
            fs_removeDirNode(   context,
                                toDirInode,
                                toInode,
                                toName,
                                req->toNameLen,
								&list1,
								&list2,
                                &list3);
        }
        else
        {
            //printf("removing existing non-dir toName %s\n", toName);
            fs_removeNonDirNode(context,
                                toDirInode,
                                toInode,
                                toName,
                                req->toNameLen,
								&list1,
								&list2,
								&list3,
                                &wasDestroyed);
            if (!wasDestroyed)
            {
                // Return the handle and updated attributes for
                // the inode so the NFS server can sync up its
                // attribute cache.
                fs_getRespAttrs(context, toInode,
                                FS_HANDLE_GET_INUM(req->toDirHandle.inum),
                                FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);

                fs_handleInit(context->fileSystem,
                              &resp->fileHandle,
                              FS_HANDLE_GET_SNAP(req->fromDirHandle.inum),
                              toInode->inum,
                              toInode->dinode.genNum,
                              req->fromDirHandle.mntPnt);
            }
        }
        //status = fs_removeLink(context, 0, toInode, toDirInode->inum);
        //FS_ASSERT(status == FS_ERR_NONE);
    }

    //printf("removing fromName %s\n", fromName);
    // Remove from dirent.
	parentNumBlocks = fromDirInode->dinode.numblocks;
    rc = fs_dir_removeent(  context,
                            fromDirInode,
                            fromName,
                            req->fromNameLen);
    status = fs_removeLink(context, 0, fromInode, fromDirInode->inum);
    if (status != FS_ERR_NONE) {
        rc = NFX_ERR;
    }

    /*
     * Hack to make incremental dump make happy. Update the inode
     * change time so that incremental dump sees this file for
     * backup
     */

    if (fromInode->dinode.linkCnt > 1 &&
        !FS_I_IS_DIR(fromInode->dinode.attrs))
    {
        fromInode->dinode.ctime = fs_getTimeOfDay();
        fromInode->dinode.version++;
        fs_inodeWrite(context, fromInode, INO_WRITE_LOG);
    }
/* TODO
    if (rc != FS_BT_REMOVE_OK &&
        rc != FS_BT_REMOVE_OK_EMPTY)
    {
        fs_error(context, FS_INTERNAL_ERR, "rename remove dirent");
    }
*/
	//Check if directory size has changed
	parentNumBlocks = fromDirInode->dinode.numblocks - parentNumBlocks;
	if (parentNumBlocks != 0)	{
		fs_QuotaAdjustmentNoCheck(	context,
									context->fileSystem,
									fromDirInode,
									NULL,
									parentNumBlocks,
									FS_COUNT_BLOCKS,
									fromDirInode->sd->u.nfs.uid,
									fromDirInode->sd->u.nfs.gid,
									&list4);
	}

    //printf("creating new dirent %s\n", toName);
    // Create new dirent.
	parentNumBlocks = toDirInode->dinode.numblocks;
    rc = fs_dir_addent( context,
                        toDirInode,
                        toName,
                        req->toNameLen,
                        fromInode->inum);
    status = fs_addLink(context, 0, fromInode, toDirInode->inum);
    if (status != FS_ERR_NONE) {
        rc = NFX_ERR;
    }
/* TODO
    if (rc != FS_BT_INSERT_OK &&
        rc != FS_BT_INSERT_DUP)
    {
        fs_error(context, FS_INTERNAL_ERR, "add dirent");
    }
*/
	//Check if directory size has changed
	parentNumBlocks = toDirInode->dinode.numblocks - parentNumBlocks;
	if (parentNumBlocks != 0)	{
		fs_QuotaAdjustmentNoCheck(	context,
									context->fileSystem,
									toDirInode,
									NULL,
									parentNumBlocks,
									FS_COUNT_BLOCKS,
									toDirInode->sd->u.nfs.uid,
									toDirInode->sd->u.nfs.gid,
									&list5);
	}

    if (FS_I_IS_DIR(fromInode->dinode.attrs) && fromDirInode != toDirInode)
    {
        // Decr parent link count (for .. of fromName).
        fromDirInode->dinode.linkCnt -= 1;

        // Incr parent link count (for .. of toName).
        toDirInode->dinode.linkCnt += 1;

        // Update the moved inode's btree with its new
        // parent inode number.
        fs_dirroot_set_parent(context, fromInode, toDirInode->inum);
    }
    
    // fromInode may not be modified, so be sure it's unlocked during
    // commit.
    fs_inodeLockTransaction(context, fromInode, context->logSummary);

    toDirInode->dinode.ctime = fs_getTimeOfDay();
    toDirInode->dinode.mtime = toDirInode->dinode.ctime;
    toDirInode->dinode.version++;
    fs_inodeWrite(context, toDirInode, INO_WRITE_LOG);

    if (fromDirInode != toDirInode)
    {
        fromDirInode->dinode.ctime = fs_getTimeOfDay();
        fromDirInode->dinode.mtime = fromDirInode->dinode.ctime;
        fromDirInode->dinode.version++;
        fs_inodeWrite(context, fromDirInode, INO_WRITE_LOG);
    }

    madeMods = 1;

renameDone:

    fs_getRespAttrs(context, fromDirInode,
                    FS_HANDLE_GET_INUM(req->fromDirHandle.inum),
                    FS_POST_FPAR_ATTRS, &attrp,
                    &resp->numAttrs, TRUE);
    fs_getRespAttrs(context, toDirInode,
                    FS_HANDLE_GET_INUM(req->toDirHandle.inum),
                    FS_POST_TPAR_ATTRS, &attrp,
                    &resp->numAttrs, TRUE);

    if (madeMods)
    {
			
	    /*
	     * If we did any quota reservations, commit them.
		 * We commit the directory size change lists first since they are less
		 * likely to exist. Therefore, we prevent the unnecessary hold of
		 * quota inodes
	     */
	    if (list2 != NULL) {
	        fs_commitContextList(context, context->fileSystem, list2, TRUE);
	        fs_freeQuotaContextList(context, context->fileSystem, list2);
	    }       
	    if (list4 != NULL) {
	        fs_commitContextList(context, context->fileSystem, list4, TRUE);
	        fs_freeQuotaContextList(context, context->fileSystem, list4);
	    }
	    if (list5 != NULL) {
	        fs_commitContextList(context, context->fileSystem, list5, TRUE);
	        fs_freeQuotaContextList(context, context->fileSystem, list5);
	    }
	    if (list1 != NULL) {
	        fs_commitContextList(context, context->fileSystem, list1, TRUE);
	        fs_freeQuotaContextList(context, context->fileSystem, list1);
	    }       
	    if (list3 != NULL) {
	        fs_commitContextList(context, context->fileSystem, list3, FALSE);
	        fs_freeQuotaContextList(context, context->fileSystem, list3);
	    }
	
        rc = fs_logCommitEntry(context);
        if (rc != NFX_OK)
        {
            fs_error(context, FS_INTERNAL_ERR, "commit");
        }
    }
    else
    {
        rc = NFX_OK;
        goto renameUnlock;
    }

    return NFX_OK;

renameErr:

    if (fromDirInode)
    {
        fs_getRespAttrs(context, fromDirInode,
                        FS_HANDLE_GET_INUM(req->fromDirHandle.inum),
                        FS_POST_FPAR_ATTRS, &attrp,
                        &resp->numAttrs, TRUE);
    }
    if (toDirInode)
    {
        fs_getRespAttrs(context, toDirInode,
                        FS_HANDLE_GET_INUM(req->toDirHandle.inum),
                        FS_POST_TPAR_ATTRS, &attrp,
                        &resp->numAttrs, TRUE);
    }

renameUnlock:

    if (fromInode != NULL)
    {
        if (!fs_inodeWasModified(context, fromInode)) {
            fs_inodeReleaseExclusive(context, fromInode);
        }
    }
    if (toInode && toInode != fromInode)
    {
        fs_inodeReleaseExclusive(context, toInode);
    }
    for (i = 0; i < 2; i++)
    {
        if (inodePtrs[i] != NULL)
        {
            fs_inodeReleaseExclusive(context, inodePtrs[i]);
        }
    }

    if (ownModifyLock)
    {
        fs_unlockExclusive(context, &context->fileSystem->modifyLock);
    }

    return rc;

} /* fs_renameThread */


/**
 * Handle the HARD_LINK request.
 */
int32
fs_hardLinkThread(
    fs_Context      *context)
{
    fs_MsgHardLinkResp  *resp;
    fs_MsgHardLinkReq   *req;
    fs_Inode            *inode, *parentInode;
    fs_Attr             *attrp;
    uint64              inum;
    int32               rc;
    fs_FileSetIncoreRecord_t    *fset, *parentDirFset;
    fs_Status           status;
	int64				parentNumBlocks;
    fs_QuotaContextList_t   *list1 = NULL;

    resp = (fs_MsgHardLinkResp *) context->respHdr;
    req = (fs_MsgHardLinkReq *) context->reqHdr;
    resp->numAttrs = 0;
    context->respSize = sizeof(fs_MsgHardLinkResp);

    if (context->fileSystem->state == FS_VOL_STATE_ONLINE_RDONLY) {
        return FS_ERR_ROFS;
    }
    attrp = &(resp->attr[0]);
    inode = NULL;

    FS_TRACE(context, FS_TRACE_REQ,
             ("%s xid 0x%x hardlink snap %lld inum %lld gen %d name %s",
             fs_protoStr[context->reqHdr->proto],
             context->reqHdr->clientId,
             FS_HANDLE_GET_SNAP(req->fileHandle.inum),
             FS_HANDLE_GET_INUM(req->fileHandle.inum),
             req->fileHandle.gen,
             req->linkName));

    if (!fs_handleVolMatch(req->fileHandle, req->parentDirHandle))
    {
        return FS_ERR_STALE;
    }

    if (fs_handleIsSnapMagic(&req->parentDirHandle))
    {
        return FS_ERR_ACCESS;
    }

    parentInode = I_GET_EXCL(   FS_HANDLE_GET_SNAP(req->parentDirHandle.inum),
                                FS_HANDLE_GET_INUM(req->parentDirHandle.inum));
    if (parentInode == NULL)
    {
        return FS_ERR_STALE;
    }
    if(!fs_handleIsValid(context, &req->parentDirHandle, parentInode))
    {
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_STALE;
    }
    if (!FS_I_IS_DIR(parentInode->dinode.attrs))
    {
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_NOTDIR;
    }

    fs_getRespAttrs(context, parentInode,
                    FS_HANDLE_GET_INUM(req->parentDirHandle.inum),
                    FS_PRE_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);

    rc = fs_nameIsValid(req->linkName, req->nameLen);

    if (!fs_permissionOkay(context, parentInode, FS_PERM_DIRENT))
    {
        rc = FS_ERR_ACCESS;
    }
	if (context->flags & FS_CONTEXT_ENOSPC_CONDITION)
    {
		rc = FS_ERR_NOSPC;
	}
    if (rc != NFX_OK)
    {
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentDirHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_inodeReleaseExclusive(context, parentInode);
        return rc;
    }

    inum = fs_dir_lookupent(    context,
                                parentInode,
                                req->linkName,
                                req->nameLen);
    if (inum != 0)
    {
        // Name already exists.
		FS_LOCK_ORDER_SET_PARENT(parentInode->inum);
        inode = I_GET_SHARED( FS_HANDLE_GET_SNAP(inum),
                              FS_HANDLE_GET_INUM(inum));
        if (inode == NULL)
        {
            fs_error(context, FS_INTERNAL_ERR, "i_get");
        }
        fs_getRespAttrs(context, inode,
                        FS_HANDLE_GET_INUM(req->parentDirHandle.inum),
                        FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentDirHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_inodeRelease(context, inode);
        fs_inodeReleaseExclusive(context, parentInode);
        return FS_ERR_EXIST;
    }

	FS_LOCK_ORDER_SET_PARENT(parentInode->inum);
    inode = I_GET_EXCL( FS_HANDLE_GET_SNAP(req->fileHandle.inum),
                        FS_HANDLE_GET_INUM(req->fileHandle.inum));
    if (inode == NULL)
    {
        fs_error(context, FS_INTERNAL_ERR, "i_get");
    }

    if (FS_I_IS_DIR(inode->dinode.attrs))
    {
        // No hard link on directories.
        fs_getRespAttrs(context, inode,
                        FS_HANDLE_GET_INUM(req->parentDirHandle.inum),
                        FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentDirHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_inodeReleaseExclusive(context, parentInode);
        fs_inodeReleaseExclusive(context, inode);
        return FS_ERR_ISDIR;
    }

    if (inode->dinode.linkCnt == FS_INODE_LINK_MAX)
    {
        fs_getRespAttrs(context, inode,
                        FS_HANDLE_GET_INUM(req->parentDirHandle.inum),
                        FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentDirHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_inodeReleaseExclusive(context, parentInode);
        fs_inodeReleaseExclusive(context, inode);
        return FS_ERR_MLINK;
    }

    fset = fs_getFset(context, context->fileSystem, inode);
    parentDirFset = fs_getFset(context, context->fileSystem, parentInode);

    if (fset != parentDirFset) {
        fs_getRespAttrs(context, inode,
                        FS_HANDLE_GET_INUM(req->parentDirHandle.inum),
                        FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_getRespAttrs(context, parentInode,
                        FS_HANDLE_GET_INUM(req->parentDirHandle.inum),
                        FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);
        fs_inodeReleaseExclusive(context, parentInode);
        fs_inodeReleaseExclusive(context, inode);
        return FS_ERR_PERM;
    }

    // Insert dirent.
	parentNumBlocks = parentInode->dinode.numblocks;
    rc = fs_dir_addent( context,
                        parentInode,
                        req->linkName,
                        req->nameLen,
                        inode->inum);
    status = fs_addLink(context, 0, inode, parentInode->inum);
    if (status != FS_ERR_NONE) {
        rc = NFX_ERR;
    }
/* TODO
    if (rc != FS_BT_INSERT_OK &&
        rc != FS_BT_INSERT_DUP)
    {
        fs_error(context, FS_INTERNAL_ERR, "add dirent");
        return FS_ERR_IO;
    }
*/
	//Check if directory size has changed
	parentNumBlocks = parentInode->dinode.numblocks - parentNumBlocks;
	if (parentNumBlocks != 0)	{
		fs_QuotaAdjustmentNoCheck(	context,
									context->fileSystem,
									parentInode,
									NULL,
									parentNumBlocks,
									FS_COUNT_BLOCKS,
									parentInode->sd->u.nfs.uid,
									parentInode->sd->u.nfs.gid,
									&list1);
	}

    parentInode->dinode.mtime = fs_getTimeOfDay();
    fs_inodeWrite(context, parentInode, INO_WRITE_LOG);

    // Incr link count.
    inode->dinode.linkCnt += 1;
    inode->dinode.ctime = fs_getTimeOfDay();
    inode->dinode.version++;
    fs_inodeWrite(context, inode, INO_WRITE_LOG);

    fs_getRespAttrs(context, inode,
                    FS_HANDLE_GET_INUM(req->parentDirHandle.inum),
                    FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);
    fs_getRespAttrs(context, parentInode,
                    FS_HANDLE_GET_INUM(req->parentDirHandle.inum),
                    FS_POST_PAR_ATTRS, &attrp, &resp->numAttrs, TRUE);

    if (list1) {
	    fs_commitContextList(context, context->fileSystem, list1, FALSE);
	    fs_freeQuotaContextList(context, context->fileSystem, list1);
	}
	
    rc = fs_logCommitEntry(context);
    if (rc != NFX_OK)
    {
        fs_error(context, FS_INTERNAL_ERR, "commit");
    }

    return NFX_OK;

} /* fs_hardLinkThread */


/**
 * Handle the READDIR request.
 */
int32
fs_readDirThread(
    fs_Context      *context)
{
    fs_MsgReadDirReq    *req;
    fs_MsgReadDirResp   *resp;
    fs_DirEntry         *dirent;
    fs_Inode            *inode;
    fs_Attr             *attrp;
    int32               i, numEntries, eof, spaceMax;
    fs_Inumber          inum;
    int                 snapnum;
    boolean             showDotSnapshot = FALSE;
    fs_SnapType         snapType = FS_NO_SNAPSHOT_TYPE;
    uint64              snapInfo = 0;
    boolean             processSnapBoundary = FALSE;

#define FS_READDIR_SPACE_SMALL  512

    resp = (fs_MsgReadDirResp *) context->respHdr;
    req = (fs_MsgReadDirReq *) context->reqHdr;
    attrp = &resp->dirAttr;

    if (fs_handleIsSnapMagic(&req->fileHandle))
    {
        inum = FS_SNAPSHOT_DIR_INODE;
        snapnum = 0;
    }
    else
    {
        inum = FS_HANDLE_GET_INUM(req->fileHandle.inum);
        snapnum = FS_HANDLE_GET_SNAP(req->fileHandle.inum);
        if (FS_HANDLE_IS_SNAP_BOUNDARY(req->fileHandle))
        {
            processSnapBoundary = TRUE;
        }
    }

    // If this is the live fs and a mountpoint, then include
    // ".snapshot" in the response.
    if (snapnum == 0 &&
        inum == req->fileHandle.mntPnt)
    {
        showDotSnapshot = TRUE;
    }

    // Also include ".snapshot" in the response if magic2 set
    if (fs_handleIsSnapMagic2(&req->fileHandle))
    {
        showDotSnapshot = TRUE;
    }
    
    FS_TRACE(context, FS_TRACE_REQ,
             ("%s xid 0x%x readdir snap %d inum %lld gen %d cookie %lld max %d",
              fs_protoStr[context->reqHdr->proto],
              context->reqHdr->clientId,
              snapnum, inum,
              req->fileHandle.gen,
              req->cookie,
              req->maxBytes));

    context->respSize = sizeof(fs_MsgReadDirResp);
    resp->numEntriesThis = 0;
    resp->eof = TRUE;

    inode = I_GET_SHARED(snapnum, inum);
    if (inode == NULL)
    {
        return FS_ERR_STALE;
    }
    if(!fs_handleIsValid(context, &req->fileHandle, inode))
    {
        fs_inodeRelease(context, inode);
        return FS_ERR_STALE;
    }
    if (!FS_I_IS_DIR(inode->dinode.attrs))
    {
        fs_inodeRelease(context, inode);
        return FS_ERR_NOTDIR;
    }
    if (!fs_permissionOkay(context, inode, FS_PERM_READDIR))
    {
        fs_getRespAttrs(context, inode,
                        FS_HANDLE_GET_INUM(req->fileHandle.inum),
                        FS_POST_ATTRS, &attrp, NULL, FALSE);
        fs_inodeRelease(context, inode);
        return FS_ERR_ACCESS;
    }

    spaceMax = fs_incoreTune->readdirMaxSpace; 
    if (req->cookie == 0)
    {
        // This helps the sfs benchmark and doesn't hurt
        // other cases too much (sfs always reads from 0
        // and doesn't care how big the response is).
        spaceMax = MIN(FS_READDIR_SPACE_SMALL, spaceMax);
    }

    fs_dir_readdir( context,
                    inode,
                    &resp->dirEntry[0],
                    MIN(req->maxBytes, spaceMax),
                    req->cookie,
                    showDotSnapshot,
                    &numEntries,
                    &eof);

    fs_inodeAtimeUpdate(context, inode);

    // set the correct directory attributes
    if (inode->inum == FS_SNAPSHOT_DIR_INODE)
    {
        snapType = FS_SNAPSHOT_DIR_TYPE;
        snapInfo = FS_HANDLE_GET_INUM(req->fileHandle.inum);
    }
    else if (FS_HANDLE_IS_SNAP_BOUNDARY(req->fileHandle))
    {
        snapType = FS_SNAPSHOT_BOUNDARY_TYPE;
        snapInfo = snapnum + FS_SNAPSHOT_ID_INODE_FIRST;
    }
    fs_getSnapRespAttrs(context, inode, snapType, snapInfo,
                        FS_POST_ATTRS, &attrp, NULL, FALSE);

    fs_inodeRelease(context, inode);

    resp->numEntriesThis = numEntries;
    resp->eof = eof;
    resp->cookieVerif = 0;

    context->respSize = sizeof(fs_MsgReadDirResp);
    dirent = &resp->dirEntry[0];
    for (i = 0; i < numEntries; i += 1)
    {
        //printf("name: %s  len: %d  cookie: %lld  inode %llx\n",
        //      dirent->name, dirent->nameLen, dirent->cookie, dirent->inode);

        if (processSnapBoundary)
        {
            if (!strcmp(dirent->name, "."))
            {
                // inode is snapnum inum
                dirent->inode = FS_HANDLE_GET_SNAP(req->fileHandle.inum) + FS_SNAPSHOT_ID_INODE_FIRST;
            }
            else if (!strcmp(dirent->name, ".."))
            {
                // parent is .snapshot dir
                dirent->inode = FS_SNAPSHOT_DIR_INODE;
            }
        }

        FS_READDIR_ADV_ENTRY_PTR(dirent);
    }
    context->respSize += ((char *)dirent - (char *)&(resp->dirEntry[0]));

    return NFX_OK;

} /* fs_readDirThread */


/**
 * Handle the READDIR_PLUS request.
 */
int32
fs_readDirPlusThread(
    fs_Context      *context)
{
    fs_MsgReadDirPlusReq    *req;
    fs_MsgReadDirPlusResp   *resp;
    fs_DirEntryPlus         *direntPlus;
    fs_DirEntry             *dirent;
    fs_Inode                *inode;
    fs_Attr                 *attrp;
    int32                   numThis, numEntries, eof, dirCount, totalCount;
    int32                   spaceMax;
    char                    *tmpDirents;
    fs_Inumber              parentInum;
    int                     parentSnapnum;
    boolean                 showDotSnapshot = FALSE;
    fs_SnapType             snapType = FS_NO_SNAPSHOT_TYPE;
    uint64                  snapInfo = 0;

    resp = (fs_MsgReadDirPlusResp *) context->respHdr;
    req = (fs_MsgReadDirPlusReq *) context->reqHdr;
    attrp = &resp->attr;

    if (fs_handleIsSnapMagic(&req->fileHandle))
    {
        parentInum = FS_SNAPSHOT_DIR_INODE;
        parentSnapnum = 0;
    }
    else
    {
        parentInum = FS_HANDLE_GET_INUM(req->fileHandle.inum);
        parentSnapnum = FS_HANDLE_GET_SNAP(req->fileHandle.inum);
    }

    // If this is the live fs and a mountpoint, then include
    // ".snapshot" in the response.
    if (parentSnapnum == 0 &&
        parentInum == req->fileHandle.mntPnt)
    {
        showDotSnapshot = TRUE;
    }

    // Also include ".snapshot" in the response if magic2 set
    if (fs_handleIsSnapMagic2(&req->fileHandle))
    {
        showDotSnapshot = TRUE;
    }

    FS_TRACE(context, FS_TRACE_REQ,
             ("%s xid 0x%x readdirplus snap %d inum %lld gen %d cookie %lld"
              " maxBytes %d maxDirBytes %d",
              fs_protoStr[context->reqHdr->proto],
              context->reqHdr->clientId,
              parentSnapnum, parentInum,
              req->fileHandle.gen,
              req->cookie,
              req->maxBytes,
              req->maxDirBytes));

    context->respSize = sizeof(fs_MsgReadDirPlusResp);
    resp->numEntriesThis = 0;
    resp->eof = TRUE;

    if (fs_handleIsSnapMagic(&req->fileHandle))
    {
        inode = I_GET_SHARED(0, FS_SNAPSHOT_DIR_INODE);
    }
    else
    {
        inode = I_GET_SHARED(parentSnapnum, parentInum);
    }
    if (inode == NULL)
    {
        return FS_ERR_STALE;
    }
    if(!fs_handleIsValid(context, &req->fileHandle, inode))
    {
        fs_inodeRelease(context, inode);
        return FS_ERR_STALE;
    }
    if (!FS_I_IS_DIR(inode->dinode.attrs))
    {
        fs_inodeRelease(context, inode);
        return FS_ERR_NOTDIR;
    }
    if (!fs_permissionOkay(context, inode, FS_PERM_READDIR))
    {
        fs_getRespAttrs(context, inode,
                        FS_HANDLE_GET_INUM(req->fileHandle.inum),
                        FS_POST_ATTRS, &attrp, NULL, FALSE);
        fs_inodeRelease(context, inode);
        return FS_ERR_ACCESS;
    }

    // For simplicity, just build the non-plus style response in tmp
    // space and then construct the plus style from it.  Allocate space
	// based on the hint (maxDirBytes) in the request with our own
	// in-core tunable as max.   We read only 50% of the maxDirBytes
    // indicated since when we expand the entries into readDirPlus format
    // we would end up using fewer entries than we read.  

    spaceMax = MIN(req->maxDirBytes, fs_incoreTune->readdirMaxSpace);
    spaceMax /= 2;

    if (req->cookie == 0)
    {
        // This helps the sfs benchmark and doesn't hurt
        // other cases too much (sfs always reads from 0
        // and doesn't care how big the response is).
        spaceMax = MIN(FS_READDIR_SPACE_SMALL, spaceMax);
    }

    tmpDirents = fs_ramAlloc(FS_RAM_CORE_MISC, spaceMax);

    fs_dir_readdir( context,
                    inode,
                    (fs_DirEntry *) tmpDirents,
                    spaceMax,
                    req->cookie,
                    showDotSnapshot,
                    &numEntries,
                    &eof);

    fs_inodeAtimeUpdate(context, inode);

    // set the correct directory attributes
    if (inode->inum == FS_SNAPSHOT_DIR_INODE)
    {
        snapType = FS_SNAPSHOT_DIR_TYPE;
        snapInfo = FS_HANDLE_GET_INUM(req->fileHandle.inum);
    }
    else if (FS_HANDLE_IS_SNAP_BOUNDARY(req->fileHandle))
    {
        snapType = FS_SNAPSHOT_BOUNDARY_TYPE;
        snapInfo = parentSnapnum + FS_SNAPSHOT_ID_INODE_FIRST;
    }

    fs_getSnapRespAttrs(context, inode, snapType, snapInfo,
                        FS_POST_ATTRS, &attrp, NULL, FALSE);

    fs_inodeRelease(context, inode);

    dirent = (fs_DirEntry *) tmpDirents;
    direntPlus = &(resp->dirEntry[0]);
    for (numThis = 0, dirCount = 0, totalCount = 0;
         numThis < numEntries;
         numThis += 1)
    {
        fs_Inode    *entryInode;
        uint64      effectiveInum;
        int32       snapnum;

        FS_READDIR_COUNT_NEXT_ENTRY_PTR(dirent, dirCount);
        if (dirCount > req->maxDirBytes)
        {
            break;
        }
        
        totalCount = ((char *)direntPlus - (char *)resp)
                     + sizeof(fs_DirEntryPlus) 
                     + dirent->nameLen
                     + sizeof(uint64);
        if (totalCount > MIN(req->maxBytes, spaceMax))
        {
            break;
        }

        // Determine the snapshot number of the inode we're
        // looking for.  If the parent directory is a .snapshot
        // directory, then the snapnum is the inum (minus offset),
        // otherwise the snapnum is the same as the parent dir's
        // snapnum.

        // Determine the effective inum that we're looking for.

        if (parentInum == FS_SNAPSHOT_DIR_INODE)
        {
            if (!strcmp(dirent->name, ".."))
            {
                // ".." entry for a ".snapshot" directory should
                // point to its actual parent (which we've strategically
                // placed in the inum field of the handle given out
                // for the ".snapshot" directory).

                dirent->inode = FS_HANDLE_GET_INUM(req->fileHandle.inum);
                snapnum = 0;
                effectiveInum = dirent->inode;
            }
            else if (!strcmp(dirent->name, "."))
            {
                snapnum = 0;
                effectiveInum = dirent->inode;
            }
            else
            {
                snapnum = dirent->inode - FS_SNAPSHOT_ID_INODE_FIRST;
                effectiveInum = FS_HANDLE_GET_INUM(req->fileHandle.inum);
            }
        }
        else
        {
            snapnum = parentSnapnum;
            effectiveInum = dirent->inode;
        }

        entryInode = I_GET_SHARED(snapnum, effectiveInum);

        if (entryInode == NULL)
        {
            if (parentInum == FS_SNAPSHOT_DIR_INODE)
            {
                fs_Inode    *nullInode = I_GET_SHARED(snapnum, FS_NULL_INODE);

                // File doesn't exist for the indicated snapshot.
                // Return the NULL inode.

                fs_handleInit(context->fileSystem,
                              &direntPlus->handle,
                              snapnum,
                              FS_NULL_INODE,
                              0,
                              req->fileHandle.mntPnt);

                fs_doGetAttr(context->volId, &direntPlus->attr, nullInode,
                             FS_SNAPSHOT_DIR_TYPE, FS_HANDLE_GET_INUM(req->fileHandle.inum),
                             context->reqHdr->proto, FS_POST_ATTRS);

                fs_inodeRelease(context, nullInode);
                dirent->inode = FS_NULL_INODE;
            }
            else
            {
                // Dirent removed while we were looping without
                // holding the parent inode.  Return !eof to force
                // client to request more from this point.
                break;
            }
        }
        else
        {
            uint64      handleInum;
            uint32      genNum;

            snapType = FS_NO_SNAPSHOT_TYPE;
            snapInfo = 0;

            if (effectiveInum == FS_SNAPSHOT_DIR_INODE)
            {
                // The inum in the handle needs to be the parent inum for the
                // .snapshot dir - essentially a link to the directory
                // contents for the snapshot.
                handleInum = FS_HANDLE_GET_INUM(req->fileHandle.inum);
                genNum = FS_INODE_GENNUM_SNAP_MAGIC;
                snapType = FS_SNAPSHOT_DIR_TYPE;
                snapInfo = handleInum;
            }
            else if (parentInum == FS_SNAPSHOT_DIR_INODE)
            {
                // The inum in the handle needs to be the parent inum for a
                // .snapshot dirent - essentially a link to the directory
                // contents for the snapshot.
                handleInum = FS_HANDLE_GET_INUM(req->fileHandle.inum);
                genNum = entryInode->dinode.genNum;
                if ((strcmp(dirent->name, "..") != 0) &&
                    (handleInum == req->fileHandle.mntPnt))
                {
                    snapType = FS_SNAPSHOT_BOUNDARY_TYPE;
                    snapInfo = snapnum + FS_SNAPSHOT_ID_INODE_FIRST;
                }
                // setup magic2 bit if parent entry
                if (!strcmp(dirent->name, "..") &&
                    (FS_HANDLE_GET_INUM(req->fileHandle.inum) != req->fileHandle.mntPnt))
                {
                    genNum = FS_INODE_GENNUM_SNAP_MAGIC2;
                }
            }
            else if (FS_HANDLE_IS_SNAP_BOUNDARY(req->fileHandle))
            {
                handleInum = parentInum;
                if (!strcmp(dirent->name, ".."))
                {
                    // traversing up through a snapshot boundary
                    // parent is the .snapshot dir
                    genNum = FS_INODE_GENNUM_SNAP_MAGIC;
                    snapnum = 0;
                    snapType = FS_SNAPSHOT_DIR_TYPE;
                    snapInfo = parentInum;
                    dirent->inode = FS_SNAPSHOT_DIR_INODE;
                }
                else if (!strcmp(dirent->name, "."))
                {
                    // set up snapshot boundary attributes
                    genNum = entryInode->dinode.genNum;
                    snapType = FS_SNAPSHOT_BOUNDARY_TYPE;
                    snapInfo = parentSnapnum + FS_SNAPSHOT_ID_INODE_FIRST;
                    // also reset dirent inum
                    dirent->inode = snapInfo;
                }
                else
                {
                    handleInum = effectiveInum;
                    genNum = entryInode->dinode.genNum;
                }
            }
            else if (snapnum &&
                     !strcmp(dirent->name, "..") &&
                     (dirent->inode == req->fileHandle.mntPnt))
            {
                // dotdot of the dirent crosses the snap boundary
                // ie dotdot from /mnt/d0/.snapshot/snap -> /mnt/.snapshot/snap
                handleInum = effectiveInum;
                genNum = entryInode->dinode.genNum;
                snapType = FS_SNAPSHOT_BOUNDARY_TYPE;
                snapInfo = parentSnapnum + FS_SNAPSHOT_ID_INODE_FIRST;
            }
            else
            {
                handleInum = effectiveInum;
                genNum = entryInode->dinode.genNum;
            }

            fs_handleInit(context->fileSystem,
                          &direntPlus->handle,
                          snapnum,
                          handleInum,
                          genNum,
                          req->fileHandle.mntPnt);

            fs_doGetAttr(context->volId, &direntPlus->attr, entryInode,
                         snapType, snapInfo,
                         context->reqHdr->proto, FS_POST_ATTRS);

            fs_inodeRelease(context, entryInode);
        }

        direntPlus->dirEntry = *dirent;
        memcpy(direntPlus->dirEntry.name, dirent->name, dirent->nameLen);

        FS_READDIR_ADV_ENTRY_PTR(dirent);
        FS_READDIRPLUS_ADV_ENTRY_PTR(direntPlus);
    }
    context->respSize += ((char *)direntPlus - (char *)&(resp->dirEntry[0]));

    resp->numEntriesThis = numThis;
    resp->eof = numThis < numEntries ? 0 : eof;
    resp->cookieVerif = 0;

    fs_ramDealloc(FS_RAM_CORE_MISC, tmpDirents);

    return NFX_OK;

} /* fs_readDirPlusThread */


/**
 * Handle the STAT request.
 */
int32
fs_statThread(
    fs_Context      *context)
{
    fs_FileSystem   *fileSystem;
    fs_MsgStatResp  *resp;
    fs_MsgStatReq   *req;
    fs_Inode        *inode;
    fs_Attr         *attrp;
    int32           rc;

    resp = (fs_MsgStatResp *) context->respHdr;
    req = (fs_MsgStatReq *) context->reqHdr;
    attrp = &resp->attr;

    FS_TRACE(context, FS_TRACE_REQ,
             ("%s xid 0x%x statfs snap %lld inum %lld gen %d",
             fs_protoStr[context->reqHdr->proto],
             context->reqHdr->clientId,
             FS_HANDLE_GET_SNAP(req->fsHandle.inum),
             FS_HANDLE_GET_INUM(req->fsHandle.inum),
             req->fsHandle.gen));

    if (fs_handleIsSnapMagic(&req->fsHandle))
    {
        inode = I_GET_SHARED(0, FS_SNAPSHOT_DIR_INODE);
    }
    else
    {
        inode = I_GET_SHARED(FS_HANDLE_GET_SNAP(req->fsHandle.inum),
                             FS_HANDLE_GET_INUM(req->fsHandle.inum));
    }
    if (inode == NULL)
    {
        return FS_ERR_STALE;
    }
    if(!fs_handleIsValid(context, &req->fsHandle, inode))
    {
        fs_inodeRelease(context, inode);
        return FS_ERR_STALE;
    }

    fileSystem = fs_fileSystemForVolume(context->volId);
    if (fileSystem != NULL)
    {
        uint64  avail;

        rc = NFX_OK;

        // Blocks
        resp->stat.fsBlockSize = fileSystem->super->blkSize;
        resp->stat.numBlks = fs_numBlocksUserMax(fileSystem);
        if (!fs_haveMinfreeAvail(fileSystem))
        {
            avail = 0;
        }
        else
        {
            avail = fs_numBlocksAvail(fileSystem); 
        }
        resp->stat.numBlksAvailUser = resp->stat.numBlksAvail = avail;

        // Inodes
        if (!fs_haveMininodesAvail(fileSystem))
        {
            avail = 0;
        }
        else
        {
            avail = fs_numInodesAvail(fileSystem); 
        }
        resp->stat.numInodes = fs_numInodesUserMax(fileSystem);
        resp->stat.numInodesAvail = resp->stat.numInodesAvailUser = avail;

        // Other stuff
#define FS_PREF (8 * 1024)
        resp->stat.avgReadSize = FS_PREF;
        resp->stat.avgWriteSize = FS_PREF;
#undef FS_PREF
    }
    else
    {
        rc = FS_ERR_STALE;
    }

    fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->fsHandle.inum),
                    FS_POST_ATTRS, &attrp, NULL, FALSE);
    fs_inodeRelease(context, inode);

    context->respSize = sizeof(fs_MsgStatResp);

    return rc;

} /* fs_statThread */


/**
 * Handle the FSINFO request.
 */
int32
fs_fsinfoThread(
    fs_Context      *context)
{
    fs_FileSystem       *fileSystem;
    fs_MsgFsinfoResp    *resp;
    fs_MsgFsinfoReq     *req;
    fs_Inode            *inode;
    fs_Attr             *attrp;
    int32               rc;

    resp = (fs_MsgFsinfoResp *) context->respHdr;
    req = (fs_MsgFsinfoReq *) context->reqHdr;
    attrp = &resp->attr;

    FS_TRACE(context, FS_TRACE_REQ,
             ("%s xid 0x%x fsinfo snap %lld inum %lld gen %d",
             fs_protoStr[context->reqHdr->proto],
             context->reqHdr->clientId,
             FS_HANDLE_GET_SNAP(req->fsHandle.inum),
             FS_HANDLE_GET_INUM(req->fsHandle.inum),
             req->fsHandle.gen));

    if (fs_handleIsSnapMagic(&req->fsHandle))
    {
        inode = I_GET_SHARED(0, FS_SNAPSHOT_DIR_INODE);
    }
    else
    {
        inode = I_GET_SHARED(FS_HANDLE_GET_SNAP(req->fsHandle.inum),
                             FS_HANDLE_GET_INUM(req->fsHandle.inum));
    }
    if (inode == NULL)
    {
        return FS_ERR_STALE;
    }
    if(!fs_handleIsValid(context, &req->fsHandle, inode))
    {
        fs_inodeRelease(context, inode);
        return FS_ERR_STALE;
    }

    fileSystem = fs_fileSystemForVolume(context->volId);
    if (fileSystem != NULL)
    {
        rc = NFX_OK;

// 64K support (HP defaults to this) is not ready yet,
// BUGID is 720
#define FS_PREF (32 * 1024)
        // TODO - correct response here
        resp->fsinfo.maxFileSize = -1;
        resp->fsinfo.readMax = FS_PREF;
        resp->fsinfo.readPref = FS_PREF;
        resp->fsinfo.readMult =  FS_PREF;
        resp->fsinfo.writeMax = FS_PREF;
        resp->fsinfo.writePref = FS_PREF;
        resp->fsinfo.writeMult = FS_PREF;
        resp->fsinfo.readDirPref = 8192;
        resp->fsinfo.timeDelta.secs = 0;
        resp->fsinfo.timeDelta.nsecs = 1;
        resp->fsinfo.fsProperties = (FS_PROP_LINK |
                                     FS_PROP_SYMLINK |
                                     FS_PROP_HOMO |
                                     FS_PROP_CANSETTIME);
#undef FS_PREF

    }
    else
    {
        rc = FS_ERR_STALE;
    }

    fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->fsHandle.inum),
                    FS_POST_ATTRS, &attrp, NULL, FALSE);
    fs_inodeRelease(context, inode);

    context->respSize = sizeof(fs_MsgFsinfoResp);

    return rc;

} /* fs_fsinfoThread */


/**
 * Handle the PATHCONF request.
 */
int32
fs_pathconfThread(
    fs_Context      *context)
{
    fs_FileSystem       *fileSystem;
    fs_MsgPathconfResp  *resp;
    fs_MsgPathconfReq   *req;
    fs_Inode            *inode;
    fs_Attr             *attrp;
    int32               rc;
    fs_SnapType         snapType = FS_NO_SNAPSHOT_TYPE;
    uint64              snapInfo = 0;

    resp = (fs_MsgPathconfResp *) context->respHdr;
    req = (fs_MsgPathconfReq *) context->reqHdr;
    attrp = &resp->attr;

    FS_TRACE(context, FS_TRACE_REQ,
             ("%s xid 0x%x pathconf snap %lld inum %lld gen %d",
             fs_protoStr[context->reqHdr->proto],
             context->reqHdr->clientId,
             FS_HANDLE_GET_SNAP(req->fsHandle.inum),
             FS_HANDLE_GET_INUM(req->fsHandle.inum),
             req->fsHandle.gen));

    if (fs_handleIsSnapMagic(&req->fsHandle))
    {
        inode = I_GET_SHARED(0, FS_SNAPSHOT_DIR_INODE);
    }
    else
    {
        inode = I_GET_SHARED(FS_HANDLE_GET_SNAP(req->fsHandle.inum),
                             FS_HANDLE_GET_INUM(req->fsHandle.inum));
    }
    if (inode == NULL)
    {
        return FS_ERR_STALE;
    }
    if(!fs_handleIsValid(context, &req->fsHandle, inode))
    {
        fs_inodeRelease(context, inode);
        return FS_ERR_STALE;
    }

    fileSystem = fs_fileSystemForVolume(context->volId);
    if (fileSystem != NULL)
    {
        rc = NFX_OK;

        // TODO - correct response here
        resp->pathconf.linkMax = FS_INODE_LINK_MAX;
        resp->pathconf.nameMax = FS_NAME_MAX;
        resp->pathconf.noTrunc = TRUE;
        resp->pathconf.chownRestricted = TRUE;
        resp->pathconf.caseInsensitive = FALSE;
        resp->pathconf.casePreserving = TRUE;
    }
    else
    {
        rc = FS_ERR_STALE;
    }

    if (inode->inum == FS_SNAPSHOT_DIR_INODE)
    {
        snapType = FS_SNAPSHOT_DIR_TYPE;
        snapInfo = FS_HANDLE_GET_INUM(req->fsHandle.inum);
    }
    else if (FS_HANDLE_IS_SNAP_BOUNDARY(req->fsHandle))
    {
        snapType = FS_SNAPSHOT_BOUNDARY_TYPE;
        snapInfo = FS_HANDLE_GET_SNAP(req->fsHandle.inum) + FS_SNAPSHOT_ID_INODE_FIRST;
    }

    fs_getSnapRespAttrs(context, inode, snapType, snapInfo,
                        FS_POST_ATTRS, &attrp, NULL, FALSE);
    fs_inodeRelease(context, inode);

    context->respSize = sizeof(fs_MsgPathconfResp);

    return rc;

} /* fs_pathconfThread */


/**
 * Handle the COMMIT request.
 */
int32
fs_commitThread(
    fs_Context      *context)
{
    fs_MsgCommitResp    *resp;
    fs_MsgCommitReq     *req;
    fs_Inode            *inode;
    fs_Attr             *attrp;
    uint64              inum;

    resp = (fs_MsgCommitResp *) context->respHdr;
    req = (fs_MsgCommitReq *) context->reqHdr;
    resp->numAttrs = 0;
    attrp = &resp->attr[0];
    context->respSize = sizeof(fs_MsgCommitResp);

    if (context->fileSystem->state == FS_VOL_STATE_ONLINE_RDONLY) {
        return FS_ERR_ROFS;
    }

    FS_TRACE(context, FS_TRACE_REQ,
             ("%s xid 0x%x commit snap %lld inum %lld gen %d",
             fs_protoStr[context->reqHdr->proto],
             context->reqHdr->clientId,
             FS_HANDLE_GET_SNAP(req->fsHandle.inum),
             FS_HANDLE_GET_INUM(req->fsHandle.inum),
             req->fsHandle.gen));

    if (fs_handleIsSnapMagic(&req->fsHandle))
    {
        inode = I_GET_SHARED(0, FS_SNAPSHOT_DIR_INODE);
    }
    else
    {
        inode = I_GET_SHARED(FS_HANDLE_GET_SNAP(req->fsHandle.inum),
                             FS_HANDLE_GET_INUM(req->fsHandle.inum));
    }
    if (inode == NULL)
    {
        return FS_ERR_STALE;
    }
    if(!fs_handleIsValid(context, &req->fsHandle, inode))
    {
        fs_inodeRelease(context, inode);
        return FS_ERR_STALE;
    }
    if (!FS_I_IS_REG(inode->dinode.attrs))
    {
        fs_inodeRelease(context, inode);
        return FS_ERR_STALE;
    }

    // printf ("Commit      inum %8lld off %8lld len %5d vers %d \n",
    //   req->fsHandle.inum, req->offset, req->count, req->hdr.proto+1);

    fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->fsHandle.inum),
                    FS_PRE_ATTRS, &attrp, &resp->numAttrs, TRUE);

    // Block until all buffers affecting this inode
    // are stable on disk.
    if (fs_logSyncInodeBuf(context, inode) != NFX_OK)
    {
        // Timed out trying to sync data to disk.
        fs_inodeRelease(context, inode);

	    NFX_SPIN_LOCK(&fs_errStatsSpinId);
	    fs_errStats.resource[FS_RESOURCE_IO] += 1;
        context->fileSystem->nfsStats.errs[context->reqHdr->requestType] += 1;
		context->fileSystem->errStats.resource[FS_RESOURCE_IO] += 1;
	    NFX_SPIN_UNLOCK(&fs_errStatsSpinId);

        return FS_ERR_NO_RESOURCE;
    }

#ifdef NEW_ZF
    /* Don't zero fill (cancel previous zero fill records for this inode).
     */
    fs_logAddUserZeroRecord(context, inode, 0, 0, 0, FALSE, TRUE, NULL);
#endif // NEW_ZF
    
    fs_getRespAttrs(context, inode, FS_HANDLE_GET_INUM(req->fsHandle.inum),
                    FS_POST_ATTRS, &attrp, &resp->numAttrs, TRUE);

    inum = inode->inum;
    fs_inodeRelease(context, inode);

    // Block until all transactions affecting this inode
    // are stable in the log.
    fs_logSyncInodeLog(context, inum);

    resp->verifier = context->fileSystem->super->mountTime[0];

    return NFX_OK;

} /* fs_commitThread */

/**
 * Multi Component Lookup (MCL).
 */
static int32
fs_mcl(fs_Context * context,
       fs_Handle *  reqParentHandle,
       char *       reqObjName,
       fs_Handle *  rspParentHandle,
       fs_Handle *  rspHandle,
       int32 *      rspNumAttrs,
       fs_Attr *    rspAttr)
    
{
    fs_Inode *      inode = NULL;
    fs_Inode *      parentInode = NULL;
    fs_Handle       handle = *reqParentHandle;
    fs_Handle       parentHandle = *reqParentHandle;
    uint64          inum = 0;
    int32           rc = NFX_OK;
    uint64          parentInum = FS_HANDLE_GET_INUM(reqParentHandle->inum);
    uint32          parentSnap = FS_HANDLE_GET_SNAP(reqParentHandle->inum);
    char *          sep = "\\/";
    char *          component = NULL;
    char            name[FS_PATH_MAX];

    /* Check if the parent handle is correct.
     */
    parentInode = I_GET_SHARED(parentSnap, parentInum);
    inode = parentInode;
    if (parentInode == NULL)
    {
        rc = FS_ERR_BADHANDLE;
        goto fs_mcl_exit;
    }
    if (!fs_handleIsValid(context, reqParentHandle, parentInode))
    {
        rc = FS_ERR_STALE;
        goto fs_mcl_exit;
    }

    /* Go through each component of the object.
     */
    strcpy(name, reqObjName);
    for (component = strtok(name, sep);
         component;
         component = strtok(NULL, sep))
    {
        if (inum != 0)
        {
            fs_inodeRelease(context, parentInode);
            parentInode = inode;
            parentInum = inum;
        }
        
        /* Check if we have access to the parent inode.
         */
        if (!fs_permissionOkay(context, parentInode, FS_PERM_LOOKUP))
        {
            rc = FS_ERR_ACCESS;
            goto fs_mcl_exit;
        }

        /* Lookup for the component's inode in the parent inode.
         */
        inum = fs_dir_lookupent(context, parentInode,
                                component, strlen(component));
        if (inum == 0)
        {
            /* Component not found.
             */
            if (strtok(NULL, sep) == NULL)
            {
                rc = FS_ERR_NOENT;
            }
            else
            {
                rc = FS_ERR_BADPATH;
            }
            goto fs_mcl_exit;
        }
        else
        {
            /* Component found.
             * Get the inode of the component.
             */
            FS_LOCK_ORDER_SET_PARENT(parentInode->inum);
            inode = I_GET_SHARED((parentInum == FS_SNAPSHOT_DIR_INODE)
                                 ? FS_HANDLE_GET_SNAP(inum)
                                 : parentSnap,
                                 FS_HANDLE_GET_INUM(inum));
            assert(inode != NULL);
            parentHandle = handle;
            fs_handleInit(context->fileSystem,
                          &handle,
                          FS_HANDLE_GET_SNAP(parentInode->inum),
                          inum,
                          inode->dinode.genNum,
                          reqParentHandle->mntPnt);
        }
    } // for
    
  fs_mcl_exit:

    *rspHandle = handle;
    if (rspParentHandle != NULL)
    {
        *rspParentHandle = parentHandle;
    }

    if (rspAttr != NULL)
    {
        if (inode != NULL)
        {
            fs_getRespAttrs(context, inode, parentInum, FS_POST_ATTRS,
                            &rspAttr, rspNumAttrs, TRUE);
            fs_inodeRelease(context, inode);
        }
        
        if ((parentInode != NULL) && (parentInode != inode))
        {
            fs_getRespAttrs(context, parentInode, parentInum, FS_POST_PAR_ATTRS,
                            &rspAttr, rspNumAttrs, TRUE);
            fs_inodeRelease(context, parentInode);
        }
    }

    return(rc);

} /* fs_mcl */

/**
 * Handle the MULTI COMPONENT LOOKUP (MCL) request.
 */
int32
fs_mclThread(fs_Context * context)
{
    fs_MsgMclResp * resp = (fs_MsgMclResp *) context->respHdr;
    fs_MsgMclReq *  req = (fs_MsgMclReq *) context->reqHdr;
    int32           rc = NFX_OK;

    context->respSize = sizeof(fs_MsgMclResp);

    rc = fs_mcl(context,
                &(req->parentHandle),
                req->objName,
                &(resp->parentHandle),
                &(resp->handle),
                &(resp->numAttrs),
                &(resp->attr[0]));

    return(rc);
} /* fs_mclThread */

/*
 * Read ahead
 */

EXPORTED void
fs_readAhead(
    fs_Context      *context,
    fs_Inode        *inode,
    uint64          reqOffset,
    uint32          reqLen,
    int             meta        // Is this meta data i/o or not
)
{
    uint64      craOffset;
    uint32      craSize;
    int32       numIo;

    //printf("In read ahead at %lld for %d trigger %lld\n", reqOffset, reqLen,
    //  inode->raTrigger);

    NFX_SPIN_LOCK(&inode->spin);
    if (inode->readAheadState == NULL)
    {
        inode->readAheadState = fs_ramAlloc(FS_RAM_READ_AHEAD,
                                            sizeof(fs_ReadAheadState));
        bzero(inode->readAheadState, sizeof(fs_ReadAheadState));
        DLL_INIT(&inode->readAheadState->pendingReadList);
    }

    if (((reqOffset + reqLen) >= inode->dinode.size) ||
        ((reqOffset + reqLen) < inode->readAheadState->raTrigger) ||
		(inode->readAheadState->raOffset > inode->dinode.size)) {

        /*
         * We are past the end of file or we have not
         * yet reached the read ahead trigger. Don't do
         * any read ahead yet
         */

        NFX_SPIN_UNLOCK(&inode->spin);
        return;
    }
	//printf("do Read ahead for inum %lld offset %ld len %ld\n",
	//				inode->inum, reqOffset >> FS_BLOCK_SHIFT);

    /*
     * Detect wether we are doing sequential read or not
     */

    if (reqOffset == 0) {

        /*
         * Read ahead 1 block
         */

        inode->readAheadState->raSize = MAX(1, reqLen >> FS_BLOCK_SHIFT);
    } else if (inode->readAheadState->uxRead == 0) {
        if (inode->readAheadState->raSize == 0) {

            /*
             * ra size was turned down earlier because of
             * unexpected read, start small
             */

            inode->readAheadState->raSize = MAX(1, reqLen >> FS_BLOCK_SHIFT);
        } else if (inode->readAheadState->raSize <
                          fs_incoreTune->readAheadDefaultSize) {

            /*
             * We had performed a small read ahead earlier
             * and we are doing read ahead, get into higher
             * speed
             */

            inode->readAheadState->raSize = fs_incoreTune->readAheadDefaultSize;
        } else {

            /*
             * Keep on doubling the speed till we hit max
             */

            inode->readAheadState->raSize <<= 2;
        }
        if (inode->readAheadState->raSize >= fs_incoreTune->readAheadMaxSize) {
            inode->readAheadState->raSize = fs_incoreTune->readAheadMaxSize;
        }
    } else {

        /*
         * There was an unexpected read. Reduce our read ahead size
         * as client is not reading data fast enough and cache manager
         * is reclaiming the read ahead buffers
         */

        //printf("Unxepected read %x\n", inode->uxRead);
        if (inode->readAheadState->uxRead > 1) {
            inode->readAheadState->raSize = 0;
        } else {
            inode->readAheadState->raSize >>= 2;
            if (inode->readAheadState->raSize < reqLen) {
                inode->readAheadState->raSize = 0;
            }
        }
        inode->readAheadState->uxRead = 0;
        
    }
    inode->readAheadState->raOffset = MAX(inode->readAheadState->raOffset, reqOffset + reqLen);

    /*
     * Round up the ra offset to the next block boundary
     */

    inode->readAheadState->raOffset =
                 (inode->readAheadState->raOffset + FS_BLOCK_SIZE - 1) &
                        ~(FS_BLOCK_SIZE - 1);

    /*
     * We have to perform read ahead
     */

    // Tweak the raTrigger point while tunning the system.

    inode->readAheadState->raTrigger = inode->readAheadState->raOffset;
    craOffset = inode->readAheadState->raOffset;
    craSize = inode->readAheadState->raSize;
	craSize = MIN(craSize,
				  ((inode->dinode.size - craOffset + FS_BLOCK_SIZE - 1) /
				   FS_BLOCK_SIZE));
    inode->readAheadState->raOffset += (inode->readAheadState->raSize << FS_BLOCK_SHIFT);
    NFX_SPIN_UNLOCK(&inode->spin);
	//printf("New RaOffset for inum %lld is %lld\n", inode->inum, inode->raOffset);


    /*
     * Do the i/o
     */

    if (craSize == 0) {

        /*
         * Nothing to do
         */


        return;
    }

    numIo = fs_bufGetPageAsync(context, inode, craOffset >> FS_BLOCK_SHIFT, craSize,
                               meta ? FS_BUFLIST_META : FS_BUFLIST_USER);

    if (meta)
    {
        NFX_SPIN_LOCK(&fs_ioSpin);
        fs_ioTotalMetaReadAheads += numIo;
        NFX_SPIN_UNLOCK(&fs_ioSpin);

        NFX_SPIN_LOCK(&context->fileSystem->spin);
        context->fileSystem->ioTotalMetaReadAheads += numIo;
        NFX_SPIN_UNLOCK(&context->fileSystem->spin);
    }
    else
    {
        NFX_SPIN_LOCK(&fs_ioSpin);
        fs_ioTotalUserReadAheads += numIo;
        NFX_SPIN_UNLOCK(&fs_ioSpin);

        NFX_SPIN_LOCK(&context->fileSystem->spin);
        context->fileSystem->ioTotalUserReadAheads += numIo;
        NFX_SPIN_UNLOCK(&context->fileSystem->spin);
    }
}

#endif // SSC
