/*
 * Name:  fs-sim.c
 *
 * RCS:  $Id: fs-sim.c,v 1.242 2002/12/10 22:25:39 bobonfeline Exp $
 *
 * Description:  Code for providing a UNIX based Eagle simulatator for
 *               filesystem testing.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */

#if !defined(SSC)

#include "nfsproto.h"
#include "sfs_c_nfs.h"      // NFS V3 prototypes
#include "mount.h"
#include "fs-includes.h"
#include "mcpu-sim.h"
#include "btrace.h"
#include "simthread.h"

#include <sys/wait.h>
#include <sys/uio.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <fcntl.h>
#include <time.h>

#include "exports/fssf-api.h"

#define NFS_VERSION_V2 2
#define NFS_VERSION_V3 3
#define MOUNTVERS_V3   3

#define FS_REQ_ALL_ATTR  0xFF

nfx_spinlock_t wait_time_lock;

pthread_mutexattr_t fs_simMutexAttr;

mountres3 *
mountproc_mnt_3(
    dirpath         *ap,
    struct svc_req  *rp);



//   M E M O R Y   A L L O C A T O R   D E B U G G I N G


/**
 * The Boehm garbage collector, used as a garbage collector.
 */
#ifdef LEAK_DETECTOR

#ifdef EFENCE
#error define exactly one of LEAK_DETECTOR or EFENCE
#endif

void *malloc(size_t n) 
{
    return GC_MALLOC(n);
}

void *calloc(size_t m, size_t n)
{
    return GC_MALLOC(m * n);
}

void *realloc(void *p, size_t n)
{
    return GC_REALLOC(p, n);
}

void free(void *p)
{
    GC_FREE(p);
}

void fs_memCheckLeaks(void)
{
    GC_gcollect();
}

static void GC_format_frame(
    char *buf, int buflen, GC_word pc, int arglen, GC_word *invertedArgs)
{
    int n = btrace_sprint_function(buf, (void *) pc);

    if (invertedArgs != NULL) {
        int i;
        n += sprintf(&buf[n], "(");
        for (i = 0; i < arglen; i++) {
            GC_word arg = ~invertedArgs[i];
            if (i > 0)
                n += sprintf(&buf[n], ", ");
            if (-0x100 < arg && arg < 0x1000)
                n += sprintf(&buf[n], "%ld", arg);
            else
                n += sprintf(&buf[n], "%#lx", arg);
        }
        n += sprintf(&buf[n], ")");
    }

    n += btrace_sprint_source_line(&buf[n], (void *) pc);
}

void fs_memInit(void)
{
    GC_find_leak = 1;
    GC_set_format_frame_func(GC_format_frame);
    btrace_init();
}

#else  // !LEAK_DETECTOR

void fs_memCheckLeaks(void)
{
}

#endif // LEAK_DETECTOR


/**
 * Electric fence.
 */
#ifdef EFENCE
extern void *ef_malloc(size_t size);
extern void ef_free(void *ptr);
pthread_mutex_t ef_alloc_lock;

void fs_memInit(void)
{
    // initialize the electric fence heap corruption detector
    pthread_mutex_init(&ef_alloc_lock, NULL);
}

#endif


/**
 * Default memory allocator.
 */
#if !defined(EFENCE) && !defined(LEAK_DETECTOR)
void fs_memInit(void)
{
}
#endif


/**
 * EEE malloc.
 */
void *
eee_ramAlloc(uint32 numBytes)
{
    return fs_simAlloc(numBytes);
}


/**
 * EEE malloc aligned.
 */
void *
eee_ramAllocAligned(uint32 numBytes)
{
    return fs_simAlloc(numBytes);
}


/**
 * EEE malloc.
 */
void *
eee_ramAllocZero(uint32 numBytes)
{
    void * p = fs_simAlloc(numBytes);

    if (p != NULL)
        bzero(p, numBytes);
    return(p);
}


/**
 * EEE free.
 */
int32
eee_ramDealloc(void *ptr)
{
    fs_simDealloc(ptr);

    return NFX_OK;
}


/**
 * EEE system buffer free.
 */
void
eee_deallocateBuffer(void *ptr)
{
    fs_simDealloc(ptr);
}

/**
 * EEE system desc free.
 */
void
eee_deallocateDesc(eee_descPtr_t edesc)
{

}


/**
 * String stuff in sm-libc but not in linux.
 */
unsigned long
atolx (
        char8   * p
)
{
        int32   digit;
        unsigned long   value;
        uint32  c;

        value = 0;

        for (; isspace (*p); p++);      /* gobble up leading whitespace */

        if ((*p == '0') && (*(p+1) == 'x'))
                p+=2;

        for (; *p; p++) {
                value *= 16;
                c = *p;
                if (c >= '0' && c <= '9')
                        digit = c - '0';
                else if (c >= 'a' && c <= 'f')
                        digit = c - 'a' + 10;
                else if (c >= 'A' && c <= 'F')
                        digit = c - 'A' + 10;
                else
                        break;
                value += digit;
        }

        return (value);

}       /* atolx() */
char8 *
strnchr (
        char8   * p,
        char8   c,
    int32   n
)
{
    char8 * stop;

        if (!p)
                return (0);

    if (n <= 0)
        return (0);

    stop = p + n;

        for (; *p; p++) {
        if (p == stop)
            break;
            if (*p == c)
                    return (p);
    }

        return (0);

}       /* strnchr() */
char           *
strnrchr (p, c, n)
     char           *p, c;
     int            n;
{
    int             i;

    if (!p)
        return (0);

    if (n <= 0)
    return (0);

    for (i = n-1; i >= 0; i--) {
            if (p[i] == c)
                return (&p[i]);
    }
    return (0);
}

/**
 * Allocate memory.
 */
void *
fs_simAlloc(numBytes)
{
#ifdef EFENCE
    void *ptr;
    int32 rc;

    rc = pthread_mutex_lock(&ef_alloc_lock);
    ptr = ef_malloc(numBytes);
    rc |= pthread_mutex_unlock(&ef_alloc_lock);

    FS_ASSERT(rc == 0);
    FS_ASSERT(((int)ptr & 0x07) == 0);

#else
    void *ptr = malloc(numBytes);
#endif

    //printf("allocing %p\n", ptr);

    return ptr;
}

/**
 * Deallocate memory.
 */
void
fs_simDealloc(void *ptr)
{
    //printf("freeing %p\n", ptr);

#ifdef EFENCE
    int32 rc;
    
    rc = pthread_mutex_lock(&ef_alloc_lock);
    ef_free(ptr);
    rc |= pthread_mutex_unlock(&ef_alloc_lock);
    FS_ASSERT(rc == 0);
#else
    free(ptr);
#endif

}


/**
 * Return the total SAC size in bytes.
 */
uint64
sbm_totalSacSize()
{
    return mcpu_sim_totalSacSize();
}


/**
 * Report the total size of the heap.
 */
uint64
eee_ramTotalHeapSize()
{
    // Simulate a limited heap space, this is
    // used to size our caches.
    return (256 * 1024 * 1024);
}

/**
 * Report the total size of the flex.
 */
uint64
eee_ramTotalFlexSize()
{
    // Simulate a limited flex space, this is
    // used to size our caches.
    return (256 * 1024 * 1024);
}


/**
 * Report available unallocated heap space.
 */
uint64
eee_ramTotalHeapAvail()
{
    // We don't try to track this in the simulator,
	// so just return something.
    return (256 * 1024 * 1024);
}


/**
 * Get the caller's address
 */
address_t
get_ra()
{
    address_t   ra_list[4];
	int		    n;

    n = backtrace((void**) ra_list, sizeof(ra_list) / sizeof(ra_list[0]));

/*
    {
        char name[64];
        int i;
        printf("get_ra, n = %d\n", n);
        for (i=0; i<n; i++)
        {
            btrace_sprint_function(name, ra_list[i]);
            printf("    %s\n", name);
        }
    }
*/

    return (ra_list[2]);
}


/* input: seconds since  1 Jan 1970
   output "Mon Oct  8 11:21:00 2001" */

char *
eee_humandate(uint32 time)
{    
    int i;
    char *date = ctime((time_t *)&time);
    for (i=20;i<30;i++)
    {
        if (date[i] == '\n')
        {
            date[i]=0;
            break;
        }
    }
    
    return (date);
}

eeeGMT_t bootGMTTime = {0,0};
eeeGMT_t sscGMTTime = {0,0};
eeeRTC_t sscRTC = {0};
eeeRTC_t eeeLastRTC = {0};

/*-----------------------------------------------------------------------------
 * Simulated volume manager stuff
 */
char *volName[FS_SIM_VOL_MAX + 1] = { "FS_EAGLE_SIM_BAD_VOL",
                                      "FS_EAGLE_SIM_VOL_1",
                                      "FS_EAGLE_SIM_VOL_2",
                                      "FS_EAGLE_SIM_VOL_3" };
int volFd[FS_SIM_VOL_MAX + 1] = {0,0,0,0};            
pthread_mutex_t volMutex;
pthread_cond_t  volCondVar;
DLL_HDR(evm_IOReq_t)   fs_simEvmHdr;
int32 fs_simEvmSignal = 1;

// Turn this on to get async EVM.
#define FS_SIM_EVM_THREAD 1

/*-----------------------------------------------------------------------------
 * Simulated e-layer state machine stuff.
 */
esm_instance_t       *fs_simEsmInstance[FS_SIM_ESM_MAX];
nfx_spinlock_t       fs_simEsmInstLock;
int32       fs_simEsmNextEntry = 0;
ushort16    fs_simEsmGeneration = 1;
DLL_HDR(esm_instance_t)   fs_simEsmReadyQueueHdr;

pthread_key_t  fs_simKeyCpuId;

/*-----------------------------------------------------------------------------
 * Simulated embedded executive
 */
uint32 nfxMySlotId;
pthread_mutex_t fs_simFpExecMutex;
fs_SimExec fs_simFpExec[FS_SIM_NUM_FP_EXEC_THREADS];
fs_SimExec fs_simSscExec;
fs_SimExec fs_simMcpuExec;

int32 fs_simNumReqThreadsWaiting = 0;
int32 fs_simNumReqThreadsCreated = 0;
fs_SimReqThread fs_simReqThread[FS_SIM_EXEC_REQ_THREADS_MAX];
pthread_key_t  fs_simKeyReqThreadNum;

// queue of pending yielding esm (threads) requests
fs_SimThreadJobQueueHead fs_simThreadJobQueue;

// queue of received filesystem requests
fs_SimMsgQueueHead  fs_simFpRxQueue[FS_SIM_NUM_FP_EXEC_THREADS];

// queue of received ssc responses
fs_SimMsgQueueHead  fs_simSscRxQueue;

// queue of received Mcpu requests
fs_SimMsgQueueHead  fs_simMcpuQueue;

// queue of received fssf responses
fs_SimMsgQueueHead  fs_simFssfQueue;


/*-----------------------------------------------------------------------------
 * NFS daemon.
 */
#define FS_SIM_NFS_TO_CLIENT_PORT       2050

// Internal sockets between the mux thread and the worker threads.
int32 fs_simNfsIntSockets[FS_SIM_NFS_THREAD_MAX];

// The external (to client) socket.
int32 fs_simNfsSrvrSocket;

// By default, don't register with portmapper.
boolean fs_simPortmapNfs = FALSE;

/*-----------------------------------------------------------------------------
 * Mount daemon.
 */
boolean fs_simPortmapMount = FALSE;
#define FS_SIM_MOUNT_PORT_MAPPED    687
#define FS_SIM_MOUNT_PORT_UNMAPPED  2060
int32 fs_simMountPort = FS_SIM_MOUNT_PORT_UNMAPPED;

extern int
pmap_set(u_long prognum, u_long versnum, u_long protocol, u_short port);

extern int
pmap_unset(u_long prognum, u_long versnum);

// just a value, diff than fs_cod_get_new_cod
// and will be made different for each NFS_RX process !!!
static int slot_id = 4 ;

fs_SimNfsThread fs_simNfsThread[FS_SIM_NFS_THREAD_MAX];

typedef struct fs_SimNfsDispatchTbl
{
    char        *name;
    bool_t      (*xdrArgument)();
    bool_t      (*xdrResult)();
    int32       (*buildReq)(void *rqst, void *arg, int32 *length, void *buf);
    void        *(*buildResp)(void *fsResp);
    void        (*freeResp)(void *fsRresp);
} fs_SimNfsDispatchTbl;

fs_SimNfsDispatchTbl fs_simNfsDispatchTbl[] = {
    { "null", xdr_void, xdr_void, fs_simNullReq2,
            fs_simNullResp2, NULL },
    { "getattr", xdr_nfs_fh, xdr_attrstat, fs_simGetAttrReq2,
            fs_simGetAttrResp2, fs_simFreeResp2 },
    { "setattr", xdr_sattrargs, xdr_attrstat, fs_simSetAttrReq2,
            fs_simSetAttrResp2, fs_simFreeResp2 },
    { "getroot", xdr_void, xdr_void, fs_simGetRootReq2,
            fs_simGetRootResp2, fs_simFreeResp2 },
    { "lookup", xdr_diropargs, xdr_diropres, fs_simLookupReq2,
            fs_simLookupResp2, fs_simFreeResp2 },
    { "readlink", xdr_nfs_fh, xdr_readlinkres, fs_simReadLinkReq2,
            fs_simReadLinkResp2, fs_simFreeReadLinkResp2 },
    { "read", xdr_readargs, xdr_readres, fs_simReadReq2,
            fs_simReadResp2, fs_simFreeReadResp2 },
    { "bad", xdr_void, xdr_void, fs_simBadReq2,
            fs_simBadResp2, fs_simFreeResp2 },
    { "write", xdr_writeargs, xdr_attrstat, fs_simWriteReq2,
            fs_simWriteResp2, fs_simFreeResp2 },
    { "create", xdr_createargs, xdr_diropres, fs_simCreateReq2,
            fs_simCreateResp2, fs_simFreeResp2 },
    { "remove", xdr_diropargs, xdr_nfsstat, fs_simRemoveReq2,
            fs_simRemoveResp2, fs_simFreeResp2 },
    { "rename", xdr_renameargs, xdr_nfsstat, fs_simRenameReq2,
            fs_simRenameResp2, fs_simFreeResp2 },
    { "link", xdr_linkargs, xdr_nfsstat, fs_simLinkReq2,
            fs_simLinkResp2, NULL },
    { "symlink", xdr_symlinkargs, xdr_nfsstat, fs_simSymlinkReq2,
            fs_simSymlinkResp2, NULL },
    { "mkdir", xdr_createargs, xdr_diropres, fs_simMkdirReq2,
            fs_simMkdirResp2, fs_simFreeResp2 },
    { "rmdir", xdr_diropargs, xdr_nfsstat, fs_simRmdirReq2,
            fs_simRmdirResp2, fs_simFreeResp2 },
    { "readdir", xdr_readdirargs, xdr_readdirres, fs_simReaddirReq2,
            fs_simReaddirResp2, fs_simFreeReaddirResp2 },
    { "statfs", xdr_nfs_fh, xdr_statfsres, fs_simStatfsReq2,
            fs_simStatfsResp2, fs_simFreeResp2 },
};

fs_SimNfsDispatchTbl fs_simNfsDispatchTblV3[] = {
    { "null", xdr_void, xdr_void, fs_simNullReq3,                                   // 0
            fs_simNullResp3, NULL },
    { "getattr", xdr_GETATTR3args, xdr_GETATTR3res, fs_simGetAttrReq3,              // 1
            fs_simGetAttrResp3, fs_simFreeResp3 },
    { "setattr", xdr_SETATTR3args, xdr_SETATTR3res, fs_simSetAttrReq3,              // 2
            fs_simSetAttrResp3, fs_simFreeResp3 },
    { "lookup", xdr_LOOKUP3args, xdr_LOOKUP3res, fs_simLookupReq3,                  // 3
            fs_simLookupResp3, fs_simFreeResp3 },
    { "access", xdr_ACCESS3args, xdr_ACCESS3res, fs_simAccessReq3,                 // 4
            fs_simAccessResp3, fs_simFreeResp3 },
    { "readlink", xdr_READLINK3args, xdr_READLINK3res, fs_simReadLinkReq3,          // 5
            fs_simReadLinkResp3, fs_simFreeReadLinkResp3 },
    { "read", xdr_READ3args, xdr_READ3res, fs_simReadReq3,                          // 6
            fs_simReadResp3, fs_simFreeReadResp3 },
    { "write", xdr_WRITE3args, xdr_WRITE3res, fs_simWriteReq3,                      // 7
            fs_simWriteResp3, fs_simFreeResp3 },
    { "create", xdr_CREATE3args, xdr_CREATE3res, fs_simCreateReq3,                  // 8
            fs_simCreateResp3, fs_simFreeResp3 },
    { "mkdir", xdr_MKDIR3args, xdr_MKDIR3res, fs_simMkdirReq3,                      // 9
            fs_simMkdirResp3, fs_simFreeResp3 },
    { "symlink", xdr_SYMLINK3args, xdr_SYMLINK3res, fs_simSymlinkReq3,              // 10
            fs_simSymlinkResp3, fs_simFreeResp3 },
    { "mknod", xdr_MKNOD3args, xdr_MKNOD3res, fs_simMknodReq3,                      // 11
            fs_simMknodResp3, fs_simFreeResp3 },
    { "remove", xdr_REMOVE3args, xdr_REMOVE3res, fs_simRemoveReq3,                  // 12
            fs_simRemoveResp3, fs_simFreeResp3 },
    { "rmdir", xdr_RMDIR3args, xdr_RMDIR3res, fs_simRmdirReq3,                      // 13
            fs_simRmdirResp3, fs_simFreeResp3 },
    { "rename", xdr_RENAME3args, xdr_RENAME3res, fs_simRenameReq3,                  // 14
            fs_simRenameResp3, fs_simFreeResp3 },
    { "link", xdr_LINK3args, xdr_LINK3res, fs_simLinkReq3,                          // 15
            fs_simLinkResp3, fs_simFreeResp3 },
    { "readdir", xdr_READDIR3args, xdr_READDIR3res, fs_simReaddirReq3,              // 16
            fs_simReaddirResp3, fs_simFreeReaddirResp3 },
    { "readdirplus", xdr_READDIRPLUS3args, xdr_READDIRPLUS3res, fs_simReadDirPlusReq3, // 17 
            fs_simReadDirPlusResp3, fs_simFreeReadDirPlusResp3 },
    { "fsstat", xdr_FSSTAT3args, xdr_FSSTAT3res, fs_simFsstatReq3,                  // 18
            fs_simFsstatResp3, fs_simFreeResp3 },
    { "fsinfo", xdr_FSINFO3args, xdr_FSINFO3res, fs_simFsinfoReq3,                  // 19
            fs_simFsinfoResp3, fs_simFreeResp3 },
    { "pathconf", xdr_PATHCONF3args, xdr_PATHCONF3res, fs_simPathConfReq3,          // 20
            fs_simPathConfResp3, fs_simFreeResp3 },
    { "commit", xdr_COMMIT3args, xdr_COMMIT3res, fs_simCommitReq3,                  // 21
            fs_simCommitResp3, fs_simFreeResp3 },
};


/*-----------------------------------------------------------------------------
 * Simulated FPNV stuff.
 */
char *fpnvName[FS_SIM_VOL_MAX + 1] =    { "FS_EAGLE_SIM_BAD_FPNV",
                                          "FS_EAGLE_SIM_FPNV_1",
                                          "FS_EAGLE_SIM_FPNV_2",
                                          "FS_EAGLE_SIM_FPNV_3" };
int fpnvFd[FS_SIM_VOL_MAX + 1] = {0,0,0,0};


/**
 * Flag indicating that we need to cleanup after readline.
 */
static boolean readlineInUse = FALSE;


/**
 * Stuff for testing exception handling.
 */
int32 bobDbgIoLate = 0;
int32 bobDbgIoFailStart = 0;
int32 bobDbgIoFailFinish = 0;


/**
 * Perform IO request to volume.
 */
int32
fs_simEvmDoReq(
    evm_IOReq_t     *ioreq)
{
    evm_IOStat_t    iostat;
    int             fd;
    int             openFlags = 0;
    ssize_t         numWritten;
    char            *datap;
    uint64          offset, size, seekset;
    int32           rc;

    // EVM expects the size and offset in units of EVM_LBLK_SIZE,
    // so convert them to byte units for our simulated EVM.
    offset = ioreq->offset * EVM_LBLK_SIZE;
    size = ioreq->reqSize * EVM_LBLK_SIZE;

    if (ioreq->reqType == EVM_WRITE_REQ)
    {
        FS_TRACE(NULL, FS_TRACE_IO,
         ("vol %llx evm_IOReq W: datap %p cod %llx offset %lld num %lld",
             ioreq->id.whole_thing, ioreq->wDataHandle.dataPtr,
             ioreq->wDataHandle.codHandle, offset, size));
        /*
         * Sanity checks.
         */
        if (ioreq->id.details.u.id > FS_SIM_VOL_MAX)
        {
            fs_error(NULL, FS_SIM_ARGS_ERR, "evm_IOReq");
        }

        /*
         * Open/create the volume.
         */
        fd = volFd[ioreq->id.details.u.parts.clusterUnique];
        if (fd == 0)
        {
            fd = open64(  volName[ioreq->id.details.u.id],
                        openFlags | O_RDWR | O_CREAT,
                        S_IREAD | S_IWRITE);
            if (fd <= 0)
            {
                fs_error(NULL, FS_SIM_OPEN_ERR, "evm_IOReq");
            }
            volFd[ioreq->id.details.u.parts.clusterUnique] = fd;
        }

        seekset = lseek64(fd,
                          offset,
                          SEEK_SET);
        if (seekset < 0)
        //if (rc != offset) // according to man page, but success rtns 0!!!
        {
            printf("failed to seek to %lld for write\n", offset);
            perror("write seek");
            fs_error(NULL, FS_SIM_LSEEK_ERR, "evm_IOReq");
        }

        if (ioreq->dataType == EVM_META_DATA_REQ)
        {
            numWritten = write( fd,
                                ioreq->wDataHandle.dataPtr,
                                size);
        }
        else
        {
            numWritten = mcpu_sim_evm_write_data (ioreq->wDataHandle.codHandle,
                                                  fd,
                                                  size);
        }
        if (numWritten != size)
        {
            perror("vol_write");
            printf("vol_write: numWritten: %d, reqSize: %lld\n",
                        (int32) numWritten, size);
            fs_error(NULL, FS_SIM_IO_ERR, "numBytes");
        }

        iostat.rDataHandle.dataPtr = NULL;
        iostat.status = NFX_OK;
        iostat.devErrorCode = 0;
        iostat.clientPrivate = ioreq->clientPrivate;
    }
    else if (ioreq->reqType == EVM_READ_REQ)
    {
        size_t  numRead;

        FS_TRACE(
            NULL, FS_TRACE_IO,
         ("vol %llx evm_IOReq R: datap %p offset %lld num %lld mod1024 %lld",
            ioreq->id.whole_thing, ioreq->wDataHandle.dataPtr, offset,
            size, offset % 1024));
    
        /*
         * Sanity checks.
         */
        if (ioreq->id.details.u.id > FS_SIM_VOL_MAX)
        {
            fs_error(NULL, FS_SIM_ARGS_ERR, "volMax");
        }
    
        /*
         * Open the volume.
         */
        fd = volFd[ioreq->id.details.u.parts.clusterUnique];
        if (fd == 0)
        {
            // fd = open(volName[ioreq->id.details.u.id], O_RDONLY);
            fd = open64(volName[ioreq->id.details.u.id], 
                        openFlags | O_RDWR | O_CREAT,
                        S_IREAD | S_IWRITE);
            if (fd <= 0)
            {
                fs_error(NULL, FS_SIM_OPEN_ERR, "evm_IOReq %s",
						volName[ioreq->id.details.u.id]);
            }
            volFd[ioreq->id.details.u.parts.clusterUnique] = fd;
        }
    
        seekset = lseek64(fd, offset, SEEK_SET);
        if (seekset != offset)
        {
            fs_error(NULL, FS_SIM_LSEEK_ERR, "evm_IOReq");
        }
    
        datap = fs_ramAlloc(FS_RAM_READ_DATA, size);

        /*
         * Do the read.
         */
        numRead = read(fd, datap, size);
        if (numRead != size)
        {
            int numWritten;

            // File hasn't been written this far yet,
            // do it now to cover this error.
            numWritten = write( fd,
                                datap,
                                size - numRead);
        }

        if (ioreq->dataType == EVM_USER_DATA_REQ)
        {
            rc = mcpu_sim_evm_read_data(    datap, 
                                            size, 
                                            &(iostat.rDataHandle.codHandle));
            if (rc != NFX_OK)
            {
                FS_ASSERT(FALSE);
            }
            fs_ramDealloc(FS_RAM_READ_DATA, datap);
            datap = NULL;
        }
        else
        {
            iostat.rDataHandle.dataPtr = datap;
        }

        iostat.status = NFX_OK;
        iostat.devErrorCode = 0;
        iostat.clientPrivate = ioreq->clientPrivate;
    }

    if (bobDbgIoFailFinish)
    {
        bobDbgIoFailFinish -= 1;
        if (bobDbgIoFailFinish == 0)
        {
            printf("bobDbgIoFailFinish now ioreq %p\n", ioreq);
            iostat.status = NFX_ERR;
        }
    }

    if (bobDbgIoLate)
    {
        bobDbgIoLate -= 1;
        if (bobDbgIoLate == 0)
        {
            printf("bobDbgIoLate now\n");
            sleep(25);
        }
    }

    ioreq->callback(&iostat);

    return NFX_OK;

} /* fs_simEvmDoReq */


/**
 * Thread to process evm IO requests.
 */
void *
fs_simEvmThread(void *arg)
{
    evm_IOReq_t *ioreq;
    int32       rc;

    while(TRUE)
    {
        rc = pthread_mutex_lock(&volMutex);
        if (rc)
        {
            perror("mutex_lock");
            fs_error(NULL, FS_SIM_OPEN_ERR, "mutex_lock");
        }

		if (DLL_IS_EMPTY(&fs_simEvmHdr))
		{
        	// wait for new reqs
            fs_simEvmSignal = 0;
        	rc = pthread_cond_wait(&volCondVar, &volMutex);
            if (rc)
            {
                perror("cond_wait");
                fs_error(NULL, FS_SIM_OPEN_ERR, "vol cond wait");
            }
		}

		assert(!DLL_IS_EMPTY(&fs_simEvmHdr));

        DLL_REMOVE_HEAD(&fs_simEvmHdr,
                        ioreq,
                        evmEntry);

        rc = pthread_mutex_unlock(&volMutex);
        if (rc)
        {
            perror("mutex_unlock");
            fs_error(NULL, FS_SIM_OPEN_ERR, "mutex_unlock");
        }

        // Do it.
        fs_simEvmDoReq(ioreq);
    }
}


/*-----------------------------------------------------------------------------
 * evm_IOReq
 *
 * Description:  Simulated version of volume I/O request function
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
int
evm_IOReq(
    evm_IOReq_t     *ioreq,
    void            (*callback)(
                        evm_IOStat_t    *iostat))
{
    int32           rc;
    int32           err = NFX_OK;

    FS_ASSERT(ioreq->id.details.u.id > 0 &&
              ioreq->id.details.u.id <= FS_SIM_VOL_MAX);

    ioreq->callback = callback;

    if (bobDbgIoFailStart)
    {
        bobDbgIoFailStart -= 1;
        if (bobDbgIoFailStart == 0)
        {
            printf("bobDbgIoFailStart now\n");
            return NFX_ERR;
        }
    }

    rc = pthread_mutex_lock(&volMutex);
    if (rc)
    {
        perror("mutex_lock");
        fs_error(NULL, FS_SIM_OPEN_ERR, "evm_IOReq mutex lock");
    }

#if defined(FS_SIM_EVM_THREAD)

    DLL_ISOLATE_ELEM(ioreq, evmEntry);
	DLL_INSERT_TAIL(&fs_simEvmHdr, ioreq, evmEntry);

	if (fs_simEvmSignal == 0)
    {
        fs_simEvmSignal = 1;
        rc = pthread_cond_signal(&volCondVar);
        if (rc)
        {
            perror("pthread_cond_signal");
            fs_error(NULL, FS_SIM_OPEN_ERR, "evm_IOReq cond signal");
        }
    }
#else
    
    err = fs_simEvmDoReq(ioreq);

#endif
    rc = pthread_mutex_unlock(&volMutex);
    if (rc)
    {
        perror("mutex_unlock");
        fs_error(NULL, FS_SIM_OPEN_ERR, "evm_IOReq mutex unlock");
    }

    return err;
}


/*-----------------------------------------------------------------------------
 * fs_simReqThreadMain
 *
 * Description:  Filesystem requests that are processed as a thread
 *               start the thread here.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
void *
fs_simReqThreadMain(void *arg)
{
    esm_actionFunc_t ***evFunArr;
    esm_actionFunc_t **evFun;
    esm_actionFunc_t *fun;
    fs_SimThreadArg  *threadArg;
    int32            cpuId, threadNum;
    int              rc;

    threadArg = (fs_SimThreadArg *) arg;
    threadNum = threadArg->threadNum;
    rc = pthread_setspecific(fs_simKeyReqThreadNum, (void *) threadNum);
    if (rc != 0)
    {
        perror("pthread_setspecific");
        FS_ASSERT(FALSE);
    }

    //printf("fp %d thread %d created =========\n",
    //      threadArg->cpuId, threadNum);

    for (;;)
    {
        FS_ASSERT(fs_simReqThread[threadNum].state == FS_SIM_THREAD_BUSY);

        cpuId = threadArg->cpuId;
        rc = pthread_setspecific(fs_simKeyCpuId, (void *) cpuId);
        if (rc)
        {
            perror("pthread_setspecific");
            FS_ASSERT(FALSE);
        }
        memcpy(&fs_simReqThread[threadNum].threadArg, threadArg,
                sizeof(fs_SimThreadArg));

        evFunArr = threadArg->smInst->table->eventVectors;
        evFun = evFunArr[threadArg->event.event];
        fun = evFun[threadArg->smInst->currentState];
    
        threadArg->smInst->currentEvent = threadArg->event;
        threadArg->smInst->threadNum = threadNum;
    
        //printf("fp %d thread %d calling thread func\n", 
        //      cpuId, threadNum);
    
        if (!setjmp(fs_simReqThread[threadNum].jmpBuf))
        {
            fun->func(threadArg->handle,
                      threadArg->smInst->currentState,
                      threadArg->event,
                      threadArg->smInst->userInfo);
        }
        
        // If the function returns, we're done.

        // Thread returned, mark it for destruction.  It doesn't
        // hurt to do this even if its already been done by the client.
        esm_destroy(threadArg->handle);

        fs_ramDealloc(FS_RAM_SIM_MISC, threadArg);

        // Wake up the exec thread.  Careful here, cpuId may have
        // changed if the action function yielded anywhere along the
        // way.
        cpuId = fs_simReqThread[threadNum].threadArg.cpuId;

        //printf("fp %d thread %d func returned, signalling exec -----\n", 
        //      cpuId, threadNum);

        rc = pthread_mutex_lock(&(fs_simFpExecMutex));
        if (rc)
        {
            perror("pthread_mutex_lock");
            FS_ASSERT(FALSE);
        }

        fs_simNumReqThreadsWaiting += 1;
        fs_simReqThread[threadNum].state = FS_SIM_THREAD_AVAIL;

        rc = pthread_cond_signal(&(fs_simFpExec[cpuId].condVar));
        if (rc)
        {
            perror("pthread_cond_signal");
            FS_ASSERT(FALSE);
        }

        /* PAUSE */
        rc = pthread_cond_wait( &(fs_simReqThread[threadNum].condVar),
                                &(fs_simFpExecMutex));
        /* RESUME */
        if (rc)
        {
            perror("pthread_cond_wait fs_simReqThreadMain");
            FS_ASSERT(FALSE);
        }
    
        DLL_REMOVE_HEAD( &(fs_simThreadJobQueue.dllHdr),
                            threadArg, dllEntry);
        if (threadArg == NULL)
        {
            printf("Failed to get threadArg from job queue!!!\n");
            FS_ASSERT(FALSE);
        }
        rc = pthread_mutex_unlock(&(fs_simFpExecMutex));
        if (rc)
        {
            perror("2pthread_mutex_unlock");
            FS_ASSERT(FALSE);
        }

        //printf("fp %d thread %d running job from queue ++++\n",
        //      threadArg->cpuId, threadNum);
    }

    /* NOTREACHED */

    // gdb has a problem if we exit the thread!!!!
    //pthread_exit(NFX_OK);

} /* fs_simReqThreadMain */

/*-----------------------------------------------------------------------------
 * fs_simEsmRunEvent
 *
 * Description:  Simulated e-layer state machine engine.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
void
fs_simEsmRunEvent(
    int32                   cpuId,
    esm_instance_t          *smInst,
    esm_event_t             event)
{
    int32               newState;
    esm_actionFunc_t    ***evFunArr;
    esm_actionFunc_t    **evFun;
    esm_actionFunc_t    *fun;
    int                 i, rc;

    //printf("fp %d runEvent called\n", cpuId);

    if ((smInst->flags & ESM_CAN_YIELD) != 0)
    {
        rc = pthread_mutex_lock(&fs_simFpExecMutex);
        if (rc != 0)
        {
            perror("pthread_mutex_lock");
        }
        if ((smInst->flags & ESM_YIELD) != 0)
        {
            //printf("fp %d smHandle %d runEvent signalling yielder\n",
            //  cpuId, smInst->handle);

            // Signal the condition variable so the
            // thread will resume.
            smInst->flags &= ~ESM_YIELD;
            smInst->currentEvent = event;

            // Set the cpuId in the thread, since it can run on
            // either FP exec thread.
            fs_simReqThread[smInst->threadNum].threadArg.cpuId = cpuId;
    
            rc = pthread_cond_signal(&(fs_simReqThread[smInst->threadNum].condVar));
            if (rc)
            {
                perror("pthread_cond_signal");
            }
        }
        else if ((smInst->flags & ESM_RUNNING) == 0)
        {
            fs_SimThreadArg     *threadArg;
            pthread_attr_t      ptAttr;

            //printf("fp %d smHandle %d runEvent starting new\n",
            //  cpuId, smInst->handle);

            evFunArr = smInst->table->eventVectors;
            evFun = evFunArr[event.event];

            threadArg = fs_ramAlloc(FS_RAM_SIM_MISC, sizeof(fs_SimThreadArg));
            threadArg->cpuId = cpuId;
            threadArg->handle = smInst->handle;
            threadArg->func = evFun[smInst->currentState];
            threadArg->event = event;
            threadArg->smInst = smInst;
			DLL_ISOLATE_ELEM(threadArg, dllEntry);

            // First event for this yielding instance.  Create a
            // (or grab a waiting) thread and call its action function.

            smInst->flags |= ESM_RUNNING;

            if (fs_simNumReqThreadsWaiting > 0)
            {
                fs_simNumReqThreadsWaiting -= 1;

                // An idle thread is available, use it.
                
                DLL_INSERT_TAIL(&(fs_simThreadJobQueue.dllHdr),
                                    threadArg, dllEntry);

                for (i = 0; i < FS_SIM_EXEC_REQ_THREADS_MAX; i++)
                {
                    if (fs_simReqThread[i].state == FS_SIM_THREAD_AVAIL)
                    {
                        //printf("fp %d using avail thread %d\n",
                        //  cpuId, i);
                        threadArg->threadNum = i;
                        fs_simReqThread[i].state = FS_SIM_THREAD_BUSY;
                        rc = pthread_cond_signal(&fs_simReqThread[i].condVar);
                        if (rc)
                        {
                            perror("pthread_cond_signal");
                            FS_ASSERT(FALSE);
                        }
                        break;
                    }
                }
                FS_ASSERT(i < FS_SIM_EXEC_REQ_THREADS_MAX);
            }
            else
            {
                if (fs_simNumReqThreadsCreated >= FS_SIM_EXEC_REQ_THREADS_MAX)
                {
                    printf("max req threads (%d) exceeded\n",
                            fs_simNumReqThreadsCreated);
                    FS_ASSERT(FALSE);
                }

                //printf("fp %d no avail threads, creating new thread %d\n",
                //      cpuId, fs_simNumReqThreadsCreated);

                // No idle threads available, create one.

                rc = pthread_attr_init(&ptAttr);
                if (rc)
                {
                    perror("pthread_attr_init");
                    FS_ASSERT(FALSE);
                }
                rc = pthread_attr_setdetachstate(&ptAttr, PTHREAD_CREATE_DETACHED);
                if (rc)
                {
                    perror("pthread_attr_setdetachstate");
                    FS_ASSERT(FALSE);
                }
                threadArg->threadNum = fs_simNumReqThreadsCreated;

                rc = pthread_cond_init( &fs_simReqThread[threadArg->threadNum].
                                            condVar,
                                        NULL);
                if (rc)
                {
                    perror("pthread_cond_init");
                    FS_ASSERT(FALSE);
                }

                rc = pthread_create(&fs_simReqThread[threadArg->threadNum].thread,
                                    &ptAttr,
                                    &fs_simReqThreadMain,
                                    threadArg);
                if (rc)
                {
                    perror("pthread_create");
                    FS_ASSERT(FALSE);
                }

                fs_simNumReqThreadsCreated += 1;
            }
        }
        else
        {
            // This is not permitted.  Events delivered to yield'able
            // ESM types when they're not in esm_yield() are not handled.

            printf("ERR: fs_simEsmRunEvent: unexpected event for yielder\n");
            FS_ASSERT(FALSE);
            return;
        }

        /* PAUSE */
        //printf("fp %d runEvent pausing\n", cpuId);
        rc = pthread_cond_wait( &(fs_simFpExec[cpuId].condVar),
                                &(fs_simFpExecMutex));
        /* RESUME */
        if (rc)
        {
            perror("pthread_cond_wait fs_simEsmRunEvent");
        }
        rc = pthread_mutex_unlock(&fs_simFpExecMutex);
        if (rc)
        {
            perror("pthread_mutex_unlock");
        }
        //printf("fp %d runEvent resuming\n", cpuId);
        return;
    }

    /*
     * Call the action function appropriate for this event
     * based on the current state.
     */
    //printf("fp %d runEvent in exec thread\n", cpuId);
    evFunArr = smInst->table->eventVectors;
    evFun = evFunArr[event.event];
    fun = evFun[smInst->currentState];

    smInst->currentEvent = event;
    newState = fun->func(   smInst->handle,
                            fun->nextState,
                            event,
                            smInst->userInfo);
    //printf("fp %d event run in exec thread returned\n", cpuId);

    /*
     * Update currentState.
     */
    smInst->currentState = newState;

} /* fs_simEsmRunEvent */

/*-----------------------------------------------------------------------------
 * fs_simEsmGetInstFromHandle
 *
 * Description:  Find the ESM instance for the given handle.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
esm_instance_t *
fs_simEsmGetInstFromHandle(
    int32           handle)
{
    int32   index;

    index = FS_SIM_ESM_GET_HANDLE_INDEX(handle);

    if (fs_simEsmInstance[index] == NULL)
    {
        printf("got NULL instance for sm handle 0x%x\n", handle);
        return NULL;
    }

    if (fs_simEsmInstance[index]->handle == handle)
    {
        return fs_simEsmInstance[index];
    }
    else
    {
        return NULL;
    }

} /* fs_simEsmGetInstFromHandle */

/*-----------------------------------------------------------------------------
 * fs_simEsmAddToReadyQueue
 *
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
void
fs_simEsmAddToReadyQueue(
    esm_instance_t   *esm)
{
    if ((esm->flags & ESM_IS_BUSY) == 0 &&
        DLL_IS_ISOLATED(esm, readyQueueEntry))
    {
        DLL_INSERT_TAIL(&fs_simEsmReadyQueueHdr, esm, readyQueueEntry);
    }

} /* fs_simEsmAddToReadyQueue */

/*-----------------------------------------------------------------------------
 * fs_simEsmEnqueue
 *
 * Description:  Simulated e-layer state machine event enqueue.
 *
 *               Caller must hold fs_simEsmInstLock.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
int32
fs_simEsmEnqueue(
    esm_instance_t *smInst,
    esm_event_t    event)
{
    if (smInst->qcount >= ESM_Q_MAX)
    {
        printf("esm queue full!!!: qcount = %d, qload = %d, qunload = %d\n",
               smInst->qcount, smInst->qload, smInst->qunload);
        return NFX_ERR;
    }

    smInst->eventQueue[smInst->qload] = event;

    smInst->qcount += 1;

    if (smInst->qcount == 1)
    {
        fs_simEsmAddToReadyQueue(smInst);
    }

    if (++smInst->qload == ESM_Q_MAX)
    {
        smInst->qload = 0;
    }

    return NFX_OK;

} /* fs_simEsmEnqueue */

/*-----------------------------------------------------------------------------
 * esm_deliverEvent
 *
 * Description:  Simulated e-layer state machine event delivery.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
int32
esm_deliverEvent(
    esm_handle_t handle,
    esm_event_t  *event)
{
    esm_instance_t  *smInst;
    int32           rc = NFX_ERR;

    NFX_SPIN_LOCK(&fs_simEsmInstLock);
    smInst = fs_simEsmGetInstFromHandle(handle);
    if (!smInst)
    {
        printf("esm_deliverEvent: stale handle 0x%llx\n", (uint64)handle);
    }
    else
    {
        rc = fs_simEsmEnqueue(smInst, *event);
        if (rc != NFX_OK)
        {
            printf("esm_deliverEvent: enqueue failed (rc: %d)\n", rc);
        }
    }
    NFX_SPIN_UNLOCK(&fs_simEsmInstLock);

    return rc;

} /* esm_deliverEvent */

int32
esm_deliverHighPriorityEvent(
    esm_handle_t handle,
    esm_event_t  *event)
{
    return esm_deliverEvent(handle, event);
}

/*-----------------------------------------------------------------------------
 * esm_longJump
 *
 * Description:  Simulated e-layer state machine long jump.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
void
esm_longJump(
    esm_handle_t    handle)
{
    int32   threadNum;

    //printf("esm_longJump on handle 0x%x\n", handle);

    threadNum = (int32) pthread_getspecific(fs_simKeyReqThreadNum);

    longjmp(fs_simReqThread[threadNum].jmpBuf, 1);

} /* esm_longJump */


/*-----------------------------------------------------------------------------
 * fs_simEsmProcessQueue
 *
 * Description:  Process all queued esm events.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
int32
fs_simEsmProcessQueue(
    int32   cpuId)
{
    esm_instance_t       *esm;
    int32                q;
    int32                numEvents;

#define MAX_EVENTS_PER_POLL 32
    
    for (numEvents = 0; numEvents < MAX_EVENTS_PER_POLL; numEvents++)
    {
        //printf("fp %d top event %d\n", cpuId, numEvents);
        NFX_SPIN_LOCK(&fs_simEsmInstLock);
        if (DLL_IS_EMPTY(&fs_simEsmReadyQueueHdr))
        {
            //printf("fp %d esm q empty\n", cpuId);
            NFX_SPIN_UNLOCK(&fs_simEsmInstLock);
            break;
        }
        DLL_REMOVE_HEAD(&fs_simEsmReadyQueueHdr, esm, readyQueueEntry);
        FS_ASSERT((esm->flags & ESM_IS_BUSY) == 0);
        esm->flags |= ESM_IS_BUSY;
        
        if (esm->flags & ESM_DESTROY)
        {
            fs_simEsmInstance[FS_SIM_ESM_GET_HANDLE_INDEX(esm->handle)] = NULL;
            NFX_SPIN_UNLOCK(&fs_simEsmInstLock);

            if (esm->qcount > 0)
            {
                int i;
                // TODO - properly drain the queue by calling
                // the trash action for the esm.
                q = esm->qunload;
                for (i = 0; i < esm->qcount; i++) {
                    esm_event_t *event = &esm->eventQueue[(q + i) % ESM_Q_MAX];
                    printf("esm %llx: pending event %d for destroyed sm:"
                           " %d (%p, %p)\n",
                           (uint64)esm->handle, i, event->event, event->p1.u.p,
                           event->p2.u.p);
                }
            }

            fs_ramDealloc(FS_RAM_SIM_MISC, esm);
        }
        else
        {
            FS_ASSERT(esm->qcount > 0);
            esm->qcount -= 1;
            NFX_SPIN_UNLOCK(&fs_simEsmInstLock);
    
            q = esm->qunload++;
            if (esm->qunload == ESM_Q_MAX)
            {
                esm->qunload = 0;
            }
            //printf("fp %d calling runevent inst %d event %d\n",
            //  cpuId, esm->handle, esm->eventQueue[q].event);
            fs_simEsmRunEvent(  cpuId,
                                esm,
                                esm->eventQueue[q]);

            NFX_SPIN_LOCK(&fs_simEsmInstLock);
            // Put it back on the ready queue if there are still 
            // events to be processed, or if a destroy was indicated.
            if (esm->qcount > 0 ||
                (esm->flags & ESM_DESTROY))
            {
                DLL_INSERT_TAIL(&fs_simEsmReadyQueueHdr, esm, readyQueueEntry);
            }
            esm->flags &= ~ESM_IS_BUSY;
            NFX_SPIN_UNLOCK(&fs_simEsmInstLock);
        }
    }

    return numEvents;

} /* fs_simEsmProcessQueue */

/*-----------------------------------------------------------------------------
 * esm_destroy
 *
 * Description:  Destroy an instance of an ESM.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
int32
esm_destroy(
    esm_handle_t        handle)
{
    esm_instance_t   *esm;
    int32               index;
    int32               rc;

    NFX_SPIN_LOCK(&fs_simEsmInstLock);
    index = FS_SIM_ESM_GET_HANDLE_INDEX(handle);
    esm = fs_simEsmInstance[index];

    //printf("destroying esm at %p, index %d\n", esm, index);

    if (esm == NULL)
    {
        printf("destroy: esm == NULL\n");
        rc = NFX_ERR;
    }
    else
    {
        if (esm->handle == handle)
        {
            if (esm->qcount)
            {
                printf("destroy handle 0x%llx with %d in queue\n",
                    (uint64)esm->handle, esm->qcount);

                btrace_print();
            }

            esm->flags |= ESM_DESTROY;
            fs_simEsmAddToReadyQueue(esm);
            rc = NFX_OK;
        }
        else
        {
            printf(
            "destroy: wrong handle %llx fs_simEsmInstance->handle == %llx\n",
                (uint64)handle, (uint64)esm->handle);
            rc = NFX_ERR;
        }
    }
    NFX_SPIN_UNLOCK(&fs_simEsmInstLock);

    return rc;

} /* esm_destroy */


/*-----------------------------------------------------------------
 * Name:    esm_getUserInfo
 *
 * Description: Return the the user info for the specified state machine.
 *              Returns NULL if the state machine handle is invalid.
 * 
 * Created by:  Paul Haahr
 * 
 * Date Created:    9/13/01
 *
 *-----------------------------------------------------------------
 */
void *
esm_getUserInfo(
    esm_handle_t   handle)
{
    esm_instance_t *esm;
    void *result;
    
    NFX_SPIN_LOCK(&fs_simEsmInstLock);
    esm = fs_simEsmInstance[FS_SIM_ESM_GET_HANDLE_INDEX(handle)];
    if (esm == NULL || esm->handle != handle)
        result = NULL;
    else
        result = esm->userInfo;
    NFX_SPIN_UNLOCK(&fs_simEsmInstLock);
    
    return result;
}


/*-----------------------------------------------------------------------------
 * esm_instantiate
 *
 * Description:  Instantiate an ESM.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
int32
esm_instantiate(
    esm_table_t             *table,
    esm_actionFunc_t        *createAction,
    esm_handle_t            *handle,
    void                    *userInfo)
{
    int                     i, n;
    esm_instance_t          *esm;
    int32                   nextState;

    if (handle == NULL)
    {
        printf("esm_instantiate:  NULL handle ptr\n");
        return NFX_ERR;
    }

    esm = fs_ramAlloc(FS_RAM_SIM_MISC, sizeof(esm_instance_t));
    bzero(esm, sizeof(esm_instance_t));    
    esm->table = table;
    esm->userInfo = userInfo;
    if (table->canYield)
    {
        esm->flags = ESM_CAN_YIELD;
    }
    esm->flags |= ESM_IS_BUSY;

    /*
     * Find the next available slot, add the entry, return handle.
     */
    NFX_SPIN_LOCK(&fs_simEsmInstLock);
    for (   i = fs_simEsmNextEntry, n = 0;
            n < FS_SIM_ESM_MAX;
            n++)
    {
        if (fs_simEsmInstance[i] == NULL)
        {
            esm_event_t event;

            /*
             * Found available slot.
             */
            fs_simEsmInstance[i] = esm;

            FS_SIM_ESM_CREATE_HANDLE(esm->handle, i);
            
            fs_simEsmNextEntry = i + 1;
            if (fs_simEsmNextEntry >= FS_SIM_ESM_MAX)
            {
                fs_simEsmNextEntry = 0;
            }
            NFX_SPIN_UNLOCK(&fs_simEsmInstLock);
            
            /*
             * Call the create action.
             */
            event.event = 0;
            nextState = createAction->func( esm->handle,
                                            createAction->nextState,
                                            event,
                                            userInfo);
            esm->currentState = nextState;
            break;
        }

        i += 1;
        if (i >= FS_SIM_ESM_MAX)
        {
            i = 0;
        }
    }

    if (n < FS_SIM_ESM_MAX)
    {
        //printf("esm create handle %d table %p\n", esm->handle, table);

        *handle = esm->handle;
        NFX_SPIN_LOCK(&fs_simEsmInstLock);
        esm->flags &= ~ESM_IS_BUSY;
        NFX_SPIN_UNLOCK(&fs_simEsmInstLock);
        return NFX_OK;
    }
    else
    {
        FS_ASSERT(FALSE);
        return NFX_ERR;
    }

} /* esm_instantiate */

/*-----------------------------------------------------------------------------
 * esm_yield
 *
 * Description:  State machine instance is yielding back the CPU.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
void
esm_yield(
	esm_event_t	*eventp,
	int32		waitmSecs)
{
    esm_instance_t   *smInst;
    esm_handle_t     myHandle;
    int32            cpuId, rc, threadNum;

    struct timeval  now;
    struct timespec timeout;

    threadNum = (int32) pthread_getspecific(fs_simKeyReqThreadNum);
    cpuId = fs_simReqThread[threadNum].threadArg.cpuId;

    myHandle = fs_simReqThread[threadNum].threadArg.handle;
    smInst = fs_simEsmGetInstFromHandle(myHandle);

    FS_ASSERT(smInst->threadNum == threadNum);

    if ((smInst->flags & ESM_CAN_YIELD) == 0)
    {
        printf("ERR: esm_yield() called from non-yieldable ESM %s\n",
                smInst->name);
        printf("   currState: %d  currEvent: %d\n",
                smInst->currentState,
                smInst->currentEvent.event);
        FS_ASSERT(FALSE);
        return;
    }

    smInst->flags |= ESM_YIELD;

    //printf("fp %d smHandle %d esm_yield signalling exec\n",
    //  cpuId, myHandle);

    // wake up the exec thread
    rc = pthread_mutex_lock(&fs_simFpExecMutex);
    if (rc != 0)
    {
        perror("pthread_mutex_lock");
        return;
    }
    rc = pthread_cond_signal(&fs_simFpExec[cpuId].condVar);
    if (rc != 0)
    {
        perror("pthread_cond_signal");
        return;
    }

    /* PAUSE */

    //printf("esm_yield: pausing\n");

	if (waitmSecs != 0)
	{
    	gettimeofday(&now, NULL);
    	timeout.tv_sec  = now.tv_sec + (waitmSecs / 1000);
    	timeout.tv_nsec = (now.tv_usec + ((waitmSecs % 1000) * 1000)) * 1000;

        if (timeout.tv_nsec > (1000 * 1000 * 1000))
        {
            timeout.tv_sec += 1;
            timeout.tv_nsec -= (1000 * 1000 * 1000);
        }

    	rc = pthread_cond_timedwait(&fs_simReqThread[threadNum].condVar,
                                	&fs_simFpExecMutex,
                                	&timeout);
    	if (rc == ETIMEDOUT)
        {
			fs_Context *context = esm_getUserInfo(myHandle);
			char       lockDesc[FS_LOCK_DESCRIPTION];
            esm_event_t timeoutEvent;

/*
  #ifdef FS_DEBUG
  fprintf(stderr, "fp%d smHandle %llx esm_yield: waited %d seconds\n",
  cpuId, (uint64)myHandle, waitSecs);
  btrace_print();
  fprintf(stderr, "Locks:\n");
  fs_lockPrintOwned(context);
  if (context->lockWait.lock != NULL) {
  fprintf(stderr, "Waiting for: %s\n", 
  fs_lockDescribe(context, context->lockWait.lock,
  lockDesc));
  }
  fprintf(stderr, "\n");
  #endif
*/

            timeoutEvent.event = THREAD_EVENT_TIMEOUT;
            rc = esm_deliverEvent(myHandle, &timeoutEvent);
            if (rc != NFX_OK)
            {
                fs_error(NULL, FS_INTERNAL_ERR,
                         "yield timeout event delivery");
            }
            rc = pthread_cond_wait(&fs_simReqThread[threadNum].condVar,
                                   &fs_simFpExecMutex);
    	}
	}
	else
	{
       	rc = pthread_cond_wait(&fs_simReqThread[threadNum].condVar,
                           	   &fs_simFpExecMutex);
	}

    if (rc != 0)
    {
        perror("pthread_cond_wait esm_yield");
        return;
    }

    /* RESUME */
    rc = pthread_mutex_unlock(&fs_simFpExecMutex);
    if (rc != 0)
    {
        perror("pthread_mutex_unlock");
        return;
    }

    // Careful, we might have migrated to the other FP.
    cpuId = fs_simReqThread[threadNum].threadArg.cpuId;

    rc = pthread_setspecific(fs_simKeyCpuId, (void *) cpuId);
    if (rc != 0)
    {
        perror("pthread_setspecific");
        return;
    }
    //printf("fp %d smHandle %d esm_yield resumed\n",
    //  cpuId, myHandle);

	if (eventp != NULL)
	{
    	*eventp = smInst->currentEvent;
	}

} /* esm_yield */

/*-----------------------------------------------------------------------------
 * Thread library stuff (simulated sm-thread)
 */

/**
 * Create a thread and run the specified function
 * in it with the specified argument.  Returns a
 * non-NULL esm_handle_t for the thread on success,
 * NULL otherwise.  A thread remains in existence
 * until it returns or thread_exit() is used.
 */ 
esm_handle_t
thread_create(
    // Opaque userInfo to deliver to the thread. 
    void            *userInfo)
{   
    esm_handle_t    handle;
    int32           rc;

    rc = simthread_sm_create(&handle, userInfo);
    if (rc != NFX_OK)
    {
        return 0;
    }
 
    return handle;

} /* thread_create */

/**
 * Start a thread with the function to run in the thread.
 * To call once a thread is created.
 */ 
int32
thread_start(
    // The handle to the thread
    esm_handle_t * handle,
    // The function to run in the thread.
    void    (*func)(esm_handle_t    handle,     // The thread's esm_handle_t.
                    void            *userInfo)) // Opaque userInfo.
{   
    esm_event_t     event;
    int32           rc;

    event.event = 0;
    event.p1.u.p = func;
    rc = esm_deliverEvent(*handle, &event); 
    if (rc != NFX_OK)
    {
        esm_destroy(*handle);
    }

    return rc;

} /* thread_start */

/**
 * Yield the cpu back to the executive.
 */ 
void
thread_yield(
    esm_event_t         *eventp,        // Return my waking event info here,
                                        // NULL if I don't care.
    int32               waitmSecs)      // Deliver TIMEOUT event if I sleep
                                        // longer than this, 0 to let me sleep
                                        // forever.
{
    esm_yield(eventp, waitmSecs);

} /* thread_yield */
 
/**
 * Sleep the thread for a while, waking it with
 * a WAKEUP event (unless another event is delivered
 * first) when the specified period has elapsed.
 * Returns NFX_ERR if unable to schedule a timer,
 * NFX_OK otherwise.
 */
int32
thread_sleep(
    esm_event_t         *eventp,        // Return my waking event info here,
                                        // NULL if I don't care. 
    int32               waitMsecs)      // Wake me up after this many
                                        // milliseconds (10 msec granularity).
{   
	sleep(waitMsecs / 1000);

	return NFX_OK;
}

/**
 * Enqueue an event to a thread.  A thread can have up to
 * THREAD_EVENT_QUEUE_MAX threads awaiting delivery at any
 * given time.  (While multiple enqueued events are supported,
 * generally, a thread will be sleeping waiting for a specific
 * event, thus only a single event would normally be enqueued
 * to a thread at any given time).  Returns NFX_ERR if unable
 * to enqueue the event, otherwise NFX_OK.
 *
 * The value of event.event is user-defined, however, event
 * numbers greater than or equal to THREAD_EVENT_USER_MAX are
 * reserved.
 */
int32
thread_event(
    esm_handle_t    handle,
    esm_event_t     *eventp)
{
    if (eventp->event >= THREAD_EVENT_USER_MAX)
    {
        return NFX_ERR;
    }

    return esm_deliverEvent(handle, eventp);

} /* thread_event */

int32
thread_highPriorityEvent(
    esm_handle_t    handle,
    esm_event_t     *eventp)
{
    return thread_event(handle, eventp);
}


/**
 * Perform a long jump of the thread back
 * to the executive.  Use this anywhere in a
 * thread to return to the executive and destroy
 * the thread.  This function does not return.
 */
void
thread_exit()
{
    int32   threadNum;

    threadNum = (int32) pthread_getspecific(fs_simKeyReqThreadNum);

    esm_longJump(fs_simReqThread[threadNum].threadArg.handle);

} /* thread_exit */


/**
 * Create function for thread state machine instantiation.
 * We don't use it, but the state machine API needs it.
 */
int32
simthread_create(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *smCookie)
{
    return nextState;

} /* simthread_create */


/**
 * Start action function for thread state machine.  This is
 * where threads first come to life outside the executive.
 */
int32
simthread_start(
    esm_handle_t        handle,
    int32               nextState, 
    esm_event_t         event,
    void                *userInfo)
{
    void    (*func)() = event.p1.u.p;
 
    // Call the user's thread function.
    // If it returns, the thread dies.
    (func)(handle, userInfo);
    
    return nextState;
    
} /* simthread_start */

    
/**
 * This should never happen.
 */
int32
simthread_bad_state(
    esm_handle_t        handle,
    int32               nextState,
    esm_event_t         event,
    void                *userInfo)
{
 
	FS_ASSERT(FALSE);

    return nextState;

} /* simthread_bad_state */


/*-----------------------------------------------------------------------------
 * fs_simFpProcessRxMsgs
 *
 * Description:  Process pending messages for the FP specified.  
 *               Must be called with fs_simFpExecMutex locked.
 *
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
int32
fs_simFpProcessRxMsgs(
    int32   cpuId)
{
    fs_SimMsgQueueHead      *qHead;
    fs_SimMsgQueueEntry     *qEntry;
    int32                   numMsgs;
    int32                   rc;

    numMsgs = 0;

    qHead = &(fs_simFpRxQueue[cpuId]);

    for (;;)
    {
        if (DLL_IS_EMPTY(&(qHead->dllHdr)))
        {
            break;
        }
        DLL_REMOVE_HEAD(&(qHead->dllHdr), qEntry, dllEntry);
        rc = pthread_mutex_unlock(&fs_simFpExecMutex);
        if (rc)
        {
            perror("6pthread_mutex_unlock");
            break;
        }

        //printf("Received req %lld\n",
        //  ((fs_MsgReqHdr *)qEntry->msg)->clientPrivate);

        fs_receiveMsg(qEntry->msg, 0, 0, 0);
        fs_ramDealloc(FS_RAM_SIM_MSG_QENTRY, qEntry);
        numMsgs += 1;

        rc = pthread_mutex_lock(&fs_simFpExecMutex);
        if (rc)
        {
            perror("pthread_mutex_lock");
            FS_ASSERT(FALSE);
        }
    }

    return numMsgs;

} /* fs_simFpProcessRxMsgs */

/*-----------------------------------------------------------------------------
 * fs_simFpExecThreadMain
 *
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
void *
fs_simFpExecThreadMain(void *arg)
{
    struct timeval  now;
    struct timespec timeout;
    int32           numEventsRun, numMsgsRcvd, cpuId;
    int32           rc;

    cpuId = (int32) arg;
    rc = pthread_setspecific(fs_simKeyCpuId, (void *) cpuId);
    if (rc)
    {
        perror("pthread_setspecific");
        return NULL;
    }

    //printf("fp exec thread %ld cpuId %d\n", pthread_self(), cpuId);

    rc = pthread_mutex_lock(&fs_simFpExecMutex);
    if (rc)
    {
        perror("pthread_mutex_lock");
        FS_ASSERT(FALSE);
    }

    for (;;)
    {
checkEvents:
        //printf("fp %d checking event q\n", cpuId);
        rc = pthread_mutex_unlock(&fs_simFpExecMutex);
        if (rc)
        {
            perror("pthread_mutex_lock");
            FS_ASSERT(FALSE);
        }
        //printf("fp %d calling EsmprocessQueue\n", cpuId);
        numEventsRun = fs_simEsmProcessQueue(cpuId);

        //printf("fp %d checking message q\n", cpuId);
        rc = pthread_mutex_lock(&fs_simFpExecMutex);
        if (rc)
        {
            perror("pthread_mutex_lock");
            FS_ASSERT(FALSE);
        }
        numMsgsRcvd = fs_simFpProcessRxMsgs(cpuId);

        if (numMsgsRcvd)
        {
            goto checkEvents;
        }

        if (!numEventsRun && !numMsgsRcvd)
        {
            /* PAUSE */
            //printf("fp %d pausing\n", cpuId);

            // We need to do this with a timeout
            // since we have to simulate the polling on the
            // embedded platform.
            gettimeofday(&now, NULL);
            timeout.tv_sec = now.tv_sec;
            timeout.tv_nsec = (now.tv_usec + 10) * 1000;

            rc = pthread_cond_timedwait(    &(fs_simFpRxQueue[cpuId].condVar),
                                            &fs_simFpExecMutex,
                                            &timeout);
            if (rc && rc != ETIMEDOUT)
            {
                perror("pthread_cond_wait fs_simFpExecThreadMain");
                FS_ASSERT(FALSE);
            }

            /* RESUME */
            //printf("fp %d resuming\n", cpuId);
        }
    }

} /* fs_simFpExecThreadMain */

/*-----------------------------------------------------------------------------
 * fs_simMountDispatch1
 *
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
void
fs_simMountDispatch1(
    struct svc_req  *rqstp,
    SVCXPRT         *transp)
{
    union
    {
        dirpath mountproc_mnt_1_arg;
        dirpath mountproc_umnt_1_arg;
    } argument;
    char                    *result;
    bool_t                  (*xdr_argument)();
    bool_t                  (*xdr_result)();
    char                    *(*local)();

    //printf("got mount req %ld \n", rqstp->rq_proc);
    //printf("mountDispatch1 called in thread %ld\n", pthread_self());

    switch (rqstp->rq_proc)
    {
    case MOUNTPROC_NULL:
        FS_ASSERT (rqstp->rq_vers == MOUNTVERS ||
                   rqstp->rq_vers == MOUNTVERS_V3);
        xdr_argument = xdr_void;
        xdr_result = xdr_void;
        local = (char *(*)()) mountproc_null_1;
        break;

    case MOUNTPROC_MNT:
        xdr_argument = xdr_dirpath;
        if (rqstp->rq_vers == MOUNTVERS_V3)
        {
            xdr_result = xdr_mntres3;
            local = (char *(*)()) mountproc_mnt_3;
        }
        else
        if (rqstp->rq_vers == MOUNTVERS)
        {
            xdr_result = xdr_fhstatus;
            local = (char *(*)()) mountproc_mnt_1;
        }
        else
            FS_ASSERT (FALSE);
        break;

    case MOUNTPROC_DUMP:
        FS_ASSERT (rqstp->rq_vers == MOUNTVERS ||
                   rqstp->rq_vers == MOUNTVERS_V3);
        xdr_argument = xdr_void;
        xdr_result = xdr_mountlist;
        local = (char *(*)()) mountproc_dump_1;
        break;

    case MOUNTPROC_UMNT:
        FS_ASSERT (rqstp->rq_vers == MOUNTVERS ||
                   rqstp->rq_vers == MOUNTVERS_V3);
        xdr_argument = xdr_dirpath;
        xdr_result = xdr_void;
        local = (char *(*)()) mountproc_umnt_1;
        break;

    case MOUNTPROC_UMNTALL:
        FS_ASSERT (rqstp->rq_vers == MOUNTVERS ||
                   rqstp->rq_vers == MOUNTVERS_V3);
        xdr_argument = xdr_void;
        xdr_result = xdr_void;
        local = (char *(*)()) mountproc_umntall_1;
        break;

    case MOUNTPROC_EXPORT:
        FS_ASSERT (rqstp->rq_vers == MOUNTVERS ||
                   rqstp->rq_vers == MOUNTVERS_V3);
        xdr_argument = xdr_void;
        xdr_result = xdr_exports;
        local = (char *(*)()) mountproc_export_1;
        break;

    case MOUNTPROC_EXPORTALL:
        FS_ASSERT (rqstp->rq_vers == MOUNTVERS ||
                   rqstp->rq_vers == MOUNTVERS_V3);
        xdr_argument = xdr_void;
        xdr_result = xdr_exports;
        local = (char *(*)()) mountproc_exportall_1;
        break;

    default:
        FS_ASSERT (rqstp->rq_vers == MOUNTVERS ||
                   rqstp->rq_vers == MOUNTVERS_V3);
        svcerr_noproc (transp);
        return;
    }
    memset ((char *)&argument, 0, sizeof (argument));
    if (!svc_getargs(   transp,
                        (xdrproc_t) xdr_argument,
                        (caddr_t) &argument))
    {
        svcerr_decode (transp);
        return;
    }
    result = (*local)(&argument, rqstp);
    if (result == NULL)
    {
        svcerr_systemerr(transp);
    }
    else if (!svc_sendreply(    transp,
                                (xdrproc_t) xdr_result,
                                (caddr_t) result))
    {
        svcerr_systemerr(transp);
    }

    if (!svc_freeargs(  transp,
                        (xdrproc_t) xdr_argument,
                        (caddr_t) &argument))
    {
        fprintf (stderr, "%s", "unable to free arguments");
        FS_ASSERT(FALSE);
    }

} /* fs_simMountDispatch1 */


/*-----------------------------------------------------------------------------
 * fs_simNfsServerMuxThread
 *
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 * 
 * In order to get some concurrency we basically multiplex a
 * single externally visible port (2049) onto multiple
 * internal ports like this:
 *
 *   +--------+          +---------------------------------+
 *   | client |<-------->| port 2049<-->+----------------+ |
 *   +--------+          |              | srvr mux       | |
 *                       |              +----------------+ |
 *                       |               |  |  |  | ...    |
 *                       |              +----------------+ |
 *                       |              |3  3  3  3      | |
 *                       |              |0  0  0  0      | |
 *                       |              |0  0  0  0 ...  | |
 *                       |              |0  1  2  3      | |
 *                       |              |                | |
 *                       |              | fs threads     | |
 *                       |              +----------------+ |
 *                       +---------------------------------+
 */

#define FS_SIM_NFS_MUX_PORT_START       3000 

void *
fs_simNfsServerMuxThread(void *threadArg)
{
    struct hostent      *host;
    struct sockaddr_in  sin, fromAddr;
    struct sockaddr_in  intAddrs[FS_SIM_NFS_THREAD_MAX];
    struct msghdr       msghdr;
    struct iovec        iovecs[2];
	socklen_t  			fromLen;
	int32				intSocks[FS_SIM_NFS_THREAD_MAX];
    int32               i, rc, msgLen;
    uint32              nextChild, usePort;
	char				msgPtr[FS_SIM_MSG_BUF_LEN];
//    char                hostName[64];

    //printf("nfs srvr mux pid %d\n", getpid());

    // Create the external socket.
    fs_simNfsSrvrSocket = socket(AF_INET, SOCK_DGRAM, 0);
    if (fs_simNfsSrvrSocket < 0)
    {
        perror("ext socket");
       	FS_ASSERT(FALSE);
    }

    // Bind the external socket.
    bzero((char *) &sin, sizeof(struct sockaddr_in));
    sin.sin_family=AF_INET;
    sin.sin_port = htons(FS_SIM_NFS_TO_CLIENT_PORT);

//     rc = gethostname(hostName, sizeof(hostName));
//     if (rc < 0)
//     {
//         perror("gethostname");
//        	FS_ASSERT(FALSE);
//     }
//     host = gethostbyname(hostName);
//     if (!host) {
//         perror ("gethostbyname");
//        	FS_ASSERT(FALSE);
//     }
// 
//     //printf("server mux binding to %s (0x%x)\n", hostName,
//     // *((uint32 *)host->h_addr));
// 
//     memcpy (&sin.sin_addr, host->h_addr, host->h_length);

    if (bind(fs_simNfsSrvrSocket,(struct sockaddr *)&sin,sizeof(sin)) != 0) {
        perror("bind ext nfs socket");
       	FS_ASSERT(FALSE);
    }

    if (fs_simPortmapNfs)
    {
        // register for version 3
        {
            // Unregister current program.
            pmap_unset(NFS_PROGRAM, NFS_VERSION_V3);

            // Register with the portmapper service.
            if (!pmap_set(NFS_PROGRAM,
                          NFS_VERSION_V3,
                          IPPROTO_UDP,
                          FS_SIM_NFS_TO_CLIENT_PORT))
            {
                perror("nfs pmap_set");
                FS_ASSERT(FALSE);
            }
        }
        // register for version 2
        {
            // Unregister current program.
            pmap_unset(NFS_PROGRAM, NFS_VERSION);

            // Register with the portmapper service.
            if (!pmap_set(NFS_PROGRAM,
                          NFS_VERSION,
                          IPPROTO_UDP,
                          FS_SIM_NFS_TO_CLIENT_PORT))
            {
                perror("nfs pmap_set");
                FS_ASSERT(FALSE);
            }
        }
    }

	// Create sockets and dest addrs for our internal
	// destination ports.
	for (i = 0; i < FS_SIM_NFS_THREAD_MAX; i++)
	{
        intSocks[i] = socket(AF_INET, SOCK_DGRAM, 0);
        if (intSocks[i]< 0)
        {
            perror("int socket");
       	    FS_ASSERT(FALSE);
        }

        bzero((char *) &intAddrs[i], sizeof(struct sockaddr_in));
        intAddrs[i].sin_family=AF_INET;
        intAddrs[i].sin_port = htons(FS_SIM_NFS_MUX_PORT_START + i);
        host = gethostbyname("localhost");
        if (!host) {
            perror ("gethostbyname");
       	    FS_ASSERT(FALSE);
        }
        memcpy (&intAddrs[i].sin_addr, host->h_addr, host->h_length);
	}

    while (TRUE)
    {
        // rx on the ext socket, sending a two part message to an internal
        // destination for each one received:  the first part describes
        // the client address, the second part is the client's rpc request.
        // This way, the internal nfs server instance doesn't need us
        // to stay around to send the reply - it will do it directly.

        //printf("srvrMux listening...\n");

		fromLen = sizeof(fromAddr);
        
    	msgLen = recvfrom(fs_simNfsSrvrSocket,
                  	      msgPtr,
                      	  FS_SIM_MSG_BUF_LEN,
						  0,
                          (struct sockaddr *) &fromAddr,
						  &fromLen);
        //printf("srvrMux got %d byte packet\n", msgLen);
    	if (msgLen <= 0)
    	{
        	perror("nfs srvr mux recv");
        	FS_ASSERT(FALSE);
    	}

        // Send it to one of our nfs listener processes.
		usePort = (++nextChild % FS_SIM_NFS_THREAD_MAX);

        msghdr.msg_name = &intAddrs[usePort];
        msghdr.msg_namelen = sizeof(struct sockaddr_in);
        msghdr.msg_iov = iovecs;
        msghdr.msg_iovlen = 2;
        msghdr.msg_control = NULL;
        msghdr.msg_controllen = 0;
        msghdr.msg_flags = 0;

        iovecs[0].iov_base = &fromAddr;
        iovecs[0].iov_len = fromLen;
        iovecs[1].iov_base = msgPtr;
        iovecs[1].iov_len = msgLen;

        //printf("sending on usePort %d\n", usePort);

        rc = sendmsg(intSocks[usePort], &msghdr, 0);
        if (rc != fromLen + msgLen) {
            printf("numSent %d\n", rc);
       		perror("nfs srvr mux internal sendto 2");
       		FS_ASSERT(FALSE);
   		}
    }

    FS_ASSERT(FALSE);

} /* fs_simNfsServerMuxThread */


/**
 * Thread to wait for fs responses to fssf requests.
 */

void *
fs_simFssfThread(void *threadArg)
{
    fs_MsgRespHdr           *respHdr;
    fs_SimMsgQueueEntry     *qEntry;
    int32                   rc;

    while (TRUE)
    {
        rc = pthread_mutex_lock(&fs_simFssfQueue.mutex);
        if (rc)
        {
            perror("pthread_mutex_lock fs_simFssfThread");
        }

        // Now wait for the next FP response.

        /* PAUSE */
        rc = pthread_cond_wait(&fs_simFssfQueue.condVar,
                               &fs_simFssfQueue.mutex);
        /* RESUME */
        if (rc)
        {
            perror("pthread_cond_wait fs_simFssfThread");
        }

        //printf("fs_simFssfThread checking for msg\n");
        DLL_REMOVE_HEAD(&fs_simFssfQueue.dllHdr,
                        qEntry, dllEntry);
        if (qEntry == NULL)
        {
            printf("fs_simFssfThread: failed to find response msg\n");
            FS_ASSERT(FALSE);
        }

        //printf("fs_simFssfThread got a msg\n");

        rc = pthread_mutex_unlock(&fs_simFssfQueue.mutex);
        if (rc)
        {
            perror("pthread_mutex_unlock in fs_simFssfThread");
        }   

        respHdr = (fs_MsgRespHdr *) qEntry->msg;
        fs_ramDealloc(FS_RAM_SIM_MSG_QENTRY, qEntry);

        // Call fssf's receive function with the fs
        // response message.

        fssf_processMsg((char *) respHdr, respHdr->src);
    }
}


/*-----------------------------------------------------------------------------
 * fs_simGetUnixCred
 *
 * Unpack the opaque credential into a struct authunix_parm.
 */
int32
fs_simGetUnixCred(
    struct authunix_parms   *aup,
    struct opaque_auth      *oa)
{
    XDR     xdrs;
    int32   *buf, rc = NFX_OK;
    int     str_len, auth_len, gid_len, i;

    FS_ASSERT(oa->oa_flavor == AUTH_UNIX);

    if (!oa->oa_base || !oa->oa_length)
        return NFX_ERR;

    auth_len = oa->oa_length;
    xdrmem_create(&xdrs, oa->oa_base, auth_len, XDR_DECODE);
    buf = XDR_INLINE(&xdrs, auth_len);
    if (buf != NULL)
    {
        aup->aup_time = IXDR_GET_LONG(buf);
        str_len = IXDR_GET_U_LONG(buf);
        if (str_len > MAX_MACHINE_NAME) {
            rc = NFX_ERR;
            goto done;
        }
        memcpy(aup->aup_machname, buf, str_len);
        aup->aup_machname[str_len] = 0;
        str_len = RNDUP(str_len);
        buf += str_len / sizeof (int32);
        aup->aup_uid = IXDR_GET_LONG(buf);
        aup->aup_gid = IXDR_GET_LONG(buf);
        gid_len = IXDR_GET_U_LONG(buf);
        if (gid_len > NGRPS) {
            rc = NFX_ERR;
            goto done;
        }
        aup->aup_len = gid_len;
        for (i = 0; i < gid_len; i++) {
            aup->aup_gids[i] = IXDR_GET_LONG(buf);
        }
        /*
         * five is the smallest unix credentials structure -
         * timestamp, hostname len (0), uid, gid, and gids len (0).
         */
        if ((5 + gid_len) * BYTES_PER_XDR_UNIT + str_len > auth_len) {
            printf("bad auth_len gid %d str %d auth %d\n", gid_len, str_len, auth_len);
            rc = NFX_ERR;
            goto done;
        }
    } else if (!xdr_authunix_parms(&xdrs, aup)) {
        rc = NFX_ERR;
    }

done:

    xdrs.x_op = XDR_FREE;
    xdr_destroy(&xdrs);
    return rc;

} /* fs_simGetUnixCred */

/*-----------------------------------------------------------------------------
 * fs_simProcessBadProgRpc *
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
void
fs_simProcessBadProgRpc(
    XDR                 *xdr, 
    struct rpc_msg      *rpccall, 
    struct sockaddr_in  *fromAddr,
    int32               threadNum)
{
    struct rpc_msg          rpcreply;
    char                    xdrRespBuf[1000];
    int32                   rc;

    //printf("fs_simProcessBadProgRpc: entered for prog %d",
    //      rpccall->rm_call.cb_prog);

    FS_TRACE(NULL, FS_TRACE_REQ, ("rcvd bad rpc program number %ld", 
                                  rpccall->rm_call.cb_prog));

    // Cleanup the XDR from the call, ready it for encoding the
    // reply.
    memset(xdr, 0 , sizeof *xdr);
    memset(&rpcreply, 0, sizeof rpcreply);

    // Put together the RPC reply header.
    rpcreply.rm_xid = rpccall->rm_xid;
    rpcreply.rm_direction = REPLY;
    rpcreply.rm_reply.rp_stat = MSG_ACCEPTED;
    rpcreply.rm_reply.rp_acpt.ar_stat = PROG_UNAVAIL;
    rpcreply.acpted_rply.ar_results.proc = (void *) xdr_void;
    xdrmem_create(xdr, xdrRespBuf, sizeof xdrRespBuf, XDR_ENCODE);

    if (!xdr_replymsg(xdr, &rpcreply))
    {
        printf("xdr_reply error\n");
        //FS_ASSERT(FALSE);
    }

    rc = sendto(fs_simNfsSrvrSocket,
                (char *) xdrRespBuf,
                xdr_getpos(xdr),
                0,
                (struct sockaddr *) fromAddr,
                sizeof(struct sockaddr_in));
    if (rc != xdr_getpos(xdr))
    {
        perror("sendto");
        FS_ASSERT(FALSE);
    }

} /* fs_simProcessBadProgRpc */

/*-----------------------------------------------------------------------------
 * fs_simProcessNfsRpc *
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
void
fs_simProcessNfsRpc(
    XDR                 *xdr, 
    struct rpc_msg      *rpccall, 
    struct sockaddr_in  *fromAddr,
    int32               threadNum)
{
    fs_SimNfsDispatchTbl    *nfsDispPtr;
    fs_MsgReqHdr            *reqHdr;
    fs_MsgRespHdr           *respHdr;
    fs_SimMsgQueueEntry     *qEntry;
    union {
        nfs_fh          getattr_2_arg;
        sattrargs       setattr_2_arg;
        diropargs       lookup_2_arg;
        nfs_fh          readlink_2_arg;
        readargs        read_2_arg;
        writeargs       write_2_arg;
        createargs      create_2_arg;
        diropargs       remove_2_arg;
        renameargs      rename_2_arg;
        linkargs        link_2_arg;
        symlinkargs     symlink_2_arg;
        createargs      mkdir_2_arg;
        diropargs       rmdir_2_arg;
        readdirargs     readdir_2_arg;
        nfs_fh          statfs_2_arg;

        GETATTR3args    GETATTR_3_args;
        SETATTR3args    SETATTR_3_args;
        LOOKUP3args     LOOKUP_3_args;
        ACCESS3args     ACCESS_3_args;
        READLINK3args   READLINK_3_args;
        READ3args       READ_3_args;
        WRITE3args      WRITE_3_args;
        CREATE3args     CREATE_3_args;
        MKDIR3args      MKDIR_3_args;
        SYMLINK3args    SYMLINK_3_args;
        MKNOD3args      MKNOD_3_args;
        REMOVE3args     REMOVE_3_args;
        RMDIR3args      RMDIR_3_args;
        RENAME3args     RENAME_3_args;
        LINK3args       LINK_3_args;
        READDIR3args    READDIR_3_args;
        READDIRPLUS3args READDIRPLUS_3_args;
        FSSTAT3args     FSSTAT_3_args;
        FSINFO3args     FSINFO_3_args;
        PATHCONF3args   PATHCONF_3_args;
        COMMIT3args     COMMIT_3_args;
    } argument;
    struct rpc_msg      rpcreply;
    int32               i, rc, uid, gid, ngids, reqLength;
    static uint32       nextFp;
    char                xdrRespBuf[65000];
    void                *xdrResp;
    struct auth_area {
        struct authunix_parms   aup;
        char                    machname[MAX_MACHINE_NAME+1];
        int                     gids[NGRPS];
    } auth_area;
    struct authunix_parms   *unix_cred;

    //printf("fs_simProcessNfsRpc: got %s request (procnum %ld)\n",
    //       fs_simNfsDispatchTbl[rpccall->rm_call.cb_proc].name,
    //       rpccall->rm_call.cb_proc);

    if (rpccall->rm_call.cb_proc != NFSPROC_NULL)
    {
        switch (rpccall->rm_call.cb_cred.oa_flavor)
        {
        case AUTH_UNIX:
            unix_cred = &auth_area.aup;
            unix_cred->aup_machname = auth_area.machname;
            unix_cred->aup_gids = auth_area.gids;
            rc = fs_simGetUnixCred(unix_cred, &rpccall->rm_call.cb_cred);
            if (rc != NFX_OK)
            {
                printf("Bad unix credential decode\n");
                return;
            }
            uid = unix_cred->aup_uid;
            gid = unix_cred->aup_gid;
            ngids = unix_cred->aup_len;
    
            //printf("got uid: %d gid: %d ngids: %d\n", uid, gid, unix_cred->aup_len);
            //for (i = 0; i < unix_cred->aup_len; i++)
            //{
            //    printf(" got gids[%d]: %d\n", i, *(unix_cred->aup_gids+i));
            //}
            break;
    
        case AUTH_NULL:
        default:
            printf("AUTH_NULL:  fssim ignoring request\n");
            return;
        }
    }

    if (rpccall->rm_call.cb_vers == NFS_VERSION_V2)
        nfsDispPtr = &(fs_simNfsDispatchTbl[rpccall->rm_call.cb_proc]);
    else
    if (rpccall->rm_call.cb_vers == NFS_VERSION_V3)
        nfsDispPtr = &(fs_simNfsDispatchTblV3[rpccall->rm_call.cb_proc]);
    else
        fs_error(NULL, FS_UNK_MSG_ERR, "wrong NFS version");

    //FS_TRACE(NULL, FS_TRACE_REQ,
    //        ("+++  NFS version %ld +++ %s +++", 
    //        rpccall->rm_call.cb_vers,
    //        (rpccall->rm_call.cb_vers == NFS_VERSION_V3) ?
    //            fs_simNfsDispatchTblV3[rpccall->rm_call.cb_proc].name :
    //            fs_simNfsDispatchTbl[rpccall->rm_call.cb_proc].name));

    memset ((char *)&argument, 0, sizeof (argument));

    if (nfsDispPtr->xdrArgument(xdr, &argument) != TRUE)
    {
        printf("rpc arg decode failed for proc %ld\n",
               rpccall->rm_call.cb_proc); 
        return;
    }

    reqHdr = fs_ramAlloc(FS_RAM_REQ_MSG, FS_SIM_MSG_BUF_LEN);
    rc = (nfsDispPtr->buildReq)(rpccall, &argument, &reqLength, reqHdr);
    if (rc != NFX_OK)
    {
        printf("fssim: buildFsReqFromNfsReq failed, dropping nfs req\n");
        return;
    }

	//printf("buildReq rtn length is %d\n", reqLength);

    if (rpccall->rm_call.cb_vers == NFS_VERSION_V2)
        reqHdr->proto = FS_PROTO_NFS_V2;
    else
    if (rpccall->rm_call.cb_vers == NFS_VERSION_V3)
        reqHdr->proto = FS_PROTO_NFS_V3;

    reqHdr->clientId = rpccall->rm_xid;
    reqHdr->nfsReqNum = rpccall->rm_call.cb_proc;
    reqHdr->uid = uid;
    reqHdr->gid = gid;
    if (rpccall->rm_call.cb_proc != NFSPROC_NULL)
    {
    	for (i = 0; i < unix_cred->aup_len; i++)
    	{
        	reqHdr->gids[i] = *(unix_cred->aup_gids + i);
    	}
    	reqHdr->numGids = unix_cred->aup_len;
	}
	else
	{
    	reqHdr->numGids = 0;
	}

    // For writes, send the data to the MCPU.
    if (reqHdr->requestType == FS_REQ_WRITE &&
        ((fs_MsgWriteReq*)reqHdr)->writeLength > 0)
    {
        mcpu_luc_msg_req_t  *mcpuReq;
        int32               mcpu_length, loop;
        char                *data_val = NULL;
        
        FS_TRACE(NULL, FS_TRACE_COD, ("NFS to MCPU send write msg, len: %d", 
                 (int32)((fs_MsgWriteReq*)reqHdr)->writeLength));

        mcpu_length = sizeof(mcpu_luc_msg_req_t) + 
          (int32)((fs_MsgWriteReq*)reqHdr)->writeLength;
        mcpuReq = fs_ramAlloc(FS_RAM_MCPU, mcpu_length);
        mcpuReq->src = FS_SIM_MSG_ADDR_NFS + threadNum;
        mcpuReq->dest = FS_SIM_MSG_ADDR_MCPU;
        mcpuReq->requestType = MCPU_LUC_WRITE_REQ;
        for (loop = 0; loop < NFX_MAX_COD_MSG; loop++)
        {
            mcpuReq->cod[loop] = ((fs_MsgWriteReq*)reqHdr)->cod[loop];
        }
        mcpuReq->nbrUsedCod = ((fs_MsgWriteReq*)reqHdr)->nbrUsedCod;
        mcpuReq->writeLength = (uint32)((fs_MsgWriteReq*)reqHdr)->writeLength;
        mcpuReq->sockIndex = threadNum;
        if (rpccall->rm_call.cb_vers == NFS_VERSION_V2)
        {
            data_val = ((writeargs*)(&argument))->data.data_val;
        }
        else
        if (rpccall->rm_call.cb_vers == NFS_VERSION_V3)
        {
            data_val = ((WRITE3args*)(&argument))->data.data_val;
        }
        
        if (data_val)
            eee_ramCopy(&(mcpuReq->data[0]), data_val, mcpuReq->writeLength);
        else
            printf ("MCPU write send msg, no argument data_val\n");

        rc = mcpu_sim_rx_data_from_LUC (mcpuReq);
        FS_ASSERT(rc == NFX_OK);
        fs_ramDealloc(FS_RAM_MCPU, mcpuReq);
    }

    //FS_TRACE(FS_TRACE_REQ, ("++++++ nfsrx:   got %s ++++++", 
    //  fs_simNfsDispatchTbl[reqHdr->nfsReqNum].name));

    // Send to the next FP.  Get our receive lock first to avoid
    // the potential race deadlock.
    rc = pthread_mutex_lock( &(fs_simNfsThread[threadNum].rxQueue.mutex));
    if (rc)
    {
        perror("pthread_mutex_lock fs_simProcessNfsRpc");
    }
    //printf("sending req to %d\n", nextFp % FS_SIM_NUM_FP_EXEC_THREADS);
    rc = fs_simSendMsg( FS_SIM_MSG_ADDR_NFS + threadNum,
                        nextFp++ % FS_SIM_NUM_FP_EXEC_THREADS,
                        (char *) reqHdr,
                        reqLength);
    if (rc != NFX_OK)
    {
        printf("nfsThreadRx: sendMsg failed\n");
        return;
    }

    // Now wait for the FP response.
    /* PAUSE */
    rc = pthread_cond_wait( &(fs_simNfsThread[threadNum].condVar),
                            &(fs_simNfsThread[threadNum].rxQueue.mutex));
    /* RESUME */
    if (rc)
    {
        perror("pthread_cond_wait fs_simProcessNfsRpc");
    }

    //printf("nfx_tx checking for msg\n");
    DLL_REMOVE_HEAD(&(fs_simNfsThread[threadNum].rxQueue.dllHdr),
                        qEntry, dllEntry);
    if (qEntry == NULL)
    {
        printf("nfs srvr thread: failed to find response msg\n");
        FS_ASSERT(FALSE);
    }

    //printf("nfs_tx got a msg\n");

    rc = pthread_mutex_unlock(&(fs_simNfsThread[threadNum].rxQueue.mutex));
    if (rc)
    {
        perror("pthread_mutex_unlock");
    }   

    respHdr = (fs_MsgRespHdr *) qEntry->msg;

    if (respHdr->requestType ==  FS_REQ_READ &&
        respHdr->fsStatus == FS_ERR_NONE)
    {
    	fs_MsgReadResp	*fsReadResp = (void *) respHdr;

        fs_ramDealloc(FS_RAM_SIM_MSG_QENTRY, qEntry);

        FS_TRACE(NULL, FS_TRACE_COD,
                 ("FS to MCPU send read msg, data len: %d", 
                 respHdr->len));

        rc = mcpu_sim_tx_data_to_LUC(&fsReadResp);
        if (rc != NFX_OK)
        {
            fsReadResp->hdr.fsStatus = FS_ERR_INVAL;
            fsReadResp->bytesRead = 0;
        }

		// Careful here, mcpu_sim may have reallocated!
		respHdr = (void *) fsReadResp;

        if (fsReadResp->bytesRead)
        {
            fs_MsgTransmitComplete *fsTxCompl;

            fsTxCompl = fs_ramAlloc(FS_RAM_REQ_MSG,
                                    sizeof(fs_MsgTransmitComplete));
            bzero (fsTxCompl, sizeof(fs_MsgTransmitComplete));
            fsTxCompl->hdr.requestType = FS_REQ_TX_COMPL ;
            memcpy (&(fsTxCompl->fileHandle), &(fsReadResp->fileHandle),
                    sizeof(fs_Handle)) ;
            fsTxCompl->fileOffset = fsReadResp->fileOffset ;
            fsTxCompl->bytesRead = fsReadResp->bytesRead ;
            fsTxCompl->fsPrivate = fsReadResp->fsPrivate ;

        	rc = fs_simSendMsg(FS_SIM_MSG_ADDR_NFS,
                               ++nextFp % FS_SIM_NUM_FP_EXEC_THREADS,
                               (char *) fsTxCompl,
                               sizeof(fs_MsgTransmitComplete));
        }
    }
	else
	{
   		fs_ramDealloc(FS_RAM_SIM_MSG_QENTRY, qEntry);
	}

    // Cleanup the XDR from the call, ready it for encoding the
    // reply.
    xdr->x_op = XDR_FREE;
    if (nfsDispPtr->xdrArgument(xdr, &argument) != TRUE)
    {
        FS_ASSERT(FALSE);
    }
    memset(xdr, 0 , sizeof *xdr);
    memset(&rpcreply, 0, sizeof rpcreply);

    // Put together the RPC reply header.
    rpcreply.rm_xid = rpccall->rm_xid;
    rpcreply.rm_direction = REPLY;
    rpcreply.rm_reply.rp_stat = MSG_ACCEPTED;
    rpcreply.rm_reply.rp_acpt.ar_stat = SUCCESS;
    rpcreply.acpted_rply.ar_results.proc = (void *) xdr_void;
    xdrmem_create(xdr, xdrRespBuf, sizeof xdrRespBuf, XDR_ENCODE);

    if (!xdr_replymsg(xdr, &rpcreply))
    {
        printf("xdr_reply error\n");
        //FS_ASSERT(FALSE);
    }

    // Encode and add the reply arguments.
    xdrResp = (nfsDispPtr->buildResp)(respHdr);

    if (respHdr->nfsReqNum != FS_REQ_NULL && xdrResp == NULL)
    {
        printf("Failed to create NFS response message.\n");
        fs_ramDealloc(FS_RAM_RESP_MSG, respHdr);
        return;
    }
    rc = (nfsDispPtr->xdrResult)(xdr, xdrResp);
    if (rc != TRUE)
    {
        printf("Failed to encode NFS response message.\n");
        fs_ramDealloc(FS_RAM_RESP_MSG, respHdr);
        return;
    }

    //FS_TRACE(FS_TRACE_REQ, ("------ nfstx: reply %s ------", 
    //      fs_simNfsDispatchTbl[fsHdr->nfsReqNum].name));

    rc = sendto(fs_simNfsSrvrSocket,
                (char *) xdrRespBuf,
                xdr_getpos(xdr),
                0,
                (struct sockaddr *) fromAddr,
                sizeof(struct sockaddr_in));
    if (rc != xdr_getpos(xdr))
    {
        perror("sendto");
        FS_ASSERT(FALSE);
    }

    if (nfsDispPtr->freeResp != NULL)
    {
        (nfsDispPtr->freeResp)(xdrResp);
    }

    fs_ramDealloc(FS_RAM_RESP_MSG, respHdr);

} /* fs_simProcessNfsRpc */

/*-----------------------------------------------------------------------------
 * fs_simNfsSrvrThread
 *
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
void *
fs_simNfsSrvrThread(void *arg)
{
    struct sockaddr_in  sockAdd, fromAddr;
    struct hostent      *host;
    int32               threadNum, msgLen, fromLen, sock;
    char                rxBuf[FS_SIM_MSG_BUF_LEN];
    XDR                 xdr;
    struct rpc_msg      rpcmsg;

    threadNum = (int32) arg;

    //printf("I'm thread num %d\n", threadNum);

    memset (&sockAdd, 0, sizeof(sockAdd));
    sockAdd.sin_family = AF_INET;

    host = gethostbyname("localhost");
    if (!host) {
        perror ("gethostbyname");
        pthread_exit((void *) 1);
    }
    memcpy (&sockAdd.sin_addr, host->h_addr, host->h_length);

    sock = socket (AF_INET, SOCK_DGRAM, 0);
    if (sock < 0) {
        perror("socket");
        pthread_exit((void *) 1);
    }
    fs_simNfsIntSockets[threadNum] = sock;
    sockAdd.sin_port = htons(FS_SIM_NFS_MUX_PORT_START + threadNum);

    //printf("binding intfsport %d to intsocket %d index %d\n",
    //    FS_SIM_NFS_MUX_PORT_START + threadNum, sock, threadNum);

    if (bind (sock, (struct sockaddr *)&sockAdd, sizeof(sockAdd)) < 0) {
        perror ("bind internal nfs socket");
        close(sock);
        pthread_exit((void *) 1);
    }

    fromLen = sizeof(fromAddr);
    for (;;)
    {
        msgLen = recv (sock, rxBuf, FS_SIM_MSG_BUF_LEN, 0);
        if (msgLen <= sizeof(fromLen))
        {
            perror("recv nfs rx");
            continue;
            //pthread_exit((void *) 1);
        }

        // Save the client's address.
        memcpy(&fromAddr, rxBuf, fromLen);

        // Decode the rpc.
        memset(&xdr, 0, sizeof xdr);
        memset(&rpcmsg, 0, sizeof rpcmsg);
        xdrmem_create(&xdr, &rxBuf[fromLen], msgLen - fromLen, XDR_DECODE);
        if (xdr_callmsg(&xdr, &rpcmsg))
        {
            //printf("xdrpos: %d\n", xdr_getpos(&xdr));
            //printf("rm_direction: %d\n", rpcmsg.rm_direction);
            //printf("rm_xid: %lx\n", rpcmsg.rm_xid);
            //printf("rm_call.cb_rpcvers: %ld\n", rpcmsg.rm_call.cb_rpcvers);
            //printf("rm_call.1016568293cb_prog: %ld\n", rpcmsg.rm_call.cb_prog);
            //printf("rm_call.cb_vers: %ld\n", rpcmsg.rm_call.cb_vers);
            //printf("rm_call.cb_proc: %ld\n", rpcmsg.rm_call.cb_proc);
            //printf("xdrpos: %d\n", xdr_getpos(&xdr));
    
            if ((rpcmsg.rm_call.cb_prog == NFS_PROGRAM) &&
                (((rpcmsg.rm_call.cb_vers == NFS_VERSION_V2) &&
                  (rpcmsg.rm_call.cb_proc >= NFSPROC_NULL) &&
                  (rpcmsg.rm_call.cb_proc <= NFSPROC_STATFS)) ||
                 ((rpcmsg.rm_call.cb_vers == NFS_VERSION_V3) &&
                  (rpcmsg.rm_call.cb_proc >= FS_REQ_NULL) &&
                  (rpcmsg.rm_call.cb_proc <= FS_REQ_COMMIT))))
            {
                // Process it and respond.
                fs_simProcessNfsRpc(&xdr, &rpcmsg, &fromAddr, threadNum);
            }
            else if (rpcmsg.rm_call.cb_prog == 100227)
            {
                // wacky nfs acl request from solaris, send
                // "program number not supported" response.
                fs_simProcessBadProgRpc(&xdr, &rpcmsg, &fromAddr, threadNum);
            }

            // else ignore it

            xdr.x_op = XDR_FREE;
            xdr_callmsg(&xdr, &rpcmsg);
            xdr_destroy(&xdr);
        }
        else
        {
            printf("nfs srvr rpc decode failed, ignoring pkt\n");
        }
    }
        
    return NULL;

} /* fs_simNfsSrvrThread */

/*-----------------------------------------------------------------------------
 * fs_simMountThread
 *
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
void *
fs_simMountThread(void *arg)
{
    struct hostent      *hp;
    struct rlimit       rl;
    struct sockaddr_in  sin;
    SVCXPRT             *tp;
    int32               svrsock;

    //printf("mount thread %d\n", pthread_self());

    //printf("nfs mount on port %d\n", fs_simMountPort);

    if ((hp=gethostbyname("localhost"))==NULL) {
        fprintf(stderr,"Can't deal with localhost\n");
        pthread_exit((void *) 1);
    }
    bzero((char *)&sin,sizeof(sin));
    sin.sin_family=AF_INET;
    /* bcopy((char *)hp->h_addr,(char *)&sin.sin_addr,hp->h_length); */
    /* sin.sin_addr = inet_makeaddr(INADDR_ANY,np);*/
    sin.sin_port = htons(fs_simMountPort);

    if ((svrsock=socket(AF_INET,SOCK_DGRAM,0)) < 0) {
        perror("socket");
        pthread_exit((void *) 1);
    }

    if (bind(svrsock,(struct sockaddr *)&sin,sizeof(sin)) != 0) {
        perror("bind mount socket");
        pthread_exit((void *) 1);
    }

    if ((tp = svcudp_create(svrsock)) == NULL) {
        fprintf(stderr,"Can't create UDP RPC Service\n");
        pthread_exit((void *) 1);
    }
    
    // Register with the local RPC dispatcher.  If we're
    // using the "standard" mount port, register with the
    // portmapper, otherwise assume another mountd is already
    // running and don't attempt to register.

    if (fs_simPortmapMount)
    {
        // Unregister current program.
        pmap_unset(MOUNTPROG, MOUNTVERS);
        pmap_unset(MOUNTPROG, MOUNTVERS_V3);
    }

    // register for version 3
    {
        if (!svc_register(  tp,
                            MOUNTPROG,
                            MOUNTVERS_V3,
                            fs_simMountDispatch1,                        // Todo make fs_simMountDispatch3
                            (fs_simPortmapMount ? IPPROTO_UDP :0 )))
        {
            perror("svc_register");
            fprintf(stderr,"Can't register fsSim Mount\n");
            pthread_exit((void *) 1);
        }
    }
    // register for version 1
    {
        if (!svc_register(  tp,
                            MOUNTPROG,
                            MOUNTVERS,
                            fs_simMountDispatch1,
                            (fs_simPortmapMount ? IPPROTO_UDP :0 )))
        {
            perror("svc_register");
            fprintf(stderr,"Can't register fsSim Mount\n");
            pthread_exit((void *) 1);
        }
    }
    
    /* first set uid to 0, if we can */
    /* now we can go back and forth easily */
    setuid(0);
    umask(0);

    /* try to make sure we don't spill a corefile */
    rl.rlim_cur=0;
    rl.rlim_max=0;
    setrlimit(RLIMIT_CORE,&rl);

    svc_run(); /* do it */

    fprintf(stderr,"fs_simMountThread:  svc_run() returned!\n");
    FS_ASSERT(FALSE);

    return NULL;

} /* fs_simMountThread */


int32
mcpu_sim_write_to_file();
int32
nvram_write_to_file();

// As defined in ../sm-nvram/nvram.c
// Default FALSE
// Set to TRUE when nvram is used
// If FALSE, the nvram and cod's are not saved to a file
//  when the fssim crashes.
// Only for standalone.
extern int nvram_ever_used;

/**
 * The signal handler for cleaning up after readline.
 */
static void fs_simCleanupHandler(
    int sig)
{
    int loop;
    
    if (readlineInUse) {        // TODO - make this atomic
        readlineInUse = FALSE;
        rl_cleanup_after_signal();
    }

    if (nvram_ever_used)
    {
        mcpu_sim_write_to_file();
        nvram_write_to_file();
    }

    for (loop = 0 ; loop < (FS_SIM_VOL_MAX + 1) ; loop++)
    {
        if (volFd[loop])
            close (volFd[loop]);
    }
    
    exit(1);
}


/**
 * Handler for printing debugging information after a signal.
 */
static void
fs_simSigHandler(
    int       sig,
    siginfo_t *siginfo,
    void      *context)
{
    struct timeval 	now;
 
    gettimeofday (&now, NULL);

    fprintf(stderr, "pid %d: %s (signal %d) at %s",
            getpid(), sys_siglist[sig], sig, ctime(&now.tv_sec));
#ifdef i386
    if (sig == SIGBUS || sig == SIGSEGV) {
        fprintf(stderr, "fault address is (probably) %p\n",
                ((void **) siginfo)[14]);
    }
#endif
    btrace_print_signal(context, fs_simSigHandler);

    if (readlineInUse) {        // TODO - make this atomic
        readlineInUse = FALSE;
        rl_cleanup_after_signal();
    }

    abort();
}


/**
 * Initalize the signal handlers for cleaning up after readline.
 */
static void fs_simInitCleanupHandler()
{
    int i;
    struct sigaction sa;

    static int signals[] = { SIGINT, SIGQUIT, SIGABRT, SIGTERM, SIGUSR2, 0 };
    
    memset(&sa, 0, sizeof(struct sigaction));
    sa.sa_handler = fs_simCleanupHandler;

    for (i = 0; signals[i] != 0; i++) {
        int sig = signals[i];
        if (sigaction(sig, &sa, NULL) == -1) {
            char buf[BUFSIZ];
            sprintf(buf, "sigaction(%d \"%s\")", sig, sys_siglist[sig]);
            perror(buf);
            exit(1);
        }
    }
}


/**
 * Initialize the signal handlers.
 */
static void fs_simInitSigHandler()
{
    int i;
    struct sigaction sa;

    static int signals[] = { SIGSEGV, SIGBUS, SIGILL, SIGFPE, 0 };
    
    memset(&sa, 0, sizeof(struct sigaction));
    sa.sa_sigaction = fs_simSigHandler;
    sa.sa_flags = SA_SIGINFO;

    for (i = 0; signals[i] != 0; i++) {
        int sig = signals[i];
        if (sigaction(sig, &sa, NULL) == -1) {
            char buf[BUFSIZ];
            sprintf(buf, "sigaction(%d \"%s\")", sig, sys_siglist[sig]);
            perror(buf);
            exit(1);
        }
    }
}


/*-----------------------------------------------------------------------------
 * fs_simSscExecThreadMain
 *
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
void *
fs_simSscExecThreadMain(void *arg)
{
    struct timeval          now;
    struct timespec         timeout;
    fs_MsgReqHdr            *reqHdr = NULL;
    fs_MsgRespHdr           *respHdr;
    fs_SimMsgQueueEntry     *respEntry;
    uint32                  msgNum;
    uint32                  nextFp, thisFp;
    int32                   reqType;
    int32                   rc, numTimedOut;
    fs_MsgReqHdr *          (*resp)(int status, fs_MsgRespHdr *);
    uint32                  sender = -1;

    //printf("ssc exec thread %d\n", pthread_self());

    fs_simInitCleanupHandler();

    thisFp = -1;
    nextFp = 0;
    msgNum = 1;
    numTimedOut = 0;

    using_history();

    while (TRUE) {
        char *inputStr;

        int waitSecs = 300;
        //waitSecs = 10;

	/*
	 * Prompt for a new command only if the previous command
	 * has finished executing.
	 */
	if (reqHdr == NULL) {
            do {
                readlineInUse = TRUE;
                inputStr = readline("fs> ");
            } while (inputStr == NULL);

            if (strlen(inputStr) > 0)
                add_history(inputStr);
            reqHdr = fs_shell_parsecmd(inputStr, &resp);
            free(inputStr);

            if (reqHdr == NULL)
                continue;
	}

        // Hack to make the memstats look right.
        fs_ramStats[FS_RAM_STATS_ALLOC][FS_RAM_REQ_MSG] += 1;

        reqType = reqHdr->requestType;

        if (reqType < FSSF_MSG_BASE)
        {
            // Send to FS

            // Send the request message.
            if (thisFp == -1)
            {
                // Round robin the FP's
                thisFp = nextFp++;
                nextFp %= FS_SIM_NUM_FP_EXEC_THREADS;
            }
            sender = thisFp;
        }
        else
        {
            // Send to FSSF
            sender = FS_SIM_MSG_ADDR_FSSF;
        }
        
#ifdef ISABELLE // don't modify the clientPrivate!
        reqHdr->clientPrivate = msgNum;
#endif 
            
        rc = pthread_mutex_lock(&(fs_simSscExec.mutex));
        if (rc)
        {
            perror("pthread_mutex_lock");
            goto sscErr;
        }

        rc = fs_simSendMsg(FS_SIM_MSG_ADDR_SSC,
                           sender,
                           (char *) reqHdr,
                           reqHdr->len);
        if (rc != NFX_OK)
        {
            printf("sscMain: sendMsg failed\n");
        }
        sender = -1;
        thisFp = -1;

        gettimeofday(&now, NULL);
        timeout.tv_sec = now.tv_sec + waitSecs;
        timeout.tv_nsec = now.tv_usec * 1000;
    
        /* PAUSE - wait for response msg */
        rc = pthread_cond_timedwait(    &(fs_simSscExec.condVar),
                                        &(fs_simSscExec.mutex),
                                        &timeout);
        if (rc == ETIMEDOUT)
        {
            //printf("FS_ERR: SCC response timeout\n");
        }
        else if (rc)
        {
            perror("pthread_cond_wait fs_simSscExecThreadMain");
            goto sscErr;
        }
        /* RESUME */
        respHdr = NULL;
        while (!DLL_IS_EMPTY(&(fs_simSscRxQueue.dllHdr)))
        {
            DLL_REMOVE_HEAD(&(fs_simSscRxQueue.dllHdr), respEntry, dllEntry);
            respHdr = (fs_MsgRespHdr *) respEntry->msg;
            fs_ramDealloc(FS_RAM_SIM_MSG_QENTRY, respEntry);

#ifdef ISABELLE // not valid for FSSF response
            if (respHdr->requestType != reqType)
            {
                printf("SCC_ERR: Wrong respType, exp %d, got %d\n",
                    reqType, respHdr->requestType);

                fs_ramDealloc(FS_RAM_RESP_MSG, respHdr);
                continue;
            }
#endif // ISABELLE
#ifdef ISABELLE // don't modify the clientPrivate!
            if (respHdr->clientPrivate != msgNum)
            {
                printf("SCC_ERR: Wrong msgNum, exp %d, got %d\n",
                    msgNum, (int32) respHdr->clientPrivate);

                fs_ramDealloc(FS_RAM_RESP_MSG, respHdr);
                continue;
            }
#endif // ISABELLE
        }
        rc = pthread_mutex_unlock(&(fs_simSscExec.mutex));
        if (rc)
        {
            perror("pthread_mutex_unlock");
            goto sscErr;
        }

        if (respHdr == NULL)
        {
            numTimedOut += 1;
            printf("sscExec: response timed out\n");
            continue;
        }

#ifdef ISABELLE // not valid for FSSF response
        if (respHdr->fsStatus != NFX_OK)
        {
            printf("FS_ERR: %s completed with %d status\n",
                   fs_reqTypeTbl[reqType].name,
                   respHdr->fsStatus);
        }
#endif // ISABELLE
        
        reqHdr = resp(respHdr->fsStatus, respHdr);

#ifdef ISABELLE // debug only!!!
        printf("msgNum = %d\n", msgNum);
#endif // ISABELLE
        fs_ramDealloc(FS_RAM_RESP_MSG, respHdr);
        msgNum += 1;
    } // while (TRUE)

sscErr:
    return NULL;

} /* fs_simSscExecThreadMain */

/*-----------------------------------------------------------------------------
 * fs_simMcpuExecThreadMain
 *
 * Created By:  
 *
 * Date Created:  7/11/2001
 */

void *
fs_simMcpuExecThreadMain(void *arg)
{
    fs_MsgReqHdr            *reqHdr = NULL;
    fs_SimMsgQueueEntry     *reqEntry = NULL ;
    int32                   rc;
    uint32                  nextFp;
    int32                   numMsgsRcvd;
    uint32                  requestType;

    if ((rc = mcpu_sim_sbm_init()) != NFX_OK)
      {
        printf("Mcpu init error %#x\n", rc);
        goto McpuErr;
      }
    
    for (;;)
      {

        rc = pthread_mutex_lock(&(fs_simMcpuQueue.mutex));
        if (rc)
          {
            perror("pthread_mutex_lock Mcpu");
            goto McpuErr;
          }

        numMsgsRcvd = 0 ;
        
        if ( ! DLL_IS_EMPTY (&(fs_simMcpuQueue.dllHdr)))
        {
            DLL_REMOVE_HEAD(&(fs_simMcpuQueue.dllHdr), reqEntry, dllEntry);
            if (reqEntry == NULL)
              {
                perror ("Failed to get qEntry from job queue!!!\n");
                goto McpuErr;
              }

            reqHdr = (fs_MsgReqHdr *) reqEntry->msg;
            fs_ramDealloc(FS_RAM_SIM_MSG_QENTRY, reqEntry);

            numMsgsRcvd++;

        }
        
        rc = pthread_mutex_unlock(&(fs_simMcpuQueue.mutex));
        if (rc)
          {
            perror("pthread_mutex_unlock Mcpu");
            goto McpuErr;
          }

        if (numMsgsRcvd > 0)
          {
            /* process received message */
            mcpu_luc_msg_req_t *mcpuReq;

            mcpuReq = (mcpu_luc_msg_req_t*)reqHdr;
            requestType = mcpuReq->requestType;

            if (requestType == MCPU_LUC_WRITE_REQ) 
            {
                FS_TRACE(NULL, FS_TRACE_COD,
                         ("fs-sim mcpu_sim_rx_data_from_LUC, nbr cod %d, len %d", 
                         mcpuReq->nbrUsedCod, mcpuReq->writeLength));

                rc = mcpu_sim_rx_data_from_LUC (mcpuReq);
                
                // mcpu_sim_sbm_show_all (MCPU_SHOW_ALL);

#ifndef SBM_REV_2
                if (rc == NFX_OK)
                {
                    uint32             loop ;
                    
                    for (loop = 0; loop < mcpuReq->nbrUsedCod; loop++)
                    {
                        mcpu_confirm_cod_t *mcpu_cnf;
                    
                        mcpu_cnf = fs_ramAlloc(FS_RAM_MCPU, sizeof (mcpu_confirm_cod_t));
                        mcpu_cnf->requestType = MCPU_CONFIRM_COD ;
                        mcpu_cnf->cod = mcpuReq->cod[loop].cod;
                        rc = fs_simSendMsg(FS_SIM_MSG_ADDR_MCPU,
                                           ++nextFp % FS_SIM_NUM_FP_EXEC_THREADS,
                                           (char *) mcpu_cnf,
                                           sizeof (mcpu_confirm_cod_t));
                        if (rc != NFX_OK)
                        {
                            printf("nfsThreadRx: sendMsg failed\n");
                            pthread_exit((void *) 1);
                        }
                    }
                    
                }
#endif // SBM_REV_2

                fs_ramDealloc(FS_RAM_MCPU, reqHdr);
            }
            else
            if (requestType == MCPU_LUC_READ_REQ) 
            {
                fs_MsgReadResp *fsReadResp = (fs_MsgReadResp*)reqHdr;
                fs_MsgTransmitComplete *fsTxCompl;
                int32 src = reqHdr->src;
                
                FS_TRACE(NULL, FS_TRACE_COD,
                         ("fs-sim mcpu_sim_tx_data_to_LUC, nbr of cod %d, off %lld, len %d", 
                          fsReadResp->nbrUsedCod, 
                          fsReadResp->fileOffset, fsReadResp->bytesRead));

                /* (fsReadResp == reqHdr) Deallocated, and a new is allocated */ 
                rc = mcpu_sim_tx_data_to_LUC (&fsReadResp);
                                
                if (rc != NFX_OK)
                {
                    fsReadResp->hdr.fsStatus = FS_ERR_INVAL;
                    fsReadResp->bytesRead = 0;
                }

                if (fsReadResp->bytesRead)
                {
                    fsTxCompl = fs_ramAlloc(FS_RAM_REQ_MSG, sizeof(fs_MsgTransmitComplete));
                    bzero (fsTxCompl, sizeof(fs_MsgTransmitComplete));
                    fsTxCompl->hdr.requestType = FS_REQ_TX_COMPL ;
                    memcpy (&(fsTxCompl->fileHandle), &(fsReadResp->fileHandle), sizeof(fs_Handle)) ;
                    fsTxCompl->fileOffset = fsReadResp->fileOffset ;
                    fsTxCompl->bytesRead = fsReadResp->bytesRead ;
                    fsTxCompl->fsPrivate = fsReadResp->fsPrivate ;
                }
                else
                    fsTxCompl = NULL;
                
                rc = fs_simSendMsg(FS_SIM_MSG_ADDR_MCPU,
                                   src,
                                   (char *) fsReadResp,
                                   fsReadResp->hdr.len);

                if (rc != NFX_OK)
                {
                    printf("nfsThreadRx: sendMsg to NFS_TX failed\n");
                    pthread_exit((void *) 1);
                }

                if (fsTxCompl)
                {
                    rc = fs_simSendMsg(FS_SIM_MSG_ADDR_NFS,
                                       ++nextFp % FS_SIM_NUM_FP_EXEC_THREADS,
                                       (char *) fsTxCompl,
                                       sizeof(fs_MsgTransmitComplete));

                    if (rc != NFX_OK)
                    {
                        printf("nfsThreadRx: sendMsg to FP failed\n");
                        pthread_exit((void *) 1);
                    }
                }
                
            }
            else 
            if (requestType == MCPU_SHOW_STAT_REQ) 
              {
                mcpu_show_stat_req_t *mcpuShow ;

                mcpuShow = (mcpu_show_stat_req_t*)reqHdr;
                mcpu_sim_sbm_show_all (mcpuShow->level);

                fs_ramDealloc(FS_RAM_REQ_MSG, reqHdr);
              }
            else
              {
                printf("nfsThreadRx: unknown requestType %d\n", requestType);
                fs_ramDealloc(FS_RAM_REQ_MSG, reqHdr);
              }
          }
        else
          {
            rc = pthread_mutex_lock(&(fs_simMcpuExec.mutex));
            if (rc)
              {
                perror("pthread_mutex_unlock Mcpu");
                goto McpuErr;
              }

            rc = pthread_cond_wait( &(fs_simMcpuExec.condVar),
                                    &(fs_simMcpuExec.mutex));

            rc = pthread_mutex_unlock(&(fs_simMcpuExec.mutex));
            if (rc)
              {
                perror("pthread_mutex_unlock Mcpu");
                goto McpuErr;
              }

            if (rc == ETIMEDOUT)
              {
                //printf("FS_ERR: Mcpu timeout\n");
              }
            else if (rc)
              {
                perror("pthread_cond_wait Mcpu");
                goto McpuErr;
              }
          }
      }

McpuErr:
    perror("pthread Mcpu stopped due to major problem");
    return NULL;

} /* fs_simMcpuExecThreadMain */

/**
 * Print some memory to standard output.
 */
void memdump(
    const char *name,
    const void *memory,
    uint32 size)
{
    uint32 i;
    const uint8 *bytes = memory;
    enum { ROWSIZE = 16 };

    printf("\n");
    printf("# %s:\n", name);
    for (i = 0; i < size; i += ROWSIZE) {
        int j;
        const uint8 *row = &bytes[i];
        boolean nonzero = FALSE;
        for (j = 0 ; (i + j) < size && j < ROWSIZE; j++) {
            if (row[j] != 0) {
                nonzero = TRUE;
                break;
            }
        }
        if (nonzero) {
            printf("#  %+5d (%p):", i, &bytes[i]);
            for (j = 0; (i + j) < size && j < 16; j++) {
                printf(" %02x", row[j]);
            }
            printf("\n");
        }
    }
    printf("\n");
    fflush(stdout);
}

/**
 * Run the file server.
 */
static void fs_simRunServer()
{
    int i;
    int rc;
    fs_SimNfsThread *thread;
    pthread_t mountThread, muxThread, evmThread, fssfThread;

    NFX_SPIN_INIT(&wait_time_lock,SPINLOCK_LEVEL_UNKNOWN);
    eee_rtcSSCTimeInit(sscGMTTime);
    
    fs_initIncoreTune();

    fs_rdwrInit();

	fs_lockOrderInit();

    // Initialize the in-core file system structure table.
    fs_initFileSystemTable();

    fs_bufCacheInit();

    fs_dnlcInit();

    // Init inode cache.
    fs_inodeCacheInit();

    fs_quotaCacheInit();

    fs_dirhash_cache_init();

    // Init btree lib.
    fs_bt_dispatch_init();

    e_log_init();

    // Init the CODs
    fs_cod_init ();
    
#ifdef NEW_ZF
    fs_zfInit();
#endif // NEW_ZF

    fs_vlcInit();

    // nvram_init();

    DLL_INIT(&fs_volExerDllHdr);

    // Initialize FSSF
    rc = fssf_init();
    if (rc != 0)
    {
        perror("fssf_init() failed.");
        exit(1);
    }

    // Init EVM server
    rc = pthread_mutex_init(&volMutex, NULL);
    if (rc != 0) {
        perror("pthread_mutex_init");
        exit(1);
    }
    rc = pthread_cond_init(&volCondVar, NULL);
    if (rc != 0) {
        perror("pthread_cond_init");
        exit(1);
    }
    DLL_INIT(&fs_simEvmHdr);
    rc = pthread_create(&evmThread,
                        NULL,
                        &fs_simEvmThread,
                        NULL);
    if (rc != 0) {
        perror("pthread_create");
        exit(1);
    }
    
    /*
     * Init fp executives.
     */
    rc = pthread_key_create(&fs_simKeyCpuId, NULL);
    if (rc)
    {
        perror("pthread_key_create");
        exit(1);
    }
    rc = pthread_key_create(&fs_simKeyReqThreadNum, NULL);
    if (rc)
    {
        perror("pthread_key_create");
        exit(1);
    }

    rc = pthread_mutex_init(&fs_simFpExecMutex, NULL);
    if (rc != 0) {
        perror("pthread_mutex_init");
        exit(1);
    }
    DLL_INIT(&fs_simThreadJobQueue.dllHdr);

    memset(fs_simReqThread, 0,
           sizeof(fs_SimReqThread) * FS_SIM_EXEC_REQ_THREADS_MAX);

    for (i = 0; i < FS_SIM_NUM_FP_EXEC_THREADS; i++)
    {
        DLL_INIT(&(fs_simFpRxQueue[i].dllHdr));

        rc = pthread_mutex_init(&(fs_simFpRxQueue[i].mutex), NULL);
        if (rc != 0) {
            perror("pthread_mutex_init");
            exit(1);
        }

        rc = pthread_cond_init(&(fs_simFpRxQueue[i].condVar), NULL);
        if (rc != 0) {
            perror("pthread_cond_init");
            exit(1);
        }

        rc = pthread_cond_init(&(fs_simFpExec[i].condVar), NULL);
        if (rc != 0) {
            perror("pthread_cond_init");
            exit(1);
        }

        rc = pthread_create(    &(fs_simFpExec[i].thread),
                                NULL,
                                &fs_simFpExecThreadMain,
                                (void *) (i % FS_SIM_NUM_FP_CPUS));
        if (rc != 0) {
            perror("pthread_create");
            exit(1);
        }
    }

    // Start the NFS server threads.
    for (i = 0; i < FS_SIM_NFS_THREAD_MAX; i++)
    {
        thread = &fs_simNfsThread[i];
        DLL_INIT(&thread->rxQueue.dllHdr);
        rc = pthread_mutex_init(&thread->rxQueue.mutex, NULL);
        if (rc != 0) {
            perror("pthread_mutex_init");
            exit(1);
        }
        rc = pthread_cond_init(&thread->condVar, NULL);
        if (rc != 0) {
            perror("pthread_cond_init");
            exit(1);
        }
        rc = pthread_create(    &thread->thread,
                                NULL,
                                &fs_simNfsSrvrThread,
                                (void *) i);
        if (rc != 0) {
            perror("pthread_create");
            exit(1);
        }
    }

    // create NFS mux thread
    rc = pthread_create(&muxThread,
                        NULL,
                        fs_simNfsServerMuxThread,
                        NULL);
    if (rc != 0) {
        perror("nfsmux pthread_create");
        exit(1);
    }

    // create the fssf thread
    DLL_INIT(&fs_simFssfQueue.dllHdr);
    rc = pthread_mutex_init(&fs_simFssfQueue.mutex, NULL);
    if (rc != 0) {
        perror("pthread_mutex_init");
        exit(1);
    }
    rc = pthread_cond_init(&fs_simFssfQueue.condVar, NULL);
    if (rc != 0) {
        perror("pthread_cond_init");
        exit(1);
    }
    rc = pthread_create(&fssfThread,
                        NULL,
                        &fs_simFssfThread,
                        NULL);
    if (rc != 0) {
        perror("pthread_create");
        exit(1);
    }

    // Start the NFS mount thread.
    rc = pthread_create(    &mountThread,
                            NULL,
                            &fs_simMountThread,
                            NULL);
    if (rc != 0) {
        perror("pthread_create");
        exit(1);
    }

    // Start SSC thread
    DLL_INIT(&(fs_simSscRxQueue.dllHdr));
    rc = pthread_mutex_init(&(fs_simSscRxQueue.mutex), NULL);
    if (rc)
    {
        perror("pthread_mutex_init");
        exit(1);
    }

    rc = pthread_create(    &(fs_simSscExec.thread),
                            NULL,
                            &fs_simSscExecThreadMain,
                            NULL);
    if (rc != 0) {
        perror("pthread_create");
        exit(1);
    }

    DLL_INIT(&(fs_simMcpuQueue.dllHdr));
    rc = pthread_mutex_init(&(fs_simMcpuQueue.mutex), NULL);
    if (rc)
    {
        perror("pthread_mutex_init Mcpu");
        exit(1);
    }

    rc = pthread_create(    &(fs_simMcpuExec.thread),
                            NULL,
                            &fs_simMcpuExecThreadMain,
                            NULL);
    if (rc != 0) {
        perror("pthread_create Mcpu");
        exit(1);
    }

    // Wait for SSC thread to exit... (Why just one)
    rc = pthread_join(fs_simSscExec.thread, NULL);
    if (rc != 0) {
        perror("pthread_join");
        exit(1);
    }
}

/*-----------------------------------------------------------------------------
 * main
 *
 * Description:  Entry point for fs simulation environment
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
int
main(
    int     argc,
    char    **argv)
{
    int rc;

    fs_memInit();
    
    if (argc > 1)
    {
        int loop;
        
        for (loop = 1; loop < argc; loop++)
        {
            if (!strncmp(argv[loop], "-p", 2))
            {
                fs_simPortmapNfs = TRUE;
                fs_simPortmapMount = TRUE;
                fs_simMountPort = FS_SIM_MOUNT_PORT_MAPPED;
            }
        }
        
    }

    // install the debugging signal handlers

    fs_simInitSigHandler();

    // initialize stack backtracing - only needed if malloc gets corrupted

    btrace_init();

    // Init FP globals - TODO: move to fs_fsForkServer?

    rc = pthread_mutexattr_init(&fs_simMutexAttr);   
    if (rc != 0)
    {
        perror("mutexattr_init");
        exit(2);
    }
    rc = pthread_mutexattr_settype(&fs_simMutexAttr,
                                   PTHREAD_MUTEX_ERRORCHECK_NP);
    if (rc != 0)
    {
        perror("mutexattr_setkind_np");
        exit(2);
    }

    bzero(fs_errorCounts, sizeof(fs_errorCounts));
    bzero(fs_simEsmInstance, sizeof(fs_simEsmInstance));
    bzero(fs_ramStats, sizeof(fs_ramStats));
    NFX_SPIN_INIT(&fs_ramStatsSpinId,SPINLOCK_LEVEL_UNKNOWN);

    bzero(&fs_nfsStats, sizeof(fs_nfsStats));
    bzero(&fs_yieldStats, sizeof(fs_yieldStats));
    bzero(&fs_recvReqStats, sizeof(fs_recvReqStats));
    bzero(&fs_replyStats, sizeof(fs_ReplyStats));
    NFX_SPIN_INIT(&fs_statsReqRcvdSpinId,SPINLOCK_LEVEL_UNKNOWN);

    DLL_INIT(&fs_simEsmReadyQueueHdr);
    NFX_SPIN_INIT(&fs_simEsmInstLock,SPINLOCK_LEVEL_UNKNOWN);

    fs_efsSscAppId = FS_SIM_MSG_ADDR_SSC;
    fs_mcpuAppId = FS_SIM_MSG_ADDR_MCPU;

    NFX_SPIN_INIT(&fs_errStatsSpinId,SPINLOCK_LEVEL_UNKNOWN);
    NFX_SPIN_INIT(&fs_ioSpin,SPINLOCK_LEVEL_UNKNOWN);

    fs_simRunServer();

    return 0;
}

/*-----------------------------------------------------------------------------
 * fs_simSendMsg
 *
 * Description:  Send a message.
 * 
 * Created By:  Bob Fozard
 *
 * Date Created:  3/20/2001
 */
int32
fs_simSendMsg(
    int32       src,
    int32       dest,
    char        *msg,
    int32       msgLen)
{
    fs_MsgReqHdr            *reqHdr;
    fs_SimMsgQueueEntry     *qEntry;
    int32                   rc;

    //printf("Sending msg\n");

    reqHdr = (fs_MsgReqHdr *) msg;
    reqHdr->src = src;
    reqHdr->dest = dest;
    reqHdr->len = msgLen;

    qEntry = fs_ramAlloc(FS_RAM_SIM_MSG_QENTRY, sizeof(fs_SimMsgQueueEntry));
    qEntry->msg = msg;
	DLL_ISOLATE_ELEM(qEntry, dllEntry);

    switch(dest)
    {
    case FS_SIM_MSG_ADDR_SSC:
        //printf("  send to ssc\n");
        rc = pthread_mutex_lock(&(fs_simSscExec.mutex));
        DLL_INSERT_TAIL(&(fs_simSscRxQueue.dllHdr), qEntry, dllEntry);
        rc |= pthread_cond_signal(&(fs_simSscExec.condVar));
        rc |= pthread_mutex_unlock(&(fs_simSscExec.mutex));
        break;

    case FS_SIM_MSG_ADDR_FP_0:
        //printf("  send to fp 0\n");
        rc = pthread_mutex_lock(&fs_simFpExecMutex);
        DLL_INSERT_TAIL(&(fs_simFpRxQueue[0].dllHdr), qEntry, dllEntry);
        rc |= pthread_cond_signal(&(fs_simFpRxQueue[0].condVar));
        rc |= pthread_mutex_unlock(&fs_simFpExecMutex);
        break;

    case FS_SIM_MSG_ADDR_FP_1:
        //printf("  send to fp 1\n");
        rc = pthread_mutex_lock(&fs_simFpExecMutex);
        DLL_INSERT_TAIL(&(fs_simFpRxQueue[1].dllHdr), qEntry, dllEntry);
        rc |= pthread_cond_signal(&(fs_simFpRxQueue[1].condVar));
        rc |= pthread_mutex_unlock(&fs_simFpExecMutex);
        break;

    case FS_SIM_MSG_ADDR_MCPU:
        //printf("  send to Mcpu\n");
        rc = pthread_mutex_lock(&(fs_simMcpuQueue.mutex));
        DLL_INSERT_TAIL(&(fs_simMcpuQueue.dllHdr), qEntry, dllEntry);
        rc |= pthread_mutex_unlock(&(fs_simMcpuQueue.mutex));
        rc |= pthread_mutex_lock(&(fs_simMcpuExec.mutex));
        rc |= pthread_cond_signal(&(fs_simMcpuExec.condVar));
        rc |= pthread_mutex_unlock(&(fs_simMcpuExec.mutex));
        break;

    case FS_SIM_MSG_ADDR_FSSF:
        //printf("  send to fssf\n");
        rc = pthread_mutex_lock(&(fs_simFssfQueue.mutex));
        DLL_INSERT_TAIL(&(fs_simFssfQueue.dllHdr), qEntry, dllEntry);
        rc |= pthread_cond_signal(&(fs_simFssfQueue.condVar));
        rc |= pthread_mutex_unlock(&(fs_simFssfQueue.mutex));
        break;

    default:
        if (dest >= FS_SIM_MSG_ADDR_NFS &&
            dest < FS_SIM_MSG_ADDR_NFS + FS_SIM_NFS_THREAD_MAX)
        {
            int32 nfsThread = dest - FS_SIM_MSG_ADDR_NFS;

            //printf("  send to nfs thread %d\n", nfsThread);

            rc = pthread_mutex_lock(
                        &(fs_simNfsThread[nfsThread].rxQueue.mutex));
            DLL_INSERT_TAIL(&(fs_simNfsThread[nfsThread].rxQueue.dllHdr),
                                qEntry, dllEntry);
            rc |= pthread_cond_signal(&(fs_simNfsThread[nfsThread].condVar));
            rc |= pthread_mutex_unlock(
                        &(fs_simNfsThread[nfsThread].rxQueue.mutex));
            break;
        }
        else
        {
            printf("fs_simSendMsg: bad dest %d\n", dest);
            FS_ASSERT(FALSE);
            return NFX_ERR;
        }
    }

    if (rc)
    {
        printf("fs_simSendMsg:  pthread err\n");
        return NFX_ERR;
    }

    //printf("sent okay\n");

    return NFX_OK;

} /* fs_simSendMsg */

int32
eee_ramCopy(void *dst, void *src, uint32 numBytes)
{
    memcpy(dst, src, numBytes);
    return NFX_OK;
}

/*-----------------------------------------------------------------------------
 * NFS <--> FS message translation functions.
 */
void
fs_simPrintNfsSattr(
    sattr       *attr)
{
    printf("nfsSattr:\n");
    printf("  mode: %o\n", attr->mode);
    printf("  uid: %d\n", attr->uid);
    printf("  gid: %d\n", attr->gid);
    printf("  size: %d\n", attr->size);

} /* fs_simPrintNfsSattr */

void
fs_simPrintNfsFattr(
    fattr       *attr)
{
    printf("nfsFattr:\n");
    printf("  type: %d\n", attr->type);
    printf("  mode: %o\n", attr->mode);
    printf("  nlink: %d\n", attr->nlink);
    printf("  uid: %d\n", attr->uid);
    printf("  gid: %d\n", attr->gid);
    printf("  size: %d\n", attr->size);
    printf("  blksz: %d\n", attr->blocksize);
    printf("  fsid: %d\n", attr->fsid);
    printf("  ino: %d\n", attr->fileid);

} /* fs_simPrintNfsFattr */

void
fs_simDoGetAttr(
    fattr       *attr,
    fs_Attr     *fsattr)
{
    //printf("fs_simDoGetAttr from: \n");
    //fs_printAttr(fsattr);

    attr->type = fsattr->type;
    attr->mode = fsattr->mode;
    attr->nlink = fsattr->numLinks;
    attr->uid = fsattr->uid;
    attr->gid = fsattr->gid;
    attr->size = fsattr->size;
    attr->blocksize = fsattr->blockSize;
    attr->rdev = fsattr->rdev;
    attr->blocks = fsattr->blocks;
    attr->fsid = fsattr->fsId;
    attr->fileid = fsattr->inode;
    attr->atime.seconds = fsattr->atime.secs;
    attr->atime.useconds = fsattr->atime.nsecs/1000;
    attr->mtime.seconds = fsattr->mtime.secs;
    attr->mtime.useconds = fsattr->mtime.nsecs/1000;
    attr->ctime.seconds = fsattr->ctime.secs;
    attr->ctime.useconds = fsattr->ctime.nsecs/1000;

} /* fs_simDoGetAttr */

// copy from uint64 into nfs_uint64_t
#define fs_simSetNfs_uint64(Dest, Src)  Dest._p._u = ((Src >> 32) & 0xffffffff) ;   \
                                        Dest._p._l = (Src & 0xffffffff);

// copy from nfs_uint64_t into uint64
#define fs_simGetNfs_uint64(Dest, Src)  Dest = (((uint64)Src._p._u) << 32) + (Src._p._l) ;

// copy from uint64 into nfs_uint64_t
//#define fs_simSetNfs_uint64(Dest, Src)  Dest._u64 = Src ;

// copy from nfs_uint64_t into uint64
//#define fs_simGetNfs_uint64(Dest, Src)  Dest = Src._u64 ; 

 

void
fs_simDoGetAttr3(
    fattr3      *attr,
    fs_Attr     *fsattr)
{
    //printf("fs_simDoGetAttr3 from: \n");
    //fs_printAttr(fsattr);

    attr->type = fsattr->type;
    attr->mode = fsattr->mode;
    attr->nlink = fsattr->numLinks;
    attr->uid = fsattr->uid;
    attr->gid = fsattr->gid;
    fs_simSetNfs_uint64 (attr->size, fsattr->size);
    fs_simSetNfs_uint64 (attr->used, ((uint64)fsattr->blocks * FS_BLOCK_SIZE));
    attr->rdev.specdata1 = fsattr->devNum.major;
    attr->rdev.specdata2 = fsattr->devNum.minor;
    fs_simSetNfs_uint64 (attr->fsid, fsattr->fsId);
    fs_simSetNfs_uint64 (attr->fileid, fsattr->inode);
    attr->atime.seconds = fsattr->atime.secs;
    attr->atime.nseconds = fsattr->atime.nsecs;
    attr->mtime.seconds = fsattr->mtime.secs;
    attr->mtime.nseconds = fsattr->mtime.nsecs;
    attr->ctime.seconds = fsattr->ctime.secs;
    attr->ctime.nseconds = fsattr->ctime.nsecs;

} /* fs_simDoGetAttr3 */

void
fs_simDoGetWccData3(
    wcc_data    *dir_wcc,
    fs_Attr     *fsattrBefore,
    fs_Attr     *fsattrAfter)
{
    dir_wcc->before.attributes = TRUE;
    fs_simSetNfs_uint64 (dir_wcc->before.attr.size, fsattrBefore->size);
    dir_wcc->before.attr.mtime.seconds = fsattrBefore->mtime.secs;
    dir_wcc->before.attr.mtime.nseconds = fsattrBefore->mtime.nsecs;
    dir_wcc->before.attr.ctime.seconds = fsattrBefore->ctime.secs;
    dir_wcc->before.attr.ctime.nseconds = fsattrBefore->ctime.nsecs;
    dir_wcc->after.attributes = TRUE;
    fs_simDoGetAttr3(&(dir_wcc->after.attr), fsattrAfter);
}

void
fs_simDoSetAttr(
    fs_Attr     *fsattr,
    sattr       *attr)
{
    //printf("fs_simDoSetAttr\n");
    //fs_simPrintNfsSattr(attr);

    if (attr->mode == -1)
    {
        fsattr->mode = -1;
    }
    else
    {
        fsattr->mode = attr->mode;
    }
    if (attr->uid == -1)
    {
        fsattr->uid = -1;
    }
    else
    {
        fsattr->uid = attr->uid;
    }
    if (attr->gid == -1)
    {
        fsattr->gid = -1;
    }
    else
    {
        fsattr->gid = attr->gid;
    }
    if (attr->size == -1)
    {
        fsattr->size = -1;
    }
    else
    {
        fsattr->size = attr->size;
    }
    if (attr->atime.seconds == -1)
    {
        fsattr->atime.secs = -1;
        fsattr->atimeType = FS_ATTR_TIME_SET_DONT;
    }
    else
    {
        fsattr->atime.secs = attr->atime.seconds;
        fsattr->atimeType = FS_ATTR_TIME_SET_CLIENT;
    }
    if (attr->atime.useconds == -1)
    {
        fsattr->atime.nsecs = -1;
    }
    else
    {
        fsattr->atime.nsecs = attr->atime.useconds;
    }
    if (attr->mtime.seconds == -1)
    {
        fsattr->mtime.secs = -1;
        fsattr->mtimeType = FS_ATTR_TIME_SET_DONT;
    }
    else
    {
        fsattr->mtime.secs = attr->mtime.seconds;
        fsattr->mtimeType = FS_ATTR_TIME_SET_CLIENT;
    }
    if (attr->mtime.useconds == -1)
    {
        fsattr->mtime.nsecs = -1;
    }
    else
    {
        fsattr->mtime.nsecs = attr->mtime.useconds;
    }

} /* fs_simDoSetAttr */

void
fs_simDoSetAttr3(
    fs_Attr     *fsattr,
    sattr3      *attr)
{
    //printf("fs_simDoSetAttr\n");
    //fs_simPrintNfsSattr(attr);

    if (attr->mode.set_it)
    {
        fsattr->mode = attr->mode.mode;
    }
    else
    {
        fsattr->mode = -1;
    }
    if (attr->uid.set_it)
    {
        fsattr->uid = attr->uid.uid;
    }
    else
    {
        fsattr->uid = -1;
    }
    if (attr->gid.set_it)
    {
        fsattr->gid = attr->gid.gid;
    }
    else
    {
        fsattr->gid = -1;
    }
    if (attr->size.set_it)
    {
        fs_simGetNfs_uint64 (fsattr->size, attr->size.size);
    }
    else
    {
        fsattr->size = -1;
    }

    if (attr->atime.set_it == SET_TO_CLIENT_TIME)
    {
        fsattr->atime.secs = attr->atime.atime.seconds;
        fsattr->atime.nsecs = attr->atime.atime.nseconds;
        fsattr->atimeType = FS_ATTR_TIME_SET_CLIENT;
    }
    else if (attr->atime.set_it == SET_TO_SERVER_TIME)
    {
        fsattr->atime.secs = -1;
        fsattr->atime.nsecs = -1;
        fsattr->atimeType = FS_ATTR_TIME_SET_SRVR;
    }
    else
    {
        fsattr->atime.secs = -1;
        fsattr->atime.nsecs = -1;
        fsattr->atimeType = FS_ATTR_TIME_SET_DONT;
    }

    if (attr->mtime.set_it == SET_TO_CLIENT_TIME)
    {
        fsattr->mtime.secs = attr->mtime.mtime.seconds;
        fsattr->mtime.nsecs = attr->mtime.mtime.nseconds;
        fsattr->mtimeType = FS_ATTR_TIME_SET_CLIENT;
    }
    else if (attr->mtime.set_it == SET_TO_SERVER_TIME)
    {
        fsattr->mtime.secs = -1;
        fsattr->mtime.nsecs = -1;
        fsattr->mtimeType = FS_ATTR_TIME_SET_SRVR;
    }
    else
    {
        fsattr->mtime.secs = -1;
        fsattr->mtime.nsecs = -1;
        fsattr->mtimeType = FS_ATTR_TIME_SET_DONT;
    }

} /* fs_simDoSetAttr3 */

int32
fs_simNullReq2(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    bzero(buf, sizeof(fs_MsgNullReq));
    *len = sizeof(fs_MsgNullReq);
    return NFX_OK;

} /* fs_simNullReq2 */

void *
fs_simNullResp2(
    void    *fsResp)
{
    return NULL;
} /* fs_simNullResp2 */

int32
fs_simGetAttrReq2(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgGetAttrReq    *fsMsg;
    nfs_fh              *fh;

    //printf("nfs getattr req rcvd\n");

    fh = (nfs_fh *) arg;

    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgGetAttrReq));
    fsMsg->hdr.requestType = FS_REQ_GET_ATTR;
    fsMsg->hdr.attrType = FS_POST_ATTRS;
    memcpy(&(fsMsg->fileHandle), fh, sizeof(fs_Handle));

    *len = sizeof(fs_MsgGetAttrReq);
    return NFX_OK;

} /* fs_simGetAttrReq2 */

void *
fs_simGetAttrResp2(
    void    *fsResp)
{
    fs_MsgGetAttrResp   *fs;
    attrstat            *as;

    //printf("nfs getattr resp called\n");

    fs = (fs_MsgGetAttrResp *) fsResp;
    as = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(attrstat));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        as->status = NFS_OK;
        fs_simDoGetAttr(&(as->attrstat_u.attributes), &(fs->attr));
    }
    else
    {
        as->status = fs->hdr.fsStatus;
    }
    
    return as;

} /* fs_simGetAttrResp2 */

int32
fs_simSetAttrReq2(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgSetAttrReq    *fsMsg;
    sattrargs           *sargs;

    //printf("nfs setattr req rcvd\n");

    sargs = (sattrargs *) arg;

    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgSetAttrReq));
    fsMsg->hdr.requestType = FS_REQ_SET_ATTR;
    fsMsg->hdr.attrType = FS_POST_ATTRS;
    memcpy(&(fsMsg->fileHandle), &(sargs->file), sizeof(fs_Handle));

    fs_simDoSetAttr(&(fsMsg->attr), &(sargs->attributes));

    *len = sizeof(fs_MsgSetAttrReq);
    return NFX_OK;

} /* fs_simSetAttrReq2 */

void *
fs_simSetAttrResp2(
    void    *fsResp)
{
    fs_MsgSetAttrResp   *fs;
    attrstat            *as;

    //printf("nfs setattr resp called\n");

    fs = (fs_MsgSetAttrResp *) fsResp;

    as = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(attrstat));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        as->status = NFS_OK;
        fs_simDoGetAttr(&(as->attrstat_u.attributes), &(fs->attr[0]));
    }
    else
    {
        as->status = fs->hdr.fsStatus;
    }

    return as;

} /* fs_simSetAttrResp2 */

int32
fs_simGetRootReq2(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    printf("nfs getroot req rcvd\n");
    *len = 0;
    return NFX_ERR;

} /* fs_simGetRootReq2 */

void *
fs_simGetRootResp2(
    void    *fsResp)
{
    return NULL;
} /* fs_simGetRootResp2 */

int32
fs_simLookupReq2(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgLookupReq     *fsMsg;
    diropargs           *diroparg;
    int32               nameLen;

    //printf("nfs lookup req rcvd\n");
    diroparg = (diropargs *) arg;

    nameLen = MIN(FS_NAME_MAX, strlen(diroparg->name));
    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgLookupReq));
    fsMsg->hdr.requestType = FS_REQ_LOOKUP;
    fsMsg->hdr.attrType = FS_POST_ATTRS;
    memcpy(&(fsMsg->parentHandle), &(diroparg->dir), sizeof(fs_Handle));
    memcpy(&(fsMsg->fileName), diroparg->name, nameLen);
    fsMsg->fileName[nameLen] = '\0';
    fsMsg->nameLen = nameLen;

    *len = sizeof(fs_MsgLookupReq) + nameLen + 1;
    return NFX_OK;

} /* fs_simLookupReq2 */

void *
fs_simLookupResp2(
    void    *fsResp)
{
    fs_MsgLookupResp    *fs;
    diropres            *resp;

    //printf("nfs lookup resp called\n");

    fs = (fs_MsgLookupResp *) fsResp;

    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(diropres));
    bzero (resp, sizeof(diropres));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;
        fs_simDoGetAttr(&(resp->diropres_u.diropres.attributes),
                        &(fs->attr[0]));
        memcpy( &(resp->diropres_u.diropres.file),
                &(fs->fileHandle),
                sizeof(fs_Handle));
    }
    else
    {
        resp->status = fs->hdr.fsStatus;
    }
    
    return resp;

} /* fs_simLookupResp2 */

int32
fs_simReadLinkReq2(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgReadLinkReq   *fsMsg;
    nfs_fh              *fh;

    fh = (nfs_fh *) arg;

    //printf("nfs readlink req rcvd\n");
    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgReadLinkReq));
    fsMsg->hdr.requestType = FS_REQ_READLINK;
    memcpy(&(fsMsg->fileHandle), fh, sizeof(fs_Handle));

    *len = sizeof(fs_MsgReadLinkReq);
    return NFX_OK;

} /* fs_simReadLinkReq2 */

void *
fs_simReadLinkResp2(
    void    *fsResp)
{
    fs_MsgReadLinkResp  *fs;
    readlinkres         *resp;

    //printf("nfs create resp called\n");

    fs = (fs_MsgReadLinkResp *) fsResp;

    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(readlinkres));
    bzero(resp, sizeof *resp);
    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;
        resp->readlinkres_u.data = fs_ramAlloc(FS_RAM_NFS_RESP_ARG,
                                                fs->nameLen + 1);
        memcpy( resp->readlinkres_u.data,
                fs->linkName,
                fs->nameLen);
        resp->readlinkres_u.data[fs->nameLen] = '\0';
    }
    else
    {
        resp->status = fs->hdr.fsStatus;
        resp->readlinkres_u.data = NULL;
    }
    
    return resp;

} /* fs_simReadLinkResp2 */

void
fs_simFreeReadLinkResp2(
    void    *resp)
{
    readlinkres         *rl = resp;

    if (rl->readlinkres_u.data != NULL)
    {
        fs_ramDealloc(FS_RAM_NFS_RESP_ARG, rl->readlinkres_u.data);
    }
        
    fs_ramDealloc(FS_RAM_NFS_RESP_ARG, rl);

} /* fs_simFreeReadLinkResp2 */

int32
fs_simReadReq2(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    readargs *rargs = (readargs *) arg;
    fs_MsgReadReq *msg ; 
    msg = buf;

    // printf("nfs read req rcvd\n");

    bzero(msg, sizeof (fs_MsgReadReq));          // TODO - remove
    memcpy(&msg->fileHandle, &rargs->file, sizeof (fs_Handle));
    msg->hdr.requestType = FS_REQ_READ;
    msg->hdr.attrType    = FS_POST_ATTRS;         // TODO - ???
    msg->fileOffset      = rargs->offset;
    msg->readLength      = rargs->count;
    // msg->readLength      = rargs->totalcount;
    
    *len = sizeof(fs_MsgReadReq);
    return NFX_OK;
} /* fs_simReadReq2 */

void *
fs_simReadResp2(
    void    *fsResp)
{
    fs_MsgReadResp      *fs;
    readres             *rs;

    //printf("nfs Read resp called\n");

    fs = (fs_MsgReadResp *) fsResp;
    rs = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(readres));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        rs->status = NFS_OK;
        fs_simDoGetAttr(&(rs->readres_u.reply.attributes), &(fs->attr[0]));
        rs->readres_u.reply.data.data_len = fs->bytesRead;
        if (fs->bytesRead)
          {
            rs->readres_u.reply.data.data_val = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, fs->bytesRead);
            memcpy(rs->readres_u.reply.data.data_val, 
                   &(((char*)fs)[fs->data_index]),
                   fs->bytesRead);
          }
        else
          {
            rs->readres_u.reply.data.data_val = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, 1);
            // dummy allocation for the moment, 
            // Todo: need changes in
            // fs_simFreeReadResp2, EFENCE and perhaps xdr_readres (xdr_bytes)
          }
    }
    else
    {
        rs->status = fs->hdr.fsStatus;
    }
    
    return rs;
} /* fs_simReadResp2 */

void
fs_simFreeReadResp2(
    void    *resp)
{
    readres             *rs;

    //printf("nfs Free Read resp called\n");

    rs = (readres *) resp;
    if (rs->status == NFS_OK)
      fs_ramDealloc(FS_RAM_NFS_RESP_ARG, rs->readres_u.reply.data.data_val);
    fs_ramDealloc(FS_RAM_NFS_RESP_ARG, rs);

} /* fs_simFreeReadResp2 */

int32
fs_simBadReq2(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    printf("nfs bad req rcvd\n");
    *len = 0;
    return NFX_ERR;

} /* fs_simBadReq2 */

void *
fs_simBadResp2(
    void    *fsResp)
{
    return NULL;
} /* fs_simBadResp2 */

nfx_cod_t
fs_sim_get_new_cod(int id)
{
#define START_COD  0x123456789ABCULL
    static nfx_cod_t curr_cod = START_COD ;   // some start value
    static nfx_spinlock_t get_cod_spinlock;

    nfx_cod_t COD;
    unsigned long long  slot;
    unsigned long long  cpu;

    if (curr_cod == START_COD)
    {
        // to make it unique after restart
        // good enough for now
        struct timeval  now;
        gettimeofday(&now, NULL);
        curr_cod += ((unsigned long long)now.tv_sec << 32) & 0x00ffffffffffffffULL ;
        NFX_SPIN_INIT(&get_cod_spinlock,SPINLOCK_LEVEL_UNKNOWN);
    }

    NFX_SPIN_LOCK(&get_cod_spinlock);
    slot = (unsigned long long)(slot_id & 0x0f) << 60 ;
    cpu = (unsigned long long)(id & 0x0f) << 56 ;
    COD = curr_cod + slot + cpu ;
    curr_cod =  (curr_cod + 1) & 0x00ffffffffffffffULL ;
    NFX_SPIN_UNLOCK(&get_cod_spinlock);

    return COD;
}

int32
fs_simWriteReq2(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    uint32 loop, bytesThis, length = 0;
    writeargs *wargs = (writeargs *) arg;
    fs_MsgWriteReq *msg = buf;

    // printf("nfs write req rcvd\n");

    bzero(msg, sizeof (fs_MsgWriteReq));          // TODO - remove
    memcpy(&msg->fileHandle, &wargs->file, sizeof (fs_Handle));
    msg->hdr.requestType = FS_REQ_WRITE;
    msg->hdr.attrType    = FS_POST_ATTRS;         // TODO - ???
    msg->fileOffset      = wargs->offset;
    msg->writeLength     = wargs->data.data_len;
    msg->nbrUsedCod      = (msg->writeLength - 1) / FS_BLOCK_SIZE + 1;
    for (loop = 0 ; loop < msg->nbrUsedCod; loop++)
    {
        bytesThis = MIN (msg->writeLength - length, FS_BLOCK_SIZE);
        msg->cod[loop].cod = fs_sim_get_new_cod(FS_SIM_MSG_ADDR_NFS);
        msg->cod[loop].offset = 0;
        msg->cod[loop].length = bytesThis;
        length += bytesThis;
    }
    FS_ASSERT (length == msg->writeLength);
    
    *len = sizeof(fs_MsgWriteReq);
    return NFX_OK;

} /* fs_simWriteReq2 */

void *
fs_simWriteResp2(
    void    *fsResp)
{
    fs_MsgWriteResp     *fs;
    attrstat            *as;

    //printf("nfs Write resp called\n");

    fs = (fs_MsgWriteResp *) fsResp;
    as = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(attrstat));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        as->status = NFS_OK;
        fs_simDoGetAttr(&(as->attrstat_u.attributes), &(fs->attr[0]));
    }
    else
    {
        as->status = fs->hdr.fsStatus;
    }
    
    return as;

} /* fs_simWriteResp2 */

int32
fs_simCreateReq2(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgCreateReq     *fsMsg;
    createargs          *cargs;
    int32               nameLen;

    //printf("nfs create req rcvd\n");
    cargs = (createargs *) arg;

    nameLen = MIN(FS_NAME_MAX, strlen(cargs->where.name));
    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgCreateReq));
    fsMsg->hdr.requestType = FS_REQ_CREATE;
    fsMsg->hdr.attrType = FS_POST_ATTRS;
    fs_simDoSetAttr(&(fsMsg->u.attr), &(cargs->attributes));
    memcpy(&(fsMsg->parentHandle), &(cargs->where.dir), sizeof(fs_Handle));
    memcpy(&(fsMsg->fileName), cargs->where.name, nameLen);
    fsMsg->fileName[nameLen] = '\0';
    fsMsg->nameLen = nameLen;

    *len = sizeof(fs_MsgCreateReq) + nameLen + 1;
    return NFX_OK;

} /* fs_simCreatReq2 */

void *
fs_simCreateResp2(
    void    *fsResp)
{
    fs_MsgCreateResp    *fs;
    diropres            *resp;

    //printf("nfs create resp called\n");

    fs = (fs_MsgCreateResp *) fsResp;

    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(diropres));
    bzero (resp, sizeof(diropres));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;
        fs_simDoGetAttr(&(resp->diropres_u.diropres.attributes),
                        &(fs->attr[0]));
        memcpy( &(resp->diropres_u.diropres.file),
                &(fs->fileHandle),
                sizeof(fs_Handle));
    }
    else
    {
        resp->status = fs->hdr.fsStatus;
    }
    
    return resp;

} /* fs_simCreatResp2 */

int32
fs_simRemoveReq2(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgRemoveReq     *fsMsg;
    diropargs           *diroparg;
    int32               nameLen;

    //printf("nfs remove req rcvd\n");
    diroparg = (diropargs *) arg;

    nameLen = MIN(FS_NAME_MAX, strlen(diroparg->name));
    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgRemoveReq));
    fsMsg->hdr.requestType = FS_REQ_REMOVE;
    memcpy(&(fsMsg->parentHandle), &(diroparg->dir), sizeof(fs_Handle));
    memcpy(&(fsMsg->name), diroparg->name, nameLen);
    fsMsg->name[nameLen] = '\0';
    fsMsg->nameLen = nameLen;

    *len = sizeof(fs_MsgRemoveReq) + nameLen + 1;
    return NFX_OK;

} /* fs_simRemoveReq2 */

void *
fs_simRemoveResp2(
    void    *fsResp)
{
    fs_MsgRemoveResp    *fs;
    nfsstat             *resp;

    //printf("nfs remove resp called\n");

    fs = (fs_MsgRemoveResp *) fsResp;

    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(nfsstat));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        *resp = NFS_OK;
    }
    else
    {
        *resp = fs->hdr.fsStatus;
    }
    
    return resp;

} /* fs_simRemoveResp2 */

int32
fs_simRenameReq2(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgRenameReq     *fsMsg;
    renameargs          *ren;
    int32               fromNameLen, toNameLen;
    char                *toPtr;

    //printf("nfs rename req rcvd\n");
    ren = (renameargs *) arg;

    fromNameLen = MIN(FS_NAME_MAX, strlen(ren->from.name));
    toNameLen = MIN(FS_NAME_MAX, strlen(ren->to.name));
    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgRenameReq));
    fsMsg->hdr.requestType = FS_REQ_RENAME;
    memcpy(&(fsMsg->fromDirHandle), &(ren->from.dir), sizeof(fs_Handle));
    memcpy(&(fsMsg->toDirHandle), &(ren->to.dir), sizeof(fs_Handle));
    memcpy(&(fsMsg->fromName), ren->from.name, fromNameLen);
    fsMsg->fromName[fromNameLen] = '\0';
    fsMsg->fromNameLen = fromNameLen;
    toPtr = &fsMsg->fromName[fromNameLen + 1];
    memcpy(toPtr, ren->to.name, toNameLen);
    toPtr[toNameLen] = '\0';
    fsMsg->toNameLen = toNameLen;

    *len = sizeof(fs_MsgRenameReq) + fromNameLen + toNameLen + 2;
    return NFX_OK;

} /* fs_simRenameReq2 */

void *
fs_simRenameResp2(
    void    *fsResp)
{
    fs_MsgRenameResp    *fs;
    nfsstat             *resp;

    //printf("nfs rename resp called\n");

    fs = (fs_MsgRenameResp *) fsResp;

    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(nfsstat));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        *resp = NFS_OK;
    }
    else
    {
        *resp = fs->hdr.fsStatus;
    }
    
    return resp;

} /* fs_simRenameResp2 */

int32
fs_simLinkReq2(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    linkargs            *largs;
    fs_MsgHardLinkReq   *fsMsg;
    int32               nameLen;

    //printf("nfs link req rcvd\n");
    largs = (linkargs *) arg;

    nameLen = MIN(FS_NAME_MAX, strlen(largs->to.name));
    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgHardLinkReq));
    fsMsg->hdr.requestType = FS_REQ_HARDLINK;
    memcpy(&(fsMsg->fileHandle), &(largs->from), sizeof(fs_Handle));
    memcpy(&(fsMsg->parentDirHandle), &(largs->to.dir), sizeof(fs_Handle));
    memcpy(&(fsMsg->linkName), largs->to.name, nameLen);
    fsMsg->linkName[nameLen] = '\0';
    fsMsg->nameLen = nameLen;

    *len = sizeof(fs_MsgHardLinkReq) + nameLen + 1;

    return NFX_OK;

} /* fs_simLinkReq2 */

void *
fs_simLinkResp2(
    void    *fsResp)
{
    fs_MsgHardLinkResp  *fs;
    static uint32       resp;

    //printf("nfs link resp called\n");

    fs = (fs_MsgHardLinkResp *) fsResp;

    resp = fs->hdr.fsStatus;
    
    return &resp;

} /* fs_simLinkResp2 */

int32
fs_simSymlinkReq2(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    symlinkargs         *slargs;
    fs_MsgSymLinkReq    *fsMsg;
    int32               linkLen, linkToLen;
    char                *msgPtr;

    slargs = (symlinkargs *) arg;

    if (slargs->from.name == (void *) -1 ||
        slargs->to == (void *) -1)
    {
        printf("nfs v2 symlink bad name(s)\n");
        return NFX_ERR;
    }

    //printf("nfs symlink req rcvd linkName %s linkToName %s\n",
    //      slargs->from.name, slargs->to);

    linkLen = MIN(FS_NAME_MAX, strlen(slargs->from.name));
    linkToLen = MIN(FS_PATH_MAX, strlen(slargs->to));

    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgSymLinkReq));
    fsMsg->hdr.requestType = FS_REQ_SYMLINK;
    memcpy(&(fsMsg->parentHandle), &(slargs->from.dir), sizeof(fs_Handle));
    msgPtr = fsMsg->linkName;
    memcpy(msgPtr, slargs->from.name, linkLen);
    msgPtr[linkLen] = '\0';
    fsMsg->linkNameLen = linkLen;
    msgPtr += linkLen + 1;
    memcpy(msgPtr, slargs->to, linkToLen);
    msgPtr[linkToLen] = '\0';
    fsMsg->linkToNameLen = linkToLen;

    *len = sizeof(fs_MsgSymLinkReq) + linkLen + linkToLen + 2;

    return NFX_OK;

} /* fs_simSymlinkReq2 */

void *
fs_simSymlinkResp2(
    void    *fsResp)
{
    fs_MsgSymLinkResp   *fs;
    static uint32       resp;

    //printf("nfs link resp called\n");

    fs = (fs_MsgSymLinkResp *) fsResp;

    resp = fs->hdr.fsStatus;
    
    return &resp;

} /* fs_simSymlinkResp2 */

int32
fs_simMkdirReq2(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgMkdirReq      *fsMsg;
    createargs          *cargs;
    int32               nameLen;

    //printf("nfs mkdir req rcvd\n");
    cargs = (createargs *) arg;

    nameLen = MIN(FS_NAME_MAX, strlen(cargs->where.name));
    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgMkdirReq));
    fsMsg->hdr.requestType = FS_REQ_MKDIR;
    fsMsg->hdr.attrType = FS_POST_ATTRS;
    fs_simDoSetAttr(&(fsMsg->dirAttr), &(cargs->attributes));
    memcpy(&(fsMsg->parentHandle), &(cargs->where.dir), sizeof(fs_Handle));
    memcpy(&(fsMsg->dirName), cargs->where.name, nameLen);
    fsMsg->dirName[nameLen] = '\0';
    fsMsg->nameLen = nameLen;

    *len = sizeof(fs_MsgMkdirReq) + nameLen + 1;
    return NFX_OK;

} /* fs_simMkdirReq2 */

void *
fs_simMkdirResp2(
    void    *fsResp)
{
    fs_MsgMkdirResp     *fs;
    diropres            *resp;

    //printf("nfs mkdir resp called\n");

    fs = (fs_MsgMkdirResp *) fsResp;

    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(diropres));
    bzero (resp, sizeof(diropres));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;
        fs_simDoGetAttr(&(resp->diropres_u.diropres.attributes),
                        &(fs->attr[0]));
        memcpy( &(resp->diropres_u.diropres.file),
                &(fs->dirHandle),
                sizeof(fs_Handle));
    }
    else
    {
        resp->status = fs->hdr.fsStatus;
    }
    
    return resp;

} /* fs_simMkdirResp2 */

int32
fs_simRmdirReq2(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgRmdirReq      *fsMsg;
    diropargs           *diroparg;
    int32               nameLen;

    //printf("nfs rmdir req rcvd\n");
    diroparg = (diropargs *) arg;

    nameLen = MIN(FS_NAME_MAX, strlen(diroparg->name));
    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgRmdirReq));
    fsMsg->hdr.requestType = FS_REQ_RMDIR;
    memcpy(&(fsMsg->parentHandle), &(diroparg->dir), sizeof(fs_Handle));
    memcpy(&(fsMsg->name), diroparg->name, nameLen);
    fsMsg->name[nameLen] = '\0';
    fsMsg->nameLen = nameLen;

    *len = sizeof(fs_MsgRmdirReq) + nameLen + 1;
    return NFX_OK;

} /* fs_simRmdirReq2 */

void *
fs_simRmdirResp2(
    void    *fsResp)
{
    fs_MsgRmdirResp     *fs;
    nfsstat             *resp;

    //printf("nfs rmdir resp called\n");

    fs = (fs_MsgRmdirResp *) fsResp;

    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(nfsstat));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        *resp = NFS_OK;
    }
    else
    {
        *resp = fs->hdr.fsStatus;
    }
    
    return resp;

} /* fs_simRmdirResp2 */

int32
fs_simReaddirReq2(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgReadDirReq    *fsMsg;
    readdirargs         *rargs;
    nfs_fh              *fh;

    //printf("nfs readdir req rcvd\n");

    rargs = (readdirargs *) arg;
    fh = &(rargs->dir);

    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgReadDirReq));
    fsMsg->hdr.requestType = FS_REQ_READDIR;
    memcpy(&(fsMsg->fileHandle), fh, sizeof(fs_Handle));
    fsMsg->cookie = rargs->cookie;
    fsMsg->maxBytes = rargs->count;

    *len = sizeof(fs_MsgReadDirReq);
    return NFX_OK;

} /* fs_simReaddirReq2 */

void *
fs_simReaddirResp2(
    void    *fsResp)
{
    fs_MsgReadDirResp       *fs;
    readdirres              *rdresp;
    entry                   *ent, *entprev;
    int32                   i;
    fs_DirEntry             *dirent;

    fs = (fs_MsgReadDirResp *) fsResp;

    //printf("nfs readdir resp called, %d, %d\n", fs->numEntriesThis, fs->eof);

    rdresp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(readdirres));
    bzero(rdresp, sizeof *rdresp);
    i = 0;
        
    if (fs->hdr.fsStatus == NFX_OK)
    {
        rdresp->status = NFS_OK;
        // Walk the fs response, allocate an ent for
        // each and link it to the response.
        dirent = &(fs->dirEntry[0]);
        rdresp->readdirres_u.reply.entries = NULL;
        entprev = NULL;
        for (i = 0; i < fs->numEntriesThis; i += 1)
        {
            if (dirent->nameLen)
            {
                ent = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(entry));
                ent->fileid = dirent->inode;
                ent->name = dirent->name;
                ent->cookie = dirent->cookie;
                if (i == 0)
                {
                    rdresp->readdirres_u.reply.entries = ent;
                }
                else
                {
                    entprev->nextentry = ent;
                }
                ent->nextentry = NULL;
                entprev = ent;
            }
            else
            {
                // early end of entries???
                break;
            }

            FS_READDIR_ADV_ENTRY_PTR(dirent);
        }
    }
    else
    {
        rdresp->status = fs->hdr.fsStatus;
    }

    rdresp->readdirres_u.reply.eof = fs->eof;

    //printf("  done readdir resp\n");
    
    return rdresp;

} /* fs_simReaddirResp2 */

void
fs_simFreeReaddirResp2(
    void    *resp)
{
    readdirres  *rdresp;
    entry       *ent, *nextEnt;

    //printf("free readdir resp2\n");

    rdresp = (readdirres *) resp;

    ent = rdresp->readdirres_u.reply.entries;
    while (ent != NULL)
    {
        nextEnt = ent->nextentry;
        fs_ramDealloc(FS_RAM_NFS_RESP_ARG, ent);
        ent = nextEnt;
    }

    fs_ramDealloc(FS_RAM_NFS_RESP_ARG, rdresp);

    //printf("  done free readdir resp2\n");

} /* fs_simFreeReaddirResp2 */

int32
fs_simStatfsReq2(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgStatReq       *fsMsg;
    nfs_fh              *fh;

    //printf("nfs statfs req rcvd\n");

    fh = (nfs_fh *) arg;

    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgStatReq));
    fsMsg->hdr.requestType = FS_REQ_STAT;
    memcpy(&(fsMsg->fsHandle), fh, sizeof(fs_Handle));

    *len = sizeof(fs_MsgStatReq);
    return NFX_OK;

} /* fs_simStatfsReq2 */

void *
fs_simStatfsResp2(
    void    *fsResp)
{
    fs_MsgStatResp      *fs;
    statfsres           *statfs;

    //printf("nfs statfs resp called\n");

    fs = (fs_MsgStatResp *) fsResp;
    statfs = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(statfsres));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        statfs->status = NFS_OK;
        statfs->statfsres_u.reply.tsize = fs->stat.fsBlockSize;
        statfs->statfsres_u.reply.bsize = fs->stat.fsBlockSize;
        statfs->statfsres_u.reply.blocks = fs->stat.numBlks;
        statfs->statfsres_u.reply.bfree = fs->stat.numBlksAvail;
        statfs->statfsres_u.reply.bavail = fs->stat.numBlksAvailUser;
    }
    else
    {
        statfs->status = fs->hdr.fsStatus;
    }
    
    return statfs;

} /* fs_simStatfsResp2 */

void
fs_simFreeResp2(
    void    *resp)
{
    fs_ramDealloc(FS_RAM_NFS_RESP_ARG, resp);

} /* fs_simFreeResp2 */

/*-----------------------------------------------------------------------------
 * NFS mount protocol actions.
 */
void *
mountproc_null_1(
    void            *ap,
    struct svc_req  *rp)
{
    static int  ret = 0;

    return (void *) &ret;

} /* mountproc_null_1 */

fhstatus *
mountproc_mnt_1(
    dirpath         *ap,
    struct svc_req  *rp)
{
    fs_FileSystem       *fileSystem;
    static fhstatus     fhs;
    fs_Handle           fsHandle;

    bzero(&fhs, sizeof(fhstatus));

    fhs.fhs_status = NFS_OK;

    //printf("Received mount request for %s\n", *ap);

    if (!strcmp(*ap, "/v1") || !strcmp(*ap, "v1"))
    {
        fsHandle.volId.whole_thing = 0;
        fsHandle.volId.details.u.parts.clusterUnique = 1;
        fsHandle.volId.details.u.parts.reserved = 0;
    }
    else
    if (!strcmp(*ap, "/v2") || !strcmp(*ap, "v2"))
    {
        fsHandle.volId.whole_thing = 0;
        fsHandle.volId.details.u.parts.clusterUnique = 2;
        fsHandle.volId.details.u.parts.reserved = 0;
    }
    else
    if (!strcmp(*ap, "/v3") || !strcmp(*ap, "v3"))
    {
        fsHandle.volId.whole_thing = 0;
        fsHandle.volId.details.u.parts.clusterUnique = 3;
        fsHandle.volId.details.u.parts.reserved = 0;
    }
    else
    {
        fsHandle.volId.whole_thing = 0;
        fsHandle.volId.details.u.parts.clusterUnique = -1;
        fsHandle.volId.details.u.parts.reserved = 0;
        fhs.fhs_status = NFSERR_NOENT;
    }

    fileSystem = fs_fileSystemForVolume(fsHandle.volId);
    if (fileSystem == NULL)
    {
        fhs.fhs_status = NFSERR_NOENT;
    }
    else
	{
    	fsHandle.inum = FS_ROOT_INODE;
    	fsHandle.gen = 0;   // root inode gen doesn't change.
    	fsHandle.mkfsTime = fileSystem->super->mkfsTime;
    	fsHandle.mntPnt = FS_ROOT_INODE;

    	memcpy(&(fhs.fhstatus_u.fhs_fhandle), &fsHandle, sizeof(fs_Handle));
	}

    if (fhs.fhs_status != NFS_OK)
    {
        printf("Volume %s is not available\n", *ap);
    }

    return &fhs;

} /* mountproc_mnt_1 */

mountres3 *
mountproc_mnt_3(
    dirpath         *ap,
    struct svc_req  *rp)
{
    fs_FileSystem       *fileSystem;
    static mountres3    mountres3;
    static char         fh_val[NFS3_FHSIZE];
    static int          auth_val = AUTH_UNIX;
    fs_Handle           fsHandle;

    bzero(&mountres3, sizeof(mountres3));

    mountres3.fhs_status = MNT_OK;

    //printf("Received mount request for %s\n", *ap);

    if (!strcmp(*ap, "/v1") || !strcmp(*ap, "v1"))
    {
        fsHandle.volId.whole_thing = 0;
        fsHandle.volId.details.u.parts.clusterUnique = 1;
        fsHandle.volId.details.u.parts.reserved = 0;
    }
    else
    if (!strcmp(*ap, "/v2") || !strcmp(*ap, "v2"))
    {
        fsHandle.volId.whole_thing = 0;
        fsHandle.volId.details.u.parts.clusterUnique = 2;
        fsHandle.volId.details.u.parts.reserved = 0;
    }
    else
    if (!strcmp(*ap, "/v3") || !strcmp(*ap, "v3"))
    {
        fsHandle.volId.whole_thing = 0;
        fsHandle.volId.details.u.parts.clusterUnique = 3;
        fsHandle.volId.details.u.parts.reserved = 0;
    }
    else
    {
        fsHandle.volId.whole_thing = 0;
        fsHandle.volId.details.u.parts.clusterUnique = -1;
        fsHandle.volId.details.u.parts.reserved = 0;
        mountres3.fhs_status = NFSERR_NOENT;
    }

    fileSystem = fs_fileSystemForVolume(fsHandle.volId);
    if (fileSystem == NULL)
    {
        mountres3.fhs_status = MNT3ERR_NOENT;
    }
    else
	{
    	fsHandle.inum = FS_ROOT_INODE;
    	fsHandle.gen = 0;   // root inode gen doesn't change.
    	fsHandle.mkfsTime = fileSystem->super->mkfsTime;
    	fsHandle.mntPnt = FS_ROOT_INODE;

    	mountres3.mntres3_u.mntinfo.fhandle.fhandle3_len = sizeof(fs_Handle);
    	memcpy(fh_val, &fsHandle, sizeof(fs_Handle));
    	mountres3.mntres3_u.mntinfo.fhandle.fhandle3_val = &fh_val[0];

        mountres3.mntres3_u.mntinfo.auth_flavors.auth_flavors_len = 1;
        mountres3.mntres3_u.mntinfo.auth_flavors.auth_flavors_val = &auth_val;
	}

    if (mountres3.fhs_status != MNT_OK)
    {
        printf("Volume %s is not available\n", *ap);
    }

    return &mountres3;

} /* mountproc_mnt_3 */

mountlist *
mountproc_dump_1(
    void            *ap,
    struct svc_req  *rp)
{
    return NULL;

} /* mountproc_dump_1 */

void *
mountproc_umnt_1(
    void            *ap,
    struct svc_req  *rp)
{
    static int  ret = 0;

    return (void *) &ret;

} /* mountproc_umnt_1 */

void *
mountproc_umntall_1(
    void            *ap,
    struct svc_req  *rp)
{
    static int  ret = 0;

    return (void *) &ret;

} /* mountproc_umntall_1 */

exports *
mountproc_export_1(
    void            *ap,
    struct svc_req  *rp)
{

    return NULL;

} /* mountproc_export_1 */

exports *
mountproc_exportall_1(
    void            *ap,
    struct svc_req  *rp)
{
    return NULL;

} /* mountproc_exportall_1 */

/*-----------------------------------------------------------------------------
 * NFS v3 
 */

int32
fs_simNullReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    bzero(buf, sizeof(fs_MsgNullReq));
    *len = sizeof(fs_MsgNullReq);
    return NFX_OK;

} /* fs_simNullReq3 */

void *
fs_simNullResp3(
    void    *fsResp)
{
    return NULL;
} /* fs_simNullResp3 */

int32
fs_simGetAttrReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgGetAttrReq    *fsMsg;
    GETATTR3args        *args = (GETATTR3args*)arg;
    nfs_fh3             *fh;

    //printf("nfs getattr req rcvd\n");

    fh = &(args->object);

    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgGetAttrReq));
    fsMsg->hdr.requestType = FS_REQ_GET_ATTR;
    fsMsg->hdr.attrType = FS_REQ_ALL_ATTR;

    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->fileHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    *len = sizeof(fs_MsgGetAttrReq);
    return NFX_OK;

} /* fs_simGetAttrReq3 */

void *
fs_simGetAttrResp3(
    void    *fsResp)
{
    fs_MsgGetAttrResp   *fs;
    GETATTR3res         *as;

    //printf("nfs getattr resp called\n");

    fs = (fs_MsgGetAttrResp *) fsResp;
    as = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(GETATTR3res));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        as->status = NFS_OK;

        FS_ASSERT (fs->attr.attrType == FS_POST_ATTRS);
        fs_simDoGetAttr3(&(as->res_u.ok.obj_attributes), &(fs->attr));
    }
    else
    {
        as->status = fs->hdr.fsStatus;
    }
    
    return as;

} /* fs_simGetAttrResp3 */

int32
fs_simSetAttrReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgSetAttrReq    *fsMsg;
    SETATTR3args        *args = (SETATTR3args*)arg;
    nfs_fh3             *fh;

    //printf("nfs setattr req rcvd\n");

    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgSetAttrReq));
    fsMsg->hdr.requestType = FS_REQ_SET_ATTR;
    fsMsg->hdr.attrType = FS_REQ_ALL_ATTR;

    fh = &(args->object);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->fileHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    fs_simDoSetAttr3(&(fsMsg->attr), &(args->new_attributes));

    if (args->guard.check)
    {
        fsMsg->checkCtime.secs = args->guard.obj_ctime.seconds;
        fsMsg->checkCtime.nsecs = args->guard.obj_ctime.nseconds;
    }

    *len = sizeof(fs_MsgSetAttrReq);
    return NFX_OK;

} /* fs_simSetAttrReq3 */

void *
fs_simSetAttrResp3(
    void    *fsResp)
{
    fs_MsgSetAttrResp   *fs;
    SETATTR3res         *as;

    //printf("nfs setattr resp called\n");

    fs = (fs_MsgSetAttrResp *) fsResp;

    as = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(SETATTR3res));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        as->status = NFS_OK;

        FS_ASSERT (fs->attr[0].attrType == FS_PRE_ATTRS);
        FS_ASSERT (fs->attr[1].attrType == FS_POST_ATTRS);
        fs_simDoGetWccData3(&(as->res_u.ok.obj_wcc), &(fs->attr[0]), &(fs->attr[1]));
    }
    else
    {
        as->status = fs->hdr.fsStatus;

        if ((fs->attr[0].attrType == FS_PRE_ATTRS) &&
            (fs->attr[1].attrType == FS_POST_ATTRS))
        {
            fs_simDoGetWccData3(&(as->res_u.fail.obj_wcc), &(fs->attr[0]), &(fs->attr[1]));
        }
        else
        {
            as->res_u.fail.obj_wcc.before.attributes = FALSE;
            as->res_u.fail.obj_wcc.after.attributes = FALSE;
            //printf ("fs_simSetAttrResp3 Error  fs->attr[0..1].attrType != FS_PRE_ATTRS..FS_POST_ATTRS\n");
        }
    }

    return as;

} /* fs_simSetAttrResp3 */

int32
fs_simLookupReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgLookupReq     *fsMsg;
    LOOKUP3args         *args = (LOOKUP3args*)arg;
    int32               nameLen;
    nfs_fh3             *fh;

    //printf("nfs lookup req rcvd\n");
    
    nameLen = MIN(FS_NAME_MAX, strlen(args->what.name));
    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgLookupReq));
    fsMsg->hdr.requestType = FS_REQ_LOOKUP;
    fsMsg->hdr.attrType = FS_REQ_ALL_ATTR;

    fh = &(args->what.dir);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->parentHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    memcpy(&(fsMsg->fileName), args->what.name, nameLen);
    fsMsg->fileName[nameLen] = '\0';
    fsMsg->nameLen = nameLen;

    *len = sizeof(fs_MsgLookupReq) + nameLen + 1;
    return NFX_OK;

} /* fs_simLookupReq3 */

void *
fs_simLookupResp3(
    void    *fsResp)
{
    fs_MsgLookupResp    *fs;
    LOOKUP3res          *resp;

    //printf("nfs lookup resp called\n");

    fs = (fs_MsgLookupResp *) fsResp;

    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(LOOKUP3res));
    bzero (resp, sizeof(LOOKUP3res));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;

        FS_ASSERT (fs->attr[0].attrType == FS_POST_ATTRS);
        resp->res_u.ok.obj_attributes.attributes = TRUE;
        fs_simDoGetAttr3(&(resp->res_u.ok.obj_attributes.attr),
                         &(fs->attr[0]));

        FS_ASSERT (fs->attr[1].attrType == FS_POST_PAR_ATTRS);
        resp->res_u.ok.dir_attributes.attributes = TRUE;
        fs_simDoGetAttr3(&(resp->res_u.ok.dir_attributes.attr),
                         &(fs->attr[1]));

        resp->res_u.ok.object.fh3_length = sizeof(fs_Handle);
        memcpy( &(resp->res_u.ok.object.fh3_u.data),
                &(fs->fileHandle),
                sizeof(fs_Handle));
    }
    else
    {
        resp->status = fs->hdr.fsStatus;

        if (fs->attr[0].attrType == FS_POST_PAR_ATTRS)
        {
            resp->res_u.fail.dir_attributes.attributes = TRUE;
            fs_simDoGetAttr3(&(resp->res_u.fail.dir_attributes.attr),
                             &(fs->attr[0]));
        }
        else
        {
            resp->res_u.fail.dir_attributes.attributes = FALSE;
        }
    }
    
    return resp;

} /* fs_simLookupResp3 */

int32
fs_simReadLinkReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgReadLinkReq   *fsMsg;
    READLINK3args       *args = (READLINK3args*)arg;
    nfs_fh3             *fh;

    fh = &(args->symlink);

    //printf("nfs readlink req rcvd\n");
    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgReadLinkReq));
    fsMsg->hdr.requestType = FS_REQ_READLINK;
    fsMsg->hdr.attrType = FS_REQ_ALL_ATTR;

    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->fileHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    *len = sizeof(fs_MsgReadLinkReq);
    return NFX_OK;

} /* fs_simReadLinkReq3 */

void *
fs_simReadLinkResp3(
    void    *fsResp)
{
    fs_MsgReadLinkResp  *fs;
    READLINK3res        *resp;

    //printf("nfs create resp called\n");

    fs = (fs_MsgReadLinkResp *) fsResp;

    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(READLINK3res));
    bzero(resp, sizeof *resp);
    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;

        FS_ASSERT (fs->attr.attrType == FS_POST_ATTRS);
        resp->res_u.ok.symlink_attributes.attributes = TRUE;
        fs_simDoGetAttr3(&(resp->res_u.ok.symlink_attributes.attr),
                         &(fs->attr));

        resp->res_u.ok.data = fs_ramAlloc(FS_RAM_NFS_RESP_ARG,
                                                fs->nameLen + 1);
        memcpy( resp->res_u.ok.data,
                fs->linkName,
                fs->nameLen);
        resp->res_u.ok.data[fs->nameLen] = '\0';
    }
    else
    {
        resp->status = fs->hdr.fsStatus;

        if (fs->attr.attrType == FS_POST_ATTRS)
        {
            resp->res_u.fail.symlink_attributes.attributes = TRUE;
            fs_simDoGetAttr3(&(resp->res_u.fail.symlink_attributes.attr),
                             &(fs->attr));
        }
        else
        {
            resp->res_u.fail.symlink_attributes.attributes = FALSE;
            //printf ("fs_simReadLinkResp3 Error fs->attr.attrType != FS_POST_ATTRS\n");
        }
    }
    
    return resp;

} /* fs_simReadLinkResp3 */

void
fs_simFreeReadLinkResp3(
    void    *fsResp)
{
    READLINK3res        *resp = fsResp;

    if ((resp->status == NFS_OK) &&
        (resp->res_u.ok.data != NULL))
    {
        fs_ramDealloc(FS_RAM_NFS_RESP_ARG, resp->res_u.ok.data);
    }
        
    fs_ramDealloc(FS_RAM_NFS_RESP_ARG, resp);

} /* fs_simFreeReadLinkResp3 */

int32
fs_simReadReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    READ3args           *args = (READ3args *) arg;
    fs_MsgReadReq       *msg = buf; 
    nfs_fh3             *fh;

    // printf("nfs read req rcvd\n");

    bzero(msg, sizeof (fs_MsgReadReq));          // TODO - remove

    msg->hdr.requestType = FS_REQ_READ;
    msg->hdr.attrType = FS_REQ_ALL_ATTR;

    fh = &(args->file);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(msg->fileHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    fs_simGetNfs_uint64 (msg->fileOffset, args->offset);
    msg->readLength      = args->count;
    
    *len = sizeof(fs_MsgReadReq);
    return NFX_OK;
} /* fs_simReadReq3 */

void *
fs_simReadResp3(
    void    *fsResp)
{
    fs_MsgReadResp      *fs;
    READ3res            *rs;

    //printf("nfs Read resp called\n");

    fs = (fs_MsgReadResp *) fsResp;
    rs = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(READ3res));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        rs->status = NFS_OK;

        FS_ASSERT (fs->attr[0].attrType == FS_POST_ATTRS);
        rs->res_u.ok.file_attributes.attributes = TRUE;
        fs_simDoGetAttr3(&(rs->res_u.ok.file_attributes.attr),
                         &(fs->attr[0]));

        rs->res_u.ok.count = fs->bytesRead;
        rs->res_u.ok.eof = fs->eof;
        // resp->res_u.ok.size ;

        rs->res_u.ok.data.data_len = fs->bytesRead;
        if (fs->bytesRead)
          {
            rs->res_u.ok.data.data_val = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, fs->bytesRead);
            memcpy(rs->res_u.ok.data.data_val, 
                   &(((char*)fs)[fs->data_index]),
                   fs->bytesRead);
          }
        else
          {
            rs->res_u.ok.data.data_val = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, 1);
            // dummy allocation for the moment, 
            // Todo: need changes in
            // fs_simFreeReadResp2, EFENCE and perhaps xdr_readres (xdr_bytes)
          }
    }
    else
    {
        rs->status = fs->hdr.fsStatus;

        if (fs->attr[0].attrType == FS_POST_ATTRS)
        {
            rs->res_u.fail.file_attributes.attributes = TRUE;
            fs_simDoGetAttr3(&(rs->res_u.fail.file_attributes.attr),
                             &(fs->attr[0]));
        }
        else
        {
            rs->res_u.fail.file_attributes.attributes = FALSE;
            //printf("fs_simReadResp3 Error fs->attr[0].attrType != FS_POST_ATTRS\n");
        }
    }
    
    return rs;
} /* fs_simReadResp3 */

void
fs_simFreeReadResp3(
    void    *fsResp)
{
    READ3res             *resp = fsResp;

    //printf("nfs Free Read resp called\n");

    if (resp->status == NFS_OK)
        fs_ramDealloc(FS_RAM_NFS_RESP_ARG, resp->res_u.ok.data.data_val);
    fs_ramDealloc(FS_RAM_NFS_RESP_ARG, resp);

} /* fs_simFreeReadResp3 */

int32
fs_simWriteReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    uint32              loop, bytesThis, length = 0;
    WRITE3args          *args = (WRITE3args *) arg;
    fs_MsgWriteReq      *msg = buf;
    nfs_fh3             *fh;

    // printf("nfs write req rcvd\n");

    bzero(msg, sizeof (fs_MsgWriteReq));          // TODO - remove

    fh = &(args->file);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(msg->fileHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    msg->hdr.requestType = FS_REQ_WRITE;
    msg->hdr.attrType    = FS_REQ_ALL_ATTR;

    fs_simGetNfs_uint64 (msg->fileOffset, args->offset);
    msg->writeLength     = args->count;
    msg->stableHow       = args->stable;

    msg->nbrUsedCod      = (msg->writeLength - 1) / FS_BLOCK_SIZE + 1;
    for (loop = 0 ; loop < msg->nbrUsedCod; loop++)
    {
        bytesThis = MIN (msg->writeLength - length, FS_BLOCK_SIZE);
        msg->cod[loop].cod = fs_sim_get_new_cod(FS_SIM_MSG_ADDR_NFS);
        msg->cod[loop].offset = 0;
        msg->cod[loop].length = bytesThis;
        length += bytesThis;
    }
    FS_ASSERT (length == msg->writeLength);
   
    *len = sizeof(fs_MsgWriteReq);
    return NFX_OK;

} /* fs_simWriteReq3 */

void *
fs_simWriteResp3(
    void    *fsResp)
{
    fs_MsgWriteResp     *fs;
    WRITE3res           *as;

    //printf("nfs Write resp called\n");

    fs = (fs_MsgWriteResp *) fsResp;
    as = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(WRITE3res));
    bzero (as, sizeof(WRITE3res));

    if (fs->hdr.fsStatus == NFX_OK)
    {
        as->status = NFS_OK;
        as->res_u.ok.count = fs->count;
        as->res_u.ok.committed = fs->stableHow;
        // Rest is set to 0
        memcpy(&(as->res_u.ok.verf), &(fs->verifier), sizeof(uint32));

        FS_ASSERT (fs->attr[0].attrType == FS_PRE_ATTRS);
        FS_ASSERT (fs->attr[1].attrType == FS_POST_ATTRS);
        fs_simDoGetWccData3(&(as->res_u.ok.file_wcc), &(fs->attr[0]), &(fs->attr[1]));
    }
    else
    {
        as->status = fs->hdr.fsStatus;

        if ((fs->attr[0].attrType == FS_PRE_ATTRS) &&
            (fs->attr[1].attrType == FS_POST_ATTRS))
        {
            fs_simDoGetWccData3(&(as->res_u.fail.file_wcc), &(fs->attr[0]), &(fs->attr[1]));
        }
        else
        {
            as->res_u.fail.file_wcc.before.attributes = FALSE;
            as->res_u.fail.file_wcc.after.attributes = FALSE;
            
            //if (fs->attr[0].attrType != FS_PRE_ATTRS)
            //  printf ("fs_simWriteResp3 Error  fs->attr[0].attrType != FS_PRE_ATTRS\n");
            //if (fs->attr[1].attrType != FS_POST_ATTRS)
            //  printf ("fs_simWriteResp3 Error  fs->attr[1].attrType != FS_POST_ATTRS\n");
        }
    }
    
    return as;

} /* fs_simWriteResp3 */

int32
fs_simCreateReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgCreateReq     *fsMsg;
    CREATE3args         *args = (CREATE3args *) arg;
    int32               nameLen;
    nfs_fh3             *fh;

    //printf("nfs create req rcvd\n");

    nameLen = MIN(FS_NAME_MAX, strlen(args->where.name));
    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgCreateReq));
    fsMsg->hdr.requestType = FS_REQ_CREATE;
    fsMsg->hdr.attrType    = FS_REQ_ALL_ATTR;

    fsMsg->flags = (ushort16)args->how.mode;
    if ((args->how.mode == GUARDED) || (args->how.mode == UNCHECKED))
    {
        fs_simDoSetAttr3(&(fsMsg->u.attr), &(args->how.createhow3_u.obj_attributes));
    }
    if (args->how.mode == EXCLUSIVE)
    {
        FS_ASSERT(sizeof(fsMsg->u.verifier) == sizeof(args->how.createhow3_u.verf));
        memcpy(&(fsMsg->u.verifier), &(args->how.createhow3_u.verf), sizeof(fsMsg->u.verifier));
    }
    
    fh = &(args->where.dir);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->parentHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    memcpy(&(fsMsg->fileName), args->where.name, nameLen);
    fsMsg->fileName[nameLen] = '\0';
    fsMsg->nameLen = (ushort16)nameLen;

    *len = sizeof(fs_MsgCreateReq) + nameLen + 1;
    return NFX_OK;

} /* fs_simCreatReq3 */

void *
fs_simCreateResp3(
    void    *fsResp)
{
    fs_MsgCreateResp    *fs;
    CREATE3res          *resp;

    //printf("nfs create resp called\n");

    fs = (fs_MsgCreateResp *) fsResp;

    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(CREATE3res));
    bzero (resp, sizeof(CREATE3res));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;
        resp->res_u.ok.obj.handle_follows = TRUE;
        resp->res_u.ok.obj.handle.fh3_length = sizeof(fs_Handle);
        memcpy( &(resp->res_u.ok.obj.handle.fh3_u.data),
                &(fs->fileHandle),
                sizeof(fs_Handle));
        
        FS_ASSERT (fs->attr[1].attrType == FS_POST_ATTRS);
        resp->res_u.ok.obj_attributes.attributes = TRUE;
        fs_simDoGetAttr3(&(resp->res_u.ok.obj_attributes.attr),
                         &(fs->attr[1]));

        FS_ASSERT (fs->attr[0].attrType == FS_PRE_PAR_ATTRS);
        FS_ASSERT (fs->attr[2].attrType == FS_POST_PAR_ATTRS);
        fs_simDoGetWccData3(&(resp->res_u.ok.dir_wcc), &(fs->attr[0]), &(fs->attr[2]));
    }
    else
    {
        resp->status = fs->hdr.fsStatus;

        fs_simDoGetWccData3(&(resp->res_u.fail.dir_wcc), &(fs->attr[0]), &(fs->attr[1]));
    }
    
    return resp;

} /* fs_simCreatResp3 */

int32
fs_simRemoveReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgRemoveReq     *fsMsg;
    REMOVE3args         *args = (REMOVE3args*)arg;
    int32               nameLen;
    nfs_fh3             *fh;

    //printf("nfs remove req rcvd\n");

    nameLen = MIN(FS_NAME_MAX, strlen(args->object.name));
    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgRemoveReq));
    fsMsg->hdr.requestType = FS_REQ_REMOVE;
    fsMsg->hdr.attrType    = FS_REQ_ALL_ATTR;

    fh = &(args->object.dir);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->parentHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    memcpy(&(fsMsg->name), args->object.name, nameLen);
    fsMsg->name[nameLen] = '\0';
    fsMsg->nameLen = nameLen;

    *len = sizeof(fs_MsgRemoveReq) + nameLen + 1;
    return NFX_OK;

} /* fs_simRemoveReq3 */

void *
fs_simRemoveResp3(
    void    *fsResp)
{
    fs_MsgRemoveResp    *fs;
    REMOVE3res          *resp;

    //printf("nfs remove resp called\n");

    fs = (fs_MsgRemoveResp *) fsResp;

    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(REMOVE3res));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;

        FS_ASSERT (fs->attr[0].attrType == FS_PRE_PAR_ATTRS);
        FS_ASSERT (fs->attr[1].attrType == FS_POST_PAR_ATTRS);
        fs_simDoGetWccData3(&(resp->res_u.ok.dir_wcc), &(fs->attr[0]), &(fs->attr[1]));
    }
    else
    {
        resp->status = fs->hdr.fsStatus;

        if (fs->attr[0].attrType == FS_PRE_PAR_ATTRS &&
            fs->attr[1].attrType == FS_POST_PAR_ATTRS)
        {
            fs_simDoGetWccData3(&(resp->res_u.fail.dir_wcc), &(fs->attr[0]), &(fs->attr[1]));
        }
    }
    
    return resp;

} /* fs_simRemoveResp3 */

int32
fs_simRenameReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgRenameReq     *fsMsg;
    RENAME3args         *args = (RENAME3args*)arg;
    int32               fromNameLen, toNameLen;
    char                *toPtr;
    nfs_fh3             *fh;

    //printf("nfs rename req rcvd\n");

    fromNameLen = MIN(FS_NAME_MAX, strlen(args->from.name));
    toNameLen = MIN(FS_NAME_MAX, strlen(args->to.name));
    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgRenameReq));
    fsMsg->hdr.requestType = FS_REQ_RENAME;
    fsMsg->hdr.attrType    = FS_REQ_ALL_ATTR;

    fh = &(args->from.dir);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->fromDirHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    fh = &(args->to.dir);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->toDirHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    memcpy(&(fsMsg->fromName), args->from.name, fromNameLen);
    fsMsg->fromName[fromNameLen] = '\0';
    fsMsg->fromNameLen = fromNameLen;
    toPtr = &fsMsg->fromName[fromNameLen + 1];
    memcpy(toPtr, args->to.name, toNameLen);
    toPtr[toNameLen] = '\0';
    fsMsg->toNameLen = toNameLen;

    *len = sizeof(fs_MsgRenameReq) + fromNameLen + toNameLen + 2;
    return NFX_OK;

} /* fs_simRenameReq3 */

void *
fs_simRenameResp3(
    void    *fsResp)
{
    fs_MsgRenameResp    *fs;
    RENAME3res          *resp;
    int                 attrIndex;
    int                 fromDirPre, fromDirPost, toDirPre, toDirPost;

    //printf("nfs rename resp called\n");

    fs = (fs_MsgRenameResp *) fsResp;

    for (attrIndex = 0; attrIndex < fs->numAttrs; attrIndex++)
    {
        switch(fs->attr[attrIndex].attrType)
        {
        case FS_PRE_FPAR_ATTRS:
            fromDirPre = attrIndex;
            break;
        case FS_PRE_TPAR_ATTRS:
            toDirPre = attrIndex;
            break;
        case FS_POST_FPAR_ATTRS:
            fromDirPost = attrIndex;
            break;
        case FS_POST_TPAR_ATTRS:
            toDirPost = attrIndex;
            break;
        default:
            // Ignore this attribute
        }
    }

    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(RENAME3res));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;

        fs_simDoGetWccData3(&(resp->res_u.ok.fromdir_wcc),
                            &(fs->attr[fromDirPre]),
                            &(fs->attr[fromDirPost]));
        fs_simDoGetWccData3(&(resp->res_u.ok.todir_wcc),
                            &(fs->attr[toDirPre]),
                            &(fs->attr[toDirPost]));
    }
    else
    {
        resp->status = fs->hdr.fsStatus;
    }
    
    return resp;

} /* fs_simRenameResp3 */

int32
fs_simLinkReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    LINK3args           *args = (LINK3args*)arg ;
    fs_MsgHardLinkReq   *fsMsg;
    int32               nameLen;
    nfs_fh3             *fh;

    //printf("nfs link req rcvd\n");

    nameLen = MIN(FS_NAME_MAX, strlen(args->link.name));
    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgHardLinkReq));
    fsMsg->hdr.requestType = FS_REQ_HARDLINK;
    fsMsg->hdr.attrType    = FS_REQ_ALL_ATTR;

    fh = &(args->link.dir);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->parentDirHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    fh = &(args->file);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->fileHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    memcpy(&(fsMsg->linkName), args->link.name, nameLen);
    fsMsg->linkName[nameLen] = '\0';
    fsMsg->nameLen = nameLen;

    *len = sizeof(fs_MsgHardLinkReq) + nameLen + 1;

    return NFX_OK;

} /* fs_simLinkReq3 */

void *
fs_simLinkResp3(
    void    *fsResp)
{
    fs_MsgHardLinkResp  *fs;
    LINK3res            *resp;

    //printf("nfs link resp called\n");

    fs = (fs_MsgHardLinkResp *) fsResp;

    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(RENAME3res));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;

        FS_ASSERT (fs->attr[1].attrType == FS_POST_ATTRS);
        resp->res_u.ok.file_attributes.attributes = TRUE;
        fs_simDoGetAttr3(&(resp->res_u.ok.file_attributes.attr), &(fs->attr[1]));

        FS_ASSERT (fs->attr[0].attrType == FS_PRE_PAR_ATTRS);
        FS_ASSERT (fs->attr[2].attrType == FS_POST_PAR_ATTRS);
        fs_simDoGetWccData3(&(resp->res_u.ok.linkdir_wcc), &(fs->attr[0]), &(fs->attr[2]));
    }
    else
    {
        resp->status = fs->hdr.fsStatus;

        if (fs->attr[1].attrType == FS_POST_ATTRS)
        {
            resp->res_u.fail.file_attributes.attributes = TRUE;
            fs_simDoGetAttr3(&(resp->res_u.fail.file_attributes.attr), &(fs->attr[1]));
        }


        if (fs->attr[0].attrType == FS_PRE_PAR_ATTRS &&
            fs->attr[2].attrType == FS_POST_PAR_ATTRS)
        {
            fs_simDoGetWccData3(&(resp->res_u.fail.linkdir_wcc), &(fs->attr[0]), &(fs->attr[2]));
        }
    }
    
    return resp;

} /* fs_simLinkResp3 */

int32
fs_simSymlinkReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    SYMLINK3args        *args = (SYMLINK3args*)arg;
    fs_MsgSymLinkReq    *fsMsg;
    int32               linkLen, linkToLen;
    char                *msgPtr;
    nfs_fh3             *fh;

    if (args->where.name == (void *) -1 ||
        args->symlink.symlink_data == (void *) -1)
    {
        printf("nfs v3 symlink bad name(s)\n");
        return NFX_ERR;
    }

    //printf("nfs symlink req rcvd linkName %s linkToName %s\n",
    //      slargs->from.name, slargs->to);

    linkLen = MIN(FS_NAME_MAX, strlen(args->where.name));
    linkToLen = MIN(FS_PATH_MAX, strlen(args->symlink.symlink_data));

    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgSymLinkReq));
    fsMsg->hdr.requestType = FS_REQ_SYMLINK;
    fsMsg->hdr.attrType    = FS_REQ_ALL_ATTR;

    fh = &(args->where.dir);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->parentHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    msgPtr = fsMsg->linkName;
    memcpy(msgPtr, args->where.name, linkLen);
    msgPtr[linkLen] = '\0';
    fsMsg->linkNameLen = linkLen;
    msgPtr += linkLen + 1;
    memcpy(msgPtr, args->symlink.symlink_data, linkToLen);
    msgPtr[linkToLen] = '\0';
    fsMsg->linkToNameLen = linkToLen;

    *len = sizeof(fs_MsgSymLinkReq) + linkLen + linkToLen + 2;

    return NFX_OK;

} /* fs_simSymlinkReq3 */

void *
fs_simSymlinkResp3(
    void    *fsResp)
{
    fs_MsgSymLinkResp   *fs;
    SYMLINK3res         *resp;

    //printf("nfs link resp called\n");

    fs = (fs_MsgSymLinkResp *) fsResp;
    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(SYMLINK3res));

    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;

        resp->res_u.ok.obj.handle_follows = TRUE;
        resp->res_u.ok.obj.handle.fh3_length = sizeof(fs_Handle);
        memcpy( &(resp->res_u.ok.obj.handle.fh3_u.data),
                &(fs->fileHandle),
                sizeof(fs_Handle));

        FS_ASSERT (fs->attr[1].attrType == FS_POST_ATTRS);
        resp->res_u.ok.obj_attributes.attributes = TRUE;
        fs_simDoGetAttr3(&(resp->res_u.ok.obj_attributes.attr), &(fs->attr[1]));

        FS_ASSERT (fs->attr[0].attrType == FS_PRE_PAR_ATTRS);
        FS_ASSERT (fs->attr[2].attrType == FS_POST_PAR_ATTRS);
        fs_simDoGetWccData3(&(resp->res_u.ok.dir_wcc), &(fs->attr[0]), &(fs->attr[2]));
    }
    else
    {
        resp->status = fs->hdr.fsStatus;
        if (fs->attr[0].attrType == FS_PRE_PAR_ATTRS &&
            fs->attr[1].attrType == FS_POST_PAR_ATTRS)
        {
            fs_simDoGetWccData3(&(resp->res_u.fail.dir_wcc), &(fs->attr[0]), &(fs->attr[1]));
        }
    }
    
    return resp;

} /* fs_simSymlinkResp3 */

int32
fs_simMkdirReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgMkdirReq      *fsMsg;
    MKDIR3args          *args = (MKDIR3args*)arg ;
    int32               nameLen;
    nfs_fh3             *fh;

    //printf("nfs mkdir req rcvd\n");

    nameLen = MIN(FS_NAME_MAX, strlen(args->where.name));
    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgMkdirReq));
    fsMsg->hdr.requestType = FS_REQ_MKDIR;
    fsMsg->hdr.attrType    = FS_REQ_ALL_ATTR;

    fh = &(args->where.dir);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->parentHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    fs_simDoSetAttr3(&(fsMsg->dirAttr), &(args->attributes));

    memcpy(&(fsMsg->dirName), args->where.name, nameLen);
    fsMsg->dirName[nameLen] = '\0';
    fsMsg->nameLen = nameLen;

    *len = sizeof(fs_MsgMkdirReq) + nameLen + 1;
    return NFX_OK;

} /* fs_simMkdirReq3 */

void *
fs_simMkdirResp3(
    void    *fsResp)
{
    fs_MsgMkdirResp     *fs;
    MKDIR3res           *resp;

    //printf("nfs mkdir resp called\n");

    fs = (fs_MsgMkdirResp *) fsResp;

    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(MKDIR3res));
    bzero (resp, sizeof(MKDIR3res));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;

        resp->res_u.ok.obj.handle_follows = TRUE;
        resp->res_u.ok.obj.handle.fh3_length = sizeof(fs_Handle);
        memcpy( &(resp->res_u.ok.obj.handle.fh3_u.data),
                &(fs->dirHandle),
                sizeof(fs_Handle));

        FS_ASSERT (fs->attr[1].attrType == FS_POST_ATTRS);
        resp->res_u.ok.obj_attributes.attributes = TRUE;
        fs_simDoGetAttr3(&(resp->res_u.ok.obj_attributes.attr), &(fs->attr[1]));

        FS_ASSERT (fs->attr[0].attrType == FS_PRE_PAR_ATTRS);
        FS_ASSERT (fs->attr[2].attrType == FS_POST_PAR_ATTRS);
        fs_simDoGetWccData3(&(resp->res_u.ok.dir_wcc), &(fs->attr[0]), &(fs->attr[2]));
    }
    else
    {
        resp->status = fs->hdr.fsStatus;

        fs_simDoGetWccData3(&(resp->res_u.fail.dir_wcc), &(fs->attr[0]), &(fs->attr[1]));
    }
    
    return resp;

} /* fs_simMkdirResp3 */

int32
fs_simRmdirReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgRmdirReq      *fsMsg;
    RMDIR3args          *args = (RMDIR3args*)arg;
    int32               nameLen;
    nfs_fh3             *fh;

    //printf("nfs rmdir req rcvd\n");

    nameLen = MIN(FS_NAME_MAX, strlen(args->object.name));
    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgRmdirReq));
    fsMsg->hdr.requestType = FS_REQ_RMDIR;
    fsMsg->hdr.attrType    = FS_REQ_ALL_ATTR;

    fh = &(args->object.dir);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->parentHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    memcpy(&(fsMsg->name), args->object.name, nameLen);
    fsMsg->name[nameLen] = '\0';
    fsMsg->nameLen = nameLen;

    *len = sizeof(fs_MsgRmdirReq) + nameLen + 1;
    return NFX_OK;

} /* fs_simRmdirReq3 */

void *
fs_simRmdirResp3(
    void    *fsResp)
{
    fs_MsgRmdirResp     *fs;
    RMDIR3res           *resp;

    //printf("nfs rmdir resp called\n");

    fs = (fs_MsgRmdirResp *) fsResp;

    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(RMDIR3res));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;

        FS_ASSERT (fs->attr[0].attrType == FS_PRE_PAR_ATTRS);
        FS_ASSERT (fs->attr[1].attrType == FS_POST_PAR_ATTRS);
        fs_simDoGetWccData3(&(resp->res_u.ok.dir_wcc), &(fs->attr[0]), &(fs->attr[1]));
    }
    else
    {
        resp->status = fs->hdr.fsStatus;
        if (fs->attr[0].attrType == FS_PRE_PAR_ATTRS &&
            fs->attr[1].attrType == FS_POST_PAR_ATTRS)
        {
            fs_simDoGetWccData3(&(resp->res_u.fail.dir_wcc), &(fs->attr[0]), &(fs->attr[1]));
        }
    }
    
    return resp;

} /* fs_simRmdirResp3 */

int32
fs_simReaddirReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgReadDirReq    *fsMsg;
    READDIR3args        *args = (READDIR3args*)arg;
    nfs_fh3             *fh;

    //printf("nfs readdir req rcvd\n");

    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgReadDirReq));
    fsMsg->hdr.requestType = FS_REQ_READDIR;
    fsMsg->hdr.attrType    = FS_REQ_ALL_ATTR;

    fh = &(args->dir);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->fileHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    fs_simGetNfs_uint64 (fsMsg->cookie, args->cookie);
    fsMsg->maxBytes = args->count;

    FS_ASSERT (sizeof(fsMsg->cookieVerif) == sizeof(args->cookieverf));
    memcpy(&(fsMsg->cookieVerif), &(args->cookieverf), sizeof(args->cookieverf));

    *len = sizeof(fs_MsgReadDirReq);
    return NFX_OK;

} /* fs_simReaddirReq3 */

void *
fs_simReaddirResp3(
    void    *fsResp)
{
    fs_MsgReadDirResp       *fs;
    READDIR3res             *resp;
    entry3                  *ent;
    int32                   i;
    fs_DirEntry             *dirent;
    int32                   nameLen;

    fs = (fs_MsgReadDirResp *) fsResp;

    //printf("nfs readdir resp called, %d, %d\n", fs->numEntriesThis, fs->eof);

    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(READDIR3res));
    bzero(resp, sizeof(READDIR3res));
    i = 0;
        
    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;

        FS_ASSERT (fs->dirAttr.attrType == FS_POST_ATTRS);
        resp->res_u.ok.dir_attributes.attributes = TRUE;
        fs_simDoGetAttr3(&(resp->res_u.ok.dir_attributes.attr), &(fs->dirAttr));

        FS_ASSERT (sizeof(resp->res_u.ok.cookieverf) == sizeof(fs->cookieVerif));
        memcpy(&(resp->res_u.ok.cookieverf), &(fs->cookieVerif), sizeof(fs->cookieVerif));

        fs_simSetNfs_uint64 (resp->res_u.ok.cookie, 0ll) ;  // ?? 
        resp->res_u.ok.count = fs->numEntriesThis;
        // ?? resp->res_u.ok.size = fs->;
        resp->res_u.ok.reply.entries = NULL;

        // Walk the fs response, allocate an ent for
        // each and link it to the response.
        dirent = &(fs->dirEntry[0]);
        if (fs->numEntriesThis)
        {
            // ent = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(entry3) * fs->numEntriesThis);
            ent = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(entry3) * 128); // SFS_MAXDIRENTS);
            bzero (ent, sizeof(entry3) * 128); // SFS_MAXDIRENTS);
            resp->res_u.ok.reply.entries = ent;
        }

        FS_ASSERT(fs->numEntriesThis <= 128);

        for (i = 0; i < fs->numEntriesThis; i++)
        {
            if (dirent->nameLen)
            {
                fs_simSetNfs_uint64 (ent[i].fileid, dirent->inode);
                nameLen = MIN(SFS_MAXNAMLEN, dirent->nameLen);
                memcpy(&(ent[i].name[0]), dirent->name, nameLen);
                ent[i].name[nameLen] = '\0';
                // ent->name = dirent->name;
                fs_simSetNfs_uint64 (ent[i].cookie, dirent->cookie);
            }
            else
            {
                // early end of entries???
                break;
            }

            FS_READDIR_ADV_ENTRY_PTR(dirent);
        }

        resp->res_u.ok.reply.eof = fs->eof;
    }
    else
    {
        resp->status = fs->hdr.fsStatus;
        
        if (fs->dirAttr.attrType == FS_POST_ATTRS)
        {
            resp->res_u.fail.dir_attributes.attributes = TRUE;
            fs_simDoGetAttr3(&(resp->res_u.fail.dir_attributes.attr), &(fs->dirAttr));
        }
        else
        {
            resp->res_u.fail.dir_attributes.attributes = TRUE;
        }
    }

    if (i != fs->numEntriesThis) {
        printf("i=%d, fs->numEntriesThis=%d\n", i, fs->numEntriesThis);
        FS_ASSERT(i == fs->numEntriesThis);
    }
    //printf("  done readdir resp\n");
    
    return resp;

} /* fs_simReaddirResp3 */

void
fs_simFreeReaddirResp3(
    void    *fsResp)
{
    READDIR3res  *resp = fsResp;

    //printf("free readdir resp3\n");

    if (resp->status == NFS_OK)
    {
        if (resp->res_u.ok.reply.entries)
            fs_ramDealloc(FS_RAM_NFS_RESP_ARG, resp->res_u.ok.reply.entries);
    }
    
    fs_ramDealloc(FS_RAM_NFS_RESP_ARG, resp);

    //printf("  done free readdir resp3\n");

} /* fs_simFreeReaddirResp3 */

int32
fs_simFsstatReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgStatReq       *fsMsg;
    FSSTAT3args         *args = (FSSTAT3args*)arg;
    nfs_fh3             *fh;

    //printf("nfs statfs req rcvd\n");

    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgStatReq));
    fsMsg->hdr.requestType = FS_REQ_STAT;
    fsMsg->hdr.attrType    = FS_REQ_ALL_ATTR;

    fh = &(args->fsroot);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->fsHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    *len = sizeof(fs_MsgStatReq);
    return NFX_OK;

} /* fs_simStatfsReq3 */

void *
fs_simFsstatResp3(
    void    *fsResp)
{
    fs_MsgStatResp      *fs;
    FSSTAT3res          *resp;

    //printf("nfs statfs resp called\n");

    fs = (fs_MsgStatResp *) fsResp;
    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(FSSTAT3res));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;

        FS_ASSERT (fs->attr.attrType == FS_POST_ATTRS);
        resp->res_u.ok.obj_attributes.attributes = TRUE;
        fs_simDoGetAttr3(&(resp->res_u.ok.obj_attributes.attr), &(fs->attr));

        fs_simSetNfs_uint64 (resp->res_u.ok.tbytes, (fs->stat.numBlks * fs->stat.fsBlockSize));
        fs_simSetNfs_uint64 (resp->res_u.ok.fbytes, (fs->stat.numBlksAvail * fs->stat.fsBlockSize));
        fs_simSetNfs_uint64 (resp->res_u.ok.abytes, (fs->stat.numBlksAvailUser * fs->stat.fsBlockSize));
        fs_simSetNfs_uint64 (resp->res_u.ok.tfiles, fs->stat.numInodes);
        fs_simSetNfs_uint64 (resp->res_u.ok.ffiles, fs->stat.numInodesAvail);
        fs_simSetNfs_uint64 (resp->res_u.ok.afiles, fs->stat.numInodesAvailUser);
        resp->res_u.ok.invarsec = 0 ; // ??
    }
    else
    {
        resp->status = fs->hdr.fsStatus;

        if (fs->attr.attrType == FS_POST_ATTRS)
        {
            resp->res_u.fail.obj_attributes.attributes = TRUE;
            fs_simDoGetAttr3(&(resp->res_u.fail.obj_attributes.attr), &(fs->attr));
        }
        else
        {
            resp->res_u.fail.obj_attributes.attributes = FALSE;
        }
    }
    
    return resp;

} /* fs_simFsstatResp3 */

int32
fs_simAccessReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgAccessReq     *fsMsg;
    ACCESS3args         *args = (ACCESS3args*)arg;
    nfs_fh3             *fh;

    //printf("nfs Access req rcvd\n");

    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgStatReq));
    fsMsg->hdr.requestType = FS_REQ_ACCESS;
    fsMsg->hdr.attrType    = FS_REQ_ALL_ATTR;

    fh = &(args->object);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->fileHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    fsMsg->accessBits = args->access;

    *len = sizeof(fs_MsgAccessReq);
    return NFX_OK;
}

void *
fs_simAccessResp3(
    void    *fsResp)
{
    fs_MsgAccessResp      *fs;
    ACCESS3res            *resp;

    //printf("nfs access resp called\n");

    fs = (fs_MsgAccessResp *) fsResp;
    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(ACCESS3res));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;

        FS_ASSERT (fs->attr[0].attrType == FS_POST_ATTRS);
        resp->res_u.ok.obj_attributes.attributes = TRUE;
        fs_simDoGetAttr3(&(resp->res_u.ok.obj_attributes.attr), &(fs->attr[0]));

        resp->res_u.ok.access = fs->accessBits;
    }
    else
    {
        resp->status = fs->hdr.fsStatus;

        if (fs->attr[0].attrType == FS_POST_ATTRS)
        {
            resp->res_u.fail.obj_attributes.attributes = TRUE;
            fs_simDoGetAttr3(&(resp->res_u.fail.obj_attributes.attr), &(fs->attr[0]));
        }
        else
        {
            //printf ("fs_simAccessResp3 Error  fs->attr[0].attrType != FS_POST_ATTRS\n");
            resp->res_u.fail.obj_attributes.attributes = FALSE;
        }
    }
    
    return resp;
}

int32
fs_simMknodReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgMknodReq      *fsMsg;
    MKNOD3args          *args = (MKNOD3args*)arg;
    nfs_fh3             *fh;
    int32               nameLen;

    //printf("nfs Mknod req rcvd\n");

    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgMknodReq));
    fsMsg->hdr.requestType = FS_REQ_MKNOD;
    fsMsg->hdr.attrType    = FS_REQ_ALL_ATTR;

    fh = &(args->where.dir);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->parentHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    nameLen = MIN(FS_NAME_MAX, strlen(args->where.name));
    memcpy(&(fsMsg->fileName), args->where.name, nameLen);
    fsMsg->fileName[nameLen] = '\0';
    fsMsg->nameLen = nameLen;

    fsMsg->attr.type = args->what.type;
    if (fsMsg->attr.type == FS_FTYPE_FIFO)
    {
        fs_simDoSetAttr3(&(fsMsg->attr), &(args->what.mknoddata3_u.pipe_attributes));
    }
    else
    {
        fs_simDoSetAttr3(&(fsMsg->attr), &(args->what.mknoddata3_u.device.dev_attributes));
        fsMsg->devNum.major = args->what.mknoddata3_u.device.spec.specdata1;
        fsMsg->devNum.minor = args->what.mknoddata3_u.device.spec.specdata2;
    }

    *len = sizeof(fs_MsgMknodReq);
    return NFX_OK;
}

void *
fs_simMknodResp3(
    void    *fsResp)
{
    fs_MsgMknodResp      *fs;
    MKNOD3res            *resp;

    //printf("nfs Mknod resp called\n");

    fs = (fs_MsgMknodResp *) fsResp;
    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(MKNOD3res));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;

        resp->res_u.ok.obj.handle_follows = TRUE;
        resp->res_u.ok.obj.handle.fh3_length = sizeof(fs_Handle);
        memcpy( &(resp->res_u.ok.obj.handle.fh3_u.data),
                &(fs->fileHandle),
                sizeof(fs_Handle));

        FS_ASSERT (fs->attr[0].attrType == FS_PRE_PAR_ATTRS);
        FS_ASSERT (fs->attr[2].attrType == FS_POST_PAR_ATTRS);
        fs_simDoGetWccData3(&(resp->res_u.ok.dir_wcc), &(fs->attr[0]), &(fs->attr[2]));

        FS_ASSERT (fs->attr[1].attrType == FS_POST_ATTRS);
        resp->res_u.ok.obj_attributes.attributes = TRUE;
        fs_simDoGetAttr3(&(resp->res_u.ok.obj_attributes.attr), &(fs->attr[1]));
    }
    else
    {
        resp->status = fs->hdr.fsStatus;

        if (fs->attr[0].attrType == FS_PRE_PAR_ATTRS &&
            fs->attr[1].attrType == FS_POST_PAR_ATTRS)
        {
            fs_simDoGetWccData3(&(resp->res_u.fail.dir_wcc), &(fs->attr[0]), &(fs->attr[1]));
        }
    }
    
    return resp;
}

int32
fs_simReadDirPlusReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgReadDirPlusReq     *fsMsg;
    READDIRPLUS3args         *args = (READDIRPLUS3args*)arg;
    nfs_fh3                  *fh;

    //printf("nfs ReadDirPlus req rcvd\n");

    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgReadDirPlusReq));
    fsMsg->hdr.requestType = FS_REQ_READDIRPLUS;
    fsMsg->hdr.attrType    = FS_REQ_ALL_ATTR;

    fh = &(args->dir);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->fileHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    fsMsg->maxDirBytes = args->dircount;
    fsMsg->maxBytes = args->maxcount;

    fs_simGetNfs_uint64 (fsMsg->cookie, args->cookie);
    FS_ASSERT (sizeof(fsMsg->cookieVerif) == sizeof(args->cookieverf));
    memcpy(&(fsMsg->cookieVerif), &(args->cookieverf), sizeof(args->cookieverf));

    *len = sizeof(fs_MsgReadDirPlusReq);
    return NFX_OK;
}

void *
fs_simReadDirPlusResp3(
    void    *fsResp)
{
    fs_MsgReadDirPlusResp      *fs;
    READDIRPLUS3res            *resp;
    entryplus3                 *ent;
    int32                      i;
    fs_DirEntryPlus            *dirent;
    int32                      nameLen;
    
    //printf("nfs ReadDirPlus resp called\n");

    fs = (fs_MsgReadDirPlusResp *) fsResp;
    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(READDIRPLUS3res));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;

        FS_ASSERT (fs->attr.attrType == FS_POST_ATTRS);
        resp->res_u.ok.dir_attributes.attributes = TRUE;
        fs_simDoGetAttr3(&(resp->res_u.ok.dir_attributes.attr), &(fs->attr));

        FS_ASSERT (sizeof(resp->res_u.ok.cookieverf) == sizeof(fs->cookieVerif));
        memcpy(&(resp->res_u.ok.cookieverf), &(fs->cookieVerif), sizeof(fs->cookieVerif));

        resp->res_u.ok.size = fs->numEntriesThis;            // ??
        resp->res_u.ok.maxcount = fs->numEntriesThis;
        // resp->res_u.ok.attributes = fs->numEntriesThis;   // ??
        // resp->res_u.ok.handles = fs->numEntriesThis;      // ??

        resp->res_u.ok.reply.entries = NULL;
        resp->res_u.ok.count = fs->numEntriesThis;

        resp->res_u.ok.reply.eof = fs->eof;

        if (fs->numEntriesThis)
        {
            ent = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(entryplus3) * fs->numEntriesThis);
            resp->res_u.ok.reply.entries = ent;
        }
        dirent = &(fs->dirEntry[0]);
        for (i = 0; i < fs->numEntriesThis; i++)
        {
            if (dirent->dirEntry.nameLen)
            {
                fs_simSetNfs_uint64 (ent[i].fileid, dirent->dirEntry.inode);
                nameLen = MIN(SFS_MAXNAMLEN, dirent->dirEntry.nameLen);
                memcpy(&(ent[i].name[0]), dirent->dirEntry.name, nameLen);
                ent[i].name[nameLen] = '\0';
                // ent->name = dirent->name;
                fs_simSetNfs_uint64 (ent[i].cookie, dirent->dirEntry.cookie);

                FS_ASSERT (dirent->attr.attrType == FS_POST_ATTRS);
                ent[i].name_attributes.attributes = TRUE;
                fs_simDoGetAttr3(&(ent[i].name_attributes.attr), &(dirent->attr));

                ent[i].name_handle.handle_follows = TRUE;
                ent[i].name_handle.handle.fh3_length = sizeof(fs_Handle);
                memcpy( &(ent[i].name_handle.handle.fh3_u.data),
                        &(dirent->handle),
                        sizeof(fs_Handle));
            }
            else
            {
                // early end of entries???
                break;
            }

            FS_READDIRPLUS_ADV_ENTRY_PTR(dirent);
        }

    }
    else
    {
        resp->status = fs->hdr.fsStatus;

        if (fs->attr.attrType == FS_POST_ATTRS)
        {
            resp->res_u.fail.dir_attributes.attributes = TRUE;
            fs_simDoGetAttr3(&(resp->res_u.fail.dir_attributes.attr), &(fs->attr));
        }
        else
        {
            resp->res_u.fail.dir_attributes.attributes = FALSE;
            //printf ("fs_simReadDirPlusResp3 Error  fs->attr.attrType != FS_POST_ATTRS\n");
        }
    }
    
    if (i != fs->numEntriesThis) {
        printf("i=%d, fs->numEntriesThis=%d\n", i, fs->numEntriesThis);
        // FS_ASSERT(i == fs->numEntriesThis);
    }

    return resp;
}

void
fs_simFreeReadDirPlusResp3(
    void    *fsResp)
{
    READDIRPLUS3res  *resp = fsResp;

    //printf("free readdir resp3\n");

    if (resp->status == NFS_OK)
    {
        if (resp->res_u.ok.reply.entries)
            fs_ramDealloc(FS_RAM_NFS_RESP_ARG, resp->res_u.ok.reply.entries);
    }
    
    fs_ramDealloc(FS_RAM_NFS_RESP_ARG, resp);

    //printf("  done free readdir resp3\n");

} /* fs_simFreeReaddirPlusResp3 */


int32
fs_simFsinfoReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgFsinfoReq     *fsMsg;
    FSINFO3args         *args = (FSINFO3args*)arg;
    nfs_fh3             *fh;

    //printf("nfs Fsinfo req rcvd\n");

    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgFsinfoReq));
    fsMsg->hdr.requestType = FS_REQ_FSINFO;
    fsMsg->hdr.attrType    = FS_REQ_ALL_ATTR;

    fh = &(args->fsroot);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->fsHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    *len = sizeof(fs_MsgFsinfoReq);
    return NFX_OK;
}

void *
fs_simFsinfoResp3(
    void    *fsResp)
{
    fs_MsgFsinfoResp      *fs;
    FSINFO3res            *resp;

    //printf("nfs Fsinfo resp called\n");

    fs = (fs_MsgFsinfoResp *) fsResp;
    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(FSINFO3res));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;

        FS_ASSERT (fs->attr.attrType == FS_POST_ATTRS);
        resp->res_u.ok.obj_attributes.attributes = TRUE;
        fs_simDoGetAttr3(&(resp->res_u.ok.obj_attributes.attr), &(fs->attr));

        resp->res_u.ok.rtmax = fs->fsinfo.readMax;
        resp->res_u.ok.rtpref = fs->fsinfo.readPref;
        resp->res_u.ok.rtmult = fs->fsinfo.readMult;
        resp->res_u.ok.wtmax = fs->fsinfo.writeMax;
        resp->res_u.ok.wtpref = fs->fsinfo.writePref;
        resp->res_u.ok.wtmult = fs->fsinfo.writeMult;
        resp->res_u.ok.dtpref = fs->fsinfo.readDirPref;
        fs_simSetNfs_uint64 (resp->res_u.ok.maxfilesize, fs->fsinfo.maxFileSize);
        resp->res_u.ok.time_delta.seconds = fs->fsinfo.timeDelta.secs;
        resp->res_u.ok.time_delta.nseconds = fs->fsinfo.timeDelta.nsecs;
        resp->res_u.ok.properties = fs->fsinfo.fsProperties;
    }
    else
    {
        resp->status = fs->hdr.fsStatus;

        if (fs->attr.attrType == FS_POST_ATTRS)
        {
            resp->res_u.fail.obj_attributes.attributes = TRUE;
            fs_simDoGetAttr3(&(resp->res_u.fail.obj_attributes.attr), &(fs->attr));
        }
    }
    
    return resp;
}

int32
fs_simPathConfReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgPathconfReq     *fsMsg;
    PATHCONF3args         *args = (PATHCONF3args*)arg;
    nfs_fh3               *fh;

    //printf("nfs PathConf req rcvd\n");

    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgPathconfReq));
    fsMsg->hdr.requestType = FS_REQ_PATHCONF;
    fsMsg->hdr.attrType    = FS_REQ_ALL_ATTR;

    fh = &(args->object);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->fsHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    *len = sizeof(fs_MsgPathconfReq);
    return NFX_OK;
}

void *
fs_simPathConfResp3(
    void    *fsResp)
{
    fs_MsgPathconfResp      *fs;
    PATHCONF3res            *resp;

    //printf("nfs PathConf resp called\n");

    fs = (fs_MsgPathconfResp *) fsResp;
    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(PATHCONF3res));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;

        FS_ASSERT (fs->attr.attrType == FS_POST_ATTRS);
        resp->res_u.ok.obj_attributes.attributes = TRUE;
        fs_simDoGetAttr3(&(resp->res_u.ok.obj_attributes.attr), &(fs->attr));

        resp->res_u.ok.link_max = fs->pathconf.linkMax;
        resp->res_u.ok.name_max = fs->pathconf.nameMax;
        resp->res_u.ok.no_trunc = fs->pathconf.noTrunc;
        resp->res_u.ok.chown_restricted = fs->pathconf.chownRestricted;
        resp->res_u.ok.case_insensitive = fs->pathconf.caseInsensitive;
        resp->res_u.ok.case_preserving = fs->pathconf.casePreserving;
    }
    else
    {
        resp->status = fs->hdr.fsStatus;

        if (fs->attr.attrType == FS_POST_ATTRS)
        {
            resp->res_u.fail.obj_attributes.attributes = TRUE;
            fs_simDoGetAttr3(&(resp->res_u.fail.obj_attributes.attr), &(fs->attr));
        }
    }
    
    return resp;
}

int32
fs_simCommitReq3(
    void    *req,
    void    *arg,
    int32   *len,
    void    *buf)
{
    fs_MsgCommitReq     *fsMsg;
    COMMIT3args         *args = (COMMIT3args*)arg;
    nfs_fh3             *fh;

    //printf("nfs Commit req rcvd\n");

    fsMsg = buf;
    bzero(fsMsg, sizeof(fs_MsgCommitReq));
    fsMsg->hdr.requestType = FS_REQ_COMMIT;
    fsMsg->hdr.attrType    = FS_REQ_ALL_ATTR;

    fh = &(args->file);
    FS_ASSERT (fh->fh3_length == sizeof(fs_Handle));
    memcpy(&(fsMsg->fsHandle), &(fh->fh3_u.data[0]), fh->fh3_length);

    fs_simGetNfs_uint64 (fsMsg->offset, args->offset);
    fsMsg->count = args->count;

    *len = sizeof(fs_MsgCommitReq);
    return NFX_OK;
}

void *
fs_simCommitResp3(
    void    *fsResp)
{
    fs_MsgCommitResp      *fs;
    COMMIT3res            *resp;

    //printf("nfs Commit resp called\n");

    fs = (fs_MsgCommitResp *) fsResp;
    resp = fs_ramAlloc(FS_RAM_NFS_RESP_ARG, sizeof(COMMIT3res));
    bzero (resp, sizeof(COMMIT3res));
    if (fs->hdr.fsStatus == NFX_OK)
    {
        resp->status = NFS_OK;

        FS_ASSERT (fs->attr[0].attrType == FS_PRE_ATTRS);
        FS_ASSERT (fs->attr[1].attrType == FS_POST_ATTRS);
        fs_simDoGetWccData3(&(resp->res_u.ok.file_wcc), &(fs->attr[0]), &(fs->attr[1]));

        // Need bzero, FS_ASSERT (sizeof(resp->res_u.ok.verf) == sizeof(fs->verifier));
        memcpy(&(resp->res_u.ok.verf), &(fs->verifier), sizeof(fs->verifier));
    }
    else
    {
        resp->status = fs->hdr.fsStatus;

        if ((fs->attr[0].attrType == FS_PRE_ATTRS) &&
            (fs->attr[1].attrType == FS_POST_ATTRS))
        {
            fs_simDoGetWccData3(&(resp->res_u.fail.file_wcc), &(fs->attr[0]), &(fs->attr[1]));
        }
        else
        {
            resp->res_u.fail.file_wcc.before.attributes = FALSE;
            resp->res_u.fail.file_wcc.after.attributes = FALSE;
            //printf ("fs_simCommitResp3 Error  fs->attr[0..1].attrType != FS_PRE_ATTRS..FS_POST_ATTRS\n");
        }
    }
    
    return resp;
} // fs_simCommitResp3

void
fs_simFreeResp3(
    void    *resp)
{
    fs_ramDealloc(FS_RAM_NFS_RESP_ARG, resp);

} /* fs_simFreeResp3 */

nfx_spinlock_t atomic_bitset;
int atomic_bitset_init;

uint32 ATOMIC_OR_INT(uint32 *p, uint32 v)
{
	uint32 tmp;

	if (!atomic_bitset_init) {
		atomic_bitset_init = 1;
		NFX_SPIN_INIT(&atomic_bitset, SPINLOCK_LEVEL_UNKNOWN);
	}

	NFX_SPIN_LOCK(&atomic_bitset);
	tmp = *p;
	*p |= v;
	NFX_SPIN_UNLOCK(&atomic_bitset);

	return tmp;
}

uint32 ATOMIC_AND_INT(uint32 *p, uint32 v)
{
	uint32 tmp;

	if (!atomic_bitset_init) {
		atomic_bitset_init = 1;
		NFX_SPIN_INIT(&atomic_bitset, SPINLOCK_LEVEL_UNKNOWN);
	}

	NFX_SPIN_LOCK(&atomic_bitset);
	tmp = *p;
	*p &= v;
	NFX_SPIN_UNLOCK(&atomic_bitset);

	return tmp;
}

/*-----------------------------------------------------------------------------
 */

#endif // SSC
