// -*- tab-width: 4; c-basic-offset: 4; indent-tabs-mode: nil -*-
// ex: set tabstop=4 shiftwidth=4 expandtab:

/*
 * Name:  fs-snap.c
 *
 * RCS:  $Id: fs-snap.c,v 1.119 2002/12/05 21:08:57 bob Exp $
 *
 * Description:  Snapshots.
 * 
 * Created By:  Paul Haahr <haahr@agilestorage.com>
 *
 * Date Created:  2001-09-21
 */

#if !defined(SSC)

#include "fs-includes.h"

/*
 * uncomment to combine the 2 snap remove phases:
 * FS_SNAP_PHASE_REMOVE_SNAP_BITS and FS_SNAP_PHASE_REMOVE_FREE_BLOCKS
 * into a single FS_SNAP_PHASE_REMOVE_SNAP_BITS_AND_FREE
 */
#define FS_SNAP_OPTIMIZE_REMOVE_FREE_BLOCKS

/*
 * uncomment for optimized version:
 * 1. manipulate refbytes in 64 bits (vs 8 bits)
 * 2. "expand transform" function calls
 */
#define FS_SNAP_OPTIMIZED

/*
 * uncomment for prefetching refbytes and snapbits pages
 * before freezing the filesystem for snapshot operations
 */
#define FS_SNAP_READAHEAD

/*
 * uncomment for optimized cache handling:
 * 1. flush blocks from the buffer cache when a remove would deallocate them
 * 2. re-enable cow in both buffer and inode caches after a snapshot operation
 * 3. flush buffer and inode cache by snapnum after a snapshot remove operation
 *    (vs flusing both caches after any snapshot operation)
 */
#define FS_SNAP_CACHE_OPTIMIZED

#ifdef FS_DEBUG_SNAP_CRASH

typedef enum {
    FS_SNAP_PHASE_DURING_REVERT_BUDDY = FS_SNAP_PHASE_REVERT_CLEAR_SNAP_BITS + 1,
    FS_SNAP_PHASE_DURING_REVERT_INODES,
} fs_SnapCrashPhase;

extern int fs_snapPhaseToCrash;
extern void fs_debugCrash();

#endif // FS_DEBUG_SNAP_CRASH

static void fs_snapPrintStatus(fs_Context *context, int32 level, char *fmt, ...);
fs_SnapStats fs_snapStats;


// used as arg0 for emap_walk

typedef struct {
    uint64 *blocks;         // list of blocks including inner nodes
    uint32 blockIndex;      // current block index for blocks
    uint32 maxBlocks;       // maxinum number of blocks for inode
} fs_SnapRemoveWorkerArg;

/*
 * Watchdog timeout support
 */

static eeeRTC_t fs_snapLastSleepTime;

#define FS_SNAP_WATCHDOG_DECL       \
    eeeRTC_t _beginTime, _delta;

#define FS_SNAP_WATCHDOG_INIT(name) \
    eee_rtc(&_beginTime);

#define FS_SNAP_WATCHDOG_CHECK(name)                                    \
    eee_rtcDuration(&_beginTime, &_delta);                              \
    if (eee_rtcCompare(&_delta, &fs_incoreTune->maxCpuUsageRTC) > 0) {  \
        eee_rtc(&fs_snapLastSleepTime);                                 \
        fs_sleep(context, 10);                                          \
        eee_rtc(&_beginTime);                                           \
    }

/*
 * Performance tuning support
 */

// uncomment for snapshot tuning
//#define FS_SNAP_TUNE

#ifdef FS_SNAP_TUNE
#define FS_SNAP_TUNE_DECL               \
    eeeRTC_t _tune_time, _tune_delta;
#else
#define FS_SNAP_TUNE_DECL
#endif

#ifdef FS_SNAP_TUNE
#define FS_SNAP_START_TUNE(name)        \
    eee_rtc(&_tune_time);               \
    printf("%s: start %s\n", eee_rtcStr(&_tune_time), name);
#else
#define FS_SNAP_START_TUNE(name)
#endif

#ifdef FS_SNAP_TUNE
#define FS_SNAP_END_TUNE(name)                  \
    eee_rtcDuration(&_tune_time, &_tune_delta); \
    eee_rtc(&_tune_time);                       \
    printf("%s: end %s duration %lld ms\n",     \
            eee_rtcStr(&_tune_time), name, eee_rtcToMS(&_tune_delta));
#else
#define FS_SNAP_END_TUNE(name)
#endif


/*
 * Return true if phase is a revert phase
 */
static boolean fs_snapIsRevertPhase(fs_SnapPhase phase)
{
    switch (phase) {
        case FS_SNAP_PHASE_REVERT_REMOVE_SNAPS:
        case FS_SNAP_PHASE_REVERT_REF_COUNTS:
        case FS_SNAP_PHASE_REVERT_BUDDY:
        case FS_SNAP_PHASE_REVERT_INODE_BLOCKS:
        case FS_SNAP_PHASE_REVERT_META:
        case FS_SNAP_PHASE_REVERT_CLEAR_SNAP_BITS:
            return TRUE;
        default:
            return FALSE;
    }
}


/**
 * Quiesce the file system block out modify access
 * For snap create and remove
 */
void fs_freeze(
    fs_Context *context)
{
    // grab the lock
    fs_lockExclusive(context, &context->fileSystem->modifyLock, NULL);

    // sync the inode and buffer caches.
    fs_syncThread(context);
    fs_syncThread(context);
    fs_syncThread(context);
}


/**
 * Unquiesce the file system, allow modify access
 */
void fs_thaw(
    fs_Context *context)
{
    // release the lock
    fs_unlockExclusive(context, &context->fileSystem->modifyLock);
}


/**
 * Quiesce the file system block out all access
 * For snap revert
 */
void fs_freezeAll(
    fs_Context *context)
{
    // grab the access lock
    fs_lockExclusive(context, &context->fileSystem->accessLock, NULL);

    fs_freeze(context);
}


/**
 * Unquiesce the file system, allow all access
 */
void fs_thawAll(
    fs_Context *context)
{
    // release the lock
    fs_unlockExclusive(context, &context->fileSystem->accessLock);

    fs_thaw(context);
}


/**
 * flush the txrx attribute cache
 * if snap, invalidate all snap entries (FS_CACHE_INV_TYPE_SNAP)
 * if inum, invalidate inum (FS_CACHE_INV_TYPE_INUM)
 */
void fs_snapFlushAttrCache(
    fs_Context *context, int32 snap, uint64 inum)
{
#ifndef FS_STANDALONE
    int32 rc;
    fs_MsgCacheInvReq msg;

    msg.hdr.requestType = FS_REQ_CACHE_INV;
    msg.hdr.len = sizeof(msg);

    // TBD: do the following when supported on txrx
    // set handle's vol, snap & inum
    // if snap, invType is FS_CACHE_INV_TYPE_SNAP
    // if inum == .snapshot, set handle's gen to magic and inum root

    msg.invType = FS_CACHE_INV_TYPE_ALL;

    rc = eee_sendMessage((void *) &msg, sizeof(msg),
                         PACK_UNICAST_PORT_ADDRESS(nfxMySlotId,
                                                   NFX_NFP_TXRX_ANY,
                                                   eee_getApplicationId("nfs_srvr"),
                                                   EEE_CONTROL,
                                                   NFX_COS_HIGH,
                                                   0),
                         fs_packedPortAddr(nfxMyCPUId, EEE_CONTROL));
#endif
}


//   S N A P S H O T S


/**
 * Phase names, for debugging purposes.
 */
static char *fs_snapPhaseName[] = {
    "ready",
    "create:alloc",
    "create:snap-bits",
    "create:ref-counts",
    "create:meta",
    "remove:ref-counts",
    "remove:snap-bits",
    "remove:snap-bits-and-free",
    "remove:free-blocks",
    "remove:free-snap-bits",
    "revert:remove-snaps",
    "revert:ref-counts",
    "revert:buddy",
    "revert:inode-blocks",
    "revert:meta",
    "revert:free-snap-bits"
};


/**
 * Allocate a free snapshot id from the snapshots bitmap in the superblock.
 * Modifies the in-core superblock, but doesn't write it to disk.  Returns
 * a negative number if no snapshot id is available.
 */
static int fs_snapAllocateId(
    fs_Context *context)
{
    fs_SuperBlock *super = context->fileSystem->super;
    uint64 freebits = ~super->snapshots & FS_SNAP_USER_ID_MASK;
    int id;

    if (freebits == 0) {
        FS_TRACE(context, FS_TRACE_SNAP,
                 ("snap allocate id: max user snapshots reached"));
        return -1;
    } else {
        uint64 bit = leastBit64(freebits);
        id = log2_64(bit);
        FS_TRACE(context, FS_TRACE_SNAP,
                ("snap allocate id: id %d snapshots %llx freebits %llx bit %llx",
                id, super->snapshots, freebits, bit));
        NFX_SPIN_LOCK(&context->fileSystem->superLock);
        super->snapshots |= bit;
        NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
        return id;
    }
}


/**
 * Enable or disable snapshots
 */
void fs_snapState(
    fs_Context *context,
    boolean enable)
{
    fs_SnapState *state = &context->fileSystem->super->snapState;

    NFX_SPIN_LOCK(&context->fileSystem->superLock);
    if (!state->times.maxWeekly && !state->times.maxDaily && !state->times.maxHourly) {
        state->times.maxWeekly = FS_SNAP_DEFAULT_MAX_WEEKLY;
        state->times.maxDaily = FS_SNAP_DEFAULT_MAX_DAILY;
        state->times.maxHourly = FS_SNAP_DEFAULT_MAX_HOURLY;
        state->times.hours = FS_SNAP_DEFAULT_HOURS;
    }
    state->times.enable = enable;
    NFX_SPIN_UNLOCK(&context->fileSystem->superLock);

    // write the superblock(s)

   fs_writeSuperblocksSync(context, context->fileSystem->super);

   fs_snapPrintStatus(context, LOG_NOTICE,
                      "%s snapshot scheduling %s",
                      context->fileSystem->super->mountName,
                      (enable ? "enabled" : "disabled"));
}


/**
 * Get the state of the snapshot creation process in the file system's
 * superblock(s).
 */
static fs_SnapPhase fs_snapGetState(
    fs_Context *context,
    int *id)
{
    fs_SnapPhase phase;
    fs_SnapState *state = &context->fileSystem->super->snapState;

    NFX_SPIN_LOCK(&context->fileSystem->superLock);
    *id = state->id;
    phase = state->phase;
    NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
    return phase;
}


/**
 * Set the state of the snapshot creation process in the file system's
 * superblock(s).
 */
static void fs_snapSetState(
    fs_Context *context,
    int id,
    fs_SnapPhase phase)
{
    fs_SnapState *state = &context->fileSystem->super->snapState;

    assert((id == 0) == (phase == FS_SNAP_PHASE_READY));
    assert(id == 0 || state->id == 0 || state->id == id);
    FS_TRACE(context, FS_TRACE_SNAP,
             ("snapSetState %s %d -> %s %d",
              fs_snapPhaseName[state->phase], state->id,
              fs_snapPhaseName[phase], id));
    NFX_SPIN_LOCK(&context->fileSystem->superLock);
    state->id = id;
    state->phase = phase;
    NFX_SPIN_UNLOCK(&context->fileSystem->superLock);

    // write the superblock(s)

    fs_writeSuperblocksSync(context, context->fileSystem->super);
}


#ifndef FS_SNAP_OPTIMIZED
/**
 * General routine that process snapbits and write ref bytes
 * transform is the routine that actually does the work
 */
static int32 fs_snapTransformRefBytes(
    fs_Context *context,
    int        id,
    boolean    (*transform)(fs_Context *context,
                            uint8 *snapBits,
                            fs_RefByte *ref,
                            uint64 blockStart))
{
    fs_Parameters *params = &context->fileSystem->params;
    uint32 blockSize = params->blockSize;
    uint32 chunkSize = blockSize / BITS_PER_BYTE;
    uint64 snapBitsPages;
    uint64 page, block;
    FS_SNAP_WATCHDOG_DECL

    fs_Inode    *snapBitsInode = I_GET_SHARED(id, FS_SNAPSHOT_BITS_INODE);
    fs_Inode    *refInode = I_GET_SHARED(0, FS_REF_COUNT_INODE);

    assert(refInode != NULL);

    // snap resume support
    if (snapBitsInode == NULL) {
        FS_TRACE(context, FS_TRACE_SNAP,
                 ("snap transform ref bytes: no snapbits inode"));
        fs_inodeRelease(context, refInode);
        return NFX_OK;
    }

    snapBitsPages = ceilingDiv64(snapBitsInode->dinode.size, blockSize);
    assert(snapBitsPages == snapBitsInode->dinode.numblocks);

    FS_SNAP_WATCHDOG_INIT("transform refbytes");

    for (page = 0; page < snapBitsPages; page++) {
        int i;
        fs_Buf *snapBuf = O_GET_SHARED(snapBitsInode, page);

        assert(snapBuf != NULL);

        // yield the CPU if necc to prevent watchdog timeouts
        FS_SNAP_WATCHDOG_CHECK("transform refbytes");

        block = (page * blockSize);
        for (i = 0; i < BITS_PER_BYTE; i++) {
            boolean changed = FALSE;
            uint8 *bits = &snapBuf->u.data[i * chunkSize];
            fs_Buf *refBuf = O_GET_EXCL(refInode, page * BITS_PER_BYTE + i);

            if (refBuf == NULL) {
                // nearing the end of the file
                assert(page + 1 == snapBitsPages);
            }
            else {
                changed = transform(context, bits, refBuf->u.refBytes, block);
            }
            if (changed)
                FS_BUF_WRITE_NO_LOG(refBuf, 0, blockSize);
            if (refBuf)
                fs_bufReleaseExclusive(context, refBuf);
        }
        fs_bufRelease(context, snapBuf);
    }

    fs_inodeRelease(context, refInode);
    fs_inodeRelease(context, snapBitsInode);

    return NFX_OK;
}
#endif


#ifndef FS_SNAP_OPTIMIZED
/**
 * General routine that process ref bytes and write snapbits
 * transform is the routine that actually does the work
 */
static int32 fs_snapTransformSnapBits(
    fs_Context *context,
    int        id,
    boolean    (*transform)(fs_Context *context,
                            uint8 *snapBits,
                            fs_RefByte *ref,
                            uint64 blockStart))
{
    fs_Parameters *params = &context->fileSystem->params;
    uint32 blockSize = params->blockSize;
    uint32 chunkSize = blockSize / BITS_PER_BYTE;
    uint64 snapBitsPages;
    uint64 page, block;
    FS_SNAP_WATCHDOG_DECL

    fs_Inode *refInode = I_GET_SHARED(0, FS_REF_COUNT_INODE);
    fs_Inode *snapBitsInode = I_GET_SHARED(id, FS_SNAPSHOT_BITS_INODE);

    assert(refInode != NULL);

    // snap resume support
    if (snapBitsInode == NULL) {
        FS_TRACE(context, FS_TRACE_SNAP,
                 ("snap transform snapbits: no snapbits inode"));
        fs_inodeRelease(context, refInode);
        return NFX_OK;
    }

    snapBitsPages = ceilingDiv64(snapBitsInode->dinode.size, blockSize);
    assert(snapBitsPages == snapBitsInode->dinode.numblocks);

    FS_SNAP_WATCHDOG_INIT("transform snapbits");

    for (page = 0; page < snapBitsPages; page++) {
        int i;
        fs_Buf *snapBuf = O_GET_EXCL(snapBitsInode, page);

        // yield the CPU if necc to prevent watchdog timeouts
        FS_SNAP_WATCHDOG_CHECK("transform snapbits");

        // mark the live bits

        assert(snapBuf != NULL);
        block = (page * blockSize);
        for (i = 0; i < BITS_PER_BYTE; i++) {
            uint8 *bits = &snapBuf->u.data[i * chunkSize];
            fs_Buf *refBuf = O_GET_SHARED(refInode, page * BITS_PER_BYTE + i);
            if (refBuf == NULL) {
                // nearing the end of the file
                assert(page + 1 == snapBitsPages);
                // zero out remaining chunks
                bzero(bits, (BITS_PER_BYTE - i) * chunkSize);
            }
            else {
                transform(context, bits, refBuf->u.refBytes, block);
                fs_bufRelease(context, refBuf);
            }
        }

        // dirty and release the buffer

        FS_BUF_WRITE_NO_LOG(snapBuf, 0, blockSize);
        fs_bufReleaseExclusive(context, snapBuf);
    }

    fs_inodeRelease(context, refInode);
    fs_inodeRelease(context, snapBitsInode);

    return NFX_OK;
}
#endif


#ifndef FS_SNAP_OPTIMIZED
/**
 * Given snapnum and blocknum, return corresponding snapbit
 *  - snapBitsInode's lock must be locked shared
 */
static int32 fs_snapGetSnapBit(
        fs_Context *context, fs_Inode *snapBitsInode,
        uint64 block, boolean *bit)
{
    fs_Parameters *params = &context->fileSystem->params;
    uint32 blockSize = params->blockSize;
    uint8 *snapbits;
    uint64 bitnum;
    fs_Buf *buf = O_GET_SHARED(snapBitsInode, block / (blockSize * BITS_PER_BYTE));

    if (buf == NULL) {
        *bit = 0;
        return NFX_OK;  // growfs extras
    }
    snapbits = buf->u.data;    
    bitnum = (block % (blockSize * BITS_PER_BYTE));
    *bit = (snapbits[bitnum >> 3] >> (bitnum & 0x7)) & 0x1;
    fs_bufRelease(context, buf);

    return NFX_OK;
}
#endif


/**
 * Recalculate ref counts from all snapbits file
 */
static void fs_snapRecalculateRefCnts(
        fs_Context *context)
{
#ifdef FS_SNAP_OPTIMIZED 
    fs_Parameters *params = &context->fileSystem->params;
    uint64 page, snapBitsPages;
    uint32 blockSize = params->blockSize;
    uint32 chunkSize = blockSize / BITS_PER_BYTE;
    uint32 id;
    fs_Inode *refInode;
    boolean initRefbytes = TRUE;

    FS_SNAP_WATCHDOG_DECL
    FS_SNAP_TUNE_DECL

    FS_SNAP_START_TUNE("recal snapcounts for refbytes");

    refInode = I_GET_SHARED(0, FS_REF_COUNT_INODE);
    assert(refInode);

    for (id = 1; id < FS_MAX_SNAPSHOTS; id++) {
        fs_Inode *snapBitsInode = I_GET_SHARED(id, FS_SNAPSHOT_BITS_INODE);
        if (snapBitsInode == NULL)
            continue;

        snapBitsPages = ceilingDiv64(snapBitsInode->dinode.size, blockSize);
        assert(snapBitsPages == snapBitsInode->dinode.numblocks);

        FS_SNAP_WATCHDOG_INIT("recal refbytes");

        for (page = 0; page < snapBitsPages; page++) {
            int i;
            fs_Buf *snapBuf = O_GET_SHARED(snapBitsInode, page);

            assert(snapBuf != NULL);

            // yield the CPU if necc to prevent watchdog timeouts
            FS_SNAP_WATCHDOG_CHECK("recal refbytes");

            for (i = 0; i < BITS_PER_BYTE; i++) {
                uint64 changed = 0;
                uint8 *snapBits = &snapBuf->u.data[i * chunkSize];
                uint64 refPage = page * BITS_PER_BYTE + i;
                fs_Buf *refBuf = O_GET_EXCL(refInode, refPage);
                fs_RefByte *refbytes;

                if (refBuf == NULL) {
                    // nearing the end of the file
                    assert(page + 1 == snapBitsPages);
                }
                else {
                    int j, k;
                    uint64 word, snapcount;

                    refbytes = refBuf->u.refBytes;
                    for (j = 0, k = 0; j < blockSize / BITS_PER_BYTE; j += 4, k += 4*sizeof (uint64)) {
                        uint64 live, live_mask, never_mask;
                        uint8 bits;

                        live_mask = ((uint64)FS_REFBYTE_IS_LIVE
                                    | ((uint64)FS_REFBYTE_IS_LIVE << 8)
                                    | ((uint64)FS_REFBYTE_IS_LIVE << 16)
                                    | ((uint64)FS_REFBYTE_IS_LIVE << 24)
                                    | ((uint64)FS_REFBYTE_IS_LIVE << 32)
                                    | ((uint64)FS_REFBYTE_IS_LIVE << 40)
                                    | ((uint64)FS_REFBYTE_IS_LIVE << 48)
                                    | ((uint64)FS_REFBYTE_IS_LIVE << 56));
                        never_mask = ((uint64)FS_REFBYTE_NEVER_SNAPSHOT
                                    |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 8)
                                    |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 16)
                                    |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 24)
                                    |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 32)
                                    |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 40)
                                    |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 48)
                                    |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 56));

#define CLEAR_SNAP_COUNT do{\
    live = word & (live_mask | never_mask); /* keep only live and never snap bits */\
    word = live;            \
} while (0)

                        // need to clear the snapcount the first go around

                        if (initRefbytes) {
                            word = *(uint64 *)&refbytes[k];
                            if (word != 0) {
                                CLEAR_SNAP_COUNT;
                                *(uint64 *)&refbytes[k] = word;
                            }

                            word = *(uint64 *)&refbytes[k+8];
                            if (word != 0) {
                                CLEAR_SNAP_COUNT;
                                *(uint64 *)&refbytes[k+8] = word;
                            }

                            word = *(uint64 *)&refbytes[k+16];
                            if (word != 0) {
                                CLEAR_SNAP_COUNT;
                                *(uint64 *)&refbytes[k+16] = word;
                            }

                            word = *(uint64 *)&refbytes[k+24];
                            if (word != 0) {
                                CLEAR_SNAP_COUNT;
                                *(uint64 *)&refbytes[k+24] = word;
                            }
                        } // if initRefbytes
/*
 * The idea is to "expand" a byte of snapbits to 64 bit value.
 * This value is then shifted to the snapcount position and
 * added to the existing snapcount.
 */

#define INCREMENT_SNAP_COUNT do{\
    snapcount = ((uint64) (bits & 0x1))                     \
                | ((uint64) ((bits & 0x2) ? 1 : 0) << 8)    \
                | ((uint64) ((bits & 0x4) ? 1 : 0) << 16)   \
                | ((uint64) ((bits & 0x8) ? 1 : 0) << 24)   \
                | ((uint64) ((bits & 0x10) ? 1 : 0) << 32)  \
                | ((uint64) ((bits & 0x20) ? 1 : 0) << 40)  \
                | ((uint64) ((bits & 0x40) ? 1 : 0) << 48)  \
                | ((uint64) ((bits & 0x80) ? 1 : 0) << 56); \
    snapcount <<= FS_REFBYTE_SNAP_COUNT_SHIFT;  \
    word += snapcount;          /* decrement snap counts            */\
    changed |= snapcount;       /* nonzero "changed" means block is dirty */\
} while (0)

                        bits = snapBits[j];

                        if (bits != 0) {
                            word = *(uint64 *)&refbytes[k];
                            INCREMENT_SNAP_COUNT;
                            *(uint64 *)&refbytes[k] = word;
                        }
                        bits = snapBits[j+1];
                        if (bits != 0) {
                            word = *(uint64 *)&refbytes[k+8];
                            INCREMENT_SNAP_COUNT;
                            *(uint64 *)&refbytes[k+8] = word;
                        }
                        bits = snapBits[j+2];
                        if (bits != 0) {
                            word = *(uint64 *)&refbytes[k+16];
                            INCREMENT_SNAP_COUNT;
                            *(uint64 *)&refbytes[k+16] = word;
                        }
                        bits = snapBits[j+3];
                        if (bits != 0) {
                            word = *(uint64 *)&refbytes[k+24];
                            INCREMENT_SNAP_COUNT;
                            *(uint64 *)&refbytes[k+24] = word;
                        }
                    }
                }
                if (changed)
                    FS_BUF_WRITE_NO_LOG(refBuf, 0, blockSize);
                if (refBuf)
                    fs_bufReleaseExclusive(context, refBuf);
            }
            fs_bufRelease(context, snapBuf);
        } // for all snapbits pages

        fs_inodeRelease(context, snapBitsInode);
        initRefbytes = FALSE;

    } // for all snaps

    fs_inodeRelease(context, refInode);

    FS_SNAP_END_TUNE("recal snapcounts for refbytes");
    return;

#else

    fs_Parameters *params = &context->fileSystem->params;
    uint32 blockSize = params->blockSize;
    uint64 page, refBytePages = params->refbytesBlockCount;
    uint32 i, id, nSnaps = 0;
    uint64 block = 0;
    char status_string[128];
    FS_SNAP_WATCHDOG_DECL
    FS_SNAP_TUNE_DECL

    fs_Inode *refInode = I_GET_EXCL(0, FS_REF_COUNT_INODE);
    fs_Inode *snapBitsInodes[FS_MAX_SNAPSHOTS];

    for (id = 0; id < FS_MAX_SNAPSHOTS; id++) {
        snapBitsInodes[id] = I_GET_SHARED(id, FS_SNAPSHOT_BITS_INODE);
        if (snapBitsInodes[id]) {
            nSnaps++;
            FS_TRACE(context, FS_TRACE_SNAP,
                    ("snap recal refcnts: found snapnum %d", id));
        }
    }

    sprintf(status_string, "snap recal refcnts: begin (pages %lld nSnaps %d)",
            refBytePages, nSnaps);
    fs_snapPrintStatus(context, LOG_INFO, status_string);

    FS_SNAP_WATCHDOG_INIT("recal refcnts");

    for (page = 0; page < refBytePages; page++, block += blockSize) {
        fs_Buf *refBuf = O_GET_EXCL(refInode, page);
        fs_RefByte *ref;

        assert(refBuf != NULL);
        ref = refBuf->u.refBytes;
        // printf("  [%lld] block %lld\n", page, fs_blockNumber(refBuf->block));

        if (page == 0) {
            FS_SNAP_START_TUNE("snap recal refcnts for page 0");
        }

        for (i = 0; i < blockSize; i++) {
            boolean changed = 0;
            int32 snapcount = 0;
            fs_RefByte byte = ref[i];
            boolean bit;

            // yield the CPU if necc to prevent watchdog timeouts
            FS_SNAP_WATCHDOG_CHECK("recal refcnts");

            for (id = 0; id < FS_MAX_SNAPSHOTS; id++) {
                if (snapBitsInodes[id] == NULL) {
                    continue;
                }
                if (fs_snapGetSnapBit(context, snapBitsInodes[id],
                                      block + i, &bit) == NFX_OK) {
                    if (bit) {
                        changed = 1;
                        snapcount++;
                    }
                }
            }
            ref[i] = fs_refbyteSetSnapCount(byte, snapcount);
        }
        // write out the ref buf and release
        FS_BUF_WRITE_NO_LOG(refBuf, 0, blockSize);
        fs_bufReleaseExclusive(context, refBuf);

        if (page == 0) {
            FS_SNAP_END_TUNE("snap recal refcnts for page 0");
        }
    }

    fs_inodeReleaseExclusive(context, refInode);

    for (id = 0; id < FS_MAX_SNAPSHOTS; id++) {
        if (snapBitsInodes[id]) {
            fs_inodeRelease(context, snapBitsInodes[id]);
        }
    }

    fs_snapPrintStatus(context, LOG_INFO, "snap recal refcnts: end");
#endif
}


//    S N A P S H O T   C R E A T E


/**
 * Snapshot creation:  allocating snap bits and sync'ing.
 */
static int32 fs_snapCreateAllocSnapBits(
    fs_Context *context,
    int id)
{
    fs_Parameters *params = &context->fileSystem->params;
    fs_Buf *inodeBuf;
    fs_Inode *inode;
    uint64 page;
    uint64 snapBitsPages
        = ceilingDiv64(params->totalBlocks, params->blockSize * BITS_PER_BYTE);
    
    // if we crashed during a snap create, the inode might or might
    // not have been created, so check for it via I_GET_SHARED()

    inode = I_GET_SHARED(id, FS_SNAPSHOT_BITS_INODE);
    if (inode) {
        FS_TRACE(context, FS_TRACE_SNAP,
                ("snapbits file: already created for id %d pages %lld size %lld",
                id, snapBitsPages, inode->dinode.size));
        fs_inodeRelease(context, inode);
        return NFX_OK;
    }

    // create a transaction context

    assert(context->logSummary == NULL);
    fs_logBeginTransaction(context);

    // create the inode, allocate space for it
    // do for both in-core and on-disk copies

    inode = fs_inodeInstantiateMetanode(context, id, FS_SNAPSHOT_BITS_INODE);
    inode->dinode.size = ceilingDiv64(params->totalBlocks, BITS_PER_BYTE);
    inode->dinode.attrs = FS_IFMT_REG;
    inode->dinode.genNum += 1;
    inode->dinode.linkCnt = 1;
    inode->dinode.ctime.secs = fs_getTimeOfDaySecs();
    inode->dinode.version++;

    inodeBuf = B_GET_EXCL(NULL, fs_firstInodeBlock(id));
    assert(inodeBuf);
    inodeBuf->u.dinode->dinode[FS_SNAPSHOT_BITS_INODE] = inode->dinode;
    FS_BUF_WRITE(inodeBuf, (FS_SNAPSHOT_BITS_INODE * sizeof (fs_Dinode)), sizeof (fs_Dinode));

#if defined(FS_DEBUG_SNAP_CRASH)
    if (fs_snapPhaseToCrash != -1) {
        if (fs_snapPhaseToCrash == FS_SNAP_PHASE_CREATE_ALLOC) {
            printf("snap crash test: %d %s\n",
                        fs_snapPhaseToCrash, fs_snapPhaseName[fs_snapPhaseToCrash]);
            fs_debugCrash();
        }
    }
#endif

    fs_emaproot_init(context, inode);
    for (page = 0; page < snapBitsPages; ) {
        uint64 remaining = snapBitsPages - page;
        fs_Extent extent
            = fs_allocateExtent(context, inode, page, remaining, FALSE);
        assert(fs_extentLength(extent) > 0);
        // TODO - handle allocation failure by aborting snapshot
        FS_TRACE(context, FS_TRACE_SNAP,
                 ("snapbits file: page %lld extent start %lld length %d",
                 page, fs_blockNumber(fs_extentStartBlock(extent)),
                 fs_extentLength(extent)));
        fs_emap_add(context, inode, page, extent);
        inode->dinode.numblocks += fs_extentLength(extent);
        page += fs_extentLength(extent);
    }

    assert(snapBitsPages == inode->dinode.numblocks);
    fs_inodeWrite(context, inode, INO_WRITE_LOG);

    // commit and release the transaction context

    fs_logCommitEntry(context);
    context->logSummary = NULL;

    return NFX_OK;
}


#ifndef FS_SNAP_OPTIMIZED
/**
 * Compress the live indication from the reference bytes for a single
 * block into an array of bits for a snap bits file.
 */
static boolean fs_snapCompressLiveBits(
    fs_Context *context,
    uint8      *snapbits,
    fs_RefByte *ref,
    uint64     blockStart)
{
    boolean changed = FALSE;
    int i, n;
    uint32 blockSize = context->fileSystem->params.blockSize;

    assert((blockSize % BITS_PER_BYTE) == 0);

    for (i = 0, n = blockSize / BITS_PER_BYTE; i < n; i++) {
        int j;
        uint8 bits = 0;
        uint8 bit;
        for (j = 0, bit = 1; j < BITS_PER_BYTE; j++, bit <<= 1) {
            fs_RefByte byte = *ref++;
            
            if ((byte & (FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT))
                == FS_REFBYTE_IS_LIVE)
            {
                bits |= bit;
                changed = TRUE;
                fs_snapStats.snapBitBlocks++;
            }
        }
        snapbits[i] = bits;
    }
    return changed;
}
#endif


/**
 * A sentinel value for a block past the end of the file system.
 */
#define LAST_BLOCK 0xffffffffffffffffULL


/**
 * Snapshot creation:  reading reference counts, writing snap bits.
 */
static int32 fs_snapCreateWriteSnapBits(
    fs_Context *context,
    int        id)
{
#ifdef FS_SNAP_OPTIMIZED 
    fs_Parameters *params = &context->fileSystem->params;
    uint32 blockSize = params->blockSize;
    uint32 chunkSize = blockSize / BITS_PER_BYTE;
    uint64 snapBitsPages;
    uint64 page, block;
    FS_SNAP_WATCHDOG_DECL
    FS_SNAP_TUNE_DECL

    fs_Inode *refInode = I_GET_SHARED(0, FS_REF_COUNT_INODE);
    fs_Inode *snapBitsInode = I_GET_SHARED(id, FS_SNAPSHOT_BITS_INODE);

    assert(refInode != NULL);

    // snap resume support
    if (snapBitsInode == NULL) {
        FS_TRACE(context, FS_TRACE_SNAP,
                 ("create write snapbits: no snapbits inode"));
        fs_inodeRelease(context, refInode);
        return NFX_OK;
    }

    snapBitsPages = ceilingDiv64(snapBitsInode->dinode.size, blockSize);
    assert(snapBitsPages == snapBitsInode->dinode.numblocks);

    FS_SNAP_START_TUNE("create write snapbits");
    FS_SNAP_WATCHDOG_INIT("create write snapbits");

    for (page = 0; page < snapBitsPages; page++) {
        int i;
        fs_Buf *snapBuf = O_GET_EXCL(snapBitsInode, page);

        // yield the CPU if necc to prevent watchdog timeouts
        FS_SNAP_WATCHDOG_CHECK("create write snapbits");

        assert(snapBuf != NULL);
        block = (page * blockSize);
        for (i = 0; i < BITS_PER_BYTE; i++) {
            uint8 *bits = &snapBuf->u.data[i * chunkSize];
            uint64 refPage = page * BITS_PER_BYTE + i;
            fs_Buf *refBuf = O_GET_SHARED(refInode, refPage);

            if (refBuf == NULL) {
                // nearing the end of the file
                assert(page + 1 == snapBitsPages);
                // zero out remaining chunks
                bzero(bits, (BITS_PER_BYTE - i) * chunkSize);
            }
            else {
                int j, k;
                fs_RefByte *refbytes = refBuf->u.refBytes;

                for (j = 0, k = 0; j < blockSize / BITS_PER_BYTE; j += 4, k += 4*sizeof (uint64)) {
                    uint8 snapbits;
                    uint64 word, live, never, live_mask, never_mask;

#define GET_SNAPBITS do{\
        live = word & live_mask;        /* find only live bits              */\
        never = (word & never_mask) >> 1;   /* mask out the NEVER_SNAPSHOT bits  */\
        live &= ~never;             /* clear live bits marked NEVER_SNAPSHOT */\
        snapbits = (live & 0x1)                             \
                    | ((live & 0x100) >> 7)                 \
                    | ((live & 0x10000) >> 14)              \
                    | ((live & 0x1000000) >> 21)            \
                    | ((live & 0x100000000) >> 28)          \
                    | ((live & 0x10000000000) >> 35)        \
                    | ((live & 0x1000000000000) >> 42)      \
                    | ((live & 0x100000000000000) >> 49);   \
} while (0)

                    live_mask = ((uint64)FS_REFBYTE_IS_LIVE
                                | ((uint64)FS_REFBYTE_IS_LIVE << 8)
                                | ((uint64)FS_REFBYTE_IS_LIVE << 16)
                                | ((uint64)FS_REFBYTE_IS_LIVE << 24)
                                | ((uint64)FS_REFBYTE_IS_LIVE << 32)
                                | ((uint64)FS_REFBYTE_IS_LIVE << 40)
                                | ((uint64)FS_REFBYTE_IS_LIVE << 48)
                                | ((uint64)FS_REFBYTE_IS_LIVE << 56));
                    never_mask = ((uint64)FS_REFBYTE_NEVER_SNAPSHOT
                                |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 8)
                                |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 16)
                                |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 24)
                                |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 32)
                                |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 40)
                                |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 48)
                                |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 56));

                    word = *(uint64 *)&refbytes[k];
                    GET_SNAPBITS;
                    bits[j] = snapbits; 

                    word = *(uint64 *)&refbytes[k+8];
                    GET_SNAPBITS;
                    bits[j+1] = snapbits; 

                    word = *(uint64 *)&refbytes[k+16];
                    GET_SNAPBITS;
                    bits[j+2] = snapbits; 

                    word = *(uint64 *)&refbytes[k+24];
                    GET_SNAPBITS;
                    bits[j+3] = snapbits; 
                }
                fs_bufRelease(context, refBuf);
            }
        }

        // dirty and release the buffer

        FS_BUF_WRITE_NO_LOG(snapBuf, 0, blockSize);
        fs_bufReleaseExclusive(context, snapBuf);
    }

    fs_inodeRelease(context, refInode);
    fs_inodeRelease(context, snapBitsInode);

    FS_SNAP_END_TUNE("create write snapbits");
    return NFX_OK;
#else
    int32   rc;
    FS_SNAP_TUNE_DECL
    FS_SNAP_START_TUNE("snap create write snapbits");

    rc = fs_snapTransformSnapBits(context, id, fs_snapCompressLiveBits);

    FS_SNAP_END_TUNE("snap create write snapbits");

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap create write snapbits: %d rc %d", id, rc));
    return rc;
#endif
}


/**
 * Snapshot creation:  incrementing reference counts.
 */
static int32 fs_snapCreateIncrementRefBytes(
    fs_Context *context,
    int id)
{
#ifdef FS_SNAP_OPTIMIZED 
    fs_Parameters *params = &context->fileSystem->params;
    uint32 blockSize = params->blockSize;
    uint64 refBytePages = params->refbytesBlockCount;
    uint64 page;
    FS_SNAP_WATCHDOG_DECL
    FS_SNAP_TUNE_DECL

    fs_Inode *inode = I_GET_SHARED(0, FS_REF_COUNT_INODE);

    FS_SNAP_START_TUNE("snap create inc refbytes");

    assert(inode != NULL);

    FS_SNAP_WATCHDOG_INIT("inc refbytes");

    for (page = 0; page < refBytePages; page++) {
        int i;
        fs_Buf *buf = O_GET_EXCL(inode, page);
        fs_RefByte *ref;
        uint64 changed = 0;
	    uint64 live_mask, never_mask;
        
        assert(buf != NULL);
        ref = buf->u.refBytes;

	    /*
	     * Unrolled loop to increment snapcount in the refbytes:
	     *
	     * The idea is to mask out the live bits, and to clear any of
	     * the live bits which also have the never snapshot bit set.
	     * This is accomplished by shifting the never snapshot bit
	     * into the position of the live bit, and doing a NAND.
	     *
	     * Finally, the live bits are shifted to the snap count
	     * position in the byte, and added to the existing snapcount.
	     *
	     * This can all be done in an 8-way "vector" operation
	     * in a dword register.
	     */
	    FS_ASSERT((FS_REFBYTE_NEVER_SNAPSHOT >> 1) == FS_REFBYTE_IS_LIVE);
	    FS_ASSERT(FS_REFBYTE_IS_LIVE == 1);

#define INCREMENT_SNAP_COUNTS do{\
		live = word & live_mask;		/* find only live bits				*/\
		never = (word & never_mask) >> 1;	/* mask out the NEVER_SNAPSHOT bits	 */\
		live &= ~never;				/* clear live bits marked NEVER_SNAPSHOT */\
		live <<= FS_REFBYTE_SNAP_COUNT_SHIFT;	/* create a snap_count addand			*/\
		word += live;				/* increment snap counts			*/\
		changed |= live;			/* nonzero "changed" means block is dirty */\
} while (0)

	    live_mask = ((uint64)FS_REFBYTE_IS_LIVE
		            | ((uint64)FS_REFBYTE_IS_LIVE << 8)
		            | ((uint64)FS_REFBYTE_IS_LIVE << 16)
		            | ((uint64)FS_REFBYTE_IS_LIVE << 24)
		            | ((uint64)FS_REFBYTE_IS_LIVE << 32)
		            | ((uint64)FS_REFBYTE_IS_LIVE << 40)
		            | ((uint64)FS_REFBYTE_IS_LIVE << 48)
		            | ((uint64)FS_REFBYTE_IS_LIVE << 56));
	    never_mask = ((uint64)FS_REFBYTE_NEVER_SNAPSHOT
		            |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 8)
		            |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 16)
		            |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 24)
		            |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 32)
		            |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 40)
		            |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 48)
		            |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 56));

	    /*
	     * Unroll by 4x.
	     */
        for (i = 0; i < blockSize; i += 4*sizeof (uint64)) {
		    uint64 word;
		    uint64 never, live;

		    word = *(uint64 *)&ref[i];
		    INCREMENT_SNAP_COUNTS;
		    *(uint64 *)&ref[i] = word;

		    word = *(uint64 *)&ref[i+8];
		    INCREMENT_SNAP_COUNTS;
		    *(uint64 *)&ref[i+8] = word;

		    word = *(uint64 *)&ref[i+16];
		    INCREMENT_SNAP_COUNTS;
		    *(uint64 *)&ref[i+16] = word;

		    word = *(uint64 *)&ref[i+24];
		    INCREMENT_SNAP_COUNTS;
		    *(uint64 *)&ref[i+24] = word;

		    // yield the CPU if necc to prevent watchdog timeouts
		    FS_SNAP_WATCHDOG_CHECK("inc refbytes");
        }

        // write the buffer

        if (changed)
            FS_BUF_WRITE_NO_LOG(buf, 0, blockSize);
        fs_bufReleaseExclusive(context, buf);
    }

    fs_inodeRelease(context, inode);

    FS_SNAP_END_TUNE("snap create inc refbytes");

    return NFX_OK;
#else
    fs_Parameters *params = &context->fileSystem->params;
    uint32 blockSize = params->blockSize;
    uint64 refBytePages = params->refbytesBlockCount;
    uint64 page;
    FS_SNAP_WATCHDOG_DECL

    fs_Inode *inode = I_GET_SHARED(0, FS_REF_COUNT_INODE);

    assert(inode != NULL);

    FS_SNAP_WATCHDOG_INIT("inc refbytes");

    for (page = 0; page < refBytePages; page++) {
        int i;
        fs_Buf *buf = O_GET_EXCL(inode, page);
        fs_RefByte *ref;
        boolean changed = FALSE;

        assert(buf != NULL);
        ref = buf->u.refBytes;

        for (i = 0; i < blockSize; i++) {
            fs_RefByte byte = ref[i];
            boolean live
                = ((byte & (FS_REFBYTE_IS_LIVE | FS_REFBYTE_NEVER_SNAPSHOT))
                   == FS_REFBYTE_IS_LIVE);

            if (live) {
                ref[i] = byte + (1U << FS_REFBYTE_SNAP_COUNT_SHIFT);
                // printf("snapcount[%lld] incremented to %d\n",
                //        firstBlock + i, fs_refbyteSnapCount(ref[i]));
                changed = TRUE;
                fs_snapStats.refCntIncs++;
            }

            // yield the CPU if necc to prevent watchdog timeouts
            FS_SNAP_WATCHDOG_CHECK("inc refbytes");
        }

        // write the buffer

        if (changed)
            FS_BUF_WRITE_NO_LOG(buf, 0, blockSize);
        fs_bufReleaseExclusive(context, buf);
    }

    fs_inodeRelease(context, inode);
    return NFX_OK;
#endif
}


/**
 * Copy the inode with the given number from the source to destination buffer 
 */
static void fs_snapCopyInode(
    fs_Context *context,
    fs_Buf *dest,
    fs_Buf *source,
    fs_Inumber inum)
{
    fs_Time now = fs_getTimeOfDay();

    assert(inum < FS_MAX_META_INODE);
    dest->u.dinode->dinode[inum] = source->u.dinode->dinode[inum];

    // set the correct times for the snapshot's root inode
    // so things like ls -l in the .snapshot dir will see the correct times

    if (inum == FS_ROOT_INODE) {
        dest->u.dinode->dinode[inum].ctime = now;
        dest->u.dinode->dinode[inum].version++;
        dest->u.dinode->dinode[inum].atime = now;
        dest->u.dinode->dinode[inum].mtime = now;
    }

    FS_BUF_WRITE_NO_LOG(dest, inum * sizeof (fs_Dinode), sizeof (fs_Dinode));
}


/**
 * Copy necessary meta inodes from the live file system to the snapshot.
 *
 * TODO - automate, use a flag to indicate what's not copied?
 */
static void fs_snapCopyMetaInodes(
    fs_Context *context,
    int id)
{
    fs_Buf *live = B_GET_SHARED(NULL, fs_firstInodeBlock(0));
    fs_Buf *snap = B_GET_EXCL(NULL, fs_firstInodeBlock(id)); 

    fs_snapCopyInode(context, snap, live, FS_INODE_INODE);
    fs_snapCopyInode(context, snap, live, FS_ROOT_INODE);
    fs_snapCopyInode(context, snap, live, FS_FILESET_INFO_INODE);
    fs_snapCopyInode(context, snap, live, FS_FILESET_USER_QUOTA_INODE);
    fs_snapCopyInode(context, snap, live, FS_FILESET_GROUP_QUOTA_INODE);
    fs_snapCopyInode(context, snap, live, FS_TRASHCAN_INODE);
    fs_snapCopyInode(context, snap, live, FS_SECURITY_INODE);
    fs_snapCopyInode(context, snap, live, FS_SNAPSHOT_DIR_INODE);
    fs_snapCopyInode(context, snap, live, FS_LINK_INODE);
    fs_snapCopyInode(context, snap, live, FS_TEMPDIR_INODE);
    fs_snapCopyInode(context, snap, live, FS_NULL_INODE);
    fs_snapCopyInode(context, snap, live, FS_TEMP_DUMP_SWAP);
    fs_snapCopyInode(context, snap, live, FS_AUDIT_INODE);
    fs_snapCopyInode(context, snap, live, FS_SYSTEM_HIDDEN_INODE);
    fs_snapCopyInode(context, snap, live, FS_EXTENDED_ATTR_INODE);
    fs_snapCopyInode(context, snap, live, FS_REF_OVERFLOW_INODE);
    fs_snapCopyInode(context, snap, live, FS_CIFS_SHARES_INODE);

    fs_bufRelease(context, live);
    fs_bufReleaseExclusive(context, snap);
}


/**
 * Clear the live bit for block
 */
static void fs_snapClearLiveBlock(
    fs_Context *context,
    fs_Inode *refInode,
    fs_Block block)
{
    fs_Parameters *params = &context->fileSystem->params;
    uint32 blockSize = params->blockSize;
    uint64 pageno = fs_blockNumber(block) / blockSize;
    uint32 offset = fs_blockNumber(block) % blockSize;
    fs_Buf *buf = O_GET_EXCL(refInode, pageno);

    if (buf == NULL) {
        // nearing the end of the file
        // assert(page + 1 == snapBitsPages);
    } else {
        fs_RefByte byte = buf->u.refBytes[offset];
        boolean live = fs_refbyteIsLive(byte);
        // for snap resume, check if bit was already cleared
        if (live)
        {
            buf->u.refBytes[offset] = fs_refbyteClearLive(byte);
            FS_BUF_WRITE_NO_LOG(buf, 0, blockSize);
        }
        FS_TRACE(context, FS_TRACE_SNAP,
                 ("snapbits clear: block %lld, ref-byte %x ref-page %lld offset %d (%s)",
                 fs_blockNumber(block), byte, pageno, offset,
                 live ? "bit cleared" : "bit already cleared"));
        fs_bufReleaseExclusive(context, buf);
    }
}


static boolean fs_snapClearLiveWorker(
    fs_Context *context,
    fs_Inode *inode,
    fs_WalkResult *result,
    void *arg0)
{
    fs_Inode *refInode = arg0;
    uint16 b, n;

    assert(inode->inum == FS_SNAPSHOT_BITS_INODE);

    if (result->type == FS_WALK_EXTENT_TYPE) {
        n = fs_extentLength(result->extent);
        FS_TRACE(context, FS_TRACE_SNAP,
                 ("snap clear worker: extent-type startBlock %lld len %d",
                 fs_blockNumber(fs_extentStartBlock(result->extent)), n));
        for (b = 0; b < n; b++) {
            fs_snapClearLiveBlock(context, refInode,
                                  fs_extentBlock(result->extent, b));
        }
    } else {
        assert((result->type == FS_WALK_INNER_NODE_TYPE) ||
               (result->type == FS_WALK_LEAF_NODE_TYPE));
        FS_TRACE(context, FS_TRACE_SNAP,
                 ("snap clear worker: %s block %lld",
                 (result->type == FS_WALK_INNER_NODE_TYPE) ? "inner-type" : "leaf-type",
                 fs_blockNumber(result->block)));
        fs_snapClearLiveBlock(context, refInode, result->block);
    }
    return TRUE;
}

/**
 * Remove the snap bits from the live file system.  That is, clear the
 * live-block indicator for the snap bits file.
 */
static void fs_snapClearLiveSnapBits(
    fs_Context *context,
    int id)
{
    fs_Inode *snapBitsInode = I_GET_SHARED(id, FS_SNAPSHOT_BITS_INODE);
    fs_Inode *refInode = I_GET_EXCL(0, FS_REF_COUNT_INODE);
    FS_SNAP_TUNE_DECL
    FS_SNAP_START_TUNE("snap create clear snapbits");

    assert(refInode != NULL);
    assert(snapBitsInode != NULL);

    /*
     * Walk through the btree index itself, and all the blocks of the snapbits
     * file, and clear the bit in fs_snapClearLiveWorker.
     */
    fs_emap_walk(context, snapBitsInode, fs_snapClearLiveWorker, refInode);

    fs_inodeReleaseExclusive(context, refInode);
    fs_inodeRelease(context, snapBitsInode);

    FS_SNAP_END_TUNE("snap create clear snapbits");
}


/**
 * Snapshot creation:  writing the global metadata.  Mostly, this means
 * copying features of page 0 of the inode file, but also includes
 * removing the snap bits from the live file system.
 */
static int32 fs_snapCreateFinish(
    fs_Context *context,
    int id)
{
#ifdef FS_SNAP_CACHE_OPTIMIZED
    FS_SNAP_TUNE_DECL
#endif

    fs_snapCopyMetaInodes(context, id);
    fs_snapClearLiveSnapBits(context, id);

#ifdef FS_SNAP_CACHE_OPTIMIZED

    FS_SNAP_START_TUNE("snap create finish: enable cow for caches");

    // optimized: don't flush entire buffer and inode caches
    // just re-enable their cow mechanism

    fs_bufCacheEnableCow(context);
    fs_inodeCacheEnableCow(context);

    FS_SNAP_END_TUNE("snap create finish: enable cow for caches");

#endif

    // flush .snapshot from the attribute cache on the txrx
    fs_snapFlushAttrCache(context, 0, FS_SNAPSHOT_DIR_INODE);

    return NFX_OK;
}


//    S N A P S H O T   R E M O V E


#ifndef FS_SNAP_OPTIMIZED
/**
 * Clear the snapbits (used by phase FS_SNAP_PHASE_REMOVE_SNAP_BITS)
 *  clear block's snap bit if any of these are true:
 *      1. live bit set
 *      2. reference byte's snap counter > 0 
 *      3. block should never be snapshotted
 *         (redundant since fs_snapCompressLiveBits() checks)
 *  if any true, this block should not be free'd later in the
 *  FS_SNAP_PHASE_REMOVE_FREE_BLOCKS phase
 */
static boolean fs_snapClearSnapBits(
    fs_Context *context,
    uint8      *snapBits,
    fs_RefByte *ref,
    uint64     blockStart)
{
    boolean changed = FALSE;
    int i, n;
    uint32 blockSize = context->fileSystem->params.blockSize;
    uint64 block = blockStart;

    assert((blockSize % BITS_PER_BYTE) == 0);

    for (i = 0, n = blockSize / BITS_PER_BYTE; i < n; i++, block += BITS_PER_BYTE) {
        int j;
        uint8 bits = snapBits[i];

        for (j = 0; j < BITS_PER_BYTE; j++, ref++) {
            uint8 mask = 1 << j;
            fs_RefByte byte = *ref;
            uint8 bit = bits & mask;
            boolean live = fs_refbyteIsLive(byte);
            boolean neversnap = fs_refbyteNeverSnapshot(byte);
            uint32 snapcount = fs_refbyteSnapCount(byte);

            if (bit &&
                (live || neversnap || snapcount)) {
                snapBits[i] &= ~mask;
                changed = TRUE;
                fs_snapStats.notFrees++;
                if (live) fs_snapStats.liveNotFrees++;
                if (snapcount) fs_snapStats.snapNotFrees++;
            }
        }
    }
    return changed;
}
#endif


#ifndef FS_SNAP_OPTIMIZED
/**
 * Decrement the reference bytes (used by phase FS_SNAP_PHASE_REMOVE_REF_COUNTS)
 *  if snap bit set for block, decrement reference byte's snap counter
 */
static boolean fs_snapDecrementRefSnapCount(
    fs_Context *context,
    uint8      *snapBits,
    fs_RefByte *ref,
    uint64     blockStart)
{
    boolean changed = FALSE;
    int i, n;
    uint32 blockSize = context->fileSystem->params.blockSize;
    uint64 block = blockStart;

    assert((blockSize % BITS_PER_BYTE) == 0);

    for (i = 0, n = blockSize / BITS_PER_BYTE; i < n; i++, block += BITS_PER_BYTE) {
        int j;
        uint8 bits = snapBits[i];

        for (j = 0; j < BITS_PER_BYTE; j++, ref++) {
            uint8 mask = 1 << j;
            fs_RefByte byte = *ref;
            uint8 bit = bits & mask;
            if (bit) {
                *ref = fs_refbyteDecrSnapCount(byte);
                changed = TRUE;
                byte = *ref;
                fs_snapStats.refCntDecs++;
            }
        }
    }

    return changed;
}
#endif


/**
 * Snapshot removal (phase FS_SNAP_PHASE_REMOVE_REF_COUNTS)
 *   decrement ref byte's for blocks in the snapbits file
 */
static int32 fs_snapRemoveDecrementRefBytes(
    fs_Context *context,
    int        id)
{
#ifdef FS_SNAP_OPTIMIZED 
    fs_Parameters *params = &context->fileSystem->params;
    uint32 blockSize = params->blockSize;
    uint32 chunkSize = blockSize / BITS_PER_BYTE;
    uint64 snapBitsPages;
    uint64 page, block;
    FS_SNAP_WATCHDOG_DECL

    fs_Inode    *snapBitsInode = I_GET_SHARED(id, FS_SNAPSHOT_BITS_INODE);
    fs_Inode    *refInode = I_GET_SHARED(0, FS_REF_COUNT_INODE);

    FS_SNAP_TUNE_DECL
    FS_SNAP_START_TUNE("snap remove dec refbytes");

    assert(refInode != NULL);

    // snap resume support
    if (snapBitsInode == NULL) {
        FS_TRACE(context, FS_TRACE_SNAP,
                 ("snap transform ref bytes: no snapbits inode"));
        fs_inodeRelease(context, refInode);
        return NFX_OK;
    }

    snapBitsPages = ceilingDiv64(snapBitsInode->dinode.size, blockSize);
    assert(snapBitsPages == snapBitsInode->dinode.numblocks);

    FS_SNAP_WATCHDOG_INIT("dec refbytes");

    for (page = 0; page < snapBitsPages; page++) {
        int i;
        fs_Buf *snapBuf = O_GET_SHARED(snapBitsInode, page);

        assert(snapBuf != NULL);

        // yield the CPU if necc to prevent watchdog timeouts
        FS_SNAP_WATCHDOG_CHECK("dec refbytes");

        block = (page * blockSize);
        for (i = 0; i < BITS_PER_BYTE; i++) {
            uint64 changed = 0;
            uint8 *snapBits = &snapBuf->u.data[i * chunkSize];
            fs_Buf *refBuf = O_GET_EXCL(refInode, page * BITS_PER_BYTE + i);
            fs_RefByte *refbytes;

            if (refBuf == NULL) {
                // nearing the end of the file
                assert(page + 1 == snapBitsPages);
            }
            else {
                int j, k;
                uint64 word, snapcount;

                refbytes = refBuf->u.refBytes;
                for (j = 0, k = 0; j < blockSize / BITS_PER_BYTE; j += 4, k += 4*sizeof (uint64)) {
                    uint8 bits;
/*
 * The idea is to "expand" a byte of snapbits to 64 bit value.
 * This value is then shifted to the snapcount position and
 * decremented to the existing snapcount.
 */

#define DECREMENT_SNAP_COUNTS do{\
    snapcount = ((uint64) (bits & 0x1))                 \
                | ((uint64) ((bits & 0x2) ? 1 : 0) << 8)    \
                | ((uint64) ((bits & 0x4) ? 1 : 0) << 16)   \
                | ((uint64) ((bits & 0x8) ? 1 : 0) << 24)   \
                | ((uint64) ((bits & 0x10) ? 1 : 0) << 32)  \
                | ((uint64) ((bits & 0x20) ? 1 : 0) << 40)  \
                | ((uint64) ((bits & 0x40) ? 1 : 0) << 48)  \
                | ((uint64) ((bits & 0x80) ? 1 : 0) << 56); \
    snapcount <<= FS_REFBYTE_SNAP_COUNT_SHIFT;  \
    word -= snapcount;          /* decrement snap counts            */\
    changed |= snapcount;       /* nonzero "changed" means block is dirty */\
} while (0)

                    bits = snapBits[j];
                    if (bits != 0) {
                        word = *(uint64 *)&refbytes[k];
                        DECREMENT_SNAP_COUNTS;
                        *(uint64 *)&refbytes[k] = word;
                    }
                    bits = snapBits[j+1];
                    if (bits != 0) {
                        word = *(uint64 *)&refbytes[k+8];
                        DECREMENT_SNAP_COUNTS;
                        *(uint64 *)&refbytes[k+8] = word;
                    }
                    bits = snapBits[j+2];
                    if (bits != 0) {
                        word = *(uint64 *)&refbytes[k+16];
                        DECREMENT_SNAP_COUNTS;
                        *(uint64 *)&refbytes[k+16] = word;
                    }
                    bits = snapBits[j+3];
                    if (bits != 0) {
                        word = *(uint64 *)&refbytes[k+24];
                        DECREMENT_SNAP_COUNTS;
                        *(uint64 *)&refbytes[k+24] = word;
                    }
                }
            }
            if (refBuf) {
                if (changed) FS_BUF_WRITE_NO_LOG(refBuf, 0, blockSize);
                fs_bufReleaseExclusive(context, refBuf);
            }
        }
        fs_bufRelease(context, snapBuf);
    }

    fs_inodeRelease(context, refInode);
    fs_inodeRelease(context, snapBitsInode);

    FS_SNAP_END_TUNE("snap remove dec refbytes");
    return NFX_OK;
#else
    int32   rc;
    FS_SNAP_TUNE_DECL
    FS_SNAP_START_TUNE("snap remove dec refbytes");

    rc = fs_snapTransformRefBytes(context, id, fs_snapDecrementRefSnapCount);

    FS_TRACE(context, FS_TRACE_SNAP,
            ("snap remove dec refs: %d rc %d", id, rc));

    FS_SNAP_END_TUNE("snap remove dec refbytes");
    return rc;
#endif
}


/**
 * worker routine for emap_walk of the snapbits file
 */
static boolean fs_snapSnapBitsEmapWorker(
    fs_Context *context,
    fs_Inode *inode,
    fs_WalkResult *result,
    void *arg0)
{
    fs_SnapRemoveWorkerArg *pBlocks = (fs_SnapRemoveWorkerArg *) arg0;
    uint16 b, n;

    assert(inode->inum == FS_SNAPSHOT_BITS_INODE);
    assert(pBlocks->blockIndex < pBlocks->maxBlocks);

    if (result->type == FS_WALK_EXTENT_TYPE) {
        n = fs_extentLength(result->extent);
        for (b = 0; b < n; b++) {
            pBlocks->blocks[pBlocks->blockIndex++] =
                fs_blockNumber(fs_extentBlock(result->extent, b));
        }
    } else {
        assert((result->type == FS_WALK_INNER_NODE_TYPE) ||
               (result->type == FS_WALK_LEAF_NODE_TYPE));
        pBlocks->blocks[pBlocks->blockIndex++] = fs_blockNumber(result->block);
    }
    return TRUE;
}


#ifdef FS_SNAP_OPTIMIZE_REMOVE_FREE_BLOCKS
/**
 * Snapshot removal (phase FS_SNAP_PHASE_REMOVE_SNAP_BITS_AND_FREE)
 *   update snapbits file and deallocate blocks:
 * 1. clear bits for all the blocks in the snapbits file including inner nodes
 * 2. clear bits for blocks that are in use
 * 3. deallocate blocks that have their snapbits set
 */
static int32 fs_snapRemoveUpdateSnapBitsAndFree(
    fs_Context *context,
    int        id)
{
    int32   b;
    fs_Parameters *params = &context->fileSystem->params;
    uint32 blockSize = params->blockSize;
    uint32 chunkSize = blockSize / BITS_PER_BYTE;
    uint32 totalBlocksFree = 0;
    uint64 totalBlocks = params->totalBlocks;
    uint64 snapBitsPages
        = ceilingDiv64(totalBlocks, params->blockSize * BITS_PER_BYTE);
    fs_SnapRemoveWorkerArg arg0;
    uint64 page, block;
    FS_SNAP_WATCHDOG_DECL
    FS_SNAP_TUNE_DECL

    fs_Inode *refInode = I_GET_SHARED(0, FS_REF_COUNT_INODE);
    fs_Inode *snapBitsInode = I_GET_SHARED(id, FS_SNAPSHOT_BITS_INODE);

    assert(refInode != NULL);

    // snap resume support
    if (snapBitsInode == NULL) {
        FS_TRACE(context, FS_TRACE_SNAP,
                 ("remove update snapbits and free: no snapbits inode"));
        fs_inodeRelease(context, refInode);
        return NFX_OK;
    }

    snapBitsPages = ceilingDiv64(snapBitsInode->dinode.size, blockSize);
    assert(snapBitsPages == snapBitsInode->dinode.numblocks);

    FS_SNAP_START_TUNE("remove update snapbits and free");
    FS_SNAP_WATCHDOG_INIT("remove update snapbits and free");

    // call emap walk to gather all blocks of the snapbits file
    // these will not get deallocated til FS_SNAP_PHASE_REMOVE_FREE_SNAP_BITS phase

    arg0.blockIndex = 0;
    arg0.maxBlocks = snapBitsPages + 8;  // include inner nodes (FS_BT_MAXLEVEL)
    arg0.blocks = (uint64 *) fs_ramAlloc(FS_RAM_SNAP_NAME,
                                         (arg0.maxBlocks*sizeof(uint64)));
    fs_emap_walk(context, snapBitsInode, fs_snapSnapBitsEmapWorker, &arg0);

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap remove update snapbits and free: found %d blocks for snapbits file",
             arg0.blockIndex));

    for (b = 0; b <arg0.blockIndex; b++) {
        uint8 *snapbits, mask;
        uint64 bitnum;
        fs_Buf *buf;

        // clear this block's bit in the snapbits file

        block = arg0.blocks[b];
        buf = O_GET_EXCL(snapBitsInode, block / (blockSize * BITS_PER_BYTE));
        assert(buf);
        snapbits = buf->u.data;
        bitnum = (block % (blockSize * BITS_PER_BYTE));
        mask = 1 << (bitnum & 0x7);
        snapbits[bitnum >> 3] ^= mask;
        FS_TRACE(context, FS_TRACE_SNAP,
                ("  block %lld bitnum %lld snapbits[%lld] = %x",
                block, bitnum, bitnum >> 3, snapbits[bitnum >> 3]));
        FS_BUF_WRITE_NO_LOG(buf, 0, blockSize);
        fs_bufReleaseExclusive(context, buf);
    }

    fs_ramDealloc(FS_RAM_SNAP_NAME, arg0.blocks);

    // create a transaction context (necc for free'ing block)

    assert(context->logSummary == NULL);
    fs_logBeginTransaction(context);

    for (page = 0; page < snapBitsPages; page++) {
        int i;
        fs_Buf *snapBuf = O_GET_EXCL(snapBitsInode, page);
        assert(snapBuf != NULL);

        for (i = 0; i < BITS_PER_BYTE; i++) {
            uint8 *bits = &snapBuf->u.data[i * chunkSize];
            uint64 refPage = page * BITS_PER_BYTE + i;
            fs_Buf *refBuf = O_GET_SHARED(refInode, refPage);

            block = refPage * FS_BLOCK_SIZE;

            if (refBuf == NULL) {
                // nearing the end of the file
                assert(page + 1 == snapBitsPages);
                // zero out remaining chunks
                bzero(bits, (BITS_PER_BYTE - i) * chunkSize);
            }
            else {
                int j, k;
                fs_RefByte *refbytes = refBuf->u.refBytes;

                for (j = 0, k = 0; j < blockSize / BITS_PER_BYTE; j += 4, k += 4*sizeof (uint64)) {
                    uint8 snapbits, oldbits;
                    uint32 len;
                    uint64 word, startBlock, endBlock;

                    // yield the CPU if necc to prevent watchdog timeouts
                    FS_SNAP_WATCHDOG_CHECK("remove update snapbits");

// determine which blocks to deallocate
// snapbits cleared if corresponding refbyte non-zero
// otherwise set for deallocation

#define UPDATE_SNAPBITS(index) do{\
        oldbits = snapbits;                                                 \
        snapbits &= ((uint64)((word & 0xff) ? 0 : 1))                       \
                | ((uint64) ((word & 0xff00) ? 0 : 1) << 1)                 \
                | ((uint64) ((word & 0xff0000) ? 0 : 1) << 2)               \
                | ((uint64) ((word & 0xff000000) ? 0 : 1) << 3)             \
                | ((uint64) ((word & 0xff00000000) ? 0 : 1) << 4)           \
                | ((uint64) ((word & 0xff0000000000) ? 0 : 1) << 5)         \
                | ((uint64) ((word & 0xff000000000000) ? 0 : 1) << 6)       \
                | ((uint64) ((word & 0xff00000000000000) ? 0 : 1) << 7);    \
} while (0)

// remove deallocated blocks from the buffer cache and buddy

#define FREE_BLOCKS(index) do{\
    if (snapbits) {                                                         \
        for (b = 0; b < BITS_PER_BYTE; ) {                                  \
            boolean bit;                                                    \
            bit = (snapbits >> b) & 0x1;                                    \
            if (!bit) {                                                     \
                b++;                                                        \
            } else {                                                        \
                len = 1;                                                    \
                startBlock = block + index + b++;                           \
                while (b < BITS_PER_BYTE) {                                 \
                    bit = (snapbits >> b++) & 0x1;                          \
                    if (!bit) break;                                        \
                    len++;                                                  \
                }                                                           \
                endBlock = startBlock + len - 1;                            \
                totalBlocksFree += len;                                     \
                fs_buddyMaybeAddExtent(context,                             \
                    fs_makeExtent(len, fs_makeBlock(startBlock)), TRUE);    \
                fs_bufCacheFlushBlocks(context, startBlock, len);           \
            }                                                               \
        }                                                                   \
    }                                                                       \
} while (0)

#define FREE_BLOCKS_DEBUG(index) do{\
    if (snapbits) {                                                         \
        printf("refbytes[%lld,%d]: %x %x %x %x %x %x %x %x\n",              \
            refPage, index,                                                 \
            refbytes[index], refbytes[index+1], refbytes[index+2],          \
            refbytes[index+3], refbytes[index+4], refbytes[index+5],        \
            refbytes[index+6], refbytes[index+7]);                          \
        for (b = 0; b < BITS_PER_BYTE; ) {                                  \
            boolean bit;                                                    \
            bit = (snapbits >> b) & 0x1;                                    \
            if (!bit) {                                                     \
                b++;                                                        \
            } else {                                                        \
                len = 1;                                                    \
                startBlock = block + index + b++;                           \
                while (b < BITS_PER_BYTE) {                                 \
                    bit = (snapbits >> b++) & 0x1;                          \
                    if (!bit) break;                                        \
                    len++;                                                  \
                }                                                           \
                endBlock = startBlock + len - 1;                            \
                printf("FREE[%d]: %x,%x %lld -> %lld len %d\n",             \
                    b, oldbits, snapbits, startBlock, endBlock, len);       \
                totalBlocksFree += len;                                     \
                fs_buddyMaybeAddExtent(context,                             \
                    fs_makeExtent(len, fs_makeBlock(startBlock)), TRUE);    \
                fs_bufCacheFlushBlocks(context, startBlock, len);           \
            }                                                               \
        }                                                                   \
    }                                                                       \
} while (0)

                    word = *(uint64 *)&refbytes[k];
                    snapbits = bits[j];
                    UPDATE_SNAPBITS(k);
                    FREE_BLOCKS(k);

                    word = *(uint64 *)&refbytes[k+8];
                    snapbits = bits[j+1];
                    UPDATE_SNAPBITS(k+8);
                    FREE_BLOCKS(k+8);

                    word = *(uint64 *)&refbytes[k+16];
                    snapbits = bits[j+2];
                    UPDATE_SNAPBITS(k+16);
                    FREE_BLOCKS(k+16);

                    word = *(uint64 *)&refbytes[k+24];
                    snapbits = bits[j+3];
                    UPDATE_SNAPBITS(k+24);
                    FREE_BLOCKS(k+24);
                }

                fs_bufRelease(context, refBuf);
            }
        }
        fs_bufReleaseExclusive(context, snapBuf);
    }

    fs_inodeRelease(context, snapBitsInode);
    fs_inodeRelease(context, refInode);

    // commit and release the transaction context

    fs_logCommitEntry(context);
    context->logSummary = NULL;

    // update blksUsed in superblock

    if (totalBlocksFree) {
        NFX_SPIN_LOCK(&context->fileSystem->superLock);
        context->fileSystem->commitedBlksUsed -= totalBlocksFree;
        context->fileSystem->super->blksUsed -= totalBlocksFree;
        NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
        fs_writeSuperblocksSync(context, context->fileSystem->super);
        FS_TRACE(context, FS_TRACE_SNAP,
                 ("snap free blks: id %d blocks free %d super blksUsed %lld commited blksUsed %lld",
                 id, totalBlocksFree, context->fileSystem->super->blksUsed,
                 context->fileSystem->commitedBlksUsed));
    }

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap remove update snapbits and free: id %d", id));
    FS_SNAP_END_TUNE("snap remove update snapbits and free");

    return NFX_OK;
}

#else

/**
 * Snapshot removal (phase FS_SNAP_PHASE_REMOVE_SNAP_BITS)
 *   update snapbits file:
 * 1. clear bits for blocks that are in use
 * 2. clear bits for all the blocks in the snapbits file including inner nodes
 */
static int32 fs_snapRemoveUpdateSnapBits(
    fs_Context *context,
    int        id)
{
#ifdef FS_SNAP_OPTIMIZED 
    int32   b;
    fs_Parameters *params = &context->fileSystem->params;
    uint32 blockSize = params->blockSize;
    uint32 chunkSize = blockSize / BITS_PER_BYTE;
    uint64 snapBitsPages
        = ceilingDiv64(params->totalBlocks, params->blockSize * BITS_PER_BYTE);
    fs_SnapRemoveWorkerArg arg0;
    uint64 page, block;
    FS_SNAP_WATCHDOG_DECL
    FS_SNAP_TUNE_DECL

    fs_Inode *refInode = I_GET_SHARED(0, FS_REF_COUNT_INODE);
    fs_Inode *snapBitsInode = I_GET_SHARED(id, FS_SNAPSHOT_BITS_INODE);

    assert(refInode != NULL);

    // snap resume support
    if (snapBitsInode == NULL) {
        FS_TRACE(context, FS_TRACE_SNAP,
                 ("remove update snapbits: no snapbits inode"));
        fs_inodeRelease(context, refInode);
        return NFX_OK;
    }

    snapBitsPages = ceilingDiv64(snapBitsInode->dinode.size, blockSize);
    assert(snapBitsPages == snapBitsInode->dinode.numblocks);

    FS_SNAP_START_TUNE("remove update snapbits");
    FS_SNAP_WATCHDOG_INIT("remove update snapbits");

    for (page = 0; page < snapBitsPages; page++) {
        int i;
        fs_Buf *snapBuf = O_GET_EXCL(snapBitsInode, page);

        // yield the CPU if necc to prevent watchdog timeouts
        FS_SNAP_WATCHDOG_CHECK("remove update snapbits");

        assert(snapBuf != NULL);
        block = (page * blockSize);
        for (i = 0; i < BITS_PER_BYTE; i++) {
            uint8 *bits = &snapBuf->u.data[i * chunkSize];
            uint64 refPage = page * BITS_PER_BYTE + i;
            fs_Buf *refBuf = O_GET_SHARED(refInode, refPage);

            if (refBuf == NULL) {
                // nearing the end of the file
                assert(page + 1 == snapBitsPages);
                // zero out remaining chunks
                bzero(bits, (BITS_PER_BYTE - i) * chunkSize);
            }
            else {
                int j, k;
                fs_RefByte *refbytes = refBuf->u.refBytes;

                for (j = 0, k = 0; j < blockSize / BITS_PER_BYTE; j += 4, k += 4*sizeof (uint64)) {
                    uint8 snapbits;
                    uint64 word;

#define UPDATE_SNAPBITS do{\
        snapbits &= ((uint64)((word & 0xff) ? 0 : 1))                       \
                | ((uint64) ((word & 0xff00) ? 0 : 1) << 1)                 \
                | ((uint64) ((word & 0xff0000) ? 0 : 1) << 2)               \
                | ((uint64) ((word & 0xff000000) ? 0 : 1) << 3)             \
                | ((uint64) ((word & 0xff00000000) ? 0 : 1) << 4)           \
                | ((uint64) ((word & 0xff0000000000) ? 0 : 1) << 5)         \
                | ((uint64) ((word & 0xff000000000000) ? 0 : 1) << 6)       \
                | ((uint64) ((word & 0xff00000000000000) ? 0 : 1) << 7);    \
} while (0)

                    word = *(uint64 *)&refbytes[k];
                    snapbits = bits[j];
                    UPDATE_SNAPBITS;
                    bits[j] = snapbits; 

                    word = *(uint64 *)&refbytes[k+8];
                    snapbits = bits[j+1];
                    UPDATE_SNAPBITS;
                    bits[j+1] = snapbits; 

                    word = *(uint64 *)&refbytes[k+16];
                    snapbits = bits[j+2];
                    UPDATE_SNAPBITS;
                    bits[j+2] = snapbits; 

                    word = *(uint64 *)&refbytes[k+24];
                    snapbits = bits[j+3];
                    UPDATE_SNAPBITS;
                    bits[j+3] = snapbits; 
                }

                fs_bufRelease(context, refBuf);
            }
        }

        // dirty and release the buffer

        FS_BUF_WRITE_NO_LOG(snapBuf, 0, blockSize);
        fs_bufReleaseExclusive(context, snapBuf);
    }

    fs_inodeRelease(context, refInode);

    // call emap walk to gather all blocks of the snapbits file
    // these will not get deallocated til FS_SNAP_PHASE_REMOVE_FREE_SNAP_BITS phase

    arg0.blockIndex = 0;
    arg0.maxBlocks = snapBitsPages + 8;  // include inner nodes (FS_BT_MAXLEVEL)
    arg0.blocks = (uint64 *) fs_ramAlloc(FS_RAM_SNAP_NAME,
                                         (arg0.maxBlocks*sizeof(uint64)));
    fs_emap_walk(context, snapBitsInode, fs_snapSnapBitsEmapWorker, &arg0);

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap remove update snapbits: found %d blocks for snapbits file",
             arg0.blockIndex));

    for (b = 0; b <arg0.blockIndex; b++) {
        uint8 *snapbits, mask;
        uint64 bitnum;
        fs_Buf *buf;

        // clear this block's bit in the snapbits file

        block = arg0.blocks[b];
        buf = O_GET_EXCL(snapBitsInode, block / (blockSize * BITS_PER_BYTE));
        assert(buf);
        snapbits = buf->u.data;
        bitnum = (block % (blockSize * BITS_PER_BYTE));
        mask = 1 << (bitnum & 0x7);
        snapbits[bitnum >> 3] ^= mask;
        FS_TRACE(context, FS_TRACE_SNAP,
                ("  block %lld bitnum %lld snapbits[%lld] = %x",
                block, bitnum, bitnum >> 3, snapbits[bitnum >> 3]));
        FS_BUF_WRITE_NO_LOG(buf, 0, blockSize);
        fs_bufReleaseExclusive(context, buf);
    }

    fs_inodeRelease(context, snapBitsInode);
#if defined(FS_DEBUG)
    for (b = 0; b <arg0.blockIndex; b++) {
        boolean bit;
        block = arg0.blocks[b];

        fs_snapGetSnapBit(context, snapBitsInode, block, &bit);
        if (bit) {
            printf("snap remove update snapbits: block %lld still SET !!\n",
                    block);
        }
    }
#endif

    fs_ramDealloc(FS_RAM_SNAP_NAME, arg0.blocks);

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap remove update snapbits: id %d", id));
    FS_SNAP_END_TUNE("snap remove update snapbits");

    return NFX_OK;

#else
    int32   rc, b;
    uint64  block;
    fs_Parameters *params = &context->fileSystem->params;
    uint32 blockSize = params->blockSize;
    uint64 snapBitsPages
        = ceilingDiv64(params->totalBlocks, params->blockSize * BITS_PER_BYTE);
    fs_SnapRemoveWorkerArg arg0;

    fs_Inode    *snapBitsInode = I_GET_SHARED(id, FS_SNAPSHOT_BITS_INODE);

    FS_SNAP_TUNE_DECL
    FS_SNAP_START_TUNE("snap remove update snapbits");

    // snap resume support
    if (snapBitsInode == NULL) {
        FS_TRACE(context, FS_TRACE_SNAP,
                 ("snap remove update snapbits: no snapbits inode"));
        return NFX_OK;
    }

    rc = fs_snapTransformSnapBits(context, id, fs_snapClearSnapBits);

    // call emap walk to gather all blocks of the snapbits file
    // these will not get deallocated til FS_SNAP_PHASE_REMOVE_FREE_SNAP_BITS phase

    arg0.blockIndex = 0;
    arg0.maxBlocks = snapBitsPages + 8;  // include inner nodes (FS_BT_MAXLEVEL)
    arg0.blocks = (uint64 *) fs_ramAlloc(FS_RAM_SNAP_NAME,
                                         (arg0.maxBlocks*sizeof(uint64)));
    fs_emap_walk(context, snapBitsInode, fs_snapSnapBitsEmapWorker, &arg0);

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap remove update snapbits: found %d blocks for snapbits file",
             arg0.blockIndex));

    for (b = 0; b <arg0.blockIndex; b++) {
        uint8 *snapbits, mask;
        uint64 bitnum;
        fs_Buf *buf;

        // clear this block's bit in the snapbits file

        block = arg0.blocks[b];
        buf = O_GET_EXCL(snapBitsInode, block / (blockSize * BITS_PER_BYTE));
        assert(buf);
        snapbits = buf->u.data;
        bitnum = (block % (blockSize * BITS_PER_BYTE));
        mask = 1 << (bitnum & 0x7);
        snapbits[bitnum >> 3] ^= mask;
        FS_TRACE(context, FS_TRACE_SNAP,
                ("  block %lld bitnum %lld snapbits[%lld] = %x",
                block, bitnum, bitnum >> 3, snapbits[bitnum >> 3]));
        FS_BUF_WRITE_NO_LOG(buf, 0, blockSize);
        fs_bufReleaseExclusive(context, buf);
    }

    fs_inodeRelease(context, snapBitsInode);

#if defined(FS_DEBUG)
    for (b = 0; b <arg0.blockIndex; b++) {
        boolean bit;
        block = arg0.blocks[b];

        fs_snapGetSnapBit(context, snapBitsInode, block, &bit);
        if (bit) {
            printf("snap remove update snapbits: block %lld still SET !!\n",
                    block);
        }
    }
#endif

    fs_ramDealloc(FS_RAM_SNAP_NAME, arg0.blocks);

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap remove update snapbits: id %d rc %d", id, rc));

    FS_SNAP_END_TUNE("snap remove update snapbits");

    return rc;
#endif
}
#endif // else FS_SNAP_OPTIMIZE_REMOVE_FREE_BLOCKS


/**
 * Snapshot removal (phase FS_SNAP_PHASE_REMOVE_FREE_BLOCKS)
 *  free blocks in the snapbits file
 */
static int32 fs_snapRemoveFreeBlocks(
    fs_Context *context,
    int        id)
{
    fs_Parameters *params = &context->fileSystem->params;
    uint32 blockSize = params->blockSize, blocksFree = 0;
    uint64 blocksPerPage = blockSize * BITS_PER_BYTE;
    uint64 page;
    uint64 snapBitsPages;
    FS_SNAP_WATCHDOG_DECL
    FS_SNAP_TUNE_DECL

    fs_Inode *snapBitsInode = I_GET_SHARED(id, FS_SNAPSHOT_BITS_INODE);

    FS_SNAP_START_TUNE("snap remove free blocks");

    // snap resume support
    if (snapBitsInode == NULL) {
        FS_TRACE(context, FS_TRACE_SNAP,
                 ("snap remove free blocks: no snapbits inode"));
        return NFX_OK;
    }

    snapBitsPages = ceilingDiv64(snapBitsInode->dinode.size, params->blockSize);
    assert(snapBitsPages == snapBitsInode->dinode.numblocks);

    // create a transaction context (necc for free'ing block)

    assert(context->logSummary == NULL);
    fs_logBeginTransaction(context);

    FS_SNAP_WATCHDOG_INIT("remove free blks");

    for (page = 0; page < snapBitsPages; page++) {
        uint64 block, blockStart, firstBlock, lastBlock;
        boolean bit;
        fs_Buf *snapBuf = O_GET_SHARED(snapBitsInode, page);
        uint8 *snapbits;

        assert(snapBuf != NULL);
        snapbits = snapBuf->u.data;

        blockStart = page * blocksPerPage;
        for (block = 0; block < blocksPerPage; ) {

            // yield the CPU if necc to prevent watchdog timeouts
            FS_SNAP_WATCHDOG_CHECK("remove free blks");

            bit = (snapbits[block >> 3] >> (block & 0x7)) & 0x1;

            if (bit == 0) {
                block++;
            } else {
                int32 len = 0;

                firstBlock = blockStart + block;
                while ((block < blocksPerPage) &&
                       ((snapbits[block >> 3] >> (block & 0x7)) & 0x1)) {
#if defined(FS_DEBUG)
                    fs_RefByte byte = fs_refGetByte(context,
                                                    fs_makeBlock(blockStart+block));
                    if (byte) {
                        // should NOT happen !!
                        // since any live blocks are removed from snapbits ....
                        printf("snap remove free: block %lld refbyte non zero %x\n",
                                blockStart+block, byte);
                    }
#endif
                    block++;
                    len++;
                }

                lastBlock = blockStart + block - 1;

                // call buddy to add extent to free tree

                fs_buddyMaybeAddExtent(context,
                                       fs_makeExtent(len, fs_makeBlock(firstBlock)),
                                       TRUE);

#ifdef FS_SNAP_CACHE_OPTIMIZED
                // delete deallocated blocks from the buffer cache
                // buffer cache should not contain any blocks that have been deallocated
                // since re-allocation of same block will produce "duplicate cache insert"
                // optimized: flush only those blocks that are deallocated

                fs_bufCacheFlushBlocks(context, firstBlock, len);
#endif
                FS_TRACE(context, FS_TRACE_SNAP,
                         ("snap free blks: id %d firstBlock %lld lastBlock %lld len %d",
                         id, firstBlock, lastBlock, len));

                fs_snapStats.blockFrees += len;
                blocksFree += len;
            }
        }
        fs_bufRelease(context, snapBuf);
    }

    fs_inodeRelease(context, snapBitsInode);

    // update blksUsed in superblock

    if (blocksFree) {
        NFX_SPIN_LOCK(&context->fileSystem->superLock);
        context->fileSystem->commitedBlksUsed -= blocksFree;
        context->fileSystem->super->blksUsed -= blocksFree;
        NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
        fs_writeSuperblocksSync(context, context->fileSystem->super);
        FS_TRACE(context, FS_TRACE_SNAP,
                 ("snap free blks: id %d blocks free %d super blksUsed %lld commited blksUsed %lld",
                 id, blocksFree, context->fileSystem->super->blksUsed,
                 context->fileSystem->commitedBlksUsed));
    }

    // commit and release the transaction context

    fs_logCommitEntry(context);
    context->logSummary = NULL;

#ifndef FS_SNAP_CACHE_OPTIMIZED
    // buffer cache should not contain any blocks that have been deallocated
    // since re-allocation of same block will produce "duplicate cache insert"
    // calling flush will also reset buf crc for any meta blocks

    fs_bufCacheFlush(context, NULL);
#endif

    FS_SNAP_END_TUNE("snap remove free blocks");
    return NFX_OK;
}


/** 
 * Snapshot removal (FS_SNAP_PHASE_REMOVE_FREE_SNAP_BITS)
 * 1. deallocate all blocks (including inner btree nodes) of the snapbits file
 * 3. free snapbits inode
 */
static int32 fs_snapRemoveDeallocSnapBits(
    fs_Context *context,
    int id)
{
    fs_Parameters *params = &context->fileSystem->params;
    fs_Buf *inodeBuf;
    fs_Inode *inode;
    uint32 b, blocksFree;
    uint64 snapBitsPages
            = ceilingDiv64(params->totalBlocks, params->blockSize * BITS_PER_BYTE);
    fs_SnapRemoveWorkerArg arg0;
    FS_SNAP_WATCHDOG_DECL
    FS_SNAP_TUNE_DECL
    FS_SNAP_START_TUNE("snap remove dealloc snapbits");

    // if we crashed during a snap remove, the inode might or might
    // not have been removed, so check for null inode

    inode = I_GET_EXCL(id, FS_SNAPSHOT_BITS_INODE);
    if (inode == NULL) {
        FS_TRACE(context, FS_TRACE_SNAP,
                ("snapbits file: already removed for id %d", id));
        return NFX_OK;
    }

    // call emap walk to gather all blocks (including the inner nodes)
    // of the snapbits file to deallocate

    arg0.blockIndex = 0;
    arg0.maxBlocks = snapBitsPages + 8;  // include inner nodes (FS_BT_MAXLEVEL)
    arg0.blocks = (uint64 *) fs_ramAlloc(FS_RAM_SNAP_NAME,
                                         (arg0.maxBlocks*sizeof(uint64)));
    fs_emap_walk(context, inode, fs_snapSnapBitsEmapWorker, &arg0);

    FS_TRACE(context, FS_TRACE_SNAP,
            ("snap remove dealloc snapbits: found %d blocks for snapbits file",
            arg0.blockIndex));

    blocksFree = arg0.blockIndex;

    // create a transaction context (necc for free'ing block)

    assert(context->logSummary == NULL);
    fs_logBeginTransaction(context);

    FS_SNAP_WATCHDOG_INIT("remove dealloc snapbits");

    for (b = 0; b <blocksFree; b++) {
        uint64 block;
        block = arg0.blocks[b];

        // yield the CPU if necc to prevent watchdog timeouts
        FS_SNAP_WATCHDOG_CHECK("remove dealloc snapbits");

        // call buddy to add block to free tree

        fs_buddyMaybeAddExtent(context,
                               fs_makeExtent(1, fs_makeBlock(block)),
                               TRUE);
#ifdef FS_SNAP_CACHE_OPTIMIZED
        // optimized: flush only those blocks that are deallocated

        fs_bufCacheFlushBlocks(context, block, 1);
#endif

        FS_TRACE(context, FS_TRACE_SNAP, ("  free'd %lld", block));
    }

#ifndef FS_SNAP_CACHE_OPTIMIZED
    // make sure there are no deallocated bufs in the cache
    // calling flush will also reset buf crc for any meta snapbit blocks

    fs_bufCacheFlush(context, NULL);
#endif

    // delete the inode by writing mode 0
    // do for both in-core and on-disk copies

    inode->dinode.size = 0;
    inode->dinode.attrs = 0;
    inode->dinode.linkCnt = 0;
    fs_inodeWrite(context, inode, INO_WRITE_LOG);

    inodeBuf = B_GET_EXCL(NULL, fs_firstInodeBlock(id));
    assert(inodeBuf);
    inodeBuf->u.dinode->dinode[FS_SNAPSHOT_BITS_INODE] = inode->dinode;
    FS_BUF_WRITE(inodeBuf, (FS_SNAPSHOT_BITS_INODE * sizeof (fs_Dinode)), sizeof (fs_Dinode));

    // update blksUsed in superblock

    if (blocksFree) {
        NFX_SPIN_LOCK(&context->fileSystem->superLock);
        context->fileSystem->commitedBlksUsed -= blocksFree;
        context->fileSystem->super->blksUsed -= blocksFree;
        NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
        fs_writeSuperblocksSync(context, context->fileSystem->super);
        FS_TRACE(context, FS_TRACE_SNAP,
                 ("snap remove dealloc snapbits: id %d blocks free %d super blksUsed %lld commited blksUsed %lld",
                 id, blocksFree, context->fileSystem->super->blksUsed,
                 context->fileSystem->commitedBlksUsed));
    }

    // commit and release the transaction context
    // also need to syncThread here to write out the inode

    fs_logCommitEntry(context);
    fs_syncThread(context);
    context->logSummary = NULL;

    // invalidate inode from cache so eek will not find this snapshot
    // fs_inodeRemoveFromCache() will destroy lock so no need to release
    inode = I_GET_EXCL(id, FS_SNAPSHOT_BITS_INODE);
    fs_inodeRemoveFromCache(context, inode, TRUE);

    fs_ramDealloc(FS_RAM_SNAP_NAME, arg0.blocks);

    FS_SNAP_END_TUNE("snap remove dealloc snapbits");

    return NFX_OK;
}


/**
 * Free the inode for a snapshot by clearing the mode field
 */
static void fs_snapFreeInode(
    fs_Context *context,
    fs_Buf *snap,
    fs_Inumber inum,
    int snapnum)
{
#ifdef FS_SNAP_CACHE_OPTIMIZED
    fs_Inode *inode;
#endif

    assert(inum < FS_MAX_META_INODE);

    snap->u.dinode->dinode[inum].attrs = 0;
    snap->u.dinode->dinode[inum].size = 0;
    snap->u.dinode->dinode[inum].linkCnt = 0;
    FS_BUF_WRITE_NO_LOG(snap, inum * sizeof (fs_Dinode), sizeof (fs_Dinode));

#ifdef FS_SNAP_CACHE_OPTIMIZED
    // also remove inode from cache
    inode = I_GET_EXCL(snapnum, inum);
    if (inode == NULL)
        return;
    fs_inodeRemoveFromCache(context, inode, TRUE);
#endif
}


/**
 * Remove necessary meta inodes from the snapshot.
 */
static void fs_snapRemoveMetaInodes(
    fs_Context *context,
    int id)
{
    fs_Buf *snap = B_GET_EXCL(NULL, fs_firstInodeBlock(id)); 

    fs_snapFreeInode(context, snap, FS_INODE_INODE, id);
    fs_snapFreeInode(context, snap, FS_ROOT_INODE, id);
    fs_snapFreeInode(context, snap, FS_FILESET_INFO_INODE, id);
    fs_snapFreeInode(context, snap, FS_FILESET_USER_QUOTA_INODE, id);
    fs_snapFreeInode(context, snap, FS_FILESET_GROUP_QUOTA_INODE, id);
    fs_snapFreeInode(context, snap, FS_TRASHCAN_INODE, id);
    fs_snapFreeInode(context, snap, FS_SECURITY_INODE, id);
    fs_snapFreeInode(context, snap, FS_SNAPSHOT_DIR_INODE, id);
    fs_snapFreeInode(context, snap, FS_LINK_INODE, id);
    fs_snapFreeInode(context, snap, FS_TEMPDIR_INODE, id);
    fs_snapFreeInode(context, snap, FS_NULL_INODE, id);
    fs_snapFreeInode(context, snap, FS_TEMP_DUMP_SWAP, id);
    fs_snapFreeInode(context, snap, FS_AUDIT_INODE, id);
    fs_snapFreeInode(context, snap, FS_SYSTEM_HIDDEN_INODE, id);
    fs_snapFreeInode(context, snap, FS_EXTENDED_ATTR_INODE, id);
    fs_snapFreeInode(context, snap, FS_REF_OVERFLOW_INODE, id);
    fs_snapFreeInode(context, snap, FS_CIFS_SHARES_INODE, id);

    fs_bufReleaseExclusive(context, snap);
}


/**
 * Snapshot remove: finish up remove
 * 1. deallocate snapbits file
 * 2. remove global meta data
 */
int32 fs_snapRemoveFinish(
    fs_Context *context,
    int32 id)
{
    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap remove finish: for snapnum %d", id));

    fs_snapRemoveDeallocSnapBits(context, id);
    fs_snapRemoveMetaInodes(context, id);

#ifdef FS_SNAP_CACHE_OPTIMIZED

    // optimized: don't flush entire buffer and inode caches
    // flush only those for removed snap

    fs_bufCacheFlushSnap(context, NULL, id);
    fs_inodeCacheFlushSnap(context, id);
#endif

    // flush all snap entries in the attribute cache on the txrx
    fs_snapFlushAttrCache(context, id, 0);

    return NFX_OK;
}


/**
 * Deallocates a snapshot id from the snapshots bitmap in the superblock.
 * Modifies the in-core superblock, but doesn't write it to disk.
 */
static void fs_snapDeallocateId(
    fs_Context *context,
    int id)
{
    fs_SuperBlock *super = context->fileSystem->super;
    uint64 bits = super->snapshots;
    uint64 bit;

    bit = 1LL << id;
    assert(bits & bit);
    NFX_SPIN_LOCK(&context->fileSystem->superLock);
    super->snapshots &= ~bit;
    NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
    FS_TRACE(context, FS_TRACE_SNAP,
            ("snap deallocate: id %d bit %llx snapshots before %llx after %llx",
            id, bit, bits, super->snapshots));
}


//    S N A P S H O T   R E V E R T


/**
 * Snapshot revert: remove snapshots that are not in reverted snapshot
 *                  that is, the newer snapshots
 */
static int32 fs_snapRevertRemoveSnaps(
        fs_Context *context,
        int32 id)
{
    fs_SuperBlock *super = context->fileSystem->super;
    int32 i, eof, entries;
    int32 snapnum, snapnums[FS_MAX_SNAPSHOTS];
    uint64 cookie = 0;
    char namebuf[FS_READDIR_SPACE_MAX];
    fs_DirEntry *dirent;
    fs_Inode *dir, *root;

    // get snapshot's snapdir to get list of other snapshots to keep
    // since this snapshot does not see the newer ones, these will be removed

    dir = I_GET_SHARED(id, FS_SNAPSHOT_DIR_INODE);
    assert(dir);

    bzero(snapnums, sizeof(snapnums));

    dirent = (void *) namebuf;
    fs_dir_readdir(context, dir, dirent, sizeof(namebuf), cookie, FALSE,
                   &entries, &eof);

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap revert remove snaps: %d (snapshots %llx) %d entries",
             id, super->snapshots, entries));

    for (i = 0; i < entries; i++) {
        uint64 inum;

        if (!fs_nameIsDotOrDotDot(dirent->name)) {
            inum = dirent->inode;
            assert((inum >= FS_SNAPSHOT_ID_INODE_FIRST) && (inum <= FS_SNAPSHOT_ID_INODE_LAST));
            snapnum = inum - FS_SNAPSHOT_ID_INODE_FIRST;
            assert((snapnum > 0) && (snapnum < FS_MAX_SNAPSHOTS));
            snapnums[snapnum] = 1;
        }

        // get the next dir entry

        cookie = dirent->cookie;
        FS_READDIR_ADV_ENTRY_PTR(dirent);
    }

    fs_inodeRelease(context, dir);

    // do the same for the internal snapshots
    // TBD: can we do this way for user snapshots ??

    root = I_GET_SHARED(id, FS_ROOT_INODE);
    for (snapnum = FS_MAX_USER_SNAPSHOTS+1; snapnum < FS_MAX_SNAPSHOTS; snapnum++) 
    {
        fs_Inode *snapbits = I_GET_SHARED(snapnum, FS_SNAPSHOT_BITS_INODE);

        if (snapbits == NULL) continue;

        // compare internal snap's create time with revert's
        // include if create time is smaller/earlier

        FS_TRACE(context, FS_TRACE_SNAP,
                ("  found internal snapnum %d ctime %d (%d)",
                snapnum, snapbits->dinode.ctime.secs, root->dinode.ctime.secs));

        if (FS_TIME_LE(snapbits->dinode.ctime, root->dinode.ctime))
        {
            snapnums[snapnum] = 1;
        }
        fs_inodeRelease(context, snapbits);
    }
    fs_inodeRelease(context, root);

    // remove all snapbits that are not part of revert
    // also update the superblock's snapshots bits

    for (i = 1; i < FS_MAX_SNAPSHOTS; i++) {
        if ((i != id ) && (snapnums[i] == 0)) {
            fs_Buf *snap = B_GET_EXCL(NULL, fs_firstInodeBlock(i));
            fs_Dinode *dp = &snap->u.dinode->dinode[FS_SNAPSHOT_BITS_INODE];

            uint64 snapbit = 1LL << i;
            assert(snap);
            NFX_SPIN_LOCK(&context->fileSystem->superLock);
            super->snapshots &= ~snapbit;
            NFX_SPIN_UNLOCK(&context->fileSystem->superLock);

            if (dp->attrs) {
                FS_TRACE(context, FS_TRACE_SNAP,
                        ("snap revert remove snaps: removing %d", i));
                dp->attrs = 0;
                dp->size = 0;
                dp->linkCnt = 0;
                FS_BUF_WRITE_NO_LOG(snap,
                        (FS_SNAPSHOT_BITS_INODE * sizeof(fs_Dinode)),
                        sizeof(fs_Dinode));
            }
            fs_bufReleaseExclusive(context, snap);
        }
    }

    // flush all the snapbits inodes we just read in
    // next get of these inodes should read disk versions

    fs_inodeCacheFlush(context);

    // write back superblock with new snapshot bits

    fs_writeSuperblocksSync(context, context->fileSystem->super);

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap revert remove snaps: %d recal snapshots %0llx",
             id, super->snapshots));
    return NFX_OK;
}


#ifndef FS_SNAP_OPTIMIZED
/**
 * Snapshot revert: set the reference bytes using snapbits file
 *   1. clear live bit if not in snapbits file
 *      unless it's a never snap block since these were not included
 *      in the first place in the snapbits file
 *   2. set live bit if in snapbits file
 */
static boolean fs_snapRevertSetRefBytes(
        fs_Context  *context,
        uint8       *snapBits,
        fs_RefByte  *ref,
        uint64      blockStart)
{
    boolean changed = FALSE;
    int i, n;
    uint32 blockSize = context->fileSystem->params.blockSize;
    uint64 block = blockStart;

    assert((blockSize % BITS_PER_BYTE) == 0);

    for (i = 0, n = blockSize / BITS_PER_BYTE; i < n; i++, block += BITS_PER_BYTE) {
        int j;
        uint8 bits = snapBits[i];

        for (j = 0; j < BITS_PER_BYTE; j++, ref++) {
            uint8 mask = 1 << j;
            fs_RefByte byte = *ref;
            boolean live = fs_refbyteIsLive(byte);
            boolean neversnap = live && fs_refbyteNeverSnapshot(byte);
            uint8 bit = bits & mask;

            if (bit) fs_snapStats.revertSnapBitBlks++;
            if (bit && live) fs_snapStats.revertAlreadyLives++;
            if (neversnap) fs_snapStats.revertNeverSnaps++;

            // clear the live bit unless it's never snapped

            if (!bit && live && !neversnap) {
                *ref = byte &~ FS_REFBYTE_IS_LIVE;
                changed = TRUE;
                fs_snapStats.revertLiveClears++;
            }

            // set the live bit if block is in snapbits

            if (bit && !live) {
                *ref = byte | FS_REFBYTE_IS_LIVE;
                changed = TRUE;
                fs_snapStats.revertLiveSets++;
            }
        }
    }

    return changed;
}
#endif


/**
 * Snapshot revert: regenerate reference count file
 */
int32 fs_snapRevertRefBytes(
    fs_Context *context,
    int32 id)
{
#ifdef FS_SNAP_OPTIMIZED 
    fs_Parameters *params = &context->fileSystem->params;
    uint32 blockSize = params->blockSize;
    uint32 chunkSize = blockSize / BITS_PER_BYTE;
    uint64 snapBitsPages;
    uint64 page, block;
    fs_Inode *snapBitsInode;
    fs_Inode *refInode;
    FS_SNAP_WATCHDOG_DECL
    FS_SNAP_TUNE_DECL

#ifdef FS_DEBUG_SNAP_REVERT
    printf("snap revert refbytes: %d before\n", id);
    fs_snapDiagBitsFile(context, NULL, id);
    fs_refBytePrintAllocated(context);
#endif

    FS_SNAP_START_TUNE("snap revert set refbytes")

    // clear or set live using snapbits file

    snapBitsInode = I_GET_SHARED(id, FS_SNAPSHOT_BITS_INODE);
    refInode = I_GET_SHARED(0, FS_REF_COUNT_INODE);
    assert(refInode != NULL);
    assert(snapBitsInode != NULL);

    snapBitsPages = ceilingDiv64(snapBitsInode->dinode.size, blockSize);
    assert(snapBitsPages == snapBitsInode->dinode.numblocks);

    FS_SNAP_WATCHDOG_INIT("revert set refbytes");

    for (page = 0; page < snapBitsPages; page++) {
        int i;
        fs_Buf *snapBuf = O_GET_SHARED(snapBitsInode, page);

        assert(snapBuf != NULL);

        // yield the CPU if necc to prevent watchdog timeouts
        FS_SNAP_WATCHDOG_CHECK("revert set refbytes");

        block = (page * blockSize);
        for (i = 0; i < BITS_PER_BYTE; i++) {
            uint8 *snapBits = &snapBuf->u.data[i * chunkSize];
            uint64 refPage = page * BITS_PER_BYTE + i;
            fs_Buf *refBuf = O_GET_EXCL(refInode, refPage);
            fs_RefByte *refbytes;

            if (refBuf == NULL) {
                // nearing the end of the file
                assert(page + 1 == snapBitsPages);
            }
            else {
                int j, k;
                uint64 word, snapbits;

                refbytes = refBuf->u.refBytes;
                for (j = 0, k = 0; j < blockSize / BITS_PER_BYTE; j += 4, k += 4*sizeof (uint64)) {
                    uint8 bits;
                    uint64 live, live_mask, never_mask;

                    live_mask = ((uint64)FS_REFBYTE_IS_LIVE
                                | ((uint64)FS_REFBYTE_IS_LIVE << 8)
                                | ((uint64)FS_REFBYTE_IS_LIVE << 16)
                                | ((uint64)FS_REFBYTE_IS_LIVE << 24)
                                | ((uint64)FS_REFBYTE_IS_LIVE << 32)
                                | ((uint64)FS_REFBYTE_IS_LIVE << 40)
                                | ((uint64)FS_REFBYTE_IS_LIVE << 48)
                                | ((uint64)FS_REFBYTE_IS_LIVE << 56));
                    never_mask = ((uint64)FS_REFBYTE_NEVER_SNAPSHOT
                                |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 8)
                                |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 16)
                                |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 24)
                                |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 32)
                                |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 40)
                                |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 48)
                                |((uint64)FS_REFBYTE_NEVER_SNAPSHOT << 56));
/*
 * Unrolled loop to set the live and never snapshot bits in the refbytes:
 *
 * The live bit is set if a) set in snapbits
 *                        b) never snap and live are both set in refbyte
 * The live bit is cleared if c) not set in snapbits
 *
 * b) is accomplished by masking out the live bits, shifting the never
 * snapshot bit into the position of the live bits and doing an AND.
 * a) and c) are accomplished by "expanding" the byte of snapbits to
 * a 64 bit value, and added to the existing bits.
 * Finally, the never mask bits are added back to the bits.
 */
#define SET_LIVE do{\
    snapbits = ((uint64) (bits & 0x1))                      \
                | ((uint64) ((bits & 0x2) ? 1 : 0) << 8)    \
                | ((uint64) ((bits & 0x4) ? 1 : 0) << 16)   \
                | ((uint64) ((bits & 0x8) ? 1 : 0) << 24)   \
                | ((uint64) ((bits & 0x10) ? 1 : 0) << 32)  \
                | ((uint64) ((bits & 0x20) ? 1 : 0) << 40)  \
                | ((uint64) ((bits & 0x40) ? 1 : 0) << 48)  \
                | ((uint64) ((bits & 0x80) ? 1 : 0) << 56); \
    live = word & live_mask;    /* mask out the LIVE bits                   */\
    live &= (word & never_mask) >> 1;   /* shift the NEVER_SNAPSHOT and AND */\
    live |= snapbits;                   /* include snapbits                 */\
    live |= (word & never_mask);        /* include never snap bits          */\
    word = live;                                                              \
} while (0)

                    bits = snapBits[j];
                    word = *(uint64 *)&refbytes[k];
                    SET_LIVE;
                    *(uint64 *)&refbytes[k] = word;

                    bits = snapBits[j+1];
                    word = *(uint64 *)&refbytes[k+8];
                    SET_LIVE;
                    *(uint64 *)&refbytes[k+8] = word;

                    bits = snapBits[j+2];
                    word = *(uint64 *)&refbytes[k+16];
                    SET_LIVE;
                    *(uint64 *)&refbytes[k+16] = word;

                    bits = snapBits[j+3];
                    word = *(uint64 *)&refbytes[k+24];
                    SET_LIVE;
                    *(uint64 *)&refbytes[k+24] = word;
                }
                FS_BUF_WRITE_NO_LOG(refBuf, 0, blockSize);
                fs_bufReleaseExclusive(context, refBuf);
            }
        }
        fs_bufRelease(context, snapBuf);
    }

    fs_inodeRelease(context, refInode);
    fs_inodeRelease(context, snapBitsInode);

    FS_SNAP_END_TUNE("snap revert set refbytes")

    // regenerate snapcount using all snapbits file
    fs_snapRecalculateRefCnts(context);

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap revert refbytes: for snapnum %d", id));

#ifdef FS_DEBUG_SNAP_REVERT
    printf("snap revert refbytes: %d after\n", id);
    fs_refBytePrintAllocated(context);
#endif

    return NFX_OK;
#else
    int32 rc;
    FS_SNAP_TUNE_DECL

#ifdef FS_DEBUG_SNAP_REVERT
    printf("snap revert refbytes: %d before\n", id);
    fs_snapDiagBitsFile(context, NULL, id);
    fs_refBytePrintAllocated(context);
#endif

    // regenerate snapcount using all snapbits file
    fs_snapRecalculateRefCnts(context);

    FS_SNAP_START_TUNE("snap revert set refbytes")

    // clear or set live using snapbits file

    rc = fs_snapTransformRefBytes(context, id, fs_snapRevertSetRefBytes);

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap revert refbytes: for snapnum %d %s",
             id, ((rc == NFX_OK) ? "ok" : "err")));

    FS_SNAP_END_TUNE("snap revert set refbytes")

#ifdef FS_DEBUG_SNAP_REVERT
    printf("snap revert refbytes: %d after\n", id);
    fs_refBytePrintAllocated(context);
#endif

    return rc;
#endif
}


/**
 * Fill all of the blocks of a file with zeros
 */
void
fs_snapZeroFile(
    fs_Context  *context,
    uint64      inum,
    uint64      startPage)
{
    fs_Inode *inode = I_GET_SHARED(0, inum);
    uint32 blockSize = context->fileSystem->params.blockSize;
    uint64 blocks = inode->dinode.size / blockSize;
    uint64 offset;
    fs_WriteQueue queue;
    char   *zeroes;

    assert(inode != NULL);
    assert(blockSize <= FS_MAX_BLOCK_SIZE);

    fs_initWriteQueue(context, &queue, THREAD_QUEUE_MAX);

    // Allocate a max block size chunk of zeroes, and ensure
    // its aligned such that the sibyte won't choke on it.

    zeroes = fs_ramAlloc(FS_RAM_CORE_MISC, FS_MAX_BLOCK_SIZE + MALLOC_ALIGNED_SIZE);
    zeroes += (MALLOC_ALIGNED_SIZE - 1);
    (address_t) zeroes &= ~(MALLOC_ALIGNED_SIZE - 1);
    bzero(zeroes, FS_MAX_BLOCK_SIZE);

    for (offset = startPage; offset < blocks; offset++)
    {
        fs_Block block = fs_emap_lookup_block(context, inode, offset);
        assert(fs_blockIsValid(block));
        fs_queueWriteBlock(&queue, block, blockSize, zeroes);
    }

    fs_writeQueueSync(&queue);
    fs_ramDealloc(FS_RAM_CORE_MISC, zeroes);
    fs_inodeRelease(context, inode);
}


/**
 * Snapshot revert: using reference count file, add free extents back to buddy free file
 */
int32 fs_snapRevertBuddy(
    fs_Context *context,
    int32 id)
{
    fs_Parameters   *params = &context->fileSystem->params;
    uint64          used = 0;
    uint64          block, firstBlock, lastBlock;
    fs_buddy_t      *buddies = context->fileSystem->buddies;
    int             loop;
    fs_Inode        *refInode;
    FS_SNAP_WATCHDOG_DECL
    FS_SNAP_TUNE_DECL
    FS_SNAP_START_TUNE("snap revert buddy");

#ifdef FS_DEBUG_SNAP_REVERT
    printf("snap revert buddy: refcnts file ->\n");
    fs_refBytePrintAllocated(context);
#endif

    refInode = I_GET_SHARED(0, FS_REF_COUNT_INODE);
    assert(refInode != NULL);

    // zero out the buddy file before adding free extents
    fs_snapZeroFile(context, FS_FREE_BUDDY_TREE_INODE, 0);

    // invalidate the buddies' summary
    for (loop = 0; loop < params->buddyRegionCount; loop++)
    {
        if (buddies[loop].summary != NULL)
        {
            NFX_SPIN_LOCK(&(buddies[loop].spin));
            fs_ramDealloc(FS_RAM_REGION_SUMMARY, buddies[loop].summary);
            buddies[loop].flags = 0;
            buddies[loop].summary = NULL;
            NFX_SPIN_UNLOCK(&(buddies[loop].spin));
        }
    }

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap revert buddy: for snapnum %d blksUsed %lld totalBlocks %lld",
             id, context->fileSystem->super->blksUsed,
             params->totalBlocks));

    // create a transaction context (necc for free'ing block)

    assert(context->logSummary == NULL);
    fs_logBeginTransaction(context);

    FS_SNAP_WATCHDOG_INIT("revert buddy");

    for (block = 0; block < params->totalBlocks; block += params->blockSize) {
        uint64 pageno = block >> params->blockBytesShift;
        fs_Buf *buf = O_GET_SHARED(refInode, pageno);
        fs_RefByte *ref;
        uint64 max_block;
        uint64 i;

        assert(buf != NULL);

        ref = buf->u.refBytes;
        if (block + params->blockSize > params->totalBlocks)
            max_block = params->totalBlocks % params->blockSize;
        else
            max_block = params->blockSize;

        for (i = 0; i < max_block; ) {
            
            // yield the CPU if necc to prevent watchdog timeouts
            FS_SNAP_WATCHDOG_CHECK("transform refbytes");

            // skip non zero (in use) reference bytes

            if (ref[i]) {
                i++;
                used++;
            } else {

                // find first and last blocks to add to buddy free tree

                int32 len = 1;    // found the firstBlock, start with next
                firstBlock = block + i++;
                while ((i < max_block) && (i < params->totalBlocks)) {
                    if (ref[i++]) {
                        used++;
                        break;
                    }
                    len++;
                }
                lastBlock = firstBlock + len - 1;
                if (len) {
                    FS_TRACE(context, FS_TRACE_SNAP,
                             ("snap revert buddy: add %lld to %lld (len %d)",
                             firstBlock, lastBlock, len));
                    fs_buddyMaybeAddExtent(context,
                                           fs_makeExtent(len, fs_makeBlock(firstBlock)),
                                           TRUE);
                    fs_snapStats.revertBuddyAdds += len;
                }
            }
        }
        fs_bufRelease(context, buf);

#if defined(FS_DEBUG_SNAP_CRASH)
        if ((fs_snapPhaseToCrash == FS_SNAP_PHASE_DURING_REVERT_BUDDY) &&
            (block > (params->totalBlocks >> 1))) {
            char log[1024];
            printf("snap force crash (freeblocks after %lld)\n",
                    (params->totalBlocks >> 1));
            fs_logMakeDumpLogRecord(context, log, 1023);
            printf("log dump record:\n");
            printf("%s\n", log);
            fs_debugCrash();
        }
#endif

    }

    fs_inodeRelease(context, refInode);

    // commit and release the transaction context

    fs_logCommitEntry(context);
    context->logSummary = NULL;

    // update superblock

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap revert buddy: recal blksUsed %lld", used));

    NFX_SPIN_LOCK(&context->fileSystem->superLock);
    context->fileSystem->super->blksUsed = used;
    NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
    fs_writeSuperblocksSync(context, context->fileSystem->super);

    fs_snapStats.totalBlocks = params->totalBlocks;
    fs_snapStats.usedBlocks = used;

    FS_SNAP_END_TUNE("snap revert buddy");

    return NFX_OK;
}


/**
 * Snapshot revert: regenerate the FS_FULL_INODE_BLOCK_INODE and
 *                  FS_DIRECTORY_IN_INODE_BLOCK_INODE files
 */
static int32 fs_snapRevertInodeBlockFiles(
        fs_Context *context,
        int32 id)
{
    fs_Inode *inodeFile = I_GET_SHARED(id, FS_INODE_INODE);
    fs_Inode *inosum;
    fs_Inode *dirsum;
    fs_Parameters *params = &context->fileSystem->params;
    int i, n;
    uint64 max_pageno, ipage, iblks, isize;
    uint64 used = 0;
    fs_Inumber base;
    boolean full, hasDir;
    FS_SNAP_TUNE_DECL
    FS_SNAP_START_TUNE("snap revert summary inodes");

    assert(inodeFile);

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap revert inode blocks: %d inodesUsed %lld maxInodeBlocks %d",
             id, context->fileSystem->super->inodesUsed,
             params->maxInodeBlocks));

    // zero out both inode block files before setting bits

    fs_snapZeroFile(context, FS_FULL_INODE_BLOCK_INODE, 0);
    fs_snapZeroFile(context, FS_DIRECTORY_IN_INODE_BLOCK_INODE, 0);

    inosum = I_GET_SHARED(0, FS_FULL_INODE_BLOCK_INODE);
    dirsum = I_GET_SHARED(0, FS_DIRECTORY_IN_INODE_BLOCK_INODE);
    assert(inosum);
    assert(dirsum);

    max_pageno = ceilingDiv64(params->maxInodeNumber, FS_INODES_PER_PAGE);
    
    // create a transaction context for fs-bitmap module

    assert(context->logSummary == NULL);
    fs_logBeginTransaction(context);

    // mark the first bits in the inode summary files
    // and mark past the end of the file system in the inode summaries

    for (i = 0, n = 0; n < FS_MAX_RESERVED_INODE; i++, n += FS_INODES_PER_PAGE)
    {
        fs_bitmapSet(context, inosum, i, NULL);
        fs_bitmapSet(context, dirsum, i, NULL);
    }
    fs_bitmapSetBitsAfter(context, inosum, params->maxInodeBlocks);
    fs_bitmapSetBitsAfter(context, dirsum, params->maxInodeBlocks);

    // init inode parameters for the superblock update
    // count the super nodes (page 0)

    iblks = 1;
    isize = FS_BLOCK_SIZE;

    for (ipage = 1; ipage < max_pageno; ipage++) {
        fs_Buf *buf = O_GET_SHARED(inodeFile, ipage);
        int i;

        if (buf == NULL)
            continue; // hole in the file
        
        // keep track of inode parameters for the superblock update

        iblks += 1;
        isize = (ipage + 1) * FS_BLOCK_SIZE;

        base = buf->page * params->inodesPerBlock;
        full = TRUE;
        hasDir = FALSE;

        for (i = 0; i < FS_INODES_PER_PAGE; i++) {
            fs_Dinode *dinode = &buf->u.dinode->dinode[i];

            if (dinode->attrs == 0) {
                full = FALSE;
            } else {
                used++;
                if (FS_I_IS_DIR(dinode->attrs)) {
                    hasDir = TRUE;
                    fs_snapStats.revertDInodes++;
                } else {
                    fs_snapStats.revertInodes++;
                }
            }
        } // for FS_INODES_PER_PAGE

        if (full) {
            fs_bitmapSet(context, inosum, ipage, NULL);
        }
        else {
            fs_bitmapClearIfSet(context, inosum, ipage, NULL);
        }

        // if there is a directory in this page, the dirsum bit
        // can be 1 or 0 since removing a directory clears this bit
        // even if there are other directories in the same inode page
        // for us, we'll just set it

        if (hasDir) {
            fs_bitmapSet(context, dirsum, ipage, NULL);
        }
        fs_bufRelease(context, buf);

#if defined(FS_DEBUG_SNAP_CRASH)
        if ((fs_snapPhaseToCrash == FS_SNAP_PHASE_DURING_REVERT_INODES) &&
            (ipage == 1)) {
            char log[1024];
            printf("snap force crash (inodes after page 1)\n");
            fs_logMakeDumpLogRecord(context, log, 1023);
            printf("log dump record:\n");
            printf("%s\n", log);
            fs_debugCrash();
        }
#endif
    }

    fs_inodeRelease(context, inodeFile);
    fs_inodeRelease(context, inosum);
    fs_inodeRelease(context, dirsum);

    // commit and release the transaction context

    fs_logCommitEntry(context);
    context->logSummary = NULL;

    // update superblock

    NFX_SPIN_LOCK(&context->fileSystem->superLock);
    context->fileSystem->super->inodesUsed = used;
    context->fileSystem->super->inodeFileBlks = iblks;
    context->fileSystem->super->inodeFileSize = isize;
    NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
    fs_writeSuperblocksSync(context, context->fileSystem->super);

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap revert inode blocks: recal inodesUsed %lld INODE_INODE blks %lld INODE_INODE size %lld",
             used, iblks, isize));

    fs_snapStats.usedInodes = used;

    FS_SNAP_END_TUNE("snap revert summary inodes");

    return NFX_OK;
}


/**
 * Snapshot revert: copy meta inodes from the snapshot to the live file system.
 */
int32 fs_snapRevertMetaData(
    fs_Context *context,
    int32 id)
{
    fs_Buf  *live = B_GET_EXCL(NULL, fs_firstInodeBlock(0));
    fs_Buf  *snap = B_GET_SHARED(NULL, fs_firstInodeBlock(id));

    // copy meta inodes from snapshot to live file system

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap revert meta: for snapnum %d", id));

    fs_snapCopyInode(context, live, snap, FS_INODE_INODE);
    fs_snapCopyInode(context, live, snap, FS_ROOT_INODE);
    fs_snapCopyInode(context, live, snap, FS_SNAPSHOT_DIR_INODE);
    fs_snapCopyInode(context, live, snap, FS_FILESET_INFO_INODE);
    fs_snapCopyInode(context, live, snap, FS_FILESET_USER_QUOTA_INODE);
    fs_snapCopyInode(context, live, snap, FS_FILESET_GROUP_QUOTA_INODE);
    fs_snapCopyInode(context, live, snap, FS_TRASHCAN_INODE);
    fs_snapCopyInode(context, live, snap, FS_SECURITY_INODE);
    fs_snapCopyInode(context, live, snap, FS_LINK_INODE);
    fs_snapCopyInode(context, live, snap, FS_TEMPDIR_INODE);
    fs_snapCopyInode(context, live, snap, FS_NULL_INODE);
    fs_snapCopyInode(context, live, snap, FS_TEMP_DUMP_SWAP);
    fs_snapCopyInode(context, live, snap, FS_AUDIT_INODE);
    fs_snapCopyInode(context, live, snap, FS_SYSTEM_HIDDEN_INODE);
    fs_snapCopyInode(context, live, snap, FS_EXTENDED_ATTR_INODE);
    fs_snapCopyInode(context, live, snap, FS_REF_OVERFLOW_INODE);
    fs_snapCopyInode(context, live, snap, FS_CIFS_SHARES_INODE);

    fs_bufRelease(context, snap);
    fs_bufReleaseExclusive(context, live);

    return NFX_OK;
}


/**
 * Snapshot revert: finish up revert
 *   remove the snap bits from the live file system.
 *   That is, clear the live-block indicator for the snap bits file.
 */
int32 fs_snapRevertFinish(
    fs_Context *context,
    int32 id)
{
    FS_SNAP_TUNE_DECL
    FS_SNAP_START_TUNE("snap revert finish");

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap revert clear snapbits: for snapnum %d", id));

    fs_snapClearLiveSnapBits(context, id);

    // PR #2629 cleanup in core filesets and quotas
    // and rebuild from reverted meta inodes

    //fs_cleanFileSetNQuotaLists(context, context->fileSystem);
    fs_buildFileSetList(context, context->fileSystem);

    // clear all caches of stale data
    fs_inodeCacheFlush(context);
    fs_bufCacheFlush(context, NULL);

    FS_SNAP_END_TUNE("snap revert finish");

    return NFX_OK;
}


/**
 * Driver for the snapshot creation and removal process.  Advances the
 * process for the specified snapshot id from the specified start phase,
 * until it fails or the operation completes successfully.
 *
 * When this function returns, the state in the superblock should always
 * be ready.
 */
static int32 fs_snapDoIt(
    fs_Context *context,
    int id,
    fs_SnapPhase phase)
{
#ifndef FS_SNAP_CACHE_OPTIMIZED
    int32 snap = id;
#endif
    int32 rc = NFX_OK;
    boolean terminate = FALSE;
    fs_FileSystem *fs;
    FS_SNAP_TUNE_DECL

    assert(0 < id && id < FS_MAX_SNAPSHOTS);
    assert(phase != FS_SNAP_PHASE_READY);
    fs = context->fileSystem;

    do {
        fs_SnapPhase nextPhase = FS_SNAP_PHASE_READY;
        
        switch (phase) {

            // No snapshot operation in progress.
            default:
                fs_error(context, FS_INTERNAL_ERR,
                         "unexpected snapshot phase: %d", phase);
                /* NOTREACHED */
                break;

            // Snapshot creation:  allocating snap bits and sync'ing.
            case FS_SNAP_PHASE_CREATE_ALLOC:
                rc = fs_snapCreateAllocSnapBits(context, id);
                nextPhase = FS_SNAP_PHASE_CREATE_SNAP_BITS;
                break;
                
            // Snapshot creation:  reading reference counts, writing snap bits.
            case FS_SNAP_PHASE_CREATE_SNAP_BITS:
                rc = fs_snapCreateWriteSnapBits(context, id);
                nextPhase = FS_SNAP_PHASE_CREATE_REF_COUNTS;
                break;
                
            // Snapshot creation:  incrementing reference counts.
            case FS_SNAP_PHASE_CREATE_REF_COUNTS:
                rc = fs_snapCreateIncrementRefBytes(context, id);
                nextPhase = FS_SNAP_PHASE_CREATE_META;
                break;
                
            // Snapshot creation:  writing the global metadata.
            case FS_SNAP_PHASE_CREATE_META:
                rc = fs_snapCreateFinish(context, id);
                nextPhase = FS_SNAP_PHASE_READY;
                break;
                

            // Snapshot removal:  decrementing the reference counts.
            case FS_SNAP_PHASE_REMOVE_REF_COUNTS:
                rc = fs_snapRemoveDecrementRefBytes(context, id);
#ifdef FS_SNAP_OPTIMIZE_REMOVE_FREE_BLOCKS
                nextPhase = FS_SNAP_PHASE_REMOVE_SNAP_BITS_AND_FREE;
#else
                nextPhase = FS_SNAP_PHASE_REMOVE_SNAP_BITS;
#endif
                break;
 
#ifndef FS_SNAP_OPTIMIZE_REMOVE_FREE_BLOCKS
            // Snapshot removal:  update snap bits.
            case FS_SNAP_PHASE_REMOVE_SNAP_BITS:
                rc = fs_snapRemoveUpdateSnapBits(context, id);
                nextPhase = FS_SNAP_PHASE_REMOVE_FREE_BLOCKS;
                break;
#endif

            // Snapshot removal:  update snap bits and free blocks.
            case FS_SNAP_PHASE_REMOVE_SNAP_BITS_AND_FREE:
                rc = fs_snapRemoveUpdateSnapBitsAndFree(context, id);
                nextPhase = FS_SNAP_PHASE_REMOVE_FREE_SNAP_BITS;
                break;
                
            // Snapshot removal:  freeing blocks based on snap bits.
            case FS_SNAP_PHASE_REMOVE_FREE_BLOCKS:
                rc = fs_snapRemoveFreeBlocks(context, id);
                nextPhase = FS_SNAP_PHASE_REMOVE_FREE_SNAP_BITS;
                break;
                
            // Snapshot removal:  freeing the snap bits and deallocate metadata.
            case FS_SNAP_PHASE_REMOVE_FREE_SNAP_BITS:
                rc = fs_snapRemoveFinish(context, id);
                nextPhase = FS_SNAP_PHASE_READY;
                break;


            // Snapshot revert: remove newer snapshots from the live system
            case FS_SNAP_PHASE_REVERT_REMOVE_SNAPS:
                rc = fs_snapRevertRemoveSnaps(context, id);
                nextPhase = FS_SNAP_PHASE_REVERT_REF_COUNTS;
                break;

            // Snapshot revert: recalculate reference counts from snapbits file
            case FS_SNAP_PHASE_REVERT_REF_COUNTS:
                rc = fs_snapRevertRefBytes(context, id);
                nextPhase = FS_SNAP_PHASE_REVERT_BUDDY;
                break;

            // Snapshot revert: recalculate free buddy tree from reference counts
            case FS_SNAP_PHASE_REVERT_BUDDY:
                rc = fs_snapRevertBuddy(context, id);
                nextPhase = FS_SNAP_PHASE_REVERT_INODE_BLOCKS;
                break;

            // Snapshot revert: regenerate inode and directory block files
            case FS_SNAP_PHASE_REVERT_INODE_BLOCKS:
                rc = fs_snapRevertInodeBlockFiles(context, id);
                nextPhase = FS_SNAP_PHASE_REVERT_META;
                break;

            // Snapshot revert: copy snapshot snodes to live system
            case FS_SNAP_PHASE_REVERT_META:
                rc = fs_snapRevertMetaData(context, id);
                nextPhase = FS_SNAP_PHASE_REVERT_CLEAR_SNAP_BITS;
                break;

            // Snapshot revert: finish the revert (clear snapbits)
            case FS_SNAP_PHASE_REVERT_CLEAR_SNAP_BITS:
                rc = fs_snapRevertFinish(context, id);
                nextPhase = FS_SNAP_PHASE_READY;
                break;
        }

        // fs_lockPrintOwned(context);

        // rc should never be NFX_ERR
        assert(rc == NFX_OK);

#if defined(FS_DEBUG_SNAP_CRASH)
        if (fs_snapPhaseToCrash != -1) {
            if (phase == fs_snapPhaseToCrash) {
                printf("snap crash test: %d %s\n",
                        fs_snapPhaseToCrash, fs_snapPhaseName[fs_snapPhaseToCrash]);
                fs_debugCrash();
            }
            if ((fs_snapPhaseToCrash == FS_SNAP_PHASE_READY) &&
                (nextPhase == FS_SNAP_PHASE_READY)) {
                fs_debugCrash();
            }
        }
#endif

        FS_SNAP_START_TUNE("snap do it syncs");

        // write out what we have so far before setting next phase

        fs_syncThread(context);
        fs_syncThread(context);
        fs_syncThread(context);

        FS_SNAP_END_TUNE("snap do it syncs");

        if (nextPhase == FS_SNAP_PHASE_READY)
            id = 0;
        fs_snapSetState(context, id, nextPhase);

        // PR #1834: don't hang up umount

        NFX_SPIN_LOCK(&fs->spin);
        if (fs->snapInfo.terminate)
        {
            terminate = TRUE;
            fs->snapInfo.terminate = FALSE;
        }
        NFX_SPIN_UNLOCK(&fs->spin);
        if (terminate) {
            printf("snapDoIt: id %d received terminate\n", id);
            fs_thawAll(context);
            if (context->reqHdr != NULL)
            {
                // Send an error response to this context's requestor.
                (fs_reqTypeTbl[context->reqHdr->requestType].respFunc)
                                                    (context, NFX_ERR);
                (fs_reqTypeTbl[context->reqHdr->requestType].clnupFunc)
                                                    (context, FALSE);
                thread_exit();
            }
            else
            {
                // snapshot system thread
                fs_threadTerminate(context);
            }
        }

        phase = nextPhase;

    } while (phase != FS_SNAP_PHASE_READY);

#ifndef FS_SNAP_CACHE_OPTIMIZED
    fs_inodeCacheFlush(context);
    fs_bufCacheFlush(context, NULL);

    FS_SNAP_START_TUNE("snap do it dnlc dirhas");

    fs_dnlcDestroySnapnum(context->fileSystem, snap);
    fs_dirhash_destroy_snapnum(context->fileSystem, snap);
#endif

    FS_SNAP_END_TUNE("snap do it dnlc dirhas");

    return NFX_OK;
}


//  U T I L I T I E S


/**
 * Snapshot "printf"
 */
static void fs_snapPrintStatus(
            fs_Context *context,
            int32 level,
            char *fmt, ...)
{
    eeeGMT_t date;
    extern int vsprintf(char *str, const char *format, va_list ap);
    char string[512], vstring[256], dates[32];
    va_list ap;
    extern int elog_enabled;

    va_start(ap, fmt);
    eee_TimeGMT (&date);

    vsnprintf(vstring, 256, fmt, ap);
    if (!elog_enabled) {
        sprintf(dates, "%s", eee_humandate(date.sec));
        sprintf(string, "%s [%s] %s",
                dates, context->fileSystem->super->mountName, vstring);
        // print out on FP console
        printf("%s\n", string);
    } else {
        sprintf(string, "%s %s",
                context->fileSystem->super->mountName, vstring);
        E_LOG(class_1, LOG_INFO, fs_applicationId, 0, 0, 0, ("%s", string));
    }
    va_end(ap);
}


/**
 * Called during mount unclean to check if snap revert
 * will recalculate blksUsed in superblock
 */
boolean fs_snapRecalculateBlksUsed(
        fs_Context *context)
{
    fs_SuperBlock *super = context->fileSystem->super;
    return fs_snapIsRevertPhase(super->snapState.phase);
}


/**
 * Called during mount unclean to check if snap revert
 * will recalculate inodesUsed in superblock
 */
boolean fs_snapRecalculateInodesUsed(
        fs_Context *context)
{
    fs_SuperBlock *super = context->fileSystem->super;
    return fs_snapIsRevertPhase(super->snapState.phase);
}


/**
 * Get the names of all snapshots for this volume.  Returns the count of
 * names returned.  The names of the snapshots are filled into the names
 * array.  The actual bytes in the names are written to the namebuf;  any
 * names which there isn't space for, according to maxBytes, are omitted.
 */
int32 fs_snapGetNames(
    fs_Context *context,
    char *names[FS_MAX_SNAPSHOTS],
    char *namebuf,
    uint32 maxBytes)
{
    char *s = namebuf;
    int32 remaining = maxBytes;
    uint64 cookie = 0;
    int32 count = 0, snapnum;
    int32 eof;
    fs_Inode *dir;

    if (context->fileSystem->state != FS_VOL_STATE_ONLINE) {
        return 0;
    }

    dir = I_GET_SHARED(0, FS_SNAPSHOT_DIR_INODE);
    do {
        int32 i;
        int32 entries;
        char buf[1024];
        fs_DirEntry *dirent = (void *) buf;
        
        fs_dir_readdir(context, dir, dirent, sizeof buf, cookie, FALSE,
                       &entries, &eof);

        for (i = 0; i < entries; i++) {
            char *name = dirent->name;
            uint64 inum = dirent->inode;
            int len = strlen(name) + 1;

            // skip dot and dot dot
            if (fs_nameIsDotOrDotDot(name)) {
                FS_READDIR_ADV_ENTRY_PTR(dirent);
                continue;
            }

            // get snapnum and do some sanity checks
            assert((inum >= FS_SNAPSHOT_ID_INODE_FIRST) && (inum <= FS_SNAPSHOT_ID_INODE_LAST));
            snapnum = inum - FS_SNAPSHOT_ID_INODE_FIRST;
            assert((snapnum > 0) && (snapnum < FS_MAX_SNAPSHOTS));

            // only return the user snapshots
            if (snapnum > FS_MAX_USER_SNAPSHOTS) {
                FS_READDIR_ADV_ENTRY_PTR(dirent);
                continue;
            }
            if (len >= remaining)
                goto breakOuterLoop;
            memcpy(s, name, len);
            names[count] = s;
            s += len;
            remaining -= len;
            count += 1;
            cookie = dirent->cookie;
            FS_READDIR_ADV_ENTRY_PTR(dirent);
        }
    } while (!eof);

  breakOuterLoop:
    fs_inodeRelease(context, dir);
    return count;
}


/**
 * Return the snapshot id for the named snapshot, or a negative value if
 * there is no snapshot with the specified name.
 */
static int fs_snapLookup(
    fs_Context *context,
    char *name)
{
     int id;
     fs_SnapPhase phase;
     fs_Inode *snapdir = I_GET_SHARED(0, FS_SNAPSHOT_DIR_INODE);
     fs_Inumber inum = fs_dir_lookupent(context, snapdir, name, strlen(name));
     fs_inodeRelease(context, snapdir);

     if (inum == 0) {
         //printf("snap lookup: %s not found\n", name);
         return -1;
     } else {
         fs_Inode *snapbits;
         int snapnum = inum - FS_SNAPSHOT_ID_INODE_FIRST;
         assert(snapnum > 0);
         assert(snapnum < FS_MAX_SNAPSHOTS);
         // PR #2424: verify snapbits exists, operation complete
         snapbits = I_GET_SHARED(snapnum, FS_SNAPSHOT_BITS_INODE);
         if (snapbits == NULL) {
            //printf("snap lookup: %s %d no snapbits\n", name, snapnum);
            return -1;
         }
         fs_inodeRelease(context, snapbits);
         phase = fs_snapGetState(context, &id);
         if ((phase != FS_SNAP_PHASE_READY) && (id == snapnum)) {
            //printf("snap lookup: %s %d not complete\n", name, snapnum);
            return -1;
         }
         return snapnum;
     }
}


/**
 * Lock down a snapshot to prevent snapshot removal while
 * an application (eg NDMP) is accessing it
 */
int32 fs_snapLockSnap(
        fs_Context *context,
        char *name)
{
    int snapnum = fs_snapLookup(context, name); 

    if (snapnum == -1) {
        return NFX_ERR;
    }
    fs_lockExclusive(context, &context->fileSystem->snapInfo.accessLocks[snapnum], NULL);
    FS_TRACE(context, FS_TRACE_SNAP, ("snap lock: acquired for %s snapnum %d\n",
            name, snapnum));
    return NFX_OK;
}


/**
 * Unlock a snapshot that was previously locked
 */
int32 fs_snapUnLockSnap(
        fs_Context *context,
        char *name)
{
    int snapnum = fs_snapLookup(context, name); 

    if (snapnum == -1) {
        return NFX_ERR;
    }
    if (!fs_lockHeldExclusive(context,
                &context->fileSystem->snapInfo.accessLocks[snapnum])) {
        return NFX_ERR;
    }

    fs_unlockExclusive(context, &context->fileSystem->snapInfo.accessLocks[snapnum]);
    FS_TRACE(context, FS_TRACE_SNAP, ("snap unlock: %s snapnum %d\n", name, snapnum));
    return NFX_OK;
}


/**
 * Rename snapshot with oldname to newname
 */
int32 fs_snapRename(
    fs_Context *context,
    char *oldname,
    char *newname)
{
    fs_Inode *snapdir;
    fs_Inumber inum;

    // newname cannot be a reserved name
    if (FS_SNAP_RESERVE_NAME(newname)) {
        return FS_ERR_SNAP_INVAL_NAME;
    }

    snapdir = I_GET_EXCL(0, FS_SNAPSHOT_DIR_INODE);
    assert(snapdir);

    inum = fs_dir_lookupent(context, snapdir, newname, strlen(newname));
    if (inum != 0) {
        fs_inodeReleaseExclusive(context, snapdir);
        return FS_ERR_SNAP_EXISTS;
    }

    inum = fs_dir_lookupent(context, snapdir, oldname, strlen(oldname));
    if (inum == 0) {
        fs_inodeReleaseExclusive(context, snapdir);
        return FS_ERR_SNAP_NO_EXIST;
    }

    // create a transaction context

    assert(context->logSummary == NULL);
    fs_logBeginTransaction(context);

    // remove old entry and add with new name but same inum snapnum

    fs_dir_removeent(context, snapdir, oldname, strlen(oldname));
    fs_dir_addent(context, snapdir, newname, strlen(newname), inum);
    snapdir->dinode.mtime = fs_getTimeOfDay();
    snapdir->dinode.atime = snapdir->dinode.mtime;
    fs_inodeWrite(context, snapdir, INO_WRITE_LOG);

    // commit and release the transaction context

    fs_logCommitEntry(context);
    context->logSummary = NULL;

    // flush .snapshot from the attribute cache on the txrx

    fs_snapFlushAttrCache(context, 0, FS_SNAPSHOT_DIR_INODE);

    fs_snapPrintStatus(context, LOG_NOTICE,
                       "snapshot %s renamed to %s",
                       oldname, newname);
    return NFX_OK;
}

        
/**
 * Return TRUE if snapname is pinned
 */
static boolean fs_snapIsPin(
    fs_Context *context,
    char *snapname)
{
    fs_Inode *snapBitsInode;
    boolean pin;
    uint32 snapnum;

    snapnum = fs_getSnapNum(context, snapname);
    if (snapnum < 0) {
        return FALSE;
    }

    snapBitsInode = I_GET_SHARED(snapnum, FS_SNAPSHOT_BITS_INODE);
    if (snapBitsInode == NULL) {
        return FALSE;
    }

    pin = (snapBitsInode->dinode.attrs & FS_IFLG_SNAP_PIN);
    fs_inodeRelease(context, snapBitsInode);

    return pin;
}


/**
 * Pin a snapshot so it can not be removed until it is un pinned
 */
int32 fs_snapPin(
        fs_Context *context,
        int snapnum)
{
    fs_Inode *snapBitsInode;
   
    snapBitsInode = I_GET_EXCL(snapnum, FS_SNAPSHOT_BITS_INODE);
    if (snapBitsInode == NULL) {
        return NFX_ERR;
    }

    if ((snapBitsInode->dinode.attrs & FS_IFLG_SNAP_PIN) != 0) {
        fs_inodeReleaseExclusive(context, snapBitsInode);
        return NFX_OK;
    }

    // create a transaction context

    assert(context->logSummary == NULL);
    fs_logBeginTransaction(context);

    snapBitsInode->dinode.attrs |= FS_IFLG_SNAP_PIN;
    fs_inodeWrite(context, snapBitsInode, INO_WRITE_LOG);

    // commit and release the transaction context

    fs_logCommitEntry(context);
    context->logSummary = NULL;

    fs_snapPrintStatus(context, LOG_NOTICE,
                       "snapnum %d pinned", snapnum);
    return NFX_OK;
}


/**
 * Un pin a snapshot that was previously pinned
 */
int32 fs_snapUnpin(
        fs_Context *context,
        int snapnum)
{
    fs_Inode *snapBitsInode;

    snapBitsInode = I_GET_EXCL(snapnum, FS_SNAPSHOT_BITS_INODE);
    if (snapBitsInode == NULL) {
        return NFX_ERR;
    }

    if ((snapBitsInode->dinode.attrs & FS_IFLG_SNAP_PIN) == 0) {
        fs_inodeReleaseExclusive(context, snapBitsInode);
        return NFX_OK;
    }

    // create a transaction context

    assert(context->logSummary == NULL);
    fs_logBeginTransaction(context);

    snapBitsInode->dinode.attrs &= ~FS_IFLG_SNAP_PIN;
    fs_inodeWrite(context, snapBitsInode, INO_WRITE_LOG);

    // commit and release the transaction context

    fs_logCommitEntry(context);
    context->logSummary = NULL;

    fs_snapPrintStatus(context, LOG_NOTICE,
                       "snapnum %d un pinned", snapnum);
    return NFX_OK;
}


/**
 * Return the time, in seconds since the start of 1970, the specified
 * snapshot was taken.
 *
 * Returns 0 if the named snapshot does not exists.
 */
uint32 fs_snapGetSnapTime(
    fs_Context *context,
    int32 id)
{
    fs_Inode *inode;
    uint32 time;

    inode = I_GET_SHARED(id, FS_SNAPSHOT_BITS_INODE);
    if (inode == NULL)
        return 0;
    time = inode->dinode.ctime.secs;
    fs_inodeRelease(context, inode);
    return time;
}


/**
 * Return the time, in seconds since the start of 1970, the specified
 * snapshot was taken.
 *
 * Returns 0 if the named snapshot does not exists.
 */
uint32 fs_snapGetTime(
    fs_Context *context,
    char *name)
{
    uint32 time;
    int id = fs_snapLookup(context, name);

    if (id < 0)
        return 0;
    time = fs_snapGetSnapTime(context, id);
    return time;
}


#ifndef FS_SNAP_OPTIMIZED
/**
 * Return the number of bits which are set in the specified snapshot
 * bits, but not live in the corresponding excerpt from the reference
 * bytes file.
 */
static uint32 fs_snapCountBitsAndBytes(
    fs_Context *context,
    uint8      *bits,
    fs_RefByte *ref)
{
    int i, n;
    uint32 count = 0;
    uint32 blockSize = context->fileSystem->params.blockSize;

    assert((blockSize % BITS_PER_BYTE) == 0);

    for (i = 0, n = blockSize / BITS_PER_BYTE; i < n; i++) {
        uint8 stripe = bits[i];

        if (stripe == 0) {
            ref += BITS_PER_BYTE;
        }
        else {
            int j;
            int bit;
            for (j = 0, bit = 1; j < BITS_PER_BYTE; j++, bit <<= 1) {
                fs_RefByte byte = *ref++;
                if (!(byte & FS_REFBYTE_IS_LIVE) && (stripe & bit) != 0)
                    count += 1;
            }
        }
    }

    return count;
}
#endif


/**
 * Return an estimate of the number of blocks the specified snapshot
 * contains that are different from the current file system.  This value
 * is intended to be an upper bound on the number of blocks that would be
 * devoted to this snapshot if it were the only snapshot in the system.
 *
 * Returns FS_BAD_COUNT if the named snapshot does not exist.
 */
uint64 fs_snapEstimateSize(
    fs_Context *context,
    int snap)
{
    fs_Inode *snapBits;
    fs_Inode *refInode;
    
    fs_Parameters *params = &context->fileSystem->params;
    uint32 blockSize = params->blockSize;
    uint32 chunkSize = blockSize / BITS_PER_BYTE;
    uint64 count = 0;
    uint64 snapBitsPages;
    uint64 page;
    FS_SNAP_WATCHDOG_DECL

    snapBits = I_GET_SHARED(snap, FS_SNAPSHOT_BITS_INODE);
    refInode = I_GET_SHARED(0, FS_REF_COUNT_INODE);

    snapBitsPages = ceilingDiv64(snapBits->dinode.size, blockSize);
    assert(snapBitsPages == snapBits->dinode.numblocks);

    FS_SNAP_WATCHDOG_INIT("estimate snap size");

    for (page = 0; page < snapBitsPages; page++) {
        int i;
        fs_Buf *snapBuf = O_GET_SHARED(snapBits, page);

        assert(snapBuf != NULL);

        // yield the CPU if necc to prevent watchdog timeouts
        FS_SNAP_WATCHDOG_CHECK("estimate snap size");

        for (i = 0; i < BITS_PER_BYTE; i++) {
            uint8 *bits = &snapBuf->u.data[i * chunkSize];
            uint64 refPage = page * BITS_PER_BYTE + i;
            fs_Buf *refBuf = O_GET_SHARED(refInode, refPage);

            if (refBuf == NULL) {
                // nearing the end of the file
                assert(page + 1 == snapBitsPages);
            }
            else {
#ifdef FS_SNAP_OPTIMIZED 
                int j, k;
                fs_RefByte *refbytes = refBuf->u.refBytes;

                for (j = 0, k = 0; j < blockSize / BITS_PER_BYTE; j += 4, k += 4*sizeof (uint64)) {
                    uint64 word, snapcount, live, live_mask;
                    uint8 snapbits;

                    live_mask = ((uint64)FS_REFBYTE_IS_LIVE
                                | ((uint64)FS_REFBYTE_IS_LIVE << 8)
                                | ((uint64)FS_REFBYTE_IS_LIVE << 16)
                                | ((uint64)FS_REFBYTE_IS_LIVE << 24)
                                | ((uint64)FS_REFBYTE_IS_LIVE << 32)
                                | ((uint64)FS_REFBYTE_IS_LIVE << 40)
                                | ((uint64)FS_REFBYTE_IS_LIVE << 48)
                                | ((uint64)FS_REFBYTE_IS_LIVE << 56));
#define COUNT_SNAPS do{\
    snapcount = ((uint64) (snapbits & 0x1))                     \
                | ((uint64) ((snapbits & 0x2) ? 1 : 0) << 8)    \
                | ((uint64) ((snapbits & 0x4) ? 1 : 0) << 16)   \
                | ((uint64) ((snapbits & 0x8) ? 1 : 0) << 24)   \
                | ((uint64) ((snapbits & 0x10) ? 1 : 0) << 32)  \
                | ((uint64) ((snapbits & 0x20) ? 1 : 0) << 40)  \
                | ((uint64) ((snapbits & 0x40) ? 1 : 0) << 48)  \
                | ((uint64) ((snapbits & 0x80) ? 1 : 0) << 56); \
    live = word & live_mask;    /* get only the live bits                   */\
    live ^= live_mask;          /* find only the live bits that are not set */\
    word = live & snapcount;    /* AND with the snapbits that are set       */\
    count += ( ((word & 1) ? 1 : 0)                 \
               + (((word >> 8) & 1) ? 1 : 0)        \
               + (((word >> 16) & 1) ? 1 : 0)       \
               + (((word >> 24) & 1) ? 1 : 0)       \
               + (((word >> 32) & 1) ? 1 : 0)       \
               + (((word >> 40) & 1) ? 1 : 0)       \
               + (((word >> 48) & 1) ? 1 : 0)       \
               + (((word >> 56) & 1) ? 1 : 0) );    \
} while (0)
                    snapbits = bits[j];
                    if (snapbits != 0) {
                        word = *(uint64 *)&refbytes[k];
                        COUNT_SNAPS;
                    }

                    snapbits = bits[j+1];
                    if (snapbits != 0) {
                        word = *(uint64 *)&refbytes[k+8];
                        COUNT_SNAPS;
                    }

                    snapbits = bits[j+2];
                    if (snapbits != 0) {
                        word = *(uint64 *)&refbytes[k+16];
                        COUNT_SNAPS;
                    }

                    snapbits = bits[j+3];
                    if (snapbits != 0) {
                        word = *(uint64 *)&refbytes[k+24];
                        COUNT_SNAPS;
                    }
                }

#else
                count += fs_snapCountBitsAndBytes(
                                context, bits, refBuf->u.refBytes);
#endif
                fs_bufRelease(context, refBuf);
            }
        }
        fs_bufRelease(context, snapBuf);
    }

    fs_inodeRelease(context, snapBits);
    fs_inodeRelease(context, refInode);

    return count;
}


/**
 * Return an estimate of the number of blocks which differ between the two
 * named snapshots.  This value is intended to be an upper bound on the
 * number of blocks that need to be copied to make one snapshot into the
 * other.  If the oldName is NULL, the number of blocks referenced by
 * snapshot newName is computed instead.
 *
 * Returns FS_BAD_COUNT if either named snapshot does not exist or if
 * the snapshot named oldname was created more recently than the one
 * specified by newname.
 */
uint64 fs_snapEstimateDeltaSize(
    fs_Context *context,
    char *oldName,
    char *newName)
{
    fs_Inode *oldBits;
    fs_Inode *newBits;
    
    fs_Parameters *params = &context->fileSystem->params;
    uint32 blockSize = params->blockSize;
    uint32 wordsPerBlock = blockSize / sizeof (uint64);
    uint64 snapBitsPages;
    uint64 page;
    uint64 count = 0;

    int newSnap = fs_snapLookup(context, newName);
    if (newSnap < 0)
        return FS_BAD_COUNT;
    newBits = I_GET_SHARED(newSnap, FS_SNAPSHOT_BITS_INODE);
    assert(newBits != NULL);
    snapBitsPages = ceilingDiv64(newBits->dinode.size, blockSize);
    assert(snapBitsPages == newBits->dinode.numblocks);

    if (oldName == NULL) {
        oldBits = NULL;
    }
    else {
        int oldSnap = fs_snapLookup(context, oldName);
        if (oldSnap < 0)
            return FS_BAD_COUNT;
        oldBits = I_GET_SHARED(oldSnap, FS_SNAPSHOT_BITS_INODE);
        assert(oldBits != NULL);
    }

    for (page = 0; page < snapBitsPages; page++) {
        fs_Buf *newBuf = O_GET_SHARED(newBits, page);
        uint64 *newData;

        assert(newBuf != NULL);
        newData = newBuf->u.rawdata;

        if (oldBits == NULL) {
            int i;
            for (i = 0; i < wordsPerBlock; i++)
                count += fs_popCount64(newData[i]);
        }
        else {
            fs_Buf *oldBuf = O_GET_SHARED(oldBits, page);
            uint64 *oldData;
            int i;
            assert(oldBuf != NULL);
            oldData = oldBuf->u.rawdata;
            for (i = 0; i < wordsPerBlock; i++)
                count += fs_popCount64(newData[i] &~ oldData[i]);
        }
    }

    fs_inodeRelease(context, oldBits);
    fs_inodeRelease(context, newBits);

    return count;
}


/**
 * Given a snapshot name find the snapshot number
 */
int fs_getSnapNum(
    fs_Context          *context,
    char                *name)
{
    return fs_snapLookup(context, name);
}


/**
 * Terminate any in progress snapshot operation
 */
boolean fs_snapTerminateActive(
    fs_Context  *context)
{
    fs_FileSystem *fs;
    boolean active = FALSE;

    fs = context->fileSystem;

    NFX_SPIN_LOCK(&fs->superLock);
    if (fs->super->snapState.phase != FS_SNAP_PHASE_READY)
        active = TRUE;
    NFX_SPIN_UNLOCK(&fs->superLock);

    if (active) {
        NFX_SPIN_LOCK(&fs->spin);
        fs->snapInfo.terminate = TRUE;
        NFX_SPIN_UNLOCK(&fs->spin);
    }

    return active;
}


#ifdef FS_SNAP_READAHEAD
/**
 * prefetch refbyte and snapbits (if snap) pages into the buffer cache
 */
void fs_snapDoReadAhead(
    fs_Context  *context,
    int32       snap)
{
    fs_Parameters *params = &context->fileSystem->params;
    fs_Inode *refInode;
    fs_Inode *snapBitsInode;
    uint64 page, refPages, snapBitsPages;
    
    refInode = I_GET_SHARED(0, FS_REF_COUNT_INODE);
    refPages = ceilingDiv64(params->totalBlocks, params->blockSize);
    for (page = 0; page < refPages; page++) {
        fs_Buf *refBuf = O_GET_SHARED(refInode, page);
        fs_bufRelease(context, refBuf);
    }
    fs_inodeRelease(context, refInode);

    if (snap != 0) {

        // prefetch by reading sync into the buffer cache
        // even though the buffers are released, for a 100k (8k non-dir)
        // cache memory footprint, there is only 1% chance these
        // buffers will be recycled before use

        snapBitsInode = I_GET_SHARED(snap, FS_SNAPSHOT_BITS_INODE);
        snapBitsPages = ceilingDiv64(params->totalBlocks, params->blockSize * BITS_PER_BYTE);

        for (page = 0; page < snapBitsPages; page++) {
            fs_Buf *snapBuf = O_GET_SHARED(snapBitsInode, page);
            fs_bufRelease(context, snapBuf);
        }
        fs_inodeRelease(context, snapBitsInode);
    }
}
#endif


//   M A N A G E M E N T


/**
 * Set default snap state
 */
void fs_snapSetDefaultState(
      fs_SnapState  *snapState)
{
    fs_SnapTimes    *times = &snapState->times;

    times->enable = TRUE;
    times->maxWeekly = FS_SNAP_DEFAULT_MAX_WEEKLY;
    times->maxDaily = FS_SNAP_DEFAULT_MAX_DAILY;
    times->maxHourly = FS_SNAP_DEFAULT_MAX_HOURLY;
    times->hours = FS_SNAP_DEFAULT_HOURS;

    snapState->version = FS_SNAP_VERSION_ONE;
    snapState->id = 0;
    snapState->phase = FS_SNAP_PHASE_READY;
}


#if !defined(FS_STANDALONE)
/**
 * Timer function to signal snapshot thread to wake up
 */
static int32
fs_snapshotTimeout(
    uint32      timer,
    void        *param1,
    void        *param2,
    void        *param3,
    void        *param4)
{
    int32 rc;
    esm_event_t event;
    fs_FileSystem *fileSystem;
    evm_volId_t volId, *vp = (evm_volId_t *) param1;

    volId.whole_thing = vp->whole_thing;

    fileSystem = fs_fileSystemForVolume(volId);
    if (fileSystem == NULL) {
        return NFX_ERR;
    }

    event.event = FS_THREAD_EVENT_START;
    event.p1.u.w = 0x1234;
    event.p2.u.w = 0x5678;

    // send event if file system online and internal snaps enabled

    if (fileSystem && (fileSystem->state == FS_VOL_STATE_ONLINE) &&
        fileSystem->snapInfo.maxInternalSnaps) {

        boolean sendEvent = TRUE;

        NFX_SPIN_LOCK(&fileSystem->snapInfo.spin);
        if (fileSystem->snapInfo.active == TRUE) {
            sendEvent = FALSE;
        }
        NFX_SPIN_UNLOCK(&fileSystem->snapInfo.spin);

        if (sendEvent == FALSE) {
            return NFX_OK;
        }

        rc = thread_event(fileSystem->snapInfo.snapThreadHandle, &event);
        if (rc == NFX_OK) {
           // printf("snap internal timeout: event sent to snap thread for volId %lld\n",
           //     volId.whole_thing);
        } else {
            fs_error(NULL, FS_EVENT_DELIVERY_FAIL, "snap timer event");
        }
    }
    return NFX_OK;
}
#endif


/**
 * Set some parameters for taking internal snapshots for
 * the specified context's file system:
 *  1. nsnaps - keep this many internal snapshots (default is 2)
 *  2. minutes - interval to take the internal snapshot (default is 10 minutes)
 *
 * Returns NFX_OK or NFX_ERR
 */
int32 fs_snapSetInternal(
    fs_Context *context,
    int32 max_snapshots,
    int32 minutes,
    boolean mkfsTime)
{
    fs_FileSystem *fileSystem = context->fileSystem;
    int32 rc = NFX_ERR, oldTimeout;

    if (max_snapshots > FS_SNAP_MAX_INTERNAL) {
        FS_TRACE(context, FS_TRACE_SNAP,
                ("snap set internal: max snaps =%d", FS_SNAP_MAX_INTERNAL));
        return rc;
    }

    if (mkfsTime)
        oldTimeout = 0;
    else
        oldTimeout = fileSystem->snapInfo.timeout;

    FS_TRACE(context, FS_TRACE_SNAP,
            ("snap set internal: max snaps =%d minutes =%d (oldTimeout =%d tics)",
            max_snapshots, minutes, oldTimeout));

    // adjust to timer tics (= 1000 ms)
    fileSystem->snapInfo.timeout = minutes * 60 * 1000;
    fileSystem->snapInfo.maxInternalSnaps = max_snapshots;

    if (mkfsTime) {
        int loop;
        
        for (loop = 0 ; loop < FS_MAX_SNAPSHOTS; loop++)
        {
            fs_lockInit(&fileSystem->snapInfo.accessLocks[loop], FS_LOCK_TYPE_SNAP);
        }
        
        NFX_SPIN_INIT(&fileSystem->snapInfo.spin,SPINLOCK_LEVEL_UNKNOWN);
        fileSystem->snapInfo.active = 0;
        fileSystem->snapInfo.timer = 0;
        fileSystem->snapInfo.snapThreadHandle = 0;
        fileSystem->snapInfo.lastSnapCreate = *(eee_rtcZero());
        fs_lockInit(&fileSystem->snapInfo.lock, FS_LOCK_TYPE_SNAP);
    }

    if ((max_snapshots == 0) || (minutes == 0)) {
        return NFX_OK;
    }

#if !defined(FS_STANDALONE)

    // create eee timer if necc

    if (fileSystem->snapInfo.timer == 0) {

        fileSystem->snapInfo.timer = eee_timerCreate(fileSystem->snapInfo.timeout,
                                                    EEE_TIMER_REPEAT);
        if (fileSystem->snapInfo.timer == TIMER_ERR_NO_TIMER_AVAILABLE)
            fs_error(NULL, FS_RAM_ALLOC_ERR, "create snap timer");
    }

    // check if we need to stop a previous timer

    if (!mkfsTime && fileSystem->snapInfo.timer &&
        (fileSystem->snapInfo.timeout != oldTimeout)) {

        rc = eee_timerStop(fileSystem->snapInfo.timer);
    }

    // restart if different timeout value

    if (fileSystem->snapInfo.timeout != oldTimeout) {

        rc = eee_timerStart(fileSystem->snapInfo.timer, fs_snapshotTimeout,
                           (void *) &fileSystem->volId, NULL, NULL, NULL);
        if (rc != NFX_OK) {
            fs_snapPrintStatus(context, LOG_ERR,
                    "snap set internal: cannot restart timer\n");
            return NFX_ERR;
        }
    }

    if (rc == NFX_OK) {
        FS_TRACE(context, FS_TRACE_SNAP,
                ("snap set internal: new timeout tics =%d",
                fileSystem->snapInfo.timeout));
    }
#endif

    return NFX_OK;
}


/**
 * Snapshot admin thread to manage internal snapshots
 */
void fs_snapshotThread(
    esm_handle_t    handle,
    void            *arg)
{
    int32 snapnum = 0;
    esm_event_t wakeEvent;
    fs_Context *context = arg;
    fs_FileSystem *fs = context->fileSystem;

    while (1) {

        // sleep until internal snapshot timeout

        // printf("snapshot thread: sleeping\n");
        fs_yield(context, &wakeEvent, FS_YIELD_TIME_FOREVER, FS_YIELD_TYPE_ADMIN);
        // printf("snapshot thread: waking up\n");
        if (wakeEvent.event != FS_THREAD_EVENT_START) {
            fs_error(NULL, FS_INTERNAL_ERR, "snapshot thread: wrong event");
            // printf("snapshot thread: wrong event %d", wakeEvent.event);
        }
        if ((wakeEvent.p1.u.w != 0x1234) || (wakeEvent.p2.u.w != 0x5678)) {
            fs_error(NULL, FS_INTERNAL_ERR, "snapshot thread: wrong event info");
            // printf("snapshot thread: wrong debug event info %d %d\n",
            //      wakeEvent.p1.u.w, wakeEvent.p2.u.w);
        }

        NFX_SPIN_LOCK(&fs->snapInfo.spin);
        fs->snapInfo.active = TRUE;
        NFX_SPIN_UNLOCK(&fs->snapInfo.spin);

        // create internal snapshot

        fs_snapCreateInternal(context, &snapnum, TRUE);

        NFX_SPIN_LOCK(&fs->snapInfo.spin);
        fs->snapInfo.active = FALSE;
        NFX_SPIN_UNLOCK(&fs->snapInfo.spin);
    }
}


/**
 * Called at mount time to initialize snapshots:
 * 1. setup default internal snapshot parameters
 * 2. start eee timer for internal snapshot management
 */
void fs_snapInit(
    fs_Context  *context)
{
    fs_FileSystem *fileSystem = context->fileSystem;

    fs_snapSetInternal(context, FS_SNAP_DEFAULT_INTERNAL_SNAPS,
                       FS_SNAP_INTERNAL_TIMEOUT, TRUE);
    fileSystem->snapInfo.terminate = FALSE;

#if !defined(FS_STANDALONE)

    // create the internal snapshot admin thread

    fileSystem->snapInfo.snapThreadHandle =
    fs_threadFork(context, fs_snapshotThread, TRUE, FALSE,
                  FS_DAEMON_TYPE_SNAP, NULL);
    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap init: enable internal snapshots max %d timeout %d (tics)",
             context->fileSystem->snapInfo.maxInternalSnaps,
             context->fileSystem->snapInfo.timeout));

#endif

}


/**
 * Allocate a free internal snapshot id from the snapshots bitmap
 * in the superblock.
 * Modifies the in-core superblock, but doesn't write it to disk.
 * Returns a negative number if no snapshot id is available.
 */
static int fs_snapAllocateInternalId(
    fs_Context *context)
{
    fs_SuperBlock *super = context->fileSystem->super;
    uint64 freebits = ~super->snapshots & FS_SNAP_INTERNAL_ID_MASK;
    int id;

    if (freebits == 0) {
        FS_TRACE(context, FS_TRACE_SNAP,
                 ("snap allocate id: max internal snapshots reached"));
        return -1;
    } else {
        uint64 bit = leastBit64(freebits);
        id = log2_64(bit);
        FS_TRACE(context, FS_TRACE_SNAP,
                ("snap allocate internal id: id %d snapshots %llx freebits %llx bit %llx",
                id, super->snapshots, freebits, bit));
        NFX_SPIN_LOCK(&context->fileSystem->superLock);
        super->snapshots |= bit;
        NFX_SPIN_UNLOCK(&context->fileSystem->superLock);
        return id;
    }
}


/**
 * Remove an internal snapshot in the specified context's file system.
 */
int32 fs_snapRemoveInternal(
    fs_Context *context,
    int32 snapnum)
{
    int32 rc;

    // deallocate snapnum and write back superblock

    fs_snapDeallocateId(context, snapnum);
    fs_writeSuperblocksSync(context, context->fileSystem->super);

    // update the superblock with current phase

    fs_snapSetState(context, snapnum, FS_SNAP_PHASE_REMOVE_REF_COUNTS);

    rc = fs_snapDoIt(context, snapnum, FS_SNAP_PHASE_REMOVE_REF_COUNTS);

    return rc;
}


/**
 * qsort routine to sort list of internal snapshots
 */
static int fs_snapInternalCompare(
    const void *s0,
    const void *s1)
{
    fs_SnapInternalInfo *snap0 = (fs_SnapInternalInfo *) s0;
    fs_SnapInternalInfo *snap1 = (fs_SnapInternalInfo *) s1;

    if (snap0->ctime.secs > snap1->ctime.secs) {
        return 1;
    }
    return -1;
}


/**
 * Get snapshot info: name, snapnum, ctime
 * usage must be big enough to hold FS_MAX_SNAPSHOTS
 * n modified to return number of snapshots found
 *
 * If internal is TRUE, get info for internal snapshots
 * Else return info for user created snapshots
 */ 
int32 fs_snapGetInternal(
    fs_Context *context,
    int32 *nSnaps,
    fs_SnapInternalInfo *snaps)
{
    uint64 bits = context->fileSystem->super->snapshots & FS_SNAP_INTERNAL_ID_MASK;
    uint64 mask;
    int32 snapnum, scount = 0;

    for (snapnum = FS_MAX_USER_SNAPSHOTS+1; snapnum < FS_MAX_SNAPSHOTS; snapnum++) {
        mask = 1ULL << snapnum;
        if (mask & bits) {
            snaps[scount].snapnum = snapnum;
            snaps[scount].ctime.secs = fs_snapGetSnapTime(context, snapnum);
            scount++;
        }
    }

    // sort with oldest to newest

    qsort(snaps, scount, sizeof(fs_SnapInternalInfo), fs_snapInternalCompare);

    FS_TRACE(context, FS_TRACE_SNAP,
            ("snap get internal: found %d oldest snapnum %d ctime %d",
            scount, snaps[0].snapnum, snaps[0].ctime.secs));

    *nSnaps = scount;
    return NFX_OK;
}


/**
 * Create an internal snapshot in the specified context's file system.
 * The oldest internal snapshot will be removed before creating this new snapshot.
 *
 * This internal snapshot is just like a user created one except
 * there is no entry for it in the .snapshot directory.
 *
 * Returns NFX_OK or NFX_ERR
 * If NFX_OK, created snapnum returned (if not NULL)
 */

int32 fs_snapCreateInternal(
    fs_Context *context,
    int32 *snapnum,
    boolean quiesce)
{
    int32 rc = NFX_OK, id, n = 0;
    eeeRTC_t delta;
    fs_SnapInfo *snapInfo = &context->fileSystem->snapInfo;
    fs_SnapInternalInfo snaps[FS_SNAP_MAX_INTERNAL];
	static boolean first = TRUE;
	static eeeRTC_t minInterval;

    // check if internal snapshot is disabled

    if (snapInfo->maxInternalSnaps == 0) {
        fs_snapPrintStatus(context, LOG_INFO,
                           "snap create internal: disabled");
        return NFX_OK;
    }

    // check last internal snapshot create time
    // don't bother if one was created within the interval

    eee_rtcDuration(&snapInfo->lastSnapCreate, &delta);
	if (first) {
		first = FALSE;
		eee_msToRTC((uint32)FS_SNAP_INTERNAL_MIN, &minInterval);
	}
	if (eee_rtcCompare(&delta, &minInterval) < 0) {
		eeeGMT_t crTime;

		eee_rtcToGMT(&snapInfo->lastSnapCreate, &crTime);
        fs_snapPrintStatus(context, LOG_NOTICE,
                "snap create internal: last already created at %s",
                eee_humandate(crTime.sec));
        return NFX_OK;
    }

    FS_LOCK_ORDER_DISABLE(context);
    
#ifdef FS_SNAP_EEK
    printf("\nEEK before snap internal create:\n");
    fs_eek(context);
#endif

    // only one snapshot operation is allowed
    // because of the snap state info in the superblock
    // since we do the FS_SNAP_PHASE_CREATE_ALLOC outside
    // the freeze, guarantee only one snapshot operation
    // by acquiring the snap lock

    fs_lockExclusive(context, &snapInfo->lock, NULL);

    id = fs_snapAllocateInternalId(context);
    if (id < 0) {

        // try to remove one if we've maxed out

        if (quiesce)
            fs_freezeAll(context);
        rc = fs_snapGetInternal(context, &n, &snaps[0]);
        if (rc == NFX_ERR)
            goto fs_snapCreateInternal_return;

        if (n >= snapInfo->maxInternalSnaps) {
            fs_snapPrintStatus(context, LOG_INFO,
                    "snap create internal: max reached, remove old snap %d begin",
                    snaps[0].snapnum);

            rc = fs_snapRemoveInternal(context, snaps[0].snapnum);

            fs_snapPrintStatus(context, LOG_INFO,
                    "snap create internal: remove old snap %d end",
                    snaps[0].snapnum);
        }
        id = fs_snapAllocateInternalId(context);
        if (id < 0) {
            rc = FS_ERR_SNAP_MAX;
            goto fs_snapCreateInternal_return;
        }
        if (quiesce)
            fs_thawAll(context);
    }

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap internal create: begin for %d", id));

    // update the superblock with the current phase

    fs_snapSetState(context, id, FS_SNAP_PHASE_CREATE_ALLOC);

    // create snapbits and prefetch before freezing the file system
    // and continue with the FS_SNAP_PHASE_CREATE_SNAP_BITS phase

    fs_snapCreateAllocSnapBits(context, id);

#ifdef FS_SNAP_READAHEAD
    //  prefetch refbyte pages before freeze
    fs_snapDoReadAhead(context, id);
#endif

    if (quiesce) {
        fs_freezeAll(context);
#ifdef FS_SNAP_EEK
        fs_syncThread(context);
        fs_syncThread(context);
#endif
    }

    fs_snapPrintStatus(context, LOG_INFO, "snap create internal %d: begin", id);

    // finish the snapshot creation process

    rc = fs_snapDoIt(context, id, FS_SNAP_PHASE_CREATE_SNAP_BITS);

    if (rc == NFX_OK) {
        if (snapnum) *snapnum = id;
        fs_snapStats.snapCreates++;
        eee_rtc(&snapInfo->lastSnapCreate);
    }

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap internal create: done (time %lld inodesUsed %lld)",
             eee_rtcToSec(&snapInfo->lastSnapCreate),
             context->fileSystem->super->inodesUsed));

    // Calculate the min free blocks for current conditions

    fs_recalcMinFreeBlks(context);

    // get number and list of internal snapshots
    // and check if oldest needs to be removed

    rc = fs_snapGetInternal(context, &n, &snaps[0]);
    if (rc == NFX_ERR)
        goto fs_snapCreateInternal_return;

    if (n > snapInfo->maxInternalSnaps) {
        FS_TRACE(context, FS_TRACE_SNAP,
                ("snap internal create: found %d with max %d -> removing old %d",
                n, snapInfo->maxInternalSnaps, snaps[0].snapnum));
        fs_snapPrintStatus(context, LOG_INFO,
                "snap create internal: remove old snap %d begin",
                snaps[0].snapnum);

        rc = fs_snapRemoveInternal(context, snaps[0].snapnum);

        fs_snapPrintStatus(context, LOG_INFO,
                "snap create internal: remove old snap %d end",
                snaps[0].snapnum);
        FS_TRACE(context, FS_TRACE_SNAP,
                ("snap internal create: old snap %d removed", snaps[0].snapnum));
    }

fs_snapCreateInternal_return:

    if (quiesce) {
#ifdef FS_SNAP_EEK
        printf("\nEEK after snap internal create:\n");
        fs_eek(context);
#endif
        fs_thawAll(context);
    }
    fs_unlockExclusive(context, &snapInfo->lock);

    fs_snapPrintStatus(context, LOG_INFO, "snap create internal %d: end", id);

    return rc;
}


/**
 * Set a snapshot schedule
 */
int32 fs_snapSetSched(
      fs_Context *context,
      fs_SnapSched *sched,
      boolean write)
{
    fs_SnapTimes    *times = &context->fileSystem->super->snapState.times;
    uint32          *hour = &sched->hours[0];

    // verify filesystem state and snap version

    if (context->fileSystem->state != FS_VOL_STATE_ONLINE) {
        return NFX_ERR;
    }

    if (FS_SNAP_CHECK_VERSION(context) == FALSE) {
        return NFX_ERR;
    }

    if (times->enable == FALSE) {
        return FS_ERR_SNAP_NOT_ENABLED;
    }

    // modify only if specified

    if (sched->maxWeeklySet)
        times->maxWeekly = sched->maxWeekly;
    if (sched->maxDailySet)
        times->maxDaily = sched->maxDaily;
    if (sched->maxHourlySet) {
        times->maxHourly = sched->maxHourly;
        // default unless hoursSet
        times->hours = FS_SNAP_ALL_HOURS;
    }

    // disable if everything is 0

    if (!times->maxWeekly && !times->maxDaily && !times->maxHourly) {
        times->enable = FALSE;
        fs_snapPrintStatus(context, LOG_NOTICE,
                           "%s snapshot scheduling disabled",
                           context->fileSystem->super->mountName);
    }
    
    if (sched->maxHourlySet && sched->hoursSet) {
        times->hours = 0;   // reset first
        while (*hour) {
            uint32 shift = *hour;
            uint32 bit;

            assert((shift > 0) && (shift <= FS_SNAP_SCHED_MAX_HOURS));
            bit = 1 << shift;
            times->hours |= bit;
            hour++;
        } // while
    }

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap sched: vol %s %s %d %d %d %x",
             context->fileSystem->super->mountName,
             times->enable ? "enable" : "disable",
             times->maxWeekly, times->maxDaily, times->maxHourly, times->hours));

    // write the superblock(s)

    if (write)
        fs_writeSuperblocksSync(context, context->fileSystem->super);

    return NFX_OK;
}

/**
 * Get snapshot schedule for volume
 */
int32 fs_snapGetSched(
    fs_Context      *context,
    fs_SnapSched    *sched)
{
    fs_SnapTimes    *times = &context->fileSystem->super->snapState.times;
    uint32          hour, i = 0;

    // verify filesystem state and snap version

    if (context->fileSystem->state != FS_VOL_STATE_ONLINE) {
        return NFX_ERR;
    }

    if (FS_SNAP_CHECK_VERSION(context) == FALSE) {
        return NFX_ERR;
    }

    sched->enable = times->enable;
    sched->maxWeekly = times->maxWeekly;
    sched->maxDaily = times->maxDaily;
    sched->maxHourly = times->maxHourly;
    bzero(sched->hours, sizeof(sched->hours));
    if (times->hours == 0) {
        return NFX_OK;
    }
    for (hour = 1; hour <= FS_SNAP_SCHED_MAX_HOURS; hour++) {
        uint32 bit;

        bit = (times->hours >> hour) & 0x1;
        if (bit) {
            sched->hours[i] = hour;
            i++;
        }
    }

    return NFX_OK;
}

/**
 * Get snapshot info: name, snapnum, ctime
 * If internal, get info for internal snapshots
 * Else return info for user created snapshots
 */
int32 fs_snapGetInfo(
    fs_Context      *context,
    fs_SnapUsage    *usage,
    int32           *nUsage,
    boolean         internal)
{
    int32 i, found;
    char *buf;
    char *names[FS_MAX_USER_SNAPSHOTS];
    fs_SnapInternalInfo snaps[FS_SNAP_MAX_INTERNAL];

    // verify filesystem state and snap version

    if (context->fileSystem->state != FS_VOL_STATE_ONLINE) {
        return NFX_ERR;
    }

    if (FS_SNAP_CHECK_VERSION(context) == FALSE) {
        return NFX_ERR;
    }
    
    if (!internal) {
        buf = fs_ramAlloc(FS_RAM_RESP_MSG, FS_MAX_SNAPSHOTS * FS_NAME_MAX);
        *nUsage = fs_snapGetNames(context, names, buf, FS_MAX_SNAPSHOTS * FS_NAME_MAX);
        if (*nUsage == 0) {
            return NFX_OK;
            fs_ramDealloc(FS_RAM_RESP_MSG, buf);
        }
        //printf("snap get info: nSnaps %d\n", *nUsage);
        found = *nUsage;
        for (i = 0; i < *nUsage; i++) {
            int32 snapnum;
            snapnum = fs_snapLookup(context, names[i]);
            if (snapnum < 0) {
                found--;
                continue;
            }
            strcpy(usage->name, names[i]);
            usage->ctime.secs = fs_snapGetSnapTime(context, snapnum);
            usage->size = fs_snapEstimateSize(context, snapnum);
            usage->pin = fs_snapIsPin(context, names[i]);
            //printf("  *** %s size %lld\n", usage->name, usage->size);
            usage++;
        }
        *nUsage = found;
        fs_ramDealloc(FS_RAM_RESP_MSG, buf);
    } else {
        int32 rc = fs_snapGetInternal(context, nUsage, &snaps[0]);
        if ((rc != NFX_OK) || (*nUsage == 0)) {
            return NFX_OK;
        }
        for (i = 0; i < *nUsage; i++) {
            // return external ids
            usage->snapnum = snaps[i].snapnum - FS_MAX_USER_SNAPSHOTS;
            usage->ctime.secs = snaps[i].ctime.secs;
            usage++;
        }
    }

    return NFX_OK;
}


/**
 * Rename snapshot names starting with prefix
 *
 * eg
 *      hourly.0 -> hourly.1
 *      hourly.1 -> hourly.2
 *      hourly.2 -> hourly.3
 *
 * Also remove oldest if max snapshots found
 */
static void fs_snapRenameSnaps(
        fs_Context *context,
        char *prefix, int32 max)
{
    int32 prefix_len = strlen(prefix);
    uint64 cookie = 0;
    int32 eof, entries = 0;
    int32 i, found = 0, max_found = 0;

    char *name = NULL;
    char *namebuf = NULL;
    fs_DirEntry *dirent;

    typedef struct {
        char name[FS_NAME_MAX];
        uint64 inum;
    } fs_SnapEntry;
    fs_SnapEntry *names = NULL;
    fs_SnapEntry *remove = NULL;

    fs_Inode *dir = I_GET_EXCL(0, FS_SNAPSHOT_DIR_INODE);
    assert(dir);
    namebuf = (char *) fs_ramAlloc(FS_RAM_SNAP_NAME, FS_READDIR_SPACE_MAX);
    dirent = (void *) namebuf;
    fs_dir_readdir(context, dir, dirent, FS_READDIR_SPACE_MAX, cookie, FALSE,
                   &entries, &eof);

    // one readdir should be enough
    assert(eof);
    if (!entries) {
        fs_inodeReleaseExclusive(context, dir);
        goto fs_snapRenameSnapsReturn;
    }

    // need to alloc everything otherwise we get the red zone

    names = (fs_SnapEntry *) fs_ramAlloc(FS_RAM_SNAP_NAME,
                                        entries * sizeof(fs_SnapEntry));
    bzero(names, entries * sizeof(fs_SnapEntry));
    name = (char *) fs_ramAlloc(FS_RAM_SNAP_NAME, FS_NAME_MAX);
    remove = (fs_SnapEntry *) fs_ramAlloc(FS_RAM_SNAP_NAME, sizeof(fs_SnapEntry));

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap sched rename: vol %llx %s max %d entries %d",
             context->volId.whole_thing, prefix, max, entries-2));

    for (i = 0; i < entries; i++) {
        int32 index;

        // check if name contains prefix

        if (strncmp(prefix, dirent->name, prefix_len) == 0) {
            index = atoi((char *) &dirent->name[prefix_len]);

            // sanity check the index found
            // user could manually create names using the same format !!

            if (index >= entries) {
                continue;
            }
            assert(names[index].inum == 0);

            // save dir entry for later rename
            // also save the max in case we need to remove it later

            strcpy(names[index].name, dirent->name);
            names[index].inum = dirent->inode;
            if (index >= max_found) {
                *remove = names[index];
                max_found = index;
            }
            found++;
        }

        // get the next dir entry

        cookie = dirent->cookie;
        FS_READDIR_ADV_ENTRY_PTR(dirent);
    }

    // remove if we've maxed out

    if (found == max) {
    
        FS_TRACE(context, FS_TRACE_SNAP,
                ("snap sched rename: max reached removing snap %s", remove->name));

        // make sure there is something to remove

        assert(remove->inum == names[max_found].inum); 
        assert(strlen(remove->name));

        // release dir before calling snap remove and require for later

        fs_inodeReleaseExclusive(context, dir);
        fs_snapRemove(context, remove->name);
        dir = I_GET_EXCL(0, FS_SNAPSHOT_DIR_INODE);
        assert(dir);

        // clear inum so we won't try to remove it again later
        names[max_found].inum = 0;
    }

    // log necc for fs_dir operations

    fs_logBeginTransaction(context);

    // start renaming from low to high
    // rename by removing old and add new with same inum but different name

    for (i = 0; i <= max_found; i++) {

        // check for holes
        // automatically generated snapshots can be manually removed

        if (!names[i].inum) {
            break;
        }
        if (names[i+1].inum) {
            fs_dir_removeent(context, dir, names[i+1].name, strlen(names[i+1].name));
        }
        sprintf(name, "%s%d", prefix, i+1);
        fs_dir_addent(context, dir, name, strlen(name), names[i].inum);

        FS_TRACE(context, FS_TRACE_SNAP,
                 ("snap sched rename: from %s to %s", names[i].name, name));
    }

    // don't forget to remove the first ....

    if (names[0].inum) {
        fs_dir_removeent(context, dir, names[0].name, strlen(names[0].name));
    }

    // reset times for the snapdir

    dir->dinode.mtime = fs_getTimeOfDay();
    dir->dinode.atime = dir->dinode.mtime;
    fs_inodeWrite(context, dir, INO_WRITE_LOG);

    fs_logCommitEntry(context);
    context->logSummary = NULL;

fs_snapRenameSnapsReturn:

    if (name) fs_ramDealloc(FS_RAM_SNAP_NAME, name);
    if (names) fs_ramDealloc(FS_RAM_SNAP_NAME, names);
    if (namebuf) fs_ramDealloc(FS_RAM_SNAP_NAME, namebuf);
    if (remove) fs_ramDealloc(FS_RAM_SNAP_NAME, remove);

    return;
}


/**
 * Snapshot management for a volume
 */
static void fs_snapSchedVol(
    fs_Context *context,
    fs_SnapTime *time)
{
    fs_SuperBlock *super = context->fileSystem->super;
    fs_SnapTimes *times = &super->snapState.times;
    char *name;

    // verify filesystem state and snap version

    if (context->fileSystem->state != FS_VOL_STATE_ONLINE) {
        return;
    }

    if (FS_SNAP_CHECK_VERSION(context) == FALSE) {
        return;
    }

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap sched: vol %llx %s time hour %d wday %d",
             context->volId.whole_thing,
             (times->enable ? "enable" : "disable"),
             time->hour, time->wday));

    // read the superblock and check if we need to schedule any snapshot operations

    if (times->enable == FALSE) {
        fs_snapPrintStatus(context, LOG_NOTICE,
                           "snapshot scheduling: not enabled");
        return;
    }

    name = (char *) fs_ramAlloc(FS_RAM_SNAP_NAME, FS_NAME_MAX);

    if (times->maxWeekly && FS_SNAP_IS_WEEKLY(time)) {
        fs_snapRenameSnaps(context, FS_SNAP_WEEKLY_NAME, times->maxWeekly);
        sprintf(name, "%s0", FS_SNAP_WEEKLY_NAME);
        fs_snapCreate(context, name);
        fs_ramDealloc(FS_RAM_SNAP_NAME, name);
        return;
    }

    // check daily next (if midnight, this will override hourly)

    if (times->maxDaily && FS_SNAP_IS_DAILY(time)) {
        fs_snapRenameSnaps(context, FS_SNAP_DAILY_NAME, times->maxDaily); 
        sprintf(name, "%s0", FS_SNAP_DAILY_NAME);
        fs_snapCreate(context, name);
        fs_ramDealloc(FS_RAM_SNAP_NAME, name);
        return;
    }

    // check hour

    if (times->maxHourly && FS_SNAP_IS_HOURLY(time->hour, times->hours, time->min)) {
        fs_snapRenameSnaps(context, FS_SNAP_HOURLY_NAME, times->maxHourly); 
        sprintf(name, "%s0", FS_SNAP_HOURLY_NAME);
        fs_snapCreate(context, name);
        fs_ramDealloc(FS_RAM_SNAP_NAME, name);
        return;
    }

    fs_ramDealloc(FS_RAM_SNAP_NAME, name);
    return;
}


/**
 * Snapshot management for all volumes
 */
#define FS_MAX_IDS  1000        // should be big enough for version one

void fs_snapSched(
    fs_Context *context,
    fs_SnapTime *time)
{
    int32       i, nVols;
    evm_volId_t vols[FS_MAX_IDS];

    nVols = fs_fileSystemGetVols(context, vols, FS_MAX_IDS);
    for (i = 0; i < nVols; i++) {

        // setup context to handle this vol

        context->volId = vols[i];
        context->fileSystem = fs_fileSystemForVolume(vols[i]);
        if (context->fileSystem == NULL)
            continue;
        if (context->fileSystem->state != FS_VOL_STATE_ONLINE)
            continue;
        fs_snapSchedVol(context, time);
    }
}


/**
 * Find the newest snapshot (user or internal)
 * Return snapnum (1 to 64)
 * -1 if no snapshots found
 */
int32 fs_snapGetNewest(
    fs_Context *context)
{
    int32 rc;
    int32 n = 0;
    int32 newest_snapnum = -1, snapnum, newest_ctime = 0, ctime;
    fs_SnapInternalInfo snaps[FS_SNAP_MAX_INTERNAL];
    int32 i, eof, entries;
    uint64 cookie = 0;
    char namebuf[FS_READDIR_SPACE_MAX];
    fs_DirEntry *dirent;
    fs_Inode *dir;

    // get list of internal snapshots to find newest

    rc = fs_snapGetInternal(context, &n, &snaps[0]);

    // oldest first so get the last one in the list

    if ((rc == NFX_OK) && n) {
        assert(n < FS_SNAP_MAX_INTERNAL);
        newest_snapnum = snaps[n-1].snapnum;
        newest_ctime = snaps[n-1].ctime.secs;
        FS_TRACE(context, FS_TRACE_SNAP,
                ("snap get newest: found newest internal id %d time %d",
                newest_snapnum, newest_ctime));
    }
    
    // readdir .snapshot to get newest user snapshot

    dir = I_GET_SHARED(0, FS_SNAPSHOT_DIR_INODE);
    assert(dir);
    dirent = (void *) namebuf;
    fs_dir_readdir(context, dir, dirent, sizeof(namebuf), cookie, FALSE,
                   &entries, &eof);
    for (i = 0; i < entries; i++) {
        uint64 inum;

        if (!fs_nameIsDotOrDotDot(dirent->name)) {
            inum = dirent->inode;
            assert((inum >= FS_SNAPSHOT_ID_INODE_FIRST) && (inum <= FS_SNAPSHOT_ID_INODE_LAST));
            snapnum = inum - FS_SNAPSHOT_ID_INODE_FIRST;
            assert((snapnum > 0) && (snapnum < FS_MAX_SNAPSHOTS));

            ctime = fs_snapGetSnapTime(context, snapnum);
            if (ctime > newest_ctime) {
                newest_snapnum = snapnum;
                newest_ctime = ctime;
                FS_TRACE(context, FS_TRACE_SNAP,
                        ("snap get newest: found newest snapname %s id %d time %d",
                        dirent->name, newest_snapnum, newest_ctime));
            }
        }

        // get the next dir entry

        cookie = dirent->cookie;
        FS_READDIR_ADV_ENTRY_PTR(dirent);
    }

    fs_inodeRelease(context, dir);
    return newest_snapnum;
}


//    D I A G N O S T I C S


/**
 * Dump out the blocks in the snapbits file that are set (allocated)
 */
void fs_snapDiagBitsFile(
        fs_Context *context,
        char *name,
        int snapnum)
{
    fs_Parameters *params = &context->fileSystem->params;
    uint64 page, snapBitsPages;
    uint64 blockSize = params->blockSize;
    uint64 block, blocksPerPage = blockSize * BITS_PER_BYTE;
    uint64 inum = 0, firstBlock, lastBlock;
    uint32 id;
    int len;
    fs_Inode *snapdir=NULL, *snapBitsInode=NULL;
    FS_SNAP_WATCHDOG_DECL

    if (snapnum == 0) {
        snapdir = I_GET_SHARED(0, FS_SNAPSHOT_DIR_INODE);
        assert(snapdir != NULL);
        len = strlen(name);
        inum = fs_dir_lookupent(context, snapdir, name, len);
        if (inum == 0)
        {
            printf("snap diag: no snap named %s\n", name);
            goto fs_snapDiagBitsFile_exit;
        }
        id = (uint32) (inum - FS_SNAPSHOT_ID_INODE_FIRST);
    }
    else {
        id = snapnum;
    }

    snapBitsInode = I_GET_SHARED(id, FS_SNAPSHOT_BITS_INODE);
    if (snapBitsInode == NULL)
    {
        printf("snap diag: no snap bits for %s inum %#llx id %d\n",
            name, inum, id);
        goto fs_snapDiagBitsFile_exit;
    }

    snapBitsPages = ceilingDiv64(snapBitsInode->dinode.size, blockSize);

    printf("Snap bits (id %d pages %lld)\n",
           id, snapBitsPages);
    printf("==========\n");

    FS_SNAP_WATCHDOG_INIT("diag");

    for (page = 0; page < snapBitsPages; page++) {
        boolean bit;
        uint8 *snapbits;
        fs_Buf *snapBuf = O_GET_SHARED(snapBitsInode, page);

        assert(snapBuf != NULL);
        snapbits = snapBuf->u.data;

        printf("page %lld buf %lld\n", page, fs_blockNumber(snapBuf->block));

        for (block = 0; block < blocksPerPage; ) {

            // yield the CPU if necc to prevent watchdog timeouts
            FS_SNAP_WATCHDOG_CHECK("diag");

            bit = (snapbits[block >> 3] >> (block & 0x7)) & 0x1;
            if (bit == 0) {
                block++;
            } else {
                firstBlock = block;

                while ((block < blocksPerPage)
                       && ((snapbits[block >> 3] >> (block & 0x7)) & 0x1))
                {
                    block++;
                }
                lastBlock = block - 1;
                if (firstBlock == lastBlock)
                    printf("%lld\n", (page * blocksPerPage) + firstBlock);
                else
                    printf("%lld-%lld\n", (page * blocksPerPage) + firstBlock,
                                          (page * blocksPerPage) + lastBlock);
            }
        }
        fs_bufRelease(context, snapBuf);
    }

fs_snapDiagBitsFile_exit:
    if (snapdir) fs_inodeRelease(context, snapdir);
    if (snapBitsInode) fs_inodeRelease(context, snapBitsInode);
}


/**
 * Get global snapshot stats
 */
void fs_snapDiagGetStats(fs_Context *context,
                        fs_SnapStats *stats)
{
    memcpy((void *) stats, (void *) &fs_snapStats, sizeof(fs_snapStats));
    return;
}


//   E X T E R N A L   A P I


/**
 * Create a snapshot in the specified context's file system, with the
 * specified name.
 */
int32 fs_snapCreate(
    fs_Context *context,
    char *name)
{
    int32 rc;
    uint64 inum;
    int id, len = strlen(name);
    fs_Time now = fs_getTimeOfDay();
    fs_Inode *snapdir;
    fs_SnapInfo *snapInfo = &context->fileSystem->snapInfo;

    // verify filesystem state and snap version

    if (context->fileSystem->state != FS_VOL_STATE_ONLINE) {
        return FS_ERR_NOT_MOUNTED;
    }

    if (FS_SNAP_CHECK_VERSION(context) == FALSE) {
        assert(FALSE);
    }
    
    // disallow user creates with names that are generated automatically

    if ((context->reqHdr->requestType == FS_REQ_SNAP_CREATE) &&
        (FS_SNAP_RESERVE_NAME(name))) {
        return FS_ERR_SNAP_INVAL_NAME;
    }

    // disallow user creates with invalid names

    if (fs_nameIsValid(name, len) != NFX_OK) {
        return FS_ERR_SNAP_INVAL_NAME;
    }

#ifdef FS_SNAP_EEK
    fs_syncThread(context);
    fs_syncThread(context);
    printf("\nEEK before snap create:\n");
    fs_eek(context);
#endif

    FS_LOCK_ORDER_DISABLE(context);

    // only one snapshot operation is allowed
    // because of the snap state info in the superblock
    // since we do the FS_SNAP_PHASE_CREATE_ALLOC outside
    // the freeze, guarantee only one snapshot operation
    // by acquiring the snap lock

    fs_lockExclusive(context, &snapInfo->lock, NULL);

    // check if name already exists

    snapdir = I_GET_EXCL(0, FS_SNAPSHOT_DIR_INODE);
    assert(snapdir);

    inum = fs_dir_lookupent(context, snapdir, name, len);
    if (inum != 0) {
        fs_inodeReleaseExclusive(context, snapdir);
        fs_unlockExclusive(context, &snapInfo->lock);
        return FS_ERR_SNAP_EXISTS;
    }

    // assign next snap id to use

    id = fs_snapAllocateId(context);
    if (id < 0) {
        fs_inodeReleaseExclusive(context, snapdir);
        fs_unlockExclusive(context, &snapInfo->lock);
        return FS_ERR_SNAP_MAX;
    }

    // create a transaction context to add entry to snapdir

    assert(context->logSummary == NULL);
    fs_logBeginTransaction(context);

    inum = id + FS_SNAPSHOT_ID_INODE_FIRST;
    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap create assign id: %d name %s len %d ino %llx",
             id, name, len, inum));

    fs_dir_addent(context, snapdir, name, len, inum);
    snapdir->dinode.atime = now;
    snapdir->dinode.mtime = now;

    fs_inodeWrite(context, snapdir, INO_WRITE_LOG);

    // commit and release the transaction context

    fs_logCommitEntry(context);
    context->logSummary = NULL;

    // update the superblock with the current phase

    fs_snapSetState(context, id, FS_SNAP_PHASE_CREATE_ALLOC);

    // create snapbits and prefetch before freezing the file system
    // and continue with the FS_SNAP_PHASE_CREATE_SNAP_BITS phase

    fs_snapCreateAllocSnapBits(context, id);

#ifdef FS_SNAP_READAHEAD
    //  prefetch refbyte pages before freeze
    fs_snapDoReadAhead(context, id);
#endif

    fs_freezeAll(context);

    // log data should be sync'd out after freeze's syncThread calls

    if (context->fileSystem->logRoot->head != context->fileSystem->logRoot->tail) {
        printf("snap create: log not sync'd: head %d tail %d\n",
                context->fileSystem->logRoot->head,
                context->fileSystem->logRoot->tail);
    }

    fs_snapPrintStatus(context, LOG_INFO, "snap create %s: begin", name);

    // finish the snapshot creation process

    rc = fs_snapDoIt(context, id, FS_SNAP_PHASE_CREATE_SNAP_BITS);

    // Calculate the min free blocks for current conditions

    fs_recalcMinFreeBlks(context);

#ifdef FS_SNAP_EEK
    printf("\nEEK after snap create:\n");
    fs_eek(context);
#endif

    fs_thawAll(context);
    fs_unlockExclusive(context, &snapInfo->lock);

    fs_snapStats.snapCreates++;

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap create inode blocks: inodesUsed %lld blksUsed %lld",
             context->fileSystem->super->inodesUsed,
             context->fileSystem->super->blksUsed));
    fs_snapPrintStatus(context, LOG_INFO, "snap create %s: end", name);

    return rc;
}


/**
 * Remove the snapshot with the specified name in the specified context's
 * file system.
 */
int32 fs_snapRemove(
    fs_Context *context,
    char *name)
{
    fs_Inode *snapdir;
    fs_Inumber inum;
    int rc, snapnum;
    fs_SnapInfo *snapInfo = &context->fileSystem->snapInfo;

    // verify filesystem state and snap version

    if (!context->eek.inUse &&
        (context->fileSystem->state != FS_VOL_STATE_ONLINE)) {
        return FS_ERR_NOT_MOUNTED;
    }

    if (FS_SNAP_CHECK_VERSION(context) == FALSE) {
        assert(FALSE);
    }

    if (fs_snapIsPin(context, name)) {
        fs_snapPrintStatus(context, LOG_INFO,
                           "snap remove %s: pinned", name);
        return FS_ERR_SNAP_PIN;
    }

    // only one snapshot operation is allowed
    // because of the snap state info in the superblock
    // since we do the FS_SNAP_PHASE_CREATE_ALLOC outside
    // the freeze, guarantee only one snapshot operation
    // by acquiring the snap lock

    fs_lockExclusive(context, &snapInfo->lock, NULL);

    snapdir = I_GET_EXCL(0, FS_SNAPSHOT_DIR_INODE);
    inum = fs_dir_lookupent(context, snapdir, name, strlen(name));
    if (inum == 0) {
        fs_inodeReleaseExclusive(context, snapdir);
        fs_unlockExclusive(context, &snapInfo->lock);
        return FS_ERR_SNAP_NO_EXIST;
    } else {
        snapnum = inum - FS_SNAPSHOT_ID_INODE_FIRST;
        assert(snapnum > 0);
        assert(snapnum < FS_MAX_SNAPSHOTS);
    }

#ifdef FS_SNAP_READAHEAD
    //  prefetch refbyte and snapbits pages before freeze
    fs_snapDoReadAhead(context, snapnum);
#endif

    // quiesce the file system

    fs_freezeAll(context);

#ifdef FS_SNAP_EEK
    printf("\nEEK before snap remove:\n");
    fs_eek(context);
#endif

    FS_LOCK_ORDER_DISABLE(context);

    fs_snapPrintStatus(context, LOG_INFO, "snap remove %s: begin", name);

    // check if an application (eg ndmp) is still accessing this snapshot

    fs_lockExclusive(context, &context->fileSystem->snapInfo.accessLocks[snapnum], NULL);

    FS_TRACE(context, FS_TRACE_SNAP,
            ("snap remove: lock acquired for %s snapnum %d",
            name, snapnum));

    // create a transaction context

    assert(context->logSummary == NULL);
    fs_logBeginTransaction(context);

    // remove snap from snapdir

    fs_dir_removeent(context, snapdir, name, strlen(name));
    snapdir->dinode.mtime = fs_getTimeOfDay();
    snapdir->dinode.atime = snapdir->dinode.mtime;
    fs_inodeWrite(context, snapdir, INO_WRITE_LOG);

    // commit and release the transaction context

    fs_logCommitEntry(context);
    fs_syncThread(context);
    context->logSummary = NULL;

    // deallocate snapnum and write back superblock

    fs_snapDeallocateId(context, snapnum);
    fs_writeSuperblocksSync(context, context->fileSystem->super);

    // update the superblock with current phase

    fs_snapSetState(context, snapnum, FS_SNAP_PHASE_REMOVE_REF_COUNTS);

    rc = fs_snapDoIt(context, snapnum, FS_SNAP_PHASE_REMOVE_REF_COUNTS);

    // release snap lock

    fs_unlockExclusive(context, &context->fileSystem->snapInfo.accessLocks[snapnum]);

    // Calculate the min free blocks for current conditions

    fs_recalcMinFreeBlks(context);

#ifdef FS_SNAP_EEK
    printf("\nEEK after snap remove:\n");
    fs_eek(context);
#endif

    fs_thawAll(context);
    fs_unlockExclusive(context, &snapInfo->lock);
    fs_snapStats.snapRemoves++;

    FS_TRACE(context, FS_TRACE_SNAP,
            ("snap remove: lock released for %s snapnum %d",
            name, snapnum));
    fs_snapPrintStatus(context, LOG_INFO, "snap remove %s: end", name);

    return NFX_OK;
}


/**
 * Remove a snapshot in the specified context's file system, with the
 * specified snapnum.  Called by Eek to remove a corrupted snapshot.
 */
int32 fs_snapRemoveSnapnum(
    fs_Context *context,
    int32 snapnum)
{
    int32 i, eof, entries;
    fs_Inode *dir;
    uint64 cookie = 0;
    int32 snap = 0;
    char namebuf[FS_READDIR_SPACE_MAX];
    fs_DirEntry *dirent;

    dir = I_GET_SHARED(0, FS_SNAPSHOT_DIR_INODE);
    assert(dir);

    dirent = (void *) namebuf;
    fs_dir_readdir(context, dir, dirent, sizeof(namebuf), cookie, FALSE,
                   &entries, &eof);

    for (i = 0; i < entries; i++) {
        uint64 inum;

        if (!fs_nameIsDotOrDotDot(dirent->name)) {
            inum = dirent->inode;
            assert((inum >= FS_SNAPSHOT_ID_INODE_FIRST) && (inum <= FS_SNAPSHOT_ID_INODE_LAST));
            snap = inum - FS_SNAPSHOT_ID_INODE_FIRST;
            assert((snap > 0) && (snap < FS_MAX_SNAPSHOTS));
            if (snap == snapnum) {
                break;
            }
        }

        // get the next dir entry

        cookie = dirent->cookie;
        FS_READDIR_ADV_ENTRY_PTR(dirent);
    }

    fs_inodeRelease(context, dir);

    if (snap == snapnum) {
        return fs_snapRemove(context, dirent->name);
    }
    return NFX_ERR;
}


/**
 * Resume a snaphot create during the FS_SNAP_PHASE_CREATE_REF_COUNTS crash phase
 */
static int32 fs_snapResumeCreateRefCnts(
        fs_Context *context, int id)
{
    int32 rc;

    fs_snapRecalculateRefCnts(context);

    // ref counters already updated
    // continue with the rest of create phases

    rc = fs_snapDoIt(context, id, FS_SNAP_PHASE_CREATE_META);

    return rc;
}


/**
 * Resume a snaphot remove during the FS_SNAP_PHASE_REMOVE_REF_COUNTS crash phase
 */
static int32 fs_snapResumeRemoveRefCnts(
        fs_Context *context, int id)
{
    int32 rc;

    fs_snapRecalculateRefCnts(context);

    // continue with the remove phases
    // start with decrementing the ref counters

    rc = fs_snapDoIt(context, id, FS_SNAP_PHASE_REMOVE_REF_COUNTS);

    return rc;
}


/**
 * Resume a snapshot creation, removal or revert operation,
 * based on the snapshot state from the superblock.
 */
int32 fs_snapResume(
    fs_Context *context,
    fs_SnapState state)
{
    int32 rc;

    FS_TRACE(context, FS_TRACE_SNAP,
             ("snap resume: %s %d", fs_snapPhaseName[state.phase], state.id));

    fs_freezeAll(context);

    FS_LOCK_ORDER_DISABLE(context);

    switch (state.phase) {
        case FS_SNAP_PHASE_CREATE_ALLOC:
        case FS_SNAP_PHASE_CREATE_SNAP_BITS:
        case FS_SNAP_PHASE_CREATE_META:
            rc = fs_snapDoIt(context, state.id, state.phase);
            break;
        case FS_SNAP_PHASE_CREATE_REF_COUNTS:
            // crash during reference count increment
            // need to recalculate from all snapbits before continuing
            rc = fs_snapResumeCreateRefCnts(context, state.id);
            break;
        case FS_SNAP_PHASE_REMOVE_REF_COUNTS:
            // crash during reference count decrement
            // need to recalculate from all snapbits before continuing
            rc = fs_snapResumeRemoveRefCnts(context, state.id);
            break;
        case FS_SNAP_PHASE_REMOVE_SNAP_BITS:
#ifdef FS_SNAP_OPTIMIZE_REMOVE_FREE_BLOCKS
        case FS_SNAP_PHASE_REMOVE_SNAP_BITS_AND_FREE:
#else
        case FS_SNAP_PHASE_REMOVE_FREE_BLOCKS:
#endif
        case FS_SNAP_PHASE_REMOVE_FREE_SNAP_BITS:
            rc = fs_snapDoIt(context, state.id, state.phase);
            break;
        case FS_SNAP_PHASE_REVERT_BUDDY:

#ifndef FS_SNAP_CACHE_OPTIMIZED
            // need to flush buffer cache since log replay might've
            // brought into the cache the old buddy pages
            // the first thing revert buddy will do is to clear
            // the buddy file without going through the buffer cache
            // hence subsequent buddy adds will be no ops and
            // the buddy will remain cleared ....

            fs_bufCacheFlush(context, NULL);
#endif
            // fall through to continue the revert

        case FS_SNAP_PHASE_REVERT_REMOVE_SNAPS:
        case FS_SNAP_PHASE_REVERT_REF_COUNTS:
        case FS_SNAP_PHASE_REVERT_INODE_BLOCKS:
        case FS_SNAP_PHASE_REVERT_META:
        case FS_SNAP_PHASE_REVERT_CLEAR_SNAP_BITS:
            rc = fs_snapDoIt(context, state.id, state.phase);
            break;
        default:
            rc = NFX_ERR;
    }

    // Calculate the min free blocks for current
    // conditions.
    fs_recalcMinFreeBlks(context);

    fs_thawAll(context);

    return rc;
}


/**
 * Revert back to a snapshot with snapnum
 *
 * Assumptions (relevant only for agile mirror):
 *
 *  1. the snapshot supernodes are on disk
 *  2. all blocks used by the snapshot are on disk
 *  3. same superblock version for both live and snapshot
 *  4. volume big enough to revert
 *     (agile mirror will grow vol and re adjust fs using fs_growfs()? if necc)
 *
 * if NFX_OK, the snapshot's view becomes the live file system
 * if NFX_ERR, the snapnum is not valid
 */
int32 fs_snapRevert(
        fs_Context *context,
        int32 snapnum)
{
    int32 rc;
    fs_Inode *root;
    fs_Time  now = fs_getTimeOfDay();
    fs_SnapInfo *snapInfo = &context->fileSystem->snapInfo;

    // only one snapshot operation is allowed
    // because of the snap state info in the superblock
    // since we do the FS_SNAP_PHASE_CREATE_ALLOC outside
    // the freeze, guarantee only one snapshot operation
    // by acquiring the snap lock

    fs_lockExclusive(context, &snapInfo->lock, NULL);

    // quiesce the file system

    fs_freezeAll(context);
    FS_LOCK_ORDER_DISABLE(context);

    // sync and flush the caches

    fs_syncThread(context);
    fs_syncThread(context);
    fs_syncThread(context);
    fs_inodeCacheFlush(context);
    fs_bufCacheFlush(context, NULL);

    // lookup the root snode for snapnum to make sure snap exists

    root = I_GET_SHARED(snapnum, FS_ROOT_INODE);
    if (!root) {
        fs_thawAll(context);
        fs_unlockExclusive(context, &snapInfo->lock);
        return FS_ERR_SNAP_NO_EXIST;
    }
    fs_inodeRelease(context, root);

    fs_snapPrintStatus(context, LOG_INFO, "snap revert %d: begin", snapnum);

    // update the superblock with current phase and snapnum

    fs_snapSetState(context, snapnum, FS_SNAP_PHASE_REVERT_REMOVE_SNAPS);

    // start the snap revert phases

    rc = fs_snapDoIt(context, snapnum, FS_SNAP_PHASE_REVERT_REMOVE_SNAPS);

    fs_snapPrintStatus(context, LOG_INFO, "snap revert %d: end", snapnum);

    fs_logStopDaemon(context, FALSE);
    context->fileSystem->super->logInitTimeSec = now.secs;
    context->fileSystem->super->logInitTimeNsec = now.nsecs;
    fs_logInitDevice(context);

    // Calculate the min free blocks for current
    // conditions.
    fs_recalcMinFreeBlks(context);

    fs_thawAll(context);
    fs_unlockExclusive(context, &snapInfo->lock);

    fs_snapStats.snapReverts++;
    return rc;
}
#endif // !SSC
